# Copyright (c) 2020, Susi Lehtola
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
# * Neither the name of the <organization> nor the
# names of its contributors may be used to endorse or promote products
# derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Routines for the implementation of the superposition of atomic
# potentials guess for electronic structure calculations, see
#
# S. Lehtola, "Assessment of Initial Guesses for Self-Consistent Field
# Calculations. Superposition of Atomic Potentials: Simple yet
# Efficient", J. Chem. Theory Comput. 15, 1593 (2019).
# DOI: 10.1021/acs.jctc.8b01089
#
# This function evaluates the effective charge of a neutral atom,
# given by exchange-only LDA on top of spherically symmetric
# unrestricted Hartree-Fock calculations as described in
#
# S. Lehtola, L. Visscher, E. Engel, Efficient implementation of the
# superposition of atomic potentials initial guess for electronic
# structure calculations in Gaussian basis sets, J. Chem. Phys., in
# press (2020).
#
# The potentials have been calculated for the ground-states of
# spherically symmetric atoms at the non-relativistic level of theory
# as described in
#
# S. Lehtola, "Fully numerical calculations on atoms with fractional
# occupations and range-separated exchange functionals", Phys. Rev. A
# 101, 012516 (2020). DOI: 10.1103/PhysRevA.101.012516
#
# using accurate finite-element calculations as described in
#
# S. Lehtola, "Fully numerical Hartree-Fock and density functional
# calculations. I. Atoms", Int. J. Quantum Chem. e25945 (2019).
# DOI: 10.1002/qua.25945

import numpy
sap_Zeff = numpy.asarray([
[ 0.00000000000000e+00, 1.90952910601727e-10, 6.09559478066679e-09, 4.61004660175235e-08, 1.93163410435251e-07, 5.85185417059306e-07, 1.44314740634262e-06, 3.08636764846659e-06, 5.94431823161112e-06, 1.05645380400433e-05, 1.76162923183837e-05, 2.78897519760029e-05, 4.22905959185128e-05, 6.18300732778089e-05, 8.76106957450797e-05, 1.20807859590111e-04, 1.62647818777058e-04, 2.14382541464425e-04, 2.77262079003254e-04, 3.52505156601008e-04, 4.41268755831824e-04, 5.44617499397453e-04, 6.63493666792960e-04, 7.98688665221507e-04, 9.50816753262807e-04, 1.12029176607439e-03, 1.30730752153578e-03, 1.51182249854004e-03, 1.73354927391704e-03, 1.97194908601520e-03, 2.22623176392513e-03, 2.49536112514821e-03, 2.77806580483838e-03, 3.07285534032323e-03, 3.37804119918327e-03, 3.69176231137635e-03, 4.01201454917522e-03, 4.33668349618189e-03, 4.66357976115339e-03, 4.99047602612490e-03, 5.31514497313156e-03, 5.63539721093044e-03, 5.94911832312352e-03, 6.25430418198355e-03, 6.54909371746840e-03, 6.83179839715858e-03, 7.10092775838166e-03, 7.35521043629159e-03, 7.59361024838974e-03, 7.81533702376675e-03, 8.01985200077100e-03, 8.20686775623240e-03, 8.37634276904398e-03, 8.52847085708528e-03, 8.66366585551382e-03, 8.78254202290933e-03, 8.88589076647496e-03, 8.97465436570578e-03, 9.04989744330353e-03, 9.11277698084236e-03, 9.16451170352972e-03, 9.20635166271667e-03, 9.23954882656170e-03, 9.26532944902898e-03, 9.28486892638827e-03, 9.29926977033078e-03, 9.30954322998840e-03, 9.31659498426674e-03, 9.32121520407517e-03, 9.32407315465832e-03, 9.32571637490044e-03, 9.32657433688972e-03, 9.32696635889635e-03, 9.32711342184076e-03, 9.32715342671200e-03, 9.32715933135387e-03, 9.32716010591841e-03, 9.32717815234475e-03, 9.32730041970285e-03, 9.32774988990819e-03, 9.32894803137414e-03, 9.33157023068491e-03, 9.33659242402233e-03, 9.34532721197674e-03, 9.35944804407490e-03, 9.38100040354926e-03, 9.41239929903636e-03, 9.45641276761913e-03, 9.51613150190375e-03, 9.59492512132892e-03, 9.69638600332746e-03, 9.82426196230115e-03, 9.98237940324620e-03, 1.01745588727976e-02, 1.04045251751296e-02, 1.06758144066244e-02, 1.09916803861568e-02, 1.13550030136215e-02, 1.17682010761543e-02, 1.22331519394712e-02, 1.27511204128277e-02, 1.33226988640867e-02, 1.39477603918012e-02, 1.46254265411583e-02, 1.53540506885909e-02, 1.61312178254622e-02, 1.69537610550243e-02, 1.78177946899570e-02, 1.87187634116783e-02, 1.96515065387106e-02, 2.06103360608126e-02, 2.15891267388584e-02, 2.25814162571620e-02, 2.35805131535374e-02, 2.45796100499127e-02, 2.55718995682163e-02, 2.65506902462622e-02, 2.75095197683641e-02, 2.84422628953965e-02, 2.93432316171178e-02, 3.02072652520504e-02, 3.10298084816126e-02, 3.18069756184839e-02, 3.25355997659164e-02, 3.32132659152736e-02, 3.38383274429880e-02, 3.44099058942470e-02, 3.49278743676035e-02, 3.53928252309205e-02, 3.58060232934533e-02, 3.61693459209180e-02, 3.64852119004504e-02, 3.67565011319451e-02, 3.69864674342771e-02, 3.71786469038285e-02, 3.73367643447736e-02, 3.74646403037473e-02, 3.75661011857458e-02, 3.76448948051710e-02, 3.77046135394556e-02, 3.77486270080384e-02, 3.77800259035255e-02, 3.78015782629999e-02, 3.78156990950980e-02, 3.78244338830524e-02, 3.78294560763898e-02, 3.78320782757006e-02, 3.78332764171666e-02, 3.78337258873719e-02, 3.78338481547300e-02, 3.78338662011563e-02, 3.78338677943125e-02, 3.78338990114269e-02, 3.78341105122543e-02, 3.78348880159300e-02, 3.78369605879999e-02, 3.78414965273755e-02, 3.78501840304259e-02, 3.78652936631901e-02, 3.78897201963173e-02, 3.79270019527330e-02, 3.79813164687220e-02, 3.80574519571201e-02, 3.81607547675479e-02, 3.82970537435376e-02, 3.84725630604126e-02, 3.86937657718628e-02, 3.89672808793569e-02, 3.92997172504388e-02, 3.96975181351890e-02, 4.01668003527027e-02, 4.07131924320912e-02, 4.13416760889992e-02, 4.20564353958445e-02, 4.28607178620895e-02, 4.37567113832465e-02, 4.47454406505750e-02, 4.58266861470961e-02, 4.69989283019068e-02, 4.82593187485059e-02, 4.96036799506086e-02, 5.10265337389549e-02, 5.25211585641778e-02, 5.40796745336850e-02, 5.56931545845249e-02, 5.73517594685833e-02, 5.90448936093834e-02, 6.07613783478282e-02, 6.24896386420494e-02, 6.42178989362706e-02, 6.59343836747154e-02, 6.76275178155154e-02, 6.92861226995739e-02, 7.08996027504137e-02, 7.24581187199210e-02, 7.39527435451439e-02, 7.53755973334902e-02, 7.67199585355928e-02, 7.79803489821920e-02, 7.91525911370027e-02, 8.02338366335238e-02, 8.12225659008523e-02, 8.21185594220093e-02, 8.29228418882542e-02, 8.36376011950996e-02, 8.42660848520075e-02, 8.48124769313960e-02, 8.52817591489097e-02, 8.56795600336600e-02, 8.60119964047418e-02, 8.62855115122360e-02, 8.65067142236862e-02, 8.66822235405611e-02, 8.68185225165509e-02, 8.69218253269787e-02, 8.69979608153767e-02, 8.70522753313658e-02, 8.70895570877814e-02, 8.71139836209086e-02, 8.71290932536728e-02, 8.71377807567233e-02, 8.71423166960989e-02, 8.71443892681687e-02, 8.71451667718444e-02, 8.71453782726719e-02, 8.71454094897863e-02, 8.71454119937853e-02, 8.71454582052761e-02, 8.71457712953469e-02, 8.71469222538740e-02, 8.71499903351853e-02, 8.71567050019657e-02, 8.71695653343703e-02, 8.71919325092340e-02, 8.72280917294016e-02, 8.72832808645706e-02, 8.73636840282826e-02, 8.74763893341569e-02, 8.76293111199353e-02, 8.78310779714063e-02, 8.80908888908394e-02, 8.84183409080102e-02, 8.88232322996593e-02, 8.93153463410200e-02, 8.99042211395692e-02, 9.05989115786682e-02, 9.14077497135578e-02, 9.23381101050022e-02, 9.33961865421474e-02, 9.45867863961021e-02, 9.59131484644118e-02, 9.73767896236902e-02, 9.89773849173530e-02, 1.00712684885825e-01, 1.02578473019505e-01, 1.04568565204852e-01, 1.06674851968149e-01, 1.08887383228386e-01, 1.11194494179524e-01, 1.13582969862523e-01, 1.16038244987378e-01, 1.18544634651912e-01, 1.21085590801869e-01, 1.23643978607465e-01, 1.26202366413060e-01, 1.28743322563018e-01, 1.31249712227552e-01, 1.33704987352407e-01, 1.36093463035406e-01, 1.38400573986544e-01, 1.40613105246780e-01, 1.42719392010078e-01, 1.44709484195425e-01, 1.46575272329105e-01, 1.48310572297577e-01, 1.49911167591240e-01, 1.51374808750518e-01, 1.52701170818828e-01, 1.53891770672782e-01, 1.54949847109928e-01, 1.55880207501372e-01, 1.56689045636262e-01, 1.57383736075361e-01, 1.57972610873910e-01, 1.58464724915270e-01, 1.58869616306920e-01, 1.59197068324090e-01, 1.59456879243523e-01, 1.59658646094994e-01, 1.59811567880773e-01, 1.59924273186647e-01, 1.60004676350359e-01, 1.60059865485528e-01, 1.60096024705696e-01, 1.60118391880560e-01, 1.60131252212964e-01, 1.60137966879745e-01, 1.60141034961056e-01, 1.60142185919583e-01, 1.60142499009654e-01, 1.60142545221145e-01, 1.60142548785420e-01, 1.60142612788376e-01, 1.60143046418402e-01, 1.60144640496984e-01, 1.60148889792324e-01, 1.60158189611776e-01, 1.60176001183575e-01, 1.60206979740620e-01, 1.60257060292657e-01, 1.60333497293867e-01, 1.60444855746995e-01, 1.60600952695698e-01, 1.60812749504776e-01, 1.61092196773205e-01, 1.61452035127297e-01, 1.61905556461813e-01, 1.62466331398736e-01, 1.63147909782953e-01, 1.63963501901786e-01, 1.64925648776731e-01, 1.66045890311693e-01, 1.67334440279880e-01, 1.68799877084757e-01, 1.70448858939578e-01, 1.72285871581820e-01, 1.74313015886939e-01, 1.76529841789778e-01, 1.78933233786826e-01, 1.81517352008545e-01, 1.84273631452188e-01, 1.87190840489457e-01, 1.90255198249318e-01, 1.93450548965050e-01, 1.96758589906653e-01, 2.00159148134535e-01, 2.03630500045255e-01, 2.07149726568976e-01, 2.10693095951234e-01, 2.14236465333492e-01, 2.17755691857213e-01, 2.21227043767933e-01, 2.24627601995815e-01, 2.27935642937417e-01, 2.31130993653150e-01, 2.34195351413011e-01, 2.37112560450280e-01, 2.39868839893923e-01, 2.42452958115642e-01, 2.44856350112690e-01, 2.47073176015529e-01, 2.49100320320648e-01, 2.50937332962890e-01, 2.52586314817711e-01, 2.54051751622588e-01, 2.55340301590775e-01, 2.56460543125737e-01, 2.57422690000682e-01, 2.58238282119515e-01, 2.58919860503732e-01, 2.59480635440655e-01, 2.59934156775171e-01, 2.60293995129263e-01, 2.60573442397692e-01, 2.60785239206770e-01, 2.60941336155473e-01, 2.61052694608601e-01, 2.61129131609811e-01, 2.61179212161848e-01, 2.61210190718893e-01, 2.61228002290692e-01, 2.61237302110144e-01, 2.61241551405484e-01, 2.61243145484066e-01, 2.61243579114092e-01, 2.61243643117048e-01, 2.61243647963159e-01, 2.61243733811555e-01, 2.61244315447847e-01, 2.61246453616100e-01, 2.61252153277648e-01, 2.61264627304301e-01, 2.61288518308045e-01, 2.61330070442282e-01, 2.61397244449703e-01, 2.61499770869196e-01, 2.61649138103636e-01, 2.61858513942445e-01, 2.62142601075000e-01, 2.62517429069501e-01, 2.63000087173004e-01, 2.63608404059534e-01, 2.64360582265302e-01, 2.65274796457795e-01, 2.66368765849416e-01, 2.67659311953440e-01, 2.69161913464872e-01, 2.70890270314110e-01, 2.72855888878673e-01, 2.75067699948026e-01, 2.77531720328100e-01, 2.80250767963542e-01, 2.83224239173301e-01, 2.86447955072614e-01, 2.89914082532175e-01, 2.93611133149108e-01, 2.97524041724410e-01, 3.01634323710778e-01, 3.05920309067656e-01, 3.10357447991359e-01, 3.14918682130121e-01, 3.19574873196930e-01, 3.24295279402716e-01, 3.29048068888888e-01, 3.33800858375061e-01, 3.38521264580847e-01, 3.43177455647656e-01, 3.47738689786418e-01, 3.52175828710121e-01, 3.56461814066999e-01, 3.60572096053367e-01, 3.64485004628669e-01, 3.68182055245602e-01, 3.71648182705163e-01, 3.74871898604476e-01, 3.77845369814235e-01, 3.80564417449677e-01, 3.83028437829751e-01, 3.85240248899104e-01, 3.87205867463667e-01, 3.88934224312905e-01, 3.90436825824337e-01, 3.91727371928361e-01, 3.92821341319982e-01, 3.93735555512475e-01, 3.94487733718243e-01, 3.95096050604772e-01, 3.95578708708276e-01, 3.95953536702777e-01, 3.96237623835332e-01, 3.96446999674141e-01, 3.96596366908581e-01, 3.96698893328074e-01, 3.96766067335495e-01, 3.96807619469732e-01, 3.96831510473476e-01, 3.96843984500129e-01, 3.96849684161676e-01, 3.96851822329930e-01, 3.96852403966222e-01, 3.96852489814618e-01, 3.96852496259285e-01, 3.96852609692754e-01, 3.96853378222230e-01, 3.96856203433365e-01, 3.96863734527908e-01, 3.96880216747647e-01, 3.96911784483105e-01, 3.96966688278498e-01, 3.97055446847149e-01, 3.97190917374997e-01, 3.97388279754301e-01, 3.97664932889584e-01, 3.98040303784153e-01, 3.98535572676975e-01, 3.99173319985182e-01, 3.99977103147848e-01, 4.00970973596802e-01, 4.02178945940278e-01, 4.03624432983181e-01, 4.05329661379811e-01, 4.07315083487650e-01, 4.09598801341387e-01, 4.12196018583595e-01, 4.15118535672820e-01, 4.18374302753806e-01, 4.21967043241952e-01, 4.25895959479530e-01, 4.30155529809514e-01, 4.34735404137097e-01, 4.39620402570021e-01, 4.44790619112629e-01, 4.50221629705321e-01, 4.55884801222635e-01, 4.61747695441524e-01, 4.67774559536377e-01, 4.73926892415072e-01, 4.80164074241151e-01, 4.86444044844104e-01, 4.92724015447057e-01, 4.98961197273136e-01, 5.05113530151831e-01, 5.11140394246684e-01, 5.17003288465573e-01, 5.22666459982887e-01, 5.28097470575579e-01, 5.33267687118187e-01, 5.38152685551111e-01, 5.42732559878694e-01, 5.46992130208678e-01, 5.50921046446256e-01, 5.54513786934401e-01, 5.57769554015388e-01, 5.60692071104612e-01, 5.63289288346821e-01, 5.65573006200558e-01, 5.67558428308397e-01, 5.69263656705027e-01, 5.70709143747930e-01, 5.71917116091406e-01, 5.72910986540360e-01, 5.73714769703026e-01, 5.74352517011232e-01, 5.74847785904055e-01, 5.75223156798624e-01, 5.75499809933907e-01, 5.75697172313211e-01, 5.75832642841059e-01, 5.75921401409710e-01, 5.75976305205103e-01, 5.76007872940561e-01, 5.76024355160300e-01, 5.76031886254843e-01, 5.76034711465978e-01, 5.76035479995454e-01, 5.76035593428923e-01, 5.76035601918516e-01, 5.76035751000057e-01, 5.76036761050791e-01, 5.76040474124483e-01, 5.76050371972002e-01, 5.76072033962409e-01, 5.76113522305210e-01, 5.76185680394834e-01, 5.76302332599034e-01, 5.76480376656276e-01, 5.76739762954657e-01, 5.77103358250526e-01, 5.77596694757745e-01, 5.78247608904937e-01, 5.79085777324680e-01, 5.80142160714465e-01, 5.81448369008738e-01, 5.83035963746022e-01, 5.84935715536299e-01, 5.87176836074347e-01, 5.89786205160249e-01, 5.92787613649124e-01, 5.96201043143252e-01, 6.00042002562177e-01, 6.04320940496073e-01, 6.09042750496286e-01, 6.14206384229870e-01, 6.19804584780979e-01, 6.25823749391131e-01, 6.32243927672227e-01, 6.39038957887938e-01, 6.46176740372493e-01, 6.53619643635784e-01, 6.61325035284385e-01, 6.69245926661566e-01, 6.77331717162472e-01, 6.85529021592532e-01, 6.93782561777778e-01, 7.02036101963024e-01, 7.10233406393084e-01, 7.18319196893990e-01, 7.26240088271171e-01, 7.33945479919772e-01, 7.41388383183063e-01, 7.48526165667618e-01, 7.55321195883329e-01, 7.61741374164425e-01, 7.67760538774577e-01, 7.73358739325686e-01, 7.78522373059270e-01, 7.83244183059483e-01, 7.87523120993379e-01, 7.91364080412304e-01, 7.94777509906432e-01, 7.97778918395307e-01, 8.00388287481209e-01, 8.02629408019257e-01, 8.04529159809534e-01, 8.06116754546818e-01, 8.07422962841091e-01, 8.08479346230876e-01, 8.09317514650619e-01, 8.09968428797812e-01, 8.10461765305029e-01, 8.10825360600899e-01, 8.11084746899280e-01, 8.11262790956522e-01, 8.11379443160722e-01, 8.11451601250346e-01, 8.11493089593147e-01, 8.11514751583554e-01, 8.11524649431073e-01, 8.11528362504765e-01, 8.11529372555499e-01, 8.11529521637040e-01, 8.11529532798784e-01, 8.11529728860443e-01, 8.11531057208815e-01, 8.11535940384711e-01, 8.11548957344360e-01, 8.11577445685336e-01, 8.11632008264029e-01, 8.11726905557525e-01, 8.11880318411393e-01, 8.12114469554298e-01, 8.12455596347074e-01, 8.12933771554950e-01, 8.13582573367217e-01, 8.14438610315931e-01, 8.15540911041212e-01, 8.16930192895915e-01, 8.18648027064120e-01, 8.20735921082956e-01, 8.23234342315431e-01, 8.26181707947852e-01, 8.29613368421049e-01, 8.33560611810586e-01, 8.38049716528020e-01, 8.43101078824098e-01, 8.48728439956853e-01, 8.54938235584165e-01, 8.61729087011563e-01, 8.69091450448790e-01, 8.77007436495350e-01, 8.85450807790408e-01, 8.94387158240569e-01, 9.03774272601226e-01, 9.13562660557712e-01, 9.23696254955634e-01, 9.34113259586532e-01, 9.44747128059346e-01, 9.55527651884583e-01, 9.66382133058090e-01, 9.77236614231597e-01, 9.88017138056833e-01, 9.98651006529648e-01, 1.00906801116055e+00, 1.01920160555847e+00, 1.02898999351495e+00, 1.03837710787561e+00, 1.04731345832577e+00, 1.05575682962083e+00, 1.06367281566739e+00, 1.07103517910462e+00, 1.07782603053201e+00, 1.08403582615933e+00, 1.08966318729208e+00, 1.09471454958816e+00, 1.09920365430559e+00, 1.10315089769513e+00, 1.10658255816833e+00, 1.10952992380075e+00, 1.11202834503322e+00, 1.11411623905206e+00, 1.11583407322026e+00, 1.11722335507497e+00, 1.11832565580025e+00, 1.11918169274896e+00, 1.11983049456123e+00, 1.12030866976911e+00, 1.12064979656188e+00, 1.12088394770479e+00, 1.12103736055865e+00, 1.12113225785215e+00, 1.12118682043084e+00, 1.12121530877182e+00, 1.12122832573147e+00, 1.12123320890736e+00, 1.12123453725574e+00, 1.12123473331740e+00, 1.12123474803548e+00, 1.12123500708624e+00, 1.12123676219562e+00, 1.12124321419870e+00, 1.12126041314168e+00, 1.12129805398602e+00, 1.12137014598792e+00, 1.12149553111498e+00, 1.12169823120755e+00, 1.12200760853249e+00, 1.12245832977734e+00, 1.12309012924169e+00, 1.12394737284343e+00, 1.12507843040717e+00, 1.12653486937832e+00, 1.12837048845095e+00, 1.13064021446234e+00, 1.13339889015484e+00, 1.13669998391797e+00, 1.14059425530038e+00, 1.14512841184613e+00, 1.15034379361017e+00, 1.15627512151923e+00, 1.16294934456623e+00, 1.17038461868924e+00, 1.17858944714218e+00, 1.18756200829478e+00, 1.19728969220517e+00, 1.20774886211121e+00, 1.21890485132536e+00, 1.23071220004332e+00, 1.24311513044871e+00, 1.25604825237957e+00, 1.26943748588041e+00, 1.28320118135761e+00, 1.29725141293477e+00, 1.31149541610773e+00, 1.32583713704658e+00, 1.34017885798544e+00, 1.35442286115840e+00, 1.36847309273555e+00, 1.38223678821275e+00, 1.39562602171359e+00, 1.40855914364445e+00, 1.42096207404985e+00, 1.43276942276780e+00, 1.44392541198196e+00, 1.45438458188800e+00, 1.46411226579839e+00, 1.47308482695099e+00, 1.48128965540392e+00, 1.48872492952694e+00, 1.49539915257394e+00, 1.50133048048299e+00, 1.50654586224704e+00, 1.51108001879278e+00, 1.51497429017520e+00, 1.51827538393833e+00, 1.52103405963083e+00, 1.52330378564222e+00, 1.52513940471485e+00, 1.52659584368599e+00, 1.52772690124974e+00, 1.52858414485147e+00, 1.52921594431582e+00, 1.52966666556068e+00, 1.52997604288562e+00, 1.53017874297818e+00, 1.53030412810525e+00, 1.53037622010714e+00, 1.53041386095149e+00, 1.53043105989446e+00, 1.53043751189755e+00, 1.53043926700692e+00, 1.53043952605768e+00, 1.53043954558710e+00, 1.53043989042458e+00, 1.53044222675237e+00, 1.53045081538756e+00, 1.53047370989720e+00, 1.53052381580632e+00, 1.53061978164763e+00, 1.53078668905608e+00, 1.53105651489537e+00, 1.53146834498040e+00, 1.53206832614783e+00, 1.53290935102636e+00, 1.53405047766020e+00, 1.53555609392585e+00, 1.53749484423810e+00, 1.53993834315617e+00, 1.54295970697606e+00, 1.54663194005015e+00, 1.55102621725011e+00, 1.55621010755256e+00, 1.56224578607601e+00, 1.56918828296319e+00, 1.57708381725166e+00, 1.58596826230754e+00, 1.59586578655212e+00, 1.60678770915941e+00, 1.61873160525180e+00, 1.63168068900496e+00, 1.64560349615511e+00, 1.66045387986558e+00, 1.67617132595656e+00, 1.69268158534447e+00, 1.70989761339548e+00, 1.72772079798805e+00, 1.74604245061668e+00, 1.76474552805216e+00, 1.78370654608750e+00, 1.80279764190358e+00, 1.82188873771966e+00, 1.84084975575500e+00, 1.85955283319048e+00, 1.87787448581911e+00, 1.89569767041169e+00, 1.91291369846269e+00, 1.92942395785060e+00, 1.94514140394158e+00, 1.95999178765205e+00, 1.97391459480220e+00, 1.98686367855536e+00, 1.99880757464776e+00, 2.00972949725504e+00, 2.01962702149962e+00, 2.02851146655550e+00, 2.03640700084397e+00, 2.04334949773115e+00, 2.04938517625460e+00, 2.05456906655706e+00, 2.05896334375701e+00, 2.06263557683110e+00, 2.06565694065099e+00, 2.06810043956906e+00, 2.07003918988132e+00, 2.07154480614696e+00, 2.07268593278081e+00, 2.07352695765933e+00, 2.07412693882676e+00, 2.07453876891179e+00, 2.07480859475108e+00, 2.07497550215953e+00, 2.07507146800084e+00, 2.07512157390996e+00, 2.07514446841960e+00, 2.07515305705479e+00, 2.07515539338258e+00, 2.07515573822006e+00, 2.07515576435827e+00, 2.07515622776600e+00, 2.07515936742575e+00, 2.07517090921038e+00, 2.07520167585650e+00, 2.07526901037463e+00, 2.07539797348087e+00, 2.07562227097627e+00, 2.07598487477262e+00, 2.07653831010217e+00, 2.07734459110783e+00, 2.07847479722370e+00, 2.08000829323968e+00, 2.08203160640800e+00, 2.08463698410373e+00, 2.08792066511228e+00, 2.09198090631884e+00, 2.09691581417381e+00, 2.10282103659120e+00, 2.10978737572511e+00, 2.11789838522654e+00, 2.12722801701473e+00, 2.13783838225931e+00, 2.14977768916297e+00, 2.16307841631015e+00, 2.17775577490330e+00, 2.19380650628557e+00, 2.21120805293004e+00, 2.22991813177913e+00, 2.24987472868994e+00, 2.27099652205367e+00, 2.29318373269557e+00, 2.31631938621923e+00, 2.34027096333100e+00, 2.36489240365051e+00, 2.39002641935330e+00, 2.41550706694645e+00, 2.44116251876607e+00, 2.46681797058569e+00, 2.49229861817885e+00, 2.51743263388163e+00, 2.54205407420114e+00, 2.56600565131291e+00, 2.58914130483657e+00, 2.61132851547847e+00, 2.63245030884221e+00, 2.65240690575301e+00, 2.67111698460210e+00, 2.68851853124657e+00, 2.70456926262884e+00, 2.71924662122199e+00, 2.73254734836917e+00, 2.74448665527283e+00, 2.75509702051741e+00, 2.76442665230560e+00, 2.77253766180703e+00, 2.77950400094095e+00, 2.78540922335833e+00, 2.79034413121330e+00, 2.79440437241986e+00, 2.79768805342842e+00, 2.80029343112415e+00, 2.80231674429246e+00, 2.80385024030844e+00, 2.80498044642432e+00, 2.80578672742997e+00, 2.80634016275952e+00, 2.80670276655587e+00, 2.80692706405127e+00, 2.80705602715752e+00, 2.80712336167564e+00, 2.80715412832176e+00, 2.80716567010639e+00, 2.80716880976614e+00, 2.80716927317387e+00, 2.80716930852243e+00, 2.80716993816209e+00, 2.80717420406922e+00, 2.80718988608100e+00, 2.80723168922692e+00, 2.80732317773938e+00, 2.80749840202375e+00, 2.80780315870386e+00, 2.80829583442000e+00, 2.80904779606537e+00, 2.81014330327150e+00, 2.81167893282952e+00, 2.81376251897912e+00, 2.81651162771490e+00, 2.82005159705603e+00, 2.82451318821626e+00, 2.83002990443455e+00, 2.83673504455167e+00, 2.84475856695471e+00, 2.85422384601747e+00, 2.86524440745381e+00, 2.87792073094745e+00, 2.89233720796172e+00, 2.90855933977107e+00, 2.92663125556005e+00, 2.94657362303854e+00, 2.96838201461609e+00, 2.99202578101149e+00, 3.01744747154198e+00, 3.04456282657602e+00, 3.07326135311204e+00, 3.10340747955128e+00, 3.13484227086578e+00, 3.16738567092109e+00, 3.20083922508651e+00, 3.23498922381917e+00, 3.26961019697776e+00, 3.30446867950146e+00, 3.33932716202516e+00, 3.37394813518375e+00, 3.40809813391641e+00, 3.44155168808183e+00, 3.47409508813714e+00, 3.50552987945164e+00, 3.53567600589088e+00, 3.56437453242690e+00, 3.59148988746094e+00, 3.61691157799142e+00, 3.64055534438683e+00, 3.66236373596438e+00, 3.68230610344287e+00, 3.70037801923185e+00, 3.71660015104120e+00, 3.73101662805547e+00, 3.74369295154911e+00, 3.75471351298545e+00, 3.76417879204821e+00, 3.77220231445125e+00, 3.77890745456837e+00, 3.78442417078666e+00, 3.78888576194689e+00, 3.79242573128802e+00, 3.79517484002380e+00, 3.79725842617340e+00, 3.79879405573142e+00, 3.79988956293755e+00, 3.80064152458292e+00, 3.80113420029905e+00, 3.80143895697917e+00, 3.80161418126354e+00, 3.80170566977600e+00, 3.80174747292192e+00, 3.80176315493370e+00, 3.80176742084082e+00, 3.80176805048049e+00, 3.80176809884808e+00, 3.80176896483002e+00, 3.80177483199299e+00, 3.80179640042291e+00, 3.80185389484619e+00, 3.80197972459283e+00, 3.80222072126789e+00, 3.80263987176905e+00, 3.80331747881363e+00, 3.80435169765467e+00, 3.80585841571676e+00, 3.80797046096740e+00, 3.81083614443161e+00, 3.81461716181212e+00, 3.81948589815240e+00, 3.82562219734712e+00, 3.83320967456591e+00, 3.84243166385706e+00, 3.85346690493865e+00, 3.86648508213255e+00, 3.88164233429630e+00, 3.89907685728428e+00, 3.91890471983743e+00, 3.94121600986455e+00, 3.96607142093212e+00, 3.99349937860558e+00, 4.02349379334899e+00, 4.05601251133138e+00, 4.09097651711504e+00, 4.12826992327540e+00, 4.16774076202960e+00, 4.20920257346629e+00, 4.25243676452099e+00, 4.29719569297971e+00, 4.34320641205133e+00, 4.39017499393101e+00, 4.43779133574358e+00, 4.48573433871212e+00, 4.53367734168066e+00, 4.58129368349323e+00, 4.62826226537290e+00, 4.67427298444453e+00, 4.71903191290325e+00, 4.76226610395795e+00, 4.80372791539464e+00, 4.84319875414884e+00, 4.88049216030919e+00, 4.91545616609286e+00, 4.94797488407525e+00, 4.97796929881866e+00, 5.00539725649212e+00, 5.03025266755968e+00, 5.05256395758681e+00, 5.07239182013995e+00, 5.08982634312794e+00, 5.10498359529169e+00, 5.11800177248559e+00, 5.12903701356718e+00, 5.13825900285833e+00, 5.14584648007711e+00, 5.15198277927184e+00, 5.15685151561212e+00, 5.16063253299263e+00, 5.16349821645683e+00, 5.16561026170748e+00, 5.16711697976957e+00, 5.16815119861061e+00, 5.16882880565519e+00, 5.16924795615635e+00, 5.16948895283141e+00, 5.16961478257805e+00, 5.16967227700133e+00, 5.16969384543124e+00, 5.16969971259421e+00, 5.16970057857616e+00, 5.16970064560674e+00, 5.16970185234318e+00, 5.16971002817113e+00, 5.16974008354421e+00, 5.16982020139927e+00, 5.16999554378603e+00, 5.17033137003803e+00, 5.17091545171385e+00, 5.17185968977880e+00, 5.17330086251307e+00, 5.17540045778552e+00, 5.17834356992774e+00, 5.18233686874399e+00, 5.18760567544173e+00, 5.19439020670875e+00, 5.20294107306107e+00, 5.21351414024542e+00, 5.22636488226903e+00, 5.24174237098982e+00, 5.25988305966932e+00, 5.28100452611150e+00, 5.30529934474010e+00, 5.33292925608631e+00, 5.36401979667365e+00, 5.39865554232852e+00, 5.43687610376795e+00, 5.47867299528969e+00, 5.52398747598761e+00, 5.57270943870657e+00, 5.62467739557770e+00, 5.67967958114439e+00, 5.73745616554315e+00, 5.79770254171024e+00, 5.86007362290727e+00, 5.92418906074211e+00, 5.98963927000769e+00, 6.05599212571190e+00, 6.12280018019287e+00, 6.18960823467384e+00, 6.25596109037805e+00, 6.32141129964362e+00, 6.38552673747847e+00, 6.44789781867550e+00, 6.50814419484259e+00, 6.56592077924135e+00, 6.62092296480804e+00, 6.67289092167917e+00, 6.72161288439812e+00, 6.76692736509605e+00, 6.80872425661778e+00, 6.84694481805722e+00, 6.88158056371208e+00, 6.91267110429943e+00, 6.94030101564564e+00, 6.96459583427423e+00, 6.98571730071642e+00, 7.00385798939592e+00, 7.01923547811670e+00, 7.03208622014032e+00, 7.04265928732466e+00, 7.05121015367698e+00, 7.05799468494401e+00, 7.06326349164174e+00, 7.06725679045800e+00, 7.07019990260022e+00, 7.07229949787266e+00, 7.07374067060694e+00, 7.07468490867189e+00, 7.07526899034770e+00, 7.07560481659971e+00, 7.07578015898647e+00, 7.07586027684153e+00, 7.07589033221461e+00, 7.07589850804255e+00, 7.07589971477899e+00, 7.07589980894264e+00, 7.07590151393263e+00, 7.07591306550621e+00, 7.07595553054537e+00, 7.07606872853638e+00, 7.07631646864258e+00, 7.07679095534755e+00, 7.07761620050970e+00, 7.07895030826772e+00, 7.08098653175501e+00, 7.08395303612042e+00, 7.08811133992896e+00, 7.09375344558918e+00, 7.10119770795464e+00, 7.11078352760525e+00, 7.12286499049248e+00, 7.13780360764876e+00, 7.15596033662038e+00, 7.17768708939890e+00, 7.20331794924395e+00, 7.23316033040524e+00, 7.26748632002074e+00, 7.30652444022432e+00, 7.35045206074525e+00, 7.39938867821294e+00, 7.45339025834923e+00, 7.51244481176116e+00, 7.57646934380843e+00, 7.64530828481495e+00, 7.71873346963184e+00, 7.79644569623682e+00, 7.87807785272294e+00, 7.96319956177120e+00, 8.05132325259625e+00, 8.14191153345390e+00, 8.23438570409613e+00, 8.32813521796083e+00, 8.42252787918641e+00, 8.51692054041200e+00, 8.61067005427670e+00, 8.70314422491893e+00, 8.79373250577658e+00, 8.88185619660163e+00, 8.96697790564989e+00, 9.04861006213601e+00, 9.12632228874099e+00, 9.19974747355788e+00, 9.26858641456440e+00, 9.33261094661167e+00, 9.39166550002360e+00, 9.44566708015989e+00, 9.49460369762758e+00, 9.53853131814851e+00, 9.57756943835209e+00, 9.61189542796759e+00, 9.64173780912888e+00, 9.66736866897393e+00, 9.68909542175245e+00, 9.70725215072407e+00, 9.72219076788035e+00, 9.73427223076758e+00, 9.74385805041819e+00, 9.75130231278365e+00, 9.75694441844387e+00, 9.76110272225241e+00, 9.76406922661782e+00, 9.76610545010511e+00, 9.76743955786313e+00, 9.76826480302528e+00, 9.76873928973025e+00, 9.76898702983645e+00, 9.76910022782746e+00, 9.76914269286662e+00, 9.76915424444020e+00, 9.76915594943019e+00, 9.76915608360535e+00, 9.76915852757683e+00, 9.76917508586551e+00, 9.76923595621982e+00, 9.76939821679837e+00, 9.76975333310918e+00, 9.77043347315912e+00, 9.77161639837056e+00, 9.77352873881689e+00, 9.77644750794114e+00, 9.78069976286442e+00, 9.78666037025420e+00, 9.79474789301366e+00, 9.80541866824081e+00, 9.81915920045716e+00, 9.83647704453031e+00, 9.85789039860804e+00, 9.88391666745844e+00, 9.91506028974544e+00, 9.95180014802261e+00, 9.99457689687719e+00, 1.00437805522099e+01, 1.00997386828523e+01, 1.01627055346141e+01, 1.02328523966857e+01, 1.03102594916074e+01, 1.03949096335102e+01, 1.04866838559865e+01, 1.05853591619192e+01, 1.06906084941872e+01, 1.08020029697971e+01, 1.09190163621800e+01, 1.10410317586844e+01, 1.11673502642416e+01, 1.12972015692856e+01, 1.14297561517004e+01, 1.15641388401379e+01, 1.16994434306509e+01, 1.18347480211640e+01, 1.19691307096015e+01, 1.21016852920163e+01, 1.22315365970603e+01, 1.23578551026175e+01, 1.24798704991219e+01, 1.25968838915048e+01, 1.27082783671147e+01, 1.28135276993827e+01, 1.29122030053154e+01, 1.30039772277917e+01, 1.30886273696945e+01, 1.31660344646162e+01, 1.32361813266878e+01, 1.32991481784496e+01, 1.33551063090920e+01, 1.34043099644247e+01, 1.34470867132793e+01, 1.34838265715565e+01, 1.35149701938435e+01, 1.35409964626939e+01, 1.35624098167716e+01, 1.35797276608447e+01, 1.35934681930611e+01, 1.36041389682882e+01, 1.36122264910477e+01, 1.36181870984375e+01, 1.36224393533608e+01, 1.36253581224850e+01, 1.36272704629313e+01, 1.36284533881428e+01, 1.36291335281927e+01, 1.36294886445035e+01, 1.36296509050821e+01, 1.36297117754364e+01, 1.36297283337251e+01, 1.36297307776965e+01, 1.36297309717289e+01, 1.36297345276215e+01, 1.36297586193493e+01, 1.36298471835744e+01, 1.36300832670107e+01, 1.36305999487561e+01, 1.36315895286132e+01, 1.36333106432010e+01, 1.36360930313074e+01, 1.36403397377515e+01, 1.36465266191442e+01, 1.36551990933054e+01, 1.36669661545414e+01, 1.36824917572839e+01, 1.37024837485048e+01, 1.37276806026895e+01, 1.37588362799217e+01, 1.37967035859457e+01, 1.38420164612802e+01, 1.38954716632017e+01, 1.39577103286400e+01, 1.40292999170169e+01, 1.41107170294638e+01, 1.42023315846942e+01, 1.43043928024568e+01, 1.44170174037271e+01, 1.45401803836701e+01, 1.46737086503457e+01, 1.48172777507898e+01, 1.49704118283915e+01, 1.51324868734758e+01, 1.53027372448887e+01, 1.54802653564145e+01, 1.56640543403033e+01, 1.58529834232229e+01, 1.60458456796624e+01, 1.62413677660837e+01, 1.64382311876085e+01, 1.66350946091333e+01, 1.68306166955546e+01, 1.70234789519942e+01, 1.72124080349137e+01, 1.73961970188025e+01, 1.75737251303284e+01, 1.77439755017412e+01, 1.79060505468255e+01, 1.80591846244272e+01, 1.82027537248713e+01, 1.83362819915469e+01, 1.84594449714899e+01, 1.85720695727602e+01, 1.86741307905228e+01, 1.87657453457532e+01, 1.88471624582002e+01, 1.89187520465770e+01, 1.89809907120154e+01, 1.90344459139368e+01, 1.90797587892713e+01, 1.91176260952953e+01, 1.91487817725275e+01, 1.91739786267122e+01, 1.91939706179331e+01, 1.92094962206756e+01, 1.92212632819116e+01, 1.92299357560729e+01, 1.92361226374656e+01, 1.92403693439096e+01, 1.92431517320160e+01, 1.92448728466038e+01, 1.92458624264609e+01, 1.92463791082063e+01, 1.92466151916426e+01, 1.92467037558678e+01, 1.92467278475955e+01, 1.92467314034881e+01, 1.92467316883823e+01, 1.92467369419827e+01, 1.92467725359463e+01, 1.92469033838386e+01, 1.92472521817871e+01, 1.92480155455184e+01, 1.92494775855052e+01, 1.92520204205589e+01, 1.92561312183292e+01, 1.92624054513670e+01, 1.92715461669184e+01, 1.92843591843989e+01, 1.93017442533549e+01, 1.93246823233492e+01, 1.93542191923235e+01, 1.93914459083817e+01, 1.94374763985939e+01, 1.94934228845684e+01, 1.95603697157677e+01, 1.96393463058285e+01, 1.97312998929363e+01, 1.98370688615422e+01, 1.99573573588732e+01, 2.00927119158096e+01, 2.02435007383489e+01, 2.04098962741537e+01, 2.05918615802039e+01, 2.07891409243967e+01, 2.10012549485440e+01, 2.12275006053997e+01, 2.14669559611853e+01, 2.17184898308079e+01, 2.19807760889130e+01, 2.22523123794240e+01, 2.25314428325127e+01, 2.28163842941012e+01, 2.31052554817886e+01, 2.33961084050000e+01, 2.36869613282115e+01, 2.39758325158988e+01, 2.42607739774874e+01, 2.45399044305761e+01, 2.48114407210870e+01, 2.50737269791921e+01, 2.53252608488147e+01, 2.55647162046004e+01, 2.57909618614561e+01, 2.60030758856033e+01, 2.62003552297961e+01, 2.63823205358463e+01, 2.65487160716512e+01, 2.66995048941905e+01, 2.68348594511269e+01, 2.69551479484578e+01, 2.70609169170637e+01, 2.71528705041716e+01, 2.72318470942323e+01, 2.72987939254317e+01, 2.73547404114062e+01, 2.74007709016183e+01, 2.74379976176765e+01, 2.74675344866508e+01, 2.74904725566452e+01, 2.75078576256011e+01, 2.75206706430817e+01, 2.75298113586330e+01, 2.75360855916709e+01, 2.75401963894411e+01, 2.75427392244949e+01, 2.75442012644816e+01, 2.75449646282129e+01, 2.75453134261614e+01, 2.75454442740537e+01, 2.75454798680173e+01, 2.75454851216177e+01, 2.75454855464949e+01, 2.75454934309369e+01, 2.75455468492639e+01, 2.75457432217608e+01, 2.75462666870061e+01, 2.75474123198421e+01, 2.75496065045374e+01, 2.75534227132652e+01, 2.75595920722670e+01, 2.75690082484533e+01, 2.75827263535320e+01, 2.76019557365256e+01, 2.76280467139111e+01, 2.76624714646559e+01, 2.77067994901817e+01, 2.77626682019637e+01, 2.78317493475266e+01, 2.79157121148908e+01, 2.80161838624160e+01, 2.81347095024626e+01, 2.82727106209986e+01, 2.84314454396510e+01, 2.86119707209453e+01, 2.88151066816365e+01, 2.90414059139735e+01, 2.92911272221104e+01, 2.95642151630975e+01, 2.98602859420515e+01, 3.01786201529305e+01, 3.05181626840266e+01, 3.08775299254490e+01, 3.12550242293620e+01, 3.16486553875750e+01, 3.20561687102434e+01, 3.24750791188009e+01, 3.29027105103893e+01, 3.33362395141778e+01, 3.37727426457581e+01, 3.42092457773385e+01, 3.46427747811270e+01, 3.50704061727154e+01, 3.54893165812729e+01, 3.58968299039413e+01, 3.62904610621543e+01, 3.66679553660673e+01, 3.70273226074896e+01, 3.73668651385858e+01, 3.76851993494648e+01, 3.79812701284188e+01, 3.82543580694059e+01, 3.85040793775428e+01, 3.87303786098798e+01, 3.89335145705710e+01, 3.91140398518653e+01, 3.92727746705177e+01, 3.94107757890537e+01, 3.95293014291003e+01, 3.96297731766255e+01, 3.97137359439897e+01, 3.97828170895526e+01, 3.98386858013346e+01, 3.98830138268604e+01, 3.99174385776052e+01, 3.99435295549907e+01, 3.99627589379842e+01, 3.99764770430629e+01, 3.99858932192493e+01, 3.99920625782511e+01, 3.99958787869789e+01, 3.99980729716742e+01, 3.99992186045101e+01, 3.99997420697555e+01, 3.99999384422524e+01, 3.99999918605793e+01, 3.99999997450214e+01],
[ 1.00000000000000e+00, 9.99999999889928e-01, 9.99999996486289e-01, 9.99999973426100e-01, 9.99999888653936e-01, 9.99999662678842e-01, 9.99999168119462e-01, 9.99998220908527e-01, 9.99996573479931e-01, 9.99993910204311e-01, 9.99989845273768e-01, 9.99983923165215e-01, 9.99975621736633e-01, 9.99964357933462e-01, 9.99949496005737e-01, 9.99930358062378e-01, 9.99906236719547e-01, 9.99876409536990e-01, 9.99840154881594e-01, 9.99796768812490e-01, 9.99745582548056e-01, 9.99685980053110e-01, 9.99617415274851e-01, 9.99539428558983e-01, 9.99451661792832e-01, 9.99353871849654e-01, 9.99245941946993e-01, 9.99127890580963e-01, 9.98999877756314e-01, 9.98862208297790e-01, 9.98715332099846e-01, 9.98559841247586e-01, 9.98396464019895e-01, 9.98226055864217e-01, 9.98049587509366e-01, 9.97868130456063e-01, 9.97682840152839e-01, 9.97494937225497e-01, 9.97305687180029e-01, 9.97116379040094e-01, 9.96928303409788e-01, 9.96742730469383e-01, 9.96560888415529e-01, 9.96383942847680e-01, 9.96212977579456e-01, 9.96048977317727e-01, 9.95892812604323e-01, 9.95745227356658e-01, 9.95606829275832e-01, 9.95478083315766e-01, 9.95359308326784e-01, 9.95250676903973e-01, 9.95152218387114e-01, 9.95063824877173e-01, 9.94985260056805e-01, 9.94916170531047e-01, 9.94856099341463e-01, 9.94804501254156e-01, 9.94760759380640e-01, 9.94724202661756e-01, 9.94694123729151e-01, 9.94669796656791e-01, 9.94650494126389e-01, 9.94635503554967e-01, 9.94624141769458e-01, 9.94615767860841e-01, 9.94609793907695e-01, 9.94605693324321e-01, 9.94603006660094e-01, 9.94601344752277e-01, 9.94600389212280e-01, 9.94599890302945e-01, 9.94599662339992e-01, 9.94599576822061e-01, 9.94599553559000e-01, 9.94599550125417e-01, 9.94599549675004e-01, 9.94599539180904e-01, 9.94599468081733e-01, 9.94599206712159e-01, 9.94598509985318e-01, 9.94596985157213e-01, 9.94594064704967e-01, 9.94588985312927e-01, 9.94580773782270e-01, 9.94568240477086e-01, 9.94549980698897e-01, 9.94524384148834e-01, 9.94489652399129e-01, 9.94443824060332e-01, 9.94384807104574e-01, 9.94310417594898e-01, 9.94218423881880e-01, 9.94106595166727e-01, 9.93972753198549e-01, 9.93814825775810e-01, 9.93630900659776e-01, 9.93419278481994e-01, 9.93178523238338e-01, 9.92907509007893e-01, 9.92605461614218e-01, 9.92271994057091e-01, 9.91907134681872e-01, 9.91511347218089e-01, 9.91085542005330e-01, 9.90631077929286e-01, 9.90149754809850e-01, 9.89643796212129e-01, 9.89115822885300e-01, 9.88568817268281e-01, 9.88006079729524e-01, 9.87431177425120e-01, 9.86847886858561e-01, 9.86260131401075e-01, 9.85671915177498e-01, 9.85087254833950e-01, 9.84510110775814e-01, 9.83944319494124e-01, 9.83393528583682e-01, 9.82861135996241e-01, 9.82350234968066e-01, 9.81863565915593e-01, 9.81403476409790e-01, 9.80971890124509e-01, 9.80570285413148e-01, 9.80199683908522e-01, 9.79860649271010e-01, 9.79553295937930e-01, 9.79277307460854e-01, 9.79031963765202e-01, 9.78816176435198e-01, 9.78628530923822e-01, 9.78467334417216e-01, 9.78330667950834e-01, 9.78216441283453e-01, 9.78122448987258e-01, 9.78046426208045e-01, 9.77986102588633e-01, 9.77939252929053e-01, 9.77903743275891e-01, 9.77877571286575e-01, 9.77858899897457e-01, 9.77846083531907e-01, 9.77837686310387e-01, 9.77832491962233e-01, 9.77829505382298e-01, 9.77827946018166e-01, 9.77827233508922e-01, 9.77826966218402e-01, 9.77826893508550e-01, 9.77826882776716e-01, 9.77826881829299e-01, 9.77826863265130e-01, 9.77826737489973e-01, 9.77826275124504e-01, 9.77825042607265e-01, 9.77822345168615e-01, 9.77817178849213e-01, 9.77808193319256e-01, 9.77793666937905e-01, 9.77771495132467e-01, 9.77739192784768e-01, 9.77693910899739e-01, 9.77632467411131e-01, 9.77551391564332e-01, 9.77446980918955e-01, 9.77315369645750e-01, 9.77152606463781e-01, 9.76954740283233e-01, 9.76717911393271e-01, 9.76438445867613e-01, 9.76112950755187e-01, 9.75738407580136e-01, 9.75312261693265e-01, 9.74832505093522e-01, 9.74297750469988e-01, 9.73707294398455e-01, 9.73061167857783e-01, 9.72360172505941e-01, 9.71605901469268e-01, 9.70800743746694e-01, 9.69947871707825e-01, 9.69051211563946e-01, 9.68115397106428e-01, 9.67145707429012e-01, 9.66147989768458e-01, 9.65128569000478e-01, 9.64094145702064e-01, 9.63051685024326e-01, 9.62008298899211e-01, 9.60971124317541e-01, 9.59947200555261e-01, 9.58943348282658e-01, 9.57966053463904e-01, 9.57021358841022e-01, 9.56114765600295e-01, 9.55251147546299e-01, 9.54434679768131e-01, 9.53668783385288e-01, 9.52956087519994e-01, 9.52298409172722e-01, 9.51696751192825e-01, 9.51151318051141e-01, 9.50661548650455e-01, 9.50226164965888e-01, 9.49843234902924e-01, 9.49510247406482e-01, 9.49224197559261e-01, 9.48981679178874e-01, 9.48778982266239e-01, 9.48612192575755e-01, 9.48477290572069e-01, 9.48370247107876e-01, 9.48287113299261e-01, 9.48224102284697e-01, 9.48177660824320e-01, 9.48144529019362e-01, 9.48121786798129e-01, 9.48106886214324e-01, 9.48097669024665e-01, 9.48092369444233e-01, 9.48089602408189e-01, 9.48088338086044e-01, 9.48087863788603e-01, 9.48087734767580e-01, 9.48087715724325e-01, 9.48087714196820e-01, 9.48087686006603e-01, 9.48087495013467e-01, 9.48086792898459e-01, 9.48084921286616e-01, 9.48080825152513e-01, 9.48072979962628e-01, 9.48059335224232e-01, 9.48037276631053e-01, 9.48003608451338e-01, 9.47954557210150e-01, 9.47885797092788e-01, 9.47792496858907e-01, 9.47669387426512e-01, 9.47510848680273e-01, 9.47311013497065e-01, 9.47063886478922e-01, 9.46763474452747e-01, 9.46403925447328e-01, 9.45979672598301e-01, 9.45485579264718e-01, 9.44917081567572e-01, 9.44270324579916e-01, 9.43542288506780e-01, 9.42730901386908e-01, 9.41835135122584e-01, 9.40855081993609e-01, 9.39792009231756e-01, 9.38648389717289e-01, 9.37427907403269e-01, 9.36135436668357e-01, 9.34776995434763e-01, 9.33359672551875e-01, 9.31891530621803e-01, 9.30381486111601e-01, 9.28839169236563e-01, 9.27274766686876e-01, 9.25698850783003e-01, 9.24122199062111e-01, 9.22555608600472e-01, 9.21009709551439e-01, 9.19494782417327e-01, 9.18020583474492e-01, 9.16596182538383e-01, 9.15229816899176e-01, 9.13928764793580e-01, 9.12699241222631e-01, 9.11546318299275e-01, 9.10473871634885e-01, 9.09484553571984e-01, 9.08579793362011e-01, 9.07759823691218e-01, 9.07023732292040e-01, 9.06369536757289e-01, 9.05794280113851e-01, 9.05294144223443e-01, 9.04864577670498e-01, 9.04500434480321e-01, 9.04196119790964e-01, 9.03945738484988e-01, 9.03743242775088e-01, 9.03582574831089e-01, 9.03457800732708e-01, 9.03363232328011e-01, 9.03293533963939e-01, 9.03243811522297e-01, 9.03209681729586e-01, 9.03187320297143e-01, 9.03173488072994e-01, 9.03165535031084e-01, 9.03161382569255e-01, 9.03159485216415e-01, 9.03158773444358e-01, 9.03158579824203e-01, 9.03158551246241e-01, 9.03158549042034e-01, 9.03158509461531e-01, 9.03158241297442e-01, 9.03157255492456e-01, 9.03154627656610e-01, 9.03148876490571e-01, 9.03137861508833e-01, 9.03118703825796e-01, 9.03087733102992e-01, 9.03040463000141e-01, 9.02971596647410e-01, 9.02875062784553e-01, 9.02744082319231e-01, 9.02571264167118e-01, 9.02348728375688e-01, 9.02068253725994e-01, 9.01721446274672e-01, 9.01299924661382e-01, 9.00795517481226e-01, 9.00200467619764e-01, 8.99507638178711e-01, 8.98710714487884e-01, 8.97804396705180e-01, 8.96784577649930e-01, 8.95648500792862e-01, 8.94394893733469e-01, 8.93024073026889e-01, 8.91538016870408e-01, 8.89940402915148e-01, 8.88236609319573e-01, 8.86433678092359e-01, 8.84540240762403e-01, 8.82566407437560e-01, 8.80523621340146e-01, 8.78424481901190e-01, 8.76282540419195e-01, 8.74112073105100e-01, 8.71927837008168e-01, 8.69744814818194e-01, 8.67577954846436e-01, 8.65441912589297e-01, 8.63350800174328e-01, 8.61317949686953e-01, 8.59355695897480e-01, 8.57475183277414e-01, 8.55686201442881e-01, 8.53997052324111e-01, 8.52414451466227e-01, 8.50943464948409e-01, 8.49587482492701e-01, 8.48348226442424e-01, 8.47225795441340e-01, 8.46218740852234e-01, 8.45324173228338e-01, 8.44537895501805e-01, 8.43854558988146e-01, 8.43267837831574e-01, 8.42770617141299e-01, 8.42355189800738e-01, 8.42013456778127e-01, 8.41737125734843e-01, 8.41517902822234e-01, 8.41347672781384e-01, 8.41218662812358e-01, 8.41123586155160e-01, 8.41055761914630e-01, 8.41009208352095e-01, 8.40978707639779e-01, 8.40959840908184e-01, 8.40948993287565e-01, 8.40943329526075e-01, 8.40940741632246e-01, 8.40939770811886e-01, 8.40939506724073e-01, 8.40939467745224e-01, 8.40939464793862e-01, 8.40939412510784e-01, 8.40939058284793e-01, 8.40937756105988e-01, 8.40934284924736e-01, 8.40926688072438e-01, 8.40912138206492e-01, 8.40886832784962e-01, 8.40845924218527e-01, 8.40783487870946e-01, 8.40692530008816e-01, 8.40565036663272e-01, 8.40392063188514e-01, 8.40163863114286e-01, 8.39870053722160e-01, 8.39499814659665e-01, 8.39042114872620e-01, 8.38485962213695e-01, 8.37820669301530e-01, 8.37036128583504e-01, 8.36123089116755e-01, 8.35073427341874e-01, 8.33880404093075e-01, 8.32538900273870e-01, 8.31045624030258e-01, 8.29399282871139e-01, 8.27600715010323e-01, 8.25652975222552e-01, 8.23561371697463e-01, 8.21333451713026e-01, 8.18978935397804e-01, 8.16509598365333e-01, 8.13939105531563e-01, 8.11282799909619e-01, 8.08557451553524e-01, 8.05780973033826e-01, 8.02972108818352e-01, 8.00150106656158e-01, 7.97334379492008e-01, 7.94544166559188e-01, 7.91798202115303e-01, 7.89114399821249e-01, 7.86509560055253e-01, 7.83999106549792e-01, 7.81596857693938e-01, 7.79314836712458e-01, 7.77163123768008e-01, 7.75149751878628e-01, 7.73280647434683e-01, 7.71559615061873e-01, 7.69988365624246e-01, 7.68566585298792e-01, 7.67292042880240e-01, 7.66160731786700e-01, 7.65167042628277e-01, 7.64303961667954e-01, 7.63563290046418e-01, 7.62935878264012e-01, 7.62411870121999e-01, 7.61980950134047e-01, 7.61632588341817e-01, 7.61356276520802e-01, 7.61141749957375e-01, 7.60979189324953e-01, 7.60859397689996e-01, 7.60773948334163e-01, 7.60715299876311e-01, 7.60676876098430e-01, 7.60653108897280e-01, 7.60639443866457e-01, 7.60632309126143e-01, 7.60629049121540e-01, 7.60627826167866e-01, 7.60627493493549e-01, 7.60627444391472e-01, 7.60627440705363e-01, 7.60627375825656e-01, 7.60626936255623e-01, 7.60625320342564e-01, 7.60621012856926e-01, 7.60611585752381e-01, 7.60593530655668e-01, 7.60562129362035e-01, 7.60511367584429e-01, 7.60433896033016e-01, 7.60321041609224e-01, 7.60162870110016e-01, 7.59948300379612e-01, 7.59665268346640e-01, 7.59300937877149e-01, 7.58841953894681e-01, 7.58274731807804e-01, 7.57585775986726e-01, 7.56762018888781e-01, 7.55791171492947e-01, 7.54662075009008e-01, 7.53365043415951e-01, 7.51892186288734e-01, 7.50237701615729e-01, 7.48398128903538e-01, 7.46372553811789e-01, 7.44162756844189e-01, 7.41773300214842e-01, 7.39211548865541e-01, 7.36487623669495e-01, 7.33614287043148e-01, 7.30606763411232e-01, 7.27482499132696e-01, 7.24260868495612e-01, 7.20962834130128e-01, 7.17610571584666e-01, 7.14227068795251e-01, 7.10835711710062e-01, 7.07459867399000e-01, 7.04122475599684e-01, 7.00845658873245e-01, 6.97650360435428e-01, 6.94556017376495e-01, 6.91580275480503e-01, 6.88738750292667e-01, 6.86044837545397e-01, 6.83509574606895e-01, 6.81141553308952e-01, 6.78946883370561e-01, 6.76929204668952e-01, 6.75089745810660e-01, 6.73427425800896e-01, 6.71938995070768e-01, 6.70619211668054e-01, 6.69461048019771e-01, 6.68455923312433e-01, 6.67593956196552e-01, 6.66864232204512e-01, 6.66255079985370e-01, 6.65754350225677e-01, 6.65349690968552e-01, 6.65028812994103e-01, 6.64779739012994e-01, 6.64591030677693e-01, 6.64451987851487e-01, 6.64352815202381e-01, 6.64284752004069e-01, 6.64240162013230e-01, 6.64212581423373e-01, 6.64196724131707e-01, 6.64188444850788e-01, 6.64184661900042e-01, 6.64183242771847e-01, 6.64182856733364e-01, 6.64182799754849e-01, 6.64182795490461e-01, 6.64182720605651e-01, 6.64182213249604e-01, 6.64180348148543e-01, 6.64175376423479e-01, 6.64164495668983e-01, 6.64143656742225e-01, 6.64107414606263e-01, 6.64048829443167e-01, 6.63959422962401e-01, 6.63829193393929e-01, 6.63646691088891e-01, 6.63399154969957e-01, 6.63072708300065e-01, 6.62652610400966e-01, 6.62123559091063e-01, 6.61470036775835e-01, 6.60676691375647e-01, 6.59728741686658e-01, 6.58612395419185e-01, 6.57315267124617e-01, 6.55826782583572e-01, 6.54138556051723e-01, 6.52244727096690e-01, 6.50142244639104e-01, 6.47831087236050e-01, 6.45314410587691e-01, 6.42598615647742e-01, 6.39693333483165e-01, 6.36611326036216e-01, 6.33368305047645e-01, 6.29982674442960e-01, 6.26475204300118e-01, 6.22868646951058e-01, 6.19187307687471e-01, 6.15456583843059e-01, 6.11702486653418e-01, 6.07951160242034e-01, 6.04228411386979e-01, 6.00559262471875e-01, 5.96967538335494e-01, 5.93475495748074e-01, 5.90103502108042e-01, 5.86869767813361e-01, 5.83790134741768e-01, 5.80877921471123e-01, 5.78143824349550e-01, 5.75595872315428e-01, 5.73239432468483e-01, 5.71077262776752e-01, 5.69109607922146e-01, 5.67334334079546e-01, 5.65747098326521e-01, 5.64341548332068e-01, 5.63109547921343e-01, 5.62041424021611e-01, 5.61126230341418e-01, 5.60352022917191e-01, 5.59706142393185e-01, 5.59175497611006e-01, 5.58746844814483e-01, 5.58407056572045e-01, 5.58143374431184e-01, 5.57943639393577e-01, 5.57796494571809e-01, 5.57691554882910e-01, 5.57619539357618e-01, 5.57572362587279e-01, 5.57543182965008e-01, 5.57526406661381e-01, 5.57517647652575e-01, 5.57513645528275e-01, 5.57512144183440e-01, 5.57511735780313e-01, 5.57511675500833e-01, 5.57511670987706e-01, 5.57511591712347e-01, 5.57511054609813e-01, 5.57509080159337e-01, 5.57503816962839e-01, 5.57492298401085e-01, 5.57470238268635e-01, 5.57431873295524e-01, 5.57369859345585e-01, 5.57275226797677e-01, 5.57137399168951e-01, 5.56944277426707e-01, 5.56682390656041e-01, 5.56337111810032e-01, 5.55892935186396e-01, 5.55333810086470e-01, 5.54643522878687e-01, 5.53806117492416e-01, 5.52806342314126e-01, 5.51630109668283e-01, 5.50264952670813e-01, 5.48700463372393e-01, 5.46928695876825e-01, 5.44944518613863e-01, 5.42745901214255e-01, 5.40334123476269e-01, 5.37713896670926e-01, 5.34893390793483e-01, 5.31884165163830e-01, 5.28701003797233e-01, 5.25361660969047e-01, 5.21886526133406e-01, 5.18298220589601e-01, 5.14621140818605e-01, 5.10880965087905e-01, 5.07104140665164e-01, 5.03317368785744e-01, 4.99547103454818e-01, 4.95819078365670e-01, 4.92157873864397e-01, 4.88586533197899e-01, 4.85126234463168e-01, 4.81796021932469e-01, 4.78612597930061e-01, 4.75590174306514e-01, 4.72740380872143e-01, 4.70072226937925e-01, 4.67592111353075e-01, 4.65303876070546e-01, 4.63208898237678e-01, 4.61306216007922e-01, 4.59592683606422e-01, 4.58063151567909e-01, 4.56710668422676e-01, 4.55526700375271e-01, 4.54501365660890e-01, 4.53623680257723e-01, 4.52881811482033e-01, 4.52263335718809e-01, 4.51755496183099e-01, 4.51345456217375e-01, 4.51020543268164e-01, 4.50768478412842e-01, 4.50577586183561e-01, 4.50436979509323e-01, 4.50336714904286e-01, 4.50267913587470e-01, 4.50222845022591e-01, 4.50194970392520e-01, 4.50178944727922e-01, 4.50170577735832e-01, 4.50166754753580e-01, 4.50165320615378e-01, 4.50164930494488e-01, 4.50164872913443e-01, 4.50164868590911e-01, 4.50164792510713e-01, 4.50164277056035e-01, 4.50162382187902e-01, 4.50157331148196e-01, 4.50146276985258e-01, 4.50125106610955e-01, 4.50088290120371e-01, 4.50028782113551e-01, 4.49937980613658e-01, 4.49805747863286e-01, 4.49620495778743e-01, 4.49369337128535e-01, 4.49038301563516e-01, 4.48612613466569e-01, 4.48077026240637e-01, 4.47416205180280e-01, 4.46615148575202e-01, 4.45659634311142e-01, 4.44536677129511e-01, 4.43234980063810e-01, 4.41745362569836e-01, 4.40061147668679e-01, 4.38178491146251e-01, 4.36096637558291e-01, 4.33818090457161e-01, 4.31348687782773e-01, 4.28697577557489e-01, 4.25877093633922e-01, 4.22902535953423e-01, 4.19791864247003e-01, 4.16565318025416e-01, 4.13244978781697e-01, 4.09854292362469e-01, 4.06417570344554e-01, 4.02959488976712e-01, 3.99504602911004e-01, 3.96076888738881e-01, 3.92699330510113e-01, 3.89393556224872e-01, 3.86179531027197e-01, 3.83075309738510e-01, 3.80096848650121e-01, 3.77257874277651e-01, 3.74569805136229e-01, 3.72041721529630e-01, 3.69680377814737e-01, 3.67490251523220e-01, 3.65473623991064e-01, 3.63630687651159e-01, 3.61959675774886e-01, 3.60457011107165e-01, 3.59117470444480e-01, 3.57934362695016e-01, 3.56899718293291e-01, 3.56004487997503e-01, 3.55238749074299e-01, 3.54591916687806e-01, 3.54052957987717e-01, 3.53610605977444e-01, 3.53253569789813e-01, 3.52970737562250e-01, 3.52751367745314e-01, 3.52585264453891e-01, 3.52462932427745e-01, 3.52375707342688e-01, 3.52315857624436e-01, 3.52276654563956e-01, 3.52252408395741e-01, 3.52238469040570e-01, 3.52231191378252e-01, 3.52227866139916e-01, 3.52226618725900e-01, 3.52226279398823e-01, 3.52226229314850e-01, 3.52226225539101e-01, 3.52226158869442e-01, 3.52225707172674e-01, 3.52224046687248e-01, 3.52219620440262e-01, 3.52209933694358e-01, 3.52191382368706e-01, 3.52159121306317e-01, 3.52106978446414e-01, 3.52027420015819e-01, 3.51911570753128e-01, 3.51749291890070e-01, 3.51529318100338e-01, 3.51239452861065e-01, 3.50866819649544e-01, 3.50398164142360e-01, 3.49820200160281e-01, 3.49119989618594e-01, 3.48285344351620e-01, 3.47305235569887e-01, 3.46170195087988e-01, 3.44872691538715e-01, 3.43407464745851e-01, 3.41771802390519e-01, 3.39965745120788e-01, 3.37992209270313e-01, 3.35857020213016e-01, 3.33568853833618e-01, 3.31139088307604e-01, 3.28581572985761e-01, 3.25912325293804e-01, 3.23149169856545e-01, 3.20311336291618e-01, 3.17419033155187e-01, 3.14493015351159e-01, 3.11554161046730e-01, 3.08623071980463e-01, 3.05719708281099e-01, 3.02863065840355e-01, 3.00070901194611e-01, 2.97359506019332e-01, 2.94743530912301e-01, 2.92235856248649e-01, 2.89847506570651e-01, 2.87587604204067e-01, 2.85463357499836e-01, 2.83480079185744e-01, 2.81641230666574e-01, 2.79948488624591e-01, 2.78401830848636e-01, 2.76999638780699e-01, 2.75738814754732e-01, 2.74614912273711e-01, 2.73622277905064e-01, 2.72754203463504e-01, 2.72003087098084e-01, 2.71360601720762e-01, 2.70817868929136e-01, 2.70365636215136e-01, 2.69994454849407e-01, 2.69694855426959e-01, 2.69457517695603e-01, 2.69273431007144e-01, 2.69134041572093e-01, 2.69031382695983e-01, 2.68958184353944e-01, 2.68907958832586e-01, 2.68875059732497e-01, 2.68854712363620e-01, 2.68843014447778e-01, 2.68836907023668e-01, 2.68834116477946e-01, 2.68833069645769e-01, 2.68832784881833e-01, 2.68832742851266e-01, 2.68832739665407e-01, 2.68832683182904e-01, 2.68832300505616e-01, 2.68830893743101e-01, 2.68827143828235e-01, 2.68818937218616e-01, 2.68803220530904e-01, 2.68775888927106e-01, 2.68731713380345e-01, 2.68664311144072e-01, 2.68566162811919e-01, 2.68428678255624e-01, 2.68242312434411e-01, 2.67996730556951e-01, 2.67681020340229e-01, 2.67283947169917e-01, 2.66794245881866e-01, 2.66200940753093e-01, 2.65493683252664e-01, 2.64663095330869e-01, 2.63701104709994e-01, 2.62601257968947e-01, 2.61358997343829e-01, 2.59971888196033e-01, 2.58439786045431e-01, 2.56764934848703e-01, 2.54951991643088e-01, 2.53007976509794e-01, 2.50942150718341e-01, 2.48765829555073e-01, 2.46492139405212e-01, 2.44135730905571e-01, 2.41712461271228e-01, 2.39239059194779e-01, 2.36732785102184e-01, 2.34211098195367e-01, 2.31691339846428e-01, 2.29190440779319e-01, 2.26724657313925e-01, 2.24309339943943e-01, 2.21958735804107e-01, 2.19685825222574e-01, 2.17502191563073e-01, 2.15417922908820e-01, 2.13441543769644e-01, 2.11579974837144e-01, 2.09838518802402e-01, 2.08220870327386e-01, 2.06729148377021e-01, 2.05363949238289e-01, 2.04124418650779e-01, 2.03008341532591e-01, 2.02012247795382e-01, 2.01131532695725e-01, 2.00360590063634e-01, 1.99692956583677e-01, 1.99121465084727e-01, 1.98638404531131e-01, 1.98235684116815e-01, 1.97904998565838e-01, 1.97637991464374e-01, 1.97426413219367e-01, 1.97262270088777e-01, 1.97137960686232e-01, 1.97046396453613e-01, 1.96981102835332e-01, 1.96936298285426e-01, 1.96906948788579e-01, 1.96888796263036e-01, 1.96878360009841e-01, 1.96872911242567e-01, 1.96870421632372e-01, 1.96869487689846e-01, 1.96869233634477e-01, 1.96869196136424e-01, 1.96869193276087e-01, 1.96869142326881e-01, 1.96868797138419e-01, 1.96867528187690e-01, 1.96864145619988e-01, 1.96856742897716e-01, 1.96842565585919e-01, 1.96817910454903e-01, 1.96778059417376e-01, 1.96717252050866e-01, 1.96628699574719e-01, 1.96504642137587e-01, 1.96336450108329e-01, 1.96114768738527e-01, 1.95829704079721e-01, 1.95471046415082e-01, 1.95028525752422e-01, 1.94492092202234e-01, 1.93852212439375e-01, 1.93100172051458e-01, 1.92228372553972e-01, 1.91230611338355e-01, 1.90102332925658e-01, 1.88840840689200e-01, 1.87445459686387e-01, 1.85917643332628e-01, 1.84261019220193e-01, 1.82481372237787e-01, 1.80586566057322e-01, 1.78586406796255e-01, 1.76492455038709e-01, 1.74317794260956e-01, 1.72076764977310e-01, 1.69784674589156e-01, 1.67457493028174e-01, 1.65111543919809e-01, 1.62763200258872e-01, 1.60428592591092e-01, 1.58123336526243e-01, 1.55862285146972e-01, 1.53659310582662e-01, 1.51527117737398e-01, 1.49477091935239e-01, 1.47519181111135e-01, 1.45661812165567e-01, 1.43911840245218e-01, 1.42274529033467e-01, 1.40753559645306e-01, 1.39351065420258e-01, 1.38067689778649e-01, 1.36902664322373e-01, 1.35853904482419e-01, 1.34918120197061e-01, 1.34090939301368e-01, 1.33367041479130e-01, 1.32740300739674e-01, 1.32203934413140e-01, 1.31750656601086e-01, 1.31372833881080e-01, 1.31062640863616e-01, 1.30812212966883e-01, 1.30613793547015e-01, 1.30459872338483e-01, 1.30343312060780e-01, 1.30257460067115e-01, 1.30196242073575e-01, 1.30154235325952e-01, 1.30126719035931e-01, 1.30109700534504e-01, 1.30099916322007e-01, 1.30094808004708e-01, 1.30092473954849e-01, 1.30091598369337e-01, 1.30091360188580e-01, 1.30091325033590e-01, 1.30091322333058e-01, 1.30091273982263e-01, 1.30090946398460e-01, 1.30089742164934e-01, 1.30086532114231e-01, 1.30079506965045e-01, 1.30066052842338e-01, 1.30042655629735e-01, 1.30004838493472e-01, 1.29947136211486e-01, 1.29863109126747e-01, 1.29745398602052e-01, 1.29585824765152e-01, 1.29375526085498e-01, 1.29105138917052e-01, 1.28765013595142e-01, 1.28345462033047e-01, 1.27837030096245e-01, 1.27230786434626e-01, 1.26518618041102e-01, 1.25693521704900e-01, 1.24749879863147e-01, 1.23683709231369e-01, 1.22492871086183e-01, 1.21177233211032e-01, 1.19738775274824e-01, 1.18181631716273e-01, 1.16512068927036e-01, 1.14738396499743e-01, 1.12870815344999e-01, 1.10921208390054e-01, 1.08902882166101e-01, 1.06830269709252e-01, 1.04718606714568e-01, 1.02583593707884e-01, 1.00441057099651e-01, 9.83066213726415e-02, 9.61954033961336e-02, 9.41217380647992e-02, 9.20989422804595e-02, 9.01391219034560e-02, 8.82530238801998e-02, 8.64499334776507e-02, 8.47376145714035e-02, 8.31222893504975e-02, 8.16086526818843e-02, 8.01999157363000e-02, 7.88978732876129e-02, 7.77029892974845e-02, 7.66144959004859e-02, 7.56305016135095e-02, 7.47481054075990e-02, 7.39635141074827e-02, 7.32721613451268e-02, 7.26688269272155e-02, 7.21477559412774e-02, 7.17027772002038e-02, 7.13274207086705e-02, 7.10150337437623e-02, 7.07588949073049e-02, 7.05523251722523e-02, 7.03887945610832e-02, 7.02620227151235e-02, 7.01660712933333e-02, 7.00954259246854e-02, 7.00450653668319e-02, 7.00105156184970e-02, 6.99878870013046e-02, 6.99738926593666e-02, 6.99658474975088e-02, 6.99616472546036e-02, 6.99597281418606e-02, 6.99590082185079e-02, 6.99588123820979e-02, 6.99587834770610e-02, 6.99587812396921e-02, 6.99587409608554e-02, 6.99584680659673e-02, 6.99574648778549e-02, 6.99547907648194e-02, 6.99489386259623e-02, 6.99377314329879e-02, 6.99182431354692e-02, 6.98867478342260e-02, 6.98387006528749e-02, 6.97687531595326e-02, 6.96708054684688e-02, 6.95380962430636e-02, 6.93633306920742e-02, 6.91388452839932e-02, 6.88568063090371e-02, 6.85094376391948e-02, 6.80892711560683e-02, 6.75894114536923e-02, 6.70038047302256e-02, 6.63275004291554e-02, 6.55568933512263e-02, 6.46899337892790e-02, 6.37262938573236e-02, 6.26674796512425e-02, 6.15168811754313e-02, 6.02797549979114e-02, 5.89631381736434e-02, 5.75756958480859e-02, 5.61275088165781e-02, 5.46298108460520e-02, 5.30946884556279e-02, 5.15347578453836e-02, 4.99628345854054e-02, 4.83916114639135e-02, 4.68333585928238e-02, 4.52996576394643e-02, 4.38011791393990e-02, 4.23475085469120e-02, 4.09470233116082e-02, 3.96068201238500e-02, 3.83326887859757e-02, 3.71291271026177e-02, 3.59993898209212e-02, 3.49455639897298e-02, 3.39686630800060e-02, 3.30687327049314e-02, 3.22449616601735e-02, 3.14957931300178e-02, 3.08190321401968e-02, 3.02119465699037e-02, 2.96713601749559e-02, 2.91937370585723e-02, 2.87752578167078e-02, 2.84118881627142e-02, 2.80994411988875e-02, 2.78336346605447e-02, 2.76101444315547e-02, 2.74246554461812e-02, 2.72729107839804e-02, 2.71507593703910e-02, 2.70542022567903e-02, 2.69794370129786e-02, 2.69228993643047e-02, 2.68813008832087e-02, 2.68516613324190e-02, 2.68313341761958e-02, 2.68180238366545e-02, 2.68097934709314e-02, 2.68050623653983e-02, 2.68025924578646e-02, 2.68014639723172e-02, 2.68010406441891e-02, 2.68009254892176e-02, 2.68009084926047e-02, 2.68009071663312e-02, 2.68008831519467e-02, 2.68007204512672e-02, 2.68001223509698e-02, 2.67985280705972e-02, 2.67950392067673e-02, 2.67883583040308e-02, 2.67767423138647e-02, 2.67579735737201e-02, 2.67293507718824e-02, 2.66877021189480e-02, 2.66294225696652e-02, 2.65505363731351e-02, 2.64467854255614e-02, 2.63137428274729e-02, 2.61469497069576e-02, 2.59420718031638e-02, 2.56950705983836e-02, 2.54023820781873e-02, 2.50610946631477e-02, 2.46691166929532e-02, 2.42253232552511e-02, 2.37296723098568e-02, 2.31832810768946e-02, 2.25884555603284e-02, 2.19486687854618e-02, 2.12684866488707e-02, 2.05534439226829e-02, 1.98098765656630e-02, 1.90447196925377e-02, 1.82652829923294e-02, 1.74790168002993e-02, 1.66932822733948e-02, 1.59151381971641e-02, 1.51511550064083e-02, 1.44072638927889e-02, 1.36886457342843e-02, 1.29996613698729e-02, 1.23438217852230e-02, 1.17237943288167e-02, 1.11414393053104e-02, 1.05978702543706e-02, 1.00935308866600e-02, 9.62828191127729e-03, 9.20149170546480e-03, 8.81212578772594e-03, 8.45883120783353e-03, 8.14001313349411e-03, 7.85390199734615e-03, 7.59861050652377e-03, 7.37218057771494e-03, 7.17262083549108e-03, 6.99793571020990e-03, 6.84614741388223e-03, 6.71531218249166e-03, 6.60353217608756e-03, 6.50896434617820e-03, 6.42982743313025e-03, 6.36440806118843e-03, 6.31106667008408e-03, 6.26824377375401e-03, 6.23446678210460e-03, 6.20835737521197e-03, 6.18863919536261e-03, 6.17414543523109e-03, 6.16382576302410e-03, 6.15675194745091e-03, 6.15212153226441e-03, 6.14925896173157e-03, 6.14761366862593e-03, 6.14675479357707e-03, 6.14636239301114e-03, 6.14621519430480e-03, 6.14617515309290e-03, 6.14616924310862e-03, 6.14616877801910e-03, 6.14616030652326e-03, 6.14610291111339e-03, 6.14589192281456e-03, 6.14532953082547e-03, 6.14409887834366e-03, 6.14174252447408e-03, 6.13764634735015e-03, 6.13102997502535e-03, 6.12094483165504e-03, 6.10628084884823e-03, 6.08578278026539e-03, 6.05807683326443e-03, 6.02170796044599e-03, 5.97518761410187e-03, 5.91705105637208e-03, 5.84592246437643e-03, 5.76058513342492e-03, 5.66005315735640e-03, 5.54364017581621e-03, 5.41102026084972e-03, 5.26227590029693e-03, 5.09792842263601e-03, 4.91894714041119e-03, 4.72673493491971e-03, 4.52308984767924e-03, 4.31014429253940e-03, 4.09028551222101e-03, 3.86606261554845e-03, 3.64008671654970e-03, 3.41493119415714e-03, 3.19303884206223e-03, 2.97664173453938e-03, 2.76769814850053e-03, 2.56784907722940e-03, 2.37839499606374e-03, 2.20029182559478e-03, 2.03416366132281e-03, 1.88032890557910e-03, 1.73883597840385e-03, 1.60950476406274e-03, 1.49197028524817e-03, 1.38572567532036e-03, 1.29016222087974e-03, 1.20460496647534e-03, 1.12834303166121e-03, 1.06065434157876e-03, 1.00082489743758e-03, 9.48163013654413e-04, 9.02009135907100e-04, 8.61741946683980e-04, 8.26781482953454e-04, 7.96589956285176e-04, 7.70670899625792e-04, 7.48567183445625e-04, 7.29858358421542e-04, 7.14157698526563e-04, 7.01109240096587e-04, 6.90385039829877e-04, 6.81682807563311e-04, 6.74724007693528e-04, 6.69252466125148e-04, 6.65033467923681e-04, 6.61853285318648e-04, 6.59519037772616e-04, 6.57858757276530e-04, 6.56721514622816e-04, 6.55977457373114e-04, 6.55517617996759e-04, 6.55253370364206e-04, 6.55115442447896e-04, 6.55052429614833e-04, 6.55028792550749e-04, 6.55022362811564e-04, 6.55021413800028e-04, 6.55021338455852e-04, 6.55019957679470e-04, 6.55010602778394e-04, 6.54976214114811e-04, 6.54884553799362e-04, 6.54683994153915e-04, 6.54300039089173e-04, 6.53632777960711e-04, 6.52555486947315e-04, 6.50914607075670e-04, 6.48531332541613e-04, 6.45205027622775e-04, 6.40718643970861e-04, 6.34846222236196e-04, 6.27362426293465e-04, 6.18053874950064e-04, 6.06731814343231e-04, 5.93245434927558e-04, 5.77494911696230e-04, 5.59443075031196e-04, 5.39124544735925e-04, 5.16651217674524e-04, 4.92213210615655e-04, 4.66074721092946e-04, 4.38564748882042e-04, 4.10063161265461e-04, 3.80983110442101e-04, 3.51751239080045e-04, 3.22787367142974e-04, 2.94485390780364e-04, 2.67196931212910e-04, 2.41218882986782e-04, 2.16785504142614e-04, 1.94065165332935e-04, 1.73161418924983e-04, 1.54117709917134e-04, 1.36924830334784e-04, 1.21530101741762e-04, 1.07847256445859e-04, 9.57660970143381e-05, 8.51612498378174e-05, 7.58996381100552e-05, 6.78465779544446e-05, 6.08705579331348e-05, 5.48467898875415e-05, 4.96595970828073e-05, 4.52037252944848e-05, 4.13847460402073e-05, 3.81188105404995e-05, 3.53320339914198e-05, 3.29597230841472e-05, 3.09455528426650e-05, 2.92407178529297e-05, 2.78030590185052e-05, 2.65961840121287e-05, 2.55886213660350e-05, 2.47530510578686e-05, 2.40656386183424e-05, 2.35054787846023e-05, 2.30541403021256e-05, 2.26952987143481e-05, 2.24144452124486e-05, 2.21986620542181e-05, 2.20364559972985e-05, 2.19176404344656e-05, 2.18332554758893e-05, 2.17755141016518e-05, 2.17377622431236e-05, 2.17144413569015e-05, 2.17010435429321e-05, 2.16940513955155e-05, 2.16908572460150e-05, 2.16896591106464e-05, 2.16893331985712e-05, 2.16892850947170e-05, 2.16892812406888e-05, 2.16892101638777e-05, 2.16887286154099e-05, 2.16869584718227e-05, 2.16822405370021e-05, 2.16719185868186e-05, 2.16521628250144e-05, 2.16178449233828e-05, 2.15624786726876e-05, 2.14782419569737e-05, 2.13560966120241e-05, 2.11860216833282e-05, 2.09573717150358e-05, 2.06593641816077e-05, 2.02816888242863e-05, 1.98152170676202e-05, 1.92527732564285e-05, 1.85899128444245e-05, 1.78256372390662e-05, 1.69629615780487e-05, 1.60092420198321e-05, 1.49761686636385e-05, 1.38793507608037e-05, 1.27374770909627e-05, 1.15711311216016e-05, 1.04014534614283e-05, 9.24891099585157e-06, 8.13237859476068e-06, 7.06854412513280e-06, 6.07140793862371e-06, 5.15156586411436e-06, 4.31520511012806e-06, 3.56322979433443e-06, 2.89127104025066e-06, 2.29103765314598e-06, 1.75242674926857e-06, 1.26454147486577e-06, 8.12336200861630e-07, 3.55027195420021e-07, 3.17268591176756e-07, 6.09459165357151e-07, 7.95757475802183e-07, 9.31858308961253e-07, 1.03466044454326e-06, 1.11048509543465e-06, 1.16353955148529e-06, 1.19916561103217e-06, 1.22381691480822e-06, 1.24339406426532e-06, 1.26167249114850e-06, 1.27990102150033e-06, 1.29756782751844e-06, 1.31359279553767e-06, 1.32720566192734e-06, 1.33823469761207e-06, 1.34695995424039e-06, 1.35382948207585e-06, 1.35925079058907e-06, 1.36351734147944e-06, 1.36682818807632e-06, 1.36933527039584e-06, 1.37117688980837e-06, 1.37248767850817e-06, 1.37339360584132e-06, 1.37400359345730e-06, 1.37440463443550e-06, 1.37466184746415e-06, 1.37482186413074e-06, 1.37491732810080e-06, 1.37497096608374e-06, 1.37499860297652e-06, 1.37501114605421e-06, 1.37501583774569e-06, 1.37501711261478e-06, 1.37501730079759e-06, 1.37501730879119e-06, 1.37501736929835e-06, 1.37501778041393e-06, 1.37501930541628e-06, 1.37502347619112e-06, 1.37503313757392e-06, 1.37505365949142e-06, 1.37509551678683e-06, 1.37517906595441e-06, 1.37534192901079e-06, 1.37564786018007e-06, 1.37619284878987e-06, 1.37709986502976e-06, 1.37849038472915e-06, 1.38042361730317e-06, 1.38280972394877e-06, 1.38533410032338e-06, 1.38746647160293e-06, 1.38864243048875e-06, 1.38865475074468e-06, 1.38815111261348e-06, 1.38893738899082e-06, 1.39366331386448e-06, 1.40463164288729e-06, 1.42205204067825e-06, 1.44280532898833e-06, 1.46099428299795e-06, 1.47069797218347e-06, 1.46976823067746e-06, 1.46221308361394e-06, 1.45664650463839e-06, 1.46017378832575e-06, 1.47080640222619e-06, 1.47421973817341e-06, 1.44835135118537e-06, 1.37340372918793e-06, 1.24141884039197e-06, 1.06079006390036e-06, 8.54207010569930e-07, 6.51304105092620e-07, 4.78886024057879e-07, 3.51917520791112e-07, 2.68591853980205e-07, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15, -2.88657986402541e-15],
[ 2.00000000000000e+00, 1.99999999964370e+00, 1.99999998862625e+00, 1.99999991398131e+00, 1.99999963957704e+00, 1.99999890810403e+00, 1.99999730723247e+00, 1.99999424113900e+00, 1.99998890844779e+00, 1.99998028744353e+00, 1.99996712920193e+00, 1.99994795905176e+00, 1.99992108653847e+00, 1.99988462380985e+00, 1.99983651209700e+00, 1.99977455572542e+00, 1.99969646286867e+00, 1.99959989205712e+00, 1.99948250328137e+00, 1.99934201238966e+00, 1.99917624737336e+00, 1.99898320506736e+00, 1.99876110676375e+00, 1.99850845124786e+00, 1.99822406381497e+00, 1.99790713991127e+00, 1.99755728216287e+00, 1.99717452970706e+00, 1.99675937891880e+00, 1.99631279482701e+00, 1.99583621273587e+00, 1.99533152980142e+00, 1.99480108655751e+00, 1.99424763863251e+00, 1.99367431914368e+00, 1.99308459249308e+00, 1.99248220051225e+00, 1.99187110210599e+00, 1.99125540772254e+00, 1.99063931012367e+00, 1.99002701303758e+00, 1.98942265934700e+00, 1.98883026049110e+00, 1.98825362874073e+00, 1.98769631394225e+00, 1.98716154621609e+00, 1.98665218594521e+00, 1.98617068219960e+00, 1.98571904052067e+00, 1.98529880074115e+00, 1.98491102524799e+00, 1.98455629781706e+00, 1.98423473286650e+00, 1.98394599469934e+00, 1.98368932604355e+00, 1.98346358495695e+00, 1.98326728895186e+00, 1.98309866501666e+00, 1.98295570407258e+00, 1.98283621830842e+00, 1.98273789978492e+00, 1.98265837869542e+00, 1.98259527970949e+00, 1.98254627490945e+00, 1.98250913195336e+00, 1.98248175625757e+00, 1.98246222618309e+00, 1.98244882042616e+00, 1.98244003704930e+00, 1.98243460383780e+00, 1.98243147992080e+00, 1.98242984885017e+00, 1.98242910357668e+00, 1.98242882399500e+00, 1.98242874794165e+00, 1.98242873671632e+00, 1.98242873524379e+00, 1.98242870093568e+00, 1.98242846849287e+00, 1.98242761400334e+00, 1.98242533620854e+00, 1.98242035111011e+00, 1.98241080328292e+00, 1.98239419712985e+00, 1.98236735072249e+00, 1.98232637420084e+00, 1.98226667397447e+00, 1.98218298320197e+00, 1.98206941824885e+00, 1.98191956006136e+00, 1.98172655866940e+00, 1.98148325836650e+00, 1.98118234052897e+00, 1.98081648054206e+00, 1.98037851491112e+00, 1.97986161435377e+00, 1.97925945849787e+00, 1.97856640774650e+00, 1.97777766791071e+00, 1.97688944334676e+00, 1.97589907455882e+00, 1.97480515653532e+00, 1.97360763446884e+00, 1.97230787396267e+00, 1.97090870334692e+00, 1.96941442630931e+00, 1.96783080368672e+00, 1.96616500395639e+00, 1.96442552270163e+00, 1.96262207209203e+00, 1.96076544219670e+00, 1.95886733671796e+00, 1.95694018646864e+00, 1.95499694459090e+00, 1.95305086810096e+00, 1.95111529081573e+00, 1.94920339305165e+00, 1.94732797366408e+00, 1.94550123000610e+00, 1.94373455122429e+00, 1.94203832997748e+00, 1.94042179717433e+00, 1.93889288369216e+00, 1.93745811228504e+00, 1.93612252204096e+00, 1.93488962683279e+00, 1.93376140825908e+00, 1.93273834261609e+00, 1.93181946051449e+00, 1.93100243687812e+00, 1.93028370826584e+00, 1.92965861376060e+00, 1.92912155509289e+00, 1.92866617122119e+00, 1.92828552229148e+00, 1.92797227774545e+00, 1.92771890334546e+00, 1.92751784202851e+00, 1.92736168378610e+00, 1.92724332017889e+00, 1.92715607962235e+00, 1.92709384020306e+00, 1.92705111748805e+00, 1.92702312554977e+00, 1.92700581022679e+00, 1.92699585445356e+00, 1.92699065629979e+00, 1.92698828114210e+00, 1.92698739012578e+00, 1.92698714774654e+00, 1.92698711197183e+00, 1.92698710881360e+00, 1.92698704692970e+00, 1.92698662765654e+00, 1.92698508635474e+00, 1.92698097773861e+00, 1.92697198577024e+00, 1.92695476368610e+00, 1.92692480999016e+00, 1.92687638518020e+00, 1.92680247276199e+00, 1.92669478678825e+00, 1.92654382677844e+00, 1.92633897947691e+00, 1.92606866553614e+00, 1.92572052791244e+00, 1.92528165757322e+00, 1.92473885106964e+00, 1.92407889365308e+00, 1.92328886092184e+00, 1.92235643148586e+00, 1.92127020282791e+00, 1.92002000241476e+00, 1.91859718615733e+00, 1.91699491652265e+00, 1.91520841295094e+00, 1.91323516771844e+00, 1.91107512100872e+00, 1.90873078971070e+00, 1.90620734535542e+00, 1.90351263763827e+00, 1.90065716114967e+00, 1.89765396424747e+00, 1.89451850043152e+00, 1.89126842409433e+00, 1.88792333407804e+00, 1.88450447001253e+00, 1.88103436787790e+00, 1.87753648256131e+00, 1.87403478629607e+00, 1.87055335272472e+00, 1.86711593687058e+00, 1.86374556150743e+00, 1.86046412027258e+00, 1.85729200738390e+00, 1.85424778302064e+00, 1.85134788235076e+00, 1.84860637488122e+00, 1.84603477932945e+00, 1.84364193761886e+00, 1.84143394994567e+00, 1.83941417120105e+00, 1.83758326740767e+00, 1.83593932928555e+00, 1.83447803863392e+00, 1.83319288193416e+00, 1.83207540446868e+00, 1.83111549733461e+00, 1.83030170902551e+00, 1.82962157277380e+00, 1.82906194059947e+00, 1.82860931500355e+00, 1.82825016947494e+00, 1.82797124944274e+00, 1.82775984598912e+00, 1.82760403552282e+00, 1.82749287967555e+00, 1.82741658089434e+00, 1.82736659052809e+00, 1.82733566760959e+00, 1.82731788797679e+00, 1.82730860481855e+00, 1.82730436313270e+00, 1.82730277190819e+00, 1.82730233905450e+00, 1.82730227516612e+00, 1.82730227004149e+00, 1.82730217546590e+00, 1.82730153470147e+00, 1.82729917917034e+00, 1.82729290008682e+00, 1.82727915794666e+00, 1.82725283811368e+00, 1.82720706147346e+00, 1.82713305752356e+00, 1.82702010545104e+00, 1.82685554676508e+00, 1.82662487096409e+00, 1.82631187358493e+00, 1.82589888387145e+00, 1.82536705727264e+00, 1.82469672609613e+00, 1.82386779995197e+00, 1.82286020616708e+00, 1.82165435916388e+00, 1.82023164689741e+00, 1.81857492184198e+00, 1.81666898371214e+00, 1.81450104108650e+00, 1.81206113936861e+00, 1.80934254306053e+00, 1.80634206113641e+00, 1.80306030538805e+00, 1.79950187297389e+00, 1.79567544604085e+00, 1.79159380320016e+00, 1.78727373980788e+00, 1.78273589639178e+00, 1.77800449712231e+00, 1.77310700286322e+00, 1.76807368595637e+00, 1.76293713637764e+00, 1.75773171112533e+00, 1.75249294055418e+00, 1.74725690674804e+00, 1.74205960986326e+00, 1.73693633863457e+00, 1.73192106091628e+00, 1.72704584926895e+00, 1.72234035525935e+00, 1.71783134440772e+00, 1.71354230169048e+00, 1.70949311529116e+00, 1.70569984398563e+00, 1.70217457123442e+00, 1.69892534680515e+00, 1.69595621461314e+00, 1.69326732348413e+00, 1.69085511573159e+00, 1.68871258681434e+00, 1.68682960790565e+00, 1.68519330196737e+00, 1.68378846289042e+00, 1.68259800644581e+00, 1.68160344120347e+00, 1.68078534723776e+00, 1.68012385036620e+00, 1.67959907987880e+00, 1.67919159822238e+00, 1.67888279190927e+00, 1.67865521401933e+00, 1.67849287003892e+00, 1.67838144040195e+00, 1.67830843492464e+00, 1.67826327630707e+00, 1.67823731195110e+00, 1.67822375545285e+00, 1.67821756120226e+00, 1.67821523749677e+00, 1.67821460538987e+00, 1.67821451209212e+00, 1.67821450489610e+00, 1.67821437567863e+00, 1.67821350021025e+00, 1.67821028187998e+00, 1.67820170287119e+00, 1.67818292730174e+00, 1.67814696746972e+00, 1.67808442560194e+00, 1.67798332139695e+00, 1.67782901325844e+00, 1.67760421848480e+00, 1.67728913488102e+00, 1.67686166335475e+00, 1.67629772812192e+00, 1.67557168825202e+00, 1.67465683150503e+00, 1.67352593882590e+00, 1.67215190554179e+00, 1.67050840331718e+00, 1.66857056532299e+00, 1.66631567591638e+00, 1.66372384544827e+00, 1.66077865064632e+00, 1.65746772138379e+00, 1.65378325555504e+00, 1.64972244524270e+00, 1.64528779937844e+00, 1.64048735065326e+00, 1.63533473749057e+00, 1.62984915539670e+00, 1.62405517585882e+00, 1.61798243504658e+00, 1.61166519873381e+00, 1.60514181390850e+00, 1.59845406128615e+00, 1.59164642618886e+00, 1.58476530782497e+00, 1.57785818876591e+00, 1.57097278728345e+00, 1.56415621515998e+00, 1.55745416265662e+00, 1.55091013061888e+00, 1.54456472736032e+00, 1.53845504516685e+00, 1.53261412819204e+00, 1.52707054034612e+00, 1.52184803867470e+00, 1.51696535479963e+00, 1.51243608433878e+00, 1.50826868187800e+00, 1.50446655704398e+00, 1.50102826550375e+00, 1.49794778725295e+00, 1.49521488330661e+00, 1.49281552082392e+00, 1.49073235574887e+00, 1.48894526120880e+00, 1.48743188918186e+00, 1.48616825233982e+00, 1.48512931252707e+00, 1.48428956209685e+00, 1.48362358434266e+00, 1.48310657958693e+00, 1.48271484416081e+00, 1.48242619055701e+00, 1.48222029846695e+00, 1.48207898821114e+00, 1.48198641019707e+00, 1.48192914643364e+00, 1.48189622271530e+00, 1.48187903277211e+00, 1.48187117836353e+00, 1.48186823187434e+00, 1.48186743035506e+00, 1.48186731205240e+00, 1.48186730309488e+00, 1.48186714441325e+00, 1.48186606932084e+00, 1.48186211715190e+00, 1.48185158200121e+00, 1.48182852547606e+00, 1.48178436715167e+00, 1.48170756819338e+00, 1.48158342116150e+00, 1.48139395622001e+00, 1.48111797087464e+00, 1.48073118701328e+00, 1.48020653546323e+00, 1.47951456457256e+00, 1.47862396553794e+00, 1.47750220342419e+00, 1.47611623915322e+00, 1.47443332429848e+00, 1.47242184743385e+00, 1.47005220818594e+00, 1.46729769316027e+00, 1.46413532667852e+00, 1.46054666888596e+00, 1.45651853434999e+00, 1.45204360582863e+00, 1.44712092045742e+00, 1.44175620915732e+00, 1.43596207452609e+00, 1.42975799770527e+00, 1.42317017052505e+00, 1.41623115537289e+00, 1.40897938142167e+00, 1.40145849177290e+00, 1.39371656139868e+00, 1.38580521020485e+00, 1.37777863883742e+00, 1.36969261684233e+00, 1.36160345338418e+00, 1.35356697995566e+00, 1.34563757248773e+00, 1.33786723720702e+00, 1.33030478075199e+00, 1.32299508075360e+00, 1.31597846861465e+00, 1.30929023186311e+00, 1.30296023943756e+00, 1.29701268975279e+00, 1.29146597848655e+00, 1.28633268074883e+00, 1.28161964061020e+00, 1.27732815979289e+00, 1.27345427655636e+00, 1.26998912531161e+00, 1.26691936715259e+00, 1.26422768119164e+00, 1.26189330624842e+00, 1.25989262202261e+00, 1.25819975836991e+00, 1.25678722072449e+00, 1.25562651912708e+00, 1.25468878780761e+00, 1.25394538192968e+00, 1.25336843803049e+00, 1.25293138497369e+00, 1.25260939294511e+00, 1.25237974920728e+00, 1.25222215099637e+00, 1.25211890807053e+00, 1.25205504993811e+00, 1.25201833561690e+00, 1.25199916678377e+00, 1.25199040823747e+00, 1.25198712258040e+00, 1.25198622879976e+00, 1.25198609687952e+00, 1.25198608697623e+00, 1.25198591266700e+00, 1.25198473169557e+00, 1.25198039030780e+00, 1.25196881767101e+00, 1.25194349077693e+00, 1.25189498497641e+00, 1.25181062744838e+00, 1.25167426839880e+00, 1.25146618191079e+00, 1.25116310513943e+00, 1.25073842097233e+00, 1.25016248536817e+00, 1.24940309636616e+00, 1.24842609727661e+00, 1.24719610190298e+00, 1.24567732493331e+00, 1.24383449604385e+00, 1.24163383198813e+00, 1.23904403723902e+00, 1.23603730086298e+00, 1.23259025548228e+00, 1.22868486364642e+00, 1.22430919785868e+00, 1.21945808299115e+00, 1.21413357388143e+00, 1.20834524644463e+00, 1.20211028745306e+00, 1.19545337592485e+00, 1.18840635741933e+00, 1.18100772099264e+00, 1.17330189661597e+00, 1.16533839800628e+00, 1.15717084161973e+00, 1.14885587666359e+00, 1.14045206317299e+00, 1.13201873540845e+00, 1.12361488614581e+00, 1.11529810408760e+00, 1.10712359197112e+00, 1.09914328741029e+00, 1.09140510253938e+00, 1.08395229256687e+00, 1.07682295778087e+00, 1.07004967867195e+00, 1.06365927985493e+00, 1.05767271546878e+00, 1.05210506671067e+00, 1.04696564102753e+00, 1.04225816209938e+00, 1.03798103991728e+00, 1.03412771078353e+00, 1.03068703774804e+00, 1.02764376266837e+00, 1.02497900160339e+00, 1.02267077552486e+00, 1.02069456830940e+00, 1.01902390365365e+00, 1.01763093198058e+00, 1.01648701765838e+00, 1.01556331604439e+00, 1.01483132912771e+00, 1.01426342800874e+00, 1.01383333025221e+00, 1.01351652038792e+00, 1.01329060258797e+00, 1.01313557585355e+00, 1.01303402389091e+00, 1.01297121418749e+00, 1.01293510352073e+00, 1.01291625011418e+00, 1.01290763574930e+00, 1.01290440418944e+00, 1.01290352512546e+00, 1.01290339537741e+00, 1.01290338566680e+00, 1.01290321514363e+00, 1.01290205982368e+00, 1.01289781273928e+00, 1.01288649152532e+00, 1.01286171509068e+00, 1.01281426435469e+00, 1.01273174431181e+00, 1.01259836242837e+00, 1.01239483576883e+00, 1.01209843627356e+00, 1.01168318022495e+00, 1.01112016410429e+00, 1.01037804473823e+00, 1.00942365687536e+00, 1.00822275619511e+00, 1.00674087037142e+00, 1.00494423540964e+00, 1.00280078933220e+00, 1.00028119075130e+00, 9.97359826311941e-01, 9.94015768790454e-01, 9.90233647129582e-01, 9.86004391130693e-01, 9.81325817038514e-01, 9.76203025818674e-01, 9.70648593347980e-01, 9.64682540643179e-01, 9.58332082124054e-01, 9.51631160103907e-01, 9.44619783528111e-01, 9.37343197749266e-01, 9.29850919221909e-01, 9.22195673947008e-01, 9.14432281012147e-01, 9.06616522588794e-01, 8.98804039412126e-01, 8.91049286422986e-01, 8.83404577383415e-01, 8.75919240460702e-01, 8.68638899604237e-01, 8.61604889567765e-01, 8.54853806119605e-01, 8.48417187674191e-01, 8.42321320471135e-01, 8.36587156589618e-01, 8.31230332464487e-01, 8.26261275021909e-01, 8.21685382870445e-01, 8.17503270928401e-01, 8.13711068194765e-01, 8.10300759847920e-01, 8.07260566283101e-01, 8.04575352915633e-01, 8.02227065467515e-01, 8.00195185952089e-01, 7.98457204654041e-01, 7.96989103092017e-01, 7.95765842309738e-01, 7.94761849962944e-01, 7.93951498673165e-01, 7.93309567140699e-01, 7.92811674688893e-01, 7.92434679383859e-01, 7.92157029753273e-01, 7.91959060499960e-01, 7.91823223516750e-01, 7.91734246959632e-01, 7.91679217081889e-01, 7.91647579884625e-01, 7.91631062276650e-01, 7.91623515213215e-01, 7.91620684044001e-01, 7.91619913896841e-01, 7.91619800224717e-01, 7.91619791714079e-01, 7.91619642220464e-01, 7.91618629379167e-01, 7.91614906066511e-01, 7.91604981075000e-01, 7.91583260438407e-01, 7.91541662685147e-01, 7.91469323418812e-01, 7.91352402794541e-01, 7.91174007335966e-01, 7.90914235067423e-01, 7.90550350047048e-01, 7.90057089002214e-01, 7.89407098833601e-01, 7.88571499255939e-01, 7.87520559837721e-01, 7.86224475329921e-01, 7.84654217671527e-01, 7.82782437756788e-01, 7.80584385351594e-01, 7.78038811904617e-01, 7.75128818863461e-01, 7.71842613871826e-01, 7.68174139173530e-01, 7.64123540799777e-01, 7.59697453580377e-01, 7.54909085388902e-01, 7.49778093787291e-01, 7.44330258686971e-01, 7.38596964997141e-01, 7.32614518672160e-01, 7.26423327357626e-01, 7.20066982384955e-01, 7.13591281817547e-01, 7.07043234508639e-01, 7.00470082854056e-01, 6.93918377503358e-01, 6.87433131289702e-01, 6.81057072702303e-01, 6.74830012015504e-01, 6.68788326301476e-01, 6.62964563466999e-01, 6.57387160495181e-01, 6.52080267407179e-01, 6.47063666106540e-01, 6.42352772128297e-01, 6.37958707199795e-01, 6.33888431194002e-01, 6.30144923266754e-01, 6.26727403474756e-01, 6.23631587757488e-01, 6.20849970658913e-01, 6.18372131432339e-01, 6.16185060125233e-01, 6.14273500830711e-01, 6.12620309504781e-01, 6.11206823599455e-01, 6.10013240294612e-01, 6.09018999391623e-01, 6.08203166043499e-01, 6.07544807539057e-01, 6.07023357440077e-01, 6.06618959601565e-01, 6.06312784091109e-01, 6.06087306854928e-01, 6.05926545223624e-01, 6.05816242047577e-01, 6.05743992402060e-01, 6.05699308369723e-01, 6.05673619321497e-01, 6.05660207274807e-01, 6.05654079188848e-01, 6.05651780330451e-01, 6.05651154985060e-01, 6.05651062685384e-01, 6.05651055756570e-01, 6.05650933803644e-01, 6.05650107554676e-01, 6.05647070176858e-01, 6.05638973648838e-01, 6.05621254626088e-01, 6.05587320695732e-01, 6.05528309675178e-01, 6.05432933168084e-01, 6.05287413912645e-01, 6.05075524486047e-01, 6.04778732589628e-01, 6.04376455352762e-01, 6.03846421797628e-01, 6.03165138781124e-01, 6.02308451417223e-01, 6.01252184309620e-01, 5.99972845110936e-01, 5.98448367284433e-01, 5.96658864866790e-01, 5.94587368949377e-01, 5.92220513939245e-01, 5.89549141796072e-01, 5.86568794609682e-01, 5.83280070149311e-01, 5.79688821231686e-01, 5.75806187549222e-01, 5.71648457402466e-01, 5.67236765880015e-01, 5.62596644651683e-01, 5.57757445948772e-01, 5.52751668890788e-01, 5.47614219677915e-01, 5.42381638151573e-01, 5.37091321938104e-01, 5.31780776167731e-01, 5.26486912103298e-01, 5.21245412508935e-01, 5.16090175830170e-01, 5.11052845768763e-01, 5.06162428023180e-01, 5.01444992089262e-01, 4.96923453188240e-01, 4.92617427595431e-01, 4.88543153769757e-01, 4.84713471557775e-01, 4.81137852165107e-01, 4.77822472353873e-01, 4.74770327258294e-01, 4.71981377165710e-01, 4.69452724476030e-01, 4.67178817752516e-01, 4.65151680264323e-01, 4.63361160674289e-01, 4.61795203541812e-01, 4.60440137103491e-01, 4.59280975389980e-01, 4.58301731174010e-01, 4.57485735570060e-01, 4.56815959378565e-01, 4.56275330552259e-01, 4.55847041530129e-01, 4.55514839707202e-01, 4.55263294054286e-01, 4.55078030928738e-01, 4.54945932467917e-01, 4.54855291652063e-01, 4.54795919158689e-01, 4.54759198475110e-01, 4.54738087332854e-01, 4.54727065299253e-01, 4.54722029214852e-01, 4.54720140001487e-01, 4.54719626089244e-01, 4.54719550236860e-01, 4.54719544518472e-01, 4.54719443546997e-01, 4.54718759450338e-01, 4.54716244637710e-01, 4.54709541063318e-01, 4.54694870417982e-01, 4.54666774236231e-01, 4.54617914299285e-01, 4.54538942465521e-01, 4.54418447609718e-01, 4.54242985664913e-01, 4.53997196785684e-01, 4.53664011335650e-01, 4.53224943704743e-01, 4.52660469874020e-01, 4.51950481207241e-01, 4.51074803263039e-01, 4.50013764665912e-01, 4.48748797496868e-01, 4.47263047571749e-01, 4.45541970703930e-01, 4.43573889925541e-01, 4.41350488937113e-01, 4.38867218931607e-01, 4.36123599408387e-01, 4.33123398496168e-01, 4.29874684309598e-01, 4.26389745497621e-01, 4.22684885841168e-01, 4.18780103946431e-01, 4.14698674244439e-01, 4.10466649265819e-01, 4.06112305306160e-01, 4.01665554119919e-01, 3.97157342342400e-01, 3.92619058236573e-01, 3.88081962465525e-01, 3.83576656286585e-01, 3.79132597194782e-01, 3.74777668881491e-01, 3.70537809598079e-01, 3.66436700714753e-01, 3.62495515458749e-01, 3.58732726469185e-01, 3.55163969854984e-01, 3.51801962813498e-01, 3.48656471489395e-01, 3.45734325562842e-01, 3.43039475999682e-01, 3.40573092428475e-01, 3.38333696689473e-01, 3.36317329191373e-01, 3.34517744777971e-01, 3.32926634817007e-01, 3.31533872151460e-01, 3.30327775381834e-01, 3.29295388670609e-01, 3.28422772884645e-01, 3.27695303440552e-01, 3.27097969728612e-01, 3.26615670511659e-01, 3.26233499283781e-01, 3.25937013290202e-01, 3.25712479812504e-01, 3.25547093461535e-01, 3.25429158629249e-01, 3.25348231946089e-01, 3.25295220566727e-01, 3.25262433335532e-01, 3.25243583314632e-01, 3.25233741725429e-01, 3.25229244981045e-01, 3.25227558090350e-01, 3.25227099214554e-01, 3.25227031485425e-01, 3.25227026351650e-01, 3.25226935334300e-01, 3.25226318678191e-01, 3.25224051782582e-01, 3.25218009051808e-01, 3.25204784585007e-01, 3.25179457772311e-01, 3.25135413053565e-01, 3.25064221912601e-01, 3.24955593812498e-01, 3.24797401205754e-01, 3.24575781948506e-01, 3.24275320379094e-01, 3.23879305968402e-01, 3.23370065815714e-01, 3.22729364388482e-01, 3.21938860872634e-01, 3.20980611448187e-01, 3.19837600918854e-01, 3.18494285628714e-01, 3.16937127737589e-01, 3.15155099934028e-01, 3.13140139733224e-01, 3.10887533754031e-01, 3.08396214810162e-01, 3.05668958185161e-01, 3.02712467875330e-01, 2.99537348572682e-01, 2.96157964357494e-01, 2.92592190100770e-01, 2.88861066097071e-01, 2.84988370183648e-01, 2.81000124370826e-01, 2.76924054728257e-01, 2.72789023948626e-01, 2.68624455722383e-01, 2.64459768930216e-01, 2.60323837847862e-01, 2.56244492225911e-01, 2.52248068422890e-01, 2.48359019896699e-01, 2.44599592454098e-01, 2.40989566866746e-01, 2.37546068916039e-01, 2.34283444736963e-01, 2.31213197573122e-01, 2.28343980776216e-01, 2.25681641090059e-01, 2.23229305921525e-01, 2.20987508355070e-01, 2.18954344026054e-01, 2.17125654527678e-01, 2.15495232679473e-01, 2.14055045630129e-01, 2.12795472316545e-01, 2.11705552188900e-01, 2.10773242298560e-01, 2.09985679820535e-01, 2.09329446861142e-01, 2.08790834025717e-01, 2.08356098751779e-01, 2.08011713924706e-01, 2.07744601866539e-01, 2.07542348502517e-01, 2.07393392432881e-01, 2.07287183820593e-01, 2.07214308477746e-01, 2.07166573297443e-01, 2.07137050210820e-01, 2.07120077103424e-01, 2.07111215535095e-01, 2.07107166593838e-01, 2.07105647693358e-01, 2.07105234515332e-01, 2.07105173531111e-01, 2.07105168879260e-01, 2.07105086019026e-01, 2.07104524628743e-01, 2.07102460898491e-01, 2.07096959749927e-01, 2.07084920613699e-01, 2.07061864200223e-01, 2.07021768818558e-01, 2.06956963795248e-01, 2.06858086380997e-01, 2.06714107177597e-01, 2.06512427529044e-01, 2.06239050456377e-01, 2.05878824544005e-01, 2.05415757699707e-01, 2.04833394941812e-01, 2.04115251392905e-01, 2.03245288608719e-01, 2.02208419421863e-01, 2.00991023850943e-01, 1.99581456556818e-01, 1.97970525058538e-01, 1.96151917661463e-01, 1.94122560950730e-01, 1.91882888832696e-01, 1.89437008432358e-01, 1.86792752535728e-01, 1.83961613460496e-01, 1.80958558919883e-01, 1.77801736231535e-01, 1.74512076711900e-01, 1.71112816898879e-01, 1.67628957024465e-01, 1.64086679658844e-01, 1.60512752516189e-01, 1.56933939015604e-01, 1.53376438412800e-01, 1.49865374353733e-01, 1.46424346834682e-01, 1.43075058127828e-01, 1.39837018614665e-01, 1.36727334015156e-01, 1.33760571514990e-01, 1.30948699010265e-01, 1.28301089254921e-01, 1.25824579166853e-01, 1.23523573895229e-01, 1.21400185376481e-01, 1.19454395862956e-01, 1.17684238121722e-01, 1.16085985489795e-01, 1.14654346562864e-01, 1.13382660835657e-01, 1.12263092979812e-01, 1.11286824548992e-01, 1.10444242683998e-01, 1.09725125828978e-01, 1.09118826570165e-01, 1.08614451504782e-01, 1.08201037596717e-01, 1.07867723852238e-01, 1.07603916440217e-01, 1.07399444679543e-01, 1.07244704711693e-01, 1.07130787249393e-01, 1.07049585606530e-01, 1.06993880310751e-01, 1.06957396992550e-01, 1.06934834919135e-01, 1.06921864457859e-01, 1.06915092850491e-01, 1.06911998879657e-01, 1.06910838229640e-01, 1.06910522505190e-01, 1.06910475904940e-01, 1.06910472325209e-01, 1.06910408233067e-01, 1.06909973999702e-01, 1.06908377716753e-01, 1.06904122649627e-01, 1.06894810727765e-01, 1.06876978005683e-01, 1.06845969013076e-01, 1.06795856309049e-01, 1.06719410839493e-01, 1.06608127809969e-01, 1.06452311654860e-01, 1.06241222229106e-01, 1.05963282506681e-01, 1.05606345806286e-01, 1.05158017893277e-01, 1.04606026303107e-01, 1.03938626040278e-01, 1.03145027640686e-01, 1.02215830714744e-01, 1.01143443819625e-01, 9.99224701515873e-02, 9.85500383797389e-02, 9.70260591620273e-02, 9.53533905812074e-02, 9.35378998608978e-02, 9.15884140628889e-02, 8.95165586718594e-02, 8.73364895659630e-02, 8.50645302993904e-02, 8.27187323188778e-02, 8.03183801855667e-02, 7.78834666781481e-02, 7.54341635813833e-02, 7.29903129801002e-02, 7.05709611446088e-02, 6.81939529440294e-02, 6.58755996192866e-02, 6.36304272100654e-02, 6.14710074779583e-02, 5.94078682541725e-02, 5.74494761069266e-02, 5.56022812791548e-02, 5.38708130627814e-02, 5.22578131021796e-02, 5.07643944136482e-02, 4.93902149668246e-02, 4.81336562680328e-02, 4.69919992871488e-02, 4.59615920752985e-02, 4.50380053645754e-02, 4.42161741989795e-02, 4.34905251345197e-02, 4.28550897180866e-02, 4.23036057904251e-02, 4.18296086623335e-02, 4.14265144058996e-02, 4.10876974179946e-02, 4.08065640945308e-02, 4.05766239522432e-02, 4.03915589070722e-02, 4.02452907264628e-02, 4.01320459811729e-02, 4.00464171937054e-02, 3.99834183732775e-02, 3.99385327893931e-02, 3.99077507012058e-02, 3.98875948431998e-02, 3.98751317645574e-02, 3.98679676049269e-02, 3.98642275212171e-02, 3.98625187039079e-02, 3.98618776768958e-02, 3.98617033028894e-02, 3.98616775656830e-02, 3.98616755735173e-02, 3.98616397090279e-02, 3.98613967223118e-02, 3.98605034845922e-02, 3.98581224957277e-02, 3.98529120386362e-02, 3.98429344583180e-02, 3.98255867056057e-02, 3.97975571012443e-02, 3.97548120369970e-02, 3.96926159719500e-02, 3.96055875215833e-02, 3.94877935868361e-02, 3.93328822537100e-02, 3.91342535636574e-02, 3.88852652115019e-02, 3.85794678351568e-02, 3.82108619575863e-02, 3.77741660381627e-02, 3.72650827619381e-02, 3.66805489527597e-02, 3.60189536502458e-02, 3.52803092044536e-02, 3.44663618819483e-02, 3.35806314692288e-02, 3.26283735629338e-02, 3.16164633381750e-02, 3.05532051266391e-02, 2.94480775592287e-02, 2.83114287563186e-02, 2.71541395648496e-02, 2.59872747657761e-02, 2.48217423227888e-02, 2.36679791498777e-02, 2.25356787884172e-02, 2.14335722192081e-02, 2.03692683051748e-02, 1.93491555987149e-02, 1.83783629293088e-02, 1.74607726726463e-02, 1.65990781076102e-02, 1.57948748586494e-02, 1.50487760346567e-02, 1.43605411504910e-02, 1.37292100348012e-02, 1.31532344507492e-02, 1.26306018629241e-02, 1.21589474908066e-02, 1.17356523623713e-02, 1.13579264355264e-02, 1.10228769476983e-02, 1.07275629758725e-02, 1.04690377558179e-02, 1.02443806494616e-02, 1.00507207996205e-02, 9.88525450766153e-03, 9.74525824509964e-03, 9.62809899240180e-03, 9.53124331118715e-03, 9.45226622044104e-03, 9.38886058306698e-03, 9.33884733600321e-03, 9.30018653594811e-03, 9.27098886479749e-03, 9.24952696518355e-03, 9.23424577582388e-03, 9.22377092316395e-03, 9.21691420803850e-03, 9.21267530345315e-03, 9.21023894329354e-03, 9.20896711458585e-03, 9.20838604455643e-03, 9.20816807143909e-03, 9.20810877806155e-03, 9.20810002650496e-03, 9.20809934360767e-03, 9.20808697860553e-03, 9.20800320420989e-03, 9.20769524536191e-03, 9.20687437618062e-03, 9.20507810692861e-03, 9.20163873711122e-03, 9.19565983642556e-03, 9.18600222765509e-03, 9.17128104250242e-03, 9.14987536269041e-03, 9.11995179366332e-03, 9.07950299238225e-03, 9.02640163524304e-03, 8.95846953557089e-03, 8.87356060044087e-03, 8.76965509259131e-03, 8.64496132158554e-03, 8.49801956448837e-03, 8.32780188512316e-03, 8.13380077774140e-03, 7.91609939182814e-03, 7.67541664133509e-03, 7.41312182531773e-03, 7.13121544181283e-03, 6.83227549862253e-03, 6.51937154161342e-03, 6.19595148783025e-03, 5.86570880073012e-03, 5.53243925242941e-03, 5.19989725686165e-03, 4.87166144310969e-03, 4.55101784145739e-03, 4.24086698759596e-03, 3.94365872131863e-03, 3.66135581028604e-03, 3.39542508987467e-03, 3.14685283007377e-03, 2.91617967645763e-03, 2.70354981604659e-03, 2.50876894535090e-03, 2.33136604799689e-03, 2.17065476316503e-03, 2.02579107945811e-03, 1.89582508173536e-03, 1.77974540797909e-03, 1.67651587259754e-03, 1.58510434447434e-03, 1.50450442038208e-03, 1.43375071745555e-03, 1.37192875038727e-03, 1.31818039649501e-03, 1.27170591951953e-03, 1.23176344738130e-03, 1.19766669796939e-03, 1.16878163258360e-03, 1.14452259816811e-03, 1.12434840392273e-03, 1.10775866944746e-03, 1.09429068147593e-03, 1.08351690406838e-03, 1.07504320192531e-03, 1.06850775846157e-03, 1.06358060083145e-03, 1.05996358586280e-03, 1.05739065690846e-03, 1.05562815483462e-03, 1.05447495868227e-03, 1.05376224347942e-03, 1.05335267283868e-03, 1.05313888930669e-03, 1.05304122109007e-03, 1.05300458418922e-03, 1.05299461823538e-03, 1.05299314728780e-03, 1.05299303153106e-03, 1.05299092304922e-03, 1.05297663786530e-03, 1.05292412537517e-03, 1.05278415677712e-03, 1.05247789179197e-03, 1.05189156025198e-03, 1.05087255836334e-03, 1.04922728034595e-03, 1.04672102486703e-03, 1.04308032051326e-03, 1.03799799070181e-03, 1.03114120972680e-03, 1.02216267352961e-03, 1.01071481191140e-03, 9.96466703274024e-04, 9.79123031845885e-04, 9.58444079636411e-04, 9.34265415858837e-04, 9.06515691785348e-04, 8.75230829687101e-04, 8.40562963386970e-04, 8.02782776684552e-04, 7.62274392512419e-04, 7.19522647922632e-04, 6.75093366645063e-04, 6.29608003313242e-04, 5.83714664820878e-04, 5.38057913163081e-04, 4.93249855333611e-04, 4.49844813957156e-04, 4.08319384889166e-04, 3.69059008149542e-04, 3.32351418937948e-04, 2.98386625552816e-04, 2.67262482506192e-04, 2.38994552599348e-04, 2.13528795562690e-04, 1.90755655161690e-04, 1.70524290362017e-04, 1.52655955410586e-04, 1.36955832782171e-04, 1.23222923606114e-04, 1.11257867300596e-04, 1.00868766463869e-04, 9.18752209004836e-05, 8.41108341445018e-05, 7.74244713146466e-05, 7.16805428238043e-05, 6.67585755416145e-05, 6.25523094950342e-05, 5.89685198053580e-05, 5.59257135095415e-05, 5.33528005979278e-05, 5.11877995867938e-05, 4.93766148859098e-05, 4.78719123995219e-05, 4.66321140371040e-05, 4.56205255703335e-05, 4.48046041963135e-05, 4.41553635690539e-05, 4.36469069193812e-05, 4.32560740546428e-05, 4.29621852036099e-05, 4.27468630854744e-05, 4.25939136474707e-05, 4.24892454382597e-05, 4.24208078151356e-05, 4.23785293133250e-05, 4.23542397867127e-05, 4.23415632249124e-05, 4.23357722856288e-05, 4.23336000803243e-05, 4.23330092043095e-05, 4.23329219929602e-05, 4.23329150680551e-05, 4.23327881564628e-05, 4.23319283207046e-05, 4.23287676050155e-05, 4.23203433166375e-05, 4.23019121345014e-05, 4.22666340089695e-05, 4.22053469399941e-05, 4.21064566813456e-05, 4.19559681728021e-05, 4.17376869061847e-05, 4.14336168923501e-05, 4.10245757149763e-05, 4.04910349154353e-05, 3.98141748709779e-05, 3.89771178492815e-05, 3.79662731171404e-05, 3.67726974748184e-05, 3.53933483192570e-05, 3.38320896133837e-05, 3.21003091683369e-05, 3.02170227348686e-05, 2.82083800551280e-05, 2.61065515949621e-05, 2.39480585688678e-05, 2.17717001957940e-05, 1.96163068117983e-05, 1.75185760313834e-05, 1.55112134101465e-05, 1.36215110595472e-05, 1.18704027347594e-05, 1.02719849688349e-05, 8.83350523395166e-06, 7.55584187772662e-06, 6.43446082970733e-06, 5.46070966289669e-06, 4.62318390193062e-06, 3.90890356882778e-06, 3.30420513439122e-06, 2.79545014336691e-06, 2.36968019273753e-06, 2.01517083953107e-06, 1.72166049106082e-06, 1.48010957046196e-06, 1.28218227768429e-06, 1.11992230156233e-06, 9.85977194334353e-07, 8.74244611814845e-07, 7.80402898081789e-07, 7.01842958417132e-07, 6.36974408996238e-07, 5.84331587338838e-07, 5.42013535831387e-07, 5.07712781816494e-07, 4.79167100442268e-07, 4.54642274272388e-07, 4.33146933209727e-07, 4.14329934361746e-07, 3.98195762807418e-07, 3.84813174436616e-07, 3.74130707747611e-07, 3.65924237355486e-07, 3.59838513874777e-07, 3.55464539136818e-07, 3.52408033643670e-07, 3.50329691922724e-07, 3.48958166362934e-07, 3.48085751999960e-07, 3.47556994295672e-07, 3.47256863264533e-07, 3.47101305253616e-07, 3.47030491898437e-07, 3.47003970002646e-07, 3.46996759326146e-07, 3.46995695288399e-07, 3.46995585820409e-07, 3.46993269895179e-07, 3.46977581333618e-07, 3.46919937221912e-07, 3.46766499292883e-07, 3.46431819675530e-07, 3.45795118983361e-07, 3.44701047483298e-07, 3.42967405142502e-07, 3.40402372067672e-07, 3.36832568281764e-07, 3.32139732828196e-07, 3.26297457942104e-07, 3.19390990988211e-07, 3.11596496382904e-07, 3.03099050302791e-07, 2.93950670249288e-07, 2.83913655252377e-07, 2.72383606558790e-07, 2.58501134942080e-07, 2.41503353981543e-07, 2.21236655661983e-07, 1.98599345191397e-07, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 2.22044604925031e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00],
[ 3.00000000000000e+00, 2.99999999935979e+00, 2.99999997956324e+00, 2.99999984543851e+00, 2.99999935237897e+00, 2.99999803804152e+00, 2.99999516153577e+00, 2.99998965226006e+00, 2.99998007025930e+00, 2.99996457964287e+00, 2.99994093621856e+00, 2.99990649008146e+00, 2.99985820345737e+00, 2.99979268365246e+00, 2.99970623051673e+00, 2.99959489740298e+00, 2.99945456420594e+00, 2.99928102071070e+00, 2.99907005817431e+00, 2.99881756681647e+00, 2.99851963671183e+00, 2.99817265945925e+00, 2.99777342795551e+00, 2.99731923162079e+00, 2.99680794451035e+00, 2.99623810389611e+00, 2.99560897711088e+00, 2.99492061470974e+00, 2.99417388831321e+00, 2.99337051184770e+00, 2.99251304528405e+00, 2.99160488038692e+00, 2.99065020841798e+00, 2.98965397017564e+00, 2.98862178919391e+00, 2.98755988935173e+00, 2.98647499855213e+00, 2.98537424050515e+00, 2.98426501697949e+00, 2.98315488316392e+00, 2.98205141899071e+00, 2.98096209941167e+00, 2.97989416667619e+00, 2.97885450763577e+00, 2.97784953899000e+00, 2.97688510319584e+00, 2.97596637749055e+00, 2.97509779813526e+00, 2.97428300158135e+00, 2.97352478380788e+00, 2.97282507858785e+00, 2.97218495493099e+00, 2.97160463343465e+00, 2.97108352076933e+00, 2.97062026104626e+00, 2.97021280237472e+00, 2.96985847653009e+00, 2.96955408933035e+00, 2.96929601906708e+00, 2.96908032016431e+00, 2.96890282914741e+00, 2.96875926999677e+00, 2.96864535603510e+00, 2.96855688565058e+00, 2.96848982938308e+00, 2.96844040619209e+00, 2.96840514707123e+00, 2.96838094456732e+00, 2.96836508718845e+00, 2.96835527813522e+00, 2.96834963824922e+00, 2.96834669353059e+00, 2.96834534802097e+00, 2.96834484326684e+00, 2.96834470596082e+00, 2.96834468569471e+00, 2.96834468303623e+00, 2.96834462109669e+00, 2.96834420144658e+00, 2.96834265875895e+00, 2.96833854644765e+00, 2.96832954638717e+00, 2.96831230878526e+00, 2.96828232803993e+00, 2.96823385933936e+00, 2.96815987954560e+00, 2.96805209457914e+00, 2.96790099413635e+00, 2.96769595317086e+00, 2.96742537820000e+00, 2.96707689520252e+00, 2.96663757469361e+00, 2.96609418853067e+00, 2.96543349214219e+00, 2.96464252519899e+00, 2.96370892326631e+00, 2.96262123268554e+00, 2.96136922082612e+00, 2.95994417390798e+00, 2.95833917480797e+00, 2.95654935361660e+00, 2.95457210419399e+00, 2.95240726058077e+00, 2.95005722785098e+00, 2.94752706285171e+00, 2.94482450126396e+00, 2.94195992854199e+00, 2.93894629354030e+00, 2.93579896500560e+00, 2.93253553256993e+00, 2.92917555539347e+00, 2.92574026312186e+00, 2.92225221528412e+00, 2.91873492660039e+00, 2.91521246682868e+00, 2.91170904469904e+00, 2.90824858611166e+00, 2.90485431707951e+00, 2.90154836185667e+00, 2.89835136631046e+00, 2.89528215588580e+00, 2.89235743650375e+00, 2.88959154547734e+00, 2.88699625806661e+00, 2.88458065368851e+00, 2.88235104410234e+00, 2.88031096416474e+00, 2.87846122404254e+00, 2.87680002013573e+00, 2.87532310043879e+00, 2.87402397869566e+00, 2.87289419051165e+00, 2.87192358360329e+00, 2.87110063361312e+00, 2.87041277640734e+00, 2.86984674751886e+00, 2.86938891939887e+00, 2.86902562739440e+00, 2.86874347586643e+00, 2.86852961658716e+00, 2.86837199248364e+00, 2.86825954090031e+00, 2.86818235180332e+00, 2.86813177770904e+00, 2.86810049355093e+00, 2.86808250616267e+00, 2.86807311451457e+00, 2.86806882325361e+00, 2.86806721343092e+00, 2.86806677551799e+00, 2.86806671088288e+00, 2.86806670517683e+00, 2.86806659336956e+00, 2.86806583585777e+00, 2.86806305114786e+00, 2.86805562800731e+00, 2.86803938200462e+00, 2.86800826650900e+00, 2.86795414871169e+00, 2.86786665931498e+00, 2.86773312235902e+00, 2.86753856928953e+00, 2.86726583888575e+00, 2.86689576214871e+00, 2.86640742877470e+00, 2.86577852948679e+00, 2.86498576633845e+00, 2.86400532119598e+00, 2.86281337099684e+00, 2.86138663709423e+00, 2.85970295504799e+00, 2.85774185060117e+00, 2.85548510727770e+00, 2.85291731102706e+00, 2.85002635760749e+00, 2.84680390892755e+00, 2.84324578535439e+00, 2.83935228205664e+00, 2.83512839880265e+00, 2.83058397430468e+00, 2.82573371820999e+00, 2.82059713619480e+00, 2.81519834629913e+00, 2.80956578759786e+00, 2.80373182544738e+00, 2.79773226075784e+00, 2.79160575386941e+00, 2.78539317649617e+00, 2.77913690768473e+00, 2.77288009167383e+00, 2.76666587682803e+00, 2.76053665538733e+00, 2.75453332360674e+00, 2.74869458098623e+00, 2.74305628578388e+00, 2.73765088196753e+00, 2.73250691031744e+00, 2.72764861367534e+00, 2.72309564347136e+00, 2.71886287176260e+00, 2.71496031018090e+00, 2.71139313448348e+00, 2.70816181088184e+00, 2.70526231802282e+00, 2.70268645643324e+00, 2.70042223542610e+00, 2.69845432591272e+00, 2.69676456627904e+00, 2.69533250748035e+00, 2.69413598280139e+00, 2.69315168733817e+00, 2.69235575220264e+00, 2.69172429874804e+00, 2.69123395877142e+00, 2.69086234767112e+00, 2.69058847890956e+00, 2.69039310983083e+00, 2.69025901087039e+00, 2.69017115241873e+00, 2.69011680600133e+00, 2.69008555894224e+00, 2.69006924421393e+00, 2.69006178966348e+00, 2.69005899316924e+00, 2.69005823245170e+00, 2.69005812017127e+00, 2.69005811116499e+00, 2.69005794495340e+00, 2.69005681884403e+00, 2.69005267912582e+00, 2.69004164399549e+00, 2.69001749306314e+00, 2.68997123799794e+00, 2.68989079022308e+00, 2.68976073884023e+00, 2.68956224885916e+00, 2.68927308640676e+00, 2.68886777397750e+00, 2.68831787505714e+00, 2.68759240369907e+00, 2.68665835095804e+00, 2.68548131658769e+00, 2.68402623117920e+00, 2.68225815104573e+00, 2.68014310571294e+00, 2.67764897592033e+00, 2.67474637861569e+00, 2.67140953456809e+00, 2.66761709395790e+00, 2.66335289564499e+00, 2.65860663678953e+00, 2.65337443112616e+00, 2.64765923649323e+00, 2.64147113520947e+00, 2.63482745456458e+00, 2.62775271901093e+00, 2.62027843052405e+00, 2.61244267890041e+00, 2.60428958928114e+00, 2.59586861968043e+00, 2.58723372647227e+00, 2.57844242035805e+00, 2.56955473902722e+00, 2.56063216531700e+00, 2.55173652103014e+00, 2.54292886663300e+00, 2.53426843587534e+00, 2.52581163208476e+00, 2.51761110969836e+00, 2.50971496075638e+00, 2.50216602186562e+00, 2.49500131280610e+00, 2.48825161372766e+00, 2.48194118394067e+00, 2.47608762176441e+00, 2.47070186181636e+00, 2.46578830351137e+00, 2.46134506235619e+00, 2.45736433380866e+00, 2.45383285794918e+00, 2.45073247191186e+00, 2.44804073588726e+00, 2.44573161750390e+00, 2.44377621851346e+00, 2.44214352696534e+00, 2.44080117750427e+00, 2.43971620212022e+00, 2.43885575369407e+00, 2.43818778508294e+00, 2.43768166733485e+00, 2.43730873195026e+00, 2.43704272393251e+00, 2.43686015467072e+00, 2.43674054643004e+00, 2.43666656330748e+00, 2.43662402684551e+00, 2.43660181795976e+00, 2.43659167030372e+00, 2.43658786352995e+00, 2.43658682799163e+00, 2.43658667514822e+00, 2.43658666335946e+00, 2.43658645167122e+00, 2.43658501745085e+00, 2.43657974508482e+00, 2.43656569073711e+00, 2.43653493235119e+00, 2.43647602333641e+00, 2.43637357059566e+00, 2.43620795430290e+00, 2.43595520312903e+00, 2.43558703450175e+00, 2.43507106502695e+00, 2.43437119145835e+00, 2.43344813765508e+00, 2.43226015790533e+00, 2.43076388193063e+00, 2.42891528195653e+00, 2.42667073759473e+00, 2.42398817010020e+00, 2.42082821402210e+00, 2.41715539153751e+00, 2.41293925301073e+00, 2.40815544670874e+00, 2.40278668125046e+00, 2.39682354636094e+00, 2.39026516088412e+00, 2.38311962176400e+00, 2.37540423374727e+00, 2.36714550672592e+00, 2.35837891567112e+00, 2.34914842667049e+00, 2.33950580125403e+00, 2.32950969951431e+00, 2.31922461001127e+00, 2.30871964064983e+00, 2.29806720925067e+00, 2.28734167514107e+00, 2.27661795365927e+00, 2.26597015404118e+00, 2.25547027794178e+00, 2.24518701116933e+00, 2.23518463549709e+00, 2.22552208113121e+00, 2.21625213400956e+00, 2.20742080599261e+00, 2.19906687050566e+00, 2.19122156152132e+00, 2.18390843004051e+00, 2.17714334944717e+00, 2.17093465919428e+00, 2.16528343508480e+00, 2.16018387375722e+00, 2.15562377867806e+00, 2.15158513479770e+00, 2.14804475888582e+00, 2.14497501231465e+00, 2.14234456263892e+00, 2.14011917972003e+00, 2.13826255139876e+00, 2.13673710291962e+00, 2.13550480356673e+00, 2.13452794342290e+00, 2.13376986294991e+00, 2.13319561833914e+00, 2.13277256640062e+00, 2.13247085421127e+00, 2.13226380084948e+00, 2.13212816127283e+00, 2.13204426566833e+00, 2.13199603129777e+00, 2.13197084781779e+00, 2.13195934109703e+00, 2.13195502449957e+00, 2.13195385027764e+00, 2.13195367696484e+00, 2.13195366384212e+00, 2.13195343137438e+00, 2.13195185637025e+00, 2.13194606647433e+00, 2.13193063262886e+00, 2.13189685548399e+00, 2.13183216609724e+00, 2.13171966409741e+00, 2.13153781320141e+00, 2.13126031020027e+00, 2.13085613928075e+00, 2.13028981881837e+00, 2.12952184256548e+00, 2.12850931148377e+00, 2.12720674639031e+00, 2.12556706520947e+00, 2.12354270212119e+00, 2.12108683950730e+00, 2.11815471761975e+00, 2.11470498167455e+00, 2.11070102198310e+00, 2.10611226013732e+00, 2.10091533349750e+00, 2.09509513155304e+00, 2.08864564129052e+00, 2.08157056453165e+00, 2.07388367815821e+00, 2.06560891791913e+00, 2.05678017764934e+00, 2.04744082761403e+00, 2.03764296761488e+00, 2.02744644170273e+00, 2.01691765109759e+00, 2.00612820958337e+00, 1.99515349073852e+00, 1.98407111861621e+00, 1.97295945286987e+00, 1.96189611604066e+00, 1.95095660520501e+00, 1.94021302299938e+00, 1.92973295486224e+00, 1.91957851083944e+00, 1.90980554211018e+00, 1.90046303501972e+00, 1.89159267921567e+00, 1.88322860168098e+00, 1.87539725508861e+00, 1.86811744688473e+00, 1.86140049465358e+00, 1.85525049337524e+00, 1.84966468087323e+00, 1.84463388877923e+00, 1.84014306745817e+00, 1.83617187432291e+00, 1.83269531566234e+00, 1.82968443240981e+00, 1.82710702015075e+00, 1.82492837312920e+00, 1.82311204113689e+00, 1.82162058707183e+00, 1.82041633178989e+00, 1.81946207181242e+00, 1.81872175467413e+00, 1.81816109636319e+00, 1.81774812555905e+00, 1.81745364031133e+00, 1.81725156446977e+00, 1.81711919355927e+00, 1.81703732282268e+00, 1.81699025370135e+00, 1.81696567892046e+00, 1.81695445039560e+00, 1.81695023817137e+00, 1.81694909234265e+00, 1.81694892322068e+00, 1.81694891052464e+00, 1.81694868705991e+00, 1.81694717305310e+00, 1.81694160739710e+00, 1.81692677134934e+00, 1.81689430275633e+00, 1.81683212049962e+00, 1.81672398185188e+00, 1.81654919250725e+00, 1.81628248567433e+00, 1.81589408275210e+00, 1.81534994368740e+00, 1.81461221007285e+00, 1.81363983834767e+00, 1.81238941412504e+00, 1.81081613177964e+00, 1.80887491617931e+00, 1.80652165613363e+00, 1.80371451216143e+00, 1.80041525503664e+00, 1.79659058679283e+00, 1.79221339298753e+00, 1.78726387452157e+00, 1.78173050952846e+00, 1.77561080096054e+00, 1.76891177344004e+00, 1.76165019339627e+00, 1.75385249890623e+00, 1.74555443920595e+00, 1.73680043760587e+00, 1.72764270451958e+00, 1.71814013854329e+00, 1.70835706218506e+00, 1.69836184436420e+00, 1.68822546390338e+00, 1.67802006696654e+00, 1.66781756710844e+00, 1.65768832990006e+00, 1.64769997574723e+00, 1.63791632536366e+00, 1.62839650319704e+00, 1.61919420562757e+00, 1.61035713348055e+00, 1.60192658263586e+00, 1.59393718239922e+00, 1.58641676876888e+00, 1.57938637860104e+00, 1.57286035067353e+00, 1.56684652045476e+00, 1.56134649668653e+00, 1.55635600939266e+00, 1.55186532038831e+00, 1.54785968859633e+00, 1.54431988334893e+00, 1.54122273929077e+00, 1.53854174648059e+00, 1.53624766883541e+00, 1.53430918323585e+00, 1.53269353050632e+00, 1.53136716822136e+00, 1.53029641400834e+00, 1.52944806686866e+00, 1.52878999317602e+00, 1.52829166357333e+00, 1.52792462709898e+00, 1.52766290961300e+00, 1.52748332500536e+00, 1.52736568974148e+00, 1.52729293397166e+00, 1.52725110558778e+00, 1.52722926709978e+00, 1.52721928884528e+00, 1.52721554564638e+00, 1.52721452740446e+00, 1.52721437711405e+00, 1.52721436586601e+00, 1.52721416834477e+00, 1.52721283010902e+00, 1.52720791060905e+00, 1.52719479699903e+00, 1.52716609803565e+00, 1.52711113551795e+00, 1.52701555345612e+00, 1.52686106264671e+00, 1.52662533542573e+00, 1.52628206201418e+00, 1.52580117598629e+00, 1.52514925189557e+00, 1.52429007293130e+00, 1.52318536064235e+00, 1.52179565233580e+00, 1.52008130491889e+00, 1.51800359700969e+00, 1.51552589451588e+00, 1.51261483909032e+00, 1.50924151448943e+00, 1.50538254345135e+00, 1.50102106776305e+00, 1.49614756702914e+00, 1.49076047739422e+00, 1.48486657993521e+00, 1.47848113917001e+00, 1.47162778439582e+00, 1.46433813944753e+00, 1.45665121892047e+00, 1.44861261992368e+00, 1.44027354714977e+00, 1.43168971484913e+00, 1.42292017188291e+00, 1.41402609545163e+00, 1.40506959572164e+00, 1.39611256801860e+00, 1.38721562228974e+00, 1.37843711195803e+00, 1.36983227683648e+00, 1.36145250803149e+00, 1.35334473713980e+00, 1.34555094773193e+00, 1.33810780413108e+00, 1.33104639071321e+00, 1.32439205415058e+00, 1.31816434093542e+00, 1.31237702288895e+00, 1.30703820394969e+00, 1.30215050215378e+00, 1.29771130123188e+00, 1.29371306655980e+00, 1.29014372026379e+00, 1.28698707007706e+00, 1.28422328607743e+00, 1.28182941873114e+00, 1.27977995076578e+00, 1.27804737435174e+00, 1.27660278395573e+00, 1.27541647412571e+00, 1.27445853046814e+00, 1.27369940128637e+00, 1.27311043686607e+00, 1.27266438331253e+00, 1.27233581824517e+00, 1.27210151658747e+00, 1.27194073617594e+00, 1.27183541493069e+00, 1.27177027382255e+00, 1.27173282274181e+00, 1.27171326949283e+00, 1.27170433535832e+00, 1.27170098384088e+00, 1.27170007214545e+00, 1.27169993758106e+00, 1.27169992750622e+00, 1.27169975053655e+00, 1.27169855154058e+00, 1.27169414389807e+00, 1.27168239469152e+00, 1.27165668158732e+00, 1.27160743694690e+00, 1.27152179706867e+00, 1.27138337210521e+00, 1.27117214948012e+00, 1.27086454051673e+00, 1.27043357643939e+00, 1.26984925587040e+00, 1.26907904138950e+00, 1.26808849766280e+00, 1.26684205816063e+00, 1.26530390173292e+00, 1.26343891455524e+00, 1.26121370755049e+00, 1.25859765475793e+00, 1.25556391472634e+00, 1.25209039530068e+00, 1.24816062251895e+00, 1.24376447696007e+00, 1.23889876580449e+00, 1.23356760587126e+00, 1.22778260153063e+00, 1.22156281101327e+00, 1.21493450447831e+00, 1.20793072646934e+00, 1.20059068336806e+00, 1.19295898261328e+00, 1.18508475449144e+00, 1.17702068918808e+00, 1.16882202172126e+00, 1.16054549573591e+00, 1.15224833439339e+00, 1.14398724321611e+00, 1.13581746614183e+00, 1.12779191248797e+00, 1.11996036917168e+00, 1.11236880941258e+00, 1.10505880622346e+00, 1.09806705620874e+00, 1.09142501649355e+00, 1.08515865499990e+00, 1.07928831181997e+00, 1.07382866720204e+00, 1.06878880976649e+00, 1.06417239709467e+00, 1.05997789982914e+00, 1.05619891987780e+00, 1.05282457315565e+00, 1.04983992741209e+00, 1.04722648593502e+00, 1.04496270815629e+00, 1.04302455828827e+00, 1.04138607302561e+00, 1.04001993902785e+00, 1.03889807039204e+00, 1.03799217571181e+00, 1.03727430371750e+00, 1.03671735603763e+00, 1.03629555545021e+00, 1.03598485822386e+00, 1.03576329986919e+00, 1.03561126487607e+00, 1.03551167279992e+00, 1.03545007532825e+00, 1.03541466161581e+00, 1.03539617209427e+00, 1.03538772399421e+00, 1.03538455480647e+00, 1.03538369270926e+00, 1.03538356546547e+00, 1.03538355591345e+00, 1.03538338778985e+00, 1.03538224872733e+00, 1.03537806140729e+00, 1.03536689950530e+00, 1.03534247173192e+00, 1.03529568876577e+00, 1.03521433011145e+00, 1.03508282569425e+00, 1.03488216444361e+00, 1.03458993919428e+00, 1.03418053391064e+00, 1.03362545547708e+00, 1.03289380807906e+00, 1.03195290354438e+00, 1.03076899599812e+00, 1.02930812395048e+00, 1.02753703769725e+00, 1.02542418494848e+00, 1.02294072323913e+00, 1.02006152426300e+00, 1.01676613314131e+00, 1.01303964506122e+00, 1.00887346287971e+00, 1.00426590224058e+00, 9.99222615425769e-01, 9.93756811361409e-01, 9.87889256641286e-01, 9.81648050779061e-01, 9.75068177812061e-01, 9.68190845522698e-01, 9.61062632603842e-01, 9.53734472742898e-01, 9.46260512449117e-01, 9.38696886012642e-01, 9.31100455668765e-01, 9.23527567198843e-01, 9.16032870249056e-01, 9.08668248243770e-01, 9.01481894971707e-01, 8.94517564301290e-01, 8.87814007104632e-01, 8.81404596747887e-01, 8.75317132927611e-01, 8.69573804466448e-01, 8.64191285732206e-01, 8.59180938837123e-01, 8.54549094383578e-01, 8.50297386522640e-01, 8.46423122563534e-01, 8.42919672404002e-01, 8.39776867887498e-01, 8.36981406306313e-01, 8.34517255371527e-01, 8.32366058974345e-01, 8.30507544027856e-01, 8.28919928750661e-01, 8.27580332123687e-01, 8.26465183119894e-01, 8.25550626868642e-01, 8.24812923354907e-01, 8.24228832736046e-01, 8.23775980037591e-01, 8.23433190999158e-01, 8.23180790294110e-01, 8.23000853325150e-01, 8.22877403349194e-01, 8.22796546812580e-01, 8.22746541439995e-01, 8.22717793731426e-01, 8.22702784957191e-01, 8.22695927351772e-01, 8.22693354831676e-01, 8.22692655044255e-01, 8.22692551757113e-01, 8.22692543970464e-01, 8.22692406479020e-01, 8.22691474954592e-01, 8.22688050577260e-01, 8.22678922482980e-01, 8.22658946066913e-01, 8.22620689587531e-01, 8.22554163586951e-01, 8.22446646160525e-01, 8.22282615367230e-01, 8.22043797980556e-01, 8.21709341215368e-01, 8.21256110900991e-01, 8.20659115696811e-01, 8.19892052311358e-01, 8.18927961322648e-01, 8.17739977254309e-01, 8.16302150318627e-01, 8.14590311108720e-01, 8.12582944043558e-01, 8.10262031161897e-01, 8.07613825573849e-01, 8.04629514111901e-01, 8.01305731934368e-01, 7.97644898235560e-01, 7.93655351677106e-01, 7.89351276127600e-01, 7.84752420801748e-01, 7.79883632570722e-01, 7.74774230502183e-01, 7.69457262044356e-01, 7.63968685489299e-01, 7.58346523839336e-01, 7.52630031133830e-01, 7.46858904608623e-01, 7.41072566251797e-01, 7.35309527094446e-01, 7.29606838473817e-01, 7.23999627598690e-01, 7.18520710489107e-01, 7.13200273627000e-01, 7.08065615895304e-01, 7.03140943860474e-01, 6.98447215439240e-01, 6.94002028912532e-01, 6.89819555751968e-01, 6.85910516661187e-01, 6.82282200619750e-01, 6.78938526659795e-01, 6.75880147748233e-01, 6.73104595623331e-01, 6.70606464843698e-01, 6.68377633708435e-01, 6.66407519121139e-01, 6.64683361893858e-01, 6.63190538403941e-01, 6.61912893910891e-01, 6.60833092205520e-01, 6.59932975609096e-01, 6.59193928692832e-01, 6.58597238491793e-01, 6.58124443498961e-01, 6.57757663408248e-01, 6.57479901493197e-01, 6.57275311716223e-01, 6.57129423201309e-01, 6.57029315589909e-01, 6.56963740028010e-01, 6.56923182068100e-01, 6.56899864553358e-01, 6.56887690501741e-01, 6.56882128029616e-01, 6.56880041346616e-01, 6.56879473717522e-01, 6.56879389936631e-01, 6.56879383586155e-01, 6.56879270997748e-01, 6.56878508194543e-01, 6.56875704046486e-01, 6.56868229195925e-01, 6.56851870559782e-01, 6.56820541398626e-01, 6.56766058422418e-01, 6.56677995949996e-01, 6.56543625457636e-01, 6.56347946913201e-01, 6.56073816054650e-01, 6.55702169190789e-01, 6.55212344147730e-01, 6.54582492657515e-01, 6.53790075836941e-01, 6.52812430550621e-01, 6.51627390581862e-01, 6.50213942907159e-01, 6.48552896296933e-01, 6.46627537278533e-01, 6.44424247503894e-01, 6.41933056992467e-01, 6.39148109673342e-01, 6.36068021074434e-01, 6.32696112683031e-01, 6.29040513073357e-01, 6.25114121919355e-01, 6.20934439021803e-01, 6.16523266064839e-01, 6.11906293663028e-01, 6.07112590171494e-01, 6.02174011624721e-01, 5.97124554042645e-01, 5.91999670238668e-01, 5.86835573242262e-01, 5.81668547571250e-01, 5.76534287924996e-01, 5.71467282510346e-01, 5.66500255283923e-01, 5.61663678066010e-01, 5.56985359957998e-01, 5.52490118000979e-01, 5.48199529765495e-01, 5.44131765748114e-01, 5.40301497203099e-01, 5.36719873425903e-01, 5.33394561532040e-01, 5.30329841383019e-01, 5.27526748398866e-01, 5.24983257434194e-01, 5.22694501541336e-01, 5.20653020163196e-01, 5.18849031970482e-01, 5.17270728088056e-01, 5.15904581779205e-01, 5.14735670743101e-01, 5.13748008030506e-01, 5.12924877226056e-01, 5.12249167038182e-01, 5.11703699854730e-01, 5.11271548251330e-01, 5.10936332972117e-01, 5.10682495626392e-01, 5.10495539334888e-01, 5.10362230870685e-01, 5.10270758501043e-01, 5.10210840747040e-01, 5.10173782607114e-01, 5.10152477380183e-01, 5.10141353993499e-01, 5.10136271594702e-01, 5.10134365006344e-01, 5.10133846367601e-01, 5.10133769817592e-01, 5.10133763978388e-01, 5.10133659968667e-01, 5.10132955287423e-01, 5.10130364803405e-01, 5.10123459514704e-01, 5.10108347411555e-01, 5.10079405743328e-01, 5.10029075367635e-01, 5.09947726565944e-01, 5.09823604210997e-01, 5.09642858435147e-01, 5.09389664930225e-01, 5.09046436662187e-01, 5.08594126074545e-01, 5.08012613779111e-01, 5.07281176331092e-01, 5.06379022052961e-01, 5.05285880164069e-01, 5.03982624905976e-01, 5.02451913189954e-01, 5.00678811823828e-01, 4.98651388889898e-01, 4.96361243599033e-01, 4.93803950123088e-01, 4.90979393591736e-01, 4.87891980589363e-01, 4.84550711926295e-01, 4.80969111879180e-01, 4.77165015080727e-01, 4.73160219296481e-01, 4.68980018934019e-01, 4.64652639788254e-01, 4.60208599811665e-01, 4.55680023306924e-01, 4.51099936724049e-01, 4.46501573226105e-01, 4.41917710554423e-01, 4.37380062803213e-01, 4.32918741928983e-01, 4.28561799640996e-01, 4.24334855201283e-01, 4.20260810000038e-01, 4.16359645860855e-01, 4.12648301048685e-01, 4.09140615964595e-01, 4.05847339474037e-01, 4.02776186608270e-01, 3.99931938828737e-01, 3.97316578954542e-01, 3.94929454028139e-01, 3.92767460657305e-01, 3.90825248577427e-01, 3.89095439217519e-01, 3.87568856851989e-01, 3.86234770436855e-01, 3.85081144453610e-01, 3.84094897026754e-01, 3.83262163278403e-01, 3.82568561383931e-01, 3.81999458160487e-01, 3.81540230329505e-01, 3.81176516923706e-01, 3.80894457739509e-01, 3.80680912343315e-01, 3.80523653991586e-01, 3.80411532967782e-01, 3.80334604303187e-01, 3.80284215634934e-01, 3.80253052037624e-01, 3.80235135995428e-01, 3.80225782186977e-01, 3.80221508350110e-01, 3.80219905085613e-01, 3.80219468958738e-01, 3.80219404587318e-01, 3.80219399642445e-01, 3.80219311108525e-01, 3.80218711278323e-01, 3.80216506240460e-01, 3.80210628430107e-01, 3.80197765041614e-01, 3.80173130327315e-01, 3.80130291020532e-01, 3.80061053105348e-01, 3.79955416967969e-01, 3.79801606561763e-01, 3.79586176536217e-01, 3.79294199254038e-01, 3.78909531199448e-01, 3.78415155443396e-01, 3.77793593605309e-01, 3.77027377235225e-01, 3.76099564895678e-01, 3.74994287695222e-01, 3.73697302912620e-01, 3.72196532989374e-01, 3.70482565893499e-01, 3.68549092958703e-01, 3.66393261972765e-01, 3.64015926589203e-01, 3.61421777953159e-01, 3.58619350486670e-01, 3.55620900619100e-01, 3.52442164312960e-01, 3.49102005903026e-01, 3.45621976436445e-01, 3.42025803878655e-01, 3.38338839904328e-01, 3.34587488405389e-01, 3.30798639423898e-01, 3.26999129250676e-01, 3.23215243364966e-01, 3.19472274244049e-01, 3.15794141350560e-01, 3.12203076268599e-01, 3.08719372323654e-01, 3.05361195280307e-01, 3.02144449912104e-01, 2.99082696320803e-01, 2.96187109684384e-01, 2.93466477479721e-01, 2.90927228924129e-01, 2.88573492262021e-01, 2.86407176414115e-01, 2.84428074316187e-01, 2.82633985920635e-01, 2.81020859307128e-01, 2.79582948590607e-01, 2.78312987408412e-01, 2.77202376648556e-01, 2.76241384825131e-01, 2.75419359130389e-01, 2.74724944709141e-01, 2.74146309177230e-01, 2.73671368843977e-01, 2.73288012587919e-01, 2.72984318868907e-01, 2.72748761039302e-01, 2.72570395929990e-01, 2.72439030736317e-01, 2.72345363473047e-01, 2.72281092773872e-01, 2.72238993565084e-01, 2.72212956079645e-01, 2.72197986837855e-01, 2.72190171465136e-01, 2.72186600538757e-01, 2.72185260958341e-01, 2.72184896559647e-01, 2.72184842775145e-01, 2.72184838611753e-01, 2.72184763664064e-01, 2.72184255879521e-01, 2.72182389210369e-01, 2.72177413354386e-01, 2.72166523807464e-01, 2.72145668992184e-01, 2.72109402199726e-01, 2.72050785166206e-01, 2.71961349056867e-01, 2.71831117376112e-01, 2.71648693875530e-01, 2.71401416790654e-01, 2.71075578709714e-01, 2.70656709021146e-01, 2.70129913266270e-01, 2.69480260867018e-01, 2.68693209822588e-01, 2.67755054200184e-01, 2.66653377890889e-01, 2.65377496417134e-01, 2.63918867808518e-01, 2.62271453964944e-01, 2.60432015560667e-01, 2.58400326464158e-01, 2.56179297604778e-01, 2.53775004986082e-01, 2.51196621680033e-01, 2.48456258590828e-01, 2.45568723273341e-01, 2.42551209485462e-01, 2.39422932514744e-01, 2.36204726256538e-01, 2.32918617956181e-01, 2.29587395291224e-01, 2.26234178805256e-01, 2.22882010493991e-01, 2.19553467223101e-01, 2.16270305614656e-01, 2.13053143221889e-01, 2.09921179475704e-01, 2.06891958451937e-01, 2.03981174660537e-01, 2.01202522190802e-01, 1.98567586452708e-01, 1.96085777541569e-01, 1.93764303145870e-01, 1.91608178633031e-01, 1.89620271635548e-01, 1.87801377869239e-01, 1.86150325143706e-01, 1.84664102260612e-01, 1.83338009786504e-01, 1.82165829553426e-01, 1.81140010047320e-01, 1.80251864890442e-01, 1.79491781233498e-01, 1.78849435353502e-01, 1.78314011969984e-01, 1.77874423936345e-01, 1.77519528295993e-01, 1.77238334497527e-01, 1.77020200700789e-01, 1.76855013119044e-01, 1.76733344374173e-01, 1.76646586429346e-01, 1.76587054511573e-01, 1.76548058366842e-01, 1.76523939687638e-01, 1.76510073477609e-01, 1.76502833927542e-01, 1.76499526115757e-01, 1.76498285235192e-01, 1.76497947686669e-01, 1.76497897860385e-01, 1.76497893972511e-01, 1.76497823566602e-01, 1.76497346659178e-01, 1.76495593436865e-01, 1.76490919977468e-01, 1.76480692228894e-01, 1.76461104705427e-01, 1.76427041379137e-01, 1.76371984671753e-01, 1.76287978342959e-01, 1.76165647950078e-01, 1.75994281780690e-01, 1.75761973849073e-01, 1.75455827438583e-01, 1.75062217291659e-01, 1.74567104984910e-01, 1.73956400213000e-01, 1.73216357787969e-01, 1.72333998799576e-01, 1.71297540614868e-01, 1.70096820894455e-01, 1.68723698022997e-01, 1.67172411608812e-01, 1.65439887012191e-01, 1.63525968948317e-01, 1.61378464063635e-01, 1.59118007593784e-01, 1.56694255720409e-01, 1.54119268343269e-01, 1.51407866736489e-01, 1.48577334775543e-01, 1.45647047902070e-01, 1.42638044131862e-01, 1.39572553711137e-01, 1.36473505472571e-01, 1.33364028452846e-01, 1.30266966869990e-01, 1.27204425142417e-01, 1.24197357352431e-01, 1.21265212576922e-01, 1.18425644050926e-01, 1.15694286461113e-01, 1.13084602066140e-01, 1.10607793074326e-01, 1.08272774997321e-01, 1.06086203693909e-01, 1.04052547589792e-01, 1.02174196092922e-01, 1.00451595434567e-01, 9.88834039197535e-02, 9.74666597059364e-02, 9.61969555772231e-02, 9.50686165825934e-02, 9.40748777200699e-02, 9.32080599638883e-02, 9.24597437728258e-02, 9.18209397456806e-02, 9.12822562994271e-02, 9.08340641617085e-02, 9.04666571415649e-02, 9.01704081389934e-02, 8.99359187561255e-02, 8.97541602670215e-02, 8.96166031740826e-02, 8.95153322021929e-02, 8.94431434162577e-02, 8.93936202314976e-02, 8.93611854318954e-02, 8.93411269080673e-02, 8.93295956357343e-02, 8.93235753846877e-02, 8.93208247079120e-02, 8.93197928382401e-02, 8.93195121450621e-02, 8.93194707153508e-02, 8.93194674550171e-02, 8.93194080687567e-02, 8.93190057184166e-02, 8.93175266406407e-02, 8.93135839979555e-02, 8.93049558016386e-02, 8.92884324662924e-02, 8.92597004161417e-02, 8.92132676101483e-02, 8.91424361968269e-02, 8.90393266967018e-02, 8.88949570719864e-02, 8.86993787077541e-02, 8.84418696374203e-02, 8.81111832666051e-02, 8.76958483952510e-02, 8.71845135746239e-02, 8.65663258905380e-02, 8.58313313236270e-02, 8.49708811388066e-02, 8.39780265694894e-02, 8.28478826652344e-02, 8.15779418194027e-02, 8.01683183849522e-02, 7.86219080327664e-02, 7.69444491045728e-02, 7.51444780263610e-02, 7.32331766068550e-02, 7.12241153520390e-02, 6.91329032862380e-02, 6.69767606351437e-02, 6.47740355542248e-02, 6.25436894076250e-02, 6.03047765853111e-02, 5.80759443454979e-02, 5.58749757539503e-02, 5.37183947381923e-02, 5.16211470298700e-02, 4.95963648889144e-02, 4.76552175788671e-02, 4.58068441406185e-02, 4.40583605275910e-02, 4.24149298986869e-02, 4.08798829238948e-02, 3.94548742921841e-02, 3.81400620531247e-02, 3.69342977294496e-02, 3.58353170334942e-02, 3.48399232358765e-02, 3.39441575292203e-02, 3.31434529095156e-02, 3.24327700247440e-02, 3.18067150317205e-02, 3.12596407210735e-02, 3.07857330142074e-02, 3.03790854244297e-02, 3.00337642403461e-02, 2.97438670752381e-02, 2.95035770801642e-02, 2.93072145945095e-02, 2.91492873631563e-02, 2.90245397440980e-02, 2.89280006248536e-02, 2.88550291193097e-02, 2.88013565830223e-02, 2.87631231096208e-02, 2.87369064862775e-02, 2.87197416075839e-02, 2.87091285714647e-02, 2.87030280856020e-02, 2.86998433595409e-02, 2.86983882952274e-02, 2.86978424606135e-02, 2.86976939813313e-02, 2.86976720661358e-02, 2.86976703262396e-02, 2.86976384404229e-02, 2.86974224098926e-02, 2.86966282673164e-02, 2.86945114401136e-02, 2.86898791470920e-02, 2.86810090221894e-02, 2.86655877758935e-02, 2.86406736147238e-02, 2.86026860048074e-02, 2.85474261131808e-02, 2.84701307969448e-02, 2.83655622364560e-02, 2.82281341366772e-02, 2.80520737946364e-02, 2.78316172470658e-02, 2.75612322377130e-02, 2.72358610274135e-02, 2.68511723439504e-02, 2.64038093357444e-02, 2.58916185991178e-02, 2.53138445394190e-02, 2.46712737951746e-02, 2.39663163852035e-02, 2.32030136528851e-02, 2.23869678042270e-02, 2.15251934888512e-02, 2.06258979051435e-02, 1.96982016683291e-02, 1.87518174984498e-02, 1.77967070891194e-02, 1.68427379085272e-02, 1.58993609992750e-02, 1.49753281892249e-02, 1.40784628718582e-02, 1.32154932462019e-02, 1.23919513182851e-02, 1.16121357262018e-02, 1.08791320769215e-02, 1.01948812860728e-02, 9.56028451395285e-03, 8.97533267219286e-03, 8.43924900023829e-03, 7.95063462822698e-03, 7.50760900781122e-03, 7.10793924060127e-03, 6.74915438140733e-03, 6.42864259120390e-03, 6.14373052091333e-03, 5.89174551566973e-03, 5.67006212386145e-03, 5.47613496049504e-03, 5.30752023140124e-03, 5.16188824388841e-03, 5.03702910335191e-03, 4.93085359878620e-03, 4.84139104998516e-03, 4.76678563253374e-03, 4.70529241190754e-03, 4.65527401049703e-03, 4.61519851465519e-03, 4.58363891991942e-03, 4.55927412612533e-03, 4.54089124169066e-03, 4.52738874785918e-03, 4.51777991823743e-03, 4.51119579514492e-03, 4.50688699782997e-03, 4.50422367953962e-03, 4.50269305573814e-03, 4.50189408292223e-03, 4.50152906006229e-03, 4.50139213295353e-03, 4.50135488598580e-03, 4.50134938843050e-03, 4.50134894797038e-03, 4.50134082562403e-03, 4.50128579583220e-03, 4.50108350498368e-03, 4.50054430633040e-03, 4.49936445819699e-03, 4.49710557808647e-03, 4.49317946325900e-03, 4.48683944413553e-03, 4.47717948383808e-03, 4.46314223715971e-03, 4.44353718668644e-03, 4.41706973032785e-03, 4.38238165601401e-03, 4.33810277565172e-03, 4.28291260104885e-03, 4.21560986991842e-03, 4.13518655842449e-03, 4.04090188361117e-03, 3.93235087635313e-03, 3.80952158124614e-03, 3.67283498930737e-03, 3.52316256015062e-03, 3.36181768558541e-03, 3.19051960976635e-03, 3.01133093352712e-03, 2.82657254138696e-03, 2.63872217441063e-03, 2.45030454038631e-03, 2.26378156477880e-03, 2.08145111069058e-03, 1.90536136785946e-03, 1.73724631649197e-03, 1.57848537841687e-03, 1.43008777176767e-03, 1.29269953028910e-03, 1.16662913384280e-03, 1.05188665822054e-03, 9.48231346327155e-04, 8.55223120859616e-04, 7.72274242375026e-04, 6.98697829975004e-04, 6.33750542873290e-04, 5.76667717141710e-04, 5.26690609945923e-04, 4.83086587387493e-04, 4.45163555028216e-04, 4.12279649235181e-04, 3.83848706462242e-04, 3.59341872987695e-04, 3.38286021771772e-04, 3.20260050363341e-04, 3.04890228805199e-04, 2.91845450047568e-04, 2.80832743522641e-04, 2.71593052294072e-04, 2.63897178530481e-04, 2.57541893590840e-04, 2.52346328792452e-04, 2.48148805619852e-04, 2.44804220008721e-04, 2.42182011887859e-04, 2.40164677027810e-04, 2.38646734747494e-04, 2.37534048948884e-04, 2.36743398763117e-04, 2.36202199004332e-04, 2.35848275913053e-04, 2.35629611068155e-04, 2.35503977505935e-04, 2.35438407802047e-04, 2.35408453563846e-04, 2.35397217519751e-04, 2.35394161121061e-04, 2.35393710006360e-04, 2.35393673533313e-04, 2.35392996841721e-04, 2.35388412201054e-04, 2.35371559217157e-04, 2.35326640275701e-04, 2.35228361262774e-04, 2.35040240962547e-04, 2.34713397868180e-04, 2.34185937143305e-04, 2.33383074519367e-04, 2.32218142898422e-04, 2.30594624659020e-04, 2.28409329769530e-04, 2.25556787072367e-04, 2.21934823096070e-04, 2.17451161141025e-04, 2.12030683489406e-04, 2.05622781490611e-04, 1.98208022908819e-04, 1.89803275545852e-04, 1.80464528241497e-04, 1.70286978835588e-04, 1.59402426599176e-04, 1.47974391058625e-04, 1.36191427758892e-04, 1.24258766535323e-04, 1.12387987908580e-04, 1.00784602956594e-04, 8.96345295879897e-05, 7.91237826636149e-05, 6.93060302272563e-05, 6.02844655421819e-05, 5.21056585855995e-05, 4.47886408894149e-05, 3.83131937704562e-05, 3.26719472809600e-05, 2.77854921857390e-05, 2.35749481900527e-05, 1.99681104793292e-05, 1.68724148261745e-05, 1.42630069777105e-05, 1.20419550664330e-05, 1.01298582726450e-05, 8.46274651822299e-06, 6.92460461948485e-06, 5.46851109417545e-06, 4.09923611011109e-06, 2.86362652834882e-06, 1.82421701300228e-06, 1.03191015732307e-06, 5.10147736054734e-07, 2.74872909322710e-07, 0.00000000000000e+00, 2.81434031190742e-07, 2.87018022238072e-07, 2.63462435423634e-07, -4.44089209850063e-16, -4.44089209850063e-16, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, -4.44089209850063e-16, -4.44089209850063e-16, -4.44089209850063e-16, -4.44089209850063e-16, -4.44089209850063e-16, -4.44089209850063e-16, 0.00000000000000e+00, -4.44089209850063e-16, -4.44089209850063e-16, 0.00000000000000e+00, 0.00000000000000e+00, -4.44089209850063e-16, 0.00000000000000e+00],
[ 4.00000000000000e+00, 3.99999999901167e+00, 3.99999996845054e+00, 3.99999976139408e+00, 3.99999900022821e+00, 3.99999697120561e+00, 3.99999253056743e+00, 3.99998402555551e+00, 3.99996923320731e+00, 3.99994531930490e+00, 3.99990881925772e+00, 3.99985564205351e+00, 3.99978109773359e+00, 3.99967994815619e+00, 3.99954648012818e+00, 3.99937459932932e+00, 3.99915794284457e+00, 3.99889000757602e+00, 3.99856429133984e+00, 3.99817444307810e+00, 3.99771441833735e+00, 3.99717863599023e+00, 3.99656213210550e+00, 3.99586070690324e+00, 3.99507106086349e+00, 3.99419091628145e+00, 3.99321912087554e+00, 3.99215573044687e+00, 3.99100206805363e+00, 3.98976075769071e+00, 3.98843573104642e+00, 3.98703220653192e+00, 3.98555664043579e+00, 3.98401665073110e+00, 3.98242091474296e+00, 3.98077904255365e+00, 3.97910142866304e+00, 3.97739908501592e+00, 3.97568345903579e+00, 3.97396624074896e+00, 3.97225916342618e+00, 3.97057380239723e+00, 3.96892137679601e+00, 3.96731255896256e+00, 3.96575729606196e+00, 3.96426464818072e+00, 3.96284264673670e+00, 3.96149817650061e+00, 3.96023688389273e+00, 3.95906311350619e+00, 3.95797987404111e+00, 3.95698883403559e+00, 3.95609034697461e+00, 3.95528350457090e+00, 3.95456621626624e+00, 3.95393531231943e+00, 3.95338666724738e+00, 3.95291533988541e+00, 3.95251572594451e+00, 3.95218171867685e+00, 3.95190687312165e+00, 3.95168456939376e+00, 3.95150817059401e+00, 3.95137117115927e+00, 3.95126733182116e+00, 3.95119079779385e+00, 3.95113619735035e+00, 3.95109871855489e+00, 3.95107416258142e+00, 3.95105897274429e+00, 3.95105023907956e+00, 3.95104567902451e+00, 3.95104359543033e+00, 3.95104281379132e+00, 3.95104260116553e+00, 3.95104256978237e+00, 3.95104256566556e+00, 3.95104246974884e+00, 3.95104181989797e+00, 3.95103943096272e+00, 3.95103306282530e+00, 3.95101912573833e+00, 3.95099243234981e+00, 3.95094600543894e+00, 3.95087094872092e+00, 3.95075638618889e+00, 3.95058947340884e+00, 3.95035548203719e+00, 3.95003795666427e+00, 3.94961894097354e+00, 3.94907926821518e+00, 3.94839890918574e+00, 3.94755736933173e+00, 3.94653412529188e+00, 3.94530909017712e+00, 3.94386309616642e+00, 3.94217838255709e+00, 3.94023907723297e+00, 3.93803165957475e+00, 3.93554539311079e+00, 3.93277271667259e+00, 3.92970958346811e+00, 3.92635573832005e+00, 3.92271492434772e+00, 3.91879501162172e+00, 3.91460804181110e+00, 3.91017018458941e+00, 3.90550160356879e+00, 3.90062623176949e+00, 3.89557145905882e+00, 3.89036773653342e+00, 3.88504810537492e+00, 3.87964766016460e+00, 3.87420295887596e+00, 3.86875139365445e+00, 3.86333053793862e+00, 3.85797748639450e+00, 3.85272820447846e+00, 3.84761690419737e+00, 3.84267546181870e+00, 3.83793289194615e+00, 3.83341489059209e+00, 3.82914345773494e+00, 3.82513660744493e+00, 3.82140817109329e+00, 3.81796769652037e+00, 3.81482044341132e+00, 3.81196747258609e+00, 3.80940582451499e+00, 3.80712878017235e+00, 3.80512619537974e+00, 3.80338489810146e+00, 3.80188913676639e+00, 3.80062106662571e+00, 3.79956126043339e+00, 3.79868922937044e+00, 3.79798394013208e+00, 3.79742431445913e+00, 3.79698969811418e+00, 3.79666028736179e+00, 3.79641750238258e+00, 3.79624429869917e+00, 3.79612540957100e+00, 3.79604751437445e+00, 3.79599933016737e+00, 3.79597162588169e+00, 3.79595716083355e+00, 3.79595055142482e+00, 3.79594807197364e+00, 3.79594739749961e+00, 3.79594729794855e+00, 3.79594728916008e+00, 3.79594711695440e+00, 3.79594595023395e+00, 3.79594166122183e+00, 3.79593022810323e+00, 3.79590520605401e+00, 3.79585728220035e+00, 3.79577393083770e+00, 3.79563918232036e+00, 3.79543351572089e+00, 3.79513388174835e+00, 3.79471385861661e+00, 3.79414393968518e+00, 3.79339194787421e+00, 3.79242356819865e+00, 3.79120298638147e+00, 3.78969361848157e+00, 3.78785891388154e+00, 3.78566321186571e+00, 3.78307263040407e+00, 3.78005596463825e+00, 3.77658557192748e+00, 3.77263822013217e+00, 3.76819587607447e+00, 3.76324641181937e+00, 3.75778420758597e+00, 3.75181063177163e+00, 3.74533438080829e+00, 3.73837166442979e+00, 3.73094622544934e+00, 3.72308918732702e+00, 3.71483872758557e+00, 3.70623958037841e+00, 3.69734237702119e+00, 3.68820283880260e+00, 3.67888084158902e+00, 3.66943937631780e+00, 3.65994343315246e+00, 3.65045883962205e+00, 3.64105108434397e+00, 3.63178415788937e+00, 3.62271944104527e+00, 3.61391466830067e+00, 3.60542299104305e+00, 3.59729216094727e+00, 3.58956384963375e+00, 3.58227311611556e+00, 3.57544802905904e+00, 3.56910944661659e+00, 3.56327095266507e+00, 3.55793894476078e+00, 3.55311286601724e+00, 3.54878557041029e+00, 3.54494380868009e+00, 3.54156881999212e+00, 3.53863701280336e+00, 3.53612071693631e+00, 3.53398898769086e+00, 3.53220844194169e+00, 3.53074410561235e+00, 3.52956025173538e+00, 3.52862120855132e+00, 3.52789211781600e+00, 3.52733962470887e+00, 3.52693248248073e+00, 3.52664205723583e+00, 3.52644272097504e+00, 3.52631212416896e+00, 3.52623134259432e+00, 3.52618489684662e+00, 3.52616064671457e+00, 3.52614956633985e+00, 3.52614540966199e+00, 3.52614427894041e+00, 3.52614411204811e+00, 3.52614409866128e+00, 3.52614385160638e+00, 3.52614217777120e+00, 3.52613602454877e+00, 3.52611962210757e+00, 3.52608372470386e+00, 3.52601497281034e+00, 3.52589539995103e+00, 3.52570210421487e+00, 3.52540709995765e+00, 3.52497736028236e+00, 3.52437505561263e+00, 3.52355798814499e+00, 3.52248021630245e+00, 3.52109285763929e+00, 3.51934505310591e+00, 3.51718507031837e+00, 3.51456151863945e+00, 3.51142464461428e+00, 3.50772767275121e+00, 3.50342815392676e+00, 3.49848928193492e+00, 3.49288113800111e+00, 3.48658182353174e+00, 3.47957844305567e+00, 3.47186790230321e+00, 3.46345749070564e+00, 3.45436522328343e+00, 3.44461992386162e+00, 3.43426103965201e+00, 3.42333818621661e+00, 3.41191043130776e+00, 3.40004533560045e+00, 3.38781777736031e+00, 3.37530859607368e+00, 3.36260309649256e+00, 3.34978945900284e+00, 3.33695710444312e+00, 3.32419506140188e+00, 3.31159038170763e+00, 3.29922664558264e+00, 3.28718259217457e+00, 3.27553090441473e+00, 3.26433716990617e+00, 3.25365903231889e+00, 3.24354554098420e+00, 3.23403670035091e+00, 3.22516321588254e+00, 3.21694642890748e+00, 3.20939842984417e+00, 3.20252233699146e+00, 3.19631272652370e+00, 3.19075619826257e+00, 3.18583206101561e+00, 3.18151312060581e+00, 3.17776655304008e+00, 3.17455484450499e+00, 3.17183677902052e+00, 3.16956845366893e+00, 3.16770430043328e+00, 3.16619809295554e+00, 3.16500391609936e+00, 3.16407707622793e+00, 3.16337493071345e+00, 3.16285761648616e+00, 3.16248865946595e+00, 3.16223544950702e+00, 3.16206956897776e+00, 3.16196696719323e+00, 3.16190797746890e+00, 3.16187717838908e+00, 3.16186310577768e+00, 3.16185782661365e+00, 3.16185639054910e+00, 3.16185617858884e+00, 3.16185616224042e+00, 3.16185586867531e+00, 3.16185387972712e+00, 3.16184656812211e+00, 3.16182707790578e+00, 3.16178442322297e+00, 3.16170273125618e+00, 3.16156065839486e+00, 3.16133100444718e+00, 3.16098054482379e+00, 3.16047009515308e+00, 3.15975481675146e+00, 3.15878476479972e+00, 3.15750567401629e+00, 3.15585996916121e+00, 3.15378797999276e+00, 3.15122933253881e+00, 3.14812448100457e+00, 3.14441633764449e+00, 3.14005195184713e+00, 3.13498418490849e+00, 3.12917332388871e+00, 3.12258857691075e+00, 3.11520939355652e+00, 3.10702655783336e+00, 3.09804300758460e+00, 3.08827434310630e+00, 3.07774899883704e+00, 3.06650806485519e+00, 3.05460475891648e+00, 3.04210356411885e+00, 3.02907906112160e+00, 3.01561449628196e+00, 3.00180013728105e+00, 2.98773147512244e+00, 2.97350733536162e+00, 2.95922796190644e+00, 2.94499313386051e+00, 2.93090037008502e+00, 2.91704326807503e+00, 2.90351001417041e+00, 2.89038209188375e+00, 2.87773320501880e+00, 2.86562842294374e+00, 2.85412354737119e+00, 2.84326469358280e+00, 2.83308807432315e+00, 2.82361997150928e+00, 2.81487687925469e+00, 2.80686580119897e+00, 2.79958468543028e+00, 2.79302298104661e+00, 2.78716230131236e+00, 2.78197717916836e+00, 2.77743590135350e+00, 2.77350140747612e+00, 2.77013223998855e+00, 2.76728353019884e+00, 2.76490800429064e+00, 2.76295699196144e+00, 2.76138141891529e+00, 2.76013276326535e+00, 2.75916395512473e+00, 2.75843019848631e+00, 2.75788969507444e+00, 2.75750425130222e+00, 2.75723975183390e+00, 2.75706648650565e+00, 2.75695932140343e+00, 2.75689770957014e+00, 2.75686554190349e+00, 2.75685084406259e+00, 2.75684533036781e+00, 2.75684383050657e+00, 2.75684360913010e+00, 2.75684359236815e+00, 2.75684329543170e+00, 2.75684128364279e+00, 2.75683388808220e+00, 2.75681417414266e+00, 2.75677103020995e+00, 2.75668840266162e+00, 2.75654470716770e+00, 2.75631244229764e+00, 2.75595802685175e+00, 2.75544187710621e+00, 2.75471873423175e+00, 2.75373824545264e+00, 2.75244579502766e+00, 2.75078357288941e+00, 2.74869185990047e+00, 2.74611049941868e+00, 2.74298051558024e+00, 2.73924582990306e+00, 2.73485502007701e+00, 2.72976305879410e+00, 2.72393296683037e+00, 2.71733731390193e+00, 2.70995950351449e+00, 2.70179478432488e+00, 2.69285094037667e+00, 2.68314862559408e+00, 2.67272132344138e+00, 2.66161492972692e+00, 2.64988697399808e+00, 2.63760551158486e+00, 2.62484773288823e+00, 2.61169834792240e+00, 2.59824781163973e+00, 2.58459045879867e+00, 2.57082261609327e+00, 2.55704075435949e+00, 2.54333973564928e+00, 2.52981119978134e+00, 2.51654212369561e+00, 2.50361357557902e+00, 2.49109967517213e+00, 2.47906676256740e+00, 2.46757277056929e+00, 2.45666679044941e+00, 2.44638881761723e+00, 2.43676966209377e+00, 2.42783100836825e+00, 2.41958560984093e+00, 2.41203760421571e+00, 2.40518293755404e+00, 2.39900988595396e+00, 2.39349966476267e+00, 2.38862711573578e+00, 2.38436146255434e+00, 2.38066712459942e+00, 2.37750457791134e+00, 2.37483125091703e+00, 2.37260244092176e+00, 2.37077223568412e+00, 2.36929442279779e+00, 2.36812336827416e+00, 2.36721484483539e+00, 2.36652679015385e+00, 2.36601997574701e+00, 2.36565856854815e+00, 2.36541056936088e+00, 2.36524811544196e+00, 2.36514763825233e+00, 2.36508987181750e+00, 2.36505971194411e+00, 2.36504593151449e+00, 2.36504076197784e+00, 2.36503935573648e+00, 2.36503914817813e+00, 2.36503913259665e+00, 2.36503885834507e+00, 2.36503700024965e+00, 2.36503016968538e+00, 2.36501196184008e+00, 2.36497211406685e+00, 2.36489579947607e+00, 2.36476308357484e+00, 2.36454856814949e+00, 2.36422124241470e+00, 2.36374455667082e+00, 2.36307672824047e+00, 2.36217128321470e+00, 2.36097783047429e+00, 2.35944305657768e+00, 2.35751192153822e+00, 2.35512902649717e+00, 2.35224011522826e+00, 2.34879366282174e+00, 2.34474249744654e+00, 2.34004539549633e+00, 2.33466858740003e+00, 2.32858711152474e+00, 2.32178595732738e+00, 2.31426094634056e+00, 2.30601931048122e+00, 2.29707994095348e+00, 2.28747329674918e+00, 2.27724097824697e+00, 2.26643498736568e+00, 2.25511670986214e+00, 2.24335566659015e+00, 2.23122808809372e+00, 2.21881537044725e+00, 2.20620246987068e+00, 2.19347628983360e+00, 2.18072410791593e+00, 2.16803208158089e+00, 2.15548386322806e+00, 2.14315934631046e+00, 2.13113355659751e+00, 2.11947569626741e+00, 2.10824834358649e+00, 2.09750680742949e+00, 2.08729863360788e+00, 2.07766325861341e+00, 2.06863180564088e+00, 2.06022701734579e+00, 2.05246331949570e+00, 2.04534700933235e+00, 2.03887656198130e+00, 2.03304304758066e+00, 2.02783065093954e+00, 2.02321728448627e+00, 2.01917528404498e+00, 2.01567217561513e+00, 2.01267149985897e+00, 2.01013367948033e+00, 2.00801691317609e+00, 2.00627807845065e+00, 2.00487362441360e+00, 2.00376043485541e+00, 2.00289664153903e+00, 2.00224236787511e+00, 2.00176038405862e+00, 2.00141665639441e+00, 2.00118077594440e+00, 2.00102625475128e+00, 2.00093068164535e+00, 2.00087573388169e+00, 2.00084704540848e+00, 2.00083393722805e+00, 2.00082901986788e+00, 2.00082768222394e+00, 2.00082748479040e+00, 2.00082747001407e+00, 2.00082721053429e+00, 2.00082545251973e+00, 2.00081898985853e+00, 2.00080176267139e+00, 2.00076406086020e+00, 2.00069185533685e+00, 2.00056628292692e+00, 2.00036330734381e+00, 2.00005357448797e+00, 1.99960247562964e+00, 1.99897042675189e+00, 1.99811336643165e+00, 1.99698346811132e+00, 1.99553005550734e+00, 1.99370070233735e+00, 1.99144248974890e+00, 1.98870338713692e+00, 1.98543371489058e+00, 1.98158764154033e+00, 1.97712466335229e+00, 1.97201101217785e+00, 1.96622093774711e+00, 1.95973781384694e+00, 1.95255502395045e+00, 1.94467659058522e+00, 1.93611752348533e+00, 1.92690387358495e+00, 1.91707249226767e+00, 1.90667050706540e+00, 1.89575453540273e+00, 1.88438966643012e+00, 1.87264824719501e+00, 1.86060851336994e+00, 1.84835310674263e+00, 1.83596752207387e+00, 1.82353852517941e+00, 1.81115258255216e+00, 1.79889434072541e+00, 1.78684519093111e+00, 1.77508195133845e+00, 1.76367569511065e+00, 1.75269074756750e+00, 1.74218386988057e+00, 1.73220364012224e+00, 1.72279003548430e+00, 1.71397421255642e+00, 1.70577847624358e+00, 1.69821642269118e+00, 1.69129323782707e+00, 1.68500613097160e+00, 1.67934488234579e+00, 1.67429248396333e+00, 1.66982585493491e+00, 1.66591661418796e+00, 1.66253189557888e+00, 1.65963519199078e+00, 1.65718721603528e+00, 1.65514676531771e+00, 1.65347157992355e+00, 1.65211917899674e+00, 1.65104766224541e+00, 1.65021646120753e+00, 1.64958702442183e+00, 1.64912342052714e+00, 1.64879284394685e+00, 1.64856600931224e+00, 1.64841742315886e+00, 1.64832552462631e+00, 1.64827269075810e+00, 1.64824510632961e+00, 1.64823250268786e+00, 1.64822777461239e+00, 1.64822648845987e+00, 1.64822629862640e+00, 1.64822628441356e+00, 1.64822603475788e+00, 1.64822434330356e+00, 1.64821812533364e+00, 1.64820155047083e+00, 1.64816527659662e+00, 1.64809580705142e+00, 1.64797499662690e+00, 1.64777972869250e+00, 1.64748178213890e+00, 1.64704790262084e+00, 1.64644008776227e+00, 1.64561609047587e+00, 1.64453013830337e+00, 1.64313385972356e+00, 1.64137740080279e+00, 1.63921070758066e+00, 1.63658494149126e+00, 1.63345398730646e+00, 1.62977600600479e+00, 1.62551497909558e+00, 1.62064218673163e+00, 1.61513755983801e+00, 1.60899084680079e+00, 1.60220253822551e+00, 1.59478449901734e+00, 1.58676026558825e+00, 1.57816497732129e+00, 1.56904492540325e+00, 1.55945671855241e+00, 1.54946608361205e+00, 1.53914633874673e+00, 1.52857659693913e+00, 1.51783977601754e+00, 1.50702050648544e+00, 1.49620303770970e+00, 1.48546924451668e+00, 1.47489682868347e+00, 1.46455779321690e+00, 1.45451724325864e+00, 1.44483253894672e+00, 1.43555279649917e+00, 1.42671870812508e+00, 1.41836263227478e+00, 1.41050889495743e+00, 1.40317424049224e+00, 1.39636837482367e+00, 1.39009455423060e+00, 1.38435018442931e+00, 1.37912740748886e+00, 1.37441366499778e+00, 1.37019223457463e+00, 1.36644274275679e+00, 1.36314166064401e+00, 1.36026278981084e+00, 1.35777774544563e+00, 1.35565644193838e+00, 1.35386758367966e+00, 1.35237916101356e+00, 1.35115894839380e+00, 1.35017499903769e+00, 1.34939612792692e+00, 1.34879237300535e+00, 1.34833542299025e+00, 1.34799899944591e+00, 1.34775918074177e+00, 1.34759465627430e+00, 1.34748690086828e+00, 1.34742026153238e+00, 1.34738195161170e+00, 1.34736195069378e+00, 1.34735281218069e+00, 1.34734938402049e+00, 1.34734845147889e+00, 1.34734831383774e+00, 1.34734830350521e+00, 1.34734812164391e+00, 1.34734688950752e+00, 1.34734236005359e+00, 1.34733028626767e+00, 1.34730386346438e+00, 1.34725326209808e+00, 1.34716527047363e+00, 1.34702306489252e+00, 1.34680612257068e+00, 1.34649028996915e+00, 1.34604801578529e+00, 1.34544875354159e+00, 1.34465953329464e+00, 1.34364569536785e+00, 1.34237177123110e+00, 1.34080248794961e+00, 1.33890386347593e+00, 1.33664435117940e+00, 1.33399598432310e+00, 1.33093546577599e+00, 1.32744514616340e+00, 1.32351383583939e+00, 1.31913740308367e+00, 1.31431912282241e+00, 1.30906975629088e+00, 1.30340736099715e+00, 1.29735685005071e+00, 1.29094933792349e+00, 1.28422132355505e+00, 1.27721376944096e+00, 1.26997113595225e+00, 1.26254042385959e+00, 1.25497026634467e+00, 1.24731009705231e+00, 1.23960940571289e+00, 1.23191708003304e+00, 1.22428082363414e+00, 1.21674663550802e+00, 1.20935833644986e+00, 1.20215713117269e+00, 1.19518119989359e+00, 1.18846531870508e+00, 1.18204051288447e+00, 1.17593375074411e+00, 1.17016768741116e+00, 1.16476046811871e+00, 1.15972559948948e+00, 1.15507189529887e+00, 1.15080350071982e+00, 1.14691999641447e+00, 1.14341658129283e+00, 1.14028433044901e+00, 1.13751052277542e+00, 1.13507903104416e+00, 1.13297076580275e+00, 1.13116416321556e+00, 1.12963570595781e+00, 1.12836046541842e+00, 1.12731265279549e+00, 1.12646616619911e+00, 1.12579512065319e+00, 1.12527434795935e+00, 1.12487985380368e+00, 1.12458922029009e+00, 1.12438194329768e+00, 1.12423969568339e+00, 1.12414650936039e+00, 1.12408887162293e+00, 1.12405573368179e+00, 1.12403843211981e+00, 1.12403052676302e+00, 1.12402756116969e+00, 1.12402675445441e+00, 1.12402663538486e+00, 1.12402662640839e+00, 1.12402646790803e+00, 1.12402539404408e+00, 1.12402144639625e+00, 1.12401092333707e+00, 1.12398789347634e+00, 1.12394378699309e+00, 1.12386708061339e+00, 1.12374308972448e+00, 1.12355387851060e+00, 1.12327829605724e+00, 1.12289214309499e+00, 1.12236847043621e+00, 1.12167800620726e+00, 1.12078970473415e+00, 1.11967140548532e+00, 1.11829058595405e+00, 1.11661518798368e+00, 1.11461449306765e+00, 1.11226001890159e+00, 1.10952640723785e+00, 1.10639227216327e+00, 1.10284097846435e+00, 1.09886132180121e+00, 1.09444808586524e+00, 1.08960245628370e+00, 1.08433227639340e+00, 1.07865213573768e+00, 1.07258328788911e+00, 1.06615339970745e+00, 1.05939613927023e+00, 1.05235061442779e+00, 1.04506067824934e+00, 1.03757412154000e+00, 1.02994177604893e+00, 1.02221655479626e+00, 1.01445245789223e+00, 1.00670357306360e+00, 9.99023099651519e-01, 9.91462423029598e-01, 9.84070263288547e-01, 9.76891917875577e-01, 9.69968613010733e-01, 9.63336973542917e-01, 9.57028615870119e-01, 9.51069863990928e-01, 9.45481584940945e-01, 9.40279136944425e-01, 9.35472421607941e-01, 9.31066030327877e-01, 9.27059474631343e-01, 9.23447490231243e-01, 9.20220404946762e-01, 9.17364561127859e-01, 9.14862783664490e-01, 9.12694884939452e-01, 9.10838198129077e-01, 9.09268130050692e-01, 9.07958724328881e-01, 9.06883225074130e-01, 9.06014630637066e-01, 9.05326226437881e-01, 9.04792085497477e-01, 9.04387525230791e-01, 9.04089509400077e-01, 9.03876984932945e-01, 9.03731144615501e-01, 9.03635608462684e-01, 9.03576518792612e-01, 9.03542546600132e-01, 9.03524809618979e-01, 9.03516705346453e-01, 9.03513665137968e-01, 9.03512838126027e-01, 9.03512716060743e-01, 9.03512706808362e-01, 9.03512542771715e-01, 9.03511431398822e-01, 9.03507345867376e-01, 9.03496455285807e-01, 9.03472621236671e-01, 9.03426975117040e-01, 9.03347592802211e-01, 9.03219280939441e-01, 9.03023486623672e-01, 9.02738339304196e-01, 9.02338830494860e-01, 9.01797133186189e-01, 9.01083058778138e-01, 9.00164644891586e-01, 8.99008862646893e-01, 8.97582427046440e-01, 8.95852689153305e-01, 8.93788584064086e-01, 8.91361604519664e-01, 8.88546766697900e-01, 8.85323532602547e-01, 8.81676652790488e-01, 8.77596894193557e-01, 8.73081620633123e-01, 8.68135198323197e-01, 8.62769205106544e-01, 8.57002430119958e-01, 8.50860659648227e-01, 8.44376254580989e-01, 8.37587534514126e-01, 8.30537992467909e-01, 8.23275371770383e-01, 8.15850642301819e-01, 8.08316916594027e-01, 8.00728347016815e-01, 7.93139043495801e-01, 7.85602047165163e-01, 7.78168389545142e-01, 7.70886259866637e-01, 7.63800295720522e-01, 7.56951004936570e-01, 7.50374320043991e-01, 7.44101281229741e-01, 7.38157839615111e-01, 7.32564769969314e-01, 7.27337680580366e-01, 7.22487107712054e-01, 7.18018682632100e-01, 7.13933360318622e-01, 7.10227700368061e-01, 7.06894192099120e-01, 7.03921617182559e-01, 7.01295444186680e-01, 6.98998250126362e-01, 6.97010164401264e-01, 6.95309330411343e-01, 6.93872379686907e-01, 6.92674912638913e-01, 6.91691979120596e-01, 6.90898551008914e-01, 6.90269978089442e-01, 6.89782417788206e-01, 6.89413228857657e-01, 6.89141319092385e-01, 6.88947437596342e-01, 6.88814403084719e-01, 6.88727261176629e-01, 6.88673365573664e-01, 6.88642380339105e-01, 6.88626203075919e-01, 6.88618811510036e-01, 6.88616038671899e-01, 6.88615284392074e-01, 6.88615173061930e-01, 6.88615164569711e-01, 6.88615013303666e-01, 6.88613988453870e-01, 6.88610220996363e-01, 6.88600178328306e-01, 6.88578200146341e-01, 6.88536109112555e-01, 6.88462911869548e-01, 6.88344604102584e-01, 6.88164091184238e-01, 6.87901233424087e-01, 6.87533022010038e-01, 6.87033888293114e-01, 6.86376145077563e-01, 6.85530554021105e-01, 6.84467008181813e-01, 6.83155313311796e-01, 6.81566045934269e-01, 6.79671460886557e-01, 6.77446416283737e-01, 6.74869280220298e-01, 6.71922781445847e-01, 6.68594766135340e-01, 6.64878825016398e-01, 6.60774759633681e-01, 6.56288863321537e-01, 6.51434001181064e-01, 6.46229483446294e-01, 6.40700737320852e-01, 6.34878792805023e-01, 6.28799607336345e-01, 6.22503261455423e-01, 6.16033062602303e-01, 6.09434596241790e-01, 6.02754762812531e-01, 5.96040835796177e-01, 5.89339571052430e-01, 5.82696391157098e-01, 5.76154661553330e-01, 5.69755068565950e-01, 5.63535103285901e-01, 5.57528650376308e-01, 5.51765677160872e-01, 5.46272015933285e-01, 5.41069231143434e-01, 5.36174562758402e-01, 5.31600937413146e-01, 5.27357039709388e-01, 5.23447436972768e-01, 5.19872751761910e-01, 5.16629877309053e-01, 5.13712231773302e-01, 5.11110047652675e-01, 5.08810692906176e-01, 5.06799020279434e-01, 5.05057741021335e-01, 5.03567818653524e-01, 5.02308877752223e-01, 5.01259621878528e-01, 5.00398253917639e-01, 4.99702891238528e-01, 4.99151967349712e-01, 4.98724611192220e-01, 4.98400994961118e-01, 4.98162641453133e-01, 4.97992682450728e-01, 4.97876060596865e-01, 4.97799668581792e-01, 4.97752421213165e-01, 4.97725258001194e-01, 4.97711076163129e-01, 4.97704596320009e-01, 4.97702165499999e-01, 4.97701504257342e-01, 4.97701406659288e-01, 4.97701399162011e-01, 4.97701264929475e-01, 4.97700355484175e-01, 4.97697012265252e-01, 4.97688100457275e-01, 4.97668597113504e-01, 4.97631245608700e-01, 4.97566290188520e-01, 4.97461302703704e-01, 4.97301111669406e-01, 4.97067841529912e-01, 4.96741067374422e-01, 4.96298087274964e-01, 4.95714310854391e-01, 4.94963758631809e-01, 4.94019662189106e-01, 4.92855150391082e-01, 4.91444002002173e-01, 4.89761440394642e-01, 4.87784942030125e-01, 4.85495027450147e-01, 4.82876002050403e-01, 4.79916614281662e-01, 4.76610601322056e-01, 4.72957096714575e-01, 4.68960880750998e-01, 4.64632462080302e-01, 4.59987987516143e-01, 4.55048985596350e-01, 4.49841957377467e-01, 4.44397834587984e-01, 4.38751330149327e-01, 4.32940208968797e-01, 4.27004507821247e-01, 4.20985732287790e-01, 4.14926056481205e-01, 4.08867548100897e-01, 4.02851437659285e-01, 3.96917446873945e-01, 3.91103187493971e-01, 3.85443638389649e-01, 3.79970705660568e-01, 3.74712867830211e-01, 3.69694905888481e-01, 3.64937716005181e-01, 3.60458201160436e-01, 3.56269236721486e-01, 3.52379704138281e-01, 3.48794586421923e-01, 3.45515118881233e-01, 3.42538988672788e-01, 3.39860576996040e-01, 3.37471238150375e-01, 3.35359610071828e-01, 3.33511951295576e-01, 3.31912499474449e-01, 3.30543846576193e-01, 3.29387325666408e-01, 3.28423403775946e-01, 3.27632074798298e-01, 3.26993245737274e-01, 3.26487109020082e-01, 3.26094493105924e-01, 3.25797183353917e-01, 3.25578205151790e-01, 3.25422061711649e-01, 3.25314919744450e-01, 3.25244737430216e-01, 3.25201330671095e-01, 3.25176375482092e-01, 3.25163346446681e-01, 3.25157393332200e-01, 3.25155160106867e-01, 3.25154552614809e-01, 3.25154462950231e-01, 3.25154456009824e-01, 3.25154331063243e-01, 3.25153484531683e-01, 3.25150372590651e-01, 3.25142077287129e-01, 3.25123923161039e-01, 3.25089155618522e-01, 3.25028693884548e-01, 3.24930969808317e-01, 3.24781861993904e-01, 3.24564732642157e-01, 3.24260572063646e-01, 3.23848253000099e-01, 3.23304893633816e-01, 3.22606324466900e-01, 3.21727650165231e-01, 3.20643893104726e-01, 3.19330700923671e-01, 3.17765096141160e-01, 3.15926242170713e-01, 3.13796197191272e-01, 3.11360625675934e-01, 3.08609437207517e-01, 3.05537323715670e-01, 3.02144169504105e-01, 2.98435313296211e-01, 2.94421647758849e-01, 2.90119549185467e-01, 2.85550637762049e-01, 2.80741376597716e-01, 2.75722524984026e-01, 2.70528467715150e-01, 2.65196447399053e-01, 2.59765730256908e-01, 2.54276737786131e-01, 2.48770176794916e-01, 2.43286198749229e-01, 2.37863616253988e-01, 2.32539200063147e-01, 2.27347074610509e-01, 2.22318224073307e-01, 2.17480114859530e-01, 2.12856434579671e-01, 2.08466942409753e-01, 2.04327421575975e-01, 2.00449721681454e-01, 1.96841876819855e-01, 1.93508284830757e-01, 1.90449933502253e-01, 1.87664660802549e-01, 1.85147438069828e-01, 1.82890667243725e-01, 1.80884485432528e-01, 1.79117072161938e-01, 1.77574956374843e-01, 1.76243321531578e-01, 1.75106307936148e-01, 1.74147311678801e-01, 1.73349279379527e-01, 1.72694997322640e-01, 1.72167372702796e-01, 1.71749703692753e-01, 1.71425934037014e-01, 1.71180887014572e-01, 1.71000473023462e-01, 1.70871864818404e-01, 1.70783634642371e-01, 1.70725848154059e-01, 1.70690111143218e-01, 1.70669566481431e-01, 1.70658840482105e-01, 1.70653939722249e-01, 1.70652101284153e-01, 1.70651601184993e-01, 1.70651527371426e-01, 1.70651521611632e-01, 1.70651417320976e-01, 1.70650710736905e-01, 1.70648113266056e-01, 1.70641189415744e-01, 1.70626037009230e-01, 1.70597019374917e-01, 1.70546560710997e-01, 1.70465014813929e-01, 1.70340616161107e-01, 1.70159520868839e-01, 1.69905943194737e-01, 1.69562390902407e-01, 1.69109999860671e-01, 1.68528964660631e-01, 1.67799057825848e-01, 1.66900225479930e-01, 1.65813242340492e-01, 1.64520403948412e-01, 1.63006229520505e-01, 1.61258145189305e-01, 1.59267115131628e-01, 1.57028187608644e-01, 1.54540924562859e-01, 1.51809687303722e-01, 1.48843756927869e-01, 1.45657276204037e-01, 1.42269009226847e-01, 1.38701925533233e-01, 1.34982625765947e-01, 1.31140635484049e-01, 1.27207601518545e-01, 1.23216430643475e-01, 1.19200412793122e-01, 1.15192370406664e-01, 1.11223871841213e-01, 1.07324540585960e-01, 1.03521483908118e-01, 9.98388554039664e-02, 9.62975566071593e-02, 9.29150741527676e-02, 8.97054416806093e-02, 8.66793101333023e-02, 8.38441065665418e-02, 8.12042600181808e-02, 7.87614731731865e-02, 7.65150201823257e-02, 7.44620536457945e-02, 7.25979070486078e-02, 7.09163824544645e-02, 6.94100167135447e-02, 6.80703225796813e-02, 6.68880038038178e-02, 6.58531453827425e-02, 6.49553816595647e-02, 6.41840459021799e-02, 6.35283053659945e-02, 6.29772857380995e-02, 6.25201883390387e-02, 6.21464026153551e-02, 6.18456153910674e-02, 6.16079171659196e-02, 6.14239045625351e-02, 6.12847769412750e-02, 6.11824243190902e-02, 6.11095031287046e-02, 6.10594960940460e-02, 6.10267526034827e-02, 6.10065064263869e-02, 6.09948683998369e-02, 6.09887927406034e-02, 6.09860168220644e-02, 6.09849754954195e-02, 6.09846922308557e-02, 6.09846504216547e-02, 6.09846471314568e-02, 6.09845872012462e-02, 6.09841811661371e-02, 6.09826885513911e-02, 6.09787098904100e-02, 6.09700032053353e-02, 6.09533308435708e-02, 6.09243436680633e-02, 6.08775093363234e-02, 6.08060910523895e-02, 6.07021826190031e-02, 6.05568047047607e-02, 6.03600658143475e-02, 6.01013893697688e-02, 5.97698054895854e-02, 5.93543024891434e-02, 5.88442289289905e-02, 5.82297324534040e-02, 5.75022170616819e-02, 5.66547963293162e-02, 5.56827170008254e-02, 5.45837258687252e-02, 5.33583534102733e-02, 5.20100905813550e-02, 5.05454405228409e-02, 4.89738344757256e-02, 4.73074103719320e-02, 4.55606625385294e-02, 4.37499807102979e-02, 4.18931050211486e-02, 4.00085298718693e-02, 3.81148928352610e-02, 3.62303847156107e-02, 3.43722136095694e-02, 3.25561498090114e-02, 3.07961704477728e-02, 2.91042139060935e-02, 2.74900451572728e-02, 2.59612253675865e-02, 2.45231728214197e-02, 2.31792980334520e-02, 2.19311938186175e-02, 2.07788609368231e-02, 1.97209513363892e-02, 1.87550135172199e-02, 1.78777276498292e-02, 1.70851214079870e-02, 1.63727606776880e-02, 1.57359121635801e-02, 1.51696772855736e-02, 1.46690985846099e-02, 1.42292411471101e-02, 1.38452523697419e-02, 1.35124038007741e-02, 1.32261188960494e-02, 1.29819903918098e-02, 1.27757906850419e-02, 1.26034781734070e-02, 1.24612019797454e-02, 1.23453069023047e-02, 1.22523398194221e-02, 1.21790581634658e-02, 1.21224404916753e-02, 1.20796986506897e-02, 1.20482905870269e-02, 1.20259325254386e-02, 1.20106090429699e-02, 1.20005795214038e-02, 1.19943795650488e-02, 1.19908162111915e-02, 1.19889561109643e-02, 1.19881062823892e-02, 1.19877874931227e-02, 1.19877007757294e-02, 1.19876879764607e-02, 1.19876869602984e-02, 1.19876683378362e-02, 1.19875421684821e-02, 1.19870783650726e-02, 1.19858421006414e-02, 1.19831368968537e-02, 1.19779573749752e-02, 1.19689541068912e-02, 1.19544130706122e-02, 1.19322522599554e-02, 1.19000379498466e-02, 1.18550228855954e-02, 1.17942081483133e-02, 1.17144295609064e-02, 1.16124681823351e-02, 1.14851826883720e-02, 1.13296593257077e-02, 1.11433728122181e-02, 1.09243492881737e-02, 1.06713205221101e-02, 1.03838573967643e-02, 1.00624705789776e-02, 9.70866745686783e-03, 9.32495699563773e-03, 8.91479800216421e-03, 8.48249105803545e-03, 8.03301954560132e-03, 7.57185009179562e-03, 7.10470671119534e-03, 6.63733536838551e-03, 6.17527624568437e-03, 5.72365963830235e-03, 5.28703835977407e-03, 4.86926533189092e-03, 4.47342029278497e-03, 4.10178491891688e-03, 3.75586165197728e-03, 3.43642853069070e-03, 3.14362056884487e-03, 2.87702773969478e-03, 2.63580027586130e-03, 2.41875344854936e-03, 2.22446583524105e-03, 2.05136695399455e-03, 1.89781182130622e-03, 1.76214141950259e-03, 1.64272924308140e-03, 1.53801501195794e-03, 1.44652724136041e-03, 1.36689661912470e-03, 1.29786211635396e-03, 1.23827156625556e-03, 1.18707820792574e-03, 1.14333447241854e-03, 1.10618409474483e-03, 1.07485344947911e-03, 1.04864281947492e-03, 1.02691812404565e-03, 1.00910346912952e-03, 9.94674745395052e-04, 9.83154389083918e-04, 9.74107327345042e-04, 9.67138048095784e-04, 9.61888661889621e-04, 9.58037761229491e-04, 9.55299835239121e-04, 9.53424968373806e-04, 9.52198543797689e-04, 9.51440686197458e-04, 9.51005213793543e-04, 9.50777921885582e-04, 9.50674084931880e-04, 9.50635134465472e-04, 9.50624539217682e-04, 9.50622975389948e-04, 9.50622850097282e-04, 9.50620539653890e-04, 9.50604886182926e-04, 9.50547344235275e-04, 9.50393972867669e-04, 9.50058396126607e-04, 9.49416005206150e-04, 9.48299753636750e-04, 9.46497932479584e-04, 9.43754345752446e-04, 9.39771320876748e-04, 9.34215963982243e-04, 9.26729982644758e-04, 9.16943228526979e-04, 9.04490846770578e-04, 8.89033558850727e-04, 8.70280171912263e-04, 8.48010944207722e-04, 8.22100009669313e-04, 7.92534757324770e-04, 7.59429959807267e-04, 7.23034622646779e-04, 6.83730022283946e-04, 6.42018202759331e-04, 5.98501228654236e-04, 5.53852595970739e-04, 5.08783186254025e-04, 4.64004815980079e-04, 4.20194654624950e-04, 3.77963553843230e-04, 3.37830758522895e-04, 3.00206713192530e-04, 2.65384830174575e-04, 2.33542150353205e-04, 2.04747802656335e-04, 1.78977206917619e-04, 1.56129406672090e-04, 1.36045082017588e-04, 1.18523640244561e-04, 1.03338801190844e-04, 9.02526243660162e-05, 7.90277222462166e-05, 6.94369076938983e-05, 6.12694726820351e-05, 5.43340085585875e-05, 4.84587045499474e-05, 4.34905930388751e-05, 3.92948484031486e-05, 3.57543521962711e-05, 3.27690540871117e-05, 3.02546314019381e-05, 2.81404164486077e-05, 2.63670169986518e-05, 2.48841502963337e-05, 2.36489747824642e-05, 2.26248965553566e-05, 2.17806640727680e-05, 2.10895802048583e-05, 2.05287623726846e-05, 2.00784661541498e-05, 1.97215132402562e-05, 1.94428462769736e-05, 1.92292043790232e-05, 1.90688954910812e-05, 1.89516385082911e-05, 1.88684527433480e-05, 1.88115782195908e-05, 1.87744141530999e-05, 1.87514648777309e-05, 1.87382835350469e-05, 1.87314052331189e-05, 1.87282632895247e-05, 1.87270847700205e-05, 1.87267641966748e-05, 1.87267168811900e-05, 1.87267130544733e-05, 1.87266420699217e-05, 1.87261611466205e-05, 1.87243933100767e-05, 1.87196816128221e-05, 1.87093737511468e-05, 1.86896466582986e-05, 1.86553839367498e-05, 1.86001213831055e-05, 1.85160782200100e-05, 1.83942934666170e-05, 1.82248870608248e-05, 1.79974623883972e-05, 1.77016591313972e-05, 1.73278499762120e-05, 1.68679499386037e-05, 1.63162724531318e-05, 1.56703277838233e-05, 1.49314316009175e-05, 1.41049980744157e-05, 1.32004548945019e-05, 1.22308365639867e-05, 1.12122404671844e-05, 1.01633721549810e-05, 9.10526091413288e-06, 8.06088276439354e-06, 7.05406694523347e-06, 6.10705925208066e-06, 5.23681476405358e-06, 4.45138058147165e-06, 3.74875754749482e-06, 3.12017531012287e-06, 2.55725085462899e-06, 2.05927714702625e-06, 1.63557193255670e-06, 1.29992007558144e-06, 1.05839457331669e-06, 8.97385477038881e-07, 7.84230873573932e-07, 6.84104315862299e-07, 5.76629567650144e-07, 4.60022234971547e-07, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -3.55271367880050e-15, -3.55271367880050e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -3.55271367880050e-15, -2.66453525910038e-15, -3.55271367880050e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -3.55271367880050e-15, -2.66453525910038e-15, -3.55271367880050e-15, -3.55271367880050e-15, -3.55271367880050e-15, -3.55271367880050e-15, -3.55271367880050e-15, -2.66453525910038e-15, -3.55271367880050e-15, -2.66453525910038e-15, -2.66453525910038e-15, -3.55271367880050e-15, -2.66453525910038e-15, -3.55271367880050e-15],
[ 5.00000000000000e+00, 4.99999999862353e+00, 4.99999995606027e+00, 4.99999966768755e+00, 4.99999860759274e+00, 4.99999578172166e+00, 4.99998959713080e+00, 4.99997775197544e+00, 4.99995715025904e+00, 4.99992384464222e+00, 4.99987300978396e+00, 4.99979894778734e+00, 4.99969512637673e+00, 4.99955424946914e+00, 4.99936835885180e+00, 4.99912896476729e+00, 4.99882720236504e+00, 4.99845401022578e+00, 4.99800032652485e+00, 4.99745729788419e+00, 4.99681649558373e+00, 4.99607013356439e+00, 4.99521128255941e+00, 4.99423407473499e+00, 4.99313389339972e+00, 4.99190754264744e+00, 4.99055339222101e+00, 4.98907149341508e+00, 4.98746366246416e+00, 4.98573352857843e+00, 4.98388654458099e+00, 4.98192995895631e+00, 4.97987274902513e+00, 4.97772551589943e+00, 4.97550034282387e+00, 4.97321061945357e+00, 4.97087083552823e+00, 4.96849634825112e+00, 4.96610312844075e+00, 4.96370749116542e+00, 4.96132581707013e+00, 4.95897427094001e+00, 4.95666852419836e+00, 4.95442348799860e+00, 4.95225306333711e+00, 4.95016991419025e+00, 4.94818526907662e+00, 4.94630875568341e+00, 4.94454827229622e+00, 4.94290989876572e+00, 4.94139784866257e+00, 4.94001446314887e+00, 4.93876024596484e+00, 4.93763393782758e+00, 4.93663262749688e+00, 4.93575189581168e+00, 4.93498598816603e+00, 4.93432801019734e+00, 4.93377014092150e+00, 4.93330385718004e+00, 4.93292016307328e+00, 4.93260981804143e+00, 4.93236355742090e+00, 4.93217229963697e+00, 4.93202733468526e+00, 4.93192048918546e+00, 4.93184426404285e+00, 4.93179194160336e+00, 4.93175766011205e+00, 4.93173645425691e+00, 4.93172426157447e+00, 4.93171789548302e+00, 4.93171498666908e+00, 4.93171389545731e+00, 4.93171359861981e+00, 4.93171355480716e+00, 4.93171354905987e+00, 4.93171341515476e+00, 4.93171250792660e+00, 4.93170917283925e+00, 4.93170028256241e+00, 4.93168082560486e+00, 4.93164356012199e+00, 4.93157874549051e+00, 4.93147396191653e+00, 4.93131402570205e+00, 4.93108100492877e+00, 4.93075433732314e+00, 4.93031104904675e+00, 4.92972607021549e+00, 4.92897264018682e+00, 4.92802279315366e+00, 4.92684791241376e+00, 4.92541933989214e+00, 4.92370902610353e+00, 4.92169020475075e+00, 4.91933807554162e+00, 4.91663047853658e+00, 4.91354854336997e+00, 4.91007729698340e+00, 4.90620621404637e+00, 4.90192969501313e+00, 4.89724745779470e+00, 4.89216483034959e+00, 4.88669293316666e+00, 4.88084874267992e+00, 4.87465502915343e+00, 4.86814016551190e+00, 4.86133780693233e+00, 4.85428644467238e+00, 4.84702884145809e+00, 4.83961135961629e+00, 4.83208319681695e+00, 4.82449554758206e+00, 4.81690071142975e+00, 4.80935117049554e+00, 4.80189866059767e+00, 4.79459325993681e+00, 4.78748251894838e+00, 4.78061065332156e+00, 4.77401781996759e+00, 4.76773949290104e+00, 4.76180595274559e+00, 4.75624190004730e+00, 4.75106619892145e+00, 4.74629175390262e+00, 4.74192551931920e+00, 4.73796863715602e+00, 4.73441669626503e+00, 4.73126010297854e+00, 4.72848455070572e+00, 4.72607157397630e+00, 4.72399917065798e+00, 4.72224247473698e+00, 4.72077446113594e+00, 4.71956666356856e+00, 4.71858988641365e+00, 4.71781489204135e+00, 4.71721304594219e+00, 4.71675690338379e+00, 4.71642072312409e+00, 4.71618089590445e+00, 4.71601627797488e+00, 4.71590842270102e+00, 4.71584170628946e+00, 4.71580334676043e+00, 4.71578331840575e+00, 4.71577416700569e+00, 4.71577073395438e+00, 4.71576980007692e+00, 4.71576966223841e+00, 4.71576965006989e+00, 4.71576941163373e+00, 4.71576779619154e+00, 4.71576185762308e+00, 4.71574602733350e+00, 4.71571138188160e+00, 4.71564502690509e+00, 4.71552962004929e+00, 4.71534305162486e+00, 4.71505829719562e+00, 4.71464345135786e+00, 4.71406194675606e+00, 4.71327295703874e+00, 4.71223197714658e+00, 4.71089156919719e+00, 4.70920225743110e+00, 4.70711355133974e+00, 4.70457507230809e+00, 4.70153775594500e+00, 4.69795509978046e+00, 4.69378442419379e+00, 4.68898811329278e+00, 4.68353480198334e+00, 4.67740047565460e+00, 4.67056944978710e+00, 4.66303519843494e+00, 4.65480100303095e+00, 4.64588039642291e+00, 4.63629738156428e+00, 4.62608640989714e+00, 4.61529211114127e+00, 4.60396877378979e+00, 4.59217958383538e+00, 4.57999563772156e+00, 4.56749475375460e+00, 4.55476011370745e+00, 4.54187877260152e+00, 4.52894007924565e+00, 4.51603405275611e+00, 4.50324976084700e+00, 4.49067374421308e+00, 4.47838852803066e+00, 4.46647125681023e+00, 4.45499248296349e+00, 4.44401513294510e+00, 4.43359366812694e+00, 4.42377345103381e+00, 4.41459032150153e+00, 4.40607038190253e+00, 4.39822998590895e+00, 4.39107592133654e+00, 4.38460577436987e+00, 4.37880845980867e+00, 4.37366489977241e+00, 4.36914883144198e+00, 4.36522772280755e+00, 4.36186377397640e+00, 4.35901498035677e+00, 4.35663623300220e+00, 4.35468043063841e+00, 4.35309957749068e+00, 4.35184584108718e+00, 4.35087254483205e+00, 4.35013507140568e+00, 4.34959165501488e+00, 4.34920404319783e+00, 4.34893801226267e+00, 4.34876372443233e+00, 4.34865591926812e+00, 4.34859393680343e+00, 4.34856157486228e+00, 4.34854678807959e+00, 4.34854124099118e+00, 4.34853973204337e+00, 4.34853950932564e+00, 4.34853949146092e+00, 4.34853916176622e+00, 4.34853692803403e+00, 4.34852871656691e+00, 4.34850682757006e+00, 4.34845892281621e+00, 4.34836717480812e+00, 4.34820760962730e+00, 4.34794967060763e+00, 4.34755602420685e+00, 4.34698262305643e+00, 4.34617903424154e+00, 4.34508903349712e+00, 4.34365145832053e+00, 4.34180130514091e+00, 4.33947104782426e+00, 4.33659214713773e+00, 4.33309671357843e+00, 4.32891927944330e+00, 4.32399863043688e+00, 4.31827964274391e+00, 4.31171506858458e+00, 4.30426721205715e+00, 4.29590943776670e+00, 4.28662745751495e+00, 4.27642034530853e+00, 4.26530123817574e+00, 4.25329768970256e+00, 4.24045165460118e+00, 4.22681909563517e+00, 4.21246921829084e+00, 4.19748335298803e+00, 4.18195351852457e+00, 4.16598071296251e+00, 4.14967298844523e+00, 4.13314337378475e+00, 4.11650771260647e+00, 4.09988248520617e+00, 4.08338267919422e+00, 4.06711976790177e+00, 4.05119984706165e+00, 4.03572197025888e+00, 4.02077671292588e+00, 4.00644498404146e+00, 3.99279709485863e+00, 3.97989208543569e+00, 3.96777730277286e+00, 3.95648821906068e+00, 3.94604847485025e+00, 3.93647012964765e+00, 3.92775410121960e+00, 3.91989077443713e+00, 3.91286076044579e+00, 3.90663578703568e+00, 3.90117970105823e+00, 3.89644956343800e+00, 3.89239681668222e+00, 3.88896850380532e+00, 3.88610851634295e+00, 3.88375884777635e+00, 3.88186082741617e+00, 3.88035630882271e+00, 3.87918878639039e+00, 3.87830441399989e+00, 3.87765290080607e+00, 3.87718826138555e+00, 3.87686940064880e+00, 3.87666051808098e+00, 3.87653132089119e+00, 3.87645704133087e+00, 3.87641825954424e+00, 3.87640053956030e+00, 3.87639389213963e+00, 3.87639208387598e+00, 3.87639181697991e+00, 3.87639179639431e+00, 3.87639142674308e+00, 3.87638892230028e+00, 3.87637971568342e+00, 3.87635517406801e+00, 3.87630146455608e+00, 3.87619860140027e+00, 3.87601971207448e+00, 3.87573055403637e+00, 3.87528930811335e+00, 3.87464666775985e+00, 3.87374623595375e+00, 3.87252523320071e+00, 3.87091551100731e+00, 3.86884485534340e+00, 3.86623855420677e+00, 3.86302119271125e+00, 3.85911862853984e+00, 3.85446009065076e+00, 3.84898033539728e+00, 3.84262178737544e+00, 3.83533658800090e+00, 3.82708847363201e+00, 3.81785440746875e+00, 3.80762589574388e+00, 3.79640992891374e+00, 3.78422950238499e+00, 3.77112368820537e+00, 3.75714724821579e+00, 3.74236979926979e+00, 3.72687456095669e+00, 3.71075673444582e+00, 3.69412157631103e+00, 3.67708224243298e+00, 3.65975748359888e+00, 3.64226927594705e+00, 3.62474046612201e+00, 3.60729250353180e+00, 3.59004332138774e+00, 3.57310541542393e+00, 3.55658415556265e+00, 3.54057635245343e+00, 3.52516908870729e+00, 3.51043881444050e+00, 3.49645069879756e+00, 3.48325822351935e+00, 3.47090300119500e+00, 3.45941479925080e+00, 3.44881175054833e+00, 3.43910073221612e+00, 3.43027789557426e+00, 3.42232933133839e+00, 3.41523185539804e+00, 3.40895390113747e+00, 3.40345650437725e+00, 3.39869436652032e+00, 3.39461698042016e+00, 3.39116980194884e+00, 3.38829544838076e+00, 3.38593490271547e+00, 3.38402870116348e+00, 3.38251807944645e+00, 3.38134605255329e+00, 3.38045840235864e+00, 3.37980454822431e+00, 3.37933827748455e+00, 3.37901831560670e+00, 3.37880871978413e+00, 3.37867908463900e+00, 3.37860455439248e+00, 3.37856564204255e+00, 3.37854786247565e+00, 3.37854119271508e+00, 3.37853937837554e+00, 3.37853911058272e+00, 3.37853909030626e+00, 3.37853873111068e+00, 3.37853629750693e+00, 3.37852735130897e+00, 3.37850350392893e+00, 3.37845131400907e+00, 3.37835136210724e+00, 3.37817753877197e+00, 3.37789657756475e+00, 3.37746785907309e+00, 3.37684350565462e+00, 3.37596877947217e+00, 3.37478278822914e+00, 3.37321949384053e+00, 3.37120900910440e+00, 3.36867915642459e+00, 3.36555725108912e+00, 3.36177206001913e+00, 3.35725587593154e+00, 3.35194663730781e+00, 3.34579001731324e+00, 3.33874140074205e+00, 3.33076766793158e+00, 3.32184870892708e+00, 3.31197860018252e+00, 3.30116638953958e+00, 3.28943645247540e+00, 3.27682840256818e+00, 3.26339656037537e+00, 3.24920900583441e+00, 3.23434625824146e+00, 3.21889964338186e+00, 3.20296941837480e+00, 3.18666273064211e+00, 3.17009148805989e+00, 3.15337021327584e+00, 3.13661394729064e+00, 3.11993625690618e+00, 3.10344738883909e+00, 3.08725260141058e+00, 3.07145069376091e+00, 3.05613274319752e+00, 3.04138105393129e+00, 3.02726831513503e+00, 3.01385696277782e+00, 3.00119873769319e+00, 2.98933443140575e+00, 2.97829381094908e+00, 2.96809571389324e+00, 2.95874830478257e+00, 2.95024948397001e+00, 2.94258743931295e+00, 2.93574133031821e+00, 2.92968209309086e+00, 2.92437335288460e+00, 2.91977242922732e+00, 2.91583141658012e+00, 2.91249832137914e+00, 2.90971823422279e+00, 2.90743451404055e+00, 2.90558995946718e+00, 2.90412794151095e+00, 2.90299347111019e+00, 2.90213417546622e+00, 2.90150115824020e+00, 2.90104972087947e+00, 2.90073992551010e+00, 2.90053698394532e+00, 2.90041146229560e+00, 2.90033929624265e+00, 2.90030161802394e+00, 2.90028440229818e+00, 2.90027794404684e+00, 2.90027618724228e+00, 2.90027592794148e+00, 2.90027590847568e+00, 2.90027556585560e+00, 2.90027324455297e+00, 2.90026471118104e+00, 2.90024196420692e+00, 2.90019218231225e+00, 2.90009684135291e+00, 2.89993103425062e+00, 2.89966302340325e+00, 2.89925404973661e+00, 2.89865841632270e+00, 2.89782385751263e+00, 2.89669219673188e+00, 2.89520028744688e+00, 2.89328122239042e+00, 2.89086578608935e+00, 2.88788411536891e+00, 2.88426752226222e+00, 2.87995042422649e+00, 2.87487231847263e+00, 2.86897973131687e+00, 2.86222807049789e+00, 2.85458330896864e+00, 2.84602343312343e+00, 2.83653959678024e+00, 2.82613693414163e+00, 2.81483499965665e+00, 2.80266781913324e+00, 2.78968355333782e+00, 2.77594379136572e+00, 2.76152250511214e+00, 2.74650470735599e+00, 2.73098486382492e+00, 2.71506511410555e+00, 2.69885335774217e+00, 2.68246126093847e+00, 2.66600223665569e+00, 2.64958944725734e+00, 2.63333387468734e+00, 2.61734249875934e+00, 2.60171661952912e+00, 2.58655035481517e+00, 2.57192933856295e+00, 2.55792963980785e+00, 2.54461691550671e+00, 2.53204580366891e+00, 2.52025955637064e+00, 2.50928990580702e+00, 2.49915715095295e+00, 2.48987044799934e+00, 2.48142828468464e+00, 2.47381911694186e+00, 2.46702214574516e+00, 2.46100821235020e+00, 2.45574079089935e+00, 2.45117705822655e+00, 2.44726902133812e+00, 2.44396468325849e+00, 2.44120922764305e+00, 2.43894620182809e+00, 2.43711867697734e+00, 2.43567036294281e+00, 2.43454665466873e+00, 2.43369558671980e+00, 2.43306867305537e+00, 2.43262161068231e+00, 2.43231482839257e+00, 2.43211386542240e+00, 2.43198956945507e+00, 2.43191810874112e+00, 2.43188079897430e+00, 2.43186375164326e+00, 2.43185735657004e+00, 2.43185561695216e+00, 2.43185536018810e+00, 2.43185534097136e+00, 2.43185500351564e+00, 2.43185271720258e+00, 2.43184431246036e+00, 2.43182190839932e+00, 2.43177287713151e+00, 2.43167897436422e+00, 2.43151567033294e+00, 2.43125171061350e+00, 2.43084893114371e+00, 2.43026234586146e+00, 2.42944051867398e+00, 2.42832622396029e+00, 2.42685739146131e+00, 2.42496832228504e+00, 2.42259115300090e+00, 2.41965753468518e+00, 2.41610048369844e+00, 2.41185635141750e+00, 2.40686685168034e+00, 2.40108107792998e+00, 2.39445743753483e+00, 2.38696542900295e+00, 2.37858718914443e+00, 2.36931874184795e+00, 2.35917088802451e+00, 2.34816968727912e+00, 2.33635649571970e+00, 2.32378754064417e+00, 2.31053303123103e+00, 2.29667582427555e+00, 2.28230968477397e+00, 2.26753720180156e+00, 2.25246743933950e+00, 2.23721341779000e+00, 2.22188953295055e+00, 2.20660902332787e+00, 2.19148159248230e+00, 2.17661128018771e+00, 2.16209465543654e+00, 2.14801937792921e+00, 2.13446314590685e+00, 2.12149302068986e+00, 2.10916509541115e+00, 2.09752445952332e+00, 2.08660540266781e+00, 2.07643180097491e+00, 2.06701763427948e+00, 2.05836759196240e+00, 2.05047773596535e+00, 2.04333620010899e+00, 2.03692391385911e+00, 2.03121534541446e+00, 2.02617926323310e+00, 2.02177951703891e+00, 2.01797583934047e+00, 2.01472466702230e+00, 2.01197998010401e+00, 2.00969415173402e+00, 2.00781880025313e+00, 2.00630563103646e+00, 2.00510725305591e+00, 2.00417795292134e+00, 2.00347440774214e+00, 2.00295631765525e+00, 2.00258693939832e+00, 2.00233350391721e+00, 2.00216750367070e+00, 2.00206483893558e+00, 2.00200581685735e+00, 2.00197500199869e+00, 2.00196092243188e+00, 2.00195564069974e+00, 2.00195420394049e+00, 2.00195399187783e+00, 2.00195397600069e+00, 2.00195369711075e+00, 2.00195180759064e+00, 2.00194486151859e+00, 2.00192634586961e+00, 2.00188582497114e+00, 2.00180822316382e+00, 2.00167327472308e+00, 2.00145516686594e+00, 2.00112239789057e+00, 2.00063786802702e+00, 1.99995921521391e+00, 1.99903940154421e+00, 1.99782754836432e+00, 1.99627000889031e+00, 1.99431165680533e+00, 1.99189735792633e+00, 1.98897358022928e+00, 1.98549008611645e+00, 1.98140164084070e+00, 1.97666966367240e+00, 1.97126374493262e+00, 1.96516295351958e+00, 1.95835686678160e+00, 1.95084626775870e+00, 1.94264347342032e+00, 1.93377228024101e+00, 1.92426753815390e+00, 1.91417438788721e+00, 1.90354721700124e+00, 1.89244840400241e+00, 1.88094692599364e+00, 1.86911690300243e+00, 1.85703614244298e+00, 1.84478473239847e+00, 1.83244371556434e+00, 1.82009385981010e+00, 1.80781452879401e+00, 1.79568264822880e+00, 1.78377176039001e+00, 1.77215116044301e+00, 1.76088511171394e+00, 1.75003214157910e+00, 1.73964442384966e+00, 1.72976725647640e+00, 1.72043864467198e+00, 1.71168899915125e+00, 1.70354095741623e+00, 1.69600933329798e+00, 1.68910119676249e+00, 1.68281608267396e+00, 1.67714632405360e+00, 1.67207750252623e+00, 1.66758900615933e+00, 1.66365468274946e+00, 1.66024357474530e+00, 1.65732072035435e+00, 1.65484800392240e+00, 1.65278503739112e+00, 1.65109005356030e+00, 1.64972079106821e+00, 1.64863535054534e+00, 1.64779300139187e+00, 1.64715491917525e+00, 1.64668483481872e+00, 1.64634957860431e+00, 1.64611950455031e+00, 1.64596878390570e+00, 1.64587556024522e+00, 1.64582196281820e+00, 1.64579397923763e+00, 1.64578119310452e+00, 1.64577639655156e+00, 1.64577509176974e+00, 1.64577489918656e+00, 1.64577488472960e+00, 1.64577463027490e+00, 1.64577290630584e+00, 1.64576656879710e+00, 1.64574967520204e+00, 1.64571270335535e+00, 1.64564189544276e+00, 1.64551875236724e+00, 1.64531970021984e+00, 1.64501594640214e+00, 1.64457353825380e+00, 1.64395363180305e+00, 1.64311297247279e+00, 1.64200458325368e+00, 1.64057864905926e+00, 1.63878357883696e+00, 1.63656721974416e+00, 1.63387819062910e+00, 1.63066729559201e+00, 1.62688897301604e+00, 1.62250273163735e+00, 1.61747452340460e+00, 1.61177800337145e+00, 1.60539562979352e+00, 1.59831956286258e+00, 1.59055232778086e+00, 1.58210721665758e+00, 1.57300841340663e+00, 1.56329083584981e+00, 1.55299969908398e+00, 1.54218981348451e+00, 1.53092463925548e+00, 1.51927512704906e+00, 1.50731838073177e+00, 1.49513618370944e+00, 1.48281343411252e+00, 1.47043653632867e+00, 1.45809179661557e+00, 1.44586386869342e+00, 1.43383429133324e+00, 1.42208015424562e+00, 1.41067292145303e+00, 1.39967743335111e+00, 1.38915110045409e+00, 1.37914329397923e+00, 1.36969493146417e+00, 1.36083824989112e+00, 1.35259675450163e+00, 1.34498532864303e+00, 1.33801048846402e+00, 1.33167076583443e+00, 1.32595720320295e+00, 1.32085394490277e+00, 1.31633891036738e+00, 1.31238453556728e+00, 1.30895856953571e+00, 1.30602491300477e+00, 1.30354448589297e+00, 1.30147610971641e+00, 1.29977739005159e+00, 1.29840558311875e+00, 1.29731842957643e+00, 1.29647493792355e+00, 1.29583609968842e+00, 1.29536551900714e+00, 1.29502994036846e+00, 1.29479966028295e+00, 1.29464881140307e+00, 1.29455551109845e+00, 1.29450187053136e+00, 1.29447386469559e+00, 1.29446106845507e+00, 1.29445626812037e+00, 1.29445496231076e+00, 1.29445476957591e+00, 1.29445475504595e+00, 1.29445449848549e+00, 1.29445276024997e+00, 1.29444637030237e+00, 1.29442933697149e+00, 1.29439205955460e+00, 1.29432066735170e+00, 1.29419651105003e+00, 1.29399582894636e+00, 1.29368960654222e+00, 1.29324364354490e+00, 1.29261883720581e+00, 1.29177168517918e+00, 1.29065500466637e+00, 1.28921885755706e+00, 1.28741166369855e+00, 1.28518147652179e+00, 1.28247738733501e+00, 1.27925101706188e+00, 1.27545804752890e+00, 1.27105973911505e+00, 1.26602437818608e+00, 1.26032859671671e+00, 1.25395850822420e+00, 1.24691060881307e+00, 1.23919239978750e+00, 1.23082269872578e+00, 1.22183161869943e+00, 1.21226020980024e+00, 1.20215977246142e+00, 1.19159086723584e+00, 1.18062205967793e+00, 1.16932845075935e+00, 1.15779005196823e+00, 1.14609006927915e+00, 1.13431316123899e+00, 1.12254373355884e+00, 1.11086432625491e+00, 1.09935414026599e+00, 1.08808773952206e+00, 1.07713395266842e+00, 1.06655498705119e+00, 1.05640575699416e+00, 1.04673341948277e+00, 1.03757710350731e+00, 1.02896781464081e+00, 1.02092849385598e+00, 1.01347420886354e+00, 1.00661245701453e+00, 1.00034356062775e+00, 9.94661138051440e-01, 9.89552636447416e-01, 9.84999914858320e-01, 9.80979868317322e-01, 9.77465085398642e-01, 9.74424532579846e-01, 9.71824259061920e-01, 9.69628115309593e-01, 9.67798477632018e-01, 9.66296969774008e-01, 9.65085170920450e-01, 9.64125297946129e-01, 9.63380848392162e-01, 9.62817189730707e-01, 9.62402080174265e-01, 9.62106106730845e-01, 9.61903027477232e-01, 9.61770007124973e-01, 9.61687737821908e-01, 9.61640440632879e-01, 9.61615747090067e-01, 9.61604464372687e-01, 9.61600231830420e-01, 9.61599080475786e-01, 9.61598910538235e-01, 9.61598897657202e-01, 9.61598669287745e-01, 9.61597122051107e-01, 9.61591434246245e-01, 9.61576272652000e-01, 9.61543091900452e-01, 9.61479546893859e-01, 9.61369041914282e-01, 9.61190437351028e-01, 9.60917931803722e-01, 9.60521133614081e-01, 9.59965331508802e-01, 9.59211968865556e-01, 9.58219320042051e-01, 9.56943360204434e-01, 9.55338812236377e-01, 9.53360345824235e-01, 9.50963895073599e-01, 9.48108052548395e-01, 9.44755490092550e-01, 9.40874350929454e-01, 9.36439554073588e-01, 9.31433951683750e-01, 9.25849283111517e-01, 9.19686876256431e-01, 9.12958057301412e-01, 9.05684243470269e-01, 8.97896709272040e-01, 8.89636033629607e-01, 8.80951252005573e-01, 8.71898752771914e-01, 8.62540969379319e-01, 8.52944928400457e-01, 8.43180717674271e-01, 8.33319938448033e-01, 8.23434200937093e-01, 8.13593714824949e-01, 8.03866015906587e-01, 7.94314858462855e-01, 7.84999291150496e-01, 7.75972923166273e-01, 7.67283377927970e-01, 7.58971923973013e-01, 7.51073267399610e-01, 7.43615486924441e-01, 7.36620091285573e-01, 7.30102178944778e-01, 7.24070681450113e-01, 7.18528674010843e-01, 7.13473739444705e-01, 7.08898374364402e-01, 7.04790429017947e-01, 7.01133574391000e-01, 6.97907791882326e-01, 6.95089881995180e-01, 6.92653989016531e-01, 6.90572138596528e-01, 6.88814784548343e-01, 6.87351360156216e-01, 6.86150827933574e-01, 6.85182220266194e-01, 6.84415161878579e-01, 6.83820363753168e-01, 6.83370077183737e-01, 6.83038496207170e-01, 6.82802096847330e-01, 6.82639902488138e-01, 6.82533666280816e-01, 6.82467963732316e-01, 6.82430191411387e-01, 6.82410470890638e-01, 6.82401460429785e-01, 6.82398080296132e-01, 6.82397160817943e-01, 6.82397025104894e-01, 6.82397014752755e-01, 6.82396830357281e-01, 6.82395581051112e-01, 6.82390988473927e-01, 6.82378746383507e-01, 6.82351955041526e-01, 6.82300647081282e-01, 6.82211424188082e-01, 6.82067221857281e-01, 6.81847217077629e-01, 6.81526890912962e-01, 6.81078254481948e-01, 6.80470242614570e-01, 6.79669274375716e-01, 6.78639973641053e-01, 6.77346036048544e-01, 6.75751221134511e-01, 6.73820440661767e-01, 6.71520906585439e-01, 6.68823295437634e-01, 6.65702880877386e-01, 6.62140583471141e-01, 6.58123887054527e-01, 6.53647574676413e-01, 6.48714244232599e-01, 6.43334574185150e-01, 6.37527322586242e-01, 6.31319057007757e-01, 6.24743627721383e-01, 6.17841410302383e-01, 6.10658355552899e-01, 6.03244893306451e-01, 5.95654741689848e-01, 5.87943674602393e-01, 5.80168297755425e-01, 5.72384878170751e-01, 5.64648264372120e-01, 5.57010925533857e-01, 5.49522128472881e-01, 5.42227262368133e-01, 5.35167313069410e-01, 5.28378482215454e-01, 5.21891941318817e-01, 5.15733707519808e-01, 5.09924625758847e-01, 5.04480441470234e-01, 4.99411948313575e-01, 4.94725196665023e-01, 4.90421750325435e-01, 4.86498980922578e-01, 4.82950391574704e-01, 4.79765963362563e-01, 4.76932519882130e-01, 4.74434106514073e-01, 4.72252381977813e-01, 4.70367020203786e-01, 4.68756120557956e-01, 4.67396624022879e-01, 4.66264732146245e-01, 4.65336324506906e-01, 4.64587369239235e-01, 4.63994319935684e-01, 4.63534491158318e-01, 4.63186403972996e-01, 4.62930092498722e-01, 4.62747362535692e-01, 4.62621993956138e-01, 4.62539879723963e-01, 4.62489096115370e-01, 4.62459900859685e-01, 4.62444658385294e-01, 4.62437693991160e-01, 4.62435081409446e-01, 4.62434370724154e-01, 4.62434265828503e-01, 4.62434257770640e-01, 4.62434113501295e-01, 4.62433136055591e-01, 4.62429542866841e-01, 4.62419964783719e-01, 4.62399003580195e-01, 4.62358861123152e-01, 4.62289055417179e-01, 4.62176237184134e-01, 4.62004118801723e-01, 4.61753525226269e-01, 4.61402573835895e-01, 4.60926986820041e-01, 4.60300535703054e-01, 4.59495612775495e-01, 4.58483918662634e-01, 4.57237249152332e-01, 4.55728358039486e-01, 4.53931866562900e-01, 4.51825184565289e-01, 4.49389404411933e-01, 4.46610126561406e-01, 4.43478175997023e-01, 4.39990171818428e-01, 4.36148918214743e-01, 4.31963593543769e-01, 4.27449724768730e-01, 4.22628946237286e-01, 4.17528553741572e-01, 4.12180875939865e-01, 4.06622494602716e-01, 4.00893352023483e-01, 3.95035787847668e-01, 3.89093548395448e-01, 3.83110809462421e-01, 3.77131249033734e-01, 3.71197199970006e-01, 3.65348905243125e-01, 3.59623890440213e-01, 3.54056460663188e-01, 3.48677322149643e-01, 3.43513323291443e-01, 3.38587305431106e-01, 3.33918050924960e-01, 3.29520314404756e-01, 3.25404922784227e-01, 3.21578930125869e-01, 3.18045814761158e-01, 3.14805707799448e-01, 3.11855644139164e-01, 3.09189829110989e-01, 3.06799915773692e-01, 3.04675289523087e-01, 3.02803357974151e-01, 3.01169844980474e-01, 2.99759088140804e-01, 2.98554339215718e-01, 2.97538066571589e-01, 2.96692258142287e-01, 2.95998722531627e-01, 2.95439384869906e-01, 2.94996572997446e-01, 2.94653288593873e-01, 2.94393457118757e-01, 2.94202149979085e-01, 2.94065772272226e-01, 2.93972209819179e-01, 2.93910930014888e-01, 2.93873032254439e-01, 2.93851245281824e-01, 2.93839870656852e-01, 2.93834673528308e-01, 2.93832723911072e-01, 2.93832193568394e-01, 2.93832115290943e-01, 2.93832109231946e-01, 2.93832000153216e-01, 2.93831261129174e-01, 2.93828544404643e-01, 2.93821302660734e-01, 2.93805454515994e-01, 2.93775104322142e-01, 2.93722327894431e-01, 2.93637034711704e-01, 2.93506915815810e-01, 2.93317485047597e-01, 2.93052219263754e-01, 2.92692800665828e-01, 2.92219461240297e-01, 2.91611425507463e-01, 2.90847443337751e-01, 2.89906399651641e-01, 2.88767982628253e-01, 2.87413386979449e-01, 2.85826024370120e-01, 2.83992209704232e-01, 2.81901790264981e-01, 2.79548685027780e-01, 2.76931304127176e-01, 2.74052823498117e-01, 2.70921296907408e-01, 2.67549596455613e-01, 2.63955182449995e-01, 2.60159713467250e-01, 2.56188516545112e-01, 2.52069944949302e-01, 2.47834656227481e-01, 2.43514845918607e-01, 2.39143472264621e-01, 2.34753504787424e-01, 2.30377225096867e-01, 2.26045602388109e-01, 2.21787759436354e-01, 2.17630538147206e-01, 2.13598167424186e-01, 2.09712030687180e-01, 2.05990526077868e-01, 2.02449009331564e-01, 1.99099807461038e-01, 1.95952290670638e-01, 1.93012990118108e-01, 1.90285750054312e-01, 1.87771904278740e-01, 1.85470468545485e-01, 1.83378342359462e-01, 1.81490515366294e-01, 1.79800275144850e-01, 1.78299414573355e-01, 1.76978438001601e-01, 1.75826766191699e-01, 1.74832940378726e-01, 1.73984825861785e-01, 1.73269815294288e-01, 1.72675031346173e-01, 1.72187527721345e-01, 1.71794486705583e-01, 1.71483410577136e-01, 1.71242303423144e-01, 1.71059839256917e-01, 1.70925511902849e-01, 1.70829761970062e-01, 1.70764076413113e-01, 1.70721056691110e-01, 1.70694452368312e-01, 1.70679158104127e-01, 1.70671173288026e-01, 1.70667524997991e-01, 1.70666156404658e-01, 1.70665784114673e-01, 1.70665729165473e-01, 1.70665724877696e-01, 1.70665647240370e-01, 1.70665121236490e-01, 1.70663187596894e-01, 1.70658033272160e-01, 1.70646753404601e-01, 1.70625152051629e-01, 1.70587590067877e-01, 1.70526887836901e-01, 1.70434289688139e-01, 1.70299495764755e-01, 1.70110765723836e-01, 1.69855096847548e-01, 1.69518476833569e-01, 1.69086208667826e-01, 1.68543301581346e-01, 1.67874918261767e-01, 1.67066864435052e-01, 1.66106102959965e-01, 1.64981271074103e-01, 1.63683176828215e-01, 1.62205249469316e-01, 1.60543918933370e-01, 1.58698901896916e-01, 1.56673376031472e-01, 1.54474029997484e-01, 1.52110983876186e-01, 1.49597582555741e-01, 1.46950072341286e-01, 1.44187178012644e-01, 1.41329603061496e-01, 1.38399479429345e-01, 1.35419794499066e-01, 1.32413822378223e-01, 1.29404583896210e-01, 1.26414355642134e-01, 1.23464243325454e-01, 1.20573829302078e-01, 1.17760898787751e-01, 1.15041244494191e-01, 1.12428545462913e-01, 1.09934312895678e-01, 1.07567893826530e-01, 1.05336522489236e-01, 1.03245409077073e-01, 1.01297856102598e-01, 9.94953935623961e-02, 9.78379254210751e-02, 9.63238813944347e-02, 9.49503695035707e-02, 9.37133262872774e-02, 9.26076628230890e-02, 9.16274057645063e-02, 9.07658334200816e-02, 9.00156074626457e-02, 8.93689011615395e-02, 8.88175250881327e-02, 8.83530510768065e-02, 8.79669348635144e-02, 8.76506373149724e-02, 8.73957435565238e-02, 8.71940786648224e-02, 8.70378179766202e-02, 8.69195895398782e-02, 8.68325658561240e-02, 8.67705418802602e-02, 8.67279962871361e-02, 8.67001332949497e-02, 8.66829028442355e-02, 8.66729976373195e-02, 8.66678263981582e-02, 8.66654636554394e-02, 8.66645773154087e-02, 8.66643362100268e-02, 8.66643006234016e-02, 8.66642978228906e-02, 8.66642468122389e-02, 8.66639012079595e-02, 8.66626307360283e-02, 8.66592441700824e-02, 8.66518329855923e-02, 8.66376405706237e-02, 8.66129626266732e-02, 8.65730839846766e-02, 8.65122565655696e-02, 8.64237225685649e-02, 8.62997860919155e-02, 8.61319351944729e-02, 8.59110148173912e-02, 8.56274489620779e-02, 8.52715080726192e-02, 8.48336147771027e-02, 8.43046781606231e-02, 8.36764438100959e-02, 8.29418442889240e-02, 8.20953328047063e-02, 8.11331819539660e-02, 8.00537298319055e-02, 7.88575576368586e-02, 7.75475861758546e-02, 7.61290832083983e-02, 7.46095789952781e-02, 7.29986932563520e-02, 7.13078824202462e-02, 6.95501210091773e-02, 6.77395347681911e-02, 6.58910053950930e-02, 6.40197673192988e-02, 6.21410159749827e-02, 6.02695446475270e-02, 5.84194235973063e-02, 5.66037311994352e-02, 5.48343427026765e-02, 5.31217782760933e-02, 5.14751085651781e-02, 4.99019132077638e-02, 4.84082857538555e-02, 4.69988771994361e-02, 4.56769698241839e-02, 4.44445731141663e-02, 4.33025341244084e-02, 4.22506555554421e-02, 4.12878159499996e-02, 4.04120876443850e-02, 3.96208493373065e-02, 3.89108912919971e-02, 3.82785122103426e-02, 3.77196076721855e-02, 3.72297506963957e-02, 3.68042654428109e-02, 3.64382953366622e-02, 3.61268669705259e-02, 3.58649510407103e-02, 3.56475213306444e-02, 3.54696123938210e-02, 3.53263761500280e-02, 3.52131371317403e-02, 3.51254456460994e-02, 3.50591276959733e-02, 3.50103301710982e-02, 3.49755596115422e-02, 3.49517127836769e-02, 3.49360974042261e-02, 3.49264415969790e-02, 3.49208910506311e-02, 3.49179933346297e-02, 3.49166693829357e-02, 3.49161727292060e-02, 3.49160376279416e-02, 3.49160176872978e-02, 3.49160161041642e-02, 3.49159870912548e-02, 3.49157905248729e-02, 3.49150679315580e-02, 3.49131418052373e-02, 3.49089267388338e-02, 3.49008552021637e-02, 3.48868212973077e-02, 3.48641456625396e-02, 3.48295644910888e-02, 3.47792453291982e-02, 3.47088318551974e-02, 3.46135191473147e-02, 3.44881599634137e-02, 3.43274012429431e-02, 3.41258484033125e-02, 3.38782530933264e-02, 3.35797180001256e-02, 3.32259102581247e-02, 3.28132732033843e-02, 3.23392249055976e-02, 3.18023313385449e-02, 3.12024424161015e-02, 3.05407805367004e-02, 2.98199737357407e-02, 2.90440288973262e-02, 2.82182444473840e-02, 2.73490661615918e-02, 2.64438937438785e-02, 2.55108492442417e-02, 2.45585208377239e-02, 2.35956967453754e-02, 2.26311040549207e-02, 2.16731659571927e-02, 2.07297886574116e-02, 1.98081862558590e-02, 1.89147485832883e-02, 1.80549536768249e-02, 1.72333235884086e-02, 1.64534197377355e-02, 1.57178721772366e-02, 1.50284359856530e-02, 1.43860675447396e-02, 1.37910136158936e-02, 1.32429067876512e-02, 1.27408618351019e-02, 1.22835686442908e-02, 1.18693784841648e-02, 1.14963814820523e-02, 1.11624741408551e-02, 1.08654165937327e-02, 1.06028799896771e-02, 1.03724849150000e-02, 1.01718320856579e-02, 9.99852671050316e-03, 9.85019796869846e-03, 9.72451498810489e-03, 9.61920057796650e-03, 9.53204377055528e-03, 9.46091197791610e-03, 9.40376328262005e-03, 9.35865908411149e-03, 9.32377702585274e-03, 9.29742386382504e-03, 9.27804771396534e-03, 9.26424895700428e-03, 9.25478899580501e-03, 9.24859605339279e-03, 9.24476727807821e-03, 9.24256656732059e-03, 9.24141772991405e-03, 9.24089284678775e-03, 9.24069594987031e-03, 9.24064238954969e-03, 9.24063448419332e-03, 9.24063385082885e-03, 9.24062217117516e-03, 9.24054304008237e-03, 9.24025214989133e-03, 9.23947677477788e-03, 9.23778004431242e-03, 9.23453121180096e-03, 9.22888337112493e-03, 9.21976007617431e-03, 9.20585230032245e-03, 9.18562710923077e-03, 9.15734925972966e-03, 9.11911663429876e-03, 9.06890993576859e-03, 9.00465636667658e-03, 8.92430610563988e-03, 8.82591930120125e-03, 8.70776010728580e-03, 8.56839310273383e-03, 8.40677642368792e-03, 8.22234526266996e-03, 8.01507920872968e-03, 7.78554735042380e-03, 7.53492617344342e-03, 7.26498703340184e-03, 6.97805223225778e-03, 6.67692127540409e-03, 6.36477145749392e-03, 6.04503921834443e-03, 5.72129039813341e-03, 5.39708832876418e-03, 5.07586847477004e-03, 4.76082717412840e-03, 4.45483022693782e-03, 4.16034508188989e-03, 3.87939844910434e-03, 3.61355934245822e-03, 3.36394565854370e-03, 3.13125047240526e-03, 2.91578273598425e-03, 2.71751669709452e-03, 2.53614531322288e-03, 2.37113454656601e-03, 2.22177659157108e-03, 2.08724012028050e-03, 1.96661522922970e-03, 1.85895099752997e-03, 1.76328485798738e-03, 1.67866442905051e-03, 1.60416329492730e-03, 1.53889216542069e-03, 1.48200637515483e-03, 1.43271020859093e-03, 1.39025844571439e-03, 1.35395568999819e-03, 1.32315421959461e-03, 1.29725108725331e-03, 1.27568505694686e-03, 1.25793375860450e-03, 1.24351130316125e-03, 1.23196648013746e-03, 1.22288158322004e-03, 1.21587185017091e-03, 1.21058543168751e-03, 1.20670375157150e-03, 1.20394207138297e-03, 1.20205003759999e-03, 1.20081198634736e-03, 1.20004678661534e-03, 1.19960704113620e-03, 1.19937750299570e-03, 1.19927263626263e-03, 1.19923329892213e-03, 1.19922259837235e-03, 1.19922101901171e-03, 1.19922089128988e-03, 1.19921852083671e-03, 1.19920246138427e-03, 1.19914342633454e-03, 1.19898607223590e-03, 1.19864176528051e-03, 1.19798260069537e-03, 1.19683700981010e-03, 1.19498731379419e-03, 1.19216960009538e-03, 1.18807630776008e-03, 1.18236187055665e-03, 1.17465169694864e-03, 1.16455461776699e-03, 1.15167871758626e-03, 1.13565018173212e-03, 1.11613444722458e-03, 1.09285858326125e-03, 1.06563346440769e-03, 1.03437399298034e-03, 9.99115411114637e-04, 9.60023703253654e-04, 9.17398327755947e-04, 8.71666180299435e-04, 8.23366781820489e-04, 7.73129939534378e-04, 7.21648041631440e-04, 6.69645374808958e-04, 6.17846448881743e-04, 5.66944692642402e-04, 5.17572621808071e-04, 4.70275239861095e-04, 4.25489926941580e-04, 3.83536670824469e-04, 3.44620169917320e-04, 3.08841292423701e-04, 2.76213082690546e-04, 2.46677036650489e-04, 2.20117206001902e-04, 1.96371473194290e-04, 1.75240610881922e-04, 1.56496587555566e-04, 1.39892095227800e-04, 1.25173197876194e-04, 1.12099059901283e-04, 1.00496920013349e-04, 9.01449460544157e-05, 8.08257314250937e-05, 7.23853572877076e-05, 6.46987711059310e-05, 5.76659321875894e-05, 5.12081145371823e-05, 4.52595293962688e-05, 3.98337329547971e-05, 3.48164207801815e-05, 3.01995702747249e-05, 2.59688891333099e-05, 2.21130865876162e-05, 1.86229678478256e-05, 1.54903927827021e-05, 1.27064752142303e-05, 1.02595707485165e-05, 8.13534641697089e-06, 6.31754734392587e-06, 4.78779745805014e-06, 3.52540153603087e-06, 2.50745497698546e-06, 1.70914817143597e-06, 1.10417747389846e-06, 6.65218493622888e-07, 3.64441909539437e-07, -6.21724893790088e-15, -6.21724893790088e-15, -6.21724893790088e-15, -6.21724893790088e-15],
[ 6.00000000000000e+00, 5.99999999814398e+00, 5.99999994075207e+00, 5.99999955191290e+00, 5.99999812249059e+00, 5.99999431211119e+00, 5.99998597286262e+00, 5.99997000095285e+00, 5.99994222175216e+00, 5.99989731266337e+00, 5.99982876714950e+00, 5.99972890203367e+00, 5.99958890890969e+00, 5.99939894920284e+00, 5.99914829113896e+00, 5.99882548565512e+00, 5.99841857715241e+00, 5.99791534398307e+00, 5.99730356270537e+00, 5.99657128945113e+00, 5.99570715124480e+00, 5.99470063979613e+00, 5.99354240016193e+00, 5.99222450673057e+00, 5.99074071921930e+00, 5.98908671177653e+00, 5.98726026883963e+00, 5.98526144209858e+00, 5.98309266374746e+00, 5.98075881215448e+00, 5.97826722713590e+00, 5.97562767316364e+00, 5.97285225005387e+00, 5.96995525195267e+00, 5.96695297672866e+00, 5.96386348917106e+00, 5.96070634263817e+00, 5.95750226496848e+00, 5.95427281551283e+00, 5.95104002103217e+00, 5.94782599889445e+00, 5.94465257646559e+00, 5.94154091579901e+00, 5.93851115267338e+00, 5.93558205870398e+00, 5.93277073466851e+00, 5.93009234235894e+00, 5.92755988122525e+00, 5.92518401484817e+00, 5.92297295090732e+00, 5.92093237684198e+00, 5.91906545188169e+00, 5.91737285459872e+00, 5.91585288365031e+00, 5.91450160797891e+00, 5.91331306146167e+00, 5.91227947588206e+00, 5.91139154516530e+00, 5.91063871309874e+00, 5.91000947626549e+00, 5.90949169366415e+00, 5.90907289447296e+00, 5.90874057563965e+00, 5.90848248142750e+00, 5.90828685770927e+00, 5.90814267464982e+00, 5.90803981243126e+00, 5.90796920581940e+00, 5.90792294461640e+00, 5.90789432835490e+00, 5.90787787493044e+00, 5.90786928420427e+00, 5.90786535890362e+00, 5.90786388636729e+00, 5.90786348579977e+00, 5.90786342667677e+00, 5.90786341892108e+00, 5.90786323822276e+00, 5.90786201396323e+00, 5.90785751342791e+00, 5.90784551643997e+00, 5.90781926024360e+00, 5.90776897234124e+00, 5.90768150828401e+00, 5.90754010827732e+00, 5.90732428290419e+00, 5.90700983440517e+00, 5.90656901591722e+00, 5.90597082700500e+00, 5.90518143985622e+00, 5.90416474679159e+00, 5.90288301637472e+00, 5.90129764248919e+00, 5.89936996834113e+00, 5.89706216547032e+00, 5.89433814650627e+00, 5.89116448955165e+00, 5.88751135166183e+00, 5.88335334885674e+00, 5.87867038040272e+00, 5.87344837571096e+00, 5.86767994312420e+00, 5.86136490114787e+00, 5.85451067439873e+00, 5.84713253878523e+00, 5.83925370328410e+00, 5.83090521919548e+00, 5.82212571195102e+00, 5.81296093535209e+00, 5.80346315338731e+00, 5.79369036030560e+00, 5.78370535512698e+00, 5.77357469195159e+00, 5.76336753196832e+00, 5.75315442668985e+00, 5.74300606443426e+00, 5.73299201329970e+00, 5.72317949379432e+00, 5.71363221293858e+00, 5.70440928917775e+00, 5.69556429401809e+00, 5.68714443215564e+00, 5.67918987723873e+00, 5.67173327552281e+00, 5.66479942474138e+00, 5.65840513069538e+00, 5.65255923948194e+00, 5.64726283902946e+00, 5.64250961973483e+00, 5.63828638054256e+00, 5.63457366378195e+00, 5.63134649949986e+00, 5.62857523790639e+00, 5.62622644690682e+00, 5.62426385055150e+00, 5.62264928362276e+00, 5.62134363752571e+00, 5.62030777318039e+00, 5.61950337774131e+00, 5.61889374369379e+00, 5.61844445117416e+00, 5.61812393718868e+00, 5.61790393869933e+00, 5.61775980021841e+00, 5.61767064050476e+00, 5.61761937706949e+00, 5.61759261135422e+00, 5.61758038151656e+00, 5.61757579362378e+00, 5.61757454560019e+00, 5.61757436139432e+00, 5.61757434513245e+00, 5.61757402648900e+00, 5.61757186763002e+00, 5.61756393139522e+00, 5.61754277599334e+00, 5.61749647632203e+00, 5.61740780081295e+00, 5.61725357427011e+00, 5.61700425192036e+00, 5.61662372392622e+00, 5.61606936308056e+00, 5.61529232148148e+00, 5.61423807486141e+00, 5.61284720610969e+00, 5.61105641258001e+00, 5.60879971520123e+00, 5.60600984138425e+00, 5.60261974838240e+00, 5.59856424922729e+00, 5.59378169969693e+00, 5.58821570201938e+00, 5.58181677919194e+00, 5.57454397292542e+00, 5.56636631834513e+00, 5.55726414977336e+00, 5.54723019430268e+00, 5.53627041359729e+00, 5.52440455958643e+00, 5.51166641655294e+00, 5.49810371059402e+00, 5.48377767741483e+00, 5.46876229061368e+00, 5.45314316454380e+00, 5.43701615785228e+00, 5.42048571515714e+00, 5.40366299426648e+00, 5.38666383417743e+00, 5.36960662428570e+00, 5.35261013748037e+00, 5.33579138904439e+00, 5.31926357973587e+00, 5.30313417551528e+00, 5.28750316869169e+00, 5.27246155644914e+00, 5.25809006344585e+00, 5.24445812604655e+00, 5.23162314723092e+00, 5.21963002364406e+00, 5.20851093979654e+00, 5.19828541911325e+00, 5.18896061729407e+00, 5.18053184012400e+00, 5.17298326525477e+00, 5.16628884536558e+00, 5.16041336831145e+00, 5.15531364824548e+00, 5.15093982017506e+00, 5.14723670896877e+00, 5.14414524251759e+00, 5.14160387767329e+00, 5.13955000687730e+00, 5.13792131321448e+00, 5.13665704213546e+00, 5.13569915942394e+00, 5.13499336724065e+00, 5.13448995329432e+00, 5.13414445235616e+00, 5.13391810436352e+00, 5.13377809910726e+00, 5.13369760377022e+00, 5.13365557614631e+00, 5.13363637296598e+00, 5.13362916912344e+00, 5.13362720949731e+00, 5.13362692026037e+00, 5.13362689705999e+00, 5.13362646889522e+00, 5.13362356801349e+00, 5.13361290402911e+00, 5.13358447748350e+00, 5.13352226527020e+00, 5.13340311602280e+00, 5.13319589753254e+00, 5.13286093185654e+00, 5.13234974645182e+00, 5.13160516151275e+00, 5.13056172478000e+00, 5.12914649551838e+00, 5.12728016924766e+00, 5.12487852431676e+00, 5.12185416075501e+00, 5.11811849129443e+00, 5.11358393436923e+00, 5.10816624965101e+00, 5.10178694868752e+00, 5.09437570691631e+00, 5.08587269914981e+00, 5.07623077898197e+00, 5.06541742379954e+00, 5.05341637146978e+00, 5.04022888248021e+00, 5.02587457232760e+00, 5.01039177309470e+00, 4.99383739997620e+00, 4.97628631731974e+00, 4.95783021858538e+00, 4.93857605434469e+00, 4.91864406077905e+00, 4.89816545684939e+00, 4.87727989030785e+00, 4.85613272020966e+00, 4.83487222615889e+00, 4.81364683222773e+00, 4.79260242680358e+00, 4.77187984938057e+00, 4.75161260260738e+00, 4.73192483391857e+00, 4.71292961696097e+00, 4.69472754976863e+00, 4.67740567498531e+00, 4.66103671784503e+00, 4.64567863027738e+00, 4.63137442433894e+00, 4.61815227492336e+00, 4.60602586998850e+00, 4.59499498591293e+00, 4.58504626561201e+00, 4.57615417729990e+00, 4.56828213195016e+00, 4.56138373734585e+00, 4.55540416598413e+00, 4.55028161297358e+00, 4.54594881849164e+00, 4.54233462749297e+00, 4.53936555738176e+00, 4.53696734253214e+00, 4.53506642313691e+00, 4.53359134515961e+00, 4.53247403840835e+00, 4.53165094114003e+00, 4.53106394226739e+00, 4.53066111621653e+00, 4.53039723071184e+00, 4.53023401409215e+00, 4.53014017594624e+00, 4.53009118258564e+00, 4.53006879679676e+00, 4.53006039906310e+00, 4.53005811467107e+00, 4.53005777749942e+00, 4.53005775149350e+00, 4.53005728451055e+00, 4.53005412063081e+00, 4.53004248985058e+00, 4.53001148627922e+00, 4.52994363480067e+00, 4.52981368757298e+00, 4.52958769721023e+00, 4.52922240705317e+00, 4.52866499153023e+00, 4.52785317121748e+00, 4.52671571777267e+00, 4.52517335339322e+00, 4.52314003782074e+00, 4.52052462326721e+00, 4.51723284417680e+00, 4.51316959483138e+00, 4.50824143399883e+00, 4.50235924281819e+00, 4.49544095075795e+00, 4.48741423569099e+00, 4.47821909882247e+00, 4.46781021420365e+00, 4.45615895648772e+00, 4.44325501975097e+00, 4.42910755454921e+00, 4.41374576939960e+00, 4.39721896561724e+00, 4.37959599953153e+00, 4.36096419188874e+00, 4.34142772888819e+00, 4.32110562100665e+00, 4.30012930298105e+00, 4.27863996990574e+00, 4.25678574977074e+00, 4.23471881195246e+00, 4.21259250478273e+00, 4.19055860446321e+00, 4.16876474366896e+00, 4.14735207271809e+00, 4.12645319061580e+00, 4.10619036881361e+00, 4.08667407802271e+00, 4.06800181837033e+00, 4.05025724572160e+00, 4.03350958194441e+00, 4.01781329390311e+00, 4.00320802455014e+00, 3.98971875912854e+00, 3.97735620972400e+00, 3.96611740180218e+00, 3.95598644661831e+00, 3.94693548326972e+00, 3.93892577354779e+00, 3.93190893158264e+00, 3.92582826858377e+00, 3.92062023084066e+00, 3.91621590669354e+00, 3.91254257559063e+00, 3.90952526982731e+00, 3.90708831735264e+00, 3.90515683237895e+00, 3.90365811968708e+00, 3.90252295870343e+00, 3.90168673481294e+00, 3.90109038807411e+00, 3.90068115354907e+00, 3.90041307278980e+00, 3.90024726247472e+00, 3.90015193352370e+00, 3.90010216191895e+00, 3.90007942056454e+00, 3.90007088944931e+00, 3.90006856877458e+00, 3.90006822624768e+00, 3.90006820031258e+00, 3.90006774087477e+00, 3.90006462811454e+00, 3.90005318525676e+00, 3.90002268263096e+00, 3.89995592752696e+00, 3.89982808024589e+00, 3.89960574251449e+00, 3.89924635813996e+00, 3.89869795853771e+00, 3.89789927752793e+00, 3.89678025049590e+00, 3.89526290256770e+00, 3.89326261883573e+00, 3.89068977693825e+00, 3.88745170866237e+00, 3.88345494309369e+00, 3.87860766977849e+00, 3.87282234719778e+00, 3.86601837054069e+00, 3.85812470435510e+00, 3.84908238116849e+00, 3.83884676747431e+00, 3.82738950413179e+00, 3.81470003936830e+00, 3.80078668883622e+00, 3.78567717766245e+00, 3.76941864276825e+00, 3.75207709821369e+00, 3.73373639007022e+00, 3.71449668855759e+00, 3.69447258241299e+00, 3.67379085268555e+00, 3.65258800995080e+00, 3.63100768049058e+00, 3.60919792396419e+00, 3.58730855853305e+00, 3.56548856047331e+00, 3.54388359516104e+00, 3.52263372590252e+00, 3.50187133709921e+00, 3.48171929909494e+00, 3.46228939392210e+00, 3.44368101404988e+00, 3.42598014006119e+00, 3.40925859784556e+00, 3.39357359130939e+00, 3.37896750271551e+00, 3.36546794952523e+00, 3.35308808397631e+00, 3.34182711950873e+00, 3.33167106643349e+00, 3.32259365778279e+00, 3.31455744492730e+00, 3.30751504115113e+00, 3.30141048982715e+00, 3.29618073207971e+00, 3.29175714687131e+00, 3.28806713439120e+00, 3.28503571160930e+00, 3.28258708709626e+00, 3.28064618093804e+00, 3.27914005504231e+00, 3.27799921957199e+00, 3.27715878283752e+00, 3.27655941485001e+00, 3.27614809891779e+00, 3.27587865109703e+00, 3.27571199382684e+00, 3.27561617744834e+00, 3.27556615121392e+00, 3.27554329348319e+00, 3.27553471870569e+00, 3.27553238615325e+00, 3.27553204187321e+00, 3.27553201602800e+00, 3.27553156112291e+00, 3.27552847907234e+00, 3.27551714910295e+00, 3.27548694736881e+00, 3.27542085062000e+00, 3.27529426363603e+00, 3.27507411590821e+00, 3.27471826669796e+00, 3.27417525028989e+00, 3.27338438499292e+00, 3.27227626018594e+00, 3.27077360542060e+00, 3.26879253429172e+00, 3.26624414355488e+00, 3.26303643502676e+00, 3.25907651451708e+00, 3.25427300895002e+00, 3.24853863063681e+00, 3.24179280717982e+00, 3.23396428759817e+00, 3.22499363080411e+00, 3.21483548222794e+00, 3.20346054863830e+00, 3.19085719015814e+00, 3.17703256187890e+00, 3.16201325469955e+00, 3.14584540511223e+00, 3.12859426544958e+00, 3.11034324831617e+00, 3.09119248027301e+00, 3.07125691915031e+00, 3.05066410560922e+00, 3.02955163192583e+00, 3.00806441881315e+00, 2.98635189405432e+00, 2.96456516471777e+00, 2.94285426804016e+00, 2.92136557537017e+00, 2.90023940991593e+00, 2.87960792376322e+00, 2.85959326417077e+00, 2.84030604483794e+00, 2.82184412571316e+00, 2.80429169556598e+00, 2.78771864511568e+00, 2.77218021472181e+00, 2.75771689895004e+00, 2.74435459005745e+00, 2.73210494293024e+00, 2.72096594469425e+00, 2.71092267268825e+00, 2.70194822448369e+00, 2.69400480302760e+00, 2.68704493875773e+00, 2.68101282875307e+00, 2.67584577075204e+00, 2.67147566735406e+00, 2.66783057310697e+00, 2.66483625468840e+00, 2.66241773224563e+00, 2.66050076840702e+00, 2.65901327074883e+00, 2.65788657380416e+00, 2.65705656819658e+00, 2.65646464727168e+00, 2.65605844570550e+00, 2.65579234992253e+00, 2.65562776659989e+00, 2.65553314282728e+00, 2.65548373932527e+00, 2.65546116614543e+00, 2.65545269811593e+00, 2.65545039460183e+00, 2.65545005460780e+00, 2.65545002916195e+00, 2.65544958232008e+00, 2.65544655489907e+00, 2.65543542575546e+00, 2.65540575936241e+00, 2.65534083426202e+00, 2.65521649138358e+00, 2.65500024698281e+00, 2.65465070875524e+00, 2.65411732654074e+00, 2.65334050094841e+00, 2.65225206413925e+00, 2.65077613685800e+00, 2.64883035459198e+00, 2.64632744352854e+00, 2.64317711399010e+00, 2.63928822561797e+00, 2.63457116531226e+00, 2.62894036655362e+00, 2.62231688811751e+00, 2.61463096230546e+00, 2.60582441860024e+00, 2.59585288890231e+00, 2.58468770575076e+00, 2.57231741531097e+00, 2.55874884208572e+00, 2.54400766143701e+00, 2.52813845779040e+00, 2.51120426919900e+00, 2.49328564099560e+00, 2.47447923088460e+00, 2.45489602368226e+00, 2.43465922518285e+00, 2.41390191109076e+00, 2.39276450895873e+00, 2.37139218937045e+00, 2.34993223816888e+00, 2.32853147531018e+00, 2.30733377869457e+00, 2.28647776360929e+00, 2.26609466049899e+00, 2.24630642577230e+00, 2.22722411232071e+00, 2.20894651844510e+00, 2.19155912610927e+00, 2.17513333208481e+00, 2.15972596886992e+00, 2.14537910648733e+00, 2.13212012155454e+00, 2.11996201642865e+00, 2.10890396869758e+00, 2.09893208965281e+00, 2.09002036939357e+00, 2.08213178560075e+00, 2.07521955250753e+00, 2.06922848595871e+00, 2.06409645954003e+00, 2.05975592551765e+00, 2.05613547280096e+00, 2.05316139246675e+00, 2.05075921977385e+00, 2.04885522031957e+00, 2.04737778732639e+00, 2.04625871726800e+00, 2.04543433237277e+00, 2.04484642112952e+00, 2.04444297182246e+00, 2.04417867929375e+00, 2.04401521142511e+00, 2.04392122900199e+00, 2.04387216036386e+00, 2.04384974019106e+00, 2.04384132956059e+00, 2.04383904166048e+00, 2.04383870397105e+00, 2.04383867868823e+00, 2.04383823458270e+00, 2.04383522570083e+00, 2.04382416470951e+00, 2.04379467998890e+00, 2.04373015249167e+00, 2.04360657113812e+00, 2.04339165124760e+00, 2.04304425435744e+00, 2.04251414070999e+00, 2.04174207768509e+00, 2.04066031847809e+00, 2.03919345533392e+00, 2.03725964064927e+00, 2.03477215733482e+00, 2.03164130719453e+00, 2.02777657308518e+00, 2.02308899778025e+00, 2.01749371043875e+00, 2.01091252114533e+00, 2.00327649596662e+00, 1.99452842015575e+00, 1.98462505620716e+00, 1.97353910688167e+00, 1.96126080126838e+00, 1.94779903427844e+00, 1.93318200619321e+00, 1.91745732823304e+00, 1.90069158154724e+00, 1.88296933937662e+00, 1.86439168416434e+00, 1.84507427186969e+00, 1.82514501354872e+00, 1.80474145843498e+00, 1.78400797253103e+00, 1.76309281161203e+00, 1.74214518736451e+00, 1.72131242029063e+00, 1.70073726349885e+00, 1.68055546839278e+00, 1.66089364762696e+00, 1.64186747373735e+00, 1.62358023482395e+00, 1.60612175271461e+00, 1.58956765512596e+00, 1.57397898213109e+00, 1.55940209910125e+00, 1.54586888325292e+00, 1.53339714876669e+00, 1.52199127569939e+00, 1.51164300999402e+00, 1.50233240514310e+00, 1.49402887982877e+00, 1.48669236955591e+00, 1.48027455343092e+00, 1.47472013947279e+00, 1.46996819297938e+00, 1.46595349247091e+00, 1.46260789669546e+00, 1.45986170433814e+00, 1.45764498575067e+00, 1.45588886360154e+00, 1.45452671725330e+00, 1.45349528430101e+00, 1.45273563239896e+00, 1.45219397551623e+00, 1.45182231123518e+00, 1.45157885965079e+00, 1.45142828972135e+00, 1.45134172532237e+00, 1.45129653042884e+00, 1.45127588040003e+00, 1.45126813384373e+00, 1.45126602659092e+00, 1.45126571556477e+00, 1.45126569221645e+00, 1.45126528126645e+00, 1.45126249701944e+00, 1.45125226182859e+00, 1.45122497853431e+00, 1.45116526954313e+00, 1.45105091903473e+00, 1.45085206051788e+00, 1.45053064686406e+00, 1.45004023404971e+00, 1.44932610300910e+00, 1.44832573606352e+00, 1.44696965517129e+00, 1.44518261860784e+00, 1.44288516062644e+00, 1.43999544531978e+00, 1.43643139166637e+00, 1.43211301219231e+00, 1.42696489361814e+00, 1.42091873527936e+00, 1.41391585109973e+00, 1.40590953456224e+00, 1.39686718445515e+00, 1.38677209293090e+00, 1.37562480701638e+00, 1.36344399012423e+00, 1.35026673080995e+00, 1.33614827096516e+00, 1.32116115334445e+00, 1.30539381694715e+00, 1.28894869627589e+00, 1.27193990482381e+00, 1.25449060245764e+00, 1.23673015922467e+00, 1.21879123364087e+00, 1.20080688150701e+00, 1.18290780222296e+00, 1.16521981452695e+00, 1.14786163415218e+00, 1.13094300393081e+00, 1.11456320431194e+00, 1.09880995087880e+00, 1.08375866671490e+00, 1.06947210241778e+00, 1.05600026575091e+00, 1.04338061646256e+00, 1.03163847939393e+00, 1.02078763006214e+00, 1.01083101066572e+00, 1.00176154007046e+00, 9.93562987948245e-01, 9.86210890097652e-01, 9.79673488425894e-01, 9.73912684618401e-01, 9.68885000818323e-01, 9.64542543485035e-01, 9.60833967943254e-01, 9.57705441038964e-01, 9.55101597968513e-01, 9.52966487013123e-01, 9.51244492937906e-01, 9.49881226596285e-01, 9.48824365225093e-01, 9.48024425424391e-01, 9.47435449238180e-01, 9.47015583364766e-01, 9.46727532495788e-01, 9.46538870161369e-01, 9.46422194162579e-01, 9.46355118502826e-01, 9.46320099385281e-01, 9.46304098957564e-01, 9.46298096660302e-01, 9.46296463891390e-01, 9.46296222898164e-01, 9.46296204730085e-01, 9.46295883930180e-01, 9.46293710464659e-01, 9.46285720590364e-01, 9.46264422633202e-01, 9.46217813154759e-01, 9.46128552649647e-01, 9.45973334837523e-01, 9.45722480268123e-01, 9.45339781026352e-01, 9.44782618572864e-01, 9.44002370892088e-01, 9.42945117829053e-01, 9.41552644524985e-01, 9.39763732089864e-01, 9.37515712163313e-01, 9.34746248170189e-01, 9.31395291528025e-01, 9.27407146765064e-01, 9.22732566658691e-01, 9.17330788460741e-01, 9.11171416401644e-01, 9.04236055166379e-01, 8.96519604775388e-01, 8.88031139638408e-01, 8.78794313195774e-01, 8.68847253558115e-01, 8.58241943271180e-01, 8.47043105629786e-01, 8.35326648386111e-01, 8.23177740761309e-01, 8.10688619171596e-01, 7.97956229356648e-01, 7.85079816756920e-01, 7.72158572983265e-01, 7.59289434869084e-01, 7.46565115399255e-01, 7.34072424769215e-01, 7.21890917127733e-01, 7.10091876322834e-01, 6.98737633995059e-01, 6.87881196971519e-01, 6.77566148870531e-01, 6.67826783354127e-01, 6.58688423306814e-01, 6.50167880789106e-01, 6.42274016088060e-01, 6.35008359686158e-01, 6.28365767628336e-01, 6.22335087807066e-01, 6.16899821498149e-01, 6.12038770598281e-01, 6.07726666111740e-01, 6.03934777355734e-01, 6.00631503997028e-01, 5.97782954445603e-01, 5.95353514375446e-01, 5.93306408364009e-01, 5.91604256013188e-01, 5.90209621658818e-01, 5.89085554129969e-01, 5.88196110247632e-01, 5.87506853137724e-01, 5.86985314225732e-01, 5.86601406252251e-01, 5.86327773945004e-01, 5.86140069280940e-01, 5.86017139587552e-01, 5.85941119009005e-01, 5.85897417001696e-01, 5.85874601275709e-01, 5.85864176732569e-01, 5.85860266151128e-01, 5.85859202380580e-01, 5.85859045370144e-01, 5.85859033469104e-01, 5.85858822472015e-01, 5.85857392936483e-01, 5.85852137832481e-01, 5.85838129800604e-01, 5.85807474309144e-01, 5.85748768224541e-01, 5.85646686765799e-01, 5.85481719924854e-01, 5.85230077182949e-01, 5.84863776512399e-01, 5.84350929918332e-01, 5.83656232882771e-01, 5.82741658718160e-01, 5.81567350920520e-01, 5.80092697152431e-01, 5.78277557781710e-01, 5.76083610523967e-01, 5.73475761545261e-01, 5.70423563456937e-01, 5.66902573214553e-01, 5.62895579226039e-01, 5.58393628005906e-01, 5.53396787142959e-01, 5.47914593274792e-01, 5.41966150640069e-01, 5.35579866463032e-01, 5.28792832062828e-01, 5.21649881186452e-01, 5.14202377272286e-01, 5.06506797253946e-01, 4.98623189611626e-01, 4.90613587740365e-01, 4.82540456522460e-01, 4.74465240907843e-01, 4.66447071739293e-01, 4.58541667837199e-01, 4.50800456298321e-01, 4.43269916790328e-01, 4.35991141709750e-01, 4.28999593337251e-01, 4.22325031741546e-01, 4.15991583535240e-01, 4.10017920659421e-01, 4.04417520008070e-01, 3.99198978208614e-01, 3.94366359839514e-01, 3.89919562614014e-01, 3.85854687110113e-01, 3.82164403356556e-01, 3.78838309861301e-01, 3.75863283606608e-01, 3.73223821616886e-01, 3.70902376052147e-01, 3.68879685258991e-01, 3.67135103565134e-01, 3.65646931652616e-01, 3.64392748859526e-01, 3.63349747102288e-01, 3.62495064884579e-01, 3.61806118025602e-01, 3.61260922344352e-01, 3.60838401895845e-01, 3.60518675323365e-01, 3.60283312160311e-01, 3.60115550398865e-01, 3.60000467516916e-01, 3.59925097395485e-01, 3.59878487500658e-01, 3.59851692679290e-01, 3.59837703734173e-01, 3.59831312136987e-01, 3.59828914444642e-01, 3.59828262216690e-01, 3.59828165955040e-01, 3.59828158610341e-01, 3.59828027803352e-01, 3.59827141622373e-01, 3.59823883900749e-01, 3.59815200099495e-01, 3.59796196254946e-01, 3.59759803252069e-01, 3.59696520682596e-01, 3.59594253052146e-01, 3.59438250181226e-01, 3.59211161267888e-01, 3.58893210334303e-01, 3.58462497002538e-01, 3.57895423375590e-01, 3.57167242073952e-01, 3.56252714985201e-01, 3.55126865614195e-01, 3.53765800858394e-01, 3.52147570820722e-01, 3.50253030143278e-01, 3.48066659311796e-01, 3.45577302954414e-01, 3.42778783935159e-01, 3.39670355942829e-01, 3.36256966443897e-01, 3.32549311636217e-01, 3.28563679560644e-01, 3.24321589275077e-01, 3.19849248413028e-01, 3.15176861296379e-01, 3.10337827771525e-01, 3.05367876886909e-01, 3.00304179944745e-01, 2.95184481946150e-01, 2.90046288283589e-01, 2.84926130317956e-01, 2.79772439798170e-01, 2.74799607146107e-01, 2.69941880364403e-01, 2.65226648147045e-01, 2.60678217288145e-01, 2.56317677236349e-01, 2.52162839561893e-01, 2.48228240441549e-01, 2.44525194886863e-01, 2.41061892563258e-01, 2.37843526436970e-01, 2.34872446969434e-01, 2.32148336033325e-01, 2.29668396073424e-01, 2.27427551228528e-01, 2.25418658135033e-01, 2.23632724924897e-01, 2.22059137491915e-01, 2.20685892417296e-01, 2.19499836012396e-01, 2.18486908757310e-01, 2.17632394006931e-01, 2.16921169234804e-01, 2.16337957340706e-01, 2.15867574726407e-01, 2.15495172024556e-01, 2.15206462633885e-01, 2.14987933655784e-01, 2.14827033520992e-01, 2.14712330602491e-01, 2.14633637470236e-01, 2.14582096164905e-01, 2.14550220929683e-01, 2.14531896189288e-01, 2.14522329126216e-01, 2.14517957879564e-01, 2.14516318077831e-01, 2.14515872012359e-01, 2.14515806174043e-01, 2.14515801116484e-01, 2.14515710565061e-01, 2.14515097066139e-01, 2.14512841783652e-01, 2.14506830062731e-01, 2.14493673731878e-01, 2.14468478448570e-01, 2.14424665717635e-01, 2.14353858274773e-01, 2.14245836421685e-01, 2.14088571555071e-01, 2.13868341457978e-01, 2.13569929821731e-01, 2.13176909860043e-01, 2.12672008725590e-01, 2.12037545772364e-01, 2.11255933642808e-01, 2.10310226916286e-01, 2.09184698954584e-01, 2.07865424019845e-01, 2.06340839158879e-01, 2.04602259164713e-01, 2.02644318472875e-01, 2.00465316302755e-01, 1.98067445683336e-01, 1.95456892949139e-01, 1.92643801387177e-01, 1.89642100334473e-01, 1.86469208476079e-01, 1.83145626714565e-01, 1.79694441236549e-01, 1.76140760941823e-01, 1.72511115094263e-01, 1.68832836972812e-01, 1.65133457677721e-01, 1.61440131414846e-01, 1.57779109911590e-01, 1.54175279475763e-01, 1.50651769916018e-01, 1.47229640367461e-01, 1.43927643219319e-01, 1.40762063980646e-01, 1.37746632155599e-01, 1.34892496098130e-01, 1.32208253402731e-01, 1.29700027648635e-01, 1.27371582198328e-01, 1.25224462173423e-01, 1.23258156582824e-01, 1.21470273734367e-01, 1.19856724390389e-01, 1.18411908502518e-01, 1.17128902665930e-01, 1.15999646572068e-01, 1.15015127638110e-01, 1.14165563604141e-01, 1.13440583194169e-01, 1.12829404940338e-01, 1.12321014000681e-01, 1.11904336310748e-01, 1.11568408767341e-01, 1.11302543429390e-01, 1.11096483024207e-01, 1.10940544452841e-01, 1.10825746573945e-01, 1.10743918373168e-01, 1.10687783734723e-01, 1.10651019437232e-01, 1.10628283683804e-01, 1.10615213406415e-01, 1.10608389695498e-01, 1.10605271920281e-01, 1.10604102340766e-01, 1.10603784187310e-01, 1.10603737228546e-01, 1.10603733593744e-01, 1.10603668157249e-01, 1.10603224815717e-01, 1.10601595050678e-01, 1.10597250736371e-01, 1.10587743518413e-01, 1.10569536854379e-01, 1.10537877803178e-01, 1.10486715047183e-01, 1.10408668912931e-01, 1.10295058260179e-01, 1.10135987954059e-01, 1.09920499136306e-01, 1.09636782597495e-01, 1.09272453185489e-01, 1.08814880380617e-01, 1.08251567009444e-01, 1.07570564717347e-01, 1.06760911513454e-01, 1.05813073740265e-01, 1.04719372541354e-01, 1.03474373637702e-01, 1.02075219261428e-01, 1.00521882623349e-01, 9.88173283595968e-02, 9.69675669032242e-02, 9.49815963846188e-02, 9.28712320612908e-02, 9.06508298926036e-02, 8.83369171445016e-02, 8.59477482914297e-02, 8.35028085204739e-02, 8.10222895153263e-02, 7.85265627526792e-02, 7.60356743088693e-02, 7.35688823468097e-02, 7.11442543681287e-02, 6.87783364036605e-02, 6.64859010497638e-02, 6.42797761002463e-02, 6.21707508768736e-02, 6.01675535325077e-02, 5.82768897802257e-02, 5.65035317589748e-02, 5.48504450425424e-02, 5.33189420095210e-02, 5.19088507368854e-02, 5.06186900509871e-02, 4.94458431574207e-02, 4.83867241835272e-02, 4.74369338427243e-02, 4.65914021439904e-02, 4.58445175355600e-02, 4.51902430367781e-02, 4.46222207523590e-02, 4.41338666783428e-02, 4.37184579157819e-02, 4.33692143371536e-02, 4.30793764431670e-02, 4.28422806541908e-02, 4.26514326583165e-02, 4.25005787500652e-02, 4.23837744048878e-02, 4.22954487096900e-02, 4.22304627673684e-02, 4.21841598628312e-02, 4.21524050530566e-02, 4.21316119396966e-02, 4.21187546946866e-02, 4.21113639115509e-02, 4.21075055050197e-02, 4.21057426239120e-02, 4.21050813155217e-02, 4.21049014244703e-02, 4.21048748729591e-02, 4.21048728011026e-02, 4.21048352866746e-02, 4.21045811214027e-02, 4.21036467903049e-02, 4.21011562629099e-02, 4.20957060901852e-02, 4.20852694516602e-02, 4.20671234739469e-02, 4.20378039434786e-02, 4.19930912806654e-02, 4.19280312628354e-02, 4.18369933943250e-02, 4.17137689336640e-02, 4.15517093216256e-02, 4.13439040610273e-02, 4.10833949809515e-02, 4.07634213407215e-02, 4.03776875350381e-02, 3.99206424697969e-02, 3.93877572707515e-02, 3.87757861851190e-02, 3.80829946587324e-02, 3.73093388883099e-02, 3.64565828302457e-02, 3.55283417201200e-02, 3.45300454760498e-02, 3.34688206038010e-02, 3.23532949224461e-02, 3.11933350197107e-02, 2.99997312373517e-02, 2.87838486508445e-02, 2.75572645543471e-02, 2.63314131964343e-02, 2.51172569611660e-02, 2.39250000933362e-02, 2.27638568393695e-02, 2.16418810355155e-02, 2.05658592676086e-02, 1.95412652435349e-02, 1.85722693379917e-02, 1.76617946183049e-02, 1.68116091165489e-02, 1.60224436219396e-02, 1.52941246748419e-02, 1.46257135354784e-02, 1.40156434386469e-02, 1.34618492000804e-02, 1.29618850142847e-02, 1.25130279301473e-02, 1.21123659152778e-02, 1.17568705764999e-02, 1.14434554821230e-02, 1.11690216456628e-02, 1.09304921086659e-02, 1.07248377366611e-02, 1.05490963529853e-02, 1.04003872140650e-02, 1.02759226067892e-02, 1.01730180498523e-02, 1.00891022292506e-02, 1.00217274147871e-02, 9.96858071098394e-03, 9.92749611410293e-03, 9.89646700027347e-03, 9.87365838003029e-03, 9.85741804243201e-03, 9.84628559231915e-03, 9.83899836584801e-03, 9.83449329182484e-03, 9.83190394001454e-03, 9.83055224439067e-03, 9.82993468445148e-03, 9.82970302302366e-03, 9.82964000612441e-03, 9.82963070498322e-03, 9.82962997302650e-03, 9.82961664057669e-03, 9.82952631160661e-03, 9.82919425785678e-03, 9.82830916365351e-03, 9.82637235544281e-03, 9.82266389775344e-03, 9.81621724200288e-03, 9.80580412934895e-03, 9.78993144698759e-03, 9.76685167525027e-03, 9.73458838569563e-03, 9.69097789863582e-03, 9.63372762940473e-03, 9.56049081189558e-03, 9.46895618321886e-03, 9.35694988680602e-03, 9.22254539693412e-03, 9.06417583202224e-03, 8.88074179694431e-03, 8.67170708772580e-03, 8.43717440761438e-03, 8.17793383590715e-03, 7.89547822767478e-03, 7.59198195511335e-03, 7.27024225120232e-03, 6.93358558746127e-03, 6.58574463214023e-03, 6.23071399311037e-03, 5.87259479778179e-03, 5.51543895811868e-03, 5.16310362398631e-03, 4.81912492046010e-03, 4.48661781846660e-03, 4.16820622945213e-03, 3.86598451587705e-03, 3.58150892954523e-03, 3.31581532125735e-03, 3.06945799114011e-03, 2.84256382274428e-03, 2.63489579842702e-03, 2.44592048032199e-03, 2.27487488122247e-03, 2.12082917672696e-03, 1.98274278893251e-03, 1.85951239734816e-03, 1.75001132370767e-03, 1.65312043852150e-03, 1.56775122701003e-03, 1.49286194480958e-03, 1.42746792999304e-03, 1.37064716557234e-03, 1.32154214594848e-03, 1.27935901756615e-03, 1.24336485447252e-03, 1.21288380439211e-03, 1.18729271089002e-03, 1.16601669008620e-03, 1.14852502204688e-03, 1.13432760853005e-03, 1.12297214986778e-03, 1.11404210291077e-03, 1.10715539914086e-03, 1.10196382844840e-03, 1.09815293234039e-03, 1.09544220376723e-03, 1.09358536227422e-03, 1.09237046491284e-03, 1.09161962596183e-03, 1.09118815040787e-03, 1.09096293414801e-03, 1.09086004304881e-03, 1.09082144699713e-03, 1.09081094811891e-03, 1.09080939851403e-03, 1.09080927548799e-03, 1.09080702087194e-03, 1.09079174561710e-03, 1.09073559366824e-03, 1.09058592474121e-03, 1.09025843655619e-03, 1.08963148202257e-03, 1.08854190511476e-03, 1.08678274340246e-03, 1.08410316689955e-03, 1.08021102950939e-03, 1.07477838403547e-03, 1.06745023679000e-03, 1.05785667701497e-03, 1.04562829920507e-03, 1.03041454330643e-03, 1.01190422342778e-03, 9.89847132626842e-04, 9.64075249430429e-04, 9.34521793868548e-04, 9.01236254566840e-04, 8.64393592558521e-04, 8.24296157841076e-04, 7.81367429179447e-04, 7.36137457513664e-04, 6.89220762746245e-04, 6.41288273105722e-04, 5.93035568893896e-04, 5.45150088025359e-04, 4.98280014497254e-04, 4.53007310758480e-04, 4.09826826166082e-04, 3.69132691190543e-04, 3.31212372584488e-04, 2.96247919330028e-04, 2.64323208926776e-04, 2.35435557261354e-04, 2.09509969094412e-04, 1.86414527629353e-04, 1.65975767077775e-04, 1.47993148143399e-04, 1.32251906122427e-04, 1.18533674881682e-04, 1.06624553559698e-04, 9.63206845145947e-05, 8.74317834362870e-05, 7.97832093306639e-05, 7.32170401152032e-05, 6.75923780510956e-05, 6.28214061864085e-05, 5.87263480831268e-05, 5.52335669112836e-05, 5.22482269813551e-05, 4.97828312502335e-05, 4.77035105896917e-05, 4.59588231418451e-05, 4.45031338545121e-05, 4.32965216710102e-05, 4.22942221778300e-05, 4.15336270389588e-05, 4.09231492302098e-05, 4.04432355667339e-05, 4.00731557306244e-05, 3.97940273551001e-05, 3.95889582005537e-05, 3.94429539980479e-05, 3.93428567448950e-05, 3.92773209476616e-05, 3.92367987558373e-05, 3.92135052358356e-05, 3.92013446299799e-05, 3.91957885002014e-05, 3.91937042287793e-05, 3.91931372591969e-05, 3.91930535759144e-05, 3.91930469705315e-05, 3.91929264482727e-05, 3.91921098783499e-05, 3.91891080502305e-05, 3.91811060476144e-05, 3.91635927412537e-05, 3.91300488429280e-05, 3.90717063574186e-05, 3.89773950120187e-05, 3.88334997731832e-05, 3.86240675549487e-05, 3.83311144664944e-05, 3.79351876462763e-05, 3.74162222716024e-05, 3.67547018411685e-05, 3.59330629127186e-05, 3.49370886052469e-05, 3.37554080687141e-05, 3.24347536100333e-05, 3.09471823243257e-05, 2.92865425501887e-05, 2.74725997861935e-05, 2.55308066350324e-05, 2.35364601417487e-05, 2.15014814806125e-05, 1.94428448301309e-05, 1.73991813179342e-05, 1.54555190681194e-05, 1.36026806529799e-05, 1.18544947138588e-05, 1.02583608638795e-05, 8.81894397775085e-06, 7.51542241239633e-06, 6.36307484480625e-06, 5.38378686965046e-06, 4.52916137927417e-06, 3.78418316770990e-06, 3.19222603195612e-06, 2.67441325085827e-06, 2.23090761952704e-06, 1.88100891662657e-06, 1.57899378461224e-06, 1.34870426204969e-06, 1.17227910312323e-06, 1.01407316943636e-06, 8.66225851048341e-07, 7.75579708012231e-07, 6.80115441298312e-07, 5.80280992146243e-07, 5.16220503676834e-07, 4.93673063317601e-07, 4.70066245483736e-07, 4.40541798596428e-07, 4.01735444910400e-07, 3.49738884963813e-07, 3.75223705439964e-07, 3.85112000067522e-07, 3.88168481357809e-07, 3.86364684246132e-07, 3.81441986441189e-07, 3.75000082541987e-07, 3.68292519148383e-07, 3.62137377862837e-07, 3.56956365799022e-07, 3.52877494336212e-07, 3.49845489466816e-07, 3.47710948922497e-07, 3.46290820907313e-07, 3.45404366441926e-07, 3.44892048254053e-07, 3.44623812154055e-07, 3.44501082771842e-07, 3.44455015621747e-07, 3.44442481647889e-07, 3.44440631572240e-07, 3.44440526767187e-07, 3.44439141208852e-07, 3.44429746057529e-07, 3.44395114204588e-07, 3.44302071297875e-07, 3.44094776671966e-07, 3.43683856485200e-07, 3.42926501417651e-07, 3.41591151808984e-07, 3.39301307050732e-07, 3.35457948530404e-07, 3.29148014621694e-07, 3.19046153940405e-07, 3.03257833955684e-07, 2.78571439693565e-07, 2.53578730458059e-07, 2.75596244492249e-07, 2.85571894131920e-07, 2.82650024274744e-07, 2.61421591041255e-07, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -3.55271367880050e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -3.55271367880050e-15, -3.55271367880050e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -3.55271367880050e-15, -2.66453525910038e-15, -3.55271367880050e-15, -1.77635683940025e-15, -2.66453525910038e-15, -2.66453525910038e-15, -1.77635683940025e-15, -2.66453525910038e-15, -3.55271367880050e-15, -3.55271367880050e-15, -2.66453525910038e-15, -2.66453525910038e-15, -1.77635683940025e-15, -1.77635683940025e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -2.66453525910038e-15, -3.55271367880050e-15, -2.66453525910038e-15],
[ 7.00000000000000e+00, 6.99999999760877e+00, 6.99999992366718e+00, 6.99999942270128e+00, 6.99999758108681e+00, 6.99999267193552e+00, 6.99998192795548e+00, 6.99996135032922e+00, 6.99992556059699e+00, 6.99986770126620e+00, 6.99977938942980e+00, 6.99965072611062e+00, 6.99947036240280e+00, 6.99922562180859e+00, 6.99890267651971e+00, 6.99848677381790e+00, 6.99796250731597e+00, 6.99731412646873e+00, 6.99652587668565e+00, 6.99558236149911e+00, 6.99446891759844e+00, 6.99317199313808e+00, 6.99167951956749e+00, 6.98998126730351e+00, 6.98806917586240e+00, 6.98593764957372e+00, 6.98358381069837e+00, 6.98100770265317e+00, 6.97821243709096e+00, 6.97520427978546e+00, 6.97199267160950e+00, 6.96859018235817e+00, 6.96501239673333e+00, 6.96127773344777e+00, 6.95740720009075e+00, 6.95342408808191e+00, 6.94935361367848e+00, 6.94522251253849e+00, 6.94105859672408e+00, 6.93689028419996e+00, 6.93274611179076e+00, 6.92865424316745e+00, 6.92464198370378e+00, 6.92073531396387e+00, 6.91695845314646e+00, 6.91333346303344e+00, 6.90987990189456e+00, 6.90661453642482e+00, 6.90355111818282e+00, 6.90070022921197e+00, 6.89806919961982e+00, 6.89566209792548e+00, 6.89347979301998e+00, 6.89152008467804e+00, 6.88977789776606e+00, 6.88824553365838e+00, 6.88691297094607e+00, 6.88576820633496e+00, 6.88479762571169e+00, 6.88398639472893e+00, 6.88331885793679e+00, 6.88277893547007e+00, 6.88235050658780e+00, 6.88201776993810e+00, 6.88176557126959e+00, 6.88157969040127e+00, 6.88144708056427e+00, 6.88135605470188e+00, 6.88129641491727e+00, 6.88125952294646e+00, 6.88123831126231e+00, 6.88122723613661e+00, 6.88122217565694e+00, 6.88122027726990e+00, 6.88121976086010e+00, 6.88121968463900e+00, 6.88121967464040e+00, 6.88121944168497e+00, 6.88121786337525e+00, 6.88121206130618e+00, 6.88119659484691e+00, 6.88116274549189e+00, 6.88109791461168e+00, 6.88098515653036e+00, 6.88080286481880e+00, 6.88052462511233e+00, 6.88011924278871e+00, 6.87955094864784e+00, 6.87877978049767e+00, 6.87776213345306e+00, 6.87645146696665e+00, 6.87479915228219e+00, 6.87275544024785e+00, 6.87027052632416e+00, 6.86729568719753e+00, 6.86378446164977e+00, 6.85969384718328e+00, 6.85498548328587e+00, 6.84962679205986e+00, 6.84359204717841e+00, 6.83686334274615e+00, 6.82943143466303e+00, 6.82129642860258e+00, 6.81246829084830e+00, 6.80296716113177e+00, 6.79282345041975e+00, 6.78207771138995e+00, 6.77078027511730e+00, 6.75899065415403e+00, 6.74677671949934e+00, 6.73421366658004e+00, 6.72138279288321e+00, 6.70837011683924e+00, 6.69526487350481e+00, 6.68215792715878e+00, 6.66914014382249e+00, 6.65630076780417e+00, 6.64372584563854e+00, 6.63149673837446e+00, 6.61968875928893e+00, 6.60836996908394e+00, 6.59760015479729e+00, 6.58743001237455e+00, 6.57790054641779e+00, 6.56904269431026e+00, 6.56087717590867e+00, 6.55341456443036e+00, 6.54665556911348e+00, 6.54059151572387e+00, 6.53520500702140e+00, 6.53047074186594e+00, 6.52635646872389e+00, 6.52282404692555e+00, 6.51983058713220e+00, 6.51732964112715e+00, 6.51527241028952e+00, 6.51360894199087e+00, 6.51228928372072e+00, 6.51126456603489e+00, 6.51048798744641e+00, 6.50991567713485e+00, 6.50950741479678e+00, 6.50922719102526e+00, 6.50904359619022e+00, 6.50893003076391e+00, 6.50886473525044e+00, 6.50883064317272e+00, 6.50881506577894e+00, 6.50880922208925e+00, 6.50880763245722e+00, 6.50880739783062e+00, 6.50880737711755e+00, 6.50880697125519e+00, 6.50880422147470e+00, 6.50879411294180e+00, 6.50876716692850e+00, 6.50870819432972e+00, 6.50859524743770e+00, 6.50839880896770e+00, 6.50808125075082e+00, 6.50759658702347e+00, 6.50689053903138e+00, 6.50590091892447e+00, 6.50455833186188e+00, 6.50278718609959e+00, 6.50050699183215e+00, 6.49763392093916e+00, 6.49408259176756e+00, 6.48976803586005e+00, 6.48460779727899e+00, 6.47852410999889e+00, 6.47144609483875e+00, 6.46331191465311e+00, 6.45407082507639e+00, 6.44368505812839e+00, 6.43213147759467e+00, 6.41940294849644e+00, 6.40550936838744e+00, 6.39047831586462e+00, 6.37435528166906e+00, 6.35720346003669e+00, 6.33910309225747e+00, 6.32015037018022e+00, 6.30045592386416e+00, 6.28014293373687e+00, 6.25934492238232e+00, 6.23820329339550e+00, 6.21686469371156e+00, 6.19547828084212e+00, 6.17419297728363e+00, 6.15315479114056e+00, 6.13250427521963e+00, 6.11237418726228e+00, 6.09288740251869e+00, 6.07415511749827e+00, 6.05627537136307e+00, 6.03933189981372e+00, 6.02339332599562e+00, 6.00851268423993e+00, 5.99472726544871e+00, 5.98205876754917e+00, 5.97051373045744e+00, 5.96008423210659e+00, 5.95074881997520e+00, 5.94247365089004e+00, 5.93521381040686e+00, 5.92891478161398e+00, 5.92351403165914e+00, 5.91894268266792e+00, 5.91512723208331e+00, 5.91199128596465e+00, 5.90945726764073e+00, 5.90744806354636e+00, 5.90588856831030e+00, 5.90470709240642e+00, 5.90383659807753e+00, 5.90321573287319e+00, 5.90278963500506e+00, 5.90251049072635e+00, 5.90233783091175e+00, 5.90223856170523e+00, 5.90218673221456e+00, 5.90216305043126e+00, 5.90215416650218e+00, 5.90215174985103e+00, 5.90215139315813e+00, 5.90215136454695e+00, 5.90215083652542e+00, 5.90214725909928e+00, 5.90213410806341e+00, 5.90209905193107e+00, 5.90202233092494e+00, 5.90187539520343e+00, 5.90161985454979e+00, 5.90120678405145e+00, 5.90057642147746e+00, 5.89965828218906e+00, 5.89837170654409e+00, 5.89662684293863e+00, 5.89432605696895e+00, 5.89136574385887e+00, 5.88763850753912e+00, 5.88303565594641e+00, 5.87744994869839e+00, 5.87077852085563e+00, 5.86292589563981e+00, 5.85380699041167e+00, 5.84335001460178e+00, 5.83149915626673e+00, 5.81821695605448e+00, 5.80348627399676e+00, 5.78731176588076e+00, 5.76972080188775e+00, 5.75076378028592e+00, 5.73051381242391e+00, 5.70906578091162e+00, 5.68653479919974e+00, 5.66305412607189e+00, 5.63877261108808e+00, 5.61385176515698e+00, 5.58846256290086e+00, 5.56278208955646e+00, 5.53699014467582e+00, 5.51126590832332e+00, 5.48578476382005e+00, 5.46071535576538e+00, 5.43621694466351e+00, 5.41243710158281e+00, 5.38950976926603e+00, 5.36755370105629e+00, 5.34667127658133e+00, 5.32694768362748e+00, 5.30845044896842e+00, 5.29122929675884e+00, 5.27531631095650e+00, 5.26072637751520e+00, 5.24745788221713e+00, 5.23549364046471e+00, 5.22480203570820e+00, 5.21533834314008e+00, 5.20704621465662e+00, 5.19985929980275e+00, 5.19370297551473e+00, 5.18849615508456e+00, 5.18415314409264e+00, 5.18058550835371e+00, 5.17770391649488e+00, 5.17541991795237e+00, 5.17364761624756e+00, 5.17230519766164e+00, 5.17131627709322e+00, 5.17061102609371e+00, 5.17012705286731e+00, 5.16981001032430e+00, 5.16961391589795e+00, 5.16950117548903e+00, 5.16944231321471e+00, 5.16941541818536e+00, 5.16940532887145e+00, 5.16940258432779e+00, 5.16940217923869e+00, 5.16940214799432e+00, 5.16940158694567e+00, 5.16939778575673e+00, 5.16938381215694e+00, 5.16934656346197e+00, 5.16926504454678e+00, 5.16910892202033e+00, 5.16883741095730e+00, 5.16839854285017e+00, 5.16772885458848e+00, 5.16675352821573e+00, 5.16538699987665e+00, 5.16353404366591e+00, 5.16109132199484e+00, 5.15794937868755e+00, 5.15399503454988e+00, 5.14911412808558e+00, 5.14319452705394e+00, 5.13612932059179e+00, 5.12782008776058e+00, 5.11818012783127e+00, 5.10713753158333e+00, 5.09463797241733e+00, 5.08064710191544e+00, 5.06515244694742e+00, 5.04816472427676e+00, 5.02971851303796e+00, 5.00987225400227e+00, 4.98870757530277e+00, 4.96632797500953e+00, 4.94285691930845e+00, 4.91843543891444e+00, 4.89321932406352e+00, 4.86737602896215e+00, 4.84108139968188e+00, 4.81451633569554e+00, 4.78786348573561e+00, 4.76130406503103e+00, 4.73501486504789e+00, 4.70916551034883e+00, 4.68391600153827e+00, 4.65941456949924e+00, 4.63579585481184e+00, 4.61317941750502e+00, 4.59166857592765e+00, 4.57134956911377e+00, 4.55229103404776e+00, 4.53454378720080e+00, 4.51814089817004e+00, 4.50309804186924e+00, 4.48941411425685e+00, 4.47707209489957e+00, 4.46604013767759e+00, 4.45627286861470e+00, 4.44771286717309e+00, 4.44029230443226e+00, 4.43393470845356e+00, 4.42855682393431e+00, 4.42407053013878e+00, 4.42038477825741e+00, 4.41740750702700e+00, 4.41504749389591e+00, 4.41321609849582e+00, 4.41182885591483e+00, 4.41080687943811e+00, 4.41007803613979e+00, 4.40957786398821e+00, 4.40925020587982e+00, 4.40904754404716e+00, 4.40893102730513e+00, 4.40887019323373e+00, 4.40884239722585e+00, 4.40883196991600e+00, 4.40882913342851e+00, 4.40882871476867e+00, 4.40882868306901e+00, 4.40882812151262e+00, 4.40882431688328e+00, 4.40881033063104e+00, 4.40877304816934e+00, 4.40869145515382e+00, 4.40853518993954e+00, 4.40826342830909e+00, 4.40782414868335e+00, 4.40715381696413e+00, 4.40617751990964e+00, 4.40480956446409e+00, 4.40295454770943e+00, 4.40050888808366e+00, 4.39736279332811e+00, 4.39340262455286e+00, 4.38851359931564e+00, 4.38258276038974e+00, 4.37550212185834e+00, 4.36717189139298e+00, 4.35750365820366e+00, 4.34642343128995e+00, 4.33387441317394e+00, 4.31981940079836e+00, 4.30424271777942e+00, 4.28715160020571e+00, 4.26857698060237e+00, 4.24857363999885e+00, 4.22721972442231e+00, 4.20461564769748e+00, 4.18088242546934e+00, 4.15615950457123e+00, 4.13060216646849e+00, 4.10437859330349e+00, 4.07766669032084e+00, 4.05065075975266e+00, 4.02351811931103e+00, 3.99645575394346e+00, 3.96964708297464e+00, 3.94326891652701e+00, 3.91748866539701e+00, 3.89246185755869e+00, 3.86833000244214e+00, 3.84521883150806e+00, 3.82323693097512e+00, 3.80247477050470e+00, 3.78300412084815e+00, 3.76487784443623e+00, 3.74813003595445e+00, 3.73277648516909e+00, 3.71881543147247e+00, 3.70622857842910e+00, 3.69498233653382e+00, 3.68502926290097e+00, 3.67630966719488e+00, 3.66875335338417e+00, 3.66228146660012e+00, 3.65680841340241e+00, 3.65224382216507e+00, 3.64849450829412e+00, 3.64546640688732e+00, 3.64306643362880e+00, 3.64120423358340e+00, 3.63979377750300e+00, 3.63875476660052e+00, 3.63801380970753e+00, 3.63750534140070e+00, 3.63717225601667e+00, 3.63696624028690e+00, 3.63684779630821e+00, 3.63678595631302e+00, 3.63675770075134e+00, 3.63674710105689e+00, 3.63674421767754e+00, 3.63674379209659e+00, 3.63674376014809e+00, 3.63674319781827e+00, 3.63673938794921e+00, 3.63672538244116e+00, 3.63668804869634e+00, 3.63660634368145e+00, 3.63644986486633e+00, 3.63617773457466e+00, 3.63573786660577e+00, 3.63506665515372e+00, 3.63408911602794e+00, 3.63271949889492e+00, 3.63086237664287e+00, 3.62841420400569e+00, 3.62526532236467e+00, 3.62130237142779e+00, 3.61641105171456e+00, 3.61047916509564e+00, 3.60339984489423e+00, 3.59507487323369e+00, 3.58541797247578e+00, 3.57435795077551e+00, 3.56184157989469e+00, 3.54783608714819e+00, 3.53233115306149e+00, 3.51534032196821e+00, 3.49690175392951e+00, 3.47707827216808e+00, 3.45595668946190e+00, 3.43364642811929e+00, 3.41027747951500e+00, 3.38599777883370e+00, 3.36097009675193e+00, 3.33536857050305e+00, 3.30937501056782e+00, 3.28317512496443e+00, 3.25695480017710e+00, 3.23089656626261e+00, 3.20517635452229e+00, 3.17996063108614e+00, 3.15540396132102e+00, 3.13164703110597e+00, 3.10881512468201e+00, 3.08701703746219e+00, 3.06634438742582e+00, 3.04687128091546e+00, 3.02865428710061e+00, 3.01173267856435e+00, 2.99612890153257e+00, 2.98184924637467e+00, 2.96888469568685e+00, 2.95721193254878e+00, 2.94679449493727e+00, 2.93758406370265e+00, 2.92952187114659e+00, 2.92254021541227e+00, 2.91656406299469e+00, 2.91151271808861e+00, 2.90730153358775e+00, 2.90384363468151e+00, 2.90105162249922e+00, 2.89883922244446e+00, 2.89712284004319e+00, 2.89582298656420e+00, 2.89486553757690e+00, 2.89418279013105e+00, 2.89371428842681e+00, 2.89340739362927e+00, 2.89321758069471e+00, 2.89310845342654e+00, 2.89305147809035e+00, 2.89302544534488e+00, 2.89301567952386e+00, 2.89301302298004e+00, 2.89301263087960e+00, 2.89301260153400e+00, 2.89301208621059e+00, 2.89300859481753e+00, 2.89299576007120e+00, 2.89296154722576e+00, 2.89288667262112e+00, 2.89274327638431e+00, 2.89249390157676e+00, 2.89209082531347e+00, 2.89147577955916e+00, 2.89058009032679e+00, 2.88932525370389e+00, 2.88762395534114e+00, 2.88538152663955e+00, 2.88249781587971e+00, 2.87886943616593e+00, 2.87439233478715e+00, 2.86896461120901e+00, 2.86248949448797e+00, 2.85487837677471e+00, 2.84605378922963e+00, 2.83595220157433e+00, 2.82452652789666e+00, 2.81174823001048e+00, 2.79760892577857e+00, 2.78212143264186e+00, 2.76532020458400e+00, 2.74726115153193e+00, 2.72802086087313e+00, 2.70769526839282e+00, 2.68639784795923e+00, 2.66425740407532e+00, 2.64141555853276e+00, 2.61802402263540e+00, 2.59424174152685e+00, 2.57023198921249e+00, 2.54615948392627e+00, 2.52218758498698e+00, 2.49847562485983e+00, 2.47517642370612e+00, 2.45243402772932e+00, 2.43038170645032e+00, 2.40914023718036e+00, 2.38881649725080e+00, 2.36950237620735e+00, 2.35127401164961e+00, 2.33419134428454e+00, 2.31829798061827e+00, 2.30362134594203e+00, 2.29017310606632e+00, 2.27794983359006e+00, 2.26693389313531e+00, 2.25709451957659e+00, 2.24838906343139e+00, 2.24076437784545e+00, 2.23415832166083e+00, 2.22850135266211e+00, 2.22371818415048e+00, 2.21972947653090e+00, 2.21645353377085e+00, 2.21380797266059e+00, 2.21171133110327e+00, 2.21008458054521e+00, 2.20885250746626e+00, 2.20794492987909e+00, 2.20729771723768e+00, 2.20685358612324e+00, 2.20656264952759e+00, 2.20638270433265e+00, 2.20627924942109e+00, 2.20622523538840e+00, 2.20620055564638e+00, 2.20619129737710e+00, 2.20618877889903e+00, 2.20618840717679e+00, 2.20618837934593e+00, 2.20618789048283e+00, 2.20618457836208e+00, 2.20617240263539e+00, 2.20613994646128e+00, 2.20606891612111e+00, 2.20593288146839e+00, 2.20569630651036e+00, 2.20531391283160e+00, 2.20473041098019e+00, 2.20388062629927e+00, 2.20269003591523e+00, 2.20107572260074e+00, 2.19894773898291e+00, 2.19621086213195e+00, 2.19276670416451e+00, 2.18851612953468e+00, 2.18336191477175e+00, 2.17721157235199e+00, 2.16998024809856e+00, 2.16159359199434e+00, 2.15199049653571e+00, 2.14112559557638e+00, 2.12897142057565e+00, 2.11552012050011e+00, 2.10078466616420e+00, 2.08479947896989e+00, 2.06762044689944e+00, 2.04932431602546e+00, 2.03000747233357e+00, 2.00978415480980e+00, 1.98878416502238e+00, 1.96715015939524e+00, 1.94503462676098e+00, 1.92259666458305e+00, 1.89999867179191e+00, 1.87740307426890e+00, 1.85496919090905e+00, 1.83285033467332e+00, 1.81119122531369e+00, 1.79012577004775e+00, 1.76977524705021e+00, 1.75024690585503e+00, 1.73163298005517e+00, 1.71401009213263e+00, 1.69743901852014e+00, 1.68196477532032e+00, 1.66761698132556e+00, 1.65441045461167e+00, 1.64234600131510e+00, 1.63141135943103e+00, 1.62158226574260e+00, 1.61282361952898e+00, 1.60509072182690e+00, 1.59833057322202e+00, 1.59248321607161e+00, 1.58748310854139e+00, 1.58326051787281e+00, 1.57974291903856e+00, 1.57685638267637e+00, 1.57452693329966e+00, 1.57268185571907e+00, 1.57125092484461e+00, 1.57016753203255e+00, 1.56936968028843e+00, 1.56880082123864e+00, 1.56841050900596e+00, 1.56815484999807e+00, 1.56799673301727e+00, 1.56790583076741e+00, 1.56785837140163e+00, 1.56783668677997e+00, 1.56782855212239e+00, 1.56782633929760e+00, 1.56782601268927e+00, 1.56782598817122e+00, 1.56782555663296e+00, 1.56782263289789e+00, 1.56781188494067e+00, 1.56778323484957e+00, 1.56772053491077e+00, 1.56760045735254e+00, 1.56739164249188e+00, 1.56705414491275e+00, 1.56653921135912e+00, 1.56578941707058e+00, 1.56473917908010e+00, 1.56331565528135e+00, 1.56144002666808e+00, 1.55902914699618e+00, 1.55599752936919e+00, 1.55225962330456e+00, 1.54773231938794e+00, 1.54233760262537e+00, 1.53600526124596e+00, 1.52867554631422e+00, 1.52030167042388e+00, 1.51085203216979e+00, 1.50031205794741e+00, 1.48868556438348e+00, 1.47599556326772e+00, 1.46228445551318e+00, 1.44761359006397e+00, 1.43206219587448e+00, 1.41572572777284e+00, 1.39871369766758e+00, 1.38114708870726e+00, 1.36315546955075e+00, 1.34487393733835e+00, 1.32644002053940e+00, 1.30799066672166e+00, 1.28965942640731e+00, 1.27157392421482e+00, 1.25385368456235e+00, 1.23660835365180e+00, 1.21993633449455e+00, 1.20392382928749e+00, 1.18864426490972e+00, 1.17415806351969e+00, 1.16051271144821e+00, 1.14774307556666e+00, 1.13587191644173e+00, 1.12491055100938e+00, 1.11485962324235e+00, 1.10570994838642e+00, 1.09744340392648e+00, 1.09003384778307e+00, 1.08344805076564e+00, 1.07764663562236e+00, 1.07258501888724e+00, 1.06821435403484e+00, 1.06448247523364e+00, 1.06133484036909e+00, 1.05871547020221e+00, 1.05656787782754e+00, 1.05483597934139e+00, 1.05346497320859e+00, 1.05240217262148e+00, 1.05159777256000e+00, 1.05100553162719e+00, 1.05058334831466e+00, 1.05029371231325e+00, 1.05010401387085e+00, 1.04998669793123e+00, 1.04991925466317e+00, 1.04988404370623e+00, 1.04986795564496e+00, 1.04986192047649e+00, 1.04986027876610e+00, 1.04986003645315e+00, 1.04986001818557e+00, 1.04985969562891e+00, 1.04985751026117e+00, 1.04984947663502e+00, 1.04982806206661e+00, 1.04978119746381e+00, 1.04969144865832e+00, 1.04953538260451e+00, 1.04928315950909e+00, 1.04889837806470e+00, 1.04833819623228e+00, 1.04755374412960e+00, 1.04649083819796e+00, 1.04509099665690e+00, 1.04329274517901e+00, 1.04103318880999e+00, 1.03824981179960e+00, 1.03488245190340e+00, 1.03087538089802e+00, 1.02617940983157e+00, 1.02075392737816e+00, 1.01456877405709e+00, 1.00760585529286e+00, 9.99860403208984e-01, 9.91341810950109e-01, 9.82073983771290e-01, 9.72095176908580e-01, 9.61457319436018e-01, 9.50224853478638e-01, 9.38473146597472e-01, 9.26286559283044e-01, 9.13756267104851e-01, 9.00977946697742e-01, 8.88049435854017e-01, 8.75068470917762e-01, 8.62130590690634e-01, 8.49327277047465e-01, 8.36744380668751e-01, 8.24460858003432e-01, 8.12547824805218e-01, 8.01067913902440e-01, 7.90074911251375e-01, 7.79613635182383e-01, 7.69720018952050e-01, 7.60421355741284e-01, 7.51736667318598e-01, 7.43677161843867e-01, 7.36246751858683e-01, 7.29442609630220e-01, 7.23255743066492e-01, 7.17671580942620e-01, 7.12670560863523e-01, 7.08228717062691e-01, 7.04318267727815e-01, 7.00908203058194e-01, 6.97964875758250e-01, 6.95452595259389e-01, 6.93334225773905e-01, 6.91571786481511e-01, 6.90127049918020e-01, 6.88962132184779e-01, 6.88040066151603e-01, 6.87325346617372e-01, 6.86784434647389e-01, 6.86386207228799e-01, 6.86102338136096e-01, 6.85907596580134e-01, 6.85780051853468e-01, 6.85701174729044e-01, 6.85655829684748e-01, 6.85632155911389e-01, 6.85621339266145e-01, 6.85617281584826e-01, 6.85616177798609e-01, 6.85616014882132e-01, 6.85616002533286e-01, 6.85615783599078e-01, 6.85614300288536e-01, 6.85608847495870e-01, 6.85594312458607e-01, 6.85562503422878e-01, 6.85501587374810e-01, 6.85395660303904e-01, 6.85224471304076e-01, 6.84963319255647e-01, 6.84583137678064e-01, 6.84050779349515e-01, 6.83329507193084e-01, 6.82379691507545e-01, 6.81159705760349e-01, 6.79627003920088e-01, 6.77739351975180e-01, 6.75456175419571e-01, 6.72739973903029e-01, 6.69557744979562e-01, 6.65882352098816e-01, 6.61693768804449e-01, 6.56980132463178e-01, 6.51738547301401e-01, 6.45975588093616e-01, 6.39707471930418e-01, 6.32959884882125e-01, 6.25767471354409e-01, 6.18173014498884e-01, 6.10226354171119e-01, 6.01983102871507e-01, 5.93503228639239e-01, 5.84849576466707e-01, 5.76086396669503e-01, 5.67277940652670e-01, 5.58487173001041e-01, 5.49774635371007e-01, 5.41197483840044e-01, 5.32808708522702e-01, 5.24656533368216e-01, 5.16783985676367e-01, 5.09228619167306e-01, 5.02022371256166e-01, 4.95191534135330e-01, 4.88756819865594e-01, 4.82733501412988e-01, 4.77131613976959e-01, 4.71956203656505e-01, 4.67207613205536e-01, 4.62881797136014e-01, 4.58970660607661e-01, 4.55462418319157e-01, 4.52341970946827e-01, 4.49591297545823e-01, 4.47189862736201e-01, 4.45115037456926e-01, 4.43342531616262e-01, 4.41846836146410e-01, 4.40601670854917e-01, 4.39580433147786e-01, 4.38756641293822e-01, 4.38104364536380e-01, 4.37598631177302e-01, 4.37215804895620e-01, 4.36933919145022e-01, 4.36732959595834e-01, 4.36595085308195e-01, 4.36504780654347e-01, 4.36448931909537e-01, 4.36416824812570e-01, 4.36400062126214e-01, 4.36392403141890e-01, 4.36389529996579e-01, 4.36388748431780e-01, 4.36388633074468e-01, 4.36388624275059e-01, 4.36388467537253e-01, 4.36387405615935e-01, 4.36383501886567e-01, 4.36373096021316e-01, 4.36350323262914e-01, 4.36306711561363e-01, 4.36230873121792e-01, 4.36108305221640e-01, 4.35921313167978e-01, 4.35649065874868e-01, 4.35267791604500e-01, 4.34751117784567e-01, 4.34070554364488e-01, 4.33196114849085e-01, 4.32097063002665e-01, 4.30742766444261e-01, 4.29103631315900e-01, 4.27152085420033e-01, 4.24863571340462e-01, 4.22217506798551e-01, 4.19198167541349e-01, 4.15795448959606e-01, 4.12005466691152e-01, 4.07830963642205e-01, 4.03281500742374e-01, 3.98373420568598e-01, 3.93129585690323e-01, 3.87578906034033e-01, 3.81755680609820e-01, 3.75698787655382e-01, 3.69450763025624e-01, 3.63056809262560e-01, 3.56563777365295e-01, 3.50019160287721e-01, 3.43470132256035e-01, 3.36962661822820e-01, 3.30540719828570e-01, 3.24245596689399e-01, 3.18115337095474e-01, 3.12184294573409e-01, 3.06482803601522e-01, 3.01036963143887e-01, 2.95868522598405e-01, 2.90994859205167e-01, 2.86429034873470e-01, 2.82179920074205e-01, 2.78252372802001e-01, 2.74647461511265e-01, 2.71362722228020e-01, 2.68392441581085e-01, 2.65727959124777e-01, 2.63357983892134e-01, 2.61268921490165e-01, 2.59445209120553e-01, 2.57869656606045e-01, 2.56523791787240e-01, 2.55388208528215e-01, 2.54442915072206e-01, 2.53667679695738e-01, 2.53042369625115e-01, 2.52547278128336e-01, 2.52163433713513e-01, 2.51872884585474e-01, 2.51658951054909e-01, 2.51506438552846e-01, 2.51401804335138e-01, 2.51333271882939e-01, 2.51290888386567e-01, 2.51266522471453e-01, 2.51253801380056e-01, 2.51247989033943e-01, 2.51245808626395e-01, 2.51245215503066e-01, 2.51245127959319e-01, 2.51245121234394e-01, 2.51245000830176e-01, 2.51244185074237e-01, 2.51241186274433e-01, 2.51233192616020e-01, 2.51215698897409e-01, 2.51182197127920e-01, 2.51123939756087e-01, 2.51029787032367e-01, 2.50886148522639e-01, 2.50677026994083e-01, 2.50384170682072e-01, 2.49987337199143e-01, 2.49464668970284e-01, 2.48793176029031e-01, 2.47949317322655e-01, 2.46909666492045e-01, 2.45651642668915e-01, 2.44154281540292e-01, 2.42399017230395e-01, 2.40370441956444e-01, 2.38057008432161e-01, 2.35451640045065e-01, 2.32552216195503e-01, 2.29361904920607e-01, 2.25889321856448e-01, 2.22148503293838e-01, 2.18158690928041e-01, 2.13943936121384e-01, 2.09532541270915e-01, 2.04956364425021e-01, 2.00250019974792e-01, 1.95450012600326e-01, 1.90593843447450e-01, 1.85719126747859e-01, 1.80862751991812e-01, 1.76060121711890e-01, 1.71344488461076e-01, 1.66746407262449e-01, 1.62293312277387e-01, 1.58009219249040e-01, 1.53914548904430e-01, 1.50026061296668e-01, 1.46356887252904e-01, 1.42916640736526e-01, 1.39711594977237e-01, 1.36744905509679e-01, 1.34016864556805e-01, 1.31525173225536e-01, 1.29265220467740e-01, 1.27230360433198e-01, 1.25412182470408e-01, 1.23800770427271e-01, 1.22384949925936e-01, 1.21152523841966e-01, 1.20090497259090e-01, 1.19185293690074e-01, 1.18422964379420e-01, 1.17789392093948e-01, 1.17270490047721e-01, 1.16852395604041e-01, 1.16521657269270e-01, 1.16265412367938e-01, 1.16071551790601e-01, 1.15928867447741e-01, 1.15827177635912e-01, 1.15757425486837e-01, 1.15711746050503e-01, 1.15683498344698e-01, 1.15667259833304e-01, 1.15658782190247e-01, 1.15654908758947e-01, 1.15653455713085e-01, 1.15653060450506e-01, 1.15653002110617e-01, 1.15652997594871e-01, 1.15652916298919e-01, 1.15652365507551e-01, 1.15650340750191e-01, 1.15644943567061e-01, 1.15633132352505e-01, 1.15610513953233e-01, 1.15571185028884e-01, 1.15507631549398e-01, 1.15410694035062e-01, 1.15269605139746e-01, 1.15072104805559e-01, 1.14804636342003e-01, 1.14452624287412e-01, 1.14000831728050e-01, 1.13433790892325e-01, 1.12736296418831e-01, 1.11893945949954e-01, 1.10893707976960e-01, 1.09724492605619e-01, 1.08377697630235e-01, 1.06847700506865e-01, 1.05132266943075e-01, 1.03232849169181e-01, 1.01154751613919e-01, 9.89071485134581e-02, 9.65029465102889e-02, 9.39584948834726e-02, 9.12931558571941e-02, 8.85287565465314e-02, 8.56889516414840e-02, 8.27985311775867e-02, 7.98827102161734e-02, 7.69664367911336e-02, 7.40737512287977e-02, 7.12272243658330e-02, 6.84474949527880e-02, 6.57529184377807e-02, 6.31593311898158e-02, 6.06799267718108e-02, 5.83252346738501e-02, 5.61031873244895e-02, 5.40192583518788e-02, 5.20766539105253e-02, 5.02765392129385e-02, 4.86182838979001e-02, 4.70997121748669e-02, 4.57173464581277e-02, 4.44666361359189e-02, 4.33421659609126e-02, 4.23378411193500e-02, 4.14470482212836e-02, 4.06627931948620e-02, 3.99778183473822e-02, 3.93847016902749e-02, 3.88759420494011e-02, 3.84440335427190e-02, 3.80815327580413e-02, 3.77811214609851e-02, 3.75356669682434e-02, 3.73382814975782e-02, 3.71823809209042e-02, 3.70617424702138e-02, 3.69705601468908e-02, 3.69034959282910e-02, 3.68557244058874e-02, 3.68229682668453e-02, 3.68015220646640e-02, 3.67882620111510e-02, 3.67806400340136e-02, 3.67766610338567e-02, 3.67748430777493e-02, 3.67741611129047e-02, 3.67739756031833e-02, 3.67739482223817e-02, 3.67739460858134e-02, 3.67739073996960e-02, 3.67736452962326e-02, 3.67726817863625e-02, 3.67701135010217e-02, 3.67644932682820e-02, 3.67537313711042e-02, 3.67350211005544e-02, 3.67047931047431e-02, 3.66587028645302e-02, 3.65916553137087e-02, 3.64978700508658e-02, 3.63709896418056e-02, 3.62042320785241e-02, 3.59905864757533e-02, 3.57230485466422e-02, 3.53948893905089e-02, 3.49999478361118e-02, 3.45329333049884e-02, 3.39897232694888e-02, 3.33676373060419e-02, 3.26656689056444e-02, 3.18846569426361e-02, 3.10273812128585e-02, 3.00985707180246e-02, 2.91048191365526e-02, 2.80544086874981e-02, 2.69570506787735e-02, 2.58235576485326e-02, 2.46654673875604e-02, 2.34946426355318e-02, 2.23228714791359e-02, 2.11614923528103e-02, 2.00210642737018e-02, 1.89110980172531e-02, 1.78398580204524e-02, 1.68142386131578e-02, 1.58397124066623e-02, 1.49203438623742e-02, 1.40588575793181e-02, 1.32567488170485e-02, 1.25144231477492e-02, 1.18313526880653e-02, 1.12062377896436e-02, 1.06371650379753e-02, 1.01217546216947e-02, 9.65729234312906e-03, 9.24084356278332e-03, 8.86934808630890e-03, 8.53969635327712e-03, 8.24878826722131e-03, 7.99357664918254e-03, 7.77109765761619e-03, 7.57849065539862e-03, 7.41300997476912e-03, 7.27203087840955e-03, 7.15305177548498e-03, 7.05369445308257e-03, 6.97170374717526e-03, 6.90494771913208e-03, 6.85141903732589e-03, 6.80923790140664e-03, 6.77665650325032e-03, 6.75206470898626e-03, 6.73399638739092e-03, 6.72113561694765e-03, 6.71232189021698e-03, 6.70655340703252e-03, 6.70298760854937e-03, 6.70093824436346e-03, 6.69986847019732e-03, 6.69937972232759e-03, 6.69919638272631e-03, 6.69914651053372e-03, 6.69913914952414e-03, 6.69913857024529e-03, 6.69912801882333e-03, 6.69905653172709e-03, 6.69879374259974e-03, 6.69809328169269e-03, 6.69656053508483e-03, 6.69362588891875e-03, 6.68852485144544e-03, 6.68028649012253e-03, 6.66773165317736e-03, 6.64948242260621e-03, 6.62398411413978e-03, 6.58954084138141e-03, 6.54436514320444e-03, 6.48664140205657e-03, 6.41460175394126e-03, 6.32661195208861e-03, 6.22126329201578e-03, 6.09746538635747e-03, 5.95453348644437e-03, 5.79226339680616e-03, 5.61098701457308e-03, 5.41160228626669e-03, 5.19557294953898e-03, 4.96489573165082e-03, 4.72203549303440e-03, 4.46983180615046e-03, 4.21138325408066e-03, 3.94991792520738e-03, 3.68865984343181e-03, 3.43070121742706e-03, 3.17888941134026e-03, 2.93573562017801e-03, 2.70334971393282e-03, 2.48340299453442e-03, 2.27711804806052e-03, 2.08528272882358e-03, 1.90828373587859e-03, 1.74615432554415e-03, 1.59863048682674e-03, 1.46521034993707e-03, 1.34521253605779e-03, 1.23783031614533e-03, 1.14217955178475e-03, 1.05733928493468e-03, 9.82384526701630e-04, 9.16411350107893e-04, 8.58554867057215e-04, 8.08001034064532e-04, 7.63993420730635e-04, 7.25836073179664e-04, 6.92893469103062e-04, 6.64588384005960e-04, 6.40398336094528e-04, 6.19851165634344e-04, 6.02520215028157e-04, 5.88019488239055e-04, 5.75999076942502e-04, 5.66141052910751e-04, 5.58155949520689e-04, 5.51779892368209e-04, 5.46772387238548e-04, 5.42914729168231e-04, 5.40008957246485e-04, 5.37877246528318e-04, 5.36361602737401e-04, 5.35323709764945e-04, 5.34644775841642e-04, 5.34225232160956e-04, 5.33984157276457e-04, 5.33858329572645e-04, 5.33800845869514e-04, 5.33779283045099e-04, 5.33773417557448e-04, 5.33772551827738e-04, 5.33772483093387e-04, 5.33771223470758e-04, 5.33762689402195e-04, 5.33731318252961e-04, 5.33647701919548e-04, 5.33464748316170e-04, 5.33114518541211e-04, 5.32505929723648e-04, 5.31523532234068e-04, 5.30027587084980e-04, 5.27855676559774e-04, 5.24826067065831e-04, 5.20742995989032e-04, 5.15403963074590e-04, 5.08608965122725e-04, 5.00171421833784e-04, 4.89930311761810e-04, 4.77762793757996e-04, 4.63596364676633e-04, 4.47419438897256e-04, 4.29289171760772e-04, 4.09335424494728e-04, 3.87760009844307e-04, 3.64830771948377e-04, 3.40870616653888e-04, 3.16242247135889e-04, 2.91329950290908e-04, 2.66520167940598e-04, 2.42182651367884e-04, 2.18653727241502e-04, 1.96222739112351e-04, 1.75122302092490e-04, 1.55522777309791e-04, 1.37531268784130e-04, 1.21195194082269e-04, 1.06509848367686e-04, 9.34284926685791e-05, 8.18728938662971e-05, 7.17425020360096e-05, 6.29216030505475e-05, 5.52851283153188e-05, 4.87043578001334e-05, 4.30530617370195e-05, 3.82132876879382e-05, 3.40792372872656e-05, 3.05581873289995e-05, 2.75688063045010e-05, 2.50384708113671e-05, 2.29013466759298e-05, 2.10980056998267e-05, 1.95761620354062e-05, 1.82915330739775e-05, 1.72078973168510e-05, 1.62960280682967e-05, 1.55319872243709e-05, 1.48955632797154e-05, 1.43692441518795e-05, 1.39376287089377e-05, 1.35870699935481e-05, 1.33054531703536e-05, 1.30820612973537e-05, 1.29074670542906e-05, 1.27734058539275e-05, 1.26726331384219e-05, 1.25987946555384e-05, 1.25463308631169e-05, 1.25104171413426e-05, 1.24869286075580e-05, 1.24724153298672e-05, 1.24640762102501e-05, 1.24597237478596e-05, 1.24577353695088e-05, 1.24569895092463e-05, 1.24567866208736e-05, 1.24567566759382e-05, 1.24567542991727e-05, 1.24567107606666e-05, 1.24564157841789e-05, 1.24553314542197e-05, 1.24524413216420e-05, 1.24461178039326e-05, 1.24340131462120e-05, 1.24129807623774e-05, 1.23790347483776e-05, 1.23273568046756e-05, 1.22523618024672e-05, 1.21478369825567e-05, 1.20071758402673e-05, 1.18237356137030e-05, 1.15913493941022e-05, 1.13050004060256e-05, 1.09615901378035e-05, 1.05606009244141e-05, 1.01043472513496e-05, 9.59759856389297e-06, 9.04671123436174e-06, 8.45869494892071e-06, 7.84043594848072e-06, 7.19816317484856e-06, 6.53803057026892e-06, 5.86893779974673e-06, 5.20647980639666e-06, 4.57431572886691e-06, 3.99958513597198e-06, 3.50338580723530e-06, 3.09025806544128e-06, 2.73884805501723e-06, 2.41683586921226e-06, 2.10712416848224e-06, 1.81318997238833e-06, 1.61058185543794e-06, 1.50355801586954e-06, 1.40166115603080e-06, 1.40310964980017e-06, 1.37759065044207e-06, 1.30383684915358e-06, 1.19608247395320e-06, 1.28015474309962e-06, 1.30874720216667e-06, 1.26227504626542e-06, 1.20474859599540e-06, 1.31235194889001e-06, 1.34435047716153e-06, 1.31176564188706e-06, 1.23791516060123e-06, 1.28208041250133e-06, 1.35358050368239e-06, 1.38340632727108e-06, 1.38065128396647e-06, 1.36691680285850e-06, 1.36636111669475e-06, 1.38169464936055e-06, 1.39778830821058e-06, 1.40628377565122e-06, 1.40698770323411e-06, 1.40262328418572e-06, 1.39610753269181e-06, 1.38957554618457e-06, 1.38418465844836e-06, 1.38027795504314e-06, 1.37770169494900e-06, 1.37611426165307e-06, 1.37518268950032e-06, 1.37465775207346e-06, 1.37437560443487e-06, 1.37423433343997e-06, 1.37417120615879e-06, 1.37414775913669e-06, 1.37414140155556e-06, 1.37414046452733e-06, 1.37414047252093e-06, 1.37414164846916e-06, 1.37414960654780e-06, 1.37417866152845e-06, 1.37425460255969e-06, 1.37441318948106e-06, 1.37468820327058e-06, 1.37507927444602e-06, 1.37548907552798e-06, 1.37562910396127e-06, 1.37492550411622e-06, 1.37253409793203e-06, 1.36774821335450e-06, 1.36145448337288e-06, 1.35937681200460e-06, 1.37270611411111e-06, 1.40461812137005e-06, 1.44251792910666e-06, 1.46962923697203e-06, 1.47120087223840e-06, 1.43614023695449e-06, 1.35662367828360e-06, 1.26005262490736e-06, 1.39975327151376e-06, 1.47763012581237e-06, 1.49002892513295e-06, 1.42045382744982e-06, 1.40561278527684e-06, 1.54247030081933e-06, 1.58615960810238e-06, 1.53747921771696e-06, 1.42786078960455e-06, 1.35868341022416e-06, 1.52810578146756e-06, 1.83350408722305e-06, 2.14702653611454e-06, 2.37370147981864e-06, 2.42131776051480e-06, 2.24062858578122e-06, 1.85844670586022e-06, 1.37113885845253e-06, 9.06361145247558e-07, 5.88925392897011e-07, 4.68507227679993e-07, 3.81002045024559e-07, 6.21724893790088e-15, 5.57923671706817e-07, 7.73550332233697e-07, 8.66125828835607e-07, 8.26110641227729e-07, 6.84539582351817e-07, 4.96387301929246e-07, 7.10542735760100e-15, 6.21724893790088e-15, 6.21724893790088e-15, 7.10542735760100e-15, 6.21724893790088e-15, 7.10542735760100e-15, 6.21724893790088e-15, 6.21724893790088e-15, 6.21724893790088e-15, 6.21724893790088e-15, 6.21724893790088e-15, 6.21724893790088e-15, 6.21724893790088e-15, 5.32907051820075e-15, 6.21724893790088e-15, 6.21724893790088e-15, 6.21724893790088e-15, 6.21724893790088e-15, 6.21724893790088e-15, 5.32907051820075e-15, 6.21724893790088e-15, 6.21724893790088e-15, 6.21724893790088e-15],
[ 8.00000000000000e+00, 7.99999999704240e+00, 7.99999990558770e+00, 7.99999928596774e+00, 7.99999700816574e+00, 7.99999093627851e+00, 7.99997764757700e+00, 7.99995219611277e+00, 7.99990792950872e+00, 7.99983636600523e+00, 7.99972713705114e+00, 7.99956799879099e+00, 7.99934491376311e+00, 7.99904220205464e+00, 7.99864275912175e+00, 7.99812833554084e+00, 7.99747987216509e+00, 7.99667788257216e+00, 7.99570287334150e+00, 7.99453579162449e+00, 7.99315848868432e+00, 7.99155418759228e+00, 7.98970794307142e+00, 7.98760708156535e+00, 7.98524160996690e+00, 7.98260458204930e+00, 7.97969241248530e+00, 7.97650512940111e+00, 7.97304655767771e+00, 7.96932442666997e+00, 7.96535039764865e+00, 7.96114000806728e+00, 7.95671253168939e+00, 7.95209075565290e+00, 7.94730067765308e+00, 7.94237112854255e+00, 7.93733332771213e+00, 7.93222038056047e+00, 7.92706672910936e+00, 7.92190756830302e+00, 7.91677824167804e+00, 7.91171363085081e+00, 7.90674755360316e+00, 7.90191218523239e+00, 7.89723751726900e+00, 7.89275086667055e+00, 7.88847644720986e+00, 7.88443501303977e+00, 7.88064358239723e+00, 7.87711524717597e+00, 7.87385907172417e+00, 7.87088008178604e+00, 7.86817934207812e+00, 7.86575411864258e+00, 7.86359811991523e+00, 7.86170180844379e+00, 7.86005277344342e+00, 7.85863615292364e+00, 7.85743509299901e+00, 7.85643123122893e+00, 7.85560519043656e+00, 7.85493706943740e+00, 7.85440691746110e+00, 7.85399517975985e+00, 7.85368310294055e+00, 7.85345308990209e+00, 7.85328899586347e+00, 7.85317635878639e+00, 7.85310255947598e+00, 7.85305690872982e+00, 7.85303066104229e+00, 7.85301695650148e+00, 7.85301069458209e+00, 7.85300834548734e+00, 7.85300770647353e+00, 7.85300761215632e+00, 7.85300759978389e+00, 7.85300731152108e+00, 7.85300535849523e+00, 7.85299817892183e+00, 7.85297904047894e+00, 7.85293715475877e+00, 7.85285693205621e+00, 7.85271740364346e+00, 7.85249183369365e+00, 7.85214753735912e+00, 7.85164591536926e+00, 7.85094270909623e+00, 7.84998847356788e+00, 7.84872925960931e+00, 7.84710749037910e+00, 7.84506301221771e+00, 7.84253429508534e+00, 7.83945975402401e+00, 7.83577916006229e+00, 7.83143510676377e+00, 7.82637449712757e+00, 7.82055001467955e+00, 7.81392154224970e+00, 7.80645749204369e+00, 7.79813601117725e+00, 7.78894602791356e+00, 7.77888810556644e+00, 7.76797507360391e+00, 7.75623240912839e+00, 7.74369834681400e+00, 7.73042370166290e+00, 7.71647139658449e+00, 7.70191569562737e+00, 7.68684115335849e+00, 7.67134130089337e+00, 7.65551709883895e+00, 7.63947519628488e+00, 7.62332604237328e+00, 7.60718190240353e+00, 7.59115483355883e+00, 7.57535467603991e+00, 7.55988711371997e+00, 7.54485185462871e+00, 7.53034097600598e+00, 7.51643747179086e+00, 7.50321403271146e+00, 7.49073208107575e+00, 7.47904107433004e+00, 7.46817808376148e+00, 7.45816764758896e+00, 7.44902189123417e+00, 7.44074090183727e+00, 7.43331333906490e+00, 7.42671725990773e+00, 7.42092113142177e+00, 7.41588500218530e+00, 7.41156180059458e+00, 7.40789872600964e+00, 7.40483869721928e+00, 7.40232182178024e+00, 7.40028684957185e+00, 7.39867257447055e+00, 7.39741914945451e+00, 7.39646928273806e+00, 7.39576928571967e+00, 7.39526994757147e+00, 7.39492721613080e+00, 7.39470267025533e+00, 7.39456377482064e+00, 7.39448391588835e+00, 7.39444222005142e+00, 7.39442316835914e+00, 7.39441602132750e+00, 7.39441407715366e+00, 7.39441379019738e+00, 7.39441376486460e+00, 7.39441326848121e+00, 7.39440990540691e+00, 7.39439754233361e+00, 7.39436458648587e+00, 7.39429246121981e+00, 7.39415432427904e+00, 7.39391407660655e+00, 7.39352570168593e+00, 7.39293296635993e+00, 7.39206950414851e+00, 7.39085929147776e+00, 7.38921751616823e+00, 7.38705182627029e+00, 7.38426393614497e+00, 7.38075155584987e+00, 7.37641059967745e+00, 7.37113762036404e+00, 7.36483240727118e+00, 7.35740067992789e+00, 7.34875680287811e+00, 7.33882644395098e+00, 7.32754909602698e+00, 7.31488038231110e+00, 7.30079406729771e+00, 7.28528370031756e+00, 7.26836382609840e+00, 7.25007070738202e+00, 7.23046251840526e+00, 7.20961898479648e+00, 7.18764046466566e+00, 7.16464648652120e+00, 7.14077378095627e+00, 7.11617386341572e+00, 7.09101024332280e+00, 7.06545534907195e+00, 7.03968726784013e+00, 7.01388640323438e+00, 6.98823215239020e+00, 6.96289969769193e+00, 6.93805699765802e+00, 6.91386204788403e+00, 6.89046046755777e+00, 6.86798345122755e+00, 6.84654611032269e+00, 6.82624621524468e+00, 6.80716333720841e+00, 6.78935837965243e+00, 6.77287348191822e+00, 6.75773227277762e+00, 6.74394044788221e+00, 6.73148664287367e+00, 6.72034357228278e+00, 6.71046940306050e+00, 6.70180933031902e+00, 6.69429732140939e+00, 6.68785799274081e+00, 6.68240858178010e+00, 6.67786097458540e+00, 6.67412374723598e+00, 6.67110417788518e+00, 6.66871018518066e+00, 6.66685214875715e+00, 6.66544456866978e+00, 6.66440752319598e+00, 6.66366788850252e+00, 6.66316028925967e+00, 6.66282775628726e+00, 6.66262207553418e+00, 6.66250382182608e+00, 6.66244208049973e+00, 6.66241386986776e+00, 6.66240328700326e+00, 6.66240040819972e+00, 6.66239998329407e+00, 6.66239994921138e+00, 6.66239932021280e+00, 6.66239505865212e+00, 6.66237939266267e+00, 6.66233763256714e+00, 6.66224623999859e+00, 6.66207120608532e+00, 6.66176680120575e+00, 6.66127474981251e+00, 6.66052387349451e+00, 6.65943023383701e+00, 6.65789779376944e+00, 6.65581960198500e+00, 6.65307948974920e+00, 6.64955425315523e+00, 6.64511627692731e+00, 6.63963653866689e+00, 6.63298791558838e+00, 6.62504870005577e+00, 6.61570621648253e+00, 6.60486042133763e+00, 6.59242736106327e+00, 6.57834236053824e+00, 6.56256281804977e+00, 6.54507049205509e+00, 6.52587318047500e+00, 6.50500571459489e+00, 6.48253021609742e+00, 6.45853559605001e+00, 6.43313630708517e+00, 6.40647039244942e+00, 6.37869690579130e+00, 6.34999280131352e+00, 6.32054941336367e+00, 6.29056865638906e+00, 6.26025907988505e+00, 6.22983190878098e+00, 6.19949718864334e+00, 6.16946013873148e+00, 6.13991779625767e+00, 6.11105601419529e+00, 6.08304685447890e+00, 6.05604639990556e+00, 6.03019299245024e+00, 6.00560589351256e+00, 5.98238435282461e+00, 5.96060706701859e+00, 5.94033200559851e+00, 5.92159658059413e+00, 5.90441813580458e+00, 5.88879473164550e+00, 5.87470620169537e+00, 5.86211545672901e+00, 5.85097001110439e+00, 5.84120370473285e+00, 5.83273859152584e+00, 5.82548696227979e+00, 5.81935346662054e+00, 5.81423729512856e+00, 5.81003437940441e+00, 5.80663956493999e+00, 5.80394870957864e+00, 5.80186065940629e+00, 5.80027905441320e+00, 5.79911391842969e+00, 5.79828299181086e+00, 5.79771277115143e+00, 5.79733922785725e+00, 5.79710818646077e+00, 5.79697535380686e+00, 5.79690600121958e+00, 5.79687431299959e+00, 5.79686242558146e+00, 5.79685919190862e+00, 5.79685871462504e+00, 5.79685867781233e+00, 5.79685801677428e+00, 5.79685353814195e+00, 5.79683707418229e+00, 5.79679318705344e+00, 5.79669713984063e+00, 5.79651319293770e+00, 5.79619329208147e+00, 5.79567620522781e+00, 5.79488715496275e+00, 5.79373798123114e+00, 5.79212785575195e+00, 5.78994455448146e+00, 5.78706627785469e+00, 5.78336399038063e+00, 5.77870423178974e+00, 5.77295233188172e+00, 5.76597594133256e+00, 5.75764877209665e+00, 5.74785442501001e+00, 5.73649017020983e+00, 5.72347053945584e+00, 5.70873058958672e+00, 5.69222870399721e+00, 5.67394881443706e+00, 5.65390194818078e+00, 5.63212703452617e+00, 5.60869093778747e+00, 5.58368771906562e+00, 5.55723716343095e+00, 5.52948264011427e+00, 5.50058838861074e+00, 5.47073634165927e+00, 5.44012260611792e+00, 5.40895372496538e+00, 5.37744283897968e+00, 5.34580585662933e+00, 5.31425772720342e+00, 5.28300889704098e+00, 5.25226201345611e+00, 5.22220892671904e+00, 5.19302802787552e+00, 5.16488194945535e+00, 5.13791564710102e+00, 5.11225487252511e+00, 5.08800504163258e+00, 5.06525049582606e+00, 5.04405414926021e+00, 5.02445751003485e+00, 5.00648105898997e+00, 4.99012496588487e+00, 4.97537011927951e+00, 4.96217944332212e+00, 4.95049947177093e+00, 4.94026214680929e+00, 4.93138680743130e+00, 4.92378232930306e+00, 4.91734937504165e+00, 4.91198271088745e+00, 4.90757354295797e+00, 4.90401182392321e+00, 4.90118847934996e+00, 4.89899750245992e+00, 4.89733786694901e+00, 4.89611521007750e+00, 4.89524324262348e+00, 4.89464484853725e+00, 4.89425284514710e+00, 4.89401038432091e+00, 4.89387098573298e+00, 4.89379820487657e+00, 4.89376495019277e+00, 4.89375247513038e+00, 4.89374908160316e+00, 4.89374858072537e+00, 4.89374854280041e+00, 4.89374787096352e+00, 4.89374331916711e+00, 4.89372658624241e+00, 4.89368198210460e+00, 4.89358436549704e+00, 4.89339741210290e+00, 4.89307228007843e+00, 4.89254673060469e+00, 4.89174475018790e+00, 4.89057670916806e+00, 4.88894007718646e+00, 4.88672070118383e+00, 4.88379463487953e+00, 4.88003049076775e+00, 4.87529226678387e+00, 4.86944258049090e+00, 4.86234622469853e+00, 4.85387394086625e+00, 4.84390629163875e+00, 4.83233750266135e+00, 4.81907913760840e+00, 4.80406347008338e+00, 4.78724642229589e+00, 4.76860995327029e+00, 4.74816379832175e+00, 4.72594648564267e+00, 4.70202558365000e+00, 4.67649716257458e+00, 4.64948448391210e+00, 4.62113596023465e+00, 4.59162245419499e+00, 4.56113400837847e+00, 4.52987611627809e+00, 4.49806565857570e+00, 4.46592663767006e+00, 4.43368584660356e+00, 4.40156860588603e+00, 4.36979469309454e+00, 4.33857457581874e+00, 4.30810603931500e+00, 4.27857127747304e+00, 4.25013449117944e+00, 4.22294001390113e+00, 4.19711096225167e+00, 4.17274839101690e+00, 4.14993091858394e+00, 4.12871478023399e+00, 4.10913426296508e+00, 4.09120247554400e+00, 4.07491241018757e+00, 4.06023825640533e+00, 4.04713693196720e+00, 4.03554979979348e+00, 4.02540454219233e+00, 4.01661716498454e+00, 4.00909410361285e+00, 4.00273440150739e+00, 3.99743192810170e+00, 3.99307760040769e+00, 3.98956156846686e+00, 3.98677532182412e+00, 3.98461367193213e+00, 3.98297656454420e+00, 3.98177067706478e+00, 3.98091075874705e+00, 3.98032067666301e+00, 3.97993413746513e+00, 3.97969506389009e+00, 3.97955761535691e+00, 3.97948585339449e+00, 3.97945306443208e+00, 3.97944076410705e+00, 3.97943741811537e+00, 3.97943692425383e+00, 3.97943688717949e+00, 3.97943623462897e+00, 3.97943181350175e+00, 3.97941556094816e+00, 3.97937223743672e+00, 3.97927742412524e+00, 3.97909584197145e+00, 3.97878005869457e+00, 3.97826964107788e+00, 3.97749080015255e+00, 3.97635656542369e+00, 3.97476751132236e+00, 3.97261304432146e+00, 3.96977324253569e+00, 3.96612122117924e+00, 3.96152597724386e+00, 3.95585564572232e+00, 3.94898107848447e+00, 3.94077963665639e+00, 3.93113906945336e+00, 3.91996133842862e+00, 3.90716623761155e+00, 3.89269465848539e+00, 3.87651135535780e+00, 3.85860708208686e+00, 3.83899999536179e+00, 3.81773625202323e+00, 3.79488976657655e+00, 3.77056113754564e+00, 3.74487579431401e+00, 3.71798145580037e+00, 3.69004502488951e+00, 3.66124906471668e+00, 3.63178801259623e+00, 3.60186428417722e+00, 3.57168440575096e+00, 3.54145528961823e+00, 3.51138074014825e+00, 3.48165825084593e+00, 3.45247612885221e+00, 3.42401096493628e+00, 3.39642545475937e+00, 3.36986657026335e+00, 3.34446407694757e+00, 3.32032939181964e+00, 3.29755477647044e+00, 3.27621285905800e+00, 3.25635647755502e+00, 3.23801883440704e+00, 3.22121394999584e+00, 3.20593739930838e+00, 3.19216731322507e+00, 3.17986562299858e+00, 3.16897952380902e+00, 3.15944313068219e+00, 3.15117929743491e+00, 3.14410156657637e+00, 3.13811621521818e+00, 3.13312435908876e+00, 3.12902407386688e+00, 3.12571249048957e+00, 3.12308781915857e+00, 3.12105125580269e+00, 3.11950872507618e+00, 3.11837241585723e+00, 3.11756206883933e+00, 3.11700598124889e+00, 3.11664170090759e+00, 3.11641639058498e+00, 3.11628685352670e+00, 3.11621922176866e+00, 3.11618831986258e+00, 3.11617672742649e+00, 3.11617357399673e+00, 3.11617310855688e+00, 3.11617307372242e+00, 3.11617246201165e+00, 3.11616831757918e+00, 3.11615308217744e+00, 3.11611246996255e+00, 3.11602358996911e+00, 3.11585336972464e+00, 3.11555734144550e+00, 3.11507884386600e+00, 3.11434868309295e+00, 3.11328528490613e+00, 3.11179535855225e+00, 3.10977507821455e+00, 3.10711177295237e+00, 3.10368609909121e+00, 3.09937465109386e+00, 3.09405294839092e+00, 3.08759871729345e+00, 3.07989537000696e+00, 3.07083556815881e+00, 3.06032474745975e+00, 3.04828447439005e+00, 3.03465550611145e+00, 3.01940043171379e+00, 3.00250578640143e+00, 2.98398354969960e+00, 2.96387196304286e+00, 2.94223562961703e+00, 2.91916488828126e+00, 2.89477448204883e+00, 2.86920156842527e+00, 2.84260314269249e+00, 2.81515296513501e+00, 2.78703809866006e+00, 2.75845517389310e+00, 2.72960650438209e+00, 2.70069617487748e+00, 2.67192622078232e+00, 2.64349300703836e+00, 2.61558390049158e+00, 2.58837431206616e+00, 2.56202516507868e+00, 2.53668082514433e+00, 2.51246750680806e+00, 2.48949215359117e+00, 2.46784177262115e+00, 2.44758319308017e+00, 2.42876320962531e+00, 2.41140906756894e+00, 2.39552924549827e+00, 2.38111449246504e+00, 2.36813908006514e+00, 2.35656223380260e+00, 2.34632971230553e+00, 2.33737550657265e+00, 2.32962363399652e+00, 2.32299000315227e+00, 2.31738432518452e+00, 2.31271204619291e+00, 2.30887627260164e+00, 2.30577965852491e+00, 2.30332622112198e+00, 2.30142304741587e+00, 2.29998185455268e+00, 2.29892036545319e+00, 2.29816346358204e+00, 2.29764409430803e+00, 2.29730388605201e+00, 2.29709347194317e+00, 2.29697250169912e+00, 2.29690934345217e+00, 2.29688048572713e+00, 2.29686966016577e+00, 2.29686671534753e+00, 2.29686628069842e+00, 2.29686624815622e+00, 2.29686567653614e+00, 2.29686180372526e+00, 2.29684756685172e+00, 2.29680961648883e+00, 2.29672656261619e+00, 2.29656750286041e+00, 2.29629089143774e+00, 2.29584379996323e+00, 2.29516161289960e+00, 2.29416819075367e+00, 2.29277652298732e+00, 2.29088987978699e+00, 2.28840345719599e+00, 2.28520649366702e+00, 2.28118481806470e+00, 2.27622377003058e+00, 2.27021141419633e+00, 2.26304195105019e+00, 2.25461921060409e+00, 2.24486010178360e+00, 2.23369788207993e+00, 2.22108510973341e+00, 2.20699614553065e+00, 2.19142908372294e+00, 2.17440701158663e+00, 2.15597852409852e+00, 2.13621745280891e+00, 2.11522180440147e+00, 2.09311194231183e+00, 2.07002808153803e+00, 2.04612719976661e+00, 2.02157949471684e+00, 1.99656453619294e+00, 1.97126727043413e+00, 1.94587403351097e+00, 1.92056872017337e+00, 1.89552923599779e+00, 1.87092433587766e+00, 1.84691092326163e+00, 1.82363185462755e+00, 1.80121426490652e+00, 1.77976840397603e+00, 1.75938695340689e+00, 1.74014477722432e+00, 1.72209905075726e+00, 1.70528970738543e+00, 1.68974014342255e+00, 1.67545812549825e+00, 1.66243685150044e+00, 1.65065612429272e+00, 1.64008360600013e+00, 1.63067612878979e+00, 1.62238104508200e+00, 1.61513760553565e+00, 1.60887835668819e+00, 1.60353055171075e+00, 1.59901756745502e+00, 1.59526031906041e+00, 1.59217866022760e+00, 1.58969275331346e+00, 1.58772438919246e+00, 1.58619823291563e+00, 1.58504296811451e+00, 1.58419231131409e+00, 1.58358586719807e+00, 1.58316979763056e+00, 1.58289728093244e+00, 1.58272874341678e+00, 1.58263185222110e+00, 1.58258126661584e+00, 1.58255815371055e+00, 1.58254948327405e+00, 1.58254712470625e+00, 1.58254677658663e+00, 1.58254675045375e+00, 1.58254629049326e+00, 1.58254317419420e+00, 1.58253171836371e+00, 1.58250118143677e+00, 1.58243435268319e+00, 1.58230636978724e+00, 1.58208381322468e+00, 1.58172412082998e+00, 1.58117536005324e+00, 1.58037638837401e+00, 1.57925742379848e+00, 1.57774103693261e+00, 1.57574356348891e+00, 1.57317692111270e+00, 1.56995079724508e+00, 1.56597515585733e+00, 1.56116299117071e+00, 1.55543323717450e+00, 1.54871372445169e+00, 1.54094406228959e+00, 1.53207831608004e+00, 1.52208734918389e+00, 1.51096070588894e+00, 1.49870792831434e+00, 1.48535922477077e+00, 1.47096543894160e+00, 1.45559730621509e+00, 1.43934402278372e+00, 1.42231119152274e+00, 1.40461824289233e+00, 1.38639545622727e+00, 1.36778072453117e+00, 1.34891621299901e+00, 1.32994505779261e+00, 1.31100823803923e+00, 1.29224173250730e+00, 1.27377404547997e+00, 1.25572415683969e+00, 1.23819992206440e+00, 1.22129692110708e+00, 1.20509773277144e+00, 1.18967159428547e+00, 1.17507439468796e+00, 1.16134894514682e+00, 1.14852546872868e+00, 1.13662225545543e+00, 1.12564643460718e+00, 1.11559482406829e+00, 1.10645482507418e+00, 1.09820533917909e+00, 1.09081769199371e+00, 1.08425655478758e+00, 1.07848086013279e+00, 1.07344471124656e+00, 1.06909828655267e+00, 1.06538874130470e+00, 1.06226110706223e+00, 1.05965918761376e+00, 1.05752644689198e+00, 1.05580688086648e+00, 1.05444586170178e+00, 1.05339093902401e+00, 1.05259258032566e+00, 1.05200483069502e+00, 1.05158587145257e+00, 1.05129845808241e+00, 1.05111022011418e+00, 1.05099380926258e+00, 1.05092688696750e+00, 1.05089194819022e+00, 1.05087598453228e+00, 1.05086999603862e+00, 1.05086836702556e+00, 1.05086812658673e+00, 1.05086810846044e+00, 1.05086778839853e+00, 1.05086561993344e+00, 1.05085764844738e+00, 1.05083639955356e+00, 1.05078989768749e+00, 1.05070084418674e+00, 1.05054598921480e+00, 1.05029572875800e+00, 1.04991395421675e+00, 1.04935817754612e+00, 1.04857994911510e+00, 1.04752557806620e+00, 1.04613715564642e+00, 1.04435387058334e+00, 1.04211359217514e+00, 1.03935468175693e+00, 1.03601797737303e+00, 1.03204888096860e+00, 1.02739946366489e+00, 1.02203049432616e+00, 1.01591329127733e+00, 1.00903129805729e+00, 1.00138129237308e+00, 9.92974153133473e-01, 9.83835132932152e-01, 9.74003611106006e-01, 9.63532333258720e-01, 9.52486174156519e-01, 9.40940489262932e-01, 9.28979143206694e-01, 9.16692319099461e-01, 9.04174219643483e-01, 8.91520769216885e-01, 8.78827416408186e-01, 8.66187120419585e-01, 8.53688584525050e-01, 8.41414777700610e-01, 8.29441763855296e-01, 8.17837838609188e-01, 8.06662957572989e-01, 7.95968428268424e-01, 7.85796830325423e-01, 7.76182125089067e-01, 7.67149915651209e-01, 7.58717820811793e-01, 7.50895930759015e-01, 7.43687317576711e-01, 7.37088579393346e-01, 7.31090402568067e-01, 7.25678131390447e-01, 7.20832339100236e-01, 7.16529397462198e-01, 7.12742044589763e-01, 7.09439952191304e-01, 7.06590293953085e-01, 7.04158316449771e-01, 7.02107912896485e-01, 7.00402198367176e-01, 6.99004082973747e-01, 6.97876837127628e-01, 6.96984640610068e-01, 6.96293104990094e-01, 6.95769757176966e-01, 6.95384470782709e-01, 6.95109831664106e-01, 6.94921424617358e-01, 6.94798029743625e-01, 6.94721719439734e-01, 6.94677850166368e-01, 6.94654946908611e-01, 6.94644482325200e-01, 6.94640556715545e-01, 6.94639488856129e-01, 6.94639331242374e-01, 6.94639319295470e-01, 6.94639107487295e-01, 6.94637672456707e-01, 6.94632397146839e-01, 6.94618335217936e-01, 6.94587561590535e-01, 6.94528628558569e-01, 6.94426150236277e-01, 6.94260536048245e-01, 6.94007891530892e-01, 6.93640101696269e-01, 6.93125108366237e-01, 6.92427387953446e-01, 6.91508629923047e-01, 6.90328608500376e-01, 6.88846231151981e-01, 6.87020737246075e-01, 6.84813009641895e-01, 6.82186951574476e-01, 6.79110872115167e-01, 6.75558816836722e-01, 6.71511777205069e-01, 6.66958713564046e-01, 6.61897332901193e-01, 6.56334573881515e-01, 6.50286767308046e-01, 6.43779459017168e-01, 6.36846902576582e-01, 6.29531249094565e-01, 6.21881479045953e-01, 6.13952134627369e-01, 6.05801919635544e-01, 5.97492236731259e-01, 5.89085729397499e-01, 5.80644888706698e-01, 5.72230774384130e-01, 5.63901887008281e-01, 5.55713214936157e-01, 5.47715466913623e-01, 5.39954490264588e-01, 5.32470865638770e-01, 5.25299662792656e-01, 5.18470337755929e-01, 5.12006749764808e-01, 5.05927276165790e-01, 5.00245004694142e-01, 4.94967984694065e-01, 4.90099521589516e-01, 4.85638501908101e-01, 4.81579739140922e-01, 4.77914333486063e-01, 4.74630040925790e-01, 4.71711649028141e-01, 4.69141358281925e-01, 4.66899168643352e-01, 4.64963271294396e-01, 4.63310445416973e-01, 4.61916459128221e-01, 4.60756472681282e-01, 4.59805440717891e-01, 4.59038508889670e-01, 4.58431398686590e-01, 4.57960772974226e-01, 4.57604573694305e-01, 4.57342322559688e-01, 4.57155375481061e-01, 4.57027121965090e-01, 4.56943121840162e-01, 4.56891173361175e-01, 4.56861308932538e-01, 4.56845717238025e-01, 4.56838593316229e-01, 4.56835920895664e-01, 4.56835193933273e-01, 4.56835086635175e-01, 4.56835078450520e-01, 4.56834932662903e-01, 4.56833944930908e-01, 4.56830313931887e-01, 4.56820635085197e-01, 4.56799453488560e-01, 4.56758889433147e-01, 4.56688352068259e-01, 4.56574355316623e-01, 4.56400448402430e-01, 4.56147271226739e-01, 4.55792742152606e-01, 4.55312382379256e-01, 4.54679776852663e-01, 4.53867166509850e-01, 4.52846160628753e-01, 4.51588551344387e-01, 4.50067205354147e-01, 4.48257000996144e-01, 4.46135772906244e-01, 4.43685222075551e-01, 4.40891747053477e-01, 4.37747152851165e-01, 4.34249198131100e-01, 4.30401948518261e-01, 4.26215913951203e-01, 4.21707960150740e-01, 4.16900997469089e-01, 4.11823463380532e-01, 4.06508626477550e-01, 4.00993749005795e-01, 3.95319150979196e-01, 3.89527221425962e-01, 3.83661421399061e-01, 3.77765319470452e-01, 3.71881694208749e-01, 3.66051730445688e-01, 3.60314327817294e-01, 3.54705531886884e-01, 3.49258090731277e-01, 3.44001133621940e-01, 3.38959963589265e-01, 3.34155952280125e-01, 3.29606523531510e-01, 3.25325211320363e-01, 3.21321777992703e-01, 3.17602379683988e-01, 3.14169767379251e-01, 3.11023513905908e-01, 3.08160259112387e-01, 3.05573967401555e-01, 3.03256193533186e-01, 3.01196354087788e-01, 2.99382003128152e-01, 2.97799111363563e-01, 2.96432348498873e-01, 2.95265368444686e-01, 2.94281096707545e-01, 2.93462018624295e-01, 2.92790466227692e-01, 2.92248900522910e-01, 2.91820184922437e-01, 2.91487844641278e-01, 2.91236306105612e-01, 2.91051109974322e-01, 2.90919091292827e-01, 2.90828520640694e-01, 2.90769200912401e-01, 2.90732515559751e-01, 2.90711425662859e-01, 2.90700414992104e-01, 2.90695384161329e-01, 2.90693496928788e-01, 2.90692983556320e-01, 2.90692907783639e-01, 2.90692901962943e-01, 2.90692797748187e-01, 2.90692091678293e-01, 2.90689496096706e-01, 2.90682577275504e-01, 2.90667435839670e-01, 2.90638439079678e-01, 2.90588016290183e-01, 2.90506527226448e-01, 2.90382212532335e-01, 2.90201233479408e-01, 2.89947806453908e-01, 2.89604435219387e-01, 2.89152240969133e-01, 2.88571386515329e-01, 2.87841586670653e-01, 2.86942692095964e-01, 2.85855328866260e-01, 2.84561571114232e-01, 2.83045619810236e-01, 2.81294457543982e-01, 2.79298447589671e-01, 2.77051845976676e-01, 2.74553197999966e-01, 2.71805595607446e-01, 2.68816779149923e-01, 2.65599075570341e-01, 2.62169174527573e-01, 2.58547753357854e-01, 2.54758970320187e-01, 2.50829852496082e-01, 2.46789609462255e-01, 2.42668906130610e-01, 2.38499127936912e-01, 2.34311669097316e-01, 2.30137270378147e-01, 2.26005427299389e-01, 2.21943883520471e-01, 2.17978217914396e-01, 2.14131528014882e-01, 2.10424207489793e-01, 2.06873811284449e-01, 2.03494999190777e-01, 2.00299546823447e-01, 1.97296412229636e-01, 1.94491846474842e-01, 1.91889537347695e-01, 1.89490776612579e-01, 1.87294642815450e-01, 1.85298193342481e-01, 1.83496661098800e-01, 1.81883652703082e-01, 1.80451346398930e-01, 1.79190688907873e-01, 1.78091591156917e-01, 1.77143123194659e-01, 1.76333708672987e-01, 1.75651319045271e-01, 1.75083667160920e-01, 1.74618399280486e-01, 1.74243283765668e-01, 1.73946393895514e-01, 1.73716281508026e-01, 1.73542137548877e-01, 1.73413935200996e-01, 1.73322551130547e-01, 1.73259860554605e-01, 1.73218802325865e-01, 1.73193411023871e-01, 1.73178814096697e-01, 1.73171193344801e-01, 1.73167711396756e-01, 1.73166405203729e-01, 1.73166049888111e-01, 1.73165997444291e-01, 1.73165993384923e-01, 1.73165920305065e-01, 1.73165425178646e-01, 1.73163605047262e-01, 1.73158753289048e-01, 1.73148135557089e-01, 1.73127802181717e-01, 1.73092444935996e-01, 1.73035305380977e-01, 1.72948141193544e-01, 1.72821255865053e-01, 1.72643595852291e-01, 1.72402917567259e-01, 1.72086024427995e-01, 1.71679071515471e-01, 1.71167932209835e-01, 1.70538617611331e-01, 1.69777735777535e-01, 1.68872974111560e-01, 1.67813584972590e-01, 1.66590852149937e-01, 1.65198514636343e-01, 1.63633124481597e-01, 1.61894317602257e-01, 1.59984980288646e-01, 1.57911299600995e-01, 1.55682692490726e-01, 1.53311615759621e-01, 1.50813266214382e-01, 1.48205186907226e-01, 1.45506800570149e-01, 1.42738894801726e-01, 1.39923085008972e-01, 1.37081280532108e-01, 1.34235176994780e-01, 1.31405794112593e-01, 1.28613073455488e-01, 1.25875545525287e-01, 1.23210070470183e-01, 1.20631652213484e-01, 1.18153322007986e-01, 1.15786084601858e-01, 1.13538918364524e-01, 1.11418819819849e-01, 1.09430882941405e-01, 1.07578404107002e-01, 1.05863004601630e-01, 1.04284763826183e-01, 1.02842357765835e-01, 1.01533198674975e-01, 1.00353573248890e-01, 9.92987777038197e-02, 9.83632491291937e-02, 9.75406931832792e-02, 9.68242086691058e-02, 9.62064097566335e-02, 9.56795466243978e-02, 9.52356251020472e-02, 9.48665255358003e-02, 9.45641206119587e-02, 9.43203913075967e-02, 9.41275395467009e-02, 9.39780955820133e-02, 9.38650176543030e-02, 9.37817811536741e-02, 9.37224543634354e-02, 9.36817579338411e-02, 9.36551055202388e-02, 9.36386235184186e-02, 9.36291485101162e-02, 9.36242018486881e-02, 9.36219417117448e-02, 9.36210938617714e-02, 9.36208632265334e-02, 9.36208291852747e-02, 9.36208265289791e-02, 9.36207784323368e-02, 9.36204525708293e-02, 9.36192546744685e-02, 9.36160615609944e-02, 9.36090737052071e-02, 9.35956918724186e-02, 9.35724230903707e-02, 9.35348208275730e-02, 9.34774637555771e-02, 9.33939768433580e-02, 9.32770976912893e-02, 9.31187898849641e-02, 9.29104036698858e-02, 9.26428823806287e-02, 9.23070108127035e-02, 9.18936991739638e-02, 9.13942935383840e-02, 9.08009010629653e-02, 9.01067158907027e-02, 8.93063299535299e-02, 8.83960121048695e-02, 8.73739393926503e-02, 8.62403659680542e-02, 8.49977181086814e-02, 8.36506079522996e-02, 8.22057634675648e-02, 8.06718774853508e-02, 7.90593837621856e-02, 7.73801725308099e-02, 7.56472613732546e-02, 7.38744392347126e-02, 7.20759018723651e-02, 7.02658960797304e-02, 6.84583878748306e-02, 6.66667668169820e-02, 6.49035950932007e-02, 6.31804063614529e-02, 6.15075558915947e-02, 5.98941205727748e-02, 5.83478450243300e-02, 5.68751284089002e-02, 5.54810455652035e-02, 5.41693956706615e-02, 5.29427717183992e-02, 5.18026445549369e-02, 5.07494559720767e-02, 4.97827162647946e-02, 4.89011026472133e-02, 4.81025558772599e-02, 4.73843733273700e-02, 4.67432975274313e-02, 4.61755998764346e-02, 4.56771597494958e-02, 4.52435395969291e-02, 4.48700568338527e-02, 4.45518533576177e-02, 4.42839634249879e-02, 4.40613803952372e-02, 4.38791225274731e-02, 4.37322976407089e-02, 4.36161660381842e-02, 4.35262007007289e-02, 4.34581434067276e-02, 4.34080551754104e-02, 4.33723592858337e-02, 4.33478751171261e-02, 4.33318411944184e-02, 4.33219261044773e-02, 4.33162263474607e-02, 4.33132506868583e-02, 4.33118911117720e-02, 4.33113810928827e-02, 4.33112423558164e-02, 4.33112218785325e-02, 4.33112202670634e-02, 4.33111909144905e-02, 4.33109920467709e-02, 4.33102609921496e-02, 4.33083123008471e-02, 4.33040477994995e-02, 4.32958813896471e-02, 4.32816818757198e-02, 4.32587368979931e-02, 4.32237407873375e-02, 4.31728088094490e-02, 4.31015197734563e-02, 4.30049882866816e-02, 4.28779669949311e-02, 4.27149779274822e-02, 4.25104705723438e-02, 4.22590025893577e-02, 4.19554372261839e-02, 4.15951496859117e-02, 4.11742330980278e-02, 4.06896935813528e-02, 4.01396233671800e-02, 3.95233412402591e-02, 3.88414907467736e-02, 3.80960887014634e-02, 3.72905193782938e-02, 3.64294731610215e-02, 3.55188320533451e-02, 3.45655079513687e-02, 3.35772426155305e-02, 3.25623805444222e-02, 3.15296272412349e-02, 3.04878055917097e-02, 2.94456222965316e-02, 2.84114546957657e-02, 2.73931661345124e-02, 2.63979555061091e-02, 2.54322440019621e-02, 2.45015995921838e-02, 2.36106975327148e-02, 2.27633134002101e-02, 2.19623439060790e-02, 2.12098500635287e-02, 2.05071171083357e-02, 1.98547257790089e-02, 1.92526300227431e-02, 1.87002368315632e-02, 1.81964846856557e-02, 1.77399179455593e-02, 1.73287554191432e-02, 1.69609521486560e-02, 1.66342541481868e-02, 1.63462463457718e-02, 1.60943943592624e-02, 1.58760809837268e-02, 1.56886384114889e-02, 1.55293772447607e-02, 1.53956133019451e-02, 1.52846930645438e-02, 1.51940183845936e-02, 1.51210707937537e-02, 1.50634354574404e-02, 1.50188245216816e-02, 1.49850993374709e-02, 1.49602908327271e-02, 1.49426171583338e-02, 1.49304976715889e-02, 1.49225623475884e-02, 1.49176558213684e-02, 1.49148354567199e-02, 1.49133630884508e-02, 1.49126903777521e-02, 1.49124380250516e-02, 1.49123693794024e-02, 1.49123592474538e-02, 1.49123584430511e-02, 1.49123437014094e-02, 1.49122438248561e-02, 1.49118766716834e-02, 1.49108980038166e-02, 1.49087563537158e-02, 1.49046553761885e-02, 1.48975254273029e-02, 1.48860060588891e-02, 1.48684410431326e-02, 1.48428874127902e-02, 1.48071397806673e-02, 1.47587708406505e-02, 1.46951884173321e-02, 1.46137086943643e-02, 1.45116443174560e-02, 1.43864049656033e-02, 1.42356067841583e-02, 1.40571858820637e-02, 1.38495100453273e-02, 1.36114820663478e-02, 1.33426277789459e-02, 1.30431621524529e-02, 1.27140277058277e-02, 1.23569010531144e-02, 1.19741654848076e-02, 1.15688499326421e-02, 1.11445371894776e-02, 1.07052465880706e-02, 1.02552982260962e-02, 9.79916708577466e-03, 9.34133591519437e-03, 8.88615545831506e-03, 8.43771957858230e-03, 7.99976115386958e-03, 7.57557260185227e-03, 7.16795279282767e-03, 6.77918013137191e-03, 6.41100983176024e-03, 6.06469196438741e-03, 5.74100587205084e-03, 5.44030623849157e-03, 5.16257641336448e-03, 4.90748527324225e-03, 4.67444453492760e-03, 4.46266384610094e-03, 4.27120132519221e-03, 4.09900782745520e-03, 3.94496409059197e-03, 3.80791087731946e-03, 3.68667287763724e-03, 3.58007732293508e-03, 3.48696812740901e-03, 3.40621618919101e-03, 3.33672643793381e-03, 3.27744230765337e-03, 3.22734839864314e-03, 3.18547211744313e-03, 3.15088497031901e-03, 3.12270399094317e-03, 3.10009360380459e-03, 3.08226804674927e-03, 3.06849426727851e-03, 3.05809520103306e-03, 3.05045311660468e-03, 3.04501277912106e-03, 3.04128402448534e-03, 3.03884342630845e-03, 3.03733469806655e-03, 3.03646756731446e-03, 3.03601491525018e-03, 3.03580810931248e-03, 3.03573053328776e-03, 3.03570943056375e-03, 3.03570631581973e-03, 3.03570606606485e-03, 3.03570146413179e-03, 3.03567028447382e-03, 3.03555566409663e-03, 3.03525014767381e-03, 3.03458161352577e-03, 3.03330160066917e-03, 3.03107662530344e-03, 3.02748309073131e-03, 3.02200644633022e-03, 3.01404519248738e-03, 3.00292027023996e-03, 2.98789029482993e-03, 2.96817278458139e-03, 2.94297127939114e-03, 2.91044804290053e-03, 2.87203778168266e-03, 2.82602119203812e-03, 2.77190475004652e-03, 2.70936727334981e-03, 2.63829298795049e-03, 2.55879667785397e-03, 2.47123806633631e-03, 2.37622317355868e-03, 2.27459146956299e-03, 2.16738909897263e-03, 2.05582998498333e-03, 1.94124779085580e-03, 1.82504222160595e-03, 1.70862307191300e-03, 1.59335523752713e-03, 1.48050804884292e-03, 1.37121263274498e-03, 1.26643078426447e-03, 1.16693728512551e-03, 1.07331499385044e-03, 9.85959798030400e-04, 9.05092174133593e-04, 8.30773648842431e-04, 7.62928092876258e-04, 7.01367541792841e-04, 6.45820134132613e-04, 5.95956002658049e-04, 5.51407770218226e-04, 5.11785463539915e-04, 4.76688514500445e-04, 4.45717600568329e-04, 4.18486563569687e-04, 3.94632104926629e-04, 3.73818749919153e-04, 3.55738714335807e-04, 3.40108766817337e-04, 3.26666987191082e-04, 3.15171186865193e-04, 3.05398917367583e-04, 2.97147855268243e-04, 2.90235408287209e-04, 2.84497172548193e-04, 2.79784628399504e-04, 2.75962754909820e-04, 2.72908094689051e-04, 2.70507474370518e-04, 2.68657317884724e-04, 2.67263369624793e-04, 2.66240643297166e-04, 2.65513461259914e-04, 2.65015496745846e-04, 2.64689759166714e-04, 2.64488474065239e-04, 2.64372814402059e-04, 2.64312446868864e-04, 2.64284868385367e-04, 2.64274523382291e-04, 2.64271709349728e-04, 2.64271294006413e-04, 2.64271260442150e-04, 2.64270637943653e-04, 2.64266420439974e-04, 2.64250916854891e-04, 2.64209593108689e-04, 2.64119172919131e-04, 2.63946068169929e-04, 2.63645228216980e-04, 2.63159505512256e-04, 2.62419644939449e-04, 2.61345015156955e-04, 2.59845209731324e-04, 2.57822647711237e-04, 2.55176283826763e-04, 2.51806472883587e-04, 2.47620888670141e-04, 2.42541152852382e-04, 2.36509509243810e-04, 2.29494597007829e-04, 2.21495338609934e-04, 2.12542374005587e-04, 2.02697357766901e-04, 1.92051402986948e-04, 1.80724226076734e-04, 1.68864406754565e-04, 1.56648793494085e-04, 1.44277042867458e-04, 1.32009367135311e-04, 1.19889045329558e-04, 1.08236560074637e-04, 9.71674775946951e-05, 8.66960780383508e-05, 7.68922072857947e-05, 6.79623324604250e-05, 5.98265829934164e-05, 5.24951940343854e-05, 4.59172232840643e-05, 4.00039916588213e-05, 3.48407144885599e-05, 3.02259607227739e-05, 2.61589957419872e-05, 2.25872583676789e-05, 1.93855653707331e-05, 1.63867041758081e-05, 1.34593095530988e-05, 1.05180407050653e-05, 7.63061718433988e-06, 4.96673017646998e-06, 2.72900731435044e-06, 1.19827472477851e-06, 9.51153298345275e-07, 7.31368925421805e-07, 3.47372995257444e-07, 5.67280273422455e-07, 6.74019853263985e-07, 6.36943179976868e-07, 5.43656779150581e-07, 4.57141201337663e-07, 4.05058743879749e-07, 3.83043112783810e-07, 3.71305519664133e-07, 3.52996076813383e-07, 3.21565053695849e-07, 2.78513405227443e-07, -5.32907051820075e-15, -3.55271367880050e-15, -5.32907051820075e-15, -5.32907051820075e-15, -5.32907051820075e-15, -5.32907051820075e-15, -5.32907051820075e-15, -3.55271367880050e-15, -3.55271367880050e-15, -5.32907051820075e-15, -3.55271367880050e-15],
[ 9.00000000000000e+00, 8.99999999638873e+00, 8.99999988472125e+00, 8.99999912815649e+00, 8.99999634692794e+00, 8.99998893306711e+00, 8.99997270737137e+00, 8.99994163076019e+00, 8.99988758056493e+00, 8.99980020038123e+00, 8.99966682993652e+00, 8.99947251905669e+00, 8.99920012732913e+00, 8.99883050853317e+00, 8.99834277642386e+00, 8.99771464608662e+00, 8.99692284290019e+00, 8.99594356921281e+00, 8.99475301720195e+00, 8.99332791508370e+00, 8.99164609288703e+00, 8.98968705341482e+00, 8.98743253377533e+00, 8.98486704296958e+00, 8.98197836144363e+00, 8.97875798923907e+00, 8.97520153037990e+00, 8.97130900240247e+00, 8.96708506145256e+00, 8.96253913512858e+00, 8.95768545722809e+00, 8.95254300074050e+00, 8.94713530779485e+00, 8.94149021778073e+00, 8.93563949746123e+00, 8.92961837952022e+00, 8.92346501855377e+00, 8.91721987593420e+00, 8.91092504715154e+00, 8.90462354707674e+00, 8.89835857001096e+00, 8.89217274231584e+00, 8.88610738581444e+00, 8.88020180998646e+00, 8.87449265025691e+00, 8.86901326842042e+00, 8.86379322950173e+00, 8.85885786719270e+00, 8.85422794750744e+00, 8.84991943754709e+00, 8.84594338335669e+00, 8.84230589788083e+00, 8.83900825707040e+00, 8.83604709934360e+00, 8.83341472093668e+00, 8.83109945726154e+00, 8.82908613827770e+00, 8.82735660413538e+00, 8.82589026599431e+00, 8.82466469599899e+00, 8.82365622991456e+00, 8.82284056590463e+00, 8.82219334335959e+00, 8.82169068654392e+00, 8.82130969909634e+00, 8.82102889704828e+00, 8.82082856997629e+00, 8.82069106211582e+00, 8.82060096767519e+00, 8.82054523713238e+00, 8.82051319390432e+00, 8.82049646337803e+00, 8.82048881881726e+00, 8.82048595103894e+00, 8.82048517092996e+00, 8.82048505578737e+00, 8.82048504068309e+00, 8.82048468877144e+00, 8.82048230451470e+00, 8.82047353968172e+00, 8.82045017544813e+00, 8.82039904132958e+00, 8.82030110546279e+00, 8.82013076940615e+00, 8.81985539456061e+00, 8.81943508042414e+00, 8.81882270741471e+00, 8.81796424917988e+00, 8.81679935142624e+00, 8.81526216662042e+00, 8.81328242669256e+00, 8.81078672933672e+00, 8.80770000782200e+00, 8.80394714950771e+00, 8.79945472452717e+00, 8.79415278332572e+00, 8.78797667981256e+00, 8.78086887568585e+00, 8.77278068089648e+00, 8.76367388515512e+00, 8.75352223586850e+00, 8.74231271902363e+00, 8.73004660153421e+00, 8.71674019670515e+00, 8.70242531906388e+00, 8.68714940110816e+00, 8.67097525266615e+00, 8.65398045350439e+00, 8.63625638128207e+00, 8.61790688943227e+00, 8.59904666236265e+00, 8.57979928769428e+00, 8.56029509625857e+00, 8.54066882948376e+00, 8.52105720002577e+00, 8.50159641466964e+00, 8.48241972854867e+00, 8.46365509676282e+00, 8.44542298390989e+00, 8.42783438441934e+00, 8.41098909752689e+00, 8.39497429088872e+00, 8.37986337678990e+00, 8.36571521513425e+00, 8.35257364826674e+00, 8.34046736439299e+00, 8.32941007902357e+00, 8.31940101748139e+00, 8.31042567600404e+00, 8.30245683424694e+00, 8.29545578794474e+00, 8.28937376703161e+00, 8.28415350161101e+00, 8.27973089579771e+00, 8.27603676767395e+00, 8.27299861248532e+00, 8.27054234585553e+00, 8.26859398433123e+00, 8.26708122207986e+00, 8.26593486512655e+00, 8.26509008815960e+00, 8.26448748363747e+00, 8.26407387861405e+00, 8.26380290123043e+00, 8.26363528601540e+00, 8.26353891477154e+00, 8.26348859764570e+00, 8.26346560672988e+00, 8.26345698194567e+00, 8.26345463578600e+00, 8.26345428949741e+00, 8.26345425892671e+00, 8.26345365990896e+00, 8.26344960147124e+00, 8.26343468216051e+00, 8.26339491226118e+00, 8.26330787429976e+00, 8.26314117657507e+00, 8.26285125774158e+00, 8.26238259103857e+00, 8.26166732600544e+00, 8.26062539367145e+00, 8.25916508848503e+00, 8.25718412690713e+00, 8.25457116891126e+00, 8.25120777491767e+00, 8.24697075726743e+00, 8.24173487255335e+00, 8.23537578930039e+00, 8.22777325494210e+00, 8.21881437705672e+00, 8.20839692666980e+00, 8.19643256635728e+00, 8.18284990316883e+00, 8.16759726635711e+00, 8.15064511291398e+00, 8.13198797037909e+00, 8.11164583666988e+00, 8.08966497102584e+00, 8.06611802855983e+00, 8.04110351300002e+00, 8.01474454718097e+00, 7.98718698744644e+00, 7.95859693472386e+00, 7.92915771977109e+00, 7.89906646113091e+00, 7.86853031005892e+00, 7.83776250598998e+00, 7.80697836847310e+00, 7.77639134711585e+00, 7.74620924075026e+00, 7.71663068207102e+00, 7.68784196600606e+00, 7.66001428073431e+00, 7.63330138111003e+00, 7.60783772654494e+00, 7.58373709001573e+00, 7.56109163227645e+00, 7.53997142565627e+00, 7.52042440480171e+00, 7.50247671695115e+00, 7.48613344126280e+00, 7.47137964477909e+00, 7.45818174126193e+00, 7.44648911792869e+00, 7.43623599373034e+00, 7.42734347105163e+00, 7.41972174052506e+00, 7.41327239610846e+00, 7.40789081486522e+00, 7.40346855327627e+00, 7.39989570973372e+00, 7.39706320147304e+00, 7.39486490394395e+00, 7.39319960180658e+00, 7.39197270360417e+00, 7.39109767683945e+00, 7.39049716668344e+00, 7.39010376975724e+00, 7.38986044410869e+00, 7.38972054730457e+00, 7.38964750603588e+00, 7.38961413229941e+00, 7.38960161256522e+00, 7.38959820688515e+00, 7.38959770421358e+00, 7.38959766389310e+00, 7.38959691977569e+00, 7.38959187826765e+00, 7.38957334510477e+00, 7.38952394214349e+00, 7.38941582317731e+00, 7.38920875554900e+00, 7.38884864169651e+00, 7.38826654422312e+00, 7.38737826597568e+00, 7.38608452297437e+00, 7.38427173285498e+00, 7.38181342475394e+00, 7.37857225834537e+00, 7.37440262020333e+00, 7.36915374515465e+00, 7.36267328935642e+00, 7.35481126124233e+00, 7.34542419720698e+00, 7.33437945207595e+00, 7.32155946128230e+00, 7.30686582348826e+00, 7.29022305030192e+00, 7.27158183466604e+00, 7.25092170200137e+00, 7.22825292834777e+00, 7.20361763704926e+00, 7.17709001880883e+00, 7.14877565738723e+00, 7.11880998247753e+00, 7.08735590963622e+00, 7.05460076178633e+00, 7.02075259516417e+00, 6.98603607265657e+00, 6.95068803809824e+00, 6.91495294607102e+00, 6.87907829385085e+00, 6.84331018699247e+00, 6.80788914979556e+00, 6.77304626895401e+00, 6.73899973533966e+00, 6.70595182702133e+00, 6.67408635762201e+00, 6.64356659870902e+00, 6.61453367326061e+00, 6.58710540906535e+00, 6.56137563559431e+00, 6.53741390469987e+00, 6.51526561368079e+00, 6.49495250813308e+00, 6.47647354103177e+00, 6.45980606326130e+00, 6.44490731907848e+00, 6.43171621762388e+00, 6.42015534857560e+00, 6.41013320642964e+00, 6.40154658383703e+00, 6.39428309013645e+00, 6.38822374695721e+00, 6.38324560884535e+00, 6.37922435363110e+00, 6.37603678507377e+00, 6.37356318954160e+00, 6.37168948942400e+00, 6.37030913887349e+00, 6.36932471247660e+00, 6.36864914457363e+00, 6.36820658606191e+00, 6.36793285636168e+00, 6.36777548040044e+00, 6.36769331348085e+00, 6.36765577016922e+00, 6.36764168628508e+00, 6.36763785511844e+00, 6.36763728964596e+00, 6.36763724603128e+00, 6.36763646285154e+00, 6.36763115669065e+00, 6.36761165063860e+00, 6.36755965433253e+00, 6.36744585999895e+00, 6.36722792376217e+00, 6.36684891055421e+00, 6.36623627003044e+00, 6.36530139832038e+00, 6.36393982620680e+00, 6.36203205834966e+00, 6.35944507037922e+00, 6.35603445106512e+00, 6.35164715546669e+00, 6.34612481234986e+00, 6.33930750587205e+00, 6.33103792856113e+00, 6.32116578123341e+00, 6.30955227723681e+00, 6.29607459492501e+00, 6.28063011516977e+00, 6.26314028134296e+00, 6.24355392840597e+00, 6.22184994570704e+00, 6.19803916417554e+00, 6.17216539134402e+00, 6.14430555480841e+00, 6.11456895362112e+00, 6.08309565476589e+00, 6.05005410553451e+00, 6.01563806009753e+00, 5.98006293843288e+00, 5.94356174758999e+00, 5.90638069947393e+00, 5.86877465709499e+00, 5.83100253411222e+00, 5.79332276214401e+00, 5.75598892814232e+00, 5.71924567114604e+00, 5.68332491451351e+00, 5.64844249648869e+00, 5.61479524868244e+00, 5.58255855872926e+00, 5.55188444010934e+00, 5.52290011918333e+00, 5.49570713730636e+00, 5.47038095496326e+00, 5.44697103563446e+00, 5.42550137982975e+00, 5.40597147444678e+00, 5.38835761911560e+00, 5.37261458907412e+00, 5.35867759286513e+00, 5.34646448222121e+00, 5.33587817044999e+00, 5.32680921413626e+00, 5.31913851090765e+00, 5.31274006343102e+00, 5.30748375695771e+00, 5.30323809499986e+00, 5.29987283556741e+00, 5.29726146933084e+00, 5.29528348157179e+00, 5.29382634223412e+00, 5.29278717305106e+00, 5.29207404769635e+00, 5.29160689010239e+00, 5.29131794724302e+00, 5.29115182536600e+00, 5.29106509233104e+00, 5.29102546273293e+00, 5.29101059621223e+00, 5.29100655214961e+00, 5.29100595525414e+00, 5.29100591005901e+00, 5.29100510943179e+00, 5.29099968506172e+00, 5.29097974446134e+00, 5.29092658981757e+00, 5.29081026053393e+00, 5.29058746973235e+00, 5.29020001492405e+00, 5.28957373196397e+00, 5.28861804864843e+00, 5.28722617988127e+00, 5.28527599014952e+00, 5.28263153105001e+00, 5.27914524180264e+00, 5.27466077923252e+00, 5.26901642097349e+00, 5.26204896227406e+00, 5.25359800368386e+00, 5.24351050523611e+00, 5.23164546388031e+00, 5.21787855631180e+00, 5.20210658038900e+00, 5.18425152621934e+00, 5.16426411357303e+00, 5.14212664592635e+00, 5.11785505299340e+00, 5.09150002240337e+00, 5.06314715607093e+00, 5.03291612625166e+00, 5.00095884841916e+00, 4.96745673082732e+00, 4.93261710157343e+00, 4.89666895058859e+00, 4.85985815357217e+00, 4.82244236488970e+00, 4.78468577477764e+00, 4.74685392169816e+00, 4.70920873358617e+00, 4.67200394385092e+00, 4.63548099263882e+00, 4.59986548531401e+00, 4.56536424285587e+00, 4.53216294675926e+00, 4.50042435660816e+00, 4.47028706273758e+00, 4.44186472876627e+00, 4.41524577765416e+00, 4.39049347816036e+00, 4.36764639398721e+00, 4.34671916368425e+00, 4.32770358428194e+00, 4.31056997490298e+00, 4.29526879798645e+00, 4.28173251529919e+00, 4.26987765383128e+00, 4.25960705329903e+00, 4.25081226266660e+00, 4.24337604821606e+00, 4.23717497062245e+00, 4.23208198361523e+00, 4.22796900253693e+00, 4.22470938785426e+00, 4.22218028684053e+00, 4.22026477659561e+00, 4.21885375359698e+00, 4.21784751927294e+00, 4.21715701771224e+00, 4.21670469047382e+00, 4.21642492427242e+00, 4.21626407967667e+00, 4.21618010232615e+00, 4.21614173192987e+00, 4.21612733779652e+00, 4.21612342223635e+00, 4.21612284430762e+00, 4.21612280092233e+00, 4.21612203729191e+00, 4.21611686358132e+00, 4.21609784444141e+00, 4.21604714614767e+00, 4.21593619290509e+00, 4.21572369931744e+00, 4.21535415599130e+00, 4.21475683492387e+00, 4.21384536942071e+00, 4.21251794971614e+00, 4.21065816006289e+00, 4.20813646565072e+00, 4.20481233821191e+00, 4.20053698765315e+00, 4.19515664385650e+00, 4.18851630859903e+00, 4.18046387341271e+00, 4.17085447658511e+00, 4.15955495313395e+00, 4.14644821738497e+00, 4.13143741060831e+00, 4.11444964757886e+00, 4.09543920688814e+00, 4.07439003047233e+00, 4.05131742724260e+00, 4.02626891197460e+00, 3.99932415090187e+00, 3.97059402637434e+00, 3.94021887102260e+00, 3.90836595412158e+00, 3.87522632722420e+00, 3.84101115181741e+00, 3.80594763914595e+00, 3.77027473281442e+00, 3.73423866016338e+00, 3.69808847054959e+00, 3.66207166895562e+00, 3.62643004262244e+00, 3.59139576689776e+00, 3.55718786413862e+00, 3.52400907615121e+00, 3.49204319631091e+00, 3.46145289248923e+00, 3.43237803681242e+00, 3.40493454385869e+00, 3.37921370595653e+00, 3.35528200342976e+00, 3.33318135934854e+00, 3.31292980268633e+00, 3.29452250054546e+00, 3.27793311885141e+00, 3.26311547102006e+00, 3.25000541490819e+00, 3.23852295923707e+00, 3.22857454111601e+00, 3.22005543594594e+00, 3.21285225970982e+00, 3.20684552152349e+00, 3.20191218158253e+00, 3.19792816670560e+00, 3.19477079305263e+00, 3.19232104383703e+00, 3.19046564948383e+00, 3.18909891915760e+00, 3.18812427620773e+00, 3.18745545599827e+00, 3.18701733275745e+00, 3.18674635226206e+00, 3.18659055895550e+00, 3.18650921894392e+00, 3.18647205359489e+00, 3.18645811152048e+00, 3.18645431893169e+00, 3.18645375915329e+00, 3.18645371725833e+00, 3.18645298156183e+00, 3.18644799710770e+00, 3.18642967369647e+00, 3.18638082997458e+00, 3.18627393549151e+00, 3.18606921519954e+00, 3.18571319059802e+00, 3.18513772181474e+00, 3.18425960350432e+00, 3.18298075445611e+00, 3.18118902611175e+00, 3.17875963866710e+00, 3.17555723502410e+00, 3.17143852266792e+00, 3.16625545190254e+00, 3.15985885638714e+00, 3.15210245947871e+00, 3.14284712867874e+00, 3.13196524189709e+00, 3.11934501477761e+00, 3.10489462943685e+00, 3.08854600290461e+00, 3.07025803921828e+00, 3.05001922291251e+00, 3.02784943338369e+00, 3.00380088851417e+00, 2.97795816067180e+00, 2.95043724694148e+00, 2.92138371604354e+00, 2.89096999449757e+00, 2.85939189181157e+00, 2.82686449654155e+00, 2.79361759998050e+00, 2.75989082043819e+00, 2.72592860759311e+00, 2.69197530297170e+00, 2.65827041974779e+00, 2.62504428402205e+00, 2.59251415244879e+00, 2.56088088988489e+00, 2.53032625819405e+00, 2.50101083591553e+00, 2.47307256033068e+00, 2.44662586013554e+00, 2.42176132942610e+00, 2.39854588234383e+00, 2.37702332224640e+00, 2.35721525891956e+00, 2.33912231106729e+00, 2.32272553786982e+00, 2.30798805151740e+00, 2.29485677112737e+00, 2.28326428632391e+00, 2.27313080521780e+00, 2.26436616604233e+00, 2.25687189401927e+00, 2.25054328513992e+00, 2.24527149668185e+00, 2.24094562086654e+00, 2.23745471368213e+00, 2.23468974622704e+00, 2.23254544169625e+00, 2.23092195802589e+00, 2.22972637483533e+00, 2.22887394411675e+00, 2.22828906737988e+00, 2.22790596770946e+00, 2.22766903325090e+00, 2.22753281861774e+00, 2.22746170205789e+00, 2.22742920825827e+00, 2.22741701870316e+00, 2.22741370284776e+00, 2.22741321343442e+00, 2.22741317679202e+00, 2.22741253314990e+00, 2.22740817238011e+00, 2.22739214173305e+00, 2.22734940994432e+00, 2.22725589262559e+00, 2.22707679636992e+00, 2.22676534835111e+00, 2.22626197059531e+00, 2.22549394926334e+00, 2.22437563893291e+00, 2.22280922964365e+00, 2.22068608998292e+00, 2.21788868246083e+00, 2.21429302776238e+00, 2.20977167239577e+00, 2.20419709038434e+00, 2.19744542500443e+00, 2.18940045259575e+00, 2.17995762894620e+00, 2.16902806167056e+00, 2.15654224137574e+00, 2.14245336204648e+00, 2.12674006836552e+00, 2.10940848530792e+00, 2.09049341317955e+00, 2.07005860821566e+00, 2.04819611287457e+00, 2.02502464817312e+00, 2.00068712932155e+00, 1.97534741175458e+00, 1.94918641376101e+00, 1.92239779115174e+00, 1.89518335653014e+00, 1.86774843965531e+00, 1.84029737631523e+00, 1.81302929248590e+00, 1.78613432080512e+00, 1.75979035068753e+00, 1.73416037518848e+00, 1.70939046030178e+00, 1.68560832858896e+00, 1.66292252099946e+00, 1.64142207972669e+00, 1.62117668138192e+00, 1.60223714335308e+00, 1.58463622607226e+00, 1.56838965880152e+00, 1.55349732504265e+00, 1.53994455435443e+00, 1.52770347890761e+00, 1.51673442440504e+00, 1.50698731515984e+00, 1.49840308152129e+00, 1.49091506406553e+00, 1.48445041283995e+00, 1.47893148147045e+00, 1.47427721527972e+00, 1.47040453003308e+00, 1.46722967395976e+00, 1.46466956081233e+00, 1.46264305651474e+00, 1.46107219701879e+00, 1.45988331093462e+00, 1.45900801784780e+00, 1.45838407240005e+00, 1.45795602545058e+00, 1.45767567702987e+00, 1.45750230122975e+00, 1.45740263034374e+00, 1.45735059402419e+00, 1.45732681839761e+00, 1.45731789937122e+00, 1.45731547318281e+00, 1.45731511508259e+00, 1.45731508820049e+00, 1.45731461505295e+00, 1.45731140941055e+00, 1.45729962515659e+00, 1.45726821286675e+00, 1.45719946887700e+00, 1.45706781989145e+00, 1.45683889365312e+00, 1.45646892142969e+00, 1.45590451243254e+00, 1.45508283341941e+00, 1.45393221814302e+00, 1.45237322048852e+00, 1.45032011167013e+00, 1.44768280552010e+00, 1.44436917682402e+00, 1.44028771639045e+00, 1.43535044413469e+00, 1.42947597943732e+00, 1.42259264834465e+00, 1.41464149199131e+00, 1.40557903213286e+00, 1.39537964978181e+00, 1.38403744295670e+00, 1.37156744992226e+00, 1.35800615439862e+00, 1.34341122729329e+00, 1.32786050276886e+00, 1.31145023132542e+00, 1.29429269508958e+00, 1.27651330680201e+00, 1.25824734081977e+00, 1.23963645955998e+00, 1.22082520126849e+00, 1.20195758525582e+00, 1.18317397052366e+00, 1.16460827570689e+00, 1.14638563573225e+00, 1.12862053690606e+00, 1.11141544033306e+00, 1.09485987605018e+00, 1.07902996863123e+00, 1.06398834002402e+00, 1.04978432698677e+00, 1.03645444805029e+00, 1.02402305738964e+00, 1.01250312908564e+00, 1.00189712370531e+00, 9.92197898750149e-01, 9.83389634316044e-01, 9.75448754505724e-01, 9.68344833185455e-01, 9.62041479232767e-01, 9.56497201302255e-01, 9.51666255294048e-01, 9.47499479197317e-01, 9.43945119931675e-01, 9.40949655414828e-01, 9.38458612591763e-01, 9.36417378860655e-01, 9.34772000555082e-01, 9.33469958251761e-01, 9.32460905043165e-01, 9.31697350911591e-01, 9.31135274297612e-01, 9.30734641137466e-01, 9.30459812224143e-01, 9.30279821782570e-01, 9.30168513570969e-01, 9.30104525434205e-01, 9.30071118739654e-01, 9.30055855145815e-01, 9.30050129277296e-01, 9.30048571705294e-01, 9.30048341811041e-01, 9.30048324479687e-01, 9.30048018454283e-01, 9.30045945088811e-01, 9.30038323203034e-01, 9.30018006246156e-01, 9.29973544058379e-01, 9.29888397387327e-01, 9.29740338307440e-01, 9.29501066870181e-01, 9.29136070945585e-01, 9.28604751997913e-01, 9.27860834076176e-01, 9.26853066176641e-01, 9.25526218995131e-01, 9.23822365741239e-01, 9.21682423205214e-01, 9.19047914071504e-01, 9.15862895379018e-01, 9.12075982255038e-01, 9.07642382145308e-01, 9.02525844461209e-01, 8.96700425574657e-01, 8.90151970834646e-01, 8.82879224625370e-01, 8.74894496521248e-01, 8.66223835488688e-01, 8.56906693088009e-01, 8.46995088224205e-01, 8.36552317193627e-01, 8.25651280513162e-01, 8.14372519594283e-01, 8.02802069774973e-01, 7.91029240593080e-01, 7.79144429626513e-01, 7.67237063909560e-01, 7.55393744812292e-01, 7.43696650732785e-01, 7.32222229460406e-01, 7.21040190823132e-01, 7.10212791931999e-01, 6.99794393057312e-01, 6.89831252356226e-01, 6.80361522212895e-01, 6.71415408332663e-01, 6.63015454184153e-01, 6.55176917042969e-01, 6.47908206915258e-01, 6.41211365270610e-01, 6.35082566200156e-01, 6.29512627903267e-01, 6.24487527006451e-01, 6.19988911963466e-01, 6.15994614602359e-01, 6.12479160764305e-01, 6.09414281958436e-01, 6.06769430103878e-01, 6.04512296836369e-01, 6.02609337632744e-01, 6.01026299284669e-01, 5.99728747185356e-01, 5.98682586654117e-01, 5.97854570304499e-01, 5.97212781463142e-01, 5.96727082067371e-01, 5.96369512490135e-01, 5.96114630506449e-01, 5.95939777218485e-01, 5.95825259223700e-01, 5.95754438595279e-01, 5.95713725228165e-01, 5.95692469605368e-01, 5.95682757832552e-01, 5.95679114626941e-01, 5.95678123588154e-01, 5.95677977312963e-01, 5.95677966225505e-01, 5.95677769654577e-01, 5.95676437858515e-01, 5.95671542048652e-01, 5.95658491719542e-01, 5.95629931913512e-01, 5.95575238460841e-01, 5.95480132321923e-01, 5.95326432250308e-01, 5.95091962727819e-01, 5.94750631316733e-01, 5.94272685995401e-01, 5.93625158444731e-01, 5.92772493423849e-01, 5.91677357205022e-01, 5.90301609585495e-01, 5.88607414529855e-01, 5.86558454545033e-01, 5.84121204225243e-01, 5.81266210003154e-01, 5.77969317091004e-01, 5.74212781940073e-01, 5.69986210131223e-01, 5.65287265896991e-01, 5.60122110412221e-01, 5.54505540934521e-01, 5.48460820624918e-01, 5.42019207797303e-01, 5.35219211580651e-01, 5.28105616729606e-01, 5.20728332102614e-01, 5.13141124187433e-01, 5.05400298680796e-01, 4.97563389855518e-01, 4.89687910145173e-01, 4.81830202265318e-01, 4.74044424628428e-01, 4.66381689083164e-01, 4.58889359177011e-01, 4.51610507947745e-01, 4.44583527098276e-01, 4.37841874402903e-01, 4.31413943197592e-01, 4.25323036537737e-01, 4.19587428705132e-01, 4.14220497840427e-01, 4.09230915226811e-01, 4.04622878864632e-01, 4.00396381223022e-01, 3.96547503254629e-01, 3.93068728781193e-01, 3.89949275103362e-01, 3.87175437089539e-01, 3.84730943008137e-01, 3.82597320959182e-01, 3.80754274927243e-01, 3.79180069233094e-01, 3.77851919543216e-01, 3.76746387667009e-01, 3.75839776217113e-01, 3.75108517937065e-01, 3.74529553237446e-01, 3.74080688360278e-01, 3.73740925744265e-01, 3.73490757709416e-01, 3.73312414612586e-01, 3.73190059202994e-01, 3.73109920042038e-01, 3.73060358510360e-01, 3.73031866025809e-01, 3.73016990518778e-01, 3.73010193808186e-01, 3.73007644132478e-01, 3.73006950559345e-01, 3.73006848189421e-01, 3.73006840380686e-01, 3.73006701289050e-01, 3.73005758923169e-01, 3.73002294692492e-01, 3.72993060371428e-01, 3.72972851519210e-01, 3.72934150044523e-01, 3.72866850655447e-01, 3.72758084467046e-01, 3.72592151310071e-01, 3.72350570193786e-01, 3.72012254792052e-01, 3.71553817603926e-01, 3.70950002472668e-01, 3.70174240376354e-01, 3.69199317873635e-01, 3.67998141474251e-01, 3.66544574828445e-01, 3.64814319461537e-01, 3.62785804406279e-01, 3.60441046153575e-01, 3.57766438476025e-01, 3.54753432382097e-01, 3.51399070014518e-01, 3.47706342701173e-01, 3.43684352250435e-01, 3.39348265287594e-01, 3.34719062038939e-01, 3.29823092442709e-01, 3.24691462774819e-01, 3.19359284253002e-01, 3.13864820724412e-01, 3.08248575274394e-01, 3.02552355473820e-01, 2.96818354345348e-01, 2.91088279508465e-01, 2.85402556992748e-01, 2.79799629531482e-01, 2.74315362345103e-01, 2.68982562972576e-01, 2.63830615961073e-01, 2.58885228409200e-01, 2.54168278596548e-01, 2.49697757251415e-01, 2.45487789368429e-01, 2.41548723794610e-01, 2.37887277925152e-01, 2.34506725633832e-01, 2.31407117838099e-01, 2.28585526693895e-01, 2.26036306162595e-01, 2.23751363437852e-01, 2.21720437328480e-01, 2.19931381053547e-01, 2.18370447935220e-01, 2.17022579121318e-01, 2.15871692714003e-01, 2.14900973536254e-01, 2.14093162278203e-01, 2.13430842003268e-01, 2.12896719055420e-01, 2.12473894406319e-01, 2.12146120535369e-01, 2.11898038166364e-01, 2.11715386698209e-01, 2.11585182048648e-01, 2.11495855932490e-01, 2.11437351335082e-01, 2.11401170094023e-01, 2.11380370006300e-01, 2.11369510640639e-01, 2.11364548942093e-01, 2.11362687643385e-01, 2.11362181325544e-01, 2.11362106594112e-01, 2.11362100853401e-01, 2.11361998070737e-01, 2.11361301703468e-01, 2.11358741789692e-01, 2.11351918045278e-01, 2.11336984680166e-01, 2.11308386390957e-01, 2.11258656515998e-01, 2.11178287308734e-01, 2.11055681067419e-01, 2.10877189371837e-01, 2.10627245812503e-01, 2.10288595242256e-01, 2.09842619639749e-01, 2.09269757103574e-01, 2.08550006322511e-01, 2.07663504217777e-01, 2.06591159562597e-01, 2.05315320597437e-01, 2.03820450407903e-01, 2.02093780599103e-01, 2.00125912058855e-01, 1.97911331752417e-01, 1.95448816788369e-01, 1.92741701493171e-01, 1.89797989749945e-01, 1.86630302978209e-01, 1.83255663238100e-01, 1.79695120284814e-01, 1.75973240182671e-01, 1.72117480573291e-01, 1.68157483279222e-01, 1.64124318219159e-01, 1.60049713458664e-01, 1.55965304705298e-01, 1.51901933974623e-01, 1.47889021958129e-01, 1.43954032368654e-01, 1.40122039808375e-01, 1.36415406056164e-01, 1.32853563585996e-01, 1.29452899974469e-01, 1.26226732872276e-01, 1.23185362508400e-01, 1.20336187252960e-01, 1.17683867473620e-01, 1.15230523598736e-01, 1.12975955724325e-01, 1.10917874032719e-01, 1.09052131496888e-01, 1.07372952618112e-01, 1.05873154111547e-01, 1.04544355378806e-01, 1.03377178192121e-01, 1.02361436201335e-01, 1.01486315636548e-01, 1.00740548918202e-01, 1.00112582831562e-01, 9.95907425243967e-02, 9.91633919149137e-02, 9.88190902372192e-02, 9.85467434994121e-02, 9.83357486872052e-02, 9.81761277134332e-02, 9.80586474822296e-02, 9.79749220797377e-02, 9.79174930707263e-02, 9.78798841920963e-02, 9.78566273781478e-02, 9.78432579854136e-02, 9.78362782377538e-02, 9.78330891973851e-02, 9.78318928885962e-02, 9.78315674645760e-02, 9.78315194327095e-02, 9.78315157148444e-02, 9.78314487829994e-02, 9.78309953104635e-02, 9.78293283060196e-02, 9.78248847499597e-02, 9.78151604698247e-02, 9.77965385884758e-02, 9.77641589813629e-02, 9.77118356704665e-02, 9.76320284615735e-02, 9.75158743125490e-02, 9.73532827926835e-02, 9.71330984394090e-02, 9.68433307391479e-02, 9.64714497948336e-02, 9.60047425039594e-02, 9.54307203560578e-02, 9.47375659682859e-02, 9.39146015140508e-02, 9.29527586498722e-02, 9.18450268461051e-02, 9.05868556134788e-02, 8.91764863594453e-02, 8.76151917473393e-02, 8.59074045208565e-02, 8.40607236286672e-02, 8.20857927419549e-02, 7.99960543062603e-02, 7.78073903773482e-02, 7.55376688817044e-02, 7.32062198952104e-02, 7.08332704793602e-02, 6.84393682137632e-02, 6.60448227455408e-02, 6.36691916390877e-02, 6.13308319812713e-02, 5.90465331640697e-02, 5.68312396895241e-02, 5.46978663670128e-02, 5.26572024586720e-02, 5.07178965901325e-02, 4.88865108239960e-02, 4.71676302633224e-02, 4.55640138320668e-02, 4.40767722790572e-02, 4.27055607172342e-02, 4.14487748647865e-02, 4.03037423354835e-02, 3.92669026029377e-02, 3.83339714544793e-02, 3.75000877226714e-02, 3.67599417524165e-02, 3.61078863865334e-02, 3.55380322232151e-02, 3.50443295271941e-02, 3.46206394896900e-02, 3.42607975654623e-02, 3.39586714081239e-02, 3.37082155209991e-02, 3.35035241863988e-02, 3.33388835817257e-02, 3.32088232894012e-02, 3.31081667147366e-02, 3.30320792972678e-02, 3.29761128882833e-02, 3.29362443146390e-02, 3.29089059878331e-02, 3.28910064633163e-02, 3.28799391045766e-02, 3.28735774375257e-02, 3.28702563540020e-02, 3.28687389874425e-02, 3.28681697812048e-02, 3.28680149442828e-02, 3.28679920907113e-02, 3.28679903074107e-02, 3.28679580177820e-02, 3.28677392513157e-02, 3.28669350506523e-02, 3.28647914084002e-02, 3.28601004116198e-02, 3.28511177865263e-02, 3.28355006651559e-02, 3.28102692826722e-02, 3.27717961545648e-02, 3.27158260053917e-02, 3.26375292355117e-02, 3.25315909227868e-02, 3.23923361784662e-02, 3.22138910562995e-02, 3.19903761512581e-02, 3.17161275882754e-02, 3.13859374447141e-02, 3.09953030078773e-02, 3.05406719452215e-02, 3.00196688024847e-02, 2.94312875792713e-02, 2.87760357345430e-02, 2.80560169968336e-02, 2.72749437780408e-02, 2.64380746034050e-02, 2.55520773726641e-02, 2.46248249110010e-02, 2.36651345304875e-02, 2.26824676001449e-02, 2.16866079242877e-02, 2.06873387529782e-02, 1.96941374270079e-02, 1.87159041622404e-02, 1.77607376787332e-02, 1.68357657918676e-02, 1.59470342748236e-02, 1.50994528072186e-02, 1.42967930651654e-02, 1.35417312459474e-02, 1.28359256501280e-02, 1.21801193182769e-02, 1.15742579896327e-02, 1.10176146095764e-02, 1.05089130309306e-02, 1.00464452033844e-02, 9.62817782616199e-03, 9.25184599782014e-03, 8.91503273750871e-03, 8.61523432971367e-03, 8.34991225634063e-03, 8.11653304559634e-03, 7.91259771899000e-03, 7.73566269065107e-03, 7.58335400315069e-03, 7.45337670327473e-03, 7.34352099846269e-03, 7.25166661260701e-03, 7.17578649681627e-03, 7.11395076068477e-03, 7.06433138566176e-03, 7.02520797689843e-03, 6.99497451767250e-03, 6.97214682626424e-03, 6.95537019468517e-03, 6.94342652637658e-03, 6.93524019741076e-03, 6.92988184945342e-03, 6.92656938292302e-03, 6.92466554861326e-03, 6.92367172201180e-03, 6.92321766787529e-03, 6.92304734190330e-03, 6.92300100961774e-03, 6.92299417108977e-03, 6.92299363292648e-03, 6.92298383043699e-03, 6.92291741738771e-03, 6.92267328013685e-03, 6.92202253161689e-03, 6.92059854555005e-03, 6.91787205929906e-03, 6.91313260849569e-03, 6.90547758025239e-03, 6.89381018372615e-03, 6.87684760137053e-03, 6.85314045648511e-03, 6.82110446458317e-03, 6.77906468414768e-03, 6.72531211848870e-03, 6.65817154380832e-03, 6.57607838434160e-03, 6.47766130357219e-03, 6.36182605349411e-03, 6.22783518108427e-03, 6.07537760659582e-03, 5.90462202628572e-03, 5.71624867015608e-03, 5.51145520452856e-03, 5.29193444787701e-03, 5.05982389680959e-03, 4.81762957301690e-03, 4.56812908154269e-03, 4.31426069443930e-03, 4.05900647816893e-03, 3.80527781157447e-03, 3.55581107091574e-03, 3.31307989794993e-03, 3.07922853332343e-03, 2.85602847449340e-03, 2.64485850967411e-03, 2.44670625910892e-03, 2.26218791794608e-03, 2.09158202897797e-03, 1.93487280155047e-03, 1.79179864528045e-03, 1.66190208338790e-03, 1.54457793002294e-03, 1.43911744563852e-03, 1.34474701607701e-03, 1.26066063984176e-03, 1.18604609245310e-03, 1.12010505175242e-03, 1.06206773788742e-03, 1.01120278362110e-03, 9.66823132467098e-04, 9.28288777238606e-04, 8.95007109534163e-04, 8.66431566821646e-04, 8.42059159896991e-04, 8.21427359277038e-04, 8.04110725344032e-04, 7.89717585050198e-04, 7.77886984737464e-04, 7.68286079816605e-04, 7.60608056872059e-04, 7.54570622625295e-04, 7.49915038081639e-04, 7.46405627277014e-04, 7.43829649422878e-04, 7.41997393049587e-04, 7.40742332556721e-04, 7.39921182685066e-04, 7.39413695411528e-04, 7.39122065635200e-04, 7.38969844862325e-04, 7.38900302209800e-04, 7.38874215699070e-04, 7.38867119663311e-04, 7.38866072309108e-04, 7.38865989156068e-04, 7.38864465299471e-04, 7.38854141010492e-04, 7.38816188855296e-04, 7.38715030255577e-04, 7.38493687263642e-04, 7.38069942672226e-04, 7.37333528119777e-04, 7.36144576499953e-04, 7.34333590305170e-04, 7.31703182088594e-04, 7.28031826191966e-04, 7.23079809535676e-04, 7.16597472523617e-04, 7.08335681649075e-04, 6.98058273361113e-04, 6.85555964313522e-04, 6.70660960144787e-04, 6.53261248872994e-04, 6.33313382252254e-04, 6.10852475714907e-04, 5.85998232521590e-04, 5.58956035638047e-04, 5.30012537060642e-04, 4.99525670038636e-04, 4.67909561708169e-04, 4.35615381652710e-04, 4.03109671642810e-04, 3.70852085957551e-04, 3.39274608919027e-04, 3.08764076304868e-04, 2.79649177736374e-04, 2.52192244630578e-04, 2.26585422961278e-04, 2.02950650097122e-04, 1.81343185650107e-04, 1.61758770426346e-04, 1.44144135386881e-04, 1.28409450324440e-04, 1.14440222347412e-04, 1.02106231036458e-04, 9.12665093171228e-05, 8.17709401808031e-05, 7.34577150804938e-05, 6.61202570295671e-05, 6.04127052650227e-05, 5.57247318528908e-05, 5.16193054753700e-05, 4.80555038375741e-05, 4.49747964168523e-05, 4.23228956716315e-05, 4.00518648078219e-05, 3.81184419744329e-05, 3.64824802971953e-05, 3.51065571528864e-05, 3.39564051774488e-05, 3.30014135094814e-05, 3.22147228821734e-05, 3.15728804345383e-05, 3.10552769047234e-05, 3.06436035106827e-05, 3.03214541101937e-05, 3.00740847585246e-05, 2.98882842386661e-05, 2.97523025736268e-05, 2.96558018977322e-05, 2.95898127351535e-05, 2.95466891149232e-05, 2.95200586180044e-05, 2.95047625851197e-05, 2.94967807175084e-05, 2.94931346598304e-05, 2.94917670480288e-05, 2.94913950398268e-05, 2.94913401308605e-05, 2.94913357361537e-05, 2.94912547662562e-05, 2.94907061828553e-05, 2.94886895986224e-05, 2.94833146128326e-05, 2.94715540736235e-05, 2.94490407970471e-05, 2.94099201685327e-05, 2.93467727612295e-05, 2.92506236725387e-05, 2.91110587244958e-05, 2.89164717361956e-05, 2.86544695757840e-05, 2.83124575073401e-05, 2.78784073302063e-05, 2.73417651506236e-05, 2.66943842905931e-05, 2.59312966477410e-05, 2.50511229324957e-05, 2.40560427364755e-05, 2.29515185505846e-05, 2.17462695584203e-05, 2.04530479965825e-05, 1.90903190464553e-05, 1.76839900660752e-05, 1.62673471937325e-05, 1.48772746992876e-05, 1.35466241157900e-05, 1.22960034669717e-05, 1.11305854311183e-05, 1.00455154043999e-05, 9.03757614523215e-06, 8.11532741096244e-06, 7.29921904252251e-06, 6.60902306215405e-06, 6.04639620327418e-06, 5.58726862465164e-06, 5.19289287481683e-06, 4.83294554953773e-06, 4.50349732794564e-06, 4.22590133375422e-06, 4.02627997964089e-06, 3.90989916709827e-06, 3.85056534568662e-06, 3.80463703031353e-06, 3.73738421188818e-06, 3.64014789333567e-06, 3.52928266167396e-06, 3.43256910895207e-06, 3.37301756836439e-06, 3.35754375235808e-06, 3.37563241359362e-06, 3.40889666539113e-06, 3.44254408446432e-06, 3.46972692355507e-06, 3.48969864027993e-06, 3.50430146411895e-06, 3.51543803844834e-06, 3.52413037951749e-06, 3.53067413705332e-06, 3.53513621753621e-06, 3.53770411187782e-06, 3.53876563075062e-06, 3.53881559433944e-06, 3.53832422561595e-06, 3.53764903771037e-06, 3.53700932897993e-06, 3.53650582241016e-06, 3.53615909709504e-06, 3.53594712443339e-06, 3.53583271639479e-06, 3.53577957845630e-06, 3.53575950562401e-06, 3.53575403089224e-06, 3.53575322264987e-06, 3.53575311784482e-06, 3.53575069844680e-06, 3.53573431333132e-06, 3.53567416766509e-06, 3.53551454423950e-06, 3.53516876394622e-06, 3.53451998336141e-06, 3.53343279080320e-06, 3.53178132783682e-06, 3.52949682280723e-06, 3.52662675950910e-06, 3.52337598741315e-06, 3.52006869697163e-06, 3.51694472122688e-06, 3.51372004026018e-06, 3.50894623046827e-06, 3.49941028687795e-06, 3.48003843697597e-06, 3.44478838520956e-06, 3.38846244574142e-06, 3.30616070876033e-06, 3.30493283406952e-06, 3.35798151240851e-06, 3.38278554323779e-06, 3.37570449637781e-06, 3.51716641766586e-06, 3.64053685331100e-06, 3.68696333019614e-06, 3.63907825828846e-06, 3.50268874349524e-06, 3.54979478700557e-06, 3.57293199826358e-06, 3.58011912382494e-06, 3.78673900236492e-06, 3.85694497140321e-06, 3.70912938585377e-06, 3.30136601078834e-06, 2.63684480472648e-06, 2.19370731002755e-06, 1.71242863089560e-06, 1.23797615891874e-06, 8.52974290310726e-07, 5.95107009004892e-07, 4.46581115198796e-07, 3.69045428882941e-07, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00],
[ 1.00000000000000e+01, 9.99999999566217e+00, 9.99999986152802e+00, 9.99999895274808e+00, 9.99999561195702e+00, 9.99998670648217e+00, 9.99996721629456e+00, 9.99992988728867e+00, 9.99986496255468e+00, 9.99976000200414e+00, 9.99959979809400e+00, 9.99936639254734e+00, 9.99903919598156e+00, 9.99859520932108e+00, 9.99800934288877e+00, 9.99725482622977e+00, 9.99630369910916e+00, 9.99512737181232e+00, 9.99369724092219e+00, 9.99198534519178e+00, 9.98996504499488e+00, 9.98761170813144e+00, 9.98490338447697e+00, 9.98182145208109e+00, 9.97835121781565e+00, 9.97448245652121e+00, 9.97020987378141e+00, 9.96553347894812e+00, 9.96045885683337e+00, 9.95499732856651e+00, 9.94916599447494e+00, 9.94298765446708e+00, 9.93649060424872e+00, 9.92970830874208e+00, 9.92267895723905e+00, 9.91544490801938e+00, 9.90805203330015e+00, 9.90054897833986e+00, 9.89298635117743e+00, 9.88541586172709e+00, 9.87788943066948e+00, 9.87045828969303e+00, 9.86317209509165e+00, 9.85607807648804e+00, 9.84922024153363e+00, 9.84263865587311e+00, 9.83636881551601e+00, 9.83044112611662e+00, 9.82488050062491e+00, 9.81970608344402e+00, 9.81493110572723e+00, 9.81056287287713e+00, 9.80660288177889e+00, 9.80304706190369e+00, 9.79988613124631e+00, 9.79710605519087e+00, 9.79468859389730e+00, 9.79261192172734e+00, 9.79085130062657e+00, 9.78937978828450e+00, 9.78816896132923e+00, 9.78718963378612e+00, 9.78641255153730e+00, 9.78580904454202e+00, 9.78535162008531e+00, 9.78501448226942e+00, 9.78477396529211e+00, 9.78460887070199e+00, 9.78450070170881e+00, 9.78443379067629e+00, 9.78439531905248e+00, 9.78437523211412e+00, 9.78436605393296e+00, 9.78436261083292e+00, 9.78436167422175e+00, 9.78436153597974e+00, 9.78436151784530e+00, 9.78436109533462e+00, 9.78435823275835e+00, 9.78434770956223e+00, 9.78431965810120e+00, 9.78425826571008e+00, 9.78414068252975e+00, 9.78393617495361e+00, 9.78360555729186e+00, 9.78310092600690e+00, 9.78236571191902e+00, 9.78133505641720e+00, 9.77993650825734e+00, 9.77809102831251e+00, 9.77571428096457e+00, 9.77271818296056e+00, 9.76901267370424e+00, 9.76450766523525e+00, 9.75911512559891e+00, 9.75275124587162e+00, 9.74533863865897e+00, 9.73680851427094e+00, 9.72710277986126e+00, 9.71617600652365e+00, 9.70399720970249e+00, 9.69055138946968e+00, 9.67584077954171e+00, 9.65988575774722e+00, 9.64272537640871e+00, 9.62441747909637e+00, 9.60503838059908e+00, 9.58468209963199e+00, 9.56345914835827e+00, 9.54149489854583e+00, 9.51892756017366e+00, 9.49590582347422e+00, 9.47258622867402e+00, 9.44913033812584e+00, 9.42570179241181e+00, 9.40246333495740e+00, 9.37957388870679e+00, 9.35718576377512e+00, 9.33544206728773e+00, 9.31447437659440e+00, 9.29440072553649e+00, 9.27532394125701e+00, 9.25733035688745e+00, 9.24048891387974e+00, 9.22485065716024e+00, 9.21044861687196e+00, 9.19729806229955e+00, 9.18539710657837e+00, 9.17472763484237e+00, 9.16525652340169e+00, 9.15693711320491e+00, 9.14971089710978e+00, 9.14350937729682e+00, 9.13825604650564e+00, 9.13386844471450e+00, 9.13026024152256e+00, 9.12734329397434e+00, 9.12502963003442e+00, 9.12323330951888e+00, 9.12187211712245e+00, 9.12086904630217e+00, 9.12015353817872e+00, 9.11966244621547e+00, 9.11934070507926e+00, 9.11914169056200e+00, 9.11902726647837e+00, 9.11896752374902e+00, 9.11894022610707e+00, 9.11892998570665e+00, 9.11892720005882e+00, 9.11892678890265e+00, 9.11892675260538e+00, 9.11892604137824e+00, 9.11892122270488e+00, 9.11890350867936e+00, 9.11885628903406e+00, 9.11875294716893e+00, 9.11855502430400e+00, 9.11821080089056e+00, 9.11765435335087e+00, 9.11680513159701e+00, 9.11556808883815e+00, 9.11383438016279e+00, 9.11148263054513e+00, 9.10838075657174e+00, 9.10438830965839e+00, 9.09935929220743e+00, 9.09314538245276e+00, 9.08559948907295e+00, 9.07657954344080e+00, 9.06595242602242e+00, 9.05359791433130e+00, 9.03941253337808e+00, 9.02331318615406e+00, 9.00524044181033e+00, 8.98516136335093e+00, 8.96307176535247e+00, 8.93899780587631e+00, 8.91299683556158e+00, 8.88515745072943e+00, 8.85559872555318e+00, 8.82446862975420e+00, 8.79194167112706e+00, 8.75821583431778e+00, 8.72350891633624e+00, 8.68805438304715e+00, 8.65209688754749e+00, 8.61588759977342e+00, 8.57967949658925e+00, 8.54372275355069e+00, 8.50826036479867e+00, 8.47352409794610e+00, 8.43973086843240e+00, 8.40707959467271e+00, 8.37574857318366e+00, 8.34589339306045e+00, 8.31764539253792e+00, 8.29111064722107e+00, 8.26636946982057e+00, 8.24347639445583e+00, 8.22246061417876e+00, 8.20332683762653e+00, 8.18605652895214e+00, 8.17060949381014e+00, 8.15692577272921e+00, 8.14492780137178e+00, 8.13452279481091e+00, 8.12560531004517e+00, 8.11805993765482e+00, 8.11176407002378e+00, 8.10659069024274e+00, 8.10241112305926e+00, 8.09909768746149e+00, 8.09652619007416e+00, 8.09457819985714e+00, 8.09314304789440e+00, 8.09211950149439e+00, 8.09141706940364e+00, 8.09095690452932e+00, 8.09067228188949e+00, 8.09050864215162e+00, 8.09042320456487e+00, 8.09038416676451e+00, 8.09036952223086e+00, 8.09036553855253e+00, 8.09036495056959e+00, 8.09036490340609e+00, 8.09036403300009e+00, 8.09035813586791e+00, 8.09033645733413e+00, 8.09027866992231e+00, 8.09015220155498e+00, 8.08990999172307e+00, 8.08948876245017e+00, 8.08880787859894e+00, 8.08776885796450e+00, 8.08625557568961e+00, 8.08413518978572e+00, 8.08125979492443e+00, 8.07746879021965e+00, 8.07259192364219e+00, 8.06645295139449e+00, 8.05887382568196e+00, 8.04967929978618e+00, 8.03870181638188e+00, 8.02578652506691e+00, 8.01079625964988e+00, 7.99361629642355e+00, 7.97415871286015e+00, 7.95236617299253e+00, 7.92821498179724e+00, 7.90171727611941e+00, 7.87292225325359e+00, 7.84191637857761e+00, 7.80882255821929e+00, 7.77379830860872e+00, 7.73703299860266e+00, 7.69874427838401e+00, 7.65917383971438e+00, 7.61858267235295e+00, 7.57724599063591e+00, 7.53544800260483e+00, 7.49347668303609e+00, 7.45161869344024e+00, 7.41015456921015e+00, 7.36935426928088e+00, 7.32947315928841e+00, 7.29074847706495e+00, 7.25339631047911e+00, 7.21760910253673e+00, 7.18355368716106e+00, 7.15136985065142e+00, 7.12116940776848e+00, 7.09303577697242e+00, 7.06702403588743e+00, 7.04316143505758e+00, 7.02144834511873e+00, 7.00185960939145e+00, 6.98434627045686e+00, 6.96883763544260e+00, 6.95524364050924e+00, 6.94345747043589e+00, 6.93335838435961e+00, 6.92481469379848e+00, 6.91768683432254e+00, 6.91183046794027e+00, 6.90709954980574e+00, 6.90334929062075e+00, 6.90043894550460e+00, 6.89823436149019e+00, 6.89661021945857e+00, 6.89545191240770e+00, 6.89465701048093e+00, 6.89413627401088e+00, 6.89381418865268e+00, 6.89362901102427e+00, 6.89353232855868e+00, 6.89348815284398e+00, 6.89347158089541e+00, 6.89346707291293e+00, 6.89346640754377e+00, 6.89346635622410e+00, 6.89346543468739e+00, 6.89345919113702e+00, 6.89343623913113e+00, 6.89337505709374e+00, 6.89324115958758e+00, 6.89298472174459e+00, 6.89253874883279e+00, 6.89181786996533e+00, 6.89071782004464e+00, 6.88911565718330e+00, 6.88687074399381e+00, 6.88382650023359e+00, 6.87981291135131e+00, 6.87464975269484e+00, 6.86815046295322e+00, 6.86012657357609e+00, 6.85039257456068e+00, 6.83877107257328e+00, 6.82509807659603e+00, 6.80922823099831e+00, 6.79103980788542e+00, 6.77043927121330e+00, 6.74736523535694e+00, 6.72179166068899e+00, 6.69373015748223e+00, 6.66323130542175e+00, 6.63038493677895e+00, 6.59531937396554e+00, 6.55819965375631e+00, 6.51922480824138e+00, 6.47862430447227e+00, 6.43665376957558e+00, 6.39359014547196e+00, 6.34972642763436e+00, 6.30536614637833e+00, 6.26081774795372e+00, 6.21638902701393e+00, 6.17238175240463e+00, 6.12908661490954e+00, 6.08677860880313e+00, 6.04571293911764e+00, 6.00612152411385e+00, 5.96821013865576e+00, 5.93215622048258e+00, 5.89810733933327e+00, 5.86618030994074e+00, 5.83646091508395e+00, 5.80900419459264e+00, 5.78383525024528e+00, 5.76095051417644e+00, 5.74031942866932e+00, 5.72188648689000e+00, 5.70557358613614e+00, 5.69128264665412e+00, 5.67889844944886e+00, 5.66829164550981e+00, 5.65932188653744e+00, 5.65184102385631e+00, 5.64569631821229e+00, 5.64073359916505e+00, 5.63680030945648e+00, 5.63374836771522e+00, 5.63143678273968e+00, 5.62973395486365e+00, 5.62851960486658e+00, 5.62768627866041e+00, 5.62714038646947e+00, 5.62680274810426e+00, 5.62660863070774e+00, 5.62650728136121e+00, 5.62646097342054e+00, 5.62644360161826e+00, 5.62643887605787e+00, 5.62643817857474e+00, 5.62643812576341e+00, 5.62643719021607e+00, 5.62643085174230e+00, 5.62640755079946e+00, 5.62634543871217e+00, 5.62620950626083e+00, 5.62594917296505e+00, 5.62549643118486e+00, 5.62476462642469e+00, 5.62364794056106e+00, 5.62202162805529e+00, 5.61974303641648e+00, 5.61665342110499e+00, 5.61258054166403e+00, 5.60734200026992e+00, 5.60074925650622e+00, 5.59261222373375e+00, 5.58274432405045e+00, 5.57096785196271e+00, 5.55711947325342e+00, 5.54105566702804e+00, 5.52265790745333e+00, 5.50183737900950e+00, 5.47853902652252e+00, 5.45274475965344e+00, 5.42447566100178e+00, 5.39379308679478e+00, 5.36079859760495e+00, 5.32563271100952e+00, 5.28847252499676e+00, 5.24952831591661e+00, 5.20903926318392e+00, 5.16726849029540e+00, 5.12449763446372e+00, 5.08102116339470e+00, 5.03714064767766e+00, 4.99315917341296e+00, 4.94937604636900e+00, 4.90608190139058e+00, 4.86355429407076e+00, 4.82205381982639e+00, 4.78182078075080e+00, 4.74307240343639e+00, 4.70600060042166e+00, 4.67077026226752e+00, 4.63751806456871e+00, 4.60635177280599e+00, 4.57735002671594e+00, 4.55056258421184e+00, 4.52601100266703e+00, 4.50368973265002e+00, 4.48356759613568e+00, 4.46558961793418e+00, 4.44967917562576e+00, 4.43574042964026e+00, 4.42366099121835e+00, 4.41331478180441e+00, 4.40456503298765e+00, 4.39726737157473e+00, 4.39127293000665e+00, 4.38643141849786e+00, 4.38259409242333e+00, 4.37961654709196e+00, 4.37736127257811e+00, 4.37569990412008e+00, 4.37451510898535e+00, 4.37370205873376e+00, 4.37316944637332e+00, 4.37284002070017e+00, 4.37265062466102e+00, 4.37255174025208e+00, 4.37250655855340e+00, 4.37248960924204e+00, 4.37248499860929e+00, 4.37248431808922e+00, 4.37248426700236e+00, 4.37248336781566e+00, 4.37247727569046e+00, 4.37245488034926e+00, 4.37239518224981e+00, 4.37226453268262e+00, 4.37201431647633e+00, 4.37157916756576e+00, 4.37087579530948e+00, 4.36980248520602e+00, 4.36823932149633e+00, 4.36604916092999e+00, 4.36307936658928e+00, 4.35916428806996e+00, 4.35412844965647e+00, 4.34779038182613e+00, 4.33996700424794e+00, 4.33047844156903e+00, 4.31915312823964e+00, 4.30583303722735e+00, 4.29037885164868e+00, 4.27267488993394e+00, 4.25263359560906e+00, 4.23019941297972e+00, 4.20535188997343e+00, 4.17810787825944e+00, 4.14852273679065e+00, 4.11669048569932e+00, 4.08274290028388e+00, 4.04684757691790e+00, 4.00920504170530e+00, 3.97004500676903e+00, 3.92962190698952e+00, 3.88820987114829e+00, 3.84609729551098e+00, 3.80358119485483e+00, 3.76096150587371e+00, 3.71853551095907e+00, 3.67659253692660e+00, 3.63540906401116e+00, 3.59524435644967e+00, 3.55633669867409e+00, 3.51890029230672e+00, 3.48312284068424e+00, 3.44916382134186e+00, 3.41715342427819e+00, 3.38719211594742e+00, 3.35935077629716e+00, 3.33367134874595e+00, 3.31016794024774e+00, 3.28882830962855e+00, 3.26961568608645e+00, 3.25247086492692e+00, 3.23731453312479e+00, 3.22404978218949e+00, 3.21256476933255e+00, 3.20273548965264e+00, 3.19442862182461e+00, 3.18750440775592e+00, 3.18181952326387e+00, 3.17722989263771e+00, 3.17359339571860e+00, 3.17077241264982e+00, 3.16863614947770e+00, 3.16706268797469e+00, 3.16594070588266e+00, 3.16517081948875e+00, 3.16466650904230e+00, 3.16435459874123e+00, 3.16417527637211e+00, 3.16408165250801e+00, 3.16403887466547e+00, 3.16402282716719e+00, 3.16401846185358e+00, 3.16401781754226e+00, 3.16401776932069e+00, 3.16401692252565e+00, 3.16401118536338e+00, 3.16399009492872e+00, 3.16393387542533e+00, 3.16381083966662e+00, 3.16357520859398e+00, 3.16316543518615e+00, 3.16250310828380e+00, 3.16149249931717e+00, 3.16002080051517e+00, 3.15795908661759e+00, 3.15516401312853e+00, 3.15148024256348e+00, 3.14674356598751e+00, 3.14078466075856e+00, 3.13343339749809e+00, 3.12452358103972e+00, 3.11389798297387e+00, 3.10141349924176e+00, 3.08694624704512e+00, 3.07039640313767e+00, 3.05169258215138e+00, 3.03079556036028e+00, 3.00770116794885e+00, 2.98244220140380e+00, 2.95508924619117e+00, 2.92575034663519e+00, 2.89456951229867e+00, 2.86172410494370e+00, 2.82742120369092e+00, 2.79189309455268e+00, 2.75539207058799e+00, 2.71818475760162e+00, 2.68054619557367e+00, 2.64275390698913e+00, 2.60508217032107e+00, 2.56779669170562e+00, 2.53114983296668e+00, 2.49537651295128e+00, 2.46069085529413e+00, 2.42728361281649e+00, 2.39532035989062e+00, 2.36494041163558e+00, 2.33625640423054e+00, 2.30935445449990e+00, 2.28429480897364e+00, 2.26111289192931e+00, 2.23982066709193e+00, 2.22040823707668e+00, 2.20284561662202e+00, 2.18708462859192e+00, 2.17306088424705e+00, 2.16069582029117e+00, 2.14989877389298e+00, 2.14056908277968e+00, 2.13259820041295e+00, 2.12587181628392e+00, 2.12027196884212e+00, 2.11567913404999e+00, 2.11197426673801e+00, 2.10904076563633e+00, 2.10676632703162e+00, 2.10504464726680e+00, 2.10377693149393e+00, 2.10287316576578e+00, 2.10225311205175e+00, 2.10184699117389e+00, 2.10159582679963e+00, 2.10145143407392e+00, 2.10137604859483e+00, 2.10134160446934e+00, 2.10132868330855e+00, 2.10132516844140e+00, 2.10132464965444e+00, 2.10132461081283e+00, 2.10132392854063e+00, 2.10131930604753e+00, 2.10130231329584e+00, 2.10125701702692e+00, 2.10115788810400e+00, 2.10096804762954e+00, 2.10063792369374e+00, 2.10010438316298e+00, 2.09929039419631e+00, 2.09810526476995e+00, 2.09644548985033e+00, 2.09419622441492e+00, 2.09123338082287e+00, 2.08742632688667e+00, 2.08264113559509e+00, 2.07674430949736e+00, 2.06960687355947e+00, 2.06110870168465e+00, 2.05114291635737e+00, 2.03962018061571e+00, 2.02647268940142e+00, 2.01165766561439e+00, 1.99516017661924e+00, 1.97699511028953e+00, 1.95720818552800e+00, 1.93587591887152e+00, 1.91310452332369e+00, 1.88902777395739e+00, 1.86380393242809e+00, 1.83761187450015e+00, 1.81064660657309e+00, 1.78311438549690e+00, 1.75522766854673e+00, 1.72720011677107e+00, 1.69924185616799e+00, 1.67155516986875e+00, 1.64433075436834e+00, 1.61774462804819e+00, 1.59195573501022e+00, 1.56710424533874e+00, 1.54331051725002e+00, 1.52067465906491e+00, 1.49927661036646e+00, 1.47917665193232e+00, 1.46041625214778e+00, 1.44301916218485e+00, 1.42699268158847e+00, 1.41232902831485e+00, 1.39900676111865e+00, 1.38699221611534e+00, 1.37624093227127e+00, 1.36669905170932e+00, 1.35830468955227e+00, 1.35098927429038e+00, 1.34467886328967e+00, 1.33929543915627e+00, 1.33475819148466e+00, 1.33098478540420e+00, 1.32789261375970e+00, 1.32540002425614e+00, 1.32342750705995e+00, 1.32189882279534e+00, 1.32074204620698e+00, 1.31989049751930e+00, 1.31928353213842e+00, 1.31886716009479e+00, 1.31859446960592e+00, 1.31842583424472e+00, 1.31832889013157e+00, 1.31827827786573e+00, 1.31825515299980e+00, 1.31824647811218e+00, 1.31824411833694e+00, 1.31824377003922e+00, 1.31824374389298e+00, 1.31824328369723e+00, 1.31824016580571e+00, 1.31822870414277e+00, 1.31819815183496e+00, 1.31813129026430e+00, 1.31800324775210e+00, 1.31778059763015e+00, 1.31742078116660e+00, 1.31687189570419e+00, 1.31607288200111e+00, 1.31495413657194e+00, 1.31343856420689e+00, 1.31144307229855e+00, 1.30888049185374e+00, 1.30566189021949e+00, 1.30169921820057e+00, 1.29690821054914e+00, 1.29121143547882e+00, 1.28454136807930e+00, 1.27684334673948e+00, 1.26807826337697e+00, 1.25822483951346e+00, 1.24728135240514e+00, 1.23526669887043e+00, 1.22222071823826e+00, 1.20820373775576e+00, 1.19329535048552e+00, 1.17759248303297e+00, 1.16120685392947e+00, 1.14426195900288e+00, 1.12688974430305e+00, 1.10922713812744e+00, 1.09141261097187e+00, 1.07358291692630e+00, 1.05587014457158e+00, 1.03839917314018e+00, 1.02128559429510e+00, 1.00463412494438e+00, 9.88537505071442e-01, 9.73075848810847e-01, 9.58316398158338e-01, 9.44313617071344e-01, 9.31109558818999e-01, 9.18734440256692e-01, 9.07207361900600e-01, 8.96537120844496e-01, 8.86723073359057e-01, 8.77756014300898e-01, 8.69619050342486e-01, 8.62288452879252e-01, 8.55734483876091e-01, 8.49922193674267e-01, 8.44812193831720e-01, 8.40361410460099e-01, 8.36523824368431e-01, 8.33251203792994e-01, 8.30493833790412e-01, 8.28201243731986e-01, 8.26322931028843e-01, 8.24809075534469e-01, 8.23611235333313e-01, 8.22683011162747e-01, 8.21980663857859e-01, 8.21463667252315e-01, 8.21095178155318e-01, 8.20842405513911e-01, 8.20676862714997e-01, 8.20574490120974e-01, 8.20515639188015e-01, 8.20484914608977e-01, 8.20470876503203e-01, 8.20465610358756e-01, 8.20464177842956e-01, 8.20463966406741e-01, 8.20463950466911e-01, 8.20463669012062e-01, 8.20461762115750e-01, 8.20454752188535e-01, 8.20436066490995e-01, 8.20395174286993e-01, 8.20316864610186e-01, 8.20180695270949e-01, 8.19960641216166e-01, 8.19624967338445e-01, 8.19136347015970e-01, 8.18452242661120e-01, 8.17525557953793e-01, 8.16305562882212e-01, 8.14739082016837e-01, 8.12771923666698e-01, 8.10350513121008e-01, 8.07423677867616e-01, 8.03944517706380e-01, 7.99872279567865e-01, 7.95174147313988e-01, 7.89826852504856e-01, 7.83818014427290e-01, 7.77147127380436e-01, 7.69826130284097e-01, 7.61879517171911e-01, 7.53343975203508e-01, 7.44267566900787e-01, 7.34708502407772e-01, 7.24733572733385e-01, 7.14416333659385e-01, 7.03835140600299e-01, 6.93071136556449e-01, 6.82206288865377e-01, 6.71321557144561e-01, 6.60495256710760e-01, 6.49801661241810e-01, 6.39309867835021e-01, 6.29082928864033e-01, 6.19177239532906e-01, 6.09642158529393e-01, 6.00519831850924e-01, 5.91845186402162e-01, 5.83646059707936e-01, 5.75943434256438e-01, 5.68751748770028e-01, 5.62079263357273e-01, 5.55928460425747e-01, 5.50296467980592e-01, 5.45175496194044e-01, 5.40553281724664e-01, 5.36413537102694e-01, 5.32736404555873e-01, 5.29498914945332e-01, 5.26675453056628e-01, 5.24238230405924e-01, 5.22157766048942e-01, 5.20403374708396e-01, 5.18943659969963e-01, 5.17747008462331e-01, 5.16782078980254e-01, 5.16018278593355e-01, 5.15426216081709e-01, 5.14978121724342e-01, 5.14648221696124e-01, 5.14413055228173e-01, 5.14251723336681e-01, 5.14146059346174e-01, 5.14080713582150e-01, 5.14043147376761e-01, 5.14023534756900e-01, 5.14014573663063e-01, 5.14011212059394e-01, 5.14010297622818e-01, 5.14010162653939e-01, 5.14010152423483e-01, 5.14009971046461e-01, 5.14008742191196e-01, 5.14004224800019e-01, 5.13992183176464e-01, 5.13965830797806e-01, 5.13915364449170e-01, 5.13827608079687e-01, 5.13685783967656e-01, 5.13469426632334e-01, 5.13154452289978e-01, 5.12713393318881e-01, 5.12115802956385e-01, 5.11328830119627e-01, 5.10317957729576e-01, 5.09047890279133e-01, 5.07483567869896e-01, 5.05591275021208e-01, 5.03339803923851e-01, 5.00701624343121e-01, 4.97654007036843e-01, 4.94180045269427e-01, 4.90269520507523e-01, 4.85919564063378e-01, 4.81135076228764e-01, 4.75928877712743e-01, 4.70321583867678e-01, 4.64341208800807e-01, 4.58022522399849e-01, 4.51406197001457e-01, 4.44537790657819e-01, 4.37466619945576e-01, 4.30244576811786e-01, 4.22924941403531e-01, 4.15561236952659e-01, 4.08206164616349e-01, 4.00910646814964e-01, 3.93722998078433e-01, 3.86688233525827e-01, 3.79847517414056e-01, 3.73237747996953e-01, 3.66891270311035e-01, 3.60835705366702e-01, 3.55093882395149e-01, 3.49683860061603e-01, 3.44619022679087e-01, 3.39908238233381e-01, 3.35556066271522e-01, 3.31563005248963e-01, 3.27925770628802e-01, 3.24637596750826e-01, 3.21688557121870e-01, 3.19065899219073e-01, 3.16754391056225e-01, 3.14736677573070e-01, 3.12993645324477e-01, 3.11504793956180e-01, 3.10248612572549e-01, 3.09202958378831e-01, 3.08345433994683e-01, 3.07653758694315e-01, 3.07106127657207e-01, 3.06681552248415e-01, 3.06360173524659e-01, 3.06123540702705e-01, 3.05954846326292e-01, 3.05839110385266e-01, 3.05763306691317e-01, 3.05716426365956e-01, 3.05689475271318e-01, 3.05675404497885e-01, 3.05668975474248e-01, 3.05666563730210e-01, 3.05665907677795e-01, 3.05665810845852e-01, 3.05665803459551e-01, 3.05665671892458e-01, 3.05664780506403e-01, 3.05661503682570e-01, 3.05652768916239e-01, 3.05633653311590e-01, 3.05597045476526e-01, 3.05533386756078e-01, 3.05430504352225e-01, 3.05273547241754e-01, 3.05045033829060e-01, 3.04725017835636e-01, 3.04291375903439e-01, 3.03720216669959e-01, 3.02986406610190e-01, 3.02064202776442e-01, 3.00927976860550e-01, 2.99553009047806e-01, 2.97916324352027e-01, 2.95997539046656e-01, 2.93779681022642e-01, 2.91249945977741e-01, 2.88400351738172e-01, 2.85228256007873e-01, 2.81736708451772e-01, 2.77934615974699e-01, 2.73836709803074e-01, 2.69463313733985e-01, 2.64839923779370e-01, 2.59996619484493e-01, 2.54967335628219e-01, 2.49789029206886e-01, 2.44500780210926e-01, 2.39142865642435e-01, 2.33755844659456e-01, 2.28379689030286e-01, 2.23052987730899e-01, 2.17812248076466e-01, 2.12691308807177e-01, 2.07720873572415e-01, 2.02928166727208e-01, 1.98336707627883e-01, 1.93966194942540e-01, 1.89832489021061e-01, 1.85947678141575e-01, 1.82320213416933e-01, 1.78955097180953e-01, 1.75854110599758e-01, 1.73016067852155e-01, 1.70437086265574e-01, 1.68110864055055e-01, 1.66028959586752e-01, 1.64181068198104e-01, 1.62555294414661e-01, 1.61138418808147e-01, 1.59916159682108e-01, 1.58873430229988e-01, 1.57994591802995e-01, 1.57263703502370e-01, 1.56664767553885e-01, 1.56181968935451e-01, 1.55799906633938e-01, 1.55503812833611e-01, 1.55279755414226e-01, 1.55114818476713e-01, 1.54997255310628e-01, 1.54916608330220e-01, 1.54863791056453e-01, 1.54831128191752e-01, 1.54812351164180e-01, 1.54802548117791e-01, 1.54798069082981e-01, 1.54796388851777e-01, 1.54795931789037e-01, 1.54795864327568e-01, 1.54795859145322e-01, 1.54795766361485e-01, 1.54795137737873e-01, 1.54792826859156e-01, 1.54786666967667e-01, 1.54773186514150e-01, 1.54747371052890e-01, 1.54702481567032e-01, 1.54629938476111e-01, 1.54519279582258e-01, 1.54358198960477e-01, 1.54132672172036e-01, 1.53827171051853e-01, 1.53424968584570e-01, 1.52908530982401e-01, 1.52259990030425e-01, 1.51461684192260e-01, 1.50496752105312e-01, 1.49349757286938e-01, 1.48007318561358e-01, 1.46458717397246e-01, 1.44696451510258e-01, 1.42716704156888e-01, 1.40519700808328e-01, 1.38109929426939e-01, 1.35496207214134e-01, 1.32691585049217e-01, 1.29713090265930e-01, 1.26581318131471e-01, 1.23319891548277e-01, 1.19954816272784e-01, 1.16513764653750e-01, 1.13025324068161e-01, 1.09518246675531e-01, 1.06020734904693e-01, 1.02559792573745e-01, 9.91606652704711e-02, 9.58463862645527e-02, 9.26374365017306e-02, 8.95515198214216e-02, 8.66034480081321e-02, 8.38051250404437e-02, 8.11656161668104e-02, 7.86912852684782e-02, 7.63859832706277e-02, 7.42512709258278e-02, 7.22866608418471e-02, 7.04898658538244e-02, 6.88570434507643e-02, 6.73830286945432e-02, 6.60615507001321e-02, 6.48854301184247e-02, 6.38467570683794e-02, 6.29370505406683e-02, 6.21474014151158e-02, 6.14686019033499e-02, 6.08912644733941e-02, 6.04059331788420e-02, 6.00031898599998e-02, 5.96737569772934e-02, 5.94085979572387e-02, 5.91990149635588e-02, 5.90367430410197e-02, 5.89140387046339e-02, 5.88237603443265e-02, 5.87594373520517e-02, 5.87153247018595e-02, 5.86864398452445e-02, 5.86685792169188e-02, 5.86583123450843e-02, 5.86529524654562e-02, 5.86505035727249e-02, 5.86495849215449e-02, 5.86493350273827e-02, 5.86492981435711e-02, 5.86492952886122e-02, 5.86492438914625e-02, 5.86488956688704e-02, 5.86476155757936e-02, 5.86442033944152e-02, 5.86367363085625e-02, 5.86224374389026e-02, 5.85975762567266e-02, 5.85574065207979e-02, 5.84961470296914e-02, 5.84070099596232e-02, 5.82822807485677e-02, 5.81134522679161e-02, 5.78914142853169e-02, 5.76066969003541e-02, 5.72497637297253e-02, 5.68113472230820e-02, 5.62828148008805e-02, 5.56565508237004e-02, 5.49263361178518e-02, 5.40877043363501e-02, 5.31382532656206e-02, 5.20778896646270e-02, 5.09089885724929e-02, 4.96364522706934e-02, 4.82676600271787e-02, 4.68123069323223e-02, 4.52821379088597e-02, 4.36905905603275e-02, 4.20523671133193e-02, 4.03829605970802e-02, 3.86981630749634e-02, 3.70135839531045e-02, 3.53442042017935e-02, 3.37039880813563e-02, 3.21055682369291e-02, 3.05600135166166e-02, 2.90766822990562e-02, 2.76631581355069e-02, 2.63252596113031e-02, 2.50671128103690e-02, 2.38912727229224e-02, 2.27988792926812e-02, 2.17898343461620e-02, 2.08629870922135e-02, 2.00163179085031e-02, 1.92471124421019e-02, 1.85521203923571e-02, 1.79276955267049e-02, 1.73699153803462e-02, 1.68746806437348e-02, 1.64377954283825e-02, 1.60550304366076e-02, 1.57221715800855e-02, 1.54350568411914e-02, 1.51896041976194e-02, 1.49818332785863e-02, 1.48078831291105e-02, 1.46640280609684e-02, 1.45466930959621e-02, 1.44524699884077e-02, 1.43781342791716e-02, 1.43206633145621e-02, 1.42772546914465e-02, 1.42453441969259e-02, 1.42226220238868e-02, 1.42070458845343e-02, 1.41968496235769e-02, 1.41905460508909e-02, 1.41869229568883e-02, 1.41850316173180e-02, 1.41841675039043e-02, 1.41838433540702e-02, 1.41837551782960e-02, 1.41837421637696e-02, 1.41837411482264e-02, 1.41837227601052e-02, 1.41835981783274e-02, 1.41831402092087e-02, 1.41819194829029e-02, 1.41792482131571e-02, 1.41741334078986e-02, 1.41652418345259e-02, 1.41508790515239e-02, 1.41289848025643e-02, 1.40971469730538e-02, 1.40526360600592e-02, 1.39924616268985e-02, 1.39134514341865e-02, 1.38123528150338e-02, 1.36859543906791e-02, 1.35312244594310e-02, 1.33454604596537e-02, 1.31264419986685e-02, 1.28725783023338e-02, 1.25830398524904e-02, 1.22578637062603e-02, 1.18980227363252e-02, 1.15054508851440e-02, 1.10830194298313e-02, 1.06344629808213e-02, 1.01642581079169e-02, 9.67746161603600e-03, 9.17951906138370e-03, 8.67605664159221e-03, 8.17267078022610e-03, 7.67472942201053e-03, 7.18719734263651e-03, 6.71449493745868e-03, 6.26039640822107e-03, 5.82796949959707e-03, 5.41955540772854e-03, 5.03678456221657e-03, 4.68062190825513e-03, 4.35143417716333e-03, 4.04907138452515e-03, 3.77295528297950e-03, 3.52216850787634e-03, 3.29553946178507e-03, 3.09171938650898e-03, 2.90924941273829e-03, 2.74661655162411e-03, 2.60229853915028e-03, 2.47479814144747e-03, 2.36266798844120e-03, 2.26452725891235e-03, 2.17907163823483e-03, 2.10507795650194e-03, 2.04140482717463e-03, 1.98699047336071e-03, 1.94084877086809e-03, 1.90206436924534e-03, 1.86978758442358e-03, 1.84322959534633e-03, 1.82165832442216e-03, 1.80439523795251e-03, 1.79081316857399e-03, 1.78033513801168e-03, 1.77243404822214e-03, 1.76663301701652e-03, 1.76250606451056e-03, 1.75967881487793e-03, 1.75782886556597e-03, 1.75668549543495e-03, 1.75602843068745e-03, 1.75568545894578e-03, 1.75552876951279e-03, 1.75546999267340e-03, 1.75545400420951e-03, 1.75545164435675e-03, 1.75545145864575e-03, 1.75544807598627e-03, 1.75542515810889e-03, 1.75534091164842e-03, 1.75511635747583e-03, 1.75462500713230e-03, 1.75368432267220e-03, 1.75204943493057e-03, 1.74940962554970e-03, 1.74538810455793e-03, 1.73954562082734e-03, 1.73138840352927e-03, 1.72038082560455e-03, 1.70596298153924e-03, 1.68757306836298e-03, 1.66467404974036e-03, 1.63678358886266e-03, 1.60350569955270e-03, 1.56456205446531e-03, 1.51982048958565e-03, 1.46931804752981e-03, 1.41327598942098e-03, 1.35210462609514e-03, 1.28639657523699e-03, 1.21690808168218e-03, 1.14452922548658e-03, 1.07024502409203e-03, 9.95090433894674e-04, 9.20102914768961e-04, 8.46276433254189e-04, 7.74520514809751e-04, 7.05627262382436e-04, 6.40248256683407e-04, 5.78882106998080e-04, 5.21872306553206e-04, 4.69414117709732e-04, 4.21568573205633e-04, 3.78281369554756e-04, 3.39404423167267e-04, 3.04718091166833e-04, 2.73952441210312e-04, 2.46806406229538e-04, 2.22964113323343e-04, 2.02108080912922e-04, 1.83929300147767e-04, 1.68134440423628e-04, 1.54450549450402e-04, 1.42627676272156e-04, 1.32439854773736e-04, 1.23684864027140e-04, 1.16183138459292e-04, 1.09776140869045e-04, 1.04324442167680e-04, 9.97056846383515e-05, 9.58125503967011e-05, 9.25508176958800e-05, 8.98375621325442e-05, 8.75995419953313e-05, 8.57717923228307e-05, 8.42964385192602e-05, 8.31217278918217e-05, 8.22012672632155e-05, 8.14934468014172e-05, 8.09610244179026e-05, 8.05708407831673e-05, 8.02936318997638e-05, 8.01039045654761e-05, 7.99798399597762e-05, 7.99031926455029e-05, 7.98591563135886e-05, 7.98361735778741e-05, 7.98256744456438e-05, 7.98217361648312e-05, 7.98206648831012e-05, 7.98205067642499e-05, 7.98204942125125e-05, 7.98202641405510e-05, 7.98187053856481e-05, 7.98129754642218e-05, 7.97977033553110e-05, 7.97642896444017e-05, 7.97003325754275e-05, 7.95892175915469e-05, 7.94099129013404e-05, 7.91370182842144e-05, 7.87411167610941e-05, 7.81894760937263e-05, 7.74471365421192e-05, 7.64784005085062e-05, 7.52487066009166e-05, 7.37268266384206e-05, 7.18872713125762e-05, 6.97127358169070e-05, 6.71963687128851e-05, 6.43436165841393e-05, 6.11733936768388e-05, 5.77183577110674e-05, 5.40241450437406e-05, 5.01475279541808e-05, 4.61535937255064e-05, 4.21121879412567e-05, 3.80939807769209e-05, 3.41665716128148e-05, 3.03910214327630e-05, 2.68191021977060e-05, 2.34914173091738e-05, 2.04364317379913e-05, 1.76703850822690e-05, 1.51980288780607e-05, 1.30140838763282e-05, 1.11052273332746e-05, 9.45233033711190e-06, 8.03264495630174e-06, 6.82172546540016e-06, 5.79499356945234e-06, 4.92892240977483e-06, 4.20179707560919e-06, 3.59399649063619e-06, 3.08784352043290e-06, 2.66727547248991e-06, 2.31768764535900e-06, 2.02613440869470e-06, 1.78176199128188e-06, 1.57611157192150e-06, 1.40293516892598e-06, 1.25752309898530e-06, 1.13595702977420e-06, 1.03465950118675e-06, 9.50250948150710e-07, 8.79610096760075e-07, 8.20096081710631e-07, 7.69780637455142e-07, 7.27452416526830e-07, 6.92352292830378e-07, 6.63841614567673e-07, 6.41207959617418e-07, 6.23644341501972e-07, 6.10316426730151e-07, 6.00432798592010e-07, 5.93285815142508e-07, 5.88265994139192e-07, 5.84861309604889e-07, 5.82649693825488e-07, 5.81289500090065e-07, 5.80510219450048e-07, 5.80104160263772e-07, 5.79918811638436e-07, 5.79849311677094e-07, 5.79830407687609e-07, 5.79827618807371e-07, 5.79827386104625e-07, 5.79822934554386e-07, 5.79792782673394e-07, 5.79681957546541e-07, 5.79386707499907e-07, 5.78741412127215e-07, 5.77508817656280e-07, 5.75375215916551e-07, 5.71952391226205e-07, 5.66787836220328e-07, 5.59382945297671e-07, 5.49216165879329e-07, 5.35765460440984e-07, 5.18525414605620e-07, 4.97021646950202e-07, 4.70831690080331e-07, 4.39608935920432e-07, 4.03059738474099e-07, 3.60781369224128e-07, 3.11934927665902e-07, 2.54915246244991e-07, 3.55271367880050e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 2.01513119080232e-07, 2.48336352370870e-07, 2.91658048823251e-07, 3.33515085770841e-07, 3.72393044401065e-07, 4.04539663989567e-07, 4.25771725431900e-07, 4.34300202911686e-07, 4.32647354386972e-07, 4.26921134888403e-07, 4.24173018487295e-07, 4.29173756089085e-07, 4.41855332766750e-07, 4.58091522048676e-07, 4.73175948201288e-07, 4.83934355699489e-07, 4.88983507906937e-07, 4.88496020523144e-07, 4.83614300605950e-07, 4.76139655702923e-07, 4.68801053088441e-07, 4.64567296631913e-07, 4.64885648199242e-07, 4.69183847329191e-07, 4.75824670331804e-07, 4.82776504995286e-07, 4.88141067123138e-07, 4.91047225992247e-07, 4.92042632416201e-07, 4.92435177079642e-07, 4.93282801272699e-07, 4.94917992810429e-07, 4.97118950448794e-07, 4.99484398730488e-07, 5.01677641651099e-07, 5.03503931881255e-07, 5.04899910325207e-07, 5.05893384072920e-07, 5.06559397095430e-07, 5.06984816794898e-07, 5.07246632253100e-07, 5.07403289162767e-07, 5.07494782198137e-07, 5.07546690897698e-07, 5.07574855035386e-07, 5.07589062337388e-07, 5.07595439458441e-07, 5.07597812671179e-07, 5.07598455712355e-07, 5.07598551635624e-07, 5.07598549859267e-07, 5.07598436172430e-07, 5.07597668786275e-07, 5.07594871024253e-07, 5.07587605724780e-07, 5.07572677221901e-07, 5.07547765593586e-07, 5.07515720116203e-07, 5.07493149726201e-07, 5.07524484660848e-07, 5.07700573137981e-07, 5.08176698943430e-07, 5.09178875063299e-07, 5.10980939338879e-07, 5.13830311632546e-07, 5.17802169142101e-07, 5.22587411566633e-07, 5.27296270647071e-07, 5.30471011117584e-07, 5.30520887664920e-07, 5.26565486325126e-07, 5.19256342812469e-07, 5.10900374450785e-07, 5.04535458745181e-07, 5.02358311393891e-07, 5.04265790013392e-07, 5.07367419189109e-07, 5.07607943234234e-07, 5.03075000324316e-07, 4.95821897317228e-07, 4.90026110355757e-07, 4.88421578737075e-07, 4.89306556161750e-07, 4.85192606092255e-07, 4.66620464933953e-07, 4.29622700082177e-07, 3.78529113831405e-07, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 0.00000000000000e+00, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 0.00000000000000e+00, 1.77635683940025e-15, 3.55271367880050e-15, 1.77635683940025e-15, 0.00000000000000e+00, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 0.00000000000000e+00, 0.00000000000000e+00, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 0.00000000000000e+00, 0.00000000000000e+00, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 1.77635683940025e-15, 0.00000000000000e+00],
[ 1.10000000000000e+01, 1.09999999950043e+01, 1.09999998405264e+01, 1.09999987939147e+01, 1.09999949464365e+01, 1.09999846902957e+01, 1.09999622440876e+01, 1.09999192534849e+01, 1.09998444817341e+01, 1.09997236018862e+01, 1.09995390996675e+01, 1.09992702926168e+01, 1.09988934676727e+01, 1.09983821359002e+01, 1.09977073996183e+01, 1.09968384239214e+01, 1.09957430015983e+01, 1.09943881977975e+01, 1.09927410585611e+01, 1.09907693655729e+01, 1.09884424181757e+01, 1.09857318229134e+01, 1.09826122705209e+01, 1.09790622804151e+01, 1.09750648932859e+01, 1.09706082933322e+01, 1.09656863430030e+01, 1.09602990147712e+01, 1.09544527064830e+01, 1.09481604291742e+01, 1.09414418589300e+01, 1.09343232473627e+01, 1.09268371885758e+01, 1.09190222440065e+01, 1.09109224302353e+01, 1.09025865786003e+01, 1.08940675791461e+01, 1.08854215249207e+01, 1.08767067757661e+01, 1.08679829633849e+01, 1.08593099614782e+01, 1.08507468460425e+01, 1.08423508714141e+01, 1.08341764873383e+01, 1.08262744212229e+01, 1.08186908478687e+01, 1.08114666664281e+01, 1.08046369012326e+01, 1.07982302395795e+01, 1.07922687156950e+01, 1.07867675460372e+01, 1.07817351169866e+01, 1.07771731219218e+01, 1.07730768407888e+01, 1.07694355516571e+01, 1.07662330604875e+01, 1.07634483324926e+01, 1.07610562061183e+01, 1.07590281688535e+01, 1.07573331728342e+01, 1.07559384675681e+01, 1.07548104270736e+01, 1.07539153493090e+01, 1.07532202069349e+01, 1.07526933301789e+01, 1.07523050047973e+01, 1.07520279708041e+01, 1.07518378106699e+01, 1.07517132190154e+01, 1.07516361493279e+01, 1.07515918368329e+01, 1.07515687002439e+01, 1.07515581286088e+01, 1.07515541627691e+01, 1.07515530839595e+01, 1.07515529247292e+01, 1.07515529038416e+01, 1.07515524171843e+01, 1.07515491200050e+01, 1.07515369991533e+01, 1.07515046888604e+01, 1.07514339757721e+01, 1.07512985410092e+01, 1.07510629852114e+01, 1.07506821741742e+01, 1.07501009332164e+01, 1.07492541081315e+01, 1.07480669998113e+01, 1.07464561687652e+01, 1.07443305951519e+01, 1.07415931699563e+01, 1.07381424838794e+01, 1.07338748726003e+01, 1.07286866704470e+01, 1.07224766192102e+01, 1.07151483747662e+01, 1.07066130511970e+01, 1.06967917400160e+01, 1.06856179407685e+01, 1.06730398386174e+01, 1.06590223646090e+01, 1.06435489754139e+01, 1.06266230918433e+01, 1.06082691398780e+01, 1.05885331448358e+01, 1.05674828390345e+01, 1.05452072560617e+01, 1.05218158003421e+01, 1.04974367985412e+01, 1.04722155585247e+01, 1.04463119808714e+01, 1.04198977859671e+01, 1.03931534351373e+01, 1.03662648359859e+01, 1.03394199293015e+01, 1.03128052572179e+01, 1.02866026098529e+01, 1.02609858409130e+01, 1.02361179325364e+01, 1.02121483769613e+01, 1.01892109285027e+01, 1.01674217647776e+01, 1.01468780819803e+01, 1.01276571358804e+01, 1.01098157284888e+01, 1.00933901301832e+01, 1.00783964184786e+01, 1.00648312074408e+01, 1.00526727357350e+01, 1.00418822762547e+01, 1.00324058259471e+01, 1.00241760306949e+01, 1.00171142968298e+01, 1.00111330380394e+01, 1.00061380041572e+01, 1.00020306367370e+01, 9.99871039557842e+00, 9.99607700069815e+00, 9.99403253580209e+00, 9.99248336227059e+00, 9.99134179709137e+00, 9.99052751408533e+00, 9.98996863508422e+00, 9.98960248627677e+00, 9.98937600449607e+00, 9.98924578846919e+00, 9.98917780063383e+00, 9.98914673566900e+00, 9.98913508200504e+00, 9.98913191191425e+00, 9.98913144401507e+00, 9.98913140270848e+00, 9.98913059332602e+00, 9.98912510963560e+00, 9.98910495093207e+00, 9.98905121463476e+00, 9.98893361104330e+00, 9.98870837452755e+00, 9.98831665013496e+00, 9.98768342218180e+00, 9.98671703722350e+00, 9.98530935810282e+00, 9.98333656878708e+00, 9.98066063168655e+00, 9.97713138039857e+00, 9.97258921170599e+00, 9.96686832159635e+00, 9.95980041153338e+00, 9.95121877371849e+00, 9.94096264815820e+00, 9.92888173054456e+00, 9.91484069880464e+00, 9.89872361824512e+00, 9.88043808110455e+00, 9.85991893666844e+00, 9.83713147355934e+00, 9.81207392700423e+00, 9.78477920126286e+00, 9.75531572110941e+00, 9.72378735594335e+00, 9.69033239477326e+00, 9.65512158829024e+00, 9.61835531321787e+00, 9.58025995135978e+00, 9.54108360842113e+00, 9.50109132318606e+00, 9.46055993405893e+00, 9.41977277631304e+00, 9.37901437968119e+00, 9.33856532322542e+00, 9.29869738461591e+00, 9.25966909641138e+00, 9.22172179520039e+00, 9.18507622288478e+00, 9.14992971487244e+00, 9.11645398881543e+00, 9.08479353046592e+00, 9.05506456033500e+00, 9.02735455577661e+00, 9.00172229722399e+00, 8.97819840376247e+00, 8.95678632119106e+00, 8.93746372443924e+00, 8.92018429503466e+00, 8.90487983280428e+00, 8.89146265885880e+00, 8.87982826405308e+00, 8.86985815358938e+00, 8.86142283442564e+00, 8.85438488796954e+00, 8.84860206656846e+00, 8.84393034898589e+00, 8.84022688784741e+00, 8.83735278138934e+00, 8.83517560313941e+00, 8.83357162669773e+00, 8.83242768873741e+00, 8.83164264172264e+00, 8.83112835850095e+00, 8.83081026354865e+00, 8.83062737977501e+00, 8.83053189484102e+00, 8.83048826627303e+00, 8.83047189957560e+00, 8.83046744742617e+00, 8.83046679029785e+00, 8.83046673758803e+00, 8.83046576482438e+00, 8.83045917420390e+00, 8.83043494633082e+00, 8.83037036329404e+00, 8.83022902275406e+00, 8.82995833049683e+00, 8.82948756868370e+00, 8.82872662341479e+00, 8.82756543957543e+00, 8.82587425324745e+00, 8.82350463216742e+00, 8.82029133273606e+00, 8.81605495794288e+00, 8.81060537445754e+00, 8.80374581949527e+00, 8.79527759964726e+00, 8.78500525578312e+00, 8.77274204182051e+00, 8.75831554233226e+00, 8.74157323649756e+00, 8.72238780567747e+00, 8.70066198058143e+00, 8.67633273284293e+00, 8.64937463544928e+00, 8.61980224664531e+00, 8.58767141146814e+00, 8.55307942180408e+00, 8.51616402675074e+00, 8.47710133646874e+00, 8.43610271073419e+00, 8.39341076441528e+00, 8.34929465315940e+00, 8.30404482186477e+00, 8.25796740552079e+00, 8.21137846757198e+00, 8.16459824705558e+00, 8.11794556508569e+00, 8.07173251677919e+00, 8.02625954921737e+00, 7.98181100173819e+00, 7.93865116323544e+00, 7.89702088290749e+00, 7.85713475609885e+00, 7.81917889511092e+00, 7.78330928549955e+00, 7.74965072076955e+00, 7.71829630196281e+00, 7.68930748301059e+00, 7.66271463761926e+00, 7.63851811872255e+00, 7.61668977704973e+00, 7.59717490102575e+00, 7.57989453592512e+00, 7.56474813583513e+00, 7.55161649746270e+00, 7.54036492011873e+00, 7.53084653140329e+00, 7.52290571337640e+00, 7.51638155962473e+00, 7.51111129001880e+00, 7.50693354755244e+00, 7.50369150094324e+00, 7.50123567808810e+00, 7.49942645936576e+00, 7.49813616637423e+00, 7.49725069103051e+00, 7.49667062189551e+00, 7.49631183877720e+00, 7.49610556259438e+00, 7.49599786449583e+00, 7.49594865558211e+00, 7.49593019549626e+00, 7.49592517389366e+00, 7.49592443271522e+00, 7.49592437554840e+00, 7.49592334901578e+00, 7.49591639410235e+00, 7.49589082704597e+00, 7.49582267420832e+00, 7.49567352109496e+00, 7.49538786624436e+00, 7.49489108253769e+00, 7.49408807408525e+00, 7.49286270029642e+00, 7.49107802003475e+00, 7.48857738782512e+00, 7.48518641084902e+00, 7.48071574995921e+00, 7.47496472038025e+00, 7.46772561862931e+00, 7.45878867234165e+00, 7.44794748034914e+00, 7.43500478313877e+00, 7.41977838058730e+00, 7.40210699662119e+00, 7.38185588108206e+00, 7.35892193911924e+00, 7.33323818880367e+00, 7.30477736846261e+00, 7.27355454565310e+00, 7.23962861801337e+00, 7.20310264003500e+00, 7.16412295626090e+00, 7.12287716767478e+00, 7.07959100156950e+00, 7.03452419399721e+00, 6.98796552670558e+00, 6.94022718650927e+00, 6.89163863393731e+00, 6.84254017943038e+00, 6.79327646896330e+00, 6.74419007630402e+00, 6.69561538591527e+00, 6.64787292897614e+00, 6.60126430617987e+00, 6.55606779684972e+00, 6.51253471733421e+00, 6.47088655584830e+00, 6.43131287901397e+00, 6.39396997970066e+00, 6.35898021764710e+00, 6.32643199383002e+00, 6.29638029565885e+00, 6.26884775115980e+00, 6.24382613443396e+00, 6.22127826998744e+00, 6.20114028853072e+00, 6.18332419047690e+00, 6.16772067504622e+00, 6.15420219242865e+00, 6.14262617400717e+00, 6.13283839157812e+00, 6.12467639135678e+00, 6.11797294296375e+00, 6.11255943824289e+00, 6.10826917037006e+00, 6.10494042094884e+00, 6.10241928225785e+00, 6.10056214399203e+00, 6.09923777904367e+00, 6.09832897120570e+00, 6.09773363903587e+00, 6.09736542414778e+00, 6.09715372833813e+00, 6.09704320149054e+00, 6.09699270027718e+00, 6.09697375543409e+00, 6.09696860196851e+00, 6.09696784132747e+00, 6.09696778373402e+00, 6.09696676347179e+00, 6.09695985104234e+00, 6.09693444017952e+00, 6.09686670383432e+00, 6.09671846290354e+00, 6.09643455767060e+00, 6.09594082494795e+00, 6.09514277021251e+00, 6.09392500823168e+00, 6.09215152766407e+00, 6.08966681466785e+00, 6.08629784743673e+00, 6.08185694768597e+00, 6.07614544652657e+00, 6.06895809132473e+00, 6.06008808789610e+00, 6.04933264005009e+00, 6.03649881785786e+00, 6.02140955923370e+00, 6.00390958889808e+00, 5.98387102696265e+00, 5.96119845843179e+00, 5.93583324649736e+00, 5.90775689740103e+00, 5.87699332250983e+00, 5.84360989242161e+00, 5.80771723532890e+00, 5.76946779322239e+00, 5.72905320965797e+00, 5.68670067637964e+00, 5.64266840831601e+00, 5.59724044399842e+00, 5.55072097997601e+00, 5.50342844428298e+00, 5.45568949837932e+00, 5.40783313340339e+00, 5.36018499946008e+00, 5.31306207975390e+00, 5.26676779711039e+00, 5.22158761981390e+00, 5.17778521654221e+00, 5.13559919563560e+00, 5.09524045099013e+00, 5.05689012476330e+00, 5.02069818551792e+00, 4.98678260952516e+00, 4.95522914308423e+00, 4.92609161533756e+00, 4.89939276448877e+00, 4.87512553565259e+00, 4.85325480560221e+00, 4.83371948802271e+00, 4.81643497198035e+00, 4.80129584559279e+00, 4.78817885581641e+00, 4.77694605349126e+00, 4.76744807014408e+00, 4.75952746961982e+00, 4.75302211368800e+00, 4.74776847683291e+00, 4.74360484209390e+00, 4.74037430773696e+00, 4.73792753435037e+00, 4.73612516420832e+00, 4.73483984981263e+00, 4.73395783657087e+00, 4.73338005552507e+00, 4.73302269559422e+00, 4.73281724040493e+00, 4.73270997175206e+00, 4.73266095924131e+00, 4.73264257286303e+00, 4.73263757131351e+00, 4.73263683309496e+00, 4.73263677767666e+00, 4.73263580225164e+00, 4.73262919360088e+00, 4.73260489945997e+00, 4.73254013988042e+00, 4.73239841350326e+00, 4.73212698432356e+00, 4.73165494726257e+00, 4.73089195768761e+00, 4.72972769484138e+00, 4.72803211186939e+00, 4.72565650559320e+00, 4.72243541719697e+00, 4.71818935050340e+00, 4.71272826765041e+00, 4.70585579318557e+00, 4.69737402769541e+00, 4.68708884231906e+00, 4.67481549748474e+00, 4.66038440487135e+00, 4.64364683298935e+00, 4.62448034586433e+00, 4.60279376273481e+00, 4.57853143551682e+00, 4.55167666036443e+00, 4.52225406944040e+00, 4.49033088764739e+00, 4.45601698449407e+00, 4.41946370092444e+00, 4.38086148201021e+00, 4.34043639606072e+00, 4.29844566626534e+00, 4.25517238005259e+00, 4.21091957189698e+00, 4.16600389573718e+00, 4.12074911242792e+00, 4.07547961531339e+00, 4.03051420341297e+00, 3.98616028798759e+00, 3.94270868632598e+00, 3.90042911901420e+00, 3.85956648668453e+00, 3.82033796231855e+00, 3.78293089839576e+00, 3.74750151679047e+00, 3.71417432485777e+00, 3.68304218432847e+00, 3.65416695038229e+00, 3.62758059584311e+00, 3.60328673860089e+00, 3.58126249757183e+00, 3.56146061212368e+00, 3.54381177033911e+00, 3.52822710139088e+00, 3.51460079556434e+00, 3.50281282133420e+00, 3.49273171198605e+00, 3.48421739452573e+00, 3.47712403131406e+00, 3.47130284055165e+00, 3.46660485617694e+00, 3.46288358182681e+00, 3.45999748819250e+00, 3.45781229929671e+00, 3.45620301171396e+00, 3.45505559214967e+00, 3.45426830342259e+00, 3.45375261680820e+00, 3.45343367964830e+00, 3.45325032059623e+00, 3.45315459010849e+00, 3.45311084993515e+00, 3.45309444146618e+00, 3.45308997796290e+00, 3.45308931915910e+00, 3.45308926985288e+00, 3.45308840401095e+00, 3.45308253780452e+00, 3.45306097300984e+00, 3.45300348919008e+00, 3.45287768728378e+00, 3.45263676168845e+00, 3.45221779039016e+00, 3.45154062230262e+00, 3.45050742891638e+00, 3.44900297328896e+00, 3.44689563287592e+00, 3.44403919324433e+00, 3.44027540697689e+00, 3.43543728598196e+00, 3.42935306627914e+00, 3.42185075292711e+00, 3.41276312043456e+00, 3.40193301260083e+00, 3.38921875755264e+00, 3.37449949132206e+00, 3.35768016922032e+00, 3.33869604082582e+00, 3.31751637339463e+00, 3.29414723086559e+00, 3.26863315126027e+00, 3.24105761287339e+00, 3.21154223670681e+00, 3.18024473553278e+00, 3.14735568436900e+00, 3.11309424816986e+00, 3.07770305539633e+00, 3.04144244660292e+00, 3.00458435211530e+00, 2.96740606054947e+00, 2.93018413027688e+00, 2.89318867056101e+00, 2.85667818099992e+00, 2.82089509116199e+00, 2.78606209145502e+00, 2.75237929583233e+00, 2.72002223087147e+00, 2.68914060709658e+00, 2.65985779906197e+00, 2.63227094141607e+00, 2.60645153862486e+00, 2.58244648513304e+00, 2.56027939881056e+00, 2.53995218163267e+00, 2.52144673569878e+00, 2.50472677808834e+00, 2.48973971312953e+00, 2.47641853420163e+00, 2.46468373833440e+00, 2.45444524507147e+00, 2.44560431609130e+00, 2.43805547395452e+00, 2.43168841731693e+00, 2.42638992644976e+00, 2.42204574753363e+00, 2.41854243764797e+00, 2.41576914545147e+00, 2.41361929605515e+00, 2.41199214331574e+00, 2.41079414941914e+00, 2.40994015072664e+00, 2.40935427075767e+00, 2.40897054597873e+00, 2.40873323759333e+00, 2.40859681237548e+00, 2.40852558714372e+00, 2.40849304398151e+00, 2.40848083595578e+00, 2.40847751508036e+00, 2.40847702492624e+00, 2.40847698822838e+00, 2.40847634361210e+00, 2.40847197624411e+00, 2.40845592136962e+00, 2.40841312521622e+00, 2.40831946813691e+00, 2.40814010843755e+00, 2.40782821547493e+00, 2.40732415396995e+00, 2.40655517365416e+00, 2.40543564887875e+00, 2.40386790103412e+00, 2.40174362189822e+00, 2.39894589780677e+00, 2.39535181255598e+00, 2.39083558137917e+00, 2.38527213992323e+00, 2.37854108227193e+00, 2.37053081273900e+00, 2.36114274993885e+00, 2.35029540141185e+00, 2.33792811572355e+00, 2.32400431899657e+00, 2.30851405602425e+00, 2.29147568307342e+00, 2.27293659942635e+00, 2.25297295535789e+00, 2.23168833191978e+00, 2.20921144787470e+00, 2.18569300610650e+00, 2.16130184064260e+00, 2.13622056163966e+00, 2.11064091622846e+00, 2.08475908666153e+00, 2.05877113431372e+00, 2.03286877105215e+00, 2.00723560196951e+00, 1.98204393990060e+00, 1.95745224710880e+00, 1.93360321719751e+00, 1.91062247394840e+00, 1.88861783555015e+00, 1.86767907349142e+00, 1.84787808510260e+00, 1.82926939634930e+00, 1.81189091546716e+00, 1.79576486659490e+00, 1.78089884392631e+00, 1.76728693946629e+00, 1.75491090993724e+00, 1.74374135977498e+00, 1.73373892682883e+00, 1.72485546497256e+00, 1.71703522319343e+00, 1.71021602387285e+00, 1.70433044403721e+00, 1.69930700256034e+00, 1.69507135391757e+00, 1.69154748546090e+00, 1.68865891068259e+00, 1.68632984598491e+00, 1.68448635352794e+00, 1.68305742825563e+00, 1.68197600365639e+00, 1.68117984860945e+00, 1.68061232712637e+00, 1.68022299412479e+00, 1.67996800362391e+00, 1.67981031082304e+00, 1.67971965614738e+00, 1.67967232711449e+00, 1.67965070228884e+00, 1.67964259010259e+00, 1.67964038339430e+00, 1.67964005768889e+00, 1.67964003323862e+00, 1.67963960289342e+00, 1.67963668724310e+00, 1.67962596903035e+00, 1.67959739841197e+00, 1.67953487333555e+00, 1.67941513425445e+00, 1.67920691925384e+00, 1.67887042138447e+00, 1.67835708493308e+00, 1.67760977113398e+00, 1.67656331596542e+00, 1.67514549266872e+00, 1.67327837905095e+00, 1.67088011426397e+00, 1.66786701171154e+00, 1.66415597458349e+00, 1.65966713932708e+00, 1.65432665172308e+00, 1.64806946214615e+00, 1.64084201329331e+00, 1.63260468736167e+00, 1.62333388216449e+00, 1.61302359808573e+00, 1.60168644015407e+00, 1.58935397074119e+00, 1.57607638613224e+00, 1.56192153117098e+00, 1.54697330644954e+00, 1.53132955813251e+00, 1.51509956802654e+00, 1.49840127849106e+00, 1.48135839214123e+00, 1.46409748038515e+00, 1.44674521934127e+00, 1.42942584924064e+00, 1.41225892716444e+00, 1.39535741601589e+00, 1.37882612764795e+00, 1.36276051699611e+00, 1.34724580797208e+00, 1.33235642101760e+00, 1.31815566619964e+00, 1.30469566372889e+00, 1.29201745478483e+00, 1.28015126852617e+00, 1.26911691528219e+00, 1.25892428049197e+00, 1.24957389852546e+00, 1.24105758979713e+00, 1.23335914840508e+00, 1.22645507079916e+00, 1.22031531863186e+00, 1.21490411091389e+00, 1.21018074182281e+00, 1.20610042095217e+00, 1.20261513242249e+00, 1.19967450813528e+00, 1.19722670862702e+00, 1.19521930263010e+00, 1.19360013379159e+00, 1.19231816031078e+00, 1.19132425083272e+00, 1.19057191808066e+00, 1.19001797070313e+00, 1.18962306386775e+00, 1.18935213039225e+00, 1.18917467669650e+00, 1.18906493151429e+00, 1.18900183994097e+00, 1.18896890074641e+00, 1.18895385062317e+00, 1.18894820481316e+00, 1.18894666901696e+00, 1.18894644233669e+00, 1.18894642524764e+00, 1.18894612350053e+00, 1.18894407912024e+00, 1.18893656377366e+00, 1.18891653070820e+00, 1.18887268928748e+00, 1.18878872945791e+00, 1.18864272804170e+00, 1.18840676552024e+00, 1.18804677817354e+00, 1.18752266587054e+00, 1.18678867019034e+00, 1.18579403060921e+00, 1.18448391797915e+00, 1.18280063434764e+00, 1.18068505642442e+00, 1.17807828706156e+00, 1.17492346565130e+00, 1.17116767533593e+00, 1.16676387357712e+00, 1.16167276427763e+00, 1.15586452556220e+00, 1.14932030852605e+00, 1.14203342931080e+00, 1.13401018971910e+00, 1.12527027950317e+00, 1.11584673509361e+00, 1.10578545304369e+00, 1.09514427980709e+00, 1.08399172069494e+00, 1.07240532839004e+00, 1.06046984421507e+00, 1.04827517306185e+00, 1.03591427565120e+00, 1.02348106014357e+00, 1.01106834976421e+00, 9.98765994729954e-01, 9.86659185942422e-01, 9.74827015126433e-01, 9.63341311823054e-01, 9.52265772509454e-01, 9.41655381936750e-01, 9.31556112575132e-01, 9.22004875927000e-01, 9.13029690381720e-01, 9.04650024879466e-01, 8.96877276098937e-01, 8.89715338863192e-01, 8.83161234238669e-01, 8.77205766409155e-01, 8.71834186805371e-01, 8.67026851235497e-01, 8.62759862176578e-01, 8.59005693490374e-01, 8.55733798400038e-01, 8.52911203584005e-01, 8.50503092831531e-01, 8.48473383069585e-01, 8.46785293974866e-01, 8.45401910111036e-01, 8.44286731873046e-01, 8.43404208768197e-01, 8.42720245998443e-01, 8.42202673190855e-01, 8.41821662679864e-01, 8.41550084152656e-01, 8.41363782839169e-01, 8.41241769792932e-01, 8.41166315116215e-01, 8.41122938097218e-01, 8.41100291943258e-01, 8.41089944856142e-01, 8.41086063327198e-01, 8.41085007459160e-01, 8.41084851615321e-01, 8.41084839802575e-01, 8.41084630372894e-01, 8.41083211457116e-01, 8.41077995389321e-01, 8.41064091395976e-01, 8.41033663497708e-01, 8.40975392927314e-01, 8.40874067723792e-01, 8.40710320218539e-01, 8.40460530812351e-01, 8.40096913547937e-01, 8.39587795248478e-01, 8.38898095139573e-01, 8.37990005661798e-01, 8.36823867483083e-01, 8.35359222573334e-01, 8.33556018873056e-01, 8.31375929097197e-01, 8.28783735351575e-01, 8.25748721515534e-01, 8.22246007920981e-01, 8.18257758921400e-01, 8.13774194553368e-01, 8.08794343405545e-01, 8.03326485315868e-01, 7.97388249295526e-01, 7.91006353116241e-01, 7.84215994594160e-01, 7.77059928487541e-01, 7.69587284499204e-01, 7.61852198578730e-01, 7.53912339468240e-01, 7.45827414040942e-01, 7.37657728447616e-01, 7.29462868673441e-01, 7.21300546109982e-01, 7.13225634020887e-01, 7.05289402143794e-01, 6.97538941355742e-01, 6.90016759620043e-01, 6.82760524592647e-01, 6.75802926707188e-01, 6.69171638173056e-01, 6.62889346841325e-01, 6.56973848180954e-01, 6.51438182820023e-01, 6.46290810743142e-01, 6.41535816092114e-01, 6.37173138596415e-01, 6.33198829078378e-01, 6.29605327385715e-01, 6.26381761633921e-01, 6.23514267886918e-01, 6.20986329415215e-01, 6.18779134463797e-01, 6.16871951035774e-01, 6.15242516550275e-01, 6.13867439373502e-01, 6.12722608181439e-01, 6.11783603946671e-01, 6.11026108132169e-01, 6.10426299522903e-01, 6.09961231147471e-01, 6.09609178053168e-01, 6.09349946407008e-01, 6.09165134586817e-01, 6.09038337652592e-01, 6.08955287858404e-01, 6.08903925643506e-01, 6.08874397746964e-01, 6.08858981604104e-01, 6.08851937859416e-01, 6.08849295510566e-01, 6.08848576727892e-01, 6.08848470637085e-01, 6.08848462544522e-01, 6.08848318397264e-01, 6.08847341778688e-01, 6.08843751630230e-01, 6.08834181648909e-01, 6.08813238163261e-01, 6.08773129589203e-01, 6.08703382652280e-01, 6.08590658989186e-01, 6.08418683892262e-01, 6.08168296740564e-01, 6.07817629942614e-01, 6.07342419863846e-01, 6.06716449115758e-01, 6.05912114695535e-01, 6.04901110837552e-01, 6.03655209258019e-01, 6.02147113081028e-01, 6.00351354616981e-01, 5.98245201928400e-01, 5.95809535420001e-01, 5.93029654143958e-01, 5.89895972581997e-01, 5.86404572565796e-01, 5.82557581614017e-01, 5.78363357844713e-01, 5.73836472014035e-01, 5.68997488194793e-01, 5.63872555153200e-01, 5.58492829725513e-01, 5.52893760779504e-01, 5.47114267297724e-01, 5.41195846657647e-01, 5.35181649459126e-01, 5.29115555566507e-01, 5.23041282769423e-01, 5.17001555015423e-01, 5.11037351909192e-01, 5.05187255465842e-01, 4.99486904290162e-01, 4.93968559752085e-01, 4.88660783633668e-01, 4.83588222382849e-01, 4.78771489705538e-01, 4.74227136860387e-01, 4.69967698699493e-01, 4.66001803147581e-01, 4.62334332284144e-01, 4.58966624292991e-01, 4.55896707053604e-01, 4.53119555854315e-01, 4.50627369413276e-01, 4.48409859938993e-01, 4.46454554225255e-01, 4.44747103676381e-01, 4.43271601657235e-01, 4.42010906657114e-01, 4.40946969477510e-01, 4.40061162061035e-01, 4.39334604754743e-01, 4.38748487847496e-01, 4.38284382249741e-01, 4.37924533313128e-01, 4.37652131128592e-01, 4.37451550294471e-01, 4.37308552186952e-01, 4.37210443239147e-01, 4.37146183651468e-01, 4.37106442285589e-01, 4.37083595171334e-01, 4.37071666983899e-01, 4.37066216911433e-01, 4.37064172403502e-01, 4.37063616247997e-01, 4.37063534160622e-01, 4.37063527854841e-01, 4.37063414955096e-01, 4.37062650042966e-01, 4.37059838149082e-01, 4.37052342703232e-01, 4.37035939259916e-01, 4.37004525303635e-01, 4.36949898047859e-01, 4.36861610921438e-01, 4.36726918004226e-01, 4.36530813955741e-01, 4.36256174879489e-01, 4.35884002978540e-01, 4.35393774701112e-01, 4.34763888310682e-01, 4.33972202477637e-01, 4.32996652707443e-01, 4.31815927428302e-01, 4.30410180700326e-01, 4.28761754204759e-01, 4.26855877903316e-01, 4.24681316996187e-01, 4.22230932956230e-01, 4.19502128742408e-01, 4.16497152869235e-01, 4.13223243669163e-01, 4.09692603429368e-01, 4.05922201496145e-01, 4.01933415147434e-01, 3.97751526195954e-01, 3.93405099093288e-01, 3.88925272082325e-01, 3.84344996231398e-01, 3.79698257788714e-01, 3.75019317319815e-01, 3.70341994897320e-01, 3.65699024772926e-01, 3.61121496171274e-01, 3.56638389826934e-01, 3.52276213296893e-01, 3.48058732440915e-01, 3.44006792105391e-01, 3.40138216113916e-01, 3.36467775124973e-01, 3.33007210594060e-01, 3.29765303716876e-01, 3.26747979536291e-01, 3.23958438076639e-01, 3.21397306168695e-01, 3.19062805346379e-01, 3.16950932692441e-01, 3.15055652700543e-01, 3.13369099068536e-01, 3.11881785837770e-01, 3.10582827466085e-01, 3.09460167304254e-01, 3.08500813580270e-01, 3.07691081438728e-01, 3.07016838886465e-01, 3.06463753725277e-01, 3.06017537769328e-01, 3.05664183919468e-01, 3.05390191064333e-01, 3.05182771367029e-01, 3.05030034331386e-01, 3.04921142168954e-01, 3.04846431431114e-01, 3.04797496631696e-01, 3.04767232641669e-01, 3.04749833938800e-01, 3.04740750272879e-01, 3.04736599876581e-01, 3.04735042920569e-01, 3.04734619390867e-01, 3.04734556878778e-01, 3.04734552040077e-01, 3.04734464930194e-01, 3.04733874748052e-01, 3.04731705178737e-01, 3.04725921924126e-01, 3.04713265517350e-01, 3.04689027384661e-01, 3.04646878320916e-01, 3.04578757588571e-01, 3.04474829725175e-01, 3.04323514797531e-01, 3.04111596193287e-01, 3.03824408014677e-01, 3.03446101666129e-01, 3.02959988278785e-01, 3.02348950218537e-01, 3.01595911187412e-01, 3.00684350548929e-01, 2.99598843772774e-01, 2.98325607665351e-01, 2.96853026730519e-01, 2.95172135977769e-01, 2.93277036082538e-01, 2.91165219180170e-01, 2.88837787735602e-01, 2.86299554640365e-01, 2.83559019503796e-01, 2.80628223406261e-01, 2.77522491456947e-01, 2.74260078643435e-01, 2.70861739086241e-01, 2.67350241530865e-01, 2.63749854597979e-01, 2.60085824103975e-01, 2.56383862026739e-01, 2.52669662941820e-01, 2.48968459563327e-01, 2.45304624921413e-01, 2.41701325097125e-01, 2.38180223566570e-01, 2.34761236157965e-01, 2.31462334347221e-01, 2.28299393972852e-01, 2.25286086254981e-01, 2.22433808078954e-01, 2.19751648703392e-01, 2.17246390272571e-01, 2.14922539691949e-01, 2.12782389566870e-01, 2.10826105985097e-01, 2.09051840981965e-01, 2.07455867554225e-01, 2.06032735090727e-01, 2.04775443053599e-01, 2.03675630668947e-01, 2.02723780235219e-01, 2.01909431439958e-01, 2.01221403785249e-01, 2.00648023843410e-01, 2.00177353672958e-01, 1.99797416294626e-01, 1.99496413758023e-01, 1.99262933033905e-01, 1.99086134829432e-01, 1.98955920451032e-01, 1.98863072108844e-01, 1.98799362534078e-01, 1.98757630532883e-01, 1.98731820021255e-01, 1.98716981222049e-01, 1.98709233939649e-01, 1.98705694121173e-01, 1.98704366209650e-01, 1.98704004985158e-01, 1.98703951669174e-01, 1.98703947508726e-01, 1.98703872179074e-01, 1.98703361808676e-01, 1.98701485632679e-01, 1.98696484433594e-01, 1.98685539415647e-01, 1.98664578335856e-01, 1.98628126650464e-01, 1.98569210520651e-01, 1.98479317447164e-01, 1.98348419053625e-01, 1.98165059082100e-01, 1.97916507934824e-01, 1.97588983082019e-01, 1.97167932366963e-01, 1.96638374643582e-01, 1.95985289452508e-01, 1.95194044625564e-01, 1.94250848047451e-01, 1.93143207531147e-01, 1.91860381066965e-01, 1.90393798908502e-01, 1.88737439162260e-01, 1.86888139882909e-01, 1.84845833118624e-01, 1.82613689752955e-01, 1.80198168090648e-01, 1.77608963639742e-01, 1.74858862067873e-01, 1.71963501544161e-01, 1.68941054353180e-01, 1.65811840619495e-01, 1.62597889116538e-01, 1.59322461494067e-01, 1.56009556862674e-01, 1.52683413601778e-01, 1.49368024523781e-01, 1.46086680213793e-01, 1.42861553388975e-01, 1.39713334766094e-01, 1.36660928077742e-01, 1.33721208884314e-01, 1.30908848838548e-01, 1.28236204156064e-01, 1.25713264566938e-01, 1.23347656992127e-01, 1.21144696787891e-01, 1.19107478659865e-01, 1.17236999312809e-01, 1.15532304111403e-01, 1.13990651186374e-01, 1.12607687262035e-01, 1.11377630800101e-01, 1.10293459211318e-01, 1.09347097855785e-01, 1.08529609412676e-01, 1.07831382518134e-01, 1.07242319237924e-01, 1.06752020059931e-01, 1.06349965724281e-01, 1.06025693967730e-01, 1.05768969225103e-01, 1.05569942253810e-01, 1.05419296743817e-01, 1.05308379247424e-01, 1.05229308587239e-01, 1.05175061595393e-01, 1.05139531813490e-01, 1.05117558788113e-01, 1.05104926761488e-01, 1.05098331765950e-01, 1.05095318471932e-01, 1.05094188085879e-01, 1.05093880591124e-01, 1.05093835208450e-01, 1.05093831643911e-01, 1.05093766591029e-01, 1.05093325814719e-01, 1.05091705524490e-01, 1.05087386456804e-01, 1.05077934408760e-01, 1.05059833202397e-01, 1.05028356837080e-01, 1.04977487515201e-01, 1.04899884633591e-01, 1.04786909878898e-01, 1.04628711371937e-01, 1.04414369106928e-01, 1.04132102001051e-01, 1.03769534110656e-01, 1.03314015990085e-01, 1.02752993246325e-01, 1.02074412314961e-01, 1.01267149604755e-01, 1.00321448323696e-01, 9.92293441409249e-02, 9.79850602727694e-02, 9.65853516548485e-02, 9.50297789866319e-02, 9.33208954186302e-02, 9.14643324694762e-02, 8.94376102057244e-02, 8.73176747575819e-02, 8.50845276584256e-02, 8.27547215977038e-02, 8.03467041442083e-02, 7.78803197991422e-02, 7.53762705388947e-02, 7.28555612062696e-02, 7.03389563162826e-02, 6.78464731059236e-02, 6.53969321645995e-02, 6.30075821020206e-02, 6.06938089468780e-02, 5.84689348997482e-02, 5.63441052625731e-02, 5.43282573326813e-02, 5.24281611395843e-02, 5.06485193021398e-02, 4.89921120093406e-02, 4.74599730615335e-02, 4.60515838429263e-02, 4.47650737739735e-02, 4.35974179392336e-02, 4.25446249379906e-02, 4.16019103310052e-02, 4.07638531812271e-02, 4.00245349938544e-02, 3.93776617916810e-02, 3.88166710955531e-02, 3.83348262237977e-02, 3.79253006030336e-02, 3.75812547341887e-02, 3.72959081318225e-02, 3.70626080129171e-02, 3.68748958203859e-02, 3.67265718982193e-02, 3.66117578621221e-02, 3.65249555007168e-02, 3.64611004618371e-02, 3.64156085777427e-02, 3.63844124995278e-02, 3.63639863613638e-02, 3.63513564750075e-02, 3.63440965385973e-02, 3.63403064859096e-02, 3.63385748453862e-02, 3.63379252578948e-02, 3.63377485553560e-02, 3.63377224744674e-02, 3.63377204038464e-02, 3.63376824570878e-02, 3.63374253628113e-02, 3.63364802660477e-02, 3.63339610541367e-02, 3.63284481709893e-02, 3.63178916833338e-02, 3.62995380656805e-02, 3.62698850246197e-02, 3.62246684829657e-02, 3.61588855104831e-02, 3.60668564215842e-02, 3.59423283548335e-02, 3.57786213120583e-02, 3.55688158066592e-02, 3.53059789493404e-02, 3.49834230561878e-02, 3.45949878573659e-02, 3.41353343598616e-02, 3.36002356946157e-02, 3.29868482238496e-02, 3.22939451695472e-02, 3.15220953718516e-02, 3.06737717113830e-02, 2.97533772830718e-02, 2.87671824350912e-02, 2.77231719099937e-02, 2.66308079810873e-02, 2.55007219786023e-02, 2.43443522378346e-02, 2.31735506388357e-02, 2.20001820625573e-02, 2.08357409940252e-02, 1.96910071600271e-02, 1.85757578030579e-02, 1.74985485620560e-02, 1.64665687318521e-02, 1.54855706800365e-02, 1.45598680046444e-02, 1.36923929720663e-02, 1.28848010575808e-02, 1.21376091039362e-02, 1.14503536983346e-02, 1.08217576476370e-02, 1.02498945004612e-02, 9.73234343044460e-03, 9.26632907652447e-03, 8.84884297644106e-03, 8.47674503148710e-03, 8.14684501141194e-03, 7.85596537758870e-03, 7.60098756681771e-03, 7.37888430951372e-03, 7.18674064260405e-03, 7.02176616751160e-03, 6.88130091858952e-03, 6.76281699207770e-03, 6.66391783020259e-03, 6.58233673899744e-03, 6.51593584501242e-03, 6.46270629038881e-03, 6.42077006262376e-03, 6.38838348132076e-03, 6.36394203800172e-03, 6.34598601821068e-03, 6.33320613793309e-03, 6.32444830866241e-03, 6.31871661409633e-03, 6.31517363768985e-03, 6.31313741809691e-03, 6.31207451360538e-03, 6.31158890614891e-03, 6.31140674488329e-03, 6.31135719325115e-03, 6.31134987955662e-03, 6.31134929358446e-03, 6.31133848791343e-03, 6.31126527825465e-03, 6.31099615735486e-03, 6.31027882251622e-03, 6.30870916926618e-03, 6.30570392498342e-03, 6.30048037487185e-03, 6.29204469281142e-03, 6.27919043761516e-03, 6.26050876400797e-03, 6.23441176722039e-03, 6.19917006583570e-03, 6.15296517214681e-03, 6.09395636587351e-03, 6.02036067564882e-03, 5.93054323260134e-03, 5.82311379782574e-03, 5.69702384141380e-03, 5.55165736886565e-03, 5.38690797443664e-03, 5.20323456636973e-03, 5.00168902772202e-03, 4.78391083349727e-03, 4.55208628262582e-03, 4.30887327465435e-03, 4.05729601934723e-03, 3.80061714650637e-03, 3.54219686556689e-03, 3.28664672745838e-03, 3.03453224755934e-03, 2.78994771402097e-03, 2.55534418811187e-03, 2.33272979336618e-03, 2.12363105480939e-03, 1.92908318607898e-03, 1.74965364133328e-03, 1.58549534827657e-03, 1.43641640643644e-03, 1.30195136341271e-03, 1.18142451776748e-03, 1.07400299838645e-03, 9.78744078587113e-04, 8.94642749388552e-04, 8.20679121178003e-04, 7.55858705833390e-04, 6.99239898596105e-04, 6.49948219496821e-04, 6.07180438207067e-04, 5.70202971770684e-04, 5.38348598967175e-04, 5.11013697057194e-04, 4.87655902148276e-04, 4.67791064449941e-04, 4.50988889236115e-04, 4.36867514750361e-04, 4.25087659644774e-04, 4.15346929793614e-04, 4.07374685055473e-04, 4.00927679402585e-04, 3.95786543476717e-04, 3.91753082539026e-04, 3.88648306060446e-04, 3.86311074917600e-04, 3.84597235477457e-04, 3.83379100169279e-04, 3.82545129957279e-04, 3.81999676147160e-04, 3.81662649324710e-04, 3.81469001206725e-04, 3.81367930998522e-04, 3.81321758387898e-04, 3.81304438615615e-04, 3.81299727324702e-04, 3.81299031952054e-04, 3.81298975723254e-04, 3.81297932335656e-04, 3.81290863288086e-04, 3.81264877677978e-04, 3.81195616890651e-04, 3.81044078421766e-04, 3.80754006137352e-04, 3.80250012211647e-04, 3.79436615737916e-04, 3.78198402119168e-04, 3.76401517362623e-04, 3.73896702408061e-04, 3.70524033526110e-04, 3.66119455728509e-04, 3.60523070385810e-04, 3.53588963230678e-04, 3.45196135558723e-04, 3.35259836342416e-04, 3.23742304336960e-04, 3.10661664933320e-04, 2.96097572427101e-04, 2.80192260591505e-04, 2.63146108290258e-04, 2.45207841789608e-04, 2.26661082660229e-04, 2.07810167276890e-04, 1.88967391906658e-04, 1.70441314576664e-04, 1.52524283437927e-04, 1.35478405299594e-04, 1.19520566286724e-04, 1.04807470625801e-04, 9.14242336396143e-05, 7.93838736381502e-05, 6.86395821620778e-05, 5.91024742000457e-05, 5.06614642947767e-05, 4.32083328814059e-05, 3.66632740309569e-05, 3.09834380889384e-05, 2.61398449055150e-05, 2.20722336052859e-05, 1.86553486631880e-05, 1.56679806604387e-05, 1.28860938186648e-05, 1.02453517794032e-05, 7.74879763909553e-06, 5.45924740968928e-06, 3.49217265416257e-06, 1.95247898560069e-06, 8.96781228831856e-07, 3.51476293403152e-07, 3.23232294263676e-07, 4.61827417197469e-07, 5.62756770960959e-07, 5.68030364078709e-07, 5.14688373698391e-07, 4.41559491193289e-07, 3.73587283775123e-07, 3.19986074970302e-07, 2.79961545501806e-07, 1.77635683940025e-15, 3.55271367880050e-15, 0.00000000000000e+00, 1.77635683940025e-15, 3.55271367880050e-15, 1.77635683940025e-15, 3.55271367880050e-15, 1.77635683940025e-15, 0.00000000000000e+00, 3.55271367880050e-15, 1.77635683940025e-15, 0.00000000000000e+00, 3.55271367880050e-15, 3.55271367880050e-15],
[ 1.20000000000000e+01, 1.19999999943145e+01, 1.19999998185066e+01, 1.19999986273805e+01, 1.19999942486486e+01, 1.19999825763558e+01, 1.19999570308059e+01, 1.19999081041036e+01, 1.19998230078950e+01, 1.19996854368675e+01, 1.19994754581318e+01, 1.19991695328715e+01, 1.19987406727273e+01, 1.19981587294065e+01, 1.19973908121117e+01, 1.19964018236746e+01, 1.19951551028874e+01, 1.19936131575264e+01, 1.19917384700432e+01, 1.19894943559037e+01, 1.19868458530999e+01, 1.19837606204625e+01, 1.19802098220295e+01, 1.19761689748583e+01, 1.19716187382697e+01, 1.19665456235456e+01, 1.19609426045505e+01, 1.19548096115842e+01, 1.19481538930079e+01, 1.19409902318058e+01, 1.19333410072635e+01, 1.19252360953422e+01, 1.19167126050826e+01, 1.19078144524263e+01, 1.18985917771095e+01, 1.18891002126323e+01, 1.18794000236049e+01, 1.18695551288240e+01, 1.18596320320826e+01, 1.18496986857761e+01, 1.18398233146914e+01, 1.18300732288395e+01, 1.18205136547338e+01, 1.18112066141092e+01, 1.18022098777299e+01, 1.17935760197272e+01, 1.17853515949286e+01, 1.17775764580272e+01, 1.17702832393326e+01, 1.17634969873969e+01, 1.17572349841760e+01, 1.17515067337038e+01, 1.17463141206641e+01, 1.17416517308613e+01, 1.17375073215139e+01, 1.17338624256268e+01, 1.17306930715099e+01, 1.17279705958689e+01, 1.17256625268600e+01, 1.17237335121065e+01, 1.17221462659565e+01, 1.17208625102295e+01, 1.17198438833518e+01, 1.17190527940974e+01, 1.17184531981003e+01, 1.17180112778198e+01, 1.17176960096674e+01, 1.17174796054450e+01, 1.17173378190113e+01, 1.17172501130759e+01, 1.17171996851174e+01, 1.17171733555089e+01, 1.17171613249121e+01, 1.17171568117582e+01, 1.17171555840651e+01, 1.17171554028599e+01, 1.17171553790896e+01, 1.17171548252702e+01, 1.17171510730568e+01, 1.17171372794423e+01, 1.17171005101083e+01, 1.17170200381630e+01, 1.17168659126625e+01, 1.17165978492634e+01, 1.17161644858847e+01, 1.17155030355170e+01, 1.17145393566372e+01, 1.17131884494777e+01, 1.17113553740453e+01, 1.17089365737367e+01, 1.17058215770308e+01, 1.17018950394102e+01, 1.16970390786321e+01, 1.16911358488779e+01, 1.16840702931797e+01, 1.16757330087550e+01, 1.16660231562894e+01, 1.16548513415700e+01, 1.16421423960219e+01, 1.16278379815783e+01, 1.16118989450543e+01, 1.15943073481576e+01, 1.15750681019799e+01, 1.15542101399557e+01, 1.15317870714842e+01, 1.15078772701779e+01, 1.14825833661757e+01, 1.14560311308705e+01, 1.14283677639570e+01, 1.13997596156931e+01, 1.13703894001198e+01, 1.13404529760057e+01, 1.13101557898612e+01, 1.12797090881825e+01, 1.12493260132884e+01, 1.12192176983999e+01, 1.11895894732440e+01, 1.11606372821846e+01, 1.11325444037842e+01, 1.11054785450580e+01, 1.10795893667958e+01, 1.10550064793561e+01, 1.10318379322286e+01, 1.10101692060688e+01, 1.09900627031953e+01, 1.09715577218160e+01, 1.09546708903926e+01, 1.09393970313357e+01, 1.09257104173049e+01, 1.09135663784605e+01, 1.09029032147773e+01, 1.08936443637896e+01, 1.08857007707729e+01, 1.08789734053923e+01, 1.08733558663384e+01, 1.08687370136154e+01, 1.08650035671581e+01, 1.08620426105889e+01, 1.08597439404083e+01, 1.08580022039454e+01, 1.08567187740898e+01, 1.08558033152200e+01, 1.08551750027627e+01, 1.08547633682892e+01, 1.08545087527093e+01, 1.08543623616030e+01, 1.08542859286462e+01, 1.08542510049735e+01, 1.08542379037650e+01, 1.08542343399060e+01, 1.08542338138874e+01, 1.08542337674500e+01, 1.08542328575313e+01, 1.08542266926930e+01, 1.08542040300135e+01, 1.08541436190036e+01, 1.08540114078039e+01, 1.08537581953470e+01, 1.08533178186066e+01, 1.08526059502873e+01, 1.08515195665820e+01, 1.08499371271534e+01, 1.08477194905141e+01, 1.08447115677202e+01, 1.08407446960535e+01, 1.08356396925311e+01, 1.08292105250982e+01, 1.08212685177830e+01, 1.08116269855303e+01, 1.08001061755607e+01, 1.07865383756126e+01, 1.07707730360628e+01, 1.07526817434380e+01, 1.07321628780441e+01, 1.07091457891763e+01, 1.06835943284690e+01, 1.06555095961131e+01, 1.06249317763851e+01, 1.05919409681917e+01, 1.05566569525181e+01, 1.05192378803222e+01, 1.04798779093173e+01, 1.04388038632647e+01, 1.03962710295254e+01, 1.03525582463574e+01, 1.03079624579322e+01, 1.02627929303396e+01, 1.02173653251783e+01, 1.01719958192169e+01, 1.01269954407204e+01, 1.00826647678991e+01, 1.00392891055127e+01, 9.99713422492231e+00, 9.95644272341339e+00, 9.91743103238318e+00, 9.88028708218365e+00, 9.84516861450050e+00, 9.81220212095054e+00, 9.78148237846262e+00, 9.75307254707371e+00, 9.72700479303096e+00, 9.70328139859799e+00, 9.68187631891097e+00, 9.66273714499499e+00, 9.64578743021531e+00, 9.63092933478077e+00, 9.61804653939541e+00, 9.60700737488488e+00, 9.59766810985229e+00, 9.58987633349675e+00, 9.58347436609059e+00, 9.57830262574218e+00, 9.57420287747363e+00, 9.57102128979930e+00, 9.56861122532615e+00, 9.56683569573287e+00, 9.56556941800649e+00, 9.56470041804199e+00, 9.56413113947635e+00, 9.56377902958928e+00, 9.56357658974473e+00, 9.56347089452944e+00, 9.56342260074255e+00, 9.56340448395121e+00, 9.56339955573294e+00, 9.56339882833782e+00, 9.56339876999173e+00, 9.56339769321038e+00, 9.56339039785439e+00, 9.56336357929554e+00, 9.56329209041460e+00, 9.56313563646633e+00, 9.56283599969803e+00, 9.56231490149005e+00, 9.56147259479853e+00, 9.56018726244250e+00, 9.55831527726246e+00, 9.55569235828202e+00, 9.55213563262558e+00, 9.54744658607565e+00, 9.54141485598415e+00, 9.53382278927185e+00, 9.52445065632058e+00, 9.51308237996124e+00, 9.49951160916327e+00, 9.48354794144544e+00, 9.46502307864725e+00, 9.44379668970727e+00, 9.41976175340774e+00, 9.39284916504457e+00, 9.36303141424303e+00, 9.33032517621201e+00, 9.29479270402193e+00, 9.25654196227964e+00, 9.21572549919826e+00, 9.17253811027207e+00, 9.12721339822749e+00, 9.08001937675826e+00, 9.03125329688099e+00, 8.98123589305410e+00, 8.93030525150045e+00, 8.87881049689019e+00, 8.82710547817020e+00, 8.77554261292137e+00, 8.72446702524428e+00, 8.67421108741775e+00, 8.62508945228352e+00, 8.57739464248307e+00, 8.53139324456609e+00, 8.48732274035635e+00, 8.44538899431712e+00, 8.40576440352752e+00, 8.36858670593644e+00, 8.33395843267326e+00, 8.30194698138308e+00, 8.27258527989826e+00, 8.24587300309461e+00, 8.22177830043933e+00, 8.20023998731437e+00, 8.18117014938765e+00, 8.16445710575074e+00, 8.14996867292595e+00, 8.13755566794832e+00, 8.12705558448969e+00, 8.11829637153301e+00, 8.11110023973121e+00, 8.10528741674829e+00, 8.10067977013906e+00, 8.09710421526875e+00, 8.09439582696282e+00, 8.09240057746662e+00, 8.09097763017742e+00, 8.09000112857672e+00, 8.08936143269806e+00, 8.08896577094733e+00, 8.08873829257402e+00, 8.08861952483600e+00, 8.08856525807029e+00, 8.08854490060270e+00, 8.08853936286561e+00, 8.08853854550678e+00, 8.08853848246419e+00, 8.08853735042170e+00, 8.08852968066314e+00, 8.08850148575832e+00, 8.08842632802496e+00, 8.08826184478451e+00, 8.08794683058424e+00, 8.08739898971491e+00, 8.08651345661548e+00, 8.08516216366049e+00, 8.08319411534611e+00, 8.08043660492629e+00, 8.07669738400442e+00, 8.07176776746751e+00, 8.06542662568165e+00, 8.05744518358668e+00, 8.04759251316195e+00, 8.03564157303383e+00, 8.02137561850227e+00, 8.00459477901935e+00, 7.98512258031793e+00, 7.96281217701871e+00, 7.93755206031923e+00, 7.90927101534178e+00, 7.87794212412426e+00, 7.84358564240501e+00, 7.80627061975226e+00, 7.76611518101547e+00, 7.72328543991937e+00, 7.67799307015700e+00, 7.63049161294701e+00, 7.58107165029397e+00, 7.53005501788543e+00, 7.47778826848744e+00, 7.42463562364842e+00, 7.37097166634905e+00, 7.31717402819637e+00, 7.26361631103251e+00, 7.21066145504999e+00, 7.15865572610047e+00, 7.10792344794330e+00, 7.05876255586701e+00, 7.01144100170537e+00, 6.96619400113837e+00, 6.92322208503960e+00, 6.88268989828303e+00, 6.84472568088835e+00, 6.80942136550766e+00, 6.77683322936627e+00, 6.74698304527966e+00, 6.71985968316914e+00, 6.69542111912461e+00, 6.67359681267824e+00, 6.65429041422915e+00, 6.63738276352734e+00, 6.62273513704399e+00, 6.61019269731969e+00, 6.59958809147287e+00, 6.59074513950869e+00, 6.58348254648821e+00, 6.57761756663147e+00, 6.57296954269080e+00, 6.56936324108162e+00, 6.56663190288170e+00, 6.56461993337852e+00, 6.56318515867037e+00, 6.56220058701606e+00, 6.56155562505062e+00, 6.56115671426997e+00, 6.56092737073798e+00, 6.56080763000536e+00, 6.56075291884990e+00, 6.56073239470541e+00, 6.56072681163092e+00, 6.56072598758047e+00, 6.56072592518586e+00, 6.56072481987141e+00, 6.56071733120043e+00, 6.56068980201009e+00, 6.56061641896044e+00, 6.56045582024853e+00, 6.56014824797762e+00, 6.55961335679676e+00, 6.55874877560500e+00, 6.55742950201913e+00, 6.55550818978592e+00, 6.55281636712699e+00, 6.54916659887380e+00, 6.54435557703169e+00, 6.53816809320409e+00, 6.53038181259181e+00, 6.52077273402855e+00, 6.50912118524244e+00, 6.49521816930659e+00, 6.47887184956361e+00, 6.45991393893265e+00, 6.43820574815027e+00, 6.41364364847694e+00, 6.38616371925238e+00, 6.35574537978080e+00, 6.32241384732411e+00, 6.28624131591141e+00, 6.24734681030296e+00, 6.20589473089503e+00, 6.16209216345062e+00, 6.11618507762368e+00, 6.06845357688899e+00, 6.01920638802415e+00, 5.96877479089544e+00, 5.91750619070481e+00, 5.86575752768347e+00, 5.81388870623696e+00, 5.76225620903504e+00, 5.71120704295303e+00, 5.66107314376329e+00, 5.61216634518029e+00, 5.56477399529148e+00, 5.51915527978736e+00, 5.47553828735897e+00, 5.43411782915769e+00, 5.39505400251587e+00, 5.35847147038155e+00, 5.32445941303033e+00, 5.29307209806277e+00, 5.26433000845765e+00, 5.23822146605026e+00, 5.21470468839936e+00, 5.19371021953984e+00, 5.17514367847866e+00, 5.15888877245540e+00, 5.14481052413665e+00, 5.13275866251158e+00, 5.12257112609329e+00, 5.11407762422081e+00, 5.10710319821001e+00, 5.10147171946094e+00, 5.09700925719648e+00, 5.09354724514729e+00, 5.09092537504765e+00, 5.08899414597804e+00, 5.08761700289688e+00, 5.08667200540060e+00, 5.08605297880441e+00, 5.08567011372106e+00, 5.08544999684508e+00, 5.08533507383130e+00, 5.08528256406224e+00, 5.08526286575497e+00, 5.08525750733022e+00, 5.08525671643768e+00, 5.08525665706512e+00, 5.08525561204079e+00, 5.08524853184501e+00, 5.08522250426391e+00, 5.08515312403951e+00, 5.08500128585083e+00, 5.08471049270967e+00, 5.08420478583828e+00, 5.08338738840345e+00, 5.08214013968164e+00, 5.08032377814824e+00, 5.07777911134623e+00, 5.07432908692055e+00, 5.06978175274778e+00, 5.06393406473233e+00, 5.05657646902634e+00, 5.04749815204527e+00, 5.03649281805786e+00, 5.02336482213726e+00, 5.00793545803234e+00, 4.99004917839851e+00, 4.96957951111871e+00, 4.94643443215339e+00, 4.92056096393506e+00, 4.89194878945018e+00, 4.86063270559897e+00, 4.82669378399681e+00, 4.79025916096997e+00, 4.75150043821154e+00, 4.71063073792904e+00, 4.66790051752035e+00, 4.62359230497270e+00, 4.57801456361264e+00, 4.53149493036420e+00, 4.48437309291010e+00, 4.43699357675804e+00, 4.38969870309041e+00, 4.34282195363362e+00, 4.29668194206178e+00, 4.25157714607447e+00, 4.20778150428700e+00, 4.16554093161321e+00, 4.12507075975074e+00, 4.08655406882603e+00, 4.05014084436879e+00, 4.01594787162390e+00, 3.98405926678490e+00, 3.95452754117817e+00, 3.92737509822414e+00, 3.90259607225560e+00, 3.88015843095645e+00, 3.86000627734974e+00, 3.84206230120004e+00, 3.82623034201970e+00, 3.81239803558377e+00, 3.80043952235550e+00, 3.79021819927577e+00, 3.78158949608945e+00, 3.77440365418824e+00, 3.76850848051002e+00, 3.76375204220490e+00, 3.75998526053614e+00, 3.75706435584716e+00, 3.75485309037369e+00, 3.75322475306341e+00, 3.75206383103466e+00, 3.75126731624092e+00, 3.75074560338396e+00, 3.75042294589344e+00, 3.75023745032562e+00, 3.75014060505274e+00, 3.75009635567752e+00, 3.75007975621405e+00, 3.75007524075796e+00, 3.75007457428612e+00, 3.75007452440601e+00, 3.75007364848629e+00, 3.75006771400233e+00, 3.75004589822719e+00, 3.74998774550051e+00, 3.74986048029047e+00, 3.74961675452625e+00, 3.74919292069727e+00, 3.74850791222245e+00, 3.74746280093403e+00, 3.74594108775293e+00, 3.74380976616046e+00, 3.74092117827532e+00, 3.73711565974062e+00, 3.73222494203040e+00, 3.72607624947960e+00, 3.71849699419490e+00, 3.70931993653926e+00, 3.69838864426771e+00, 3.68556305232695e+00, 3.67072490085026e+00, 3.65378281405043e+00, 3.63467678028854e+00, 3.61338180561723e+00, 3.58991054054567e+00, 3.56431472227437e+00, 3.53668533032645e+00, 3.50715141899018e+00, 3.47587766064802e+00, 3.44306070439052e+00, 3.40892451849485e+00, 3.37371493793471e+00, 3.33769367465404e+00, 3.30113206602759e+00, 3.26430483483877e+00, 3.22748411335731e+00, 3.19093394767742e+00, 3.15490545076532e+00, 3.11963271883665e+00, 3.08532957100373e+00, 3.05218712132888e+00, 3.02037214918199e+00, 2.99002620050781e+00, 2.96126533025556e+00, 2.93418038456481e+00, 2.90883771911196e+00, 2.88528025543939e+00, 2.86352878795445e+00, 2.84358346843244e+00, 2.82542541033028e+00, 2.80901837040258e+00, 2.79431047878870e+00, 2.78123600007973e+00, 2.76971711639658e+00, 2.75966572902465e+00, 2.75098527769369e+00, 2.74357257638402e+00, 2.73731966193724e+00, 2.73211564722414e+00, 2.72784856473551e+00, 2.72440717984862e+00, 2.72168274636661e+00, 2.71957067092908e+00, 2.71797204822144e+00, 2.71679502616426e+00, 2.71595595990449e+00, 2.71538031574428e+00, 2.71500329119422e+00, 2.71477012496225e+00, 2.71463608049131e+00, 2.71456609805557e+00, 2.71453412269864e+00, 2.71452212767005e+00, 2.71451886473442e+00, 2.71451838313207e+00, 2.71451834707448e+00, 2.71451771370489e+00, 2.71451342253456e+00, 2.71449764776683e+00, 2.71445559824522e+00, 2.71436357499913e+00, 2.71418734370221e+00, 2.71388088922908e+00, 2.71338561284870e+00, 2.71263002476538e+00, 2.71152997525079e+00, 2.70998945590007e+00, 2.70790198795271e+00, 2.70515259661589e+00, 2.70162034866134e+00, 2.69718140541677e+00, 2.69171251539359e+00, 2.68509484159152e+00, 2.67721799005354e+00, 2.66798408111318e+00, 2.65731168589100e+00, 2.64513944085362e+00, 2.63142915510656e+00, 2.61616824014616e+00, 2.59937132039366e+00, 2.58108092382294e+00, 2.56136720266687e+00, 2.54032669047363e+00, 2.51808015867456e+00, 2.49476968798874e+00, 2.47055511249331e+00, 2.44561002318373e+00, 2.42011753110659e+00, 2.39426598734303e+00, 2.36824483979775e+00, 2.34224077799914e+00, 2.31643428102799e+00, 2.29099664468755e+00, 2.26608752621915e+00, 2.24185301155496e+00, 2.21842418345550e+00, 2.19591614988750e+00, 2.17442748056218e+00, 2.15403999477482e+00, 2.13481884417494e+00, 2.11681283829260e+00, 2.10005496707160e+00, 2.08456308207399e+00, 2.07034070549519e+00, 2.05737794304214e+00, 2.04565248271019e+00, 2.03513066635649e+00, 2.02576862462995e+00, 2.01751346825508e+00, 2.01030452988321e+00, 2.00407465074549e+00, 1.99875150522277e+00, 1.99425895429255e+00, 1.99051841579382e+00, 1.98745023581098e+00, 1.98497504154062e+00, 1.98301505215144e+00, 1.98149532079554e+00, 1.98034487849811e+00, 1.97949774951918e+00, 1.97889380824294e+00, 1.97847944988487e+00, 1.97820805135032e+00, 1.97804020430877e+00, 1.97794370969712e+00, 1.97789333103692e+00, 1.97787031266195e+00, 1.97786167768306e+00, 1.97785932876022e+00, 1.97785898206416e+00, 1.97785895603815e+00, 1.97785849795856e+00, 1.97785539440275e+00, 1.97784398541534e+00, 1.97781357333612e+00, 1.97774701771350e+00, 1.97761955753072e+00, 1.97739790882477e+00, 1.97703968067800e+00, 1.97649314658618e+00, 1.97569740108297e+00, 1.97458292286020e+00, 1.97307255508191e+00, 1.97108290083467e+00, 1.96852611655924e+00, 1.96531206903028e+00, 1.96135080253421e+00, 1.95655524331439e+00, 1.95084404950817e+00, 1.94414449843125e+00, 1.93639529108238e+00, 1.92754914798511e+00, 1.91757507242522e+00, 1.90646016760319e+00, 1.89421091312797e+00, 1.88085383256235e+00, 1.86643551536890e+00, 1.85102199085461e+00, 1.83469748548109e+00, 1.81756262521640e+00, 1.79973216903610e+00, 1.78133237671559e+00, 1.76249812322521e+00, 1.74336987386768e+00, 1.72409063003965e+00, 1.70480294677332e+00, 1.68564611160373e+00, 1.66675356105842e+00, 1.64825059694055e+00, 1.63025244988359e+00, 1.61286272244147e+00, 1.59617222827485e+00, 1.58025822805966e+00, 1.56518404717954e+00, 1.55099904599656e+00, 1.53773890161114e+00, 1.52542615152741e+00, 1.51407094521174e+00, 1.50367194935673e+00, 1.49421735638220e+00, 1.48568595250177e+00, 1.47804821043557e+00, 1.47126738134039e+00, 1.46530056962541e+00, 1.46009978211709e+00, 1.45561294891630e+00, 1.45178491694965e+00, 1.44855841862976e+00, 1.44587501741138e+00, 1.44367602973723e+00, 1.44190341940032e+00, 1.44050065626067e+00, 1.43941352711253e+00, 1.43859088284644e+00, 1.43798530336045e+00, 1.43755366031974e+00, 1.43725755807821e+00, 1.43706363494655e+00, 1.43694371044308e+00, 1.43687476897771e+00, 1.43683877625276e+00, 1.43682233109252e+00, 1.43681616197964e+00, 1.43681448383461e+00, 1.43681423614404e+00, 1.43681421747105e+00, 1.43681388775595e+00, 1.43681165388938e+00, 1.43680344198620e+00, 1.43678155227194e+00, 1.43673364820608e+00, 1.43664191016866e+00, 1.43648238940799e+00, 1.43622459490633e+00, 1.43583134220517e+00, 1.43525888873609e+00, 1.43445737443127e+00, 1.43337157892840e+00, 1.43194199719186e+00, 1.43010622365930e+00, 1.42780062114878e+00, 1.42496223509533e+00, 1.42153089694839e+00, 1.41745144381554e+00, 1.41267596604061e+00, 1.40716598190469e+00, 1.40089443064072e+00, 1.39384737294665e+00, 1.38602529337475e+00, 1.37744391212004e+00, 1.36813443498828e+00, 1.35814319914436e+00, 1.34753070727631e+00, 1.33637008186327e+00, 1.32474501127259e+00, 1.31274729668077e+00, 1.30047413914215e+00, 1.28802532544007e+00, 1.27550047639365e+00, 1.26299651042880e+00, 1.25060544911160e+00, 1.23841265315901e+00, 1.22649553250163e+00, 1.21492272874112e+00, 1.20375372908060e+00, 1.19303884225183e+00, 1.18281945147302e+00, 1.17312845682754e+00, 1.16399082733674e+00, 1.15542419786647e+00, 1.14743946406745e+00, 1.14004134654029e+00, 1.13322891104356e+00, 1.12699604359843e+00, 1.12133188747286e+00, 1.11622125357797e+00, 1.11164501745082e+00, 1.10758051551153e+00, 1.10400195139310e+00, 1.10088082043849e+00, 1.09818635737046e+00, 1.09588600894880e+00, 1.09394593031315e+00, 1.09233150076972e+00, 1.09100785208696e+00, 1.08994039997652e+00, 1.08909536741919e+00, 1.08844028693072e+00, 1.08794446783998e+00, 1.08757941426491e+00, 1.08731917979822e+00, 1.08714064600485e+00, 1.08702371369004e+00, 1.08695139847137e+00, 1.08690982536788e+00, 1.08688812074132e+00, 1.08687820378210e+00, 1.08687448359866e+00, 1.08687347161954e+00, 1.08687332225358e+00, 1.08687331093185e+00, 1.08687311020740e+00, 1.08687175027044e+00, 1.08686675100839e+00, 1.08685342488868e+00, 1.08682426139449e+00, 1.08676841131901e+00, 1.08667129227438e+00, 1.08651433469530e+00, 1.08627488515392e+00, 1.08592628078415e+00, 1.08543810481183e+00, 1.08477662836065e+00, 1.08390543764328e+00, 1.08278623825267e+00, 1.08137981961408e+00, 1.07964715306476e+00, 1.07755058709943e+00, 1.07505509394549e+00, 1.07212951392607e+00, 1.06874773923564e+00, 1.06488977789970e+00, 1.06054264258769e+00, 1.05570101781768e+00, 1.05036767243519e+00, 1.04455360082637e+00, 1.03827789427733e+00, 1.03156736102992e+00, 1.02445592780201e+00, 1.01698386522972e+00, 1.00919688407498e+00, 1.00114514831458e+00, 9.92882246442377e-01, 9.84464155083318e-01, 9.75948221074947e-01, 9.67392180986222e-01, 9.58853231526579e-01, 9.50387160729534e-01, 9.42047547908020e-01, 9.33885039576831e-01, 9.25946708136626e-01, 9.18275499546368e-01, 9.10909775142436e-01, 9.03882951117215e-01, 8.97223237081020e-01, 8.90953472853344e-01, 8.85091060444720e-01, 8.79647986334174e-01, 8.74630927766495e-01, 8.70041435933565e-01, 8.65876188520733e-01, 8.62127304082412e-01, 8.58782710914857e-01, 8.55826563364342e-01, 8.53239698708951e-01, 8.51000127777350e-01, 8.49083552257580e-01, 8.47463901188496e-01, 8.46113878446380e-01, 8.45005512209514e-01, 8.44110696504028e-01, 8.43401714127021e-01, 8.42851729636150e-01, 8.42435240813053e-01, 8.42128477157855e-01, 8.41909734630400e-01, 8.41759637059811e-01, 8.41661316392186e-01, 8.41600506185664e-01, 8.41565545400639e-01, 8.41547292443169e-01, 8.41538952452893e-01, 8.41535823824380e-01, 8.41534972760600e-01, 8.41534847145336e-01, 8.41534837563453e-01, 8.41534666887984e-01, 8.41533510536339e-01, 8.41529259662174e-01, 8.41517928367780e-01, 8.41493129982453e-01, 8.41445637630031e-01, 8.41363046535273e-01, 8.41229553353584e-01, 8.41025865301566e-01, 8.40729248970586e-01, 8.40313725355113e-01, 8.39750413751435e-01, 8.39008022732617e-01, 8.38053481380030e-01, 8.36852698422151e-01, 8.35371431060683e-01, 8.33576239318921e-01, 8.31435496091707e-01, 8.28920418151309e-01, 8.26006079646284e-01, 8.22672367578708e-01, 8.18904838737266e-01, 8.14695439833606e-01, 8.10043057187544e-01, 8.04953869072145e-01, 7.99441482402024e-01, 7.93526845296968e-01, 7.87237937532405e-01, 7.80609251295107e-01, 7.73681084298682e-01, 7.66498675539017e-01, 7.59111220251347e-01, 7.51570804581625e-01, 7.43931301900732e-01, 7.36247271570790e-01, 7.28572897532469e-01, 7.20960998732702e-01, 7.13462136722020e-01, 7.06123838376467e-01, 6.98989944304483e-01, 6.92100086674177e-01, 6.85489294395078e-01, 6.79187719092409e-01, 6.73220472225594e-01, 6.67607561970287e-01, 6.62363917926651e-01, 6.57499492078834e-01, 6.53019425423370e-01, 6.48924271025248e-01, 6.45210265702580e-01, 6.41869643886533e-01, 6.38890988308113e-01, 6.36259612936545e-01, 6.33957973988897e-01, 6.31966104841915e-01, 6.30262070331971e-01, 6.28822435282173e-01, 6.27622741224236e-01, 6.26637984281675e-01, 6.25843086158023e-01, 6.25213349244888e-01, 6.24724886145591e-01, 6.24355013510099e-01, 6.24082600087487e-01, 6.23888359390358e-01, 6.23755078364431e-01, 6.23667774960127e-01, 6.23613779465801e-01, 6.23582736799232e-01, 6.23566529550065e-01, 6.23559124283041e-01, 6.23556346305087e-01, 6.23555590627101e-01, 6.23555479090591e-01, 6.23555470522591e-01, 6.23555317119653e-01, 6.23554277792110e-01, 6.23550457113071e-01, 6.23540272577479e-01, 6.23517983929652e-01, 6.23475298346298e-01, 6.23401067269215e-01, 6.23281088798660e-01, 6.23098027371075e-01, 6.22831459898059e-01, 6.22458054559244e-01, 6.21951884938037e-01, 6.21284878103914e-01, 6.20427390562755e-01, 6.19348900775586e-01, 6.18018801338025e-01, 6.16407268177626e-01, 6.14486178629466e-01, 6.12230045443939e-01, 6.09616930167272e-01, 6.06629297411288e-01, 6.03254771719358e-01, 5.99486761321119e-01, 5.95324918124758e-01, 5.90775410646545e-01, 5.85850995785290e-01, 5.80570885739405e-01, 5.74960417089137e-01, 5.69050539205545e-01, 5.62877147824397e-01, 5.56480296123002e-01, 5.49903319505571e-01, 5.43191911396766e-01, 5.36393185840708e-01, 5.29554759051718e-01, 5.22723876896089e-01, 5.15946609309466e-01, 5.09267126549441e-01, 5.02727066505063e-01, 4.96364997417224e-01, 4.90215976501389e-01, 4.84311202134144e-01, 4.78677755373843e-01, 4.73338425468855e-01, 4.68311613476473e-01, 4.63611307991311e-01, 4.59247127111167e-01, 4.55224421033504e-01, 4.51544429992190e-01, 4.48204492553669e-01, 4.45198299550974e-01, 4.42516189107081e-01, 4.40145478253724e-01, 4.38070826560088e-01, 4.36274626929565e-01, 4.34737418297384e-01, 4.33438314381398e-01, 4.32355441937512e-01, 4.31466381204526e-01, 4.30748600464968e-01, 4.30179875983866e-01, 4.29738688110037e-01, 4.29404584122782e-01, 4.29158498558325e-01, 4.28983022310318e-01, 4.28862612794200e-01, 4.28783738889177e-01, 4.28734956183225e-01, 4.28706910171101e-01, 4.28692267393041e-01, 4.28685576933782e-01, 4.28683067102956e-01, 4.28682384367306e-01, 4.28682283596924e-01, 4.28682275796879e-01, 4.28682135374487e-01, 4.28681183991948e-01, 4.28677686606624e-01, 4.28668363842540e-01, 4.28647961105549e-01, 4.28608887056205e-01, 4.28540935807227e-01, 4.28431105436863e-01, 4.28263523351852e-01, 4.28019486654419e-01, 4.27677622908504e-01, 4.27214173500131e-01, 4.26603398102612e-01, 4.25818094585503e-01, 4.24830224105982e-01, 4.23611626231118e-01, 4.22134803988177e-01, 4.20373754042618e-01, 4.18304813143832e-01, 4.15907488973508e-01, 4.13165241980607e-01, 4.10066185013669e-01, 4.06603669753894e-01, 4.02776733117829e-01, 3.98590382737103e-01, 3.94055707940151e-01, 3.89189810807119e-01, 3.84015560212900e-01, 3.78561179685157e-01, 3.72859686839801e-01, 3.66948207717694e-01, 3.60867193310510e-01, 3.54659567878169e-01, 3.48369839410989e-01, 3.42043201951805e-01, 3.35724657683235e-01, 3.29458183920657e-01, 3.23285966634252e-01, 3.17247718040857e-01, 3.11380091346052e-01, 3.05716201082756e-01, 3.00285252907972e-01, 2.95112282416792e-01, 2.90217998733350e-01, 2.85618725527689e-01, 2.81326429815271e-01, 2.77348827482848e-01, 2.73689553932488e-01, 2.70348388453250e-01, 2.67321521761763e-01, 2.64601857403987e-01, 2.62179339169373e-01, 2.60041298133494e-01, 2.58172814241092e-01, 2.56557088334679e-01, 2.55175821139607e-01, 2.54009595903318e-01, 2.53038261174192e-01, 2.52241309661557e-01, 2.51598248347506e-01, 2.51088954153996e-01, 2.50694008647704e-01, 2.50395004630249e-01, 2.50174817135683e-01, 2.50017831435557e-01, 2.49910121190279e-01, 2.49839570897283e-01, 2.49795938239897e-01, 2.49770853764080e-01, 2.49757757398257e-01, 2.49751773558538e-01, 2.49749528813350e-01, 2.49748918188203e-01, 2.49748828061209e-01, 2.49748821028449e-01, 2.49748693688661e-01, 2.49747830943052e-01, 2.49744659402007e-01, 2.49736205261613e-01, 2.49717703701942e-01, 2.49682271461033e-01, 2.49620655934802e-01, 2.49521072728877e-01, 2.49369141925353e-01, 2.49147931357786e-01, 2.48838112820509e-01, 2.48418234280217e-01, 2.47865107736391e-01, 2.47154308364403e-01, 2.46260776012898e-01, 2.45159505141439e-01, 2.43826304096133e-01, 2.42238599541817e-01, 2.40376257283971e-01, 2.38222387049973e-01, 2.35764096487024e-01, 2.32993159053295e-01, 2.29906561913188e-01, 2.26506903537132e-01, 2.22802616419585e-01, 2.18807997946149e-01, 2.14543041561861e-01, 2.10033070458831e-01, 2.05308186337163e-01, 2.00402555659711e-01, 1.95353564479955e-01, 1.90200879702076e-01, 1.84985458993488e-01, 1.79748553163229e-01, 1.74530743526164e-01, 1.69371052733231e-01, 1.64306161156931e-01, 1.59369752793323e-01, 1.54592005540199e-01, 1.49999231450098e-01, 1.45613663903603e-01, 1.41453381238495e-01, 1.37532350653940e-01, 1.33860572424547e-01, 1.30444302632906e-01, 1.27286332598592e-01, 1.24386304641090e-01, 1.21741046375039e-01, 1.19344908988483e-01, 1.17190098512962e-01, 1.15266992626752e-01, 1.13564438772984e-01, 1.12070032126919e-01, 1.10770374084074e-01, 1.09651313402059e-01, 1.08698172909479e-01, 1.07895964839127e-01, 1.07229597430088e-01, 1.06684074583791e-01, 1.06244689182651e-01, 1.05897209331935e-01, 1.05628055416734e-01, 1.05424464624740e-01, 1.05274638605065e-01, 1.05167869322042e-01, 1.05094637993501e-01, 1.05046682306172e-01, 1.05017027862333e-01, 1.04999980973872e-01, 1.04991081388894e-01, 1.04987015192540e-01, 1.04985489837681e-01, 1.04985074905665e-01, 1.04985013662622e-01, 1.04985008843069e-01, 1.04984921055960e-01, 1.04984326286298e-01, 1.04982139863832e-01, 1.04976311769747e-01, 1.04963557663593e-01, 1.04939134075455e-01, 1.04896667669392e-01, 1.04828047903419e-01, 1.04723391691859e-01, 1.04571087687901e-01, 1.04357926383495e-01, 1.04069320196832e-01, 1.03689614935174e-01, 1.03202490392563e-01, 1.02591443373221e-01, 1.01840341251329e-01, 1.00934028558587e-01, 9.98589634457776e-02, 9.86038557302180e-02, 9.71602742303510e-02, 9.55231888269079e-02, 9.36914127311290e-02, 9.16679131785312e-02, 8.94599643525833e-02, 8.70791246224076e-02, 8.45410306641927e-02, 8.18650129245277e-02, 7.90735491224197e-02, 7.61915838875300e-02, 7.32457520195346e-02, 7.02635492253272e-02, 6.72724968688545e-02, 6.42993460031906e-02, 6.13693609682677e-02, 5.85057147873904e-02, 5.57290184677530e-02, 5.30569952659423e-02, 5.05043001870362e-02, 4.80824754638594e-02, 4.58000252637909e-02, 4.36625878140067e-02, 4.16731806014621e-02, 3.98324940760073e-02, 3.81392109391339e-02, 3.65903311065239e-02, 3.51814862439461e-02, 3.39072319016953e-02, 3.27613093108585e-02, 3.17368725628064e-02, 3.08266799857169e-02, 3.00232509740663e-02, 2.93189913118574e-02, 2.87062912063689e-02, 2.81776008923291e-02, 2.77254888605736e-02, 2.73426875923395e-02, 2.70221312139878e-02, 2.67569887958725e-02, 2.65406961664105e-02, 2.63669881602624e-02, 2.62299322291728e-02, 2.61239633777102e-02, 2.60439195056996e-02, 2.59850755041420e-02, 2.59431739137685e-02, 2.59144496542980e-02, 2.58956462890563e-02, 2.58840215021738e-02, 2.58773399094814e-02, 2.58738519526283e-02, 2.58722583760349e-02, 2.58716605865708e-02, 2.58714979748635e-02, 2.58714739737727e-02, 2.58714720682729e-02, 2.58714371475612e-02, 2.58712005554553e-02, 2.58703308307684e-02, 2.58680125633290e-02, 2.58629395953900e-02, 2.58532261846671e-02, 2.58363405198594e-02, 2.58090650034468e-02, 2.57674876971485e-02, 2.57070291522226e-02, 2.56225083812236e-02, 2.55082508202076e-02, 2.53582396478453e-02, 2.51663096836072e-02, 2.49263802746640e-02, 2.46327202032539e-02, 2.42802339351282e-02, 2.38647548476631e-02, 2.33833278949405e-02, 2.28344620161440e-02, 2.22183319863074e-02, 2.15369107474093e-02, 2.07940167412541e-02, 1.99952663070864e-02, 1.91479283973148e-02, 1.82606869896027e-02, 1.73433246951831e-02, 1.64063481616772e-02, 1.54605810123094e-02, 1.45167525721543e-02, 1.35851101958497e-02, 1.26750797378659e-02, 1.17949931001675e-02, 1.09518946672829e-02, 1.01514307845800e-02, 9.39781924058281e-03, 8.69388978866148e-03, 8.04118258528952e-03, 7.44008918378647e-03, 6.89002027306351e-03, 6.38958538855050e-03, 5.93677195826992e-03, 5.52911384697197e-03, 5.16384257495162e-03, 4.83801721543564e-03, 4.54863134578254e-03, 4.29269724715020e-03, 4.06730885151596e-03, 3.86968580593461e-03, 3.69720152536779e-03, 3.54739827992034e-03, 3.41799226617034e-03, 3.30687133338614e-03, 3.21208767402936e-03, 3.13184742095451e-03, 3.06449875456671e-03, 3.00851981157457e-03, 2.96250739065407e-03, 2.92516716300462e-03, 2.89530582013064e-03, 2.87182533464225e-03, 2.85371927973621e-03, 2.84007095377703e-03, 2.83005289336025e-03, 2.82292723624877e-03, 2.81804632121840e-03, 2.81485289002603e-03, 2.81287928909535e-03, 2.81174515221494e-03, 2.81115317199365e-03, 2.81088272362773e-03, 2.81078127433609e-03, 2.81075367816186e-03, 2.81074960504135e-03, 2.81074927870328e-03, 2.81074326088415e-03, 2.81070248954940e-03, 2.81055261396901e-03, 2.81015313262500e-03, 2.80927903946271e-03, 2.80760566709048e-03, 2.80469759971069e-03, 2.80000259781588e-03, 2.79285153971465e-03, 2.78246540593408e-03, 2.76797026592313e-03, 2.74842102249018e-03, 2.72283428315667e-03, 2.69023013117398e-03, 2.64968176286473e-03, 2.60037098456678e-03, 2.54164651238220e-03, 2.47308102910004e-03, 2.39452219827463e-03, 2.30613249897615e-03, 2.20841298683006e-03, 2.10220700506270e-03, 1.98868146192055e-03, 1.86928543177345e-03, 1.74568826792765e-03, 1.61970177360260e-03, 1.49319286098226e-03, 1.36799419490430e-03, 1.24582038599286e-03, 1.12819640176731e-03, 1.01640322372631e-03, 9.11443696248071e-04, 8.14029276018147e-04, 7.24586238396441e-04, 6.43278075111553e-04, 5.70039603237760e-04, 5.04617920991990e-04, 4.46615785172355e-04, 3.95533935732217e-04, 3.50809899559934e-04, 3.11851580169886e-04, 2.78064536766820e-04, 2.48872510960396e-04, 2.23731552235762e-04, 2.02138794897166e-04, 1.83637220999344e-04, 1.67817526513403e-04, 1.54317765741041e-04, 1.42821161722040e-04, 1.33052491738894e-04, 1.24773630732378e-04, 1.17778900946774e-04, 1.11890719152186e-04, 1.06955749481941e-04, 1.02841535545295e-04, 9.94334995620960e-05, 9.66322340616443e-05, 9.43510860302865e-05, 9.25140750496212e-05, 9.10541802436882e-05, 8.99119967066042e-05, 8.90347278996728e-05, 8.83754600291553e-05, 8.78926581560080e-05, 8.75498252295159e-05, 8.73152693685597e-05, 8.71619290183645e-05, 8.70672102450953e-05, 8.70127965377776e-05, 8.69843993314134e-05, 8.69714270734789e-05, 8.69665611631376e-05, 8.69652375570240e-05, 8.69650421968515e-05, 8.69650264032629e-05, 8.69647333576751e-05, 8.69627479627155e-05, 8.69554497704428e-05, 8.69359980573137e-05, 8.68934415816369e-05, 8.68119911139331e-05, 8.66705053361727e-05, 8.64422487207861e-05, 8.60949878731532e-05, 8.55914978465933e-05, 8.48905482744300e-05, 8.39484262655787e-05, 8.27210242864140e-05, 8.11664716664495e-05, 7.92482155720364e-05, 7.69383628185238e-05, 7.42209888571210e-05, 7.10950292006629e-05, 6.75763289113007e-05, 6.36984780779670e-05, 5.95122243591817e-05, 5.50834956918322e-05, 5.04902931162121e-05, 4.58188052085973e-05, 4.11589939641743e-05, 3.65997051048339e-05, 3.22233121341498e-05, 2.81002376691930e-05, 2.42843306956786e-05, 2.08105015744309e-05, 1.76955660595723e-05, 1.49417843964272e-05, 1.25409848639180e-05, 1.04769180655495e-05, 8.72531016682387e-06, 7.25374577825733e-06, 6.02439308394764e-06, 5.00008959924969e-06, 4.15021593447307e-06, 3.45139475399492e-06, 2.88177408513945e-06, 2.41381663279583e-06, 2.01321111603647e-06, 1.64688202630714e-06, 1.29459985309666e-06, 9.55511294975508e-07, 6.46221506528377e-07, 3.94962736294246e-07, -1.77635683940025e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.95399252334028e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.59872115546023e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.59872115546023e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.59872115546023e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.77635683940025e-14, -1.59872115546023e-14],
[ 1.30000000000000e+01, 1.29999999936151e+01, 1.29999997961824e+01, 1.29999984585449e+01, 1.29999935412182e+01, 1.29999804332035e+01, 1.29999517454810e+01, 1.29998968006422e+01, 1.29998012372789e+01, 1.29996467443444e+01, 1.29994109368404e+01, 1.29990673799795e+01, 1.29985857645737e+01, 1.29979322319344e+01, 1.29970698421962e+01, 1.29959591758291e+01, 1.29945590543021e+01, 1.29928273625205e+01, 1.29907219528531e+01, 1.29882016083457e+01, 1.29852270411084e+01, 1.29817619008690e+01, 1.29777737682661e+01, 1.29732351075941e+01, 1.29681241543545e+01, 1.29624257140817e+01, 1.29561318504786e+01, 1.29492424428891e+01, 1.29417655955762e+01, 1.29337178841571e+01, 1.29251244278933e+01, 1.29160187803343e+01, 1.29064426350466e+01, 1.28964453477613e+01, 1.28860832811375e+01, 1.28754189833297e+01, 1.28645202164741e+01, 1.28534588558768e+01, 1.28423096848709e+01, 1.28311491138196e+01, 1.28200538543849e+01, 1.28090995818330e+01, 1.27983596187236e+01, 1.27879036727979e+01, 1.27777966602867e+01, 1.27680976432737e+01, 1.27588589063080e+01, 1.27501251933118e+01, 1.27419331211422e+01, 1.27343107811312e+01, 1.27272775346956e+01, 1.27208440038609e+01, 1.27150122524097e+01, 1.27097761484852e+01, 1.27051218949563e+01, 1.27010287097915e+01, 1.26974696351584e+01, 1.26944124510538e+01, 1.26918206670172e+01, 1.26896545639457e+01, 1.26878722572279e+01, 1.26864307523864e+01, 1.26852869651369e+01, 1.26843986792391e+01, 1.26837254176809e+01, 1.26832292055454e+01, 1.26828752062868e+01, 1.26826322169932e+01, 1.26824730124271e+01, 1.26823745321049e+01, 1.26823179092657e+01, 1.26822883451743e+01, 1.26822748366705e+01, 1.26822697690955e+01, 1.26822683905857e+01, 1.26822681871203e+01, 1.26822681604299e+01, 1.26822675385762e+01, 1.26822633254200e+01, 1.26822478373222e+01, 1.26822065510488e+01, 1.26821161935432e+01, 1.26819431346747e+01, 1.26816421418042e+01, 1.26811555444097e+01, 1.26804128449901e+01, 1.26793307993996e+01, 1.26778139760814e+01, 1.26757557898328e+01, 1.26730399922094e+01, 1.26695425879165e+01, 1.26651341349161e+01, 1.26596823757908e+01, 1.26530551393077e+01, 1.26451234441236e+01, 1.26357647310453e+01, 1.26248661459804e+01, 1.26123277924379e+01, 1.25980658699730e+01, 1.25820156132929e+01, 1.25641339460513e+01, 1.25444017641323e+01, 1.25228257661610e+01, 1.24994397549066e+01, 1.24743053429612e+01, 1.24475120101408e+01, 1.24191764786020e+01, 1.23894413942751e+01, 1.23584733288500e+01, 1.23264601436464e+01, 1.22936077833124e+01, 1.22601365913593e+01, 1.22262772591538e+01, 1.21922665336514e+01, 1.21583428159921e+01, 1.21247417828876e+01, 1.20916921559972e+01, 1.20594117322577e+01, 1.20281037718138e+01, 1.19979538213827e+01, 1.19691270311110e+01, 1.19417660036015e+01, 1.19159891958195e+01, 1.18918898787078e+01, 1.18695356458644e+01, 1.18489684515643e+01, 1.18302051495320e+01, 1.18132384968103e+01, 1.17980385814075e+01, 1.17845546277183e+01, 1.17727171296536e+01, 1.17624402577394e+01, 1.17536244830282e+01, 1.17461593575238e+01, 1.17399263880569e+01, 1.17348019383966e+01, 1.17306600930949e+01, 1.17273754164589e+01, 1.17248255413950e+01, 1.17228935259297e+01, 1.17214699201241e+01, 1.17204544929326e+01, 1.17197575772373e+01, 1.17193010016442e+01, 1.17190185893442e+01, 1.17188562170491e+01, 1.17187714402293e+01, 1.17187327041213e+01, 1.17187181727277e+01, 1.17187142198229e+01, 1.17187136363820e+01, 1.17187135848752e+01, 1.17187125756261e+01, 1.17187057378087e+01, 1.17186806011841e+01, 1.17186135955208e+01, 1.17184669519628e+01, 1.17181860993500e+01, 1.17176976546975e+01, 1.17169080919652e+01, 1.17157031560221e+01, 1.17139480695239e+01, 1.17114885594853e+01, 1.17081527078412e+01, 1.17037536065611e+01, 1.16980927732615e+01, 1.16909642582765e+01, 1.16821593493919e+01, 1.16714717566711e+01, 1.16587031378053e+01, 1.16436688051072e+01, 1.16262034395768e+01, 1.16061666264071e+01, 1.15834480209038e+01, 1.15579719551236e+01, 1.15297013045996e+01, 1.14986404521185e+01, 1.14648372120878e+01, 1.14283836143958e+01, 1.13894154898179e+01, 1.13481108480105e+01, 1.13046870910465e+01, 1.12593971566589e+01, 1.12125247319042e+01, 1.11643787160400e+01, 1.11152871379891e+01, 1.10655907469774e+01, 1.10156364944115e+01, 1.09657711119062e+01, 1.09163349669632e+01, 1.08676563473457e+01, 1.08200462912076e+01, 1.07737940458620e+01, 1.07291632064325e+01, 1.06863885584186e+01, 1.06456736264052e+01, 1.06071889148955e+01, 1.05710708160595e+01, 1.05374211521420e+01, 1.05063073162572e+01, 1.04777629731932e+01, 1.04517892806577e+01, 1.04283565903249e+01, 1.04074065865088e+01, 1.03888548179443e+01, 1.03725935748485e+01, 1.03584950591788e+01, 1.03464147909711e+01, 1.03361951881222e+01, 1.03276692513522e+01, 1.03206642808348e+01, 1.03150055466084e+01, 1.03105198319412e+01, 1.03070387678165e+01, 1.03044018781020e+01, 1.03024592591084e+01, 1.03010738243315e+01, 1.03001230552342e+01, 1.02995002117744e+01, 1.02991149716500e+01, 1.02988934843550e+01, 1.02987778444063e+01, 1.02987250067319e+01, 1.02987051853624e+01, 1.02986997934567e+01, 1.02986989976222e+01, 1.02986989337864e+01, 1.02986977556926e+01, 1.02986897739295e+01, 1.02986604320621e+01, 1.02985822169396e+01, 1.02984110426377e+01, 1.02980832141003e+01, 1.02975130885072e+01, 1.02965915360608e+01, 1.02951852830939e+01, 1.02931371991002e+01, 1.02902675653532e+01, 1.02863763360618e+01, 1.02812463734907e+01, 1.02746476063396e+01, 1.02663420264712e+01, 1.02560894037578e+01, 1.02436535638406e+01, 1.02288090408614e+01, 1.02113478890340e+01, 1.01910864157682e+01, 1.01678715874117e+01, 1.01415868586723e+01, 1.01121571898930e+01, 1.00795530430863e+01, 1.00437931872724e+01, 1.00049461942081e+01, 9.96313056374643e+00, 9.91851347962563e+00, 9.87130825666802e+00, 9.82177059454573e+00, 9.77019379747731e+00, 9.71690315079080e+00, 9.66224966313465e+00, 9.60660338779613e+00, 9.55034653004014e+00, 9.49386653244331e+00, 9.43754930998287e+00, 9.38177278396333e+00, 9.32690084083526e+00, 9.27327781981037e+00, 9.22122361234329e+00, 9.17102943693317e+00, 9.12295433400194e+00, 9.07722240762734e+00, 9.03402082371254e+00, 8.99349855811358e+00, 8.95576587384860e+00, 8.92089449430710e+00, 8.88891842971154e+00, 8.85983540700805e+00, 8.83360884860200e+00, 8.81017034235834e+00, 8.78942254334748e+00, 8.77124244620624e+00, 8.75548496508086e+00, 8.74198675551242e+00, 8.73057020916521e+00, 8.72104754811294e+00, 8.71322494086217e+00, 8.70690655798300e+00, 8.70189848183256e+00, 8.69801238313763e+00, 8.69506887784607e+00, 8.69290048123337e+00, 8.69135408314187e+00, 8.69029287857287e+00, 8.68959770151918e+00, 8.68916772652139e+00, 8.68892052132366e+00, 8.68879145437231e+00, 8.68873248180340e+00, 8.68871035902533e+00, 8.68870434108089e+00, 8.68870345284444e+00, 8.68870338433510e+00, 8.68870215412709e+00, 8.68869381928284e+00, 8.68866317945409e+00, 8.68858150446201e+00, 8.68840275841539e+00, 8.68806042944651e+00, 8.68746508862854e+00, 8.68650278477101e+00, 8.68503436238133e+00, 8.68289576806280e+00, 8.67989938583376e+00, 8.67583641414671e+00, 8.67048026672415e+00, 8.66359094592037e+00, 8.65492030172610e+00, 8.64421805275673e+00, 8.63123840907411e+00, 8.61574710239493e+00, 8.59752859937717e+00, 8.57639325064296e+00, 8.55218411430317e+00, 8.52478318995439e+00, 8.49411680878348e+00, 8.46015994810418e+00, 8.42293927404110e+00, 8.38253476296679e+00, 8.33907980870654e+00, 8.29275978583353e+00, 8.24380910645184e+00, 8.19250687512828e+00, 8.13917131009203e+00, 8.08415315413418e+00, 8.02782834135268e+00, 7.97059021192948e+00, 7.91284157355487e+00, 7.85498689395783e+00, 7.79742487588355e+00, 7.74054161808607e+00, 7.68470450984429e+00, 7.63025694937481e+00, 7.57751392487982e+00, 7.52675845552531e+00, 7.47823886055128e+00, 7.43216680766681e+00, 7.38871608476351e+00, 7.34802203877893e+00, 7.31018162920281e+00, 7.27525404873020e+00, 7.24326186820263e+00, 7.21419266628042e+00, 7.18800110587216e+00, 7.16461141918475e+00, 7.14392026148422e+00, 7.12579989046297e+00, 7.11010162368839e+00, 7.09665952117307e+00, 7.08529423392084e+00, 7.07581695271464e+00, 7.06803338487759e+00, 7.06174768081735e+00, 7.05676622749403e+00, 7.05290122319907e+00, 7.04997394782777e+00, 7.04781764570006e+00, 7.04627994428979e+00, 7.04522474209803e+00, 7.04453351222726e+00, 7.04410598459475e+00, 7.04386018852919e+00, 7.04373185789248e+00, 7.04367322189350e+00, 7.04365122539736e+00, 7.04364524180682e+00, 7.04364435864102e+00, 7.04364429177037e+00, 7.04364310716337e+00, 7.04363508127354e+00, 7.04360557720666e+00, 7.04352692983679e+00, 7.04335481016465e+00, 7.04302517340087e+00, 7.04245191039296e+00, 7.04152530613007e+00, 7.04011139073713e+00, 7.03805224745348e+00, 7.03516731831393e+00, 7.03125572133288e+00, 7.02609956278356e+00, 7.01946819478045e+00, 7.01112333234692e+00, 7.00082490654404e+00, 6.98833749270649e+00, 6.97343711760344e+00, 6.95591821909820e+00, 6.93560050952021e+00, 6.91233548227979e+00, 6.88601230253147e+00, 6.85656283828887e+00, 6.82396561841214e+00, 6.78824854697030e+00, 6.74949025684124e+00, 6.70782004512763e+00, 6.66341639449171e+00, 6.61650414333542e+00, 6.56735042004433e+00, 6.51625949959953e+00, 6.46356677340237e+00, 6.40963204500765e+00, 6.35483237629047e+00, 6.29955471138165e+00, 6.24418850045761e+00, 6.18911853289704e+00, 6.13471816999546e+00, 6.08134314199197e+00, 6.02932604355822e+00, 5.97897162755723e+00, 5.93055296070990e+00, 5.88430846904148e+00, 5.84043986788156e+00, 5.79911094274722e+00, 5.76044712508796e+00, 5.72453579134487e+00, 5.69142720506078e+00, 5.66113601918609e+00, 5.63364325806125e+00, 5.60889870431495e+00, 5.58682362348895e+00, 5.56731376706590e+00, 5.55024260144408e+00, 5.53546471531884e+00, 5.52281936032798e+00, 5.51213407953631e+00, 5.50322837559304e+00, 5.49591736575007e+00, 5.49001536519045e+00, 5.48533933426624e+00, 5.48171212033048e+00, 5.47896542187440e+00, 5.47694240251445e+00, 5.47549988564835e+00, 5.47451006765559e+00, 5.47386169836217e+00, 5.47346069179426e+00, 5.47323014737851e+00, 5.47310978083315e+00, 5.47305478400289e+00, 5.47303415273467e+00, 5.47302854052396e+00, 5.47302771217321e+00, 5.47302764998866e+00, 5.47302655546999e+00, 5.47301913994369e+00, 5.47299187966599e+00, 5.47291921363128e+00, 5.47276018506626e+00, 5.47245562356054e+00, 5.47192597985541e+00, 5.47106991150577e+00, 5.46976370165719e+00, 5.46786157475459e+00, 5.46519695171154e+00, 5.46158466249916e+00, 5.45682410598034e+00, 5.45070331512241e+00, 5.44300385089935e+00, 5.43350641119698e+00, 5.42199700338825e+00, 5.40827349302208e+00, 5.39215230877978e+00, 5.37347505826129e+00, 5.35211479303499e+00, 5.32798165718819e+00, 5.30102766322013e+00, 5.27125036351633e+00, 5.23869522473518e+00, 5.20345656491808e+00, 5.16567697648625e+00, 5.12554522889317e+00, 5.08329271807943e+00, 5.03918860101058e+00, 4.99353381734496e+00, 4.94665425191902e+00, 4.89889332735275e+00, 4.85060433307090e+00, 4.80214279444432e+00, 4.75385916440948e+00, 4.70609208237262e+00, 4.65916239542225e+00, 4.61336807976495e+00, 4.56898014109583e+00, 4.52623951625139e+00, 4.48535494907482e+00, 4.44650177386082e+00, 4.40982151158669e+00, 4.37542216760212e+00, 4.34337911364081e+00, 4.31373644020917e+00, 4.28650867539260e+00, 4.26168278052432e+00, 4.23922034971154e+00, 4.21905995691433e+00, 4.20111960952387e+00, 4.18529928002099e+00, 4.17148349657342e+00, 4.15954397900702e+00, 4.14934230846274e+00, 4.14073261752400e+00, 4.13356428320891e+00, 4.12768459870620e+00, 4.12294139197669e+00, 4.11918555131795e+00, 4.11627341070041e+00, 4.11406894208516e+00, 4.11244569886339e+00, 4.11128845465407e+00, 4.11049448534123e+00, 4.10997444950207e+00, 4.10965283303285e+00, 4.10946793728103e+00, 4.10937140555247e+00, 4.10932729952644e+00, 4.10931075385246e+00, 4.10930625302977e+00, 4.10930558871783e+00, 4.10930553899938e+00, 4.10930466591836e+00, 4.10929875066764e+00, 4.10927700560439e+00, 4.10921904143504e+00, 4.10909218920974e+00, 4.10884925563123e+00, 4.10842680339397e+00, 4.10774403860081e+00, 4.10670237598652e+00, 4.10518573920040e+00, 4.10306163708164e+00, 4.10018303698738e+00, 4.09639103223192e+00, 4.09151827231849e+00, 4.08539309216032e+00, 4.07784424081734e+00, 4.06870607308547e+00, 4.05782403096735e+00, 4.04506020964098e+00, 4.03029877741649e+00, 4.01345100474758e+00, 3.99445965665628e+00, 3.97330251808567e+00, 3.94999485357332e+00, 3.92459065048190e+00, 3.89718255632092e+00, 3.86790049125515e+00, 3.83690899119023e+00, 3.80440340854208e+00, 3.77060516057463e+00, 3.73575626342451e+00, 3.70011341950389e+00, 3.66394193480738e+00, 3.62750973096746e+00, 3.59108168713868e+00, 3.55491450317819e+00, 3.51925222349982e+00, 3.48432250613012e+00, 3.45033366919038e+00, 3.41747250153438e+00, 3.38590278843574e+00, 3.35576447835293e+00, 3.32717340280965e+00, 3.30022145710663e+00, 3.27497715300937e+00, 3.25148646349100e+00, 3.22977389183188e+00, 3.20984371092532e+00, 3.19168133195067e+00, 3.17525477352500e+00, 3.16051621231843e+00, 3.14740360353991e+00, 3.13584236455444e+00, 3.12574711724526e+00, 3.11702348477622e+00, 3.10956993641782e+00, 3.10327967042092e+00, 3.09804251996186e+00, 3.09374686140700e+00, 3.09028149807481e+00, 3.08753748686948e+00, 3.08540987020272e+00, 3.08379927206736e+00, 3.08261331547539e+00, 3.08176781910739e+00, 3.08118773416381e+00, 3.08080778809533e+00, 3.08057280994079e+00, 3.08043772203625e+00, 3.08036719432403e+00, 3.08033496970928e+00, 3.08032288115663e+00, 3.08031959277842e+00, 3.08031910742073e+00, 3.08031907108198e+00, 3.08031843277359e+00, 3.08031410814153e+00, 3.08029821035368e+00, 3.08025583281925e+00, 3.08016309128697e+00, 3.07998548270677e+00, 3.07967662786994e+00, 3.07917745774354e+00, 3.07841589509485e+00, 3.07730707307003e+00, 3.07575412033072e+00, 3.07364952761058e+00, 3.07087709268357e+00, 3.06731441954169e+00, 3.06283592327626e+00, 3.05731626549105e+00, 3.05063411741330e+00, 3.04267612118181e+00, 3.03334089655362e+00, 3.02254292322944e+00, 3.01021612083398e+00, 2.99631695151470e+00, 2.98082688545927e+00, 2.96375409744525e+00, 2.94513430143456e+00, 2.92503067736058e+00, 2.90353289558071e+00, 2.88075529522129e+00, 2.85683431800608e+00, 2.83192533495708e+00, 2.80619902666954e+00, 2.77983748743622e+00, 2.75303021984379e+00, 2.72597017166426e+00, 2.69884994408625e+00, 2.67185827321625e+00, 2.64517685880309e+00, 2.61897758809646e+00, 2.59342018045138e+00, 2.56865026050238e+00, 2.54479785433564e+00, 2.52197629338373e+00, 2.50028150382171e+00, 2.47979165418922e+00, 2.46056713020176e+00, 2.44265080297258e+00, 2.42606855514743e+00, 2.41083002890382e+00, 2.39692956052711e+00, 2.38434726836029e+00, 2.37305026415134e+00, 2.36299396181488e+00, 2.35412346186529e+00, 2.34637499369557e+00, 2.33967740094202e+00, 2.33395365698673e+00, 2.32912239799019e+00, 2.32509945970668e+00, 2.32179940191395e+00, 2.31913700096167e+00, 2.31702868721440e+00, 2.31539390061016e+00, 2.31415633475236e+00, 2.31324503842475e+00, 2.31259534357125e+00, 2.31214959086773e+00, 2.31185762809748e+00, 2.31167706250640e+00, 2.31157325586088e+00, 2.31151905962233e+00, 2.31149429695569e+00, 2.31148500763139e+00, 2.31148248071060e+00, 2.31148210774241e+00, 2.31148207974419e+00, 2.31148158695198e+00, 2.31147824821308e+00, 2.31146597466743e+00, 2.31143325799102e+00, 2.31136165880206e+00, 2.31122453954177e+00, 2.31098609348301e+00, 2.31060071599912e+00, 2.31001275777653e+00, 2.30915669200780e+00, 2.30795771840006e+00, 2.30633281570778e+00, 2.30419224115862e+00, 2.30144145947999e+00, 2.29798346642309e+00, 2.29372145224248e+00, 2.28856173043400e+00, 2.28241683744292e+00, 2.27520869160090e+00, 2.26687168594951e+00, 2.25735558152320e+00, 2.24662806647567e+00, 2.23467685302831e+00, 2.22151119883203e+00, 2.20716276148690e+00, 2.19168572352024e+00, 2.17515615844230e+00, 2.15767064465500e+00, 2.13934417098106e+00, 2.12030741346258e+00, 2.10070349598003e+00, 2.08068437528403e+00, 2.06040701222007e+00, 2.04002950314742e+00, 2.01970734677866e+00, 1.99959001040829e+00, 1.97981793541894e+00, 1.96052008642067e+00, 1.94181210469708e+00, 1.92379507973738e+00, 1.90655490826051e+00, 1.89016217364413e+00, 1.87467245391500e+00, 1.86012695506820e+00, 1.84655336774805e+00, 1.83396685665838e+00, 1.82237110984517e+00, 1.81175939543543e+00, 1.80211559331107e+00, 1.79341518629171e+00, 1.78562620848449e+00, 1.77871015720992e+00, 1.77262287964037e+00, 1.76731544665369e+00, 1.76273502519657e+00, 1.75882575743503e+00, 1.75552965080189e+00, 1.75278747827113e+00, 1.75053968322207e+00, 1.74872727844834e+00, 1.74729272451236e+00, 1.74618076901028e+00, 1.74533922563787e+00, 1.74471967044950e+00, 1.74427803254881e+00, 1.74397505774212e+00, 1.74377662643404e+00, 1.74365391115195e+00, 1.74358336434643e+00, 1.74354653322201e+00, 1.74352970492905e+00, 1.74352339207998e+00, 1.74352167483428e+00, 1.74352142137250e+00, 1.74352140226443e+00, 1.74352106486692e+00, 1.74351877895054e+00, 1.74351037570079e+00, 1.74348797588109e+00, 1.74343895522963e+00, 1.74334507790592e+00, 1.74318183411268e+00, 1.74291801466394e+00, 1.74251555128333e+00, 1.74192964649901e+00, 1.74110921075691e+00, 1.73999761637558e+00, 1.73853376785194e+00, 1.73665347560008e+00, 1.73429110555378e+00, 1.73138146065512e+00, 1.72786183300575e+00, 1.72367414875528e+00, 1.71876711341518e+00, 1.71309825522812e+00, 1.70663576048423e+00, 1.69935999891030e+00, 1.69126465041242e+00, 1.68235736643669e+00, 1.67265992866659e+00, 1.66220790203876e+00, 1.65104981439128e+00, 1.63924592715313e+00, 1.62686668620289e+00, 1.61399095622193e+00, 1.60070414409449e+00, 1.58709630777206e+00, 1.57326032907049e+00, 1.55929020599370e+00, 1.54527949667975e+00, 1.53131992669819e+00, 1.51750015666020e+00, 1.50390469880746e+00, 1.49061296883513e+00, 1.47769846109299e+00, 1.46522803951516e+00, 1.45326134131167e+00, 1.44185029426706e+00, 1.43103875071660e+00, 1.42086224177000e+00, 1.41134785440103e+00, 1.40251423210793e+00, 1.39437169749929e+00, 1.38692249281067e+00, 1.38016113229546e+00, 1.37407485878637e+00, 1.36864419548001e+00, 1.36384358305495e+00, 1.35964209144606e+00, 1.35600419481668e+00, 1.35289059738086e+00, 1.35025909666766e+00, 1.34806546959264e+00, 1.34626436537114e+00, 1.34481018799628e+00, 1.34365794986860e+00, 1.34276407738503e+00, 1.34208714904806e+00, 1.34158854709836e+00, 1.34123300491692e+00, 1.34098903454163e+00, 1.34082922158677e+00, 1.34073037856006e+00, 1.34067355189415e+00, 1.34064388274824e+00, 1.34063032655510e+00, 1.34062524114039e+00, 1.34062385778252e+00, 1.34062365360173e+00, 1.34062363812511e+00, 1.34062336373808e+00, 1.34062150472518e+00, 1.34061467079030e+00, 1.34059645397704e+00, 1.34055658666301e+00, 1.34048023497640e+00, 1.34034745558882e+00, 1.34013284030063e+00, 1.33980536874078e+00, 1.33932848478820e+00, 1.33866040688771e+00, 1.33775467617812e+00, 1.33656093920543e+00, 1.33502595396612e+00, 1.33309479923745e+00, 1.33071225787700e+00, 1.32782433545202e+00, 1.32437986678775e+00, 1.32033215551903e+00, 1.31564058625051e+00, 1.31027214618095e+00, 1.30420279356348e+00, 1.29741861442266e+00, 1.28991671646553e+00, 1.28170581969816e+00, 1.27280651618875e+00, 1.26325118580949e+00, 1.25308356968252e+00, 1.24235801753913e+00, 1.23113843849637e+00, 1.21949699623946e+00, 1.20751259880296e+00, 1.19526923971965e+00, 1.18285425100948e+00, 1.17035652917447e+00, 1.15786479305242e+00, 1.14546592724656e+00, 1.13324345728151e+00, 1.12127619323283e+00, 1.10963706808857e+00, 1.09839218635036e+00, 1.08760008816777e+00, 1.07731122530394e+00, 1.06756763794431e+00, 1.05840281604971e+00, 1.04984172566428e+00, 1.04190097916270e+00, 1.03458912856612e+00, 1.02790706238348e+00, 1.02184848852874e+00, 1.01640048831467e+00, 1.01154412897095e+00, 1.00725512428248e+00, 1.00350453457870e+00, 1.00025949829497e+00, 9.97483987623108e-01, 9.95139580396671e-01, 9.93186239414209e-01, 9.91583089041052e-01, 9.90289177345236e-01, 9.89264210438860e-01, 9.88469244345481e-01, 9.87867318820316e-01, 9.87424017303331e-01, 9.87107937729109e-01, 9.86891060328784e-01, 9.86749000843265e-01, 9.86661140652380e-01, 9.86610629070816e-01, 9.86584257271108e-01, 9.86572207731893e-01, 9.86567687526275e-01, 9.86566457920045e-01, 9.86566276432717e-01, 9.86566262588935e-01, 9.86566015999209e-01, 9.86564345317529e-01, 9.86558203715024e-01, 9.86541832458169e-01, 9.86506004365701e-01, 9.86437389293073e-01, 9.86318067139919e-01, 9.86125210911975e-01, 9.85830958924216e-01, 9.85402491210854e-01, 9.84802320476955e-01, 9.83988802348014e-01, 9.82916863143718e-01, 9.81538935892749e-01, 9.79806086904965e-01, 9.77669306161559e-01, 9.75080925473245e-01, 9.71996119350901e-01, 9.68374435525426e-01, 9.64181295814324e-01, 9.59389404334161e-01, 9.53979999583177e-01, 9.47943890159717e-01, 9.41282221055122e-01, 9.34006928432598e-01, 9.26140855066446e-01, 9.17717515294365e-01, 9.08780516251113e-01, 8.99382659934252e-01, 8.89584766880729e-01, 8.79454275598881e-01, 8.69063681363629e-01, 8.58488882895806e-01, 8.47807505628676e-01, 8.37097266008154e-01, 8.26434433287282e-01, 8.15892434572518e-01, 8.05540636628244e-01, 7.95443325320182e-01, 7.85658891618738e-01, 7.76239222603753e-01, 7.67229287447176e-01, 7.58666902153470e-01, 7.50582652911710e-01, 7.42999956066393e-01, 7.35935232621275e-01, 7.29398176456510e-01, 7.23392097648896e-01, 7.17914325041734e-01, 7.12956655162202e-01, 7.08505837435141e-01, 7.04544088160556e-01, 7.01049627738103e-01, 6.97997237026490e-01, 6.95358829466807e-01, 6.93104035677939e-01, 6.91200796700821e-01, 6.89615961023357e-01, 6.88315879096823e-01, 6.87266987426730e-01, 6.86436372678497e-01, 6.85792304781522e-01, 6.85304726938398e-01, 6.84945689920159e-01, 6.84689718184325e-01, 6.84514096267623e-01, 6.84399065593052e-01, 6.84327924238117e-01, 6.84287025221144e-01, 6.84265672305793e-01, 6.84255915995507e-01, 6.84252256068813e-01, 6.84251260480211e-01, 6.84251113533428e-01, 6.84251102245291e-01, 6.84250900140468e-01, 6.84249530850945e-01, 6.84244497203565e-01, 6.84231079391934e-01, 6.84201715063576e-01, 6.84145479721046e-01, 6.84047688677317e-01, 6.83889639572175e-01, 6.83648511840142e-01, 6.83297436400311e-01, 6.82805745052958e-01, 6.82139404447117e-01, 6.81261633890768e-01, 6.80133699651973e-01, 6.78715870810395e-01, 6.76968513387470e-01, 6.74853290815463e-01, 6.72334430383398e-01, 6.69380007850908e-01, 6.65963196762926e-01, 6.62063425933825e-01, 6.57667388785049e-01, 6.52769852166658e-01, 6.47374220085771e-01, 6.41492819118318e-01, 6.35146886496139e-01, 6.28366257876914e-01, 6.21188768304025e-01, 6.13659395434247e-01, 6.05829187420833e-01, 5.97754027806399e-01, 5.89493295702038e-01, 5.81108481155566e-01, 5.72661813143284e-01, 5.64214951643152e-01, 5.55827786628312e-01, 5.47557376565324e-01, 5.39457048129286e-01, 5.31575668275059e-01, 5.23957090266308e-01, 5.16639767279562e-01, 5.09656521058529e-01, 5.03034448879307e-01, 4.96794949715094e-01, 4.90953849749488e-01, 4.85521607991313e-01, 4.80503584362440e-01, 4.75900354928489e-01, 4.71708061607355e-01, 4.67918786443285e-01, 4.64520943143333e-01, 4.61499680855521e-01, 4.58837296991286e-01, 4.56513657174833e-01, 4.54506621100263e-01, 4.52792473198340e-01, 4.51346356603393e-01, 4.50142708047242e-01, 4.49155690104696e-01, 4.48359615812571e-01, 4.47729359239363e-01, 4.47240744260968e-01, 4.46870902761072e-01, 4.46598592866943e-01, 4.46404467764570e-01, 4.46271286181307e-01, 4.46184056796110e-01, 4.46130110600212e-01, 4.46099097493621e-01, 4.46082906030867e-01, 4.46075508057492e-01, 4.46072732828695e-01, 4.46071977899777e-01, 4.46071866473874e-01, 4.46071857849047e-01, 4.46071702578326e-01, 4.46070650596877e-01, 4.46066783409790e-01, 4.46056474970311e-01, 4.46033915514487e-01, 4.45990712652733e-01, 4.45915586253097e-01, 4.45794172058664e-01, 4.45608947102652e-01, 4.45339287700454e-01, 4.44961667952404e-01, 4.44450003119050e-01, 4.43776137797283e-01, 4.42910473462884e-01, 4.41822723698388e-01, 4.40482778482547e-01, 4.38861651633268e-01, 4.36932478385751e-01, 4.34671523802187e-01, 4.32059157974377e-01, 4.29080751511412e-01, 4.25727445202673e-01, 4.21996751393973e-01, 4.17892951583145e-01, 4.13427264737150e-01, 4.08617773193733e-01, 4.03489106759091e-01, 3.98071899564833e-01, 3.92402047157816e-01, 3.86519802019954e-01, 3.80468753353357e-01, 3.74294740975127e-01, 3.68044753410805e-01, 3.61765857018847e-01, 3.55504196820341e-01, 3.49304101475738e-01, 3.43207315477425e-01, 3.37252372029477e-01, 3.31474111064958e-01, 3.25903339023979e-01, 3.20566620783373e-01, 3.15486189675083e-01, 3.10679958845078e-01, 3.06161616127042e-01, 3.01940784873102e-01, 2.98023234485919e-01, 2.94411126410495e-01, 2.91103283768786e-01, 2.88095475391826e-01, 2.85380707505128e-01, 2.82949518585617e-01, 2.80790274811729e-01, 2.78889464995489e-01, 2.77231994876447e-01, 2.75801481164775e-01, 2.74580545764575e-01, 2.73551110233171e-01, 2.72694689805576e-01, 2.71992685323188e-01, 2.71426670257727e-01, 2.70978668833681e-01, 2.70631420149297e-01, 2.70368622299593e-01, 2.70175149924759e-01, 2.70037238430680e-01, 2.69942628411364e-01, 2.69880664562079e-01, 2.69842344582411e-01, 2.69820315164749e-01, 2.69808814045469e-01, 2.69803559139776e-01, 2.69801587851434e-01, 2.69801051613982e-01, 2.69800972466488e-01, 2.69800966290479e-01, 2.69800854463556e-01, 2.69800096820273e-01, 2.69797311651997e-01, 2.69789887481835e-01, 2.69773640201343e-01, 2.69742525996579e-01, 2.69688422142089e-01, 2.69600986691428e-01, 2.69467606841863e-01, 2.69273444192534e-01, 2.69001583111109e-01, 2.68633285827296e-01, 2.68148354558647e-01, 2.67525596879658e-01, 2.66743385696339e-01, 2.65780299737283e-01, 2.64615824710104e-01, 2.63231089633102e-01, 2.61609607883649e-01, 2.59737988819882e-01, 2.57606584031549e-01, 2.55210032856159e-01, 2.52547675058512e-01, 2.49623804540084e-01, 2.46447746325018e-01, 2.43033749250131e-01, 2.39400697894970e-01, 2.35571658296299e-01, 2.31573281851897e-01, 2.27435099609483e-01, 2.23188744188610e-01, 2.18867138561549e-01, 2.14503689850082e-01, 2.10131522535301e-01, 2.05782779638097e-01, 2.01488013269453e-01, 1.97275678271676e-01, 1.93171735208271e-01, 1.89199362300794e-01, 1.85378770455518e-01, 1.81727111470995e-01, 1.78258466898729e-01, 1.74983903739607e-01, 1.71911583010898e-01, 1.69046907976544e-01, 1.66392700242284e-01, 1.63949393729034e-01, 1.61715238531933e-01, 1.59686508669292e-01, 1.57857709591317e-01, 1.56221782959481e-01, 1.54770307561295e-01, 1.53493696253355e-01, 1.52381389508609e-01, 1.51422046481116e-01, 1.50603734509177e-01, 1.49914117688629e-01, 1.49340644609508e-01, 1.48870734621898e-01, 1.48491961153347e-01, 1.48192229721500e-01, 1.47959947459176e-01, 1.47784180280132e-01, 1.47654793341372e-01, 1.47562570265132e-01, 1.47499306712362e-01, 1.47457874363971e-01, 1.47432252150848e-01, 1.47417522634280e-01, 1.47409832705636e-01, 1.47406319160865e-01, 1.47405001116544e-01, 1.47404642577252e-01, 1.47404589657629e-01, 1.47404585493090e-01, 1.47404509636939e-01, 1.47403995700863e-01, 1.47402106425071e-01, 1.47397070367678e-01, 1.47386049374711e-01, 1.47364943986856e-01, 1.47328245082859e-01, 1.47268939407661e-01, 1.47178475848921e-01, 1.47046798322648e-01, 1.46862449797352e-01, 1.46612750209531e-01, 1.46284048701304e-01, 1.45862047688615e-01, 1.45332192762387e-01, 1.44680118448196e-01, 1.43892135625654e-01, 1.42955742269800e-01, 1.41860135545990e-01, 1.40596700637703e-01, 1.39159450471134e-01, 1.37545391092887e-01, 1.35754790068527e-01, 1.33791329891503e-01, 1.31662134763939e-01, 1.29377666731088e-01, 1.26951495327294e-01, 1.24399952829227e-01, 1.21741694130558e-01, 1.18997185494807e-01, 1.16188149573206e-01, 1.13336994924364e-01, 1.10466256945591e-01, 1.07598073951365e-01, 1.04753717598056e-01, 1.01953191520575e-01, 9.92149064866137e-02, 9.65554351004982e-02, 9.39893444967215e-02, 9.15291017959650e-02, 8.91850444378868e-02, 8.69654058207665e-02, 8.48763858680925e-02, 8.29222560879010e-02, 8.11054892626597e-02, 7.94269049595329e-02, 7.78858234138013e-02, 7.64802218430134e-02, 7.52068887550177e-02, 7.40615732264551e-02, 7.30391273949262e-02, 7.21336414925116e-02, 7.13385716180301e-02, 7.06468610709443e-02, 7.00510564335346e-02, 6.95434196915770e-02, 6.91160375483353e-02, 6.87609287479169e-02, 6.84701497272684e-02, 6.82358983136737e-02, 6.80506145348865e-02, 6.79070769743753e-02, 6.77984925483148e-02, 6.77185771623670e-02, 6.76616244738231e-02, 6.76225599713991e-02, 6.75969778044490e-02, 6.75811582375623e-02, 6.75720642468640e-02, 6.75673165641424e-02, 6.75651473543617e-02, 6.75643336162679e-02, 6.75641122604755e-02, 6.75640795888466e-02, 6.75640769949748e-02, 6.75640294589375e-02, 6.75637073957525e-02, 6.75625234643054e-02, 6.75593675906523e-02, 6.75524613068301e-02, 6.75392359758558e-02, 6.75162402334823e-02, 6.74790816427517e-02, 6.74224071445746e-02, 6.73399263277492e-02, 6.72244807476190e-02, 6.70681614008704e-02, 6.68624749349505e-02, 6.65985571868060e-02, 6.62674302149018e-02, 6.58602961897916e-02, 6.53688585040264e-02, 6.47856574967829e-02, 6.41044055736089e-02, 6.33203045858917e-02, 6.24303274580083e-02, 6.14334464898079e-02, 6.03307926771972e-02, 5.91257337779556e-02, 5.78238635065453e-02, 5.64328997905861e-02, 5.49624959442720e-02, 5.34239743308671e-02, 5.18299970472373e-02, 5.01941919184841e-02, 4.85307543304128e-02, 4.68540459906315e-02, 4.51782105922920e-02, 4.35168237355885e-02, 4.18825907176661e-02, 4.02871014583184e-02, 3.87406474314957e-02, 3.72521014216396e-02, 3.58288574025298e-02, 3.44768248970162e-02, 3.32004698909909e-02, 3.20028929107981e-02, 3.08859343837700e-02, 2.98502978438613e-02, 2.88956826368327e-02, 2.80209191322847e-02, 2.72241007997334e-02, 2.65027087967891e-02, 2.58537260176013e-02, 2.52737388746578e-02, 2.47590263398525e-02, 2.43056368030459e-02, 2.39094540207123e-02, 2.35662538360977e-02, 2.32717535196514e-02, 2.30216555800631e-02, 2.28116877642464e-02, 2.26376407137874e-02, 2.24954043825178e-02, 2.23810038758820e-02, 2.22906348807399e-02, 2.22206983674553e-02, 2.21678337994629e-02, 2.21289497254471e-02, 2.21012503790341e-02, 2.20822567968373e-02, 2.20698209940302e-02, 2.20621319093830e-02, 2.20577121301648e-02, 2.20554048045702e-02, 2.20543506135780e-02, 2.20539551574390e-02, 2.20538475844450e-02, 2.20538317069483e-02, 2.20538304348494e-02, 2.20538069768832e-02, 2.20536480461089e-02, 2.20530638061387e-02, 2.20515064852354e-02, 2.20480985722791e-02, 2.20415729530590e-02, 2.20302277931061e-02, 2.20118989088363e-02, 2.19839522688972e-02, 2.19432990835404e-02, 2.18864355923891e-02, 2.18095090849850e-02, 2.17084108086727e-02, 2.15788951998892e-02, 2.14167233095566e-02, 2.12178264410117e-02, 2.09784839912910e-02, 2.06955074772956e-02, 2.03664209691325e-02, 1.99896269217881e-02, 1.95645459408436e-02, 1.90917195380127e-02, 1.85728665138960e-02, 1.80108862085415e-02, 1.74098053306757e-02, 1.67746691494717e-02, 1.61113821715837e-02, 1.54265075890674e-02, 1.47270382313831e-02, 1.40201539090139e-02, 1.33129804766394e-02, 1.26123646590521e-02, 1.19246761841723e-02, 1.12556458520920e-02, 1.06102453913781e-02, 9.99261226337822e-03, 9.40601943800168e-03, 8.85288637527815e-03, 8.33482377159278e-03, 7.85270250922743e-03, 7.40673765803557e-03, 6.99658068645270e-03, 6.62141541105932e-03, 6.28005407789978e-03, 5.97102933761207e-03, 5.69267732709733e-03, 5.44320807663645e-03, 5.22076199394128e-03, 5.02345363953438e-03, 4.84940525100619e-03, 4.69677222493914e-03, 4.56376191026742e-03, 4.44864641510456e-03, 4.34977018025862e-03, 4.26555345294055e-03, 4.19449307351449e-03, 4.13516196977604e-03, 4.08620841284879e-03, 4.04635570598977e-03, 4.01440272485409e-03, 3.98922540342284e-03, 3.96977918581243e-03, 3.95510222037920e-03, 3.94431896810055e-03, 3.93664378286296e-03, 3.93138391716441e-03, 3.92794143124853e-03, 3.92581346661380e-03, 3.92459047111693e-03, 3.92395206449692e-03, 3.92366039574910e-03, 3.92355098591324e-03, 3.92352122384665e-03, 3.92351683119685e-03, 3.92351647614575e-03, 3.92350988306411e-03, 3.92346521516629e-03, 3.92330101651162e-03, 3.92286335029723e-03, 3.92190565937511e-03, 3.92007207219258e-03, 3.91688502340060e-03, 3.91173814066015e-03, 3.90389530454982e-03, 3.89249686072013e-03, 3.87657382593076e-03, 3.85507073300140e-03, 3.82687747418764e-03, 3.78949215334501e-03, 3.74462208117876e-03, 3.68985644399089e-03, 3.62434380220122e-03, 3.54743948578751e-03, 3.45876054980465e-03, 3.35823229038290e-03, 3.24612177786854e-03, 3.12305440633054e-03, 2.99001054763082e-03, 2.84830096945043e-03, 2.69952157561093e-03, 2.54549002381310e-03, 2.38816859971180e-03, 2.22957911213228e-03, 2.07171631342540e-03, 1.91646633240516e-03, 1.76553585003347e-03, 1.62039638617806e-03, 1.48224632668637e-03, 1.35199147229415e-03, 1.23024318950549e-03, 1.11733188551355e-03, 1.01333263297398e-03, 9.18099369838998e-04, 8.31304149663481e-04, 7.52478311461857e-04, 6.81053030969991e-04, 6.16397358232490e-04, 5.57852431258610e-04, 5.04761032081191e-04, 4.56492009558218e-04, 4.12459359992567e-04, 3.72135986657085e-04, 3.35062364863248e-04, 3.00850485297488e-04, 2.69183548660124e-04, 2.39812006858742e-04, 2.12546696936045e-04, 1.87249924509914e-04, 1.63825374297488e-04, 1.42207668803351e-04, 1.22352288769889e-04, 1.04226422680043e-04, 8.78011391325373e-05, 7.30450959167683e-05, 5.99198369251042e-05, 4.83765985190843e-05, 3.83544579936057e-05, 2.97796083845725e-05, 2.25739560235638e-05, 1.66205973197719e-05, 1.18207812374038e-05, 8.05704146245034e-06, 5.20503738243860e-06, 3.13576064137067e-06, 1.71792240521995e-06, 8.20528395450992e-07, 3.15705039000136e-07, -7.10542735760100e-15, -7.10542735760100e-15],
[ 1.40000000000000e+01, 1.39999999928760e+01, 1.39999997725868e+01, 1.39999982800932e+01, 1.39999927934958e+01, 1.39999781679871e+01, 1.39999461591278e+01, 1.39998848533853e+01, 1.39997782266997e+01, 1.39996058480449e+01, 1.39993427406337e+01, 1.39989594087293e+01, 1.39984220331065e+01, 1.39976928332333e+01, 1.39967305893665e+01, 1.39954913131356e+01, 1.39939290509716e+01, 1.39919968010243e+01, 1.39896475211077e+01, 1.39868352027638e+01, 1.39835159847567e+01, 1.39796492782141e+01, 1.39751988751669e+01, 1.39701340123745e+01, 1.39644303630064e+01, 1.39580709299438e+01, 1.39510468161440e+01, 1.39433578496624e+01, 1.39350130435752e+01, 1.39260308742048e+01, 1.39164393647377e+01, 1.39062759655612e+01, 1.38955872273863e+01, 1.38844282684199e+01, 1.38728620423654e+01, 1.38609584197103e+01, 1.38487931003779e+01, 1.38364463811419e+01, 1.38240018059736e+01, 1.38115447314674e+01, 1.37991608424740e+01, 1.37869346548996e+01, 1.37749480432180e+01, 1.37632788295723e+01, 1.37519994694553e+01, 1.37411758659610e+01, 1.37308663406506e+01, 1.37211207843463e+01, 1.37119800058710e+01, 1.37034752910777e+01, 1.36956281786675e+01, 1.36884504534451e+01, 1.36819443519709e+01, 1.36761029701780e+01, 1.36709108575378e+01, 1.36663447778936e+01, 1.36623746132101e+01, 1.36589643832890e+01, 1.36560733520328e+01, 1.36536571891496e+01, 1.36516691553189e+01, 1.36500612787992e+01, 1.36487854922554e+01, 1.36477947001973e+01, 1.36470437498181e+01, 1.36464902811280e+01, 1.36460954360246e+01, 1.36458244102164e+01, 1.36456468366056e+01, 1.36455369937081e+01, 1.36454738378080e+01, 1.36454408626645e+01, 1.36454257955760e+01, 1.36454201433144e+01, 1.36454186057549e+01, 1.36454183788141e+01, 1.36454183490443e+01, 1.36454176554424e+01, 1.36454129561810e+01, 1.36453956810997e+01, 1.36453496313141e+01, 1.36452488486197e+01, 1.36450558228933e+01, 1.36447201031894e+01, 1.36441773664568e+01, 1.36433489844318e+01, 1.36421421147400e+01, 1.36404503270129e+01, 1.36381547594214e+01, 1.36351257859525e+01, 1.36312251605099e+01, 1.36263085909268e+01, 1.36202286845464e+01, 1.36128381973350e+01, 1.36039935105315e+01, 1.35935582524559e+01, 1.35814069780493e+01, 1.35674288147027e+01, 1.35515309797700e+01, 1.35336420728521e+01, 1.35137150447793e+01, 1.34917297458143e+01, 1.34676949588315e+01, 1.34416498301121e+01, 1.34136646218738e+01, 1.33838407273554e+01, 1.33523099112901e+01, 1.33192327653307e+01, 1.32847963980885e+01, 1.32492114109626e+01, 1.32127082414901e+01, 1.31755329830900e+01, 1.31379428115892e+01, 1.31002011631773e+01, 1.30625728145391e+01, 1.30253190138412e+01, 1.29886928017515e+01, 1.29529346461381e+01, 1.29182684943048e+01, 1.28848983244765e+01, 1.28530052555499e+01, 1.28227452523712e+01, 1.27942474440883e+01, 1.27676130561088e+01, 1.27429149421039e+01, 1.27201976912355e+01, 1.26994782769894e+01, 1.26807472071814e+01, 1.26639701293233e+01, 1.26490898410996e+01, 1.26360286517921e+01, 1.26246910368402e+01, 1.26149665242020e+01, 1.26067327478110e+01, 1.25998586003574e+01, 1.25942074151227e+01, 1.25896401049770e+01, 1.25860181862760e+01, 1.25832066166076e+01, 1.25810763784100e+01, 1.25795067456350e+01, 1.25783871779129e+01, 1.25776187960525e+01, 1.25771154039819e+01, 1.25768040350697e+01, 1.25766250147431e+01, 1.25765315459025e+01, 1.25764888382818e+01, 1.25764728170275e+01, 1.25764684588436e+01, 1.25764678155843e+01, 1.25764677587967e+01, 1.25764666460724e+01, 1.25764591071955e+01, 1.25764313933986e+01, 1.25763575179168e+01, 1.25761958398095e+01, 1.25758861937785e+01, 1.25753476756982e+01, 1.25744771771510e+01, 1.25731487434744e+01, 1.25712138088549e+01, 1.25685023390250e+01, 1.25648248873142e+01, 1.25599755434387e+01, 1.25537357268830e+01, 1.25458787485491e+01, 1.25361750362382e+01, 1.25243978923285e+01, 1.25103296267149e+01, 1.24937678857798e+01, 1.24745319800654e+01, 1.24524690006706e+01, 1.24274595085125e+01, 1.23994225827572e+01, 1.23683200261106e+01, 1.23341595461461e+01, 1.22969967637834e+01, 1.22569359419900e+01, 1.22141293783008e+01, 1.21687754612894e+01, 1.21211154500500e+01, 1.20714290927365e+01, 1.20200292506639e+01, 1.19672557342476e+01, 1.19134685829925e+01, 1.18590410322231e+01, 1.18043524043984e+01, 1.17497811444703e+01, 1.16956981899455e+01, 1.16424608309600e+01, 1.15904071777927e+01, 1.15398513163989e+01, 1.14910791994915e+01, 1.14443452931318e+01, 1.13998699774032e+01, 1.13578376843156e+01, 1.13183957458242e+01, 1.12816539185747e+01, 1.12476845484555e+01, 1.12165233360721e+01, 1.11881706628824e+01, 1.11625934362056e+01, 1.11397274091524e+01, 1.11194799284385e+01, 1.11017330589534e+01, 1.10863470289212e+01, 1.10731639337069e+01, 1.10620116301021e+01, 1.10527077466644e+01, 1.10450637298904e+01, 1.10388888411977e+01, 1.10339940164643e+01, 1.10301954987675e+01, 1.10273181564758e+01, 1.10251984033562e+01, 1.10236866450788e+01, 1.10226491874695e+01, 1.10219695558728e+01, 1.10215491916449e+01, 1.10213075104796e+01, 1.10211813272210e+01, 1.10211236721551e+01, 1.10211020436050e+01, 1.10210961601013e+01, 1.10210952917080e+01, 1.10210952220521e+01, 1.10210939365476e+01, 1.10210852270608e+01, 1.10210532099996e+01, 1.10209678637579e+01, 1.10207810829995e+01, 1.10204233655548e+01, 1.10198012607843e+01, 1.10187956904824e+01, 1.10172612329252e+01, 1.10150264374586e+01, 1.10118952109099e+01, 1.10076492879300e+01, 1.10020517650880e+01, 1.09948516433528e+01, 1.09857892860981e+01, 1.09746026610295e+01, 1.09610341960662e+01, 1.09448380433616e+01, 1.09257875148664e+01, 1.09036824299271e+01, 1.08783561030811e+01, 1.08496817007988e+01, 1.08175777109465e+01, 1.07820122986202e+01, 1.07430063657517e+01, 1.07006351870887e+01, 1.06550285580484e+01, 1.06063694558283e+01, 1.05548912789172e+01, 1.05008737869826e+01, 1.04446379091485e+01, 1.03865396215707e+01, 1.03269631143071e+01, 1.02663134737706e+01, 1.02050091027403e+01, 1.01434740877989e+01, 1.00821307069100e+01, 1.00213922498179e+01, 9.96165630237136e+00, 9.90329862329268e+00, 9.84666771837143e+00, 9.79208019251154e+00, 9.73981693471434e+00, 9.69012016561605e+00, 9.64319135272163e+00, 9.59918997630991e+00, 9.55823311034598e+00, 9.52039576845090e+00, 9.48571195528819e+00, 9.45417635825999e+00, 9.42574661229187e+00, 9.40034607055031e+00, 9.37786701493758e+00, 9.35817424103881e+00, 9.34110895202963e+00, 9.32649289442569e+00, 9.31413266538070e+00, 9.30382411677423e+00, 9.29535677612006e+00, 9.28851819911404e+00, 9.28309816429271e+00, 9.27889261767333e+00, 9.27570727521144e+00, 9.27336079411166e+00, 9.27168743090635e+00, 9.27053911494994e+00, 9.26978688043621e+00, 9.26932161780583e+00, 9.26905412577611e+00, 9.26891446728551e+00, 9.26885065535834e+00, 9.26882671717048e+00, 9.26882020539061e+00, 9.26881924426510e+00, 9.26881917013384e+00, 9.26881783897444e+00, 9.26880882016984e+00, 9.26877566603313e+00, 9.26868728883969e+00, 9.26849387529329e+00, 9.26812345670624e+00, 9.26747926905456e+00, 9.26643801908099e+00, 9.26484915189906e+00, 9.26253519197690e+00, 9.25929320271553e+00, 9.25489737985056e+00, 9.24910276049938e+00, 9.24164999302342e+00, 9.23227107355792e+00, 9.22069591415951e+00, 9.20665956665639e+00, 9.18990988759030e+00, 9.17021539565300e+00, 9.14737304649759e+00, 9.12121563347876e+00, 9.09161851916489e+00, 9.05850541320770e+00, 9.02185293836045e+00, 8.98169376811115e+00, 8.93811817543320e+00, 8.89127390027193e+00, 8.84136432011865e+00, 8.78864498876240e+00, 8.73341868743040e+00, 8.67602920372176e+00, 8.61685411061275e+00, 8.55629685480656e+00, 8.49477847717005e+00, 8.43272927719218e+00, 8.37058070093584e+00, 8.30875768353495e+00, 8.24767162056835e+00, 8.18771408558449e+00, 8.12925136027091e+00, 8.07261980337822e+00, 8.01812205581832e+00, 7.96602406128632e+00, 7.91655287180197e+00, 7.86989520281287e+00, 7.82619670041481e+00, 7.78556188206852e+00, 7.74805471099681e+00, 7.71369976292540e+00, 7.68248394201559e+00, 7.65435870084596e+00, 7.62924271716905e+00, 7.60702497780281e+00, 7.58756821721192e+00, 7.57071265486338e+00, 7.55627997113885e+00, 7.54407745643878e+00, 7.53390226229901e+00, 7.52554567725375e+00, 7.51879734440153e+00, 7.51344933288337e+00, 7.50929997254820e+00, 7.50615736071549e+00, 7.50384245279300e+00, 7.50219165501604e+00, 7.50105884793402e+00, 7.50031678337650e+00, 7.49985781507981e+00, 7.49959394326387e+00, 7.49945617531426e+00, 7.49939322729661e+00, 7.49936961320814e+00, 7.49936318959278e+00, 7.49936224148022e+00, 7.49936216969199e+00, 7.49936089797068e+00, 7.49935228186843e+00, 7.49932060811654e+00, 7.49923617714965e+00, 7.49905140014919e+00, 7.49869752282821e+00, 7.49808210447818e+00, 7.49708736380800e+00, 7.49556948350998e+00, 7.49335894311488e+00, 7.49026192425889e+00, 7.48606280368317e+00, 7.48052771696252e+00, 7.47340914024499e+00, 7.46445139875479e+00, 7.45339697059840e+00, 7.43999341426894e+00, 7.42400071050290e+00, 7.40519877659541e+00, 7.38339488692015e+00, 7.35843072008505e+00, 7.33018875320383e+00, 7.29859773856619e+00, 7.26363702769716e+00, 7.22533955118053e+00, 7.18379331714246e+00, 7.13914135339101e+00, 7.09158008376882e+00, 7.04135619414845e+00, 6.98876210397060e+00, 6.93413021231597e+00, 6.87782613106190e+00, 6.82024115028684e+00, 6.76178420188156e+00, 6.70287359583392e+00, 6.64392879975727e+00, 6.58536251625532e+00, 6.52757328559333e+00, 6.47093880455898e+00, 6.41581010881252e+00, 6.36250671854658e+00, 6.31131279934359e+00, 6.26247434511455e+00, 6.21619735084387e+00, 6.17264691167487e+00, 6.13194716279381e+00, 6.09418196173914e+00, 6.05939621041092e+00, 6.02759771672515e+00, 5.99875950366651e+00, 5.97282248439185e+00, 5.94969843404685e+00, 5.92927320034643e+00, 5.91141010437505e+00, 5.89595348955801e+00, 5.88273237987034e+00, 5.87156420805770e+00, 5.86225857131327e+00, 5.85462096619211e+00, 5.84845644750882e+00, 5.84357314867626e+00, 5.83978559457919e+00, 5.83691773376427e+00, 5.83480561543576e+00, 5.83329963919632e+00, 5.83226631206276e+00, 5.83158945805295e+00, 5.83117084024677e+00, 5.83093017299788e+00, 5.83080452196677e+00, 5.83074711073587e+00, 5.83072557376276e+00, 5.83071971518026e+00, 5.83071885046563e+00, 5.83071878555123e+00, 5.83071764298415e+00, 5.83070990192372e+00, 5.83068144496104e+00, 5.83060558910300e+00, 5.83043958019224e+00, 5.83012165242309e+00, 5.82956877072516e+00, 5.82867515985023e+00, 5.82731170782992e+00, 5.82532631267924e+00, 5.82254521925380e+00, 5.81877536746137e+00, 5.81380774317072e+00, 5.80742168907002e+00, 5.79939009482165e+00, 5.78948534517606e+00, 5.77748586296721e+00, 5.76318304349723e+00, 5.74638834065407e+00, 5.72694023644821e+00, 5.70471080778442e+00, 5.67961160016136e+00, 5.65159852986119e+00, 5.62067556524219e+00, 5.58689698381777e+00, 5.55036806321593e+00, 5.51124413764758e+00, 5.46972803256336e+00, 5.42606597308195e+00, 5.38054214028989e+00, 5.33347211744869e+00, 5.28519552001811e+00, 5.23606813509388e+00, 5.18645390512868e+00, 5.13671707760907e+00, 5.08721480886135e+00, 5.03829046044564e+00, 4.99026776612919e+00, 4.94344598231625e+00, 4.89809607102180e+00, 4.85445790715200e+00, 4.81273845479060e+00, 4.77311082258575e+00, 4.73571408677866e+00, 4.70065376111908e+00, 4.66800279404302e+00, 4.63780298259853e+00, 4.61006670702561e+00, 4.58477890706422e+00, 4.56189923874927e+00, 4.54136436688598e+00, 4.52309036231179e+00, 4.50697518365123e+00, 4.49290123017869e+00, 4.48073795558854e+00, 4.47034453217079e+00, 4.46157255155618e+00, 4.45426874244771e+00, 4.44827767834686e+00, 4.44344444005795e+00, 4.43961718961993e+00, 4.43664960518524e+00, 4.43440312111153e+00, 4.43274891491116e+00, 4.43156958329301e+00, 4.43076045366862e+00, 4.43023048523145e+00, 4.42990272480208e+00, 4.42971429652208e+00, 4.42961592039048e+00, 4.42957097161748e+00, 4.42955410979614e+00, 4.42954952297333e+00, 4.42954884596795e+00, 4.42954879529949e+00, 4.42954790553593e+00, 4.42954187725836e+00, 4.42951971669440e+00, 4.42946064493667e+00, 4.42933136869993e+00, 4.42908379253731e+00, 4.42865326582789e+00, 4.42795744790011e+00, 4.42689586301088e+00, 4.42535020304425e+00, 4.42318541903456e+00, 4.42025162223031e+00, 4.41638679080756e+00, 4.41142024926503e+00, 4.40517685426687e+00, 4.39748178421630e+00, 4.38816579190461e+00, 4.37707074272813e+00, 4.36405522839737e+00, 4.34900002136126e+00, 4.33181312191562e+00, 4.31243415124709e+00, 4.29083786156994e+00, 4.26703656965763e+00, 4.24108137126837e+00, 4.21306205812136e+00, 4.18310573135861e+00, 4.15137417970216e+00, 4.11806016007770e+00, 4.08338277689201e+00, 4.04758219813995e+00, 4.01091396868762e+00, 3.97364318239021e+00, 3.93603875655533e+00, 3.89836801819597e+00, 3.86089176658446e+00, 3.82385992655975e+00, 3.78750785741384e+00, 3.75205333762944e+00, 3.71769420948740e+00, 3.68460664122797e+00, 3.65294394814724e+00, 3.62283590665711e+00, 3.59438849509984e+00, 3.56768399984315e+00, 3.54278143281904e+00, 3.51971721546711e+00, 3.49850609270709e+00, 3.47914224825842e+00, 3.46160059888990e+00, 3.44583824983725e+00, 3.43179609664788e+00, 3.41940056015983e+00, 3.40856544128211e+00, 3.39919388081983e+00, 3.39118040690728e+00, 3.38441304885220e+00, 3.37877549160243e+00, 3.37414923995684e+00, 3.37041575647592e+00, 3.36745853229531e+00, 3.36516504623512e+00, 3.36342856525052e+00, 3.36214973884679e+00, 3.36123794193544e+00, 3.36061232493036e+00, 3.36020253667821e+00, 3.35994909488554e+00, 3.35980338965277e+00, 3.35972731799288e+00, 3.35969256013243e+00, 3.35967952124431e+00, 3.35967597434921e+00, 3.35967545083488e+00, 3.35967541163934e+00, 3.35967472314997e+00, 3.35967005853332e+00, 3.35965291091032e+00, 3.35960720165298e+00, 3.35950716811099e+00, 3.35931559210102e+00, 3.35898244040509e+00, 3.35844398027679e+00, 3.35762242335001e+00, 3.35642614009062e+00, 3.35475047361663e+00, 3.35247916590779e+00, 3.34948639079588e+00, 3.34563936657342e+00, 3.34080149684865e+00, 3.33483596216045e+00, 3.32760965814727e+00, 3.31899735057385e+00, 3.30888589552174e+00, 3.29717835701669e+00, 3.28379784666916e+00, 3.26869091243688e+00, 3.25183031742458e+00, 3.23321707461058e+00, 3.21288163813336e+00, 3.19088419366143e+00, 3.16731403585721e+00, 3.14228806605018e+00, 3.11594848414945e+00, 3.08845978249557e+00, 3.06000517390182e+00, 3.03078260103202e+00, 3.00100048017395e+00, 2.97087333086172e+00, 2.94061743540504e+00, 2.91044666068119e+00, 2.88056855946871e+00, 2.85118085052097e+00, 2.82246835558140e+00, 2.79460044787755e+00, 2.76772904110016e+00, 2.74198712201111e+00, 2.71748780573528e+00, 2.69432387271828e+00, 2.67256773209485e+00, 2.65227174875307e+00, 2.63346887055546e+00, 2.61617349686906e+00, 2.60038253799581e+00, 2.58607662530368e+00, 2.57322144203238e+00, 2.56176915350969e+00, 2.55165992201527e+00, 2.54282349541600e+00, 2.53518086001105e+00, 2.52864594707340e+00, 2.52312737981913e+00, 2.51853024351166e+00, 2.51475785667071e+00, 2.51171351645805e+00, 2.50930218678307e+00, 2.50743209400926e+00, 2.50601619279322e+00, 2.50497346392089e+00, 2.50423000729028e+00, 2.50371989656173e+00, 2.50338576744020e+00, 2.50317911889196e+00, 2.50306031549599e+00, 2.50299828912609e+00, 2.50296994869904e+00, 2.50295931721569e+00, 2.50295642519344e+00, 2.50295599833697e+00, 2.50295596629343e+00, 2.50295540230017e+00, 2.50295158116298e+00, 2.50293753426063e+00, 2.50290009039288e+00, 2.50281814548826e+00, 2.50266121151473e+00, 2.50238830302438e+00, 2.50194721304567e+00, 2.50127422221290e+00, 2.50029427584283e+00, 2.49892165279018e+00, 2.49706113697300e+00, 2.49460968716824e+00, 2.49145858293603e+00, 2.48749600452636e+00, 2.48260998293037e+00, 2.47669163387005e+00, 2.46963856793031e+00, 2.46135835005731e+00, 2.45177186734979e+00, 2.44081645655870e+00, 2.42844864385448e+00, 2.41464636054665e+00, 2.39941052003455e+00, 2.38276587271165e+00, 2.36476109496255e+00, 2.34546811264298e+00, 2.32498070434961e+00, 2.30341247064629e+00, 2.28089428765821e+00, 2.25757138352651e+00, 2.23360018236295e+00, 2.20914505302729e+00, 2.18437508192856e+00, 2.15946096436111e+00, 2.13457208241462e+00, 2.10987381347037e+00, 2.08552509450643e+00, 2.06167625488938e+00, 2.03846712342315e+00, 2.01602541251306e+00, 1.99446538140595e+00, 1.97388677987522e+00, 1.95437407234208e+00, 1.93599593987068e+00, 1.91880505392677e+00, 1.90283811176205e+00, 1.88811611938394e+00, 1.87464490479639e+00, 1.86241584183890e+00, 1.85140676356037e+00, 1.84158304348035e+00, 1.83289882301575e+00, 1.82529836343099e+00, 1.81871750056575e+00, 1.81308518005269e+00, 1.80832504962085e+00, 1.80435708339148e+00, 1.80109921094953e+00, 1.79846892166838e+00, 1.79638481261363e+00, 1.79476804672986e+00, 1.79354368729557e+00, 1.79264187514091e+00, 1.79199881709878e+00, 1.79155755772611e+00, 1.79126851147780e+00, 1.79108973909577e+00, 1.79098695971794e+00, 1.79093329874683e+00, 1.79090878040653e+00, 1.79089958269835e+00, 1.79089708069566e+00, 1.79089671140520e+00, 1.79089668356500e+00, 1.79089619198119e+00, 1.79089286142779e+00, 1.79088061794947e+00, 1.79084798124137e+00, 1.79077655614686e+00, 1.79063976679465e+00, 1.79040188348279e+00, 1.79001738609535e+00, 1.78943070052850e+00, 1.78857633619294e+00, 1.78737944298331e+00, 1.78575679498876e+00, 1.78361819555332e+00, 1.78086828409483e+00, 1.77740870957266e+00, 1.77314061911553e+00, 1.76796739381319e+00, 1.76179754802406e+00, 1.75454769493670e+00, 1.74614547080822e+00, 1.73653230448847e+00, 1.72566591849077e+00, 1.71352245357028e+00, 1.70009812059910e+00, 1.68541030102756e+00, 1.66949803944087e+00, 1.65242189733140e+00, 1.63426316466398e+00, 1.61512245351890e+00, 1.59511772453269e+00, 1.57438182065163e+00, 1.55305960267394e+00, 1.53130479617469e+00, 1.50927666887275e+00, 1.48713666075335e+00, 1.46504508607694e+00, 1.44315801698465e+00, 1.42162444343425e+00, 1.40058378483164e+00, 1.38016380651370e+00, 1.36047897097513e+00, 1.34162923121986e+00, 1.32369925347705e+00, 1.30675804000499e+00, 1.29085891059599e+00, 1.27603979396189e+00, 1.26232377722579e+00, 1.24971986269173e+00, 1.23822388506901e+00, 1.22781954841931e+00, 1.21847954927972e+00, 1.21016675977847e+00, 1.20283545132965e+00, 1.19643254507782e+00, 1.19089887927522e+00, 1.18617048602441e+00, 1.18217987031100e+00, 1.17885728315594e+00, 1.17613197834889e+00, 1.17393343900663e+00, 1.17219255662693e+00, 1.17084274190142e+00, 1.16982094381545e+00, 1.16906855194270e+00, 1.16853215667938e+00, 1.16816414366275e+00, 1.16792310183668e+00, 1.16777402945379e+00, 1.16768832848910e+00, 1.16764358510956e+00, 1.16762314154644e+00, 1.16761547246926e+00, 1.16761338629519e+00, 1.16761307838030e+00, 1.16761305504077e+00, 1.16761264125140e+00, 1.16760983776840e+00, 1.16759953188152e+00, 1.16757206027200e+00, 1.16751193983967e+00, 1.16739680398627e+00, 1.16719658792876e+00, 1.16687300209646e+00, 1.16637932737764e+00, 1.16566055967581e+00, 1.16465392359058e+00, 1.16328976571453e+00, 1.16149282676732e+00, 1.15918387845863e+00, 1.15628169570151e+00, 1.15270531801023e+00, 1.14837653635782e+00, 1.14322252451558e+00, 1.13717851829084e+00, 1.13019043363492e+00, 1.12221730684919e+00, 1.11323343846300e+00, 1.10323012785145e+00, 1.09221689884708e+00, 1.08022213737227e+00, 1.06729308963460e+00, 1.05349520210361e+00, 1.03891082008530e+00, 1.02363729751686e+00, 1.00778460368868e+00, 9.91472540134033e-01, 9.74827700516379e-01, 9.57980316330550e-01, 9.41061130918520e-01, 9.24198434034306e-01, 9.07515370334179e-01, 8.91127609890265e-01, 8.75141439814362e-01, 8.59652306176027e-01, 8.44743807247255e-01, 8.30487114853772e-01, 8.16940781728858e-01, 8.04150879964759e-01, 7.92151408948200e-01, 7.80964909976435e-01, 7.70603228101988e-01, 7.61068368452836e-01, 7.52353403083491e-01, 7.44443394179214e-01, 7.37316309187774e-01, 7.30943912434146e-01, 7.25292625449399e-01, 7.20324354292138e-01, 7.15997286403804e-01, 7.12266662008588e-01, 7.09085525844234e-01, 7.06405464274013e-01, 7.04177330836000e-01, 7.02351960338488e-01, 7.00880868055972e-01, 6.99716926795301e-01, 6.98815010971149e-01, 6.98132593731808e-01, 6.97630280942978e-01, 6.97272264745047e-01, 6.97026679619407e-01, 6.96865845499914e-01, 6.96766385384276e-01, 6.96709208957635e-01, 6.96679358656844e-01, 6.96665720024042e-01, 6.96660603736989e-01, 6.96659211986164e-01, 6.96659006566774e-01, 6.96658990897475e-01, 6.96658711790946e-01, 6.96656820804995e-01, 6.96649869368517e-01, 6.96631339605368e-01, 6.96590788759137e-01, 6.96513133203670e-01, 6.96378102578642e-01, 6.96159892195221e-01, 6.95827038976443e-01, 6.95342542525839e-01, 6.94664248053062e-01, 6.93745501411511e-01, 6.92536078056232e-01, 6.90983377113099e-01, 6.89033859010435e-01, 6.86634690647184e-01, 6.83735546622611e-01, 6.80290499778993e-01, 6.76259920681883e-01, 6.71612295342566e-01, 6.66325865141349e-01, 6.60389993962934e-01, 6.53806175949619e-01, 6.46588613240054e-01, 6.38764315958667e-01, 6.30372705012297e-01, 6.21464729597786e-01, 6.12101542825627e-01, 6.02352807413578e-01, 5.92294726116426e-01, 5.82007906173537e-01, 5.71575172283950e-01, 5.61079438295286e-01, 5.50601734904912e-01, 5.40219471135895e-01, 5.30004983721463e-01, 5.20024403563980e-01, 5.10336844716562e-01, 5.00993900943344e-01, 4.92039419227595e-01, 4.83509509222831e-01, 4.75432742494993e-01, 4.67830494846471e-01, 4.60717388072233e-01, 4.54101793058609e-01, 4.47986363113797e-01, 4.42368573863471e-01, 4.37241253241426e-01, 4.32593091529640e-01, 4.28409126765901e-01, 4.24671205014416e-01, 4.21358417977167e-01, 4.18447522277337e-01, 4.15913345607059e-01, 4.13729184904323e-01, 4.11867200978069e-01, 4.10298812647895e-01, 4.08995091670555e-01, 4.07927157621501e-01, 4.07066569651249e-01, 4.06385709804216e-01, 4.05858150549392e-01, 4.05458997504894e-01, 4.05165197189243e-01, 4.04955799159085e-01, 4.04812162138448e-01, 4.04718094812473e-01, 4.04659923720848e-01, 4.04626483137394e-01, 4.04609024724149e-01, 4.04601047964437e-01, 4.04598055628361e-01, 4.04597241642126e-01, 4.04597121499551e-01, 4.04597112270507e-01, 4.04596947031266e-01, 4.04595827511928e-01, 4.04591712057631e-01, 4.04580741904446e-01, 4.04556734684345e-01, 4.04510760636621e-01, 4.04430819676842e-01, 4.04301636095871e-01, 4.04104585998496e-01, 4.03817770422362e-01, 4.03416244170398e-01, 4.02872406459483e-01, 4.02156554285263e-01, 4.01237592816042e-01, 4.00083889166144e-01, 3.98664246865611e-01, 3.96948968734248e-01, 3.94910966510379e-01, 3.92526867513196e-01, 3.89778062924462e-01, 3.86651640118490e-01, 3.83141143655022e-01, 3.79247116552918e-01, 3.74977385167597e-01, 3.70347066575425e-01, 3.65378295504369e-01, 3.60099686680574e-01, 3.54545565940505e-01, 3.48755017728950e-01, 3.42770806147410e-01, 3.36638230738979e-01, 3.30403976641469e-01, 3.24115012336485e-01, 3.17817578153800e-01, 3.11556296695155e-01, 3.05373423748671e-01, 2.99308246747527e-01, 2.93396628165029e-01, 2.87670683959810e-01, 2.82158582609203e-01, 2.76884447873352e-01, 2.71868347907500e-01, 2.67126354271204e-01, 2.62670655961598e-01, 2.58509715749618e-01, 2.54648458246187e-01, 2.51088481303922e-01, 2.47828284401535e-01, 2.44863509132372e-01, 2.42187188656896e-01, 2.39790003904663e-01, 2.37660545499661e-01, 2.35785580791791e-01, 2.34150325939515e-01, 2.32738722977755e-01, 2.31533721678041e-01, 2.30517565273376e-01, 2.29672078816503e-01, 2.28978957409959e-01, 2.28420051086854e-01, 2.27977641611819e-01, 2.27634705909264e-01, 2.27375159543179e-01, 2.27184073939021e-01, 2.27047860147946e-01, 2.26954413027347e-01, 2.26893210017442e-01, 2.26855360251090e-01, 2.26833601048696e-01, 2.26822240968652e-01, 2.26817050499685e-01, 2.26815103376751e-01, 2.26814573720596e-01, 2.26814495548155e-01, 2.26814489496073e-01, 2.26814380550602e-01, 2.26813642478358e-01, 2.26810929238052e-01, 2.26803696797610e-01, 2.26787869061162e-01, 2.26757558113169e-01, 2.26704850548314e-01, 2.26619670198364e-01, 2.26489727160356e-01, 2.26300560401340e-01, 2.26035680186380e-01, 2.25676813433704e-01, 2.25204252021024e-01, 2.24597300395136e-01, 2.23834813836159e-01, 2.22895813629220e-01, 2.21760160206600e-01, 2.20409259769966e-01, 2.18826775422778e-01, 2.16999310741086e-01, 2.14917032051130e-01, 2.12574196802356e-01, 2.09969558475885e-01, 2.07106624505023e-01, 2.03993751279887e-01, 2.00644069666755e-01, 1.97075243971177e-01, 1.93309076905269e-01, 1.89370981048784e-01, 1.85289343742841e-01, 1.81094816504356e-01, 1.76819561305742e-01, 1.72438418201059e-01, 1.68106958274915e-01, 1.63792190172884e-01, 1.59525014004025e-01, 1.55334470857992e-01, 1.51247287893522e-01, 1.47287523994818e-01, 1.43476316676901e-01, 1.39831726435686e-01, 1.36368671028228e-01, 1.33098939358343e-01, 1.30031272795462e-01, 1.27171500867437e-01, 1.24522718272237e-01, 1.22085490925746e-01, 1.19858080137154e-01, 1.17836675790214e-01, 1.16015631412640e-01, 1.14387696052855e-01, 1.12944239791796e-01, 1.11675471368361e-01, 1.10570647696488e-01, 1.09618275943701e-01, 1.08806309304859e-01, 1.08122337653500e-01, 1.07553773930480e-01, 1.07088036503184e-01, 1.06712726889311e-01, 1.06415801289360e-01, 1.06185733423713e-01, 1.06011665331954e-01, 1.05883542163610e-01, 1.05792226650228e-01, 1.05729588951801e-01, 1.05688567937063e-01, 1.05663200673124e-01, 1.05648617919456e-01, 1.05641004670137e-01, 1.05637526173542e-01, 1.05636221279067e-01, 1.05635866317046e-01, 1.05635813925428e-01, 1.05635809837223e-01, 1.05635735813495e-01, 1.05635234292414e-01, 1.05633390656504e-01, 1.05628476266345e-01, 1.05617721580062e-01, 1.05597126358262e-01, 1.05561315128874e-01, 1.05503445494337e-01, 1.05415176147359e-01, 1.05286700531460e-01, 1.05106850825136e-01, 1.04863275223272e-01, 1.04542689234441e-01, 1.04131198849650e-01, 1.03614689979111e-01, 1.02979274602351e-01, 1.02211779841523e-01, 1.01300261959045e-01, 1.00234523495789e-01, 9.90066088519619e-02, 9.76112520149801e-02, 9.60462502393362e-02, 9.43127395367505e-02, 9.24153519124200e-02, 9.03622402388056e-02, 8.81649641212601e-02, 8.58382385104388e-02, 8.33995554427460e-02, 8.08686973672650e-02, 7.82671672998365e-02, 7.56175659066098e-02, 7.29429481301818e-02, 7.02661919495515e-02, 6.76094093942172e-02, 6.49934253475557e-02, 6.24373435115775e-02, 5.99582118286950e-02, 5.75707923727133e-02, 5.52874338899230e-02, 5.31180393330501e-02, 5.10701162563301e-02, 4.91488950108181e-02, 4.73574982945504e-02, 4.56971456216895e-02, 4.41673774213509e-02, 4.27662854491775e-02, 4.14907386710031e-02, 4.03365964635380e-02, 3.92989036276479e-02, 3.83720641419369e-02, 3.75499926767002e-02, 3.68262445736303e-02, 3.61941262519547e-02, 3.56467888352761e-02, 3.51773082354416e-02, 3.47787550224314e-02, 3.44442572030914e-02, 3.41670585807350e-02, 3.39405747303356e-02, 3.37584478615582e-02, 3.36146010183285e-02, 3.35032912452782e-02, 3.34191606035876e-02, 3.33572833025393e-02, 3.33132067799671e-02, 3.32829843520699e-02, 3.32631970785755e-02, 3.32509627500048e-02, 3.32439303739918e-02, 3.32402591729828e-02, 3.32385818469607e-02, 3.32379526363660e-02, 3.32377814769913e-02, 3.32377562142661e-02, 3.32377542262030e-02, 3.32377180140710e-02, 3.32374726722229e-02, 3.32365707794242e-02, 3.32341667395859e-02, 3.32289059389623e-02, 3.32188323392533e-02, 3.32013188422167e-02, 3.31730246632791e-02, 3.31298838031007e-02, 3.30671284213100e-02, 3.29793503882581e-02, 3.28606034027388e-02, 3.27045467147915e-02, 3.25046296176925e-02, 3.22543134692914e-02, 3.19473251499875e-02, 3.15779327404915e-02, 3.11412310831951e-02, 3.06334221344855e-02, 3.00520730279530e-02, 2.93963339495509e-02, 2.86670986051085e-02, 2.78670924265736e-02, 2.70008777100497e-02, 2.60747703675897e-02, 2.50966694410284e-02, 2.40758073200293e-02, 2.30224349760366e-02, 2.19474617233608e-02, 2.08620724262421e-02, 1.97773462887980e-02, 1.87039002866527e-02, 1.76515771223986e-02, 1.66291927784545e-02, 1.56443529482999e-02, 1.47033415729645e-02, 1.38110790740598e-02, 1.29711432012680e-02, 1.21858420483623e-02, 1.14563268695314e-02, 1.07827317835465e-02, 1.01643280747581e-02, 9.59968228530883e-03, 9.08680931065042e-03, 8.62331394929861e-03, 8.20651656094462e-03, 7.83356047089612e-03, 7.50150041297992e-03, 7.20737258368942e-03, 6.94824779290215e-03, 6.72126977848464e-03, 6.52368105956214e-03, 6.35283879516102e-03, 6.20622304797536e-03, 6.08143967494712e-03, 5.97621981401097e-03, 5.88841763744163e-03, 5.81600771578117e-03, 5.75708299560596e-03, 5.70985405122570e-03, 5.67264993390459e-03, 5.64392062541152e-03, 5.62224081740581e-03, 5.60631449961768e-03, 5.59497966117917e-03, 5.58721230264148e-03, 5.58212892821786e-03, 5.57898673901569e-03, 5.57718087232395e-03, 5.57623821541497e-03, 5.57580754609610e-03, 5.57564599338534e-03, 5.57560204771512e-03, 5.57559556144760e-03, 5.57559504648708e-03, 5.57558560921123e-03, 5.57552167059505e-03, 5.57528663033047e-03, 5.57466013656871e-03, 5.57328925950884e-03, 5.57066459088063e-03, 5.56610253802781e-03, 5.55873513390281e-03, 5.54750872649024e-03, 5.53119290139925e-03, 5.50840087778681e-03, 5.47762234291760e-03, 5.43726920435539e-03, 5.38573400728204e-03, 5.32145978660026e-03, 5.24301894134105e-03, 5.14919742767006e-03, 5.03907931316583e-03, 4.91212570892152e-03, 4.76824150416277e-03, 4.60782336185339e-03, 4.43178322464988e-03, 4.24154316241676e-03, 4.03899967157528e-03, 3.82645828767458e-03, 3.60654225919355e-03, 3.38208165918985e-03, 3.15599129715061e-03, 2.93114684002127e-03, 2.71026850572653e-03, 2.49582058003384e-03, 2.28993301930558e-03, 2.09434885683457e-03, 1.91039840926699e-03, 1.73899875574079e-03, 1.58067494070124e-03, 1.43559802315529e-03, 1.30363451821935e-03, 1.18440188175128e-03, 1.07732531311377e-03, 9.81692090299191e-04, 8.96700711599863e-04, 8.21503150111980e-04, 7.55239436974975e-04, 6.97064527386715e-04, 6.46167949616583e-04, 6.01787092167072e-04, 5.63215165101383e-04, 5.29804910941678e-04, 5.00969080679425e-04, 4.76178573014963e-04, 4.54958994197341e-04, 4.36886254409075e-04, 4.21581686367745e-04, 4.08707058090485e-04, 3.97959755620292e-04, 3.89068331834608e-04, 3.81788552122231e-04, 3.75900014025987e-04, 3.71203373109807e-04, 3.67518168753023e-04, 3.64681210257345e-04, 3.62545454843044e-04, 3.60979284995011e-04, 3.59866074795434e-04, 3.59103924079562e-04, 3.58605436941772e-04, 3.58297427524690e-04, 3.58120451307897e-04, 3.58028082366957e-04, 3.57985884765100e-04, 3.57970056041879e-04, 3.57965750344746e-04, 3.57965114838876e-04, 3.57965063921384e-04, 3.57964124999555e-04, 3.57957763720052e-04, 3.57934379827896e-04, 3.57872053330155e-04, 3.57735685716065e-04, 3.57474649510081e-04, 3.57021093448040e-04, 3.56289067223514e-04, 3.55174645711998e-04, 3.53557240272195e-04, 3.51302274321341e-04, 3.48265362418942e-04, 3.44298056406345e-04, 3.39255103698477e-04, 3.33003001939502e-04, 3.25429441289771e-04, 3.16453020902330e-04, 3.06032441667270e-04, 2.94174251314061e-04, 2.80938190146074e-04, 2.66439288660436e-04, 2.50846115404713e-04, 2.34374952544414e-04, 2.17280142040366e-04, 1.99841322041650e-04, 1.82348670019650e-04, 1.65087505401118e-04, 1.48323640615899e-04, 1.32290726268636e-04, 1.17180581465703e-04, 1.03137201895365e-04, 9.02548073664633e-05, 7.85798774085578e-05, 6.81165868066103e-05, 5.88344755865222e-05, 5.06768148884618e-05, 4.35682407662341e-05, 3.74208909441620e-05, 3.21529744340410e-05, 2.76424321334900e-05, 2.37884398792687e-05, 2.05186666324408e-05, 1.77477955229932e-05, 1.53841452146253e-05, 1.33713042096417e-05, 1.16931388003394e-05, 1.02713962455425e-05, 9.07168088382093e-06, 8.07245117151467e-06, 7.24883959790645e-06, 6.56343897809109e-06, 5.99042390447835e-06, 5.54537111341347e-06, 5.18840932706155e-06, 4.89618725474372e-06, 4.65666479954052e-06, 4.46068235149255e-06, 4.30131623119223e-06, 4.17315044387578e-06, 4.07166722382613e-06, 3.99284074958928e-06, 3.93294011757916e-06, 3.88850020982545e-06, 3.85638371191988e-06, 3.83385029856242e-06, 3.81858382603184e-06, 3.80867579075073e-06, 3.80258588883464e-06, 3.79909823955416e-06, 3.79728132671175e-06, 3.79645207715384e-06, 3.79614114187632e-06, 3.79605657663262e-06, 3.79604409239676e-06, 3.79604308697878e-06, 3.79602443345561e-06, 3.79589804211378e-06, 3.79543342354793e-06, 3.79419492269051e-06, 3.79148456985945e-06, 3.78629421327048e-06, 3.77726939326806e-06, 3.76268761215215e-06, 3.74045569451198e-06, 3.70813458872021e-06, 3.66300714205181e-06, 3.60221633144420e-06, 3.52303507433760e-06, 3.42354447724347e-06, 3.30550270177810e-06, 3.17631453761180e-06, 3.03184131311696e-06, 2.86771050461709e-06, 2.68449885076905e-06, 2.48461056706617e-06, 2.27146355769037e-06, 2.07777446625812e-06, 1.89279836604328e-06, 1.70273586874714e-06, 1.51157620464915e-06, 1.32850932388351e-06, 1.15547761403434e-06, 9.84997212682970e-07, 8.18859474449596e-07, 6.78580446944466e-07, 5.59869297589444e-07, 4.55100119367557e-07, 3.87755505926179e-07, 3.36634593978147e-07, 2.88919776991747e-07, 2.42214547085950e-07, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -3.55271367880050e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -3.55271367880050e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -3.55271367880050e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15, -1.77635683940025e-15],
[ 1.50000000000000e+01, 1.49999999921008e+01, 1.49999997478407e+01, 1.49999980929402e+01, 1.49999920093142e+01, 1.49999757923176e+01, 1.49999403003806e+01, 1.49998723235697e+01, 1.49997540940944e+01, 1.49995629575588e+01, 1.49992712189674e+01, 1.49988461722939e+01, 1.49982503169633e+01, 1.49974417590893e+01, 1.49963747899037e+01, 1.49950006287113e+01, 1.49932683130363e+01, 1.49911257145445e+01, 1.49885206559011e+01, 1.49854021010403e+01, 1.49817213893749e+01, 1.49774334832696e+01, 1.49724981975856e+01, 1.49668813802375e+01, 1.49605560134174e+01, 1.49535032064082e+01, 1.49457130526876e+01, 1.49371853263341e+01, 1.49279299955964e+01, 1.49179675349238e+01, 1.49073290208035e+01, 1.48960560014389e+01, 1.48842001355969e+01, 1.48718226017819e+01, 1.48589932851230e+01, 1.48457897557852e+01, 1.48322960590892e+01, 1.48186013435544e+01, 1.48047983584778e+01, 1.47909818571432e+01, 1.47772469450844e+01, 1.47636874148359e+01, 1.47503941091856e+01, 1.47374533541017e+01, 1.47249455002830e+01, 1.47129436088277e+01, 1.47015123120041e+01, 1.46907068747550e+01, 1.46805724766103e+01, 1.46711437273440e+01, 1.46624444232167e+01, 1.46544875441758e+01, 1.46472754861287e+01, 1.46408005164862e+01, 1.46350454357325e+01, 1.46299844229033e+01, 1.46255840386385e+01, 1.46218043559974e+01, 1.46186001865323e+01, 1.46159223672801e+01, 1.46137190733752e+01, 1.46119371209441e+01, 1.46105232258103e+01, 1.46094251853057e+01, 1.46085929531127e+01, 1.46079795804778e+01, 1.46075420012591e+01, 1.46072416429913e+01, 1.46070448513255e+01, 1.46069231207070e+01, 1.46068531298163e+01, 1.46068165859796e+01, 1.46067998882772e+01, 1.46067936243081e+01, 1.46067919203487e+01, 1.46067916688476e+01, 1.46067916358559e+01, 1.46067908671901e+01, 1.46067856593585e+01, 1.46067665147104e+01, 1.46067154812745e+01, 1.46066037916214e+01, 1.46063898763770e+01, 1.46060178252855e+01, 1.46054163557201e+01, 1.46044983337847e+01, 1.46031608771073e+01, 1.46012860516008e+01, 1.45987421574234e+01, 1.45953855826732e+01, 1.45910631875416e+01, 1.45856151671962e+01, 1.45788783289079e+01, 1.45706897080544e+01, 1.45608904386158e+01, 1.45493297864469e+01, 1.45358692476617e+01, 1.45203866096059e+01, 1.45027798679079e+01, 1.44829708900593e+01, 1.44609087142731e+01, 1.44365723727762e+01, 1.44099731323035e+01, 1.43811560525853e+01, 1.43502007771643e+01, 1.43172214906344e+01, 1.42823660023659e+01, 1.42458139481512e+01, 1.42077741362550e+01, 1.41684811006665e+01, 1.41281909590786e+01, 1.40871767033576e+01, 1.40457230735323e+01, 1.40041211808182e+01, 1.39626630500936e+01, 1.39216362477395e+01, 1.38813187479547e+01, 1.38419741713492e+01, 1.38038475059977e+01, 1.37671613954392e+01, 1.37321130524027e+01, 1.36988718329550e+01, 1.36675774844487e+01, 1.36383390626885e+01, 1.36112344992668e+01, 1.35863107887779e+01, 1.35635847571171e+01, 1.35430443656518e+01, 1.35246505010804e+01, 1.35083391966715e+01, 1.34940242268546e+01, 1.34816000135028e+01, 1.34709447786031e+01, 1.34619238743759e+01, 1.34543932184919e+01, 1.34482027591362e+01, 1.34431998926685e+01, 1.34392327559486e+01, 1.34361533164144e+01, 1.34338201860738e+01, 1.34321010909134e+01, 1.34308749349637e+01, 1.34300334083204e+01, 1.34294821006217e+01, 1.34291410954517e+01, 1.34289450364116e+01, 1.34288426715925e+01, 1.34287958992723e+01, 1.34287783532002e+01, 1.34287735802281e+01, 1.34287728757468e+01, 1.34287728135546e+01, 1.34287715949272e+01, 1.34287633385417e+01, 1.34287329871051e+01, 1.34286520806257e+01, 1.34284750152588e+01, 1.34281359004276e+01, 1.34275461347383e+01, 1.34265928034300e+01, 1.34251379797403e+01, 1.34230189897965e+01, 1.34200496755814e+01, 1.34160226633271e+01, 1.34107126156040e+01, 1.34038804147687e+01, 1.33952781938998e+01, 1.33846550996563e+01, 1.33717636406367e+01, 1.33563664459807e+01, 1.33382432334749e+01, 1.33171977657830e+01, 1.32930645591648e+01, 1.32657151027917e+01, 1.32350633500353e+01, 1.32010702572528e+01, 1.31637471715328e+01, 1.31231579068143e+01, 1.30794193969816e+01, 1.30327008729846e+01, 1.29832215754555e+01, 1.29312470803080e+01, 1.28770843772966e+01, 1.28210758953334e+01, 1.27635927090037e+01, 1.27050271850976e+01, 1.26457853348261e+01, 1.25862791274942e+01, 1.25269189973758e+01, 1.24681067412680e+01, 1.24102289642569e+01, 1.23536511900015e+01, 1.22987127130816e+01, 1.22457222372586e+01, 1.21949543162817e+01, 1.21466465934041e+01, 1.21009978214523e+01, 1.20581666359913e+01, 1.20182710484681e+01, 1.19813886228603e+01, 1.19475572971518e+01, 1.19167768090060e+01, 1.18890106826925e+01, 1.18641887312644e+01, 1.18422100239895e+01, 1.18229462640866e+01, 1.18062455159894e+01, 1.17919362148674e+01, 1.17798313842704e+01, 1.17697329809530e+01, 1.17614362796515e+01, 1.17547342054143e+01, 1.17494215176125e+01, 1.17452987485816e+01, 1.17421758014953e+01, 1.17398751169628e+01, 1.17382343262161e+01, 1.17371083206419e+01, 1.17363706826182e+01, 1.17359144406835e+01, 1.17356521323528e+01, 1.17355151795513e+01, 1.17354526037230e+01, 1.17354291292138e+01, 1.17354227435634e+01, 1.17354218010543e+01, 1.17354217254534e+01, 1.17354203302333e+01, 1.17354108774073e+01, 1.17353761277474e+01, 1.17352834973639e+01, 1.17350807752507e+01, 1.17346925274812e+01, 1.17340173279569e+01, 1.17329259363391e+01, 1.17312605216133e+01, 1.17288350045656e+01, 1.17254365641210e+01, 1.17208283208151e+01, 1.17147531755816e+01, 1.17069387437772e+01, 1.16971032835586e+01, 1.16849624755724e+01, 1.16702368691863e+01, 1.16526597715512e+01, 1.16319853224530e+01, 1.16079964732574e+01, 1.15805125752297e+01, 1.15493962836091e+01, 1.15145595005897e+01, 1.14759681130952e+01, 1.14336453286783e+01, 1.13876734722153e+01, 1.13381941731058e+01, 1.12854069423547e+01, 1.12295662058790e+01, 1.11709769198485e+01, 1.11099889422346e+01, 1.10469903700504e+01, 1.09824000738185e+01, 1.09166596708703e+01, 1.08502251793319e+01, 1.07835585874253e+01, 1.07171195599293e+01, 1.06513574863990e+01, 1.05867040544006e+01, 1.05235665054765e+01, 1.04623217018748e+01, 1.04033110988507e+01, 1.03468366820529e+01, 1.02931578943193e+01, 1.02424895436608e+01, 1.01950006566396e+01, 1.01508142203635e+01, 1.01100077425382e+01, 1.00726145520017e+01, 1.00386257606265e+01, 1.00079928096370e+01, 9.98063052734375e+00, 9.95642062934034e+00, 9.93521559503600e+00, 9.91684285520334e+00, 9.90110922369780e+00, 9.88780550278827e+00, 9.87671118611737e+00, 9.86759917693778e+00, 9.86024043286721e+00, 9.85440844296731e+00, 9.84988343946887e+00, 9.84645624583369e+00, 9.84393166578577e+00, 9.84213132494226e+00, 9.84089588791656e+00, 9.84008658911402e+00, 9.83958603443292e+00, 9.83929825296344e+00, 9.83914800155236e+00, 9.83907934963132e+00, 9.83905359579450e+00, 9.83904659011407e+00, 9.83904555608987e+00, 9.83904547633597e+00, 9.83904404421186e+00, 9.83903434135663e+00, 9.83899867257966e+00, 9.83890359229527e+00, 9.83869550928329e+00, 9.83829699724034e+00, 9.83760395585168e+00, 9.83648374798375e+00, 9.83477441654245e+00, 9.83228505937759e+00, 9.82879741202860e+00, 9.82406865578044e+00, 9.81783543226229e+00, 9.80981900581211e+00, 9.79973147164818e+00, 9.78728286270930e+00, 9.77218896265455e+00, 9.75417958936984e+00, 9.73300707532956e+00, 9.70845464153067e+00, 9.68034434370956e+00, 9.64854426613187e+00, 9.61297465173350e+00, 9.57361268912202e+00, 9.53049572695227e+00, 9.48372275293442e+00, 9.43345405494150e+00, 9.37990907030273e+00, 9.32336251976233e+00, 9.26413900704636e+00, 9.20260633564142e+00, 9.13916784444190e+00, 9.07425408894898e+00, 9.00831419372882e+00, 8.94180717769071e+00, 8.87519351249726e+00, 8.80892712396673e+00, 8.74344799455174e+00, 8.67917547807246e+00, 8.61650239936970e+00, 8.55578998220158e+00, 8.49736362724586e+00, 8.44150954618823e+00, 8.38847224526526e+00, 8.33845284063632e+00, 8.29160817780218e+00, 8.24805071793961e+00, 8.20784914586449e+00, 8.17102964782367e+00, 8.13757780269436e+00, 8.10744102735904e+00, 8.08053151560830e+00, 8.05672960926028e+00, 8.03588753953710e+00, 8.01783347543650e+00, 8.00237581339950e+00, 7.98930763880087e+00, 7.97841128479154e+00, 7.96946290822249e+00, 7.96223699646239e+00, 7.95651071374679e+00, 7.95206799221247e+00, 7.94870327189057e+00, 7.94622479644082e+00, 7.94445737785875e+00, 7.94324455403627e+00, 7.94245007782022e+00, 7.94195869466970e+00, 7.94167618742368e+00, 7.94152869006894e+00, 7.94146129659862e+00, 7.94143601486624e+00, 7.94142913761119e+00, 7.94142812254236e+00, 7.94142804568441e+00, 7.94142668415339e+00, 7.94141745957725e+00, 7.94138354901181e+00, 7.94129315552796e+00, 7.94109532973896e+00, 7.94071646255748e+00, 7.94005758704107e+00, 7.93899260914857e+00, 7.93736756642312e+00, 7.93500098983521e+00, 7.93168541693053e+00, 7.92719007369457e+00, 7.92126470798310e+00, 7.91364451910491e+00, 7.90405608678512e+00, 7.89222415947935e+00, 7.87787911869582e+00, 7.86076489506807e+00, 7.84064707633032e+00, 7.81732092021980e+00, 7.79061896967811e+00, 7.76041796603946e+00, 7.72664476977038e+00, 7.68928102816479e+00, 7.64836637426999e+00, 7.60399999900788e+00, 7.55634050569074e+00, 7.50560402894468e+00, 7.45206067418727e+00, 7.39602940506346e+00, 7.33787157075926e+00, 7.27798331948940e+00, 7.21678718585554e+00, 7.15472316599488e+00, 7.09223960405361e+00, 7.02978420605301e+00, 6.96779547334558e+00, 6.90669480952828e+00, 6.84687950504755e+00, 6.78871674692314e+00, 6.73253874168763e+00, 6.67863898242635e+00, 6.62726963978921e+00, 6.57864001511312e+00, 6.53291596316283e+00, 6.49022017296936e+00, 6.45063318713339e+00, 6.41419504115149e+00, 6.38090741260958e+00, 6.35073618300811e+00, 6.32361433011020e+00, 6.29944508387827e+00, 6.27810529253078e+00, 6.25944895574491e+00, 6.24331088878779e+00, 6.22951048410210e+00, 6.21785553575867e+00, 6.20814608775707e+00, 6.20017826024745e+00, 6.19374799943136e+00, 6.18865468838279e+00, 6.18470454856076e+00, 6.18171375654519e+00, 6.17951119854031e+00, 6.17794078721413e+00, 6.17686327190531e+00, 6.17615748417441e+00, 6.17572097575692e+00, 6.17547002448088e+00, 6.17533900463721e+00, 6.17527914043746e+00, 6.17525668328495e+00, 6.17525057439266e+00, 6.17524967273287e+00, 6.17524960504498e+00, 6.17524841366146e+00, 6.17524034186302e+00, 6.17521066908162e+00, 6.17513157236770e+00, 6.17495847128068e+00, 6.17462696251544e+00, 6.17405046763741e+00, 6.17311870308945e+00, 6.17169706511467e+00, 6.16962700292742e+00, 6.16672743069629e+00, 6.16279720199122e+00, 6.15761863876898e+00, 6.15096207057783e+00, 6.14259129886487e+00, 6.13226985719546e+00, 6.11976789276202e+00, 6.10486945046444e+00, 6.08737990142472e+00, 6.06713322679188e+00, 6.04399884887753e+00, 6.01788769843225e+00, 5.98875722177219e+00, 5.95661506574956e+00, 5.92152123185081e+00, 5.88358856077669e+00, 5.84298149164177e+00, 5.79991312973992e+00, 5.75464074682040e+00, 5.70745992068704e+00, 5.65869758968877e+00, 5.60870434650127e+00, 5.55784632063014e+00, 5.50649699886464e+00, 5.45502930869952e+00, 5.40380824520122e+00, 5.35318426252343e+00, 5.30348758397900e+00, 5.25502351612908e+00, 5.20806878892614e+00, 5.16286889031935e+00, 5.11963632287863e+00, 5.07854968302836e+00, 5.03975344984529e+00, 5.00335836824347e+00, 4.96944231814318e+00, 4.93805157400821e+00, 4.90920237515363e+00, 4.88288274407499e+00, 4.85905450586260e+00, 4.83765547523935e+00, 4.81860178811691e+00, 4.80179036145406e+00, 4.78710146860489e+00, 4.77440141748700e+00, 4.76354531617244e+00, 4.75437990542471e+00, 4.74674643088687e+00, 4.74048351976062e+00, 4.73543001866216e+00, 4.73142774169405e+00, 4.72832407143916e+00, 4.72597435132601e+00, 4.72424400629746e+00, 4.72301033043649e+00, 4.72216388544121e+00, 4.72160946260696e+00, 4.72126657299186e+00, 4.72106944516414e+00, 4.72096652657150e+00, 4.72091950219333e+00, 4.72090186172188e+00, 4.72089706308556e+00, 4.72089635481688e+00, 4.72089630180862e+00, 4.72089537095682e+00, 4.72088906429967e+00, 4.72086588037217e+00, 4.72080408062023e+00, 4.72066883381513e+00, 4.72040982169090e+00, 4.71995940271120e+00, 4.71923142005191e+00, 4.71812072725498e+00, 4.71650349010109e+00, 4.71423830542827e+00, 4.71116815621728e+00, 4.70712319673795e+00, 4.70192433169808e+00, 4.69538751948414e+00, 4.68732869274640e+00, 4.67756915156920e+00, 4.66594124787341e+00, 4.65229414776512e+00, 4.63649943488141e+00, 4.61845630594617e+00, 4.59809611272559e+00, 4.57538602424654e+00, 4.55033161981135e+00, 4.52297827542419e+00, 4.49341127021114e+00, 4.46175461006621e+00, 4.42816863677646e+00, 4.39284655565739e+00, 4.35601006727913e+00, 4.31790432469601e+00, 4.27879245436652e+00, 4.23894987676517e+00, 4.19865864394325e+00, 4.15820198016896e+00, 4.11785917341463e+00, 4.07790092506666e+00, 4.03858522726081e+00, 4.00015380481032e+00, 3.96282913335403e+00, 3.92681202720327e+00, 3.89227977841951e+00, 3.85938482133160e+00, 3.82825389234471e+00, 3.79898765214316e+00, 3.77166073539422e+00, 3.74632219148184e+00, 3.72299627868821e+00, 3.70168357382303e+00, 3.68236235977867e+00, 3.66499025489175e+00, 3.64950605011892e+00, 3.63583172247611e+00, 3.62387459542013e+00, 3.61352961833376e+00, 3.60468173758263e+00, 3.59720833050586e+00, 3.59098167118213e+00, 3.58587139312262e+00, 3.58174690964310e+00, 3.57847974817804e+00, 3.57594575092700e+00, 3.57402709167628e+00, 3.57261405804843e+00, 3.57160655028017e+00, 3.57091525219149e+00, 3.57046243732097e+00, 3.57018238305469e+00, 3.57002137752896e+00, 3.56993731751611e+00, 3.56989890966076e+00, 3.56988450152539e+00, 3.56988058216107e+00, 3.56988000367103e+00, 3.56987996035945e+00, 3.56987919956988e+00, 3.56987404510845e+00, 3.56985509676040e+00, 3.56980458740175e+00, 3.56969404880598e+00, 3.56948235378922e+00, 3.56911421318526e+00, 3.56851919690060e+00, 3.56761133756790e+00, 3.56628936281730e+00, 3.56443758753024e+00, 3.56192748023137e+00, 3.55861989771635e+00, 3.55436795905068e+00, 3.54902050450854e+00, 3.54242605760988e+00, 3.53443718041860e+00, 3.52491508542396e+00, 3.51373434374779e+00, 3.50078751138089e+00, 3.48598948482760e+00, 3.46928139669680e+00, 3.45063387148623e+00, 3.43004948222351e+00, 3.40756427893087e+00, 3.38324829837004e+00, 3.35720500887000e+00, 3.32956969162799e+00, 3.30050680813035e+00, 3.27020644993688e+00, 3.23888000993773e+00, 3.20675525132837e+00, 3.17407097978991e+00, 3.14107154323852e+00, 3.10800138942753e+00, 3.07509990249058e+00, 3.04259671432595e+00, 3.01070764673305e+00, 2.97963138907156e+00, 2.94954695969715e+00, 2.92061194440053e+00, 2.89296145805224e+00, 2.86670774142884e+00, 2.84194028599560e+00, 2.81872637479304e+00, 2.79711193492837e+00, 2.77712261269296e+00, 2.75876500194981e+00, 2.74202797659795e+00, 2.72688409600030e+00, 2.71329106668288e+00, 2.70119325373210e+00, 2.69052324118392e+00, 2.68120344279668e+00, 2.67314776361946e+00, 2.66626330945700e+00, 2.66045213640089e+00, 2.65561302668864e+00, 2.65164327084455e+00, 2.64844042987323e+00, 2.64590404570869e+00, 2.64393726362984e+00, 2.64244832735903e+00, 2.64135190642788e+00, 2.64057021639533e+00, 2.64003389578165e+00, 2.63968260913124e+00, 2.63946535325189e+00, 2.63934045305020e+00, 2.63927524399160e+00, 2.63924544945758e+00, 2.63923427250304e+00, 2.63923123210076e+00, 2.63923078334376e+00, 2.63923074965618e+00, 2.63923015672639e+00, 2.63922613954061e+00, 2.63921137195215e+00, 2.63917200707081e+00, 2.63908585843515e+00, 2.63892087519382e+00, 2.63863397327575e+00, 2.63817027777773e+00, 2.63746282389526e+00, 2.63643275433248e+00, 2.63499003979446e+00, 2.63303473540494e+00, 2.63045876930672e+00, 2.62714824011933e+00, 2.62298617754248e+00, 2.61785569588495e+00, 2.61164344491268e+00, 2.60424323796495e+00, 2.59555971608631e+00, 2.58551189155907e+00, 2.57403640727920e+00, 2.56109035208219e+00, 2.54665348775089e+00, 2.53072977115230e+00, 2.51334809336163e+00, 2.49456220375063e+00, 2.47444983645761e+00, 2.45311110416128e+00, 2.43066626432345e+00, 2.40725299159287e+00, 2.38302330417759e+00, 2.35814029132683e+00, 2.33277477564123e+00, 2.30710202168230e+00, 2.28129857615315e+00, 2.25553929946611e+00, 2.22999462731002e+00, 2.20482808564921e+00, 2.18019407341605e+00, 2.15623592264881e+00, 2.13308424397479e+00, 2.11085556418734e+00, 2.08965126080123e+00, 2.06955679523867e+00, 2.05064124175621e+00, 2.03295710392000e+00, 2.01654040512968e+00, 2.00141103509009e+00, 1.98757333072639e+00, 1.97501686802662e+00, 1.96371744058954e+00, 1.95363820094177e+00, 1.94473094153297e+00, 1.93693749325048e+00, 1.93019121990295e+00, 1.92441858710212e+00, 1.91954078317260e+00, 1.91547536814925e+00, 1.91213792474370e+00, 1.90944368266397e+00, 1.90730908523287e+00, 1.90565326529410e+00, 1.90439939633590e+00, 1.90347588496651e+00, 1.90281737261294e+00, 1.90236551773174e+00, 1.90206953491667e+00, 1.90188647392052e+00, 1.90178122949248e+00, 1.90172628168146e+00, 1.90170117540497e+00, 1.90169175714674e+00, 1.90168919514958e+00, 1.90168881700408e+00, 1.90168878849630e+00, 1.90168828512505e+00, 1.90168487471019e+00, 1.90167233765600e+00, 1.90163891840809e+00, 1.90156578087452e+00, 1.90142571249555e+00, 1.90118212849021e+00, 1.90078842146097e+00, 1.90018769418894e+00, 1.89931290496241e+00, 1.89808744495165e+00, 1.89642615576567e+00, 1.89423678226256e+00, 1.89142184093042e+00, 1.88788086792400e+00, 1.88351299360879e+00, 1.87821977300143e+00, 1.87190818485264e+00, 1.86449369755618e+00, 1.85590328894429e+00, 1.84607830064386e+00, 1.83497700708098e+00, 1.82257678509729e+00, 1.80887578262225e+00, 1.79389400350859e+00, 1.77767374952791e+00, 1.76027938826213e+00, 1.74179644556181e+00, 1.72233005160498e+00, 1.70200279862108e+00, 1.68095209438349e+00, 1.65932711712354e+00, 1.63728549329152e+00, 1.61498982859331e+00, 1.59260422436039e+00, 1.57029090546682e+00, 1.54820707317762e+00, 1.52650207759614e+00, 1.50531498139994e+00, 1.48477256130927e+00, 1.46498776833505e+00, 1.44605864429381e+00, 1.42806767198362e+00, 1.41108152090119e+00, 1.39515113997584e+00, 1.38031214346771e+00, 1.36658543542312e+00, 1.35397802105943e+00, 1.34248395914118e+00, 1.33208541673068e+00, 1.32275379563575e+00, 1.31445090755821e+00, 1.30713018167680e+00, 1.30073789368942e+00, 1.29521440890492e+00, 1.29049543371824e+00, 1.28651326979685e+00, 1.28319806377081e+00, 1.28047904249081e+00, 1.27828572042791e+00, 1.27654906202110e+00, 1.27520257824128e+00, 1.27418333382105e+00, 1.27343283992592e+00, 1.27289780684506e+00, 1.27253073275704e+00, 1.27229030782847e+00, 1.27214161772394e+00, 1.27205613679113e+00, 1.27201150836338e+00, 1.27199111733987e+00, 1.27198346797489e+00, 1.27198138716328e+00, 1.27198108003988e+00, 1.27198105676035e+00, 1.27198064403463e+00, 1.27197784775815e+00, 1.27196756836491e+00, 1.27194016739096e+00, 1.27188020160996e+00, 1.27176536218955e+00, 1.27156566243551e+00, 1.27124291324224e+00, 1.27075052015233e+00, 1.27003362968384e+00, 1.26902964496621e+00, 1.26766912137482e+00, 1.26587704147956e+00, 1.26357445509465e+00, 1.26068045465344e+00, 1.25711443896705e+00, 1.25279860045051e+00, 1.24766055325228e+00, 1.24163600383657e+00, 1.23467135305837e+00, 1.22672611127588e+00, 1.21777500703126e+00, 1.20780967637948e+00, 1.19683983455734e+00, 1.18489385412475e+00, 1.17201870292295e+00, 1.15827922932177e+00, 1.14375681873209e+00, 1.12854748125823e+00, 1.11275946254770e+00, 1.09651049550008e+00, 1.07992482726473e+00, 1.06313016254379e+00, 1.04625466036357e+00, 1.02942410804291e+00, 1.01275937490259e+00, 9.96374221855767e-01, 9.80373514254437e-01, 9.64851857041797e-01, 9.49892645780444e-01, 9.35567506255799e-01, 9.21936080105491e-01, 9.09046104566627e-01, 8.96933730598416e-01, 8.85624024500448e-01, 8.75131602614971e-01, 8.65461355594741e-01, 8.56609226917673e-01, 8.48563018868862e-01, 8.41303207329954e-01, 8.34803753868163e-01, 8.29032909455291e-01, 8.23954008489757e-01, 8.19526254593427e-01, 8.15705500960613e-01, 8.12445027974421e-01, 8.09696319557183e-01, 8.07409837515635e-01, 8.05535790244127e-01, 8.04024888859489e-01, 8.02829080481320e-01, 8.01902245272888e-01, 8.01200841341913e-01, 8.00684479955930e-01, 8.00316412982799e-01, 8.00063915168847e-01, 7.99898545859188e-01, 7.99796277978794e-01, 7.99737486352496e-01, 7.99706792477794e-01, 7.99692768341934e-01, 7.99687507428514e-01, 7.99686076334774e-01, 7.99685865108422e-01, 7.99685848996162e-01, 7.99685561999596e-01, 7.99683617557044e-01, 7.99676469601620e-01, 7.99657415938805e-01, 7.99615718285649e-01, 7.99535865440779e-01, 7.99397010528057e-01, 7.99172610529443e-01, 7.98830293077522e-01, 7.98331971447020e-01, 7.97634223519983e-01, 7.96688943831533e-01, 7.95444269320781e-01, 7.93845768948653e-01, 7.91837874970673e-01, 7.89365519748811e-01, 7.86375927290907e-01, 7.82820494314025e-01, 7.78656682933239e-01, 7.73849837642057e-01, 7.68374834623096e-01, 7.62217472904400e-01, 7.55375525267800e-01, 7.47859382230581e-01, 7.39692244141871e-01, 7.30909842899734e-01, 7.21559703726351e-01, 7.11699986103744e-01, 7.01397968541304e-01, 6.90728261795480e-01, 6.79770847640222e-01, 6.68609044379760e-01, 6.57327496131215e-01, 6.46010271583201e-01, 6.34739141305349e-01, 6.23592083005200e-01, 6.12642043720044e-01, 6.01955968835346e-01, 5.91594091578658e-01, 5.81609464171075e-01, 5.72047703440070e-01, 5.62946919207629e-01, 5.54337792606074e-01, 5.46243772884965e-01, 5.38681364453678e-01, 5.31660480113336e-01, 5.25184841053365e-01, 5.19252408756318e-01, 5.13855838148993e-01, 5.08982944958646e-01, 5.04617183171524e-01, 5.00738130701258e-01, 4.97321982850499e-01, 4.94342053910897e-01, 4.91769287332540e-01, 4.89572774361934e-01, 4.87720279975074e-01, 4.86178773421127e-01, 4.84914958871400e-01, 4.83895799691885e-01, 4.83089027901411e-01, 4.82463628629214e-01, 4.81990288032781e-01, 4.81641792348906e-01, 4.81393365663887e-01, 4.81222934688102e-01, 4.81111310328222e-01, 4.81042278120928e-01, 4.81002592510960e-01, 4.80981873351841e-01, 4.80972406666725e-01, 4.80968855397608e-01, 4.80967889367369e-01, 4.80967746783367e-01, 4.80967735830369e-01, 4.80967539725961e-01, 4.80966211090527e-01, 4.80961326897759e-01, 4.80948307522526e-01, 4.80919815392134e-01, 4.80865251292887e-01, 4.80770369307463e-01, 4.80617029398420e-01, 4.80383104330430e-01, 4.80042554773053e-01, 4.79565682829348e-01, 4.78919569662910e-01, 4.78068697153198e-01, 4.76975746499775e-01, 4.75602558478508e-01, 4.73911230893426e-01, 4.71865319170339e-01, 4.69431096742099e-01, 4.66578823810025e-01, 4.63283967258111e-01, 4.59528311932852e-01, 4.55300904967016e-01, 4.50598780740208e-01, 4.45427424360808e-01, 4.39800945599334e-01, 4.33741951860608e-01, 4.27281126522644e-01, 4.20456536087094e-01, 4.13312704449341e-01, 4.05899503890623e-01, 3.98270919304171e-01, 3.90483744455707e-01, 3.82596267067024e-01, 3.74666993957590e-01, 3.66753459403023e-01, 3.58911150362564e-01, 3.51192572293693e-01, 3.43646469718594e-01, 3.36317207102244e-01, 3.29244308282725e-01, 3.22462146808947e-01, 3.15999775104432e-01, 3.09880877319634e-01, 3.04123828945036e-01, 2.98741845593415e-01, 2.93743203666416e-01, 2.89131516728776e-01, 2.84906053140453e-01, 2.81062082648910e-01, 2.77591242021575e-01, 2.74481912206371e-01, 2.71719601762209e-01, 2.69287333243611e-01, 2.67166030726994e-01, 2.65334907646480e-01, 2.63771854523862e-01, 2.62453826038536e-01, 2.61357226243048e-01, 2.60458289685188e-01, 2.59733454881140e-01, 2.59159725154758e-01, 2.58715010487645e-01, 2.58378442884441e-01, 2.58130657001823e-01, 2.57954027541434e-01, 2.57832855244555e-01, 2.57753494276139e-01, 2.57704415316883e-01, 2.57676200710582e-01, 2.57661470410772e-01, 2.57654740076186e-01, 2.57652215305203e-01, 2.57651528507214e-01, 2.57651427137304e-01, 2.57651419290854e-01, 2.57651278033050e-01, 2.57650320991317e-01, 2.57646802814669e-01, 2.57637424724237e-01, 2.57616901392449e-01, 2.57577598250403e-01, 2.57509254430742e-01, 2.57398805199616e-01, 2.57230316123957e-01, 2.56985039234047e-01, 2.56641598912124e-01, 2.56176311994389e-01, 2.55563642472136e-01, 2.54776786145067e-01, 2.53788374652711e-01, 2.52571281668986e-01, 2.51099507034402e-01, 2.49349107723617e-01, 2.47299138437326e-01, 2.44932559971206e-01, 2.42237071042480e-01, 2.39205819521858e-01, 2.35837952386754e-01, 2.32138970243049e-01, 2.28120861677921e-01, 2.23802004387908e-01, 2.19206833076612e-01, 2.14365287459620e-01, 2.09312066226667e-01, 2.04085723453993e-01, 1.98727651896579e-01, 1.93281002267071e-01, 1.87789588798445e-01, 1.82296829187543e-01, 1.76844761815222e-01, 1.71473175553734e-01, 1.66218878272598e-01, 1.61115120178460e-01, 1.56191178194545e-01, 1.51472098435550e-01, 1.46978586053173e-01, 1.42727025717548e-01, 1.38729611961846e-01, 1.34994566554314e-01, 1.31526419810212e-01, 1.28326334026569e-01, 1.25392449653127e-01, 1.22720238014555e-01, 1.20302847997500e-01, 1.18131437778667e-01, 1.16195486123779e-01, 1.14483080825117e-01, 1.12981184327422e-01, 1.11675878438051e-01, 1.10552591200225e-01, 1.09596309543283e-01, 1.08791781262685e-01, 1.08123709305618e-01, 1.07576940350381e-01, 1.07136648394100e-01, 1.06788512641327e-01, 1.06518887561384e-01, 1.06314961697160e-01, 1.06164900792709e-01, 1.06057970168225e-01, 1.05984631081608e-01, 1.05936606107567e-01, 1.05906909324922e-01, 1.05889838272970e-01, 1.05880926123833e-01, 1.05876854198529e-01, 1.05875326696440e-01, 1.05874911180507e-01, 1.05874849851286e-01, 1.05874845065667e-01, 1.05874758414071e-01, 1.05874171337749e-01, 1.05872013197567e-01, 1.05866260499493e-01, 1.05853671425150e-01, 1.05829564005626e-01, 1.05787647772472e-01, 1.05719918178092e-01, 1.05616622399030e-01, 1.05466304183707e-01, 1.05255934007387e-01, 1.04971128768117e-01, 1.04596462438822e-01, 1.04115865386447e-01, 1.03513105485323e-01, 1.02772338829352e-01, 1.01878712082605e-01, 1.00818992751829e-01, 9.95821984989576e-02, 9.81601926970477e-02, 9.65482114272742e-02, 9.47452875801869e-02, 9.27545410251334e-02, 9.05833100266271e-02, 8.82431079519659e-02, 8.57494002140928e-02, 8.31212084025079e-02, 8.03805605369199e-02, 7.75518171274499e-02, 7.46609111288219e-02, 7.17345450711377e-02, 6.87993901475359e-02, 6.58813297883629e-02, 6.30047846359858e-02, 6.01921476021730e-02, 5.74633478347764e-02, 5.48355520373427e-02, 5.23230017220317e-02, 4.99369765029538e-02, 4.76858670520155e-02, 4.55753371268131e-02, 4.36085521264129e-02, 4.17864516870612e-02, 4.01080454969645e-02, 3.85707143202829e-02, 3.71705017056119e-02, 3.59023855879528e-02, 3.47605226257315e-02, 3.37384613925682e-02, 3.28293233132051e-02, 3.20259524276594e-02, 3.13210366860588e-02, 3.07072045572028e-02, 3.01771013365464e-02, 2.97234497314118e-02, 2.93390991513789e-02, 2.90170677059240e-02, 2.87505802711738e-02, 2.85331051922171e-02, 2.83583912938354e-02, 2.82205059396485e-02, 2.81138739670581e-02, 2.80333164945397e-02, 2.79740879067063e-02, 2.79319088227403e-02, 2.79029925834546e-02, 2.78840627712498e-02, 2.78723595044781e-02, 2.78656326976616e-02, 2.78621211071144e-02, 2.78605167257417e-02, 2.78599148820007e-02, 2.78597511673357e-02, 2.78597270034435e-02, 2.78597251018553e-02, 2.78596904648261e-02, 2.78594557946246e-02, 2.78585931338959e-02, 2.78562936873428e-02, 2.78512618626170e-02, 2.78416270704884e-02, 2.78248775747407e-02, 2.77978206694236e-02, 2.77565734031207e-02, 2.76965878660036e-02, 2.76127145778702e-02, 2.74993066275471e-02, 2.73503657935112e-02, 2.71597298445343e-02, 2.69212975720325e-02, 2.66292849394993e-02, 2.62785022631729e-02, 2.58646388971862e-02, 2.53845389197842e-02, 2.48364492915538e-02, 2.42202213514524e-02, 2.35374476993915e-02, 2.27915196629596e-02, 2.19875955832780e-02, 2.11324767135324e-02, 2.02343949667618e-02, 1.93027242411414e-02, 1.83476336764468e-02, 1.73797061056291e-02, 1.64095475277559e-02, 1.54474133402545e-02, 1.45028744162214e-02, 1.35845413452138e-02, 1.26998589871743e-02, 1.18549767523479e-02, 1.10546935267255e-02, 1.03024705938886e-02, 9.60050174627369e-03, 8.94982728433114e-03, 8.35047777327880e-03, 7.80163403337930e-03, 7.30179148772336e-03, 6.84891923983066e-03, 6.44060672001423e-03, 6.07419315469393e-03, 5.74687730428458e-03, 5.45580676845070e-03, 5.19814759523562e-03, 4.97113592274090e-03, 4.77211396160371e-03, 4.59855287685507e-03, 4.44806515098151e-03, 4.31840889092427e-03, 4.20748633549550e-03, 4.11333855849527e-03, 4.03413806761499e-03, 3.96818068860405e-03, 3.91387781339425e-03, 3.86974978947308e-03, 3.83442094075548e-03, 3.80661644008207e-03, 3.78516100404092e-03, 3.76897915818653e-03, 3.75709663428658e-03, 3.74864232081862e-03, 3.74285010287245e-03, 3.73905990371526e-03, 3.73671727860092e-03, 3.73537100789001e-03, 3.73466828003899e-03, 3.73434723139177e-03, 3.73422680037905e-03, 3.73419404073694e-03, 3.73418920550073e-03, 3.73418882162646e-03, 3.73418178653750e-03, 3.73413412305723e-03, 3.73395891125838e-03, 3.73349189318617e-03, 3.73246999921228e-03, 3.73051357697385e-03, 3.72711330356879e-03, 3.72162282036470e-03, 3.71325818102974e-03, 3.70110521118328e-03, 3.68413579658622e-03, 3.66123389658135e-03, 3.63123167566037e-03, 3.59295553003669e-03, 3.54528095747142e-03, 3.48719421709909e-03, 3.41785763839653e-03, 3.33667439919161e-03, 3.24334777137203e-03, 3.13792941410540e-03, 3.02085144136477e-03, 2.89293780925348e-03, 2.75539206934639e-03, 2.60976060838125e-03, 2.45787290948485e-03, 2.30176279542782e-03, 2.14357667799803e-03, 1.98547620520628e-03, 1.82954315517492e-03, 1.67769393065598e-03, 1.53160968944910e-03, 1.39268624413802e-03, 1.26200565624579e-03, 1.14032920276230e-03, 1.02810937010389e-03, 9.25517002460907e-04, 8.32478929424951e-04, 7.48721398002417e-04, 6.73815268012135e-04, 6.07219837664985e-04, 5.48323007320661e-04, 4.96476147395342e-04, 4.51022629547992e-04, 4.11319659068710e-04, 3.76753782008521e-04, 3.46751019074887e-04, 3.20782798434394e-04, 2.98368724443776e-04, 2.79076933111355e-04, 2.62522571810919e-04, 2.48364879793073e-04, 2.36303363848833e-04, 2.26073549034922e-04, 2.17442692836300e-04, 2.10205715340805e-04, 2.04181476318510e-04, 1.99209455004734e-04, 1.95146855805461e-04, 1.91866150537123e-04, 1.89253057426342e-04, 1.87204942044161e-04, 1.85629607900140e-04, 1.84444427855368e-04, 1.83575754185483e-04, 1.82958536251832e-04, 1.82536069742056e-04, 1.82259801295714e-04, 1.82089116893991e-04, 1.81991051420027e-04, 1.81939870119052e-04, 1.81916489005118e-04, 1.81907718598495e-04, 1.81905332896193e-04, 1.81904980772529e-04, 1.81904952563983e-04, 1.81904432370317e-04, 1.81900908023280e-04, 1.81887952653170e-04, 1.81853422326128e-04, 1.81777873448752e-04, 1.81633264556780e-04, 1.81382026941179e-04, 1.80976599750693e-04, 1.80359535733032e-04, 1.79464289280062e-04, 1.78216793150554e-04, 1.76537907545793e-04, 1.74346779168388e-04, 1.71565074010616e-04, 1.68121946321875e-04, 1.63959482661014e-04, 1.59038229314845e-04, 1.53342297476655e-04, 1.46883474322124e-04, 1.39703782032541e-04, 1.31876038874879e-04, 1.23502178078994e-04, 1.14709328329354e-04, 1.05643889199314e-04, 9.64639960887581e-05, 8.73308684194996e-05, 7.83996470197224e-05, 6.98105413974304e-05, 6.16814006626498e-05, 5.41029393730241e-05, 4.71374167076988e-05, 4.08205014306162e-05, 3.51648754381984e-05, 3.01636384314463e-05, 2.57923120123849e-05, 2.20101702517184e-05, 1.87633991384217e-05, 1.59916153101847e-05, 1.36358118165703e-05, 1.16442460154076e-05, 9.97401057745151e-06, 8.58699391770301e-06, 7.44245407879873e-06, 6.49580715794684e-06, 5.70609227068530e-06, 5.04112871446694e-06, 4.47736385922326e-06, 3.99891571944977e-06, 3.59825721396589e-06, 3.27822855972215e-06, 3.02847068667234e-06, 2.82910133542202e-06, 2.66892871358948e-06, 2.53893178303599e-06, 2.43278411637959e-06, 2.34589045788880e-06, 2.27481864634171e-06, 2.21700158853366e-06, 2.17044322781135e-06, 2.13346995714403e-06, 2.10459324812007e-06, 2.08246526511857e-06, 2.06587495021893e-06, 2.05374986883555e-06, 2.04515343682488e-06, 2.03927862862940e-06, 2.03544089494301e-06, 2.03307146762199e-06, 2.03171067703067e-06, 2.03100062456940e-06, 2.03067629023224e-06, 2.03055463643409e-06, 2.03052154468253e-06, 2.03051665970122e-06, 2.03051624758643e-06, 2.03050847069619e-06, 2.03045576263605e-06, 2.03026203848822e-06, 2.02974587182325e-06, 2.02861742693017e-06, 2.02646080182944e-06, 2.02272423344141e-06, 2.01672122557284e-06, 2.00764554314503e-06, 1.99460214744818e-06, 1.97665426959759e-06, 1.95288473747723e-06, 1.92246477404012e-06, 1.88470526296669e-06, 1.83903008377229e-06, 1.78480805068659e-06, 1.72108832252604e-06, 1.64643726030533e-06, 1.55925360267872e-06, 1.46390603461555e-06, 1.40840973372747e-06, 1.35898810071922e-06, 1.28408230715138e-06, 1.18302652829527e-06, 1.11984971518098e-06, 1.12491792592095e-06, 1.12153795939207e-06, 1.09154512095699e-06, 1.02845774918592e-06, 9.83985888325378e-07, 9.89126169415044e-07, 9.90596161543067e-07, 1.06925444498529e-06, 1.12107725769306e-06, 1.09611545262567e-06, 9.68135658752090e-07, 7.40912922125858e-07, 6.61645703559088e-07, 6.22200825972641e-07, 5.53278436399296e-07, 4.44228671625524e-07, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 8.88178419700125e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 8.88178419700125e-15, 7.10542735760100e-15, 8.88178419700125e-15, 7.10542735760100e-15, 5.32907051820075e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 8.88178419700125e-15, 5.32907051820075e-15, 7.10542735760100e-15, 5.32907051820075e-15, 7.10542735760100e-15],
[ 1.60000000000000e+01, 1.59999999913149e+01, 1.59999997227538e+01, 1.59999979032100e+01, 1.59999912143338e+01, 1.59999733839334e+01, 1.59999343609530e+01, 1.59998596212026e+01, 1.59997296291378e+01, 1.59995194763478e+01, 1.59991987121359e+01, 1.59987313758526e+01, 1.59980762346447e+01, 1.59971872242203e+01, 1.59960140843038e+01, 1.59945031748484e+01, 1.59925984539733e+01, 1.59902425941187e+01, 1.59873782091901e+01, 1.59839491625330e+01, 1.59799019234659e+01, 1.59751869387907e+01, 1.59697599851289e+01, 1.59635834680534e+01, 1.59566276347292e+01, 1.59488716680962e+01, 1.59403046325025e+01, 1.59309262431396e+01, 1.59207474346757e+01, 1.59097907081856e+01, 1.58980902398865e+01, 1.58856917403341e+01, 1.58726520586000e+01, 1.58590385324436e+01, 1.58449280924667e+01, 1.58304061354504e+01, 1.58155651892385e+01, 1.58005033983087e+01, 1.57853228652240e+01, 1.57701278881596e+01, 1.57550231383831e+01, 1.57401118237420e+01, 1.57254938847736e+01, 1.57112642690041e+01, 1.56975113264158e+01, 1.56843153651146e+01, 1.56717474011219e+01, 1.56598681302132e+01, 1.56487271430863e+01, 1.56383623981229e+01, 1.56287999588536e+01, 1.56200539961532e+01, 1.56121270483649e+01, 1.56050105261241e+01, 1.55986854427594e+01, 1.55931233458757e+01, 1.55882874211805e+01, 1.55841337358530e+01, 1.55806125858573e+01, 1.55776699096133e+01, 1.55752487294031e+01, 1.55732905818410e+01, 1.55717368996756e+01, 1.55705303091079e+01, 1.55696158096669e+01, 1.55689418074054e+01, 1.55684609766850e+01, 1.55681309309703e+01, 1.55679146887325e+01, 1.55677809265898e+01, 1.55677040180359e+01, 1.55676638623433e+01, 1.55676455143057e+01, 1.55676386312311e+01, 1.55676367588593e+01, 1.55676364825008e+01, 1.55676364462484e+01, 1.55676356016107e+01, 1.55676298790583e+01, 1.55676088422327e+01, 1.55675527648766e+01, 1.55674300363729e+01, 1.55671949790502e+01, 1.55667861574562e+01, 1.55661252455721e+01, 1.55651165017962e+01, 1.55636468839532e+01, 1.55615868181560e+01, 1.55587916166437e+01, 1.55551035213833e+01, 1.55503543328162e+01, 1.55443685671870e+01, 1.55369670717625e+01, 1.55279710151272e+01, 1.55172061596096e+01, 1.55045073145294e+01, 1.54897228620370e+01, 1.54727192415061e+01, 1.54533852735429e+01, 1.54316362008254e+01, 1.54074173206828e+01, 1.53807070845573e+01, 1.53515195435503e+01, 1.53199060285872e+01, 1.52859559696066e+01, 1.52497967813166e+01, 1.52115927733777e+01, 1.51715430792434e+01, 1.51298786382061e+01, 1.50868583065022e+01, 1.50427642122652e+01, 1.49978965023376e+01, 1.49525676536631e+01, 1.49070965363098e+01, 1.48618024183946e+01, 1.48169990957811e+01, 1.47729893129074e+01, 1.47300596177240e+01, 1.46884757660949e+01, 1.46484787617398e+01, 1.46102815891705e+01, 1.45740666708403e+01, 1.45399840570367e+01, 1.45081503383979e+01, 1.44786482563207e+01, 1.44515269755190e+01, 1.44268029749228e+01, 1.44044615071964e+01, 1.43844585726406e+01, 1.43667233494543e+01, 1.43511610187826e+01, 1.43376559193427e+01, 1.43260749626054e+01, 1.43162712355760e+01, 1.43080877144099e+01, 1.43013610087726e+01, 1.42959250544409e+01, 1.42916146706212e+01, 1.42882688992795e+01, 1.42857340468153e+01, 1.42838663539480e+01, 1.42825342278433e+01, 1.42816199812497e+01, 1.42810210365252e+01, 1.42806505675354e+01, 1.42804375689627e+01, 1.42803263599601e+01, 1.42802755466088e+01, 1.42802564845983e+01, 1.42802512992534e+01, 1.42802505339068e+01, 1.42802504663412e+01, 1.42802491424276e+01, 1.42802401727119e+01, 1.42802071989960e+01, 1.42801193024682e+01, 1.42799269394766e+01, 1.42795585275775e+01, 1.42789178131453e+01, 1.42778821327177e+01, 1.42763016583502e+01, 1.42739996926959e+01, 1.42707740524066e+01, 1.42663995488685e+01, 1.42606315434582e+01, 1.42532105208030e+01, 1.42438675885689e+01, 1.42323307769510e+01, 1.42183319764388e+01, 1.42016143199699e+01, 1.41819397868703e+01, 1.41590967827840e+01, 1.41329074340007e+01, 1.41032343281128e+01, 1.40699864375532e+01, 1.40331239798059e+01, 1.39926619988370e+01, 1.39486724966166e+01, 1.39012850002932e+01, 1.38506855170532e+01, 1.37971139009015e+01, 1.37408597283236e+01, 1.36822568471993e+01, 1.36216768197344e+01, 1.35595215208822e+01, 1.34962151757969e+01, 1.34321961225691e+01, 1.33679085713704e+01, 1.33037946016469e+01, 1.32402865998295e+01, 1.31778002962893e+01, 1.31167285166556e+01, 1.30574357228358e+01, 1.30002533854766e+01, 1.29454762031861e+01, 1.28933591644095e+01, 1.28441154344508e+01, 1.27979150413522e+01, 1.27548843287199e+01, 1.27151061397814e+01, 1.26786206939575e+01, 1.26454271142791e+01, 1.26154855606133e+01, 1.25887199196240e+01, 1.25650209975532e+01, 1.25442501562622e+01, 1.25262433265857e+01, 1.25108153260906e+01, 1.24977644010708e+01, 1.24868769054157e+01, 1.24779320223672e+01, 1.24707064297020e+01, 1.24649788051829e+01, 1.24605340678574e+01, 1.24571672525246e+01, 1.24546869199066e+01, 1.24529180140241e+01, 1.24517040910294e+01, 1.24509088600943e+01, 1.24504169963984e+01, 1.24501342081160e+01, 1.24499865626508e+01, 1.24499191011915e+01, 1.24498937939029e+01, 1.24498869096911e+01, 1.24498858935953e+01, 1.24498858120919e+01, 1.24498843079396e+01, 1.24498741170818e+01, 1.24498366543346e+01, 1.24497367918233e+01, 1.24495182422081e+01, 1.24490996822379e+01, 1.24483717675356e+01, 1.24471951687885e+01, 1.24453997356477e+01, 1.24427848675577e+01, 1.24391211402219e+01, 1.24341532022715e+01, 1.24276039188573e+01, 1.24191796975892e+01, 1.24085768881836e+01, 1.23954891016513e+01, 1.23796152497976e+01, 1.23606680637900e+01, 1.23383828146383e+01, 1.23125259319322e+01, 1.22829032032698e+01, 1.22493672380890e+01, 1.22118238976646e+01, 1.21702374280240e+01, 1.21246340829761e+01, 1.20751040872247e+01, 1.20218018601732e+01, 1.19649444942438e+01, 1.19048085520013e+01, 1.18417253094781e+01, 1.17760746255773e+01, 1.17082776576866e+01, 1.16387886716560e+01, 1.15680862111644e+01, 1.14966638987180e+01, 1.14250211393161e+01, 1.13536539886994e+01, 1.12830464309053e+01, 1.12136622841686e+01, 1.11459379200547e+01, 1.10802759391853e+01, 1.10170399004807e+01, 1.09565501531359e+01, 1.08990807757152e+01, 1.08448575885791e+01, 1.07940571769832e+01, 1.07468068436435e+01, 1.07031854007448e+01, 1.06632247104308e+01, 1.06269118871959e+01, 1.05941920826240e+01, 1.05649717802336e+01, 1.05391225340397e+01, 1.05164850877310e+01, 1.04968738116292e+01, 1.04800813919091e+01, 1.04658837013555e+01, 1.04540447739631e+01, 1.04443217978510e+01, 1.04364700333012e+01, 1.04302475562913e+01, 1.04254197236993e+01, 1.04217632553402e+01, 1.04190698308815e+01, 1.04171491069683e+01, 1.04158310717609e+01, 1.04149676703884e+01, 1.04144336550575e+01, 1.04141266369159e+01, 1.04139663421931e+01, 1.04138931013920e+01, 1.04138656261065e+01, 1.04138581521502e+01, 1.04138570490094e+01, 1.04138569639246e+01, 1.04138554360740e+01, 1.04138450846581e+01, 1.04138070317049e+01, 1.04137055960919e+01, 1.04134836046495e+01, 1.04130584564925e+01, 1.04123190952845e+01, 1.04111240238958e+01, 1.04093004725622e+01, 1.04066448049393e+01, 1.04029242159911e+01, 1.03978797410583e+01, 1.03912305564669e+01, 1.03826795087778e+01, 1.03719197628706e+01, 1.03586424097813e+01, 1.03425448256210e+01, 1.03233395257274e+01, 1.03007632168151e+01, 1.02745857180352e+01, 1.02446184033224e+01, 1.02107218156374e+01, 1.01728121213059e+01, 1.01308661109208e+01, 1.00849245117509e+01, 1.00350934527576e+01, 9.98154401241480e+00, 9.92450987478937e+00, 9.86428321257559e+00, 9.80120899840207e+00, 9.73567801012709e+00, 9.66811883677114e+00, 9.59898920741276e+00, 9.52876695777410e+00, 9.45794092349097e+00, 9.38700201211133e+00, 9.31643466461890e+00, 9.24670887714715e+00, 9.17827291760081e+00, 9.11154684088291e+00, 9.04691687945591e+00, 8.98473076153330e+00, 8.92529398603108e+00, 8.86886706099113e+00, 8.81566369083591e+00, 8.76584987842572e+00, 8.71954389154435e+00, 8.67681703100314e+00, 8.63769512941620e+00, 8.60216070567411e+00, 8.57015569951386e+00, 8.54158471226512e+00, 8.51631868259234e+00, 8.49419892863484e+00, 8.47504148936901e+00, 8.45864169760929e+00, 8.44477891456118e+00, 8.43322135138620e+00, 8.42373089736264e+00, 8.41606786776024e+00, 8.40999557852951e+00, 8.40528465046345e+00, 8.40171694370813e+00, 8.39908902530736e+00, 8.39721507853359e+00, 8.39592917340764e+00, 8.39508683298999e+00, 8.39456584929209e+00, 8.39426632521463e+00, 8.39410994369824e+00, 8.39403849101717e+00, 8.39401168654053e+00, 8.39400439506268e+00, 8.39400331885544e+00, 8.39400323736828e+00, 8.39400179383119e+00, 8.39399201365219e+00, 8.39395606063696e+00, 8.39386022277100e+00, 8.39365048228354e+00, 8.39324879779062e+00, 8.39255024516306e+00, 8.39142114492294e+00, 8.38969828221811e+00, 8.38718930049545e+00, 8.38367432148187e+00, 8.37890881157510e+00, 8.37262767799607e+00, 8.36455053716165e+00, 8.35438805331843e+00, 8.34184919879229e+00, 8.32664924021462e+00, 8.30851821040074e+00, 8.28720958631204e+00, 8.26250886311745e+00, 8.23424169607808e+00, 8.20228127862089e+00, 8.16655463848084e+00, 8.12704756492818e+00, 8.08380792825053e+00, 8.03694721583638e+00, 7.98664018417998e+00, 7.93312260869199e+00, 7.87668719850809e+00, 7.81767782638684e+00, 7.75648229915623e+00, 7.69352395722237e+00, 7.62925243827658e+00, 7.56413396741657e+00, 7.49864154167713e+00, 7.43324536133618e+00, 7.36840382502035e+00, 7.30455535401853e+00, 7.24211124818657e+00, 7.18144970712539e+00, 7.12291108185251e+00, 7.06679435933918e+00, 7.01335482929824e+00, 6.96280284221307e+00, 6.91530354088726e+00, 6.87097743434049e+00, 6.82990168103476e+00, 6.79211195573257e+00, 6.75760478792294e+00, 6.72634027684821e+00, 6.69824510614458e+00, 6.67321579784829e+00, 6.65112215944064e+00, 6.63181088769229e+00, 6.61510929882963e+00, 6.60082915595888e+00, 6.58877056211547e+00, 6.57872588144901e+00, 6.57048364285197e+00, 6.56383237091814e+00, 6.55856427970511e+00, 6.55447875662639e+00, 6.55138555809346e+00, 6.54910763628230e+00, 6.54748351836934e+00, 6.54636916619184e+00, 6.54563925557455e+00, 6.54518783016680e+00, 6.54492830381522e+00, 6.54479280721590e+00, 6.54473089759668e+00, 6.54470767314434e+00, 6.54470135552861e+00, 6.54470042306171e+00, 6.54470035306112e+00, 6.54469912097147e+00, 6.54469077338337e+00, 6.54466008677523e+00, 6.54457828761178e+00, 6.54439927248447e+00, 6.54405643840764e+00, 6.54346025116636e+00, 6.54249666509021e+00, 6.54102649071871e+00, 6.53888578698550e+00, 6.53588733103242e+00, 6.53182319079505e+00, 6.52646839247308e+00, 6.51958563671230e+00, 6.51093097419545e+00, 6.50026030457848e+00, 6.48733651441526e+00, 6.47193702286521e+00, 6.45386146228314e+00, 6.43293918841079e+00, 6.40903629602066e+00, 6.38206181423808e+00, 6.35197277410097e+00, 6.31877788039493e+00, 6.28253957967928e+00, 6.24337439380737e+00, 6.20145147816343e+00, 6.15698945955635e+00, 6.11025170238281e+00, 6.06154023519953e+00, 6.01118863587029e+00, 5.95955421632624e+00, 5.90700986452995e+00, 5.85393589121906e+00, 5.80071219510300e+00, 5.74771100755426e+00, 5.69529041329818e+00, 5.64378877461144e+00, 5.59352012011768e+00, 5.54477050113591e+00, 5.49779527253100e+00, 5.45281722289187e+00, 5.41002546045848e+00, 5.36957495486315e+00, 5.33158663783059e+00, 5.29614797550557e+00, 5.26331393815771e+00, 5.23310830719553e+00, 5.20552527284691e+00, 5.18053128729543e+00, 5.15806714680966e+00, 5.13805028219858e+00, 5.12037723979000e+00, 5.10492633526091e+00, 5.09156046035416e+00, 5.08013001816681e+00, 5.07047595672406e+00, 5.06243286345247e+00, 5.05583207550842e+00, 5.05050475334026e+00, 5.04628485806634e+00, 5.04301196795591e+00, 5.04053386619961e+00, 5.03870883185712e+00, 5.03740756882029e+00, 5.03651471406787e+00, 5.03592987639285e+00, 5.03556816985622e+00, 5.03536022190224e+00, 5.03525165357144e+00, 5.03520204763120e+00, 5.03518343870240e+00, 5.03517837661992e+00, 5.03517762946705e+00, 5.03517757354861e+00, 5.03517659159271e+00, 5.03516993869679e+00, 5.03514548194514e+00, 5.03508028920529e+00, 5.03493761639089e+00, 5.03466438050777e+00, 5.03418921983629e+00, 5.03342123068795e+00, 5.03224945631410e+00, 5.03054318734371e+00, 5.02815311403808e+00, 5.02491334808847e+00, 5.02064430520191e+00, 5.01515640919178e+00, 5.00825454409463e+00, 4.99974314402135e+00, 4.98943177282693e+00, 4.97714100980745e+00, 4.96270842667050e+00, 4.94599441845939e+00, 4.92688764031852e+00, 4.90530980571918e+00, 4.88121962162297e+00, 4.85461567203965e+00, 4.82553811168395e+00, 4.79406909226100e+00, 4.76033191014058e+00, 4.72448892980693e+00, 4.68673839650043e+00, 4.64731029883876e+00, 4.60646147455256e+00, 4.56447016860328e+00, 4.52163025389307e+00, 4.47824531338508e+00, 4.43462276268043e+00, 4.39106816809201e+00, 4.34787989050576e+00, 4.30534416207458e+00, 4.26373068185242e+00, 4.22328879743045e+00, 4.18424432137333e+00, 4.14679701263562e+00, 4.11111873358370e+00, 4.07735227304146e+00, 4.04561080605749e+00, 4.01597794352294e+00, 3.98850831105605e+00, 3.96322858796997e+00, 3.94013893413833e+00, 3.91921473477761e+00, 3.90040859944812e+00, 3.88365256033861e+00, 3.86886042441942e+00, 3.85593024276243e+00, 3.84474686704170e+00, 3.83518456722965e+00, 3.82710968555356e+00, 3.82038330007678e+00, 3.81486386735837e+00, 3.81040980831177e+00, 3.80688199555453e+00, 3.80414609518679e+00, 3.80207471198092e+00, 3.80054928520040e+00, 3.79946168328708e+00, 3.79871544980269e+00, 3.79822666033898e+00, 3.79792436038925e+00, 3.79775056690615e+00, 3.79765983074684e+00, 3.79761837257806e+00, 3.79760282017663e+00, 3.79759858954531e+00, 3.79759796511295e+00, 3.79759791836167e+00, 3.79759709715201e+00, 3.79759153333607e+00, 3.79757108016554e+00, 3.79751655956392e+00, 3.79739724277212e+00, 3.79716873765218e+00, 3.79677136787156e+00, 3.79612911950246e+00, 3.79514921930386e+00, 3.79372239575828e+00, 3.79172385833829e+00, 3.78901501198959e+00, 3.78544590206606e+00, 3.78085835982410e+00, 3.77508979035580e+00, 3.76797751427524e+00, 3.75936354287008e+00, 3.74909963560410e+00, 3.73705246104032e+00, 3.72310865993515e+00, 3.70717959494604e+00, 3.68920556738704e+00, 3.66915928957304e+00, 3.64704842263850e+00, 3.62291702457113e+00, 3.59684580086084e+00, 3.56895110884923e+00, 3.53938273361692e+00, 3.50832052384326e+00, 3.47597004503123e+00, 3.44255746830247e+00, 3.40832395872210e+00, 3.37351985155938e+00, 3.33839890381813e+00, 3.30321288093775e+00, 3.26820668807374e+00, 3.23361418906297e+00, 3.19965478394722e+00, 3.16653074819400e+00, 3.13442528221622e+00, 3.10350118372673e+00, 3.07390003905116e+00, 3.04574183030020e+00, 3.01912486842920e+00, 2.99412598201879e+00, 2.97080091294836e+00, 2.94918488922912e+00, 2.92929335997932e+00, 2.91112288718639e+00, 2.89465219386762e+00, 2.87984336941749e+00, 2.86664323137386e+00, 2.85498483949128e+00, 2.84478915359593e+00, 2.83596682169344e+00, 2.82842007950143e+00, 2.82204473717482e+00, 2.81673222364210e+00, 2.81237165386267e+00, 2.80885187969135e+00, 2.80606348119956e+00, 2.80390065260037e+00, 2.80226293571079e+00, 2.80105675448095e+00, 2.80019670676668e+00, 2.79960657434033e+00, 2.79922001909360e+00, 2.79898094228999e+00, 2.79884349421606e+00, 2.79877173316476e+00, 2.79873894477191e+00, 2.79872664468532e+00, 2.79872329876085e+00, 2.79872280490931e+00, 2.79872276783654e+00, 2.79872211532457e+00, 2.79871769445951e+00, 2.79870144288469e+00, 2.79865812209760e+00, 2.79856331533341e+00, 2.79838174795982e+00, 2.79806599739524e+00, 2.79755565144926e+00, 2.79677696455745e+00, 2.79564305041210e+00, 2.79405463664612e+00, 2.79190139215853e+00, 2.78906382011263e+00, 2.78541568953721e+00, 2.78082695609882e+00, 2.77516709871824e+00, 2.76830877451358e+00, 2.76013167180201e+00, 2.75052642170504e+00, 2.73939841556568e+00, 2.72667137002950e+00, 2.71229048585341e+00, 2.69622506095873e+00, 2.67847044242284e+00, 2.65904923421974e+00, 2.63801171468652e+00, 2.61543545635029e+00, 2.59142417728105e+00, 2.56610588451887e+00, 2.53963039445735e+00, 2.51216633179431e+00, 2.48389771844863e+00, 2.45502026812860e+00, 2.42573750264810e+00, 2.39625680389369e+00, 2.36678551115712e+00, 2.33752716730411e+00, 2.30867800847983e+00, 2.28042378025812e+00, 2.25293694817375e+00, 2.22637435283652e+00, 2.20087534029559e+00, 2.17656037839724e+00, 2.15353015110848e+00, 2.13186510657875e+00, 2.11162542213853e+00, 2.09285134104567e+00, 2.07556383162875e+00, 2.05976551911404e+00, 2.04544184309443e+00, 2.03256239833642e+00, 2.02108242240226e+00, 2.01094439942968e+00, 2.00207975455584e+00, 1.99441061730479e+00, 1.98785163442916e+00, 1.98231181310208e+00, 1.97769637412145e+00, 1.97390859223627e+00, 1.97085159731074e+00, 1.96843010638219e+00, 1.96655205336097e+00, 1.96513008076247e+00, 1.96408285697500e+00, 1.96333618354224e+00, 1.96282386000102e+00, 1.96248827899611e+00, 1.96228073153584e+00, 1.96216141101765e+00, 1.96209911456589e+00, 1.96207065071399e+00, 1.96205997292613e+00, 1.96205706830758e+00, 1.96205663959192e+00, 1.96205660727175e+00, 1.96205603658367e+00, 1.96205217008804e+00, 1.96203795644169e+00, 1.96200006808374e+00, 1.96191715036277e+00, 1.96175835309893e+00, 1.96148220361380e+00, 1.96103587342057e+00, 1.96035488282206e+00, 1.95936327839785e+00, 1.95797430747320e+00, 1.95609160084568e+00, 1.95361086007160e+00, 1.95042202843100e+00, 1.94641190550130e+00, 1.94146714460492e+00, 1.93547755113613e+00, 1.92833957914065e+00, 1.91995990501339e+00, 1.91025894245567e+00, 1.89917415355564e+00, 1.88666300851216e+00, 1.87270545223322e+00, 1.85730575039253e+00, 1.84049361044766e+00, 1.82232450380372e+00, 1.80287915221213e+00, 1.78226218242647e+00, 1.76059999538220e+00, 1.73803793669654e+00, 1.71473689101964e+00, 1.69086945088294e+00, 1.66661582891781e+00, 1.64215968923377e+00, 1.61768406897368e+00, 1.59336754534725e+00, 1.56938077857904e+00, 1.54588352981616e+00, 1.52302221823032e+00, 1.50092804651537e+00, 1.47971569162074e+00, 1.45948253016396e+00, 1.44030834702376e+00, 1.42225546177583e+00, 1.40536920075662e+00, 1.38967864186919e+00, 1.37519756359506e+00, 1.36192553765531e+00, 1.34984911494843e+00, 1.33894306546992e+00, 1.32917164377531e+00, 1.32048986131671e+00, 1.31284475504839e+00, 1.30617664768327e+00, 1.30042039872871e+00, 1.29550664695489e+00, 1.29136304442846e+00, 1.28791547997832e+00, 1.28508928635787e+00, 1.28281042091242e+00, 1.28100660479202e+00, 1.27960840122419e+00, 1.27855020961755e+00, 1.27777114978197e+00, 1.27721580970016e+00, 1.27683483130409e+00, 1.27658531166936e+00, 1.27643100183810e+00, 1.27634229184667e+00, 1.27629597806005e+00, 1.27627481709647e+00, 1.27626687891890e+00, 1.27626471954500e+00, 1.27626440082604e+00, 1.27626437666759e+00, 1.27626394835927e+00, 1.27626104650920e+00, 1.27625037902814e+00, 1.27622194364574e+00, 1.27615971454984e+00, 1.27604054238345e+00, 1.27583331349487e+00, 1.27549841010870e+00, 1.27498750808419e+00, 1.27424374279640e+00, 1.27320226595308e+00, 1.27179120636073e+00, 1.26993303537089e+00, 1.26754632276715e+00, 1.26454785129637e+00, 1.26085503841395e+00, 1.25638859306664e+00, 1.25107531490060e+00, 1.24485092496073e+00, 1.23766280278103e+00, 1.22947249680919e+00, 1.22025787516572e+00, 1.21001479306721e+00, 1.19875817229817e+00, 1.18652241634621e+00, 1.17336112059702e+00, 1.15934607768474e+00, 1.14456562030647e+00, 1.12912238373201e+00, 1.11313060413788e+00, 1.09671309360028e+00, 1.07999804592338e+00, 1.06311582857960e+00, 1.04619590540757e+00, 1.02936401413552e+00, 1.01273969502545e+00, 9.96434235249744e-01, 9.80549061362584e-01, 9.65174582377518e-01, 9.50389460783674e-01, 9.36260269730086e-01, 9.22841482106165e-01, 9.10175731102083e-01, 8.98294281227070e-01, 8.87217652527060e-01, 8.76956347572156e-01, 8.67511639413893e-01, 8.58876388034693e-01, 8.51035861944373e-01, 8.43968549875649e-01, 8.37646954572351e-01, 8.32038366216578e-01, 8.27105617016302e-01, 8.22807820891947e-01, 8.19101103151960e-01, 8.15939324680850e-01, 8.13274803666639e-01, 8.11059035490434e-01, 8.09243408345168e-01, 8.07779908727401e-01, 8.06621807464474e-01, 8.05724313718803e-01, 8.05045181766614e-01, 8.04545253564120e-01, 8.04188919420843e-01, 8.03944479646209e-01, 8.03784391871114e-01, 8.03685391808722e-01, 8.03628479336650e-01, 8.03598766690005e-01, 8.03585190917358e-01, 8.03580098205632e-01, 8.03578712867326e-01, 8.03578508394397e-01, 8.03578492797282e-01, 8.03578214976710e-01, 8.03576332703056e-01, 8.03569413290221e-01, 8.03550968862128e-01, 8.03510604626496e-01, 8.03433305904040e-01, 8.03298894096612e-01, 8.03081679268329e-01, 8.02750334008483e-01, 8.02268009456238e-01, 8.01592709419532e-01, 8.00677932017324e-01, 7.99473579876279e-01, 7.97927129435115e-01, 7.95985037416704e-01, 7.93594348416031e-01, 7.90704452601263e-01, 7.87268927897344e-01, 7.83247388195154e-01, 7.78607249757364e-01, 7.73325323693117e-01, 7.67389144477574e-01, 7.60797953776672e-01, 7.53563275310258e-01, 7.45709039261031e-01, 7.37271242044796e-01, 7.28297156617725e-01, 7.18844137023309e-01, 7.08978085689331e-01, 6.98771670610830e-01, 6.88302390323704e-01, 6.77650586825488e-01, 6.66897500741651e-01, 6.56123450410204e-01, 6.45406199165595e-01, 6.34819555251291e-01, 6.24432228729315e-01, 6.14306951386880e-01, 6.04499850325038e-01, 5.95060054405669e-01, 5.86029505210306e-01, 5.77442940350679e-01, 5.69328016283382e-01, 5.61705539485036e-01, 5.54589778182120e-01, 5.47988831110136e-01, 5.41905034427637e-01, 5.36335392495452e-01, 5.31272022443714e-01, 5.26702606094558e-01, 5.22610845775624e-01, 5.18976922799999e-01, 5.15777958892865e-01, 5.12988481633196e-01, 5.10580895089722e-01, 5.08525956324476e-01, 5.06793257393070e-01, 5.05351710991958e-01, 5.04170036115489e-01, 5.03217238139833e-01, 5.02463075816594e-01, 5.01878505916718e-01, 5.01436094895045e-01, 5.01110386111952e-01, 5.00878210982115e-01, 5.00718933005970e-01, 5.00614615000719e-01, 5.00550101942569e-01, 5.00513014552904e-01, 5.00493651944851e-01, 5.00484805089611e-01, 5.00481486341773e-01, 5.00480583563119e-01, 5.00480450314928e-01, 5.00480440079086e-01, 5.00480256814781e-01, 5.00479015173003e-01, 5.00474450778951e-01, 5.00462283876111e-01, 5.00435657393297e-01, 5.00384666357249e-01, 5.00295998349495e-01, 5.00152702929322e-01, 4.99934105832072e-01, 4.99615881136950e-01, 4.99170291262832e-01, 4.98566600362826e-01, 4.97771661252706e-01, 4.96750669324232e-01, 4.95468069014180e-01, 4.93888589580299e-01, 4.91978377669348e-01, 4.89706185174073e-01, 4.87044563080580e-01, 4.83971006413615e-01, 4.80468992966042e-01, 4.76528860020338e-01, 4.72148469143043e-01, 4.67333619280680e-01, 4.62098182186109e-01, 4.56463950525880e-01, 4.50460206348930e-01, 4.44123034248582e-01, 4.37494417910383e-01, 4.30621169506935e-01, 4.23553747767750e-01, 4.16345022284576e-01, 4.09049039035047e-01, 4.01719835989161e-01, 3.94410349048448e-01, 3.87171438590928e-01, 3.80051056617578e-01, 3.73093564791223e-01, 3.66339205149675e-01, 3.59823718327359e-01, 3.53578098869129e-01, 3.47628473629404e-01, 3.41996087167463e-01, 3.36697377262517e-01, 3.31744123942421e-01, 3.27143656505003e-01, 3.22899104679058e-01, 3.19009682106509e-01, 3.15470992535872e-01, 3.12275351330552e-01, 3.09412116970217e-01, 3.06868029037771e-01, 3.04627550641650e-01, 3.02673214253586e-01, 3.00985970504961e-01, 2.99545539574124e-01, 2.98330764438168e-01, 2.97319964514372e-01, 2.96491287167614e-01, 2.95823053324463e-01, 2.95294092145944e-01, 2.94884058512064e-01, 2.94573726104968e-01, 2.94345248271449e-01, 2.94182378702862e-01, 2.94070644354676e-01, 2.93997463959592e-01, 2.93952206936734e-01, 2.93926189389813e-01, 2.93912606102342e-01, 2.93906399835604e-01, 2.93904071658691e-01, 2.93903438338877e-01, 2.93903344862226e-01, 2.93903337626746e-01, 2.93903207368109e-01, 2.93902324847238e-01, 2.93899080615645e-01, 2.93890432750450e-01, 2.93871507439416e-01, 2.93835264470875e-01, 2.93772241606565e-01, 2.93670390540248e-01, 2.93515015632918e-01, 2.93288824753938e-01, 2.92972099199000e-01, 2.92542986645579e-01, 2.91977917314606e-01, 2.91252138867570e-01, 2.90340360117762e-01, 2.89217487542432e-01, 2.87859432172803e-01, 2.86243958183819e-01, 2.84351539001030e-01, 2.82166182647828e-01, 2.79676186015150e-01, 2.76874778273594e-01, 2.73760617076951e-01, 2.70338107544244e-01, 2.66617522938034e-01, 2.62614916839167e-01, 2.58351828552515e-01, 2.53854795417663e-01, 2.49154696585364e-01, 2.44285961725442e-01, 2.39285684361928e-01, 2.34192682703556e-01, 2.29046550896012e-01, 2.23886740817349e-01, 2.18751709363756e-01, 2.13678159286026e-01, 2.08700393762173e-01, 2.03849796737936e-01, 1.99154443262946e-01, 1.94638837103671e-01, 1.90323767191543e-01, 1.86226270180072e-01, 1.82359683605824e-01, 1.78733772808796e-01, 1.75354914699492e-01, 1.72226322429500e-01, 1.69348296767952e-01, 1.66718492241603e-01, 1.64332188610480e-01, 1.62182560798717e-01, 1.60260942792670e-01, 1.58557083110500e-01, 1.57059391139899e-01, 1.55755174874310e-01, 1.54630871331465e-01, 1.53672271218273e-01, 1.52864739247644e-01, 1.52193430974844e-01, 1.51643506184206e-01, 1.51200337820610e-01, 1.50849714337021e-01, 1.50578032237881e-01, 1.50372474655633e-01, 1.50221171110038e-01, 1.50113333252717e-01, 1.50039361449485e-01, 1.49990917521931e-01, 1.49960959840003e-01, 1.49943738176370e-01, 1.49934747214344e-01, 1.49930639237674e-01, 1.49929098204828e-01, 1.49928679007566e-01, 1.49928617134968e-01, 1.49928612306949e-01, 1.49928524887610e-01, 1.49927932609394e-01, 1.49925755341918e-01, 1.49919951622259e-01, 1.49907250709628e-01, 1.49882928421201e-01, 1.49840636384917e-01, 1.49772293634921e-01, 1.49668048616494e-01, 1.49516318553989e-01, 1.49303911633419e-01, 1.49016235418030e-01, 1.48637592228750e-01, 1.48151558812817e-01, 1.47541443507977e-01, 1.46790809415133e-01, 1.45884047072826e-01, 1.44806975167464e-01, 1.43547443400671e-01, 1.42095908320153e-01, 1.40445951244319e-01, 1.38594707829172e-01, 1.36543181622919e-01, 1.34296419173525e-01, 1.31863531659533e-01, 1.29257557084204e-01, 1.26495167045350e-01, 1.23596232050259e-01, 1.20583268327600e-01, 1.17480796225669e-01, 1.14314644918689e-01, 1.11111239908361e-01, 1.07896908679434e-01, 1.04697236134692e-01, 1.01536495636907e-01, 9.84371743307211e-02, 9.54196036856327e-02, 9.25016986507501e-02, 8.96988020867742e-02, 8.70236256980128e-02, 8.44862747646147e-02, 8.20943416092259e-02, 7.98530518109661e-02, 7.77654474982974e-02, 7.58325933498600e-02, 7.40537929139506e-02, 7.24268052226016e-02, 7.09480541546270e-02, 6.96128253863257e-02, 6.84154479228560e-02, 6.73494590504955e-02, 6.64077530461054e-02, 6.55827151025115e-02, 6.48663426699478e-02, 6.42503567881434e-02, 6.37263060229110e-02, 6.32856653716409e-02, 6.29199320164542e-02, 6.26207191395416e-02, 6.23798482324602e-02, 6.21894394961391e-02, 6.20419991106491e-02, 6.19305014246034e-02, 6.18484635403505e-02, 6.17900094084280e-02, 6.17499204328116e-02, 6.17236697442305e-02, 6.17074377160769e-02, 6.16981069446201e-02, 6.16932357406466e-02, 6.16910101157302e-02, 6.16901752180556e-02, 6.16899481066895e-02, 6.16899145855641e-02, 6.16899119476031e-02, 6.16898638977101e-02, 6.16895383532263e-02, 6.16883416265193e-02, 6.16851516635375e-02, 6.16781708680740e-02, 6.16648031910003e-02, 6.16415610107719e-02, 6.16040070720612e-02, 6.15467364147335e-02, 6.14634027296361e-02, 6.13467928102516e-02, 6.11889516135058e-02, 6.09813587964894e-02, 6.07151553952416e-02, 6.03814165612988e-02, 5.99714630625048e-02, 5.94772007818261e-02, 5.88914740102062e-02, 5.82084153097888e-02, 5.74237725538786e-02, 5.65351928484308e-02, 5.55424437459298e-02, 5.44475546619445e-02, 5.32548656736740e-02, 5.19709766557401e-02, 5.06045965007136e-02, 4.91662993129260e-02, 4.76682011983929e-02, 4.61235768645221e-02, 4.45464390838897e-02, 4.29511057835938e-02, 4.13517789936737e-02, 3.97621573135112e-02, 3.81950993693678e-02, 3.66623505311203e-02, 3.51743395698758e-02, 3.37400465588402e-02, 3.23669386167644e-02, 3.10609663910775e-02, 2.98266116450527e-02, 2.86669749710136e-02, 2.75838923828964e-02, 2.65780701326079e-02, 2.56492282868948e-02, 2.47962451491066e-02, 2.40172963166412e-02, 2.33099838830118e-02, 2.26714529131584e-02, 2.20984937548003e-02, 2.15876299364464e-02, 2.11351923191518e-02, 2.07373808229541e-02, 2.03903154727971e-02, 2.00900787392282e-02, 1.98327512150307e-02, 1.96144425909868e-02, 1.94313196888878e-02, 1.92796329988560e-02, 1.91557427765137e-02, 1.90561453136926e-02, 1.89774995359127e-02, 1.89166536328607e-02, 1.88706710252777e-02, 1.88368546432827e-02, 1.88127682619861e-02, 1.87962535305051e-02, 1.87854413498165e-02, 1.87787564024280e-02, 1.87749138992483e-02, 1.87729079628998e-02, 1.87719914789035e-02, 1.87716476813762e-02, 1.87715541607965e-02, 1.87715403573812e-02, 1.87715392615004e-02, 1.87715191780011e-02, 1.87713831097298e-02, 1.87708829143212e-02, 1.87695496227303e-02, 1.87666319786608e-02, 1.87610452300113e-02, 1.87513326080833e-02, 1.87356418931035e-02, 1.87117193995476e-02, 1.86769235689646e-02, 1.86282600656895e-02, 1.85624397603377e-02, 1.84759601946798e-02, 1.83652100097618e-02, 1.82265943867588e-02, 1.80566778511153e-02, 1.78523389404006e-02, 1.76109294136459e-02, 1.73304291123753e-02, 1.70095865254538e-02, 1.66480348051774e-02, 1.62463736227032e-02, 1.58062089281454e-02, 1.53301453457182e-02, 1.48217293725690e-02, 1.42853453957006e-02, 1.37260703343642e-02, 1.31494959984586e-02, 1.25615306661917e-02, 1.19681927080215e-02, 1.13754092243319e-02, 1.07888316130449e-02, 1.02136778282951e-02, 9.65460809017671e-03, 9.11563742839938e-03, 8.60008526942124e-03, 8.11055974890351e-03, 7.64897264947706e-03, 7.21657963915057e-03, 6.81403963237237e-03, 6.44148667747402e-03, 6.09860803139561e-03, 5.78472309051747e-03, 5.49885928841221e-03, 5.23982240138388e-03, 5.00625956714451e-03, 4.79671380787217e-03, 4.60966931332152e-03, 4.44358741443374e-03, 4.29693395526165e-03, 4.16819934389601e-03, 4.05591273366923e-03, 3.95865162150422e-03, 3.87504788754356e-03, 3.80379113207852e-03, 3.74363012624279e-03, 3.69337319351004e-03, 3.65188827988838e-03, 3.61810331618351e-03, 3.59100725648887e-03, 3.56965194725589e-03, 3.55315478101659e-03, 3.54070193221645e-03, 3.53155185766596e-03, 3.52503866520948e-03, 3.52057491018520e-03, 3.51765337181398e-03, 3.51584739325439e-03, 3.51480943850291e-03, 3.51426762064300e-03, 3.51402007969703e-03, 3.51392722179611e-03, 3.51390196250101e-03, 3.51389823429393e-03, 3.51389793557999e-03, 3.51389242700506e-03, 3.51385510576918e-03, 3.51371791160382e-03, 3.51335222267224e-03, 3.51255202458312e-03, 3.51101994486847e-03, 3.50835688213103e-03, 3.50405598323178e-03, 3.49750173922736e-03, 3.48797495827746e-03, 3.47466430241461e-03, 3.45668491392637e-03, 3.43310437915711e-03, 3.40297586804716e-03, 3.36537774998291e-03, 3.31945834917491e-03, 3.26448381680500e-03, 3.19988643484592e-03, 3.12531010823136e-03, 3.04064943778926e-03, 2.94607868958607e-03, 2.84206729161518e-03, 2.72937928834516e-03, 2.60905549537682e-03, 2.48237880367697e-03, 2.35082488141991e-03, 2.21600199644101e-03, 2.07958450918966e-03, 1.94324473656238e-03, 1.80858763891578e-03, 1.67709245543080e-03, 1.55006501391064e-03, 1.42860361620123e-03, 1.31357986305147e-03, 1.20563381882910e-03, 1.10518132982840e-03, 1.01243080076330e-03, 9.27407154931004e-04, 8.49980900097691e-04, 7.79899490089164e-04, 7.16817691390759e-04, 6.60324654440103e-04, 6.09967029504332e-04, 5.65268621080861e-04, 5.25747515311537e-04, 4.90931111544413e-04, 4.60368402221079e-04, 4.33638139499593e-04, 4.10351938123554e-04, 3.90153079905531e-04, 3.72713481107567e-04, 3.57731166811703e-04, 3.44929049610698e-04, 3.34054538321027e-04, 3.24879104068643e-04, 3.17197109684031e-04, 3.10823704930741e-04, 3.05592133550192e-04, 3.01351060389976e-04, 2.97962409787189e-04, 2.95299900246349e-04, 2.93248203679042e-04, 2.91702545911932e-04, 2.90568566461147e-04, 2.89762300297980e-04, 2.89210185947297e-04, 2.88849028368787e-04, 2.88625856230240e-04, 2.88497620166694e-04, 2.88430688486940e-04, 2.88400111216447e-04, 2.88388641331849e-04, 2.88385521310275e-04, 2.88385060807528e-04, 2.88385023583970e-04, 2.88384333115843e-04, 2.88379655112081e-04, 2.88362458768177e-04, 2.88316623368701e-04, 2.88216332696933e-04, 2.88024336885684e-04, 2.87690682887032e-04, 2.87152024350590e-04, 2.86331632706194e-04, 2.85140241370385e-04, 2.83477860868686e-04, 2.81236697267317e-04, 2.78305277580770e-04, 2.74573812504286e-04, 2.69940682890990e-04, 2.64319705397753e-04, 2.57647538123962e-04, 2.49890318785972e-04, 2.41048537871436e-04, 2.31159391223201e-04, 2.20296484451765e-04, 2.08567617812960e-04, 1.96111993940207e-04, 1.83097818183597e-04, 1.69719502835619e-04, 1.56191649997695e-04, 1.42737040190966e-04, 1.29569091180670e-04, 1.16873884415725e-04, 1.04799122564003e-04, 9.34541186712323e-05, 8.29171090028780e-05, 7.32399532079597e-05, 6.44428687728293e-05, 5.65040865918576e-05, 4.93610663010458e-05, 4.29357482421011e-05, 3.71728141725924e-05, 3.20578383661285e-05, 2.75891168488585e-05, 2.37162693839110e-05, 2.02988790913849e-05, 1.71297923454716e-05, 1.40220864484064e-05, 1.09248571504850e-05, 7.95255002827844e-06, 5.23478004588185e-06, 2.91605928914862e-06, 1.08700147372076e-06, 7.24331195911532e-07, 6.42960589658514e-07, -3.55271367880050e-15, 5.83234459838877e-07, 7.05936496814275e-07, 6.73477986268267e-07, 5.78873484258224e-07, 4.88026076439496e-07, 4.30843677179382e-07, 4.04414524979302e-07, 3.89517021304187e-07, 3.68927315363976e-07, 3.35433693265941e-07, 2.90236869204819e-07, -3.55271367880050e-15, -3.55271367880050e-15, 0.00000000000000e+00, -3.55271367880050e-15, -3.55271367880050e-15, -3.55271367880050e-15, -3.55271367880050e-15, -3.55271367880050e-15, -3.55271367880050e-15, -3.55271367880050e-15, -3.55271367880050e-15],
[ 1.70000000000000e+01, 1.69999999904899e+01, 1.69999996964177e+01, 1.69999977040323e+01, 1.69999903797683e+01, 1.69999708556265e+01, 1.69999281257782e+01, 1.69998462863337e+01, 1.69997039459683e+01, 1.69994738300112e+01, 1.69991225948277e+01, 1.69986108630453e+01, 1.69978934836139e+01, 1.69969200140485e+01, 1.69956354157238e+01, 1.69939809469653e+01, 1.69918952331186e+01, 1.69893154879081e+01, 1.69861788563505e+01, 1.69824238463115e+01, 1.69779918135093e+01, 1.69728284633416e+01, 1.69668853322903e+01, 1.69601212117615e+01, 1.69525034779804e+01, 1.69440092929318e+01, 1.69346266432997e+01, 1.69243551869303e+01, 1.69132068795856e+01, 1.69012063587304e+01, 1.68883910658846e+01, 1.68748110947113e+01, 1.68605287584820e+01, 1.68456178777765e+01, 1.68301627970472e+01, 1.68142571467440e+01, 1.67980023757087e+01, 1.67815060861254e+01, 1.67648802100620e+01, 1.67482390721837e+01, 1.67316973872680e+01, 1.67153682434856e+01, 1.66993611229305e+01, 1.66837800095879e+01, 1.66687216319429e+01, 1.66542738829325e+01, 1.66405144542014e+01, 1.66275097149193e+01, 1.66153138580540e+01, 1.66039683292650e+01, 1.65935015457416e+01, 1.65839289045891e+01, 1.65752530729564e+01, 1.65674645451522e+01, 1.65605424456372e+01, 1.65544555511175e+01, 1.65491635000724e+01, 1.65446181540189e+01, 1.65407650716936e+01, 1.65375450551967e+01, 1.65348957260239e+01, 1.65327530888540e+01, 1.65310530419696e+01, 1.65297327952568e+01, 1.65287321598208e+01, 1.65279946772923e+01, 1.65274685617941e+01, 1.65271074331506e+01, 1.65268708261134e+01, 1.65267244669584e+01, 1.65266403156160e+01, 1.65265963783154e+01, 1.65265763023796e+01, 1.65265687711027e+01, 1.65265667224033e+01, 1.65265664200192e+01, 1.65265663803528e+01, 1.65265654561727e+01, 1.65265591947086e+01, 1.65265361767773e+01, 1.65264748184498e+01, 1.65263405323096e+01, 1.65260833393521e+01, 1.65256360193914e+01, 1.65249128723372e+01, 1.65238091451134e+01, 1.65222011598318e+01, 1.65199471587251e+01, 1.65168888609163e+01, 1.65128537060323e+01, 1.65076577405907e+01, 1.65011090855544e+01, 1.64930119078494e+01, 1.64831708051849e+01, 1.64713955021523e+01, 1.64575057460802e+01, 1.64413362831170e+01, 1.64227417881413e+01, 1.64016016161954e+01, 1.63778242383920e+01, 1.63513512223280e+01, 1.63221606171246e+01, 1.62902696078626e+01, 1.62557363150626e+01, 1.62186606333989e+01, 1.61791840307884e+01, 1.61374882641185e+01, 1.60937930097251e+01, 1.60483524526860e+01, 1.60014509255506e+01, 1.59533977303443e+01, 1.59045213137731e+01, 1.58551629914311e+01, 1.58056704305598e+01, 1.57563911019724e+01, 1.57076659009792e+01, 1.56598231165108e+01, 1.56131728998838e+01, 1.55680023528617e+01, 1.55245713217771e+01, 1.54831089530204e+01, 1.54438110369905e+01, 1.54068381437447e+01, 1.53723145344668e+01, 1.53403278183034e+01, 1.53109293134840e+01, 1.52841350641039e+01, 1.52599274585790e+01, 1.52382573917164e+01, 1.52190469088514e+01, 1.52021922670511e+01, 1.51875673446744e+01, 1.51750273265121e+01, 1.51644125874262e+01, 1.51555526931467e+01, 1.51482704330791e+01, 1.51423857971220e+01, 1.51377198071184e+01, 1.51340981141682e+01, 1.51313542760507e+01, 1.51293326347442e+01, 1.51278907226428e+01, 1.51269011375258e+01, 1.51262528404014e+01, 1.51258518466165e+01, 1.51256212985267e+01, 1.51255009268650e+01, 1.51254459269667e+01, 1.51254252944294e+01, 1.51254196818624e+01, 1.51254188534586e+01, 1.51254187803263e+01, 1.51254173473351e+01, 1.51254076386024e+01, 1.51253719481799e+01, 1.51252768098944e+01, 1.51250685984149e+01, 1.51246698344657e+01, 1.51239763366257e+01, 1.51228553416592e+01, 1.51211446918237e+01, 1.51186531611896e+01, 1.51151619637454e+01, 1.51104274539978e+01, 1.51041849960890e+01, 1.50961539405200e+01, 1.50860436090307e+01, 1.50735601490161e+01, 1.50584140803504e+01, 1.50403283212620e+01, 1.50190464478479e+01, 1.49943409160352e+01, 1.49660209575221e+01, 1.49339398546936e+01, 1.48980013057652e+01, 1.48581646120993e+01, 1.48144484556413e+01, 1.47669330855200e+01, 1.47157607973772e+01, 1.46611346636439e+01, 1.46033155528975e+01, 1.45426175555645e+01, 1.44794020050206e+01, 1.44140703414469e+01, 1.43470561058702e+01, 1.42788163710113e+01, 1.42098229138120e+01, 1.41405534141986e+01, 1.40714829300886e+01, 1.40030758552977e+01, 1.39357785204000e+01, 1.38700125515652e+01, 1.38061690624507e+01, 1.37446037213133e+01, 1.36856327101092e+01, 1.36295295738420e+01, 1.35765229454811e+01, 1.35267951228424e+01, 1.34804814674571e+01, 1.34376705904593e+01, 1.33984052860857e+01, 1.33626841689819e+01, 1.33304639668502e+01, 1.33016624149006e+01, 1.32761616929838e+01, 1.32538123401384e+01, 1.32344375745846e+01, 1.32178379400156e+01, 1.32037961915807e+01, 1.31920823275435e+01, 1.31824586657329e+01, 1.31746848581665e+01, 1.31685227333207e+01, 1.31637408541427e+01, 1.31601186817022e+01, 1.31574502398855e+01, 1.31555471860653e+01, 1.31542412062986e+01, 1.31533856711071e+01, 1.31528565087609e+01, 1.31525522764720e+01, 1.31523934350167e+01, 1.31523208579517e+01, 1.31522936316123e+01, 1.31522862253711e+01, 1.31522851322248e+01, 1.31522850445410e+01, 1.31522834263290e+01, 1.31522724626992e+01, 1.31522321591564e+01, 1.31521247240948e+01, 1.31518896019747e+01, 1.31514393030476e+01, 1.31506561920702e+01, 1.31493903762000e+01, 1.31474588066479e+01, 1.31446456795837e+01, 1.31407041890137e+01, 1.31353596476492e+01, 1.31283139510815e+01, 1.31192513161339e+01, 1.31078451768285e+01, 1.30937660723770e+01, 1.30766903130854e+01, 1.30563091648140e+01, 1.30323382539571e+01, 1.30045268663228e+01, 1.29726667981754e+01, 1.29366004187632e+01, 1.28962276225030e+01, 1.28515113857033e+01, 1.28024816956460e+01, 1.27492376857109e+01, 1.26919478844199e+01, 1.26308485635334e+01, 1.25662402454900e+01, 1.24984824992189e+01, 1.24279872127129e+01, 1.23552105792176e+01, 1.22806440711392e+01, 1.22048047019853e+01, 1.21282248918722e+01, 1.20514422558018e+01, 1.19749896249111e+01, 1.18993855881137e+01, 1.18251258047850e+01, 1.17526752899291e+01, 1.16824618152627e+01, 1.16148705082865e+01, 1.15502396728481e+01, 1.14888578045491e+01, 1.14309617364744e+01, 1.13767358266147e+01, 1.13263120871899e+01, 1.12797711552844e+01, 1.12371440103161e+01, 1.11984143533858e+01, 1.11635215734997e+01, 1.11323642339041e+01, 1.11048040170340e+01, 1.10806700683350e+01, 1.10597636775153e+01, 1.10418632310408e+01, 1.10267293625991e+01, 1.10141102196669e+01, 1.10037467551894e+01, 1.09953779447402e+01, 1.09887458224274e+01, 1.09836002242600e+01, 1.09797031266068e+01, 1.09768324705038e+01, 1.09747853703944e+01, 1.09733806186027e+01, 1.09724604142650e+01, 1.09718912670679e+01, 1.09715640510801e+01, 1.09713932111128e+01, 1.09713151520684e+01, 1.09712858692881e+01, 1.09712779036488e+01, 1.09712767279366e+01, 1.09712766372544e+01, 1.09712750088924e+01, 1.09712639764968e+01, 1.09712234201888e+01, 1.09711153115404e+01, 1.09708787162695e+01, 1.09704255998280e+01, 1.09696376008924e+01, 1.09683639162627e+01, 1.09664204158518e+01, 1.09635900780196e+01, 1.09596248030356e+01, 1.09542486254777e+01, 1.09471623045991e+01, 1.09380492252198e+01, 1.09265824911994e+01, 1.09124330404467e+01, 1.08952785569906e+01, 1.08748129049657e+01, 1.08507557652481e+01, 1.08228621221161e+01, 1.07909312289272e+01, 1.07548146821082e+01, 1.07144232543715e+01, 1.06697321819343e+01, 1.06207846654271e+01, 1.05676934264766e+01, 1.05106402556720e+01, 1.04498735849468e+01, 1.03857042097152e+01, 1.03184993652488e+01, 1.02486754215650e+01, 1.01766894983221e+01, 1.01030303161455e+01, 1.00282085967197e+01, 9.95274730602113e+00, 9.87717200871423e+00, 9.80200157146636e+00, 9.72773942146279e+00, 9.65486553467195e+00, 9.58382929618707e+00, 9.51504334170654e+00, 9.44887845477687e+00, 9.38565955948075e+00, 9.32566281426385e+00, 9.26911378145656e+00, 9.21618662063762e+00, 9.16700423371868e+00, 9.12163927617232e+00, 9.08011594201625e+00, 9.04241242908995e+00, 9.00846399436027e+00, 8.97816651473402e+00, 8.95138047538242e+00, 8.92793531334363e+00, 8.90763404797879e+00, 8.89025813099007e+00, 8.87557244694016e+00, 8.86333039079356e+00, 8.85327894260614e+00, 8.84516365212172e+00, 8.83873343890846e+00, 8.83374510807868e+00, 8.82996747882893e+00, 8.82718502407191e+00, 8.82520092507819e+00, 8.82383945569133e+00, 8.82294762629180e+00, 8.82239603779848e+00, 8.82207891975901e+00, 8.82191335282836e+00, 8.82183770330579e+00, 8.82180932445712e+00, 8.82180160471266e+00, 8.82180046529421e+00, 8.82180037902088e+00, 8.82179885069741e+00, 8.82178849607858e+00, 8.82175043136836e+00, 8.82164896454928e+00, 8.82142690549142e+00, 8.82100163044304e+00, 8.82026205671143e+00, 8.81906666371011e+00, 8.81724267385843e+00, 8.81458648073557e+00, 8.81086538191046e+00, 8.80582063948093e+00, 8.79917185223129e+00, 8.79062257951221e+00, 8.77986710908534e+00, 8.76659821059263e+00, 8.75051566511325e+00, 8.73133531230152e+00, 8.70879831328114e+00, 8.68268029354419e+00, 8.65280000926012e+00, 8.61902717586836e+00, 8.58128911195467e+00, 8.53957588529966e+00, 8.49394370125934e+00, 8.44451634432816e+00, 8.39148456836524e+00, 8.33510342473611e+00, 8.27568761474068e+00, 8.21360504678145e+00, 8.14926886327571e+00, 8.08312827125664e+00, 8.01565855885599e+00, 7.94735070388451e+00, 7.87870097902553e+00, 7.81020093155251e+00, 7.74232806712551e+00, 7.67553750228242e+00, 7.61025477527475e+00, 7.54686992701514e+00, 7.48573288986230e+00, 7.42715015738563e+00, 7.37138265703259e+00, 7.31864471172120e+00, 7.26910395586540e+00, 7.22288206470899e+00, 7.18005616049890e+00, 7.14066077179906e+00, 7.10469023985984e+00, 7.07210148541430e+00, 7.04281706807499e+00, 7.01672848678428e+00, 6.99369968228819e+00, 6.97357071070753e+00, 6.95616156081396e+00, 6.94127608683105e+00, 6.92870602403916e+00, 6.91823504698027e+00, 6.90964282063348e+00, 6.90270898469159e+00, 6.89721700121282e+00, 6.89295778765005e+00, 6.88973305169965e+00, 6.88735824252656e+00, 6.88566503542601e+00, 6.88450327426547e+00, 6.88374230812401e+00, 6.88327167502528e+00, 6.88300110578516e+00, 6.88285984369427e+00, 6.88279529974952e+00, 6.88277108706672e+00, 6.88276450062796e+00, 6.88276352848332e+00, 6.88276345550412e+00, 6.88276217098738e+00, 6.88275346819794e+00, 6.88272147583047e+00, 6.88263619597952e+00, 6.88244956338925e+00, 6.88209214071565e+00, 6.88147058301655e+00, 6.88046598946865e+00, 6.87893324259294e+00, 6.87670141448103e+00, 6.87357529487828e+00, 6.86933806667667e+00, 6.86375511976316e+00, 6.85657895404287e+00, 6.84755507725313e+00, 6.83642875419851e+00, 6.82295241352834e+00, 6.80689346935684e+00, 6.78804227190625e+00, 6.76621986842317e+00, 6.74128523738522e+00, 6.71314165941685e+00, 6.68174191012541e+00, 6.64709200423834e+00, 6.60925328576314e+00, 6.56834274181059e+00, 6.52453151236990e+00, 6.47804166708265e+00, 6.42914141434700e+00, 6.37813898939357e+00, 6.32537552902181e+00, 6.27121727641523e+00, 6.21604746776809e+00, 6.16025823448431e+00, 6.10424281456099e+00, 6.04838831080462e+00, 5.99306916929903e+00, 5.93864148659682e+00, 5.88543819491511e+00, 5.83376512580156e+00, 5.78389791671746e+00, 5.73607970204925e+00, 5.69051951878615e+00, 5.64739135497643e+00, 5.60683377315636e+00, 5.56895004842703e+00, 5.53380876944401e+00, 5.50144485868053e+00, 5.47186097500037e+00, 5.44502926644284e+00, 5.42089344414126e+00, 5.39937114959506e+00, 5.38035658725343e+00, 5.36372339267422e+00, 5.34932770348199e+00, 5.33701139605378e+00, 5.32660544545335e+00, 5.31793335986996e+00, 5.31081463409292e+00, 5.30506815991363e+00, 5.30051552545790e+00, 5.29698413104524e+00, 5.29431004697647e+00, 5.29234053930827e+00, 5.29093619365841e+00, 5.28997257464707e+00, 5.28934136968636e+00, 5.28895098016497e+00, 5.28872654004215e+00, 5.28860936064555e+00, 5.28855582008583e+00, 5.28853573512078e+00, 5.28853027151833e+00, 5.28852946510187e+00, 5.28852940474802e+00, 5.28852834490409e+00, 5.28852116430447e+00, 5.28849476764452e+00, 5.28842440371658e+00, 5.28827041335183e+00, 5.28797550096861e+00, 5.28746263830754e+00, 5.28663369561258e+00, 5.28536888007257e+00, 5.28352704344677e+00, 5.28094689999599e+00, 5.27744917164116e+00, 5.27283964932233e+00, 5.26691312779407e+00, 5.25945813601623e+00, 5.25026234796158e+00, 5.23911852085324e+00, 5.22583077204430e+00, 5.21022097497781e+00, 5.19213503218944e+00, 5.17144877224880e+00, 5.14807322034165e+00, 5.12195901020250e+00, 5.09309973812541e+00, 5.06153410592661e+00, 5.02734675553847e+00, 4.99066775874508e+00, 4.95167078630341e+00, 4.91057003653748e+00, 4.86761605075342e+00, 4.82309057944584e+00, 4.77730068897361e+00, 4.73057231437264e+00, 4.68324347221770e+00, 4.63565734982323e+00, 4.58815548458131e+00, 4.54107123958748e+00, 4.49472376748998e+00, 4.44941263188651e+00, 4.40541322338755e+00, 4.36297306608507e+00, 4.32230906217997e+00, 4.28360567245056e+00, 4.24701398366155e+00, 4.21265157623312e+00, 4.18060308030945e+00, 4.15092129726948e+00, 4.12362876576525e+00, 4.09871966369282e+00, 4.07616195618942e+00, 4.05589972078556e+00, 4.03785560080490e+00, 4.02193335458596e+00, 4.00802047977496e+00, 3.99599089843926e+00, 3.98570769040808e+00, 3.97702585984665e+00, 3.96979511462434e+00, 3.96386263065155e+00, 3.95907576513086e+00, 3.95528467464160e+00, 3.95234478711334e+00, 3.95011907188674e+00, 3.94848004990492e+00, 3.94731148712364e+00, 3.94650971874159e+00, 3.94598455983175e+00, 3.94565976912146e+00, 3.94547304648010e+00, 3.94537556036555e+00, 3.94533101813908e+00, 3.94531430880964e+00, 3.94530976346666e+00, 3.94530909258354e+00, 3.94530904235449e+00, 3.94530816005607e+00, 3.94530218235518e+00, 3.94528020770581e+00, 3.94522163144628e+00, 3.94509343918011e+00, 3.94484793736007e+00, 3.94442101297249e+00, 3.94373100399346e+00, 3.94267825036071e+00, 3.94114538113859e+00, 3.93899837675857e+00, 3.93608842507013e+00, 3.93225456593765e+00, 3.92732709106345e+00, 3.92113163386387e+00, 3.91349384943249e+00, 3.90424454843630e+00, 3.89322511344333e+00, 3.88029299452708e+00, 3.86532705632013e+00, 3.84823253444408e+00, 3.82894535866582e+00, 3.80743561581547e+00, 3.78370995890049e+00, 3.75781281977188e+00, 3.72982634885386e+00, 3.69986908219624e+00, 3.66809341651686e+00, 3.63468204832976e+00, 3.59984359451903e+00, 3.56380765077048e+00, 3.52681955604260e+00, 3.48913511511803e+00, 3.45101549155301e+00, 3.41272242849606e+00, 3.37451389548608e+00, 3.33664020560223e+00, 3.29934060656788e+00, 3.26284032483117e+00, 3.22734803235970e+00, 3.19305370801174e+00, 3.16012687368551e+00, 3.12871519510074e+00, 3.09894344450113e+00, 3.07091282602629e+00, 3.04470066382305e+00, 3.02036044907497e+00, 2.99792223643442e+00, 2.97739337423690e+00, 2.95875954740630e+00, 2.94198610769045e+00, 2.92701966289743e+00, 2.91378989488688e+00, 2.90221157475326e+00, 2.89218674242107e+00, 2.88360701632939e+00, 2.87635599675005e+00, 2.87031172349947e+00, 2.86534914551717e+00, 2.86134255632667e+00, 2.85816794623509e+00, 2.85570521980678e+00, 2.85384022620723e+00, 2.85246655093520e+00, 2.85148702059700e+00, 2.85081487790944e+00, 2.85037459203279e+00, 2.85010227941069e+00, 2.84994572212370e+00, 2.84986398378497e+00, 2.84982663656016e+00, 2.84981262627778e+00, 2.84980881513665e+00, 2.84980825262004e+00, 2.84980821039267e+00, 2.84980746715547e+00, 2.84980243161289e+00, 2.84978392040744e+00, 2.84973457618988e+00, 2.84962658686510e+00, 2.84941977166251e+00, 2.84906010975703e+00, 2.84847877758922e+00, 2.84759174995708e+00, 2.84630000720834e+00, 2.84449037619584e+00, 2.84203701580935e+00, 2.83880353869304e+00, 2.83464573922791e+00, 2.82941487433276e+00, 2.82296141883144e+00, 2.81513919218567e+00, 2.80580972983800e+00, 2.79484675208125e+00, 2.78214056821397e+00, 2.76760224557363e+00, 2.75116737326628e+00, 2.73279925979390e+00, 2.71249142225797e+00, 2.69026925149191e+00, 2.66619077071831e+00, 2.64034644303494e+00, 2.61285802291320e+00, 2.58387648673987e+00, 2.55357911529668e+00, 2.52216583529546e+00, 2.48985495624304e+00, 2.45687846170969e+00, 2.42347702932462e+00, 2.38989496049242e+00, 2.35637519822929e+00, 2.32315459954453e+00, 2.29045960809278e+00, 2.25850244489193e+00, 2.22747790198208e+00, 2.19756078873331e+00, 2.16890404597924e+00, 2.14163751190814e+00, 2.11586729778171e+00, 2.09167571240986e+00, 2.06912166239352e+00, 2.04824145019039e+00, 2.02904989316951e+00, 2.01154169268727e+00, 1.99569299131719e+00, 1.98146306715387e+00, 1.96879612517302e+00, 1.95762315577919e+00, 1.94786383901789e+00, 1.93942847888722e+00, 1.93221995548636e+00, 1.92613568340028e+00, 1.92106956301433e+00, 1.91691390785876e+00, 1.91356132624591e+00, 1.91090653012409e+00, 1.90884803901744e+00, 1.90728974290414e+00, 1.90614228557051e+00, 1.90532422987035e+00, 1.90476296872506e+00, 1.90439535069252e+00, 1.90416799634825e+00, 1.90403729117507e+00, 1.90396905158528e+00, 1.90393787241932e+00, 1.90392617604846e+00, 1.90392299435300e+00, 1.90392252474145e+00, 1.90392248933820e+00, 1.90392186421130e+00, 1.90391762888569e+00, 1.90390205939763e+00, 1.90386055698865e+00, 1.90376973064068e+00, 1.90359578984449e+00, 1.90329331324863e+00, 1.90280445190565e+00, 1.90205861773105e+00, 1.90097269816739e+00, 1.89945182643541e+00, 1.89739072267161e+00, 1.89467560431622e+00, 1.89118664416992e+00, 1.88680093177069e+00, 1.88139586877675e+00, 1.87485290298349e+00, 1.86706148002886e+00, 1.85792306872644e+00, 1.84735509752766e+00, 1.83529462811640e+00, 1.82170158963097e+00, 1.80656140505453e+00, 1.78988686074169e+00, 1.77171910072339e+00, 1.75212766814126e+00, 1.73120956459139e+00, 1.70908735100642e+00, 1.68590636695098e+00, 1.66183119447935e+00, 1.63704153377338e+00, 1.61172768704914e+00, 1.58608586220396e+00, 1.56031350732829e+00, 1.53460487207675e+00, 1.50914696404415e+00, 1.48411603100299e+00, 1.45967465715729e+00, 1.43596951768025e+00, 1.41312979462134e+00, 1.39126622191919e+00, 1.37047069983026e+00, 1.35081640055538e+00, 1.33235827715427e+00, 1.31513388608799e+00, 1.29916443846442e+00, 1.28445600453967e+00, 1.27100080846746e+00, 1.25877856404960e+00, 1.24775781594792e+00, 1.23789726340634e+00, 1.22914705425629e+00, 1.22145004538124e+00, 1.21474303168526e+00, 1.20895794893033e+00, 1.20402305671852e+00, 1.19986410666121e+00, 1.19640549776000e+00, 1.19357141666562e+00, 1.19128695528736e+00, 1.18947919274694e+00, 1.18807822347260e+00, 1.18701810886282e+00, 1.18623772690689e+00, 1.18568149282461e+00, 1.18529992443481e+00, 1.18505002866703e+00, 1.18489549029594e+00, 1.18480665032934e+00, 1.18476026909366e+00, 1.18473907740560e+00, 1.18473112771734e+00, 1.18472896521368e+00, 1.18472864603283e+00, 1.18472862183936e+00, 1.18472819291036e+00, 1.18472528685565e+00, 1.18471460392740e+00, 1.18468612744138e+00, 1.18462380876122e+00, 1.18450446644800e+00, 1.18429694610796e+00, 1.18396158356198e+00, 1.18345000929124e+00, 1.18270532625831e+00, 1.18166268560559e+00, 1.18025027480984e+00, 1.17839072012595e+00, 1.17600288931453e+00, 1.17300406180766e+00, 1.16931241219189e+00, 1.16484973028625e+00, 1.15954427882132e+00, 1.15333366989255e+00, 1.14616762634913e+00, 1.13801048646727e+00, 1.12884331168159e+00, 1.11866546914892e+00, 1.10749558381623e+00, 1.09537178754944e+00, 1.08235123355953e+00, 1.06850888950835e+00, 1.05393566817008e+00, 1.03873599596360e+00, 1.02302495297013e+00, 1.00692514000922e+00, 9.90563437113849e-01, 9.74067813057115e-01, 9.57564328751481e-01, 9.41174450990516e-01, 9.25012760614646e-01, 9.09185104544157e-01, 8.93787207788179e-01, 8.78903732393685e-01, 8.64607747293721e-01, 8.50960557082498e-01, 8.38011828885840e-01, 8.25799953975107e-01, 8.14352583381133e-01, 8.03687283075234e-01, 7.93812262873519e-01, 7.84727142824170e-01, 7.76423730437973e-01, 7.68886790981377e-01, 7.62094800689290e-01, 7.56020678913728e-01, 7.50632499811225e-01, 7.45894187210908e-01, 7.41766197893526e-01, 7.38206198790792e-01, 7.35169742753325e-01, 7.32610945721181e-01, 7.30483165571304e-01, 7.28739679839691e-01, 7.27334356174353e-01, 7.26222306041326e-01, 7.25360509164606e-01, 7.24708393714256e-01, 7.24228355622909e-01, 7.23886199819546e-01, 7.23651486757738e-01, 7.23497769427713e-01, 7.23402709019325e-01, 7.23348061382630e-01, 7.23319531162510e-01, 7.23306495645364e-01, 7.23301605601325e-01, 7.23300275393470e-01, 7.23300079057676e-01, 7.23300064081265e-01, 7.23299797316745e-01, 7.23297989949440e-01, 7.23291345899536e-01, 7.23273635482304e-01, 7.23234877582986e-01, 7.23160655077436e-01, 7.23031592494308e-01, 7.22823022511619e-01, 7.22504864998651e-01, 7.22041738673166e-01, 7.21393320675599e-01, 7.20514963077772e-01, 7.19358567209284e-01, 7.17873706519242e-01, 7.16008976558996e-01, 7.13713536970221e-01, 7.10938795871918e-01, 7.07640172921597e-01, 7.03778865055689e-01, 6.99323530132812e-01, 6.94251800000639e-01, 6.88551537157075e-01, 6.82221758872050e-01, 6.75273169272739e-01, 6.67728262461701e-01, 6.59620986355815e-01, 6.50995985060735e-01, 6.41907464362578e-01, 6.32417747557671e-01, 6.22595605108682e-01, 6.12514450127538e-01, 6.02250492100382e-01, 5.91880934276055e-01, 5.81482287269402e-01, 5.71128854742035e-01, 5.60891428733267e-01, 5.50836214356728e-01, 5.41023987751963e-01, 5.31509478414812e-01, 5.22340957775594e-01, 5.13560010103440e-01, 5.05201459085480e-01, 4.97293423156751e-01, 4.89857474172446e-01, 4.82908876689677e-01, 4.76456888430778e-01, 4.70505106025691e-01, 4.65051843585446e-01, 4.60090534846831e-01, 4.55610152425681e-01, 4.51595640045507e-01, 4.48028355421762e-01, 4.44886522749247e-01, 4.42145694448453e-01, 4.39779221976071e-01, 4.37758735119285e-01, 4.36054628320107e-01, 4.34636551292868e-01, 4.33473899613730e-01, 4.32536299214590e-01, 4.31794076966163e-01, 4.31218707963069e-01, 4.30783228902001e-01, 4.30462606233529e-01, 4.30234047696235e-01, 4.30077246489390e-01, 4.29974548729270e-01, 4.29911036922697e-01, 4.29874524874464e-01, 4.29855462565627e-01, 4.29846752901121e-01, 4.29843485615276e-01, 4.29842596835243e-01, 4.29842465653209e-01, 4.29842455576086e-01, 4.29842275153490e-01, 4.29841052764559e-01, 4.29836559144011e-01, 4.29824580881807e-01, 4.29798367155602e-01, 4.29748166295951e-01, 4.29660871467078e-01, 4.29519792934308e-01, 4.29304572243449e-01, 4.28991251008391e-01, 4.28552503693741e-01, 4.27958039579995e-01, 4.27175173836790e-01, 4.26169561216987e-01, 4.24906078377450e-01, 4.23349832474063e-01, 4.21467264927340e-01, 4.19227310781718e-01, 4.16602566726937e-01, 4.13570415538814e-01, 4.10114052315414e-01, 4.06223359133183e-01, 4.01895579997259e-01, 3.97135757141601e-01, 3.91956902290641e-01, 3.86379891417512e-01, 3.80433087481673e-01, 3.74151711115704e-01, 3.67576992837414e-01, 3.60755150944506e-01, 3.53736246073066e-01, 3.46572966218194e-01, 3.39319395028753e-01, 3.32029811962009e-01, 3.24757566205523e-01, 3.17554058013741e-01, 3.10467852100690e-01, 3.03543938700411e-01, 2.96823149422075e-01, 2.90341727497925e-01, 2.84131045715640e-01, 2.78217460385516e-01, 2.72622286163301e-01, 2.67361874395252e-01, 2.62447776774586e-01, 2.57886976345649e-01, 2.53682169071094e-01, 2.49832081066625e-01, 2.46331808970201e-01, 2.43173173511124e-01, 2.40345078954419e-01, 2.37833873519332e-01, 2.35623707943841e-01, 2.33696890967650e-01, 2.32034241557859e-01, 2.30615438173537e-01, 2.29419365274207e-01, 2.28424456682156e-01, 2.27609034407241e-01, 2.26951640267220e-01, 2.26431356238987e-01, 2.26028108122662e-01, 2.25722945954196e-01, 2.25498293810958e-01, 2.25338161336683e-01, 2.25228309543070e-01, 2.25156364252122e-01, 2.25111871898793e-01, 2.25086294240111e-01, 2.25072940694975e-01, 2.25066839417902e-01, 2.25064550629234e-01, 2.25063928024262e-01, 2.25063836129110e-01, 2.25063829016044e-01, 2.25063700961211e-01, 2.25062833371577e-01, 2.25059644030740e-01, 2.25051142499254e-01, 2.25032537512583e-01, 2.24996908297186e-01, 2.24934953674190e-01, 2.24834831640432e-01, 2.24682100710943e-01, 2.24459772550958e-01, 2.24148483160270e-01, 2.23726786929138e-01, 2.23171574122631e-01, 2.22458607693930e-01, 2.21563169807123e-01, 2.20460802233593e-01, 2.19128118196071e-01, 2.17543656746983e-01, 2.15688744984348e-01, 2.13548329002027e-01, 2.11111732103948e-01, 2.08373299043817e-01, 2.05332888239575e-01, 2.01996180133133e-01, 1.98374778867937e-01, 1.94486095658817e-01, 1.90353014766913e-01, 1.86003355817508e-01, 1.81469158196457e-01, 1.76785823389778e-01, 1.71991158544564e-01, 1.67124368669558e-01, 1.62225045544648e-01, 1.57332198697201e-01, 1.52483368165633e-01, 1.47713850880599e-01, 1.43056063175575e-01, 1.38539052071156e-01, 1.34188158402115e-01, 1.30024826293429e-01, 1.26066546477848e-01, 1.22326915802944e-01, 1.18815792102790e-01, 1.15539522321523e-01, 1.12501222143109e-01, 1.09701087079120e-01, 1.07136717625547e-01, 1.04803444345958e-01, 1.02694642225625e-01, 1.00802027074625e-01, 9.91159299075548e-02, 9.76255479302850e-02, 9.63191729195927e-02, 9.51843993414627e-02, 9.42083155119207e-02, 9.33776814877412e-02, 9.26790972367719e-02, 9.20991640544244e-02, 9.16246412579866e-02, 9.12425990130892e-02, 9.09405668463314e-02, 9.07066761010498e-02, 9.05297934228990e-02, 9.03996414321142e-02, 9.03069021437410e-02, 9.02432985000132e-02, 9.02016496099556e-02, 9.01758959381880e-02, 9.01610916985653e-02, 9.01533630051183e-02, 9.01498318048688e-02, 9.01485071465054e-02, 9.01481468089003e-02, 9.01480936238812e-02, 9.01480894737681e-02, 9.01480143290527e-02, 9.01475052134906e-02, 9.01456336643953e-02, 9.01406449057802e-02, 9.01297276580735e-02, 9.01088218478172e-02, 9.00724728057050e-02, 9.00137401051069e-02, 8.99241686404579e-02, 8.97938287484372e-02, 8.96114309748697e-02, 8.93645193167316e-02, 8.90397442734994e-02, 8.86232137350085e-02, 8.81009156185719e-02, 8.74592013749371e-02, 8.66853142847290e-02, 8.57679412825298e-02, 8.46977624140948e-02, 8.34679685648716e-02, 8.20747164013440e-02, 8.05174900545360e-02, 7.87993422676614e-02, 7.69269935758068e-02, 7.49107763071457e-02, 7.27644201832298e-02, 7.05046871622521e-02, 6.81508738471379e-02, 6.57242091669339e-02, 6.32471821483200e-02, 6.07428386924660e-02, 5.82340869823703e-02, 5.57430484944632e-02, 5.32904859964063e-02, 5.08953321353296e-02, 4.85743332211221e-02, 4.63418136107379e-02, 4.42095576191690e-02, 4.21867988185873e-02, 4.02803013595623e-02, 3.84945147158469e-02, 3.68317819553212e-02, 3.52925820480223e-02, 3.38757884825682e-02, 3.25789291506986e-02, 3.13984356347632e-02, 3.03298733097925e-02, 2.93681467607598e-02, 2.85076777367301e-02, 2.77425551256201e-02, 2.70666582067278e-02, 2.64737557307200e-02, 2.59575842210964e-02, 2.55119093364229e-02, 2.51305742386734e-02, 2.48075387408022e-02, 2.45369126142663e-02, 2.43129858766196e-02, 2.41302581975944e-02, 2.39834688048610e-02, 2.38676274847194e-02, 2.37780465064077e-02, 2.37103726020429e-02, 2.36606175553433e-02, 2.36251855340903e-02, 2.36008950755817e-02, 2.35849936182788e-02, 2.35751626650931e-02, 2.35695120436041e-02, 2.35665622581358e-02, 2.35652145562995e-02, 2.35647089996647e-02, 2.35645714772232e-02, 2.35645511792377e-02, 2.35645495818808e-02, 2.35645204863175e-02, 2.35643233602865e-02, 2.35635987141372e-02, 2.35616671502967e-02, 2.35574403595500e-02, 2.35493470418966e-02, 2.35352773636528e-02, 2.35125495557327e-02, 2.34779022133935e-02, 2.34275156894057e-02, 2.33570655712754e-02, 2.32618104822713e-02, 2.31367152417441e-02, 2.29766086952985e-02, 2.27763732764785e-02, 2.25311606721128e-02, 2.22366250181523e-02, 2.18891621396473e-02, 2.14861408399400e-02, 2.10261105565266e-02, 2.05089692347435e-02, 1.99360763371708e-02, 1.93102986528579e-02, 1.86359809186065e-02, 1.79188388834213e-02, 1.71657787677510e-02, 1.63846533572496e-02, 1.55839704426839e-02, 1.47725732793056e-02, 1.39593147095702e-02, 1.31527463763881e-02, 1.23608421468226e-02, 1.15907708204190e-02, 1.08487279613456e-02, 1.01398309549658e-02, 9.46807588639942e-03, 8.83635023828333e-03, 8.24649215737594e-03, 7.69938527599834e-03, 7.19507762862648e-03, 6.73291375320062e-03, 6.31167029676760e-03, 5.92968713609565e-03, 5.58498801115093e-03, 5.27538676174544e-03, 4.99857721417740e-03, 4.75220636012352e-03, 4.53393159135018e-03, 4.34146343359743e-03, 4.17259558544103e-03, 4.02522424316842e-03, 3.89735875267405e-03, 3.78712559469818e-03, 3.69276758193493e-03, 3.61263993837113e-03, 3.54520467392661e-03, 3.48902439429111e-03, 3.44275642069647e-03, 3.40514784571155e-03, 3.37503191862965e-03, 3.35132593489362e-03, 3.33303059945322e-03, 3.31923064954154e-03, 3.30909636653232e-03, 3.30188548927524e-03, 3.29694497068189e-03, 3.29371200019324e-03, 3.29171374825776e-03, 3.29056537120920e-03, 3.28996593653130e-03, 3.28969207764374e-03, 3.28958934821344e-03, 3.28956140373649e-03, 3.28955727920999e-03, 3.28955695174926e-03, 3.28955095072558e-03, 3.28951029313629e-03, 3.28936083500864e-03, 3.28896246129418e-03, 3.28809076600933e-03, 3.28642188568296e-03, 3.28352131684184e-03, 3.27883759491243e-03, 3.27170175257052e-03, 3.26133347917690e-03, 3.24685483267118e-03, 3.22731216917305e-03, 3.20170661688834e-03, 3.16903290572412e-03, 3.12832567225030e-03, 3.07871152374517e-03, 3.01946423727983e-03, 2.95005960010641e-03, 2.87022570758211e-03, 2.77998417811176e-03, 2.67967786048118e-03, 2.56998129095720e-03, 2.45189141721625e-03, 2.32669784665163e-03, 2.19593388563055e-03, 2.06131161392875e-03, 1.92464587173902e-03, 1.78777309044875e-03, 1.65247128452606e-03, 1.52038728188941e-03, 1.39297647535841e-03, 1.27145903883985e-03, 1.15679465458030e-03, 1.04967549710366e-03, 9.50535012936626e-04, 8.59568642351149e-04, 7.76762517485707e-04, 7.01927075997588e-04, 6.34733508203311e-04, 5.74751122812245e-04, 5.21483083929297e-04, 4.74397523625214e-04, 4.32951800053871e-04, 3.96609632975498e-04, 3.64852812015926e-04, 3.37189813759409e-04, 3.13162731668370e-04, 2.92352427052833e-04, 2.74381047113081e-04, 2.58911489027014e-04, 2.45644416629887e-04, 2.34314123808588e-04, 2.24684445448986e-04, 2.16545295252502e-04, 2.09709797733382e-04, 2.04011724420639e-04, 1.99303026651165e-04, 1.95451460541562e-04, 1.92338430998973e-04, 1.89857186228437e-04, 1.87911422035114e-04, 1.86414276722502e-04, 1.85287649955512e-04, 1.84461764312971e-04, 1.83874891330760e-04, 1.83473171038173e-04, 1.83210460019012e-04, 1.83048148290510e-04, 1.82954892164133e-04, 1.82906220604906e-04, 1.82883985946347e-04, 1.82875645570846e-04, 1.82873376843418e-04, 1.82873041985943e-04, 1.82873015155849e-04, 1.82872520351651e-04, 1.82869168018129e-04, 1.82856844944013e-04, 1.82823999654858e-04, 1.82752136314690e-04, 1.82614577528994e-04, 1.82375574691918e-04, 1.81989854215914e-04, 1.81402697222666e-04, 1.80550655382916e-04, 1.79363001389987e-04, 1.77763992887492e-04, 1.75675987538426e-04, 1.73023378415849e-04, 1.69737221391131e-04, 1.65760303655560e-04, 1.61052273465856e-04, 1.55594348967725e-04, 1.49393094670103e-04, 1.42482834419866e-04, 1.34926451362816e-04, 1.26814527689589e-04, 1.18262864738483e-04, 1.09408300986047e-04, 1.00402490026852e-04, 9.14032461238889e-05, 8.25636738284174e-05, 7.40207375002910e-05, 6.58865057339142e-05, 5.82454349888906e-05, 5.11584924005604e-05, 4.46705827670257e-05, 3.88147865066912e-05, 3.36083576293333e-05, 2.90419930131236e-05, 2.50727025203901e-05, 2.16328444153646e-05, 1.86550603444857e-05, 1.60955289061349e-05, 1.39336999893658e-05, 1.21421432481839e-05, 1.06662659007384e-05, 9.44185146778409e-06, 8.41015787855781e-06, 7.52655480340536e-06, 6.76819363221171e-06, 6.12736923955026e-06, 5.59572741920533e-06, 5.15858981486872e-06, 4.80022861637508e-06, 4.50680868979703e-06, 4.26596198721541e-06, 4.06731505009361e-06, 3.90315435794264e-06, 3.76780745625638e-06, 3.65666514667851e-06, 3.56577136173541e-06, 3.49179957126466e-06, 3.43202052377478e-06, 3.38418927015027e-06, 3.34642532706653e-06, 3.31712021406361e-06, 3.29486843853033e-06, 3.27841845049193e-06, 3.26664499539220e-06, 3.25854063731867e-06, 3.25321911276433e-06, 3.24992232236809e-06, 3.24802495654808e-06, 3.24703375298441e-06, 3.24658072159423e-06, 3.24641075266641e-06, 3.24636451409788e-06, 3.24635768933490e-06, 3.24635712800614e-06, 3.24634668302792e-06, 3.24627590231330e-06, 3.24601574774874e-06, 3.24532250672860e-06, 3.24380660288170e-06, 3.24090815340128e-06, 3.23588207962189e-06, 3.22779592210054e-06, 3.21554299986815e-06, 3.19787220348644e-06, 3.17343394584668e-06, 3.14084145713878e-06, 3.09874572934632e-06, 3.04590830424445e-06, 2.98121364750159e-06, 2.90353183274306e-06, 2.81147369918244e-06, 2.70347506159396e-06, 2.57890313548614e-06, 2.44012351302558e-06, 2.29371226012631e-06, 2.14863042558022e-06, 2.01264399635193e-06, 1.89086263446825e-06, 1.78479001178289e-06, 1.68699506275516e-06, 1.57957994417757e-06, 1.44891014386417e-06, 1.30451068258708e-06, 1.17702902713290e-06, 1.09656467728314e-06, 1.07844606134222e-06, 1.11619491605097e-06, 1.15215204843366e-06, 1.10572943157194e-06, 9.40729144360830e-07, 6.83679068913534e-07, 5.58767911940095e-07, 5.92577521274507e-07, 5.61313758140614e-07, 4.43936301053327e-07, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, 5.20183938590435e-07, 7.06128339800216e-07, 7.98474321328513e-07, 7.87984681949183e-07, 6.91629303162244e-07, 5.41276193644080e-07, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14],
[ 1.80000000000000e+01, 1.79999999896303e+01, 1.79999996689778e+01, 1.79999974965070e+01, 1.79999895102257e+01, 1.79999682213567e+01, 1.79999216292830e+01, 1.79998323925896e+01, 1.79996771863897e+01, 1.79994262705698e+01, 1.79990432872857e+01, 1.79984852991917e+01, 1.79977030727275e+01, 1.79966416035510e+01, 1.79952408741463e+01, 1.79934368269763e+01, 1.79911625305034e+01, 1.79883495101236e+01, 1.79849292116788e+01, 1.79808345617814e+01, 1.79760015867178e+01, 1.79703710501594e+01, 1.79638900692172e+01, 1.79565136684666e+01, 1.79482062323341e+01, 1.79389428176578e+01, 1.79287102902667e+01, 1.79175082521263e+01, 1.79053497290250e+01, 1.78922615930410e+01, 1.78782846992075e+01, 1.78634737219498e+01, 1.78478966839952e+01, 1.78316341784318e+01, 1.78147782932350e+01, 1.77974312565495e+01, 1.77797038299448e+01, 1.77617134852878e+01, 1.77435824083581e+01, 1.77254353784505e+01, 1.77073975776268e+01, 1.76895923857604e+01, 1.76721392179706e+01, 1.76551514594694e+01, 1.76387345494041e+01, 1.76229842601905e+01, 1.76079852123989e+01, 1.75938096578096e+01, 1.75805165551358e+01, 1.75681509544334e+01, 1.75567436976734e+01, 1.75463114345775e+01, 1.75368569448184e+01, 1.75283697502091e+01, 1.75208269936896e+01, 1.75141945558529e+01, 1.75084283745244e+01, 1.75034759285948e+01, 1.74992778439730e+01, 1.74957695772284e+01, 1.74928831312945e+01, 1.74905487575335e+01, 1.74886965995405e+01, 1.74872582362906e+01, 1.74861680855547e+01, 1.74853646328790e+01, 1.74847914567092e+01, 1.74843980263327e+01, 1.74841402560296e+01, 1.74839808059824e+01, 1.74838891278988e+01, 1.74838412607352e+01, 1.74838193891610e+01, 1.74838111842698e+01, 1.74838089523301e+01, 1.74838086229002e+01, 1.74838085796859e+01, 1.74838075728452e+01, 1.74838007513420e+01, 1.74837756746389e+01, 1.74837088283101e+01, 1.74835625314517e+01, 1.74832823351463e+01, 1.74827950078348e+01, 1.74820071866287e+01, 1.74808047545102e+01, 1.74790529825780e+01, 1.74765974546468e+01, 1.74732657692894e+01, 1.74688699925430e+01, 1.74632098136618e+01, 1.74560763370927e+01, 1.74472564266893e+01, 1.74365375032825e+01, 1.74237126840390e+01, 1.74085861412860e+01, 1.73909785492773e+01, 1.73707324793320e+01, 1.73477175967646e+01, 1.73218355073164e+01, 1.72930240972439e+01, 1.72612612112172e+01, 1.72265675175708e+01, 1.71890084231302e+01, 1.71486949214106e+01, 1.71057832891830e+01, 1.70604735868092e+01, 1.70130069655035e+01, 1.69636618366280e+01, 1.69127490101313e+01, 1.68606059567557e+01, 1.68075903874136e+01, 1.67540733698330e+01, 1.67004322152788e+01, 1.66470433665555e+01, 1.65942755038710e+01, 1.65424830599830e+01, 1.64920003036738e+01, 1.64431361145216e+01, 1.63961695354800e+01, 1.63513461556123e+01, 1.63088753453529e+01, 1.62689283418839e+01, 1.62316371628506e+01, 1.61970943123097e+01, 1.61653532326788e+01, 1.61364294495263e+01, 1.61103023512384e+01, 1.60869175419543e+01, 1.60661897029065e+01, 1.60480058938809e+01, 1.60322292226389e+01, 1.60187028057480e+01, 1.60072539395233e+01, 1.59976983950121e+01, 1.59898447466284e+01, 1.59834986407222e+01, 1.59784669086213e+01, 1.59745614290798e+01, 1.59716026480677e+01, 1.59694226698039e+01, 1.59678678420222e+01, 1.59668007706574e+01, 1.59661017141941e+01, 1.59656693254054e+01, 1.59654207274898e+01, 1.59652909319555e+01, 1.59652316261580e+01, 1.59652093783205e+01, 1.59652033263517e+01, 1.59652024330932e+01, 1.59652023542355e+01, 1.59652008090570e+01, 1.59651903402390e+01, 1.59651518556605e+01, 1.59650492691649e+01, 1.59648247573782e+01, 1.59643947760782e+01, 1.59636469900244e+01, 1.59624382482429e+01, 1.59605937116067e+01, 1.59579072112396e+01, 1.59541428847986e+01, 1.59490381030590e+01, 1.59423076615931e+01, 1.59336491720457e+01, 1.59227495452700e+01, 1.59092924154570e+01, 1.58929663118411e+01, 1.58734733444560e+01, 1.58505381349440e+01, 1.58239166950508e+01, 1.57934049367520e+01, 1.57588464915425e+01, 1.57201395245733e+01, 1.56772422538085e+01, 1.56301769259822e+01, 1.55790320593384e+01, 1.55239628356912e+01, 1.54651896072163e+01, 1.54029945708196e+01, 1.53377167480439e+01, 1.52697454840372e+01, 1.51995127386474e+01, 1.51274844815117e+01, 1.50541515189074e+01, 1.49800200738191e+01, 1.49056024154410e+01, 1.48314077953869e+01, 1.47579339012685e+01, 1.46856589898915e+01, 1.46150348168594e+01, 1.45464804400147e+01, 1.44803769423001e+01, 1.44170630952256e+01, 1.43568319660990e+01, 1.42999284589737e+01, 1.42465477692993e+01, 1.41968347242440e+01, 1.41508839736908e+01, 1.41087409904722e+01, 1.40704038322485e+01, 1.40358256114230e+01, 1.40049176135155e+01, 1.39775529983676e+01, 1.39535710122527e+01, 1.39327816322955e+01, 1.39149705575136e+01, 1.38999044533732e+01, 1.38873363492851e+01, 1.38770110814296e+01, 1.38686706673229e+01, 1.38620594943818e+01, 1.38569292031869e+01, 1.38530431479345e+01, 1.38501803222906e+01, 1.38481386488991e+01, 1.38467375452508e+01, 1.38458196972625e+01, 1.38452519942272e+01, 1.38449256039396e+01, 1.38447551937425e+01, 1.38446773307622e+01, 1.38446481214835e+01, 1.38446401758339e+01, 1.38446390030720e+01, 1.38446389090020e+01, 1.38446371729330e+01, 1.38446254108046e+01, 1.38445821718930e+01, 1.38444669121881e+01, 1.38442146658868e+01, 1.38437315715654e+01, 1.38428914274369e+01, 1.38415334263104e+01, 1.38394611934741e+01, 1.38364432196120e+01, 1.38322147457744e+01, 1.38264811180952e+01, 1.38189225863016e+01, 1.38092004723612e+01, 1.37969645846692e+01, 1.37818617005008e+01, 1.37635448872746e+01, 1.37416833844702e+01, 1.37159727263135e+01, 1.36861447543391e+01, 1.36519771521992e+01, 1.36133021354647e+01, 1.35700139483147e+01, 1.35220748569919e+01, 1.34695193850572e+01, 1.34124566044831e+01, 1.33510703749657e+01, 1.32856175064463e+01, 1.32164239018280e+01, 1.31438788141811e+01, 1.30684274223884e+01, 1.29905619892066e+01, 1.29108119146138e+01, 1.28297330334709e+01, 1.27478965277779e+01, 1.26658778276050e+01, 1.25842458588779e+01, 1.25035529598735e+01, 1.24243257333935e+01, 1.23470570330253e+01, 1.22721992070948e+01, 1.22001586513447e+01, 1.21312916587085e+01, 1.20659015070576e+01, 1.20042366954705e+01, 1.19464902252887e+01, 1.18927998206316e+01, 1.18432489897580e+01, 1.17978688394071e+01, 1.17566405655231e+01, 1.17194985532125e+01, 1.16863340250598e+01, 1.16569991795693e+01, 1.16313117605596e+01, 1.16090599942206e+01, 1.15900078238368e+01, 1.15739003635749e+01, 1.15604694830428e+01, 1.15494394244109e+01, 1.15405323447307e+01, 1.15334736686841e+01, 1.15279971323654e+01, 1.15238493977618e+01, 1.15207941211175e+01, 1.15186153668613e+01, 1.15171202724283e+01, 1.15161408879513e+01, 1.15155351378638e+01, 1.15151868780481e+01, 1.15150050510737e+01, 1.15149219719220e+01, 1.15148908059175e+01, 1.15148823279947e+01, 1.15148810766706e+01, 1.15148809801564e+01, 1.15148792470718e+01, 1.15148675051646e+01, 1.15148243406135e+01, 1.15147092793212e+01, 1.15144574682289e+01, 1.15139752111699e+01, 1.15131365348922e+01, 1.15117809380018e+01, 1.15097124491527e+01, 1.15067000905506e+01, 1.15024798082434e+01, 1.14967578912646e+01, 1.14892158571410e+01, 1.14795167316862e+01, 1.14673125971133e+01, 1.14522532258761e+01, 1.14339955607703e+01, 1.14122137480959e+01, 1.13866093842299e+01, 1.13569216013390e+01, 1.13229365996676e+01, 1.12844962356594e+01, 1.12415052995799e+01, 1.11939371637978e+01, 1.11418375515559e+01, 1.10853262615016e+01, 1.10245967788573e+01, 1.09599138018759e+01, 1.08916088039466e+01, 1.08200738302902e+01, 1.07457537888065e+01, 1.06691375353764e+01, 1.05907480756555e+01, 1.05111322109519e+01, 1.04308499487938e+01, 1.03504639825227e+01, 1.02705295209026e+01, 1.01915847193519e+01, 1.01141419293415e+01, 1.00386799420814e+01, 9.96563735778813e+00, 9.89540716442313e+00, 9.82833256245001e+00, 9.76470402789228e+00, 9.70475756763705e+00, 9.64867409068944e+00, 9.59657979810686e+00, 9.54854748267730e+00, 9.50459862608546e+00, 9.46470618460937e+00, 9.42879796214350e+00, 9.39676047928002e+00, 9.36844325719186e+00, 9.34366344342364e+00, 9.32221071218482e+00, 9.30385237369208e+00, 9.28833862542300e+00, 9.27540787322843e+00, 9.26479204295005e+00, 9.25622179466782e+00, 9.24943154333370e+00, 9.24416418277035e+00, 9.24017540620022e+00, 9.23723751678622e+00, 9.23514262695922e+00, 9.23370515601715e+00, 9.23276355158454e+00, 9.23218118153186e+00, 9.23184636795934e+00, 9.23167156259984e+00, 9.23159169201628e+00, 9.23156172971057e+00, 9.23155357922822e+00, 9.23155237623378e+00, 9.23155228514667e+00, 9.23155067154755e+00, 9.23153973917416e+00, 9.23149955058396e+00, 9.23139242233369e+00, 9.23115797364137e+00, 9.23070897199265e+00, 9.22992814073346e+00, 9.22866607327241e+00, 9.22674037489043e+00, 9.22393612590552e+00, 9.22000772846007e+00, 9.21468216286715e+00, 9.20766363784715e+00, 9.19863957210872e+00, 9.18728779321188e+00, 9.17328478491805e+00, 9.15631475857049e+00, 9.13607927057450e+00, 9.11230706054939e+00, 9.08476374737547e+00, 9.05326099732879e+00, 9.01766477348661e+00, 8.97790229136210e+00, 8.93396734368626e+00, 8.88592371710680e+00, 8.83390650316411e+00, 8.77812120122288e+00, 8.71884061640247e+00, 8.65639966398246e+00, 8.59118829551992e+00, 8.52364285317600e+00, 8.45423623037481e+00, 8.38346726325986e+00, 8.31184979503586e+00, 8.23990184347217e+00, 8.16813526279324e+00, 8.09704622977100e+00, 8.02710680700680e+00, 7.95875775221030e+00, 7.89240265885070e+00, 7.82840343798189e+00, 7.76707708865011e+00, 7.70869365818323e+00, 7.65347526467004e+00, 7.60159604092851e+00, 7.55318285964701e+00, 7.50831670974266e+00, 7.46703461065944e+00, 7.42933197090112e+00, 7.39516531669016e+00, 7.36445533409709e+00, 7.33709018185735e+00, 7.31292904160282e+00, 7.29180587714433e+00, 7.27353337492812e+00, 7.25790703433947e+00, 7.24470936985000e+00, 7.23371417797443e+00, 7.22469081161070e+00, 7.21740839366428e+00, 7.21163989202442e+00, 7.20716597008392e+00, 7.20377852211330e+00, 7.20128380180258e+00, 7.19950505581856e+00, 7.19828458263098e+00, 7.19748515011030e+00, 7.19699072307668e+00, 7.19670647330171e+00, 7.19655806833675e+00, 7.19649026066203e+00, 7.19646482362658e+00, 7.19645790413281e+00, 7.19645688282990e+00, 7.19645680616037e+00, 7.19645545668967e+00, 7.19644631382706e+00, 7.19641270369608e+00, 7.19632311142010e+00, 7.19612704088108e+00, 7.19575154221044e+00, 7.19509854645512e+00, 7.19404313058161e+00, 7.19243281704073e+00, 7.19008798996742e+00, 7.18680348396850e+00, 7.18235137055055e+00, 7.17648493093664e+00, 7.16894376212175e+00, 7.15945991601192e+00, 7.14776492075870e+00, 7.13359748132362e+00, 7.11671160627560e+00, 7.09688486405793e+00, 7.07392643916662e+00, 7.04768464153260e+00, 7.01805352490340e+00, 6.98497829483479e+00, 6.94845923471722e+00, 6.90855394736551e+00, 6.86537779583379e+00, 6.81910252371406e+00, 6.76995313398917e+00, 6.71820319759030e+00, 6.66416883969459e+00, 6.60820170672716e+00, 6.55068124593415e+00, 6.49200663148114e+00, 6.43258864882641e+00, 6.37284180791639e+00, 6.31317690261576e+00, 6.25399417629176e+00, 6.19567719846310e+00, 6.13858751002815e+00, 6.08306005769181e+00, 6.02939941243037e+00, 5.97787675092432e+00, 5.92872757044655e+00, 5.88215010393250e+00, 5.83830440038465e+00, 5.79731203460801e+00, 5.75925640863774e+00, 5.72418360496925e+00, 5.69210374919085e+00, 5.66299283736672e+00, 5.63679498192587e+00, 5.61342502899783e+00, 5.59277149989260e+00, 5.57469980929249e+00, 5.55905571214135e+00, 5.54566892966420e+00, 5.53435690210215e+00, 5.52492861155811e+00, 5.51718841308715e+00, 5.51093980639390e+00, 5.50598907500153e+00, 5.50214871544119e+00, 5.49924057679632e+00, 5.49709863163745e+00, 5.49557130360967e+00, 5.49452328501565e+00, 5.49383678967145e+00, 5.49341220176473e+00, 5.49316809976779e+00, 5.49304065475353e+00, 5.49298242366903e+00, 5.49296057911337e+00, 5.49295463685840e+00, 5.49295375979379e+00, 5.49295369415249e+00, 5.49295254145824e+00, 5.49294473178362e+00, 5.49291602257110e+00, 5.49283949417841e+00, 5.49267201278652e+00, 5.49235126252723e+00, 5.49179346462892e+00, 5.49089188673008e+00, 5.48951622855797e+00, 5.48751295061647e+00, 5.48470658993996e+00, 5.48090208090445e+00, 5.47588806924267e+00, 5.46944117380546e+00, 5.46133111374303e+00, 5.45132657975081e+00, 5.43920168860243e+00, 5.42474282277574e+00, 5.40775562443623e+00, 5.38807188844691e+00, 5.36555608531305e+00, 5.34011124432775e+00, 5.31168394093195e+00, 5.28026816042709e+00, 5.24590785130575e+00, 5.20869803304946e+00, 5.16878438199072e+00, 5.12636128128843e+00, 5.08166838414948e+00, 5.03498580080614e+00, 4.98662807779368e+00, 4.93693719138862e+00, 4.88627482382100e+00, 4.83501422806411e+00, 4.78353201025423e+00, 4.73220016292624e+00, 4.68137866274531e+00, 4.63140890134956e+00, 4.58260814965434e+00, 4.53526517173841e+00, 4.48963701530778e+00, 4.44594692416056e+00, 4.40438325479101e+00, 4.36509924075851e+00, 4.32821343579615e+00, 4.29381067607550e+00, 4.26194342669084e+00, 4.23263340946091e+00, 4.20587344166401e+00, 4.18162944341092e+00, 4.15984259237206e+00, 4.14043161779255e+00, 4.12329523176025e+00, 4.10831469584043e+00, 4.09535651696923e+00, 4.08427525932914e+00, 4.07491645002175e+00, 4.06711954668393e+00, 4.06072092556490e+00, 4.05555683967880e+00, 4.05146628909218e+00, 4.04829373978121e+00, 4.04589162433279e+00, 4.04412255752661e+00, 4.04286120285209e+00, 4.04199573244079e+00, 4.04142883266757e+00, 4.04107822048341e+00, 4.04087665084173e+00, 4.04077141261140e+00, 4.04072332823219e+00, 4.04070529009518e+00, 4.04070038328207e+00, 4.04069965904664e+00, 4.04069960482310e+00, 4.04069865235942e+00, 4.04069219927771e+00, 4.04066847706491e+00, 4.04060524234109e+00, 4.04046685462304e+00, 4.04020182531196e+00, 4.03974093624826e+00, 4.03899601463420e+00, 4.03785943768568e+00, 4.03620442464022e+00, 4.03388615516011e+00, 4.03074373113176e+00, 4.02660297290497e+00, 4.02128001098288e+00, 4.01458560042212e+00, 4.00633004876593e+00, 3.99632861103747e+00, 3.98440716978580e+00, 3.97040798770066e+00, 3.95419529858099e+00, 3.93566049309301e+00, 3.91472666180408e+00, 3.89135228117923e+00, 3.86553386847654e+00, 3.83730748640336e+00, 3.80674904330512e+00, 3.77397340296147e+00, 3.73913238221756e+00, 3.70241176753936e+00, 3.66402751787104e+00, 3.62422133863329e+00, 3.58325581155422e+00, 3.54140925150572e+00, 3.49897044058973e+00, 3.45623336738240e+00, 3.41349207999277e+00, 3.37103574751441e+00, 3.32914401518425e+00, 3.28808273200105e+00, 3.24810012292204e+00, 3.20942346866766e+00, 3.17225634333831e+00, 3.13677644353558e+00, 3.10313402367622e+00, 3.07145093256262e+00, 3.04182022804870e+00, 3.01430633152155e+00, 2.98894567300568e+00, 2.96574777136752e+00, 2.94469669206199e+00, 2.92575282630727e+00, 2.90885493937325e+00, 2.89392244059029e+00, 2.88085783257009e+00, 2.86954930103622e+00, 2.85987340894227e+00, 2.85169785890812e+00, 2.84488428646319e+00, 2.83929104348803e+00, 2.83477592717457e+00, 2.83119880552425e+00, 2.82842408669264e+00, 2.82632297717107e+00, 2.82477547356805e+00, 2.82367203513594e+00, 2.82291488946809e+00, 2.82241893199261e+00, 2.82211219076970e+00, 2.82193584117846e+00, 2.82184376968308e+00, 2.82180170120161e+00, 2.82178591982062e+00, 2.82178162689846e+00, 2.82178099327201e+00, 2.82178094570651e+00, 2.82178010851392e+00, 2.82177443641014e+00, 2.82175358514450e+00, 2.82169800323284e+00, 2.82157636310679e+00, 2.82134340590594e+00, 2.82093828572674e+00, 2.82028348865593e+00, 2.81928438977855e+00, 2.81782949733068e+00, 2.81579141693379e+00, 2.81302855006557e+00, 2.80938751936363e+00, 2.80470628902298e+00, 2.79881792177010e+00, 2.79155488542508e+00, 2.78275379309781e+00, 2.77226043322478e+00, 2.75993492089190e+00, 2.74565678232853e+00, 2.72932977218298e+00, 2.71088622001001e+00, 2.69029070962500e+00, 2.66754291322935e+00, 2.64267943131335e+00, 2.61577452829193e+00, 2.58693970084332e+00, 2.55632206853843e+00, 2.52410163157035e+00, 2.49048749484109e+00, 2.45571320777477e+00, 2.42003141148895e+00, 2.38370801619788e+00, 2.34701614944973e+00, 2.31023011853715e+00, 2.27361961798121e+00, 2.23744438659834e+00, 2.20194948088885e+00, 2.16736128596098e+00, 2.13388433615537e+00, 2.10169896925866e+00, 2.07095979452622e+00, 2.04179491863932e+00, 2.01430584704584e+00, 1.98856796154933e+00, 1.96463146812081e+00, 1.94252271048014e+00, 1.92224575324039e+00, 1.90378415128520e+00, 1.88710283749870e+00, 1.87215007711844e+00, 1.85885945226694e+00, 1.84715185343741e+00, 1.83693746504813e+00, 1.82811773918591e+00, 1.82058735521321e+00, 1.81423616318499e+00, 1.80895110642286e+00, 1.80461811373936e+00, 1.80112394545392e+00, 1.79835797035352e+00, 1.79621384401733e+00, 1.79459105330798e+00, 1.79339628809423e+00, 1.79254460000545e+00, 1.79196030959917e+00, 1.79157762786242e+00, 1.79134096530631e+00, 1.79120491163611e+00, 1.79113388046126e+00, 1.79110142598251e+00, 1.79108925122782e+00, 1.79108593940319e+00, 1.79108545058494e+00, 1.79108541373374e+00, 1.79108476303975e+00, 1.79108035449464e+00, 1.79106414825222e+00, 1.79102094863488e+00, 1.79092640872364e+00, 1.79074535882860e+00, 1.79043052827162e+00, 1.78992172332611e+00, 1.78914551639190e+00, 1.78801549203910e+00, 1.78643308378570e+00, 1.78428902070185e+00, 1.78146538460132e+00, 1.77783825641181e+00, 1.77328090447228e+00, 1.76766743871086e+00, 1.76087682426805e+00, 1.75279711812799e+00, 1.74332976521794e+00, 1.73239376907470e+00, 1.71992953945203e+00, 1.70590221774495e+00, 1.69030429274671e+00, 1.67315734490155e+00, 1.65451279641956e+00, 1.63445159545621e+00, 1.61308282166604e+00, 1.59054126325041e+00, 1.56698407679990e+00, 1.54258669527696e+00, 1.51753819139878e+00, 1.49203632963882e+00, 1.46628254793534e+00, 1.44047709983436e+00, 1.41481456104192e+00, 1.38947986474330e+00, 1.36464498223469e+00, 1.34046631453633e+00, 1.31708281161641e+00, 1.29461479273346e+00, 1.27316340707637e+00, 1.25281064985016e+00, 1.23361983542020e+00, 1.21563642520577e+00, 1.19888911207147e+00, 1.18339107296388e+00, 1.16914131536087e+00, 1.15612605877429e+00, 1.14432010842427e+00, 1.13368819301920e+00, 1.12418625145326e+00, 1.11576266364762e+00, 1.10835942847116e+00, 1.10191329665843e+00, 1.09635686903036e+00, 1.09161967036124e+00, 1.08762920724782e+00, 1.08431201470746e+00, 1.08159469140253e+00, 1.07940491783711e+00, 1.07767244611896e+00, 1.07633004445475e+00, 1.07531437497064e+00, 1.07456678019332e+00, 1.07403395197155e+00, 1.07366845701515e+00, 1.07342909566864e+00, 1.07328107493200e+00, 1.07319598282639e+00, 1.07315155856374e+00, 1.07313126109030e+00, 1.07312364686212e+00, 1.07312157561262e+00, 1.07312126990072e+00, 1.07312124672818e+00, 1.07312083589935e+00, 1.07311805247639e+00, 1.07310782036086e+00, 1.07308054561834e+00, 1.07302085713888e+00, 1.07290655279617e+00, 1.07270779609864e+00, 1.07238660482156e+00, 1.07189666888736e+00, 1.07118352935618e+00, 1.07018514319914e+00, 1.06883284891348e+00, 1.06705273545566e+00, 1.06476740110797e+00, 1.06189806983565e+00, 1.05836701103206e+00, 1.05410018547710e+00, 1.04903001763612e+00, 1.04309817438340e+00, 1.03625821541312e+00, 1.02847797353947e+00, 1.01974152590518e+00, 1.01005063111613e+00, 9.99425532628148e-01, 9.87905064085275e-01, 9.75546035101271e-01, 9.62421922378940e-01, 9.48620936535523e-01, 9.34243574901142e-01, 9.19399800791055e-01, 9.04206007402600e-01, 8.88781928225836e-01, 8.73247646102445e-01, 8.57720831847246e-01, 8.42314313894658e-01, 8.27134046625034e-01, 8.12277510765089e-01, 7.97832547932405e-01, 7.83876605499493e-01, 7.70476348911735e-01, 7.57687586781650e-01, 7.45555449002318e-01, 7.34114758673769e-01, 7.23390543397539e-01, 7.13398638987162e-01, 7.04146347519821e-01, 6.95633120818794e-01, 6.87851249102806e-01, 6.80786542143192e-01, 6.74418996547832e-01, 6.68723447635401e-01, 6.63670207790194e-01, 6.59225695283055e-01, 6.55353058432805e-01, 6.52012799805224e-01, 6.49163404050771e-01, 6.46761971125155e-01, 6.44764854186874e-01, 6.43128298613725e-01, 6.41809075536841e-01, 6.40765100292587e-01, 6.39956023486707e-01, 6.39343780202882e-01, 6.38893081497606e-01, 6.38571831894399e-01, 6.38351457243829e-01, 6.38207129092162e-01, 6.38117874543902e-01, 6.38066564362585e-01, 6.38039776496985e-01, 6.38027537053631e-01, 6.38022945640738e-01, 6.38021696667451e-01, 6.38021512321700e-01, 6.38021498259882e-01, 6.38021247786416e-01, 6.38019550793214e-01, 6.38013312487530e-01, 6.37996683615164e-01, 6.37960292547447e-01, 6.37890602413297e-01, 6.37769420428100e-01, 6.37573584272705e-01, 6.37274847623594e-01, 6.36839983472886e-01, 6.36231119367832e-01, 6.35406312786365e-01, 6.34320367242996e-01, 6.32925880223972e-01, 6.31174502748930e-01, 6.29018377625510e-01, 6.26411710039584e-01, 6.23312411099594e-01, 6.19683743711068e-01, 6.15495892225390e-01, 6.10727374150073e-01, 6.05366214976538e-01, 5.99410816459542e-01, 5.92870464291764e-01, 5.85765442030457e-01, 5.78126742532170e-01, 5.69995393642845e-01, 5.61421438843443e-01, 5.52462633504561e-01, 5.43182931448175e-01, 5.33650843577064e-01, 5.23937750297232e-01, 5.14116243106354e-01, 5.04258559519677e-01, 4.94435161292621e-01, 4.84713490558750e-01, 4.75156923673246e-01, 4.65823929452061e-01, 4.56767427848430e-01, 4.48034337171823e-01, 4.39665292632181e-01, 4.31694515952955e-01, 4.24149814603364e-01, 4.17052689402812e-01, 4.10418530447171e-01, 4.04256883169342e-01, 3.98571768621611e-01, 3.93362044559467e-01, 3.88621796459990e-01, 3.84340750089386e-01, 3.80504699522906e-01, 3.77095946504401e-01, 3.74093748614474e-01, 3.71474774814651e-01, 3.69213567497074e-01, 3.67283010175239e-01, 3.65654799420259e-01, 3.64299918641645e-01, 3.63189109935185e-01, 3.62293338613934e-01, 3.61584243371517e-01, 3.61034563483955e-01, 3.60618533223132e-01, 3.60312232899400e-01, 3.60093885807427e-01, 3.59944090902804e-01, 3.59845982313441e-01, 3.59785308749249e-01, 3.59750428416223e-01, 3.59732218012034e-01, 3.59723897590317e-01, 3.59720776322963e-01, 3.59719927263516e-01, 3.59719801944159e-01, 3.59719792317389e-01, 3.59719619958089e-01, 3.59718452199139e-01, 3.59714159404060e-01, 3.59702716467918e-01, 3.59677674285109e-01, 3.59629717053682e-01, 3.59546323858776e-01, 3.59411551260216e-01, 3.59205951575483e-01, 3.58906639056656e-01, 3.58487513020076e-01, 3.57919643002937e-01, 3.57171816026568e-01, 3.56211239960626e-01, 3.55004389856923e-01, 3.53517976181408e-01, 3.51720005543349e-01, 3.49580896420349e-01, 3.47074605283222e-01, 3.44179713291091e-01, 3.40880421178252e-01, 3.37167400753959e-01, 3.33038455957546e-01, 3.28498954640224e-01, 3.23562003733713e-01, 3.18248354360094e-01, 3.12586038571919e-01, 3.06609754464393e-01, 3.00360030072781e-01, 2.93882207649517e-01, 2.87225297816086e-01, 2.80440757299424e-01, 2.73581244468531e-01, 2.66699403995446e-01, 2.59846726233249e-01, 2.53072519041183e-01, 2.46423020566429e-01, 2.39940671676678e-01, 2.33663557030976e-01, 2.27625014783893e-01, 2.21853407122911e-01, 2.16372037584286e-01, 2.11199196573013e-01, 2.06348313776576e-01, 2.01828195125131e-01, 1.97643322418603e-01, 1.93794195440354e-01, 1.90277698986250e-01, 1.87087480419592e-01, 1.84214326790986e-01, 1.81646533945845e-01, 1.79370263137635e-01, 1.77369883286669e-01, 1.75628299044298e-01, 1.74127266171453e-01, 1.72847696401149e-01, 1.71769953958410e-01, 1.70874145333528e-01, 1.70140402856763e-01, 1.69549161247524e-01, 1.69081424771431e-01, 1.68719021108114e-01, 1.68444836682575e-01, 1.68243027198027e-01, 1.68099196554433e-01, 1.68000537326833e-01, 1.67935926545226e-01, 1.67895971640917e-01, 1.67873003030238e-01, 1.67861011778296e-01, 1.67855532975071e-01, 1.67853477702824e-01, 1.67852918620440e-01, 1.67852836101112e-01, 1.67852829713773e-01, 1.67852714724045e-01, 1.67851935652628e-01, 1.67849071717171e-01, 1.67841437620392e-01, 1.67824731112322e-01, 1.67792738166433e-01, 1.67737108473226e-01, 1.67647213038268e-01, 1.67510094561418e-01, 1.67310519831062e-01, 1.67031141412288e-01, 1.66652773253322e-01, 1.66154781319509e-01, 1.65515585897129e-01, 1.64713266808768e-01, 1.63726256610904e-01, 1.62534100229362e-01, 1.61118252919355e-01, 1.59462882540183e-01, 1.57555637614454e-01, 1.55388340201963e-01, 1.52957562859456e-01, 1.50265052276364e-01, 1.47317968677200e-01, 1.44128919516881e-01, 1.40715777760892e-01, 1.37101288213604e-01, 1.33312478791098e-01, 1.29379906093703e-01, 1.25336774947929e-01, 1.21217978798850e-01, 1.17059111317538e-01, 1.12895499125443e-01, 1.08761301334447e-01, 1.04688714244553e-01, 1.00707309908703e-01, 9.68435264103853e-02, 9.31203166792933e-02, 8.95569524675039e-02, 8.61689714967291e-02, 8.29682492721453e-02, 7.99631728643106e-02, 7.71588920487289e-02, 7.45576233152114e-02, 7.21589840270980e-02, 6.99603369632307e-02, 6.79571291464462e-02, 6.61432128253310e-02, 6.45111403816081e-02, 6.30524285195264e-02, 6.17577901831190e-02, 6.06173351427799e-02, 5.96207420589216e-02, 5.87574060759550e-02, 5.80165666639800e-02, 5.73874205681300e-02, 5.68592244176216e-02, 5.64213908682127e-02, 5.60635811897683e-02, 5.57757960568246e-02, 5.55484650535654e-02, 5.53725341663025e-02, 5.52395494076379e-02, 5.51417337911637e-02, 5.50720542327632e-02, 5.50242746514940e-02, 5.49929916074561e-02, 5.49736492396242e-02, 5.49625310144783e-02, 5.49567267966573e-02, 5.49540749152371e-02, 5.49530801213507e-02, 5.49528095150258e-02, 5.49527695741645e-02, 5.49527664575145e-02, 5.49527100253790e-02, 5.49523276903194e-02, 5.49509222004012e-02, 5.49471757893549e-02, 5.49389774137623e-02, 5.49232786640310e-02, 5.48959850412558e-02, 5.48518891023377e-02, 5.47846517055532e-02, 5.46868371122322e-02, 5.45500068594755e-02, 5.43648759447137e-02, 5.41215328061142e-02, 5.38097217500564e-02, 5.34191828644346e-02, 5.29400401848434e-02, 5.23632242135754e-02, 5.16809102350884e-02, 5.08869497561619e-02, 4.99772694205554e-02, 4.89502104956188e-02, 4.78067829848321e-02, 4.65508118649645e-02, 4.51889588706393e-02, 4.37306113078222e-02, 4.21876388958786e-02, 4.05740296720261e-02, 3.89054254511798e-02, 3.71985851364975e-02, 3.54708094140896e-02, 3.37393624521134e-02, 3.20209249845753e-02, 3.03311088516800e-02, 2.86840563322741e-02, 2.70921393457364e-02, 2.55657648405538e-02, 2.41132843988865e-02, 2.27409990656149e-02, 2.14532451858211e-02, 2.02525438384598e-02, 1.91397952424168e-02, 1.81145000212268e-02, 1.71749910432837e-02, 1.63186622471372e-02, 1.55421839755689e-02, 1.48416975029200e-02, 1.42129843696459e-02, 1.36516086568719e-02, 1.31530323532800e-02, 1.27127054705518e-02, 1.23261335853009e-02, 1.19889260906483e-02, 1.16968287064623e-02, 1.14457438020388e-02, 1.12317418958590e-02, 1.10510673700084e-02, 1.09001410136642e-02, 1.07755615223191e-02, 1.06741075528944e-02, 1.05927413895053e-02, 1.05286147332997e-02, 1.04790766135139e-02, 1.04416829514840e-02, 1.04142069194992e-02, 1.03946489468001e-02, 1.03812450547380e-02, 1.03724721637768e-02, 1.03670491070460e-02, 1.03639322975582e-02, 1.03623053059572e-02, 1.03615619802966e-02, 1.03612831429238e-02, 1.03612072933146e-02, 1.03611960981098e-02, 1.03611952170937e-02, 1.03611791696494e-02, 1.03610704463293e-02, 1.03606707753059e-02, 1.03596054559425e-02, 1.03572743190519e-02, 1.03528110182793e-02, 1.03450527585771e-02, 1.03325226282607e-02, 1.03134267428615e-02, 1.02856683696366e-02, 1.02468809999152e-02, 1.01944818878756e-02, 1.01257468002167e-02, 1.00379055720410e-02, 9.92825653715101e-03, 9.79429605916238e-03, 9.63385737459177e-03, 9.44525099093951e-03, 9.22739724486377e-03, 8.97994063053886e-03, 8.70333544777679e-03, 8.39889340121402e-03, 8.06878607386352e-03, 7.71599858272509e-03, 7.34423488893654e-03, 6.95777969951195e-03, 6.56132608288473e-03, 6.15978124368866e-03, 5.75806488778596e-03, 5.36091497540880e-03, 4.97271438240787e-03, 4.59734929557243e-03, 4.23810649579792e-03, 3.89761258679044e-03, 3.57781427183568e-03, 3.27999544000690e-03, 3.00482441175376e-03, 2.75242334531001e-03, 2.52245148399766e-03, 2.31419446742009e-03, 2.12665309289406e-03, 1.95862642775424e-03, 1.80878578302313e-03, 1.67573756230155e-03, 1.55807426620314e-03, 1.45441390451850e-03, 1.36342874361617e-03, 1.28386472831110e-03, 1.21455311241903e-03, 1.15441586088494e-03, 1.10246629818889e-03, 1.05780631820451e-03, 1.01962127781619e-03, 9.87173499353844e-04, 9.59795122099649e-04, 9.36880878906976e-04, 9.17881231078610e-04, 9.02296171314987e-04, 8.89669897375711e-04, 8.79586464531457e-04, 8.71666441206287e-04, 8.65564517621209e-04, 8.60967952434066e-04, 8.57595688373181e-04, 8.55197927439377e-04, 8.53555931545458e-04, 8.52481808038874e-04, 8.51818051980757e-04, 8.51436647273829e-04, 8.51237574494235e-04, 8.51146629013755e-04, 8.51112514247632e-04, 8.51103234396078e-04, 8.51101864714821e-04, 8.51101755976913e-04, 8.51099763146124e-04, 8.51086261519640e-04, 8.51036629669011e-04, 8.50904341277214e-04, 8.50614891735546e-04, 8.50060792476626e-04, 8.49097932491105e-04, 8.47543635039472e-04, 8.45176764265432e-04, 8.41740247572886e-04, 8.36946357857471e-04, 8.30485026394712e-04, 8.22035315305669e-04, 8.11279957197542e-04, 7.97922569368836e-04, 7.81706788803405e-04, 7.62436186615645e-04, 7.39993464041078e-04, 7.14357174004476e-04, 6.85614125146117e-04, 6.53965767440923e-04, 6.19727258587943e-04, 5.83318552479284e-04, 5.45247669588633e-04, 5.06087196466609e-04, 4.66445881571076e-04, 4.26937813045924e-04, 3.88151965605488e-04, 3.50624823241219e-04, 3.14818322607380e-04, 2.81104600610149e-04, 2.49758125491439e-04, 2.20954934345485e-04, 1.94778047085009e-04, 1.71227727303602e-04, 1.50235061521897e-04, 1.31677251964391e-04, 1.15393055608592e-04, 1.01197013094634e-04, 8.88915476942032e-05, 7.82766049383099e-05, 6.91570107242967e-05, 6.13479295559216e-05, 5.46787127113646e-05, 4.89952346178768e-05, 4.41607403409705e-05, 4.00553826338523e-05, 3.65748938726540e-05, 3.36289515026067e-05, 3.11396448431367e-05, 2.90401556490849e-05, 2.72735311312999e-05, 2.57914144548010e-05, 2.45527530076117e-05, 2.35226007454514e-05, 2.26710787956108e-05, 2.19724869872095e-05, 2.14045571667043e-05, 2.09478548427455e-05, 2.05853205486051e-05, 2.03019202196231e-05, 2.00843736593015e-05, 1.99209444780024e-05, 1.98012842425044e-05, 1.97163236741460e-05, 1.96581992746303e-05, 1.96202015168012e-05, 1.95967304392752e-05, 1.95832469174206e-05, 1.95762102102037e-05, 1.95729957397361e-05, 1.95717899877934e-05, 1.95714620048193e-05, 1.95714135955427e-05, 1.95714097195321e-05, 1.95713381927476e-05, 1.95708536381289e-05, 1.95690724495989e-05, 1.95643251466038e-05, 1.95539393494926e-05, 1.95340628827978e-05, 1.94995399560582e-05, 1.94438551908149e-05, 1.93591622235090e-05, 1.92364142463930e-05, 1.90656120651056e-05, 1.88361813364679e-05, 1.85374834416052e-05, 1.81594544557129e-05, 1.76933512925359e-05, 1.71325608206985e-05, 1.64733997287669e-05, 1.57158177778172e-05, 1.48639322290478e-05, 1.39263369760556e-05, 1.29160739312795e-05, 1.18500494998841e-05, 1.07477751285501e-05, 9.62972517015714e-06, 8.51579188321239e-06, 7.42385758201181e-06, 6.36865530267983e-06, 5.36238455950411e-06, 4.41680929696986e-06, 3.54139873692816e-06, 2.73699181718712e-06, 1.99781397469678e-06, 1.31573180794931e-06, 6.51913914850866e-07, 6.58154892363427e-07, 1.17206973371253e-06, 1.49742479749193e-06, 1.67185249111412e-06, 1.79254267962392e-06, 1.91784003433781e-06, 2.01910027541885e-06, 2.08839050586107e-06, 2.15894569777220e-06, 2.22731259214015e-06, 2.26520103296934e-06, 2.28057188067510e-06, 2.30110243393256e-06, 2.32948702105773e-06, 2.35251950186921e-06, 2.36761474070590e-06, 2.38000452057463e-06, 2.39222962505892e-06, 2.40530372508374e-06, 2.42134235506342e-06, 2.44095205559347e-06, 2.46149636495829e-06, 2.47930704233568e-06, 2.49231462134958e-06, 2.50051956385278e-06, 2.50504037424548e-06, 2.50716094285508e-06, 2.50789343425595e-06, 2.50791363853864e-06, 2.50763137543686e-06, 2.50727428152686e-06, 2.50695324055528e-06, 2.50670904122785e-06, 2.50654446176668e-06, 2.50644513499765e-06, 2.50639190468860e-06, 2.50636726661924e-06, 2.50635797627297e-06, 2.50635544318811e-06, 2.50635507015318e-06, 2.50635503462604e-06, 2.50635434539959e-06, 2.50634968423924e-06, 2.50633256726474e-06, 2.50628708187151e-06, 2.50618827735138e-06, 2.50600164264370e-06, 2.50568412596408e-06, 2.50518554878454e-06, 2.50444704974484e-06, 2.50339410712286e-06, 2.50194065287701e-06, 2.50007094493299e-06, 2.49810927854810e-06, 2.49715914790727e-06, 2.49920397621395e-06, 2.50587125094626e-06, 2.51582168786513e-06, 2.52383339116591e-06, 2.52572446513000e-06, 2.52611477336018e-06, 2.53588025245222e-06, 2.55538451554571e-06, 2.57194933794835e-06, 2.58932864838357e-06, 2.63472754724603e-06, 2.69989129364490e-06, 2.72499212883304e-06, 2.69538646691103e-06, 2.68336844655437e-06, 2.71359285264339e-06, 2.71729710377144e-06, 2.70983840167105e-06, 2.74125273946879e-06, 2.69582840317639e-06, 2.64576510389247e-06, 3.18392162768077e-06, 4.14825723282775e-06, 4.89917535162476e-06, 5.15898119246572e-06, 4.91249612011302e-06, 4.29894900122463e-06, 3.52546732429460e-06, 2.77596275211067e-06, 2.13345387223285e-06, 1.61116670582828e-06, 1.35825451863525e-06, 1.47579814679943e-06, 1.62076368681596e-06, 1.59566792490295e-06, 1.39247176278445e-06, 1.07009612904108e-06, 7.09271962051616e-07, 4.23044305364328e-07, 3.55271367880050e-15, 4.22082070627994e-07, 4.08105492510913e-07, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15],
[ 1.90000000000000e+01, 1.89999999888123e+01, 1.89999996428678e+01, 1.89999972990389e+01, 1.89999886828232e+01, 1.89999657147492e+01, 1.89999154476190e+01, 1.89998191721488e+01, 1.89996517235669e+01, 1.89993810157277e+01, 1.89989678224950e+01, 1.89983658187540e+01, 1.89975218856836e+01, 1.89963766771198e+01, 1.89948654362350e+01, 1.89929190445910e+01, 1.89904652791315e+01, 1.89874302470208e+01, 1.89837399635507e+01, 1.89793220346772e+01, 1.89741074031216e+01, 1.89680321153221e+01, 1.89610390657796e+01, 1.89530796753880e+01, 1.89441154611107e+01, 1.89341194557848e+01, 1.89230774389141e+01, 1.89109889420908e+01, 1.88978679962633e+01, 1.88837435925694e+01, 1.88686598339891e+01, 1.88526757617280e+01, 1.88358648479935e+01, 1.88183141555819e+01, 1.88001231741916e+01, 1.87814023532841e+01, 1.87622713611631e+01, 1.87428571092336e+01, 1.87232915886262e+01, 1.87037095730577e+01, 1.86842462465695e+01, 1.86650348174010e+01, 1.86462041795993e+01, 1.86278766820941e+01, 1.86101660610354e+01, 1.85931755854936e+01, 1.85769964594807e+01, 1.85617065150639e+01, 1.85473692224694e+01, 1.85340330338794e+01, 1.85217310684024e+01, 1.85104811367157e+01, 1.85002860953331e+01, 1.84911345125041e+01, 1.84830016205175e+01, 1.84758505227706e+01, 1.84696336184384e+01, 1.84642942030200e+01, 1.84597681995133e+01, 1.84559859725403e+01, 1.84528741764652e+01, 1.84503575884676e+01, 1.84483608786737e+01, 1.84468102718044e+01, 1.84456350583477e+01, 1.84447689179152e+01, 1.84441510231141e+01, 1.84437268987888e+01, 1.84434490187063e+01, 1.84432771294611e+01, 1.84431782993492e+01, 1.84431266979721e+01, 1.84431031201569e+01, 1.84430942751880e+01, 1.84430918691310e+01, 1.84430915140017e+01, 1.84430914674162e+01, 1.84430903820301e+01, 1.84430830283702e+01, 1.84430559953942e+01, 1.84429839342966e+01, 1.84428262247205e+01, 1.84425241704551e+01, 1.84419988280879e+01, 1.84411495540330e+01, 1.84398533348722e+01, 1.84379649436228e+01, 1.84353179410585e+01, 1.84317265173013e+01, 1.84269881453190e+01, 1.84208869954496e+01, 1.84131980392473e+01, 1.84036917522645e+01, 1.83921393090622e+01, 1.83783181497069e+01, 1.83620177849714e+01, 1.83430456969770e+01, 1.83212331827045e+01, 1.82964409795684e+01, 1.82685645054784e+01, 1.82375385415377e+01, 1.82033411854308e+01, 1.81659969097334e+01, 1.81255785740322e+01, 1.80822082645822e+01, 1.80360568710003e+01, 1.79873423555051e+01, 1.79363267241679e+01, 1.78833117677622e+01, 1.78286336972565e+01, 1.77726568506336e+01, 1.77157666888106e+01, 1.76583623254311e+01, 1.76008488463517e+01, 1.75436296697625e+01, 1.74870991788446e+01, 1.74316358287847e+01, 1.73775958927176e+01, 1.73253079707673e+01, 1.72750683464387e+01, 1.72271372380150e+01, 1.71817359612185e+01, 1.71390449940711e+01, 1.70992029156749e+01, 1.70623061768831e+01, 1.70284096515254e+01, 1.69975279107649e+01, 1.69696371591052e+01, 1.69446777674789e+01, 1.69225573359181e+01, 1.69031542149777e+01, 1.68863214110828e+01, 1.68718907962790e+01, 1.68596775376902e+01, 1.68494846567089e+01, 1.68411076230984e+01, 1.68343388853681e+01, 1.68289722366385e+01, 1.68248069153417e+01, 1.68216513430300e+01, 1.68193264076729e+01, 1.68176682102937e+01, 1.68165302056315e+01, 1.68157846834467e+01, 1.68153235556589e+01, 1.68150584350543e+01, 1.68149200130079e+01, 1.68148567656448e+01, 1.68148330391834e+01, 1.68148265849924e+01, 1.68148256323667e+01, 1.68148255482680e+01, 1.68148239003949e+01, 1.68148127358055e+01, 1.68147716934974e+01, 1.68146622890298e+01, 1.68144228562596e+01, 1.68139642994188e+01, 1.68131668200341e+01, 1.68118777600104e+01, 1.68099106755984e+01, 1.68070457269464e+01, 1.68030314344637e+01, 1.67975878162381e+01, 1.67904108803446e+01, 1.67811784024175e+01, 1.67695568730434e+01, 1.67552094525495e+01, 1.67378047242444e+01, 1.67170259932446e+01, 1.66925808391979e+01, 1.66642106003392e+01, 1.66316994463488e+01, 1.65948826913571e+01, 1.65536540087125e+01, 1.65079712376363e+01, 1.64578605193072e+01, 1.64034185653268e+01, 1.63448129420551e+01, 1.62822803451105e+01, 1.62161229327522e+01, 1.61467028771872e+01, 1.60744353712507e+01, 1.59997803876985e+01, 1.59232335250228e+01, 1.58453162857031e+01, 1.57665661217125e+01, 1.56875265521700e+01, 1.56087376153023e+01, 1.55307268679216e+01, 1.54540010964169e+01, 1.53790388582911e+01, 1.53062839351551e+01, 1.52361397475798e+01, 1.51689647586623e+01, 1.51050688751269e+01, 1.50447108405840e+01, 1.49880966037830e+01, 1.49353786343441e+01, 1.48866561490520e+01, 1.48419762031930e+01, 1.48013355936266e+01, 1.47646835133022e+01, 1.47319248906272e+01, 1.47029243412152e+01, 1.46775106537365e+01, 1.46554817255319e+01, 1.46366098571057e+01, 1.46206473075780e+01, 1.46073320058495e+01, 1.45963933051208e+01, 1.45875576621693e+01, 1.45805541182888e+01, 1.45751194568986e+01, 1.45710029144007e+01, 1.45679703265634e+01, 1.45658076030062e+01, 1.45643234373789e+01, 1.45633511803601e+01, 1.45627498260855e+01, 1.45624040891914e+01, 1.45622235781813e+01, 1.45621411000240e+01, 1.45621101594279e+01, 1.45621017428175e+01, 1.45621005005428e+01, 1.45621004008970e+01, 1.45620985619264e+01, 1.45620861026244e+01, 1.45620403008285e+01, 1.45619182094112e+01, 1.45616510120553e+01, 1.45611392846026e+01, 1.45602493469908e+01, 1.45588108657509e+01, 1.45566158381920e+01, 1.45534190566057e+01, 1.45489401137793e+01, 1.45428669695012e+01, 1.45348610514758e+01, 1.45245638133886e+01, 1.45116046186462e+01, 1.44956097621364e+01, 1.44762123865994e+01, 1.44530629980133e+01, 1.44258402394898e+01, 1.43942615494946e+01, 1.43580933114957e+01, 1.43171601013915e+01, 1.42713526580668e+01, 1.42206342413228e+01, 1.41650450986791e+01, 1.41047048350224e+01, 1.40398125625148e+01, 1.39706447978323e+01, 1.38975511650328e+01, 1.38209480510547e+01, 1.37413104434629e+01, 1.36591622532823e+01, 1.35750654859861e+01, 1.34896086667788e+01, 1.34033949477163e+01, 1.33170303200928e+01, 1.32311123241535e+01, 1.31462195913843e+01, 1.30629024783136e+01, 1.29816749643194e+01, 1.29030079002769e+01, 1.28273236199526e+01, 1.27549918686739e+01, 1.26863269666220e+01, 1.26215861057692e+01, 1.25609686759157e+01, 1.25046165212512e+01, 1.24526150393753e+01, 1.24049950459064e+01, 1.23617353372176e+01, 1.23227658902211e+01, 1.22879716411280e+01, 1.22571967848645e+01, 1.22302495336930e+01, 1.22069072679883e+01, 1.21869220045177e+01, 1.21700260984838e+01, 1.21559380856266e+01, 1.21443685606059e+01, 1.21350259785841e+01, 1.21276222594110e+01, 1.21218780690955e+01, 1.21175276523356e+01, 1.21143230935675e+01, 1.21120378928685e+01, 1.21104697573176e+01, 1.21094425279287e+01, 1.21088071864338e+01, 1.21084419140750e+01, 1.21082512048974e+01, 1.21081640673656e+01, 1.21081313789182e+01, 1.21081224868536e+01, 1.21081211744030e+01, 1.21081210731742e+01, 1.21081192554293e+01, 1.21081069399367e+01, 1.21080616668238e+01, 1.21079409848714e+01, 1.21076768730595e+01, 1.21071710584876e+01, 1.21062914149865e+01, 1.21048696027995e+01, 1.21027000820632e+01, 1.20995406011340e+01, 1.20951142245615e+01, 1.20891129241740e+01, 1.20812027099669e+01, 1.20710302253820e+01, 1.20582306749280e+01, 1.20424368925852e+01, 1.20232892997278e+01, 1.20004464449951e+01, 1.19735957700455e+01, 1.19424642092060e+01, 1.19068282123257e+01, 1.18665227824501e+01, 1.18214491455533e+01, 1.17715807186690e+01, 1.17169671130550e+01, 1.16577359957906e+01, 1.15940927297878e+01, 1.15263178109208e+01, 1.14547622143374e+01, 1.13798408438285e+01, 1.13020243443259e+01, 1.12218295864805e+01, 1.11398091641022e+01, 1.10565402614521e+01, 1.09726132496365e+01, 1.08886203609611e+01, 1.08051447678382e+01, 1.07227503592497e+01, 1.06419724637754e+01, 1.05633097155713e+01, 1.04872172013096e+01, 1.04141009657765e+01, 1.03443138958198e+01, 1.02781529506894e+01, 1.02158576647143e+01, 1.01576098176736e+01, 1.01035341496400e+01, 1.00536999896938e+01, 1.00081236698499e+01, 9.96677160428475e+00, 9.92956392671699e+00, 9.89637859292207e+00, 9.86705586853666e+00, 9.84140313278651e+00, 9.81919993540153e+00, 9.80020324625613e+00, 9.78415283526450e+00, 9.77077671435573e+00, 9.75979656495775e+00, 9.75093306464282e+00, 9.74391101692545e+00, 9.73846418020215e+00, 9.73433968694430e+00, 9.73130194374228e+00, 9.72913590756040e+00, 9.72764964407277e+00, 9.72667609020461e+00, 9.72607396449804e+00, 9.72572779470434e+00, 9.72554706075631e+00, 9.72546448143899e+00, 9.72543350300785e+00, 9.72542507611608e+00, 9.72542383232424e+00, 9.72542373814807e+00, 9.72542206982665e+00, 9.72541076670228e+00, 9.72536921520203e+00, 9.72525845400911e+00, 9.72501605509207e+00, 9.72455182927336e+00, 9.72374452762800e+00, 9.72243968926360e+00, 9.72044876200013e+00, 9.71754959463354e+00, 9.71348836710649e+00, 9.70798298727747e+00, 9.70072793954161e+00, 9.69140052124302e+00, 9.67966834825534e+00, 9.66519795279824e+00, 9.64766423697281e+00, 9.62676048804711e+00, 9.60220861038286e+00, 9.57376918870499e+00, 9.54125097284688e+00, 9.50451936933653e+00, 9.46350354336794e+00, 9.41820177742948e+00, 9.36868479978780e+00, 9.31509688461616e+00, 9.25765463106299e+00, 9.19664344423956e+00, 9.13241185871053e+00, 9.06536395553558e+00, 8.99595021827688e+00, 8.92465724380924e+00, 8.85199676450154e+00, 8.77849444654026e+00, 8.70467890535862e+00, 8.63107132719281e+00, 8.55817601241593e+00, 8.48647207008244e+00, 8.41640640327698e+00, 8.34838804000498e+00, 8.28278379139616e+00, 8.21991516243788e+00, 8.16005640221593e+00, 8.10343356020962e+00, 8.05022441018583e+00, 8.00055911011976e+00, 7.95452148134832e+00, 7.91215080900457e+00, 7.87344408543414e+00, 7.83835863629132e+00, 7.80681508371675e+00, 7.77870061151454e+00, 7.75387250327322e+00, 7.73216192604092e+00, 7.71337792989237e+00, 7.69731162812388e+00, 7.68374051462161e+00, 7.67243286500298e+00, 7.66315215734764e+00, 7.65566143769535e+00, 7.64972754601848e+00, 7.64512511109933e+00, 7.64164021861692e+00, 7.63907365659503e+00, 7.63724364678006e+00, 7.63598797980340e+00, 7.63516548608473e+00, 7.63465679291758e+00, 7.63436434025353e+00, 7.63421165227006e+00, 7.63414188756404e+00, 7.63411571636487e+00, 7.63410859715948e+00, 7.63410754637950e+00, 7.63410746749712e+00, 7.63410607907769e+00, 7.63409667233117e+00, 7.63406209212732e+00, 7.63396991393398e+00, 7.63376818387744e+00, 7.63338184540136e+00, 7.63270999552558e+00, 7.63162409668577e+00, 7.62996725035367e+00, 7.62755461557617e+00, 7.62417502849826e+00, 7.61959384707536e+00, 7.61355700794375e+00, 7.60579623964488e+00, 7.59603532862215e+00, 7.58399728305716e+00, 7.56941218714526e+00, 7.55202548830211e+00, 7.53160641635747e+00, 7.50795620180656e+00, 7.48091574434456e+00, 7.45037238713278e+00, 7.41626547898196e+00, 7.37859045622305e+00, 7.33740124628062e+00, 7.29281088108657e+00, 7.24499030345050e+00, 7.19416544484683e+00, 7.14061273994824e+00, 7.08465331371633e+00, 7.02664612517993e+00, 6.96698037543675e+00, 6.90606748653525e+00, 6.84433293647314e+00, 6.78220819958846e+00, 6.72012299813038e+00, 6.65849802643113e+00, 6.59773826898257e+00, 6.53822700073624e+00, 6.48032053261689e+00, 6.42434374616889e+00, 6.37058644596763e+00, 6.31930054434105e+00, 6.27069807820178e+00, 6.22495004173174e+00, 6.18218600187265e+00, 6.14249444753313e+00, 6.10592380991803e+00, 6.07248408192862e+00, 6.04214895996288e+00, 6.01485843154613e+00, 5.99052173609960e+00, 5.96902063230825e+00, 5.95021291227699e+00, 5.93393610840378e+00, 5.92001134248016e+00, 5.90824726731889e+00, 5.89844404913677e+00, 5.89039733443147e+00, 5.88390213903228e+00, 5.87875659048988e+00, 5.87476544922763e+00, 5.87174333011226e+00, 5.86951754535848e+00, 5.86793049274119e+00, 5.86684152038597e+00, 5.86612821099172e+00, 5.86568704386552e+00, 5.86543341192244e+00, 5.86530099184092e+00, 5.86524048770383e+00, 5.86521779046333e+00, 5.86521161625888e+00, 5.86521070495906e+00, 5.86521063675553e+00, 5.86520943906722e+00, 5.86520132455188e+00, 5.86517149472268e+00, 5.86509197925446e+00, 5.86491796111223e+00, 5.86458469369453e+00, 5.86400513305227e+00, 5.86306839340970e+00, 5.86163911702883e+00, 5.85955782955354e+00, 5.85664232882974e+00, 5.85269012829316e+00, 5.84748194497860e+00, 5.84078618699782e+00, 5.83236435642352e+00, 5.82197724198561e+00, 5.80939173358455e+00, 5.79438804973168e+00, 5.77676713249738e+00, 5.75635793548278e+00, 5.73302431175136e+00, 5.70667120317293e+00, 5.67724984214403e+00, 5.64476170215386e+00, 5.60926097520355e+00, 5.57085541081273e+00, 5.52970542163949e+00, 5.48602144223408e+00, 5.44005961686449e+00, 5.39211598509990e+00, 5.34251942352430e+00, 5.29162368018805e+00, 5.23979889533236e+00, 5.18742302781338e+00, 5.13487359445836e+00, 5.08252007771293e+00, 5.03071727108959e+00, 4.97979972490290e+00, 4.93007734385353e+00, 4.88183209063929e+00, 4.83531567909420e+00, 4.79074810238662e+00, 4.74831683495220e+00, 4.70817656375682e+00, 4.67044933515784e+00, 4.63522503837338e+00, 4.60256217806379e+00, 4.57248891261058e+00, 4.54500435004173e+00, 4.52008010090844e+00, 4.49766208850294e+00, 4.47767261359104e+00, 4.46001266501922e+00, 4.44456446037556e+00, 4.43119419309989e+00, 4.41975495446203e+00, 4.41008979088181e+00, 4.40203484931861e+00, 4.39542255613762e+00, 4.39008476828776e+00, 4.38585583023720e+00, 4.38257546641537e+00, 4.38009143743481e+00, 4.37826188959044e+00, 4.37695733141751e+00, 4.37606217860569e+00, 4.37547581926552e+00, 4.37511316512211e+00, 4.37490467013254e+00, 4.37479581554532e+00, 4.37474607866268e+00, 4.37472742058869e+00, 4.37472234513524e+00, 4.37472159600876e+00, 4.37472153992163e+00, 4.37472055472295e+00, 4.37471387985580e+00, 4.37468934232119e+00, 4.37462393413306e+00, 4.37448078925433e+00, 4.37420664716139e+00, 4.37372990385298e+00, 4.37295933866358e+00, 4.37178359079328e+00, 4.37007144272382e+00, 4.36767294821523e+00, 4.36442141990389e+00, 4.36013626531160e+00, 4.35462663028540e+00, 4.34769577595501e+00, 4.33914608028273e+00, 4.32878451996606e+00, 4.31642845530893e+00, 4.30191151271177e+00, 4.28508933987558e+00, 4.26584500066912e+00, 4.24409378210023e+00, 4.21978720589438e+00, 4.19291607107723e+00, 4.16351239916536e+00, 4.13165020605884e+00, 4.09744507958137e+00, 4.06105259390210e+00, 4.02266563778321e+00, 3.98251077031154e+00, 3.94084374490645e+00, 3.89794436091008e+00, 3.85411081375311e+00, 3.80965372131925e+00, 3.76489000674988e+00, 3.72013681656425e+00, 3.67570564680584e+00, 3.63189683783008e+00, 3.58899457953284e+00, 3.54726254335471e+00, 3.50694022645662e+00, 3.46824005921434e+00, 3.43134529240737e+00, 3.39640864808105e+00, 3.36355169055085e+00, 3.33286485316337e+00, 3.30440804306099e+00, 3.27821174022090e+00, 3.25427850757312e+00, 3.23258483461323e+00, 3.21308324589246e+00, 3.19570461632257e+00, 3.18036064575922e+00, 3.16694645450301e+00, 3.15534326823150e+00, 3.14542116490856e+00, 3.13704185726847e+00, 3.13006148276915e+00, 3.12433336898606e+00, 3.11971073704997e+00, 3.11604929984033e+00, 3.11320970621813e+00, 3.11105977856694e+00, 3.10947648913652e+00, 3.10834762177352e+00, 3.10757306993815e+00, 3.10706572951011e+00, 3.10675195554043e+00, 3.10657156529605e+00, 3.10647738493454e+00, 3.10643435305888e+00, 3.10641821030204e+00, 3.10641381907939e+00, 3.10641317094408e+00, 3.10641312228943e+00, 3.10641226592674e+00, 3.10640646394381e+00, 3.10638513524511e+00, 3.10632828079806e+00, 3.10620385639818e+00, 3.10596556942771e+00, 3.10555118839594e+00, 3.10488144390215e+00, 3.10385958786858e+00, 3.10237166410273e+00, 3.10028752699145e+00, 3.09746262591518e+00, 3.09374055043702e+00, 3.08895630536664e+00, 3.08294025572240e+00, 3.07552265020555e+00, 3.06653859938939e+00, 3.05583335326338e+00, 3.04326769433488e+00, 3.02872323971407e+00, 3.01210743105565e+00, 2.99335798722724e+00, 2.97244660291741e+00, 2.94938169809766e+00, 2.92421005832080e+00, 2.89701725314942e+00, 2.86792677726186e+00, 2.83709792258335e+00, 2.80472245583865e+00, 2.77102023933989e+00, 2.73623398858764e+00, 2.70062340372582e+00, 2.66445893933412e+00, 2.62801548616921e+00, 2.59156622876676e+00, 2.55537691567960e+00, 2.51970073770207e+00, 2.48477395820513e+00, 2.45081238389866e+00, 2.41800870919646e+00, 2.38653071751522e+00, 2.35652028179779e+00, 2.32809307642623e+00, 2.30133889415736e+00, 2.27632245418606e+00, 2.25308458937359e+00, 2.23164370993617e+00, 2.21199745510850e+00, 2.19412446118298e+00, 2.17798619185794e+00, 2.16352879338401e+00, 2.15068495139047e+00, 2.13937573773070e+00, 2.12951244381074e+00, 2.12099840155932e+00, 2.11373079461602e+00, 2.10760246079939e+00, 2.10250368297026e+00, 2.09832395964743e+00, 2.09495373987672e+00, 2.09228609967237e+00, 2.09021833063261e+00, 2.08865340583881e+00, 2.08750128455495e+00, 2.08668001607834e+00, 2.08611660468765e+00, 2.08574760208449e+00, 2.08551940087655e+00, 2.08538821210547e+00, 2.08531972098086e+00, 2.08528842710272e+00, 2.08527668773434e+00, 2.08527349434581e+00, 2.08527302300851e+00, 2.08527298747516e+00, 2.08527236005106e+00, 2.08526810916304e+00, 2.08525248248743e+00, 2.08521082780291e+00, 2.08511966904317e+00, 2.08494509485516e+00, 2.08464152684794e+00, 2.08415092849632e+00, 2.08340250830183e+00, 2.08231296179196e+00, 2.08078728557897e+00, 2.07872018228602e+00, 2.07599805708477e+00, 2.07250158456843e+00, 2.06810879886116e+00, 2.06269863099547e+00, 2.05615478711343e+00, 2.04836983109247e+00, 2.03924930846098e+00, 2.02871572800642e+00, 2.01671220635479e+00, 2.00320558168642e+00, 1.98818881747544e+00, 1.97168254628080e+00, 1.95373564621671e+00, 1.93442479618901e+00, 1.91385301618726e+00, 1.89214726062823e+00, 1.86945519018835e+00, 1.84594129519749e+00, 1.82178257693928e+00, 1.79716400920568e+00, 1.77227400032750e+00, 1.74730005693172e+00, 1.72242481802073e+00, 1.69782258611527e+00, 1.67365643623836e+00, 1.65007593836742e+00, 1.62721548876551e+00, 1.60519321320297e+00, 1.58411038197222e+00, 1.56405126293378e+00, 1.54508333370018e+00, 1.52725777586489e+00, 1.51061018099972e+00, 1.49516140806762e+00, 1.48091854325710e+00, 1.46787592472300e+00, 1.45601620538546e+00, 1.44531143620974e+00, 1.43572415995296e+00, 1.42720851110001e+00, 1.41971132162248e+00, 1.41317323435196e+00, 1.40752982628016e+00, 1.40271274313078e+00, 1.39865084428373e+00, 1.39527135380719e+00, 1.39250100926434e+00, 1.39026719546004e+00, 1.38849904578309e+00, 1.38712848971164e+00, 1.38609122181590e+00, 1.38532756560554e+00, 1.38478320514683e+00, 1.38440975871632e+00, 1.38416517192325e+00, 1.38401391262235e+00, 1.38392695629971e+00, 1.38388155806189e+00, 1.38386081541455e+00, 1.38385303416092e+00, 1.38385091747406e+00, 1.38385060505562e+00, 1.38385058137473e+00, 1.38385016153332e+00, 1.38384731704771e+00, 1.38383686044376e+00, 1.38380898717677e+00, 1.38374798822853e+00, 1.38363117183369e+00, 1.38342803930958e+00, 1.38309975580615e+00, 1.38259895186428e+00, 1.38186988561008e+00, 1.38084898774937e+00, 1.37946580175976e+00, 1.37764431938895e+00, 1.37530469654527e+00, 1.37236531703210e+00, 1.36874515189668e+00, 1.36436634150826e+00, 1.35915690742768e+00, 1.35305348366464e+00, 1.34600394424061e+00, 1.33796979821016e+00, 1.32892822615266e+00, 1.31887364457314e+00, 1.30781870654442e+00, 1.29579467699244e+00, 1.28285115684813e+00, 1.26905516856814e+00, 1.25448965254949e+00, 1.23925145615060e+00, 1.22344892149335e+00, 1.20719919318085e+00, 1.19062537207155e+00, 1.17385363712252e+00, 1.15701044585342e+00, 1.14021990752049e+00, 1.12360140400270e+00, 1.10726751370477e+00, 1.09132227488820e+00, 1.07585980753040e+00, 1.06096329736471e+00, 1.04670433218180e+00, 1.03314256875772e+00, 1.02032569902088e+00, 1.00828967654096e+00, 9.97059159462125e-01, 9.86648123889967e-01, 9.77060602551898e-01, 9.68291507064340e-01, 9.60327497841455e-01, 9.53147872828513e-01, 9.46725453992393e-01, 9.41027458031417e-01, 9.36016344384161e-01, 9.31650638823356e-01, 9.27885734432763e-01, 9.24674673500906e-01, 9.21968913917482e-01, 9.19719082249536e-01, 9.17875713125973e-01, 9.16389971253281e-01, 9.15214348740996e-01, 9.14303326857002e-01, 9.13613988267450e-01, 9.13106563601989e-01, 9.12744895106282e-01, 9.12496800383188e-01, 9.12334320847041e-01, 9.12233842460150e-01, 9.12176080405356e-01, 9.12145924289174e-01, 9.12132145912018e-01, 9.12126977199648e-01, 9.12125571187634e-01, 9.12125363663318e-01, 9.12125347833452e-01, 9.12125065866920e-01, 9.12123155503924e-01, 9.12116132833894e-01, 9.12097413176905e-01, 9.12056446695221e-01, 9.11977994928080e-01, 9.11841578996391e-01, 9.11621127745260e-01, 9.11284851123128e-01, 9.10795360340998e-01, 9.10110051431683e-01, 9.09181762334878e-01, 9.07959705227626e-01, 9.06390665310269e-01, 9.04420444694807e-01, 9.01995515788979e-01, 8.99064833345481e-01, 8.95581739226898e-01, 8.91505880300056e-01, 8.86805049278557e-01, 8.81456852361787e-01, 8.75450107572004e-01, 8.68785884777719e-01, 8.61478112940009e-01, 8.53553701833587e-01, 8.45052153282971e-01, 8.36024668927703e-01, 8.26532795093929e-01, 8.16646677384789e-01, 8.06443024755730e-01, 7.96002901929128e-01, 7.85409477506818e-01, 7.74745851727356e-01, 7.64093072755664e-01, 7.53528425704445e-01, 7.43124047798801e-01, 7.32945890615923e-01, 7.23053020559075e-01, 7.13497225163906e-01, 7.04322877478340e-01, 6.95567003906294e-01, 6.87259501424201e-01, 6.79423455939499e-01, 6.72075522462066e-01, 6.65226337642654e-01, 6.58880944577486e-01, 6.53039217686526e-01, 6.47696281592484e-01, 6.42842922287276e-01, 6.38465991716405e-01, 6.34548808554605e-01, 6.31071558684749e-01, 6.28011698960798e-01, 6.25344367395790e-01, 6.23042802071506e-01, 6.21078769882178e-01, 6.19423004752459e-01, 6.18045653267671e-01, 6.16916723797196e-01, 6.16006533281951e-01, 6.15286144020047e-01, 6.14727781165954e-01, 6.14305220408866e-01, 6.13994134559661e-01, 6.13772387670316e-01, 6.13620265913095e-01, 6.13520635782219e-01, 6.13459022209994e-01, 6.13423601812972e-01, 6.13405109547358e-01, 6.13396660363790e-01, 6.13393490797140e-01, 6.13392628599463e-01, 6.13392501340943e-01, 6.13392491565211e-01, 6.13392316538871e-01, 6.13391130710486e-01, 6.13386771492749e-01, 6.13375151517349e-01, 6.13349721985699e-01, 6.13301023307240e-01, 6.13216341893406e-01, 6.13079490308035e-01, 6.12870725985907e-01, 6.12566821250269e-01, 6.12141294166662e-01, 6.11564805626333e-01, 6.10805722758517e-01, 6.09830842222433e-01, 6.08606259183304e-01, 6.07098359102284e-01, 6.05274900385602e-01, 6.03106147207328e-01, 6.00566004396214e-01, 5.97633101200753e-01, 5.94291768986295e-01, 5.90532860190269e-01, 5.86354362456344e-01, 5.81761772527180e-01, 5.76768208363802e-01, 5.71394253751684e-01, 5.65667545730125e-01, 5.59622129899726e-01, 5.53297620632225e-01, 5.46738211508845e-01, 5.39991585600273e-01, 5.33107775667794e-01, 5.26138021608439e-01, 5.19133667295328e-01, 5.12145132212272e-01, 5.05220985684161e-01, 4.98407143643004e-01, 4.91746200179186e-01, 4.85276898938451e-01, 4.79033743000944e-01, 4.73046736439347e-01, 4.67341246472337e-01, 4.61937972118005e-01, 4.56853003536967e-01, 4.52097955771386e-01, 4.47680161179630e-01, 4.43602906313277e-01, 4.39865701019862e-01, 4.36464569907397e-01, 4.33392358720791e-01, 4.30639050439229e-01, 4.28192087840639e-01, 4.26036700780848e-01, 4.24156237438094e-01, 4.22532499260118e-01, 4.21146079341867e-01, 4.19976703509622e-01, 4.19003572570933e-01, 4.18205703109418e-01, 4.17562262977381e-01, 4.17052896394175e-01, 4.16658032428582e-01, 4.16359169756095e-01, 4.16139130051750e-01, 4.15982272295221e-01, 4.15874660682597e-01, 4.15804179773051e-01, 4.15760591918175e-01, 4.15735533853443e-01, 4.15722451466266e-01, 4.15716474056815e-01, 4.15714231730842e-01, 4.15713621764439e-01, 4.15713531734703e-01, 4.15713524766030e-01, 4.15713399310619e-01, 4.15712549332259e-01, 4.15709424729378e-01, 4.15701095742676e-01, 4.15682868250151e-01, 4.15647961537051e-01, 4.15587261961043e-01, 4.15489164680508e-01, 4.15339514130544e-01, 4.15121652113232e-01, 4.14816580078107e-01, 4.14403239260885e-01, 4.13858908701556e-01, 4.13159716715128e-01, 4.12281256174136e-01, 4.11199288141486e-01, 4.09890512272472e-01, 4.08333376433841e-01, 4.06508892735754e-01, 4.04401423271921e-01, 4.01999396937835e-01, 3.99295919265743e-01, 3.96289240597202e-01, 3.92983054151756e-01, 3.89386604358346e-01, 3.85514596583935e-01, 3.81386911232916e-01, 3.77028137035506e-01, 3.72466949078103e-01, 3.67735365738948e-01, 3.62867924394614e-01, 3.57900818112061e-01, 3.52871034482583e-01, 3.47815533620594e-01, 3.42770495801709e-01, 3.37770661128715e-01, 3.32848774954083e-01, 3.28035144465755e-01, 3.23357304604336e-01, 3.18839785826874e-01, 3.14503972398970e-01, 3.10368037862155e-01, 3.06446943869254e-01, 3.02752489359808e-01, 2.99293398663181e-01, 2.96075439182012e-01, 2.93101561495039e-01, 2.90372056778665e-01, 2.87884728218295e-01, 2.85635074477739e-01, 2.83616484289563e-01, 2.81820441834636e-01, 2.80236742830610e-01, 2.78853721193755e-01, 2.77658485826571e-01, 2.76637166563997e-01, 2.75775167634396e-01, 2.75057426207631e-01, 2.74468672766844e-01, 2.73993689211750e-01, 2.73617559844290e-01, 2.73325909767944e-01, 2.73105124816016e-01, 2.72942546969759e-01, 2.72826639379968e-01, 2.72747115591056e-01, 2.72695028387570e-01, 2.72662814814517e-01, 2.72644295313704e-01, 2.72634626495670e-01, 2.72630208739777e-01, 2.72628551488143e-01, 2.72628100675625e-01, 2.72628034136645e-01, 2.72628028944510e-01, 2.72627934932050e-01, 2.72627297984080e-01, 2.72624956499193e-01, 2.72618714984780e-01, 2.72605055708269e-01, 2.72578897038944e-01, 2.72533408392544e-01, 2.72459890667974e-01, 2.72347729862702e-01, 2.72184430106343e-01, 2.71955730574309e-01, 2.71645808544211e-01, 2.71237568161428e-01, 2.70713011251175e-01, 2.70053682787061e-01, 2.69241179521476e-01, 2.68257706028578e-01, 2.67086658336659e-01, 2.65713211844801e-01, 2.64124887784028e-01, 2.62312071527980e-01, 2.60268456917476e-01, 2.57991393606797e-01, 2.55482119199300e-01, 2.52745864303069e-01, 2.49791826044298e-01, 2.46633013305559e-01, 2.43285974212803e-01, 2.39770422451297e-01, 2.36108783291872e-01, 2.32325682472705e-01, 2.28447401320064e-01, 2.24501319993593e-01, 2.20515367987907e-01, 2.16517497572514e-01, 2.12535192245014e-01, 2.08595018915304e-01, 2.04722229686936e-01, 2.00940416829056e-01, 1.97271222794953e-01, 1.93734105804172e-01, 1.90346160426042e-01, 1.87121991661265e-01, 1.84073640156676e-01, 1.81210555397072e-01, 1.78539613038630e-01, 1.76065172026576e-01, 1.73789166830467e-01, 1.71711230053692e-01, 1.69828840818223e-01, 1.68137494666777e-01, 1.66630891171799e-01, 1.65301135946940e-01, 1.64138954199764e-01, 1.63133913318969e-01, 1.62274652152426e-01, 1.61549114630173e-01, 1.60944785182654e-01, 1.60448923040626e-01, 1.60048792056450e-01, 1.59731882170473e-01, 1.59486118210939e-01, 1.59300051364387e-01, 1.59163028526592e-01, 1.59065334832608e-01, 1.58998305053263e-01, 1.58954400199651e-01, 1.58927246608730e-01, 1.58911635930469e-01, 1.58903485732164e-01, 1.58899761835400e-01, 1.58898364872318e-01, 1.58897984864510e-01, 1.58897928776113e-01, 1.58897924362211e-01, 1.58897843963800e-01, 1.58897299253702e-01, 1.58895296843088e-01, 1.58889959165290e-01, 1.58878277858499e-01, 1.58855907030045e-01, 1.58817004819774e-01, 1.58754130784825e-01, 1.58658205560748e-01, 1.58518538010274e-01, 1.58322923561570e-01, 1.58057815619898e-01, 1.57708569753382e-01, 1.57259757801711e-01, 1.56695546139520e-01, 1.56000129166937e-01, 1.55158205808270e-01, 1.54155483630277e-01, 1.52979192367074e-01, 1.51618586474552e-01, 1.50065415089589e-01, 1.48314337660437e-01, 1.46363264709930e-01, 1.44213605675493e-01, 1.41870409500953e-01, 1.39342388369776e-01, 1.36641820433461e-01, 1.33784333171242e-01, 1.30788574854904e-01, 1.27675787021378e-01, 1.24469295628117e-01, 1.21193942390931e-01, 1.17875480390964e-01, 1.14539959311220e-01, 1.11213125413773e-01, 1.07919859659770e-01, 1.04683674303967e-01, 1.01526284072143e-01, 9.84672629443892e-02, 9.55237922497290e-02, 9.27105003381321e-02, 9.00393893004079e-02, 8.75198402906499e-02, 8.51586859478211e-02, 8.29603369156828e-02, 8.09269487727988e-02, 7.90586162850602e-02, 7.73535832122469e-02, 7.58084577478222e-02, 7.44184258806797e-02, 7.31774573003747e-02, 7.20785005235847e-02, 7.11136657109996e-02, 7.02743953864058e-02, 6.95516238913108e-02, 6.89359274896901e-02, 6.84176667372398e-02, 6.79871229400817e-02, 6.76346296651076e-02, 6.73506998508699e-02, 6.71261479935019e-02, 6.69522059734433e-02, 6.68206305495005e-02, 6.67237997918946e-02, 6.66547949996392e-02, 6.66074651921278e-02, 6.65764709451828e-02, 6.65573048817585e-02, 6.65462872202482e-02, 6.65405352776567e-02, 6.65379072344443e-02, 6.65369213673976e-02, 6.65366531783889e-02, 6.65366135954848e-02, 6.65366104615011e-02, 6.65365528736679e-02, 6.65361626784495e-02, 6.65347283315292e-02, 6.65309049225691e-02, 6.65225378404557e-02, 6.65065152320921e-02, 6.64786560043744e-02, 6.64336393947771e-02, 6.63649818083165e-02, 6.62650659373440e-02, 6.61252260461289e-02, 6.59358924694935e-02, 6.56867960966601e-02, 6.53672313778522e-02, 6.49663736130179e-02, 6.44736428018255e-02, 6.38791026354681e-02, 6.31738795700230e-02, 6.23505835282252e-02, 6.14037091611124e-02, 6.03299948599272e-02, 5.91287167372663e-02, 5.78018964265006e-02, 5.63344331787761e-02, 5.47758369868028e-02, 5.31146673327285e-02, 5.13637539079248e-02, 4.95380169703239e-02, 4.76540149039906e-02, 4.57294250361429e-02, 4.37824870118249e-02, 4.18314402626123e-02, 3.98939866422836e-02, 3.79868062114035e-02, 3.61251488259278e-02, 3.43225172754238e-02, 3.25904500366434e-02, 3.09384041034626e-02, 2.93737315821900e-02, 2.79017383723179e-02, 2.65258096148813e-02, 2.52475847624005e-02, 2.40671649630819e-02, 2.29833366584202e-02, 2.19937974806754e-02, 2.10953732998220e-02, 2.02842182403167e-02, 1.95559923691562e-02, 1.89060143377695e-02, 1.83293884161380e-02, 1.78211070369407e-02, 1.73761311741529e-02, 1.69894516537568e-02, 1.66561348935517e-02, 1.63713566591319e-02, 1.61304272665035e-02, 1.59288113149962e-02, 1.57621445460485e-02, 1.56262498367745e-02, 1.55171536921408e-02, 1.54311039339916e-02, 1.53645886387395e-02, 1.53143557862201e-02, 1.52774325891372e-02, 1.52511431080207e-02, 1.52331225467748e-02, 1.52213265821750e-02, 1.52140342063980e-02, 1.52098428392407e-02, 1.52076548651081e-02, 1.52066552295302e-02, 1.52062802426904e-02, 1.52061782382162e-02, 1.52061631826044e-02, 1.52061619763657e-02, 1.52061397327117e-02, 1.52059890292584e-02, 1.52054350364921e-02, 1.52039583647436e-02, 1.52007270577776e-02, 1.51945400710467e-02, 1.51837850814154e-02, 1.51664135186671e-02, 1.51399357260367e-02, 1.51014390712092e-02, 1.50476315429344e-02, 1.49749127787118e-02, 1.48794734815354e-02, 1.47574227382243e-02, 1.46049408427764e-02, 1.44184529263676e-02, 1.41948161634673e-02, 1.39315108137374e-02, 1.36268232034098e-02, 1.32800073176078e-02, 1.28914113264891e-02, 1.24625563805694e-02, 1.19961575168652e-02, 1.14960804446937e-02, 1.09672330280581e-02, 1.04153959341815e-02, 9.84700250316450e-03, 9.26888268555714e-03, 8.68798923238501e-03, 8.11112573533990e-03, 7.54469541076475e-03, 6.99448682788528e-03, 6.46550852003358e-03, 5.96187922338487e-03, 5.48677508549744e-03, 5.04243026572126e-03, 4.63018344717270e-03, 4.25056021530423e-03, 3.90338011104419e-03, 3.58787730856491e-03, 3.30282502200063e-03, 3.04665556206629e-03, 2.81757008926675e-03, 2.61363424194982e-03, 2.43285773242974e-03, 2.27325756859997e-03, 2.13290571957359e-03, 2.00996281127175e-03, 1.90269986250158e-03, 1.80951022503351e-03, 1.72891384862339e-03, 1.65955582282962e-03, 1.60020090975266e-03, 1.54972551710486e-03, 1.50710829806400e-03, 1.47142031933356e-03, 1.44181551810263e-03, 1.41752197411549e-03, 1.39783435109564e-03, 1.38210770983349e-03, 1.36975275932372e-03, 1.36023249145723e-03, 1.35306003920732e-03, 1.34779751064684e-03, 1.34405548483585e-03, 1.34149281542761e-03, 1.33981637575076e-03, 1.33878039745383e-03, 1.33818510101591e-03, 1.33787438620558e-03, 1.33773243712909e-03, 1.33767919021466e-03, 1.33766470605678e-03, 1.33766256823264e-03, 1.33766239533983e-03, 1.33765918697293e-03, 1.33763744996429e-03, 1.33755754495368e-03, 1.33734456694512e-03, 1.33687856830278e-03, 1.33598650254996e-03, 1.33443637313135e-03, 1.33193411848609e-03, 1.32812380897107e-03, 1.32259175022753e-03, 1.31487505300854e-03, 1.30447511149612e-03, 1.29087620195989e-03, 1.27356905452203e-03, 1.25207876342159e-03, 1.22599581266059e-03, 1.19500836312270e-03, 1.15893336329620e-03, 1.11774362083850e-03, 1.07158782331851e-03, 1.02080072140964e-03, 9.65901336357433e-04, 9.07578104946793e-04, 8.46661224841228e-04, 7.84083929730173e-04, 7.20835779166151e-04, 6.57912058223076e-04, 5.96263863187119e-04, 5.36753307450510e-04, 4.80117540128333e-04, 4.26944066749257e-04, 3.77658416013560e-04, 3.32523758054037e-04, 2.91650870341442e-04, 2.55016019146126e-04, 2.22566942781555e-04, 1.93945361573356e-04, 1.68898826505171e-04, 1.47127944281067e-04, 1.28296128739436e-04, 1.12038889344745e-04, 9.80673486346006e-05, 8.60756977871802e-05, 7.56917701920656e-05, 6.67947424055626e-05, 5.91832174485774e-05, 5.25173370924392e-05, 4.66185211465131e-05, 4.13458333845540e-05, 3.65952474865594e-05, 3.22924611104725e-05, 2.83843247146365e-05, 2.48330707286470e-05, 2.16303091491454e-05, 1.87442413768224e-05, 1.61071171476124e-05, 1.37073074064631e-05, 1.15395737800839e-05, 9.59775347908476e-06, 7.87408298918990e-06, 6.35956480365962e-06, 5.04454271776922e-06, 3.91870226934543e-06, 2.97071446908603e-06, 2.18796197160032e-06, 1.55647902388978e-06, 1.06107077613160e-06, 6.85548776857559e-07, 4.13032488211229e-07, -3.55271367880050e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -3.55271367880050e-15],
[ 2.00000000000000e+01, 1.99999999879758e+01, 1.99999996161623e+01, 1.99999970970675e+01, 1.99999878365514e+01, 1.99999631509773e+01, 1.99999091249780e+01, 1.99998056502034e+01, 1.99996256800282e+01, 1.99993347287495e+01, 1.99988906364755e+01, 1.99982436129588e+01, 1.99973365653193e+01, 1.99961057062202e+01, 1.99944814308956e+01, 1.99923894437531e+01, 1.99897521083193e+01, 1.99864899882605e+01, 1.99825235422155e+01, 1.99777749312907e+01, 1.99721698952736e+01, 1.99656396518743e+01, 1.99581227724875e+01, 1.99495669879865e+01, 1.99399308788042e+01, 1.99291854049763e+01, 1.99173152339234e+01, 1.99043198265878e+01, 1.98902142462591e+01, 1.98750296591525e+01, 1.98588135017110e+01, 1.98416292967647e+01, 1.98235561091068e+01, 1.98046876406073e+01, 1.97851309754078e+01, 1.97650049966262e+01, 1.97444385068265e+01, 1.97235680947127e+01, 1.97025357994993e+01, 1.96814866316778e+01, 1.96605660140282e+01, 1.96399172094440e+01, 1.96196788023571e+01, 1.95999822983216e+01, 1.95809499018614e+01, 1.95626925263192e+01, 1.95453080815630e+01, 1.95288800764351e+01, 1.95134765631792e+01, 1.94991494411464e+01, 1.94859341271777e+01, 1.94738495904679e+01, 1.94628987406339e+01, 1.94530691492936e+01, 1.94443340778380e+01, 1.94366537773175e+01, 1.94299770205561e+01, 1.94242428218138e+01, 1.94193822956027e+01, 1.94153206037012e+01, 1.94119789380516e+01, 1.94092764871369e+01, 1.94071323346245e+01, 1.94054672415604e+01, 1.94042052671548e+01, 1.94032751881510e+01, 1.94026116828125e+01, 1.94021562525301e+01, 1.94018578617766e+01, 1.94016732853873e+01, 1.94015671606991e+01, 1.94015117506844e+01, 1.94014864326231e+01, 1.94014769348199e+01, 1.94014743511755e+01, 1.94014739698345e+01, 1.94014739198106e+01, 1.94014727543139e+01, 1.94014648578916e+01, 1.94014358296555e+01, 1.94013584498647e+01, 1.94011891001433e+01, 1.94008647523982e+01, 1.94003006377008e+01, 1.93993886871031e+01, 1.93979968144658e+01, 1.93959690874977e+01, 1.93931268080655e+01, 1.93892704971952e+01, 1.93841827548576e+01, 1.93776319403853e+01, 1.93693765968657e+01, 1.93591705225832e+01, 1.93467683747481e+01, 1.93319316752813e+01, 1.93144350749817e+01, 1.92940727205371e+01, 1.92706645581482e+01, 1.92440623979834e+01, 1.92141555557746e+01, 1.91808758828709e+01, 1.91442019959284e+01, 1.91041625245768e+01, 1.90608382122984e+01, 1.90143627342577e+01, 1.89649221366009e+01, 1.89127528538216e+01, 1.88581383213433e+01, 1.88014042650257e+01, 1.87429128123164e+01, 1.86830556254019e+01, 1.86222462998052e+01, 1.85609122987007e+01, 1.84994867020556e+01, 1.84384000409821e+01, 1.83780724637631e+01, 1.83189064446592e+01, 1.82612802043130e+01, 1.82055419658402e+01, 1.81520051274629e+01, 1.81009443937583e+01, 1.80525928750723e+01, 1.80071401390970e+01, 1.79647311797992e+01, 1.79254662559378e+01, 1.78894015430433e+01, 1.78565505375627e+01, 1.78268861485673e+01, 1.78003434098559e+01, 1.77768227426193e+01, 1.77561936955233e+01, 1.77382990848675e+01, 1.77229594524225e+01, 1.77099777528865e+01, 1.76991441770791e+01, 1.76902410115873e+01, 1.76830474312547e+01, 1.76773441183452e+01, 1.76729176020806e+01, 1.76695642150997e+01, 1.76670935696389e+01, 1.76653314661001e+01, 1.76641221601436e+01, 1.76633299312628e+01, 1.76628399154677e+01, 1.76625581865177e+01, 1.76624110932415e+01, 1.76623438838814e+01, 1.76623186711315e+01, 1.76623118126335e+01, 1.76623108003329e+01, 1.76623107109660e+01, 1.76623089598659e+01, 1.76622970958987e+01, 1.76622534826032e+01, 1.76621372248246e+01, 1.76618827937369e+01, 1.76613955132710e+01, 1.76605480830948e+01, 1.76591782888641e+01, 1.76570880223914e+01, 1.76540437026616e+01, 1.76497781531275e+01, 1.76439939513605e+01, 1.76363682239235e+01, 1.76265588126418e+01, 1.76142116890024e+01, 1.75989694425096e+01, 1.75804806182824e+01, 1.75584096313873e+01, 1.75324469432262e+01, 1.75023191519512e+01, 1.74677986277483e+01, 1.74287123181771e+01, 1.73849493613788e+01, 1.73364671777669e+01, 1.72832957642852e+01, 1.72255399880830e+01, 1.71633797649828e+01, 1.70970681066334e+01, 1.70269271212298e+01, 1.69533421475876e+01, 1.68767542828961e+01, 1.67976516238304e+01, 1.67165595747601e+01, 1.66340305847469e+01, 1.65506336594231e+01, 1.64669439599270e+01, 1.63835327556072e+01, 1.63009579471142e+01, 1.62197553276535e+01, 1.61404307065811e+01, 1.60634529830711e+01, 1.59892482283185e+01, 1.59181948114401e+01, 1.58506195852108e+01, 1.57867951314156e+01, 1.57269380508456e+01, 1.56712082693567e+01, 1.56197093189823e+01, 1.55724895421133e+01, 1.55295441574768e+01, 1.54908181190894e+01, 1.54562096932658e+01, 1.54255746736325e+01, 1.53987311493129e+01, 1.53754647364061e+01, 1.53555341771901e+01, 1.53386772049731e+01, 1.53246165654103e+01, 1.53130660778769e+01, 1.53037366139257e+01, 1.52963418649027e+01, 1.52906037684413e+01, 1.52862574647855e+01, 1.52830556594646e+01, 1.52807722792989e+01, 1.52792053242257e+01, 1.52781788377927e+01, 1.52775439438035e+01, 1.52771789245706e+01, 1.52769883463452e+01, 1.52769012683720e+01, 1.52768686022226e+01, 1.52768597162194e+01, 1.52768584046633e+01, 1.52768582994603e+01, 1.52768563579307e+01, 1.52768432037767e+01, 1.52767948476317e+01, 1.52766659472634e+01, 1.52763838487135e+01, 1.52758435838019e+01, 1.52749040196558e+01, 1.52733853296786e+01, 1.52710679251258e+01, 1.52676929520286e+01, 1.52629644192389e+01, 1.52565529794995e+01, 1.52481013365293e+01, 1.52372311974010e+01, 1.52235516318745e+01, 1.52066686405021e+01, 1.51861956737533e+01, 1.51617647884469e+01, 1.51330380793372e+01, 1.50997189869592e+01, 1.50615630617561e+01, 1.50183877623065e+01, 1.49700808841394e+01, 1.49166072555795e+01, 1.48580133970472e+01, 1.47944299174560e+01, 1.47260715119270e+01, 1.46532345244649e+01, 1.45762921428165e+01, 1.44956873956757e+01, 1.44119242196712e+01, 1.43255569495271e+01, 1.42371786529607e+01, 1.41474087754904e+01, 1.40568805733968e+01, 1.39662287923682e+01, 1.38760779960798e+01, 1.37870318695794e+01, 1.36996637279069e+01, 1.36145083640087e+01, 1.35320552839791e+01, 1.34527433105026e+01, 1.33769564905759e+01, 1.33050212198033e+01, 1.32372044881699e+01, 1.31737131552672e+01, 1.31146941709755e+01, 1.30602356665169e+01, 1.30103688481747e+01, 1.29650706308087e+01, 1.29242669504666e+01, 1.28878366951516e+01, 1.28556161904942e+01, 1.28274041729278e+01, 1.28029671771298e+01, 1.27820452570969e+01, 1.27643579515130e+01, 1.27496103944506e+01, 1.27374994625721e+01, 1.27277198407330e+01, 1.27199698802762e+01, 1.27139571194456e+01, 1.27094033343328e+01, 1.27060489924977e+01, 1.27036569905221e+01, 1.27020155715330e+01, 1.27009403390194e+01, 1.27002753084709e+01, 1.26998929675023e+01, 1.26996933468701e+01, 1.26996021375979e+01, 1.26995679216962e+01, 1.26995586141267e+01, 1.26995572403484e+01, 1.26995571343894e+01, 1.26995552317057e+01, 1.26995423407398e+01, 1.26994949521258e+01, 1.26993686310195e+01, 1.26990921780345e+01, 1.26985627285532e+01, 1.26976419834757e+01, 1.26961537394911e+01, 1.26938828581000e+01, 1.26905757795389e+01, 1.26859426498903e+01, 1.26796610864422e+01, 1.26713815574211e+01, 1.26607342976773e+01, 1.26473376225932e+01, 1.26308074401651e+01, 1.26107676986908e+01, 1.25868614485548e+01, 1.25587621458213e+01, 1.25261847876633e+01, 1.24888964497993e+01, 1.24467257979003e+01, 1.23995711705923e+01, 1.23474068812620e+01, 1.22902874570199e+01, 1.22283496212247e+01, 1.21618119245457e+01, 1.20909720314641e+01, 1.20162017674427e+01, 1.19379401209635e+01, 1.18566844701521e+01, 1.17729803634174e+01, 1.16874102263642e+01, 1.16005813928380e+01, 1.15131138661985e+01, 1.14256282077354e+01, 1.13387339227681e+01, 1.12530186724391e+01, 1.11690385827630e+01, 1.10873098558035e+01, 1.10083018158240e+01, 1.09324314514426e+01, 1.08600594486554e+01, 1.07914876536255e+01, 1.07269578614109e+01, 1.06666517986513e+01, 1.06106921542299e+01, 1.05591445102512e+01, 1.05120200335526e+01, 1.04692788021616e+01, 1.04308336583763e+01, 1.03965544976028e+01, 1.03662729173858e+01, 1.03397871626004e+01, 1.03168673095636e+01, 1.02972606336106e+01, 1.02806971017368e+01, 1.02668949250187e+01, 1.02555660958468e+01, 1.02464218239516e+01, 1.02391777743593e+01, 1.02335590013909e+01, 1.02293044671146e+01, 1.02261710315705e+01, 1.02239368065396e+01, 1.02224037751220e+01, 1.02213995959191e+01, 1.02207785326609e+01, 1.02204214766957e+01, 1.02202350594868e+01, 1.02201498835096e+01, 1.02201179309928e+01, 1.02201092391272e+01, 1.02201079562258e+01, 1.02201078590883e+01, 1.02201061383086e+01, 1.02200944797717e+01, 1.02200516217452e+01, 1.02199373778936e+01, 1.02196873575705e+01, 1.02192085370246e+01, 1.02183758587092e+01, 1.02170300147828e+01, 1.02149765464080e+01, 1.02119863619943e+01, 1.02077977442739e+01, 1.02021198771748e+01, 1.01946378782308e+01, 1.01850192706107e+01, 1.01729217713788e+01, 1.01580022109106e+01, 1.01399263351382e+01, 1.01183791812103e+01, 1.00930756627978e+01, 1.00637709587804e+01, 1.00302702735774e+01, 9.99243753356874e+00, 9.95020260535158e+00, 9.90356666967884e+00, 9.85260545928647e+00, 9.79747016632508e+00, 9.73838594013361e+00, 9.67564802071340e+00, 9.60961567793801e+00, 9.54070424101101e+00, 9.46937559714005e+00, 9.39612760466748e+00, 9.32148289860260e+00, 9.24597756383748e+00, 9.17015011539056e+00, 9.09453116151484e+00, 9.01963404288546e+00, 8.94594664925853e+00, 8.87392452387419e+00, 8.80398528398484e+00, 8.73650431936101e+00, 8.67181168271331e+00, 8.61019005708216e+00, 8.55187367345814e+00, 8.49704805370686e+00, 8.44585046506857e+00, 8.39837098890840e+00, 8.35465412441728e+00, 8.31470086489114e+00, 8.27847119827496e+00, 8.24588699390892e+00, 8.21683524355871e+00, 8.19117162698156e+00, 8.16872437085094e+00, 8.14929836534241e+00, 8.13267949564080e+00, 8.11863913668972e+00, 8.10693874935583e+00, 8.09733450559677e+00, 8.08958186006206e+00, 8.08343997675430e+00, 8.07867591289007e+00, 8.07506845884400e+00, 8.07241153382406e+00, 8.07051704228811e+00, 8.06921710635025e+00, 8.06836560447061e+00, 8.06783896612593e+00, 8.06753619511377e+00, 8.06737811954310e+00, 8.06730589309448e+00, 8.06727879839295e+00, 8.06727142797084e+00, 8.06727034011176e+00, 8.06727025844583e+00, 8.06726882103289e+00, 8.06725908234749e+00, 8.06722328189123e+00, 8.06712785087629e+00, 8.06691900171015e+00, 8.06651902779640e+00, 8.06582346111321e+00, 8.06469921707409e+00, 8.06298383533655e+00, 8.06048590103191e+00, 8.05698670266641e+00, 8.05224314997727e+00, 8.04599193678328e+00, 8.03795489025022e+00, 8.02784539957888e+00, 8.01537576539653e+00, 8.00026525858568e+00, 7.98224862741346e+00, 7.96108474893959e+00, 7.93656508950483e+00, 7.90852162421260e+00, 7.87683387043653e+00, 7.84143471766958e+00, 7.80231478544266e+00, 7.75952511001032e+00, 7.71317804393465e+00, 7.66344634346452e+00, 7.61056050842641e+00, 7.55480452000936e+00, 7.49651018649161e+00, 7.43605035125339e+00, 7.37383124011823e+00, 7.31028422810667e+00, 7.24585729353830e+00, 7.18100640588465e+00, 7.11618706854866e+00, 7.05184621307271e+00, 6.98841461910228e+00, 6.92630001426447e+00, 6.86588098764626e+00, 6.80750182681903e+00, 6.75146835909749e+00, 6.69804484246555e+00, 6.64745191208637e+00, 6.59986554815820e+00, 6.55541699461431e+00, 6.51419352999788e+00, 6.47623997455888e+00, 6.44156081198820e+00, 6.41012280897616e+00, 6.38185802819958e+00, 6.35666714690873e+00, 6.33442301054273e+00, 6.31497436590831e+00, 6.29814972950224e+00, 6.28376135262513e+00, 6.27160924598404e+00, 6.26148522315484e+00, 6.25317691567194e+00, 6.24647170400169e+00, 6.24116049970665e+00, 6.23704130617375e+00, 6.23392247967729e+00, 6.23162561037873e+00, 6.22998794492276e+00, 6.22886427900741e+00, 6.22812825971220e+00, 6.22767305308771e+00, 6.22741135179166e+00, 6.22727471935051e+00, 6.22721229068163e+00, 6.22718887150385e+00, 6.22718250091694e+00, 6.22718156063156e+00, 6.22718149025869e+00, 6.22718025447574e+00, 6.22717188186443e+00, 6.22714110326034e+00, 6.22705905880777e+00, 6.22687950655851e+00, 6.22653564263090e+00, 6.22593766072410e+00, 6.22497116394399e+00, 6.22349652465884e+00, 6.22134926798602e+00, 6.21834153085546e+00, 6.21426462071678e+00, 6.20889266533549e+00, 6.20198730760578e+00, 6.19330335729337e+00, 6.18259526620659e+00, 6.16962424633968e+00, 6.15416580465506e+00, 6.13601742662662e+00, 6.11500610714571e+00, 6.09099540575003e+00, 6.06389169706036e+00, 6.03364929998108e+00, 6.00027420293521e+00, 5.96382615818689e+00, 5.92441899544238e+00, 5.88221910044607e+00, 5.83744211247208e+00, 5.79034800667949e+00, 5.74123483171930e+00, 5.69043145667847e+00, 5.63828973208093e+00, 5.58517647866395e+00, 5.53146568307976e+00, 5.47753120778930e+00, 5.42374022686976e+00, 5.37044749810629e+00, 5.31799049199717e+00, 5.26668533275311e+00, 5.21682347020899e+00, 5.16866899258966e+00, 5.12245650084901e+00, 5.07838948633243e+00, 5.03663917613222e+00, 4.99734382870516e+00, 4.96060847327137e+00, 4.92650509009120e+00, 4.89507322651226e+00, 4.86632103796099e+00, 4.84022673596192e+00, 4.81674041835642e+00, 4.79578625098110e+00, 4.77726496532369e+00, 4.76105663282071e+00, 4.74702367299891e+00, 4.73501404908939e+00, 4.72486460070589e+00, 4.71640445856231e+00, 4.70945848116512e+00, 4.70385064835775e+00, 4.69940734210981e+00, 4.69596044173623e+00, 4.69335015951476e+00, 4.69142754408063e+00, 4.69005658348526e+00, 4.68911584765412e+00, 4.68849962113992e+00, 4.68811849123349e+00, 4.68789937309919e+00, 4.68778497186267e+00, 4.68773270057649e+00, 4.68771309174507e+00, 4.68770775766113e+00, 4.68770697036128e+00, 4.68770691141610e+00, 4.68770587601446e+00, 4.68769886101429e+00, 4.68767307310635e+00, 4.68760433181783e+00, 4.68745389220482e+00, 4.68716577857603e+00, 4.68666473475124e+00, 4.68585488301865e+00, 4.68461916852398e+00, 4.68281964805761e+00, 4.68029866173973e+00, 4.67688090256108e+00, 4.67237637186736e+00, 4.66658417871223e+00, 4.65929710801165e+00, 4.65030684763468e+00, 4.63940972959743e+00, 4.62641280760603e+00, 4.61114006496880e+00, 4.59343852617506e+00, 4.57318403481057e+00, 4.55028646191444e+00, 4.52469412333873e+00, 4.49639721186816e+00, 4.46543008823850e+00, 4.43187232215369e+00, 4.39584842670600e+00, 4.35752628389341e+00, 4.31711431218412e+00, 4.27485747687338e+00, 4.23103228849818e+00, 4.18594097241028e+00, 4.13990502247849e+00, 4.09325837245635e+00, 4.04634042842680e+00, 3.99948920375010e+00, 3.95303478353528e+00, 3.90729331927756e+00, 3.86256171763813e+00, 3.81911314325835e+00, 3.77719340773418e+00, 3.73701826950702e+00, 3.69877162628857e+00, 3.66260454580660e+00, 3.62863505411222e+00, 3.59694858418147e+00, 3.56759898068453e+00, 3.54060995828516e+00, 3.51597691877063e+00, 3.49366904452221e+00, 3.47363160015168e+00, 3.45578838860439e+00, 3.44004432108180e+00, 3.42628807061504e+00, 3.41439478631901e+00, 3.40422884899399e+00, 3.39564664891685e+00, 3.38849936380272e+00, 3.38263570970510e+00, 3.37790463093868e+00, 3.37415788795255e+00, 3.37125249548826e+00, 3.36905295831950e+00, 3.36743324924591e+00, 3.36627847445765e+00, 3.36548617528135e+00, 3.36496722273593e+00, 3.36464627202858e+00, 3.36446175756711e+00, 3.36436542448407e+00, 3.36432140912438e+00, 3.36430489744689e+00, 3.36430040587059e+00, 3.36429974292335e+00, 3.36429969315678e+00, 3.36429881722355e+00, 3.36429288264802e+00, 3.36427106653621e+00, 3.36421291291155e+00, 3.36408564573360e+00, 3.36384191619026e+00, 3.36341807575765e+00, 3.36273305652685e+00, 3.36168792863949e+00, 3.36016619091217e+00, 3.35803483427281e+00, 3.35514619786736e+00, 3.35134061414065e+00, 3.34644981167336e+00, 3.34030101334236e+00, 3.33272163332129e+00, 3.32354444104859e+00, 3.31261302569084e+00, 3.29978736347497e+00, 3.28494926554270e+00, 3.26800746873728e+00, 3.24890212872240e+00, 3.22760848616211e+00, 3.20413950345364e+00, 3.17854731150367e+00, 3.15092336158849e+00, 3.12139724324901e+00, 3.09013420089386e+00, 3.05733145376890e+00, 3.02321349018537e+00, 2.98802656157610e+00, 2.95203264019807e+00, 2.91550312284119e+00, 2.87871256053456e+00, 2.84193267202005e+00, 2.80542685986533e+00, 2.76944539732477e+00, 2.73422139716320e+00, 2.69996761646525e+00, 2.66687409912766e+00, 2.63510661418076e+00, 2.60480581565064e+00, 2.57608702909659e+00, 2.54904056059537e+00, 2.52373242413579e+00, 2.50020539088871e+00, 2.47848027620433e+00, 2.45855739517591e+00, 2.44041813325908e+00, 2.42402659325417e+00, 2.40933129290305e+00, 2.39626689778080e+00, 2.38475598175842e+00, 2.37471081200588e+00, 2.36603515740267e+00, 2.35862611856819e+00, 2.35237597486539e+00, 2.34717403910815e+00, 2.34290850483946e+00, 2.33946826453964e+00, 2.33674467062049e+00, 2.33463320523452e+00, 2.33303502043052e+00, 2.33185830759856e+00, 2.33101945492359e+00, 2.33044395398681e+00, 2.33006702178153e+00, 2.32983391207973e+00, 2.32969989990673e+00, 2.32962993427501e+00, 2.32959796658264e+00, 2.32958597442713e+00, 2.32958271227285e+00, 2.32958223078582e+00, 2.32958219448730e+00, 2.32958155355228e+00, 2.32957721112513e+00, 2.32956124793234e+00, 2.32951869614200e+00, 2.32942557370986e+00, 2.32924723740445e+00, 2.32893712250468e+00, 2.32843593043935e+00, 2.32767131771795e+00, 2.32655812991986e+00, 2.32499921265041e+00, 2.32288681658742e+00, 2.32010459549388e+00, 2.31653017410382e+00, 2.31203823728834e+00, 2.30650406365248e+00, 2.29980739714781e+00, 2.29183652152433e+00, 2.28249237715571e+00, 2.27169254093455e+00, 2.25937488049247e+00, 2.24550069642425e+00, 2.23005718203084e+00, 2.21305905956939e+00, 2.19454929376504e+00, 2.17459883445589e+00, 2.15330539640313e+00, 2.13079134032235e+00, 2.10720076971729e+00, 2.08269599834964e+00, 2.05745356967868e+00, 2.03166002067039e+00, 2.00550757830650e+00, 1.97918996003091e+00, 1.95289842269711e+00, 1.92681817242101e+00, 1.90112521412087e+00, 1.87598368778431e+00, 1.85154371092090e+00, 1.82793972435386e+00, 1.80528932157144e+00, 1.78369252969761e+00, 1.76323150182763e+00, 1.74397057509774e+00, 1.72595664573493e+00, 1.70921981107811e+00, 1.69377422904001e+00, 1.67961914766933e+00, 1.66674006133125e+00, 1.65510995536506e+00, 1.64469060752501e+00, 1.63543392151388e+00, 1.62728327481780e+00, 1.62017486917739e+00, 1.61403907679661e+00, 1.60880177837767e+00, 1.60438569007174e+00, 1.60071167548013e+00, 1.59770003617875e+00, 1.59527177031084e+00, 1.59334978417354e+00, 1.59186003706310e+00, 1.59073259559727e+00, 1.58990257089738e+00, 1.58931091086312e+00, 1.58890502061900e+00, 1.58863918716534e+00, 1.58847478923537e+00, 1.58838028005805e+00, 1.58833093870999e+00, 1.58830839445970e+00, 1.58829993736856e+00, 1.58829763683806e+00, 1.58829729728471e+00, 1.58829727154703e+00, 1.58829681524055e+00, 1.58829372369905e+00, 1.58828235889587e+00, 1.58825206473684e+00, 1.58818576789642e+00, 1.58805880605978e+00, 1.58783803258320e+00, 1.58748124215232e+00, 1.58693695677129e+00, 1.58614460497470e+00, 1.58503511902237e+00, 1.58353196449936e+00, 1.58155260377571e+00, 1.57901037897386e+00, 1.57581678156854e+00, 1.57188405504974e+00, 1.56712805523429e+00, 1.56147127128682e+00, 1.55484589114631e+00, 1.54719677987825e+00, 1.53848423047137e+00, 1.52868634544039e+00, 1.51780091537869e+00, 1.50584667764637e+00, 1.49286386414696e+00, 1.47891398029008e+00, 1.46407879576932e+00, 1.44845856931365e+00, 1.43216957154497e+00, 1.41534100995090e+00, 1.39811149527893e+00, 1.38062521692637e+00, 1.36302801369997e+00, 1.34546353331371e+00, 1.32806966729073e+00, 1.31097542662767e+00, 1.29429838837929e+00, 1.27814279699773e+00, 1.26259835162420e+00, 1.24773965780095e+00, 1.23362627572980e+00, 1.22030326264729e+00, 1.20780208731854e+00, 1.19614179056337e+00, 1.18533027509340e+00, 1.17536562699278e+00, 1.16623739545458e+00, 1.15792778269927e+00, 1.15041271910534e+00, 1.14366281744947e+00, 1.13764421393470e+00, 1.13231931244036e+00, 1.12764745279452e+00, 1.12358552475370e+00, 1.12008854771289e+00, 1.11711023279780e+00, 1.11460353958167e+00, 1.11252123473133e+00, 1.11081645479846e+00, 1.10944327042034e+00, 1.10835724461821e+00, 1.10751597389920e+00, 1.10687959768735e+00, 1.10641125942588e+00, 1.10607750167511e+00, 1.10584857778947e+00, 1.10569866433106e+00, 1.10560596120377e+00, 1.10555267040747e+00, 1.10552484905526e+00, 1.10551213753414e+00, 1.10550736904974e+00, 1.10550607191070e+00, 1.10550588045585e+00, 1.10550586585176e+00, 1.10550560571909e+00, 1.10550384328428e+00, 1.10549736442209e+00, 1.10548009441993e+00, 1.10544230079099e+00, 1.10536992663624e+00, 1.10524408353513e+00, 1.10504073099705e+00, 1.10473056710585e+00, 1.10427915176186e+00, 1.10364727939070e+00, 1.10279161162209e+00, 1.10166557179153e+00, 1.10022049190822e+00, 1.09840698900712e+00, 1.09617653203373e+00, 1.09348314358159e+00, 1.09028516442505e+00, 1.08654699481581e+00, 1.08224071715177e+00, 1.07734750203820e+00, 1.07185870567066e+00, 1.06577658174436e+00, 1.05911455537851e+00, 1.05189703803340e+00, 1.04415879786888e+00, 1.03594393515059e+00, 1.02730454244265e+00, 1.01829915018622e+00, 1.00899106697703e+00, 9.99446719598289e-01, 9.89734082110200e-01, 9.79921259482378e-01, 9.70075263976977e-01, 9.60260996425230e-01, 9.50540423432848e-01, 9.40971927552589e-01, 9.31609800962459e-01, 9.22503853103368e-01, 9.13699107091041e-01, 9.05235566314314e-01, 8.97148039544049e-01, 8.89466018770651e-01, 8.82213608193521e-01, 8.75409505194309e-01, 8.69067035006395e-01, 8.63194240591294e-01, 8.57794028411675e-01, 8.52864369730071e-01, 8.48398556012224e-01, 8.44385506097758e-01, 8.40810122036139e-01, 8.37653689825565e-01, 8.34894320656730e-01, 8.32507427572644e-01, 8.30466231650274e-01, 8.28742290863957e-01, 8.27306043716799e-01, 8.26127358575842e-01, 8.25176078504128e-01, 8.24422550356335e-01, 8.23838126117810e-01, 8.23395624041662e-01, 8.23069737186373e-01, 8.22837377559605e-01, 8.22677945279850e-01, 8.22573513975257e-01, 8.22508925999752e-01, 8.22471793863112e-01, 8.22452407406626e-01, 8.22443549543745e-01, 8.22440226648595e-01, 8.22439322740074e-01, 8.22439189325053e-01, 8.22439179076394e-01, 8.22438995582615e-01, 8.22437752385319e-01, 8.22433182261232e-01, 8.22420999993625e-01, 8.22394339425760e-01, 8.22343281351426e-01, 8.22254491251584e-01, 8.22110983698522e-01, 8.21892027698176e-01, 8.21573204243773e-01, 8.21126623793340e-01, 8.20521307136641e-01, 8.19723728088583e-01, 8.18698510605376e-01, 8.17409266345003e-01, 8.15819551602335e-01, 8.13893915319952e-01, 8.11599003039387e-01, 8.08904675861889e-01, 8.05785099422163e-01, 8.02219756176058e-01, 7.98194335425684e-01, 7.93701459645007e-01, 7.88741212684659e-01, 7.83321444849204e-01, 7.77457840900809e-01, 7.71173748834954e-01, 7.64499778855679e-01, 7.57473192509146e-01, 7.50137110760331e-01, 7.42539576484237e-01, 7.34732511164687e-01, 7.26770607499937e-01, 7.18710199192241e-01, 7.10608146651190e-01, 7.02520772981455e-01, 6.94502878868821e-01, 6.86606858330421e-01, 6.78881930308862e-01, 6.71373494316505e-01, 6.64122612256730e-01, 6.57165613522796e-01, 6.50533816706442e-01, 6.44253358781697e-01, 6.38345121368076e-01, 6.32824743421335e-01, 6.27702710191290e-01, 6.22984509252994e-01, 6.18670845606371e-01, 6.14757909040438e-01, 6.11237688010362e-01, 6.08098325069552e-01, 6.05324509370728e-01, 6.02897901873153e-01, 6.00797588674901e-01, 5.99000557361460e-01, 5.97482190479607e-01, 5.96216769283451e-01, 5.95177979846682e-01, 5.94339412595495e-01, 5.93675045400143e-01, 5.93159699682747e-01, 5.92769458659525e-01, 5.92482036926860e-01, 5.92277091186777e-01, 5.92136463019010e-01, 5.92044346236225e-01, 5.91987373456778e-01, 5.91954619008167e-01, 5.91937518014245e-01, 5.91929704378668e-01, 5.91926773206080e-01, 5.91925975855297e-01, 5.91925858167961e-01, 5.91925849058477e-01, 5.91925685062506e-01, 5.91924573965553e-01, 5.91920489454026e-01, 5.91909601633681e-01, 5.91885773843007e-01, 5.91840140535073e-01, 5.91760783083718e-01, 5.91632518339829e-01, 5.91436812406204e-01, 5.91151829358580e-01, 5.90752621434863e-01, 5.90211463459784e-01, 5.89498329903815e-01, 5.88581507979207e-01, 5.87428334616934e-01, 5.86006039211735e-01, 5.84282667951005e-01, 5.82228059752079e-01, 5.79814838810741e-01, 5.77019385048832e-01, 5.73822741867694e-01, 5.70211421003535e-01, 5.66178067213443e-01, 5.61721951034244e-01, 5.56849265714881e-01, 5.51573214110885e-01, 5.45913882094105e-01, 5.39897905964608e-01, 5.33557951507706e-01, 5.26932030860024e-01, 5.20062689583611e-01, 5.12996099937439e-01, 5.05781097249553e-01, 4.98468194805032e-01, 4.91108609283380e-01, 4.83753324140334e-01, 4.76452213082098e-01, 4.69253240486353e-01, 4.62201750694270e-01, 4.55339853772891e-01, 4.48705911715528e-01, 4.42334126088287e-01, 4.36254225761193e-01, 4.30491251484622e-01, 4.25065432602441e-01, 4.19992150078134e-01, 4.15281979220410e-01, 4.10940805015432e-01, 4.06970002781801e-01, 4.03366676924904e-01, 4.00123950816315e-01, 3.97231301181009e-01, 3.94674930747247e-01, 3.92438173208912e-01, 3.90501924689445e-01, 3.88845095825101e-01, 3.87445078282656e-01, 3.86278219008222e-01, 3.85320294822694e-01, 3.84546979220620e-01, 3.83934292500552e-01, 3.83459025774989e-01, 3.83099129093818e-01, 3.82834053969084e-01, 3.82645041085947e-01, 3.82515344965483e-01, 3.82430388808555e-01, 3.82377844653764e-01, 3.82347636245779e-01, 3.82331864520715e-01, 3.82324658240478e-01, 3.82321954908079e-01, 3.82321219535374e-01, 3.82321110995875e-01, 3.82321102526355e-01, 3.82320949171657e-01, 3.82319910170914e-01, 3.82316090692786e-01, 3.82305909354905e-01, 3.82283627687610e-01, 3.82240955405539e-01, 3.82166747249478e-01, 3.82046805263933e-01, 3.81863798175889e-01, 3.81597306988624e-01, 3.81224002929780e-01, 3.80717960424253e-01, 3.80051103747434e-01, 3.79193781446389e-01, 3.78115457558131e-01, 3.76785503253849e-01, 3.75174067041939e-01, 3.73252996406972e-01, 3.70996779156485e-01, 3.68383469234988e-01, 3.65395559772708e-01, 3.62020766016485e-01, 3.58252682761645e-01, 3.54091285008195e-01, 3.49543246648071e-01, 3.44622059706882e-01, 3.39347945517801e-01, 3.33747558607413e-01, 3.27853493410981e-01, 3.21703612636068e-01, 3.15340223674198e-01, 3.08809135548056e-01, 3.02158633213367e-01, 2.95438408448021e-01, 2.88698486978895e-01, 2.81988189928960e-01, 2.75355164221768e-01, 2.68844511479813e-01, 2.62498038537704e-01, 2.56353645408790e-01, 2.50444858913788e-01, 2.44800512749585e-01, 2.39444568066403e-01, 2.34396063067688e-01, 2.29669176054934e-01, 2.25273383847714e-01, 2.21213696590226e-01, 2.17490950437664e-01, 2.14102141224107e-01, 2.11040784616436e-01, 2.08297291105723e-01, 2.05859347142702e-01, 2.03712296493133e-01, 2.01839518237538e-01, 2.00222799608277e-01, 1.98842702960416e-01, 1.97678926599721e-01, 1.96710658993858e-01, 1.95916925176107e-01, 1.95276923060586e-01, 1.94770346097599e-01, 1.94377687392059e-01, 1.94080519269935e-01, 1.93861741467661e-01, 1.93705790766565e-01, 1.93598805078604e-01, 1.93528735739907e-01, 1.93485403058332e-01, 1.93460491912518e-01, 1.93447486294055e-01, 1.93441543975442e-01, 1.93439314815617e-01, 1.93438708430946e-01, 1.93438618929875e-01, 1.93438611886535e-01, 1.93438483593724e-01, 1.93437614391705e-01, 1.93434419123221e-01, 1.93425901787933e-01, 1.93407262200576e-01, 1.93371566665508e-01, 1.93309496540252e-01, 1.93209187367522e-01, 1.93056169839313e-01, 1.92833422101128e-01, 1.92521540660309e-01, 1.92099034273930e-01, 1.91542741530434e-01, 1.90828368336931e-01, 1.89931136206624e-01, 1.88826526254534e-01, 1.87491097430630e-01, 1.85903351157414e-01, 1.84044608711471e-01, 1.81899862984164e-01, 1.79458563279052e-01, 1.76715291102159e-01, 1.73670286899558e-01, 1.70329792622553e-01, 1.66706182830850e-01, 1.62817867469556e-01, 1.58688961877317e-01, 1.54348733136892e-01, 1.49830845503669e-01, 1.45172440175266e-01, 1.40413094950045e-01, 1.35593716377034e-01, 1.30755420128107e-01, 1.25938454227700e-01, 1.21181214604245e-01, 1.16519393770066e-01, 1.11985292237417e-01, 1.07607309718315e-01, 1.03409620484999e-01, 9.94120256517377e-02, 9.56299655078148e-02, 9.20746679902784e-02, 8.87534051723975e-02, 8.56698281633435e-02, 8.28243517118743e-02, 8.02145625662938e-02, 7.78356296795089e-02, 7.56806990849590e-02, 7.37412611886334e-02, 7.20074829074022e-02, 7.04685012343695e-02, 6.91126782464018e-02, 6.79278202049076e-02, 6.69013652305885e-02, 6.60205450961158e-02, 6.52725270479131e-02, 6.46445413292120e-02, 6.41239993364273e-02, 6.36986062145191e-02, 6.33564703090563e-02, 6.30862103724219e-02, 6.28770599002486e-02, 6.27189665796344e-02, 6.26026836802183e-02, 6.25198494107444e-02, 6.24630498674819e-02, 6.24258612496043e-02, 6.24028675042574e-02, 6.23896504410517e-02, 6.23827505396797e-02, 6.23795980540081e-02, 6.23784154698512e-02, 6.23780937803637e-02, 6.23780462997416e-02, 6.23780425301526e-02, 6.23779734475711e-02, 6.23775054041431e-02, 6.23757848453153e-02, 6.23711986078668e-02, 6.23611624533318e-02, 6.23419447524611e-02, 6.23085335699685e-02, 6.22545551993383e-02, 6.21722519024850e-02, 6.20525260589062e-02, 6.18850569231988e-02, 6.16584945146101e-02, 6.13607326282164e-02, 6.09792594449523e-02, 6.05015797179682e-02, 5.99156971714621e-02, 5.92106398851087e-02, 5.83770055530479e-02, 5.74074982532551e-02, 5.62974244848746e-02, 5.50451144744351e-02, 5.36522357487499e-02, 5.21239701191121e-02, 5.04690325678183e-02, 4.86995206974470e-02, 4.68305955865276e-02, 4.48800079088372e-02, 4.28674956102704e-02, 4.08140898650124e-02, 3.87413732088255e-02, 3.66707368068049e-02, 3.46226823997995e-02, 3.26162088081539e-02, 3.06683137441901e-02, 2.87936303469785e-02, 2.70042057982671e-02, 2.53094180457794e-02, 2.37160171743511e-02, 2.22282710084265e-02, 2.08481903749309e-02, 1.95758080661328e-02, 1.84094866263251e-02, 1.73462331044938e-02, 1.63820031282853e-02, 1.55119812823479e-02, 1.47308292031987e-02, 1.40328967050394e-02, 1.34123945245967e-02, 1.28635299004642e-02, 1.23806081478328e-02, 1.19581046251014e-02, 1.15907120609151e-02, 1.12733682688102e-02, 1.10012690173242e-02, 1.07698704051771e-02, 1.05748845938187e-02, 1.04122721924078e-02, 1.02782339713556e-02, 1.01692039133461e-02, 1.00818449238034e-02, 1.00130478519027e-02, 9.95993385075522e-03, 9.91985955569419e-03, 9.89042410512297e-03, 9.86947668980775e-03, 9.85512311173409e-03, 9.84572977611720e-03, 9.83992363171993e-03, 9.83658680374333e-03, 9.83484500614651e-03, 9.83404923916709e-03, 9.83375073158754e-03, 9.83366953142806e-03, 9.83365754649412e-03, 9.83365658627733e-03, 9.83363887918642e-03, 9.83351891169804e-03, 9.83307790827581e-03, 9.83190242994780e-03, 9.82933030429223e-03, 9.82440583152666e-03, 9.81584668556579e-03, 9.80202504933558e-03, 9.78096562465680e-03, 9.75036315621125e-03, 9.70762190044638e-03, 9.64991894089096e-03, 9.57429229788431e-03, 9.47775333875711e-03, 9.35742106161541e-03, 9.21067349046112e-03, 9.03530887221393e-03, 8.82970690193829e-03, 8.59297819456017e-03, 8.32508907876317e-03, 8.02694889174305e-03, 7.70044857850039e-03, 7.34844263800483e-03, 6.97467114461148e-03, 6.58362423480696e-03, 6.18035733386435e-03, 5.77027059294721e-03, 5.35886966884291e-03, 4.95152657327580e-03, 4.55325875361012e-03, 4.16854211661288e-03, 3.80116978558220e-03, 3.45416338680238e-03, 3.12973804840766e-03, 2.82931679577203e-03, 2.55358566814223e-03, 2.30257855830729e-03, 2.07578063893976e-03, 1.87224049765078e-03, 1.69068262351857e-03, 1.52961312837263e-03, 1.38741293224953e-03, 1.26241474819366e-03, 1.15296299601653e-03, 1.05745824962611e-03, 9.74388902420742e-04, 9.02352315645061e-04, 8.40066765409375e-04, 7.86375136087258e-04, 7.40241805626596e-04, 7.00744911579676e-04, 6.67066383808645e-04, 6.38481552215353e-04, 6.14349193863717e-04, 5.94102150898834e-04, 5.77238403131730e-04, 5.63312600849741e-04, 5.51928266236956e-04, 5.42730952972192e-04, 5.35402579203037e-04, 5.29656997251493e-04, 5.25236720882560e-04, 5.21910640049583e-04, 5.19472512291230e-04, 5.17740009144063e-04, 5.16554098691557e-04, 5.15778554337487e-04, 5.15299396003144e-04, 5.15024095371075e-04, 5.14880412300300e-04, 5.14814773435290e-04, 5.14790151829203e-04, 5.14783454324430e-04, 5.14782465792507e-04, 5.14782385874213e-04, 5.14780903170475e-04, 5.14770857744651e-04, 5.14733931083811e-04, 5.14635508935157e-04, 5.14420169107410e-04, 5.14007976921249e-04, 5.13291826393925e-04, 5.12136095721871e-04, 5.10376916551536e-04, 5.07824375990396e-04, 5.04266964725275e-04, 4.99478534869979e-04, 4.93227916560812e-04, 4.85291140190469e-04, 4.75465902336936e-04, 4.63587498625628e-04, 4.49544967580096e-04, 4.33295756245400e-04, 4.14877009269787e-04, 3.94411792257188e-04, 3.72109270141863e-04, 3.48258895375153e-04, 3.23219542728737e-04, 2.97404696667058e-04, 2.71264089146683e-04, 2.45261261472507e-04, 2.19846721389416e-04, 1.95428645454854e-04, 1.72346626804654e-04, 1.50855882928624e-04, 1.31126494743228e-04, 1.13254676545438e-04, 9.72758790673822e-05, 8.31700365857557e-05, 7.08594613634261e-05, 6.02111208252154e-05, 5.10546280807489e-05, 4.32132371059879e-05, 3.65298366844513e-05, 3.08679669878131e-05, 2.60851783266958e-05, 2.20037148572771e-05, 1.84149498814179e-05, 1.51280311584401e-05, 1.20314647347186e-05, 9.12436082245449e-06, 6.49957079801311e-06, 4.29325967132854e-06, 2.62921770755042e-06, 1.56449004506953e-06, 9.91576563080798e-07, 6.91198401625570e-07, 6.20986501331799e-07, 6.49687258658105e-07, 6.34504669250191e-07, 5.70917205067190e-07, 4.88827836875316e-07, 4.11649367038081e-07, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.42108547152020e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -1.06581410364015e-14, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15],
[ 2.10000000000000e+01, 2.09999999871344e+01, 2.09999995893051e+01, 2.09999968939491e+01, 2.09999869854741e+01, 2.09999605726468e+01, 2.09999027664324e+01, 2.09997920514650e+01, 2.09995994885667e+01, 2.09992881788181e+01, 2.09988130118370e+01, 2.09981207124144e+01, 2.09971501907608e+01, 2.09958331926533e+01, 2.09940952370536e+01, 2.09918568205725e+01, 2.09890348607528e+01, 2.09855443437252e+01, 2.09813001364973e+01, 2.09762189200212e+01, 2.09702211962374e+01, 2.09632333204374e+01, 2.09551895094060e+01, 2.09460337757740e+01, 2.09357217397235e+01, 2.09242222705844e+01, 2.09115189129593e+01, 2.08976110548976e+01, 2.08825147994697e+01, 2.08662635060478e+01, 2.08489079738662e+01, 2.08305162481235e+01, 2.08111730380059e+01, 2.07909787464210e+01, 2.07700481226163e+01, 2.07485085607736e+01, 2.07264980795196e+01, 2.07041630284411e+01, 2.06816555774840e+01, 2.06591310529764e+01, 2.06367451894872e+01, 2.06146513695440e+01, 2.05929979232820e+01, 2.05719255574812e+01, 2.05515649784188e+01, 2.05320347658949e+01, 2.05134395471245e+01, 2.04958685094124e+01, 2.04793942800814e+01, 2.04640721914479e+01, 2.04499399380553e+01, 2.04370176231842e+01, 2.04253081820552e+01, 2.04147981602806e+01, 2.04054588181151e+01, 2.03972475239774e+01, 2.03901093946373e+01, 2.03839791344461e+01, 2.03787830220982e+01, 2.03744409907197e+01, 2.03708687456475e+01, 2.03679798641535e+01, 2.03656878226162e+01, 2.03639078992641e+01, 2.03625589045773e+01, 2.03615646966781e+01, 2.03608554454451e+01, 2.03603686164980e+01, 2.03600496544219e+01, 2.03598523534063e+01, 2.03597389126030e+01, 2.03596796827085e+01, 2.03596526192649e+01, 2.03596424667009e+01, 2.03596397049449e+01, 2.03596392973150e+01, 2.03596392438426e+01, 2.03596379979987e+01, 2.03596295572123e+01, 2.03595985278285e+01, 2.03595158136536e+01, 2.03593347894626e+01, 2.03589880825616e+01, 2.03583850815063e+01, 2.03574102705612e+01, 2.03559224651848e+01, 2.03537549984102e+01, 2.03507168811991e+01, 2.03465949324813e+01, 2.03411568475085e+01, 2.03341551471472e+01, 2.03253319265294e+01, 2.03144242995594e+01, 2.03011704163678e+01, 2.02853159138148e+01, 2.02666206442142e+01, 2.02448655140838e+01, 2.02198592525578e+01, 2.01914449181201e+01, 2.01595059432234e+01, 2.01239715106237e+01, 2.00848210551714e+01, 2.00420876931065e+01, 1.99958604003286e+01, 1.99462847936454e+01, 1.98935624152134e+01, 1.98379484789428e+01, 1.97797481051233e+01, 1.97193111407188e+01, 1.96570257313367e+01, 1.95933108703097e+01, 1.95286081950139e+01, 1.94633733266710e+01, 1.93980670559162e+01, 1.93331466632966e+01, 1.92690576346038e+01, 1.92062259900407e+01, 1.91450513987871e+01, 1.90859012015523e+01, 1.90291054173772e+01, 1.89749527702796e+01, 1.89236877380357e+01, 1.88755085999282e+01, 1.88305664421667e+01, 1.87889650677281e+01, 1.87507617500814e+01, 1.87159687660547e+01, 1.86845556405557e+01, 1.86564520337800e+01, 1.86315511990826e+01, 1.86097139363397e+01, 1.85907729611712e+01, 1.85745376049281e+01, 1.85607987541638e+01, 1.85493339319178e+01, 1.85399124171644e+01, 1.85323002939352e+01, 1.85262653186515e+01, 1.85215814937972e+01, 1.85180332388316e+01, 1.85154190556318e+01, 1.85135545960001e+01, 1.85122750528741e+01, 1.85114368145649e+01, 1.85109183421004e+01, 1.85106202528182e+01, 1.85104646178078e+01, 1.85103935056209e+01, 1.85103668287798e+01, 1.85103595720129e+01, 1.85103585009286e+01, 1.85103584063722e+01, 1.85103565535869e+01, 1.85103440006868e+01, 1.85102978548013e+01, 1.85101748460671e+01, 1.85099056406766e+01, 1.85093900656468e+01, 1.85084934309654e+01, 1.85070441091125e+01, 1.85048325024602e+01, 1.85016114956518e+01, 1.84970984525933e+01, 1.84909787759712e+01, 1.84829110012287e+01, 1.84725333470174e+01, 1.84594715910710e+01, 1.84433480856185e+01, 1.84237916718798e+01, 1.84004482015438e+01, 1.83729913276262e+01, 1.83411331913727e+01, 1.83046346096943e+01, 1.82633143626077e+01, 1.82170571953369e+01, 1.81658201870145e+01, 1.81096371975752e+01, 1.80486211845741e+01, 1.79829642779969e+01, 1.79129356069997e+01, 1.78388769794772e+01, 1.77611966141439e+01, 1.76803612067278e+01, 1.75968866701975e+01, 1.75113279200549e+01, 1.74242680797261e+01, 1.73363074615371e+01, 1.72480526418414e+01, 1.71601059019280e+01, 1.70730552564279e+01, 1.69874652435683e+01, 1.69038686100813e+01, 1.68227589888787e+01, 1.67445846390508e+01, 1.66697432936510e+01, 1.65985781393533e+01, 1.65313749321534e+01, 1.64683602343702e+01, 1.64097007405368e+01, 1.63555036440020e+01, 1.63058179828627e+01, 1.62606368935954e+01, 1.62199006933552e+01, 1.61835007068291e+01, 1.61512837498796e+01, 1.61230571790307e+01, 1.60985944122429e+01, 1.60776408217924e+01, 1.60599198941940e+01, 1.60451395451717e+01, 1.60329984702597e+01, 1.60231924045927e+01, 1.60154201598814e+01, 1.60093893036270e+01, 1.60048213463687e+01, 1.60014563080838e+01, 1.59990565453581e+01, 1.59974097368455e+01, 1.59963309456371e+01, 1.59956637028640e+01, 1.59952800861727e+01, 1.59950797983632e+01, 1.59949882839930e+01, 1.59949539535966e+01, 1.59949446148778e+01, 1.59949432365018e+01, 1.59949431259390e+01, 1.59949410854943e+01, 1.59949272611758e+01, 1.59948764414375e+01, 1.59947409740521e+01, 1.59944445038378e+01, 1.59938767158214e+01, 1.59928892898681e+01, 1.59912932454526e+01, 1.59888578237145e+01, 1.59853110179864e+01, 1.59803418231935e+01, 1.59736042284633e+01, 1.59647229257504e+01, 1.59533006505164e+01, 1.59389270094042e+01, 1.59211885861812e+01, 1.58996800536166e+01, 1.58740159589409e+01, 1.58438427982321e+01, 1.58088509549439e+01, 1.57687860540679e+01, 1.57234592796633e+01, 1.56727562220248e+01, 1.56166438624268e+01, 1.55551753673381e+01, 1.54884924479107e+01, 1.54168251407850e+01, 1.53404889782622e+01, 1.52598796342451e+01, 1.51754652507046e+01, 1.50877767602709e+01, 1.49973966152054e+01, 1.49049464022130e+01, 1.48110738579399e+01, 1.47164397962555e+01, 1.46217054154616e+01, 1.45275203776752e+01, 1.44345119557129e+01, 1.43432754398104e+01, 1.42543659017702e+01, 1.41682913380162e+01, 1.40855071601401e+01, 1.40064119708985e+01, 1.39313445504059e+01, 1.38605819751546e+01, 1.37943387956325e+01, 1.37327672025686e+01, 1.36759581149946e+01, 1.36239431246868e+01, 1.35766972313612e+01, 1.35341423017052e+01, 1.34961511833080e+01, 1.34625524018404e+01, 1.34331353662615e+01, 1.34076560020584e+01, 1.33858427263242e+01, 1.33674026707875e+01, 1.33520280499958e+01, 1.33394025623187e+01, 1.33292077021798e+01, 1.33211288540847e+01, 1.33148610338123e+01, 1.33101141407894e+01, 1.33066175892049e+01, 1.33041241945638e+01, 1.33024132074691e+01, 1.33012924073239e+01, 1.33005991947538e+01, 1.33002006518381e+01, 1.32999925723053e+01, 1.32998974980860e+01, 1.32998618323084e+01, 1.32998521303374e+01, 1.32998506983463e+01, 1.32998505878973e+01, 1.32998486045889e+01, 1.32998351673781e+01, 1.32997857707115e+01, 1.32996540968840e+01, 1.32993659296479e+01, 1.32988140461597e+01, 1.32978542887381e+01, 1.32963029920029e+01, 1.32939359110392e+01, 1.32904887618926e+01, 1.32856594463541e+01, 1.32791119880148e+01, 1.32704821555781e+01, 1.32593846925121e+01, 1.32454220101845e+01, 1.32281941364818e+01, 1.32073096464818e+01, 1.31823972399613e+01, 1.31531175770984e+01, 1.31191749437683e+01, 1.30803282961598e+01, 1.30364012349933e+01, 1.29872904846560e+01, 1.29329725022716e+01, 1.28735079138733e+01, 1.28090435650848e+01, 1.27398120760401e+01, 1.26661288978943e+01, 1.25883869744868e+01, 1.25070492116023e+01, 1.24226390430681e+01, 1.23357294540169e+01, 1.22469308743177e+01, 1.21568783873451e+01, 1.20662187094069e+01, 1.19755973825559e+01, 1.18856465886958e+01, 1.17969739380333e+01, 1.17101525141089e+01, 1.16257123765970e+01, 1.15441336385981e+01, 1.14658411542176e+01, 1.13912007809630e+01, 1.13205171244557e+01, 1.12540326326133e+01, 1.11919278831681e+01, 1.11343229006323e+01, 1.10812793437365e+01, 1.10328034183237e+01, 1.09888493898831e+01, 1.09493235908228e+01, 1.09140888372819e+01, 1.08829691865798e+01, 1.08557549779034e+01, 1.08322081048801e+01, 1.08120674692691e+01, 1.07950545606533e+01, 1.07808790986825e+01, 1.07692446633819e+01, 1.07598542267965e+01, 1.07524154874066e+01, 1.07466458989512e+01, 1.07422772790413e+01, 1.07390598815331e+01, 1.07367658209847e+01, 1.07351917481502e+01, 1.07341606923457e+01, 1.07335230091359e+01, 1.07331563990972e+01, 1.07329649940178e+01, 1.07328775390908e+01, 1.07328447316693e+01, 1.07328358072496e+01, 1.07328344900238e+01, 1.07328343902874e+01, 1.07328326234676e+01, 1.07328206530031e+01, 1.07327766483035e+01, 1.07326593478871e+01, 1.07324026385731e+01, 1.07319110086774e+01, 1.07310560575091e+01, 1.07296742220105e+01, 1.07275658564542e+01, 1.07244957720621e+01, 1.07201953093108e+01, 1.07143659755920e+01, 1.07066846340624e+01, 1.06968101757730e+01, 1.06843915471283e+01, 1.06690769400729e+01, 1.06505238859895e+01, 1.06284099301279e+01, 1.06024435064485e+01, 1.05723745886265e+01, 1.05380046672606e+01, 1.04991956010754e+01, 1.04558769147864e+01, 1.04080511699722e+01, 1.03557971168729e+01, 1.02992704407037e+01, 1.02387020392025e+01, 1.01743938996228e+01, 1.01067127724840e+01, 1.00360819548965e+01, 9.96297158784033e+00, 9.88788793135943e+00, 9.81136210466699e+00, 9.73393876425001e+00, 9.65616514609812e+00, 9.57858082568437e+00, 9.50170846124571e+00, 9.42604569303065e+00, 9.35205828347732e+00, 9.28017450856764e+00, 9.21078075367575e+00, 9.14421822968732e+00, 9.08078070577643e+00, 9.02071315072098e+00, 8.96421118069767e+00, 8.91142122368093e+00, 8.86244132501927e+00, 8.81732253260689e+00, 8.77607081151928e+00, 8.73864944620449e+00, 8.70498189316403e+00, 8.67495504876459e+00, 8.64842289579070e+00, 8.62521048897716e+00, 8.60511823439133e+00, 8.58792641047522e+00, 8.57339987013944e+00, 8.56129285395356e+00, 8.55135383477865e+00, 8.54333030487932e+00, 8.53697340851672e+00, 8.53204231722025e+00, 8.52830824231137e+00, 8.52555798063903e+00, 8.52359689549353e+00, 8.52225124559711e+00, 8.52136979086037e+00, 8.52082462378769e+00, 8.52051119918913e+00, 8.52034756107795e+00, 8.52027279296540e+00, 8.52024474478291e+00, 8.52023711498944e+00, 8.52023598884763e+00, 8.52023590430780e+00, 8.52023441631101e+00, 8.52022433491205e+00, 8.52018727459607e+00, 8.52008848519350e+00, 8.51987228595237e+00, 8.51945823451747e+00, 8.51873818300180e+00, 8.51757435448407e+00, 8.51579855141844e+00, 8.51321258217289e+00, 8.50958996259750e+00, 8.50467891550962e+00, 8.49820665180295e+00, 8.48988487250288e+00, 8.47941638213802e+00, 8.46650265178702e+00, 8.45085211758907e+00, 8.43218895083417e+00, 8.41026199319103e+00, 8.38485351975899e+00, 8.35578747779756e+00, 8.32293685363489e+00, 8.28622984621644e+00, 8.24565457269833e+00, 8.20126209664882e+00, 8.15316764773524e+00, 8.10154998647871e+00, 8.04664895128831e+00, 7.98876130059969e+00, 7.92823502532967e+00, 7.86546235340009e+00, 7.80087169903012e+00, 7.73491882748965e+00, 7.66807751495998e+00, 7.60082998671292e+00, 7.53365741694946e+00, 7.46703076972065e+00, 7.40140224932589e+00, 7.33719760611268e+00, 7.27480950600903e+00, 7.21459211837625e+00, 7.15685700977515e+00, 7.10187035776178e+00, 7.04985142805297e+00, 7.00097219955845e+00, 6.95535798174065e+00, 6.91308885048439e+00, 6.87420173085785e+00, 6.83869297321456e+00, 6.80652129667298e+00, 6.77761100468033e+00, 6.75185540583187e+00, 6.72912039585696e+00, 6.70924817198659e+00, 6.69206105856380e+00, 6.67736542354392e+00, 6.66495566082215e+00, 6.65461820472038e+00, 6.64613553208379e+00, 6.63929009581409e+00, 6.63386812270841e+00, 6.62966319943518e+00, 6.62647956444733e+00, 6.62413502148747e+00, 6.62246339269820e+00, 6.62131643652716e+00, 6.62056516756443e+00, 6.62010053175820e+00, 6.61983341035515e+00, 6.61969394834281e+00, 6.61963022686809e+00, 6.61960632272208e+00, 6.61959982021279e+00, 6.61959886045597e+00, 6.61959878862581e+00, 6.61959752725227e+00, 6.61958898126126e+00, 6.61955756530010e+00, 6.61947382192665e+00, 6.61929055182849e+00, 6.61893956853277e+00, 6.61832920834904e+00, 6.61734271130244e+00, 6.61583757136497e+00, 6.61364593382410e+00, 6.61057609337727e+00, 6.60641511770021e+00, 6.60093258716685e+00, 6.59388540185850e+00, 6.58502356238138e+00, 6.57409678252277e+00, 6.56086174156786e+00, 6.54508973528215e+00, 6.52657444110304e+00, 6.50513947956101e+00, 6.48064543517745e+00, 6.45299600058395e+00, 6.42214293095369e+00, 6.38808954391125e+00, 6.35089257230085e+00, 6.31066226980230e+00, 6.26756077515241e+00, 6.22179884918964e+00, 6.17363119752614e+00, 6.12335066775502e+00, 6.07128165376673e+00, 6.01777304628207e+00, 5.96319104027752e+00, 5.90791205538368e+00, 5.85231595781462e+00, 5.79677970603014e+00, 5.74167148842486e+00, 5.68734538584986e+00, 5.63413657479852e+00, 5.58235708412854e+00, 5.53229212250350e+00, 5.48419699876647e+00, 5.43829465857370e+00, 5.39477385579790e+00, 5.35378796682456e+00, 5.31545444186156e+00, 5.27985487233974e+00, 5.24703563978494e+00, 5.21700910086260e+00, 5.18975525639503e+00, 5.16522384892859e+00, 5.14333683314203e+00, 5.12399116492596e+00, 5.10706185712532e+00, 5.09240525164408e+00, 5.07986245807428e+00, 5.06926290780434e+00, 5.06042796965121e+00, 5.05317456877411e+00, 5.04731874558888e+00, 5.04267908643750e+00, 5.03907995380377e+00, 5.03635444181554e+00, 5.03434698343187e+00, 5.03291553966823e+00, 5.03193330877648e+00, 5.03128990446908e+00, 5.03089196673952e+00, 5.03066318597405e+00, 5.03054374003098e+00, 5.03048916378993e+00, 5.03046869029251e+00, 5.03046312099864e+00, 5.03046229898231e+00, 5.03046223743791e+00, 5.03046115637958e+00, 5.03045383204881e+00, 5.03042690701080e+00, 5.03035513456603e+00, 5.03019806141802e+00, 5.02989724406924e+00, 5.02937410928973e+00, 5.02852855519079e+00, 5.02723837346257e+00, 5.02535955494200e+00, 5.02272751948638e+00, 5.01915928675074e+00, 5.01445657664930e+00, 5.00840979705495e+00, 5.00080284211456e+00, 4.99141858842581e+00, 4.98004493981097e+00, 4.96648123665851e+00, 4.95054481532585e+00, 4.93207747963202e+00, 4.91095163262630e+00, 4.88707581474739e+00, 4.86039940557167e+00, 4.83091627098199e+00, 4.79866717509231e+00, 4.76374082495130e+00, 4.72627347340954e+00, 4.68644706847988e+00, 4.64448600260650e+00, 4.60065257886862e+00, 4.55524136960552e+00, 4.50857269264598e+00, 4.46098546788054e+00, 4.41282973947679e+00, 4.36445915466351e+00, 4.31622367806337e+00, 4.26846279199935e+00, 4.22149939065770e+00, 4.17563452351526e+00, 4.13114308601040e+00, 4.08827049828931e+00, 4.04723036077537e+00, 4.00820303203088e+00, 3.97133504224643e+00, 3.93673923553475e+00, 3.90449552548550e+00, 3.87465214955890e+00, 3.84722731659678e+00, 3.82221115547607e+00, 3.79956788924621e+00, 3.77923817580056e+00, 3.76114157148932e+00, 3.74517908683047e+00, 3.73123581283649e+00, 3.71918360209663e+00, 3.70888379065565e+00, 3.70018994523070e+00, 3.69295061597315e+00, 3.68701206856821e+00, 3.68222096186456e+00, 3.67842692941321e+00, 3.67548501626423e+00, 3.67325791706455e+00, 3.67161795873737e+00, 3.67044877142794e+00, 3.66964659533278e+00, 3.66912117855600e+00, 3.66879623200812e+00, 3.66860942103701e+00, 3.66851188917245e+00, 3.66846732612602e+00, 3.66845060899982e+00, 3.66844606153721e+00, 3.66844539034129e+00, 3.66844533995551e+00, 3.66844445312354e+00, 3.66843844470783e+00, 3.66841635715637e+00, 3.66835748000950e+00, 3.66822862960472e+00, 3.66798186869653e+00, 3.66755275889570e+00, 3.66685922886384e+00, 3.66580112954854e+00, 3.66426053392054e+00, 3.66210282202382e+00, 3.65917857198801e+00, 3.65532625343149e+00, 3.65037569117940e+00, 3.64415223470570e+00, 3.63648153309857e+00, 3.62719477828062e+00, 3.61613424304345e+00, 3.60315890813428e+00, 3.58814994749145e+00, 3.57101582613588e+00, 3.55169676418970e+00, 3.53016833519272e+00, 3.50644399826383e+00, 3.48057641106920e+00, 3.45265743162478e+00, 3.42281678757837e+00, 3.39121946627279e+00, 3.35806195119611e+00, 3.32356749385535e+00, 3.28798065884632e+00, 3.25156140963277e+00, 3.21457901112736e+00, 3.17730601286975e+00, 3.14001254608248e+00, 3.10296112371881e+00, 3.06640208051807e+00, 3.03056973602236e+00, 2.99567931284571e+00, 2.96192459931682e+00, 2.92947631240890e+00, 2.89848109444160e+00, 2.86906106479999e+00, 2.84131384428532e+00, 2.81531297260623e+00, 2.79110864676612e+00, 2.76872871776431e+00, 2.74817989356524e+00, 2.72944910660131e+00, 2.71250501344828e+00, 2.69729960232448e+00, 2.68376989047847e+00, 2.67183969820809e+00, 2.66142148912267e+00, 2.65241826728753e+00, 2.64472552110966e+00, 2.63823320135701e+00, 2.63282771678560e+00, 2.62839392584371e+00, 2.62481709731737e+00, 2.62198480716360e+00, 2.61978873379181e+00, 2.61812631034415e+00, 2.61690219067404e+00, 2.61602948618655e+00, 2.61543073375940e+00, 2.61503856067175e+00, 2.61479602065510e+00, 2.61465658545444e+00, 2.61458378806888e+00, 2.61455052642325e+00, 2.61453804884477e+00, 2.61453465464213e+00, 2.61453415366497e+00, 2.61453411589712e+00, 2.61453344901766e+00, 2.61452893081247e+00, 2.61451232143130e+00, 2.61446804706572e+00, 2.61437115444280e+00, 2.61418559632656e+00, 2.61386291811767e+00, 2.61334140827361e+00, 2.61254576663667e+00, 2.61138733445311e+00, 2.60976491735414e+00, 2.60756621728194e+00, 2.60466987082607e+00, 2.60094806949790e+00, 2.59626971234119e+00, 2.59050401373315e+00, 2.58352446066494e+00, 2.57521298620751e+00, 2.56546420175532e+00, 2.55418951274197e+00, 2.54132093349138e+00, 2.52681441886552e+00, 2.51065254464564e+00, 2.49284639514633e+00, 2.47343655399985e+00, 2.45249313958803e+00, 2.43011487639985e+00, 2.40642724327609e+00, 2.38157978479674e+00, 2.35574270946090e+00, 2.32910292554245e+00, 2.30185968179556e+00, 2.27421998612214e+00, 2.24639397246488e+00, 2.21859037652640e+00, 2.19101226624435e+00, 2.16385315452075e+00, 2.13729360010193e+00, 2.11149837787564e+00, 2.08661427232683e+00, 2.06276851805703e+00, 2.04006788047464e+00, 2.01859834012079e+00, 1.99842531817752e+00, 1.97959436098217e+00, 1.96213218964054e+00, 1.94604801778973e+00, 1.93133504567130e+00, 1.91797205033370e+00, 1.90592500770993e+00, 1.89514870002730e+00, 1.88558827923771e+00, 1.87718077215072e+00, 1.86985652457992e+00, 1.86354058956339e+00, 1.85815406858858e+00, 1.85361541511183e+00, 1.84984170712050e+00, 1.84674989078327e+00, 1.84425799116508e+00, 1.84228627935051e+00, 1.84075837888942e+00, 1.83960228895653e+00, 1.83875129760574e+00, 1.83814475646842e+00, 1.83772868848808e+00, 1.83745620289993e+00, 1.83728769653862e+00, 1.83719082737604e+00, 1.83714025446998e+00, 1.83711714764013e+00, 1.83710847952692e+00, 1.83710612159528e+00, 1.83710577356970e+00, 1.83710574718984e+00, 1.83710527949806e+00, 1.83710211082036e+00, 1.83709046246966e+00, 1.83705941257942e+00, 1.83699146235922e+00, 1.83686133609283e+00, 1.83663506587347e+00, 1.83626940795776e+00, 1.83571163269599e+00, 1.83489972377124e+00, 1.83376301432603e+00, 1.83222327680540e+00, 1.83019626943343e+00, 1.82759372484003e+00, 1.82432574552644e+00, 1.82030354716349e+00, 1.81544246529176e+00, 1.80966511555583e+00, 1.80290457438404e+00, 1.79510742860825e+00, 1.78623653161871e+00, 1.77627330281500e+00, 1.76521941838667e+00, 1.75309776601736e+00, 1.73995257392450e+00, 1.72584867418544e+00, 1.71086991831266e+00, 1.69511682455868e+00, 1.67870359504547e+00, 1.66175468934057e+00, 1.64440117266599e+00, 1.62677706633276e+00, 1.60901591308807e+00, 1.59124773263768e+00, 1.57359648843835e+00, 1.55617812467754e+00, 1.53909917198076e+00, 1.52245587049064e+00, 1.50633372528965e+00, 1.49080739360256e+00, 1.47594080411878e+00, 1.46178742189466e+00, 1.44839059234812e+00, 1.43578391985139e+00, 1.42399165656313e+00, 1.41302909316131e+00, 1.40290295422713e+00, 1.39361180743539e+00, 1.38514649829283e+00, 1.37749062201047e+00, 1.37062104221551e+00, 1.36450846340567e+00, 1.35911806086982e+00, 1.35441016855807e+00, 1.35034102222249e+00, 1.34686355210358e+00, 1.34392821650179e+00, 1.34148386474547e+00, 1.33947861537809e+00, 1.33786073291603e+00, 1.33657948439821e+00, 1.33558595531891e+00, 1.33483380358109e+00, 1.33427992999763e+00, 1.33388504473762e+00, 1.33361411104587e+00, 1.33343665056040e+00, 1.33332689853938e+00, 1.33326380211985e+00, 1.33323086012961e+00, 1.33321580866834e+00, 1.33321016234657e+00, 1.33320862641024e+00, 1.33320839970926e+00, 1.33320838241660e+00, 1.33320807439438e+00, 1.33320598749896e+00, 1.33319831585731e+00, 1.33317786612874e+00, 1.33313311265864e+00, 1.33304740540787e+00, 1.33289836290040e+00, 1.33265747894732e+00, 1.33228996781591e+00, 1.33175486763130e+00, 1.33100541586937e+00, 1.32998970355041e+00, 1.32865160595156e+00, 1.32693197723269e+00, 1.32477008444776e+00, 1.32210524341782e+00, 1.31887860564352e+00, 1.31503503293285e+00, 1.31052498606453e+00, 1.30530634703218e+00, 1.29934609250666e+00, 1.29262173998651e+00, 1.28512249789814e+00, 1.27685006608021e+00, 1.26781905226621e+00, 1.25805699139251e+00, 1.24760397556918e+00, 1.23651192128071e+00, 1.22484351530465e+00, 1.21267089119871e+00, 1.20007409407873e+00, 1.18713939347928e+00, 1.17395750335603e+00, 1.16062176572232e+00, 1.14722635070144e+00, 1.13386452124637e+00, 1.12062700545263e+00, 1.10760051315623e+00, 1.09486642630406e+00, 1.08249968452919e+00, 1.07056787880028e+00, 1.05913055746461e+00, 1.04823874106582e+00, 1.03793463554107e+00, 1.02825152819729e+00, 1.01921384743827e+00, 1.01083736555612e+00, 1.00312952382206e+00, 9.96089860284343e-01, 9.89710522709174e-01, 9.83976851565881e-01, 9.78868020476238e-01, 9.74357723793524e-01, 9.70414902707031e-01, 9.67004502325977e-01, 9.64088252519357e-01, 9.61625464892279e-01, 9.59573837261594e-01, 9.57890255510801e-01, 9.56531580965073e-01, 9.55455409670193e-01, 9.54620788435356e-01, 9.53988871449745e-01, 9.53523500916305e-01, 9.53191695619029e-01, 9.52964032742720e-01, 9.52814910604367e-01, 9.52722683162389e-01, 9.52669661094390e-01, 9.52641978666826e-01, 9.52629330307428e-01, 9.52624585466356e-01, 9.52623294754087e-01, 9.52623104247643e-01, 9.52623089613343e-01, 9.52622827598272e-01, 9.52621052407764e-01, 9.52614526622188e-01, 9.52597131293061e-01, 9.52559062138580e-01, 9.52486155528483e-01, 9.52359371495849e-01, 9.52154457947628e-01, 9.51841816674236e-01, 9.51386587561093e-01, 9.50748962417805e-01, 9.49884733893416e-01, 9.48746077906257e-01, 9.47282559840666e-01, 9.45442345541860e-01, 9.43173588138727e-01, 9.40425951403252e-01, 9.37152220361128e-01, 9.33309941005135e-01, 9.28863024121519e-01, 9.23783244302996e-01, 9.18051564956404e-01, 9.11659224046822e-01, 9.04608523647930e-01, 8.96913278878820e-01, 8.88598897862561e-01, 8.79702082906313e-01, 8.70270162830540e-01, 8.60360085735106e-01, 8.50037118931407e-01, 8.39373316886967e-01, 8.28445827729812e-01, 8.17335113488880e-01, 8.06123158662054e-01, 7.94891736282594e-01, 7.83720791229140e-01, 7.72686988254879e-01, 7.61862458421788e-01, 7.51313763628918e-01, 7.41101085870891e-01, 7.31277636630519e-01, 7.21889272950396e-01, 7.12974300468897e-01, 7.04563439989993e-01, 6.96679932716368e-01, 6.89339759706293e-01, 6.82551952950892e-01, 6.76318978236349e-01, 6.70637173220427e-01, 6.65497227536086e-01, 6.60884694932449e-01, 6.56780530240546e-01, 6.53161646143335e-01, 6.50001486236459e-01, 6.47270611644771e-01, 6.44937298521139e-01, 6.42968143159482e-01, 6.41328670308603e-01, 6.39983938724836e-01, 6.38899136227472e-01, 6.38040154725246e-01, 6.37374134075340e-01, 6.36869962424498e-01, 6.36498720046433e-01, 6.36234053770675e-01, 6.36052469979951e-01, 6.35933535852146e-01, 6.35859980986030e-01, 6.35817694654417e-01, 6.35795617498054e-01, 6.35785530290288e-01, 6.35781746234827e-01, 6.35780716880404e-01, 6.35780564949862e-01, 6.35780553189818e-01, 6.35780341476341e-01, 6.35778907086937e-01, 6.35773634127730e-01, 6.35759578415811e-01, 6.35728818145388e-01, 6.35669909740592e-01, 6.35567471263592e-01, 6.35401913474510e-01, 6.35149335970393e-01, 6.34781602623239e-01, 6.34266606603624e-01, 6.33568730421555e-01, 6.32649500476926e-01, 6.31468428516481e-01, 6.29984024206212e-01, 6.28154953992947e-01, 6.25941311995579e-01, 6.23305959486306e-01, 6.20215881413795e-01, 6.16643502294735e-01, 6.12567900559853e-01, 6.07975860846757e-01, 6.02862708276437e-01, 5.97232877537508e-01, 5.91100182280091e-01, 5.84487766069941e-01, 5.77427733748969e-01, 5.69960479997370e-01, 5.62133748613412e-01, 5.54001470059788e-01, 5.45622435017609e-01, 5.37058867340804e-01, 5.28374960728957e-01, 5.19635439955103e-01, 5.10904200327406e-01, 5.02243069242908e-01, 4.93710722337919e-01, 4.85361774960932e-01, 4.77246058452451e-01, 4.69408080757731e-01, 4.61886662718975e-01, 4.54714735244139e-01, 4.47919278467779e-01, 4.41521381873823e-01, 4.35536403891287e-01, 4.29974210381253e-01, 4.24839473356098e-01, 4.20132013862272e-01, 4.15847175892079e-01, 4.11976221184847e-01, 4.08506737597730e-01, 4.05423056186937e-01, 4.02706674106483e-01, 4.00336681819152e-01, 3.98290193883600e-01, 3.96542782736820e-01, 3.95068914477907e-01, 3.93842384760589e-01, 3.92836751634736e-01, 3.92025760686611e-01, 3.91383756279069e-01, 3.90886071260272e-01, 3.90509386364492e-01, 3.90232049823265e-01, 3.90034347561006e-01, 3.89898714840559e-01, 3.89809881372926e-01, 3.89754943673704e-01, 3.89723360744327e-01, 3.89706871838769e-01, 3.89699337974111e-01, 3.89696511769873e-01, 3.89695742974599e-01, 3.89695629502057e-01, 3.89695620647608e-01, 3.89695460323072e-01, 3.89694374101413e-01, 3.89690381045725e-01, 3.89679737102380e-01, 3.89656443483776e-01, 3.89611834916042e-01, 3.89534264931807e-01, 3.89408903805272e-01, 3.89217662851394e-01, 3.88939256508873e-01, 3.88549410719307e-01, 3.88021222417116e-01, 3.87325670299933e-01, 3.86432271378361e-01, 3.85309871150533e-01, 3.83927547814011e-01, 3.82255603101601e-01, 3.80266604675757e-01, 3.77936438254011e-01, 3.75245322563707e-01, 3.72178737623681e-01, 3.68728217379015e-01, 3.64891961782320e-01, 3.60675231081970e-01, 3.56090496002611e-01, 3.51157330935568e-01, 3.45902052109416e-01, 3.40357117677200e-01, 3.34560320381438e-01, 3.28553814711878e-01, 3.22383028310142e-01, 3.16095511255291e-01, 3.09739776686996e-01, 3.03364182319509e-01, 2.97015895451736e-01, 2.90739974999134e-01, 2.84578593874119e-01, 2.78570414699843e-01, 2.72750122202225e-01, 2.67148107323795e-01, 2.61790291545228e-01, 2.56698075260175e-01, 2.51888391301392e-01, 2.47373843684802e-01, 2.43162912043747e-01, 2.39260203740365e-01, 2.35666737923022e-01, 2.32380248529118e-01, 2.29395496128792e-01, 2.26704581334662e-01, 2.24297255084338e-01, 2.22161223302596e-01, 2.20282445181226e-01, 2.18645425526926e-01, 2.17233502306691e-01, 2.16029130681683e-01, 2.15014164508158e-01, 2.14170135567468e-01, 2.13478529758333e-01, 2.12921058256956e-01, 2.12479920350408e-01, 2.12138053411529e-01, 2.11879364441963e-01, 2.11688936885249e-01, 2.11553206100238e-01, 2.11460097051216e-01, 2.11399118436294e-01, 2.11361408617666e-01, 2.11339730270105e-01, 2.11328412524630e-01, 2.11323241423035e-01, 2.11321301575587e-01, 2.11320773891114e-01, 2.11320696006027e-01, 2.11320689876821e-01, 2.11320578234623e-01, 2.11319821843091e-01, 2.11317041278988e-01, 2.11309629402052e-01, 2.11293409127368e-01, 2.11262347034264e-01, 2.11208335030125e-01, 2.11121051330590e-01, 2.10987910866255e-01, 2.10794113755249e-01, 2.10522798529649e-01, 2.10155304202125e-01, 2.09671541860839e-01, 2.09050472206126e-01, 2.08270680314335e-01, 2.07311033094214e-01, 2.06151398708748e-01, 2.04773401147211e-01, 2.03161177754172e-01, 2.01302103532626e-01, 1.99187444101721e-01, 1.96812899855232e-01, 1.94179007464143e-01, 1.91291371419222e-01, 1.88160707483796e-01, 1.84802691043242e-01, 1.81237615426660e-01, 1.77489877204525e-01, 1.73587316081520e-01, 1.69560445300807e-01, 1.65441613727289e-01, 1.61264142628568e-01, 1.57061478666861e-01, 1.52866400160232e-01, 1.48710306953298e-01, 1.44622616102335e-01, 1.40630276908229e-01, 1.36757410431272e-01, 1.33025071143354e-01, 1.29451122265944e-01, 1.26050211844358e-01, 1.22833833766720e-01, 1.19810456642057e-01, 1.16985703489057e-01, 1.14362566277673e-01, 1.11941641211402e-01, 1.09721372952212e-01, 1.07698298518574e-01, 1.05867284120311e-01, 1.04221750568183e-01, 1.02753884993859e-01, 1.01454838358023e-01, 1.00314909565977e-01, 9.93237179334088e-02, 9.84703662538351e-02, 9.77435968360325e-02, 9.71319426423456e-02, 9.66238751202120e-02, 9.62079495448585e-02, 9.58729477601885e-02, 9.56080172044480e-02, 9.54028041314672e-02, 9.52475780798601e-02, 9.51333439881203e-02, 9.50519379735368e-02, 9.49961027364878e-02, 9.49595388386939e-02, 9.49369287244188e-02, 9.49239312650931e-02, 9.49171457423290e-02, 9.49140454545301e-02, 9.49128824414558e-02, 9.49125660748251e-02, 9.49125193798110e-02, 9.49125156725898e-02, 9.49124477330336e-02, 9.49119874332851e-02, 9.49102953335519e-02, 9.49057849001278e-02, 9.48959143492196e-02, 9.48770126816036e-02, 9.48441475865955e-02, 9.47910424495575e-02, 9.47100491430426e-02, 9.45921825116685e-02, 9.44272213408368e-02, 9.42038789860256e-02, 9.39100446093200e-02, 9.35330930481122e-02, 9.30602577372035e-02, 9.24790569371616e-02, 9.17777590349971e-02, 9.09458682497508e-02, 8.99746081728345e-02, 8.88573777333441e-02, 8.75901529226191e-02, 8.61718083612999e-02, 8.46043357831512e-02, 8.28929417277884e-02, 8.10460138741753e-02, 7.90749539260887e-02, 7.69938839806059e-02, 7.48192419579752e-02, 7.25692890514722e-02, 7.02635575208639e-02, 6.79222700100546e-02, 6.55657617435352e-02, 6.32139345979681e-02, 6.08857675850736e-02, 5.85989023484110e-02, 5.63693155849130e-02, 5.42110835576395e-02, 5.21362376735475e-02, 5.01547049114919e-02, 4.82743229743754e-02, 4.65009175033551e-02, 4.48384274784388e-02, 4.32890648740312e-02, 4.18534955057766e-02, 4.05310295377213e-02, 3.93198120618585e-02, 3.82170062923954e-02, 3.72189640481260e-02, 3.63213801897047e-02, 3.55194294355705e-02, 3.48078854451828e-02, 3.41812232039977e-02, 3.36337065720791e-02, 3.31594633810752e-02, 3.27525507075990e-02, 3.24070129452778e-02, 3.21169350757771e-02, 3.18764931396842e-02, 3.16800033714415e-02, 3.15219708339178e-02, 3.13971377174802e-02, 3.13005308062593e-02, 3.12275070141226e-02, 3.11737954029887e-02, 3.11355337595600e-02, 3.11092976532414e-02, 3.10921199431426e-02, 3.10814989431272e-02, 3.10753938689707e-02, 3.10722067445290e-02, 3.10707505837229e-02, 3.10702043376700e-02, 3.10700557464578e-02, 3.10700338147427e-02, 3.10700320575918e-02, 3.10699996549317e-02, 3.10697801227313e-02, 3.10689731078959e-02, 3.10668219699153e-02, 3.10621145984591e-02, 3.10531007270782e-02, 3.10374296221063e-02, 3.10121119373719e-02, 3.09735093907371e-02, 3.09173556610141e-02, 3.08388114243634e-02, 3.07325556444944e-02, 3.05929140072401e-02, 3.04140236895485e-02, 3.01900314748629e-02, 2.99153196479764e-02, 2.95847512914982e-02, 2.91939238224934e-02, 2.87394171861344e-02, 2.82190214375291e-02, 2.76319278531680e-02, 2.69788685029972e-02, 2.62621915247188e-02, 2.54858631241390e-02, 2.46553923162303e-02, 2.37776801578065e-02, 2.28608010957281e-02, 2.19137293968146e-02, 2.09460278158780e-02, 1.99675182198362e-02, 1.89879545629879e-02, 1.80167173948576e-02, 1.70625462117364e-02, 1.61333218586215e-02, 1.52359063893570e-02, 1.43760428747406e-02, 1.35583131374410e-02, 1.27861476981543e-02, 1.20618795813137e-02, 1.13868321158854e-02, 1.07614303982260e-02, 1.01853264883864e-02, 9.65752947916698e-03, 9.17653308896149e-03, 8.74043516795453e-03, 8.34704526690899e-03, 7.99397803010393e-03, 7.67873152654985e-03, 7.39875069042029e-03, 7.15147681512107e-03, 6.93438457864204e-03, 6.74500840843351e-03, 6.58096015460075e-03, 6.43994005407578e-03, 6.31974286674364e-03, 6.21826087548527e-03, 6.13348519872048e-03, 6.06350658538446e-03, 6.00651656842643e-03, 5.96080955033784e-03, 5.92478609615199e-03, 5.89695742312912e-03, 5.87595081865189e-03, 5.86051550037325e-03, 5.84952827210472e-03, 5.84199823405740e-03, 5.83706978475007e-03, 5.83402320498294e-03, 5.83227223308214e-03, 5.83135821559111e-03, 5.83094062715261e-03, 5.83078398075187e-03, 5.83074136966744e-03, 5.83073508034815e-03, 5.83073457172389e-03, 5.83072513297722e-03, 5.83066118436903e-03, 5.83042610714557e-03, 5.82979951288465e-03, 5.82842840603703e-03, 5.82580325943383e-03, 5.82124025698505e-03, 5.81387100041297e-03, 5.80264101130723e-03, 5.78631834241250e-03, 5.76351349531734e-03, 5.73271157256983e-03, 5.69231711938301e-03, 5.64071140624378e-03, 5.57632096227323e-03, 5.49769503593112e-03, 5.40358841842092e-03, 5.29304485981541e-03, 5.16547531534073e-03, 5.02072467889647e-03, 4.85912067038541e-03, 4.68149927300487e-03, 4.48920259014685e-03, 4.28404712322461e-03, 4.06826304308794e-03, 3.84440773379779e-03, 3.61525937790219e-03, 3.38369831342789e-03, 3.15258505412430e-03, 2.92464407103310e-03, 2.70236162994664e-03, 2.48790423594514e-03, 2.28306172528647e-03, 2.08921616175672e-03, 1.90733503971074e-03, 1.73798533413105e-03, 1.58136383864971e-03, 1.43733928732104e-03, 1.30550300927368e-03, 1.18522640270413e-03, 1.07536224725990e-03, 9.75818656915095e-04, 8.85237384665061e-04, 8.02693170115276e-04, 7.27291720068024e-04, 6.58199141792437e-04, 5.94663562271336e-04, 5.36229306668190e-04, 4.81742441866828e-04, 4.31353223184772e-04, 3.84508312492926e-04, 3.40941997727384e-04, 3.00462856451134e-04, 2.62938805072110e-04, 2.28281562822730e-04, 1.96431630801186e-04, 1.67344654872892e-04, 1.40979772890404e-04, 1.17290273461634e-04, 9.62166481457416e-05, 7.76819226580017e-05, 6.15890170649891e-05, 4.78384895465922e-05, 3.62510474616329e-05, 2.66905657362315e-05, 1.89826132519499e-05, 1.29385168960994e-05, 8.35857349557045e-06, 5.03559595088632e-06, 2.75874358024453e-06, 1.31765380828597e-06, 5.06978089731547e-07, -7.10542735760100e-15, -3.55271367880050e-15],
[ 2.20000000000000e+01, 2.19999999861574e+01, 2.19999995581155e+01, 2.19999966580646e+01, 2.19999859971047e+01, 2.19999575783922e+01, 2.19998953821595e+01, 2.19997762590634e+01, 2.19995690721105e+01, 2.19992341199642e+01, 2.19987228659922e+01, 2.19979779882524e+01, 2.19969337560430e+01, 2.19955167290079e+01, 2.19936467654099e+01, 2.19912383173924e+01, 2.19882019831100e+01, 2.19844462787415e+01, 2.19798795877368e+01, 2.19744122402654e+01, 2.19679586726849e+01, 2.19604396148650e+01, 2.19517842522342e+01, 2.19419323093370e+01, 2.19308360023690e+01, 2.19184618095477e+01, 2.19047920103000e+01, 2.18898259472106e+01, 2.18735809686779e+01, 2.18560930154709e+01, 2.18374168210997e+01, 2.18176257042307e+01, 2.17968109413101e+01, 2.17750807189415e+01, 2.17525586780562e+01, 2.17293820749861e+01, 2.17056995975441e+01, 2.16816688864151e+01, 2.16574538228261e+01, 2.16332216519617e+01, 2.16091400174402e+01, 2.15853739850459e+01, 2.15620831337725e+01, 2.15394187891650e+01, 2.15175214682780e+01, 2.14965185977101e+01, 2.14765225566396e+01, 2.14576290861100e+01, 2.14399160944824e+01, 2.14234428774583e+01, 2.14082497597292e+01, 2.13943581544316e+01, 2.13817710263848e+01, 2.13704737357068e+01, 2.13604352299492e+01, 2.13516095454281e+01, 2.13439375720267e+01, 2.13373490304529e+01, 2.13317646068310e+01, 2.13270981866532e+01, 2.13232591285918e+01, 2.13201545185429e+01, 2.13176913455804e+01, 2.13157785442695e+01, 2.13143288519968e+01, 2.13132604355533e+01, 2.13124982480390e+01, 2.13119750850842e+01, 2.13116323181888e+01, 2.13114202924185e+01, 2.13112983855079e+01, 2.13112347353001e+01, 2.13112056521234e+01, 2.13111947418789e+01, 2.13111917740145e+01, 2.13111913359635e+01, 2.13111912785004e+01, 2.13111899396799e+01, 2.13111808689626e+01, 2.13111475238770e+01, 2.13110586368217e+01, 2.13108641030799e+01, 2.13104915224203e+01, 2.13098435224500e+01, 2.13087959699388e+01, 2.13071971514083e+01, 2.13048679770735e+01, 2.13016032335083e+01, 2.12971738810406e+01, 2.12913303627390e+01, 2.12838068638128e+01, 2.12743264340323e+01, 2.12626068619367e+01, 2.12483671683492e+01, 2.12313345679678e+01, 2.12112517311645e+01, 2.11878841630877e+01, 2.11610275033801e+01, 2.11305145373671e+01, 2.10962216992991e+01, 2.10580748418918e+01, 2.10160540466061e+01, 2.09701972589470e+01, 2.09206025555144e+01, 2.08674288867048e+01, 2.08108951912537e+01, 2.07512778444818e+01, 2.06889064770586e+01, 2.06241582793063e+01, 2.05574509803506e+01, 2.04892347546179e+01, 2.04199833542509e+01, 2.03501847910959e+01, 2.02803318947700e+01, 2.02109130554553e+01, 2.01424034252355e+01, 2.00752568052249e+01, 2.00098983932123e+01, 1.99467185134702e+01, 1.98860674011916e+01, 1.98282510717043e+01, 1.97735282707295e+01, 1.97221084766892e+01, 1.96741509086237e+01, 1.96297644822019e+01, 1.95890086498679e+01, 1.95518950576225e+01, 1.95183899487770e+01, 1.94884172430375e+01, 1.94618622166382e+01, 1.94385757054885e+01, 1.94183787482705e+01, 1.94010675802787e+01, 1.93864188818909e+01, 1.93741951784625e+01, 1.93641502818198e+01, 1.93560346581601e+01, 1.93496006038276e+01, 1.93446071098605e+01, 1.93408242990416e+01, 1.93380373258996e+01, 1.93360496409560e+01, 1.93346855354860e+01, 1.93337919018823e+01, 1.93332391668134e+01, 1.93329213789656e+01, 1.93327554592472e+01, 1.93326796478279e+01, 1.93326512081321e+01, 1.93326434718250e+01, 1.93326423299616e+01, 1.93326422291568e+01, 1.93326402539360e+01, 1.93326268715181e+01, 1.93325776762342e+01, 1.93324465388992e+01, 1.93321595441408e+01, 1.93316098998693e+01, 1.93306540170062e+01, 1.93291089303506e+01, 1.93267512001244e+01, 1.93233173982217e+01, 1.93185062432661e+01, 1.93119824040304e+01, 1.93033819415416e+01, 1.92923193066038e+01, 1.92783957523546e+01, 1.92612089623471e+01, 1.92403636357352e+01, 1.92154827154229e+01, 1.91862188960777e+01, 1.91522660107725e+01, 1.91133698718754e+01, 1.90693381376317e+01, 1.90200487938704e+01, 1.89654568823341e+01, 1.89055991732285e+01, 1.88405965672857e+01, 1.87706541168271e+01, 1.86960586684297e+01, 1.86171742425044e+01, 1.85344353674759e+01, 1.84483386693469e+01, 1.83594330747167e+01, 1.82683090138711e+01, 1.81755870114088e+01, 1.80819060296257e+01, 1.79879118915322e+01, 1.78942460636681e+01, 1.78015350307190e+01, 1.77103804491965e+01, 1.76213502284803e+01, 1.75349706543994e+01, 1.74517196418074e+01, 1.73720211763472e+01, 1.72962409802603e+01, 1.72246834120575e+01, 1.71575895855214e+01, 1.70951366709489e+01, 1.70374383220387e+01, 1.69845461563821e+01, 1.69364522064673e+01, 1.68930922510984e+01, 1.68543499331916e+01, 1.68200615677156e+01, 1.67900215417601e+01, 1.67639882061845e+01, 1.67416901543100e+01, 1.67228327774226e+01, 1.67071049797075e+01, 1.66941859273352e+01, 1.66837516987677e+01, 1.66754816971718e+01, 1.66690646823859e+01, 1.66642042803657e+01, 1.66606238334166e+01, 1.66580704654491e+01, 1.66563182532170e+01, 1.66551704168159e+01, 1.66544604699827e+01, 1.66540523018890e+01, 1.66538391958004e+01, 1.66537418245990e+01, 1.66537052970944e+01, 1.66536953607081e+01, 1.66536938941176e+01, 1.66536937764789e+01, 1.66536916054480e+01, 1.66536768963890e+01, 1.66536228242488e+01, 1.66534786871373e+01, 1.66531632433539e+01, 1.66525591183552e+01, 1.66515085010604e+01, 1.66498103198893e+01, 1.66472190554918e+01, 1.66434453089761e+01, 1.66381582011091e+01, 1.66309896289722e+01, 1.66215403516524e+01, 1.66093878159755e+01, 1.65940955679881e+01, 1.65752240277081e+01, 1.65523423363928e+01, 1.65250409209392e+01, 1.64929443634461e+01, 1.64557241201886e+01, 1.64131106079042e+01, 1.63649041703152e+01, 1.63109844570309e+01, 1.62513177916855e+01, 1.61859621760141e+01, 1.61150696694979e+01, 1.60388859965752e+01, 1.59577473599906e+01, 1.58720745727465e+01, 1.57823647536304e+01, 1.56891809519725e+01, 1.55931401646560e+01, 1.54949002715530e+01, 1.53951464365807e+01, 1.52945774981200e+01, 1.51938928094307e+01, 1.50937798985163e+01, 1.49949032134554e+01, 1.48978941195360e+01, 1.48033422309496e+01, 1.47117880985535e+01, 1.46237172363310e+01, 1.45395554481479e+01, 1.44596654068064e+01, 1.43843444332066e+01, 1.43138234204600e+01, 1.42482668440241e+01, 1.41877737940621e+01, 1.41323799609125e+01, 1.40820604996286e+01, 1.40367336955514e+01, 1.39962653498336e+01, 1.39604738012991e+01, 1.39291354982843e+01, 1.39019910304128e+01, 1.38787515250365e+01, 1.38591053061060e+01, 1.38427247047100e+01, 1.38292729010672e+01, 1.38184106683629e+01, 1.38098028807362e+01, 1.38031246423319e+01, 1.37980668929363e+01, 1.37943413494844e+01, 1.37916846524405e+01, 1.37898616020926e+01, 1.37886673920388e+01, 1.37879287749239e+01, 1.37875041277215e+01, 1.37872824190557e+01, 1.37871811174947e+01, 1.37871431156135e+01, 1.37871327781645e+01, 1.37871312523781e+01, 1.37871311346948e+01, 1.37871290214798e+01, 1.37871147041331e+01, 1.37870620719880e+01, 1.37869217735206e+01, 1.37866147312573e+01, 1.37860266990157e+01, 1.37850040760482e+01, 1.37833511650092e+01, 1.37808290287224e+01, 1.37771560641108e+01, 1.37720103697350e+01, 1.37650339355438e+01, 1.37558386292863e+01, 1.37440138936974e+01, 1.37291360029782e+01, 1.37107786581762e+01, 1.36885246318644e+01, 1.36619781070930e+01, 1.36307772988387e+01, 1.35946069033453e+01, 1.35532098968429e+01, 1.35063982042245e+01, 1.34540617828144e+01, 1.33961757167676e+01, 1.33328049919670e+01, 1.32641067154907e+01, 1.31903296520976e+01, 1.31118110660821e+01, 1.30289709734182e+01, 1.29423040198370e+01, 1.28523692995306e+01, 1.27597785114596e+01, 1.26651829114434e+01, 1.25692595548274e+01, 1.24726973342001e+01, 1.23761832985862e+01, 1.22803896961127e+01, 1.21859621150079e+01, 1.20935090139051e+01, 1.20035928393926e+01, 1.19167228348601e+01, 1.18333495577936e+01, 1.17538610491013e+01, 1.16785805419703e+01, 1.16077655607875e+01, 1.15416082421414e+01, 1.14802367073019e+01, 1.14237173251412e+01, 1.13720577220085e+01, 1.13252104165034e+01, 1.12830769788555e+01, 1.12455126339408e+01, 1.12123312419377e+01, 1.11833106001890e+01, 1.11581980136376e+01, 1.11367160795113e+01, 1.11185686254853e+01, 1.11034467304286e+01, 1.10910347444229e+01, 1.10810162115402e+01, 1.10730795865234e+01, 1.10669236266150e+01, 1.10622623338435e+01, 1.10588293223356e+01, 1.10563814906044e+01, 1.10547018907184e+01, 1.10536017047378e+01, 1.10529212632395e+01, 1.10525300700435e+01, 1.10523258299380e+01, 1.10522325104896e+01, 1.10521975030688e+01, 1.10521879801945e+01, 1.10521865746380e+01, 1.10521864682135e+01, 1.10521845829141e+01, 1.10521718097321e+01, 1.10521248541512e+01, 1.10519996877131e+01, 1.10517257635228e+01, 1.10512011640037e+01, 1.10502888750155e+01, 1.10488143573026e+01, 1.10465645593800e+01, 1.10432884847061e+01, 1.10386993894018e+01, 1.10324786438714e+01, 1.10242812417228e+01, 1.10137428821845e+01, 1.10004884884787e+01, 1.09841419561107e+01, 1.09643368548602e+01, 1.09407277406890e+01, 1.09130016741611e+01, 1.08808894962818e+01, 1.08441763868962e+01, 1.08027112302653e+01, 1.07564143409383e+01, 1.07052831621086e+01, 1.06493956369392e+01, 1.05889110663646e+01, 1.05240683970533e+01, 1.04551820204129e+01, 1.03826352960924e+01, 1.03068721295346e+01, 1.02283870223050e+01, 1.01477140685751e+01, 1.00654153877122e+01, 9.98206946242364e+00, 9.89825979972196e+00, 9.81456425693814e+00, 9.73154528784655e+00, 9.64974127554982e+00, 9.56965903842089e+00, 9.49176752967127e+00, 9.41649270322650e+00, 9.34421348855447e+00, 9.27525880241437e+00, 9.20990552219778e+00, 9.14837734923355e+00, 9.09084449721544e+00, 9.03742414789024e+00, 8.98818162160768e+00, 8.94313221362081e+00, 8.90224364819774e+00, 8.86543910206949e+00, 8.83260074692896e+00, 8.80357375788513e+00, 8.77817073102065e+00, 8.75617644839628e+00, 8.73735292285529e+00, 8.72144464776933e+00, 8.70818396861451e+00, 8.69729648442939e+00, 8.68850637851763e+00, 8.68154157017563e+00, 8.67613857382919e+00, 8.67204694983839e+00, 8.66903323332838e+00, 8.66688423442553e+00, 8.66540961557138e+00, 8.66444366908764e+00, 8.66384624037597e+00, 8.66350276816134e+00, 8.66332344180121e+00, 8.66324150544165e+00, 8.66321076817691e+00, 8.66320240688548e+00, 8.66320117277601e+00, 8.66320108013102e+00, 8.66319944947409e+00, 8.66318840153081e+00, 8.66314778808252e+00, 8.66303952723278e+00, 8.66280259944888e+00, 8.66234884841205e+00, 8.66155975281568e+00, 8.66028431423285e+00, 8.65833818186121e+00, 8.65550410696452e+00, 8.65153378759493e+00, 8.64615112710725e+00, 8.63905688766989e+00, 8.62993467248235e+00, 8.61845811844380e+00, 8.60429912616733e+00, 8.58713689904720e+00, 8.56666751103309e+00, 8.54261367802381e+00, 8.51473437484536e+00, 8.48283392284455e+00, 8.44677017550435e+00, 8.40646145289985e+00, 8.36189191994465e+00, 8.31311516568580e+00, 8.26025581685817e+00, 8.20350910263654e+00, 8.14313837279720e+00, 8.07947065282069e+00, 8.01289039295876e+00, 7.94383163211076e+00, 7.87276885142172e+00, 7.80020683754594e+00, 7.72666991169615e+00, 7.65269090635358e+00, 7.57880028305873e+00, 7.50551577672289e+00, 7.43333291943920e+00, 7.36271673770642e+00, 7.29409483410347e+00, 7.22785196610652e+00, 7.16432613347158e+00, 7.10380609517537e+00, 7.04653016874738e+00, 6.99268612527812e+00, 6.94241198268004e+00, 6.89579751288299e+00, 6.85288630742461e+00, 6.81367828155269e+00, 6.77813253200944e+00, 6.74617049299223e+00, 6.71767935577370e+00, 6.69251572957058e+00, 6.67050952529394e+00, 6.65146804131091e+00, 6.63518022301610e+00, 6.62142105749575e+00, 6.60995605232375e+00, 6.60054573483009e+00, 6.59295009615449e+00, 6.58693289407286e+00, 6.58226572090792e+00, 6.57873173866267e+00, 6.57612898355410e+00, 6.57427314687818e+00, 6.57299974883069e+00, 6.57216563643857e+00, 6.57164975565833e+00, 6.57135317016461e+00, 6.57119832427833e+00, 6.57112757356455e+00, 6.57110103247269e+00, 6.57109381264684e+00, 6.57109274701548e+00, 6.57109266726146e+00, 6.57109126674079e+00, 6.57108177800630e+00, 6.57104689640106e+00, 6.57095391474454e+00, 6.57075042613693e+00, 6.57036071909102e+00, 6.56968300894531e+00, 6.56858763217373e+00, 6.56691631006679e+00, 6.56448256482242e+00, 6.56107334364475e+00, 6.55645187425582e+00, 6.55036173770307e+00, 6.54253210134333e+00, 6.53268400692604e+00, 6.52053755719968e+00, 6.50581979187005e+00, 6.48827299353023e+00, 6.46766312073297e+00, 6.44378803360878e+00, 6.41648516227852e+00, 6.38563827400455e+00, 6.35118302429121e+00, 6.31311103038400e+00, 6.27147228027850e+00, 6.22637578069160e+00, 6.17798844489302e+00, 6.12653231551532e+00, 6.07228029810740e+00, 6.01555063997745e+00, 5.95670042138980e+00, 5.89611833292469e+00, 5.83421699884370e+00, 5.77142507966121e+00, 5.70817935639999e+00, 5.64491697101576e+00, 5.58206797569482e+00, 5.52004832807962e+00, 5.45925345722504e+00, 5.40005251243234e+00, 5.34278339077268e+00, 5.28774861742616e+00, 5.23521212621803e+00, 5.18539695787372e+00, 5.13848386342571e+00, 5.09461077289963e+00, 5.05387306733701e+00, 5.01632457685474e+00, 4.98197921916706e+00, 4.95081319117348e+00, 4.92276762947774e+00, 4.89775166226082e+00, 4.87564578289840e+00, 4.85630548336537e+00, 4.83956509143881e+00, 4.82524175909233e+00, 4.81313954988133e+00, 4.80305357065619e+00, 4.79477408815694e+00, 4.78809056482722e+00, 4.78279554165926e+00, 4.77868829026231e+00, 4.77557815282226e+00, 4.77328748821659e+00, 4.77165414603120e+00, 4.77053339801217e+00, 4.76979926860818e+00, 4.76934522235834e+00, 4.76908418523955e+00, 4.76894789871089e+00, 4.76888562789867e+00, 4.76886226790706e+00, 4.76885591341727e+00, 4.76885497550768e+00, 4.76885490528635e+00, 4.76885367181334e+00, 4.76884531485107e+00, 4.76881459375571e+00, 4.76873270245645e+00, 4.76855348464934e+00, 4.76821025842091e+00, 4.76761337672462e+00, 4.76664863467776e+00, 4.76517661675597e+00, 4.76303305629817e+00, 4.76003025580832e+00, 4.75595958885722e+00, 4.75059507260970e+00, 4.74369796412664e+00, 4.73502229410524e+00, 4.72432120972294e+00, 4.71135395550147e+00, 4.69589327999099e+00, 4.67773301946963e+00, 4.65669558090141e+00, 4.63263902819805e+00, 4.60546347110346e+00, 4.57511646676203e+00, 4.54159717123343e+00, 4.50495902168904e+00, 4.46531078831369e+00, 4.42281590537825e+00, 4.37769006978850e+00, 4.33019717801820e+00, 4.28064375343666e+00, 4.22937209008882e+00, 4.17675240055012e+00, 4.12317429975566e+00, 4.06903798008173e+00, 4.01474543352979e+00, 3.96069205481217e+00, 3.90725891684391e+00, 3.85480595197326e+00, 3.80366620405007e+00, 3.75414124463443e+00, 3.70649777762090e+00, 3.66096539569034e+00, 3.61773540320101e+00, 3.57696058552674e+00, 3.53875578483027e+00, 3.50319913572564e+00, 3.47033381905502e+00, 3.44017020527662e+00, 3.41268827772859e+00, 3.38784024743693e+00, 3.36555329268918e+00, 3.34573237631950e+00, 3.32826311011392e+00, 3.31301464804316e+00, 3.29984259774583e+00, 3.28859194280870e+00, 3.27909996727162e+00, 3.27119916905799e+00, 3.26472014157019e+00, 3.25949439353924e+00, 3.25535706753712e+00, 3.25214950854388e+00, 3.24972162675378e+00, 3.24793399441452e+00, 3.24665961569762e+00, 3.24578531187389e+00, 3.24521267153492e+00, 3.24485852700668e+00, 3.24465493283936e+00, 3.24454863942971e+00, 3.24450007334269e+00, 3.24448185456718e+00, 3.24447689862231e+00, 3.24447616713537e+00, 3.24447611222363e+00, 3.24447514573090e+00, 3.24446859760305e+00, 3.24444452603175e+00, 3.24438036036263e+00, 3.24423993689825e+00, 3.24397101503989e+00, 3.24350337605987e+00, 3.24274759652884e+00, 3.24159457636422e+00, 3.23991588696214e+00, 3.23756498707140e+00, 3.23437933206301e+00, 3.23018337531158e+00, 3.22479242819216e+00, 3.21801730797887e+00, 3.20966966176450e+00, 3.19956781134994e+00, 3.18754292170329e+00, 3.17344525762913e+00, 3.15715026374305e+00, 3.13856418578769e+00, 3.11762895036061e+00, 3.09432603783996e+00, 3.06867912075180e+00, 3.04075529615364e+00, 3.01066481283695e+00, 2.97855927727245e+00, 2.94462840960905e+00, 2.90909550510818e+00, 2.87221182952690e+00, 2.83425023246303e+00, 2.79549829569746e+00, 2.75625134178147e+00, 2.71680561199332e+00, 2.67745188546745e+00, 2.63846975803242e+00, 2.60012273659384e+00, 2.56265423956776e+00, 2.52628453208253e+00, 2.49120857126852e+00, 2.45759469507770e+00, 2.42558405906276e+00, 2.39529070917007e+00, 2.36680217343753e+00, 2.34018045937608e+00, 2.31546335428469e+00, 2.29266594037689e+00, 2.27178225323227e+00, 2.25278702897433e+00, 2.23563750139078e+00, 2.22027522402670e+00, 2.20662790350434e+00, 2.19461123863887e+00, 2.18413076521651e+00, 2.17508370864051e+00, 2.16736084622546e+00, 2.16084837805197e+00, 2.15542980041393e+00, 2.15098776954246e+00, 2.14740593610988e+00, 2.14457072371239e+00, 2.14237301784446e+00, 2.14070972654143e+00, 2.13948517053444e+00, 2.13861225995572e+00, 2.13801341668028e+00, 2.13762120638541e+00, 2.13737865219066e+00, 2.13723921189504e+00, 2.13716641273592e+00, 2.13713315048237e+00, 2.13712067270858e+00, 2.13711727845590e+00, 2.13711677747147e+00, 2.13711673970307e+00, 2.13711607281398e+00, 2.13711155454485e+00, 2.13709494494880e+00, 2.13705067016508e+00, 2.13695377741192e+00, 2.13676822205352e+00, 2.13644555805037e+00, 2.13592409642711e+00, 2.13512858849251e+00, 2.13397048087119e+00, 2.13234877723517e+00, 2.13015152619107e+00, 2.12725793545082e+00, 2.12354108886587e+00, 2.11887121532455e+00, 2.11311942774903e+00, 2.10616181809896e+00, 2.09788376276394e+00, 2.08818426500524e+00, 2.07698014052816e+00, 2.06420984211121e+00, 2.04983672220818e+00, 2.03385155022929e+00, 2.01627413395368e+00, 1.99715394064130e+00, 1.97656966956753e+00, 1.95462778913787e+00, 1.93146011282427e+00, 1.90722054316706e+00, 1.88208115696566e+00, 1.85622783389139e+00, 1.82985564329139e+00, 1.80316420008855e+00, 1.77635318237049e+00, 1.74961817377653e+00, 1.72314695712701e+00, 1.69711634599107e+00, 1.67168960174395e+00, 1.64701444805812e+00, 1.62322166473353e+00, 1.60042421942519e+00, 1.57871687950354e+00, 1.55817623669170e+00, 1.53886107352479e+00, 1.52081300205613e+00, 1.50405731044598e+00, 1.48860396094302e+00, 1.47444869220713e+00, 1.46157418894079e+00, 1.44995129157194e+00, 1.43954022760429e+00, 1.43029185373627e+00, 1.42214890361907e+00, 1.41504723999699e+00, 1.40891711190010e+00, 1.40368441760609e+00, 1.39927197243145e+00, 1.39560077732444e+00, 1.39259128007551e+00, 1.39016461617220e+00, 1.38824381139788e+00, 1.38675492372366e+00, 1.38562809838232e+00, 1.38479850769745e+00, 1.38420714663694e+00, 1.38380145639528e+00, 1.38353575165392e+00, 1.38337143241343e+00, 1.38327696815695e+00, 1.38322765016915e+00, 1.38320511657122e+00, 1.38319666347273e+00, 1.38319436402800e+00, 1.38319402463489e+00, 1.38319399890936e+00, 1.38319354281822e+00, 1.38319045273554e+00, 1.38317909329299e+00, 1.38314881340691e+00, 1.38308254771855e+00, 1.38295564521951e+00, 1.38273497391913e+00, 1.38237834589651e+00, 1.38183430173175e+00, 1.38104228674331e+00, 1.37993324316597e+00, 1.37843063185381e+00, 1.37645188383755e+00, 1.37391026588076e+00, 1.37071712518734e+00, 1.36678445720263e+00, 1.36202771813646e+00, 1.35636878205803e+00, 1.34973892323877e+00, 1.34208169015340e+00, 1.33335553045126e+00, 1.32353602815060e+00, 1.31261762642956e+00, 1.30061473180590e+00, 1.28756212714758e+00, 1.27351465959445e+00, 1.25854621187824e+00, 1.24274800788866e+00, 1.22622634172650e+00, 1.20909985041301e+00, 1.19149647127770e+00, 1.17355023442663e+00, 1.15539803853433e+00, 1.13717654569385e+00, 1.11901931037517e+00, 1.10105423146237e+00, 1.08340138782684e+00, 1.06617128969871e+00, 1.04946355246926e+00, 1.03336597806333e+00, 1.01795401251619e+00, 1.00329053708754e+00, 9.89425943854542e-01, 9.76398444620830e-01, 9.64234563351681e-01, 9.52949766330100e-01, 9.42549189982877e-01, 9.33028433090321e-01, 9.24374387214559e-01, 9.16566086122280e-01, 9.09575561313478e-01, 9.03368696176258e-01, 8.97906075541108e-01, 8.93143830365243e-01, 8.89034478882404e-01, 8.85527765828140e-01, 8.82571500397471e-01, 8.80112391586316e-01, 8.78096876754586e-01, 8.76471935929768e-01, 8.75185880891724e-01, 8.74189104810526e-01, 8.73434775513964e-01, 8.72879453669736e-01, 8.72483616542834e-01, 8.72212068701693e-01, 8.72034223151893e-01, 8.71924239805683e-01, 8.71861012758924e-01, 8.71828003254876e-01, 8.71812921102514e-01, 8.71807263292787e-01, 8.71805724233841e-01, 8.71805497072060e-01, 8.71805479744250e-01, 8.71805171095961e-01, 8.71803079959953e-01, 8.71795392745682e-01, 8.71774901630943e-01, 8.71730058221246e-01, 8.71644181151641e-01, 8.71494850923604e-01, 8.71253522346755e-01, 8.70885381476977e-01, 8.70349469622262e-01, 8.69599091452034e-01, 8.68582517099412e-01, 8.67243979058337e-01, 8.65524953473304e-01, 8.63365702178275e-01, 8.60707036973935e-01, 8.57492251929717e-01, 8.53669154126926e-01, 8.49192109696567e-01, 8.44024011894419e-01, 8.38138072906332e-01, 8.31519342420027e-01, 8.24165864542156e-01, 8.16089400457365e-01, 8.07315666554810e-01, 7.97884064993568e-01, 7.87846913569563e-01, 7.77268211606764e-01, 7.66222005688704e-01, 7.54790440945719e-01, 7.43061598505601e-01, 7.31127226613459e-01, 7.19080471734877e-01, 7.07013707411239e-01, 6.95016544131999e-01, 6.83174084847657e-01, 6.71565469961845e-01, 6.60262734646224e-01, 6.49329981798644e-01, 6.38822857213572e-01, 6.28788300400277e-01, 6.19264535357583e-01, 6.10281260470202e-01, 6.01859995174273e-01, 5.94014542561663e-01, 5.86751530932609e-01, 5.80071002704504e-01, 5.73967025315778e-01, 5.68428305193144e-01, 5.63438791965691e-01, 5.58978265531696e-01, 5.55022903060458e-01, 5.51545826400979e-01, 5.48517632620033e-01, 5.45906911525556e-01, 5.43680754124590e-01, 5.41805255141060e-01, 5.40246011138297e-01, 5.38968613647175e-01, 5.37939134211907e-01, 5.37124595672378e-01, 5.36493421551310e-01, 5.36015853351813e-01, 5.35664324115544e-01, 5.35413775922922e-01, 5.35241909252246e-01, 5.35129353295140e-01, 5.35059748411548e-01, 5.35019734773776e-01, 5.34998844703949e-01, 5.34989300008096e-01, 5.34985719487363e-01, 5.34984745501198e-01, 5.34984601742959e-01, 5.34984590699761e-01, 5.34984392980373e-01, 5.34983053403810e-01, 5.34978128998805e-01, 5.34965002497842e-01, 5.34936276257024e-01, 5.34881265079051e-01, 5.34785609596938e-01, 5.34631030190965e-01, 5.34395239382008e-01, 5.34052028044098e-01, 5.33571537319823e-01, 5.32920723494016e-01, 5.32064017022684e-01, 5.30964169275755e-01, 5.29583271375195e-01, 5.27883919082562e-01, 5.25830486588834e-01, 5.23390461146732e-01, 5.20535780862975e-01, 5.17244110867772e-01, 5.13499989715697e-01, 5.09295779251108e-01, 5.04632357898821e-01, 4.99519509460360e-01, 4.93975976397412e-01, 4.88029167001677e-01, 4.81714527980447e-01, 4.75074615695593e-01, 4.68157918403314e-01, 4.61017496438114e-01, 4.53709515991115e-01, 4.46291754301615e-01, 4.38822149860570e-01, 4.31357461499186e-01, 4.23952086451610e-01, 4.16657071406007e-01, 4.09519334017869e-01, 4.02581096992957e-01, 3.95879523918712e-01, 3.89446536311937e-01, 3.83308785136485e-01, 3.77487747167450e-01, 3.71999916563979e-01, 3.66857064204329e-01, 3.62066541016937e-01, 3.57631606028512e-01, 3.53551764556578e-01, 3.49823106450398e-01, 3.46438638228097e-01, 3.43388606201394e-01, 3.40660810163378e-01, 3.38240908947334e-01, 3.36112720198024e-01, 3.34258517102271e-01, 3.32659324686112e-01, 3.31295217687668e-01, 3.30145621048004e-01, 3.29189612818531e-01, 3.28406227860359e-01, 3.27774759210662e-01, 3.27275052526048e-01, 3.26887787699402e-01, 3.26594740700415e-01, 3.26379018019608e-01, 3.26225255888186e-01, 3.26119776759032e-01, 3.26050696385266e-01, 3.26007976195477e-01, 3.25983417466940e-01, 3.25970595927583e-01, 3.25964737735745e-01, 3.25962540137933e-01, 3.25961942339191e-01, 3.25961854105397e-01, 3.25961847275728e-01, 3.25961724322958e-01, 3.25960891300522e-01, 3.25957829033097e-01, 3.25949666237928e-01, 3.25931802593981e-01, 3.25897593240359e-01, 3.25838108085790e-01, 3.25741978213408e-01, 3.25595340344996e-01, 3.25381888654512e-01, 3.25083041016573e-01, 3.24678223856971e-01, 3.24145275977930e-01, 3.23460966976313e-01, 3.22601620187054e-01, 3.21543823660711e-01, 3.20265205901229e-01, 3.18745246509263e-01, 3.16966086215299e-01, 3.14913296832078e-01, 3.12576570168506e-01, 3.09950286517417e-01, 3.07033928253453e-01, 3.03832312249412e-01, 3.00355625684038e-01, 2.96619262399130e-01, 2.92643469995731e-01, 2.88452829944202e-01, 2.84075602832271e-01, 2.79542977507550e-01, 2.74888265763924e-01, 2.70146083375380e-01, 2.65351554179446e-01, 2.60539567414213e-01, 2.55744110673547e-01, 2.50997692723974e-01, 2.46330862919979e-01, 2.41771827677773e-01, 2.37346159723351e-01, 2.33076592645052e-01, 2.28982891490947e-01, 2.25081789483184e-01, 2.21386981063787e-01, 2.17909162147922e-01, 2.14656109384293e-01, 2.11632791262229e-01, 2.08841504933748e-01, 2.06282033632366e-01, 2.03951820519798e-01, 2.01846155683498e-01, 1.99958373803128e-01, 1.98280060680403e-01, 1.96801267349230e-01, 1.95510730816544e-01, 1.94396100628616e-01, 1.93444170368135e-01, 1.92641112896176e-01, 1.91972717656114e-01, 1.91424627721617e-01, 1.90982573522330e-01, 1.90632599438594e-01, 1.90361278768979e-01, 1.90155912044954e-01, 1.90004703372594e-01, 1.89896909471969e-01, 1.89822956414375e-01, 1.89774519727489e-01, 1.89744564527057e-01, 1.89727343599312e-01, 1.89718352821085e-01, 1.89714244882406e-01, 1.89712703856596e-01, 1.89712284660356e-01, 1.89712222787964e-01, 1.89712217960032e-01, 1.89712130541089e-01, 1.89711538263715e-01, 1.89709360994929e-01, 1.89703557232253e-01, 1.89690856025400e-01, 1.89666532407092e-01, 1.89624235659934e-01, 1.89555878855685e-01, 1.89451597054994e-01, 1.89299780275757e-01, 1.89087185728024e-01, 1.88799131800145e-01, 1.88419773741625e-01, 1.87932457937478e-01, 1.87320148084783e-01, 1.86565912616743e-01, 1.85653458567188e-01, 1.84567693050841e-01, 1.83295290049486e-01, 1.81825237649850e-01, 1.80149339701270e-01, 1.78262646351218e-01, 1.76163790259380e-01, 1.73855209399260e-01, 1.71343243061564e-01, 1.68638094442468e-01, 1.65753660528551e-01, 1.62707237192389e-01, 1.59519113900185e-01, 1.56212077715452e-01, 1.52810849987144e-01, 1.49341481216418e-01, 1.45830729985462e-01, 1.42305450844880e-01, 1.38792013745171e-01, 1.35315774299094e-01, 1.31900610100296e-01, 1.28568533836383e-01, 1.25339389294417e-01, 1.22230631942603e-01, 1.19257191779358e-01, 1.16431412656560e-01, 1.13763059672738e-01, 1.11259384241730e-01, 1.08925235728766e-01, 1.06763208186724e-01, 1.04773811755461e-01, 1.02955659253499e-01, 1.01305660125259e-01, 9.98192156988367e-02, 9.84904113123726e-02, 9.73122027548463e-02, 9.62765955982121e-02, 9.53748174225311e-02, 9.45974836570436e-02, 9.39347577639751e-02, 9.33765072716319e-02, 9.29124559190413e-02, 9.25323325767131e-02, 9.22260162055828e-02, 9.19836757901393e-02, 9.17959031466040e-02, 9.16538358518366e-02, 9.15492671770934e-02, 9.14747394837576e-02, 9.14236174257610e-02, 9.13901379472293e-02, 9.13694343607254e-02, 9.13575325948024e-02, 9.13513190134907e-02, 9.13484800224289e-02, 9.13474150405271e-02, 9.13471253322520e-02, 9.13470825722591e-02, 9.13470792011530e-02, 9.13470179109304e-02, 9.13466026470466e-02, 9.13450760937451e-02, 9.13410069359344e-02, 9.13321019410596e-02, 9.13150489251180e-02, 9.12853970534719e-02, 9.12374810184922e-02, 9.11643948191951e-02, 9.10580198271198e-02, 9.09091114997729e-02, 9.07074468258955e-02, 9.04420333482605e-02, 9.01013780292317e-02, 8.96738111962421e-02, 8.91478576164459e-02, 8.85126432846910e-02, 8.77583226221468e-02, 8.68765081335034e-02, 8.58606814386782e-02, 8.47065639624063e-02, 8.34124249228232e-02, 8.19793066470567e-02, 8.04111504225382e-02, 7.87148107919329e-02, 7.68721390392813e-02, 7.49538090035600e-02, 7.29436543710733e-02, 7.08580307836684e-02, 6.87146826558198e-02, 6.65322232513859e-02, 6.43296005371461e-02, 6.21255764548856e-02, 5.99382448056289e-02, 5.77846086941847e-02, 5.56802329674966e-02, 5.36389808850046e-02, 5.16728379738538e-02, 4.97918202125334e-02, 4.80039588590131e-02, 4.63153507604410e-02, 4.47302609913400e-02, 4.32512640542413e-02, 4.18794103532143e-02, 4.06144058962248e-02, 3.94547949405712e-02, 3.83981373851441e-02, 3.74411749597314e-02, 3.65799824643354e-02, 3.58101022653408e-02, 3.51266618239876e-02, 3.45244751848703e-02, 3.39981301404784e-02, 3.35420632872250e-02, 3.31506254478953e-02, 3.28181399708818e-02, 3.25389562350757e-02, 3.23075003103526e-02, 3.21183241929504e-02, 3.19661544087140e-02, 3.18459401181634e-02, 3.17529002186205e-02, 3.16825683697779e-02, 3.16308344115548e-02, 3.15939803323104e-02, 3.15687088083259e-02, 3.15521623854806e-02, 3.15419316086434e-02, 3.15360508038509e-02, 3.15329807478690e-02, 3.15315780719594e-02, 3.15310518891820e-02, 3.15309087555988e-02, 3.15308876294118e-02, 3.15308859521579e-02, 3.15308552141929e-02, 3.15306469604977e-02, 3.15298814057279e-02, 3.15278407779935e-02, 3.15233752196455e-02, 3.15148243118948e-02, 3.14999578746864e-02, 3.14759395930402e-02, 3.14393168765577e-02, 3.13860400999602e-02, 3.13115139103992e-02, 3.12106825289078e-02, 3.10781498342969e-02, 3.09083334478117e-02, 3.06956500364102e-02, 3.04347266919152e-02, 3.01206306723145e-02, 2.97491072376310e-02, 2.93168130682098e-02, 2.88215311477948e-02, 2.82623523608514e-02, 2.76398096814816e-02, 2.69559529003622e-02, 2.62143553456404e-02, 2.54200487550094e-02, 2.45793878464617e-02, 2.36998515505320e-02, 2.27897926629943e-02, 2.18581514036025e-02, 2.09141507971538e-02, 1.99669927829547e-02, 1.90255732913371e-02, 1.80982319898000e-02, 1.71925481444468e-02, 1.63151888631532e-02, 1.54718111259307e-02, 1.46670154021322e-02, 1.39043463023789e-02, 1.31863338536569e-02, 1.25145670670221e-02, 1.18897899044583e-02, 1.13120095871686e-02, 1.07806088887230e-02, 1.02944567721970e-02, 9.85201390153279e-03, 9.45143032181761e-03, 9.09063251339148e-03, 8.76739735567256e-03, 8.47941195582180e-03, 8.22432027248610e-03, 7.99975887817439e-03, 7.80338441688500e-03, 7.63289462001282e-03, 7.48604394901164e-03, 7.36065462718827e-03, 7.25462402341392e-03, 7.16592968545626e-03, 7.09263342781341e-03, 7.03288564567472e-03, 6.98493057345928e-03, 6.94711272083737e-03, 6.91788432898477e-03, 6.89581342592049e-03, 6.87959189709630e-03, 6.86804288988796e-03, 6.86012682261961e-03, 6.85494527018449e-03, 6.85174205639782e-03, 6.84990100124594e-03, 6.84893994226599e-03, 6.84850085793087e-03, 6.84833614733193e-03, 6.84829134253206e-03, 6.84828472945398e-03, 6.84828419957029e-03, 6.84827442765013e-03, 6.84820822167609e-03, 6.84796484571848e-03, 6.84731612693668e-03, 6.84589658489898e-03, 6.84317861713524e-03, 6.83845400350336e-03, 6.83082301707927e-03, 6.81919244043883e-03, 6.80228374709912e-03, 6.77865257028287e-03, 6.74672029771628e-03, 6.70481816161228e-03, 6.65124352123314e-03, 6.58432717922253e-03, 6.50250958509702e-03, 6.40442273866526e-03, 6.28897359242941e-03, 6.15542383430778e-03, 6.00346018502051e-03, 5.83324891108461e-03, 5.64546841321700e-03, 5.44131488116051e-03, 5.22247837443501e-03, 4.99109008127618e-03, 4.74964501452035e-03, 4.50090670756609e-03, 4.24780074554576e-03, 3.99330276331611e-03, 3.74032572117855e-03, 3.49161244590945e-03, 3.24964200559563e-03, 3.01655940968004e-03, 2.79413443357157e-03, 2.58374777198611e-03, 2.38639627375292e-03, 2.20270894927665e-03, 2.03297159379545e-03, 1.87716369606861e-03, 1.73500995052223e-03, 1.60604050791235e-03, 1.48964722703937e-03, 1.38512552061698e-03, 1.29170196960970e-03, 1.20855727635316e-03, 1.13485379571188e-03, 1.06976770376122e-03, 1.01251693707738e-03, 9.62375491614154e-04, 9.18672295817657e-04, 8.80781462537783e-04, 8.48113469594125e-04, 8.20112854871269e-04, 7.96261705108492e-04, 7.76084422039247e-04, 7.59149595673136e-04, 7.45067629857488e-04, 7.33485423516811e-04, 7.24080427733753e-04, 7.16555887901649e-04, 7.10637965497085e-04, 7.06074531276357e-04, 7.02635033178467e-04, 7.00110861053815e-04, 6.98315815295558e-04, 6.97086459883423e-04, 6.96282237569790e-04, 6.95785259086534e-04, 6.95499685438250e-04, 6.95350630898872e-04, 6.95282535950525e-04, 6.95256992688797e-04, 6.95250044440598e-04, 6.95249018903610e-04, 6.95248936047221e-04, 6.95247399544741e-04, 6.95236989589176e-04, 6.95198722308277e-04, 6.95096722267863e-04, 6.94873530552087e-04, 6.94446218080458e-04, 6.93703515704414e-04, 6.92504188528176e-04, 6.90676904831378e-04, 6.88021881334322e-04, 6.84314623118354e-04, 6.79312105123842e-04, 6.72761721659754e-04, 6.64413181940660e-04, 6.54033144883215e-04, 6.41421688325039e-04, 6.26428770569731e-04, 6.08968030100243e-04, 5.89025242828001e-04, 5.66660180261636e-04, 5.42003492327581e-04, 5.15253244454783e-04, 4.86676278043774e-04, 4.56615359894386e-04, 4.25494880165189e-04, 3.93811083846174e-04, 3.62095374246252e-04, 3.30854724062135e-04, 3.00514034734078e-04, 2.71393412564436e-04, 2.43734174336652e-04, 2.17748436934784e-04, 1.93640936153372e-04, 1.71570370692820e-04, 1.51576442117829e-04, 1.33547059576244e-04, 1.17280078203663e-04, 1.02605759131791e-04, 8.94596405238701e-05, 7.78188414329861e-05, 6.75375374719067e-05, 5.82622869664817e-05, 4.94144837368538e-05, 4.05716436056025e-05, 3.16019033306247e-05, 2.27607060665491e-05, 1.45827998352388e-05, 7.63283660987213e-06, 2.13580411667635e-06, 2.13186595487969e-06, 1.76547816010952e-06, 5.80213505685379e-07, 1.73983854523385e-06, 2.05054524826664e-06, 1.91670368465680e-06, 1.62086609023504e-06, 1.35792490496556e-06, 1.21078857517887e-06, 1.15828169100496e-06, 1.13318944983121e-06, 1.08322583614040e-06, 9.89768519588097e-07, 8.58700914818655e-07, 7.06811075446012e-07, 5.52135293929723e-07, 4.08967984810715e-07, 2.86265930071750e-07, -3.55271367880050e-15, -3.55271367880050e-15, -7.10542735760100e-15, -3.55271367880050e-15, -7.10542735760100e-15, -3.55271367880050e-15, -7.10542735760100e-15],
[ 2.30000000000000e+01, 2.29999999851617e+01, 2.29999995263306e+01, 2.29999964176783e+01, 2.29999849898723e+01, 2.29999545269920e+01, 2.29998878569562e+01, 2.29997601652567e+01, 2.29995380751303e+01, 2.29991790293162e+01, 2.29986309994908e+01, 2.29978325396429e+01, 2.29967131893947e+01, 2.29951942229466e+01, 2.29931897292877e+01, 2.29906079999064e+01, 2.29873531917589e+01, 2.29833272259259e+01, 2.29784318763682e+01, 2.29725709985244e+01, 2.29656528441561e+01, 2.29575924067219e+01, 2.29483137405110e+01, 2.29377521966212e+01, 2.29258565195051e+01, 2.29125907491730e+01, 2.28979358762673e+01, 2.28818912002544e+01, 2.28644753451332e+01, 2.28457268926026e+01, 2.28257045998072e+01, 2.28044871777520e+01, 2.27821726172636e+01, 2.27588770617846e+01, 2.27347332399384e+01, 2.27098884850931e+01, 2.26845023833511e+01, 2.26587441046777e+01, 2.26327894834513e+01, 2.26068179238733e+01, 2.25810092118731e+01, 2.25555403180905e+01, 2.25305822761283e+01, 2.25062972167149e+01, 2.24828356320510e+01, 2.24603339359198e+01, 2.24389123746983e+01, 2.24186733327903e+01, 2.23997000637703e+01, 2.23820558661576e+01, 2.23657837106286e+01, 2.23509063139160e+01, 2.23374266438497e+01, 2.23253288301051e+01, 2.23145794463311e+01, 2.23051291214982e+01, 2.22969144315823e+01, 2.22898600171489e+01, 2.22838808680806e+01, 2.22788847136823e+01, 2.22747744547763e+01, 2.22714505742471e+01, 2.22688134638594e+01, 2.22667656080894e+01, 2.22652135701568e+01, 2.22640697313595e+01, 2.22632537420775e+01, 2.22626936512518e+01, 2.22623266905383e+01, 2.22620996994256e+01, 2.22619691880910e+01, 2.22619010453552e+01, 2.22618699094499e+01, 2.22618582291470e+01, 2.22618550518073e+01, 2.22618545828381e+01, 2.22618545213192e+01, 2.22618530880031e+01, 2.22618433770641e+01, 2.22618076784470e+01, 2.22617125176731e+01, 2.22615042537076e+01, 2.22611053767097e+01, 2.22604116430652e+01, 2.22592901617191e+01, 2.22575785192575e+01, 2.22550850021790e+01, 2.22515899441344e+01, 2.22468481945093e+01, 2.22405926734929e+01, 2.22325389486300e+01, 2.22223907395796e+01, 2.22098462319534e+01, 2.21946050579324e+01, 2.21763757807434e+01, 2.21548837016141e+01, 2.21298787910131e+01, 2.21011435304614e+01, 2.20685004371893e+01, 2.20318190324194e+01, 2.19910220071444e+01, 2.19460903398785e+01, 2.18970671324645e+01, 2.18440599558481e+01, 2.17872415399490e+01, 2.17268487005975e+01, 2.16631794697080e+01, 2.15965884776065e+01, 2.15274807217927e+01, 2.14563039363647e+01, 2.13835398429993e+01, 2.13096946113840e+01, 2.12352888804602e+01, 2.11608476909302e+01, 2.10868906563649e+01, 2.10139226594802e+01, 2.09424253077534e+01, 2.08728493249439e+01, 2.08056079981080e+01, 2.07410717479372e+01, 2.06795638466508e+01, 2.06213572735495e+01, 2.05666726735641e+01, 2.05156773675813e+01, 2.04684853532499e+01, 2.04251582294053e+01, 2.03857069743328e+01, 2.03500945062407e+01, 2.03182389523197e+01, 2.02900175498452e+01, 2.02652710985163e+01, 2.02438088775428e+01, 2.02254139341104e+01, 2.02098486421865e+01, 2.01968604228024e+01, 2.01861875096825e+01, 2.01775646381972e+01, 2.01707285319155e+01, 2.01654230603079e+01, 2.01614039440682e+01, 2.01584428915752e+01, 2.01563310614846e+01, 2.01548817622998e+01, 2.01539323197390e+01, 2.01533450661859e+01, 2.01530074326510e+01, 2.01528311514092e+01, 2.01527506056590e+01, 2.01527203899416e+01, 2.01527121705126e+01, 2.01527109573413e+01, 2.01527108502414e+01, 2.01527087516708e+01, 2.01526945335394e+01, 2.01526422660807e+01, 2.01525029394305e+01, 2.01521980224279e+01, 2.01516140544225e+01, 2.01505984807574e+01, 2.01489569137548e+01, 2.01464519678422e+01, 2.01428037797984e+01, 2.01376922827059e+01, 2.01307612547878e+01, 2.01216241117979e+01, 2.01098713543359e+01, 2.00950795202651e+01, 2.00768214289595e+01, 2.00546774408481e+01, 2.00282473959231e+01, 1.99971628425112e+01, 1.99610991271257e+01, 1.99197868922357e+01, 1.98730225255792e+01, 1.98206771256056e+01, 1.97627035945834e+01, 1.96991415434853e+01, 1.96301197879023e+01, 1.95558563260019e+01, 1.94766558094383e+01, 1.93929046359703e+01, 1.93050638978734e+01, 1.92136605039690e+01, 1.91192768490868e+01, 1.90225394309049e+01, 1.89241068124530e+01, 1.88246573047391e+01, 1.87248767056216e+01, 1.86254463860914e+01, 1.85270319700030e+01, 1.84302728119469e+01, 1.83357724416088e+01, 1.82440901106526e+01, 1.81557335478639e+01, 1.80711529980536e+01, 1.79907365890890e+01, 1.79148070397616e+01, 1.78436196905232e+01, 1.77773618114513e+01, 1.77161531190051e+01, 1.76600474163060e+01, 1.76090352609144e+01, 1.75630475584989e+01, 1.75219599788361e+01, 1.74855980904018e+01, 1.74537431096836e+01, 1.74261381599148e+01, 1.74024949304322e+01, 1.73825006221082e+01, 1.73658250567047e+01, 1.73521278193764e+01, 1.73410652950941e+01, 1.73322974528253e+01, 1.73254942272862e+01, 1.73203413482557e+01, 1.73165454728577e+01, 1.73138384875750e+01, 1.73119808643110e+01, 1.73107639783541e+01, 1.73100113249309e+01, 1.73095786040384e+01, 1.73093526789584e+01, 1.73092494506043e+01, 1.73092107258714e+01, 1.73092001917850e+01, 1.73091986369751e+01, 1.73091985122602e+01, 1.73091962106360e+01, 1.73091806167883e+01, 1.73091232920697e+01, 1.73089704847648e+01, 1.73086360664081e+01, 1.73079956025748e+01, 1.73068817909370e+01, 1.73050814687481e+01, 1.73023343565831e+01, 1.72983336648305e+01, 1.72927286435453e+01, 1.72851291043203e+01, 1.72751118844988e+01, 1.72622291595757e+01, 1.72460184399720e+01, 1.72260140154866e+01, 1.72017595375672e+01, 1.71728213600877e+01, 1.71388021982874e+01, 1.70993546180731e+01, 1.70541938390850e+01, 1.70031093292617e+01, 1.69459746894618e+01, 1.68827553758563e+01, 1.68135138852932e+01, 1.67384121326715e+01, 1.66577108753430e+01, 1.65717661811539e+01, 1.64810230848314e+01, 1.63860067205040e+01, 1.62873113429370e+01, 1.61855877432926e+01, 1.60815296161126e+01, 1.59758594375090e+01, 1.58693143725162e+01, 1.57626326522975e+01, 1.56565407649970e+01, 1.55517417045446e+01, 1.54489044338005e+01, 1.53486546502195e+01, 1.52515668952352e+01, 1.51581580191135e+01, 1.50688819949099e+01, 1.49841260623411e+01, 1.49042081706380e+01, 1.48293756769754e+01, 1.47598052439819e+01, 1.46956038673323e+01, 1.46368109538515e+01, 1.45834013627982e+01, 1.45352893181949e+01, 1.44923330976303e+01, 1.44543404018381e+01, 1.44210743082802e+01, 1.43922597098128e+01, 1.43675901354716e+01, 1.43467348441701e+01, 1.43293460738479e+01, 1.43150663190537e+01, 1.43035355002087e+01, 1.42943978792724e+01, 1.42873085707388e+01, 1.42819394952938e+01, 1.42779846273115e+01, 1.42751643975370e+01, 1.42732291292051e+01, 1.42719614093508e+01, 1.42711773264890e+01, 1.42707265400144e+01, 1.42704911840080e+01, 1.42703836467992e+01, 1.42703433057008e+01, 1.42703323319270e+01, 1.42703307122203e+01, 1.42703305872930e+01, 1.42703283439984e+01, 1.42703131453431e+01, 1.42702572734102e+01, 1.42701083388268e+01, 1.42697823964772e+01, 1.42691581677039e+01, 1.42680725967721e+01, 1.42663179400130e+01, 1.42636405521655e+01, 1.42597414955769e+01, 1.42542790534668e+01, 1.42468731777461e+01, 1.42371118445106e+01, 1.42245592264498e+01, 1.42087655219951e+01, 1.41892782081022e+01, 1.41656544102703e+01, 1.41374740139908e+01, 1.41043530813603e+01, 1.40659570905418e+01, 1.40220134893503e+01, 1.39723230517999e+01, 1.39167695507293e+01, 1.38553273113189e+01, 1.37880662878430e+01, 1.37151544055863e+01, 1.36368570259661e+01, 1.35535335187595e+01, 1.34656310536108e+01, 1.33736758462099e+01, 1.32782622055280e+01, 1.31800398206119e+01, 1.30796997928553e+01, 1.29779599578229e+01, 1.28755500468647e+01, 1.27731972126281e+01, 1.26716123867273e+01, 1.25714778577339e+01, 1.24734363611995e+01, 1.23780818699225e+01, 1.22859521716022e+01, 1.21975232307175e+01, 1.21132052580767e+01, 1.20333403584229e+01, 1.19582015943463e+01, 1.18879932917413e+01, 1.18228524146213e+01, 1.17628508508307e+01, 1.17079984704493e+01, 1.16582468412824e+01, 1.16134935073345e+01, 1.15735867541150e+01, 1.15383307974574e+01, 1.15074913395583e+01, 1.14808014371220e+01, 1.14579676223437e+01, 1.14386762088757e+01, 1.14225997030937e+01, 1.14094032273177e+01, 1.13987508476717e+01, 1.13903116866096e+01, 1.13837656903479e+01, 1.13788089160026e+01, 1.13751582033427e+01, 1.13725551026164e+01, 1.13707689433108e+01, 1.13695989488799e+01, 1.13688753287853e+01, 1.13684593105123e+01, 1.13682421089942e+01, 1.13681428672586e+01, 1.13681056381654e+01, 1.13680955109407e+01, 1.13680940161833e+01, 1.13680939030047e+01, 1.13680918980584e+01, 1.13680783142506e+01, 1.13680283787140e+01, 1.13678952687386e+01, 1.13676039599570e+01, 1.13670460655091e+01, 1.13660758721123e+01, 1.13645077540641e+01, 1.13621151173170e+01, 1.13586309922210e+01, 1.13537503543928e+01, 1.13471342074423e+01, 1.13384154082326e+01, 1.13272061548130e+01, 1.13131069898907e+01, 1.12957171006362e+01, 1.12746456219892e+01, 1.12495235800348e+01, 1.12200160501069e+01, 1.11858340573828e+01, 1.11467457221802e+01, 1.11025861534742e+01, 1.10532656261410e+01, 1.09987756414369e+01, 1.09391925644017e+01, 1.08746786510231e+01, 1.08054804136336e+01, 1.07319244141311e+01, 1.06544107088584e+01, 1.05734042841895e+01, 1.04894249077903e+01, 1.04030358703145e+01, 1.03148321036077e+01, 1.02254281367674e+01, 1.01354462973861e+01, 1.00455054917378e+01, 9.95621081554003e+00, 9.86814416657283e+00, 9.78185595993448e+00, 9.69785799092545e+00, 9.61661745088192e+00, 9.53855207623680e+00, 9.46402639725772e+00, 9.39334904618292e+00, 9.32677108108386e+00, 9.26448527902465e+00, 9.20662634857011e+00, 9.15327200718714e+00, 9.10444486410617e+00, 9.06011504452800e+00, 9.02020348727888e+00, 8.98458584531441e+00, 8.95309691657465e+00, 8.92553553096894e+00, 8.90166981695380e+00, 8.88124276758495e+00, 8.86397802068165e+00, 8.84958576084786e+00, 8.83776864299581e+00, 8.82822762855689e+00, 8.82066761787823e+00, 8.81480275668340e+00, 8.81036129220755e+00, 8.80708985682436e+00, 8.80475706448893e+00, 8.80315631854026e+00, 8.80210774828746e+00, 8.80145921580824e+00, 8.80108636253885e+00, 8.80089169619606e+00, 8.80080275077489e+00, 8.80076938415349e+00, 8.80076030761181e+00, 8.80075896793286e+00, 8.80075886736274e+00, 8.80075709721443e+00, 8.80074510419602e+00, 8.80070101654428e+00, 8.80058349468725e+00, 8.80032629912916e+00, 8.79983373178360e+00, 8.79897713087215e+00, 8.79759257656124e+00, 8.79547993799271e+00, 8.79240336190341e+00, 8.78809326806137e+00, 8.78224987669429e+00, 8.77454824758262e+00, 8.76464475962312e+00, 8.75218490421484e+00, 8.73681220742618e+00, 8.71817803716746e+00, 8.69595199602307e+00, 8.66983255217175e+00, 8.63955752441089e+00, 8.60491401691132e+00, 8.56574739830680e+00, 8.52196893997218e+00, 8.47356176993990e+00, 8.42058485998817e+00, 8.36317484047961e+00, 8.30154552593551e+00, 8.23598512919039e+00, 8.16685123878146e+00, 8.09456372924832e+00, 8.01959586410858e+00, 7.94246393319877e+00, 7.86371583558141e+00, 7.78391907045594e+00, 7.70364862428853e+00, 7.62347523570071e+00, 7.54395447623279e+00, 7.46561700593732e+00, 7.38896025566820e+00, 7.31444166687897e+00, 7.24247350194056e+00, 7.17341913962164e+00, 7.10759070258141e+00, 7.04524783032424e+00, 6.98659740869953e+00, 6.93179408733497e+00, 6.88094144906567e+00, 6.83409373101762e+00, 6.79125802864399e+00, 6.75239693789975e+00, 6.71743160573451e+00, 6.68624516583102e+00, 6.65868653655292e+00, 6.63457455312630e+00, 6.61370239772161e+00, 6.59584228056501e+00, 6.58075031345236e+00, 6.56817150485605e+00, 6.55784479394241e+00, 6.54950803002133e+00, 6.54290279505388e+00, 6.53777896070994e+00, 6.53389886893417e+00, 6.53104102677442e+00, 6.52900321287663e+00, 6.52760490479078e+00, 6.52668895294893e+00, 6.52612244837326e+00, 6.52579675599549e+00, 6.52562671275592e+00, 6.52554901803200e+00, 6.52551987196760e+00, 6.52551194352279e+00, 6.52551077330064e+00, 6.52551068571883e+00, 6.52550914773824e+00, 6.52549872769126e+00, 6.52546042246406e+00, 6.52535831454444e+00, 6.52513485260275e+00, 6.52470689143323e+00, 6.52396264828475e+00, 6.52275971296750e+00, 6.52092422636376e+00, 6.51825131567701e+00, 6.51450684430055e+00, 6.50943049891959e+00, 6.50274019590368e+00, 6.49413774316639e+00, 6.48331564311574e+00, 6.46996486865010e+00, 6.45378338995657e+00, 6.43448517874591e+00, 6.41180937293686e+00, 6.38552925352185e+00, 6.35546067105718e+00, 6.32146956562454e+00, 6.28347825317983e+00, 6.24147020250087e+00, 6.19549309727605e+00, 6.14566006141411e+00, 6.09214901466249e+00, 6.03520021168345e+00, 5.97511209330290e+00, 5.91223563845029e+00, 5.84696744724444e+00, 5.77974181081765e+00, 5.71102203516556e+00, 5.64129128875001e+00, 5.57104324026428e+00, 5.50077274568494e+00, 5.43096683235176e+00, 5.36209621085898e+00, 5.29460752123402e+00, 5.22891648721820e+00, 5.16540211192896e+00, 5.10440200189765e+00, 5.04620885785204e+00, 4.99106812365313e+00, 4.93917674331841e+00, 4.89068294304094e+00, 4.84568693226984e+00, 4.80424240561110e+00, 4.76635872466583e+00, 4.73200366413757e+00, 4.70110661724457e+00, 4.67356216911502e+00, 4.64923396100425e+00, 4.62795878081057e+00, 4.60955082496655e+00, 4.59380608243465e+00, 4.58050679294433e+00, 4.56942592904623e+00, 4.56033164579220e+00, 4.55299163402006e+00, 4.54717730471636e+00, 4.54266772423872e+00, 4.53925321474014e+00, 4.53673853220242e+00, 4.53494553699099e+00, 4.53371527932268e+00, 4.53290943456591e+00, 4.53241104048579e+00, 4.53212450959673e+00, 4.53197491355304e+00, 4.53190656163466e+00, 4.53188092042978e+00, 4.53187394539653e+00, 4.53187291589609e+00, 4.53187283881734e+00, 4.53187148489051e+00, 4.53186231183669e+00, 4.53182859071602e+00, 4.53173870255471e+00, 4.53154198433816e+00, 4.53116524475398e+00, 4.53051008938047e+00, 4.52945117858392e+00, 4.52783552232558e+00, 4.52548289391337e+00, 4.52218741880887e+00, 4.51772036424637e+00, 4.51183412015384e+00, 4.50426732159426e+00, 4.49475101816605e+00, 4.48301574773264e+00, 4.46879932250575e+00, 4.45185508764988e+00, 4.43196036958796e+00, 4.40892479681045e+00, 4.38259815398825e+00, 4.35287742395323e+00, 4.31971268425111e+00, 4.28311155694674e+00, 4.24314196218417e+00, 4.19993299606475e+00, 4.15367383843352e+00, 4.10461069132901e+00, 4.05304184803726e+00, 3.99931108890874e+00, 3.94379968606984e+00, 3.88691736801960e+00, 3.82909264115907e+00, 3.77076288476902e+00, 3.71236462751580e+00, 3.65432437863678e+00, 3.59705032956819e+00, 3.54092516807048e+00, 3.48630016431459e+00, 3.43349060466134e+00, 3.38277257107945e+00, 3.33438099795043e+00, 3.28850888711885e+00, 3.24530752811203e+00, 3.20488755321312e+00, 3.16732065473721e+00, 3.13264180161508e+00, 3.10085181090093e+00, 3.07192015368200e+00, 3.04578790094215e+00, 3.02237074055746e+00, 3.00156201970631e+00, 2.98323578608786e+00, 2.96724981556002e+00, 2.95344862273286e+00, 2.94166645472472e+00, 2.93173026711041e+00, 2.92346267577569e+00, 2.91668486990195e+00, 2.91121946079503e+00, 2.90689323002170e+00, 2.90353972965490e+00, 2.90100167863474e+00, 2.89913309346424e+00, 2.89780108956582e+00, 2.89688729218715e+00, 2.89628880292771e+00, 2.89591867951230e+00, 2.89570590171815e+00, 2.89559481439543e+00, 2.89554405810526e+00, 2.89552501773728e+00, 2.89551983830203e+00, 2.89551907382847e+00, 2.89551901644047e+00, 2.89551800636350e+00, 2.89551116294726e+00, 2.89548600588241e+00, 2.89541894682318e+00, 2.89527219194112e+00, 2.89499114732691e+00, 2.89450243511691e+00, 2.89371261804770e+00, 2.89250771813006e+00, 2.89075360049054e+00, 2.88829727511568e+00, 2.88496914616394e+00, 2.88058620978115e+00, 2.87495616651530e+00, 2.86788237373158e+00, 2.85916951804951e+00, 2.84862983996957e+00, 2.83608969576454e+00, 2.82139619948698e+00, 2.80442365522345e+00, 2.78507947118423e+00, 2.76330924696301e+00, 2.73910074622171e+00, 2.71248651022626e+00, 2.68354493188969e+00, 2.65239969164389e+00, 2.61921754964414e+00, 2.58420458581863e+00, 2.54760107147702e+00, 2.50967523507535e+00, 2.47071624302350e+00, 2.43102674906737e+00, 2.39091537063447e+00, 2.35068942859155e+00, 2.31064824200644e+00, 2.27107720782313e+00, 2.23224282417520e+00, 2.19438874287525e+00, 2.15773286812648e+00, 2.12246545989192e+00, 2.08874815492029e+00, 2.05671378752237e+00, 2.02646687548099e+00, 1.99808463234001e+00, 1.97161837333457e+00, 1.94709519563586e+00, 1.92451983166189e+00, 1.90387659451093e+00, 1.88513135509919e+00, 1.86823350979518e+00, 1.85311791416345e+00, 1.83970677217622e+00, 1.82791148056032e+00, 1.81763443470764e+00, 1.80877080587270e+00, 1.80121029945060e+00, 1.79483890134199e+00, 1.78954061426173e+00, 1.78519917893077e+00, 1.78169976710662e+00, 1.77893062512291e+00, 1.77678463883397e+00, 1.77516078438908e+00, 1.77396542481650e+00, 1.77311341055891e+00, 1.77252894325140e+00, 1.77214616629633e+00, 1.77190945302458e+00, 1.77177337302395e+00, 1.77170232892174e+00, 1.77166986872346e+00, 1.77165769185342e+00, 1.77165437945622e+00, 1.77165389055358e+00, 1.77165385369601e+00, 1.77165320288971e+00, 1.77164879358489e+00, 1.77163258456834e+00, 1.77158937769914e+00, 1.77149482264256e+00, 1.77131374652045e+00, 1.77099887904305e+00, 1.77049003773054e+00, 1.76971383071643e+00, 1.76858392569694e+00, 1.76700192200538e+00, 1.76485884765261e+00, 1.76203728372541e+00, 1.75841409436299e+00, 1.75386371179138e+00, 1.74826189348172e+00, 1.74148983416091e+00, 1.73343848174383e+00, 1.72401287657605e+00, 1.71313631137246e+00, 1.70075409854331e+00, 1.68683673523241e+00, 1.67138227617955e+00, 1.65441776062232e+00, 1.63599959007851e+00, 1.61621281526910e+00, 1.59516935731730e+00, 1.57300525439369e+00, 1.54987708375886e+00, 1.52595775506048e+00, 1.50143189973049e+00, 1.47649109146515e+00, 1.45132912439371e+00, 1.42613755105135e+00, 1.40110164561742e+00, 1.37639691385849e+00, 1.35218622473893e+00, 1.32861759411968e+00, 1.30582261178602e+00, 1.28391547146920e+00, 1.26299254061039e+00, 1.24313239237869e+00, 1.22439621613330e+00, 1.20682852284427e+00, 1.19045806743839e+00, 1.17529891906729e+00, 1.16135162146156e+00, 1.14860439760003e+00, 1.13703436488433e+00, 1.12660873810301e+00, 1.11728600715294e+00, 1.10901708439934e+00, 1.10174642249690e+00, 1.09541310738274e+00, 1.08995193301283e+00, 1.08529446436179e+00, 1.08137009343866e+00, 1.07810708986492e+00, 1.07543364326604e+00, 1.07327888976256e+00, 1.07157390968025e+00, 1.07025267873048e+00, 1.06925295083665e+00, 1.06851704795189e+00, 1.06799253099404e+00, 1.06763272665817e+00, 1.06739708743921e+00, 1.06725136662275e+00, 1.06716759602710e+00, 1.06712386150411e+00, 1.06710387913101e+00, 1.06709638310017e+00, 1.06709434400249e+00, 1.06709404303611e+00, 1.06709402022327e+00, 1.06709361577166e+00, 1.06709087555495e+00, 1.06708080226528e+00, 1.06705395085637e+00, 1.06699518864033e+00, 1.06688265746782e+00, 1.06668698202796e+00, 1.06637076470685e+00, 1.06588840313012e+00, 1.06518626105785e+00, 1.06420321614429e+00, 1.06287159863418e+00, 1.06111852283903e+00, 1.05886759795168e+00, 1.05604098650577e+00, 1.05256175815173e+00, 1.04835646453970e+00, 1.04335783966811e+00, 1.03750751122300e+00, 1.03075859459396e+00, 1.02307803474446e+00, 1.01444856384805e+00, 1.00487015568811e+00, 9.94360881264178e-01, 9.82957102595190e-01, 9.70712980869003e-01, 9.57699317388446e-01, 9.44001787190775e-01, 9.29718661760688e-01, 9.14958145478099e-01, 8.99835467940644e-01, 8.84469880007011e-01, 8.68981695641462e-01, 8.53489505955267e-01, 8.38107668697663e-01, 8.22944148788011e-01, 8.08098756296474e-01, 7.93661800253759e-01, 7.79713151911182e-01, 7.66321690994360e-01, 7.53545093799996e-01, 7.41429912753116e-01, 7.30011892854289e-01, 7.19316470556350e-01, 7.09359404088392e-01, 7.00147490116240e-01, 6.91679328962064e-01, 6.83946108586238e-01, 6.76932385486374e-01, 6.70616848073418e-01, 6.64973054566765e-01, 6.59970142764127e-01, 6.55573513048164e-01, 6.51745488639371e-01, 6.48445958415365e-01, 6.45633007669730e-01, 6.43263541113665e-01, 6.41293900412123e-01, 6.39680475820501e-01, 6.38380308314261e-01, 6.37351675281423e-01, 6.36554649690751e-01, 6.35951619970033e-01, 6.35507755914567e-01, 6.35191405029740e-01, 6.34974403947254e-01, 6.34832291001732e-01, 6.34744408661614e-01, 6.34693888122204e-01, 6.34667512740958e-01, 6.34655461815381e-01, 6.34650941130161e-01, 6.34649711397273e-01, 6.34649529891394e-01, 6.34649516046199e-01, 6.34649269431328e-01, 6.34647598580983e-01, 6.34641456384372e-01, 6.34625083742939e-01, 6.34589253630569e-01, 6.34520638560371e-01, 6.34401328522930e-01, 6.34208524388796e-01, 6.33914429267108e-01, 6.33486357096043e-01, 6.32887073439260e-01, 6.32075377683911e-01, 6.31006928244378e-01, 6.29635302783605e-01, 6.27913273921770e-01, 6.25794267779384e-01, 6.23233958733284e-01, 6.20191940049448e-01, 6.16633397993571e-01, 6.12530708171537e-01, 6.07864868741139e-01, 6.02626687011213e-01, 5.96817644528027e-01, 5.90450381043553e-01, 5.83548758936612e-01, 5.76147495081123e-01, 5.68291374542802e-01, 5.60034087197881e-01, 5.51436751767156e-01, 5.42566209586486e-01, 5.33493181118828e-01, 5.24290381066024e-01, 5.15030683203175e-01, 5.05785414807224e-01, 4.96622844469432e-01, 4.87606908183054e-01, 4.78796198932869e-01, 4.70243226413587e-01, 4.61993937399040e-01, 4.54087474607189e-01, 4.46556143083825e-01, 4.39425548134626e-01, 4.32714867294830e-01, 4.26437220130214e-01, 4.20600103089630e-01, 4.15205861442580e-01, 4.10252175860418e-01, 4.05732546869807e-01, 4.01636765784446e-01, 3.97951365497384e-01, 3.94660048499233e-01, 3.91744092580794e-01, 3.89182736856839e-01, 3.86953552039795e-01, 3.85032799359589e-01, 3.83395782257551e-01, 3.82017194084128e-01, 3.80871463624114e-01, 3.79933098498750e-01, 3.79177024504102e-01, 3.78578916907610e-01, 3.78115517813956e-01, 3.77764932101208e-01, 3.77506893277996e-01, 3.77322990053358e-01, 3.77196844530683e-01, 3.77114233769763e-01, 3.77063147976475e-01, 3.77033780689295e-01, 3.77018448890539e-01, 3.77011443795599e-01, 3.77008815963766e-01, 3.77008101131786e-01, 3.77007995624158e-01, 3.77007987519285e-01, 3.77007842408272e-01, 3.77006859260952e-01, 3.77003245124403e-01, 3.76993611293830e-01, 3.76972528553626e-01, 3.76932155121960e-01, 3.76861953326721e-01, 3.76748509878990e-01, 3.76575473197875e-01, 3.76323619394260e-01, 3.75971056004669e-01, 3.75493569128512e-01, 3.74865115045711e-01, 3.74058451553239e-01, 3.73045897186753e-01, 3.71800198420065e-01, 3.70295476329996e-01, 3.68508215782725e-01, 3.66418252839580e-01, 3.64009710798854e-01, 3.61271833058982e-01, 3.58199662592067e-01, 3.54794523684095e-01, 3.51064271675806e-01, 3.47023290119132e-01, 3.42692230913190e-01, 3.38097510056361e-01, 3.33270587897484e-01, 3.28247076487003e-01, 3.23065726421522e-01, 3.17767350553140e-01, 3.12393741840303e-01, 3.06986637823815e-01, 3.01586775634121e-01, 2.96233070345831e-01, 2.90961937307344e-01, 2.85806767132105e-01, 2.80797551446472e-01, 2.75960649005832e-01, 2.71318675793928e-01, 2.66890499238624e-01, 2.62691315465158e-01, 2.58732789138477e-01, 2.55023237413372e-01, 2.51567842305363e-01, 2.48368878968449e-01, 2.45425950558577e-01, 2.42736223318740e-01, 2.40294658080462e-01, 2.38094236456472e-01, 2.36126181575450e-01, 2.34380174297190e-01, 2.32844566481315e-01, 2.31506593112055e-01, 2.30352584956091e-01, 2.29368183002030e-01, 2.28538555252729e-01, 2.27848615572292e-01, 2.27283243292593e-01, 2.26827501229756e-01, 2.26466848728805e-01, 2.26187345427633e-01, 2.25975840694264e-01, 2.25820143220918e-01, 2.25709165120115e-01, 2.25633035101126e-01, 2.25583175925284e-01, 2.25552342324868e-01, 2.25534616872149e-01, 2.25525362822076e-01, 2.25521134623524e-01, 2.25519548488904e-01, 2.25519117022642e-01, 2.25519053339159e-01, 2.25519048409794e-01, 2.25518959667589e-01, 2.25518358426648e-01, 2.25516148209262e-01, 2.25510256641989e-01, 2.25497363399622e-01, 2.25472672480052e-01, 2.25429738461784e-01, 2.25360355598962e-01, 2.25254517652441e-01, 2.25100455142723e-01, 2.24884755118019e-01, 2.24592566424189e-01, 2.24207890729691e-01, 2.23713956135789e-01, 2.23093666114082e-01, 2.22330111898572e-01, 2.21407131592354e-01, 2.20309894533955e-01, 2.19025485419852e-01, 2.17543459863517e-01, 2.15856342025511e-01, 2.13960036091144e-01, 2.11854126900903e-01, 2.09542050865871e-01, 2.07031126019572e-01, 2.04332438972656e-01, 2.01460595747996e-01, 1.98433352008017e-01, 1.95271145153644e-01, 1.91996555505455e-01, 1.88633725926469e-01, 1.85207768824874e-01, 1.81744186821437e-01, 1.78268329058213e-01, 1.74804899859051e-01, 1.71377530916708e-01, 1.68008422960888e-01, 1.64718058367193e-01, 1.61524982625110e-01, 1.58445650044492e-01, 1.55494327471761e-01, 1.52683048958043e-01, 1.50021614093472e-01, 1.47517622923214e-01, 1.45176540850926e-01, 1.43001787601225e-01, 1.40994845080435e-01, 1.39155379785787e-01, 1.37481376228259e-01, 1.35969278616919e-01, 1.34614138772914e-01, 1.33409768864123e-01, 1.32348898047706e-01, 1.31423332449653e-01, 1.30624118077211e-01, 1.29941706244317e-01, 1.29366120895629e-01, 1.28887126861592e-01, 1.28494397600303e-01, 1.28177680426237e-01, 1.27926956656882e-01, 1.27732593577718e-01, 1.27585484712295e-01, 1.27477174630819e-01, 1.27399964498114e-01, 1.27346994767450e-01, 1.27312301888384e-01, 1.27290846594043e-01, 1.27278512234984e-01, 1.27272072676252e-01, 1.27269130408898e-01, 1.27268026666108e-01, 1.27267726421589e-01, 1.27267682106140e-01, 1.27267678648135e-01, 1.27267616035152e-01, 1.27267191823300e-01, 1.27265632379501e-01, 1.27261475504625e-01, 1.27252378437031e-01, 1.27234957043992e-01, 1.27204662909456e-01, 1.27155704404331e-01, 1.27081017093268e-01, 1.26972287928123e-01, 1.26820036539957e-01, 1.26613755490041e-01, 1.26342109519861e-01, 1.25993191612508e-01, 1.25554831073774e-01, 1.25014945930580e-01, 1.24361928895773e-01, 1.23585053187455e-01, 1.22674881903915e-01, 1.21623662763909e-01, 1.20425689120154e-01, 1.19077608493885e-01, 1.17578661575749e-01, 1.15930837691401e-01, 1.14138936949981e-01, 1.12210534367811e-01, 1.10155846744139e-01, 1.07987508487646e-01, 1.05720267480713e-01, 1.03370616040554e-01, 1.00956374819116e-01, 9.84962489650343e-02, 9.60093760356138e-02, 9.35148841457121e-02, 9.10314768295635e-02, 8.85770583494256e-02, 8.61684099408784e-02, 8.38209239983527e-02, 8.15483997199991e-02, 7.93629004598984e-02, 7.72746701632663e-02, 7.52921038956949e-02, 7.34217657681029e-02, 7.16684464473580e-02, 7.00352520182932e-02, 6.85237160211898e-02, 6.71339271191904e-02, 6.58646657570507e-02, 6.47135443604796e-02, 6.36771469231299e-02, 6.27511650878532e-02, 6.19305290266041e-02, 6.12095324232698e-02, 6.05819516402448e-02, 6.00411596404768e-02, 5.95802356253081e-02, 5.91920711622897e-02, 5.88694736247675e-02, 5.86052671594430e-02, 5.83923910551256e-02, 5.82239947228018e-02, 5.80935279690706e-02, 5.79948247175963e-02, 5.79221779521824e-02, 5.78704034457900e-02, 5.78348898137584e-02, 5.78116326065157e-02, 5.77972505966038e-02, 5.77889829425438e-02, 5.77846666470982e-02, 5.77826945345876e-02, 5.77819547330485e-02, 5.77817534884701e-02, 5.77817237866576e-02, 5.77817214499134e-02, 5.77816788719048e-02, 5.77813904065145e-02, 5.77803299823216e-02, 5.77775033263208e-02, 5.77713174746286e-02, 5.77594716596543e-02, 5.77388742954490e-02, 5.77055904812589e-02, 5.76548240555255e-02, 5.75809377285061e-02, 5.74775140628851e-02, 5.73374591563933e-02, 5.71531495034598e-02, 5.69166208306058e-02, 5.66197956607759e-02, 5.62547439774477e-02, 5.58139687905914e-02, 5.52907059323466e-02, 5.46792250413901e-02, 5.39751170469742e-02, 5.31755524270849e-02, 5.22794946459797e-02, 5.12878545259845e-02, 5.02035737217099e-02, 4.90316292215738e-02, 4.77789554109442e-02, 4.64542852673873e-02, 4.50679175688578e-02, 4.36314218260918e-02, 4.21572966661579e-02, 4.06586000908860e-02, 3.91485714080595e-02, 3.76402642557814e-02, 3.61341758393614e-02, 3.46677431861231e-02, 3.32377143118912e-02, 3.18535171640164e-02, 3.05232244094711e-02, 2.92534948090228e-02, 2.80495697099070e-02, 2.69153167160567e-02, 2.58533110681434e-02, 2.48649446902043e-02, 2.39505531027788e-02, 2.31095512735422e-02, 2.23405707777822e-02, 2.16415921865156e-02, 2.10100682306766e-02, 2.04430348710183e-02, 1.99372088296883e-02, 1.94890713384126e-02, 1.90949387963428e-02, 1.87510217038067e-02, 1.84534736655557e-02, 1.81984324655318e-02, 1.79820552334853e-02, 1.78005495811924e-02, 1.76502023111986e-02, 1.75274069260389e-02, 1.74286907233032e-02, 1.73507417887571e-02, 1.72904357306649e-02, 1.72448615708127e-02, 1.72113458536671e-02, 1.71874737838031e-02, 1.71711060728228e-02, 1.71603901820099e-02, 1.71537647830284e-02, 1.71499565132684e-02, 1.71479684495637e-02, 1.71470601316628e-02, 1.71467193975090e-02, 1.71466267102431e-02, 1.71466130298192e-02, 1.71466119437049e-02, 1.71465920391896e-02, 1.71464571835855e-02, 1.71459614460652e-02, 1.71446400374577e-02, 1.71417483984939e-02, 1.71362114508433e-02, 1.71265854275191e-02, 1.71110346643637e-02, 1.70873256722928e-02, 1.70528406700221e-02, 1.70046124736416e-02, 1.69393821437822e-02, 1.68536800213985e-02, 1.67439296998495e-02, 1.66065730819760e-02, 1.64382130036635e-02, 1.62357680779159e-02, 1.59966325919356e-02, 1.57188326913769e-02, 1.54011689559113e-02, 1.50433350463430e-02, 1.46460025794646e-02, 1.42108638658165e-02, 1.37406266132523e-02, 1.32389580011427e-02, 1.27103793794454e-02, 1.21601168453545e-02, 1.15939166347161e-02, 1.10178371746272e-02, 1.04380313995946e-02, 9.86053332492887e-03, 9.29106190002926e-03, 8.73485303539212e-03, 8.19652774302426e-03, 7.68000091776955e-03, 7.18843177131845e-03, 6.72421365257847e-03, 6.28899829145269e-03, 5.88374768529576e-03, 5.50880604384218e-03, 5.16398431759058e-03, 4.84865055594952e-03, 4.56182035254571e-03, 4.30224274446189e-03, 4.06847810349475e-03, 3.85896579400224e-03, 3.67208059403978e-03, 3.50617792302543e-03, 3.35962865027994e-03, 3.23084465141221e-03, 3.11829642907213e-03, 3.02052414731691e-03, 2.93614342322002e-03, 2.86384718550536e-03, 2.80240482874916e-03, 2.75065975166555e-03, 2.70752618541437e-03, 2.67198601766339e-03, 2.64308612192110e-03, 2.61993651808012e-03, 2.60170951966643e-03, 2.58763986524357e-03, 2.57702568902118e-03, 2.56923006485721e-03, 2.56368276586016e-03, 2.55988182621181e-03, 2.55739447576175e-03, 2.55585704261918e-03, 2.55497348073774e-03, 2.55451227157621e-03, 2.55430156178704e-03, 2.55422252059390e-03, 2.55420101978387e-03, 2.55419784632238e-03, 2.55419759206887e-03, 2.55419290347447e-03, 2.55416113773066e-03, 2.55404436615692e-03, 2.55373311653173e-03, 2.55305205839207e-03, 2.55174814927273e-03, 2.54948188753090e-03, 2.54582235263712e-03, 2.54024674310926e-03, 2.53214512450839e-03, 2.52083104478018e-03, 2.50555853104828e-03, 2.48554572310411e-03, 2.46000500513688e-03, 2.42817897093772e-03, 2.38938092072516e-03, 2.34303789524759e-03, 2.28873358274839e-03, 2.22624789035919e-03, 2.15558966536733e-03, 2.07701909285518e-03, 1.99105677083011e-03, 1.89847740400140e-03, 1.80028741405280e-03, 1.69768737173470e-03, 1.59202175018791e-03, 1.48471977832187e-03, 1.37723192707284e-03, 1.27096676957805e-03, 1.16723277465880e-03, 1.06718918268101e-03, 9.71809437942284e-04, 8.81859461802748e-04, 7.97891182237720e-04, 7.20249509836179e-04, 6.49089232634736e-04, 5.84397953435456e-04, 5.26022274147664e-04, 4.73695941373364e-04, 4.27069276543790e-04, 3.85738467933550e-04, 3.49272128506328e-04, 3.17232402593959e-04, 2.89189420048075e-04, 2.64730126549040e-04, 2.43463870280181e-04, 2.25026776007553e-04, 2.09085477642645e-04, 1.95339546042561e-04, 1.83521804093090e-04, 1.73396467989306e-04, 1.64755872027911e-04, 1.57416804160704e-04, 1.51217178650853e-04, 1.46013275784895e-04, 1.41677441334309e-04, 1.38096065871451e-04, 1.35167753036569e-04, 1.32852169617337e-04, 1.30966735365945e-04, 1.29487998293598e-04, 1.28350069868333e-04, 1.27493683166335e-04, 1.26865859240155e-04, 1.26419708724512e-04, 1.26114305434299e-04, 1.25914578653408e-04, 1.25791179286239e-04, 1.25720279839925e-04, 1.25683276348099e-04, 1.25666371967981e-04, 1.25660031045527e-04, 1.25658306224352e-04, 1.25658051629784e-04, 1.25658031070230e-04, 1.25657650077216e-04, 1.25655068721642e-04, 1.25645579739597e-04, 1.25620288017103e-04, 1.25564950060664e-04, 1.25459018473606e-04, 1.25274950669052e-04, 1.24977847512753e-04, 1.24525496328687e-04, 1.23868900910651e-04, 1.22953387226232e-04, 1.21720370430012e-04, 1.20109850254124e-04, 1.18021257137002e-04, 1.15491055108663e-04, 1.12431194366991e-04, 1.08812430895000e-04, 1.04622546267308e-04, 9.98687462576697e-05, 9.45786664807713e-05, 8.88000371261910e-05, 8.25994913604688e-05, 7.60938471131567e-05, 6.93238984652567e-05, 6.24299137541584e-05, 5.55357448810412e-05, 4.87603068570763e-05, 4.22068823056065e-05, 3.59520654491519e-05, 3.00395829775368e-05, 2.44870077672488e-05, 1.93177313541071e-05, 1.45687472610234e-05, 1.03205949137930e-05, 6.78037633150552e-06, 4.41538732331992e-06, 3.57064296352405e-06, 3.13777765015288e-06, 2.41929306099564e-06, 1.53546898928880e-06, 7.25299095449827e-07, 2.95099159330903e-07, 0.00000000000000e+00, 4.20106292864375e-07, 4.47297605177255e-07, -3.55271367880050e-15, 4.14300689755009e-07, 5.00324532026752e-07, 4.21879164491656e-07, -3.55271367880050e-15, 3.32554279225405e-07, 4.18690227377283e-07, 4.08976429611130e-07, 3.51004832310764e-07, 2.82014671171282e-07, 0.00000000000000e+00, -3.55271367880050e-15, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, -3.55271367880050e-15, 0.00000000000000e+00, -3.55271367880050e-15, 0.00000000000000e+00, 0.00000000000000e+00, 3.55271367880050e-15, 0.00000000000000e+00, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00],
[ 2.40000000000000e+01, 2.39999999842000e+01, 2.39999994956341e+01, 2.39999961855226e+01, 2.39999840171267e+01, 2.39999515800688e+01, 2.39998805894060e+01, 2.39997446224701e+01, 2.39995081393889e+01, 2.39991258246932e+01, 2.39985422777672e+01, 2.39976920694240e+01, 2.39965001709682e+01, 2.39948827511150e+01, 2.39927483254609e+01, 2.39899992332143e+01, 2.39865334068988e+01, 2.39822463929942e+01, 2.39770335751140e+01, 2.39707925464038e+01, 2.39634255743223e+01, 2.39548420987217e+01, 2.39449612030012e+01, 2.39337139978901e+01, 2.39210458579884e+01, 2.39069184524990e+01, 2.38913115136762e+01, 2.38742242895668e+01, 2.38556766318843e+01, 2.38357096756567e+01, 2.38143860748982e+01, 2.37917897681706e+01, 2.37680252595396e+01, 2.37432164138866e+01, 2.37175047803700e+01, 2.36910474733690e+01, 2.36640146556728e+01, 2.36365866830667e+01, 2.36089509819554e+01, 2.35812987414526e+01, 2.35538215079064e+01, 2.35267077727889e+01, 2.35001396442125e+01, 2.34742896882420e+01, 2.34493180190705e+01, 2.34253697075653e+01, 2.34025725663150e+01, 2.33810353567522e+01, 2.33608464507965e+01, 2.33420729662664e+01, 2.33247603824662e+01, 2.33089326301548e+01, 2.32945926387713e+01, 2.32817233134452e+01, 2.32702889050400e+01, 2.32602367283195e+01, 2.32514991763297e+01, 2.32439959733001e+01, 2.32376366038639e+01, 2.32323228532377e+01, 2.32279513912989e+01, 2.32244163333118e+01, 2.32216117114709e+01, 2.32194337944724e+01, 2.32177831969937e+01, 2.32165667271834e+01, 2.32156989279375e+01, 2.32151032766567e+01, 2.32147130181412e+01, 2.32144716159792e+01, 2.32143328189336e+01, 2.32142603500811e+01, 2.32142272374851e+01, 2.32142148156489e+01, 2.32142114365933e+01, 2.32142109378512e+01, 2.32142108724268e+01, 2.32142093481156e+01, 2.32141990206705e+01, 2.32141610557040e+01, 2.32140598536142e+01, 2.32138383680707e+01, 2.32134141689960e+01, 2.32126763963782e+01, 2.32114837264976e+01, 2.32096634440357e+01, 2.32070116828034e+01, 2.32032948653147e+01, 2.31982523380705e+01, 2.31916001661675e+01, 2.31830360186545e+01, 2.31722450457520e+01, 2.31589066212068e+01, 2.31427017979285e+01, 2.31233213025271e+01, 2.31004738740187e+01, 2.30738947332711e+01, 2.30433539524215e+01, 2.30086644778240e+01, 2.29696895473148e+01, 2.29263492350933e+01, 2.28786258586263e+01, 2.28265679955352e+01, 2.27702928879133e+01, 2.27099870591522e+01, 2.26459050341154e+01, 2.25783661345661e+01, 2.25077494124942e+01, 2.24344868765258e+01, 2.23590552519918e+01, 2.22819665849450e+01, 2.22037580477355e+01, 2.21249813249070e+01, 2.20461919528199e+01, 2.19679389574540e+01, 2.18907550877411e+01, 2.18151478833733e+01, 2.17415917533915e+01, 2.16705211812111e+01, 2.16023251177742e+01, 2.15373425800505e+01, 2.14758594381611e+01, 2.14181063504922e+01, 2.13642577907961e+01, 2.13144321024049e+01, 2.12686925101289e+01, 2.12270490181917e+01, 2.11894611210487e+01, 2.11558412519461e+01, 2.11260588908851e+01, 2.10999452488984e+01, 2.10772984392386e+01, 2.10578890384787e+01, 2.10414659321349e+01, 2.10277623308861e+01, 2.10165018355700e+01, 2.10074044227212e+01, 2.10001922183489e+01, 2.09945949267400e+01, 2.09903547840317e+01, 2.09872309136311e+01, 2.09850029725674e+01, 2.09834739945282e+01, 2.09824723563484e+01, 2.09818528194586e+01, 2.09814966253770e+01, 2.09813106536152e+01, 2.09812256801311e+01, 2.09811938034090e+01, 2.09811851321455e+01, 2.09811838522844e+01, 2.09811837392970e+01, 2.09811815253649e+01, 2.09811665256418e+01, 2.09811113849672e+01, 2.09809643993623e+01, 2.09806427208429e+01, 2.09800266522150e+01, 2.09789552543174e+01, 2.09772234578773e+01, 2.09745808372461e+01, 2.09707321636649e+01, 2.09653398110090e+01, 2.09580280368802e+01, 2.09483891062936e+01, 2.09359911643650e+01, 2.09203876992833e+01, 2.09011283692077e+01, 2.08777708992438e+01, 2.08498936908918e+01, 2.08171087306910e+01, 2.07790743420920e+01, 2.07355072998824e+01, 2.06861938243750e+01, 2.06309989965673e+01, 2.05698741873303e+01, 2.05028621726086e+01, 2.04300997089096e+01, 2.03518174621996e+01, 2.02683373093199e+01, 2.01800671532262e+01, 2.00874935008451e+01, 1.99911721360889e+01, 1.98917172749934e+01, 1.97897896138491e+01, 1.96860836777875e+01, 1.95813148531031e+01, 1.94762064497830e+01, 1.93714770989885e+01, 1.92678287492438e+01, 1.91659354875186e+01, 1.90664333771524e+01, 1.89699114716916e+01, 1.88769041298176e+01, 1.87878847200864e+01, 1.87032607652930e+01, 1.86233705366633e+01, 1.85484810706308e+01, 1.84787875486449e+01, 1.84144139555317e+01, 1.83554149153008e+01, 1.83017785944557e+01, 1.82534305601445e+01, 1.82102384815850e+01, 1.81720175656782e+01, 1.81385366195382e+01, 1.81095246323422e+01, 1.80846777657451e+01, 1.80636666360674e+01, 1.80461437631539e+01, 1.80317510512499e+01, 1.80201271577927e+01, 1.80109145981778e+01, 1.80037664298232e+01, 1.79983523585992e+01, 1.79943641160139e+01, 1.79915199671775e+01, 1.79895682277905e+01, 1.79882896929812e+01, 1.79874989110218e+01, 1.79870442696065e+01, 1.79868068999219e+01, 1.79866984424069e+01, 1.79866577560309e+01, 1.79866466883296e+01, 1.79866450547594e+01, 1.79866449237268e+01, 1.79866425055116e+01, 1.79866261217423e+01, 1.79865658931889e+01, 1.79864053453316e+01, 1.79860539869911e+01, 1.79853810811601e+01, 1.79842108529228e+01, 1.79823193473320e+01, 1.79794331131761e+01, 1.79752298463383e+01, 1.79693410783216e+01, 1.79613569408803e+01, 1.79508329761836e+01, 1.79372988937506e+01, 1.79202691015202e+01, 1.78992547609079e+01, 1.78737770376918e+01, 1.78433811463133e+01, 1.78076507197775e+01, 1.77662219864087e+01, 1.77187972039032e+01, 1.76651567954863e+01, 1.76051696564889e+01, 1.75388011546131e+01, 1.74661184337133e+01, 1.73872927467445e+01, 1.73025986832924e+01, 1.72124103123171e+01, 1.71171944196718e+01, 1.70175011683252e+01, 1.69139526317936e+01, 1.68072297345710e+01, 1.66980581687091e+01, 1.65871938420723e+01, 1.64754083585504e+01, 1.63634749478162e+01, 1.62521551692430e+01, 1.61421866268149e+01, 1.60342718594214e+01, 1.59290685171615e+01, 1.58271808967860e+01, 1.57291528825751e+01, 1.56354623168885e+01, 1.55465168031623e+01, 1.54626509216390e+01, 1.53841248151693e+01, 1.53111240808529e+01, 1.52437608851010e+01, 1.51820762063403e+01, 1.51260431014615e+01, 1.50755708886772e+01, 1.50305101393639e+01, 1.49906583730022e+01, 1.49557663507555e+01, 1.49255448631050e+01, 1.48996719043205e+01, 1.48778001210619e+01, 1.48595644143342e+01, 1.48445895641949e+01, 1.48324977363002e+01, 1.48229157201401e+01, 1.48154817423237e+01, 1.48098516961428e+01, 1.48057046322195e+01, 1.48027473653067e+01, 1.48007180697019e+01, 1.47993887601347e+01, 1.47985665856882e+01, 1.47980939000652e+01, 1.47978471106515e+01, 1.47977343494067e+01, 1.47976920485891e+01, 1.47976805417239e+01, 1.47976788433341e+01, 1.47976787123379e+01, 1.47976763600671e+01, 1.47976604230817e+01, 1.47976018369700e+01, 1.47974456673867e+01, 1.47971038914381e+01, 1.47964493395113e+01, 1.47953110366213e+01, 1.47934711517030e+01, 1.47906637278809e+01, 1.47865753261471e+01, 1.47808476683709e+01, 1.47730823124460e+01, 1.47628473321892e+01, 1.47496859075038e+01, 1.47331266572814e+01, 1.47126954706677e+01, 1.46879285149151e+01, 1.46583860245357e+01, 1.46236664121297e+01, 1.45834201918295e+01, 1.45373631773131e+01, 1.44852884124545e+01, 1.44270763170351e+01, 1.43627025835907e+01, 1.42922434431740e+01, 1.42158780239943e+01, 1.41338876519970e+01, 1.40466520793215e+01, 1.39546427670789e+01, 1.38584134843437e+01, 1.37585886069601e+01, 1.36558495995058e+01, 1.35509202344227e+01, 1.34445511386492e+01, 1.33375042574568e+01, 1.32305377882598e+01, 1.31243920681514e+01, 1.30197768053039e+01, 1.29173599359768e+01, 1.28177582765756e+01, 1.27215300343074e+01, 1.26291691489237e+01, 1.25411013674107e+01, 1.24576819056267e+01, 1.23791945251803e+01, 1.23058518472549e+01, 1.22377967331753e+01, 1.21751045792091e+01, 1.21177863955719e+01, 1.20657925627269e+01, 1.20190171787104e+01, 1.19773029272896e+01, 1.19404464071556e+01, 1.19082038667322e+01, 1.18802972878053e+01, 1.18564207547610e+01, 1.18362470357700e+01, 1.18194342890109e+01, 1.18056327923900e+01, 1.17944915806933e+01, 1.17856648612690e+01, 1.17788180697076e+01, 1.17736334219983e+01, 1.17698148204634e+01, 1.17670919782710e+01, 1.17652236418767e+01, 1.17639998122502e+01, 1.17632428935202e+01, 1.17628077305593e+01, 1.17625805334175e+01, 1.17624767245141e+01, 1.17624377821049e+01, 1.17624271888154e+01, 1.17624256252677e+01, 1.17624255068806e+01, 1.17624234096646e+01, 1.17624092007158e+01, 1.17623569670881e+01, 1.17622177311923e+01, 1.17619130157132e+01, 1.17613294447993e+01, 1.17603145965670e+01, 1.17586742957091e+01, 1.17561715045472e+01, 1.17525269355608e+01, 1.17474214673137e+01, 1.17405003977843e+01, 1.17313795136341e+01, 1.17196528908081e+01, 1.17049022718276e+01, 1.16867077902759e+01, 1.16646597366968e+01, 1.16383709872420e+01, 1.16074896528679e+01, 1.15717114593121e+01, 1.15307913430488e+01, 1.14845537515852e+01, 1.14329011715951e+01, 1.13758204765275e+01, 1.13133867841276e+01, 1.12457646378219e+01, 1.11732064648545e+01, 1.10960484067079e+01, 1.10147037508810e+01, 1.09296543053960e+01, 1.08414401387563e+01, 1.07506481527312e+01, 1.06578999623568e+01, 1.05638395308374e+01, 1.04691209544862e+01, 1.03743967247128e+01, 1.02803067209671e+01, 1.01874681195741e+01, 1.00964663447326e+01, 1.00078471422460e+01, 9.92210982326916e+00, 9.83970170164174e+00, 9.76101373048198e+00, 9.68637732817595e+00, 9.61606236849414e+00, 9.55027629360799e+00, 9.48916429292892e+00, 9.43281047637636e+00, 9.38123995940065e+00, 9.33442176983088e+00, 9.29227248354548e+00, 9.25466049637697e+00, 9.22141084222739e+00, 9.19231047046561e+00, 9.16711389775760e+00, 9.14554914937896e+00, 9.12732390216881e+00, 9.11213173562671e+00, 9.09965838983905e+00, 9.08958792003815e+00, 9.08160862905824e+00, 9.07541865231205e+00, 9.07073106668748e+00, 9.06727839626143e+00, 9.06481639490022e+00, 9.06312699915330e+00, 9.06202036431301e+00, 9.06133592154678e+00, 9.06094242355596e+00, 9.06073697884530e+00, 9.06064310870734e+00, 9.06060789465604e+00, 9.06059831556877e+00, 9.06059690171514e+00, 9.06059679557670e+00, 9.06059492741972e+00, 9.06058227037830e+00, 9.06053574171005e+00, 9.06041171301725e+00, 9.06014027749965e+00, 9.05962043927894e+00, 9.05871641518753e+00, 9.05725521537794e+00, 9.05502564132492e+00, 9.05177880784677e+00, 9.04723025904639e+00, 9.04106370593472e+00, 9.03293636554051e+00, 9.02248582760113e+00, 9.00933831625291e+00, 8.99311815217348e+00, 8.97345815802415e+00, 8.95001069036208e+00, 8.92245892870631e+00, 8.89052801173645e+00, 8.85399558609100e+00, 8.81270132860073e+00, 8.76655502050218e+00, 8.71554279311921e+00, 8.65973122790079e+00, 8.59926907726404e+00, 8.53438647292155e+00, 8.46539160091470e+00, 8.39266494234980e+00, 8.31665129987127e+00, 8.23784994493430e+00, 8.15680332105306e+00, 8.07408481310189e+00, 7.99028613203551e+00, 7.90600485996314e+00, 7.82183264950966e+00, 7.73834447857773e+00, 7.65608923978612e+00, 7.57558181165658e+00, 7.49729663616222e+00, 7.42166273101786e+00, 7.34906000411767e+00, 7.27981671217883e+00, 7.21420790918277e+00, 7.15245475204397e+00, 7.09472456023793e+00, 7.04113155436040e+00, 6.99173822069716e+00, 6.94655726303640e+00, 6.90555410964450e+00, 6.86864994419539e+00, 6.83572522636396e+00, 6.80662366228869e+00, 6.78115657815702e+00, 6.75910764225677e+00, 6.74023787215302e+00, 6.72429085429522e+00, 6.71099809352983e+00, 6.70008440009767e+00, 6.69127321237964e+00, 6.68429174577886e+00, 6.67887585267739e+00, 6.67477447637822e+00, 6.67175358422228e+00, 6.66959947230364e+00, 6.66812134670926e+00, 6.66715310390046e+00, 6.66655425523041e+00, 6.66620996675872e+00, 6.66603021426135e+00, 6.66594808320128e+00, 6.66591727289848e+00, 6.66590889173901e+00, 6.66590765469707e+00, 6.66590756211432e+00, 6.66590593631489e+00, 6.66589492128179e+00, 6.66585442881464e+00, 6.66574649045461e+00, 6.66551026842352e+00, 6.66505786895661e+00, 6.66427112365230e+00, 6.66299948349071e+00, 6.66105914590696e+00, 6.65823350735890e+00, 6.65427500162080e+00, 6.64890834695979e+00, 6.64183518299907e+00, 6.63274003047458e+00, 6.62129745496962e+00, 6.60718026067519e+00, 6.59006848487798e+00, 6.56965891178900e+00, 6.54567477979853e+00, 6.51787532390041e+00, 6.48606477920869e+00, 6.45010047554412e+00, 6.40989967866914e+00, 6.36544488039166e+00, 6.31678730457877e+00, 6.26404847419451e+00, 6.20741976953346e+00, 6.14715999330263e+00, 6.08359103834444e+00, 6.01709182445389e+00, 5.94809072971047e+00, 5.87705678844754e+00, 5.80448996274194e+00, 5.73091081749708e+00, 5.65684994061887e+00, 5.58283744857758e+00, 5.50939290267029e+00, 5.43701593186924e+00, 5.36617781477729e+00, 5.29731421807024e+00, 5.23081922567416e+00, 5.16704072675962e+00, 5.10627716680381e+00, 5.04877560942467e+00, 4.99473101125331e+00, 4.94428658005207e+00, 4.89753506818684e+00, 4.85452084848493e+00, 4.81524262532831e+00, 4.77965664768760e+00, 4.74768030953768e+00, 4.71919604364516e+00, 4.69405543439181e+00, 4.67208349197756e+00, 4.65308304257239e+00, 4.63683919599405e+00, 4.62312385416428e+00, 4.61170022042182e+00, 4.60232726271398e+00, 4.59476407409434e+00, 4.58877406340402e+00, 4.58412889918025e+00, 4.58061212233246e+00, 4.57802233935553e+00, 4.57617590887940e+00, 4.57490904080795e+00, 4.57407923930351e+00, 4.57356603804263e+00, 4.57327099765130e+00, 4.57311695979628e+00, 4.57304657858670e+00, 4.57302017615847e+00, 4.57301299405720e+00, 4.57301193399407e+00, 4.57301185462710e+00, 4.57301046050640e+00, 4.57300101513468e+00, 4.57296629296547e+00, 4.57287373653677e+00, 4.57267117969672e+00, 4.57228326157093e+00, 4.57160867638692e+00, 4.57051838807884e+00, 4.56885491974579e+00, 4.56643280589437e+00, 4.56304026910822e+00, 4.55844215147207e+00, 4.55238409406315e+00, 4.54459791488830e+00, 4.53480808725232e+00, 4.52273916800767e+00, 4.50812397070757e+00, 4.49071222552843e+00, 4.47027941985654e+00, 4.44663547503153e+00, 4.41963289035954e+00, 4.38917397919966e+00, 4.35521683682721e+00, 4.31777971763108e+00, 4.27694355996578e+00, 4.23285247856098e+00, 4.18571214258623e+00, 4.13578606611449e+00, 4.08338994910777e+00, 4.02888431253983e+00, 3.97266576219663e+00, 3.91515728429447e+00, 3.85679801645254e+00, 3.79803294656561e+00, 3.73930296972840e+00, 3.68103568277238e+00, 3.62363722318504e+00, 3.56748537214257e+00, 3.51292404890042e+00, 3.46025923435117e+00, 3.40975628235477e+00, 3.36163851363672e+00, 3.31608694146405e+00, 3.27324095150078e+00, 3.23319974885468e+00, 3.19602439061757e+00, 3.16174023865123e+00, 3.13033969122160e+00, 3.10178507976537e+00, 3.07601164549585e+00, 3.05293053725568e+00, 3.03243179520334e+00, 3.01438730340958e+00, 2.99865370761495e+00, 2.98507530209890e+00, 2.97348689204628e+00, 2.96371663548460e+00, 2.95558886257660e+00, 2.94892686076144e+00, 2.94355560308349e+00, 2.93930438527549e+00, 2.93600932606165e+00, 2.93351567596326e+00, 2.93167987373478e+00, 2.93037128731227e+00, 2.92947357837349e+00, 2.92888563646557e+00, 2.92852203990909e+00, 2.92831301570365e+00, 2.92820388846748e+00, 2.92815402784264e+00, 2.92813532348351e+00, 2.92813023545210e+00, 2.92812948446961e+00, 2.92812942809437e+00, 2.92812843584278e+00, 2.92812171319664e+00, 2.92809700010359e+00, 2.92803112457251e+00, 2.92788696012907e+00, 2.92761087772622e+00, 2.92713079863283e+00, 2.92635494514135e+00, 2.92517137445972e+00, 2.92344836703204e+00, 2.92103572265547e+00, 2.91776699495715e+00, 2.91346266608595e+00, 2.90793422835170e+00, 2.90098909824951e+00, 2.89243624205627e+00, 2.88209234331055e+00, 2.86978829442463e+00, 2.85537575179786e+00, 2.83873346103037e+00, 2.81977304111440e+00, 2.79844391807642e+00, 2.77473712234406e+00, 2.74868771096643e+00, 2.72037564409136e+00, 2.68992503053538e+00, 2.65750175323246e+00, 2.62330958344366e+00, 2.58758498374079e+00, 2.55059087527481e+00, 2.51260969767446e+00, 2.47393611567238e+00, 2.43486972399922e+00, 2.39570807325385e+00, 2.35674028919913e+00, 2.31824149305370e+00, 2.28046815851958e+00, 2.24365446993528e+00, 2.20800968134145e+00, 2.17371642289371e+00, 2.14092986145657e+00, 2.10977759696492e+00, 2.08036016429760e+00, 2.05275200995099e+00, 2.02700282113677e+00, 2.00313909932558e+00, 1.98116588816528e+00, 1.96106858494893e+00, 1.94281478368177e+00, 1.92635611505439e+00, 1.91163006343555e+00, 1.89856175285764e+00, 1.88706570264301e+00, 1.87704755877370e+00, 1.86840580945000e+00, 1.86103349274983e+00, 1.85481990123300e+00, 1.84965228317831e+00, 1.84541753344302e+00, 1.84200385933084e+00, 1.83930239905549e+00, 1.83720876313962e+00, 1.83562446313329e+00, 1.83445818804343e+00, 1.83362688737918e+00, 1.83305662108784e+00, 1.83268314100056e+00, 1.83245217557977e+00, 1.83231939936695e+00, 1.83225007994945e+00, 1.83221840773042e+00, 1.83220652645157e+00, 1.83220329446168e+00, 1.83220281742696e+00, 1.83220278146409e+00, 1.83220214645586e+00, 1.83219784418497e+00, 1.83218202863189e+00, 1.83213987055661e+00, 1.83204761058120e+00, 1.83187092912935e+00, 1.83156370187696e+00, 1.83106720344172e+00, 1.83030981497147e+00, 1.82920728246654e+00, 1.82766356128737e+00, 1.82557226750812e+00, 1.82281873786113e+00, 1.81928267659240e+00, 1.81484133977500e+00, 1.80937317638038e+00, 1.80276181238834e+00, 1.79490023194506e+00, 1.78569498124753e+00, 1.77507020000445e+00, 1.76297127547596e+00, 1.74936791808446e+00, 1.73425647709017e+00, 1.71766134987696e+00, 1.69963538714461e+00, 1.68025925504894e+00, 1.65963977887958e+00, 1.63790735518298e+00, 1.61521257428707e+00, 1.59172223778188e+00, 1.56761498205609e+00, 1.54307672791268e+00, 1.51829616813406e+00, 1.49346048205408e+00, 1.46875143247015e+00, 1.44434196001426e+00, 1.42039334782373e+00, 1.39705298888977e+00, 1.37445275279257e+00, 1.35270791957483e+00, 1.33191662715263e+00, 1.31215976493003e+00, 1.29350123953660e+00, 1.27598853779010e+00, 1.25965351583978e+00, 1.24451335066489e+00, 1.23057159946095e+00, 1.21781932286042e+00, 1.20623623850073e+00, 1.19579188145340e+00, 1.18644675691975e+00, 1.17815347799697e+00, 1.17085788697448e+00, 1.16450016240609e+00, 1.15901591609649e+00, 1.15433728421500e+00, 1.15039401517150e+00, 1.14711455391322e+00, 1.14442711826050e+00, 1.14226075820424e+00, 1.14054638419752e+00, 1.13921774588142e+00, 1.13821233887711e+00, 1.13747221469604e+00, 1.13694466782389e+00, 1.13658277485559e+00, 1.13634576327542e+00, 1.13619919200078e+00, 1.13611493188561e+00, 1.13607094161801e+00, 1.13605084235398e+00, 1.13604330246699e+00, 1.13604125143877e+00, 1.13604094871145e+00, 1.13604092576513e+00, 1.13604051894708e+00, 1.13603776269712e+00, 1.13602763046361e+00, 1.13600062190190e+00, 1.13594151560795e+00, 1.13582832489929e+00, 1.13563150068775e+00, 1.13531342172674e+00, 1.13482820802770e+00, 1.13412188762343e+00, 1.13313293959345e+00, 1.13179322680647e+00, 1.13002931971426e+00, 1.12776419750122e+00, 1.12491929505523e+00, 1.12141684414043e+00, 1.11718243591177e+00, 1.11214771112819e+00, 1.10625306614052e+00, 1.09945024922511e+00, 1.09170471532711e+00, 1.08299760958943e+00, 1.07332726224694e+00, 1.06271009960302e+00, 1.05118090673004e+00, 1.03879241494577e+00, 1.02561422778066e+00, 1.01173113930513e+00, 9.97240934542475e-01, 9.82251789961119e-01, 9.66879410359081e-01, 9.51244045671693e-01, 9.35467527483659e-01, 9.19670451596396e-01, 9.03969612060266e-01, 8.88475766301010e-01, 8.73291783141756e-01, 8.58511198215943e-01, 8.44217176580614e-01, 8.30481861761378e-01, 8.17366074831668e-01, 8.04919316694892e-01, 7.93180021254521e-01, 7.82176006023651e-01, 7.71925069126357e-01, 7.62435686668734e-01, 7.53707771206699e-01, 7.45733459684477e-01, 7.38497907056196e-01, 7.31980069265848e-01, 7.26153465918657e-01, 7.20986918533395e-01, 7.16445264536105e-01, 7.12490050056804e-01, 7.09080206118468e-01, 7.06172713029289e-01, 7.03723256834049e-01, 7.01686879732407e-01, 7.00018623666789e-01, 6.98674163098570e-01, 6.97610419639812e-01, 6.96786148018038e-01, 6.96162480148178e-01, 6.95703412164892e-01, 6.95376218370239e-01, 6.95151776334676e-01, 6.95004788911952e-01, 6.94913891638045e-01, 6.94861637720440e-01, 6.94834357333278e-01, 6.94821892897028e-01, 6.94817217088698e-01, 6.94815945158481e-01, 6.94815757424362e-01, 6.94815743104080e-01, 6.94815488026787e-01, 6.94813759842326e-01, 6.94807406878969e-01, 6.94790472407281e-01, 6.94753412725394e-01, 6.94682442814557e-01, 6.94559037477205e-01, 6.94359613680817e-01, 6.94055417196470e-01, 6.93612633255569e-01, 6.92992736483525e-01, 6.92153089413974e-01, 6.91047791109082e-01, 6.89628767624679e-01, 6.87847084316854e-01, 6.85654446660951e-01, 6.83004842077516e-01, 6.79856261330276e-01, 6.76172425773171e-01, 6.71924437635980e-01, 6.67092266180283e-01, 6.61665984190250e-01, 6.55646677638522e-01, 6.49046966530356e-01, 6.41891096104388e-01, 6.34214583173343e-01, 6.26063430184146e-01, 6.17492946939272e-01, 6.08566244190591e-01, 5.99352482168502e-01, 5.89924968866566e-01, 5.80359206750007e-01, 5.70730982607405e-01, 5.61114584482343e-01, 5.51581213626736e-01, 5.42197640214589e-01, 5.33025131242869e-01, 5.24118659549384e-01, 5.15526385748970e-01, 5.07289391189975e-01, 4.99441630304648e-01, 4.92010065028261e-01, 4.85014941962753e-01, 4.78470174059407e-01, 4.72383792042361e-01, 4.66758435805275e-01, 4.61591861851783e-01, 4.56877448889987e-01, 4.52604689441849e-01, 4.48759660441244e-01, 4.45325470052047e-01, 4.42282681229852e-01, 4.39609714853567e-01, 4.37283236599260e-01, 4.35278532195291e-01, 4.33569875384524e-01, 4.32130891947498e-01, 4.30934921643736e-01, 4.29955378055606e-01, 4.29166104233357e-01, 4.28541719914016e-01, 4.28057954102606e-01, 4.27691955138798e-01, 4.27422569190234e-01, 4.27230577548322e-01, 4.27098883243666e-01, 4.27012638379573e-01, 4.26959305172240e-01, 4.26928645891888e-01, 4.26912639569810e-01, 4.26905326282693e-01, 4.26902582837833e-01, 4.26901836556354e-01, 4.26901726406840e-01, 4.26901717945395e-01, 4.26901566450127e-01, 4.26900540048564e-01, 4.26896766905298e-01, 4.26886709224789e-01, 4.26864698917537e-01, 4.26822549153631e-01, 4.26749258519504e-01, 4.26630823271307e-01, 4.26450171693151e-01, 4.26187232631364e-01, 4.25819147642510e-01, 4.25320632602514e-01, 4.24664489852908e-01, 4.23822265879050e-01, 4.22765042153870e-01, 4.21464338381295e-01, 4.19893098421582e-01, 4.18026720410513e-01, 4.15844084929169e-01, 4.13328529585403e-01, 4.10468716047507e-01, 4.07259337245335e-01, 4.03701618578403e-01, 3.99803577496225e-01, 3.95580020117443e-01, 3.91052270450611e-01, 3.86247645654713e-01, 3.81198707791782e-01, 3.75942336878982e-01, 3.70518680256815e-01, 3.64970038401012e-01, 3.59339747033129e-01, 3.53671110122253e-01, 3.48006429096834e-01, 3.42386161673339e-01, 3.36848230692635e-01, 3.31427490695450e-01, 3.26155348862393e-01, 3.21059528204422e-01, 3.16163954906354e-01, 3.11488748499738e-01, 3.07050292763577e-01, 3.02861366418455e-01, 2.98931315196729e-01, 2.95266250157169e-01, 2.91869260660203e-01, 2.88740633843016e-01, 2.85878075470535e-01, 2.83276929540285e-01, 2.80930395932522e-01, 2.78829746726689e-01, 2.76964542603430e-01, 2.75322851081949e-01, 2.73891468283537e-01, 2.72656145531023e-01, 2.71601821457416e-01, 2.70712859463952e-01, 2.69973289397452e-01, 2.69367051268677e-01, 2.68878237774821e-01, 2.68491331390951e-01, 2.68191430933800e-01, 2.67964461852692e-01, 2.67797364138392e-01, 2.67678251716710e-01, 2.67596537547863e-01, 2.67543019393266e-01, 2.67509922316602e-01, 2.67490895402016e-01, 2.67480961817071e-01, 2.67476423120961e-01, 2.67474720505991e-01, 2.67474257354063e-01, 2.67474188993823e-01, 2.67474183702461e-01, 2.67474088443262e-01, 2.67473443048640e-01, 2.67471070516446e-01, 2.67464746268271e-01, 2.67450906067431e-01, 2.67424401429402e-01, 2.67378312776994e-01, 2.67303829675384e-01, 2.67190206388843e-01, 2.67024799639621e-01, 2.66793193702068e-01, 2.66479415703024e-01, 2.66066241113762e-01, 2.65535585854007e-01, 2.64868977212650e-01, 2.64048091067679e-01, 2.63055337941758e-01, 2.61874475671210e-01, 2.60491222418505e-01, 2.58893840992812e-01, 2.57073664501736e-01, 2.55025534649249e-01, 2.52748127706283e-01, 2.50244149191577e-01, 2.47520386174671e-01, 2.44587615109744e-01, 2.41460372296345e-01, 2.38156602448502e-01, 2.34697207548038e-01, 2.31105522524107e-01, 2.27406746054950e-01, 2.23627354015594e-01, 2.19794520230593e-01, 2.15935564876858e-01, 2.12077445859979e-01, 2.08246303443712e-01, 2.04467063876773e-01, 2.00763104070372e-01, 1.97155976645924e-01, 1.93665192854457e-01, 1.90308059795186e-01, 1.87099567825303e-01, 1.84052323852310e-01, 1.81176526179925e-01, 1.78479976642063e-01, 1.75968125881504e-01, 1.73644147783484e-01, 1.71509039289969e-01, 1.69561742084223e-01, 1.67799282974027e-01, 1.66216930163582e-01, 1.64808362988513e-01, 1.63565853039799e-01, 1.62480454869467e-01, 1.61542204624983e-01, 1.60740324947177e-01, 1.60063434318200e-01, 1.59499758723882e-01, 1.59037343065929e-01, 1.58664259252973e-01, 1.58368807387536e-01, 1.58139705992717e-01, 1.57966266902243e-01, 1.57838550263985e-01, 1.57747495206323e-01, 1.57685022049453e-01, 1.57644102559026e-01, 1.57618795606773e-01, 1.57604246669319e-01, 1.57596650826179e-01, 1.57593180226076e-01, 1.57591878284563e-01, 1.57591524125166e-01, 1.57591471851791e-01, 1.57591467772924e-01, 1.57591393916611e-01, 1.57590893528532e-01, 1.57589054050451e-01, 1.57584150696096e-01, 1.57573419912687e-01, 1.57552869516291e-01, 1.57517133261102e-01, 1.57459376809385e-01, 1.57371261108363e-01, 1.57242968054501e-01, 1.57063291928473e-01, 1.56819798410577e-01, 1.56499050919642e-01, 1.56086901591696e-01, 1.55568841429265e-01, 1.54930401081245e-01, 1.54157590560825e-01, 1.53237363144985e-01, 1.52158086026663e-01, 1.50909998309277e-01, 1.49485635924421e-01, 1.47880203229331e-01, 1.46091872551594e-01, 1.44121995764806e-01, 1.41975215952989e-01, 1.39659472092216e-01, 1.37185895020853e-01, 1.34568598445686e-01, 1.31824373847881e-01, 1.28972302659385e-01, 1.26033302652946e-01, 1.23029628023172e-01, 1.19984343965676e-01, 1.16920796802834e-01, 1.13862099774060e-01, 1.10830652722129e-01, 1.07847711130219e-01, 1.04933016574179e-01, 1.02104496762287e-01, 9.93780393946366e-02, 9.67673400343259e-02, 9.42838206680889e-02, 9.19366128429964e-02, 8.97325968665434e-02, 8.76764875065987e-02, 8.57709558861437e-02, 8.40167776985368e-02, 8.24129984288753e-02, 8.09571078431119e-02, 7.96452175229412e-02, 7.84722365612609e-02, 7.74320427700381e-02, 7.65176475980844e-02, 7.57213541067649e-02, 7.50349087332332e-02, 7.44496470369924e-02, 7.39566350168630e-02, 7.35468060138480e-02, 7.32110942409037e-02, 7.29405640832539e-02, 7.27265347180612e-02, 7.25606978783020e-02, 7.24352268030692e-02, 7.23428737267078e-02, 7.22770522469567e-02, 7.22319021724900e-02, 7.22023337292796e-02, 7.21840486654912e-02, 7.21735372488652e-02, 7.21680495363834e-02, 7.21655421879568e-02, 7.21646016081863e-02, 7.21643457493570e-02, 7.21643079878511e-02, 7.21643050288350e-02, 7.21642508884806e-02, 7.21638841226131e-02, 7.21625358980518e-02, 7.21589420802999e-02, 7.21510773592087e-02, 7.21360164307576e-02, 7.21098284038639e-02, 7.20675097970229e-02, 7.20029613372866e-02, 7.18826035895752e-02, 7.17512807427561e-02, 7.15734303479358e-02, 7.13393567285721e-02, 7.10389222061742e-02, 7.06618329794182e-02, 7.01979669647024e-02, 6.96377336477703e-02, 6.89724529011499e-02, 6.81947368995282e-02, 6.72988570305471e-02, 6.62810763728992e-02, 6.51399281846956e-02, 6.38764221232506e-02, 6.24941626963533e-02, 6.09993686691261e-02, 5.94007876030389e-02, 5.77095060072068e-02, 5.59386622301403e-02, 5.41030756218284e-02, 5.22188110450017e-02, 5.03027019615452e-02, 4.83718576460888e-02, 4.64431803584553e-02, 4.45329165371895e-02, 4.26562624887481e-02, 4.08270400659916e-02, 3.90574520185147e-02, 3.73579206713188e-02, 3.57370079324575e-02, 3.42014098266645e-02, 3.27560151283777e-02, 3.14040153784667e-02, 3.01470526030307e-02, 2.89853912613793e-02, 2.79181020950787e-02, 2.69432473463631e-02, 2.60580589788439e-02, 2.52591038129779e-02, 2.45424316812404e-02, 2.39037046699799e-02, 2.33383071587348e-02, 2.28414376538808e-02, 2.24081843387900e-02, 2.20335868475310e-02, 2.17126870501723e-02, 2.14405716551589e-02, 2.12124092338364e-02, 2.10234838972632e-02, 2.08692273509961e-02, 2.07452504641346e-02, 2.06473748602178e-02, 2.05716644159075e-02, 2.05144559846424e-02, 2.04723881897095e-02, 2.04424267907832e-02, 2.04218849471722e-02, 2.04084366931667e-02, 2.04001221042276e-02, 2.03953429493211e-02, 2.03928480631816e-02, 2.03917081892477e-02, 2.03912805929178e-02, 2.03911642772638e-02, 2.03911471093505e-02, 2.03911457463555e-02, 2.03911207676661e-02, 2.03909515339227e-02, 2.03903294207031e-02, 2.03886711560521e-02, 2.03850423929630e-02, 2.03780940597191e-02, 2.03660145490048e-02, 2.03465007783308e-02, 2.03167510938265e-02, 2.02734828276938e-02, 2.02129769718979e-02, 2.01311518045273e-02, 2.00236663197408e-02, 1.98860529145684e-02, 1.97138769718954e-02, 1.95029188114120e-02, 1.92493710982973e-02, 1.89500424219062e-02, 1.86025556677762e-02, 1.82055283283447e-02, 1.77587213481019e-02, 1.72631437355015e-02, 1.67211021415135e-02, 1.61361878842925e-02, 1.55131982844097e-02, 1.48579942715727e-02, 1.41773014906761e-02, 1.34784669565775e-02, 1.27691870861426e-02, 1.20572251929723e-02, 1.13501369834133e-02, 1.06550212073522e-02, 9.97830961328106e-03, 9.32560616812950e-03, 8.70158070563321e-03, 8.10991737573019e-03, 7.55331402999104e-03, 7.03352539139246e-03, 6.55144072833735e-03, 6.10718580918501e-03, 5.70023903648220e-03, 5.32955263366119e-03, 4.99367130671402e-03, 4.69084263990283e-03, 4.41911534623429e-03, 4.17642318642919e-03, 3.96065378543042e-03, 3.76970267612720e-03, 3.60151369227424e-03, 3.45410734338714e-03, 3.32559906908259e-03, 3.21420933750360e-03, 3.11826747724453e-03, 3.03621097015849e-03, 2.96658172652542e-03, 2.90802064052897e-03, 2.85926149506466e-03, 2.81912505486304e-03, 2.78651396035556e-03, 2.76040881388795e-03, 2.73986564150164e-03, 2.72401472080475e-03, 2.71206059560924e-03, 2.70328295641775e-03, 2.69703795970955e-03, 2.69275949392878e-03, 2.68995988116671e-03, 2.68822953066561e-03, 2.68723513119795e-03, 2.68671607534188e-03, 2.68647893961216e-03, 2.68638998588600e-03, 2.68636578869774e-03, 2.68636221725771e-03, 2.68636193112570e-03, 2.68635665454298e-03, 2.68632090515908e-03, 2.68618948988930e-03, 2.68583921008059e-03, 2.68507275838559e-03, 2.68360539860169e-03, 2.68105517074702e-03, 2.67693741141173e-03, 2.67066442183861e-03, 2.66155112519328e-03, 2.64882749439010e-03, 2.63165836500434e-03, 2.60917093801893e-03, 2.58048980395742e-03, 2.54477868148228e-03, 2.50128729123134e-03, 2.44940094720647e-03, 2.38868964601124e-03, 2.31895280211347e-03, 2.24025545593420e-03, 2.15295189939013e-03, 2.05769330248273e-03, 1.95541709478420e-03, 1.84731747308220e-03, 1.73479829152967e-03, 1.61941148407863e-03, 1.50278577788043e-03, 1.38655151654277e-03, 1.27226775162370e-03, 1.16135734235456e-03, 1.05505472290446e-03, 9.54369442737857e-04, 8.60066790181691e-04, 7.72665012501506e-04, 6.92447085516079e-04, 6.19483875716753e-04, 5.53665012137827e-04, 4.94733834752736e-04, 4.42323246975462e-04, 3.95989941793573e-04, 3.55245130624127e-04, 3.19580546282339e-04, 2.88489144011805e-04, 2.61480560773464e-04, 2.38091890810921e-04, 2.17990791554712e-04, 2.00605133695575e-04, 1.85551674135809e-04, 1.72817654750190e-04, 1.61931992668229e-04, 1.52626751894758e-04, 1.44688627685241e-04, 1.37937653985176e-04, 1.32313144135310e-04, 1.27664793257765e-04, 1.23790117534384e-04, 1.20583714345202e-04, 1.17955560675398e-04, 1.15826741861724e-04, 1.14126859276809e-04, 1.12792195885447e-04, 1.11764675413184e-04, 1.10991443467157e-04, 1.10424819112609e-04, 1.10022384305353e-04, 1.09747056807663e-04, 1.09567080333051e-04, 1.09455920007662e-04, 1.09392066061531e-04, 1.09358743863197e-04, 1.09343522197491e-04, 1.09337812595811e-04, 1.09336259498605e-04, 1.09336030270413e-04, 1.09336011789196e-04, 1.09335669574051e-04, 1.09333351034024e-04, 1.09324828002855e-04, 1.09302110029574e-04, 1.09252398946325e-04, 1.09157221920952e-04, 1.08991789819157e-04, 1.08724631957102e-04, 1.08317566532179e-04, 1.07726076592485e-04, 1.06900165452117e-04, 1.05785730688979e-04, 1.04326442841085e-04, 1.02479605445183e-04, 1.00284716332055e-04, 9.76307909574814e-05, 9.44797536419628e-05, 9.08193711417482e-05, 8.66554867364755e-05, 8.20153119889255e-05, 7.69476858266671e-05, 7.16166649681327e-05, 6.61305408868884e-05, 6.04609025032232e-05, 5.47169177167461e-05, 4.90368539125541e-05, 4.36072033771495e-05, 3.84149729342198e-05, 3.35120980388126e-05, 2.90193022998153e-05, 2.49700745378334e-05, 2.12990868639906e-05, 1.81018308715863e-05, 1.53785260010864e-05, 1.30407501863772e-05, 1.10443471044164e-05, 9.32978976919685e-06, 7.84715323121077e-06, 6.56702480483773e-06, 5.47004082562808e-06, 4.53285069212939e-06, 3.77596572675998e-06, 3.17453314835348e-06, 2.57767775124762e-06, 2.11231335711659e-06, 1.75829660165050e-06, 1.41150158228243e-06, 1.07654857117723e-06, 7.70901319668837e-07, 5.14079221858310e-07, -1.06581410364015e-14, -7.10542735760100e-15, -7.10542735760100e-15, -3.55271367880050e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -3.55271367880050e-15, -3.55271367880050e-15, -7.10542735760100e-15, -3.55271367880050e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -1.06581410364015e-14, -1.06581410364015e-14, -7.10542735760100e-15, -1.06581410364015e-14, -1.06581410364015e-14, -7.10542735760100e-15, -7.10542735760100e-15],
[ 2.50000000000000e+01, 2.49999999832335e+01, 2.49999994647798e+01, 2.49999959521747e+01, 2.49999830393859e+01, 2.49999486180123e+01, 2.49998732845335e+01, 2.49997289998588e+01, 2.49994780498853e+01, 2.49990723467383e+01, 2.49984531001169e+01, 2.49975508770514e+01, 2.49962860567862e+01, 2.49945696758418e+01, 2.49923046470017e+01, 2.49893873253965e+01, 2.49857093853552e+01, 2.49811599635194e+01, 2.49756280170203e+01, 2.49690048403533e+01, 2.49611866808868e+01, 2.49520773905743e+01, 2.49415910501995e+01, 2.49296545021822e+01, 2.49162097284624e+01, 2.49012160112019e+01, 2.48846518160717e+01, 2.48665163409395e+01, 2.48468306771302e+01, 2.48256385364704e+01, 2.48030065053809e+01, 2.47790237975537e+01, 2.47538014892767e+01, 2.47274712360155e+01, 2.47001834849306e+01, 2.46721052148541e+01, 2.46434172519493e+01, 2.46143112248193e+01, 2.45849862362429e+01, 2.45556453391537e+01, 2.45264919113329e+01, 2.44977260262346e+01, 2.44695409163647e+01, 2.44421196209623e+01, 2.44156319018362e+01, 2.43902315007431e+01, 2.43660537993449e+01, 2.43432139292674e+01, 2.43218053657381e+01, 2.43018990242613e+01, 2.42835428662213e+01, 2.42667620064869e+01, 2.42515593042366e+01, 2.42379164074455e+01, 2.42257952118311e+01, 2.42151396865932e+01, 2.42058780120263e+01, 2.41979249680756e+01, 2.41911845082235e+01, 2.41855524497952e+01, 2.41809192099804e+01, 2.41771725166538e+01, 2.41742000245319e+01, 2.41718917703737e+01, 2.41701424058084e+01, 2.41688531529044e+01, 2.41679334356572e+01, 2.41673021499839e+01, 2.41668885453203e+01, 2.41666327022421e+01, 2.41664856023263e+01, 2.41664087984019e+01, 2.41663737050242e+01, 2.41663605401200e+01, 2.41663569589311e+01, 2.41663564303546e+01, 2.41663563610165e+01, 2.41663547455218e+01, 2.41663438002949e+01, 2.41663035642897e+01, 2.41661963083876e+01, 2.41659615739346e+01, 2.41655120005056e+01, 2.41647300982077e+01, 2.41634660937355e+01, 2.41615369497348e+01, 2.41587266245947e+01, 2.41547876054354e+01, 2.41494437117368e+01, 2.41423941317726e+01, 2.41333186197491e+01, 2.41218837491648e+01, 2.41077500880071e+01, 2.40905801342522e+01, 2.40700468255980e+01, 2.40458424150167e+01, 2.40176874830225e+01, 2.39853398382914e+01, 2.39486030408523e+01, 2.39073342680046e+01, 2.38614512350929e+01, 2.38109378850350e+01, 2.37558485762518e+01, 2.36963105321564e+01, 2.36325243688296e+01, 2.35647625906190e+01, 2.34933660327040e+01, 2.34187383286237e+01, 2.33413385805024e+01, 2.32616725003426e+01, 2.31802823630946e+01, 2.30977361592767e+01, 2.30146163531210e+01, 2.29315086418082e+01, 2.28489910760917e+01, 2.27676238488996e+01, 2.26879399940263e+01, 2.26104371695069e+01, 2.25355706362457e+01, 2.24637474866911e+01, 2.23953221334183e+01, 2.23305930340825e+01, 2.22698006064338e+01, 2.22131262730806e+01, 2.21606925680920e+01, 2.21125642339801e+01, 2.20687502360136e+01, 2.20292066195524e+01, 2.19938401340069e+01, 2.19625125434622e+01, 2.19350455386892e+01, 2.19112261582727e+01, 2.18908126182555e+01, 2.18735404405618e+01, 2.18591287612329e+01, 2.18472866909874e+01, 2.18377195936967e+01, 2.18301351439409e+01, 2.18242490237225e+01, 2.18197901214120e+01, 2.18165051036169e+01, 2.18141622432091e+01, 2.18125544042058e+01, 2.18115011062561e+01, 2.18108496174833e+01, 2.18104750534261e+01, 2.18102794906449e+01, 2.18101901348907e+01, 2.18101566142191e+01, 2.18101474957607e+01, 2.18101461498945e+01, 2.18101460310801e+01, 2.18101437029709e+01, 2.18101279296815e+01, 2.18100699452936e+01, 2.18099153793776e+01, 2.18095771114565e+01, 2.18089292718454e+01, 2.18078026230210e+01, 2.18059815239171e+01, 2.18032026429809e+01, 2.17991555442465e+01, 2.17934852237177e+01, 2.17857966206116e+01, 2.17756610693133e+01, 2.17626245935303e+01, 2.17462178751125e+01, 2.17259676582031e+01, 2.17014092777069e+01, 2.16720999334034e+01, 2.16376322721296e+01, 2.15976477956439e+01, 2.15518495864547e+01, 2.15000138429643e+01, 2.14419997423779e+01, 2.13777572065387e+01, 2.13073322309819e+01, 2.12308695466194e+01, 2.11486125089038e+01, 2.10609002407833e+01, 2.09681621816168e+01, 2.08709103032937e+01, 2.07697293382584e+01, 2.06652654171352e+01, 2.05582135360432e+01, 2.04493042697964e+01, 2.03392901243358e+01, 2.02289318882717e+01, 2.01189853064793e+01, 2.00101883627058e+01, 1.99032494242721e+01, 1.97988364686256e+01, 1.96975675759616e+01, 1.96000028320309e+01, 1.95066377400350e+01, 1.94178981918873e+01, 1.93341370006413e+01, 1.92556319518529e+01, 1.91825852958359e+01, 1.91151245774232e+01, 1.90533046852659e+01, 1.89971109973865e+01, 1.89464635009727e+01, 1.89012217691204e+01, 1.88611906824727e+01, 1.88261267872038e+01, 1.87957451811537e+01, 1.87697268165954e+01, 1.87477261013636e+01, 1.87293786707021e+01, 1.87143091914611e+01, 1.87021390496997e+01, 1.86924937639391e+01, 1.86850099608590e+01, 1.86793417495602e+01, 1.86751663357757e+01, 1.86721887293529e+01, 1.86701454172356e+01, 1.86688068997921e+01, 1.86679790199043e+01, 1.86675030505041e+01, 1.86672545455849e+01, 1.86671410002653e+01, 1.86670984052743e+01, 1.86670868183831e+01, 1.86670851081815e+01, 1.86670849710022e+01, 1.86670824393476e+01, 1.86670652870103e+01, 1.86670022331205e+01, 1.86668341539604e+01, 1.86664663135823e+01, 1.86657618428086e+01, 1.86645367232810e+01, 1.86625564998681e+01, 1.86595349062474e+01, 1.86551345443995e+01, 1.86489697084315e+01, 1.86406113858621e+01, 1.86295944048390e+01, 1.86154266238574e+01, 1.85975999824538e+01, 1.85756031491863e+01, 1.85489354203592e+01, 1.85171214439545e+01, 1.84797262736465e+01, 1.84363702037509e+01, 1.83867428037414e+01, 1.83306155662658e+01, 1.82678526098859e+01, 1.81984189396469e+01, 1.81223858650397e+01, 1.80399333027564e+01, 1.79513488439235e+01, 1.78570236314104e+01, 1.77574452580832e+01, 1.76531880452037e+01, 1.75449011757704e+01, 1.74332952284738e+01, 1.73191276790221e+01, 1.72031879104173e+01, 1.70862822137266e+01, 1.69692191820209e+01, 1.68527958185125e+01, 1.67377846073952e+01, 1.66249217379244e+01, 1.65148966281763e+01, 1.64083428600958e+01, 1.63058306063388e+01, 1.62078605979375e+01, 1.61148596484533e+01, 1.60271777160180e+01, 1.59450864520055e+01, 1.58687791568913e+01, 1.57983720423331e+01, 1.57339066846539e+01, 1.56753535483723e+01, 1.56226164578042e+01, 1.55755378979485e+01, 1.55339050305706e+01, 1.54974563154784e+01, 1.54658886288159e+01, 1.54388647687596e+01, 1.54160212340131e+01, 1.53969761523409e+01, 1.53813372260363e+01, 1.53687095501289e+01, 1.53587031490103e+01, 1.53509400698374e+01, 1.53450608682752e+01, 1.53407303253573e+01, 1.53376422444917e+01, 1.53355231954365e+01, 1.53341350972776e+01, 1.53332765643313e+01, 1.53327829761543e+01, 1.53325252737158e+01, 1.53324075262182e+01, 1.53323633548846e+01, 1.53323513391937e+01, 1.53323495657023e+01, 1.53323494289136e+01, 1.53323469726270e+01, 1.53323303309197e+01, 1.53322691541779e+01, 1.53321060789425e+01, 1.53317491902278e+01, 1.53310656958705e+01, 1.53298770628670e+01, 1.53279558338377e+01, 1.53250243049898e+01, 1.53207552047346e+01, 1.53147744636755e+01, 1.53066661112396e+01, 1.52959792712727e+01, 1.52822371585691e+01, 1.52649479015869e+01, 1.52436169356860e+01, 1.52177606295442e+01, 1.51869207295953e+01, 1.51506791389455e+01, 1.51086724943344e+01, 1.50606059732027e+01, 1.50062657579127e+01, 1.49455296091599e+01, 1.48783750570772e+01, 1.48048848054824e+01, 1.47252490586700e+01, 1.46397646152892e+01, 1.45488307224294e+01, 1.44529418358372e+01, 1.43526775791473e+01, 1.42486903258507e+01, 1.41416909327872e+01, 1.40324332251339e+01, 1.39216978646068e+01, 1.38102762228730e+01, 1.36989548329910e+01, 1.35885009091981e+01, 1.34796493190580e+01, 1.33730912735310e+01, 1.32694648819765e+01, 1.31693476112167e+01, 1.30732505986696e+01, 1.29816147038325e+01, 1.28948081411495e+01, 1.28131255185398e+01, 1.27367881054730e+01, 1.26659451671964e+01, 1.26006762221354e+01, 1.25409941027735e+01, 1.24868487226335e+01, 1.24381314706415e+01, 1.23946801675777e+01, 1.23562845268105e+01, 1.23226920630997e+01, 1.22936143894157e+01, 1.22687338332890e+01, 1.22477102922233e+01, 1.22301882333915e+01, 1.22158037275415e+01, 1.22041913922239e+01, 1.21949911066257e+01, 1.21878543509516e+01, 1.21824500188239e+01, 1.21784695527488e+01, 1.21756312611308e+01, 1.21736836910033e+01, 1.21724079534327e+01, 1.21716189277425e+01, 1.21711653051011e+01, 1.21709284699458e+01, 1.21708202572638e+01, 1.21707796628306e+01, 1.21707686201491e+01, 1.21707669902721e+01, 1.21707668668627e+01, 1.21707646806779e+01, 1.21707498689516e+01, 1.21706954194425e+01, 1.21705502767623e+01, 1.21702326341345e+01, 1.21696243050520e+01, 1.21685663986115e+01, 1.21668564947164e+01, 1.21642474833511e+01, 1.21604481951042e+01, 1.21551259070975e+01, 1.21479107584777e+01, 1.21384020519121e+01, 1.21261763519886e+01, 1.21107972189307e+01, 1.20918263387286e+01, 1.20688357322378e+01, 1.20414206510272e+01, 1.20092127029293e+01, 1.19718927022221e+01, 1.19292027148818e+01, 1.18809567740980e+01, 1.18270497788951e+01, 1.17674641599533e+01, 1.17022739987125e+01, 1.16316464119609e+01, 1.15558401544446e+01, 1.14752015343939e+01, 1.13901578683351e+01, 1.13012088103885e+01, 1.12089159686349e+01, 1.11138912626150e+01, 1.10167844821123e+01, 1.09182704832096e+01, 1.08190364117053e+01, 1.07197692863244e+01, 1.06211442141117e+01, 1.05238134549568e+01, 1.04283965049840e+01, 1.03354713298033e+01, 1.02455668461018e+01, 1.01591567204017e+01, 1.00766545240976e+01, 9.99841025265792e+00, 9.92470818460972e+00, 9.85576602468288e+00, 9.79173524812575e+00, 9.73270254234736e+00, 9.67869222931661e+00, 9.62966954772583e+00, 9.58554467657786e+00, 9.54617738932997e+00, 9.51138223731585e+00, 9.48093417026068e+00, 9.45457450822388e+00, 9.43201718194947e+00, 9.41295515692186e+00, 9.39706695075193e+00, 9.38402314483628e+00, 9.37349278098746e+00, 9.36514952363943e+00, 9.35867746008502e+00, 9.35377640669494e+00, 9.35016658963238e+00, 9.34759257524490e+00, 9.34582633861501e+00, 9.34466937864913e+00, 9.34395381398186e+00, 9.34354242475017e+00, 9.34332763942048e+00, 9.34322950152704e+00, 9.34319268650148e+00, 9.34318267191117e+00, 9.34318119377823e+00, 9.34318108281432e+00, 9.34317912972365e+00, 9.34316589724436e+00, 9.34311725321193e+00, 9.34298758578292e+00, 9.34270381027288e+00, 9.34216034042496e+00, 9.34121522366120e+00, 9.33968761502859e+00, 9.33735673581654e+00, 9.33396242970432e+00, 9.32920739108271e+00, 9.32276109615036e+00, 9.31426541701584e+00, 9.30334184235030e+00, 9.28960016590208e+00, 9.27264843811873e+00, 9.25210390897678e+00, 9.22760462563356e+00, 9.19882129121553e+00, 9.16546894590232e+00, 9.12731800338627e+00, 9.08420416915841e+00, 9.03603678522893e+00, 8.98280519072183e+00, 8.92458275947850e+00, 8.86152837270195e+00, 8.79388520315839e+00, 8.72197682181044e+00, 8.64620077995987e+00, 8.56701995955236e+00, 8.48495210858683e+00, 8.40055807382741e+00, 8.31442929655688e+00, 8.22717514038020e+00, 8.13941057172306e+00, 8.05174462129768e+00, 7.96476993435256e+00, 7.87905358975128e+00, 7.79512925317951e+00, 7.71349064284389e+00, 7.63458623307399e+00, 7.55881509981579e+00, 7.48652381363592e+00, 7.41800429983212e+00, 7.35349260204728e+00, 7.29316849909004e+00, 7.23715593178065e+00, 7.18552419775029e+00, 7.13828986903045e+00, 7.09541938219195e+00, 7.05683224552390e+00, 7.02240480326178e+00, 6.99197449334148e+00, 6.96534453212945e+00, 6.94228895632724e+00, 6.92255794805570e+00, 6.90588336352755e+00, 6.89198437864357e+00, 6.88057315666438e+00, 6.87136043458626e+00, 6.86406091705831e+00, 6.85839836086347e+00, 6.85411023039835e+00, 6.85095180633243e+00, 6.84869963652711e+00, 6.84715423077150e+00, 6.84614191891530e+00, 6.84551581503813e+00, 6.84515585744654e+00, 6.84496792423289e+00, 6.84488205533063e+00, 6.84484984283333e+00, 6.84484108024287e+00, 6.84483978690251e+00, 6.84483969010628e+00, 6.84483799031587e+00, 6.84482647398314e+00, 6.84478413868603e+00, 6.84467128802227e+00, 6.84442431555660e+00, 6.84395132780651e+00, 6.84312877957542e+00, 6.84179927311061e+00, 6.83977064730956e+00, 6.83681645412559e+00, 6.83267788583442e+00, 6.82706717830490e+00, 6.81967247146410e+00, 6.81016405881018e+00, 6.79820190385154e+00, 6.78344424436311e+00, 6.76555704794984e+00, 6.74422402811173e+00, 6.71915688303187e+00, 6.69010538416236e+00, 6.65686692260403e+00, 6.61929512160120e+00, 6.57730714505494e+00, 6.53088937474586e+00, 6.48010119089029e+00, 6.42507666790122e+00, 6.36602408475079e+00, 6.30322324162857e+00, 6.23702066638688e+00, 6.16782288100237e+00, 6.09608797623979e+00, 6.02231580883492e+00, 5.94703718712946e+00, 5.87080244566584e+00, 5.79416982447420e+00, 5.71769406296506e+00, 5.64191559097847e+00, 5.56735065188237e+00, 5.49448262792845e+00, 5.42375476151117e+00, 5.35556438398334e+00, 5.29025868314759e+00, 5.22813196783669e+00, 5.16942432815600e+00, 5.11432154616269e+00, 5.06295608513365e+00, 5.01540897537474e+00, 4.97171241849375e+00, 4.93185294696546e+00, 4.89577499797067e+00, 4.86338478621771e+00, 4.83455438644883e+00, 4.80912595988298e+00, 4.78691607796089e+00, 4.76772011018748e+00, 4.75131665008749e+00, 4.73747195442087e+00, 4.72594436653315e+00, 4.71648868618349e+00, 4.70886043688660e+00, 4.70281996942441e+00, 4.69813632850986e+00, 4.69459080032499e+00, 4.69198005330463e+00, 4.69011878423738e+00, 4.68884178719588e+00, 4.68800537415108e+00, 4.68748809299425e+00, 4.68719071020970e+00, 4.68703545032569e+00, 4.68696451096908e+00, 4.68693789919431e+00, 4.68693066014887e+00, 4.68692959168102e+00, 4.68692951168478e+00, 4.68692810651083e+00, 4.68691858625304e+00, 4.68688358881509e+00, 4.68679029877742e+00, 4.68658613724599e+00, 4.68619514898634e+00, 4.68551523433299e+00, 4.68441635774152e+00, 4.68273984659233e+00, 4.68029887217475e+00, 4.67688017863671e+00, 4.67224709399698e+00, 4.66614381922498e+00, 4.65830094655049e+00, 4.64844210699164e+00, 4.63629159099004e+00, 4.62158272750208e+00, 4.60406674947506e+00, 4.58352182178312e+00, 4.55976186642632e+00, 4.53264479421721e+00, 4.50207974692576e+00, 4.46803297246390e+00, 4.43053200001408e+00, 4.38966785171333e+00, 4.34559511978937e+00, 4.29852984757802e+00, 4.24874527205162e+00, 4.19656560511545e+00, 4.14235814098948e+00, 4.08652406784605e+00, 4.02948842543681e+00, 3.97168968125480e+00, 3.91356939379549e+00, 3.85556239436680e+00, 3.79808785379194e+00, 3.74154151509517e+00, 3.68628927725871e+00, 3.63266221802942e+00, 3.58095305417153e+00, 3.53141396218415e+00, 3.48425562551078e+00, 3.43964733725258e+00, 3.39771796965196e+00, 3.35855762068189e+00, 3.32221976042034e+00, 3.28872372152831e+00, 3.25805740519325e+00, 3.23018010291656e+00, 3.20502536274172e+00, 3.18250385389193e+00, 3.16250620494383e+00, 3.14490580683195e+00, 3.12956158285186e+00, 3.11632073347496e+00, 3.10502146455607e+00, 3.09549570398293e+00, 3.08757180475045e+00, 3.08107722276716e+00, 3.07584114647537e+00, 3.07169704374099e+00, 3.06848508063731e+00, 3.06605435786728e+00, 3.06426490468005e+00, 3.06298936806532e+00, 3.06211433727902e+00, 3.06154125052850e+00, 3.06118684169165e+00, 3.06098309964743e+00, 3.06087673021440e+00, 3.06082812966148e+00, 3.06080989800026e+00, 3.06080493855429e+00, 3.06080420655073e+00, 3.06080415160022e+00, 3.06080318442497e+00, 3.06079663167475e+00, 3.06077254313757e+00, 3.06070833244682e+00, 3.06056781148453e+00, 3.06029870685563e+00, 3.05983076239202e+00, 3.05907452227651e+00, 3.05792087825690e+00, 3.05624145069497e+00, 3.05388985574820e+00, 3.05070388863578e+00, 3.04650862470574e+00, 3.04112040537942e+00, 3.03435163524356e+00, 3.02601627082127e+00, 3.01593583325344e+00, 3.00394572976605e+00, 2.98990162683873e+00, 2.97368558642800e+00, 2.95521166040651e+00, 2.93443064177818e+00, 2.91133369693491e+00, 2.88595465178012e+00, 2.85837077390451e+00, 2.82870197842557e+00, 2.79710847952452e+00, 2.76378700450051e+00, 2.72896577321263e+00, 2.69289851478910e+00, 2.65585783912442e+00, 2.61812829950404e+00, 2.57999947452768e+00, 2.54175936533913e+00, 2.50368835351652e+00, 2.46605390284373e+00, 2.42910612194551e+00, 2.39307424109743e+00, 2.35816400061485e+00, 2.32455590338443e+00, 2.29240425172363e+00, 2.26183686859715e+00, 2.23295539388172e+00, 2.20583604579447e+00, 2.18053074353884e+00, 2.15706849756125e+00, 2.13545698676753e+00, 2.11568425621254e+00, 2.09772048309149e+00, 2.08151977253979e+00, 2.06702195718819e+00, 2.05415438514444e+00, 2.04283368969746e+00, 2.03296754026749e+00, 2.02445637776530e+00, 2.01719513851933e+00, 2.01107496937924e+00, 2.00598493278525e+00, 2.00181369494943e+00, 1.99845118343583e+00, 1.99579019307663e+00, 1.99372791211709e+00, 1.99216733454627e+00, 1.99101852047361e+00, 1.99019966474829e+00, 1.98963793517621e+00, 1.98927004580601e+00, 1.98904253770091e+00, 1.98891174899078e+00, 1.98884346719688e+00, 1.98881226906937e+00, 1.98880056563744e+00, 1.98879738202612e+00, 1.98879691213197e+00, 1.98879687670743e+00, 1.98879625120441e+00, 1.98879201333274e+00, 1.98877643451713e+00, 1.98873490749029e+00, 1.98864402851478e+00, 1.98846999171233e+00, 1.98816736312083e+00, 1.98767829626186e+00, 1.98693224404868e+00, 1.98584621320019e+00, 1.98432559617766e+00, 1.98226560121493e+00, 1.97955328258451e+00, 1.97607015047406e+00, 1.97169531290085e+00, 1.96630907180774e+00, 1.95979686348025e+00, 1.95205340207937e+00, 1.94298685739898e+00, 1.93252287725656e+00, 1.92060825444934e+00, 1.90721404060264e+00, 1.89233792606607e+00, 1.87600573635264e+00, 1.85827193986500e+00, 1.83921911559803e+00, 1.81895638868349e+00, 1.79761690092163e+00, 1.77535443774214e+00, 1.75233937799251e+00, 1.72875416542616e+00, 1.70478851906685e+00, 1.68063460345636e+00, 1.65648236996131e+00, 1.63251525839257e+00, 1.60890641622181e+00, 1.58581555304969e+00, 1.56338650342694e+00, 1.54174552479649e+00, 1.52100031270984e+00, 1.50123967615854e+00, 1.48253378504680e+00, 1.46493488176666e+00, 1.44847834039129e+00, 1.43318395951421e+00, 1.41905738624735e+00, 1.40609158651434e+00, 1.39426829745831e+00, 1.38355941873194e+00, 1.37392831850159e+00, 1.36533104580938e+00, 1.35771745286130e+00, 1.35103223878726e+00, 1.34521593076906e+00, 1.34020581967340e+00, 1.33593686604919e+00, 1.33234258913130e+00, 1.32935594687839e+00, 1.32691020956089e+00, 1.32493982347711e+00, 1.32338125546825e+00, 1.32217380347163e+00, 1.32126035382450e+00, 1.32058806279755e+00, 1.32010893821567e+00, 1.31978029722576e+00, 1.31956507837471e+00, 1.31943199008113e+00, 1.31935548308404e+00, 1.31931554114826e+00, 1.31929729171588e+00, 1.31929044578327e+00, 1.31928858352924e+00, 1.31928830866464e+00, 1.31928828783029e+00, 1.31928791845542e+00, 1.31928541588926e+00, 1.31927621623528e+00, 1.31925169367288e+00, 1.31919802835733e+00, 1.31909525941949e+00, 1.31891656390439e+00, 1.31862779955295e+00, 1.31818734575804e+00, 1.31754627234119e+00, 1.31664885938786e+00, 1.31543348259657e+00, 1.31383386687326e+00, 1.31178069590883e+00, 1.30920354727351e+00, 1.30603310173227e+00, 1.30220355322771e+00, 1.29765512411015e+00, 1.29233657106303e+00, 1.28620755342760e+00, 1.27924072990320e+00, 1.27142345407041e+00, 1.26275895514108e+00, 1.25326691776975e+00, 1.24298341208970e+00, 1.23196016920393e+00, 1.22026324368252e+00, 1.20797114796221e+00, 1.19517257875223e+00, 1.18196387843881e+00, 1.16844638264589e+00, 1.15472379839347e+00, 1.14089973778361e+00, 1.12707550366528e+00, 1.11334819099424e+00, 1.09980913523331e+00, 1.08654271083300e+00, 1.07362546089610e+00, 1.06112552438823e+00, 1.04910231930520e+00, 1.03760643785219e+00, 1.02667971139449e+00, 1.01635540719924e+00, 1.00665852454858e+00, 9.97606163752451e-01, 9.89207947333785e-01, 9.81466477871830e-01, 9.74377821516029e-01, 9.67932009974348e-01, 9.62113556835686e-01, 9.56901986421197e-01, 9.52272374992148e-01, 9.48195905087523e-01, 9.44640434039666e-01, 9.41571077348055e-01, 9.38950806624405e-01, 9.36741060324813e-01, 9.34902363558525e-01, 9.33394951040434e-01, 9.32179384901847e-01, 9.31217156784331e-01, 9.30471261622650e-01, 9.29906728985184e-01, 9.29491096976172e-01, 9.29194813668403e-01, 9.28991551925034e-01, 9.28858425312278e-01, 9.28776095544890e-01, 9.28728765411833e-01, 9.28704055196210e-01, 9.28692764981196e-01, 9.28688529645765e-01, 9.28687377533162e-01, 9.28687207483804e-01, 9.28687194512509e-01, 9.28686963463765e-01, 9.28685398075437e-01, 9.28679643555867e-01, 9.28664304223282e-01, 9.28630734995895e-01, 9.28566447927835e-01, 9.28454658491347e-01, 9.28273993987695e-01, 9.27998383721143e-01, 9.27597147195840e-01, 9.27035292335621e-01, 9.26274030306352e-01, 9.25271506708746e-01, 9.23983740590479e-01, 9.22365752974585e-01, 9.20372855741086e-01, 9.17962060347897e-01, 9.15093554936515e-01, 9.11732188968266e-01, 9.07848897913812e-01, 9.03421997843267e-01, 8.98438281937246e-01, 8.92893858396413e-01, 8.86794681793255e-01, 8.80156746750885e-01, 8.73005932518964e-01, 8.65377507706224e-01, 8.57315324132500e-01, 8.48870745638013e-01, 8.40101370297894e-01, 8.31069611980425e-01, 8.21841209348150e-01, 8.12483727616979e-01, 8.03065111508204e-01, 7.93652337986241e-01, 7.84310205810051e-01, 7.75100286827939e-01, 7.66080052337863e-01, 7.57302177511502e-01, 7.48814018371899e-01, 7.40657249398247e-01, 7.32867645554364e-01, 7.25474990261354e-01, 7.18503090285115e-01, 7.11969879329008e-01, 7.05887593919858e-01, 7.00263007568850e-01, 6.95097711833853e-01, 6.90388435520603e-01, 6.86127395617984e-01, 6.82302675515015e-01, 6.78898627500978e-01, 6.75896297463090e-01, 6.73273870065838e-01, 6.71007132552930e-01, 6.69069954711041e-01, 6.67434781551915e-01, 6.66073134000534e-01, 6.64956111435263e-01, 6.64054888437871e-01, 6.63341196714896e-01, 6.62787781987948e-01, 6.62368824855587e-01, 6.62060314323110e-01, 6.61840362970565e-01, 6.61689453636015e-01, 6.61590609035425e-01, 6.61529477874822e-01, 6.61494333637730e-01, 6.61475985210565e-01, 6.61467601669905e-01, 6.61464456715549e-01, 6.61463601211818e-01, 6.61463474941261e-01, 6.61463465241422e-01, 6.61463291573874e-01, 6.61462114950972e-01, 6.61457789565237e-01, 6.61446259709631e-01, 6.61421027078703e-01, 6.61372704246350e-01, 6.61288672545290e-01, 6.61152860637021e-01, 6.60945657790744e-01, 6.60643973179532e-01, 6.60221449209885e-01, 6.59648832875575e-01, 6.58894504244845e-01, 6.57925155398917e-01, 6.56706606498094e-01, 6.55204738368294e-01, 6.53386513455878e-01, 6.51221049740183e-01, 6.48680705909239e-01, 6.45742131551970e-01, 6.42387234034615e-01, 6.38604014703315e-01, 6.34387231426626e-01, 6.29738852246017e-01, 6.24668275623790e-01, 6.19192305660352e-01, 6.13334884583086e-01, 6.07126598499860e-01, 6.00603984583046e-01, 5.93808677406788e-01, 5.86786438342351e-01, 5.79586114391685e-01, 5.72258571728749e-01, 5.64855645038914e-01, 5.57429137305242e-01, 5.50029896935943e-01, 5.42706991003019e-01, 5.35506985680527e-01, 5.28473338328215e-01, 5.21645900406519e-01, 5.15060526636969e-01, 5.08748783447832e-01, 5.02737748548938e-01, 4.97049893180225e-01, 4.91703038888456e-01, 4.86710381351717e-01, 4.82080574592665e-01, 4.77817869756823e-01, 4.73922303389802e-01, 4.70389930772349e-01, 4.67213100332160e-01, 4.64380765427126e-01, 4.61878829872383e-01, 4.59690523454391e-01, 4.57796803336795e-01, 4.56176776724206e-01, 4.54808139434615e-01, 4.53667624181431e-01, 4.52731451443832e-01, 4.51975774888663e-01, 4.51377112491844e-01, 4.50912753891171e-01, 4.50561134183001e-01, 4.50302164436437e-01, 4.50117509703322e-01, 4.49990806282447e-01, 4.49907811447805e-01, 4.49856480731153e-01, 4.49826970080835e-01, 4.49811562691977e-01, 4.49804522889963e-01, 4.49801882010906e-01, 4.49801163627178e-01, 4.49801057595220e-01, 4.49801049387919e-01, 4.49800901633584e-01, 4.49799900576188e-01, 4.49796220582574e-01, 4.49786411065073e-01, 4.49764943147365e-01, 4.49723829407048e-01, 4.49652331924042e-01, 4.49536772058757e-01, 4.49360453285841e-01, 4.49103705961090e-01, 4.48744060022328e-01, 4.48256548235531e-01, 4.47614138632751e-01, 4.46788290232035e-01, 4.45749621042069e-01, 4.44468671893226e-01, 4.42916744072317e-01, 4.41066783448253e-01, 4.38894279224250e-01, 4.36378142143237e-01, 4.33501525387310e-01, 4.30252551930518e-01, 4.26624914932553e-01, 4.22618322873738e-01, 4.18238768244343e-01, 4.13498607189876e-01, 4.08416446873439e-01, 4.03016846660286e-01, 3.97329847797767e-01, 3.91390353434893e-01, 3.85237386183348e-01, 3.78913253794739e-01, 3.72462654967499e-01, 3.65931757022143e-01, 3.59367275512504e-01, 3.52815583111823e-01, 3.46321871636736e-01, 3.39929387101396e-01, 3.33678753426430e-01, 3.27607396024337e-01, 3.21749072092498e-01, 3.16133510220183e-01, 3.10786158019916e-01, 3.05728033089423e-01, 3.00975669843758e-01, 2.96541152735017e-01, 2.92432225148257e-01, 2.88652462811203e-01, 2.85201500797974e-01, 2.82075304007076e-01, 2.79266472173614e-01, 2.76764571843316e-01, 2.74556489101737e-01, 2.72626798050208e-01, 2.70958140922758e-01, 2.69531616263031e-01, 2.68327171698267e-01, 2.67323997578806e-01, 2.66500917163633e-01, 2.65836768229374e-01, 2.65310770092043e-01, 2.64902869199496e-01, 2.64594055818254e-01, 2.64366644025092e-01, 2.64204507315519e-01, 2.64093262715544e-01, 2.64020397341714e-01, 2.63975332865122e-01, 2.63949425223920e-01, 2.63935899088224e-01, 2.63929718882341e-01, 2.63927400473300e-01, 2.63926769809785e-01, 2.63926676725170e-01, 2.63926669461625e-01, 2.63926537943036e-01, 2.63925646885426e-01, 2.63922371266499e-01, 2.63913639692856e-01, 2.63894530978160e-01, 2.63857935966040e-01, 2.63794298381143e-01, 2.63691447017738e-01, 2.63534529853967e-01, 2.63306058640470e-01, 2.62986070061249e-01, 2.62552406604726e-01, 2.61981116738461e-01, 2.61246969823670e-01, 2.60324076478337e-01, 2.59186599934463e-01, 2.57809538568669e-01, 2.56169554545021e-01, 2.54245818801557e-01, 2.52020838887255e-01, 2.49481233862682e-01, 2.46618420005685e-01, 2.43429172694153e-01, 2.39916033693930e-01, 2.36087539096172e-01, 2.31958251064462e-01, 2.27548585921799e-01, 2.22884441338106e-01, 2.17996635776167e-01, 2.12920183182419e-01, 2.07693434454836e-01, 2.02357123848710e-01, 1.96953362695840e-01, 1.91524624302755e-01, 1.86112762568278e-01, 1.80758102871948e-01, 1.75498637514824e-01, 1.70369350027528e-01, 1.65401683717363e-01, 1.60623160684349e-01, 1.56057148939468e-01, 1.51722767838081e-01, 1.47634916252166e-01, 1.43804403996270e-01, 1.40238165033075e-01, 1.36939530773162e-01, 1.33908543068696e-01, 1.31142288916795e-01, 1.28635242038563e-01, 1.26379599998291e-01, 1.24365609035952e-01, 1.22581872026895e-01, 1.21015637752624e-01, 1.19653071831916e-01, 1.18479511154579e-01, 1.17479704470917e-01, 1.16638041955792e-01, 1.15938776164654e-01, 1.15366235939025e-01, 1.14905033634439e-01, 1.14540264683168e-01, 1.14257697124270e-01, 1.14043947486120e-01, 1.13886638429335e-01, 1.13774532960981e-01, 1.13697639889693e-01, 1.13647285534991e-01, 1.13616147518794e-01, 1.13598247691947e-01, 1.13588902787097e-01, 1.13584633118325e-01, 1.13583031433599e-01, 1.13582595737977e-01, 1.13582531430271e-01, 1.13582526369537e-01, 1.13582434189464e-01, 1.13581809656807e-01, 1.13579513822334e-01, 1.13573394075090e-01, 1.13560001688558e-01, 1.13534355699642e-01, 1.13489763473666e-01, 1.13417707670934e-01, 1.13307808545478e-01, 1.13147869443999e-01, 1.12924011867182e-01, 1.12620904332481e-01, 1.12222086387259e-01, 1.11710385374074e-01, 1.11068418953376e-01, 1.10279171075298e-01, 1.09326623337971e-01, 1.08196417894657e-01, 1.06876522818009e-01, 1.05357866717505e-01, 1.03634907065661e-01, 1.01706096682271e-01, 9.95742155514989e-02, 9.72465407636101e-02, 9.47348357306765e-02, 9.20551504427785e-02, 8.92274366064392e-02, 8.62749939892851e-02, 8.32237759921242e-02, 8.01015921975292e-02, 7.69372523834093e-02, 7.37596995358665e-02, 7.05971784625206e-02, 6.74764818774563e-02, 6.44223079225057e-02, 6.14567529714449e-02, 5.85989523920212e-02, 5.58648708996472e-02, 5.32672342714982e-02, 5.08155862888451e-02, 4.85164492964998e-02, 4.63735638336544e-02, 4.43881822227077e-02, 4.25593924105385e-02, 4.08844512191493e-02, 3.93591099306683e-02, 3.79779192926328e-02, 3.67345051616610e-02, 3.56218097943994e-02, 3.46322970442969e-02, 3.37581223311361e-02, 3.29912701953141e-02, 3.23236635646857e-02, 3.17472496135842e-02, 3.12540673625428e-02, 3.08363020352580e-02, 3.04863307354850e-02, 3.01967633034685e-02, 2.99604813281711e-02, 2.97706772941098e-02, 2.96208947974925e-02, 2.95050697437169e-02, 2.94175715052489e-02, 2.93532422409584e-02, 2.93074320116773e-02, 2.92760270144754e-02, 2.92554682220896e-02, 2.92427579522077e-02, 2.92354523766392e-02, 2.92316386624911e-02, 2.92298962488182e-02, 2.92292426260872e-02, 2.92290648264455e-02, 2.92290385836438e-02, 2.92290365001691e-02, 2.92289983178655e-02, 2.92287396280244e-02, 2.92277886703189e-02, 2.92252538699991e-02, 2.92197070481031e-02, 2.92090862383638e-02, 2.91906228744594e-02, 2.91607981165072e-02, 2.91153330424834e-02, 2.90492172232888e-02, 2.89567796838881e-02, 2.88318052708085e-02, 2.86676978594613e-02, 2.84576895520736e-02, 2.81950920325542e-02, 2.78735826651051e-02, 2.74875139911472e-02, 2.70322313722993e-02, 2.65043801379434e-02, 2.59021812776830e-02, 2.52256540103204e-02, 2.44767648868134e-02, 2.36594866598736e-02, 2.27797559001708e-02, 2.18453258486981e-02, 2.08655195405250e-02, 1.98508968550684e-02, 1.88128567799808e-02, 1.77632018321745e-02, 1.67136945131752e-02, 1.56756355211272e-02, 1.46594902733739e-02, 1.36745846172701e-02, 1.27288832585997e-02, 1.18288564302240e-02, 1.09794326687371e-02, 1.01840291091015e-02, 9.44464600909711e-03, 8.76200951771366e-03, 8.13574594334199e-03, 7.56457166700386e-03, 7.04648494918203e-03, 6.57894870570175e-03, 6.15905641388537e-03, 5.78367627435483e-03, 5.44957133870128e-03, 5.15349538818199e-03, 4.89226588171476e-03, 4.66281631602072e-03, 4.46223092320608e-03, 4.28776485376403e-03, 4.13685294876842e-03, 4.00710998361120e-03, 3.89632494693970e-03, 3.80245155862369e-03, 3.72359686420509e-03, 3.65800938560312e-03, 3.60406796519896e-03, 3.56027211439169e-03, 3.52523437052810e-03, 3.49767487999131e-03, 3.47641816459188e-03, 3.46039179801849e-03, 3.44862652798383e-03, 3.44025723560293e-03, 3.43452403540212e-03, 3.43077279384829e-03, 3.42845438159856e-03, 3.42712207206475e-03, 3.42642664514159e-03, 3.42610893508066e-03, 3.42598975692709e-03, 3.42595733813766e-03, 3.42595255320788e-03, 3.42595216984520e-03, 3.42594510042460e-03, 3.42589720435882e-03, 3.42572113803286e-03, 3.42525184576914e-03, 3.42422499335981e-03, 3.42225914630490e-03, 3.41884270558879e-03, 3.41332668838845e-03, 3.40492450790464e-03, 3.39271992648804e-03, 3.37568427686818e-03, 3.35270381347641e-03, 3.32261761741748e-03, 3.28426580497876e-03, 3.23654687834107e-03, 3.17848195632564e-03, 3.10928243144559e-03, 3.02841647130592e-03, 2.93566891151187e-03, 2.83118868177112e-03, 2.71551814817172e-03, 2.58959975490214e-03, 2.45475710559617e-03, 2.31264999820624e-03, 2.16520563556344e-03, 2.01453088818226e-03, 1.86281266040922e-03, 1.71221473633665e-03, 1.56477973692049e-03, 1.42234398775898e-03, 1.28647135712256e-03, 1.15840978531523e-03, 1.03907164662687e-03, 9.29036607974609e-04, 8.28573587664749e-04, 7.37677036742213e-04, 6.56112220372762e-04, 5.83464440392589e-04, 5.19187974486357e-04, 4.62651567399774e-04, 4.13178319956131e-04, 3.70078693819664e-04, 3.32676134167542e-04, 3.00325547016200e-04, 2.72425473703208e-04, 2.48425138789088e-04, 2.27827564355465e-04, 2.10189751179257e-04, 1.95120702638718e-04, 1.82277929450692e-04, 1.71363021042481e-04, 1.62116812479240e-04, 1.54314556592539e-04, 1.47761356782183e-04, 1.42287984694178e-04, 1.37747126636611e-04, 1.34010068407520e-04, 1.30963821717955e-04, 1.28508695176777e-04, 1.26556304181236e-04, 1.25027997778915e-04, 1.23853662568507e-04, 1.22970850785720e-04, 1.22324171545785e-04, 1.21864880743061e-04, 1.21550603555676e-04, 1.21345125652539e-04, 1.21218193374517e-04, 1.21145270977507e-04, 1.21107213722382e-04, 1.21089828432730e-04, 1.21083307153214e-04, 1.21081533261957e-04, 1.21081271434065e-04, 1.21081250266997e-04, 1.21080857471867e-04, 1.21078196233526e-04, 1.21068413690750e-04, 1.21042340367694e-04, 1.20985296458542e-04, 1.20876115417445e-04, 1.20686451570151e-04, 1.20380448670687e-04, 1.19914856067993e-04, 1.19239684885031e-04, 1.18299493561125e-04, 1.17035374245233e-04, 1.15387673410794e-04, 1.13299416437229e-04, 1.10720315621649e-04, 1.07611126683338e-04, 1.03947995640397e-04, 9.97263310544838e-05, 9.49636852887181e-05, 8.97011704132922e-05, 8.40030891282595e-05, 7.79547095959288e-05, 7.16583867266252e-05, 6.52284225708399e-05, 5.87850651925237e-05, 5.24478955483687e-05, 4.63287943155422e-05, 4.05249926274109e-05, 3.51132555316269e-05, 3.01464901966142e-05, 2.56537825791270e-05, 2.16441026275049e-05, 1.81122870692718e-05, 1.50441367381404e-05, 1.24185556167333e-05, 1.02082130020165e-05, 8.38043058237758e-06, 6.89849029100742e-06, 5.72035319024167e-06, 4.77529084719208e-06, 3.98226290698744e-06, 3.32297912919444e-06, 2.77853377994575e-06, 2.22958304618714e-06, 1.84655358737018e-06, 1.61867767900503e-06, 1.37398696153923e-06, 1.11802403779393e-06, 8.63924420002604e-07, 6.19176127258925e-07, 3.84660655328162e-07, -1.06581410364015e-14, -7.10542735760100e-15, -1.06581410364015e-14, -1.06581410364015e-14, -7.10542735760100e-15, -1.06581410364015e-14, -7.10542735760100e-15, -1.06581410364015e-14, -3.55271367880050e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -3.55271367880050e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15],
[ 2.60000000000000e+01, 2.59999999822682e+01, 2.59999994339645e+01, 2.59999957191206e+01, 2.59999820628758e+01, 2.59999456596841e+01, 2.59998659888539e+01, 2.59997133969015e+01, 2.59994479982152e+01, 2.59990189359654e+01, 2.59983640343456e+01, 2.59974098614673e+01, 2.59960722099663e+01, 2.59942569900597e+01, 2.59918615178440e+01, 2.59887761704773e+01, 2.59848863698842e+01, 2.59800748480291e+01, 2.59742241397848e+01, 2.59672192440210e+01, 2.59589503896649e+01, 2.59493158409984e+01, 2.59382246751158e+01, 2.59255994640705e+01, 2.59113787946220e+01, 2.58955195596111e+01, 2.58779989569298e+01, 2.58588161350604e+01, 2.58379934285885e+01, 2.58155771333618e+01, 2.57916377794546e+01, 2.57662698710472e+01, 2.57395910757729e+01, 2.57117408617657e+01, 2.56828785979949e+01, 2.56531811516766e+01, 2.56228400345686e+01, 2.55920581666815e+01, 2.55610463402908e+01, 2.55300194782177e+01, 2.54991927874971e+01, 2.54687779124433e+01, 2.54389791897642e+01, 2.54099901030527e+01, 2.53819900252624e+01, 2.53551413263507e+01, 2.53295869099235e+01, 2.53054482282168e+01, 2.52828238097923e+01, 2.52617883194831e+01, 2.52423921558447e+01, 2.52246615779597e+01, 2.52085993410930e+01, 2.51941858095088e+01, 2.51813805047872e+01, 2.51701240392300e+01, 2.51603403764560e+01, 2.51519393550705e+01, 2.51448194064429e+01, 2.51388703941935e+01, 2.51339765011106e+01, 2.51300190889719e+01, 2.51268794582365e+01, 2.51244414378524e+01, 2.51225937405084e+01, 2.51212320254784e+01, 2.51202606196613e+01, 2.51195938572978e+01, 2.51191570099077e+01, 2.51188867899255e+01, 2.51187314239576e+01, 2.51186503041885e+01, 2.51186132388155e+01, 2.51185993341396e+01, 2.51185955517142e+01, 2.51185949934355e+01, 2.51185949202011e+01, 2.51185932139275e+01, 2.51185816536593e+01, 2.51185391566905e+01, 2.51184258738365e+01, 2.51181779492778e+01, 2.51177031141877e+01, 2.51168772784261e+01, 2.51155422571499e+01, 2.51135047380257e+01, 2.51105365606103e+01, 2.51063763432859e+01, 2.51007324558740e+01, 2.50932872987320e+01, 2.50837028127887e+01, 2.50716271104784e+01, 2.50567020855284e+01, 2.50385718303071e+01, 2.50168916628122e+01, 2.49913375409343e+01, 2.49616156188540e+01, 2.49274716791403e+01, 2.48887001549045e+01, 2.48451524409788e+01, 2.47967441846003e+01, 2.47434612486544e+01, 2.46853640587619e+01, 2.46225900833774e+01, 2.45553542557909e+01, 2.44839472277985e+01, 2.44087314426708e+01, 2.43301351223928e+01, 2.42486443710386e+01, 2.41647936917487e+01, 2.40791552892645e+01, 2.39923275761606e+01, 2.39049233155182e+01, 2.38175578167178e+01, 2.37308375590368e+01, 2.36453495571887e+01, 2.35616517124246e+01, 2.34802643206238e+01, 2.34016628417782e+01, 2.33262719781040e+01, 2.32544610630513e+01, 2.31865407309754e+01, 2.31227608158625e+01, 2.30633094149975e+01, 2.30083130471777e+01, 2.29578378324953e+01, 2.29118916196566e+01, 2.28704269856851e+01, 2.28333450305740e+01, 2.28004998854313e+01, 2.27717038467363e+01, 2.27467330416202e+01, 2.27253335200167e+01, 2.27072276596628e+01, 2.26921207600188e+01, 2.26797076920569e+01, 2.26696794634524e+01, 2.26617295539409e+01, 2.26555598743440e+01, 2.26508862058026e+01, 2.26474429836414e+01, 2.26449873033542e+01, 2.26433020444354e+01, 2.26421980308575e+01, 2.26415151742179e+01, 2.26411225759716e+01, 2.26409175974919e+01, 2.26408239395756e+01, 2.26407888050060e+01, 2.26407792475275e+01, 2.26407778368628e+01, 2.26407777123280e+01, 2.26407752721289e+01, 2.26407587394142e+01, 2.26406979632968e+01, 2.26405359556420e+01, 2.26401814015909e+01, 2.26395023718380e+01, 2.26383214822968e+01, 2.26364127136260e+01, 2.26335000648812e+01, 2.26292581720782e+01, 2.26233149732624e+01, 2.26152564463293e+01, 2.26046333841070e+01, 2.25909701033866e+01, 2.25737749117090e+01, 2.25525520798202e+01, 2.25268149919026e+01, 2.24961000741978e+01, 2.24599810405902e+01, 2.24180829468667e+01, 2.23700955195042e+01, 2.23157852251362e+01, 2.22550055770381e+01, 2.21877052364354e+01, 2.21139335575168e+01, 2.20338433405610e+01, 2.19476906890397e+01, 2.18558320028154e+01, 2.17587182684370e+01, 2.16568869177269e+01, 2.15509516089513e+01, 2.14415903368944e+01, 2.13295323001302e+01, 2.12155439510501e+01, 2.11004146346584e+01, 2.09849421939209e+01, 2.08699188886528e+01, 2.07561179443359e+01, 2.06442810161650e+01, 2.05351068191241e+01, 2.04292411337254e+01, 2.03272683475737e+01, 2.02297046362859e+01, 2.01369928274236e+01, 2.00494989335990e+01, 1.99675102914621e+01, 1.98912352060122e+01, 1.98208039762806e+01, 1.97562711680416e+01, 1.96976189989803e+01, 1.96447617078483e+01, 1.95975507876600e+01, 1.95557809706816e+01, 1.95191968575823e+01, 1.94875000835019e+01, 1.94603569097311e+01, 1.94374061217580e+01, 1.94182671037056e+01, 1.94025479471068e+01, 1.93898534401926e+01, 1.93797927740869e+01, 1.93719867961702e+01, 1.93660746398417e+01, 1.93617195651504e+01, 1.93586138570472e+01, 1.93564826476047e+01, 1.93550865551989e+01, 1.93542230665457e+01, 1.93537266253204e+01, 1.93534674321702e+01, 1.93533490032884e+01, 1.93533045763012e+01, 1.93532924910622e+01, 1.93532907073055e+01, 1.93532905642262e+01, 1.93532879236862e+01, 1.93532700336340e+01, 1.93532042678275e+01, 1.93530289597012e+01, 1.93526452989479e+01, 1.93519105304186e+01, 1.93506327233244e+01, 1.93485673440926e+01, 1.93454158262333e+01, 1.93408262847354e+01, 1.93343964700553e+01, 1.93256789969171e+01, 1.93141888153952e+01, 1.92994128162131e+01, 1.92808213799408e+01, 1.92578815930541e+01, 1.92300717662420e+01, 1.91968968068257e+01, 1.91579039237101e+01, 1.91126980866663e+01, 1.90609566288130e+01, 1.90024423782127e+01, 1.89370147364066e+01, 1.88646381910790e+01, 1.87853878563216e+01, 1.86994517726913e+01, 1.86071298615885e+01, 1.85088296013401e+01, 1.84050586595163e+01, 1.82964148602608e+01, 1.81835739718883e+01, 1.80672758594776e+01, 1.79483095586047e+01, 1.78274977970568e+01, 1.77056814348663e+01, 1.75837042248062e+01, 1.74623982286170e+01, 1.73425701661907e+01, 1.72249889270427e+01, 1.71103744324638e+01, 1.69993879978293e+01, 1.68926243036766e+01, 1.67906050398825e+01, 1.66937742408631e+01, 1.66024952843770e+01, 1.65170494860554e+01, 1.64376361894356e+01, 1.63643742288894e+01, 1.62973046306172e+01, 1.62363944135490e+01, 1.61815413552529e+01, 1.61325795950001e+01, 1.60892859541874e+01, 1.60513868609813e+01, 1.60185657695929e+01, 1.59904709640465e+01, 1.59667236315044e+01, 1.59469260816792e+01, 1.59306699777807e+01, 1.59175444323802e+01, 1.59071438104344e+01, 1.58990750734389e+01, 1.58929644951445e+01, 1.58884635820253e+01, 1.58852540418524e+01, 1.58830516618214e+01, 1.58816089836087e+01, 1.58807166956996e+01, 1.58802037019939e+01, 1.58799358681853e+01, 1.58798134915957e+01, 1.58797675837324e+01, 1.58797550956635e+01, 1.58797532524501e+01, 1.58797531102837e+01, 1.58797505574321e+01, 1.58797332614831e+01, 1.58796696796833e+01, 1.58795001934755e+01, 1.58791292746520e+01, 1.58784189115228e+01, 1.58771835552690e+01, 1.58751868137987e+01, 1.58721400824541e+01, 1.58677032572474e+01, 1.58614876260940e+01, 1.58530609758869e+01, 1.58419548876346e+01, 1.58276741183550e+01, 1.58097078880082e+01, 1.57875428047262e+01, 1.57606770755469e+01, 1.57286355676515e+01, 1.56909852126255e+01, 1.56473501899472e+01, 1.55974262921058e+01, 1.55409938680037e+01, 1.54779287675728e+01, 1.54082107706574e+01, 1.53319290763789e+01, 1.52492845518445e+01, 1.51605885849389e+01, 1.50662585464304e+01, 1.49668100312731e+01, 1.48628462062085e+01, 1.47550447287662e+01, 1.46441428106247e+01, 1.45309210672598e+01, 1.44161868204911e+01, 1.43007574998981e+01, 1.41854447268428e+01, 1.40710395692030e+01, 1.39582993374292e+01, 1.38479361664031e+01, 1.37406075056512e+01, 1.36369085335208e+01, 1.35373664262570e+01, 1.34424363539073e+01, 1.33524990412385e+01, 1.32678597200620e+01, 1.31887483043907e+01, 1.31153206359073e+01, 1.30476606687991e+01, 1.29857834855692e+01, 1.29296390556147e+01, 1.28791166640834e+01, 1.28340499487268e+01, 1.27942224868986e+01, 1.27593738737939e+01, 1.27292062270968e+01, 1.27033910432435e+01, 1.26815763175058e+01, 1.26633938251769e+01, 1.26484664455687e+01, 1.26364153956793e+01, 1.26268672277307e+01, 1.26194604357747e+01, 1.26138515126120e+01, 1.26097203005158e+01, 1.26067744885414e+01, 1.26047531258996e+01, 1.26034290448184e+01, 1.26026101168061e+01, 1.26021393019933e+01, 1.26018934905965e+01, 1.26017811765092e+01, 1.26017390434832e+01, 1.26017275822659e+01, 1.26017258906138e+01, 1.26017257625270e+01, 1.26017234934821e+01, 1.26017081203662e+01, 1.26016516071221e+01, 1.26015009632261e+01, 1.26011712810387e+01, 1.26005398936727e+01, 1.25994418852825e+01, 1.25976671568394e+01, 1.25949592170362e+01, 1.25910158293362e+01, 1.25854916014033e+01, 1.25780025512439e+01, 1.25681326246469e+01, 1.25554420708465e+01, 1.25394775086930e+01, 1.25197834362147e+01, 1.24959148560094e+01, 1.24674506125749e+01, 1.24340069718326e+01, 1.23952509247084e+01, 1.23509126725390e+01, 1.23007967579446e+01, 1.22447913441253e+01, 1.21828752186483e+01, 1.21151222013673e+01, 1.20417027630213e+01, 1.19628828009803e+01, 1.18790196592018e+01, 1.17905556081116e+01, 1.16980091058186e+01, 1.16019642372252e+01, 1.15030587691543e+01, 1.14019712693763e+01, 1.12994077212823e+01, 1.11960880322181e+01, 1.10927327908807e+01, 1.09900505846710e+01, 1.08887261455356e+01, 1.07894095532698e+01, 1.06927066866091e+01, 1.05991710717358e+01, 1.05092972327800e+01, 1.04235155991723e+01, 1.03421889724448e+01, 1.02656105043657e+01, 1.01940030940325e+01, 1.01275200779982e+01, 1.00662470672027e+01, 1.00102047776433e+01, 9.95935270643857e+00, 9.91359351775788e+00, 9.87277801994936e+00, 9.83671063220293e+00, 9.80515525313264e+00, 9.77784145274412e+00, 9.75447091248093e+00, 9.73472403553852e+00, 9.71826664229501e+00, 9.70475665498735e+00, 9.69385066344763e+00, 9.68521025174153e+00, 9.67850795584328e+00, 9.67343271679366e+00, 9.66969469357960e+00, 9.66702930628323e+00, 9.66520039339384e+00, 9.66400238750684e+00, 9.66326144030843e+00, 9.66283545959257e+00, 9.66261305646888e+00, 9.66251143800538e+00, 9.66247331730802e+00, 9.66246294754428e+00, 9.66246141698852e+00, 9.66246130208922e+00, 9.66245927973110e+00, 9.66244557795518e+00, 9.66239520874592e+00, 9.66226094268442e+00, 9.66196710346159e+00, 9.66140436147987e+00, 9.66042573369277e+00, 9.65884397140492e+00, 9.65643048830967e+00, 9.65291594700333e+00, 9.64799258276179e+00, 9.64131829764045e+00, 9.63252250535368e+00, 9.62121364783863e+00, 9.60698823865726e+00, 9.58944121822509e+00, 9.56817733419716e+00, 9.54282319123255e+00, 9.51303955276676e+00, 9.47853342890412e+00, 9.43906945477287e+00, 9.39448005789200e+00, 9.34467393547139e+00, 9.28964241566873e+00, 9.22946336104838e+00, 9.16430238552830e+00, 9.09441129252735e+00, 9.02012379296761e+00, 8.94184871500833e+00, 8.86006105802243e+00, 8.77529135560435e+00, 8.68811388240013e+00, 8.59913425854674e+00, 8.50897697284170e+00, 8.41827327000474e+00, 8.32764974482037e+00, 8.23771787650328e+00, 8.14906463860805e+00, 8.06224424534182e+00, 7.97777104806303e+00, 7.89611357240647e+00, 7.81768967869059e+00, 7.74286282697409e+00, 7.67193942626398e+00, 7.60516724123661e+00, 7.54273481900469e+00, 7.48477188478070e+00, 7.43135064149116e+00, 7.38248789698809e+00, 7.33814793504597e+00, 7.29824604321404e+00, 7.26265261116676e+00, 7.23119771610524e+00, 7.20367611536221e+00, 7.17985256907740e+00, 7.15946741642846e+00, 7.14224232674716e+00, 7.12788614185298e+00, 7.11610071861182e+00, 7.10658667208206e+00, 7.09904891100048e+00, 7.09320185031688e+00, 7.08877418152987e+00, 7.08551308204291e+00, 7.08318775064577e+00, 7.08159216807242e+00, 7.08054699941280e+00, 7.07990057845299e+00, 7.07952894178712e+00, 7.07933491140151e+00, 7.07924625673266e+00, 7.07921299921135e+00, 7.07920395235027e+00, 7.07920261705220e+00, 7.07920251711577e+00, 7.07920076218190e+00, 7.07918887224464e+00, 7.07914516354654e+00, 7.07902865196696e+00, 7.07877366801668e+00, 7.07828533861904e+00, 7.07743611521865e+00, 7.07606350557405e+00, 7.07396914091462e+00, 7.07091928051192e+00, 7.06664682145825e+00, 7.06085484187451e+00, 7.05322166041974e+00, 7.04340734386314e+00, 7.03106153833817e+00, 7.01583244036520e+00, 6.99737666342244e+00, 6.97536969831434e+00, 6.94951661508964e+00, 6.91956261537972e+00, 6.88530302115312e+00, 6.84659228263409e+00, 6.80335160682341e+00, 6.75557484929821e+00, 6.70333237450998e+00, 6.64677267066213e+00, 6.58612160005408e+00, 6.52167926923049e+00, 6.45381460961761e+00, 6.38295786270448e+00, 6.30959125847865e+00, 6.23423825623430e+00, 6.15745177792824e+00, 6.07980190160190e+00, 6.00186349290261e+00, 5.92420423517239e+00, 5.84737347403533e+00, 5.77189222459771e+00, 5.69824460432421e+00, 5.62687086010169e+00, 5.55816206236467e+00, 5.49245645041459e+00, 5.43003733777006e+00, 5.37113242896604e+00, 5.31591436169883e+00, 5.26450227031140e+00, 5.21696416616244e+00, 5.17331994396018e+00, 5.13354484653723e+00, 5.09757324956894e+00, 5.06530265850926e+00, 5.03659783927566e+00, 5.01129502947563e+00, 4.98920619655675e+00, 4.97012332227430e+00, 4.95382269909179e+00, 4.94006922393628e+00, 4.92862066899535e+00, 4.91923189922671e+00, 4.91165899351260e+00, 4.90566321265916e+00, 4.90101474451191e+00, 4.89749614604627e+00, 4.89490539590524e+00, 4.89305846965231e+00, 4.89179135469804e+00, 4.89096143262240e+00, 4.89044817308709e+00, 4.89015310483082e+00, 4.88999905406290e+00, 4.88992866732668e+00, 4.88990226288561e+00, 4.88989508024249e+00, 4.88989402009960e+00, 4.88989394072666e+00, 4.88989254650114e+00, 4.88988310042167e+00, 4.88984837568697e+00, 4.88975581269807e+00, 4.88955324291412e+00, 4.88916530541317e+00, 4.88849070347534e+00, 4.88740043356002e+00, 4.88573710151611e+00, 4.88331541984273e+00, 4.87992395412003e+00, 4.87532815511515e+00, 4.86927467453111e+00, 4.86149691644956e+00, 4.85172172361113e+00, 4.83967703925397e+00, 4.82510032398159e+00, 4.80774744695168e+00, 4.78740171647579e+00, 4.76388267244455e+00, 4.73705423748924e+00, 4.70683182052693e+00, 4.67318798907682e+00, 4.63615637724505e+00, 4.59583357376572e+00, 4.55237883530318e+00, 4.50601158787907e+00, 4.45700680488376e+00, 4.40568847316926e+00, 4.35242146825454e+00, 4.29760224566060e+00, 4.24164881002053e+00, 4.18499044240042e+00, 4.12805764871892e+00, 4.07127274178677e+00, 4.01504139327522e+00, 3.95974539918474e+00, 3.90573680339698e+00, 3.85333342843197e+00, 3.80281577867641e+00, 3.75442521467096e+00, 3.70836325032228e+00, 3.66479179829271e+00, 3.62383418036378e+00, 3.58557672592592e+00, 3.55007079890389e+00, 3.51733511734121e+00, 3.48735825690909e+00, 3.46010125682434e+00, 3.43550027186533e+00, 3.41346923589720e+00, 3.39390251969698e+00, 3.37667757852078e+00, 3.36165759273883e+00, 3.34869410818808e+00, 3.33762968205122e+00, 3.32830053561002e+00, 3.32053920782401e+00, 3.31417719416736e+00, 3.30904754445178e+00, 3.30498738249657e+00, 3.30184030055202e+00, 3.29945857338971e+00, 3.29770513189040e+00, 3.29645523454172e+00, 3.29559777798277e+00, 3.29503619474021e+00, 3.29468889735952e+00, 3.29448924265466e+00, 3.29438500687881e+00, 3.29433738114242e+00, 3.29431951515737e+00, 3.29431465518313e+00, 3.29431393786136e+00, 3.29431388401299e+00, 3.29431293623633e+00, 3.29430651491403e+00, 3.29428290951266e+00, 3.29421998663034e+00, 3.29408228372982e+00, 3.29381857497052e+00, 3.29336001066389e+00, 3.29261892249401e+00, 3.29148837552664e+00, 3.28984253437267e+00, 3.28753789362121e+00, 3.28441540069986e+00, 3.28030347172134e+00, 3.27502186715905e+00, 3.26838635461710e+00, 3.26021404174443e+00, 3.25032921582282e+00, 3.23856948116944e+00, 3.22479194558642e+00, 3.20887917751958e+00, 3.19074464113633e+00, 3.17033732117137e+00, 3.14764527553655e+00, 3.12269790157495e+00, 3.09556676910411e+00, 3.06636495510624e+00, 3.03524490400075e+00, 3.00239492549682e+00, 2.96803452053839e+00, 2.93240878737620e+00, 2.89578219906706e+00, 2.85843205834795e+00, 2.82064192661096e+00, 2.78269529412437e+00, 2.74486971417874e+00, 2.70743157086000e+00, 2.67063159486546e+00, 2.63470118934083e+00, 2.59984958173419e+00, 2.56626178007396e+00, 2.53409728338282e+00, 2.50348947563053e+00, 2.47454561967141e+00, 2.44734736084760e+00, 2.42195164832647e+00, 2.39839198492911e+00, 2.37667992248698e+00, 2.35680672895319e+00, 2.33874516484189e+00, 2.32245131920220e+00, 2.30786646828961e+00, 2.29491893238886e+00, 2.28352591695952e+00, 2.27359533267886e+00, 2.26502759454455e+00, 2.25771740274354e+00, 2.25155550754316e+00, 2.24643045732242e+00, 2.24223032356262e+00, 2.23884438985780e+00, 2.23616478460320e+00, 2.23408802984573e+00, 2.23251647268846e+00, 2.23135956140146e+00, 2.23053492661591e+00, 2.22996922906241e+00, 2.22959873939849e+00, 2.22936962261305e+00, 2.22923790889846e+00, 2.22916914411876e+00, 2.22913772529962e+00, 2.22912593907679e+00, 2.22912273294416e+00, 2.22912225972591e+00, 2.22912222405076e+00, 2.22912159412284e+00, 2.22911732627170e+00, 2.22910163724730e+00, 2.22905981643611e+00, 2.22896829447934e+00, 2.22879302613239e+00, 2.22848825537994e+00, 2.22799572489646e+00, 2.22724438489593e+00, 2.22615064766219e+00, 2.22461922283575e+00, 2.22254455405658e+00, 2.21981285915810e+00, 2.21630475342257e+00, 2.21189840859236e+00, 2.20647317018875e+00, 2.19991352376067e+00, 2.19211326924450e+00, 2.18297973453007e+00, 2.17243783780518e+00, 2.16043379645839e+00, 2.14693828095024e+00, 2.13194882688335e+00, 2.11549134799920e+00, 2.09762063599795e+00, 2.07841978741413e+00, 2.05799855948136e+00, 2.03649072123477e+00, 2.01405052779134e+00, 1.99084849955207e+00, 1.96706672915874e+00, 1.94289396345837e+00, 1.91852071285162e+00, 1.89413462538045e+00, 1.86991632905605e+00, 1.84603589686571e+00, 1.82265003030490e+00, 1.79989999614288e+00, 1.77791029457740e+00, 1.75678799096764e+00, 1.73662261179030e+00, 1.71748648952805e+00, 1.69943543958312e+00, 1.68250966192868e+00, 1.66673477714501e+00, 1.65212292684950e+00, 1.63867388910995e+00, 1.62637617799034e+00, 1.61520811166026e+00, 1.60513884505416e+00, 1.59612937100309e+00, 1.58813349849674e+00, 1.58109881880971e+00, 1.57496767017940e+00, 1.56967811002914e+00, 1.56516490078344e+00, 1.56136051145362e+00, 1.55819613267214e+00, 1.55560269800468e+00, 1.55351189946408e+00, 1.55185718050244e+00, 1.55057468570490e+00, 1.54960414328491e+00, 1.54888965460371e+00, 1.54838036456201e+00, 1.54803098800743e+00, 1.54780217031122e+00, 1.54766066490080e+00, 1.54757931655492e+00, 1.54753684631915e+00, 1.54751744153043e+00, 1.54751016215867e+00, 1.54750818199631e+00, 1.54750788972861e+00, 1.54750786757512e+00, 1.54750747481317e+00, 1.54750481379502e+00, 1.54749503163944e+00, 1.54746895622403e+00, 1.54741189193066e+00, 1.54730261125592e+00, 1.54711258461560e+00, 1.54680548692473e+00, 1.54633701506355e+00, 1.54565504476825e+00, 1.54470014831329e+00, 1.54340648505257e+00, 1.54170306500457e+00, 1.53951537096551e+00, 1.53676730722659e+00, 1.53338342344485e+00, 1.52929134171229e+00, 1.52442429507390e+00, 1.51872366880134e+00, 1.51214142399497e+00, 1.50464227881967e+00, 1.49620552762653e+00, 1.48682639320976e+00, 1.47651683210666e+00, 1.46530574541384e+00, 1.45323858505453e+00, 1.44037638390377e+00, 1.42679427348599e+00, 1.41257958134353e+00, 1.39782961893043e+00, 1.38264927879428e+00, 1.36714855725502e+00, 1.35144010753528e+00, 1.33563691100213e+00, 1.31985013380735e+00, 1.30418721546404e+00, 1.28875021683869e+00, 1.27363443892575e+00, 1.25892731110498e+00, 1.24470753827634e+00, 1.23104448989502e+00, 1.21799780993479e+00, 1.20561722465983e+00, 1.19394252435039e+00, 1.18300369548588e+00, 1.17282118110120e+00, 1.16340624890618e+00, 1.15476144910950e+00, 1.14688114652878e+00, 1.13975211429246e+00, 1.13335417903925e+00, 1.12766090979520e+00, 1.12264034448071e+00, 1.11825574913219e+00, 1.11446640533044e+00, 1.11122842098592e+00, 1.10849555857854e+00, 1.10622007329201e+00, 1.10435355138204e+00, 1.10284773678882e+00, 1.10165533169596e+00, 1.10073075471917e+00, 1.10003083894062e+00, 1.09951545132502e+00, 1.09914801534172e+00, 1.09889591998339e+00, 1.09873080083629e+00, 1.09862868234441e+00, 1.09856997475839e+00, 1.09853932422556e+00, 1.09852531977086e+00, 1.09852006622076e+00, 1.09851863712816e+00, 1.09851842619711e+00, 1.09851841010737e+00, 1.09851812351200e+00, 1.09851618178676e+00, 1.09850904380777e+00, 1.09849001664122e+00, 1.09844837648092e+00, 1.09836863185042e+00, 1.09822995920489e+00, 1.09800583790807e+00, 1.09766390785436e+00, 1.09716606857767e+00, 1.09646883330057e+00, 1.09552394477940e+00, 1.09427925179052e+00, 1.09267983559441e+00, 1.09066936480360e+00, 1.08819164510913e+00, 1.08519231790001e+00, 1.08162064982305e+00, 1.07743134487439e+00, 1.07258630289257e+00, 1.06705624450024e+00, 1.06082212355733e+00, 1.05387625458726e+00, 1.04622309442826e+00, 1.03787963395482e+00, 1.02887537594721e+00, 1.01925189744889e+00, 1.00906201737902e+00, 9.98368610881553e-01, 9.87243129230002e-01, 9.75763896809397e-01, 9.64014264021952e-01, 9.52080696727627e-01, 9.40050879370002e-01, 9.28011901018611e-01, 9.16048582253048e-01, 9.04241987361058e-01, 8.92668151990506e-01, 8.81397042358159e-01, 8.70491749355605e-01, 8.60007910126129e-01, 8.49993341348704e-01, 8.40487862710223e-01, 8.31523285785636e-01, 8.23123542501193e-01, 8.15304928127514e-01, 8.08076435874781e-01, 8.01440163171041e-01, 7.95391773158407e-01, 7.89920998461323e-01, 7.85012177558805e-01, 7.80644816894632e-01, 7.76794174026307e-01, 7.73431858549632e-01, 7.70526448207455e-01, 7.68044117519011e-01, 7.65949275521141e-01, 7.64205207909740e-01, 7.62774717164994e-01, 7.61620752324550e-01, 7.60707018145563e-01, 7.59998551688906e-01, 7.59462253079842e-01, 7.59067356538115e-01, 7.58785827871613e-01, 7.58592675579564e-01, 7.58466164530351e-01, 7.58387923808197e-01, 7.58342943633551e-01, 7.58319460065415e-01, 7.58308730260389e-01, 7.58304705145136e-01, 7.58303610216668e-01, 7.58303448607538e-01, 7.58303436193067e-01, 7.58303213922236e-01, 7.58301708005192e-01, 7.58296172102941e-01, 7.58281415482355e-01, 7.58249121295233e-01, 7.58187275274789e-01, 7.58079728316666e-01, 7.57905914016320e-01, 7.57640739420580e-01, 7.57254663871315e-01, 7.56713976677723e-01, 7.55981279244313e-01, 7.55016171035329e-01, 7.53776131296689e-01, 7.52217579852474e-01, 7.50297090804793e-01, 7.47972723070596e-01, 7.45205422078747e-01, 7.41960438493052e-01, 7.38208703488336e-01, 7.33928096836042e-01, 7.29104544649893e-01, 7.23732888601226e-01, 7.17817477822386e-01, 7.11372448172025e-01, 7.04421670142683e-01, 6.96998365132853e-01, 6.89144408495647e-01, 6.80909355051508e-01, 6.72349237093645e-01, 6.63525195144693e-01, 6.54502007149095e-01, 6.45346582266718e-01, 6.36126481358637e-01, 6.26908518446047e-01, 6.17757486984907e-01, 6.08735042949288e-01, 5.99898764627881e-01, 5.91301397706054e-01, 5.82990284371192e-01, 5.75006967298734e-01, 5.67386953637861e-01, 5.60159620482153e-01, 5.53348241580817e-01, 5.46970114912973e-01, 5.41036771854305e-01, 5.35554250646157e-01, 5.30523419388146e-01, 5.25940336514580e-01, 5.21796639428988e-01, 5.18079954459111e-01, 5.14774323403000e-01, 5.11860643559498e-01, 5.09317119208063e-01, 5.07119722998397e-01, 5.05242665644275e-01, 5.03658871736437e-01, 5.02340458481363e-01, 5.01259212850357e-01, 5.00387061128457e-01, 4.99696523344817e-01, 4.99161143714556e-01, 4.98755887191479e-01, 4.98457491668653e-01, 4.98244765384229e-01, 4.98098819762902e-01, 4.98003229264441e-01, 4.97944111778345e-01, 4.97910125599173e-01, 4.97892381897255e-01, 4.97884274686740e-01, 4.97881233397592e-01, 4.97880406093714e-01, 4.97880283985413e-01, 4.97880274533728e-01, 4.97880104377209e-01, 4.97878951542258e-01, 4.97874713606702e-01, 4.97863416885398e-01, 4.97838694586154e-01, 4.97791349571305e-01, 4.97709019778522e-01, 4.97575962570835e-01, 4.97372972185005e-01, 4.97077441905947e-01, 4.96663579448509e-01, 4.96102780106597e-01, 4.95364157388881e-01, 4.94415225035052e-01, 4.93222717546960e-01, 4.91753528882885e-01, 4.89975741136256e-01, 4.87859707412223e-01, 4.85379146431985e-01, 4.82512201416338e-01, 4.79242413293154e-01, 4.75559558871339e-01, 4.71460308712562e-01, 4.66948667053217e-01, 4.62036166925749e-01, 4.56741806847120e-01, 4.51091730005274e-01, 4.45118661504740e-01, 4.38861132621046e-01, 4.32362531991256e-01, 4.25670031357949e-01, 4.18833437401155e-01, 4.11904021280581e-01, 4.04933374111067e-01, 3.97972330366112e-01, 3.91069993000645e-01, 3.84272884828956e-01, 3.77624241262385e-01, 3.71163450648037e-01, 3.64925640711718e-01, 3.58941403352798e-01, 3.53236645426477e-01, 3.47832550176509e-01, 3.42745632516891e-01, 3.37987871186375e-01, 3.33566901653061e-01, 3.29486255249122e-01, 3.25745632100901e-01, 3.22341197740851e-01, 3.19265895633130e-01, 3.16509770037175e-01, 3.14060295537089e-01, 3.11902711077799e-01, 3.10020357410242e-01, 3.08395017426314e-01, 3.07007258964866e-01, 3.05836779325617e-01, 3.04862750004304e-01, 3.04064159149181e-01, 3.03420148050812e-01, 3.02910336743697e-01, 3.02515132657337e-01, 3.02216015341848e-01, 3.01995789729684e-01, 3.01838800277768e-01, 3.01731098721561e-01, 3.01660559083405e-01, 3.01616934980633e-01, 3.01591856102466e-01, 3.01578762856213e-01, 3.01572780486847e-01, 3.01570536300520e-01, 3.01569925828080e-01, 3.01569835723651e-01, 3.01569828692656e-01, 3.01569701384761e-01, 3.01568838855562e-01, 3.01565668115128e-01, 3.01557216147586e-01, 3.01538719539618e-01, 3.01503297534509e-01, 3.01441702162851e-01, 3.01342157848850e-01, 3.01190301407445e-01, 3.00969231523048e-01, 3.00659674494042e-01, 3.00240270087567e-01, 2.99687977652212e-01, 2.98978598117582e-01, 2.98087402204224e-01, 2.96989849238265e-01, 2.95662374722831e-01, 2.94083218718804e-01, 2.92233261705302e-01, 2.90096830566522e-01, 2.87662435313194e-01, 2.84923397616094e-01, 2.81878335524631e-01, 2.78531474904508e-01, 2.74892766866927e-01, 2.70977801167557e-01, 2.66807517348489e-01, 2.62407727237587e-01, 2.57808473235063e-01, 2.53043255644499e-01, 2.48148168420105e-01, 2.43160985698086e-01, 2.38120241303452e-01, 2.33064340349653e-01, 2.28030736614510e-01, 2.23055202289910e-01, 2.18171208740603e-01, 2.13409428816664e-01, 2.08797363690632e-01, 2.04359090624330e-01, 2.00115122828784e-01, 1.96082368809023e-01, 1.92274176282975e-01, 1.88700444812220e-01, 1.85367791484794e-01, 1.82279755100975e-01, 1.79437026070207e-01, 1.76837691377699e-01, 1.74477486295359e-01, 1.72350046801579e-01, 1.70447158786381e-01, 1.68759001937147e-01, 1.67274387647389e-01, 1.65980991320158e-01, 1.64865580032778e-01, 1.63914236699675e-01, 1.63112581650601e-01, 1.62445991987337e-01, 1.61899818268353e-01, 1.61459597088989e-01, 1.61111257077739e-01, 1.60841314826431e-01, 1.60637056421717e-01, 1.60486699645155e-01, 1.60379531638632e-01, 1.60306016942179e-01, 1.60257871321612e-01, 1.60228097695832e-01, 1.60210981695680e-01, 1.60202045855840e-01, 1.60197963054980e-01, 1.60196431464836e-01, 1.60196014836064e-01, 1.60195953342566e-01, 1.60195948503297e-01, 1.60195860357163e-01, 1.60195263154741e-01, 1.60193067784661e-01, 1.60187215804722e-01, 1.60174409248359e-01, 1.60149884534952e-01, 1.60107240152488e-01, 1.60038327037689e-01, 1.59933209683196e-01, 1.59780204851849e-01, 1.59566003245637e-01, 1.59275877423418e-01, 1.58893976569757e-01, 1.58403705322037e-01, 1.57788179781615e-01, 1.57030749195371e-01, 1.56115566850776e-01, 1.55028188860644e-01, 1.53756175206365e-01, 1.52289664202197e-01, 1.50621889969827e-01, 1.48749613013326e-01, 1.46673436826752e-01, 1.44397988675369e-01, 1.41931950008175e-01, 1.39287930849147e-01, 1.36482192208994e-01, 1.33534230154318e-01, 1.30466243742813e-01, 1.27302515763496e-01, 1.24068739513472e-01, 1.20791326390254e-01, 1.17496727899692e-01, 1.14210802084916e-01, 1.10958248919452e-01, 1.07762132560218e-01, 1.04643501247576e-01, 1.01621108738836e-01, 9.87112350148038e-02, 9.59275989884283e-02, 9.32813522745555e-02, 9.07811407926964e-02, 8.84332199769240e-02, 8.62416094747047e-02, 8.42082741920329e-02, 8.23333201330918e-02, 8.06151954477912e-02, 7.90508892266288e-02, 7.76361226961768e-02, 7.63655294347529e-02, 7.52328229534811e-02, 7.42309514117565e-02, 7.33522403261411e-02, 7.25885248783023e-02, 7.19312738361637e-02, 7.13717071944906e-02, 7.09009094476372e-02, 7.05099399716005e-02, 7.01899413678362e-02, 6.99322458697047e-02, 6.97284791043202e-02, 6.95706597112355e-02, 6.94512926190960e-02, 6.93634532407543e-02, 6.93008595222508e-02, 6.92579287110711e-02, 6.92298159112354e-02, 6.92124319586078e-02, 6.92024388473378e-02, 6.91972218139725e-02, 6.91948381702900e-02, 6.91939439932732e-02, 6.91937007563901e-02, 6.91936648551703e-02, 6.91936620048921e-02, 6.91936097699148e-02, 6.91932558708572e-02, 6.91919549094493e-02, 6.91884870928625e-02, 6.91808982106075e-02, 6.91663659325030e-02, 6.91410983697303e-02, 6.91002705214530e-02, 6.90380038266092e-02, 6.89473934683917e-02, 6.88205872284620e-02, 6.86489184427366e-02, 6.84230938703863e-02, 6.81334349903437e-02, 6.77701683922187e-02, 6.73237576237966e-02, 6.67852652929035e-02, 6.61467306956922e-02, 6.54015451404533e-02, 6.45448048844450e-02, 6.35736206203461e-02, 6.24873630768867e-02, 6.12878267260335e-02, 5.99792978008189e-02, 5.85685185741944e-02, 5.70645466547788e-02, 5.54785152689661e-02, 5.38233073819683e-02, 5.21131623399604e-02, 5.03632378952297e-02, 4.85891526170512e-02, 4.68065336623482e-02, 4.50305928151167e-02, 4.32757499629020e-02, 4.15553182836561e-02, 3.98812599619163e-02, 3.82640158163632e-02, 3.67124072866218e-02, 3.52336051456525e-02, 3.38331562777121e-02, 3.25150579584736e-02, 3.12818682457490e-02, 3.01348411990787e-02, 2.90740764909003e-02, 2.80986743249443e-02, 2.72068882201530e-02, 2.63962699728175e-02, 2.56638028372080e-02, 2.50060205652716e-02, 2.44191113436614e-02, 2.38990068132452e-02, 2.34414572308133e-02, 2.30420944433014e-02, 2.26964847134496e-02, 2.24001735927040e-02, 2.21487250084920e-02, 2.19377565452312e-02, 2.17629725775588e-02, 2.16201964905274e-02, 2.15054027291650e-02, 2.14147488966923e-02, 2.13446076063875e-02, 2.12915973252343e-02, 2.12526110648668e-02, 2.12248415072374e-02, 2.12058010228766e-02, 2.11933350589533e-02, 2.11856275424367e-02, 2.11811972433154e-02, 2.11788844476430e-02, 2.11778277623722e-02, 2.11774313713633e-02, 2.11773235441690e-02, 2.11773076291237e-02, 2.11773063540264e-02, 2.11772828405969e-02, 2.11771235341018e-02, 2.11765379134938e-02, 2.11749769172194e-02, 2.11715609852057e-02, 2.11650201019289e-02, 2.11536486899817e-02, 2.11352781594201e-02, 2.11072698410355e-02, 2.10665308615816e-02, 2.10095551795639e-02, 2.09324914081748e-02, 2.08312381362070e-02, 2.07015661705050e-02, 2.05392654653380e-02, 2.03403125333104e-02, 2.01010519816016e-02, 1.98183836888113e-02, 1.94899452926016e-02, 1.91142783886882e-02, 1.86909664273855e-02, 1.82207329556370e-02, 1.77054906986953e-02, 1.71483349657926e-02, 1.65534787801427e-02, 1.59261315994677e-02, 1.52723280140421e-02, 1.45987168559110e-02, 1.39123242431261e-02, 1.32203058643832e-02, 1.25297041214303e-02, 1.18472246105696e-02, 1.11790440125112e-02, 1.05306580900049e-02, 9.90677462192124e-03, 9.31125221758577e-03, 8.74708251209810e-03, 8.21641060314704e-03, 7.72058684028210e-03, 7.26024210446496e-03, 6.83537850959937e-03, 6.44546790504563e-03, 6.08955146989842e-03, 5.76633514638303e-03, 5.47427731004646e-03, 5.21166624839964e-03, 4.97668613396129e-03, 4.76747137898670e-03, 4.58214959257219e-03, 4.41887354974213e-03, 4.27584380807033e-03, 4.15132423599829e-03, 4.04365173395149e-03, 3.95124090836774e-03, 3.87258522960465e-03, 3.80625650274524e-03, 3.75090369342601e-03, 3.70525149539702e-03, 3.66809905010612e-03, 3.63831936036974e-03, 3.61485976604570e-03, 3.59674352295869e-03, 3.58307226562360e-03, 3.57302898163780e-03, 3.56588104120448e-03, 3.56098277308092e-03, 3.55777706699811e-03, 3.55579551505159e-03, 3.55465668287636e-03, 3.55406221525101e-03, 3.55379062216699e-03, 3.55368874209461e-03, 3.55366102862575e-03, 3.55365693818044e-03, 3.55365660738016e-03, 3.55365046872791e-03, 3.55360887869338e-03, 3.55345599262336e-03, 3.55304847944637e-03, 3.55215677348042e-03, 3.55044953546013e-03, 3.54748215298883e-03, 3.54269015051401e-03, 3.53538840643708e-03, 3.52477707680166e-03, 3.50995505052865e-03, 3.48994158613181e-03, 3.46370645693028e-03, 3.43020843678232e-03, 3.38844127879057e-03, 3.33748550649204e-03, 3.27656347042904e-03, 3.20509439708516e-03, 3.12274568332072e-03, 3.02947629853989e-03, 2.92556761121077e-03, 2.81163685151142e-03, 2.68863042652256e-03, 2.55779840648174e-03, 2.42065303436689e-03, 2.27891008163184e-03, 2.13441152655847e-03, 1.98904014587953e-03, 1.84464201304735e-03, 1.70295413792587e-03, 1.56553250424807e-03, 1.43370401772458e-03, 1.30854498596733e-03, 1.19085501330574e-03, 1.08115129396325e-03, 9.79700617431689e-04, 8.86529339652498e-04, 8.01454529135270e-04, 7.24160139622398e-04, 6.54188001437461e-04, 5.90966002057769e-04, 5.33949861875271e-04, 4.82508900091716e-04, 4.35885604723296e-04, 3.93752202278108e-04, 3.55799308287885e-04, 3.21060429900655e-04, 2.89035565295137e-04, 2.59372892351450e-04, 2.31826445517669e-04, 2.06228002703313e-04, 1.82444316642716e-04, 1.60756255620953e-04, 1.40723267382015e-04, 1.22165590632051e-04, 1.05097659623965e-04, 8.95199117962875e-05, 7.54182368112311e-05, 6.27619494544263e-05, 5.15082647538634e-05, 4.16078010694321e-05, 3.30054756325637e-05, 2.56386249084528e-05, 1.94352678732912e-05, 1.43134896610775e-05, 1.01819027484851e-05, 6.94089076347382e-06, 4.48434885669258e-06, 2.70171395655439e-06, 1.48016799883521e-06, 7.06978369180433e-07, 2.72016926317065e-07, 3.55271367880050e-15, 7.10542735760100e-15],
[ 2.70000000000000e+01, 2.69999999811787e+01, 2.69999993991861e+01, 2.69999954560951e+01, 2.69999809607841e+01, 2.69999423209078e+01, 2.69998577549362e+01, 2.69996957873885e+01, 2.69994140819171e+01, 2.69989586567301e+01, 2.69982635153271e+01, 2.69972507130272e+01, 2.69958308667538e+01, 2.69939041024667e+01, 2.69913614219696e+01, 2.69880864590131e+01, 2.69839575840330e+01, 2.69788503077895e+01, 2.69726399267775e+01, 2.69652043475868e+01, 2.69564270233188e+01, 2.69461999325233e+01, 2.69344265296706e+01, 2.69210245956659e+01, 2.69059289171975e+01, 2.68890937247203e+01, 2.68704948207504e+01, 2.68501313331402e+01, 2.68280270325591e+01, 2.68042311599601e+01, 2.67788187188286e+01, 2.67518901987347e+01, 2.67235707111998e+01, 2.66940085358293e+01, 2.66633730934446e+01, 2.66318523826370e+01, 2.65996499356186e+01, 2.65669813672484e+01, 2.65340706064771e+01, 2.65011459112115e+01, 2.64684357750556e+01, 2.64361648372031e+01, 2.64045499049781e+01, 2.63737961924988e+01, 2.63440938693088e+01, 2.63156150003588e+01, 2.62885109442933e+01, 2.62629102614322e+01, 2.62389171668828e+01, 2.62166105484801e+01, 2.61960435541933e+01, 2.61772437395622e+01, 2.61602137528140e+01, 2.61449325236607e+01, 2.61313569114034e+01, 2.61194237588895e+01, 2.61090522910848e+01, 2.61001467905649e+01, 2.60925994771649e+01, 2.60862935154389e+01, 2.60811060715891e+01, 2.60769113412438e+01, 2.60735834709908e+01, 2.60709992999917e+01, 2.60690408533113e+01, 2.60675975257610e+01, 2.60665679039455e+01, 2.60658611846237e+01, 2.60653981591771e+01, 2.60651117466038e+01, 2.60649470706397e+01, 2.60648610899629e+01, 2.60648218035438e+01, 2.60648070656675e+01, 2.60648030565905e+01, 2.60648024648585e+01, 2.60648023872357e+01, 2.60648005787183e+01, 2.60647883257329e+01, 2.60647432822540e+01, 2.60646232112773e+01, 2.60643604307638e+01, 2.60638571435389e+01, 2.60629818255099e+01, 2.60615668176296e+01, 2.60594072341520e+01, 2.60562612636986e+01, 2.60518519014106e+01, 2.60458701109876e+01, 2.60379793758017e+01, 2.60278215595544e+01, 2.60150239600699e+01, 2.59992074054281e+01, 2.59799952100401e+01, 2.59570227793045e+01, 2.59299476247296e+01, 2.58984595263630e+01, 2.58622905559427e+01, 2.58212246531282e+01, 2.57751064305329e+01, 2.57238488745088e+01, 2.56674396123402e+01, 2.56059454376607e+01, 2.55395148287371e+01, 2.54683782608662e+01, 2.53928462033848e+01, 2.53133047987464e+01, 2.52302093372059e+01, 2.51440757548416e+01, 2.50554704831283e+01, 2.49649990546199e+01, 2.48732939142045e+01, 2.47810018959578e+01, 2.46887718035525e+01, 2.45972424832481e+01, 2.45070317110986e+01, 2.44187261396372e+01, 2.43328724727676e+01, 2.42499699679634e+01, 2.41704643067203e+01, 2.40947428295198e+01, 2.40231311001329e+01, 2.39558907441436e+01, 2.38932184954221e+01, 2.38352463790480e+01, 2.37820429571885e+01, 2.37336155635056e+01, 2.36899134502437e+01, 2.36508317692494e+01, 2.36162163033249e+01, 2.35858688574497e+01, 2.35595532107339e+01, 2.35370015199555e+01, 2.35179210548028e+01, 2.35020011342659e+01, 2.34889201238675e+01, 2.34783523455359e+01, 2.34699747468610e+01, 2.34634731751500e+01, 2.34585481049026e+01, 2.34549196756599e+01, 2.34523319109614e+01, 2.34505560083806e+01, 2.34493926149341e+01, 2.34486730308655e+01, 2.34482593168431e+01, 2.34480433136913e+01, 2.34479446184350e+01, 2.34479075941729e+01, 2.34478975226510e+01, 2.34478960361146e+01, 2.34478959048817e+01, 2.34478933334380e+01, 2.34478759115212e+01, 2.34478118665979e+01, 2.34476411454713e+01, 2.34472675220195e+01, 2.34465519713897e+01, 2.34453075695670e+01, 2.34432961423315e+01, 2.34402268473973e+01, 2.34357568277283e+01, 2.34294940221834e+01, 2.34210021612513e+01, 2.34098079104479e+01, 2.33954100522594e+01, 2.33772905204351e+01, 2.33549270201287e+01, 2.33278068872220e+01, 2.32954417646662e+01, 2.32573826083916e+01, 2.32132344864120e+01, 2.31626706083754e+01, 2.31054450243976e+01, 2.30414034653383e+01, 2.29704918628884e+01, 2.28927621847701e+01, 2.28083753420960e+01, 2.27176010631240e+01, 2.26208147684229e+01, 2.25184916142099e+01, 2.24111979820681e+01, 2.22995807765751e+01, 2.21843549446226e+01, 2.20662896535259e+01, 2.19461935656654e+01, 2.18248996335022e+01, 2.17032498178141e+01, 2.15820801084923e+01, 2.14622062019186e+01, 2.13444101591805e+01, 2.12294283310795e+01, 2.11179407859296e+01, 2.10105624143560e+01, 2.09078358152830e+01, 2.08102259956181e+01, 2.07181168504974e+01, 2.06318093379804e+01, 2.05515212256288e+01, 2.04773882669537e+01, 2.04094666607601e+01, 2.03477366516018e+01, 2.02921071399694e+01, 2.02424211820531e+01, 2.01984622676776e+01, 2.01599612693405e+01, 2.01266039544623e+01, 2.00980389472078e+01, 2.00738860164547e+01, 2.00537445540149e+01, 2.00372020936566e+01, 2.00238427085451e+01, 2.00132551141649e+01, 2.00050402972692e+01, 1.99988184903784e+01, 1.99942353170017e+01, 1.99909669458366e+01, 1.99887241129641e+01, 1.99872548992289e+01, 1.99863461846973e+01, 1.99858237419660e+01, 1.99855509733624e+01, 1.99854263416697e+01, 1.99853795877820e+01, 1.99853668695686e+01, 1.99853649923861e+01, 1.99853648418129e+01, 1.99853620629724e+01, 1.99853432359138e+01, 1.99852740255681e+01, 1.99850895355307e+01, 1.99846857802157e+01, 1.99839125275309e+01, 1.99825677940815e+01, 1.99803942384570e+01, 1.99770776560482e+01, 1.99722477305620e+01, 1.99654811421421e+01, 1.99563070689067e+01, 1.99442150472404e+01, 1.99286650764332e+01, 1.99090997663647e+01, 1.98849582353016e+01, 1.98556913723776e+01, 1.98207779912527e+01, 1.97797413243376e+01, 1.97321652481340e+01, 1.96777095971245e+01, 1.96161239230129e+01, 1.95472590930233e+01, 1.94710761977790e+01, 1.93876523545800e+01, 1.92971831397924e+01, 1.91999815538254e+01, 1.90964735989314e+01, 1.89871907164750e+01, 1.88727594694782e+01, 1.87538889551694e+01, 1.86313564849034e+01, 1.85059920775029e+01, 1.83786622864227e+01, 1.82502538347787e+01, 1.81216574784718e+01, 1.79937524655639e+01, 1.78673919132086e+01, 1.77433893803033e+01, 1.76225068705215e+01, 1.75054444526397e+01, 1.73928316314753e+01, 1.72852205446243e+01, 1.71830810012523e+01, 1.70867973242537e+01, 1.69966669106569e+01, 1.69129003902892e+01, 1.68356232406268e+01, 1.67648787056975e+01, 1.67006318667103e+01, 1.66427747186382e+01, 1.65911321169677e+01, 1.65454684691004e+01, 1.65054950529754e+01, 1.64708778496401e+01, 1.64412457758969e+01, 1.64161991977852e+01, 1.63953185962368e+01, 1.63781732441049e+01, 1.63643297406532e+01, 1.63533602375356e+01, 1.63448501813511e+01, 1.63384053939912e+01, 1.63336583148200e+01, 1.63302732394089e+01, 1.63279504086159e+01, 1.63264288291420e+01, 1.63254877414735e+01, 1.63249466919050e+01, 1.63246642101620e+01, 1.63245351407540e+01, 1.63244867221744e+01, 1.63244735511289e+01, 1.63244716071096e+01, 1.63244714571681e+01, 1.63244687647003e+01, 1.63244505228302e+01, 1.63243834637218e+01, 1.63242047082637e+01, 1.63238135038004e+01, 1.63230642908151e+01, 1.63217613729910e+01, 1.63196554304518e+01, 1.63164420759406e+01, 1.63117626076571e+01, 1.63052070595086e+01, 1.62963195887232e+01, 1.62846061715040e+01, 1.62695444997349e+01, 1.62505958867690e+01, 1.62272189004116e+01, 1.61988843501310e+01, 1.61650911684413e+01, 1.61253826495364e+01, 1.60793624484718e+01, 1.60267097082934e+01, 1.59671926765167e+01, 1.59006802006287e+01, 1.58271505569481e+01, 1.57466971674798e+01, 1.56595308915684e+01, 1.55659787363084e+01, 1.54664790022057e+01, 1.53615730567134e+01, 1.52518940949420e+01, 1.51381533909038e+01, 1.50211246522325e+01, 1.49016271573822e+01, 1.47805083719626e+01, 1.46586267103369e+01, 1.45368350354092e+01, 1.44159653837790e+01, 1.42968152783755e+01, 1.41801358605710e+01, 1.40666219517747e+01, 1.39569040508684e+01, 1.38515421947762e+01, 1.37510215569252e+01, 1.36557496305894e+01, 1.35660548367100e+01, 1.34821864029977e+01, 1.34043153770263e+01, 1.33325366554325e+01, 1.32668719302600e+01, 1.32072734692929e+01, 1.31536286584487e+01, 1.31057652404094e+01, 1.30634571847031e+01, 1.30264311208108e+01, 1.29943732581514e+01, 1.29669367056973e+01, 1.29437490903399e+01, 1.29244203579137e+01, 1.29085506251742e+01, 1.28957379363193e+01, 1.28855857652945e+01, 1.28777100966352e+01, 1.28717459143863e+01, 1.28673529319075e+01, 1.28642204059835e+01, 1.28620708969816e+01, 1.28606628626701e+01, 1.28597920059533e+01, 1.28592913348873e+01, 1.28590299351381e+01, 1.28589104984444e+01, 1.28588656934572e+01, 1.28588535053992e+01, 1.28588517064668e+01, 1.28588515702571e+01, 1.28588491573151e+01, 1.28588328092746e+01, 1.28587727120911e+01, 1.28586125146243e+01, 1.28582619241528e+01, 1.28575904928980e+01, 1.28564228415168e+01, 1.28545355344832e+01, 1.28516557863424e+01, 1.28474621364082e+01, 1.28415871816664e+01, 1.28336224025284e+01, 1.28231250529731e+01, 1.28096270155304e+01, 1.27926454435185e+01, 1.27716949302874e+01, 1.27463008618024e+01, 1.27160135300372e+01, 1.26804225168350e+01, 1.26391708082526e+01, 1.25919680747829e+01, 1.25386025588854e+01, 1.24789510512237e+01, 1.24129865109975e+01, 1.23407829900884e+01, 1.22625176480164e+01, 1.21784697845127e+01, 1.20890169567507e+01, 1.19946283769428e+01, 1.18958558931324e+01, 1.17933229352223e+01, 1.16877118576862e+01, 1.15797501324426e+01, 1.14701958455927e+01, 1.13598229369280e+01, 1.12494065972041e+01, 1.11397092085891e+01, 1.10314671787701e+01, 1.09253789769491e+01, 1.08220946276565e+01, 1.07222068546372e+01, 1.06262439935791e+01, 1.05346647140996e+01, 1.04478545155516e+01, 1.03661238958281e+01, 1.02897080438732e+01, 1.02187678783777e+01, 1.01533922468556e+01, 1.00936011074702e+01, 1.00393495352154e+01, 9.99053241839305e+00, 9.94698973547808e+00, 9.90851232249645e+00, 9.87484805460942e+00, 9.84570837183296e+00, 9.82077507798110e+00, 9.79970733507035e+00, 9.78214876410560e+00, 9.76773454920265e+00, 9.75609842731798e+00, 9.74687943238748e+00, 9.73972825226265e+00, 9.73431305112876e+00, 9.73032461044671e+00, 9.72748064882845e+00, 9.72552919607977e+00, 9.72425091880588e+00, 9.72346032377084e+00, 9.72300579938520e+00, 9.72276849357591e+00, 9.72266006587972e+00, 9.72261939079236e+00, 9.72260832617149e+00, 9.72260669305624e+00, 9.72260657045776e+00, 9.72260441258544e+00, 9.72258979268054e+00, 9.72253604832697e+00, 9.72239278533130e+00, 9.72207925634180e+00, 9.72147880534858e+00, 9.72043459891434e+00, 9.71874683774447e+00, 9.71617161065016e+00, 9.71242151647931e+00, 9.70716813730008e+00, 9.70004639709552e+00, 9.69066078379279e+00, 9.67859334852020e+00, 9.66341332536498e+00, 9.64468813954517e+00, 9.62199549505471e+00, 9.59493615901754e+00, 9.56314699481709e+00, 9.52631374589242e+00, 9.48418304339718e+00, 9.43657310962166e+00, 9.38338265969177e+00, 9.32459756881553e+00, 9.26029497036188e+00, 9.19064457667094e+00, 9.11590716098312e+00, 9.03643029276980e+00, 8.95264156501999e+00, 8.86503967520731e+00, 8.77418380865026e+00, 8.68068181601019e+00, 8.58517767532375e+00, 8.48833869041816e+00, 8.39084281469388e+00, 8.29336641707648e+00, 8.19657273850870e+00, 8.10110123078192e+00, 8.00755792689807e+00, 7.91650696065388e+00, 7.82846332734680e+00, 7.74388695208125e+00, 7.66317810376366e+00, 7.58667416089181e+00, 7.51464770143621e+00, 7.44730585661643e+00, 7.38479084054048e+00, 7.32718154708820e+00, 7.27449609336592e+00, 7.22669518539664e+00, 7.18368618507329e+00, 7.14532776561141e+00, 7.11143505325259e+00, 7.08178516332918e+00, 7.05612304694872e+00, 7.03416756907767e+00, 7.01561773902035e+00, 7.00015901027985e+00, 6.98746955931240e+00, 6.97722644305188e+00, 6.96911152496805e+00, 6.96281705068300e+00, 6.95805074862258e+00, 6.95454033040846e+00, 6.95203727088691e+00, 6.95031975949869e+00, 6.94919473318225e+00, 6.94849892562429e+00, 6.94809889733376e+00, 6.94789004418365e+00, 6.94779461692205e+00, 6.94775881877560e+00, 6.94774908080365e+00, 6.94774764349910e+00, 6.94774753592830e+00, 6.94774564693102e+00, 6.94773284869490e+00, 6.94768580100098e+00, 6.94756038893731e+00, 6.94728592676352e+00, 6.94676029526612e+00, 6.94584620642356e+00, 6.94436876564842e+00, 6.94211447460091e+00, 6.93883178602388e+00, 6.93423328545772e+00, 6.92799953225751e+00, 6.91978454350889e+00, 6.90922284919628e+00, 6.89593798607037e+00, 6.87955223280606e+00, 6.85969732300218e+00, 6.83602580910461e+00, 6.80822269396595e+00, 6.77601690241535e+00, 6.73919213773345e+00, 6.69759666147493e+00, 6.65115155261349e+00, 6.59985704481766e+00, 6.54379660821637e+00, 6.48313853175894e+00, 6.41813486990228e+00, 6.34911773706506e+00, 6.27649305814018e+00, 6.20073200569178e+00, 6.12236046631466e+00, 6.04194697223077e+00, 5.96008960257130e+00, 5.87740239649815e+00, 5.79450182418182e+00, 5.71199383139143e+00, 5.63046191198662e+00, 5.55045657589910e+00, 5.47248647662536e+00, 5.39701135153758e+00, 5.32443682025112e+00, 5.25511098951079e+00, 5.18932273411549e+00, 5.12730146618532e+00, 5.06921817074886e+00, 5.01518747292405e+00, 4.96527050777845e+00, 4.91947838402299e+00, 4.87777606230091e+00, 4.84008650336409e+00, 4.80629497679084e+00, 4.77625345375855e+00, 4.74978503527545e+00, 4.72668838859134e+00, 4.70674217843652e+00, 4.68970948616869e+00, 4.67534220932203e+00, 4.66338542742722e+00, 4.65358170868098e+00, 4.64567531775903e+00, 4.63941626964761e+00, 4.63456415974033e+00, 4.63089168846003e+00, 4.62818779095040e+00, 4.62626028020873e+00, 4.62493791618573e+00, 4.62407182408283e+00, 4.62353620196124e+00, 4.62322827991106e+00, 4.62306751902164e+00, 4.62299406653925e+00, 4.62296651205745e+00, 4.62295901657788e+00, 4.62295791026107e+00, 4.62295782743108e+00, 4.62295637248091e+00, 4.62294651498406e+00, 4.62291027785238e+00, 4.62281368349736e+00, 4.62260229182184e+00, 4.62219746185134e+00, 4.62149349161600e+00, 4.62035577579669e+00, 4.61862010288371e+00, 4.61609319211545e+00, 4.61255454389982e+00, 4.60775964433967e+00, 4.60144452370777e+00, 4.59333161968370e+00, 4.58313683925619e+00, 4.57057765001446e+00, 4.55538196508173e+00, 4.53729752052711e+00, 4.51610138522424e+00, 4.49160919701686e+00, 4.46368369201215e+00, 4.43224209146711e+00, 4.39726193715664e+00, 4.35878502308800e+00, 4.31691915779756e+00, 4.27183760283148e+00, 4.22377616182774e+00, 4.17302803074639e+00, 4.11993665152501e+00, 4.06488692681720e+00, 4.00829524184209e+00, 3.95059879279393e+00, 3.89224473562262e+00, 3.83367964457157e+00, 3.77533971118648e+00, 3.71764202961867e+00, 3.66097721318728e+00, 3.60570348127299e+00, 3.55214225482690e+00, 3.50057521124088e+00, 3.45124268047586e+00, 3.40434321671148e+00, 3.36003415323805e+00, 3.31843294066698e+00, 3.27961907625436e+00, 3.24363645107326e+00, 3.21049596779061e+00, 3.18017831119126e+00, 3.15263678330037e+00, 3.12780014269220e+00, 3.10557541175267e+00, 3.08585063531346e+00, 3.06849758870490e+00, 3.05337444276466e+00, 3.04032839784471e+00, 3.02919829875066e+00, 3.01981723837956e+00, 3.01201515030904e+00, 3.00562138059947e+00, 3.00046721760658e+00, 2.99638834676994e+00, 2.99322718630306e+00, 2.99083505060586e+00, 2.98907408207927e+00, 2.98781888967760e+00, 2.98695783452040e+00, 2.98639390937992e+00, 2.98604516964244e+00, 2.98584468780688e+00, 2.98574002079805e+00, 2.98569219816563e+00, 2.98567425834037e+00, 2.98566937828191e+00, 2.98566865799583e+00, 2.98566860392493e+00, 2.98566765223165e+00, 2.98566120437451e+00, 2.98563750144186e+00, 2.98557431868122e+00, 2.98543604757240e+00, 2.98517125265353e+00, 2.98471080585073e+00, 2.98396669210192e+00, 2.98283156939362e+00, 2.98117915304555e+00, 2.97886547701032e+00, 2.97573106286302e+00, 2.97160399899243e+00, 2.96630389771465e+00, 2.95964665684581e+00, 2.95144990597004e+00, 2.94153896869523e+00, 2.92975312426818e+00, 2.91595190967181e+00, 2.90002117195478e+00, 2.88187856516978e+00, 2.86147819118385e+00, 2.83881411137556e+00, 2.81392250710876e+00, 2.78688233838657e+00, 2.75781443700782e+00, 2.72687906535625e+00, 2.69427206575191e+00, 2.66021980900354e+00, 2.62497321645352e+00, 2.58880117168581e+00, 2.55198365346147e+00, 2.51480491088145e+00, 2.47754696869887e+00, 2.44048370070644e+00, 2.40387564892043e+00, 2.36796570264007e+00, 2.33297569031551e+00, 2.29910388303562e+00, 2.26652336423226e+00, 2.23538118720398e+00, 2.20579822030871e+00, 2.17786956825431e+00, 2.15166545535451e+00, 2.12723246117511e+00, 2.10459500887568e+00, 2.08375702003792e+00, 2.06470366529975e+00, 2.04740315631387e+00, 2.03180854026190e+00, 2.01785947244340e+00, 2.00548395460788e+00, 1.99460003620926e+00, 1.98511748234765e+00, 1.97693941572722e+00, 1.96996394058640e+00, 1.96408575449830e+00, 1.95919774959499e+00, 1.95519259867423e+00, 1.95196431444695e+00, 1.94940976260825e+00, 1.94743010223301e+00, 1.94593212097094e+00, 1.94482942833309e+00, 1.94404346857946e+00, 1.94350431569823e+00, 1.94315121683843e+00, 1.94293285719039e+00, 1.94280732831858e+00, 1.94274179274563e+00, 1.94271184941455e+00, 1.94270061670420e+00, 1.94269756114095e+00, 1.94269711014644e+00, 1.94269707614670e+00, 1.94269647580210e+00, 1.94269240838246e+00, 1.94267745616865e+00, 1.94263759944839e+00, 1.94255037598025e+00, 1.94238334022689e+00, 1.94209288750366e+00, 1.94162350272525e+00, 1.94090748690251e+00, 1.93986520634008e+00, 1.93840589882623e+00, 1.93642905716148e+00, 1.93382639243659e+00, 1.93048435716559e+00, 1.92628718184536e+00, 1.92112034857580e+00, 1.91487439364767e+00, 1.90744889987965e+00, 1.89875651199601e+00, 1.88872678786817e+00, 1.87730968832086e+00, 1.86447851118067e+00, 1.85023209299807e+00, 1.83459613457549e+00, 1.81762355251664e+00, 1.79939381518968e+00, 1.78001128306008e+00, 1.75960263472116e+00, 1.73831351539172e+00, 1.71630458900849e+00, 1.69374720441329e+00, 1.67081889832832e+00, 1.64769895256082e+00, 1.62456420177312e+00, 1.60158525434992e+00, 1.57892324661288e+00, 1.55672720458067e+00, 1.53513204222457e+00, 1.51425718464413e+00, 1.49420577164415e+00, 1.47506437345279e+00, 1.45690313614924e+00, 1.43977626908617e+00, 1.42372278874564e+00, 1.40876744119426e+00, 1.39492173664326e+00, 1.38218504276313e+00, 1.37054569685509e+00, 1.35998210961983e+00, 1.35046384433708e+00, 1.34195266434559e+00, 1.33440354860782e+00, 1.32776567985038e+00, 1.32198341239771e+00, 1.31699722754147e+00, 1.31274468333913e+00, 1.30916136336976e+00, 1.30618182548781e+00, 1.30374054733359e+00, 1.30177286064763e+00, 1.30021586169027e+00, 1.29900928070235e+00, 1.29809628976882e+00, 1.29742422603723e+00, 1.29694520630501e+00, 1.29661660972064e+00, 1.29640140781317e+00, 1.29626832518934e+00, 1.29619181978807e+00, 1.29615187820287e+00, 1.29613362882048e+00, 1.29612678288881e+00, 1.29612492063334e+00, 1.29612464576848e+00, 1.29612462493410e+00, 1.29612425555884e+00, 1.29612175298928e+00, 1.29611255331162e+00, 1.29608803060033e+00, 1.29603436452352e+00, 1.29593159245954e+00, 1.29575288628846e+00, 1.29546409070386e+00, 1.29502355591180e+00, 1.29438229255553e+00, 1.29348447011991e+00, 1.29226827138967e+00, 1.29066710430104e+00, 1.28861115848626e+00, 1.28602927705240e+00, 1.28285109520391e+00, 1.27900937727793e+00, 1.27444246420346e+00, 1.26909672633907e+00, 1.26292890433866e+00, 1.25590821532762e+00, 1.24801810500890e+00, 1.23925753934547e+00, 1.22964175208847e+00, 1.21920239531406e+00, 1.20798707678147e+00, 1.19605830688793e+00, 1.18349191535777e+00, 1.17037502974517e+00, 1.15680373121105e+00, 1.14288051585564e+00, 1.12871169155368e+00, 1.11440483155947e+00, 1.10006638909810e+00, 1.08579955446365e+00, 1.07170241078659e+00, 1.05786641940283e+00, 1.04437524290501e+00, 1.03130389501470e+00, 1.01871819215746e+00, 1.00667447216887e+00, 9.95219540527813e-01, 9.84390803215067e-01, 9.74216546899246e-01, 9.64716330813499e-01, 9.55901459633022e-01, 9.47775512233402e-01, 9.40334906876323e-01, 9.33569488729276e-01, 9.27463130392642e-01, 9.21994340089629e-01, 9.17136875255395e-01, 9.12860361388507e-01, 9.09130917194005e-01, 9.05911787290417e-01, 9.03163983145095e-01, 9.00846931553623e-01, 8.98919128032208e-01, 8.97338790123730e-01, 8.96064503036293e-01, 8.95055847466988e-01, 8.94273997158216e-01, 8.93682271927887e-01, 8.93246630826905e-01, 8.92936089880742e-01, 8.92723049676931e-01, 8.92583519897009e-01, 8.92497230702567e-01, 8.92447624527179e-01, 8.92421726080570e-01, 8.92409892968580e-01, 8.92405453976266e-01, 8.92404246464302e-01, 8.92404068238111e-01, 8.92404054643091e-01, 8.92403812484378e-01, 8.92402171824461e-01, 8.92396140600400e-01, 8.92380063689330e-01, 8.92344880360948e-01, 8.92277502370408e-01, 8.92160338651738e-01, 8.91970990184937e-01, 8.91682135776669e-01, 8.91261626004368e-01, 8.90672798147740e-01, 8.89875019276602e-01, 8.88824457530433e-01, 8.87475072887039e-01, 8.85779808444244e-01, 8.83691951760959e-01, 8.81166623753074e-01, 8.78162340945508e-01, 8.74642586714405e-01, 8.70577319809762e-01, 8.65944345178594e-01, 8.60730473915552e-01, 8.54932406602266e-01, 8.48557287298391e-01, 8.41622893265676e-01, 8.34157446737247e-01, 8.26199057778396e-01, 8.17794829361997e-01, 8.08999675087811e-01, 7.99874914745857e-01, 7.90486721977395e-01, 7.80904501156350e-01, 7.71199267557584e-01, 7.61442096777262e-01, 7.51702697546698e-01, 7.42048148073373e-01, 7.32541821392854e-01, 7.23242511284134e-01, 7.14203758143132e-01, 7.05473364510912e-01, 6.97093083024903e-01, 6.89098455382137e-01, 6.81518779196686e-01, 6.74377179953279e-01, 6.67690767081918e-01, 6.61470855971451e-01, 6.55723241019945e-01, 6.50448508182645e-01, 6.45642378618238e-01, 6.41296077739341e-01, 6.37396726118848e-01, 6.33927750230814e-01, 6.30869311906633e-01, 6.28198755691329e-01, 6.25891073044059e-01, 6.23919381612012e-01, 6.22255416703993e-01, 6.20870030700374e-01, 6.19733694574403e-01, 6.18816994095642e-01, 6.18091111778121e-01, 6.17528284367982e-01, 6.17102224777021e-01, 6.16788496984483e-01, 6.16564832644890e-01, 6.16411379011549e-01, 6.16310869320330e-01, 6.16248708929913e-01, 6.16212973182609e-01, 6.16194315990281e-01, 6.16185791385160e-01, 6.16182593514740e-01, 6.16181723616755e-01, 6.16181595221647e-01, 6.16181585358607e-01, 6.16181408769030e-01, 6.16180212349072e-01, 6.16175814188690e-01, 6.16164090353596e-01, 6.16138433259120e-01, 6.16089297742178e-01, 6.16003853444873e-01, 6.15865760201142e-01, 6.15655080857668e-01, 6.15348343245689e-01, 6.14918759691818e-01, 6.14336608347568e-01, 6.13569775589013e-01, 6.12584452728726e-01, 6.11345973349177e-01, 6.09819769932908e-01, 6.07972420537095e-01, 6.05772748632237e-01, 6.03192932639502e-01, 6.00209576978912e-01, 5.96804694371965e-01, 5.92966550359499e-01, 5.88690325844222e-01, 5.83978561901333e-01, 5.78841362632012e-01, 5.73296345551711e-01, 5.67368343668964e-01, 5.61088877617546e-01, 5.54495428585685e-01, 5.47630552210457e-01, 5.40540879341894e-01, 5.33276051389279e-01, 5.25887636092826e-01, 5.18428064677568e-01, 5.10949624367651e-01, 5.03503532188454e-01, 4.96139107819662e-01, 4.88903055754932e-01, 4.81838860691425e-01, 4.74986295159770e-01, 4.68381034922917e-01, 4.62054375460500e-01, 4.56033041666004e-01, 4.50339082447957e-01, 4.44989842001306e-01, 4.39997999905909e-01, 4.35371672774135e-01, 4.31114570811051e-01, 4.27226203306450e-01, 4.23702127705869e-01, 4.20534237471248e-01, 4.17711084403194e-01, 4.15218231413288e-01, 4.13038631864275e-01, 4.11153031505606e-01, 4.09540388705317e-01, 4.08178308124626e-01, 4.07043482233188e-01, 4.06112134180731e-01, 4.05360454611358e-01, 4.04765024131020e-01, 4.04303212430392e-01, 4.03953544634184e-01, 4.03696025392733e-01, 4.03512411625623e-01, 4.03386425711613e-01, 4.03303902295203e-01, 4.03252863707877e-01, 4.03223521202225e-01, 4.03208201657872e-01, 4.03201202005945e-01, 4.03198576190707e-01, 4.03197861904911e-01, 4.03197756477809e-01, 4.03197748317311e-01, 4.03197601405839e-01, 4.03196606059026e-01, 4.03192947059534e-01, 4.03183193515570e-01, 4.03161848149942e-01, 4.03120969323719e-01, 4.03049881022550e-01, 4.02934984288414e-01, 4.02759681547046e-01, 4.02504422895895e-01, 4.02146880590792e-01, 4.01662254606606e-01, 4.01023708176755e-01, 4.00202927615773e-01, 3.99170795555566e-01, 3.97898161140052e-01, 3.96356685000072e-01, 3.94519731360084e-01, 3.92363274886954e-01, 3.89866786399164e-01, 3.87014059807431e-01, 3.83793943051316e-01, 3.80200938550527e-01, 3.76235643792157e-01, 3.71905009854423e-01, 3.67222404413543e-01, 3.62207475401824e-01, 3.56885821207484e-01, 3.51288482376290e-01, 3.45451277559579e-01, 3.39414012504800e-01, 3.33219594965961e-01, 3.26913090483185e-01, 3.20540754159254e-01, 3.14149072068169e-01, 3.07783843040237e-01, 3.01489327567094e-01, 2.95307485738544e-01, 2.89277320735376e-01, 2.83434338737241e-01, 2.77810130457926e-01, 2.72432074192643e-01, 2.67323155537333e-01, 2.62501895053049e-01, 2.57982372259946e-01, 2.53774332532615e-01, 2.49883362720436e-01, 2.46311121541744e-01, 2.43055611833178e-01, 2.40111483356674e-01, 2.37470356829629e-01, 2.35121161903503e-01, 2.33050483754766e-01, 2.31242914592460e-01, 2.29681407604250e-01, 2.28347631585194e-01, 2.27222324704709e-01, 2.26285645601241e-01, 2.25517519333621e-01, 2.24897974781015e-01, 2.24407469014338e-01, 2.24027193116331e-01, 2.23739353058424e-01, 2.23527418684931e-01, 2.23376333716050e-01, 2.23272680025552e-01, 2.23204790297626e-01, 2.23162804492084e-01, 2.23138667276103e-01, 2.23126065605559e-01, 2.23120307828236e-01, 2.23118147892034e-01, 2.23117560337478e-01, 2.23117473615673e-01, 2.23117466848624e-01, 2.23117344320016e-01, 2.23116514171068e-01, 2.23113462462262e-01, 2.23105327777251e-01, 2.23087525472934e-01, 2.23053432907214e-01, 2.22994148706928e-01, 2.22898337891692e-01, 2.22752173201300e-01, 2.22539381193272e-01, 2.22241399465332e-01, 2.21837648591467e-01, 2.21305918939418e-01, 2.20622868414555e-01, 2.19764622381884e-01, 2.18707461690268e-01, 2.17428579109683e-01, 2.15906878953827e-01, 2.14123789649790e-01, 2.12064055040035e-01, 2.09716467770598e-01, 2.07074507682314e-01, 2.04136850006730e-01, 2.00907712522984e-01, 1.97397017586241e-01, 1.93620353787864e-01, 1.89598732422350e-01, 1.85358145195391e-01, 1.80928940858923e-01, 1.76345048800961e-01, 1.71643086195647e-01, 1.66861391417612e-01, 1.62039029549252e-01, 1.57214815730665e-01, 1.52426398878109e-01, 1.47709442275932e-01, 1.43096929343940e-01, 1.38618613322258e-01, 1.34300619626664e-01, 1.30165200103903e-01, 1.26230630096909e-01, 1.22511232633070e-01, 1.19017509467440e-01, 1.15756356231579e-01, 1.12731338454282e-01, 1.09943006448070e-01, 1.07389229578708e-01, 1.05065533779264e-01, 1.02965429893377e-01, 1.01080724178200e-01, 9.94018058019748e-02, 9.79179092427103e-02, 9.66173519943396e-02, 9.54877498301379e-02, 9.45162130337600e-02, 9.36895275151350e-02, 9.29943246536240e-02, 9.24172431501837e-02, 9.19450852322115e-02, 9.15649683570905e-02, 9.12644722268681e-02, 9.10317795872508e-02, 9.08558080674844e-02, 9.07263293426261e-02, 9.06340712626168e-02, 9.05707983616466e-02, 9.05293663645814e-02, 9.05037469346119e-02, 9.04890199060588e-02, 9.04813315340149e-02, 9.04778187592932e-02, 9.04765010133666e-02, 9.04761425561382e-02, 9.04760896486607e-02, 9.04760854850792e-02, 9.04760096463768e-02, 9.04754958289722e-02, 9.04736069962162e-02, 9.04685721719183e-02, 9.04575541420769e-02, 9.04364554426813e-02, 9.03997713425824e-02, 9.03404981143758e-02, 9.02501043542898e-02, 9.01185723325533e-02, 8.99345151619571e-02, 8.96853737876597e-02, 8.93576952714845e-02, 8.89374904739029e-02, 8.84106650271441e-02, 8.77635125715734e-02, 8.69832538690645e-02, 8.60586000350594e-02, 8.49803132947216e-02, 8.37417349952894e-02, 8.23392487289816e-02, 8.07726468843697e-02, 7.90453721095936e-02, 7.71646111281719e-02, 7.51412268561253e-02, 7.29895252396915e-02, 7.07268647703394e-02, 6.83731281200615e-02, 6.59500855689998e-02, 6.34806877345078e-02, 6.09883296609119e-02, 5.84961290890682e-02, 5.60262586751712e-02, 5.35993655483473e-02, 5.12341027662835e-02, 4.89467870640610e-02, 4.67511869525978e-02, 4.46584357544815e-02, 4.26770563978351e-02, 4.08130792757007e-02, 3.90702314488749e-02, 3.74501748083205e-02, 3.59527721034603e-02, 3.45763623940698e-02, 3.33180309004248e-02, 3.21738619247185e-02, 3.11391671583721e-02, 3.02086850349887e-02, 2.93767496530108e-02, 2.86374300502530e-02, 2.79846422195611e-02, 2.74122372665957e-02, 2.69140696435777e-02, 2.64840495732876e-02, 2.61161836986190e-02, 2.58046077077623e-02, 2.55436142242651e-02, 2.53276786467218e-02, 2.51514849151029e-02, 2.50099524171929e-02, 2.48982644782991e-02, 2.48118981441756e-02, 2.47466543139261e-02, 2.46986867444186e-02, 2.46645280664595e-02, 2.46411107532438e-02, 2.46257809789441e-02, 2.46163034993572e-02, 2.46108560589775e-02, 2.46080123433394e-02, 2.46067131036121e-02, 2.46062257264619e-02, 2.46060931492451e-02, 2.46060735811682e-02, 2.46060720276127e-02, 2.46060435566768e-02, 2.46058506626028e-02, 2.46051415736375e-02, 2.46032514802010e-02, 2.45991154496892e-02, 2.45911959612393e-02, 2.45774286123037e-02, 2.45551895520073e-02, 2.45212881617540e-02, 2.44719883437341e-02, 2.44030613817721e-02, 2.43098725869793e-02, 2.41875027269636e-02, 2.40309034938768e-02, 2.38350839933439e-02, 2.35953225216967e-02, 2.33073949394189e-02, 2.29678080432620e-02, 2.25740238734780e-02, 2.21246592936772e-02, 2.16196448613211e-02, 2.10603282889537e-02, 2.04495108275786e-02, 1.97914095790068e-02, 1.90915446511895e-02, 1.83565565108168e-02, 1.75939649970474e-02, 1.68118864164093e-02, 1.60187283667845e-02, 1.52228831845100e-02, 1.44324401585720e-02, 1.36549339927861e-02, 1.28971426091304e-02, 1.21649417623537e-02, 1.14632180202037e-02, 1.07958366254302e-02, 1.01656573973834e-02, 9.57459010157535e-03, 9.02367989344555e-03, 8.51321292529761e-03, 8.04283212777790e-03, 7.61165416642484e-03, 7.21838084857040e-03, 6.86140107685063e-03, 6.53888160254468e-03, 6.24884569606721e-03, 5.98923889569036e-03, 5.75798123134774e-03, 5.55300628596100e-03, 5.37228877617579e-03, 5.21386319622508e-03, 5.07583602436767e-03, 4.95639328920205e-03, 4.85380455162243e-03, 4.76642401940452e-03, 4.69268958297064e-03, 4.63112075753358e-03, 4.58031656205904e-03, 4.53895417236794e-03, 4.50578884071007e-03, 4.47965521343718e-03, 4.45946989405144e-03, 4.44423491016011e-03, 4.43304162815750e-03, 4.42507459175800e-03, 4.41961472620633e-03, 4.41604135030360e-03, 4.41383247930460e-03, 4.41256298564596e-03, 4.41190030806027e-03, 4.41159755089515e-03, 4.41148398038393e-03, 4.41145308684199e-03, 4.41144852703346e-03, 4.41144816166883e-03, 4.41144142369509e-03, 4.41139577315042e-03, 4.41122796030413e-03, 4.41078065879807e-03, 4.40980188204421e-03, 4.40792791176747e-03, 4.40467063898353e-03, 4.39941025111779e-03, 4.39139422341839e-03, 4.37974355774884e-03, 4.36346711897428e-03, 4.34148470877815e-03, 4.31265915603873e-03, 4.27583717827673e-03, 4.22989809367635e-03, 4.17380868342931e-03, 4.10668168753503e-03, 4.02783463794520e-03, 3.93684505862879e-03, 3.83359756342116e-03, 3.71831817707857e-03, 3.59159149684984e-03, 3.45435737025568e-03, 3.30788574765961e-03, 3.15373106593597e-03, 2.99367019433916e-03, 2.82962960760358e-03, 2.66360742502414e-03, 2.49759472353617e-03, 2.33349954465467e-03, 2.17307752208740e-03, 2.01787465245928e-03, 1.86918817412618e-03, 1.72804867524690e-03, 1.59522100979359e-03, 1.47121709195730e-03, 1.35631396026170e-03, 1.25057539529649e-03, 1.15387992548222e-03, 1.06595723076452e-03, 9.86429104649744e-04, 9.14846248260659e-04, 8.50713902391220e-04, 7.93506937021959e-04, 7.42681717039062e-04, 6.97691673341438e-04, 6.58006719557847e-04, 6.23130179608467e-04, 5.92606552391572e-04, 5.66018884843089e-04, 5.42980560201300e-04, 5.23128201972867e-04, 5.06119565969243e-04, 4.91635808408830e-04, 4.79384829223051e-04, 4.69102663576138e-04, 4.60551922241592e-04, 4.53518180041357e-04, 4.47805933713852e-04, 4.43235388612351e-04, 4.39640548204778e-04, 4.36868448680627e-04, 4.34779114755202e-04, 4.33245833818319e-04, 4.32155482027241e-04, 4.31408761802032e-04, 4.30920280287239e-04, 4.30618421628282e-04, 4.30444969207855e-04, 4.30354436559810e-04, 4.30313077195876e-04, 4.30297562804327e-04, 4.30293342599697e-04, 4.30292719709513e-04, 4.30292669395982e-04, 4.30291736435606e-04, 4.30285415522036e-04, 4.30262179680341e-04, 4.30200245169488e-04, 4.30064721893331e-04, 4.29805251325632e-04, 4.29354258894676e-04, 4.28625957333395e-04, 4.27516251573223e-04, 4.25903717580667e-04, 4.23651855154361e-04, 4.20612840038359e-04, 4.16632996167010e-04, 4.11560118106280e-04, 4.05252520270238e-04, 3.97589217744354e-04, 3.88480006183300e-04, 3.77873662582573e-04, 3.65762499601630e-04, 3.52182535110046e-04, 3.37210582006975e-04, 3.20961686767873e-04, 3.03590716150381e-04, 2.85298830274172e-04, 2.66339487154710e-04, 2.47013525395801e-04, 2.27645056973813e-04, 2.08541962685160e-04, 1.89959027341757e-04, 1.72084120521276e-04, 1.55054698566914e-04, 1.38994998074793e-04, 1.24047640319702e-04, 1.10351250640406e-04, 9.79282095130429e-05, 8.66229574469912e-05, 7.62660549042948e-05, 6.71631557906949e-05, 5.90470309838054e-05, 5.16700503077061e-05, 4.49914773561488e-05, 3.88355387990202e-05, 3.29247312507164e-05, 2.70228118495197e-05, 2.11270372751926e-05, 1.52982171428562e-05, 9.87477445590912e-06, 5.24657868439249e-06, 1.64230141663779e-06, 1.58573197595047e-06, 1.27435328423076e-06, 3.66433411613798e-07, 1.20223422683807e-06, 1.44774848465090e-06, 1.37410182787789e-06, 1.17680162503575e-06, 9.90616211282713e-07, 8.75654368570622e-07, 8.22274422773717e-07, 7.89883586094220e-07, 7.46491991776566e-07, 6.79140114812071e-07, 5.88814561552908e-07, 4.84867364036745e-07, 3.78969915004745e-07, 2.80827702425768e-07, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 3.55271367880050e-15, 1.06581410364015e-14, 7.10542735760100e-15, 7.10542735760100e-15],
[ 2.80000000000000e+01, 2.79999999800081e+01, 2.79999993618179e+01, 2.79999951734821e+01, 2.79999797766197e+01, 2.79999387334937e+01, 2.79998489078419e+01, 2.79996768665084e+01, 2.79993776399185e+01, 2.79988938886413e+01, 2.79981555110431e+01, 2.79970797137607e+01, 2.79955715529022e+01, 2.79935249399987e+01, 2.79908240932975e+01, 2.79873454024700e+01, 2.79829596636043e+01, 2.79775346317585e+01, 2.79709378305432e+01, 2.79630395521953e+01, 2.79537159773103e+01, 2.79428523405925e+01, 2.79303460673993e+01, 2.79161098052405e+01, 2.79000742745586e+01, 2.78821908640268e+01, 2.78624338974009e+01, 2.78408025019644e+01, 2.78173220132999e+01, 2.77920448580117e+01, 2.77650508656183e+01, 2.77364469734197e+01, 2.77063663037639e+01, 2.76749666114489e+01, 2.76424281193112e+01, 2.76089507813291e+01, 2.75747510335520e+01, 2.75400581125167e+01, 2.75051100372327e+01, 2.74701493632783e+01, 2.74354188252953e+01, 2.74011569868906e+01, 2.73675940147062e+01, 2.73349476866422e+01, 2.73034197336073e+01, 2.72731926006170e+01, 2.72444266974791e+01, 2.72172581926261e+01, 2.71917973866484e+01, 2.71681276854044e+01, 2.71463051767034e+01, 2.71263587997758e+01, 2.71082910832420e+01, 2.70920794151359e+01, 2.70776777977382e+01, 2.70650190305304e+01, 2.70540172564731e+01, 2.70445708000856e+01, 2.70365652205048e+01, 2.70298764989389e+01, 2.70243742778166e+01, 2.70199250686128e+01, 2.70163953468996e+01, 2.70136544567325e+01, 2.70115772520432e+01, 2.70100464102374e+01, 2.70089543625612e+01, 2.70082047968001e+01, 2.70077137002293e+01, 2.70074099240946e+01, 2.70072352648795e+01, 2.70071440717768e+01, 2.70071024036957e+01, 2.70070867723660e+01, 2.70070825202470e+01, 2.70070818926424e+01, 2.70070818103140e+01, 2.70070798921590e+01, 2.70070668963622e+01, 2.70070191222215e+01, 2.70068917722362e+01, 2.70066130614523e+01, 2.70060792645589e+01, 2.70051508859086e+01, 2.70036501070869e+01, 2.70013596319234e+01, 2.69980230107728e+01, 2.69933464854168e+01, 2.69870023541055e+01, 2.69786338141523e+01, 2.69678611982342e+01, 2.69542894810913e+01, 2.69375168963564e+01, 2.69171444690894e+01, 2.68927862381121e+01, 2.68640799129974e+01, 2.68306976831233e+01, 2.67923568705327e+01, 2.67488300954186e+01, 2.66999546052124e+01, 2.66456404093739e+01, 2.65858768670970e+01, 2.65207373996443e+01, 2.64503820472865e+01, 2.63750576648456e+01, 2.62950956479850e+01, 2.62109071988236e+01, 2.61229762645109e+01, 2.60318504038772e+01, 2.59381299424136e+01, 2.58424558536372e+01, 2.57454968480277e+01, 2.56479361567608e+01, 2.55504584690262e+01, 2.54537374256586e+01, 2.53584239976319e+01, 2.52651359959542e+01, 2.51744488790258e+01, 2.50868879517393e+01, 2.50029219918838e+01, 2.49229582953389e+01, 2.48473391013896e+01, 2.47763393409874e+01, 2.47101656408860e+01, 2.46489565121516e+01, 2.45927836498351e+01, 2.45416542694302e+01, 2.44955144036727e+01, 2.44542530793972e+01, 2.44177072882239e+01, 2.43856676568188e+01, 2.43578847126735e+01, 2.43340756302800e+01, 2.43139313309115e+01, 2.42971237977371e+01, 2.42833134576125e+01, 2.42721564725461e+01, 2.42633117785372e+01, 2.42564477081600e+01, 2.42512480367345e+01, 2.42474173008069e+01, 2.42446852522914e+01, 2.42428103319929e+01, 2.42415820719607e+01, 2.42408223664632e+01, 2.42403855852157e+01, 2.42401575384909e+01, 2.42400533403347e+01, 2.42400142517301e+01, 2.42400036186555e+01, 2.42400020492351e+01, 2.42400019106851e+01, 2.42399991958668e+01, 2.42399808025651e+01, 2.42399131867219e+01, 2.42397329467838e+01, 2.42393384913921e+01, 2.42385830441326e+01, 2.42372692586905e+01, 2.42351456808977e+01, 2.42319052516802e+01, 2.42271859965462e+01, 2.42205739924846e+01, 2.42116086414393e+01, 2.41997902106271e+01, 2.41845895241665e+01, 2.41654596089903e+01, 2.41418490131344e+01, 2.41132164298249e+01, 2.40790461812088e+01, 2.40388640469940e+01, 2.39922528722721e+01, 2.39388673619047e+01, 2.38784474717264e+01, 2.38108298432060e+01, 2.37359567990321e+01, 2.36538825196176e+01, 2.35647761481361e+01, 2.34689217142217e+01, 2.33667149114096e+01, 2.32586568978502e+01, 2.31453454027654e+01, 2.30274635054720e+01, 2.29057665076964e+01, 2.27810673466232e+01, 2.26542210025207e+01, 2.25261083488176e+01, 2.23976198803712e+01, 2.22696397398590e+01, 2.21430304410652e+01, 2.20186186567820e+01, 2.18971823933647e+01, 2.17794398115331e+01, 2.16660398760767e+01, 2.15575549324820e+01, 2.14544752256172e+01, 2.13572053039664e+01, 2.12660621993337e+01, 2.11812752391724e+01, 2.11029873352915e+01, 2.10312575941769e+01, 2.09660651046526e+01, 2.09073137724020e+01, 2.08548380833995e+01, 2.08084096866453e+01, 2.07677446893257e+01, 2.07325115544744e+01, 2.07023394830655e+01, 2.06768271504491e+01, 2.06555516526651e+01, 2.06380775031325e+01, 2.06239655062529e+01, 2.06127813233374e+01, 2.06041035396028e+01, 2.05975310402570e+01, 2.05926895100439e+01, 2.05892368848083e+01, 2.05868676058910e+01, 2.05853155581613e+01, 2.05843556093168e+01, 2.05838037102952e+01, 2.05835155623752e+01, 2.05833839036366e+01, 2.05833345136437e+01, 2.05833210783441e+01, 2.05833190953212e+01, 2.05833189362582e+01, 2.05833160007394e+01, 2.05832961121615e+01, 2.05832229995503e+01, 2.05830281074432e+01, 2.05826015871348e+01, 2.05817847355437e+01, 2.05803641794090e+01, 2.05780680638967e+01, 2.05745644600814e+01, 2.05694621546295e+01, 2.05623139265043e+01, 2.05526223487568e+01, 2.05398480778822e+01, 2.05234205089456e+01, 2.05027505829077e+01, 2.04772454359020e+01, 2.04463244827722e+01, 2.04094364346793e+01, 2.03660766700501e+01, 2.03158043174253e+01, 2.02582583758204e+01, 2.01931722001168e+01, 2.01203857208221e+01, 2.00398548511378e+01, 1.99516576572569e+01, 1.98559970227954e+01, 1.97531997129845e+01, 1.96437119226189e+01, 1.95280915562524e+01, 1.94069976239606e+01, 1.92811772305564e+01, 1.91514506872854e+01, 1.90186952873971e+01, 1.88838282708462e+01, 1.87477894709084e+01, 1.86115240967468e+01, 1.84759660663455e+01, 1.83420222640862e+01, 1.82105580536280e+01, 1.80823843259966e+01, 1.79582463029992e+01, 1.78388142483161e+01, 1.77246761668290e+01, 1.76163325025902e+01, 1.75141927831794e+01, 1.74185741076671e+01, 1.73297013396018e+01, 1.72477088456668e+01, 1.71726436132223e+01, 1.71044695827966e+01, 1.70430730409705e+01, 1.69882689312629e+01, 1.69398079522623e+01, 1.68973843208613e+01, 1.68606440824207e+01, 1.68291938483576e+01, 1.68026098351930e+01, 1.67804470684249e+01, 1.67622486012028e+01, 1.67475545835670e+01, 1.67359110051081e+01, 1.67268779244733e+01, 1.67200369951980e+01, 1.67149981005720e+01, 1.67114049218165e+01, 1.67089392843087e+01, 1.67073241557757e+01, 1.67063252074089e+01, 1.67057508921955e+01, 1.67054510423558e+01, 1.67053140372224e+01, 1.67052626416622e+01, 1.67052486608044e+01, 1.67052465972583e+01, 1.67052464380978e+01, 1.67052435800855e+01, 1.67052242166266e+01, 1.67051530344321e+01, 1.67049632882875e+01, 1.67045480307082e+01, 1.67037527521016e+01, 1.67023697225638e+01, 1.67001342895697e+01, 1.66967233432120e+01, 1.66917561116550e+01, 1.66847973931687e+01, 1.66753632665118e+01, 1.66629292479095e+01, 1.66469407804827e+01, 1.66268258518938e+01, 1.66020094406701e+01, 1.65719293951657e+01, 1.65360532568633e+01, 1.64938954583249e+01, 1.64450342628810e+01, 1.63891277753845e+01, 1.63259283475095e+01, 1.62552947319133e+01, 1.61772014094131e+01, 1.60917446213936e+01, 1.59991447817846e+01, 1.58997451115260e+01, 1.57940065227454e+01, 1.56824989667945e+01, 1.55658896353795e+01, 1.54449285529677e+01, 1.53204322089329e+01, 1.51932659405252e+01, 1.50643257886206e+01, 1.49345205090002e+01, 1.48047543398979e+01, 1.46759110136053e+01, 1.45488393705549e+01, 1.44243408033746e+01, 1.43031586388827e+01, 1.41859694672031e+01, 1.40733763540230e+01, 1.39659038250014e+01, 1.38639944875224e+01, 1.37680071491988e+01, 1.36782162988092e+01, 1.35948128280747e+01, 1.35179058872979e+01, 1.34475257812299e+01, 1.33836278217402e+01, 1.33260970601132e+01, 1.32747538238938e+01, 1.32293599813540e+01, 1.31896258511630e+01, 1.31552176661051e+01, 1.31257654881613e+01, 1.31008714585439e+01, 1.30801182511422e+01, 1.30630775823660e+01, 1.30493186158456e+01, 1.30384160883696e+01, 1.30299579753823e+01, 1.30235525118283e+01, 1.30188343884314e+01, 1.30154699555238e+01, 1.30131612867093e+01, 1.30116489827628e+01, 1.30107136314053e+01, 1.30101758795602e+01, 1.30098951195003e+01, 1.30097668367377e+01, 1.30097187132549e+01, 1.30097056224829e+01, 1.30097036903117e+01, 1.30097035440136e+01, 1.30097009523559e+01, 1.30096833934889e+01, 1.30096188451637e+01, 1.30094467824697e+01, 1.30090702246409e+01, 1.30083490604988e+01, 1.30070949160401e+01, 1.30050677939526e+01, 1.30019746758543e+01, 1.29974702332248e+01, 1.29911597415000e+01, 1.29826042327170e+01, 1.29713278548511e+01, 1.29568273306433e+01, 1.29385833264167e+01, 1.29160734545868e+01, 1.28887865462908e+01, 1.28562377482591e+01, 1.28179839273639e+01, 1.27736388144190e+01, 1.27228872926621e+01, 1.26654982415162e+01, 1.26013353858610e+01, 1.25303656750881e+01, 1.24526648208548e+01, 1.23684197503260e+01, 1.22779278726382e+01, 1.21815931988258e+01, 1.20799194884372e+01, 1.19735007106406e+01, 1.18630091983948e+01, 1.17491819397393e+01, 1.16328054922747e+01, 1.15147000290314e+01, 1.13957030295948e+01, 1.12766531212184e+01, 1.11583745499288e+01, 1.10416627188867e+01, 1.09272711682778e+01, 1.08159002880736e+01, 1.07081879565351e+01, 1.06047021920020e+01, 1.05059358045867e+01, 1.04123029490132e+01, 1.03241374181661e+01, 1.02416924823437e+01, 1.01651420701032e+01, 1.00945830974543e+01, 1.00300387755418e+01, 9.97146275540857e+00, 9.91874399581927e+00, 9.87171226233029e+00, 9.83014418064888e+00, 9.79376977426984e+00, 9.76227941592230e+00, 9.73533111565454e+00, 9.71255805690927e+00, 9.69357627732659e+00, 9.67799237496026e+00, 9.66541110484579e+00, 9.65544271715635e+00, 9.64770987820821e+00, 9.64185401086603e+00, 9.63754089272547e+00, 9.63446535970660e+00, 9.63235497977648e+00, 9.63097258626203e+00, 9.63011759185143e+00, 9.62962604159016e+00, 9.62936940417388e+00, 9.62925214353251e+00, 9.62920815487426e+00, 9.62919618887912e+00, 9.62919442272281e+00, 9.62919429013689e+00, 9.62919195647438e+00, 9.62917614556409e+00, 9.62911802293432e+00, 9.62896308895130e+00, 9.62862401773073e+00, 9.62797464816003e+00, 9.62684536517989e+00, 9.62502007961099e+00, 9.62223498024484e+00, 9.61817919204898e+00, 9.61249742828585e+00, 9.60479467118439e+00, 9.59464285488060e+00, 9.58158945562397e+00, 9.56516781865874e+00, 9.54490897090249e+00, 9.52035458713579e+00, 9.49107069986016e+00, 9.45666167542598e+00, 9.41678392833705e+00, 9.37115881867341e+00, 9.31958418032530e+00, 9.26194396428566e+00, 9.19821555280066e+00, 9.12847440421924e+00, 9.05289581831593e+00, 8.97175375724941e+00, 8.88541680495684e+00, 8.79434148357296e+00, 8.69906325684103e+00, 8.60018562883188e+00, 8.49836778861305e+00, 8.39431126073066e+00, 8.28874600502868e+00, 8.18241637707924e+00, 8.07606732122123e+00, 7.97043112781885e+00, 7.86621504697247e+00, 7.76409001148463e+00, 7.66468067980913e+00, 7.56855696270801e+00, 7.47622714480934e+00, 7.38813265578183e+00, 7.30464448884892e+00, 7.22606121121517e+00, 7.15260846589245e+00, 7.08443983056448e+00, 7.02163887809996e+00, 6.96422227497763e+00, 6.91214375657442e+00, 6.86529882926013e+00, 6.82353006525183e+00, 6.78663287385642e+00, 6.75436164908213e+00, 6.72643620628902e+00, 6.70254842807080e+00, 6.68236904130248e+00, 6.66555444350165e+00, 6.65175348834643e+00, 6.64061412897804e+00, 6.63178980559716e+00, 6.62494545302049e+00, 6.61976299642064e+00, 6.61594620127792e+00, 6.61322474801322e+00, 6.61135741363067e+00, 6.61013426207980e+00, 6.60937777136468e+00, 6.60894285747022e+00, 6.60871579123509e+00, 6.60861204234314e+00, 6.60857312247855e+00, 6.60856253532479e+00, 6.60856097268282e+00, 6.60856085573151e+00, 6.60855880200785e+00, 6.60854488772685e+00, 6.60849373734395e+00, 6.60835738909475e+00, 6.60805899373719e+00, 6.60748752882769e+00, 6.60649373998769e+00, 6.60488749543676e+00, 6.60243671087922e+00, 6.59886796796322e+00, 6.59386891004946e+00, 6.58709245297597e+00, 6.57816279520929e+00, 6.56668315113876e+00, 6.55224506407655e+00, 6.53443908362551e+00, 6.51286651849095e+00, 6.48715190467494e+00, 6.45695576532455e+00, 6.42198718782692e+00, 6.38201571155389e+00, 6.33688201083297e+00, 6.28650687591789e+00, 6.23089804186878e+00, 6.17015449113737e+00, 6.10446795780681e+00, 6.03412148511437e+00, 5.95948502629619e+00, 5.88100822347382e+00, 5.79921064065322e+00, 5.71466985488090e+00, 5.62800791453052e+00, 5.53987674724896e+00, 5.45094313625867e+00, 5.36187387965743e+00, 5.27332170396514e+00, 5.18591242513066e+00, 5.10023374554066e+00, 5.01682595460558e+00, 4.93617467455674e+00, 4.85870567304519e+00, 4.78478165912923e+00, 4.71470089579965e+00, 4.64869740380892e+00, 4.58694249877151e+00, 4.52954739430400e+00, 4.47656661457036e+00, 4.42800198512549e+00, 4.38380700620090e+00, 4.34389145256463e+00, 4.30812608445201e+00, 4.27634739128546e+00, 4.24836232138961e+00, 4.22395297498226e+00, 4.20288125353519e+00, 4.18489346604464e+00, 4.16972489235115e+00, 4.15710429646059e+00, 4.14675837033145e+00, 4.13841607263319e+00, 4.13181280958385e+00, 4.12669438826514e+00, 4.12282065884029e+00, 4.11996875268335e+00, 4.11793581999565e+00, 4.11654117393554e+00, 4.11562775890241e+00, 4.11506287801975e+00, 4.11473813806377e+00, 4.11456859757536e+00, 4.11449113381887e+00, 4.11446207460157e+00, 4.11445416980046e+00, 4.11445300306876e+00, 4.11445291571550e+00, 4.11445138131195e+00, 4.11444098550810e+00, 4.11440276952323e+00, 4.11430090042820e+00, 4.11407796590483e+00, 4.11365103303364e+00, 4.11290863516298e+00, 4.11170883556366e+00, 4.10987849844694e+00, 4.10721387580369e+00, 4.10348259082172e+00, 4.09842706369658e+00, 4.09176938209219e+00, 4.08321756564223e+00, 4.07247311208832e+00, 4.05923964368326e+00, 4.04323239966052e+00, 4.02418824873982e+00, 4.00187583096247e+00, 3.97610538759120e+00, 3.94673780842463e+00, 3.91369242391688e+00, 3.87695309945930e+00, 3.83657225292209e+00, 3.79267251259898e+00, 3.74544585594000e+00, 3.69515021131257e+00, 3.64210365412332e+00, 3.58667647196922e+00, 3.52928149798547e+00, 3.47036320573010e+00, 3.41038611442668e+00, 3.34982306607863e+00, 3.28914390653896e+00, 3.22880503628922e+00, 3.16924020232612e+00, 3.11085279130083e+00, 3.05400976766741e+00, 2.99903728991678e+00, 2.94621794171640e+00, 2.89578943880462e+00, 2.84794461960693e+00, 2.80283249776986e+00, 2.76056014595944e+00, 2.72119518869196e+00, 2.68476870327629e+00, 2.65127835767521e+00, 2.62069164813033e+00, 2.59294913428164e+00, 2.56796760255324e+00, 2.54564311783344e+00, 2.52585394769987e+00, 2.50846336193524e+00, 2.49332232261224e+00, 2.48027208670206e+00, 2.46914674436501e+00, 2.45977571242025e+00, 2.45198619476771e+00, 2.44560561072698e+00, 2.44046397949403e+00, 2.43639623544180e+00, 2.43324443610850e+00, 2.43085981371095e+00, 2.42910461305932e+00, 2.42785365478108e+00, 2.42699556341704e+00, 2.42643360546527e+00, 2.42608609262231e+00, 2.42588631968171e+00, 2.42578202381054e+00, 2.42573437098932e+00, 2.42571649490418e+00, 2.42571163218818e+00, 2.42571091446193e+00, 2.42571086058318e+00, 2.42570991227221e+00, 2.42570348733216e+00, 2.42567986866822e+00, 2.42561691071375e+00, 2.42547913248479e+00, 2.42521528492724e+00, 2.42475649635115e+00, 2.42401509162273e+00, 2.42288417101183e+00, 2.42123802210795e+00, 2.41893342165636e+00, 2.41581186208078e+00, 2.41170270861404e+00, 2.40642725666572e+00, 2.39980361550426e+00, 2.39165229487298e+00, 2.38180231848235e+00, 2.37009763644517e+00, 2.35640356277175e+00, 2.34061292977828e+00, 2.32265163436906e+00, 2.30248325639235e+00, 2.28011245958203e+00, 2.25558694135158e+00, 2.22899777623770e+00, 2.20047809338357e+00, 2.17020013283172e+00, 2.13837082869730e+00, 2.10522615936274e+00, 2.07102457666706e+00, 2.03603987106142e+00, 2.00055384457687e+00, 1.96484914854000e+00, 1.92920260198949e+00, 1.89387924599547e+00, 1.85912731636757e+00, 1.82517424065437e+00, 1.79222369222012e+00, 1.76045367031544e+00, 1.73001552425665e+00, 1.70103380389121e+00, 1.67360679746007e+00, 1.64780761039102e+00, 1.62368564218672e+00, 1.60126833067832e+00, 1.58056305071605e+00, 1.56155907529713e+00, 1.54422952901599e+00, 1.52853328485331e+00, 1.51441677445804e+00, 1.50181569839817e+00, 1.49065663589251e+00, 1.48085856309062e+00, 1.47233429505256e+00, 1.46499186935672e+00, 1.45873588901260e+00, 1.45346883944930e+00, 1.44909238925204e+00, 1.44550867757110e+00, 1.44262158335187e+00, 1.44033796341496e+00, 1.43856883867397e+00, 1.43723050113239e+00, 1.43624550941262e+00, 1.43554353798199e+00, 1.43506204532382e+00, 1.43474672920049e+00, 1.43455174276816e+00, 1.43443965327976e+00, 1.43438113488678e+00, 1.43435439790262e+00, 1.43434436802624e+00, 1.43434163966512e+00, 1.43434123696504e+00, 1.43434120660615e+00, 1.43434067054906e+00, 1.43433703868769e+00, 1.43432368764492e+00, 1.43428809915716e+00, 1.43421021711068e+00, 1.43406107320994e+00, 1.43380174055221e+00, 1.43338267013165e+00, 1.43274346012664e+00, 1.43181310105892e+00, 1.43051073067899e+00, 1.42874691980373e+00, 1.42642549351160e+00, 1.42344587063177e+00, 1.41970587844501e+00, 1.41510496978803e+00, 1.40954773801445e+00, 1.40294759406637e+00, 1.39523044250289e+00, 1.38633817335327e+00, 1.37623177766245e+00, 1.36489389946792e+00, 1.35233065734650e+00, 1.33857260454466e+00, 1.32367474604098e+00, 1.30771558976445e+00, 1.29079527217442e+00, 1.27303285924802e+00, 1.25456297642615e+00, 1.23553196000613e+00, 1.21609374430923e+00, 1.19640570232806e+00, 1.17662464335451e+00, 1.15690314215387e+00, 1.13738633484496e+00, 1.11820927171180e+00, 1.09949487160952e+00, 1.08135248067184e+00, 1.06387700283214e+00, 1.04714854311828e+00, 1.03123248740812e+00, 1.01617993391910e+00, 1.00202839096070e+00, 9.88802660760953e-01, 9.76515838685430e-01, 9.65170369167545e-01, 9.54759112662966e-01, 9.45266390737707e-01, 9.36668988139537e-01, 9.28937100831710e-01, 9.22035227185781e-01, 9.15923005730722e-01, 9.10556007065686e-01, 9.05886489882029e-01, 9.01864131675332e-01, 8.98436743864941e-01, 8.95550978905909e-01, 8.93153033830345e-01, 8.91189350774390e-01, 8.89607310745834e-01, 8.88355912507297e-01, 8.87386424349817e-01, 8.86652993071007e-01, 8.86113191984148e-01, 8.85728488544704e-01, 8.85464612376623e-01, 8.85291806184583e-01, 8.85184945192456e-01, 8.85123515164345e-01, 8.85091444438824e-01, 8.85076791352912e-01, 8.85071294521978e-01, 8.85069799255163e-01, 8.85069578557076e-01, 8.85069561828473e-01, 8.85069265244880e-01, 8.85067255849769e-01, 8.85059869135844e-01, 8.85040179137977e-01, 8.84997089426363e-01, 8.84914572712098e-01, 8.84771091843866e-01, 8.84539232611715e-01, 8.84185575898030e-01, 8.83670832949598e-01, 8.82950262399934e-01, 8.81974380629025e-01, 8.80689967660256e-01, 8.79041358805303e-01, 8.76971997737137e-01, 8.74426210082074e-01, 8.71351138948668e-01, 8.67698766518519e-01, 8.63427930777704e-01, 8.58506235743068e-01, 8.52911749143534e-01, 8.46634385052134e-01, 8.39676881284053e-01, 8.32055302314817e-01, 8.23799026718916e-01, 8.14950211281204e-01, 8.05562758766605e-01, 7.95700849238891e-01, 7.85437122335793e-01, 7.74850617253435e-01, 7.64024586673184e-01, 7.53044300090014e-01, 7.41994941868992e-01, 7.30959691815283e-01, 7.20018053705864e-01, 7.09244472930543e-01, 6.98707260751863e-01, 6.88467821826197e-01, 6.78580164936790e-01, 6.69090665043228e-01, 6.60038037762316e-01, 6.51453484771999e-01, 6.43360969564494e-01, 6.35777586520341e-01, 6.28713991487274e-01, 6.22174868090582e-01, 6.16159410195856e-01, 6.10661806791715e-01, 6.05671720726342e-01, 6.01174757026691e-01, 5.97152919870961e-01, 5.93585059668293e-01, 5.90447313167317e-01, 5.87713540141813e-01, 5.85355760074950e-01, 5.83344591484661e-01, 5.81649695210231e-01, 5.80240221235989e-01, 5.79085256596759e-01, 5.78154269744648e-01, 5.77417544623021e-01, 5.76846595769450e-01, 5.76414554229185e-01, 5.76096513068784e-01, 5.75869820966940e-01, 5.75714312813780e-01, 5.75612467498903e-01, 5.75549485072507e-01, 5.75513278118102e-01, 5.75494375314072e-01, 5.75485738578479e-01, 5.75482498658847e-01, 5.75481617323852e-01, 5.75481487240715e-01, 5.75481477318025e-01, 5.75481300572037e-01, 5.75480103092911e-01, 5.75475701048035e-01, 5.75463966928204e-01, 5.75438287680530e-01, 5.75389111096598e-01, 5.75303599635919e-01, 5.75165409253650e-01, 5.74954608871394e-01, 5.74647753696020e-01, 5.74218122412258e-01, 5.73636124095714e-01, 5.72869875293105e-01, 5.71885940978824e-01, 5.70650225067691e-01, 5.69128987112009e-01, 5.67289952262787e-01, 5.65103472311321e-01, 5.62543687636818e-01, 5.59589634260135e-01, 5.56226237989268e-01, 5.52445139657390e-01, 5.48245302112672e-01, 5.43633360791532e-01, 5.38623694657446e-01, 5.33238211723699e-01, 5.27505861595760e-01, 5.21461904598517e-01, 5.15146981337892e-01, 5.08606036611194e-01, 5.01887156634286e-01, 4.95040378471096e-01, 4.88116525848337e-01, 4.81166117229126e-01, 4.74238381412391e-01, 4.67380404397019e-01, 4.60636420046697e-01, 4.54047247183723e-01, 4.47649867753803e-01, 4.41477134915992e-01, 4.35557596301422e-01, 4.29915416010520e-01, 4.24570378811197e-01, 4.19537961040724e-01, 4.14829454495685e-01, 4.10452131766149e-01, 4.06409443751873e-01, 4.02701242287247e-01, 3.99324022760705e-01, 3.96271183260598e-01, 3.93533298070626e-01, 3.91098404254915e-01, 3.88952300617277e-01, 3.87078858502175e-01, 3.85460343747983e-01, 3.84077748638219e-01, 3.82911131966360e-01, 3.81939964392583e-01, 3.81143475197430e-01, 3.80500995414895e-01, 3.79992291253881e-01, 3.79597880795576e-01, 3.79299326289537e-01, 3.79079494055404e-01, 3.78922774103351e-01, 3.78815252159054e-01, 3.78744827825635e-01, 3.78701274105659e-01, 3.78676235369028e-01, 3.78663162986932e-01, 3.78657190129260e-01, 3.78654949507641e-01, 3.78654340004562e-01, 3.78654250043201e-01, 3.78654243132562e-01, 3.78654119403244e-01, 3.78653281119252e-01, 3.78650199503593e-01, 3.78641985085846e-01, 3.78624008227263e-01, 3.78589581134719e-01, 3.78529714453936e-01, 3.78432960213114e-01, 3.78285351334441e-01, 3.78070446054632e-01, 3.77769483279085e-01, 3.77361652013146e-01, 3.76824474465376e-01, 3.76134298153271e-01, 3.75266887403829e-01, 3.74198099189407e-01, 3.72904622570537e-01, 3.71364755562848e-01, 3.69559188543356e-01, 3.67471759954945e-01, 3.65090148620300e-01, 3.62406467883201e-01, 3.59417730284402e-01, 3.56126157496060e-01, 3.52539318409892e-01, 3.48670087934835e-01, 3.44536429328912e-01, 3.40161012795271e-01, 3.35570691690322e-01, 3.30795864286426e-01, 3.25869753145774e-01, 3.20827635667644e-01, 3.15706058447400e-01, 3.10542065155250e-01, 3.05372463274139e-01, 3.00233149841223e-01, 2.95158510880878e-01, 2.90180903958593e-01, 2.85330228548755e-01, 2.80633584887166e-01, 2.76115018750943e-01, 2.71795347171246e-01, 2.67692058385158e-01, 2.63819278290335e-01, 2.60187795189090e-01, 2.56805134603226e-01, 2.53675676315343e-01, 2.50800806453167e-01, 2.48179098290688e-01, 2.45806516399753e-01, 2.43676639760004e-01, 2.41780900338213e-01, 2.40108834405991e-01, 2.38648344416429e-01, 2.37385969564205e-01, 2.36307163191118e-01, 2.35396574974324e-01, 2.34638335379479e-01, 2.34016339227857e-01, 2.33514524490307e-01, 2.33117141668775e-01, 2.32809008453692e-01, 2.32575743846553e-01, 2.32403975694680e-01, 2.32281515664884e-01, 2.32197496122446e-01, 2.32142464190037e-01, 2.32108429413241e-01, 2.32088862899580e-01, 2.32078647446077e-01, 2.32073979927840e-01, 2.32072228981718e-01, 2.32071752682025e-01, 2.32071682381193e-01, 2.32071676939622e-01, 2.32071578976210e-01, 2.32070915259893e-01, 2.32068475371577e-01, 2.32061971548656e-01, 2.32047738210706e-01, 2.32020480123449e-01, 2.31973079513256e-01, 2.31896471371172e-01, 2.31779594927502e-01, 2.31609427774025e-01, 2.31371107280712e-01, 2.31048141712300e-01, 2.30622710728912e-01, 2.30076051724968e-01, 2.29388924754986e-01, 2.28542144724997e-01, 2.27517165298650e-01, 2.26296694867621e-01, 2.24865321329474e-01, 2.23210119702017e-01, 2.21321215169478e-01, 2.19192274323458e-01, 2.16820899326045e-01, 2.14208903510112e-01, 2.11362452374413e-01, 2.08292060672331e-01, 2.05012443821996e-01, 2.01542229578134e-01, 1.97903543180718e-01, 1.94121485464496e-01, 1.90223528232256e-01, 1.86238854274997e-01, 1.82197670652080e-01, 1.78130523278487e-01, 1.74067638697270e-01, 1.70038315440841e-01, 1.66070382966865e-01, 1.62189741183020e-01, 1.58419988433575e-01, 1.54782140863329e-01, 1.51294441598189e-01, 1.47972254416580e-01, 1.44828033682103e-01, 1.41871360334598e-01, 1.39109032686363e-01, 1.36545200569600e-01, 1.34181531902751e-01, 1.32017401829561e-01, 1.30050096056436e-01, 1.28275021695160e-01, 1.26685920642291e-01, 1.25275082150942e-01, 1.24033552659540e-01, 1.22951342051660e-01, 1.22017626279835e-01, 1.21220946675340e-01, 1.20549406295357e-01, 1.19990863367079e-01, 1.19533121336339e-01, 1.19164114297099e-01, 1.18872085760369e-01, 1.18645757917658e-01, 1.18474487862297e-01, 1.18348406740001e-01, 1.18258537578935e-01, 1.18196887643805e-01, 1.18156511584971e-01, 1.18131542397393e-01, 1.18117188220122e-01, 1.18109694227638e-01, 1.18106270202063e-01, 1.18104985738569e-01, 1.18104636333989e-01, 1.18104584762623e-01, 1.18104580738429e-01, 1.18104507873628e-01, 1.18104014204327e-01, 1.18102199430300e-01, 1.18097361959236e-01, 1.18086775526898e-01, 1.18066502218827e-01, 1.18031249821563e-01, 1.17974280781098e-01, 1.17887379269163e-01, 1.17760881880571e-01, 1.17583776211632e-01, 1.17343869901692e-01, 1.17028030572182e-01, 1.16622494425453e-01, 1.16113238083940e-01, 1.15486404651989e-01, 1.14728771151650e-01, 1.13828240689855e-01, 1.12774339302604e-01, 1.11558694778420e-01, 1.10175473277291e-01, 1.08621749566360e-01, 1.06897788418824e-01, 1.05007218240942e-01, 1.02957083190866e-01, 1.00757766624021e-01, 9.84227861600928e-02, 9.59684684017219e-02, 9.34135186711593e-02, 9.07785074090235e-02, 8.80852995398236e-02, 8.53564557568305e-02, 8.26146351308523e-02, 7.98820267489901e-02, 7.71798345009920e-02, 7.45278340866875e-02, 7.19440153641102e-02, 6.94443168823007e-02, 6.70424534023546e-02, 6.47498318769379e-02, 6.25755470737737e-02, 6.05264449891436e-02, 5.86072404514049e-02, 5.68206747846176e-02, 5.51676999129143e-02, 5.36476766035463e-02, 5.22585764137276e-02, 5.09971790744643e-02, 4.98592592958076e-02, 4.88397591354826e-02, 4.79329440074920e-02, 4.71325420333919e-02, 4.64318677084279e-02, 4.58239317519293e-02, 4.53015395425389e-02, 4.48573807289918e-02, 4.44841124936559e-02, 4.41744385767073e-02, 4.39211855991992e-02, 4.37173775164226e-02, 4.35563082562958e-02, 4.34316118219442e-02, 4.33373284326670e-02, 4.32679647069634e-02, 4.32185455094327e-02, 4.31846549264208e-02, 4.31624639216359e-02, 4.31487425468404e-02, 4.31408551170058e-02, 4.31367374573703e-02, 4.31348561307843e-02, 4.31341503909337e-02, 4.31339584135912e-02, 4.31339300781772e-02, 4.31339278483023e-02, 4.31338872317042e-02, 4.31336120488517e-02, 4.31326004563637e-02, 4.31299039895912e-02, 4.31240031799973e-02, 4.31127037342058e-02, 4.30930580039899e-02, 4.30613162908493e-02, 4.30129122821583e-02, 4.29424865300057e-02, 4.28439512587673e-02, 4.27105988157521e-02, 4.25352546694846e-02, 4.23104739526146e-02, 4.20287781418018e-02, 4.16829256423483e-02, 4.12662069664940e-02, 4.07727521198602e-02, 4.01978350710159e-02, 3.95381581531602e-02, 3.87920983075816e-02, 3.79598975388546e-02, 3.70437820033978e-02, 3.60479978135047e-02, 3.49787567211237e-02, 3.38440909538704e-02, 3.26536230410177e-02, 3.14182628117372e-02, 3.01498491788053e-02, 2.88607582317475e-02, 2.75635011304125e-02, 2.62703351353153e-02, 2.49929089290752e-02, 2.37419595112556e-02, 2.25270729031664e-02, 2.13565152863389e-02, 2.02371356285482e-02, 1.91743358634433e-02, 1.81721006988589e-02, 1.72330763885320e-02, 1.63586863860594e-02, 1.55492716132457e-02, 1.48042438888645e-02, 1.41222425765335e-02, 1.35012864163855e-02, 1.29389145416638e-02, 1.24323126538570e-02, 1.19784221094150e-02, 1.15740311805084e-02, 1.12158489559491e-02, 1.09005632390868e-02, 1.06248843965524e-02, 1.03855774505774e-02, 1.01794848319017e-02, 1.00035421662774e-02, 9.85478929425909e-03, 9.73037845569635e-03, 9.62758123289120e-03, 9.54379546360684e-03, 9.47655292569038e-03, 9.42352818042735e-03, 9.38254856251319e-03, 9.35160494221421e-03, 9.32886258278742e-03, 9.31267119311840e-03, 9.30157314701319e-03, 9.29430881648585e-03, 9.28981804572260e-03, 9.28723696632971e-03, 9.28588960398713e-03, 9.28527402724200e-03, 9.28504311030309e-03, 9.28498029598046e-03, 9.28497102474424e-03, 9.28497028867881e-03, 9.28495679936958e-03, 9.28486540768958e-03, 9.28452944788205e-03, 9.28363394594811e-03, 9.28167438286209e-03, 9.27792243195569e-03, 9.27140040724694e-03, 9.26086619218225e-03, 9.24481043401926e-03, 9.22146774076893e-03, 9.18884344336490e-03, 9.14475712050233e-03, 9.08690346728136e-03, 9.01293018423033e-03, 8.92053136764304e-03, 8.80755344235240e-03, 8.67210910121585e-03, 8.51269316581948e-03, 8.32829297443638e-03, 8.11848507301960e-03, 7.88350985571995e-03, 7.62431653869200e-03, 7.34257251603765e-03, 7.04063366658403e-03, 6.72147533897061e-03, 6.38858719726798e-03, 6.04583844693352e-03, 5.69732275405244e-03, 5.34719404967632e-03, 4.99950510829095e-03, 4.65806017934867e-03, 4.32629108142990e-03, 4.00716329660611e-03, 3.70311519615996e-03, 3.41603019347048e-03, 3.14723892716628e-03, 2.89754684392207e-03, 2.66728172726260e-03, 2.45635547493706e-03, 2.26433442395191e-03, 2.09051265814608e-03, 1.93398317367155e-03, 1.79370273331614e-03, 1.66854768905722e-03, 1.55735966827564e-03, 1.45898135937728e-03, 1.37228341648310e-03, 1.29618375035534e-03, 1.22966041267247e-03, 1.17175915472600e-03, 1.12159666460698e-03, 1.07836045032172e-03, 1.04130628670873e-03, 1.00975405310777e-03, 9.83082665261037e-04, 9.60724672129487e-04, 9.42160963273153e-04, 9.26915918984861e-04, 9.14553230781934e-04, 9.04672522924699e-04, 8.96906817267507e-04, 8.90920805790785e-04, 8.86409829224277e-04, 8.83099407136712e-04, 8.80745126043081e-04, 8.79132669016514e-04, 8.78077764429719e-04, 8.77425843491153e-04, 8.77051225174341e-04, 8.76855690382428e-04, 8.76766360271120e-04, 8.76732851299522e-04, 8.76723736226381e-04, 8.76722390863449e-04, 8.76722283074116e-04, 8.76720295423183e-04, 8.76706828911722e-04, 8.76657326038810e-04, 8.76525380718363e-04, 8.76236678305276e-04, 8.75683995804621e-04, 8.74723555590151e-04, 8.73173051228804e-04, 8.70811687949669e-04, 8.67382588971566e-04, 8.62597897231865e-04, 8.56146831704763e-04, 8.47706822831640e-04, 8.36957641734415e-04, 8.23598152706495e-04, 8.07364974310332e-04, 7.88051964761394e-04, 7.65529108115004e-04, 7.39759139449347e-04, 7.10810181050903e-04, 6.78862819576409e-04, 6.44210439574522e-04, 6.07252181183071e-04, 5.68478497289249e-04, 5.28449839947598e-04, 4.87769493755508e-04, 4.47052128549785e-04, 4.06890478512167e-04, 3.67823706120873e-04, 3.30312040421887e-04, 2.94722215837595e-04, 2.61326066816281e-04, 2.30310296874592e-04, 2.01790675188107e-04, 1.75821853382274e-04, 1.52397149566497e-04, 1.31440456840437e-04, 1.12800478220265e-04, 9.62596519578085e-05, 8.15632063719818e-05, 6.84612328853973e-05, 5.67471103458672e-05, 4.62761293782421e-05, 3.69583772368287e-05, 2.87323028764774e-05, 2.15312630871267e-05, 1.52504938135678e-05, 9.69846196596791e-06, 4.26768448491543e-06, 3.68544634454793e-06, 5.45340039437292e-06, 7.09415786914747e-06, 7.99227124659296e-06, 8.48678527987090e-06, 8.74486907775918e-06, 8.86290321489014e-06, 8.89940188386618e-06, 8.89070263454528e-06, 8.85942769812686e-06, 8.81939059382830e-06, 8.77868892246170e-06, 8.74174482490275e-06, 8.71064118612708e-06, 8.68598605663351e-06, 8.66748544581242e-06, 8.65434037322643e-06, 8.64552849932920e-06, 8.63999795086556e-06, 8.63679031937181e-06, 8.63510744864016e-06, 8.63433666253854e-06, 8.63404721229699e-06, 8.63396844508202e-06, 8.63395681349743e-06, 8.63395592176630e-06, 8.63393995231831e-06, 8.63383171534338e-06, 8.63343364443381e-06, 8.63237104553605e-06, 8.63003801399032e-06, 8.62554121994208e-06, 8.61763198045651e-06, 8.60461235774324e-06, 8.58419837257429e-06, 8.55332887539362e-06, 8.50792688211754e-06, 8.44265420951729e-06, 8.35074193261676e-06, 8.22399930555662e-06, 8.05306451923116e-06, 7.82786745645581e-06, 7.53826304134009e-06, 7.17508584457960e-06, 6.73240257498264e-06, 6.21168250347637e-06, 5.62679768023600e-06, 5.00544612336284e-06, 4.38178464534644e-06, 3.78465733064104e-06, 3.24099161730373e-06, 2.79532489244616e-06, 2.48100772992643e-06, 2.24691776296027e-06, 2.02635803958628e-06, 1.80118945891650e-06, 1.53066881480868e-06, 1.16296257957060e-06, 7.76678053426849e-07, 6.84990720145606e-07, 5.45887662184441e-07, 4.83150326857640e-07, 5.24735405349475e-07, 4.49118754630717e-07, 5.00999295383053e-07, 4.87334805399087e-07, 3.55271367880050e-15, 3.89587064830721e-07, 5.03287353126325e-07, 3.56229200804137e-07, 6.30473746099369e-07, 1.13751652008887e-06, 1.43886457237841e-06, 1.50022458811350e-06, 1.33894801734868e-06, 1.01774918448427e-06, 6.16997379410122e-07, 3.55271367880050e-15, 3.55271367880050e-15, 3.57507516923761e-07, 3.34196204931914e-07, 3.55271367880050e-15, 3.55271367880050e-15, 7.10542735760100e-15, 0.00000000000000e+00, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 7.10542735760100e-15, 7.10542735760100e-15, 3.55271367880050e-15, 7.10542735760100e-15, 3.55271367880050e-15, 7.10542735760100e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15, 3.55271367880050e-15],
[ 2.90000000000000e+01, 2.89999999789295e+01, 2.89999993273880e+01, 2.89999949130916e+01, 2.89999786855689e+01, 2.89999354281654e+01, 2.89998407564082e+01, 2.89996594333865e+01, 2.89993440633070e+01, 2.89988342129964e+01, 2.89980559982089e+01, 2.89969221576056e+01, 2.89953326227013e+01, 2.89931755774485e+01, 2.89903289869945e+01, 2.89866625619832e+01, 2.89820401130160e+01, 2.89763222398361e+01, 2.89693692916302e+01, 2.89610445285789e+01, 2.89512174102945e+01, 2.89397669338340e+01, 2.89265849422679e+01, 2.89115793240344e+01, 2.88946770233242e+01, 2.88758267824979e+01, 2.88550015391927e+01, 2.88322004037143e+01, 2.88074501470622e+01, 2.87808061370967e+01, 2.87523526704668e+01, 2.87222026613227e+01, 2.86904966645592e+01, 2.86574012310086e+01, 2.86231066138433e+01, 2.85878238683180e+01, 2.85517814094933e+01, 2.85152211132563e+01, 2.84783940634203e+01, 2.84415560607933e+01, 2.84049630180979e+01, 2.83688663671885e+01, 2.83335086022360e+01, 2.82991190749684e+01, 2.82659101464324e+01, 2.82340737850682e+01, 2.82037786841742e+01, 2.81751679540576e+01, 2.81483574261723e+01, 2.81234345889914e+01, 2.81004581587446e+01, 2.80794582727627e+01, 2.80604372791704e+01, 2.80433710841111e+01, 2.80282110065532e+01, 2.80148860809997e+01, 2.80033057400744e+01, 2.79933628020132e+01, 2.79849366826080e+01, 2.79778967472360e+01, 2.79721057163920e+01, 2.79674230377668e+01, 2.79637081395202e+01, 2.79608234830619e+01, 2.79586373394309e+01, 2.79570262212014e+01, 2.79558769116287e+01, 2.79550880442628e+01, 2.79545711992147e+01, 2.79542514962950e+01, 2.79540676799402e+01, 2.79539717057475e+01, 2.79539278530927e+01, 2.79539114022451e+01, 2.79539069271965e+01, 2.79539062666880e+01, 2.79539061800432e+01, 2.79539041613235e+01, 2.79538904841851e+01, 2.79538402053540e+01, 2.79537061787309e+01, 2.79534128560202e+01, 2.79528510745114e+01, 2.79518740271668e+01, 2.79502945781765e+01, 2.79478840507621e+01, 2.79443725741046e+01, 2.79394510337881e+01, 2.79327746261533e+01, 2.79239679725617e+01, 2.79126317059311e+01, 2.78983503999942e+01, 2.78807016722992e+01, 2.78592662553118e+01, 2.78336387959903e+01, 2.78034391124556e+01, 2.77683236064626e+01, 2.77279965023967e+01, 2.76822205586546e+01, 2.76308268781864e+01, 2.75737234360109e+01, 2.75109019481573e+01, 2.74424427345228e+01, 2.73685172821126e+01, 2.72893882968760e+01, 2.72054071396337e+01, 2.71170086676434e+01, 2.70247036372963e+01, 2.69290689518781e+01, 2.68307361475417e+01, 2.67303785891195e+01, 2.66286978878748e+01, 2.65264100539519e+01, 2.64242318607590e+01, 2.63228678348623e+01, 2.62229982037968e+01, 2.61252680466544e+01, 2.60302778082057e+01, 2.59385752638393e+01, 2.58506489638607e+01, 2.57669231427313e+01, 2.56877540503789e+01, 2.56134276459795e+01, 2.55441585861821e+01, 2.54800904362767e+01, 2.54212970314640e+01, 2.53677849140745e+01, 2.53194967699632e+01, 2.52763157826524e+01, 2.52380708168993e+01, 2.52045423343590e+01, 2.51754689332630e+01, 2.51505543920911e+01, 2.51294750847631e+01, 2.51118876227186e+01, 2.50974365683033e+01, 2.50857620551143e+01, 2.50765071453962e+01, 2.50693247531802e+01, 2.50638839654838e+01, 2.50598756031712e+01, 2.50570168783672e+01, 2.50550550266226e+01, 2.50537698189426e+01, 2.50529748905357e+01, 2.50525178585610e+01, 2.50522792388263e+01, 2.50521702097049e+01, 2.50521293088271e+01, 2.50521181827692e+01, 2.50521165405854e+01, 2.50521163956118e+01, 2.50521135549259e+01, 2.50520943088522e+01, 2.50520235581251e+01, 2.50518349617076e+01, 2.50514222182145e+01, 2.50506317464750e+01, 2.50492570513593e+01, 2.50470350224782e+01, 2.50436443689814e+01, 2.50387063431064e+01, 2.50317878478333e+01, 2.50224069593854e+01, 2.50100408230980e+01, 2.49941358016594e+01, 2.49741196691726e+01, 2.49494155553526e+01, 2.49194572553021e+01, 2.48837054368804e+01, 2.48416642060159e+01, 2.47928974373649e+01, 2.47370442503717e+01, 2.46738330148664e+01, 2.46030933096029e+01, 2.45247653321900e+01, 2.44389063664423e+01, 2.43456940459139e+01, 2.42454262993857e+01, 2.41385180123376e+01, 2.40254945749719e+01, 2.39069826015635e+01, 2.37836981917003e+01, 2.36564331606714e+01, 2.35260396981372e+01, 2.33934139284863e+01, 2.32594788501015e+01, 2.31251671284476e+01, 2.29914042092216e+01, 2.28590921984379e+01, 2.27290949200547e+01, 2.26022245040082e+01, 2.24792297783006e+01, 2.23607866440059e+01, 2.22474905123361e+01, 2.21398507906721e+01, 2.20382873302719e+01, 2.19431286982540e+01, 2.18546121110153e+01, 2.17728848614439e+01, 2.16980070814712e+01, 2.16299556974818e+01, 2.15686294527316e+01, 2.15138548839845e+01, 2.14653931467549e+01, 2.14229475841737e+01, 2.13861719290032e+01, 2.13546790178165e+01, 2.13280498822708e+01, 2.13058430663735e+01, 2.12876040023440e+01, 2.12728742628524e+01, 2.12612004957756e+01, 2.12521428407809e+01, 2.12452826264735e+01, 2.12402291536867e+01, 2.12366253854973e+01, 2.12341523879386e+01, 2.12325323968181e+01, 2.12315304245695e+01, 2.12309543651895e+01, 2.12306536030734e+01, 2.12305161807333e+01, 2.12304646286030e+01, 2.12304506051485e+01, 2.12304485353150e+01, 2.12304483692888e+01, 2.12304453052623e+01, 2.12304245460254e+01, 2.12303482327750e+01, 2.12301448089051e+01, 2.12296996168695e+01, 2.12288470060240e+01, 2.12273642621695e+01, 2.12249676291852e+01, 2.12213106465536e+01, 2.12159849731621e+01, 2.12085238066436e+01, 2.11984079377399e+01, 2.11850744001451e+01, 2.11679275881253e+01, 2.11463526182117e+01, 2.11197306101565e+01, 2.10874554605652e+01, 2.10489515862543e+01, 2.10036920309326e+01, 2.09512162666158e+01, 2.08911469886202e+01, 2.08232052073215e+01, 2.07472229860112e+01, 2.06631532633978e+01, 2.05710763281148e+01, 2.04712026723489e+01, 2.03638721289326e+01, 2.02495493743548e+01, 2.01288160420933e+01, 2.00023598223844e+01, 1.98709610177909e+01, 1.97354770778794e+01, 1.95968256567223e+01, 1.94559667334080e+01, 1.93138843179291e+01, 1.91715682392741e+01, 1.90299964813114e+01, 1.88901184933905e+01, 1.87528398532638e+01, 1.86190085977419e+01, 1.84894034620763e+01, 1.83647241862958e+01, 1.82455839619434e+01, 1.81325040130616e+01, 1.80259102372613e+01, 1.79261317806611e+01, 1.78334013861625e+01, 1.77478573371776e+01, 1.76695468158064e+01, 1.75984305016355e+01, 1.75343882504530e+01, 1.74772257071513e+01, 1.74266817205183e+01, 1.73824364370419e+01, 1.73441199548358e+01, 1.73113214168590e+01, 1.72835984151496e+01, 1.72604865659112e+01, 1.72415091006476e+01, 1.72261863031405e+01, 1.72140446081679e+01, 1.72046251676917e+01, 1.71974916858982e+01, 1.71922373276796e+01, 1.71884905171023e+01, 1.71859194636728e+01, 1.71842352846095e+01, 1.71831936299314e+01, 1.71825947623840e+01, 1.71822820938840e+01, 1.71821392317693e+01, 1.71820856390485e+01, 1.71820710605092e+01, 1.71820689087466e+01, 1.71820687427820e+01, 1.71820657625897e+01, 1.71820455713436e+01, 1.71819713461155e+01, 1.71817734883573e+01, 1.71813404786939e+01, 1.71805112026612e+01, 1.71790690511210e+01, 1.71767380607311e+01, 1.71731813155122e+01, 1.71680017795414e+01, 1.71607456721117e+01, 1.71509084292731e+01, 1.71379432188137e+01, 1.71212718894994e+01, 1.71002981409478e+01, 1.70744226005001e+01, 1.70430593920817e+01, 1.70056536850791e+01, 1.69616996257358e+01, 1.69107579872721e+01, 1.68524728356529e+01, 1.67865865026074e+01, 1.67129521912696e+01, 1.66315436151364e+01, 1.65424611870313e+01, 1.64459344267419e+01, 1.63423204354769e+01, 1.62320984803135e+01, 1.61158609278698e+01, 1.59943009477404e+01, 1.58681975574731e+01, 1.57383986891385e+01, 1.56058030142183e+01, 1.54713412657036e+01, 1.53359577472947e+01, 1.52005926286710e+01, 1.50661655064754e+01, 1.49335605785756e+01, 1.48036136494281e+01, 1.46771010693426e+01, 1.45547306181803e+01, 1.44371342778033e+01, 1.43248627964086e+01, 1.42183819275984e+01, 1.41180702219600e+01, 1.40242182531557e+01, 1.39370291690485e+01, 1.38566204675723e+01, 1.37830269047198e+01, 1.37162044471455e+01, 1.36560351841921e+01, 1.36023331137132e+01, 1.35548507130044e+01, 1.35132862005427e+01, 1.34772913860726e+01, 1.34464799959745e+01, 1.34204363480672e+01, 1.33987242356412e+01, 1.33808958655374e+01, 1.33665006807938e+01, 1.33550938863356e+01, 1.33462444880956e+01, 1.33395426534600e+01, 1.33346062054628e+01, 1.33310860757062e+01, 1.33286705620261e+01, 1.33270882662696e+01, 1.33261096243518e+01, 1.33255469835604e+01, 1.33252532288873e+01, 1.33251190087010e+01, 1.33250686578731e+01, 1.33250549612080e+01, 1.33250529396082e+01, 1.33250527865388e+01, 1.33250500749288e+01, 1.33250317033672e+01, 1.33249641674848e+01, 1.33247841410164e+01, 1.33243901544390e+01, 1.33236356113039e+01, 1.33223234178245e+01, 1.33202024659213e+01, 1.33169661690593e+01, 1.33122532032690e+01, 1.33056505518866e+01, 1.32966988908613e+01, 1.32849002814110e+01, 1.32697280584487e+01, 1.32506387177654e+01, 1.32270855149868e+01, 1.31985333989321e+01, 1.31644748167665e+01, 1.31244458549863e+01, 1.30780421260816e+01, 1.30249337825977e+01, 1.29648790438193e+01, 1.28977356585513e+01, 1.28234698003819e+01, 1.27421619956921e+01, 1.26540098124696e+01, 1.25593271801733e+01, 1.24585403568896e+01, 1.23521806998702e+01, 1.22408745213505e+01, 1.21253304183586e+01, 1.20063245508829e+01, 1.18846844069116e+01, 1.17612716353434e+01, 1.16369645473018e+01, 1.15126408800265e+01, 1.13891613816735e+01, 1.12673547079026e+01, 1.11480040238994e+01, 1.10318355858211e+01, 1.09195094460945e+01, 1.08116123027886e+01, 1.07086524087222e+01, 1.06110563809364e+01, 1.05191677091285e+01, 1.04332467498204e+01, 1.03534720041725e+01, 1.02799425023433e+01, 1.02126811475438e+01, 1.01516389018112e+01, 1.00966997189144e+01, 1.00476861457863e+01, 1.00043655221726e+01, 9.96645670941196e+00, 9.93363727449012e+00, 9.90555104593948e+00, 9.88181594502447e+00, 9.86203198024590e+00, 9.84578927689173e+00, 9.83267599765176e+00, 9.82228599679830e+00, 9.81422604074426e+00, 9.80812242346833e+00, 9.80362680766070e+00, 9.80042113243028e+00, 9.79822144646084e+00, 9.79678055145001e+00, 9.79588937373300e+00, 9.79537702081803e+00, 9.79510952229643e+00, 9.79498729907031e+00, 9.79494144876613e+00, 9.79492897635730e+00, 9.79492713545536e+00, 9.79492699725827e+00, 9.79492456483264e+00, 9.79490808478751e+00, 9.79484750234331e+00, 9.79468601136758e+00, 9.79433259017390e+00, 9.79365573809423e+00, 9.79247866094157e+00, 9.79057612229016e+00, 9.78767314139970e+00, 9.78344567699105e+00, 9.77752338803501e+00, 9.76949450729760e+00, 9.75891279993192e+00, 9.74530650791253e+00, 9.72818910257866e+00, 9.70707158430430e+00, 9.68147598414172e+00, 9.65094964244026e+00, 9.61507977036721e+00, 9.57350774904601e+00, 9.52594259463193e+00, 9.47217302187868e+00, 9.41207757721179e+00, 9.34563238543168e+00, 9.27291615850999e+00, 9.19411224319569e+00, 9.10950762573517e+00, 9.01948895451463e+00, 8.92453577258642e+00, 8.82521126185276e+00, 8.72215088311661e+00, 8.61604934988177e+00, 8.50764640138340e+00, 8.39771184699713e+00, 8.28703034540778e+00, 8.17638636179461e+00, 8.06654971662162e+00, 7.95826210024748e+00, 7.85222487749810e+00, 7.74908844517273e+00, 7.64944333446819e+00, 7.55381317287641e+00, 7.46264954132843e+00, 7.37632868811999e+00, 7.29514999704765e+00, 7.21933605750925e+00, 7.14903415146322e+00, 7.08431895638104e+00, 7.02519626304072e+00, 6.97160751913140e+00, 6.92343503025122e+00, 6.88050767480281e+00, 6.84260701459207e+00, 6.80947370530957e+00, 6.78081412808416e+00, 6.75630717349437e+00, 6.73561111234749e+00, 6.71837048366096e+00, 6.70422292086614e+00, 6.69280582417848e+00, 6.68376277261625e+00, 6.67674955574648e+00, 6.67143969526284e+00, 6.66752932202341e+00, 6.66474127676181e+00, 6.66282831324336e+00, 6.66157530133849e+00, 6.66080035374458e+00, 6.66035483263590e+00, 6.66012222954045e+00, 6.66001595105367e+00, 6.65997608228995e+00, 6.65996523701632e+00, 6.65996363627654e+00, 6.65996351647392e+00, 6.65996141267972e+00, 6.65994715916519e+00, 6.65989476174816e+00, 6.65975508954995e+00, 6.65944942075601e+00, 6.65886402996609e+00, 6.65784603523467e+00, 6.65620069796159e+00, 6.65369033980941e+00, 6.65003500694487e+00, 6.64491497196869e+00, 6.63797511671988e+00, 6.62883118383255e+00, 6.61707782124796e+00, 6.60229827257999e+00, 6.58407548924297e+00, 6.56200436084223e+00, 6.53570468303325e+00, 6.50483441244401e+00, 6.46910270251031e+00, 6.42828217843460e+00, 6.38221989956851e+00, 6.33084647766211e+00, 6.27418287201506e+00, 6.21234446757543e+00, 6.14554215671463e+00, 6.07408028423205e+00, 5.99835147000574e+00, 5.91882848439224e+00, 5.83605350640842e+00, 5.75062523193701e+00, 5.66318440752249e+00, 5.57439843566653e+00, 5.48494572403202e+00, 5.39550043188478e+00, 5.30671820520250e+00, 5.21922339417335e+00, 5.13359812366992e+00, 5.05037345107954e+00, 4.97002270931077e+00, 4.89295700735918e+00, 4.81952275549686e+00, 4.75000100270450e+00, 4.68460832272073e+00, 4.62349896125454e+00, 4.56676795726135e+00, 4.51445497088122e+00, 4.46654858408550e+00, 4.42299088164946e+00, 4.38368216464796e+00, 4.34848569198354e+00, 4.31723238421644e+00, 4.28972545588611e+00, 4.26574496623387e+00, 4.24505229319614e+00, 4.22739454186476e+00, 4.21250889699599e+00, 4.20012692075361e+00, 4.18997878323918e+00, 4.18179739632799e+00, 4.17532240292506e+00, 4.17030395608556e+00, 4.16650620754624e+00, 4.16371041490474e+00, 4.16171757239755e+00, 4.16035047288164e+00, 4.15945511852215e+00, 4.15890141447131e+00, 4.15858310251160e+00, 4.15841691872073e+00, 4.15834098882783e+00, 4.15831250504172e+00, 4.15830475677431e+00, 4.15830361314675e+00, 4.15830352752331e+00, 4.15830202350472e+00, 4.15829183356456e+00, 4.15825437437153e+00, 4.15815452272217e+00, 4.15793600392033e+00, 4.15751752994858e+00, 4.15678984925932e+00, 4.15561385554984e+00, 4.15381988573006e+00, 4.15120831612311e+00, 4.14755153904191e+00, 4.14259736745651e+00, 4.13607387199983e+00, 4.12769560126422e+00, 4.11717107356289e+00, 4.10421135788558e+00, 4.08853948715536e+00, 4.06990037324597e+00, 4.04807082710022e+00, 4.02286923611288e+00, 3.99416442218626e+00, 3.96188320409083e+00, 3.92601622155467e+00, 3.88662164750624e+00, 3.84382651710375e+00, 3.79782553166704e+00, 3.74887734279136e+00, 3.69729847436081e+00, 3.64345518405139e+00, 3.58775368766734e+00, 3.53062925783024e+00, 3.47253475535400e+00, 3.41392915403385e+00, 3.35526657956279e+00, 3.29698630747596e+00, 3.23950406348900e+00, 3.18320485427637e+00, 3.12843743969955e+00, 3.07551044938624e+00, 3.02469005547458e+00, 2.97619904424085e+00, 2.93021708404925e+00, 2.88688196469770e+00, 2.84629158086517e+00, 2.80850644585430e+00, 2.77355254656197e+00, 2.74142438212931e+00, 2.71208806309290e+00, 2.68548438190947e+00, 2.66153179710262e+00, 2.64012930038747e+00, 2.62115915802107e+00, 2.60448953386630e+00, 2.58997701219378e+00, 2.57746904331145e+00, 2.56680633515076e+00, 2.55782520957129e+00, 2.55035993414356e+00, 2.54424502944835e+00, 2.53931753953269e+00, 2.53541924023435e+00, 2.53239874782768e+00, 2.53011348004130e+00, 2.52843141403471e+00, 2.52723258227933e+00, 2.52641024805608e+00, 2.52587170767625e+00, 2.52553867636345e+00, 2.52534722840994e+00, 2.52524727880028e+00, 2.52520161179167e+00, 2.52518448064853e+00, 2.52517982057434e+00, 2.52517913275759e+00, 2.52517908112411e+00, 2.52517817233169e+00, 2.52517201513539e+00, 2.52514938072224e+00, 2.52508904639164e+00, 2.52495700976040e+00, 2.52470415751674e+00, 2.52426448821563e+00, 2.52355398083377e+00, 2.52247019144225e+00, 2.52089264838210e+00, 2.51868410042038e+00, 2.51569265111785e+00, 2.51175478485563e+00, 2.50669925497692e+00, 2.50035176248762e+00, 2.49254030605219e+00, 2.48310103325230e+00, 2.47188437318665e+00, 2.45876118650447e+00, 2.44362863650892e+00, 2.42641546957170e+00, 2.40708639933961e+00, 2.38564531982531e+00, 2.36213712761774e+00, 2.33664801027414e+00, 2.30930415064342e+00, 2.28026889715161e+00, 2.24973854826105e+00, 2.21793698559776e+00, 2.18510945613260e+00, 2.15151584333677e+00, 2.11742377776187e+00, 2.08310191991750e+00, 2.04881370671120e+00, 2.01481179348953e+00, 1.98133335449988e+00, 1.94859633305559e+00, 1.91679666556028e+00, 1.88610644593887e+00, 1.85667295210364e+00, 1.82861842512418e+00, 1.80204047441740e+00, 1.77701297698641e+00, 1.75358734324445e+00, 1.73179403369525e+00, 1.71164422717074e+00, 1.69313156018542e+00, 1.67623387637546e+00, 1.66091494350965e+00, 1.64712611216039e+00, 1.63480790414234e+00, 1.62389152989734e+00, 1.61430034198363e+00, 1.60595123674330e+00, 1.59875601822763e+00, 1.59262273780451e+00, 1.58745701989296e+00, 1.58316337937705e+00, 1.57964652994121e+00, 1.57681267540161e+00, 1.57457076871310e+00, 1.57283371636519e+00, 1.57151950000388e+00, 1.57055218293170e+00, 1.56986276714146e+00, 1.56938986706216e+00, 1.56908016936209e+00, 1.56888865385733e+00, 1.56877855847707e+00, 1.56872108080224e+00, 1.56869481924265e+00, 1.56868496770020e+00, 1.56868228784888e+00, 1.56868189230870e+00, 1.56868186248957e+00, 1.56868133596341e+00, 1.56867776867488e+00, 1.56866465499982e+00, 1.56862969917864e+00, 1.56855320135956e+00, 1.56840670708668e+00, 1.56815197798183e+00, 1.56774033692383e+00, 1.56711243566334e+00, 1.56619848591407e+00, 1.56491898574166e+00, 1.56318596069077e+00, 1.56090472282957e+00, 1.55797613034643e+00, 1.55429930561849e+00, 1.54977474158670e+00, 1.54430769640999e+00, 1.53781174711493e+00, 1.53021234733856e+00, 1.52145021565782e+00, 1.51148437269717e+00, 1.50029464986662e+00, 1.48788351170793e+00, 1.47427706731952e+00, 1.45952519227582e+00, 1.44370073718113e+00, 1.42689785745635e+00, 1.40922955538230e+00, 1.39082457414127e+00, 1.37182381985334e+00, 1.35237650822822e+00, 1.33263623626386e+00, 1.31275716727244e+00, 1.29289049201706e+00, 1.27318129374683e+00, 1.25376590485482e+00, 1.23476980211415e+00, 1.21630604972530e+00, 1.19847426752063e+00, 1.18136007728934e+00, 1.16503496389037e+00, 1.14955647927888e+00, 1.13496871578515e+00, 1.12130297856243e+00, 1.10857859452736e+00, 1.09680380487235e+00, 1.08597669903856e+00, 1.07608615886159e+00, 1.06711279167152e+00, 1.05902983993041e+00, 1.05180406222169e+00, 1.04539658592916e+00, 1.03976373573969e+00, 1.03485784423656e+00, 1.03062805143844e+00, 1.02702109934328e+00, 1.02398212555201e+00, 1.02145545710662e+00, 1.01938540204682e+00, 1.01771703217635e+00, 1.01639694646880e+00, 1.01537400078899e+00, 1.01459998651320e+00, 1.01403023852428e+00, 1.01362415220110e+00, 1.01334558959066e+00, 1.01316315700788e+00, 1.01305033978352e+00, 1.01298548458270e+00, 1.01295162533838e+00, 1.01293615500181e+00, 1.01293035158197e+00, 1.01292877291461e+00, 1.01292853990674e+00, 1.01292852224506e+00, 1.01292820911904e+00, 1.01292608764611e+00, 1.01291828891632e+00, 1.01289750058974e+00, 1.01285200698059e+00, 1.01276488562606e+00, 1.01261339446848e+00, 1.01236858101327e+00, 1.01199514150093e+00, 1.01145155390984e+00, 1.01069050304548e+00, 1.00965960860033e+00, 1.00830245748593e+00, 1.00655992972373e+00, 1.00437179280877e+00, 1.00167852320148e+00, 9.98423296382935e-01, 9.94554070110272e-01, 9.90025670882041e-01, 9.84801783115309e-01, 9.78856736029126e-01, 9.72176986231005e-01, 9.64762205333471e-01, 9.56625901469039e-01, 9.47795530146287e-01, 9.38312081310638e-01, 9.28229162782682e-01, 9.17611632112418e-01, 9.06533856084909e-01, 8.95077697006428e-01, 8.83330335821416e-01, 8.71382043559951e-01, 8.59324005258586e-01, 8.47246285968318e-01, 8.35236009027280e-01, 8.23375794939462e-01, 8.11742487337732e-01, 8.00406172578352e-01, 7.89429482886582e-01, 7.78867160393744e-01, 7.68765851046485e-01, 7.59164092948794e-01, 7.50092462636680e-01, 7.41573844348711e-01, 7.33623790773759e-01, 7.26250948314476e-01, 7.19457524989743e-01, 7.13239784225941e-01, 7.07588552594569e-01, 7.02489733794000e-01, 6.97924824688144e-01, 6.93871431918332e-01, 6.90303789463684e-01, 6.87193278547621e-01, 6.84508951512171e-01, 6.82218060775831e-01, 6.80286592847946e-01, 6.78679805714459e-01, 6.77362765885320e-01, 6.76300879177450e-01, 6.75460407093428e-01, 6.74808958653035e-01, 6.74315945946834e-01, 6.73952990694431e-01, 6.73694268855638e-01, 6.73516780955886e-01, 6.73400537275718e-01, 6.73328649371985e-01, 6.73287322425132e-01, 6.73265746457709e-01, 6.73255888316625e-01, 6.73252190203200e-01, 6.73251184228160e-01, 6.73251035748404e-01, 6.73251024422434e-01, 6.73250822680661e-01, 6.73249455851359e-01, 6.73244431256272e-01, 6.73231037637947e-01, 6.73201726576806e-01, 6.73145594476267e-01, 6.73047986815671e-01, 6.72890244422884e-01, 6.72649609202495e-01, 6.72299303884795e-01, 6.71808796695665e-01, 6.71144257128489e-01, 6.70269203011905e-01, 6.69145331697141e-01, 6.67733519493307e-01, 6.65994963739134e-01, 6.63892431657334e-01, 6.61391570188890e-01, 6.58462222356782e-01, 6.55079689475386e-01, 6.51225875776476e-01, 6.46890253624772e-01, 6.42070593913630e-01, 6.36773417398896e-01, 6.31014137990668e-01, 6.24816887150789e-01, 6.18214027885497e-01, 6.11245385512852e-01, 6.03957238615013e-01, 5.96401125865647e-01, 5.88632531782103e-01, 5.80709516548115e-01, 5.72691352168086e-01, 5.64637220157792e-01, 5.56605015906740e-01, 5.48650293079419e-01, 5.40825369234000e-01, 5.33178602319872e-01, 5.25753837689006e-01, 5.18590017226870e-01, 5.11720936387366e-01, 5.05175131268523e-01, 4.98975876172953e-01, 4.93141272029511e-01, 4.87684407227288e-01, 4.82613574445342e-01, 4.77932529594181e-01, 4.73640781715936e-01, 4.69733905371960e-01, 4.66203869496706e-01, 4.63039378782781e-01, 4.60226225301291e-01, 4.57747649207494e-01, 4.55584708021089e-01, 4.53716654114729e-01, 4.52121319730502e-01, 4.50775508128810e-01, 4.49655388437066e-01, 4.48736890504016e-01, 4.47996094696300e-01, 4.47409610224565e-01, 4.46954934394117e-01, 4.46610784270881e-01, 4.46357391757516e-01, 4.46176753080195e-01, 4.46052824251243e-01, 4.45971655209959e-01, 4.45921457017935e-01, 4.45892598613717e-01, 4.45877532091718e-01, 4.45870648111217e-01, 4.45868065698590e-01, 4.45867363220323e-01, 4.45867259536023e-01, 4.45867251571215e-01, 4.45867108967921e-01, 4.45866142810072e-01, 4.45862591118718e-01, 4.45853123668066e-01, 4.45832404660429e-01, 4.45792726341729e-01, 4.45723728727792e-01, 4.45612219138596e-01, 4.45442103499321e-01, 4.45194439251036e-01, 4.44847617077158e-01, 4.44377675337776e-01, 4.43758746985900e-01, 4.42963633754069e-01, 4.41964496590639e-01, 4.40733644876143e-01, 4.39244400207311e-01, 4.37472004004352e-01, 4.35394532511388e-01, 4.32993778621832e-01, 4.30256058052461e-01, 4.27172898259705e-01, 4.23741572452403e-01, 4.19965448090245e-01, 4.15854128981270e-01, 4.11423381761974e-01, 4.06694850137175e-01, 4.01695572601714e-01, 3.96457330302677e-01, 3.91015860244384e-01, 3.85409974512164e-01, 3.79680628304360e-01, 3.73869978421709e-01, 3.68020469923174e-01, 3.62173982618678e-01, 3.56371061757891e-01, 3.50650249516903e-01, 3.45047526408582e-01, 3.39595865103711e-01, 3.34324893708605e-01, 3.29260661462882e-01, 3.24425497097284e-01, 3.19837948600725e-01, 3.15512792689464e-01, 3.11461102616168e-01, 3.07690363870417e-01, 3.04204628593137e-01, 3.01004700976286e-01, 2.98088347400590e-01, 2.95450526466290e-01, 2.93083635308751e-01, 2.90977769600527e-01, 2.89120995379900e-01, 2.87499631287382e-01, 2.86098539926517e-01, 2.84901426901659e-01, 2.83891145650326e-01, 2.83050005526551e-01, 2.82360079768523e-01, 2.81803509079982e-01, 2.81362795661884e-01, 2.81021081749323e-01, 2.80762406134460e-01, 2.80571931877457e-01, 2.80436138490934e-01, 2.80342972370693e-01, 2.80281950143277e-01, 2.80244210882987e-01, 2.80222514754993e-01, 2.80211187478656e-01, 2.80206011965731e-01, 2.80204070454285e-01, 2.80203542316301e-01, 2.80203464364249e-01, 2.80203458330433e-01, 2.80203349705140e-01, 2.80202613753040e-01, 2.80199908319286e-01, 2.80192696656972e-01, 2.80176914266583e-01, 2.80146689699379e-01, 2.80094130795959e-01, 2.80009186661385e-01, 2.79879594303914e-01, 2.79690917210910e-01, 2.79426681107640e-01, 2.79068609637537e-01, 2.78596959634079e-01, 2.77990951985288e-01, 2.77229289858372e-01, 2.76290751386586e-01, 2.75154839064847e-01, 2.73802463406884e-01, 2.72216634323001e-01, 2.70383130666936e-01, 2.68291116942951e-01, 2.65933676636806e-01, 2.63308234256868e-01, 2.60416842940771e-01, 2.57266321147505e-01, 2.53868230024452e-01, 2.50238691834141e-01, 2.46398058567568e-01, 2.42370447786136e-01, 2.38183169160305e-01, 2.33866069644556e-01, 2.29450827524037e-01, 2.24970225714820e-01, 2.20457432937945e-01, 2.15945318107497e-01, 2.11465818941111e-01, 2.07049380889373e-01, 2.02724477414648e-01, 1.98517217778043e-01, 1.94451044076281e-01, 1.90546515483703e-01, 1.86821174598077e-01, 1.83289488503089e-01, 1.79962855634770e-01, 1.76849668730068e-01, 1.73955423970177e-01, 1.71282866815790e-01, 1.68832165854788e-01, 1.66601107120847e-01, 1.64585302663653e-01, 1.62778408528677e-01, 1.61172348618550e-01, 1.59757542056592e-01, 1.58523132577312e-01, 1.57457219074239e-01, 1.56547086716969e-01, 1.55779438008910e-01, 1.55140622825837e-01, 1.54616865909215e-01, 1.54194489564503e-01, 1.53860128525270e-01, 1.53600933189260e-01, 1.53404756810016e-01, 1.53260321826643e-01, 1.53157360403586e-01, 1.53086724476950e-01, 1.53040461176591e-01, 1.53011850396329e-01, 1.52995402468854e-01, 1.52986815293197e-01, 1.52982891770211e-01, 1.52981419925890e-01, 1.52981019548989e-01, 1.52980960454215e-01, 1.52980955842949e-01, 1.52980872348365e-01, 1.52980306660634e-01, 1.52978227138043e-01, 1.52972683935452e-01, 1.52960552980939e-01, 1.52937321551804e-01, 1.52896924405333e-01, 1.52831638600407e-01, 1.52732044176251e-01, 1.52587056664519e-01, 1.52384035908849e-01, 1.52108973756917e-01, 1.51746760794602e-01, 1.51281529370259e-01, 1.50697066731535e-01, 1.49977288270268e-01, 1.49106756835401e-01, 1.48071230112297e-01, 1.46858214533388e-01, 1.45457501475594e-01, 1.43861660004323e-01, 1.42066460472613e-01, 1.40071205091381e-01, 1.37878945207060e-01, 1.35496570319464e-01, 1.32934760519667e-01, 1.30207801534979e-01, 1.27333269334827e-01, 1.24331598635710e-01, 1.21225556036375e-01, 1.18039643408245e-01, 1.14799460205063e-01, 1.11531054364733e-01, 1.08260290470277e-01, 1.05012261007786e-01, 1.01810762249062e-01, 9.86778509419288e-02, 9.56334921120892e-02, 9.26953023637580e-02, 8.98783875577429e-02, 8.71952690049724e-02, 8.46558885938009e-02, 8.22676807108209e-02, 8.00356974407279e-02, 7.79627732618664e-02, 7.60497161245688e-02, 7.42955131916041e-02, 7.26975413947208e-02, 7.12517750925556e-02, 6.99529853002296e-02, 6.87949270389616e-02, 6.77705131955832e-02, 6.68719747989748e-02, 6.60910087583275e-02, 6.54189148454307e-02, 6.48467240473742e-02, 6.43653203979433e-02, 6.39655580643215e-02, 6.36383748839151e-02, 6.33749027905708e-02, 6.31665747207677e-02, 6.30052267394525e-02, 6.28831933549350e-02, 6.27933933836573e-02, 6.27294033417094e-02, 6.26855152252190e-02, 6.26567757152081e-02, 6.26390042958214e-02, 6.26287884742958e-02, 6.26234551790859e-02, 6.26210184173708e-02, 6.26201043144192e-02, 6.26198556572461e-02, 6.26198189560050e-02, 6.26198160677802e-02, 6.26197634594234e-02, 6.26194070306525e-02, 6.26180967697749e-02, 6.26146041642741e-02, 6.26069610323796e-02, 6.25923248609119e-02, 6.25668766308394e-02, 6.25257567890039e-02, 6.24630446194736e-02, 6.23717856958983e-02, 6.22440712512926e-02, 6.20711720703824e-02, 6.18437277943151e-02, 6.15519902688817e-02, 6.11861167684253e-02, 6.07365056741322e-02, 6.01941636577017e-02, 5.95510898953791e-02, 5.88006596676927e-02, 5.79379872913357e-02, 5.69602470843087e-02, 5.58669313305771e-02, 5.46600262116463e-02, 5.33440904642610e-02, 5.19262269475540e-02, 5.04159439881455e-02, 4.88249107510583e-02, 4.71666182617412e-02, 4.54559643385544e-02, 4.37087858979375e-02, 4.19413653326970e-02, 4.01699386251160e-02, 3.84102314961758e-02, 3.66770464287391e-02, 3.49839182881873e-02, 3.33428501165187e-02, 3.17641341925814e-02, 3.02562573056164e-02, 2.88258839327931e-02, 2.74779070069080e-02, 2.62155533644126e-02, 2.50405297566410e-02, 2.39531953421377e-02, 2.29527476325195e-02, 2.20374106626444e-02, 2.12046163974122e-02, 2.04511727795165e-02, 1.97734141182373e-02, 1.91673315565204e-02, 1.86286830502844e-02, 1.81530836311232e-02, 1.77360777141402e-02, 1.73731958776919e-02, 1.70599989116340e-02, 1.67921120396244e-02, 1.65652521131072e-02, 1.63752502961678e-02, 1.62180723549241e-02, 1.60898381703447e-02, 1.59868415442013e-02, 1.59055707973863e-02, 1.58427301018413e-02, 1.57952609787273e-02, 1.57603629704433e-02, 1.57355121844915e-02, 1.57184762354632e-02, 1.57073240910748e-02, 1.57004294572083e-02, 1.56964666015149e-02, 1.56943978872768e-02, 1.56934527325419e-02, 1.56930981817425e-02, 1.56930017362065e-02, 1.56929875010761e-02, 1.56929863709188e-02, 1.56929656593370e-02, 1.56928253358366e-02, 1.56923094992862e-02, 1.56909345261909e-02, 1.56879257241798e-02, 1.56821646309773e-02, 1.56721495539642e-02, 1.56559719549065e-02, 1.56313112672990e-02, 1.55954506263676e-02, 1.55453157129983e-02, 1.54775383748955e-02, 1.53885458144920e-02, 1.52746748722095e-02, 1.51323092823894e-02, 1.49580357993564e-02, 1.47488129172544e-02, 1.45021437542816e-02, 1.42162428145447e-02, 1.38901850895579e-02, 1.35240256159044e-02, 1.31188783997835e-02, 1.26769456660938e-02, 1.22014916322293e-02, 1.16967591985855e-02, 1.11678326615099e-02, 1.06204542428578e-02, 1.00608063100012e-02, 9.49527411900419e-03, 8.93020538811840e-03, 8.37168283215206e-03, 7.82532399230007e-03, 7.29611951110343e-03, 6.78831685778292e-03, 6.30535200324900e-03, 5.84982734482864e-03, 5.42353084117053e-03, 5.02748909909556e-03, 4.66204597820763e-03, 4.32695791023008e-03, 4.02149739919011e-03, 3.74455705045307e-03, 3.49474798233018e-03, 3.27048842278188e-03, 3.07008022855726e-03, 2.89177256355089e-03, 2.73381291398778e-03, 2.59448609634916e-03, 2.47214221099412e-03, 2.36521489644304e-03, 2.27223161725476e-03, 2.19181778313526e-03, 2.12269634439366e-03, 2.06368434605153e-03, 2.01368767944032e-03, 1.97169492404825e-03, 1.93677092414646e-03, 1.90805066351629e-03, 1.88473393024324e-03, 1.86608109665798e-03, 1.85141013440671e-03, 1.84009481977654e-03, 1.83156396797202e-03, 1.82530144567394e-03, 1.82084663927995e-03, 1.81779500814372e-03, 1.81579833738965e-03, 1.81456432411409e-03, 1.81385518254729e-03, 1.81348503187451e-03, 1.81331592640888e-03, 1.81325249225495e-03, 1.81323523693067e-03, 1.81323269008971e-03, 1.81323248603960e-03, 1.81322872326817e-03, 1.81320323010326e-03, 1.81310951703750e-03, 1.81285973163980e-03, 1.81231317995767e-03, 1.81126684187305e-03, 1.80944842029263e-03, 1.80651250434494e-03, 1.80204046263199e-03, 1.79554469369947e-03, 1.78647781902441e-03, 1.77424728044784e-03, 1.75823557207622e-03, 1.73782597771321e-03, 1.71243319396908e-03, 1.68153761546108e-03, 1.64472139907446e-03, 1.60170381936453e-03, 1.55237301623501e-03, 1.49681111043520e-03, 1.43530979963913e-03, 1.36837392170719e-03, 1.29671127604425e-03, 1.22120846838314e-03, 1.14289426822722e-03, 1.06289294758355e-03, 9.82370515046682e-04, 9.02478330761625e-04, 8.24300421090385e-04, 7.48809195325606e-04, 6.76831155029589e-04, 6.09025087857873e-04, 5.45874023547555e-04, 4.87685364959134e-04, 4.34597940095927e-04, 3.86602167875338e-04, 3.43564558249909e-04, 3.05252534491984e-04, 2.71373342179970e-04, 2.41598775691187e-04, 2.15565275713203e-04, 1.92914220832563e-04, 1.73297927787530e-04, 1.56328767126013e-04, 1.41687319057837e-04, 1.29155613173282e-04, 1.18394126182864e-04, 1.09080116530436e-04, 1.01235414227574e-04, 9.47007590497151e-05, 8.90707609535468e-05, 8.41995415115093e-05, 7.99693903168475e-05, 7.65793416093175e-05, 7.39192516654441e-05, 7.16674102640980e-05, 6.97748515428032e-05, 6.81993435307504e-05, 6.69053116233442e-05, 6.58606381627180e-05, 6.50343215262694e-05, 6.43956240367061e-05, 6.39143876455250e-05, 6.35619269715448e-05, 6.33119529069859e-05, 6.31412084715066e-05, 6.30297541626135e-05, 6.29609886431126e-05, 6.29215151519702e-05, 6.29009243127143e-05, 6.28915203328972e-05, 6.28879932520476e-05, 6.28870338701404e-05, 6.28868922625259e-05, 6.28868809329219e-05, 6.28866721008592e-05, 6.28852571935568e-05, 6.28800556192743e-05, 6.28661881876269e-05, 6.28358299437082e-05, 6.27776536390456e-05, 6.26763732505253e-05, 6.25123928799098e-05, 6.22615720970998e-05, 6.18951383621891e-05, 6.13797964668095e-05, 6.06780327849776e-05, 5.97483850484082e-05, 5.85615862043198e-05, 5.73233536158568e-05, 5.58468056546246e-05, 5.40519711300647e-05, 5.19210314919860e-05, 4.94450970300875e-05, 4.66299305337259e-05, 4.34939640889809e-05, 4.03298654347850e-05, 3.72397086856324e-05, 3.39462374761013e-05, 3.05229162655962e-05, 2.70880165551546e-05, 2.41072304163481e-05, 2.11499430520234e-05, 1.82606684049347e-05, 1.57324138747583e-05, 1.35170525865647e-05, 1.14152145265223e-05, 9.65050715961979e-06, 8.18572659611050e-06, 6.83847027005413e-06, 5.70350454509594e-06, 4.78517809199275e-06, 3.91894621287747e-06, 3.34755278430521e-06, 2.86082235945173e-06, 2.31884271784111e-06, 2.03658597186518e-06, 1.81822766620598e-06, 1.37837784919270e-06, 1.21880998094070e-06, 1.72706435463965e-06, 2.20681944895773e-06, 2.34661180087414e-06, 2.12831025692140e-06, 1.64148841363954e-06, 1.01153486653516e-06, -3.55271367880050e-15, 4.44356217599307e-07, 5.86640144462081e-07, 5.59229462737676e-07, 4.75330111271433e-07, 3.97910135063739e-07, -7.10542735760100e-15, -3.55271367880050e-15, -7.10542735760100e-15, -3.55271367880050e-15, -7.10542735760100e-15, -3.55271367880050e-15, -7.10542735760100e-15, -7.10542735760100e-15, -3.55271367880050e-15, -3.55271367880050e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15],
[ 3.00000000000000e+01, 2.99999999778409e+01, 2.99999992926375e+01, 2.99999946502767e+01, 2.99999775843595e+01, 2.99999320920614e+01, 2.99998325290760e+01, 2.99996418379391e+01, 2.99993101740376e+01, 2.99987739816213e+01, 2.99979555585483e+01, 2.99967631337736e+01, 2.99950914662618e+01, 2.99928229586412e+01, 2.99898292640612e+01, 2.99859733509181e+01, 2.99811119777931e+01, 2.99750985204429e+01, 2.99677860841635e+01, 2.99590308283173e+01, 2.99486954251394e+01, 2.99366525718359e+01, 2.99227884731483e+01, 2.99070062106587e+01, 2.98892289149566e+01, 2.98694026573587e+01, 2.98474989793605e+01, 2.98235169808432e+01, 2.97974848928546e+01, 2.97694610682029e+01, 2.97395343337477e+01, 2.97078236625190e+01, 2.96744771416756e+01, 2.96396702334023e+01, 2.96036033492858e+01, 2.95664987832336e+01, 2.95285970720901e+01, 2.94901528751533e+01, 2.94514304823049e+01, 2.94126990742224e+01, 2.93742278663552e+01, 2.93362812706941e+01, 2.92991142060174e+01, 2.92629676788388e+01, 2.92280647445951e+01, 2.91946069427772e+01, 2.91627712818098e+01, 2.91327078306032e+01, 2.91045379546886e+01, 2.90783532164384e+01, 2.90542149415237e+01, 2.90321544377895e+01, 2.90121738382504e+01, 2.89942475269717e+01, 2.89783240951571e+01, 2.89643287647680e+01, 2.89521662084235e+01, 2.89417236871792e+01, 2.89328744221241e+01, 2.89254811116660e+01, 2.89193995040617e+01, 2.89144819343170e+01, 2.89105807362093e+01, 2.89075514439577e+01, 2.89052557040454e+01, 2.89035638258448e+01, 2.89023569098938e+01, 2.89015285046991e+01, 2.89009857565042e+01, 2.89006500311652e+01, 2.89004570026983e+01, 2.89003562187174e+01, 2.89003101683767e+01, 2.89002928930926e+01, 2.89002881937765e+01, 2.89002875001665e+01, 2.89002874091795e+01, 2.89002852892914e+01, 2.89002709267223e+01, 2.89002181281687e+01, 2.89000773848405e+01, 2.88997693625326e+01, 2.88991794285693e+01, 2.88981534208443e+01, 2.88964948307475e+01, 2.88939635337455e+01, 2.88902761596463e+01, 2.88851081500793e+01, 2.88780975045475e+01, 2.88688501697037e+01, 2.88569469804345e+01, 2.88419520169223e+01, 2.88234221998640e+01, 2.88009179067709e+01, 2.87740143556647e+01, 2.87423134681085e+01, 2.87054558910141e+01, 2.86631328262868e+01, 2.86150972905154e+01, 2.85611744066338e+01, 2.85012703205531e+01, 2.84353793441777e+01, 2.83635889581423e+01, 2.82860823677286e+01, 2.82031383953394e+01, 2.81151286097322e+01, 2.80225117281277e+01, 2.79258254701925e+01, 2.78256761780843e+01, 2.77227266295980e+01, 2.76176825499832e+01, 2.75112783651326e+01, 2.74042627334786e+01, 2.72973843508512e+01, 2.71913784510586e+01, 2.70869543368220e+01, 2.69847841828871e+01, 2.68854932658332e+01, 2.67896517004629e+01, 2.66977677043613e+01, 2.66102823707692e+01, 2.65275659034144e+01, 2.64499152520780e+01, 2.63775530806776e+01, 2.63106279969853e+01, 2.62492159719391e+01, 2.61933228748501e+01, 2.61428880474514e+01, 2.60977888341383e+01, 2.60578459778267e+01, 2.60228297808370e+01, 2.59924669185186e+01, 2.59664477805491e+01, 2.59444342016438e+01, 2.59260674306249e+01, 2.59109761753217e+01, 2.58987845516180e+01, 2.58891197591484e+01, 2.58816193046945e+01, 2.58759375981017e+01, 2.58717517552220e+01, 2.58687664583402e+01, 2.58667177467705e+01, 2.58653756384080e+01, 2.58645455161594e+01, 2.58640682501759e+01, 2.58638190661620e+01, 2.58637052100650e+01, 2.58636624984108e+01, 2.58636508797768e+01, 2.58636491648897e+01, 2.58636490134978e+01, 2.58636460470482e+01, 2.58636259489068e+01, 2.58635520658849e+01, 2.58633551198906e+01, 2.58629241034048e+01, 2.58620986361320e+01, 2.58606630817168e+01, 2.58583426832627e+01, 2.58548019300923e+01, 2.58496453154287e+01, 2.58424205848495e+01, 2.58326245078578e+01, 2.58197111293843e+01, 2.58031023748166e+01, 2.57822007925563e+01, 2.57564041247537e+01, 2.57251213038190e+01, 2.56877893850829e+01, 2.56438908512480e+01, 2.55929706693980e+01, 2.55346524534843e+01, 2.54686530904180e+01, 2.53947952298235e+01, 2.53130171164692e+01, 2.52233793565375e+01, 2.51260683462032e+01, 2.50213962419061e+01, 2.49097975028793e+01, 2.47918221750299e+01, 2.46681262012722e+01, 2.45394591318660e+01, 2.44066496697823e+01, 2.42705895256524e+01, 2.41322160814995e+01, 2.39924943778898e+01, 2.38523989470043e+01, 2.37128960109953e+01, 2.35749265438139e+01, 2.34393906483036e+01, 2.33071336255496e+01, 2.31789340138382e+01, 2.30554937606959e+01, 2.29374305781717e+01, 2.28252724335431e+01, 2.27194540554502e+01, 2.26203152930015e+01, 2.25281011501358e+01, 2.24429633224815e+01, 2.23649630804905e+01, 2.22940753627191e+01, 2.22301939608920e+01, 2.21731376903264e+01, 2.21226574440725e+01, 2.20784440268777e+01, 2.20401366567962e+01, 2.20073320093515e+01, 2.19795936631707e+01, 2.19564617885495e+01, 2.19374629031084e+01, 2.19221195032339e+01, 2.19099593680149e+01, 2.19005243254985e+01, 2.18933782707550e+01, 2.18881142326080e+01, 2.18843603017270e+01, 2.18817842573139e+01, 2.18800967624784e+01, 2.18790530386121e+01, 2.18784529750770e+01, 2.18781396803034e+01, 2.18779965316142e+01, 2.18779428313207e+01, 2.18779282235126e+01, 2.18779260674297e+01, 2.18779258944852e+01, 2.18779227027816e+01, 2.18779010785126e+01, 2.18778215853085e+01, 2.18776096848119e+01, 2.18771459416972e+01, 2.18762578025921e+01, 2.18747132725286e+01, 2.18722167705095e+01, 2.18684073972087e+01, 2.18628597921428e+01, 2.18550876939219e+01, 2.18445502444366e+01, 2.18306609954233e+01, 2.18127994839262e+01, 2.17903251430855e+01, 2.17625932093489e+01, 2.17289721813223e+01, 2.16888622855020e+01, 2.16417143179869e+01, 2.15870481677415e+01, 2.15244702947304e+01, 2.14536894426344e+01, 2.13745299155233e+01, 2.12869418413951e+01, 2.11910079784801e+01, 2.10869467829609e+01, 2.09751116351592e+01, 2.08559862987287e+01, 2.07301768479805e+01, 2.05984004298666e+01, 2.04614713232029e+01, 2.03202848189318e+01, 2.01757994773351e+01, 2.00290183289629e+01, 1.98809695822654e+01, 1.97326873857150e+01, 1.95851931650418e+01, 1.94394780143330e+01, 1.92964865606921e+01, 1.91571026458389e+01, 1.90221370778873e+01, 1.88923176091523e+01, 1.87682811995301e+01, 1.86505685380046e+01, 1.85396207236744e+01, 1.84357779561003e+01, 1.83392800534839e+01, 1.82502686042874e+01, 1.81687905596078e+01, 1.80948030852184e+01, 1.80281795088346e+01, 1.79687162155562e+01, 1.79161403591767e+01, 1.78701182668239e+01, 1.78302644179710e+01, 1.77961508760060e+01, 1.77673170418629e+01, 1.77432795859956e+01, 1.77235423990000e+01, 1.77076063846061e+01, 1.76949789038842e+01, 1.76851826686577e+01, 1.76777638774140e+01, 1.76722993902348e+01, 1.76684027516388e+01, 1.76657288923223e+01, 1.76639773723847e+01, 1.76628940687839e+01, 1.76622712568077e+01, 1.76619460870319e+01, 1.76617975129523e+01, 1.76617417774715e+01, 1.76617266160492e+01, 1.76617243782543e+01, 1.76617242056540e+01, 1.76617211063069e+01, 1.76617001077693e+01, 1.76616229148517e+01, 1.76614171463201e+01, 1.76609668241070e+01, 1.76601043924110e+01, 1.76586045827065e+01, 1.76561804008859e+01, 1.76524814667264e+01, 1.76470948828853e+01, 1.76395487495949e+01, 1.76293183710285e+01, 1.76158351191725e+01, 1.75984978309959e+01, 1.75766865159769e+01, 1.75497780463829e+01, 1.75171633965384e+01, 1.74782658957608e+01, 1.74325598701547e+01, 1.73795889793114e+01, 1.73189835134759e+01, 1.72504759123139e+01, 1.71739138034956e+01, 1.70892699404621e+01, 1.69966485428220e+01, 1.68962877045430e+01, 1.67885577249663e+01, 1.66739554224643e+01, 1.65530946943855e+01, 1.64266937727965e+01, 1.62955597773676e+01, 1.61605712715826e+01, 1.60226595783663e+01, 1.58827896047162e+01, 1.57419408671968e+01, 1.56010893122909e+01, 1.54611904025975e+01, 1.53231638080137e+01, 1.51878799152231e+01, 1.50561482602990e+01, 1.49287079044488e+01, 1.48062197133877e+01, 1.46892604640362e+01, 1.45783186831841e+01, 1.44737921156204e+01, 1.43759867186494e+01, 1.42851170819350e+01, 1.42013081738103e+01, 1.41245983165022e+01, 1.40549432929402e+01, 1.39922214870459e+01, 1.39362399577148e+01, 1.38867413438908e+01, 1.38434114937890e+01, 1.38058877049347e+01, 1.37737674528874e+01, 1.37466174752569e+01, 1.37239830643161e+01, 1.37053974070730e+01, 1.36903907974662e+01, 1.36784995330928e+01, 1.36692743004440e+01, 1.36622878498292e+01, 1.36571417655973e+01, 1.36534721500347e+01, 1.36509540609653e+01, 1.36493045734367e+01, 1.36482843740737e+01, 1.36476978411905e+01, 1.36473916124931e+01, 1.36472516927765e+01, 1.36471992038506e+01, 1.36471849255703e+01, 1.36471828181254e+01, 1.36471826585560e+01, 1.36471798318002e+01, 1.36471606801087e+01, 1.36470902763839e+01, 1.36469026052807e+01, 1.36464918885073e+01, 1.36457053046641e+01, 1.36443373909327e+01, 1.36421263770999e+01, 1.36387526598392e+01, 1.36338395763066e+01, 1.36269565800520e+01, 1.36176248580916e+01, 1.36053253551489e+01, 1.35895090896810e+01, 1.35696095576118e+01, 1.35450569263129e+01, 1.35152936275250e+01, 1.34797908692487e+01, 1.34380655100061e+01, 1.33896966816071e+01, 1.33343415156449e+01, 1.32717493299970e+01, 1.32017736678535e+01, 1.31243816533719e+01, 1.30396602316856e+01, 1.29478189902978e+01, 1.28491894052718e+01, 1.27442205095702e+01, 1.26334711333221e+01, 1.25175990090806e+01, 1.23973471634052e+01, 1.22735281246841e+01, 1.21470065615533e+01, 1.20186810213212e+01, 1.18894654574780e+01, 1.17602712141638e+01, 1.16319900709060e+01, 1.15054788461515e+01, 1.13815459232683e+01, 1.12609399140188e+01, 1.11443405310049e+01, 1.10323516194400e+01, 1.09254962111353e+01, 1.08242134132525e+01, 1.07288569273415e+01, 1.06396950018110e+01, 1.05569116430438e+01, 1.04806089377035e+01, 1.04108103648032e+01, 1.03474649971000e+01, 1.02904525059534e+01, 1.02395888920058e+01, 1.01946328666393e+01, 1.01552928069758e+01, 1.01212342009545e+01, 1.00920874893669e+01, 1.00674561992756e+01, 1.00469252487743e+01, 1.00300692875988e+01, 1.00164609230025e+01, 1.00056786671394e+01, 9.99731443265135e+00, 9.99098039890928e+00, 9.98631507385253e+00, 9.98298838666970e+00, 9.98070566517885e+00, 9.97921037857682e+00, 9.97828556045078e+00, 9.97775386717133e+00, 9.97747627110030e+00, 9.97734943418937e+00, 9.97730185312771e+00, 9.97728890991040e+00, 9.97728699951809e+00, 9.97728685610434e+00, 9.97728433185952e+00, 9.97726722972572e+00, 9.97720436041613e+00, 9.97703677348436e+00, 9.97667001137631e+00, 9.97596760960583e+00, 9.97474610056502e+00, 9.97277174613194e+00, 9.96975918680242e+00, 9.96537215192215e+00, 9.95922632579847e+00, 9.95089440706040e+00, 9.93991333294484e+00, 9.92579356623525e+00, 9.90803026143691e+00, 9.88611604086523e+00, 9.85955502459184e+00, 9.82787767605248e+00, 9.79065595415716e+00, 9.74751821012872e+00, 9.69816323993564e+00, 9.64237290669459e+00, 9.58002278506624e+00, 9.51109035150348e+00, 9.43566034660124e+00, 9.35392706157115e+00, 9.26619344016190e+00, 9.17286702907056e+00, 9.07445294348481e+00, 8.97154413147314e+00, 8.86480931631030e+00, 8.75497906759755e+00, 8.64283050081817e+00, 8.52917113251584e+00, 8.41482242618384e+00, 8.30060355286517e+00, 8.18731586019065e+00, 8.07572849385559e+00, 7.96656554711420e+00, 7.86049502942821e+00, 7.75811984985331e+00, 7.65997091083670e+00, 7.56650231148139e+00, 7.47808857334289e+00, 7.39502373236784e+00, 7.31752209135605e+00, 7.24572039950720e+00, 7.17968121799841e+00, 7.11939724002703e+00, 7.06479635599559e+00, 7.01574728461788e+00, 6.97206562385651e+00, 6.93352020748266e+00, 6.89983968026500e+00, 6.87071922497314e+00, 6.84582738623981e+00, 6.82481293963372e+00, 6.80731174976546e+00, 6.79295355038593e+00, 6.78136856435051e+00, 6.77219386450866e+00, 6.76507936068863e+00, 6.75969328554328e+00, 6.75572704534580e+00, 6.75289930258053e+00, 6.75095916638145e+00, 6.74968838475693e+00, 6.74890245852752e+00, 6.74845062968112e+00, 6.74821473450595e+00, 6.74810695209829e+00, 6.74806651920360e+00, 6.74805552047659e+00, 6.74805389708754e+00, 6.74805377558981e+00, 6.74805164202859e+00, 6.74803718683973e+00, 6.74798404807195e+00, 6.74784239990002e+00, 6.74753240772627e+00, 6.74693874094201e+00, 6.74590636606430e+00, 6.74423781854558e+00, 6.74169212309413e+00, 6.73798549880482e+00, 6.73279394538810e+00, 6.72575775886866e+00, 6.71648796819130e+00, 6.70457461783237e+00, 6.68959674645825e+00, 6.67113382993893e+00, 6.64877837214496e+00, 6.62214924390207e+00, 6.59090529541689e+00, 6.55475870732730e+00, 6.51348750728791e+00, 6.46694666904885e+00, 6.41507723425261e+00, 6.35791295639103e+00, 6.29558406145165e+00, 6.22831784745204e+00, 6.15643599872532e+00, 6.08034866084231e+00, 6.00054549637365e+00, 5.91758410681944e+00, 5.83207634835478e+00, 5.74467317633454e+00, 5.65604871644716e+00, 5.56688427386150e+00, 5.47785295551036e+00, 5.38960549973629e+00, 5.30275779130483e+00, 5.21788040086045e+00, 5.13549034029554e+00, 5.05604508284392e+00, 4.97993877055790e+00, 4.90750043053312e+00, 4.83899394939102e+00, 4.77461951402533e+00, 4.71451621324999e+00, 4.65876550529341e+00, 4.60739528432936e+00, 4.56038431928955e+00, 4.51766688426506e+00, 4.47913744685857e+00, 4.44465532492451e+00, 4.41404926035026e+00, 4.38712188903773e+00, 4.36365410808443e+00, 4.34340935411203e+00, 4.32613781112281e+00, 4.31158056302459e+00, 4.29947369628707e+00, 4.28955234362253e+00, 4.28155464191282e+00, 4.27522555878851e+00, 4.27032052432593e+00, 4.26660878922027e+00, 4.26387642027069e+00, 4.26192883946731e+00, 4.26059281529753e+00, 4.25971782439030e+00, 4.25917671794742e+00, 4.25886564961197e+00, 4.25870324800839e+00, 4.25862904631089e+00, 4.25860121084393e+00, 4.25859363893632e+00, 4.25859252133912e+00, 4.25859243766459e+00, 4.25859096787936e+00, 4.25858100987569e+00, 4.25854440331329e+00, 4.25844682452200e+00, 4.25823328010808e+00, 4.25782433380887e+00, 4.25711322533885e+00, 4.25596402617213e+00, 4.25421096079116e+00, 4.25165900152338e+00, 4.24808581890750e+00, 4.24324513570232e+00, 4.23687148959250e+00, 4.22868635651369e+00, 4.21840552363992e+00, 4.20574753030512e+00, 4.19044292007133e+00, 4.17224397315310e+00, 4.15093452227032e+00, 4.12633940442313e+00, 4.09833307378439e+00, 4.06684690357183e+00, 4.03187474191775e+00, 3.99347635973586e+00, 3.95177853483457e+00, 3.90697364936064e+00, 3.85931582665276e+00, 3.80911478559446e+00, 3.75672773126628e+00, 3.70254971649986e+00, 3.64700298881870e+00, 3.59052587434041e+00, 3.53356174271677e+00, 3.47654854847860e+00, 3.41990936209293e+00, 3.36404419957640e+00, 3.30932334493973e+00, 3.25608224692345e+00, 3.20461797036682e+00, 3.15518710018153e+00, 3.10800493607116e+00, 3.06324577955554e+00, 3.02104409973909e+00, 2.98149636711447e+00, 2.94466336115691e+00, 2.91057278301809e+00, 2.87922203512097e+00, 2.85058106141591e+00, 2.82459517281973e+00, 2.80118781004913e+00, 2.78026321948830e+00, 2.76170903625768e+00, 2.74539878204035e+00, 2.73119429353037e+00, 2.71894810085414e+00, 2.70850577440377e+00, 2.69970825376529e+00, 2.69239416451149e+00, 2.68640211837907e+00, 2.68157298070201e+00, 2.67775207697325e+00, 2.67479129915554e+00, 2.67255106295347e+00, 2.67090206069272e+00, 2.66972675154526e+00, 2.66892053212844e+00, 2.66839253618564e+00, 2.66806602193408e+00, 2.66787831917055e+00, 2.66778032446018e+00, 2.66773555056362e+00, 2.66771875444184e+00, 2.66771418550010e+00, 2.66771351113426e+00, 2.66771346051052e+00, 2.66771256949037e+00, 2.66770653270317e+00, 2.66768434091678e+00, 2.66762518639105e+00, 2.66749573135036e+00, 2.66724782169168e+00, 2.66681674298720e+00, 2.66612010793914e+00, 2.66505745496523e+00, 2.66351062558580e+00, 2.66134497251341e+00, 2.65841142950653e+00, 2.65454944702619e+00, 2.64959076391608e+00, 2.64336394491826e+00, 2.63569956812632e+00, 2.62643589799962e+00, 2.61542483209718e+00, 2.60253786803433e+00, 2.58767180667993e+00, 2.57075389352454e+00, 2.55174610673907e+00, 2.53064833018226e+00, 2.50750020248345e+00, 2.48238150642842e+00, 2.45541105050848e+00, 2.42674408872188e+00, 2.39656841637898e+00, 2.36509935966455e+00, 2.33257393734806e+00, 2.29924450909797e+00, 2.26537223429562e+00, 2.23122064925959e+00, 2.19704963339726e+00, 2.16310998195890e+00, 2.12963874163640e+00, 2.09685540192560e+00, 2.06495897568185e+00, 2.03412595087497e+00, 2.00450905475674e+00, 1.97623674249213e+00, 1.94941330450390e+00, 1.92411947916221e+00, 1.90041345830760e+00, 1.87833218048891e+00, 1.85789281879403e+00, 1.83909438496661e+00, 1.82191938758771e+00, 1.80633549817430e+00, 1.79229719409614e+00, 1.77974736048017e+00, 1.76861884422995e+00, 1.75883596160934e+00, 1.75031596637273e+00, 1.74297048816931e+00, 1.73670695104196e+00, 1.73142997954256e+00, 1.72704279567205e+00, 1.72344860400511e+00, 1.72055195555630e+00, 1.71826007383081e+00, 1.71648411976835e+00, 1.71514036661321e+00, 1.71415125175017e+00, 1.71344627073988e+00, 1.71296267950329e+00, 1.71264597396444e+00, 1.71245012235173e+00, 1.71233753343571e+00, 1.71227875370625e+00, 1.71225189718052e+00, 1.71224182243817e+00, 1.71223908187036e+00, 1.71223867736853e+00, 1.71223864687380e+00, 1.71223810841826e+00, 1.71223446030611e+00, 1.71222104951045e+00, 1.71218530164074e+00, 1.71210707026975e+00, 1.71195725537459e+00, 1.71169674960606e+00, 1.71127576636098e+00, 1.71063359794293e+00, 1.70969884487952e+00, 1.70839014796627e+00, 1.70661744203640e+00, 1.70428373413199e+00, 1.70128738830648e+00, 1.69752487484070e+00, 1.69289391398248e+00, 1.68729691494926e+00, 1.68064458213926e+00, 1.67285953516443e+00, 1.66387977066580e+00, 1.65366178503952e+00, 1.64218318076395e+00, 1.62944459647819e+00, 1.61547083235884e+00, 1.60031108602028e+00, 1.58403826683053e+00, 1.56674741357664e+00, 1.54855329646980e+00, 1.52958733416271e+00, 1.50999399508000e+00, 1.48992687658965e+00, 1.46954466375157e+00, 1.44900716179959e+00, 1.42847157506504e+00, 1.40808917295787e+00, 1.38800244488823e+00, 1.36834280485501e+00, 1.34922886675615e+00, 1.33076527647914e+00, 1.31304205872980e+00, 1.29613441651152e+00, 1.28010290933413e+00, 1.26499393194882e+00, 1.25084041743144e+00, 1.23766269521929e+00, 1.22546944462193e+00, 1.21425869586897e+00, 1.20401884265772e+00, 1.19472964144780e+00, 1.18636318274053e+00, 1.17888482786690e+00, 1.17225411118846e+00, 1.16642561205299e+00, 1.16134980340700e+00, 1.15697388479536e+00, 1.15324260676408e+00, 1.15009909165537e+00, 1.14748565270857e+00, 1.14534460955183e+00, 1.14361909392121e+00, 1.14225383513429e+00, 1.14119591085127e+00, 1.14039544535180e+00, 1.13980623527970e+00, 1.13938628183987e+00, 1.13909820895537e+00, 1.13890954897242e+00, 1.13879288106694e+00, 1.13872581235592e+00, 1.13869079753206e+00, 1.13867479921816e+00, 1.13866879773847e+00, 1.13866716519427e+00, 1.13866692423430e+00, 1.13866690596987e+00, 1.13866658215748e+00, 1.13866438828294e+00, 1.13865632339905e+00, 1.13863482561765e+00, 1.13858777945671e+00, 1.13849768505832e+00, 1.13834102462428e+00, 1.13808785858496e+00, 1.13770168061273e+00, 1.13713955625562e+00, 1.13635256423579e+00, 1.13528655183739e+00, 1.13388320587860e+00, 1.13208142828720e+00, 1.12981899035426e+00, 1.12703442280890e+00, 1.12366908089941e+00, 1.11966930611436e+00, 1.11498859085409e+00, 1.10958964129896e+00, 1.10344622888988e+00, 1.09654472383789e+00, 1.08888521580391e+00, 1.08048214726149e+00, 1.07136441288839e+00, 1.06157491136846e+00, 1.05116957113073e+00, 1.04021590527772e+00, 1.02879117980800e+00, 1.01698030035566e+00, 1.00487353420439e+00, 9.92564185691233e-01, 9.80146334974481e-01, 9.67712734255528e-01, 9.55352934385793e-01, 9.43151691108753e-01, 9.31187676598189e-01, 9.19532500608749e-01, 9.08250027949087e-01, 8.97395965908615e-01, 8.87017686861228e-01, 8.77154247175785e-01, 8.67836563096930e-01, 8.59087706571476e-01, 8.50923288209898e-01, 8.43351899885523e-01, 8.36375595208416e-01, 8.29990391740704e-01, 8.24186783964887e-01, 8.18950260429808e-01, 8.14261822038262e-01, 8.10098501047278e-01, 8.06433882026905e-01, 8.03238626807737e-01, 8.00481005413431e-01, 7.98127434216244e-01, 7.96143021185415e-01, 7.94492116255263e-01, 7.93138862675750e-01, 7.92047742899591e-01, 7.91184110296829e-01, 7.90514695972252e-01, 7.90008078390411e-01, 7.89635102568067e-01, 7.89369235418626e-01, 7.89186844520131e-01, 7.89067389151530e-01, 7.88993514853324e-01, 7.88951045894887e-01, 7.88928873685190e-01, 7.88918743113967e-01, 7.88914942802244e-01, 7.88913909026750e-01, 7.88913756443709e-01, 7.88913744804738e-01, 7.88913537487751e-01, 7.88912132885621e-01, 7.88906969433476e-01, 7.88893205672363e-01, 7.88863084558436e-01, 7.88805401093192e-01, 7.88705095543726e-01, 7.88542992502446e-01, 7.88295703635157e-01, 7.87935708995732e-01, 7.87431627999645e-01, 7.86748686283147e-01, 7.85849378509958e-01, 7.84694319606505e-01, 7.83243267969382e-01, 7.81456294187549e-01, 7.79295058316173e-01, 7.76724148557950e-01, 7.73712425394070e-01, 7.70234308917146e-01, 7.66270944463500e-01, 7.61811183506293e-01, 7.56852323624457e-01, 7.51400563116245e-01, 7.45471141765080e-01, 7.39088158081241e-01, 7.32284073280571e-01, 7.25098931357103e-01, 7.17579340967099e-01, 7.09777276929493e-01, 7.01748765987350e-01, 6.93552522773089e-01, 6.85248598054425e-01, 6.76897093244968e-01, 6.68556984147315e-01, 6.60285084408393e-01, 6.52135166592767e-01, 6.44157247277160e-01, 6.36397032963160e-01, 6.28895516354643e-01, 6.21688707767152e-01, 6.14807483964974e-01, 6.08277536200212e-01, 6.02119400194979e-01, 5.96348552774504e-01, 5.90975562372993e-01, 5.86006283314070e-01, 5.81442086325179e-01, 5.77280119981260e-01, 5.73513599569836e-01, 5.70132121175202e-01, 5.67121999584884e-01, 5.64466628952726e-01, 5.62146865049982e-01, 5.60141427453857e-01, 5.58427319222719e-01, 5.56980260557559e-01, 5.55775131728442e-01, 5.54786419239800e-01, 5.53988657917497e-01, 5.53356860429933e-01, 5.52866924814928e-01, 5.52496009980505e-01, 5.52222868976102e-01, 5.52028130162796e-01, 5.51894517288037e-01, 5.51807000892342e-01, 5.51752875400972e-01, 5.51721758597228e-01, 5.51705512815271e-01, 5.51698089981901e-01, 5.51695305420608e-01, 5.51694547952398e-01, 5.51694436151681e-01, 5.51694427563387e-01, 5.51694273797068e-01, 5.51693232007882e-01, 5.51689402283795e-01, 5.51679193669788e-01, 5.51656852491153e-01, 5.51614066934139e-01, 5.51539663969908e-01, 5.51419412963995e-01, 5.51235948270389e-01, 5.50968820678996e-01, 5.50594684645137e-01, 5.50087624653031e-01, 5.49419619745589e-01, 5.48561140148450e-01, 5.47481864081700e-01, 5.46151496477233e-01, 5.44540664743295e-01, 5.42621860431346e-01, 5.40370390275442e-01, 5.37765296261977e-01, 5.34790202788074e-01, 5.31434050083789e-01, 5.27691677165805e-01, 5.23564224591034e-01, 5.19059336754989e-01, 5.14191154658615e-01, 5.08980101921296e-01, 5.03452478208288e-01, 4.97639884085437e-01, 4.91578508749274e-01, 4.85308316584799e-01, 4.78872169954606e-01, 4.72314924282241e-01, 4.65682527923267e-01, 4.59021154257702e-01, 4.52376387662710e-01, 4.45792479225418e-01, 4.39311682752770e-01, 4.32973677143266e-01, 4.26815077622631e-01, 4.20869035674933e-01, 4.15164925590865e-01, 4.09728114234323e-01, 4.04579809735292e-01, 3.99736984223274e-01, 3.95212365336725e-01, 3.91014491032792e-01, 3.87147822154560e-01, 3.83612907272070e-01, 3.80406594476451e-01, 3.77522285037699e-01, 3.74950224085989e-01, 3.72677823686193e-01, 3.70690013786017e-01, 3.68969616480054e-01, 3.67497738812624e-01, 3.66254178934355e-01, 3.65217839851596e-01, 3.64367144313164e-01, 3.63680443639264e-01, 3.63136412604362e-01, 3.62714421940691e-01, 3.62394879729941e-01, 3.62159532985991e-01, 3.61991721164721e-01, 3.61876574202959e-01, 3.61801148986778e-01, 3.61754499841545e-01, 3.61727680651352e-01, 3.61713678453494e-01, 3.61707280696287e-01, 3.61704880670640e-01, 3.61704227804900e-01, 3.61704131443258e-01, 3.61704123984470e-01, 3.61703989705621e-01, 3.61703079946572e-01, 3.61699735575161e-01, 3.61690820701991e-01, 3.61671310685203e-01, 3.61633946535225e-01, 3.61568969546138e-01, 3.61463948332791e-01, 3.61303708540220e-01, 3.61070373266035e-01, 3.60743519630475e-01, 3.60300453872970e-01, 3.59716603796695e-01, 3.58966023313741e-01, 3.58021999317984e-01, 3.56857746266950e-01, 3.55447168918769e-01, 3.53765668962847e-01, 3.51790967191761e-01, 3.49503909807261e-01, 3.46889225836321e-01, 3.43936202776781e-01, 3.40639249683541e-01, 3.36998320951974e-01, 3.33019179864593e-01, 3.28713488163491e-01, 3.24098715976319e-01, 3.19197874766626e-01, 3.14039084009956e-01, 3.08654989497683e-01, 3.03082057139573e-01, 2.97359770613184e-01, 2.91529764062219e-01, 2.85634922256484e-01, 2.79718480252470e-01, 2.73823152754737e-01, 2.67990320240820e-01, 2.62259294690928e-01, 2.56666682721175e-01, 2.51245858368723e-01, 2.46026552066645e-01, 2.41034556836635e-01, 2.36291547760146e-01, 2.31815006649434e-01, 2.27618240731100e-01, 2.23710482176639e-01, 2.20097054458908e-01, 2.16779591677010e-01, 2.13756297993534e-01, 2.11022235936120e-01, 2.08569634279019e-01, 2.06388208291383e-01, 2.04465487098386e-01, 2.02787144568955e-01, 2.01337331395141e-01, 2.00099006790058e-01, 1.99054268489125e-01, 1.98184679529145e-01, 1.97471589679814e-01, 1.96896448529618e-01, 1.96441106218931e-01, 1.96088096818329e-01, 1.95820898514551e-01, 1.95624164220000e-01, 1.95483916066333e-01, 1.95387697540220e-01, 1.95324677790875e-01, 1.95285703857294e-01, 1.95263298162640e-01, 1.95251600500548e-01, 1.95246255771753e-01, 1.95244250784253e-01, 1.95243705379500e-01, 1.95243624878913e-01, 1.95243618597317e-01, 1.95243504858613e-01, 1.95242734262642e-01, 1.95239901476580e-01, 1.95232350357120e-01, 1.95215825161164e-01, 1.95184178370194e-01, 1.95129147265892e-01, 1.95040210244208e-01, 1.94904532500530e-01, 1.94707009073415e-01, 1.94430411189138e-01, 1.94055639287399e-01, 1.93562082937923e-01, 1.92928084032005e-01, 1.92131495181975e-01, 1.91150320308772e-01, 1.89963419170819e-01, 1.88551252416644e-01, 1.86896639044338e-01, 1.84985494389558e-01, 1.82807514414982e-01, 1.80356771561243e-01, 1.77632189062400e-01, 1.74637864586330e-01, 1.71383220292249e-01, 1.67882964644878e-01, 1.64156861109333e-01, 1.60229309507876e-01, 1.56128756547904e-01, 1.51886961964710e-01, 1.47538155024311e-01, 1.43118122075787e-01, 1.38663268908900e-01, 1.34209701607844e-01, 1.29792366451515e-01, 1.25444283554376e-01, 1.21195900994227e-01, 1.17074586940998e-01, 1.13104267672604e-01, 1.09305210197316e-01, 1.05693940224171e-01, 1.02283279959874e-01, 9.90824859476547e-02, 9.60974649512316e-02, 9.33310455690055e-02, 9.07832845347158e-02, 8.84517891263386e-02, 8.63320403430308e-02, 8.44177051189483e-02, 8.27009294751058e-02, 8.11726078935742e-02, 7.98226271354636e-02, 7.86400850937170e-02, 7.76134870137177e-02, 7.67309225206319e-02, 7.59802273919483e-02, 7.53491339658794e-02, 7.48254135630226e-02, 7.43970134179825e-02, 7.40521894812893e-02, 7.37796351812854e-02, 7.35686049542252e-02, 7.34090301819279e-02, 7.32916242299240e-02, 7.32079726474240e-02, 7.31506043363161e-02, 7.31130396487352e-02, 7.30898119220420e-02, 7.30764598595037e-02, 7.30694893361203e-02, 7.30663045506681e-02, 7.30651098445598e-02, 7.30647848571238e-02, 7.30647368897230e-02, 7.30647331148973e-02, 7.30646643573785e-02, 7.30641985159934e-02, 7.30624860477533e-02, 7.30579213434623e-02, 7.30479321442381e-02, 7.30288037191045e-02, 7.29955457564451e-02, 7.29418095776779e-02, 7.28598628339867e-02, 7.27406281703580e-02, 7.25737912765787e-02, 7.23479821229240e-02, 7.20510308517248e-02, 7.16702966729414e-02, 7.11930641916005e-02, 7.06069969968262e-02, 6.99006333257834e-02, 6.90639035799521e-02, 6.80886449349565e-02, 6.69690848409665e-02, 6.57022634554991e-02, 6.42883654996709e-02, 6.27309350242591e-02, 6.10369521979912e-02, 5.92167592573425e-02, 5.72838326235185e-02, 5.52544090406535e-02, 5.31469843533827e-02, 5.09817130802688e-02, 4.87797441948601e-02, 4.65625326833425e-02, 4.43511670615742e-02, 4.21657501023809e-02, 4.00248639831275e-02, 3.79451427177635e-02, 3.59409651242828e-02, 3.40242718032613e-02, 3.22045006790148e-02, 3.04886283686123e-02, 2.88812994804353e-02, 2.73850230637933e-02, 2.60004147090136e-02, 2.47264638901825e-02, 2.35608085795853e-02, 2.25000024350770e-02, 2.15397634917522e-02, 2.06751968835377e-02, 1.99009873927132e-02, 1.92115603979417e-02, 1.86012119856116e-02, 1.80642105995581e-02, 1.75948736757832e-02, 1.71876233167154e-02, 1.68370252856427e-02, 1.65378155278688e-02, 1.62849181235210e-02, 1.60734581044757e-02, 1.58987719762003e-02, 1.57564181121863e-02, 1.56421884717659e-02, 1.55521223660458e-02, 1.54825222981962e-02, 1.54299712706418e-02, 1.53913504218970e-02, 1.53638554626703e-02, 1.53450101527284e-02, 1.53326750113685e-02, 1.53250495847495e-02, 1.53206668877495e-02, 1.53183790631708e-02, 1.53173338144548e-02, 1.53169417180727e-02, 1.53168350595259e-02, 1.53168193169897e-02, 1.53168180671521e-02, 1.53167951623452e-02, 1.53166399795772e-02, 1.53160695211909e-02, 1.53145489644828e-02, 1.53112216484317e-02, 1.53048509157294e-02, 1.52937767747368e-02, 1.52758904095123e-02, 1.52486296025067e-02, 1.52089980599364e-02, 1.51536113383450e-02, 1.50787714477367e-02, 1.49805711496960e-02, 1.48550274110981e-02, 1.46982414106205e-02, 1.45065800080033e-02, 1.42768708567935e-02, 1.40066006570407e-02, 1.36941037741103e-02, 1.33387270050740e-02, 1.29409560457496e-02, 1.25024904901672e-02, 1.20262570956449e-02, 1.15163554538960e-02, 1.09779357451387e-02, 1.04170143162605e-02, 9.84023865923334e-03, 9.25461818906115e-03, 8.66724035448030e-03, 8.08499262211271e-03, 7.51430963632060e-03, 6.96096159840209e-03, 6.42989514808434e-03, 5.92513247729798e-03, 5.44972882770978e-03, 5.00578360997750e-03, 4.59449664506906e-03, 4.21625873654108e-03, 3.87076495836425e-03, 3.55713949874925e-03, 3.27406230865179e-03, 3.01988978600320e-03, 2.79276392627636e-03, 2.59070650195170e-03, 2.41169670461261e-03, 2.25373218632541e-03, 2.11487453405468e-03, 1.99328091762396e-03, 1.88722402003449e-03, 1.79510246174175e-03, 1.71544384742006e-03, 1.64690236673337e-03, 1.58825262748863e-03, 1.53838112862559e-03, 1.49627651861195e-03, 1.46101954264566e-03, 1.43177336681788e-03, 1.40777477779253e-03, 1.38832659056121e-03, 1.37279144983538e-03, 1.36058707975906e-03, 1.35118291986203e-03, 1.34409798319624e-03, 1.33889968751078e-03, 1.33520334640735e-03, 1.33267196842013e-03, 1.33101600100716e-03, 1.32999267465195e-03, 1.32940464859033e-03, 1.32909772863243e-03, 1.32895751324114e-03, 1.32890491665805e-03, 1.32889060940400e-03, 1.32888849768875e-03, 1.32888832850142e-03, 1.32888520858998e-03, 1.32886407089217e-03, 1.32878636890510e-03, 1.32857926248420e-03, 1.32812610929989e-03, 1.32725862739491e-03, 1.32575119716449e-03, 1.32331781534845e-03, 1.31961225542909e-03, 1.31423200572911e-03, 1.30672652307595e-03, 1.29661022544525e-03, 1.28338042646092e-03, 1.26654006983173e-03, 1.24562465524036e-03, 1.22023218375489e-03, 1.19005434706310e-03, 1.15490662603079e-03, 1.11475455629062e-03, 1.06973327410032e-03, 1.02015766752217e-03, 9.66521075003612e-04, 9.09481476629281e-04, 8.49835412623179e-04, 7.88481263430896e-04, 7.26374811584662e-04, 6.64480951492408e-04, 6.03725848289827e-04, 5.44953704533668e-04, 4.88891624719656e-04, 4.36125016108235e-04, 3.87084707384133e-04, 3.42045664513790e-04, 3.01135970563138e-04, 2.64353751799007e-04, 2.31589137197119e-04, 2.02648277507933e-04, 1.77276943112759e-04, 1.55182070340487e-04, 1.36050482694827e-04, 1.19564570397301e-04, 1.05414886224509e-04, 9.33095772168713e-05, 8.29806010926859e-05, 7.41869160130193e-05, 6.67151608908512e-05, 6.03785550516989e-05, 5.50147659375000e-05, 5.04833632284374e-05, 4.66632664561928e-05, 4.34503823925070e-05, 4.07554798442789e-05, 3.85022738136342e-05, 3.66256740846893e-05, 3.50701724762814e-05, 3.37883715602061e-05, 3.27396735855245e-05, 3.18891463031434e-05, 3.12065697372077e-05, 3.06656544601935e-05, 3.02434146668418e-05, 2.99196775159771e-05, 2.96767109588814e-05, 2.94989517861666e-05, 2.93728139055816e-05, 2.92865553177535e-05, 2.92301818127783e-05, 2.91953667321820e-05, 2.91753686845198e-05, 2.91649328829635e-05, 2.91601658091167e-05, 2.91583776999005e-05, 2.91578913085289e-05, 2.91578195188436e-05, 2.91578137137094e-05, 2.91577059314818e-05, 2.91569757138177e-05, 2.91542915071830e-05, 2.91471375675201e-05, 2.91314872775672e-05, 2.91015378479642e-05, 2.90495266064283e-05, 2.89656530689797e-05, 2.88381342592459e-05, 2.86534228735036e-05, 2.83966157539339e-05, 2.80520728850320e-05, 2.76042518088104e-05, 2.70387393612737e-05, 2.63434328253709e-05, 2.55097912997826e-05, 2.45340532316618e-05, 2.34183081992967e-05, 2.21713226018494e-05, 2.08090420237284e-05, 1.93547071098976e-05, 1.78385117628466e-05, 1.62966766410477e-05, 1.47696233625538e-05, 1.32986439496108e-05, 1.19208195457077e-05, 1.06643407598028e-05, 9.54969927136062e-06, 8.59988522350363e-06, 7.85036481687484e-06, 7.33743522829400e-06, 7.05383748922372e-06, 6.90601416408754e-06, 6.73211257051776e-06, 6.37496974320584e-06, 5.74919351592484e-06, 4.86888357897897e-06, 3.83670346693066e-06, 2.80563963173108e-06, 1.92775672047674e-06, 1.30332822934065e-06, 9.39556048962231e-07, 7.40591563186399e-07, 5.89248930538133e-07, 5.02930006973656e-07, 5.74561919819416e-07, 6.57947065718645e-07, 6.53323866117717e-07, 5.59635932262381e-07, 4.10591852073594e-07, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 3.55271367880050e-15, 7.10542735760100e-15, 3.55271367880050e-15, 7.10542735760100e-15, 3.55271367880050e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 3.55271367880050e-15, 3.55271367880050e-15, 7.10542735760100e-15, 3.55271367880050e-15, 3.55271367880050e-15, 7.10542735760100e-15, 3.55271367880050e-15, 7.10542735760100e-15, 3.55271367880050e-15, 3.55271367880050e-15],
[ 3.10000000000000e+01, 3.09999999767594e+01, 3.09999992581150e+01, 3.09999943891853e+01, 3.09999764903717e+01, 3.09999287778350e+01, 3.09998243556955e+01, 3.09996243578689e+01, 3.09992765069636e+01, 3.09987141450858e+01, 3.09978557771354e+01, 3.09966051517331e+01, 3.09948518888400e+01, 3.09924726470230e+01, 3.09893328077712e+01, 3.09852886398258e+01, 3.09801898936288e+01, 3.09738827650457e+01, 3.09662131586536e+01, 3.09570301741131e+01, 3.09461897342783e+01, 3.09335582704573e+01, 3.09190163782517e+01, 3.09024623563443e+01, 3.08838155402546e+01, 3.08630193434369e+01, 3.08400439193796e+01, 3.08148883610824e+01, 3.07875823590973e+01, 3.07581872469891e+01, 3.07267963742475e+01, 3.06935347618002e+01, 3.06585580143398e+01, 3.06220504862430e+01, 3.05842227229436e+01, 3.05453082258588e+01, 3.05055596146781e+01, 3.04652442842663e+01, 3.04246396729954e+01, 3.03840282737060e+01, 3.03436925268904e+01, 3.03039097377785e+01, 3.02649471550287e+01, 3.02270573393369e+01, 3.01904739364802e+01, 3.01554079522647e+01, 3.01220446077640e+01, 3.00905408332157e+01, 3.00610234389475e+01, 3.00335879824333e+01, 3.00082983325430e+01, 2.99851869155117e+01, 2.99642556122449e+01, 2.99454772632884e+01, 2.99287977260638e+01, 2.99141384187436e+01, 2.99013992763479e+01, 2.98904620373105e+01, 2.98811937729271e+01, 2.98734505678950e+01, 2.98670812577223e+01, 2.98619311283060e+01, 2.98578454846162e+01, 2.98546729992926e+01, 2.98522687581313e+01, 2.98504969278814e+01, 2.98492329823680e+01, 2.98483654354909e+01, 2.98477970437952e+01, 2.98474454566852e+01, 2.98472433084865e+01, 2.98471377629549e+01, 2.98470895369701e+01, 2.98470714455184e+01, 2.98470665241841e+01, 2.98470657978047e+01, 2.98470657025190e+01, 2.98470634824773e+01, 2.98470484413521e+01, 2.98469931483489e+01, 2.98468457556887e+01, 2.98465231812511e+01, 2.98459053773399e+01, 2.98448309007942e+01, 2.98430939648290e+01, 2.98404431133610e+01, 2.98365816150393e+01, 2.98311696277206e+01, 2.98238281358022e+01, 2.98141446138142e+01, 2.98016803211698e+01, 2.97859790860303e+01, 2.97665773916354e+01, 2.97430155365358e+01, 2.97148496008468e+01, 2.96816639135430e+01, 2.96430836806447e+01, 2.95987874012984e+01, 2.95485186698834e+01, 2.94920969408428e+01, 2.94294268241786e+01, 2.93605054899791e+01, 2.92854277964866e+01, 2.92043888228785e+01, 2.91176835864182e+01, 2.90257038503144e+01, 2.89289320745676e+01, 2.88279327138283e+01, 2.87233412079293e+01, 2.86158511266893e+01, 2.85062000084991e+01, 2.83951544652657e+01, 2.82834951142891e+01, 2.81720018465814e+01, 2.80614398616609e+01, 2.79525468038927e+01, 2.78460212377298e+01, 2.77425126092209e+01, 2.76426127658879e+01, 2.75468490497399e+01, 2.74556789386250e+01, 2.73694861867785e+01, 2.72885784024825e+01, 2.72131859950929e+01, 2.71434624216749e+01, 2.70794856623079e+01, 2.70212608509239e+01, 2.69687239843139e+01, 2.69217466253080e+01, 2.68801415071534e+01, 2.68436689351104e+01, 2.68120438687113e+01, 2.67849435545741e+01, 2.67620155658064e+01, 2.67428860906801e+01, 2.67271683013090e+01, 2.67144706235418e+01, 2.67044047232392e+01, 2.66965930225813e+01, 2.66906755639619e+01, 2.66863160490761e+01, 2.66832068973867e+01, 2.66810731912718e+01, 2.66796754043930e+01, 2.66788108443270e+01, 2.66783137790299e+01, 2.66780542577033e+01, 2.66779356783373e+01, 2.66778911948112e+01, 2.66778790941841e+01, 2.66778773081558e+01, 2.66778771504835e+01, 2.66778740609722e+01, 2.66778531290699e+01, 2.66777761810514e+01, 2.66775710648757e+01, 2.66771221680179e+01, 2.66762624571962e+01, 2.66747673514038e+01, 2.66723506979049e+01, 2.66686630727027e+01, 2.66632925720979e+01, 2.66557681996745e+01, 2.66455658827046e+01, 2.66321170732106e+01, 2.66148198020711e+01, 2.65930519610093e+01, 2.65661864898042e+01, 2.65336080489263e+01, 2.64947306668264e+01, 2.64490157733670e+01, 2.63959899740698e+01, 2.63352618914588e+01, 2.62665374059325e+01, 2.61896326729174e+01, 2.61044843754840e+01, 2.60111567877847e+01, 2.59098453659300e+01, 2.58008767372090e+01, 2.56847051123495e+01, 2.55619052862526e+01, 2.54331625113135e+01, 2.52992596202633e+01, 2.51610618440984e+01, 2.50194998205720e+01, 2.48755513260749e+01, 2.47302222918883e+01, 2.45845276829441e+01, 2.44394728163979e+01, 2.42960356692070e+01, 2.41551506615386e+01, 2.40176943060502e+01, 2.38844729905840e+01, 2.37562130296127e+01, 2.36335529965500e+01, 2.35170382506321e+01, 2.34071175072174e+01, 2.33041412694755e+01, 2.32083619362168e+01, 2.31199354152243e+01, 2.30389240938174e+01, 2.29653010405099e+01, 2.28989553284534e+01, 2.28396983806081e+01, 2.27872712379940e+01, 2.27413526468427e+01, 2.27015678494091e+01, 2.26674979481551e+01, 2.26386896954877e+01, 2.26146655427265e+01, 2.25949337639880e+01, 2.25789984547904e+01, 2.25663691930044e+01, 2.25565701429150e+01, 2.25491483830675e+01, 2.25436812464486e+01, 2.25397824781786e+01, 2.25371070415000e+01, 2.25353544370718e+01, 2.25342704424055e+01, 2.25336472260899e+01, 2.25333218431680e+01, 2.25331731712332e+01, 2.25331173989680e+01, 2.25331022275323e+01, 2.25330999882592e+01, 2.25330998086418e+01, 2.25330964937894e+01, 2.25330740351697e+01, 2.25329914747993e+01, 2.25327713983286e+01, 2.25322897621159e+01, 2.25313673549037e+01, 2.25297632301727e+01, 2.25271704015722e+01, 2.25232140433510e+01, 2.25174523798585e+01, 2.25093803819578e+01, 2.24984363125958e+01, 2.24840110781376e+01, 2.24654602465709e+01, 2.24421184897348e+01, 2.24133160974255e+01, 2.23783971015034e+01, 2.23367384447820e+01, 2.22877695408009e+01, 2.22309915057156e+01, 2.21659953113430e+01, 2.20924781162484e+01, 2.20102570839462e+01, 2.19192800938106e+01, 2.18196328859644e+01, 2.17115423458958e+01, 2.15953758135194e+01, 2.14716364788328e+01, 2.13409550874772e+01, 2.12040783138026e+01, 2.10618542616915e+01, 2.09152156253578e+01, 2.07651610882829e+01, 2.06127355637994e+01, 2.04590098890922e+01, 2.03050605758770e+01, 2.01519501935307e+01, 2.00007089112123e+01, 1.98523176535965e+01, 1.97076932327444e+01, 1.95676757126275e+01, 1.94330181519213e+01, 1.93043787647685e+01, 1.91823154471105e+01, 1.90672825440671e+01, 1.89596296845582e+01, 1.88596024824279e+01, 1.87673448956595e+01, 1.86829030420897e+01, 1.86062302859648e+01, 1.85371934295188e+01, 1.84755798631139e+01, 1.84211055430668e+01, 1.83734236759806e+01, 1.83321339912522e+01, 1.82967924794173e+01, 1.82669214639572e+01, 1.82420198595637e+01, 1.82215734525610e+01, 1.82050650214388e+01, 1.81919840996515e+01, 1.81818361713372e+01, 1.81741510855940e+01, 1.81684904782019e+01, 1.81644540024574e+01, 1.81616841936499e+01, 1.81598698244479e+01, 1.81587476500766e+01, 1.81581024909262e+01, 1.81577656538126e+01, 1.81576117488073e+01, 1.81575540135086e+01, 1.81575383080869e+01, 1.81575359899989e+01, 1.81575358112056e+01, 1.81575326006524e+01, 1.81575108486773e+01, 1.81574308860437e+01, 1.81572177344669e+01, 1.81567512546432e+01, 1.81558578792024e+01, 1.81543042581036e+01, 1.81517931028591e+01, 1.81479614683762e+01, 1.81423816563839e+01, 1.81345648611931e+01, 1.81239676057303e+01, 1.81100009325473e+01, 1.80920422208493e+01, 1.80694493977608e+01, 1.80415772029323e+01, 1.80077950548678e+01, 1.79675059614421e+01, 1.79201658238915e+01, 1.78653024118750e+01, 1.78025332458301e+01, 1.77315816196275e+01, 1.76522900371763e+01, 1.75646304237156e+01, 1.74687106045581e+01, 1.73647767150269e+01, 1.72532114046867e+01, 1.71345279122127e+01, 1.70093602971910e+01, 1.68784503037945e+01, 1.67426314820886e+01, 1.66028112929734e+01, 1.64599519655878e+01, 1.63150508614846e+01, 1.61691210351464e+01, 1.60231725781996e+01, 1.58781952110759e+01, 1.57351424573202e+01, 1.55949176163293e+01, 1.54583616497266e+01, 1.53262430193462e+01, 1.51992494606070e+01, 1.50779816401694e+01, 1.49629486257549e+01, 1.48545650834758e+01, 1.47531501097702e+01, 1.46589275985478e+01, 1.45720280384884e+01, 1.44924916306293e+01, 1.44202726128380e+01, 1.43552446756675e+01, 1.42972073531221e+01, 1.42458932711989e+01, 1.42009761355874e+01, 1.41620793364607e+01, 1.41287850420361e+01, 1.41006436431931e+01, 1.40771833992301e+01, 1.40579201207788e+01, 1.40423667114931e+01, 1.40300423773022e+01, 1.40204813028407e+01, 1.40132405911823e+01, 1.40079072669421e+01, 1.40041041554435e+01, 1.40014944725602e+01, 1.39997849909160e+01, 1.39987276874522e+01, 1.39981198233578e+01, 1.39978024578351e+01, 1.39976574496087e+01, 1.39976030518067e+01, 1.39975882542658e+01, 1.39975860701790e+01, 1.39975859048065e+01, 1.39975829752497e+01, 1.39975631270660e+01, 1.39974901629638e+01, 1.39972956668465e+01, 1.39968700137582e+01, 1.39960548253204e+01, 1.39946371686987e+01, 1.39923457591276e+01, 1.39888493821012e+01, 1.39837577008751e+01, 1.39766245569951e+01, 1.39669538052094e+01, 1.39542076487013e+01, 1.39378173562436e+01, 1.39171961508619e+01, 1.38917539626255e+01, 1.38609136405134e+01, 1.38241281257463e+01, 1.37808980084588e+01, 1.37307888286058e+01, 1.36734474476958e+01, 1.36086168161514e+01, 1.35361484953888e+01, 1.34560123647164e+01, 1.33683030483186e+01, 1.32732427314449e+01, 1.31711801898270e+01, 1.30625860234986e+01, 1.29480442565464e+01, 1.28282406290656e+01, 1.27039480580839e+01, 1.25760098714990e+01, 1.24453215135811e+01, 1.23128114726684e+01, 1.21794221833893e+01, 1.20460916036361e+01, 1.19137360643546e+01, 1.17832348501593e+01, 1.16554168098678e+01, 1.15310491401408e+01, 1.14108283519388e+01, 1.12953733309704e+01, 1.11852203433418e+01, 1.10808198119027e+01, 1.09825346880597e+01, 1.08906402575016e+01, 1.08053252373261e+01, 1.07266940403972e+01, 1.06547700974306e+01, 1.05895001375814e+01, 1.05307593346928e+01, 1.04783572296565e+01, 1.04320443400726e+01, 1.03915193667305e+01, 1.03564369021840e+01, 1.03264155396298e+01, 1.03010462703859e+01, 1.02799010458373e+01, 1.02625413656223e+01, 1.02485267393703e+01, 1.02374228561753e+01, 1.02288092860768e+01, 1.02222865330243e+01, 1.02174822607662e+01, 1.02140565230665e+01, 1.02117058482375e+01, 1.02101660551257e+01, 1.02092137126109e+01, 1.02086661959330e+01, 1.02083803387359e+01, 1.02082497272743e+01, 1.02082007302525e+01, 1.02081874018601e+01, 1.02081854346168e+01, 1.02081852869353e+01, 1.02081826875719e+01, 1.02081650764992e+01, 1.02081003362802e+01, 1.02079277622194e+01, 1.02075500860460e+01, 1.02068267830895e+01, 1.02055689284501e+01, 1.02035358345503e+01, 1.02004336641256e+01, 1.01959161684961e+01, 1.01895876498915e+01, 1.01810081876455e+01, 1.01697011005535e+01, 1.01551625416125e+01, 1.01368730378362e+01, 1.01143106993662e+01, 1.00869657326985e+01, 1.00543558080409e+01, 1.00160417572601e+01, 9.97164302378317e+00, 9.92085225606004e+00, 9.86344843733208e+00, 9.79930797970826e+00, 9.72841328005020e+00, 9.65085833559480e+00, 9.56685114208586e+00, 9.47671273758341e+00, 9.38087290116164e+00, 9.27986265786742e+00, 9.17430387164139e+00, 9.06489632063115e+00, 8.95240274077602e+00, 8.83763239150097e+00, 8.72142374027154e+00, 8.60462687920337e+00, 8.48808627560452e+00, 8.37262441868705e+00, 8.25902685774898e+00, 8.14802903628019e+00, 8.04030521769702e+00, 7.93645967999186e+00, 7.83702023796110e+00, 7.74243404220858e+00, 7.65306551175995e+00, 7.56919618739455e+00, 7.49102624831252e+00, 7.41867741520364e+00, 7.35219696540941e+00, 7.29156260615939e+00, 7.23668798436070e+00, 7.18742865047368e+00, 7.14358833437684e+00, 7.10492542839903e+00, 7.07115960354000e+00, 7.04197850714105e+00, 7.01704450288270e+00, 6.99600141701571e+00, 6.97848124913779e+00, 6.96411079333503e+00, 6.95251809840730e+00, 6.94333867685092e+00, 6.93622135407037e+00, 6.93083363462969e+00, 6.92686645359197e+00, 6.92403817994621e+00, 6.92209774688360e+00, 6.92082680054109e+00, 6.92004078421788e+00, 6.91958890765171e+00, 6.91935298874472e+00, 6.91924519576374e+00, 6.91920475894631e+00, 6.91919375915635e+00, 6.91919213561056e+00, 6.91919201410110e+00, 6.91918988033394e+00, 6.91917542375158e+00, 6.91912227988729e+00, 6.91898061833107e+00, 6.91867059788655e+00, 6.91807688087060e+00, 6.91704443087574e+00, 6.91537579480003e+00, 6.91283004244909e+00, 6.90912350430691e+00, 6.90393240853282e+00, 6.89689747019204e+00, 6.88763042782984e+00, 6.87572245420338e+00, 6.86075428992197e+00, 6.84230786304648e+00, 6.81997906815471e+00, 6.79339129040660e+00, 6.76220918051800e+00, 6.72615212290930e+00, 6.68500679941702e+00, 6.63863824192005e+00, 6.58699879447465e+00, 6.53013447199296e+00, 6.46818830781011e+00, 6.40140042262706e+00, 6.33010471445672e+00, 6.25472225206423e+00, 6.17575163909700e+00, 6.09375678743781e+00, 6.00935268138531e+00, 5.92318981626937e+00, 5.83593804702926e+00, 5.74827058018591e+00, 5.66084878829660e+00, 5.57430842666415e+00, 5.48924769952953e+00, 5.40621747172641e+00, 5.32571376708227e+00, 5.24817255071367e+00, 5.17396666991287e+00, 5.10340473482394e+00, 5.03673165868041e+00, 4.97413054736094e+00, 4.91572562577466e+00, 4.86158590855517e+00, 4.81172935827806e+00, 4.76612731950869e+00, 4.72470906568814e+00, 4.68736634348160e+00, 4.65395784224178e+00, 4.62431355232770e+00, 4.59823900382460e+00, 4.57551939620959e+00, 4.55592363981368e+00, 4.53920833213786e+00, 4.52512168715439e+00, 4.51340742492298e+00, 4.50380861367683e+00, 4.49607143867902e+00, 4.48994885344947e+00, 4.48520405130062e+00, 4.48161368032853e+00, 4.47897071472987e+00, 4.47708689086820e+00, 4.47579461874544e+00, 4.47494828873256e+00, 4.47442490924737e+00, 4.47412403261039e+00, 4.47396695212458e+00, 4.47389518172002e+00, 4.47386825831988e+00, 4.47386093451701e+00, 4.47385985353959e+00, 4.47385977260675e+00, 4.47385835098127e+00, 4.47384871926711e+00, 4.47381331218168e+00, 4.47371893077275e+00, 4.47351238383882e+00, 4.47311683888351e+00, 4.47242903650102e+00, 4.47131750894257e+00, 4.46962192793611e+00, 4.46715368723991e+00, 4.46369780011326e+00, 4.45901615898806e+00, 4.45285216252162e+00, 4.44493666303624e+00, 4.43499512541918e+00, 4.42275581873442e+00, 4.40795878773230e+00, 4.39036527855986e+00, 4.36976722818949e+00, 4.34599637819467e+00, 4.31893254828613e+00, 4.28851061012348e+00, 4.25472574166285e+00, 4.21763661766202e+00, 4.17736629985194e+00, 4.13410072339784e+00, 4.08808482360408e+00, 4.03961649481023e+00, 3.98903870768804e+00, 3.93673021842269e+00, 3.88309537332460e+00, 3.82855353955902e+00, 3.77352867643264e+00, 3.71843950670037e+00, 3.66369066247234e+00, 3.60966507693649e+00, 3.55671778343821e+00, 3.50517117865814e+00, 3.45531171553902e+00, 3.40738791996714e+00, 3.36160957550900e+00, 3.31814789237075e+00, 3.27713646768473e+00, 3.23867285043851e+00, 3.20282054156874e+00, 3.16961128380860e+00, 3.13904752322500e+00, 3.11110495217795e+00, 3.08573506962500e+00, 3.06286771790694e+00, 3.04241357456817e+00, 3.02426659296694e+00, 3.00830639628014e+00, 2.99440063605927e+00, 2.98240732894563e+00, 2.97217718382364e+00, 2.96355592702318e+00, 2.95638662574956e+00, 2.95051200043581e+00, 2.94577670602229e+00, 2.94202955123010e+00, 2.93912561472864e+00, 2.93692820872244e+00, 2.93531063483309e+00, 2.93415767498660e+00, 2.93336676183168e+00, 2.93284877918013e+00, 2.93252845288425e+00, 2.93234430589590e+00, 2.93224816713185e+00, 2.93220424112088e+00, 2.93218776305220e+00, 2.93218328062691e+00, 2.93218261903065e+00, 2.93218256936552e+00, 2.93218169521742e+00, 2.93217577274019e+00, 2.93215400115808e+00, 2.93209596665669e+00, 2.93196896232540e+00, 2.93172574439587e+00, 2.93130281941378e+00, 2.93061934889582e+00, 2.92957674836839e+00, 2.92805904559409e+00, 2.92593404714022e+00, 2.92305534203594e+00, 2.91926514509163e+00, 2.91439794997353e+00, 2.90828492356271e+00, 2.90075892971526e+00, 2.89166002467959e+00, 2.88084122167086e+00, 2.86817428295107e+00, 2.85355526921255e+00, 2.83690956296695e+00, 2.81819608892689e+00, 2.79741048224879e+00, 2.77458700485345e+00, 2.74979907805510e+00, 2.72315838105061e+00, 2.69481255209379e+00, 2.66494161403390e+00, 2.63375332013422e+00, 2.60147767287939e+00, 2.56836090331241e+00, 2.53465920962221e+00, 2.50063254235278e+00, 2.46653869319875e+00, 2.43262789991411e+00, 2.39913812707045e+00, 2.36629112676361e+00, 2.33428932961040e+00, 2.30331356812205e+00, 2.27352159421245e+00, 2.24504732152255e+00, 2.21800070181029e+00, 2.19246813253111e+00, 2.16851328902687e+00, 2.14617827816361e+00, 2.12548501927107e+00, 2.10643677117979e+00, 2.08901973937500e+00, 2.07320471326552e+00, 2.05894869899353e+00, 2.04619652705098e+00, 2.03488242548355e+00, 2.02493155820077e+00, 2.01626153367422e+00, 2.00878389211043e+00, 2.00240557922031e+00, 1.99703041229786e+00, 1.99256053990415e+00, 1.98889789055918e+00, 1.98594559907422e+00, 1.98360939216885e+00, 1.98179890848789e+00, 1.98042892272613e+00, 1.97942043988295e+00, 1.97870162418218e+00, 1.97820852821993e+00, 1.97788559154631e+00, 1.97768588402300e+00, 1.97757107758002e+00, 1.97751113988079e+00, 1.97748375421868e+00, 1.97747348097086e+00, 1.97747068640397e+00, 1.97747027393196e+00, 1.97747024283637e+00, 1.97746969377125e+00, 1.97746597377736e+00, 1.97745229873195e+00, 1.97741584643222e+00, 1.97733607324468e+00, 1.97718330486317e+00, 1.97691766069403e+00, 1.97648836635108e+00, 1.97583350272792e+00, 1.97488023256762e+00, 1.97354553565671e+00, 1.97173747097432e+00, 1.96935696768404e+00, 1.96630012641817e+00, 1.96246098799891e+00, 1.95773469932725e+00, 1.95202097713155e+00, 1.94522774181087e+00, 1.93727476853557e+00, 1.92809718421777e+00, 1.91764863002303e+00, 1.90590391232727e+00, 1.89286098199269e+00, 1.87854211269997e+00, 1.86299419234325e+00, 1.84628809403483e+00, 1.82851715056346e+00, 1.80979481282188e+00, 1.79025162320744e+00, 1.77003167432088e+00, 1.74928874773492e+00, 1.72818233525320e+00, 1.70687373604094e+00, 1.68552239935063e+00, 1.66428264793942e+00, 1.64330087630446e+00, 1.62271327540959e+00, 1.60264409603990e+00, 1.58320442966995e+00, 1.56449146080718e+00, 1.54658812883403e+00, 1.52956312990056e+00, 1.51347118905341e+00, 1.49835353772857e+00, 1.48423854010273e+00, 1.47114242190332e+00, 1.45907006576550e+00, 1.44801584713122e+00, 1.43796449340045e+00, 1.42889195624433e+00, 1.42076629255485e+00, 1.41354855344569e+00, 1.40719368311643e+00, 1.40165143036788e+00, 1.39686727524884e+00, 1.39278337187957e+00, 1.38933950611136e+00, 1.38647406355067e+00, 1.38412499984381e+00, 1.38223080126428e+00, 1.38073141988709e+00, 1.37956916430571e+00, 1.37868952428468e+00, 1.37804190624791e+00, 1.37758025632470e+00, 1.37726354896724e+00, 1.37705612195555e+00, 1.37692784283210e+00, 1.37685409724840e+00, 1.37681559603825e+00, 1.37679800466129e+00, 1.37679140555176e+00, 1.37678961043617e+00, 1.37678934548094e+00, 1.37678932539770e+00, 1.37678896933942e+00, 1.37678655699357e+00, 1.37677768897599e+00, 1.37675405025825e+00, 1.37670231832060e+00, 1.37660324860519e+00, 1.37643097534379e+00, 1.37615256226906e+00, 1.37572783349184e+00, 1.37510950963845e+00, 1.37424366779354e+00, 1.37307053546394e+00, 1.37152561831240e+00, 1.36954114862833e+00, 1.36704782651356e+00, 1.36397680905305e+00, 1.36026188524376e+00, 1.35584175753333e+00, 1.35066233619176e+00, 1.34467894229872e+00, 1.33785831069606e+00, 1.33018028726634e+00, 1.32163912610713e+00, 1.31224431140716e+00, 1.30202085489374e+00, 1.29100905044823e+00, 1.27926370000196e+00, 1.26685285594326e+00, 1.25385615196825e+00, 1.24036281419258e+00, 1.22646945595933e+00, 1.21227776278089e+00, 1.19789216892722e+00, 1.18341761581933e+00, 1.16895746657722e+00, 1.15461163290493e+00, 1.14047495189461e+00, 1.12663583285017e+00, 1.11317517895501e+00, 1.10016557616676e+00, 1.08767073234298e+00, 1.07574514321719e+00, 1.06443395817765e+00, 1.05377301745870e+00, 1.04378903288415e+00, 1.03449988624088e+00, 1.02591502226549e+00, 1.01803591668940e+00, 1.01085660345201e+00, 1.00436424875319e+00, 9.98539762827658e-01, 9.93358442992797e-01, 9.88790643509663e-01, 9.84802469018700e-01, 9.81356488736200e-01, 9.78412468242848e-01, 9.75928114634538e-01, 9.73859829158506e-01, 9.72163459391584e-01, 9.70795040739027e-01, 9.69711514776421e-01, 9.68871409971705e-01, 9.68235468853472e-01, 9.67767204951361e-01, 9.67433372993934e-01, 9.67204337009875e-01, 9.67054323160546e-01, 9.66961547268561e-01, 9.66908210947047e-01, 9.66880364755575e-01, 9.66867641640668e-01, 9.66862868767485e-01, 9.66861570430869e-01, 9.66861378799123e-01, 9.66861364181526e-01, 9.66861103808448e-01, 9.66859339743120e-01, 9.66852854861642e-01, 9.66835568619683e-01, 9.66797738462979e-01, 9.66725290570466e-01, 9.66599307405556e-01, 9.66395696735539e-01, 9.66085063471489e-01, 9.65632801708495e-01, 9.64999419477994e-01, 9.64141102718489e-01, 9.63010517628742e-01, 9.61557841799277e-01, 9.59732004408700e-01, 9.57482104634170e-01, 9.54758965834895e-01, 9.51516771854219e-01, 9.47714721989634e-01, 9.43318633948088e-01, 9.38302420551665e-01, 9.32649366995996e-01, 9.26353141627040e-01, 9.19418484513773e-01, 9.11861533991726e-01, 9.03709770704648e-01, 8.95001579889072e-01, 8.85785453877169e-01, 8.76118876162277e-01, 8.66066944230095e-01, 8.55700799490446e-01, 8.45095938397918e-01, 8.34330479176216e-01, 8.23483453964130e-01, 8.12633187612970e-01, 8.01855812982062e-01, 7.91223959687507e-01, 7.80805640080040e-01, 7.70663343795562e-01, 7.60853341313283e-01, 7.51425188065664e-01, 7.42421413995469e-01, 7.33877379045438e-01, 7.25821272720271e-01, 7.18274235286511e-01, 7.11250579019982e-01, 7.04758089794787e-01, 6.98798391868355e-01, 6.93367361619288e-01, 6.88455578950009e-01, 6.84048807833634e-01, 6.80128499875110e-01, 6.76672316629741e-01, 6.73654667687259e-01, 6.71047262143887e-01, 6.68819671052244e-01, 6.66939897809506e-01, 6.65374952311613e-01, 6.64091423197419e-01, 6.63056040796729e-01, 6.62236221669893e-01, 6.61600584081246e-01, 6.61119422581507e-01, 6.60765129256191e-01, 6.60512549264080e-01, 6.60339259122445e-01, 6.60225757811791e-01, 6.60155563121521e-01, 6.60115208624116e-01, 6.60094140078677e-01, 6.60084513717983e-01, 6.60080902542880e-01, 6.60079920216177e-01, 6.60079775226833e-01, 6.60079764089058e-01, 6.60079564676430e-01, 6.60078213627010e-01, 6.60073247034099e-01, 6.60060007979837e-01, 6.60031034932462e-01, 6.59975549226612e-01, 6.59879062704533e-01, 6.59723124460804e-01, 6.59485223123362e-01, 6.59138858084734e-01, 6.58653789390439e-01, 6.57996471407898e-01, 6.57130669785900e-01, 6.56018254499497e-01, 6.54620154022865e-01, 6.52897447116572e-01, 6.50812559791092e-01, 6.48330526340850e-01, 6.45420265719249e-01, 6.42055818817308e-01, 6.38217489269927e-01, 6.33892830956000e-01, 6.29077429820189e-01, 6.23775436087925e-01, 6.17999814994551e-01, 6.11772298993031e-01, 6.05123040866680e-01, 5.98089983850876e-01, 5.90717980307041e-01, 5.83057703375957e-01, 5.75164405350957e-01, 5.67096581654258e-01, 5.58914600133896e-01, 5.50679352224641e-01, 5.42450975993077e-01, 5.34287692113278e-01, 5.26244783402863e-01, 5.18373737680093e-01, 5.10721563258265e-01, 5.03330277067377e-01, 4.96236557665476e-01, 4.89471549536727e-01, 4.83060801134826e-01, 4.77024317021662e-01, 4.71376703953236e-01, 4.66127391580830e-01, 4.61280910233352e-01, 4.56837210689542e-01, 4.52792013622865e-01, 4.49137179231503e-01, 4.45861090222152e-01, 4.42949043623120e-01, 4.40383648732254e-01, 4.38145229782211e-01, 4.36212232595611e-01, 4.34561634616077e-01, 4.33169357283337e-01, 4.32010678854130e-01, 4.31060644566685e-01, 4.30294469641449e-01, 4.29687929160114e-01, 4.29217727530151e-01, 4.28861839182595e-01, 4.28599811507890e-01, 4.28413020920100e-01, 4.28284873422360e-01, 4.28200942145676e-01, 4.28149036012215e-01, 4.28119195847110e-01, 4.28103616796896e-01, 4.28096498647033e-01, 4.28093828390846e-01, 4.28093102017137e-01, 4.28092994805926e-01, 4.28092986507338e-01, 4.28092837109748e-01, 4.28091824919644e-01, 4.28088104010424e-01, 4.28078185498936e-01, 4.28056479416011e-01, 4.28014910961323e-01, 4.27942627109221e-01, 4.27825808032352e-01, 4.27647596036351e-01, 4.27388152530479e-01, 4.27024850836418e-01, 4.26532609193078e-01, 4.25884364010116e-01, 4.25051678186129e-01, 4.24005473181204e-01, 4.22716866703727e-01, 4.21158090690945e-01, 4.19303457247779e-01, 4.17130334011858e-01, 4.14620085760689e-01, 4.11758936686521e-01, 4.08538708235191e-01, 4.04957391114401e-01, 4.01019517084709e-01, 3.96736306147687e-01, 3.92125577060451e-01, 3.87211422742801e-01, 3.82023665908044e-01, 3.76597122893308e-01, 3.70970714053112e-01, 3.65186466331366e-01, 3.59288457226523e-01, 3.53321749212952e-01, 3.47331360079409e-01, 3.41361308225835e-01, 3.35453763576222e-01, 3.29648325364875e-01, 3.23981438553108e-01, 3.18485951821010e-01, 3.13190812545699e-01, 3.08120888292951e-01, 3.03296900258363e-01, 2.98735451756265e-01, 2.94449134081223e-01, 2.90446692580439e-01, 2.86733237255373e-01, 2.83310484333903e-01, 2.80177017723688e-01, 2.77328561817370e-01, 2.74758259566518e-01, 2.72456951920393e-01, 2.70413456532332e-01, 2.68614845007079e-01, 2.67046718867384e-01, 2.65693484857959e-01, 2.64538630203393e-01, 2.63564998040149e-01, 2.62755062517577e-01, 2.62091202093437e-01, 2.61555968435111e-01, 2.61132347190475e-01, 2.60804005828692e-01, 2.60555522886047e-01, 2.60372592388254e-01, 2.60242197042082e-01, 2.60152744046859e-01, 2.60094158088400e-01, 2.60057927219240e-01, 2.60037098839057e-01, 2.60026224771426e-01, 2.60021256371232e-01, 2.60019392561070e-01, 2.60018885560289e-01, 2.60018810728067e-01, 2.60018804888784e-01, 2.60018699158877e-01, 2.60017982823779e-01, 2.60015349508642e-01, 2.60008330122343e-01, 2.59992968697119e-01, 2.59963550994609e-01, 2.59912397319923e-01, 2.59829730004562e-01, 2.59703624849124e-01, 2.59520054353157e-01, 2.59263028668435e-01, 2.58914837750392e-01, 2.58456395029363e-01, 2.57867679014719e-01, 2.57128264605949e-01, 2.56217930661979e-01, 2.55117324856478e-01, 2.53808661443141e-01, 2.52276422803757e-01, 2.50508032148510e-01, 2.48494463063391e-01, 2.46230752234577e-01, 2.43716384907227e-01, 2.40955528470824e-01, 2.37957097705486e-01, 2.34734645063664e-01, 2.31306080041779e-01, 2.27693232209997e-01, 2.23921281800813e-01, 2.20018089025853e-01, 2.16013457866513e-01, 2.11938371674453e-01, 2.07824236582447e-01, 2.03702164849293e-01, 1.99602324466646e-01, 1.95553374398290e-01, 1.91581997478735e-01, 1.87712535963328e-01, 1.83966728540515e-01, 1.80363542644677e-01, 1.76919092310310e-01, 1.73646629586170e-01, 1.70556596554302e-01, 1.67656725067303e-01, 1.64952172181454e-01, 1.62445680674423e-01, 1.60137755763259e-01, 1.58026850988673e-01, 1.56109558050577e-01, 1.54380797054202e-01, 1.52834005077054e-01, 1.51461322142900e-01, 1.50253774562643e-01, 1.49201456161020e-01, 1.48293708159123e-01, 1.47519298441424e-01, 1.46866600634233e-01, 1.46323772902083e-01, 1.45878935683864e-01, 1.45520346807725e-01, 1.45236571616017e-01, 1.45016644978082e-01, 1.44850221448106e-01, 1.44727709408354e-01, 1.44640384883147e-01, 1.44580480853236e-01, 1.44541248356585e-01, 1.44516986414700e-01, 1.44503038832447e-01, 1.44495757118875e-01, 1.44492430085204e-01, 1.44491182006711e-01, 1.44490842499732e-01, 1.44490792389238e-01, 1.44490788445761e-01, 1.44490716616271e-01, 1.44490229961384e-01, 1.44488440973316e-01, 1.44483672240248e-01, 1.44473236249791e-01, 1.44453251095900e-01, 1.44418499925660e-01, 1.44362341332602e-01, 1.44276677145989e-01, 1.44151983103416e-01, 1.43977406624664e-01, 1.43740934221196e-01, 1.43429628885688e-01, 1.43029935059033e-01, 1.42528045486003e-01, 1.41910320550583e-01, 1.41163746739164e-01, 1.40276417024893e-01, 1.39238012596152e-01, 1.38040262905765e-01, 1.36677359920597e-01, 1.35146303046685e-01, 1.33447153682660e-01, 1.31583182695827e-01, 1.29560900066885e-01, 1.27389963033234e-01, 1.25082966625229e-01, 1.22655127797969e-01, 1.20123880693125e-01, 1.17508405327431e-01, 1.14829114808771e-01, 1.12107126892013e-01, 1.09363744421380e-01, 1.06619966289276e-01, 1.03896046436962e-01, 1.01211113651605e-01, 9.85828599815193e-02, 9.60273009274353e-02, 9.35586064899496e-02, 9.11889988608898e-02, 8.89287101177381e-02, 8.67859917067584e-02, 8.47671667047649e-02, 8.28767157213086e-02, 8.11173877117035e-02, 7.94903277842991e-02, 7.79952151723222e-02, 7.66304057834439e-02, 7.53930750346399e-02, 7.42793579408882e-02, 7.32844845840823e-02, 7.24029100909647e-02, 7.16284390543827e-02, 7.09543449161458e-02, 7.03734851778108e-02, 6.98784134204686e-02, 6.94614890089760e-02, 6.91149850596915e-02, 6.88311948006266e-02, 6.86025358997711e-02, 6.84216517370473e-02, 6.82815080095978e-02, 6.81754825495275e-02, 6.80974458550025e-02, 6.80418296347085e-02, 6.80036806735203e-02, 6.79786975563168e-02, 6.79632482290877e-02, 6.79543670033134e-02, 6.79497303780749e-02, 6.79476119056694e-02, 6.79468171999922e-02, 6.79466010213901e-02, 6.79465691139001e-02, 6.79465665806980e-02, 6.79465201564575e-02, 6.79462056257876e-02, 6.79450493839866e-02, 6.79419673152921e-02, 6.79352225240315e-02, 6.79223063642276e-02, 6.78998479402075e-02, 6.78635569052339e-02, 6.78082039620058e-02, 6.77276429054920e-02, 6.76148772645568e-02, 6.74621735089325e-02, 6.72612213200061e-02, 6.70033395337768e-02, 6.66797240617569e-02, 6.62817314535005e-02, 6.58011889347598e-02, 6.52307189634698e-02, 6.45640638810505e-02, 6.37963944204714e-02, 6.29245849781768e-02, 6.19474389193790e-02, 6.08658489118561e-02, 5.96828803705804e-02, 5.84037703738467e-02, 5.70358395490231e-02, 5.55883199642224e-02, 5.40721074754238e-02, 5.24994517446231e-02, 5.08836008214075e-02, 4.92384194510684e-02, 4.75780009942000e-02, 4.59162920341178e-02, 4.42667465929887e-02, 4.26420236729506e-02, 4.10537379674096e-02, 3.95122694489345e-02, 3.80266335102490e-02, 3.66044097167730e-02, 3.52517242420838e-02, 3.39732788330132e-02, 3.27724177444537e-02, 3.16512234769775e-02, 3.06106322613537e-02, 2.96505609234892e-02, 2.87700378581803e-02, 2.79673321718619e-02, 2.72400764804104e-02, 2.65853802588865e-02, 2.59999319497730e-02, 2.54800891832403e-02, 2.50219574107184e-02, 2.46214579644359e-02, 2.42743870320865e-02, 2.39764672940801e-02, 2.37233940252324e-02, 2.35108773509971e-02, 2.33346820914271e-02, 2.31906662597012e-02, 2.30748188375358e-02, 2.29832969683308e-02, 2.29124622294066e-02, 2.28589152057666e-02, 2.28195272343434e-02, 2.27914679374095e-02, 2.27722270558850e-02, 2.27596291255949e-02, 2.27518397133757e-02, 2.27473622357088e-02, 2.27450247804626e-02, 2.27439568216496e-02, 2.27435562004885e-02, 2.27434472225063e-02, 2.27434311376129e-02, 2.27434298489193e-02, 2.27434060845084e-02, 2.27432450779297e-02, 2.27426532064108e-02, 2.27410755418056e-02, 2.27376231012180e-02, 2.27310121837689e-02, 2.27195186190805e-02, 2.27009496846158e-02, 2.26726363374823e-02, 2.26314482129801e-02, 2.25738334901884e-02, 2.24958851396870e-02, 2.23934341973724e-02, 2.22621694918494e-02, 2.20977817068650e-02, 2.18961278301535e-02, 2.16534100334762e-02, 2.13663610422508e-02, 2.10324262954593e-02, 2.06499319459823e-02, 2.02182272500870e-02, 1.97377903522664e-02, 1.92102879928946e-02, 1.86385822374966e-02, 1.80266807807463e-02, 1.73796314273318e-02, 1.67033656189233e-02, 1.60044998789886e-02, 1.52901073387035e-02, 1.45674737682313e-02, 1.38438535278276e-02, 1.31262402339658e-02, 1.24211645595018e-02, 1.17345281004795e-02, 1.10714784820267e-02, 1.04363271922985e-02, 9.83250988267415e-03, 9.26259048970834e-03, 8.72830799289304e-03, 8.23065193543826e-03, 7.76994425420696e-03, 7.34591504120274e-03, 6.95778107375489e-03, 6.60434698208334e-03, 6.28412527581546e-03, 5.99544278320252e-03, 5.73651288714672e-03, 5.50547912853716e-03, 5.30044526657036e-03, 5.11950320456478e-03, 4.96076214176711e-03, 4.82237548637698e-03, 4.70256054509122e-03, 4.59960915317836e-03, 4.51189114378536e-03, 4.43785401676422e-03, 4.37602159990291e-03, 4.32499321805935e-03, 4.28344386889279e-03, 4.25012542201841e-03, 4.22386874041436e-03, 4.20358661661169e-03, 4.18827736098137e-03, 4.17702874671022e-03, 4.16902188640833e-03, 4.16353450017937e-03, 4.15994300584188e-03, 4.15772289115424e-03, 4.15644691995354e-03, 4.15578085658197e-03, 4.15547655138226e-03, 4.15536240010894e-03, 4.15533134847834e-03, 4.15532676541019e-03, 4.15532639468452e-03, 4.15531951473724e-03, 4.15527290286022e-03, 4.15510155618293e-03, 4.15464483577210e-03, 4.15364545222019e-03, 4.15173204104136e-03, 4.14840625324331e-03, 4.14303532048876e-03, 4.13485107805656e-03, 4.12295641121929e-03, 4.10633998149379e-03, 4.08389986342428e-03, 4.05447634939549e-03, 4.01689367849301e-03, 3.97000985345386e-03, 3.91277309110194e-03, 3.84428278227844e-03, 3.76385205831298e-03, 3.67106797881789e-03, 3.56584394640791e-03, 3.44845750501577e-03, 3.31956614891382e-03, 3.18019584625873e-03, 3.03170299791589e-03, 2.87571856555147e-03, 2.71408672206874e-03, 2.54880526471624e-03, 2.38196540551883e-03, 2.21568147460616e-03, 2.05199958996261e-03, 1.89278266646298e-03, 1.73959384573408e-03, 1.59362591161383e-03, 1.45570780702542e-03, 1.32636776610795e-03, 1.20590806188758e-03, 1.09446031764548e-03, 9.92011343186050e-04, 8.98404533021591e-04, 8.13331517306892e-04, 7.36330646461170e-04, 6.66784633558848e-04, 6.03892181853638e-04, 5.46787231836987e-04, 4.94746192416073e-04, 4.47263380660701e-04, 4.03785721658778e-04, 3.63718732444340e-04, 3.26667670154990e-04, 2.92343944067852e-04, 2.60487225737904e-04, 2.30869619954888e-04, 2.03305608167170e-04, 1.78025161336137e-04, 1.54599489562202e-04, 1.33049427049059e-04, 1.13349319693867e-04, 9.54788116800387e-05, 7.94140986819514e-05, 6.51234507031972e-05, 5.25751149069720e-05, 4.16977635033788e-05, 3.23864150821862e-05, 2.45473217006520e-05, 1.80764535038236e-05, 1.28576514626388e-05, 8.76442467045990e-06, 5.66228609599762e-06, 3.41132022541046e-06, 1.86891355369312e-06, 8.92651755179941e-07, 3.43455909046497e-07, -1.77635683940025e-14, -1.77635683940025e-14],
[ 3.20000000000000e+01, 3.19999999756516e+01, 3.19999992227504e+01, 3.19999941217262e+01, 3.19999753697024e+01, 3.19999253827773e+01, 3.19998159829726e+01, 3.19996064514716e+01, 3.19992420187673e+01, 3.19986528491534e+01, 3.19977535620427e+01, 3.19964433164065e+01, 3.19946064677853e+01, 3.19921137904565e+01, 3.19888242410736e+01, 3.19845872250099e+01, 3.19792453132186e+01, 3.19726373460254e+01, 3.19646018510616e+01, 3.19549806955148e+01, 3.19436228878304e+01, 3.19303884405989e+01, 3.19151522042334e+01, 3.18978075798017e+01, 3.18782700188159e+01, 3.18564802178944e+01, 3.18324069172724e+01, 3.18060492147047e+01, 3.17774383111264e+01, 3.17466386123561e+01, 3.17137481228667e+01, 3.16788980836753e+01, 3.16422518267252e+01, 3.16040028422365e+01, 3.15643720823240e+01, 3.15236045521972e+01, 3.14819652676527e+01, 3.14397346824502e+01, 3.13972037097892e+01, 3.13546684771194e+01, 3.13124249620490e+01, 3.12707636589041e+01, 3.12299644208129e+01, 3.11902916118158e+01, 3.11519896885290e+01, 3.11152793125894e+01, 3.10803540747905e+01, 3.10473778906531e+01, 3.10164831061664e+01, 3.09877693323129e+01, 3.09613030082589e+01, 3.09371176760080e+01, 3.09152149339777e+01, 3.08955660233310e+01, 3.08781139889008e+01, 3.08627763460798e+01, 3.08494481760547e+01, 3.08380055642392e+01, 3.08283092907525e+01, 3.08202086774392e+01, 3.08135454933840e+01, 3.08081578203353e+01, 3.08038837810982e+01, 3.08005650379206e+01, 3.07980499742579e+01, 3.07961964820411e+01, 3.07948742875763e+01, 3.07939667622407e+01, 3.07933721788955e+01, 3.07930043909820e+01, 3.07927929281330e+01, 3.07926825192832e+01, 3.07926320711555e+01, 3.07926131460926e+01, 3.07926079979950e+01, 3.07926072381458e+01, 3.07926071384697e+01, 3.07926048161338e+01, 3.07925890819503e+01, 3.07925312411856e+01, 3.07923770570850e+01, 3.07920396195232e+01, 3.07913933500976e+01, 3.07902693691144e+01, 3.07884524136592e+01, 3.07856794546824e+01, 3.07816401156675e+01, 3.07759789469513e+01, 3.07682995592354e+01, 3.07581705684511e+01, 3.07451332531214e+01, 3.07287107757979e+01, 3.07084187728215e+01, 3.06837770719589e+01, 3.06543222553086e+01, 3.06196207449209e+01, 3.05792820506291e+01, 3.05329717841854e+01, 3.04804240129057e+01, 3.04214525034984e+01, 3.03559603983437e+01, 3.02839478792192e+01, 3.02055174142179e+01, 3.01208762572630e+01, 3.00303359771295e+01, 2.99343089297874e+01, 2.98333017440441e+01, 2.97279060510471e+01, 2.96187868359960e+01, 2.95066689089443e+01, 2.93923220682195e+01, 2.92765455583427e+01, 2.91601524051488e+01, 2.90439541514400e+01, 2.89287464289929e+01, 2.88152957011633e+01, 2.87043274081415e+01, 2.85965156547809e+01, 2.84924745055872e+01, 2.83927508955368e+01, 2.82978191280028e+01, 2.82080769090039e+01, 2.81238428558928e+01, 2.80453554140918e+01, 2.79727731138307e+01, 2.79061760973424e+01, 2.78455688439663e+01, 2.77908840152845e+01, 2.77419873346164e+01, 2.76986834050880e+01, 2.76607223584824e+01, 2.76278072136505e+01, 2.75996018089853e+01, 2.75757391589992e+01, 2.75558300711754e+01, 2.75394718468853e+01, 2.75262568803078e+01, 2.75157809630337e+01, 2.75076511004548e+01, 2.75014926500969e+01, 2.74969556024786e+01, 2.74937198422946e+01, 2.74914992517459e+01, 2.74900445482336e+01, 2.74891447845274e+01, 2.74886274796115e+01, 2.74883573910836e+01, 2.74882339834082e+01, 2.74881876886057e+01, 2.74881750952665e+01, 2.74881732365149e+01, 2.74881730724225e+01, 2.74881698571128e+01, 2.74881480729076e+01, 2.74880679917298e+01, 2.74878545236749e+01, 2.74873873487961e+01, 2.74864926328500e+01, 2.74849366512019e+01, 2.74824216017010e+01, 2.74785838372932e+01, 2.74729946920678e+01, 2.74651640096410e+01, 2.74545464095620e+01, 2.74405502453818e+01, 2.74225491175098e+01, 2.73998957064330e+01, 2.73719375902089e+01, 2.73380346088759e+01, 2.72975772436958e+01, 2.72500053983407e+01, 2.71948269103229e+01, 2.71316350918852e+01, 2.70601246064712e+01, 2.69801050332757e+01, 2.68915115578578e+01, 2.67944123465028e+01, 2.66890123067433e+01, 2.65756530937859e+01, 2.64548093790548e+01, 2.63270815406836e+01, 2.61931850586114e+01, 2.60539369964352e+01, 2.59102400308519e+01, 2.57630645526872e+01, 2.56134294155768e+01, 2.54623819494173e+01, 2.53109778798265e+01, 2.51602617915519e+01, 2.50112487324367e+01, 2.48649074701467e+01, 2.47221457911949e+01, 2.45837980859999e+01, 2.44506153163618e+01, 2.43232573342611e+01, 2.42022874282910e+01, 2.40881689215255e+01, 2.39812636280088e+01, 2.38818319840729e+01, 2.37900346932511e+01, 2.37059357491427e+01, 2.36295067220513e+01, 2.35606322089637e+01, 2.34991163516058e+01, 2.34446903246933e+01, 2.33970206875962e+01, 2.33557184790614e+01, 2.33203489178677e+01, 2.32904415536735e+01, 2.32655006931360e+01, 2.32450159080056e+01, 2.32284724157984e+01, 2.32153611114005e+01, 2.32051880211810e+01, 2.31974829513770e+01, 2.31918071109056e+01, 2.31877595061711e+01, 2.31849819321362e+01, 2.31831624195232e+01, 2.31820370414771e+01, 2.31813900326813e+01, 2.31810522276080e+01, 2.31808978798137e+01, 2.31808399783255e+01, 2.31808242276883e+01, 2.31808219029262e+01, 2.31808217164515e+01, 2.31808182750477e+01, 2.31807949590228e+01, 2.31807092467352e+01, 2.31804807683713e+01, 2.31799807446229e+01, 2.31790231223200e+01, 2.31773577558728e+01, 2.31746659379891e+01, 2.31705585307061e+01, 2.31645768876703e+01, 2.31561966898224e+01, 2.31448347381518e+01, 2.31298586583032e+01, 2.31105993727363e+01, 2.30863660883705e+01, 2.30564634344188e+01, 2.30202102716170e+01, 2.29769595873598e+01, 2.29261188000034e+01, 2.28671697291637e+01, 2.27996874562870e+01, 2.27233573083617e+01, 2.26379892513998e+01, 2.25435290787326e+01, 2.24400659165286e+01, 2.23278357346264e+01, 2.22072207306334e+01, 2.20787446340675e+01, 2.19430641416338e+01, 2.18009568352415e+01, 2.16533060470082e+01, 2.15010832207526e+01, 2.13453283802461e+01, 2.11871293534229e+01, 2.10276004194029e+01, 2.08678610396340e+01, 2.07090153026951e+01, 2.05521326523374e+01, 2.03982303813780e+01, 2.02482582653964e+01, 2.01030855888556e+01, 1.99634906933657e+01, 1.98301530643718e+01, 1.97036478775007e+01, 1.95844428546792e+01, 1.94728972346120e+01, 1.93692626405304e+01, 1.92736856260098e+01, 1.91862116914644e+01, 1.91067905837356e+01, 1.90352827135761e+01, 1.89714665464631e+01, 1.89150468379329e+01, 1.88656635937025e+01, 1.88229016365308e+01, 1.87863006563264e+01, 1.87553656084427e+01, 1.87295773089993e+01, 1.87084030574523e+01, 1.86913070977930e+01, 1.86777607131172e+01, 1.86672517362740e+01, 1.86592932540782e+01, 1.86534312859676e+01, 1.86492512312743e+01, 1.86463829030115e+01, 1.86445040000418e+01, 1.86433419126334e+01, 1.86426738069470e+01, 1.86423249895276e+01, 1.86421656105883e+01, 1.86421058218226e+01, 1.86420895578076e+01, 1.86420871572723e+01, 1.86420869721199e+01, 1.86420836473773e+01, 1.86420611217518e+01, 1.86419783150967e+01, 1.86417575823950e+01, 1.86412745114256e+01, 1.86403493618059e+01, 1.86387404846718e+01, 1.86361400197112e+01, 1.86321721170239e+01, 1.86263938745344e+01, 1.86182991162010e+01, 1.86073250607658e+01, 1.85928618443993e+01, 1.85742647632805e+01, 1.85508689952244e+01, 1.85220064459451e+01, 1.84870242503128e+01, 1.84453043488349e+01, 1.83962834628806e+01, 1.83394727181685e+01, 1.82744761240418e+01, 1.82010071143027e+01, 1.81189023997875e+01, 1.80281324759239e+01, 1.79288082683054e+01, 1.78211835790311e+01, 1.77056532045319e+01, 1.75827468160781e+01, 1.74531189088072e+01, 1.73175353151240e+01, 1.71768569269986e+01, 1.70320213669037e+01, 1.68840233832895e+01, 1.67338947254926e+01, 1.65826841837166e+01, 1.64314383763963e+01, 1.62811837465025e+01, 1.61329101061965e+01, 1.59875559583622e+01, 1.58459957316189e+01, 1.57090289948138e+01, 1.55773716657147e+01, 1.54516491921078e+01, 1.53323916565410e+01, 1.52200307342984e+01, 1.51148984154059e+01, 1.50172273848983e+01, 1.49271529418457e+01, 1.48447163277077e+01, 1.47698693290851e+01, 1.47024800186478e+01, 1.46423394998335e+01, 1.45891695240761e+01, 1.45426308518449e+01, 1.45023322288670e+01, 1.44678398453342e+01, 1.44386871381870e+01, 1.44143847850924e+01, 1.43944307246010e+01, 1.43783200219503e+01, 1.43655543862027e+01, 1.43556511342108e+01, 1.43481513925025e+01, 1.43426273314921e+01, 1.43386882388136e+01, 1.43359852607095e+01, 1.43342146721811e+01, 1.43331195770662e+01, 1.43324899867135e+01, 1.43321612781807e+01, 1.43320110872517e+01, 1.43319547452412e+01, 1.43319394188285e+01, 1.43319371566814e+01, 1.43319369853985e+01, 1.43319339511379e+01, 1.43319133935707e+01, 1.43318378217030e+01, 1.43316363742642e+01, 1.43311955085002e+01, 1.43303511864347e+01, 1.43288828676499e+01, 1.43265095784832e+01, 1.43228882822745e+01, 1.43176147232553e+01, 1.43102268568457e+01, 1.43002109104771e+01, 1.42870100409103e+01, 1.42700354664736e+01, 1.42486798569718e+01, 1.42223326629716e+01, 1.41903969641385e+01, 1.41523073192155e+01, 1.41075480152394e+01, 1.40556710484057e+01, 1.39963131310899e+01, 1.39292110151612e+01, 1.38542144550393e+01, 1.37712962063029e+01, 1.36805585653824e+01, 1.35822360983147e+01, 1.34766943746191e+01, 1.33644247070896e+01, 1.32460350892782e+01, 1.31222377080070e+01, 1.29938335755236e+01, 1.28616949611543e+01, 1.27267463920042e+01, 1.25899450254976e+01, 1.24522611680803e+01, 1.23146596274669e+01, 1.21780824531514e+01, 1.20434334619179e+01, 1.19115647853453e+01, 1.17832655357327e+01, 1.16592525791101e+01, 1.15401633332675e+01, 1.14265504710089e+01, 1.13188783950751e+01, 1.12175213511871e+01, 1.11227630510721e+01, 1.10347976829843e+01, 1.09537321910979e+01, 1.08795897073760e+01, 1.08123140211307e+01, 1.07517749734639e+01, 1.06977746664016e+01, 1.06500543793665e+01, 1.06083020876628e+01, 1.05721604776482e+01, 1.05412353502305e+01, 1.05151042976597e+01, 1.04933255284144e+01, 1.04754467020784e+01, 1.04610136219169e+01, 1.04495786193017e+01, 1.04407084533465e+01, 1.04339915432774e+01, 1.04290443520598e+01, 1.04255167490612e+01, 1.04230961978062e+01, 1.04215106421735e+01, 1.04205299999082e+01, 1.04199662145713e+01, 1.04196718639258e+01, 1.04195373717963e+01, 1.04194869190116e+01, 1.04194731946173e+01, 1.04194711689250e+01, 1.04194710168556e+01, 1.04194683402622e+01, 1.04194502059454e+01, 1.04193835422325e+01, 1.04192058408966e+01, 1.04188169440785e+01, 1.04180721532858e+01, 1.04167769343798e+01, 1.04146834586642e+01, 1.04114891810625e+01, 1.04068376111772e+01, 1.04003213807128e+01, 1.03914876502324e+01, 1.03798458287456e+01, 1.03648775012057e+01, 1.03460483727399e+01, 1.03228219468915e+01, 1.02946745624385e+01, 1.02611113250600e+01, 1.02216823930198e+01, 1.01759990174049e+01, 1.01237487042856e+01, 1.00647088641245e+01, 9.99875834633223e+00, 9.92588632454167e+00, 9.84619809826451e+00, 9.75991750338338e+00, 9.66738576945298e+00, 9.56905681685376e+00, 9.46548914219064e+00, 9.35733458746689e+00, 9.24532442020859e+00, 9.13025326179767e+00, 9.01296148466985e+00, 8.89431675139212e+00, 8.77519538650066e+00, 8.65646425334425e+00, 8.53896375327194e+00, 8.42349247651254e+00, 8.31079391943981e+00, 8.20154555089081e+00, 8.09635037193652e+00, 7.99573098050531e+00, 7.90012603486865e+00, 7.80988891608579e+00, 7.72528832367870e+00, 7.64651050223112e+00, 7.57366278733069e+00, 7.50677817298261e+00, 7.44582063353064e+00, 7.39069097487750e+00, 7.34123303624708e+00, 7.29724010947601e+00, 7.25846148355144e+00, 7.22460905474909e+00, 7.19536396544056e+00, 7.17038324675292e+00, 7.14930644212816e+00, 7.13176218165584e+00, 7.11737466275211e+00, 7.10576997375717e+00, 7.09658217607338e+00, 7.08945904042967e+00, 7.08406731648767e+00, 7.08009740471531e+00, 7.07726729711152e+00, 7.07532566011346e+00, 7.07405394917011e+00, 7.07326746946781e+00, 7.07281532978993e+00, 7.07257927446786e+00, 7.07247141937529e+00, 7.07243095929285e+00, 7.07241995317757e+00, 7.07241832869830e+00, 7.07241820711897e+00, 7.07241607212501e+00, 7.07240160723236e+00, 7.07234843283965e+00, 7.07220669006805e+00, 7.07189649270871e+00, 7.07130244003323e+00, 7.07026941619065e+00, 7.06859987911002e+00, 7.06605281507870e+00, 7.06234450308997e+00, 7.05715119383304e+00, 7.05011375994017e+00, 7.04084431528032e+00, 7.02893473100226e+00, 7.01396689527317e+00, 6.99552447445419e+00, 6.97320583985260e+00, 6.94663773203980e+00, 6.91548915138926e+00, 6.87948489729345e+00, 6.83841813796189e+00, 6.79216138546967e+00, 6.74067528264614e+00, 6.68401468241431e+00, 6.62233161556082e+00, 6.55587489467755e+00, 6.48498628097537e+00, 6.41009333389617e+00, 6.33169925535424e+00, 6.25037021436429e+00, 6.16672077797239e+00, 6.08139816790515e+00, 5.99506610098089e+00, 5.90838895276847e+00, 5.82201691218096e+00, 5.73657267928773e+00, 5.65264011347907e+00, 5.57075508026148e+00, 5.49139858834063e+00, 5.41499216806203e+00, 5.34189532788434e+00, 5.27240484313041e+00, 5.20675558221436e+00, 5.14512255758877e+00, 5.08762389684356e+00, 5.03432445728266e+00, 4.98523984807195e+00, 4.94034067131189e+00, 4.89955684175736e+00, 4.86278189025636e+00, 4.82987719545349e+00, 4.80067612018318e+00, 4.77498805246839e+00, 4.75260236604609e+00, 4.73329232226540e+00, 4.71681893480869e+00, 4.70293481198891e+00, 4.69138797960528e+00, 4.68192567189844e+00, 4.67429806059407e+00, 4.66826187403187e+00, 4.66358384166746e+00, 4.66004388548369e+00, 4.65743797056584e+00, 4.65558052348998e+00, 4.65430633002556e+00, 4.65347183321169e+00, 4.65295576879963e+00, 4.65265909646280e+00, 4.65250421066951e+00, 4.65243344296246e+00, 4.65240689569654e+00, 4.65239967421014e+00, 4.65239860833437e+00, 4.65239852853221e+00, 4.65239712676735e+00, 4.65238762961130e+00, 4.65235271716811e+00, 4.65225965423511e+00, 4.65205599245581e+00, 4.65166597177553e+00, 4.65098777257973e+00, 4.64989175671609e+00, 4.64821981990312e+00, 4.64578595847444e+00, 4.64237812826677e+00, 4.63776144089091e+00, 4.63168270107701e+00, 4.62387623742365e+00, 4.61407091803817e+00, 4.60199817410481e+00, 4.58740078199702e+00, 4.57004208357004e+00, 4.54971526161175e+00, 4.52625224082362e+00, 4.49953176180327e+00, 4.46948618273673e+00, 4.43610660489794e+00, 4.39944599415392e+00, 4.35962007793462e+00, 4.31680592781014e+00, 4.27123828062487e+00, 4.22320379249432e+00, 4.17303354583700e+00, 4.12109422749339e+00, 4.06777845677166e+00, 4.01349476156010e+00, 3.95865767928295e+00, 3.90367840311355e+00, 3.84895631181621e+00, 3.79487162514363e+00, 3.74177932725444e+00, 3.69000440801133e+00, 3.63983839352510e+00, 3.59153707707178e+00, 3.54531932071088e+00, 3.50136677535715e+00, 3.45982435991490e+00, 3.42080134482772e+00, 3.38437289841748e+00, 3.35058197246823e+00, 3.31944142405341e+00, 3.29093629166136e+00, 3.26502616386199e+00, 3.24164759712501e+00, 3.22071655529197e+00, 3.20213085616292e+00, 3.18577262036792e+00, 3.17151072396004e+00, 3.15920325892666e+00, 3.14870000515940e+00, 3.13984491361350e+00, 3.13247859387657e+00, 3.12644079079941e+00, 3.12157282503220e+00, 3.11771996221038e+00, 3.11473366617155e+00, 3.11247368398414e+00, 3.11080990566425e+00, 3.10962394000103e+00, 3.10881035038693e+00, 3.10827750111170e+00, 3.10794797502082e+00, 3.10775853720432e+00, 3.10765963561174e+00, 3.10761444711845e+00, 3.10759749542915e+00, 3.10759288416569e+00, 3.10759220355312e+00, 3.10759215246045e+00, 3.10759125318662e+00, 3.10758516047802e+00, 3.10756276309575e+00, 3.10750306035112e+00, 3.10737240465497e+00, 3.10712219218023e+00, 3.10668709816680e+00, 3.10598394460873e+00, 3.10491127763444e+00, 3.10334971903490e+00, 3.10116314008146e+00, 3.09820068491729e+00, 3.09429964449217e+00, 3.08928914977423e+00, 3.08299461506737e+00, 3.07524281988706e+00, 3.06586747328068e+00, 3.05471506106931e+00, 3.04165073844782e+00, 3.02656400237655e+00, 3.00937386474743e+00, 2.99003325207570e+00, 2.96853238254616e+00, 2.94490091659392e+00, 2.91920874037973e+00, 2.89156531779260e+00, 2.86211762946948e+00, 2.83104679933036e+00, 2.79856358295636e+00, 2.76490295153729e+00, 2.73031804566183e+00, 2.69507379271778e+00, 2.65944048012340e+00, 2.62368755594461e+00, 2.58807789200083e+00, 2.55286269655620e+00, 2.51827720879215e+00, 2.48453725017898e+00, 2.45183665305936e+00, 2.42034553811484e+00, 2.39020937293964e+00, 2.36154871562406e+00, 2.33445953072559e+00, 2.30901395972098e+00, 2.28526143238268e+00, 2.26323001720545e+00, 2.24292792541693e+00, 2.22434510171337e+00, 2.20745485350438e+00, 2.19221548748585e+00, 2.17857193670482e+00, 2.16645737235383e+00, 2.15579480215598e+00, 2.14649866148107e+00, 2.13847640455587e+00, 2.13163010169565e+00, 2.12585804485149e+00, 2.12105635844882e+00, 2.11712060603296e+00, 2.11394737622436e+00, 2.11143582453583e+00, 2.10948914134458e+00, 2.10801591135660e+00, 2.10693132679664e+00, 2.10615821574376e+00, 2.10562784878647e+00, 2.10528049156894e+00, 2.10506567769555e+00, 2.10494218549057e+00, 2.10487771271849e+00, 2.10484825487054e+00, 2.10483720426235e+00, 2.10483419823347e+00, 2.10483375455007e+00, 2.10483372110150e+00, 2.10483313048910e+00, 2.10482912900577e+00, 2.10481441916876e+00, 2.10477520844359e+00, 2.10468939825064e+00, 2.10452506728974e+00, 2.10423931259307e+00, 2.10377750594929e+00, 2.10307301649225e+00, 2.10204744178974e+00, 2.10061137901361e+00, 2.09866575398969e+00, 2.09610370849787e+00, 2.09281302479594e+00, 2.08867904113202e+00, 2.08358798378364e+00, 2.07743061144230e+00, 2.07010603881696e+00, 2.06152558102380e+00, 2.05161644187909e+00, 2.04032506076029e+00, 2.02761993684985e+00, 2.01349376785733e+00, 1.99796477276605e+00, 1.98107711304930e+00, 1.96290038067179e+00, 1.94352817911856e+00, 1.92307587991819e+00, 1.90167768587242e+00, 1.87948316855623e+00, 1.85665346833762e+00, 1.83335734905066e+00, 1.80976728764846e+00, 1.78605575469199e+00, 1.76239180868850e+00, 1.73893809079051e+00, 1.71584827052727e+00, 1.69326496140017e+00, 1.67131809937010e+00, 1.65012375827037e+00, 1.62978336373207e+00, 1.61038326039683e+00, 1.59199458480541e+00, 1.57467339717701e+00, 1.55846102828840e+00, 1.54338460200803e+00, 1.52945769914399e+00, 1.51668113369265e+00, 1.50504381800204e+00, 1.49452369851783e+00, 1.48508874841785e+00, 1.47669800734118e+00, 1.46930266138579e+00, 1.46284715843246e+00, 1.45727035455910e+00, 1.45250668682172e+00, 1.44848736606376e+00, 1.44514158083259e+00, 1.44239770017921e+00, 1.44018445941626e+00, 1.43843210919457e+00, 1.43707350493985e+00, 1.43604511117832e+00, 1.43528789394065e+00, 1.43474807455643e+00, 1.43437771991314e+00, 1.43413514769786e+00, 1.43398513015948e+00, 1.43389888628293e+00, 1.43385385958729e+00, 1.43383328660499e+00, 1.43382556898312e+00, 1.43382346960417e+00, 1.43382315974029e+00, 1.43382313625303e+00, 1.43382271984452e+00, 1.43381989861670e+00, 1.43380952750107e+00, 1.43378188203959e+00, 1.43372138125783e+00, 1.43360551745415e+00, 1.43340403693462e+00, 1.43307841127123e+00, 1.43258163342671e+00, 1.43185836678344e+00, 1.43084546736034e+00, 1.42947288989706e+00, 1.42766497670265e+00, 1.42534211405653e+00, 1.42242272464869e+00, 1.41882554656475e+00, 1.41447213059470e+00, 1.40928946952672e+00, 1.40321265726832e+00, 1.39618746397659e+00, 1.38817270767161e+00, 1.37914230450234e+00, 1.36908688976039e+00, 1.35801491988722e+00, 1.34595319114693e+00, 1.33294674150360e+00, 1.31905813601899e+00, 1.30436616987608e+00, 1.28896405405236e+00, 1.27295717421374e+00, 1.25646053174559e+00, 1.23959598598478e+00, 1.22248941852459e+00, 1.20526793455991e+00, 1.18805720384504e+00, 1.17097902656960e+00, 1.15414918912815e+00, 1.13767565318243e+00, 1.12165710026387e+00, 1.10618183484552e+00, 1.09132703239326e+00, 1.07715830606663e+00, 1.06372955677828e+00, 1.05108306618402e+00, 1.03924979051826e+00, 1.02824981445815e+00, 1.01809292770942e+00, 1.00877929203314e+00, 1.00030017227273e+00, 9.92638710968095e-01, 9.85770731837576e-01, 9.79665562365263e-01, 9.74286869667580e-01, 9.69593506557981e-01, 9.65540366214004e-01, 9.62079244102494e-01, 9.59159704942579e-01, 9.56729950661426e-01, 9.54737682766861e-01, 9.53130949613335e-01, 9.51858966000643e-01, 9.50872889757743e-01, 9.50126537758639e-01, 9.49577022487823e-01, 9.49185290051275e-01, 9.48916541557260e-01, 9.48740522108338e-01, 9.48631665174958e-01, 9.48569084680162e-01, 9.48536412437573e-01, 9.48521484314963e-01, 9.48515884276144e-01, 9.48514360931256e-01, 9.48514136088793e-01, 9.48514118937894e-01, 9.48513813440897e-01, 9.48511743654830e-01, 9.48504134918146e-01, 9.48483852941365e-01, 9.48439466948685e-01, 9.48354464828689e-01, 9.48206652832578e-01, 9.47967769228601e-01, 9.47603337628024e-01, 9.47072781206039e-01, 9.46329813442524e-01, 9.45323113976627e-01, 9.43997289437000e-01, 9.42294108543553e-01, 9.40153988462495e-01, 9.37517695706511e-01, 9.34328210501040e-01, 9.30532689506123e-01, 9.26084449373178e-01, 9.20944884263552e-01, 9.15085225576771e-01, 9.08488052927638e-01, 9.01148472638432e-01, 8.93074893829564e-01, 8.84289352031832e-01, 8.74827354803205e-01, 8.64737251174098e-01, 8.54079154503719e-01, 8.42923474001616e-01, 8.31349131432901e-01, 8.19441554547836e-01, 8.07290546439685e-01, 7.94988130058577e-01, 7.82626460045318e-01, 7.70295881152514e-01, 7.58083195533771e-01, 7.46070182070167e-01, 7.34332391591099e-01, 7.22938223993612e-01, 7.11948278111358e-01, 7.01414953483177e-01, 6.91382275193774e-01, 6.81885908570305e-01, 6.72953329281849e-01, 6.64604115687595e-01, 6.56850333422803e-01, 6.49696986520016e-01, 6.43142514231265e-01, 6.37179317643231e-01, 6.31794304778978e-01, 6.26969446884743e-01, 6.22682341832942e-01, 6.18906782938062e-01, 6.15613332945689e-01, 6.12769903536751e-01, 6.10342340447016e-01, 6.08295013337745e-01, 6.06591407997040e-01, 6.05194716472695e-01, 6.04068418526261e-01, 6.03176845572730e-01, 6.02485716253835e-01, 6.01962631207162e-01, 6.01577513635295e-01, 6.01302982106777e-01, 6.01114642731240e-01, 6.00991289471168e-01, 6.00915003829979e-01, 6.00871148363787e-01, 6.00848252206884e-01, 6.00837790843304e-01, 6.00833866437551e-01, 6.00832798905241e-01, 6.00832641339757e-01, 6.00832629235910e-01, 6.00832412526557e-01, 6.00830944289697e-01, 6.00825546908776e-01, 6.00811159573787e-01, 6.00779673754875e-01, 6.00719376606129e-01, 6.00614525494986e-01, 6.00445074697173e-01, 6.00186572742405e-01, 5.99810245760899e-01, 5.99283278224718e-01, 5.98569297446826e-01, 5.97629061882952e-01, 5.96421345567450e-01, 5.94904001966388e-01, 5.93035180430256e-01, 5.90774657815992e-01, 5.88085237513308e-01, 5.84934159056260e-01, 5.81294454839195e-01, 5.77146187275741e-01, 5.72477500933658e-01, 5.67285430274662e-01, 5.61576414647405e-01, 5.55366487555450e-01, 5.48681125807519e-01, 5.41554764339512e-01, 5.34030002396690e-01, 5.26156544482088e-01, 5.17989933372188e-01, 5.09590141415476e-01, 5.01020089732016e-01, 4.92344162957796e-01, 4.83626780548228e-01, 4.74931075524093e-01, 4.66317719284124e-01, 4.57843918109546e-01, 4.49562594470478e-01, 4.41521755153754e-01, 4.33764039169329e-01, 4.26326431629494e-01, 4.19240125326215e-01, 4.12530509349786e-01, 4.06217263471230e-01, 4.00314537772346e-01, 3.94831198768866e-01, 3.89771125679331e-01, 3.85133543236741e-01, 3.80913380267856e-01, 3.77101645975216e-01, 3.73685818294359e-01, 3.70650240746816e-01, 3.67976525783853e-01, 3.65643963662883e-01, 3.63629936395370e-01, 3.61910336260589e-01, 3.60459987836226e-01, 3.59253071528414e-01, 3.58263545295639e-01, 3.57465559783950e-01, 3.56833860578774e-01, 3.56344169891628e-01, 3.55973538901964e-01, 3.55700661308841e-01, 3.55506138530814e-01, 3.55372687497194e-01, 3.55285283120480e-01, 3.55231229292336e-01, 3.55200154513007e-01, 3.55183930911014e-01, 3.55176518266539e-01, 3.55173737536305e-01, 3.55172981111068e-01, 3.55172869464319e-01, 3.55172860822403e-01, 3.55172705243938e-01, 3.55171651177585e-01, 3.55167776327654e-01, 3.55157447473289e-01, 3.55134843397952e-01, 3.55091555306618e-01, 3.55016281386817e-01, 3.54894630624692e-01, 3.54709049156622e-01, 3.54438880190724e-01, 3.54060565699022e-01, 3.53547994479392e-01, 3.52872996665177e-01, 3.52005979243977e-01, 3.50916690680151e-01, 3.49575095520677e-01, 3.47952332286116e-01, 3.46021720570214e-01, 3.43759776801697e-01, 3.41147193344682e-01, 3.38169733293203e-01, 3.34818994071082e-01, 3.31092997143351e-01, 3.26996568784399e-01, 3.22541487531112e-01, 3.17746386885723e-01, 3.12636415921599e-01, 3.07242674416695e-01, 3.01601451727386e-01, 2.95753308729523e-01, 2.89742049016866e-01, 2.83613628778372e-01, 2.77415054401537e-01, 2.71193313257392e-01, 2.64994376970570e-01, 2.58862308572478e-01, 2.52838496117850e-01, 2.46961026393421e-01, 2.41264203908898e-01, 2.35778212925691e-01, 2.30528914171405e-01, 2.25537763280418e-01, 2.20821834927435e-01, 2.16393935003953e-01, 2.12262782865125e-01, 2.08433246430218e-01, 2.04906614511106e-01, 2.01680892918230e-01, 1.98751113408278e-01, 1.96109647162881e-01, 1.93746517021030e-01, 1.91649704962970e-01, 1.89805453232793e-01, 1.88198558904052e-01, 1.86812662592232e-01, 1.85630532388426e-01, 1.84634343957317e-01, 1.83805957167834e-01, 1.83127188696741e-01, 1.82580078877809e-01, 1.82147149796911e-01, 1.81811650396224e-01, 1.81557783291566e-01, 1.81370907251626e-01, 1.81237708939562e-01, 1.81146337642701e-01, 1.81086497340033e-01, 1.81049491557456e-01, 1.81028217972901e-01, 1.81017111553746e-01, 1.81012037009708e-01, 1.81010133384461e-01, 1.81009615553318e-01, 1.81009539122567e-01, 1.81009533158552e-01, 1.81009425170092e-01, 1.81008693533173e-01, 1.81006003969173e-01, 1.80998834661267e-01, 1.80983145223667e-01, 1.80953099673154e-01, 1.80900855214045e-01, 1.80816427718270e-01, 1.80687643665962e-01, 1.80500186890221e-01, 1.80237746590311e-01, 1.79882270579981e-01, 1.79414324485538e-01, 1.78813553525970e-01, 1.78059238618477e-01, 1.77130933005152e-01, 1.76009159689265e-01, 1.74676144141081e-01, 1.73116551545039e-01, 1.71318193939303e-01, 1.69272670562403e-01, 1.66975905090496e-01, 1.64428546546080e-01, 1.61636206542681e-01, 1.58609513939222e-01, 1.55363978339050e-01, 1.51919665341467e-01, 1.48300698025487e-01, 1.44534609758786e-01, 1.40651582108042e-01, 1.36683607626416e-01, 1.32663620155427e-01, 1.28624634896283e-01, 1.24598937108605e-01, 1.20617352390415e-01, 1.16708623793308e-01, 1.12898912345720e-01, 1.09211428720027e-01, 1.05666195529704e-01, 1.02279932666860e-01, 9.90660525745817e-02, 9.60347485711175e-02, 9.31931573005187e-02, 9.05455759318201e-02, 8.80937156161750e-02, 8.58369746250602e-02, 8.37727171901044e-02, 8.18965470168713e-02, 8.02025674685503e-02, 7.86836232900328e-02, 7.73315213120078e-02, 7.61372297327654e-02, 7.50910572607992e-02, 7.41828145855479e-02, 7.34019613328769e-02, 7.27377418849926e-02, 7.21793132544377e-02, 7.17158676628777e-02, 7.13367516649761e-02, 7.10315826617176e-02, 7.07903625599400e-02, 7.06035872535793e-02, 7.04623496272703e-02, 7.03584330032641e-02, 7.02843914463855e-02, 7.02336131596972e-02, 7.02003633714554e-02, 7.01798036217767e-02, 7.01679851672345e-02, 7.01618152633756e-02, 7.01589962734275e-02, 7.01579387875491e-02, 7.01576511271256e-02, 7.01576086691027e-02, 7.01576053278430e-02, 7.01575444675946e-02, 7.01571321312571e-02, 7.01556163513821e-02, 7.01515759317779e-02, 7.01427340443210e-02, 7.01258025808436e-02, 7.00963642838062e-02, 7.00487993650931e-02, 6.99762629146825e-02, 6.98707187888061e-02, 6.97230347102824e-02, 6.95231418607491e-02, 6.92602601657342e-02, 6.89231876873180e-02, 6.85006490479623e-02, 6.79816937260860e-02, 6.73561306303512e-02, 6.66149809454240e-02, 6.57509273250199e-02, 6.47587346272580e-02, 6.36356160739631e-02, 6.23815194095130e-02, 6.09993105972784e-02, 5.94948378330677e-02, 5.78768658924780e-02, 5.61568794919296e-02, 5.43487636207125e-02, 5.24683777463899e-02, 5.05330484501734e-02, 4.85610105817074e-02, 4.65708298601086e-02, 4.45808397509317e-02, 4.26086225497500e-02, 4.06705593662586e-02, 3.87814668400779e-02, 3.69543307653828e-02, 3.52001391902910e-02, 3.35278107139416e-02, 3.19442081726393e-02, 3.04542240043730e-02, 2.90609213969049e-02, 2.77657147391466e-02, 2.65685736406738e-02, 2.54682365100685e-02, 2.44624220280087e-02, 2.35480294890316e-02, 2.27213216466247e-02, 2.19780861760377e-02, 2.13137740262574e-02, 2.07236146907022e-02, 2.02027097629269e-02, 1.97461070807243e-02, 1.93488583450971e-02, 1.90060633842357e-02, 1.87129042676624e-02, 1.84646723075836e-02, 1.82567906539752e-02, 1.80848347330915e-02, 1.79445522320556e-02, 1.78318837305014e-02, 1.77429844595629e-02, 1.76742470657558e-02, 1.76223247106897e-02, 1.75841533828809e-02, 1.75569719680091e-02, 1.75383384439840e-02, 1.75261405497054e-02, 1.75185994219973e-02, 1.75142649868469e-02, 1.75120023008297e-02, 1.75109685247783e-02, 1.75105807300184e-02, 1.75104752414228e-02, 1.75104596715592e-02, 1.75104584354351e-02, 1.75104357818476e-02, 1.75102823010285e-02, 1.75097180976067e-02, 1.75082142020528e-02, 1.75049232864097e-02, 1.74986220267606e-02, 1.74876679552654e-02, 1.74699736559170e-02, 1.74430011421123e-02, 1.74037791362203e-02, 1.73489456650486e-02, 1.72748177906996e-02, 1.71774893289935e-02, 1.70529560119661e-02, 1.68972657231379e-02, 1.67066892450229e-02, 1.64779045617287e-02, 1.62081853970619e-02, 1.58955826496694e-02, 1.55390860587907e-02, 1.51387531254805e-02, 1.46957932774683e-02, 1.42125976093794e-02, 1.36927081738918e-02, 1.31407254482276e-02, 1.25621577673378e-02, 1.19632215829419e-02, 1.13506057324493e-02, 1.07312159270734e-02, 1.01119170226838e-02, 9.49929018535656e-03, 8.89941991691856e-03, 8.31772238763406e-03, 7.75882213284973e-03, 7.22647951066691e-03, 6.72356700099996e-03, 6.25208894346940e-03, 5.81323692792779e-03, 5.40747178766310e-03, 5.03462286334866e-03, 4.69399572308049e-03, 4.38448067256303e-03, 4.10465598382714e-03, 3.85288156411789e-03, 3.62738049624411e-03, 3.42630729419113e-03, 3.24780276682191e-03, 3.09003613237380e-03, 2.95123557087607e-03, 2.82970878042477e-03, 2.72385530257679e-03, 2.63217239197999e-03, 2.55325606485357e-03, 2.48579874788035e-03, 2.42858473214369e-03, 2.38048444620276e-03, 2.34044839052316e-03, 2.30750140397618e-03, 2.28073775265614e-03, 2.25931734682305e-03, 2.24246321391419e-03, 2.22946019563963e-03, 2.21965469891572e-03, 2.21245521963098e-03, 2.20733327535427e-03, 2.20382433398214e-03, 2.20152831132125e-03, 2.20010923427694e-03, 2.19929372488181e-03, 2.19886804800140e-03, 2.19867357383308e-03, 2.19860062328081e-03, 2.19858077928237e-03, 2.19857785036126e-03, 2.19857761569742e-03, 2.19857328834649e-03, 2.19854397008135e-03, 2.19843619577631e-03, 2.19814893002734e-03, 2.19752036240806e-03, 2.19631698798040e-03, 2.19422558531690e-03, 2.19084874500197e-03, 2.18570463849943e-03, 2.17823172525655e-03, 2.16779904659603e-03, 2.15372261454050e-03, 2.13528814427733e-03, 2.11177998080814e-03, 2.08251553531369e-03, 2.04688390062202e-03, 2.00438661763513e-03, 1.95467790272730e-03, 1.89760113253357e-03, 1.83321813408455e-03, 1.76182795001267e-03, 1.68397230555328e-03, 1.60042600058219e-03, 1.51217181127450e-03, 1.42036105525278e-03, 1.32626253352086e-03, 1.23120386778197e-03, 1.13651008750537e-03, 1.04344454557292e-03, 9.53156835603153e-04, 8.66641423339587e-04, 7.84709318818244e-04, 7.07973523997651e-04, 6.36847538675767e-04, 5.71555141906543e-04, 5.12149037760423e-04, 4.58535812690286e-04, 4.10504662504252e-04, 3.67757075778741e-04, 3.29934841278856e-04, 2.96644596527784e-04, 2.67477739203770e-04, 2.42025809992441e-04, 2.19892673182898e-04, 2.00703669413826e-04, 1.84111631750028e-04, 1.69800926226316e-04, 1.57489233515662e-04, 1.46925943557363e-04, 1.37889041685213e-04, 1.30182958585578e-04, 1.23636456546450e-04, 1.18099188068044e-04, 1.13438342783212e-04, 1.09536935809729e-04, 1.06292888407467e-04, 1.03617413913071e-04, 1.01432659644018e-04, 9.96694962758227e-05, 9.82660047448292e-05, 9.71665744522454e-05, 9.63213168780896e-05, 9.56856209839430e-05, 9.52198108556956e-05, 9.48889020904176e-05, 9.46624360729231e-05, 9.45143543624738e-05, 9.44228719994555e-05, 9.43703134232976e-05, 9.43428831945425e-05, 9.43303523932570e-05, 9.43256520322677e-05, 9.43243734496946e-05, 9.43241847437548e-05, 9.43241694741914e-05, 9.43238862696205e-05, 9.43219675448859e-05, 9.43149144063682e-05, 9.42961155772082e-05, 9.42549862301689e-05, 9.41762624186993e-05, 9.40394981086001e-05, 9.38188182928457e-05, 9.34829882304200e-05, 9.29958622144511e-05, 9.23172728981569e-05, 9.14044089519450e-05, 9.02137025029504e-05, 8.87032048630942e-05, 8.68353744536421e-05, 8.45801421966996e-05, 8.19180598590208e-05, 7.88432558422869e-05, 7.53657860990131e-05, 7.15128294537237e-05, 6.73282890701898e-05, 6.28709578620601e-05, 5.82120158618693e-05, 5.34321555498707e-05, 4.86174464029432e-05, 4.38536756064423e-05, 3.92212000548398e-05, 3.47917439604828e-05, 3.06255429478597e-05, 2.67682678014580e-05, 2.32500886774289e-05, 2.00861877850400e-05, 1.72754663587682e-05, 1.48002509519074e-05, 1.26301262781681e-05, 1.07272360665434e-05, 9.05478426815876e-06, 7.58960225510918e-06, 6.32490162644217e-06, 5.25417807040185e-06, 4.37041681422556e-06, 3.66811494956210e-06, 3.03514566368790e-06, 2.42734846622739e-06, 2.11303288111253e-06, 1.78930552152679e-06, 1.44350512343294e-06, 1.12143343500293e-06, 8.57822644917405e-07, 5.95798571367823e-07, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -1.42108547152020e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14],
[ 3.30000000000000e+01, 3.29999999745243e+01, 3.29999991867640e+01, 3.29999938495636e+01, 3.29999742293256e+01, 3.29999219280154e+01, 3.29998074630094e+01, 3.29995882301752e+01, 3.29992069240578e+01, 3.29985904752378e+01, 3.29976495492746e+01, 3.29962786347013e+01, 3.29943567299191e+01, 3.29917486211944e+01, 3.29883067269736e+01, 3.29838734679837e+01, 3.29782841086112e+01, 3.29713700031999e+01, 3.29629621713537e+01, 3.29528951190494e+01, 3.29410108171276e+01, 3.29271627451833e+01, 3.29112199065801e+01, 3.28930707188811e+01, 3.28726266831912e+01, 3.28498257357511e+01, 3.28246351859312e+01, 3.27970541471707e+01, 3.27671153722305e+01, 3.27348864122984e+01, 3.27004700318146e+01, 3.26640038278624e+01, 3.26256590246102e+01, 3.25856384390074e+01, 3.25441736425714e+01, 3.25015213739740e+01, 3.24579592862713e+01, 3.24137811389777e+01, 3.23692915668905e+01, 3.23248005731936e+01, 3.22806179030157e+01, 3.22370474550440e+01, 3.21943818833640e+01, 3.21528975302686e+01, 3.21128498145753e+01, 3.20744691803938e+01, 3.20379576896952e+01, 3.20034863197043e+01, 3.19711930041054e+01, 3.19411814360881e+01, 3.19135206318362e+01, 3.18882452354544e+01, 3.18653565305694e+01, 3.18448241099000e+01, 3.18265881418321e+01, 3.18105621623490e+01, 3.17966363114764e+01, 3.17846809256847e+01, 3.17745503915188e+01, 3.17660871612178e+01, 3.17591258284321e+01, 3.17534971615400e+01, 3.17490319937161e+01, 3.17455648729552e+01, 3.17429373818057e+01, 3.17410010456023e+01, 3.17396197593950e+01, 3.17386716773233e+01, 3.17380505235451e+01, 3.17376663005603e+01, 3.17374453883785e+01, 3.17373300458937e+01, 3.17372773434945e+01, 3.17372575727680e+01, 3.17372521946286e+01, 3.17372514008257e+01, 3.17372512966955e+01, 3.17372488705865e+01, 3.17372324333244e+01, 3.17371720079628e+01, 3.17370109342338e+01, 3.17366584186762e+01, 3.17359832722695e+01, 3.17348090714254e+01, 3.17329109404888e+01, 3.17300141110081e+01, 3.17257943658193e+01, 3.17198804254533e+01, 3.17118582816076e+01, 3.17012774286288e+01, 3.16876588903589e+01, 3.16705048874198e+01, 3.16493099398654e+01, 3.16235731525325e+01, 3.15928113853016e+01, 3.15565729675635e+01, 3.15144515753475e+01, 3.14660998515510e+01, 3.14112423167865e+01, 3.13496870948167e+01, 3.12813359686758e+01, 3.12061922988970e+01, 3.11243663810212e+01, 3.10360779006603e+01, 3.09416552613423e+01, 3.08415317078289e+01, 3.07362383341249e+01, 3.06263942347255e+01, 3.05126942112417e+01, 3.03958945671135e+01, 3.02767975978287e+01, 3.01562354070494e+01, 3.00350536521681e+01, 2.99140957548646e+01, 2.97941880166972e+01, 2.96761259718602e+01, 2.95606622030552e+01, 2.94484957527418e+01, 2.93402631871923e+01, 2.92365313167134e+01, 2.91377915404640e+01, 2.90444557645700e+01, 2.89568538328844e+01, 2.88752324061364e+01, 2.87997552236225e+01, 2.87305046794706e+01, 2.86674846414202e+01, 2.86106244334657e+01, 2.85597838946184e+01, 2.85147594147761e+01, 2.84752908356998e+01, 2.84410690908678e+01, 2.84117444430601e+01, 2.83869351635261e+01, 2.83662364822657e+01, 2.83492296261932e+01, 2.83354907517893e+01, 2.83245995723927e+01, 2.83161474786503e+01, 2.83097449548393e+01, 2.83050281045743e+01, 2.83016641172602e+01, 2.82993555315798e+01, 2.82978431838701e+01, 2.82969077665314e+01, 2.82963699632880e+01, 2.82960891724871e+01, 2.82959608747866e+01, 2.82959127455558e+01, 2.82958996532066e+01, 2.82958977208021e+01, 2.82958975502076e+01, 2.82958942074915e+01, 2.82958715600892e+01, 2.82957883057042e+01, 2.82955663790165e+01, 2.82950806924900e+01, 2.82941505239987e+01, 2.82925328884324e+01, 2.82899181853917e+01, 2.82859283631250e+01, 2.82801177795868e+01, 2.82719768741731e+01, 2.82609386875591e+01, 2.82463881817218e+01, 2.82276742180114e+01, 2.82041239495780e+01, 2.81750592785983e+01, 2.81398149233483e+01, 2.80977575416365e+01, 2.80483052732084e+01, 2.79909470028160e+01, 2.79252606157183e+01, 2.78509295247650e+01, 2.77677567963029e+01, 2.76756762902147e+01, 2.75747603521339e+01, 2.74652237436449e+01, 2.73474236565017e+01, 2.72218558163701e+01, 2.70891468291517e+01, 2.69500430518468e+01, 2.68053963787612e+01, 2.66561474256390e+01, 2.65033066733092e+01, 2.63479342003477e+01, 2.61911186869145e+01, 2.60339563989383e+01, 2.58775308497473e+01, 2.57228937747650e+01, 2.55710479430674e+01, 2.54229321791110e+01, 2.52794088011050e+01, 2.51412535256814e+01, 2.50091477639057e+01, 2.48836731532205e+01, 2.47653081336118e+01, 2.46544263748266e+01, 2.45512968811543e+01, 2.44560856278339e+01, 2.43688586087238e+01, 2.42895861929961e+01, 2.42181486975000e+01, 2.41543430816824e+01, 2.40978906651955e+01, 2.40484457563979e+01, 2.40056050644174e+01, 2.39689177495526e+01, 2.39378959475910e+01, 2.39120255841685e+01, 2.38907772767950e+01, 2.38736171060009e+01, 2.38600170248127e+01, 2.38494646690800e+01, 2.38414723316114e+01, 2.38355848719530e+01, 2.38313863518037e+01, 2.38285052138191e+01, 2.38266178585183e+01, 2.38254505191059e+01, 2.38247793855337e+01, 2.38244289848171e+01, 2.38242688818927e+01, 2.38242088214441e+01, 2.38241924835161e+01, 2.38241900720711e+01, 2.38241898786434e+01, 2.38241863089206e+01, 2.38241621235159e+01, 2.38240732152907e+01, 2.38238362176913e+01, 2.38233175496047e+01, 2.38223242203996e+01, 2.38205967569578e+01, 2.38178045673987e+01, 2.38135440017823e+01, 2.38073393077069e+01, 2.37986466073731e+01, 2.37868609412875e+01, 2.37713263316577e+01, 2.37513487158915e+01, 2.37262114890523e+01, 2.36951932770192e+01, 2.36575874448383e+01, 2.36127227346849e+01, 2.35599843338768e+01, 2.34988346051021e+01, 2.34288326775972e+01, 2.33496521066491e+01, 2.32610958632636e+01, 2.31631080153135e+01, 2.30557815999724e+01, 2.29393623541407e+01, 2.28142481512101e+01, 2.26809841745715e+01, 2.25402540283835e+01, 2.23928671357969e+01, 2.22397429000128e+01, 2.20818922035475e+01, 2.19203968965327e+01, 2.17563879756143e+01, 2.15910231789873e+01, 2.14254647168685e+01, 2.12608578171777e+01, 2.10983106928922e+01, 2.09388764341498e+01, 2.07835372029132e+01, 2.06331909724448e+01, 2.04886409208160e+01, 2.03505874689047e+01, 2.02196228574830e+01, 2.00962280896201e+01, 1.99807720240710e+01, 1.98735123895070e+01, 1.97745984929580e+01, 1.96840754122886e+01, 1.96018894855999e+01, 1.95278949346655e+01, 1.94618614806881e+01, 1.94034828259861e+01, 1.93523858831819e+01, 1.93081406336589e+01, 1.92702704899750e+01, 1.92382630237474e+01, 1.92115809029523e+01, 1.91896728627074e+01, 1.91719845137500e+01, 1.91579687754583e+01, 1.91470957077992e+01, 1.91388615112576e+01, 1.91327964674322e+01, 1.91284716068600e+01, 1.91255039153016e+01, 1.91235599249541e+01, 1.91223575817674e+01, 1.91216663323801e+01, 1.91213054316813e+01, 1.91211405317504e+01, 1.91210786718634e+01, 1.91210618444526e+01, 1.91210593607612e+01, 1.91210591691950e+01, 1.91210557292812e+01, 1.91210324233538e+01, 1.91209467482244e+01, 1.91207183692091e+01, 1.91202185643971e+01, 1.91192613671888e+01, 1.91175967581581e+01, 1.91149062132896e+01, 1.91108008647122e+01, 1.91048224713342e+01, 1.90964473289925e+01, 1.90850931715300e+01, 1.90701290246571e+01, 1.90508878735193e+01, 1.90266818939211e+01, 1.89968198793106e+01, 1.89606263760458e+01, 1.89174619252803e+01, 1.88667437097962e+01, 1.88079658280230e+01, 1.87407183750973e+01, 1.86647045106955e+01, 1.85797547415963e+01, 1.84858377458173e+01, 1.83830672122923e+01, 1.82717043577438e+01, 1.81521559978900e+01, 1.80249682764669e+01, 1.78908163734340e+01, 1.77504907038405e+01, 1.76048802646022e+01, 1.74549538766638e+01, 1.73017401006291e+01, 1.71463065786428e+01, 1.69897394846755e+01, 1.68331236645559e+01, 1.66775239327527e+01, 1.65239678799173e+01, 1.63734304442086e+01, 1.62268204156927e+01, 1.60849689768249e+01, 1.59486203298476e+01, 1.58184244191667e+01, 1.56949317192851e+01, 1.55785900244482e+01, 1.54697431446976e+01, 1.53686313860103e+01, 1.52753936716190e+01, 1.51900711489015e+01, 1.51126121216116e+01, 1.50428781494905e+01, 1.49806511641129e+01, 1.49256414582700e+01, 1.48774964134080e+01, 1.48358098333203e+01, 1.48001317510162e+01, 1.47699785690488e+01, 1.47448433821688e+01, 1.47242063163717e+01, 1.47075447022748e+01, 1.46943428856975e+01, 1.46841014668709e+01, 1.46763457542663e+01, 1.46706332216572e+01, 1.46665597691552e+01, 1.46637646113175e+01, 1.46619336478898e+01, 1.46608012143577e+01, 1.46601501584944e+01, 1.46598102431383e+01, 1.46596549317439e+01, 1.46595966688739e+01, 1.46595808199413e+01, 1.46595784806714e+01, 1.46595783035489e+01, 1.46595751658420e+01, 1.46595539074112e+01, 1.46594757590990e+01, 1.46592674438384e+01, 1.46588115481825e+01, 1.46579384427209e+01, 1.46564200713883e+01, 1.46539658894252e+01, 1.46502211832061e+01, 1.46447679515714e+01, 1.46371284665941e+01, 1.46267715607523e+01, 1.46131216065018e+01, 1.45955700633199e+01, 1.45734893676787e+01, 1.45462488359433e+01, 1.45132321433830e+01, 1.44738558404743e+01, 1.44275882778989e+01, 1.43739682421923e+01, 1.43126225628497e+01, 1.42432819457142e+01, 1.41657943215498e+01, 1.40801350750224e+01, 1.39864136369037e+01, 1.38848760769454e+01, 1.37759035191223e+01, 1.36600064044211e+01, 1.35378148358952e+01, 1.34100654407515e+01, 1.32775853576304e+01, 1.31412740869412e+01, 1.30020840139140e+01, 1.28610004200192e+01, 1.27190217398623e+01, 1.25771407091963e+01, 1.24363269052969e+01, 1.22975110274251e+01, 1.21615711241357e+01, 1.20293208604129e+01, 1.19014998363730e+01, 1.17787659176044e+01, 1.16616895071124e+01, 1.15507496711333e+01, 1.14463320184928e+01, 1.13487282216856e+01, 1.12581370565785e+01, 1.11746668276975e+01, 1.10983390392527e+01, 1.10290931696647e+01, 1.09667924096502e+01, 1.09112302299762e+01, 1.08621376530175e+01, 1.08191911100499e+01, 1.07820207716531e+01, 1.07502192400566e+01, 1.07233504888625e+01, 1.07009589273508e+01, 1.06825784544168e+01, 1.06677413527514e+01, 1.06559868593528e+01, 1.06468692363031e+01, 1.06399651584120e+01, 1.06348802339680e+01, 1.06312544830614e+01, 1.06287666156477e+01, 1.06271369787562e+01, 1.06261290782346e+01, 1.06255496235829e+01, 1.06252470925713e+01, 1.06251088628629e+01, 1.06250570079991e+01, 1.06250429022068e+01, 1.06250408202210e+01, 1.06250406639256e+01, 1.06250379129504e+01, 1.06250192746864e+01, 1.06249507584209e+01, 1.06247681189426e+01, 1.06243684155392e+01, 1.06236029304479e+01, 1.06222717290089e+01, 1.06201201094668e+01, 1.06168371506110e+01, 1.06120565194475e+01, 1.06053596476741e+01, 1.05962813237027e+01, 1.05843176755078e+01, 1.05689364386201e+01, 1.05495893142544e+01, 1.05257261274524e+01, 1.04968103984301e+01, 1.04623358477922e+01, 1.04218432748086e+01, 1.03749371850481e+01, 1.03213015065343e+01, 1.02607137282589e+01, 1.01930568252463e+01, 1.01183284015180e+01, 1.00366465842492e+01, 9.94825233415978e+00, 9.85350799136593e+00, 9.75289204353194e+00, 9.64699027451053e+00, 9.53648361697789e+00, 9.42213318253820e+00, 9.30476306871601e+00, 9.18524163639439e+00, 9.06446200709257e+00, 8.94332254247051e+00, 8.82270803687798e+00, 8.70347227952259e+00, 8.58642253180269e+00, 8.47230632715845e+00, 8.36180084779379e+00, 8.25550497834234e+00, 8.15393399389358e+00, 8.05751671938959e+00, 7.96659490643575e+00, 7.88142451528440e+00, 7.80217856359682e+00, 7.72895120590175e+00, 7.66176273264879e+00, 7.60056521852967e+00, 7.54524859931328e+00, 7.49564700842441e+00, 7.45154525358192e+00, 7.41268535600071e+00, 7.37877310730093e+00, 7.34948462097216e+00, 7.32447286573869e+00, 7.30337416813829e+00, 7.28581466250768e+00, 7.27141665041846e+00, 7.25980481092702e+00, 7.25061218056608e+00, 7.24348580067490e+00, 7.23809191217994e+00, 7.23412056670781e+00, 7.23128951980348e+00, 7.22934727819313e+00, 7.22807518879417e+00, 7.22728848198631e+00, 7.22683621415051e+00, 7.22660009261604e+00, 7.22649220742973e+00, 7.22645173608386e+00, 7.22644072690708e+00, 7.22643910197603e+00, 7.22643898036290e+00, 7.22643684477521e+00, 7.22642237586101e+00, 7.22636918670016e+00, 7.22622740468078e+00, 7.22591712203077e+00, 7.22532290832048e+00, 7.22428961165393e+00, 7.22261965298269e+00, 7.22007199177099e+00, 7.21636290960871e+00, 7.21116871934041e+00, 7.20413045879077e+00, 7.19486056853355e+00, 7.18295148156758e+00, 7.16798596978869e+00, 7.14954900009821e+00, 7.12724075615404e+00, 7.10069038639569e+00, 7.06956995295709e+00, 7.03360798836709e+00, 6.99260202663857e+00, 6.94642947064911e+00, 6.89505619469114e+00, 6.83854236254458e+00, 6.77704506607384e+00, 6.71081755119331e+00, 6.64020498648966e+00, 6.56563693032811e+00, 6.48761684782339e+00, 6.40670920186868e+00, 6.32352477626296e+00, 6.23870497137217e+00, 6.15290583674692e+00, 6.06678257031993e+00, 5.98097512645962e+00, 5.89609544712601e+00, 5.81271667727974e+00, 5.73136456454918e+00, 5.65251108995998e+00, 5.57657024425609e+00, 5.50389576166926e+00, 5.43478055387998e+00, 5.36945755094651e+00, 5.30810164933185e+00, 5.25083248381807e+00, 5.19771777311762e+00, 5.14877703155097e+00, 5.10398548529523e+00, 5.06327807672098e+00, 5.02655348088106e+00, 4.99367809220880e+00, 4.96448996586078e+00, 4.93880271660185e+00, 4.91640938890408e+00, 4.89708631559060e+00, 4.88059697969704e+00, 4.86669588621732e+00, 4.85513243815991e+00, 4.84565479612595e+00, 4.83801368382796e+00, 4.83196608510945e+00, 4.82727876266616e+00, 4.82373151634587e+00, 4.82112009100129e+00, 4.81925864150090e+00, 4.81798166637205e+00, 4.81714533187046e+00, 4.81662812469646e+00, 4.81633079323031e+00, 4.81617556268692e+00, 4.81610463731791e+00, 4.81607803088446e+00, 4.81607079330086e+00, 4.81606972504911e+00, 4.81606964506905e+00, 4.81606824017942e+00, 4.81605872185168e+00, 4.81602373156531e+00, 4.81593046102481e+00, 4.81572634436286e+00, 4.81533545046327e+00, 4.81465572626896e+00, 4.81355722824157e+00, 4.81188146287552e+00, 4.80944193709304e+00, 4.80602599382942e+00, 4.80139797543036e+00, 4.79530371617192e+00, 4.78747631452019e+00, 4.77764307604548e+00, 4.76553345106123e+00, 4.75088772066129e+00, 4.73346611618311e+00, 4.71305799695559e+00, 4.68949066700901e+00, 4.66263739060373e+00, 4.63242417506422e+00, 4.59883493110975e+00, 4.56191469581712e+00, 4.52177070764347e+00, 4.47857124865715e+00, 4.43254230504294e+00, 4.38396222990945e+00, 4.33315470919262e+00, 4.28048042066913e+00, 4.22632783001159e+00, 4.17110358336646e+00, 4.11522293497575e+00, 4.05910059704046e+00, 4.00314232641164e+00, 3.94773747906252e+00, 3.89325267838508e+00, 3.84002666504485e+00, 3.78836632968412e+00, 3.73854387773884e+00, 3.69079503828966e+00, 3.64531820489090e+00, 3.60227438360484e+00, 3.56178781973443e+00, 3.52394717793633e+00, 3.48880715879816e+00, 3.45639044718628e+00, 3.42668990250566e+00, 3.39967091732957e+00, 3.37527388753271e+00, 3.35341675299546e+00, 3.33399758211082e+00, 3.31689718484454e+00, 3.30198174732016e+00, 3.28910548544416e+00, 3.27811331586705e+00, 3.26884353980370e+00, 3.26113052939008e+00, 3.25480739805723e+00, 3.24970862676781e+00, 3.24567260793454e+00, 3.24254405953275e+00, 3.24017625442021e+00, 3.23843300516365e+00, 3.23719034351613e+00, 3.23633783658120e+00, 3.23577948878215e+00, 3.23543418980131e+00, 3.23523568308087e+00, 3.23513204639619e+00, 3.23508469433012e+00, 3.23506693099865e+00, 3.23506209894776e+00, 3.23506138574739e+00, 3.23506133220839e+00, 3.23506038987722e+00, 3.23505400544837e+00, 3.23503053566535e+00, 3.23496797424641e+00, 3.23483106216961e+00, 3.23456886705092e+00, 3.23411293222112e+00, 3.23337608658855e+00, 3.23225199512430e+00, 3.23061551606905e+00, 3.22832391374807e+00, 3.22521895414863e+00, 3.22112988329485e+00, 3.21587725551918e+00, 3.20927754024152e+00, 3.20114839300732e+00, 3.19131443151876e+00, 3.17961331346251e+00, 3.16590187416803e+00, 3.15006205305309e+00, 3.13200632286446e+00, 3.11168233859126e+00, 3.08907654590436e+00, 3.06421653241103e+00, 3.03717196703967e+00, 3.00805404941147e+00, 2.97701347620691e+00, 2.94423701818492e+00, 2.90994288217079e+00, 2.87437509995858e+00, 2.83779723487839e+00, 2.80048572280599e+00, 2.76272316596966e+00, 2.72479187584079e+00, 2.68696791891641e+00, 2.64951586163340e+00, 2.61268434473387e+00, 2.57670255036903e+00, 2.54177756378241e+00, 2.50809258077500e+00, 2.47580587545621e+00, 2.44505042088818e+00, 2.41593404701078e+00, 2.38854002316041e+00, 2.36292796335217e+00, 2.33913496803041e+00, 2.31717693339624e+00, 2.29704997655270e+00, 2.27873194015315e+00, 2.26218395319889e+00, 2.24735203477648e+00, 2.23416873481979e+00, 2.22255481055559e+00, 2.21242093935454e+00, 2.20366946850061e+00, 2.19619620017993e+00, 2.18989220607680e+00, 2.18464566070896e+00, 2.18034367646570e+00, 2.17687411673476e+00, 2.17412735708855e+00, 2.17199795885243e+00, 2.17038621510358e+00, 2.16919952679469e+00, 2.16835356669748e+00, 2.16777319148722e+00, 2.16739306760230e+00, 2.16715798435089e+00, 2.16702283771292e+00, 2.16695227982588e+00, 2.16692004153573e+00, 2.16690794787104e+00, 2.16690465810394e+00, 2.16690417254131e+00, 2.16690413593553e+00, 2.16690348957526e+00, 2.16689911039130e+00, 2.16688301207752e+00, 2.16684010011146e+00, 2.16674618948021e+00, 2.16656634381923e+00, 2.16625360457400e+00, 2.16574817203270e+00, 2.16497709210580e+00, 2.16385449305893e+00, 2.16228240440386e+00, 2.16015217513893e+00, 2.15734648994611e+00, 2.15374195952634e+00, 2.14921223519428e+00, 2.14363156898809e+00, 2.13687871043012e+00, 2.12884100196185e+00, 2.11941850981157e+00, 2.10852800877298e+00, 2.09610663110677e+00, 2.08211499395327e+00, 2.06653963759219e+00, 2.04939463844120e+00, 2.03072230399151e+00, 2.01059290846748e+00, 1.98910348318376e+00, 1.96637572909049e+00, 1.94255316579961e+00, 1.91779766739970e+00, 1.89228555807870e+00, 1.86620344927502e+00, 1.83974399576768e+00, 1.81310173309782e+00, 1.78646913600732e+00, 1.76003301034058e+00, 1.73397130189569e+00, 1.70845037725172e+00, 1.68362280519607e+00, 1.65962564401610e+00, 1.63657922016930e+00, 1.61458636799691e+00, 1.59373208830774e+00, 1.57408357580686e+00, 1.55569056130618e+00, 1.53858591411005e+00, 1.52278645243724e+00, 1.50829391459993e+00, 1.49509605019171e+00, 1.48316779797843e+00, 1.47247252479217e+00, 1.46296330682812e+00, 1.45458424077911e+00, 1.44727177679730e+00, 1.44095606809426e+00, 1.43556233298534e+00, 1.43101222441919e+00, 1.42722519972171e+00, 1.42411987976646e+00, 1.42161538250665e+00, 1.41963261128301e+00, 1.41809547410827e+00, 1.41693200676014e+00, 1.41607537047236e+00, 1.41546469467303e+00, 1.41504573681684e+00, 1.41477133495583e+00, 1.41460163417896e+00, 1.41450407513575e+00, 1.41445314111845e+00, 1.41442986908120e+00, 1.41442113895858e+00, 1.41441876415563e+00, 1.41441841363977e+00, 1.41441838707114e+00, 1.41441791603276e+00, 1.41441472467985e+00, 1.41440299295126e+00, 1.41437172064095e+00, 1.41430328282166e+00, 1.41417221953256e+00, 1.41394430976996e+00, 1.41357597492060e+00, 1.41301405100741e+00, 1.41219596213484e+00, 1.41105031887409e+00, 1.40949795454543e+00, 1.40745339904615e+00, 1.40482677385480e+00, 1.40152607325321e+00, 1.39745977620129e+00, 1.39253971170452e+00, 1.38668407942077e+00, 1.37982050853430e+00, 1.37188902366171e+00, 1.36284477880440e+00, 1.35266042086044e+00, 1.34132795408593e+00, 1.32885999642809e+00, 1.31529034711052e+00, 1.30067382049440e+00, 1.28508534144361e+00, 1.26861833895500e+00, 1.25138251419665e+00, 1.23350109300878e+00, 1.21510769858162e+00, 1.19634299549691e+00, 1.17735126071052e+00, 1.15827703056640e+00, 1.13926195680023e+00, 1.12044198078607e+00, 1.10194490662672e+00, 1.08388842294624e+00, 1.06637859315757e+00, 1.04950880690659e+00, 1.03335916308382e+00, 1.01799623828317e+00, 1.00347318419763e+00, 9.89830092868658e-01, 9.77094569174724e-01, 9.65282454374595e-01, 9.54398651718755e-01, 9.44438013930771e-01, 9.35386261698206e-01, 9.27220911347220e-01, 9.19912197968706e-01, 9.13423986988100e-01, 9.07714672276860e-01, 9.02738062297047e-01, 8.98444257472043e-01, 8.94780522104526e-01, 8.91692152905641e-01, 8.89123343806908e-01, 8.87018043501797e-01, 8.85320798446806e-01, 8.83977570217503e-01, 8.82936512546209e-01, 8.82148690439621e-01, 8.81568721823378e-01, 8.81155321452951e-01, 8.80871727538569e-01, 8.80685993710188e-01, 8.80571132520934e-01, 8.80505101456649e-01, 8.80470628077539e-01, 8.80454877085633e-01, 8.80448968373869e-01, 8.80447361063752e-01, 8.80447123828240e-01, 8.80447105732017e-01, 8.80446783396408e-01, 8.80444599526918e-01, 8.80436571416602e-01, 8.80415171613812e-01, 8.80368339641990e-01, 8.80278654542408e-01, 8.80122703070498e-01, 8.79870675358553e-01, 8.79486216260503e-01, 8.78926556111310e-01, 8.78142938860648e-01, 8.77081358055669e-01, 8.75683601589081e-01, 8.73888594305782e-01, 8.71634013548658e-01, 8.68858136953300e-01, 8.65501865110637e-01, 8.61510845349862e-01, 8.56837608434184e-01, 8.51443619185432e-01, 8.45301136715683e-01, 8.38394781521558e-01, 8.30722716111943e-01, 8.22297363227506e-01, 8.13145610275228e-01, 8.03308478596307e-01, 7.92840269062388e-01, 7.81807228179900e-01, 7.70285808185122e-01, 7.58360617661147e-01, 7.46122173864109e-01, 7.33664573099091e-01, 7.21083191184469e-01, 7.08472513437094e-01, 6.95924174713426e-01, 6.83525267386159e-01, 6.71356951364992e-01, 6.59493377775789e-01, 6.48000918552469e-01, 6.36937679147074e-01, 6.26353261300771e-01, 6.16288737205984e-01, 6.06776794838439e-01, 5.97842015890940e-01, 5.89501251648706e-01, 5.81764067400798e-01, 5.74633231807070e-01, 5.68105233432718e-01, 5.62170811995649e-01, 5.56815496476226e-01, 5.52020145967560e-01, 5.47761491938708e-01, 5.44012682445171e-01, 5.40743829788290e-01, 5.37922563261937e-01, 5.35514588015367e-01, 5.33484249809391e-01, 5.31795103676451e-01, 5.30410482367728e-01, 5.29294058158548e-01, 5.28410389289654e-01, 5.27725440254898e-01, 5.27207063520606e-01, 5.26825429266047e-01, 5.26553389526718e-01, 5.26366763796524e-01, 5.26244534733003e-01, 5.26168945059830e-01, 5.26125489947070e-01, 5.26102802880587e-01, 5.26092437068108e-01, 5.26088548509442e-01, 5.26087490728656e-01, 5.26087334602487e-01, 5.26087322609200e-01, 5.26087107879434e-01, 5.26085653054608e-01, 5.26080304979445e-01, 5.26066049088136e-01, 5.26034850994840e-01, 5.25975105121404e-01, 5.25871213453080e-01, 5.25703315435884e-01, 5.25447187560431e-01, 5.25074328117682e-01, 5.24552238707287e-01, 5.23844909054837e-01, 5.22913505320851e-01, 5.21717254241295e-01, 5.20214506188871e-01, 5.18363949885845e-01, 5.16125940610820e-01, 5.13463893174702e-01, 5.10345681770183e-01, 5.06744982212751e-01, 5.02642489235619e-01, 4.98026943289119e-01, 4.92895908217065e-01, 4.87256253153234e-01, 4.81124308273223e-01, 4.74525683307043e-01, 4.67494758170346e-01, 4.60073874691034e-01, 4.52312275268831e-01, 4.44264846853727e-01, 4.35990735913748e-01, 4.27551901829339e-01, 4.19011672833292e-01, 4.10433361180431e-01, 4.01878983973901e-01, 3.93408124374417e-01, 3.85076956034844e-01, 3.76937442526625e-01, 3.69036713927223e-01, 3.61416614961669e-01, 3.54113413226074e-01, 3.47157651964999e-01, 3.40574129431111e-01, 3.34381985778222e-01, 3.28594878484857e-01, 3.23221228243014e-01, 3.18264518863401e-01, 3.13723636846824e-01, 3.09593238663211e-01, 3.05864136283738e-01, 3.02523693955614e-01, 2.99556231435155e-01, 2.96943430763534e-01, 2.94664745073796e-01, 2.92697808781334e-01, 2.91018848803425e-01, 2.89603096187427e-01, 2.88425196760251e-01, 2.87459618244846e-01, 2.86681049857251e-01, 2.86064788866568e-01, 2.85587107144877e-01, 2.85225589534839e-01, 2.84959435077553e-01, 2.84769711902925e-01, 2.84639556969154e-01, 2.84554312875507e-01, 2.84501595629507e-01, 2.84471289436567e-01, 2.84455467159439e-01, 2.84448237895838e-01, 2.84445525960074e-01, 2.84444788248791e-01, 2.84444679364185e-01, 2.84444670936068e-01, 2.84444519206630e-01, 2.84443491218063e-01, 2.84439712234061e-01, 2.84429638932600e-01, 2.84407594176962e-01, 2.84365377435307e-01, 2.84291967182661e-01, 2.84173330164002e-01, 2.83992350669330e-01, 2.83728890833594e-01, 2.83359990251540e-01, 2.82860209700743e-01, 2.82202119341498e-01, 2.81356926329401e-01, 2.80295230375863e-01, 2.78987888622446e-01, 2.77406963634739e-01, 2.75526720924525e-01, 2.73324635871440e-01, 2.70782365010305e-01, 2.67886634128672e-01, 2.64629996100076e-01, 2.61011415231941e-01, 2.57036642168337e-01, 2.52718353708282e-01, 2.48076044561692e-01, 2.43135672053441e-01, 2.37929068913566e-01, 2.32493152350528e-01, 2.26868968513358e-01, 2.21100619379378e-01, 2.15234123555305e-01, 2.09316263304842e-01, 2.03393467499474e-01, 1.97510774595607e-01, 1.91710911835386e-01, 1.86033517438872e-01, 1.80514522431181e-01, 1.75185698726686e-01, 1.70074370874346e-01, 1.65203281010847e-01, 1.60590590452792e-01, 1.56249997178804e-01, 1.52190946222291e-01, 1.48418909570921e-01, 1.44935713286337e-01, 1.41739891872618e-01, 1.38827053057142e-01, 1.36190239729302e-01, 1.33820279472786e-01, 1.31706115643340e-01, 1.29835117066342e-01, 1.28193366003650e-01, 1.26765925972741e-01, 1.25537092255129e-01, 1.24490628511865e-01, 1.23609992880368e-01, 1.22878556338428e-01, 1.22279815098331e-01, 1.21797597466930e-01, 1.21416264123781e-01, 1.21120899284065e-01, 1.20897488878065e-01, 1.20733080834562e-01, 1.20615921913490e-01, 1.20535565372940e-01, 1.20482944108495e-01, 1.20450404754415e-01, 1.20431699524360e-01, 1.20421934192258e-01, 1.20417472441822e-01, 1.20415798703057e-01, 1.20415343407224e-01, 1.20415276206572e-01, 1.20415270962795e-01, 1.20415176015499e-01, 1.20414532734607e-01, 1.20412167980831e-01, 1.20405864529467e-01, 1.20392070158651e-01, 1.20365654504511e-01, 1.20319724395706e-01, 1.20245507751015e-01, 1.20132315341564e-01, 1.19967588644030e-01, 1.19737040460720e-01, 1.19424892766425e-01, 1.19014213194504e-01, 1.18487347597977e-01, 1.17826441223613e-01, 1.17014035367625e-01, 1.16033720254364e-01, 1.14870818768203e-01, 1.13513070184467e-01, 1.11951278879545e-01, 1.10179890842268e-01, 1.08197461247478e-01, 1.06006979769269e-01, 1.03616026791997e-01, 1.01036742951059e-01, 9.82856058672397e-02, 9.53830205746229e-02, 9.23527428079964e-02, 8.92211657666948e-02, 8.60165100457237e-02, 8.27679622134099e-02, 7.95048094762194e-02, 7.62556159144907e-02, 7.30474802522778e-02, 6.99054067829508e-02, 6.68518109190899e-02, 6.39061699937002e-02, 6.10848194932601e-02, 5.84008857652591e-02, 5.58643390661970e-02, 5.34821459514845e-02, 5.12584975107373e-02, 4.91950896180740e-02, 4.72914328118108e-02, 4.55451721570981e-02, 4.39524009770977e-02, 4.25079562092208e-02, 4.12056869819040e-02, 4.00386915397632e-02, 3.89995206930465e-02, 3.80803484347325e-02, 3.72731122291512e-02, 3.65696267527511e-02, 3.59616756140042e-02, 3.54410858651235e-02, 3.49997900155046e-02, 3.46298798352294e-02, 3.43236555649113e-02, 3.40736732903650e-02, 3.38727922630397e-02, 3.37142229189169e-02, 3.35915753384839e-02, 3.34989069706637e-02, 3.34307676800663e-02, 3.33822396287644e-02, 3.33489692138045e-02, 3.33271882718336e-02, 3.33137220305488e-02, 3.33059818049293e-02, 3.33019411522173e-02, 3.33000950456963e-02, 3.32994025237454e-02, 3.32992141425308e-02, 3.32991863379135e-02, 3.32991841498185e-02, 3.32991442940838e-02, 3.32988742665066e-02, 3.32978816293448e-02, 3.32952357213046e-02, 3.32894457145230e-02, 3.32783590680208e-02, 3.32590852703021e-02, 3.32279497241217e-02, 3.31804825336590e-02, 3.31114466988822e-02, 3.30149096632937e-02, 3.28843611514600e-02, 3.27128786380939e-02, 3.24933395258498e-02, 3.22186761697196e-02, 3.18821663766755e-02, 3.14777481633257e-02, 3.10003437359825e-02, 3.04461743374347e-02, 2.98130453138796e-02, 2.91005800070465e-02, 2.83103822777662e-02, 2.74461108237816e-02, 2.65134538976852e-02, 2.55200001784388e-02, 2.44750097245188e-02, 2.33890972411004e-02, 2.22738473377930e-02, 2.11413871110082e-02, 2.00039445463673e-02, 1.88734215312465e-02, 1.77610077245234e-02, 1.66768565693545e-02, 1.56298380912219e-02, 1.46273757077040e-02, 1.36753670209231e-02, 1.27781822582804e-02, 1.19387292208657e-02, 1.11585705514656e-02, 1.04380778485620e-02, 9.77660743097886e-03, 9.17268406706029e-03, 8.62418131774234e-03, 8.12848987967385e-03, 7.68266807623519e-03, 7.28357114970635e-03, 6.92795810109459e-03, 6.61257645253244e-03, 6.33422648679982e-03, 6.08980729321473e-03, 5.87634737528475e-03, 5.69102271722954e-03, 5.53116513678731e-03, 5.39426354131223e-03, 5.27796041393458e-03, 5.18004552696283e-03, 5.09844853439034e-03, 5.03123174451758e-03, 4.97658402570522e-03, 4.93281645565702e-03, 4.89835999709243e-03, 4.87176517707866e-03, 4.85170347458563e-03, 4.83696989334703e-03, 4.82648602631031e-03, 4.81930281554810e-03, 4.81460218273355e-03, 4.81169675366289e-03, 4.81002701766187e-03, 4.80915543818838e-03, 4.80875724587548e-03, 4.80860787659054e-03, 4.80856724511369e-03, 4.80856124802642e-03, 4.80856077189884e-03, 4.80855204634167e-03, 4.80849292978291e-03, 4.80827561601416e-03, 4.80769637543688e-03, 4.80642891517391e-03, 4.80400231879230e-03, 4.79978476734999e-03, 4.79297430839409e-03, 4.78259799469782e-03, 4.76752071797648e-03, 4.74646496060416e-03, 4.71804242143037e-03, 4.68079798837095e-03, 4.63326579525614e-03, 4.57403611372342e-03, 4.50183063806264e-03, 4.41558242238216e-03, 4.31451548409001e-03, 4.19821809369125e-03, 4.06670324667147e-03, 3.92044995111718e-03, 3.76041989095199e-03, 3.58804574992888e-03, 3.40518988527805e-03, 3.21407486182324e-03, 3.01719023568126e-03, 2.81718250177221e-03, 2.61673690529562e-03, 2.41846057514294e-03, 2.22477603619353e-03, 2.03783267053836e-03, 1.85944139229832e-03, 1.69103509114876e-03, 1.53365472979772e-03, 1.38795870919495e-03, 1.25425144270253e-03, 1.13252606556813e-03, 1.02251583462021e-03, 9.23749013132635e-04, 8.35602810703051e-04, 7.57353091159985e-04, 6.88217803073599e-04, 6.27393160314682e-04, 5.74082362703621e-04, 5.27517131573063e-04, 4.86972666564611e-04, 4.51776897371303e-04, 4.21315096879482e-04, 3.95030979973399e-04, 3.72425314687064e-04, 3.53052874920934e-04, 3.36518351751636e-04, 3.22471678437353e-04, 3.10603121711495e-04, 3.00638423162525e-04, 2.92334211408729e-04, 2.85473838609107e-04, 2.79863733105401e-04, 2.75330309136734e-04, 2.71717437819063e-04, 2.68884456858132e-04, 2.66704674608320e-04, 2.65064305750684e-04, 2.63861759620454e-04, 2.63007189225561e-04, 2.62422201330992e-04, 2.62039626420574e-04, 2.61803252989523e-04, 2.61667442650548e-04, 2.61596560875432e-04, 2.61564179766083e-04, 2.61552033357759e-04, 2.61548729319827e-04, 2.61548241653031e-04, 2.61548202587392e-04, 2.61547482089952e-04, 2.61542600661357e-04, 2.61524656686163e-04, 2.61476829891194e-04, 2.61372188646192e-04, 2.61171889853529e-04, 2.60823886577555e-04, 2.60262273648948e-04, 2.59407414453960e-04, 2.58166999273612e-04, 2.56438177196117e-04, 2.54110872909052e-04, 2.51072337256630e-04, 2.47212883564885e-04, 2.42432630045641e-04, 2.36648911595694e-04, 2.29803859497224e-04, 2.21871500187376e-04, 2.12863626757098e-04, 2.02833678315528e-04, 1.91877950946662e-04, 1.80133672962768e-04, 1.67773801535986e-04, 1.54998801413342e-04, 1.42026079146262e-04, 1.29078081869238e-04, 1.16370252918330e-04, 1.04100003362362e-04, 9.24376040956076e-05, 8.15195743939512e-05, 7.14449145746698e-05, 6.22742971287948e-05, 5.40320033550756e-05, 4.67102531445107e-05, 4.02753437427350e-05, 3.46743532588789e-05, 2.98414247623668e-05, 2.57033460044909e-05, 2.21836236633521e-05, 1.92054939063269e-05, 1.66954104940942e-05, 1.45857270794636e-05, 1.28160706580616e-05, 1.13348920791623e-05, 1.00987304207933e-05, 9.06899958152962e-06, 8.21161718533858e-06, 7.49829646196076e-06, 6.90444637996279e-06, 6.40746057456454e-06, 5.98978292742913e-06, 5.64079419262953e-06, 5.35269312251785e-06, 5.11590054941280e-06, 4.91985280604013e-06, 4.75642258379594e-06, 4.62095521669426e-06, 4.51064119744160e-06, 4.42270614087192e-06, 4.35387227071260e-06, 4.30075426294252e-06, 4.26032666922538e-06, 4.23008212635523e-06, 4.20796590816508e-06, 4.19226171288756e-06, 4.18151248027243e-06, 4.17448181622149e-06, 4.17013739451022e-06, 4.16764105892753e-06, 4.16633810118583e-06, 4.16574285111437e-06, 4.16551956305966e-06, 4.16545882586661e-06, 4.16544985881728e-06, 4.16544914827455e-06, 4.16543598902308e-06, 4.16534686564773e-06, 4.16501924860313e-06, 4.16414606263515e-06, 4.16223580401720e-06, 4.15858001190372e-06, 4.15223065886039e-06, 4.14199023879291e-06, 4.12641799840685e-06, 4.10385599280971e-06, 4.07247841138769e-06, 4.03036545293389e-06, 3.97559935549907e-06, 3.90637705294239e-06, 3.82114181007864e-06, 3.71876844695862e-06, 3.59887560108518e-06, 3.46228658543168e-06, 3.31137675146920e-06, 3.14963330083629e-06, 2.97996114539956e-06, 2.80288775655890e-06, 2.61731047856983e-06, 2.42416133033885e-06, 2.22875768685071e-06, 2.03870543913354e-06, 1.86144476543859e-06, 1.70500280916031e-06, 1.57526070410086e-06, 1.46600012129738e-06, 1.35873097661943e-06, 1.23750798053379e-06, 1.09299530492990e-06, 9.11697121352972e-07, 9.06601592021161e-07, 8.53622999841264e-07, 7.00659434471618e-07, 6.81558816495453e-07, 6.46570015305770e-07, 5.49133453375816e-07, 4.40775480115008e-07, 4.43455398624337e-07, -1.42108547152020e-14, -1.42108547152020e-14, 7.72277068961102e-07, 1.10880441894778e-06, 1.25913394555255e-06, 1.20845899687083e-06, 9.88657305356355e-07, 6.62395287065465e-07, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14],
[ 3.40000000000000e+01, 3.39999999733952e+01, 3.39999991507230e+01, 3.39999935769892e+01, 3.39999730872234e+01, 3.39999184680264e+01, 3.39997989301540e+01, 3.39995699813006e+01, 3.39991717762127e+01, 3.39985280068292e+01, 3.39975453787925e+01, 3.39961137029685e+01, 3.39941066122101e+01, 3.39913828951893e+01, 3.39877884214049e+01, 3.39831586150743e+01, 3.39773214210882e+01, 3.39701006939101e+01, 3.39613199304075e+01, 3.39508062600766e+01, 3.39383946007041e+01, 3.39239318838013e+01, 3.39072812516828e+01, 3.38883261264194e+01, 3.38669740498326e+01, 3.38431601932462e+01, 3.38168504362321e+01, 3.37880439157827e+01, 3.37567749521428e+01, 3.37231142659517e+01, 3.36871694142741e+01, 3.36490843910630e+01, 3.36090383606012e+01, 3.35672435198702e+01, 3.35239421163011e+01, 3.34794026791527e+01, 3.34339155536878e+01, 3.33877878551718e+01, 3.33413379825088e+01, 3.32948898475383e+01, 3.32487669847279e+01, 3.32032867070101e+01, 3.31587544672627e+01, 3.31154585723874e+01, 3.30736653794570e+01, 3.30336150824638e+01, 3.29955181753141e+01, 3.29595526532122e+01, 3.29258619915328e+01, 3.28945539194795e+01, 3.28656999857550e+01, 3.28393358953449e+01, 3.28154625803866e+01, 3.27940479538541e+01, 3.27750292822962e+01, 3.27583161029838e+01, 3.27437936014411e+01, 3.27313263574459e+01, 3.27207623612449e+01, 3.27119371970742e+01, 3.27046782882989e+01, 3.26988090978147e+01, 3.26941531789972e+01, 3.26905379766202e+01, 3.26877982838909e+01, 3.26857792710738e+01, 3.26843390129811e+01, 3.26833504566629e+01, 3.26827027865942e+01, 3.26823021620677e+01, 3.26820718198447e+01, 3.26819515538013e+01, 3.26818966017356e+01, 3.26818759870731e+01, 3.26818703793619e+01, 3.26818695516747e+01, 3.26818694430996e+01, 3.26818669134294e+01, 3.26818497745251e+01, 3.26817867698477e+01, 3.26816188205716e+01, 3.26812512578533e+01, 3.26805472935919e+01, 3.26793229759364e+01, 3.26773438366882e+01, 3.26743233932267e+01, 3.26699236185547e+01, 3.26637574404895e+01, 3.26553932760296e+01, 3.26443615506999e+01, 3.26301630964831e+01, 3.26122792669019e+01, 3.25901835547913e+01, 3.25633544477035e+01, 3.25312892077184e+01, 3.24935182164349e+01, 3.24496194821050e+01, 3.23992328651809e+01, 3.23420735435863e+01, 3.22779442145448e+01, 3.22067455226465e+01, 3.21284842220362e+01, 3.20432786317439e+01, 3.19513610321415e+01, 3.18530767772965e+01, 3.17488800562876e+01, 3.16393264135217e+01, 3.15250623159738e+01, 3.14068122142529e+01, 3.12853636663571e+01, 3.11615511650305e+01, 3.10362393265546e+01, 3.09103060637019e+01, 3.07846262887998e+01, 3.06600565893414e+01, 3.05374212046837e+01, 3.04174995227436e+01, 3.03010152209740e+01, 3.01886271021447e+01, 3.00809216236975e+01, 2.99784070872131e+01, 2.98815094371845e+01, 2.97905696105585e+01, 2.97058423755566e+01, 2.96274965964650e+01, 2.95556168580488e+01, 2.94902063778603e+01, 2.94311911266934e+01, 2.93784250670045e+01, 2.93316964067178e+01, 2.92907347519308e+01, 2.92552190271165e+01, 2.92247860159588e+01, 2.91990393605229e+01, 2.91775588417636e+01, 2.91599097512811e+01, 2.91456521537880e+01, 2.91343498331066e+01, 2.91255787128139e+01, 2.91189345469689e+01, 2.91140396874930e+01, 2.91105487532300e+01, 2.91081530515073e+01, 2.91065836357308e+01, 2.91056129212038e+01, 2.91050548245979e+01, 2.91047634385474e+01, 2.91046302997247e+01, 2.91045803544113e+01, 2.91045667680425e+01, 2.91045647627218e+01, 2.91045645856901e+01, 2.91045611168419e+01, 2.91045376148752e+01, 2.91044512190180e+01, 2.91042209182954e+01, 2.91037169052775e+01, 2.91027516389006e+01, 2.91010729663706e+01, 2.90983596074858e+01, 2.90942192508764e+01, 2.90881894502214e+01, 2.90797414386825e+01, 2.90682869008035e+01, 2.90531876525097e+01, 2.90337680819680e+01, 2.90093300985320e+01, 2.89791702269754e+01, 2.89425983746935e+01, 2.88989576972273e+01, 2.88476449004064e+01, 2.87881302542881e+01, 2.87199765631076e+01, 2.86428563430366e+01, 2.85565665089393e+01, 2.84610399614954e+01, 2.83563535913413e+01, 2.82427323673435e+01, 2.81205493392379e+01, 2.79903215479364e+01, 2.78527019896960e+01, 2.77084679175525e+01, 2.75585058844970e+01, 2.74037940405949e+01, 2.72453822929869e+01, 2.70843710212656e+01, 2.69218891019574e+01, 2.67590720199434e+01, 2.65970408164985e+01, 2.64368825353538e+01, 2.62796326853230e+01, 2.61262600605731e+01, 2.59776540768884e+01, 2.58346146236132e+01, 2.56978443167810e+01, 2.55679429758831e+01, 2.54454041285611e+01, 2.53306133600328e+01, 2.52238483512885e+01, 2.51252804789243e+01, 2.50349778717967e+01, 2.49529098322540e+01, 2.48789525326429e+01, 2.48128958929360e+01, 2.47544515348608e+01, 2.47032616936740e+01, 2.46589089519984e+01, 2.46209266417966e+01, 2.45888097413035e+01, 2.45620260743774e+01, 2.45400276013041e+01, 2.45222615739570e+01, 2.45081813159691e+01, 2.44972563819299e+01, 2.44889818502344e+01, 2.44828865134615e+01, 2.44785397489842e+01, 2.44755568812372e+01, 2.44736028853140e+01, 2.44723943282335e+01, 2.44716994975305e+01, 2.44713367244673e+01, 2.44711709684449e+01, 2.44711087873127e+01, 2.44710918725063e+01, 2.44710893759152e+01, 2.44710891756577e+01, 2.44710854798910e+01, 2.44710604405201e+01, 2.44709683930202e+01, 2.44707230272330e+01, 2.44701860453994e+01, 2.44691576425156e+01, 2.44673691835803e+01, 2.44644784035263e+01, 2.44600673992906e+01, 2.44536436189079e+01, 2.44446439786723e+01, 2.44324421551908e+01, 2.44163590038695e+01, 2.43956759492685e+01, 2.43696510774946e+01, 2.43375375398816e+01, 2.42986037562170e+01, 2.42521547922800e+01, 2.41975541895829e+01, 2.41342454547605e+01, 2.40617723812566e+01, 2.39797973839912e+01, 2.38881170821814e+01, 2.37866744654499e+01, 2.36755671177673e+01, 2.35550511423056e+01, 2.34255406148645e+01, 2.32876025807252e+01, 2.31419477880440e+01, 2.29894175120352e+01, 2.28309669635129e+01, 2.26676458905183e+01, 2.25005770708566e+01, 2.23309334535428e+01, 2.21599147342452e+01, 2.19887241394933e+01, 2.18185461442437e+01, 2.16505257588677e+01, 2.14857499011980e+01, 2.13252312280061e+01, 2.11698946520361e+01, 2.10205666296752e+01, 2.08779671824714e+01, 2.07427045210542e+01, 2.06152720758921e+01, 2.04960477046573e+01, 2.03852948363278e+01, 2.02831653211297e+01, 2.01897037759980e+01, 2.01048532408771e+01, 2.00284619865099e+01, 1.99602913354569e+01, 1.99000243724816e+01, 1.98472754269654e+01, 1.98016002085607e+01, 1.97625064685418e+01, 1.97294650445858e+01, 1.97019211277866e+01, 1.96793055696894e+01, 1.96610460263958e+01, 1.96465777187932e+01, 1.96353535751525e+01, 1.96268535169515e+01, 1.96205926526789e+01, 1.96161281588313e+01, 1.96130646529134e+01, 1.96110578997160e+01, 1.96098167383717e+01, 1.96091031717698e+01, 1.96087306192913e+01, 1.96085603955243e+01, 1.96084965384758e+01, 1.96084791677880e+01, 1.96084766039101e+01, 1.96084764061591e+01, 1.96084728551869e+01, 1.96084487968221e+01, 1.96083603556514e+01, 1.96081246033703e+01, 1.96076086622696e+01, 1.96066205618434e+01, 1.96049022109337e+01, 1.96021248024144e+01, 1.95978869148496e+01, 1.95917155149426e+01, 1.95830699945749e+01, 1.95713492954263e+01, 1.95559020815765e+01, 1.95360398160792e+01, 1.95110524826719e+01, 1.94802265718300e+01, 1.94428648266754e+01, 1.93983071262752e+01, 1.93459517808054e+01, 1.92852764351260e+01, 1.92158577346922e+01, 1.91373889092941e+01, 1.90496944820122e+01, 1.89527414150915e+01, 1.88466461581829e+01, 1.87316772590210e+01, 1.86082534182734e+01, 1.84769371010351e+01, 1.83384240371655e+01, 1.81935291318482e+01, 1.80431694503144e+01, 1.78883450265928e+01, 1.77301182730318e+01, 1.75695927406719e+01, 1.74078919121893e+01, 1.72461386146876e+01, 1.70854355349291e+01, 1.69268472176672e+01, 1.67713838366089e+01, 1.66199869497931e+01, 1.64735173850754e+01, 1.63327453430019e+01, 1.61983427498196e+01, 1.60708778408089e+01, 1.59508119041945e+01, 1.58384980712653e+01, 1.57341820024997e+01, 1.56380042952161e+01, 1.55500044266381e+01, 1.54701260462490e+01, 1.53982234400939e+01, 1.53340690033926e+01, 1.52773615723669e+01, 1.52277354780335e+01, 1.51847701913344e+01, 1.51480004290805e+01, 1.51169265836721e+01, 1.50910253273750e+01, 1.50697602258424e+01, 1.50525921779108e+01, 1.50389894821040e+01, 1.50284373174658e+01, 1.50204464198365e+01, 1.50145607366111e+01, 1.50103638548871e+01, 1.50074840204582e+01, 1.50055975982297e+01, 1.50044308673057e+01, 1.50037600944306e+01, 1.50034098851867e+01, 1.50032498704684e+01, 1.50031898432256e+01, 1.50031735143414e+01, 1.50031711042316e+01, 1.50031709217454e+01, 1.50031676890197e+01, 1.50031457868230e+01, 1.50030652719650e+01, 1.50028506484048e+01, 1.50023809474072e+01, 1.50014814038979e+01, 1.49999170591867e+01, 1.49973885790248e+01, 1.49935305282996e+01, 1.49879122887147e+01, 1.49800417423726e+01, 1.49693717710297e+01, 1.49553095370812e+01, 1.49372284180485e+01, 1.49144823627700e+01, 1.48864223275314e+01, 1.48524143387016e+01, 1.48118586214938e+01, 1.47642091400581e+01, 1.47089928208687e+01, 1.46458276878785e+01, 1.45744391317944e+01, 1.44946735727186e+01, 1.44065088581201e+01, 1.43100608659801e+01, 1.42055859513706e+01, 1.40934790749367e+01, 1.39742676708905e+01, 1.38486015336361e+01, 1.37172392068556e+01, 1.35810315270002e+01, 1.34409030873012e+01, 1.32978324373684e+01, 1.31528318151809e+01, 1.30069271314044e+01, 1.28611388081880e+01, 1.27164639383529e+01, 1.25738600976818e+01, 1.24342310282667e+01, 1.22984143214086e+01, 1.21671711631436e+01, 1.20411781577842e+01, 1.19210212072514e+01, 1.18071913905072e+01, 1.17000827553606e+01, 1.15999919046679e+01, 1.15071192327958e+01, 1.14215716488812e+01, 1.13433666128727e+01, 1.12724373090331e+01, 1.12086387884041e+01, 1.11517549242929e+01, 1.11015060400214e+01, 1.10575570827753e+01, 1.10195262285849e+01, 1.09869938092918e+01, 1.09595114518036e+01, 1.09366113131302e+01, 1.09178152827234e+01, 1.09026440084193e+01, 1.08906255863215e+01, 1.08813037410006e+01, 1.08742453131759e+01, 1.08690468699656e+01, 1.08653402596574e+01, 1.08627969497709e+01, 1.08611310140906e+01, 1.08601006705312e+01, 1.08595083155527e+01, 1.08591990500438e+01, 1.08590577434332e+01, 1.08590047343407e+01, 1.08589903145724e+01, 1.08589881862445e+01, 1.08589880264702e+01, 1.08589852142622e+01, 1.08589661611376e+01, 1.08588961198159e+01, 1.08587094152142e+01, 1.08583008159797e+01, 1.08575182965758e+01, 1.08561574794265e+01, 1.08539580121077e+01, 1.08506020945539e+01, 1.08457153247843e+01, 1.08388699774504e+01, 1.08295907662255e+01, 1.08173630677458e+01, 1.08016435013386e+01, 1.07818726662344e+01, 1.07574897389921e+01, 1.07279485327928e+01, 1.06927345229805e+01, 1.06513822569225e+01, 1.06034924987237e+01, 1.05487484181742e+01, 1.04869301250446e+01, 1.04179268789491e+01, 1.03417463732278e+01, 1.02585205973123e+01, 1.01685079213830e+01, 1.00720912126175e+01, 9.96977197466607e+00, 9.86216069043728e+00, 9.74996373134781e+00, 9.63396736224424e+00, 9.51501950905029e+00, 9.39401005581740e+00, 9.27185049134911e+00, 9.14945372818043e+00, 9.02771486783398e+00, 8.90749359013354e+00, 8.78959870984725e+00, 8.67477528383901e+00, 8.56369448129957e+00, 8.45694626385739e+00, 8.35503477505339e+00, 8.25837622006683e+00, 8.16729893267964e+00, 8.08204527878086e+00, 8.00277503182493e+00, 7.92956987019993e+00, 7.86243868231457e+00, 7.80132341465880e+00, 7.74610525377041e+00, 7.69661098859766e+00, 7.65261945020131e+00, 7.61386796757936e+00, 7.58005880951517e+00, 7.55086560187091e+00, 7.52593971782711e+00, 7.50491663619021e+00, 7.48742225169281e+00, 7.47307910334212e+00, 7.46151246484072e+00, 7.45235621764375e+00, 7.44525840512128e+00, 7.43988634823488e+00, 7.43593119146204e+00, 7.43311174423846e+00, 7.43117748906958e+00, 7.42991064300537e+00, 7.42912718383074e+00, 7.42867678473419e+00, 7.42844163933500e+00, 7.42833420026368e+00, 7.42829389628885e+00, 7.42828293264275e+00, 7.42828131443200e+00, 7.42828119332181e+00, 7.42827906656640e+00, 7.42826465749279e+00, 7.42821168832286e+00, 7.42807049280000e+00, 7.42776149409654e+00, 7.42716974088058e+00, 7.42614072798517e+00, 7.42447770630508e+00, 7.42194066083318e+00, 7.41824710454977e+00, 7.41307479711083e+00, 7.40606644876089e+00, 7.39683640981569e+00, 7.38497927353211e+00, 7.37008023574672e+00, 7.35172696071796e+00, 7.32952260363961e+00, 7.30309954288882e+00, 7.27213328757451e+00, 7.23635595782219e+00, 7.19556869606493e+00, 7.14965236594810e+00, 7.09857593746418e+00, 7.04240204517600e+00, 6.98128933886238e+00, 6.91549141558036e+00, 6.84535231705749e+00, 6.77129878060320e+00, 6.69382962710523e+00, 6.61350283772314e+00, 6.53092099544006e+00, 6.44671583707452e+00, 6.36153267036216e+00, 6.27601536103675e+00, 6.19079249495437e+00, 6.10646518400530e+00, 6.02359682871320e+00, 5.94270499237625e+00, 5.86425539674657e+00, 5.78865792915028e+00, 5.71626446241724e+00, 5.64736823396937e+00, 5.58220450673405e+00, 5.52095223709622e+00, 5.46373649720910e+00, 5.41063143366071e+00, 5.36166358538499e+00, 5.31681542565138e+00, 5.27602903220965e+00, 5.23920982378949e+00, 5.20623032884601e+00, 5.17693397320568e+00, 5.15113888714173e+00, 5.12864173975942e+00, 5.10922160995194e+00, 5.09264389927869e+00, 5.07866428371338e+00, 5.06703268924065e+00, 5.05749726182651e+00, 5.04980828659535e+00, 5.04372199548988e+00, 5.03900418872717e+00, 5.03543358443552e+00, 5.03280480426879e+00, 5.03093090159433e+00, 5.02964534369540e+00, 5.02880337049869e+00, 5.02828266928088e+00, 5.02798332675625e+00, 5.02782704558304e+00, 5.02775564001822e+00, 5.02772885342176e+00, 5.02772156682711e+00, 5.02772049134133e+00, 5.02772041081967e+00, 5.02771899641633e+00, 5.02770941363086e+00, 5.02767418637613e+00, 5.02758028405095e+00, 5.02737478416480e+00, 5.02698123900706e+00, 5.02629689731013e+00, 5.02519091752819e+00, 5.02350369240580e+00, 5.02104738379392e+00, 5.01760774067799e+00, 5.01294724051161e+00, 5.00680955286431e+00, 4.99892527442692e+00, 4.98901882598775e+00, 4.97681633694744e+00, 4.96205427477502e+00, 4.94448851065333e+00, 4.92390345488727e+00, 4.90012085363617e+00, 4.87300781916590e+00, 4.84248367477033e+00, 4.80852523601327e+00, 4.77117022184158e+00, 4.73051858838578e+00, 4.68673169709647e+00, 4.64002935649767e+00, 4.59068490095795e+00, 4.53901857843180e+00, 4.48538960214790e+00, 4.43018727233916e+00, 4.37382159132702e+00, 4.31671378092576e+00, 4.25928707102765e+00, 4.20195807021955e+00, 4.14512896165528e+00, 4.08918069750309e+00, 4.03446729861520e+00, 3.98131130606516e+00, 3.93000037953736e+00, 3.88078499474056e+00, 3.83387715804206e+00, 3.78945003129057e+00, 3.74763834328630e+00, 3.70853945649032e+00, 3.67221495797929e+00, 3.63869265148955e+00, 3.60796884123942e+00, 3.58001081618800e+00, 3.55475946335934e+00, 3.53213195874088e+00, 3.51202450223193e+00, 3.49431507778294e+00, 3.47886623034374e+00, 3.46552785713737e+00, 3.45414001214232e+00, 3.44453571990889e+00, 3.43654378864255e+00, 3.42999160376261e+00, 3.42470787293966e+00, 3.42052528307380e+00, 3.41728301996547e+00, 3.41482909367349e+00, 3.41302240773863e+00, 3.41173450934944e+00, 3.41085096062137e+00, 3.41027227857727e+00, 3.40991440291201e+00, 3.40970866558489e+00, 3.40960125379591e+00, 3.40955217683615e+00, 3.40953376643469e+00, 3.40952875836482e+00, 3.40952801918440e+00, 3.40952796369512e+00, 3.40952698703725e+00, 3.40952037003992e+00, 3.40949604530942e+00, 3.40943120490536e+00, 3.40928930527965e+00, 3.40901755823297e+00, 3.40854501197236e+00, 3.40778131638129e+00, 3.40661625484100e+00, 3.40492011060852e+00, 3.40254491786625e+00, 3.39932662505719e+00, 3.39508817013396e+00, 3.38964343344275e+00, 3.38280199439958e+00, 3.37437457407714e+00, 3.36417899953128e+00, 3.35204648042237e+00, 3.33782794835254e+00, 3.32140017897297e+00, 3.30267140091646e+00, 3.28158609791769e+00, 3.25812873370796e+00, 3.23232617412305e+00, 3.20424864575890e+00, 3.17400915146985e+00, 3.14176135386223e+00, 3.10769603088044e+00, 3.07203629397891e+00, 3.03503183079361e+00, 2.99695248356558e+00, 2.95808149710655e+00, 2.91870876430163e+00, 2.87912436503947e+00, 2.83961264137632e+00, 2.80044698544780e+00, 2.76188544598370e+00, 2.72416719262803e+00, 2.68750982115918e+00, 2.65210744101807e+00, 2.61812946032487e+00, 2.58571997145294e+00, 2.55499763933825e+00, 2.52605600152749e+00, 2.49896410017696e+00, 2.47376737913831e+00, 2.45048879208582e+00, 2.42913007931244e+00, 2.40967318089011e+00, 2.39208176224027e+00, 2.37630283482814e+00, 2.36226845972854e+00, 2.34989752522476e+00, 2.33909759136376e+00, 2.32976679446788e+00, 2.32179580300081e+00, 2.31506981300179e+00, 2.30947056675349e+00, 2.30487837279071e+00, 2.30117409927258e+00, 2.29824110671713e+00, 2.29596708078332e+00, 2.29424572183125e+00, 2.29297824598929e+00, 2.29207465287179e+00, 2.29145471821029e+00, 2.29104867554844e+00, 2.29079755963158e+00, 2.29065319479095e+00, 2.29057782387781e+00, 2.29054338640928e+00, 2.29053046774600e+00, 2.29052695355826e+00, 2.29052643487158e+00, 2.29052639576864e+00, 2.29052570531502e+00, 2.29052102739175e+00, 2.29050383087472e+00, 2.29045799145643e+00, 2.29035767393606e+00, 2.29016555742925e+00, 2.28983147630619e+00, 2.28929154217331e+00, 2.28846780341081e+00, 2.28726848821505e+00, 2.28558886416690e+00, 2.28331273141750e+00, 2.28031454713312e+00, 2.27646215564394e+00, 2.27162007203110e+00, 2.26565323751211e+00, 2.25843113448281e+00, 2.24983211966617e+00, 2.23974780826432e+00, 2.22808732333833e+00, 2.21478121574780e+00, 2.19978486317341e+00, 2.18308117326753e+00, 2.16468244567687e+00, 2.14463128884881e+00, 2.12300053702672e+00, 2.09989216646272e+00, 2.07543526296873e+00, 2.04978314103412e+00, 2.02310975420971e+00, 1.99560556486066e+00, 1.96747305763771e+00, 1.93892208526824e+00, 1.91016522863099e+00, 1.88141333726803e+00, 1.85287139351706e+00, 1.82473481542443e+00, 1.79718628261068e+00, 1.77039313732737e+00, 1.74450538197550e+00, 1.71965426608115e+00, 1.69595143160058e+00, 1.67348856652239e+00, 1.65233750364867e+00, 1.63255069426148e+00, 1.61416198472549e+00, 1.59718762715672e+00, 1.58162746203799e+00, 1.56746621988832e+00, 1.55467489958615e+00, 1.54321219159852e+00, 1.53302592423845e+00, 1.52405451943433e+00, 1.51622845084824e+00, 1.50947170124803e+00, 1.50370321775542e+00, 1.49883836308084e+00, 1.49479035840378e+00, 1.49147170959133e+00, 1.48879560350299e+00, 1.48667725581542e+00, 1.48503518675528e+00, 1.48379239697559e+00, 1.48287741310394e+00, 1.48222517166646e+00, 1.48177771142362e+00, 1.48148464772590e+00, 1.48130340818950e+00, 1.48119921649143e+00, 1.48114481991664e+00, 1.48111996587362e+00, 1.48111064229659e+00, 1.48110810606064e+00, 1.48110773171764e+00, 1.48110770334296e+00, 1.48110720028465e+00, 1.48110379199255e+00, 1.48109126277969e+00, 1.48105786472443e+00, 1.48098477504497e+00, 1.48084480396260e+00, 1.48060140683887e+00, 1.48020804933714e+00, 1.47960796841048e+00, 1.47873436465136e+00, 1.47751105242727e+00, 1.47585358271772e+00, 1.47367083920317e+00, 1.47086709070681e+00, 1.46734446268511e+00, 1.46300576770441e+00, 1.45775761085283e+00, 1.45151366247410e+00, 1.44419796955618e+00, 1.43574816090305e+00, 1.42611839219921e+00, 1.41528187725869e+00, 1.40323286250323e+00, 1.38998792347392e+00, 1.37558649425090e+00, 1.36009058116406e+00, 1.34358365817361e+00, 1.32616878903857e+00, 1.30796606673799e+00, 1.28910949952492e+00, 1.26974350203009e+00, 1.25001916656780e+00, 1.23009049311992e+00, 1.21011074671065e+00, 1.19022908970262e+00, 1.17058760669386e+00, 1.15131880459810e+00, 1.13254363377420e+00, 1.11437004110967e+00, 1.09689203550317e+00, 1.08018922211720e+00, 1.06432674501912e+00, 1.04935556844829e+00, 1.03531302426740e+00, 1.02222355601333e+00, 1.01009959693436e+00, 9.98942529016787e-01, 9.88743680910282e-01, 9.79485333718941e-01, 9.71141713957074e-01, 9.63679961964175e-01, 9.57061071355135e-01, 9.51240800468945e-01, 9.46170560239331e-01, 9.41798284524147e-01, 9.38069288857790e-01, 9.34927122055903e-01, 9.32314412379526e-01, 9.30173706374937e-01, 9.28448294402862e-01, 9.27083012642974e-01, 9.26025007405237e-01, 9.25224444290336e-01, 9.24635142469363e-01, 9.24215113384449e-01, 9.23926983691942e-01, 9.23738284341283e-01, 9.23621591234884e-01, 9.23554507740874e-01, 9.23519485113466e-01, 9.23503483214532e-01, 9.23497480386814e-01, 9.23495847475621e-01, 9.23495606461472e-01, 9.23495588077024e-01, 9.23495260607332e-01, 9.23493041954018e-01, 9.23484885979889e-01, 9.23463145378946e-01, 9.23415567780779e-01, 9.23324455491837e-01, 9.23166024548273e-01, 9.22909995853352e-01, 9.22519447691741e-01, 9.21950954503806e-01, 9.21155031012837e-01, 9.20076893146721e-01, 9.18657537273376e-01, 9.16835126850010e-01, 9.14546660740960e-01, 9.11729880652658e-01, 9.08325357287168e-01, 9.04278677301349e-01, 8.99542637728707e-01, 8.94079343139524e-01, 8.87862095415933e-01, 8.80876968212050e-01, 8.73123968924880e-01, 8.64617710398846e-01, 8.55387541645953e-01, 8.45477119556378e-01, 8.34943438996149e-01, 8.23855373403646e-01, 8.12291808509791e-01, 8.00339475059538e-01, 7.88090600236011e-01, 7.75640500908736e-01, 7.63085235161327e-01, 7.50519413299656e-01, 7.38034248071351e-01, 7.25715898975253e-01, 7.13644140152489e-01, 7.01891357908515e-01, 6.90521864253725e-01, 6.79591498047188e-01, 6.69147475707135e-01, 6.59228448723347e-01, 6.49864724634732e-01, 6.41078610758065e-01, 6.32884844744630e-01, 6.25291082052613e-01, 6.18298416859858e-01, 6.11901919199653e-01, 6.06091176764885e-01, 6.00850834639211e-01, 5.96161130051762e-01, 5.91998422078149e-01, 5.88335718046196e-01, 5.85143199308611e-01, 5.82388749096708e-01, 5.80038484465021e-01, 5.78057292992021e-01, 5.76409373047831e-01, 5.75058774231785e-01, 5.73969932197386e-01, 5.73108189716962e-01, 5.72440293701405e-01, 5.71934856190445e-01, 5.71562766255077e-01, 5.71297539462144e-01, 5.71115592138710e-01, 5.70996429173171e-01, 5.70922736452566e-01, 5.70880372138561e-01, 5.70858254636775e-01, 5.70848149078955e-01, 5.70844358153366e-01, 5.70843326931382e-01, 5.70843174725248e-01, 5.70843163033089e-01, 5.70842953694786e-01, 5.70841535398017e-01, 5.70836321605427e-01, 5.70822423674223e-01, 5.70792009028906e-01, 5.70733763786073e-01, 5.70632482458947e-01, 5.70468805411338e-01, 5.70219122393063e-01, 5.69855657563814e-01, 5.69346747633467e-01, 5.68657319823402e-01, 5.67749570046352e-01, 5.66583833878738e-01, 5.65119633618309e-01, 5.63316874286336e-01, 5.61137150441070e-01, 5.58545114997379e-01, 5.55509852006693e-01, 5.52006188771699e-01, 5.48015879932393e-01, 5.43528598198499e-01, 5.38542673707532e-01, 5.33065536441832e-01, 5.27113832957674e-01, 5.20713208443240e-01, 5.13897765945565e-01, 5.06709234393611e-01, 4.99195893809485e-01, 4.91411318264760e-01, 4.83413003762983e-01, 4.75260949115217e-01, 4.67016253537430e-01, 4.58739786208028e-01, 4.50490971783047e-01, 4.42326723354071e-01, 4.34300541893236e-01, 4.26461789913738e-01, 4.18855137577609e-01, 4.11520172138935e-01, 4.04491156454604e-01, 3.97796919143168e-01, 3.91460857526141e-01, 3.85501034399326e-01, 3.79930350617393e-01, 3.74756777129107e-01, 3.69983632210612e-01, 3.65609891990189e-01, 3.61630524761814e-01, 3.58036841898652e-01, 3.54816860282071e-01, 3.51955672955427e-01, 3.49435826113378e-01, 3.47237701485845e-01, 3.45339903637168e-01, 3.43719651671805e-01, 3.42353174345838e-01, 3.41216106692706e-01, 3.40283885073838e-01, 3.39532136183074e-01, 3.38937054109394e-01, 3.38475758251455e-01, 3.38126623831627e-01, 3.37869576117726e-01, 3.37686339337388e-01, 3.37560631733716e-01, 3.37478299282381e-01, 3.37427382235383e-01, 3.37398110792265e-01, 3.37382828693862e-01, 3.37375846230053e-01, 3.37373226875414e-01, 3.37372514348260e-01, 3.37372409180780e-01, 3.37372401040383e-01, 3.37372254490688e-01, 3.37371261595642e-01, 3.37367611617665e-01, 3.37357882187625e-01, 3.37336589929038e-01, 3.37295814090197e-01, 3.37224908893752e-01, 3.37110318868589e-01, 3.36935509716852e-01, 3.36681025163038e-01, 3.36324677568143e-01, 3.35841876697209e-01, 3.35206096756167e-01, 3.34389476593216e-01, 3.33363541831083e-01, 3.32100030855209e-01, 3.30571799396374e-01, 3.28753771453172e-01, 3.26623898175612e-01, 3.24164081824350e-01, 3.21361019750434e-01, 3.18206924092294e-01, 3.14700076890226e-01, 3.10845187563025e-01, 3.06653529784377e-01, 3.02142846981191e-01, 2.97337028914448e-01, 2.92265574916456e-01, 2.86962871178957e-01, 2.81467319024003e-01, 2.75820357635958e-01, 2.70065427938619e-01, 2.64246924171481e-01, 2.58409176580493e-01, 2.52595503052802e-01, 2.46847360202139e-01, 2.41203616102695e-01, 2.35699958290951e-01, 2.30368442430262e-01, 2.25237179662557e-01, 2.20330154523701e-01, 2.15667160599764e-01, 2.11263837937885e-01, 2.07131794553334e-01, 2.03278794044628e-01, 1.99708992121678e-01, 1.96423206514602e-01, 1.93419206993632e-01, 1.90692014833985e-01, 1.88234203756501e-01, 1.86036196953893e-01, 1.84086557103363e-01, 1.82372268145777e-01, 1.80879009004840e-01, 1.79591420290357e-01, 1.78493365377712e-01, 1.77568187114048e-01, 1.76798960832095e-01, 1.76168743443533e-01, 1.75660817247866e-01, 1.75258925856568e-01, 1.74947498430463e-01, 1.74711857394620e-01, 1.74538404048683e-01, 1.74414776128550e-01, 1.74329971461148e-01, 1.74274432411679e-01, 1.74240086831631e-01, 1.74220342615754e-01, 1.74210034669628e-01, 1.74205324955601e-01, 1.74203558190861e-01, 1.74203077589112e-01, 1.74203006653343e-01, 1.74203001118101e-01, 1.74202900893476e-01, 1.74202221857584e-01, 1.74199725660415e-01, 1.74193071796893e-01, 1.74178510393624e-01, 1.74150625151881e-01, 1.74102137450561e-01, 1.74023781773698e-01, 1.73904261616848e-01, 1.73730293727715e-01, 1.73486746804308e-01, 1.73156878449483e-01, 1.72722671134181e-01, 1.72165264075609e-01, 1.71465473312530e-01, 1.70604386997745e-01, 1.69564017312432e-01, 1.68327984859218e-01, 1.66882206450168e-01, 1.65215553477111e-01, 1.63320446130456e-01, 1.61193349123337e-01, 1.58835137596526e-01, 1.56251307582735e-01, 1.53452013548666e-01, 1.50451925550762e-01, 1.47269909616611e-01, 1.43928546100234e-01, 1.40453510907882e-01, 1.36872852686516e-01, 1.33216204563560e-01, 1.29513971392697e-01, 1.25796532621472e-01, 1.22093497122179e-01, 1.18433040164192e-01, 1.14841344898686e-01, 1.11342162121048e-01, 1.07956493521343e-01, 1.04702395874995e-01, 1.01594897215008e-01, 9.86460112738854e-02, 9.58648334694701e-02, 9.32577003482038e-02, 9.08283944915595e-02, 8.85783781662326e-02, 8.65070411298063e-02, 8.46119506444225e-02, 8.28890945788032e-02, 8.13331112397577e-02, 7.99375020994688e-02, 7.86948257734537e-02, 7.75968733978658e-02, 7.66348269167167e-02, 7.57994027059468e-02, 7.50809834416302e-02, 7.44697412008151e-02, 7.39557545180531e-02, 7.35291215740261e-02, 7.31800709297801e-02, 7.28990703165735e-02, 7.26769330217252e-02, 7.25049204601120e-02, 7.23748386720828e-02, 7.22791258193851e-02, 7.22109273286762e-02, 7.21641552012287e-02, 7.21335281895037e-02, 7.21145900270415e-02, 7.21037036523029e-02, 7.20980203282551e-02, 7.20954236492588e-02, 7.20944495580440e-02, 7.20941845828236e-02, 7.20941454730735e-02, 7.20941423953150e-02, 7.20940863345803e-02, 7.20937065155240e-02, 7.20923102711595e-02, 7.20885884782305e-02, 7.20804438409814e-02, 7.20648474786429e-02, 7.20377302562483e-02, 7.19939149918076e-02, 7.19270952778004e-02, 7.18298660734220e-02, 7.16938103282203e-02, 7.15096445534584e-02, 7.12674243462459e-02, 7.09568083215544e-02, 7.05673757218293e-02, 7.00889892585650e-02, 6.95121907182923e-02, 6.88286128783915e-02, 6.80313877670926e-02, 6.71155287607732e-02, 6.60782629262400e-02, 6.49192907768992e-02, 6.36409534441853e-02, 6.22482921500946e-02, 6.07489915148278e-02, 5.91532060926525e-02, 5.74732778377509e-02, 5.57233600912994e-02, 5.39189702908658e-02, 5.20764982138715e-02, 5.02126987094869e-02, 4.83441974013488e-02, 4.64870349544810e-02, 4.46562706817630e-02, 4.28656602008886e-02, 4.11274152843575e-02, 3.94520476690232e-02, 3.78482929750206e-02, 3.63231064323273e-02, 3.48817190517821e-02, 3.35277412442707e-02, 3.22633005533817e-02, 3.10892008602721e-02, 3.00050918313630e-02, 2.90096392162980e-02, 2.81006886351278e-02, 2.72754175475924e-02, 2.65304720368889e-02, 2.58620867547137e-02, 2.52661877847373e-02, 2.47384792583603e-02, 2.42745153161437e-02, 2.38697594949926e-02, 2.35196338831329e-02, 2.32195604557788e-02, 2.29649969081578e-02, 2.27514690572264e-02, 2.25746015149966e-02, 2.24301478732372e-02, 2.23140211199322e-02, 2.22223244674069e-02, 2.21513822488006e-02, 2.20977700690455e-02, 2.20583430125814e-02, 2.20302604429108e-02, 2.20110058006568e-02, 2.19983998291013e-02, 2.19906058284067e-02, 2.19861258457712e-02, 2.19837871213571e-02, 2.19827185914312e-02, 2.19823177574625e-02, 2.19822087217096e-02, 2.19821926282933e-02, 2.19821913505953e-02, 2.19821679352137e-02, 2.19820092930192e-02, 2.19814261144649e-02, 2.19798716281048e-02, 2.19764699442848e-02, 2.19699563533098e-02, 2.19586324194267e-02, 2.19403386699497e-02, 2.19124476034622e-02, 2.18718795724513e-02, 2.18151437542815e-02, 2.17384058288062e-02, 2.16375830559343e-02, 2.15084661466562e-02, 2.13468656459455e-02, 2.11487785587465e-02, 2.09105687870874e-02, 2.06291528151112e-02, 2.03021802467092e-02, 1.99281975655978e-02, 1.95067831351636e-02, 1.90386422119388e-02, 1.85256527216495e-02, 1.79708556814404e-02, 1.73783881956382e-02, 1.67533614668258e-02, 1.61016907026834e-02, 1.54298876202148e-02, 1.47448290337806e-02, 1.40535165186932e-02, 1.33628422706522e-02, 1.26793750277940e-02, 1.20091773806337e-02, 1.13576622564580e-02, 1.07294923831560e-02, 1.01285228095165e-02, 9.55778362270365e-03, 9.01949797962942e-03, 8.51512921455111e-03, 8.04544985349764e-03, 7.61062491700670e-03, 7.21030220721985e-03, 6.84370345281593e-03, 6.50971185991267e-03, 6.20695315203079e-03, 5.93386817106278e-03, 5.68877566333725e-03, 5.46992441346106e-03, 5.27553467601649e-03, 5.10382975417656e-03, 4.95305922152056e-03, 4.82151545536169e-03, 4.70754493967007e-03, 4.60955548628306e-03, 4.52602033889349e-03, 4.45548009404462e-03, 4.39654338635620e-03, 4.34788721987189e-03, 4.30825764546228e-03, 4.27647122531738e-03, 4.25141745531477e-03, 4.23206208316174e-03, 4.21745107933447e-03, 4.20671488628699e-03, 4.19907248203089e-03, 4.19383474423540e-03, 4.19040659343750e-03, 4.18828743190147e-03, 4.18706947536407e-03, 4.18643369408755e-03, 4.18614322361321e-03, 4.18603426187047e-03, 4.18600462199237e-03, 4.18600024720917e-03, 4.18599989669133e-03, 4.18599343274195e-03, 4.18594963873176e-03, 4.18578865036068e-03, 4.18535953804877e-03, 4.18442055571688e-03, 4.18262274998682e-03, 4.17949778265125e-03, 4.17445084669765e-03, 4.16675957471568e-03, 4.15557982212533e-03, 4.13995912356313e-03, 4.11885843110582e-03, 4.09118241863382e-03, 4.05581816011136e-03, 4.01168136610863e-03, 3.95776862558250e-03, 3.89321330938230e-03, 3.81734202644424e-03, 3.72972787398851e-03, 3.63023628703729e-03, 3.51905918043371e-03, 3.39673342173796e-03, 3.26414060391045e-03, 3.12248665021286e-03, 2.97326183320479e-03, 2.81818391919586e-03, 2.65912880720975e-03, 2.49805381837120e-03, 2.33691877128450e-03, 2.17760967087344e-03, 2.02186975153751e-03, 1.87124264603966e-03, 1.72703179141820e-03, 1.59027809696255e-03, 1.46175476574228e-03, 1.34197561747129e-03, 1.23121289523453e-03, 1.12952218834295e-03, 1.03677369163080e-03, 9.52688537580570e-04, 8.76876665735438e-04, 8.08871179643234e-04, 7.48155655003302e-04, 6.94184742627613e-04, 6.46401345008485e-04, 6.04253208500438e-04, 5.67208804106656e-04, 5.34769591929773e-04, 5.06475677994445e-04, 4.81904966633806e-04, 4.60669182068330e-04, 4.42410159934070e-04, 4.26797652572475e-04, 4.13528572643429e-04, 4.02327089012999e-04, 3.92944320601885e-04, 3.85156482863636e-04, 3.78761630365432e-04, 3.73576170808576e-04, 3.69432169094353e-04, 3.66175702382066e-04, 3.63666013846853e-04, 3.61775137598386e-04, 3.60387774648530e-04, 3.59401289671268e-04, 3.58725734535881e-04, 3.58283816204619e-04, 3.58010733648939e-04, 3.57853816865372e-04, 3.57771914977434e-04, 3.57734498557249e-04, 3.57720463206590e-04, 3.57716645325468e-04, 3.57716081822446e-04, 3.57716036290867e-04, 3.57715191867669e-04, 3.57709470875989e-04, 3.57688440445258e-04, 3.57632385295403e-04, 3.57509731649941e-04, 3.57274918499684e-04, 3.56866837684322e-04, 3.56207975357847e-04, 3.55204400548814e-04, 3.53746760602291e-04, 3.51712454637720e-04, 3.48969156362955e-04, 3.45379828530668e-04, 3.40809283798649e-04, 3.35132160827811e-04, 3.28241879230973e-04, 3.20059743174284e-04, 3.10543003926966e-04, 2.99690565142896e-04, 2.87545333350181e-04, 2.74193112296928e-04, 2.59759453186348e-04, 2.44407512212774e-04, 2.28339198905303e-04, 2.11795701105189e-04, 1.95047099346368e-04, 1.78369519204580e-04, 1.62023671059330e-04, 1.46238993998793e-04, 1.31199751102429e-04, 1.17031341524410e-04, 1.03847465972251e-04, 9.17343040711671e-05, 8.07241167564143e-05, 7.08220145426708e-05, 6.19602449418721e-05, 5.40192027358444e-05, 4.70007392152638e-05, 4.07883879418591e-05, 3.53037027522873e-05, 3.05023067497245e-05, 2.62352294981838e-05, 2.21977152250474e-05, 1.82615716823875e-05, 1.44162988675589e-05, 1.06292259047791e-05, 7.10002306902879e-06, 4.06343283287924e-06, 1.67532653705393e-06, 6.79336324083124e-07, 6.95403173267550e-07, 0.00000000000000e+00, 7.51314750857546e-07, 8.96890831825203e-07, 8.46505479046300e-07, 7.21003686976474e-07, 6.05785622553867e-07, 5.37655672872006e-07, 5.09456278052767e-07, 4.93294372461150e-07, 4.67855905128545e-07, 4.25324977015862e-07, 3.67855044203225e-07, 3.02215717340459e-07, -7.10542735760100e-15, -7.10542735760100e-15, 0.00000000000000e+00, 0.00000000000000e+00, -7.10542735760100e-15, 0.00000000000000e+00, -7.10542735760100e-15, -7.10542735760100e-15, 0.00000000000000e+00, 0.00000000000000e+00],
[ 3.50000000000000e+01, 3.49999999722420e+01, 3.49999991139108e+01, 3.49999932985816e+01, 3.49999719206792e+01, 3.49999149339905e+01, 3.49997902146872e+01, 3.49995513418808e+01, 3.49991358761607e+01, 3.49984642015034e+01, 3.49974389788647e+01, 3.49959452412884e+01, 3.49938511411627e+01, 3.49910093410616e+01, 3.49872590211471e+01, 3.49824284590497e+01, 3.49763381230515e+01, 3.49688042066439e+01, 3.49596425223018e+01, 3.49486726645399e+01, 3.49357223467109e+01, 3.49206318121388e+01, 3.49032582175292e+01, 3.48834798847367e+01, 3.48612003156126e+01, 3.48363518638851e+01, 3.48088989582315e+01, 3.47788407726726e+01, 3.47462132451921e+01, 3.47110903541554e+01, 3.46735845756591e+01, 3.46338464639432e+01, 3.45920633214371e+01, 3.45484569541726e+01, 3.45032805407645e+01, 3.44568146769407e+01, 3.44093626903818e+01, 3.43612453500182e+01, 3.43127951178135e+01, 3.42643501078424e+01, 3.42162479262092e+01, 3.41688195658960e+01, 3.41223835234930e+01, 3.40772402910473e+01, 3.40336673574340e+01, 3.39919148313315e+01, 3.39522017736713e+01, 3.39147133027312e+01, 3.38795985109826e+01, 3.38469692101698e+01, 3.38168995003783e+01, 3.37894261402374e+01, 3.37645496788987e+01, 3.37422362959143e+01, 3.37224202824184e+01, 3.37050070859395e+01, 3.36898768316047e+01, 3.36768882244375e+01, 3.36658827309360e+01, 3.36566889333115e+01, 3.36491269468685e+01, 3.36430127902592e+01, 3.36381625999802e+01, 3.36343965846928e+01, 3.36315426218524e+01, 3.36294394087442e+01, 3.36279390922293e+01, 3.36269093160741e+01, 3.36262346413079e+01, 3.36258173131609e+01, 3.36255773672251e+01, 3.36254520869490e+01, 3.36253948437866e+01, 3.36253733696462e+01, 3.36253675281354e+01, 3.36253666659398e+01, 3.36253665528379e+01, 3.36253639176995e+01, 3.36253460642316e+01, 3.36252804327381e+01, 3.36251054812978e+01, 3.36247225943283e+01, 3.36239892815788e+01, 3.36227139244245e+01, 3.36206522857936e+01, 3.36175059539002e+01, 3.36129228390748e+01, 3.36064997883525e+01, 3.35977873252010e+01, 3.35862964630610e+01, 3.35715074824908e+01, 3.35528805038136e+01, 3.35298676311399e+01, 3.35019263899357e+01, 3.34685341289432e+01, 3.34292030080721e+01, 3.33834951469849e+01, 3.33310374656800e+01, 3.32715357113882e+01, 3.32047871408009e+01, 3.31306913204192e+01, 3.30492585292310e+01, 3.29606153048534e+01, 3.28650067715488e+01, 3.27627955255629e+01, 3.26544570226318e+01, 3.25405715999990e+01, 3.24218134515467e+01, 3.22989370385754e+01, 3.21727615415036e+01, 3.20441540264660e+01, 3.19140120110109e+01, 3.17832460693561e+01, 3.16527630319028e+01, 3.15234502223944e+01, 3.13961610566591e+01, 3.12717022143949e+01, 3.11508225005442e+01, 3.10342034406886e+01, 3.09224516058387e+01, 3.08160926326258e+01, 3.07155668898847e+01, 3.06212267362561e+01, 3.05333353107903e+01, 3.04520667960936e+01, 3.03775080892639e+01, 3.03096618089023e+01, 3.02484505568822e+01, 3.01937223417164e+01, 3.01452570568286e+01, 3.01027738922939e+01, 3.00659395430882e+01, 3.00343770609841e+01, 3.00076751814475e+01, 2.99853979418741e+01, 2.99670943941237e+01, 2.99523082036059e+01, 2.99405869203348e+01, 2.99314907056055e+01, 2.99246003023637e+01, 2.99195240488208e+01, 2.99159037539113e+01, 2.99134192798728e+01, 2.99117917110817e+01, 2.99107850283334e+01, 2.99102062525752e+01, 2.99099040698545e+01, 2.99097659978780e+01, 2.99097142019576e+01, 2.99097001121778e+01, 2.99096980325547e+01, 2.99096978489635e+01, 2.99096942515853e+01, 2.99096698788087e+01, 2.99095802817591e+01, 2.99093414478091e+01, 2.99088187599118e+01, 2.99078177284466e+01, 2.99060768586878e+01, 2.99032629686794e+01, 2.98989692165470e+01, 2.98927160326419e+01, 2.98839550795572e+01, 2.98720762811560e+01, 2.98564178698253e+01, 2.98362792995912e+01, 2.98109367631675e+01, 2.97796609367719e+01, 2.97417364628430e+01, 2.96964825745768e+01, 2.96432741758838e+01, 2.95815626248684e+01, 2.95108954366886e+01, 2.94309341291033e+01, 2.93414694842925e+01, 2.92424335923939e+01, 2.91339081696608e+01, 2.90161287971995e+01, 2.88894848926286e+01, 2.87545153947747e+01, 2.86119003017159e+01, 2.84624483507867e+01, 2.83070812655970e+01, 2.81468151214325e+01, 2.79827394958591e+01, 2.78159951688181e+01, 2.76477512014242e+01, 2.74791822365930e+01, 2.73114468126836e+01, 2.71456673608161e+01, 2.69829123815767e+01, 2.68241810964124e+01, 2.66703906780987e+01, 2.65223660128855e+01, 2.63808318500581e+01, 2.62464071523163e+01, 2.61196014598696e+01, 2.60008131044869e+01, 2.58903291401934e+01, 2.57883268833568e+01, 2.56948769710169e+01, 2.56099478517116e+01, 2.55334116196039e+01, 2.54650510931109e+01, 2.54045680258155e+01, 2.53515923217534e+01, 2.53056921100379e+01, 2.52663845156060e+01, 2.52331469439908e+01, 2.52054286790033e+01, 2.51826625739882e+01, 2.51642766012390e+01, 2.51497050118861e+01, 2.51383988519079e+01, 2.51298355806285e+01, 2.51235275476355e+01, 2.51190291034778e+01, 2.51159421491671e+01, 2.51139199690254e+01, 2.51126692396659e+01, 2.51119501630404e+01, 2.51115747311161e+01, 2.51114031910861e+01, 2.51113388401616e+01, 2.51113213351180e+01, 2.51113187514091e+01, 2.51113185441637e+01, 2.51113147194344e+01, 2.51112888063220e+01, 2.51111935468520e+01, 2.51109396190994e+01, 2.51103838994498e+01, 2.51093196107774e+01, 2.51074687441861e+01, 2.51044770914876e+01, 2.50999121672957e+01, 2.50932642328179e+01, 2.50839505570486e+01, 2.50713229639769e+01, 2.50546786159367e+01, 2.50332738735965e+01, 2.50063409540433e+01, 2.49731069836004e+01, 2.49328149171634e+01, 2.48847456787249e+01, 2.48282407777123e+01, 2.47627245827467e+01, 2.46877253977758e+01, 2.46028944924243e+01, 2.45080222924500e+01, 2.44030510364058e+01, 2.42880833451722e+01, 2.41633863220074e+01, 2.40293909898002e+01, 2.38866870665758e+01, 2.37360132689390e+01, 2.35782435076330e+01, 2.34143694939386e+01, 2.32454804057822e+01, 2.30727403637038e+01, 2.28973645339093e+01, 2.27205947027389e+01, 2.25436751494673e+01, 2.23678295811968e+01, 2.21942397885980e+01, 2.20240265437470e+01, 2.18582331050990e+01, 2.16978115355620e+01, 2.15436118927805e+01, 2.13963742279588e+01, 2.12567232378908e+01, 2.11251653562162e+01, 2.10020880417042e+01, 2.08877610178488e+01, 2.07823392319689e+01, 2.06858673258924e+01, 2.05982854375945e+01, 2.05194361787112e+01, 2.04490726531123e+01, 2.03868673946346e+01, 2.03324221068197e+01, 2.02852780842275e+01, 2.02449271844483e+01, 2.02108232037216e+01, 2.01823934888939e+01, 2.01590505964361e+01, 2.01402037877493e+01, 2.01252701314899e+01, 2.01136849705916e+01, 2.01049115063154e+01, 2.00984492558954e+01, 2.00938411554697e+01, 2.00906791065663e+01, 2.00886078021719e+01, 2.00873267162120e+01, 2.00865901961412e+01, 2.00862056596571e+01, 2.00860299602473e+01, 2.00859640490879e+01, 2.00859461196313e+01, 2.00859434732802e+01, 2.00859432691681e+01, 2.00859396039705e+01, 2.00859147717121e+01, 2.00858234856236e+01, 2.00855801498134e+01, 2.00850476122287e+01, 2.00840277271056e+01, 2.00822541009487e+01, 2.00793873491897e+01, 2.00750131363871e+01, 2.00686432102281e+01, 2.00597195672724e+01, 2.00476218049070e+01, 2.00316776183179e+01, 2.00111762934024e+01, 1.99853849278990e+01, 1.99535669870334e+01, 1.99150026722851e+01, 1.98690104602664e+01, 1.98149690627614e+01, 1.97523389793416e+01, 1.96806827711991e+01, 1.95996831880481e+01, 1.95091583352745e+01, 1.94090731777764e+01, 1.92995468365253e+01, 1.91808553343812e+01, 1.90534296742891e+01, 1.89178493668399e+01, 1.87748317447114e+01, 1.86252175892960e+01, 1.84699537346110e+01, 1.83100733967770e+01, 1.81466750033034e+01, 1.79809002721369e+01, 1.78139122286194e+01, 1.76468737643806e+01, 1.74809272499451e+01, 1.73171756227500e+01, 1.71566652892740e+01, 1.70003711040640e+01, 1.68491836165904e+01, 1.67038987056489e+01, 1.65652096489505e+01, 1.64337016041455e+01, 1.63098484110490e+01, 1.61940115687366e+01, 1.60864412002374e+01, 1.59872787942747e+01, 1.58965615074674e+01, 1.58142278185043e+01, 1.57401243431694e+01, 1.56740136403164e+01, 1.56155828590140e+01, 1.55644530923467e+01, 1.55201893114387e+01, 1.54823107533338e+01, 1.54503016288087e+01, 1.54236220024203e+01, 1.54017186792085e+01, 1.53840359130541e+01, 1.53700257335485e+01, 1.53591576741561e+01, 1.53509276770502e+01, 1.53448659514311e+01, 1.53405435739386e+01, 1.53375776426886e+01, 1.53356348303675e+01, 1.53344332256692e+01, 1.53337424042759e+01, 1.53333817280259e+01, 1.53332169308752e+01, 1.53331551095815e+01, 1.53331382926725e+01, 1.53331358105313e+01, 1.53331356225910e+01, 1.53331322932482e+01, 1.53331097364568e+01, 1.53330268152436e+01, 1.53328057772600e+01, 1.53323220386344e+01, 1.53313956120812e+01, 1.53297845196986e+01, 1.53271804886879e+01, 1.53232071793714e+01, 1.53174211384892e+01, 1.53093156055172e+01, 1.52983272243313e+01, 1.52838456259885e+01, 1.52652257506599e+01, 1.52418026691136e+01, 1.52129085496133e+01, 1.51778912995851e+01, 1.51361342996716e+01, 1.50870765491559e+01, 1.50302324654153e+01, 1.49652105353655e+01, 1.48917300120839e+01, 1.48096348911892e+01, 1.47189044921892e+01, 1.46196601088842e+01, 1.45121673742923e+01, 1.43968341986206e+01, 1.42742043678748e+01, 1.41449471162770e+01, 1.40098431864628e+01, 1.38697680475189e+01, 1.37256730374919e+01, 1.35785652277972e+01, 1.34294867761059e+01, 1.32794944554507e+01, 1.31296399399980e+01, 1.29809513123451e+01, 1.28344161489988e+01, 1.26909664481600e+01, 1.25514655878681e+01, 1.24166974387005e+01, 1.22873576977108e+01, 1.21640474547377e+01, 1.20472689472843e+01, 1.19374234075840e+01, 1.18348108589058e+01, 1.17396316816141e+01, 1.16519897459518e+01, 1.15718968990634e+01, 1.14992785974402e+01, 1.14339804900372e+01, 1.13757757779427e+01, 1.13243731994494e+01, 1.12794255106864e+01, 1.12405383484286e+01, 1.12072793711441e+01, 1.11791875758717e+01, 1.11557826824036e+01, 1.11365744638593e+01, 1.11210718862722e+01, 1.11087919020637e+01, 1.10992677262973e+01, 1.10920564133696e+01, 1.10867455479156e+01, 1.10829588691641e+01, 1.10803606638951e+01, 1.10786587897067e+01, 1.10776062266951e+01, 1.10770011001940e+01, 1.10766851675121e+01, 1.10765408147778e+01, 1.10764866630054e+01, 1.10764719324030e+01, 1.10764697581967e+01, 1.10764695949783e+01, 1.10764667221502e+01, 1.10764472583166e+01, 1.10763757072022e+01, 1.10761849781772e+01, 1.10757675722489e+01, 1.10749681894960e+01, 1.10735780547396e+01, 1.10713312235930e+01, 1.10679030907998e+01, 1.10629112759026e+01, 1.10559190063113e+01, 1.10464410532665e+01, 1.10339522009144e+01, 1.10178981424969e+01, 1.09977086015849e+01, 1.09728123730336e+01, 1.09426538723922e+01, 1.09067106800056e+01, 1.08645114745217e+01, 1.08156536781913e+01, 1.07598200913693e+01, 1.06967937831063e+01, 1.06264705337914e+01, 1.05488681968856e+01, 1.04641324590452e+01, 1.03725386271951e+01, 1.02744892499985e+01, 1.01705075796058e+01, 1.00612270853431e+01, 9.94737743060479e+00, 9.82976750360525e+00, 9.70926623834153e+00, 9.58678206239033e+00, 9.46324185439270e+00, 9.33957028209040e+00, 9.21667032279564e+00, 9.09540564914088e+00, 8.97658540595083e+00, 8.86095172486689e+00, 8.74917013926787e+00, 8.64182288962215e+00, 8.53940496234175e+00, 8.44232259278937e+00, 8.35089388962709e+00, 8.26535120273042e+00, 8.18584485608224e+00, 8.11244789345360e+00, 8.04516153006445e+00, 7.98392105953488e+00, 7.92860202486509e+00, 7.87902651888243e+00, 7.83496952912699e+00, 7.79616528164762e+00, 7.76231356617098e+00, 7.73308604125807e+00, 7.70813252294658e+00, 7.68708725525123e+00, 7.66957514754143e+00, 7.65521794441208e+00, 7.64364027067019e+00, 7.63447547011136e+00, 7.62737113452625e+00, 7.62199420140821e+00, 7.61803548737286e+00, 7.61521352109909e+00, 7.61327754573989e+00, 7.61200957652532e+00, 7.61122542413446e+00, 7.61077462699295e+00, 7.61053927391903e+00, 7.61043173999147e+00, 7.61039140043798e+00, 7.61038042711411e+00, 7.61037880747494e+00, 7.61037868625786e+00, 7.61037655762516e+00, 7.61036213583285e+00, 7.61030911991080e+00, 7.61016779978800e+00, 7.60985852852268e+00, 7.60926625378519e+00, 7.60823633541733e+00, 7.60657185415033e+00, 7.60403259091508e+00, 7.60033582498942e+00, 7.59515906036158e+00, 7.58814474089383e+00, 7.57890695427707e+00, 7.56704005144351e+00, 7.55212902259547e+00, 7.53376137578352e+00, 7.51154016369051e+00, 7.48509770579846e+00, 7.45410946513359e+00, 7.41830747114545e+00, 7.37749264291877e+00, 7.33154536857770e+00, 7.28043374352950e+00, 7.22421896428897e+00, 7.16305751327270e+00, 7.09719994505931e+00, 7.02698628303846e+00, 6.95283824007506e+00, 6.87524866873504e+00, 6.79476880710856e+00, 6.71199399969554e+00, 6.62754862887829e+00, 6.54207098777384e+00, 6.45619876366086e+00, 6.37055569322970e+00, 6.28573981187894e+00, 6.20231356664665e+00, 6.12079591306942e+00, 6.04165638437768e+00, 5.96531101658466e+00, 5.89211993979409e+00, 5.82238640434093e+00, 5.75635699645954e+00, 5.69422280609083e+00, 5.63612133242310e+00, 5.58213894437991e+00, 5.53231374816887e+00, 5.48663874824372e+00, 5.44506521904491e+00, 5.40750623124592e+00, 5.37384029732927e+00, 5.34391511701602e+00, 5.31755141349748e+00, 5.29454685679352e+00, 5.27468007117601e+00, 5.25771471981413e+00, 5.24340365211513e+00, 5.23149308833028e+00, 5.22172680277216e+00, 5.21385025257459e+00, 5.20761458463412e+00, 5.20278044063159e+00, 5.19912147027160e+00, 5.19642745739967e+00, 5.19450696349702e+00, 5.19318939885494e+00, 5.19232644366173e+00, 5.19179275892775e+00, 5.19148594975564e+00, 5.19132576961903e+00, 5.19125258242570e+00, 5.19122512745299e+00, 5.19121765904136e+00, 5.19121655671968e+00, 5.19121647418881e+00, 5.19121502449275e+00, 5.19120520259287e+00, 5.19116909631331e+00, 5.19107285071534e+00, 5.19086222204885e+00, 5.19045885242120e+00, 5.18975741887360e+00, 5.18862379528685e+00, 5.18689434754983e+00, 5.18437646066755e+00, 5.18085036956852e+00, 5.17607233182490e+00, 5.16977913911160e+00, 5.16169391443340e+00, 5.15153308440981e+00, 5.13901435205555e+00, 5.12386542891583e+00, 5.10583322103748e+00, 5.08469310723986e+00, 5.06025790729495e+00, 5.03238611848590e+00, 5.00098900684393e+00, 4.96603617706240e+00, 4.92755931235785e+00, 4.88565386849100e+00, 4.84047861747290e+00, 4.79225305629487e+00, 4.74125281328551e+00, 4.68780328884898e+00, 4.63227184988689e+00, 4.57505895288943e+00, 4.51658859791321e+00, 4.45729851618837e+00, 4.39763047202776e+00, 4.33802102023940e+00, 4.27889300848762e+00, 4.22064805431786e+00, 4.16366016223187e+00, 4.10827058007284e+00, 4.05478392882710e+00, 4.00346557892363e+00, 3.95454019272967e+00, 3.90819131064363e+00, 3.86456182963676e+00, 3.82375520951711e+00, 3.78583724302985e+00, 3.75083823893328e+00, 3.71875548897200e+00, 3.68955591627330e+00, 3.66317883035713e+00, 3.63953873957913e+00, 3.61852819324493e+00, 3.60002064161229e+00, 3.58387331211604e+00, 3.56993010459975e+00, 3.55802450771766e+00, 3.54798253382760e+00, 3.53962566162104e+00, 3.53277376549639e+00, 3.52724799939270e+00, 3.52287359159344e+00, 3.51948249700489e+00, 3.51691584566426e+00, 3.51502612166336e+00, 3.51367900600220e+00, 3.51275482054881e+00, 3.51214951837245e+00, 3.51177517796241e+00, 3.51155997461039e+00, 3.51144762057797e+00, 3.51139628543773e+00, 3.51137702791018e+00, 3.51137178940103e+00, 3.51137101620824e+00, 3.51137095816569e+00, 3.51136993656824e+00, 3.51136301509851e+00, 3.51133757109313e+00, 3.51126974709690e+00, 3.51112131787007e+00, 3.51083706553917e+00, 3.51034277149054e+00, 3.50954392301504e+00, 3.50832522021894e+00, 3.50655095354883e+00, 3.50406630341799e+00, 3.50069958978851e+00, 3.49626547025925e+00, 3.49056904978534e+00, 3.48341082389079e+00, 3.47459233136864e+00, 3.46392234433169e+00, 3.45122337652489e+00, 3.43633824941266e+00, 3.41913642464564e+00, 3.39951979602414e+00, 3.37742763824991e+00, 3.35284043637622e+00, 3.32578236957192e+00, 3.29632229359981e+00, 3.26457315343927e+00, 3.23068985334005e+00, 3.19486570700766e+00, 3.15732767564829e+00, 3.11833066712224e+00, 3.07815120873282e+00, 3.03708081611652e+00, 2.99541936247313e+00, 2.95346871116055e+00, 2.91152681851613e+00, 2.86988245180186e+00, 2.82881060777967e+00, 2.78856866673706e+00, 2.74939327797050e+00, 2.71149794613526e+00, 2.67507127176750e+00, 2.64027579094906e+00, 2.60724735574959e+00, 2.57609499656756e+00, 2.54690120846735e+00, 2.51972260549809e+00, 2.49459088970144e+00, 2.47151408518418e+00, 2.45047799230674e+00, 2.43144782259328e+00, 2.41436998104963e+00, 2.39917396865511e+00, 2.38577438326128e+00, 2.37407300139040e+00, 2.36396092599765e+00, 2.35532078584146e+00, 2.34802897061845e+00, 2.34195788262127e+00, 2.33697818075200e+00, 2.33296098684661e+00, 2.32978001814936e+00, 2.32731360420806e+00, 2.32544654222455e+00, 2.32407174269441e+00, 2.32309161755518e+00, 2.32241916636931e+00, 2.32197872237764e+00, 2.32170632938283e+00, 2.32154973191577e+00, 2.32146797434772e+00, 2.32143061873607e+00, 2.32141660537210e+00, 2.32141279339879e+00, 2.32141223075957e+00, 2.32141218834312e+00, 2.32141143938173e+00, 2.32140636505969e+00, 2.32138771133699e+00, 2.32133798752961e+00, 2.32122916904905e+00, 2.32102077191962e+00, 2.32065837782458e+00, 2.32007268017713e+00, 2.31917911204159e+00, 2.31787810518970e+00, 2.31605601556623e+00, 2.31358673332259e+00, 2.31033397485449e+00, 2.30615422973640e+00, 2.30090030733970e+00, 2.29442539710766e+00, 2.28658752447314e+00, 2.27725425343411e+00, 2.26630745956074e+00, 2.25364797667427e+00, 2.23919990948745e+00, 2.22291440548592e+00, 2.20477269367614e+00, 2.18478822569376e+00, 2.16300779492189e+00, 2.13951155914281e+00, 2.11441194821055e+00, 2.08785149603829e+00, 2.05999969147056e+00, 2.03104899132985e+00, 2.00121017778158e+00, 1.97070726877706e+00, 1.93977220338714e+00, 1.90863952305351e+00, 1.87754125589809e+00, 1.84670218590230e+00, 1.81633565444974e+00, 1.78664000146722e+00, 1.75779571057769e+00, 1.72996328068206e+00, 1.70328180826552e+00, 1.67786823290676e+00, 1.65381717454290e+00, 1.63120127568569e+00, 1.61007195479996e+00, 1.59046047752780e+00, 1.57237925895475e+00, 1.55582332098056e+00, 1.54077184233937e+00, 1.52718975329734e+00, 1.51502934116024e+00, 1.50423184536579e+00, 1.49472903134888e+00, 1.48644474007944e+00, 1.47929641498338e+00, 1.47319660989069e+00, 1.46805448093748e+00, 1.46377726236319e+00, 1.46027172140738e+00, 1.45744558164404e+00, 1.45520889777971e+00, 1.45347535890181e+00, 1.45216349208793e+00, 1.45119773479406e+00, 1.45050934301944e+00, 1.45003710321396e+00, 1.44972781935017e+00, 1.44953655240794e+00, 1.44942659737480e+00, 1.44936919223386e+00, 1.44934296364667e+00, 1.44933312444597e+00, 1.44933044794934e+00, 1.44933005290419e+00, 1.44933002296033e+00, 1.44932949208162e+00, 1.44932589530288e+00, 1.44931267320329e+00, 1.44927742824389e+00, 1.44920029705303e+00, 1.44905258749941e+00, 1.44879573782370e+00, 1.44838065007907e+00, 1.44774744416338e+00, 1.44682567132539e+00, 1.44553501656257e+00, 1.44378650713489e+00, 1.44148422894039e+00, 1.43852753343818e+00, 1.43481369528362e+00, 1.43024095552039e+00, 1.42471185833504e+00, 1.41813676291226e+00, 1.41043738824057e+00, 1.40155023052288e+00, 1.39142968286694e+00, 1.38005068752753e+00, 1.36741076375394e+00, 1.35353127977019e+00, 1.33845787474036e+00, 1.32225998348568e+00, 1.30502946966183e+00, 1.28687842754596e+00, 1.26793626354473e+00, 1.24834621120423e+00, 1.22826146387018e+00, 1.20784112454961e+00, 1.18724617199683e+00, 1.16663562642653e+00, 1.14616307000730e+00, 1.12597364009560e+00, 1.10620157136665e+00, 1.08696832095015e+00, 1.06838127222370e+00, 1.05053298095341e+00, 1.03350090374134e+00, 1.01734753382522e+00, 1.00212086273680e+00, 9.87855086983160e-01, 9.74571485139556e-01, 9.62279400755378e-01, 9.50977278575706e-01, 9.40653714310145e-01, 9.31288490396625e-01, 9.22853581106850e-01, 9.15314119424146e-01, 9.08629325150308e-01, 9.02753398610656e-01, 8.97636387204173e-01, 8.93225033053540e-01, 8.89463609367134e-01, 8.86294751090631e-01, 8.83660282288432e-01, 8.81502038773860e-01, 8.79762680149867e-01, 8.78386481001307e-01, 8.77320086882044e-01, 8.76513217342705e-01, 8.75919295896310e-01, 8.75495985798075e-01, 8.75205611008923e-01, 8.75015443784378e-01, 8.74897843908968e-01, 8.74830239476850e-01, 8.74794944978333e-01, 8.74778818883357e-01, 8.74772769469303e-01, 8.74771123885871e-01, 8.74770881001346e-01, 8.74770862474207e-01, 8.74770532463216e-01, 8.74768296592286e-01, 8.74760077326975e-01, 8.74738168034519e-01, 8.74690221358328e-01, 8.74598402620954e-01, 8.74438744346314e-01, 8.74180735145828e-01, 8.73787172784553e-01, 8.73214306861406e-01, 8.72412290718032e-01, 8.71325954422403e-01, 8.69895900514976e-01, 8.68059911411649e-01, 8.65754642000219e-01, 8.62917553539511e-01, 8.59489026457275e-01, 8.55414571519226e-01, 8.50647042985536e-01, 8.45148745885673e-01, 8.38893324480757e-01, 8.31867322021658e-01, 8.24071314008727e-01, 8.15520538275713e-01, 8.06244974161679e-01, 7.96288857474124e-01, 7.85709654657744e-01, 7.74576554847627e-01, 7.62968568626086e-01, 7.50972344240878e-01, 7.38679823760471e-01, 7.26185862472960e-01, 7.13585925499594e-01, 7.00973958013329e-01, 6.88440502354041e-01, 6.76071109772252e-01, 6.63945069413579e-01, 6.52134454840862e-01, 6.40703470476332e-01, 6.29708067593867e-01, 6.19195791911437e-01, 6.09205821868173e-01, 5.99769157403756e-01, 5.90908922444378e-01, 5.82640749321342e-01, 5.74973219135529e-01, 5.67908337959331e-01, 5.61442034270179e-01, 5.55564667844116e-01, 5.50261544354669e-01, 5.45513433062546e-01, 5.41297087249532e-01, 5.37585768476283e-01, 5.34349776380303e-01, 5.31556985637003e-01, 5.29173390953495e-01, 5.27163659642049e-01, 5.25491689536267e-01, 5.24121167905093e-01, 5.23016124752672e-01, 5.22141471653178e-01, 5.21463515261573e-01, 5.20950433068052e-01, 5.20572698012110e-01, 5.20303438393235e-01, 5.20118720205083e-01, 5.19997740608027e-01, 5.19922923691610e-01, 5.19879912840999e-01, 5.19857457720484e-01, 5.19847197886278e-01, 5.19843349083757e-01, 5.19842302117610e-01, 5.19842147587660e-01, 5.19842135716999e-01, 5.19841923182589e-01, 5.19840483231768e-01, 5.19835189834950e-01, 5.19821079695895e-01, 5.19790200576963e-01, 5.19731065574334e-01, 5.19628236204937e-01, 5.19462055116023e-01, 5.19208546786835e-01, 5.18839501565182e-01, 5.18322754511125e-01, 5.17622665533878e-01, 5.16700800974327e-01, 5.15516808985630e-01, 5.14029471860503e-01, 5.12197908151727e-01, 5.09982886616100e-01, 5.07348203543643e-01, 5.04262065996677e-01, 5.00698417071384e-01, 4.96638136644663e-01, 4.92070053073149e-01, 4.86991708414770e-01, 4.81409831806999e-01, 4.75340491869460e-01, 4.68808918016315e-01, 4.61849000552057e-01, 4.54502498403961e-01, 4.46817999469381e-01, 4.38849690396481e-01, 4.30655999350279e-01, 4.22298176805917e-01, 4.13838876152418e-01, 4.05340788849649e-01, 3.96865379281465e-01, 3.88471753547073e-01, 3.80215685346613e-01, 3.72148811708300e-01, 3.64318002148231e-01, 3.56764897258969e-01, 3.49525606789101e-01, 3.42630552961460e-01, 3.36104441959932e-01, 3.29966345028097e-01, 3.24229870289841e-01, 3.18903407034469e-01, 3.13990425617703e-01, 3.09489818119090e-01, 3.05396267265181e-01, 3.01700633675274e-01, 2.98390354016533e-01, 2.95449844988383e-01, 2.92860910035792e-01, 2.90603147197459e-01, 2.88654357446049e-01, 2.86990953238835e-01, 2.85588366781838e-01, 2.84421456775476e-01, 2.83464911261589e-01, 2.82693642765608e-01, 2.82083170393292e-01, 2.81609982076738e-01, 2.81251868949454e-01, 2.80988223027485e-01, 2.80800289111042e-01, 2.80671362183661e-01, 2.80586922598687e-01, 2.80534702975466e-01, 2.80504682889053e-01, 2.80489009991960e-01, 2.80481848982838e-01, 2.80479162651936e-01, 2.80478431905806e-01, 2.80478324049248e-01, 2.80478315700712e-01, 2.80478165403835e-01, 2.80477147121140e-01, 2.80473403817098e-01, 2.80463425626294e-01, 2.80441589023134e-01, 2.80399770938018e-01, 2.80327054013945e-01, 2.80209537764122e-01, 2.80030268688506e-01, 2.79769300298454e-01, 2.79403891307815e-01, 2.78908846788219e-01, 2.78257002702318e-01, 2.77419848839145e-01, 2.76368278821600e-01, 2.75073448732499e-01, 2.73507718385986e-01, 2.71645641904044e-01, 2.69464967732290e-01, 2.66947603311117e-01, 2.64080497055517e-01, 2.60856390708355e-01, 2.57274398895298e-01, 2.53340379875034e-01, 2.49067071716723e-01, 2.44473980753014e-01, 2.39587023156304e-01, 2.34437934689218e-01, 2.29063476875183e-01, 2.23504478915174e-01, 2.17804762778592e-01, 2.12010003472628e-01, 2.06166577361870e-01, 2.00320448731183e-01, 1.94516139024238e-01, 1.88795815038141e-01, 1.83198522637923e-01, 1.77759582148695e-01, 1.72510151328460e-01, 1.67476952458486e-01, 1.62682152187493e-01, 1.58143376717277e-01, 1.53873840895685e-01, 1.49882567770135e-01, 1.46174674966410e-01, 1.42751705593696e-01, 1.39611983870559e-01, 1.36750978928767e-01, 1.34161663909310e-01, 1.31834861184480e-01, 1.29759568045948e-01, 1.27923260281776e-01, 1.26312173583116e-01, 1.24911564591017e-01, 1.23705954586192e-01, 1.22679359354883e-01, 1.21815508682154e-01, 1.21098058313194e-01, 1.20510796189770e-01, 1.20037843443349e-01, 1.19663849150655e-01, 1.19374176384206e-01, 1.19155075767310e-01, 1.18993841707450e-01, 1.18878945843484e-01, 1.18800142077951e-01, 1.18748537907386e-01, 1.18716627596804e-01, 1.18698284008119e-01, 1.18688707486498e-01, 1.18684332005387e-01, 1.18682690629278e-01, 1.18682244136870e-01, 1.18682178235581e-01, 1.18682173093198e-01, 1.18682079981760e-01, 1.18681449139139e-01, 1.18679130109726e-01, 1.18672948542574e-01, 1.18659420911634e-01, 1.18633516093333e-01, 1.18588474323168e-01, 1.18515693444913e-01, 1.18404691609101e-01, 1.18243154529857e-01, 1.18017073892133e-01, 1.17710981322404e-01, 1.17308279332129e-01, 1.16791666714903e-01, 1.16143651032907e-01, 1.15347135216858e-01, 1.14386059236580e-01, 1.13246071751220e-01, 1.11915201229138e-01, 1.10384491927320e-01, 1.08648568027768e-01, 1.06706089726799e-01, 1.04560068540401e-01, 1.02218015599711e-01, 9.96919059842512e-02, 9.69979535273282e-02, 9.41562030599385e-02, 9.11899595564165e-02, 8.81250848565713e-02, 8.49892014260831e-02, 8.18108480933759e-02, 7.86186343732282e-02, 7.54404377995357e-02, 7.23026830305074e-02, 6.92297331120244e-02, 6.62434132034164e-02, 6.33626764011552e-02, 6.06034111191320e-02, 5.79783806889154e-02, 5.54972790428039e-02, 5.31668818225626e-02, 5.09912700167519e-02, 4.89721030374426e-02, 4.71089196302259e-02, 4.53994477209534e-02, 4.38399077694385e-02, 4.24252979827600e-02, 4.11496534659150e-02, 4.00062747816534e-02, 3.89879242912770e-02, 3.80869909788899e-02, 3.72956262161921e-02, 3.66058541308547e-02, 3.60096609474496e-02, 3.54990679358167e-02, 3.50661924927209e-02, 3.47033014661733e-02, 3.44028601750779e-02, 3.41575797426970e-02, 3.39604644163174e-02, 3.38048595507345e-02, 3.36844999584045e-02, 3.35935574411224e-02, 3.35266855828138e-02, 3.34790593571057e-02, 3.34464068276858e-02, 3.34250302171810e-02, 3.34118138872910e-02, 3.34042172834259e-02, 3.34002515990122e-02, 3.33984397426335e-02, 3.33977600685458e-02, 3.33975751821995e-02, 3.33975478934221e-02, 3.33975457459132e-02, 3.33975066295906e-02, 3.33972416115884e-02, 3.33962673897261e-02, 3.33936705668663e-02, 3.33879879653836e-02, 3.33771069499136e-02, 3.33581905492579e-02, 3.33276321227132e-02, 3.32810442186471e-02, 3.32132860157230e-02, 3.31185331862116e-02, 3.29903930296425e-02, 3.28220661662755e-02, 3.26065538705080e-02, 3.23369072594062e-02, 3.20065111350658e-02, 3.16093915378630e-02, 3.11405323560123e-02, 3.05961831138291e-02, 2.99741378420180e-02, 2.92739642162516e-02, 2.84971633280833e-02, 2.76472437229387e-02, 2.67296986323942e-02, 2.57518822703560e-02, 2.47227889959802e-02, 2.36527471943191e-02, 2.25530469257293e-02, 2.14355258422643e-02, 2.03121408839095e-02, 1.91945535128255e-02, 1.80937537807466e-02, 1.70197437877704e-02, 1.59812947739795e-02, 1.49857850212243e-02, 1.40391187643942e-02, 1.31457201722611e-02, 1.23085917382539e-02, 1.15294234601961e-02, 1.08087380192927e-02, 1.01460575374404e-02, 9.54007896157805e-03, 8.98884724287541e-03, 8.48991790167730e-03, 8.04050306799553e-03, 7.63759749018078e-03, 7.27808313519063e-03, 6.95881270143417e-03, 6.67667354739621e-03, 6.42863424093321e-03, 6.21177627463965e-03, 6.02331360493480e-03, 5.86060264325994e-03, 5.72114519400913e-03, 5.60258661008817e-03, 5.50271114025236e-03, 5.41943610217999e-03, 5.35080615967587e-03, 5.29498863228639e-03, 5.25027042844783e-03, 5.21505687284929e-03, 5.18787239987972e-03, 5.16736281606711e-03, 5.15229860748434e-03, 5.14157859873876e-03, 5.13423316745332e-03, 5.12942619189261e-03, 5.12645495900443e-03, 5.12474738020785e-03, 5.12385603988008e-03, 5.12344881786220e-03, 5.12329606109319e-03, 5.12325450813478e-03, 5.12324837502831e-03, 5.12324788810758e-03, 5.12323896466427e-03, 5.12317850735400e-03, 5.12295626477055e-03, 5.12236388512122e-03, 5.12106766690579e-03, 5.11858598257930e-03, 5.11427259204567e-03, 5.10730712251473e-03, 5.09669404067381e-03, 5.08127144105686e-03, 5.05973086784195e-03, 5.03064911853812e-03, 4.99253249579823e-03, 4.94387324396683e-03, 4.88321693109839e-03, 4.80923835750247e-03, 4.72082228969839e-03, 4.61714408261571e-03, 4.49774425982952e-03, 4.36259057811128e-03, 4.21212120553349e-03, 4.04726352007856e-03, 3.86942472010787e-03, 3.68045281708618e-03, 3.48256939454217e-03, 3.27827837808314e-03, 3.07025752713486e-03, 2.86124104970753e-03, 2.65390243534824e-03, 2.45074628564623e-03, 2.25401672881986e-03, 2.06562811136735e-03, 1.88712122203327e-03, 1.71964552747994e-03, 1.56396516650403e-03, 1.42048432154951e-03, 1.28928656216942e-03, 1.17018291189908e-03, 1.06276424951801e-03, 9.66454497074665e-04, 8.80561481771736e-04, 8.04322616353659e-04, 7.36943176548266e-04, 6.77626202850945e-04, 6.25594517941863e-04, 5.80106318032847e-04, 5.40465879829810e-04, 5.06030384741507e-04, 4.76213315678820e-04, 4.50484755049274e-04, 4.28369157013719e-04, 4.09441440467617e-04, 3.93322278831931e-04, 3.79673238306566e-04, 3.68192113377575e-04, 3.58608591014331e-04, 3.50680291418826e-04, 3.44189233672410e-04, 3.38938786320853e-04, 3.34751151171986e-04, 3.31465392413577e-04, 3.28935981222855e-04, 3.27031790597232e-04, 3.25635451680739e-04, 3.24642968173805e-04, 3.23963479239353e-04, 3.23519060735578e-04, 3.23244461114314e-04, 3.23086682030294e-04, 3.23004332770438e-04, 3.22966712559491e-04, 3.22952600875226e-04, 3.22948762246256e-04, 3.22948195666584e-04, 3.22948150277114e-04, 3.22947313279087e-04, 3.22941642508567e-04, 3.22920796890003e-04, 3.22865235787617e-04, 3.22743670395198e-04, 3.22510968139511e-04, 3.22106641071684e-04, 3.21454064206250e-04, 3.20460584077864e-04, 3.19018684990624e-04, 3.17008372270777e-04, 3.14300899148634e-04, 3.10763897168442e-04, 3.06267863358300e-04, 3.00693808341634e-04, 2.93941686521748e-04, 2.85939032941940e-04, 2.76649059976819e-04, 2.66077370724815e-04, 2.54276474109361e-04, 2.41347461965802e-04, 2.27438494967203e-04, 2.12740052027982e-04, 1.97477118398126e-04, 1.81898592281016e-04, 1.66264297675411e-04, 1.50830332458440e-04, 1.35834186167472e-04, 1.21481878629481e-04, 1.07939600702878e-04, 9.53312564391240e-05, 8.37409742615591e-05, 7.32174023951870e-05, 6.37763415767267e-05, 5.54008117674698e-05, 4.80414887533698e-05, 4.16222582231285e-05, 3.60527375420361e-05, 3.12434745737278e-05, 2.71159854250413e-05, 2.36025984037269e-05, 2.06376927636143e-05, 1.81484675110255e-05, 1.60552554291371e-05, 1.42820847841563e-05, 1.27676809427157e-05, 1.14704017875056e-05, 1.03667262294493e-05, 9.44249094203542e-06, 8.67934094372913e-06, 8.05329939623789e-06, 7.54708838712759e-06, 7.14421997827230e-06, 6.81751735953640e-06, 6.54177549819224e-06, 6.30411502555717e-06, 6.10152657998242e-06, 5.93482273814061e-06, 5.80324260823772e-06, 5.70230935181826e-06, 5.62562843953174e-06, 5.56747084345943e-06, 5.52366790884662e-06, 5.49126358606600e-06, 5.46797504341612e-06, 5.45187122469315e-06, 5.44125801127393e-06, 5.43466597235920e-06, 5.43086600401921e-06, 5.42887903520750e-06, 5.42797047842214e-06, 5.42762953159581e-06, 5.42753677734709e-06, 5.42752307808314e-06, 5.42752198384733e-06, 5.42750163390338e-06, 5.42736378150721e-06, 5.42685710769319e-06, 5.42550718307666e-06, 5.42255637725475e-06, 5.41691838407132e-06, 5.40715399210967e-06, 5.39147538347606e-06, 5.36778283333206e-06, 5.33372701028156e-06, 5.28676952171736e-06, 5.22419011161901e-06, 5.14298464082685e-06, 5.03964857045958e-06, 4.90995261515081e-06, 4.74892898694179e-06, 4.55132284571391e-06, 4.31274751377941e-06, 4.03193943299129e-06, 3.71657931452773e-06, 3.40774644769226e-06, 3.14156354619399e-06, 2.85622514439865e-06, 2.55021132744560e-06, 2.30788521804470e-06, 2.10202795614123e-06, 1.87119882610887e-06, 1.59960060130970e-06, 1.29620597988378e-06, 1.06259810905840e-06, 9.05527230088410e-07, 7.23255816126311e-07, 7.54956694493103e-07, 7.81254058779268e-07, 7.49476605221844e-07, 6.23404261546057e-07, 3.73606027892492e-07, -7.10542735760100e-15, 4.18778952848697e-07, 4.44671726995693e-07, 4.13189980008610e-07, 4.11653083176589e-07, -7.10542735760100e-15, -7.10542735760100e-15, 7.01768612998421e-07, 9.82704193575046e-07, 1.10073239056874e-06, 1.04456108829254e-06, 8.43878027012579e-07, 5.54533166052806e-07, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15],
[ 3.60000000000000e+01, 3.59999999710689e+01, 3.59999990764622e+01, 3.59999930153608e+01, 3.59999707339679e+01, 3.59999113388582e+01, 3.59997813485471e+01, 3.59995323802196e+01, 3.59990993554544e+01, 3.59983992930623e+01, 3.59973307393539e+01, 3.59957738668831e+01, 3.59935912527007e+01, 3.59906293271582e+01, 3.59867204650916e+01, 3.59816856734824e+01, 3.59753378140340e+01, 3.59674852858928e+01, 3.59579360831736e+01, 3.59465021339226e+01, 3.59330038213579e+01, 3.59172745841918e+01, 3.58991654899969e+01, 3.58785496734662e+01, 3.58553265297556e+01, 3.58294255519662e+01, 3.58008097016950e+01, 3.57694782033043e+01, 3.57354686572927e+01, 3.56988583770845e+01, 3.56597648677673e+01, 3.56183453853932e+01, 3.55747955414047e+01, 3.55293469477454e+01, 3.54822639327269e+01, 3.54338393935732e+01, 3.53843898862455e+01, 3.53342500841088e+01, 3.52837667619800e+01, 3.52332924794201e+01, 3.51831791458571e+01, 3.51337716501372e+01, 3.50854017290154e+01, 3.50383822341371e+01, 3.49930019368323e+01, 3.49495209862902e+01, 3.49081671111157e+01, 3.48691326283593e+01, 3.48325722990312e+01, 3.47986020456583e+01, 3.47672985260877e+01, 3.47386995386524e+01, 3.47128052169615e+01, 3.46895799577912e+01, 3.46689550126232e+01, 3.46508316621064e+01, 3.46350848829830e+01, 3.46215674087753e+01, 3.46101140788521e+01, 3.46005463655215e+01, 3.45926769657758e+01, 3.45863143434796e+01, 3.45812671094184e+01, 3.45773481309079e+01, 3.45743782697452e+01, 3.45721896571825e+01, 3.45706284272107e+01, 3.45695568445303e+01, 3.45688547807695e+01, 3.45684205113226e+01, 3.45681708250295e+01, 3.45680404591794e+01, 3.45679808923237e+01, 3.45679585464773e+01, 3.45679524678405e+01, 3.45679515706455e+01, 3.45679514529524e+01, 3.45679487108452e+01, 3.45679301326474e+01, 3.45678618369654e+01, 3.45676797837422e+01, 3.45672813545173e+01, 3.45665182758117e+01, 3.45651911532907e+01, 3.45630458423830e+01, 3.45597718332215e+01, 3.45550027750212e+01, 3.45483192082950e+01, 3.45392535137125e+01, 3.45272970251599e+01, 3.45119091929289e+01, 3.44925286221297e+01, 3.44685857523000e+01, 3.44395168872476e+01, 3.44047792294988e+01, 3.43638665212147e+01, 3.43163248433775e+01, 3.42617680788467e+01, 3.41998925058943e+01, 3.41304899628184e+01, 3.40534590191595e+01, 3.39688136140029e+01, 3.38766886849880e+01, 3.37773424176625e+01, 3.36711548924955e+01, 3.35586230876076e+01, 3.34403523933161e+01, 3.33170449889947e+01, 3.31894856011432e+01, 3.30585252844308e+01, 3.29250639321233e+01, 3.27900322252405e+01, 3.26543736770166e+01, 3.25190273344267e+01, 3.23849115796732e+01, 3.22529093500281e+01, 3.21238549797562e+01, 3.19985227733183e+01, 3.18776173491395e+01, 3.17617657472183e+01, 3.16515112674585e+01, 3.15473089927862e+01, 3.14495229458216e+01, 3.13584248251384e+01, 3.12741942636778e+01, 3.11969205459707e+01, 3.11266057120489e+01, 3.10631689645885e+01, 3.10064522825542e+01, 3.09562271299975e+01, 3.09122021331813e+01, 3.08740315831916e+01, 3.08413246049700e+01, 3.08136548176575e+01, 3.07905702958840e+01, 3.07716036280026e+01, 3.07562818563228e+01, 3.07441360773715e+01, 3.07347104783618e+01, 3.07275705905646e+01, 3.07223105520731e+01, 3.07185591920855e+01, 3.07159847763840e+01, 3.07142982886949e+01, 3.07132551640715e+01, 3.07126554368625e+01, 3.07123423153087e+01, 3.07121992452233e+01, 3.07121455743291e+01, 3.07121309745100e+01, 3.07121288196061e+01, 3.07121286293691e+01, 3.07121249017685e+01, 3.07120996467161e+01, 3.07120068063247e+01, 3.07117593267936e+01, 3.07112177181398e+01, 3.07101804507266e+01, 3.07083765649875e+01, 3.07054608207432e+01, 3.07010116548354e+01, 3.06945321499373e+01, 3.06854541424966e+01, 3.06731455129261e+01, 3.06569206058867e+01, 3.06360536231901e+01, 3.06097947182074e+01, 3.05773884021732e+01, 3.05380937548091e+01, 3.04912058215078e+01, 3.04360774855856e+01, 3.03721410360391e+01, 3.02989286174367e+01, 3.02160907555284e+01, 3.01234122029188e+01, 3.00208244422838e+01, 2.99084143140097e+01, 2.97864283908936e+01, 2.96552728928937e+01, 2.95155091089088e+01, 2.93678444624512e+01, 2.92131195204979e+01, 2.90522913995823e+01, 2.88864141699885e+01, 2.87166169926243e+01, 2.85440808308976e+01, 2.83700146418740e+01, 2.81956319465603e+01, 2.80221285964146e+01, 2.78506623972762e+01, 2.76823350471630e+01, 2.75181766281256e+01, 2.73591327026661e+01, 2.72060539288108e+01, 2.70596880334055e+01, 2.69206739622938e+01, 2.67895380403961e+01, 2.66666920042001e+01, 2.65524327980940e+01, 2.64469440455473e+01, 2.63502991138756e+01, 2.62624656888131e+01, 2.61833117655544e+01, 2.61126129494208e+01, 2.60500609439536e+01, 2.59952730880725e+01, 2.59478027871600e+01, 2.59071506654734e+01, 2.58727762491075e+01, 2.58441099702193e+01, 2.58205652652555e+01, 2.58015505238722e+01, 2.57864806328534e+01, 2.57747878525687e+01, 2.57659317642216e+01, 2.57594080359455e+01, 2.57547557757710e+01, 2.57515632700389e+01, 2.57494719466010e+01, 2.57481784519707e+01, 2.57474347885668e+01, 2.57470465198399e+01, 2.57468691145029e+01, 2.57468025632887e+01, 2.57467844597119e+01, 2.57467817876607e+01, 2.57467815733292e+01, 2.57467776178246e+01, 2.57467508186900e+01, 2.57466523021053e+01, 2.57463896920521e+01, 2.57458149712558e+01, 2.57447142925619e+01, 2.57428001416903e+01, 2.57397062001370e+01, 2.57349851967403e+01, 2.57281099673913e+01, 2.57184778634388e+01, 2.57054185593576e+01, 2.56882052085192e+01, 2.56660687826867e+01, 2.56382153079719e+01, 2.56038455811518e+01, 2.55621768213711e+01, 2.55124655912976e+01, 2.54540312182687e+01, 2.53862788700073e+01, 2.53087214005352e+01, 2.52209990872568e+01, 2.51228964335143e+01, 2.50143553114197e+01, 2.48954838619680e+01, 2.47665607439238e+01, 2.46280345180303e+01, 2.44805181565299e+01, 2.43247788688404e+01, 2.41617236234320e+01, 2.39923809161268e+01, 2.38178794794272e+01, 2.36394247390551e+01, 2.34582738951668e+01, 2.32757105298810e+01, 2.30930196156029e+01, 2.29114637206959e+01, 2.27322610868769e+01, 2.25565660986265e+01, 2.23854524952237e+01, 2.22198995081994e+01, 2.20607809566506e+01, 2.19088572112550e+01, 2.17647698507547e+01, 2.16290387824384e+01, 2.15020615766170e+01, 2.13841147673570e+01, 2.12753568898207e+01, 2.11758330507863e+01, 2.10854808568666e+01, 2.10041375498358e+01, 2.09315482172318e+01, 2.08673749573953e+01, 2.08112068808562e+01, 2.07625708248406e+01, 2.07209426456133e+01, 2.06857589358291e+01, 2.06564289928216e+01, 2.06323468409271e+01, 2.06129030888461e+01, 2.05974963841585e+01, 2.05855442139020e+01, 2.05764927948778e+01, 2.05698258019665e+01, 2.05650716985600e+01, 2.05618094608055e+01, 2.05596725264408e+01, 2.05583508483967e+01, 2.05575909910284e+01, 2.05571942701165e+01, 2.05570130034827e+01, 2.05569450038559e+01, 2.05569265062847e+01, 2.05569237760811e+01, 2.05569235655015e+01, 2.05569197841680e+01, 2.05568941650720e+01, 2.05567999864812e+01, 2.05565489402939e+01, 2.05559995285955e+01, 2.05549473269813e+01, 2.05531175004692e+01, 2.05501599094095e+01, 2.05456470867435e+01, 2.05390753033270e+01, 2.05298688633757e+01, 2.05173876859217e+01, 2.05009381294407e+01, 2.04797869054813e+01, 2.04531778047307e+01, 2.04203508290491e+01, 2.03805631914517e+01, 2.03331115209130e+01, 2.02773545001837e+01, 2.02127350835618e+01, 2.01388013986261e+01, 2.00552254406200e+01, 1.99618187265520e+01, 1.98585441896301e+01, 1.97455237591477e+01, 1.96230412761540e+01, 1.94915406255062e+01, 1.93516192007339e+01, 1.92040170387953e+01, 1.90496021482843e+01, 1.88893526928505e+01, 1.87243367745073e+01, 1.85556905901468e+01, 1.83845957171091e+01, 1.82122562328263e+01, 1.80398763032560e+01, 1.78686387965343e+01, 1.76996853988171e+01, 1.75340986302595e+01, 1.73728860784489e+01, 1.72169670813195e+01, 1.70671620004849e+01, 1.69241841314835e+01, 1.67886342056844e+01, 1.66609973575406e+01, 1.65416423679671e+01, 1.64308229545500e+01, 1.63286808627949e+01, 1.62352505166020e+01, 1.61504650048313e+01, 1.60741632073116e+01, 1.60060978914467e+01, 1.59459446344918e+01, 1.58933114432112e+01, 1.58477489503203e+01, 1.58087610656975e+01, 1.57758159508002e+01, 1.57483571688376e+01, 1.57258148434019e+01, 1.57076166371427e+01, 1.56931983425528e+01, 1.56820138618490e+01, 1.56735443448859e+01, 1.56673062552130e+01, 1.56628581463144e+01, 1.56598059535175e+01, 1.56578066418580e+01, 1.56565700952849e+01, 1.56558591860365e+01, 1.56554880221891e+01, 1.56553184331853e+01, 1.56552548143048e+01, 1.56552375084088e+01, 1.56552349540941e+01, 1.56552347606891e+01, 1.56552313345387e+01, 1.56552081218619e+01, 1.56551227895481e+01, 1.56548953244813e+01, 1.56543975204740e+01, 1.56534441575394e+01, 1.56517862241497e+01, 1.56491064898711e+01, 1.56450176856749e+01, 1.56390634913296e+01, 1.56307224621909e+01, 1.56194149499865e+01, 1.56045129827947e+01, 1.55853529682496e+01, 1.55612509723788e+01, 1.55315202075384e+01, 1.54954902418021e+01, 1.54525273260763e+01, 1.54020551329537e+01, 1.53435751227696e+01, 1.52766857074092e+01, 1.52010993800174e+01, 1.51166570253951e+01, 1.50233387244267e+01, 1.49212705145062e+01, 1.48107267592905e+01, 1.46921280021523e+01, 1.45660344104651e+01, 1.44331351411353e+01, 1.42942341501948e+01, 1.41502331128653e+01, 1.40021122045104e+01, 1.38509095160095e+01, 1.36976998475748e+01, 1.35435735580298e+01, 1.33896160595695e+01, 1.32368884560878e+01, 1.30864097355313e+01, 1.29391408461413e+01, 1.27959709107444e+01, 1.26577057584837e+01, 1.25250588767244e+01, 1.23986448074021e+01, 1.22789749349746e+01, 1.21664555426197e+01, 1.20613879550066e+01, 1.19639705442744e+01, 1.18743023529990e+01, 1.17923880835402e+01, 1.17181442145672e+01, 1.16514060284709e+01, 1.15919353626659e+01, 1.15394289283710e+01, 1.14935270678480e+01, 1.14538228418961e+01, 1.14198713515263e+01, 1.13911992004713e+01, 1.13673139990429e+01, 1.13477137964940e+01, 1.13318963109664e+01, 1.13193678063703e+01, 1.13096514474291e+01, 1.13022949508518e+01, 1.12968773449323e+01, 1.12930146539629e+01, 1.12903643388890e+01, 1.12886283518619e+01, 1.12875546989440e+01, 1.12869374504495e+01, 1.12866151897514e+01, 1.12864679458671e+01, 1.12864127095526e+01, 1.12863976839310e+01, 1.12863954661806e+01, 1.12863952996934e+01, 1.12863923693294e+01, 1.12863725156836e+01, 1.12862995315963e+01, 1.12861049829283e+01, 1.12856792184996e+01, 1.12848638308817e+01, 1.12834458718663e+01, 1.12811540917080e+01, 1.12776574308702e+01, 1.12725659457897e+01, 1.12654342944864e+01, 1.12557678419750e+01, 1.12430313681497e+01, 1.12266602717435e+01, 1.12060740641625e+01, 1.11806918392261e+01, 1.11499492937487e+01, 1.11133167659626e+01, 1.10703176619235e+01, 1.10205465630509e+01, 1.09636862595976e+01, 1.08995229429238e+01, 1.08279588197769e+01, 1.07490214873511e+01, 1.06628695282484e+01, 1.05697939453806e+01, 1.04702152504153e+01, 1.03646762343151e+01, 1.02538306708797e+01, 1.01384284182295e+01, 1.00192975725332e+01, 9.89732447754537e+00, 9.77343248979056e+00, 9.64856043390239e+00, 9.52364165278959e+00, 9.39958446654523e+00, 9.27725471234107e+00, 9.15746086018362e+00, 9.04094200478119e+00, 8.92835884092565e+00, 8.82028755581858e+00, 8.71721642971785e+00, 8.61954483393540e+00, 8.52758425486256e+00, 8.44156095174614e+00, 8.36161986817245e+00, 8.28782945430894e+00, 8.22018710991389e+00, 8.15862501849083e+00, 8.10301620357903e+00, 8.05318069357282e+00, 8.00889172794778e+00, 7.96988197322206e+00, 7.93584974059043e+00, 7.90646520913450e+00, 7.88137665977183e+00, 7.86021671712390e+00, 7.84260858111250e+00, 7.82817220947890e+00, 7.81653038891504e+00, 7.80731460858846e+00, 7.80017062804228e+00, 7.79476361415472e+00, 7.79078271115977e+00, 7.78794490531016e+00, 7.78599805261800e+00, 7.78472295449362e+00, 7.78393439145378e+00, 7.78348105803721e+00, 7.78324438062774e+00, 7.78313624156116e+00, 7.78309567499279e+00, 7.78308463991475e+00, 7.78308301116079e+00, 7.78308288926155e+00, 7.78308074864957e+00, 7.78306624569574e+00, 7.78301293141168e+00, 7.78287081593746e+00, 7.78255980392349e+00, 7.78196419490125e+00, 7.78092847648154e+00, 7.77925461619986e+00, 7.77670103185173e+00, 7.77298338860009e+00, 7.76777733184956e+00, 7.76072321560691e+00, 7.75143282490475e+00, 7.73949801692102e+00, 7.72450111924873e+00, 7.70602682789845e+00, 7.68367524684179e+00, 7.65707561227197e+00, 7.62590015720952e+00, 7.58987750570876e+00, 7.54880495084666e+00, 7.50255897566782e+00, 7.45110342722781e+00, 7.39449485251100e+00, 7.33288464777766e+00, 7.26651785074695e+00, 7.19572860380234e+00, 7.12093251825815e+00, 7.04261635521918e+00, 6.96132558933345e+00, 6.87765052315696e+00, 6.79221166327936e+00, 6.70564505344553e+00, 6.61858819066295e+00, 6.53166703988768e+00, 6.44548452745685e+00, 6.36061075027214e+00, 6.27757500256533e+00, 6.19685960692697e+00, 6.11889544831125e+00, 6.04405905112202e+00, 5.97267100822144e+00, 5.90499556194516e+00, 5.84124114482070e+00, 5.78156170557969e+00, 5.72605866910481e+00, 5.67478340342957e+00, 5.62774009047429e+00, 5.58488891861974e+00, 5.54614953395590e+00, 5.51140470288412e+00, 5.48050415153945e+00, 5.45326855701384e+00, 5.42949367131244e+00, 5.40895456109420e+00, 5.39140994440097e+00, 5.37660659987830e+00, 5.36428381486719e+00, 5.35417782698615e+00, 5.34602620053363e+00, 5.33957206559252e+00, 5.33456813562317e+00, 5.33078041011409e+00, 5.32799146390220e+00, 5.32600322517054e+00, 5.32463915055516e+00, 5.32374571841464e+00, 5.32319317975622e+00, 5.32287552967341e+00, 5.32270968908928e+00, 5.32263391547311e+00, 5.32260549022488e+00, 5.32259775787299e+00, 5.32259661659422e+00, 5.32259653114663e+00, 5.32259503021691e+00, 5.32258486120068e+00, 5.32254747887407e+00, 5.32244783172123e+00, 5.32222975841943e+00, 5.32181212983215e+00, 5.32108589485352e+00, 5.31991217220003e+00, 5.31812151116447e+00, 5.31551442145306e+00, 5.31186324421003e+00, 5.30691540182207e+00, 5.30039802196964e+00, 5.29202388116374e+00, 5.28149855517986e+00, 5.26852860018250e+00, 5.25283052220215e+00, 5.23414022870405e+00, 5.21222260014401e+00, 5.18688077808010e+00, 5.15796474578902e+00, 5.12537878237847e+00, 5.08908740487352e+00, 5.04911947455281e+00, 5.00557023056912e+00, 4.95860111927891e+00, 4.90843740321541e+00, 4.85536364979312e+00, 4.79971730755164e+00, 4.74188066952529e+00, 4.68227159393076e+00, 4.62133339892024e+00, 4.55952436969524e+00, 4.49730731309642e+00, 4.43513956786676e+00, 4.37346382973856e+00, 4.31270008195527e+00, 4.25323883811682e+00, 4.19543581172226e+00, 4.13960803370600e+00, 4.08603135460832e+00, 4.03493919989416e+00, 3.98652240090604e+00, 3.94092990197403e+00, 3.89827014456004e+00, 3.85861294732178e+00, 3.82199173043070e+00, 3.78840596703772e+00, 3.75782377910982e+00, 3.73018462518136e+00, 3.70540205174107e+00, 3.68336649724601e+00, 3.66394814830375e+00, 3.64699985212357e+00, 3.63236008885964e+00, 3.61985600294879e+00, 3.60930648493662e+00, 3.60052528550051e+00, 3.59332413231223e+00, 3.58751580895076e+00, 3.58291714422346e+00, 3.57935185093987e+00, 3.57665314634038e+00, 3.57466608284351e+00, 3.57324951819363e+00, 3.57227765885646e+00, 3.57164111968834e+00, 3.57124745620388e+00, 3.57102114253833e+00, 3.57090298749593e+00, 3.57084900172925e+00, 3.57082874984450e+00, 3.57082324084460e+00, 3.57082242772790e+00, 3.57082236668831e+00, 3.57082129234050e+00, 3.57081401347875e+00, 3.57078725565376e+00, 3.57071592942997e+00, 3.57055983532437e+00, 3.57026090253262e+00, 3.56974107496563e+00, 3.56890094658743e+00, 3.56761923447647e+00, 3.56575316168584e+00, 3.56313980186154e+00, 3.55959841264414e+00, 3.55493375425643e+00, 3.54894035296267e+00, 3.54140762660968e+00, 3.53212574255529e+00, 3.52089202942088e+00, 3.50751771688357e+00, 3.49183473670251e+00, 3.47370228843751e+00, 3.45301285984511e+00, 3.42969739881993e+00, 3.40372936335699e+00, 3.37512742827089e+00, 3.34395669935844e+00, 3.31032837149088e+00, 3.27439785861262e+00, 3.23636151143705e+00, 3.19645211368761e+00, 3.15493340283453e+00, 3.11209389237682e+00, 3.06824027954378e+00, 3.02369070807239e+00, 2.97876812615150e+00, 2.93379394132349e+00, 2.88908213311309e+00, 2.84493394474919e+00, 2.80163323983817e+00, 2.75944257873969e+00, 2.71860004211975e+00, 2.67931680485691e+00, 2.64177544160805e+00, 2.60612892592734e+00, 2.57250026841343e+00, 2.54098272672407e+00, 2.51164051215729e+00, 2.48450991421877e+00, 2.45960076605828e+00, 2.43689817924108e+00, 2.41636448500408e+00, 2.39794132966450e+00, 2.38155188286741e+00, 2.36710312762924e+00, 2.35448820962388e+00, 2.34358882911741e+00, 2.33427766197623e+00, 2.32642079617468e+00, 2.31988016745874e+00, 2.31451597281000e+00, 2.31018903384690e+00, 2.30676307520489e+00, 2.30410687622719e+00, 2.30209624893282e+00, 2.30061579206122e+00, 2.29956037068695e+00, 2.29883627386613e+00, 2.29836200912534e+00, 2.29806870214018e+00, 2.29790008217752e+00, 2.29781204806545e+00, 2.29777182471346e+00, 2.29775673557150e+00, 2.29775263096148e+00, 2.29775202512974e+00, 2.29775197945707e+00, 2.29775117299965e+00, 2.29774570913394e+00, 2.29772562341489e+00, 2.29767208248759e+00, 2.29755491062043e+00, 2.29733051685671e+00, 2.29694030792617e+00, 2.29630966337262e+00, 2.29534753916465e+00, 2.29394675385083e+00, 2.29198499641528e+00, 2.28932657667557e+00, 2.28582491676780e+00, 2.28132575573689e+00, 2.27567100873967e+00, 2.26870318876630e+00, 2.26027026369587e+00, 2.25023078722641e+00, 2.23845911162780e+00, 2.22485046653922e+00, 2.20932567432050e+00, 2.19183527146347e+00, 2.17236281910871e+00, 2.15092721440955e+00, 2.12758385753211e+00, 2.10242458420218e+00, 2.07557633732542e+00, 2.04719861874540e+00, 2.01747982854405e+00, 1.98663265926227e+00, 1.95488876131318e+00, 1.92249292987937e+00, 1.88969708023113e+00, 1.85675427673917e+00, 1.82391306162373e+00, 1.79141229503011e+00, 1.75947667204252e+00, 1.72831302934664e+00, 1.69810749937920e+00, 1.66902351767813e+00, 1.64120064372589e+00, 1.61475411967916e+00, 1.58977506650184e+00, 1.56633120337443e+00, 1.54446797294582e+00, 1.52420996033307e+00, 1.50556250560889e+00, 1.48851342555825e+00, 1.47303477857210e+00, 1.45908462480887e+00, 1.44660875072390e+00, 1.43554234167911e+00, 1.42581159793686e+00, 1.41733529757813e+00, 1.41002631471208e+00, 1.40379310292699e+00, 1.39854115260601e+00, 1.39417442697197e+00, 1.39059677611547e+00, 1.38771332147248e+00, 1.38543179597762e+00, 1.38366381818096e+00, 1.38232607270711e+00, 1.38134136521170e+00, 1.38063951797541e+00, 1.38015807281233e+00, 1.37984277117826e+00, 1.37964778712746e+00, 1.37953569673279e+00, 1.37947717720674e+00, 1.37945043955420e+00, 1.37944040940273e+00, 1.37943768096448e+00, 1.37943727825292e+00, 1.37943724772796e+00, 1.37943670654681e+00, 1.37943303996829e+00, 1.37941956128829e+00, 1.37938363246402e+00, 1.37930500507933e+00, 1.37915443177427e+00, 1.37889260713808e+00, 1.37846949229725e+00, 1.37782407171677e+00, 1.37688458350125e+00, 1.37556925499367e+00, 1.37378756294316e+00, 1.37144202128745e+00, 1.36843047929926e+00, 1.36464888858833e+00, 1.35999446995562e+00, 1.35436918180659e+00, 1.34768336289777e+00, 1.33985939636594e+00, 1.33083522238361e+00, 1.32056751650781e+00, 1.30903435251724e+00, 1.29623718403620e+00, 1.28220200897393e+00, 1.26697962361774e+00, 1.25064492628422e+00, 1.23329528943520e+00, 1.21504807873213e+00, 1.19603745185356e+00, 1.17641061362828e+00, 1.15632373285927e+00, 1.13593773759568e+00, 1.11541419910421e+00, 1.09491149205522e+00, 1.07458138291699e+00, 1.05456615488085e+00, 1.03499633092844e+00, 1.01598901173302e+00, 9.97646805919914e-01, 9.80057299491889e-01, 9.63292990250366e-01, 9.47411601813187e-01, 9.32456689312204e-01, 9.18458453346247e-01, 9.05434688255013e-01, 8.93391803266354e-01, 8.82325868778210e-01, 8.72223653559558e-01, 8.63063630982282e-01, 8.54816942897727e-01, 8.47448318133424e-01, 8.40916948726374e-01, 8.35177331010676e-01, 8.30180080712374e-01, 8.25872731500127e-01, 8.22200525245215e-01, 8.19107199822085e-01, 8.16535776906726e-01, 8.14429348200299e-01, 8.12731854135372e-01, 8.11388844754049e-01, 8.10348208434270e-01, 8.09560850835858e-01, 8.08981304165989e-01, 8.08568245890832e-01, 8.08284906523227e-01, 8.08099348158819e-01, 8.07984598944863e-01, 8.07918633450939e-01, 8.07884194652196e-01, 8.07868459539513e-01, 8.07862556797396e-01, 8.07860951112374e-01, 8.07860714116771e-01, 8.07860696038844e-01, 8.07860374029218e-01, 8.07858192368840e-01, 8.07850172387688e-01, 8.07828794317196e-01, 8.07782010225388e-01, 8.07692418045363e-01, 8.07536631988427e-01, 8.07284881897260e-01, 8.06900870790514e-01, 8.06341915602353e-01, 8.05559390440237e-01, 8.04499484054347e-01, 8.03104273184196e-01, 8.01313100773065e-01, 7.99064232789974e-01, 7.96296750077190e-01, 7.92952613256617e-01, 7.88978820788344e-01, 7.84329564703206e-01, 7.78968277463818e-01, 7.72869458919750e-01, 7.66020176061183e-01, 7.58421141127990e-01, 7.50087295423107e-01, 7.41047855541531e-01, 7.31345813148444e-01, 7.21036915540253e-01, 7.10188188215675e-01, 6.98876088944971e-01, 6.87184402463707e-01, 6.75201994220295e-01, 6.63020540275561e-01, 6.50732339568044e-01, 6.38428296513787e-01, 6.26196139156043e-01, 6.14118913825138e-01, 6.02273774199858e-01, 5.90731062819110e-01, 5.79553667697752e-01, 5.68796626173828e-01, 5.58506942206868e-01, 5.48723581357649e-01, 5.39477608682546e-01, 5.30792437808579e-01, 5.22684163663243e-01, 5.15161956035826e-01, 5.08228495855271e-01, 5.01880440472256e-01, 4.96108908145423e-01, 4.90899975262828e-01, 4.86235182535466e-01, 4.82092048463862e-01, 4.78444589790307e-01, 4.75263849409735e-01, 4.72518432325735e-01, 4.70175049733697e-01, 4.68199070240892e-01, 4.66555075679700e-01, 4.65207417058096e-01, 4.64120764079738e-01, 4.63260639543137e-01, 4.62593928002370e-01, 4.62089346549121e-01, 4.61717864649884e-01, 4.61453059797286e-01, 4.61271396409792e-01, 4.61152416970187e-01, 4.61078836782356e-01, 4.61036536827756e-01, 4.61014452830007e-01, 4.61004362558228e-01, 4.61000577363357e-01, 4.60999547699927e-01, 4.60999395723825e-01, 4.60999384049330e-01, 4.60999175027389e-01, 4.60997758873887e-01, 4.60992552957784e-01, 4.60978676004196e-01, 4.60948307175627e-01, 4.60890149323831e-01, 4.60789018864496e-01, 4.60625582691769e-01, 4.60376260131369e-01, 4.60013304942564e-01, 4.59505078573777e-01, 4.58816520995299e-01, 4.57909819254937e-01, 4.56745266318279e-01, 4.55282293827331e-01, 4.53480652432702e-01, 4.51301702885871e-01, 4.48709770939715e-01, 4.45673510326522e-01, 4.42167211794484e-01, 4.38171993463321e-01, 4.33676809447803e-01, 4.28679220222534e-01, 4.23185879440481e-01, 4.17212707150618e-01, 4.10784737318615e-01, 4.03935646581679e-01, 3.96706989458281e-01, 3.89147181098174e-01, 3.81310280774436e-01, 3.73254636891552e-01, 3.65041457103040e-01, 3.56733365524185e-01, 3.48393003718947e-01, 3.40081724088598e-01, 3.31858414510556e-01, 3.23778482508928e-01, 3.15893016679304e-01, 3.08248133153022e-01, 3.00884506018832e-01, 2.93837073123917e-01, 2.87134902735254e-01, 2.80801202242976e-01, 2.74853447426430e-01, 2.69303609698980e-01, 2.64158459043898e-01, 2.59419921828851e-01, 2.55085475075724e-01, 2.51148561767472e-01, 2.47599015094842e-01, 2.44423482895229e-01, 2.41605846658146e-01, 2.39127632159715e-01, 2.36968410884160e-01, 2.35106192798391e-01, 2.33517811727822e-01, 2.32179304556290e-01, 2.31066284813494e-01, 2.30154310038650e-01, 2.29419240777219e-01, 2.28837587363387e-01, 2.28386838964319e-01, 2.28045767909713e-01, 2.27794701280551e-01, 2.27615751228413e-01, 2.27492995634357e-01, 2.27412601530965e-01, 2.27362885172994e-01, 2.27334304663650e-01, 2.27319383485543e-01, 2.27312565972724e-01, 2.27310008503935e-01, 2.27309312811897e-01, 2.27309210129263e-01, 2.27309202181210e-01, 2.27309059094146e-01, 2.27308089659203e-01, 2.27304525927700e-01, 2.27295026434938e-01, 2.27274237570242e-01, 2.27234426487705e-01, 2.27165201191696e-01, 2.27053332191794e-01, 2.26882688634461e-01, 2.26634299930573e-01, 2.26286551410880e-01, 2.25815519247647e-01, 2.25195445604392e-01, 2.24399349619752e-01, 2.23399763426137e-01, 2.22169575142864e-01, 2.20682953047699e-01, 2.18916317486340e-01, 2.16849320243512e-01, 2.14465785875575e-01, 2.11754566688448e-01, 2.08710263308028e-01, 2.05333766568188e-01, 2.01632583833323e-01, 1.97620923586037e-01, 1.93319525456005e-01, 1.88755237815059e-01, 1.83960360354895e-01, 1.78971783354314e-01, 1.73829967363865e-01, 1.68577815748193e-01, 1.63259497233398e-01, 1.57919276026945e-01, 1.52600403363124e-01, 1.47344117019948e-01, 1.42188785314467e-01, 1.37169220366985e-01, 1.32316173157804e-01, 1.27656011149206e-01, 1.23210568910032e-01, 1.18997153913227e-01, 1.15028683849481e-01, 1.11313928499911e-01, 1.07857828279258e-01, 1.04661862656243e-01, 1.01724444320297e-01, 9.90413186929544e-02, 9.66059526903464e-02, 9.44099011017485e-02, 9.24431432110282e-02, 9.06943860986544e-02, 8.91513342534367e-02, 8.78009276066507e-02, 8.66295518452915e-02, 8.56232258859677e-02, 8.47677717455895e-02, 8.40489718070074e-02, 8.34527177409541e-02, 8.29651542195080e-02, 8.25728191636443e-02, 8.22627807395335e-02, 8.20227697924238e-02, 8.18413050179458e-02, 8.17078070442463e-02, 8.16126968375315e-02, 8.15474735216668e-02, 8.15047668546924e-02, 8.14783602200748e-02, 8.14631810173196e-02, 8.14552566827942e-02, 8.14516361207254e-02, 8.14502779442066e-02, 8.14499084892830e-02, 8.14498539585813e-02, 8.14498497034677e-02, 8.14497726574572e-02, 8.14492506605546e-02, 8.14473317614954e-02, 8.14422168098972e-02, 8.14310235169913e-02, 8.14095895263733e-02, 8.13723234735448e-02, 8.13121126645342e-02, 8.12202955410086e-02, 8.10867063709892e-02, 8.08997984200772e-02, 8.06468500147588e-02, 8.03142552287710e-02, 7.98878972954142e-02, 7.93535982718794e-02, 7.86976331032818e-02, 7.79072903721030e-02, 7.69714561630153e-02, 7.58811922577749e-02, 7.46302760189792e-02, 7.32156675379727e-02, 7.16378705071747e-02, 6.99011571941810e-02, 6.80136348779712e-02, 6.59871407954853e-02, 6.38369642867787e-02, 6.15814073329801e-02, 5.92412067652930e-02, 5.68388517681697e-02, 5.43978377534486e-02, 5.19419014309435e-02, 4.94942815946757e-02, 4.70770459436309e-02, 4.47105167969895e-02, 4.24128188430188e-02, 4.01995612717059e-02, 3.80836559921462e-02, 3.60752641783080e-02, 3.41818558946017e-02, 3.24083624537010e-02, 3.07573985387748e-02, 2.92295307691717e-02, 2.78235708930765e-02, 2.65368746259966e-02, 2.53656307828578e-02, 2.43051292796110e-02, 2.33500004105807e-02, 2.24944212557077e-02, 2.17322879678505e-02, 2.10573549634319e-02, 2.04633436965338e-02, 1.99440247904406e-02, 1.94932779118986e-02, 1.91051339894344e-02, 1.87738042817358e-02, 1.84937004671113e-02, 1.82594494132218e-02, 1.80659056452726e-02, 1.79081638052523e-02, 1.77815726204287e-02, 1.76817511149352e-02, 1.76046070437508e-02, 1.75463568438303e-02, 1.75035458250861e-02, 1.74730669016014e-02, 1.74521759206385e-02, 1.74385016000826e-02, 1.74300482341465e-02, 1.74251896545883e-02, 1.74226534055464e-02, 1.74214946562827e-02, 1.74210599830005e-02, 1.74209417426141e-02, 1.74209242906187e-02, 1.74209229172249e-02, 1.74208979011681e-02, 1.74207284144856e-02, 1.74201053738514e-02, 1.74184446560162e-02, 1.74148106202452e-02, 1.74078525572199e-02, 1.73957572773418e-02, 1.73762211069075e-02, 1.73464445854066e-02, 1.73031530637004e-02, 1.72426460785076e-02, 1.71608777028709e-02, 1.70535689383158e-02, 1.69163515497459e-02, 1.67449405476390e-02, 1.65353298771294e-02, 1.62840029762350e-02, 1.59881470168202e-02, 1.56458572318243e-02, 1.52563161958099e-02, 1.48199326714789e-02, 1.43384259686741e-02, 1.38148448030577e-02, 1.32535142659265e-02, 1.26599103285372e-02, 1.20404676761439e-02, 1.14023328107393e-02, 1.07530794625745e-02, 1.01004067067905e-02, 9.45184133649235e-03, 8.81446487002080e-03, 8.19468229744302e-03, 7.59804482500925e-03, 7.02913318188081e-03, 6.49150228686324e-03, 5.98768293868090e-03, 5.51923221113526e-03, 5.08682169792962e-03, 4.69035170322485e-03, 4.32907975267938e-03, 4.00175309211193e-03, 3.70673677179667e-03, 3.44213111865344e-03, 3.20587458124777e-03, 2.99582988917990e-03, 2.80985307963277e-03, 2.64584615788976e-03, 2.50179497528080e-03, 2.37579436887358e-03, 2.26606278150143e-03, 2.17094855164390e-03, 2.08892990506371e-03, 2.01861045472640e-03, 1.95871175976947e-03, 1.90806423435674e-03, 1.86559744631865e-03, 1.83033061308890e-03, 1.80136389138141e-03, 1.77787086840908e-03, 1.75909249135486e-03, 1.74433251852690e-03, 1.73295443626387e-03, 1.72437966368477e-03, 1.71808676532947e-03, 1.71361131469894e-03, 1.71054600472331e-03, 1.70854058777081e-03, 1.70730124967378e-03, 1.70658907601506e-03, 1.70621735077248e-03, 1.70604752779013e-03, 1.70598382479170e-03, 1.70596649635968e-03, 1.70596393872557e-03, 1.70596373568088e-03, 1.70596001442647e-03, 1.70593480256542e-03, 1.70584212374791e-03, 1.70559509643198e-03, 1.70505458659420e-03, 1.70401984172486e-03, 1.70222165076694e-03, 1.69931862029671e-03, 1.69489719888105e-03, 1.68847609674572e-03, 1.67951570649194e-03, 1.66743300201944e-03, 1.65162214654657e-03, 1.63148066046404e-03, 1.60644048405345e-03, 1.57600264832780e-03, 1.53977359750712e-03, 1.49750057985898e-03, 1.44910305382950e-03, 1.39469686398996e-03, 1.33460812887165e-03, 1.26937440765573e-03, 1.19973176290955e-03, 1.12658771577401e-03, 1.05098162217132e-03, 9.74035452976807e-04, 8.96899095671699e-04, 8.20694913180375e-04, 7.46466294060610e-04, 6.75134321973303e-04, 6.07465602385560e-04, 5.44052901574332e-04, 4.85308789272665e-04, 4.31471130688976e-04, 3.82618222793951e-04, 3.38690744897008e-04, 2.99517552477369e-04, 2.64842632802242e-04, 2.34351097226693e-04, 2.07692706517548e-04, 1.84501970025508e-04, 1.64414301686122e-04, 1.47078109115739e-04, 1.32163066538737e-04, 1.19365132718485e-04, 1.08409034730528e-04, 9.90489130998640e-05, 9.10676738001825e-05, 8.42754301615400e-05, 7.85073192730579e-05, 7.36209461322801e-05, 6.94936923366640e-05, 6.60200860096438e-05, 6.31093619531953e-05, 6.06832718546002e-05, 5.86741567971671e-05, 5.70232740173537e-05, 5.56793662340738e-05, 5.45974631620538e-05, 5.37379033644925e-05, 5.30655598396379e-05, 5.25492477478906e-05, 5.21612880675093e-05, 5.18771984019395e-05, 5.16754810320208e-05, 5.15374774892052e-05, 5.14472602191063e-05, 5.13915335886850e-05, 5.13595202278339e-05, 5.13428133004368e-05, 5.13351813538065e-05, 5.13323186055459e-05, 5.13315398933401e-05, 5.13314249559471e-05, 5.13314157473133e-05, 5.13312459560211e-05, 5.13300955233831e-05, 5.13258666572369e-05, 5.13145955736150e-05, 5.12899367919317e-05, 5.12427418613015e-05, 5.11607620268251e-05, 5.10285089916351e-05, 5.08273123287495e-05, 5.05356147257885e-05, 5.01295440997751e-05, 4.95837930785115e-05, 4.88728182332920e-05, 4.79723419246625e-05, 4.68611003654473e-05, 4.55227328544083e-05, 4.39476580993414e-05, 4.21347423937846e-05, 4.00925451131684e-05, 3.78399414344699e-05, 3.54059761491499e-05, 3.28288893953754e-05, 3.01543553860029e-05, 2.74330623852848e-05, 2.47178189098918e-05, 2.20604044329775e-05, 1.95084225538267e-05, 1.71024830848410e-05, 1.48741095031824e-05, 1.28447455480796e-05, 1.10260164944975e-05, 9.42099711664923e-06, 8.02585879000617e-06, 6.83122588185370e-06, 5.82300302198746e-06, 4.98308196483777e-06, 4.29065874385515e-06, 3.72442999463374e-06, 3.26494436109215e-06, 2.89589956992131e-06, 2.60372473803727e-06, 2.37592349350280e-06, 2.19981313875905e-06, 2.06299692706580e-06, 1.95480038911455e-06, 1.86730969886639e-06, 1.79579728865065e-06, 1.73809139880632e-06, 1.69271527283854e-06, 1.65761547066268e-06, 1.63067502967351e-06, 1.61036064127984e-06, 1.59521682974173e-06, 1.58339010170039e-06, 1.57319917803989e-06, 1.56389972261195e-06, 1.55563247972168e-06, 1.54877082536586e-06, 1.54342366442961e-06, 1.53939085123511e-06, 1.53635720323564e-06, 1.53405336789092e-06, 1.53230049448894e-06, 1.53098592647893e-06, 1.53002887515186e-06, 1.52936026154293e-06, 1.52891639970676e-06, 1.52863938751580e-06, 1.52847923828858e-06, 1.52839536582405e-06, 1.52835698230547e-06, 1.52834257960421e-06, 1.52833865740831e-06, 1.52833807476327e-06, 1.52833803213070e-06, 1.52833717237399e-06, 1.52833137434527e-06, 1.52831004385234e-06, 1.52825324306605e-06, 1.52812916098810e-06, 1.52789239393769e-06, 1.52748322790330e-06, 1.52682845566687e-06, 1.52584348711571e-06, 1.52443506351574e-06, 1.52250183305114e-06, 1.51992882280183e-06, 1.51657199864985e-06, 1.51223477473650e-06, 1.50665067621958e-06, 1.49950091810069e-06, 1.49049636632981e-06, 1.47951804763125e-06, 1.46673995260471e-06, 1.45262282558178e-06, 1.43777563721414e-06, 1.42287660054308e-06, 1.40874340814889e-06, 1.39609535665386e-06, 1.38453385289949e-06, 1.37167017300044e-06, 1.35410529367164e-06, 1.32969871913247e-06, 1.29859507325136e-06, 1.26265072708520e-06, 1.22460584606188e-06, 1.18597012743749e-06, 1.14385732530309e-06, 1.09149207361270e-06, 1.02187549089194e-06, 9.29732657084514e-07, 8.14576026186842e-07, 6.85085282725595e-07, 5.53075011566762e-07, 4.23250803294195e-07, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, 4.49067286467653e-07, 6.00299038922003e-07, 6.56394448128594e-07, 6.11499451963482e-07, 4.84424589330956e-07, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14],
[ 3.70000000000000e+01, 3.69999999699348e+01, 3.69999990402601e+01, 3.69999927415675e+01, 3.69999695867583e+01, 3.69999078633954e+01, 3.69997727775267e+01, 3.69995140497060e+01, 3.69990640503123e+01, 3.69983365449227e+01, 3.69972261019796e+01, 3.69956081949600e+01, 3.69933400104612e+01, 3.69902619528487e+01, 3.69861998161841e+01, 3.69809675757799e+01, 3.69743707355259e+01, 3.69662101535120e+01, 3.69562862575787e+01, 3.69444035541853e+01, 3.69303753280159e+01, 3.69140284255475e+01, 3.68952080127638e+01, 3.68737821948175e+01, 3.68496463834385e+01, 3.68227272963560e+01, 3.67929864724760e+01, 3.67604231879873e+01, 3.67250766632084e+01, 3.66870274591622e+01, 3.66463979777186e+01, 3.66033520003307e+01, 3.65580932278671e+01, 3.65108628169229e+01, 3.64619359445874e+01, 3.64116174716068e+01, 3.63602368104835e+01, 3.63081421375872e+01, 3.62556941144014e+01, 3.62032593008532e+01, 3.61512034523234e+01, 3.60998848913374e+01, 3.60496481358366e+01, 3.60008179496646e+01, 3.59536939592270e+01, 3.59085459550793e+01, 3.58656099702688e+01, 3.58250852001576e+01, 3.57871318023869e+01, 3.57518695914207e+01, 3.57193776201720e+01, 3.56896946217174e+01, 3.56628202669604e+01, 3.56387171791110e+01, 3.56173136327563e+01, 3.55985068538878e+01, 3.55821668273707e+01, 3.55681405099534e+01, 3.55562563400781e+01, 3.55463289306464e+01, 3.55381638277338e+01, 3.55315622173385e+01, 3.55263254638421e+01, 3.55222593682066e+01, 3.55191780411611e+01, 3.55169072967922e+01, 3.55152874849389e+01, 3.55141756963590e+01, 3.55134472924106e+01, 3.55129967304677e+01, 3.55127376768700e+01, 3.55126024202287e+01, 3.55125406186813e+01, 3.55125174345170e+01, 3.55125111278366e+01, 3.55125101969829e+01, 3.55125100748745e+01, 3.55125072298955e+01, 3.55124879547267e+01, 3.55124170969017e+01, 3.55122282139189e+01, 3.55118148377951e+01, 3.55110231334649e+01, 3.55096462291013e+01, 3.55074204531037e+01, 3.55040236631274e+01, 3.54990757980180e+01, 3.54921417248036e+01, 3.54827363910811e+01, 3.54703322294237e+01, 3.54543686961044e+01, 3.54342637626923e+01, 3.54094271168476e+01, 3.53792747684563e+01, 3.53432446992166e+01, 3.53008131379061e+01, 3.52515109902690e+01, 3.51949399034499e+01, 3.51307874039301e+01, 3.50588405212978e+01, 3.49789973064551e+01, 3.48912756816488e+01, 3.47958191294020e+01, 3.46928988425396e+01, 3.45829121160888e+01, 3.44663769540451e+01, 3.43439230724450e+01, 3.42162796823341e+01, 3.40842606084287e+01, 3.39487474214452e+01, 3.38106713218227e+01, 3.36709945074717e+01, 3.35306916959198e+01, 3.33907323672246e+01, 3.32520641678197e+01, 3.31155977864081e+01, 3.29821934968260e+01, 3.28526494693518e+01, 3.27276918847984e+01, 3.26079668431605e+01, 3.24940340353043e+01, 3.23863621354997e+01, 3.22853258681356e+01, 3.21912046988367e+01, 3.21041830954017e+01, 3.20243522962487e+01, 3.19517135133481e+01, 3.18861824835759e+01, 3.18275952678233e+01, 3.17757151816896e+01, 3.17302407255822e+01, 3.16908143657719e+01, 3.16570320015869e+01, 3.16284529377869e+01, 3.16046101657580e+01, 3.15850207433447e+01, 3.15691960519620e+01, 3.15566517023917e+01, 3.15469168586903e+01, 3.15395427541644e+01, 3.15341101853959e+01, 3.15302357904204e+01, 3.15275769454760e+01, 3.15258351507889e+01, 3.15247578186122e+01, 3.15241384246401e+01, 3.15238150350267e+01, 3.15236672733259e+01, 3.15236118424361e+01, 3.15235967638547e+01, 3.15235945382864e+01, 3.15235943418111e+01, 3.15235904919737e+01, 3.15235644087494e+01, 3.15234685239085e+01, 3.15232129289819e+01, 3.15226535599261e+01, 3.15215822789989e+01, 3.15197192427868e+01, 3.15167078938111e+01, 3.15121128536988e+01, 3.15054209280044e+01, 3.14960453554955e+01, 3.14833333471268e+01, 3.14665768614987e+01, 3.14450264547032e+01, 3.14179079249049e+01, 3.13844413493580e+01, 3.13438619894224e+01, 3.12954424250103e+01, 3.12385151826944e+01, 3.11724950509260e+01, 3.10969002401888e+01, 3.10113715520308e+01, 3.09156887717431e+01, 3.08097835934894e+01, 3.06937485178282e+01, 3.05678413202252e+01, 3.04324848643017e+01, 3.02882622153858e+01, 3.01359071918773e+01, 2.99762906711726e+01, 2.98104031422239e+01, 2.96393341645863e+01, 2.94642495435061e+01, 2.92863671431195e+01, 2.91069323109472e+01, 2.89271938560943e+01, 2.87483814048487e+01, 2.85716847660389e+01, 2.83982357094871e+01, 2.82290923379698e+01, 2.80652260542401e+01, 2.79075110104038e+01, 2.77567158771305e+01, 2.76134977688645e+01, 2.74783981859521e+01, 2.73518408651369e+01, 2.72341314530202e+01, 2.71254589275583e+01, 2.70258986910646e+01, 2.69354172480659e+01, 2.68538783667530e+01, 2.67810506066951e+01, 2.67166160795145e+01, 2.66601802936504e+01, 2.66112829187436e+01, 2.65694092889630e+01, 2.65340024474393e+01, 2.65044755161119e+01, 2.64802241575907e+01, 2.64606388795397e+01, 2.64451169194795e+01, 2.64330734408609e+01, 2.64239517717928e+01, 2.64172324276412e+01, 2.64124406790097e+01, 2.64091524578391e+01, 2.64069984361674e+01, 2.64056661632764e+01, 2.64049002055729e+01, 2.64045002970239e+01, 2.64043175733130e+01, 2.64042490269861e+01, 2.64042303806897e+01, 2.64042276285342e+01, 2.64042274077773e+01, 2.64042233336922e+01, 2.64041957311578e+01, 2.64040942611888e+01, 2.64038237784876e+01, 2.64032318285425e+01, 2.64020981537728e+01, 2.64001266218404e+01, 2.63969399356119e+01, 2.63920774222994e+01, 2.63849961283165e+01, 2.63750753651910e+01, 2.63616247592921e+01, 2.63438957534149e+01, 2.63210963917113e+01, 2.62924090929048e+01, 2.62570109840049e+01, 2.62140962338933e+01, 2.61628997013075e+01, 2.61027211045985e+01, 2.60329488414953e+01, 2.59530825455824e+01, 2.58627534697254e+01, 2.57617418391570e+01, 2.56499904178585e+01, 2.55276136761991e+01, 2.53949021266862e+01, 2.52523215969840e+01, 2.51005074231874e+01, 2.49402537604796e+01, 2.47724984126550e+01, 2.45983037673006e+01, 2.44188345805485e+01, 2.42353334747888e+01, 2.40490950853465e+01, 2.38614398105022e+01, 2.36736880800543e+01, 2.34871359635563e+01, 2.33030327998407e+01, 2.31225613596792e+01, 2.29468208720377e+01, 2.27768130700665e+01, 2.26134312613824e+01, 2.24574523087369e+01, 2.23095313260739e+01, 2.21701988498604e+01, 2.20398602308032e+01, 2.19187969987047e+01, 2.18071699747547e+01, 2.17050239333260e+01, 2.16122936432218e+01, 2.15288111420552e+01, 2.14543141144395e+01, 2.13884552537529e+01, 2.13308124881975e+01, 2.12808999452295e+01, 2.12381795151744e+01, 2.12020728563761e+01, 2.11719736622712e+01, 2.11472599873869e+01, 2.11273064067226e+01, 2.11114957637665e+01, 2.10992302489994e+01, 2.10899415454810e+01, 2.10830997829530e+01, 2.10782210581845e+01, 2.10748733076441e+01, 2.10726803587130e+01, 2.10713240363633e+01, 2.10705442614368e+01, 2.10701371416129e+01, 2.10699511235956e+01, 2.10698813415553e+01, 2.10698623591238e+01, 2.10698595573558e+01, 2.10698593412565e+01, 2.10698554608063e+01, 2.10698291701800e+01, 2.10697325229713e+01, 2.10694748963438e+01, 2.10689110834553e+01, 2.10678313016173e+01, 2.10659535121707e+01, 2.10629183983914e+01, 2.10582872902450e+01, 2.10515432584911e+01, 2.10420955240213e+01, 2.10292872398697e+01, 2.10124066019538e+01, 2.09907011302807e+01, 2.09633948366446e+01, 2.09297078614633e+01, 2.08888780273977e+01, 2.08401836291655e+01, 2.07829666678031e+01, 2.07166556548922e+01, 2.06407870691091e+01, 2.05550245532803e+01, 2.04591750009297e+01, 2.03532007981925e+01, 2.02372276551392e+01, 2.01115476690420e+01, 1.99766174946134e+01, 1.98330517329151e+01, 1.96816118708697e+01, 1.95231912889684e+01, 1.93587969930846e+01, 1.91895288119430e+01, 1.90165568371264e+01, 1.88410978761575e+01, 1.86643916529923e+01, 1.84876774355744e+01, 1.83121717047235e+01, 1.81390474054520e+01, 1.79694152395213e+01, 1.78043073638835e+01, 1.76446637525700e+01, 1.74913213627910e+01, 1.73450061273903e+01, 1.72063276862813e+01, 1.70757766799250e+01, 1.69537243660081e+01, 1.68404242888507e+01, 1.67360157271228e+01, 1.66405286626218e+01, 1.65538900427926e+01, 1.64759311440990e+01, 1.64063958755509e+01, 1.63449498870527e+01, 1.62911903632461e+01, 1.62446563894424e+01, 1.62048397725986e+01, 1.61711961885362e+01, 1.61431565087403e+01, 1.61201381385696e+01, 1.61015561762343e+01, 1.60868341813108e+01, 1.60754143256566e+01, 1.60667666909810e+01, 1.60603974781992e+01, 1.60558559056134e+01, 1.60527395966948e+01, 1.60506982936425e+01, 1.60494357788646e+01, 1.60487099411035e+01, 1.60483309834088e+01, 1.60481578333733e+01, 1.60480928786335e+01, 1.60480752093521e+01, 1.60480726014025e+01, 1.60480724039364e+01, 1.60480689058445e+01, 1.60480452057538e+01, 1.60479580816604e+01, 1.60477258404091e+01, 1.60472175840514e+01, 1.60462442044304e+01, 1.60445514644040e+01, 1.60418154801847e+01, 1.60376408683662e+01, 1.60315617622887e+01, 1.60230458339858e+01, 1.60115013784949e+01, 1.59962874256891e+01, 1.59767267408508e+01, 1.59521214602969e+01, 1.59217709857671e+01, 1.58849916364062e+01, 1.58411374375748e+01, 1.57896213207417e+01, 1.57299359287527e+01, 1.56616731762986e+01, 1.55845417156141e+01, 1.54983815088234e+01, 1.54031748132567e+01, 1.52990530415219e+01, 1.51862991551713e+01, 1.50653454746787e+01, 1.49367670198562e+01, 1.48012707124461e+01, 1.46596809564943e+01, 1.45129222473856e+01, 1.43619995402069e+01, 1.42079771344375e+01, 1.40519568143973e+01, 1.38950559369468e+01, 1.37383860928307e+01, 1.35830328951720e+01, 1.34300373720776e+01, 1.32803793598297e+01, 1.31349632065124e+01, 1.29946060017928e+01, 1.28600284483224e+01, 1.27318483884344e+01, 1.26105769033646e+01, 1.24966168186331e+01, 1.23902633848199e+01, 1.22917068616013e+01, 1.22010367153787e+01, 1.21182471449749e+01, 1.20432436712767e+01, 1.19758505596235e+01, 1.19158188820281e+01, 1.18628350642494e+01, 1.18165297956097e+01, 1.17764872038675e+01, 1.17422542115408e+01, 1.17133499933599e+01, 1.16892754478614e+01, 1.16695225814705e+01, 1.16535836834955e+01, 1.16409601485824e+01, 1.16311707828271e+01, 1.16237594142892e+01, 1.16183016210352e+01, 1.16144103922921e+01, 1.16117405521150e+01, 1.16099918004432e+01, 1.16089102627154e+01, 1.16082884845666e+01, 1.16079638599465e+01, 1.16078155361871e+01, 1.16077598948099e+01, 1.16077447590048e+01, 1.16077425249916e+01, 1.16077423572836e+01, 1.16077394054314e+01, 1.16077194062009e+01, 1.16076458869507e+01, 1.16074499119036e+01, 1.16070210267345e+01, 1.16061996658034e+01, 1.16047713294046e+01, 1.16024628041837e+01, 1.15989406600134e+01, 1.15938122097776e+01, 1.15866290628339e+01, 1.15768933365287e+01, 1.15640665115615e+01, 1.15475808257670e+01, 1.15268529979764e+01, 1.15012999618790e+01, 1.14703561740620e+01, 1.14334919474909e+01, 1.13902321600085e+01, 1.13401746063706e+01, 1.12830072113530e+01, 1.12185233090866e+01, 1.11466342264534e+01, 1.10673784896008e+01, 1.09809271020718e+01, 1.08875845162184e+01, 1.07877851276521e+01, 1.06820853527673e+01, 1.05711515859148e+01, 1.04557445576474e+01, 1.03367008102003e+01, 1.02149121540770e+01, 1.00913040568846e+01, 9.96681393450501e+00, 9.84237026428183e+00, 9.71887332654295e+00, 9.59717821765902e+00, 9.47808058314715e+00, 9.36230531370328e+00, 9.25049825058964e+00, 9.14322077642841e+00, 9.04094703494806e+00, 8.94406343438808e+00, 8.85287004408644e+00, 8.76758348751191e+00, 8.68834095974768e+00, 8.61520504379006e+00, 8.54816905860249e+00, 8.48716273436260e+00, 8.43205807017176e+00, 8.38267528200088e+00, 8.33878879096221e+00, 8.30013323278525e+00, 8.26640948844593e+00, 8.23729074388600e+00, 8.21242858485001e+00, 8.19145912262188e+00, 8.17400912968724e+00, 8.15970214311907e+00, 8.14816446997908e+00, 8.13903100555884e+00, 8.13195075418708e+00, 8.12659192584080e+00, 8.12264647189955e+00, 8.11983392162244e+00, 8.11790438825966e+00, 8.11664063036046e+00, 8.11585907924553e+00, 8.11540977644968e+00, 8.11517520323063e+00, 8.11506802555019e+00, 8.11502781962515e+00, 8.11501688265031e+00, 8.11501526837615e+00, 8.11501514756059e+00, 8.11501302597889e+00, 8.11499865195770e+00, 8.11494581163995e+00, 8.11480495955699e+00, 8.11449671229747e+00, 8.11390639754897e+00, 8.11287988415466e+00, 8.11122089689849e+00, 8.10868999386347e+00, 8.10500535278765e+00, 8.09984547375073e+00, 8.09285385651656e+00, 8.08364565006481e+00, 8.07181619834511e+00, 8.05695132066422e+00, 8.03863906994561e+00, 8.01648261223625e+00, 7.99011377341311e+00, 7.95920671310895e+00, 7.92349112160981e+00, 7.88276430300519e+00, 7.83690151583357e+00, 7.78586399644237e+00, 7.72970419141707e+00, 7.66856786947848e+00, 7.60269296032591e+00, 7.53240516316858e+00, 7.45811056290118e+00, 7.38028566776030e+00, 7.29946542128631e+00, 7.21622983035479e+00, 7.13118988323697e+00, 7.04497340783496e+00, 6.95821144822609e+00, 6.87152563069269e+00, 6.78551686466759e+00, 6.70075559566728e+00, 6.61777370994096e+00, 6.53705809355360e+00, 6.45904577631890e+00, 6.38412054326779e+00, 6.31261086977433e+00, 6.24478902597541e+00, 6.18087119644899e+00, 6.12101846788401e+00, 6.06533854770667e+00, 6.01388808870301e+00, 5.96667550794526e+00, 5.92366420257979e+00, 5.88477608004145e+00, 5.84989533550540e+00, 5.81887242400176e+00, 5.79152818748751e+00, 5.76765810713958e+00, 5.74703665722911e+00, 5.72942173854494e+00, 5.71455916631194e+00, 5.70218718026061e+00, 5.69204093379783e+00, 5.68385690635108e+00, 5.67737716945125e+00, 5.67235342466176e+00, 5.66855072172951e+00, 5.66575075982420e+00, 5.66375467462424e+00, 5.66238522002766e+00, 5.66148826560494e+00, 5.66093354919319e+00, 5.66061464734083e+00, 5.66044815328492e+00, 5.66037208110690e+00, 5.66034354386021e+00, 5.66033578104226e+00, 5.66033463526677e+00, 5.66033454948250e+00, 5.66033304263900e+00, 5.66032283355624e+00, 5.66028530394231e+00, 5.66018526418728e+00, 5.65996633174480e+00, 5.65954705802919e+00, 5.65881796287774e+00, 5.65763961936879e+00, 5.65584191269901e+00, 5.65322457392746e+00, 5.64955906138637e+00, 5.64459182807531e+00, 5.63804897128803e+00, 5.62964221058423e+00, 5.61907608249613e+00, 5.60605617680318e+00, 5.59029817298440e+00, 5.57153737112435e+00, 5.54953835476313e+00, 5.52410438017296e+00, 5.49508606335273e+00, 5.46238893762944e+00, 5.42597948419538e+00, 5.38588929568091e+00, 5.34221711660547e+00, 5.29512860921848e+00, 5.24485381168108e+00, 5.19168237932560e+00, 5.13595682005683e+00, 5.07806404334122e+00, 5.01842563094087e+00, 4.95748729984462e+00, 4.89570805824859e+00, 4.83354955027654e+00, 4.77146604348801e+00, 4.70989543799124e+00, 4.64925157451734e+00, 4.58991800261646e+00, 4.53224325322092e+00, 4.47653755608695e+00, 4.42307086301926e+00, 4.37207198818554e+00, 4.32372865740921e+00, 4.27818826441184e+00, 4.23555915597068e+00, 4.19591230155016e+00, 4.15928323891516e+00, 4.12567422036106e+00, 4.09505651157915e+00, 4.06737281578464e+00, 4.04253980979765e+00, 4.02045078714448e+00, 4.00097840696756e+00, 3.98397754751745e+00, 3.96928825997730e+00, 3.95673881289359e+00, 3.94614881002647e+00, 3.93733235546722e+00, 3.93010122996279e+00, 3.92426803222235e+00, 3.91964922936628e+00, 3.91606805249870e+00, 3.91335716754135e+00, 3.91136104877921e+00, 3.90993798368762e+00, 3.90896164292927e+00, 3.90832215899768e+00, 3.90792667055367e+00, 3.90769930642362e+00, 3.90758060256869e+00, 3.90752636595931e+00, 3.90750601996093e+00, 3.90750048535853e+00, 3.90749966846290e+00, 3.90749960713965e+00, 3.90749852779884e+00, 3.90749121510820e+00, 3.90746433291635e+00, 3.90739267511223e+00, 3.90723585502674e+00, 3.90693553065332e+00, 3.90641327922338e+00, 3.90556922279190e+00, 3.90428149254429e+00, 3.90240660284101e+00, 3.89978078565387e+00, 3.89622231104362e+00, 3.89153478889012e+00, 3.88551141086834e+00, 3.87794005004679e+00, 3.86860908980527e+00, 3.85731380645941e+00, 3.84386308455657e+00, 3.82808620465523e+00, 3.80983941533168e+00, 3.78901198877544e+00, 3.76553146625074e+00, 3.73936782776642e+00, 3.71053636898785e+00, 3.67909913458376e+00, 3.64516483524338e+00, 3.60888725823461e+00, 3.57046226073979e+00, 3.53012350416544e+00, 3.48813714092435e+00, 3.44479570016445e+00, 3.40041143554944e+00, 3.35530939856927e+00, 3.30982048823086e+00, 3.26427470580276e+00, 3.21899481448174e+00, 3.17429057053043e+00, 3.13045365604717e+00, 3.08775340522833e+00, 3.04643337702573e+00, 3.00670878907351e+00, 2.96876479254361e+00, 2.93275553714491e+00, 2.89880395150509e+00, 2.86700214779658e+00, 2.83741235107077e+00, 2.81006825293328e+00, 2.78497669484053e+00, 2.76211959684991e+00, 2.74145606130179e+00, 2.72292459582849e+00, 2.70644541462073e+00, 2.69192278968066e+00, 2.67924743386339e+00, 2.66829890424687e+00, 2.65894801754515e+00, 2.65105926899911e+00, 2.64449324285125e+00, 2.63910899679060e+00, 2.63476639548306e+00, 2.63132836044938e+00, 2.62866299612951e+00, 2.62664554596760e+00, 2.62516012863248e+00, 2.62410120373545e+00, 2.62337471903741e+00, 2.62289889726775e+00, 2.62260463010579e+00, 2.62243545910005e+00, 2.62234713757326e+00, 2.62230678296292e+00, 2.62229164459173e+00, 2.62228752659115e+00, 2.62228691878301e+00, 2.62228687296135e+00, 2.62228606387307e+00, 2.62228058218329e+00, 2.62226043094751e+00, 2.62220671542568e+00, 2.62208916171179e+00, 2.62186403762030e+00, 2.62147256163018e+00, 2.62083987718924e+00, 2.61987465969051e+00, 2.61846941138430e+00, 2.61650148464478e+00, 2.61383485559393e+00, 2.61032264746544e+00, 2.60581037597313e+00, 2.60013985755892e+00, 2.59315368664063e+00, 2.58470015152061e+00, 2.57463842286553e+00, 2.56284381665091e+00, 2.54921290857873e+00, 2.53366826261859e+00, 2.51616253546012e+00, 2.49668173337569e+00, 2.47524742907197e+00, 2.45191779277859e+00, 2.42678735163487e+00, 2.39998546037987e+00, 2.37167353917405e+00, 2.34204120510007e+00, 2.31130148642605e+00, 2.27968535759366e+00, 2.24743586389997e+00, 2.21480211555605e+00, 2.18203342091877e+00, 2.14937380004492e+00, 2.11705707603998e+00, 2.08530268805773e+00, 2.05431231198896e+00, 2.02426731847631e+00, 1.99532704770937e+00, 1.96762783997841e+00, 1.94128273212174e+00, 1.91638171317668e+00, 1.89299242681778e+00, 1.87116121169262e+00, 1.85091438119008e+00, 1.83225965902985e+00, 1.81518770405868e+00, 1.79967367486827e+00, 1.78567880086573e+00, 1.77315194024496e+00, 1.76203111634862e+00, 1.75224503192609e+00, 1.74371456577423e+00, 1.73635425835644e+00, 1.73007379251659e+00, 1.72477947270538e+00, 1.72037570166270e+00, 1.71676644775572e+00, 1.71385668972228e+00, 1.71155381902070e+00, 1.70976897396358e+00, 1.70841827492789e+00, 1.70742392673510e+00, 1.70671515322148e+00, 1.70622893035239e+00, 1.70591048805954e+00, 1.70571355716731e+00, 1.70560034598531e+00, 1.70554124085838e+00, 1.70551423553746e+00, 1.70550410495775e+00, 1.70550134919902e+00, 1.70550094245496e+00, 1.70550091162433e+00, 1.70550036502411e+00, 1.70549666172986e+00, 1.70548304806867e+00, 1.70544675935501e+00, 1.70536734395969e+00, 1.70521525998141e+00, 1.70495080345497e+00, 1.70452342183800e+00, 1.70387146033537e+00, 1.70292238094034e+00, 1.70159348422509e+00, 1.69979315119483e+00, 1.69742260633436e+00, 1.69437818300378e+00, 1.69055404871855e+00, 1.68584532131188e+00, 1.68015147894614e+00, 1.67337993952798e+00, 1.66544966096961e+00, 1.65629459593474e+00, 1.64586682617350e+00, 1.63413920475801e+00, 1.62110735098189e+00, 1.60679087252109e+00, 1.59123373119252e+00, 1.57450371918527e+00, 1.55669106751220e+00, 1.53790626233688e+00, 1.51827719236315e+00, 1.49794578690542e+00, 1.47706432625833e+00, 1.45579161210669e+00, 1.43428917655537e+00, 1.41271768634791e+00, 1.39123366774329e+00, 1.36998664172188e+00, 1.34911672297298e+00, 1.32875270302937e+00, 1.30901061038780e+00, 1.28999271967671e+00, 1.27178696797193e+00, 1.25446672846946e+00, 1.23809088869746e+00, 1.22270418097795e+00, 1.20833771575703e+00, 1.19500967281054e+00, 1.18272611058288e+00, 1.17148185964279e+00, 1.16126147219156e+00, 1.15204020552877e+00, 1.14378502314912e+00, 1.13645560245243e+00, 1.13000534261351e+00, 1.12438236970334e+00, 1.11953053845163e+00, 1.11539043095411e+00, 1.11190035213266e+00, 1.10899731995430e+00, 1.10661804553257e+00, 1.10469989460629e+00, 1.10318181793218e+00, 1.10200523429877e+00, 1.10111484663645e+00, 1.10045936948475e+00, 1.09999214522630e+00, 1.09967162722980e+00, 1.09946171044077e+00, 1.09933189394445e+00, 1.09925726537494e+00, 1.09921830341364e+00, 1.09920050157212e+00, 1.09919382351920e+00, 1.09919200692995e+00, 1.09919173880529e+00, 1.09919171835285e+00, 1.09919135404766e+00, 1.09918888582833e+00, 1.09917981242089e+00, 1.09915562626254e+00, 1.09910269653533e+00, 1.09900133394790e+00, 1.09882507661472e+00, 1.09854023296400e+00, 1.09810571353802e+00, 1.09747317849001e+00, 1.09658752109065e+00, 1.09538769916630e+00, 1.09380791571274e+00, 1.09177913676081e+00, 1.08923091900752e+00, 1.08609350225603e+00, 1.08230010325279e+00, 1.07778932942819e+00, 1.07250761504009e+00, 1.06641157016795e+00, 1.05947012672939e+00, 1.05166636667196e+00, 1.04299892660386e+00, 1.03348289042349e+00, 1.02315010614475e+00, 1.01204889340613e+00, 1.00024314177718e+00, 9.87810834291473e-01, 9.74842063007294e-01, 9.61436631485221e-01, 9.47701360994216e-01, 9.33747231639302e-01, 9.19686495532687e-01, 9.05629896128531e-01, 8.91684115905413e-01, 8.77949554323230e-01, 8.64518510824361e-01, 8.51473815912094e-01, 8.38887920133686e-01, 8.26822419616846e-01, 8.15327970941901e-01, 8.04444530010798e-01, 7.94201840329031e-01, 7.84620095446826e-01, 7.75710706711806e-01, 7.67477118730547e-01, 7.59915628605960e-01, 7.53016178968259e-01, 7.46763107465974e-01, 7.41135845783518e-01, 7.36109569027484e-01, 7.31655801518620e-01, 7.27742987930881e-01, 7.24337039736149e-01, 7.21401866452439e-01, 7.18899899624738e-01, 7.16792615103557e-01, 7.15041056287305e-01, 7.13606357789729e-01, 7.12450265693931e-01, 7.11535647372358e-01, 7.10826981000722e-01, 7.10290812585775e-01, 7.09896166760345e-01, 7.09614896937566e-01, 7.09421960770243e-01, 7.09295608266935e-01, 7.09217472324617e-01, 7.09172554714250e-01, 7.09149104485022e-01, 7.09138390066862e-01, 7.09134370748671e-01, 7.09133277399502e-01, 7.09133116023565e-01, 7.09133103627011e-01, 7.09132881676929e-01, 7.09131377934114e-01, 7.09125850040955e-01, 7.09111114895215e-01, 7.09078868342523e-01, 7.09017115989184e-01, 7.08909739556780e-01, 7.08736221367587e-01, 7.08471547326582e-01, 7.08086305876535e-01, 7.07546996492859e-01, 7.06816555972644e-01, 7.05855103730798e-01, 7.04620898354712e-01, 7.03071486836343e-01, 7.01165015561905e-01, 6.98861659032900e-01, 6.96125109532780e-01, 6.92924059968995e-01, 6.89233604507059e-01, 6.85036478920324e-01, 6.80324066034402e-01, 6.75097101873085e-01, 6.69366034892768e-01, 6.63151012838789e-01, 6.56481497155973e-01, 6.49395530741813e-01, 6.41938708095040e-01, 6.34162914818624e-01, 6.26124914027649e-01, 6.17884859722885e-01, 6.09504812153602e-01, 6.01047319255912e-01, 5.92574113737882e-01, 5.84144959738488e-01, 5.75816668286642e-01, 5.67642288409360e-01, 5.59670471289039e-01, 5.51944998302389e-01, 5.44504459629856e-01, 5.37382067795576e-01, 5.30605589421398e-01, 5.24197378242540e-01, 5.18174492789868e-01, 5.12548882976375e-01, 5.07327631073245e-01, 5.02513234175268e-01, 4.98103917161366e-01, 4.94093967242705e-01, 4.90474083321423e-01, 4.87231735406588e-01, 4.84351531110036e-01, 4.81815587651930e-01, 4.79603908758484e-01, 4.77694766283719e-01, 4.76065086326848e-01, 4.74690839077915e-01, 4.73547430675680e-01, 4.72610094104404e-01, 4.71854274716534e-01, 4.71256004493462e-01, 4.70792257802401e-01, 4.70441280327570e-01, 4.70182882188496e-01, 4.69998686116547e-01, 4.69872322015483e-01, 4.69789560300654e-01, 4.69738378064520e-01, 4.69708954265393e-01, 4.69693592652838e-01, 4.69686573865083e-01, 4.69683940885162e-01, 4.69683224651696e-01, 4.69683118937169e-01, 4.69683110754431e-01, 4.69682963442459e-01, 4.69681965382826e-01, 4.69678296420042e-01, 4.69668516389255e-01, 4.69647113421374e-01, 4.69606125672911e-01, 4.69534852313792e-01, 4.69419668178354e-01, 4.69243954787700e-01, 4.68988158353397e-01, 4.68629982700492e-01, 4.68144721617847e-01, 4.67505730802600e-01, 4.66685034233471e-01, 4.65654053526180e-01, 4.64384441795026e-01, 4.62848996151031e-01, 4.61022615773899e-01, 4.58883266211366e-01, 4.56412905967241e-01, 4.53598329295033e-01, 4.50431880021611e-01, 4.46911995526655e-01, 4.43043547680183e-01, 4.38837958179874e-01, 4.34313078539056e-01, 4.29492838873813e-01, 4.24406683380859e-01, 4.19088822745806e-01, 4.13577343585793e-01, 4.07913221604801e-01, 4.02139287983950e-01, 3.96299197594580e-01, 3.90436443246699e-01, 3.84593453024188e-01, 3.78810798697799e-01, 3.73126533249533e-01, 3.67575665678324e-01, 3.62189772354100e-01, 3.56996736891702e-01, 3.52020605204999e-01, 3.47281539178425e-01, 3.42795851122752e-01, 3.38576101556370e-01, 3.34631244463310e-01, 3.30966806581081e-01, 3.27585090050363e-01, 3.24485390561229e-01, 3.21664225697006e-01, 3.19115570338219e-01, 3.16831097659502e-01, 3.14800425410986e-01, 3.13011367841533e-01, 3.11450193840173e-01, 3.10101891703077e-01, 3.08950440441016e-01, 3.07979086797197e-01, 3.07170626217932e-01, 3.06507684986293e-01, 3.05972999671070e-01, 3.05549689044497e-01, 3.05221512768846e-01, 3.04973110527811e-01, 3.04790214958729e-01, 3.04659831786125e-01, 3.04570381000374e-01, 3.04511793775170e-01, 3.04475561046360e-01, 3.04454731224467e-01, 3.04443856296004e-01, 3.04438887477794e-01, 3.04437023506821e-01, 3.04436516461919e-01, 3.04436441623174e-01, 3.04436435783380e-01, 3.04436330044247e-01, 3.04435613646461e-01, 3.04432980098106e-01, 3.04425960068862e-01, 3.04410597128673e-01, 3.04381176111299e-01, 3.04330015377310e-01, 3.04247333173485e-01, 3.04121197009223e-01, 3.03937563409860e-01, 3.03680413445932e-01, 3.03331986954319e-01, 3.02873115259416e-01, 3.02283648417031e-01, 3.01542968574424e-01, 3.00630576109789e-01, 2.99526730060720e-01, 2.98213119377486e-01, 2.96673537256247e-01, 2.94894527785480e-01, 2.92865972922556e-01, 2.90581588839700e-01, 2.88039304167995e-01, 2.85241498570187e-01, 2.82195088017453e-01, 2.78911452456242e-01, 2.75406211324103e-01, 2.71698861598118e-01, 2.67812300770643e-01, 2.63772262579330e-01, 2.59606696027376e-01, 2.55345118158345e-01, 2.51017968519704e-01, 2.46655988888989e-01, 2.42289646440454e-01, 2.37948612899764e-01, 2.33661307036130e-01, 2.29454503512102e-01, 2.25353007848454e-01, 2.21379395061582e-01, 2.17553808255779e-01, 2.13893812888507e-01, 2.10414302325340e-01, 2.07127450428125e-01, 2.04042707086884e-01, 2.01166832729974e-01, 1.98503967914931e-01, 1.96055734180817e-01, 1.93821362476044e-01, 1.91797845695312e-01, 1.89980112139921e-01, 1.88361217018965e-01, 1.86932549389276e-01, 1.85684052144069e-01, 1.84604452770387e-01, 1.83681502568447e-01, 1.82902221849751e-01, 1.82253148300639e-01, 1.81720585241841e-01, 1.81290845985693e-01, 1.80950489949353e-01, 1.80686545708035e-01, 1.80486715839585e-01, 1.80339558287109e-01, 1.80234639107418e-01, 1.80162651902926e-01, 1.80115499969297e-01, 1.80086338198244e-01, 1.80069573019125e-01, 1.80060820068334e-01, 1.80056820769018e-01, 1.80055320493210e-01, 1.80054912381728e-01, 1.80054852145339e-01, 1.80054847405010e-01, 1.80054761060816e-01, 1.80054176066328e-01, 1.80052025566702e-01, 1.80046293139668e-01, 1.80033747941543e-01, 1.80009722686030e-01, 1.79967943503129e-01, 1.79900419755064e-01, 1.79797400752150e-01, 1.79647404964008e-01, 1.79437325694614e-01, 1.79152615218513e-01, 1.78777547002028e-01, 1.78295552840297e-01, 1.77689628559037e-01, 1.76942798460153e-01, 1.76038625105569e-01, 1.74961747587155e-01, 1.73698428423911e-01, 1.72237086994556e-01, 1.70568796263119e-01, 1.68687719740483e-01, 1.66591467258506e-01, 1.64281351221959e-01, 1.61762529341914e-01, 1.59044025158728e-01, 1.56138623518160e-01, 1.53062644154403e-01, 1.49835602261923e-01, 1.46479770058271e-01, 1.43019657591893e-01, 1.39481434213501e-01, 1.35892314089681e-01, 1.32279929863564e-01, 1.28671718126782e-01, 1.25094338861160e-01, 1.21573148539056e-01, 1.18131743129219e-01, 1.14791582936761e-01, 1.11571706181167e-01, 1.08488533025216e-01, 1.05555756989233e-01, 1.02784316804772e-01, 1.00182439018475e-01, 9.77557398282229e-02, 9.55073736906158e-02, 9.34382160152794e-02, 9.15470679164017e-02, 8.98308724467185e-02, 8.82849338992955e-02, 8.69031340472475e-02, 8.56781414489731e-02, 8.46016117710064e-02, 8.36643784838316e-02, 8.28566343049886e-02, 8.21681044088294e-02, 8.15882126974046e-02, 8.11062422689162e-02, 8.07114907923889e-02, 8.03934207885248e-02, 8.01418039151116e-02, 7.99468575188484e-02, 7.97993708660343e-02, 7.96908178895137e-02, 7.96134529263526e-02, 7.95603858146734e-02, 7.95256330631986e-02, 7.95041422906024e-02, 7.94917880711097e-02, 7.94853382999463e-02, 7.94823914013421e-02, 7.94812859287717e-02, 7.94809852135430e-02, 7.94809408292068e-02, 7.94809373046448e-02, 7.94808727251208e-02, 7.94804351979010e-02, 7.94788268090230e-02, 7.94745395016818e-02, 7.94651571708869e-02, 7.94471901700220e-02, 7.94159494264903e-02, 7.93654668868626e-02, 7.92884684098638e-02, 7.91764045100152e-02, 7.90195431930272e-02, 7.88071277929205e-02, 7.85276005862485e-02, 7.81688905640578e-02, 7.77187605956300e-02, 7.71652056804086e-02, 7.64968901404188e-02, 7.57036077765747e-02, 7.47767454417172e-02, 7.37097277424894e-02, 7.24984188226330e-02, 7.11414570881317e-02, 6.96405000966678e-02, 6.80003603180026e-02, 6.62290173849485e-02, 6.43374994791444e-02, 6.23396343182634e-02, 6.02516791547814e-02, 5.80918477491963e-02, 5.58797595177367e-02, 5.36358408835724e-02, 5.13807111978934e-02, 4.91345849323466e-02, 4.69167201428036e-02, 4.47288965974550e-02, 4.26215519758486e-02, 4.05898744847661e-02, 3.86455400772334e-02, 3.67978946746703e-02, 3.50539800280032e-02, 3.34186387106854e-02, 3.18946806280067e-02, 3.04830925334159e-02, 2.91832727314443e-02, 2.79932750448069e-02, 2.69100488156226e-02, 2.59296647916472e-02, 2.50475198734890e-02, 2.42585166058760e-02, 2.35572158140727e-02, 2.29379628211817e-02, 2.23949892126001e-02, 2.19224931558628e-02, 2.15147018909647e-02, 2.11659202373866e-02, 2.08705688866786e-02, 2.06232159248714e-02, 2.04186045160810e-02, 2.02516790295988e-02, 2.01176111590584e-02, 2.00118268135157e-02, 1.99300338067090e-02, 1.98682496823750e-02, 1.98228284375404e-02, 1.97904844828329e-02, 1.97683119409788e-02, 1.97537973458708e-02, 1.97448239656808e-02, 1.97396663129012e-02, 1.97369738868503e-02, 1.97357437712924e-02, 1.97352823248096e-02, 1.97351568013318e-02, 1.97351382743776e-02, 1.97351367900183e-02, 1.97351094176668e-02, 1.97349239666735e-02, 1.97342422394229e-02, 1.97324250835607e-02, 1.97284486816471e-02, 1.97208349260549e-02, 1.97075993276954e-02, 1.96862199302856e-02, 1.96536306911597e-02, 1.96062426586536e-02, 1.95399959154869e-02, 1.94504445488164e-02, 1.93328757478071e-02, 1.91824624472332e-02, 1.89944467270919e-02, 1.87643485203779e-02, 1.84881912546686e-02, 1.81627331388583e-02, 1.77856902723761e-02, 1.73559360161306e-02, 1.68736605281765e-02, 1.63404753579997e-02, 1.57594506794609e-02, 1.51350770746177e-02, 1.44731494560162e-02, 1.37805771880295e-02, 1.30651310052698e-02, 1.23351431111516e-02, 1.15991811026319e-02, 1.08657185223677e-02, 1.01428245938351e-02, 9.43789310761645e-03, 8.75742589898465e-03, 8.10688055552333e-03, 7.49058572692007e-03, 6.91172148215458e-03, 6.37235723988994e-03, 5.87353632553089e-03, 5.41539435641880e-03, 4.99729834729123e-03, 4.61799439636934e-03, 4.27575367920952e-03, 3.96850883869604e-03, 3.69397534681326e-03, 3.44975474051523e-03, 3.23341861071214e-03, 3.04257375756123e-03, 2.87491000172935e-03, 2.72823280054979e-03, 2.60048313289474e-03, 2.48974717003136e-03, 2.39425812549854e-03, 2.31239244271109e-03, 2.24266219144909e-03, 2.18370524026312e-03, 2.13427447614123e-03, 2.09322706664494e-03, 2.05951450805486e-03, 2.03217397519495e-03, 2.01032128197198e-03, 1.99314557424657e-03, 1.97990570814000e-03, 1.96992811899577e-03, 1.96260586301378e-03, 1.95739842092735e-03, 1.95383179588760e-03, 1.95149842057418e-03, 1.95005641250390e-03, 1.94922777883022e-03, 1.94879526694081e-03, 1.94859767368882e-03, 1.94852355369335e-03, 1.94850339163111e-03, 1.94850041577155e-03, 1.94850017510362e-03, 1.94849570902988e-03, 1.94846545092986e-03, 1.94835422228579e-03, 1.94805775306861e-03, 1.94740906694335e-03, 1.94616725153907e-03, 1.94400927595950e-03, 1.94052557693425e-03, 1.93522019000625e-03, 1.92751622389409e-03, 1.91676742799984e-03, 1.90227644518615e-03, 1.88403299073769e-03, 1.85990104009193e-03, 1.82989532941491e-03, 1.79342100398117e-03, 1.74941572752374e-03, 1.69890128667305e-03, 1.64113622942352e-03, 1.57629029656903e-03, 1.50478940372523e-03, 1.42731560438136e-03, 1.34478951171246e-03, 1.25833527102515e-03, 1.16923010596537e-03, 1.07932382547915e-03, 9.89045340439532e-04, 9.00178430960352e-04, 8.14008016163825e-04, 7.31586789918026e-04, 6.53794507968541e-04, 5.81418365378283e-04, 5.14800150007488e-04, 4.54097443835622e-04, 3.99361094267192e-04, 3.50401706818104e-04, 3.06890404438320e-04, 2.68719086832903e-04, 2.35465804593105e-04, 2.06321787352692e-04, 1.81190550364363e-04, 1.59507703372697e-04, 1.40390132379764e-04, 1.24433138033453e-04, 1.10761385734293e-04, 9.87115751769352e-05, 8.79287379618177e-05, 7.81554725577394e-05, 6.92329227689470e-05, 6.10726107197479e-05, 5.35972233848270e-05, 4.73788051067459e-05, 4.16826131655057e-05, 3.63396167557539e-05, 3.13962348883479e-05, 2.68869169772756e-05, 2.28556986741069e-05, 1.92783748929060e-05, 1.61056837555407e-05, 1.32767605265371e-05, 1.07712316079756e-05, 8.57723764369211e-06, 6.68423314920119e-06, 5.07982858266587e-06, 3.74825178539595e-06, 2.66991370523328e-06, 1.82169333129423e-06, 1.17761999973709e-06, 7.09720922031920e-07, 3.88896438607844e-07, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15],
[ 3.80000000000000e+01, 3.79999999687886e+01, 3.79999990036692e+01, 3.79999924648332e+01, 3.79999684272255e+01, 3.79999043505997e+01, 3.79997641144366e+01, 3.79994955222825e+01, 3.79990283659030e+01, 3.79982731226721e+01, 3.79971203403808e+01, 3.79954407428462e+01, 3.79930860681107e+01, 3.79898906294849e+01, 3.79856735690236e+01, 3.79802417540181e+01, 3.79733932503471e+01, 3.79649212924873e+01, 3.79546186587860e+01, 3.79422823521299e+01, 3.79277184800122e+01, 3.79107472236287e+01, 3.78912077823880e+01, 3.78689631775505e+01, 3.78439047963323e+01, 3.78159565558507e+01, 3.77850785653299e+01, 3.77512701660990e+01, 3.77145722334455e+01, 3.76750686338077e+01, 3.76328867463066e+01, 3.75881969799553e+01, 3.75412112469562e+01, 3.74921803873402e+01, 3.74413905789635e+01, 3.73891588070110e+01, 3.73358275057459e+01, 3.72817585193671e+01, 3.72273265559581e+01, 3.71729123268070e+01, 3.71188955718909e+01, 3.70656481710606e+01, 3.70135275302631e+01, 3.69628704145093e+01, 3.69139873761173e+01, 3.68671579000623e+01, 3.68226263599539e+01, 3.67805988498676e+01, 3.67412409302071e+01, 3.67046763008008e+01, 3.66709863919437e+01, 3.66402108442136e+01, 3.66123488304734e+01, 3.65873611582982e+01, 3.65651730778239e+01, 3.65456777084769e+01, 3.65287399880299e+01, 3.65142010388991e+01, 3.65018828396079e+01, 3.64915930840845e+01, 3.64831301081742e+01, 3.64762877617435e+01, 3.64708601063180e+01, 3.64666458225900e+01, 3.64634522195103e+01, 3.64610987470975e+01, 3.64594199284535e+01, 3.64582676425274e+01, 3.64575127075333e+01, 3.64570457350717e+01, 3.64567772463233e+01, 3.64566370634708e+01, 3.64565730110483e+01, 3.64565489824951e+01, 3.64565424461198e+01, 3.64565414813636e+01, 3.64565413548079e+01, 3.64565384062122e+01, 3.64565184290250e+01, 3.64564449905033e+01, 3.64562492283059e+01, 3.64558207970657e+01, 3.64550002599132e+01, 3.64535732135503e+01, 3.64512663907876e+01, 3.64477459335335e+01, 3.64426179729440e+01, 3.64354315927339e+01, 3.64256841875906e+01, 3.64128291624188e+01, 3.63962858510687e+01, 3.63754514664699e+01, 3.63497148286758e+01, 3.63184715537716e+01, 3.62811403250971e+01, 3.62371798088714e+01, 3.61861057197034e+01, 3.61275074896122e+01, 3.60610639512694e+01, 3.59865574190690e+01, 3.59038855494958e+01, 3.58130703952145e+01, 3.57142641439647e+01, 3.56077511579856e+01, 3.54939460995387e+01, 3.53733881319424e+01, 3.52467314043304e+01, 3.51147322378977e+01, 3.49782336068304e+01, 3.48381476278868e+01, 3.46954368266987e+01, 3.45510949351745e+01, 3.44061279023061e+01, 3.42615356875350e+01, 3.41182952726560e+01, 3.39773451952545e+01, 3.38395717897077e+01, 3.37057972301014e+01, 3.35767694056142e+01, 3.34531536200638e+01, 3.33355260871580e+01, 3.32243691841968e+01, 3.31200684229603e+01, 3.30229110925389e+01, 3.29330865223131e+01, 3.28506879033945e+01, 3.27757155940046e+01, 3.27080818194467e+01, 3.26476166614517e+01, 3.25940752153935e+01, 3.25471457775033e+01, 3.25064589078190e+01, 3.24715971982465e+01, 3.24421055589562e+01, 3.24175018208664e+01, 3.23972874379745e+01, 3.23809580619014e+01, 3.23680137535502e+01, 3.23579685946456e+01, 3.23503594664415e+01, 3.23447537751057e+01, 3.23407559238768e+01, 3.23380123611439e+01, 3.23362150706786e+01, 3.23351034142817e+01, 3.23344642865044e+01, 3.23341305938379e+01, 3.23339781245350e+01, 3.23339209276526e+01, 3.23339053686777e+01, 3.23339030722043e+01, 3.23339028694693e+01, 3.23338988969788e+01, 3.23338719827607e+01, 3.23337730430978e+01, 3.23335093051285e+01, 3.23329321151798e+01, 3.23318267049447e+01, 3.23299043170608e+01, 3.23267970386139e+01, 3.23220556298989e+01, 3.23151505658376e+01, 3.23054764278150e+01, 3.22923596933625e+01, 3.22750698695089e+01, 3.22528338031495e+01, 3.22248528802620e+01, 3.21903226989632e+01, 3.21484546750087e+01, 3.20984989201351e+01, 3.20397676328187e+01, 3.19716581672881e+01, 3.18936749088965e+01, 3.18054490888867e+01, 3.17067557221458e+01, 3.15975269462638e+01, 3.14778611731928e+01, 3.13480276269741e+01, 3.12084660223097e+01, 3.10597813305837e+01, 3.09027337767322e+01, 3.07382244091719e+01, 3.05672767825579e+01, 3.03910154812514e+01, 3.02106423731722e+01, 3.00274115940566e+01, 2.98426042940612e+01, 2.96575041146397e+01, 2.94733742064380e+01, 2.92914363756289e+01, 2.91128527012357e+01, 2.89387097468557e+01, 2.87700053308850e+01, 2.86076377316753e+01, 2.84523971784525e+01, 2.83049594929472e+01, 2.81658817761265e+01, 2.80356000607013e+01, 2.79144288637326e+01, 2.78025625732119e+01, 2.77000785913032e+01, 2.76069421400044e+01, 2.75230126165741e+01, 2.74480513687515e+01, 2.73817307443103e+01, 2.73236442553559e+01, 2.72733176838923e+01, 2.72302209405159e+01, 2.71937804720816e+01, 2.71633919969681e+01, 2.71384333290640e+01, 2.71182770353480e+01, 2.71023026589822e+01, 2.70899082323969e+01, 2.70805208050992e+01, 2.70736057207047e+01, 2.70686743982447e+01, 2.70652904046521e+01, 2.70630736481252e+01, 2.70617025746032e+01, 2.70609143098360e+01, 2.70605027548287e+01, 2.70603147097182e+01, 2.70602441671446e+01, 2.70602249778202e+01, 2.70602221455148e+01, 2.70602219183289e+01, 2.70602177255961e+01, 2.70601893192054e+01, 2.70600848941752e+01, 2.70598065343616e+01, 2.70591973455287e+01, 2.70580306561496e+01, 2.70560017112953e+01, 2.70527222298866e+01, 2.70477181319615e+01, 2.70404306696872e+01, 2.70302211124032e+01, 2.70163790407961e+01, 2.69981341975893e+01, 2.69746717221836e+01, 2.69451504663847e+01, 2.69087239516158e+01, 2.68645633910249e+01, 2.68118820709336e+01, 2.67499602750360e+01, 2.66781698521483e+01, 2.65959974839576e+01, 2.65030657107894e+01, 2.63991508251237e+01, 2.62841968443216e+01, 2.61583249212523e+01, 2.60218377359574e+01, 2.58752186224165e+01, 2.57191254100652e+01, 2.55543791881625e+01, 2.53819484209309e+01, 2.52029290412659e+01, 2.50185213191465e+01, 2.48300044260108e+01, 2.46387096877432e+01, 2.44459935291115e+01, 2.42532110595521e+01, 2.40616911392813e+01, 2.38727136082091e+01, 2.36874891760749e+01, 2.35071422811881e+01, 2.33326970466727e+01, 2.31650663124968e+01, 2.30050436079410e+01, 2.28532978550366e+01, 2.27103705557336e+01, 2.25766752071082e+01, 2.24524987010042e+01, 2.23380044883464e+01, 2.22332373164862e+01, 2.21381293747504e+01, 2.20525077051829e+01, 2.19761027502922e+01, 2.19085579166454e+01, 2.18494400324191e+01, 2.17982505691007e+01, 2.17544374833954e+01, 2.17174075163164e+01, 2.16865387640536e+01, 2.16611933114768e+01, 2.16407296962968e+01, 2.16245149524497e+01, 2.16119359677089e+01, 2.16024098852513e+01, 2.15953932839240e+01, 2.15903898886856e+01, 2.15869565917791e+01, 2.15847076063397e+01, 2.15833166261358e+01, 2.15825169258697e+01, 2.15820994030734e+01, 2.15819086318180e+01, 2.15818370666684e+01, 2.15818175991875e+01, 2.15818147258272e+01, 2.15818145042060e+01, 2.15818105246004e+01, 2.15817835621817e+01, 2.15816844453919e+01, 2.15814202357566e+01, 2.15808420160610e+01, 2.15797346432511e+01, 2.15778088722062e+01, 2.15746962056174e+01, 2.15699467662689e+01, 2.15630304200314e+01, 2.15533412993214e+01, 2.15402057860686e+01, 2.15228939092619e+01, 2.15006339948659e+01, 2.14726302769741e+01, 2.14380830422929e+01, 2.13962107417029e+01, 2.13462733713575e+01, 2.12875963121770e+01, 2.12195937322981e+01, 2.11417906134343e+01, 2.10538424686489e+01, 2.09555518816136e+01, 2.08468811167584e+01, 2.07279602204927e+01, 2.05990902443452e+01, 2.04607414547198e+01, 2.03135466311015e+01, 2.01582897749200e+01, 1.99958907378347e+01, 1.98273864197369e+01, 1.96539092793847e+01, 1.94766639474814e+01, 1.92969027412760e+01, 1.91159008609950e+01, 1.89349320092185e+01, 1.87552451180282e+01, 1.85780427944746e+01, 1.84044619997197e+01, 1.82355573598486e+01, 1.80722873703312e+01, 1.79155036110119e+01, 1.77659429483282e+01, 1.76242225810817e+01, 1.74908376971065e+01, 1.73661614560345e+01, 1.72504469965669e+01, 1.71438311783004e+01, 1.70463397983727e+01, 1.69578940619166e+01, 1.68783181241266e+01, 1.68073475547706e+01, 1.67446385998709e+01, 1.66897781287381e+01, 1.66422941577076e+01, 1.66016668358375e+01, 1.65673397641024e+01, 1.65387315002833e+01, 1.65152470791272e+01, 1.64962893540313e+01, 1.64812699452456e+01, 1.64696195631572e+01, 1.64607974662388e+01, 1.64542998139287e+01, 1.64496666866701e+01, 1.64464875693818e+01, 1.64444051306134e+01, 1.64431171764774e+01, 1.64423767141153e+01, 1.64419901212070e+01, 1.64418134826126e+01, 1.64417472191994e+01, 1.64417291939276e+01, 1.64417265334346e+01, 1.64417263319901e+01, 1.64417227634207e+01, 1.64416985858351e+01, 1.64416097064246e+01, 1.64413727861778e+01, 1.64408542901326e+01, 1.64398613009739e+01, 1.64381344619510e+01, 1.64353433704688e+01, 1.64310846919698e+01, 1.64248832037966e+01, 1.64161959090090e+01, 1.64044192790090e+01, 1.63888995900882e+01, 1.63689462124977e+01, 1.63438475927204e+01, 1.63128895437253e+01, 1.62753753297589e+01, 1.62306469096117e+01, 1.61781065950354e+01, 1.61172382999543e+01, 1.60476275120960e+01, 1.59689791209791e+01, 1.58811322910778e+01, 1.57840716780876e+01, 1.56779344454974e+01, 1.55630127379277e+01, 1.54397514911788e+01, 1.53087416871622e+01, 1.51707093745173e+01, 1.50265009550447e+01, 1.48770653702230e+01, 1.47234339067884e+01, 1.45666983788621e+01, 1.44079884447905e+01, 1.42484487897373e+01, 1.40892168584256e+01, 1.39314017608323e+01, 1.37760648980712e+01, 1.36242027652982e+01, 1.34767322830255e+01, 1.33344788901884e+01, 1.31981675074936e+01, 1.30684163565399e+01, 1.29457335085405e+01, 1.28305159448700e+01, 1.27230508461846e+01, 1.26235187901241e+01, 1.25319985285606e+01, 1.24484730299871e+01, 1.23728365048455e+01, 1.23049021744100e+01, 1.22444105903289e+01, 1.21910383560521e+01, 1.21444071383830e+01, 1.21040928841465e+01, 1.20696351718637e+01, 1.20405466312891e+01, 1.20163223558577e+01, 1.19964492167521e+01, 1.19804149654026e+01, 1.19677169872504e+01, 1.19578705471684e+01, 1.19504163495216e+01, 1.19449272265063e+01, 1.19410137694339e+01, 1.19383287303936e+01, 1.19365700465398e+01, 1.19354823752831e+01, 1.19348570741245e+01, 1.19345306110925e+01, 1.19343814475569e+01, 1.19343254911845e+01, 1.19343102696963e+01, 1.19343080230366e+01, 1.19343078543791e+01, 1.19343048858168e+01, 1.19342847733743e+01, 1.19342108379654e+01, 1.19340137537518e+01, 1.19335824420087e+01, 1.19327564370007e+01, 1.19313200342229e+01, 1.19289984976575e+01, 1.19254565633696e+01, 1.19202994303061e+01, 1.19130763735410e+01, 1.19032870483189e+01, 1.18903904734586e+01, 1.18738165893550e+01, 1.18529801799114e+01, 1.18272968321386e+01, 1.17962004869451e+01, 1.17591620170021e+01, 1.17157081614411e+01, 1.16654400625694e+01, 1.16080505969479e+01, 1.15433396812953e+01, 1.14712267698245e+01, 1.13917598475892e+01, 1.13051203638643e+01, 1.12116237354285e+01, 1.11117152719106e+01, 1.10059616196950e+01, 1.08950380692545e+01, 1.07797123031215e+01, 1.06608253578553e+01, 1.05392707153907e+01, 1.04159725137898e+01, 1.02918638680831e+01, 1.01678662199904e+01, 1.00448705003001e+01, 9.92372070591216e+00, 9.80520028581731e+00, 9.69002151885788e+00, 9.57881787141717e+00, 9.47213916126098e+00, 9.37044923452845e+00, 9.27412578977712e+00, 9.18346195361661e+00, 9.09866922004550e+00, 9.01988140083873e+00, 8.94715928717397e+00, 8.88049578378333e+00, 8.81982133853530e+00, 8.76500954678407e+00, 8.71588285738714e+00, 8.67221834404934e+00, 8.63375353100607e+00, 8.60019227634740e+00, 8.57121072044236e+00, 8.54646330224904e+00, 8.52558883432040e+00, 8.50821660974450e+00, 8.49397249295427e+00, 8.48248492331627e+00, 8.47339073778714e+00, 8.46634069887768e+00, 8.46100459878180e+00, 8.45707580167515e+00, 8.45427508532972e+00, 8.45235365121989e+00, 8.45109518941403e+00, 8.45031691019008e+00, 8.44986948718295e+00, 8.44963589502786e+00, 8.44952916552254e+00, 8.44948912771033e+00, 8.44947823646500e+00, 8.44947662894039e+00, 8.44947650862998e+00, 8.44947439591894e+00, 8.44946008199720e+00, 8.44940746260337e+00, 8.44926719936105e+00, 8.44896024045310e+00, 8.44837239185681e+00, 8.44735016337939e+00, 8.44569809162900e+00, 8.44317771600096e+00, 8.43950835219137e+00, 8.43436976860487e+00, 8.42740682211749e+00, 8.41823604917051e+00, 8.40645413522673e+00, 8.39164810083311e+00, 8.37340694850223e+00, 8.35133441622135e+00, 8.32506238776880e+00, 8.29426442619703e+00, 8.25866883494450e+00, 8.21807062104881e+00, 8.17234174521320e+00, 8.12143909918569e+00, 8.06540975262539e+00, 8.00439315446554e+00, 7.93862014728076e+00, 7.86840884213209e+00, 7.79415758764048e+00, 7.71633543200969e+00, 7.63547060411738e+00, 7.55213761828128e+00, 7.46694363225230e+00, 7.38051466202165e+00, 7.29348218922090e+00, 7.20647060071672e+00, 7.12008579058325e+00, 7.03490514593247e+00, 6.95146904067093e+00, 6.87027388115510e+00, 6.79176668646907e+00, 6.71634114174308e+00, 6.64433503185742e+00, 6.57602894114359e+00, 6.51164608942198e+00, 6.45135316451266e+00, 6.39526200613918e+00, 6.34343199649632e+00, 6.29587301917783e+00, 6.25254886050153e+00, 6.21338094446896e+00, 6.17825231278548e+00, 6.14701178221233e+00, 6.11947823064347e+00, 6.09544497866141e+00, 6.07468424349292e+00, 6.05695164654141e+00, 6.04199075402502e+00, 6.02953762334038e+00, 6.01932531675740e+00, 6.01108833046167e+00, 6.00456687255541e+00, 5.99951091024352e+00, 5.99568389584402e+00, 5.99286607502741e+00, 5.99085728001694e+00, 5.98947911610762e+00, 5.98857646197471e+00, 5.98801822245233e+00, 5.98769729583120e+00, 5.98752974485795e+00, 5.98745318981009e+00, 5.98742447142943e+00, 5.98741665933932e+00, 5.98741550629138e+00, 5.98741541996264e+00, 5.98741390355494e+00, 5.98740362967366e+00, 5.98736586185790e+00, 5.98726518717852e+00, 5.98704486539323e+00, 5.98662293158534e+00, 5.98588921259474e+00, 5.98470340133179e+00, 5.98289431401655e+00, 5.98026043249148e+00, 5.97657180568108e+00, 5.97157334960863e+00, 5.96498954298301e+00, 5.95653046471248e+00, 5.94589906123473e+00, 5.93279946698163e+00, 5.91694613375034e+00, 5.89807345874890e+00, 5.87594554228281e+00, 5.85036566075171e+00, 5.82118501508451e+00, 5.78831031434394e+00, 5.75170978268407e+00, 5.71141723650932e+00, 5.66753396615663e+00, 5.62022826843262e+00, 5.56973260592987e+00, 5.51633850699031e+00, 5.46038945554481e+00, 5.40227214092694e+00, 5.34240653224392e+00, 5.28123529948620e+00, 5.21921311690677e+00, 5.15679635094199e+00, 5.09443355909359e+00, 5.03255711831901e+00, 4.97157617720836e+00, 4.91187100319571e+00, 4.85378869013766e+00, 4.79764011365474e+00, 4.74369797582492e+00, 4.69219576487254e+00, 4.64332746261005e+00, 4.59724785385232e+00, 4.55407331959399e+00, 4.51388302319583e+00, 4.47672042247835e+00, 4.44259505903938e+00, 4.41148458942106e+00, 4.38333703184115e+00, 4.35807320811856e+00, 4.33558936398749e+00, 4.31575995267841e+00, 4.29844056660107e+00, 4.28347100015566e+00, 4.27067842304187e+00, 4.25988063794613e+00, 4.25088938935847e+00, 4.24351368193000e+00, 4.23756305787853e+00, 4.23285077432105e+00, 4.22919681400377e+00, 4.22643065767593e+00, 4.22439374419243e+00, 4.22294154601979e+00, 4.22194519357456e+00, 4.22129259181844e+00, 4.22088898648182e+00, 4.22065695454856e+00, 4.22053581327486e+00, 4.22048046289628e+00, 4.22045969906967e+00, 4.22045405080646e+00, 4.22045321713469e+00, 4.22045315455205e+00, 4.22045205304538e+00, 4.22044459017714e+00, 4.22041715590780e+00, 4.22034402640087e+00, 4.22018398519003e+00, 4.21987749067485e+00, 4.21934450524591e+00, 4.21848308889721e+00, 4.21716884561938e+00, 4.21525529356148e+00, 4.21257520905758e+00, 4.20894296925361e+00, 4.20415788752823e+00, 4.19800849990354e+00, 4.19027771959181e+00, 4.18074873203263e+00, 4.16921145656342e+00, 4.15546935655640e+00, 4.13934634154312e+00, 4.12069347699627e+00, 4.09939520427160e+00, 4.07537477802945e+00, 4.04859865294718e+00, 4.01907959530725e+00, 3.98687835544361e+00, 3.95210380938209e+00, 3.91491155626942e+00, 3.87550103582969e+00, 3.83411130108767e+00, 3.79101564118965e+00, 3.74651529425002e+00, 3.70093251934799e+00, 3.65460331004444e+00, 3.60787002993576e+00, 3.56107423518383e+00, 3.51454992136096e+00, 3.46861739436883e+00, 3.42357792011337e+00, 3.37970925802839e+00, 3.33726213283196e+00, 3.29645765059605e+00, 3.25748562257908e+00, 3.22050372590829e+00, 3.18563740572381e+00, 3.15298040930811e+00, 3.12259583845803e+00, 3.09451761048516e+00, 3.06875222876785e+00, 3.04528077849254e+00, 3.02406107992736e+00, 3.00502994832075e+00, 2.98810552475103e+00, 2.97318965482751e+00, 2.96017030134885e+00, 2.94892398253282e+00, 2.93931822925774e+00, 2.93121405319422e+00, 2.92446841328777e+00, 2.91893666149596e+00, 2.91447494084692e+00, 2.91094250071599e+00, 2.90820388669417e+00, 2.90613095648208e+00, 2.90460466971674e+00, 2.90351659916271e+00, 2.90277011366844e+00, 2.90228118979557e+00, 2.90197881886025e+00, 2.90180498876738e+00, 2.90171423471267e+00, 2.90167276864548e+00, 2.90165721332607e+00, 2.90165298190524e+00, 2.90165235735650e+00, 2.90165231027279e+00, 2.90165147890013e+00, 2.90164584623034e+00, 2.90162513997495e+00, 2.90156994496621e+00, 2.90144915338838e+00, 2.90121782823522e+00, 2.90081556802929e+00, 2.90016545203235e+00, 2.89917363482146e+00, 2.89772964663469e+00, 2.89570744054711e+00, 2.89296720853044e+00, 2.88935796513023e+00, 2.88472086955748e+00, 2.87889322467422e+00, 2.87171305559399e+00, 2.86302413315151e+00, 2.85268127084328e+00, 2.84055569118134e+00, 2.82654023230565e+00, 2.81055415174698e+00, 2.79254728453159e+00, 2.77250332952152e+00, 2.75044207172063e+00, 2.72642039824441e+00, 2.70053202889222e+00, 2.67290595419678e+00, 2.64370364855598e+00, 2.61311519698729e+00, 2.58135453469734e+00, 2.54865404348315e+00, 2.51525877412907e+00, 2.48142056781113e+00, 2.44739233288156e+00, 2.41342269935578e+00, 2.37975122678233e+00, 2.34660428772986e+00, 2.31419169475620e+00, 2.28270408858895e+00, 2.25231106319120e+00, 2.22315997163054e+00, 2.19537533583259e+00, 2.16905877266160e+00, 2.14428934668599e+00, 2.12112426436049e+00, 2.09959983303592e+00, 2.07973261927982e+00, 2.06152075293034e+00, 2.04494533501008e+00, 2.02997191836393e+00, 2.01655203920293e+00, 2.00462478536307e+00, 1.99411839286101e+00, 1.98495186615116e+00, 1.97703661931916e+00, 1.97027813531302e+00, 1.96457763833486e+00, 1.95983377092737e+00, 1.95594426244966e+00, 1.95280757002740e+00, 1.95032446726130e+00, 1.94839955062345e+00, 1.94694262920149e+00, 1.94586996085334e+00, 1.94510529737346e+00, 1.94458070324247e+00, 1.94423711702207e+00, 1.94402463132642e+00, 1.94390247619734e+00, 1.94383870109484e+00, 1.94380956192553e+00, 1.94379863084977e+00, 1.94379565733519e+00, 1.94379521845079e+00, 1.94379518518395e+00, 1.94379459539210e+00, 1.94379059946758e+00, 1.94377591005741e+00, 1.94373675372157e+00, 1.94365106224838e+00, 1.94348695746267e+00, 1.94320159237344e+00, 1.94274040548093e+00, 1.94203683794927e+00, 1.94101255469237e+00, 1.93957819961877e+00, 1.93763470170290e+00, 1.93507513167753e+00, 1.93178708872055e+00, 1.92765557272305e+00, 1.92256627131537e+00, 1.91640916317837e+00, 1.90908231229375e+00, 1.90049570421200e+00, 1.89057495793081e+00, 1.87926473828398e+00, 1.86653169603969e+00, 1.85236677746061e+00, 1.83678677187656e+00, 1.81983500339948e+00, 1.80158111841406e+00, 1.78211996997262e+00, 1.76156964923604e+00, 1.74006875828829e+00, 1.71777305444454e+00, 1.69485162128389e+00, 1.67148273528221e+00, 1.64784959971178e+00, 1.62413611104207e+00, 1.60052280952222e+00, 1.57718314697132e+00, 1.55428018261508e+00, 1.53196379308011e+00, 1.51036845597105e+00, 1.48961163832266e+00, 1.46979279244602e+00, 1.45099293360376e+00, 1.43327474838323e+00, 1.41668316163478e+00, 1.40124627524020e+00, 1.38697658494571e+00, 1.37387238225723e+00, 1.36191925616314e+00, 1.35109162259907e+00, 1.34135422599542e+00, 1.33266357474847e+00, 1.32496928908185e+00, 1.31821535403314e+00, 1.31234128128876e+00, 1.30728319090282e+00, 1.30297482761309e+00, 1.29934852686280e+00, 1.29633614330382e+00, 1.29386995015842e+00, 1.29188351206608e+00, 1.29031252765762e+00, 1.28909563177745e+00, 1.28817514166058e+00, 1.28749772702541e+00, 1.28701498140963e+00, 1.28668387144749e+00, 1.28646704227595e+00, 1.28633296078618e+00, 1.28625588374403e+00, 1.28621564446755e+00, 1.28619725924040e+00, 1.28619036237662e+00, 1.28618848626903e+00, 1.28618820935971e+00, 1.28618818823719e+00, 1.28618781199621e+00, 1.28618526291172e+00, 1.28617589225951e+00, 1.28615091394082e+00, 1.28609625152934e+00, 1.28599157424355e+00, 1.28580956388634e+00, 1.28551545206718e+00, 1.28506686368004e+00, 1.28441399814100e+00, 1.28350017363800e+00, 1.28226275092172e+00, 1.28063444136884e+00, 1.27854498881678e+00, 1.27592319601727e+00, 1.27269924500045e+00, 1.26880723727033e+00, 1.26418785623470e+00, 1.25879103278807e+00, 1.25257847804201e+00, 1.24552593747274e+00, 1.23762502067200e+00, 1.22888447234446e+00, 1.21933077419975e+00, 1.20900800372691e+00, 1.19797692282813e+00, 1.18631332363091e+00, 1.17410571564275e+00, 1.16145249167715e+00, 1.14845875300855e+00, 1.13523300078548e+00, 1.12188390628134e+00, 1.10851735538337e+00, 1.09523392472083e+00, 1.08212689354811e+00, 1.06928083513696e+00, 1.05677077318447e+00, 1.04466184076774e+00, 1.03300934713221e+00, 1.02185914299558e+00, 1.01124817658940e+00, 1.00120514640794e+00, 9.91751177558996e-01, 9.82900471852389e-01, 9.74660903512401e-01, 9.67034550282150e-01, 9.60018162700720e-01, 9.53603582542435e-01, 9.47778125528437e-01, 9.42524944453254e-01, 9.37823387808919e-01, 9.33649366683596e-01, 9.29975739778882e-01, 9.26772723239921e-01, 9.24008328854875e-01, 9.21648831167417e-01, 9.19659261196841e-01, 9.18003921789641e-01, 9.16646917151148e-01, 9.15552686861105e-01, 9.14686532732745e-01, 9.14015125322017e-01, 9.13506975844044e-01, 9.13132858822152e-01, 9.12866171079557e-01, 9.12683213749958e-01, 9.12563385846468e-01, 9.12489280542012e-01, 9.12446678566489e-01, 9.12424436848397e-01, 9.12414274504137e-01, 9.12410462270977e-01, 9.12409425252349e-01, 9.12409272190622e-01, 9.12409260432739e-01, 9.12409049917677e-01, 9.12407623648015e-01, 9.12402380542950e-01, 9.12388404450361e-01, 9.12357818623363e-01, 9.12299245058421e-01, 9.12197391244312e-01, 9.12032784779761e-01, 9.11781673824400e-01, 9.11416108257058e-01, 9.10904212465176e-01, 9.10210655599592e-01, 9.09297318659981e-01, 9.08124149811037e-01, 9.06650189950518e-01, 9.04834740121615e-01, 9.02638631569097e-01, 9.00025549085925e-01, 8.96963350026567e-01, 8.93425316300934e-01, 8.89391275982419e-01, 8.84848535636259e-01, 8.79792574218534e-01, 8.74227463751744e-01, 8.68165999509252e-01, 8.61629541127868e-01, 8.54647583680801e-01, 8.47257092264471e-01, 8.39501643677139e-01, 8.31430423752792e-01, 8.23097129205088e-01, 8.14558819468701e-01, 8.05874758384533e-01, 7.97105279014218e-01, 7.88310698427821e-01, 7.79550303568044e-01, 7.70881424409147e-01, 7.62358606460673e-01, 7.54032890967707e-01, 7.45951207724175e-01, 7.38155882162282e-01, 7.30684255362824e-01, 7.23568412983276e-01, 7.16835016979211e-01, 7.10505232510101e-01, 7.04594741608403e-01, 6.99113835001860e-01, 6.94067573797014e-01, 6.89456013393972e-01, 6.85274482834359e-01, 6.81513913615312e-01, 6.78161212694476e-01, 6.75199674858540e-01, 6.72609429769210e-01, 6.70367918810776e-01, 6.68450396354054e-01, 6.66830449264864e-01, 6.65480527491233e-01, 6.64372477453831e-01, 6.63478068850033e-01, 6.62769504482888e-01, 6.62219901967099e-01, 6.61803735762120e-01, 6.61497228037213e-01, 6.61278677458846e-01, 6.61128716143182e-01, 6.61030486729523e-01, 6.60969733754385e-01, 6.60934806146415e-01, 6.60916570590025e-01, 6.60908238568631e-01, 6.60905112932461e-01, 6.60904262682948e-01, 6.60904137187877e-01, 6.60904127474041e-01, 6.60903952598076e-01, 6.60902767787626e-01, 6.60898412298273e-01, 6.60886802156860e-01, 6.60861393616081e-01, 6.60812733104862e-01, 6.60728111704046e-01, 6.60591340044654e-01, 6.60382657045247e-01, 6.60078783010377e-01, 6.59653124208020e-01, 6.59076132030862e-01, 6.58315815146743e-01, 6.57338397639037e-01, 6.56109110094071e-01, 6.54593094096320e-01, 6.52756393956700e-01, 6.50567003170330e-01, 6.47995927623427e-01, 6.45018223545627e-01, 6.41613966211260e-01, 6.37769105906344e-01, 6.33476170976536e-01, 6.28734783876560e-01, 6.23551964773561e-01, 6.17942207845132e-01, 6.11927327140094e-01, 6.05536080771877e-01, 5.98803593283264e-01, 5.91770605345033e-01, 5.84482586814033e-01, 5.76988753154708e-01, 5.69341026229345e-01, 5.61592978718238e-01, 5.53798797447911e-01, 5.46012295364150e-01, 5.38285995535475e-01, 5.30670304110913e-01, 5.23212783153561e-01, 5.15957529108142e-01, 5.08944658525230e-01, 5.02209899567902e-01, 4.95784285659212e-01, 4.89693946212924e-01, 4.83959988536803e-01, 4.78598464543133e-01, 4.73620415719388e-01, 4.69031989814113e-01, 4.64834622818728e-01, 4.61025280023705e-01, 4.57596750152995e-01, 4.54537986784956e-01, 4.51834491400511e-01, 4.49468732410686e-01, 4.47420594362150e-01, 4.45667851176204e-01, 4.44186656740499e-01, 4.42952045471728e-01, 4.41938434658510e-01, 4.41120119560985e-01, 4.40471751490314e-01, 4.39968788529036e-01, 4.39587908293120e-01, 4.39307372271294e-01, 4.39107331879235e-01, 4.38970067467814e-01, 4.38880153122710e-01, 4.38824542140445e-01, 4.38792570478185e-01, 4.38775878136880e-01, 4.38768251210661e-01, 4.38765390077997e-01, 4.38764611779661e-01, 4.38764496904426e-01, 4.38764487940517e-01, 4.38764325634097e-01, 4.38763225984061e-01, 4.38759183551923e-01, 4.38748407894728e-01, 4.38724825544632e-01, 4.38679662186260e-01, 4.38601121673315e-01, 4.38474176494339e-01, 4.38280481942634e-01, 4.37998423516738e-01, 4.37603303906471e-01, 4.37067672229801e-01, 4.36361793926558e-01, 4.35454254879573e-01, 4.34312687990776e-01, 4.32904604749119e-01, 4.31198308554769e-01, 4.29163861073484e-01, 4.26774068132019e-01, 4.24005448107820e-01, 4.20839143875504e-01, 4.17261739519489e-01, 4.13265945423838e-01, 4.08851120018355e-01, 4.04023603175027e-01, 3.98796844575166e-01, 3.93191319701785e-01, 3.87234235761255e-01, 3.80959039120903e-01, 3.74404744171557e-01, 3.67615110443609e-01, 3.60637700055193e-01, 3.53522851037411e-01, 3.46322603771675e-01, 3.39089617772551e-01, 3.31876114474163e-01, 3.24732878676528e-01, 3.17708347044139e-01, 3.10847806733342e-01, 3.04192721125609e-01, 2.97780193102277e-01, 2.91642569715670e-01, 2.85807185917655e-01, 2.80296239599458e-01, 2.75126785896035e-01, 2.70310835714540e-01, 2.65855541827946e-01, 2.61763455565358e-01, 2.58032837953778e-01, 2.54658010860680e-01, 2.51629735942529e-01, 2.48935611701725e-01, 2.46560481395377e-01, 2.44486846675507e-01, 2.42695283492289e-01, 2.41164857850222e-01, 2.39873539440616e-01, 2.38798611014857e-01, 2.37917070707702e-01, 2.37206023507099e-01, 2.36643056871145e-01, 2.36206594299261e-01, 2.35876219658806e-01, 2.35632964412225e-01, 2.35459549711514e-01, 2.35340575704591e-01, 2.35262651357388e-01, 2.35214459603377e-01, 2.35186754606630e-01, 2.35172290239795e-01, 2.35165681382696e-01, 2.35163202177354e-01, 2.35162527773866e-01, 2.35162428233345e-01, 2.35162420399952e-01, 2.35162277716370e-01, 2.35161311014586e-01, 2.35157757324281e-01, 2.35148284549588e-01, 2.35127553913095e-01, 2.35087853410917e-01, 2.35018817494520e-01, 2.34907246745529e-01, 2.34737039605115e-01, 2.34489246153949e-01, 2.34142251375943e-01, 2.33672092140161e-01, 2.33052908185421e-01, 2.32257522637283e-01, 2.31258142064775e-01, 2.30027159943567e-01, 2.28538040910777e-01, 2.26766256771960e-01, 2.24690239353926e-01, 2.22292310550451e-01, 2.19559546854526e-01, 2.16484534829867e-01, 2.13065975743611e-01, 2.09309102184164e-01, 2.05225876916877e-01, 2.00834954229904e-01, 1.96161396076157e-01, 1.91236148663386e-01, 1.86095298825322e-01, 1.80779142443328e-01, 1.75331108291530e-01, 1.69796588968389e-01, 1.64221735297893e-01, 1.58652271283131e-01, 1.53132383296885e-01, 1.47703730022144e-01, 1.42404609441328e-01, 1.37269306980535e-01, 1.32327635991246e-01, 1.27604669324555e-01, 1.23120649851373e-01, 1.18891059140125e-01, 1.14926817544578e-01, 1.11234585812255e-01, 1.07817137853779e-01, 1.04673776118304e-01, 1.01800764512724e-01, 9.91917583337880e-02, 9.68382156574279e-02, 9.47297795832114e-02, 9.28546253279237e-02, 9.11997701586387e-02, 8.97513473842224e-02, 8.84948479858210e-02, 8.74153349427900e-02, 8.64976359516092e-02, 8.57265201439787e-02, 8.50868637130588e-02, 8.45638081910920e-02, 8.41429136305010e-02, 8.38103072735947e-02, 8.35528266101875e-02, 8.33581541725152e-02, 8.32149401462985e-02, 8.31129080053543e-02, 8.30429379861286e-02, 8.29971233472975e-02, 8.29687949962477e-02, 8.29525111502036e-02, 8.29440101366856e-02, 8.29401260953446e-02, 8.29386690803204e-02, 8.29382727390637e-02, 8.29382142400092e-02, 8.29382095956461e-02, 8.29381244816929e-02, 8.29375478234340e-02, 8.29354279861789e-02, 8.29297774280349e-02, 8.29174120792047e-02, 8.28937338672446e-02, 8.28525663653892e-02, 8.27860531908158e-02, 8.26846283055573e-02, 8.25370666049352e-02, 8.23306218986204e-02, 8.20512573680077e-02, 8.16839706190322e-02, 8.12132114118640e-02, 8.06233850339098e-02, 7.98994282443104e-02, 7.90274381027132e-02, 7.79953273364953e-02, 7.67934739014962e-02, 7.54153278512888e-02, 7.38579363599356e-02, 7.21223484361317e-02, 7.02138649788964e-02, 6.81421074632667e-02, 6.59208893954073e-02, 6.35678879892367e-02, 6.11041281341684e-02, 5.85533051840343e-02, 5.59409857884461e-02, 5.32937353921525e-02, 5.06382259844216e-02, 4.80003776379121e-02, 4.54045824856593e-02, 4.28730508156292e-02, 4.04253070989995e-02, 3.80778503891293e-02, 3.58439800882948e-02, 3.37337760090648e-02, 3.17542121810988e-02, 2.99093777218715e-02, 2.82007753811584e-02, 2.66276685762463e-02, 2.51874500911740e-02, 2.38760094350425e-02, 2.26880806267502e-02, 2.16175574144089e-02, 2.06577680836517e-02, 1.98017064876836e-02, 1.90422193890072e-02, 1.83721526190723e-02, 1.77844601592696e-02, 1.72722812738613e-02, 1.68289914125666e-02, 1.64482327690081e-02, 1.61239301469109e-02, 1.58502972226771e-02, 1.56218375250177e-02, 1.54333435974578e-02, 1.52798969391839e-02, 1.51568704602596e-02, 1.50599343480167e-02, 1.49850654378270e-02, 1.49285594455293e-02, 1.48870447934542e-02, 1.48574962905244e-02, 1.48372466435163e-02, 1.48239937001051e-02, 1.48158014554056e-02, 1.48110931730940e-02, 1.48086354444956e-02, 1.48075125836797e-02, 1.48070913753031e-02, 1.48069767978711e-02, 1.48069598865419e-02, 1.48069585316222e-02, 1.48069335464243e-02, 1.48067642689043e-02, 1.48061419993653e-02, 1.48044833536574e-02, 1.48008539374302e-02, 1.47939050459414e-02, 1.47818267305908e-02, 1.47623207026086e-02, 1.47325966464606e-02, 1.46893954839342e-02, 1.46290428293483e-02, 1.45475351625564e-02, 1.44406599863913e-02, 1.43041493439569e-02, 1.41338635413035e-02, 1.39259988564930e-02, 1.36773096553640e-02, 1.33853320498787e-02, 1.30485935133393e-02, 1.26667912481082e-02, 1.22409220899868e-02, 1.17733486859990e-02, 1.12677906974028e-02, 1.07292356184985e-02, 1.01637708828974e-02, 9.57834640125554e-03, 8.98048357092307e-03, 8.37795220194693e-03, 7.77843998537975e-03, 7.18923961427720e-03, 6.61697628208913e-03, 6.06739323599470e-03, 5.54520606241482e-03, 5.05402864523319e-03, 4.59636672234609e-03, 4.17366981688616e-03, 3.78642935079654e-03, 3.43430943920708e-03, 3.11629647958966e-03, 2.83085400854333e-03, 2.57607089281464e-03, 2.34979415830594e-03, 2.14974206456731e-03, 1.97359683922116e-03, 1.81907837981043e-03, 1.68400016703885e-03, 1.56630798306168e-03, 1.46410230619409e-03, 1.37564674825796e-03, 1.29936652835028e-03, 1.23384140653116e-03, 1.17779636905624e-03, 1.13009148888921e-03, 1.08971097632349e-03, 1.05575110352163e-03, 1.02740718941163e-03, 1.00396047894691e-03, 9.84766008009785e-04, 9.69242315129293e-04, 9.56863374689476e-04, 9.47152672331697e-04, 9.39679079714040e-04, 9.34054110174998e-04, 9.29930167750115e-04, 9.26999454620159e-04, 9.24993238129446e-04, 9.23681193256698e-04, 9.22870548365040e-04, 9.22404786116715e-04, 9.22161696962576e-04, 9.22050645947081e-04, 9.22008989867606e-04, 9.21997658679174e-04, 9.21995986232105e-04, 9.21995850916346e-04, 9.21993338906191e-04, 9.21976319865792e-04, 9.21913758439530e-04, 9.21747010735885e-04, 9.21382179313923e-04, 9.20683835303748e-04, 9.19470509344933e-04, 9.17512402047294e-04, 9.14531811048391e-04, 9.10206792930524e-04, 9.04178547138201e-04, 8.96062895044736e-04, 8.85466002777946e-04, 8.72004146657446e-04, 8.55326853482552e-04, 8.35142206689454e-04, 8.11242577782423e-04, 7.83528636645769e-04, 7.52029330648440e-04, 7.16915656482797e-04, 6.78506421657232e-04, 6.37264624991474e-04, 5.93783414004179e-04, 5.48760882502108e-04, 5.02963754968277e-04, 4.57181965259679e-04, 4.12179405692825e-04, 3.68649341261573e-04, 3.27183325438796e-04, 2.88257546728232e-04, 2.52231635407441e-04, 2.19348325110502e-04, 1.89725550264086e-04, 1.63346449177482e-04, 1.40066332498634e-04, 1.19652204396914e-04, 1.01844691009489e-04, 8.64034442287220e-05, 7.30974613745161e-05, 6.16442419669738e-05, 5.16574571562955e-05, 4.26758678528927e-05, 3.42873344223449e-05, 2.62768508108024e-05, 1.86996118429761e-05, 1.18297000923917e-05, 5.99992488758971e-06, 1.31991289009648e-06, 2.13286050865236e-06, 1.80168756003241e-06, 4.37384159113208e-07, 1.17329045679071e-06, 1.40303478701753e-06, 1.14186769195612e-06, 6.66967352458414e-07, 3.55271367880050e-14, 3.55271367880050e-14, 3.55271367880050e-14, 4.98243807101062e-07, 6.82461767098630e-07, 7.63436609929613e-07, 7.55613520198040e-07, 6.84558258967627e-07, 5.77509915444807e-07, 4.57697062472562e-07, 4.26325641456060e-14, 4.26325641456060e-14, 3.55271367880050e-14, 3.55271367880050e-14, 4.26325641456060e-14, 3.55271367880050e-14, 3.55271367880050e-14, 3.55271367880050e-14, 3.55271367880050e-14],
[ 3.90000000000000e+01, 3.89999999676422e+01, 3.89999989670735e+01, 3.89999921880625e+01, 3.89999672675405e+01, 3.89999008373422e+01, 3.89997554502065e+01, 3.89994769924151e+01, 3.89989926767665e+01, 3.89982096919623e+01, 3.89970145645304e+01, 3.89952732678340e+01, 3.89928320903315e+01, 3.89895192529369e+01, 3.89851472439508e+01, 3.89795158203800e+01, 3.89724156072030e+01, 3.89636322116778e+01, 3.89529507582967e+01, 3.89401607413103e+01, 3.89250610854409e+01, 3.89074653008701e+01, 3.88872066151130e+01, 3.88641429614159e+01, 3.88381617005325e+01, 3.88091839503034e+01, 3.87771683960436e+01, 3.87421144554970e+01, 3.87040646765276e+01, 3.86631062553727e+01, 3.86193715794910e+01, 3.85730377225664e+01, 3.85243248499577e+01, 3.84734935297641e+01, 3.84208409856728e+01, 3.83666963701313e+01, 3.83114151769980e+01, 3.82553729485578e+01, 3.81989584599727e+01, 3.81425665829632e+01, 3.80865910388441e+01, 3.80314172490426e+01, 3.79774154798675e+01, 3.79249344592350e+01, 3.78742956183274e+01, 3.78257880829355e+01, 3.77796645095299e+01, 3.77361378316114e+01, 3.76953789538867e+01, 3.76575154061025e+01, 3.76226309453668e+01, 3.75907660755394e+01, 3.75619194346307e+01, 3.75360499858079e+01, 3.75130799342535e+01, 3.74928982804645e+01, 3.74753649104510e+01, 3.74603151146231e+01, 3.74475644200199e+01, 3.74369136151248e+01, 3.74281538430917e+01, 3.74210716381074e+01, 3.74154537811423e+01, 3.74110918557752e+01, 3.74077863923007e+01, 3.74053504989885e+01, 3.74036128930895e+01, 3.74024202607105e+01, 3.74016388936235e+01, 3.74011555718883e+01, 3.74008776832154e+01, 3.74007325925474e+01, 3.74006662976562e+01, 3.74006414278679e+01, 3.74006346626555e+01, 3.74006336641233e+01, 3.74006335331369e+01, 3.74006304813116e+01, 3.74006098047281e+01, 3.74005337951489e+01, 3.74003311794354e+01, 3.73998877493381e+01, 3.73990384871780e+01, 3.73975614865303e+01, 3.73951739209996e+01, 3.73915302617239e+01, 3.73862228862142e+01, 3.73787851584494e+01, 3.73686969938451e+01, 3.73553928539935e+01, 3.73382720462170e+01, 3.73167111332134e+01, 3.72900781893933e+01, 3.72577485735051e+01, 3.72191218220796e+01, 3.71736392052836e+01, 3.71208014267914e+01, 3.70601858945552e+01, 3.69914629445489e+01, 3.69144103721154e+01, 3.68289256252479e+01, 3.67350350515988e+01, 3.66328996750303e+01, 3.65228171120735e+01, 3.64052194200407e+01, 3.62806668841242e+01, 3.61498379798013e+01, 3.60135159633936e+01, 3.58725727216479e+01, 3.57279506297811e+01, 3.55806432151671e+01, 3.54316754009909e+01, 3.52820840220658e+01, 3.51328991828352e+01, 3.49851268878839e+01, 3.48397332390807e+01, 3.46976303764958e+01, 3.45596642510632e+01, 3.44266042569740e+01, 3.42991347168758e+01, 3.41778481958506e+01, 3.40632406129326e+01, 3.39557081148994e+01, 3.38555456717482e+01, 3.37629473445408e+01, 3.36780081639417e+01, 3.36007275426732e+01, 3.35310141285177e+01, 3.34686919874651e+01, 3.34135079897337e+01, 3.33651402548578e+01, 3.33232074957310e+01, 3.32872790852839e+01, 3.32568856534305e+01, 3.32315300064314e+01, 3.32106981466960e+01, 3.31938701594217e+01, 3.31805307247534e+01, 3.31701790118164e+01, 3.31623377154414e+01, 3.31565610087771e+01, 3.31524412059840e+01, 3.31496139589646e+01, 3.31477618501571e+01, 3.31466162887029e+01, 3.31459576681653e+01, 3.31456137983126e+01, 3.31454566789216e+01, 3.31453977376229e+01, 3.31453817041237e+01, 3.31453793376115e+01, 3.31453791286935e+01, 3.31453750350482e+01, 3.31453472999893e+01, 3.31452453428283e+01, 3.31449735613136e+01, 3.31443787682448e+01, 3.31432396458616e+01, 3.31412586320595e+01, 3.31380565979834e+01, 3.31331706136479e+01, 3.31260550263202e+01, 3.31160859944425e+01, 3.31025695268307e+01, 3.30847529721377e+01, 3.30618397875221e+01, 3.30330072899854e+01, 3.29974269627767e+01, 3.29542867585494e+01, 3.29028147185714e+01, 3.28423031226548e+01, 3.27721323075416e+01, 3.26917932513388e+01, 3.26009080250006e+01, 3.24992472618904e+01, 3.23867438917826e+01, 3.22635025208049e+01, 3.21298040054498e+01, 3.19861049579606e+01, 3.18330321246801e+01, 3.16713717933676e+01, 3.15020546061450e+01, 3.13261363750529e+01, 3.11447757032632e+01, 3.09592093831710e+01, 3.07707266426091e+01, 3.05806433150336e+01, 3.03902769073679e+01, 3.02009233443556e+01, 3.00138359197394e+01, 2.98302067343619e+01, 2.96511506962755e+01, 2.94776920249261e+01, 2.93107531418591e+01, 2.91511458258305e+01, 2.89995645339682e+01, 2.88565818186890e+01, 2.87226457873929e+01, 2.85980795534643e+01, 2.84830826149941e+01, 2.83777340775076e+01, 2.82819976145752e+01, 2.81957280395125e+01, 2.81186793441715e+01, 2.80505140469097e+01, 2.79908136798405e+01, 2.79390902337284e+01, 2.78947983659669e+01, 2.78573481622358e+01, 2.78261182258112e+01, 2.78004688510774e+01, 2.77797550212657e+01, 2.77633389570067e+01, 2.77506019343313e+01, 2.77409550906009e+01, 2.77338489464529e+01, 2.77287813925626e+01, 2.77253039224100e+01, 2.77230259359636e+01, 2.77216169930095e+01, 2.77208069566829e+01, 2.77203840348447e+01, 2.77201907961090e+01, 2.77201183052230e+01, 2.77200985859100e+01, 2.77200956753795e+01, 2.77200954419189e+01, 2.77200911333873e+01, 2.77200619424431e+01, 2.77199546333117e+01, 2.77196685855485e+01, 2.77190425718520e+01, 2.77178436609471e+01, 2.77157586829055e+01, 2.77123886378269e+01, 2.77072463635920e+01, 2.76997577036311e+01, 2.76892663300435e+01, 2.76750422796696e+01, 2.76562941500386e+01, 2.76321847787725e+01, 2.76018500959334e+01, 2.75644206979610e+01, 2.75190455505630e+01, 2.74649170946570e+01, 2.74012969143168e+01, 2.73275410393856e+01, 2.72431239080777e+01, 2.71476600144929e+01, 2.70409223171018e+01, 2.69228565873074e+01, 2.67935910281499e+01, 2.66534406843296e+01, 2.65029063854942e+01, 2.63426682031566e+01, 2.61735736449722e+01, 2.59966210452887e+01, 2.58129388243142e+01, 2.56237614659690e+01, 2.54304031928836e+01, 2.52342303846245e+01, 2.50366337850737e+01, 2.48390014768223e+01, 2.46426934723575e+01, 2.44490185991547e+01, 2.42592141591648e+01, 2.40744286448084e+01, 2.38957076133583e+01, 2.37239826740578e+01, 2.35600634350141e+01, 2.34046321904442e+01, 2.32582410983248e+01, 2.31213115956275e+01, 2.29941358136913e+01, 2.28768797811360e+01, 2.27695882289999e+01, 2.26721908375874e+01, 2.25845097838486e+01, 2.25062684605331e+01, 2.24371012434250e+01, 2.23765641808453e+01, 2.23241464708451e+01, 2.22792825769396e+01, 2.22413648139630e+01, 2.22097562131426e+01, 2.21838034516348e+01, 2.21628496087735e+01, 2.21462464916163e+01, 2.21333662586276e+01, 2.21236120649686e+01, 2.21164274579645e+01, 2.21113042683756e+01, 2.21077887727891e+01, 2.21054859445113e+01, 2.21040616635251e+01, 2.21032428182303e+01, 2.21028152998918e+01, 2.21026199615651e+01, 2.21025466831452e+01, 2.21025267496129e+01, 2.21025238074643e+01, 2.21025235805375e+01, 2.21025195056602e+01, 2.21024918977615e+01, 2.21023904081182e+01, 2.21021198733222e+01, 2.21015278111442e+01, 2.21003939282365e+01, 2.20984220556349e+01, 2.20952348759387e+01, 2.20903717457633e+01, 2.20832898474955e+01, 2.20733688239913e+01, 2.20599189570087e+01, 2.20421928437079e+01, 2.20194004054689e+01, 2.19907269312572e+01, 2.19553537177449e+01, 2.19124807268108e+01, 2.18613505467772e+01, 2.18012728276769e+01, 2.17316482748444e+01, 2.16519912407546e+01, 2.15619499617643e+01, 2.14613235499813e+01, 2.13500749713048e+01, 2.12283394130007e+01, 2.10964276560922e+01, 2.09548243024987e+01, 2.08041809446981e+01, 2.06453045875224e+01, 2.04791418218360e+01, 2.03067593985604e+01, 2.01293219559279e+01, 1.99480677160324e+01, 1.97642829949075e+01, 1.95792763694363e+01, 1.93943533171623e+01, 1.92107920902719e+01, 1.90298214991607e+01, 1.88526011619642e+01, 1.86802046274843e+01, 1.85136056107289e+01, 1.83536674097402e+01, 1.82011354187766e+01, 1.80566325325392e+01, 1.79206571582259e+01, 1.77935835173366e+01, 1.76756639205454e+01, 1.75670327255201e+01, 1.74677117273958e+01, 1.73776167744941e+01, 1.72965654406269e+01, 1.72242856157591e+01, 1.71604248971546e+01, 1.71045606732192e+01, 1.70562107927580e+01, 1.70148447044808e+01, 1.69798949367762e+01, 1.69507687678241e+01, 1.69268599131047e+01, 1.69075600337265e+01, 1.68922698473899e+01, 1.68804096070168e+01, 1.68714287027574e+01, 1.68648141435409e+01, 1.68600976862403e+01, 1.68568614047481e+01, 1.68547415276983e+01, 1.68534304211268e+01, 1.68526766489883e+01, 1.68522831073440e+01, 1.68521032938529e+01, 1.68520358394325e+01, 1.68520174901784e+01, 1.68520147818663e+01, 1.68520145768011e+01, 1.68520109440911e+01, 1.68519863319437e+01, 1.68518958550430e+01, 1.68516546765120e+01, 1.68511268615103e+01, 1.68501160261788e+01, 1.68483581548399e+01, 1.68455169131356e+01, 1.68411817323425e+01, 1.68348688792794e+01, 1.68260256553728e+01, 1.68140377858465e+01, 1.67982399644027e+01, 1.67779294099576e+01, 1.67523821714250e+01, 1.67208717877386e+01, 1.66826897791513e+01, 1.66371673205783e+01, 1.65836973384506e+01, 1.65217561903954e+01, 1.64509240430473e+01, 1.63709030666804e+01, 1.62815326220402e+01, 1.61828007257667e+01, 1.60748512412896e+01, 1.59579864411743e+01, 1.58326648086862e+01, 1.56994941719890e+01, 1.55592204751286e+01, 1.54127126703053e+01, 1.52609443560975e+01, 1.51049728831165e+01, 1.49459167044890e+01, 1.47849317691809e+01, 1.46231877472780e+01, 1.44618448417384e+01, 1.43020318820846e+01, 1.41448263118025e+01, 1.39912365733073e+01, 1.38421872652694e+01, 1.36985073036017e+01, 1.35609211694830e+01, 1.34300431871675e+01, 1.33063746523213e+01, 1.31903035371149e+01, 1.30821064363838e+01, 1.29819523906555e+01, 1.28899082237217e+01, 1.28059450587481e+01, 1.27299457201208e+01, 1.26617127803291e+01, 1.26009770647764e+01, 1.25474064763624e+01, 1.25006150414561e+01, 1.24601721066886e+01, 1.24256116306905e+01, 1.23964415167665e+01, 1.23721529230770e+01, 1.23522294686502e+01, 1.23361562296560e+01, 1.23234283944145e+01, 1.23135594213086e+01, 1.23060885246614e+01, 1.23005873028077e+01, 1.22966653223966e+01, 1.22939744847989e+01, 1.22922120246864e+01, 1.22911220265934e+01, 1.22904953907207e+01, 1.22901682317137e+01, 1.22900187503792e+01, 1.22899626748213e+01, 1.22899474209148e+01, 1.22899451694704e+01, 1.22899450004537e+01, 1.22899420255693e+01, 1.22899218702944e+01, 1.22898477774489e+01, 1.22896502737165e+01, 1.22892180446261e+01, 1.22883902856972e+01, 1.22869508419384e+01, 1.22846244147722e+01, 1.22810750767558e+01, 1.22759072885081e+01, 1.22686695568234e+01, 1.22588608057616e+01, 1.22459394517140e+01, 1.22293350783150e+01, 1.22084624984782e+01, 1.21827378717756e+01, 1.21515964211368e+01, 1.21145111710075e+01, 1.20710120191642e+01, 1.20207043670312e+01, 1.19632864794935e+01, 1.18985647347319e+01, 1.18264659651015e+01, 1.17470461855159e+01, 1.16604951555857e+01, 1.15671364200039e+01, 1.14674227070743e+01, 1.13619268216105e+01, 1.12513284256384e+01, 1.11363973364559e+01, 1.10179741650842e+01, 1.08969492506578e+01, 1.07742409052829e+01, 1.06507739646967e+01, 1.05274595468467e+01, 1.04051767659395e+01, 1.02847569528398e+01, 1.01669707169118e+01, 1.00525179726167e+01, 9.94202086648241e+00, 9.83601939085327e+00, 9.73496936773357e+00, 9.63924242985132e+00, 9.54912761202556e+00, 9.46483418547498e+00, 9.38649541060154e+00, 9.31417293982371e+00, 9.24786166238142e+00, 9.18749484090561e+00, 9.13294944036411e+00, 9.08405159118031e+00, 9.04058215883899e+00, 9.00228241218758e+00, 8.96885979260501e+00, 8.93999378729752e+00, 8.91534190342776e+00, 8.89454572695191e+00, 8.87723703241698e+00, 8.86304388919653e+00, 8.85159668752648e+00, 8.84253398620086e+00, 8.83550806492969e+00, 8.83019005013833e+00, 8.82627447518089e+00, 8.82348313595722e+00, 8.82156811161968e+00, 8.82031383761650e+00, 8.81953814418497e+00, 8.81909220629863e+00, 8.81885938908878e+00, 8.81875301356605e+00, 8.81871310852935e+00, 8.81870225340072e+00, 8.81870065120680e+00, 8.81870053129534e+00, 8.81869842559021e+00, 8.81868415913396e+00, 8.81863171421880e+00, 8.81849191600009e+00, 8.81818597441690e+00, 8.81760007270889e+00, 8.81658122547191e+00, 8.81493460684646e+00, 8.81242252312002e+00, 8.80876517224976e+00, 8.80364329359107e+00, 8.79670276234526e+00, 8.78756112317867e+00, 8.77581598529161e+00, 8.76105511758561e+00, 8.74286799009381e+00, 8.72085841134922e+00, 8.69465781794368e+00, 8.66393869108779e+00, 8.62842751543989e+00, 8.58791666752977e+00, 8.54227463273581e+00, 8.49145400566200e+00, 8.43549682895376e+00, 8.37453696472232e+00, 8.30879936006781e+00, 8.23859624857772e+00, 8.16432050556171e+00, 8.08643652873198e+00, 8.00546913351804e+00, 7.92199102393246e+00, 7.83660942323995e+00, 7.74995242769673e+00, 7.66265559078644e+00, 7.57534916719217e+00, 7.48864635794999e+00, 7.40313281129551e+00, 7.31935755427282e+00, 7.23782546073214e+00, 7.15899130121546e+00, 7.08325536721868e+00, 7.01096061448611e+00, 6.94239122688106e+00, 6.87777246533660e+00, 6.81727163799971e+00, 6.76100001063392e+00, 6.70901547229943e+00, 6.66132578016915e+00, 6.61789222705143e+00, 6.57863360230439e+00, 6.54343034718353e+00, 6.51212883521647e+00, 6.48454573363319e+00, 6.46047242094767e+00, 6.43967944736585e+00, 6.42192102866754e+00, 6.40693956123825e+00, 6.39447013723315e+00, 6.38424502601836e+00, 6.37599807284040e+00, 6.36946894999432e+00, 6.36440718144762e+00, 6.36057585065757e+00, 6.35775489468730e+00, 6.35574388682051e+00, 6.35436421538189e+00, 6.35346057854392e+00, 6.35290173312584e+00, 6.35258045882065e+00, 6.35241272651221e+00, 6.35233608865354e+00, 6.35230733921460e+00, 6.35229951867652e+00, 6.35229836438173e+00, 6.35229827795963e+00, 6.35229675991214e+00, 6.35228647492127e+00, 6.35224866626974e+00, 6.35214788276917e+00, 6.35192732299397e+00, 6.35150493402963e+00, 6.35077042546495e+00, 6.34958334325971e+00, 6.34777232918917e+00, 6.34513566875244e+00, 6.34144320234221e+00, 6.33643963964526e+00, 6.32984927446547e+00, 6.32138204585622e+00, 6.31074083202682e+00, 6.29762979765979e+00, 6.28176354621588e+00, 6.26287676116345e+00, 6.24073395972940e+00, 6.21513893638262e+00, 6.18594344756426e+00, 6.15305469014634e+00, 6.11644115805490e+00, 6.07613652625918e+00, 6.03224130755343e+00, 5.98492215029894e+00, 5.93440878594618e+00, 5.88098878183383e+00, 5.82500039322885e+00, 5.76682392382889e+00, 5.70687208243905e+00, 5.64557985569363e+00, 5.58339439928919e+00, 5.52076538763384e+00, 5.45813616575612e+00, 5.39593593434149e+00, 5.33457308667267e+00, 5.27442972008181e+00, 5.21585727360906e+00, 5.15917320049889e+00, 5.10465856574520e+00, 5.05255645838191e+00, 5.00307111794116e+00, 4.95636768793564e+00, 4.91257252205099e+00, 4.87177397897541e+00, 4.83402364924794e+00, 4.79933796296394e+00, 4.76770013167359e+00, 4.73906238213402e+00, 4.71334844404482e+00, 4.69045625834923e+00, 4.67026087662275e+00, 4.65261752485067e+00, 4.63736480591004e+00, 4.62432801389242e+00, 4.61332252988498e+00, 4.60415726313483e+00, 4.59663809413599e+00, 4.59057126784910e+00, 4.58576667693739e+00, 4.58204096763247e+00, 4.57922039568265e+00, 4.57714335772844e+00, 4.57566252512768e+00, 4.57464651314956e+00, 4.57398102865361e+00, 4.57356945356850e+00, 4.57333283903217e+00, 4.57320930499792e+00, 4.57315286129508e+00, 4.57313168731635e+00, 4.57312592748102e+00, 4.57312507734139e+00, 4.57312501352255e+00, 4.57312389025736e+00, 4.57311627997178e+00, 4.57308830377590e+00, 4.57301372965502e+00, 4.57285052675560e+00, 4.57253797651491e+00, 4.57199445789752e+00, 4.57111601113556e+00, 4.56977576934010e+00, 4.56782432956927e+00, 4.56509111360777e+00, 4.56138674300010e+00, 4.55650642212288e+00, 4.55023428704160e+00, 4.54234863698501e+00, 4.53262792075094e+00, 4.52085730445206e+00, 4.50683560286862e+00, 4.49038231817437e+00, 4.47134450116999e+00, 4.44960313543049e+00, 4.42507874718735e+00, 4.39773596515770e+00, 4.36758679491344e+00, 4.33469242976398e+00, 4.29916349067986e+00, 4.26115866631205e+00, 4.22088180475264e+00, 4.17857758543574e+00, 4.13452596720667e+00, 4.08903566283365e+00, 4.04243692805728e+00, 3.99507397289194e+00, 3.94729730377725e+00, 3.89945628803410e+00, 3.85189219885029e+00, 3.80493195283775e+00, 3.75888269721023e+00, 3.71402734462892e+00, 3.67062109571269e+00, 3.62888893663416e+00, 3.58902405569392e+00, 3.55118709053136e+00, 3.51550609748312e+00, 3.48207712598912e+00, 3.45096528226186e+00, 3.42220617540406e+00, 3.39580765325108e+00, 3.37175175196641e+00, 3.34999680069257e+00, 3.33047963865705e+00, 3.31311791583864e+00, 3.29781245885212e+00, 3.28444969073220e+00, 3.27290409673792e+00, 3.26304072835018e+00, 3.25471773469844e+00, 3.24778890529213e+00, 3.24210620084841e+00, 3.23752224100805e+00, 3.23389270969545e+00, 3.23107863170655e+00, 3.22894846866882e+00, 3.22737997957070e+00, 3.22626179119429e+00, 3.22549462735628e+00, 3.22499215293884e+00, 3.22468139904479e+00, 3.22450274872343e+00, 3.22440947782555e+00, 3.22436686173494e+00, 3.22435087499282e+00, 3.22434652621368e+00, 3.22434588434307e+00, 3.22434583595350e+00, 3.22434498152268e+00, 3.22433919263017e+00, 3.22431791207821e+00, 3.22426118616504e+00, 3.22413704403410e+00, 3.22389930138033e+00, 3.22348587861144e+00, 3.22281771429112e+00, 3.22179834380812e+00, 3.22031420017264e+00, 3.21823567790284e+00, 3.21541898136089e+00, 3.21170875499789e+00, 3.20694146448594e+00, 3.20094946491322e+00, 3.19356565606729e+00, 3.18462858711373e+00, 3.17398783627910e+00, 3.16150945872464e+00, 3.14708127129386e+00, 3.13061772986115e+00, 3.11206415670257e+00, 3.09140009369870e+00, 3.06864159278273e+00, 3.04384230652279e+00, 3.01709330579740e+00, 2.98852162316527e+00, 2.95828759349338e+00, 2.92658113095743e+00, 2.89361713735978e+00, 2.85963027583369e+00, 2.82486936347534e+00, 2.78959163576265e+00, 2.75405711671356e+00, 2.71852329553384e+00, 2.68324026818231e+00, 2.64844645634629e+00, 2.61436497172449e+00, 2.58120065400574e+00, 2.54913777871280e+00, 2.51833840684266e+00, 2.48894133146732e+00, 2.46106156592823e+00, 2.43479031249159e+00, 2.41019534801026e+00, 2.38732176330272e+00, 2.36619299504336e+00, 2.34681209268493e+00, 2.32916316814467e+00, 2.31321298247917e+00, 2.29891263119412e+00, 2.28619929765457e+00, 2.27499805161182e+00, 2.26522367643978e+00, 2.25678251363784e+00, 2.24957431603992e+00, 2.24349410173964e+00, 2.23843399903959e+00, 2.23428506906886e+00, 2.23093908762845e+00, 2.22829026203183e+00, 2.22623685301615e+00, 2.22468266702415e+00, 2.22353838104180e+00, 2.22272266130784e+00, 2.22216303896956e+00, 2.22179651026497e+00, 2.22156983593012e+00, 2.22143952387557e+00, 2.22137149016174e+00, 2.22134040520665e+00, 2.22132874420090e+00, 2.22132557212782e+00, 2.22132510393659e+00, 2.22132506844834e+00, 2.22132443927271e+00, 2.22132017651748e+00, 2.22130450621097e+00, 2.22126273517806e+00, 2.22117132156720e+00, 2.22099625844525e+00, 2.22069183747796e+00, 2.22019985327906e+00, 2.21944930142810e+00, 2.21835661396479e+00, 2.21682646475162e+00, 2.21475316282835e+00, 2.21202263424120e+00, 2.20851497158088e+00, 2.20410750579163e+00, 2.19867832745523e+00, 2.19211015605012e+00, 2.18429442756930e+00, 2.17513544574611e+00, 2.16455442262389e+00, 2.15249322287700e+00, 2.13891762527621e+00, 2.12381992534019e+00, 2.10722072583334e+00, 2.08916979551505e+00, 2.06974591950413e+00, 2.04905571406784e+00, 2.02723143140684e+00, 2.00442783284742e+00, 1.98081825872244e+00, 1.95659006735456e+00, 1.93193965134721e+00, 1.90706726425137e+00, 1.88217190196842e+00, 1.85744647854327e+00, 1.83307351368300e+00, 1.80922150948086e+00, 1.78604213895151e+00, 1.76366830433469e+00, 1.74221305613477e+00, 1.72176930283368e+00, 1.70241019355828e+00, 1.68419002660579e+00, 1.66714552710700e+00, 1.65129734530648e+00, 1.63665164850335e+00, 1.62320170897789e+00, 1.61092942172015e+00, 1.59980671507514e+00, 1.58979684168228e+00, 1.58085555501787e+00, 1.57293218838636e+00, 1.56597065908170e+00, 1.55991042177196e+00, 1.55468739313815e+00, 1.55023486549386e+00, 1.54648442140206e+00, 1.54336685486228e+00, 1.54081309797726e+00, 1.53875514552377e+00, 1.53712696389151e+00, 1.53586536572898e+00, 1.53491082764642e+00, 1.53420822575102e+00, 1.53370746285492e+00, 1.53336396205458e+00, 1.53313900406729e+00, 1.53299989012822e+00, 1.53291991816057e+00, 1.53287816695902e+00, 1.53285909080359e+00, 1.53285193472981e+00, 1.53284998810901e+00, 1.53284970079204e+00, 1.53284967887561e+00, 1.53284928849356e+00, 1.53284664360032e+00, 1.53283692073314e+00, 1.53281100344743e+00, 1.53275428566648e+00, 1.53264567032635e+00, 1.53245680609493e+00, 1.53215160161875e+00, 1.53168605272506e+00, 1.53100841280959e+00, 1.53005972966101e+00, 1.52877476255447e+00, 1.52708328101410e+00, 1.52491173087058e+00, 1.52218523420341e+00, 1.51882986810002e+00, 1.51477514418946e+00, 1.50995658859988e+00, 1.50431830290689e+00, 1.49781537367244e+00, 1.49041599415886e+00, 1.48210316901717e+00, 1.47287589238086e+00, 1.46274972144612e+00, 1.45175670901431e+00, 1.43994470548145e+00, 1.42737608782662e+00, 1.41412601414348e+00, 1.40028033160073e+00, 1.38593327958686e+00, 1.37118512701490e+00, 1.35613986512571e+00, 1.34090304897178e+00, 1.32557984790130e+00, 1.31027333368316e+00, 1.29508300907283e+00, 1.28010356221074e+00, 1.26542382362148e+00, 1.25112590128783e+00, 1.23728447282362e+00, 1.22396621949435e+00, 1.21122939253041e+00, 1.19912350645051e+00, 1.18768915641345e+00, 1.17695795706793e+00, 1.16695259949500e+00, 1.15768702128956e+00, 1.14916668319547e+00, 1.14138894439501e+00, 1.13434352775497e+00, 1.12801306604363e+00, 1.12237372022295e+00, 1.11739586117601e+00, 1.11304480642598e+00, 1.10928160334341e+00, 1.10606384989051e+00, 1.10334654305316e+00, 1.10108294379295e+00, 1.09922544570968e+00, 1.09772643280809e+00, 1.09653911002122e+00, 1.09561828869388e+00, 1.09492110830666e+00, 1.09440767553485e+00, 1.09404160244560e+00, 1.09379042733428e+00, 1.09362590439506e+00, 1.09352415203789e+00, 1.09346565404458e+00, 1.09343511267739e+00, 1.09342115804187e+00, 1.09341592317104e+00, 1.09341449915875e+00, 1.09341428897751e+00, 1.09341427283182e+00, 1.09341398375679e+00, 1.09341202523112e+00, 1.09340482548640e+00, 1.09338563363163e+00, 1.09334363283499e+00, 1.09326319669891e+00, 1.09312331887853e+00, 1.09289724260810e+00, 1.09255231285079e+00, 1.09205006913789e+00, 1.09134659161995e+00, 1.09039310622087e+00, 1.08913684690667e+00, 1.08752216390678e+00, 1.08549185636727e+00, 1.08298869670489e+00, 1.07995710243339e+00, 1.07634490025409e+00, 1.07210511771565e+00, 1.06719773080938e+00, 1.06159129246290e+00, 1.05526436778896e+00, 1.04820670753490e+00, 1.04042010138738e+00, 1.03191886705847e+00, 1.02272994842441e+00, 1.01289261512326e+00, 1.00245777552860e+00, 9.91486933528741e-01, 9.80050835842754e-01, 9.68227869730349e-01, 9.56102280205904e-01, 9.43762280841035e-01, 9.31298132806312e-01, 9.18800263129626e-01, 9.06357485696979e-01, 8.94055378030330e-01, 8.81974854312183e-01, 8.70190961582210e-01, 8.58771912618479e-01, 8.47778356727396e-01, 8.37262879279756e-01, 8.27269712838685e-01, 8.17834637321546e-01, 8.08985043746198e-01, 8.00740135440172e-01, 7.93111241710797e-01, 7.86102221389427e-01, 7.79709936871441e-01, 7.73924782817957e-01, 7.68731257170650e-01, 7.64108565248620e-01, 7.60031250219988e-01, 7.56469845026430e-01, 7.53391541817940e-01, 7.50760875128535e-01, 7.48540414454574e-01, 7.46691460702955e-01, 7.45174739323929e-01, 7.43951081034560e-01, 7.42982079103456e-01, 7.42230710443799e-01, 7.41661906478797e-01, 7.41243059105308e-01, 7.40944447240373e-01, 7.40739570485005e-01, 7.40605378394513e-01, 7.40522386644869e-01, 7.40474674891331e-01, 7.40449765131984e-01, 7.40438383674515e-01, 7.40434114099692e-01, 7.40432952672080e-01, 7.40432781247804e-01, 7.40432767978852e-01, 7.40432529101092e-01, 7.40430910669538e-01, 7.40424961151547e-01, 7.40409101978820e-01, 7.40374394861128e-01, 7.40307927704897e-01, 7.40192344287074e-01, 7.40005540219492e-01, 7.39720544675144e-01, 7.39305603683619e-01, 7.38724476344693e-01, 7.37936949846699e-01, 7.36899572544147e-01, 7.35566596473312e-01, 7.33891111638194e-01, 7.31826344450809e-01, 7.29327082340781e-01, 7.26351176458373e-01, 7.22861065458439e-01, 7.18825256545834e-01, 7.14219696258297e-01, 7.09028963690209e-01, 7.03247223572824e-01, 6.96878885976467e-01, 6.89938933057469e-01, 6.82452890407461e-01, 6.74456439898776e-01, 6.65994690860032e-01, 6.57121145234932e-01, 6.47896408430704e-01, 6.38386709508254e-01, 6.28662301325370e-01, 6.18795812906455e-01, 6.08860622938167e-01, 5.98929315637847e-01, 5.89072269422466e-01, 5.79356416106819e-01, 5.69844195064519e-01, 5.60592714029603e-01, 5.51653116882264e-01, 5.43070149439437e-01, 5.34881907254075e-01, 5.27119744756703e-01, 5.19808322590826e-01, 5.12965769407991e-01, 5.06603935329437e-01, 5.00728716351347e-01, 4.95340431781237e-01, 4.90434239985213e-01, 4.86000580988339e-01, 4.82025637544311e-01, 4.78491808973509e-01, 4.75378194211146e-01, 4.72661082014760e-01, 4.70314447107967e-01, 4.68310451187335e-01, 4.66619947242705e-01, 4.65212984627748e-01, 4.64059310899636e-01, 4.63128864786214e-01, 4.62392252920019e-01, 4.61821201399879e-01, 4.61388971990502e-01, 4.61070732023558e-01, 4.60843866951500e-01, 4.60688225113863e-01, 4.60586285624814e-01, 4.60523242338823e-01, 4.60486999491408e-01, 4.60468077685633e-01, 4.60459432208069e-01, 4.60456188999338e-01, 4.60455306768715e-01, 4.60455176553353e-01, 4.60455166392443e-01, 4.60454982412095e-01, 4.60453735918868e-01, 4.60449153688494e-01, 4.60436939225858e-01, 4.60410208585024e-01, 4.60359017788321e-01, 4.60270001502465e-01, 4.60126140780034e-01, 4.59906675488824e-01, 4.59587174378555e-01, 4.59139772382478e-01, 4.58533580512622e-01, 4.57735268412975e-01, 4.56709813198948e-01, 4.55421400685275e-01, 4.53834456717409e-01, 4.51914777506339e-01, 4.49630719256469e-01, 4.46954399769474e-01, 4.43862858993540e-01, 4.40339122542809e-01, 4.36373112767072e-01, 4.31962356465512e-01, 4.27112446880862e-01, 4.21837229805071e-01, 4.16158698644715e-01, 4.10106599926230e-01, 4.03717767519829e-01, 3.97035219318916e-01, 3.90107062862064e-01, 3.82985265371708e-01, 3.75724348292358e-01, 3.68380066515016e-01, 3.61008128419407e-01, 3.53663005382749e-01, 3.46396869481872e-01, 3.39258686857107e-01, 3.32293482688222e-01, 3.25541782891506e-01, 3.19039228208332e-01, 3.12816348791401e-01, 3.06898481931498e-01, 3.01305812206998e-01, 2.96053511914046e-01, 2.91151959854993e-01, 2.86607018074108e-01, 2.82420348555767e-01, 2.78589754880571e-01, 2.75109537047342e-01, 2.71970850841946e-01, 2.69162066053518e-01, 2.66669120343018e-01, 2.64475867553202e-01, 2.62564420646626e-01, 2.60915490249843e-01, 2.59508719982335e-01, 2.58323019404315e-01, 2.57336894609999e-01, 2.56528775326409e-01, 2.55877335983804e-01, 2.55361806748525e-01, 2.54962269108240e-01, 2.54659929427326e-01, 2.54437363084129e-01, 2.54278721474833e-01, 2.54169894391509e-01, 2.54098621080708e-01, 2.54054544635366e-01, 2.54029206191277e-01, 2.54015977569246e-01, 2.54009933377787e-01, 2.54007666005059e-01, 2.54007049225848e-01, 2.54006958190573e-01, 2.54006951026497e-01, 2.54006820534528e-01, 2.54005936433046e-01, 2.54002686394863e-01, 2.53994023081226e-01, 2.53975064113021e-01, 2.53938757266909e-01, 2.53875625133787e-01, 2.53773602332522e-01, 2.53617976999678e-01, 2.53391446534266e-01, 2.53074295283916e-01, 2.52644698824412e-01, 2.52079155544038e-01, 2.51353041308533e-01, 2.50441277080256e-01, 2.49319092679933e-01, 2.47962862775942e-01, 2.46350984203737e-01, 2.44464757545856e-01, 2.42289231308227e-01, 2.39813964767912e-01, 2.37033666265518e-01, 2.33948667753630e-01, 2.30565203824362e-01, 2.26895473861248e-01, 2.22957478630164e-01, 2.18774636454384e-01, 2.14375197826186e-01, 2.09791489581534e-01, 2.05059029449629e-01, 2.00215558038863e-01, 1.95300037705103e-01, 1.90351666286666e-01, 1.85408948821674e-01, 1.80508862845862e-01, 1.75686143655881e-01, 1.70972706016315e-01, 1.66397209117655e-01, 1.61984762914045e-01, 1.57756766817890e-01, 1.53730866395222e-01, 1.49921010264535e-01, 1.46337587742941e-01, 1.42987627673904e-01, 1.39875039998792e-01, 1.37000883664257e-01, 1.34363647057782e-01, 1.31959530041208e-01, 1.29782719557817e-01, 1.27825653529264e-01, 1.26079270192847e-01, 1.24533242060494e-01, 1.23176195252526e-01, 1.21995916046053e-01, 1.20979547082428e-01, 1.20113775823881e-01, 1.19385017580811e-01, 1.18779594808636e-01, 1.18283913475530e-01, 1.17884636220992e-01, 1.17568850864160e-01, 1.17324231689871e-01, 1.17139189949263e-01, 1.17003009261687e-01, 1.16905961177501e-01, 1.16839396115807e-01, 1.16795805248039e-01, 1.16768849646775e-01, 1.16753354107956e-01, 1.16745264404344e-01, 1.16741568235092e-01, 1.16740181687852e-01, 1.16739804514651e-01, 1.16739748844729e-01, 1.16739744424962e-01, 1.16739663427232e-01, 1.16739114656490e-01, 1.16737097329576e-01, 1.16731719971696e-01, 1.16719952237631e-01, 1.16697417466597e-01, 1.16658235091720e-01, 1.16594921477144e-01, 1.16498357074974e-01, 1.16357826859392e-01, 1.16161139651673e-01, 1.15894830030783e-01, 1.15544443889469e-01, 1.15094905257649e-01, 1.14530957786805e-01, 1.13837669407445e-01, 1.13000983421976e-01, 1.12008294109636e-01, 1.10849020350635e-01, 1.09515147441435e-01, 1.08001705776971e-01, 1.06307155907082e-01, 1.04433652916370e-01, 1.02387169113371e-01, 1.00177462304302e-01, 9.78178867983530e-02, 9.53250548256435e-02, 9.27183661893523e-02, 9.00194326765202e-02, 8.72514301422882e-02, 8.44384146717729e-02, 8.16046395922925e-02, 7.87739075129323e-02, 7.59689864929811e-02, 7.32111126041843e-02, 7.05195933818317e-02, 6.79115187804555e-02, 6.54015789732227e-02, 6.30019822018681e-02, 6.07224612244721e-02, 5.85703538456954e-02, 5.65507414934956e-02, 5.46666296409342e-02, 5.29191547982961e-02, 5.13078045206754e-02, 4.98306390984453e-02, 4.84845060548338e-02, 4.72652410479029e-02, 4.61678510935002e-02, 4.51866780811230e-02, 4.43155422756902e-02, 4.35478668579705e-02, 4.28767855471648e-02, 4.22952359840565e-02, 4.17960418563581e-02, 4.13719867526865e-02, 4.10158824780424e-02, 4.07206340965445e-02, 4.04793033357791e-02, 4.02851712457775e-02, 4.01318002128548e-02, 4.00130946466177e-02, 3.99233589503964e-02, 3.98573508110616e-02, 3.98103274539707e-02, 3.97780823386924e-02, 3.97569698407878e-02, 3.97439157677226e-02, 3.97364120696082e-02, 3.97324947813473e-02, 3.97307050127012e-02, 3.97300336204367e-02, 3.97298509865891e-02, 3.97298240302391e-02, 3.97298218705302e-02, 3.97297820443328e-02, 3.97295122166739e-02, 3.97285203118898e-02, 3.97258763365116e-02, 3.97200904600155e-02, 3.97090113429712e-02, 3.96897494484918e-02, 3.96586299456132e-02, 3.96111796292686e-02, 3.95421519482113e-02, 3.94455936359037e-02, 3.93149554554313e-02, 3.91432480899709e-02, 3.89232421667174e-02, 3.86477087940378e-02, 3.83096939107404e-02, 3.79028163919060e-02, 3.74215765341575e-02, 3.68616586494923e-02, 3.62202094735551e-02, 3.54960733652447e-02, 3.46899661777158e-02, 3.38045723820670e-02, 3.28445544726677e-02, 3.18164695683052e-02, 3.07285949086307e-02, 2.95906709203351e-02, 2.84135769109000e-02, 2.72089595079237e-02, 2.59888371292689e-02, 2.47652047239981e-02, 2.35496617439210e-02, 2.23530830521881e-02, 2.11853477475046e-02, 2.00551353125675e-02, 1.89697927519177e-02, 1.79352710629246e-02, 1.69561249426948e-02, 1.60355663597613e-02, 1.51755606280730e-02, 1.43769528750681e-02, 1.36396131323409e-02, 1.29625894419547e-02, 1.23442600703072e-02, 1.17824778716766e-02, 1.12747018232469e-02, 1.08181126089164e-02, 1.04097107663463e-02, 1.00463972688729e-02, 9.72503745737185e-03, 9.44250996272444e-03, 9.19574269612156e-03, 8.98173818728054e-03, 8.79759058258145e-03, 8.64049652462739e-03, 8.50776195219538e-03, 8.39680660085662e-03, 8.30516766669120e-03, 8.23050373514889e-03, 8.17059969558187e-03, 8.12337297819710e-03, 8.08688108598687e-03, 8.05933006668624e-03, 8.03908329795178e-03, 8.02466975811456e-03, 8.01479083930445e-03, 8.00832473709079e-03, 8.00432752073732e-03, 8.00203014519241e-03, 8.00083088996928e-03, 8.00028298185396e-03, 8.00007744933140e-03, 8.00002154013413e-03, 8.00001328806843e-03, 8.00001262070538e-03, 8.00000023625103e-03, 7.99991633011388e-03, 7.99960788798160e-03, 7.99878573825907e-03, 7.99698670560645e-03, 7.99354219324044e-03, 7.98755481760338e-03, 7.97788476631212e-03, 7.96314758051864e-03, 7.94172503234591e-03, 7.91179061109659e-03, 7.87135078495993e-03, 7.81830260620353e-03, 7.75050734046090e-03, 7.66587862164414e-03, 7.56248221382094e-03, 7.43864290893725e-03, 7.29305257650026e-03, 7.12487212526014e-03, 6.93381937745841e-03, 6.72023481493511e-03, 6.48511798701890e-03, 6.23012912674170e-03, 5.95755311384494e-03, 5.67022611862456e-03, 5.37142870053486e-03, 5.06475237617110e-03, 4.75394925648231e-03, 4.44277586807118e-03, 4.13484247763307e-03, 3.83347815658652e-03, 3.54161975450040e-03, 3.26173028920351e-03, 2.99574943150560e-03, 2.74507606491881e-03, 2.51058033744300e-03, 2.29264028204312e-03, 2.09119665716884e-03, 1.90581947146740e-03, 1.73578047863288e-03, 1.58012795611739e-03, 1.43776142012086e-03, 1.30750266114177e-03, 1.18816081064921e-03, 1.07859511034292e-03, 9.77766706434124e-04, 8.84696881357172e-04, 7.98453568734203e-04, 7.18282268884707e-04, 6.43376923292749e-04, 5.73859362553719e-04, 5.09078148496656e-04, 4.48794255987650e-04, 3.92846123283164e-04, 3.41128351855957e-04, 2.93571084768018e-04, 2.50121289639083e-04, 2.10726792083449e-04, 1.75323522284998e-04, 1.43826093022881e-04, 1.16168683469198e-04, 9.21104598958777e-05, 7.15227954799502e-05, 5.41996311014259e-05, 3.99061571769721e-05, 2.83819707433963e-05, 1.93452329853017e-05, 1.24975086990275e-05, 7.52910241175186e-06, 4.12481250577912e-06, 1.97012803937469e-06, 7.58023062985558e-07, 0.00000000000000e+00, -7.10542735760100e-15],
[ 4.00000000000000e+01, 3.99999999663870e+01, 3.99999989270080e+01, 3.99999918850505e+01, 3.99999659979027e+01, 3.99998969909846e+01, 3.99997459645043e+01, 3.99994567057121e+01, 3.99989536039614e+01, 3.99981402475765e+01, 3.99968987607866e+01, 3.99950899165664e+01, 3.99925540379752e+01, 3.99891126774189e+01, 3.99845710410843e+01, 3.99787211054392e+01, 3.99713453548364e+01, 3.99622210543369e+01, 3.99511249600278e+01, 3.99378383601758e+01, 3.99221523340605e+01, 3.99038731105978e+01, 3.98828274051501e+01, 3.98588676096189e+01, 3.98318767077172e+01, 3.98017727844141e+01, 3.97685129966651e+01, 3.97320968729589e+01, 3.96925688135450e+01, 3.96500196731633e+01, 3.96045873250932e+01, 3.95564561301727e+01, 3.95058552669679e+01, 3.94530559183277e+01, 3.93983673529410e+01, 3.93421319852502e+01, 3.92847195398153e+01, 3.92265204836284e+01, 3.91679389191578e+01, 3.91093851500511e+01, 3.90512681395379e+01, 3.89939880787868e+01, 3.89379292698863e+01, 3.88834535075521e+01, 3.88308941173195e+01, 3.87805507781633e+01, 3.87326852263283e+01, 3.86875179064173e+01, 3.86452256067659e+01, 3.86059400896507e+01, 3.85697477033005e+01, 3.85366899420303e+01, 3.85067649028902e+01, 3.84799295716657e+01, 3.84561028575374e+01, 3.84351692838831e+01, 3.84169832324051e+01, 3.84013736289101e+01, 3.83881489517498e+01, 3.83771024383398e+01, 3.83680173616030e+01, 3.83606722469805e+01, 3.83548459021440e+01, 3.83503221360427e+01, 3.83468940515973e+01, 3.83443678073136e+01, 3.83425657572182e+01, 3.83413288955854e+01, 3.83405185525106e+01, 3.83400173079397e+01, 3.83397291146089e+01, 3.83395786437467e+01, 3.83395098905468e+01, 3.83394840985535e+01, 3.83394770824781e+01, 3.83394760469191e+01, 3.83394759110756e+01, 3.83394727460846e+01, 3.83394513027869e+01, 3.83393724746812e+01, 3.83391623457827e+01, 3.83387024731182e+01, 3.83378217209181e+01, 3.83362899570857e+01, 3.83338138750235e+01, 3.83300351497275e+01, 3.83245310770923e+01, 3.83168177804781e+01, 3.83063559998261e+01, 3.82925594071321e+01, 3.82748053192058e+01, 3.82524476055897e+01, 3.82248315173171e+01, 3.81913100914984e+01, 3.81512617178712e+01, 3.81041083867659e+01, 3.80493340744589e+01, 3.79865026642693e+01, 3.79152747551311e+01, 3.78354226818043e+01, 3.77468430727618e+01, 3.76495663142056e+01, 3.75437623805720e+01, 3.74297426368375e+01, 3.73079574112669e+01, 3.71789893647884e+01, 3.70435429223341e+01, 3.69024302547189e+01, 3.67565544797640e+01, 3.66068908672241e+01, 3.64544668729690e+01, 3.63003417956097e+01, 3.61455867566514e+01, 3.59912655744650e+01, 3.58384169568717e+01, 3.56880382985673e+01, 3.55410712535176e+01, 3.53983891664734e+01, 3.52607863920216e+01, 3.51289694986900e+01, 3.50035503413059e+01, 3.48850409786542e+01, 3.47738504087606e+01, 3.46702830867510e+01, 3.45745391786875e+01, 3.44867164894477e+01, 3.44068139849940e+01, 3.43347368107918e+01, 3.42703026898299e+01, 3.42132495661861e+01, 3.41632443433522e+01, 3.41198925502738e+01, 3.40827487519699e+01, 3.40513275055987e+01, 3.40251146472941e+01, 3.40035786807866e+01, 3.39861820269593e+01, 3.39723918855286e+01, 3.39616904575660e+01, 3.39535842820684e+01, 3.39476124524300e+01, 3.39433535002361e+01, 3.39404307644356e+01, 3.39385161032134e+01, 3.39373318526340e+01, 3.39366509886779e+01, 3.39362955054622e+01, 3.39361330797627e+01, 3.39360721478737e+01, 3.39360555728845e+01, 3.39360531264495e+01, 3.39360529104758e+01, 3.39360486785784e+01, 3.39360200068400e+01, 3.39359146063437e+01, 3.39356336461408e+01, 3.39350187656336e+01, 3.39338411730058e+01, 3.39317932578840e+01, 3.39284830901586e+01, 3.39234321119621e+01, 3.39160762555525e+01, 3.39057706352031e+01, 3.38917978650607e+01, 3.38733799466616e+01, 3.38496935497836e+01, 3.38198883805082e+01, 3.37831081947115e+01, 3.37385138798304e+01, 3.36853079008410e+01, 3.36227592976281e+01, 3.35502283404762e+01, 3.34671899075728e+01, 3.33732546500953e+01, 3.32681870599093e+01, 3.31519196511420e+01, 3.30245626049732e+01, 3.28864083995341e+01, 3.27379311460211e+01, 3.25797805715574e+01, 3.24127708239567e+01, 3.22378645175280e+01, 3.20561526813358e+01, 3.18688314914884e+01, 3.16771768369020e+01, 3.14825178499758e+01, 3.12862105050823e+01, 3.10896122465138e+01, 3.08940583801703e+01, 3.07008406995175e+01, 3.05111885720609e+01, 3.03262525303832e+01, 3.01470903085273e+01, 2.99746552310279e+01, 2.98097868727617e+01, 2.96532039344332e+01, 2.95054992993404e+01, 2.93671372415908e+01, 2.92384527435825e+01, 2.91196528567036e+01, 2.90108200109541e+01, 2.89119171524489e+01, 2.88227945658448e+01, 2.87431982223679e+01, 2.86727794821191e+01, 2.86111059695748e+01, 2.85576734314424e+01, 2.85119183745831e+01, 2.84732312677505e+01, 2.84409700745786e+01, 2.84144738677063e+01, 2.83930762569890e+01, 2.83761183507905e+01, 2.83629609609318e+01, 2.83529957614472e+01, 2.83456551209450e+01, 2.83404203495181e+01, 2.83368281343904e+01, 2.83344749834949e+01, 2.83330195516678e+01, 2.83321827879454e+01, 2.83317459117340e+01, 2.83315462970696e+01, 2.83314714143429e+01, 2.83314510443903e+01, 2.83314480378267e+01, 2.83314477966631e+01, 2.83314433459713e+01, 2.83314131918708e+01, 2.83313023420726e+01, 2.83310068561712e+01, 2.83303601872377e+01, 2.83291217186901e+01, 2.83269679483488e+01, 2.83234867133362e+01, 2.83181747821961e+01, 2.83104390634477e+01, 2.82996015914398e+01, 2.82849083494893e+01, 2.82655418760291e+01, 2.82406374721406e+01, 2.82093026903580e+01, 2.81706396391055e+01, 2.81237694910195e+01, 2.80678584453550e+01, 2.80021442750267e+01, 2.79259624985845e+01, 2.78387711669912e+01, 2.77401732528198e+01, 2.76299356804593e+01, 2.75080041410075e+01, 2.73745129912474e+01, 2.72297897350951e+01, 2.70743538177560e+01, 2.69089097150837e+01, 2.67343345594341e+01, 2.65516607937235e+01, 2.63620545715898e+01, 2.61667908073501e+01, 2.59672259096463e+01, 2.57647692953144e+01, 2.55608547689939e+01, 2.53569127712367e+01, 2.51543443540868e+01, 2.49544975565614e+01, 2.47586466460217e+01, 2.45679744884041e+01, 2.43835581304868e+01, 2.42063575341872e+01, 2.40372073020833e+01, 2.38768111736714e+01, 2.37257390468451e+01, 2.35844262795187e+01, 2.34531750425452e+01, 2.33321575186666e+01, 2.32214207668295e+01, 2.31208930926715e+01, 2.30303917820615e+01, 2.29496320642916e+01, 2.28782371746405e+01, 2.28157493827483e+01, 2.27616418439306e+01, 2.27153311158414e+01, 2.26761901636533e+01, 2.26435616544841e+01, 2.26167713179081e+01, 2.25951411262638e+01, 2.25780020286279e+01, 2.25647059584418e+01, 2.25546368294198e+01, 2.25472202397296e+01, 2.25419316220791e+01, 2.25383026080036e+01, 2.25359254180502e+01, 2.25344551444468e+01, 2.25336098569749e+01, 2.25331685331502e+01, 2.25329668869172e+01, 2.25328912421734e+01, 2.25328706649425e+01, 2.25328676277854e+01, 2.25328673935306e+01, 2.25328631870664e+01, 2.25328346876464e+01, 2.25327299206729e+01, 2.25324506496865e+01, 2.25318394684108e+01, 2.25306689695363e+01, 2.25286334195603e+01, 2.25253433151981e+01, 2.25203231346050e+01, 2.25130125248169e+01, 2.25027710838403e+01, 2.24888868000395e+01, 2.24705881017220e+01, 2.24470593461118e+01, 2.24174594409839e+01, 2.23809431481822e+01, 2.23366844726622e+01, 2.22839014027367e+01, 2.22218811479457e+01, 2.21500049325616e+01, 2.20677713568505e+01, 2.19748173443937e+01, 2.18709357576420e+01, 2.17560888856651e+01, 2.16304171817822e+01, 2.14942428424475e+01, 2.13480680557444e+01, 2.11925679892086e+01, 2.10285788142845e+01, 2.08570812640443e+01, 2.06791803830020e+01, 2.04960822499969e+01, 2.03090685385504e+01, 2.01194698265497e+01, 1.99286385796937e+01, 1.97379227087604e+01, 1.95486405351377e+01, 1.93620578890873e+01, 1.91793679132682e+01, 1.90016739609199e+01, 1.88299757826584e+01, 1.86651590109940e+01, 1.85079877984884e+01, 1.83591003576016e+01, 1.82190070913939e+01, 1.80880909887492e+01, 1.79666099741706e+01, 1.78547009371772e+01, 1.77523852081591e+01, 1.76595752877567e+01, 1.75760826704658e+01, 1.75016266280201e+01, 1.74358438337131e+01, 1.73782987155588e+01, 1.73284944247730e+01, 1.72858842973418e+01, 1.72498836714888e+01, 1.72198819040472e+01, 1.71952544059046e+01, 1.71753744931067e+01, 1.71596248285261e+01, 1.71474082120624e+01, 1.71381574679161e+01, 1.71313441779833e+01, 1.71264860226390e+01, 1.71231525150561e+01, 1.71209689526652e+01, 1.71196184582836e+01, 1.71188420417396e+01, 1.71184366775803e+01, 1.71182514622629e+01, 1.71181819814274e+01, 1.71181630809389e+01, 1.71181602912658e+01, 1.71181600800402e+01, 1.71181563381991e+01, 1.71181309866724e+01, 1.71180377917349e+01, 1.71177893679139e+01, 1.71172456967624e+01, 1.71162044949699e+01, 1.71143938159844e+01, 1.71114672227761e+01, 1.71070018153958e+01, 1.71004993343793e+01, 1.70913904881190e+01, 1.70790425672949e+01, 1.70627703110482e+01, 1.70418498775923e+01, 1.70155356479004e+01, 1.69830794586399e+01, 1.69437517257510e+01, 1.68968637914502e+01, 1.68417907153716e+01, 1.67779936464168e+01, 1.67050408667752e+01, 1.66226266026850e+01, 1.65305867534740e+01, 1.64289108018835e+01, 1.63177493294449e+01, 1.61974167598299e+01, 1.60683891751682e+01, 1.59312972775575e+01, 1.57869147831418e+01, 1.56361427251704e+01, 1.54799902963138e+01, 1.53195529753769e+01, 1.51559887595712e+01, 1.49904933626665e+01, 1.48242752434233e+01, 1.46585312982614e+01, 1.44944239868935e+01, 1.43330605601742e+01, 1.41754749288276e+01, 1.40226125570448e+01, 1.38753185970966e+01, 1.37343293135647e+01, 1.36002666924568e+01, 1.34736360033179e+01, 1.33548259896807e+01, 1.32441113082958e+01, 1.31416568193186e+01, 1.30475233429366e+01, 1.29616745351540e+01, 1.28839845878681e+01, 1.28142465173363e+01, 1.27521808630978e+01, 1.26974446704374e+01, 1.26496406694710e+01, 1.26083265906191e+01, 1.25730245689627e+01, 1.25432305895013e+01, 1.25184239134921e+01, 1.24980764055610e+01, 1.24816616554297e+01, 1.24686637605577e+01, 1.24585856104685e+01, 1.24509564935718e+01, 1.24453388359657e+01, 1.24413338813542e+01, 1.24385861332327e+01, 1.24367864051850e+01, 1.24356733616619e+01, 1.24350334781013e+01, 1.24346994029396e+01, 1.24345467616326e+01, 1.24344895006723e+01, 1.24344739243088e+01, 1.24344716252704e+01, 1.24344714526808e+01, 1.24344684149094e+01, 1.24344478335669e+01, 1.24343721744598e+01, 1.24341704957138e+01, 1.24337291300352e+01, 1.24328838747007e+01, 1.24314140084658e+01, 1.24290384210026e+01, 1.24254141011796e+01, 1.24201371837323e+01, 1.24127466973027e+01, 1.24027310879997e+01, 1.23895375095983e+01, 1.23725837734212e+01, 1.23512727387180e+01, 1.23250088010704e+01, 1.22932160075790e+01, 1.22553572013264e+01, 1.22109534839265e+01, 1.21596031952327e+01, 1.21009995550815e+01, 1.20349461037936e+01, 1.19613691239836e+01, 1.18803263300517e+01, 1.17920112724594e+01, 1.16967531146339e+01, 1.15950116882703e+01, 1.14873679998105e+01, 1.13745106249937e+01, 1.12572186660111e+01, 1.11363421345080e+01, 1.10127807452307e+01, 1.08874621482070e+01, 1.07613205895848e+01, 1.06352768796938e+01, 1.05102203771859e+01, 1.03869934919756e+01, 1.02663789915734e+01, 1.01490901886173e+01, 1.00357639109191e+01, 9.92695602155783e+00, 9.82313917048307e+00, 9.72470241901770e+00, 9.63195237766416e+00, 9.54511552584912e+00, 9.46434142891556e+00, 9.38970662288098e+00, 9.32121899325363e+00, 9.25882252493513e+00, 9.20240234262473e+00, 9.15178999371084e+00, 9.10676894822153e+00, 9.06708030359850e+00, 9.03242868681593e+00, 9.00248834376717e+00, 8.97690939699773e+00, 8.95532423892417e+00, 8.93735400989651e+00, 8.92261509025658e+00, 8.91072551453465e+00, 8.90131119591537e+00, 8.89401183200070e+00, 8.88848635059000e+00, 8.88441774838686e+00, 8.88151717756084e+00, 8.87952714573631e+00, 8.87822371438378e+00, 8.87741760815733e+00, 8.87695418215593e+00, 8.87671223347788e+00, 8.87660168546163e+00, 8.87656021513806e+00, 8.87654893420962e+00, 8.87654726916840e+00, 8.87654714455333e+00, 8.87654495625023e+00, 8.87653013017897e+00, 8.87647562803686e+00, 8.87633034582913e+00, 8.87601240177089e+00, 8.87540351044114e+00, 8.87434467327375e+00, 8.87263339132618e+00, 8.87002257794039e+00, 8.86622131502890e+00, 8.86089755544148e+00, 8.85368282325134e+00, 8.84417890225217e+00, 8.83196642978873e+00, 8.81661522899912e+00, 8.79769612027344e+00, 8.77479385705429e+00, 8.74752073900332e+00, 8.71553037576790e+00, 8.67853101676149e+00, 8.63629783586024e+00, 8.58868357225451e+00, 8.53562698406788e+00, 8.47715866935033e+00, 8.41340394396059e+00, 8.34458262685716e+00, 8.27100575564036e+00, 8.19306942213395e+00, 8.11124606334889e+00, 8.02607365463838e+00, 7.93814332192247e+00, 7.84808591756740e+00, 7.75655809479511e+00, 7.66422837747674e+00, 7.57176366649433e+00, 7.47981656005690e+00, 7.38901379980194e+00, 7.29994608972238e+00, 7.21315947034786e+00, 7.12914836432211e+00, 7.04835034056919e+00, 6.97114257408344e+00, 6.89783991125690e+00, 6.82869439266437e+00, 6.76389604246716e+00, 6.70357471029480e+00, 6.64780274869855e+00, 6.59659832476608e+00, 6.54992919337150e+00, 6.50771679565985e+00, 6.46984058367197e+00, 6.43614250555905e+00, 6.40643161229784e+00, 6.38048876453939e+00, 6.35807142695024e+00, 6.33891853787452e+00, 6.32275543572426e+00, 6.30929881188503e+00, 6.29826164493925e+00, 6.28935805452784e+00, 6.28230799702381e+00, 6.27684171113319e+00, 6.27270381116312e+00, 6.26965692036806e+00, 6.26748473752509e+00, 6.26599443728277e+00, 6.26501831894863e+00, 6.26441463874633e+00, 6.26406758617711e+00, 6.26388639450376e+00, 6.26380360677016e+00, 6.26377255027172e+00, 6.26376410215476e+00, 6.26376285523052e+00, 6.26376276187325e+00, 6.26376112200600e+00, 6.26375001166723e+00, 6.26370916893384e+00, 6.26360029756443e+00, 6.26336203707449e+00, 6.26290574677957e+00, 6.26211227617750e+00, 6.26082987592565e+00, 6.25887338243671e+00, 6.25602478375732e+00, 6.25203524356758e+00, 6.24662862242079e+00, 6.23950648956961e+00, 6.23035456436557e+00, 6.21885046375461e+00, 6.20467256362551e+00, 6.18750971002198e+00, 6.16707144635394e+00, 6.14309836089328e+00, 6.11537211200458e+00, 6.08372466384583e+00, 6.04804626898100e+00, 6.00829177096102e+00, 5.96448487109733e+00, 5.91672010732385e+00, 5.86516242294030e+00, 5.81004434864533e+00, 5.75166096859686e+00, 5.69036297436831e+00, 5.62654821421081e+00, 5.56065220700395e+00, 5.49313810482468e+00, 5.42448655707712e+00, 5.35518586206635e+00, 5.28572270358666e+00, 5.21657367703796e+00, 5.14819772600628e+00, 5.08102954514691e+00, 5.01547396131259e+00, 4.95190127953615e+00, 4.89064356792807e+00, 4.83199184938911e+00, 4.77619416322152e+00, 4.72345445349935e+00, 4.67393223295350e+00, 4.62774296224709e+00, 4.58495907664389e+00, 4.54561158684823e+00, 4.50969217920899e+00, 4.47715574267305e+00, 4.44792325520671e+00, 4.42188496969450e+00, 4.39890384713112e+00, 4.37881919182275e+00, 4.36145044812843e+00, 4.34660112020734e+00, 4.33406277496147e+00, 4.32361908402044e+00, 4.31504985378117e+00, 4.30813498412427e+00, 4.30265828766571e+00, 4.29841109356698e+00, 4.29519555431701e+00, 4.29282757165612e+00, 4.29113925981713e+00, 4.28998087102708e+00, 4.28922211983915e+00, 4.28875285900669e+00, 4.28848307951683e+00, 4.28834222998877e+00, 4.28827787460266e+00, 4.28825373266367e+00, 4.28824716546979e+00, 4.28824619616569e+00, 4.28824612340134e+00, 4.28824484268750e+00, 4.28823616566375e+00, 4.28820426801904e+00, 4.28811924070531e+00, 4.28793316099542e+00, 4.28757679795472e+00, 4.28695708547511e+00, 4.28595548104270e+00, 4.28442731096277e+00, 4.28220218225987e+00, 4.27908551710775e+00, 4.27486123659340e+00, 4.26929558624384e+00, 4.26214205584203e+00, 4.25314730110274e+00, 4.24205792610093e+00, 4.22862793522728e+00, 4.21262661509213e+00, 4.19384656413044e+00, 4.17211155496526e+00, 4.14728389601454e+00, 4.11927095782857e+00, 4.08803054842314e+00, 4.05357486096734e+00, 4.01597277526443e+00, 3.97535036840112e+00, 3.93188957510486e+00, 3.88582502909095e+00, 3.83743920686930e+00, 3.78705607902261e+00, 3.73503354523956e+00, 3.68175498360651e+00, 3.62762027820316e+00, 3.57303669975597e+00, 3.51841000155965e+00, 3.46413605858736e+00, 3.41059332507788e+00, 3.35813631995138e+00, 3.30709027628741e+00, 3.25774701727926e+00, 3.21036205259726e+00, 3.16515283085420e+00, 3.12229803916289e+00, 3.08193781110803e+00, 3.04417468962389e+00, 3.00907518969777e+00, 2.97667181498742e+00, 2.94696539935029e+00, 2.91992766586918e+00, 2.89550391940006e+00, 2.87361581158107e+00, 2.85416413777950e+00, 2.83703164231832e+00, 2.82208582072882e+00, 2.80918171539163e+00, 2.79816470381059e+00, 2.78887327729534e+00, 2.78114180266593e+00, 2.77480325161520e+00, 2.76969187262292e+00, 2.76564576997560e+00, 2.76250934471410e+00, 2.76013554437255e+00, 2.75838786322344e+00, 2.75714203321159e+00, 2.75628734835357e+00, 2.75572757224428e+00, 2.75538138921330e+00, 2.75518237401204e+00, 2.75507847178520e+00, 2.75503099837493e+00, 2.75501318952050e+00, 2.75500834508602e+00, 2.75500763005785e+00, 2.75500757615304e+00, 2.75500662433804e+00, 2.75500017565489e+00, 2.75497646966657e+00, 2.75491327861276e+00, 2.75477498860463e+00, 2.75451015461827e+00, 2.75404963089015e+00, 2.75330536869968e+00, 2.75216996217029e+00, 2.75051700888313e+00, 2.74820233480339e+00, 2.74506611124384e+00, 2.74093586448325e+00, 2.73563034579737e+00, 2.72896419129055e+00, 2.72075325815211e+00, 2.71082047890319e+00, 2.69900203092499e+00, 2.68515357896801e+00, 2.66915631783002e+00, 2.65092252537451e+00, 2.63040033639694e+00, 2.60757746819858e+00, 2.58248367001174e+00, 2.55519172938536e+00, 2.52581694572724e+00, 2.49451506870699e+00, 2.46147878983017e+00, 2.42693296106945e+00, 2.39112878706065e+00, 2.35433729037742e+00, 2.31684237828947e+00, 2.27893384245297e+00, 2.24090060137777e+00, 2.20302445311274e+00, 2.16557454819742e+00, 2.12880272738561e+00, 2.09293980182450e+00, 2.05819279129196e+00, 2.02474308331239e+00, 1.99274543527872e+00, 1.96232771414673e+00, 1.93359125338871e+00, 1.90661170314824e+00, 1.88144025472988e+00, 1.85810513223617e+00, 1.83661325994803e+00, 1.81695203181997e+00, 1.79909112749197e+00, 1.78298433617643e+00, 1.76857136470656e+00, 1.75577961830874e+00, 1.74452595193277e+00, 1.73471839610038e+00, 1.72625786424971e+00, 1.71903984864620e+00, 1.71295610940668e+00, 1.70789635647906e+00, 1.70374991807940e+00, 1.70040738175412e+00, 1.69776218662906e+00, 1.69571213828419e+00, 1.69416081180372e+00, 1.69301880456958e+00, 1.69220479883997e+00, 1.69164639543276e+00, 1.69128068403685e+00, 1.69105452265924e+00, 1.69092450809197e+00, 1.69085663044372e+00, 1.69082561696759e+00, 1.69081398280377e+00, 1.69081081803496e+00, 1.69081035092193e+00, 1.69081031551541e+00, 1.69080968778875e+00, 1.69080543485159e+00, 1.69078980065470e+00, 1.69074812600994e+00, 1.69065692401792e+00, 1.69048226875985e+00, 1.69017856516693e+00, 1.68968776215102e+00, 1.68893906412576e+00, 1.68784918735931e+00, 1.68632319638280e+00, 1.68425594109484e+00, 1.68153409653947e+00, 1.67803878421629e+00, 1.67364872635655e+00, 1.66824385372603e+00, 1.66170925485459e+00, 1.65393932260592e+00, 1.64484192581437e+00, 1.63434241281661e+00, 1.62238724351953e+00, 1.60894705000850e+00, 1.59401894432139e+00, 1.57762792602976e+00, 1.55982728998288e+00, 1.54069799247397e+00, 1.52034699717913e+00, 1.49890468461719e+00, 1.47652146460311e+00, 1.45336377499019e+00, 1.42960967814880e+00, 1.40544427728789e+00, 1.38105516813998e+00, 1.35662811983536e+00, 1.33234314552865e+00, 1.30837108284163e+00, 1.28487076091831e+00, 1.26198678889610e+00, 1.23984796310285e+00, 1.21856625948531e+00, 1.19823635479367e+00, 1.17893560508539e+00, 1.16072440262748e+00, 1.14364683123308e+00, 1.12773154415051e+00, 1.11299279643296e+00, 1.09943157390176e+00, 1.08703677217350e+00, 1.07578639074175e+00, 1.06564871799468e+00, 1.05658349272059e+00, 1.04854303569753e+00, 1.04147335112651e+00, 1.03531520182955e+00, 1.03000516428435e+00, 1.02547666979548e+00, 1.02166103659422e+00, 1.01848849467866e+00, 1.01588920110385e+00, 1.01379423862207e+00, 1.01213658552759e+00, 1.01085203977530e+00, 1.00988007641329e+00, 1.00916461454710e+00, 1.00865466880089e+00, 1.00830486080135e+00, 1.00807576866809e+00, 1.00793409676762e+00, 1.00785265382832e+00, 1.00781013453192e+00, 1.00779070740136e+00, 1.00778341966045e+00, 1.00778143722260e+00, 1.00778114461909e+00, 1.00778112229942e+00, 1.00778072473445e+00, 1.00777803117580e+00, 1.00776812940682e+00, 1.00774173521483e+00, 1.00768397363594e+00, 1.00757335891485e+00, 1.00738101654949e+00, 1.00707018730223e+00, 1.00659604887170e+00, 1.00590588560294e+00, 1.00493962961340e+00, 1.00363078744065e+00, 1.00190775427316e+00, 9.99695502734156e-01, 9.96917615138500e-01, 9.93498607705142e-01, 9.89366473508099e-01, 9.84455349659626e-01, 9.78708195468123e-01, 9.72079354470743e-01, 9.64536866626659e-01, 9.56064399475494e-01, 9.46662679838695e-01, 9.36350330711448e-01, 9.25164050152986e-01, 9.13158107813267e-01, 9.00403176829570e-01, 8.86984560225976e-01, 8.72999907663392e-01, 8.58556546960855e-01, 8.43768572784604e-01, 8.28753841120623e-01, 8.13631012829831e-01, 7.98516774191413e-01, 7.83523339252632e-01, 7.68756310929042e-01, 7.54312948141127e-01, 7.40280857571918e-01, 7.26737103094031e-01, 7.13747705067853e-01, 7.01367486364198e-01, 6.89640212267165e-01, 6.78598966991771e-01, 6.68266709669197e-01, 6.58656956361945e-01, 6.49774540949636e-01, 6.41616415585638e-01, 6.34172459976192e-01, 6.27426277242719e-01, 6.21355962028424e-01, 6.15934833399109e-01, 6.11132130712448e-01, 6.06913674853324e-01, 6.03242500018432e-01, 6.00079462615660e-01, 5.97383833922841e-01, 5.95113882071921e-01, 5.93227446884704e-01, 5.91682508317263e-01, 5.90437746042888e-01, 5.89453084315821e-01, 5.88690213022716e-01, 5.88113073056867e-01, 5.87688292125343e-01, 5.87385556054450e-01, 5.87177900750262e-01, 5.87041911260172e-01, 5.86957816820977e-01, 5.86909474220299e-01, 5.86884235998212e-01, 5.86872704668060e-01, 5.86868378903930e-01, 5.86867202194583e-01, 5.86867028514881e-01, 5.86867015173162e-01, 5.86866776300980e-01, 5.86865157908697e-01, 5.86859208556227e-01, 5.86843349987454e-01, 5.86808645017271e-01, 5.86742185137837e-01, 5.86626624274636e-01, 5.86439883237226e-01, 5.86155047117515e-01, 5.85740474883764e-01, 5.85160136311323e-01, 5.84374185768205e-01, 5.83339774876009e-01, 5.82012096491567e-01, 5.80345640843859e-01, 5.78295631389544e-01, 5.75819593771946e-01, 5.72878997319968e-01, 5.69440896253070e-01, 5.65479488787169e-01, 5.60977508248556e-01, 5.55927362392488e-01, 5.50331946140602e-01, 5.44205068886974e-01, 5.37571459503908e-01, 5.30466338479812e-01, 5.22934574842864e-01, 5.15029472885132e-01, 5.06811257425312e-01, 4.98345344063829e-01, 4.89700490963834e-01, 4.80946930466530e-01, 4.72154572659811e-01, 4.63391360093560e-01, 4.54721835048964e-01, 4.46205960335064e-01, 4.37898213733909e-01, 4.29846956922830e-01, 4.22094063475008e-01, 4.14674778305638e-01, 4.07617773015673e-01, 4.00945357805988e-01, 3.94673810439890e-01, 3.88813785357833e-01, 3.83370770656768e-01, 3.78345566414531e-01, 3.73734764036165e-01, 3.69531212334863e-01, 3.65724461518120e-01, 3.62301180876202e-01, 3.59245549651057e-01, 3.56539623275090e-01, 3.54163678947145e-01, 3.52096545427095e-01, 3.50315922074088e-01, 3.48798691634187e-01, 3.47521230217943e-01, 3.46459716421251e-01, 3.45590439760976e-01, 3.44890106658553e-01, 3.44336140254043e-01, 3.43906968524053e-01, 3.43582293660276e-01, 3.43343334584979e-01, 3.43173033952134e-01, 3.43056221088645e-01, 3.42979723100001e-01, 3.42932417771920e-01, 3.42905223861870e-01, 3.42891026757478e-01, 3.42884540112777e-01, 3.42882106769359e-01, 3.42881444842924e-01, 3.42881347144044e-01, 3.42881339581751e-01, 3.42881203439497e-01, 3.42880281056864e-01, 3.42876890298093e-01, 3.42867851913311e-01, 3.42848072305273e-01, 3.42810194579656e-01, 3.42744333000645e-01, 3.42637904930562e-01, 3.42475573055545e-01, 3.42239309004533e-01, 3.41908587118112e-01, 3.41460713864628e-01, 3.40871294013162e-01, 3.40114829030483e-01, 3.39165436296909e-01, 3.37997669868869e-01, 3.36587415127866e-01, 3.34912821459284e-01, 3.32955230013638e-01, 3.30700048628721e-01, 3.28137524109891e-01, 3.25263364049242e-01, 3.22079166603636e-01, 3.18592627034896e-01, 3.14817503646481e-01, 3.10773341801784e-01, 3.06484971346237e-01, 3.01981808185104e-01, 2.97297003304557e-01, 2.92466490861486e-01, 2.87527990370677e-01, 2.82520016418282e-01, 2.77480943295117e-01, 2.72448162559840e-01, 2.67457360212639e-01, 2.62541928359383e-01, 2.57732515279379e-01, 2.53056708638198e-01, 2.48538839745578e-01, 2.44199892396061e-01, 2.40057497784726e-01, 2.36125996902388e-01, 2.32416553196153e-01, 2.28937300605232e-01, 2.25693514851905e-01, 2.22687798693116e-01, 2.19920274459838e-01, 2.17388779498116e-01, 2.15089062015181e-01, 2.13014976315257e-01, 2.11158677482082e-01, 2.09510816256063e-01, 2.08060735198700e-01, 2.06796667285332e-01, 2.05705937860934e-01, 2.04775170467478e-01, 2.03990496442010e-01, 2.03337767422120e-01, 2.02802769030356e-01, 2.02371433099863e-01, 2.02030044914153e-01, 2.01765441148211e-01, 2.01565193583882e-01, 2.01417773304506e-01, 2.01312690009274e-01, 2.01240601359999e-01, 2.01193387897483e-01, 2.01164190020201e-01, 2.01147404756952e-01, 2.01138641515726e-01, 2.01134637559285e-01, 2.01133135543579e-01, 2.01132726959521e-01, 2.01132666653386e-01, 2.01132661947625e-01, 2.01132576741536e-01, 2.01131999458219e-01, 2.01129877310485e-01, 2.01124220495871e-01, 2.01111840959825e-01, 2.01088133680024e-01, 2.01046909689900e-01, 2.00980289265352e-01, 2.00878662750320e-01, 2.00730725307160e-01, 2.00523590332217e-01, 2.00242984258786e-01, 1.99873522874093e-01, 1.99399066050702e-01, 1.98803143961733e-01, 1.98069443544910e-01, 1.97182339460362e-01, 1.96127449428211e-01, 1.94892190113620e-01, 1.93466307167597e-01, 1.91842352120524e-01, 1.90016079928803e-01, 1.87986744249599e-01, 1.85757272853017e-01, 1.83334312586410e-01, 1.80728141332558e-01, 1.77952452655852e-01, 1.75024026437825e-01, 1.71962305003269e-01, 1.68788898454558e-01, 1.65527044917965e-01, 1.62201051246328e-01, 1.58835737771369e-01, 1.55455907482271e-01, 1.52085856033544e-01, 1.48748934716558e-01, 1.45467174335245e-01, 1.42260974110378e-01, 1.39148856529999e-01, 1.36147286526224e-01, 1.33270551447076e-01, 1.30530696839443e-01, 1.27937511998120e-01, 1.25498558536592e-01, 1.23219234963656e-01, 1.21102870397458e-01, 1.19150841030212e-01, 1.17362703621694e-01, 1.15736341103670e-01, 1.14268116184554e-01, 1.12953029796159e-01, 1.11784882105560e-01, 1.10756434700690e-01, 1.09859573192281e-01, 1.09085469954337e-01, 1.08424746829208e-01, 1.07867637598574e-01, 1.07404149706205e-01, 1.07024224283066e-01, 1.06717893002148e-01, 1.06475429663696e-01, 1.06287493849280e-01, 1.06145263503457e-01, 1.06040552953630e-01, 1.05965912785642e-01, 1.05914708097679e-01, 1.05881172097568e-01, 1.05860432596181e-01, 1.05848509853651e-01, 1.05842285226466e-01, 1.05839441170495e-01, 1.05838374270860e-01, 1.05838084049211e-01, 1.05838041213019e-01, 1.05838037842233e-01, 1.05837976439787e-01, 1.05837560431254e-01, 1.05836031143731e-01, 1.05831954662214e-01, 1.05823033572577e-01, 1.05805949347548e-01, 1.05776242017569e-01, 1.05728233198498e-01, 1.05654997870474e-01, 1.05548389452032e-01, 1.05399121628849e-01, 1.05196908981760e-01, 1.04930666649196e-01, 1.04588767035978e-01, 1.04159348946496e-01, 1.03630671542987e-01, 1.02991502381229e-01, 1.02231525729692e-01, 1.01341754596618e-01, 1.00314927872965e-01, 9.91458729062416e-02, 9.78318139976437e-02, 9.63726089032448e-02, 9.47708983522162e-02, 9.30321578675262e-02, 9.11646463646250e-02, 8.91792517847705e-02, 8.70892399780772e-02, 8.49099185699203e-02, 8.26582323892637e-02, 8.03523104733301e-02, 7.80109867163361e-02, 7.56533166061644e-02, 7.32981112331927e-02, 7.09635075617854e-02, 6.86665902803369e-02, 6.64230770919332e-02, 6.42470743544337e-02, 6.21509060294230e-02, 6.01450138849629e-02, 5.82379235997195e-02, 5.64165990664591e-02, 5.47274984575026e-02, 5.31513987260723e-02, 5.16898665171581e-02, 5.03430366455859e-02, 4.91097722053695e-02, 4.79878253529264e-02, 4.69739931495923e-02, 4.60642646707825e-02, 4.52539572092121e-02, 4.45378407341437e-02, 4.39102508197493e-02, 4.33651910501851e-02, 4.28964264548952e-02, 4.24975698090009e-02, 4.21621626403024e-02, 4.18837525340976e-02, 4.16559678694313e-02, 4.14725905258280e-02, 4.13276264439162e-02, 4.12153732735163e-02, 4.11304837613855e-02, 4.10680230682061e-02, 4.10235179046268e-02, 4.09929952683683e-02, 4.09730086640820e-02, 4.09606499902395e-02, 4.09535457589527e-02, 4.09498369359724e-02, 4.09481423957203e-02, 4.09475067235192e-02, 4.09473338060522e-02, 4.09473082838261e-02, 4.09473062575572e-02, 4.09472691234924e-02, 4.09470175352524e-02, 4.09460926775367e-02, 4.09436274022781e-02, 4.09382324911718e-02, 4.09279016772501e-02, 4.09099397043988e-02, 4.08809175075433e-02, 4.08366583443254e-02, 4.07722583335755e-02, 4.06821442735890e-02, 4.05601707341390e-02, 4.03997571646642e-02, 4.01940640890217e-02, 3.99362053681855e-02, 3.96194910727772e-02, 3.92376928549396e-02, 3.87853210527780e-02, 3.82579003653447e-02, 3.76522291084527e-02, 3.69666061180993e-02, 3.62010096047456e-02, 3.53572138912028e-02, 3.44388330533363e-02, 3.34512848623305e-02, 3.24016736929522e-02, 3.12985966480142e-02, 3.01518824861731e-02, 2.89722776119845e-02, 2.77710971230647e-02, 2.65598614134746e-02, 2.53499396585681e-02, 2.41522201895705e-02, 2.29768242672463e-02, 2.18328747967007e-02, 2.07283263356430e-02, 1.96698583254502e-02, 1.86628299032066e-02, 1.77112913273021e-02, 1.68180436342809e-02, 1.59847352965414e-02, 1.52119835714259e-02, 1.44995094248088e-02, 1.38462774637560e-02, 1.32506345787391e-02, 1.27104420502349e-02, 1.22231962280992e-02, 1.17861337617100e-02, 1.13963193375923e-02, 1.10507163577296e-02, 1.07462428026679e-02, 1.04798150193233e-02, 1.02483816693848e-02, 1.00489493970670e-02, 9.87860153277609e-03, 9.73451136401593e-03, 9.61395181145264e-03, 9.51430338665915e-03, 9.43306198102078e-03, 9.36784747050012e-03, 9.31641352116941e-03, 9.27665847918746e-03, 9.24663686529925e-03, 9.22457074914718e-03, 9.20886011896016e-03, 9.19809127858429e-03, 9.19104229515000e-03, 9.18668459736693e-03, 9.18417998366294e-03, 9.18287253231398e-03, 9.18227518872072e-03, 9.18205111126724e-03, 9.18199015740129e-03, 9.18198116076496e-03, 9.18198043989094e-03, 9.18196714638242e-03, 9.18187708123952e-03, 9.18154599782639e-03, 9.18066349372282e-03, 9.17873237187905e-03, 9.17503487313098e-03, 9.16860749007498e-03, 9.15822610210881e-03, 9.14240318396509e-03, 9.11939878590573e-03, 9.08724680202511e-03, 9.04379766959096e-03, 8.98677802392456e-03, 8.91386694478769e-03, 8.82278728764163e-03, 8.71140925991654e-03, 8.57786197708066e-03, 8.42064733052439e-03, 8.23874923132450e-03, 8.03173028872095e-03, 7.79980743786268e-03, 7.54389828130542e-03, 7.26563136976210e-03, 6.96731659464689e-03, 6.65187604869999e-03, 6.32274013776879e-03, 5.98371696511890e-03, 5.63884410766491e-03, 5.29223131854195e-03, 4.94987360301735e-03, 4.61163366860973e-03, 4.28287499125446e-03, 3.96656372547000e-03, 3.66514267837204e-03, 3.38049898524417e-03, 3.11396541730602e-03, 2.86634995183732e-03, 2.63798575341667e-03, 2.42879489336900e-03, 2.23836043513614e-03, 2.06599975312827e-03, 1.91083113381296e-03, 1.77182993301983e-03, 1.64787661993415e-03, 1.53780160821526e-03, 1.44042952284451e-03, 1.35461993021835e-03, 1.27929692281725e-03, 1.21346154929824e-03, 1.15618863768674e-03, 1.10661657213740e-03, 1.06393933781845e-03, 1.02740511276522e-03, 9.96319663499889e-04, 9.70049859418509e-04, 9.48023544786736e-04, 9.29724938281140e-04, 9.14687250009649e-04, 9.02485029349975e-04, 8.92728096218320e-04, 8.85057711748516e-04, 8.79144714026836e-04, 8.74688956450598e-04, 8.71419412298735e-04, 8.69094497602418e-04, 8.67502338948611e-04, 8.66460809895386e-04, 8.65817196839203e-04, 8.65447367921490e-04, 8.65254337604426e-04, 8.65166152742347e-04, 8.65133073553181e-04, 8.65124075389190e-04, 8.65122747285341e-04, 8.65122639936544e-04, 8.65120648761319e-04, 8.65107158318779e-04, 8.65057567288829e-04, 8.64925385457127e-04, 8.64636157892562e-04, 8.64082441033531e-04, 8.63120114601656e-04, 8.61566336510577e-04, 8.59199478227879e-04, 8.55761410193168e-04, 8.50962533974098e-04, 8.44489961878025e-04, 8.36019184369263e-04, 8.25229367983127e-04, 8.11821993856654e-04, 7.95541810305167e-04, 7.76198098684233e-04, 7.53683366866653e-04, 7.27986398423752e-04, 6.99197734995494e-04, 6.67508229931002e-04, 6.33204182030056e-04, 5.96663551192478e-04, 5.58354908186232e-04, 5.18834351730391e-04, 4.78729491163676e-04, 4.38700457102925e-04, 3.99380466340915e-04, 3.61317276123430e-04, 3.24944691818985e-04, 2.90595765250146e-04, 2.58535588308462e-04, 2.28972603132149e-04, 2.02025150670693e-04, 1.77667448241436e-04, 1.55720410290883e-04, 1.35934626989354e-04, 1.18125692559090e-04, 1.02244421462672e-04, 8.82919530127424e-05, 7.61241576228144e-05, 6.53163968422632e-05, 5.52185132534078e-05, 4.52258544711981e-05, 3.51472949731146e-05, 2.52723720564063e-05, 1.61768820419184e-05, 8.46192774872634e-06, 2.36397141151201e-06, 2.36610666348724e-06, 1.96207049185659e-06, 5.86708182481743e-07, 1.92769177687069e-06, 2.27076467496090e-06, 2.14097925521628e-06, 1.82447346475101e-06, 1.53287153636938e-06, 1.36014302398735e-06, 1.28789687181552e-06, 1.25329788858153e-06, 1.20235149125847e-06, 1.10246354267929e-06, 9.58562978325972e-07, 7.90051515764389e-07, 6.17655793178074e-07, 4.57723643876307e-07, 3.20488439342625e-07, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, 0.00000000000000e+00, -7.10542735760100e-15, -7.10542735760100e-15],
[ 4.10000000000000e+01, 4.09999999651547e+01, 4.09999988876689e+01, 4.09999915875327e+01, 4.09999647512858e+01, 4.09998932143679e+01, 4.09997366507913e+01, 4.09994367868249e+01, 4.09989152395449e+01, 4.09980720621128e+01, 4.09967850561307e+01, 4.09949098881747e+01, 4.09922810235061e+01, 4.09887134659632e+01, 4.09840052700815e+01, 4.09779407704543e+01, 4.09702944549173e+01, 4.09608353927787e+01, 4.09493321171517e+01, 4.09355578512774e+01, 4.09192959620454e+01, 4.09003455189878e+01, 4.08785268330625e+01, 4.08536868458823e+01, 4.08257042364031e+01, 4.07944941086558e+01, 4.07600121217171e+01, 4.07222579231530e+01, 4.06812777513872e+01, 4.06371660826919e+01, 4.05900662162912e+01, 4.05401697172352e+01, 4.04877146710870e+01, 4.04329827457560e+01, 4.03762951016387e+01, 4.03180072383838e+01, 4.02585029114897e+01, 4.01981872910353e+01, 4.01374795651823e+01, 4.00768052106196e+01, 4.00165881599416e+01, 3.99572430922951e+01, 3.98991680597423e+01, 3.98427376396537e+01, 3.97882967754302e+01, 3.97361554364254e+01, 3.96865841953212e+01, 3.96398107892459e+01, 3.95960177009284e+01, 3.95553407689726e+01, 3.95178688122353e+01, 3.94836442322983e+01, 3.94526645398535e+01, 3.94248847351066e+01, 3.94002204586337e+01, 3.93785518171725e+01, 3.93597277783801e+01, 3.93435710195713e+01, 3.93298831079581e+01, 3.93184498841450e+01, 3.93090469169133e+01, 3.93014448960115e+01, 3.92954148311212e+01, 3.92907329297068e+01, 3.92871850342925e+01, 3.92845705109342e+01, 3.92827054951798e+01, 3.92814254193904e+01, 3.92805867655046e+01, 3.92800680095982e+01, 3.92797697482210e+01, 3.92796140207087e+01, 3.92795428656511e+01, 3.92795161726297e+01, 3.92795089114521e+01, 3.92795078397165e+01, 3.92795076991274e+01, 3.92795044235695e+01, 3.92794822311641e+01, 3.92794006492568e+01, 3.92791831797116e+01, 3.92787072421081e+01, 3.92777957231936e+01, 3.92762104544298e+01, 3.92736478905376e+01, 3.92697372061389e+01, 3.92640409657425e+01, 3.92560584553042e+01, 3.92452316924046e+01, 3.92309540579049e+01, 3.92125814160400e+01, 3.91894455135620e+01, 3.91608693728000e+01, 3.91261843190674e+01, 3.90847482101385e+01, 3.90359643650088e+01, 3.89793006221313e+01, 3.89143078967934e+01, 3.88406375589250e+01, 3.87580569250754e+01, 3.86664621625973e+01, 3.85658879517824e+01, 3.84565133518553e+01, 3.83386634724492e+01, 3.82128067576892e+01, 3.80795479295319e+01, 3.79396168860791e+01, 3.77938540799918e+01, 3.76431930836237e+01, 3.74886411597743e+01, 3.73312586902274e+01, 3.71721382720221e+01, 3.70123841891141e+01, 3.68530928278734e+01, 3.66953344541023e+01, 3.65401366290939e+01, 3.63884694278426e+01, 3.62412325404312e+01, 3.60992442865746e+01, 3.59632325465346e+01, 3.58338275999043e+01, 3.57115568582459e+01, 3.55968414716351e+01, 3.54899947792279e+01, 3.53912225592124e+01, 3.53006250150026e+01, 3.52182004141749e+01, 3.51438502762222e+01, 3.50773859858850e+01, 3.50185366909659e+01, 3.49669583269111e+01, 3.49222435944924e+01, 3.48839327011279e+01, 3.48515246605711e+01, 3.48244889301465e+01, 3.48022771502485e+01, 3.47843347386790e+01, 3.47701120841595e+01, 3.47590750806609e+01, 3.47507147486361e+01, 3.47445557020754e+01, 3.47401632423462e+01, 3.47371488912102e+01, 3.47351742157770e+01, 3.47339528462879e+01, 3.47332506417826e+01, 3.47328840166268e+01, 3.47327165000257e+01, 3.47326536583545e+01, 3.47326365638575e+01, 3.47326340407441e+01, 3.47326338180012e+01, 3.47326294534639e+01, 3.47325998830711e+01, 3.47324911790245e+01, 3.47322014127592e+01, 3.47315672603482e+01, 3.47303527595213e+01, 3.47282406601131e+01, 3.47248267518606e+01, 3.47196174866391e+01, 3.47120311357772e+01, 3.47014026358915e+01, 3.46869921775741e+01, 3.46679974795790e+01, 3.46435695672695e+01, 3.46128317400051e+01, 3.45749012719176e+01, 3.45289132504422e+01, 3.44740458254422e+01, 3.44095460287006e+01, 3.43347552393676e+01, 3.42491333251500e+01, 3.41522804887608e+01, 3.40439558979642e+01, 3.39240922749276e+01, 3.37928057623796e+01, 3.36504005639534e+01, 3.34973680674366e+01, 3.33343803966072e+01, 3.31622785941549e+01, 3.29820559059328e+01, 3.27948368988109e+01, 3.26018533732788e+01, 3.24044181918057e+01, 3.22038982001627e+01, 3.20016873526188e+01, 3.17991809724975e+01, 3.15977518265819e+01, 3.13987284233076e+01, 3.12033757164618e+01, 3.10128782429589e+01, 3.08283256505736e+01, 3.06507005603130e+01, 3.04808687281241e+01, 3.03195714938741e+01, 3.01674205145163e+01, 3.00248947677240e+01, 2.98923397857366e+01, 2.97699690445969e+01, 2.96578673993119e+01, 2.95559964262165e+01, 2.94642015122591e+01, 2.93822205166350e+01, 2.93096938209505e+01, 2.92461755770613e+01, 2.91911459542114e+01, 2.91440241771983e+01, 2.91041821341677e+01, 2.90709583165462e+01, 2.90436718358154e+01, 2.90216362443490e+01, 2.90041728729108e+01, 2.89906233883769e+01, 2.89803612743982e+01, 2.89728019472553e+01, 2.89674112405692e+01, 2.89637120264768e+01, 2.89612887870130e+01, 2.89597900064378e+01, 2.89589283208887e+01, 2.89584784330634e+01, 2.89582728732454e+01, 2.89581957602799e+01, 2.89581747836476e+01, 2.89581716875394e+01, 2.89581714391932e+01, 2.89581668559462e+01, 2.89581358037646e+01, 2.89580216525247e+01, 2.89577173661889e+01, 2.89570514377029e+01, 2.89557760847434e+01, 2.89535581718994e+01, 2.89499732650160e+01, 2.89445031542864e+01, 2.89365371033511e+01, 2.89253769907990e+01, 2.89102464066398e+01, 2.88903036486544e+01, 2.88646584322971e+01, 2.88323919850091e+01, 2.87925800456018e+01, 2.87443181383724e+01, 2.86867483486839e+01, 2.86190867024624e+01, 2.85406501577620e+01, 2.84508821630095e+01, 2.83493757325648e+01, 2.82358930413846e+01, 2.81103806482732e+01, 2.79729796184717e+01, 2.78240300241064e+01, 2.76640695447858e+01, 2.74938261572589e+01, 2.73142051772896e+01, 2.71262711818339e+01, 2.69312255771122e+01, 2.67303807699891e+01, 2.65251320297322e+01, 2.63169281826753e+01, 2.61072422587502e+01, 2.58975431105984e+01, 2.56892688666980e+01, 2.54838028807088e+01, 2.52824526248715e+01, 2.50864317698442e+01, 2.48968455161114e+01, 2.47146791049031e+01, 2.45407893431889e+01, 2.43758989245443e+01, 2.42205933076298e+01, 2.40753199167551e+01, 2.39403894448960e+01, 2.38159790607424e+01, 2.37021373424241e+01, 2.35987907782559e+01, 2.35057516876150e+01, 2.34227274223558e+01, 2.33493307108779e+01, 2.32850910031511e+01, 2.32294666657935e+01, 2.31818578619988e+01, 2.31416199323538e+01, 2.31080770705028e+01, 2.30805360638619e+01, 2.30582998463832e+01, 2.30406805903013e+01, 2.30270120496526e+01, 2.30166608628285e+01, 2.30090365268313e+01, 2.30035997738929e+01, 2.29998691124885e+01, 2.29974253392743e+01, 2.29959138848145e+01, 2.29950449218375e+01, 2.29945912370908e+01, 2.29943839430108e+01, 2.29943061795580e+01, 2.29942850259863e+01, 2.29942819037625e+01, 2.29942816629465e+01, 2.29942773386649e+01, 2.29942480410143e+01, 2.29941403396589e+01, 2.29938532466754e+01, 2.29932249470873e+01, 2.29920216643554e+01, 2.29899291022222e+01, 2.29865468490377e+01, 2.29813860667160e+01, 2.29738707126806e+01, 2.29633424579192e+01, 2.29490693658809e+01, 2.29302582842364e+01, 2.29060707744848e+01, 2.28756422648604e+01, 2.28381039641333e+01, 2.27926069245011e+01, 2.27383475001673e+01, 2.26745933257251e+01, 2.26007088474371e+01, 2.25161793927372e+01, 2.24206327683712e+01, 2.23138574411091e+01, 2.21958164770199e+01, 2.20666565899099e+01, 2.19267118648393e+01, 2.17765019625134e+01, 2.16167248567540e+01, 2.14482443931779e+01, 2.12720731691550e+01, 2.10893514145036e+01, 2.09013226950545e+01, 2.07093073655001e+01, 2.05146747618440e+01, 2.03188151430657e+01, 2.01231123601904e+01, 1.99289181436809e+01, 1.97375287574289e+01, 1.95501645794541e+01, 1.93679529555613e+01, 1.91919144589121e+01, 1.90229525017922e+01, 1.88618461048396e+01, 1.87092455415170e+01, 1.85656705382822e+01, 1.84315107124215e+01, 1.83070279554540e+01, 1.81923605070614e+01, 1.80875285028523e+01, 1.79924408131734e+01, 1.79069030170914e+01, 1.78306263750119e+01, 1.77632376754312e+01, 1.77042898363809e+01, 1.76532731405189e+01, 1.76096269747970e+01, 1.75727519316985e+01, 1.75420221101719e+01, 1.75167974321665e+01, 1.74964357673163e+01, 1.74803046365148e+01, 1.74677922478435e+01, 1.74583176084914e+01, 1.74513394565057e+01, 1.74463637683782e+01, 1.74429496236100e+01, 1.74407132454947e+01, 1.74393300872811e+01, 1.74385348924875e+01, 1.74381197244508e+01, 1.74379300296748e+01, 1.74378588684421e+01, 1.74378395108450e+01, 1.74378366537036e+01, 1.74378364373695e+01, 1.74378326050320e+01, 1.74378066403787e+01, 1.74377111915263e+01, 1.74374567596189e+01, 1.74368999400346e+01, 1.74358335578705e+01, 1.74339790914186e+01, 1.74309817298199e+01, 1.74264083559635e+01, 1.74197486817404e+01, 1.74104196881955e+01, 1.73977734361312e+01, 1.73811082119532e+01, 1.73596828588353e+01, 1.73327340161529e+01, 1.72994958543655e+01, 1.72592217543838e+01, 1.72112072486027e+01, 1.71548134258073e+01, 1.70894899156199e+01, 1.70147965212159e+01, 1.69304225708749e+01, 1.68362031150337e+01, 1.67321312064060e+01, 1.66183656612391e+01, 1.64952338993230e+01, 1.63632296842375e+01, 1.62230058168837e+01, 1.60753620585239e+01, 1.59212287610356e+01, 1.57616468525454e+01, 1.55977449605526e+01, 1.54307145493195e+01, 1.52617840020307e+01, 1.50921925893881e+01, 1.49231652333187e+01, 1.47558888968494e+01, 1.45914913112823e+01, 1.44310225961603e+01, 1.42754401472787e+01, 1.41255969779981e+01, 1.39822335156858e+01, 1.38459726935885e+01, 1.37173180507192e+01, 1.35966544653782e+01, 1.34842511031884e+01, 1.33802661545749e+01, 1.32847529623528e+01, 1.31976671883066e+01, 1.31188747286456e+01, 1.30481601529877e+01, 1.29852355025820e+01, 1.29297493353135e+01, 1.28812959441069e+01, 1.28394246999993e+01, 1.28036494812847e+01, 1.27734581469304e+01, 1.27483219980665e+01, 1.27277051486625e+01, 1.27110736988795e+01, 1.26979045756958e+01, 1.26876938789244e+01, 1.26799645501723e+01, 1.26742731706446e+01, 1.26702156932791e+01, 1.26674319268703e+01, 1.26656086148920e+01, 1.26644809888027e+01, 1.26638327228074e+01, 1.26634942715762e+01, 1.26633396308793e+01, 1.26632816198905e+01, 1.26632658395021e+01, 1.26632635103501e+01, 1.26632633354999e+01, 1.26632602579387e+01, 1.26632394070152e+01, 1.26631627569076e+01, 1.26629584365772e+01, 1.26625112901558e+01, 1.26616549651556e+01, 1.26601658522706e+01, 1.26577591667312e+01, 1.26540874215877e+01, 1.26487414960714e+01, 1.26412544448797e+01, 1.26311081242700e+01, 1.26177426261532e+01, 1.26005684110475e+01, 1.25789809156605e+01, 1.25523772842303e+01, 1.25201747403810e+01, 1.24818299865409e+01, 1.24368589015288e+01, 1.23848557157350e+01, 1.23255107896197e+01, 1.22586261160414e+01, 1.21841277184100e+01, 1.21020742287018e+01, 1.20126611002774e+01, 1.19162201321538e+01, 1.18132142394942e+01, 1.17042276796648e+01, 1.15899522105468e+01, 1.14711698930344e+01, 1.13487334299184e+01, 1.12235450410591e+01, 1.10965349003703e+01, 1.09686401038809e+01, 1.08407850101547e+01, 1.07138636130837e+01, 1.05887243963233e+01, 1.04661579036359e+01, 1.03468870629105e+01, 1.02315601407776e+01, 1.01207460892943e+01, 1.00149319782506e+01, 9.91452218190718e+00, 9.81983899866198e+00, 9.73112441548209e+00, 9.64854277527225e+00, 9.57218415566461e+00, 9.50206831545001e+00, 9.43814910599865e+00, 9.38031927767420e+00, 9.32841563508311e+00, 9.28222451054661e+00, 9.24148753330890e+00, 9.20590767360131e+00, 9.17515553656448e+00, 9.14887587191402e+00, 9.12669425186919e+00, 9.10822385316315e+00, 9.09307226007856e+00, 9.08084818586363e+00, 9.07116799129689e+00, 9.06366186346144e+00, 9.05797950684515e+00, 9.05379519440991e+00, 9.05081202960250e+00, 9.04876528220367e+00, 9.04742468154817e+00, 9.04659557938529e+00, 9.04611893016995e+00, 9.04587007696046e+00, 9.04575637401905e+00, 9.04571372014377e+00, 9.04570211725770e+00, 9.04570040469612e+00, 9.04570027652451e+00, 9.04569802576707e+00, 9.04568277655727e+00, 9.04562671889584e+00, 9.04547729011188e+00, 9.04515027069897e+00, 9.04452399639677e+00, 9.04343492201358e+00, 9.04167474699429e+00, 9.03898928290671e+00, 9.03507920757839e+00, 9.02960280784675e+00, 9.02218076098096e+00, 9.01240294233524e+00, 8.99983717312214e+00, 8.98403973817038e+00, 8.96456741181023e+00, 8.94099063537156e+00, 8.91290739900051e+00, 8.87995730207759e+00, 8.84183520974060e+00, 8.79830389678107e+00, 8.74920508153949e+00, 8.69446830511221e+00, 8.63411720446559e+00, 8.56827285618957e+00, 8.49715402018407e+00, 8.42107427573031e+00, 8.34043620099845e+00, 8.25572288711038e+00, 8.16748718895661e+00, 8.07633919170755e+00, 7.98293241444133e+00, 7.88794928498471e+00, 7.79208640959109e+00, 7.69604013375258e+00, 7.60049285012566e+00, 7.50610045693445e+00, 7.41348130374237e+00, 7.32320687959269e+00, 7.23579440223084e+00, 7.15170136174530e+00, 7.07132196699956e+00, 6.99498535036680e+00, 6.92295531628071e+00, 6.85543137883406e+00, 6.79255082438037e+00, 6.73439155269343e+00, 6.68097548681613e+00, 6.63227238796099e+00, 6.58820395910717e+00, 6.54864816279393e+00, 6.51344371115513e+00, 6.48239470796369e+00, 6.45527543359894e+00, 6.43183526571071e+00, 6.41180372272272e+00, 6.39489560610374e+00, 6.38081620238869e+00, 6.36926648899663e+00, 6.35994827061215e+00, 6.35256915684482e+00, 6.34684727855765e+00, 6.34251563104449e+00, 6.33932592833153e+00, 6.33705185517804e+00, 6.33549161235404e+00, 6.33446966649237e+00, 6.33383763773262e+00, 6.33347428546531e+00, 6.33328458327329e+00, 6.33319790687469e+00, 6.33316539158130e+00, 6.33315654663463e+00, 6.33315524113910e+00, 6.33315514339659e+00, 6.33315342650042e+00, 6.33314179427861e+00, 6.33309903303560e+00, 6.33298504752914e+00, 6.33273559439407e+00, 6.33225786699995e+00, 6.33142711132563e+00, 6.33008443295645e+00, 6.32803593251745e+00, 6.32505327978127e+00, 6.32087580402360e+00, 6.31521413855631e+00, 6.30775540961675e+00, 6.29816990345179e+00, 6.28611908106173e+00, 6.27126473959179e+00, 6.25327904629022e+00, 6.23185510040529e+00, 6.20671761683328e+00, 6.17763328005430e+00, 6.14442029519603e+00, 6.10695667119357e+00, 6.06518681297543e+00, 6.01912607603937e+00, 5.96886304410094e+00, 5.91455942057809e+00, 5.85644756533998e+00, 5.79482584460242e+00, 5.73005207899234e+00, 5.66253546002738e+00, 5.59272735110687e+00, 5.52111139489924e+00, 5.44819332080878e+00, 5.37449079502276e+00, 5.30052359432476e+00, 5.22680432474942e+00, 5.15382985458291e+00, 5.08207359064252e+00, 5.01197869553707e+00, 4.94395231779455e+00, 4.87836088224493e+00, 4.81552646216543e+00, 4.75572422682410e+00, 4.69918092949603e+00, 4.64607437423534e+00, 4.59653377736102e+00, 4.55064092388702e+00, 4.50843201108970e+00, 4.46990007097521e+00, 4.43499786946766e+00, 4.40364119084825e+00, 4.37571242914284e+00, 4.35106442158842e+00, 4.32952447108431e+00, 4.31089851319245e+00, 4.29497538789879e+00, 4.28153117668212e+00, 4.27033356172372e+00, 4.26114615709029e+00, 4.25373275258268e+00, 4.24786140107631e+00, 4.24330827111515e+00, 4.23986117975760e+00, 4.23732271752898e+00, 4.23551287881835e+00, 4.23427111776508e+00, 4.23345776173360e+00, 4.23295473150426e+00, 4.23266553849891e+00, 4.23251455352141e+00, 4.23244556720022e+00, 4.23241968804332e+00, 4.23241264828460e+00, 4.23241160923093e+00, 4.23241153123057e+00, 4.23241015835865e+00, 4.23240085695058e+00, 4.23236666400795e+00, 4.23227551831030e+00, 4.23207604885767e+00, 4.23169404358962e+00, 4.23102974135436e+00, 4.22995607419166e+00, 4.22831796845401e+00, 4.22593279147969e+00, 4.22259200457295e+00, 4.21806405204016e+00, 4.21209847927923e+00, 4.20443123019565e+00, 4.19479102600028e+00, 4.18290667511407e+00, 4.16851510983341e+00, 4.15136989292854e+00, 4.13124989055320e+00, 4.10796777125725e+00, 4.08137796897147e+00, 4.05138374443221e+00, 4.01794299731753e+00, 3.98107252145930e+00, 3.94085045705454e+00, 3.89741677404546e+00, 3.85097171518274e+00, 3.80177222968270e+00, 3.75012653180841e+00, 3.69638701570447e+00, 3.64094184115810e+00, 3.58420556825197e+00, 3.52660925721304e+00, 3.46859046030360e+00, 3.41058351499208e+00, 3.35301050416758e+00, 3.29627318457081e+00, 3.24074610555807e+00, 3.18677105450880e+00, 3.13465288042106e+00, 3.08465667041932e+00, 3.03700619030641e+00, 2.99188345315779e+00, 2.94942925039243e+00, 2.90974446698777e+00, 2.87289200432198e+00, 2.83889914741079e+00, 2.80776023456307e+00, 2.77943951326876e+00, 2.75387409338011e+00, 2.73097693482074e+00, 2.71063983023818e+00, 2.69273636189163e+00, 2.67712482587076e+00, 2.66365112518393e+00, 2.65215163644410e+00, 2.64245605326181e+00, 2.63439020375316e+00, 2.62777883074259e+00, 2.62244831242229e+00, 2.61822928968432e+00, 2.61495915537029e+00, 2.61248435156044e+00, 2.61066241487124e+00, 2.60936370744559e+00, 2.60847277346510e+00, 2.60788926776873e+00, 2.60752841429003e+00, 2.60732096689419e+00, 2.60721266284030e+00, 2.60716317832186e+00, 2.60714461505120e+00, 2.60713956539909e+00, 2.60713882008130e+00, 2.60713876389301e+00, 2.60713777175778e+00, 2.60713104990002e+00, 2.60710633970574e+00, 2.60704047190668e+00, 2.60689632440967e+00, 2.60662027455759e+00, 2.60614025237253e+00, 2.60536449167572e+00, 2.60418106453398e+00, 2.60245827039878e+00, 2.60004593346982e+00, 2.59677763906383e+00, 2.59247391112541e+00, 2.58694629797450e+00, 2.58000229218357e+00, 2.57145096432114e+00, 2.56110914149462e+00, 2.54880791357461e+00, 2.53439920700455e+00, 2.51776213306737e+00, 2.49880879928828e+00, 2.47748927355174e+00, 2.45379541341276e+00, 2.42776331891962e+00, 2.39947423454291e+00, 2.36905381045297e+00, 2.33666972897143e+00, 2.30252780036233e+00, 2.26686672423164e+00, 2.22995179002120e+00, 2.19206784526462e+00, 2.15351188875682e+00, 2.11458564600569e+00, 2.07558845793380e+00, 2.03681076526388e+00, 1.99852840688552e+00, 1.96099787828800e+00, 1.92445262330224e+00, 1.88910036535904e+00, 1.85512142803473e+00, 1.82266795168280e+00, 1.79186388442433e+00, 1.76280561108094e+00, 1.73556308102332e+00, 1.71018130293176e+00, 1.68668208843145e+00, 1.66506594486351e+00, 1.64531403777377e+00, 1.62739016415323e+00, 1.61124269660358e+00, 1.59680647540242e+00, 1.58400463924691e+00, 1.57275039589840e+00, 1.56294874090582e+00, 1.55449813611284e+00, 1.54729215996512e+00, 1.54122113907691e+00, 1.53617376555799e+00, 1.53203869782694e+00, 1.52870613472712e+00, 1.52606934447758e+00, 1.52402612213687e+00, 1.52248014262452e+00, 1.52134217165653e+00, 1.52053109479376e+00, 1.51997472556224e+00, 1.51961035742168e+00, 1.51938503110208e+00, 1.51925549810073e+00, 1.51918787230567e+00, 1.51915697400207e+00, 1.51914538305941e+00, 1.51914223004932e+00, 1.51914176467190e+00, 1.51914172939694e+00, 1.51914110400269e+00, 1.51913686686868e+00, 1.51912129077581e+00, 1.51907977109158e+00, 1.51898890861546e+00, 1.51881490505962e+00, 1.51851233944088e+00, 1.51802338820080e+00, 1.51727754531581e+00, 1.51619189032510e+00, 1.51467194120955e+00, 1.51261311447080e+00, 1.50990279578731e+00, 1.50642300082813e+00, 1.50205357731816e+00, 1.49667586716804e+00, 1.49017671317642e+00, 1.48245266112226e+00, 1.47341417834865e+00, 1.46298968795404e+00, 1.45112920713910e+00, 1.43780738215931e+00, 1.42302573252621e+00, 1.40681395370686e+00, 1.38923017871640e+00, 1.37036016085374e+00, 1.35031540695691e+00, 1.32923035657442e+00, 1.30725876091197e+00, 1.28456946070255e+00, 1.26134179027030e+00, 1.23776084414973e+00, 1.21401283309622e+00, 1.19028073064813e+00, 1.16674037357396e+00, 1.14355713439348e+00, 1.12088323664780e+00, 1.09885573814238e+00, 1.07759516747831e+00, 1.05720476712381e+00, 1.03777027316576e+00, 1.01936014777507e+00, 1.00202617455702e+00, 9.85804328036878e-01, 9.70715834972914e-01, 9.56768355357262e-01, 9.43957223326862e-01, 9.32266701435751e-01, 9.21671214762441e-01, 9.12136543332252e-01, 9.03620961756651e-01, 8.96076323472087e-01, 8.89449093312592e-01, 8.83681336320969e-01, 8.78711672748764e-01, 8.74476209257239e-01, 8.70909454629285e-01, 8.67945225123393e-01, 8.65517540299706e-01, 8.63561505138748e-01, 8.62014169013747e-01, 8.60815347057716e-01, 8.59908385168175e-01, 8.59240846761622e-01, 8.58765097790901e-01, 8.58438766715111e-01, 8.58225058160912e-01, 8.58092902862026e-01, 8.58016931876676e-01, 8.57977269680823e-01, 8.57959148039598e-01, 8.57952350041955e-01, 8.57950500827009e-01, 8.57950227886938e-01, 8.57950207067191e-01, 8.57949836219284e-01, 8.57947323672974e-01, 8.57938087329906e-01, 8.57913466959992e-01, 8.57859587555069e-01, 8.57756408457448e-01, 8.57576999193796e-01, 8.57287079960635e-01, 8.56844861086934e-01, 8.56201211082094e-01, 8.55300176474167e-01, 8.54079867925925e-01, 8.52473715547184e-01, 8.50412081594435e-01, 8.47824200933552e-01, 8.44640399312780e-01, 8.40794517804220e-01, 8.36226450452678e-01, 8.30884683426596e-01, 8.24728710259379e-01, 8.17731191505032e-01, 8.09879730269216e-01, 8.01178148743134e-01, 7.91647175040801e-01, 7.81324482977936e-01, 7.70264067312986e-01, 7.58534979760825e-01, 7.46219492598556e-01, 7.33410792750135e-01, 7.20210336328876e-01, 7.06725009320770e-01, 6.93064243417872e-01, 6.79337227455697e-01, 6.65650336277174e-01, 6.52104872888110e-01, 6.38795189722828e-01, 6.25807223947490e-01, 6.13217452847692e-01, 6.01092250657850e-01, 5.89487609069543e-01, 5.78449170670993e-01, 5.68012517586425e-01, 5.58203655933994e-01, 5.49039639396909e-01, 5.40529281047448e-01, 5.32673910423867e-01, 5.25468141695427e-01, 5.18900627703331e-01, 5.12954783083053e-01, 5.07609467105787e-01, 5.02839623051493e-01, 4.98616875712315e-01, 4.94910092000346e-01, 4.91685911647657e-01, 4.88909255740893e-01, 4.86543820461243e-01, 4.84552562064053e-01, 4.82898177012792e-01, 4.81543578487774e-01, 4.80452367449317e-01, 4.79589293299867e-01, 4.78920696223334e-01, 4.78414920745273e-01, 4.78042688197306e-01, 4.77777414785933e-01, 4.77595461994859e-01, 4.77476307147391e-01, 4.77402624082742e-01, 4.77360266923803e-01, 4.77338153622874e-01, 4.77328050090698e-01, 4.77324259942179e-01, 4.77323228933209e-01, 4.77323076758566e-01, 4.77323065068838e-01, 4.77322855773870e-01, 4.77321437771543e-01, 4.77316225072620e-01, 4.77302330143374e-01, 4.77271922507491e-01, 4.77213692373539e-01, 4.77112442593516e-01, 4.76948830690411e-01, 4.76699280769502e-01, 4.76336082562895e-01, 4.75827691281346e-01, 4.75139236967536e-01, 4.74233245322850e-01, 4.73070563315481e-01, 4.71611472310393e-01, 4.69816959339937e-01, 4.67650104176158e-01, 4.65077527149639e-01, 4.62070831565015e-01, 4.58607966628023e-01, 4.54674433512210e-01, 4.50264259751087e-01, 4.45380676188719e-01, 4.40036446155631e-01, 4.34253817405441e-01, 4.28064091887983e-01, 4.21506834265855e-01, 4.14628764520927e-01, 4.07482400456075e-01, 4.00124530271846e-01, 3.92614602379872e-01, 3.85013118868493e-01, 3.77380111193226e-01, 3.69773763165760e-01, 3.62249229132409e-01, 3.54857676526358e-01, 3.47645563785861e-01, 3.40654148602589e-01, 3.33919208721554e-01, 3.27470948596556e-01, 3.21334060129935e-01, 3.15527904098417e-01, 3.10066780032841e-01, 3.04960255503524e-01, 3.00213530199024e-01, 2.95827815193675e-01, 2.91800712840228e-01, 2.88126587408030e-01, 2.84796920671667e-01, 2.81800650008968e-01, 2.79124489153489e-01, 2.76753233580969e-01, 2.74670053645245e-01, 2.72856779083398e-01, 2.71294178457616e-01, 2.69962236569249e-01, 2.68840431947659e-01, 2.67908015268965e-01, 2.67144288090087e-01, 2.66528879699450e-01, 2.66042018302556e-01, 2.65664791304530e-01, 2.65379388248775e-01, 2.65169319140249e-01, 2.65019600519658e-01, 2.64916901829594e-01, 2.64849645348079e-01, 2.64808054236610e-01, 2.64784144987260e-01, 2.64771662642268e-01, 2.64765959455730e-01, 2.64763820009740e-01, 2.64763238030078e-01, 2.64763152131152e-01, 2.64763145482227e-01, 2.64763025783076e-01, 2.64762214804662e-01, 2.64759233575930e-01, 2.64751286820179e-01, 2.64733896074731e-01, 2.64700592767170e-01, 2.64642684451189e-01, 2.64549106436121e-01, 2.64406369865668e-01, 2.64198615762105e-01, 2.63907782457323e-01, 2.63513890971922e-01, 2.62995449099698e-01, 2.62329970103465e-01, 2.61494596058697e-01, 2.60466809194391e-01, 2.59225207482402e-01, 2.57750313821468e-01, 2.56025382222205e-01, 2.54037160271800e-01, 2.51776565654374e-01, 2.49239236248549e-01, 2.46425918618257e-01, 2.43342668419913e-01, 2.40000847754665e-01, 2.36416917748272e-01, 2.32612038286049e-01, 2.28611499402682e-01, 2.24444018987924e-01, 2.20140948216901e-01, 2.15735428934671e-01, 2.11261546145991e-01, 2.06753514306605e-01, 2.02244929167591e-01, 1.97768108555906e-01, 1.93353536746976e-01, 1.89029418900589e-01, 1.84821345037626e-01, 1.80752057592890e-01, 1.76841312779857e-01, 1.73105823741295e-01, 1.69559272496265e-01, 1.66212377740678e-01, 1.63073006325014e-01, 1.60146317466051e-01, 1.57434930248066e-01, 1.54939106580478e-01, 1.52656943396387e-01, 1.50584569422641e-01, 1.48716343270692e-01, 1.47045050844696e-01, 1.45562101101206e-01, 1.44257719993078e-01, 1.43121142965491e-01, 1.42140806627957e-01, 1.41304540206455e-01, 1.40599757094918e-01, 1.40013646309335e-01, 1.39533362951241e-01, 1.39146215972815e-01, 1.38839850684711e-01, 1.38602422643238e-01, 1.38422758885660e-01, 1.38290502026457e-01, 1.38196232554577e-01, 1.38131564819012e-01, 1.38089212676441e-01, 1.38063021584550e-01, 1.38047965011395e-01, 1.38040104327494e-01, 1.38036512762696e-01, 1.38035165450312e-01, 1.38034798949363e-01, 1.38034744854622e-01, 1.38034740633536e-01, 1.38034664203467e-01, 1.38034146378992e-01, 1.38032242808691e-01, 1.38027168644463e-01, 1.38016064240958e-01, 1.37994799055662e-01, 1.37957822086449e-01, 1.37898066510942e-01, 1.37806915385895e-01, 1.37674234239938e-01, 1.37488475043057e-01, 1.37236854247000e-01, 1.36905605265852e-01, 1.36480302852448e-01, 1.35946253344905e-01, 1.35288940820892e-01, 1.34494515031818e-01, 1.33550302925002e-01, 1.32445322017851e-01, 1.31170771325003e-01, 1.29720474397494e-01, 1.28091249660677e-01, 1.26283185814422e-01, 1.24299804549963e-01, 1.22148098960629e-01, 1.19838443281964e-01, 1.17384377328619e-01, 1.14802276475849e-01, 1.12110924572896e-01, 1.09331012214973e-01, 1.06484585986081e-01, 1.03594475481358e-01, 1.00683724212395e-01, 9.77750481494510e-02, 9.48903420266092e-02, 9.20502490576141e-02, 8.92738048029003e-02, 8.65781609810412e-02, 8.39783903678679e-02, 8.14873698331340e-02, 7.91157352213503e-02, 7.68718992947441e-02, 7.47621223638575e-02, 7.27906244982535e-02, 7.09597282512533e-02, 6.92700215246518e-02, 6.77205313832374e-02, 6.63089011541800e-02, 6.50315648416679e-02, 6.38839146158716e-02, 6.28604587708423e-02, 6.19549689840326e-02, 6.11606169008923e-02, 6.04701009419202e-02, 5.98757648051134e-02, 5.93697093765257e-02, 5.89438997282912e-02, 5.85902685756992e-02, 5.83008170650388e-02, 5.80677131097644e-02, 5.78833867713158e-02, 5.77406214565244e-02, 5.76326390487765e-02, 5.75531765901758e-02, 5.74965518057127e-02, 5.74577146867199e-02, 5.74322825172047e-02, 5.74165561390956e-02, 5.74075158682703e-02, 5.74027962745447e-02, 5.74006399084865e-02, 5.73998309898442e-02, 5.73996109452750e-02, 5.73995784671624e-02, 5.73995759114823e-02, 5.73995293565517e-02, 5.73992139415367e-02, 5.73980544499619e-02, 5.73949637301823e-02, 5.73882000656525e-02, 5.73752479883183e-02, 5.73527278149015e-02, 5.73163388861886e-02, 5.72608411296187e-02, 5.71800790420625e-02, 5.70670514403560e-02, 5.69140292299579e-02, 5.67127219305732e-02, 5.64544917015937e-02, 5.61306111522413e-02, 5.57325583677724e-02, 5.52523394935847e-02, 5.46828261508949e-02, 5.40180922134184e-02, 5.32537324129052e-02, 5.23871442164321e-02, 5.14177547309060e-02, 5.03471761927443e-02, 4.91792769782293e-02, 4.79201597903298e-02, 4.65780444808317e-02, 4.51630592556569e-02, 4.36869503208683e-02, 4.21627255749755e-02, 4.06042523294659e-02, 3.90258317295178e-02, 3.74417732895864e-02, 3.58659918943189e-02, 3.43116466598659e-02, 3.27908369301824e-02, 3.13143655436150e-02, 2.98915741782508e-02, 2.85302504220155e-02, 2.72366017689691e-02, 2.60152882547189e-02, 2.48695032478210e-02, 2.38010906133965e-02, 2.28106866066184e-02, 2.18978753741297e-02, 2.10613485159072e-02, 2.02990607638966e-02, 1.96083758301882e-02, 1.89793630357826e-02, 1.84228283889070e-02, 1.79275877559562e-02, 1.74898091359594e-02, 1.71055628038772e-02, 1.67708753569755e-02, 1.64817727500690e-02, 1.62343147394424e-02, 1.60246230974863e-02, 1.58489057475819e-02, 1.57034786351318e-02, 1.55847867272030e-02, 1.54894250525359e-02, 1.54141601891524e-02, 1.53559521115056e-02, 1.53119758576423e-02, 1.52796421015040e-02, 1.52566154438816e-02, 1.52408290889241e-02, 1.52304945621538e-02, 1.52241052500202e-02, 1.52204327856893e-02, 1.52185156475397e-02, 1.52176397413371e-02, 1.52173111667793e-02, 1.52172217872959e-02, 1.52172085951037e-02, 1.52172075477424e-02, 1.52171883535814e-02, 1.52170583107463e-02, 1.52165802660775e-02, 1.52153060247215e-02, 1.52125176335005e-02, 1.52071785070831e-02, 1.51978967763782e-02, 1.51829032321871e-02, 1.51600462066028e-02, 1.51268056225007e-02, 1.50803281779943e-02, 1.50174851416338e-02, 1.49349534473870e-02, 1.48293196527263e-02, 1.46972048558069e-02, 1.45354069107668e-02, 1.43410543534586e-02, 1.41117645372688e-02, 1.38457968204051e-02, 1.35421905123465e-02, 1.32008769370699e-02, 1.28227556103511e-02, 1.24097262589942e-02, 1.19646711913717e-02, 1.14913861648134e-02, 1.09944620378144e-02, 1.04791236823729e-02, 9.95103635463579e-03, 9.41609249821340e-03, 8.88019342156809e-03, 8.34904027126271e-03, 7.82794726562486e-03, 7.32168751143547e-03, 6.83437829658828e-03, 6.36940898542093e-03, 5.92941097708888e-03, 5.51626600000077e-03, 5.13114660878955e-03, 4.77458130809083e-03, 4.44653627646829e-03, 4.14650605655709e-03, 3.87360654459457e-03, 3.62666487221475e-03, 3.40430212143161e-03, 3.20500614623853e-03, 3.02719302943899e-03, 2.86925680754990e-03, 2.72960796370114e-03, 2.60670176081135e-03, 2.49905778316872e-03, 2.40527214805297e-03, 2.32402383140595e-03, 2.25407648069620e-03, 2.19427698407770e-03, 2.14355192881754e-03, 2.10090291649578e-03, 2.06540152024814e-03, 2.03618448607301e-03, 2.01244960639002e-03, 1.99345253397354e-03, 1.97850465659144e-03, 1.96697201696594e-03, 1.95827514183833e-03, 1.95188954125314e-03, 1.94734656184892e-03, 1.94423423070589e-03, 1.94219771182702e-03, 1.94093901854586e-03, 1.94021567614300e-03, 1.93983810777354e-03, 1.93966561229786e-03, 1.93960090631151e-03, 1.93958330499555e-03, 1.93958070708078e-03, 1.93958049894150e-03, 1.93957666071043e-03, 1.93955065623186e-03, 1.93945506345727e-03, 1.93920026674022e-03, 1.93864274433508e-03, 1.93757538220041e-03, 1.93572035657752e-03, 1.93272516338538e-03, 1.92816240522120e-03, 1.92153395316552e-03, 1.91228005821387e-03, 1.89979386259864e-03, 1.88344153393416e-03, 1.86258789480576e-03, 1.83662694907127e-03, 1.80501613656503e-03, 1.76731253007034e-03, 1.72320859894626e-03, 1.67256469914889e-03, 1.61543521591057e-03, 1.55208537863416e-03, 1.48299625374904e-03, 1.40885631386567e-03, 1.33053921460657e-03, 1.24906883506526e-03, 1.16557402778739e-03, 1.08123663552817e-03, 9.97236978456328e-04, 9.14701141041974e-04, 8.34654066373730e-04, 7.57981819177189e-04, 6.85405462249378e-04, 6.17467771476754e-04, 5.54532503855398e-04, 4.96794375543175e-04, 4.44296796132448e-04, 3.96954191351995e-04, 3.54576424058450e-04, 3.16893806321161e-04, 2.83581729398463e-04, 2.54283753108098e-04, 2.28631576831617e-04, 2.06341878509875e-04, 1.86892070736633e-04, 1.69978200823095e-04, 1.55528547750805e-04, 1.43011736930987e-04, 1.32208556117064e-04, 1.22965287722820e-04, 1.15110097425486e-04, 1.08387736069915e-04, 1.02635918679539e-04, 9.77259811776321e-05, 9.36243612201793e-05, 9.02733731109606e-05, 8.74745116732356e-05, 8.51522252034442e-05, 8.32439010736152e-05, 8.16947624784348e-05, 8.04557108367021e-05, 7.94818317970680e-05, 7.87316646650993e-05, 7.81670616376573e-05, 7.77533546880704e-05, 7.74595821937396e-05, 7.72586383988028e-05, 7.71273088915336e-05, 7.70462056678411e-05, 7.69996215765900e-05, 7.69753129077344e-05, 7.69642089437639e-05, 7.69600439269880e-05, 7.69589109808066e-05, 7.69587437687846e-05, 7.69587303039998e-05, 7.69584811592949e-05, 7.69567931726556e-05, 7.69505880668930e-05, 7.69340481596714e-05, 7.68978542851073e-05, 7.68285516130618e-05, 7.67080754755511e-05, 7.65134722229277e-05, 7.62168577495004e-05, 7.57856711643967e-05, 7.51833073096009e-05, 7.43702702479254e-05, 7.33063581108695e-05, 7.19590609890020e-05, 7.03364866652123e-05, 6.84068320211395e-05, 6.61169210260937e-05, 6.34506295398296e-05, 6.04084284532291e-05, 5.70063569469426e-05, 5.32774391146518e-05, 4.93374159233895e-05, 4.52826990979815e-05, 4.10628501228416e-05, 3.67511982446445e-05, 3.24480670386151e-05, 2.83478021358974e-05, 2.43851541981144e-05, 2.05925736835866e-05, 1.70841981059766e-05, 1.38333229671161e-05, 1.07832449600664e-05, 8.02443997116598e-06, 5.58100462200173e-06, 3.45575155336064e-06, 1.74577208156279e-06, 4.18011303793264e-07, 6.32709344472460e-07, 6.23030103952260e-07, 4.53138547129583e-07, 0.00000000000000e+00, 0.00000000000000e+00, 7.10542735760100e-15, 7.10542735760100e-15, 0.00000000000000e+00, 4.73958422730902e-07, 6.06991839902093e-07, 6.43290732682544e-07, 5.82646926261532e-07, 4.53567693625700e-07, 7.10542735760100e-15, 0.00000000000000e+00, 7.10542735760100e-15, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 0.00000000000000e+00, 7.10542735760100e-15, 7.10542735760100e-15, 0.00000000000000e+00, 0.00000000000000e+00, 7.10542735760100e-15, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 7.10542735760100e-15, 0.00000000000000e+00, 7.10542735760100e-15, 0.00000000000000e+00, 7.10542735760100e-15],
[ 4.20000000000000e+01, 4.19999999639375e+01, 4.19999988488139e+01, 4.19999912936750e+01, 4.19999635200047e+01, 4.19998894842108e+01, 4.19997274516525e+01, 4.19994171129641e+01, 4.19988773470264e+01, 4.19980047152701e+01, 4.19966727497181e+01, 4.19947320731060e+01, 4.19920113644759e+01, 4.19883191588234e+01, 4.19834464456603e+01, 4.19771700096505e+01, 4.19692564375005e+01, 4.19594666994957e+01, 4.19475612016059e+01, 4.19333051946961e+01, 4.19164744205188e+01, 4.18968608690397e+01, 4.18742785174239e+01, 4.18485689169774e+01, 4.18196064902068e+01, 4.17873033961654e+01, 4.17516138193146e+01, 4.17125375367362e+01, 4.16701226226153e+01, 4.16244671594366e+01, 4.15757198439359e+01, 4.15240794033981e+01, 4.14697927741825e+01, 4.14131520379565e+01, 4.13544901594567e+01, 4.12941756192284e+01, 4.12326060818663e+01, 4.11702012810628e+01, 4.11073953341427e+01, 4.10446287186236e+01, 4.09823401507982e+01, 4.09209586017338e+01, 4.08608956708163e+01, 4.08025385131950e+01, 4.07462434877618e+01, 4.06923306592238e+01, 4.06410792537534e+01, 4.05927241345198e+01, 4.05474533324674e+01, 4.05054066398046e+01, 4.04666752490964e+01, 4.04313023995564e+01, 4.03992849737704e+01, 4.03705759722465e+01, 4.03450877794119e+01, 4.03226961226165e+01, 4.03032446151312e+01, 4.02865497649567e+01, 4.02724063236043e+01, 4.02605928430659e+01, 4.02508773053226e+01, 4.02430226873053e+01, 4.02367923256235e+01, 4.02319549499477e+01, 4.02282892619082e+01, 4.02255879478450e+01, 4.02236610286513e+01, 4.02223384680239e+01, 4.02214719812582e+01, 4.02209360096883e+01, 4.02206278502780e+01, 4.02204669548869e+01, 4.02203934385340e+01, 4.02203658597014e+01, 4.02203583575609e+01, 4.02203572502596e+01, 4.02203571050050e+01, 4.02203537207474e+01, 4.02203307918857e+01, 4.02202465026844e+01, 4.02200218164704e+01, 4.02195300852770e+01, 4.02185883194170e+01, 4.02169504499250e+01, 4.02143028664974e+01, 4.02102624561569e+01, 4.02043773041955e+01, 4.01961301498468e+01, 4.01849446155746e+01, 4.01701941519890e+01, 4.01512135614913e+01, 4.01273128840806e+01, 4.00977933494143e+01, 4.00619650209714e+01, 4.00191656815217e+01, 3.99687804347087e+01, 3.99102614269333e+01, 3.98431470302887e+01, 3.97670797772458e+01, 3.96818223104151e+01, 3.95872706177010e+01, 3.94834638765074e+01, 3.93705903394514e+01, 3.92489888608504e+01, 3.91191458811326e+01, 3.89816879377940e+01, 3.88373700302090e+01, 3.86870604006105e+01, 3.85317224756347e+01, 3.83723948206907e+01, 3.82101699843249e+01, 3.80461730571099e+01, 3.78815406570228e+01, 3.77174009059688e+01, 3.75548548067266e+01, 3.73949592886777e+01, 3.72387120788145e+01, 3.70870384770092e+01, 3.69407800685511e+01, 3.68006853843494e+01, 3.66674025097927e+01, 3.65414736378576e+01, 3.64233315543125e+01, 3.63132980298136e+01, 3.62115840753371e+01, 3.61182919955311e+01, 3.60334191516008e+01, 3.59568633232724e+01, 3.58884295393317e+01, 3.58278382283649e+01, 3.57747345251068e+01, 3.57286985523863e+01, 3.56892564833402e+01, 3.56558921730300e+01, 3.56280591330898e+01, 3.56051926084022e+01, 3.55867215023646e+01, 3.55720798887211e+01, 3.55607178449783e+01, 3.55521113467367e+01, 3.55457709752328e+01, 3.55412492128289e+01, 3.55381461333489e+01, 3.55361133355573e+01, 3.55348560175422e+01, 3.55341331454623e+01, 3.55337557297550e+01, 3.55335832828126e+01, 3.55335185915922e+01, 3.55335009939722e+01, 3.55334983965989e+01, 3.55334981673003e+01, 3.55334936743065e+01, 3.55334632336023e+01, 3.55333513301987e+01, 3.55330530356206e+01, 3.55324002191961e+01, 3.55311499745136e+01, 3.55289757161265e+01, 3.55254613421365e+01, 3.55200987896564e+01, 3.55122892316098e+01, 3.55013480736698e+01, 3.54865138081637e+01, 3.54669606667390e+01, 3.54418148858822e+01, 3.54101742610397e+01, 3.53711305202586e+01, 3.53237939034046e+01, 3.52673191967316e+01, 3.52009323550322e+01, 3.51239567554511e+01, 3.50358380780682e+01, 3.49361668060086e+01, 3.48246973861051e+01, 3.47013631901343e+01, 3.45662865631748e+01, 3.44197834345370e+01, 3.42623621923541e+01, 3.40947167796106e+01, 3.39177142496598e+01, 3.37323773100122e+01, 3.35398626612076e+01, 3.33414361685754e+01, 3.31384460485879e+01, 3.29322952755142e+01, 3.27244143081769e+01, 3.25162350233149e+01, 3.23091664727638e+01, 3.21045728193338e+01, 3.19037536027086e+01, 3.17079263660166e+01, 3.15182116295660e+01, 3.13356202035241e+01, 3.11610428535967e+01, 3.09952423477971e+01, 3.08388479059501e+01, 3.06923520465848e+01, 3.05561097857048e+01, 3.04303400982596e+01, 3.03151295140128e+01, 3.02104376895357e+01, 3.01161047782930e+01, 3.00318604095541e+01, 2.99573340809975e+01, 2.98920667658537e+01, 2.98355235302443e+01, 2.97871069479802e+01, 2.97461710876901e+01, 2.97120358311344e+01, 2.96840012633570e+01, 2.96613618571641e+01, 2.96434201589992e+01, 2.96294996735184e+01, 2.96189566427857e+01, 2.96111904253153e+01, 2.96056522017580e+01, 2.96018517685579e+01, 2.95993622280456e+01, 2.95978224418357e+01, 2.95969371817533e+01, 2.95964749857912e+01, 2.95962638022597e+01, 2.95961845796412e+01, 2.95961630291318e+01, 2.95961598483204e+01, 2.95961595931800e+01, 2.95961548845450e+01, 2.95961229828407e+01, 2.95960057086670e+01, 2.95956930977458e+01, 2.95950089511568e+01, 2.95936987085371e+01, 2.95914201228657e+01, 2.95877371541386e+01, 2.95821174283912e+01, 2.95739335272128e+01, 2.95624682796735e+01, 2.95469241219682e+01, 2.95264364691361e+01, 2.95000909081712e+01, 2.94669438747327e+01, 2.94260463207642e+01, 2.93764697243834e+01, 2.93173336455004e+01, 2.92478339016228e+01, 2.91672703399084e+01, 2.90750731249275e+01, 2.89708264561140e+01, 2.88542886806093e+01, 2.87254078781279e+01, 2.85843321620568e+01, 2.84314141583583e+01, 2.82672093801941e+01, 2.80924684974772e+01, 2.79081237899578e+01, 2.77152703509379e+01, 2.75151428558959e+01, 2.73090889060725e+01, 2.70985400839779e+01, 2.68849819041553e+01, 2.66699238052322e+01, 2.64548702153231e+01, 2.62412935488976e+01, 2.60306097830218e+01, 2.58241570408381e+01, 2.56231774046431e+01, 2.54288020082331e+01, 2.52420393283023e+01, 2.50637665091809e+01, 2.48947235089323e+01, 2.47355098386447e+01, 2.45865836703113e+01, 2.44482631027377e+01, 2.43207293924774e+01, 2.42040319734961e+01, 2.40980951028536e+01, 2.40027259792891e+01, 2.39176241869620e+01, 2.38423923175958e+01, 2.37765476207023e+01, 2.37195345231635e+01, 2.36707378461660e+01, 2.36294965296397e+01, 2.35951176528878e+01, 2.35668905166069e+01, 2.35441005282166e+01, 2.35260426120432e+01, 2.35120338513491e+01, 2.35014250633227e+01, 2.34936110133754e+01, 2.34880389932239e+01, 2.34842155190955e+01, 2.34817109517787e+01, 2.34801618978210e+01, 2.34792713186081e+01, 2.34788063481633e+01, 2.34785938975375e+01, 2.34785141996836e+01, 2.34784925199077e+01, 2.34784893200172e+01, 2.34784890732108e+01, 2.34784846413608e+01, 2.34784546149189e+01, 2.34783442344499e+01, 2.34780499999379e+01, 2.34774060713129e+01, 2.34761728571960e+01, 2.34740282444178e+01, 2.34705618643244e+01, 2.34652727264358e+01, 2.34575704745141e+01, 2.34467804328509e+01, 2.34321525098420e+01, 2.34128739104623e+01, 2.33880854791046e+01, 2.33569013512537e+01, 2.33184314409186e+01, 2.32718061375643e+01, 2.32162024409903e+01, 2.31508706367354e+01, 2.30751605206311e+01, 2.29885461310499e+01, 2.28906479508485e+01, 2.27812516035943e+01, 2.26603221905686e+01, 2.25280135904360e+01, 2.23846722609665e+01, 2.22308353264155e+01, 2.20672229881086e+01, 2.18947255434383e+01, 2.17143855268638e+01, 2.15273756862283e+01, 2.13349736721012e+01, 2.11385344407648e+01, 2.09394614454645e+01, 2.07391777065898e+01, 2.05390978014698e+01, 2.03406016958339e+01, 2.01450111591892e+01, 1.99535692847566e+01, 1.97674233996609e+01, 1.95876114335328e+01, 1.94150516384954e+01, 1.92505354339061e+01, 1.90947230852969e+01, 1.89481419085760e+01, 1.88111867026945e+01, 1.86841221418783e+01, 1.85670868910008e+01, 1.84600992381115e+01, 1.83630640637279e+01, 1.82757809867425e+01, 1.81979535420215e+01, 1.81291992552143e+01, 1.80690604857695e+01, 1.80170159091550e+01, 1.79724925033228e+01, 1.79348778924779e+01, 1.79035328838122e+01, 1.78778040114844e+01, 1.78570358789814e+01, 1.78405830689112e+01, 1.78278213713892e+01, 1.78181580716422e+01, 1.78110410370149e+01, 1.78059663553184e+01, 1.78024843015261e+01, 1.78002034482334e+01, 1.77987927859215e+01, 1.77979817797099e+01, 1.77975583568935e+01, 1.77973648904873e+01, 1.77972923143934e+01, 1.77972725719205e+01, 1.77972696579723e+01, 1.77972694373370e+01, 1.77972655288034e+01, 1.77972390479110e+01, 1.77971417013150e+01, 1.77968822107667e+01, 1.77963143207078e+01, 1.77952267381703e+01, 1.77933354069403e+01, 1.77902784697200e+01, 1.77856142160743e+01, 1.77788222489764e+01, 1.77693080269279e+01, 1.77564108500167e+01, 1.77394152555019e+01, 1.77175656713715e+01, 1.76900840463623e+01, 1.76561900360326e+01, 1.76151231830147e+01, 1.75661663943824e+01, 1.75086699009049e+01, 1.74420747935223e+01, 1.73659351829339e+01, 1.72799380279596e+01, 1.71839197328668e+01, 1.70778787237958e+01, 1.69619833749197e+01, 1.68365748560526e+01, 1.67021647012465e+01, 1.65594271367862e+01, 1.64091864415160e+01, 1.62523998292497e+01, 1.60901365316253e+01, 1.59235539123459e+01, 1.57538715542921e+01, 1.55823443244626e+01, 1.54102354338854e+01, 1.52387904678827e+01, 1.50692132668118e+01, 1.49026443936811e+01, 1.47401427432792e+01, 1.45826706429184e+01, 1.44310825858593e+01, 1.42861175434933e+01, 1.41483946373986e+01, 1.40184118285899e+01, 1.38965472038129e+01, 1.37830624067568e+01, 1.36781077698259e+01, 1.35817287405774e+01, 1.34938732555790e+01, 1.34143997828569e+01, 1.33430858231916e+01, 1.32796367231266e+01, 1.32236947037608e+01, 1.31748480463672e+01, 1.31326403976094e+01, 1.30965801640930e+01, 1.30661499597597e+01, 1.30408160526337e+01, 1.30200377326601e+01, 1.30032764932282e+01, 1.29900048890508e+01, 1.29797149059758e+01, 1.29719256574349e+01, 1.29661902105175e+01, 1.29621013443398e+01, 1.29592960557763e+01, 1.29574586529979e+01, 1.29563223147474e+01, 1.29556690409680e+01, 1.29553279754665e+01, 1.29551721403437e+01, 1.29551136812935e+01, 1.29550977790224e+01, 1.29550954318808e+01, 1.29550952556801e+01, 1.29550921543489e+01, 1.29550711423802e+01, 1.29549939002591e+01, 1.29547880018829e+01, 1.29543374021740e+01, 1.29534744645848e+01, 1.29519738550689e+01, 1.29495485951960e+01, 1.29458485273512e+01, 1.29404613974776e+01, 1.29329167030423e+01, 1.29226923831509e+01, 1.29092243418479e+01, 1.28919186936141e+01, 1.28701665028285e+01, 1.28433606597722e+01, 1.28109144006885e+01, 1.27722808472196e+01, 1.27269728223169e+01, 1.26745821080052e+01, 1.26147972579434e+01, 1.25474190759525e+01, 1.24723729289178e+01, 1.23897171822990e+01, 1.22996472264262e+01, 1.22024947925553e+01, 1.20987225232825e+01, 1.19889140407653e+01, 1.18737600230694e+01, 1.17540410281648e+01, 1.16306079738427e+01, 1.15043612737261e+01, 1.13762296372987e+01, 1.12471494685941e+01, 1.11180456570511e+01, 1.09898143660140e+01, 1.08633082146885e+01, 1.07393240432983e+01, 1.06185932700643e+01, 1.05017747070508e+01, 1.03894496064822e+01, 1.02821186588671e+01, 1.01802006523743e+01, 1.00840325192359e+01, 9.99387052850013e+00, 9.90989242541839e+00, 9.83220035871727e+00, 9.76082447336310e+00, 9.69572707603164e+00, 9.63680730298849e+00, 9.58390623607192e+00, 9.53681242284598e+00, 9.49526776250986e+00, 9.45897372024982e+00, 9.42759782958431e+00, 9.40078043497295e+00, 9.37814161575881e+00, 9.35928821791784e+00, 9.34382090304220e+00, 9.33134110585687e+00, 9.32145777409804e+00, 9.31379374971571e+00, 9.30799164008922e+00, 9.30371902405788e+00, 9.30067284147840e+00, 9.29858282755414e+00, 9.29721387447643e+00, 9.29636723235239e+00, 9.29588049763464e+00, 9.29562637837171e+00, 9.29551026921006e+00, 9.29546671265957e+00, 9.29545486422217e+00, 9.29545311541766e+00, 9.29545298453357e+00, 9.29545068614333e+00, 9.29543511421367e+00, 9.29537787018925e+00, 9.29522527891539e+00, 9.29489133805581e+00, 9.29425180676607e+00, 9.29313967340004e+00, 9.29134221560172e+00, 9.28859983526267e+00, 9.28460681430855e+00, 9.27901409348417e+00, 9.27143412352797e+00, 9.26144777500403e+00, 9.24861321909980e+00, 9.23247660776377e+00, 9.21258429029207e+00, 9.18849620953012e+00, 9.15980003084933e+00, 9.12612547914905e+00, 9.08715830220412e+00, 9.04265325116988e+00, 8.99244547758233e+00, 8.93645979419092e+00, 8.87471733368231e+00, 8.80733925947423e+00, 8.73454732685533e+00, 8.65666124862818e+00, 8.57409297411184e+00, 8.48733813232074e+00, 8.39696501099025e+00, 8.30360153850665e+00, 8.20792080488722e+00, 8.11062570182367e+00, 8.01243328131753e+00, 7.91405942660407e+00, 7.81620439482020e+00, 7.71953972470811e+00, 7.62469690357853e+00, 7.53225806030852e+00, 7.44274880669550e+00, 7.35663320491618e+00, 7.27431071269505e+00, 7.19611486540681e+00, 7.12231340391794e+00, 7.05310954809105e+00, 6.98864414075251e+00, 6.92899843354907e+00, 6.87419734183403e+00, 6.82421304999602e+00, 6.77896889448092e+00, 6.73834348577677e+00, 6.70217505229954e+00, 6.67026599966304e+00, 6.64238768026351e+00, 6.61828536259698e+00, 6.59768337917551e+00, 6.58029041792359e+00, 6.56580490588596e+00, 6.55392041720172e+00, 6.54433102081535e+00, 6.53673646855861e+00, 6.53084711232931e+00, 6.52638843136305e+00, 6.52310504815473e+00, 6.52076411528954e+00, 6.51915796576068e+00, 6.51810593628264e+00, 6.51745529613218e+00, 6.51708124213380e+00, 6.51688595208478e+00, 6.51679672241367e+00, 6.51676324927786e+00, 6.51675414377278e+00, 6.51675279981929e+00, 6.51675269919743e+00, 6.51675093172403e+00, 6.51673895683321e+00, 6.51669493589036e+00, 6.51657759239201e+00, 6.51632078989354e+00, 6.51582898574863e+00, 6.51497374472450e+00, 6.51359147604052e+00, 6.51148253341031e+00, 6.50841178933980e+00, 6.50411076100384e+00, 6.49828132417406e+00, 6.49060100284557e+00, 6.48072976500244e+00, 6.46831819001707e+00, 6.45301680252228e+00, 6.43448629485551e+00, 6.41240829057088e+00, 6.38649624158117e+00, 6.35650600847260e+00, 6.32224565447602e+00, 6.28358399417840e+00, 6.24045748146206e+00, 6.19287509689123e+00, 6.14092099812245e+00, 6.08475481906225e+00, 6.02460963244577e+00, 5.96078771296346e+00, 5.89365434178166e+00, 5.82362996947287e+00, 5.75118109923739e+00, 5.67681026753018e+00, 5.60104549080914e+00, 5.52442952323666e+00, 5.44750923855421e+00, 5.37082541540373e+00, 5.29490317130448e+00, 5.22024325580314e+00, 5.14731437610826e+00, 5.07654668710001e+00, 5.00832653164392e+00, 4.94299246814267e+00, 4.88083257328421e+00, 4.82208296281686e+00, 4.76692743556919e+00, 4.71549811856198e+00, 4.66787697525704e+00, 4.62409803458118e+00, 4.58415020388486e+00, 4.54798054206138e+00, 4.51549788681741e+00, 4.48657674965299e+00, 4.46106141086504e+00, 4.43877016273131e+00, 4.41949966049011e+00, 4.40302934701451e+00, 4.38912591805082e+00, 4.37754779100857e+00, 4.36804953251060e+00, 4.36038618958031e+00, 4.35431745804826e+00, 4.34961161120115e+00, 4.34604910351313e+00, 4.34342575993902e+00, 4.34155546181564e+00, 4.34027224657190e+00, 4.33943175034098e+00, 4.33891193982273e+00, 4.33861310149699e+00, 4.33845708128032e+00, 4.33838579442804e+00, 4.33835905228052e+00, 4.33835177776912e+00, 4.33835070406651e+00, 4.33835062346510e+00, 4.33834920481255e+00, 4.33833959323474e+00, 4.33830426008698e+00, 4.33821007511026e+00, 4.33800395472237e+00, 4.33760921385228e+00, 4.33692276972806e+00, 4.33581332997143e+00, 4.33412067861016e+00, 4.33165615145280e+00, 4.32820436568934e+00, 4.32352623656165e+00, 4.31736327568543e+00, 4.30944312079898e+00, 4.29948619582478e+00, 4.28721334458543e+00, 4.27235422381516e+00, 4.25465618479078e+00, 4.23389332236363e+00, 4.20987533029647e+00, 4.18245577750059e+00, 4.15153941536121e+00, 4.11708814500648e+00, 4.07912531659239e+00, 4.03773809984954e+00, 3.99307775341249e+00, 3.94535772479152e+00, 3.89484962632227e+00, 3.84187724676332e+00, 3.78680886450422e+00, 3.73004821793005e+00, 3.67202455387580e+00, 3.61318221087555e+00, 3.55397019744444e+00, 3.49483219751178e+00, 3.43619737912992e+00, 3.37847230521467e+00, 3.32203415465405e+00, 3.26722536767819e+00, 3.21434973944541e+00, 3.16366990741250e+00, 3.11540611609635e+00, 3.06973609975592e+00, 3.02679589948185e+00, 2.98668142446327e+00, 2.94945057479890e+00, 2.91512576145884e+00, 2.88369668407281e+00, 2.85512325559477e+00, 2.82933859157539e+00, 2.80625200842383e+00, 2.78575199797750e+00, 2.76770916385424e+00, 2.75197911788571e+00, 2.73840534230313e+00, 2.72682202549974e+00, 2.71705687664557e+00, 2.70893391794232e+00, 2.70227624383954e+00, 2.69690872521792e+00, 2.69266062461443e+00, 2.68936807730724e+00, 2.68687638373899e+00, 2.68504205245598e+00, 2.68373453035946e+00, 2.68283755917272e+00, 2.68225010379998e+00, 2.68188680945640e+00, 2.68167795944653e+00, 2.68156892328758e+00, 2.68151910430668e+00, 2.68150041557448e+00, 2.68149533179445e+00, 2.68149458143946e+00, 2.68149452487143e+00, 2.68149352603091e+00, 2.68148675874417e+00, 2.68146188155252e+00, 2.68139556863299e+00, 2.68125044716945e+00, 2.68097253270125e+00, 2.68048926995429e+00, 2.67970827706839e+00, 2.67851687974712e+00, 2.67678250777918e+00, 2.67435400845403e+00, 2.67106390839232e+00, 2.66673162620872e+00, 2.66116760268003e+00, 2.65417827294424e+00, 2.64557175796057e+00, 2.63516410243514e+00, 2.62278583723060e+00, 2.60828860047515e+00, 2.59155151827707e+00, 2.57248702824366e+00, 2.55104583128296e+00, 2.52722068233850e+00, 2.50104877952793e+00, 2.47261258177339e+00, 2.44203897285459e+00, 2.40949678788593e+00, 2.37519281781830e+00, 2.33936649935078e+00, 2.30228357286574e+00, 2.26422904274460e+00, 2.22549979852236e+00, 2.18639725092474e+00, 2.14722030630050e+00, 2.10825895144526e+00, 2.06978865525960e+00, 2.03206572173856e+00, 1.99532365771854e+00, 1.95977055463547e+00, 1.92558743052784e+00, 1.89292743896885e+00, 1.86191582605262e+00, 1.83265050407774e+00, 1.80520310926725e+00, 1.77962041831779e+00, 1.75592601222247e+00, 1.73412209325117e+00, 1.71419138010734e+00, 1.69609902541917e+00, 1.67979451757804e+00, 1.66521354459522e+00, 1.65227981050725e+00, 1.64090680458087e+00, 1.63099953001528e+00, 1.62245620204592e+00, 1.61516992550268e+00, 1.60903035927821e+00, 1.60392537025749e+00, 1.59974267260690e+00, 1.59637144058149e+00, 1.59370387492959e+00, 1.59163669534272e+00, 1.59007252500059e+00, 1.58892112880662e+00, 1.58810046497850e+00, 1.58753751062967e+00, 1.58716882598542e+00, 1.58694082878674e+00, 1.58680975984799e+00, 1.58674133202548e+00, 1.58671006723927e+00, 1.58669833881155e+00, 1.58669514840174e+00, 1.58669467750418e+00, 1.58669464181080e+00, 1.58669400899835e+00, 1.58668972160471e+00, 1.58667396074923e+00, 1.58663194853369e+00, 1.58654000805294e+00, 1.58636393956059e+00, 1.58605778165826e+00, 1.58556302073014e+00, 1.58480830507899e+00, 1.58370971167066e+00, 1.58217160220239e+00, 1.58008809046322e+00, 1.57734512388105e+00, 1.57382315832638e+00, 1.56940037681279e+00, 1.56395637059665e+00, 1.55737616709180e+00, 1.54955445559527e+00, 1.54039983239105e+00, 1.52983886506608e+00, 1.51781976546830e+00, 1.50431546464229e+00, 1.48932590306524e+00, 1.47287938564605e+00, 1.45503290136617e+00, 1.43587136833764e+00, 1.41550583108685e+00, 1.39407070181282e+00, 1.37172019495054e+00, 1.34862414916667e+00, 1.32496345909438e+00, 1.30092534889141e+00, 1.27669871148004e+00, 1.25246971343412e+00, 1.22841782969811e+00, 1.20471242921852e+00, 1.18150998681452e+00, 1.15895195242410e+00, 1.13716326957839e+00, 1.11625150285012e+00, 1.09630651026580e+00, 1.07740058145828e+00, 1.05958895508876e+00, 1.04291062867109e+00, 1.02738937897790e+00, 1.01303492018616e+00, 9.99844138377540e-01, 9.87802353645506e-01, 9.76884573793200e-01, 9.67056715578572e-01, 9.58276780049530e-01, 9.50495977284533e-01, 9.43659802563516e-01, 9.37709070529735e-01, 9.32580916280493e-01, 9.28209772653872e-01, 9.24528331467990e-01, 9.21468493400560e-01, 9.18962306935420e-01, 9.16942891775726e-01, 9.15345336812166e-01, 9.14107557646624e-01, 9.13171094313185e-01, 9.12481826670849e-01, 9.11990583343297e-01, 9.11653620303589e-01, 9.11432947342163e-01, 9.11296484640999e-01, 9.11218037260895e-01, 9.11177082144626e-01, 9.11158369753025e-01, 9.11151350143797e-01, 9.11149440645069e-01, 9.11149158807241e-01, 9.11149137308769e-01, 9.11148754371325e-01, 9.11146159916647e-01, 9.11136622469485e-01, 9.11111199464962e-01, 9.11055563505464e-01, 9.10949020363610e-01, 9.10763760898845e-01, 9.10464385918601e-01, 9.10007739220120e-01, 9.09343079132398e-01, 9.08412612249712e-01, 9.07152404077586e-01, 9.05493669462395e-01, 9.03364430641346e-01, 9.00691512605157e-01, 8.97402824790412e-01, 8.93429856069076e-01, 8.88710288314535e-01, 8.83190614705022e-01, 8.76828634846959e-01, 8.69595692176745e-01, 8.61478521914762e-01, 8.52480591261013e-01, 8.42622837568442e-01, 8.31943743628329e-01, 8.20498729396654e-01, 8.08358882897103e-01, 7.95609095491017e-01, 7.82345704024777e-01, 7.68673770941298e-01, 7.54704150732799e-01, 7.40550495957883e-01, 7.26326348759706e-01, 7.12142446039060e-01, 6.98104340787310e-01, 6.84310411752222e-01, 6.70850301871425e-01, 6.57803795721357e-01, 6.45240119940318e-01, 6.33217629758576e-01, 6.21783830165128e-01, 6.10975671881171e-01, 6.00820059639098e-01, 5.91334512327315e-01, 5.82527920194764e-01, 5.74401352312194e-01, 5.66948876741854e-01, 5.60158365420641e-01, 5.54012264872021e-01, 5.48488321995883e-01, 5.43560261010100e-01, 5.39198412959166e-01, 5.35370303017146e-01, 5.32041203139343e-01, 5.29174658562233e-01, 5.26732996363336e-01, 5.24677822947382e-01, 5.22970515123326e-01, 5.21572706605582e-01, 5.20446768559019e-01, 5.19556279480597e-01, 5.18866476550777e-01, 5.18344677874815e-01, 5.17960663022642e-01, 5.17686998181766e-01, 5.17499292203873e-01, 5.17376370911244e-01, 5.17300359195751e-01, 5.17256663552160e-01, 5.17233851516572e-01, 5.17223428743883e-01, 5.17219518839994e-01, 5.17218455254913e-01, 5.17218298272127e-01, 5.17218286213037e-01, 5.17218070305120e-01, 5.17216607499272e-01, 5.17211230099790e-01, 5.17196896154353e-01, 5.17165527818726e-01, 5.17105458122600e-01, 5.17001010244776e-01, 5.16832231825525e-01, 5.16574804901275e-01, 5.16200149435875e-01, 5.15675734756442e-01, 5.14965608063441e-01, 5.14031142216723e-01, 5.12831995989423e-01, 5.11327268965964e-01, 5.09476820602103e-01, 5.07242709409169e-01, 5.04590694899420e-01, 5.01491733274577e-01, 4.97923389469001e-01, 4.93871084637753e-01, 4.89329100775713e-01, 4.84301273564739e-01, 4.78801320691730e-01, 4.72852774794198e-01, 4.66488516044045e-01, 4.59749926684644e-01, 4.52685715758960e-01, 4.45350484045100e-01, 4.37803114574777e-01, 4.30105081597134e-01, 4.22318770039404e-01, 4.14505889025499e-01, 4.06726048349711e-01, 3.99035548083020e-01, 3.91486411067902e-01, 3.84125668188503e-01, 3.76994888840308e-01, 3.70129935239902e-01, 3.63560909741686e-01, 3.57312259192213e-01, 3.51402999104742e-01, 3.45847022313833e-01, 3.40653460865092e-01, 3.35827075301857e-01, 3.31368651431916e-01, 3.27275390461963e-01, 3.23541283635983e-01, 3.20157466940600e-01, 3.17112554932983e-01, 3.14392955294444e-01, 3.11983167380589e-01, 3.09866068924215e-01, 3.08023195263679e-01, 3.06435015128336e-01, 3.05081206218219e-01, 3.03940932665974e-01, 3.02993125055607e-01, 3.02216762094901e-01, 3.01591151391307e-01, 3.01096205179256e-01, 3.00712705400606e-01, 3.00422551369891e-01, 3.00208982468028e-01, 3.00056767997987e-01, 2.99952356559984e-01, 2.99883978088943e-01, 2.99841693014450e-01, 2.99817384785470e-01, 2.99804694131403e-01, 2.99798895765399e-01, 2.99796720614097e-01, 2.99796128921692e-01, 2.99796041589190e-01, 2.99796034829299e-01, 2.99795913132463e-01, 2.99795088619483e-01, 2.99792057636182e-01, 2.99783978252229e-01, 2.99766297249448e-01, 2.99732438049787e-01, 2.99673562986861e-01, 2.99578422326782e-01, 2.99433301232206e-01, 2.99222074282511e-01, 2.98926375016208e-01, 2.98525885035794e-01, 2.97998743327668e-01, 2.97322071495273e-01, 2.96472604613626e-01, 2.95427410596417e-01, 2.94164673743573e-01, 2.92664511138753e-01, 2.90909784590063e-01, 2.88886866737570e-01, 2.86586318621829e-01, 2.84003438057454e-01, 2.81138643872019e-01, 2.77997670278076e-01, 2.74591557660258e-01, 2.70936439738847e-01, 2.67053140959156e-01, 2.62966610502211e-01, 2.58705229125574e-01, 2.54300031122547e-01, 2.49783885582886e-01, 2.45190678994661e-01, 2.40554535730020e-01, 2.35909105163188e-01, 2.31286935291692e-01, 2.26718943931665e-01, 2.22233990757871e-01, 2.17858547253158e-01, 2.13616457262646e-01, 2.09528778254708e-01, 2.05613692305917e-01, 2.01886475874446e-01, 1.98359518210310e-01, 1.95042379431925e-01, 1.91941880614600e-01, 1.89062219521681e-01, 1.86405106786111e-01, 1.83969918384278e-01, 1.81753861151925e-01, 1.79752148871216e-01, 1.77958187128603e-01, 1.76363765681856e-01, 1.74959257479635e-01, 1.73733823718401e-01, 1.72675624378677e-01, 1.71772033551136e-01, 1.71009858534774e-01, 1.70375561190745e-01, 1.69855479396404e-01, 1.69436045730180e-01, 1.69103999785328e-01, 1.68846589850993e-01, 1.68651759184613e-01, 1.68508311804850e-01, 1.68406052719526e-01, 1.68335897804184e-01, 1.68289949177712e-01, 1.68261532856199e-01, 1.68245196669439e-01, 1.68236667816956e-01, 1.68232770938673e-01, 1.68231309089137e-01, 1.68230911431017e-01, 1.68230852737551e-01, 1.68230848157584e-01, 1.68230765229964e-01, 1.68230203383487e-01, 1.68228137981352e-01, 1.68222632415677e-01, 1.68210583812289e-01, 1.68187510033263e-01, 1.68147386846506e-01, 1.68082543306191e-01, 1.67983622406702e-01, 1.67839612865492e-01, 1.67637956378421e-01, 1.67364732756070e-01, 1.67004922934844e-01, 1.66542746902429e-01, 1.65962070125381e-01, 1.65246868219938e-01, 1.64381735592166e-01, 1.63352419887303e-01, 1.62146360735321e-01, 1.60753208867305e-01, 1.59165300618078e-01, 1.57378063418513e-01, 1.55390330271835e-01, 1.53204545345552e-01, 1.50826848410155e-01, 1.48267032469931e-01, 1.45538375949670e-01, 1.42657357600292e-01, 1.39643268297775e-01, 1.36517738677583e-01, 1.33304204838673e-01, 1.30027336063073e-01, 1.26712448723140e-01, 1.23384929441649e-01, 1.20069688346987e-01, 1.16790660190787e-01, 1.13570367397017e-01, 1.10429555063781e-01, 1.07386903806862e-01, 1.04458822306476e-01, 1.01659317705220e-01, 9.89999387653882e-02, 9.64897840783578e-02, 9.41355657361242e-02, 9.19417178865345e-02, 8.99105393409485e-02, 8.80423600204026e-02, 8.63357220307748e-02, 8.47875676670071e-02, 8.33934282720463e-02, 8.21476095329388e-02, 8.10433704799536e-02, 8.00730948061528e-02, 7.92284543590611e-02, 7.85005654104580e-02, 7.78801387720165e-02, 7.73576249186121e-02, 7.69233550743493e-02, 7.65676786167973e-02, 7.62810967184180e-02, 7.60543910784506e-02, 7.58787461882591e-02, 7.57458625228864e-02, 7.56480578229315e-02, 7.55783531730501e-02, 7.55305405523501e-02, 7.54992288946781e-02, 7.54798660667291e-02, 7.54687351260230e-02, 7.54629239928093e-02, 7.54602688859833e-02, 7.54592728729833e-02, 7.54590019319465e-02, 7.54589619425303e-02, 7.54589587972418e-02, 7.54589014714782e-02, 7.54585131040812e-02, 7.54570854289938e-02, 7.54532798263057e-02, 7.54449516515194e-02, 7.54290033099139e-02, 7.54012724799296e-02, 7.53564614098963e-02, 7.52881127021183e-02, 7.51886363099672e-02, 7.50493915157762e-02, 7.48608263207586e-02, 7.46126749576135e-02, 7.42942118963086e-02, 7.38945580578019e-02, 7.34030317318926e-02, 7.28095333216530e-02, 7.21049496447534e-02, 7.12815604613013e-02, 7.03334275164309e-02, 6.92567450003381e-02, 6.80501304561290e-02, 6.67148365851773e-02, 6.52548678171740e-02, 6.36769902713539e-02, 6.19906296855817e-02, 6.02076590919793e-02, 5.83420847332050e-02, 5.64096456258838e-02, 5.44273476745119e-02, 5.24129570616836e-02, 5.03844801578737e-02, 4.83596567701881e-02, 4.63399980152204e-02, 4.43744964711215e-02, 4.24596336107470e-02, 4.06080958764576e-02, 3.88306883919398e-02, 3.71362574777265e-02, 3.55316908015766e-02, 3.40219839523570e-02, 3.26103599444210e-02, 3.12984272521248e-02, 3.00863622753624e-02, 2.89731033967797e-02, 2.79565457127973e-02, 2.70337278083730e-02, 2.62010043390646e-02, 2.54542004732272e-02, 2.47887462872356e-02, 2.41997909082130e-02, 2.36822975246724e-02, 2.32311213342342e-02, 2.28410730926640e-02, 2.25069712077115e-02, 2.22236853289601e-02, 2.19861741672815e-02, 2.17895198802083e-02, 2.16289608299078e-02, 2.14999239024891e-02, 2.13980569202832e-02, 2.13192610286725e-02, 2.12597223452349e-02, 2.12159416641100e-02, 2.11847606532913e-02, 2.11633827922171e-02, 2.11493872885171e-02, 2.11407343821648e-02, 2.11357607750173e-02, 2.11331643808208e-02, 2.11319781301000e-02, 2.11315331367388e-02, 2.11314120887209e-02, 2.11313942223157e-02, 2.11313928038734e-02, 2.11313668089161e-02, 2.11311906898288e-02, 2.11305432657198e-02, 2.11288175346382e-02, 2.11250411384256e-02, 2.11178101296028e-02, 2.11052392298825e-02, 2.10849317436725e-02, 2.10539722293603e-02, 2.10089448648887e-02, 2.09459800831482e-02, 2.08608313991405e-02, 2.07489833208996e-02, 2.06057897726524e-02, 2.04266405591440e-02, 2.02071511289716e-02, 1.99433683998862e-02, 1.96319829204370e-02, 1.92705354592277e-02, 1.88576045755227e-02, 1.83929611640394e-02, 1.78776766539315e-02, 1.73141736240865e-02, 1.67062110521456e-02, 1.60588010197031e-02, 1.53780590372605e-02, 1.46709956578519e-02, 1.39452620720633e-02, 1.32088662920111e-02, 1.24698788388855e-02, 1.17361472593274e-02, 1.10150372904130e-02, 1.03132153157617e-02, 9.63648236908199e-03, 8.98966494426645e-03, 8.37656288774014e-03, 7.79995023958691e-03, 7.26162147055476e-03, 6.76247336244984e-03, 6.30261183706438e-03, 5.88147322068977e-03, 5.49795048994639e-03, 5.15051667566979e-03, 4.83733950544973e-03, 4.55638328926256e-03, 4.30549580990203e-03, 4.08247945881612e-03, 3.88514700566844e-03, 3.71136320352150e-03, 3.55907394239807e-03, 3.42632492225192e-03, 3.31127187980229e-03, 3.21218433037274e-03, 3.12744462376457e-03, 3.05554390052976e-03, 2.99507629797802e-03, 2.94473251114624e-03, 2.90329357113706e-03, 2.86962546761060e-03, 2.84267501628221e-03, 2.82146715866816e-03, 2.80510368435216e-03, 2.79276319051291e-03, 2.78370194706667e-03, 2.77725522538930e-03, 2.77283858114430e-03, 2.76994856183421e-03, 2.76816233809285e-03, 2.76713583076571e-03, 2.76660001569695e-03, 2.76635522344293e-03, 2.76626339763908e-03, 2.76623841919843e-03, 2.76623473244797e-03, 2.76623443706114e-03, 2.76622899014001e-03, 2.76619208661799e-03, 2.76605642878280e-03, 2.76569484072553e-03, 2.76490364574755e-03, 2.76338891704597e-03, 2.76075636881501e-03, 2.75650570420538e-03, 2.75003028129106e-03, 2.74062296713140e-03, 2.72748898889574e-03, 2.70976641787257e-03, 2.68655460238421e-03, 2.65695037510483e-03, 2.62009119901307e-03, 2.57520361587638e-03, 2.52165449141728e-03, 2.45900172027547e-03, 2.38704039993110e-03, 2.30584015375257e-03, 2.21576941205370e-03, 2.11750312984549e-03, 2.01201163795162e-03, 1.90052999693791e-03, 1.78450916694572e-03, 1.66555225444398e-03, 1.54534075865342e-03, 1.42555684661261e-03, 1.30780805058350e-03, 1.19356034907980e-03, 1.08408444476993e-03, 9.80418395798210e-04, 8.83703595427221e-04, 7.93403464513176e-04, 7.10872979311716e-04, 6.36102823790452e-04, 5.68515394839153e-04, 5.07892837390500e-04, 4.53878931025997e-04, 4.06330541451894e-04, 3.64593748940933e-04, 3.27993448365760e-04, 2.96025783796949e-04, 2.68217205452004e-04, 2.44178649367655e-04, 2.23546243489636e-04, 2.05799140694296e-04, 1.90525292111943e-04, 1.77396364094307e-04, 1.66138720608444e-04, 1.56522437414708e-04, 1.48349474947906e-04, 1.41441675630460e-04, 1.35634447040900e-04, 1.30779528170422e-04, 1.26748590183468e-04, 1.23431068288937e-04, 1.20729123750607e-04, 1.18554161147699e-04, 1.16826153629290e-04, 1.15473890886619e-04, 1.14434554291165e-04, 1.13652766266625e-04, 1.13079687949380e-04, 1.12672402551084e-04, 1.12393560286250e-04, 1.12211177096810e-04, 1.12098480720135e-04, 1.12033725855554e-04, 1.11999927881357e-04, 1.11984487560335e-04, 1.11978695727544e-04, 1.11977120248241e-04, 1.11976887701815e-04, 1.11976868836905e-04, 1.11976517885637e-04, 1.11974140175164e-04, 1.11965400066083e-04, 1.11942106855167e-04, 1.11891153999011e-04, 1.11793663911897e-04, 1.11624409647959e-04, 1.11351596522979e-04, 1.10937088116714e-04, 1.10337134927363e-04, 1.09503588852533e-04, 1.08385479173023e-04, 1.06930813387862e-04, 1.05088719891455e-04, 1.02812365803118e-04, 1.00062969984549e-04, 9.68148228253085e-05, 9.30607054030475e-05, 8.88154535516605e-05, 8.41133512068382e-05, 7.90000879149488e-05, 7.35300927203753e-05, 6.77736148162467e-05, 6.18417560218631e-05, 5.60276054386577e-05, 5.03369112507812e-05, 4.47187488887835e-05, 3.92126232426904e-05, 3.44005459353980e-05, 2.99071376943516e-05, 2.56767121342705e-05, 2.17668125586101e-05, 1.86924626177642e-05, 1.57716333788471e-05, 1.34077119042786e-05, 1.17471373570766e-05, 1.03043497361455e-05, 8.98307279584287e-06, 7.72191432218960e-06, 6.51271179918922e-06, 5.39020924605893e-06, 4.40794482159390e-06, 3.61170987872583e-06, 3.01475328257084e-06, 2.58122431517904e-06, 2.24646618818269e-06, 1.98712845644877e-06, 1.81543020261188e-06, 1.70040350155887e-06, 1.57538947576086e-06, 1.40483445676409e-06, 1.25765657799093e-06, 1.13062343132242e-06, 9.85614917681232e-07, 8.41557124431347e-07, 7.11954285748106e-07, 6.01549814405189e-07, 5.09449947116991e-07, 4.32252029725078e-07, 3.65857168560524e-07, 3.06716906095517e-07, -1.42108547152020e-14, -1.42108547152020e-14, -2.13162820728030e-14, -7.10542735760100e-15, 0.00000000000000e+00, -1.42108547152020e-14, 0.00000000000000e+00, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15],
[ 4.30000000000000e+01, 4.29999999627168e+01, 4.29999988098465e+01, 4.29999909989683e+01, 4.29999622851668e+01, 4.29998857432780e+01, 4.29997182259378e+01, 4.29993973822612e+01, 4.29988393450103e+01, 4.29979371737647e+01, 4.29965601185603e+01, 4.29945537435712e+01, 4.29917409246277e+01, 4.29879237087183e+01, 4.29828859991547e+01, 4.29763970077022e+01, 4.29682153954599e+01, 4.29580940080693e+01, 4.29457850980629e+01, 4.29310459175635e+01, 4.29136445575147e+01, 4.28933659042960e+01, 4.28700175800743e+01, 4.28434357288158e+01, 4.28134905052250e+01, 4.27800911192868e+01, 4.27431902855442e+01, 4.27027879254050e+01, 4.26589339746920e+01, 4.26117301594581e+01, 4.25613306225298e+01, 4.25079413122183e+01, 4.24518180829096e+01, 4.23932635032354e+01, 4.23326224184534e+01, 4.22702763658350e+01, 4.22066369911595e+01, 4.21421386568881e+01, 4.20772304649835e+01, 4.20123679374768e+01, 4.19480046049124e+01, 4.18845837471760e+01, 4.18225305144870e+01, 4.17622446308682e+01, 4.17040938509196e+01, 4.16484083059815e+01, 4.15954758402267e+01, 4.15455384028367e+01, 4.14987895305550e+01, 4.14553729263481e+01, 4.14153821148977e+01, 4.13788611340730e+01, 4.13458062030161e+01, 4.13161682915238e+01, 4.12898565015596e+01, 4.12667421595716e+01, 4.12466635076026e+01, 4.12294308718628e+01, 4.12148321796102e+01, 4.12026386890722e+01, 4.11926107931097e+01, 4.11845037557722e+01, 4.11780732422348e+01, 4.11730805072060e+01, 4.11692971149980e+01, 4.11665090761765e+01, 4.11645203009788e+01, 4.11631552882610e+01, 4.11622609901537e+01, 4.11617078162762e+01, 4.11613897665233e+01, 4.11612237072721e+01, 4.11611478314609e+01, 4.11611193675062e+01, 4.11611116245904e+01, 4.11611104817511e+01, 4.11611103318347e+01, 4.11611068389620e+01, 4.11610831742169e+01, 4.11609961798289e+01, 4.11607642825662e+01, 4.11602567700985e+01, 4.11592847809389e+01, 4.11575943520286e+01, 4.11548618163415e+01, 4.11506917839149e+01, 4.11446178742242e+01, 4.11361062966416e+01, 4.11245622997708e+01, 4.11093394308670e+01, 4.10897514645918e+01, 4.10650867773087e+01, 4.10346248601146e+01, 4.09976545816597e+01, 4.09534937311394e+01, 4.09015092934700e+01, 4.08411378343964e+01, 4.07719053069420e+01, 4.06934455389658e+01, 4.06055166345883e+01, 4.05080145322098e+01, 4.04009830211886e+01, 4.02846196370624e+01, 4.01592770334365e+01, 4.00254596591036e+01, 3.98838158323437e+01, 3.97351255723219e+01, 3.95802847875231e+01, 3.94202866030645e+01, 3.92562007113829e+01, 3.90891516467000e+01, 3.89202968202547e+01, 3.87508050305152e+01, 3.85818360076262e+01, 3.84145213921287e+01, 3.82499474072532e+01, 3.80891393756375e+01, 3.79330481590339e+01, 3.77825385589630e+01, 3.76383796977256e+01, 3.75012373916557e+01, 3.73716685225804e+01, 3.72501174031861e+01, 3.71369141152257e+01, 3.70322747771369e+01, 3.69363036722463e+01, 3.68489971432038e+01, 3.67702491347771e+01, 3.66998582466582e+01, 3.66375361403428e+01, 3.65829171285868e+01, 3.65355687612593e+01, 3.64950032066847e+01, 3.64606892123514e+01, 3.64320644134107e+01, 3.64085477425562e+01, 3.63895516820998e+01, 3.63744940900867e+01, 3.63628093290031e+01, 3.63539584297934e+01, 3.63474380369393e+01, 3.63427879031714e+01, 3.63395967352468e+01, 3.63375062346347e+01, 3.63362132277214e+01, 3.63354698373827e+01, 3.63350817090959e+01, 3.63349043674421e+01, 3.63348378400391e+01, 3.63348197429322e+01, 3.63348170718357e+01, 3.63348168360287e+01, 3.63348122155068e+01, 3.63347809107811e+01, 3.63346658311433e+01, 3.63343590698934e+01, 3.63336877243767e+01, 3.63324019943128e+01, 3.63301660266968e+01, 3.63265519145862e+01, 3.63210371866860e+01, 3.63130060440390e+01, 3.63017545080507e+01, 3.62864995392424e+01, 3.62663920677562e+01, 3.62405337450844e+01, 3.62079970838436e+01, 3.61678485028896e+01, 3.61191736453039e+01, 3.60611041955747e+01, 3.59928453000713e+01, 3.59137026025901e+01, 3.58231078544498e+01, 3.57206420541743e+01, 3.56060551197946e+01, 3.54792811981441e+01, 3.53404488680928e+01, 3.51898856944535e+01, 3.50281168313851e+01, 3.48558576522913e+01, 3.46740006872982e+01, 3.44835974610051e+01, 3.42858361121918e+01, 3.40820159029956e+01, 3.38735198457470e+01, 3.36617866632610e+01, 3.34482831542293e+01, 3.32344777954683e+01, 3.30218161380022e+01, 3.28116983081912e+01, 3.26054587523812e+01, 3.24043482757665e+01, 3.22095184060313e+01, 3.20220081271809e+01, 3.18427330467371e+01, 3.16724770594378e+01, 3.15118865463728e+01, 3.13614671049452e+01, 3.12215827525195e+01, 3.10924574957217e+01, 3.09741791156703e+01, 3.08667049903403e+01, 3.07698697584446e+01, 3.06833946217516e+01, 3.06068980806058e+01, 3.05399078964957e+01, 3.04818740725657e+01, 3.04321826359233e+01, 3.03901699937356e+01, 3.03551376189948e+01, 3.03263668030768e+01, 3.03031331932411e+01, 3.02847208169166e+01, 3.02704352840412e+01, 3.02596158567491e+01, 3.02516460847262e+01, 3.02459627262496e+01, 3.02420627099971e+01, 3.02395079408307e+01, 3.02379278125283e+01, 3.02370193594904e+01, 3.02365450546405e+01, 3.02363283384410e+01, 3.02362470403250e+01, 3.02362249252297e+01, 3.02362216610866e+01, 3.02362213992619e+01, 3.02362165672689e+01, 3.02361838297961e+01, 3.02360634832509e+01, 3.02357426825311e+01, 3.02350406128480e+01, 3.02336960456652e+01, 3.02313577702494e+01, 3.02275783296059e+01, 3.02218114170749e+01, 3.02134132066182e+01, 3.02016477942943e+01, 3.01856968190437e+01, 3.01646732066738e+01, 3.01376388420669e+01, 3.01036258231619e+01, 3.00616607905718e+01, 3.00107916656831e+01, 2.99501159770832e+01, 2.98788098213562e+01, 2.97961564017702e+01, 2.97015730288295e+01, 2.95946354599365e+01, 2.94750985081711e+01, 2.93429119649848e+01, 2.91982310562466e+01, 2.90414208787516e+01, 2.88730545337364e+01, 2.86939049699976e+01, 2.85049308532883e+01, 2.83072570695942e+01, 2.81021507251015e+01, 2.78909937034188e+01, 2.76752529628921e+01, 2.74564497927398e+01, 2.72361291951271e+01, 2.70158304310741e+01, 2.67970595809433e+01, 2.65812647512766e+01, 2.63698143365908e+01, 2.61639785415482e+01, 2.59649142025472e+01, 2.57736528260314e+01, 2.55910916829924e+01, 2.54179877582708e+01, 2.52549543392275e+01, 2.51024600308013e+01, 2.49608299943659e+01, 2.48302492203652e+01, 2.47107676562883e+01, 2.46023070208836e+01, 2.45046691423049e+01, 2.44175456620170e+01, 2.43405289474166e+01, 2.42731240536128e+01, 2.42147615679334e+01, 2.41648111589649e+01, 2.41225956353379e+01, 2.40874052987834e+01, 2.40585123527858e+01, 2.40351851047380e+01, 2.40167016786965e+01, 2.40023629407197e+01, 2.39915043323743e+01, 2.39835063129053e+01, 2.39778031286733e+01, 2.39738896606912e+01, 2.39713261472461e+01, 2.39697406373309e+01, 2.39688290995263e+01, 2.39683531867975e+01, 2.39681357365494e+01, 2.39680541631642e+01, 2.39680319731982e+01, 2.39680286980048e+01, 2.39680284453903e+01, 2.39680239092457e+01, 2.39679931761922e+01, 2.39678801981467e+01, 2.39675790394656e+01, 2.39669199575484e+01, 2.39656577233801e+01, 2.39634626454803e+01, 2.39599147030038e+01, 2.39545011269225e+01, 2.39466176910991e+01, 2.39355738859063e+01, 2.39206020440249e+01, 2.39008703699213e+01, 2.38754996912477e+01, 2.38435836039337e+01, 2.38042115274051e+01, 2.37564940292408e+01, 2.36995896293571e+01, 2.36327321642118e+01, 2.35552576942303e+01, 2.34666298847927e+01, 2.33664627924835e+01, 2.32545400496556e+01, 2.31308295621277e+01, 2.29954930116139e+01, 2.28488896755883e+01, 2.26915743282515e+01, 2.25242892510640e+01, 2.23479506445254e+01, 2.21636299810390e+01, 2.19725310603752e+01, 2.17759637137271e+01, 2.15753152381974e+01, 2.13720207180604e+01, 2.11675333903861e+01, 2.09632961338238e+01, 2.07607150040903e+01, 2.05611355249890e+01, 2.03658221984413e+01, 2.01759414553595e+01, 1.99925480619352e+01, 1.98165748425514e+01, 1.96488254859956e+01, 1.94899701583589e+01, 1.93405436392086e+01, 1.92009457115044e+01, 1.90714435577479e+01, 1.89521759374689e+01, 1.88431589412700e+01, 1.87442931339659e+01, 1.86553719145219e+01, 1.85760909339133e+01, 1.85060584232574e+01, 1.84448062924733e+01, 1.83918018628409e+01, 1.83464600939227e+01, 1.83081561558115e+01, 1.82762381819307e+01, 1.82500400170383e+01, 1.82288937519819e+01, 1.82121418141300e+01, 1.81991483636459e+01, 1.81893097342331e+01, 1.81820636556477e+01, 1.81768970063793e+01, 1.81733518696923e+01, 1.81710297047649e+01, 1.81695934958460e+01, 1.81687678039137e+01, 1.81683367141298e+01, 1.81681397446919e+01, 1.81680658545011e+01, 1.81680457545631e+01, 1.81680427878538e+01, 1.81680425632235e+01, 1.81680385839205e+01, 1.81680116235556e+01, 1.81679125143761e+01, 1.81676483254903e+01, 1.81670701535555e+01, 1.81659628812251e+01, 1.81640373131306e+01, 1.81609250503863e+01, 1.81561764079898e+01, 1.81492616136682e+01, 1.81395754497597e+01, 1.81264454096226e+01, 1.81091433353760e+01, 1.80869003842446e+01, 1.80589250379830e+01, 1.80244237277164e+01, 1.79826235015515e+01, 1.79327960234674e+01, 1.78742820702027e+01, 1.78065155999949e+01, 1.77290464145066e+01, 1.76415604325032e+01, 1.75438966465972e+01, 1.74360599436512e+01, 1.73182291309217e+01, 1.71907597144682e+01, 1.70541812107031e+01, 1.69091890211468e+01, 1.67566311491683e+01, 1.65974902716008e+01, 1.64328618852489e+01, 1.62639294178597e+01, 1.60919373158600e+01, 1.59181631891740e+01, 1.57438901007832e+01, 1.55703800327397e+01, 1.53988494434263e+01, 1.52304476613830e+01, 1.50662386534959e+01, 1.49071864789487e+01, 1.47541445162456e+01, 1.46078483488294e+01, 1.44689120312003e+01, 1.43378273417263e+01, 1.42149655634304e+01, 1.41005813166185e+01, 1.39948179892294e+01, 1.38977143614450e+01, 1.38092120888448e+01, 1.37291637824435e+01, 1.36573414954132e+01, 1.35934454886056e+01, 1.35371131959680e+01, 1.34879283445651e+01, 1.34454302018555e+01, 1.34091229261342e+01, 1.33784849865860e+01, 1.33529785998091e+01, 1.33320591030202e+01, 1.33151841537967e+01, 1.33018226156457e+01, 1.32914629613860e+01, 1.32836210055542e+01, 1.32778467656259e+01, 1.32737302519214e+01, 1.32709059989218e+01, 1.32690561765746e+01, 1.32679121581401e+01, 1.32672544693025e+01, 1.32669110988252e+01, 1.32667542105586e+01, 1.32666953564409e+01, 1.32666793467021e+01, 1.32666769836985e+01, 1.32666768063071e+01, 1.32666736840170e+01, 1.32666525300495e+01, 1.32665747659278e+01, 1.32663674761062e+01, 1.32659138313372e+01, 1.32650450624148e+01, 1.32635343131629e+01, 1.32610926674509e+01, 1.32573676050430e+01, 1.32519440932665e+01, 1.32443484632080e+01, 1.32340551476678e+01, 1.32204962711896e+01, 1.32030739790966e+01, 1.31811752735261e+01, 1.31541889933500e+01, 1.31215244378507e+01, 1.30826310001839e+01, 1.30370180575048e+01, 1.29842742731808e+01, 1.29240854159549e+01, 1.28562498029209e+01, 1.27806905360380e+01, 1.26974638288179e+01, 1.26067629074231e+01, 1.25089172081633e+01, 1.24043868637491e+01, 1.22937527505195e+01, 1.21777026319183e+01, 1.20570141536677e+01, 1.19325356012331e+01, 1.18051654055912e+01, 1.16758313742864e+01, 1.15454705374347e+01, 1.14150103487010e+01, 1.12853517923419e+01, 1.11573547450804e+01, 1.10318257502381e+01, 1.09095082002753e+01, 1.07910748040117e+01, 1.06771221394224e+01, 1.05681670579548e+01, 1.04646447029338e+01, 1.03669079219440e+01, 1.02752278807456e+01, 1.01897957162171e+01, 1.01107250929778e+01, 1.00380555505371e+01, 9.97175654494200e+00, 9.91173210195247e+00, 9.85782600892193e+00, 9.80982748055284e+00, 9.76747723964081e+00, 9.73047395735941e+00, 9.69848099783323e+00, 9.67113340801858e+00, 9.64804508592910e+00, 9.62881604822162e+00, 9.61303970293986e+00, 9.60031001621815e+00, 9.59022844485070e+00, 9.58241049196784e+00, 9.57649173279059e+00, 9.57213315351125e+00, 9.56902565028678e+00, 9.56689354803390e+00, 9.56549702033758e+00, 9.56463332170649e+00, 9.56413678025991e+00, 9.56387754073092e+00, 9.56375909201584e+00, 9.56371465780813e+00, 9.56370257062575e+00, 9.56370078658293e+00, 9.56370065306155e+00, 9.56369830835888e+00, 9.56368242265540e+00, 9.56362402515987e+00, 9.56346835911487e+00, 9.56312768898765e+00, 9.56247526955559e+00, 9.56134072113435e+00, 9.55950702803901e+00, 9.55670934567973e+00, 9.55263576430638e+00, 9.54693013275351e+00, 9.53919699194122e+00, 9.52900860406690e+00, 9.51591398892871e+00, 9.49944979454138e+00, 9.47915273774070e+00, 9.45457325636318e+00, 9.42528992423467e+00, 9.39092410146075e+00, 9.35115423390075e+00, 9.30572918531052e+00, 9.25447998982431e+00, 9.19732945467454e+00, 9.13429912301022e+00, 9.06551322006984e+00, 8.99119934502104e+00, 8.91168582549699e+00, 8.82739581164594e+00, 8.73883834173222e+00, 8.64659675420891e+00, 8.55131494570974e+00, 8.45368207526362e+00, 8.35441638621240e+00, 8.25424885135720e+00, 8.15390733573867e+00, 8.05410190930082e+00, 7.95551182859145e+00, 7.85877455173559e+00, 7.76447697269015e+00, 7.67314888389392e+00, 7.58525852562035e+00, 7.50120997401181e+00, 7.42134206566099e+00, 7.34592855119261e+00, 7.27517920167197e+00, 7.20924164418732e+00, 7.14820376178452e+00, 7.09209654701301e+00, 7.04089734139611e+00, 6.99453342313187e+00, 6.95288592310051e+00, 6.91579405705749e+00, 6.88305966225957e+00, 6.85445202186820e+00, 6.82971295187038e+00, 6.80856211399836e+00, 6.79070250494362e+00, 6.77582605768765e+00, 6.76361927573434e+00, 6.75376880634312e+00, 6.74596684563438e+00, 6.73991625792416e+00, 6.73533528511891e+00, 6.73196172062969e+00, 6.72955642694962e+00, 6.72790608728628e+00, 6.72682509946896e+00, 6.72615654423980e+00, 6.72577218895078e+00, 6.72557152014876e+00, 6.72547983275519e+00, 6.72544543762248e+00, 6.72543608130997e+00, 6.72543470033776e+00, 6.72543459694430e+00, 6.72543278078645e+00, 6.72542047605052e+00, 6.72537524255064e+00, 6.72525466673209e+00, 6.72499078993170e+00, 6.72448543593485e+00, 6.72360662623093e+00, 6.72218624987109e+00, 6.72001912976627e+00, 6.71686359778698e+00, 6.71244365597810e+00, 6.70645275935844e+00, 6.69855920617115e+00, 6.68841306377590e+00, 6.67565449334001e+00, 6.65992326616604e+00, 6.64086919250968e+00, 6.61816311523383e+00, 6.59150806196329e+00, 6.56065010746884e+00, 6.52538847938710e+00, 6.48558445015687e+00, 6.44116859871416e+00, 6.39214609605234e+00, 6.33859976442691e+00, 6.28069077254037e+00, 6.21865694799137e+00, 6.15280880263116e+00, 6.08352346673269e+00, 6.01123680763048e+00, 5.93643406518971e+00, 5.85963937112112e+00, 5.78140453507448e+00, 5.70229748151777e+00, 5.62289071065799e+00, 5.54375013549158e+00, 5.46542461548777e+00, 5.38843646490885e+00, 5.31327316050295e+00, 5.24038041092474e+00, 5.17015668217941e+00, 5.10294920448207e+00, 5.03905142156785e+00, 4.97870178864671e+00, 4.92208378356882e+00, 4.86932696928096e+00, 4.82050893437183e+00, 4.77565794084002e+00, 4.73475612140977e+00, 4.69774308934910e+00, 4.66451984825124e+00, 4.63495291432523e+00, 4.60887858664348e+00, 4.58610731944383e+00, 4.56642816366182e+00, 4.54961325178320e+00, 4.53542230093603e+00, 4.52360710454415e+00, 4.51391597397597e+00, 4.50609807993619e+00, 4.49990763058671e+00, 4.49510781135385e+00, 4.49147440182627e+00, 4.48879897960065e+00, 4.48689162056654e+00, 4.48558301064740e+00, 4.48472589560936e+00, 4.48419581282888e+00, 4.48389107098397e+00, 4.48373196918185e+00, 4.48365927446098e+00, 4.48363200419492e+00, 4.48362458602448e+00, 4.48362349111819e+00, 4.48362340892504e+00, 4.48362196225668e+00, 4.48361216086856e+00, 4.48357612997150e+00, 4.48348008515497e+00, 4.48326989507898e+00, 4.48286736228445e+00, 4.48216737424449e+00, 4.48103606085238e+00, 4.47931007590594e+00, 4.47679709795906e+00, 4.47327761904737e+00, 4.46850805709146e+00, 4.46222518830948e+00, 4.45415184933823e+00, 4.44400380540849e+00, 4.43149762233098e+00, 4.41635931888556e+00, 4.39833351627382e+00, 4.37719274735587e+00, 4.35274654578780e+00, 4.32484990930978e+00, 4.29341072712970e+00, 4.25839578218230e+00, 4.21983498673043e+00, 4.17782358367862e+00, 4.13252214285816e+00, 4.08415429560962e+00, 4.03300227420051e+00, 3.97940044537472e+00, 3.92372713940323e+00, 3.86639516765814e+00, 3.80784148490432e+00, 3.74851648187557e+00, 3.68887338745203e+00, 3.62935821984494e+00, 3.57040065809819e+00, 3.51240611711691e+00, 3.45574921102190e+00, 3.40076869074125e+00, 3.34776385108871e+00, 3.29699232676089e+00, 3.24866913972789e+00, 3.20296682377035e+00, 3.16001643451503e+00, 3.11990925263046e+00, 3.08269900025493e+00, 3.04840441227418e+00, 3.01701203093472e+00, 2.98847912114523e+00, 2.96273663201745e+00, 2.93969215574702e+00, 2.91923285648920e+00, 2.90122835863470e+00, 2.88553359546411e+00, 2.87199162552729e+00, 2.86043642551148e+00, 2.85069566532672e+00, 2.84259346435841e+00, 2.83595311822522e+00, 2.83059977400305e+00, 2.82636301993264e+00, 2.82307934436975e+00, 2.82059440939562e+00, 2.81876507818797e+00, 2.81746113284539e+00, 2.81656662142902e+00, 2.81598077972003e+00, 2.81561848435978e+00, 2.81541020901048e+00, 2.81530147297597e+00, 2.81525179114720e+00, 2.81523315386919e+00, 2.81522808408629e+00, 2.81522733579727e+00, 2.81522727938498e+00, 2.81522628329459e+00, 2.81521953464049e+00, 2.81519472594663e+00, 2.81512859563411e+00, 2.81498387388812e+00, 2.81470672525843e+00, 2.81422479534801e+00, 2.81344595946116e+00, 2.81225785981025e+00, 2.81052830390817e+00, 2.80810657891750e+00, 2.80482571373837e+00, 2.80050569144538e+00, 2.79495757877677e+00, 2.78798849698118e+00, 2.77940731072272e+00, 2.76903086141370e+00, 2.75669052194729e+00, 2.74223880598976e+00, 2.72555573197911e+00, 2.70655462493701e+00, 2.68518704250780e+00, 2.66144653808551e+00, 2.63537102400215e+00, 2.60704356935153e+00, 2.57659155517678e+00, 2.54418420714170e+00, 2.51002862357343e+00, 2.47436450563799e+00, 2.43745786801460e+00, 2.39959405646992e+00, 2.36107041983934e+00, 2.32218897802840e+00, 2.28324939776264e+00, 2.24454253931967e+00, 2.20634477709008e+00, 2.16891323146898e+00, 2.13248198545969e+00, 2.09725930124016e+00, 2.06342580288971e+00, 2.03113355298828e+00, 2.00050592310815e+00, 1.97163814070351e+00, 1.94459838650793e+00, 1.91942931603879e+00, 1.89614988490960e+00, 1.87475736906938e+00, 1.85522948648548e+00, 1.83752654474741e+00, 1.82159355814379e+00, 1.80736229652426e+00, 1.79475324541046e+00, 1.78367747129959e+00, 1.77403839715060e+00, 1.76573350023206e+00, 1.75865594775122e+00, 1.75269618518747e+00, 1.74774348848819e+00, 1.74368748492349e+00, 1.74041963926678e+00, 1.73783469299394e+00, 1.73583203534412e+00, 1.73431697730102e+00, 1.73320189368805e+00, 1.73240719530809e+00, 1.73186209285849e+00, 1.73150511739532e+00, 1.73128436829003e+00, 1.73115746851031e+00, 1.73109121801018e+00, 1.73106094820803e+00, 1.73104959305702e+00, 1.73104650418978e+00, 1.73104604827977e+00, 1.73104601372242e+00, 1.73104540105093e+00, 1.73104125011621e+00, 1.73102599091277e+00, 1.73098531602945e+00, 1.73089630283634e+00, 1.73072584262982e+00, 1.73042944440748e+00, 1.72995047604949e+00, 1.72921989975338e+00, 1.72815655085443e+00, 1.72666799870647e+00, 1.72465201408498e+00, 1.72199864874010e+00, 1.71859290877123e+00, 1.71431797451494e+00, 1.70905888650628e+00, 1.70270658155160e+00, 1.69516212768943e+00, 1.68634097519482e+00, 1.67617701662136e+00, 1.66462623597533e+00, 1.65166972873974e+00, 1.63731589278314e+00, 1.62160162582019e+00, 1.60459241681495e+00, 1.58638128342110e+00, 1.56708658040567e+00, 1.54684877889992e+00, 1.52582638643288e+00, 1.50419123617108e+00, 1.48212341444714e+00, 1.45980611381275e+00, 1.43742069201477e+00, 1.41514218584572e+00, 1.39313547631964e+00, 1.37155223460785e+00, 1.35052870526105e+00, 1.33018431367669e+00, 1.31062102665015e+00, 1.29192335366664e+00, 1.27415885444272e+00, 1.25737901391387e+00, 1.24162035575709e+00, 1.22690568480439e+00, 1.21324537251166e+00, 1.20063862400391e+00, 1.18907468737387e+00, 1.17853398435592e+00, 1.16898915573088e+00, 1.16040602500448e+00, 1.15274449056303e+00, 1.14595936028112e+00, 1.14000114404313e+00, 1.13481681934994e+00, 1.13035058350840e+00, 1.12654460314457e+00, 1.12333976818270e+00, 1.12067645320636e+00, 1.11849528448737e+00, 1.11673790618120e+00, 1.11534773452692e+00, 1.11427068467082e+00, 1.11345585128232e+00, 1.11285612176446e+00, 1.11242869984357e+00, 1.11213551783886e+00, 1.11194351803065e+00, 1.11182478720220e+00, 1.11175653342335e+00, 1.11172090015053e+00, 1.11170461932229e+00, 1.11169851187142e+00, 1.11169685050142e+00, 1.11169660528685e+00, 1.11169658658199e+00, 1.11169625340515e+00, 1.11169399608557e+00, 1.11168569797856e+00, 1.11166357856774e+00, 1.11161517227919e+00, 1.11152247419678e+00, 1.11136128944611e+00, 1.11110082032518e+00, 1.11070352198291e+00, 1.11012525315801e+00, 1.10931574263315e+00, 1.10821938404384e+00, 1.10677636099405e+00, 1.10492409078410e+00, 1.10259895853034e+00, 1.09973829463988e+00, 1.09628252866386e+00, 1.09217743324970e+00, 1.08737635546505e+00, 1.08184232161836e+00, 1.07554989815727e+00, 1.06848669701339e+00, 1.06065442961901e+00, 1.05206943917808e+00, 1.04276267365884e+00, 1.03277909917913e+00, 1.02217659097890e+00, 1.01102437286514e+00, 9.99401102275208e-01, 9.87392714480031e-01, 9.75090144986297e-01, 9.62587044460761e-01, 9.49977587308503e-01, 9.37354455979360e-01, 9.24807060983071e-01, 9.12420034068319e-01, 9.00272011136849e-01, 8.88434703626913e-01, 8.76972243064230e-01, 8.65940773453779e-01, 8.55388259979321e-01, 8.45354479638516e-01, 8.35871159378712e-01, 8.26962229364533e-01, 8.18644162567402e-01, 8.10926376329775e-01, 8.03811676412245e-01, 7.97296728857873e-01, 7.91372549478133e-01, 7.86025004645921e-01, 7.81235320221796e-01, 7.76980597759646e-01, 7.73234338615104e-01, 7.69966977233253e-01, 7.67146424778346e-01, 7.64738623467608e-01, 7.62708110591596e-01, 7.61018589373094e-01, 7.59633501686700e-01, 7.58516595404437e-01, 7.57632476938149e-01, 7.56947137614389e-01, 7.56428441038437e-01, 7.56046557756697e-01, 7.55774333452862e-01, 7.55587577701846e-01, 7.55465261977022e-01, 7.55389618113135e-01, 7.55346131640351e-01, 7.55323428141416e-01, 7.55313054807175e-01, 7.55309163424627e-01, 7.55308104875468e-01, 7.55307948635881e-01, 7.55307936633884e-01, 7.55307721748125e-01, 7.55306265866317e-01, 7.55300913904158e-01, 7.55286647640439e-01, 7.55255426791031e-01, 7.55195637120977e-01, 7.55091668613787e-01, 7.54923644587336e-01, 7.54667320118678e-01, 7.54294165026032e-01, 7.53771642709196e-01, 7.53063691105766e-01, 7.52131405657380e-01, 7.50933916386067e-01, 7.49429442016996e-01, 7.47576493854659e-01, 7.45335191397416e-01, 7.42668641308185e-01, 7.39544322411149e-01, 7.35935413031498e-01, 7.31821994363614e-01, 7.27192065528854e-01, 7.22042313031515e-01, 7.16378589328443e-01, 7.10216071450731e-01, 7.03579089714005e-01, 6.96500636745284e-01, 6.89021586354635e-01, 6.81189668287914e-01, 6.73058257073464e-01, 6.64685040052063e-01, 6.56130630971454e-01, 6.47457191625698e-01, 6.38727115879853e-01, 6.30001819369724e-01, 6.21340665679313e-01, 6.12800047283415e-01, 6.04432628166840e-01, 5.96286745612694e-01, 5.88405961625391e-01, 5.80828749894067e-01, 5.73588301900436e-01, 5.66712435333365e-01, 5.60223588904300e-01, 5.54138889468668e-01, 5.48470279613866e-01, 5.43224696224854e-01, 5.38404292736480e-01, 5.34006699670556e-01, 5.30025319552266e-01, 5.26449653373547e-01, 5.23265656423156e-01, 5.20456121552165e-01, 5.18001087820785e-01, 5.15878272010667e-01, 5.14063519730144e-01, 5.12531271840032e-01, 5.11255040750619e-01, 5.10207889868958e-01, 5.09362908205958e-01, 5.08693670996166e-01, 5.08174676256651e-01, 5.07781746628375e-01, 5.07492385707309e-01, 5.07286078460737e-01, 5.07144526278701e-01, 5.07051808734779e-01, 5.06994466177346e-01, 5.06961499755064e-01, 5.06944288269281e-01, 5.06936424189576e-01, 5.06933474100222e-01, 5.06932671604233e-01, 5.06932553157505e-01, 5.06932543989237e-01, 5.06932378935055e-01, 5.06931260668914e-01, 5.06927149807147e-01, 5.06916191780860e-01, 5.06892210519339e-01, 5.06846283959781e-01, 5.06766418616927e-01, 5.06637338555670e-01, 5.06440401897812e-01, 5.06153655370099e-01, 5.05752034240182e-01, 5.05207711160509e-01, 5.04490592839858e-01, 5.03568958048078e-01, 5.02410224292419e-01, 5.00981823783711e-01, 4.99252162391784e-01, 4.97191628684021e-01, 4.94773614480906e-01, 4.91975504339237e-01, 4.88779589646981e-01, 4.85173864092289e-01, 4.81152661417255e-01, 4.76717103502573e-01, 4.71875336508162e-01, 4.66642544224015e-01, 4.61040739927682e-01, 4.55098349764583e-01, 4.48849610896495e-01, 4.42333815570905e-01, 4.35594437357160e-01, 4.28678177973282e-01, 4.21633972642475e-01, 4.14511989285742e-01, 4.07362652727457e-01, 4.00235720118609e-01, 3.93179428527887e-01, 3.86239730518277e-01, 3.79459628759442e-01, 3.72878616430576e-01, 3.66532226366530e-01, 3.60451688577790e-01, 3.54663692918841e-01, 3.49190251299042e-01, 3.44048651951987e-01, 3.39251496938012e-01, 3.34806813270333e-01, 3.30718227819695e-01, 3.26985196414654e-01, 3.23603278224688e-01, 3.20564447462544e-01, 3.17857435525788e-01, 3.15468097761723e-01, 3.13379799947384e-01, 3.11573820216090e-01, 3.10029762462079e-01, 3.08725977189553e-01, 3.07639985362201e-01, 3.06748900118251e-01, 3.06029840346639e-01, 3.05460329197672e-01, 3.05018669768124e-01, 3.04684289592807e-01, 3.04438045317013e-01, 3.04262479109447e-01, 3.04142019059505e-01, 3.04063116996339e-01, 3.04014318834824e-01, 3.03986264615467e-01, 3.03971617749291e-01, 3.03964925467071e-01, 3.03962414959663e-01, 3.03961732040648e-01, 3.03961631243233e-01, 3.03961623377837e-01, 3.03961480961881e-01, 3.03960516072969e-01, 3.03956969042083e-01, 3.03947513981406e-01, 3.03926821920818e-01, 3.03887194565014e-01, 3.03818283562023e-01, 3.03706908557068e-01, 3.03536985418425e-01, 3.03289573789662e-01, 3.02943050446636e-01, 3.02473411713983e-01, 3.01854704344706e-01, 3.01059579758814e-01, 3.00059961421688e-01, 2.98827809571215e-01, 2.97335961735151e-01, 2.95559021885978e-01, 2.93474266113783e-01, 2.91062528856472e-01, 2.88309031499679e-01, 2.85204114953743e-01, 2.81743839903079e-01, 2.77930422879344e-01, 2.73772482991859e-01, 2.69285082696051e-01, 2.64489555850695e-01, 2.59413126835923e-01, 2.54088334948904e-01, 2.48552287947199e-01, 2.42845776828204e-01, 2.37012290189128e-01, 2.31096970428261e-01, 2.25145555411238e-01, 2.19203348000711e-01, 2.13314252185896e-01, 2.07519908774707e-01, 2.01858956217777e-01, 1.96366433727732e-01, 1.91073335117281e-01, 1.86006313377177e-01, 1.81187528553345e-01, 1.76634625427120e-01, 1.72360823153909e-01, 1.68375096478137e-01, 1.64682427353242e-01, 1.61284106536883e-01, 1.58178066691960e-01, 1.55359231332767e-01, 1.52819867238456e-01, 1.50549931361979e-01, 1.48537406495592e-01, 1.46768622780293e-01, 1.45228564404768e-01, 1.43901162434688e-01, 1.42769575615887e-01, 1.41816461227720e-01, 1.41024237700066e-01, 1.40375339856561e-01, 1.39852466451103e-01, 1.39438818280659e-01, 1.39118323756819e-01, 1.38875847562502e-01, 1.38697377059366e-01, 1.38570180561835e-01, 1.38482931535421e-01, 1.38425793240181e-01, 1.38390459306564e-01, 1.38370147133770e-01, 1.38359542735003e-01, 1.38354697585775e-01, 1.38352880017280e-01, 1.38352385595930e-01, 1.38352312620412e-01, 1.38352306877565e-01, 1.38352202272870e-01, 1.38351493561572e-01, 1.38348888277001e-01, 1.38341943645642e-01, 1.38326746001333e-01, 1.38297642654720e-01, 1.38247037826218e-01, 1.38165263446638e-01, 1.38040534654586e-01, 1.37858998540537e-01, 1.37604882939584e-01, 1.37260749681396e-01, 1.36807853519002e-01, 1.36226603889767e-01, 1.35497121725351e-01, 1.34599877863828e-01, 1.33516393512195e-01, 1.32229977093473e-01, 1.30726466245214e-01, 1.28994939346455e-01, 1.27028358384479e-01, 1.24824104804304e-01, 1.22384372628616e-01, 1.19716388770435e-01, 1.16832438958035e-01, 1.13749688594602e-01, 1.10489800405972e-01, 1.07078363854775e-01, 1.03544163825937e-01, 9.99183268103465e-02, 9.62333906551152e-02, 9.25223481181021e-02, 8.88177145630351e-02, 8.51506662204713e-02, 8.15502880324530e-02, 7.80429600849004e-02, 7.46519001481545e-02, 7.13968681029939e-02, 6.82940271546784e-02, 6.53559476326535e-02, 6.25917324351164e-02, 6.00072390587201e-02, 5.76053715913929e-02, 5.53864167180365e-02, 5.33484002068860e-02, 5.14874439480053e-02, 4.97981078585568e-02, 4.82737053750597e-02, 4.69065854492570e-02, 4.56883776881369e-02, 4.46102003749900e-02, 4.36628335106164e-02, 4.28368607222822e-02, 4.21227849518715e-02, 4.15111233233389e-02, 4.09924865883369e-02, 4.05576481386731e-02, 4.01976068418861e-02, 3.99036469826726e-02, 3.96673974621535e-02, 3.94808912054998e-02, 3.93366245425568e-02, 3.92276152420692e-02, 3.91474569802028e-02, 3.90903673758274e-02, 3.90512263821279e-02, 3.90256018089588e-02, 3.90097590637595e-02, 3.90006528024145e-02, 3.89958990192127e-02, 3.89937270919134e-02, 3.89929123456980e-02, 3.89926907167322e-02, 3.89926580048225e-02, 3.89926554077462e-02, 3.89926078131211e-02, 3.89922853535367e-02, 3.89910999731740e-02, 3.89879403010056e-02, 3.89810260278125e-02, 3.89677866360572e-02, 3.89447703210237e-02, 3.89075888917816e-02, 3.88509043281502e-02, 3.87684623647999e-02, 3.86531778103176e-02, 3.84972751126398e-02, 3.82924857946634e-02, 3.80303017052270e-02, 3.77022795523558e-02, 3.73003880298342e-02, 3.68173842836370e-02, 3.62472019135254e-02, 3.55853287183621e-02, 3.48291495895054e-02, 3.39782289473192e-02, 3.30345083921415e-02, 3.20023990724678e-02, 3.08887546134144e-02, 2.97027188856163e-02, 2.84554526602960e-02, 2.71597532445540e-02, 2.58295903371177e-02, 2.44795884509372e-02, 2.31244904114192e-02, 2.17786371396613e-02, 2.04554961368331e-02, 1.91672652447608e-02, 1.79245702261426e-02, 1.67362655895928e-02, 1.56093390397487e-02, 1.45489119790696e-02, 1.35583223558484e-02, 1.26392722186495e-02, 1.17920206403497e-02, 1.10156029846777e-02, 1.03080593857641e-02, 9.66665826801005e-03, 9.08810420057904e-03, 8.56872286961874e-03, 8.10461910986504e-03, 7.69180657363222e-03, 7.32630966562198e-03, 7.00423985167475e-03, 6.72184940374621e-03, 6.47556614017475e-03, 6.26201284620009e-03, 6.07801491963755e-03, 5.92059948371570e-03, 5.78698882328155e-03, 5.67459058200370e-03, 5.58098672899376e-03, 5.50392287247092e-03, 5.44129907162727e-03, 5.39116288337027e-03, 5.35170498542215e-03, 5.32125735362854e-03, 5.29829364803902e-03, 5.28143119375812e-03, 5.26943374127598e-03, 5.26121406775104e-03, 5.25583544605013e-03, 5.25251106111568e-03, 5.25060058932070e-03, 5.24960335953750e-03, 5.24914776465124e-03, 5.24897686298686e-03, 5.24893037432150e-03, 5.24892351271689e-03, 5.24892296297708e-03, 5.24891282539386e-03, 5.24884414218008e-03, 5.24859166159786e-03, 5.24791868848240e-03, 5.24644614397829e-03, 5.24362696709346e-03, 5.23872727909946e-03, 5.23081581978602e-03, 5.21876324368264e-03, 5.20125287831519e-03, 5.17680443122259e-03, 5.14381181267254e-03, 5.10059565219478e-03, 5.04547018691426e-03, 4.97682298694713e-03, 4.89320451725916e-03, 4.79342294178053e-03, 4.67663805412855e-03, 4.54244701612083e-03, 4.39095397320699e-03, 4.22281583359307e-03, 4.03925770161351e-03, 3.84205366308521e-03, 3.63347167292005e-03, 3.41618485675355e-03, 3.19315512476948e-03, 2.96749805691121e-03, 2.74234005108553e-03, 2.52067939153733e-03, 2.30526211353066e-03, 2.09848149594904e-03, 1.90230708945904e-03, 1.71824585641644e-03, 1.54733470669299e-03, 1.39016083834775e-03, 1.24690413763773e-03, 1.11739471201133e-03, 1.00117852093717e-03, 8.97584873925439e-04, 8.05790941384998e-04, 7.24879967229697e-04, 6.53891175957710e-04, 5.91860299223868e-04, 5.37850461405753e-04, 4.90973979701437e-04, 4.50406216764065e-04, 4.15392944077553e-04, 3.85252862251662e-04, 3.59376795991295e-04, 3.37224588207619e-04, 3.18320322939769e-04, 3.02246455206046e-04, 2.88637440661432e-04, 2.77173343675940e-04, 2.67573754065609e-04, 2.59592233696537e-04, 2.53011415296100e-04, 2.47638785602078e-04, 2.43303133729000e-04, 2.39851629764587e-04, 2.37147495340650e-04, 2.35068219893719e-04, 2.33504260961581e-04, 2.32358147947309e-04, 2.31543893129071e-04, 2.30986603646954e-04, 2.30622187181950e-04, 2.30397050223985e-04, 2.30267701859077e-04, 2.30200194422991e-04, 2.30169355219800e-04, 2.30157787250107e-04, 2.30154640568969e-04, 2.30154176136921e-04, 2.30154138577632e-04, 2.30153441847847e-04, 2.30148721392709e-04, 2.30131369235664e-04, 2.30085120257684e-04, 2.29983933458300e-04, 2.29790255744433e-04, 2.29453783568090e-04, 2.28910853749653e-04, 2.28084610199630e-04, 2.26886102211665e-04, 2.25216463881850e-04, 2.22970294480263e-04, 2.20040296881052e-04, 2.16323128583440e-04, 2.11726274812918e-04, 2.06175571968004e-04, 1.99622809766709e-04, 1.92052657276065e-04, 1.83488039667168e-04, 1.73993106322712e-04, 1.63673154943922e-04, 1.52671344409328e-04, 1.41162568162656e-04, 1.29345030821071e-04, 1.17429783671241e-04, 1.05628800675106e-04, 9.41437788597455e-05, 8.31573238073702e-05, 7.28225065103061e-05, 6.32474808668348e-05, 5.44910612916283e-05, 4.65842233694502e-05, 3.95329674987011e-05, 3.32727724128290e-05, 2.78284950354646e-05, 2.39540955888629e-05, 2.11863770118725e-05, 1.85810795585439e-05, 1.59456947486092e-05, 1.33124274412921e-05, 1.08260594942067e-05, 8.67566583195867e-06, 7.05100058695507e-06, 5.97092696352775e-06, 5.08366782270286e-06, 4.15467629011346e-06, 3.31576526235722e-06, 2.77827225403371e-06, 2.24555829220208e-06, 1.69296838947730e-06, 1.19495256001301e-06, 1.16399502303466e-06, 1.11088752419164e-06, 1.03216050462152e-06, 9.28291555624128e-07, 8.07845921713124e-07, 6.86311658171235e-07, 5.76269769680948e-07, 4.82278004199088e-07, 4.02557688516936e-07, -2.13162820728030e-14, -2.13162820728030e-14, -2.84217094304040e-14, -2.13162820728030e-14, -2.84217094304040e-14, -1.42108547152020e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14, -2.13162820728030e-14],
[ 4.40000000000000e+01, 4.39999999614637e+01, 4.39999987698448e+01, 4.39999906964385e+01, 4.39999610175493e+01, 4.39998819030394e+01, 4.39997087553201e+01, 4.39993771277938e+01, 4.39988003342087e+01, 4.39978678393397e+01, 4.39964444975792e+01, 4.39943706802840e+01, 4.39914633060728e+01, 4.39875177617932e+01, 4.39823106766146e+01, 4.39756034886578e+01, 4.39671467234881e+01, 4.39566848871726e+01, 4.39439618635268e+01, 4.39287266953616e+01, 4.39107396223234e+01, 4.38897782423900e+01, 4.38656436592838e+01, 4.38381664732176e+01, 4.38072124671818e+01, 4.37726878357541e+01, 4.37345437992492e+01, 4.36927804447164e+01, 4.36474496390517e+01, 4.35986568706129e+01, 4.35465618960511e+01, 4.34913780995630e+01, 4.34333705121051e+01, 4.33728524865997e+01, 4.33101810787758e+01, 4.32457512380764e+01, 4.31799889646434e+01, 4.31133436325848e+01, 4.30462797131233e+01, 4.29792681515973e+01, 4.29127776588248e+01, 4.28472661706141e+01, 4.27831727109561e+01, 4.27209098671979e+01, 4.26608570522004e+01, 4.26033546920374e+01, 4.25486994407623e+01, 4.24971404881754e+01, 4.24488769937449e+01, 4.24040566506232e+01, 4.23627753582681e+01, 4.23250779603423e+01, 4.22909599858902e+01, 4.22603703157494e+01, 4.22332146822061e+01, 4.22093598976572e+01, 4.21886386972145e+01, 4.21708550707215e+01, 4.21557899516416e+01, 4.21432071239889e+01, 4.21328592042678e+01, 4.21244935537101e+01, 4.21178579773797e+01, 4.21127060713357e+01, 4.21088020872851e+01, 4.21059251961325e+01, 4.21038730474877e+01, 4.21024645412684e+01, 4.21015417495750e+01, 4.21009709514003e+01, 4.21006427686862e+01, 4.21004714189010e+01, 4.21003931257522e+01, 4.21003637549628e+01, 4.21003557653651e+01, 4.21003545861161e+01, 4.21003544314235e+01, 4.21003508272714e+01, 4.21003264085911e+01, 4.21002366426561e+01, 4.20999973574526e+01, 4.20994736765859e+01, 4.20984707227358e+01, 4.20967264451180e+01, 4.20939068733263e+01, 4.20896040393392e+01, 4.20833367421358e+01, 4.20745542560508e+01, 4.20626430056069e+01, 4.20469361471883e+01, 4.20267259128146e+01, 4.20012784847445e+01, 4.19698510828393e+01, 4.19317108604384e+01, 4.18861551196891e+01, 4.18325322748258e+01, 4.17702629139265e+01, 4.16988602404509e+01, 4.16179491227256e+01, 4.15272829531182e+01, 4.14267575318895e+01, 4.13164212564433e+01, 4.11964810239391e+01, 4.10673034452455e+01, 4.09294112114010e+01, 4.07834747289863e+01, 4.06302994177530e+01, 4.04708093083058e+01, 4.03060277586440e+01, 4.01370562050962e+01, 3.99650518695492e+01, 3.97912052704858e+01, 3.96167182525609e+01, 3.94427830874570e+01, 3.92705630366021e+01, 3.91011746267677e+01, 3.89356717854376e+01, 3.87750319163744e+01, 3.86201439607721e+01, 3.84717984746774e+01, 3.83306797471428e+01, 3.81973599763682e+01, 3.80722955075126e+01, 3.79558251147110e+01, 3.78481702830201e+01, 3.77494374169193e+01, 3.76596218739447e+01, 3.75786136972906e+01, 3.75062049005583e+01, 3.74420981407975e+01, 3.73859166012772e+01, 3.73372148916054e+01, 3.72954907587165e+01, 3.72601973873543e+01, 3.72307560532171e+01, 3.72065688768521e+01, 3.71870314132056e+01, 3.71715448023325e+01, 3.71595272031308e+01, 3.71504242359559e+01, 3.71437181730943e+01, 3.71389356392929e+01, 3.71356536181255e+01, 3.71335036034529e+01, 3.71321737873417e+01, 3.71314092346416e+01, 3.71310100574840e+01, 3.71308276674616e+01, 3.71307592462335e+01, 3.71307406339595e+01, 3.71307378868254e+01, 3.71307376443057e+01, 3.71307328922523e+01, 3.71307006963811e+01, 3.71305823407985e+01, 3.71302668470830e+01, 3.71295763907830e+01, 3.71282540612981e+01, 3.71259544470303e+01, 3.71222374651430e+01, 3.71165657830335e+01, 3.71083060958833e+01, 3.70967344300599e+01, 3.70810455343557e+01, 3.70603662992202e+01, 3.70337730086122e+01, 3.70003120819784e+01, 3.69590238094687e+01, 3.69089684286223e+01, 3.68492537444050e+01, 3.67790633673481e+01, 3.66976845478608e+01, 3.66045345290779e+01, 3.64991843342059e+01, 3.63813789525235e+01, 3.62510529928497e+01, 3.61083410334230e+01, 3.59535821097576e+01, 3.57873180425330e+01, 3.56102856083681e+01, 3.54234028835536e+01, 3.52277504199717e+01, 3.50245482064424e+01, 3.48151295822381e+01, 3.46009133612244e+01, 3.43833753749166e+01, 3.41640204648816e+01, 3.39443556978912e+01, 3.37258653080502e+01, 3.35099876494713e+01, 3.32980943044157e+01, 3.30914714345383e+01, 3.28913034605906e+01, 3.26986591730700e+01, 3.25144803835102e+01, 3.23395732083621e+01, 3.21746020342656e+01, 3.20200861540817e+01, 3.18763989998254e+01, 3.17437698424472e+01, 3.16222877858907e+01, 3.15119078559470e+01, 3.14124589713382e+01, 3.13236535811342e+01, 3.12450987542032e+01, 3.11763085084215e+01, 3.11167171664028e+01, 3.10656935185161e+01, 3.10225555623785e+01, 3.09865855715662e+01, 3.09570452268169e+01, 3.09331905230993e+01, 3.09142861486761e+01, 3.08996190208843e+01, 3.08885106608052e+01, 3.08803280977919e+01, 3.08744930166883e+01, 3.08704888962664e+01, 3.08678659365829e+01, 3.08662436341799e+01, 3.08653109348898e+01, 3.08648239712147e+01, 3.08646014710715e+01, 3.08645180031950e+01, 3.08644952978718e+01, 3.08644919466124e+01, 3.08644916777999e+01, 3.08644867168462e+01, 3.08644531056466e+01, 3.08643295471908e+01, 3.08640001847089e+01, 3.08632793778828e+01, 3.08618989270735e+01, 3.08594982504949e+01, 3.08556179550396e+01, 3.08496971672310e+01, 3.08410749054837e+01, 3.08289956768867e+01, 3.08126193694479e+01, 3.07910353830605e+01, 3.07632807995715e+01, 3.07283622363025e+01, 3.06852808626783e+01, 3.06330598933311e+01, 3.05707737127735e+01, 3.04975776480050e+01, 3.04127372987687e+01, 3.03156562728032e+01, 3.02059011657670e+01, 3.00832226798091e+01, 2.99475718940522e+01, 2.97991108826734e+01, 2.96382171148819e+01, 2.94654813540037e+01, 2.92816990836484e+01, 2.90878558070999e+01, 2.88851068682974e+01, 2.86747527044570e+01, 2.84582106384534e+01, 2.82369844353272e+01, 2.80126328719329e+01, 2.77867385030204e+01, 2.75608776637801e+01, 2.73365925506417e+01, 2.71153659970227e+01, 2.68985993371313e+01, 2.66875935522449e+01, 2.64835337351196e+01, 2.62874767947170e+01, 2.61003422522982e+01, 2.59229059425909e+01, 2.57557964195153e+01, 2.55994938649349e+01, 2.54543313036957e+01, 2.53204979344952e+01, 2.51980443920469e+01, 2.50868897612256e+01, 2.49868301684836e+01, 2.48975487795961e+01, 2.48186270349514e+01, 2.47495569529321e+01, 2.46897543271282e+01, 2.46385726332392e+01, 2.45953174462956e+01, 2.45592611489115e+01, 2.45296576882591e+01, 2.45057571157456e+01, 2.44868196219709e+01, 2.44721287637391e+01, 2.44610035728719e+01, 2.44528092410903e+01, 2.44469660933053e+01, 2.44429565942547e+01, 2.44403301804047e+01, 2.44387057690042e+01, 2.44377718666849e+01, 2.44372842776570e+01, 2.44370614924099e+01, 2.44369779176822e+01, 2.44369551833025e+01, 2.44369518277547e+01, 2.44369515689425e+01, 2.44369469215071e+01, 2.44369154344425e+01, 2.44367996845744e+01, 2.44364911372502e+01, 2.44358158855603e+01, 2.44345226846961e+01, 2.44322737565593e+01, 2.44286387807644e+01, 2.44230924198996e+01, 2.44150156480803e+01, 2.44037010606268e+01, 2.43883622372451e+01, 2.43681471100721e+01, 2.43421551513432e+01, 2.43094580451895e+01, 2.42691233486092e+01, 2.42202404851919e+01, 2.41619482615770e+01, 2.40934629628249e+01, 2.40141059817686e+01, 2.39233298813511e+01, 2.38207417879836e+01, 2.37061230741704e+01, 2.35794444109143e+01, 2.34408754499945e+01, 2.32907886232328e+01, 2.31297568067706e+01, 2.29585448777298e+01, 2.27780954727832e+01, 2.25895095279801e+01, 2.23940224217483e+01, 2.21929767426520e+01, 2.19877928434562e+01, 2.17799384070397e+01, 2.15708982259665e+01, 2.13621452843823e+01, 2.11551140408804e+01, 2.09511765709495e+01, 2.07516219733498e+01, 2.05576392117598e+01, 2.03703033772611e+01, 2.01905652297740e+01, 2.00192438039351e+01, 1.98570218340171e+01, 1.97044437471928e+01, 1.95619159810796e+01, 1.94297093914969e+01, 1.93079635262870e+01, 1.91966925507963e+01, 1.90957926214068e+01, 1.90050505161659e+01, 1.89241533457546e+01, 1.88526991822018e+01, 1.87902084546087e+01, 1.87361359682735e+01, 1.86898834041482e+01, 1.86508121485630e+01, 1.86182562889074e+01, 1.85915355908964e+01, 1.85699682496844e+01, 1.85528831836639e+01, 1.85396316199236e+01, 1.85295977076568e+01, 1.85222078934805e+01, 1.85169388030478e+01, 1.85133233978609e+01, 1.85109552149347e+01, 1.85094905487870e+01, 1.85086484978423e+01, 1.85082088674899e+01, 1.85080079958665e+01, 1.85079326418427e+01, 1.85079121437076e+01, 1.85079091182254e+01, 1.85079088891450e+01, 1.85079048310087e+01, 1.85078773365370e+01, 1.85077762639332e+01, 1.85075068413448e+01, 1.85069172159133e+01, 1.85057880100444e+01, 1.85038243033439e+01, 1.85006504092291e+01, 1.84958077578407e+01, 1.84887561300834e+01, 1.84788784093918e+01, 1.84654889260433e+01, 1.84478453620027e+01, 1.84251640620669e+01, 1.83966384610856e+01, 1.83614601912339e+01, 1.83188422843261e+01, 1.82680437411223e+01, 1.82083946136054e+01, 1.81393206493879e+01, 1.80603664914278e+01, 1.79712164207849e+01, 1.78717116814299e+01, 1.77618635356325e+01, 1.76418613625465e+01, 1.75120753227142e+01, 1.73730533548631e+01, 1.72255125336904e+01, 1.70703250825058e+01, 1.69084995870565e+01, 1.67411581818851e+01, 1.65695106645733e+01, 1.63948266240059e+01, 1.62184067362038e+01, 1.60415543786011e+01, 1.58655486392439e+01, 1.56916196562812e+01, 1.55209270275161e+01, 1.53545417982993e+01, 1.51934322913481e+01, 1.50384538075837e+01, 1.48903420234543e+01, 1.47497097524543e+01, 1.46170466344148e+01, 1.44927212659697e+01, 1.43769852834850e+01, 1.42699789454768e+01, 1.41717378228848e+01, 1.40822002802660e+01, 1.40012155082452e+01, 1.39285519390110e+01, 1.38639059365195e+01, 1.38069106980713e+01, 1.37571453328759e+01, 1.37141440964733e+01, 1.36774057589482e+01, 1.36464030719779e+01, 1.36205922776418e+01, 1.35994225736304e+01, 1.35823454182901e+01, 1.35688235281726e+01, 1.35583393937375e+01, 1.35504031186982e+01, 1.35445593778864e+01, 1.35403932894943e+01, 1.35375350113323e+01, 1.35356628974917e+01, 1.35345050906369e+01, 1.35338394741286e+01, 1.35334919644899e+01, 1.35333331849599e+01, 1.35332736213565e+01, 1.35332574186197e+01, 1.35332550271300e+01, 1.35332548476001e+01, 1.35332516876707e+01, 1.35332302786905e+01, 1.35331515771117e+01, 1.35329417883429e+01, 1.35324826745341e+01, 1.35316034311524e+01, 1.35300744648649e+01, 1.35276033705589e+01, 1.35238333646495e+01, 1.35183443831959e+01, 1.35106569941317e+01, 1.35002392003537e+01, 1.34865161222275e+01, 1.34688824432476e+01, 1.34467173820498e+01, 1.34194018212344e+01, 1.33863370849429e+01, 1.33469647222148e+01, 1.33007865336955e+01, 1.32473839886519e+01, 1.31864361309909e+01, 1.31177350788500e+01, 1.30411982905013e+01, 1.29568769024010e+01, 1.28649596389639e+01, 1.27657720362166e+01, 1.26597709939859e+01, 1.25475349493361e+01, 1.24297502205972e+01, 1.23071942805479e+01, 1.21807168581301e+01, 1.20512198278965e+01, 1.19196368233071e+01, 1.17869134130923e+01, 1.16539885276470e+01, 1.15217776391342e+01, 1.13911580103846e+01, 1.12629561563551e+01, 1.11379375238161e+01, 1.10167982976186e+01, 1.09001591847315e+01, 1.07885610033299e+01, 1.06824619032688e+01, 1.05822360556362e+01, 1.04881736640963e+01, 1.04004821640308e+01, 1.03192884850741e+01, 1.02446422590611e+01, 1.01765198605093e+01, 1.01148291724286e+01, 1.00594149775698e+01, 1.00100648841055e+01, 9.96651570407824e+00, 9.92846021102483e+00, 9.89555420818991e+00, 9.86742383924833e+00, 9.84367306873636e+00, 9.82389124955905e+00, 9.80766068092895e+00, 9.79456404353846e+00, 9.78419158180216e+00, 9.77614788800884e+00, 9.77005813256771e+00, 9.76557358033224e+00, 9.76237623685869e+00, 9.76018248136383e+00, 9.75874556525442e+00, 9.75785688576448e+00, 9.75734598203741e+00, 9.75707924388135e+00, 9.75695736893088e+00, 9.75691164941577e+00, 9.75689921259785e+00, 9.75689737694952e+00, 9.75689723956587e+00, 9.75689482703996e+00, 9.75687848182373e+00, 9.75681839510678e+00, 9.75665822618751e+00, 9.75630770147697e+00, 9.75563640875288e+00, 9.75446903768778e+00, 9.75258228957395e+00, 9.74970364640722e+00, 9.74551215092254e+00, 9.73964130323868e+00, 9.73168412483911e+00, 9.72120037542741e+00, 9.70772583232939e+00, 9.69078345645921e+00, 9.66989617599115e+00, 9.64460092322377e+00, 9.61446346792704e+00, 9.57909350936534e+00, 9.53815942748846e+00, 9.49140205942778e+00, 9.43864686687990e+00, 9.37981389711172e+00, 9.31492501582861e+00, 9.24410800122061e+00, 9.16759722933445e+00, 9.08573084355987e+00, 8.99894447635582e+00, 8.90776176976158e+00, 8.81278211263949e+00, 8.71466616583062e+00, 8.61411986868253e+00, 8.51187769763671e+00, 8.40868596643609e+00, 8.30528690939573e+00, 8.20240417463130e+00, 8.10073018586545e+00, 8.00091563361652e+00, 7.90356115988084e+00, 7.80921113417022e+00, 7.71834930332449e+00, 7.63139603992902e+00, 7.54870690844157e+00, 7.47057229943566e+00, 7.39721793284836e+00, 7.32880608503666e+00, 7.26543744112011e+00, 7.20715350848515e+00, 7.15393954893567e+00, 7.10572799797238e+00, 7.06240234317861e+00, 7.02380143251481e+00, 6.98972417943613e+00, 6.95993462615866e+00, 6.93416731939225e+00, 6.91213294434636e+00, 6.89352415267181e+00, 6.87802150832390e+00, 6.86529946257158e+00, 6.85503225638432e+00, 6.84689963638107e+00, 6.84059226082363e+00, 6.83581666621814e+00, 6.83229966427235e+00, 6.82979204425739e+00, 6.82807146781961e+00, 6.82694446202362e+00, 6.82624744135364e+00, 6.82584671948573e+00, 6.82563750533923e+00, 6.82554191338956e+00, 6.82550605350402e+00, 6.82549629874144e+00, 6.82549485895875e+00, 6.82549475116215e+00, 6.82549285766088e+00, 6.82548002891120e+00, 6.82543286907102e+00, 6.82530715825752e+00, 6.82503204324539e+00, 6.82450516519305e+00, 6.82358892017374e+00, 6.82210802551216e+00, 6.81984853842032e+00, 6.81655844240531e+00, 6.81194988029479e+00, 6.80570306935162e+00, 6.79747188295378e+00, 6.78689102446840e+00, 6.77358465302492e+00, 6.75717624997997e+00, 6.73729944253133e+00, 6.71360943217547e+00, 6.68579461660995e+00, 6.65358795083114e+00, 6.61677757272660e+00, 6.57521622511713e+00, 6.52882904223637e+00, 6.47761933318655e+00, 6.42167208380541e+00, 6.36115500461158e+00, 6.29631706730509e+00, 6.22748458664050e+00, 6.15505501050284e+00, 6.07948867303145e+00, 6.00129884048636e+00, 5.92104043609243e+00, 5.83929786827086e+00, 5.75667240654198e+00, 5.67376955059363e+00, 5.59118681988100e+00, 5.50950235332136e+00, 5.42926465207625e+00, 5.35098372589541e+00, 5.27512381993397e+00, 5.20209781088131e+00, 5.13226327588248e+00, 5.06592016189127e+00, 5.00330992212930e+00, 4.94461594349747e+00, 4.88996506504474e+00, 4.83942998176411e+00, 4.79303233725047e+00, 4.75074632929441e+00, 4.71250268009091e+00, 4.67819285333277e+00, 4.64767343046735e+00, 4.62077058497999e+00, 4.59728461471362e+00, 4.57699450674787e+00, 4.55966251681288e+00, 4.54503874585406e+00, 4.53286569101336e+00, 4.52288273823901e+00, 4.51483055061495e+00, 4.50845529216500e+00, 4.50351261327189e+00, 4.49977131281865e+00, 4.49701658534475e+00, 4.49505276017766e+00, 4.49370544442598e+00, 4.49282299311054e+00, 4.49227724717966e+00, 4.49196350277106e+00, 4.49179970145541e+00, 4.49172485963323e+00, 4.49169678393883e+00, 4.49168914667460e+00, 4.49168801943059e+00, 4.49168793480989e+00, 4.49168644541468e+00, 4.49167635454682e+00, 4.49163925950650e+00, 4.49154037818914e+00, 4.49132398108087e+00, 4.49090956330119e+00, 4.49018891413284e+00, 4.48902422555397e+00, 4.48724736260276e+00, 4.48466039599747e+00, 4.48103747002883e+00, 4.47612804501389e+00, 4.46966151247372e+00, 4.46135313218929e+00, 4.45091118381599e+00, 4.43804516337721e+00, 4.42247478957569e+00, 4.40393952060776e+00, 4.38220822424906e+00, 4.35708859828232e+00, 4.32843591082286e+00, 4.29616062611210e+00, 4.26023450485744e+00, 4.22069482107187e+00, 4.17764641882032e+00, 4.13126143856778e+00, 4.08177666726328e+00, 4.02948859969605e+00, 3.97474643013683e+00, 3.91794331138991e+00, 3.85950631249479e+00, 3.79988556797324e+00, 3.73954313560110e+00, 3.67894206515539e+00, 3.61853613062413e+00, 3.55876059984796e+00, 3.50002431800746e+00, 3.44270327549452e+00, 3.38713572692247e+00, 3.33361883514281e+00, 3.28240673845625e+00, 3.23370988417879e+00, 3.18769543804830e+00, 3.14448856508097e+00, 3.10417438029607e+00, 3.06680038328399e+00, 3.03237921476948e+00, 3.00089160228038e+00, 2.97228939250077e+00, 2.94649859729004e+00, 2.92342240679494e+00, 2.90294414529813e+00, 2.88493016267386e+00, 2.86923266621157e+00, 2.85569250410596e+00, 2.84414191335780e+00, 2.83440724168233e+00, 2.82631164600092e+00, 2.81967776011712e+00, 2.81433031233802e+00, 2.81009866132074e+00, 2.80681920658347e+00, 2.80433762018826e+00, 2.80251083922120e+00, 2.80120875578575e+00, 2.80031554288202e+00, 2.79973056097614e+00, 2.79936880104597e+00, 2.79916083478755e+00, 2.79905226049526e+00, 2.79900265265207e+00, 2.79898404314225e+00, 2.79897898091424e+00, 2.79897823374036e+00, 2.79897817741213e+00, 2.79897718280619e+00, 2.79897044420994e+00, 2.79894567249816e+00, 2.79887964083048e+00, 2.79873513529198e+00, 2.79845840197332e+00, 2.79797719819679e+00, 2.79719954641715e+00, 2.79601327838006e+00, 2.79428644343506e+00, 2.79186863721564e+00, 2.78859328388799e+00, 2.78428087570985e+00, 2.77874313712883e+00, 2.77178803731724e+00, 2.76322552621319e+00, 2.75287381739339e+00, 2.74056599026709e+00, 2.72615663901053e+00, 2.70952826180567e+00, 2.69059706669645e+00, 2.66931787426054e+00, 2.64568782516973e+00, 2.61974865303309e+00, 2.59158735731830e+00, 2.56133520239643e+00, 2.52916506920089e+00, 2.49528728644302e+00, 2.45994415926481e+00, 2.42340348611174e+00, 2.38595140313028e+00, 2.34788491619749e+00, 2.30950447372997e+00, 2.27110690170158e+00, 2.23297897115096e+00, 2.19539180454863e+00, 2.15859625768152e+00, 2.12281934458733e+00, 2.08826170973719e+00, 2.05509609786637e+00, 2.02346672984652e+00, 1.99348946370254e+00, 1.96525260312109e+00, 1.93881821053834e+00, 1.91422378646640e+00, 1.89148418904860e+00, 1.87059368564813e+00, 1.85152804928608e+00, 1.83424663483910e+00, 1.81869439126960e+00, 1.80480378537223e+00, 1.79249662858223e+00, 1.78168581071196e+00, 1.77227695282320e+00, 1.76416999585484e+00, 1.75726074237755e+00, 1.75144236637915e+00, 1.74660690085963e+00, 1.74264670590288e+00, 1.73945591153775e+00, 1.73693182090928e+00, 1.73497625087931e+00, 1.73349677997476e+00, 1.73240786833510e+00, 1.73163181157451e+00, 1.73109949066831e+00, 1.73075088324688e+00, 1.73053530791378e+00, 1.73041138200853e+00, 1.73034668398473e+00, 1.73031712348659e+00, 1.73030603441433e+00, 1.73030301792642e+00, 1.73030257269943e+00, 1.73030253895184e+00, 1.73030194063664e+00, 1.73029788696768e+00, 1.73028298531958e+00, 1.73024326351813e+00, 1.73015633597196e+00, 1.72998986945591e+00, 1.72970041449936e+00, 1.72923266347217e+00, 1.72851919057388e+00, 1.72748072128731e+00, 1.72602696852245e+00, 1.72405805865550e+00, 1.72146655210769e+00, 1.71814003952166e+00, 1.71396426608290e+00, 1.70882670399488e+00, 1.70262045834752e+00, 1.69524835733166e+00, 1.68662704741059e+00, 1.67669089156617e+00, 1.66539545794667e+00, 1.65272039038152e+00, 1.63867147329113e+00, 1.62328174174184e+00, 1.60661154095345e+00, 1.58874750453048e+00, 1.56980049135512e+00, 1.54990259059903e+00, 1.52920336556472e+00, 1.50786555366732e+00, 1.48606046708384e+00, 1.46396334408363e+00, 1.44174888518783e+00, 1.41958717410543e+00, 1.39764013597188e+00, 1.37605863110193e+00, 1.35498022775541e+00, 1.33452764797247e+00, 1.31480784047044e+00, 1.29591160610748e+00, 1.27791368485409e+00, 1.26087320742106e+00, 1.24483441757783e+00, 1.22982758025501e+00, 1.21587000331662e+00, 1.20296711531723e+00, 1.19111355602354e+00, 1.18029424988197e+00, 1.17048544429046e+00, 1.16165570416130e+00, 1.15376686174761e+00, 1.14677492610204e+00, 1.14063095997069e+00, 1.13528193356692e+00, 1.13067156470414e+00, 1.12674115339367e+00, 1.12343041644579e+00, 1.12067832409353e+00, 1.11842393646087e+00, 1.11660723313604e+00, 1.11516992453463e+00, 1.11405622952118e+00, 1.11321360027834e+00, 1.11259337302051e+00, 1.11215132213221e+00, 1.11184809586663e+00, 1.11164951393417e+00, 1.11152671107782e+00, 1.11145611585737e+00, 1.11141926001640e+00, 1.11140242055823e+00, 1.11139610354139e+00, 1.11139438516393e+00, 1.11139413153517e+00, 1.11139411218847e+00, 1.11139376757911e+00, 1.11139143280241e+00, 1.11138284995211e+00, 1.11135997149930e+00, 1.11130990395986e+00, 1.11121402397714e+00, 1.11104730460585e+00, 1.11077788666553e+00, 1.11036692650566e+00, 1.10976874698516e+00, 1.10893131239087e+00, 1.10779703984318e+00, 1.10630394898197e+00, 1.10438713812378e+00, 1.10198055871408e+00, 1.09901904131677e+00, 1.09544050668404e+00, 1.09118827626895e+00, 1.08621337995653e+00, 1.08047674709488e+00, 1.07395116230404e+00, 1.06662287175288e+00, 1.05849273948905e+00, 1.04957687670723e+00, 1.03990669799759e+00, 1.02952839490546e+00, 1.01850185498598e+00, 1.00689909004472e+00, 9.94802266768055e-01, 9.82301453630058e-01, 9.69492208197195e-01, 9.56473128503205e-01, 9.43343482098634e-01, 9.30201008762374e-01, 9.17139970350860e-01, 9.04249496650174e-01, 8.91612251936621e-01, 8.79303425277435e-01, 8.67390029798493e-01, 8.55930482923199e-01, 8.44974431067591e-01, 8.34562778117601e-01, 8.24727876556544e-01, 8.15493842528859e-01, 8.06876960547690e-01, 7.98886149157127e-01, 7.91523464933142e-01, 7.84784628181704e-01, 7.78659559150512e-01, 7.73132918234509e-01, 7.68184647368024e-01, 7.63790512495099e-01, 7.59922648700929e-01, 7.56550110325037e-01, 7.53639428250082e-01, 7.51155175675969e-01, 7.49060542177332e-01, 7.47317913849095e-01, 7.45889455033542e-01, 7.44737684677183e-01, 7.43826037984562e-01, 7.43119401927331e-01, 7.42584611534433e-01, 7.42190892918543e-01, 7.41910238835771e-01, 7.41717703321939e-01, 7.41591603627221e-01, 7.41513620227337e-01, 7.41468788995775e-01, 7.41445383481782e-01, 7.41434689407107e-01, 7.41430677706340e-01, 7.41429586427955e-01, 7.41429425357602e-01, 7.41429412984537e-01, 7.41429191454735e-01, 7.41427690558794e-01, 7.41422173121691e-01, 7.41407465776774e-01, 7.41375279703888e-01, 7.41313641791081e-01, 7.41206460036871e-01, 7.41033244864404e-01, 7.40769005466070e-01, 7.40384337130614e-01, 7.39845711460063e-01, 7.39115976171938e-01, 7.38155064563372e-01, 7.36920906560869e-01, 7.35370523656101e-01, 7.33461279251024e-01, 7.31152244615124e-01, 7.28405629696304e-01, 7.25188218562124e-01, 7.21472742539049e-01, 7.17239121373069e-01, 7.12475504911197e-01, 7.07179055397589e-01, 7.01356423375067e-01, 6.95023887572617e-01, 6.88207149588877e-01, 6.80940795707500e-01, 6.73267458642918e-01, 6.65236729346489e-01, 6.56903881544793e-01, 6.48328478415131e-01, 6.39572931479215e-01, 6.30701076916864e-01, 6.21776825188135e-01, 6.12862927615751e-01, 6.04019890063178e-01, 5.95305050572740e-01, 5.86771826023408e-01, 5.78469123338472e-01, 5.70440903890031e-01, 5.62725885497414e-01, 5.55357364493958e-01, 5.48363140264748e-01, 5.41765525891279e-01, 5.35581430561798e-01, 5.29822501778909e-01, 5.24495317791285e-01, 5.19601622871846e-01, 5.15138599933195e-01, 5.11099176452781e-01, 5.07472360758506e-01, 5.04243606406831e-01, 5.01395202687945e-01, 4.98906689234822e-01, 4.96755292320636e-01, 4.94916379732189e-01, 4.93363930149265e-01, 4.92071011801897e-01, 4.91010263898104e-01, 4.90154373014825e-01, 4.89476535437944e-01, 4.88950895448298e-01, 4.88552948902282e-01, 4.88259901257152e-01, 4.88050969528338e-01, 4.87907618586021e-01, 4.87813723707362e-01, 4.87755653352117e-01, 4.87722268635061e-01, 4.87704838795565e-01, 4.87696874955972e-01, 4.87693887444493e-01, 4.87693074768899e-01, 4.87692954819678e-01, 4.87692945535123e-01, 4.87692778387256e-01, 4.87691645936096e-01, 4.87687482929587e-01, 4.87676385911428e-01, 4.87652100504377e-01, 4.87605591603064e-01, 4.87524713980221e-01, 4.87393998925846e-01, 4.87194570330438e-01, 4.86904200994850e-01, 4.86497516766782e-01, 4.85946352228801e-01, 4.85220256980988e-01, 4.84287146026176e-01, 4.83114081440554e-01, 4.81668165599999e-01, 4.79917519103601e-01, 4.77832309718025e-01, 4.75385792817896e-01, 4.72555319633742e-01, 4.69323267820165e-01, 4.65677849950616e-01, 4.61613759794545e-01, 4.57132623552731e-01, 4.52243233150007e-01, 4.46961550396502e-01, 4.41310483265426e-01, 4.35319447543421e-01, 4.29023737614202e-01, 4.22463738306973e-01, 4.15684015095479e-01, 4.08732322373169e-01, 4.01658569302981e-01, 3.94513780339210e-01, 3.87349083558739e-01, 3.80214755029066e-01, 3.73159342099470e-01, 3.66228883107119e-01, 3.59466235781738e-01, 3.52910521747333e-01, 3.46596690030736e-01, 3.40555198456201e-01, 3.34811808296756e-01, 3.29387484645061e-01, 3.24298392736836e-01, 3.19555978966761e-01, 3.15167124599810e-01, 3.11134360168118e-01, 3.07456129166162e-01, 3.04127090776980e-01, 3.01138452805368e-01, 2.98478327569349e-01, 2.96132105020938e-01, 2.94082838667009e-01, 2.92311640812535e-01, 2.90798084172835e-01, 2.89520606971251e-01, 2.88456918280225e-01, 2.87584399650108e-01, 2.86880498116503e-01, 2.86323104623186e-01, 2.85890910900370e-01, 2.85563737050737e-01, 2.85322821654738e-01, 2.85151066219903e-01, 2.85033226329773e-01, 2.84956042915226e-01, 2.84908308640013e-01, 2.84880866385365e-01, 2.84866539117303e-01, 2.84859992884364e-01, 2.84857537168676e-01, 2.84856869154680e-01, 2.84856770557226e-01, 2.84856762863505e-01, 2.84856623555854e-01, 2.84855679726412e-01, 2.84852210114735e-01, 2.84842961443772e-01, 2.84822721156061e-01, 2.84783959359785e-01, 2.84716554700040e-01, 2.84607617375606e-01, 2.84441420767251e-01, 2.84199451042589e-01, 2.83860581536892e-01, 2.83401375556473e-01, 2.82796517414504e-01, 2.82019366946336e-01, 2.81042627504981e-01, 2.79839111661659e-01, 2.78382582804007e-01, 2.76648644927604e-01, 2.74615647641134e-01, 2.72265569289743e-01, 2.69584838674518e-01, 2.66565055563049e-01, 2.63203572351685e-01, 2.59503903960230e-01, 2.55475940181846e-01, 2.51135943904934e-01, 2.46506329290476e-01, 2.41615225392451e-01, 2.36495842035566e-01, 2.31185665188434e-01, 2.25725517840033e-01, 2.20158528873036e-01, 2.14529056184645e-01, 2.08881611094689e-01, 2.03259828900627e-01, 1.97705525524576e-01, 1.92257873007371e-01, 1.86952717785701e-01, 1.81822056008116e-01, 1.76893670415879e-01, 1.72190924299599e-01, 1.67732700379496e-01, 1.63533466598210e-01, 1.59603446994510e-01, 1.55948874060194e-01, 1.52572299100001e-01, 1.49472938795519e-01, 1.46647039018198e-01, 1.44088240507443e-01, 1.41787934912699e-01, 1.39735603530951e-01, 1.37919134564022e-01, 1.36325117665024e-01, 1.34939116811438e-01, 1.33745924075079e-01, 1.32729797656680e-01, 1.31874687666233e-01, 1.31164452648505e-01, 1.30583068894417e-01, 1.30114833286257e-01, 1.29744558953583e-01, 1.29457761528002e-01, 1.29240832437780e-01, 1.29081194618273e-01, 1.28967435346290e-01, 1.28889410713384e-01, 1.28838316568675e-01, 1.28806721568694e-01, 1.28788559210832e-01, 1.28779077296024e-01, 1.28774745038797e-01, 1.28773119877025e-01, 1.28772677795261e-01, 1.28772612544978e-01, 1.28772607410077e-01, 1.28772513878893e-01, 1.28771880192424e-01, 1.28769550709009e-01, 1.28763341277462e-01, 1.28749752676868e-01, 1.28723731136162e-01, 1.28678486515881e-01, 1.28605378131610e-01, 1.28493877448811e-01, 1.28331615838981e-01, 1.28104524078374e-01, 1.27797068102389e-01, 1.27392582536153e-01, 1.26873699606364e-01, 1.26222866204422e-01, 1.25422936259227e-01, 1.24457819484654e-01, 1.23313161450085e-01, 1.21977024367290e-01, 1.20440533669722e-01, 1.18698453064184e-01, 1.16749650865252e-01, 1.14597423506567e-01, 1.12249648317686e-01, 1.09718746779862e-01, 1.07021450985144e-01, 1.04178379026870e-01, 1.01213438386218e-01, 9.81530887392807e-02, 9.50255057019405e-02, 9.18596937524754e-02, 8.86845991760055e-02, 8.55282720828896e-02, 8.24171206111899e-02, 7.93752910777172e-02, 7.64241962256946e-02, 7.35822011881879e-02, 7.08644647162586e-02, 6.82829228099990e-02, 6.58463940032377e-02, 6.35607806412182e-02, 6.14293385769216e-02, 5.94529884530459e-02, 5.76306445379942e-02, 5.59595412273097e-02, 5.44355421023326e-02, 5.30534212494516e-02, 5.18071109432086e-02, 5.06899135079877e-02, 4.96946780752552e-02, 4.88139450406351e-02, 4.80400623715767e-02, 4.73652786350414e-02, 4.67818178312314e-02, 4.62819409504220e-02, 4.58579987139629e-02, 4.55024792982428e-02, 4.52080540336297e-02, 4.49676231720915e-02, 4.47743628750601e-02, 4.46217736328833e-02, 4.45037294396045e-02, 4.44145262613631e-02, 4.43489277053501e-02, 4.43022053625697e-02, 4.42701710976579e-02, 4.42491986088811e-02, 4.42362318786067e-02, 4.42287786541371e-02, 4.42248877931348e-02, 4.42231101165191e-02, 4.42224432632088e-02, 4.42222618643058e-02, 4.42222350902597e-02, 4.42222329646071e-02, 4.42221940093503e-02, 4.42219300825286e-02, 4.42209598715024e-02, 4.42183737363990e-02, 4.42127145061235e-02, 4.42018781776810e-02, 4.41830392634657e-02, 4.41526054703090e-02, 4.41062062773767e-02, 4.40387197032734e-02, 4.39443409798841e-02, 4.38166958384514e-02, 4.36489995318112e-02, 4.34342604849220e-02, 4.31655245822711e-02, 4.28361526761591e-02, 4.24401201777798e-02, 4.19723239488050e-02, 4.14288786304979e-02, 4.08073825692270e-02, 4.01071331331480e-02, 3.93292728449524e-02, 3.84768515326499e-02, 3.75547954665763e-02, 3.65697817114849e-02, 3.55300238823801e-02, 3.44449831594389e-02, 3.33250247696100e-02, 3.21810443156778e-02, 3.10240897897458e-02, 2.98650037442627e-02, 2.87141062556557e-02, 2.75809337161022e-02, 2.64740420466651e-02, 2.54008765763700e-02, 2.43677053450497e-02, 2.33796084635571e-02, 2.24405135895438e-02, 2.15532664775893e-02, 2.07197256881884e-02, 1.99408715508014e-02, 1.92169210252047e-02, 1.85474418785745e-02, 1.79314613532213e-02, 1.73675660823278e-02, 1.68539913437158e-02, 1.63886988094575e-02, 1.59694427742139e-02, 1.55938254565058e-02, 1.52593423947636e-02, 1.49634192295878e-02, 1.47034413033609e-02, 1.44767775453118e-02, 1.42808000712691e-02, 1.41129008285290e-02, 1.39705064697822e-02, 1.38510924507287e-02, 1.37521971197927e-02, 1.36714363122152e-02, 1.36065186862382e-02, 1.35552617594712e-02, 1.35156083356236e-02, 1.34856427723378e-02, 1.34636063478837e-02, 1.34479108532801e-02, 1.34371494799623e-02, 1.34301040963791e-02, 1.34257481124322e-02, 1.34232443064377e-02, 1.34219372252389e-02, 1.34213400389882e-02, 1.34211160186481e-02, 1.34210550801441e-02, 1.34210460857673e-02, 1.34210453651491e-02, 1.34210320765362e-02, 1.34209420445615e-02, 1.34206110816208e-02, 1.34197288874915e-02, 1.34177983875006e-02, 1.34141018631411e-02, 1.34076754973833e-02, 1.33972939374871e-02, 1.33814664403502e-02, 1.33584460707681e-02, 1.33262532268787e-02, 1.32827144156025e-02, 1.32255166528950e-02, 1.31522770887926e-02, 1.30606264581132e-02, 1.29483037766676e-02, 1.28132584289276e-02, 1.26537545623719e-02, 1.24684716881944e-02, 1.22565947728788e-02, 1.20178870577163e-02, 1.17527394771031e-02, 1.14621918874818e-02, 1.11479232858898e-02, 1.08122105984521e-02, 1.04578581758972e-02, 1.00881025223671e-02, 9.70649869839235e-03, 9.31679604827451e-03, 8.92281129979011e-03, 8.52830669462179e-03, 8.13687976670252e-03, 7.75186988433774e-03, 7.37628491204134e-03, 7.01274951271813e-03, 6.66347487263863e-03, 6.33024816843175e-03, 6.01443906631260e-03, 5.71702001040819e-03, 5.43859695131488e-03, 5.17944732876430e-03, 4.93956236775261e-03, 4.71869098214484e-03, 4.51638288753031e-03, 4.33202902777197e-03, 4.16489806634246e-03, 4.01416833225454e-03, 3.87895512105274e-03, 3.75833354563326e-03, 3.65135720676335e-03, 3.55707296635188e-03, 3.47453220260263e-03, 3.40279908080277e-03, 3.34095649077426e-03, 3.28811033088527e-03, 3.24339276089347e-03, 3.20596493127567e-03, 3.17501956623545e-03, 3.14978365987884e-03, 3.12952144128076e-03, 3.11353766741718e-03, 3.10118120575709e-03, 3.09184877648505e-03, 3.08498864255569e-03, 3.08010397561276e-03, 3.07675559146503e-03, 3.07456374576276e-03, 3.07320870698646e-03, 3.07242987796741e-03, 3.07202331232759e-03, 3.07183756130058e-03, 3.07176788163588e-03, 3.07174892726181e-03, 3.07174612964900e-03, 3.07174590340509e-03, 3.07174170489333e-03, 3.07171325946598e-03, 3.07160869274981e-03, 3.07132996929482e-03, 3.07072005577425e-03, 3.06955225100580e-03, 3.06752221901974e-03, 3.06424326865340e-03, 3.05924542566771e-03, 3.05197882548924e-03, 3.04182189875490e-03, 3.02809470725407e-03, 3.01007759762939e-03, 2.98703506103237e-03, 2.95824432040348e-03, 2.92302772479047e-03, 2.88078754839916e-03, 2.83104131843004e-03, 2.77345540107632e-03, 2.70787433003505e-03, 2.63434334829071e-03, 2.55312192033585e-03, 2.46468658330912e-03, 2.36972234061028e-03, 2.26910263774016e-03, 2.16385869553193e-03, 2.05513983843986e-03, 1.94416746409587e-03, 1.83218557238263e-03, 1.72041016283941e-03, 1.60998137357637e-03, 1.50192678321304e-03, 1.39714156282622e-03, 1.29637513608571e-03, 1.20020661928777e-03, 1.10901463477830e-03, 1.02271311911295e-03, 9.43042941202066e-04, 8.68258042238779e-04, 7.98341485811704e-04, 7.33136831136960e-04, 6.72434756737061e-04, 6.15996926910611e-04, 5.63294285342408e-04, 5.14562279704478e-04, 4.68970901074783e-04, 4.26237483942771e-04, 3.86089039793092e-04, 3.48322165244497e-04, 3.12792486106162e-04, 2.79404068230349e-04, 2.48100120685990e-04, 2.18945731909059e-04, 1.91843553736248e-04, 1.66700711645262e-04, 1.43523326514128e-04, 1.22313500781956e-04, 1.03064428046196e-04, 8.57575037969127e-05, 7.03583286352227e-05, 5.68132778298036e-05, 4.50507176310566e-05, 3.49840045714700e-05, 2.65124187990295e-05, 1.95215774283497e-05, 1.38846056714215e-05, 9.46402415991088e-06, 6.11409524964301e-06, 3.68346002943554e-06, 2.01799029042604e-06, 9.63851817914474e-07, 3.70850266051548e-07, -1.42108547152020e-14, -1.42108547152020e-14],
[ 4.50000000000000e+01, 4.49999999601259e+01, 4.49999987271391e+01, 4.49999903734589e+01, 4.49999596642458e+01, 4.49998778032166e+01, 4.49996986445309e+01, 4.49993555042270e+01, 4.49987586865147e+01, 4.49977938184459e+01, 4.49963210618125e+01, 4.49941752444852e+01, 4.49911669258344e+01, 4.49870843834945e+01, 4.49816964826399e+01, 4.49747563651600e+01, 4.49660058754314e+01, 4.49551806222965e+01, 4.49420155634413e+01, 4.49262509883047e+01, 4.49076387682289e+01, 4.48859487367986e+01, 4.48609750581962e+01, 4.48325424361166e+01, 4.48005120100048e+01, 4.47647867795041e+01, 4.47253163932098e+01, 4.46821011360497e+01, 4.46351949532510e+01, 4.45847073603385e+01, 4.45308041099127e+01, 4.44737065180675e+01, 4.44136893958468e+01, 4.43510775822594e+01, 4.42862411318008e+01, 4.42195892669489e+01, 4.41515632600384e+01, 4.40826284548757e+01, 4.40132656728576e+01, 4.39439622689411e+01, 4.38752031088095e+01, 4.38074617306909e+01, 4.37411919354399e+01, 4.36768200194277e+01, 4.36147378296218e+01, 4.35552967820435e+01, 4.34988029462369e+01, 4.34455132615568e+01, 4.33956329173623e+01, 4.33493138993174e+01, 4.33066546781178e+01, 4.32677009948349e+01, 4.32324476782038e+01, 4.32008414129974e+01, 4.31727843645547e+01, 4.31481385521375e+01, 4.31267308527943e+01, 4.31083585077484e+01, 4.30927949951054e+01, 4.30797961261791e+01, 4.30691062183442e+01, 4.30604641955172e+01, 4.30536094685825e+01, 4.30482874527456e+01, 4.30442545871879e+01, 4.30412827346553e+01, 4.30391628546904e+01, 4.30377078638379e+01, 4.30367546188662e+01, 4.30361649842019e+01, 4.30358259715669e+01, 4.30356489673521e+01, 4.30355680905940e+01, 4.30355377505949e+01, 4.30355294973479e+01, 4.30355282791847e+01, 4.30355281193874e+01, 4.30355243963016e+01, 4.30354991718281e+01, 4.30354064437092e+01, 4.30351592623912e+01, 4.30346183009702e+01, 4.30335822524049e+01, 4.30317804215920e+01, 4.30288678252996e+01, 4.30244230506132e+01, 4.30179490532013e+01, 4.30088770010933e+01, 4.29965731886453e+01, 4.29803489600099e+01, 4.29594734929791e+01, 4.29331892038446e+01, 4.29007294431144e+01, 4.28613380615278e+01, 4.28142903366725e+01, 4.27589146638524e+01, 4.26946143331170e+01, 4.26208886422627e+01, 4.25373525411139e+01, 4.24437539767836e+01, 4.23399881265025e+01, 4.22261077771477e+01, 4.21023292477263e+01, 4.19690334532196e+01, 4.18267619643436e+01, 4.16762082048490e+01, 4.15182042136028e+01, 4.13537036470781e+01, 4.11837618773712e+01, 4.10095141311055e+01, 4.08321526110643e+01, 4.06529034570584e+01, 4.04730042601783e+01, 4.02936826763769e+01, 4.01161365212750e+01, 3.99415155907844e+01, 3.97709053532177e+01, 3.96053125984568e+01, 3.94456531003365e+01, 3.92927413371520e+01, 3.91472823095472e+01, 3.90098654856281e+01, 3.88809608853637e+01, 3.87609172900751e+01, 3.86499625310778e+01, 3.85482057785425e+01, 3.84556417210855e+01, 3.83721565007638e+01, 3.82975352475095e+01, 3.82314710406826e+01, 3.81735751116499e+01, 3.81233880883222e+01, 3.80803920690388e+01, 3.80440232985092e+01, 3.80136852029360e+01, 3.79887615260137e+01, 3.79686292938513e+01, 3.79526713270012e+01, 3.79402880138101e+01, 3.79309080631797e+01, 3.79239979681287e+01, 3.79190699352794e+01, 3.79156880698484e+01, 3.79134726504037e+01, 3.79121023812861e+01, 3.79113145711765e+01, 3.79109032512802e+01, 3.79107153130774e+01, 3.79106448105251e+01, 3.79106256320795e+01, 3.79106228013796e+01, 3.79106225514827e+01, 3.79106176548754e+01, 3.79105844796300e+01, 3.79104625237627e+01, 3.79101374329982e+01, 3.79094259737315e+01, 3.79080634208227e+01, 3.79056938569649e+01, 3.79018638160164e+01, 3.78960196283896e+01, 3.78875087414303e+01, 3.78755851903395e+01, 3.78594192841306e+01, 3.78381114455150e+01, 3.78107100037723e+01, 3.77762325877330e+01, 3.77336906064116e+01, 3.76821161444981e+01, 3.76205904481230e+01, 3.75482730440028e+01, 3.74644304338829e+01, 3.73684632470736e+01, 3.72599307258115e+01, 3.71385714672632e+01, 3.70043194551465e+01, 3.68573145833042e+01, 3.66979071009156e+01, 3.65266556894807e+01, 3.63443192054904e+01, 3.61518424712959e+01, 3.59503368390977e+01, 3.57410565457707e+01, 3.55253720712186e+01, 3.53047417723523e+01, 3.50806829786588e+01, 3.48547435311346e+01, 3.46284744829427e+01, 3.44034044267754e+01, 3.41810157244411e+01, 3.39627228099213e+01, 3.37498527057406e+01, 3.35436279009588e+01, 3.33451517516419e+01, 3.31553965563997e+01, 3.29751944217873e+01, 3.28052309701029e+01, 3.26460418680267e+01, 3.24980120823741e+01, 3.23613777094547e+01, 3.22362301824569e+01, 3.21225226373057e+01, 3.20200782083552e+01, 3.19286000259863e+01, 3.18476826931808e+01, 3.17768250226560e+01, 3.17154438166701e+01, 3.16628884662264e+01, 3.16184561347238e+01, 3.15814072740542e+01, 3.15509812007845e+01, 3.15264114391590e+01, 3.15069405194618e+01, 3.14918339081266e+01, 3.14803927429833e+01, 3.14719650557825e+01, 3.14659551864054e+01, 3.14618311297367e+01, 3.14591296066906e+01, 3.14574587138594e+01, 3.14564980791740e+01, 3.14559965304950e+01, 3.14557673662950e+01, 3.14556813984930e+01, 3.14556580131287e+01, 3.14556545614966e+01, 3.14556542846330e+01, 3.14556491750952e+01, 3.14556145572159e+01, 3.14554872980981e+01, 3.14551480710031e+01, 3.14544056756773e+01, 3.14529838801740e+01, 3.14505113044048e+01, 3.14465147996146e+01, 3.14404167016299e+01, 3.14315362479850e+01, 3.14190953481964e+01, 3.14022287808143e+01, 3.13799987592959e+01, 3.13514136613762e+01, 3.13154505555673e+01, 3.12710809882881e+01, 3.12172993231824e+01, 3.11531527603304e+01, 3.10777720192122e+01, 3.09904015584501e+01, 3.08904281402867e+01, 3.07774065394658e+01, 3.06510812525069e+01, 3.05114031877847e+01, 3.03585405076717e+01, 3.01928830440973e+01, 3.00150400055367e+01, 2.98258310187815e+01, 2.96262708805228e+01, 2.94175487063235e+01, 2.92010024315749e+01, 2.89780898164318e+01, 2.87503572164615e+01, 2.85194073945362e+01, 2.82868675710374e+01, 2.80543587544688e+01, 2.78234671881950e+01, 2.75957185209172e+01, 2.73725550870994e+01, 2.71553164910620e+01, 2.69452235376211e+01, 2.67433654458999e+01, 2.65506902162403e+01, 2.63679979833464e+01, 2.61959371712394e+01, 2.60350032583528e+01, 2.58855399583024e+01, 2.57477426208140e+01, 2.56216636574164e+01, 2.55072197981587e+01, 2.54042009889419e+01, 2.53122807434638e+01, 2.52310277678067e+01, 2.51599186774355e+01, 2.50983516239056e+01, 2.50456606405487e+01, 2.50011305023582e+01, 2.49640118758458e+01, 2.49335365114948e+01, 2.49089322071635e+01, 2.48894372486233e+01, 2.48743140168219e+01, 2.48628614438262e+01, 2.48544260036438e+01, 2.48484109423345e+01, 2.48442834850771e+01, 2.48415798059746e+01, 2.48399076079670e+01, 2.48389462327130e+01, 2.48384443002271e+01, 2.48382149613103e+01, 2.48381289280707e+01, 2.48381055249158e+01, 2.48381020706582e+01, 2.48381018042326e+01, 2.48380970200840e+01, 2.48380646067676e+01, 2.48379454519004e+01, 2.48376278281017e+01, 2.48369327127846e+01, 2.48356014707568e+01, 2.48332863889024e+01, 2.48295444916162e+01, 2.48238349954960e+01, 2.48155206807649e+01, 2.48038733631187e+01, 2.47880835412084e+01, 2.47672741704223e+01, 2.47405183729855e+01, 2.47068607396179e+01, 2.46653417136207e+01, 2.46150243816571e+01, 2.45550228367405e+01, 2.44845311402203e+01, 2.44028518040658e+01, 2.43094226551800e+01, 2.42038409402872e+01, 2.40858835897192e+01, 2.39555226826950e+01, 2.38129353414694e+01, 2.36585075179287e+01, 2.34928314108228e+01, 2.33166965492545e+01, 2.31310748812870e+01, 2.29371004974923e+01, 2.27360448785506e+01, 2.25292887629403e+01, 2.23182918646569e+01, 2.21045617148383e+01, 2.18896228472655e+01, 2.16749874022586e+01, 2.14621280086029e+01, 2.12524535531567e+01, 2.10472882007176e+01, 2.08478538142180e+01, 2.06552557653022e+01, 2.04704720201390e+01, 2.02943453252607e+01, 2.01275782882274e+01, 1.99707311337851e+01, 1.98242219084492e+01, 1.96883289011033e+01, 1.95631950442967e+01, 1.94488340622650e+01, 1.93451381388081e+01, 1.92518868911125e+01, 1.91687574528035e+01, 1.90953354880506e+01, 1.90311269750189e+01, 1.89755706081700e+01, 1.89280506725505e+01, 1.88879102381851e+01, 1.88544645093268e+01, 1.88270141432447e+01, 1.88048583291901e+01, 1.87873073936970e+01, 1.87736946773089e+01, 1.87633874140443e+01, 1.87557963417597e+01, 1.87503837815993e+01, 1.87466699493484e+01, 1.87442373008965e+01, 1.87427327669842e+01, 1.87418677965936e+01, 1.87414162003875e+01, 1.87412098615039e+01, 1.87411324565280e+01, 1.87411114004848e+01, 1.87411082926565e+01, 1.87411080573411e+01, 1.87411038887525e+01, 1.87410756459507e+01, 1.87409718224118e+01, 1.87406950668845e+01, 1.87400893937177e+01, 1.87389294553684e+01, 1.87369123072970e+01, 1.87336520448981e+01, 1.87286776334099e+01, 1.87214341833553e+01, 1.87112878438826e+01, 1.86975343921880e+01, 1.86794114874034e+01, 1.86561144315848e+01, 1.86268151403420e+01, 1.85906838752620e+01, 1.85469131363537e+01, 1.84947429646329e+01, 1.84334867740363e+01, 1.83625567305208e+01, 1.82814876365263e+01, 1.81899582711050e+01, 1.80878091866764e+01, 1.79750560746896e+01, 1.78518979814385e+01, 1.77187198738761e+01, 1.75760893112721e+01, 1.74247472567727e+01, 1.72655933464222e+01, 1.70996662044552e+01, 1.69281196352900e+01, 1.67521957181234e+01, 1.65731959646145e+01, 1.63924517622754e+01, 1.62112953092837e+01, 1.60310321507610e+01, 1.58529162604160e+01, 1.56781283910638e+01, 1.55077581653671e+01, 1.53427901197880e+01, 1.51840936750460e+01, 1.50324168058414e+01, 1.48883830346786e+01, 1.47524912846108e+01, 1.46251180912493e+01, 1.45065216870006e+01, 1.43968475181575e+01, 1.42961348248420e+01, 1.42043239923543e+01, 1.41212644598127e+01, 1.40467230404766e+01, 1.39803925630423e+01, 1.39219007822131e+01, 1.38708195296114e+01, 1.38266740836560e+01, 1.37889527312769e+01, 1.37571164777498e+01, 1.37306088362799e+01, 1.37088655992417e+01, 1.36913244612975e+01, 1.36774343341574e+01, 1.36666641666288e+01, 1.36585110647636e+01, 1.36525074979010e+01, 1.36482273791273e+01, 1.36452908242565e+01, 1.36433674219765e+01, 1.36421778883851e+01, 1.36414940297769e+01, 1.36411369954282e+01, 1.36409738638220e+01, 1.36409126675790e+01, 1.36408960207223e+01, 1.36408935636813e+01, 1.36408933792305e+01, 1.36408901326867e+01, 1.36408681368802e+01, 1.36407872780519e+01, 1.36405717387395e+01, 1.36401000394564e+01, 1.36391966912759e+01, 1.36376257999350e+01, 1.36350869258391e+01, 1.36312134619145e+01, 1.36255737406805e+01, 1.36176750206303e+01, 1.36069704272753e+01, 1.35928688338034e+01, 1.35747475592435e+01, 1.35519676393340e+01, 1.35238912905778e+01, 1.34899010478511e+01, 1.34494199200309e+01, 1.34019317885529e+01, 1.33470011842818e+01, 1.32842915322408e+01, 1.32135809634147e+01, 1.31347748658639e+01, 1.30479144857826e+01, 1.29531810876580e+01, 1.28508954283434e+01, 1.27415125725328e+01, 1.26256123515548e+01, 1.25038860162923e+01, 1.23771198329931e+01, 1.22461764982918e+01, 1.21119752967227e+01, 1.19754718912536e+01, 1.18376385367010e+01, 1.16994453576456e+01, 1.15618431615232e+01, 1.14257480885664e+01, 1.12920282531280e+01, 1.11614924178326e+01, 1.10348806663754e+01, 1.09128569983702e+01, 1.07960037511553e+01, 1.06848177480371e+01, 1.05797080706700e+01, 1.04809953492633e+01, 1.03889124562445e+01, 1.03036064784346e+01, 1.02251418330237e+01, 1.01535043870471e+01, 1.00886064407870e+01, 1.00302924427788e+01, 9.97834531633980e+00, 9.93249329200024e+00, 9.89241715374914e+00, 9.85775781679902e+00, 9.82812415866294e+00, 9.80310102286906e+00, 9.78225730587450e+00, 9.76515402392241e+00, 9.75135223969392e+00, 9.74042071109138e+00, 9.73194310899487e+00, 9.72552463992206e+00, 9.72079790542988e+00, 9.71742783450977e+00, 9.71511553911908e+00, 9.71360096650088e+00, 9.71266425436735e+00, 9.71212573485472e+00, 9.71184457825208e+00, 9.71171611527609e+00, 9.71166792434606e+00, 9.71165481524155e+00, 9.71165288036507e+00, 9.71165273555498e+00, 9.71165019261707e+00, 9.71163296383978e+00, 9.71156962904936e+00, 9.71140080188526e+00, 9.71103132844504e+00, 9.71032374520910e+00, 9.70909326017390e+00, 9.70710449112549e+00, 9.70407016006090e+00, 9.69965190179052e+00, 9.69346329619346e+00, 9.68507517585512e+00, 9.67402319265809e+00, 9.65981754794304e+00, 9.64195470201465e+00, 9.61993078261287e+00, 9.59325631290869e+00, 9.56147178379275e+00, 9.52416351027574e+00, 9.48097914599487e+00, 9.43164219120946e+00, 9.37596482513352e+00, 9.31385842770649e+00, 9.24534123084502e+00, 9.17054265367387e+00, 9.08970402606647e+00, 9.00317558322593e+00, 8.91140981170859e+00, 8.81495143230264e+00, 8.71442450255967e+00, 8.61051729344500e+00, 8.50396572039640e+00, 8.39553616996293e+00, 8.28600854602718e+00, 8.17616026273978e+00, 8.06675174795770e+00, 7.95851381837551e+00, 7.85213708076560e+00, 7.74826333642856e+00, 7.64747884129633e+00, 7.55030920990689e+00, 7.45721574062514e+00, 7.36859296527700e+00, 7.28476726926282e+00, 7.20599647176022e+00, 7.13247028955971e+00, 7.06431162865571e+00, 7.00157865609770e+00, 6.94426760450872e+00, 6.89231625730625e+00, 6.84560805732764e+00, 6.80397677712853e+00, 6.76721168613162e+00, 6.73506314747072e+00, 6.70724857471565e+00, 6.68345867456195e+00, 6.66336389525100e+00, 6.64662099172168e+00, 6.63287960768014e+00, 6.62178876287670e+00, 6.61300312231447e+00, 6.60618891455553e+00, 6.60102936045553e+00, 6.59722947309653e+00, 6.59452009558981e+00, 6.59266105646253e+00, 6.59144334259233e+00, 6.59069021654921e+00, 6.59025723758401e+00, 6.59003118172288e+00, 6.58992789451709e+00, 6.58988914785305e+00, 6.58987860781462e+00, 6.58987705212679e+00, 6.58987693565236e+00, 6.58987488972075e+00, 6.58986102823224e+00, 6.58981007192387e+00, 6.58967424103455e+00, 6.58937697805058e+00, 6.58880768232237e+00, 6.58781766741811e+00, 6.58621752709431e+00, 6.58377606674473e+00, 6.58022092339043e+00, 6.57524095765186e+00, 6.56849045461585e+00, 6.55959511627421e+00, 6.54815976612145e+00, 6.53377761740220e+00, 6.51604088253343e+00, 6.49455242594968e+00, 6.46893809095103e+00, 6.43885926902089e+00, 6.40402523389616e+00, 6.36420473838575e+00, 6.31923637416931e+00, 6.26903722598829e+00, 6.21360941140762e+00, 6.15304418243110e+00, 6.08752336999252e+00, 6.01731806944719e+00, 5.94278458693864e+00, 5.86435778575920e+00, 5.78254208259213e+00, 5.69790044111163e+00, 5.61104179103431e+00, 5.52260736099250e+00, 5.43325645031110e+00, 5.34365217501900e+00, 5.25444770528836e+00, 5.16627346489676e+00, 5.07972569075505e+00, 4.99535665737653e+00, 4.91366676515043e+00, 4.83509858163819e+00, 4.76003282117474e+00, 4.68878615797287e+00, 4.62161069762578e+00, 4.55869488454825e+00, 4.50016559886661e+00, 4.44609119353477e+00, 4.39648523725280e+00, 4.35131075633474e+00, 4.31048480397030e+00, 4.27388322358124e+00, 4.24134551008343e+00, 4.21267970560707e+00, 4.18766729233548e+00, 4.16606806327407e+00, 4.14762496149903e+00, 4.13206888006238e+00, 4.11912340919843e+00, 4.10850950625902e+00, 4.09995004877089e+00, 4.09317421427711e+00, 4.08792161441275e+00, 4.08394609710800e+00, 4.08101912180179e+00, 4.07893260955428e+00, 4.07750117388998e+00, 4.07656364936301e+00, 4.07598385282223e+00, 4.07565053615348e+00, 4.07547651735786e+00, 4.07539700731389e+00, 4.07536718044541e+00, 4.07535906682287e+00, 4.07535786926950e+00, 4.07535777937078e+00, 4.07535619707787e+00, 4.07534547681604e+00, 4.07530606808081e+00, 4.07520101948577e+00, 4.07497112637257e+00, 4.07453086569802e+00, 4.07376528668984e+00, 4.07252800954816e+00, 4.07064046537973e+00, 4.06789248742097e+00, 4.06404433491704e+00, 4.05883019519506e+00, 4.05196316520311e+00, 4.04314166045409e+00, 4.03205713731966e+00, 4.01840294568500e+00, 4.00188405630450e+00, 3.98222733544963e+00, 3.95919197450927e+00, 3.93257963075749e+00, 3.90224380434140e+00, 3.86809797178169e+00, 3.83012202249093e+00, 3.78836660420192e+00, 3.74295507492815e+00, 3.69408287891970e+00, 3.64201430441855e+00, 3.58707673134456e+00, 3.52965262489263e+00, 3.47016966330598e+00, 3.40908949268853e+00, 3.34689566904722e+00, 3.28408137211044e+00, 3.22113745581934e+00, 3.15854134051800e+00, 3.09674715973868e+00, 3.03617746101207e+00, 2.97721663758016e+00, 2.92020614814453e+00, 2.86544147492830e+00, 2.81317068363761e+00, 2.76359438641661e+00, 2.71686687150036e+00, 2.67309814926949e+00, 2.63235667023129e+00, 2.59467249147614e+00, 2.56004069947426e+00, 2.52842493403183e+00, 2.49976089680995e+00, 2.47395976483443e+00, 2.45091146256143e+00, 2.43048777377010e+00, 2.41254529596336e+00, 2.39692825474037e+00, 2.38347120384982e+00, 2.37200163875507e+00, 2.36234254821359e+00, 2.35431492048441e+00, 2.34774020940536e+00, 2.34244275196511e+00, 2.33825211448554e+00, 2.33500533054386e+00, 2.33254898168554e+00, 2.33074106307231e+00, 2.32945257149812e+00, 2.32856875337266e+00, 2.32798995558159e+00, 2.32763203238439e+00, 2.32742627605386e+00, 2.32731885675782e+00, 2.32726977691934e+00, 2.32725136552717e+00, 2.32724635719622e+00, 2.32724561797757e+00, 2.32724556224909e+00, 2.32724457823289e+00, 2.32723791138558e+00, 2.32721340345786e+00, 2.32714807515230e+00, 2.32700510997325e+00, 2.32673133072225e+00, 2.32625527690838e+00, 2.32548598324031e+00, 2.32431254957447e+00, 2.32260458004961e+00, 2.32021355214624e+00, 2.31697515348788e+00, 2.31271259399397e+00, 2.30724086273284e+00, 2.30037185268080e+00, 2.29192022399402e+00, 2.28170982019775e+00, 2.26958039618371e+00, 2.25539436756547e+00, 2.23904325395436e+00, 2.22045347022060e+00, 2.19959112496985e+00, 2.17646551756035e+00, 2.15113108453978e+00, 2.12368763067215e+00, 2.09427878272218e+00, 2.06308871700990e+00, 2.03033732376651e+00, 1.99627407140812e+00, 1.96117091202377e+00, 1.92531461821913e+00, 1.88899895712621e+00, 1.85251709005468e+00, 1.81615453989541e+00, 1.78018299994997e+00, 1.74485517609324e+00, 1.71040076825567e+00, 1.67702361550187e+00, 1.64489995811929e+00, 1.61417771448772e+00, 1.58497663210078e+00, 1.55738915086145e+00, 1.53148181094538e+00, 1.50729704432152e+00, 1.48485520513705e+00, 1.46415671627890e+00, 1.44518423449523e+00, 1.42790476199276e+00, 1.41227165651438e+00, 1.39822651323383e+00, 1.38570090958343e+00, 1.37461801797988e+00, 1.36489410126156e+00, 1.35643991164790e+00, 1.34916201646446e+00, 1.34296407312958e+00, 1.33774807242389e+00, 1.33341556337168e+00, 1.32986886572908e+00, 1.32701226773220e+00, 1.32475319809593e+00, 1.32300335299899e+00, 1.32167975165887e+00, 1.32070568876527e+00, 1.32001154905186e+00, 1.31953544903089e+00, 1.31922367354443e+00, 1.31903088020897e+00, 1.31892005269329e+00, 1.31886219350019e+00, 1.31883575778670e+00, 1.31882584093977e+00, 1.31882314332658e+00, 1.31882274516487e+00, 1.31882271498477e+00, 1.31882217991794e+00, 1.31881855476643e+00, 1.31880522839938e+00, 1.31876970576182e+00, 1.31869196820115e+00, 1.31854310244207e+00, 1.31828425796147e+00, 1.31786598860560e+00, 1.31722802948185e+00, 1.31629955369805e+00, 1.31499994444783e+00, 1.31324010522787e+00, 1.31092431373084e+00, 1.30795260266979e+00, 1.30422362353723e+00, 1.29963791798657e+00, 1.29410148797878e+00, 1.28752952284363e+00, 1.27985011251414e+00, 1.27100775537812e+00, 1.26096646031974e+00, 1.24971224872385e+00, 1.23725488522295e+00, 1.22362870560719e+00, 1.20889246422515e+00, 1.19312818690813e+00, 1.17643908283713e+00, 1.15894663288417e+00, 1.14078702601319e+00, 1.12210715371886e+00, 1.10306039164173e+00, 1.08380239635115e+00, 1.06448712531332e+00, 1.04526325289763e+00, 1.02627111002045e+00, 1.00764022536617e+00, 9.89487497482749e-01, 9.71915983894498e-01, 9.55014258798471e-01, 9.38856266490234e-01, 9.23501583561674e-01, 9.08995998180686e-01, 8.95372317707924e-01, 8.82651324485089e-01, 8.70842811772647e-01, 8.59946645663015e-01, 8.49953812842081e-01, 8.40847427212481e-01, 8.32603679909759e-01, 8.25192726746849e-01, 8.18579514477776e-01, 8.12724552513771e-01, 8.07584639997664e-01, 8.03113559643585e-01, 7.99262749701015e-01, 7.95981964032400e-01, 7.93219927832851e-01, 7.90924993210986e-01, 7.89045794944158e-01, 7.87531902501392e-01, 7.86334460195761e-01, 7.85406803407760e-01, 7.84705035537442e-01, 7.84188548000216e-01, 7.83820464434505e-01, 7.83567990511408e-01, 7.83402652397641e-01, 7.83300409966401e-01, 7.83241635099884e-01, 7.83210950592775e-01, 7.83196930878027e-01, 7.83191671645923e-01, 7.83190241011695e-01, 7.83190029853245e-01, 7.83190013746172e-01, 7.83189726841897e-01, 7.83187783025610e-01, 7.83180637387026e-01, 7.83161590012249e-01, 7.83119906689613e-01, 7.83040083470496e-01, 7.82901286900120e-01, 7.82676999506137e-01, 7.82334897410216e-01, 7.81836983326897e-01, 7.81139993241972e-01, 7.80196088192611e-01, 7.78953833347607e-01, 7.77359454747348e-01, 7.75358349666973e-01, 7.72896810108357e-01, 7.69923901402798e-01, 7.66393420818432e-01, 7.62265846323487e-01, 7.57510175375629e-01, 7.52105549814246e-01, 7.46042567207859e-01, 7.39324192130752e-01, 7.31966202515530e-01, 7.23997134962318e-01, 7.15457726088047e-01, 7.06399881287538e-01, 6.96885233960280e-01, 6.86983383910430e-01, 6.76769920638066e-01, 6.66324344195914e-01, 6.55727993150272e-01, 6.45062077206298e-01, 6.34405893467360e-01, 6.23835282866388e-01, 6.13421359868312e-01, 6.03229526578190e-01, 5.93318763748599e-01, 5.83741176950213e-01, 5.74541766704883e-01, 5.65758386424854e-01, 5.57421850885703e-01, 5.49556159789411e-01, 5.42178804810334e-01, 5.35301133496112e-01, 5.28928748806685e-01, 5.23061928384109e-01, 5.17696052495168e-01, 5.12822033762319e-01, 5.08426745196275e-01, 5.04493445637110e-01, 5.01002203522098e-01, 4.97930320971065e-01, 4.95252760572633e-01, 4.92942577030995e-01, 4.90971355064524e-01, 4.89309653715495e-01, 4.87927455631286e-01, 4.86794618026906e-01, 4.85881320071400e-01, 4.85158499510739e-01, 4.84598269611674e-01, 4.84174306151957e-01, 4.83862193345061e-01, 4.83639717397274e-01, 4.83487096933473e-01, 4.83387140818081e-01, 4.83325325902314e-01, 4.83289789844562e-01, 4.83271237220265e-01, 4.83262760464406e-01, 4.83259580555426e-01, 4.83258715544480e-01, 4.83258587870722e-01, 4.83258578063094e-01, 4.83258402465658e-01, 4.83257212768109e-01, 4.83252839327562e-01, 4.83241181444413e-01, 4.83215668978296e-01, 4.83166811568914e-01, 4.83081854429237e-01, 4.82944558027327e-01, 4.82735117002484e-01, 4.82430231196510e-01, 4.82003338497222e-01, 4.81425015047066e-01, 4.80663543060487e-01, 4.79685639912091e-01, 4.78457334353806e-01, 4.76944966956651e-01, 4.75116282661695e-01, 4.72941574415792e-01, 4.70394829211479e-01, 4.67454822485571e-01, 4.64106104750066e-01, 4.60339826294266e-01, 4.56154352177407e-01, 4.51555630371921e-01, 4.46557290089828e-01, 4.41180463784008e-01, 4.35453343448621e-01, 4.29410497912400e-01, 4.23091991222513e-01, 4.16542351739743e-01, 4.09809446539533e-01, 4.02943316083096e-01, 3.95995020390004e-01, 3.89015541023220e-01, 3.82054774226084e-01, 3.75160640687952e-01, 3.68378327678869e-01, 3.61749670464221e-01, 3.55312672482157e-01, 3.49101157972399e-01, 3.43144546600278e-01, 3.37467736992501e-01, 3.32091084775499e-01, 3.27030460436077e-01, 3.22297372861051e-01, 3.17899145537666e-01, 3.13839133914001e-01, 3.10116974164075e-01, 3.06728855434223e-01, 3.03667809442452e-01, 3.00924012956614e-01, 2.98485100098709e-01, 2.96336482539466e-01, 2.94461676405227e-01, 2.92842635084803e-01, 2.91460087093675e-01, 2.90293877745590e-01, 2.89323312647966e-01, 2.88527500055196e-01, 2.87885687982374e-01, 2.87377590822686e-01, 2.86983699150447e-01, 2.86685565559175e-01, 2.86466058893637e-01, 2.86309579179147e-01, 2.86202225987275e-01, 2.86131913920627e-01, 2.86088430321179e-01, 2.86063432135151e-01, 2.86050380993267e-01, 2.86044417856210e-01, 2.86042180883697e-01, 2.86041572373506e-01, 2.86041482558694e-01, 2.86041475606666e-01, 2.86041350450752e-01, 2.86040502501557e-01, 2.86037385357623e-01, 2.86029076250500e-01, 2.86010892247397e-01, 2.85976068758366e-01, 2.85915513711430e-01, 2.85817649496309e-01, 2.85668353282780e-01, 2.85451004494618e-01, 2.85146645929835e-01, 2.84734262124061e-01, 2.84191174939586e-01, 2.83493551970068e-01, 2.82617018219014e-01, 2.81537355799898e-01, 2.80231270420629e-01, 2.78677197599769e-01, 2.76856116469453e-01, 2.74752335266051e-01, 2.72354210784101e-01, 2.69654764652103e-01, 2.66652162551729e-01, 2.63350028441700e-01, 2.59757574144615e-01, 2.55889534700081e-01, 2.51765910834145e-01, 2.47411530782720e-01, 2.42855453579004e-01, 2.38130243950550e-01, 2.33271154582702e-01, 2.28315254388249e-01, 2.23300541571419e-01, 2.18265077936607e-01, 2.13246176513721e-01, 2.08279668713736e-01, 2.03399270494423e-01, 1.98636059991117e-01, 1.94018072266402e-01, 1.89570010674373e-01, 1.85313069126721e-01, 1.81264855479704e-01, 1.77439403416443e-01, 1.73847258561395e-01, 1.70495624038885e-01, 1.67388551118620e-01, 1.64527161782217e-01, 1.61909891781711e-01, 1.59532744826429e-01, 1.57389550727210e-01, 1.55472222465612e-01, 1.53771009095081e-01, 1.52274743004192e-01, 1.50971081303517e-01, 1.49846741892311e-01, 1.48887735110776e-01, 1.48079591808504e-01, 1.47407588211642e-01, 1.46856967224132e-01, 1.46413154848617e-01, 1.46061969371132e-01, 1.45789819937072e-01, 1.45583890271737e-01, 1.45432302671210e-01, 1.45324257093208e-01, 1.45250140268189e-01, 1.45201600248477e-01, 1.45171582700534e-01, 1.45154326468791e-01, 1.45145317415292e-01, 1.45141201162460e-01, 1.45139657023336e-01, 1.45139236980931e-01, 1.45139174983584e-01, 1.45139170145825e-01, 1.45139082550237e-01, 1.45138489077780e-01, 1.45136307419037e-01, 1.45130491985313e-01, 1.45117765393195e-01, 1.45093393758792e-01, 1.45051015387871e-01, 1.44982531698552e-01, 1.44878068311051e-01, 1.44726013090100e-01, 1.44513136405173e-01, 1.44224796850466e-01, 1.43845233031044e-01, 1.43357938711510e-01, 1.42746114655999e-01, 1.41993185985172e-01, 1.41083369074714e-01, 1.40002267275896e-01, 1.38737470504552e-01, 1.37279130525719e-01, 1.35620482061491e-01, 1.33758280091023e-01, 1.31693126169303e-01, 1.29429661321694e-01, 1.26976609878071e-01, 1.24346667041692e-01, 1.21556232371354e-01, 1.18625000867084e-01, 1.15575432125723e-01, 1.12432125261044e-01, 1.09221132325239e-01, 1.05969245422301e-01, 1.02703292454258e-01, 9.94494736580762e-02, 9.62327661898215e-02, 9.30764175874828e-02, 9.00015416867319e-02, 8.70268231710867e-02, 8.41683300416989e-02, 8.14394273909684e-02, 7.88507812928003e-02, 7.64104385362643e-02, 7.41239663201796e-02, 7.19946357532351e-02, 7.00236338389502e-02, 6.82102902892652e-02, 6.65523077189789e-02, 6.50459862575516e-02, 6.36864361442235e-02, 6.24677742628279e-02, 6.13833026923487e-02, 6.04256691125116e-02, 5.95870102646288e-02, 5.88590806144396e-02, 5.82333689061869e-02, 5.77012054630828e-02, 5.72538629222521e-02, 5.68826526424786e-02, 5.65790183548174e-02, 5.63346278081980e-02, 5.61414622706877e-02, 5.59919028633189e-02, 5.58788119053588e-02, 5.57956068109959e-02, 5.57363236595947e-02, 5.56956674057076e-02, 5.56690458229028e-02, 5.56525846779365e-02, 5.56431222793634e-02, 5.56381823809602e-02, 5.56359253750998e-02, 5.56350787063522e-02, 5.56348483930549e-02, 5.56348143993333e-02, 5.56348117241825e-02, 5.56347629968528e-02, 5.56344328626750e-02, 5.56332192645002e-02, 5.56299843339332e-02, 5.56229051562838e-02, 5.56093491738281e-02, 5.55857798679824e-02, 5.55476981181471e-02, 5.54896243191010e-02, 5.54051258569359e-02, 5.52868937809876e-02, 5.51268713434396e-02, 5.49164353995408e-02, 5.46466294067685e-02, 5.43084439280364e-02, 5.38931372275684e-02, 5.33925849443406e-02, 5.27996442399470e-02, 5.21085146374034e-02, 5.13150754361078e-02, 5.04171785403429e-02, 4.94148761348896e-02, 4.83105650928763e-02, 4.71090343127898e-02, 4.58174071183777e-02, 4.44449779396408e-02, 4.30029500577191e-02, 4.15040884552909e-02, 3.99623079850215e-02, 3.83922214737495e-02, 3.68086745647531e-02, 3.52262938829639e-02, 3.36590726127710e-02, 3.21200131961135e-02, 3.06208411767557e-02, 2.91717979262103e-02, 2.77815137775264e-02, 2.64569575695930e-02, 2.52034542113506e-02, 2.40247588587366e-02, 2.29231746981071e-02, 2.18997010152719e-02, 2.09541989509319e-02, 2.00855638101700e-02, 1.92918947187408e-02, 1.85706545548570e-02, 1.79188152284766e-02, 1.73329853692437e-02, 1.68095192092395e-02, 1.63446068398443e-02, 1.59343470685371e-02, 1.55748048198987e-02, 1.52620554568159e-02, 1.49922185919777e-02, 1.47614839604060e-02, 1.45661317707706e-02, 1.44025496781239e-02, 1.42672481522865e-02, 1.41568755810155e-02, 1.40682339706970e-02, 1.39982956177747e-02, 1.39442206485896e-02, 1.39033748944115e-02, 1.38733472102217e-02, 1.38519650853581e-02, 1.38373072527429e-02, 1.38277119892152e-02, 1.38217799146929e-02, 1.38183703277761e-02, 1.38165904385801e-02, 1.38157772432947e-02, 1.38154721938832e-02, 1.38153892138249e-02, 1.38153769661713e-02, 1.38153759938007e-02, 1.38153581739218e-02, 1.38152374420173e-02, 1.38147936253432e-02, 1.38136106236360e-02, 1.38110219098948e-02, 1.38060651964764e-02, 1.37974485060965e-02, 1.37835298992997e-02, 1.37623130989368e-02, 1.37314611614201e-02, 1.36883301037329e-02, 1.36300239256144e-02, 1.35534717006109e-02, 1.34555263031331e-02, 1.33330828884013e-02, 1.31832135047603e-02, 1.30033123172737e-02, 1.27912440493887e-02, 1.25454866566486e-02, 1.22652582072220e-02, 1.19506177209630e-02, 1.16025305069059e-02, 1.12228904199441e-02, 1.08144943628261e-02, 1.03809680591311e-02, 9.92664623719719e-03, 9.45641440804224e-03, 8.97552286194525e-03, 8.48938586019443e-03, 8.00337992331634e-03, 7.52265451530576e-03, 7.05196648706874e-03, 6.59554679267416e-03, 6.15700472981473e-03, 5.73927168303356e-03, 5.34458328372267e-03, 4.97449615277645e-03, 4.62993311993642e-03, 4.31124930084081e-03, 4.01831095506111e-03, 3.75057971162818e-03, 3.50719623441620e-03, 3.28705915486438e-03, 3.08889662262857e-03, 2.91132890426127e-03, 2.75292119065540e-03, 2.61222635753455e-03, 2.48781797512265e-03, 2.37831435898528e-03, 2.28239481272396e-03, 2.19880939172867e-03, 2.12638354178551e-03, 2.06401889476382e-03, 2.01069137926879e-03, 1.96544765213957e-03, 1.92740068743547e-03, 1.89572518493009e-03, 1.86965329419309e-03, 1.84847099870922e-03, 1.83151536946013e-03, 1.81817277412932e-03, 1.80787801527060e-03, 1.80011426849802e-03, 1.79441360429422e-03, 1.79035781054893e-03, 1.78757919209005e-03, 1.78576101169625e-03, 1.78463725593758e-03, 1.78399145595165e-03, 1.78365436208594e-03, 1.78350035752572e-03, 1.78344258775809e-03, 1.78342687323152e-03, 1.78342455381397e-03, 1.78342436796441e-03, 1.78342094118022e-03, 1.78339772433134e-03, 1.78331237884066e-03, 1.78308489547163e-03, 1.78258713656021e-03, 1.78163418624422e-03, 1.77997799104190e-03, 1.77730380468688e-03, 1.77322997903673e-03, 1.76731164845023e-03, 1.75904881597688e-03, 1.74789924029994e-03, 1.73329631962105e-03, 1.71467185924712e-03, 1.69148319210422e-03, 1.66324361808279e-03, 1.62955458281289e-03, 1.59013749722448e-03, 1.54486269055099e-03, 1.49377277774221e-03, 1.43709778253509e-03, 1.37525973944719e-03, 1.30886522548224e-03, 1.23868534343785e-03, 1.16562405606402e-03, 1.09067733402668e-03, 1.01488704433450e-03, 9.39294370787991e-04, 8.64897236731110e-04, 7.92614421335713e-04, 7.23256382073600e-04, 6.57500766394037e-04, 5.95871055701025e-04, 5.38720367231349e-04, 4.86227028559938e-04, 4.38409412574003e-04, 3.95161066485628e-04, 3.56296210796359e-04, 3.21588992918009e-04, 2.90793700784775e-04, 2.63645024602965e-04, 2.39849243122592e-04, 2.19081241937147e-04, 2.00995876170396e-04, 1.85250530982728e-04, 1.71528444397495e-04, 1.59553918187783e-04, 1.49096830966755e-04, 1.39968908925425e-04, 1.32015898650195e-04, 1.25109108601862e-04, 1.19138249132789e-04, 1.14006134694478e-04, 1.09625071409880e-04, 1.05914554495712e-04, 1.02799922835572e-04, 1.00211658647709e-04, 9.80850605003525e-05, 9.63600841359380e-05, 9.49812305108821e-05, 9.38974353488220e-05, 9.30619632910634e-05, 9.24323254949400e-05, 9.19702342585538e-05, 9.16415915881430e-05, 9.14164939018747e-05, 9.12692257344361e-05, 9.11782135588624e-05, 9.11259138618448e-05, 9.10986154565308e-05, 9.10861441241195e-05, 9.10814659533798e-05, 9.10801933926564e-05, 9.10800055677896e-05, 9.10799903977022e-05, 9.10797089730409e-05, 9.10778022671366e-05, 9.10707932177957e-05, 9.10521112089668e-05, 9.10112338488034e-05, 9.09329785727664e-05, 9.07969853471968e-05, 9.05774354151845e-05, 9.02430560500989e-05, 8.97574644227461e-05, 8.90799065587089e-05, 8.81664456642284e-05, 8.69716442508661e-05, 8.54507623131440e-05, 8.35624485020503e-05, 8.12718294582737e-05, 7.85538060057434e-05, 7.53962676895981e-05, 7.18028802495496e-05, 6.77951283307721e-05, 6.34134079646742e-05, 5.87170836539030e-05, 5.37834092639855e-05, 4.87049638380199e-05, 4.35848713777887e-05, 3.85288704336517e-05, 3.36337397186526e-05, 2.89733856675412e-05, 2.45876472391160e-05, 2.04824883311971e-05, 1.66466775510798e-05, 1.30757381242574e-05, 9.79229435671414e-06, 6.92346845454495e-06, 5.06810788181156e-06, 4.90594315039061e-06, 5.14718733768404e-06, 4.85123207027982e-06, 3.95291034038792e-06, 2.68687809779067e-06, 1.38530911186763e-06, 3.93401577980512e-07, 3.51156742794956e-07, 5.69428351582246e-07, 7.20514258034655e-07, 6.96506845088152e-07, 9.08734975268999e-07, 1.09048389163036e-06, 9.83774953056127e-07, 6.40198805967884e-07, 5.54821234288738e-07, 6.52923283439577e-07, 7.03634285059707e-07, 6.58961653243750e-07, 5.43554385501466e-07, 4.25304087059430e-07, 3.47977348269524e-07, 3.21075170006679e-07, 3.25298792347439e-07, 3.33003285390987e-07, 3.26995213129067e-07, 3.03119477962355e-07, -1.42108547152020e-14, 0.00000000000000e+00, -7.10542735760100e-15, 0.00000000000000e+00, -7.10542735760100e-15, 0.00000000000000e+00, -7.10542735760100e-15, -7.10542735760100e-15, 0.00000000000000e+00, -7.10542735760100e-15, 0.00000000000000e+00, 0.00000000000000e+00],
[ 4.60000000000000e+01, 4.59999999588177e+01, 4.59999986853803e+01, 4.59999900576403e+01, 4.59999583409477e+01, 4.59998737942942e+01, 4.59996887579127e+01, 4.59993343600738e+01, 4.59987179621458e+01, 4.59977214384693e+01, 4.59962003621063e+01, 4.59939841400011e+01, 4.59908771126227e+01, 4.59866606048951e+01, 4.59810958883247e+01, 4.59739279897393e+01, 4.59648902607950e+01, 4.59537096048747e+01, 4.59401122442763e+01, 4.59238299003022e+01, 4.59046062512451e+01, 4.58822035272729e+01, 4.58564090957685e+01, 4.58270418849232e+01, 4.57939584869921e+01, 4.57570587760398e+01, 4.57162908695318e+01, 4.56716552608380e+01, 4.56232079531940e+01, 4.55710624375061e+01, 4.55153903786814e+01, 4.54564209089317e+01, 4.53944384713029e+01, 4.53297792104995e+01, 4.52628259673829e+01, 4.51940019938181e+01, 4.51237635608611e+01, 4.50525916809821e+01, 4.49809832003921e+01, 4.49094415382616e+01, 4.48384673549905e+01, 4.47685494225484e+01, 4.47001559483777e+01, 4.46337265733905e+01, 4.45696652275196e+01, 4.45083339863213e+01, 4.44500480320689e+01, 4.43950717847360e+01, 4.43436162336608e+01, 4.42958374701766e+01, 4.42518363952245e+01, 4.42116595535982e+01, 4.41753010274795e+01, 4.41427053056415e+01, 4.41137710305383e+01, 4.40883555129955e+01, 4.40662798930764e+01, 4.40473348158606e+01, 4.40312864824526e+01, 4.40178829298348e+01, 4.40068603886119e+01, 4.39979495657380e+01, 4.39908817004724e+01, 4.39853942464934e+01, 4.39812360416288e+01, 4.39781718391859e+01, 4.39759860913309e+01, 4.39744858951179e+01, 4.39735030351216e+01, 4.39728950825260e+01, 4.39725455381351e+01, 4.39723630351729e+01, 4.39722796459410e+01, 4.39722483634181e+01, 4.39722398537809e+01, 4.39722385977752e+01, 4.39722384330137e+01, 4.39722345942689e+01, 4.39722085861887e+01, 4.39721129774470e+01, 4.39718581174332e+01, 4.39713003513074e+01, 4.39702321194667e+01, 4.39683743208702e+01, 4.39653712636244e+01, 4.39607884622158e+01, 4.39541134695026e+01, 4.39447598537434e+01, 4.39320743473619e+01, 4.39153471058603e+01, 4.38938249235284e+01, 4.38667271586796e+01, 4.38332640263197e+01, 4.37926568214849e+01, 4.37441595429718e+01, 4.36870812962659e+01, 4.36208087689409e+01, 4.35448279968568e+01, 4.34587445837127e+01, 4.33623015119235e+01, 4.32553937035948e+01, 4.31380785700798e+01, 4.30105819359384e+01, 4.28732989376712e+01, 4.27267897667831e+01, 4.25717704254349e+01, 4.24090989567517e+01, 4.22397578634252e+01, 4.20648336053326e+01, 4.18854941497810e+01, 4.17029655339431e+01, 4.15185083024898e+01, 4.13333945316975e+01, 4.11488859774987e+01, 4.09662137199271e+01, 4.07865595423691e+01, 4.06110391911950e+01, 4.04406876081281e+01, 4.02764462040101e+01, 4.01191522344702e+01, 3.99695303322320e+01, 3.98281862383505e+01, 3.96956027519046e+01, 3.95721378857533e+01, 3.94580251790891e+01, 3.93533760807038e+01, 3.92581842841756e+01, 3.91723318697347e+01, 3.90955970874395e+01, 3.90276636010726e+01, 3.89681309996601e+01, 3.89165263715988e+01, 3.88723167235067e+01, 3.88349220114589e+01, 3.88037285365487e+01, 3.87781024408711e+01, 3.87574030258437e+01, 3.87409956043851e+01, 3.87282635940901e+01, 3.87186195622259e+01, 3.87115149467734e+01, 3.87064482019049e+01, 3.87029711515266e+01, 3.87006933803280e+01, 3.86992845468395e+01, 3.86984745652589e+01, 3.86980516696282e+01, 3.86978584423247e+01, 3.86977859556396e+01, 3.86977662374610e+01, 3.86977633270977e+01, 3.86977630701680e+01, 3.86977580357570e+01, 3.86977239268715e+01, 3.86975985388444e+01, 3.86972642992006e+01, 3.86965328178285e+01, 3.86951319201673e+01, 3.86926956743833e+01, 3.86887578575024e+01, 3.86827492292232e+01, 3.86739988947348e+01, 3.86617399378214e+01, 3.86451193913617e+01, 3.86232124830686e+01, 3.85950409502603e+01, 3.85595950607642e+01, 3.85158588122610e+01, 3.84628376165560e+01, 3.83995876179385e+01, 3.83252456572251e+01, 3.82390587873328e+01, 3.81404121837902e+01, 3.80288542841931e+01, 3.79041180412751e+01, 3.77661372890420e+01, 3.76150574014695e+01, 3.74512396668604e+01, 3.72752591022741e+01, 3.70878957786771e+01, 3.68901200947049e+01, 3.66830727875087e+01, 3.64680407543653e+01, 3.62464299297021e+01, 3.60197364873478e+01, 3.57895175190224e+01, 3.55573621175418e+01, 3.53248635333067e+01, 3.50935928431058e+01, 3.48650744153298e+01, 3.46407633838266e+01, 3.44220253313353e+01, 3.42101183956442e+01, 3.40061780135840e+01, 3.38112044905520e+01, 3.36260535241915e+01, 3.34514297301904e+01, 3.32878831312248e+01, 3.31358084911681e+01, 3.29954473153746e+01, 3.28668922978889e+01, 3.27500939765670e+01, 3.26448693527447e+01, 3.25509122371919e+01, 3.24678050926352e+01, 3.23950321500876e+01, 3.23319935781414e+01, 3.22780204794774e+01, 3.22323904769284e+01, 3.21943436336758e+01, 3.21630984308030e+01, 3.21378675034341e+01, 3.21178728174661e+01, 3.21023599559046e+01, 3.20906111802846e+01, 3.20819569412379e+01, 3.20757855348199e+01, 3.20715506385061e+01, 3.20687765124663e+01, 3.20670607162940e+01, 3.20660742662110e+01, 3.20655592394416e+01, 3.20653239169620e+01, 3.20652356389720e+01, 3.20652116251802e+01, 3.20652080807935e+01, 3.20652077964898e+01, 3.20652025496450e+01, 3.20651670014908e+01, 3.20650363225853e+01, 3.20646879796001e+01, 3.20639256344132e+01, 3.20624656326992e+01, 3.20599266165103e+01, 3.20558227276851e+01, 3.20495607906987e+01, 3.20404417736129e+01, 3.20276667223285e+01, 3.20103472445823e+01, 3.19875204848435e+01, 3.19581683795961e+01, 3.19212408165704e+01, 3.18756821460143e+01, 3.18204603145659e+01, 3.17545977229004e+01, 3.16772027594322e+01, 3.15875008474488e+01, 3.14848637754973e+01, 3.13688360722400e+01, 3.12391572457714e+01, 3.10957788374173e+01, 3.09388754397605e+01, 3.07688490905080e+01, 3.05863267641789e+01, 3.03921510230228e+01, 3.01873642327466e+01, 2.99731870699922e+01, 2.97509923190578e+01, 2.95222751502571e+01, 2.92886211740989e+01, 2.90516735676411e+01, 2.88131004783431e+01, 2.85745637451895e+01, 2.83376897642603e+01, 2.81040430970860e+01, 2.78751032031385e+01, 2.76522444930831e+01, 2.74367197571639e+01, 2.72296469236338e+01, 2.70319990385010e+01, 2.68445973196586e+01, 2.66681071155023e+01, 2.65030365829287e+01, 2.63497378881894e+01, 2.62084107254780e+01, 2.60791079429993e+01, 2.59617430654360e+01, 2.58560995051260e+01, 2.57618412606319e+01, 2.56785249084113e+01, 2.56056126982533e+01, 2.55424865635113e+01, 2.54884628512209e+01, 2.54428075643495e+01, 2.54047518893646e+01, 2.53735077588678e+01, 2.53482831740437e+01, 2.53282969885298e+01, 2.53127928377510e+01, 2.53010518893184e+01, 2.52924040938457e+01, 2.52862376336857e+01, 2.52820063007507e+01, 2.52792345835977e+01, 2.52775203068962e+01, 2.52765347406246e+01, 2.52760201782628e+01, 2.52757850686549e+01, 2.52756968706305e+01, 2.52756728786017e+01, 2.52756693374275e+01, 2.52756690642981e+01, 2.52756641597700e+01, 2.52756309308653e+01, 2.52755087778122e+01, 2.52751831619745e+01, 2.52744705563971e+01, 2.52731058188935e+01, 2.52707324895833e+01, 2.52668964528432e+01, 2.52610433319136e+01, 2.52525199015196e+01, 2.52405797078681e+01, 2.52243929742224e+01, 2.52030607425034e+01, 2.51756330569527e+01, 2.51411308368110e+01, 2.50985709158255e+01, 2.50469935547348e+01, 2.49854915689724e+01, 2.49132400701598e+01, 2.48295257100734e+01, 2.47337742527701e+01, 2.46255752954344e+01, 2.45047030184688e+01, 2.43711319726653e+01, 2.42250471027602e+01, 2.40668474538443e+01, 2.38971432970692e+01, 2.37167467274345e+01, 2.35266561098840e+01, 2.33280350586373e+01, 2.31221869045730e+01, 2.29105258116121e+01, 2.26945458229089e+01, 2.24757891363990e+01, 2.22558148253893e+01, 2.20361690486156e+01, 2.18183575658313e+01, 2.16038211278314e+01, 2.13939140799569e+01, 2.11898863303463e+01, 2.09928686976155e+01, 2.08038615623916e+01, 2.06237266902349e+01, 2.04531820558136e+01, 2.02927994694281e+01, 2.01430047824630e+01, 2.00040804282172e+01, 1.98761700413789e+01, 1.97592848957729e+01, 1.96533119068773e+01, 1.95580229617720e+01, 1.94730853615199e+01, 1.93980731852524e+01, 1.93324794070194e+01, 1.92757286119711e+01, 1.92271901650865e+01, 1.91861916823537e+01, 1.91520326415147e+01, 1.91239979490738e+01, 1.91013712552669e+01, 1.90834477828615e+01, 1.90695464131579e+01, 1.90590207574235e+01, 1.90512689377683e+01, 1.90457418108213e+01, 1.90419493920005e+01, 1.90394652778637e+01, 1.90379289177952e+01, 1.90370456516803e+01, 1.90365845037925e+01, 1.90363738007432e+01, 1.90362947586304e+01, 1.90362732572475e+01, 1.90362700836880e+01, 1.90362698433956e+01, 1.90362655866406e+01, 1.90362367464989e+01, 1.90361307270845e+01, 1.90358481182306e+01, 1.90352296355642e+01, 1.90340451668791e+01, 1.90319853646943e+01, 1.90286561735874e+01, 1.90235766213073e+01, 1.90161801331695e+01, 1.90058195475784e+01, 1.89917759143966e+01, 1.89732710455579e+01, 1.89494836584161e+01, 1.89195688084145e+01, 1.88826801528910e+01, 1.88379944290898e+01, 1.87847373763111e+01, 1.87222101962677e+01, 1.86498155399199e+01, 1.85670819458484e+01, 1.84736856452896e+01, 1.83694686996386e+01, 1.82544525503088e+01, 1.81288462358954e+01, 1.79930487599986e+01, 1.78476453625555e+01, 1.76933977422354e+01, 1.75312285789671e+01, 1.73622009942096e+01, 1.71874938420037e+01, 1.70083739269070e+01, 1.68261663788006e+01, 1.66422244669288e+01, 1.64579001006937e+01, 1.62745161454635e+01, 1.60933414899495e+01, 1.59155695578876e+01, 1.57423006868916e+01, 1.55745285294218e+01, 1.54131303904652e+01, 1.52588612234062e+01, 1.51123508710090e+01, 1.49741040649479e+01, 1.48445026796164e+01, 1.47238097632000e+01, 1.46121749275013e+01, 1.45096407537113e+01, 1.44161499517883e+01, 1.43315530866294e+01, 1.42556167482203e+01, 1.41880320917927e+01, 1.41284237064313e+01, 1.40763587870333e+01, 1.40313565865922e+01, 1.39928981156908e+01, 1.39604360363815e+01, 1.39334046710507e+01, 1.39112300162482e+01, 1.38933396197437e+01, 1.38791721492239e+01, 1.38681864560132e+01, 1.38598699197462e+01, 1.38537458524565e+01, 1.38493797448763e+01, 1.38463841549086e+01, 1.38444220682900e+01, 1.38432086034467e+01, 1.38425109844884e+01, 1.38421467653419e+01, 1.38419803507927e+01, 1.38419179229803e+01, 1.38419009411045e+01, 1.38418984346153e+01, 1.38418982464524e+01, 1.38418949345715e+01, 1.38418724960961e+01, 1.38417900099559e+01, 1.38415701327225e+01, 1.38410889394731e+01, 1.38401674071701e+01, 1.38385648870957e+01, 1.38359748753797e+01, 1.38320233467810e+01, 1.38262698647464e+01, 1.38182116178561e+01, 1.38072904563913e+01, 1.37929029107579e+01, 1.37744130649473e+01, 1.37511680342011e+01, 1.37225156602863e+01, 1.36878238970373e+01, 1.36465012228698e+01, 1.35980172981368e+01, 1.35419229973098e+01, 1.34778689027460e+01, 1.34056213599119e+01, 1.33250752708243e+01, 1.32362629444590e+01, 1.31393585237548e+01, 1.30346777545502e+01, 1.29226731312559e+01, 1.28039247215048e+01, 1.26791272103433e+01, 1.25490738894708e+01, 1.24146384312108e+01, 1.22767553228696e+01, 1.21363997987733e+01, 1.19945680088327e+01, 1.18522580256063e+01, 1.17104521409806e+01, 1.15701007610620e+01, 1.14321080896073e+01, 1.12973197035741e+01, 1.11665120676775e+01, 1.10403840002892e+01, 1.09195500797626e+01, 1.08045359584615e+01, 1.06957755253336e+01, 1.05936098257873e+01, 1.04982876132311e+01, 1.04099673756798e+01, 1.03287206590277e+01, 1.02545364996341e+01, 1.01873267833998e+01, 1.01269323642487e+01, 1.00731297975519e+01, 1.00256385684741e+01, 9.98412871680993e+00, 9.94822877513917e+00, 9.91753394413454e+00, 9.89161442720435e+00, 9.87002383725636e+00, 9.85230757309441e+00, 9.83801104430992e+00, 9.82668760432357e+00, 9.81790603434411e+00, 9.81125740912314e+00, 9.80636117078409e+00, 9.80287024137800e+00, 9.80047501911542e+00, 9.79890612754247e+00, 9.79793582050316e+00, 9.79737798698623e+00, 9.79708674665461e+00, 9.79695367632170e+00, 9.79690375701014e+00, 9.79689017774367e+00, 9.79688817347227e+00, 9.79688802346851e+00, 9.79688538932738e+00, 9.79686754263469e+00, 9.79680193632152e+00, 9.79662705410870e+00, 9.79624432932674e+00, 9.79551136812591e+00, 9.79423675040878e+00, 9.79217665070908e+00, 9.78903348502675e+00, 9.78445674714947e+00, 9.77804614766635e+00, 9.76935710905416e+00, 9.75790859961113e+00, 9.74319320720445e+00, 9.72468926167499e+00, 9.70187471488601e+00, 9.67424238425686e+00, 9.64131606546690e+00, 9.60266693061941e+00, 9.55792955817552e+00, 9.50681689911119e+00, 9.44913347781404e+00, 9.38478616218624e+00, 9.31379191856894e+00, 9.23628209349884e+00, 9.15250293221217e+00, 9.06281224523186e+00, 8.96767235656562e+00, 8.86763969225156e+00, 8.76335157395768e+00, 8.65551094439304e+00, 8.54486984643744e+00, 8.43221249061651e+00, 8.31833867346809e+00, 8.20404816629780e+00, 8.09012650806487e+00, 7.97733244342046e+00, 7.86638708064166e+00, 7.75796472652696e+00, 7.65268529271452e+00, 7.55110815238438e+00, 7.45372734110695e+00, 7.36096802210398e+00, 7.27318415965461e+00, 7.19065735670191e+00, 7.11359681233036e+00, 7.04214034472469e+00, 6.97635641058125e+00, 6.91624703770131e+00, 6.86175157714818e+00, 6.81275117643008e+00, 6.76907387539095e+00, 6.73050023029367e+00, 6.69676937676519e+00, 6.66758544662220e+00, 6.64262425530568e+00, 6.62154017463646e+00, 6.60397309960008e+00, 6.58955540843218e+00, 6.57791880365026e+00, 6.56870090961118e+00, 6.56155149168300e+00, 6.55613815523703e+00, 6.55215138120788e+00, 6.54930876032043e+00, 6.54735830104138e+00, 6.54608070699448e+00, 6.54529054739844e+00, 6.54483627780962e+00, 6.54459910632225e+00, 6.54449074029461e+00, 6.54445008838909e+00, 6.54443903007877e+00, 6.54443739789510e+00, 6.54443727569344e+00, 6.54443512915984e+00, 6.54442058607830e+00, 6.54436712416319e+00, 6.54422461425732e+00, 6.54391273446524e+00, 6.54331544601729e+00, 6.54227675222101e+00, 6.54059793572254e+00, 6.53803643896240e+00, 6.53430651739463e+00, 6.52908175367466e+00, 6.52199947250023e+00, 6.51266703930828e+00, 6.50066996144587e+00, 6.48558163861227e+00, 6.46697453245500e+00, 6.44443244673620e+00, 6.41756353433549e+00, 6.38601358139441e+00, 6.34947906845490e+00, 6.30771947950103e+00, 6.26056832722725e+00, 6.20794238954383e+00, 6.14984870875662e+00, 6.08638898889451e+00, 6.01776113384024e+00, 5.94425779306677e+00, 5.86626191573022e+00, 5.78423945015401e+00, 5.69872945699566e+00, 5.61033202352368e+00, 5.51969446662245e+00, 5.42749638695163e+00, 5.33443418045748e+00, 5.24120562203426e+00, 5.14849510786663e+00, 5.05696007938837e+00, 4.96721905798068e+00, 4.87984160387238e+00, 4.79534038582461e+00, 4.71416542152657e+00, 4.63670043287137e+00, 4.56326116396605e+00, 4.49409543838826e+00, 4.42938468804981e+00, 4.36924666820262e+00, 4.31373907833103e+00, 4.26286383201459e+00, 4.21657175467545e+00, 4.17476753079074e+00, 4.13731476653955e+00, 4.10404107573084e+00, 4.07474313302104e+00, 4.04919166673252e+00, 4.02713638286868e+00, 4.00831082190996e+00, 3.99243715114485e+00, 3.97923088874839e+00, 3.96840554318622e+00, 3.95967713479216e+00, 3.95276854778970e+00, 3.94741364293841e+00, 3.94336104564275e+00, 3.94037751376622e+00, 3.93825078511582e+00, 3.93679180760004e+00, 3.93583626573665e+00, 3.93524533511469e+00, 3.93490562054445e+00, 3.93472826236391e+00, 3.93464722673205e+00, 3.93461682759600e+00, 3.93460855830592e+00, 3.93460733777647e+00, 3.93460724615295e+00, 3.93460563350244e+00, 3.93459470756477e+00, 3.93455454276389e+00, 3.93444747893309e+00, 3.93421317632562e+00, 3.93376447400735e+00, 3.93298422450199e+00, 3.93172326168886e+00, 3.92979964019155e+00, 3.92699926151822e+00, 3.92307797330791e+00, 3.91776519140585e+00, 3.91076904864552e+00, 3.90178301840147e+00, 3.89049389556614e+00, 3.87659094443328e+00, 3.85977594551422e+00, 3.83977379667456e+00, 3.81634325463993e+00, 3.78928734823073e+00, 3.75846296213672e+00, 3.72378908634462e+00, 3.68525325643518e+00, 3.64291577612261e+00, 3.59691141443149e+00, 3.54744840078236e+00, 3.49480469321978e+00, 3.43932165616484e+00, 3.38139544051988e+00, 3.32146649639417e+00, 3.26000775425331e+00, 3.19751207414494e+00, 3.13447957956060e+00, 3.07140546244145e+00, 3.00876877398965e+00, 2.94702261179100e+00, 2.88658598959434e+00, 2.82783754523047e+00, 2.77111111707603e+00, 2.71669311037351e+00, 2.66482148859309e+00, 2.61568616532296e+00, 2.56943053908651e+00, 2.52615390456640e+00, 2.48591448475040e+00, 2.44873285439805e+00, 2.41459556071330e+00, 2.38345878738463e+00, 2.35525194917131e+00, 2.32988114285130e+00, 2.30723241439100e+00, 2.28717483026300e+00, 2.26956336220606e+00, 2.25424160920409e+00, 2.24104438825597e+00, 2.22980022711588e+00, 2.22033378832722e+00, 2.21246824547299e+00, 2.20602762070591e+00, 2.20083907853530e+00, 2.19673515587968e+00, 2.19355589394657e+00, 2.19115082495507e+00, 2.18938075733668e+00, 2.18811929786439e+00, 2.18725404886807e+00, 2.18668742357027e+00, 2.18633703243587e+00, 2.18613560762759e+00, 2.18603045016347e+00, 2.18598240386304e+00, 2.18596438020096e+00, 2.18595947734333e+00, 2.18595875369238e+00, 2.18595869913753e+00, 2.18595773584440e+00, 2.18595120939949e+00, 2.18592721761437e+00, 2.18586326522028e+00, 2.18572331151902e+00, 2.18545530086952e+00, 2.18498928264792e+00, 2.18423621975456e+00, 2.18308757554351e+00, 2.18141575699835e+00, 2.17907547550951e+00, 2.17590606406165e+00, 2.17173475959125e+00, 2.16638092090043e+00, 2.15966110596973e+00, 2.15139487925567e+00, 2.14141116253703e+00, 2.12955488653319e+00, 2.11569365054962e+00, 2.09972406022054e+00, 2.08157739535863e+00, 2.06122426629564e+00, 2.03867795118449e+00, 2.01399616891734e+00, 1.98728112956806e+00, 1.95867781012318e+00, 1.92837051836789e+00, 1.89657792099528e+00, 1.86354681186391e+00, 1.82954497272807e+00, 1.79485352439254e+00, 1.75975917760468e+00, 1.72454677079200e+00, 1.68949243056728e+00, 1.65485761830581e+00, 1.62088424126369e+00, 1.58779091911455e+00, 1.55577041487722e+00, 1.52498816949832e+00, 1.49558182594093e+00, 1.46766159316272e+00, 1.44131128235704e+00, 1.41658984518195e+00, 1.39353325332484e+00, 1.37215657714385e+00, 1.35245614486032e+00, 1.33441168981908e+00, 1.31798841924036e+00, 1.30313896186586e+00, 1.28980517276517e+00, 1.27791979066066e+00, 1.26740795619067e+00, 1.25818860860096e+00, 1.25017578365468e+00, 1.24327983742620e+00, 1.23740861950188e+00, 1.23246861541083e+00, 1.22836607234780e+00, 1.22500811497720e+00, 1.22230384991737e+00, 1.22016544904495e+00, 1.21850919371040e+00, 1.21725645499620e+00, 1.21633457991356e+00, 1.21567765044880e+00, 1.21522708200491e+00, 1.21493203018926e+00, 1.21474957998937e+00, 1.21464469882066e+00, 1.21458994417714e+00, 1.21456492696270e+00, 1.21455554224450e+00, 1.21455298938329e+00, 1.21455261258669e+00, 1.21455258402602e+00, 1.21455207767054e+00, 1.21454864704256e+00, 1.21453603576460e+00, 1.21450241929170e+00, 1.21442885332417e+00, 1.21428797665892e+00, 1.21404302525993e+00, 1.21364721076233e+00, 1.21304351221136e+00, 1.21216492383743e+00, 1.21093519312999e+00, 1.20927007138230e+00, 1.20707908235599e+00, 1.20426779324094e+00, 1.20074054573825e+00, 1.19640357475274e+00, 1.19116840965560e+00, 1.18495542110740e+00, 1.17769734852344e+00, 1.16934262333154e+00, 1.15985829501741e+00, 1.14923237361838e+00, 1.13747542546145e+00, 1.12462129827235e+00, 1.11072690486680e+00, 1.09587105694066e+00, 1.08015240589398e+00, 1.06368660926606e+00, 1.04660289259430e+00, 1.02904021197078e+00, 1.01114323896506e+00, 9.93058386137470e-01, 9.74930069822371e-01, 9.56897371039297e-01, 9.39091210455608e-01, 9.21632104909207e-01, 9.04628526351075e-01, 8.88175843408042e-01, 8.72355793895686e-01, 8.57236414821259e-01, 8.42872344638728e-01, 8.29305409630457e-01, 8.16565410519402e-01, 8.04671034702935e-01, 7.93630831821467e-01, 7.83444203966326e-01, 7.74102375291520e-01, 7.65589318137692e-01, 7.57882623391751e-01, 7.50954311394239e-01, 7.44771586204045e-01, 7.39297540523204e-01, 7.34491821249406e-01, 7.30311266674931e-01, 7.26710526013541e-01, 7.23642670432817e-01, 7.21059802315580e-01, 7.18913666289026e-01, 7.17156261875836e-01, 7.15740453685726e-01, 7.14620571153247e-01, 7.13752986225515e-01, 7.13096654406598e-01, 7.12613602448904e-01, 7.12269344976214e-01, 7.12033212587571e-01, 7.11878575586056e-01, 7.11782950351214e-01, 7.11727979363040e-01, 7.11699280720971e-01, 7.11686168341011e-01, 7.11681249477721e-01, 7.11679911431723e-01, 7.11679713939091e-01, 7.11679698874427e-01, 7.11679430538098e-01, 7.11677612522244e-01, 7.11670929336535e-01, 7.11653114669978e-01, 7.11614128990526e-01, 7.11539471664771e-01, 7.11409657324246e-01, 7.11199884105376e-01, 7.10879918824105e-01, 7.10414220758672e-01, 7.09762321046050e-01, 7.08879468257955e-01, 7.07717542086989e-01, 7.06226225998385e-01, 7.04354416246439e-01, 7.02051829282922e-01, 6.99270753236810e-01, 6.95967873234714e-01, 6.92106086653489e-01, 6.87656214939651e-01, 6.82598515285683e-01, 6.76923899688759e-01, 6.70634781427708e-01, 6.63745489459110e-01, 6.56282218201014e-01, 6.48282511164531e-01, 6.39794308702960e-01, 6.30874619312031e-01, 6.21587897261620e-01, 6.12004224492857e-01, 6.02197400460142e-01, 5.92243040017259e-01, 5.82216767820427e-01, 5.72192580214605e-01, 5.62241424807937e-01, 5.52430026549899e-01, 5.42819969364899e-01, 5.33467025874899e-01, 5.24420715372834e-01, 5.15724062195041e-01, 5.07413522660585e-01, 4.99519048110272e-01, 4.92064253440084e-01, 4.85066664019513e-01, 4.78538018255072e-01, 4.72484607696380e-01, 4.66907641056288e-01, 4.61803622548857e-01, 4.57164738389736e-01, 4.52979248088752e-01, 4.49231879286543e-01, 4.45904226367958e-01, 4.42975153959253e-01, 4.40421206722739e-01, 4.38217026646747e-01, 4.36335778339220e-01, 4.34749581731765e-01, 4.33429950169675e-01, 4.32348230199381e-01, 4.31476037594699e-01, 4.30785682426261e-01, 4.30250574431497e-01, 4.29845598739597e-01, 4.29547451292727e-01, 4.29334923195697e-01, 4.29189123798693e-01, 4.29093633590412e-01, 4.29034579916248e-01, 4.29000631040324e-01, 4.28982906993198e-01, 4.28974808804014e-01, 4.28971770905704e-01, 4.28970944524835e-01, 4.28970822552792e-01, 4.28970813183163e-01, 4.28970645427626e-01, 4.28969508860000e-01, 4.28965330728651e-01, 4.28954193452924e-01, 4.28929820246132e-01, 4.28883144358572e-01, 4.28801979977315e-01, 4.28670811195161e-01, 4.28470713528313e-01, 4.28179420065916e-01, 4.27771541271525e-01, 4.27218943511967e-01, 4.26491286378415e-01, 4.25556712696938e-01, 4.24382677849238e-01, 4.22936896886185e-01, 4.21188379382926e-01, 4.19108513745378e-01, 4.16672155596203e-01, 4.13858669893891e-01, 4.10652874468106e-01, 4.07045834381023e-01, 4.03035462277579e-01, 3.98626889525957e-01, 3.93832585812994e-01, 3.88672219825551e-01, 3.83172269270773e-01, 3.77365403212977e-01, 3.71289672108219e-01, 3.64987549917714e-01, 3.58504877689136e-01, 3.51889758947635e-01, 3.45191454559242e-01, 3.38459319186342e-01, 3.31741813999237e-01, 3.25085621903895e-01, 3.18534883040158e-01, 3.12130560349402e-01, 3.05909938035953e-01, 2.99906249980943e-01, 2.94148430682156e-01, 2.88660978049592e-01, 2.83463915285132e-01, 2.78572837982331e-01, 2.73999032357246e-01, 2.69749651013491e-01, 2.65827933705182e-01, 2.62233462039681e-01, 2.58962438805739e-01, 2.56007984473342e-01, 2.53360445246670e-01, 2.51007708730384e-01, 2.48935524679979e-01, 2.47127829362732e-01, 2.45567072697156e-01, 2.44234547544352e-01, 2.43110720302361e-01, 2.42175561353562e-01, 2.41408873010748e-01, 2.40790611510519e-01, 2.40301198434508e-01, 2.39921815841477e-01, 2.39634678499797e-01, 2.39423276044896e-01, 2.39272577746661e-01, 2.39169192919661e-01, 2.39101480864427e-01, 2.39059605562154e-01, 2.39035532088707e-01, 2.39022963759581e-01, 2.39017221230391e-01, 2.39015067016581e-01, 2.39014481018877e-01, 2.39014394526869e-01, 2.39014387832043e-01, 2.39014267306402e-01, 2.39013450728102e-01, 2.39010448907521e-01, 2.39002447214908e-01, 2.38984936005252e-01, 2.38951401089430e-01, 2.38893087174290e-01, 2.38798846009786e-01, 2.38655079576262e-01, 2.38445786920423e-01, 2.38152721068595e-01, 2.37755659671812e-01, 2.37232789552273e-01, 2.36561201064937e-01, 2.35717483197270e-01, 2.34678404756046e-01, 2.33421661126044e-01, 2.31926660355178e-01, 2.30175317258642e-01, 2.28152820433330e-01, 2.25848335109511e-01, 2.23255605115114e-01, 2.20373420176394e-01, 2.17205920358644e-01, 2.13762717400087e-01, 2.10058822465996e-01, 2.06114380673071e-01, 2.01954223693505e-01, 1.97607261911394e-01, 1.93105746128758e-01, 1.88484435053482e-01, 1.83779708345746e-01, 1.79028665727962e-01, 1.74268250715585e-01, 1.69534433261795e-01, 1.64861479557487e-01, 1.60281330006555e-01, 1.55823098656931e-01, 1.51512699719120e-01, 1.47372599778912e-01, 1.43421688321311e-01, 1.39675254509314e-01, 1.36145054942247e-01, 1.32839455361150e-01, 1.29763628861184e-01, 1.26919793914993e-01, 1.24307477150765e-01, 1.21923788086484e-01, 1.19763695626311e-01, 1.17820298827894e-01, 1.16085087042521e-01, 1.14548186847969e-01, 1.13198595116003e-01, 1.12024399006749e-01, 1.11012984626342e-01, 1.10151236520238e-01, 1.09425730138057e-01, 1.08822918953173e-01, 1.08329317136551e-01, 1.07931677669235e-01, 1.07617164646904e-01, 1.07373517403751e-01, 1.07189203081070e-01, 1.07053553497337e-01, 1.06956881728870e-01, 1.06890573743726e-01, 1.06847150770129e-01, 1.06820298811080e-01, 1.06804862789389e-01, 1.06796804139698e-01, 1.06793122154819e-01, 1.06791740927918e-01, 1.06791365201900e-01, 1.06791309745581e-01, 1.06791305418227e-01, 1.06791227064399e-01, 1.06790696206382e-01, 1.06788744727041e-01, 1.06783542882944e-01, 1.06772159197348e-01, 1.06750359698403e-01, 1.06712455259178e-01, 1.06651205227912e-01, 1.06557784841598e-01, 1.06421822983499e-01, 1.06231515541218e-01, 1.05973817800169e-01, 1.05634716844932e-01, 1.05199581762903e-01, 1.04653585563753e-01, 1.03982188259614e-01, 1.03171665740447e-01, 1.02209664292197e-01, 1.01085756314049e-01, 9.97919695265423e-02, 9.83232602552491e-02, 9.66779016752142e-02, 9.48577605027765e-02, 9.28684405897471e-02, 9.07192790059241e-02, 8.84231890066687e-02, 8.59963540514670e-02, 8.34577868746038e-02, 8.08287765628606e-02, 7.81322537839699e-02, 7.53921090354126e-02, 7.26325006311441e-02, 6.98771880305102e-02, 6.71489222860515e-02, 6.44689193721746e-02, 6.18564346706094e-02, 5.93284487405867e-02, 5.68994664933555e-02, 5.45814247221443e-02, 5.23836971432345e-02, 5.03131819884928e-02, 4.83744548356100e-02, 4.65699686478374e-02, 4.49002836617041e-02, 4.33643114792588e-02, 4.19595601448108e-02, 4.06823697892946e-02, 3.95281313217382e-02, 3.84914834067231e-02, 3.75664854202924e-02, 3.67467661245158e-02, 3.60256493981126e-02, 3.53962595075998e-02, 3.48516091283599e-02, 3.43846736671694e-02, 3.39884554479610e-02, 3.36560410481610e-02, 3.33806545711752e-02, 3.31557089634273e-02, 3.29748566934214e-02, 3.28320402652551e-02, 3.27215422049036e-02, 3.26380333969638e-02, 3.25766180221763e-02, 3.25328729039427e-02, 3.25028788506572e-02, 3.24832416000262e-02, 3.24711002279017e-02, 3.24641213525112e-02, 3.24604781010223e-02, 3.24588135493187e-02, 3.24581891314608e-02, 3.24580192758930e-02, 3.24579942056005e-02, 3.24579922326791e-02, 3.24579562964047e-02, 3.24577128235077e-02, 3.24568178016449e-02, 3.24544320806766e-02, 3.24492113879273e-02, 3.24392146666739e-02, 3.24218350715881e-02, 3.23937578761289e-02, 3.23509494167880e-02, 3.22886809315932e-02, 3.22015906300379e-02, 3.20837864397419e-02, 3.19289905031184e-02, 3.17307245788427e-02, 3.14825330331772e-02, 3.11782371713178e-02, 3.08122114480298e-02, 3.03796689008706e-02, 2.98769403502348e-02, 2.93017299342679e-02, 2.86533288109823e-02, 2.79327696989782e-02, 2.71429075218208e-02, 2.62884157333900e-02, 2.53756936550218e-02, 2.44126868345731e-02, 2.34086293505200e-02, 2.23737233428025e-02, 2.13187760972247e-02, 2.02548181227300e-02, 1.91927264796377e-02, 1.81428760966611e-02, 1.71148382389390e-02, 1.61171402051821e-02, 1.51570944373205e-02, 1.42406992323245e-02, 1.33726077807950e-02, 1.25561578028552e-02, 1.17934509174518e-02, 1.10854692027189e-02, 1.04322161179482e-02, 9.83286982882703e-03, 9.28593864681204e-03, 8.78941037755254e-03, 8.34088955642898e-03, 7.93771862173287e-03, 7.57708092643128e-03, 7.25608505242548e-03, 6.97183112747979e-03, 6.72146072914614e-03, 6.50219250550776e-03, 6.31134589738735e-03, 6.14635538510555e-03, 6.00477757775764e-03, 5.88429326975159e-03, 5.78270634165534e-03, 5.69794109766519e-03, 5.62803932178468e-03, 5.57115800832264e-03, 5.52556839132734e-03, 5.48965657345946e-03, 5.46192574880422e-03, 5.44099974150214e-03, 5.42562735275709e-03, 5.41468683868374e-03, 5.40718973981313e-03, 5.40228325679948e-03, 5.39925041712763e-03, 5.39750739797995e-03, 5.39659754804234e-03, 5.39618186734003e-03, 5.39602593718769e-03, 5.39598352097670e-03, 5.39597726044150e-03, 5.39597676340975e-03, 5.39596765457873e-03, 5.39590594121364e-03, 5.39567908106875e-03, 5.39507439137310e-03, 5.39375122614416e-03, 5.39121790913555e-03, 5.38681464782798e-03, 5.37970370604057e-03, 5.36886815329041e-03, 5.35312050048731e-03, 5.33112242162304e-03, 5.30141649435478e-03, 5.26247042090233e-03, 5.21273348075368e-03, 5.15070401743856e-03, 5.07500561503349e-03, 4.98446836929389e-03, 4.87821044556824e-03, 4.75571412981424e-03, 4.61689001711107e-03, 4.46212303027949e-03, 4.29229474055148e-03, 4.10877800047871e-03, 3.91340210370572e-03, 3.70838934468054e-03, 3.49626665059333e-03, 3.27975852314211e-03, 3.06166946966613e-03, 2.84476509273901e-03, 2.63166085564137e-03, 2.42472628546864e-03, 2.22601030127834e-03, 2.03719091044263e-03, 1.85955013796502e-03, 1.69397298319041e-03, 1.54096746577892e-03, 1.40070139944726e-03, 1.27305053383964e-03, 1.15765240831678e-03, 1.05396086524223e-03, 9.61297507615200e-04, 8.78897917253596e-04, 8.05951595815202e-04, 7.41635149758224e-04, 6.85138465684076e-04, 6.35683915703567e-04, 5.92539144577131e-04, 5.55024518163805e-04, 5.22516556479502e-04, 4.94448532535330e-04, 4.70309065846664e-04, 4.49639234609833e-04, 4.32028607086465e-04, 4.17110604558957e-04, 4.04557626062285e-04, 3.94076314435665e-04, 3.85403226658809e-04, 3.78301047014418e-04, 3.72555389532181e-04, 3.67972188051624e-04, 3.64375648715054e-04, 3.61606724162300e-04, 3.59522049393490e-04, 3.57993257239286e-04, 3.56906570971205e-04, 3.56162557878292e-04, 3.55675923835008e-04, 3.55375234327937e-04, 3.55202462586135e-04, 3.55112287685699e-04, 3.55071092307924e-04, 3.55055639516877e-04, 3.55051436081055e-04, 3.55050815656455e-04, 3.55050765953990e-04, 3.55049849353861e-04, 3.55043639267194e-04, 3.55020811156237e-04, 3.54959965974899e-04, 3.54826839178202e-04, 3.54572005683451e-04, 3.54129224199085e-04, 3.53414580317235e-04, 3.52326603326958e-04, 3.50747538639951e-04, 3.48545947638002e-04, 3.45580769746334e-04, 3.41706908244532e-04, 3.36782288123061e-04, 3.30676180702483e-04, 3.23278407783789e-04, 3.14508844041939e-04, 3.04326458589799e-04, 2.92737003995569e-04, 2.79798406538134e-04, 2.65622969195078e-04, 2.50375700019845e-04, 2.34268449339936e-04, 2.17550074737005e-04, 2.00493484939557e-04, 1.83380988779902e-04, 1.66489676907133e-04, 1.50078408267973e-04, 1.34377351933779e-04, 1.19580201250358e-04, 1.05838580765294e-04, 9.32581795396459e-05, 8.18967542812743e-05, 7.17648839696494e-05, 6.28305708048060e-05, 5.50280573818895e-05, 4.82697332344628e-05, 4.24585441933800e-05, 3.74978796315872e-05, 3.32968615879281e-05, 2.97707611451870e-05, 2.68380540475732e-05, 2.44167852017085e-05, 2.24229716252466e-05, 2.07724774625717e-05, 1.93858887342913e-05, 1.81946258592802e-05, 1.71460522295774e-05, 1.62054883716678e-05, 1.53542529020001e-05, 1.45847914936326e-05, 1.38951214481153e-05, 1.32844743561122e-05, 1.27509001472959e-05, 1.22906295061398e-05, 1.18984840682401e-05, 1.15685670394328e-05, 1.12947616628389e-05, 1.10709510749984e-05, 1.08910890190828e-05, 1.07492554022315e-05, 1.06397453976115e-05, 1.05571778235003e-05, 1.04965935605605e-05, 1.04535263858452e-05, 1.04240439284808e-05, 1.04047640050453e-05, 1.03928529142649e-05, 1.03860097482311e-05, 1.03824383330675e-05, 1.03808068274702e-05, 1.03801948441173e-05, 1.03800283781652e-05, 1.03800038075974e-05, 1.03800018322886e-05, 1.03799652890757e-05, 1.03797177359866e-05, 1.03788077012723e-05, 1.03763818017910e-05, 1.03710724630446e-05, 1.03609032464647e-05, 1.03432154006100e-05, 1.03146191108294e-05, 1.02709737106466e-05, 1.02074046850475e-05, 1.01183712430952e-05, 9.99780774435521e-06, 9.83937111698197e-06, 9.63682479948602e-06, 9.38456432919565e-06, 9.07822882822984e-06, 8.71526012247159e-06, 8.29522170420205e-06, 7.81975231234355e-06, 7.29223147999392e-06, 6.71748461655852e-06, 6.10201990269843e-06, 5.45531528928223e-06, 4.79240829776018e-06, 4.13698830215026e-06, 3.52208991927228e-06, 2.98355102756886e-06, 2.54354612394536e-06, 2.19265953660397e-06, 1.88832294867325e-06, 1.57458545402278e-06, 1.20680220305758e-06, 7.86834881694176e-07, 8.38765089383742e-07, 7.49266796162829e-07, 5.63578936407794e-07, 3.41342492049534e-07, 5.12248433892637e-07, 5.81654170161983e-07, 5.63477712489657e-07, 4.34556184814028e-07, 3.73370319550759e-07, 3.64098688976355e-07, -7.10542735760100e-15, -7.10542735760100e-15, 6.09202146506505e-07, 7.85107417300424e-07, 8.36045188634671e-07, 7.66049339517849e-07, 6.05908851980530e-07, 3.98327173911639e-07, 0.00000000000000e+00, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -1.42108547152020e-14, -7.10542735760100e-15, -1.42108547152020e-14, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, 0.00000000000000e+00, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15],
[ 4.70000000000000e+01, 4.69999999575351e+01, 4.69999986444371e+01, 4.69999897479904e+01, 4.69999570434967e+01, 4.69998698636747e+01, 4.69996790643990e+01, 4.69993136288947e+01, 4.69986780331406e+01, 4.69976504719932e+01, 4.69960820192511e+01, 4.69937967664891e+01, 4.69905929561380e+01, 4.69862450951691e+01, 4.69805070080513e+01, 4.69731157624776e+01, 4.69637963795415e+01, 4.69522672221047e+01, 4.69382459410776e+01, 4.69214558488464e+01, 4.69016325812770e+01, 4.68785309035080e+01, 4.68519315089547e+01, 4.68216476546875e+01, 4.67875314693096e+01, 4.67494797621223e+01, 4.67074391561573e+01, 4.66614103648097e+01, 4.66114514350931e+01, 4.65576797927241e+01, 4.65002729475492e+01, 4.64394677532931e+01, 4.63755581627388e+01, 4.63088914760279e+01, 4.62398631420302e+01, 4.61689102358460e+01, 4.60965037942203e+01, 4.60231402400878e+01, 4.59493321637577e+01, 4.58755987490379e+01, 4.58024561372554e+01, 4.57304080116484e+01, 4.56599366613145e+01, 4.55914947509865e+01, 4.55254979838848e+01, 4.54623188031575e+01, 4.54022812358596e+01, 4.53456569442124e+01, 4.52926625134239e+01, 4.52434579742888e+01, 4.51981465321761e+01, 4.51567754514783e+01, 4.51193380255129e+01, 4.50857765455448e+01, 4.50559861683849e+01, 4.50298195694380e+01, 4.50070922568092e+01, 4.49875884120807e+01, 4.49710671147620e+01, 4.49572688005125e+01, 4.49459217984865e+01, 4.49367487910464e+01, 4.49294730401698e+01, 4.49238242295649e+01, 4.49195437801644e+01, 4.49163895094320e+01, 4.49141395217595e+01, 4.49125952378865e+01, 4.49115834952495e+01, 4.49109576777886e+01, 4.49105978621498e+01, 4.49104099964604e+01, 4.49103241569063e+01, 4.49102919551751e+01, 4.49102831954903e+01, 4.49102819025779e+01, 4.49102817329751e+01, 4.49102777814324e+01, 4.49102510091301e+01, 4.49101525910259e+01, 4.49098902422760e+01, 4.49093160871980e+01, 4.49082164684135e+01, 4.49063040872228e+01, 4.49032128088298e+01, 4.48984954004175e+01, 4.48916243969423e+01, 4.48819962095264e+01, 4.48689385052517e+01, 4.48517207959791e+01, 4.48295680787627e+01, 4.48016772728184e+01, 4.47672360991338e+01, 4.47254439498272e+01, 4.46755341964175e+01, 4.46167972909534e+01, 4.45486039246009e+01, 4.44704274305423e+01, 4.43818645611190e+01, 4.42826537457862e+01, 4.41726899614621e+01, 4.40520354340648e+01, 4.39209255479026e+01, 4.37797695667516e+01, 4.36291460526835e+01, 4.34697931787914e+01, 4.33025944334166e+01, 4.31285604674428e+01, 4.29488080100095e+01, 4.27645368527060e+01, 4.25770058771817e+01, 4.23875089931884e+01, 4.21973516933051e+01, 4.20078287519496e+01, 4.18202034313196e+01, 4.16356884272009e+01, 4.14554287016627e+01, 4.12804863037846e+01, 4.11118272615190e+01, 4.09503106222111e+01, 4.07966797125742e+01, 4.06515556724950e+01, 4.05154332884806e+01, 4.03886791144647e+01, 4.02715318255176e+01, 4.01641047095686e+01, 4.00663901678223e+01, 3.99782660680334e+01, 3.98995037757350e+01, 3.98297776748634e+01, 3.97686759782897e+01, 3.97157126180791e+01, 3.96703399931593e+01, 3.96319623378444e+01, 3.95999494587396e+01, 3.95736505712241e+01, 3.95524079519169e+01, 3.95355701125272e+01, 3.95225041956327e+01, 3.95126072963600e+01, 3.95053164273729e+01, 3.95001168691079e+01, 3.94965486831534e+01, 3.94942112135290e+01, 3.94927654570426e+01, 3.94919342477539e+01, 3.94915002691577e+01, 3.94913019779126e+01, 3.94912275915613e+01, 3.94912073566268e+01, 3.94912043699912e+01, 3.94912041063282e+01, 3.94911989399800e+01, 3.94911639372008e+01, 3.94910352631264e+01, 3.94906922640950e+01, 3.94899416130553e+01, 3.94885040033338e+01, 3.94860039154278e+01, 3.94819629139370e+01, 3.94757968536954e+01, 3.94668172846351e+01, 3.94542372423546e+01, 3.94371814941845e+01, 3.94147011777182e+01, 3.93857926205646e+01, 3.93494199686646e+01, 3.93045410804902e+01, 3.92501359730992e+01, 3.91852369430782e+01, 3.91089593425428e+01, 3.90205318800749e+01, 3.89193252509190e+01, 3.88048778904812e+01, 3.86769176982520e+01, 3.85353787008707e+01, 3.83804118152224e+01, 3.82123891335026e+01, 3.80319014746705e+01, 3.78397493139706e+01, 3.76369275844075e+01, 3.74246051972201e+01, 3.72041004000033e+01, 3.69768532341298e+01, 3.67443963444152e+01, 3.65083252480539e+01, 3.62702689381267e+01, 3.60318614499939e+01, 3.57947148193044e+01, 3.55603937408223e+01, 3.53303921938341e+01, 3.51061123020032e+01, 3.48888457046158e+01, 3.46797577024326e+01, 3.44798743922105e+01, 3.42900729233509e+01, 3.41110749124279e+01, 3.39434429536974e+01, 3.37875800802931e+01, 3.36437319699219e+01, 3.35119916524839e+01, 3.33923064622312e+01, 3.32844869779718e+01, 3.31882177045245e+01, 3.31030692606857e+01, 3.30285118482478e+01, 3.29639297797513e+01, 3.29086368380474e+01, 3.28618922284669e+01, 3.28229168658013e+01, 3.27909097158647e+01, 3.27650638882585e+01, 3.27445821566446e+01, 3.27286915689409e+01, 3.27166568057117e+01, 3.27077919533633e+01, 3.27014703814825e+01, 3.26971324515732e+01, 3.26942908372013e+01, 3.26925333015815e+01, 3.26915228552599e+01, 3.26909953002330e+01, 3.26907542534757e+01, 3.26906638281094e+01, 3.26906392301790e+01, 3.26906355995746e+01, 3.26906353083552e+01, 3.26906299338803e+01, 3.26905935210133e+01, 3.26904596633348e+01, 3.26901028469284e+01, 3.26893219579701e+01, 3.26878264431118e+01, 3.26852256705080e+01, 3.26810219702300e+01, 3.26746077529051e+01, 3.26652670134557e+01, 3.26521814209504e+01, 3.26344410738945e+01, 3.26110598614900e+01, 3.25809952155806e+01, 3.25431718673182e+01, 3.24965090418752e+01, 3.24399503414753e+01, 3.23724953921740e+01, 3.22932321759638e+01, 3.22013688509707e+01, 3.20962637926824e+01, 3.19774525805751e+01, 3.18446707161247e+01, 3.16978709942626e+01, 3.15372346592797e+01, 3.13631757499564e+01, 3.11763383625427e+01, 3.09775869130370e+01, 3.07679898356665e+01, 3.05487974830642e+01, 3.03214152659201e+01, 3.00873732606711e+01, 2.98482936065118e+01, 2.96058570033006e+01, 2.93617695190919e+01, 2.91177307416078e+01, 2.88754040915121e+01, 2.86363898883474e+01, 2.84022015495698e+01, 2.81742451274281e+01, 2.79538022548290e+01, 2.77420164774822e+01, 2.75398828869297e+01, 2.73482409268868e+01, 2.71677702145367e+01, 2.69989891935006e+01, 2.68422564143663e+01, 2.66977742227403e+01, 2.65655946254433e+01, 2.64456271036562e+01, 2.63376481469270e+01, 2.62413122917282e+01, 2.61561644593731e+01, 2.60816533969652e+01, 2.60171460285772e+01, 2.59619425200437e+01, 2.59152918490000e+01, 2.58764076529064e+01, 2.58444841038551e+01, 2.58187115329920e+01, 2.57982915030984e+01, 2.57824510091548e+01, 2.57704554773020e+01, 2.57616202357119e+01, 2.57553201487905e+01, 2.57509971400144e+01, 2.57481653784206e+01, 2.57464139678398e+01, 2.57454070537025e+01, 2.57448813459550e+01, 2.57446411439567e+01, 2.57445510356054e+01, 2.57445265239217e+01, 2.57445229060476e+01, 2.57445226270023e+01, 2.57445176162448e+01, 2.57444836676203e+01, 2.57443588688068e+01, 2.57440262003851e+01, 2.57432981606130e+01, 2.57419038655966e+01, 2.57394791380789e+01, 2.57355600350377e+01, 2.57295801912781e+01, 2.57208722717843e+01, 2.57086737276795e+01, 2.56921369371060e+01, 2.56703436816400e+01, 2.56423237609033e+01, 2.56070773847981e+01, 2.55636008089958e+01, 2.55109145026123e+01, 2.54480929679511e+01, 2.53742951835859e+01, 2.52887945278426e+01, 2.51910069735653e+01, 2.50805163385024e+01, 2.49570954366771e+01, 2.48207221077570e+01, 2.46715893011920e+01, 2.45101086516130e+01, 2.43369072883282e+01, 2.41528179566578e+01, 2.39588628699893e+01, 2.37562320328442e+01, 2.35462570485829e+01, 2.33303816230503e+01, 2.31101300755212e+01, 2.28870751606045e+01, 2.26628063956295e+01, 2.24388999009715e+01, 2.22168905313798e+01, 2.19982468429377e+01, 2.17843492335102e+01, 2.15764714308062e+01, 2.13757653827214e+01, 2.11832495199722e+01, 2.09998002979865e+01, 2.08261468727226e+01, 2.06628687181561e+01, 2.05103959516698e+01, 2.03690121008715e+01, 2.02388590253829e+01, 2.01199437022608e+01, 2.00121465936420e+01, 1.99152313370861e+01, 1.98288555282020e+01, 1.97525823960066e+01, 1.96858931988338e+01, 1.96282001882934e+01, 1.95788599981001e+01, 1.95371873125796e+01, 1.95024686570270e+01, 1.94739761309917e+01, 1.94509808793044e+01, 1.94327660682786e+01, 1.94186391103841e+01, 1.94079428640356e+01, 1.94000655296588e+01, 1.93944489716780e+01, 1.93905952200761e+01, 1.93880709449258e+01, 1.93865097515649e+01, 1.93856122104023e+01, 1.93851436101258e+01, 1.93849295021266e+01, 1.93848491827170e+01, 1.93848273338801e+01, 1.93848241090371e+01, 1.93848238648618e+01, 1.93848195393194e+01, 1.93847902331337e+01, 1.93846825005045e+01, 1.93843953249423e+01, 1.93837668486187e+01, 1.93825632426957e+01, 1.93804701663618e+01, 1.93770872106092e+01, 1.93719256626850e+01, 1.93644098569388e+01, 1.93538822950680e+01, 1.93396126210959e+01, 1.93208104219445e+01, 1.92966416927101e+01, 1.92662486582960e+01, 1.92287724840727e+01, 1.91833782447968e+01, 1.91292813629251e+01, 1.90657745867506e+01, 1.89922544686740e+01, 1.89082462374869e+01, 1.88134259470622e+01, 1.87076388353453e+01, 1.85909129454369e+01, 1.84634672429251e+01, 1.83257137030327e+01, 1.81782531249909e+01, 1.80218647424226e+01, 1.78574900172228e+01, 1.76862113082157e+01, 1.75092263718754e+01, 1.73278198589236e+01, 1.71433330993117e+01, 1.69571335065520e+01, 1.67705848764878e+01, 1.65850197112620e+01, 1.64017144824275e+01, 1.62218684821997e+01, 1.60465866280892e+01, 1.58768663133169e+01, 1.57135881592069e+01, 1.55575103442654e+01, 1.54092660665890e+01, 1.52693636409889e+01, 1.51381887315836e+01, 1.50160082615400e+01, 1.49029756091501e+01, 1.47991367791724e+01, 1.47044373183980e+01, 1.46187298160530e+01, 1.45417818876036e+01, 1.44732845823667e+01, 1.44128611808131e+01, 1.43600763577527e+01, 1.43144456847114e+01, 1.42754454310972e+01, 1.42425226017306e+01, 1.42151051205404e+01, 1.41926120393353e+01, 1.41744636192544e+01, 1.41600911035547e+01, 1.41489459765677e+01, 1.41405084875908e+01, 1.41342952124501e+01, 1.41298654312391e+01, 1.41268261192446e+01, 1.41248353793798e+01, 1.41236041875379e+01, 1.41228963752045e+01, 1.41225268335848e+01, 1.41223579870138e+01, 1.41222946468434e+01, 1.41222774167819e+01, 1.41222748736610e+01, 1.41222746827481e+01, 1.41222713224649e+01, 1.41222485560551e+01, 1.41221648643809e+01, 1.41219417735369e+01, 1.41214535468733e+01, 1.41205185428023e+01, 1.41188925891417e+01, 1.41162646862147e+01, 1.41122553058710e+01, 1.41064175002105e+01, 1.40982409696187e+01, 1.40871591621533e+01, 1.40725593832517e+01, 1.40537957853052e+01, 1.40302049820267e+01, 1.40011238965155e+01, 1.39659093113535e+01, 1.39239584538265e+01, 1.38747298318684e+01, 1.38177634504222e+01, 1.37526994972305e+01, 1.36792946030020e+01, 1.35974348603859e+01, 1.35071449297907e+01, 1.34085927607086e+01, 1.33020896998750e+01, 1.31880860206242e+01, 1.30671621653420e+01, 1.29400162186424e+01, 1.28074483002852e+01, 1.26703426691069e+01, 1.25296483580025e+01, 1.23863591220880e+01, 1.22414933938617e+01, 1.20960748224665e+01, 1.19511138516894e+01, 1.18075906814191e+01, 1.16664398702130e+01, 1.15285367734860e+01, 1.13946859659135e+01, 1.12656117571390e+01, 1.11419508664340e+01, 1.10242472686123e+01, 1.09129491606624e+01, 1.08084079324810e+01, 1.07108789653140e+01, 1.06205240373334e+01, 1.05374150932580e+01, 1.04615391353007e+01, 1.03928040125503e+01, 1.03310449185824e+01, 1.02760314448390e+01, 1.02274750730015e+01, 1.01850370179003e+01, 1.01483363504088e+01, 1.01169583363601e+01, 1.00904629235825e+01, 1.00683932966242e+01, 1.00502844002568e+01, 1.00356713113747e+01, 1.00240973174939e+01, 1.00151215415956e+01, 1.00083259402158e+01, 1.00033214966321e+01, 9.99975343534577e+00, 9.99730529867019e+00, 9.99570175115027e+00, 9.99471001192965e+00, 9.99413985743428e+00, 9.99384218455665e+00, 9.99370617516663e+00, 9.99365515331753e+00, 9.99364127413451e+00, 9.99363922559606e+00, 9.99363907227928e+00, 9.99363637995958e+00, 9.99361813909884e+00, 9.99355108378565e+00, 9.99337233909677e+00, 9.99298116151255e+00, 9.99223201269007e+00, 9.99092924614344e+00, 9.98882365468292e+00, 9.98561108954242e+00, 9.98093331875972e+00, 9.97438124089970e+00, 9.96550050887748e+00, 9.95379954607154e+00, 9.93875985261928e+00, 9.91984840462812e+00, 9.89653184563798e+00, 9.86829206264379e+00, 9.83464263490927e+00, 9.79514555093659e+00, 9.74942751666648e+00, 9.69719513590758e+00, 9.63824824101449e+00, 9.57249069495161e+00, 9.49993807902693e+00, 9.42072182346624e+00, 9.33508952500563e+00, 9.24340141501368e+00, 9.14612317507319e+00, 9.04381552067249e+00, 8.93712116102795e+00, 8.82674986922972e+00, 8.71346244477602e+00, 8.59805431628257e+00, 8.48133942770758e+00, 8.36413490254100e+00, 8.24724681972347e+00, 8.13145729293316e+00, 8.01751294124445e+00, 7.90611477966962e+00, 7.79790953546673e+00, 7.69348239540557e+00, 7.59335119435061e+00, 7.49796205445165e+00, 7.40768647127430e+00, 7.32281981885711e+00, 7.24358121462862e+00, 7.17011465363379e+00, 7.10249129529805e+00, 7.04071276874910e+00, 6.98471535585702e+00, 6.93437491379182e+00, 6.88951240860993e+00, 6.84989994493352e+00, 6.81526719083805e+00, 6.78530810871542e+00, 6.75968791004257e+00, 6.73805015355262e+00, 6.72002390218621e+00, 6.70523084522354e+00, 6.69329227974413e+00, 6.68383583212258e+00, 6.67650178799882e+00, 6.67094889039399e+00, 6.66685946244688e+00, 6.66394371519819e+00, 6.66194311287329e+00, 6.66063268838352e+00, 6.65982222968887e+00, 6.65935629194676e+00, 6.65911302915802e+00, 6.65900188008469e+00, 6.65896018418209e+00, 6.65894884188110e+00, 6.65894716778106e+00, 6.65894704244111e+00, 6.65894484078209e+00, 6.65892992421846e+00, 6.65887508935616e+00, 6.65872891977303e+00, 6.65840903135351e+00, 6.65779640729171e+00, 6.65673105057768e+00, 6.65500915542635e+00, 6.65238196849074e+00, 6.64855647495796e+00, 6.64319800258306e+00, 6.63593478715802e+00, 6.62636448495107e+00, 6.61406255107988e+00, 6.59859232844226e+00, 6.57951661185913e+00, 6.55641037005091e+00, 6.52887422888467e+00, 6.49654824895164e+00, 6.45912547543924e+00, 6.41636470473539e+00, 6.36810190555777e+00, 6.31425975624327e+00, 6.25485481550947e+00, 6.19000193035839e+00, 6.11991559822283e+00, 6.04490813523403e+00, 5.96538465122658e+00, 5.88183498621598e+00, 5.79482291323889e+00, 5.70497304891175e+00, 5.61295602622773e+00, 5.51947256517025e+00, 5.42523711864025e+00, 5.33096176978824e+00, 5.23734101174310e+00, 5.14503795590642e+00, 5.05467239856668e+00, 4.96681103896453e+00, 4.88195999820503e+00, 4.80055965057077e+00, 4.72298165804708e+00, 4.64952800335533e+00, 4.58043175092446e+00, 4.51585922972925e+00, 4.45591332435563e+00, 4.40063757631727e+00, 4.35002083056379e+00, 4.30400220594662e+00, 4.26247621716446e+00, 4.22529792425724e+00, 4.19228803001253e+00, 4.16323788280592e+00, 4.13791437063275e+00, 4.11606471059099e+00, 4.09742114686101e+00, 4.08170556999757e+00, 4.06863406236016e+00, 4.05792136046622e+00, 4.04928520699475e+00, 4.04245054535208e+00, 4.03715349047071e+00, 4.03314499310117e+00, 4.03019410327951e+00, 4.02809073349402e+00, 4.02664782435011e+00, 4.02570282558067e+00, 4.02511842270518e+00, 4.02478246345202e+00, 4.02460706661374e+00, 4.02452692730164e+00, 4.02449686443298e+00, 4.02448868661833e+00, 4.02448747959054e+00, 4.02448738898059e+00, 4.02448579416949e+00, 4.02447498909753e+00, 4.02443526862784e+00, 4.02432938934392e+00, 4.02409767971273e+00, 4.02365394564679e+00, 4.02288234345887e+00, 4.02163537709711e+00, 4.01973315861181e+00, 4.01696404860039e+00, 4.01308676504576e+00, 4.00783401254902e+00, 4.00091763782959e+00, 3.99203526065576e+00, 3.98087826242120e+00, 3.96714093930519e+00, 3.95053054702603e+00, 3.93077788506579e+00, 3.90764799684727e+00, 3.88095050656442e+00, 3.85054908120104e+00, 3.81636950483030e+00, 3.77840588667510e+00, 3.73672459671555e+00, 3.69146563112003e+00, 3.64284124853051e+00, 3.59113187731214e+00, 3.53667946012824e+00, 3.47987856062650e+00, 3.42116569260667e+00, 3.36100743173261e+00, 3.29988792439732e+00, 3.23829641361500e+00, 3.17671535933204e+00, 3.11560964718308e+00, 3.05541726648581e+00, 2.99654170857576e+00, 2.93934620424015e+00, 2.88414979637272e+00, 2.83122514056370e+00, 2.78079784809647e+00, 2.73304713500639e+00, 2.68810751650786e+00, 2.64607128484148e+00, 2.60699152559393e+00, 2.57088545739729e+00, 2.53773791736864e+00, 2.50750485511421e+00, 2.48011673791061e+00, 2.45548180609244e+00, 2.43348914893811e+00, 2.41401159644332e+00, 2.39690844089483e+00, 2.38202801413588e+00, 2.36921015219227e+00, 2.35828857906692e+00, 2.34909323674535e+00, 2.34145257966160e+00, 2.33519584006481e+00, 2.33015525703548e+00, 2.32616824756139e+00, 2.32307948438594e+00, 2.32074283356044e+00, 2.31902309593630e+00, 2.31779749218019e+00, 2.31695683093196e+00, 2.31640630470693e+00, 2.31606586788454e+00, 2.31587016501094e+00, 2.31576799467839e+00, 2.31572131316958e+00, 2.31570380147930e+00, 2.31569903788962e+00, 2.31569833479429e+00, 2.31569828178908e+00, 2.31569734585871e+00, 2.31569100480016e+00, 2.31566769450909e+00, 2.31560555868362e+00, 2.31546958027344e+00, 2.31520918191971e+00, 2.31475639884454e+00, 2.31402472021069e+00, 2.31290868593371e+00, 2.31128431411597e+00, 2.30901041817268e+00, 2.30593085073073e+00, 2.30187768212345e+00, 2.29667528398361e+00, 2.29014524321965e+00, 2.28211197998244e+00, 2.27240888799884e+00, 2.26088476123988e+00, 2.24741022290027e+00, 2.23188383737443e+00, 2.21423756949691e+00, 2.19444126287819e+00, 2.17250584374435e+00, 2.14848501841009e+00, 2.12247531811349e+00, 2.09461444779212e+00, 2.06507800611012e+00, 2.03407475169053e+00, 2.00184068403146e+00, 1.96863227749688e+00, 1.93471924652305e+00, 1.90037722706149e+00, 1.86588073466386e+00, 1.83149670851235e+00, 1.79747888074502e+00, 1.76406313055231e+00, 1.73146390158610e+00, 1.69987168685439e+00, 1.66945152308766e+00, 1.64034238983914e+00, 1.61265737833569e+00, 1.58648448046845e+00, 1.56188784711065e+00, 1.53890937423062e+00, 1.51757049186335e+00, 1.49787405191173e+00, 1.47980623339331e+00, 1.46333840610104e+00, 1.44842891424290e+00, 1.43502475951630e+00, 1.42306317772731e+00, 1.41247311427502e+00, 1.40317661159986e+00, 1.39509012620779e+00, 1.38812579438761e+00, 1.38219266456200e+00, 1.37719791073325e+00, 1.37304803613385e+00, 1.36965006946409e+00, 1.36691274854677e+00, 1.36474767846358e+00, 1.36307044390713e+00, 1.36180164923248e+00, 1.36086785512739e+00, 1.36020237843836e+00, 1.35974592183192e+00, 1.35944700277390e+00, 1.35926215667637e+00, 1.35915589666477e+00, 1.35910042172337e+00, 1.35907507530352e+00, 1.35906556707359e+00, 1.35906298061273e+00, 1.35906259885684e+00, 1.35906256992026e+00, 1.35906205690026e+00, 1.35905858111845e+00, 1.35904580384074e+00, 1.35901174480006e+00, 1.35893720991168e+00, 1.35879447622107e+00, 1.35854629097737e+00, 1.35814523774858e+00, 1.35753351763994e+00, 1.35664318726856e+00, 1.35539688631652e+00, 1.35370907530653e+00, 1.35148778792092e+00, 1.34863688125518e+00, 1.34505874196137e+00, 1.34065737713764e+00, 1.33534178777005e+00, 1.32902949210508e+00, 1.32165003980508e+00, 1.31314833876788e+00, 1.30348760860819e+00, 1.29265178082350e+00, 1.28064718711529e+00, 1.26750341393889e+00, 1.25327325083580e+00, 1.23803171827522e+00, 1.22187422197039e+00, 1.20491393863256e+00, 1.18727858682334e+00, 1.16910677107263e+00, 1.15054410473412e+00, 1.13173931639371e+00, 1.11284052750229e+00, 1.09399185857314e+00, 1.07533048224406e+00, 1.05698419864444e+00, 1.03906956641221e+00, 1.02169058510565e+00, 1.00493789421269e+00, 9.88888431783970e-01, 9.73605482082789e-01, 9.59139035840295e-01, 9.45526387429751e-01, 9.32792898938075e-01, 9.20952870090737e-01, 9.10010463785952e-01, 8.99960648377935e-01, 8.90790128865021e-01, 8.82478249122130e-01, 8.74997855853451e-01, 8.68316121815383e-01, 8.62395331001835e-01, 8.57193631927146e-01, 8.52665766982554e-01, 8.48763786215521e-01, 8.45437752956592e-01, 8.42636446682839e-01, 8.40308065581581e-01, 8.38400927710268e-01, 8.36864165721174e-01, 8.35648406144905e-01, 8.34706420535873e-01, 8.33993732707931e-01, 8.33469164138528e-01, 8.33095298646370e-01, 8.32838847825684e-01, 8.32670900508198e-01, 8.32567042668863e-01, 8.32507338508684e-01, 8.32476168654154e-01, 8.32461927140784e-01, 8.32456584698001e-01, 8.32455131427849e-01, 8.32454916928356e-01, 8.32454900566432e-01, 8.32454609122628e-01, 8.32452634550094e-01, 8.32445375844316e-01, 8.32426027043255e-01, 8.32383683901092e-01, 8.32302596453005e-01, 8.32161599514649e-01, 8.31933750741840e-01, 8.31586202948301e-01, 8.31080333588112e-01, 8.30372149200556e-01, 8.29412975703043e-01, 8.28150436284112e-01, 8.26529707060722e-01, 8.24495026685135e-01, 8.21991420186151e-01, 8.18966580441320e-01, 8.15372834199806e-01, 8.11169105286297e-01, 8.06322777483679e-01, 8.00811355512685e-01, 7.94623825992161e-01, 7.87761632052195e-01, 7.80239195182411e-01, 7.72083944646830e-01, 7.63335846083770e-01, 7.54046453682967e-01, 7.44277541215418e-01, 7.34099392984682e-01, 7.23588853936953e-01, 7.12827247207109e-01, 7.01898266993496e-01, 6.90885945714690e-01, 6.79872778711591e-01, 6.68938069690675e-01, 6.58156538198995e-01, 6.47597209012140e-01, 6.37322584262066e-01, 6.27388083671086e-01, 6.17841726990953e-01, 6.08724025704284e-01, 6.00068047840736e-01, 5.91899619729212e-01, 5.84237630871456e-01, 5.77094412097388e-01, 5.70476162045608e-01, 5.64383402222148e-01, 5.58811445982542e-01, 5.53750871440670e-01, 5.49187992329529e-01, 5.45105324106551e-01, 5.41482045055993e-01, 5.38294453790094e-01, 5.35516425410734e-01, 5.33119868718572e-01, 5.31075186314560e-01, 5.29351738321139e-01, 5.27918308868784e-01, 5.26743572582660e-01, 5.25796556220264e-01, 5.25047088529746e-01, 5.24466229505990e-01, 5.24026668703620e-01, 5.23703081292581e-01, 5.23472430250948e-01, 5.23314203570614e-01, 5.23210576632664e-01, 5.23146491950605e-01, 5.23109651173080e-01, 5.23090417416952e-01, 5.23081629457636e-01, 5.23078332807316e-01, 5.23077436040076e-01, 5.23077303679166e-01, 5.23077293511484e-01, 5.23077111467522e-01, 5.23075878093870e-01, 5.23071344096721e-01, 5.23059258238931e-01, 5.23032809212957e-01, 5.22982158379506e-01, 5.22894083086662e-01, 5.22751748603120e-01, 5.22534624688447e-01, 5.22218560300381e-01, 5.21776026663403e-01, 5.21176534640212e-01, 5.20387226840732e-01, 5.19373638056990e-01, 5.18100609474395e-01, 5.16533332949749e-01, 5.14638491986844e-01, 5.12385456655892e-01, 5.09747481572354e-01, 5.06702850256154e-01, 5.03235906771707e-01, 4.99337917324091e-01, 4.95007710883911e-01, 4.90252058834045e-01, 4.85085768390093e-01, 4.79531481906058e-01, 4.73619192498802e-01, 4.67385503889339e-01, 4.60872677232011e-01, 4.54127518598341e-01, 4.47200166845953e-01, 4.40142842608942e-01, 4.33008615478236e-01, 4.25850238973077e-01, 4.18719092823707e-01, 4.11664260688198e-01, 4.04731759931060e-01, 3.97963929506474e-01, 3.91398973025204e-01, 3.85070647143742e-01, 3.79008080610404e-01, 3.73235706525477e-01, 3.67773289353899e-01, 3.62636028618105e-01, 3.57834722617049e-01, 3.53375977603250e-01, 3.49262450288016e-01, 3.45493114075779e-01, 3.42063541855389e-01, 3.38966200357113e-01, 3.36190752922818e-01, 3.33724368974686e-01, 3.31552039473550e-01, 3.29656898223220e-01, 3.28020549010326e-01, 3.26623398301798e-01, 3.25444992595664e-01, 3.24464358605510e-01, 3.23660343333344e-01, 3.23011949855136e-01, 3.22498663417910e-01, 3.22100761352267e-01, 3.21799599456710e-01, 3.21577867021013e-01, 3.21419802610528e-01, 3.21311363185870e-01, 3.21240340097297e-01, 3.21196416940204e-01, 3.21171166114745e-01, 3.21157983090487e-01, 3.21151959699414e-01, 3.21149700124217e-01, 3.21149085465599e-01, 3.21148994743297e-01, 3.21148987721017e-01, 3.21148861300536e-01, 3.21148004783659e-01, 3.21144856144564e-01, 3.21136463086567e-01, 3.21118095377770e-01, 3.21082920142068e-01, 3.21021753628855e-01, 3.20922901721772e-01, 3.20772099950574e-01, 3.20552561933312e-01, 3.20245142859648e-01, 3.19828621679399e-01, 3.19280101961965e-01, 3.18575526876288e-01, 3.17690298447403e-01, 3.16599985343878e-01, 3.15281097264112e-01, 3.13711897992114e-01, 3.11873223982026e-01, 3.09749271563113e-01, 3.07328314166902e-01, 3.04603311874651e-01, 3.01572379338516e-01, 2.98239084711902e-01, 2.94612561243888e-01, 2.90707423934819e-01, 2.86543495145551e-01, 2.82145354212361e-01, 2.77541735888413e-01, 2.72764809941989e-01, 2.67849378923458e-01, 2.62832032764344e-01, 2.57750297642701e-01, 2.52641812894055e-01, 2.47543564314640e-01, 2.42491195729457e-01, 2.37518413888402e-01, 2.32656495207678e-01, 2.27933897024826e-01, 2.23375971150411e-01, 2.19004773693726e-01, 2.14838962417005e-01, 2.10893771160379e-01, 2.07181050059425e-01, 2.03709360205842e-01, 2.00484111923529e-01, 1.97507736798016e-01, 1.94779884860992e-01, 1.92297639769507e-01, 1.90055746312026e-01, 1.88046846019503e-01, 1.86261717965152e-01, 1.84689522922220e-01, 1.83318049853881e-01, 1.82133964192673e-01, 1.81123057514533e-01, 1.80270498034318e-01, 1.79561080882614e-01, 1.78979476428459e-01, 1.78510474069483e-01, 1.78139218016618e-01, 1.77851430759318e-01, 1.77633619211214e-01, 1.77473258100797e-01, 1.77358945059737e-01, 1.77280522120874e-01, 1.77229158983046e-01, 1.77197394410022e-01, 1.77179133453095e-01, 1.77169599738249e-01, 1.77165243739530e-01, 1.77163609659587e-01, 1.77163165150745e-01, 1.77163099542177e-01, 1.77163094422639e-01, 1.77163001724786e-01, 1.77162373683622e-01, 1.77160064944111e-01, 1.77153910740223e-01, 1.77140442659088e-01, 1.77114650605738e-01, 1.77069800928990e-01, 1.76997319769171e-01, 1.76886749621353e-01, 1.76725785791618e-01, 1.76500397754012e-01, 1.76195038279531e-01, 1.75792940514960e-01, 1.75276499889023e-01, 1.74627733843643e-01, 1.73828808047894e-01, 1.72862613180051e-01, 1.71713371893503e-01, 1.70367251634154e-01, 1.68812956018385e-01, 1.67042265962671e-01, 1.65050502054008e-01, 1.62836881987047e-01, 1.60404751290635e-01, 1.57761671813034e-01, 1.54919360098326e-01, 1.51893476250493e-01, 1.48703272423916e-01, 1.45371117957275e-01, 1.41921924718453e-01, 1.38382500946101e-01, 1.34780864471544e-01, 1.31145546624609e-01, 1.27504916517061e-01, 1.23886552057805e-01, 1.20316679375925e-01, 1.16819696758618e-01, 1.13417793222688e-01, 1.10130665906610e-01, 1.06975335028771e-01, 1.03966050549062e-01, 1.01114281098596e-01, 9.84287732838283e-02, 9.59156680967155e-02, 9.35786607903637e-02, 9.14191910876028e-02, 8.94366518226519e-02, 8.76286058770219e-02, 8.59910033413982e-02, 8.45183929979285e-02, 8.32041242910364e-02, 8.20405378021718e-02, 8.10191437762597e-02, 8.01307894110579e-02, 7.93658163884245e-02, 7.87142104982195e-02, 7.81657452003586e-02, 7.77101206350821e-02, 7.73370989868098e-02, 7.70366363161230e-02, 7.67990100852529e-02, 7.66149407124530e-02, 7.64757046929674e-02, 7.63732362039065e-02, 7.63002137373761e-02, 7.62501282273789e-02, 7.62173293705999e-02, 7.61970473810010e-02, 7.61853882223846e-02, 7.61793013726546e-02, 7.61765203050899e-02, 7.61754770410548e-02, 7.61751932489148e-02, 7.61751513618307e-02, 7.61751480655022e-02, 7.61750880236320e-02, 7.61746812317980e-02, 7.61731858310455e-02, 7.61691997117282e-02, 7.61604765425545e-02, 7.61437719920721e-02, 7.61147269011460e-02, 7.60677937576446e-02, 7.59962123211864e-02, 7.58920395574947e-02, 7.57462380573628e-02, 7.55488257169148e-02, 7.52890875404830e-02, 7.49558479356196e-02, 7.45377988065670e-02, 7.40238752119140e-02, 7.34036665281792e-02, 7.26678472495479e-02, 7.18086081326916e-02, 7.08200657947415e-02, 6.96986275171625e-02, 6.84432882600632e-02, 6.70558389829381e-02, 6.55409693495628e-02, 6.39062536030934e-02, 6.21620154555202e-02, 6.03210756729879e-02, 5.83983939478969e-02, 5.64106238536439e-02, 5.43756054348279e-02, 5.23118236932660e-02, 5.02378625414792e-02, 4.81718826807480e-02, 4.61311485925151e-02, 4.41316248853454e-02, 4.21876561735388e-02, 4.03117380461140e-02, 3.85143800850827e-02, 3.68040558966314e-02, 3.51872302913137e-02, 3.36684504815921e-02, 3.22504865612459e-02, 3.09345063689932e-02, 2.97202707212634e-02, 2.86063365467371e-02, 2.75902574244284e-02, 2.66687732864384e-02, 2.58379834640508e-02, 2.50934996376699e-02, 2.44305773713123e-02, 2.38442266049219e-02, 2.33293026905699e-02, 2.28805803451309e-02, 2.24928133425664e-02, 2.21607829598796e-02, 2.18793381629823e-02, 2.16434302933379e-02, 2.14481446155119e-02, 2.12887305468072e-02, 2.11606317592228e-02, 2.10595166754857e-02, 2.09813092244531e-02, 2.09222191249694e-02, 2.08787704742193e-02, 2.08478270613810e-02, 2.08266126386007e-02, 2.08127243714245e-02, 2.08041378596917e-02, 2.07992024481101e-02, 2.07966260027561e-02, 2.07954488685118e-02, 2.07950072953196e-02, 2.07948871777006e-02, 2.07948694486149e-02, 2.07948680410723e-02, 2.07948422458983e-02, 2.07946674804020e-02, 2.07940250324228e-02, 2.07923125673446e-02, 2.07885652108786e-02, 2.07813898450908e-02, 2.07689157980155e-02, 2.07487650897988e-02, 2.07180453468538e-02, 2.06733683372491e-02, 2.06108967234755e-02, 2.05264209737805e-02, 2.04154673322776e-02, 2.02734362648656e-02, 2.00957688674634e-02, 1.98781364160965e-02, 1.96166457069111e-02, 1.93080503206176e-02, 1.89499557612365e-02, 1.85410049123362e-02, 1.80810297597276e-02, 1.75711561082537e-02, 1.70138502019412e-02, 1.64128997041075e-02, 1.57733261861281e-02, 1.51012317124284e-02, 1.44035877281681e-02, 1.36879795582843e-02, 1.29623236655121e-02, 1.22345767723075e-02, 1.15124557659669e-02, 1.08031852258037e-02, 1.01132860892008e-02, 9.44841509002003e-03, 8.81326045662689e-03, 8.21149484318795e-03, 7.64578156717732e-03, 7.11782561297980e-03, 6.62845786349919e-03, 6.17774069817045e-03, 5.76508516490293e-03, 5.38937269037376e-03, 5.04907589438375e-03, 4.74237309673953e-03, 4.46725084692901e-03, 4.22158993722377e-03, 4.00323319847473e-03, 3.81003656374901e-03, 3.63990669972480e-03, 3.49082836814318e-03, 3.36088341711616e-03, 3.24826223324948e-03, 3.15126836151336e-03, 3.06831758169324e-03, 2.99793328137099e-03, 2.93874000768568e-03, 2.88945666623164e-03, 2.84889028847601e-03, 2.81593085269094e-03, 2.78954738097070e-03, 2.76878539022363e-03, 2.75276566505767e-03, 2.74068419756901e-03, 2.73181300435965e-03, 2.72550140833516e-03, 2.72117729146970e-03, 2.71834779482560e-03, 2.71659896881715e-03, 2.71559394988685e-03, 2.71506935034438e-03, 2.71482968177139e-03, 2.71473977792169e-03, 2.71471532226997e-03, 2.71471171268445e-03, 2.71471142350066e-03, 2.71470609065716e-03, 2.71466996012748e-03, 2.71453714375980e-03, 2.71418312954808e-03, 2.71340850715518e-03, 2.71192550704313e-03, 2.70934810394863e-03, 2.70518648465412e-03, 2.69884672234610e-03, 2.68963650854914e-03, 2.67677773095443e-03, 2.65942651438422e-03, 2.63670102726365e-03, 2.60771687953110e-03, 2.57162930186183e-03, 2.52768053237418e-03, 2.47525000807514e-03, 2.41390415024512e-03, 2.34344185364677e-03, 2.26393135706360e-03, 2.17573413010541e-03, 2.07951194575884e-03, 1.97621464693754e-03, 1.86704834368356e-03, 1.75342649966836e-03, 1.63690858493482e-03, 1.51913171449536e-03, 1.40173988388170e-03, 1.28631384674094e-03, 1.17430358825743e-03, 1.06696658058070e-03, 9.65319640663154e-04, 8.70115861346221e-04, 7.81852365904001e-04, 7.00801363414882e-04, 6.27047896863075e-04, 5.60517897007173e-04, 5.00987740785774e-04, 4.48085796492137e-04, 4.01319581044390e-04, 3.60144236083215e-04, 3.24027822202311e-04, 2.92478748690428e-04, 2.64987560299801e-04, 2.41355930775455e-04, 2.21065017065314e-04, 2.03569606668452e-04, 1.88515797603372e-04, 1.75582820588716e-04, 1.64501712156095e-04, 1.55033106977953e-04, 1.46959202197650e-04, 1.40086027393238e-04, 1.34405886448974e-04, 1.29725733948760e-04, 1.25818471381933e-04, 1.22582710019969e-04, 1.19929313505907e-04, 1.17779659831285e-04, 1.16063122383991e-04, 1.14715334902371e-04, 1.13677452247885e-04, 1.12896009142105e-04, 1.12322947316557e-04, 1.11915617729608e-04, 1.11636740918186e-04, 1.11454340121497e-04, 1.11341636696238e-04, 1.11276879508182e-04, 1.11243080887391e-04, 1.11227640410050e-04, 1.11221848548837e-04, 1.11220273048218e-04, 1.11220040508897e-04, 1.11220021750569e-04, 1.11219674181484e-04, 1.11217319336276e-04, 1.11208662943341e-04, 1.11185590142782e-04, 1.11135105889559e-04, 1.11038460829604e-04, 1.10870514205885e-04, 1.10599390644950e-04, 1.10186495795972e-04, 1.09586974524234e-04, 1.08750745610564e-04, 1.07624389137584e-04, 1.06154593041197e-04, 1.04295126710952e-04, 1.02017834628043e-04, 9.93009086585062e-05, 9.61033089552643e-05, 9.23966829589062e-05, 8.81845922506841e-05, 8.35070161002704e-05, 7.84409440441891e-05, 7.30635652885780e-05, 6.74293940647885e-05, 6.16132056379115e-05, 5.57808803876014e-05, 5.01282096792011e-05, 4.46104807494407e-05, 3.92660239967313e-05, 3.42153347361318e-05, 2.97660377412967e-05, 2.56251051382606e-05, 2.18073561129017e-05, 1.87343461846012e-05, 1.59963885479897e-05, 1.35241770280459e-05, 1.12670673431126e-05, 9.40293502793565e-06, 8.11368186504069e-06, 6.93210751734341e-06, 5.85087509819004e-06, 4.87315686825696e-06, 4.02676754163167e-06, 3.34174539062815e-06, 2.71695223830193e-06, 2.23286081535434e-06, 1.91861790455050e-06, 1.67231647196786e-06, 1.48307336189646e-06, 1.27077249345575e-06, 9.81040365388708e-07, 6.33355888624010e-07, 0.00000000000000e+00, -7.10542735760100e-15, -7.10542735760100e-15, 0.00000000000000e+00, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, 0.00000000000000e+00, 0.00000000000000e+00, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -1.42108547152020e-14, -7.10542735760100e-15, -7.10542735760100e-15, -1.42108547152020e-14, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -1.42108547152020e-14, -7.10542735760100e-15],
[ 4.80000000000000e+01, 4.79999999562444e+01, 4.79999986032360e+01, 4.79999894363899e+01, 4.79999557378724e+01, 4.79998659082945e+01, 4.79996693098202e+01, 4.79992927671136e+01, 4.79986378525764e+01, 4.79975790583748e+01, 4.79959629306399e+01, 4.79936082119691e+01, 4.79903070081131e+01, 4.79858269647507e+01, 4.79799144117801e+01, 4.79722984066912e+01, 4.79626955862325e+01, 4.79508157172280e+01, 4.79363678231061e+01, 4.79190667520070e+01, 4.78986400443546e+01, 4.78748349513211e+01, 4.78474254494633e+01, 4.78162190900256e+01, 4.77810635136754e+01, 4.77418524533045e+01, 4.76985310405437e+01, 4.76511002282083e+01, 4.75996201439699e+01, 4.75442122031043e+01, 4.74850598325181e+01, 4.74224076954927e+01, 4.73565593561307e+01, 4.72878733819243e+01, 4.72167579481635e+01, 4.71436640739011e+01, 4.70690776803452e+01, 4.69935107137219e+01, 4.69174916118266e+01, 4.68415554142812e+01, 4.67662338203740e+01, 4.66920454864502e+01, 4.66194868296932e+01, 4.65490235701966e+01, 4.64810832022166e+01, 4.64160485419600e+01, 4.63542524561992e+01, 4.62959738356424e+01, 4.62414348407058e+01, 4.61907994157397e+01, 4.61441730408568e+01, 4.61016036678215e+01, 4.60630837673268e+01, 4.60285533986255e+01, 4.59979041982399e+01, 4.59709841718097e+01, 4.59476031617602e+01, 4.59275388533105e+01, 4.59105431725382e+01, 4.58963489231133e+01, 4.58846765033676e+01, 4.58752405431198e+01, 4.58677563006617e+01, 4.58619456650126e+01, 4.58575426173166e+01, 4.58542980182632e+01, 4.58519836056278e+01, 4.58503951071829e+01, 4.58493543988303e+01, 4.58487106651464e+01, 4.58483405487412e+01, 4.58481473049154e+01, 4.58480590079988e+01, 4.58480258844184e+01, 4.58480168739675e+01, 4.58480155440426e+01, 4.58480153695845e+01, 4.58480113049199e+01, 4.58479837661999e+01, 4.58478825306623e+01, 4.58476126716649e+01, 4.58470220805696e+01, 4.58458909847538e+01, 4.58439238644544e+01, 4.58407441123438e+01, 4.58358917125137e+01, 4.58288241405628e+01, 4.58189206043357e+01, 4.58054896562710e+01, 4.57877801142176e+01, 4.57649951291878e+01, 4.57363091371420e+01, 4.57008873289062e+01, 4.56579071689323e+01, 4.56065813911541e+01, 4.55461818006039e+01, 4.54760631162685e+01, 4.53956860101545e+01, 4.53046384395926e+01, 4.52026543479021e+01, 4.50896288380763e+01, 4.49656290192345e+01, 4.48308998943434e+01, 4.46858648977431e+01, 4.45311209863193e+01, 4.43674285093993e+01, 4.41956963910210e+01, 4.40169634138417e+01, 4.38323765636014e+01, 4.36431674588670e+01, 4.34506278541406e+01, 4.32560850851507e+01, 4.30608781558318e+01, 4.28663349839292e+01, 4.26737511583930e+01, 4.24843704374692e+01, 4.22993671382430e+01, 4.21198305302141e+01, 4.19467513329161e+01, 4.17810104139091e+01, 4.16233697747935e+01, 4.14744658914434e+01, 4.13348054394108e+01, 4.12047633906445e+01, 4.10845834200775e+01, 4.09743805167966e+01, 4.08741456588272e+01, 4.07837523845090e+01, 4.07029650759159e+01, 4.06314487580487e+01, 4.05687802083750e+01, 4.05144601619013e+01, 4.04679263855221e+01, 4.04285673812870e+01, 4.03957364619979e+01, 4.03687659256409e+01, 4.03469810396619e+01, 4.03297135344310e+01, 4.03163142998622e+01, 4.03061649823089e+01, 4.02986881923070e+01, 4.02933560586354e+01, 4.02896969008312e+01, 4.02872998402145e+01, 4.02858172272497e+01, 4.02849648285213e+01, 4.02845197869469e+01, 4.02843164408942e+01, 4.02842401583034e+01, 4.02842194075451e+01, 4.02842163447751e+01, 4.02842160743907e+01, 4.02842107763434e+01, 4.02841748812829e+01, 4.02840429270897e+01, 4.02836911844751e+01, 4.02829213983666e+01, 4.02814471428870e+01, 4.02788833282878e+01, 4.02747393290964e+01, 4.02684161228894e+01, 4.02592077374575e+01, 4.02463071990511e+01, 4.02288170541934e+01, 4.02057644011349e+01, 4.01761202145998e+01, 4.01388225812469e+01, 4.00928032879236e+01, 4.00370170277831e+01, 3.99704723206906e+01, 3.98922630961147e+01, 3.98015997719476e+01, 3.96978385942900e+01, 3.95805079926652e+01, 3.94493307615437e+01, 3.93042410087248e+01, 3.91453950167079e+01, 3.89731754425278e+01, 3.87881886248666e+01, 3.85912551533611e+01, 3.83833942479763e+01, 3.81658028465475e+01, 3.79398305516572e+01, 3.77069517010804e+01, 3.74687357861901e+01, 3.72268172769460e+01, 3.69828656806605e+01, 3.67385564357921e+01, 3.64955430762624e+01, 3.62554310163045e+01, 3.60197532854723e+01, 3.57899485519702e+01, 3.55673417722101e+01, 3.53531277710591e+01, 3.51483579848341e+01, 3.49539304974840e+01, 3.47705833873804e+01, 3.45988912959659e+01, 3.44392650436864e+01, 3.42919540597910e+01, 3.41570513608647e+01, 3.40345008037933e+01, 3.39241063451922e+01, 3.38255430535435e+01, 3.37383696356227e+01, 3.36620422501056e+01, 3.35959293853061e+01, 3.35393275734543e+01, 3.34914777010598e+01, 3.34515816553532e+01, 3.34188190231901e+01, 3.33923635344431e+01, 3.33713989204844e+01, 3.33551338435624e+01, 3.33428155481268e+01, 3.33337418932613e+01, 3.33272714482870e+01, 3.33228313721401e+01, 3.33199228509394e+01, 3.33181239356214e+01, 3.33170896999201e+01, 3.33165497246460e+01, 3.33163030029644e+01, 3.33162104487363e+01, 3.33161852717037e+01, 3.33161815556251e+01, 3.33161812575496e+01, 3.33161757565449e+01, 3.33161384864208e+01, 3.33160014773773e+01, 3.33156362606112e+01, 3.33148369877854e+01, 3.33133062660598e+01, 3.33106442697893e+01, 3.33063416194692e+01, 3.32997764369355e+01, 3.32902158918860e+01, 3.32768224465334e+01, 3.32586648806056e+01, 3.32347340365970e+01, 3.32039630651956e+01, 3.31652517753489e+01, 3.31174945074345e+01, 3.30596107594012e+01, 3.29905776154346e+01, 3.29094628679384e+01, 3.28154576010161e+01, 3.27079069317829e+01, 3.25863375976014e+01, 3.24504811422119e+01, 3.23002915962418e+01, 3.21359567660556e+01, 3.19579025306663e+01, 3.17667898834755e+01, 3.15635048211108e+01, 3.13491415472305e+01, 3.11249797935294e+01, 3.08924573327206e+01, 3.06531389437570e+01, 3.04086831726920e+01, 3.01608082114222e+01, 2.99112581033433e+01, 2.96617703039810e+01, 2.94140454071451e+01, 2.91697196246490e+01, 2.89303404057316e+01, 2.86973454164104e+01, 2.84720449733434e+01, 2.82556079365453e+01, 2.80490510007005e+01, 2.78532312755877e+01, 2.76688420049285e+01, 2.74964112367059e+01, 2.73363032273291e+01, 2.71887223394427e+01, 2.70537191810559e+01, 2.69311987326267e+01, 2.68209302171937e+01, 2.67225584831140e+01, 2.66356166849611e+01, 2.65595400612152e+01, 2.64936806140061e+01, 2.64373224942083e+01, 2.63896978841509e+01, 2.63500031511247e+01, 2.63174150200938e+01, 2.62911064868831e+01, 2.62702621675061e+01, 2.62540927593412e+01, 2.62418482794421e+01, 2.62328297476920e+01, 2.62263990001655e+01, 2.62219863521616e+01, 2.62190958807647e+01, 2.62173081619851e+01, 2.62162803747564e+01, 2.62157437695242e+01, 2.62154985884094e+01, 2.62154066122252e+01, 2.62153815924462e+01, 2.62153778995783e+01, 2.62153776147488e+01, 2.62153725001249e+01, 2.62153378477906e+01, 2.62152104620713e+01, 2.62148708979552e+01, 2.62141277673602e+01, 2.62127045725164e+01, 2.62102295907852e+01, 2.62062292711722e+01, 2.62001255285583e+01, 2.61912372357521e+01, 2.61787861136895e+01, 2.61619071042021e+01, 2.61396631761482e+01, 2.61110643641559e+01, 2.60750906717843e+01, 2.60307182923403e+01, 2.59769484187519e+01, 2.59128377396072e+01, 2.58375295649027e+01, 2.57502844066568e+01, 2.56505087705758e+01, 2.55377809077122e+01, 2.54118723382540e+01, 2.52727640969844e+01, 2.51206568595024e+01, 2.49559743817276e+01, 2.47793600080703e+01, 2.45916663555378e+01, 2.43939386361441e+01, 2.41873924083274e+01, 2.39733868182302e+01, 2.37533945753960e+01, 2.35289699857527e+01, 2.33017163338914e+01, 2.30732537807840e+01, 2.28451887518507e+01, 2.26190855710272e+01, 2.23964408838446e+01, 2.21786612297447e+01, 2.19670439791130e+01, 2.17627617392911e+01, 2.15668502450811e+01, 2.13801996721553e+01, 2.12035492402292e+01, 2.10374849066097e+01, 2.08824398938407e+01, 2.07386977531107e+01, 2.06063976420643e+01, 2.04855414929737e+01, 2.03760027628770e+01, 2.02775364865955e+01, 2.01897903902533e+01, 2.01123168605001e+01, 2.00445855972397e+01, 1.99859968009169e+01, 1.99358947567168e+01, 1.98935816766654e+01, 1.98583316476075e+01, 1.98294045108492e+01, 1.98060594714821e+01, 1.97875682063299e+01, 1.97732272136420e+01, 1.97623691294982e+01, 1.97543727291973e+01, 1.97486713395505e+01, 1.97447594116221e+01, 1.97421970432684e+01, 1.97406122956061e+01, 1.97397012147340e+01, 1.97392255459949e+01, 1.97390082084676e+01, 1.97389266775686e+01, 1.97389044991788e+01, 1.97389012256946e+01, 1.97389009778363e+01, 1.97388965870504e+01, 1.97388668388320e+01, 1.97387574812562e+01, 1.97384659742776e+01, 1.97378280192372e+01, 1.97366062622838e+01, 1.97344816269345e+01, 1.97310476789412e+01, 1.97258083660817e+01, 1.97181794046165e+01, 1.97074935004269e+01, 1.96930094936727e+01, 1.96739253992033e+01, 1.96493951803446e+01, 1.96185489426332e+01, 1.95805160708147e+01, 1.95344506642050e+01, 1.94795584624238e+01, 1.94151243079784e+01, 1.93405390779323e+01, 1.92553249476069e+01, 1.91591578369032e+01, 1.90518859430817e+01, 1.89335433867232e+01, 1.88043581885552e+01, 1.86647540463817e+01, 1.85153456802654e+01, 1.83569278421097e+01, 1.81904584205937e+01, 1.80170363892327e+01, 1.78378756186738e+01, 1.76542757804038e+01, 1.74675916885861e+01, 1.72792024479441e+01, 1.70904816966738e+01, 1.69027700637018e+01, 1.67173507193427e+01, 1.65354286157266e+01, 1.63581137202943e+01, 1.61864082729923e+01, 1.60211978705030e+01, 1.58632460147997e+01, 1.57131916640979e+01, 1.55715492879851e+01, 1.54387109441405e+01, 1.53149499466238e+01, 1.52004257694249e+01, 1.50951899096605e+01, 1.49991925114848e+01, 1.49122896171443e+01, 1.48342509617893e+01, 1.47647682625207e+01, 1.47034639704058e+01, 1.46499004586052e+01, 1.46035896125307e+01, 1.45640027714263e+01, 1.45305809472475e+01, 1.45027452184948e+01, 1.44799071660913e+01, 1.44614791879270e+01, 1.44468845009191e+01, 1.44355666169577e+01, 1.44269980644712e+01, 1.44206881227605e+01, 1.44161893434141e+01, 1.44131026529660e+01, 1.44110808634870e+01, 1.44098304620298e+01, 1.44091116037714e+01, 1.44087362945150e+01, 1.44085648125105e+01, 1.44085004836732e+01, 1.44084829846682e+01, 1.44084804018516e+01, 1.44084802079588e+01, 1.44084767952248e+01, 1.44084536734526e+01, 1.44083686754144e+01, 1.44081421021748e+01, 1.44076462538218e+01, 1.44066966511890e+01, 1.44050453038869e+01, 1.44023763403035e+01, 1.43983042689110e+01, 1.43923750855016e+01, 1.43840703754808e+01, 1.43728144815619e+01, 1.43579847131139e+01, 1.43389244632506e+01, 1.43149589747328e+01, 1.42854133597759e+01, 1.42496323387923e+01, 1.42070010288918e+01, 1.41569659970005e+01, 1.40990557084995e+01, 1.40328994637397e+01, 1.39582439326503e+01, 1.38749664782696e+01, 1.37830846032085e+01, 1.36827610509047e+01, 1.35743043304166e+01, 1.34581646875303e+01, 1.33349257907378e+01, 1.32052926132657e+01, 1.30700761516167e+01, 1.29301757153163e+01, 1.27865595506425e+01, 1.26402445325227e+01, 1.24922755909333e+01, 1.23437054518499e+01, 1.21955751868804e+01, 1.20488959921897e+01, 1.19046325589481e+01, 1.17636883482464e+01, 1.16268930320016e+01, 1.14949922967184e+01, 1.13686401233004e+01, 1.12483935559928e+01, 1.11347098674818e+01, 1.10279459302236e+01, 1.09283595306871e+01, 1.08361123227900e+01, 1.07512741113510e+01, 1.06738281807452e+01, 1.06036774282533e+01, 1.05406511144088e+01, 1.04845120937375e+01, 1.04349644312680e+01, 1.03916613390308e+01, 1.03542133813435e+01, 1.03221968990190e+01, 1.02951625930380e+01, 1.02726441906276e+01, 1.02541670942016e+01, 1.02392568893472e+01, 1.02274475649117e+01, 1.02182892789740e+01, 1.02113554914893e+01, 1.02062492796541e+01, 1.02026086570069e+01, 1.02001107327012e+01, 1.01984745732401e+01, 1.01974626643729e+01, 1.01968809141759e+01, 1.01965771872510e+01, 1.01964384117061e+01, 1.01963863521752e+01, 1.01963721907174e+01, 1.01963701005157e+01, 1.01963699440808e+01, 1.01963671970047e+01, 1.01963485851613e+01, 1.01962801660763e+01, 1.01960977860708e+01, 1.01956986526760e+01, 1.01949342674884e+01, 1.01936050046577e+01, 1.01914565878215e+01, 1.01881786805736e+01, 1.01834057612071e+01, 1.01767203997840e+01, 1.01676589927200e+01, 1.01557199355864e+01, 1.01403741281852e+01, 1.01210776080324e+01, 1.00972860019349e+01, 1.00684703752575e+01, 1.00341339517689e+01, 9.99382908252866e+00, 9.94717377029692e+00, 9.89386701687263e+00, 9.83370226415100e+00, 9.76657825269819e+00, 9.69250672745894e+00, 9.61161657652721e+00, 9.52415418667794e+00, 9.43048002238922e+00, 9.33106166091108e+00, 9.22646371852366e+00, 9.11733525741035e+00, 9.00439535012219e+00, 8.88841749274453e+00, 8.77021350615872e+00, 8.65061746736661e+00, 8.53047009668661e+00, 8.41060391755870e+00, 8.29182942125894e+00, 8.17492241466813e+00, 8.06061269999199e+00, 7.94957421898373e+00, 7.84241677777132e+00, 7.73967944246155e+00, 7.64182565681492e+00, 7.54924008372367e+00, 7.46222711810644e+00, 7.38101096764096e+00, 7.30573715611750e+00, 7.23647527644861e+00, 7.17322280803529e+00, 7.11590981518336e+00, 7.06440435656254e+00, 7.01851845619424e+00, 6.97801450975240e+00, 6.94261202205894e+00, 6.91199458937585e+00, 6.88581705134923e+00, 6.86371274131817e+00, 6.84530076034591e+00, 6.83019319091554e+00, 6.81800215266023e+00, 6.80834658718565e+00, 6.80085864462743e+00, 6.79518953366539e+00, 6.79101469157096e+00, 6.78803813323697e+00, 6.78599584906509e+00, 6.78465814129040e+00, 6.78383081620704e+00, 6.78335518447452e+00, 6.78310686127708e+00, 6.78299340022746e+00, 6.78295083704963e+00, 6.78293925883128e+00, 6.78293754991045e+00, 6.78293742196346e+00, 6.78293517451057e+00, 6.78291994768795e+00, 6.78286397229761e+00, 6.78271476261467e+00, 6.78238822166657e+00, 6.78176285972935e+00, 6.78067535960940e+00, 6.77891769596346e+00, 6.77623598529189e+00, 6.77233120734149e+00, 6.76686189614806e+00, 6.75944884847257e+00, 6.74968183791180e+00, 6.73712825419615e+00, 6.72134350992215e+00, 6.70188297336719e+00, 6.67831509975544e+00, 6.65023534946546e+00, 6.61728040640413e+00, 6.57914214994983e+00, 6.53558079626504e+00, 6.48643661532947e+00, 6.43163965310035e+00, 6.37121694590257e+00, 6.30529680602020e+00, 6.23410988028180e+00, 6.15798683133134e+00, 6.07735265601018e+00, 5.99271782664673e+00, 5.90466660747776e+00, 5.81384304766934e+00, 5.72093527256175e+00, 5.62665877547415e+00, 5.53173944622278e+00, 5.43689705613551e+00, 5.34282985447165e+00, 5.25020082455587e+00, 5.15962601075754e+00, 5.07166517371821e+00, 4.98681487588006e+00, 4.90550395621016e+00, 4.82809123302216e+00, 4.75486518405744e+00, 4.68604529636141e+00, 4.62178475398269e+00, 4.56217413511335e+00, 4.50724581603301e+00, 4.45697882032177e+00, 4.41130390154683e+00, 4.37010870001772e+00, 4.33324286436538e+00, 4.30052307300693e+00, 4.27173792660001e+00, 4.24665270904227e+00, 4.22501403101027e+00, 4.20655437676736e+00, 4.19099657289645e+00, 4.17805818809060e+00, 4.16745585789262e+00, 4.15890950932639e+00, 4.15214643990490e+00, 4.14690518579474e+00, 4.14293909715440e+00, 4.14001952679236e+00, 4.13793853286261e+00, 4.13651099832784e+00, 4.13557607971444e+00, 4.13499791489793e+00, 4.13466554325094e+00, 4.13449201985544e+00, 4.13441273662364e+00, 4.13438299491462e+00, 4.13437490446452e+00, 4.13437371033162e+00, 4.13437362068967e+00, 4.13437204291622e+00, 4.13436135327743e+00, 4.13432205716067e+00, 4.13421730911072e+00, 4.13398807548634e+00, 4.13354908452954e+00, 4.13278573458816e+00, 4.13155211661714e+00, 4.12967028967818e+00, 4.12693092645979e+00, 4.12309541790526e+00, 4.11789948981758e+00, 4.11105833808541e+00, 4.10227323201083e+00, 4.09123946744794e+00, 4.07765547505025e+00, 4.06123280771052e+00, 4.04170665096360e+00, 4.01884642799030e+00, 3.99246601516478e+00, 3.96243305317023e+00, 3.92867683988656e+00, 3.89119432962995e+00, 3.85005384067949e+00, 3.80539618701323e+00, 3.75743309399238e+00, 3.70644292041764e+00, 3.65276387681645e+00, 3.59678508631092e+00, 3.53893596472553e+00, 3.47967448803613e+00, 3.41947495954035e+00, 3.35881588351249e+00, 3.29816849973401e+00, 3.23798644252138e+00, 3.17869687081578e+00, 3.12069328645646e+00, 3.06433012955247e+00, 3.00991912448038e+00, 2.95772725579267e+00, 2.90797618481361e+00, 2.86084287591469e+00, 2.81646118445417e+00, 2.77492416216085e+00, 2.73628685531470e+00, 2.70056940118621e+00, 2.66776026401718e+00, 2.63781948932028e+00, 2.61068189134788e+00, 2.58626012105947e+00, 2.56444758943531e+00, 2.54512124281618e+00, 2.52814420284326e+00, 2.51336829361140e+00, 2.50063648315349e+00, 2.48978526583323e+00, 2.48064700727670e+00, 2.47305226489323e+00, 2.46683208573571e+00, 2.46182027047111e+00, 2.45785557872819e+00, 2.45478383827550e+00, 2.45245990956929e+00, 2.45074944931070e+00, 2.44953041267459e+00, 2.44869423443205e+00, 2.44814663451957e+00, 2.44780800352319e+00, 2.44761333744197e+00, 2.44751170800863e+00, 2.44746527354982e+00, 2.44744785452177e+00, 2.44744311613704e+00, 2.44744241676187e+00, 2.44744236403713e+00, 2.44744143305884e+00, 2.44743512555078e+00, 2.44741193858571e+00, 2.44735013143168e+00, 2.44721487195365e+00, 2.44695584907193e+00, 2.44650545366952e+00, 2.44577762264394e+00, 2.44466743149044e+00, 2.44305150877189e+00, 2.44078932964613e+00, 2.43772542376368e+00, 2.43369250401777e+00, 2.42851548608578e+00, 2.42201632463446e+00, 2.41401954188411e+00, 2.40435827076494e+00, 2.39288058239667e+00, 2.37945582147830e+00, 2.36398063943582e+00, 2.34638439974501e+00, 2.32663363755532e+00, 2.30473528937941e+00, 2.28073846814145e+00, 2.25473464099110e+00, 2.22685616550140e+00, 2.19727324515209e+00, 2.16618946695009e+00, 2.13383617234588e+00, 2.10046597856982e+00, 2.06634580531713e+00, 2.03174976915781e+00, 1.99695228668205e+00, 1.96222168196829e+00, 1.92781453145564e+00, 1.89397090767527e+00, 1.86091061029698e+00, 1.82883040515090e+00, 1.79790223411009e+00, 1.76827231381717e+00, 1.74006101022558e+00, 1.71336335831988e+00, 1.68825009063836e+00, 1.66476904218396e+00, 1.64294681056557e+00, 1.62279056638774e+00, 1.60428992784573e+00, 1.58741883335596e+00, 1.57213736538840e+00, 1.55839349636455e+00, 1.54612474276306e+00, 1.53525972594706e+00, 1.52571964743892e+00, 1.51741969235734e+00, 1.51027037759164e+00, 1.50417886123414e+00, 1.49905022715400e+00, 1.49478875379771e+00, 1.49129916986411e+00, 1.48848789202748e+00, 1.48626423204109e+00, 1.48454155305554e+00, 1.48323834853553e+00, 1.48227921240862e+00, 1.48159566656122e+00, 1.48112681187671e+00, 1.48081977183083e+00, 1.48062990312121e+00, 1.48052075557163e+00, 1.48046377306518e+00, 1.48043773782415e+00, 1.48042797119323e+00, 1.48042531444101e+00, 1.48042492231025e+00, 1.48042489258727e+00, 1.48042436562508e+00, 1.48042079538290e+00, 1.48040767085892e+00, 1.48037268618732e+00, 1.48029612557919e+00, 1.48014951239643e+00, 1.47989458069563e+00, 1.47948262343763e+00, 1.47885426651843e+00, 1.47793971116238e+00, 1.47665947816128e+00, 1.47492567502448e+00, 1.47264379052999e+00, 1.46971499998204e+00, 1.46603893878358e+00, 1.46151687255481e+00, 1.45605516065435e+00, 1.44956887911485e+00, 1.44198544191141e+00, 1.43324803973792e+00, 1.42331870657692e+00, 1.41218082916225e+00, 1.39984093456990e+00, 1.38632962656743e+00, 1.37170159000766e+00, 1.35603464061833e+00, 1.33942785968819e+00, 1.32199891327512e+00, 1.30388070760695e+00, 1.28521757117724e+00, 1.26616117614958e+00, 1.24686641563727e+00, 1.22748743993940e+00, 1.20817402653281e+00, 1.18906841959935e+00, 1.17030272992212e+00, 1.15199693999318e+00, 1.13425751647143e+00, 1.11717659605508e+00, 1.10083168348940e+00, 1.08528578264958e+00, 1.07058787311858e+00, 1.05677364425301e+00, 1.04386640468555e+00, 1.03187809561426e+00, 1.02081034919033e+00, 1.01065554718760e+00, 1.00139784861903e+00, 9.93014167136195e-01, 9.85475089341755e-01, 9.78745733289308e-01, 9.72786552387610e-01, 9.67554093764967e-01, 9.63001722067226e-01, 9.59080319890617e-01, 9.55738974836898e-01, 9.52925660782213e-01, 9.50587917638046e-01, 9.48673529934815e-01, 9.47131200274129e-01, 9.45911209396201e-01, 9.44966050626938e-01, 9.44251023140147e-01, 9.43724766102477e-01, 9.43349714619842e-01, 9.43092458648358e-01, 9.42923987736513e-01, 9.42819807570856e-01, 9.42759918622556e-01, 9.42728652442064e-01, 9.42714366950838e-01, 9.42709008015989e-01, 9.42707550260117e-01, 9.42707335098568e-01, 9.42707318686139e-01, 9.42707026342774e-01, 9.42705045675915e-01, 9.42697764570269e-01, 9.42678356086873e-01, 9.42635882470420e-01, 9.42554545681389e-01, 9.42413116804211e-01, 9.42184574377414e-01, 9.41835978878395e-01, 9.41328606896278e-01, 9.40618363449182e-01, 9.39656483923599e-01, 9.38390527780982e-01, 9.36765654227884e-01, 9.34726155542052e-01, 9.32217207174858e-01, 9.29186776098774e-01, 9.25587611626973e-01, 9.21379227995203e-01, 9.16529777477507e-01, 9.11017708740282e-01, 9.04833109136511e-01, 8.97978642538355e-01, 8.90470015880283e-01, 8.82335936429619e-01, 8.73617555386097e-01, 8.64367428371615e-01, 8.54648055974636e-01, 8.44530094222463e-01, 8.34090342893333e-01, 8.23409627351253e-01, 8.12570686890496e-01, 8.01656170561671e-01, 7.90746822339507e-01, 7.79919914135853e-01, 7.69247960550793e-01, 7.58797726067975e-01, 7.48629515677038e-01, 7.38796724892921e-01, 7.29345615281794e-01, 7.20315276667925e-01, 7.11737736470539e-01, 7.03638179123566e-01, 6.96035243205486e-01, 6.88941369762354e-01, 6.82363181519044e-01, 6.76301878617842e-01, 6.70753641784145e-01, 6.65710038158693e-01, 6.61158428362882e-01, 6.57082375684674e-01, 6.53462059661727e-01, 6.50274696897995e-01, 6.47494971800661e-01, 6.45095479181819e-01, 6.43047179448452e-01, 6.41319865517332e-01, 6.39882638746705e-01, 6.38704389196207e-01, 6.37754273533375e-01, 6.37002182039794e-01, 6.36419184575338e-01, 6.35977944179523e-01, 6.35653086357976e-01, 6.35421512124481e-01, 6.35262643613466e-01, 6.35158592558021e-01, 6.35094244101076e-01, 6.35057251168526e-01, 6.35037937825579e-01, 6.35029113468541e-01, 6.35025803158705e-01, 6.35024902675227e-01, 6.35024769765778e-01, 6.35024759555961e-01, 6.35024576757566e-01, 6.35023338272291e-01, 6.35018785480696e-01, 6.35006649496326e-01, 6.34980090630037e-01, 6.34929228904547e-01, 6.34840785194612e-01, 6.34697850752929e-01, 6.34479800723184e-01, 6.34162364580142e-01, 6.33717863055757e-01, 6.33115616780763e-01, 6.32322526400920e-01, 6.31303817213613e-01, 6.30023933474497e-01, 6.28447558712821e-01, 6.26540729183972e-01, 6.24271998737626e-01, 6.21613605843599e-01, 6.18542588343743e-01, 6.15041789676020e-01, 6.11100702604894e-01, 6.06716103235094e-01, 6.01892439092261e-01, 5.96641949540029e-01, 5.90984513452149e-01, 5.84947236187098e-01, 5.78563803717401e-01, 5.71873644610100e-01, 5.64920949225275e-01, 5.57753599389009e-01, 5.50422060961289e-01, 5.42978286818965e-01, 5.35474669904644e-01, 5.27963076466136e-01, 5.20493979721401e-01, 5.13115705019295e-01, 5.05873789889449e-01, 4.98810456578006e-01, 4.91964190808353e-01, 4.85369418385240e-01, 4.79056270517397e-01, 4.73050428951709e-01, 4.67373042797611e-01, 4.62040709949470e-01, 4.57065517063754e-01, 4.52455132980297e-01, 4.48212951232065e-01, 4.44338277854918e-01, 4.40826561097872e-01, 4.37669659862223e-01, 4.34856147775889e-01, 4.32371649734762e-01, 4.30199207509908e-01, 4.28319670618237e-01, 4.26712108081979e-01, 4.25354235972399e-01, 4.24222854775962e-01, 4.23294289690737e-01, 4.22544826032549e-01, 4.21951131092314e-01, 4.21490653140282e-01, 4.21141987915455e-01, 4.20885202958821e-01, 4.20702110610442e-01, 4.20576481431091e-01, 4.20494191224563e-01, 4.20443296689122e-01, 4.20414036936862e-01, 4.20398760579197e-01, 4.20391780655493e-01, 4.20389162240333e-01, 4.20388449967483e-01, 4.20388344837477e-01, 4.20388336699986e-01, 4.20388190202502e-01, 4.20387197660631e-01, 4.20383548971955e-01, 4.20373822909269e-01, 4.20352537667824e-01, 4.20311773917859e-01, 4.20240885519739e-01, 4.20126311301900e-01, 4.19951499255426e-01, 4.19696952100992e-01, 4.19340400344794e-01, 4.18857105569380e-01, 4.18220292729494e-01, 4.17401705623448e-01, 4.16372274544749e-01, 4.15102879552784e-01, 4.13565187119872e-01, 4.11732532515806e-01, 4.09580815669770e-01, 4.07089374936398e-01, 4.04241801683121e-01, 4.01026659296683e-01, 3.97438073262940e-01, 3.93476164344854e-01, 3.89147304231116e-01, 3.84464181774327e-01, 3.79445677340073e-01, 3.74116552045592e-01, 3.68506967041142e-01, 3.62651854908869e-01, 3.56590170376975e-01, 3.50364050764256e-01, 3.44017917978938e-01, 3.37597553720919e-01, 3.31149178064855e-01, 3.24718559108568e-01, 3.18350178092381e-01, 3.12086470518629e-01, 3.05967159484332e-01, 3.00028692841110e-01, 2.94303791090528e-01, 2.88821108316611e-01, 2.83605004174881e-01, 2.78675421222253e-01, 2.74047858875022e-01, 2.69733433155011e-01, 2.65739010179047e-01, 2.62067401034017e-01, 2.58717606148167e-01, 2.55685098349225e-01, 2.52962135283312e-01, 2.50538093532668e-01, 2.48399818407208e-01, 2.46531984811938e-01, 2.44917465675520e-01, 2.43537705078765e-01, 2.42373093419566e-01, 2.41403341715113e-01, 2.40607851548219e-01, 2.39966076320762e-01, 2.39457868521953e-01, 2.39063806800324e-01, 2.38765495893261e-01, 2.38545832046412e-01, 2.38389226547653e-01, 2.38281780468029e-01, 2.38211404661328e-01, 2.38167880494537e-01, 2.38142858589939e-01, 2.38129794949920e-01, 2.38123826076134e-01, 2.38121586947322e-01, 2.38120977850166e-01, 2.38120887948710e-01, 2.38120880933558e-01, 2.38120753912433e-01, 2.38119893325901e-01, 2.38116729723323e-01, 2.38108296755307e-01, 2.38089841588383e-01, 2.38054498415558e-01, 2.37993038469213e-01, 2.37893708573516e-01, 2.37742168669051e-01, 2.37521536824815e-01, 2.37212547917800e-01, 2.36793829302222e-01, 2.36242293321979e-01, 2.35533642407397e-01, 2.34642977776723e-01, 2.33545497568727e-01, 2.32217264800695e-01, 2.30636020213801e-01, 2.28782010254562e-01, 2.26638796613379e-01, 2.24194011367622e-01, 2.21440021267654e-01, 2.18374466362008e-01, 2.15000642120131e-01, 2.11327700424228e-01, 2.07370653000829e-01, 2.03150170587719e-01, 1.98692181755490e-01, 1.94027286081763e-01, 1.89190006513847e-01, 1.84217914471766e-01, 1.79150667846230e-01, 1.74029006001895e-01, 1.68893746889495e-01, 1.63784829347847e-01, 1.58740438860356e-01, 1.53796247903131e-01, 1.48984793272462e-01, 1.44335003225059e-01, 1.39871877751183e-01, 1.35616316605301e-01, 1.31585082466501e-01, 1.27790881205101e-01, 1.24242537875070e-01, 1.20945245695445e-01, 1.17900865707718e-01, 1.15108256657017e-01, 1.12563617532182e-01, 1.10260828697371e-01, 1.08191781265248e-01, 1.06346687975091e-01, 1.04714372091969e-01, 1.03282533559529e-01, 1.02037993713587e-01, 1.00966921252208e-01, 1.00055042869315e-01, 9.92878420425782e-02, 9.86507490046122e-02, 9.81293240294718e-02, 9.77094349614802e-02, 9.73774285429769e-02, 9.71202937099420e-02, 9.69258137620557e-02, 9.67827033082216e-02, 9.66807252446458e-02, 9.66107828139897e-02, 9.65649820494860e-02, 9.65366606170619e-02, 9.65203801735370e-02, 9.65118807696399e-02, 9.65079974256682e-02, 9.65065406660841e-02, 9.65061443937429e-02, 9.65060859048350e-02, 9.65060813020031e-02, 9.65059974627351e-02, 9.65054294402563e-02, 9.65033413438832e-02, 9.64977753558145e-02, 9.64855948864241e-02, 9.64622699738058e-02, 9.64217144629771e-02, 9.63561839877798e-02, 9.62562431361320e-02, 9.61108092735543e-02, 9.59072791614659e-02, 9.56317425587301e-02, 9.52692842816560e-02, 9.48043726256245e-02, 9.42213276237993e-02, 9.35048574651276e-02, 9.26406457924145e-02, 9.16159669741461e-02, 9.04203013423981e-02, 8.90459184381882e-02, 8.74883941463338e-02, 8.57470277908519e-02, 8.38251281924798e-02, 8.17301434929050e-02, 7.94736180389748e-02, 7.70709702640104e-02, 7.45410974465699e-02, 7.19058253753673e-02, 6.91892320877514e-02, 6.64168838028729e-02, 6.36150269722577e-02, 6.08097823951965e-02, 5.80263854257197e-02, 5.52885107307901e-02, 5.26177115828190e-02, 5.00329933287063e-02, 4.75505296672480e-02, 4.51835198463684e-02, 4.29421758343764e-02, 4.08338215979143e-02, 3.88630821727034e-02, 3.70321382497067e-02, 3.53410222565245e-02, 3.37879339397702e-02, 3.23695567012550e-02, 3.10813598570547e-02, 2.99178760904013e-02, 2.88729472803908e-02, 2.79399353536220e-02, 2.71118976793687e-02, 2.63817287549557e-02, 2.57422715184177e-02, 2.51864026347235e-02, 2.47070966109675e-02, 2.42974736906731e-02, 2.39508362425553e-02, 2.36606978680385e-02, 2.34208087684138e-02, 2.32251800951744e-02, 2.30681091073990e-02, 2.29442060324061e-02, 2.28484226242003e-02, 2.27760815931504e-02, 2.27229053946942e-02, 2.26850423599956e-02, 2.26590878658399e-02, 2.26420981920654e-02, 2.26315949030607e-02, 2.26255579915389e-02, 2.26224065963976e-02, 2.26209667930846e-02, 2.26204266885475e-02, 2.26202797684678e-02, 2.26202580834354e-02, 2.26202563618187e-02, 2.26202248108791e-02, 2.26200110494688e-02, 2.26192252521713e-02, 2.26171306994019e-02, 2.26125473004117e-02, 2.26037713764455e-02, 2.25885157074543e-02, 2.25638738877336e-02, 2.25263130399256e-02, 2.24716987965579e-02, 2.23953560445054e-02, 2.22921680959587e-02, 2.21567155749725e-02, 2.19834543070050e-02, 2.17669288658300e-02, 2.15020152640477e-02, 2.11841827948973e-02, 2.08097615897032e-02, 2.03761995004612e-02, 1.98822899563993e-02, 1.93283519623151e-02, 1.87163447820495e-02, 1.80499032527095e-02, 1.73342850020788e-02, 1.65762276793657e-02, 1.57837219621584e-02, 1.49657136697385e-02, 1.41317548263373e-02, 1.32916281011219e-02, 1.24549710752575e-02, 1.16309259918026e-02, 1.08278371924655e-02, 1.00530128910208e-02, 9.31256111168466e-03, 8.61130248219411e-03, 7.95275600556522e-03, 7.33918863090111e-03, 6.77171581730818e-03, 6.25043846983431e-03, 5.77460150818609e-03, 5.34276060332672e-03, 4.95294585440575e-03, 4.60281390702733e-03, 4.28978280483250e-03, 4.01114640349221e-03, 3.76416728308726e-03, 3.54614871307035e-03, 3.35448736820609e-03, 3.18670921039654e-03, 3.04049130042472e-03, 2.91367235841022e-03, 2.80425473683010e-03, 2.71040019074320e-03, 2.63042149794757e-03, 2.56277164410079e-03, 2.50603196855792e-03, 2.45890037009389e-03, 2.42018039985226e-03, 2.38877181696751e-03, 2.36366295133195e-03, 2.34392500570380e-03, 2.32870823895581e-03, 2.31723980525800e-03, 2.30882288510514e-03, 2.30283664052422e-03, 2.29873646239298e-03, 2.29605395976762e-03, 2.29439616850868e-03, 2.29344352874250e-03, 2.29294628836385e-03, 2.29271912330375e-03, 2.29263391041457e-03, 2.29261073086207e-03, 2.29260730962011e-03, 2.29260703552114e-03, 2.29260198084802e-03, 2.29256773501874e-03, 2.29244184730959e-03, 2.29210630471499e-03, 2.29137211907471e-03, 2.28996660641201e-03, 2.28752409920929e-03, 2.28358089525926e-03, 2.27757530348072e-03, 2.26885366885909e-03, 2.25668320226191e-03, 2.24027226637702e-03, 2.21879843440576e-03, 2.19144411997974e-03, 2.15743887577702e-03, 2.11610661141748e-03, 2.06691506917167e-03, 2.00952404068033e-03, 1.94382816459893e-03, 1.86998987416587e-03, 1.78845830763663e-03, 1.69997083398954e-03, 1.60553527059903e-03, 1.50639275976516e-03, 1.40396338867532e-03, 1.29977866244246e-03, 1.19540652396211e-03, 1.09237545865426e-03, 9.92104179147191e-04, 8.95842490621135e-04, 8.04627414098036e-04, 7.19256797992784e-04, 6.40280758261724e-04, 5.68009545233394e-04, 5.02534962407708e-04, 4.43761406963006e-04, 3.91442152903210e-04, 3.45216797533965e-04, 3.04646754713644e-04, 2.69246930614031e-04, 2.38512762813059e-04, 2.11942337635662e-04, 1.89053412704254e-04, 1.69395239886683e-04, 1.52555418658551e-04, 1.38162573669831e-04, 1.25886118183871e-04, 1.15434423570093e-04, 1.06552331963883e-04, 9.90183992115590e-05, 9.26418783464555e-05, 8.72593848413317e-05, 8.27313205959967e-05, 7.89382883610301e-05, 7.57777726363429e-05, 7.31612780029423e-05, 7.10119766154094e-05, 6.92628105198878e-05, 6.78549582531218e-05, 6.67365901350081e-05, 6.58618653659460e-05, 6.51901444257419e-05, 6.46853952943616e-05, 6.43157677373551e-05, 6.40533015427991e-05, 6.38737294238467e-05, 6.37563335601499e-05, 6.36838171956811e-05, 6.36421580182400e-05, 6.36204170092469e-05, 6.36104853839470e-05, 6.36067600154888e-05, 6.36057466465445e-05, 6.36055970844041e-05, 6.36055849980721e-05, 6.36053607294684e-05, 6.36038413119877e-05, 6.35982560481807e-05, 6.35833698012789e-05, 6.35508019399822e-05, 6.34884701185001e-05, 6.33801981564375e-05, 6.32055334719439e-05, 6.29398255824754e-05, 6.25546210955008e-05, 6.20184287853931e-05, 6.12978978438150e-05, 6.03594288648424e-05, 5.91711984654353e-05, 5.77055215487121e-05, 5.59414026781724e-05, 5.38670465815017e-05, 5.14820244461589e-05, 4.87987545341184e-05, 4.58429844769626e-05, 4.26530872204012e-05, 3.92782096554356e-05, 3.57755988673603e-05, 3.22076451482189e-05, 2.86391079029613e-05, 2.51345082702414e-05, 2.17550321224280e-05, 1.85540899479975e-05, 1.55713639315991e-05, 1.28266760626161e-05, 1.03169269323189e-05, 8.02019780365981e-06, 5.90776522102487e-06, 3.95632960703551e-06, 2.13878162469427e-06, 0.00000000000000e+00, 1.36572476350239e-06, 1.63004831676972e-06, 1.51259980185614e-06, 1.22711087158223e-06, 9.15286612723776e-07, 6.85188751958776e-07, 5.74942703224224e-07, 4.53438126157835e-07, 4.35611319460349e-07, 7.20577354229590e-07, 9.43077850479312e-07, 1.01410135044944e-06, 9.33420245985417e-07, 7.39004015315459e-07, 4.86017398770855e-07, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 7.10542735760100e-15, 7.10542735760100e-15, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 7.10542735760100e-15, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 7.10542735760100e-15, 0.00000000000000e+00, 7.10542735760100e-15, 7.10542735760100e-15, 7.10542735760100e-15, 0.00000000000000e+00, 7.10542735760100e-15],
[ 4.90000000000000e+01, 4.89999999549600e+01, 4.89999985622344e+01, 4.89999891262981e+01, 4.89999544385693e+01, 4.89998619720638e+01, 4.89996596024653e+01, 4.89992720063213e+01, 4.89985978664960e+01, 4.89975079903458e+01, 4.89958444181523e+01, 4.89934205692264e+01, 4.89900224419667e+01, 4.89854108533222e+01, 4.89793246738483e+01, 4.89714849880149e+01, 4.89616000865341e+01, 4.89493711787295e+01, 4.89344986984009e+01, 4.89166890657551e+01, 4.88956617598148e+01, 4.88711565490188e+01, 4.88429407212008e+01, 4.88108161467984e+01, 4.87746260006813e+01, 4.87342609590204e+01, 4.86896646798082e+01, 4.86408383715880e+01, 4.85878442578117e+01, 4.85308077571636e+01, 4.84699182256296e+01, 4.84054281451563e+01, 4.83376506957801e+01, 4.82669557104833e+01, 4.81937640804435e+01, 4.81185407472803e+01, 4.80417864825169e+01, 4.79640287073680e+01, 4.78858116439830e+01, 4.78076861100053e+01, 4.77301992712919e+01, 4.76538846542129e+01, 4.75792526919028e+01, 4.75067820418215e+01, 4.74369118689384e+01, 4.73700352435163e+01, 4.73064937579146e+01, 4.72465734253410e+01, 4.71905018864144e+01, 4.71384469173247e+01, 4.70905162062087e+01, 4.70467583415714e+01, 4.70071649374174e+01, 4.69716738034011e+01, 4.69401730540394e+01, 4.69125060383058e+01, 4.68884769594363e+01, 4.68678570444560e+01, 4.68503911139382e+01, 4.68358043952007e+01, 4.68238094170005e+01, 4.68141128213863e+01, 4.68064219292623e+01, 4.68004509009169e+01, 4.67959263416416e+01, 4.67925922158075e+01, 4.67902139503412e+01, 4.67885816301921e+01, 4.67875122135935e+01, 4.67868507229775e+01, 4.67864703974019e+01, 4.67862718232666e+01, 4.67861810908413e+01, 4.67861470536101e+01, 4.67861377946234e+01, 4.67861364280151e+01, 4.67861362487449e+01, 4.67861320719644e+01, 4.67861037736427e+01, 4.67859997457290e+01, 4.67857224432808e+01, 4.67851155623694e+01, 4.67839532695834e+01, 4.67819318974185e+01, 4.67786644596801e+01, 4.67736782722359e+01, 4.67664158876514e+01, 4.67562394679627e+01, 4.67424386287466e+01, 4.67242416906426e+01, 4.67008301727248e+01, 4.66713562569641e+01, 4.66349628458417e+01, 4.65908057273210e+01, 4.65380772543451e+01, 4.64760308420257e+01, 4.64040054886615e+01, 4.63214494435065e+01, 4.62279420853953e+01, 4.61232130560995e+01, 4.60071577266675e+01, 4.58798481782856e+01, 4.57415390591459e+01, 4.55926679318570e+01, 4.54338500342713e+01, 4.52658677086473e+01, 4.50896550691443e+01, 4.49062787341047e+01, 4.47169156142402e+01, 4.45228288040468e+01, 4.43253425752670e+01, 4.41258173407277e+01, 4.39256252796567e+01, 4.37261271301052e+01, 4.35286504927428e+01, 4.33344698726028e+01, 4.31447886157458e+01, 4.29607228675572e+01, 4.27832876719381e+01, 4.26133853279890e+01, 4.24517961090147e+01, 4.22991714210586e+01, 4.21560294353864e+01, 4.20227531773894e+01, 4.18995910013899e+01, 4.17866593340034e+01, 4.16839475323769e+01, 4.15913246786776e+01, 4.15085481167749e+01, 4.14352735276352e+01, 4.13710663327730e+01, 4.13154142070497e+01, 4.12677404713161e+01, 4.12274181213136e+01, 4.11937842325931e+01, 4.11661544636101e+01, 4.11438373628786e+01, 4.11261481736692e+01, 4.11124218238056e+01, 4.11020247909500e+01, 4.10943655472275e+01, 4.10889033122726e+01, 4.10851548811455e+01, 4.10826993425106e+01, 4.10811805615725e+01, 4.10803073692868e+01, 4.10798514714379e+01, 4.10796431650264e+01, 4.10795650216277e+01, 4.10795437646839e+01, 4.10795406272020e+01, 4.10795403502220e+01, 4.10795349229363e+01, 4.10794981522672e+01, 4.10793629792458e+01, 4.10790026564306e+01, 4.10782140928323e+01, 4.10767038764784e+01, 4.10740775264463e+01, 4.10698324555946e+01, 4.10633550436313e+01, 4.10539221264466e+01, 4.10407071000864e+01, 4.10227907145632e+01, 4.09991764927091e+01, 4.09688105526402e+01, 4.09306054413271e+01, 4.08834674060448e+01, 4.08263263477418e+01, 4.07581675259723e+01, 4.06780639314657e+01, 4.05852081233437e+01, 4.04789422571256e+01, 4.03587850194606e+01, 4.02244542462752e+01, 4.00758841399178e+01, 3.99132362209143e+01, 3.97369034479108e+01, 3.95475073024868e+01, 3.93458880375130e+01, 3.91330886867170e+01, 3.89103337751537e+01, 3.86790039010426e+01, 3.84406074424781e+01, 3.81967505761711e+01, 3.79491066180431e+01, 3.76993854733954e+01, 3.74493037858820e+01, 3.72005562441532e+01, 3.69547884503104e+01, 3.67135717506522e+01, 3.64783804372691e+01, 3.62505717140462e+01, 3.60313687645155e+01, 3.58218471627763e+01, 3.56229247472651e+01, 3.54353549509266e+01, 3.52597234689449e+01, 3.50964480589631e+01, 3.49457812134422e+01, 3.48078154177033e+01, 3.46824907041501e+01, 3.45696042250286e+01, 3.44688215846752e+01, 3.43796896905648e+01, 3.43016508955147e+01, 3.42340582081211e+01, 3.41761913437947e+01, 3.41272733751652e+01, 3.40864877200039e+01, 3.40529951799639e+01, 3.40259507178493e+01, 3.40045196385663e+01, 3.39878928232044e+01, 3.39753006603256e+01, 3.39660253263948e+01, 3.39594110903411e+01, 3.39548723563664e+01, 3.39518992139368e+01, 3.39500603327495e+01, 3.39490031205750e+01, 3.39484511494918e+01, 3.39481989468216e+01, 3.39481043364847e+01, 3.39480786001406e+01, 3.39480748015087e+01, 3.39480744968115e+01, 3.39480688736013e+01, 3.39480307755166e+01, 3.39478907228062e+01, 3.39475173927689e+01, 3.39467003643865e+01, 3.39451356391784e+01, 3.39424145122317e+01, 3.39380162954884e+01, 3.39313053122114e+01, 3.39215324857238e+01, 3.39078417349529e+01, 3.38892812618900e+01, 3.38648196703286e+01, 3.38333666911579e+01, 3.37937981093010e+01, 3.37449842961412e+01, 3.36858215571779e+01, 3.36152653189371e+01, 3.35323640155910e+01, 3.34362924095158e+01, 3.33263830063633e+01, 3.32021542176493e+01, 3.30633339923257e+01, 3.29098777880671e+01, 3.27419799811591e+01, 3.25600781114915e+01, 3.23648497088985e+01, 3.21572018243247e+01, 3.19382537637066e+01, 3.17093138610666e+01, 3.14718513986700e+01, 3.12274649612904e+01, 3.09778485850179e+01, 3.07247570292796e+01, 3.04699713790713e+01, 3.02152659995522e+01, 2.99623776495182e+01, 2.97129773448323e+01, 2.94686453710130e+01, 2.92308496881046e+01, 2.90009278517070e+01, 2.87800724845978e+01, 2.85693202633955e+01, 2.83695443251330e+01, 2.81814499445781e+01, 2.80055732844861e+01, 2.78422829808185e+01, 2.76917842972345e+01, 2.75541255703533e+01, 2.74292066692020e+01, 2.73167892058772e+01, 2.72165082548256e+01, 2.71278853596026e+01, 2.70503426233489e+01, 2.69832176886799e+01, 2.69257794121593e+01, 2.68772440277481e+01, 2.68367915739516e+01, 2.68035823334930e+01, 2.67767730058145e+01, 2.67555323056694e+01, 2.67390556598055e+01, 2.67265786622580e+01, 2.67173889504917e+01, 2.67108361819987e+01, 2.67063398252154e+01, 2.67033945296356e+01, 2.67015729062288e+01, 2.67005256278854e+01, 2.66999788467319e+01, 2.66997290161997e+01, 2.66996352958658e+01, 2.66996098016365e+01, 2.66996060387407e+01, 2.66996057485099e+01, 2.66996005368975e+01, 2.66995652274517e+01, 2.66994354261281e+01, 2.66990894229522e+01, 2.66983322009068e+01, 2.66968820203025e+01, 2.66943601134017e+01, 2.66902839590856e+01, 2.66840645320196e+01, 2.66750078334949e+01, 2.66623209096147e+01, 2.66451224443548e+01, 2.66224578785665e+01, 2.65933188509416e+01, 2.65566665853470e+01, 2.65114586656216e+01, 2.64566784519747e+01, 2.63913662136466e+01, 2.63146508941237e+01, 2.62257813029485e+01, 2.61241554569581e+01, 2.60093467866692e+01, 2.58811259899826e+01, 2.57394774597209e+01, 2.55846094318060e+01, 2.54169572882276e+01, 2.52371797871915e+01, 2.50461483587534e+01, 2.48449299685137e+01, 2.46347643813679e+01, 2.44170369189431e+01, 2.41932479714061e+01, 2.39649805826434e+01, 2.37338673804862e+01, 2.35015579912807e+01, 2.32696878934713e+01, 2.30398494635397e+01, 2.28135657781449e+01, 2.25922675738946e+01, 2.23772736324241e+01, 2.21697747458253e+01, 2.19708213160859e+01, 2.17813145454682e+01, 2.16020010819122e+01, 2.14334708990736e+01, 2.12761581214669e+01, 2.11303444577951e+01, 2.09961648834232e+01, 2.08736152156889e+01, 2.07625612493591e+01, 2.06627491575167e+01, 2.05738169078804e+01, 2.04953064886485e+01, 2.04266767752223e+01, 2.03673168952490e+01, 2.03165599619646e+01, 2.02736970444514e+01, 2.02379912295258e+01, 2.02086916063519e+01, 2.01850469753877e+01, 2.01663190524119e+01, 2.01517949108678e+01, 2.01407983861216e+01, 2.01327001573571e+01, 2.01269262296705e+01, 2.01229645621399e+01, 2.01203696274696e+01, 2.01187647440001e+01, 2.01178420888280e+01, 2.01173603777804e+01, 2.01171402795911e+01, 2.01170577130915e+01, 2.01170352529951e+01, 2.01170319379317e+01, 2.01170316869251e+01, 2.01170272403683e+01, 2.01169971142941e+01, 2.01168863676934e+01, 2.01165911581839e+01, 2.01159451007504e+01, 2.01147078286806e+01, 2.01125562184711e+01, 2.01090786881954e+01, 2.01037729179753e+01, 2.00960472698181e+01, 2.00852260995182e+01, 2.00705590528071e+01, 2.00512343193548e+01, 2.00263956809063e+01, 1.99951630353056e+01, 1.99566559107452e+01, 1.99100193116838e+01, 1.98544510699258e+01, 1.97892297242127e+01, 1.97137418336357e+01, 1.96275075585713e+01, 1.95302033303198e+01, 1.94216804865748e+01, 1.93019788787713e+01, 1.91713346579345e+01, 1.90301817100741e+01, 1.88791465263084e+01, 1.87190366368064e+01, 1.85508230867992e+01, 1.83756177599302e+01, 1.81946466312119e+01, 1.80092202334304e+01, 1.78207027274167e+01, 1.76304809677280e+01, 1.74399348519881e+01, 1.72504100477691e+01, 1.70631939297209e+01, 1.68794952633553e+01, 1.67004278745692e+01, 1.65269982767000e+01, 1.63600970130825e+01, 1.62004933257467e+01, 1.60488326821387e+01, 1.59056366743189e+01, 1.57713048353463e+01, 1.56461179790650e+01, 1.55302427461537e+01, 1.54237371176629e+01, 1.53265567277415e+01, 1.52385618642052e+01, 1.51595250867588e+01, 1.50891394181702e+01, 1.50270270750401e+01, 1.49727487041467e+01, 1.49258130798899e+01, 1.48856872001987e+01, 1.48518066943475e+01, 1.48235864283598e+01, 1.48004311640750e+01, 1.47817460987217e+01, 1.47669470853907e+01, 1.47554703136593e+01, 1.47467812162691e+01, 1.47403823644122e+01, 1.47358201225156e+01, 1.47326898543630e+01, 1.47306395059659e+01, 1.47293714357728e+01, 1.47286424176096e+01, 1.47282618033474e+01, 1.47280878973005e+01, 1.47280226590984e+01, 1.47280049127218e+01, 1.47280022933937e+01, 1.47280020967599e+01, 1.47279986357823e+01, 1.47279751871516e+01, 1.47278889875338e+01, 1.47276592112234e+01, 1.47271563525166e+01, 1.47261933222563e+01, 1.47245186180605e+01, 1.47218118870495e+01, 1.47176821524097e+01, 1.47116689192045e+01, 1.47032463081729e+01, 1.46918302864367e+01, 1.46767889691308e+01, 1.46574558553200e+01, 1.46331457364587e+01, 1.46031728800176e+01, 1.45668709516603e+01, 1.45236140063894e+01, 1.44728377646877e+01, 1.44140603074094e+01, 1.43469012860839e+01, 1.42710987638869e+01, 1.41865228825070e+01, 1.40931856906147e+01, 1.39912466620809e+01, 1.38810136605832e+01, 1.37629393501301e+01, 1.36376132840852e+01, 1.35057501056717e+01, 1.33681744435086e+01, 1.32258031782351e+01, 1.30796257928260e+01, 1.29306835111468e+01, 1.27800478935947e+01, 1.26287995125291e+01, 1.24780072856165e+01, 1.23287090052181e+01, 1.21818935601155e+01, 1.20384852897131e+01, 1.18993308279615e+01, 1.17651886789849e+01, 1.16367216241274e+01, 1.15144919070774e+01, 1.13989590022666e+01, 1.12904796632400e+01, 1.11893098859158e+01, 1.10956084093699e+01, 1.10094414063990e+01, 1.09307880734330e+01, 1.08595468983485e+01, 1.07955424515202e+01, 1.07385326005393e+01, 1.06882160876835e+01, 1.06442404305564e+01, 1.06062101120028e+01, 1.05736950183985e+01, 1.05462390690589e+01, 1.05233689570825e+01, 1.05046028964482e+01, 1.04894592443804e+01, 1.04774648444166e+01, 1.04681629166192e+01, 1.04611203090856e+01, 1.04559339211055e+01, 1.04522361143130e+01, 1.04496989446454e+01, 1.04480370747887e+01, 1.04470092631716e+01, 1.04464183698179e+01, 1.04461098691508e+01, 1.04459689124003e+01, 1.04459160346167e+01, 1.04459016505737e+01, 1.04458995275190e+01, 1.04458993686252e+01, 1.04458965783714e+01, 1.04458776739926e+01, 1.04458081795127e+01, 1.04456229328718e+01, 1.04452175257751e+01, 1.04444411251600e+01, 1.04430909657119e+01, 1.04409087698324e+01, 1.04375793129631e+01, 1.04327313069306e+01, 1.04259407202354e+01, 1.04167365901090e+01, 1.04046093052259e+01, 1.03890212496405e+01, 1.03694195987883e+01, 1.03452509501433e+01, 1.03159773595469e+01, 1.02810932467761e+01, 1.02401425400324e+01, 1.01927353594132e+01, 1.01385635049100e+01, 1.00774140245357e+01, 1.00091801992174e+01, 9.93386939455757e+00, 9.85160739054926e+00, 9.76263899668785e+00, 9.66732497285388e+00, 9.56613548255672e+00, 9.45964048078961e+00, 9.34849756447912e+00, 9.23343787913925e+00, 9.11525068256537e+00, 8.99476712814178e+00, 8.87284376647310e+00, 8.75034619507274e+00, 8.62813322669712e+00, 8.50704190343125e+00, 8.38787365262026e+00, 8.27138185323484e+00, 8.15826104719446e+00, 8.04913798254132e+00, 7.94456461248651e+00, 7.84501309996718e+00, 7.75087279895753e+00, 7.66244911022972e+00, 7.57996404849490e+00, 7.50355831528539e+00, 7.43329464977000e+00, 7.36916222704857e+00, 7.31108188705697e+00, 7.25891200224807e+00, 7.21245482336782e+00, 7.17146317481519e+00, 7.13564739995693e+00, 7.10468247918966e+00, 7.07821525759789e+00, 7.05587172410618e+00, 7.03726428057178e+00, 7.02199892877837e+00, 7.00968228797323e+00, 6.99992833811898e+00, 6.99236476728112e+00, 6.98663878834530e+00, 6.98242228300757e+00, 6.97941613158780e+00, 6.97735359681167e+00, 6.97600264856427e+00, 6.97516714414450e+00, 6.97468681334980e+00, 6.97443603775966e+00, 6.97432145640649e+00, 6.97427847299893e+00, 6.97426678047117e+00, 6.97426505467862e+00, 6.97426492546845e+00, 6.97426265582713e+00, 6.97424727867598e+00, 6.97419075068328e+00, 6.97404006812816e+00, 6.97371030466165e+00, 6.97307877438293e+00, 6.97198055737734e+00, 6.97020559901769e+00, 6.96749756419678e+00, 6.96355459210889e+00, 6.95803205472229e+00, 6.95054737113012e+00, 6.94068686933122e+00, 6.92801461625267e+00, 6.91208305672222e+00, 6.89244521483275e+00, 6.86866812051337e+00, 6.84034703547254e+00, 6.80711997260419e+00, 6.76868193878719e+00, 6.72479829015350e+00, 6.67531657798911e+00, 6.62017628752972e+00, 6.55941593378629e+00, 6.49317707809065e+00, 6.42170496302243e+00, 6.34534562525673e+00, 6.26453952611723e+00, 6.17981192614641e+00, 6.09176040897906e+00, 6.00104011678188e+00, 5.90834738088891e+00, 5.81440250589112e+00, 5.71993248629460e+00, 5.62565440024765e+00, 5.53226013894482e+00, 5.44040300286023e+00, 5.35068654089124e+00, 5.26365584198851e+00, 5.17979132703085e+00, 5.09950494548614e+00, 5.02313856692609e+00, 4.95096427730646e+00, 4.88318624513891e+00, 4.81994381048731e+00, 4.76131546462742e+00, 4.70732342323698e+00, 4.65793854380326e+00, 4.61308539173627e+00, 4.57264731371594e+00, 4.53647142667138e+00, 4.50437347342469e+00, 4.47614252958584e+00, 4.45154556992596e+00, 4.43033191614821e+00, 4.41223759229825e+00, 4.39698961005064e+00, 4.38431019518832e+00, 4.37392095045969e+00, 4.36554693058648e+00, 4.35892058458932e+00, 4.35378550094209e+00, 4.34989987443425e+00, 4.34703960187309e+00, 4.34500090837413e+00, 4.34360240793207e+00, 4.34268651156317e+00, 4.34212011321002e+00, 4.34179450678692e+00, 4.34162451564386e+00, 4.34154684637208e+00, 4.34151771012239e+00, 4.34150978437260e+00, 4.34150861454914e+00, 4.34150852673206e+00, 4.34150698107801e+00, 4.34149650905302e+00, 4.34145801291157e+00, 4.34135539733190e+00, 4.34113083071792e+00, 4.34070077822381e+00, 4.33995297415446e+00, 4.33874448729315e+00, 4.33690101587512e+00, 4.33421752783192e+00, 4.33046033459238e+00, 4.32537065105321e+00, 4.31866964855305e+00, 4.31006495083434e+00, 4.29925845537595e+00, 4.28595528613464e+00, 4.26987360257675e+00, 4.25075490977962e+00, 4.22837444282490e+00, 4.20255114417071e+00, 4.17315672368971e+00, 4.14012329496145e+00, 4.10344912316223e+00, 4.06320210089918e+00, 4.01952068561491e+00, 3.97261217815650e+00, 3.92274838498530e+00, 3.87025887153015e+00, 3.81552216536730e+00, 3.75895539021074e+00, 3.70100289319206e+00, 3.64212446154085e+00, 3.58278370933659e+00, 3.52343715514936e+00, 3.46452441668868e+00, 3.40645983180242e+00, 3.34962569017876e+00, 3.29436713982645e+00, 3.24098872721166e+00, 3.18975244667374e+00, 3.14087711656522e+00, 3.09453886627412e+00, 3.05087250717702e+00, 3.00997356738446e+00, 2.97190078999964e+00, 2.93667892275393e+00, 2.90430165914862e+00, 2.87473462429163e+00, 2.84791833001617e+00, 2.82377105193878e+00, 2.80219160482611e+00, 2.78306201144239e+00, 2.76625007373698e+00, 2.75161186383622e+00, 2.73899415604348e+00, 2.72823682029085e+00, 2.71917519273912e+00, 2.71164243115074e+00, 2.70547185208120e+00, 2.70049923481197e+00, 2.69656506436889e+00, 2.69351667409167e+00, 2.69121023820152e+00, 2.68951255771593e+00, 2.68830257975617e+00, 2.68747259136094e+00, 2.68692903457711e+00, 2.68659289965356e+00, 2.68639966702926e+00, 2.68629878554369e+00, 2.68625269272659e+00, 2.68623540184366e+00, 2.68623069831597e+00, 2.68623000408556e+00, 2.68622995174866e+00, 2.68622902761885e+00, 2.68622276650952e+00, 2.68619975010081e+00, 2.68613839750876e+00, 2.68600413242931e+00, 2.68574701241010e+00, 2.68529992131474e+00, 2.68457741807613e+00, 2.68347532524675e+00, 2.68187112860878e+00, 2.67962524286929e+00, 2.67658317681816e+00, 2.67257860305709e+00, 2.66743730184856e+00, 2.66098190598630e+00, 2.65303732523009e+00, 2.64343667748567e+00, 2.63202750359573e+00, 2.61867799847716e+00, 2.60328295912188e+00, 2.58576913524179e+00, 2.56609967560798e+00, 2.54427739497287e+00, 2.52034664275342e+00, 2.49439363220331e+00, 2.46654518147620e+00, 2.43696591733583e+00, 2.40585408866099e+00, 2.37343622086431e+00, 2.33996090597472e+00, 2.30569206116083e+00, 2.27090199887569e+00, 2.23586463584414e+00, 2.20084912981766e+00, 2.16611417830210e+00, 2.13190314909659e+00, 2.09844014514509e+00, 2.06592704173870e+00, 2.03454147707362e+00, 2.00443573065966e+00, 1.97573638973715e+00, 1.94854468206682e+00, 1.92293734351679e+00, 1.89896788929825e+00, 1.87666816647834e+00, 1.85605008024188e+00, 1.83710740497543e+00, 1.81981761147162e+00, 1.80414366158420e+00, 1.79003574008765e+00, 1.77743290931460e+00, 1.76626468476157e+00, 1.75645253899984e+00, 1.74791134690530e+00, 1.74055078761135e+00, 1.73427671802027e+00, 1.72899252959758e+00, 1.72460049501649e+00, 1.72100310457488e+00, 1.71810438479107e+00, 1.71581118385753e+00, 1.71403440137801e+00, 1.71269013370530e+00, 1.71170070184815e+00, 1.71099552682638e+00, 1.71051181785804e+00, 1.71019504198321e+00, 1.70999914955383e+00, 1.70988653810028e+00, 1.70982774687332e+00, 1.70980088515572e+00, 1.70979080847569e+00, 1.70978806738172e+00, 1.70978766280225e+00, 1.70978763213568e+00, 1.70978708844437e+00, 1.70978340485962e+00, 1.70976986367084e+00, 1.70973376828163e+00, 1.70965477668081e+00, 1.70950350703476e+00, 1.70924047513417e+00, 1.70881541889448e+00, 1.70816705963011e+00, 1.70722334274930e+00, 1.70590219102186e+00, 1.70411279076528e+00, 1.70175741442228e+00, 1.69873376162894e+00, 1.69493777510873e+00, 1.69026685840240e+00, 1.68462339121431e+00, 1.67791840751581e+00, 1.67007527467304e+00, 1.66103319231568e+00, 1.65075032091811e+00, 1.63920635497173e+00, 1.62640437584144e+00, 1.61237185490288e+00, 1.59716072640254e+00, 1.58084650782322e+00, 1.56352650798898e+00, 1.54531722344808e+00, 1.52635107556190e+00, 1.50677267884679e+00, 1.48673485181620e+00, 1.46639458344700e+00, 1.44590915242856e+00, 1.42543256566783e+00, 1.40511244177816e+00, 1.38508741986045e+00, 1.36548512904427e+00, 1.34642071437632e+00, 1.32799588274804e+00, 1.31029841012808e+00, 1.29340203847143e+00, 1.27736668626790e+00, 1.26223889904726e+00, 1.24805247328690e+00, 1.23482919715398e+00, 1.22257966272256e+00, 1.21130411549233e+00, 1.20099331734224e+00, 1.19162940797312e+00, 1.18318675718054e+00, 1.17563280588931e+00, 1.16892889782046e+00, 1.16303110605177e+00, 1.15789105970557e+00, 1.15345677569551e+00, 1.14967349904117e+00, 1.14648455289151e+00, 1.14383219628037e+00, 1.14165848400667e+00, 1.13990611915478e+00, 1.13851928495269e+00, 1.13744443923100e+00, 1.13663105202104e+00, 1.13603226511594e+00, 1.13560545196177e+00, 1.13531265720199e+00, 1.13512089762215e+00, 1.13500231005300e+00, 1.13493413679654e+00, 1.13489854503072e+00, 1.13488228304571e+00, 1.13487618264391e+00, 1.13487452318955e+00, 1.13487427825768e+00, 1.13487425957437e+00, 1.13487392678158e+00, 1.13487167206318e+00, 1.13486338350572e+00, 1.13484128945358e+00, 1.13479293817016e+00, 1.13470034354519e+00, 1.13453933281434e+00, 1.13427912914048e+00, 1.13388219818350e+00, 1.13330438312849e+00, 1.13249534662056e+00, 1.13139933021113e+00, 1.12995623186615e+00, 1.12810298961570e+00, 1.12577524462949e+00, 1.12290924037822e+00, 1.11944389704478e+00, 1.11532298339785e+00, 1.11049729372272e+00, 1.10492672708973e+00, 1.09858216211419e+00, 1.09144702389039e+00, 1.08351845170273e+00, 1.07480799618232e+00, 1.06534180146149e+00, 1.05516025925686e+00, 1.04431715464585e+00, 1.03287835427783e+00, 1.02092011378620e+00, 1.00852709984955e+00, 9.95790232322015e-01, 9.82804452910322e-01, 9.69666519894027e-01, 9.56472915061163e-01, 9.43317931521300e-01, 9.30291991608897e-01, 9.17480224726084e-01, 9.04961317288148e-01, 8.92806632003065e-01, 8.81079582089527e-01, 8.69835237837442e-01, 8.59120137924151e-01, 8.48972275697207e-01, 8.39421230688281e-01, 8.30488417377168e-01, 8.22187426148460e-01, 8.14524435006135e-01, 8.07498674534308e-01, 8.01102932489776e-01, 7.95324088032217e-01, 7.90143668746680e-01, 7.85538426154396e-01, 7.81480927249376e-01, 7.77940160690811e-01, 7.74882156616826e-01, 7.72270618652563e-01, 7.70067565638293e-01, 7.68233979012287e-01, 7.66730449800626e-01, 7.65517816978793e-01, 7.64557786790171e-01, 7.63813520661245e-01, 7.63250177865913e-01, 7.62835398263029e-01, 7.62539710420647e-01, 7.62336851346703e-01, 7.62203985891247e-01, 7.62121816614609e-01, 7.62074578393580e-01, 7.62049916061635e-01, 7.62038647701623e-01, 7.62034420560965e-01, 7.62033270677179e-01, 7.62033100956785e-01, 7.62033087919221e-01, 7.62032854492467e-01, 7.62031272992687e-01, 7.62025459243937e-01, 7.62009962014474e-01, 7.61976047164509e-01, 7.61911097923111e-01, 7.61798156134091e-01, 7.61615626904572e-01, 7.61337166249767e-01, 7.60931767871050e-01, 7.60364060895370e-01, 7.59594824988376e-01, 7.58581722524198e-01, 7.57280239347878e-01, 7.55644816175661e-01, 7.53630142144161e-01, 7.51192571005518e-01, 7.48291609794499e-01, 7.44891420496820e-01, 7.40962268464322e-01, 7.36481848152508e-01, 7.31436418083959e-01, 7.25821683264883e-01, 7.19643374571511e-01, 7.12917490255634e-01, 7.05670183507557e-01, 6.97937300333251e-01, 6.89763592000070e-01, 6.81201644152786e-01, 6.72310578854059e-01, 6.63154595187187e-01, 6.53801418195336e-01, 6.44320724937216e-01, 6.34782610953209e-01, 6.25256151463716e-01, 6.15808100366877e-01, 6.06501757786937e-01, 5.97396024659425e-01, 5.88544651518383e-01, 5.79995678914237e-01, 5.71791059113124e-01, 5.63966443051406e-01, 5.56551112893402e-01, 5.49568038782652e-01, 5.43034038209619e-01, 5.36960017518268e-01, 5.31351277099773e-01, 5.26207864439861e-01, 5.21524962094638e-01, 5.17293300601636e-01, 5.13499589067401e-01, 5.10126958535487e-01, 5.07155415100989e-01, 5.04562301016264e-01, 5.02322762689673e-01, 5.00410224520053e-01, 4.98796866981088e-01, 4.97454106357651e-01, 4.96353072161746e-01, 4.95465076669419e-01, 4.94762069393524e-01, 4.94217067822447e-01, 4.93804554589801e-01, 4.93500830554545e-01, 4.93284313190493e-01, 4.93135770285583e-01, 4.93038480255763e-01, 4.92978312344739e-01, 4.92943722506723e-01, 4.92925663707553e-01, 4.92917412541793e-01, 4.92914317252563e-01, 4.92913475259577e-01, 4.92913350983194e-01, 4.92913341363703e-01, 4.92913168186000e-01, 4.92911994882228e-01, 4.92907681702839e-01, 4.92896184422506e-01, 4.92871023272592e-01, 4.92822838066282e-01, 4.92739047987342e-01, 4.92603632754339e-01, 4.92397049793020e-01, 4.92096299515161e-01, 4.91675147631625e-01, 4.91104509431430e-01, 4.90352995980572e-01, 4.89387616155511e-01, 4.88174621377468e-01, 4.86680472071441e-01, 4.84872896646166e-01, 4.82722005773503e-01, 4.80201417688534e-01, 4.77289344966572e-01, 4.73969590576921e-01, 4.70232401639919e-01, 4.66075133634789e-01, 4.61502685866812e-01, 4.56527680426085e-01, 4.51170370860503e-01, 4.45458282205550e-01, 4.39425599509327e-01, 4.33112336180308e-01, 4.26563325118870e-01, 4.19827083730574e-01, 4.12954608021607e-01, 4.05998150958219e-01, 3.99010036470635e-01, 3.92041553584114e-01, 3.85141966074663e-01, 3.78357662785049e-01, 3.71731463273903e-01, 3.65302083654719e-01, 3.59103758962483e-01, 3.53166011584996e-01, 3.47513550408380e-01, 3.42166282346923e-01, 3.37139416690150e-01, 3.32443642935296e-01, 3.28085364148635e-01, 3.24066970071591e-01, 3.20387136827129e-01, 3.17041142900699e-01, 3.14021193826498e-01, 3.11316750516156e-01, 3.08914858286855e-01, 3.06800475290437e-01, 3.04956800163644e-01, 3.03365599303376e-01, 3.02007534235997e-01, 3.00862489147008e-01, 2.99909897839278e-01, 2.99129068293055e-01, 2.98499501727015e-01, 2.98001201737350e-01, 2.97614967862202e-01, 2.97322666917069e-01, 2.97107474796391e-01, 2.96954081235413e-01, 2.96848850336985e-01, 2.96779930508855e-01, 2.96737308801575e-01, 2.96712806411271e-01, 2.96700014204532e-01, 2.96694169396531e-01, 2.96691976816462e-01, 2.96691380382391e-01, 2.96691292350012e-01, 2.96691285480698e-01, 2.96691161100377e-01, 2.96690318406043e-01, 2.96687220582136e-01, 2.96678962994662e-01, 2.96660891840531e-01, 2.96626284851833e-01, 2.96566107493341e-01, 2.96468856925067e-01, 2.96320504612922e-01, 2.96104546739187e-01, 2.95802169336902e-01, 2.95392532184934e-01, 2.94853171815348e-01, 2.94160519429276e-01, 2.93290524120962e-01, 2.92219365742994e-01, 2.90924235330699e-01, 2.89384154725361e-01, 2.87580801502088e-01, 2.85499301200495e-01, 2.83128946843270e-01, 2.80463806361205e-01, 2.77503182157680e-01, 2.74251893669984e-01, 2.70720363092472e-01, 2.66924495741947e-01, 2.62885358902302e-01, 2.58628675244431e-01, 2.54184157917152e-01, 2.49584723119170e-01, 2.44865621641672e-01, 2.40063533136826e-01, 2.35215665738913e-01, 2.30358899533485e-01, 2.25529005906694e-01, 2.20759966855837e-01, 2.16083409782712e-01, 2.11528164939310e-01, 2.07119945206806e-01, 2.02881141720049e-01, 1.98830724242555e-01, 1.94984232198685e-01, 1.91353840777026e-01, 1.87948486321012e-01, 1.84774036050754e-01, 1.81833488727918e-01, 1.79127194906286e-01, 1.76653087661251e-01, 1.74406916961601e-01, 1.72382482980247e-01, 1.70571865524870e-01, 1.68965648327266e-01, 1.67553138115764e-01, 1.66322579186584e-01, 1.65261364589320e-01, 1.64356245065960e-01, 1.63593536571206e-01, 1.62959326605055e-01, 1.62439678775762e-01, 1.62020834064379e-01, 1.61689406271520e-01, 1.61432568192375e-01, 1.61238224282812e-01, 1.61095165036734e-01, 1.60993198063871e-01, 1.60923250984730e-01, 1.60877441762580e-01, 1.60849112953606e-01, 1.60832827526548e-01, 1.60824325305313e-01, 1.60820440624846e-01, 1.60818983355931e-01, 1.60818586944274e-01, 1.60818528434795e-01, 1.60818523830343e-01, 1.60818439961538e-01, 1.60817871738665e-01, 1.60815782899888e-01, 1.60810214889281e-01, 1.60798029772295e-01, 1.60774695110831e-01, 1.60734119991886e-01, 1.60668550693565e-01, 1.60568533615674e-01, 1.60422952002662e-01, 1.60219141520543e-01, 1.59943087784164e-01, 1.59579706354442e-01, 1.59113202473868e-01, 1.58527503900821e-01, 1.57806755764796e-01, 1.56935861644662e-01, 1.55901050443532e-01, 1.54690444569866e-01, 1.53294601962081e-01, 1.51707003128699e-01, 1.49924455032526e-01, 1.47947386567672e-01, 1.45780015550642e-01, 1.43430374285238e-01, 1.40910189301763e-01, 1.38234620041374e-01, 1.35421870166681e-01, 1.32492692952965e-01, 1.29469818108930e-01, 1.26377330890399e-01, 1.23240035301428e-01, 1.20082831631784e-01, 1.16930134930783e-01, 1.13805355831140e-01, 1.10730459066453e-01, 1.07725608725104e-01, 1.04808903315764e-01, 1.01996198517050e-01, 9.93010113172161e-02, 9.67344962449559e-02, 9.43054825400580e-02, 9.20205603129887e-02, 8.98842038261307e-02, 8.78989208045482e-02, 8.60654179465428e-02, 8.43827743720453e-02, 8.28486164560829e-02, 8.14592892142727e-02, 8.02100210299841e-02, 7.90950799600267e-02, 7.81079210704902e-02, 7.72413252012996e-02, 7.64875302190759e-02, 7.58383561901965e-02, 7.52853259999213e-02, 7.48197827815531e-02, 7.44330051377418e-02, 7.41163205798969e-02, 7.38612169384538e-02, 7.36594507720483e-02, 7.35031510963182e-02, 7.33849161334277e-02, 7.32979003156515e-02, 7.32358885132314e-02, 7.31933544344585e-02, 7.31655003791616e-02, 7.31482760082542e-02, 7.31383744878613e-02, 7.31332052288636e-02, 7.31308434035824e-02, 7.31299574097477e-02, 7.31297163987463e-02, 7.31296808260495e-02, 7.31296780018553e-02, 7.31296262448708e-02, 7.31292755842219e-02, 7.31279865263232e-02, 7.31245504307765e-02, 7.31170309218001e-02, 7.31026313200331e-02, 7.30775939137445e-02, 7.30371365326903e-02, 7.29754314475883e-02, 7.28856310158577e-02, 7.27599436328745e-02, 7.25897623236662e-02, 7.23658466317474e-02, 7.20785562790596e-02, 7.17181323883267e-02, 7.12750189672207e-02, 7.07402140322486e-02, 7.01056364720358e-02, 6.93644918657057e-02, 6.85116183693282e-02, 6.75437928426135e-02, 6.64599779135955e-02, 6.52614928419268e-02, 6.39520948296166e-02, 6.25379626175544e-02, 6.10275803793243e-02, 5.94315265042127e-02, 5.77621782002637e-02, 5.60333483061513e-02, 5.42598747467906e-02, 5.24571853449274e-02, 5.06408610737736e-02, 4.88262193925593e-02, 4.70279363301671e-02, 4.52597218911279e-02, 4.35340586384285e-02, 4.18620084383363e-02, 4.02530877612861e-02, 3.87152079539490e-02, 3.72546737632433e-02, 3.58762312182321e-02, 3.45831547697486e-02, 3.33773632621330e-02, 3.22595547190119e-02, 3.12293508861217e-02, 3.02854438119340e-02, 2.94257382971495e-02, 2.86474856628374e-02, 2.79474058527143e-02, 2.73217963016776e-02, 2.67666272074365e-02, 2.62776237991389e-02, 2.58503369015060e-02, 2.54802035560573e-02, 2.51625996975449e-02, 2.48928869170726e-02, 2.46664551972415e-02, 2.44787632104817e-02, 2.43253773610164e-02, 2.42020102623215e-02, 2.41045588154591e-02, 2.40291415320826e-02, 2.39721342717942e-02, 2.39302031760147e-02, 2.39003333121985e-02, 2.38798514188403e-02, 2.38664411737304e-02, 2.38581495933019e-02, 2.38533834935239e-02, 2.38508953733145e-02, 2.38497585793596e-02, 2.38493321365709e-02, 2.38492161345505e-02, 2.38491990129077e-02, 2.38491976411552e-02, 2.38491723450736e-02, 2.38490009608796e-02, 2.38483709412733e-02, 2.38466915926310e-02, 2.38430166447401e-02, 2.38359796975161e-02, 2.38237455712493e-02, 2.38039805180250e-02, 2.37738441455093e-02, 2.37300057037046e-02, 2.36686870127372e-02, 2.35857336844774e-02, 2.34767153414666e-02, 2.33370542190627e-02, 2.31621798473540e-02, 2.29477055098357e-02, 2.26896199922351e-02, 2.23844859667821e-02, 2.20296344591802e-02, 2.16233435055955e-02, 2.11649886028056e-02, 2.06551531085779e-02, 2.00956884859025e-02, 1.94897171741530e-02, 1.88415747287820e-02, 1.81566923406322e-02, 1.74414254665720e-02, 1.67028385621535e-02, 1.59484592972703e-02, 1.51860177610672e-02, 1.44231868267397e-02, 1.36673390758943e-02, 1.29253335663435e-02, 1.22033424749688e-02, 1.15067237730955e-02, 1.08399421747052e-02, 1.02065368161419e-02, 9.60913080223236e-03, 9.04947579424942e-03, 8.52852412401717e-03, 8.04652007487761e-03, 7.60310147015986e-03, 7.19740473822128e-03, 6.82816892260263e-03, 6.49383273129445e-03, 6.19261872301990e-03, 5.92260535351130e-03, 5.68179092678633e-03, 5.46814641440108e-03, 5.27965159633936e-03, 5.11431894783243e-03, 4.97021432053657e-03, 4.84547441659799e-03, 4.73831557655302e-03, 4.64703460337290e-03, 4.57000760455628e-03, 4.50569027678682e-03, 4.45261894109450e-03, 4.40941116062987e-03, 4.37476629398503e-03, 4.34746686352838e-03, 4.32638110523698e-03, 4.31046652452238e-03, 4.29877395324496e-03, 4.29045155099317e-03, 4.28474816047952e-03, 4.28101540697412e-03, 4.27870802032260e-03, 4.27738190779081e-03, 4.27668967555661e-03, 4.27637341574894e-03, 4.27625478001659e-03, 4.27622250811766e-03, 4.27621774534259e-03, 4.27621735983053e-03, 4.27621021178481e-03, 4.27616177984191e-03, 4.27598374388793e-03, 4.27550919204123e-03, 4.27447078738652e-03, 4.27248265904012e-03, 4.26902697599019e-03, 4.26344620726837e-03, 4.25494208374033e-03, 4.24258229631391e-03, 4.22531585529384e-03, 4.20199786749720e-03, 4.17142411903626e-03, 4.13237523204657e-03, 4.08366932527571e-03, 4.02422092213328e-03, 3.95310287162687e-03, 3.86960799081493e-03, 3.77330844470691e-03, 3.66411205909145e-03, 3.54231308136832e-03, 3.40747944352415e-03, 3.26310350623515e-03, 3.10928189145443e-03, 2.94773749930499e-03, 2.78043494550673e-03, 2.60949746662220e-03, 2.43811772003966e-03, 2.26654506724344e-03, 2.09753919591549e-03, 1.93316038308922e-03, 1.77525973766990e-03, 1.62537085933678e-03, 1.48435135167091e-03, 1.35252335501690e-03, 1.22996202673420e-03, 1.11656779778713e-03, 1.01214437146524e-03, 9.16450829812732e-04, 8.29208679476778e-04, 7.50075371087178e-04, 6.78624004287087e-04, 6.14273922018072e-04, 5.56034770518465e-04, 5.03023623615206e-04, 4.54756063867023e-04, 4.10635626678868e-04, 3.69955546879908e-04, 3.32241746598072e-04, 2.97196506799935e-04, 2.64674114802688e-04, 2.34613722113863e-04, 2.06757120238876e-04, 1.80906815934634e-04, 1.57011118893990e-04, 1.35059839266205e-04, 1.15042856478453e-04, 9.69333274696282e-05, 8.06840794780328e-05, 6.62327917666516e-05, 5.35136965496008e-05, 4.24573584112409e-05, 3.29853614360331e-05, 2.50069927005825e-05, 1.84182507112496e-05, 1.31024889356013e-05, 8.93211950625528e-06, 5.77095471498978e-06, 3.47690478719187e-06, 1.90487785545201e-06, 9.09836757045923e-07, 3.50069051080482e-07, -2.13162820728030e-14, -2.13162820728030e-14],
[ 5.00000000000000e+01, 4.99999999536583e+01, 4.99999985206815e+01, 4.99999888120367e+01, 4.99999531217960e+01, 4.99998579829068e+01, 4.99996497645857e+01, 4.99992509663783e+01, 4.99985573427549e+01, 4.99974359667058e+01, 4.99957243120543e+01, 4.99932304032011e+01, 4.99897340489890e+01, 4.99849891456771e+01, 4.99787270039629e+01, 4.99706606277847e+01, 4.99604898492052e+01, 4.99479072045391e+01, 4.99326044220519e+01, 4.99142793804824e+01, 4.98926433893017e+01, 4.98674286346711e+01, 4.98383956281440e+01, 4.98053404872384e+01, 4.97681018677790e+01, 4.97265673580657e+01, 4.96806791362575e+01, 4.96304386876632e+01, 4.95759103812701e+01, 4.95172237181562e+01, 4.94545740910035e+01, 4.93882219349112e+01, 4.93184902043160e+01, 4.92457601762512e+01, 4.91704656517735e+01, 4.90930856993371e+01, 4.90141361499902e+01, 4.89341601088863e+01, 4.88537177864200e+01, 4.87733759729035e+01, 4.86936974827238e+01, 4.86152308789026e+01, 4.85385007599438e+01, 4.84639988517038e+01, 4.83921761019179e+01, 4.83234359278506e+01, 4.82581287215034e+01, 4.81965476742074e+01, 4.81389259446013e+01, 4.80854351614666e+01, 4.80361852254832e+01, 4.79912253510986e+01, 4.79505462705191e+01, 4.79140835054868e+01, 4.78817215982145e+01, 4.78532991800678e+01, 4.78286147449703e+01, 4.78074329840224e+01, 4.77894915286250e+01, 4.77745079418759e+01, 4.77621867926638e+01, 4.77522266443201e+01, 4.77443267904885e+01, 4.77381935755653e+01, 4.77335461460485e+01, 4.77301214926015e+01, 4.77276786605863e+01, 4.77260020289626e+01, 4.77249035832836e+01, 4.77242241372979e+01, 4.77238334884651e+01, 4.77236295244554e+01, 4.77235363293054e+01, 4.77235013682136e+01, 4.77234918579137e+01, 4.77234904542121e+01, 4.77234902700760e+01, 4.77234859799268e+01, 4.77234569135152e+01, 4.77233500620217e+01, 4.77230652329460e+01, 4.77224418801921e+01, 4.77212480418484e+01, 4.77191718116678e+01, 4.77158157087256e+01, 4.77106942393631e+01, 4.77032348668760e+01, 4.76927825179716e+01, 4.76786076613896e+01, 4.76599178941102e+01, 4.76358728654021e+01, 4.76056022599355e+01, 4.75682264497389e+01, 4.75228793123732e+01, 4.74687326009916e+01, 4.74050211434915e+01, 4.73310680470827e+01, 4.72463089987454e+01, 4.71503146925033e+01, 4.70428103961300e+01, 4.69236917094189e+01, 4.67930356779718e+01, 4.66511066178598e+01, 4.64983562727206e+01, 4.63354182461135e+01, 4.61630969944462e+01, 4.59823519867432e+01, 4.57942778940467e+01, 4.56000818301552e+01, 4.54010587114642e+01, 4.51985657431669e+01, 4.49939968977306e+01, 4.47887580672973e+01, 4.45842433845171e+01, 4.43818130487011e+01, 4.41827728842089e+01, 4.39883557975720e+01, 4.37997052775462e+01, 4.36178610792766e+01, 4.34437472309626e+01, 4.32781624851580e+01, 4.31217733017729e+01, 4.29751093986244e+01, 4.28385618459317e+01, 4.27123836229427e+01, 4.25966925056071e+01, 4.24914761179397e+01, 4.23965989565899e+01, 4.23118111853139e+01, 4.22367589892459e+01, 4.21709962737632e+01, 4.21139974859796e+01, 4.20651713266009e+01, 4.20238751056511e+01, 4.19894294783239e+01, 4.19611332787926e+01, 4.19382781526580e+01, 4.19201626755352e+01, 4.19061056387640e+01, 4.18954581857528e+01, 4.18876144959382e+01, 4.18820207389361e+01, 4.18781820595421e+01, 4.18756674042383e+01, 4.18741120603211e+01, 4.18732178473802e+01, 4.18727509746886e+01, 4.18725376537429e+01, 4.18724576292132e+01, 4.18724358605556e+01, 4.18724326475457e+01, 4.18724323638981e+01, 4.18724268059626e+01, 4.18723891501217e+01, 4.18722507231170e+01, 4.18718817263907e+01, 4.18710741802187e+01, 4.18695276102285e+01, 4.18668380417407e+01, 4.18624907952638e+01, 4.18558574931097e+01, 4.18461975912935e+01, 4.18326646408203e+01, 4.18143173562434e+01, 4.17901354257424e+01, 4.17590398360265e+01, 4.17199173093305e+01, 4.16716482635508e+01, 4.16131375179921e+01, 4.15433467869759e+01, 4.14613278446784e+01, 4.13662551214048e+01, 4.12574564186206e+01, 4.11344404210019e+01, 4.09969197498712e+01, 4.08448284516213e+01, 4.06783330499830e+01, 4.04978366074090e+01, 4.03039756220168e+01, 4.00976100008789e+01, 3.98798067506667e+01, 3.96518183559100e+01, 3.94150570213936e+01, 3.91710660101999e+01, 3.89214892223965e+01, 3.86680399793885e+01, 3.84124697744486e+01, 3.81565375834646e+01, 3.79019802344618e+01, 3.76504843056924e+01, 3.74036600282264e+01, 3.71630176700156e+01, 3.69299468440669e+01, 3.67056991025221e+01, 3.64913740586115e+01, 3.62879091388569e+01, 3.60960729307757e+01, 3.59164619750187e+01, 3.57495007661191e+01, 3.55954446756136e+01, 3.54543854915199e+01, 3.53262592714451e+01, 3.52108562238854e+01, 3.51078323549248e+01, 3.50167226385227e+01, 3.49369554829602e+01, 3.48678682710323e+01, 3.48087237464243e+01, 3.47587270041853e+01, 3.47170428213868e+01, 3.46828130379155e+01, 3.46551736704933e+01, 3.46332714193471e+01, 3.46162792103579e+01, 3.46034104095714e+01, 3.45939313545951e+01, 3.45871718706577e+01, 3.45825334788598e+01, 3.45794950599993e+01, 3.45776158076439e+01, 3.45765353859612e+01, 3.45759712973954e+01, 3.45757135581409e+01, 3.45756168708391e+01, 3.45755905695101e+01, 3.45755866874876e+01, 3.45755863761014e+01, 3.45755806294460e+01, 3.45755416950024e+01, 3.45753985677511e+01, 3.45750170421428e+01, 3.45741820780850e+01, 3.45725830044165e+01, 3.45698021467846e+01, 3.45653073936017e+01, 3.45584491289886e+01, 3.45484618638631e+01, 3.45344707834287e+01, 3.45155032990356e+01, 3.44905055431454e+01, 3.44583635778956e+01, 3.44179289027821e+01, 3.43680476504399e+01, 3.43075926598644e+01, 3.42354974253052e+01, 3.41507907507578e+01, 3.40526308102799e+01, 3.39403372391570e+01, 3.38134198743359e+01, 3.36716028349557e+01, 3.35148427901008e+01, 3.33433404988651e+01, 3.31575450171227e+01, 3.29581503272758e+01, 3.27460845352381e+01, 3.25224921608308e+01, 3.22887103891772e+01, 3.20462404196277e+01, 3.17967152210299e+01, 3.15418650661019e+01, 3.12834821767378e+01, 3.10233856844020e+01, 3.07633879241906e+01, 3.05052628705522e+01, 3.02507173167500e+01, 3.00013652193929e+01, 2.97587054824143e+01, 2.95241033396323e+01, 2.92987754027400e+01, 2.90837783622216e+01, 2.88800012553558e+01, 2.86881611465220e+01, 2.85088020035714e+01, 2.83422965054998e+01, 2.81888504858915e+01, 2.80485097055836e+01, 2.79211686550552e+01, 2.78065811073837e+01, 2.77043721696587e+01, 2.76140516077638e+01, 2.75350282407144e+01, 2.74666252124156e+01, 2.74080959489299e+01, 2.73586405982375e+01, 2.73174227287745e+01, 2.72835860357283e+01, 2.72562707740018e+01, 2.72346296082462e+01, 2.72178425478043e+01, 2.72051306219379e+01, 2.71957679518020e+01, 2.71890918927895e+01, 2.71845109553612e+01, 2.71815102641672e+01, 2.71796543825815e+01, 2.71785874098167e+01, 2.71780303465829e+01, 2.71777758181250e+01, 2.71776803354436e+01, 2.71776543618138e+01, 2.71776505281595e+01, 2.71776502324712e+01, 2.71776449228581e+01, 2.71776089494447e+01, 2.71774767073128e+01, 2.71771241978892e+01, 2.71763527373417e+01, 2.71748752893050e+01, 2.71723059672475e+01, 2.71681531858544e+01, 2.71618168643444e+01, 2.71525899959691e+01, 2.71396647946540e+01, 2.71221435093277e+01, 2.70990538571218e+01, 2.70693688680091e+01, 2.70320307575888e+01, 2.69859782566107e+01, 2.69301766337247e+01, 2.68636494633405e+01, 2.67855110274879e+01, 2.66949981148321e+01, 2.65914999071543e+01, 2.64745846374588e+01, 2.63440217745258e+01, 2.61997986408476e+01, 2.60421306022634e+01, 2.58714642682242e+01, 2.56884734931718e+01, 2.54940483458567e+01, 2.52892775821411e+01, 2.50754254826978e+01, 2.48539041673054e+01, 2.46262426485506e+01, 2.43940539311585e+01, 2.41590014081461e+01, 2.39227656762687e+01, 2.36870127235563e+01, 2.34533642621981e+01, 2.32233708118845e+01, 2.29984879900786e+01, 2.27800563338003e+01, 2.25692848545644e+01, 2.23672384075397e+01, 2.21748288364516e+01, 2.19928097417326e+01, 2.18217746192441e+01, 2.16621580393172e+01, 2.15142394871862e+01, 2.13781494681356e+01, 2.12538774913906e+01, 2.11412815798183e+01, 2.10400989996627e+01, 2.09499579570283e+01, 2.08703900577104e+01, 2.08008433679416e+01, 2.07406959415368e+01, 2.06892696917885e+01, 2.06458444843510e+01, 2.06096723120377e+01, 2.05799913870661e+01, 2.05560399549353e+01, 2.05370696014711e+01, 2.05223577954707e+01, 2.05112193884135e+01, 2.05030167838427e+01, 2.04971684952335e+01, 2.04931558341372e+01, 2.04905275103605e+01, 2.04889019816464e+01, 2.04879674590995e+01, 2.04874795526741e+01, 2.04872566238709e+01, 2.04871729955274e+01, 2.04871502465853e+01, 2.04871468888889e+01, 2.04871466346543e+01, 2.04871421309131e+01, 2.04871116174072e+01, 2.04869994465782e+01, 2.04867004406885e+01, 2.04860460754238e+01, 2.04847928946033e+01, 2.04826136243863e+01, 2.04790914027969e+01, 2.04737174791545e+01, 2.04658926660491e+01, 2.04549327410099e+01, 2.04400778930942e+01, 2.04205061889945e+01, 2.03953508931666e+01, 2.03637213184256e+01, 2.03247267118460e+01, 2.02775025032476e+01, 2.02212380708243e+01, 2.01552050239221e+01, 2.00787848814991e+01, 1.99914949514277e+01, 1.98930112038966e+01, 1.97831869917391e+01, 1.96620666064858e+01, 1.95298928699976e+01, 1.93871082393612e+01, 1.92343492319714e+01, 1.90724343365686e+01, 1.89023459375398e+01, 1.87252071141343e+01, 1.85422544534016e+01, 1.83548082091303e+01, 1.81642412296299e+01, 1.79719480563712e+01, 1.77793154676938e+01, 1.75876955244732e+01, 1.73983818961067e+01, 1.72125899401272e+01, 1.70314407126234e+01, 1.68559488299114e+01, 1.66870139056917e+01, 1.65254151618353e+01, 1.63718087532936e+01, 1.62267273478648e+01, 1.60905815437784e+01, 1.59636627748715e+01, 1.58461474287591e+01, 1.57381019757107e+01, 1.56394889671694e+01, 1.55501738093011e+01, 1.54699322480236e+01, 1.53984585189441e+01, 1.53353741206766e+01, 1.52802371653529e+01, 1.52325522477527e+01, 1.51917807559689e+01, 1.51573515233281e+01, 1.51286716948219e+01, 1.51051376532405e+01, 1.50861458224935e+01, 1.50711031405574e+01, 1.50594369746070e+01, 1.50506042386679e+01, 1.50440994718242e+01, 1.50394616443586e+01, 1.50362794811427e+01, 1.50341951261492e+01, 1.50329060181208e+01, 1.50321649031813e+01, 1.50317779726998e+01, 1.50316011805803e+01, 1.50315348596901e+01, 1.50315168187940e+01, 1.50315141559953e+01, 1.50315139560982e+01, 1.50315104376819e+01, 1.50314865998951e+01, 1.50313989696854e+01, 1.50311653798438e+01, 1.50306541748177e+01, 1.50296751584441e+01, 1.50279726482770e+01, 1.50252209593569e+01, 1.50210225915276e+01, 1.50149093371449e+01, 1.50063464580593e+01, 1.49947399997627e+01, 1.49794472147543e+01, 1.49597899559296e+01, 1.49350707755014e+01, 1.49045913295828e+01, 1.48676725500135e+01, 1.48236759130664e+01, 1.47720250214637e+01, 1.47122266348722e+01, 1.46438902473838e+01, 1.45667453283313e+01, 1.44806554202239e+01, 1.43856284231299e+01, 1.42818225795971e+01, 1.41695478924136e+01, 1.40492629384222e+01, 1.39215672628735e+01, 1.37871897304907e+01, 1.36469733579444e+01, 1.35018572531789e+01, 1.33528563446240e+01, 1.32010396093175e+01, 1.30475075169521e+01, 1.28933694067393e+01, 1.27397215072874e+01, 1.25876262881384e+01, 1.24380937803266e+01, 1.22920654078401e+01, 1.21504007264724e+01, 1.20138672800023e+01, 1.18831335791396e+01, 1.17587650176916e+01, 1.16412223926990e+01, 1.15308626097719e+01, 1.14279411349724e+01, 1.13326157895157e+01, 1.12449515538115e+01, 1.11649261316131e+01, 1.10924361056229e+01, 1.10273035814645e+01, 1.09692832625378e+01, 1.09180699238320e+01, 1.08733062610164e+01, 1.08345910858861e+01, 1.08014878243044e+01, 1.07735332514629e+01, 1.07502463743030e+01, 1.07311373446546e+01, 1.07157162611610e+01, 1.07035016954371e+01, 1.06940287601726e+01, 1.06868565259560e+01, 1.06815745911451e+01, 1.06778086164000e+01, 1.06752246531864e+01, 1.06735321235738e+01, 1.06724853461229e+01, 1.06718835479174e+01, 1.06715693535343e+01, 1.06714257951864e+01, 1.06713719414390e+01, 1.06713572919087e+01, 1.06713551296685e+01, 1.06713549678421e+01, 1.06713521260883e+01, 1.06713328727890e+01, 1.06712620956324e+01, 1.06710734297838e+01, 1.06706605395178e+01, 1.06698698065233e+01, 1.06684947191426e+01, 1.06662722228686e+01, 1.06628812528873e+01, 1.06579436230666e+01, 1.06510273908185e+01, 1.06416527518779e+01, 1.06293004417813e+01, 1.06134225312381e+01, 1.05934554016953e+01, 1.05688345782945e+01, 1.05390109855007e+01, 1.05034680836979e+01, 1.04617392526975e+01, 1.04134247212226e+01, 1.03582073106971e+01, 1.02958662759054e+01, 1.02262885894480e+01, 1.01494771310898e+01, 1.00655554001872e+01, 9.97476855578619e+00, 9.87748078581483e+00, 9.77416919286895e+00, 9.66541454005771e+00, 9.55188931301637e+00, 9.43434361934037e+00, 9.31358946966538e+00, 9.19048397767889e+00, 9.06591199402191e+00, 8.94076866304113e+00, 8.81594236708835e+00, 8.69229849895827e+00, 8.57066447263223e+00, 8.45181633809961e+00, 8.33646730251861e+00, 8.22525837705439e+00, 8.11875127132986e+00, 8.01742355438215e+00, 7.92166600278844e+00, 7.83178197287273e+00, 7.74798857167133e+00, 7.67041936409292e+00, 7.59912834130535e+00, 7.53409488466197e+00, 7.47522948526701e+00, 7.42238001544660e+00, 7.37533838880221e+00, 7.33384748473660e+00, 7.29760824706150e+00, 7.26628689157266e+00, 7.23952217274720e+00, 7.21693266470652e+00, 7.19812400714714e+00, 7.18269605483993e+00, 7.17024985198740e+00, 7.16039433308568e+00, 7.15275263300783e+00, 7.14696787374195e+00, 7.14270828615292e+00, 7.13967152426336e+00, 7.13758803806322e+00, 7.13622338906798e+00, 7.13537942015567e+00, 7.13489422619065e+00, 7.13464091247867e+00, 7.13452517164020e+00, 7.13448175330257e+00, 7.13446994246641e+00, 7.13446819921191e+00, 7.13446806869436e+00, 7.13446577608834e+00, 7.13445024334931e+00, 7.13439314341923e+00, 7.13424093644343e+00, 7.13390783760299e+00, 7.13326992268645e+00, 7.13216061213774e+00, 7.13036774889193e+00, 7.12763245559857e+00, 7.12364992155715e+00, 7.11807222715612e+00, 7.11051326206607e+00, 7.10055573150476e+00, 7.08776017201501e+00, 7.07167581506621e+00, 7.05185304567085e+00, 7.02785710809376e+00, 6.99928261730243e+00, 6.96576835014003e+00, 6.92701172224300e+00, 6.88278231356088e+00, 6.83293379432161e+00, 6.77741363012804e+00, 6.71627001276945e+00, 6.64965557226606e+00, 6.57782757193625e+00, 6.50114446434932e+00, 6.42005888075551e+00, 6.33510732585398e+00, 6.24689703759658e+00, 6.15609063184388e+00, 6.06338926937122e+00, 5.96951514678464e+00, 5.87519411743801e+00, 5.78113919390707e+00, 5.68803557697500e+00, 5.59652771006475e+00, 5.50720868905342e+00, 5.42061218322809e+00, 5.33720686042608e+00, 5.25739317149620e+00, 5.18150224483994e+00, 5.10979657459649e+00, 5.04247215504981e+00, 4.97966171448478e+00, 4.92143872718258e+00, 4.86782192477649e+00, 4.81878008024794e+00, 4.77423689284964e+00, 4.73407585500460e+00, 4.69814502903585e+00, 4.66626170009830e+00, 4.63821690072521e+00, 4.61377982166994e+00, 4.59270213356544e+00, 4.57472224515316e+00, 4.55956951759304e+00, 4.54696844206463e+00, 4.53664277112631e+00, 4.52831957491052e+00, 4.52173317312385e+00, 4.51662887496344e+00, 4.51276644337268e+00, 4.50992318925845e+00, 4.50789659675534e+00, 4.50650638323724e+00, 4.50559590783668e+00, 4.50503285937602e+00, 4.50470917785132e+00, 4.50454019140081e+00, 4.50446298111784e+00, 4.50443401704045e+00, 4.50442613812474e+00, 4.50442497521384e+00, 4.50442488791568e+00, 4.50442335139499e+00, 4.50441294124952e+00, 4.50437467257737e+00, 4.50427266329845e+00, 4.50404942331062e+00, 4.50362191051266e+00, 4.50287852001511e+00, 4.50167715864021e+00, 4.49984453985164e+00, 4.49717681305451e+00, 4.49344161382086e+00, 4.48838158597047e+00, 4.48171938082480e+00, 4.47316408241304e+00, 4.46241894028358e+00, 4.44919021589772e+00, 4.43319686833758e+00, 4.41418072617537e+00, 4.39191672238950e+00, 4.36622271669948e+00, 4.33696840312249e+00, 4.30408280715815e+00, 4.26755992143791e+00, 4.22746211185058e+00, 4.18392104448948e+00, 4.13713602900464e+00, 4.08736983376028e+00, 4.03494218753808e+00, 3.98022132556659e+00, 3.92361405005229e+00, 3.86555484641014e+00, 3.80649462052568e+00, 3.74688960003211e+00, 3.68719087976736e+00, 3.62783499863350e+00, 3.56923582478530e+00, 3.51177791148155e+00, 3.45581137849336e+00, 3.40164828220042e+00, 3.34956036643555e+00, 3.29977803740366e+00, 3.25249037850838e+00, 3.20784601174864e+00, 3.16595461765003e+00, 3.12688894144336e+00, 3.09068713567102e+00, 3.05735531543157e+00, 3.02687022952647e+00, 2.99918197691053e+00, 2.97421672163046e+00, 2.95187937984788e+00, 2.93205626891422e+00, 2.91461772041556e+00, 2.89942066650714e+00, 2.88631121180197e+00, 2.87512720185747e+00, 2.86570079439400e+00, 2.85786103142914e+00, 2.85143640032999e+00, 2.84625736032212e+00, 2.84215879930371e+00, 2.83898237500840e+00, 2.83657868573863e+00, 2.83480921004782e+00, 2.83354795268941e+00, 2.83268273639284e+00, 2.83211608574098e+00, 2.83176566039031e+00, 2.83156420949650e+00, 2.83145903655234e+00, 2.83141098275399e+00, 2.83139295621041e+00, 2.83138805256247e+00, 2.83138732879463e+00, 2.83138727423096e+00, 2.83138631078214e+00, 2.83137978327957e+00, 2.83135578756409e+00, 2.83129182436839e+00, 2.83115184538021e+00, 2.83088377998830e+00, 2.83041764680550e+00, 2.82966434503069e+00, 2.82851520994008e+00, 2.82684240335896e+00, 2.82450019264409e+00, 2.82132714921953e+00, 2.81714926896721e+00, 2.81178398119687e+00, 2.80504497075655e+00, 2.79674769043201e+00, 2.78671539072279e+00, 2.77478544521998e+00, 2.76081570701827e+00, 2.74469060028989e+00, 2.72632663660830e+00, 2.70567705213297e+00, 2.68273529183961e+00, 2.65753712059404e+00, 2.63016121516272e+00, 2.60072818060662e+00, 2.56939803108865e+00, 2.53636626987379e+00, 2.50185878711068e+00, 2.46612585895078e+00, 2.42943557201063e+00, 2.39206701035468e+00, 2.35430352846131e+00, 2.31642639637219e+00, 2.27870904811150e+00, 2.24141209867343e+00, 2.20477922603757e+00, 2.16903394981718e+00, 2.13437728283587e+00, 2.10098618968768e+00, 2.06901275847147e+00, 2.03858397773211e+00, 2.00980200805812e+00, 1.98274484387255e+00, 1.95746727265750e+00, 1.93400205346285e+00, 1.91236125195176e+00, 1.89253768398667e+00, 1.87450643298627e+00, 1.85822641755151e+00, 1.84364199498536e+00, 1.83068459327965e+00, 1.81927436892401e+00, 1.80932189054273e+00, 1.80072984894115e+00, 1.79339479274708e+00, 1.78720888564062e+00, 1.78206167645542e+00, 1.77784186759406e+00, 1.77443906073138e+00, 1.77174545227556e+00, 1.76965744516343e+00, 1.76807713893631e+00, 1.76691365727196e+00, 1.76608427171445e+00, 1.76551528255095e+00, 1.76514262272030e+00, 1.76491215815282e+00, 1.76477966765142e+00, 1.76471049675422e+00, 1.76467889224764e+00, 1.76466703634618e+00, 1.76466381125730e+00, 1.76466333524107e+00, 1.76466329915970e+00, 1.76466265946852e+00, 1.76465832546904e+00, 1.76464239326511e+00, 1.76459992414002e+00, 1.76450698292475e+00, 1.76432899484369e+00, 1.76401948917652e+00, 1.76351929163437e+00, 1.76275621972655e+00, 1.76164532606391e+00, 1.76008972281542e+00, 1.75798200567652e+00, 1.75520627710755e+00, 1.75164074596186e+00, 1.74716085413754e+00, 1.74164285140859e+00, 1.73496770872577e+00, 1.72702523036742e+00, 1.71771819930932e+00, 1.70696637130881e+00, 1.69471012457005e+00, 1.68091357594661e+00, 1.66556699276489e+00, 1.64868836131058e+00, 1.63032401689035e+00, 1.61054829265908e+00, 1.58946220041850e+00, 1.56719121116065e+00, 1.54388225132386e+00, 1.51970006857736e+00, 1.49482314593785e+00, 1.46943935427939e+00, 1.44374153147599e+00, 1.41792316333508e+00, 1.39217431964567e+00, 1.36667797082642e+00, 1.34160677946358e+00, 1.31712042882512e+00, 1.29336351920289e+00, 1.27046403427731e+00, 1.24853235487488e+00, 1.22766077741034e+00, 1.20792347953298e+00, 1.18937686622153e+00, 1.17206022561248e+00, 1.15599662468254e+00, 1.14119397973521e+00, 1.12764624447626e+00, 1.11533466822826e+00, 1.10422908746700e+00, 1.09428922439708e+00, 1.08546597590353e+00, 1.07770268429593e+00, 1.07093638737081e+00, 1.06509904922793e+00, 1.06011877493773e+00, 1.05592101169200e+00, 1.05242973674429e+00, 1.04956862865269e+00, 1.04726221356841e+00, 1.04543697312792e+00, 1.04402239549696e+00, 1.04295194686290e+00, 1.04216393769713e+00, 1.04160225683076e+00, 1.04121694707167e+00, 1.04096459883516e+00, 1.04080854296696e+00, 1.04071883033376e+00, 1.04067199341313e+00, 1.04065059350189e+00, 1.04064256569918e+00, 1.04064038194624e+00, 1.04064005962903e+00, 1.04064003504280e+00, 1.04063959710535e+00, 1.04063663001709e+00, 1.04062572272283e+00, 1.04059664815778e+00, 1.04053302058806e+00, 1.04041117156325e+00, 1.04019929193316e+00, 1.03985688335707e+00, 1.03933455793770e+00, 1.03857421861402e+00, 1.03750964520035e+00, 1.03606750088014e+00, 1.03416876106526e+00, 1.03173055055206e+00, 1.02866835594395e+00, 1.02489855893359e+00, 1.02034121335139e+00, 1.01492296652313e+00, 1.00858000551091e+00, 1.00126089356485e+00, 9.92929153896995e-01, 9.83565458677042e-01, 9.73169292236065e-01, 9.61759979182979e-01, 9.49376999679906e-01, 9.36079553496995e-01, 9.21945378654591e-01, 9.07068875708141e-01, 8.91558630985578e-01, 8.75534467490120e-01, 8.59124177538583e-01, 8.42460104395158e-01, 8.25675740367153e-01, 8.08902496628860e-01, 7.92266777272943e-01, 7.75887459580609e-01, 7.59873847658334e-01, 7.44324130940740e-01, 7.29324345827450e-01, 7.14947810467777e-01, 7.01254981120677e-01, 6.88293664344769e-01, 6.76099512393179e-01, 6.64696728786957e-01, 6.54098915841487e-01, 6.44310004430629e-01, 6.35325216991831e-01, 6.27132026360236e-01, 6.19711084374124e-01, 6.13037104516664e-01, 6.07079691623987e-01, 6.01804118615767e-01, 5.97172055199572e-01, 5.93142256608793e-01, 5.89671221792003e-01, 5.86713830274505e-01, 5.84223965394685e-01, 5.82155129039904e-01, 5.80461049658091e-01, 5.79096281509905e-01, 5.78016789185362e-01, 5.77180507675756e-01, 5.76547865103386e-01, 5.76082252872425e-01, 5.75750426766263e-01, 5.75522822549644e-01, 5.75373771001537e-01, 5.75281599958863e-01, 5.75228614723144e-01, 5.75200952804060e-01, 5.75188314107123e-01, 5.75183572938002e-01, 5.75182283229061e-01, 5.75182092870868e-01, 5.75182078247956e-01, 5.75181816436860e-01, 5.75180042630336e-01, 5.75173521963443e-01, 5.75156140516974e-01, 5.75118102952644e-01, 5.75045261471139e-01, 5.74918605183491e-01, 5.74713936989106e-01, 5.74401762613071e-01, 5.73947413220075e-01, 5.73311418544407e-01, 5.72450141277841e-01, 5.71316675125694e-01, 5.69861998236249e-01, 5.68036360671748e-01, 5.65790869641170e-01, 5.63079220215208e-01, 5.59859503492618e-01, 5.56096010348746e-01, 5.51760938844218e-01, 5.46835908939393e-01, 5.41313190859000e-01, 5.35196564156749e-01, 5.28501743234173e-01, 5.21256330728008e-01, 5.13499290732931e-01, 5.05279966381480e-01, 4.96656697502893e-01, 4.87695120581911e-01, 4.78466252208747e-01, 4.69044466781398e-01, 4.59505478755815e-01, 4.49924429904442e-01, 4.40374164639437e-01, 4.30923754069681e-01, 4.21637305005852e-01, 4.12573066391502e-01, 4.03782824895970e-01, 3.95311565189971e-01, 3.87197359441615e-01, 3.79471444772314e-01, 3.72158446159432e-01, 3.65276704572082e-01, 3.58838674847014e-01, 3.52851363880937e-01, 3.47316786237791e-01, 3.42232420581219e-01, 3.37591656012293e-01, 3.33384222182993e-01, 3.29596600878240e-01, 3.26212419621122e-01, 3.23212829820555e-01, 3.20576873138478e-01, 3.18281840197876e-01, 3.16303625573802e-01, 3.14617082292330e-01, 3.13196377891614e-01, 3.12015352566490e-01, 3.11047878129671e-01, 3.10268214599716e-01, 3.09651359311097e-01, 3.09173381684985e-01, 3.08811735359875e-01, 3.08545538401880e-01, 3.08355811920570e-01, 3.08225667686578e-01, 3.08140436312776e-01, 3.08087729193751e-01, 3.08057429604311e-01, 3.08041611001428e-01, 3.08034383468346e-01, 3.08031672190126e-01, 3.08030934658504e-01, 3.08030825800444e-01, 3.08030817374387e-01, 3.08030665681954e-01, 3.08029637944479e-01, 3.08025859889163e-01, 3.08015789106527e-01, 3.07993750082964e-01, 3.07951545160485e-01, 3.07878158093281e-01, 3.07759565607107e-01, 3.07578670835973e-01, 3.07315370527682e-01, 3.06946765237669e-01, 3.06447518112869e-01, 3.05790363088619e-01, 3.04946757256893e-01, 3.03887664852873e-01, 3.02584452002669e-01, 3.01009862578873e-01, 2.99139036976804e-01, 2.96950528299647e-01, 2.94427265372242e-01, 2.91557410181483e-01, 2.88335059520413e-01, 2.84760747142450e-01, 2.80841713415676e-01, 2.76591923531143e-01, 2.72031831455465e-01, 2.67187903368807e-01, 2.62091929522910e-01, 2.56780165698970e-01, 2.51292353583480e-01, 2.45670672868734e-01, 2.39958676846705e-01, 2.34200258404400e-01, 2.28438685730069e-01, 2.22715737909525e-01, 2.17070961036065e-01, 2.11541056344416e-01, 2.06159403807064e-01, 2.00955717963573e-01, 1.95955827654444e-01, 1.91181567782841e-01, 1.86650769063711e-01, 1.82377330667833e-01, 1.78371360458392e-01, 1.74639367942952e-01, 1.71184496019364e-01, 1.68006779009431e-01, 1.65103416281887e-01, 1.62469052831732e-01, 1.60096060341907e-01, 1.57974814342204e-01, 1.56093964951502e-01, 1.54440700251065e-01, 1.53001002512340e-01, 1.51759898257374e-01, 1.50701703436653e-01, 1.49810264881445e-01, 1.49069198663064e-01, 1.48462125140057e-01, 1.47972899397985e-01, 1.47585834600818e-01, 1.47285914613448e-01, 1.47058991244478e-01, 1.46891960721845e-01, 1.46772913647432e-01, 1.46691252738954e-01, 1.46637773194186e-01, 1.46604701473684e-01, 1.46585689648077e-01, 1.46575764089405e-01, 1.46571229094491e-01, 1.46569527873439e-01, 1.46569065101232e-01, 1.46568996797072e-01, 1.46568991467156e-01, 1.46568894960716e-01, 1.46568241115894e-01, 1.46565837523944e-01, 1.46559430514060e-01, 1.46545409353834e-01, 1.46518558780478e-01, 1.46471870495887e-01, 1.46396423475998e-01, 1.46281341992719e-01, 1.46113839018099e-01, 1.45879351018287e-01, 1.45561767901569e-01, 1.45143758919509e-01, 1.44607191560652e-01, 1.43933635945658e-01, 1.43104942068177e-01, 1.42103871723137e-01, 1.40914761536393e-01, 1.39524188708684e-01, 1.37921607501283e-01, 1.36099922711210e-01, 1.34055966895879e-01, 1.31790851197756e-01, 1.29310165300602e-01, 1.26624010020549e-01, 1.23746855670511e-01, 1.20697229795113e-01, 1.17497248211443e-01, 1.14172012653498e-01, 1.10748905997951e-01, 1.07256821525077e-01, 1.03725365511139e-01, 1.00184072330492e-01, 9.66616680683359e-02, 9.31854127843295e-02, 8.97805439268495e-02, 8.64698352366133e-02, 8.32732777958398e-02, 8.02078831094803e-02, 7.72876020668178e-02, 7.45233481557719e-02, 7.19231086456347e-02, 6.94921243799413e-02, 6.72331179868522e-02, 6.51465516893595e-02, 6.32308987070971e-02, 6.14829154522241e-02, 5.98979046413959e-02, 5.84699619953568e-02, 5.71922015501585e-02, 5.60569569886766e-02, 5.50559586511383e-02, 5.41804878158700e-02, 5.34215111846095e-02, 5.27697992439329e-02, 5.22160323424856e-02, 5.17508980830712e-02, 5.13651830643482e-02, 5.10498612345813e-02, 5.07961801698116e-02, 5.05957455798196e-02, 5.04406033051055e-02, 5.03233171358133e-02, 5.02370400039354e-02, 5.01755755772137e-02, 5.01334270407483e-02, 5.01058299179604e-02, 5.00887661946692e-02, 5.00789576341916e-02, 5.00738370833673e-02, 5.00714975536027e-02, 5.00706199302812e-02, 5.00703811969458e-02, 5.00703459598952e-02, 5.00703431871656e-02, 5.00702926784129e-02, 5.00699504741036e-02, 5.00686925065850e-02, 5.00653393191328e-02, 5.00580014199059e-02, 5.00439502612124e-02, 5.00195207995446e-02, 4.99800513867399e-02, 4.99198664823766e-02, 4.98323072316040e-02, 4.97098141901233e-02, 4.95440652522348e-02, 4.93261700442247e-02, 4.90469195669903e-02, 4.86970867644558e-02, 4.82677699939131e-02, 4.77507673819346e-02, 4.71389660078785e-02, 4.64267262655866e-02, 4.56102390423538e-02, 4.46878320737554e-02, 4.36602024072528e-02, 4.25305547757162e-02, 4.13046308809371e-02, 3.99906219188111e-02, 3.85989651815706e-02, 3.71420341816346e-02, 3.56337391524875e-02, 3.40890603190331e-02, 3.25235402968644e-02, 3.09527648887595e-02, 2.93918635321688e-02, 2.78550605451073e-02, 2.63553037862749e-02, 2.49039874838175e-02, 2.35107725578629e-02, 2.21834964181653e-02, 2.09281592106976e-02, 1.97489756600717e-02, 1.86484858519762e-02, 1.76277190749374e-02, 1.66863982753256e-02, 1.58231636217465e-02, 1.50357908964125e-02, 1.43163580106247e-02, 1.36721049104764e-02, 1.30935531387237e-02, 1.25767218815085e-02, 1.21175165899245e-02, 1.17118143828989e-02, 1.13555287203937e-02, 1.10446562171234e-02, 1.07753087638542e-02, 1.05437341754850e-02, 1.03463284490317e-02, 1.01796424389633e-02, 1.00403853806341e-02, 9.92542724862489e-03, 9.83180145171758e-03, 9.75670886283808e-03, 9.69752368041554e-03, 9.65180113926323e-03, 9.61728665673434e-03, 9.59192563760070e-03, 9.57387289179223e-03, 9.56150045990967e-03, 9.55340260385640e-03, 9.54839680382946e-03, 9.54551979744878e-03, 9.54401798203719e-03, 9.54333184523648e-03, 9.54307446054514e-03, 9.54300444650613e-03, 9.54299411261417e-03, 9.54299329218600e-03, 9.54297825675354e-03, 9.54287638992923e-03, 9.54250192373252e-03, 9.54150378814944e-03, 9.53931966591881e-03, 9.53513785684379e-03, 9.52786892910495e-03, 9.51612919427447e-03, 9.49823808473838e-03, 9.47223147702658e-03, 9.43589281816770e-03, 9.38680349079135e-03, 9.32241312595039e-03, 9.24012949116815e-03, 9.13742614271484e-03, 9.01196429542495e-03, 8.86172346172742e-03, 8.68513355205636e-03, 8.48119957614557e-03, 8.24960913480055e-03, 7.99081281953562e-03, 7.70606864427492e-03, 7.39744378906693e-03, 7.06777013604665e-03, 6.72055403187244e-03, 6.35984495973929e-03, 5.99007179093292e-03, 5.61585844926782e-03, 5.24183270782430e-03, 4.87244219035432e-03, 4.51179045438010e-03, 4.16350350314332e-03, 3.83063359809199e-03, 3.51560331314715e-03, 3.22018891372977e-03, 2.94553884686621e-03, 2.69222076948239e-03, 2.46028929726805e-03, 2.24936647539664e-03, 2.05872761030435e-03, 1.88738624780171e-03, 1.73417346856297e-03, 1.59780812985844e-03, 1.47695610643694e-03, 1.37027787292254e-03, 1.27646480790844e-03, 1.19477329918993e-03, 1.12308218380264e-03, 1.06058539453358e-03, 1.00600232268988e-03, 9.59346588686572e-04, 9.19667947250957e-04, 8.85561009162927e-04, 8.56557570166672e-04, 8.32072540191575e-04, 8.11574454935737e-04, 7.94580088488317e-04, 7.80647330948625e-04, 7.69368020939964e-04, 7.60365191943890e-04, 7.53294441864227e-04, 7.47844602749126e-04, 7.43736233879133e-04, 7.40719520152311e-04, 7.38572758095302e-04, 7.37101593188072e-04, 7.36138690932364e-04, 7.35543436022112e-04, 7.35201309993272e-04, 7.35022713790556e-04, 7.34941117002563e-04, 7.34910508263908e-04, 7.34902181847019e-04, 7.34900953048623e-04, 7.34900854915566e-04, 7.34899049277260e-04, 7.34886815294544e-04, 7.34841844270306e-04, 7.34721979874564e-04, 7.34459722728786e-04, 7.33957708007438e-04, 7.33085440643322e-04, 7.31677576517598e-04, 7.29534037503754e-04, 7.26422207435462e-04, 7.22081415126752e-04, 7.16229992796968e-04, 7.08575432277314e-04, 6.98828571486843e-04, 6.86725046762149e-04, 6.71844635682817e-04, 6.54770415096095e-04, 6.34682791208263e-04, 6.11743260854780e-04, 5.86016975063330e-04, 5.57679823145918e-04, 5.27012133261451e-04, 4.94386968050264e-04, 4.60603503995571e-04, 4.26132429936388e-04, 3.91222204378039e-04, 3.56410223183445e-04, 3.22249435022570e-04, 2.89839978002249e-04, 2.58910874961771e-04, 2.29739015956909e-04, 2.02801036039091e-04, 1.78505966736964e-04, 1.56344815579246e-04, 1.36263873272924e-04, 1.19016289204410e-04, 1.03725473593386e-04, 9.01461295512718e-05, 7.85334079438371e-05, 6.87701244146410e-05, 6.01716686077225e-05, 5.25262238468827e-05, 4.66393105256202e-05, 4.15288641661959e-05, 3.69887365181398e-05, 3.29090114732367e-05, 2.99772571281665e-05, 2.74789327221470e-05, 2.52315627164990e-05, 2.32159015851607e-05, 2.14397236391051e-05, 2.00515175805549e-05, 1.90668844126662e-05, 1.82396818004804e-05, 1.75411686385019e-05, 1.69594305958753e-05, 1.64794385426603e-05, 1.60861668661028e-05, 1.57662919662016e-05, 1.55082595583167e-05, 1.53022404916214e-05, 1.51400808050539e-05, 1.50149900264296e-05, 1.49210676596567e-05, 1.48529185750590e-05, 1.48054876945025e-05, 1.47740889175907e-05, 1.47545287987327e-05, 1.47432325547925e-05, 1.47373196739409e-05, 1.47346145169536e-05, 1.47335991371733e-05, 1.47333228781577e-05, 1.47332821001100e-05, 1.47332788316135e-05, 1.47332183786375e-05, 1.47328088218046e-05, 1.47313034730701e-05, 1.47272923740616e-05, 1.47185223653423e-05, 1.47017576423991e-05, 1.46726982350742e-05, 1.46259748916577e-05, 1.45552349195555e-05, 1.44533191672735e-05, 1.43125153897472e-05, 1.41248864693466e-05, 1.38827544304831e-05, 1.35795823510421e-05, 1.32118327798025e-05, 1.27839861434609e-05, 1.23210184739264e-05, 1.18560219917185e-05, 1.13730995110473e-05, 1.08384641777093e-05, 1.02294683799187e-05, 9.53643014156569e-06, 8.77088331918685e-06, 8.03707063568027e-06, 7.92663935555993e-06, 7.94479362298262e-06, 7.90804537587064e-06, 7.66598295598442e-06, 7.81506945912724e-06, 8.01579274423148e-06, 8.03819413164319e-06, 8.08659688544822e-06, 9.13088290133146e-06, 9.84339428811154e-06, 9.84774204937366e-06, 9.93215147815363e-06, 9.60472794986345e-06, 8.38263962776864e-06, 6.50883106345646e-06, 4.39077091130002e-06, 2.49265950458266e-06, 1.25261587413661e-06, 9.79898558739478e-07, 9.53865118447084e-07, 9.81028129842798e-07, 1.36557578400698e-06, 1.79931999610972e-06, 1.96581945743901e-06, 1.81642129604143e-06, 1.43508421501792e-06, 9.91556355245393e-07, 7.39395105142648e-07, 7.29824328971063e-07, 6.79612142562291e-07, 5.83512758112192e-07, 4.97057570214565e-07, 4.25008714444175e-07, -2.84217094304040e-14, -2.84217094304040e-14, 3.62277646104303e-07, 3.59350877943143e-07, -2.84217094304040e-14, -3.55271367880050e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -3.55271367880050e-14, -2.84217094304040e-14, -2.84217094304040e-14, -3.55271367880050e-14, -2.84217094304040e-14, -2.84217094304040e-14, -3.55271367880050e-14],
[ 5.10000000000000e+01, 5.09999999523397e+01, 5.09999984785894e+01, 5.09999884936976e+01, 5.09999517879364e+01, 5.09998539419874e+01, 5.09996397990510e+01, 5.09992296534208e+01, 5.09985162931674e+01, 5.09973630084400e+01, 5.09956026473087e+01, 5.09930377691779e+01, 5.09894419128679e+01, 5.09845619638895e+01, 5.09781215745849e+01, 5.09698255629228e+01, 5.09593651917663e+01, 5.09464242109006e+01, 5.09306855290078e+01, 5.09118383715140e+01, 5.08895857717065e+01, 5.08636522353140e+01, 5.08337914114050e+01, 5.07997935939277e+01, 5.07614928681870e+01, 5.07187737057798e+01, 5.06715768019492e+01, 5.06199039439439e+01, 5.05638217013795e+01, 5.05034637433290e+01, 5.04390316146371e+01, 5.03707938469261e+01, 5.02990833370602e+01, 5.02242929944145e+01, 5.01468697331923e+01, 5.00673069610686e+01, 4.99861357840718e+01, 4.99039152039431e+01, 4.98212216238016e+01, 4.97386379983479e+01, 4.96567429658082e+01, 4.95761002821181e+01, 4.94972488467222e+01, 4.94206935680090e+01, 4.93468972691907e+01, 4.92762737864213e+01, 4.92091823634302e+01, 4.91459234032413e+01, 4.90867355989955e+01, 4.90317944329346e+01, 4.89812120050002e+01, 4.89350381295670e+01, 4.88932626196601e+01, 4.88558186616763e+01, 4.88225871693302e+01, 4.87934019926989e+01, 4.87680558465072e+01, 4.87463068111348e+01, 4.87278852504205e+01, 4.87125009825939e+01, 4.86998505351109e+01, 4.86896243114375e+01, 4.86815134985234e+01, 4.86752165483924e+01, 4.86704450763651e+01, 4.86669290321307e+01, 4.86644210181996e+01, 4.86626996529102e+01, 4.86615719016246e+01, 4.86608743292522e+01, 4.86604732588455e+01, 4.86602638536315e+01, 4.86601681723039e+01, 4.86601322785535e+01, 4.86601225145471e+01, 4.86601210733990e+01, 4.86601208843507e+01, 4.86601164797531e+01, 4.86600866379371e+01, 4.86599769359751e+01, 4.86596845085985e+01, 4.86590445271573e+01, 4.86578188429201e+01, 4.86556872325640e+01, 4.86522416208095e+01, 4.86469835831936e+01, 4.86393253514777e+01, 4.86285944555705e+01, 4.86140420396503e+01, 4.85948547871899e+01, 4.85701702809242e+01, 4.85390955108207e+01, 4.85007281272928e+01, 4.84541799198144e+01, 4.83986018846293e+01, 4.83332101322232e+01, 4.82573117804715e+01, 4.81703298909281e+01, 4.80718264455922e+01, 4.79615223453646e+01, 4.78393134564211e+01, 4.77052818514982e+01, 4.75597015963375e+01, 4.74030387111563e+01, 4.72359452712353e+01, 4.70592479633704e+01, 4.68739317412102e+01, 4.66811194782756e+01, 4.64820486697573e+01, 4.62780462693775e+01, 4.60705026747174e+01, 4.58608457223460e+01, 4.56505153634249e+01, 4.54409395027305e+01, 4.52335113314239e+01, 4.50295683828042e+01, 4.48303734898407e+01, 4.46370978090855e+01, 4.44508060765407e+01, 4.42724442571409e+01, 4.41028297275407e+01, 4.39426440882282e+01, 4.37924286405240e+01, 4.36525824967644e+01, 4.35233632287080e+01, 4.34048899079982e+01, 4.32971483569667e+01, 4.31999984073446e+01, 4.31131829546917e+01, 4.30363385924007e+01, 4.29690076061516e+01, 4.29106511041323e+01, 4.28606630483879e+01, 4.28183849381123e+01, 4.27831208777072e+01, 4.27541527430706e+01, 4.27307551414812e+01, 4.27122098464454e+01, 4.26978193817657e+01, 4.26869194313111e+01, 4.26788897644764e+01, 4.26731633932952e+01, 4.26692337159965e+01, 4.26666594528727e+01, 4.26650672421634e+01, 4.26641518338858e+01, 4.26636738951464e+01, 4.26634555179928e+01, 4.26633735966985e+01, 4.26633513120744e+01, 4.26633480229089e+01, 4.26633477325382e+01, 4.26633420428671e+01, 4.26633034944981e+01, 4.26631617864670e+01, 4.26627840437415e+01, 4.26619573572273e+01, 4.26603741314328e+01, 4.26576208189474e+01, 4.26531705472418e+01, 4.26463800581527e+01, 4.26364912824840e+01, 4.26226377595513e+01, 4.26038559822029e+01, 4.25791016005133e+01, 4.25472702520271e+01, 4.25072226053629e+01, 4.24578130121850e+01, 4.23979209680799e+01, 4.23264843968219e+01, 4.22425336084106e+01, 4.21452246542434e+01, 4.20338707283432e+01, 4.19079702563124e+01, 4.17672303859717e+01, 4.16115847539682e+01, 4.14412046534018e+01, 4.12565030617120e+01, 4.10581313852369e+01, 4.08469692001306e+01, 4.06241076667352e+01, 4.03908276078569e+01, 4.01485734227161e+01, 3.98989240385727e+01, 3.96435620019147e+01, 3.93842416361532e+01, 3.91227570117957e+01, 3.88609103427042e+01, 3.86004813592729e+01, 3.83431982013664e+01, 3.80907103837442e+01, 3.78445643756175e+01, 3.76061822792811e+01, 3.73768439863602e+01, 3.71576730476027e+01, 3.69496263361903e+01, 3.67534874386060e+01, 3.65698635887227e+01, 3.63991858789405e+01, 3.62417124375392e+01, 3.60975342483930e+01, 3.59665832988669e+01, 3.58486427642136e+01, 3.57433589631618e+01, 3.56502548426136e+01, 3.55687447647282e+01, 3.54981503747064e+01, 3.54377173217764e+01, 3.53866325902864e+01, 3.53440421747035e+01, 3.53090688048433e+01, 3.52808293995374e+01, 3.52584519021771e+01, 3.52410911341219e+01, 3.52279432954724e+01, 3.52182587501755e+01, 3.52113527559109e+01, 3.52066138396162e+01, 3.52035095764349e+01, 3.52015896016067e+01, 3.52004857684202e+01, 3.51999094569030e+01, 3.51996461328728e+01, 3.51995473505275e+01, 3.51995204792956e+01, 3.51995165131567e+01, 3.51995161950233e+01, 3.51995103238482e+01, 3.51994705457655e+01, 3.51993243172111e+01, 3.51989345246913e+01, 3.51980814688233e+01, 3.51964477476459e+01, 3.51936066391819e+01, 3.51890145085971e+01, 3.51820076789043e+01, 3.51718041044596e+01, 3.51575100712297e+01, 3.51381320146223e+01, 3.51125933929105e+01, 3.50797563818581e+01, 3.50384479663450e+01, 3.49874898029402e+01, 3.49257310221844e+01, 3.48520829428686e+01, 3.47655544976206e+01, 3.46652870360786e+01, 3.45505870953713e+01, 3.44209557222261e+01, 3.42761130076170e+01, 3.41160166583568e+01, 3.39408736778103e+01, 3.37511445486330e+01, 3.35475396838684e+01, 3.33310083105694e+01, 3.31027203383979e+01, 3.28640421087378e+01, 3.26165071855379e+01, 3.23617835142460e+01, 3.21016383303498e+01, 3.18379021505865e+01, 3.15724330487075e+01, 3.13070822342880e+01, 3.10436617502686e+01, 3.07839149103354e+01, 3.05294899279391e+01, 3.02819170495032e+01, 3.00425893901564e+01, 2.98127475713509e+01, 2.95934681672131e+01, 2.93856558765786e+01, 2.91900392525971e+01, 2.90071697480589e+01, 2.88374237795690e+01, 2.86810074824622e+01, 2.85379638216353e+01, 2.84081817376271e+01, 2.82914070354233e+01, 2.81872547574776e+01, 2.80952228146718e+01, 2.80147066734480e+01, 2.79450149104427e+01, 2.78853854462343e+01, 2.78350022578380e+01, 2.77930123474940e+01, 2.77585427163913e+01, 2.77307170602982e+01, 2.77086718740964e+01, 2.76915716284446e+01, 2.76786226684215e+01, 2.76690854845557e+01, 2.76622850236932e+01, 2.76576187419481e+01, 2.76545621544149e+01, 2.76526717047648e+01, 2.76515848593532e+01, 2.76510174209687e+01, 2.76507581520555e+01, 2.76506608910718e+01, 2.76506344337011e+01, 2.76506305286477e+01, 2.76506302274524e+01, 2.76506248189516e+01, 2.76505881755593e+01, 2.76504534705203e+01, 2.76500943959387e+01, 2.76493085679459e+01, 2.76478036054613e+01, 2.76451864384539e+01, 2.76409563347013e+01, 2.76345020564251e+01, 2.76251034680860e+01, 2.76119378115157e+01, 2.75940907418325e+01, 2.75705720752918e+01, 2.75403360379966e+01, 2.75023056242756e+01, 2.74554004806104e+01, 2.73985675338012e+01, 2.73308133924553e+01, 2.72512373835299e+01, 2.71590639568337e+01, 2.70536731164652e+01, 2.69346275336254e+01, 2.68016950706879e+01, 2.66548656065728e+01, 2.64943612955449e+01, 2.63206397040405e+01, 2.61343896324307e+01, 2.59365198118666e+01, 2.57281410358155e+01, 2.55105426055834e+01, 2.52851642079594e+01, 2.50535644816493e+01, 2.48173875643305e+01, 2.45783288578857e+01, 2.43381011319838e+01, 2.40984019358986e+01, 2.38608831303978e+01, 2.36271231998611e+01, 2.33986028621426e+01, 2.31766843553571e+01, 2.29625946407979e+01, 2.27574126176317e+01, 2.25620603019909e+01, 2.23772977897962e+01, 2.22037217102251e+01, 2.20417667947163e+01, 2.18917101400952e+01, 2.17536777338685e+01, 2.16276528301809e+01, 2.15134858081722e+01, 2.14109052007350e+01, 2.13195296412999e+01, 2.12388805309681e+01, 2.11683952719018e+01, 2.11074409415727e+01, 2.10553282947769e+01, 2.10113259767846e+01, 2.09746748137970e+01, 2.09446020194824e+01, 2.09203351230743e+01, 2.09011153901428e+01, 2.08862104766095e+01, 2.08749260344985e+01, 2.08666159782920e+01, 2.08606911265607e+01, 2.08566259564647e+01, 2.08539632489926e+01, 2.08523164591332e+01, 2.08513697148571e+01, 2.08508754279700e+01, 2.08506495839684e+01, 2.08505648620474e+01, 2.08505418156271e+01, 2.08505384140236e+01, 2.08505381564644e+01, 2.08505335938299e+01, 2.08505026813130e+01, 2.08503890436875e+01, 2.08500861279326e+01, 2.08494232064075e+01, 2.08481536407960e+01, 2.08459458818929e+01, 2.08423776273632e+01, 2.08369334979014e+01, 2.08290065195890e+01, 2.08179036137518e+01, 2.08028551922063e+01, 2.07830288330501e+01, 2.07575468692588e+01, 2.07255075607724e+01, 2.06860093449956e+01, 2.06381774785382e+01, 2.05811922057147e+01, 2.05143174306649e+01, 2.04369287455027e+01, 2.03485395923542e+01, 2.02488243266005e+01, 2.01376370126432e+01, 2.00150249272959e+01, 1.98812359679966e+01, 1.97367194543545e+01, 1.95821201553256e+01, 1.94182657466766e+01, 1.92461482750516e+01, 1.90669005436142e+01, 1.88817686082020e+01, 1.86920817550947e+01, 1.84992214036328e+01, 1.83045903330950e+01, 1.81095834816262e+01, 1.79155613273215e+01, 1.77238265698834e+01, 1.75356045227375e+01, 1.73520273361143e+01, 1.71741219304149e+01, 1.70028013442514e+01, 1.68388590985200e+01, 1.66829661411232e+01, 1.65356699527642e+01, 1.63973954448973e+01, 1.62684473486988e+01, 1.61490138640602e+01, 1.60391713999020e+01, 1.59388902860942e+01, 1.58480413714262e+01, 1.57664034424710e+01, 1.56936714071344e+01, 1.56294651866456e+01, 1.55733392527324e+01, 1.55247927341342e+01, 1.54832799993371e+01, 1.54482216012065e+01, 1.54190154449373e+01, 1.53950480147177e+01, 1.53757054685132e+01, 1.53603843866671e+01, 1.53485019411543e+01, 1.53395052409167e+01, 1.53328796071252e+01, 1.53281555422691e+01, 1.53249141796884e+01, 1.53227910355834e+01, 1.53214779326360e+01, 1.53207230211382e+01, 1.53203288870697e+01, 1.53201488034485e+01, 1.53200812477825e+01, 1.53200628709956e+01, 1.53200601586200e+01, 1.53200599550011e+01, 1.53200563710778e+01, 1.53200320894704e+01, 1.53199428277180e+01, 1.53197048886974e+01, 1.53191841651680e+01, 1.53181869180681e+01, 1.53164526993421e+01, 1.53136497470857e+01, 1.53093731307450e+01, 1.53031458654699e+01, 1.52944231453795e+01, 1.52825997632362e+01, 1.52670206871212e+01, 1.52469946526769e+01, 1.52218105037795e+01, 1.51907558791051e+01, 1.51531377038116e+01, 1.51083038141302e+01, 1.50556649298824e+01, 1.49947161086623e+01, 1.49250567777701e+01, 1.48464084552860e+01, 1.47586293444529e+01, 1.46617251139493e+01, 1.45558553518431e+01, 1.44413353880429e+01, 1.43186333998806e+01, 1.41883629285428e+01, 1.40512711241526e+01, 1.39082231947371e+01, 1.37601836574994e+01, 1.36081950850308e+01, 1.34533551125798e+01, 1.32967925308010e+01, 1.31396433293065e+01, 1.29830275678250e+01, 1.28280279152352e+01, 1.26756705950168e+01, 1.25269093031031e+01, 1.23826124336964e+01, 1.22435536918827e+01, 1.21104059307068e+01, 1.19837378637208e+01, 1.18640131959524e+01, 1.17515916901284e+01, 1.16467317257985e+01, 1.15495939911572e+01, 1.14602460440528e+01, 1.13786675687789e+01, 1.13047562262944e+01, 1.12383340432281e+01, 1.11791543107594e+01, 1.11269089724668e+01, 1.10812364753350e+01, 1.10417300445070e+01, 1.10079463231716e+01, 1.09794142963879e+01, 1.09556443932674e+01, 1.09361376372058e+01, 1.09203946902614e+01, 1.09079246170421e+01, 1.08982531774214e+01, 1.08909304479710e+01, 1.08855375708711e+01, 1.08816924375569e+01, 1.08790541331894e+01, 1.08773259971599e+01, 1.08762571933824e+01, 1.08756427304391e+01, 1.08753219234030e+01, 1.08751753435608e+01, 1.08751203563275e+01, 1.08751053984598e+01, 1.08751031907096e+01, 1.08751030254771e+01, 1.08751001239111e+01, 1.08750804653746e+01, 1.08750081985126e+01, 1.08748155615747e+01, 1.08743939802634e+01, 1.08735866012928e+01, 1.08721825614600e+01, 1.08699132570089e+01, 1.08664508372962e+01, 1.08614090990678e+01, 1.08543469013319e+01, 1.08447741541190e+01, 1.08321603559775e+01, 1.08159455644784e+01, 1.07955535824564e+01, 1.07704070333718e+01, 1.07399438875763e+01, 1.07036348951959e+01, 1.06610012905024e+01, 1.06116320676075e+01, 1.05552000982461e+01, 1.04914763772665e+01, 1.04203417441952e+01, 1.03417955384162e+01, 1.02559607934252e+01, 1.01630857489690e+01, 1.00635416415733e+01, 9.95781690626347e+00, 9.84650807044921e+00, 9.73030773616625e+00, 9.60999012766045e+00, 9.48639473242423e+00, 9.36040859333198e+00, 9.23294782495833e+00, 9.10493893227671e+00, 8.97730050358646e+00, 8.85092582675367e+00, 8.72666693271114e+00, 8.60532049884289e+00, 8.48761594806770e+00, 8.37420596300660e+00, 8.26565950862904e+00, 8.16245733333773e+00, 8.06498980924285e+00, 7.97355688651042e+00, 7.88836987943134e+00, 7.80955477437919e+00, 7.73715674957576e+00, 7.67114561837021e+00, 7.61142194489383e+00, 7.55782362644947e+00, 7.51013278432465e+00, 7.46808284847115e+00, 7.43136575766489e+00, 7.39963922282239e+00, 7.37253401598822e+00, 7.34966125125688e+00, 7.33061961773988e+00, 7.31500251065351e+00, 7.30240498731322e+00, 7.29243045327267e+00, 7.28469696316513e+00, 7.27884300398559e+00, 7.27453261821216e+00, 7.27145972232003e+00, 7.26935148411200e+00, 7.26797064020435e+00, 7.26711666235747e+00, 7.26662571666327e+00, 7.26636940073062e+00, 7.26625228831262e+00, 7.26620835547405e+00, 7.26619640468359e+00, 7.26619464077221e+00, 7.26619450870809e+00, 7.26619218893563e+00, 7.26617647214109e+00, 7.26611869561938e+00, 7.26596468522102e+00, 7.26562764025508e+00, 7.26498217041419e+00, 7.26385972922944e+00, 7.26204566335807e+00, 7.25927806749476e+00, 7.25524859946908e+00, 7.24960536729448e+00, 7.24195794763835e+00, 7.23188453184673e+00, 7.21894112061176e+00, 7.20267260227434e+00, 7.18262545495585e+00, 7.15836171342473e+00, 7.12947374388049e+00, 7.09559928127300e+00, 7.05643611294396e+00, 7.01175574805595e+00, 6.96141540245997e+00, 6.90536765954904e+00, 6.84366724264073e+00, 6.77647445327609e+00, 6.70405498804787e+00, 6.62677603533948e+00, 6.54509875977295e+00, 6.45956749017298e+00, 6.37079611869965e+00, 6.27945237697312e+00, 6.18624076447710e+00, 6.09188495488301e+00, 5.99711049309860e+00, 5.90262852284372e+00, 5.80912116121542e+00, 5.71722897817346e+00, 5.62754086330173e+00, 5.54058638799210e+00, 5.45683061446157e+00, 5.37667117564722e+00, 5.30043735872302e+00, 5.22839087127377e+00, 5.16072795019794e+00, 5.09758248330215e+00, 5.03902984480296e+00, 4.98509119078633e+00, 4.93573801199887e+00, 4.89089679333804e+00, 4.85045367774093e+00, 4.81425907395452e+00, 4.78213218124139e+00, 4.75386542872013e+00, 4.72922884273348e+00, 4.70797436282960e+00, 4.68984012641966e+00, 4.67455473498246e+00, 4.66184150208792e+00, 4.65142266697541e+00, 4.64302353860751e+00, 4.63637651581934e+00, 4.63122491127053e+00, 4.62732649220789e+00, 4.62445664121288e+00, 4.62241103646078e+00, 4.62100775440933e+00, 4.62008870853432e+00, 4.61952035537029e+00, 4.61919362271281e+00, 4.61902304287381e+00, 4.61894510446204e+00, 4.61891586722280e+00, 4.61890791399897e+00, 4.61890674012029e+00, 4.61890665199880e+00, 4.61890510098669e+00, 4.61889459265903e+00, 4.61885596304951e+00, 4.61875299157320e+00, 4.61852764548017e+00, 4.61809609785998e+00, 4.61734568645664e+00, 4.61613296583689e+00, 4.61428298794308e+00, 4.61158992430725e+00, 4.60781911532701e+00, 4.60271059601168e+00, 4.59598410181989e+00, 4.58734550173583e+00, 4.57649453894906e+00, 4.56313368447250e+00, 4.54697782973776e+00, 4.52776446655597e+00, 4.50526393433352e+00, 4.47928926350647e+00, 4.44970511912515e+00, 4.41643535635193e+00, 4.37946874463700e+00, 4.33886250003437e+00, 4.29474338153921e+00, 4.24730624896016e+00, 4.19681013434427e+00, 4.14357203164919e+00, 4.08795874514488e+00, 4.03037724272434e+00, 3.97126402652786e+00, 3.91107405581714e+00, 3.85026973719839e+00, 3.78931044135891e+00, 3.72864292322887e+00, 3.66869292530729e+00, 3.60985814289614e+00, 3.55250263446995e+00, 3.49695267699092e+00, 3.44349399843919e+00, 3.39237026946275e+00, 3.34378270235297e+00, 3.29789058692217e+00, 3.25481258717348e+00, 3.21462862762419e+00, 3.17738221147066e+00, 3.14308303221136e+00, 3.11170976366886e+00, 3.08321293845958e+00, 3.05751784989998e+00, 3.03452743540930e+00, 3.01412511930540e+00, 2.99617760850828e+00, 2.98053764548369e+00, 2.96704672857583e+00, 2.95553781083546e+00, 2.94583798499391e+00, 2.93777115505895e+00, 2.93116068502896e+00, 2.92583200351515e+00, 2.92161513082613e+00, 2.91834708355515e+00, 2.91587410214098e+00, 2.91405364035312e+00, 2.91275605306073e+00, 2.91186592055918e+00, 2.91128295433618e+00, 2.91092244024018e+00, 2.91071518994008e+00, 2.91060698936335e+00, 2.91055755225581e+00, 2.91053900679189e+00, 2.91053396198564e+00, 2.91053321738315e+00, 2.91053316124879e+00, 2.91053217006579e+00, 2.91052545466035e+00, 2.91050076819857e+00, 2.91043496376183e+00, 2.91029095544019e+00, 2.91001517407337e+00, 2.90953562487768e+00, 2.90876064499383e+00, 2.90757844801289e+00, 2.90585752882367e+00, 2.90344798442239e+00, 2.90018378330915e+00, 2.89588598664730e+00, 2.89036688787927e+00, 2.88343499421273e+00, 2.87490072465344e+00, 2.86458264766825e+00, 2.85231403097330e+00, 2.83794943125186e+00, 2.82137101828680e+00, 2.80249431143086e+00, 2.78127301102825e+00, 2.75770263612820e+00, 2.73182273298086e+00, 2.70371749396461e+00, 2.67351471853637e+00, 2.64138314891848e+00, 2.60752831437820e+00, 2.57218710951513e+00, 2.53562140513387e+00, 2.49811103817264e+00, 2.45994654579102e+00, 2.42142199750351e+00, 2.38282824106219e+00, 2.34444681851509e+00, 2.30654473643233e+00, 2.26937019743436e+00, 2.23314932708737e+00, 2.19808386749649e+00, 2.16434976070106e+00, 2.13209651287147e+00, 2.10144721367583e+00, 2.07249908175726e+00, 2.04532441389249e+00, 2.01997182876572e+00, 1.99646771337711e+00, 1.97481779849939e+00, 1.95500880755174e+00, 1.93701013964084e+00, 1.92077556166124e+00, 1.90624489592009e+00, 1.89334569862801e+00, 1.88199493077193e+00, 1.87210062642311e+00, 1.86356356455367e+00, 1.85627894912101e+00, 1.85013809878657e+00, 1.84503014251553e+00, 1.84084371090532e+00, 1.83746860596737e+00, 1.83479742486955e+00, 1.83272710652190e+00, 1.83116036453829e+00, 1.83000696666112e+00, 1.82918481970114e+00, 1.82862082074919e+00, 1.82825143996578e+00, 1.82802300750525e+00, 1.82789168669795e+00, 1.82782312690257e+00, 1.82779180170846e+00, 1.82778005060212e+00, 1.82777685402151e+00, 1.82777638221310e+00, 1.82777634645067e+00, 1.82777571241414e+00, 1.82777141672648e+00, 1.82775562537037e+00, 1.82771353176706e+00, 1.82762141274350e+00, 1.82744500068132e+00, 1.82713824017806e+00, 1.82664249150714e+00, 1.82588623596834e+00, 1.82478532928172e+00, 1.82324383772033e+00, 1.82115547688223e+00, 1.81840565417535e+00, 1.81487409287227e+00, 1.81043798815659e+00, 1.80497561481835e+00, 1.79837027386355e+00, 1.79051443379465e+00, 1.78131389486420e+00, 1.77069178470445e+00, 1.75859218477653e+00, 1.74498319176458e+00, 1.72985923783676e+00, 1.71324252841538e+00, 1.69518350367868e+00, 1.67576028663656e+00, 1.65507714112984e+00, 1.63326202173150e+00, 1.61046334873067e+00, 1.58684618063732e+00, 1.56258798106232e+00, 1.53787418541233e+00, 1.51289376643067e+00, 1.48783497851598e+00, 1.46288143219163e+00, 1.43820861567043e+00, 1.41398094363104e+00, 1.39034937710744e+00, 1.36744962517524e+00, 1.34540091060422e+00, 1.32430525889466e+00, 1.30424725360801e+00, 1.28529419062718e+00, 1.26749655951899e+00, 1.25088878078132e+00, 1.23549013248164e+00, 1.22130580755017e+00, 1.20832805267126e+00, 1.19653735027251e+00, 1.18590361561607e+00, 1.17638739069055e+00, 1.16794102490805e+00, 1.16050983913801e+00, 1.15403327413723e+00, 1.14844602689691e+00, 1.14367917889125e+00, 1.13966131886043e+00, 1.13631965987789e+00, 1.13358114641051e+00, 1.13137354231983e+00, 1.12962648576794e+00, 1.12827249229769e+00, 1.12724788345970e+00, 1.12649361571555e+00, 1.12595598332143e+00, 1.12558716972602e+00, 1.12534562477948e+00, 1.12519624965297e+00, 1.12511037755171e+00, 1.12506554566772e+00, 1.12504506185893e+00, 1.12503737771540e+00, 1.12503528744585e+00, 1.12503497892659e+00, 1.12503495539287e+00, 1.12503453620290e+00, 1.12503169613147e+00, 1.12502125576253e+00, 1.12499342583388e+00, 1.12493252204403e+00, 1.12481588909036e+00, 1.12461307933270e+00, 1.12428532760750e+00, 1.12378535910062e+00, 1.12305756141043e+00, 1.12203854446159e+00, 1.12065810219963e+00, 1.11884057752826e+00, 1.11650661648871e+00, 1.11357527932226e+00, 1.10996645537251e+00, 1.10560350687802e+00, 1.10041604522060e+00, 1.09434272420848e+00, 1.08733392081569e+00, 1.07935416675684e+00, 1.07038419625881e+00, 1.06042248757414e+00, 1.04948619831347e+00, 1.03761142650963e+00, 1.02485276821574e+00, 1.01128218513483e+00, 9.96987238392144e-01, 9.82068783064904e-01, 9.66638248838201e-01, 9.50814652393831e-01, 9.34721495322179e-01, 9.18483697328398e-01, 9.02224699484549e-01, 8.86063848573563e-01, 8.70114144226839e-01, 8.54480398902631e-01, 8.39257829952118e-01, 8.24531075700939e-01, 8.10373605486049e-01, 7.96847477903960e-01, 7.84003392288575e-01, 7.71880975102015e-01, 7.60509244478968e-01, 7.49907201357431e-01, 7.40084503158968e-01, 7.31042184665050e-01, 7.22773399594118e-01, 7.15264164684271e-01, 7.08494095337009e-01, 7.02437127795420e-01, 6.97062227294680e-01, 6.92334084629294e-01, 6.88213805202373e-01, 6.84659594977724e-01, 6.81627446998576e-01, 6.79071830441728e-01, 6.76946381743640e-01, 6.75204594387729e-01, 6.73800500729534e-01, 6.72689336028490e-01, 6.71828171934330e-01, 6.71176504322766e-01, 6.70696778842533e-01, 6.70354837031688e-01, 6.70120266517124e-01, 6.69966640670388e-01, 6.69871636094335e-01, 6.69817020297472e-01, 6.69788506625721e-01, 6.69775478652568e-01, 6.69770591435771e-01, 6.69769261996734e-01, 6.69769065774410e-01, 6.69769050701021e-01, 6.69768780824590e-01, 6.69766952373578e-01, 6.69760230816941e-01, 6.69742313789165e-01, 6.69703103697003e-01, 6.69628015070288e-01, 6.69497445926886e-01, 6.69286439953829e-01, 6.68964563353931e-01, 6.68496016490820e-01, 6.67839995627006e-01, 6.66951313703720e-01, 6.65781280933324e-01, 6.64278835785481e-01, 6.62391904817568e-01, 6.60068956121115e-01, 6.57260696702821e-01, 6.53921850042465e-01, 6.50012937827732e-01, 6.45501981082148e-01, 6.40366032148087e-01, 6.34592451526515e-01, 6.28179853099120e-01, 6.21138657674280e-01, 6.13491217118671e-01, 6.05271497717190e-01, 5.96524339342032e-01, 5.87304333675817e-01, 5.77674387383951e-01, 5.67704052540591e-01, 5.57467715395376e-01, 5.47042735369594e-01, 5.36507619654593e-01, 5.25940306448838e-01, 5.15416613733549e-01, 5.05008892720163e-01, 4.94784907691681e-01, 4.84806948452785e-01, 4.75131168984248e-01, 4.65807136586605e-01, 4.56877569738040e-01, 4.48378239697796e-01, 4.40338009998655e-01, 4.32778988788115e-01, 4.25716770946380e-01, 4.19160749577436e-01, 4.13114479491242e-01, 4.07576078423808e-01, 4.02538654798072e-01, 3.97990753679942e-01, 3.93916815126495e-01, 3.90297641262968e-01, 3.87110870084889e-01, 3.84331455098121e-01, 3.81932150439255e-01, 3.79884001054272e-01, 3.78156836878425e-01, 3.76719768825780e-01, 3.75541682872424e-01, 3.74591726751660e-01, 3.73839781953386e-01, 3.73256912027443e-01, 3.72815776830350e-01, 3.72491001506049e-01, 3.72259488797141e-01, 3.72100663838005e-01, 3.71996641910521e-01, 3.71932311709223e-01, 3.71895329355901e-01, 3.71876021560745e-01, 3.71867199744152e-01, 3.71863890388234e-01, 3.71862990164317e-01, 3.71862857293195e-01, 3.71862847008423e-01, 3.71862661854045e-01, 3.71861407406698e-01, 3.71856795938115e-01, 3.71844503551955e-01, 3.71817602452673e-01, 3.71766085496148e-01, 3.71676502917850e-01, 3.71531729336986e-01, 3.71310877000255e-01, 3.70989368592994e-01, 3.70539179594140e-01, 3.69929255855141e-01, 3.69126106655230e-01, 3.68094566826635e-01, 3.66798713709208e-01, 3.65202915935143e-01, 3.63272981837859e-01, 3.60977366325706e-01, 3.58288387236584e-01, 3.55183396461186e-01, 3.51645848419864e-01, 3.47666209539049e-01, 3.43242657606176e-01, 3.38381529239456e-01, 3.33097486655163e-01, 3.27413390442814e-01, 3.21359881825664e-01, 3.14974694380673e-01, 3.08301729966026e-01, 3.01389945447433e-01, 2.94292104911314e-01, 2.87063456036542e-01, 2.79760389241517e-01, 2.72439134556365e-01, 2.65154544565519e-01, 2.57959003028652e-01, 2.50901488747346e-01, 2.44026813700458e-01, 2.37375044128278e-01, 2.30981103709951e-01, 2.24874549729215e-01, 2.19079506503448e-01, 2.13614735568008e-01, 2.08493819218724e-01, 2.03725432945660e-01, 1.99313682844604e-01, 1.95258485983715e-01, 1.91555974588098e-01, 1.88198908419650e-01, 1.85177083519115e-01, 1.82477729219933e-01, 1.80085888772467e-01, 1.77984781825465e-01, 1.76156149257572e-01, 1.74580582359184e-01, 1.73237839113050e-01, 1.72107150348594e-01, 1.71167517929241e-01, 1.70398005998962e-01, 1.69778024815450e-01, 1.69287605012656e-01, 1.68907658450784e-01, 1.68620220313990e-01, 1.68408665974162e-01, 1.68257895488068e-01, 1.68154478528685e-01, 1.68086753104980e-01, 1.68044872577653e-01, 1.68020797153346e-01, 1.68008228111205e-01, 1.68002485325999e-01, 1.68000331027443e-01, 1.67999745007741e-01, 1.67999658512521e-01, 1.67999651763168e-01, 1.67999529554706e-01, 1.67998701575556e-01, 1.67995657852522e-01, 1.67987544526817e-01, 1.67969789333135e-01, 1.67935788393450e-01, 1.67876667923252e-01, 1.67781133528315e-01, 1.67635418649482e-01, 1.67423342229476e-01, 1.67126483627754e-01, 1.66724479980147e-01, 1.66195447406182e-01, 1.65516522602552e-01, 1.64664515440450e-01, 1.63616656390523e-01, 1.62351415292818e-01, 1.60849360722075e-01, 1.59094022670622e-01, 1.57072716275351e-01, 1.54777281640214e-01, 1.52204695125285e-01, 1.49357511230555e-01, 1.46244101509417e-01, 1.42878667520996e-01, 1.39281017984537e-01, 1.35476114990276e-01, 1.31493409056752e-01, 1.27365996613428e-01, 1.23129644813389e-01, 1.18821736368623e-01, 1.14480190653900e-01, 1.10142416415371e-01, 1.05844346295996e-01, 1.01619594743099e-01, 9.74987697132050e-02, 9.35089561485540e-02, 8.96733767092002e-02, 8.60112238245136e-02, 8.25376476526998e-02, 7.92638775608978e-02, 7.61974504742113e-02, 7.33425177976699e-02, 7.07002032548232e-02, 6.82689864292101e-02, 6.60450904822909e-02, 6.40228569120183e-02, 6.21950948195575e-02, 6.05533965819518e-02, 5.90884158031102e-02, 5.77901067661841e-02, 5.66479272509284e-02, 5.56510085004973e-02, 5.47882973533547e-02, 5.40486761604484e-02, 5.34210661628620e-02, 5.28945195951565e-02, 5.24583049941540e-02, 5.21019891235923e-02, 5.18155176682882e-02, 5.15892955053019e-02, 5.14142660285302e-02, 5.12819877879167e-02, 5.11847057007131e-02, 5.11154133826963e-02, 5.10679027901233e-02, 5.10367973875105e-02, 5.10175654597447e-02, 5.10065109258520e-02, 5.10007400183881e-02, 5.09981033699276e-02, 5.09971142925920e-02, 5.09968452415066e-02, 5.09968055302039e-02, 5.09968024050949e-02, 5.09967454820597e-02, 5.09963598212053e-02, 5.09949421073443e-02, 5.09911631274491e-02, 5.09828935552719e-02, 5.09670587603779e-02, 5.09395295074881e-02, 5.08950552867304e-02, 5.08272468112594e-02, 5.07286137957550e-02, 5.05906633074744e-02, 5.04040625723761e-02, 5.01588679509766e-02, 4.98448187659690e-02, 4.94516907601934e-02, 4.89696993220861e-02, 4.83899375330026e-02, 4.77048290248021e-02, 4.69085711846517e-02, 4.59975410791884e-02, 4.49706352539749e-02, 4.38295158291737e-02, 4.25787393576371e-02, 4.12257516961745e-02, 3.97807412407545e-02, 3.82563534898495e-02, 3.66672809268422e-02, 3.50297524005967e-02, 3.33609543134301e-02, 3.16784210087704e-02, 2.99994331930478e-02, 2.83404609320428e-02, 2.67166821397993e-02, 2.51415993572053e-02, 2.36267680959230e-02, 2.21816402770685e-02, 2.08135173992829e-02, 1.95276008677183e-02, 1.83271219151209e-02, 1.72135309078953e-02, 1.61867253924513e-02, 1.52452975960529e-02, 1.43867847252750e-02, 1.36079087594680e-02, 1.29047960391020e-02, 1.22731704215653e-02, 1.17085168589739e-02, 1.12062147828595e-02, 1.07616425999097e-02, 1.03702559278958e-02, 1.00276430068433e-02, 9.72956110805256e-03, 9.47195783535193e-03, 9.25098105797417e-03, 9.06298090321656e-03, 8.90450682453547e-03, 8.77230228491044e-03, 8.66329908551933e-03, 8.57461284569183e-03, 8.50354061564218e-03, 8.44756109479050e-03, 8.40433744890845e-03, 8.37172229004324e-03, 8.34776402276560e-03, 8.33071349168080e-03, 8.31902970317344e-03, 8.31138335281878e-03, 8.30665696725674e-03, 8.30394065728512e-03, 8.30252275751064e-03, 8.30187496649870e-03, 8.30163196727796e-03, 8.30156586649622e-03, 8.30155611019734e-03, 8.30155533562760e-03, 8.30154114056114e-03, 8.30144496745788e-03, 8.30109143239355e-03, 8.30014909499965e-03, 8.29808710336977e-03, 8.29413925647771e-03, 8.28727740182700e-03, 8.27619617049891e-03, 8.25931112015610e-03, 8.23477234649772e-03, 8.20049545242796e-03, 8.15421134721817e-03, 8.09353560338621e-03, 8.01605697971297e-03, 7.91944321884586e-03, 7.80156041721369e-03, 7.66060028780657e-03, 7.49520772392742e-03, 7.30459951556384e-03, 7.08866419132903e-03, 6.84803304955750e-03, 6.58411370203282e-03, 6.29907992828294e-03, 5.99581517334968e-03, 5.67781125817390e-03, 5.34902828281503e-03, 5.01372567167380e-03, 4.67627725050335e-03, 4.34098471063038e-03, 4.01190360001635e-03, 3.69269415117657e-03, 3.38650613421976e-03, 3.09590300773266e-03, 2.82282651192389e-03, 2.56859906922813e-03, 2.33395837504702e-03, 2.11911665674336e-03, 1.92383632382587e-03, 1.74751399240591e-03, 1.58926587128860e-03, 1.44800893552599e-03, 1.32253390559356e-03, 1.21156760044272e-03, 1.11382361625090e-03, 1.02804139932289e-03, 9.53014593434887e-04, 8.87610053077026e-04, 8.30779107317881e-04, 7.81562627018673e-04, 7.39091372544465e-04, 7.02582969211107e-04, 6.71336596589356e-04, 6.44726241610272e-04, 6.22193263282611e-04, 6.03238893802427e-04, 5.87417074143559e-04, 5.74327841547984e-04, 5.63611433058497e-04, 5.54943244281958e-04, 5.48029713314691e-04, 5.42605111569117e-04, 5.38429164890886e-04, 5.35285395173446e-04, 5.32980047537990e-04, 5.31341444407474e-04, 5.30219588924297e-04, 5.29485836253230e-04, 5.29032459169798e-04, 5.28771957597485e-04, 5.28635994399451e-04, 5.28573881304339e-04, 5.28550582089338e-04, 5.28544244289719e-04, 5.28543308853102e-04, 5.28543233897949e-04, 5.28541851878117e-04, 5.28532488530686e-04, 5.28498069094496e-04, 5.28406328932363e-04, 5.28205605547782e-04, 5.27821379058935e-04, 5.27153777731826e-04, 5.26076289915522e-04, 5.24435947475865e-04, 5.22055256347187e-04, 5.18736135248332e-04, 5.14266067817459e-04, 5.08426563818887e-04, 5.01003851915982e-04, 4.91801488408328e-04, 4.80654272877246e-04, 4.67442547112285e-04, 4.52105682001047e-04, 4.34653409350005e-04, 4.15173665629709e-04, 3.93835743729198e-04, 3.70887775254403e-04, 3.46648057515608e-04, 3.21490659494827e-04, 2.95826680556388e-04, 2.70082732072296e-04, 2.44678047451430e-04, 2.20002494224047e-04, 1.96398572839485e-04, 1.74148338409452e-04, 1.53464150898230e-04, 1.34485413937568e-04, 1.17283653700895e-04, 1.01869764208118e-04, 8.82006128861690e-05, 7.61943667484388e-05, 6.57458488859675e-05, 5.67218063523001e-05, 4.89616572920681e-05, 4.23078552245215e-05, 3.66285017889822e-05, 3.18356933064479e-05, 2.79349311682608e-05, 2.48922003720509e-05, 2.24462010720572e-05, 2.03543818884100e-05, 1.84738820863117e-05, 1.67626660143583e-05, 1.52903187355946e-05, 1.41420502686174e-05, 1.32445956992910e-05, 1.24944787742720e-05, 1.19029700869078e-05, 1.15016333523954e-05, 1.12396464402309e-05, 1.10400616435413e-05, 1.08611834477301e-05, 1.06936550210435e-05, 1.05422798881705e-05, 1.04135191421051e-05, 1.03102458695048e-05, 1.02313641505702e-05, 1.01734094073436e-05, 1.01321822612022e-05, 1.01037279875982e-05, 1.00847223691858e-05, 1.00725216611863e-05, 1.00650794294666e-05, 1.00608377806566e-05, 1.00586341460485e-05, 1.00576298081023e-05, 1.00572534549315e-05, 1.00571511154612e-05, 1.00571360093227e-05, 1.00571346308698e-05, 1.00571070120736e-05, 1.00569199261713e-05, 1.00562322131736e-05, 1.00543992971325e-05, 1.00503894344683e-05, 1.00427156866090e-05, 1.00293892657533e-05, 1.00079032776534e-05, 9.97526464630027e-06, 9.92811600042387e-06, 9.86303671624000e-06, 9.77715539107749e-06, 9.66908147859158e-06, 9.53952877580377e-06, 9.38982245912712e-06, 9.21640766904375e-06, 9.00346282151077e-06, 8.72168272536555e-06, 8.34064997690120e-06, 7.85425910265758e-06, 7.30875157017863e-06, 6.80220811943855e-06, 6.41826553504643e-06, 6.17702529126518e-06, 6.08382138977959e-06, 6.11620435364557e-06, 6.17790031753884e-06, 6.15618474597568e-06, 6.02273775029971e-06, 5.88054719941056e-06, 5.90306208181346e-06, 6.22856187959542e-06, 6.85999646066193e-06, 7.58490303809367e-06, 8.07109692146923e-06, 8.05258561342725e-06, 7.39709843600167e-06, 6.13830978579699e-06, 4.53176110681852e-06, 2.92453692907202e-06, 1.50630337003577e-06, 9.76248415440750e-07, 1.28951865718818e-06, 1.00413736703331e-06, 1.23123553663618e-06, 2.44515830161163e-06, 3.26921009019543e-06, 3.53972867372931e-06, 3.26780447323927e-06, 2.59291845594589e-06, 1.72061965741932e-06, 8.79302632483814e-07, 6.60392096563100e-07, 8.31157294101104e-07, 8.06527268082391e-07, 7.03593720174922e-07, 5.97635377630468e-07, 5.24110923549870e-07, 4.91560527393631e-07, 4.79701093070162e-07, 4.60994151296745e-07, 4.23465166932147e-07, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -1.42108547152020e-14, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15],
[ 5.20000000000000e+01, 5.19999999510232e+01, 5.19999984365631e+01, 5.19999881758563e+01, 5.19999504561625e+01, 5.19998499073858e+01, 5.19996298490958e+01, 5.19992083737765e+01, 5.19984753077219e+01, 5.19972901641162e+01, 5.19954811724236e+01, 5.19928454354240e+01, 5.19891502314003e+01, 5.19841354455304e+01, 5.19775170829002e+01, 5.19689917869912e+01, 5.19582422629989e+01, 5.19449434854282e+01, 5.19287695540557e+01, 5.19094010509058e+01, 5.18865327421830e+01, 5.18598814616262e+01, 5.18291940039807e+01, 5.17942548481146e+01, 5.17548935184246e+01, 5.17109913814420e+01, 5.16624876640424e+01, 5.16093844735877e+01, 5.15517506024967e+01, 5.14897239139020e+01, 5.14235121340598e+01, 5.13533919222070e+01, 5.12797061486409e+01, 5.12028593836056e+01, 5.11233116778540e+01, 5.10415707939120e+01, 5.09581831182615e+01, 5.08737235426825e+01, 5.07887846432933e+01, 5.07039655059560e+01, 5.06198605465498e+01, 5.05370486561277e+01, 5.04560829677015e+01, 5.03774814977815e+01, 5.03017188664509e+01, 5.02292192488925e+01, 5.01603506622937e+01, 5.00954206472930e+01, 5.00346733638819e+01, 4.99782880883335e+01, 4.99263790699663e+01, 4.98789966835855e+01, 4.98361297943144e+01, 4.97977092352347e+01, 4.97636122839580e+01, 4.97336680113438e+01, 4.97076633637356e+01, 4.96853498292444e+01, 4.96664505289982e+01, 4.96506675663025e+01, 4.96376894608891e+01, 4.96271984925242e+01, 4.96188777788385e+01, 4.96124179169137e+01, 4.96075230273436e+01, 4.96039160534210e+01, 4.96013431867720e+01, 4.95995773139017e+01, 4.95984204051912e+01, 4.95977047981147e+01, 4.95972933588558e+01, 4.95970785399668e+01, 4.95969803850408e+01, 4.95969435633508e+01, 4.95969335469217e+01, 4.95969320685164e+01, 4.95969318745808e+01, 4.95969273561140e+01, 4.95968967428169e+01, 4.95967842048097e+01, 4.95964842175820e+01, 4.95958276916130e+01, 4.95945703227073e+01, 4.95923836127038e+01, 4.95888489459631e+01, 4.95834550335171e+01, 4.95755989544265e+01, 4.95645909331145e+01, 4.95496628926874e+01, 4.95299807182520e+01, 4.95046600520677e+01, 4.94727853253962e+01, 4.94334316116702e+01, 4.93856887637929e+01, 4.93286871770955e+01, 4.92616244018740e+01, 4.91837917208168e+01, 4.90945997156644e+01, 4.89936017869094e+01, 4.88805145766350e+01, 4.87552342954056e+01, 4.86178480841309e+01, 4.84686397571448e+01, 4.83080895658162e+01, 4.81368679689902e+01, 4.79558237589001e+01, 4.77659672220332e+01, 4.75684492686145e+01, 4.73645376092607e+01, 4.71555910811615e+01, 4.69430331408513e+01, 4.67283253783545e+01, 4.65129417114910e+01, 4.62983437321664e+01, 4.60859575302976e+01, 4.58771522299049e+01, 4.56732204319387e+01, 4.54753607520639e+01, 4.52846626455957e+01, 4.51020937052709e+01, 4.49284895882751e+01, 4.47645466753804e+01, 4.46108174946240e+01, 4.44677088668489e+01, 4.43354826626377e+01, 4.42142590079109e+01, 4.41040217415652e+01, 4.40046259110157e+01, 4.39158070853267e+01, 4.38371922647482e+01, 4.37683121646110e+01, 4.37086146470411e+01, 4.36574790641372e+01, 4.36142312611846e+01, 4.35781589695791e+01, 4.35485272987164e+01, 4.35245940169786e+01, 4.35056242970970e+01, 4.34909045934476e+01, 4.34797553207571e+01, 4.34715420172371e+01, 4.34656847015303e+01, 4.34616651724176e+01, 4.34590320523909e+01, 4.34574034393927e+01, 4.34564671028527e+01, 4.34559782375115e+01, 4.34557548678662e+01, 4.34556710737099e+01, 4.34556482796213e+01, 4.34556449152599e+01, 4.34556446182508e+01, 4.34556387985042e+01, 4.34555993688544e+01, 4.34554544211440e+01, 4.34550680426372e+01, 4.34542224569992e+01, 4.34526030373992e+01, 4.34497867844488e+01, 4.34452347865394e+01, 4.34382890932296e+01, 4.34281743323164e+01, 4.34140042860977e+01, 4.33947935099773e+01, 4.33694739255345e+01, 4.33369161504756e+01, 4.32959551417604e+01, 4.32454195307937e+01, 4.31841638291998e+01, 4.31111024918704e+01, 4.30252446548270e+01, 4.29257282348977e+01, 4.28118520027647e+01, 4.26831042362512e+01, 4.25391866400089e+01, 4.23800323897916e+01, 4.22058174257045e+01, 4.20169644695388e+01, 4.18141396517506e+01, 4.15982420619160e+01, 4.13703869270569e+01, 4.11318834172480e+01, 4.08842082351484e+01, 4.06289761563150e+01, 4.03679085809537e+01, 4.01028009955205e+01, 3.98354900901460e+01, 3.95678211801910e+01, 3.93016165465385e+01, 3.90386453158562e+01, 3.87805955092571e+01, 3.85290488596761e+01, 3.82854589166514e+01, 3.80511328249363e+01, 3.78272169993538e+01, 3.76146867482141e+01, 3.74143397452527e+01, 3.72267931317144e+01, 3.70524839528165e+01, 3.68916725948494e+01, 3.67444488833009e+01, 3.66107405184373e+01, 3.64903235522269e+01, 3.63828346402071e+01, 3.62877848269696e+01, 3.62045746399340e+01, 3.61325102708562e+01, 3.60708206178667e+01, 3.60186749440528e+01, 3.59752008841675e+01, 3.59395025022287e+01, 3.59106780734097e+01, 3.58878372377744e+01, 3.58701171551319e+01, 3.58566972832694e+01, 3.58468124091210e+01, 3.58397635861342e+01, 3.58349266721473e+01, 3.58317582200566e+01, 3.58297985467035e+01, 3.58286718907307e+01, 3.58280836636399e+01, 3.58278148952805e+01, 3.58277140705786e+01, 3.58276866437762e+01, 3.58276825956362e+01, 3.58276822709253e+01, 3.58276762783620e+01, 3.58276356778565e+01, 3.58274864259923e+01, 3.58270885744653e+01, 3.58262178817920e+01, 3.58245503845383e+01, 3.58216505408077e+01, 3.58169634827443e+01, 3.58098118268364e+01, 3.57993973892089e+01, 3.57848080355232e+01, 3.57650297590748e+01, 3.57389639244632e+01, 3.57054494376508e+01, 3.56632894086613e+01, 3.56112816660668e+01, 3.55482522717637e+01, 3.54730909828023e+01, 3.53847874295443e+01, 3.52824666432425e+01, 3.51654224884449e+01, 3.50331475516793e+01, 3.48853581188251e+01, 3.47220130443307e+01, 3.45433255728289e+01, 3.43497675053510e+01, 3.41420654864320e+01, 3.39211895948620e+01, 3.36883348141070e+01, 3.34448963018945e+01, 3.31924396400729e+01, 3.29326674039154e+01, 3.26673834375618e+01, 3.23984561686177e+01, 3.21277821633927e+01, 3.18572509464315e+01, 3.15887119157529e+01, 3.13239440033399e+01, 3.10646285719113e+01, 3.08123259047398e+01, 3.05684555277920e+01, 3.03342804929710e+01, 3.01108956414071e+01, 2.98992197568399e+01, 2.96999914177219e+01, 2.95137682724318e+01, 2.93409294035740e+01, 2.91816804190650e+01, 2.90360609082554e+01, 2.89039539245084e+01, 2.87850971924298e+01, 2.86790957788000e+01, 2.85854360030004e+01, 2.85035003895336e+01, 2.84325834788645e+01, 2.83719083124349e+01, 2.83206433943823e+01, 2.82779199087663e+01, 2.82428489405055e+01, 2.82145384149679e+01, 2.81921094397891e+01, 2.81747117075833e+01, 2.81615376040033e+01, 2.81518346656872e+01, 2.81449160495668e+01, 2.81401687101242e+01, 2.81370590343154e+01, 2.81351357534091e+01, 2.81340300338885e+01, 2.81334527416930e+01, 2.81331889705323e+01, 2.81330900206064e+01, 2.81330631038032e+01, 2.81330591309385e+01, 2.81330588245130e+01, 2.81330533220935e+01, 2.81330160423892e+01, 2.81328789982064e+01, 2.81325136883754e+01, 2.81317142149426e+01, 2.81301831207681e+01, 2.81275205135371e+01, 2.81232169743643e+01, 2.81166506704779e+01, 2.81070890000204e+01, 2.80936949901830e+01, 2.80755385449383e+01, 2.80516124935122e+01, 2.80208532250237e+01, 2.79821655104648e+01, 2.79344509156290e+01, 2.78766390065072e+01, 2.78077203543517e+01, 2.77267801762457e+01, 2.76330313155769e+01, 2.75258451923214e+01, 2.74047793506041e+01, 2.72696003114951e+01, 2.71203006069653e+01, 2.69571091226459e+01, 2.67804941993190e+01, 2.65911593131011e+01, 2.63900315409300e+01, 2.61782433852416e+01, 2.59571088441239e+01, 2.57280948424519e+01, 2.54927892705493e+01, 2.52528669112895e+01, 2.50100544908055e+01, 2.47660959875715e+01, 2.45227192052863e+01, 2.42816044752998e+01, 2.40443562123139e+01, 2.38124779020046e+01, 2.35873509466512e+01, 2.33702176325156e+01, 2.31621683140032e+01, 2.29641327440078e+01, 2.27768753303529e+01, 2.26009939777714e+01, 2.24369220926951e+01, 2.22849332878714e+01, 2.21451483231532e+01, 2.20175438505282e+01, 2.19019625853922e+01, 2.17981245911646e+01, 2.17056394302891e+01, 2.16240189931056e+01, 2.15526908611750e+01, 2.14910120901004e+01, 2.14382833078163e+01, 2.13937630187536e+01, 2.13566819848504e+01, 2.13262575247886e+01, 2.13017075375990e+01, 2.12822640207749e+01, 2.12671858212163e+01, 2.12557703343247e+01, 2.12473638563582e+01, 2.12413703006931e+01, 2.12372580116190e+01, 2.12345644499065e+01, 2.12328985813321e+01, 2.12319408698694e+01, 2.12314408574648e+01, 2.12312123975050e+01, 2.12311266942605e+01, 2.12311033808975e+01, 2.12310999398939e+01, 2.12310996793514e+01, 2.12310950638688e+01, 2.12310637932988e+01, 2.12309488394409e+01, 2.12306424151614e+01, 2.12299718156600e+01, 2.12286875471221e+01, 2.12264542238255e+01, 2.12228446614867e+01, 2.12173375325046e+01, 2.12093188755016e+01, 2.11980876641367e+01, 2.11828655353585e+01, 2.11628106525694e+01, 2.11370355336392e+01, 2.11046285088174e+01, 2.10646782939435e+01, 2.10163009780132e+01, 2.09586685426515e+01, 2.08910378687514e+01, 2.08127790586639e+01, 2.07234018272828e+01, 2.06225787069049e+01, 2.05101638799025e+01, 2.03862066054021e+01, 2.02509584394910e+01, 2.01048737529345e+01, 1.99486034075710e+01, 1.97829818364505e+01, 1.96090081516974e+01, 1.94278222436317e+01, 1.92406771018516e+01, 1.90489087567096e+01, 1.88539052912759e+01, 1.86570763066540e+01, 1.84598240496540e+01, 1.82635171572556e+01, 1.80694676724515e+01, 1.78789116798494e+01, 1.76929936324764e+01, 1.75127542202759e+01, 1.73391214800819e+01, 1.71729047676925e+01, 1.70147911954144e+01, 1.68653441665878e+01, 1.67250036931969e+01, 1.65940882464589e+01, 1.64727979502945e+01, 1.63612189761670e+01, 1.62593290321790e+01, 1.61670038603121e+01, 1.60840246658932e+01, 1.60100864057745e+01, 1.59448068586753e+01, 1.58877363940485e+01, 1.58383683451671e+01, 1.57961498778485e+01, 1.57604932282873e+01, 1.57307871621822e+01, 1.57064084836969e+01, 1.56867333984895e+01, 1.56711485124019e+01, 1.56590612291333e+01, 1.56499092991338e+01, 1.56431692705806e+01, 1.56383636036089e+01, 1.56350662320645e+01, 1.56329063930153e+01, 1.56315705920778e+01, 1.56308026302397e+01, 1.56304016823552e+01, 1.56302184853624e+01, 1.56301497617499e+01, 1.56301310672519e+01, 1.56301283079829e+01, 1.56301281008436e+01, 1.56301244549588e+01, 1.56300997535526e+01, 1.56300089485684e+01, 1.56297668958036e+01, 1.56292371691668e+01, 1.56282226789722e+01, 1.56264584710256e+01, 1.56236070392813e+01, 1.56192564336980e+01, 1.56129213851027e+01, 1.56040475983404e+01, 1.55920192815857e+01, 1.55761698817856e+01, 1.55557958833808e+01, 1.55301734012071e+01, 1.54985771627888e+01, 1.54603013368660e+01, 1.54146815334192e+01, 1.53611171871757e+01, 1.52990934540434e+01, 1.52282017096929e+01, 1.51481577504493e+01, 1.50588168629007e+01, 1.49601850483572e+01, 1.48524258533425e+01, 1.47358624542422e+01, 1.46109748567266e+01, 1.44783922826692e+01, 1.43388810167404e+01, 1.41933281652914e+01, 1.40427219408213e+01, 1.38881292282871e+01, 1.37306713144770e+01, 1.35714987609222e+01, 1.34117664571560e+01, 1.32526098810864e+01, 1.30951234962634e+01, 1.29403420258359e+01, 1.27892250769403e+01, 1.26426452868650e+01, 1.25013798757369e+01, 1.23661052677820e+01, 1.22373943135188e+01, 1.21157156118352e+01, 1.20014344746180e+01, 1.18948151661298e+01, 1.17960241529123e+01, 1.17051341936932e+01, 1.16221291696839e+01, 1.15469096006299e+01, 1.14792988142201e+01, 1.14190497418741e+01, 1.13658523083668e+01, 1.13193413706908e+01, 1.12791051457511e+01, 1.12446940483225e+01, 1.12156298408222e+01, 1.11914149753230e+01, 1.11715419865556e+01, 1.11555027735860e+01, 1.11427975889997e+01, 1.11329435397346e+01, 1.11254823952028e+01, 1.11199874979224e+01, 1.11160695809838e+01, 1.11133813161274e+01, 1.11116204460067e+01, 1.11105313935147e+01, 1.11099052881980e+01, 1.11095784023954e+01, 1.11094290450238e+01, 1.11093730158286e+01, 1.11093577745205e+01, 1.11093555249351e+01, 1.11093553565716e+01, 1.11093524000230e+01, 1.11093323689710e+01, 1.11092587326947e+01, 1.11090624453394e+01, 1.11086328748669e+01, 1.11078101945003e+01, 1.11063795416712e+01, 1.11040672135121e+01, 1.11005391261314e+01, 1.10954017160849e+01, 1.10882054050265e+01, 1.10784506807686e+01, 1.10655967686514e+01, 1.10490727758835e+01, 1.10282910896217e+01, 1.10026627002302e+01, 1.09716140099099e+01, 1.09346045814180e+01, 1.08911451913398e+01, 1.08408154874938e+01, 1.07832805200880e+01, 1.07183054284265e+01, 1.06457676225896e+01, 1.05656659007661e+01, 1.04781260805310e+01, 1.03834028844594e+01, 1.02818779925222e+01, 1.01740543413469e+01, 1.00605469021627e+01, 9.94207029836675e+00, 9.81942372840597e+00, 9.69347374183764e+00, 9.56513547868534e+00, 9.43535302555088e+00, 9.30507956434337e+00, 9.17525798624302e+00, 9.04680260926063e+00, 8.92058256915087e+00, 8.79740735146456e+00, 8.67801480347747e+00, 8.56306181958228e+00, 8.45311774621420e+00, 8.34866041641265e+00, 8.25007461138288e+00, 8.15765266464545e+00, 8.07159687663739e+00, 7.99202339265278e+00, 7.91896721002625e+00, 7.85238801454909e+00, 7.79217659348196e+00, 7.73816162563779e+00, 7.69011670130309e+00, 7.64776747112688e+00, 7.61079886000263e+00, 7.57886230740410e+00, 7.55158300899377e+00, 7.52856713614673e+00, 7.50940900181314e+00, 7.49369812509463e+00, 7.48102612577266e+00, 7.47099335684555e+00, 7.46321516104490e+00, 7.45732761930342e+00, 7.45299264785190e+00, 7.44990229804831e+00, 7.44778212041713e+00, 7.44639347203093e+00, 7.44553467368926e+00, 7.44504095885675e+00, 7.44478319781297e+00, 7.44466542525530e+00, 7.44462124479876e+00, 7.44460922665240e+00, 7.44460745279950e+00, 7.44460731999107e+00, 7.44460498714427e+00, 7.44458918177007e+00, 7.44453107963428e+00, 7.44437620137685e+00, 7.44403725767309e+00, 7.44338815363143e+00, 7.44225939913970e+00, 7.44043514699312e+00, 7.43765205144790e+00, 7.43360010482025e+00, 7.42792556689309e+00, 7.42023604911044e+00, 7.41010775114358e+00, 7.39709477062609e+00, 7.38074031831012e+00, 7.36058957302353e+00, 7.33620380804665e+00, 7.30717531930707e+00, 7.27314259421865e+00, 7.23380508726044e+00, 7.18893692391106e+00, 7.13839884688575e+00, 7.08214775431171e+00, 7.02024326191385e+00, 6.95285084975330e+00, 6.88024132312856e+00, 6.80278651663337e+00, 6.72095138543656e+00, 6.63528284079940e+00, 6.54639587852404e+00, 6.45495770124343e+00, 6.36167063348289e+00, 6.26725466319407e+00, 6.17243041311603e+00, 6.07790325558321e+00, 5.98434914773643e+00, 5.89240259769814e+00, 5.80264699534277e+00, 5.71560737212173e+00, 5.63174550791439e+00, 5.55145718901624e+00, 5.47507134458692e+00, 5.40285074855375e+00, 5.33499396562326e+00, 5.27163823685379e+00, 5.21286303447050e+00, 5.15869405987034e+00, 5.10910750677613e+00, 5.06403445842925e+00, 5.02336533020721e+00, 4.98695430505033e+00, 4.95462373750531e+00, 4.92616852267153e+00, 4.90136043896470e+00, 4.87995247880654e+00, 4.86168317975402e+00, 4.84628096106222e+00, 4.83346845830250e+00, 4.82296683273756e+00, 4.81450001420032e+00, 4.80779881792547e+00, 4.80260485892569e+00, 4.79867417385336e+00, 4.79578045144240e+00, 4.79371776986653e+00, 4.79230274350122e+00, 4.79137599288790e+00, 4.79080286980149e+00, 4.79047339323519e+00, 4.79030138034192e+00, 4.79022278704471e+00, 4.79019330411737e+00, 4.79018528405870e+00, 4.79018410031529e+00, 4.79018401145326e+00, 4.79018244740712e+00, 4.79017185077131e+00, 4.79013289652097e+00, 4.79002905958930e+00, 4.78980181905171e+00, 4.78936664174672e+00, 4.78860991329350e+00, 4.78738696922119e+00, 4.78552136083395e+00, 4.78280546811764e+00, 4.77900254365237e+00, 4.77385023522432e+00, 4.76706558903330e+00, 4.75835147924398e+00, 4.74740434354663e+00, 4.73392303043863e+00, 4.71761848605121e+00, 4.69822393234050e+00, 4.67550512169046e+00, 4.64927020364146e+00, 4.61937871559985e+00, 4.58574921763462e+00, 4.54836513577999e+00, 4.50727845883428e+00, 4.46261104630654e+00, 4.41455344142319e+00, 4.36336123098072e+00, 4.30934913935074e+00, 4.25288317314703e+00, 4.19437123407262e+00, 4.13425268220225e+00, 4.07298735716813e+00, 4.01104455203634e+00, 3.94889238979704e+00, 3.88698798366014e+00, 3.82576867900326e+00, 3.76564458556545e+00, 3.70699252051905e+00, 3.65015140171465e+00, 3.59541905929471e+00, 3.54305037528534e+00, 3.49325661604057e+00, 3.44620579213307e+00, 3.40202386436383e+00, 3.36079661212469e+00, 3.32257198969267e+00, 3.28736281472438e+00, 3.25514965831094e+00, 3.22588383433827e+00, 3.19949041464852e+00, 3.17587122314808e+00, 3.15490778471248e+00, 3.13646422232380e+00, 3.12039010777931e+00, 3.10652327748519e+00, 3.09469262564779e+00, 3.08472088324801e+00, 3.07642738340343e+00, 3.06963080312335e+00, 3.06415185920917e+00, 3.05981592338491e+00, 3.05645550993273e+00, 3.05391257938190e+00, 3.05204059525339e+00, 3.05070626837040e+00, 3.04979092536922e+00, 3.04919144496020e+00, 3.04882071696013e+00, 3.04860759449501e+00, 3.04849632806540e+00, 3.04844549013214e+00, 3.04842641916814e+00, 3.04842123141295e+00, 3.04842046571149e+00, 3.04842040798652e+00, 3.04841938871746e+00, 3.04841248302480e+00, 3.04838709704634e+00, 3.04831942794913e+00, 3.04817133883434e+00, 3.04788774214788e+00, 3.04739460169333e+00, 3.04659765377441e+00, 3.04538193659442e+00, 3.04361220283833e+00, 3.04113427209481e+00, 3.03777735610510e+00, 3.03335735961948e+00, 3.02768112229445e+00, 3.02055152281253e+00, 3.01177331660689e+00, 3.00115952586121e+00, 2.98853814874830e+00, 2.97375890915105e+00, 2.95669973397635e+00, 2.93727262817950e+00, 2.91542862245804e+00, 2.89116149823334e+00, 2.86451004955597e+00, 2.83555871955587e+00, 2.80443654460457e+00, 2.77131444441277e+00, 2.73640100099303e+00, 2.69993696332754e+00, 2.66218878809892e+00, 2.62344157264841e+00, 2.58399175053972e+00, 2.54413990290820e+00, 2.50418399434766e+00, 2.46441327782123e+00, 2.42510303820235e+00, 2.38651026797208e+00, 2.34887029939344e+00, 2.31239436080395e+00, 2.27726798332016e+00, 2.24365015839623e+00, 2.21167313451960e+00, 2.18144273995646e+00, 2.15303912468607e+00, 2.12651782559264e+00, 2.10191107233432e+00, 2.07922926546756e+00, 2.05846257233033e+00, 2.03958259924124e+00, 2.02254411034905e+00, 2.00728677367923e+00, 1.99373692332673e+00, 1.98180933311826e+00, 1.97140900123664e+00, 1.96243294715619e+00, 1.95477202177996e+00, 1.94831272902630e+00, 1.94293905256504e+00, 1.93853427538387e+00, 1.93498277294399e+00, 1.93217175354123e+00, 1.92999291286367e+00, 1.92834396437370e+00, 1.92713000372199e+00, 1.92626466446685e+00, 1.92567102427859e+00, 1.92528222566512e+00, 1.92504178291264e+00, 1.92490355698902e+00, 1.92483139198724e+00, 1.92479841951869e+00, 1.92478605045864e+00, 1.92478268577925e+00, 1.92478218915979e+00, 1.92478215151671e+00, 1.92478148413803e+00, 1.92477696255219e+00, 1.92476034076842e+00, 1.92471603353432e+00, 1.92461906995506e+00, 1.92443337968711e+00, 1.92411048336201e+00, 1.92358865238315e+00, 1.92279259550988e+00, 1.92163372011128e+00, 1.92001100387019e+00, 1.91781249723616e+00, 1.91491745738524e+00, 1.91119909050143e+00, 1.90652785103814e+00, 1.90077521514769e+00, 1.89381781238605e+00, 1.88554176760831e+00, 1.87584707680691e+00, 1.86465181997588e+00, 1.85189600425148e+00, 1.83754483426765e+00, 1.82159122537063e+00, 1.80405740896581e+00, 1.78499552594523e+00, 1.76448716031916e+00, 1.74264182603547e+00, 1.71959448008014e+00, 1.69550218899506e+00, 1.67054011943787e+00, 1.64489705319816e+00, 1.61877064163739e+00, 1.59236261388315e+00, 1.56587413870344e+00, 1.53950151424276e+00, 1.51343232579413e+00, 1.48784217285727e+00, 1.46289202622441e+00, 1.43872623679275e+00, 1.41547118280430e+00, 1.39323451320653e+00, 1.37210492303247e+00, 1.35215238259005e+00, 1.33342873560245e+00, 1.31596858144915e+00, 1.29979036209919e+00, 1.28489758375827e+00, 1.27128011516746e+00, 1.25891551749494e+00, 1.24777037364410e+00, 1.23780159661322e+00, 1.22895770659425e+00, 1.22118007435653e+00, 1.21440413390868e+00, 1.20856057043737e+00, 1.20357649020152e+00, 1.19937657765563e+00, 1.19588424192884e+00, 1.19302275032143e+00, 1.19071634119465e+00, 1.18889130306458e+00, 1.18747700143291e+00, 1.18640683044669e+00, 1.18561906337512e+00, 1.18505757452267e+00, 1.18467240582455e+00, 1.18442015408435e+00, 1.18426415951787e+00, 1.18417448268795e+00, 1.18412766462249e+00, 1.18410627336345e+00, 1.18409824881248e+00, 1.18409606594466e+00, 1.18409574375811e+00, 1.18409571918186e+00, 1.18409528142197e+00, 1.18409231553683e+00, 1.18408141266924e+00, 1.18405234993363e+00, 1.18398874840204e+00, 1.18386694981695e+00, 1.18365515969533e+00, 1.18331290059746e+00, 1.18279081467068e+00, 1.18203084868490e+00, 1.18096684705662e+00, 1.17952556818363e+00, 1.17762812624620e+00, 1.17519184418023e+00, 1.17213248388440e+00, 1.16836679749381e+00, 1.16381531992820e+00, 1.15840529968153e+00, 1.15207364422633e+00, 1.14476974100183e+00, 1.13645800725732e+00, 1.12712002414333e+00, 1.11675612371295e+00, 1.10538632212302e+00, 1.09305052719611e+00, 1.07980799116913e+00, 1.06573602631261e+00, 1.05092804781575e+00, 1.03549105035749e+00, 1.01954265800863e+00, 1.00320790840939e+00, 9.86615939839311e-01, 9.69896743764380e-01, 9.53178127207345e-01, 9.36583001621841e-01, 9.20227081448964e-01, 9.04217040051165e-01, 8.88649136826530e-01, 8.73608299886413e-01, 8.59167625644652e-01, 8.45388240902039e-01, 8.32319464452787e-01, 8.19999203097296e-01, 8.08454519956172e-01, 7.97702319676965e-01, 7.87750104059356e-01, 7.78596761516354e-01, 7.70233363619553e-01, 7.62643951008940e-01, 7.55806298712670e-01, 7.49692657172446e-01, 7.44270469921531e-01, 7.39503071947048e-01, 7.35350374385042e-01, 7.31769541490003e-01, 7.28715664956560e-01, 7.26142438838806e-01, 7.24002835717101e-01, 7.22249781635604e-01, 7.20836823927826e-01, 7.19718782639703e-01, 7.18852373139853e-01, 7.18196784961741e-01, 7.17714200217152e-01, 7.17370234269588e-01, 7.17134281901551e-01, 7.16979753998608e-01, 7.16884192746754e-01, 7.16829257335284e-01, 7.16800576917358e-01, 7.16787472784212e-01, 7.16782557001679e-01, 7.16781219792480e-01, 7.16781022423305e-01, 7.16781007261822e-01, 7.16780735808079e-01, 7.16778896670725e-01, 7.16772135832926e-01, 7.16754114118864e-01, 7.16714675039405e-01, 7.16639148319437e-01, 7.16507818719847e-01, 7.16295587382582e-01, 7.15971850060029e-01, 7.15500612938619e-01, 7.14840861943493e-01, 7.13947195000486e-01, 7.12770718378408e-01, 7.11260197744132e-01, 7.09363441997517e-01, 7.07028883730288e-01, 7.04207305088183e-01, 7.00853643135673e-01, 6.96928796076193e-01, 6.92401342597812e-01, 6.87249082859388e-01, 6.81460312538213e-01, 6.75034751634897e-01, 6.67984067240617e-01, 6.60331953121961e-01, 6.52113756769381e-01, 6.43375673796022e-01, 6.34173557253433e-01, 6.24571412666739e-01, 6.14639666076975e-01, 6.04453300727435e-01, 5.94089957961465e-01, 5.83628090198708e-01, 5.73145240177055e-01, 5.62716503153048e-01, 5.52413209743648e-01, 5.42301848647583e-01, 5.32443232218803e-01, 5.22891894800857e-01, 5.13695704291173e-01, 5.04895661513473e-01, 4.96525859202031e-01, 4.88613572124294e-01, 4.81179451419870e-01, 4.74237798990927e-01, 4.67796901216140e-01, 4.61859404965814e-01, 4.56422722568071e-01, 4.51479455805867e-01, 4.47017832076305e-01, 4.43022148427914e-01, 4.39473221245606e-01, 4.36348840837717e-01, 4.33624231065913e-01, 4.31272514443336e-01, 4.29265182828154e-01, 4.27572573007836e-01, 4.26164345185228e-01, 4.25009960756334e-01, 4.24079153953926e-01, 4.23342390092309e-01, 4.22771301468373e-01, 4.22339090637259e-01, 4.22020889956599e-01, 4.21794066115353e-01, 4.21638458918736e-01, 4.21536544917224e-01, 4.21473518501458e-01, 4.21437285735578e-01, 4.21418369304263e-01, 4.21409726307694e-01, 4.21406484033760e-01, 4.21405602057824e-01, 4.21405471880064e-01, 4.21405461803765e-01, 4.21405280402553e-01, 4.21404051383533e-01, 4.21399533392268e-01, 4.21387490183903e-01, 4.21361134411967e-01, 4.21310661852353e-01, 4.21222895582147e-01, 4.21081057886241e-01, 4.20864685588406e-01, 4.20549701928749e-01, 4.20108654822194e-01, 4.19511127099483e-01, 4.18724318961686e-01, 4.17713796270597e-01, 4.16444390502839e-01, 4.14881227470339e-01, 4.12990852737821e-01, 4.10742412764570e-01, 4.08108843058720e-01, 4.05068009045578e-01, 4.01603742864793e-01, 3.97706720671657e-01, 3.93375130621941e-01, 3.88615091491197e-01, 3.83440795221361e-01, 3.77874362518405e-01, 3.71945417510126e-01, 3.65690403824544e-01, 3.59151678755673e-01, 3.52376433224386e-01, 3.45415492238310e-01, 3.38322053234130e-01, 3.31150418257273e-01, 3.23954770998682e-01, 3.16788042122006e-01, 3.09700897036514e-01, 3.02740870239042e-01, 2.95951660370505e-01, 2.89372590855720e-01, 2.83038232882859e-01, 2.76978180839698e-01, 2.71216965320995e-01, 2.65774085486882e-01, 2.60664140806128e-01, 2.55897041887422e-01, 2.51478280960988e-01, 2.47409244362927e-01, 2.43687551827911e-01, 2.40307410242998e-01, 2.37259972495018e-01, 2.34533694922774e-01, 2.32114689466847e-01, 2.29987068747754e-01, 2.28133283900860e-01, 2.26534455999932e-01, 2.25170702304126e-01, 2.24021458390951e-01, 2.23065796562743e-01, 2.22282739834995e-01, 2.21651569466076e-01, 2.21152122521609e-01, 2.20765074545021e-01, 2.20472201191001e-01, 2.20256611815607e-01, 2.20102947621903e-01, 2.19997537111979e-01, 2.19928502320045e-01, 2.19885810575192e-01, 2.19861268288177e-01, 2.19848455358530e-01, 2.19842601106421e-01, 2.19840404987501e-01, 2.19839807591143e-01, 2.19839719416740e-01, 2.19839712536348e-01, 2.19839587955377e-01, 2.19838743901803e-01, 2.19835641083925e-01, 2.19827370205394e-01, 2.19809270070790e-01, 2.19774607990026e-01, 2.19714336106136e-01, 2.19616936198335e-01, 2.19468364222919e-01, 2.19252104213311e-01, 2.18949338989610e-01, 2.18539242244063e-01, 2.17999392811763e-01, 2.17306307220369e-01, 2.16436080964243e-01, 2.15365122539957e-01, 2.14070957458119e-01, 2.12533072719438e-01, 2.10733766263289e-01, 2.08658961389744e-01, 2.06298943837915e-01, 2.03648979670405e-01, 2.00709775739838e-01, 1.97487751354593e-01, 1.93995099512961e-01, 1.90249628068187e-01, 1.86274384446804e-01, 1.82097080894522e-01, 1.77749349439772e-01, 1.73265865718015e-01, 1.68683387618856e-01, 1.64039757890066e-01, 1.59372919251524e-01, 1.54719986516326e-01, 1.50116413280401e-01, 1.45595281748726e-01, 1.41186734169374e-01, 1.36917554125432e-01, 1.32810896475789e-01, 1.28886156723048e-01, 1.25158964438491e-01, 1.21641281252224e-01, 1.18341581778978e-01, 1.15265095516236e-01, 1.12414088933399e-01, 1.09788169299740e-01, 1.07384594859603e-01, 1.05198579353896e-01, 1.03223582283874e-01, 1.01451579484937e-01, 9.98733113809678e-02, 9.84785086263287e-02, 9.72560966507459e-02, 9.61943818697577e-02, 9.52812230135507e-02, 9.45041911994551e-02, 9.38507220952403e-02, 9.33082628689803e-02, 9.28644156863712e-02, 9.25070783899287e-02, 9.22245817787584e-02, 9.20058217092077e-02, 9.18403831649073e-02, 9.17186525995675e-02, 9.16319143229600e-02, 9.15724265370699e-02, 9.15334728621318e-02, 9.15093858100207e-02, 9.14955396196433e-02, 9.14883110876517e-02, 9.14850084092365e-02, 9.14837694763193e-02, 9.14834324580127e-02, 9.14833827148769e-02, 9.14833788003122e-02, 9.14833074974339e-02, 9.14828244108179e-02, 9.14810485466120e-02, 9.14763148465525e-02, 9.14659557768687e-02, 9.14461189006630e-02, 9.14116286069415e-02, 9.13558997689705e-02, 9.12709105928968e-02, 9.11472411694874e-02, 9.09741831019346e-02, 9.07399238799087e-02, 9.04318072610408e-02, 9.00366677142159e-02, 8.95412329793501e-02, 8.89325841370763e-02, 8.81986575324518e-02, 8.73287678850332e-02, 8.63141274847479e-02, 8.51483331347254e-02, 8.38277910595622e-02, 8.23520508468860e-02, 8.07240229171740e-02, 7.89500600220805e-02, 7.70398915148576e-02, 7.50064089580107e-02, 7.28653121028131e-02, 7.06346343192195e-02, 6.83341750947903e-02, 6.59848733293487e-02, 6.36081581997558e-02, 6.12253141143881e-02, 5.88568929102706e-02, 5.65222005372021e-02, 5.42388778583813e-02, 5.20225868143172e-02, 4.98868049437888e-02, 4.78427238815726e-02, 4.58992415088986e-02, 4.40630332536642e-02, 4.23386857196704e-02, 4.07288752371855e-02, 3.92345747469420e-02, 3.78552742430216e-02, 3.65892023983321e-02, 3.54335396594152e-02, 3.43846157925469e-02, 3.34380874174158e-02, 3.25890933335558e-02, 3.18323873246129e-02, 3.11624495594529e-02, 3.05735787162931e-02, 3.00599675887767e-02, 2.96157652649924e-02, 2.92351290507042e-02, 2.89122691759331e-02, 2.86414890010107e-02, 2.84172229578488e-02, 2.82340738621656e-02, 2.80868505583385e-02, 2.79706061606291e-02, 2.78806764782047e-02, 2.78127176048315e-02, 2.77627411583055e-02, 2.77271453089725e-02, 2.77027395681628e-02, 2.76867613319567e-02, 2.76768823913756e-02, 2.76712040079801e-02, 2.76682396778654e-02, 2.76668853189435e-02, 2.76663772631522e-02, 2.76662390606859e-02, 2.76662186623255e-02, 2.76662170428708e-02, 2.76661873639483e-02, 2.76659862854771e-02, 2.76652471087360e-02, 2.76632768039988e-02, 2.76589651907031e-02, 2.76507092781202e-02, 2.76363563778332e-02, 2.76131695511168e-02, 2.75778188648488e-02, 2.75264017202659e-02, 2.74544950886124e-02, 2.73572417308827e-02, 2.72294712982983e-02, 2.70658555482939e-02, 2.68610947654508e-02, 2.66101299298924e-02, 2.63083724011537e-02, 2.59519401521047e-02, 2.55378872376397e-02, 2.50644116015764e-02, 2.45310258794049e-02, 2.39386768374814e-02, 2.32898016395851e-02, 2.25883131784741e-02, 2.18395119258759e-02, 2.10499275796181e-02, 2.02270995140452e-02, 1.93793099693167e-02, 1.85152874924768e-02, 1.76439000385287e-02, 1.67738572116676e-02, 1.59134393703653e-02, 1.50702678918293e-02, 1.42511262211826e-02, 1.34618361817900e-02, 1.27071892951776e-02, 1.19909291955409e-02, 1.13157787068374e-02, 1.06835034313590e-02, 1.00950024909281e-02, 9.55041650659183e-03, 9.04924340491675e-03, 8.59045429969285e-03, 8.17260397747077e-03, 7.79393253989014e-03, 7.45245600069921e-03, 7.14604428284815e-03, 6.87248569236232e-03, 6.62953790877197e-03, 6.41496664557906e-03, 6.22657392282377e-03, 6.06221807144891e-03, 5.91982729309137e-03, 5.79740819747343e-03, 5.69305051084967e-03, 5.60492911813526e-03, 5.53130462242279e-03, 5.47052352560229e-03, 5.42101890290780e-03, 5.38131211445858e-03, 5.35001575490668e-03, 5.32583775048323e-03, 5.30758628976002e-03, 5.29417511298647e-03, 5.28462857595002e-03, 5.27808584433842e-03, 5.27380356562190e-03, 5.27115641374820e-03, 5.26963500308852e-03, 5.26884081689616e-03, 5.26847797573993e-03, 5.26834186619141e-03, 5.26830484154317e-03, 5.26829937679452e-03, 5.26829893893677e-03, 5.26829086445701e-03, 5.26823615898309e-03, 5.26803505969298e-03, 5.26749903226431e-03, 5.26632609875577e-03, 5.26408036314052e-03, 5.26017680056157e-03, 5.25387240970332e-03, 5.24426485179674e-03, 5.23029967849453e-03, 5.21078714561440e-03, 5.18442937171670e-03, 5.14985819410896e-03, 5.10568347749540e-03, 5.05055085034201e-03, 4.98320692202725e-03, 4.90256904744513e-03, 4.80779575374868e-03, 4.69835314062550e-03, 4.57407202740967e-03, 4.43519048298668e-03, 4.28237679325605e-03, 4.11672906370342e-03, 3.93974959351340e-03, 3.75329472172581e-03, 3.55950353867485e-03, 3.36071101504132e-03, 3.15935223756725e-03, 2.95786453032321e-03, 2.75859373732601e-03, 2.56371042571146e-03, 2.37514126347804e-03, 2.19451952564498e-03, 2.02315608883197e-03, 1.86202939001134e-03, 1.71179127186605e-03, 1.57278617474077e-03, 1.44508297913148e-03, 1.32851884140450e-03, 1.22275056098431e-03, 1.12730510305425e-03, 1.04162081225212e-03, 9.65074968426904e-04, 8.97001590658419e-04, 8.36708793762853e-04, 7.83500348667587e-04, 7.36699644171779e-04, 6.95671451978797e-04, 6.59834368605061e-04, 6.28660389153879e-04, 6.01667550483853e-04, 5.78413727218674e-04, 5.58493323673304e-04, 5.41534861604021e-04, 5.27198845809096e-04, 5.15176198561562e-04, 5.05186563543703e-04, 4.96975586067094e-04, 4.90311427128631e-04, 4.84981557420383e-04, 4.80790582038537e-04, 4.77559157495477e-04, 4.75123666049626e-04, 4.73336277153180e-04, 4.72065139838662e-04, 4.71194552559950e-04, 4.70625010066783e-04, 4.70273045692693e-04, 4.70070796652067e-04, 4.69965232191782e-04, 4.69917005261777e-04, 4.69898914701616e-04, 4.69893993731318e-04, 4.69893267407429e-04, 4.69893208723704e-04, 4.69892120634086e-04, 4.69884748689253e-04, 4.69857649264327e-04, 4.69785417081425e-04, 4.69627363692382e-04, 4.69324768459956e-04, 4.68798854214469e-04, 4.67949651820732e-04, 4.66655934239668e-04, 4.64776420763258e-04, 4.62152474796085e-04, 4.58612533130065e-04, 4.53978483392348e-04, 4.48074084125949e-04, 4.40735217864585e-04, 4.31821258800369e-04, 4.21226317698142e-04, 4.08889022438075e-04, 3.94800044091426e-04, 3.79007199072134e-04, 3.61617594634822e-04, 3.42795442463739e-04, 3.22756071241770e-04, 3.01762544339113e-04, 2.80130879652063e-04, 2.58230219351674e-04, 2.36447570699738e-04, 2.15127450161390e-04, 1.94552423963046e-04, 1.74952217115276e-04, 1.56483733107393e-04, 1.39258538922604e-04, 1.23392967218194e-04, 1.08951931451884e-04, 9.59089089178633e-05, 8.41780194491548e-05, 7.36229025406487e-05, 6.41119489586117e-05, 5.56275517098470e-05, 4.81813425849964e-05, 4.15725173681381e-05, 3.58776956232987e-05, 3.06613215030893e-05, 2.52093208104043e-05, 2.00531819416483e-05, 1.54644887473410e-05, 1.10732650071554e-05, 7.17207402800568e-06, 4.05059941499530e-06, 2.25410389731451e-06, 1.61593741410115e-06, 6.81452561934748e-07, 1.06185696324701e-06, 1.42304909900304e-06, 1.41118204055601e-06, 1.24200494866500e-06, 1.05564492969279e-06, 9.20447014607362e-07, 8.40658110234926e-07, 7.86469634306286e-07, 7.28243961134467e-07, 6.52201229911498e-07, 5.58980431719647e-07, 4.56719767782943e-07, 3.55170591603837e-07, -2.13162820728030e-14, -7.10542735760100e-15, -1.42108547152020e-14, -2.13162820728030e-14, -1.42108547152020e-14, -2.13162820728030e-14, -2.13162820728030e-14, -1.42108547152020e-14, -2.13162820728030e-14],
[ 5.30000000000000e+01, 5.29999999496902e+01, 5.29999983940120e+01, 5.29999878540460e+01, 5.29999491077585e+01, 5.29998458224036e+01, 5.29996197748935e+01, 5.29991868284066e+01, 5.29984338104710e+01, 5.29972164101290e+01, 5.29953581805350e+01, 5.29926506996475e+01, 5.29888549069264e+01, 5.29837035996204e+01, 5.29769050397852e+01, 5.29681475939371e+01, 5.29571053022188e+01, 5.29434442536630e+01, 5.29268296285333e+01, 5.29069332570979e+01, 5.28834415353029e+01, 5.28560635300575e+01, 5.28245390986206e+01, 5.27886468367339e+01, 5.27482116583827e+01, 5.27031117973188e+01, 5.26532850089963e+01, 5.25987337447632e+01, 5.25395290720806e+01, 5.24758131291631e+01, 5.24077999327327e+01, 5.23357744047638e+01, 5.22600895470405e+01, 5.21811617675165e+01, 5.20994644442079e+01, 5.20155198937183e+01, 5.19298899852541e+01, 5.18431657007066e+01, 5.17559559823362e+01, 5.16688762293701e+01, 5.15825368034381e+01, 5.14975318824112e+01, 5.14144289667106e+01, 5.13337592962190e+01, 5.12560093844056e+01, 5.11816138235575e+01, 5.11109494645588e+01, 5.10443310288513e+01, 5.09820081703052e+01, 5.09241639710309e+01, 5.08709148272643e+01, 5.08223116584915e+01, 5.07783423538967e+01, 5.07389353539694e+01, 5.07039642508114e+01, 5.06732532777208e+01, 5.06465835466533e+01, 5.06236998811468e+01, 5.06043180824555e+01, 5.05881324584353e+01, 5.05748234387357e+01, 5.05640650967624e+01, 5.05555323993601e+01, 5.05489080098164e+01, 5.05438884790749e+01, 5.05401896741985e+01, 5.05375513121669e+01, 5.05357404907280e+01, 5.05345541357249e+01, 5.05338203152825e+01, 5.05333984044556e+01, 5.05331781182639e+01, 5.05330774652363e+01, 5.05330397064148e+01, 5.05330294350625e+01, 5.05330279190311e+01, 5.05330277201597e+01, 5.05330230866956e+01, 5.05329916942754e+01, 5.05328762921299e+01, 5.05325686701663e+01, 5.05318954357910e+01, 5.05306060683914e+01, 5.05283637130603e+01, 5.05247391095439e+01, 5.05192079772835e+01, 5.05111520932382e+01, 5.04998642062113e+01, 5.04845568300481e+01, 5.04643748490062e+01, 5.04384117528593e+01, 5.04057291982452e+01, 5.03653794680191e+01, 5.03164302737450e+01, 5.02579912202702e+01, 5.01892411291001e+01, 5.01094553048109e+01, 5.00180317353433e+01, 4.99145151562315e+01, 4.97986178977953e+01, 4.96702364912561e+01, 4.95294631493242e+01, 4.93765914644422e+01, 4.92121159745040e+01, 4.90367256054380e+01, 4.88512913716214e+01, 4.86568490498057e+01, 4.84545777939864e+01, 4.82457757953570e+01, 4.80318341034378e+01, 4.78142096267576e+01, 4.75943981595591e+01, 4.73739080807548e+01, 4.71542351866143e+01, 4.69368389803620e+01, 4.67231206618244e+01, 4.65144030312210e+01, 4.63119125223321e+01, 4.61167635861323e+01, 4.59299456352918e+01, 4.57523127217507e+01, 4.55845760548329e+01, 4.54272993863483e+01, 4.52808972062488e+01, 4.51456356207315e+01, 4.50216357323010e+01, 4.49088793099941e+01, 4.48072165243901e+01, 4.47163755198363e+01, 4.46359735986493e+01, 4.45655297932094e+01, 4.45044785982045e+01, 4.44521846253483e+01, 4.44079579270714e+01, 4.43710697156949e+01, 4.43407681830024e+01, 4.43162941049188e+01, 4.42968959002683e+01, 4.42818438042593e+01, 4.42704428189876e+01, 4.42620441168483e+01, 4.42560545995404e+01, 4.42519443556877e+01, 4.42492518133511e+01, 4.42475864483057e+01, 4.42466289824007e+01, 4.42461290854562e+01, 4.42459006753413e+01, 4.42458149903245e+01, 4.42457916818755e+01, 4.42457882415956e+01, 4.42457879378843e+01, 4.42457819868121e+01, 4.42457416674123e+01, 4.42455934488885e+01, 4.42451983516065e+01, 4.42443336852061e+01, 4.42426777239085e+01, 4.42397979254037e+01, 4.42351432230291e+01, 4.42280408322437e+01, 4.42176979102657e+01, 4.42032082900533e+01, 4.41835643743472e+01, 4.41576741207125e+01, 4.41243828743648e+01, 4.40824996143244e+01, 4.40308269753731e+01, 4.39681942020121e+01, 4.38934919930612e+01, 4.38057080215275e+01, 4.37039617806543e+01, 4.35875373311495e+01, 4.34559125231432e+01, 4.33087833533674e+01, 4.31460823020936e+01, 4.29679897755530e+01, 4.27749381452147e+01, 4.25676082961615e+01, 4.23469190264950e+01, 4.21140100202969e+01, 4.18702193923305e+01, 4.16170569381076e+01, 4.13561742187229e+01, 4.10893325048650e+01, 4.08183694612640e+01, 4.05451653320932e+01, 4.02716093238275e+01, 3.99995668729419e+01, 3.97308485010003e+01, 3.94671809585472e+01, 3.92101813102315e+01, 3.89613345054405e+01, 3.87219748209966e+01, 3.84932713787248e+01, 3.82762177587175e+01, 3.80716255725515e+01, 3.78801217442193e+01, 3.77021491748165e+01, 3.75379704363815e+01, 3.73876741417120e+01, 3.72511836592048e+01, 3.71282678737795e+01, 3.70185537275270e+01, 3.69215403001930e+01, 3.68366142058460e+01, 3.67630660863141e+01, 3.67001079742802e+01, 3.66468912808599e+01, 3.66025251366710e+01, 3.65660947852466e+01, 3.65366796970550e+01, 3.65133710455066e+01, 3.64952881672655e+01, 3.64815936216875e+01, 3.64715064713767e+01, 3.64643134298289e+01, 3.64593775638468e+01, 3.64561442974616e+01, 3.64541445386819e+01, 3.64529948374488e+01, 3.64523945786203e+01, 3.64521203128526e+01, 3.64520174258817e+01, 3.64519894380920e+01, 3.64519853071515e+01, 3.64519849757990e+01, 3.64519788606640e+01, 3.64519374297172e+01, 3.64517851250663e+01, 3.64513791359548e+01, 3.64504906344990e+01, 3.64487890317651e+01, 3.64458298797834e+01, 3.64410469677483e+01, 3.64337490710692e+01, 3.64231217084253e+01, 3.64082341420620e+01, 3.63880517176746e+01, 3.63614534805154e+01, 3.63272548236511e+01, 3.62842347249534e+01, 3.62311669170493e+01, 3.61668541183759e+01, 3.60901642465439e+01, 3.60000673533124e+01, 3.58956718813009e+01, 3.57762587639311e+01, 3.56413118877855e+01, 3.54905435220242e+01, 3.53239134975106e+01, 3.51416411850918e+01, 3.49442096645161e+01, 3.47323618693746e+01, 3.45070889074286e+01, 3.42696111527466e+01, 3.40213530489620e+01, 3.37639128201385e+01, 3.34990284372316e+01, 3.32285412297016e+01, 3.29543584756388e+01, 3.26784161752422e+01, 3.24026430435320e+01, 3.21289265788703e+01, 3.18590818952924e+01, 3.15948238573143e+01, 3.13377429229138e+01, 3.10892849746732e+01, 3.08507352920985e+01, 3.06232066873421e+01, 3.04076316972247e+01, 3.02047586076843e+01, 3.00151509949040e+01, 2.98391904092728e+01, 2.96770818065281e+01, 2.95288613407240e+01, 2.93944061670559e+01, 2.92734459479646e+01, 2.91655758028197e+01, 2.90702704815495e+01, 2.89868995704866e+01, 2.89147435518870e+01, 2.88530105368734e+01, 2.88008534765203e+01, 2.87573876302803e+01, 2.87217080386700e+01, 2.86929067123816e+01, 2.86700892173556e+01, 2.86523903094406e+01, 2.86389882573585e+01, 2.86291174923954e+01, 2.86220792401728e+01, 2.86172498253417e+01, 2.86140863939346e+01, 2.86121298687674e+01, 2.86110050376001e+01, 2.86104177675482e+01, 2.86101494374557e+01, 2.86100487773225e+01, 2.86100213953019e+01, 2.86100173537722e+01, 2.86100170420505e+01, 2.86100114445299e+01, 2.86099735205016e+01, 2.86098341077214e+01, 2.86094624841234e+01, 2.86086491933894e+01, 2.86070916382867e+01, 2.86043830179586e+01, 2.86000051165770e+01, 2.85933253709430e+01, 2.85835985517871e+01, 2.85699732849397e+01, 2.85515035113384e+01, 2.85271648368774e+01, 2.84958755539105e+01, 2.84565219278720e+01, 2.84079871403571e+01, 2.83491830731771e+01, 2.82790839190797e+01, 2.81967604297682e+01, 2.81014134781367e+01, 2.79924055369433e+01, 2.78692886759976e+01, 2.77318277653714e+01, 2.75800177472310e+01, 2.74140940988373e+01, 2.72345359391821e+01, 2.70420616068263e+01, 2.68376169241438e+01, 2.66223567270078e+01, 2.63976205450221e+01, 2.61649035406742e+01, 2.59258239453554e+01, 2.56820882707420e+01, 2.54354555426789e+01, 2.51877017240765e+01, 2.49405853836341e+01, 2.46958155407780e+01, 2.44550224774606e+01, 2.42197321523101e+01, 2.39913446798474e+01, 2.37711171495399e+01, 2.35601508653602e+01, 2.33593829004904e+01, 2.31695816998515e+01, 2.29913463388450e+01, 2.28251089682727e+01, 2.26711399442082e+01, 2.25295551526807e+01, 2.24003250827926e+01, 2.22832852663857e+01, 2.21781477754739e+01, 2.20845135397358e+01, 2.20018853072345e+01, 2.19296811168658e+01, 2.18672481781900e+01, 2.18138770630104e+01, 2.17688161049074e+01, 2.17312858809194e+01, 2.17004936176709e+01, 2.16756473271040e+01, 2.16559694395025e+01, 2.16407096686420e+01, 2.16291568202466e+01, 2.16206492444130e+01, 2.16145836382023e+01, 2.16104219278481e+01, 2.16076960011711e+01, 2.16060101184798e+01, 2.16050409016939e+01, 2.16045348827074e+01, 2.16043036783441e+01, 2.16042169455878e+01, 2.16041933521734e+01, 2.16041898698348e+01, 2.16041896061627e+01, 2.16041849352366e+01, 2.16041532890302e+01, 2.16040369542993e+01, 2.16037268491684e+01, 2.16030481944617e+01, 2.16017485001921e+01, 2.15994883543882e+01, 2.15958354477063e+01, 2.15902622045789e+01, 2.15821473176471e+01, 2.15707813937408e+01, 2.15553768133937e+01, 2.15350817791695e+01, 2.15089983799224e+01, 2.14762043299946e+01, 2.14357778589627e+01, 2.13868250372108e+01, 2.13285086372315e+01, 2.12600774650794e+01, 2.11808949676558e+01, 2.10904658465985e+01, 2.09884594038396e+01, 2.08747284187941e+01, 2.07493225180597e+01, 2.06124952430577e+01, 2.04647043380407e+01, 2.03066051501367e+01, 2.01390374263837e+01, 1.99630061760583e+01, 1.97796576038275e+01, 1.95902513765123e+01, 1.93961306371142e+01, 1.91986912098237e+01, 1.89993513497554e+01, 1.87995231976551e+01, 1.86005868327889e+01, 1.84038675149247e+01, 1.82106164085868e+01, 1.80219948239487e+01, 1.78390618122664e+01, 1.76627648294491e+01, 1.74939331257709e+01, 1.73332735193482e+01, 1.71813682469357e+01, 1.70386746385066e+01, 1.69055264160715e+01, 1.67821364619693e+01, 1.66686009330199e+01, 1.65649046147766e+01, 1.64709274177284e+01, 1.63864519185572e+01, 1.63111718476334e+01, 1.62447014205612e+01, 1.61865854070368e+01, 1.61363098237457e+01, 1.60933131283668e+01, 1.60569977780642e+01, 1.60267419979734e+01, 1.60019115839295e+01, 1.59818715407785e+01, 1.59659973355808e+01, 1.59536855268288e+01, 1.59443635195160e+01, 1.59374981942870e+01, 1.59326031690943e+01, 1.59292444750015e+01, 1.59270444641390e+01, 1.59256838163020e+01, 1.59249015691976e+01, 1.59244931628947e+01, 1.59243065580384e+01, 1.59242365560103e+01, 1.59242175137516e+01, 1.59242147031537e+01, 1.59242144921613e+01, 1.59242107784546e+01, 1.59241856175447e+01, 1.59240931233726e+01, 1.59238465678228e+01, 1.59233069867988e+01, 1.59222736236900e+01, 1.59204765937649e+01, 1.59175721081063e+01, 1.59131405434932e+01, 1.59066875836557e+01, 1.58976485855729e+01, 1.58853962398358e+01, 1.58692514949723e+01, 1.58484976016671e+01, 1.58223970055239e+01, 1.57902106803612e+01, 1.57512193546622e+01, 1.57047459509647e+01, 1.56501784428768e+01, 1.55869922490396e+01, 1.55147712388903e+01, 1.54332264300331e+01, 1.53422115155467e+01, 1.52417344704068e+01, 1.51319646426600e+01, 1.50132349260989e+01, 1.48860388240535e+01, 1.47510224364931e+01, 1.46089716259237e+01, 1.44607948360614e+01, 1.43075022468928e+01, 1.41501821435892e+01, 1.39899755405338e+01, 1.38280502119148e+01, 1.36655753079109e+01, 1.35036976559403e+01, 1.33435206527178e+01, 1.31860863649436e+01, 1.30323611209314e+01, 1.28832245534277e+01, 1.27394618014314e+01, 1.26017584296885e+01, 1.24706975822116e+01, 1.23467589282225e+01, 1.22303190494226e+01, 1.21216530207967e+01, 1.20209370274024e+01, 1.19282519239523e+01, 1.18435876808295e+01, 1.17668486746768e+01, 1.16978597814881e+01, 1.16363732221051e+01, 1.15820760988849e+01, 1.15345985503440e+01, 1.14935224382646e+01, 1.14583904685427e+01, 1.14287156322506e+01, 1.14039908366343e+01, 1.13836985774299e+01, 1.13673204850714e+01, 1.13543465598471e+01, 1.13442838970277e+01, 1.13366646947511e+01, 1.13310533371192e+01, 1.13270523541878e+01, 1.13243070802362e+01, 1.13225088619658e+01, 1.13213967082475e+01, 1.13207573210278e+01, 1.13204235005876e+01, 1.13202709746571e+01, 1.13202137568155e+01, 1.13201981921658e+01, 1.13201958948558e+01, 1.13201957229205e+01, 1.13201927036492e+01, 1.13201722476413e+01, 1.13200970491766e+01, 1.13198965975590e+01, 1.13194579134881e+01, 1.13186177786787e+01, 1.13171567698911e+01, 1.13147953716447e+01, 1.13111923990256e+01, 1.13059459120558e+01, 1.12985967444858e+01, 1.12886347009947e+01, 1.12755073963967e+01, 1.12586316183406e+01, 1.12374069924184e+01, 1.12112316188006e+01, 1.11795192379189e+01, 1.11417173768332e+01, 1.10973258368549e+01, 1.10459148164919e+01, 1.09871419308759e+01, 1.09207673963294e+01, 1.08466666997900e+01, 1.07648401659848e+01, 1.06754189643965e+01, 1.05786672532169e+01, 1.04749803266359e+01, 1.03648788031257e+01, 1.02489990561082e+01, 1.01280802379231e+01, 1.00029483795577e+01, 9.87449815980514e+00, 9.74367302573468e+00, 9.61144440732299e+00, 9.47879079703378e+00, 9.34667745405745e+00, 9.21603743897299e+00, 9.08775458820310e+00, 8.96264890480764e+00, 8.84146468425152e+00, 8.72486152582211e+00, 8.61340821850533e+00, 8.50757934888312e+00, 8.40775436871633e+00, 8.31421878705381e+00, 8.22716711657551e+00, 8.14670720270680e+00, 8.07286559009129e+00, 8.00559362599268e+00, 7.94477405562898e+00, 7.89022792263793e+00, 7.84172164256851e+00, 7.79897416390653e+00, 7.76166416677268e+00, 7.72943727269335e+00, 7.70191324963431e+00, 7.67869319573224e+00, 7.65936667459777e+00, 7.64351875703868e+00, 7.63073690134283e+00, 7.62061757988936e+00, 7.61277253691183e+00, 7.60683454364980e+00, 7.60246250546313e+00, 7.59934577271999e+00, 7.59720751461738e+00, 7.59580703286478e+00, 7.59494091973946e+00, 7.59444300090674e+00, 7.59418304536539e+00, 7.59406427020797e+00, 7.59401971365529e+00, 7.59400759320295e+00, 7.59400580424995e+00, 7.59400567031096e+00, 7.59400331760558e+00, 7.59398737768717e+00, 7.59392878096112e+00, 7.59377258437273e+00, 7.59343075586423e+00, 7.59277612834146e+00, 7.59163777242568e+00, 7.58979801243298e+00, 7.58699128084876e+00, 7.58290497143898e+00, 7.57718240795691e+00, 7.56942799195469e+00, 7.55921452768967e+00, 7.54609264367497e+00, 7.52960213968496e+00, 7.50928498751415e+00, 7.48469960816041e+00, 7.45543594414725e+00, 7.42113075200056e+00, 7.38148246619996e+00, 7.33626494218866e+00, 7.28533938123517e+00, 7.22866378081655e+00, 7.16629934393076e+00, 7.09841341797919e+00, 7.02527871204153e+00, 6.94726874874365e+00, 6.86484972725508e+00, 6.77856918801092e+00, 6.68904205748098e+00, 6.59693479445631e+00, 6.50294844408485e+00, 6.40780142506708e+00, 6.31221282964086e+00, 6.21688691360108e+00, 6.12249930929885e+00, 6.02968532687444e+00, 5.93903053725852e+00, 5.85106367199924e+00, 5.76625174272064e+00, 5.68499718451398e+00, 5.60763676485880e+00, 5.53444197024604e+00, 5.46562058091465e+00, 5.40131916289784e+00, 5.34162623868180e+00, 5.28657593683914e+00, 5.23615196199912e+00, 5.19029176587836e+00, 5.14889083550333e+00, 5.11180704483645e+00, 5.07886504004463e+00, 5.04986064628505e+00, 5.02456529505040e+00, 5.00273047592586e+00, 4.98409221538168e+00, 4.96837557851428e+00, 4.95529917829450e+00, 4.94457966202452e+00, 4.93593612776964e+00, 4.92909440617141e+00, 4.92379112705216e+00, 4.91977747736296e+00, 4.91682254893052e+00, 4.91471617241570e+00, 4.91327113873356e+00, 4.91232472114935e+00, 4.91173942997925e+00, 4.91140295629903e+00, 4.91122728979610e+00, 4.91114702702362e+00, 4.91111691780050e+00, 4.91110872737253e+00, 4.91110751848290e+00, 4.91110742773319e+00, 4.91110583046203e+00, 4.91109500872125e+00, 4.91105522695085e+00, 4.91094918407499e+00, 4.91071711547890e+00, 4.91027269031793e+00, 4.90949987492748e+00, 4.90825091710260e+00, 4.90634558708519e+00, 4.90357178792103e+00, 4.89968761870873e+00, 4.89442493734475e+00, 4.88749442315752e+00, 4.87859208356872e+00, 4.86740708307873e+00, 4.85363069950207e+00, 4.83696613535335e+00, 4.81713883728280e+00, 4.79390691062452e+00, 4.76707116737870e+00, 4.73648432201546e+00, 4.70205885667014e+00, 4.66377311928201e+00, 4.62167529508496e+00, 4.57588499953433e+00, 4.52659237120641e+00, 4.47405468526642e+00, 4.41859064895581e+00, 4.36057266758419e+00, 4.30041747211330e+00, 4.23857557034941e+00, 4.17552001980653e+00, 4.11173502210426e+00, 4.04770480997905e+00, 3.98390324412384e+00, 3.92078446439235e+00, 3.85877485470839e+00, 3.79826648937305e+00, 3.73961213624352e+00, 3.68312180524327e+00, 3.62906075436338e+00, 3.57764880451696e+00, 3.52906077269255e+00, 3.48342781120942e+00, 3.44083943863667e+00, 3.40134606217408e+00, 3.36496181764774e+00, 3.33166758678233e+00, 3.30141408732479e+00, 3.27412496596428e+00, 3.24969985398887e+00, 3.22801736954872e+00, 3.20893806754762e+00, 3.19230734857945e+00, 3.17795834246832e+00, 3.16571478064687e+00, 3.15539386576258e+00, 3.14680913756585e+00, 3.13977332240305e+00, 3.13410114066021e+00, 3.12961203346944e+00, 3.12613275810401e+00, 3.12349979191018e+00, 3.12156147839041e+00, 3.12017984699742e+00, 3.11923204084174e+00, 3.11861129401783e+00, 3.11822741235169e+00, 3.11800672740879e+00, 3.11789151256055e+00, 3.11783887053394e+00, 3.11781912278766e+00, 3.11781375093085e+00, 3.11781295805638e+00, 3.11781289828286e+00, 3.11781184284222e+00, 3.11780469208158e+00, 3.11777840520510e+00, 3.11770833462756e+00, 3.11755498984534e+00, 3.11726132759549e+00, 3.11675068200255e+00, 3.11592543832774e+00, 3.11466654193553e+00, 3.11283391919869e+00, 3.11026786799962e+00, 3.10679145089448e+00, 3.10221389274260e+00, 3.09633494605497e+00, 3.08895014174111e+00, 3.07985679174612e+00, 3.06886055601256e+00, 3.05578233335373e+00, 3.04046518938207e+00, 3.02278100043992e+00, 3.00263647631608e+00, 2.97997823125037e+00, 2.95479660531853e+00, 2.92712799705450e+00, 2.89705555019139e+00, 2.86470813646765e+00, 2.83025768351033e+00, 2.79391500107634e+00, 2.75592434937647e+00, 2.71655706032713e+00, 2.67610456001720e+00, 2.63487114632888e+00, 2.59316685191069e+00, 2.55130067579883e+00, 2.50957440554102e+00, 2.46827718485745e+00, 2.42768091767365e+00, 2.38803654343486e+00, 2.34957117388122e+00, 2.31248604842298e+00, 2.27695524274306e+00, 2.24312505133161e+00, 2.21111395733445e+00, 2.18101310075281e+00, 2.15288715753167e+00, 2.12677554667619e+00, 2.10269388970232e+00, 2.08063565594865e+00, 2.06057393795476e+00, 2.04246331253596e+00, 2.02624175455819e+00, 2.01183258093457e+00, 1.99914641128623e+00, 1.98808313843547e+00, 1.97853390602545e+00, 1.97038309190538e+00, 1.96351029454036e+00, 1.95779231588857e+00, 1.95310512843144e+00, 1.94932580701797e+00, 1.94633439868964e+00, 1.94401569655244e+00, 1.94226087791409e+00, 1.94096896308416e+00, 1.94004805006027e+00, 1.93941628219193e+00, 1.93900251095226e+00, 1.93874662399815e+00, 1.93859951932471e+00, 1.93852271885394e+00, 1.93848762841013e+00, 1.93847446482589e+00, 1.93847088401677e+00, 1.93847035549702e+00, 1.93847031543596e+00, 1.93846960518827e+00, 1.93846479315885e+00, 1.93844710367537e+00, 1.93839995035866e+00, 1.93829675826220e+00, 1.93809913988798e+00, 1.93775550126809e+00, 1.93720014733476e+00, 1.93635294754753e+00, 1.93511961259512e+00, 1.93339262015393e+00, 1.93105281132740e+00, 1.92797165882143e+00, 1.92401418277584e+00, 1.91904246067037e+00, 1.91291964476227e+00, 1.90551436581330e+00, 1.89670536794159e+00, 1.88638618944883e+00, 1.87446968193140e+00, 1.86089214829917e+00, 1.84561688232359e+00, 1.82863690973350e+00, 1.80997676389279e+00, 1.78969317623066e+00, 1.76787461969227e+00, 1.74463970801123e+00, 1.72013451921867e+00, 1.69452897294271e+00, 1.66801244261792e+00, 1.64078882163174e+00, 1.61307128400394e+00, 1.58507698436076e+00, 1.55702192929535e+00, 1.52911622475801e+00, 1.50155986515798e+00, 1.47453918346356e+00, 1.44822403219322e+00, 1.42276571710138e+00, 1.39829566230839e+00, 1.37492475041125e+00, 1.35274325542881e+00, 1.33182127081233e+00, 1.31220952868978e+00, 1.29394050870616e+00, 1.27702974346378e+00, 1.26147724063878e+00, 1.24726895737729e+00, 1.23437827881198e+00, 1.22276746808029e+00, 1.21238906903483e+00, 1.20318725424111e+00, 1.19509911949450e+00, 1.18805593184641e+00, 1.18198434108563e+00, 1.17680756499643e+00, 1.17244655683549e+00, 1.16882115974302e+00, 1.16585124770816e+00, 1.16345784677588e+00, 1.16156422398782e+00, 1.16009692568322e+00, 1.15898674182638e+00, 1.15816956949469e+00, 1.15758714698081e+00, 1.15718763040657e+00, 1.15692598742017e+00, 1.15676418735022e+00, 1.15667117383559e+00, 1.15662261398200e+00, 1.15660042694540e+00, 1.15659210388049e+00, 1.15658983981054e+00, 1.15658950563875e+00, 1.15658948014828e+00, 1.15658902610387e+00, 1.15658594988924e+00, 1.15657464144527e+00, 1.15654449764195e+00, 1.15647853047430e+00, 1.15635220238675e+00, 1.15613253846860e+00, 1.15577756139836e+00, 1.15523609036102e+00, 1.15444793920023e+00, 1.15334454113466e+00, 1.15185001675711e+00, 1.14988268814346e+00, 1.14735702438580e+00, 1.14418598281782e+00, 1.14028368627608e+00, 1.13556835124540e+00, 1.12996535663405e+00, 1.12341032072008e+00, 1.11585203732766e+00, 1.10725511432217e+00, 1.09760216039499e+00, 1.08689538129602e+00, 1.07515747433952e+00, 1.06243174883896e+00, 1.04878144726872e+00, 1.03428829327125e+00, 1.01905034317200e+00, 1.00317926231295e+00, 9.86797181715282e-01, 9.70033311004549e-01, 9.53020488567368e-01, 9.35891839852665e-01, 9.18777691679963e-01, 9.01802857855195e-01, 8.85084373594019e-01, 8.68729717596779e-01, 8.52835524985053e-01, 8.37486764617054e-01, 8.22756332276974e-01, 8.08704997416925e-01, 7.95381635038353e-01, 7.82823674728562e-01, 7.71057704214670e-01, 7.60100173366439e-01, 7.49958154808269e-01, 7.40630127903508e-01, 7.32106762901331e-01, 7.24371690868310e-01, 7.17402252323367e-01, 7.11170223131333e-01, 7.05642520205188e-01, 7.00781892025319e-01, 6.96547600054060e-01, 6.92896096972575e-01, 6.89781706467322e-01, 6.87157307228127e-01, 6.84975021079083e-01, 6.83186901965243e-01, 6.81745619094997e-01, 6.80605124156017e-01, 6.79721289456488e-01, 6.79052501371970e-01, 6.78560191859063e-01, 6.78209290239664e-01, 6.77968578102565e-01, 6.77810932058719e-01, 6.77713442161661e-01, 6.77657397901967e-01, 6.77628138545359e-01, 6.77614769885786e-01, 6.77609754869501e-01, 6.77608390666258e-01, 6.77608189312814e-01, 6.77608173845279e-01, 6.77607896911738e-01, 6.77606020647850e-01, 6.77599123328655e-01, 6.77580737802501e-01, 6.77540502514880e-01, 6.77463450906117e-01, 6.77329469339902e-01, 6.77112951255360e-01, 6.76782672275152e-01, 6.76301907243349e-01, 6.75628805217904e-01, 6.74717031885812e-01, 6.73516680346118e-01, 6.71975440521642e-01, 6.70040004685596e-01, 6.67657672152728e-01, 6.64778100930953e-01, 6.61355139315226e-01, 6.57348657623459e-01, 6.52726291286079e-01, 6.47465002994828e-01, 6.41552374915278e-01, 6.34987552736206e-01, 6.27781781347309e-01, 6.19958495999761e-01, 6.11552960791734e-01, 6.02611475447546e-01, 5.93190198577602e-01, 5.83353658075318e-01, 5.73173034846938e-01, 5.62724313484289e-01, 5.52086392710642e-01, 5.41339240430915e-01, 5.30562164763708e-01, 5.19832255701992e-01, 5.09223034240648e-01, 4.98803328811796e-01, 4.88636384057266e-01, 4.78779195121390e-01, 4.69282052007557e-01, 4.60188272932754e-01, 4.51534102604597e-01, 4.43348750394641e-01, 4.35654543955387e-01, 4.28467175460945e-01, 4.21796019972938e-01, 4.15644508164355e-01, 4.10010538564862e-01, 4.04886917455435e-01, 4.00261817400136e-01, 3.96119248027091e-01, 3.92439534944877e-01, 3.89199804493551e-01, 3.86374473297280e-01, 3.83935742246749e-01, 3.81854094573860e-01, 3.80098797108538e-01, 3.78638402697597e-01, 3.77441250232479e-01, 3.76475956931507e-01, 3.75711895642333e-01, 3.75119648173317e-01, 3.74671424239125e-01, 3.74341434699033e-01, 3.74106207531057e-01, 3.73944835514166e-01, 3.73839145918943e-01, 3.73773784589368e-01, 3.73736209529092e-01, 3.73716592313436e-01, 3.73707629125960e-01, 3.73704266737903e-01, 3.73703352088057e-01, 3.73703217087701e-01, 3.73703206638112e-01, 3.73703018516679e-01, 3.73701743967104e-01, 3.73697058601273e-01, 3.73684569237703e-01, 3.73657237086107e-01, 3.73604894719783e-01, 3.73513877081777e-01, 3.73366784962258e-01, 3.73142397235668e-01, 3.72815745485617e-01, 3.72358361252203e-01, 3.71738701774412e-01, 3.70922754574266e-01, 3.69874814437466e-01, 3.68558418342566e-01, 3.66937414931357e-01, 3.64977135686566e-01, 3.62645625802074e-01, 3.59914884693893e-01, 3.56762060193510e-01, 3.53170537641859e-01, 3.49130866120419e-01, 3.44641469337830e-01, 3.39709098203059e-01, 3.34348995324248e-01, 3.28584757570567e-01, 3.22447900065654e-01, 3.15977142017211e-01, 3.09217450138917e-01, 3.02218887831380e-01, 2.95035326880139e-01, 2.87723082768643e-01, 2.80339534804021e-01, 2.72941788498898e-01, 2.65585430710338e-01, 2.58323418736076e-01, 2.51205133810394e-01, 2.44275618109640e-01, 2.37575003284100e-01, 2.31138128369430e-01, 2.24994336243228e-01, 2.19167430956048e-01, 2.13675773480354e-01, 2.08532490712678e-01, 2.03745771802858e-01, 1.99319226817529e-01, 1.95252285029333e-01, 1.91540613374251e-01, 1.88176539440725e-01, 1.85149467376796e-01, 1.82446279002640e-01, 1.80051715933978e-01, 1.77948741463197e-01, 1.76118883182980e-01, 1.74542558807680e-01, 1.73199388343953e-01, 1.72068495729192e-01, 1.71128802380061e-01, 1.70359313902821e-01, 1.69739399664842e-01, 1.69249063193085e-01, 1.68869199637037e-01, 1.68581834996402e-01, 1.68370340638035e-01, 1.68219615946576e-01, 1.68116231865042e-01, 1.68048528618790e-01, 1.68006662061742e-01, 1.67982594756971e-01, 1.67970029979486e-01, 1.67964289148649e-01, 1.67962135584197e-01, 1.67961549764264e-01, 1.67961463298539e-01, 1.67961456551481e-01, 1.67961334384700e-01, 1.67960506687898e-01, 1.67957464003472e-01, 1.67949353451341e-01, 1.67931604352937e-01, 1.67897615183797e-01, 1.67838515487148e-01, 1.67743015484866e-01, 1.67597355024483e-01, 1.67385362038260e-01, 1.67088628650966e-01, 1.66686810235802e-01, 1.66158048892484e-01, 1.65479517895058e-01, 1.64628077647187e-01, 1.63581026771688e-01, 1.62316924524546e-01, 1.60816453327371e-01, 1.59063283580039e-01, 1.57044897866577e-01, 1.54753329013069e-01, 1.52185766892359e-01, 1.49344992843432e-01, 1.46239608187066e-01, 1.42884034263091e-01, 1.39298274953234e-01, 1.35507447686621e-01, 1.31541104096705e-01, 1.27432375366681e-01, 1.23216988516226e-01, 1.18932207370818e-01, 1.14615755052391e-01, 1.10304773369336e-01, 1.06034868795135e-01, 1.01839285583203e-01, 9.77482350619852e-02, 9.37883975461631e-02, 8.99826008378923e-02, 8.63496680752291e-02, 8.29044185373817e-02, 7.96577984331250e-02, 7.66171148316843e-02, 7.37863446032350e-02, 7.11664911510326e-02, 6.87559643379316e-02, 6.65509627822942e-02, 6.45458420865523e-02, 6.27334570995615e-02, 6.11054706358800e-02, 5.96526249231957e-02, 5.83649752694910e-02, 5.72320879647776e-02, 5.62432062514091e-02, 5.53873893497610e-02, 5.46536300740570e-02, 5.40309565923209e-02, 5.35085234584756e-02, 5.30756962571175e-02, 5.27221331434831e-02, 5.24378653224957e-02, 5.22133771897373e-02, 5.20396855518541e-02, 5.19084161552001e-02, 5.18118747716088e-02, 5.17431093999505e-02, 5.16959597992468e-02, 5.16650906019649e-02, 5.16460046611869e-02, 5.16350340218565e-02, 5.16293069051841e-02, 5.16266902628928e-02, 5.16257086901959e-02, 5.16254416805282e-02, 5.16254022705311e-02, 5.16253991691187e-02, 5.16253426779869e-02, 5.16249599432967e-02, 5.16235529860509e-02, 5.16198026772727e-02, 5.16115958407042e-02, 5.15958811519255e-02, 5.15685606392466e-02, 5.15244234613874e-02, 5.14571284298526e-02, 5.13592413176767e-02, 5.12223321775949e-02, 5.10371364915017e-02, 5.07937818233017e-02, 5.04820786427373e-02, 5.00918701235307e-02, 4.96134311280656e-02, 4.90379015678641e-02, 4.83577343294215e-02, 4.75671335716399e-02, 4.66624561009112e-02, 4.56425473650484e-02, 4.45089849020306e-02, 4.32662061210252e-02, 4.19215040297374e-02, 4.04848835232627e-02, 3.89687813093929e-02, 3.73876633724777e-02, 3.57575238370487e-02, 3.40953169930245e-02, 3.24183591333949e-02, 3.07437381736548e-02, 2.90877666999521e-02, 2.74655085477562e-02, 2.58904010656522e-02, 2.43739859444290e-02, 2.29257520351140e-02, 2.15530849803329e-02, 2.02613115424342e-02, 1.90538217018172e-02, 1.79322490554341e-02, 1.68966896007987e-02, 1.59459402629665e-02, 1.50777410187501e-02, 1.42890076797997e-02, 1.35760458529433e-02, 1.29347399338613e-02, 1.23607139501303e-02, 1.18494635036441e-02, 1.13964599200784e-02, 1.09972290152101e-02, 1.06474077008301e-02, 1.03427820649742e-02, 1.00793106628601e-02, 9.85313662827991e-03, 9.66059193160618e-03, 9.49819672123198e-03, 9.36265623121812e-03, 9.25085724491481e-03, 9.15986559155613e-03, 9.08692563647406e-03, 9.02946222001333e-03, 8.98508502293538e-03, 8.95159490791286e-03, 8.92699142753628e-03, 8.90948042250983e-03, 8.89748047492844e-03, 8.88962694285311e-03, 8.88477238484597e-03, 8.88198237505833e-03, 8.88052599345457e-03, 8.87986061887602e-03, 8.87961102332468e-03, 8.87954312816674e-03, 8.87953310702017e-03, 8.87953231142546e-03, 8.87951773101747e-03, 8.87941894715283e-03, 8.87905581458881e-03, 8.87808789304501e-03, 8.87596990750694e-03, 8.87191481159277e-03, 8.86486640635553e-03, 8.85348354261595e-03, 8.83613799592098e-03, 8.81092808322848e-03, 8.77570991294618e-03, 8.72814773585162e-03, 8.66578411820740e-03, 8.58612953891225e-03, 8.48676951812166e-03, 8.36548558027772e-03, 8.22038438634110e-03, 8.05002746211869e-03, 7.85355239225538e-03, 7.63077545814639e-03, 7.38226576144285e-03, 7.10938209205381e-03, 6.81426621478920e-03, 6.49978971647158e-03, 6.16945572603811e-03, 5.82726118328480e-03, 5.47752929093548e-03, 5.12472476491155e-03, 4.77326605685136e-03, 4.42734863770511e-03, 4.09079174327331e-03, 3.76691798702211e-03, 3.45847143135103e-03, 3.16757566108095e-03, 2.89572968592466e-03, 2.64383654696587e-03, 2.41225756157348e-03, 2.20088425399467e-03, 2.00922007705628e-03, 1.83646483571209e-03, 1.68159603795459e-03, 1.54344297042286e-03, 1.42075088279370e-03, 1.31223406319947e-03, 1.21661768039161e-03, 1.13266903391462e-03, 1.05921933933928e-03, 9.95177447876472e-04, 9.39537006267699e-04, 8.91378536842069e-04, 8.49867787515279e-04, 8.14251507449626e-04, 7.83851592174756e-04, 7.58058344395351e-04, 7.36323430032826e-04, 7.18152971394659e-04, 7.03101104804205e-04, 6.90764231293883e-04, 6.80776103038738e-04, 6.72803812520328e-04, 6.66544685202553e-04, 6.61724019167309e-04, 6.58093566016760e-04, 6.55430606762764e-04, 6.53537446183350e-04, 6.52241131149367e-04, 6.51393193933814e-04, 6.50869232430296e-04, 6.50568163628407e-04, 6.50411024125219e-04, 6.50339236173636e-04, 6.50312307733714e-04, 6.50304982691807e-04, 6.50303901544191e-04, 6.50303814929032e-04, 6.50302217593435e-04, 6.50291395430713e-04, 6.50251613365072e-04, 6.50145579186301e-04, 6.49913577703387e-04, 6.49469465216157e-04, 6.48697769555895e-04, 6.47452166347762e-04, 6.45555626881844e-04, 6.42802543133314e-04, 6.38963118376523e-04, 6.33790249466415e-04, 6.27029005421775e-04, 6.18428617343625e-04, 6.07756637506895e-04, 5.94814617457473e-04, 5.79454327507278e-04, 5.61593242217384e-04, 5.41227808305678e-04, 5.18442952149201e-04, 4.93416424355075e-04, 4.66416939971737e-04, 4.37795644934624e-04, 4.07971164875676e-04, 3.77409275067464e-04, 3.46598937063902e-04, 3.16026935443858e-04, 2.86153502329967e-04, 2.57391089988346e-04, 2.30087905102039e-04, 2.04517116337399e-04, 1.80871986827924e-04, 1.59266671488467e-04, 1.39742018596678e-04, 1.22275331037258e-04, 1.06792632450947e-04, 9.31817011533553e-05, 8.13042849401313e-05, 7.10065680635807e-05, 6.21277779799811e-05, 5.45073668192231e-05, 4.79911274311462e-05, 4.24359445432287e-05, 3.77125838895154e-05, 3.37064111803898e-05, 3.03163753088143e-05, 2.74531160258107e-05, 2.50376926160811e-05, 2.30013664079820e-05, 2.12852484935411e-05, 1.98394970354343e-05, 1.86227901934899e-05, 1.76016257498190e-05, 1.67485597017958e-05, 1.60400363711233e-05, 1.54551730489061e-05, 1.49756423937220e-05, 1.45857339433064e-05, 1.42720397349194e-05, 1.40229396450309e-05, 1.38281929054074e-05, 1.36787088678147e-05, 1.35664241014410e-05, 1.34842199699392e-05, 1.34258584623126e-05, 1.33859334994213e-05, 1.33598341633956e-05, 1.33437130145353e-05, 1.33344519426259e-05, 1.33296188522536e-05, 1.33274110396542e-05, 1.33265828878848e-05, 1.33263576103104e-05, 1.33263243711212e-05, 1.33263216923751e-05, 1.33262720680705e-05, 1.33259359031968e-05, 1.33247001627979e-05, 1.33214065769494e-05, 1.33142008067466e-05, 1.33004092646161e-05, 1.32764518880890e-05, 1.32378012551726e-05, 1.31789999500143e-05, 1.30937506455098e-05, 1.29750960482511e-05, 1.28157073930879e-05, 1.26082974318820e-05, 1.23461562182570e-05, 1.20237617835528e-05, 1.16373446630291e-05, 1.11852484891983e-05, 1.06680713543028e-05, 1.00889228491496e-05, 9.45419028397509e-06, 8.77419470413088e-06, 8.06180480594776e-06, 7.32879029641254e-06, 6.58439217460227e-06, 5.83950442489822e-06, 5.11305906059079e-06, 4.43446526077196e-06, 3.81947790373260e-06, 3.25638713150056e-06, 2.78829873678887e-06, 2.38051244849657e-06, 1.99962478575344e-06, 1.68779554599041e-06, 1.58543906536579e-06, 1.46909718523602e-06, 1.29894671374586e-06, 1.04616063367757e-06, 7.57381847904526e-07, 6.51209546731479e-07, 5.21501362982235e-07, 3.99544049400902e-07, -3.55271367880050e-14, -3.55271367880050e-14, -3.55271367880050e-14, -3.55271367880050e-14, 3.90736332178676e-07, 4.89634970790576e-07, 5.10670496112198e-07, 4.54341503086653e-07, -2.84217094304040e-14, -3.55271367880050e-14, -3.55271367880050e-14, -3.55271367880050e-14, -3.55271367880050e-14, -2.84217094304040e-14, -3.55271367880050e-14, -2.84217094304040e-14, -3.55271367880050e-14, -3.55271367880050e-14, -3.55271367880050e-14, -3.55271367880050e-14, -3.55271367880050e-14, -3.55271367880050e-14, -2.84217094304040e-14, -2.84217094304040e-14, -3.55271367880050e-14, -4.26325641456060e-14, -3.55271367880050e-14, -3.55271367880050e-14, -4.26325641456060e-14, -3.55271367880050e-14, -3.55271367880050e-14, -2.84217094304040e-14, -3.55271367880050e-14],
[ 5.40000000000000e+01, 5.39999999483440e+01, 5.39999983510373e+01, 5.39999875290319e+01, 5.39999477459303e+01, 5.39998416967529e+01, 5.39996096003956e+01, 5.39991650685351e+01, 5.39983919000716e+01, 5.39971419218160e+01, 5.39952339640175e+01, 5.39924540247294e+01, 5.39885566413351e+01, 5.39832674523415e+01, 5.39762868992987e+01, 5.39672949887906e+01, 5.39559570088794e+01, 5.39419300737253e+01, 5.39248703542433e+01, 5.39044408408615e+01, 5.38803194753672e+01, 5.38522074807857e+01, 5.38198377095369e+01, 5.37829828195692e+01, 5.37414630754778e+01, 5.36951535578131e+01, 5.36439905512866e+01, 5.35879768750143e+01, 5.35271859196246e+01, 5.34617641711674e+01, 5.33919320334121e+01, 5.33179828095423e+01, 5.32402797701770e+01, 5.31592513132665e+01, 5.30753843067121e+01, 5.29892157891900e+01, 5.29013232810217e+01, 5.28123140183228e+01, 5.27228134652336e+01, 5.26334534783814e+01, 5.25448604950170e+01, 5.24576440939524e+01, 5.23723862406219e+01, 5.22896314792765e+01, 5.22098782816034e+01, 5.21335717064735e+01, 5.20610974736247e+01, 5.19927775072628e+01, 5.19288669650338e+01, 5.18695527338018e+01, 5.18149533456648e+01, 5.17651202446984e+01, 5.17200403158952e+01, 5.16796395715787e+01, 5.16437878762781e+01, 5.16123045780223e+01, 5.15849649019064e+01, 5.15615069505756e+01, 5.15416391462072e+01, 5.15250479401238e+01, 5.15114056099525e+01, 5.15003779609739e+01, 5.14916317486713e+01, 5.14848416441243e+01, 5.14796965732725e+01, 5.14759052754603e+01, 5.14732009460759e+01, 5.14713448522417e+01, 5.14701288388389e+01, 5.14693766738504e+01, 5.14689442160397e+01, 5.14687184231713e+01, 5.14686152540520e+01, 5.14685765513502e+01, 5.14685660232388e+01, 5.14685644693102e+01, 5.14685642654676e+01, 5.14685595161781e+01, 5.14685273390240e+01, 5.14684090521143e+01, 5.14680937404349e+01, 5.14674036773385e+01, 5.14660820810530e+01, 5.14637836798211e+01, 5.14600684923112e+01, 5.14543991532469e+01, 5.14461420281587e+01, 5.14345722659924e+01, 5.14188828343308e+01, 5.13981974698165e+01, 5.13715873569912e+01, 5.13380912235825e+01, 5.12967384109220e+01, 5.12465743466493e+01, 5.11866877156981e+01, 5.11162384986567e+01, 5.10344859302682e+01, 5.09408153350903e+01, 5.08347627364704e+01, 5.07160361269037e+01, 5.05845323512141e+01, 5.04403487034737e+01, 5.02837885787803e+01, 5.01153608412405e+01, 4.99357729414718e+01, 4.97459181972445e+01, 4.95468579887329e+01, 4.93397998683164e+01, 4.91260727126531e+01, 4.89071000444148e+01, 4.86843725410330e+01, 4.84594205669577e+01, 4.82337873626561e+01, 4.80090033425694e+01, 4.77865618252886e+01, 4.75678964513005e+01, 4.73543605258314e+01, 4.71472085323691e+01, 4.69475800688427e+01, 4.67564864418076e+01, 4.65748001051363e+01, 4.64032470525633e+01, 4.62424021813555e+01, 4.60926875540110e+01, 4.59543734102060e+01, 4.58275817297770e+01, 4.57122921198224e+01, 4.56083497900513e+01, 4.55154753826517e+01, 4.54332764285081e+01, 4.53612602045862e+01, 4.52988477642014e+01, 4.52453889015482e+01, 4.52001777950474e+01, 4.51624690528178e+01, 4.51314938607077e+01, 4.51064759119775e+01, 4.50866467811015e+01, 4.50712603952413e+01, 4.50596062583349e+01, 4.50510210964435e+01, 4.50448986202453e+01, 4.50406971417048e+01, 4.50379448363320e+01, 4.50362425084498e+01, 4.50352637919742e+01, 4.50347528001399e+01, 4.50345193206378e+01, 4.50344317339123e+01, 4.50344079081518e+01, 4.50344043915177e+01, 4.50344040810659e+01, 4.50343979979146e+01, 4.50343567836601e+01, 4.50342052755571e+01, 4.50338014094758e+01, 4.50329175528287e+01, 4.50312248401275e+01, 4.50282811312932e+01, 4.50235231346243e+01, 4.50162631458121e+01, 4.50056907501093e+01, 4.49908797144799e+01, 4.49708001579632e+01, 4.49443359299074e+01, 4.49103069470917e+01, 4.48674960443500e+01, 4.48146796845633e+01, 4.47506616617131e+01, 4.46743087275138e+01, 4.45845868934705e+01, 4.44805970236970e+01, 4.43616082579610e+01, 4.42270878068025e+01, 4.40767257556359e+01, 4.39104537108075e+01, 4.37284564159472e+01, 4.35311758455892e+01, 4.33193077113781e+01, 4.30937907440339e+01, 4.28557894826196e+01, 4.26066715591199e+01, 4.23479805833461e+01, 4.20814057205787e+01, 4.18087489573651e+01, 4.15318909319556e+01, 4.12527561183642e+01, 4.09732781203578e+01, 4.06953658420589e+01, 4.04208713196442e+01, 4.01515599841787e+01, 3.98890840527657e+01, 3.96349596096865e+01, 3.93905477569054e+01, 3.91570400117887e+01, 3.89354479383114e+01, 3.87265968393694e+01, 3.85311232248125e+01, 3.83494757048067e+01, 3.81819189352974e+01, 3.80285402510806e+01, 3.78892586501005e+01, 3.77638358287024e+01, 3.76518890025270e+01, 3.75529052751406e+01, 3.74662573326209e+01, 3.73912202457517e+01, 3.73269891525898e+01, 3.72726975747332e+01, 3.72274360934119e+01, 3.71902710800374e+01, 3.71602631440761e+01, 3.71364849332369e+01, 3.71180379011663e+01, 3.71040676499042e+01, 3.70937774614434e+01, 3.70864396570016e+01, 3.70814044650355e+01, 3.70781061391659e+01, 3.70760661432287e+01, 3.70748933094740e+01, 3.70742809733404e+01, 3.70740011893422e+01, 3.70738962322909e+01, 3.70738676813890e+01, 3.70738634673342e+01, 3.70738631293149e+01, 3.70738568911439e+01, 3.70738146266103e+01, 3.70736592576087e+01, 3.70732451000857e+01, 3.70723387223743e+01, 3.70706028849048e+01, 3.70675841997838e+01, 3.70627050699959e+01, 3.70552603767836e+01, 3.70444192783035e+01, 3.70292323603924e+01, 3.70086442384977e+01, 3.69815115466226e+01, 3.69466260642659e+01, 3.69027425282283e+01, 3.68486104585685e+01, 3.67830091065241e+01, 3.67047844201502e+01, 3.66128867372257e+01, 3.65064077729607e+01, 3.63846153906240e+01, 3.62469846426939e+01, 3.60932236601554e+01, 3.59232931527404e+01, 3.57374185587548e+01, 3.55360942349280e+01, 3.53200794794863e+01, 3.50903866020110e+01, 3.48482616533826e+01, 3.45951587706563e+01, 3.43327093443950e+01, 3.40626873619250e+01, 3.37869723177107e+01, 3.35075110264544e+01, 3.32262795525513e+01, 3.29452463125732e+01, 3.26663372428102e+01, 3.23914037682034e+01, 3.21221941660012e+01, 3.18603287809932e+01, 3.16072794096561e+01, 3.13643530222260e+01, 3.11326798371607e+01, 3.09132056126469e+01, 3.07066878901494e+01, 3.05136958294995e+01, 3.03346132215203e+01, 3.01696442522505e+01, 3.00188216148157e+01, 2.98820166090061e+01, 2.97589509218955e+01, 2.96492098343847e+01, 2.95522566404807e+01, 2.94674480938711e+01, 2.93940507082283e+01, 2.93312577341431e+01, 2.92782066186467e+01, 2.92339967259944e+01, 2.91977070645444e+01, 2.91684137284927e+01, 2.91452067295027e+01, 2.91272058665042e+01, 2.91135752664428e+01, 2.91035362281763e+01, 2.90963780187081e+01, 2.90914663068497e+01, 2.90882489740986e+01, 2.90862591143055e+01, 2.90851151194621e+01, 2.90845178443671e+01, 2.90842449429157e+01, 2.90841425679126e+01, 2.90841147194062e+01, 2.90841106090242e+01, 2.90841102919920e+01, 2.90841045991109e+01, 2.90840660290018e+01, 2.90839242411651e+01, 2.90835462865787e+01, 2.90827191408333e+01, 2.90811350525000e+01, 2.90783802930815e+01, 2.90739278251572e+01, 2.90671343243564e+01, 2.90572418965504e+01, 2.90433847212991e+01, 2.90246007231344e+01, 2.89998482214817e+01, 2.89680273373950e+01, 2.89280057429281e+01, 2.88786481324188e+01, 2.88188485835721e+01, 2.87475647731447e+01, 2.86638528335382e+01, 2.85669015008920e+01, 2.84560641305195e+01, 2.83308871576352e+01, 2.81911336712217e+01, 2.80368009500910e+01, 2.78681310767389e+01, 2.76856140800290e+01, 2.74899834357908e+01, 2.72822041416029e+01, 2.70634539425917e+01, 2.68350985871784e+01, 2.65986622136495e+01, 2.63557941025127e+01, 2.61082330821043e+01, 2.58577708616558e+01, 2.56062155054631e+01, 2.53553561683284e+01, 2.51069300930476e+01, 2.48625927259059e+01, 2.46238916343488e+01, 2.43922447136407e+01, 2.41689229542610e+01, 2.39550378234654e+01, 2.37515331111718e+01, 2.35591809200419e+01, 2.33785813557507e+01, 2.32101654023388e+01, 2.30542004479571e+01, 2.29107979505183e+01, 2.27799227888647e+01, 2.26614039195581e+01, 2.25549460394169e+01, 2.24601420287729e+01, 2.23764860123333e+01, 2.23033869188559e+01, 2.22401824456753e+01, 2.21861533398641e+01, 2.21405378965952e+01, 2.21025465504449e+01, 2.20713764011444e+01, 2.20462254763434e+01, 2.20263064952235e+01, 2.20108598631640e+01, 2.19991656036447e+01, 2.19905539230149e+01, 2.19844141095510e+01, 2.19802014919522e+01, 2.19774422242567e+01, 2.19757357227073e+01, 2.19747546525547e+01, 2.19742424451667e+01, 2.19740084132999e+01, 2.19739206198540e+01, 2.19738967379068e+01, 2.19738932129814e+01, 2.19738929460847e+01, 2.19738882180361e+01, 2.19738561848166e+01, 2.19737384273885e+01, 2.19734245298987e+01, 2.19727375758580e+01, 2.19714219879478e+01, 2.19691342047706e+01, 2.19654366335657e+01, 2.19597952545312e+01, 2.19515811775245e+01, 2.19400763622818e+01, 2.19244836054830e+01, 2.19039407692708e+01, 2.18775390751127e+01, 2.18443451156910e+01, 2.18034260506020e+01, 2.17538772576332e+01, 2.16948515225648e+01, 2.16255886822770e+01, 2.15454445059133e+01, 2.14539175246981e+01, 2.13506725186236e+01, 2.12355594493143e+01, 2.11086267981797e+01, 2.09701285244151e+01, 2.08205241859600e+01, 2.06604721461180e+01, 2.04908161887494e+01, 2.03125662497336e+01, 2.01268743040891e+01, 1.99350066926279e+01, 1.97383143041130e+01, 1.95382020369792e+01, 1.93360988534833e+01, 1.91334295295681e+01, 1.89315889294581e+01, 1.87319193355686e+01, 1.85356910814810e+01, 1.83440865005669e+01, 1.81581870345441e+01, 1.79789632494916e+01, 1.78072674732313e+01, 1.76438287803423e+01, 1.74892500888457e+01, 1.73440071770414e+01, 1.72084494666916e+01, 1.70828024431261e+01, 1.69671715934101e+01, 1.68615477440880e+01, 1.67658136753761e+01, 1.66797518836020e+01, 1.66030533608385e+01, 1.65353272604428e+01, 1.64761113181881e+01, 1.64248828984907e+01, 1.63810705315373e+01, 1.63440657982589e+01, 1.63132354056105e+01, 1.62879332753245e+01, 1.62675124471909e+01, 1.62513365758293e+01, 1.62387907811696e+01, 1.62292916008363e+01, 1.62222957903308e+01, 1.62173077266514e+01, 1.62138851940855e+01, 1.62116433675187e+01, 1.62102568577000e+01, 1.62094597423113e+01, 1.62090435733626e+01, 1.62088534216753e+01, 1.62087820891063e+01, 1.62087626849079e+01, 1.62087598208885e+01, 1.62087596058856e+01, 1.62087558215918e+01, 1.62087301824440e+01, 1.62086359302183e+01, 1.62083846883411e+01, 1.62078348513930e+01, 1.62067818469220e+01, 1.62049506603800e+01, 1.62019909682393e+01, 1.61974751711585e+01, 1.61908995570091e+01, 1.61816887505986e+01, 1.61692035209493e+01, 1.61527519164278e+01, 1.61316035825236e+01, 1.61050069881567e+01, 1.60722091479591e+01, 1.60324772865516e+01, 1.59851217554887e+01, 1.59295193951663e+01, 1.58651364442967e+01, 1.57915500494754e+01, 1.57084674254708e+01, 1.56157417678452e+01, 1.55133841234393e+01, 1.54015705766388e+01, 1.52806443024074e+01, 1.51511122615001e+01, 1.50136365599817e+01, 1.48690207558489e+01, 1.47181916611409e+01, 1.45621774452784e+01, 1.44020830734983e+01, 1.42390642831191e+01, 1.40743013752361e+01, 1.39089740515179e+01, 1.37442383460863e+01, 1.35812064120841e+01, 1.34209295722555e+01, 1.32643846998878e+01, 1.31124637230607e+01, 1.29659658791961e+01, 1.28255922930530e+01, 1.26919424857061e+01, 1.25655125060425e+01, 1.24466944718068e+01, 1.23357773872027e+01, 1.22329491560362e+01, 1.21382997331818e+01, 1.20518253598637e+01, 1.19734338191533e+01, 1.19029506351780e+01, 1.18401261278158e+01, 1.17846432260007e+01, 1.17361259367421e+01, 1.16941483618006e+01, 1.16582441476417e+01, 1.16279162452127e+01, 1.16026468435773e+01, 1.15819073258396e+01, 1.15651680784325e+01, 1.15519079678050e+01, 1.15416232843664e+01, 1.15338359448455e+01, 1.15281007433952e+01, 1.15240114506459e+01, 1.15212055795375e+01, 1.15193676672661e+01, 1.15182309632201e+01, 1.15175774616017e+01, 1.15172362720582e+01, 1.15170803790974e+01, 1.15170218981621e+01, 1.15170059899200e+01, 1.15170036418964e+01, 1.15170034661656e+01, 1.15170003802434e+01, 1.15169794726652e+01, 1.15169026141788e+01, 1.15166977375410e+01, 1.15162493693603e+01, 1.15153906880893e+01, 1.15138974261252e+01, 1.15114838962505e+01, 1.15078013788211e+01, 1.15024390551125e+01, 1.14949276124827e+01, 1.14847455769249e+01, 1.14713283469682e+01, 1.14540798095504e+01, 1.14323863147188e+01, 1.14056326750215e+01, 1.13732197425560e+01, 1.13345830091615e+01, 1.12892115820205e+01, 1.12366668173730e+01, 1.11765998579334e+01, 1.11087673216532e+01, 1.10330444341474e+01, 1.09494349837868e+01, 1.08580776024294e+01, 1.07592480279608e+01, 1.06533571772349e+01, 1.05409450393330e+01, 1.04226705799384e+01, 1.02992980201386e+01, 1.01716800103005e+01, 1.00407383551166e+01, 9.90744305214096e+00, 9.77279047503155e+00, 9.63778155651599e+00, 9.50340079928974e+00, 9.37059686450986e+00, 9.24026536205746e+00, 9.11323430519610e+00, 8.99025251035362e+00, 8.87198103854028e+00, 8.75898760579377e+00, 8.65174375011743e+00, 8.55062444065693e+00, 8.45590975462332e+00, 8.36778822673409e+00, 8.28636148871316e+00, 8.21164985426515e+00, 8.14359855869666e+00, 8.08208442324140e+00, 8.02692277472148e+00, 7.97787450575759e+00, 7.93465320547009e+00, 7.89693232331684e+00, 7.86435234863320e+00, 7.83652799595156e+00, 7.81305538252462e+00, 7.79351917156669e+00, 7.77749963495412e+00, 7.76457956523574e+00, 7.75435094173687e+00, 7.74642123228071e+00, 7.74041919342983e+00, 7.73600002064639e+00, 7.73284969729558e+00, 7.73068839913981e+00, 7.72927282918126e+00, 7.72839738577758e+00, 7.72789410336839e+00, 7.72763134767458e+00, 7.72751129313125e+00, 7.72746625664109e+00, 7.72745400563453e+00, 7.72745219741201e+00, 7.72745206203032e+00, 7.72744968398303e+00, 7.72743357236967e+00, 7.72737434447777e+00, 7.72721646545583e+00, 7.72687095509863e+00, 7.72620927683899e+00, 7.72505866101810e+00, 7.72319908946344e+00, 7.72036213895959e+00, 7.71623184569715e+00, 7.71044770905551e+00, 7.70260989883238e+00, 7.69228666340115e+00, 7.67902385636265e+00, 7.66235640669460e+00, 7.64182145480881e+00, 7.61697276916274e+00, 7.58739595223895e+00, 7.55272384979927e+00, 7.51265150354283e+00, 7.46694994502095e+00, 7.41547812714516e+00, 7.35819233558841e+00, 7.29515251865350e+00, 7.22652511861833e+00, 7.15258217246051e+00, 7.07369666208190e+00, 6.99033431610769e+00, 6.90304227679115e+00, 6.81243522613664e+00, 6.71917969737725e+00, 6.62397736909669e+00, 6.52754814460071e+00, 6.43061376179024e+00, 6.33388256944853e+00, 6.23803596077570e+00, 6.14371679312677e+00, 6.05151996273116e+00, 5.96198516001525e+00, 5.87559171537416e+00, 5.79275536166437e+00, 5.71382668811065e+00, 5.63909103685089e+00, 5.56876959203023e+00, 5.50302142573881e+00, 5.44194628929871e+00, 5.38558796777922e+00, 5.33393804680003e+00, 5.28693997142468e+00, 5.24449330577054e+00, 5.20645812781315e+00, 5.17265951589341e+00, 5.14289210089395e+00, 5.11692467030036e+00, 5.09450481695502e+00, 5.07536362608834e+00, 5.05922038939036e+00, 5.04578732510250e+00, 5.03477426943060e+00, 5.02589328844711e+00, 5.01886314279355e+00, 5.01341352180116e+00, 5.00928895100220e+00, 5.00625226911355e+00, 5.00408756880046e+00, 5.00260250072951e+00, 5.00162985285925e+00, 5.00102833620156e+00, 5.00068253340815e+00, 5.00050199595744e+00, 5.00041950753792e+00, 5.00038856338632e+00, 5.00038014583647e+00, 5.00037890342410e+00, 5.00037881015786e+00, 5.00037716859404e+00, 5.00036604676347e+00, 5.00032516182507e+00, 5.00021617826031e+00, 4.99997767379482e+00, 4.99952092224081e+00, 4.99872666823130e+00, 4.99744305207405e+00, 4.99548482318667e+00, 4.99263395651156e+00, 4.98864175627475e+00, 4.98323249368942e+00, 4.97610857825387e+00, 4.96695720545258e+00, 4.95545835742315e+00, 4.94129395953976e+00, 4.92415791869611e+00, 4.90376669388505e+00, 4.87986998341341e+00, 4.85226106343987e+00, 4.82078628695226e+00, 4.78535325693381e+00, 4.74593722596457e+00, 4.70258534716210e+00, 4.65541850468400e+00, 4.60463057885875e+00, 4.55048514141930e+00, 4.49330971877810e+00, 4.43348789427542e+00, 4.37144963369365e+00, 4.30766030423766e+00, 4.24260891056732e+00, 4.17679608994096e+00, 4.11072239205483e+00, 4.04487731982098e+00, 3.97972952928458e+00, 3.91571848656160e+00, 3.85324776577210e+00, 3.79268005494212e+00, 3.73433382787543e+00, 3.67848154912248e+00, 3.62534921365932e+00, 3.57511698597474e+00, 3.52792069390256e+00, 3.48385394612420e+00, 3.44297067205947e+00, 3.40528792158000e+00, 3.37078880313195e+00, 3.33942547753996e+00, 3.31112215789522e+00, 3.28577809202694e+00, 3.26327052282261e+00, 3.24345763352800e+00, 3.22618149090096e+00, 3.21127099959430e+00, 3.19854487727677e+00, 3.18781465260594e+00, 3.17888767808385e+00, 3.17157013793827e+00, 3.16567001844295e+00, 3.16099999556594e+00, 3.15738018361529e+00, 3.15464067973064e+00, 3.15262383364501e+00, 3.15118617091326e+00, 3.15019990128489e+00, 3.14955395222825e+00, 3.14915448052341e+00, 3.14892483166076e+00, 3.14880493649164e+00, 3.14875015590852e+00, 3.14872960590305e+00, 3.14872401581108e+00, 3.14872319072550e+00, 3.14872312852365e+00, 3.14872203020494e+00, 3.14871458893866e+00, 3.14868723412552e+00, 3.14861431676579e+00, 3.14845474159066e+00, 3.14814914632086e+00, 3.14761774574960e+00, 3.14675894749424e+00, 3.14544883338803e+00, 3.14354158463922e+00, 3.14087090882001e+00, 3.13725250114125e+00, 3.13248754020410e+00, 3.12636717892990e+00, 3.11867794496222e+00, 3.10920791299312e+00, 3.09775345702111e+00, 3.08412633767742e+00, 3.06816083374666e+00, 3.04972059373959e+00, 3.02870486857443e+00, 3.00505379472133e+00, 2.97875243114232e+00, 2.94983331273986e+00, 2.91837736423647e+00, 2.88451311468627e+00, 2.84841425494748e+00, 2.81029567805977e+00, 2.77040822572104e+00, 2.72903242525211e+00, 2.68647153629391e+00, 2.64304423460804e+00, 2.59907724489042e+00, 2.55489820102652e+00, 2.51082896725332e+00, 2.46717960336604e+00, 2.42424310620831e+00, 2.38229101134016e+00, 2.34156989469342e+00, 2.30229877497658e+00, 2.26466738397896e+00, 2.22883524411381e+00, 2.19493147098148e+00, 2.16305520385813e+00, 2.13327655904970e+00, 2.10563799983689e+00, 2.08015602164641e+00, 2.05682306104001e+00, 2.03560955070463e+00, 2.01646605827450e+00, 1.99932546294348e+00, 1.98410513900768e+00, 1.97070912854810e+00, 1.95903029556781e+00, 1.94895246052828e+00, 1.94035251720414e+00, 1.93310253322579e+00, 1.92707183201143e+00, 1.92212904764055e+00, 1.91814413641392e+00, 1.91499032032896e+00, 1.91254592947902e+00, 1.91069610345031e+00, 1.90933430702464e+00, 1.90836361360697e+00, 1.90769771123563e+00, 1.90726159095614e+00, 1.90699188557802e+00, 1.90683683792019e+00, 1.90675589085477e+00, 1.90671890587962e+00, 1.90670503160452e+00, 1.90670125747151e+00, 1.90670070041762e+00, 1.90670065819370e+00, 1.90669990960061e+00, 1.90669483777592e+00, 1.90667619326668e+00, 1.90662649427105e+00, 1.90651773136948e+00, 1.90630944558213e+00, 1.90594726045203e+00, 1.90536194173994e+00, 1.90446904986729e+00, 1.90316923977782e+00, 1.90134924952699e+00, 1.89888360172521e+00, 1.89563702018193e+00, 1.89146753731809e+00, 1.88623023631642e+00, 1.87978153650647e+00, 1.87198389294333e+00, 1.86271074419737e+00, 1.85185150939299e+00, 1.83931641028417e+00, 1.82504088036860e+00, 1.80898932390471e+00, 1.79115800530734e+00, 1.77157688432696e+00, 1.75031026341699e+00, 1.72745617765994e+00, 1.70314452980890e+00, 1.67753404748899e+00, 1.65080820994048e+00, 1.62317035164038e+00, 1.59483819441379e+00, 1.56603808452233e+00, 1.53699921497245e+00, 1.50794809639706e+00, 1.47910350489933e+00, 1.45067208650902e+00, 1.42284474082544e+00, 1.39579384686238e+00, 1.36967133758863e+00, 1.34460758073543e+00, 1.32071098527403e+00, 1.29806822712656e+00, 1.27674497421615e+00, 1.25678698869172e+00, 1.23822149101827e+00, 1.22105868410597e+00, 1.20529335319414e+00, 1.19090647647946e+00, 1.17786680056107e+00, 1.16613235223830e+00, 1.15565187309645e+00, 1.14636617512881e+00, 1.13820942418406e+00, 1.13111036335323e+00, 1.12499349074636e+00, 1.11978020578301e+00, 1.11538993553852e+00, 1.11174124830026e+00, 1.10875295579348e+00, 1.10634519907286e+00, 1.10444050641219e+00, 1.10296480524753e+00, 1.10184836491123e+00, 1.10102664305921e+00, 1.10044100676630e+00, 1.10003929952978e+00, 1.09977622798608e+00, 1.09961354691340e+00, 1.09952002777550e+00, 1.09947120419212e+00, 1.09944889671252e+00, 1.09944052847455e+00, 1.09943825211732e+00, 1.09943791613198e+00, 1.09943789050318e+00, 1.09943743399467e+00, 1.09943434108595e+00, 1.09942297127865e+00, 1.09939266394925e+00, 1.09932633914219e+00, 1.09919932702671e+00, 1.09897847637554e+00, 1.09862158873493e+00, 1.09807722047854e+00, 1.09728488911111e+00, 1.09617571228154e+00, 1.09467349720821e+00, 1.09269628394549e+00, 1.09015832774772e+00, 1.08697248379588e+00, 1.08305293244602e+00, 1.07831815635162e+00, 1.07269405444329e+00, 1.06611705455134e+00, 1.05853706950558e+00, 1.04992013386541e+00, 1.04025056252700e+00, 1.02953248980432e+00, 1.01779067823667e+00, 1.00507052869611e+00, 9.91437274038468e-01, 9.76974392858828e-01, 9.61781332372311e-01, 9.45970674496536e-01, 9.29664912111903e-01, 9.12993019986573e-01, 8.96087005752776e-01, 8.79078611521130e-01, 8.62096309044333e-01, 8.45262694964234e-01, 8.28692352392487e-01, 8.12490205530452e-01, 7.96750359082765e-01, 7.81555386501687e-01, 7.66976011899722e-01, 7.53071119808155e-01, 7.39888023882806e-01, 7.27462928593255e-01, 7.15821525060861e-01, 7.04979671773032e-01, 6.94944121393512e-01, 6.85713265174122e-01, 6.77277875787517e-01, 6.69621837310252e-01, 6.62722857414806e-01, 6.56553161568020e-01, 6.51080172278931e-01, 6.46267178325985e-01, 6.42073999575636e-01, 6.38457652626002e-01, 6.35373021217276e-01, 6.32773533290411e-01, 6.30611843909783e-01, 6.28840520181981e-01, 6.27412721016562e-01, 6.26282861332591e-01, 6.25407247383542e-01, 6.24744667520901e-01, 6.24256921196768e-01, 6.23909268523022e-01, 6.23670783395106e-01, 6.23514595098818e-01, 6.23418006388292e-01, 6.23362480092339e-01, 6.23333491122061e-01, 6.23320245995686e-01, 6.23315277319591e-01, 6.23313925721867e-01, 6.23313726228979e-01, 6.23313710904355e-01, 6.23313436529713e-01, 6.23311577602813e-01, 6.23304744015215e-01, 6.23286528366677e-01, 6.23246664814552e-01, 6.23170324979881e-01, 6.23037580744537e-01, 6.22823061306384e-01, 6.22495828899460e-01, 6.22019493836959e-01, 6.21352584615600e-01, 6.20449182321742e-01, 6.19259820154255e-01, 6.17732638338460e-01, 6.15814772123905e-01, 6.13453936364195e-01, 6.10600155207713e-01, 6.07207570926661e-01, 6.03236253448372e-01, 5.98653923455693e-01, 5.93437498634941e-01, 5.87574376037502e-01, 5.81063374185426e-01, 5.73915276217164e-01, 5.66152938793692e-01, 5.57810958552786e-01, 5.48934915887166e-01, 5.39580241809233e-01, 5.29810774992868e-01, 5.19697090782493e-01, 5.09314691031918e-01, 4.98742143144227e-01, 4.88059249633785e-01, 4.77345317593247e-01, 4.66677582590897e-01, 4.56129825662337e-01, 4.45771206781721e-01, 4.35665324612103e-01, 4.25869501051672e-01, 4.16434280304564e-01, 4.07403125803810e-01, 3.98812294036887e-01, 3.90690861852782e-01, 3.83060882858224e-01, 3.75937648773387e-01, 3.69330032885934e-01, 3.63240894809451e-01, 3.57667528418375e-01, 3.52602137890585e-01, 3.48032330025482e-01, 3.43941614202095e-01, 3.40309904288503e-01, 3.37114019320410e-01, 3.34328181681833e-01, 3.31924512737544e-01, 3.29873526336030e-01, 3.28144620332949e-01, 3.26706565349120e-01, 3.25527988500077e-01, 3.24577847991122e-01, 3.23825892471881e-01, 3.23243097118151e-01, 3.22802066783957e-01, 3.22477395454662e-01, 3.22245970797582e-01, 3.22087212962352e-01, 3.21983237965256e-01, 3.21918937963972e-01, 3.21881973379583e-01, 3.21862674979492e-01, 3.21853857482544e-01, 3.21850549751439e-01, 3.21849649969927e-01, 3.21849517164118e-01, 3.21849506884398e-01, 3.21849321821041e-01, 3.21848067990572e-01, 3.21843458792550e-01, 3.21831172481502e-01, 3.21804284791995e-01, 3.21752793955781e-01, 3.21663258175228e-01, 3.21518563924400e-01, 3.21297841427040e-01, 3.20976541129802e-01, 3.20526681709765e-01, 3.19917275881330e-01, 3.19114934787237e-01, 3.18084644962276e-01, 3.16790703791717e-01, 3.15197790302989e-01, 3.13272138514570e-01, 3.10982771138448e-01, 3.08302743094387e-01, 3.05210338046770e-01, 3.01690157991807e-01, 2.97734046599921e-01, 2.93341792029182e-01, 2.88521564305064e-01, 2.83290055648500e-01, 2.77672308392361e-01, 2.71701233044347e-01, 2.65416837104070e-01, 2.58865201873988e-01, 2.52097258335127e-01, 2.45167423120563e-01, 2.38132161061863e-01, 2.31048541500975e-01, 2.23972851783621e-01, 2.16959323683554e-01, 2.10059017835157e-01, 2.03318898635125e-01, 1.96781118628905e-01, 1.90482518199836e-01, 1.84454334367189e-01, 1.78722102423762e-01, 1.73305726505170e-01, 1.68219690222642e-01, 1.63473376180988e-01, 1.59071463294218e-01, 1.55014372880885e-01, 1.51298738064519e-01, 1.47917875491721e-01, 1.44862243312375e-01, 1.42119874314339e-01, 1.39676777715536e-01, 1.37517307131979e-01, 1.35624495488209e-01, 1.33980360024410e-01, 1.32566182059286e-01, 1.31362766818377e-01, 1.30350688510354e-01, 1.29510525036515e-01, 1.28823085390493e-01, 1.28269631107557e-01, 1.27832091232584e-01, 1.27493268381905e-01, 1.27237031757524e-01, 1.27048491603276e-01, 1.26914148708757e-01, 1.26822012263723e-01, 1.26761679686155e-01, 1.26724372975687e-01, 1.26702927607077e-01, 1.26691731854962e-01, 1.26686616574744e-01, 1.26684697681029e-01, 1.26684175697719e-01, 1.26684098654167e-01, 1.26684092642336e-01, 1.26683983788077e-01, 1.26683246285786e-01, 1.26680535169157e-01, 1.26673308479567e-01, 1.26657493817333e-01, 1.26627209802329e-01, 1.26574554893665e-01, 1.26489475391814e-01, 1.26359723651539e-01, 1.26170916386762e-01, 1.25906701216884e-01, 1.25549037040386e-01, 1.25078590213036e-01, 1.24475243722159e-01, 1.23718710591362e-01, 1.22789235806664e-01, 1.21668363519085e-01, 1.20339738751312e-01, 1.18789906093767e-01, 1.17009062798758e-01, 1.14991721145245e-01, 1.12737235695583e-01, 1.10250155571883e-01, 1.07540370230751e-01, 1.04623029037789e-01, 1.01518229391012e-01, 9.82504839591485e-02, 9.48479932551365e-02, 9.13417636404787e-02, 8.77646214826271e-02, 8.41501804538112e-02, 8.05318202845910e-02, 7.69417316869507e-02, 7.34100742273824e-02, 6.99642826987201e-02, 6.66285443396646e-02, 6.34234555289268e-02, 6.03658537270064e-02, 5.74688096136740e-02, 5.47417563505448e-02, 5.21907279960701e-02, 4.98186771855202e-02, 4.76258428219083e-02, 4.56101411213865e-02, 4.37675573091312e-02, 4.20925199916553e-02, 4.05782452088843e-02, 3.92170419373841e-02, 3.80005750327896e-02, 3.69200850699727e-02, 3.59665672060601e-02, 3.51309130946618e-02, 3.44040211017003e-02, 3.37768807098726e-02, 3.32406371389453e-02, 3.27866419401843e-02, 3.24064947277378e-02, 3.20920803667448e-02, 3.18356049195856e-02, 3.16296325272347e-02, 3.14671242399100e-02, 3.13414786790887e-02, 3.12465733824752e-02, 3.11768048249021e-02, 3.11271244797595e-02, 3.10930679382011e-02, 3.10707740569711e-02, 3.10569913640393e-02, 3.10490694802823e-02, 3.10449340639991e-02, 3.10430446770980e-02, 3.10423359220806e-02, 3.10421431253545e-02, 3.10421146690203e-02, 3.10421124296383e-02, 3.10420716397388e-02, 3.10417952831870e-02, 3.10407793821668e-02, 3.10380714772549e-02, 3.10321458732616e-02, 3.10207998538559e-02, 3.10010759802921e-02, 3.09692156102344e-02, 3.09206487406470e-02, 3.08500250693697e-02, 3.07512906198468e-02, 3.06178132112009e-02, 3.04425583277848e-02, 3.02183144462944e-02, 2.99379636110473e-02, 2.95947891284527e-02, 2.91828079525587e-02, 2.86971110822520e-02, 2.81341916349547e-02, 2.74922378212352e-02, 2.67713674050398e-02, 2.59737818530539e-02, 2.51038224556126e-02, 2.41679171235347e-02, 2.31744148441990e-02, 2.21333140973528e-02, 2.10559008221693e-02, 1.99543196196075e-02, 1.88411076812258e-02, 1.77287236513735e-02, 1.66291029219323e-02, 1.55532669520184e-02, 1.45110078215183e-02, 1.35106613721021e-02, 1.25589739850867e-02, 1.16610601624458e-02, 1.08204413215560e-02, 1.00391511855946e-02, 9.31789030759234e-03, 8.65621174897768e-03, 8.05272139721325e-03, 7.50527909673337e-03, 7.01118985023186e-03, 6.56737729314472e-03, 6.17053432969072e-03, 5.81724831504005e-03, 5.50410045547522e-03, 5.22774097081680e-03, 4.98494279402451e-03, 4.77263712401310e-03, 4.58793417488579e-03, 4.42813224722016e-03, 4.29071797297098e-03, 4.17336031834026e-03, 4.07390063874402e-03, 3.99034074206384e-03, 3.92083053854719e-03, 3.86365647151621e-03, 3.81723156112201e-03, 3.78008756849368e-03, 3.75086949260606e-03, 3.72833234723657e-03, 3.71133992838679e-03, 3.69886508212858e-03, 3.68999083112698e-03, 3.68391162443515e-03, 3.67993394791455e-03, 3.67747557325515e-03, 3.67606282642896e-03, 3.67532541188353e-03, 3.67498851912273e-03, 3.67486214517498e-03, 3.67482776898242e-03, 3.67482269514596e-03, 3.67482229233929e-03, 3.67481491006316e-03, 3.67476489444130e-03, 3.67458103649909e-03, 3.67409097520266e-03, 3.67301867321146e-03, 3.67096579594772e-03, 3.66739804400140e-03, 3.66163756161342e-03, 3.65286260966968e-03, 3.64011570835743e-03, 3.62232137069185e-03, 3.59831430757396e-03, 3.56687853604853e-03, 3.52679713473236e-03, 3.47691146328799e-03, 3.41618754478645e-03, 3.34378609953490e-03, 3.25913156920166e-03, 3.16197457162559e-03, 3.05244178331776e-03, 2.93106744680927e-03, 2.79880167138202e-03, 2.65699246790518e-03, 2.50734091320481e-03, 2.35183168602759e-03, 2.19264400283947e-03, 2.03205018337371e-03, 1.87231026978196e-03, 1.71557115800880e-03, 1.56377776643524e-03, 1.41860225458856e-03, 1.28139551181761e-03, 1.15316307878288e-03, 1.03456519439504e-03, 9.25937883380357e-04, 8.27329572246072e-04, 7.38546622720548e-04, 6.59201935441445e-04, 5.88762864978776e-04, 5.26596698847470e-04, 4.72012692419810e-04, 4.24299131438488e-04, 3.82753292598181e-04, 3.46702700674939e-04, 3.15517825242750e-04, 2.88618232829663e-04, 2.65474968827561e-04, 2.45611266414869e-04, 2.28602314678028e-04, 2.14073823741501e-04, 2.01699001173949e-04, 1.91194041626375e-04, 1.82312753047142e-04, 1.74841112958291e-04, 1.68592345445973e-04, 1.63402758445841e-04, 1.59128309135781e-04, 1.55641759548075e-04, 1.52830299292361e-04, 1.50593566395685e-04, 1.48842036466590e-04, 1.47495762391259e-04, 1.46483437404754e-04, 1.45741736808702e-04, 1.45214877584010e-04, 1.44854324979349e-04, 1.44618573308719e-04, 1.44472932234407e-04, 1.44389259077116e-04, 1.44345590385342e-04, 1.44325641507237e-04, 1.44318158582735e-04, 1.44316123105170e-04, 1.44315822680596e-04, 1.44315798614514e-04, 1.44315355171898e-04, 1.44312350762732e-04, 1.44301306704619e-04, 1.44271870844648e-04, 1.44207469162438e-04, 1.44084201181727e-04, 1.43870052802697e-04, 1.43524510995974e-04, 1.42998675229933e-04, 1.42235967189208e-04, 1.41173538786177e-04, 1.39744461392866e-04, 1.37880740297192e-04, 1.35517128867946e-04, 1.32595611759712e-04, 1.29070285076693e-04, 1.24912203233407e-04, 1.20113629257901e-04, 1.14691086061214e-04, 1.08686730612817e-04, 1.02167875219550e-04, 9.52248705914371e-05, 8.79678342684542e-05, 8.05226227100775e-05, 7.30259289269952e-05, 6.56187423970778e-05, 5.84373373442304e-05, 5.16021339365125e-05, 4.52069873446703e-05, 3.93131610323394e-05, 3.39514638767469e-05, 2.91320792698002e-05, 2.48562905511562e-05, 2.11217896790572e-05, 1.79172815677475e-05, 1.52110591855603e-05, 1.29450184118696e-05, 1.10439130267537e-05, 9.43819514276356e-06, 8.08500863058725e-06, 6.97109567227017e-06, 6.09507471693860e-06, 5.44037208527470e-06, 4.96017395335002e-06, 4.59051414480882e-06, 4.27675389857995e-06, 3.98945027058062e-06, 3.72715602736662e-06, 3.50738236676307e-06, 3.34542269087024e-06, 3.23877281971363e-06, 3.17390737336609e-06, 3.13939276708197e-06, 3.12660631607287e-06, 3.12559514270561e-06, 3.12667396684674e-06, 3.12448553785316e-06, 3.11863705348969e-06, 3.11123276475200e-06, 3.10436687556148e-06, 3.09909155049581e-06, 3.09553385591244e-06, 3.09336019199691e-06, 3.09213478999482e-06, 3.09148970245587e-06, 3.09116916241692e-06, 3.09101680784352e-06, 3.09094640726926e-06, 3.09091441152987e-06, 3.09090020067515e-06, 3.09089428185416e-06, 3.09089215022595e-06, 3.09089158179177e-06, 3.09089149652664e-06, 3.09089141836694e-06, 3.09088897409993e-06, 3.09087239713790e-06, 3.09081153915258e-06, 3.09064974146622e-06, 3.09029781675463e-06, 3.08963208794921e-06, 3.08849974572922e-06, 3.08673572391172e-06, 3.08419452466069e-06, 3.08079635402692e-06, 3.07657694520458e-06, 3.07171180935484e-06, 3.06645409153816e-06, 3.06088214330202e-06, 3.05433259484289e-06, 3.04452354527029e-06, 3.02682596498016e-06, 2.99482056931311e-06, 2.94337552730894e-06, 2.87384750663477e-06, 2.79768948985293e-06, 2.73298656594534e-06, 2.69438496047769e-06, 2.68725030849737e-06, 2.70979251837389e-06, 2.74919902665260e-06, 2.77453398211946e-06, 2.75123048965042e-06, 2.67148001142914e-06, 2.56558470823620e-06, 2.48011649972568e-06, 2.44843441521425e-06, 2.47526553209809e-06, 2.51939391660017e-06, 2.49636989479995e-06, 2.32554620538394e-06, 1.97939659329904e-06, 1.51053902897047e-06, 1.06885438810878e-06, 8.28003521746723e-07, 6.81904026578195e-07, 5.08392204778829e-07, 4.23877466459999e-07, -4.97379915032070e-14, 3.66364318438173e-07, 7.49681085210341e-07, 9.84086909738835e-07, 1.03630869574545e-06, 9.18551229744935e-07, 6.80102495209667e-07, 3.87008299185254e-07, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -4.97379915032070e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -4.97379915032070e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14],
[ 5.50000000000000e+01, 5.49999999470308e+01, 5.49999983091191e+01, 5.49999872120078e+01, 5.49999464175804e+01, 5.49998376725238e+01, 5.49995996760156e+01, 5.49991438435686e+01, 5.49983510198741e+01, 5.49970692643697e+01, 5.49951128003073e+01, 5.49922621826103e+01, 5.49882657032335e+01, 5.49828420166588e+01, 5.49756839341381e+01, 5.49664633047251e+01, 5.49548368750824e+01, 5.49404529990400e+01, 5.49229590518014e+01, 5.49020093917430e+01, 5.48772737034960e+01, 5.48484455476709e+01, 5.48152509333948e+01, 5.47774567185406e+01, 5.47348786288689e+01, 5.46873886723832e+01, 5.46349217116267e+01, 5.45774809482798e+01, 5.45151420758317e+01, 5.44480558716962e+01, 5.43764490331795e+01, 5.43006231134140e+01, 5.42209514823491e+01, 5.41378743200314e+01, 5.40518917383266e+01, 5.39635552151653e+01, 5.38734576043530e+01, 5.37822220470043e+01, 5.36904901527693e+01, 5.35989098378347e+01, 5.35081232025542e+01, 5.34187548071862e+01, 5.33314006640249e+01, 5.32466182134759e+01, 5.31649174956831e+01, 5.30867536728727e+01, 5.30125210042692e+01, 5.29425483276553e+01, 5.28770960605508e+01, 5.28163546997188e+01, 5.27604447696661e+01, 5.27094181479376e+01, 5.26632606761052e+01, 5.26218959492541e+01, 5.25851901625155e+01, 5.25529578801322e+01, 5.25249685803235e+01, 5.25009538178325e+01, 5.24806148357547e+01, 5.24636304495594e+01, 5.24496650197710e+01, 5.24383763263127e+01, 5.24294231577607e+01, 5.24224724333571e+01, 5.24172056850922e+01, 5.24133247417605e+01, 5.24105564766392e+01, 5.24086565050600e+01, 5.24074117470758e+01, 5.24066418028303e+01, 5.24061991230654e+01, 5.24059679932281e+01, 5.24058623855607e+01, 5.24058227680684e+01, 5.24058119911112e+01, 5.24058104004535e+01, 5.24058101917927e+01, 5.24058053302475e+01, 5.24057723925446e+01, 5.24056513097788e+01, 5.24053285453860e+01, 5.24046221722724e+01, 5.24032693405973e+01, 5.24009166215268e+01, 5.23971136440074e+01, 5.23913103634095e+01, 5.23828582130670e+01, 5.23710152924477e+01, 5.23549556391606e+01, 5.23337825167903e+01, 5.23065455275920e+01, 5.22722612297701e+01, 5.22299368051287e+01, 5.21785961864330e+01, 5.21173079176757e+01, 5.20452138888700e+01, 5.19615579668527e+01, 5.18657134455904e+01, 5.17572081787146e+01, 5.16357462521142e+01, 5.15012251245087e+01, 5.13537473235712e+01, 5.11936260380984e+01, 5.10213842805252e+01, 5.08377476781244e+01, 5.06436313396611e+01, 5.04401215843860e+01, 5.02284535644776e+01, 5.00099859300398e+01, 4.97861736726365e+01, 4.95585401610140e+01, 4.93286491935271e+01, 4.90980776866849e+01, 4.88683894434604e+01, 4.86411103268385e+01, 4.84177051092683e+01, 4.81995562622614e+01, 4.79879449646342e+01, 4.77840346133507e+01, 4.75888570965169e+01, 4.74033020268906e+01, 4.72281090436432e+01, 4.70638631867500e+01, 4.69109932511260e+01, 4.67697729510391e+01, 4.66403246761315e+01, 4.65226255973671e+01, 4.64165158776299e+01, 4.63217087485171e+01, 4.62378022236643e+01, 4.61642922235885e+01, 4.61005868842339e+01, 4.60460218104217e+01, 4.59998760173992e+01, 4.59613882810599e+01, 4.59297735931835e+01, 4.59042393955475e+01, 4.58840012492373e+01, 4.58682975859793e+01, 4.58564031894493e+01, 4.58476410682840e+01, 4.58413924101918e+01, 4.58371043484469e+01, 4.58342953275062e+01, 4.58325579218237e+01, 4.58315590385898e+01, 4.58310375177431e+01, 4.58307992274147e+01, 4.58307098359768e+01, 4.58306855192902e+01, 4.58306819301964e+01, 4.58306816133478e+01, 4.58306754048542e+01, 4.58306333413870e+01, 4.58304787114907e+01, 4.58300665238712e+01, 4.58291644558258e+01, 4.58274368664414e+01, 4.58244325077716e+01, 4.58195764875443e+01, 4.58121669447454e+01, 4.58013767931277e+01, 4.57862607649610e+01, 4.57657678457476e+01, 4.57387590284008e+01, 4.57040301323875e+01, 4.56603392319137e+01, 4.56064380229493e+01, 4.55411062410011e+01, 4.54631880329772e+01, 4.53716290034934e+01, 4.52655125170712e+01, 4.51440937623596e+01, 4.50068300911748e+01, 4.48534062485419e+01, 4.46837533176631e+01, 4.44980605120594e+01, 4.42967793365235e+01, 4.40806200713898e+01, 4.38505409578000e+01, 4.36077308163237e+01, 4.33535860696142e+01, 4.30896832422798e+01, 4.28177479959892e+01, 4.25396216745913e+01, 4.22572262427945e+01, 4.19725284476732e+01, 4.16875040275531e+01, 4.14041028173713e+01, 4.11242156145744e+01, 4.08496436377050e+01, 4.05820713109688e+01, 4.03230429450661e+01, 4.00739436790418e+01, 3.98359848307281e+01, 3.96101936045020e+01, 3.93974069463774e+01, 3.91982692284723e+01, 3.90132333875438e+01, 3.88425651276740e+01, 3.86863498132584e+01, 3.85445017122184e+01, 3.84167752892341e+01, 3.83027782857916e+01, 3.82019863519129e+01, 3.81137590102794e+01, 3.80373567360046e+01, 3.79719589252333e+01, 3.79166825049892e+01, 3.78706009081614e+01, 3.78327631047898e+01, 3.78022123479030e+01, 3.77780042633359e+01, 3.77592238924057e+01, 3.77450012879216e+01, 3.77345252709298e+01, 3.77270549800653e+01, 3.77219288883858e+01, 3.77185710236581e+01, 3.77164942054195e+01, 3.77153002025423e+01, 3.77146768142319e+01, 3.77143919804058e+01, 3.77142851289902e+01, 3.77142560627753e+01, 3.77142517726614e+01, 3.77142514285412e+01, 3.77142450777780e+01, 3.77142020504159e+01, 3.77140438771803e+01, 3.77136222446308e+01, 3.77126995081684e+01, 3.77109323423528e+01, 3.77078591789111e+01, 3.77028920028676e+01, 3.76953129845913e+01, 3.76842763183815e+01, 3.76688155138082e+01, 3.76478562416111e+01, 3.76202346694896e+01, 3.75847210343442e+01, 3.75400479888856e+01, 3.74849430379629e+01, 3.74181641533359e+01, 3.73385374386670e+01, 3.72449955262352e+01, 3.71366152422092e+01, 3.70126529972925e+01, 3.68725763608826e+01, 3.67160903714022e+01, 3.65431573274862e+01, 3.63540090890346e+01, 3.61491512777774e+01, 3.59293591774739e+01, 3.56956655593172e+01, 3.54493410593859e+01, 3.51918680745156e+01, 3.49249093913231e+01, 3.46502729047567e+01, 3.43698738188421e+01, 3.40856956705020e+01, 3.37997514051977e+01, 3.35140455906194e+01, 3.32305387052674e+01, 3.29511142940493e+01, 3.26775496420366e+01, 3.24114904716736e+01, 3.21544300095330e+01, 3.19076925948570e+01, 3.16724218225069e+01, 3.14495730444815e+01, 3.12399099158017e+01, 3.10440045765753e+01, 3.08622410177692e+01, 3.06948211792438e+01, 3.05417733635466e+01, 3.04029626032949e+01, 3.02781026799127e+01, 3.01667695461637e+01, 3.00684159473557e+01, 2.99823870627460e+01, 2.99079369987487e+01, 2.98442459600833e+01, 2.97904379062987e+01, 2.97455984723049e+01, 2.97087928964288e+01, 2.96790836623446e+01, 2.96555475265657e+01, 2.96372915756372e+01, 2.96234679410735e+01, 2.96132867991368e+01, 2.96060272994606e+01, 2.96010461026709e+01, 2.95977832624353e+01, 2.95957652600613e+01, 2.95946050867202e+01, 2.95939993651921e+01, 2.95937226045415e+01, 2.95936187818269e+01, 2.95935905395086e+01, 2.95935863710008e+01, 2.95935860494853e+01, 2.95935802761000e+01, 2.95935411605633e+01, 2.95933973676841e+01, 2.95930140684423e+01, 2.95921752263125e+01, 2.95905687390134e+01, 2.95877750307630e+01, 2.95832596196468e+01, 2.95763701028397e+01, 2.95663379070736e+01, 2.95522850399358e+01, 2.95332359462283e+01, 2.95081344203937e+01, 2.94758653503413e+01, 2.94352808720656e+01, 2.93852303038945e+01, 2.93245930136850e+01, 2.92523131653458e+01, 2.91674351094934e+01, 2.90691380455666e+01, 2.89567685079582e+01, 2.88298692327498e+01, 2.86882030552309e+01, 2.85317706743781e+01, 2.83608213918100e+01, 2.81758562717044e+01, 2.79776235474894e+01, 2.77671064872653e+01, 2.75455042882580e+01, 2.73142068717761e+01, 2.70747646753086e+01, 2.68288546819835e+01, 2.65782439958606e+01, 2.63247522775782e+01, 2.60702143122371e+01, 2.58164438990158e+01, 2.55652001327427e+01, 2.53181569908519e+01, 2.50768769453549e+01, 2.48427890947458e+01, 2.46171720686246e+01, 2.44011417177106e+01, 2.41956433853900e+01, 2.40014483828921e+01, 2.38191541710056e+01, 2.36491876909271e+01, 2.34918112811403e+01, 2.33471306556663e+01, 2.32151044875812e+01, 2.30955552255831e+01, 2.29881808572978e+01, 2.28925674104463e+01, 2.28082020448316e+01, 2.27344866304806e+01, 2.26707517291053e+01, 2.26162708984225e+01, 2.25702752243974e+01, 2.25319679588410e+01, 2.25005391033269e+01, 2.24751797398296e+01, 2.24550958687050e+01, 2.24395214804017e+01, 2.24277305630446e+01, 2.24190477375307e+01, 2.24128572178155e+01, 2.24086098182162e+01, 2.24058277719256e+01, 2.24041071841216e+01, 2.24031180162564e+01, 2.24026015812731e+01, 2.24023656178153e+01, 2.24022770997680e+01, 2.24022530207121e+01, 2.24022494666939e+01, 2.24022491975944e+01, 2.24022444305231e+01, 2.24022121329192e+01, 2.24020934035900e+01, 2.24017769153989e+01, 2.24010842918246e+01, 2.23997578466638e+01, 2.23974511844630e+01, 2.23937231046219e+01, 2.23880351883104e+01, 2.23797533718908e+01, 2.23681537201020e+01, 2.23524325033179e+01, 2.23317205529136e+01, 2.23051017160355e+01, 2.22716350573590e+01, 2.22303802656734e+01, 2.21804255261440e+01, 2.21209169275770e+01, 2.20510883039198e+01, 2.19702902786401e+01, 2.18780172078666e+01, 2.17739307195868e+01, 2.16578786336326e+01, 2.15299082255874e+01, 2.13902730631798e+01, 2.12394329822012e+01, 2.10780471564839e+01, 2.09069606205077e+01, 2.07271849858516e+01, 2.05398744153434e+01, 2.03462981475366e+01, 2.01478109757571e+01, 1.99458230721427e+01, 1.97417704168356e+01, 1.95370868710091e+01, 1.93331786567368e+01, 1.91314017187449e+01, 1.89330421815624e+01, 1.87392999090674e+01, 1.85512750360012e+01, 1.83699572715547e+01, 1.81962177601989e+01, 1.80308033041525e+01, 1.78743327840390e+01, 1.77272956425446e+01, 1.75900523107271e+01, 1.74628364562397e+01, 1.73457589211762e+01, 1.72388132014218e+01, 1.71418823060016e+01, 1.70547468283080e+01, 1.69770940623646e+01, 1.69085280046724e+01, 1.68485800920612e+01, 1.67967205342474e+01, 1.67523701029781e+01, 1.67149122356323e+01, 1.66837052994781e+01, 1.66580948446091e+01, 1.66374256513149e+01, 1.66210533546074e+01, 1.66083554085352e+01, 1.65987411394627e+01, 1.65916606339196e+01, 1.65866122153778e+01, 1.65831482867963e+01, 1.65808793521716e+01, 1.65794760795846e+01, 1.65786693281335e+01, 1.65782481285430e+01, 1.65780556783712e+01, 1.65779834835706e+01, 1.65779638448252e+01, 1.65779609461872e+01, 1.65779607285856e+01, 1.65779568985494e+01, 1.65779309494909e+01, 1.65778355580069e+01, 1.65775812793341e+01, 1.65770247966800e+01, 1.65759590658337e+01, 1.65741057509848e+01, 1.65711103016001e+01, 1.65665399666404e+01, 1.65598849798560e+01, 1.65505630762891e+01, 1.65379274132827e+01, 1.65212778676320e+01, 1.64998755636155e+01, 1.64729603559751e+01, 1.64397708512952e+01, 1.63995664071027e+01, 1.63516504092351e+01, 1.62953940053918e+01, 1.62302593778694e+01, 1.61558215823750e+01, 1.60717879716017e+01, 1.59780142678239e+01, 1.58745164501673e+01, 1.57614777777295e+01, 1.56392504746566e+01, 1.55083518507982e+01, 1.53694549128797e+01, 1.52233738242879e+01, 1.50710448787352e+01, 1.49135039377387e+01, 1.47518615084318e+01, 1.45872767668514e+01, 1.44209318293808e+01, 1.42540074291829e+01, 1.40876608845798e+01, 1.39230069034289e+01, 1.37611014211245e+01, 1.36029283858563e+01, 1.34493892248745e+01, 1.33012946572045e+01, 1.31593585387611e+01, 1.30241934962291e+01, 1.28963081873919e+01, 1.27761060901671e+01, 1.26638857581702e+01, 1.25598424880207e+01, 1.24640713314769e+01, 1.23765713646908e+01, 1.22972511065533e+01, 1.22259349637767e+01, 1.21623705737659e+01, 1.21062369162500e+01, 1.20571530681783e+01, 1.20146874801457e+01, 1.19783676536276e+01, 1.19476900946533e+01, 1.19221304105302e+01, 1.19011534024551e+01, 1.18842229899860e+01, 1.18708117858661e+01, 1.18604101244822e+01, 1.18525343372622e+01, 1.18467340661988e+01, 1.18425984144639e+01, 1.18397607519179e+01, 1.18379020233977e+01, 1.18367524481477e+01, 1.18360915478442e+01, 1.18357464958048e+01, 1.18355888381027e+01, 1.18355296951626e+01, 1.18355136068401e+01, 1.18355112322371e+01, 1.18355110545170e+01, 1.18355079336624e+01, 1.18354867894125e+01, 1.18354090609030e+01, 1.18352018651551e+01, 1.18347484219248e+01, 1.18338800223459e+01, 1.18323698633584e+01, 1.18299290320049e+01, 1.18262048796598e+01, 1.18207819749814e+01, 1.18131857642515e+01, 1.18028889971596e+01, 1.17893208937833e+01, 1.17718789343082e+01, 1.17499430481716e+01, 1.17228918669675e+01, 1.16901205908417e+01, 1.16510599084047e+01, 1.16051953140722e+01, 1.15520860934461e+01, 1.14913832056465e+01, 1.14228452883227e+01, 1.13463520504689e+01, 1.12619144005754e+01, 1.11696807797622e+01, 1.10699393249157e+01, 1.09631156668025e+01, 1.08497663629549e+01, 1.07305681647449e+01, 1.06063035124827e+01, 1.04778428313230e+01, 1.03461243535146e+01, 1.02121323079743e+01, 1.00768743857732e+01, 9.94135940167500e+00, 9.80657602359543e+00, 9.67347333619482e+00, 9.54294385109117e+00, 9.41580938974813e+00, 9.29281006514961e+00, 9.17459639487309e+00, 9.06172440870900e+00, 8.95465348132558e+00, 8.85374653134036e+00, 8.75927218311325e+00, 8.67140848210921e+00, 8.59024778082709e+00, 8.51580246043606e+00, 8.44801121371375e+00, 8.38674567904016e+00, 8.33181727620638e+00, 8.28298414774237e+00, 8.23995815131327e+00, 8.20241187801450e+00, 8.16998568791265e+00, 8.14229475867063e+00, 8.11893613700595e+00, 8.09949576802482e+00, 8.08355545651251e+00, 8.07069968968381e+00, 8.06052222552815e+00, 8.05263232757366e+00, 8.04666050836222e+00, 8.04226363256188e+00, 8.03912922829390e+00, 8.03697886305074e+00, 8.03557045880854e+00, 8.03469944896283e+00, 8.03419871604650e+00, 8.03393729160238e+00, 8.03381784536015e+00, 8.03377303707155e+00, 8.03376084814200e+00, 8.03375904908194e+00, 8.03375891438624e+00, 8.03375654838879e+00, 8.03374051841509e+00, 8.03368159064309e+00, 8.03352451166069e+00, 8.03318075232283e+00, 8.03252242805822e+00, 8.03137764668563e+00, 8.02952751003070e+00, 8.02670496648125e+00, 8.02259567639328e+00, 8.01684100873544e+00, 8.00904323319528e+00, 7.99877290523552e+00, 7.98557836135053e+00, 7.96899714866889e+00, 7.94856910976142e+00, 7.92385073505273e+00, 7.89443028892991e+00, 7.85994312079026e+00, 7.82008649938904e+00, 7.77463326857043e+00, 7.72344362414434e+00, 7.66647436187714e+00, 7.60378504763801e+00, 7.53554070972240e+00, 7.46201084148365e+00, 7.38356471537811e+00, 7.30066322883891e+00, 7.21384770736922e+00, 7.12372626081618e+00, 7.03095840843463e+00, 6.93623874668488e+00, 6.84028042812838e+00, 6.74379915575476e+00, 6.64749828696451e+00, 6.55205550209733e+00, 6.45811134208171e+00, 6.36625977507449e+00, 6.27704082545024e+00, 6.19093519760876e+00, 6.10836075410771e+00, 6.02967066086094e+00, 5.95515298737995e+00, 5.88503154221875e+00, 5.81946772821166e+00, 5.75856321505223e+00, 5.70236324561727e+00, 5.65086041536009e+00, 5.60399878964447e+00, 5.56167825067695e+00, 5.52375899219776e+00, 5.49006610467076e+00, 5.46039421474385e+00, 5.43451215881825e+00, 5.41216768065125e+00, 5.39309214652006e+00, 5.37700526868149e+00, 5.36361981932727e+00, 5.35264630413549e+00, 5.34379754844925e+00, 5.33679313195918e+00, 5.33136359155833e+00, 5.32725429877181e+00, 5.32422890963724e+00, 5.32207228254322e+00, 5.32059276422432e+00, 5.31962375613563e+00, 5.31902449240565e+00, 5.31867998546871e+00, 5.31850012475959e+00, 5.31841794559179e+00, 5.31838711745814e+00, 5.31837873146863e+00, 5.31837749371452e+00, 5.31837740079797e+00, 5.31837576538899e+00, 5.31836468525869e+00, 5.31832395361886e+00, 5.31821537872343e+00, 5.31797776878106e+00, 5.31752273096038e+00, 5.31673145907261e+00, 5.31545266799453e+00, 5.31350181337940e+00, 5.31066171140302e+00, 5.30668464351602e+00, 5.30129599364309e+00, 5.29419941851462e+00, 5.28508349511051e+00, 5.27362972299832e+00, 5.25952168565028e+00, 5.24245509736047e+00, 5.22214838659496e+00, 5.19835339932443e+00, 5.17086575469700e+00, 5.13953435775783e+00, 5.10426957605517e+00, 5.06504962288072e+00, 5.02192476046724e+00, 4.97501903904535e+00, 4.92452941609507e+00, 4.87072224533120e+00, 4.81392727587124e+00, 4.75452944672134e+00, 4.69295888847958e+00, 4.62967964245724e+00, 4.56517766869760e+00, 4.49994873282911e+00, 4.43448673515945e+00, 4.36927297620522e+00, 4.30476674822810e+00, 4.24139751424845e+00, 4.17955879958956e+00, 4.11960379221041e+00, 4.06184254069980e+00, 4.00654056198960e+00, 3.95391862784025e+00, 3.90415348754105e+00, 3.85737929751961e+00, 3.81368955820918e+00, 3.77313939625495e+00, 3.73574806920419e+00, 3.70150160567059e+00, 3.67035552415749e+00, 3.64223759747332e+00, 3.61705064717813e+00, 3.59467536443169e+00, 3.57497316072361e+00, 3.55778905490735e+00, 3.54295460224914e+00, 3.53029086726004e+00, 3.51961143532876e+00, 3.51072544911653e+00, 3.50344064495764e+00, 3.49756635294550e+00, 3.49291641293584e+00, 3.48931194842761e+00, 3.48658393226872e+00, 3.48457547337079e+00, 3.48314375290498e+00, 3.48216154230381e+00, 3.48151824395181e+00, 3.48112040845280e+00, 3.48089169915614e+00, 3.48077229420901e+00, 3.48071773754019e+00, 3.48069727152101e+00, 3.48069170427425e+00, 3.48069088256054e+00, 3.48069082061290e+00, 3.48068972678265e+00, 3.48068231592590e+00, 3.48065507289417e+00, 3.48058245344691e+00, 3.48042352996316e+00, 3.48011918167684e+00, 3.47958994623012e+00, 3.47873463829077e+00, 3.47742982790535e+00, 3.47553025526244e+00, 3.47287023844344e+00, 3.46926610529199e+00, 3.46451964869596e+00, 3.45842256583617e+00, 3.45076179664716e+00, 3.44132562632906e+00, 3.42991036400262e+00, 3.41632735855783e+00, 3.40041006847732e+00, 3.38202087050941e+00, 3.36105727787802e+00, 3.33745724647302e+00, 3.31120327937530e+00, 3.28232509562134e+00, 3.25090070478645e+00, 3.21705581846289e+00, 3.18096162475729e+00, 3.14283104351664e+00, 3.10291365978400e+00, 3.06148959458992e+00, 3.01886261195634e+00, 2.97535277817797e+00, 2.93128898576800e+00, 2.88700163325505e+00, 2.84281571735648e+00, 2.79904454998271e+00, 2.75598426271701e+00, 2.71390920907999e+00, 2.67306832290635e+00, 2.63368244218525e+00, 2.59594256419008e+00, 2.56000896174019e+00, 2.52601106353949e+00, 2.49404798450766e+00, 2.46418958478693e+00, 2.43647793777248e+00, 2.41092909651537e+00, 2.38753506219921e+00, 2.36626587596195e+00, 2.34707177407082e+00, 2.32988536458240e+00, 2.31462379971857e+00, 2.30119093126396e+00, 2.28947944574296e+00, 2.27937298172189e+00, 2.27074823337543e+00, 2.26347704277874e+00, 2.25742847877405e+00, 2.25247089341744e+00, 2.24847393877669e+00, 2.24531051814334e+00, 2.24285863750379e+00, 2.24100311630816e+00, 2.23963711201503e+00, 2.23866341122869e+00, 2.23799544190211e+00, 2.23755796618709e+00, 2.23728742189601e+00, 2.23713189172809e+00, 2.23705069268546e+00, 2.23701359256533e+00, 2.23699967509302e+00, 2.23699588920911e+00, 2.23699533042080e+00, 2.23699528806541e+00, 2.23699453714153e+00, 2.23698944952532e+00, 2.23697074696329e+00, 2.23692089320929e+00, 2.23681179156705e+00, 2.23660285683685e+00, 2.23623954251877e+00, 2.23565239693097e+00, 2.23475671335685e+00, 2.23345282876313e+00, 2.23162711215130e+00, 2.22915366629142e+00, 2.22589674465740e+00, 2.22171385843022e+00, 2.21645951663842e+00, 2.20998950681843e+00, 2.20216558583460e+00, 2.19286041341197e+00, 2.18196252794421e+00, 2.16938113910844e+00, 2.15505049859507e+00, 2.13893361212874e+00, 2.12102507507620e+00, 2.10135285085344e+00, 2.07997886466161e+00, 2.05699835139178e+00, 2.03253797064833e+00, 2.00675277727576e+00, 1.97982220554275e+00, 1.95194528259612e+00, 1.92333532654763e+00, 1.89421440316543e+00, 1.86480781161577e+00, 1.83533884558300e+00, 1.80602403518606e+00, 1.77706902284387e+00, 1.74866516884499e+00, 1.72098692595420e+00, 1.69418997212066e+00, 1.66841004991817e+00, 1.64376243268380e+00, 1.62034192066907e+00, 1.59822326474826e+00, 1.57746191829214e+00, 1.55809502724081e+00, 1.54014258172056e+00, 1.52360866758561e+00, 1.50848277132919e+00, 1.49474110572235e+00, 1.48234793557775e+00, 1.47125689283951e+00, 1.46141227764965e+00, 1.45275034717011e+00, 1.44520059685932e+00, 1.43868703976926e+00, 1.43312948843297e+00, 1.42844484128556e+00, 1.42454837158716e+00, 1.42135501184881e+00, 1.41878062122737e+00, 1.41674321773621e+00, 1.41516415194582e+00, 1.41396919464482e+00, 1.41308950818870e+00, 1.41246247037066e+00, 1.41203232086519e+00, 1.41175060370247e+00, 1.41157638472346e+00, 1.41147623024567e+00, 1.41142394178171e+00, 1.41140005102451e+00, 1.41139108882101e+00, 1.41138865088853e+00, 1.41138829105496e+00, 1.41138826360700e+00, 1.41138777469548e+00, 1.41138446225050e+00, 1.41137228539035e+00, 1.41133982664379e+00, 1.41126879295183e+00, 1.41113276060854e+00, 1.41089621697802e+00, 1.41051394724885e+00, 1.40993080902823e+00, 1.40908193076492e+00, 1.40789336120049e+00, 1.40628318563307e+00, 1.40416311017837e+00, 1.40144049714177e+00, 1.39802081320052e+00, 1.39381042803328e+00, 1.38871967566638e+00, 1.38266606611275e+00, 1.37557751336929e+00, 1.36739543028177e+00, 1.35807753386888e+00, 1.34760020855880e+00, 1.33596029059417e+00, 1.32317616442509e+00, 1.30928809954278e+00, 1.29435780077558e+00, 1.27846719237014e+00, 1.26171650150867e+00, 1.24422174580862e+00, 1.22611175828965e+00, 1.20752490022787e+00, 1.18860561691461e+00, 1.16950098485346e+00, 1.15035738379580e+00, 1.13131740620420e+00, 1.11251709313515e+00, 1.09408356144389e+00, 1.07613306410756e+00, 1.05876950403374e+00, 1.04208340214813e+00, 1.02615130286257e+00, 1.01103558438634e+00, 9.96784628215615e-01, 9.83433292206470e-01, 9.71003625603402e-01, 9.59505762711302e-01, 9.48938934563841e-01, 9.39292544426962e-01, 9.30547262307613e-01, 9.22676104570932e-01, 9.15645476013999e-01, 9.09416162167588e-01, 9.03944268342286e-01, 8.99182108477653e-01, 8.95079050984435e-01, 8.91582330536437e-01, 8.88637834410602e-01, 8.86190869852975e-01, 8.84186915503442e-01, 8.82572355621349e-01, 8.81295191209837e-01, 8.80305717625639e-01, 8.79557154341022e-01, 8.79006209599424e-01, 8.78613561096692e-01, 8.78344233738183e-01, 8.78167857045753e-01, 8.78058787853512e-01, 8.77996088335216e-01, 8.77963354823187e-01, 8.77948398905076e-01, 8.77942788471408e-01, 8.77941262301903e-01, 8.77941037042646e-01, 8.77941019738707e-01, 8.77940709926072e-01, 8.77938610902319e-01, 8.77930894706871e-01, 8.77910326453900e-01, 8.77865314812439e-01, 8.77779117794354e-01, 8.77629238169135e-01, 8.77387040405331e-01, 8.77017618097440e-01, 8.76479936981220e-01, 8.75727273367886e-01, 8.74707960604731e-01, 8.73366446537595e-01, 8.71644652637436e-01, 8.69483610496140e-01, 8.66825334231834e-01, 8.63614868877413e-01, 8.59802436427394e-01, 8.55345584661713e-01, 8.50211231178825e-01, 8.44377488320347e-01, 8.37835155690847e-01, 8.30588777078773e-01, 8.22657178279030e-01, 8.14073431091948e-01, 8.04884224963374e-01, 7.95148668514123e-01, 7.84936584758441e-01, 7.74326401644153e-01, 7.63402769064648e-01, 7.52254050716687e-01, 7.40969841544178e-01, 7.29638648552267e-01, 7.18345846455961e-01, 7.07171984104363e-01, 6.96191478461444e-01, 6.85471695904319e-01, 6.75072390475918e-01, 6.65045448347399e-01, 6.55434877743481e-01, 6.46276982695689e-01, 6.37600664758011e-01, 6.29427806417922e-01, 6.21773700838396e-01, 6.14647502936727e-01, 6.08052685593677e-01, 6.01987491612221e-01, 5.96445376987290e-01, 5.91415444420257e-01, 5.86882868190997e-01, 5.82829312815086e-01, 5.79233348604454e-01, 5.76070867458611e-01, 5.73315502009862e-01, 5.70939050654864e-01, 5.68911910035681e-01, 5.67203515204859e-01, 5.65782786064823e-01, 5.64618576788291e-01, 5.63680122918761e-01, 5.62937478860611e-01, 5.62361936666122e-01, 5.61926415585056e-01, 5.61605810926991e-01, 5.61377290534111e-01, 5.61220527664823e-01, 5.61117860377024e-01, 5.61054369538844e-01, 5.61017870279386e-01, 5.60998814860326e-01, 5.60990108392396e-01, 5.60986842313582e-01, 5.60985953862627e-01, 5.60985822729194e-01, 5.60985812578920e-01, 5.60985629846002e-01, 5.60984391804631e-01, 5.60979840649956e-01, 5.60967709068201e-01, 5.60941160035988e-01, 5.60890317906455e-01, 5.60801910660473e-01, 5.60659041562062e-01, 5.60441106489208e-01, 5.60123870715458e-01, 5.59679715636229e-01, 5.59078061592132e-01, 5.58285967306063e-01, 5.57268899395822e-01, 5.55991657005166e-01, 5.54419427111497e-01, 5.52518936093158e-01, 5.50259653526183e-01, 5.47614995992703e-01, 5.44563473077673e-01, 5.41089715779499e-01, 5.37185330049390e-01, 5.32849525416744e-01, 5.28089480362958e-01, 5.22920421365491e-01, 5.17365409915641e-01, 5.11454849585078e-01, 5.05225741622738e-01, 4.98720731105145e-01, 4.91986995305929e-01, 4.85075031215558e-01, 4.78037400082250e-01, 4.70927483925195e-01, 4.63798302916707e-01, 4.56701434187970e-01, 4.49686062844201e-01, 4.42798185603067e-01, 4.36079977241285e-01, 4.29569320605417e-01, 4.23299492852543e-01, 4.17298994230507e-01, 4.11591501308152e-01, 4.06195924160592e-01, 4.01126546457895e-01, 3.96393228404023e-01, 3.92001654640481e-01, 3.87953612137842e-01, 3.84247286348568e-01, 3.80877567134334e-01, 3.77836358944307e-01, 3.75112892222425e-01, 3.72694034951863e-01, 3.70564604564365e-01, 3.68707681157801e-01, 3.67104923121410e-01, 3.65736885933657e-01, 3.64583344155712e-01, 3.63623615589646e-01, 3.62836885309704e-01, 3.62202525919002e-01, 3.61700409052872e-01, 3.61311201963218e-01, 3.61016642094384e-01, 3.60799782004172e-01, 3.60645196877421e-01, 3.60539147276441e-01, 3.60469690685626e-01, 3.60426736807327e-01, 3.60402043380681e-01, 3.60389151415134e-01, 3.60383261021781e-01, 3.60381051340319e-01, 3.60380450254183e-01, 3.60380361535164e-01, 3.60380354612261e-01, 3.60380229261800e-01, 3.60379379994541e-01, 3.60376258007364e-01, 3.60367936005623e-01, 3.60349723862775e-01, 3.60314846788391e-01, 3.60254199511957e-01, 3.60156188793013e-01, 3.60006675143033e-01, 3.59789022870793e-01, 3.59484265287421e-01, 3.59071388962441e-01, 3.58527737169680e-01, 3.57829528025484e-01, 3.56952477335092e-01, 3.55872510003188e-01, 3.54566537373628e-01, 3.53013271551646e-01, 3.51194042275552e-01, 3.49093577962066e-01, 3.46700710853909e-01, 3.44008967314309e-01, 3.41017008563448e-01, 3.37728894512047e-01, 3.34154153400483e-01, 3.30307651892724e-01, 3.26209272990624e-01, 3.21883421341873e-01, 3.17358385936515e-01, 3.12665597755966e-01, 3.07838823949687e-01, 3.02913340348610e-01, 2.97925120853634e-01, 2.92910076163743e-01, 2.87903366414469e-01, 2.82938803677197e-01, 2.78048351955455e-01, 2.73261725140941e-01, 2.68606077903414e-01, 2.64105780897630e-01, 2.59782269874400e-01, 2.55653957951658e-01, 2.51736200990102e-01, 2.48041307294535e-01, 2.44578584375553e-01, 2.41354417014321e-01, 2.38372372222230e-01, 2.35633327792549e-01, 2.33135621976196e-01, 2.30875222391248e-01, 2.28845912640750e-01, 2.27039495308013e-01, 2.25446010055741e-01, 2.24053965481723e-01, 2.22850583179437e-01, 2.21822052108124e-01, 2.20953790902669e-01, 2.20230715164242e-01, 2.19637506106906e-01, 2.19158876242005e-01, 2.18779827124379e-01, 2.18485893633876e-01, 2.18263368894512e-01, 2.18099503806712e-01, 2.17982675339378e-01, 2.17902518225330e-01, 2.17850015534623e-01, 2.17817544738104e-01, 2.17798877269125e-01, 2.17789131172218e-01, 2.17784678100998e-01, 2.17783007600410e-01, 2.17782553183753e-01, 2.17782486112803e-01, 2.17782480834600e-01, 2.17782384693635e-01, 2.17781733324514e-01, 2.17779338825636e-01, 2.17772955992437e-01, 2.17758987428248e-01, 2.17732236378779e-01, 2.17685717399526e-01, 2.17610533904342e-01, 2.17495830398690e-01, 2.17328825711718e-01, 2.17094931724752e-01, 2.16777959866370e-01, 2.16360414933284e-01, 2.15823872575022e-01, 2.15149433083347e-01, 2.14318240078491e-01, 2.13312048503703e-01, 2.12113822344058e-01, 2.10708339069186e-01, 2.09082775380743e-01, 2.07227247824335e-01, 2.05135282491362e-01, 2.02804190522670e-01, 2.00235330352243e-01, 1.97434243297508e-01, 1.94410655726088e-01, 1.91178348003213e-01, 1.87754897106139e-01, 1.84161305621593e-01, 1.80421534407600e-01, 1.76561959319798e-01, 1.72610774126987e-01, 1.68597362281865e-01, 1.64551659834601e-01, 1.60503530671406e-01, 1.56482173492549e-01, 1.52515577533705e-01, 1.48630041044456e-01, 1.44849763150113e-01, 1.41196516147993e-01, 1.37689401688903e-01, 1.34344690734196e-01, 1.31175743722409e-01, 1.28193004220982e-01, 1.25404056828295e-01, 1.22813738519653e-01, 1.20424292143447e-01, 1.18235551178088e-01, 1.16245145924047e-01, 1.14448722708197e-01, 1.12840169309564e-01, 1.11411841501315e-01, 1.10154787308410e-01, 1.09058967052512e-01, 1.08113468445460e-01, 1.07306716700698e-01, 1.06626679946430e-01, 1.06061070075242e-01, 1.05597538729647e-01, 1.05223867434717e-01, 1.04928150081783e-01, 1.04698965136478e-01, 1.04525534205585e-01, 1.04397863065010e-01, 1.04306860952278e-01, 1.04244434014980e-01, 1.04203549168375e-01, 1.04178265374230e-01, 1.04163730350386e-01, 1.04156141945602e-01, 1.04152674783734e-01, 1.04151374138517e-01, 1.04151020332040e-01, 1.04150968110936e-01, 1.04150963964777e-01, 1.04150887985561e-01, 1.04150373211446e-01, 1.04148480856502e-01, 1.04143436589666e-01, 1.04132397625186e-01, 1.04111257796930e-01, 1.04074498933542e-01, 1.04015096170386e-01, 1.03924484074852e-01, 1.03792589445533e-01, 1.03607935363101e-01, 1.03357819336303e-01, 1.03028566180996e-01, 1.02605853488448e-01, 1.02075104231425e-01, 1.01421937311947e-01, 1.00632662822839e-01, 9.96948047859689e-02, 9.85976304456884e-02, 9.73326622305635e-02, 9.58941466309255e-02, 9.42794538613754e-02, 9.24893835358276e-02, 9.05283548468674e-02, 8.84044648976925e-02, 8.61294055462594e-02, 8.37182370655043e-02, 8.11890252974123e-02, 7.85623573582086e-02, 7.58607586301352e-02, 7.31080402164608e-02, 7.03286107625800e-02, 6.75467885948819e-02, 6.47861493023640e-02, 6.20689395919598e-02, 5.94155816861388e-02, 5.68442844329056e-02, 5.43707695589717e-02, 5.20081142407136e-02, 4.97667046537771e-02, 4.76542889703566e-02, 4.56761127513801e-02, 4.38351163438142e-02, 4.21321729958493e-02, 4.05663485454966e-02, 3.91214884381412e-02, 3.78225533153795e-02, 3.66494926977552e-02, 3.55967391873406e-02, 3.46581057574511e-02, 3.38269547182080e-02, 3.30963397682353e-02, 3.24591235925737e-02, 3.19080744814713e-02, 3.14359459778899e-02, 3.10355436743919e-02, 3.06997830425217e-02, 3.04217416596515e-02, 3.01947084709013e-02, 3.00122318581302e-02, 2.98681673556871e-02, 2.97567249264574e-02, 2.96725148604651e-02, 2.96105906499449e-02, 2.95664866817589e-02, 2.95362483122901e-02, 2.95164518717570e-02, 2.95042123787397e-02, 2.94971772084267e-02, 2.94935045992446e-02, 2.94918266413760e-02, 2.94911971956324e-02, 2.94910259724617e-02, 2.94910007003253e-02, 2.94909986755485e-02, 2.94909613376575e-02, 2.94907083687264e-02, 2.94897784401158e-02, 2.94872996839288e-02, 2.94818754536053e-02, 2.94714891929502e-02, 2.94534329974212e-02, 2.94242644156171e-02, 2.93797959540569e-02, 2.93151215132781e-02, 2.92246834698346e-02, 2.91023831884374e-02, 2.89417362800819e-02, 2.87360718351266e-02, 2.84787721480058e-02, 2.81635462027836e-02, 2.77847266094042e-02, 2.73375760918881e-02, 2.68185864508865e-02, 2.62257506345733e-02, 2.55587876312973e-02, 2.48193007428128e-02, 2.40108526464198e-02, 2.31389455084710e-02, 2.22109009929881e-02, 2.12356427543909e-02, 2.02233921097559e-02, 1.91852951144611e-02, 1.81330052740876e-02, 1.70782498274207e-02, 1.60324084383134e-02, 1.50061311311234e-02, 1.40090177027403e-02, 1.30493743067319e-02, 1.21340553346911e-02, 1.12683910949727e-02, 1.04561950023268e-02, 9.69983871631541e-03, 9.00038027514682e-03, 8.35772883218056e-03, 7.77082991141498e-03, 7.23785677472932e-03, 6.75639606337342e-03, 6.32361886895438e-03, 5.93643140197742e-03, 5.59160215676968e-03, 5.28586473996739e-03, 5.01599725421897e-03, 4.77888031894480e-03, 4.57153651932884e-03, 4.39115440123317e-03, 4.23510016144490e-03, 4.10092001962425e-03, 3.98633596925180e-03, 3.88923725374468e-03, 3.80766953644951e-03, 3.73982335738532e-03, 3.68402310549953e-03, 3.63871739052968e-03, 3.60247137457748e-03, 3.57396132131527e-03, 3.55197134503982e-03, 3.53539209740461e-03, 3.52322092576429e-03, 3.51456288337459e-03, 3.50863187711070e-03, 3.50475120902871e-03, 3.50235280551203e-03, 3.50097452761844e-03, 3.50025510648777e-03, 3.49992643450037e-03, 3.49980314435783e-03, 3.49976960701781e-03, 3.49976465699342e-03, 3.49976425668075e-03, 3.49975682784986e-03, 3.49970649681097e-03, 3.49952147957566e-03, 3.49902833001181e-03, 3.49794927932834e-03, 3.49588351550523e-03, 3.49229347255431e-03, 3.48649728088191e-03, 3.47766860374321e-03, 3.46484510854594e-03, 3.44694673998447e-03, 3.42280471610223e-03, 3.39120170431073e-03, 3.35092291735606e-03, 3.30081689898520e-03, 3.23986359437356e-03, 3.16724602615182e-03, 3.08242068869902e-03, 2.98518083808830e-03, 2.87570641752666e-03, 2.75459461443006e-03, 2.62286611524587e-03, 2.48194400522550e-03, 2.33360480120126e-03, 2.17990400637547e-03, 2.02308142915797e-03, 1.86545386654302e-03, 1.70930422179794e-03, 1.55677644666241e-03, 1.40978481230292e-03, 1.26994407749237e-03, 1.13852447770313e-03, 1.01643255447215e-03, 9.04216135197089e-04, 8.02399520054564e-04, 7.10272833870818e-04, 6.27548914344800e-04, 5.54301976819715e-04, 4.89814560957313e-04, 4.32947221234770e-04, 3.83022905424468e-04, 3.39282982459110e-04, 3.00900978658092e-04, 2.67437617907262e-04, 2.38195228490667e-04, 2.12327930441347e-04, 1.89367193129897e-04, 1.68839113825925e-04, 1.50358097869230e-04, 1.33617800358365e-04, 1.18380172004606e-04, 1.04468561772819e-04, 9.17597288250249e-05, 8.01140192621119e-05, 6.94381327335236e-05, 5.96805167134562e-05, 5.08026276122564e-05, 4.27758992458394e-05, 3.55788604480267e-05, 2.91870708579722e-05, 2.35689337344525e-05, 1.86906552315236e-05, 1.45153440271883e-05, 1.10011622638240e-05, 8.10083087543489e-06, 5.76191948908900e-06, 3.92755791978061e-06, 2.53739074906889e-06, 1.52867806235690e-06, 8.37494077643441e-07, 4.00013021817358e-07, -1.42108547152020e-14, -2.13162820728030e-14, -2.13162820728030e-14],
[ 5.60000000000000e+01, 5.59999999457086e+01, 5.59999982669102e+01, 5.59999868927854e+01, 5.59999450800197e+01, 5.59998336203903e+01, 5.59995896828183e+01, 5.59991224714206e+01, 5.59983098561858e+01, 5.59969961030310e+01, 5.59949907962119e+01, 5.59920690096718e+01, 5.59879727464332e+01, 5.59824136281916e+01, 5.59750767824641e+01, 5.59656258434327e+01, 5.59537089560685e+01, 5.59389656517725e+01, 5.59210344474773e+01, 5.58995610080488e+01, 5.58742067023973e+01, 5.58446573750743e+01, 5.58106321454314e+01, 5.57718920343506e+01, 5.57282482038805e+01, 5.56795695790340e+01, 5.56257896063098e+01, 5.55669118943035e+01, 5.55030144829134e+01, 5.54342525037545e+01, 5.53608590288846e+01, 5.52831439590244e+01, 5.52014908745889e+01, 5.51163518586189e+01, 5.50282403933311e+01, 5.49377225232899e+01, 5.48454065597761e+01, 5.47519316655587e+01, 5.46579557018522e+01, 5.45641427374454e+01, 5.44711506143341e+01, 5.43796189376632e+01, 5.42901578151256e+01, 5.42033376177494e+01, 5.41196799758154e+01, 5.40396501653436e+01, 5.39636509858866e+01, 5.38920181816568e+01, 5.38250174163893e+01, 5.37628427778683e+01, 5.37056167599929e+01, 5.36533916475022e+01, 5.36061522097133e+01, 5.35638195936327e+01, 5.35262562925942e+01, 5.34932720534693e+01, 5.34646305731650e+01, 5.34400568235519e+01, 5.34192448334591e+01, 5.34018657474351e+01, 5.33875759742970e+01, 5.33760252348294e+01, 5.33668643181098e+01, 5.33597523605055e+01, 5.33543634709226e+01, 5.33503925406934e+01, 5.33475600965702e+01, 5.33456160803985e+01, 5.33443424685742e+01, 5.33435546775097e+01, 5.33431017369528e+01, 5.33428652498707e+01, 5.33427571943913e+01, 5.33427166586333e+01, 5.33427056318849e+01, 5.33427040043585e+01, 5.33427037908614e+01, 5.33426988166341e+01, 5.33426651154932e+01, 5.33425412262504e+01, 5.33422109808347e+01, 5.33414882357576e+01, 5.33401040500907e+01, 5.33376968070317e+01, 5.33338057065789e+01, 5.33278679771244e+01, 5.33192200641741e+01, 5.33071029549786e+01, 5.32906716887629e+01, 5.32690089839474e+01, 5.32411427871646e+01, 5.32060674153829e+01, 5.31627678238470e+01, 5.31102463911050e+01, 5.30475514711988e+01, 5.29738068268127e+01, 5.28882409332624e+01, 5.27902150430501e+01, 5.26792488402441e+01, 5.25550425124832e+01, 5.24174941455533e+01, 5.22667115155869e+01, 5.21030176197931e+01, 5.19269496339854e+01, 5.17392513810185e+01, 5.15408597902128e+01, 5.13328861694275e+01, 5.11165933505426e+01, 5.08933698765997e+01, 5.06647023726070e+01, 5.04321471076379e+01, 5.01973015593241e+01, 4.99617765863380e+01, 4.97271696455948e+01, 4.94950393850276e+01, 4.92668819019978e+01, 4.90441089623829e+01, 4.88280284948382e+01, 4.86198276774837e+01, 4.84205589000077e+01, 4.82311288088675e+01, 4.80522905381398e+01, 4.78846391138621e+01, 4.77286099163438e+01, 4.75844800076864e+01, 4.74523720860984e+01, 4.73322608112839e+01, 4.72239812475370e+01, 4.71272391827943e+01, 4.70416230937622e+01, 4.69666175331931e+01, 4.69016177125183e+01, 4.68459450410955e+01, 4.67988633639116e+01, 4.67595956154385e+01, 4.67273405817102e+01, 4.67012894390191e+01, 4.66806417192267e+01, 4.66646203416242e+01, 4.66524853521822e+01, 4.66435460249182e+01, 4.66371710082269e+01, 4.66327962416848e+01, 4.66299304253761e+01, 4.66281578924780e+01, 4.66271388140550e+01, 4.66266067493463e+01, 4.66263636414053e+01, 4.66262724427089e+01, 4.66262476344036e+01, 4.66262439727480e+01, 4.66262436494935e+01, 4.66262373154807e+01, 4.66261944016028e+01, 4.66260366455085e+01, 4.66256161246081e+01, 4.66246958194566e+01, 4.66229333041211e+01, 4.66198682098307e+01, 4.66149140271025e+01, 4.66073547176284e+01, 4.65963464988783e+01, 4.65809250391820e+01, 4.65600181564425e+01, 4.65324639477455e+01, 4.64970340897725e+01, 4.64524618434866e+01, 4.63974740767812e+01, 4.63308263952605e+01, 4.62513402575369e+01, 4.61579407643637e+01, 4.60496936700096e+01, 4.59258400899246e+01, 4.57858273900378e+01, 4.56293348547930e+01, 4.54562929499108e+01, 4.52668953157634e+01, 4.50616030251803e+01, 4.48411410742499e+01, 4.46064874910241e+01, 4.43588557875549e+01, 4.40996717029566e+01, 4.38305452777565e+01, 4.35532392879270e+01, 4.32696350025659e+01, 4.29816961681078e+01, 4.26914320998983e+01, 4.24008607813286e+01, 4.21119729041693e+01, 4.18266977908912e+01, 4.15468720869177e+01, 4.12742119844002e+01, 4.10102895489350e+01, 4.07565134934568e+01, 4.05141145128989e+01, 4.02841350892874e+01, 4.00674235200306e+01, 3.98646318204736e+01, 3.96762171027640e+01, 3.95024460268283e+01, 3.93434019424162e+01, 3.91989943801329e+01, 3.90689705924793e+01, 3.89529288844464e+01, 3.88503335015388e+01, 3.87605308583658e+01, 3.86827668924263e+01, 3.86162053163676e+01, 3.85599465199136e+01, 3.85130468428485e+01, 3.84745379065108e+01, 3.84434456572818e+01, 3.84188087458200e+01, 3.83996958445468e+01, 3.83852214970458e+01, 3.83745600998312e+01, 3.83669576416288e+01, 3.83617408689037e+01, 3.83583236084301e+01, 3.83562100563539e+01, 3.83549949351966e+01, 3.83543605212490e+01, 3.83540706497201e+01, 3.83539619084885e+01, 3.83539323281979e+01, 3.83539279622075e+01, 3.83539276120011e+01, 3.83539211489160e+01, 3.83538773605569e+01, 3.83537163898200e+01, 3.83532873001900e+01, 3.83523482442454e+01, 3.83505498252677e+01, 3.83474223141873e+01, 3.83423673029325e+01, 3.83346542811863e+01, 3.83234225146451e+01, 3.83076884763553e+01, 3.82863589356446e+01, 3.82582496394599e+01, 3.82221093282601e+01, 3.81766486154864e+01, 3.81205730321350e+01, 3.80526193062551e+01, 3.79715937254547e+01, 3.78764112362659e+01, 3.77661337870323e+01, 3.76400063404342e+01, 3.74974889849853e+01, 3.73382836737339e+01, 3.71623543169980e+01, 3.69699392482141e+01, 3.67615554506619e+01, 3.65379943500094e+01, 3.63003094071424e+01, 3.60497961477887e+01, 3.57879656028254e+01, 3.55165123780163e+01, 3.52372787112515e+01, 3.49522159132613e+01, 3.46633445434129e+01, 3.43727145732696e+01, 3.40823666644876e+01, 3.37942955531458e+01, 3.35104163957490e+01, 3.32325347876969e+01, 3.29623210032903e+01, 3.27012888218076e+01, 3.24507791016921e+01, 3.22119480605662e+01, 3.19857600351732e+01, 3.17729843535891e+01, 3.15741958649858e+01, 3.13897786411557e+01, 3.12199323799334e+01, 3.10646810884399e+01, 3.09238836872762e+01, 3.07972462413026e+01, 3.06843355784776e+01, 3.05845940996409e+01, 3.04973556067701e+01, 3.04218619851261e+01, 3.03572805673078e+01, 3.03027219869785e+01, 3.02572583000151e+01, 3.02199411147260e+01, 3.01898194347725e+01, 3.01659568830019e+01, 3.01474479461906e+01, 3.01334328640599e+01, 3.01231107846520e+01, 3.01157508249848e+01, 3.01107007122966e+01, 3.01073927370339e+01, 3.01053468223276e+01, 3.01041706028238e+01, 3.01035565039431e+01, 3.01032759156544e+01, 3.01031706570676e+01, 3.01031420241579e+01, 3.01031377979996e+01, 3.01031374720376e+01, 3.01031316188065e+01, 3.01030919623026e+01, 3.01029461807775e+01, 3.01025575805857e+01, 3.01017071377097e+01, 3.01000784345634e+01, 3.00972460958031e+01, 3.00926682556584e+01, 3.00856835065548e+01, 3.00755126823401e+01, 3.00612657161915e+01, 3.00419536800568e+01, 3.00165059568665e+01, 2.99837923181318e+01, 2.99426494801179e+01, 2.98919114973794e+01, 2.98304431329649e+01, 2.97571751339716e+01, 2.96711401566172e+01, 2.95715079457583e+01, 2.94576182988164e+01, 2.93290103495377e+01, 2.91854468034368e+01, 2.90269319463289e+01, 2.88537225220513e+01, 2.86663309166731e+01, 2.84655204664888e+01, 2.82522930927963e+01, 2.80278698248520e+01, 2.77936650764175e+01, 2.75512557743317e+01, 2.73023465948681e+01, 2.70487326501235e+01, 2.67922609917260e+01, 2.65347922711520e+01, 2.62781638190411e+01, 2.60241552802610e+01, 2.57744577668566e+01, 2.55306472713249e+01, 2.52941628292876e+01, 2.50662896526407e+01, 2.48481471955404e+01, 2.46406818900177e+01, 2.44446641145711e+01, 2.42606888483939e+01, 2.40891794170460e+01, 2.39303937450579e+01, 2.37844325839331e+01, 2.36512492643799e+01, 2.35306606136224e+01, 2.34223587688664e+01, 2.33259236964311e+01, 2.32408362863852e+01, 2.31664919318263e+01, 2.31022145199619e+01, 2.30472707606501e+01, 2.30008847601087e+01, 2.29622527170261e+01, 2.29305575797450e+01, 2.29049834612287e+01, 2.28847295679434e+01, 2.28690233642610e+01, 2.28571326697867e+01, 2.28483763768252e+01, 2.28421334817163e+01, 2.28378501485108e+01, 2.28350445665044e+01, 2.28333094230789e+01, 2.28323118872749e+01, 2.28317910835016e+01, 2.28315531239139e+01, 2.28314638570505e+01, 2.28314395742985e+01, 2.28314359902152e+01, 2.28314357188392e+01, 2.28314309114411e+01, 2.28313983406166e+01, 2.28312786069012e+01, 2.28309594413956e+01, 2.28302609586368e+01, 2.28289232926380e+01, 2.28265971179319e+01, 2.28228375023571e+01, 2.28171014739878e+01, 2.28087496084729e+01, 2.27970518517317e+01, 2.27811976828452e+01, 2.27603105901889e+01, 2.27334666792101e+01, 2.26997170542760e+01, 2.26581134247391e+01, 2.26077361858422e+01, 2.25477240313618e+01, 2.24773039834358e+01, 2.23958205944514e+01, 2.23027630050135e+01, 2.21977885475719e+01, 2.20807416791170e+01, 2.19516672129678e+01, 2.18108170941159e+01, 2.16586503093159e+01, 2.14958259161844e+01, 2.13231895805169e+01, 2.11417543884482e+01, 2.09526770107954e+01, 2.07572305080021e+01, 2.05567751547148e+01, 2.03527286287229e+01, 2.01465367632580e+01, 1.99396458338573e+01, 1.97334770809005e+01, 1.95294038988601e+01, 1.93287318897099e+01, 1.91326818047121e+01, 1.89423752938041e+01, 1.87588233375811e+01, 1.85829172350307e+01, 1.84154220379396e+01, 1.82569723398330e+01, 1.81080703302476e+01, 1.79690860101505e+01, 1.78402594355649e+01, 1.77217048229184e+01, 1.76134163209850e+01, 1.75152752376376e+01, 1.74270585076606e+01, 1.73484481988946e+01, 1.72790418731732e+01, 1.72183636397176e+01, 1.71658757561464e+01, 1.71209906418414e+01, 1.70830831680107e+01, 1.70515030784328e+01, 1.70255873763382e+01, 1.70046724892782e+01, 1.69881059989731e+01, 1.69752577011432e+01, 1.69655297451390e+01, 1.69583655982034e+01, 1.69532575869436e+01, 1.69497527904866e+01, 1.69474570960262e+01, 1.69460372769369e+01, 1.69452210140668e+01, 1.69447948489999e+01, 1.69446001301391e+01, 1.69445270842874e+01, 1.69445072140367e+01, 1.69445042812291e+01, 1.69445040610623e+01, 1.69445001858770e+01, 1.69444739309271e+01, 1.69443774149624e+01, 1.69441201389007e+01, 1.69435570968598e+01, 1.69424788052495e+01, 1.69406036510294e+01, 1.69375729139031e+01, 1.69329487633448e+01, 1.69262154683700e+01, 1.69167839836977e+01, 1.69039999881642e+01, 1.68871553482339e+01, 1.68655028607237e+01, 1.68382739958186e+01, 1.68046992178516e+01, 1.67640303135294e+01, 1.67155640141586e+01, 1.66586660707806e+01, 1.65927948407929e+01, 1.65175233832736e+01, 1.64325590479990e+01, 1.63377595876575e+01, 1.62331449283487e+01, 1.61189039009266e+01, 1.59953954622615e+01, 1.58631442141338e+01, 1.57228303458573e+01, 1.55752744647845e+01, 1.54214181067643e+01, 1.52623009970624e+01, 1.50990363175006e+01, 1.49327852902607e+01, 1.47647322959643e+01, 1.45960615176809e+01, 1.44279357903985e+01, 1.42614780047058e+01, 1.40977551308602e+01, 1.39377647400609e+01, 1.37824238154810e+01, 1.36325596474542e+01, 1.34889026584415e+01, 1.33520810656345e+01, 1.32226173344192e+01, 1.31009263919872e+01, 1.29873155584364e+01, 1.28819861225029e+01, 1.27850364529208e+01, 1.26964665048247e+01, 1.26161835599881e+01, 1.25440090320075e+01, 1.24796861711537e+01, 1.24228885146803e+01, 1.23732289422318e+01, 1.23302692082584e+01, 1.22935298306857e+01, 1.22625002155841e+01, 1.22366488908301e+01, 1.22154337086983e+01, 1.21983118600712e+01, 1.21847495243289e+01, 1.21742309622449e+01, 1.21662668475728e+01, 1.21604016293513e+01, 1.21562197234488e+01, 1.21533503497975e+01, 1.21514708613530e+01, 1.21503084512006e+01, 1.21496401735893e+01, 1.21492912703618e+01, 1.21491318531186e+01, 1.21490720501299e+01, 1.21490557822597e+01, 1.21490533811558e+01, 1.21490532014523e+01, 1.21490500457687e+01, 1.21490286655477e+01, 1.21489500695933e+01, 1.21487405616333e+01, 1.21482820585757e+01, 1.21474039703277e+01, 1.21458769673784e+01, 1.21434089243844e+01, 1.21396432838872e+01, 1.21341600328413e+01, 1.21264794241633e+01, 1.21160685043202e+01, 1.21023504247829e+01, 1.20847164197231e+01, 1.20625402258364e+01, 1.20351946057639e+01, 1.20020695194214e+01, 1.19625913747725e+01, 1.19162426898054e+01, 1.18625814200204e+01, 1.18012591595241e+01, 1.17320374162766e+01, 1.16548011979387e+01, 1.15695692256279e+01, 1.14765002166619e+01, 1.13758948387361e+01, 1.12681931289390e+01, 1.11539673815721e+01, 1.10339107274504e+01, 1.09088218415562e+01, 1.07795864119446e+01, 1.06471561663823e+01, 1.05125263705124e+01, 1.03767127706321e+01, 1.02407289480557e+01, 1.01055649794202e+01, 9.97216816462290e+00, 9.84142640510309e+00, 9.71415460910361e+00, 9.59108428912864e+00, 9.47285632083946e+00, 9.36001666888981e+00, 9.25301476441458e+00, 9.15220414502365e+00, 9.05784493904802e+00, 8.97010778463335e+00, 8.88907881158495e+00, 8.81476536932416e+00, 8.74710224827545e+00, 8.68595820644160e+00, 8.63114267183969e+00, 8.58241254089570e+00, 8.53947903062138e+00, 8.50201456762004e+00, 8.46965970991517e+00, 8.44203009924367e+00, 8.41872343331038e+00, 8.39932643148573e+00, 8.38342174578456e+00, 8.37059474419422e+00, 8.36044006819267e+00, 8.35256784333419e+00, 8.34660940368586e+00, 8.34222237997680e+00, 8.33909499936608e+00, 8.33694945284248e+00, 8.33554420470600e+00, 8.33467514672279e+00, 8.33417553591025e+00, 8.33391469729849e+00, 8.33379551872629e+00, 8.33375081084969e+00, 8.33373864923460e+00, 8.33373685420612e+00, 8.33373671981225e+00, 8.33373435911682e+00, 8.33371836506508e+00, 8.33365956934581e+00, 8.33350284236563e+00, 8.33315985336679e+00, 8.33250300435751e+00, 8.33136078835100e+00, 8.32951479768822e+00, 8.32669857911565e+00, 8.32259849710100e+00, 8.31685672336077e+00, 8.30907641638142e+00, 8.29882908789588e+00, 8.28566407288632e+00, 8.26911992637928e+00, 8.24873746682096e+00, 8.22407407728605e+00, 8.19471876967561e+00, 8.16030742288687e+00, 8.12053753439166e+00, 8.07518178647111e+00, 8.02409973282371e+00, 7.96724696467034e+00, 7.90468121961532e+00, 7.83656504763813e+00, 7.76316483701135e+00, 7.68484621358791e+00, 7.60206604078680e+00, 7.51536144419386e+00, 7.42533644480741e+00, 7.33264689371073e+00, 7.23798445010006e+00, 7.14206033374384e+00, 7.04558951930091e+00, 6.94927593658148e+00, 6.85379911405337e+00, 6.75980256898227e+00, 6.66788412020707e+00, 6.57858818800700e+00, 6.49240005438803e+00, 6.40974198698419e+00, 6.33097107867716e+00, 6.25637862007946e+00, 6.18619080054833e+00, 6.12057052354828e+00, 6.05962012285146e+00, 6.00338477641935e+00, 5.95185643373767e+00, 5.90497809814986e+00, 5.86264833593413e+00, 5.82472591561893e+00, 5.79103451137123e+00, 5.76136743057359e+00, 5.73549234590492e+00, 5.71315602511028e+00, 5.69408905676658e+00, 5.67801056805209e+00, 5.66463292177747e+00, 5.65366636620089e+00, 5.64482359425716e+00, 5.63782415082566e+00, 5.63239860966394e+00, 5.62829242768757e+00, 5.62526937519021e+00, 5.62311443778874e+00, 5.62163609023777e+00, 5.62066785405315e+00, 5.62006906971739e+00, 5.61972483907727e+00, 5.61954512282102e+00, 5.61946300970082e+00, 5.61943220635130e+00, 5.61942382710439e+00, 5.61942259034550e+00, 5.61942249750366e+00, 5.61942086340966e+00, 5.61940979218869e+00, 5.61936909330515e+00, 5.61926060576030e+00, 5.61902318715707e+00, 5.61856851644257e+00, 5.61777788504858e+00, 5.61650013479875e+00, 5.61455088161939e+00, 5.61171314045579e+00, 5.60773943712769e+00, 5.60235545533450e+00, 5.59526522012754e+00, 5.58615776213672e+00, 5.57471514009405e+00, 5.56062162470025e+00, 5.54357376835549e+00, 5.52329100817214e+00, 5.49952638089592e+00, 5.47207687560806e+00, 5.44079292106516e+00, 5.40558650590543e+00, 5.36643746625424e+00, 5.32339754804101e+00, 5.27659195853236e+00, 5.22621825734905e+00, 5.17254259218521e+00, 5.11589344633577e+00, 5.05665321983118e+00, 4.99524809879909e+00, 4.93213676497838e+00, 4.86779854836583e+00, 4.80272162485497e+00, 4.73739180790126e+00, 4.67228238626526e+00, 4.60784533262184e+00, 4.54450406817585e+00, 4.48264783484478e+00, 4.42262761443455e+00, 4.36475345312117e+00, 4.30929300222754e+00, 4.25647106952383e+00, 4.20646998194471e+00, 4.15943058217355e+00, 4.11545371016045e+00, 4.07460205043840e+00, 4.03690225349465e+00, 4.00234726282509e+00, 3.97089879835757e+00, 3.94248996205035e+00, 3.91702794318717e+00, 3.89439680961080e+00, 3.87446037705281e+00, 3.85706515184400e+00, 3.84204334256393e+00, 3.82921593359474e+00, 3.81839580821035e+00, 3.80939090109558e+00, 3.80200735062816e+00, 3.79605261067539e+00, 3.79133847104163e+00, 3.78768392615186e+00, 3.78491782417782e+00, 3.78288122462192e+00, 3.78142939217957e+00, 3.78043335901489e+00, 3.77978099654162e+00, 3.77937755113901e+00, 3.77914561528569e+00, 3.77902452537357e+00, 3.77896919873636e+00, 3.77894844386026e+00, 3.77894279803598e+00, 3.77894196472432e+00, 3.77894190190233e+00, 3.77894079263338e+00, 3.77893327717662e+00, 3.77890564961505e+00, 3.77883200508255e+00, 3.77867083787044e+00, 3.77836219121804e+00, 3.77782547662866e+00, 3.77695806900651e+00, 3.77563477029312e+00, 3.77370821776063e+00, 3.77101029259662e+00, 3.76735455872876e+00, 3.76253973002937e+00, 3.75635412565811e+00, 3.74858102870218e+00, 3.73900481389001e+00, 3.72741765863054e+00, 3.71362660181727e+00, 3.69746067152784e+00, 3.67877777121554e+00, 3.65747099925908e+00, 3.63347408278653e+00, 3.60676563563597e+00, 3.57737200181581e+00, 3.54536851684254e+00, 3.51087910436107e+00, 3.47407421722216e+00, 3.43516722269619e+00, 3.39440941312940e+00, 3.35208388986806e+00, 3.30849861546321e+00, 3.26397895500700e+00, 3.21886003194357e+00, 3.17347920847170e+00, 3.12816896851455e+00, 3.08325043578797e+00, 3.03902770489781e+00, 2.99578310409240e+00, 2.95377344879983e+00, 2.91322728963047e+00, 2.87434311073329e+00, 2.83728839688401e+00, 2.80219946187488e+00, 2.76918191681793e+00, 2.73831165388072e+00, 2.70963622690346e+00, 2.68317652295361e+00, 2.65892863567370e+00, 2.63686586993485e+00, 2.61694082583981e+00, 2.59908752699563e+00, 2.58322357213250e+00, 2.56925229994180e+00, 2.55706496415990e+00, 2.54654291943709e+00, 2.53755981862312e+00, 2.52998381917804e+00, 2.52367979102443e+00, 2.51851151097200e+00, 2.51434382064290e+00, 2.51104471645608e+00, 2.50848733257070e+00, 2.50655177159639e+00, 2.50512673411322e+00, 2.50411089720719e+00, 2.50341399469450e+00, 2.50295755757814e+00, 2.50267528236425e+00, 2.50251300668027e+00, 2.50242828546401e+00, 2.50238957594085e+00, 2.50237505471091e+00, 2.50237110458885e+00, 2.50237052155929e+00, 2.50237047736645e+00, 2.50236969386609e+00, 2.50236438553890e+00, 2.50234487161111e+00, 2.50229285498607e+00, 2.50217901960948e+00, 2.50196101789330e+00, 2.50158193188351e+00, 2.50096928396203e+00, 2.50003466353898e+00, 2.49867402478257e+00, 2.49676869612053e+00, 2.49418712415078e+00, 2.49078735198703e+00, 2.48642020457108e+00, 2.48093312131849e+00, 2.47417454061065e+00, 2.46599870295648e+00, 2.45627070284814e+00, 2.44487158687317e+00, 2.43170327138104e+00, 2.41669304074793e+00, 2.39979739025363e+00, 2.38100499780410e+00, 2.36033864654277e+00, 2.33785597412167e+00, 2.31364899032644e+00, 2.28784237737548e+00, 2.26059065988292e+00, 2.23207439721840e+00, 2.20249560349873e+00, 2.17207263496690e+00, 2.14103479857258e+00, 2.10961692926036e+00, 2.07805415934871e+00, 2.04657706590679e+00, 2.01540733674251e+00, 1.98475404815235e+00, 1.95481060283952e+00, 1.92575233789998e+00, 1.89773478238851e+00, 1.87089252207001e+00, 1.84533861472919e+00, 1.82116449141029e+00, 1.79844027558562e+00, 1.77721545212930e+00, 1.75751982012564e+00, 1.73936466741117e+00, 1.72274411006775e+00, 1.70763654669990e+00, 1.69400618504527e+00, 1.68180460693478e+00, 1.67097234631886e+00, 1.66144046335955e+00, 1.65313210477427e+00, 1.64596404608848e+00, 1.63984821476591e+00, 1.63469319411274e+00, 1.63040570643668e+00, 1.62689207047965e+00, 1.62405962314745e+00, 1.62181808971661e+00, 1.62008088077977e+00, 1.61876628898698e+00, 1.61779855487173e+00, 1.61710876930318e+00, 1.61663558075941e+00, 1.61632567882328e+00, 1.61613403095878e+00, 1.61602385739458e+00, 1.61596633829472e+00, 1.61594005766938e+00, 1.61593019895233e+00, 1.61592751714723e+00, 1.61592712131859e+00, 1.61592709112495e+00, 1.61592655330638e+00, 1.61592290950940e+00, 1.61590951457069e+00, 1.61587380892322e+00, 1.61579566971825e+00, 1.61564603042743e+00, 1.61538582721020e+00, 1.61496532540269e+00, 1.61432387361257e+00, 1.61339012558113e+00, 1.61208276053512e+00, 1.61031172050758e+00, 1.60797996748535e+00, 1.60498574394281e+00, 1.60122529748129e+00, 1.59659600460246e+00, 1.59099980141256e+00, 1.58434680219897e+00, 1.57655896273479e+00, 1.56757362649547e+00, 1.55734678128420e+00, 1.54585585318368e+00, 1.53310187560542e+00, 1.51911089375267e+00, 1.50393449813851e+00, 1.48764942291858e+00, 1.47035619291422e+00, 1.45217685410420e+00, 1.43325187280205e+00, 1.41373633569828e+00, 1.39379562372276e+00, 1.37360076477208e+00, 1.35332369128156e+00, 1.33313263593797e+00, 1.31318789031211e+00, 1.29363812551542e+00, 1.27461743153832e+00, 1.25624317555010e+00, 1.23861471461505e+00, 1.22181293254435e+00, 1.20590051217663e+00, 1.19092281040079e+00, 1.17690917819486e+00, 1.16387456284355e+00, 1.15182124193620e+00, 1.14074056392132e+00, 1.13061460205392e+00, 1.12141766194988e+00, 1.11311761332784e+00, 1.10567704118341e+00, 1.09905422950334e+00, 1.09320400182138e+00, 1.08807844834952e+00, 1.08362757034057e+00, 1.07979987002596e+00, 1.07654291004918e+00, 1.07380386065797e+00, 1.07153004667875e+00, 1.06966949993391e+00, 1.06817151660949e+00, 1.06698721340348e+00, 1.06607007131670e+00, 1.06537645191939e+00, 1.06486606805238e+00, 1.06450238939729e+00, 1.06425296330725e+00, 1.06408963278023e+00, 1.06398863641746e+00, 1.06393057945693e+00, 1.06390027020809e+00, 1.06388642205390e+00, 1.06388122719657e+00, 1.06387981407539e+00, 1.06387960550190e+00, 1.06387958947973e+00, 1.06387930261603e+00, 1.06387735907546e+00, 1.06387021446256e+00, 1.06385116991493e+00, 1.06380949325149e+00, 1.06372968459380e+00, 1.06359091900088e+00, 1.06336669684103e+00, 1.06302473032311e+00, 1.06252709133761e+00, 1.06183064088791e+00, 1.06088775324845e+00, 1.05964733796972e+00, 1.05805614959329e+00, 1.05606035847677e+00, 1.05360733704052e+00, 1.05064759527673e+00, 1.04713677942116e+00, 1.04303763076951e+00, 1.03832179053617e+00, 1.03297133415021e+00, 1.02697992662960e+00, 1.02035351070774e+00, 1.01311047063442e+00, 1.00528125456281e+00, 9.96907482856855e-01, 9.88040612843385e-01, 9.78740266381521e-01, 9.69072349768943e-01, 9.59107102547065e-01, 9.48917202128037e-01, 9.38576027316763e-01, 9.28156150714948e-01, 9.17728093949123e-01, 9.07359346749217e-01, 8.97113625702254e-01, 8.87050333307812e-01, 8.77224172701048e-01, 8.67684876178650e-01, 8.58477013588164e-01, 8.49639856744041e-01, 8.41207285891137e-01, 8.33207732293282e-01, 8.25664156612632e-01, 8.18594065892015e-01, 8.12009573112405e-01, 8.05917503088665e-01, 8.00319547488492e-01, 7.95212470473956e-01, 7.90588365172106e-01, 7.86434960020017e-01, 7.82735973028167e-01, 7.79471511114075e-01, 7.76618510800546e-01, 7.74151215670649e-01, 7.72041684970212e-01, 7.70260326625340e-01, 7.68776446713204e-01, 7.67558806141160e-01, 7.66576174036658e-01, 7.65797866236596e-01, 7.65194256411931e-01, 7.64737246893716e-01, 7.64400686292234e-01, 7.64160721605712e-01, 7.63996073746739e-01, 7.63888227274784e-01, 7.63821527560523e-01, 7.63783181524580e-01, 7.63763161344059e-01, 7.63754013936754e-01, 7.63750582425004e-01, 7.63749648970261e-01, 7.63749511194291e-01, 7.63749500529869e-01, 7.63749308540611e-01, 7.63748007785274e-01, 7.63743226073274e-01, 7.63730479801588e-01, 7.63702584983498e-01, 7.63649163397965e-01, 7.63556263837955e-01, 7.63406116196975e-01, 7.63177033516065e-01, 7.62843474016748e-01, 7.62376271433538e-01, 7.61743037482432e-01, 7.60908734947286e-01, 7.59836413587273e-01, 7.58488093951215e-01, 7.56825776469000e-01, 7.54812545308766e-01, 7.52413729046857e-01, 7.49598073942067e-01, 7.46338881321421e-01, 7.42615058992676e-01, 7.38412038205993e-01, 7.33722512679243e-01, 7.28546964370388e-01, 7.22893951419074e-01, 7.16780146075834e-01, 7.10230123394965e-01, 7.03275913895325e-01, 6.95956344333489e-01, 6.88316199483339e-01, 6.80405243948449e-01, 6.72277146397548e-01, 6.63988349279840e-01, 6.55596925295463e-01, 6.47161458052963e-01, 6.38739978929642e-01, 6.30388985724231e-01, 6.22162561848519e-01, 6.14111608107237e-01, 6.06283193050707e-01, 5.98720022797039e-01, 5.91460027299391e-01, 5.84536057306359e-01, 5.77975684625997e-01, 5.71801097564524e-01, 5.66029083337000e-01, 5.60671089608796e-01, 5.55733357919614e-01, 5.51217122408190e-01, 5.47118867875497e-01, 5.43430641717151e-01, 5.40140414568576e-01, 5.37232484603486e-01, 5.34687920292320e-01, 5.32485036055391e-01, 5.30599894649399e-01, 5.29006829337213e-01, 5.27678977959098e-01, 5.26588820024472e-01, 5.25708706966597e-01, 5.25011374855218e-01, 5.24470428254546e-01, 5.24060783653084e-01, 5.23759061070024e-01, 5.23543913124769e-01, 5.23396282075581e-01, 5.23299577078319e-01, 5.23239766139987e-01, 5.23205379848626e-01, 5.23187426832571e-01, 5.23179223888903e-01, 5.23176146671453e-01, 5.23175309592744e-01, 5.23175186041648e-01, 5.23175176400748e-01, 5.23175001836265e-01, 5.23173819136019e-01, 5.23169471402348e-01, 5.23157881921698e-01, 5.23132518530531e-01, 5.23083944233349e-01, 5.22999471968369e-01, 5.22862939228816e-01, 5.22654615710472e-01, 5.22351254215003e-01, 5.21926291613738e-01, 5.21350202686548e-01, 5.20591005035023e-01, 5.19614908017921e-01, 5.18387092844399e-01, 5.16872604762433e-01, 5.15037331991586e-01, 5.12849040082578e-01, 5.10278425242475e-01, 5.07300146389483e-01, 5.03893793822165e-01, 5.00044752824088e-01, 4.95744923520078e-01, 4.90993263849461e-01, 4.85796130342898e-01, 4.80167400936388e-01, 4.74128374569311e-01, 4.67707452929517e-01, 4.60939619568350e-01, 4.53865739991336e-01, 4.46531712735336e-01, 4.38987505645450e-01, 4.31286113600365e-01, 4.23482474047006e-01, 4.15632375252606e-01, 4.07791389548542e-01, 4.00013860363480e-01, 3.92351967774871e-01, 3.84854892821977e-01, 3.77568096044229e-01, 3.70532720753324e-01, 3.63785126563812e-01, 3.57356553885460e-01, 3.51272915647854e-01, 3.45554708715895e-01, 3.40217034464636e-01, 3.35269715947753e-01, 3.30717498058092e-01, 3.26560316987283e-01, 3.22793625998798e-01, 3.19408765825628e-01, 3.16393369640295e-01, 3.13731794268953e-01, 3.11405570900739e-01, 3.09393869795130e-01, 3.07673974289244e-01, 3.06221759699888e-01, 3.05012172515880e-01, 3.04019704658444e-01, 3.03218856680957e-01, 3.02584582742504e-01, 3.02092709199620e-01, 3.01720317896823e-01, 3.01446084848457e-01, 3.01250565116248e-01, 3.01116415366700e-01, 3.01028546858213e-01, 3.00974203420381e-01, 3.00942961258123e-01, 3.00926650010162e-01, 3.00919197260860e-01, 3.00916401476421e-01, 3.00915640955182e-01, 3.00915528703840e-01, 3.00915519870159e-01, 3.00915358966584e-01, 3.00914268821458e-01, 3.00910261336583e-01, 3.00899578883381e-01, 3.00876200748831e-01, 3.00831429410842e-01, 3.00753573515010e-01, 3.00627742625892e-01, 3.00435766728285e-01, 3.00156250966829e-01, 2.99764773126768e-01, 2.99234227753885e-01, 2.98535316498381e-01, 2.97637179212302e-01, 2.96508154584863e-01, 2.95116652825023e-01, 2.93432116371960e-01, 2.91426038222227e-01, 2.89073001683526e-01, 2.86351700734471e-01, 2.83245897211089e-01, 2.79745270227643e-01, 2.75846114915304e-01, 2.71551851902295e-01, 2.66873315932287e-01, 2.61828801371507e-01, 2.56443853626045e-01, 2.50750808036436e-01, 2.44788090847862e-01, 2.38599309498248e-01, 2.32232170817198e-01, 2.25737274920178e-01, 2.19166838862897e-01, 2.12573406933934e-01, 2.06008603527629e-01, 1.99521979902919e-01, 1.93159998202397e-01, 1.86965185610561e-01, 1.80975479456627e-01, 1.75223771527328e-01, 1.69737647950527e-01, 1.64539310697464e-01, 1.59645658725367e-01, 1.55068501434876e-01, 1.50814874511148e-01, 1.46887428142350e-01, 1.43284859644289e-01, 1.40002366119312e-01, 1.37032097357732e-01, 1.34363594194511e-01, 1.31984202484304e-01, 1.29879457373733e-01, 1.28033436366458e-01, 1.26429082628654e-01, 1.25048501998947e-01, 1.23873238253289e-01, 1.22884531394220e-01, 1.22063563200008e-01, 1.21391693130640e-01, 1.20850686118530e-01, 1.20422931967340e-01, 1.20091654240653e-01, 1.19841104835864e-01, 1.19656739084640e-01, 1.19525365339364e-01, 1.19435262691390e-01, 1.19376260765151e-01, 1.19339776423523e-01, 1.19318803632638e-01, 1.19307854549945e-01, 1.19302851961599e-01, 1.19300975340330e-01, 1.19300464855947e-01, 1.19300389509597e-01, 1.19300383527687e-01, 1.19300273901473e-01, 1.19299531169091e-01, 1.19296800826092e-01, 1.19289522885879e-01, 1.19273596062285e-01, 1.19243097242389e-01, 1.19190068779240e-01, 1.19104385498446e-01, 1.18973712511654e-01, 1.18783563779992e-01, 1.18517469646186e-01, 1.18157259013017e-01, 1.17683458279849e-01, 1.17075804430179e-01, 1.16313863790737e-01, 1.15377741126615e-01, 1.14248856259742e-01, 1.12910757858153e-01, 1.11349937164945e-01, 1.09556599055104e-01, 1.07525344778395e-01, 1.05255720832503e-01, 1.02752592160677e-01, 1.00026305504755e-01, 9.70926200649842e-02, 9.39723969571560e-02, 9.06910552005016e-02, 8.72778186412546e-02, 8.37647936438231e-02, 8.01859299014538e-02, 7.65759249209381e-02, 7.29691356783633e-02, 6.93985583119741e-02, 6.58949288907920e-02, 6.24859862873137e-02, 5.91959234809991e-02, 5.60450379559612e-02, 5.30495769349315e-02, 5.02217604458508e-02, 4.75699556431337e-02, 4.50989698762498e-02, 4.28104276883516e-02, 4.07031977995658e-02, 3.87738394770381e-02, 3.70170426901737e-02, 3.54260422812942e-02, 3.39929923483879e-02, 3.27092926026893e-02, 3.15658632799298e-02, 3.05533690693593e-02, 2.96623954343644e-02, 2.88835826873353e-02, 2.82077243595609e-02, 2.76258369106017e-02, 2.71292077888887e-02, 2.67094284093687e-02, 2.63584178615801e-02, 2.60684421893203e-02, 2.58321329631102e-02, 2.56425076658076e-02, 2.54929931905039e-02, 2.53774525728474e-02, 2.52902140111786e-02, 2.52261003326169e-02, 2.51804563983669e-02, 2.51491715518810e-02, 2.51286941256907e-02, 2.51160352336370e-02, 2.51087595604389e-02, 2.51049615639403e-02, 2.51032263559736e-02, 2.51025754402647e-02, 2.51023983773635e-02, 2.51023722433246e-02, 2.51023701494901e-02, 2.51023315381929e-02, 2.51020699420792e-02, 2.51011083040851e-02, 2.50985450606152e-02, 2.50929361263985e-02, 2.50821968806960e-02, 2.50635291713763e-02, 2.50333784715480e-02, 2.49874263963363e-02, 2.49206236723651e-02, 2.48272681673285e-02, 2.47011315370429e-02, 2.45356362560187e-02, 2.43240821506987e-02, 2.40599180426315e-02, 2.37370498686218e-02, 2.33501719843616e-02, 2.28951037637017e-02, 2.23691097040799e-02, 2.17711787448494e-02, 2.11022380679395e-02, 2.03652787889013e-02, 1.95653758724958e-02, 1.87095921386842e-02, 1.78067657364593e-02, 1.68671909301992e-02, 1.59022121682355e-02, 1.49237598187568e-02, 1.39438614684337e-02, 1.29741644872823e-02, 1.20255034284966e-02, 1.11075401235752e-02, 1.02284959522621e-02, 9.39498597421107e-03, 8.61195477531140e-03, 7.88270524523682e-03, 7.20900497354648e-03, 6.59125100168723e-03, 6.02867232721849e-03, 5.51955048418762e-03, 5.06144113960261e-03, 4.65138327778192e-03, 4.28608654232221e-03, 3.96209114887824e-03, 3.67589810380053e-03, 3.42407007541823e-03, 3.20330511124212e-03, 3.01048653320635e-03, 2.84271288941085e-03, 2.69731193594680e-03, 2.57184241021946e-03, 2.46408696019529e-03, 2.37203910972283e-03, 2.29388663874630e-03, 2.22799328115997e-03, 2.17288021560336e-03, 2.12720845194525e-03, 2.08976289680152e-03, 2.05943860771640e-03, 2.03522950729962e-03, 2.01621962058596e-03, 2.00157671473988e-03, 1.99054806036258e-03, 1.98245790151930e-03, 1.97670612129031e-03, 1.97276752926712e-03, 1.97019118068908e-03, 1.96859916619019e-03, 1.96768438460282e-03, 1.96720692207464e-03, 1.96698879653923e-03, 1.96690697512736e-03, 1.96688471820039e-03, 1.96688143314816e-03, 1.96688116746202e-03, 1.96687623736835e-03, 1.96684283545068e-03, 1.96672005056087e-03, 1.96639278264144e-03, 1.96567672457348e-03, 1.96430599984154e-03, 1.96192421247332e-03, 1.95807974637319e-03, 1.95222622520674e-03, 1.94372907915152e-03, 1.93187911114734e-03, 1.91591376683675e-03, 1.89504643982730e-03, 1.86850356191570e-03, 1.83556843764876e-03, 1.79562982890502e-03, 1.74823227909826e-03, 1.69312423432189e-03, 1.63029935822578e-03, 1.56002623945284e-03, 1.48286211278759e-03, 1.39964733200770e-03, 1.31147909590368e-03, 1.21966515346372e-03, 1.12566058767527e-03, 1.03099291695941e-03, 9.37182276238957e-04, 8.45664060342699e-04, 7.57721001889422e-04, 6.74430289507200e-04, 5.96629266425452e-04, 5.24900866167854e-04, 4.59577653757037e-04, 4.00761496635482e-04, 3.48354706758869e-04, 3.02098035952270e-04, 2.61611101102233e-04, 2.26431511066494e-04, 1.96049968643308e-04, 1.69939725019219e-04, 1.47579780168883e-04, 1.28472006835523e-04, 1.12152863103177e-04, 9.82005872671721e-05, 8.62388204225795e-05, 7.59375532197737e-05, 6.70121941794832e-05, 5.92214199457430e-05, 5.23642721788065e-05, 4.62767470068570e-05, 4.08279816141999e-05, 3.59161174401379e-05, 3.14639569864994e-05, 2.74145692600314e-05, 2.37270225440511e-05, 2.03724216234491e-05, 1.73303931205737e-05, 1.45861075893094e-05, 1.21278692120086e-05, 9.94526008923913e-06, 8.02779406683385e-06, 6.36401510689666e-06, 4.94096376968400e-06, 3.74393357560621e-06, 2.75644405434150e-06, 1.96036828015167e-06, 1.33616592279395e-06, 8.63186670585492e-07, 5.20021863792408e-07, -7.10542735760100e-15, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00],
[ 5.70000000000000e+01, 5.69999999443665e+01, 5.69999982240689e+01, 5.69999865687804e+01, 5.69999437224199e+01, 5.69998295075484e+01, 5.69995795399044e+01, 5.69991007790775e+01, 5.69982680757845e+01, 5.69969218455825e+01, 5.69948669642174e+01, 5.69918729425119e+01, 5.69876754002932e+01, 5.69819788210368e+01, 5.69744605333120e+01, 5.69647758332409e+01, 5.69525641351332e+01, 5.69374560157687e+01, 5.69190810012944e+01, 5.68970759334422e+01, 5.68710937421675e+01, 5.68408124428575e+01, 5.68059441660240e+01, 5.67662440145191e+01, 5.67215185275822e+01, 5.66716335137394e+01, 5.66165209987336e+01, 5.65561850246098e+01, 5.64907060369543e+01, 5.64202436139476e+01, 5.63450373268921e+01, 5.62654055784348e+01, 5.61817423401028e+01, 5.60945118002871e+01, 5.60042410302410e+01, 5.59115108703590e+01, 5.58169453232137e+01, 5.57211998060415e+01, 5.56249486583562e+01, 5.55288723178779e+01, 5.54336445706872e+01, 5.53399202527508e+01, 5.52483237347605e+01, 5.51594384664870e+01, 5.50737977963827e+01, 5.49918772220099e+01, 5.49140881707944e+01, 5.48407733609914e+01, 5.47722037506415e+01, 5.47085770476296e+01, 5.46500177259398e+01, 5.45965784705565e+01, 5.45482429548438e+01, 5.45049298383490e+01, 5.44664978588004e+01, 5.44327518789181e+01, 5.44034497361916e+01, 5.43783097320144e+01, 5.43570185858150e+01, 5.43392396706362e+01, 5.43246213396890e+01, 5.43128051495494e+01, 5.43034337856485e+01, 5.42961585002378e+01, 5.42906458826317e+01, 5.42865837965354e+01, 5.42836863397095e+01, 5.42816977068012e+01, 5.42803948663247e+01, 5.42795889966117e+01, 5.42791256619757e+01, 5.42788837480587e+01, 5.42787732129704e+01, 5.42787317470186e+01, 5.42787204672343e+01, 5.42787188023603e+01, 5.42787185839639e+01, 5.42787134955907e+01, 5.42786790210948e+01, 5.42785522889192e+01, 5.42782144653030e+01, 5.42774751355929e+01, 5.42760591886518e+01, 5.42735967133533e+01, 5.42696163452882e+01, 5.42635424205722e+01, 5.42546962022352e+01, 5.42423013430837e+01, 5.42254936372259e+01, 5.42033349910169e+01, 5.41748314138971e+01, 5.41389546918424e+01, 5.40946672627890e+01, 5.40409496668826e+01, 5.39768297980879e+01, 5.39014130426815e+01, 5.38139122624712e+01, 5.37136764783901e+01, 5.36002170501116e+01, 5.34732301496043e+01, 5.33326144110369e+01, 5.31784828203037e+01, 5.30111681863812e+01, 5.28312218976173e+01, 5.26394060733634e+01, 5.24366796242716e+01, 5.22241790768081e+01, 5.20031952506674e+01, 5.17751469741099e+01, 5.15415529826753e+01, 5.13040030007271e+01, 5.10641288024234e+01, 5.08235758443832e+01, 5.05839759021201e+01, 5.03469210502442e+01, 5.01139393004340e+01, 4.98864722273401e+01, 4.96658549359313e+01, 4.94532987218019e+01, 4.92498767296133e+01, 4.90565128236318e+01, 4.88739737638449e+01, 4.87028646553048e+01, 4.85436275299252e+01, 4.83965428435266e+01, 4.82617336301722e+01, 4.81391720451992e+01, 4.80286880370639e+01, 4.79299799045205e+01, 4.78426265102370e+01, 4.77661009287852e+01, 4.76997853035592e+01, 4.76429866739236e+01, 4.75949535128327e+01, 4.75548926893711e+01, 4.75219865435925e+01, 4.74954097361910e+01, 4.74743455163010e+01, 4.74580010401566e+01, 4.74456213740777e+01, 4.74365018293066e+01, 4.74299983048384e+01, 4.74255353578762e+01, 4.74226117792060e+01, 4.74208035208614e+01, 4.74197639032626e+01, 4.74192211150814e+01, 4.74189731074580e+01, 4.74188800707152e+01, 4.74188547624162e+01, 4.74188510269625e+01, 4.74188506971930e+01, 4.74188442355227e+01, 4.74188004567466e+01, 4.74186395211963e+01, 4.74182105250438e+01, 4.74172716720638e+01, 4.74154736356206e+01, 4.74123467704585e+01, 4.74072927512663e+01, 4.73995811193291e+01, 4.73883511081625e+01, 4.73726189913109e+01, 4.73512910478108e+01, 4.73231820712241e+01, 4.72870391563892e+01, 4.72415702865708e+01, 4.71854770184408e+01, 4.71174903332470e+01, 4.70364085036878e+01, 4.69411356351230e+01, 4.68307193971619e+01, 4.67043863886106e+01, 4.65615735948912e+01, 4.64019545170307e+01, 4.62254587805731e+01, 4.60322843626956e+01, 4.58229019800829e+01, 4.55980516144142e+01, 4.53587315603589e+01, 4.51061807077384e+01, 4.48418549785758e+01, 4.45673989272844e+01, 4.42846135091298e+01, 4.39954209806279e+01, 4.37018278658908e+01, 4.34058869311999e+01, 4.31096591492526e+01, 4.28151766717803e+01, 4.25244078241654e+01, 4.22392250591811e+01, 4.19613766523301e+01, 4.16924627047361e+01, 4.14339157725669e+01, 4.11869862001043e+01, 4.09527320266892e+01, 4.07320131843037e+01, 4.05254896081487e+01, 4.03336228422732e+01, 4.01566807243119e+01, 3.99947447631869e+01, 3.98477198671690e+01, 3.97153461253869e+01, 3.95972123853655e+01, 3.94927713973393e+01, 3.94013563104833e+01, 3.93221983064947e+01, 3.92544451432843e+01, 3.91971803581532e+01, 3.91494428487877e+01, 3.91102465153522e+01, 3.90785996119886e+01, 3.90535234254586e+01, 3.90340698767775e+01, 3.90193376324468e+01, 3.90084863185999e+01, 3.90007484563181e+01, 3.89954387805929e+01, 3.89919606684723e+01, 3.89898094819336e+01, 3.89885727246544e+01, 3.89879270146888e+01, 3.89876319819130e+01, 3.89875213045191e+01, 3.89874911975453e+01, 3.89874867538177e+01, 3.89874863973757e+01, 3.89874798192141e+01, 3.89874352511950e+01, 3.89872714143488e+01, 3.89868346847508e+01, 3.89858789090599e+01, 3.89840484702932e+01, 3.89808652781730e+01, 3.89757202759594e+01, 3.89678699601221e+01, 3.89564382985174e+01, 3.89404243040270e+01, 3.89187153708666e+01, 3.88901063077765e+01, 3.88533238057265e+01, 3.88070558601285e+01, 3.87499854351386e+01, 3.86808274208688e+01, 3.85983677078574e+01, 3.85015030050156e+01, 3.83892798776239e+01, 3.82609314009085e+01, 3.81159098297771e+01, 3.79539137883847e+01, 3.77749086879696e+01, 3.75791393809593e+01, 3.73671344352408e+01, 3.71397018354619e+01, 3.68979163510233e+01, 3.66430992128916e+01, 3.63767910770413e+01, 3.61007194952375e+01, 3.58167622536003e+01, 3.55269079825314e+01, 3.52332154084629e+01, 3.49377725351100e+01, 3.46426569331357e+01, 3.43498981955803e+01, 3.40614434824990e+01, 3.37791269234714e+01, 3.35046434621076e+01, 3.32395275119020e+01, 3.29851365603936e+01, 3.27426396321967e+01, 3.25130103281758e+01, 3.22970240192942e+01, 3.20952586988683e+01, 3.19080989821749e+01, 3.17357427737173e+01, 3.15782101816582e+01, 3.14353543285377e+01, 3.13068737739214e+01, 3.11923263195297e+01, 3.10911440064535e+01, 3.10026491360562e+01, 3.09260711517627e+01, 3.08605642098299e+01, 3.08052252457443e+01, 3.07591123120370e+01, 3.07212629265435e+01, 3.06907121315050e+01, 3.06665099278976e+01, 3.06477377205875e+01, 3.06335233928158e+01, 3.06230546269335e+01, 3.06155901051063e+01, 3.06104682603580e+01, 3.06071133047755e+01, 3.06050383362150e+01, 3.06038454141896e+01, 3.06032225952496e+01, 3.06029380227309e+01, 3.06028312695279e+01, 3.06028022300470e+01, 3.06027979438796e+01, 3.06027976132892e+01, 3.06027916769454e+01, 3.06027514573423e+01, 3.06026036058103e+01, 3.06022094877852e+01, 3.06013469694952e+01, 3.05996951413642e+01, 3.05968225908077e+01, 3.05921797650831e+01, 3.05850958811669e+01, 3.05747807420017e+01, 3.05603317037208e+01, 3.05407459033205e+01, 3.05149376982068e+01, 3.04817610874597e+01, 3.04400366817810e+01, 3.03885825708826e+01, 3.03262482136973e+01, 3.02519502625299e+01, 3.01647090447945e+01, 3.00636842848513e+01, 2.99482085729413e+01, 2.98178170945052e+01, 2.96722722315240e+01, 2.95115818395298e+01, 2.93360102809099e+01, 2.91460816379022e+01, 2.89425749095827e+01, 2.87265113839724e+01, 2.84991347378150e+01, 2.82618847274814e+01, 2.80163655795551e+01, 2.77643103638594e+01, 2.75075427373797e+01, 2.72479374897357e+01, 2.69873813030873e+01, 2.67277350619335e+01, 2.64707989101160e+01, 2.62182810550208e+01, 2.59717710707980e+01, 2.57327181704064e+01, 2.55024146247954e+01, 2.52819842341377e+01, 2.50723755260005e+01, 2.48743591868608e+01, 2.46885291345949e+01, 2.45153066083939e+01, 2.43549466784333e+01, 2.42075466448018e+01, 2.40730558861716e+01, 2.39512868171217e+01, 2.38419267057023e+01, 2.37445501804771e+01, 2.36586323136220e+01, 2.35835622017403e+01, 2.35186569795232e+01, 2.34631761955977e+01, 2.34163364583390e+01, 2.33773262261789e+01, 2.33453205764691e+01, 2.33194957438733e+01, 2.32990431782068e+01, 2.32831828371267e+01, 2.32711754052500e+01, 2.32623331216847e+01, 2.32560289052312e+01, 2.32517034920863e+01, 2.32488703448122e+01, 2.32471181521755e+01, 2.32461108143571e+01, 2.32455848929296e+01, 2.32453445950160e+01, 2.32452544509614e+01, 2.32452299295918e+01, 2.32452263102890e+01, 2.32452260362463e+01, 2.32452211816075e+01, 2.32451882907212e+01, 2.32450673804225e+01, 2.32447450785630e+01, 2.32440397318950e+01, 2.32426889204001e+01, 2.32403398846567e+01, 2.32365433171083e+01, 2.32307509029856e+01, 2.32223169196082e+01, 2.32105041103884e+01, 2.31944939384395e+01, 2.31734011913142e+01, 2.31462927518802e+01, 2.31122101722253e+01, 2.30701954929403e+01, 2.30193195484264e+01, 2.29587118034173e+01, 2.28875905936127e+01, 2.28052925132682e+01, 2.27112996240614e+01, 2.26052631694505e+01, 2.24870225788363e+01, 2.23566187400225e+01, 2.22143008006720e+01, 2.20605261125163e+01, 2.18959533281913e+01, 2.17214290635052e+01, 2.15379689072000e+01, 2.13467338566733e+01, 2.11490034503045e+01, 2.09461469372029e+01, 2.07395937731616e+01, 2.05308045754044e+01, 2.03212434415256e+01, 2.01123522815510e+01, 1.99055275678102e+01, 1.97020997080963e+01, 1.95033151111728e+01, 1.93103209406743e+01, 1.91241525296398e+01, 1.89457234304277e+01, 1.87758180798339e+01, 1.86150870493191e+01, 1.84640448181922e+01, 1.83230699569617e+01, 1.81924075499719e+01, 1.80721736344886e+01, 1.79623613987541e+01, 1.78628488697459e+01, 1.77734078316173e+01, 1.76937137422129e+01, 1.76233564491752e+01, 1.75618515404441e+01, 1.75086521895963e+01, 1.74631613703627e+01, 1.74247443154438e+01, 1.73927410834363e+01, 1.73664790770413e+01, 1.73452853295523e+01, 1.73284983492797e+01, 1.73154792874241e+01, 1.73056221780544e+01, 1.72983629926783e+01, 1.72931872589417e+01, 1.72896360146331e+01, 1.72873099045350e+01, 1.72858712776027e+01, 1.72850442031377e+01, 1.72846123937477e+01, 1.72844150960183e+01, 1.72843410827553e+01, 1.72843209493463e+01, 1.72843179776971e+01, 1.72843177546144e+01, 1.72843138281069e+01, 1.72842872254418e+01, 1.72841894312468e+01, 1.72839287479509e+01, 1.72833582495569e+01, 1.72822656792712e+01, 1.72803656978599e+01, 1.72772948429080e+01, 1.72726095048450e+01, 1.72657871620051e+01, 1.72562310418596e+01, 1.72432782865229e+01, 1.72262115958730e+01, 1.72042742005808e+01, 1.71766878813036e+01, 1.71426736030335e+01, 1.71014741814808e+01, 1.70523782504471e+01, 1.69947446665259e+01, 1.69280263826178e+01, 1.68517927572873e+01, 1.67657492545498e+01, 1.66697535377000e+01, 1.65638270775367e+01, 1.64481615819000e+01, 1.63231198065635e+01, 1.61892306169458e+01, 1.60471785163885e+01, 1.58977882096789e+01, 1.57420050896359e+01, 1.55808727741871e+01, 1.54155089405261e+01, 1.52470806788728e+01, 1.50767804276497e+01, 1.49058032931707e+01, 1.47353262614478e+01, 1.45664895421472e+01, 1.44003800925952e+01, 1.42380172704924e+01, 1.40803405448714e+01, 1.39281992236557e+01, 1.37823441966416e+01, 1.36434217170851e+01, 1.35119692402496e+01, 1.33884133042701e+01, 1.32730693878165e+01, 1.31661436239185e+01, 1.30677362020439e+01, 1.29778462588765e+01, 1.28963780449660e+01, 1.28231481579499e+01, 1.27578936489162e+01, 1.27002808309218e+01, 1.26499146419631e+01, 1.26063484340026e+01, 1.25690940716936e+01, 1.25376322275700e+01, 1.25114227546746e+01, 1.24899150042780e+01, 1.24725579379692e+01, 1.24588098631798e+01, 1.24481476026585e+01, 1.24400748950044e+01, 1.24341298181464e+01, 1.24298910329236e+01, 1.24269826610410e+01, 1.24250776408914e+01, 1.24238994451822e+01, 1.24232220941034e+01, 1.24228684541900e+01, 1.24227068728260e+01, 1.24226462580186e+01, 1.24226297693157e+01, 1.24226273356175e+01, 1.24226271534746e+01, 1.24226239549534e+01, 1.24226022845022e+01, 1.24225226216423e+01, 1.24223102697907e+01, 1.24218455433864e+01, 1.24209555383733e+01, 1.24194078173564e+01, 1.24169063026668e+01, 1.24130896262999e+01, 1.24075321339531e+01, 1.23997476794036e+01, 1.23891962734313e+01, 1.23752935670390e+01, 1.23574230515149e+01, 1.23349507491096e+01, 1.23072420508561e+01, 1.22736802375474e+01, 1.22336861032809e+01, 1.21867379970286e+01, 1.21323915159201e+01, 1.20702980336735e+01, 1.20002212368564e+01, 1.19220508760478e+01, 1.18358130209729e+01, 1.17416762372114e+01, 1.16399532724797e+01, 1.15310980447889e+01, 1.14156979520854e+01, 1.12944617599126e+01, 1.11682035547054e+01, 1.10378234587077e+01, 1.09042859710500e+01, 1.07685969126400e+01, 1.06317799983881e+01, 1.04948540332559e+01, 1.03588116309034e+01, 1.02246001959165e+01, 1.00931057105157e+01, 9.96513964663358e+00, 9.84142910762889e+00, 9.72261011136770e+00, 9.60922377310639e+00, 9.50171504072102e+00, 9.40043357741109e+00, 9.30563637327767e+00, 9.21749168840523e+00, 9.13608397572855e+00, 9.06141949153802e+00, 8.99343236581463e+00, 8.93199096656117e+00, 8.87690444696284e+00, 8.82792940859316e+00, 8.78477664657352e+00, 8.74711796334547e+00, 8.71459304705834e+00, 8.68681640958732e+00, 8.66338436934893e+00, 8.64388204716239e+00, 8.62789032149709e+00, 8.61499266488056e+00, 8.60478175853140e+00, 8.59686576002141e+00, 8.59087408146111e+00, 8.58646252556959e+00, 8.58331762577177e+00, 8.58116004522552e+00, 8.57974690866951e+00, 8.57887296956918e+00, 8.57837055175334e+00, 8.57810824738739e+00, 8.57798839904392e+00, 8.57794343990393e+00, 8.57793120993825e+00, 8.57792940482132e+00, 8.57792926967213e+00, 8.57792689570909e+00, 8.57791081176695e+00, 8.57785168559580e+00, 8.57769407771194e+00, 8.57734916067211e+00, 8.57668861844865e+00, 8.57553997745286e+00, 8.57368359568780e+00, 8.57085150701359e+00, 8.56672828190202e+00, 8.56095402212698e+00, 8.55312955224645e+00, 8.54282380355327e+00, 8.52958330549333e+00, 8.51294360605907e+00, 8.49244233916163e+00, 8.46763354866017e+00, 8.43810277317432e+00, 8.40348230261434e+00, 8.36346594729443e+00, 8.31782262425379e+00, 8.26640807211885e+00, 8.20917406140504e+00, 8.14617457280615e+00, 8.07756856730329e+00, 8.00361915848184e+00, 7.92468920372323e+00, 7.84123353792680e+00, 7.75378826119712e+00, 7.66295764264098e+00, 7.56939930306048e+00, 7.47380838377345e+00, 7.37690139848277e+00, 7.27940040825003e+00, 7.18201806902027e+00, 7.08544399132349e+00, 6.99033273610900e+00, 6.89729365926338e+00, 6.80688271623758e+00, 6.71959624975548e+00, 6.63586670790413e+00, 6.55606017644920e+00, 6.48047555792175e+00, 6.40934519174794e+00, 6.34283668584145e+00, 6.28105572173895e+00, 6.22404960245246e+00, 6.17181133302829e+00, 6.12428405504953e+00, 6.08136569366748e+00, 6.04291371456826e+00, 6.00874992437937e+00, 5.97866527814132e+00, 5.95242467953456e+00, 5.92977177264873e+00, 5.91043372829984e+00, 5.89412602410373e+00, 5.88055720711088e+00, 5.86943361257654e+00, 5.86046399437566e+00, 5.85336400378090e+00, 5.84786043590320e+00, 5.84369514903080e+00, 5.84062855312575e+00, 5.83844256119344e+00, 5.83694290196205e+00, 5.83596070452183e+00, 5.83535328485215e+00, 5.83500408945797e+00, 5.83482178106707e+00, 5.83473848355958e+00, 5.83470723590234e+00, 5.83469873579265e+00, 5.83469748119466e+00, 5.83469738701367e+00, 5.83469572934936e+00, 5.83468449843603e+00, 5.83464321250491e+00, 5.83453316009477e+00, 5.83429231676146e+00, 5.83383108704613e+00, 5.83302904876375e+00, 5.83173285997880e+00, 5.82975546930475e+00, 5.82687674606272e+00, 5.82284561662286e+00, 5.81738375508330e+00, 5.81019082878380e+00, 5.80095124121300e+00, 5.78934224706826e+00, 5.77504323854069e+00, 5.75774592209875e+00, 5.73716502668439e+00, 5.71304911438262e+00, 5.68519101125932e+00, 5.65343734723800e+00, 5.61769669666444e+00, 5.57794585054253e+00, 5.53423382912177e+00, 5.48668335744136e+00, 5.43548967010965e+00, 5.38091667428903e+00, 5.32329066726597e+00, 5.26299196058695e+00, 5.20044488976396e+00, 5.13610677250227e+00, 5.07045640970847e+00, 5.00398269976999e+00, 4.93717386348168e+00, 4.87050766765526e+00, 4.80444290781033e+00, 4.73941228330033e+00, 4.67581668766942e+00, 4.61402085301977e+00, 4.55435023293393e+00, 4.49708898140468e+00, 4.44247887896171e+00, 4.39071906427081e+00, 4.34196644328921e+00, 4.29633666389965e+00, 4.25390555924199e+00, 4.21471097681260e+00, 4.17875492288060e+00, 4.14600596327486e+00, 4.11640183241047e+00, 4.08985221251036e+00, 4.06624165398532e+00, 4.04543261529501e+00, 4.02726860570865e+00, 4.01157741670021e+00, 3.99817442694690e+00, 3.98686596204168e+00, 3.97745268338767e+00, 3.96973297192955e+00, 3.96350626228251e+00, 3.95857627252036e+00, 3.95475406557532e+00, 3.95186087106896e+00, 3.94973059251060e+00, 3.94821192501453e+00, 3.94717001352226e+00, 3.94648759110448e+00, 3.94606555097343e+00, 3.94582292366955e+00, 3.94569625147151e+00, 3.94563837415382e+00, 3.94561666241518e+00, 3.94561075629975e+00, 3.94560988456963e+00, 3.94560981885132e+00, 3.94560865844132e+00, 3.94560079649579e+00, 3.94557189519699e+00, 3.94549485528800e+00, 3.94532625706947e+00, 3.94500337797095e+00, 3.94444190934797e+00, 3.94353448272666e+00, 3.94215010110551e+00, 3.94013455334337e+00, 3.93731186806792e+00, 3.93348683682959e+00, 3.92844860350555e+00, 3.92197527782287e+00, 3.91383948570484e+00, 3.90381471935882e+00, 3.89168229819883e+00, 3.87723870154366e+00, 3.86030299017678e+00, 3.84072400136882e+00, 3.81838698572325e+00, 3.79321935813409e+00, 3.76519526147640e+00, 3.73433869038017e+00, 3.70072499108316e+00, 3.66448063711532e+00, 3.62578127287680e+00, 3.58484811053309e+00, 3.54194285257631e+00, 3.49736138624908e+00, 3.45142655164687e+00, 3.40448031938774e+00, 3.35687572489611e+00, 3.30896889510457e+00, 3.26111147197104e+00, 3.21364368935286e+00, 3.16688830033524e+00, 3.12114548663960e+00, 3.07668881593570e+00, 3.03376225199040e+00, 2.99257817082208e+00, 2.95331629619862e+00, 2.91612344114267e+00, 2.88111392832383e+00, 2.84837055985120e+00, 2.81794601377159e+00, 2.78986455790540e+00, 2.76412398891007e+00, 2.74069772332310e+00, 2.71953698591247e+00, 2.70057305752722e+00, 2.68371955882331e+00, 2.66887475714400e+00, 2.65592389117583e+00, 2.64474151174408e+00, 2.63519383742056e+00, 2.62714112083957e+00, 2.62044001626363e+00, 2.61494593166876e+00, 2.61051534022329e+00, 2.60700801739785e+00, 2.60428916200562e+00, 2.60223135314365e+00, 2.60071629110995e+00, 2.59963626955572e+00, 2.59889532880264e+00, 2.59841004652342e+00, 2.59810993164788e+00, 2.59793739990655e+00, 2.59784732409732e+00, 2.59780616801487e+00, 2.59779072899583e+00, 2.59778652921286e+00, 2.59778590933387e+00, 2.59778586234789e+00, 2.59778502932767e+00, 2.59777938549618e+00, 2.59775863821985e+00, 2.59770333394610e+00, 2.59758230365932e+00, 2.59735052293919e+00, 2.59694747547015e+00, 2.59629610038737e+00, 2.59530239384835e+00, 2.59385572303393e+00, 2.59182989554936e+00, 2.58908500800671e+00, 2.58547007367331e+00, 2.58082640009353e+00, 2.57499165376911e+00, 2.56780451134654e+00, 2.55910975724363e+00, 2.54876364908927e+00, 2.53663933832590e+00, 2.52263210782919e+00, 2.50666417537178e+00, 2.48868881450020e+00, 2.46869356497281e+00, 2.44670234363676e+00, 2.42277632174906e+00, 2.39701350241913e+00, 2.36954700644350e+00, 2.34054214956276e+00, 2.31019246211181e+00, 2.27871485692939e+00, 2.24634418866405e+00, 2.21332746496466e+00, 2.17991796766239e+00, 2.14636952239488e+00, 2.11293112229022e+00, 2.07984207022324e+00, 2.04732775961609e+00, 2.01559616984576e+00, 1.98483511194735e+00, 1.95521022513419e+00, 1.92686369535985e+00, 1.89991364373902e+00, 1.87445411487239e+00, 1.85055558271989e+00, 1.82826588446245e+00, 1.80761149065246e+00, 1.78859902264745e+00, 1.77121693539428e+00, 1.75543729430929e+00, 1.74121758821589e+00, 1.72850253477635e+00, 1.71722584925658e+00, 1.70731196055516e+00, 1.69867766920279e+00, 1.69123374980384e+00, 1.68488650480048e+00, 1.67953927748828e+00, 1.67509393019385e+00, 1.67145228898579e+00, 1.66851754995816e+00, 1.66619563484974e+00, 1.66439647644260e+00, 1.66303520769774e+00, 1.66203322372414e+00, 1.66131908306775e+00, 1.66082921486580e+00, 1.66050840129867e+00, 1.66031001037904e+00, 1.66019596208141e+00, 1.66013642055925e+00, 1.66010921599524e+00, 1.66009901069676e+00, 1.66009623461498e+00, 1.66009582487139e+00, 1.66009579361631e+00, 1.66009523689127e+00, 1.66009146500111e+00, 1.66007759919134e+00, 1.66004063847513e+00, 1.65995975309348e+00, 1.65980485650196e+00, 1.65953551691626e+00, 1.65910026474310e+00, 1.65843634656945e+00, 1.65746996916356e+00, 1.65611706737242e+00, 1.65428461705007e+00, 1.65187249752153e+00, 1.64877588718362e+00, 1.64488815076303e+00, 1.64010414821951e+00, 1.63432386473907e+00, 1.62745623085307e+00, 1.61942297417168e+00, 1.61016232261395e+00, 1.59963236643564e+00, 1.58781388550346e+00, 1.57471246104839e+00, 1.56035971833056e+00, 1.54481358769232e+00, 1.52815752437752e+00, 1.51049868894636e+00, 1.49196515570241e+00, 1.47270228105813e+00, 1.45286842154128e+00, 1.43263023651733e+00, 1.41215783852535e+00, 1.39162006045598e+00, 1.37118009170223e+00, 1.35099169569464e+00, 1.33119616287358e+00, 1.31192008311426e+00, 1.29327394886909e+00, 1.27535153417679e+00, 1.25822994307068e+00, 1.24197018865420e+00, 1.22661815231838e+00, 1.21220577890883e+00, 1.19875238335100e+00, 1.18626597157013e+00, 1.17474450808083e+00, 1.16417709017759e+00, 1.15454501161761e+00, 1.14582271593671e+00, 1.13797865110173e+00, 1.13097604381009e+00, 1.12477361442922e+00, 1.11932625338254e+00, 1.11458567764228e+00, 1.11050108258435e+00, 1.10701980029816e+00, 1.10408797086249e+00, 1.10165122833342e+00, 1.09965539842469e+00, 1.09804720026608e+00, 1.09677494039015e+00, 1.09578918343045e+00, 1.09504338114657e+00, 1.09449443954670e+00, 1.09410320325740e+00, 1.09383483702652e+00, 1.09365908638686e+00, 1.09355040299266e+00, 1.09348792478641e+00, 1.09345530667857e+00, 1.09344040345743e+00, 1.09343481278700e+00, 1.09343329199309e+00, 1.09343306752724e+00, 1.09343305028424e+00, 1.09343274156281e+00, 1.09343064993197e+00, 1.09342296091018e+00, 1.09340246506577e+00, 1.09335761175425e+00, 1.09327171744091e+00, 1.09312236260098e+00, 1.09288100865108e+00, 1.09251286327255e+00, 1.09197701844607e+00, 1.09122688001195e+00, 1.09021090004733e+00, 1.08887361333313e+00, 1.08715696639628e+00, 1.08500191214335e+00, 1.08235022552099e+00, 1.07914647708703e+00, 1.07534008356821e+00, 1.07088733958223e+00, 1.06575332509244e+00, 1.05991358105622e+00, 1.05335545273035e+00, 1.04607901677069e+00, 1.03809753380616e+00, 1.02943740031698e+00, 1.02013760890136e+00, 1.01024876014559e+00, 9.99831698127174e-01, 9.88955861722388e-01, 9.77697453498649e-01, 9.66137526995652e-01, 9.54360083268604e-01, 9.42450251493668e-01, 9.30492609469304e-01, 9.18569680957674e-01, 9.06760630155773e-01, 8.95140160315066e-01, 8.83777613856893e-01, 8.72736264835758e-01, 8.62072790563211e-01, 8.51836906887435e-01, 8.42071150458338e-01, 8.32810790936712e-01, 8.24083856343563e-01, 8.15911255513285e-01, 8.08306982853878e-01, 8.01278392255384e-01, 7.94826528903279e-01, 7.88946509796091e-01, 7.83627945763484e-01, 7.78855399567291e-01, 7.74608876100409e-01, 7.70864341670091e-01, 7.67594269797321e-01, 7.64768210860701e-01, 7.62353382281567e-01, 7.60315274846143e-01, 7.58618269285847e-01, 7.57226255516180e-01, 7.56103245122866e-01, 7.55213965952962e-01, 7.54524426202359e-01, 7.54002434366875e-01, 7.53618060997951e-01, 7.53344028501601e-01, 7.53156016308679e-01, 7.53032870638968e-01, 7.52956710723289e-01, 7.52912926615863e-01, 7.52890067447474e-01, 7.52879622923238e-01, 7.52875704824625e-01, 7.52874639007089e-01, 7.52874481694668e-01, 7.52874469518048e-01, 7.52874250304970e-01, 7.52872765104627e-01, 7.52867305358301e-01, 7.52852751734565e-01, 7.52820901766150e-01, 7.52759906343655e-01, 7.52653838162068e-01, 7.52482412824534e-01, 7.52220880496360e-01, 7.51840102628975e-01, 7.51306824142404e-01, 7.50584146362691e-01, 7.49632199787555e-01, 7.48409008326085e-01, 7.46871528099540e-01, 7.44976834477946e-01, 7.42683421262413e-01, 7.39952566535209e-01, 7.36749711595593e-01, 7.33045793573254e-01, 7.28818469715073e-01, 7.24053172724133e-01, 7.18743942297159e-01, 7.12893988092141e-01, 7.06515953186347e-01, 6.99631863595940e-01, 6.92272767197672e-01, 6.84478082814614e-01, 6.76294695722234e-01, 6.67775848061900e-01, 6.58979880680064e-01, 6.49968886315996e-01, 6.40807332937918e-01, 6.31560710954261e-01, 6.22294249944026e-01, 6.13071740601960e-01, 6.03954486965954e-01, 5.95000403746880e-01, 5.86263264540555e-01, 5.77792099388247e-01, 5.69630734778528e-01, 5.61817465693167e-01, 5.54384847448304e-01, 5.47359594518191e-01, 5.40762573878659e-01, 5.34608881329788e-01, 5.28907990478167e-01, 5.23663965376123e-01, 5.18875729095406e-01, 5.14537381666372e-01, 5.10638561782926e-01, 5.07164847415879e-01, 5.04098190954586e-01, 5.01417384681794e-01, 4.99098552256683e-01, 4.97115661436709e-01, 4.95441052528619e-01, 4.94045976073522e-01, 4.92901132119641e-01, 4.91977202227005e-01, 4.91245364207884e-01, 4.90677778675824e-01, 4.90248035891099e-01, 4.89931551273308e-01, 4.89705898394298e-01, 4.89551069319042e-01, 4.89449653830128e-01, 4.89386931305305e-01, 4.89350871717434e-01, 4.89332045257107e-01, 4.89323443267722e-01, 4.89320216360618e-01, 4.89319338563284e-01, 4.89319209002225e-01, 4.89319198892360e-01, 4.89319015836443e-01, 4.89317775605862e-01, 4.89313216389043e-01, 4.89301063206533e-01, 4.89274466343062e-01, 4.89223530471136e-01, 4.89134953511680e-01, 4.88991792143175e-01, 4.88773368381004e-01, 4.88455328711993e-01, 4.88009863766194e-01, 4.87406092306280e-01, 4.86610608317051e-01, 4.85588184145612e-01, 4.84302616047792e-01, 4.82717691344540e-01, 4.80798249041008e-01, 4.78511298725905e-01, 4.75827156474303e-01, 4.72720552015538e-01, 4.69171659252126e-01, 4.65167002851672e-01, 4.60700197364652e-01, 4.55772482120295e-01, 4.50393024640775e-01, 4.44578976785742e-01, 4.38355280349313e-01, 4.31754231320909e-01, 4.24814823473454e-01, 4.17581901523995e-01, 4.10105161269243e-01, 4.02438038597218e-01, 3.94636531173497e-01, 3.86757996164256e-01, 3.78859964978005e-01, 3.70999012092639e-01, 3.63229709965964e-01, 3.55603696128853e-01, 3.48168872114712e-01, 3.40968747154584e-01, 3.34041932849352e-01, 3.27421788620974e-01, 3.21136211964273e-01, 3.15207562662856e-01, 3.09652706438811e-01, 3.04483161118206e-01, 2.99705327345741e-01, 2.95320786081454e-01, 2.91326646369136e-01, 2.87715928912128e-01, 2.84477973516090e-01, 2.81598861147195e-01, 2.79061843916374e-01, 2.76847778496268e-01, 2.74935560132981e-01, 2.73302555427463e-01, 2.71925032405200e-01, 2.70778586109941e-01, 2.69838557150997e-01, 2.69080439455479e-01, 2.68480272110963e-01, 2.68015008828932e-01, 2.67662857412198e-01, 2.67403580843393e-01, 2.67218751365313e-01, 2.67091949282175e-01, 2.67008899202771e-01, 2.66957538038270e-01, 2.66928011170883e-01, 2.66912595688837e-01, 2.66905552274302e-01, 2.66902910053865e-01, 2.66902191306556e-01, 2.66902085220984e-01, 2.66902076872512e-01, 2.66901924807108e-01, 2.66900894542268e-01, 2.66897107189372e-01, 2.66887011571306e-01, 2.66864917934825e-01, 2.66822607421616e-01, 2.66749033603432e-01, 2.66630130892878e-01, 2.66448742877735e-01, 2.66184681467621e-01, 2.65814925029559e-01, 2.65313960256442e-01, 2.64654268224348e-01, 2.63806949881889e-01, 2.62742480116302e-01, 2.61431572728043e-01, 2.59846131438522e-01, 2.57960254922040e-01, 2.55751257378158e-01, 2.53200661004385e-01, 2.50295113553236e-01, 2.47027183550742e-01, 2.43395988145323e-01, 2.39407614146430e-01, 2.35075301521199e-01, 2.30419370060297e-01, 2.25466883426364e-01, 2.20251059439725e-01, 2.14810450142423e-01, 2.09187928744029e-01, 2.03429531853139e-01, 1.97583213449370e-01, 1.91697571124543e-01, 1.85820604847031e-01, 1.79998563923512e-01, 1.74274929424207e-01, 1.68689567963774e-01, 1.63278079539019e-01, 1.58071348407795e-01, 1.53095293014310e-01, 1.48370799806969e-01, 1.43913817229844e-01, 1.39735580606754e-01, 1.35842936110869e-01, 1.32238732238733e-01, 1.28922249671426e-01, 1.25889644469936e-01, 1.23134384556110e-01, 1.20647664769209e-01, 1.18418790967056e-01, 1.16435528291561e-01, 1.14684412608966e-01, 1.13151027143552e-01, 1.11820248418859e-01, 1.10676466838520e-01, 1.09703787656557e-01, 1.08886217810529e-01, 1.08207843244898e-01, 1.07653000070478e-01, 1.07206441335563e-01, 1.06853499474731e-01, 1.06580242808391e-01, 1.06373622936019e-01, 1.06221608639103e-01, 1.06113301094119e-01, 1.06039024874811e-01, 1.05990389426346e-01, 1.05960316414730e-01, 1.05943029526948e-01, 1.05934004825038e-01, 1.05929881503812e-01, 1.05928334726244e-01, 1.05927913967385e-01, 1.05927851864330e-01, 1.05927846933845e-01, 1.05927756576151e-01, 1.05927144390847e-01, 1.05924893952313e-01, 1.05918895258164e-01, 1.05905768017706e-01, 1.05880630674612e-01, 1.05836925676428e-01, 1.05766311083606e-01, 1.05658628058102e-01, 1.05501952906906e-01, 1.05282740962259e-01, 1.04986067418302e-01, 1.04595967109347e-01, 1.04095870977098e-01, 1.03469131629730e-01, 1.02699624114393e-01, 1.01772401170038e-01, 1.00674375366971e-01, 9.93949944101118e-02, 9.79268713264858e-02, 9.62663291192740e-02, 9.44138204720204e-02, 9.23741876877529e-02, 9.01567363239266e-02, 8.77751075415603e-02, 8.52469481438476e-02, 8.25933923203550e-02, 7.98383836450185e-02, 7.70078782308730e-02, 7.41289786475221e-02, 7.12290522120398e-02, 6.83348861478308e-02, 6.54719261434948e-02, 6.26636349638545e-02, 5.99309953748772e-02, 5.72921683795045e-02, 5.47623051944441e-02, 5.23535008085503e-02, 5.00748691650514e-02, 4.79327153177209e-02, 4.59307781842782e-02, 4.40705182875547e-02, 4.23514274844408e-02, 4.07713414391964e-02, 3.93267398699990e-02, 3.80130238822289e-02, 3.68247636435015e-02, 3.57559130426353e-02, 3.47999907135588e-02, 3.39502288882372e-02, 3.31996930176430e-02, 3.25413760446054e-02, 3.19682717134597e-02, 3.14734314433593e-02, 3.10500091470303e-02, 3.06912980097138e-02, 3.03907626996889e-02, 3.01420698093011e-02, 2.99391185585449e-02, 2.97760729730641e-02, 2.96473959147932e-02, 2.95478845421613e-02, 2.94727060558913e-02, 2.94174319940694e-02, 2.93780689202947e-02, 2.93510831334558e-02, 2.93334170374848e-02, 2.93224950405673e-02, 2.93162172863290e-02, 2.93129401126748e-02, 2.93114428316841e-02, 2.93108811639229e-02, 2.93107283780003e-02, 2.93107058271644e-02, 2.93107040204177e-02, 2.93106707030617e-02, 2.93104449738451e-02, 2.93096151811127e-02, 2.93074033485041e-02, 2.93025632643307e-02, 2.92932956766805e-02, 2.92771847470590e-02, 2.92511599174006e-02, 2.92114873022911e-02, 2.91537945282343e-02, 2.90731324759363e-02, 2.89640764976014e-02, 2.88208682571209e-02, 2.86375972953508e-02, 2.84084187550775e-02, 2.81278005181491e-02, 2.77907895430971e-02, 2.73932838082516e-02, 2.69322934254319e-02, 2.64061727015772e-02, 2.58148046666378e-02, 2.51597211947114e-02, 2.44441454405688e-02, 2.36729487186835e-02, 2.28525206825481e-02, 2.19905589741032e-02, 2.10957915035408e-02, 2.01776502854614e-02, 1.92459195483963e-02, 1.83103821974768e-02, 1.73804875564727e-02, 1.64650599300344e-02, 1.55720624934261e-02, 1.47084250951011e-02, 1.38799385422104e-02, 1.30912125285434e-02, 1.23456900664891e-02, 1.16457083731092e-02, 1.09925946811060e-02, 1.03867852552426e-02, 9.82795672945258e-03, 9.31516041596581e-03, 8.84695215431464e-03, 8.42151228781063e-03, 8.03675226497802e-03, 7.69040602829563e-03, 7.38010570100300e-03, 7.10344210011016e-03, 6.85801131056962e-03, 6.64144899499775e-03, 6.45145433753669e-03, 6.28580558053216e-03, 6.14236905703791e-03, 6.01910348612478e-03, 5.91406111241355e-03, 5.82538705378965e-03, 5.75131798446193e-03, 5.69018103480090e-03, 5.64039353645285e-03, 5.60046399039038e-03, 5.56899439000347e-03, 5.54468380127560e-03, 5.52633289527193e-03, 5.51284895692561e-03, 5.50325076927294e-03, 5.49667270153265e-03, 5.49236732058489e-03, 5.48970589731823e-03, 5.48817628758513e-03, 5.48737782233388e-03, 5.48701302637511e-03, 5.48687618360333e-03, 5.48683895947732e-03, 5.48683346529089e-03, 5.48683302096009e-03, 5.48682477547402e-03, 5.48676891132516e-03, 5.48656355281452e-03, 5.48601617329325e-03, 5.48481840314707e-03, 5.48252512982117e-03, 5.47853898458328e-03, 5.47210134887166e-03, 5.46229104128315e-03, 5.44803182898335e-03, 5.42810980221020e-03, 5.40120140644262e-03, 5.36591250448026e-03, 5.32082820930668e-03, 5.26457239035238e-03, 5.19587474955330e-03, 5.11364227374855e-03, 5.01703082680649e-03, 4.90551181444943e-03, 4.77892841261252e-03, 4.63753595393968e-03, 4.48202181524948e-03, 4.31350154704546e-03, 4.13348993076568e-03, 3.94384792503644e-03, 3.74670877423000e-03, 3.54438857470285e-03, 3.33928803398464e-03, 3.13379281355708e-03, 2.93017964661857e-03, 2.73053444001903e-03, 2.53668699716769e-03, 2.35016510140440e-03, 2.17216876455240e-03, 2.00356372281618e-03, 1.84489192835002e-03, 1.69639592437676e-03, 1.55805360165573e-03, 1.42961983495127e-03, 1.31067177890287e-03, 1.20065503846689e-03, 1.09892841762615e-03, 1.00480542450754e-03, 9.17591150944475e-04, 8.36613537465780e-04, 7.61248370316991e-04, 6.90937677163106e-04, 6.25201520819019e-04, 5.63643482379916e-04, 5.05950363091756e-04, 4.51886883396924e-04, 4.01286414010826e-04, 3.54038935455492e-04, 3.10077440197176e-04, 2.69363911051812e-04, 2.31875865864595e-04, 1.97594254565558e-04, 1.66493245757238e-04, 1.38532186994667e-04, 1.13649801633642e-04, 9.17605114665321e-05, 7.27526557113833e-05, 5.64883116709325e-05, 4.28044051687948e-05, 3.15148192484571e-05, 2.24132528146015e-05, 1.52766362901957e-05, 9.86896911570057e-06, 5.94549991461690e-06, 3.25722467664491e-06, 1.55574059590435e-06, 5.98583646649331e-07, -3.55271367880050e-14, -4.26325641456060e-14],
[ 5.80000000000000e+01, 5.79999999429693e+01, 5.79999981794680e+01, 5.79999862314673e+01, 5.79999423090579e+01, 5.79998252257755e+01, 5.79995689803817e+01, 5.79990781957540e+01, 5.79982245793435e+01, 5.79968445382485e+01, 5.79947380464230e+01, 5.79916688233040e+01, 5.79873658437790e+01, 5.79815261606761e+01, 5.79738189845466e+01, 5.79638909329483e+01, 5.79513723338648e+01, 5.79358844458457e+01, 5.79170474407106e+01, 5.78944889822382e+01, 5.78678532244506e+01, 5.78368100438204e+01, 5.78010643089302e+01, 5.77603649774000e+01, 5.77145137930679e+01, 5.76633733378954e+01, 5.76068741760377e+01, 5.75450208166196e+01, 5.74778962223869e+01, 5.74056646086768e+01, 5.73285723147650e+01, 5.72469465888004e+01, 5.71611922063411e+01, 5.70717859359009e+01, 5.69792689652583e+01, 5.68842375004798e+01, 5.67873318364923e+01, 5.66892242658241e+01, 5.65906062354951e+01, 5.64921751787673e+01, 5.63946214394875e+01, 5.62986156756633e+01, 5.62047970811024e+01, 5.61137627056223e+01, 5.60260580915771e+01, 5.59421693824181e+01, 5.58625170015672e+01, 5.57874509493787e+01, 5.57172477233804e+01, 5.56521088321449e+01, 5.55921608451412e+01, 5.55374568983868e+01, 5.54879795572223e+01, 5.54436449217220e+01, 5.54043078460836e+01, 5.53697681301002e+01, 5.53397775281925e+01, 5.53140474094868e+01, 5.52922568914104e+01, 5.52740612598098e+01, 5.52591004814180e+01, 5.52470076104317e+01, 5.52374168908101e+01, 5.52299713604122e+01, 5.52243297728007e+01, 5.52201726677816e+01, 5.52172074425580e+01, 5.52151723014774e+01, 5.52138389931425e+01, 5.52130142782545e+01, 5.52125401088193e+01, 5.52122925379591e+01, 5.52121794181206e+01, 5.52121369825323e+01, 5.52121254389826e+01, 5.52121237351774e+01, 5.52121235116741e+01, 5.52121183043149e+01, 5.52120830236717e+01, 5.52119533280148e+01, 5.52116076048561e+01, 5.52108509872455e+01, 5.52094019321173e+01, 5.52068818820623e+01, 5.52028084596527e+01, 5.51965925605964e+01, 5.51875396186860e+01, 5.51748552116496e+01, 5.51576550625743e+01, 5.51349793668888e+01, 5.51058112407651e+01, 5.50690989446409e+01, 5.50237813873175e+01, 5.49688162643656e+01, 5.49032100330394e+01, 5.48260487800966e+01, 5.47365289074983e+01, 5.46339864568222e+01, 5.45179238338953e+01, 5.43880327014272e+01, 5.42442118995858e+01, 5.40865794463618e+01, 5.39154779617670e+01, 5.37314732342773e+01, 5.35353460664045e+01, 5.33280779455372e+01, 5.31108314283187e+01, 5.28849263532517e+01, 5.26518130809054e+01, 5.24130439082104e+01, 5.21702436463403e+01, 5.19250801437909e+01, 5.16792353348796e+01, 5.14343772442287e+01, 5.11921333009999e+01, 5.09540653060848e+01, 5.07216464222768e+01, 5.04962405831045e+01, 5.02790847069619e+01, 5.00712740423994e+01, 4.98737508616338e+01, 4.96872965828945e+01, 4.95125272657716e+01, 4.93498923115587e+01, 4.91996761265025e+01, 4.90620024713272e+01, 4.89368412171904e+01, 4.88240172434764e+01, 4.87232212337081e+01, 4.86340221426351e+01, 4.85558811146369e+01, 4.84881666291739e+01, 4.84301706340548e+01, 4.83811254043448e+01, 4.83402208371818e+01, 4.83066218642116e+01, 4.82794856373742e+01, 4.82579781237644e+01, 4.82412897343074e+01, 4.82286496116597e+01, 4.82193382170798e+01, 4.82126978852430e+01, 4.82081410604051e+01, 4.82051559862191e+01, 4.82033096930931e+01, 4.82022482084979e+01, 4.82016940035493e+01, 4.82014407794593e+01, 4.82013457858315e+01, 4.82013199452115e+01, 4.82013161311883e+01, 4.82013157944826e+01, 4.82013091969011e+01, 4.82012644973065e+01, 4.82011001767295e+01, 4.82006621573378e+01, 4.81997035572133e+01, 4.81978677026030e+01, 4.81946750712556e+01, 4.81895147562081e+01, 4.81816409417983e+01, 4.81701747709883e+01, 4.81541118526592e+01, 4.81323355065985e+01, 4.81036356704146e+01, 4.80667331963109e+01, 4.80203090490343e+01, 4.79630376855364e+01, 4.78936236622741e+01, 4.78108402922765e+01, 4.77135689796050e+01, 4.76008377148135e+01, 4.74718571435026e+01, 4.73260526410776e+01, 4.71630909547011e+01, 4.69829002121203e+01, 4.67856824354170e+01, 4.65719181063095e+01, 4.63423627616705e+01, 4.60980359969683e+01, 4.58402035696230e+01, 4.55703534936401e+01, 4.52901671043911e+01, 4.50014860830265e+01, 4.47062764146878e+01, 4.44065902567754e+01, 4.41045267295714e+01, 4.38021926961310e+01, 4.35016646347667e+01, 4.32049526865379e+01, 4.29139678580945e+01, 4.26304931771938e+01, 4.23561593566324e+01, 4.20924252576271e+01, 4.18405631928045e+01, 4.16016489009892e+01, 4.13765558772220e+01, 4.11659536549233e+01, 4.09703096054885e+01, 4.07898938303573e+01, 4.06247867562664e+01, 4.04748890916564e+01, 4.03399338496798e+01, 4.02195001830629e+01, 4.01130288035811e+01, 4.00198387721383e+01, 3.99391454443594e+01, 3.98700793425558e+01, 3.98117057002477e+01, 3.97630443932033e+01, 3.97230899348457e+01, 3.96908311779856e+01, 3.96652703335019e+01, 3.96454408941426e+01, 3.96304240420835e+01, 3.96193631255993e+01, 3.96114758155302e+01, 3.96060635972073e+01, 3.96025183177315e+01, 3.96003255900072e+01, 3.95990649503836e+01, 3.95984067715984e+01, 3.95981060416969e+01, 3.95979932271114e+01, 3.95979625387694e+01, 3.95979580092330e+01, 3.95979576459081e+01, 3.95979509407217e+01, 3.95979055120908e+01, 3.95977385115466e+01, 3.95972933487060e+01, 3.95963191171451e+01, 3.95944533334667e+01, 3.95912086771468e+01, 3.95859643348540e+01, 3.95779624556640e+01, 3.95663101108759e+01, 3.95499870209801e+01, 3.95278591596328e+01, 3.94986981679541e+01, 3.94612063117258e+01, 3.94140464917239e+01, 3.93558765798959e+01, 3.92853871121588e+01, 3.92013411372751e+01, 3.91026148191713e+01, 3.89882372378927e+01, 3.88574277527576e+01, 3.87096292979807e+01, 3.85445360880345e+01, 3.83621144205032e+01, 3.81626155704428e+01, 3.79465801526765e+01, 3.77148337565877e+01, 3.74684740937706e+01, 3.72088503021938e+01, 3.69375353859194e+01, 3.66562930129833e+01, 3.63670400378595e+01, 3.60718061675335e+01, 3.57726921721208e+01, 3.54718279764211e+01, 3.51713318766001e+01, 3.48732720134738e+01, 3.45796310960367e+01, 3.42922751958377e+01, 3.40129272188553e+01, 3.37431454134377e+01, 3.34843070120088e+01, 3.32375968611037e+01, 3.30040006985836e+01, 3.27843026079009e+01, 3.25790861211441e+01, 3.23887384454220e+01, 3.22134573325568e+01, 3.20532601797209e+01, 3.19079950212893e+01, 3.17773531378051e+01, 3.16608830599150e+01, 3.15580057807345e+01, 3.14680310091067e+01, 3.13901742996581e+01, 3.13235748850819e+01, 3.12673140138487e+01, 3.12204335651998e+01, 3.11819546761257e+01, 3.11508960759981e+01, 3.11262917880489e+01, 3.11072078276541e+01, 3.10927575099175e+01, 3.10821149772892e+01, 3.10745265748477e+01, 3.10693197379651e+01, 3.10659091143256e+01, 3.10637997179359e+01, 3.10625870035571e+01, 3.10619538512917e+01, 3.10616645574064e+01, 3.10615560330583e+01, 3.10615265117832e+01, 3.10615221545040e+01, 3.10615218184287e+01, 3.10615157835951e+01, 3.10614748967089e+01, 3.10613245921774e+01, 3.10609239353863e+01, 3.10600471073176e+01, 3.10583678748273e+01, 3.10554476696710e+01, 3.10507278263071e+01, 3.10435264435780e+01, 3.10330402362577e+01, 3.10183516259010e+01, 3.09984411830437e+01, 3.09722053720242e+01, 3.09384793648635e+01, 3.08960644840446e+01, 3.08437596117971e+01, 3.07803956760062e+01, 3.07048721047603e+01, 3.06161939508545e+01, 3.05135082441989e+01, 3.03961380535003e+01, 3.02636127450422e+01, 3.01156930258560e+01, 2.99523895522755e+01, 2.97739741636391e+01, 2.95809831454692e+01, 2.93742123095299e+01, 2.91547040686587e+01, 2.89237270523857e+01, 2.86827491313363e+01, 2.84334049792698e+01, 2.81774594947783e+01, 2.79167685291623e+01, 2.76532384233079e+01, 2.73887858436746e+01, 2.71252993235202e+01, 2.68646037593328e+01, 2.66084288887480e+01, 2.63583824984252e+01, 2.61159288011226e+01, 2.58823721097016e+01, 2.56588456523547e+01, 2.54463051436956e+01, 2.52455265669696e+01, 2.50571075386892e+01, 2.48814716128741e+01, 2.47188749241015e+01, 2.45694146487253e+01, 2.44330388632021e+01, 2.43095574809455e+01, 2.41986540418715e+01, 2.40998982035621e+01, 2.40127588356153e+01, 2.39366176483928e+01, 2.38707832955017e+01, 2.38145058788691e+01, 2.37669917599791e+01, 2.37274185449149e+01, 2.36949500687058e+01, 2.36687511605811e+01, 2.36480019305741e+01, 2.36319112838138e+01, 2.36197293458099e+01, 2.36107584734818e+01, 2.36043625351466e+01, 2.35999741695389e+01, 2.35970997791860e+01, 2.35953220756418e+01, 2.35943000703052e+01, 2.35937664907399e+01, 2.35935226936750e+01, 2.35934312369542e+01, 2.35934063585063e+01, 2.35934026864994e+01, 2.35934024084661e+01, 2.35933974831344e+01, 2.35933641132924e+01, 2.35932414422944e+01, 2.35929144470010e+01, 2.35921988285820e+01, 2.35908283443595e+01, 2.35884450941455e+01, 2.35845932176821e+01, 2.35787163933980e+01, 2.35701594502081e+01, 2.35581743100504e+01, 2.35419303656116e+01, 2.35205292623054e+01, 2.34930238948509e+01, 2.34584412483748e+01, 2.34158085171180e+01, 2.33641817300715e+01, 2.33026759159266e+01, 2.32304956669108e+01, 2.31469648318959e+01, 2.30515540032020e+01, 2.29439044757979e+01, 2.28238474636109e+01, 2.26914175585933e+01, 2.25468597067382e+01, 2.23906293326233e+01, 2.22233856407287e+01, 2.20459785200605e+01, 2.18594298373358e+01, 2.16649101846526e+01, 2.14637123209377e+01, 2.12572225984149e+01, 2.10468916001669e+01, 2.08342050550982e+01, 2.06206558786293e+01, 2.04077179537720e+01, 2.01968220571211e+01, 1.99893341751160e+01, 1.97865363578223e+01, 1.95896102128005e+01, 1.93996231294166e+01, 1.92175173180306e+01, 1.90441017264957e+01, 1.88800468466584e+01, 1.87258823477904e+01, 1.85819973849497e+01, 1.84486433463815e+01, 1.83259387419771e+01, 1.82138759045982e+01, 1.81123291790524e+01, 1.80210643032822e+01, 1.79397487319301e+01, 1.78679627019114e+01, 1.78052108828643e+01, 1.77509344856538e+01, 1.77045237164604e+01, 1.76653304625238e+01, 1.76326810807445e+01, 1.76058891358774e+01, 1.75842679054617e+01, 1.75671424386016e+01, 1.75538609296981e+01, 1.75438051502694e+01, 1.75363996753837e+01, 1.75311196483655e+01, 1.75274968495938e+01, 1.75251238724421e+01, 1.75236562604881e+01, 1.75228125226506e+01, 1.75223720135179e+01, 1.75221707408131e+01, 1.75220952364011e+01, 1.75220746973635e+01, 1.75220716658444e+01, 1.75220714382673e+01, 1.75220674326523e+01, 1.75220402940225e+01, 1.75219405295668e+01, 1.75216745942953e+01, 1.75210926021626e+01, 1.75199780203040e+01, 1.75180397615548e+01, 1.75149070428448e+01, 1.75101273216909e+01, 1.75031675581432e+01, 1.74934189728088e+01, 1.74802053805314e+01, 1.74627950720125e+01, 1.74404160914704e+01, 1.74122746186380e+01, 1.73775760117891e+01, 1.73355479116382e+01, 1.72854646532466e+01, 1.72266720962245e+01, 1.71586118758733e+01, 1.70808440133266e+01, 1.69930668142854e+01, 1.68951330444804e+01, 1.67870615026080e+01, 1.66690433196900e+01, 1.65414425914555e+01, 1.64047912813595e+01, 1.62597786887098e+01, 1.61072361204560e+01, 1.59481176911634e+01, 1.57834783603434e+01, 1.56144503708408e+01, 1.54422191731605e+01, 1.52679997353904e+01, 1.50930138976972e+01, 1.49184691934635e+01, 1.47455393743216e+01, 1.45753467669527e+01, 1.44089465506646e+01, 1.42473130508725e+01, 1.40913281619594e+01, 1.39417720164327e+01, 1.37993159917385e+01, 1.36645180914892e+01, 1.35378206644214e+01, 1.34195503468047e+01, 1.33099200461949e+01, 1.32090327360359e+01, 1.31168868058875e+01, 1.30333827100808e+01, 1.29583306738015e+01, 1.28914592433936e+01, 1.28324245000517e+01, 1.27808197867682e+01, 1.27361858226283e+01, 1.26980210932852e+01, 1.26657924105224e+01, 1.26389455276809e+01, 1.26169156832271e+01, 1.25991379247103e+01, 1.25850570433072e+01, 1.25741369288224e+01, 1.25658691400549e+01, 1.25597804790164e+01, 1.25554393619960e+01, 1.25524607973134e+01, 1.25505098090340e+01, 1.25493031870494e+01, 1.25486094947916e+01, 1.25482473236313e+01, 1.25480818443434e+01, 1.25480197673021e+01, 1.25480028808372e+01, 1.25480003884302e+01, 1.25480002018934e+01, 1.25479969262134e+01, 1.25479747330004e+01, 1.25478931484212e+01, 1.25476756740345e+01, 1.25471997374089e+01, 1.25462882647115e+01, 1.25447032150651e+01, 1.25421413777397e+01, 1.25382326882193e+01, 1.25325412664634e+01, 1.25245693184227e+01, 1.25137639681271e+01, 1.24995270009920e+01, 1.24812273989872e+01, 1.24582164361316e+01, 1.24298449815616e+01, 1.23954825324568e+01, 1.23545373777522e+01, 1.23064771848321e+01, 1.22508492152114e+01, 1.21872993214358e+01, 1.21155888647612e+01, 1.20356087280542e+01, 1.19473896840885e+01, 1.18511085155530e+01, 1.17470894651702e+01, 1.16358008138697e+01, 1.15178466296711e+01, 1.13939539841726e+01, 1.12649561790066e+01, 1.11317727415985e+01, 1.09953871187272e+01, 1.08568231011945e+01, 1.07171210423338e+01, 1.05773148838657e+01, 1.04384108806130e+01, 1.03013687354618e+01, 1.01670856391217e+01, 1.00363834804618e+01, 9.90999927660053e+00, 9.78857868733933e+00, 9.67267233897265e+00, 9.56273459328222e+00, 9.45912435678650e+00, 9.36210752585226e+00, 9.27186069475500e+00, 9.18847580500057e+00, 9.11196547487567e+00, 9.04226880991502e+00, 8.97925755175282e+00, 8.92274247090127e+00, 8.87247994643848e+00, 8.82817870178894e+00, 8.78950668099336e+00, 8.75609805488013e+00, 8.72756034248479e+00, 8.70348162129527e+00, 8.68343778205435e+00, 8.66699976181568e+00, 8.65374066487162e+00, 8.64324265739892e+00, 8.63510350065619e+00, 8.62894257170129e+00, 8.62440621201279e+00, 8.62117224476091e+00, 8.61895351178783e+00, 8.61750030183414e+00, 8.61660157153068e+00, 8.61608489867665e+00, 8.61581515115378e+00, 8.61569190180095e+00, 8.61564566679750e+00, 8.61563308976343e+00, 8.61563123341983e+00, 8.61563109443532e+00, 8.61562865310252e+00, 8.61561211271937e+00, 8.61555130860734e+00, 8.61538922783024e+00, 8.61503452136974e+00, 8.61435522892263e+00, 8.61317397389352e+00, 8.61126485973883e+00, 8.60835226779227e+00, 8.60411171946315e+00, 8.59817292011346e+00, 8.59012504573435e+00, 8.57952426491619e+00, 8.56590340670329e+00, 8.54878359050968e+00, 8.52768753021553e+00, 8.50215411613087e+00, 8.47175377335700e+00, 8.43610400281845e+00, 8.39488444261844e+00, 8.34785075296227e+00, 8.29484663654947e+00, 8.23581336352287e+00, 8.17079627642547e+00, 8.09994790105915e+00, 8.02352747274726e+00, 7.94189688847241e+00, 7.85551329493590e+00, 7.76491870176512e+00, 7.67072715159300e+00, 7.57361007360442e+00, 7.47428049036437e+00, 7.37347674237905e+00, 7.27194634965575e+00, 7.17043055643672e+00, 7.06965001637653e+00, 6.97029198045559e+00, 6.87299925495299e+00, 6.77836110442502e+00, 6.68690618537187e+00, 6.59909751051033e+00, 6.51532936313630e+00, 6.43592600975342e+00, 6.36114200203868e+00, 6.29116382113890e+00, 6.22611260117398e+00, 6.16604767461421e+00, 6.11097070674802e+00, 6.06083022425565e+00, 6.01552638737675e+00, 5.97491589995571e+00, 5.93881699162434e+00, 5.90701443808047e+00, 5.87926460710050e+00, 5.85530052928497e+00, 5.83483699433843e+00, 5.81757566735678e+00, 5.80321020696235e+00, 5.79143135019708e+00, 5.78193190995577e+00, 5.77441161151178e+00, 5.76858167739473e+00, 5.76416905641508e+00, 5.76092018462836e+00, 5.75860416474833e+00, 5.75701525671032e+00, 5.75597458591065e+00, 5.75533099662825e+00, 5.75496100516919e+00, 5.75476783874038e+00, 5.75467957996460e+00, 5.75464647114151e+00, 5.75463746474743e+00, 5.75463613542284e+00, 5.75463603563222e+00, 5.75463427923376e+00, 5.75462237938126e+00, 5.75457863434660e+00, 5.75446202677475e+00, 5.75420683715024e+00, 5.75371813073107e+00, 5.75286830446998e+00, 5.75149486213078e+00, 5.74939956422475e+00, 5.74634907324618e+00, 5.74207718226866e+00, 5.73628867446229e+00, 5.72866481100831e+00, 5.71887038446883e+00, 5.70656220438173e+00, 5.69139880374702e+00, 5.67305107302557e+00, 5.65121344799808e+00, 5.62561520660515e+00, 5.59603137596709e+00, 5.56229272253260e+00, 5.52429430309786e+00, 5.48200209742632e+00, 5.43545732613916e+00, 5.38477817782965e+00, 5.33015881947943e+00, 5.27186573192006e+00, 5.21023158112106e+00, 5.14564698832533e+00, 5.07855068004138e+00, 5.00941856908921e+00, 4.93875233357097e+00, 4.86706802369057e+00, 4.79488514710096e+00, 4.72271657832376e+00, 4.65105952533067e+00, 4.58038768299355e+00, 4.51114461974914e+00, 4.44373838466169e+00, 4.37853728530992e+00, 4.31586676705872e+00, 4.25600731470958e+00, 4.19919329271931e+00, 4.14561263674739e+00, 4.09540730607506e+00, 4.04867440387537e+00, 4.00546787156400e+00, 3.96580066561786e+00, 3.92964733083856e+00, 3.89694689282976e+00, 3.86760600355478e+00, 3.84150228589684e+00, 3.81848783465053e+00, 3.79839284091270e+00, 3.78102931329582e+00, 3.76619487205011e+00, 3.75367659083943e+00, 3.74325485582480e+00, 3.73470720357373e+00, 3.72781208919123e+00, 3.72235252527346e+00, 3.71811952226014e+00, 3.71491525293684e+00, 3.71255585949624e+00, 3.71087382170943e+00, 3.70971981000121e+00, 3.70896395772868e+00, 3.70849650241210e+00, 3.70822776529197e+00, 3.70808746124471e+00, 3.70802335538323e+00, 3.70799930709646e+00, 3.70799276538267e+00, 3.70799179983953e+00, 3.70799172704882e+00, 3.70799044175894e+00, 3.70798173373363e+00, 3.70794972215423e+00, 3.70786439136494e+00, 3.70767764868598e+00, 3.70732002050087e+00, 3.70669812207079e+00, 3.70569302254118e+00, 3.70415961020568e+00, 3.70192704347425e+00, 3.69880034908520e+00, 3.69456319979238e+00, 3.68898186814444e+00, 3.68181031067630e+00, 3.67279628827109e+00, 3.66168837508939e+00, 3.64824365292461e+00, 3.63223583388482e+00, 3.61346350658375e+00, 3.59175816466588e+00, 3.56699165647807e+00, 3.53908269516180e+00, 3.50800209197045e+00, 3.47377642274113e+00, 3.43648990635680e+00, 3.39628436065197e+00, 3.35335719959866e+00, 3.30795753858544e+00, 3.26038057451215e+00, 3.21096049686306e+00, 3.16006225837768e+00, 3.10807258430156e+00, 3.05539062414567e+00, 3.00241864816456e+00, 2.94955316335520e+00, 2.89717677386354e+00, 2.84565104338895e+00, 2.79531053904353e+00, 2.74645815436732e+00, 2.69936173084375e+00, 2.65425192831474e+00, 2.61132123954024e+00, 2.57072400520976e+00, 2.53257726345561e+00, 2.49696226112602e+00, 2.46392646030294e+00, 2.43348588964539e+00, 2.40562771273213e+00, 2.38031291147131e+00, 2.35747900906662e+00, 2.33704278177962e+00, 2.31890293019486e+00, 2.30294269783533e+00, 2.28903243723235e+00, 2.27703213077763e+00, 2.26679387606266e+00, 2.25816434340052e+00, 2.25098720751715e+00, 2.24510554687026e+00, 2.24036419373668e+00, 2.23661200723507e+00, 2.23370403100216e+00, 2.23150348846296e+00, 2.22988356255649e+00, 2.22872890421902e+00, 2.22793681541339e+00, 2.22741805820223e+00, 2.22709725109353e+00, 2.22691282707210e+00, 2.22681654349417e+00, 2.22677255127601e+00, 2.22675604836418e+00, 2.22675155918034e+00, 2.22675089658650e+00, 2.22675084636280e+00, 2.22674995594062e+00, 2.22674392320399e+00, 2.22672174629710e+00, 2.22666263135196e+00, 2.22653326251278e+00, 2.22628551633299e+00, 2.22585471689398e+00, 2.22515851973888e+00, 2.22409650287531e+00, 2.22255053082625e+00, 2.22038594188615e+00, 2.21745358820504e+00, 2.21359273208613e+00, 2.20863476941656e+00, 2.20240771265104e+00, 2.19474132238589e+00, 2.18547273053646e+00, 2.17445235278910e+00, 2.16154984757072e+00, 2.14665984803920e+00, 2.12970717730958e+00, 2.11065125938328e+00, 2.08948946174667e+00, 2.06625915104211e+00, 2.04103830885008e+00, 2.01394463617235e+00, 1.98513316613716e+00, 1.95479249663045e+00, 1.92313983927418e+00, 1.89041515026255e+00, 1.85687465557120e+00, 1.82278410410444e+00, 1.78841207666239e+00, 1.75402364849023e+00, 1.71987465349696e+00, 1.68620673560327e+00, 1.65324330436051e+00, 1.62118644488122e+00, 1.59021477194670e+00, 1.56048216892014e+00, 1.53211731593500e+00, 1.50522388916103e+00, 1.47988130279933e+00, 1.45614586591242e+00, 1.43405223485105e+00, 1.41361505638117e+00, 1.39483071429171e+00, 1.37767911126004e+00, 1.36212543646261e+00, 1.34812188665102e+00, 1.33560932335054e+00, 1.32451886097604e+00, 1.31477338973896e+00, 1.30628904317469e+00, 1.29897662303579e+00, 1.29274299437945e+00, 1.28749246123976e+00, 1.28312812873389e+00, 1.27955325131160e+00, 1.27667255971841e+00, 1.27439355176648e+00, 1.27262772489788e+00, 1.27129172247749e+00, 1.27030836140953e+00, 1.26960750655042e+00, 1.26912675784192e+00, 1.26881191924127e+00, 1.26861722427120e+00, 1.26850530100711e+00, 1.26844686901023e+00, 1.26842017141257e+00, 1.26841015629712e+00, 1.26840743194997e+00, 1.26840702984229e+00, 1.26840699916968e+00, 1.26840645281973e+00, 1.26840275122222e+00, 1.26838914381711e+00, 1.26835287192325e+00, 1.26827349406518e+00, 1.26812148476283e+00, 1.26785716682178e+00, 1.26743003261582e+00, 1.26677850436167e+00, 1.26583017630136e+00, 1.26450257220417e+00, 1.26270443879381e+00, 1.26033757884575e+00, 1.25729920635873e+00, 1.25348478049457e+00, 1.24879124572215e+00, 1.24312057443814e+00, 1.23638347772113e+00, 1.22850312293935e+00, 1.21941867716784e+00, 1.20908848619782e+00, 1.19749270313497e+00, 1.18463519981270e+00, 1.17054462849509e+00, 1.15527454874837e+00, 1.13890259123569e+00, 1.12152869139337e+00, 1.10327248553427e+00, 1.08427001396468e+00, 1.06466991514397e+00, 1.04462931828933e+00, 1.02430964767817e+00, 1.00387254088742e+00, 9.83476057836242e-01, 9.63271321611636e-01, 9.43399690169493e-01, 9.23990514641361e-01, 9.05159499117737e-01, 8.87007641468422e-01, 8.69620706945582e-01, 8.53069166840115e-01, 8.37408523224234e-01, 8.22679936988763e-01, 8.08911078691288e-01, 7.96117128668861e-01, 7.84301862915349e-01, 7.73458772983425e-01, 7.63572180452307e-01, 7.54618318371087e-01, 7.46566362852256e-01, 7.39379407194740e-01, 7.33015378290098e-01, 7.27427900491378e-01, 7.22567115598537e-01, 7.18380469217280e-01, 7.14813473631239e-01, 7.11810455694390e-01, 7.09315295360703e-01, 7.07272156628072e-01, 7.05626208236623e-01, 7.04324326820029e-01, 7.03315770774147e-01, 7.02552809293934e-01, 7.01991288216405e-01, 7.01591112807080e-01, 7.01316627654087e-01, 7.01136875471001e-01, 7.01025719800370e-01, 7.00961821154230e-01, 7.00928461702446e-01, 7.00913219813089e-01, 7.00907502105899e-01, 7.00905946755796e-01, 7.00905717189549e-01, 7.00905699554752e-01, 7.00905383818466e-01, 7.00903244661326e-01, 7.00895380934178e-01, 7.00874419437866e-01, 7.00828547304624e-01, 7.00740702769430e-01, 7.00587959434323e-01, 7.00341136733627e-01, 6.99964666376481e-01, 6.99416740655671e-01, 6.98649763120095e-01, 6.97611114639351e-01, 6.96244237841434e-01, 6.94490029989616e-01, 6.92288518639835e-01, 6.89580776397058e-01, 6.86311011817331e-01, 6.82428754555332e-01, 6.77891036248226e-01, 6.72664456599698e-01, 6.66727018867789e-01, 6.60069622231028e-01, 6.52697111296398e-01, 6.44628805234568e-01, 6.35898459381636e-01, 6.26553648158790e-01, 6.16654596446672e-01, 6.06272523255925e-01, 5.95487592881710e-01, 5.84386591576965e-01, 5.73060460066067e-01, 5.61601813237424e-01, 5.50102568761830e-01, 5.38651788025675e-01, 5.27333808304711e-01, 5.16226717565310e-01, 5.05401195629872e-01, 4.94919720214561e-01, 4.84836115399105e-01, 4.75195404498059e-01, 4.66033919392927e-01, 4.57379613841212e-01, 4.49252528337141e-01, 4.41665357742949e-01, 4.34624079033718e-01, 4.28128604058202e-01, 4.22173430298741e-01, 4.16748270500179e-01, 4.11838649215440e-01, 4.07426460461515e-01, 4.03490485630336e-01, 4.00006874506317e-01, 3.96949594737336e-01, 3.94290856471116e-01, 3.92001519213679e-01, 3.90051487421033e-01, 3.88410100040687e-01, 3.87046517328194e-01, 3.85930105943530e-01, 3.85030820763518e-01, 3.84319579228340e-01, 3.83768621579449e-01, 3.83351848254037e-01, 3.83045124170003e-01, 3.82826538829221e-01, 3.82676611196729e-01, 3.82578429223713e-01, 3.82517715642109e-01, 3.82482814159779e-01, 3.82464593252301e-01, 3.82456268154513e-01, 3.82453145152894e-01, 3.82452295623558e-01, 3.82452170234913e-01, 3.82452160529319e-01, 3.82451985801715e-01, 3.82450801998083e-01, 3.82446450235058e-01, 3.82434850219511e-01, 3.82409464817698e-01, 3.82360852370702e-01, 3.82276326288782e-01, 3.82139740191690e-01, 3.81931415299690e-01, 3.81628224672298e-01, 3.81203845915351e-01, 3.80629189788415e-01, 3.79873006445663e-01, 3.78902663658749e-01, 3.77685082332299e-01, 3.76187804237254e-01, 3.74380155785772e-01, 3.72234460803966e-01, 3.69727245838597e-01, 3.66840374911725e-01, 3.63562048110062e-01, 3.59887600988465e-01, 3.55820050000446e-01, 3.51370342889474e-01, 3.46557291267558e-01, 3.41407183851118e-01, 3.35953100868679e-01, 3.30233970636371e-01, 3.24293425986291e-01, 3.18178529395894e-01, 3.11938440319651e-01, 3.05623096262750e-01, 2.99281971300822e-01, 2.92962963444161e-01, 2.86711447271180e-01, 2.80569512499731e-01, 2.74575394318248e-01, 2.68763088653444e-01, 2.63162135901197e-01, 2.57797550278497e-01, 2.52689868727700e-01, 2.47855292767895e-01, 2.43305898213137e-01, 2.39049890593648e-01, 2.35091887797637e-01, 2.31433215397630e-01, 2.28072203962711e-01, 2.25004481144545e-01, 2.22223254319019e-01, 2.19719581997531e-01, 2.17482634082813e-01, 2.15499942354953e-01, 2.13757643378173e-01, 2.12240716367269e-01, 2.10933218494063e-01, 2.09818519695929e-01, 2.08879538321263e-01, 2.08098977967111e-01, 2.07459564703015e-01, 2.06944282613563e-01, 2.06536604328100e-01, 2.06220712040611e-01, 2.05981703563751e-01, 2.05805777305358e-01, 2.05680389787048e-01, 2.05594379494954e-01, 2.05538051486641e-01, 2.05503218251629e-01, 2.05483193780751e-01, 2.05472739544668e-01, 2.05467962996018e-01, 2.05466171160374e-01, 2.05465683738808e-01, 2.05465611796448e-01, 2.05465606182671e-01, 2.05465504535859e-01, 2.05464815864424e-01, 2.05462284246799e-01, 2.05455535971225e-01, 2.05440767979127e-01, 2.05412487208598e-01, 2.05363312033427e-01, 2.05283846161493e-01, 2.05162634344916e-01, 2.04986207940117e-01, 2.04739226571533e-01, 2.04404719727279e-01, 2.03964428896640e-01, 2.03399246763631e-01, 2.02689745023996e-01, 2.01816776765824e-01, 2.00762133389311e-01, 1.99509230250150e-01, 1.98043790240590e-01, 1.96354491082730e-01, 1.94433540874840e-01, 1.92277147913117e-01, 1.89885855244427e-01, 1.87264717654195e-01, 1.84423308329080e-01, 1.81375553372042e-01, 1.78139403567862e-01, 1.74736363114413e-01, 1.71190903383575e-01, 1.67529795381959e-01, 1.63781397061257e-01, 1.59974931025538e-01, 1.56139784911780e-01, 1.52304861471670e-01, 1.48497998981561e-01, 1.44745475869520e-01, 1.41071607053235e-01, 1.37498433883742e-01, 1.34045505002646e-01, 1.30729741880167e-01, 1.27565380262517e-01, 1.24563977157649e-01, 1.21734472264919e-01, 1.19083292792112e-01, 1.16614491226329e-01, 1.14329906626558e-01, 1.12229341214388e-01, 1.10310745361680e-01, 1.08570405472904e-01, 1.07003130697784e-01, 1.05602435817090e-01, 1.04360718920454e-01, 1.03269433539765e-01, 1.02319255661477e-01, 1.01500246483745e-01, 1.00802011928295e-01, 1.00213859775948e-01, 9.97249549091990e-02, 9.93244725569937e-02, 9.90017487098100e-02, 9.87464260772271e-02, 9.85485931862229e-02, 9.83989135432495e-02, 9.82887412908866e-02, 9.82102195433967e-02, 9.81563576241342e-02, 9.81210837718791e-02, 9.80992705194339e-02, 9.80867308373377e-02, 9.80801842171601e-02, 9.80771930634674e-02, 9.80760709867283e-02, 9.80757657554250e-02, 9.80757207039673e-02, 9.80757171585935e-02, 9.80756525808744e-02, 9.80752150575555e-02, 9.80736066832861e-02, 9.80693194073581e-02, 9.80599371099373e-02, 9.80419700414572e-02, 9.80107287683865e-02, 9.79602442569600e-02, 9.78832401072935e-02, 9.77711621387414e-02, 9.76142694224649e-02, 9.74017893463071e-02, 9.71221372906186e-02, 9.67631989320523e-02, 9.63126700957346e-02, 9.57584455372285e-02, 9.50890442603622e-02, 9.42940552597378e-02, 9.33645842942710e-02, 9.22936798797451e-02, 9.10767155308250e-02, 8.97117057361214e-02, 8.81995353908636e-02, 8.65440864414566e-02, 8.47522510703484e-02, 8.28338274289280e-02, 8.08013010992568e-02, 7.86695225066225e-02, 7.64552968114174e-02, 7.41769078948167e-02, 7.18536015239977e-02, 6.95050543111222e-02, 6.71508544656163e-02, 6.48100176647191e-02, 6.25005570940758e-02, 6.02391216364353e-02, 5.80407111156447e-02, 5.59184727862814e-02, 5.38835788152667e-02, 5.19451801232051e-02, 5.01104278185878e-02, 4.83845503481604e-02, 4.67709731388837e-02, 4.52714680176456e-02, 4.38863213362168e-02, 4.26145114644640e-02, 4.14538875908264e-02, 4.04013428639871e-02, 3.94529763100095e-02, 3.86042398984543e-02, 3.78500693372317e-02, 3.71849989595532e-02, 3.66032622745749e-02, 3.60988802597433e-02, 3.56657396442017e-02, 3.52976634530933e-02, 3.49884759540018e-02, 3.47320639403179e-02, 3.45224358074887e-02, 3.43537793287823e-02, 3.42205182770599e-02, 3.41173674049244e-02, 3.40393846001135e-02, 3.39820185881194e-02, 3.39411501757994e-02, 3.39131248987883e-02, 3.38947750039367e-02, 3.38834289456500e-02, 3.38769069962410e-02, 3.38735022140000e-02, 3.38719466008328e-02, 3.38713630458471e-02, 3.38712043069833e-02, 3.38711808778598e-02, 3.38711790167494e-02, 3.38711449277156e-02, 3.38709139675615e-02, 3.38700649462851e-02, 3.38678018248544e-02, 3.38628493605952e-02, 3.38533660208782e-02, 3.38368782378922e-02, 3.38102398662912e-02, 3.37696205649962e-02, 3.37105265157263e-02, 3.36278563963717e-02, 3.35159946640502e-02, 3.33689430170452e-02, 3.31804892365497e-02, 3.29444104165830e-02, 3.26547051412902e-02, 3.23058463583763e-02, 3.18930439776466e-02, 3.14125038046811e-02, 3.08616677076117e-02, 3.02394192549187e-02, 2.95462396271233e-02, 2.87843005862811e-02, 2.79574844678478e-02, 2.70713254537895e-02, 2.61328715888354e-02, 2.51504728106511e-02, 2.41335066093313e-02, 2.30920585995804e-02, 2.20365794364170e-02, 2.09775404253918e-02, 1.99251076038536e-02, 1.88822933868522e-02, 1.78719642463676e-02, 1.68940633172312e-02, 1.59551050768769e-02, 1.50603723234113e-02, 1.42139083390234e-02, 1.34185594243945e-02, 1.26760579432244e-02, 1.19871346178044e-02, 1.13516487046255e-02, 1.07687258772415e-02, 1.02368956508059e-02, 9.75422224280464e-03, 9.31842439018737e-03, 8.92698090143540e-03, 8.57721998264793e-03, 8.26639183027567e-03, 7.99172541528748e-03, 7.75047139493523e-03, 7.53993346651782e-03, 7.35749036578426e-03, 7.20061043993070e-03, 7.06686054124361e-03, 6.95391092651931e-03, 6.85953779174042e-03, 6.78162490195433e-03, 6.71816546465465e-03, 6.66726499149917e-03, 6.62714548261789e-03, 6.59615090030741e-03, 6.57275360280352e-03, 6.55556118090317e-03, 6.54332297887805e-03, 6.53493548895057e-03, 6.52944578872194e-03, 6.52605224790648e-03, 6.52410186002328e-03, 6.52308374453980e-03, 6.52261859633541e-03, 6.52244410918712e-03, 6.52239664503185e-03, 6.52238963943574e-03, 6.52238907813540e-03, 6.52237872726857e-03, 6.52230859892455e-03, 6.52205080462664e-03, 6.52136365733469e-03, 6.51986004331206e-03, 6.51698117442123e-03, 6.51197708679518e-03, 6.50389529455708e-03, 6.49157903856690e-03, 6.47367654515563e-03, 6.44866257869126e-03, 6.41487327427370e-03, 6.37055472606818e-03, 6.31392505236761e-03, 6.24324866595316e-03, 6.15692029560222e-03, 6.05355502367644e-03, 5.93207935469309e-03, 5.79181726241274e-03, 5.63256444740290e-03, 5.45464385840688e-03, 5.25893608541850e-03, 5.04687969575457e-03, 4.82043900344564e-03, 4.58203991738770e-03, 4.33447784761398e-03, 4.08080439816416e-03, 3.82420115263926e-03, 3.56784924365172e-03, 3.31480316221189e-03, 3.06787702545819e-03, 2.82955119689632e-03, 2.60190579669484e-03, 2.38658436090589e-03, 2.18478618761253e-03, 1.99728188243142e-03, 1.82444550424776e-03, 1.66629859312906e-03, 1.52256370218851e-03, 1.39272499544774e-03, 1.27609088482927e-03, 1.17185159827926e-03, 1.07912612970296e-03, 9.97444440159256e-04, 9.25037267052176e-04, 8.61359977257337e-04, 8.05537348391283e-04, 7.56769408503999e-04, 7.14362331841301e-04, 6.77689114368718e-04, 6.46082018441518e-04, 6.18938843558681e-04, 5.95778477915587e-04, 5.76176178022081e-04, 5.59681565242443e-04, 5.45880945175270e-04, 5.34426727710979e-04, 5.25018329064153e-04, 5.17387518030432e-04, 5.11289805871229e-04, 5.06500010104105e-04, 5.02810849312141e-04, 5.00033128851385e-04, 4.97996273786327e-04, 4.96548657409335e-04, 4.95557668543256e-04, 4.94909569503932e-04, 4.94509143614152e-04, 4.94279076754367e-04, 4.94159001306116e-04, 4.94104147044538e-04, 4.94083570778514e-04, 4.94077973712592e-04, 4.94077147578764e-04, 4.94077080745114e-04, 4.94075840350661e-04, 4.94067436598300e-04, 4.94036544367304e-04, 4.93954204138447e-04, 4.93774040585038e-04, 4.93429144121649e-04, 4.92829800464278e-04, 4.91862274671462e-04, 4.90388882326442e-04, 4.88249614448932e-04, 4.85265617001573e-04, 4.81244830659477e-04, 4.75989959880962e-04, 4.69308423127757e-04, 4.61023025529528e-04, 4.50982008779022e-04, 4.39069317522467e-04, 4.25217032145042e-04, 4.09418591743815e-04, 3.91739917724010e-04, 3.72331580138052e-04, 3.51453140503111e-04, 3.29402349265706e-04, 3.06421800175372e-04, 2.82874134931888e-04, 2.59182298364635e-04, 2.35707755059877e-04, 2.12782366318720e-04, 1.90719976941978e-04, 1.69803893179221e-04, 1.50270827944610e-04, 1.32283081192952e-04, 1.15968543695999e-04, 1.01295695294823e-04, 8.81771846081847e-05, 7.64822807113319e-05, 6.60481072785046e-05, 5.67278884560096e-05, 4.89927340439067e-05, 4.21679504043482e-05, 3.61650733964325e-05, 3.08325604549964e-05, 2.59520248917511e-05, 2.11157484102387e-05, 1.64585514141891e-05, 1.20683259439147e-05, 7.98247895517079e-06, 4.46665475806185e-06, 1.72551239074892e-06, 1.45993917044507e-06, 1.10683154730395e-06, -3.55271367880050e-14, 9.90258719468784e-07, 1.17536970378751e-06, 1.10051869484096e-06, 9.32192151026356e-07, 7.81430344432010e-07, 6.95241858750251e-07, 6.63303104886381e-07, 6.48236564870786e-07, 6.19612180230433e-07, 5.66251500799808e-07, 4.91344962938456e-07, 4.04476885762506e-07, 3.15982774168333e-07, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14],
[ 5.90000000000000e+01, 5.89999999414825e+01, 5.89999981320072e+01, 5.89999858725251e+01, 5.89999408050692e+01, 5.89998206694502e+01, 5.89995577437741e+01, 5.89990541643860e+01, 5.89981782939612e+01, 5.89967622742123e+01, 5.89946008632187e+01, 5.89914516181399e+01, 5.89870364436980e+01, 5.89810444871008e+01, 5.89731363227796e+01, 5.89629493370189e+01, 5.89501041944069e+01, 5.89342122455847e+01, 5.89148837187632e+01, 5.88917365248547e+01, 5.88644054960326e+01, 5.88325518678820e+01, 5.87958728039146e+01, 5.87541107466344e+01, 5.87070623613821e+01, 5.86545868194092e+01, 5.85966131484124e+01, 5.85331463670184e+01, 5.84642721201398e+01, 5.83901595500994e+01, 5.83110621777593e+01, 5.82273166297986e+01, 5.81393391306522e+01, 5.80476197750626e+01, 5.79527147015966e+01, 5.78552363892771e+01, 5.77558423890962e+01, 5.76552228715404e+01, 5.75540874149512e+01, 5.74531514754348e+01, 5.73531229682686e+01, 5.72546893572523e+01, 5.71585055979561e+01, 5.70651832198639e+01, 5.69752807673139e+01, 5.68892957552260e+01, 5.68076582367928e+01, 5.67307260289217e+01, 5.66587815981427e+01, 5.65920305746775e+01, 5.65306018343533e+01, 5.64745490655967e+01, 5.64238537203190e+01, 5.63784292317223e+01, 5.63381263678301e+01, 5.63027395761659e+01, 5.62720141621596e+01, 5.62456541315960e+01, 5.62233305160987e+01, 5.62046899908734e+01, 5.61893635864807e+01, 5.61769752921280e+01, 5.61671503476960e+01, 5.61595230262083e+01, 5.61537437182812e+01, 5.61494851455966e+01, 5.61464475516675e+01, 5.61443627448263e+01, 5.61429968998592e+01, 5.61421520601204e+01, 5.61416663201228e+01, 5.61414127081624e+01, 5.61412968280443e+01, 5.61412533569716e+01, 5.61412415317442e+01, 5.61412397863639e+01, 5.61412395574069e+01, 5.61412342229814e+01, 5.61411980814452e+01, 5.61410652210552e+01, 5.61407110618838e+01, 5.61399359822847e+01, 5.61384515703124e+01, 5.61358700346357e+01, 5.61316972357229e+01, 5.61253297134652e+01, 5.61160559920233e+01, 5.61030623363742e+01, 5.60854430176789e+01, 5.60622150164411e+01, 5.60323369542551e+01, 5.59947318981648e+01, 5.59483135283048e+01, 5.58920150023827e+01, 5.58248196936552e+01, 5.57457928283700e+01, 5.56541129134718e+01, 5.55491017393819e+01, 5.54302516842217e+01, 5.52972490564372e+01, 5.51499923129905e+01, 5.49886041935885e+01, 5.48134371171279e+01, 5.46250715744320e+01, 5.44243076807514e+01, 5.42121504666195e+01, 5.39897898269863e+01, 5.37585762675747e+01, 5.35199936599844e+01, 5.32756301508587e+01, 5.30271482030912e+01, 5.27762545353984e+01, 5.25246705298581e+01, 5.22741035396946e+01, 5.20262194694897e+01, 5.17826170054267e+01, 5.15448039101148e+01, 5.13141758228999e+01, 5.10919979882718e+01, 5.08793902576561e+01, 5.06773155820651e+01, 5.04865720603785e+01, 5.03077884616937e+01, 5.01414230256321e+01, 4.99877652742039e+01, 4.98469405415519e+01, 4.97189169325443e+01, 4.96035144428121e+01, 4.95004159976093e+01, 4.94091801850887e+01, 4.93292554661908e+01, 4.92599956372920e+01, 4.92006763046617e+01, 4.91505121046796e+01, 4.91086743743133e+01, 4.90743089462971e+01, 4.90465537163881e+01, 4.90245556093938e+01, 4.90074865594001e+01, 4.89945581203919e+01, 4.89850343382435e+01, 4.89782425450886e+01, 4.89735817826457e+01, 4.89705286214141e+01, 4.89686402159467e+01, 4.89675545198309e+01, 4.89669876739657e+01, 4.89667286740631e+01, 4.89666315137168e+01, 4.89666050836957e+01, 4.89666011826781e+01, 4.89666008382925e+01, 4.89665940902261e+01, 4.89665483710743e+01, 4.89663803024944e+01, 4.89659322922803e+01, 4.89649518273708e+01, 4.89630740986418e+01, 4.89598086464717e+01, 4.89545306298991e+01, 4.89464772223694e+01, 4.89347495217734e+01, 4.89183202290024e+01, 4.88960471947200e+01, 4.88666927563649e+01, 4.88289485862874e+01, 4.87814655499553e+01, 4.87228878367268e+01, 4.86518903855062e+01, 4.85672183989027e+01, 4.84677275415413e+01, 4.83524232729212e+01, 4.82204976953723e+01, 4.80713623235035e+01, 4.79046753168501e+01, 4.77203619645904e+01, 4.75186275565615e+01, 4.72999621860893e+01, 4.70651374587959e+01, 4.68151954717412e+01, 4.65514307310575e+01, 4.62753658696494e+01, 4.59887221185939e+01, 4.56933855141410e+01, 4.53913698349059e+01, 4.50847772968496e+01, 4.47757580943156e+01, 4.44664699411752e+01, 4.41590387976570e+01, 4.38555219289297e+01, 4.35578743128886e+01, 4.32679192038512e+01, 4.29873233940957e+01, 4.27175774351964e+01, 4.24599808236927e+01, 4.22156319486586e+01, 4.19854224561049e+01, 4.17700356071869e+01, 4.15699481842359e+01, 4.13854355155628e+01, 4.12165792305610e+01, 4.10632774064475e+01, 4.09252568160037e+01, 4.08020870245959e+01, 4.06931961105412e+01, 4.05978877940335e+01, 4.05153597566698e+01, 4.04447229176469e+01, 4.03850214063707e+01, 4.03352529376157e+01, 4.02943892581665e+01, 4.02613962971430e+01, 4.02352536203162e+01, 4.02149727660089e+01, 4.01996140307083e+01, 4.01883012796687e+01, 4.01802343839367e+01, 4.01746989314433e+01, 4.01710729256448e+01, 4.01688302686653e+01, 4.01675409235651e+01, 4.01668677576112e+01, 4.01665601798748e+01, 4.01664447964180e+01, 4.01664134092797e+01, 4.01664087766023e+01, 4.01664084050043e+01, 4.01664015471357e+01, 4.01663550840609e+01, 4.01661842807873e+01, 4.01657289812361e+01, 4.01647325655785e+01, 4.01628242961124e+01, 4.01595057548082e+01, 4.01541419897830e+01, 4.01459578884018e+01, 4.01340401793392e+01, 4.01173453317213e+01, 4.00947134632280e+01, 4.00648881884339e+01, 4.00265421333399e+01, 3.99783076146681e+01, 3.99188117396846e+01, 3.98467149351265e+01, 3.97607516777114e+01, 3.96597719927948e+01, 3.95427821332529e+01, 3.94089827686510e+01, 3.92578030232733e+01, 3.90889288124941e+01, 3.89023241430006e+01, 3.86982443549153e+01, 3.84772406722062e+01, 3.82401558603767e+01, 3.79881112286418e+01, 3.77224856177025e+01, 3.74448873497262e+01, 3.71571203625251e+01, 3.68611458997058e+01, 3.65590411928299e+01, 3.62529565704596e+01, 3.59450723834562e+01, 3.56375570594108e+01, 3.53325274926605e+01, 3.50320128310946e+01, 3.47379225266537e+01, 3.44520192720735e+01, 3.41758971670259e+01, 3.39109651722586e+01, 3.36584356566801e+01, 3.34193176499253e+01, 3.31944142972761e+01, 3.29843239729402e+01, 3.27894445260409e+01, 3.26099801893260e+01, 3.24459507523532e+01, 3.22972026725321e+01, 3.21634218589529e+01, 3.20441479107082e+01, 3.19387896221531e+01, 3.18466415827378e+01, 3.17669016998594e+01, 3.16986894611185e+01, 3.16410647292427e+01, 3.15930468311339e+01, 3.15536336650752e+01, 3.15218205109379e+01, 3.14966181916687e+01, 3.14770702051476e+01, 3.14622684282802e+01, 3.14513669939637e+01, 3.14435939593235e+01, 3.14382604219704e+01, 3.14347667999109e+01, 3.14326060684015e+01, 3.14313638400849e+01, 3.14307152784949e+01, 3.14304189438732e+01, 3.14303077782874e+01, 3.14302775385301e+01, 3.14302730752045e+01, 3.14302727309499e+01, 3.14302665492418e+01, 3.14302246672591e+01, 3.14300707046414e+01, 3.14296602966877e+01, 3.14287621281950e+01, 3.14270420253076e+01, 3.14240507432718e+01, 3.14192160129439e+01, 3.14118393216919e+01, 3.14010978026077e+01, 3.13860514864524e+01, 3.13656560300479e+01, 3.13387808703605e+01, 3.13042325646337e+01, 3.12607828660341e+01, 3.12072008576095e+01, 3.11422882355735e+01, 3.10649166109008e+01, 3.09740655042901e+01, 3.08688595638703e+01, 3.07486034572065e+01, 3.06128128953968e+01, 3.04612403471247e+01, 3.02938941953366e+01, 3.01110503693978e+01, 2.99132558319165e+01, 2.97013236852824e+01, 2.94763200583278e+01, 2.92395433099312e+01, 2.89924964209670e+01, 2.87368537235087e+01, 2.84744233282202e+01, 2.82071067530138e+01, 2.79368573247619e+01, 2.76656389167643e+01, 2.73953864938693e+01, 2.71279697645539e+01, 2.68651609927281e+01, 2.66086077196120e+01, 2.63598108152302e+01, 2.61201079527043e+01, 2.58906623087582e+01, 2.56724560661648e+01, 2.54662881428521e+01, 2.52727755004279e+01, 2.50923573834415e+01, 2.49253018936996e+01, 2.47717143918430e+01, 2.46315473218750e+01, 2.45046111571447e+01, 2.43905862565150e+01, 2.42890354898714e+01, 2.41994175395740e+01, 2.41211008088732e+01, 2.40533778717235e+01, 2.39954803840892e+01, 2.39465943486292e+01, 2.39058755867406e+01, 2.38724652286456e+01, 2.38455049878713e+01, 2.38241519454628e+01, 2.38075925358082e+01, 2.37950554039784e+01, 2.37858227973414e+01, 2.37792401643888e+01, 2.37747236625476e+01, 2.37717653241943e+01, 2.37699356946192e+01, 2.37688838343715e+01, 2.37683346670975e+01, 2.37680837477076e+01, 2.37679896191288e+01, 2.37679640138681e+01, 2.37679602345851e+01, 2.37679599484292e+01, 2.37679548792060e+01, 2.37679205344767e+01, 2.37677942796750e+01, 2.37674577311626e+01, 2.37667212051779e+01, 2.37653106782455e+01, 2.37628577864715e+01, 2.37588933320561e+01, 2.37528446963767e+01, 2.37440374776396e+01, 2.37317015739279e+01, 2.37149818158692e+01, 2.36929531141941e+01, 2.36646399250565e+01, 2.36290396518439e+01, 2.35851494017789e+01, 2.35319953086409e+01, 2.34686634334341e+01, 2.33943310804466e+01, 2.33082972367129e+01, 2.32100107782166e+01, 2.30990951033044e+01, 2.29753679639382e+01, 2.28388554711443e+01, 2.26897995440132e+01, 2.25286584318769e+01, 2.23561003360918e+01, 2.21729905527051e+01, 2.19803729090052e+01, 2.17794465381255e+01, 2.15715391995414e+01, 2.13580783985493e+01, 2.11405614924755e+01, 2.09205258208069e+01, 2.06995196982292e+01, 2.04790749048781e+01, 2.02606811320117e+01, 2.00457627146911e+01, 1.98356579084871e+01, 1.96316009305207e+01, 1.94347069618048e+01, 1.92459602727356e+01, 1.90662055700246e+01, 1.88961425687735e+01, 1.87363236791853e+01, 1.85871545839081e+01, 1.84488973905593e+01, 1.83216759897462e+01, 1.82054832368949e+01, 1.81001896012702e+01, 1.80055529756791e+01, 1.79212294009055e+01, 1.78467845167233e+01, 1.77817055969018e+01, 1.77254140538856e+01, 1.76772783086753e+01, 1.76366269146061e+01, 1.76027618037569e+01, 1.75749714960336e+01, 1.75525440783063e+01, 1.75347797289773e+01, 1.75210025363917e+01, 1.75105713414148e+01, 1.75028893284633e+01, 1.74974120975593e+01, 1.74936539737557e+01, 1.74911923495277e+01, 1.74896699091419e+01, 1.74887946491121e+01, 1.74883376821801e+01, 1.74881288896855e+01, 1.74880505643268e+01, 1.74880292579221e+01, 1.74880261131411e+01, 1.74880258770614e+01, 1.74880217217909e+01, 1.74879935692231e+01, 1.74878900774177e+01, 1.74876142063536e+01, 1.74870104697446e+01, 1.74858542437035e+01, 1.74838435625219e+01, 1.74805937822373e+01, 1.74756354349397e+01, 1.74684155291938e+01, 1.74583024720910e+01, 1.74445946937642e+01, 1.74265329437201e+01, 1.74033160993893e+01, 1.73741201822250e+01, 1.73381201194774e+01, 1.72945136273729e+01, 1.72425464337721e+01, 1.71815379179732e+01, 1.71109061364592e+01, 1.70301911409294e+01, 1.69390754928160e+01, 1.68374009477946e+01, 1.67251804308416e+01, 1.66026046463312e+01, 1.64700429583048e+01, 1.63280385120949e+01, 1.61772979175366e+01, 1.60186761352612e+01, 1.58531574586063e+01, 1.56818336302656e+01, 1.55058801594845e+01, 1.53265318222606e+01, 1.51450581676387e+01, 1.49627396657135e+01, 1.47808449636359e+01, 1.46006095948805e+01, 1.44232164210455e+01, 1.42497780603208e+01, 1.40813215469418e+01, 1.39187754462918e+01, 1.37629596049555e+01, 1.36145776407989e+01, 1.34742121830239e+01, 1.33423227704684e+01, 1.32192462235867e+01, 1.31051992338083e+01, 1.30002828701857e+01, 1.29044886885362e+01, 1.28177061390159e+01, 1.27397309975355e+01, 1.26702745865914e+01, 1.26089735940553e+01, 1.25554003375548e+01, 1.25090733523698e+01, 1.24694681993641e+01, 1.24360283954103e+01, 1.24081763628069e+01, 1.23853242785225e+01, 1.23668846819369e+01, 1.23522806749733e+01, 1.23409555252330e+01, 1.23323814649606e+01, 1.23260674698310e+01, 1.23215658043160e+01, 1.23184771362914e+01, 1.23164540529348e+01, 1.23152028519010e+01, 1.23144835342001e+01, 1.23141079851440e+01, 1.23139363935894e+01, 1.23138720236588e+01, 1.23138545134756e+01, 1.23138519290092e+01, 1.23138517355825e+01, 1.23138483389122e+01, 1.23138253259737e+01, 1.23137407280181e+01, 1.23135152212424e+01, 1.23130217067711e+01, 1.23120765736208e+01, 1.23104329982581e+01, 1.23077765949344e+01, 1.23037236850764e+01, 1.22978224000661e+01, 1.22895567676652e+01, 1.22783538559673e+01, 1.22635939584277e+01, 1.22446236987109e+01, 1.22207718159317e+01, 1.21913672626225e+01, 1.21557591149833e+01, 1.21133376654686e+01, 1.20635559511203e+01, 1.20059508779267e+01, 1.19401630425916e+01, 1.18659543380231e+01, 1.17832224649391e+01, 1.16920115632634e+01, 1.15925183233985e+01, 1.14850931344337e+01, 1.13702360645718e+01, 1.12485877347379e+01, 1.11209154216530e+01, 1.09880949908536e+01, 1.08510894911139e+01, 1.07109254183683e+01, 1.05686677621462e+01, 1.04253949699884e+01, 1.02821749033531e+01, 1.01400427200447e+01, 9.99998142009491e+00, 9.86290555775280e+00, 9.72964837805100e+00, 9.60095240726999e+00, 9.47746333257550e+00, 9.35972686002070e+00, 9.24818814692281e+00, 9.14319336215726e+00, 9.04499292866906e+00, 8.95374603597733e+00, 8.86952606529503e+00, 8.79232663559638e+00, 8.72206804689323e+00, 8.65860396073595e+00, 8.60172821331030e+00, 8.55118170104345e+00, 8.50665931130467e+00, 8.46781689147844e+00, 8.43427825893655e+00, 8.40564225322922e+00, 8.38148982152216e+00, 8.36139111063551e+00, 8.34491251604460e+00, 8.33162361224159e+00, 8.32110386253692e+00, 8.31294898241840e+00, 8.30677681164181e+00, 8.30223253868495e+00, 8.29899311893619e+00, 8.29677073617905e+00, 8.29531517588945e+00, 8.29441500795378e+00, 8.29389751412635e+00, 8.29362733958356e+00, 8.29350389548762e+00, 8.29345758748850e+00, 8.29344499060345e+00, 8.29344313133009e+00, 8.29344299212622e+00, 8.29344054694047e+00, 8.29342398045550e+00, 8.29336308042595e+00, 8.29320074424129e+00, 8.29284548021456e+00, 8.29216512527649e+00, 8.29098203920107e+00, 8.28907001025849e+00, 8.28615307743350e+00, 8.28190643821017e+00, 8.27595956597579e+00, 8.26790160476737e+00, 8.25728903924735e+00, 8.24365555341257e+00, 8.22652389331151e+00, 8.20541943994608e+00, 8.17988508414464e+00, 8.14949688368017e+00, 8.11387988458217e+00, 8.07272341499420e+00, 8.02579512246631e+00, 7.97295303384178e+00, 7.91415497693341e+00, 7.84946481590792e+00, 7.77905511255236e+00, 7.70320602202901e+00, 7.62230044773472e+00, 7.53681569558733e+00, 7.44731206307026e+00, 7.35441895479320e+00, 7.25881922151463e+00, 7.16123246780509e+00, 7.06239806641571e+00, 6.96305856277714e+00, 6.86394406274670e+00, 6.76575808395116e+00, 6.66916522801667e+00, 6.57478090728355e+00, 6.48316324199916e+00, 6.39480713685334e+00, 6.31014045209579e+00, 6.22952210690316e+00, 6.15324189354181e+00, 6.08152174217037e+00, 6.01451815874987e+00, 5.95232556171027e+00, 5.89498026407514e+00, 5.84246488233273e+00, 5.79471299613986e+00, 5.75161392849090e+00, 5.71301755943215e+00, 5.67873912402362e+00, 5.64856397462920e+00, 5.62225230758695e+00, 5.59954386474551e+00, 5.58016262189349e+00, 5.56382146994079e+00, 5.55022688236652e+00, 5.53908354570352e+00, 5.53009891061181e+00, 5.52298760141252e+00, 5.51747560383569e+00, 5.51330413609926e+00, 5.51023309898771e+00, 5.50804399767367e+00, 5.50654223246320e+00, 5.50555866768278e+00, 5.50495040714611e+00, 5.50460072998949e+00, 5.50441817055442e+00, 5.50433475845218e+00, 5.50430346782420e+00, 5.50429495602710e+00, 5.50429369970412e+00, 5.50429360539364e+00, 5.50429194545018e+00, 5.50428069909610e+00, 5.50423935641379e+00, 5.50412915280848e+00, 5.50388797900743e+00, 5.50342611801271e+00, 5.50262298695316e+00, 5.50132504542538e+00, 5.49934501253141e+00, 5.49646251098932e+00, 5.49242622655846e+00, 5.48695763261907e+00, 5.47975628142606e+00, 5.47050660446230e+00, 5.45888609485840e+00, 5.44457466710895e+00, 5.42726490730201e+00, 5.40667284671202e+00, 5.38254882042266e+00, 5.35468791914712e+00, 5.32293951510043e+00, 5.28721534895710e+00, 5.24749570943263e+00, 5.20383332105835e+00, 5.15635467601854e+00, 5.10525869446702e+00, 5.05081276216448e+00, 4.99334635900792e+00, 4.93324264018391e+00, 4.87092844752032e+00, 4.80686330007958e+00, 4.74152793398236e+00, 4.67541293284511e+00, 4.60900791997626e+00, 4.54279168474916e+00, 4.47722350390335e+00, 4.41273580867949e+00, 4.34972825202102e+00, 4.28856315313368e+00, 4.22956224128084e+00, 4.17300458499527e+00, 4.11912557313613e+00, 4.06811680622612e+00, 4.02012675668858e+00, 3.97526206240926e+00, 3.93358932785826e+00, 3.89513731982748e+00, 3.85989945990961e+00, 3.82783653239242e+00, 3.79887954334501e+00, 3.77293268328172e+00, 3.74987636081494e+00, 3.72957028714315e+00, 3.71185660025608e+00, 3.69656302284181e+00, 3.68350604884488e+00, 3.67249415059581e+00, 3.66333099187968e+00, 3.65581862302073e+00, 3.64976062305757e+00, 3.64496514258452e+00, 3.64124779013942e+00, 3.63843429641777e+00, 3.63636288524641e+00, 3.63488627907942e+00, 3.63387327036587e+00, 3.63320979865512e+00, 3.63279948648635e+00, 3.63256360426057e+00, 3.63244045437420e+00, 3.63238418660631e+00, 3.63236307869394e+00, 3.63235733683658e+00, 3.63235648935071e+00, 3.63235642546015e+00, 3.63235529732309e+00, 3.63234765403270e+00, 3.63231955655029e+00, 3.63224465945150e+00, 3.63208075138448e+00, 3.63176685703090e+00, 3.63122102091604e+00, 3.63033888196392e+00, 3.62899313416739e+00, 3.62703395225146e+00, 3.62429044058074e+00, 3.62057313757246e+00, 3.61567757550001e+00, 3.60938885631334e+00, 3.60148715828900e+00, 3.59175403736616e+00, 3.57997933357006e+00, 3.56596844082542e+00, 3.54954965255080e+00, 3.53058126105858e+00, 3.50895807117731e+00, 3.48461699196371e+00, 3.45754139748804e+00, 3.42776399875368e+00, 3.39536804151530e+00, 3.36048673419422e+00, 3.32330090926995e+00, 3.28403502219090e+00, 3.24295168543478e+00, 3.20034501401789e+00, 3.15653311622866e+00, 3.11185009556704e+00, 3.06663793525379e+00, 3.02123861624447e+00, 2.97598677680880e+00, 2.93120316167690e+00, 2.88718903803007e+00, 2.84422168122720e+00, 2.80255096178694e+00, 2.76239700240972e+00, 2.72394882374341e+00, 2.68736386231235e+00, 2.65276822383394e+00, 2.62025752875351e+00, 2.58989821184477e+00, 2.56172915115642e+00, 2.53576352034268e+00, 2.51199077966641e+00, 2.49037874237444e+00, 2.47087567297232e+00, 2.45341239099650e+00, 2.43790436751608e+00, 2.42425381149035e+00, 2.41235174924190e+00, 2.40208010284602e+00, 2.39331377249974e+00, 2.38592272434620e+00, 2.37977407932508e+00, 2.37473419103689e+00, 2.37067069207954e+00, 2.36745447965082e+00, 2.36496160325764e+00, 2.36307501097054e+00, 2.36168610656165e+00, 2.36069606867400e+00, 2.36001688527144e+00, 2.35957206212864e+00, 2.35929697285854e+00, 2.35913882947172e+00, 2.35905626600400e+00, 2.35901854244637e+00, 2.35900439109783e+00, 2.35900054159360e+00, 2.35899997341505e+00, 2.35899993034790e+00, 2.35899916680501e+00, 2.35899399369301e+00, 2.35897497683759e+00, 2.35892428527858e+00, 2.35881335004859e+00, 2.35860090350226e+00, 2.35823148123974e+00, 2.35763446125189e+00, 2.35672370611051e+00, 2.35539786315104e+00, 2.35354136376339e+00, 2.35102614517611e+00, 2.34771409570144e+00, 2.34346019689288e+00, 2.33811630365746e+00, 2.33153546707504e+00, 2.32357666645082e+00, 2.31410977983696e+00, 2.30302058953512e+00, 2.29021559499252e+00, 2.27562639400073e+00, 2.25921339748064e+00, 2.24096866533857e+00, 2.22091769099692e+00, 2.19912001815222e+00, 2.17566864085024e+00, 2.15068821107810e+00, 2.12433214971708e+00, 2.09677881979827e+00, 2.06822696949457e+00, 2.03889068204449e+00, 2.00899407924086e+00, 1.97876601528387e+00, 1.94843497203679e+00, 1.91822432990941e+00, 1.88834814616147e+00, 1.85900752940460e+00, 1.83038765939500e+00, 1.80265546724689e+00, 1.77595796386285e+00, 1.75042118340013e+00, 1.72614969304419e+00, 1.70322660915925e+00, 1.68171405218485e+00, 1.66165396799714e+00, 1.64306924176764e+00, 1.62596503172186e+00, 1.61033025466862e+00, 1.59613916254435e+00, 1.58335295897707e+00, 1.57192141618924e+00, 1.56178446439151e+00, 1.55287373707734e+00, 1.54511406531650e+00, 1.53842492149028e+00, 1.53272181743647e+00, 1.52791766352366e+00, 1.52392409390659e+00, 1.52065275954892e+00, 1.51801658516396e+00, 1.51593097978780e+00, 1.51431498409693e+00, 1.51309233164353e+00, 1.51219239665287e+00, 1.51155099851218e+00, 1.51111103297561e+00, 1.51082290258027e+00, 1.51064472372051e+00, 1.51054229493855e+00, 1.51048881973134e+00, 1.51046438688948e+00, 1.51045522135525e+00, 1.51045272811415e+00, 1.51045236011723e+00, 1.51045233204658e+00, 1.51045183204341e+00, 1.51044844445197e+00, 1.51043599136099e+00, 1.51040279640416e+00, 1.51033015214121e+00, 1.51019103768951e+00, 1.50994914153647e+00, 1.50955824030581e+00, 1.50896197924823e+00, 1.50809409381727e+00, 1.50687910315714e+00, 1.50523349491690e+00, 1.50306740587826e+00, 1.50028678408859e+00, 1.49679599561459e+00, 1.49250081333307e+00, 1.48731169765548e+00, 1.48114725166805e+00, 1.47393770830044e+00, 1.46562828754512e+00, 1.45618225015400e+00, 1.44558347298276e+00, 1.43383838186571e+00, 1.42097710121813e+00, 1.40705371497619e+00, 1.39214557935131e+00, 1.37635168153977e+00, 1.35979009651758e+00, 1.34259465229368e+00, 1.32491096798838e+00, 1.30689207409262e+00, 1.28869385549339e+00, 1.27047057102265e+00, 1.25237069529792e+00, 1.23453329847007e+00, 1.21708512910995e+00, 1.20013850007943e+00, 1.18379000490178e+00, 1.16812002241908e+00, 1.15319290948660e+00, 1.13905774179455e+00, 1.12574944461101e+00, 1.11329015733231e+00, 1.10169069403914e+00, 1.09095199084438e+00, 1.08106646350703e+00, 1.07201923039143e+00, 1.06378918282844e+00, 1.05634990550529e+00, 1.04967046335587e+00, 1.04371607922283e+00, 1.03844872951184e+00, 1.03382768447214e+00, 1.02981001681544e+00, 1.02635109808215e+00, 1.02340509716752e+00, 1.02092549019188e+00, 1.01886558572379e+00, 1.01717906442496e+00, 1.01582052761280e+00, 1.01474604515307e+00, 1.01391368963377e+00, 1.01328404108953e+00, 1.01282064479838e+00, 1.01249040401247e+00, 1.01226389001175e+00, 1.01211555362581e+00, 1.01202382529836e+00, 1.01197109472178e+00, 1.01194356579854e+00, 1.01193098787777e+00, 1.01192626950939e+00, 1.01192498600294e+00, 1.01192479656022e+00, 1.01192478200763e+00, 1.01192452145565e+00, 1.01192275617991e+00, 1.01191626687401e+00, 1.01189896903174e+00, 1.01186111447019e+00, 1.01178862360187e+00, 1.01166257746937e+00, 1.01145889659406e+00, 1.01114823126708e+00, 1.01069608464253e+00, 1.01006318538634e+00, 1.00920612015467e+00, 1.00807822735359e+00, 1.00663074219701e+00, 1.00481416909724e+00, 1.00257984142860e+00, 9.99881611787401e-01, 9.96677599679778e-01, 9.92931910200426e-01, 9.88616229026235e-01, 9.83711198082943e-01, 9.78207484086518e-01, 9.72106469288967e-01, 9.65420519255545e-01, 9.58172813955230e-01, 9.50396762169589e-01, 9.42135050806186e-01, 9.33438405820020e-01, 9.24364156769194e-01, 9.14974700945216e-01, 9.05335955982451e-01, 8.95515874235549e-01, 8.85583071653251e-01, 8.75605602364459e-01, 8.65649891153936e-01, 8.55779821665152e-01, 8.46055969322244e-01, 8.36534964153586e-01, 8.27268968656789e-01, 8.18305258002546e-01, 8.09685892809753e-01, 8.01447477422222e-01, 7.93620998551710e-01, 7.86231740219428e-01, 7.79299271292864e-01, 7.72837501855989e-01, 7.66854804432882e-01, 7.61354195898740e-01, 7.56333575846881e-01, 7.51786017244882e-01, 7.47700105350575e-01, 7.44060320982371e-01, 7.40847464249278e-01, 7.38039114663557e-01, 7.35610123126555e-01, 7.33533130576376e-01, 7.31779107134052e-01, 7.30317904432361e-01, 7.29118812544385e-01, 7.28151111652622e-01, 7.27384607440335e-01, 7.26790138277600e-01, 7.26340041746013e-01, 7.26008568011324e-01, 7.25772228100467e-01, 7.25610066314339e-01, 7.25503847810700e-01, 7.25438154770551e-01, 7.25400387417082e-01, 7.25380669345000e-01, 7.25371659969895e-01, 7.25368280238193e-01, 7.25367360868844e-01, 7.25367225171837e-01, 7.25367214668324e-01, 7.25367025576091e-01, 7.25365744448446e-01, 7.25361034889474e-01, 7.25348480947709e-01, 7.25321007020810e-01, 7.25268391422972e-01, 7.25176893265903e-01, 7.25029010086139e-01, 7.24803381060077e-01, 7.24474847593996e-01, 7.24014679426674e-01, 7.23390970004750e-01, 7.22569199675810e-01, 7.21512959153621e-01, 7.20184818819419e-01, 7.18547321955526e-01, 7.16564072354956e-01, 7.14200879471079e-01, 7.11426918048112e-01, 7.08215854747650e-01, 7.04546892351459e-01, 7.00405683211045e-01, 6.95785067989384e-01, 6.90685603317966e-01, 6.85115852314723e-01, 6.79092424191396e-01, 6.72639762402049e-01, 6.65789693830575e-01, 6.58580763309836e-01, 6.51057387407974e-01, 6.43268868262197e-01, 6.35268311952338e-01, 6.27111496462803e-01, 6.18855731959982e-01, 6.10558751431071e-01, 6.02277663385721e-01, 5.94067991078312e-01, 5.85982815313940e-01, 5.78072030988963e-01, 5.70381721554547e-01, 5.62953650849664e-01, 5.55824868307617e-01, 5.49027421313468e-01, 5.42588167282716e-01, 5.36528677595179e-01, 5.30865225601048e-01, 5.25608851306146e-01, 5.20765495878059e-01, 5.16336199678037e-01, 5.12317358036000e-01, 5.08701029393663e-01, 5.05475290699934e-01, 5.02624635027622e-01, 5.00130406263992e-01, 4.97971265402036e-01, 4.96123682422642e-01, 4.94562447030056e-01, 4.93261190628679e-01, 4.92192910970985e-01, 4.91330489956212e-01, 4.90647194220266e-01, 4.90117147541781e-01, 4.89715763805897e-01, 4.89420129412032e-01, 4.89209324650680e-01, 4.89064674743901e-01, 4.88969922935645e-01, 4.88911320183483e-01, 4.88877628553972e-01, 4.88860038235693e-01, 4.88852001016582e-01, 4.88848985968929e-01, 4.88848165801990e-01, 4.88848044747037e-01, 4.88848035300933e-01, 4.88847864263242e-01, 4.88846705457533e-01, 4.88842445563179e-01, 4.88831090233468e-01, 4.88806239293964e-01, 4.88758646472377e-01, 4.88675881218676e-01, 4.88542108046090e-01, 4.88337996352698e-01, 4.88040770816148e-01, 4.87624409108726e-01, 4.87059989773179e-01, 4.86316188570171e-01, 4.85359916444629e-01, 4.84157086532825e-01, 4.82673491518582e-01, 4.80875766439027e-01, 4.78732406144459e-01, 4.76214801530546e-01, 4.73298254923868e-01, 4.69962933129473e-01, 4.66194717053668e-01, 4.61985909728597e-01, 4.57335769975145e-01, 4.52250846567189e-01, 4.46745097066618e-01, 4.40839785754903e-01, 4.34563165462322e-01, 4.27949957781621e-01, 4.21040654467774e-01, 4.13880669308064e-01, 4.06519374188790e-01, 3.99009055507065e-01, 3.91403827668924e-01, 3.83758539446291e-01, 3.76127706714037e-01, 3.68564501800783e-01, 3.61119825569595e-01, 3.53841483573554e-01, 3.46773482375795e-01, 3.39955456569648e-01, 3.33422231411724e-01, 3.27203520567430e-01, 3.21323753577609e-01, 3.15802023577419e-01, 3.10652142759601e-01, 3.05882791181240e-01, 3.01497743762518e-01, 2.97496160603323e-01, 2.93872926862015e-01, 2.90619030151156e-01, 2.87721965432851e-01, 2.85166159461184e-01, 2.82933408675056e-01, 2.81003325896990e-01, 2.79353792124816e-01, 2.77961410066027e-01, 2.76801955886100e-01, 2.75850825005655e-01, 2.75083466822181e-01, 2.74475802115269e-01, 2.74004615804245e-01, 2.73647916844503e-01, 2.73385256534354e-01, 2.73197996482807e-01, 2.73069518035626e-01, 2.72985366093465e-01, 2.72933321946503e-01, 2.72903401904742e-01, 2.72887780997287e-01, 2.72880643687259e-01, 2.72877966237743e-01, 2.72877237906719e-01, 2.72877130406592e-01, 2.72877121946820e-01, 2.72876967853755e-01, 2.72875923850982e-01, 2.72872085992041e-01, 2.72861855712904e-01, 2.72839467219157e-01, 2.72796591425930e-01, 2.72722032725056e-01, 2.72601533199811e-01, 2.72417696968802e-01, 2.72150045024873e-01, 2.71775208113873e-01, 2.71267261745727e-01, 2.70598203242194e-01, 2.69738565760193e-01, 2.68658158531402e-01, 2.67326916285832e-01, 2.65715834258124e-01, 2.63797958707315e-01, 2.61549397007876e-01, 2.58950306640521e-01, 2.55985819380726e-01, 2.52646856145425e-01, 2.48930789687762e-01, 2.44841916821365e-01, 2.40391709088406e-01, 2.35598820503171e-01, 2.30488842714237e-01, 2.25093810939171e-01, 2.19451477474848e-01, 2.13604382506688e-01, 2.07598763335319e-01, 2.01483352067577e-01, 1.95308117502030e-01, 1.89123008853770e-01, 1.82976756921818e-01, 1.76915782492031e-01, 1.70983252741067e-01, 1.65218315031233e-01, 1.59655524849704e-01, 1.54324471931247e-01, 1.49249596929558e-01, 1.44450181284405e-01, 1.39940485763013e-01, 1.35730008779383e-01, 1.31823833943052e-01, 1.28223037045295e-01, 1.24925125389261e-01, 1.21924486447973e-01, 1.19212827705873e-01, 1.16779594652876e-01, 1.14612358792144e-01, 1.12697171858699e-01, 1.11018886015479e-01, 1.09561442489372e-01, 1.08308132903140e-01, 1.07241838468909e-01, 1.06345252308920e-01, 1.05601089561723e-01, 1.04992288765615e-01, 1.04502206449048e-01, 1.04114805084954e-01, 1.03814832766979e-01, 1.03587991318193e-01, 1.03421088209124e-01, 1.03302166771016e-01, 1.03220608834178e-01, 1.03167204138671e-01, 1.03134181643519e-01, 1.03115199132944e-01, 1.03105289174017e-01, 1.03100761374030e-01, 1.03099062862931e-01, 1.03098600828893e-01, 1.03098532633716e-01, 1.03098527219544e-01, 1.03098427998013e-01, 1.03097755758789e-01, 1.03095284556233e-01, 1.03088697390689e-01, 1.03074282317110e-01, 1.03046678704679e-01, 1.02998685110549e-01, 1.02921139727130e-01, 1.02802883710076e-01, 1.02630816715582e-01, 1.02390052443425e-01, 1.02064179659870e-01, 1.01635630850332e-01, 1.01086156198335e-01, 1.00397395001963e-01, 9.95515300853285e-02, 9.85320036027844e-02, 9.73242654148265e-02, 9.59165186624205e-02, 9.43004221003036e-02, 9.24717060202696e-02, 9.04306589478168e-02, 8.81824462605465e-02, 8.57372296379140e-02, 8.31100675483967e-02, 8.03205910691176e-02, 7.73924650222355e-02, 7.43526601980378e-02, 7.12305766909012e-02, 6.80570695671037e-02, 6.48634349602162e-02, 6.16804164221207e-02, 5.85372876959696e-02, 5.54610594629210e-02, 5.24758452114327e-02, 4.96024068860805e-02, 4.68578862310807e-02, 4.42557142935556e-02, 4.18056804186762e-02, 3.95141338558034e-02, 3.73842861842064e-02, 3.54165813361931e-02, 3.36091017974027e-02, 3.19579837599875e-02, 3.04578193789453e-02, 2.91020297764817e-02, 2.78831975452292e-02, 2.67933521994266e-02, 2.58242063499736e-02, 2.49673441205829e-02, 2.42143660975742e-02, 2.35569967012665e-02, 2.29871604359317e-02, 2.24970334236332e-02, 2.20790763215390e-02, 2.17260543178668e-02, 2.14310493585330e-02, 2.11874689974110e-02, 2.09890552817029e-02, 2.08298959625566e-02, 2.07044391750060e-02, 2.06075116601667e-02, 2.05343396675701e-02, 2.04805709158364e-02, 2.04422954333054e-02, 2.04160627743448e-02, 2.03988930284282e-02, 2.03882792136696e-02, 2.03821790520067e-02, 2.03789947190529e-02, 2.03775398854589e-02, 2.03769941456713e-02, 2.03768456929652e-02, 2.03768237817172e-02, 2.03768220261935e-02, 2.03767896531772e-02, 2.03765703222913e-02, 2.03757640537958e-02, 2.03736149540674e-02, 2.03689122905359e-02, 2.03599083827655e-02, 2.03442575678991e-02, 2.03189806275859e-02, 2.02804591536179e-02, 2.02244642550298e-02, 2.01462236056713e-02, 2.00405299261064e-02, 1.99018924297079e-02, 1.97247304435066e-02, 1.95036053379098e-02, 1.92334831976808e-02, 1.89100166192944e-02, 1.85298300504328e-02, 1.80907897293991e-02, 1.75922371243473e-02, 1.70351643914870e-02, 1.64223122594862e-02, 1.57581751965665e-02, 1.50489056811196e-02, 1.43021183144825e-02, 1.35266042264561e-02, 1.27319751265347e-02, 1.19282628868618e-02, 1.11255038274081e-02, 1.03333370708540e-02, 9.56064428474690e-03, 8.81525439614705e-03, 8.10373105087336e-03, 7.43125189225680e-03, 6.80157765687994e-03, 6.21709858635455e-03, 5.67893986198698e-03, 5.18710847538273e-03, 4.74066821754349e-03, 4.33793171414010e-03, 3.97665598287489e-03, 3.65422450796871e-03, 3.36780095287281e-03, 3.11444942810368e-03, 2.89122762633554e-03, 2.69526330313852e-03, 2.52381897573173e-03, 2.37434086330524e-03, 2.24448507712083e-03, 2.13211970236671e-03, 2.03531020920877e-03, 1.95229987973278e-03, 1.88149390870507e-03, 1.82144905813431e-03, 1.77086562960227e-03, 1.72857772029289e-03, 1.69354013185341e-03, 1.66481313466704e-03, 1.64154761274915e-03, 1.62297271734246e-03, 1.60838693761889e-03, 1.59715239177416e-03, 1.58869157645825e-03, 1.58248572350317e-03, 1.57807405215493e-03, 1.57505336920138e-03, 1.57307757860536e-03, 1.57185672367888e-03, 1.57115523235518e-03, 1.57078910071107e-03, 1.57062183728840e-03, 1.57055909507164e-03, 1.57054202806250e-03, 1.57053950901798e-03, 1.57053930542617e-03, 1.57053552902653e-03, 1.57050994356922e-03, 1.57041589155682e-03, 1.57016520542186e-03, 1.56961669652844e-03, 1.56856666435345e-03, 1.56674199096329e-03, 1.56379644470661e-03, 1.55931088237793e-03, 1.55279809689546e-03, 1.54371308394019e-03, 1.53146945801552e-03, 1.51546256734747e-03, 1.49509943002357e-03, 1.46983480454566e-03, 1.43921142542780e-03, 1.40290078383032e-03, 1.36073929151337e-03, 1.31275414013032e-03, 1.25917468552217e-03, 1.20042911387941e-03, 1.13713119366565e-03, 1.07006477295357e-03, 1.00017043996559e-03, 9.28528578590715e-04, 8.56321944688432e-04, 7.84760848254962e-04, 7.14973787694362e-04, 6.47898714497330e-04, 5.84228780738272e-04, 5.24443469863911e-04, 4.68893540663373e-04, 4.17855489928343e-04, 3.71491020963788e-04, 3.29743701662721e-04, 2.92293603443738e-04, 2.58664876739090e-04, 2.28434701170954e-04, 2.01361632555574e-04, 1.77290748418102e-04, 1.55896558645452e-04, 1.36510140926305e-04, 1.18245186015997e-04, 1.00386622953863e-04, 8.27836626413614e-05, 6.60232098255165e-05, 5.13970281659226e-05, 4.07027971149887e-05, 3.45739647187315e-05, 3.05906349851170e-05, 2.69509272285973e-05, 2.36098491868120e-05, 2.07188949232773e-05, 1.81472560214502e-05, 1.57551695139091e-05, 1.35305735753377e-05, 1.15077451212642e-05, 9.70317025661416e-06, 8.11385990573399e-06, 6.72618673291936e-06, 5.51684751570747e-06, 4.45857854458609e-06, 3.53067537872676e-06, 2.72341511475815e-06, 2.03441648949365e-06, 1.46266269496209e-06, 1.00445274853200e-06, 6.52019558344819e-07, 3.93913673235602e-07, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00],
[ 6.00000000000000e+01, 5.99999999399730e+01, 5.99999980838187e+01, 5.99999855080793e+01, 5.99999392780198e+01, 5.99998160432627e+01, 5.99995463348753e+01, 5.99990297645439e+01, 5.99981312988813e+01, 5.99966787488111e+01, 5.99944615765575e+01, 5.99912310825064e+01, 5.99867019927927e+01, 5.99805554277683e+01, 5.99724431933128e+01, 5.99619933029072e+01, 5.99488166096632e+01, 5.99325144046440e+01, 5.99126868205389e+01, 5.98889418669291e+01, 5.98609049131136e+01, 5.98282284244391e+01, 5.97906017460647e+01, 5.97477607126017e+01, 5.96994968429389e+01, 5.96456658584794e+01, 5.95861952435712e+01, 5.95210905543156e+01, 5.94504401821819e+01, 5.93744182975757e+01, 5.92932857396574e+01, 5.92073886834652e+01, 5.91171550014552e+01, 5.90230883381630e+01, 5.89257600252416e+01, 5.88257990695851e+01, 5.87238805396044e+01, 5.86207127456365e+01, 5.85170236544256e+01, 5.84135469925737e+01, 5.83110084812091e+01, 5.82101126081219e+01, 5.81115302903706e+01, 5.80158877167350e+01, 5.79237565919419e+01, 5.78356459388061e+01, 5.77519955542643e+01, 5.76731711630347e+01, 5.75994612691168e+01, 5.75310756701638e+01, 5.74681455717714e+01, 5.74107252163607e+01, 5.73587949229907e+01, 5.73122654186711e+01, 5.72709833274523e+01, 5.72347376700382e+01, 5.72032672135819e+01, 5.71762684987735e+01, 5.71534043596895e+01, 5.71343127417832e+01, 5.71186156156659e+01, 5.71059277798220e+01, 5.70958653450063e+01, 5.70880536975484e+01, 5.70821347487412e+01, 5.70777732932699e+01, 5.70746623212834e+01, 5.70725271559465e+01, 5.70711283205151e+01, 5.70702630752120e+01, 5.70697656032572e+01, 5.70695058659304e+01, 5.70693871870298e+01, 5.70693426660287e+01, 5.70693305551946e+01, 5.70693287676593e+01, 5.70693285331724e+01, 5.70693230699082e+01, 5.70692860554706e+01, 5.70691499862028e+01, 5.70687872733449e+01, 5.70679934742287e+01, 5.70664732122060e+01, 5.70638293331913e+01, 5.70595557714066e+01, 5.70530345105101e+01, 5.70435369288108e+01, 5.70302297084707e+01, 5.70121853682712e+01, 5.69883973478461e+01, 5.69577994289968e+01, 5.69192891282294e+01, 5.68717545361223e+01, 5.68141039165455e+01, 5.67452972164250e+01, 5.66643784811643e+01, 5.65705080319585e+01, 5.64629931535153e+01, 5.63413159833470e+01, 5.62051573089585e+01, 5.60544150877871e+01, 5.58892167196761e+01, 5.57099244209580e+01, 5.55171334504710e+01, 5.53116633777128e+01, 5.50945430035902e+01, 5.48669898839120e+01, 5.46303856166033e+01, 5.43862481136414e+01, 5.41362019993034e+01, 5.38819480994735e+01, 5.36252327728772e+01, 5.33678176450135e+01, 5.31114501825303e+01, 5.28578355037345e+01, 5.26086098426114e+01, 5.23653161296952e+01, 5.21293821779501e+01, 5.19021019314609e+01, 5.16846201383350e+01, 5.14779206609465e+01, 5.12828184676364e+01, 5.10999551949895e+01, 5.09297980547657e+01, 5.07726417949651e+01, 5.06286134058404e+01, 5.04976792747971e+01, 5.03796545220641e+01, 5.02742142772369e+01, 5.01809066757946e+01, 5.00991673602194e+01, 5.00283352621952e+01, 4.99676694228428e+01, 4.99163665805615e+01, 4.98735792246756e+01, 4.98384337815466e+01, 4.98100485717290e+01, 4.97875511554570e+01, 4.97700946722717e+01, 4.97568727815117e+01, 4.97471328256761e+01, 4.97401868695559e+01, 4.97354203147676e+01, 4.97322978511486e+01, 4.97303665815048e+01, 4.97292562415710e+01, 4.97286765290714e+01, 4.97284116502196e+01, 4.97283122844638e+01, 4.97282852545167e+01, 4.97282812649513e+01, 4.97282809127486e+01, 4.97282740115102e+01, 4.97282272545949e+01, 4.97280553710832e+01, 4.97275971916348e+01, 4.97265944714675e+01, 4.97246741207312e+01, 4.97213345468831e+01, 4.97159367255306e+01, 4.97077005140894e+01, 4.96957066046480e+01, 4.96789043758358e+01, 4.96561257453215e+01, 4.96261049426482e+01, 4.95875039161540e+01, 4.95389428604213e+01, 4.94790351086374e+01, 4.94064253885905e+01, 4.93198302075836e+01, 4.92180789302900e+01, 4.91001539672620e+01, 4.89652284236961e+01, 4.88126995885380e+01, 4.86422167861388e+01, 4.84537023671816e+01, 4.82473649663566e+01, 4.80237045666083e+01, 4.77835093339769e+01, 4.75278445686927e+01, 4.72580344136069e+01, 4.69756371522232e+01, 4.66824150295553e+01, 4.63802995784815e+01, 4.60713534764181e+01, 4.57577300199373e+01, 4.54416313870235e+01, 4.51252669294222e+01, 4.48108127602716e+01, 4.45003738414491e+01, 4.41959496188746e+01, 4.38994040160435e+01, 4.36124403096474e+01, 4.33365811179202e+01, 4.30731534703780e+01, 4.28232787232045e+01, 4.25878669489606e+01, 4.23676153601968e+01, 4.21630103120209e+01, 4.19743324522010e+01, 4.18016646321388e+01, 4.16449022437018e+01, 4.15037656948994e+01, 4.13778147750589e+01, 4.12664646840434e+01, 4.11690035091319e+01, 4.10846109280196e+01, 4.10123778986930e+01, 4.09513270691254e+01, 4.09004336049265e+01, 4.08586460948904e+01, 4.08249071568913e+01, 4.07981733341504e+01, 4.07774338489175e+01, 4.07617277711822e+01, 4.07501591675833e+01, 4.07419098226331e+01, 4.07362491717931e+01, 4.07325411533801e+01, 4.07302477717206e+01, 4.07289292638868e+01, 4.07282408720446e+01, 4.07279263374026e+01, 4.07278083441587e+01, 4.07277762470939e+01, 4.07277715096329e+01, 4.07277711296299e+01, 4.07277641166474e+01, 4.07277166026526e+01, 4.07275419360823e+01, 4.07270763383729e+01, 4.07260573853085e+01, 4.07241059534766e+01, 4.07207123507790e+01, 4.07152272620788e+01, 4.07068580391955e+01, 4.06946707464239e+01, 4.06775982354509e+01, 4.06544543637213e+01, 4.06239542855941e+01, 4.05847405356600e+01, 4.05354143911229e+01, 4.04745717519832e+01, 4.04008425252574e+01, 4.03129322585085e+01, 4.02096645581347e+01, 4.00900226709108e+01, 3.99531885246679e+01, 3.97985775339984e+01, 3.96258675912287e+01, 3.94350208838783e+01, 3.92262974979583e+01, 3.90002601600552e+01, 3.87577699078170e+01, 3.84999729192361e+01, 3.82282791366728e+01, 3.79443336593642e+01, 3.76499821287169e+01, 3.73472314905555e+01, 3.70382075984622e+01, 3.67251111407578e+01, 3.64101733471705e+01, 3.60956128669051e+01, 3.57835951028293e+01, 3.54761951252425e+01, 3.51753650653252e+01, 3.48829066092113e+01, 3.46004489033792e+01, 3.43294318788371e+01, 3.40710947449716e+01, 3.38264692222281e+01, 3.35963769854661e+01, 3.33814307679106e+01, 3.31820386082169e+01, 3.29984107860696e+01, 3.28305690641861e+01, 3.26783579224716e+01, 3.25414575258845e+01, 3.24193982086053e+01, 3.23115762832197e+01, 3.22172709956970e+01, 3.21356624459884e+01, 3.20658502811939e+01, 3.20068729448546e+01, 3.19577272340556e+01, 3.19173878785600e+01, 3.18848268168864e+01, 3.18590318075529e+01, 3.18390239844325e+01, 3.18238739480302e+01, 3.18127159836187e+01, 3.18047600156484e+01, 3.17993009473111e+01, 3.17957250945553e+01, 3.17935135034118e+01, 3.17922420347496e+01, 3.17915782067431e+01, 3.17912748966962e+01, 3.17911611143758e+01, 3.17911301628021e+01, 3.17911255944139e+01, 3.17911252420558e+01, 3.17911189148359e+01, 3.17910760469888e+01, 3.17909184602251e+01, 3.17904983915930e+01, 3.17895790807631e+01, 3.17878184870851e+01, 3.17847567890791e+01, 3.17798082420644e+01, 3.17722578802154e+01, 3.17612634450957e+01, 3.17458628003422e+01, 3.17249869482018e+01, 3.16974785954672e+01, 3.16621160234113e+01, 3.16176418010422e+01, 3.15627956497905e+01, 3.14963505314598e+01, 3.14171508050495e+01, 3.13241511005301e+01, 3.12164544091987e+01, 3.10933478106151e+01, 3.09543342615386e+01, 3.07991589724195e+01, 3.06278290926367e+01, 3.04406257074672e+01, 3.02381074988712e+01, 3.00211058126876e+01, 2.97907112778276e+01, 2.95482525110028e+01, 2.92952677907737e+01, 2.90334708809191e+01, 2.87647124142558e+01, 2.84909384059401e+01, 2.82141475424518e+01, 2.79363488817621e+01, 2.76595214970772e+01, 2.73855774029075e+01, 2.71163288299673e+01, 2.68534605877054e+01, 2.65985079023061e+01, 2.63528397798099e+01, 2.61176476518211e+01, 2.58939388395483e+01, 2.56825342327641e+01, 2.54840695227869e+01, 2.52989993411833e+01, 2.51276037201275e+01, 2.49699963853674e+01, 2.48261344989881e+01, 2.46958295708717e+01, 2.45787593440737e+01, 2.44744805242032e+01, 2.43824422642374e+01, 2.43020003347897e+01, 2.42324319082546e+01, 2.41729508669674e+01, 2.41227235144808e+01, 2.40808845293496e+01, 2.40465529565898e+01, 2.40188479874279e+01, 2.39969042372981e+01, 2.39798861994067e+01, 2.39670015303919e+01, 2.39575128189781e+01, 2.39507475004856e+01, 2.39461056108982e+01, 2.39430651238355e+01, 2.39411846806381e+01, 2.39401036048714e+01, 2.39395391836709e+01, 2.39392812944549e+01, 2.39391845512245e+01, 2.39391582347124e+01, 2.39391543504501e+01, 2.39391540563455e+01, 2.39391488463115e+01, 2.39391135475674e+01, 2.39389837856908e+01, 2.39386378884829e+01, 2.39378809024909e+01, 2.39364311898361e+01, 2.39339101463724e+01, 2.39298355208549e+01, 2.39236187432552e+01, 2.39145665861489e+01, 2.39018873828483e+01, 2.38847019060578e+01, 2.38620588684512e+01, 2.38329548406642e+01, 2.37963581945815e+01, 2.37512364761004e+01, 2.36965864017161e+01, 2.36314654716056e+01, 2.35550240163277e+01, 2.34665363649088e+01, 2.33654297589100e+01, 2.32513096570514e+01, 2.31239801889463e+01, 2.29834587262737e+01, 2.28299838359358e+01, 2.26640162411989e+01, 2.24862328116762e+01, 2.22975139922700e+01, 2.20989254240471e+01, 2.18916947706771e+01, 2.16771849183764e+01, 2.14568647580165e+01, 2.12322786966585e+01, 2.10050159102479e+01, 2.07766801774540e+01, 2.05488609651507e+01, 2.03231062980106e+01, 2.01008978508658e+01, 1.98836286469536e+01, 1.96725837073358e+01, 1.94689239503376e+01, 1.92736735635607e+01, 1.90877109572644e+01, 1.89117632656455e+01, 1.87464042136711e+01, 1.85920550380396e+01, 1.84489880631879e+01, 1.83173324973692e+01, 1.81970820271548e+01, 1.80881038387327e+01, 1.79901487637875e+01, 1.79028623197056e+01, 1.78257964754502e+01, 1.77584220180791e+01, 1.77001414178426e+01, 1.76503020930843e+01, 1.76082099630233e+01, 1.75731431512843e+01, 1.75443656704953e+01, 1.75211408831017e+01, 1.75027445002274e+01, 1.74884768530709e+01, 1.74776741536179e+01, 1.74697184563782e+01, 1.74640460425709e+01, 1.74601539737821e+01, 1.74576046035064e+01, 1.74560278907899e+01, 1.74551214278730e+01, 1.74546481697325e+01, 1.74544319335384e+01, 1.74543508157730e+01, 1.74543287497642e+01, 1.74543254928672e+01, 1.74543252483709e+01, 1.74543209449590e+01, 1.74542917887103e+01, 1.74541846072592e+01, 1.74538989008876e+01, 1.74532736396102e+01, 1.74520761899554e+01, 1.74499938164964e+01, 1.74466281514919e+01, 1.74414929658613e+01, 1.74340155043708e+01, 1.74235415648952e+01, 1.74093444033573e+01, 1.73906374304281e+01, 1.73665905326250e+01, 1.73363497005305e+01, 1.72990594846617e+01, 1.72538876323773e+01, 1.72000510975958e+01, 1.71368424722557e+01, 1.70636557794794e+01, 1.69800105089179e+01, 1.68855727790612e+01, 1.67801725902181e+01, 1.66638162902661e+01, 1.65366936096096e+01, 1.63991789180533e+01, 1.62518266892708e+01, 1.60953614932866e+01, 1.59306631339840e+01, 1.57587477701370e+01, 1.55807459802534e+01, 1.53978787491292e+01, 1.52114322853228e+01, 1.50227324586553e+01, 1.48331195156431e+01, 1.46439236207934e+01, 1.44564416978539e+01, 1.42719160029631e+01, 1.40915148331598e+01, 1.39163157369452e+01, 1.37472915320323e+01, 1.35852993434134e+01, 1.34310727580844e+01, 1.32852170649321e+01, 1.31482074261357e+01, 1.30203897249684e+01, 1.29019837642049e+01, 1.27930884533720e+01, 1.26936886200221e+01, 1.26036631039639e+01, 1.25227938354207e+01, 1.24507756492256e+01, 1.23872266389166e+01, 1.23316989001112e+01, 1.22836895470088e+01, 1.22426519066492e+01, 1.22080068020223e+01, 1.21791538284094e+01, 1.21554825099598e+01, 1.21363831990395e+01, 1.21212575534944e+01, 1.21095284010816e+01, 1.21006487801605e+01, 1.20941099350006e+01, 1.20894480455550e+01, 1.20862494869320e+01, 1.20841544434440e+01, 1.20828587450724e+01, 1.20821138482991e+01, 1.20817249453810e+01, 1.20815472525007e+01, 1.20814805937789e+01, 1.20814624609915e+01, 1.20814597846301e+01, 1.20814595843257e+01, 1.20814560668812e+01, 1.20814322356809e+01, 1.20813446297259e+01, 1.20811111048640e+01, 1.20806000436964e+01, 1.20796213091654e+01, 1.20779193089882e+01, 1.20751684978413e+01, 1.20709715972634e+01, 1.20648607579247e+01, 1.20563018228894e+01, 1.20447017714037e+01, 1.20294193287272e+01, 1.20097786179318e+01, 1.19850856055630e+01, 1.19546469579689e+01, 1.19177907848150e+01, 1.18738886090597e+01, 1.18223777786947e+01, 1.17627834362673e+01, 1.16947390991845e+01, 1.16180048875418e+01, 1.15324824748822e+01, 1.14382259354905e+01, 1.13354478197119e+01, 1.12245200013842e+01, 1.11059690984120e+01, 1.09804665532070e+01, 1.08488137540642e+01, 1.07119228582775e+01, 1.05707942186096e+01, 1.04264914940133e+01, 1.02801156255208e+01, 1.01327788690460e+01, 9.98557999846185e+00, 9.83958163514792e+00, 9.69579044381281e+00, 9.55514068447953e+00, 9.41848135481997e+00, 9.28656692104430e+00, 9.16005143908742e+00, 9.03948572289160e+00, 8.92531712697467e+00, 8.81789147288521e+00, 8.71745665516877e+00, 8.62416750095989e+00, 8.53809151658069e+00, 8.45921522393085e+00, 8.38745086036675e+00, 8.32264328187642e+00, 8.26457696652892e+00, 8.21298306107326e+00, 8.16754644718033e+00, 8.12791282508332e+00, 8.09369582183667e+00, 8.06448413008999e+00, 8.03984867255706e+00, 8.01934976899717e+00, 8.00254425862185e+00, 7.98899250381666e+00, 7.97826517360966e+00, 7.96994968022629e+00, 7.96365612209371e+00, 7.95902257426486e+00, 7.95571956437563e+00, 7.95345358015638e+00, 7.95196947352630e+00, 7.95105165578153e+00, 7.95052401678743e+00, 7.95024854605256e+00, 7.95012268219671e+00, 7.95007546648056e+00, 7.95006262267618e+00, 7.95006072695811e+00, 7.95006058502563e+00, 7.95005809191041e+00, 7.95004120069680e+00, 7.94997910694223e+00, 7.94981358881540e+00, 7.94945136164478e+00, 7.94875767321091e+00, 7.94755140557453e+00, 7.94560192383126e+00, 7.94262788218441e+00, 7.93829816117352e+00, 7.93223506422332e+00, 7.92401984417302e+00, 7.91320055885773e+00, 7.89930216802704e+00, 7.88183868259998e+00, 7.86032706455519e+00, 7.83430245745162e+00, 7.80333421218703e+00, 7.76704207089752e+00, 7.72511179592266e+00, 7.67730949239508e+00, 7.62349388223444e+00, 7.56362585034675e+00, 7.49777470150938e+00, 7.42612073326402e+00, 7.34895393421356e+00, 7.26666884107906e+00, 7.17975581104087e+00, 7.08878916712133e+00, 6.99441283518685e+00, 6.89732419849665e+00, 6.79825694371551e+00, 6.69796366257009e+00, 6.59719891392383e+00, 6.49670335425050e+00, 6.39718942413101e+00, 6.29932894745469e+00, 6.20374286901048e+00, 6.11099323270360e+00, 6.02157739177647e+00, 5.93592434749610e+00, 5.85439303623040e+00, 5.77727232856041e+00, 5.70478246926863e+00, 5.63707767379016e+00, 5.57424960355736e+00, 5.51633146649681e+00, 5.46330252528089e+00, 5.41509283957525e+00, 5.37158811424997e+00, 5.33263456877776e+00, 5.29804378034870e+00, 5.26759748233265e+00, 5.24105231952724e+00, 5.21814457200872e+00, 5.19859486095476e+00, 5.18211284364959e+00, 5.16840189249498e+00, 5.15716373596894e+00, 5.14810302001753e+00, 5.14093172835477e+00, 5.13537338163929e+00, 5.13116692045397e+00, 5.12807016719265e+00, 5.12586275872644e+00, 5.12434844596335e+00, 5.12335666837621e+00, 5.12274333085045e+00, 5.12239073572763e+00, 5.12220665308286e+00, 5.12212254506642e+00, 5.12209099338553e+00, 5.12208241057652e+00, 5.12208114377238e+00, 5.12208104867509e+00, 5.12207937488317e+00, 5.12206803470401e+00, 5.12202634711622e+00, 5.12191522416103e+00, 5.12167203859551e+00, 5.12120632563722e+00, 5.12039649850303e+00, 5.11908774111102e+00, 5.11709122193587e+00, 5.11418474912482e+00, 5.11011495668173e+00, 5.10460107279127e+00, 5.09734027248468e+00, 5.08801455695563e+00, 5.07629903133353e+00, 5.06187137374195e+00, 5.04442220510375e+00, 5.02366598746301e+00, 4.99935200627863e+00, 4.97127493787990e+00, 4.93928447571002e+00, 4.90329349550254e+00, 4.86328428496467e+00, 4.81931244896253e+00, 4.77150822315806e+00, 4.72007507928475e+00, 4.66528567119771e+00, 4.60747533683362e+00, 4.54703352046015e+00, 4.48439359652309e+00, 4.42002164920548e+00, 4.35440478451036e+00, 4.28803952520462e+00, 4.22142077093236e+00, 4.15503170885842e+00, 4.08933494929747e+00, 4.02476505023002e+00, 3.96172249559175e+00, 3.90056911151012e+00, 3.84162484456036e+00, 3.78516578549898e+00, 3.73142329771233e+00, 3.68058409812922e+00, 3.63279113630325e+00, 3.58814512227133e+00, 3.54670656389744e+00, 3.50849818847347e+00, 3.47350764033456e+00, 3.44169036510416e+00, 3.41297261075903e+00, 3.38725449476330e+00, 3.36441310383488e+00, 3.34430560737529e+00, 3.32677237633420e+00, 3.31164010572299e+00, 3.29872494090887e+00, 3.28783560535502e+00, 3.27877652111271e+00, 3.27135090393558e+00, 3.26536380345181e+00, 3.26062504667368e+00, 3.25695203161600e+00, 3.25417230829929e+00, 3.25212587816057e+00, 3.25066714086717e+00, 3.24966642035913e+00, 3.24901100983815e+00, 3.24860568812379e+00, 3.24837267661893e+00, 3.24825102600234e+00, 3.24819544337617e+00, 3.24817459250239e+00, 3.24816892056749e+00, 3.24816808340208e+00, 3.24816802028955e+00, 3.24816690589069e+00, 3.24815935567935e+00, 3.24813160037636e+00, 3.24805761548690e+00, 3.24789570419693e+00, 3.24758563555542e+00, 3.24704645753481e+00, 3.24617509366094e+00, 3.24484581897966e+00, 3.24291069522537e+00, 3.24020102492288e+00, 3.23652985847390e+00, 3.23169555595254e+00, 3.22548636580379e+00, 3.21768593620804e+00, 3.20807962297317e+00, 3.19646140315332e+00, 3.18264115013772e+00, 3.16645197869840e+00, 3.14775733297110e+00, 3.12645747198603e+00, 3.10249501064283e+00, 3.07585920168329e+00, 3.04658869661790e+00, 3.01477259834762e+00, 2.98054971036456e+00, 2.94410598970144e+00, 2.90567031466563e+00, 2.86550877493812e+00, 2.82391777267320e+00, 2.78121628226511e+00, 2.73773764920571e+00, 2.69382131333263e+00, 2.64980481981782e+00, 2.60601643595354e+00, 2.56276862858793e+00, 2.52035258265333e+00, 2.47903386283468e+00, 2.43904924495091e+00, 2.40060467694889e+00, 2.36387427583240e+00, 2.32900022878682e+00, 2.29609344474536e+00, 2.26523479555797e+00, 2.23647679143582e+00, 2.20984555039568e+00, 2.18534294274042e+00, 2.16294881607784e+00, 2.14262323134209e+00, 2.12430866367596e+00, 2.10793214240875e+00, 2.09340732084104e+00, 2.08063647871033e+00, 2.06951246800094e+00, 2.05992061638029e+00, 2.05174060237833e+00, 2.04484831299306e+00, 2.03911768833100e+00, 2.03442254989029e+00, 2.03063839996118e+00, 2.02764417020659e+00, 2.02532388867720e+00, 2.02356822717138e+00, 2.02227588575074e+00, 2.02135476899700e+00, 2.02072290965550e+00, 2.02030909979796e+00, 2.02005319739370e+00, 2.01990608675511e+00, 2.01982928401609e+00, 2.01979419272909e+00, 2.01978102885977e+00, 2.01977744797606e+00, 2.01977691944542e+00, 2.01977687938351e+00, 2.01977616912124e+00, 2.01977135699415e+00, 2.01975366717128e+00, 2.01970651310031e+00, 2.01960332011641e+00, 2.01940570296703e+00, 2.01906207562594e+00, 2.01850676447465e+00, 2.01765968837809e+00, 2.01642665961755e+00, 2.01470034708586e+00, 2.01236192655671e+00, 2.00928342132556e+00, 2.00533070959184e+00, 2.00036714276521e+00, 1.99425768244008e+00, 1.98687342508407e+00, 1.97809634550510e+00, 1.96782405669207e+00, 1.95597435892737e+00, 1.94248933934379e+00, 1.92733878784269e+00, 1.91052271862750e+00, 1.89207282872047e+00, 1.87205278361082e+00, 1.85055729122293e+00, 1.82771000236031e+00, 1.80366035116685e+00, 1.77857951528515e+00, 1.75265572557085e+00, 1.72608918469324e+00, 1.69908686062023e+00, 1.67185740565597e+00, 1.64460641781663e+00, 1.61753221429491e+00, 1.59082223295354e+00, 1.56465012352533e+00, 1.53917354083755e+00, 1.51453261171543e+00, 1.49084901723193e+00, 1.46822561295093e+00, 1.44674650072063e+00, 1.42647746455911e+00, 1.40746668805534e+00, 1.38974567938953e+00, 1.37333034079083e+00, 1.35822213067523e+00, 1.34440927793981e+00, 1.33186801837168e+00, 1.32056383252966e+00, 1.31045267256832e+00, 1.30148217215198e+00, 1.29359283872568e+00, 1.28671923086920e+00, 1.28079112519301e+00, 1.27573467724044e+00, 1.27147357922240e+00, 1.26793021432208e+00, 1.26502680306202e+00, 1.26268653223076e+00, 1.26083465160137e+00, 1.25939951867353e+00, 1.25831356747371e+00, 1.25751417455241e+00, 1.25694439412827e+00, 1.25655353511371e+00, 1.25629755561436e+00, 1.25613925533357e+00, 1.25604825286272e+00, 1.25600074266877e+00, 1.25597903516491e+00, 1.25597089197869e+00, 1.25596867683912e+00, 1.25596834988933e+00, 1.25596832494974e+00, 1.25596788071793e+00, 1.25596487098470e+00, 1.25595380692643e+00, 1.25592431451211e+00, 1.25585977262155e+00, 1.25573617306880e+00, 1.25552125083682e+00, 1.25517392884276e+00, 1.25464411677966e+00, 1.25387289755739e+00, 1.25279312699584e+00, 1.25133046207405e+00, 1.24940481967391e+00, 1.24693225108980e+00, 1.24382719771786e+00, 1.24000507093342e+00, 1.23538507545783e+00, 1.22989317235320e+00, 1.22346505748470e+00, 1.21604901642755e+00, 1.20760850983471e+00, 1.19812434619615e+00, 1.18759631278114e+00, 1.17604416023794e+00, 1.16350787037831e+00, 1.15004717741684e+00, 1.13574035679405e+00, 1.12068233871737e+00, 1.10498224192263e+00, 1.08876045379468e+00, 1.07214540383065e+00, 1.05527018755403e+00, 1.03826919749424e+00, 1.02127490760981e+00, 1.00441493894606e+00, 9.87809509100650e-01, 9.71569338193568e-01, 9.55794051725128e-01, 9.40571088395849e-01, 9.25975091186132e-01, 9.12067735102212e-01, 8.98897926817774e-01, 8.86502300975444e-01, 8.74905935169799e-01, 8.64123209674631e-01, 8.54158747175276e-01, 8.45008380152144e-01, 8.36660107196494e-01, 8.29095012800074e-01, 8.22288136881390e-01, 8.16209289818488e-01, 8.10823815792382e-01, 8.06093311851527e-01, 8.01976312517304e-01, 7.98428950283210e-01, 7.95405601360173e-01, 7.92859523822045e-01, 7.90743492225978e-01, 7.89010429122804e-01, 7.87614029931298e-01, 7.86509373727021e-01, 7.85653508892182e-01, 7.85005999585579e-01, 7.84529416889349e-01, 7.84189757496783e-01, 7.83956773069981e-01, 7.83804194970486e-01, 7.83709841894876e-01, 7.83655601860453e-01, 7.83627284722819e-01, 7.83614346628141e-01, 7.83609493140887e-01, 7.83608172878331e-01, 7.83607978010473e-01, 7.83607963041142e-01, 7.83607695027612e-01, 7.83605879198660e-01, 7.83599204050532e-01, 7.83581410796103e-01, 7.83542471910188e-01, 7.83467903945407e-01, 7.83338244209531e-01, 7.83128718742688e-01, 7.82809126414080e-01, 7.82343960534632e-01, 7.81692784715219e-01, 7.80810873322537e-01, 7.79650118391110e-01, 7.78160194007391e-01, 7.76289956039790e-01, 7.73989040078263e-01, 7.71209604445460e-01, 7.67908149472973e-01, 7.64047330601421e-01, 7.59597673170148e-01, 7.54539092884187e-01, 7.48862129400131e-01, 7.42568812105581e-01, 7.35673096866272e-01, 7.28200839050949e-01, 7.20189299184362e-01, 7.11686209919520e-01, 7.02748463023084e-01, 6.93440499249739e-01, 6.83832499625581e-01, 6.73998482323917e-01, 6.64014405081382e-01, 6.53956360500743e-01, 6.43898933214928e-01, 6.33913766805783e-01, 6.24068367482394e-01, 6.14425153066939e-01, 6.05040741141160e-01, 5.95965459695591e-01, 5.87243057020650e-01, 5.78910584200237e-01, 5.70998422603694e-01, 5.63530429482661e-01, 5.56524176588120e-01, 5.49991259229223e-01, 5.43937656130176e-01, 5.38364123619814e-01, 5.33266610959608e-01, 5.28636686839640e-01, 5.24461970110941e-01, 5.20726560538328e-01, 5.17411467629842e-01, 5.14495037327514e-01, 5.11953377461019e-01, 5.09760783343339e-01, 5.07890164732551e-01, 5.06313474647506e-01, 5.05002139292159e-01, 5.03927486735662e-01, 5.03061170163214e-01, 5.02375579629835e-01, 5.01844234500389e-01, 5.01442147329399e-01, 5.01146148992419e-01, 5.00935164561703e-01, 5.00790429813307e-01, 5.00695639387757e-01, 5.00637019468925e-01, 5.00603320294161e-01, 5.00585726710739e-01, 5.00577688153690e-01, 5.00574672629043e-01, 5.00573852334689e-01, 5.00573731261021e-01, 5.00573721889410e-01, 5.00573553174654e-01, 5.00572410108269e-01, 5.00568208089852e-01, 5.00557007163039e-01, 5.00532494774404e-01, 5.00485552785463e-01, 5.00403927055558e-01, 5.00272016382631e-01, 5.00070795755690e-01, 4.99777886563599e-01, 4.99367783392827e-01, 4.98812243091280e-01, 4.98080836672926e-01, 4.97141658278032e-01, 4.95962177830023e-01, 4.94510215490912e-01, 4.92755006993896e-01, 4.90668320139150e-01, 4.88225575076342e-01, 4.85406915460878e-01, 4.82198175114426e-01, 4.78591686204624e-01, 4.74586880592355e-01, 4.70190645818874e-01, 4.65417410645713e-01, 4.60288951054110e-01, 4.54833924738800e-01, 4.49087158789368e-01, 4.43088729879030e-01, 4.36882887554468e-01, 4.30516878220963e-01, 4.24039729720448e-01, 4.17501054095538e-01, 4.10949919776975e-01, 4.04433834953473e-01, 3.97997872437251e-01, 3.91683954161572e-01, 3.85530301699539e-01, 3.79571048843225e-01, 3.73836004011785e-01, 3.68350544440943e-01, 3.63135620801202e-01, 3.58207849895045e-01, 3.53579673990907e-01, 3.49259567649781e-01, 3.45252276051937e-01, 3.41559072343948e-01, 3.38178025003621e-01, 3.35104269373744e-01, 3.32330280165117e-01, 3.29846143782191e-01, 3.27639830760759e-01, 3.25697469449878e-01, 3.24003622377909e-01, 3.22541566585620e-01, 3.21293578667465e-01, 3.20241224415945e-01, 3.19365651895524e-01, 3.18647885570719e-01, 3.18069117865463e-01, 3.17610993337333e-01, 3.17255879607124e-01, 3.16987118391332e-01, 3.16789249527218e-01, 3.16648200830805e-01, 3.16551437029297e-01, 3.16488061872015e-01, 3.16448868821560e-01, 3.16426337394134e-01, 3.16414574157029e-01, 3.16409199482180e-01, 3.16407183263422e-01, 3.16406634803734e-01, 3.16406553852246e-01, 3.16406547535472e-01, 3.16406433159713e-01, 3.16405658247675e-01, 3.16402809597122e-01, 3.16395216202359e-01, 3.16378598557044e-01, 3.16346774979024e-01, 3.16291437266308e-01, 3.16202006891913e-01, 3.16065581633154e-01, 3.15866980900843e-01, 3.15588895943726e-01, 3.15212148402225e-01, 3.14716057246279e-01, 3.14078909885652e-01, 3.13278528220081e-01, 3.12292914769671e-01, 3.11100958111382e-01, 3.09683171129706e-01, 3.08022430667215e-01, 3.06104683709890e-01, 3.03919583903387e-01, 3.01461023481885e-01, 2.98727529857544e-01, 2.95722503096762e-01, 2.92454279840399e-01, 2.88936020102184e-01, 2.85185424717440e-01, 2.81224301815335e-01, 2.77078009410943e-01, 2.72774807176539e-01, 2.68345153163288e-01, 2.63820980689545e-01, 2.59234987215372e-01, 2.54619961543320e-01, 2.50008169044257e-01, 2.45430807724176e-01, 2.40917541606727e-01, 2.36496112671823e-01, 2.32192028764651e-01, 2.28028322499334e-01, 2.24025375018670e-01, 2.20200798181878e-01, 2.16569368965622e-01, 2.13143010291162e-01, 2.09930812979600e-01, 2.06939094023234e-01, 2.04171486829082e-01, 2.01629059522489e-01, 1.99310457789259e-01, 1.97212069085602e-01, 1.95328205376342e-01, 1.93651301880877e-01, 1.92172129597665e-01, 1.90880019600328e-01, 1.89763097199801e-01, 1.88808524011087e-01, 1.88002745729015e-01, 1.87331743023819e-01, 1.86781282438929e-01, 1.86337163571920e-01, 1.85985458210560e-01, 1.85712736553022e-01, 1.85506275252209e-01, 1.85354241850547e-01, 1.85245850276360e-01, 1.85171482490318e-01, 1.85122772102808e-01, 1.85092646815875e-01, 1.85075327823604e-01, 1.85066285765153e-01, 1.85062154377732e-01, 1.85060604552262e-01, 1.85060182962225e-01, 1.85060120736402e-01, 1.85060115839534e-01, 1.85060026643662e-01, 1.85059422328841e-01, 1.85057200806710e-01, 1.85051279066947e-01, 1.85038319599293e-01, 1.85013501108877e-01, 1.84970342924345e-01, 1.84900591525206e-01, 1.84794176475386e-01, 1.84639241683243e-01, 1.84422256241952e-01, 1.84128207054833e-01, 1.83740872959319e-01, 1.83243177096855e-01, 1.82617610874154e-01, 1.81846719124955e-01, 1.80913632214981e-01, 1.79802627106646e-01, 1.78499696167179e-01, 1.76993100129081e-01, 1.75273880452224e-01, 1.73336306652324e-01, 1.71178236096587e-01, 1.68801367281418e-01, 1.66211372498509e-01, 1.63417901705969e-01, 1.60434455875809e-01, 1.57278134576600e-01, 1.53969268565824e-01, 1.50530953297107e-01, 1.46988503243136e-01, 1.43368849713497e-01, 1.39699906464131e-01, 1.36009927920355e-01, 1.32326884237351e-01, 1.28677875611196e-01, 1.25088605163654e-01, 1.21582925595263e-01, 1.18182470067907e-01, 1.14906373012801e-01, 1.11771082076075e-01, 1.08790258245570e-01, 1.05974757406223e-01, 1.03332683215164e-01, 1.00869498842542e-01, 9.85881840572489e-02, 9.64894245269861e-02, 9.45718215658431e-02, 9.28321124386500e-02, 9.12653932962542e-02, 8.98653387199673e-02, 8.86244137858370e-02, 8.75340763477510e-02, 8.65849689120779e-02, 8.57671006454765e-02, 8.50700208480788e-02, 8.44829853458933e-02, 8.39951171344708e-02, 8.35955619670230e-02, 8.32736389237638e-02, 8.30189850556451e-02, 8.28216923534484e-02, 8.26724344628005e-02, 8.25625799408414e-02, 8.24842884661976e-02, 8.24305863544907e-02, 8.23954179836690e-02, 8.23736702867421e-02, 8.23611684031178e-02, 8.23546415510847e-02, 8.23516594362559e-02, 8.23505407509870e-02, 8.23502364414068e-02, 8.23501915268352e-02, 8.23501879608983e-02, 8.23501226111745e-02, 8.23496798549854e-02, 8.23480522460045e-02, 8.23437137065923e-02, 8.23342192699670e-02, 8.23160376257448e-02, 8.22844237887423e-02, 8.22333386798348e-02, 8.21554219081051e-02, 8.20420231759087e-02, 8.18832967304886e-02, 8.16683617691751e-02, 8.13855296511150e-02, 8.10225962745932e-02, 8.05671947397215e-02, 8.00071998560412e-02, 7.93311721006162e-02, 7.85288246950131e-02, 7.75914938626769e-02, 7.65125894106333e-02, 7.52880010427930e-02, 7.39164355598234e-02, 7.23996617799259e-02, 7.07426436275114e-02, 6.89535474768519e-02, 6.70436170020707e-02, 6.50269169271311e-02, 6.29199552869082e-02, 6.07412017592708e-02, 5.85105264668897e-02, 5.62485893260956e-02, 5.39762139485944e-02, 5.17137811231692e-02, 4.94806746074516e-02, 4.72948052715552e-02, 4.51722312887171e-02, 4.31125566087331e-02, 4.11581398893333e-02, 3.93015364092975e-02, 3.75497257779784e-02, 3.59074715944772e-02, 3.43774939840884e-02, 3.29606847735491e-02, 3.16563471648834e-02, 3.04624437621683e-02, 2.93758395994459e-02, 2.83925299993371e-02, 2.75078463041964e-02, 2.67166354975643e-02, 2.60134122956472e-02, 2.53924843450193e-02, 2.48480526893715e-02, 2.43742906908793e-02, 2.39654051639491e-02, 2.36156836717925e-02, 2.33195318172790e-02, 2.30715039995388e-02, 2.28663305645753e-02, 2.26989436081482e-02, 2.25645029429060e-02, 2.24584229673042e-02, 2.23764004211446e-02, 2.23144423283372e-02, 2.22688928580723e-02, 2.22364574201208e-02, 2.22142220784889e-02, 2.21996663350623e-02, 2.21906675004178e-02, 2.21854952117653e-02, 2.21827951438911e-02, 2.21815615365841e-02, 2.21810987802016e-02, 2.21809729004008e-02, 2.21809543208380e-02, 2.21809528322652e-02, 2.21809253822229e-02, 2.21807394047673e-02, 2.21800557422256e-02, 2.21782334274394e-02, 2.21742457347958e-02, 2.21666103539349e-02, 2.21533371422353e-02, 2.21318969329900e-02, 2.20992148638004e-02, 2.20516915557312e-02, 2.19852551084259e-02, 2.18954461625032e-02, 2.17775371106015e-02, 2.16266848452022e-02, 2.14381142068589e-02, 2.12073266218837e-02, 2.09303254747510e-02, 2.06038468331897e-02, 2.02255816063470e-02, 1.97943734898374e-02, 1.93103765478497e-02, 1.87751573315964e-02, 1.81917292029681e-02, 1.75645109719227e-02, 1.68992077503489e-02, 1.62026185167150e-02, 1.54823815172875e-02, 1.47466744635736e-02, 1.40038907254194e-02, 1.32623147610360e-02, 1.25298195639232e-02, 1.18136060088361e-02, 1.11199990685620e-02, 1.04543096698322e-02, 9.82076433400891e-03, 9.22249858895441e-03, 8.66160517828263e-03, 8.13922480841001e-03, 7.65566572825094e-03, 7.21053869519750e-03, 6.80289548085966e-03, 6.43136151780510e-03, 6.09425608008962e-03, 5.78969609391322e-03, 5.51568198875430e-03, 5.27016577039063e-03, 5.05110269709519e-03, 4.85648863023869e-03, 4.68438542073102e-03, 4.53293670083355e-03, 4.40037628649037e-03, 4.28503114248713e-03, 4.18532057767607e-03, 4.09975305903032e-03, 4.02692177436137e-03, 3.96549984181860e-03, 3.91423585655559e-03, 3.87195027518317e-03, 3.83753296332401e-03, 3.80994206576446e-03, 3.78820420220904e-03, 3.77141584688445e-03, 3.75874562028855e-03, 3.74943711537412e-03, 3.74281180230440e-03, 3.73827151473449e-03, 3.73530001965605e-03, 3.73346321244128e-03, 3.73240755752846e-03, 3.73185650531127e-03, 3.73160474665468e-03, 3.73151030677832e-03, 3.73148461717676e-03, 3.73148082546493e-03, 3.73148051880179e-03, 3.73147482832081e-03, 3.73143627462014e-03, 3.73129455009291e-03, 3.73091678712001e-03, 3.73009017670967e-03, 3.72850755795184e-03, 3.72575674406050e-03, 3.72131438225409e-03, 3.71454516943004e-03, 3.70470723692762e-03, 3.69096444322992e-03, 3.67240613618236e-03, 3.64807463562045e-03, 3.61700022219935e-03, 3.57824280049357e-03, 3.53093866738874e-03, 3.47435002895935e-03, 3.40791417214348e-03, 3.33128864304655e-03, 3.24438854235609e-03, 3.14741223165527e-03, 3.04085240873775e-03, 2.92549063168934e-03, 2.80237483705292e-03, 2.67278101769364e-03, 2.53816176013544e-03, 2.40008556143323e-03, 2.26017157235248e-03, 2.12002456783011e-03, 1.98117455397551e-03, 1.84502460076175e-03, 1.71280941383856e-03, 1.58556600467818e-03, 1.46411674004554e-03, 1.34906415247826e-03, 1.24079622568019e-03, 1.13950044147515e-03, 1.04518466422832e-03, 9.57702898141122e-04, 8.77128069738831e-04, 8.02369504938838e-04, 7.33104888951175e-04, 6.69444482880976e-04, 6.10600636640868e-04, 5.56104167586113e-04, 5.05764025902522e-04, 4.58757769294493e-04, 4.14860548303864e-04, 3.73788128364083e-04, 3.35325030384581e-04, 2.99315140026124e-04, 2.65591847067981e-04, 2.34268312873098e-04, 2.05229465592538e-04, 1.78269291460253e-04, 1.53448083963781e-04, 1.30754289457968e-04, 1.10170527534592e-04, 9.16702626483357e-05, 7.52111093262897e-05, 6.07324046839608e-05, 4.81579050983783e-05, 3.73961009856316e-05, 2.83397907097083e-05, 2.08667147845176e-05, 1.48411161902118e-05, 1.01159011762775e-05, 6.53518971915901e-06, 3.93713544610819e-06, 2.15696262273468e-06, 1.03022819786247e-06, 3.96389019385879e-07, -1.42108547152020e-14, -7.10542735760100e-15],
[ 6.10000000000000e+01, 6.09999999384495e+01, 6.09999980351863e+01, 6.09999851402765e+01, 6.09999377369043e+01, 6.09998113744621e+01, 6.09995348208851e+01, 6.09990051399436e+01, 6.09980838708988e+01, 6.09965944539755e+01, 6.09943210067228e+01, 6.09910085150426e+01, 6.09863644602184e+01, 6.09800618615723e+01, 6.09717436752977e+01, 6.09610284552129e+01, 6.09475171518880e+01, 6.09308009033291e+01, 6.09104696528749e+01, 6.08861214169316e+01, 6.08573720146666e+01, 6.08238650613651e+01, 6.07852820144980e+01, 6.07413520451101e+01, 6.06918614867951e+01, 6.06366625920844e+01, 6.05756813053648e+01, 6.05089237479635e+01, 6.04364811111044e+01, 6.03585326719389e+01, 6.02753466908898e+01, 6.01872790162584e+01, 6.00947693119091e+01, 5.99983349297008e+01, 5.98985625611287e+01, 5.97960979118000e+01, 5.96916337374736e+01, 5.95858966528233e+01, 5.94796331682265e+01, 5.93735954238110e+01, 5.92685270753577e+01, 5.91651497480625e+01, 5.90641504180913e+01, 5.89661700155160e+01, 5.88717934724048e+01, 5.87815413721902e+01, 5.86958632949553e+01, 5.86151329002175e+01, 5.85396447448601e+01, 5.84696127985669e+01, 5.84051705911648e+01, 5.83463729040263e+01, 5.82931988994273e+01, 5.82455565660095e+01, 5.82032883441348e+01, 5.81661777812166e+01, 5.81339570537887e+01, 5.81063151802142e+01, 5.80829067359683e+01, 5.80633608730191e+01, 5.80472904368029e+01, 5.80343009695670e+01, 5.80239993883245e+01, 5.80160021301279e+01, 5.80099425674488e+01, 5.80054775124960e+01, 5.80022926513799e+01, 5.80001067768471e+01, 5.79986747212221e+01, 5.79977889282679e+01, 5.79972796426671e+01, 5.79970137373241e+01, 5.79968922401538e+01, 5.79968466619140e+01, 5.79968342634846e+01, 5.79968324335008e+01, 5.79968321934455e+01, 5.79968266004455e+01, 5.79967887070287e+01, 5.79966494065408e+01, 5.79962780804405e+01, 5.79954654314744e+01, 5.79939090698118e+01, 5.79912024134634e+01, 5.79868273874679e+01, 5.79801513181258e+01, 5.79704283189651e+01, 5.79568053538109e+01, 5.79383330390073e+01, 5.79139811116053e+01, 5.78826583438705e+01, 5.78432365282293e+01, 5.77945779929852e+01, 5.77355659410129e+01, 5.76651367358300e+01, 5.75823130989608e+01, 5.74862370399649e+01, 5.73762012311824e+01, 5.72516774831800e+01, 5.71123409968874e+01, 5.69580891855479e+01, 5.67890540863807e+01, 5.66056077147917e+01, 5.64083601284047e+01, 5.61981504180674e+01, 5.59760312676454e+01, 5.57432480615761e+01, 5.55012137209840e+01, 5.52514804961485e+01, 5.49957098506416e+01, 5.47356413869636e+01, 5.44730615489775e+01, 5.42097726549717e+01, 5.39475627083922e+01, 5.36881764105492e+01, 5.34332878376290e+01, 5.31844752980162e+01, 5.29431989067208e+01, 5.27107813683543e+01, 5.24883923421781e+01, 5.22770365928044e+01, 5.20775459451310e+01, 5.18905748999920e+01, 5.17165996535851e+01, 5.15559202067505e+01, 5.14086652413794e+01, 5.12747994632671e+01, 5.11541331447590e+01, 5.10463336315837e+01, 5.09509385973604e+01, 5.08673708331103e+01, 5.07949543484302e+01, 5.07329315388101e+01, 5.06804811438115e+01, 5.06367366875391e+01, 5.06008050598817e+01, 5.05717848679709e+01, 5.05487841654805e+01, 5.05309371557646e+01, 5.05174194659411e+01, 5.05074616048651e+01, 5.05003602496981e+01, 5.04954870537254e+01, 5.04922947314003e+01, 5.04903202533882e+01, 5.04891850717149e+01, 5.04885923892350e+01, 5.04883215842112e+01, 5.04882199953308e+01, 5.04881923606386e+01, 5.04881882818140e+01, 5.04881879217314e+01, 5.04881808660905e+01, 5.04881330630757e+01, 5.04879573339874e+01, 5.04874889036134e+01, 5.04864637493922e+01, 5.04845004341689e+01, 5.04810861428684e+01, 5.04755675532705e+01, 5.04671470666273e+01, 5.04548848032056e+01, 5.04377066281032e+01, 5.04144183108502e+01, 5.03837257361389e+01, 5.03442608722454e+01, 5.02946129710928e+01, 5.02333642263105e+01, 5.01591288646264e+01, 5.00705944078945e+01, 4.99665636386936e+01, 4.98459956551829e+01, 4.97080443345194e+01, 4.95520925589547e+01, 4.93777807069810e+01, 4.91850281725755e+01, 4.89740470303058e+01, 4.87453473761003e+01, 4.84997342926347e+01, 4.82382967621303e+01, 4.79623891387905e+01, 4.76736059856875e+01, 4.73737511945844e+01, 4.70648023804947e+01, 4.67488716149707e+01, 4.64281636529417e+01, 4.61049329079166e+01, 4.57814405060237e+01, 4.54599127597789e+01, 4.51425023182562e+01, 4.48312530659386e+01, 4.45280695781068e+01, 4.42346916342868e+01, 4.39526739869359e+01, 4.36833713180679e+01, 4.34279281162644e+01, 4.31872730789997e+01, 4.29621175853402e+01, 4.27529577778137e+01, 4.25600798218504e+01, 4.23835679594370e+01, 4.22233150264062e+01, 4.20790351501595e+01, 4.19502783805896e+01, 4.18364470287019e+01, 4.17368134943082e+01, 4.16505393570348e+01, 4.15766954854959e+01, 4.15142828903127e+01, 4.14622540107341e+01, 4.14195340855384e+01, 4.13850422207084e+01, 4.13577117334733e+01, 4.13365093290903e+01, 4.13204526573848e+01, 4.13086258040330e+01, 4.13001922993321e+01, 4.12944052758412e+01, 4.12906144753434e+01, 4.12882698929972e+01, 4.12869219487318e+01, 4.12862181880941e+01, 4.12858966312482e+01, 4.12857760037213e+01, 4.12857431900664e+01, 4.12857383468382e+01, 4.12857379583514e+01, 4.12857307887990e+01, 4.12856822140199e+01, 4.12855036478906e+01, 4.12850276553590e+01, 4.12839859533293e+01, 4.12819909539289e+01, 4.12785215850526e+01, 4.12729140331970e+01, 4.12643579493830e+01, 4.12518985382338e+01, 4.12344448098474e+01, 4.12107841106846e+01, 4.11796028608530e+01, 4.11395132105538e+01, 4.10890850909149e+01, 4.10268828809365e+01, 4.09515056544986e+01, 4.08616297255923e+01, 4.07560519962198e+01, 4.06337324519695e+01, 4.04938340669732e+01, 4.03357583912190e+01, 4.01591752106434e+01, 3.99640448958180e+01, 3.97506323781404e+01, 3.95195120906496e+01, 3.92715636508959e+01, 3.90079585066898e+01, 3.87301381736022e+01, 3.84397850350079e+01, 3.81387869339503e+01, 3.78291969592994e+01, 3.75131899271763e+01, 3.71930170978828e+01, 3.68709606597291e+01, 3.65492894546098e+01, 3.62302173059542e+01, 3.59158651253559e+01, 3.56082277167463e+01, 3.53091458820430e+01, 3.50202840947010e+01, 3.47431136925603e+01, 3.44789012894562e+01, 3.42287019399574e+01, 3.39933565152389e+01, 3.37734927444319e+01, 3.35695294196575e+01, 3.33816833292903e+01, 3.32099785537325e+01, 3.30542578197739e+01, 3.29141956585725e+01, 3.27893131474338e+01, 3.26789940375540e+01, 3.25825020795717e+01, 3.24989993568053e+01, 3.24275654229816e+01, 3.23672170180429e+01, 3.23169281039585e+01, 3.22756499250864e+01, 3.22423307582566e+01, 3.22159349809036e+01, 3.21954610561707e+01, 3.21799580168017e+01, 3.21685400290737e+01, 3.21603986371744e+01, 3.21548123289929e+01, 3.21511531262277e+01, 3.21488899831950e+01, 3.21475888762078e+01, 3.21469095740032e+01, 3.21465991935940e+01, 3.21464827589247e+01, 3.21464510858468e+01, 3.21464464109659e+01, 3.21464460503941e+01, 3.21464395756820e+01, 3.21463957085528e+01, 3.21462344483178e+01, 3.21458045875252e+01, 3.21448638466139e+01, 3.21430622110861e+01, 3.21399291388307e+01, 3.21348652263427e+01, 3.21271388309390e+01, 3.21158880392867e+01, 3.21001282472732e+01, 3.20787654689793e+01, 3.20506153206493e+01, 3.20144274285058e+01, 3.19689147895582e+01, 3.19127873787456e+01, 3.18447890548869e+01, 3.17637365873641e+01, 3.16685594241186e+01, 3.15583386700090e+01, 3.14323436627610e+01, 3.12900645379047e+01, 3.11312392739364e+01, 3.09558739052659e+01, 3.07642548742049e+01, 3.05569528459707e+01, 3.03348177073585e+01, 3.00989648823518e+01, 2.98507534995280e+01, 2.95917573136835e+01, 2.93237295996052e+01, 2.90485634852852e+01, 2.87682493631527e+01, 2.84848311000454e+01, 2.82003627505090e+01, 2.79168673592858e+01, 2.76362992221761e+01, 2.73605106762147e+01, 2.70912241385592e+01, 2.68300097451357e+01, 2.65782685935354e+01, 2.63372213034757e+01, 2.61079013951954e+01, 2.58911528605791e+01, 2.56876312593836e+01, 2.54978076990663e+01, 2.53219751308803e+01, 2.51602564952996e+01, 2.50126143569858e+01, 2.48788617684906e+01, 2.47586741830552e+01, 2.46516022954046e+01, 2.45570857243539e+01, 2.44744674638531e+01, 2.44030090227893e+01, 2.43419061520764e+01, 2.42903050241651e+01, 2.42473186890162e+01, 2.42120435857635e+01, 2.41835758447629e+01, 2.41610270745606e+01, 2.41435392965491e+01, 2.41302986704935e+01, 2.41205476499710e+01, 2.41135952204819e+01, 2.41088249058304e+01, 2.41057002801997e+01, 2.41037677925659e+01, 2.41026567936276e+01, 2.41020767490257e+01, 2.41018117211638e+01, 2.41017122999509e+01, 2.41016852549598e+01, 2.41016812631754e+01, 2.41016809609296e+01, 2.41016756066742e+01, 2.41016393308077e+01, 2.41015059769097e+01, 2.41011505045678e+01, 2.41003725630010e+01, 2.40988827155055e+01, 2.40962918696797e+01, 2.40921044047419e+01, 2.40857154130850e+01, 2.40764123860987e+01, 2.40633815708942e+01, 2.40457191013500e+01, 2.40224468610716e+01, 2.39925328664593e+01, 2.39549157670397e+01, 2.39085328533238e+01, 2.38523507498473e+01, 2.37853977672973e+01, 2.37067967109194e+01, 2.36157968130481e+01, 2.35118033956551e+01, 2.33944038910963e+01, 2.32633889661786e+01, 2.31187677072895e+01, 2.29607761222898e+01, 2.27898785758834e+01, 2.26067621666254e+01, 2.24123244364534e+01, 2.22076551377058e+01, 2.19940130342968e+01, 2.17727988621206e+01, 2.15455256152891e+01, 2.13137872742927e+01, 2.10792269792112e+01, 2.08435055124784e+01, 2.06082708252863e+01, 2.03751292410114e+01, 2.01456189014647e+01, 1.99211859730772e+01, 1.97031640753749e+01, 1.94927573087095e+01, 1.92910271291582e+01, 1.90988831506387e+01, 1.89170777693492e+01, 1.87462043343615e+01, 1.85866984592661e+01, 1.84388420005604e+01, 1.83027692209672e+01, 1.81784746985074e+01, 1.80658226155528e+01, 1.79645571456432e+01, 1.78743137327355e+01, 1.77946311169957e+01, 1.77249639985446e+01, 1.76646962456931e+01, 1.76131545499951e+01, 1.75696224111676e+01, 1.75333543052569e+01, 1.75035898540161e+01, 1.74795677767225e+01, 1.74605393717657e+01, 1.74457812481580e+01, 1.74346070101170e+01, 1.74263775939451e+01, 1.74205099676580e+01, 1.74164839312631e+01, 1.74138467991019e+01, 1.74122158038569e+01, 1.74112781320513e+01, 1.74107885795286e+01, 1.74105648981757e+01, 1.74104809874546e+01, 1.74104581616923e+01, 1.74104547926572e+01, 1.74104545397426e+01, 1.74104500881601e+01, 1.74104199280329e+01, 1.74103090562089e+01, 1.74100135125842e+01, 1.74093667223454e+01, 1.74081280407214e+01, 1.74059739600359e+01, 1.74024923850296e+01, 1.73971803198404e+01, 1.73894452368971e+01, 1.73786103118322e+01, 1.73639236067398e+01, 1.73445711646736e+01, 1.73196938405744e+01, 1.72884075394041e+01, 1.72498263655664e+01, 1.72030880164193e+01, 1.71473805878190e+01, 1.70819698150493e+01, 1.70062256638885e+01, 1.69196471299814e+01, 1.68218841143882e+01, 1.67127553292019e+01, 1.65922613526292e+01, 1.64605921917767e+01, 1.63181290066756e+01, 1.61654399733892e+01, 1.60032705823882e+01, 1.58325289440539e+01, 1.56542668757891e+01, 1.54696576586370e+01, 1.52799713782931e+01, 1.50865487264860e+01, 1.48907740648286e+01, 1.46940484747433e+01, 1.44977634534440e+01, 1.43032758709294e+01, 1.41118847660198e+01, 1.39248105131047e+01, 1.37431768194214e+01, 1.35679959076031e+01, 1.34001571030015e+01, 1.32404188923303e+01, 1.30894043669755e+01, 1.29475998285646e+01, 1.28153562298371e+01, 1.26928930581756e+01, 1.25803042434024e+01, 1.24775656811631e+01, 1.23845440001634e+01, 1.23010062556992e+01, 1.22266302932464e+01, 1.21610155855131e+01, 1.21036943973983e+01, 1.20541431711541e+01, 1.20117940463623e+01, 1.19760464358420e+01, 1.19462785707803e+01, 1.19218589090792e+01, 1.19021572739969e+01, 1.18865555600974e+01, 1.18744578149749e+01, 1.18652994826621e+01, 1.18585555819454e+01, 1.18537475929755e+01, 1.18504488403241e+01, 1.18482881904328e+01, 1.18469519252374e+01, 1.18461837093872e+01, 1.18457826326272e+01, 1.18455993776048e+01, 1.18455306323616e+01, 1.18455119319926e+01, 1.18455091718575e+01, 1.18455089652833e+01, 1.18455053377381e+01, 1.18454807605903e+01, 1.18453904124734e+01, 1.18451495781659e+01, 1.18446225212238e+01, 1.18436131558153e+01, 1.18418578975008e+01, 1.18390210317932e+01, 1.18346928915981e+01, 1.18283910781534e+01, 1.18195648923531e+01, 1.18076030617195e+01, 1.17918447507306e+01, 1.17715937279352e+01, 1.17461354334123e+01, 1.17147565481761e+01, 1.16767665192837e+01, 1.16315203494126e+01, 1.15784418284691e+01, 1.15170462796000e+01, 1.14469618253309e+01, 1.13679481628260e+01, 1.12799118794097e+01, 1.11829174455756e+01, 1.10771931929293e+01, 1.09631318132514e+01, 1.08412851905086e+01, 1.07123536826522e+01, 1.05771702820112e+01, 1.04366803763687e+01, 1.02919180811258e+01, 1.01439802925594e+01, 9.99399970532256e+00, 9.84311803521318e+00, 9.69246059292356e+00, 9.54311317894661e+00, 9.39610203617713e+00, 9.25237733271903e+00, 9.11280038194398e+00, 8.97813456576160e+00, 8.84903972968749e+00, 8.72606967566901e+00, 8.60967229339172e+00, 8.50019183852601e+00, 8.39787287780351e+00, 8.30286546422879e+00, 8.21523116915716e+00, 8.13494967070292e+00, 8.06192567141554e+00, 7.99599598618421e+00, 7.93693669988944e+00, 7.88447034118434e+00, 7.83827305301771e+00, 7.79798176208297e+00, 7.76320135887064e+00, 7.73351189845635e+00, 7.70847582098341e+00, 7.68764517178013e+00, 7.67056877632359e+00, 7.65679929730363e+00, 7.64590007258948e+00, 7.63745160683057e+00, 7.63105756855148e+00, 7.62635013146831e+00, 7.62299449537363e+00, 7.62069242958528e+00, 7.61918470099086e+00, 7.61825227851572e+00, 7.61771624481598e+00, 7.61743639174192e+00, 7.61730852566269e+00, 7.61726055886016e+00, 7.61724751074367e+00, 7.61724558486963e+00, 7.61724544067936e+00, 7.61724290790511e+00, 7.61722574799658e+00, 7.61716266650171e+00, 7.61699451550354e+00, 7.61662652677309e+00, 7.61592180585179e+00, 7.61469635759867e+00, 7.61271588822446e+00, 7.60969459866782e+00, 7.60529614729576e+00, 7.59913691364847e+00, 7.59079163645511e+00, 7.57980142606127e+00, 7.56568406254065e+00, 7.54794638681715e+00, 7.52609847627845e+00, 7.49966917463929e+00, 7.46822242705326e+00, 7.43137376685272e+00, 7.38880622231498e+00, 7.34028487282369e+00, 7.28566929401723e+00, 7.22492319734670e+00, 7.15812069167736e+00, 7.08544876731881e+00, 7.00720581368459e+00, 6.92379621261714e+00, 6.83572127887047e+00, 6.74356702552877e+00, 6.64798939628879e+00, 6.54969771520915e+00, 6.44943715177295e+00, 6.34797098667612e+00, 6.24606339998448e+00, 6.14446340089335e+00, 6.04389039164749e+00, 5.94502172083526e+00, 5.84848244447480e+00, 5.75483738528091e+00, 5.66458546675516e+00, 5.57815620284185e+00, 5.49590814817653e+00, 5.41812905993954e+00, 5.34503749065119e+00, 5.27678552138021e+00, 5.21346235479071e+00, 5.15509851376105e+00, 5.10167042934530e+00, 5.05310524643556e+00, 5.00928572156369e+00, 4.97005513049950e+00, 4.93522214037266e+00, 4.90456562985590e+00, 4.87783946049060e+00, 4.85477721240789e+00, 4.83509689909076e+00, 4.81850566952571e+00, 4.80470449355270e+00, 4.79339280915713e+00, 4.78427309075918e+00, 4.77705527727726e+00, 4.77146097994179e+00, 4.76722737450343e+00, 4.76411067239489e+00, 4.76188906197147e+00, 4.76036501506796e+00, 4.75936686603750e+00, 4.75874958977096e+00, 4.75839473086708e+00, 4.75820946649971e+00, 4.75812481858443e+00, 4.75809306437535e+00, 4.75808442647437e+00, 4.75808315153878e+00, 4.75808305583107e+00, 4.75808137129532e+00, 4.75806995832542e+00, 4.75802800315529e+00, 4.75791616695509e+00, 4.75767142062958e+00, 4.75720271923917e+00, 4.75638769710519e+00, 4.75507054833646e+00, 4.75306123813610e+00, 4.75013616631579e+00, 4.74604037400930e+00, 4.74049134557040e+00, 4.73318440832664e+00, 4.72379967232564e+00, 4.71201038087463e+00, 4.69749246269924e+00, 4.67993499210687e+00, 4.65905118080681e+00, 4.63458945185716e+00, 4.60634409140511e+00, 4.57416494623109e+00, 4.53796564206131e+00, 4.49772984398693e+00, 4.45351516709906e+00, 4.40545446897129e+00, 4.35375440740731e+00, 4.29869131413485e+00, 4.24060460208010e+00, 4.17988807368994e+00, 4.11697961506827e+00, 4.05234983387314e+00, 3.98649022221341e+00, 3.91990140021389e+00, 3.85308192900196e+00, 3.78651808594513e+00, 3.72067488460436e+00, 3.65598851101364e+00, 3.59286024756162e+00, 3.53165187277178e+00, 3.47268246227065e+00, 3.41622647252508e+00, 3.36251296189121e+00, 3.31172578985165e+00, 3.26400463194067e+00, 3.21944665229946e+00, 3.17810868621711e+00, 3.14000979997740e+00, 3.10513411360181e+00, 3.07343379241978e+00, 3.04483213450465e+00, 3.01922670154671e+00, 2.99649245939568e+00, 2.97648491014456e+00, 2.95904320934141e+00, 2.94399326911371e+00, 2.93115085044095e+00, 2.92032464565376e+00, 2.91131934597565e+00, 2.90393867938706e+00, 2.89798839239129e+00, 2.89327913670492e+00, 2.88962920989801e+00, 2.88686708897484e+00, 2.88483368909039e+00, 2.88338427706022e+00, 2.88238997170441e+00, 2.88173877059479e+00, 2.88133605520559e+00, 2.88110454310696e+00, 2.88098367561382e+00, 2.88092845087134e+00, 2.88090773426309e+00, 2.88090209885272e+00, 2.88090126707829e+00, 2.88090120437219e+00, 2.88090009714960e+00, 2.88089259555890e+00, 2.88086501899729e+00, 2.88079151061782e+00, 2.88063064241786e+00, 2.88032257241304e+00, 2.87978687315113e+00, 2.87892114023557e+00, 2.87760047695377e+00, 2.87567793553988e+00, 2.87298597574443e+00, 2.86933897586301e+00, 2.86453679904695e+00, 2.85836937798713e+00, 2.85062223421679e+00, 2.84108279576346e+00, 2.82954732144648e+00, 2.81582818580931e+00, 2.79976123058965e+00, 2.78121285242695e+00, 2.76008647772407e+00, 2.73632807881010e+00, 2.70993041361744e+00, 2.68093572437827e+00, 2.64943670696699e+00, 2.61557565633471e+00, 2.57954179762474e+00, 2.54156691825037e+00, 2.50191951435270e+00, 2.46089774728478e+00, 2.41882156543679e+00, 2.37602437954579e+00, 2.33284468405344e+00, 2.28961799424564e+00, 2.24666942243163e+00, 2.20430715183152e+00, 2.16281699092024e+00, 2.12245811096657e+00, 2.08345999232771e+00, 2.04602053660641e+00, 2.01030524639430e+00, 1.97644733452562e+00, 1.94454860121916e+00, 1.91468090923218e+00, 1.88688809199326e+00, 1.86118814466334e+00, 1.83757556994183e+00, 1.81602377604023e+00, 1.79648745082498e+00, 1.77890486146894e+00, 1.76320005143344e+00, 1.74928492517940e+00, 1.73706122512264e+00, 1.72642241483766e+00, 1.71725548750392e+00, 1.70944271943095e+00, 1.70286338570230e+00, 1.69739544918724e+00, 1.69291722613077e+00, 1.68930902208655e+00, 1.68645472200765e+00, 1.68424330879610e+00, 1.68257027644707e+00, 1.68133889794636e+00, 1.68046130496246e+00, 1.67985933657357e+00, 1.67946511795596e+00, 1.67922133699517e+00, 1.67908119672564e+00, 1.67900803365309e+00, 1.67897460547520e+00, 1.67896206551421e+00, 1.67895865435046e+00, 1.67895815087019e+00, 1.67895811270709e+00, 1.67895743610862e+00, 1.67895285205951e+00, 1.67893600068653e+00, 1.67889108170041e+00, 1.67879278057782e+00, 1.67860453358647e+00, 1.67827720621927e+00, 1.67774825409815e+00, 1.67694142874213e+00, 1.67576708194613e+00, 1.67412311083424e+00, 1.67189656981538e+00, 1.66896595472073e+00, 1.66520413779219e+00, 1.66048190003856e+00, 1.65467197078456e+00, 1.64765344503063e+00, 1.63931641056485e+00, 1.62956658251726e+00, 1.61832971760946e+00, 1.60555556806870e+00, 1.59122113968384e+00, 1.57533304202701e+00, 1.55792876172281e+00, 1.53907674974577e+00, 1.51887528663489e+00, 1.49745016879790e+00, 1.47495133704631e+00, 1.45154863722454e+00, 1.42742695522697e+00, 1.40278099971142e+00, 1.37781001295010e+00, 1.35271267400548e+00, 1.32768242205104e+00, 1.30290337659161e+00, 1.27854697219923e+00, 1.25476936496387e+00, 1.23170961217293e+00, 1.20948858031110e+00, 1.18820850200523e+00, 1.16795308087974e+00, 1.14878803371198e+00, 1.13076195993989e+00, 1.11390743700951e+00, 1.09824225363634e+00, 1.08377070936563e+00, 1.07048492584772e+00, 1.05836613149732e+00, 1.04738589569358e+00, 1.03750730083797e+00, 1.02868605019300e+00, 1.02087151647076e+00, 1.01400774075803e+00, 1.00803439375586e+00, 1.00288771171302e+00, 9.98501418094222e-01, 9.94807639214940e-01, 9.91737818086698e-01, 9.89223625878076e-01, 9.87197865077668e-01, 9.85595353057413e-01, 9.84353769717337e-01, 9.83414448689679e-01, 9.82723088603869e-01, 9.82230359507781e-01, 9.81892379935843e-01, 9.81671042399640e-01, 9.81534169177749e-01, 9.81455485958591e-01, 9.81414407762159e-01, 9.81395639145397e-01, 9.81388598445974e-01, 9.81386683210374e-01, 9.81386400525821e-01, 9.81386378962746e-01, 9.81385994874820e-01, 9.81383392625524e-01, 9.81373826525740e-01, 9.81348327152951e-01, 9.81292524109257e-01, 9.81185661156914e-01, 9.80999846086824e-01, 9.80699574562038e-01, 9.80241563395715e-01, 9.79574923793308e-01, 9.78641698392515e-01, 9.77377776812894e-01, 9.75714192312104e-01, 9.73578786715990e-01, 9.70898212128148e-01, 9.67600216660315e-01, 9.63616138814729e-01, 9.58883513081517e-01, 9.53348670184731e-01, 9.46969201881458e-01, 9.39716154845982e-01, 9.31575822968519e-01, 9.22551023399151e-01, 9.12661768571390e-01, 9.01945282470265e-01, 8.90455351398543e-01, 8.78261043267337e-01, 8.65444870395500e-01, 8.52100504644767e-01, 8.38330177105298e-01, 8.24241905618045e-01, 8.09946691995577e-01, 7.95555818293948e-01, 7.81178350510302e-01, 7.66918931906453e-01, 7.52875920131608e-01, 7.39139895348650e-01, 7.25792542822923e-01, 7.12905894203537e-01, 7.00541897499718e-01, 6.88752276378160e-01, 6.77578634328086e-01, 6.67052757706870e-01, 6.57197072905682e-01, 6.48025216139828e-01, 6.39542679048574e-01, 6.31747498861735e-01, 6.24630967917433e-01, 6.18178343420212e-01, 6.12369544187793e-01, 6.07179826467096e-01, 6.02580435473122e-01, 5.98539232936893e-01, 5.95021303516695e-01, 5.91989544371259e-01, 5.89405242516342e-01, 5.87228643860605e-01, 5.85419516173637e-01, 5.83937705872188e-01, 5.82743685663999e-01, 5.81799087042235e-01, 5.81067208683216e-01, 5.80513489268014e-01, 5.80105931404582e-01, 5.79815462399637e-01, 5.79616217770813e-01, 5.79485734660324e-01, 5.79405044669286e-01, 5.79358658931966e-01, 5.79334442265569e-01, 5.79323377669070e-01, 5.79319226989014e-01, 5.79318097906430e-01, 5.79317931256320e-01, 5.79317918454606e-01, 5.79317689250651e-01, 5.79316136361818e-01, 5.79310427804266e-01, 5.79295211085508e-01, 5.79261910682604e-01, 5.79198140312336e-01, 5.79087255260497e-01, 5.78908068175451e-01, 5.78634749477111e-01, 5.78236930604625e-01, 5.77680024368568e-01, 5.76925771304268e-01, 5.75933013818634e-01, 5.74658690899192e-01, 5.73059035257849e-01, 5.71090942364570e-01, 5.68713467588957e-01, 5.65889394657873e-01, 5.62586807184346e-01, 5.58780586604598e-01, 5.54453755910522e-01, 5.49598590246667e-01, 5.44217423413812e-01, 5.38323093572188e-01, 5.31938991184781e-01, 5.25098695979487e-01, 5.17845215364616e-01, 5.10229861928025e-01, 5.02310830018729e-01, 4.94151548899772e-01, 4.85818901088997e-01, 4.77381398501713e-01, 4.68907405873694e-01, 4.60463491385532e-01, 4.52112969726301e-01, 4.43914684728703e-01, 4.35922059080227e-01, 4.28182419338903e-01, 4.20736587205305e-01, 4.13618713967999e-01, 4.06856324992859e-01, 4.00470535282601e-01, 3.94476395233163e-01, 3.88883327142132e-01, 3.83695616952600e-01, 3.78912931251598e-01, 3.74530835857520e-01, 3.70541298734139e-01, 3.66933165953981e-01, 3.63692604664557e-01, 3.60803511310444e-01, 3.58247886662035e-01, 3.56006181511788e-01, 3.54057618278368e-01, 3.52380494295858e-01, 3.50952472357932e-01, 3.49750863248403e-01, 3.48752903638228e-01, 3.47936030998291e-01, 3.47278155212045e-01, 3.46757924530735e-01, 3.46354981565064e-01, 3.46050203317098e-01, 3.45825917982303e-01, 3.45666090523750e-01, 3.45556468928898e-01, 3.45484683641565e-01, 3.45440293896502e-01, 3.45414776490713e-01, 3.45401454769423e-01, 3.45395368116584e-01, 3.45393084827499e-01, 3.45392463719833e-01, 3.45392372045715e-01, 3.45392364949767e-01, 3.45392237202979e-01, 3.45391371700991e-01, 3.45388190043693e-01, 3.45379709063678e-01, 3.45361149408525e-01, 3.45325608368931e-01, 3.45263811340615e-01, 3.45163955443255e-01, 3.45011657744742e-01, 3.44790019058500e-01, 3.44479812145458e-01, 3.44059800098272e-01, 3.43507186478597e-01, 3.42798193285660e-01, 3.41908756072172e-01, 3.40815317718629e-01, 3.39495693999865e-01, 3.37929975836730e-01, 3.36101425934856e-01, 3.33997322385841e-01, 3.31609699732248e-01, 3.28935939764307e-01, 3.25979170344390e-01, 3.22748440763263e-01, 3.19258655866641e-01, 3.15530267252974e-01, 3.11588736630334e-01, 3.07463802135899e-01, 3.03188591364957e-01, 2.98798633679432e-01, 2.94330828290235e-01, 2.89822423515652e-01, 2.85310057037940e-01, 2.80828897939877e-01, 2.76411920127813e-01, 2.72089324817664e-01, 2.67888118308136e-01, 2.63831841235628e-01, 2.59940437495565e-01, 2.56230245291157e-01, 2.52714089326979e-01, 2.49401451796388e-01, 2.46298700186848e-01, 2.43409351645298e-01, 2.40734356288911e-01, 2.38272385021368e-01, 2.36020110782505e-01, 2.33972475451715e-01, 2.32122937646068e-01, 2.30463699268938e-01, 2.28985910799551e-01, 2.27679856931061e-01, 2.26535125271447e-01, 2.25540761443497e-01, 2.24685414104542e-01, 2.23957473208678e-01, 2.23345204314825e-01, 2.22836880969645e-01, 2.22420916233645e-01, 2.22085993348088e-01, 2.21821194440672e-01, 2.21616125122132e-01, 2.21461031920541e-01, 2.21346908812315e-01, 2.21265588706444e-01, 2.21209815667429e-01, 2.21173293941888e-01, 2.21150710476159e-01, 2.21137728533002e-01, 2.21130951167851e-01, 2.21127854620292e-01, 2.21126693012415e-01, 2.21126377028227e-01, 2.21126330389673e-01, 2.21126326750401e-01, 2.21126260855137e-01, 2.21125814405802e-01, 2.21124173222222e-01, 2.21119798516973e-01, 2.21110225026152e-01, 2.21091892349484e-01, 2.21060017001051e-01, 2.21008512270849e-01, 2.20929962489379e-01, 2.20815657560834e-01, 2.20655692543488e-01, 2.20439135455543e-01, 2.20154264239632e-01, 2.19788870847481e-01, 2.19330626716925e-01, 2.18767499635383e-01, 2.18088207381690e-01, 2.17282689022262e-01, 2.16342570825610e-01, 2.15261601033276e-01, 2.14036026740644e-01, 2.12664887337297e-01, 2.11150202549753e-01, 2.09497039049616e-01, 2.07713447436355e-01, 2.05810270472220e-01, 2.03800832807055e-01, 2.01700531057227e-01, 1.99526350021223e-01, 1.97296335269300e-01, 1.95029053903788e-01, 1.92743073918216e-01, 1.90456488625209e-01, 1.88186506707474e-01, 1.85949121374271e-01, 1.83758864723011e-01, 1.81628646501522e-01, 1.79569670694882e-01, 1.77591419164003e-01, 1.75701689126420e-01, 1.73906670505062e-01, 1.72211049778561e-01, 1.70618128548519e-01, 1.69129947204674e-01, 1.67747406495621e-01, 1.66470382245507e-01, 1.65297830687209e-01, 1.64227883742257e-01, 1.63257934969216e-01, 1.62384717805807e-01, 1.61604378200849e-01, 1.60912543861215e-01, 1.60304392216318e-01, 1.59774718901303e-01, 1.59318008135699e-01, 1.58928505870655e-01, 1.58600296034869e-01, 1.58327379660975e-01, 1.58103756146957e-01, 1.57923505422161e-01, 1.57780869363272e-01, 1.57670330459027e-01, 1.57586685472637e-01, 1.57525111715671e-01, 1.57481223542675e-01, 1.57451116813199e-01, 1.57431399348404e-01, 1.57419205825960e-01, 1.57412196089886e-01, 1.57408536471188e-01, 1.57406864382921e-01, 1.57406237128413e-01, 1.57406066500307e-01, 1.57406041315959e-01, 1.57406039334063e-01, 1.57406003234293e-01, 1.57405758653191e-01, 1.57404859550759e-01, 1.57402462900883e-01, 1.57397218028713e-01, 1.57387173997620e-01, 1.57369708990238e-01, 1.57341485312834e-01, 1.57298433271087e-01, 1.57235766672400e-01, 1.57148032049101e-01, 1.57029192805389e-01, 1.56872748313397e-01, 1.56671886492134e-01, 1.56419666605636e-01, 1.56109226984981e-01, 1.55734010246903e-01, 1.55287996555785e-01, 1.54765933813614e-01, 1.54163552644029e-01, 1.53477753922111e-01, 1.52706757569305e-01, 1.51850203438819e-01, 1.50909198251853e-01, 1.49886306435569e-01, 1.48785486940639e-01, 1.47611982170432e-01, 1.46372168507625e-01, 1.45073380123605e-01, 1.43723718518373e-01, 1.42331859536412e-01, 1.40906867688471e-01, 1.39458024941682e-01, 1.37994678229610e-01, 1.36526107186086e-01, 1.35061411255386e-01, 1.33609413520979e-01, 1.32178577491992e-01, 1.30776932868748e-01, 1.29412007015510e-01, 1.28090760234940e-01, 1.26819524435263e-01, 1.25603945907528e-01, 1.24448933506166e-01, 1.23358613727817e-01, 1.22336294318558e-01, 1.21384438203123e-01, 1.20504649593727e-01, 1.19697673896567e-01, 1.18963412447037e-01, 1.18300952346786e-01, 1.17708610964520e-01, 1.17183994151183e-01, 1.16724066860286e-01, 1.16325234607473e-01, 1.15983433935178e-01, 1.15694229794300e-01, 1.15452917535592e-01, 1.15254627021564e-01, 1.15094426289346e-01, 1.14967422159459e-01, 1.14868855231187e-01, 1.14794186806215e-01, 1.14739175461651e-01, 1.14699941233461e-01, 1.14673015702792e-01, 1.14655376659528e-01, 1.14644466476122e-01, 1.14638193811921e-01, 1.14634918802210e-01, 1.14633422397517e-01, 1.14632861040043e-01, 1.14632708336806e-01, 1.14632685797922e-01, 1.14632684008669e-01, 1.14632651215508e-01, 1.14632429038146e-01, 1.14631612287461e-01, 1.14629435114296e-01, 1.14624670343602e-01, 1.14615544927240e-01, 1.14599674781608e-01, 1.14574021800124e-01, 1.14534875306568e-01, 1.14477859598594e-01, 1.14397968546108e-01, 1.14289627471116e-01, 1.14146781677753e-01, 1.13963010131513e-01, 1.13731661824602e-01, 1.13446011415050e-01, 1.13099429807988e-01, 1.12685564520838e-01, 1.12198524018133e-01, 1.11633059765303e-01, 1.10984739627433e-01, 1.10250106440944e-01, 1.09426816172338e-01, 1.08513751016226e-01, 1.07511103954678e-01, 1.06420432507349e-01, 1.05244680443079e-01, 1.03988166880832e-01, 1.02656542650699e-01, 1.01256714259250e-01, 9.97967366603163e-02, 9.82856774187368e-02, 9.67334561683941e-02, 9.51506640438922e-02, 9.35483675482942e-02, 9.19379009054637e-02, 9.03306511729269e-02, 8.87378419878289e-02, 8.71703239958919e-02, 8.56383807626457e-02, 8.41515577847431e-02, 8.27185184018049e-02, 8.13176847214478e-02, 8.00173712709906e-02, 7.87900155884529e-02, 7.76399244378965e-02, 7.65702094763085e-02, 7.55828088029489e-02, 7.46785299935055e-02, 7.38571112358599e-02, 7.31172973006551e-02, 7.24569273544589e-02, 7.18730319809353e-02, 7.13619371475502e-02, 7.09193731852409e-02, 7.05405870968363e-02, 7.02204566559885e-02, 6.99536047983358e-02, 6.97345127557014e-02, 6.95576302665444e-02, 6.94174810475019e-02, 6.93087615693813e-02, 6.92264310837771e-02, 6.91657908260765e-02, 6.91225504011825e-02, 6.90928795524641e-02, 6.90734438259568e-02, 6.90614230586633e-02, 6.90545121249500e-02, 6.90509039407274e-02, 6.90492553192357e-02, 6.90486368623908e-02, 6.90484686269670e-02, 6.90484437957508e-02, 6.90484418063093e-02, 6.90484051198297e-02, 6.90481565634755e-02, 6.90472428438014e-02, 6.90448072003704e-02, 6.90394768430735e-02, 6.90292685283111e-02, 6.90115160496276e-02, 6.89828229679605e-02, 6.89390433858534e-02, 6.88752929689542e-02, 6.87859917221374e-02, 6.86649392234671e-02, 6.85054220978927e-02, 6.83003524744166e-02, 6.80424350292697e-02, 6.77243590004082e-02, 6.73390103080678e-02, 6.68796976878667e-02, 6.63403856027003e-02, 6.57159257227065e-02, 6.50022780239041e-02, 6.41967121374947e-02, 6.32979795548607e-02, 6.23064477285098e-02, 6.12241880580982e-02, 6.00550112447564e-02, 5.88044455337879e-02, 5.74796558939070e-02, 5.60893050963287e-02, 5.46433607881625e-02, 5.31528557770713e-02, 5.16296115912880e-02, 5.00859376734297e-02, 4.85343200563761e-02, 4.69871138734987e-02, 4.54562534892204e-02, 4.39529924442397e-02, 4.24876829536345e-02, 4.10696016406433e-02, 3.97068248564878e-02, 3.84061536593308e-02, 3.71730856011681e-02, 3.60118281284940e-02, 3.49253467777118e-02, 3.39154404840301e-02, 3.29828361763234e-02, 3.21272952919784e-02, 3.13477257587635e-02, 3.06422941912032e-02, 3.00085343764920e-02, 2.94434494425602e-02, 2.89436063041393e-02, 2.85052219948767e-02, 2.81242422709127e-02, 2.77964133978372e-02, 2.75173483115552e-02, 2.72825883979877e-02, 2.70876619983014e-02, 2.69281404599582e-02, 2.67996921613189e-02, 2.66981344923138e-02, 2.66194833207791e-02, 2.65599990621439e-02, 2.65162281392008e-02, 2.64850384039264e-02, 2.64636470137845e-02, 2.64496393235518e-02, 2.64409775637162e-02, 2.64359984113298e-02, 2.64333989899299e-02, 2.64322113259112e-02, 2.64317657974757e-02, 2.64316446034485e-02, 2.64316267154712e-02, 2.64316252688204e-02, 2.64315984231231e-02, 2.64314165399711e-02, 2.64307479233707e-02, 2.64289656766152e-02, 2.64250654738589e-02, 2.64175968882583e-02, 2.64046113625866e-02, 2.63836297652276e-02, 2.63516322881827e-02, 2.63050731728498e-02, 2.62399223317686e-02, 2.61517353107195e-02, 2.60357521495251e-02, 2.58870245184966e-02, 2.57005690290768e-02, 2.54715428805810e-02, 2.51954361027558e-02, 2.48682727208234e-02, 2.44868113925207e-02, 2.40487346546558e-02, 2.35528150947033e-02, 2.29990467322523e-02, 2.23887308136455e-02, 2.17245071640022e-02, 2.10103251719573e-02, 2.02513522504262e-02, 1.94538219376241e-02, 1.86248282980017e-02, 1.77720774982646e-02, 1.69036109138361e-02, 1.60275164571431e-02, 1.51516457247993e-02, 1.42833539211082e-02, 1.34292774151703e-02, 1.25951605016255e-02, 1.17857388656830e-02, 1.10046828555213e-02, 1.02545993622343e-02, 9.53708724815527e-03, 8.85283811054194e-03, 8.20177189203264e-03, 7.58319553791154e-03, 6.99597253233009e-03, 6.43869162095712e-03, 5.90982419833352e-03, 5.40786156564366e-03, 4.93142543110991e-03, 4.47934750955881e-03, 4.05071671650603e-03, 3.64489502415211e-03, 3.26150531206082e-03, 2.90039629192762e-03, 2.56159068256068e-03, 2.24522321964571e-03, 1.95147485377589e-03, 1.68050873095638e-03, 1.43241241864445e-03, 1.20714951312806e-03, 1.00452241692039e-03, 8.24146858249719e-04, 6.65437748551767e-04, 5.27605291992472e-04, 4.09811585939224e-04, 3.10549150078998e-04, 2.28648543746601e-04, 1.62617263647746e-04, 1.10839579420485e-04, 7.16048203059927e-05, 4.31380761014566e-05, 2.36331231846520e-05, 1.12878453961684e-05, 4.34309072261385e-06, 1.12721384226688e-06, -4.26325641456060e-14],
[ 6.20000000000000e+01, 6.19999999370018e+01, 6.19999979889738e+01, 6.19999847907753e+01, 6.19999362724738e+01, 6.19998069379772e+01, 6.19995238798157e+01, 6.19989817406059e+01, 6.19980388027405e+01, 6.19965143530758e+01, 6.19941874300916e+01, 6.19907970187213e+01, 6.19860437146172e+01, 6.19795928369717e+01, 6.19710789290901e+01, 6.19601115504255e+01, 6.19462822339681e+01, 6.19291724592747e+01, 6.19083624736772e+01, 6.18834407810214e+01, 6.18540141065704e+01, 6.18197176358991e+01, 6.17802253122936e+01, 6.17352599597374e+01, 6.16846029769512e+01, 6.16281033240801e+01, 6.15656855015691e+01, 6.14973562063457e+01, 6.14232093502686e+01, 6.13434291460712e+01, 6.12582910109657e+01, 6.11681601087864e+01, 6.10734874452723e+01, 6.09748035412682e+01, 6.08727098256863e+01, 6.07678680029362e+01, 6.06609877473416e+01, 6.05528131509115e+01, 6.04441083950515e+01, 6.03356431295648e+01, 6.02281780255395e+01, 6.01224509274206e+01, 6.00191639705963e+01, 5.99189719617230e+01, 5.98224722468384e+01, 5.97301962228402e+01, 5.96426025851701e+01, 5.95600723507548e+01, 5.94829056510149e+01, 5.94113202544364e+01, 5.93454517503916e+01, 5.92853553038322e+01, 5.92310088723911e+01, 5.91823177617682e+01, 5.91391203809005e+01, 5.91011950445965e+01, 5.90682676577884e+01, 5.90400201024236e+01, 5.90160991357467e+01, 5.89961255979987e+01, 5.89797037192536e+01, 5.89664303101575e+01, 5.89559036206683e+01, 5.89477316553186e+01, 5.89415397437018e+01, 5.89369771811572e+01, 5.89337227770954e+01, 5.89314891767404e+01, 5.89300258556483e+01, 5.89291207242100e+01, 5.89286003201993e+01, 5.89283286098406e+01, 5.89282044602643e+01, 5.89281578870079e+01, 5.89281452179091e+01, 5.89281433479751e+01, 5.89281431026792e+01, 5.89281373875787e+01, 5.89280986669138e+01, 5.89279563253764e+01, 5.89275768929686e+01, 5.89267465036186e+01, 5.89251561672730e+01, 5.89223904295243e+01, 5.89179199145813e+01, 5.89110981575416e+01, 5.89011630303822e+01, 5.88872429524285e+01, 5.88683679499054e+01, 5.88434854905287e+01, 5.88114808685958e+01, 5.87712017551478e+01, 5.87214863588160e+01, 5.86611944693705e+01, 5.85892404827623e+01, 5.85046273411301e+01, 5.84064801751359e+01, 5.82940783252169e+01, 5.81668843637518e+01, 5.80245687652092e+01, 5.78670289972934e+01, 5.76944020448265e+01, 5.75070697246215e+01, 5.73056565768553e+01, 5.70910205776662e+01, 5.68642373457996e+01, 5.66265788496574e+01, 5.63794878129378e+01, 5.61245490504183e+01, 5.58634588599555e+01, 5.55979934035538e+01, 5.53299767968235e+01, 5.50612494554072e+01, 5.47936371583533e+01, 5.45289212861415e+01, 5.42688107451835e+01, 5.40149161501667e+01, 5.37687268493498e+01, 5.35315913142181e+01, 5.33047012728788e+01, 5.30890797739641e+01, 5.28855731676267e+01, 5.26948468231345e+01, 5.25173842934010e+01, 5.23534895896890e+01, 5.22032922322437e+01, 5.20667547740893e+01, 5.19436825351896e+01, 5.18337353175401e+01, 5.17364408903932e+01, 5.16512100365542e+01, 5.15773529372664e+01, 5.15140966483164e+01, 5.14606033879104e+01, 5.14159893217889e+01, 5.13793434967944e+01, 5.13497465442126e+01, 5.13262887519005e+01, 5.13080870924188e+01, 5.12943007956448e+01, 5.12841450706220e+01, 5.12769026139149e+01, 5.12719325907311e+01, 5.12686768397320e+01, 5.12666631308226e+01, 5.12655053943203e+01, 5.12649009359088e+01, 5.12646247503011e+01, 5.12645211429676e+01, 5.12644929592063e+01, 5.12644887993402e+01, 5.12644884321032e+01, 5.12644812362750e+01, 5.12644324834700e+01, 5.12642532628509e+01, 5.12637755253226e+01, 5.12627300025464e+01, 5.12607276788129e+01, 5.12572455504742e+01, 5.12516173157276e+01, 5.12430295310524e+01, 5.12305236475436e+01, 5.12130041978286e+01, 5.11892532390834e+01, 5.11579509675837e+01, 5.11177022049630e+01, 5.10670682187671e+01, 5.10046030870971e+01, 5.09288935610816e+01, 5.08386011366120e+01, 5.07325048395472e+01, 5.06095430805688e+01, 5.04688528711426e+01, 5.03098047308002e+01, 5.01320317694872e+01, 4.99354516943060e+01, 4.97202808471175e+01, 4.94870397898919e+01, 4.92365503682838e+01, 4.89699245508072e+01, 4.86885456267877e+01, 4.83940425435978e+01, 4.80882582933703e+01, 4.77732133585716e+01, 4.74510653279283e+01, 4.71240659107369e+01, 4.67945166916481e+01, 4.64647250422681e+01, 4.61369616003301e+01, 4.58134206171385e+01, 4.54961842605912e+01, 4.51871916708151e+01, 4.48882132408008e+01, 4.46008302810422e+01, 4.43264199622972e+01, 4.40661452362506e+01, 4.38209493156252e+01, 4.35915542452974e+01, 4.33784630981543e+01, 4.31819653649548e+01, 4.30021451588792e+01, 4.28388919090806e+01, 4.26919132641240e+01, 4.25607499604634e+01, 4.24447924308388e+01, 4.23432989324384e+01, 4.22554149659144e+01, 4.21801937356493e+01, 4.21166173714617e+01, 4.20636185951798e+01, 4.20201024757145e+01, 4.19849678774739e+01, 4.19571281735539e+01, 4.19355307715889e+01, 4.19191749906682e+01, 4.19071278358691e+01, 4.18985372452324e+01, 4.18926424335459e+01, 4.18887810276396e+01, 4.18863927769390e+01, 4.18850197271178e+01, 4.18843028589468e+01, 4.18839753131217e+01, 4.18838524389156e+01, 4.18838190141090e+01, 4.18838140806759e+01, 4.18838136849536e+01, 4.18838063818689e+01, 4.18837569023885e+01, 4.18835750104811e+01, 4.18830901526438e+01, 4.18820290490976e+01, 4.18799968934266e+01, 4.18764629093411e+01, 4.18707509217424e+01, 4.18620354931702e+01, 4.18493440537769e+01, 4.18315653099501e+01, 4.18074640502174e+01, 4.17757022746262e+01, 4.17348663550454e+01, 4.16834996916838e+01, 4.16201400727285e+01, 4.15433606813137e+01, 4.14518134437150e+01, 4.13442731952489e+01, 4.12196809785035e+01, 4.10771847044547e+01, 4.09161754192719e+01, 4.07363175396658e+01, 4.05375716487162e+01, 4.03202087712877e+01, 4.00848154500603e+01, 3.98322893864305e+01, 3.95638258563726e+01, 3.92808955225310e+01, 3.89852146115446e+01, 3.86787086946330e+01, 3.83634714989895e+01, 3.80417202969072e+01, 3.77157494800861e+01, 3.73878839327981e+01, 3.70604337630444e+01, 3.67356518215401e+01, 3.64156952233475e+01, 3.61025917907530e+01, 3.57982119852962e+01, 3.55042465368210e+01, 3.52221896581500e+01, 3.49533274943417e+01, 3.46987313121762e+01, 3.44592548818431e+01, 3.42355355160330e+01, 3.40279982836814e+01, 3.38368629820866e+01, 3.36621535157170e+01, 3.35037093846106e+01, 3.33611990275032e+01, 3.32341347950708e+01, 3.31218893479485e+01, 3.30237132829935e+01, 3.29387537896017e+01, 3.28660741256842e+01, 3.28046736805944e+01, 3.27535083612881e+01, 3.27115110008874e+01, 3.26776114493633e+01, 3.26507559688700e+01, 3.26299255264238e+01, 3.26141525591396e+01, 3.26025357865026e+01, 3.25942526634504e+01, 3.25885691091307e+01, 3.25848462090727e+01, 3.25825436713061e+01, 3.25812199160948e+01, 3.25805287894286e+01, 3.25802130063145e+01, 3.25800945449022e+01, 3.25800623205009e+01, 3.25800575642458e+01, 3.25800571973977e+01, 3.25800506099823e+01, 3.25800059792719e+01, 3.25798419120331e+01, 3.25794045688096e+01, 3.25784474528355e+01, 3.25766144573113e+01, 3.25734268504108e+01, 3.25682747971928e+01, 3.25604139244310e+01, 3.25489673258071e+01, 3.25329332775196e+01, 3.25111987845888e+01, 3.24825589033287e+01, 3.24457415850129e+01, 3.23994375622617e+01, 3.23423345597703e+01, 3.22731548658961e+01, 3.21906950669436e+01, 3.20938665408300e+01, 3.19817351520235e+01, 3.18535585053064e+01, 3.17088191184159e+01, 3.15472519727740e+01, 3.13688650983571e+01, 3.11739521345517e+01, 3.09630961659119e+01, 3.07371645358338e+01, 3.04972947650128e+01, 3.02448721183409e+01, 2.99814997497474e+01, 2.97089626890961e+01, 2.94291872011481e+01, 2.91441972279792e+01, 2.88560697089307e+01, 2.85668905456726e+01, 2.82787128411148e+01, 2.79935187978414e+01, 2.77131863359214e+01, 2.74394611148272e+01, 2.71739342604101e+01, 2.69180257460816e+01, 2.66729730908383e+01, 2.64398248361433e+01, 2.62194381546506e+01, 2.60124799184441e+01, 2.58194305951980e+01, 2.56405904249890e+01, 2.54760874358788e+01, 2.53258869637472e+01, 2.51898024373880e+01, 2.50675072653316e+01, 2.49585477128406e+01, 2.48623566861022e+01, 2.47782683480697e+01, 2.47055334799909e+01, 2.46433354780650e+01, 2.45908068396107e+01, 2.45470459512682e+01, 2.45111339468006e+01, 2.44821513578255e+01, 2.44591942412143e+01, 2.44413894358957e+01, 2.44279085831354e+01, 2.44179805412868e+01, 2.44109018409257e+01, 2.44060448604655e+01, 2.44028634556485e+01, 2.44008958470996e+01, 2.43997646553116e+01, 2.43991740676969e+01, 2.43989042225254e+01, 2.43988029941540e+01, 2.43987754575701e+01, 2.43987713932276e+01, 2.43987710854879e+01, 2.43987656339088e+01, 2.43987286986599e+01, 2.43985929207919e+01, 2.43982309869557e+01, 2.43974389041614e+01, 2.43959219729696e+01, 2.43932840235043e+01, 2.43890204134967e+01, 2.43825152116458e+01, 2.43730429005268e+01, 2.43597748277754e+01, 2.43417905088251e+01, 2.43180937363259e+01, 2.42876332796175e+01, 2.42493277643062e+01, 2.42020941129868e+01, 2.41448787138094e+01, 2.40766902786410e+01, 2.39966331754336e+01, 2.39039398905107e+01, 2.37980012158665e+01, 2.36783927808380e+01, 2.35448966665267e+01, 2.33975170553562e+01, 2.32364891657609e+01, 2.30622810799950e+01, 2.28755884582433e+01, 2.26773225056911e+01, 2.24685918825140e+01, 2.22506794888665e+01, 2.20250152007641e+01, 2.17931456787746e+01, 2.15567023376575e+01, 2.13173684819093e+01, 2.10768465135752e+01, 2.08368260316757e+01, 2.05989535786311e+01, 2.03648047407473e+01, 2.01358592548061e+01, 1.99134796842338e+01, 1.96988940877405e+01, 1.94931829107325e+01, 1.92972701060008e+01, 1.91119182702672e+01, 1.89377274041211e+01, 1.87751367910377e+01, 1.86244294549594e+01, 1.84857386858672e+01, 1.83590561975531e+01, 1.82442415764559e+01, 1.81410327732939e+01, 1.80490574656813e+01, 1.79678451728584e+01, 1.78968400319731e+01, 1.78354141516224e+01, 1.77828814468960e+01, 1.77385118357118e+01, 1.77015456433177e+01, 1.76712080246935e+01, 1.76467231771404e+01, 1.76273280813963e+01, 1.76122854827387e+01, 1.76008958070819e+01, 1.75925077038477e+01, 1.75865269194815e+01, 1.75824232339484e+01, 1.75797352372499e+01, 1.75780727825282e+01, 1.75771170240339e+01, 1.75766180284515e+01, 1.75763900324459e+01, 1.75763045031431e+01, 1.75762812370867e+01, 1.75762778030650e+01, 1.75762775452720e+01, 1.75762730078212e+01, 1.75762422659245e+01, 1.75761292554513e+01, 1.75758280109527e+01, 1.75751687443864e+01, 1.75739061687031e+01, 1.75717105347951e+01, 1.75681617953603e+01, 1.75627472453351e+01, 1.75548629138962e+01, 1.75438188951068e+01, 1.75288487004451e+01, 1.75091225946524e+01, 1.74837647359682e+01, 1.74518737845516e+01, 1.74125464733356e+01, 1.73649034617505e+01, 1.73081166261043e+01, 1.72414367952672e+01, 1.71642208327781e+01, 1.70759569127498e+01, 1.69762868508975e+01, 1.68650244426280e+01, 1.67421689287679e+01, 1.66079129478044e+01, 1.64626446225165e+01, 1.63069437404708e+01, 1.61415722886667e+01, 1.59674598604757e+01, 1.57856846442632e+01, 1.55974508184868e+01, 1.54040632246387e+01, 1.52069001868528e+01, 1.50073853196789e+01, 1.48069591331195e+01, 1.46070512148266e+01, 1.44090537389051e+01, 1.42142970059466e+01, 1.40240276451557e+01, 1.38393899978054e+01, 1.36614110526386e+01, 1.34909891290952e+01, 1.33288863212067e+01, 1.31757245436474e+01, 1.30319848793023e+01, 1.28980098266243e+01, 1.27740079897415e+01, 1.26600607427661e+01, 1.25561304248878e+01, 1.24620696743825e+01, 1.23776315763446e+01, 1.23024803700863e+01, 1.22362025287837e+01, 1.21783180794320e+01, 1.21282920712927e+01, 1.20855461238992e+01, 1.20494699915051e+01, 1.20194330714604e+01, 1.19947957625636e+01, 1.19749205500522e+01, 1.19591826612323e+01, 1.19469801046313e+01, 1.19377428806057e+01, 1.19309411365527e+01, 1.19260920383359e+01, 1.19227651430809e+01, 1.19205860875863e+01, 1.19192384502527e+01, 1.19184637003584e+01, 1.19180592133426e+01, 1.19178744003983e+01, 1.19178050707656e+01, 1.19177862114324e+01, 1.19177834278346e+01, 1.19177832195045e+01, 1.19177795611231e+01, 1.19177547750582e+01, 1.19176636589647e+01, 1.19174207777047e+01, 1.19168892420277e+01, 1.19158713030248e+01, 1.19141011467943e+01, 1.19112402333374e+01, 1.19068754764481e+01, 1.19005205051764e+01, 1.18916201795651e+01, 1.18795584509160e+01, 1.18636695575542e+01, 1.18432524299693e+01, 1.18175880454067e+01, 1.17859593250706e+01, 1.17476730135704e+01, 1.17020828293219e+01, 1.16486130378968e+01, 1.15867814908664e+01, 1.15162211039020e+01, 1.14366987320236e+01, 1.13481304465313e+01, 1.12505923327037e+01, 1.11443261097498e+01, 1.10297391183005e+01, 1.09073985127213e+01, 1.07780198161060e+01, 1.06424503201609e+01, 1.05016481124570e+01, 1.03566577618565e+01, 1.02085838650541e+01, 1.00585637361491e+01, 9.90774050018168e+00, 9.75723773555842e+00, 9.60813661537367e+00, 9.46145624867876e+00, 9.31813764934575e+00, 9.17903149190512e+00, 9.04488957588695e+00, 8.91635973060786e+00, 8.79398376008473e+00, 8.67819795345447e+00, 8.56933566386983e+00, 8.46763147796497e+00, 8.37322654649348e+00, 8.28617471275758e+00, 8.20644914874385e+00, 8.13394928153909e+00, 8.06850785908000e+00, 8.00989806112658e+00, 7.95784060667301e+00, 7.91201084208878e+00, 7.87204581513233e+00, 7.83755134913542e+00, 7.80810913004732e+00, 7.78328380794657e+00, 7.76263009574371e+00, 7.74569982316653e+00, 7.73204887616463e+00, 7.72124392330357e+00, 7.71286880441712e+00, 7.70653043558220e+00, 7.70186407093838e+00, 7.69853775804203e+00, 7.69625583062205e+00, 7.69476130118074e+00, 7.69383704524695e+00, 7.69330570766932e+00, 7.69302830673413e+00, 7.69290156113503e+00, 7.69285401467584e+00, 7.69284108090411e+00, 7.69283917190715e+00, 7.69283902898048e+00, 7.69283651840182e+00, 7.69281950887211e+00, 7.69275698019293e+00, 7.69259030285380e+00, 7.69222553947119e+00, 7.69152699655378e+00, 7.69031229527113e+00, 7.68834920490507e+00, 7.68535445307094e+00, 7.68099469051278e+00, 7.67488974256776e+00, 7.66661822136360e+00, 7.65572550004705e+00, 7.64173396161517e+00, 7.62415533092918e+00, 7.60250478243722e+00, 7.57631639410126e+00, 7.54515939896071e+00, 7.50865458096317e+00, 7.46649008376605e+00, 7.41843586254193e+00, 7.36435601968867e+00, 7.30421833199609e+00, 7.23810039980212e+00, 7.16619202188616e+00, 7.08879361057170e+00, 7.00631069160139e+00, 6.91924476151781e+00, 6.82818098009801e+00, 6.73377333844638e+00, 6.63672805190584e+00, 6.53778597548128e+00, 6.43770482977860e+00, 6.33724196541992e+00, 6.23713829522500e+00, 6.13810389926668e+00, 6.04080567063624e+00, 5.94585722990774e+00, 5.85381120219103e+00, 5.76515382880140e+00, 5.68030178115237e+00, 5.59960096169385e+00, 5.52332701861343e+00, 5.45168726900658e+00, 5.38482371861237e+00, 5.32281688206268e+00, 5.26569014103621e+00, 5.21341442271368e+00, 5.16591303123250e+00, 5.12306651480402e+00, 5.08471749635233e+00, 5.05067543295965e+00, 5.02072129743405e+00, 4.99461219345016e+00, 4.97208592428812e+00, 4.95286553510504e+00, 4.93666384117834e+00, 4.92318794117183e+00, 4.91214369688554e+00, 4.90324014099904e+00, 4.89619375394783e+00, 4.89073253226213e+00, 4.88659975536057e+00, 4.88355734764099e+00, 4.88138872911327e+00, 4.87990105164698e+00, 4.87892672942764e+00, 4.87832419103273e+00, 4.87797780560900e+00, 4.87779696535544e+00, 4.87771433889849e+00, 4.87768334301536e+00, 4.87767491139809e+00, 4.87767366690957e+00, 4.87767357348748e+00, 4.87767192918059e+00, 4.87766078876734e+00, 4.87761983554795e+00, 4.87751067020915e+00, 4.87727176914302e+00, 4.87681426267751e+00, 4.87601871034369e+00, 4.87473303459183e+00, 4.87277175567884e+00, 4.86991664727432e+00, 4.86591890269562e+00, 4.86050286349729e+00, 4.85337131364069e+00, 4.84421228272257e+00, 4.83270723114745e+00, 4.81854041085752e+00, 4.80140911148558e+00, 4.78103441980428e+00, 4.75717204803689e+00, 4.72962273288559e+00, 4.69824168085924e+00, 4.66294654409060e+00, 4.62372345900150e+00, 4.58063076857776e+00, 4.53380017353394e+00, 4.48343520920480e+00, 4.42980711023507e+00, 4.37324828790766e+00, 4.31414378852940e+00, 4.25292121093987e+00, 4.19003962659914e+00, 4.12597806265609e+00, 4.06122407942291e+00, 3.99626290718753e+00, 3.93156751557116e+00, 3.86758988521797e+00, 3.80475364872979e+00, 3.74344817460834e+00, 3.68402408995914e+00, 3.62679017682516e+00, 3.57201153276137e+00, 3.51990885680557e+00, 3.47065870510363e+00, 3.42439455400503e+00, 3.38120851065943e+00, 3.34115352045733e+00, 3.30424593559559e+00, 3.27046832809233e+00, 3.23977245212389e+00, 3.21208228295902e+00, 3.18729708142040e+00, 3.16529445226890e+00, 3.14593338099360e+00, 3.12905724533928e+00, 3.11449680501394e+00, 3.10207317526670e+00, 3.09160078764369e+00, 3.08289033479337e+00, 3.07575168656804e+00, 3.06999675297012e+00, 3.06544225700885e+00, 3.06191236863823e+00, 3.05924114100129e+00, 3.05727468343654e+00, 3.05587300309465e+00, 3.05491144920354e+00, 3.05428170122517e+00, 3.05389225413182e+00, 3.05366837011771e+00, 3.05355148521038e+00, 3.05349808015298e+00, 3.05347804617286e+00, 3.05347259645401e+00, 3.05347179208729e+00, 3.05347173144739e+00, 3.05347066070873e+00, 3.05346340630218e+00, 3.05343673841584e+00, 3.05336565223308e+00, 3.05321008496173e+00, 3.05291216690897e+00, 3.05239412210030e+00, 3.05155692380376e+00, 3.05027979826271e+00, 3.04842065497227e+00, 3.04581750266817e+00, 3.04229088547480e+00, 3.03764734224231e+00, 3.03168385339435e+00, 3.02419319387783e+00, 3.01497005946196e+00, 3.00381777937365e+00, 2.99055537506028e+00, 2.97502467775872e+00, 2.95709718211778e+00, 2.93668029485740e+00, 2.91372264092294e+00, 2.88821811764094e+00, 2.86020844043172e+00, 2.82978399933591e+00, 2.79708293886122e+00, 2.76228847705882e+00, 2.72562458442732e+00, 2.68735023995776e+00, 2.64775256183591e+00, 2.60713916721194e+00, 2.56583014472720e+00, 2.52415002374627e+00, 2.48242009701864e+00, 2.44095140294684e+00, 2.40003860594453e+00, 2.35995493590009e+00, 2.32094826817060e+00, 2.28323835085298e+00, 2.24701512200450e+00, 2.21243800985795e+00, 2.17963607572253e+00, 2.14870884208226e+00, 2.11972764573922e+00, 2.09273736496515e+00, 2.06775838724762e+00, 2.04478870702053e+00, 2.02380606773487e+00, 2.00477008728387e+00, 1.98762432830837e+00, 1.97229829406267e+00, 1.95870934564771e+00, 1.94676454727641e+00, 1.93636245289245e+00, 1.92739485020280e+00, 1.91974847742436e+00, 1.91330672430865e+00, 1.90795132289119e+00, 1.90356402557995e+00, 1.90002825938132e+00, 1.89723073604432e+00, 1.89506298949262e+00, 1.89342280490389e+00, 1.89221549890770e+00, 1.89135500819313e+00, 1.89076474473102e+00, 1.89037817995810e+00, 1.89013912749737e+00, 1.89000170388046e+00, 1.88992995863315e+00, 1.88989717815436e+00, 1.88988488114873e+00, 1.88988153607294e+00, 1.88988104234706e+00, 1.88988100492332e+00, 1.88988034143317e+00, 1.88987584619413e+00, 1.88985932128426e+00, 1.88981527243598e+00, 1.88971887511723e+00, 1.88953427233091e+00, 1.88921327661036e+00, 1.88869454292013e+00, 1.88790327223299e+00, 1.88675149671296e+00, 1.88513898573481e+00, 1.88295479640669e+00, 1.88007947196536e+00, 1.87638786590374e+00, 1.87175253910853e+00, 1.86604764260305e+00, 1.85915316163995e+00, 1.85095936075992e+00, 1.84137123771213e+00, 1.83031277095156e+00, 1.81773073484379e+00, 1.80359786205869e+00, 1.78791515589272e+00, 1.77071319646618e+00, 1.75205234168681e+00, 1.73202179211471e+00, 1.71073756213536e+00, 1.68833947082594e+00, 1.66498732723057e+00, 1.64085653014983e+00, 1.61613332773590e+00, 1.59100998554327e+00, 1.56568009441411e+00, 1.54033421535506e+00, 1.51515601282094e+00, 1.49031897676210e+00, 1.46598378338384e+00, 1.44229629969215e+00, 1.41938620084384e+00, 1.39736614358711e+00, 1.37633142358717e+00, 1.35636003789961e+00, 1.33751307432978e+00, 1.31983535475169e+00, 1.30335626769862e+00, 1.28809073512615e+00, 1.27404026810943e+00, 1.26119407571561e+00, 1.24953020004433e+00, 1.23901665829215e+00, 1.22961257957049e+00, 1.22126932999944e+00, 1.21393162418691e+00, 1.20753862444820e+00, 1.20202503090479e+00, 1.19732216585471e+00, 1.19335905455457e+00, 1.19006350192524e+00, 1.18736316094316e+00, 1.18518658396751e+00, 1.18346424343334e+00, 1.18212950371487e+00, 1.18111952205051e+00, 1.18037605369622e+00, 1.17984613532751e+00, 1.17948262140133e+00, 1.17924455081321e+00, 1.17909732566120e+00, 1.17901269000431e+00, 1.17896850377830e+00, 1.17894831500396e+00, 1.17894074154255e+00, 1.17893868138165e+00, 1.17893837730638e+00, 1.17893835411164e+00, 1.17893794095981e+00, 1.17893514179774e+00, 1.17892485181878e+00, 1.17889742279598e+00, 1.17883739648718e+00, 1.17872244443020e+00, 1.17852255906617e+00, 1.17819953742921e+00, 1.17770679420983e+00, 1.17698953541056e+00, 1.17598531650814e+00, 1.17462499943302e+00, 1.17283411027677e+00, 1.17053458417875e+00, 1.16764686542957e+00, 1.16409231005934e+00, 1.15979581620459e+00, 1.15468858607395e+00, 1.14871090449773e+00, 1.14181480520816e+00, 1.13396648941358e+00, 1.12514836369169e+00, 1.11536057669200e+00, 1.10462195646908e+00, 1.09297028113103e+00, 1.08046185245617e+00, 1.06717038201692e+00, 1.05318523866988e+00, 1.03860914175652e+00, 1.02355541339479e+00, 1.00814492409770e+00, 9.92502877807695e-01, 9.76755585181863e-01, 9.61027367966317e-01, 9.45437723175232e-01, 9.30098854382514e-01, 9.15113649918602e-01, 9.00574155889686e-01, 8.86560558122106e-01, 8.73140654384081e-01, 8.60369769725530e-01, 8.48291046273545e-01, 8.36936026096765e-01, 8.26325442225922e-01, 8.16470137639335e-01, 8.07372042995041e-01, 7.99025158600728e-01, 7.91416502070788e-01, 7.84526998280228e-01, 7.78332301202639e-01, 7.72803547297691e-01, 7.67908047136004e-01, 7.63609926134251e-01, 7.59870727044223e-01, 7.56649986682987e-01, 7.53905797775957e-01, 7.51595364120504e-01, 7.49675553912233e-01, 7.48103452302672e-01, 7.46836910347923e-01, 7.45835083713864e-01, 7.45058951077510e-01, 7.44471799352681e-01, 7.44039660896441e-01, 7.43731686911936e-01, 7.43520441479333e-01, 7.43382102064139e-01, 7.43296554921308e-01, 7.43247377378324e-01, 7.43221703304208e-01, 7.43209972844070e-01, 7.43205572381726e-01, 7.43204375352889e-01, 7.43204198674057e-01, 7.43204185101973e-01, 7.43203942104948e-01, 7.43202295766281e-01, 7.43196243681581e-01, 7.43180111278868e-01, 7.43144807092712e-01, 7.43077199883892e-01, 7.42959644557125e-01, 7.42769681946022e-01, 7.42479935230556e-01, 7.42058222700251e-01, 7.41467902355595e-01, 7.40668457998915e-01, 7.39616328611611e-01, 7.38265972717201e-01, 7.36571147131599e-01, 7.34486365418029e-01, 7.31968486337301e-01, 7.28978367906244e-01, 7.25482510000916e-01, 7.21454599632267e-01, 7.16876869901647e-01, 7.11741187675976e-01, 7.06049796982825e-01, 6.99815664808945e-01, 6.93062402041285e-01, 6.85823762244354e-01, 6.78142751490022e-01, 6.70070409843916e-01, 6.61664345946761e-01, 6.52987117937542e-01, 6.44104555737435e-01, 6.35084112097850e-01, 6.25993314925012e-01, 6.16898374269397e-01, 6.07862977255650e-01, 5.98947285886140e-01, 5.90207137890516e-01, 5.81693440336444e-01, 5.73451739352862e-01, 5.65521946242541e-01, 5.57938199460999e-01, 5.50728842479316e-01, 5.43916498755870e-01, 5.37518226522685e-01, 5.31545737688290e-01, 5.26005666848391e-01, 5.20899878202322e-01, 5.16225800107300e-01, 5.11976779027442e-01, 5.08142446667932e-01, 5.04709096012157e-01, 5.01660063673427e-01, 4.98976117317063e-01, 4.96635847810929e-01, 4.94616066165825e-01, 4.92892205216123e-01, 4.91438725390289e-01, 4.90229522895476e-01, 4.89238337290935e-01, 4.88439153881991e-01, 4.87806594783684e-01, 4.87316291043967e-01, 4.86945227046391e-01, 4.86672047679043e-01, 4.86477318581095e-01, 4.86343730235220e-01, 4.86256237790201e-01, 4.86202130235952e-01, 4.86171024825609e-01, 4.86154785305750e-01, 4.86147365405223e-01, 4.86144581955735e-01, 4.86143824791050e-01, 4.86143713035176e-01, 4.86143704384816e-01, 4.86143548654283e-01, 4.86142493557765e-01, 4.86138614922709e-01, 4.86128275992797e-01, 4.86105649939567e-01, 4.86062320034470e-01, 4.85986974265067e-01, 4.85865209695220e-01, 4.85679460101757e-01, 4.85409058224285e-01, 4.85030441150059e-01, 4.84517503719303e-01, 4.83842100209102e-01, 4.82974688847349e-01, 4.81885106965919e-01, 4.80543457044355e-01, 4.78921075942914e-01, 4.76991551894059e-01, 4.74731747092044e-01, 4.72122778844792e-01, 4.69150910046380e-01, 4.65808300841857e-01, 4.62093578131842e-01, 4.58012187955553e-01, 4.53576507312611e-01, 4.48805705750033e-01, 4.43725361866797e-01, 4.38366854439238e-01, 4.32766560841543e-01, 4.26964905703777e-01, 4.21005309509198e-01, 4.14933089660941e-01, 4.08794365432307e-01, 4.02635013526016e-01, 3.96499713392402e-01, 3.90431111910218e-01, 3.84469126540445e-01, 3.78650395640300e-01, 3.73007875167282e-01, 3.67570573194065e-01, 3.62363407910259e-01, 3.57407171242528e-01, 3.52718578764957e-01, 3.48310386887107e-01, 3.44191559973737e-01, 3.40367472599851e-01, 3.36840135131943e-01, 3.33608433877721e-01, 3.30668379872357e-01, 3.28013362783622e-01, 3.25634408314862e-01, 3.23520438829441e-01, 3.21658537726563e-01, 3.20034218410541e-01, 3.18631698575906e-01, 3.17434180050050e-01, 3.16424133666693e-01, 3.15583587668968e-01, 3.14894417038929e-01, 3.14338630011285e-01, 3.13898646942476e-01, 3.13557565766416e-01, 3.13299407566390e-01, 3.13109335409678e-01, 3.12973839593006e-01, 3.12880882872179e-01, 3.12820000106591e-01, 3.12782348010451e-01, 3.12760702308722e-01, 3.12749401450958e-01, 3.12744238029786e-01, 3.12742301057753e-01, 3.12741774154915e-01, 3.12741696385181e-01, 3.12741690316685e-01, 3.12741580436381e-01, 3.12740835981629e-01, 3.12738099293881e-01, 3.12730804340319e-01, 3.12714839772553e-01, 3.12684266710377e-01, 3.12631103011462e-01, 3.12545184693498e-01, 3.12414113877011e-01, 3.12223301017760e-01, 3.11956107174595e-01, 3.11594089474390e-01, 3.11117349679407e-01, 3.10504981774926e-01, 3.09735609807987e-01, 3.08788001975564e-01, 3.07641741475585e-01, 3.06277929331223e-01, 3.04679889825366e-01, 3.02833845948193e-01, 3.00729530952530e-01, 2.98360703198306e-01, 2.95725535197377e-01, 2.92826854092425e-01, 2.89672219323478e-01, 2.86273833232229e-01, 2.82648290855590e-01, 2.78816185075335e-01, 2.74801591573379e-01, 2.70631463862081e-01, 2.66334971526632e-01, 2.61942814660507e-01, 2.57486544632421e-01, 2.52997916456565e-01, 2.48508291998021e-01, 2.44048106909482e-01, 2.39646408319786e-01, 2.35330465408367e-01, 2.31125451365656e-01, 2.27054192871378e-01, 2.23136981951683e-01, 2.19391444630126e-01, 2.15832460872065e-01, 2.12472130676900e-01, 2.09319781606396e-01, 2.06382013442507e-01, 2.03662776005771e-01, 2.01163476451200e-01, 1.98883112619008e-01, 1.96818429286431e-01, 1.94964094445218e-01, 1.93312893010898e-01, 1.91855935614633e-01, 1.90582880306636e-01, 1.89482165062486e-01, 1.88541248910738e-01, 1.87746859267662e-01, 1.87085242688731e-01, 1.86542415750829e-01, 1.86104412213936e-01, 1.85757522044518e-01, 1.85488517387682e-01, 1.85284860227434e-01, 1.85134886342176e-01, 1.85027960300104e-01, 1.84954596675809e-01, 1.84906543414357e-01, 1.84876824296779e-01, 1.84859738729543e-01, 1.84850818516217e-01, 1.84846742795365e-01, 1.84845213851489e-01, 1.84844797941658e-01, 1.84844736554233e-01, 1.84844731723288e-01, 1.84844643729193e-01, 1.84844047556403e-01, 1.84841855964457e-01, 1.84836014002556e-01, 1.84823229100999e-01, 1.84798744826963e-01, 1.84756167538595e-01, 1.84687354192967e-01, 1.84582368413530e-01, 1.84429510557003e-01, 1.84215425863904e-01, 1.83925292760115e-01, 1.83543090941306e-01, 1.83051945996624e-01, 1.82434544041278e-01, 1.81673606239499e-01, 1.80752409390173e-01, 1.79655335184201e-01, 1.78368427646483e-01, 1.76879935996773e-01, 1.75180819034843e-01, 1.73265187426118e-01, 1.71130662074575e-01, 1.68778630079146e-01, 1.66214384395744e-01, 1.63447138900338e-01, 1.60489916651258e-01, 1.57359315281397e-01, 1.54075159199856e-01, 1.50660053293358e-01, 1.47138856857019e-01, 1.43538099412332e-01, 1.39885361827695e-01, 1.36208646739355e-01, 1.32535761702222e-01, 1.28893736874303e-01, 1.25308296409479e-01, 1.21803399260585e-01, 1.18400860889103e-01, 1.15120062714276e-01, 1.11977751310270e-01, 1.08987924802541e-01, 1.06161799988541e-01, 1.03507850682561e-01, 1.01031905782804e-01, 9.87372945827047e-02, 9.66250267935465e-02, 9.46939955383712e-02, 9.29411930354789e-02, 9.13619305971309e-02, 8.99500567052911e-02, 8.86981690039548e-02, 8.75978179024273e-02, 8.66397010158906e-02, 8.58138487283355e-02, 8.51098018637160e-02, 8.45167826578148e-02, 8.40238600962024e-02, 8.36201101985168e-02, 8.32947710996237e-02, 8.30373919705991e-02, 8.28379738740921e-02, 8.26870999274973e-02, 8.25760515051499e-02, 8.24969068405608e-02, 8.24426183910134e-02, 8.24070655467040e-02, 8.23850799016412e-02, 8.23724411645870e-02, 8.23658428436644e-02, 8.23628280717372e-02, 8.23616971353189e-02, 8.23613894944941e-02, 8.23613440865714e-02, 8.23613404819881e-02, 8.23612744154332e-02, 8.23608268109908e-02, 8.23591813770932e-02, 8.23547953177908e-02, 8.23451968674362e-02, 8.23268159620909e-02, 8.22948554005265e-02, 8.22432093445400e-02, 8.21644353943185e-02, 8.20497856780236e-02, 8.18893013169841e-02, 8.16719731016846e-02, 8.13859692151837e-02, 8.10189282863121e-02, 8.05583129044223e-02, 7.99918151929546e-02, 7.93078020890050e-02, 7.84957841750185e-02, 7.75468882814749e-02, 7.64543112448024e-02, 7.52137305288585e-02, 7.38236471443727e-02, 7.22856378413042e-02, 7.06044969264497e-02, 6.87882533020598e-02, 6.68480550776707e-02, 6.47979221648072e-02, 6.26543758221132e-02, 6.04359626300948e-02, 5.81626974586698e-02, 5.58554556978521e-02, 5.35353476241411e-02, 5.12231083070418e-02, 4.89385337587080e-02, 4.66999901944192e-02, 4.45240165358740e-02, 4.24250330255020e-02, 4.04016638241131e-02, 3.84927888069768e-02, 3.66898645785909e-02, 3.49980697627785e-02, 3.34204904412587e-02, 3.19583290353123e-02, 3.06111404083111e-02, 2.93770788751999e-02, 2.82531424636900e-02, 2.72354038610274e-02, 2.63192206448366e-02, 2.54994203652856e-02, 2.47704586350039e-02, 2.41265504932215e-02, 2.35617769092400e-02, 2.30701693927529e-02, 2.26457763320767e-02, 2.22827149463072e-02, 2.19752126798340e-02, 2.17176415503175e-02, 2.15045484459466e-02, 2.13306837060543e-02, 2.11910295686977e-02, 2.10808292781763e-02, 2.09956168703016e-02, 2.09312469436256e-02, 2.08839231320468e-02, 2.08502235596271e-02, 2.08271213145821e-02, 2.08119979426087e-02, 2.08026481283312e-02, 2.07972740876556e-02, 2.07944686946888e-02, 2.07931869652427e-02, 2.07927061568540e-02, 2.07925753664995e-02, 2.07925560621547e-02, 2.07925545155305e-02, 2.07925259946578e-02, 2.07923327622268e-02, 2.07916224298756e-02, 2.07897290251324e-02, 2.07855857635977e-02, 2.07776524870908e-02, 2.07638613495575e-02, 2.07415843494232e-02, 2.07076262902177e-02, 2.06582466081926e-02, 2.05892131285097e-02, 2.04958900770293e-02, 2.03733614770201e-02, 2.02165893282356e-02, 2.00206036942063e-02, 1.97807190913579e-02, 1.94927685649304e-02, 1.91533438374378e-02, 1.87600273016173e-02, 1.83115998249690e-02, 1.78082077538733e-02, 1.72514734849329e-02, 1.66445366877497e-02, 1.59920176714010e-02, 1.52999001861005e-02, 1.45753375909763e-02, 1.38263930549272e-02, 1.30617304560516e-02, 1.22902771136140e-02, 1.15208818148176e-02, 1.07619914707726e-02, 1.00251593197953e-02, 9.30967394566551e-03, 8.62507515806499e-03, 7.97596467896966e-03, 7.36575239918125e-03, 6.79668280781698e-03, 6.26991877265937e-03, 5.78568252692691e-03, 5.34339173925957e-03, 4.94180227180152e-03, 4.57916878497144e-03, 4.25337357827971e-03, 3.96204617969431e-03, 3.70269673384627e-03, 3.47280227238400e-03, 3.26985691435056e-03, 3.09144367461300e-03, 2.93528392374043e-03, 2.79922460144633e-03, 2.68123195834136e-03, 2.57942953761869e-03, 2.49211747239997e-03, 2.41774224298297e-03, 2.35486001393070e-03, 2.30213073862018e-03, 2.25832730314579e-03, 2.22233344848632e-03, 2.19313021175083e-03, 2.16978201179785e-03, 2.15142816671232e-03, 2.13727936761643e-03, 2.12661673765524e-03, 2.11879184956842e-03, 2.11322704033279e-03, 2.10941570974654e-03, 2.10692227137343e-03, 2.10538135814886e-03, 2.10449589429373e-03, 2.10403372077650e-03, 2.10382257692032e-03, 2.10374337394370e-03, 2.10372182923635e-03, 2.10371864931602e-03, 2.10371839209245e-03, 2.10371361931294e-03, 2.10368128325200e-03, 2.10356241596799e-03, 2.10324558611319e-03, 2.10255234772916e-03, 2.10122523397871e-03, 2.09891900205150e-03, 2.09519589569851e-03, 2.08952575831489e-03, 2.08129186450634e-03, 2.06980331133622e-03, 2.05431466088868e-03, 2.03405314628924e-03, 2.00825299804563e-03, 1.97619529135551e-03, 1.93725073138040e-03, 1.89092329960516e-03, 1.83689497480799e-03, 1.77507224739060e-03, 1.70562903350202e-03, 1.62902988684266e-03, 1.54601590777759e-03, 1.45756223106019e-03, 1.36484489758715e-03, 1.26921783284217e-03, 1.17214975613678e-03, 1.07511958747608e-03, 9.79543860346155e-04, 8.86749465117020e-04, 7.97920089105730e-04, 7.14019723595527e-04, 6.35767404226328e-04, 5.63637016973928e-04, 4.97840949087447e-04, 4.38367973941922e-04, 3.85056893222213e-04, 3.37626761137244e-04, 2.95714323719665e-04, 2.58950150879400e-04, 2.26922767488702e-04, 1.99066698748140e-04, 1.75245138365199e-04, 1.54804472472847e-04, 1.36607876967787e-04, 1.21285659481885e-04, 1.08509335206008e-04, 9.70507590878356e-05, 8.65898240149932e-05, 7.69234334256907e-05, 6.79610518830032e-05, 5.96756136701515e-05, 5.19917491459410e-05, 4.59893725732741e-05, 4.02792299567523e-05, 3.49210216725737e-05, 2.99849668223828e-05, 2.55077276136717e-05, 2.14941332288276e-05, 1.79203818646556e-05, 1.47514180426356e-05, 1.19575726813537e-05, 9.51782613611840e-06, 7.41580699781252e-06, 5.63543404297207e-06, 4.15818748678021e-06, 2.96195801752219e-06, 2.02099317903048e-06, 1.30647505613979e-06, 7.87386078116015e-07, 4.31455902116795e-07, -3.55271367880050e-14, -3.55271367880050e-14, -2.84217094304040e-14, -2.84217094304040e-14],
[ 6.30000000000000e+01, 6.29999999355721e+01, 6.29999979433328e+01, 6.29999844455960e+01, 6.29999348261520e+01, 6.29998025563515e+01, 6.29995130740352e+01, 6.29989586305961e+01, 6.29979942918044e+01, 6.29964352424395e+01, 6.29940555045671e+01, 6.29905881360541e+01, 6.29857269315067e+01, 6.29791296042431e+01, 6.29704223871713e+01, 6.29592059542801e+01, 6.29450625341428e+01, 6.29275640629020e+01, 6.29062812062629e+01, 6.28807930666936e+01, 6.28506973811015e+01, 6.28156210030566e+01, 6.27752304496295e+01, 6.27292422744555e+01, 6.26774330056889e+01, 6.26196483621485e+01, 6.25558114375093e+01, 6.24859295269850e+01, 6.24100992705568e+01, 6.23285098078654e+01, 6.22414436867829e+01, 6.21492753414784e+01, 6.20524670534749e+01, 6.19515624237863e+01, 6.18471775056254e+01, 6.17399898637777e+01, 6.16307259272289e+01, 6.15201470768483e+01, 6.14090349541215e+01, 6.12981764884085e+01, 6.11883491212184e+01, 6.10803066618867e+01, 6.09747661470971e+01, 6.08723960047919e+01, 6.07738057484720e+01, 6.06795373566193e+01, 6.05900584280288e+01, 6.05057571493824e+01, 6.04269390669058e+01, 6.03538256186717e+01, 6.02865543564983e+01, 6.02251807645640e+01, 6.01696815639671e+01, 6.01199593769132e+01, 6.00758486098378e+01, 6.00371224008462e+01, 6.00035004631252e+01, 5.99746576425870e+01, 5.99502329954273e+01, 5.99298391802447e+01, 5.99130719507780e+01, 5.98995195301313e+01, 5.98887716465373e+01, 5.98804280150996e+01, 5.98741060602047e+01, 5.98694476898062e+01, 5.98661249556056e+01, 5.98638444620136e+01, 5.98623504210013e+01, 5.98614262885593e+01, 5.98608949601845e+01, 5.98606175461278e+01, 5.98604907904401e+01, 5.98604432395341e+01, 5.98604303044902e+01, 5.98604283953032e+01, 5.98604281448582e+01, 5.98604223097885e+01, 5.98603827763144e+01, 5.98602374468069e+01, 5.98598500495860e+01, 5.98590022295291e+01, 5.98573785116465e+01, 5.98545547242253e+01, 5.98499903885302e+01, 5.98430254906009e+01, 5.98328819472466e+01, 5.98186699623839e+01, 5.97993993409170e+01, 5.97739956853285e+01, 5.97413212456282e+01, 5.97002000277606e+01, 5.96494465914340e+01, 5.95878977892468e+01, 5.95144465203387e+01, 5.94280764016594e+01, 5.93278961103534e+01, 5.92131720386599e+01, 5.90833578498201e+01, 5.89381195538654e+01, 5.87773548571559e+01, 5.86012057903699e+01, 5.84100639796938e+01, 5.82045683657400e+01, 5.79855956426493e+01, 5.77542441205079e+01, 5.75118120431280e+01, 5.72597715742453e+01, 5.69997396845242e+01, 5.67334470536079e+01, 5.64627059016325e+01, 5.61893774539900e+01, 5.59153395853660e+01, 5.56424551206432e+01, 5.53725412894574e+01, 5.51073409008431e+01, 5.48484958676413e+01, 5.45975237136968e+01, 5.43557976114003e+01, 5.41245303286182e+01, 5.39047622482066e+01, 5.36973534092106e+01, 5.35029793492686e+01, 5.33221304254875e+01, 5.31551142559174e+01, 5.30020609386197e+01, 5.28629307463155e+01, 5.27375240399388e+01, 5.26254931792978e+01, 5.25263562264763e+01, 5.24395122367042e+01, 5.23642579149721e+01, 5.22998053889822e+01, 5.22453008146773e+01, 5.21998434938128e+01, 5.21625051476032e+01, 5.21323489598195e+01, 5.21084479799863e+01, 5.20899024654024e+01, 5.20758557421180e+01, 5.20655081817112e+01, 5.20581289239307e+01, 5.20530650252713e+01, 5.20497477794824e+01, 5.20476960359056e+01, 5.20465164323566e+01, 5.20459005571178e+01, 5.20456191550085e+01, 5.20455135907755e+01, 5.20454848746892e+01, 5.20454806362530e+01, 5.20454802620798e+01, 5.20454729303393e+01, 5.20454232567086e+01, 5.20452406510355e+01, 5.20447538901847e+01, 5.20436886200319e+01, 5.20416484774942e+01, 5.20381005812602e+01, 5.20323660462803e+01, 5.20236160686141e+01, 5.20108740018692e+01, 5.19930237011807e+01, 5.19688242422871e+01, 5.19369309293892e+01, 5.18959222858647e+01, 5.18443324795129e+01, 5.17806883761741e+01, 5.17035501546729e+01, 5.16115541696272e+01, 5.15034565387298e+01, 5.13781757823319e+01, 5.12348327799255e+01, 5.10727863503408e+01, 5.08916629204257e+01, 5.06913790162746e+01, 5.04721556693788e+01, 5.02345242379991e+01, 4.99793235521595e+01, 4.97076886520276e+01, 4.94210316743197e+01, 4.91210156468149e+01, 4.88095220996994e+01, 4.84886135289924e+01, 4.81604918786881e+01, 4.78274543479149e+01, 4.74918479537869e+01, 4.71560243496006e+01, 4.68222963734126e+01, 4.64928976645148e+01, 4.61699464428040e+01, 4.58554142310495e+01, 4.55510999590121e+01, 4.52586095681258e+01, 4.49793409721371e+01, 4.47144740421572e+01, 4.44649651764696e+01, 4.42315459755362e+01, 4.40147255529869e+01, 4.38147960542556e+01, 4.36318410084648e+01, 4.34657461930303e+01, 4.33162127358007e+01, 4.31827722118899e+01, 4.30648035100498e+01, 4.29615512466182e+01, 4.28721454948981e+01, 4.27956225760798e+01, 4.27309466268076e+01, 4.26770316210805e+01, 4.26327634838029e+01, 4.25970218939702e+01, 4.25687013415869e+01, 4.25467309784902e+01, 4.25300927935944e+01, 4.25178376512911e+01, 4.25090987604287e+01, 4.25031021916022e+01, 4.24991741319801e+01, 4.24967446574299e+01, 4.24953479075872e+01, 4.24946186657526e+01, 4.24942854662718e+01, 4.24941604711799e+01, 4.24941264694405e+01, 4.24941214508534e+01, 4.24941210483006e+01, 4.24941136191602e+01, 4.24940632856338e+01, 4.24938782541641e+01, 4.24933850274137e+01, 4.24923056087436e+01, 4.24902383775634e+01, 4.24866433973244e+01, 4.24808328249356e+01, 4.24719669825595e+01, 4.24590565277678e+01, 4.24409710125984e+01, 4.24164539526772e+01, 4.23841443317130e+01, 4.23426042439681e+01, 4.22903521307252e+01, 4.22259008036323e+01, 4.21477991803416e+01, 4.20546764030968e+01, 4.19452867898517e+01, 4.18185539031796e+01, 4.16736119372038e+01, 4.15098426357004e+01, 4.13269060766629e+01, 4.11247638907736e+01, 4.09036938117832e+01, 4.06642948618656e+01, 4.04074829209942e+01, 4.01344768781851e+01, 3.98467759782622e+01, 3.95461293338523e+01, 3.92344988550963e+01, 3.89140170584759e+01, 3.85869413581318e+01, 3.82556065238935e+01, 3.79223770072049e+01, 3.75896007763617e+01, 3.72595661498269e+01, 3.69344628642500e+01, 3.66163482767750e+01, 3.63071192181464e+01, 3.60084896381537e+01, 3.57219738703817e+01, 3.54488751230176e+01, 3.51902786842742e+01, 3.49470492992957e+01, 3.47198322020153e+01, 3.45090573411714e+01, 3.43149464023787e+01, 3.41375222851035e+01, 3.39766207398762e+01, 3.38319039068884e+01, 3.37028755235465e+01, 3.35888975864287e+01, 3.34892082622633e+01, 3.34029408422102e+01, 3.33291435230805e+01, 3.32667997781478e+01, 3.32148490500455e+01, 3.31722074614409e+01, 3.31377881995696e+01, 3.31105211930484e+01, 3.30893716689327e+01, 3.30733571599234e+01, 3.30615625304998e+01, 3.30531526099743e+01, 3.30473820618564e+01, 3.30436021824978e+01, 3.30412644058897e+01, 3.30399203920532e+01, 3.30392186886198e+01, 3.30388980729073e+01, 3.30387777986228e+01, 3.30387450810765e+01, 3.30387402520343e+01, 3.30387398795721e+01, 3.30387331913464e+01, 3.30386878776321e+01, 3.30385212996023e+01, 3.30380772635608e+01, 3.30371055006660e+01, 3.30352444551023e+01, 3.30320080707198e+01, 3.30267771848876e+01, 3.30187960438840e+01, 3.30071743440723e+01, 3.29908950738927e+01, 3.29688282796974e+01, 3.29397507009542e+01, 3.29023710167220e+01, 3.28553602181629e+01, 3.27973863779478e+01, 3.27271528381036e+01, 3.26434385989559e+01, 3.25451394826903e+01, 3.24313084867529e+01, 3.23011936551437e+01, 3.21542717962313e+01, 3.19902764740893e+01, 3.18092188979125e+01, 3.16114006226243e+01, 3.13974173363732e+01, 3.11681534237566e+01, 3.09247674304679e+01, 3.06686689885696e+01, 3.04014881664014e+01, 3.01250385602537e+01, 2.98412757252655e+01, 2.95522527310510e+01, 2.92600747065622e+01, 2.89668541976147e+01, 2.86746689981789e+01, 2.83855238453049e+01, 2.81013170142058e+01, 2.78238124532067e+01, 2.75546177022772e+01, 2.72951674857143e+01, 2.70467125915356e+01, 2.68103134649756e+01, 2.65868378530226e+01, 2.63769618294943e+01, 2.61811735851207e+01, 2.59997794603444e+01, 2.58329118072341e+01, 2.56805383727042e+01, 2.55424729857636e+01, 2.54183874001700e+01, 2.53078241886907e+01, 2.52102106073325e+01, 2.51248733502465e+01, 2.50510541020428e+01, 2.49879257675256e+01, 2.49346092227992e+01, 2.48901903896344e+01, 2.48537373903331e+01, 2.48243174966534e+01, 2.48010135474650e+01, 2.47829394795598e+01, 2.47692545981888e+01, 2.47591762117023e+01, 2.47519902705581e+01, 2.47470596862344e+01, 2.47438300600784e+01, 2.47418326242140e+01, 2.47406842831854e+01, 2.47400847416876e+01, 2.47398108053138e+01, 2.47397080421751e+01, 2.47396800880958e+01, 2.47396759621318e+01, 2.47396756497262e+01, 2.47396701154931e+01, 2.47396326202499e+01, 2.47394947837751e+01, 2.47391273623836e+01, 2.47383232698751e+01, 2.47367833373666e+01, 2.47341053844208e+01, 2.47297771074179e+01, 2.47231732137153e+01, 2.47135571394552e+01, 2.47000875816140e+01, 2.46818299475737e+01, 2.46577726754279e+01, 2.46268482045588e+01, 2.45879581808551e+01, 2.45400022702967e+01, 2.44819097389735e+01, 2.44126727518469e+01, 2.43313801652292e+01, 2.42372504595366e+01, 2.41296623993206e+01, 2.40081820331448e+01, 2.38725847658565e+01, 2.37228714495786e+01, 2.35592777350262e+01, 2.33822762775583e+01, 2.31925717693423e+01, 2.29910891320402e+01, 2.27789555171334e+01, 2.25574769957791e+01, 2.23281109640238e+01, 2.20924353468453e+01, 2.18521156761931e+01, 2.16088710734453e+01, 2.13644401141324e+01, 2.11205475097308e+01, 2.08788725070757e+01, 2.06410198619597e+01, 2.04084941617587e+01, 2.01826781287333e+01, 1.99648153248831e+01, 1.97559974189922e+01, 1.95571559051590e+01, 1.93690579254068e+01, 1.91923056842249e+01, 1.90273388674788e+01, 1.88744394867677e+01, 1.87337386406520e+01, 1.86052247867127e+01, 1.84887532258595e+01, 1.83840565933078e+01, 1.82907562191405e+01, 1.82083742629955e+01, 1.81363465446228e+01, 1.80740359894147e+01, 1.80207465904913e+01, 1.79757377609949e+01, 1.79382389158131e+01, 1.79074640844910e+01, 1.78826263199800e+01, 1.78629516344719e+01, 1.78476921672925e+01, 1.78361382739287e+01, 1.78276292225879e+01, 1.78215621973585e+01, 1.78173993361895e+01, 1.78146725774160e+01, 1.78129861490438e+01, 1.78120166077903e+01, 1.78115104162812e+01, 1.78112791323779e+01, 1.78111923696681e+01, 1.78111687680947e+01, 1.78111652845515e+01, 1.78111650230408e+01, 1.78111604201561e+01, 1.78111292349342e+01, 1.78110145947498e+01, 1.78107090060328e+01, 1.78100402322163e+01, 1.78087594488794e+01, 1.78065321512940e+01, 1.78029322338216e+01, 1.77974395957472e+01, 1.77894415523800e+01, 1.77782382410266e+01, 1.77630521060932e+01, 1.77430414240248e+01, 1.77173176864843e+01, 1.76849665008420e+01, 1.76450714955020e+01, 1.75967405420961e+01, 1.75391334388802e+01, 1.74714900544022e+01, 1.73931578239584e+01, 1.73036174395528e+01, 1.72025055903038e+01, 1.70896337023519e+01, 1.69650017950903e+01, 1.68288068040336e+01, 1.66814450000880e+01, 1.65235084330262e+01, 1.63557756131925e+01, 1.61791968926713e+01, 1.59948751975012e+01, 1.58040428912616e+01, 1.56080356254812e+01, 1.54082640696835e+01, 1.52061844297484e+01, 1.50032686672873e+01, 1.48009753244171e+01, 1.46007218284352e+01, 1.44038590866810e+01, 1.42116490737100e+01, 1.40252459594847e+01, 1.38456811374824e+01, 1.36738523028006e+01, 1.35105165242340e+01, 1.33562870720675e+01, 1.32116336213926e+01, 1.30768853583758e+01, 1.29522364756539e+01, 1.28377535478640e+01, 1.27333843195950e+01, 1.26389675038185e+01, 1.25542432668971e+01, 1.24788641556244e+01, 1.24124062936366e+01, 1.23543807327647e+01, 1.23042448858071e+01, 1.22614139894975e+01, 1.22252725507565e+01, 1.21951857178830e+01, 1.21705104945996e+01, 1.21506066830924e+01, 1.21348474071696e+01, 1.21226290333056e+01, 1.21133802803270e+01, 1.21065702918963e+01, 1.21017154428913e+01, 1.20983846631453e+01, 1.20962030903029e+01, 1.20948539067793e+01, 1.20940782716778e+01, 1.20936733235769e+01, 1.20934883002039e+01, 1.20934188916718e+01, 1.20934000108797e+01, 1.20933972241148e+01, 1.20933970155476e+01, 1.20933933530039e+01, 1.20933685387397e+01, 1.20932773190064e+01, 1.20930341616638e+01, 1.20925020227148e+01, 1.20914829319302e+01, 1.20897107838697e+01, 1.20868466809573e+01, 1.20824771288255e+01, 1.20761153286266e+01, 1.20672057434468e+01, 1.20551320332313e+01, 1.20392283522421e+01, 1.20187938839096e+01, 1.19931103510239e+01, 1.19614620882751e+01, 1.19231581059279e+01, 1.18775554176384e+01, 1.18240827643254e+01, 1.17622637534643e+01, 1.16917383634140e+01, 1.16122817484416e+01, 1.15238193320939e+01, 1.14264373000032e+01, 1.13203877976044e+01, 1.12060883960218e+01, 1.10841156956338e+01, 1.09551932697575e+01, 1.08201744834237e+01, 1.06800210245832e+01, 1.05357782283454e+01, 1.03885484348990e+01, 1.02394636830431e+01, 1.00896589995809e+01, 9.94024740831988e+00, 9.79229757042342e+00, 9.64681470775093e+00, 9.50472518351206e+00, 9.36686484957052e+00, 9.23397104100932e+00, 9.10667792177499e+00, 8.98551476692042e+00, 8.87090670552987e+00, 8.76317743611532e+00, 8.66255345191962e+00, 8.56916936495004e+00, 8.48307398337548e+00, 8.40423686784339e+00, 8.33255516139479e+00, 8.26786055033712e+00, 8.20992626686256e+00, 8.15847408689120e+00, 8.11318130796558e+00, 8.07368771198857e+00, 8.03960252654348e+00, 8.01051139718036e+00, 7.98598337248628e+00, 7.96557788547712e+00, 7.94885169079019e+00, 7.93536568966842e+00, 7.92469154648340e+00, 7.91641797440960e+00, 7.91015654667481e+00, 7.90554687615125e+00, 7.90226100198490e+00, 7.90000682880453e+00, 7.89853048221730e+00, 7.89761747319186e+00, 7.89709260199979e+00, 7.89681857725439e+00, 7.89669337429940e+00, 7.89664640654496e+00, 7.89663363019551e+00, 7.89663174443375e+00, 7.89663160324670e+00, 7.89662912322537e+00, 7.89661232072631e+00, 7.89655055311656e+00, 7.89638590453194e+00, 7.89602558114439e+00, 7.89533554179116e+00, 7.89413562966281e+00, 7.89219644622085e+00, 7.88923817946119e+00, 7.88493156295428e+00, 7.87890109703374e+00, 7.87073060617075e+00, 7.85997113453895e+00, 7.84615109381853e+00, 7.82878847408770e+00, 7.80740481341058e+00, 7.78154050052005e+00, 7.75077086674060e+00, 7.71472241925424e+00, 7.67308849052102e+00, 7.62564354039136e+00, 7.57225535830776e+00, 7.51289447903510e+00, 7.44764024695402e+00, 7.37668313491060e+00, 7.30032313130776e+00, 7.21896423576565e+00, 7.13310532862053e+00, 7.04332788221298e+00, 6.95028114483348e+00, 6.85466553926383e+00, 6.75721507164930e+00, 6.65867954405542e+00, 6.55980731183668e+00, 6.46132923449276e+00, 6.36394434698008e+00, 6.26830763843257e+00, 6.17502017724662e+00, 6.08462167518404e+00, 5.99758544769968e+00, 5.91431561154120e+00, 5.83514627080405e+00, 5.76034238363246e+00, 5.69010197490901e+00, 5.62455936325537e+00, 5.56378909806007e+00, 5.50781034668368e+00, 5.45659152554931e+00, 5.41005502430229e+00, 5.36808192396433e+00, 5.33051665421807e+00, 5.29717156958941e+00, 5.26783144868385e+00, 5.24225793506582e+00, 5.22019394370750e+00, 5.20136805434858e+00, 5.18549890393871e+00, 5.17229957603384e+00, 5.16148196714504e+00, 5.15276109027101e+00, 5.14585925595891e+00, 5.14051005305780e+00, 5.13646203663840e+00, 5.13348202096369e+00, 5.13135787219694e+00, 5.12990069955660e+00, 5.12894635512193e+00, 5.12835617107059e+00, 5.12801688777694e+00, 5.12783975536187e+00, 5.12775882302104e+00, 5.12772846265514e+00, 5.12772020391358e+00, 5.12771898494113e+00, 5.12771889343450e+00, 5.12771728284127e+00, 5.12770637084281e+00, 5.12766625729729e+00, 5.12755933019719e+00, 5.12732532734242e+00, 5.12687720108003e+00, 5.12609795956431e+00, 5.12483864257263e+00, 5.12291757157038e+00, 5.12012099096676e+00, 5.11620518840117e+00, 5.11090014438913e+00, 5.10391471399951e+00, 5.09494328451127e+00, 5.08367378361888e+00, 5.06979683488930e+00, 5.05301577502182e+00, 5.03305716725089e+00, 5.00968137487741e+00, 4.98269270728046e+00, 4.95194862657335e+00, 4.91736751355826e+00, 4.87893454112322e+00, 4.83670529190583e+00, 4.79080688023117e+00, 4.74143648643015e+00, 4.68885737104319e+00, 4.63339259118590e+00, 4.57541677562110e+00, 4.51534641555857e+00, 4.45362918637230e+00, 4.39073282879784e+00, 4.32713409030410e+00, 4.26330816660318e+00, 4.19971900077591e+00, 4.13681070476912e+00, 4.07500027504952e+00, 4.01467168831837e+00, 3.95617138885849e+00, 3.89980511833626e+00, 3.84583599208764e+00, 3.79448369252972e+00, 3.74592462938660e+00, 3.70029290673121e+00, 3.65768193709872e+00, 3.61814655157985e+00, 3.58170547008480e+00, 3.54834401590696e+00, 3.51801698125820e+00, 3.49065157359270e+00, 3.46615039448290e+00, 3.44439442206881e+00, 3.42524598355447e+00, 3.40855171515772e+00, 3.39414551300816e+00, 3.38185147977339e+00, 3.37148686864273e+00, 3.36286501936395e+00, 3.35579827121526e+00, 3.35010082619750e+00, 3.34559152358955e+00, 3.34209647562260e+00, 3.33945150467343e+00, 3.33750431619741e+00, 3.33611633953485e+00, 3.33516417131204e+00, 3.33454056360843e+00, 3.33415491112812e+00, 3.33393320763431e+00, 3.33381746086339e+00, 3.33376457576328e+00, 3.33374473682631e+00, 3.33373934016291e+00, 3.33373854362700e+00, 3.33373848357746e+00, 3.33373742326287e+00, 3.33373023948041e+00, 3.33370383120889e+00, 3.33363343701170e+00, 3.33347938386873e+00, 3.33318436453696e+00, 3.33267135744786e+00, 3.33184229290127e+00, 3.33057755714496e+00, 3.32873641089830e+00, 3.32615837975234e+00, 3.32266564936155e+00, 3.31806646734251e+00, 3.31215951574804e+00, 3.30473917316451e+00, 3.29560153526331e+00, 3.28455100966025e+00, 3.27140724908060e+00, 3.25601214100118e+00, 3.23823653761589e+00, 3.21798639253100e+00, 3.19520797453180e+00, 3.16989185687035e+00, 3.14207543322895e+00, 3.11184378648337e+00, 3.07932882852678e+00, 3.04470673129392e+00, 3.00819377180831e+00, 2.97004080818986e+00, 2.93052668049202e+00, 2.88995088319653e+00, 2.84862588106905e+00, 2.80686943587063e+00, 2.76499728027828e+00, 2.72331642217564e+00, 2.68211929409374e+00, 2.64167888685784e+00, 2.60224493119668e+00, 2.56404112292736e+00, 2.52726333132296e+00, 2.49207868926431e+00, 2.45862543851285e+00, 2.42701339284257e+00, 2.39732488345970e+00, 2.36961606203835e+00, 2.34391845357040e+00, 2.32024067111116e+00, 2.29857022497699e+00, 2.27887537824009e+00, 2.26110701730828e+00, 2.24520052032635e+00, 2.23107761681981e+00, 2.21864823939795e+00, 2.20781237256742e+00, 2.19846190499409e+00, 2.19048249016853e+00, 2.18375541672330e+00, 2.17815948404019e+00, 2.17357287178881e+00, 2.16987498426518e+00, 2.16694824254272e+00, 2.16467979025413e+00, 2.16296307303339e+00, 2.16169924794349e+00, 2.16079837814856e+00, 2.16018037001055e+00, 2.15977561481380e+00, 2.15952530528296e+00, 2.15938140754248e+00, 2.15930628151543e+00, 2.15927195616975e+00, 2.15925907960523e+00, 2.15925557687292e+00, 2.15925505987716e+00, 2.15925502068959e+00, 2.15925432592837e+00, 2.15924961882205e+00, 2.15923231504993e+00, 2.15918618992005e+00, 2.15908524809664e+00, 2.15889193969071e+00, 2.15855579747266e+00, 2.15801256327785e+00, 2.15718386372451e+00, 2.15597748551215e+00, 2.15428828799165e+00, 2.15199977451524e+00, 2.14898632354781e+00, 2.14511605526820e+00, 2.14025427960067e+00, 2.13426743825173e+00, 2.12702741821538e+00, 2.11841608005638e+00, 2.10832981449470e+00, 2.09668391918539e+00, 2.08341657777982e+00, 2.06849222832798e+00, 2.05190412947593e+00, 2.03367597056575e+00, 2.01386242343386e+00, 1.99254859523081e+00, 1.96984840722075e+00, 1.94590198778841e+00, 1.92087222252307e+00, 1.89494064521143e+00, 1.86830287777043e+00, 1.84116383388594e+00, 1.81373289199456e+00, 1.78621922165564e+00, 1.75882741767660e+00, 1.73175356300917e+00, 1.70518180810394e+00, 1.67928152355169e+00, 1.65420505556523e+00, 1.63008609020208e+00, 1.60703861164140e+00, 1.58515642168988e+00, 1.56451317174258e+00, 1.54516284493065e+00, 1.52714061586195e+00, 1.51046400907987e+00, 1.49513427583279e+00, 1.48113791218242e+00, 1.46844824949287e+00, 1.45702705994653e+00, 1.44682613352605e+00, 1.43778879731191e+00, 1.42985136148354e+00, 1.42294448786097e+00, 1.41699448535579e+00, 1.41192454190101e+00, 1.40765590427259e+00, 1.40410901599759e+00, 1.40120461980268e+00, 1.39886482550029e+00, 1.39701413763765e+00, 1.39558043048085e+00, 1.39449585177311e+00, 1.39369763190680e+00, 1.39312877225047e+00, 1.39273858575426e+00, 1.39248306477250e+00, 1.39232505521419e+00, 1.39223422235484e+00, 1.39218680143165e+00, 1.39216513488040e+00, 1.39215700708345e+00, 1.39215479613264e+00, 1.39215446980119e+00, 1.39215444490877e+00, 1.39215400151715e+00, 1.39215099747748e+00, 1.39213995436634e+00, 1.39211051791986e+00, 1.39204609917594e+00, 1.39192273800259e+00, 1.39170823826390e+00, 1.39136162049560e+00, 1.39083293385723e+00, 1.39006346373207e+00, 1.38898636395103e+00, 1.38752773246820e+00, 1.38560813577617e+00, 1.38314457004130e+00, 1.38005282590927e+00, 1.37625019976767e+00, 1.37165846815736e+00, 1.36620701580667e+00, 1.35983598374339e+00, 1.35249928476137e+00, 1.34416732187702e+00, 1.33482924373831e+00, 1.32449458109709e+00, 1.31319413140298e+00, 1.30097999423272e+00, 1.28792470732775e+00, 1.27411948891810e+00, 1.25967165295283e+00, 1.24470132486352e+00, 1.22933764063686e+00, 1.21371465483789e+00, 1.19796720766167e+00, 1.18222700234487e+00, 1.16661912025599e+00, 1.15125915335064e+00, 1.13625106815882e+00, 1.12168584119990e+00, 1.10764083350391e+00, 1.09417981187505e+00, 1.08135348390514e+00, 1.06920039551667e+00, 1.05774804260618e+00, 1.04701406743024e+00, 1.03700743928447e+00, 1.02772955131974e+00, 1.01917519580784e+00, 1.01133340544952e+00, 1.00418816690539e+00, 9.97719024598915e-01, 9.91901598905763e-01, 9.86708044444491e-01, 9.82107472712251e-01, 9.78066359959513e-01, 9.74548956884796e-01, 9.71517712073108e-01, 9.68933716475043e-01, 9.66757171826956e-01, 9.64947881837816e-01, 9.63465761257396e-01, 9.62271354627596e-01, 9.61326353654329e-01, 9.60594099804574e-01, 9.60040057032749e-01, 9.59632238595255e-01, 9.59341571828077e-01, 9.59142185620145e-01, 9.59011607139267e-01, 9.58930857111284e-01, 9.58884436491878e-01, 9.58860201507107e-01, 9.58849128516370e-01, 9.58844974683387e-01, 9.58843844742759e-01, 9.58843677966001e-01, 9.58843665154554e-01, 9.58843435776394e-01, 9.58841881707109e-01, 9.58836168807444e-01, 9.58820940494370e-01, 9.58787614616249e-01, 9.58723795068835e-01, 9.58612823274542e-01, 9.58433492681785e-01, 9.58159947051406e-01, 9.57761780196179e-01, 9.57204350652106e-01, 9.56449319079610e-01, 9.55455408664982e-01, 9.54179379225856e-01, 9.52577194198540e-01, 9.50605346660922e-01, 9.48222296929536e-01, 9.45389961414513e-01, 9.42075182016481e-01, 9.38251099228772e-01, 9.33898351904716e-01, 9.29006033444303e-01, 9.23572348122804e-01, 9.17604931466556e-01, 9.11120822877820e-01, 9.04146104133901e-01, 8.96715240578224e-01, 8.88870179724940e-01, 8.80659272502790e-01, 8.72136084761486e-01, 8.63358161740500e-01, 8.54385797966707e-01, 8.45280852153330e-01, 8.36105633729808e-01, 8.26921876639425e-01, 8.17789808058343e-01, 8.08767314833631e-01, 7.99909208117008e-01, 7.91266585948968e-01, 7.82886293486179e-01, 7.74810480471260e-01, 7.67076255056693e-01, 7.59715432167404e-01, 7.52754373379283e-01, 7.46213914048226e-01, 7.40109372377347e-01, 7.34450634408283e-01, 7.29242308617259e-01, 7.24483943844938e-01, 7.20170304589850e-01, 7.16291698110510e-01, 7.12834348175534e-01, 7.09780810554953e-01, 7.07110425372086e-01, 7.04799801184976e-01, 7.02823325127611e-01, 7.01153692641007e-01, 6.99762449319195e-01, 6.98620536274362e-01, 6.97698829294552e-01, 6.96968661049830e-01, 6.96402314826436e-01, 6.95973477853407e-01, 6.95657642338723e-01, 6.95432442929523e-01, 6.95277920491556e-01, 6.95176703864732e-01, 6.95114103541187e-01, 6.95078113934478e-01, 6.95059323931538e-01, 6.95050738581791e-01, 6.95047517913878e-01, 6.95046641813470e-01, 6.95046512502856e-01, 6.95046502493689e-01, 6.95046322300819e-01, 6.95045101467564e-01, 6.95040613555058e-01, 6.95028650425442e-01, 6.95002469395668e-01, 6.94952329540818e-01, 6.94865135752970e-01, 6.94724206979203e-01, 6.94509182522722e-01, 6.94196077321195e-01, 6.93757493613084e-01, 6.93162992258578e-01, 6.92379622112099e-01, 6.91372600225925e-01, 6.90106129369887e-01, 6.88544332576761e-01, 6.86652277504372e-01, 6.84397056817588e-01, 6.81748885120207e-01, 6.78682168840758e-01, 6.75176503502101e-01, 6.71217553463855e-01, 6.66797772785117e-01, 6.61916932300493e-01, 6.56582427008232e-01, 6.50809348819180e-01, 6.44620321759866e-01, 6.38045108884278e-01, 6.31120011435044e-01, 6.23887090311399e-01, 6.16393246959774e-01, 6.08689204978809e-01, 6.00828434912671e-01, 5.92866063081388e-01, 5.84857801324091e-01, 5.76858928851266e-01, 5.68923350759498e-01, 5.61102750882732e-01, 5.53445850164422e-01, 5.45997776087653e-01, 5.38799544152141e-01, 5.31887649004105e-01, 5.25293760524733e-01, 5.19044518788924e-01, 5.13161421101927e-01, 5.07660794099948e-01, 5.02553843983364e-01, 4.97846778205336e-01, 4.93540992268358e-01, 4.89633315621624e-01, 4.86116310951040e-01, 4.82978621369298e-01, 4.80205360106034e-01, 4.77778537230996e-01, 4.75677517691530e-01, 4.73879504496004e-01, 4.72360040239280e-01, 4.71093519379409e-01, 4.70053702798658e-01, 4.69214225302139e-01, 4.68549085927329e-01, 4.68033110365887e-01, 4.67642374546472e-01, 4.67354578584228e-01, 4.67149360937761e-01, 4.67008543762461e-01, 4.66916302100898e-01, 4.66859251661724e-01, 4.66826452418800e-01, 4.66809327995371e-01, 4.66801503644874e-01, 4.66798568451154e-01, 4.66797770006387e-01, 4.66797652157581e-01, 4.66797642961644e-01, 4.66797476453898e-01, 4.66796348339152e-01, 4.66792201267616e-01, 4.66781146679004e-01, 4.66756953888130e-01, 4.66710621439013e-01, 4.66630047793188e-01, 4.66499816331002e-01, 4.66301107222918e-01, 4.66011745924142e-01, 4.65606394785567e-01, 4.65056890480838e-01, 4.64332725570031e-01, 4.63401667546876e-01, 4.62230503219971e-01, 4.60785890424170e-01, 4.59035293121445e-01, 4.56947970312406e-01, 4.54495984301417e-01, 4.51655190243379e-01, 4.48406167028530e-01, 4.44735049820224e-01, 4.40634227172218e-01, 4.36102870618988e-01, 4.31147271712170e-01, 4.25780970188363e-01, 4.20024666645531e-01, 4.13905923022298e-01, 4.07458663582396e-01, 4.00722497374908e-01, 3.93741889818777e-01, 3.86565215908739e-01, 3.79243730510318e-01, 3.71830492399980e-01, 3.64379278291594e-01, 3.56943521270679e-01, 3.49575305025915e-01, 3.42324441195181e-01, 3.35237652211326e-01, 3.28357876446510e-01, 3.21723706475233e-01, 3.15368965219292e-01, 3.09322418955006e-01, 3.03607621004360e-01, 2.98242875701575e-01, 2.93241309142395e-01, 2.88611031387660e-01, 2.84355374183150e-01, 2.80473188740423e-01, 2.76959189471420e-01, 2.73804331505502e-01, 2.70996212038760e-01, 2.68519487786598e-01, 2.66356302787656e-01, 2.64486722350192e-01, 2.62889169928322e-01, 2.61540864120384e-01, 2.60418252821417e-01, 2.59497440926381e-01, 2.58754607005244e-01, 2.58166403227357e-01, 2.57710331684223e-01, 2.57365089330605e-01, 2.57110873194129e-01, 2.56929637419958e-01, 2.56805294194180e-01, 2.56723851652680e-01, 2.56673483494374e-01, 2.56644527094366e-01, 2.56629409328021e-01, 2.56622501914919e-01, 2.56619910708899e-01, 2.56619205838327e-01, 2.56619101800908e-01, 2.56619093613629e-01, 2.56618944484103e-01, 2.56617934110025e-01, 2.56614219874450e-01, 2.56604319135484e-01, 2.56582651861052e-01, 2.56541157399049e-01, 2.56469001188982e-01, 2.56352385651951e-01, 2.56174477654390e-01, 2.55915462730343e-01, 2.55552733536511e-01, 2.55061216673987e-01, 2.54413837932184e-01, 2.53582121148483e-01, 2.52536910272767e-01, 2.51249198032745e-01, 2.49691038090717e-01, 2.47836511157232e-01, 2.45662709676658e-01, 2.43150700971789e-01, 2.40286425688602e-01, 2.37061487520592e-01, 2.33473791899350e-01, 2.29527995824306e-01, 2.25235738262448e-01, 2.20615630325888e-01, 2.15692996238985e-01, 2.10499369219754e-01, 2.05071759930640e-01, 1.99451728093351e-01, 1.93684299179878e-01, 1.87816776820561e-01, 1.81897506924102e-01, 1.75974650973131e-01, 1.70095023391625e-01, 1.64303041523887e-01, 1.58639827254895e-01, 1.53142487588489e-01, 1.47843588732663e-01, 1.42770825613013e-01, 1.37946877335807e-01, 1.33389429805810e-01, 1.29111340011193e-01, 1.25120912620140e-01, 1.21422258377414e-01, 1.18015704982987e-01, 1.14898234160570e-01, 1.12063922887749e-01, 1.09504371681879e-01, 1.07209107898889e-01, 1.05165956788568e-01, 1.03361377257094e-01, 1.01780762718960e-01, 1.00408709977238e-01, 9.92292607302261e-02, 9.82261211023712e-02, 9.73828646183676e-02, 9.66831233979519e-02, 9.61107711825306e-02, 9.56501002720316e-02, 9.52859927210099e-02, 9.50040843889894e-02, 9.47909188320608e-02, 9.46340867142226e-02, 9.45223455323614e-02, 9.44457140770680e-02, 9.43955362300173e-02, 9.43645094166996e-02, 9.43466742262231e-02, 9.43373632721176e-02, 9.43331091624486e-02, 9.43315133220466e-02, 9.43310792169285e-02, 9.43310151440073e-02, 9.43310100571182e-02, 9.43309168333357e-02, 9.43302852298160e-02, 9.43279634082543e-02, 9.43217744362386e-02, 9.43082307926346e-02, 9.42822960466145e-02, 9.42372045429778e-02, 9.41643494879472e-02, 9.40532492544222e-02, 9.38916008367272e-02, 9.36654279609357e-02, 9.33593291624462e-02, 9.29568279709443e-02, 9.24408230659211e-02, 9.17941308786041e-02, 9.10001067805482e-02, 9.00433240673522e-02, 8.89102829630346e-02, 8.75901155487853e-02, 8.60752476747990e-02, 8.43619763748649e-02, 8.24509217963438e-02, 8.03473166821291e-02, 7.80611041462294e-02, 7.56068255926152e-02, 7.30032943933807e-02, 7.02730661955329e-02, 6.74417319748386e-02, 6.45370735732627e-02, 6.15881319031075e-02, 5.86242440536253e-02, 5.56741064965749e-02, 5.27649173931053e-02, 4.99216422532172e-02, 4.71664350489647e-02, 4.45182328579961e-02, 4.19925278636200e-02, 3.96013075801847e-02, 3.73531436897352e-02, 3.52534025892624e-02, 3.33045468811406e-02, 3.15064963734812e-02, 2.98570191465899e-02, 2.83521271491622e-02, 2.69864558330468e-02, 2.57536127983187e-02, 2.46464857204103e-02, 2.36575045597149e-02, 2.27788569724083e-02, 2.20026588588524e-02, 2.13210841238407e-02, 2.07264590720726e-02, 2.02113275509888e-02, 1.97684931190096e-02, 1.93910442858041e-02, 1.90723683517859e-02, 1.88061586519481e-02, 1.85864191499832e-02, 1.84074693864886e-02, 1.82639518009893e-02, 1.81508424669232e-02, 1.80634653407168e-02, 1.79975092805122e-02, 1.79490463793925e-02, 1.79145496288342e-02, 1.78909076133209e-02, 1.78754338566449e-02, 1.78658685956989e-02, 1.78603711276608e-02, 1.78575014200888e-02, 1.78561903334185e-02, 1.78556985167759e-02, 1.78555647323435e-02, 1.78555449860980e-02, 1.78555434040462e-02, 1.78555142302983e-02, 1.78553165747815e-02, 1.78545899862144e-02, 1.78526532754901e-02, 1.78484153776211e-02, 1.78403013889152e-02, 1.78261976415754e-02, 1.78034198073078e-02, 1.77687081182327e-02, 1.77182537379466e-02, 1.76477599603047e-02, 1.75525410917601e-02, 1.74276604444259e-02, 1.72681067312652e-02, 1.70690053083007e-02, 1.68258572669728e-02, 1.65347956039525e-02, 1.61928439982120e-02, 1.57981606347093e-02, 1.53502476197289e-02, 1.48501063885718e-02, 1.43003215135664e-02, 1.37050596239732e-02, 1.30699765442941e-02, 1.24020336667670e-02, 1.17092330935051e-02, 1.10002891017231e-02, 1.02842598727122e-02, 9.57016723934601e-03, 8.86663288032707e-03, 8.18155684637389e-03, 7.52185897815139e-03, 6.89329651917348e-03, 6.30036313906857e-03, 5.74626679579637e-03, 5.23297734164174e-03, 4.76133016459102e-03, 4.33116972184422e-03, 3.94151647071084e-03, 3.59074206263443e-03, 3.27674020930857e-03, 2.99708378663865e-03, 2.74916200731212e-03, 2.53029447799236e-03, 2.33782142291972e-03, 2.16917119566773e-03, 2.02190741553920e-03, 1.89375873153352e-03, 1.78263444689009e-03, 1.68662915391593e-03, 1.60401924637910e-03, 1.53325378992264e-03, 1.47294181049062e-03, 1.42183765159842e-03, 1.37882568235170e-03, 1.34290531851633e-03, 1.31317705169209e-03, 1.28882996019541e-03, 1.26913099107639e-03, 1.25341614835861e-03, 1.24108358900088e-03, 1.23158851243232e-03, 1.22443962971630e-03, 1.21919691503791e-03, 1.21547028057023e-03, 1.21291877763952e-03, 1.21124991859745e-03, 1.21021873401617e-03, 1.20962622992238e-03, 1.20931698353388e-03, 1.20917570778545e-03, 1.20912271385265e-03, 1.20910829854637e-03, 1.20910617089010e-03, 1.20910599881796e-03, 1.20910280570286e-03, 1.20908117207819e-03, 1.20900164738913e-03, 1.20878968552773e-03, 1.20832592254061e-03, 1.20743818369107e-03, 1.20589572053120e-03, 1.20340623632842e-03, 1.19961631246923e-03, 1.19411587474616e-03, 1.18644730383011e-03, 1.17611966554421e-03, 1.16262827964420e-03, 1.14547944364318e-03, 1.12421958072417e-03, 1.09846742414277e-03, 1.06794715212999e-03, 1.03251975752272e-03, 9.92209503912989e-04, 9.47222221952870e-04, 8.97952541436098e-04, 8.44977981650175e-04, 7.89039090456356e-04, 7.31006386047284e-04, 6.71836483505217e-04, 6.12521207891348e-04, 5.54034445862328e-04, 4.97281782585901e-04, 4.43057537353297e-04, 3.92012727964186e-04, 3.44635985534580e-04, 3.01247790133630e-04, 2.62006845993312e-04, 2.26926206494227e-04, 1.95896134123075e-04, 1.68710623995594e-04, 1.45094694651959e-04, 1.24729897635234e-04, 1.07276353681129e-04, 9.23907300673932e-05, 7.97400949323901e-05, 6.90116139949737e-05, 5.99184796925556e-05, 5.22030942633478e-05, 4.56383808042915e-05, 4.00274350553786e-05, 3.52017921656511e-05, 3.10192261565589e-05, 2.73618000221632e-05, 2.41338769342292e-05, 2.12595735646914e-05, 1.86798668835308e-05, 1.63498831966535e-05, 1.42364605224543e-05, 1.23157868330281e-05, 1.05711007805098e-05, 8.99062783332738e-06, 7.56586737082898e-06, 6.29022728304562e-06, 5.15797039213339e-06, 4.16344743570107e-06, 3.30058671238476e-06, 2.56259348674348e-06, 1.94180720569648e-06, 1.42967282101836e-06, 1.01679402320087e-06, 6.93045080879529e-07, 4.47723252250398e-07, -2.13162820728030e-14, -1.42108547152020e-14, -2.13162820728030e-14, -7.10542735760100e-15, -7.10542735760100e-15, -7.10542735760100e-15],
[ 6.40000000000000e+01, 6.39999999341265e+01, 6.39999978971863e+01, 6.39999840965939e+01, 6.39999333638122e+01, 6.39997981261998e+01, 6.39995021485813e+01, 6.39989352646423e+01, 6.39979492879023e+01, 6.39963552556227e+01, 6.39939221178738e+01, 6.39903769397656e+01, 6.39854066394495e+01, 6.39786612399918e+01, 6.39697585716512e+01, 6.39582903242556e+01, 6.39438293185683e+01, 6.39259378412862e+01, 6.39041768701769e+01, 6.38781160023908e+01, 6.38473438878418e+01, 6.38114789579414e+01, 6.37701802252518e+01, 6.37231579101050e+01, 6.36701836259056e+01, 6.36110998279375e+01, 6.35458282056085e+01, 6.34743766816578e+01, 6.33968446812382e+01, 6.33134263556872e+01, 6.32244114947497e+01, 6.31301839380017e+01, 6.30312173979962e+01, 6.29280687267651e+01, 6.28213687831458e+01, 6.27118111787822e+01, 6.26001392838271e+01, 6.24871319499047e+01, 6.23735884519668e+01, 6.22603131607807e+01, 6.21481004364563e+01, 6.20377201864186e+01, 6.19299044662475e+01, 6.18253354270763e+01, 6.17246348363282e+01, 6.16283553255219e+01, 6.15369734537633e+01, 6.14508846204561e+01, 6.13703998160722e+01, 6.12957441646146e+01, 6.12270571840108e+01, 6.11643946690954e+01, 6.11077320841516e+01, 6.10569693365410e+01, 6.10119367885654e+01, 6.09724023506633e+01, 6.09380794851009e+01, 6.09086359356423e+01, 6.08837029857913e+01, 6.08628850368406e+01, 6.08457692880797e+01, 6.08319352960927e+01, 6.08209641891072e+01, 6.08124473167002e+01, 6.08059941255068e+01, 6.08012390683136e+01, 6.07978473771143e+01, 6.07955195600873e+01, 6.07939945173309e+01, 6.07930512095702e+01, 6.07925088566113e+01, 6.07922256865308e+01, 6.07920963008233e+01, 6.07920477632988e+01, 6.07920345598700e+01, 6.07920326110700e+01, 6.07920323554285e+01, 6.07920263992890e+01, 6.07919860455491e+01, 6.07918377006599e+01, 6.07914422655567e+01, 6.07905768548837e+01, 6.07889194492219e+01, 6.07860370795798e+01, 6.07813780624174e+01, 6.07742687104104e+01, 6.07639148388063e+01, 6.07494082689451e+01, 6.07297383987501e+01, 6.07038087645276e+01, 6.06704583598237e+01, 6.06284873068190e+01, 6.05766862963536e+01, 6.05138690280407e+01, 6.04389066978193e+01, 6.03507634053501e+01, 6.02485312006410e+01, 6.01314633759987e+01, 5.99990045583912e+01, 5.98508161932462e+01, 5.96867961549971e+01, 5.95070914827794e+01, 5.93121036132209e+01, 5.91024859342632e+01, 5.88791339599936e+01, 5.86431688588726e+01, 5.83959153911547e+01, 5.81388754807658e+01, 5.78736986520147e+01, 5.76021504312297e+01, 5.73260796084126e+01, 5.70473850483783e+01, 5.67679825984400e+01, 5.64897725933875e+01, 5.62146084999256e+01, 5.59442673269847e+01, 5.56804224925959e+01, 5.54246198273482e+01, 5.51782572837476e+01, 5.49425687237191e+01, 5.47186119173857e+01, 5.45072606600787e+01, 5.43092007452693e+01, 5.41249294385095e+01, 5.39547580761087e+01, 5.37988174401376e+01, 5.36570656115093e+01, 5.35292980526971e+01, 5.34151597070155e+01, 5.33141589167514e+01, 5.32256829583239e+01, 5.31490149728723e+01, 5.30833520401173e+01, 5.30278241067826e+01, 5.29815134424974e+01, 5.29434742596208e+01, 5.29127521021171e+01, 5.28884025855452e+01, 5.28695090582396e+01, 5.28551987553570e+01, 5.28446570346309e+01, 5.28371393166375e+01, 5.28319804033996e+01, 5.28286009163808e+01, 5.28265106763626e+01, 5.28253089402476e+01, 5.28246815095435e+01, 5.28243948275846e+01, 5.28242872826878e+01, 5.28242580278120e+01, 5.28242537098516e+01, 5.28242533286578e+01, 5.28242458593546e+01, 5.28241952537157e+01, 5.28240092218798e+01, 5.28235133281230e+01, 5.28224280707571e+01, 5.28203496500402e+01, 5.28167351868685e+01, 5.28108930596928e+01, 5.28019789166308e+01, 5.27889977920386e+01, 5.27708126087657e+01, 5.27461591756725e+01, 5.27136675923343e+01, 5.26718897488984e+01, 5.26193323618990e+01, 5.25544947240428e+01, 5.24759100803330e+01, 5.23821892924910e+01, 5.22720652410488e+01, 5.21444362649000e+01, 5.19984068761029e+01, 5.18333240328539e+01, 5.16488074150627e+01, 5.14447724188852e+01, 5.12214449450241e+01, 5.09793674605520e+01, 5.07193962171301e+01, 5.04426898661192e+01, 5.01506899978869e+01, 4.98450943493813e+01, 4.95278235941540e+01, 4.92009827837805e+01, 4.88668186692051e+01, 4.85276742905518e+01, 4.81859423550188e+01, 4.78440189827535e+01, 4.75042593545403e+01, 4.71689366291642e+01, 4.68402052271272e+01, 4.65200692392296e+01, 4.62103563632141e+01, 4.59126974467616e+01, 4.56285114552054e+01, 4.53589955039554e+01, 4.51051194980850e+01, 4.48676248915176e+01, 4.46470270960998e+01, 4.44436211161810e+01, 4.42574900398663e+01, 4.40885160714625e+01, 4.39363938332131e+01, 4.38006456945518e+01, 4.36806389026973e+01, 4.35756042898974e+01, 4.34846563212157e+01, 4.34068142241584e+01, 4.33410239098176e+01, 4.32861803573002e+01, 4.32411500923661e+01, 4.32047933513911e+01, 4.31759854874232e+01, 4.31536371508117e+01, 4.31367127670280e+01, 4.31242468425795e+01, 4.31153576589963e+01, 4.31092579659446e+01, 4.31052623571640e+01, 4.31027911049673e+01, 4.31013703367245e+01, 4.31006285550038e+01, 4.31002896259188e+01, 4.31001624814503e+01, 4.31001278950280e+01, 4.31001227901429e+01, 4.31001223806680e+01, 4.31001148237784e+01, 4.31000636247333e+01, 4.30998754115261e+01, 4.30993737034395e+01, 4.30982757236125e+01, 4.30961729457490e+01, 4.30925161500068e+01, 4.30866056686126e+01, 4.30775873921729e+01, 4.30644549806186e+01, 4.30460585748374e+01, 4.30211201327590e+01, 4.29882553143598e+01, 4.29460016133730e+01, 4.28928521824478e+01, 4.28272945306002e+01, 4.27478529995158e+01, 4.26531336660185e+01, 4.25418700931731e+01, 4.24129681855385e+01, 4.22655483178347e+01, 4.20989829194784e+01, 4.19129278211899e+01, 4.17073459045356e+01, 4.14825219287484e+01, 4.12390678171878e+01, 4.09779181349360e+01, 4.07003159420375e+01, 4.04077896296230e+01, 4.01021217137055e+01, 3.97853108614435e+01, 3.94595286556761e+01, 3.91270727686258e+01, 3.87903183140798e+01, 3.84516691688223e+01, 3.81135109806514e+01, 3.77781673963698e+01, 3.74478607494806e+01, 3.71246780707232e+01, 3.68105428758613e+01, 3.65071928049686e+01, 3.62161628867606e+01, 3.59387740063046e+01, 3.56761260621942e+01, 3.54290952863817e+01, 3.51983352346326e+01, 3.49842810097774e+01, 3.47871563350039e+01, 3.46069831417865e+01, 3.44435933750971e+01, 3.42966427487201e+01, 3.41656262073530e+01, 3.40498948699472e+01, 3.39486742393625e+01, 3.38610834651231e+01, 3.37861554374110e+01, 3.37228574709596e+01, 3.36701123082022e+01, 3.36268191344609e+01, 3.35918742580581e+01, 3.35641910699678e+01, 3.35427188664286e+01, 3.35264600992037e+01, 3.35144856165820e+01, 3.35059474772989e+01, 3.35000889612230e+01, 3.34962514649315e+01, 3.34938780553064e+01, 3.34925135562936e+01, 3.34918011578305e+01, 3.34914756554884e+01, 3.34913535480634e+01, 3.34913203318587e+01, 3.34913154292156e+01, 3.34913150510765e+01, 3.34913082609136e+01, 3.34912622565593e+01, 3.34910931396682e+01, 3.34906423359873e+01, 3.34896557624446e+01, 3.34877663533524e+01, 3.34844806465419e+01, 3.34791700478922e+01, 3.34710672970353e+01, 3.34592685454342e+01, 3.34427413269719e+01, 3.34203385469271e+01, 3.33908184349738e+01, 3.33528702009577e+01, 3.33051449013240e+01, 3.32462907760083e+01, 3.31749920618947e+01, 3.30900100455981e+01, 3.29902249049445e+01, 3.28746767264203e+01, 3.27426039956342e+01, 3.25934778563705e+01, 3.24270305315559e+01, 3.22432764982706e+01, 3.20425253012523e+01, 3.18253852589400e+01, 3.15927577400675e+01, 3.13458221404658e+01, 3.10860121412100e+01, 3.08149842532757e+01, 3.05345800243904e+01, 3.02467835759810e+01, 2.99536763291940e+01, 2.96573908502026e+01, 2.93600656857178e+01, 2.90638028714419e+01, 2.87706294962254e+01, 2.84824643252334e+01, 2.82010900699973e+01, 2.79281314891538e+01, 2.76650391535377e+01, 2.74130784435971e+01, 2.71733231800384e+01, 2.69466532178968e+01, 2.67337553443835e+01, 2.65351268885616e+01, 2.63510815508085e+01, 2.61817570694544e+01, 2.60271244438330e+01, 2.58869985165934e+01, 2.57610497786976e+01, 2.56488172975937e+01, 2.55497226847552e+01, 2.54630850164234e+01, 2.53881366044892e+01, 2.53240394863059e+01, 2.52699024658304e+01, 2.52247984967893e+01, 2.51877821546078e+01, 2.51579069009824e+01, 2.51342418068980e+01, 2.51158873704243e+01, 2.51019900485477e+01, 2.50917551209103e+01, 2.50844575201272e+01, 2.50794502996491e+01, 2.50761704658094e+01, 2.50741419740766e+01, 2.50729757774852e+01, 2.50723669133328e+01, 2.50720887172603e+01, 2.50719843561452e+01, 2.50719559673761e+01, 2.50719517772527e+01, 2.50719514599892e+01, 2.50719458396979e+01, 2.50719077613972e+01, 2.50717677815329e+01, 2.50713946465852e+01, 2.50705780497612e+01, 2.50690141686982e+01, 2.50662945649661e+01, 2.50618989583489e+01, 2.50551923099716e+01, 2.50454265559669e+01, 2.50317472243123e+01, 2.50132050382603e+01, 2.49887724577880e+01, 2.49573649349137e+01, 2.49178664617666e+01, 2.48691587780729e+01, 2.48101533878136e+01, 2.47398253282459e+01, 2.46572474568512e+01, 2.45616238935493e+01, 2.44523211964524e+01, 2.43288958753868e+01, 2.41911169670454e+01, 2.40389826075556e+01, 2.38727298293977e+01, 2.36928371556050e+01, 2.35000199319736e+01, 2.32952186909439e+01, 2.30795811452798e+01, 2.28544386419593e+01, 2.26212780582135e+01, 2.23817102003536e+01, 2.21374357930042e+01, 2.18902101475523e+01, 2.16418075941796e+01, 2.13939867566623e+01, 2.11484577293175e+01, 2.09068521534314e+01, 2.06706970578574e+01, 2.04413931119489e+01, 2.02201976514831e+01, 2.00082125188705e+01, 1.98063764612659e+01, 1.96154616036303e+01, 1.94360733874828e+01, 1.92686533419368e+01, 1.91134841103181e+01, 1.89706962598334e+01, 1.88402765200199e+01, 1.87220772030536e+01, 1.86158266416272e+01, 1.85211405340316e+01, 1.84375341135886e+01, 1.83644350657445e+01, 1.83011971063602e+01, 1.82471141135657e+01, 1.82014346765367e+01, 1.81633768905352e+01, 1.81321431912025e+01, 1.81069349851707e+01, 1.80869668016815e+01, 1.80714796643682e+01, 1.80597533669895e+01, 1.80511173346449e+01, 1.80449597651124e+01, 1.80407347746313e+01, 1.80379673186846e+01, 1.80362557197069e+01, 1.80352717075049e+01, 1.80347579607550e+01, 1.80345232247768e+01, 1.80344351670706e+01, 1.80344112132266e+01, 1.80344076776889e+01, 1.80344074122750e+01, 1.80344027406888e+01, 1.80343710900045e+01, 1.80342547387300e+01, 1.80339445888656e+01, 1.80332658330423e+01, 1.80319659327873e+01, 1.80297053901558e+01, 1.80260517383375e+01, 1.80204771109879e+01, 1.80123596731305e+01, 1.80009891059404e+01, 1.79855762308802e+01, 1.79652667334561e+01, 1.79391588027568e+01, 1.79063243417782e+01, 1.78658332302432e+01, 1.78167799446708e+01, 1.77583116717515e+01, 1.76896569054230e+01, 1.76101534116456e+01, 1.75192743934620e+01, 1.74166517050999e+01, 1.73020950544833e+01, 1.71756062972605e+01, 1.70373881517167e+01, 1.68878469332758e+01, 1.67275891942177e+01, 1.65574124318509e+01, 1.63782902740289e+01, 1.61913527508247e+01, 1.59978624124917e+01, 1.57991871627345e+01, 1.55967707529530e+01, 1.53921019354805e+01, 1.51866833030948e+01, 1.49820008417412e+01, 1.47794951823553e+01, 1.45805354458034e+01, 1.43863964286630e+01, 1.41982396830492e+01, 1.40170988168714e+01, 1.38438691042532e+01, 1.36793012736500e+01, 1.35239991546272e+01, 1.33784207278865e+01, 1.32428820432335e+01, 1.31175634451175e+01, 1.30025175671847e+01, 1.28976786141429e+01, 1.28028725278129e+01, 1.27178277217220e+01, 1.26421861540518e+01, 1.25755145838282e+01, 1.25173159143544e+01, 1.24670405680806e+01, 1.24240978576521e+01, 1.23878673199311e+01, 1.23577099659050e+01, 1.23329793732487e+01, 1.23130325142942e+01, 1.22972401751145e+01, 1.22849967862787e+01, 1.22757294572660e+01, 1.22689059886139e+01, 1.22640416317586e+01, 1.22607043781123e+01, 1.22585185867160e+01, 1.22571668028329e+01, 1.22563896757746e+01, 1.22559839496055e+01, 1.22557985709249e+01, 1.22557290291367e+01, 1.22557101120987e+01, 1.22557073199841e+01, 1.22557071110166e+01, 1.22557034414420e+01, 1.22556785795438e+01, 1.22555871847224e+01, 1.22553435608077e+01, 1.22548104015294e+01, 1.22537893595762e+01, 1.22520138274390e+01, 1.22491442791711e+01, 1.22447664763434e+01, 1.22383927864237e+01, 1.22294667943884e+01, 1.22173713052563e+01, 1.22014397337354e+01, 1.21809707560371e+01, 1.21552459591218e+01, 1.21235500680323e+01, 1.20851931695614e+01, 1.20395341904999e+01, 1.19860047439875e+01, 1.19241323426943e+01, 1.18535619077080e+01, 1.17740744907863e+01, 1.16856021855827e+01, 1.15882383360245e+01, 1.14822423560845e+01, 1.13680387459174e+01, 1.12462102079869e+01, 1.11174851094674e+01, 1.09827198746989e+01, 1.08428771925052e+01, 1.06990011582066e+01, 1.05521906157228e+01, 1.04035720076348e+01, 1.02542729791463e+01, 1.01053978269682e+01, 9.95800565858485e+00, 9.81309186059121e+00, 9.67157319844277e+00, 9.53427661284438e+00, 9.40193156216393e+00, 9.27516559877574e+00, 9.15450276401404e+00, 9.04036433759525e+00, 8.93307147381364e+00, 8.83284928683807e+00, 8.73983199917385e+00, 8.65406883040034e+00, 8.57553036952648e+00, 8.50411523798245e+00, 8.43965690769299e+00, 8.38193058767886e+00, 8.33066013202447e+00, 8.28552495125991e+00, 8.24616692804855e+00, 8.21219734667977e+00, 8.18320384470491e+00, 8.15875738499437e+00, 8.13841922886681e+00, 8.12174786756966e+00, 8.10830584255419e+00, 8.09766635732961e+00, 8.08941955804723e+00, 8.08317833919030e+00, 8.07858351740846e+00, 8.07530821269883e+00, 8.07306128312212e+00, 8.07158967745829e+00, 8.07067959906983e+00, 8.07015641220594e+00, 8.06988326668744e+00, 8.06975846542555e+00, 8.06971164835474e+00, 8.06969891299438e+00, 8.06969703328252e+00, 8.06969689254842e+00, 8.06969442048347e+00, 8.06967767188982e+00, 8.06961610243879e+00, 8.06945198204707e+00, 8.06909281446762e+00, 8.06840498813036e+00, 8.06720892290694e+00, 8.06527595292952e+00, 8.06232715652846e+00, 8.05803430869643e+00, 8.05202308652459e+00, 8.04387860092874e+00, 8.03315325630308e+00, 8.01937685230238e+00, 8.00206873945561e+00, 7.98075172594011e+00, 7.95496731262388e+00, 7.92429171625789e+00, 7.88835203760328e+00, 7.84684185468143e+00, 7.79953548330130e+00, 7.74630015744637e+00, 7.68710544684840e+00, 7.62202934840884e+00, 7.55126065593642e+00, 7.47509741665229e+00, 7.39394150590640e+00, 7.30828957378017e+00, 7.21872081912917e+00, 7.12588221144328e+00, 7.03047189688387e+00, 6.93322158605574e+00, 6.83487872715357e+00, 6.73618922324984e+00, 6.63788136391707e+00, 6.54065151826982e+00, 6.44515198934894e+00, 6.35198127013457e+00, 6.26167678185187e+00, 6.17471002832251e+00, 6.09148397746389e+00, 6.01233239131717e+00, 5.93752077348344e+00, 5.86724858684826e+00, 5.80165240976648e+00, 5.74080973723077e+00, 5.68474318561526e+00, 5.63342491665750e+00, 5.58678115149109e+00, 5.54469669409060e+00, 5.50701942287963e+00, 5.47356473857907e+00, 5.44411997582220e+00, 5.41844879642992e+00, 5.39629558458814e+00, 5.37738985965929e+00, 5.36145071219401e+00, 5.34819125415207e+00, 5.33732305676187e+00, 5.32856053035017e+00, 5.32162518148881e+00, 5.31624966564550e+00, 5.31218153987984e+00, 5.30918661153356e+00, 5.30705177655981e+00, 5.30558724590957e+00, 5.30462807047425e+00, 5.30403489407582e+00, 5.30369388890006e+00, 5.30351585704902e+00, 5.30343451364378e+00, 5.30340399905649e+00, 5.30339569836137e+00, 5.30339447319662e+00, 5.30339438122515e+00, 5.30339276245016e+00, 5.30338179501843e+00, 5.30334147768428e+00, 5.30323400728064e+00, 5.30299881502435e+00, 5.30254840940187e+00, 5.30176519932147e+00, 5.30049945534721e+00, 5.29856854799466e+00, 5.29575757884145e+00, 5.29182149072685e+00, 5.28648870560457e+00, 5.27946629100826e+00, 5.27044659773309e+00, 5.25911524267227e+00, 5.24516023395174e+00, 5.22828195475049e+00, 5.20820364363168e+00, 5.18468194063699e+00, 5.15751701849450e+00, 5.12656179553347e+00, 5.09172973818924e+00, 5.05300081026038e+00, 5.01042521315162e+00, 4.96412468121456e+00, 4.91429123933702e+00, 4.86118348281281e+00, 4.80512058708687e+00, 4.74647438253066e+00, 4.68565992510373e+00, 4.62312505040188e+00, 4.55933941449410e+00, 4.49478350362618e+00, 4.42993804372090e+00, 4.36527416903650e+00, 4.30124462687952e+00, 4.23827621013861e+00, 4.17676352786570e+00, 4.11706415041090e+00, 4.05949510231014e+00, 4.00433062466606e+00, 3.95180108982805e+00, 3.90209292486454e+00, 3.85534938619579e+00, 3.81167202482798e+00, 3.77112268836375e+00, 3.73372592036433e+00, 3.69947163741557e+00, 3.66831798702790e+00, 3.64019431302084e+00, 3.61500417733259e+00, 3.59262840667306e+00, 3.57292814796730e+00, 3.55574792741725e+00, 3.54091871396238e+00, 3.52826098902670e+00, 3.51758782111363e+00, 3.50870793673638e+00, 3.50142876927683e+00, 3.49555945576389e+00, 3.49091373949942e+00, 3.48731272523915e+00, 3.48458742453336e+00, 3.48258102298404e+00, 3.48115079949073e+00, 3.48016963060331e+00, 3.47952702105265e+00, 3.47912961412498e+00, 3.47890115210685e+00, 3.47878187652061e+00, 3.47872737901704e+00, 3.47870693520240e+00, 3.47870137399671e+00, 3.47870055317468e+00, 3.47870049129424e+00, 3.47869939865099e+00, 3.47869199583662e+00, 3.47866478237541e+00, 3.47859224179742e+00, 3.47843349114626e+00, 3.47812947473357e+00, 3.47760081917109e+00, 3.47674645588862e+00, 3.47544310453119e+00, 3.47354569489989e+00, 3.47088878525721e+00, 3.46728900959275e+00, 3.46254855477771e+00, 3.45645962988769e+00, 3.44880984463734e+00, 3.43938836322890e+00, 3.42799264659960e+00, 3.41443554392077e+00, 3.39855244813314e+00, 3.38020819574071e+00, 3.35930337339722e+00, 3.33577969749896e+00, 3.30962416088119e+00, 3.28087169328258e+00, 3.24960615723170e+00, 3.21595959337569e+00, 3.18010973160077e+00, 3.14227588760362e+00, 3.10271345939841e+00, 3.06170731585031e+00, 3.01956442283973e+00, 2.97660607794671e+00, 2.93316012069146e+00, 2.88955345469386e+00, 2.84610516572288e+00, 2.80312045252870e+00, 2.76088551347863e+00, 2.71966345899231e+00, 2.67969125401941e+00, 2.64117764088806e+00, 2.60430195320848e+00, 2.56921370654707e+00, 2.53603284005818e+00, 2.50485048285491e+00, 2.47573012679543e+00, 2.44870910072960e+00, 2.42380025760102e+00, 2.40099380314948e+00, 2.38025921186223e+00, 2.36154719129834e+00, 2.34479166933257e+00, 2.32991178984739e+00, 2.31681391071419e+00, 2.30539360342244e+00, 2.29553765640654e+00, 2.28712608405064e+00, 2.28003414071748e+00, 2.27413433428697e+00, 2.26929842709885e+00, 2.26539940451968e+00, 2.26231338336203e+00, 2.25992142491027e+00, 2.25811121119574e+00, 2.25677853918240e+00, 2.25582858629406e+00, 2.25517690263439e+00, 2.25475009045201e+00, 2.25448613972281e+00, 2.25433439970932e+00, 2.25425517932695e+00, 2.25421898324060e+00, 2.25420540489734e+00, 2.25420171126426e+00, 2.25420116609195e+00, 2.25420112476864e+00, 2.25420039214254e+00, 2.25419542849622e+00, 2.25417718165706e+00, 2.25412854266506e+00, 2.25402209935879e+00, 2.25381825511949e+00, 2.25346379151526e+00, 2.25289094648405e+00, 2.25201707126561e+00, 2.25074491888425e+00, 2.24896360375359e+00, 2.24655025502648e+00, 2.24337236490112e+00, 2.23929080683828e+00, 2.23416346774738e+00, 2.22784940363828e+00, 2.22021339186308e+00, 2.21113071761240e+00, 2.20049200120615e+00, 2.18820784972475e+00, 2.17421310534264e+00, 2.15847046633501e+00, 2.14097327683627e+00, 2.12174731799446e+00, 2.10085148416825e+00, 2.07837728933632e+00, 2.05444721555694e+00, 2.02921198101997e+00, 2.00284686402595e+00, 1.97554726619100e+00, 1.94752373008853e+00, 1.91899664217450e+00, 1.89019085187457e+00, 1.86133042425512e+00, 1.83263371967215e+00, 1.80430896226571e+00, 1.77655042290775e+00, 1.74953530348160e+00, 1.72342137000327e+00, 1.69834534369851e+00, 1.67442202334989e+00, 1.65174408076020e+00, 1.63038244579697e+00, 1.61038717970785e+00, 1.59178872618601e+00, 1.57459942914735e+00, 1.55881521356811e+00, 1.54441733941660e+00, 1.53137415655858e+00, 1.51964280820713e+00, 1.50917084987278e+00, 1.49989776811307e+00, 1.49175639748874e+00, 1.48467424436733e+00, 1.47857473242357e+00, 1.47337838708461e+00, 1.46900397523410e+00, 1.46536961284516e+00, 1.46239384757159e+00, 1.45999671643175e+00, 1.45810077132632e+00, 1.45663205796944e+00, 1.45552102757428e+00, 1.45470335592533e+00, 1.45412064177219e+00, 1.45372095611921e+00, 1.45345921607867e+00, 1.45329736141127e+00, 1.45320431838661e+00, 1.45315574367103e+00, 1.45313354996829e+00, 1.45312522442281e+00, 1.45312295967999e+00, 1.45312262540895e+00, 1.45312259991090e+00, 1.45312214573166e+00, 1.45311906860451e+00, 1.45310775681884e+00, 1.45307760420690e+00, 1.45301161826460e+00, 1.45288525614834e+00, 1.45266553908499e+00, 1.45231049231496e+00, 1.45176895351244e+00, 1.45098078709014e+00, 1.44987753412499e+00, 1.44838351701909e+00, 1.44641740372484e+00, 1.44389421799450e+00, 1.44072775970081e+00, 1.43683337347807e+00, 1.43213097611650e+00, 1.42654822539757e+00, 1.42002368809054e+00, 1.41250984574818e+00, 1.40397576691726e+00, 1.39440927623453e+00, 1.38381846662506e+00, 1.37223243120744e+00, 1.35970113569844e+00, 1.34629440744202e+00, 1.33210007922528e+00, 1.31722138882748e+00, 1.30177379189553e+00, 1.28588138930417e+00, 1.26967319475384e+00, 1.25327947030350e+00, 1.23682833630485e+00, 1.22044282085504e+00, 1.20423845870336e+00, 1.18832148890787e+00, 1.17278764320555e+00, 1.15772147038660e+00, 1.14319611054280e+00, 1.12927341815865e+00, 1.11600433290864e+00, 1.10342940796830e+00, 1.09157942309595e+00, 1.08047602945413e+00, 1.07013239185947e+00, 1.06055380988662e+00, 1.05173831117290e+00, 1.04367721840158e+00, 1.03635569632390e+00, 1.02975328755056e+00, 1.02384444643025e+00, 1.01859907973621e+00, 1.01398310152518e+00, 1.00995900769367e+00, 1.00648647358016e+00, 1.00352297553205e+00, 1.00102443472147e+00, 9.98945878705712e-01, 9.97242113370625e-01, 9.95868395077942e-01, 9.94781090221878e-01, 9.93938307164008e-01, 9.93300483859706e-01, 9.92830913604429e-01, 9.92496191371416e-01, 9.92266564290468e-01, 9.92116171959516e-01, 9.92023165427931e-01, 9.91969697699439e-01, 9.91941783246084e-01, 9.91929029020397e-01, 9.91924244489205e-01, 9.91922942982512e-01, 9.91922750882928e-01, 9.91922736126241e-01, 9.91922471920049e-01, 9.91920681885418e-01, 9.91914101547124e-01, 9.91896560921688e-01, 9.91858174404804e-01, 9.91784662363855e-01, 9.91656832850637e-01, 9.91450249293301e-01, 9.91135106975698e-01, 9.90676337318945e-01, 9.90033954104206e-01, 9.89163649469603e-01, 9.88017639466641e-01, 9.86545749027755e-01, 9.84696714461357e-01, 9.82419668487516e-01, 9.79665759166288e-01, 9.76389841049276e-01, 9.72552165988446e-01, 9.68119993869117e-01, 9.63069041559685e-01, 9.57384692618696e-01, 9.51062901097615e-01, 9.44110739568060e-01, 9.36546562836845e-01, 9.28399782519314e-01, 9.19710271179987e-01, 9.10527435645747e-01, 9.00909015383220e-01, 8.90919672345532e-01, 8.80629443189889e-01, 8.70112123809832e-01, 8.59443650827416e-01, 8.48700536388399e-01, 8.37958402576248e-01, 8.27290651061581e-01, 8.16767293001647e-01, 8.06453954231309e-01, 7.96411061803774e-01, 7.86693210202621e-01, 7.77348699238992e-01, 7.68419230885570e-01, 7.59939749118637e-01, 7.51938405187559e-01, 7.44436630454516e-01, 7.37449299806698e-01, 7.30984970335435e-01, 7.25046182179049e-01, 7.19629810827655e-01, 7.14727462525488e-01, 7.10325906470764e-01, 7.06407539154725e-01, 7.02950877307508e-01, 6.99931076481604e-01, 6.97320472306217e-01, 6.95089140923628e-01, 6.93205474145898e-01, 6.91636763547358e-01, 6.90349786161278e-01, 6.89311382826787e-01, 6.88489018691762e-01, 6.87851314090423e-01, 6.87368533137374e-01, 6.87013017056877e-01, 6.86759549604226e-01, 6.86585642996256e-01, 6.86471734557152e-01, 6.86401286748080e-01, 6.86360786273454e-01, 6.86339641381544e-01, 6.86329980107097e-01, 6.86326355829713e-01, 6.86325369938409e-01, 6.86325224422923e-01, 6.86325213159428e-01, 6.86325010385239e-01, 6.86323636560658e-01, 6.86318586241086e-01, 6.86305123981285e-01, 6.86275662347541e-01, 6.86219240532026e-01, 6.86121124829022e-01, 6.85962549960458e-01, 6.85720617576770e-01, 6.85368365087093e-01, 6.84875014129148e-01, 6.84206403302859e-01, 6.83325604122842e-01, 6.82193712436039e-01, 6.80770799870352e-01, 6.79017001469013e-01, 6.76893706956534e-01, 6.74364814699054e-01, 6.71398000146759e-01, 6.67965945235743e-01, 6.64047472717840e-01, 6.59628530353800e-01, 6.54702974741845e-01, 6.49273113258367e-01, 6.43349974712770e-01, 6.36953293979190e-01, 6.30111211827767e-01, 6.22859706999876e-01, 6.15241791809865e-01, 6.07306513952039e-01, 5.99107814845311e-01, 5.90703298326645e-01, 5.82152962874652e-01, 5.73517946336317e-01, 5.64859325212531e-01, 5.56237001989793e-01, 5.47708704849661e-01, 5.39329115262255e-01, 5.31149131148162e-01, 5.23215266867460e-01, 5.15569186400228e-01, 5.08247362653968e-01, 5.01280853671773e-01, 4.94695185360165e-01, 4.88510329946926e-01, 4.82740769492239e-01, 4.77395634248175e-01, 4.72478906374896e-01, 4.67989680392812e-01, 4.63922472700794e-01, 4.60267573445286e-01, 4.57011434899158e-01, 4.54137091224801e-01, 4.51624604976331e-01, 4.49451535884094e-01, 4.47593427327703e-01, 4.46024305436183e-01, 4.44717184993827e-01, 4.43644575346717e-01, 4.42778978408718e-01, 4.42093369783500e-01, 4.41561653099960e-01, 4.41159077042187e-01, 4.40862604369649e-01, 4.40651222566288e-01, 4.40506186682150e-01, 4.40411186446354e-01, 4.40352431790409e-01, 4.40318653433110e-01, 4.40301018009720e-01, 4.40292960221598e-01, 4.40289937464250e-01, 4.40289115200692e-01, 4.40288993836305e-01, 4.40288984366042e-01, 4.40288812891154e-01, 4.40287651123761e-01, 4.40283380346344e-01, 4.40271996040806e-01, 4.40247081863085e-01, 4.40199368601988e-01, 4.40116395997734e-01, 4.39982293317378e-01, 4.39777692275690e-01, 4.39479783163506e-01, 4.39062521787662e-01, 4.38496990910338e-01, 4.37751915210320e-01, 4.36794323337530e-01, 4.35590344458603e-01, 4.34106119986367e-01, 4.32308804277369e-01, 4.30167621437221e-01, 4.27654939569287e-01, 4.24747319465290e-01, 4.21426492471845e-01, 4.17680222561430e-01, 4.13503010773248e-01, 4.08896606178288e-01, 4.03870296070060e-01, 3.98440958611161e-01, 3.92632872874962e-01, 3.86477293198595e-01, 3.80011806093833e-01, 3.73279497853446e-01, 3.66327968852872e-01, 3.59208236037993e-01, 3.51973568090507e-01, 3.44678298339524e-01, 3.37376658812190e-01, 3.30121675139409e-01, 3.22964156662110e-01, 3.15951809392409e-01, 3.09128491903373e-01, 3.02533626218633e-01, 2.96201767826496e-01, 2.90162331522076e-01, 2.84439463322656e-01, 2.79052043580549e-01, 2.74013802888859e-01, 2.69333530526580e-01, 2.65015354932480e-01, 2.61059076802503e-01, 2.57460537554543e-01, 2.54212008750130e-01, 2.51302591250564e-01, 2.48718616084545e-01, 2.46444041920903e-01, 2.44460846444831e-01, 2.42749410677128e-01, 2.41288896287386e-01, 2.40057616242105e-01, 2.39033398770410e-01, 2.38193943739304e-01, 2.37517169268401e-01, 2.36981544956670e-01, 2.36566406638083e-01, 2.36252246316063e-01, 2.36020970018174e-01, 2.35856115897917e-01, 2.35743025073404e-01, 2.35668958466661e-01, 2.35623154261596e-01, 2.35596822452692e-01, 2.35583075183143e-01, 2.35576794012417e-01, 2.35574437739800e-01, 2.35573796777601e-01, 2.35573702172957e-01, 2.35573694728004e-01, 2.35573559119636e-01, 2.35572640353801e-01, 2.35569262885384e-01, 2.35560259886292e-01, 2.35540557508379e-01, 2.35502826881174e-01, 2.35437218586142e-01, 2.35331193131877e-01, 2.35169459503041e-01, 2.34934032036406e-01, 2.34604413517559e-01, 2.34157909312792e-01, 2.33570073401332e-01, 2.32815282271851e-01, 2.31867426819989e-01, 2.30700705794099e-01, 2.29290497303438e-01, 2.27614277900052e-01, 2.25652552372551e-01, 2.23389752331222e-01, 2.20815058611052e-01, 2.17923102092634e-01, 2.14714500197886e-01, 2.11196192268254e-01, 2.07381546185317e-01, 2.03290220515640e-01, 1.98947780395635e-01, 1.94385080277542e-01, 1.89637441299070e-01, 1.84743664099386e-01, 1.79744928142519e-01, 1.74683634998402e-01, 1.69602254933096e-01, 1.64542233401804e-01, 1.59543006977430e-01, 1.54641167707283e-01, 1.49869802060550e-01, 1.45258016894807e-01, 1.40830651635049e-01, 1.36608164354712e-01, 1.32606670578504e-01, 1.28838107852445e-01, 1.25310496463491e-01, 1.22028266784142e-01, 1.18992625983182e-01, 1.16201940614900e-01, 1.13652116229503e-01, 1.11336960064008e-01, 1.09248517613395e-01, 1.07377378107529e-01, 1.05712947428437e-01, 1.04243689716867e-01, 1.02957340847325e-01, 1.01841098160932e-01, 1.00881791410337e-01, 1.00066039874562e-01, 9.93804001240051e-02, 9.88115080483922e-02, 9.83462176002377e-02, 9.79717373617177e-02, 9.76757646252437e-02, 9.74466153020543e-02, 9.72733467494677e-02, 9.71458696404710e-02, 9.70550443729508e-02, 9.69927572934779e-02, 9.69519722181360e-02, 9.69267533619629e-02, 9.69122567888192e-02, 9.69046887831624e-02, 9.69012310157567e-02, 9.68999339067551e-02, 9.68995810634325e-02, 9.68995289845651e-02, 9.68995248499098e-02, 9.68994490770143e-02, 9.68989357055534e-02, 9.68970485140161e-02, 9.68920180788473e-02, 9.68810097267507e-02, 9.68599298380965e-02, 9.68232793149966e-02, 9.67640626746658e-02, 9.66737607594652e-02, 9.65423743231781e-02, 9.63585445661508e-02, 9.61097548518310e-02, 9.57826151640617e-02, 9.53632272488818e-02, 9.48376238308981e-02, 9.41922699513924e-02, 9.34146086774135e-02, 9.24936276801347e-02, 9.14204181324649e-02, 9.01886937689511e-02, 8.87952365092914e-02, 8.72402363690910e-02, 8.55274978015430e-02, 8.36644920983716e-02, 8.16622455510938e-02, 7.95350648269064e-02, 7.73001132092404e-02, 7.49768626184633e-02, 7.25864553747400e-02, 7.01510154982330e-02, 6.76929513819857e-02, 6.52342898092755e-02, 6.27960758995343e-02, 6.03978654980040e-02, 5.80573269759199e-02, 5.57899596984512e-02, 5.36089276618696e-02, 5.15249996711731e-02, 4.95465821642824e-02, 4.76798273951786e-02, 4.59287981401388e-02, 4.42956703187818e-02, 4.27809566508444e-02, 4.13837371481520e-02, 4.01018852109658e-02, 3.89322808989903e-02, 3.78710054641402e-02, 3.69135135634835e-02, 3.60547817585086e-02, 3.52894338162457e-02, 3.46118447450650e-02, 3.40162263272745e-02, 3.34966972467043e-02, 3.30473409558820e-02, 3.26622543422559e-02, 3.23355900823756e-02, 3.20615952804602e-02, 3.18346485356571e-02, 3.16492969796229e-02, 3.15002941280582e-02, 3.13826386702232e-02, 3.12916136421251e-02, 3.12228248382169e-02, 3.11722368484837e-02, 3.11362047908048e-02, 3.11114996634245e-02, 3.10953252854844e-02, 3.10853250215288e-02, 3.10795768831795e-02, 3.10765761329321e-02, 3.10752051329501e-02, 3.10746908344512e-02, 3.10745509338233e-02, 3.10745302848261e-02, 3.10745286304410e-02, 3.10744981226065e-02, 3.10742914281903e-02, 3.10735316069284e-02, 3.10715062738538e-02, 3.10670742443833e-02, 3.10585877663527e-02, 3.10438340543158e-02, 3.10199997932727e-02, 3.09836621876869e-02, 3.09308099140750e-02, 3.08568968871654e-02, 3.07569309579421e-02, 3.06255984274841e-02, 3.04574235232451e-02, 3.02469597473944e-02, 2.99890073611024e-02, 2.96788484019999e-02, 2.93124878285482e-02, 2.88868869987695e-02, 2.84001741146227e-02, 2.78518158779804e-02, 2.72427357232203e-02, 2.65753667866804e-02, 2.58536321865890e-02, 2.50828508446048e-02, 2.42695732790636e-02, 2.34213576361384e-02, 2.25465008616013e-02, 2.16537428776817e-02, 2.07519628989346e-02, 1.98498868085153e-02, 1.89558228962596e-02, 1.80774400367270e-02, 1.72215974922310e-02, 1.63942296188750e-02, 1.56002833115068e-02, 1.48437025297312e-02, 1.41274530030060e-02, 1.34535800341880e-02, 1.28232916814710e-02, 1.22370582456952e-02, 1.16947183053071e-02, 1.11955830400987e-02, 1.07385340697448e-02, 1.03221134443530e-02, 9.94460572588451e-03, 9.60411121494786e-03, 9.29860816960826e-03, 9.02600220627647e-03, 8.78416311643093e-03, 8.57095158453092e-03, 8.38423919653053e-03, 8.22192439570557e-03, 8.08194561065534e-03, 7.96229179035635e-03, 7.86101047445698e-03, 7.77621398107442e-03, 7.70608471746925e-03, 7.64888066263580e-03, 7.60294174556009e-03, 7.56669736719573e-03, 7.53867488699456e-03, 7.51750860742106e-03, 7.50194863038445e-03, 7.49086888822603e-03, 7.48327363421453e-03, 7.47830170391239e-03, 7.47522793088962e-03, 7.47346122132342e-03, 7.47253895712419e-03, 7.47211759381372e-03, 7.47195953022128e-03, 7.47191653348978e-03, 7.47191018729154e-03, 7.47190967406652e-03, 7.47190015053434e-03, 7.47183562743459e-03, 7.47159843731282e-03, 7.47096620354881e-03, 7.46958271737697e-03, 7.46693371997509e-03, 7.46232879398434e-03, 7.45489058128612e-03, 7.44355254396822e-03, 7.42706644632563e-03, 7.40402061699541e-03, 7.37286980709229e-03, 7.33197704779087e-03, 7.27966728521068e-03, 7.21429170848609e-03, 7.13430059952458e-03, 7.03832130988502e-03, 6.92523679950341e-03, 6.79425935075528e-03, 6.64499391265139e-03, 6.47748618926869e-03, 6.29225186656868e-03, 6.09028468294071e-03, 5.87304173508585e-03, 5.64240439194919e-03, 5.40061346075760e-03, 5.15017951693153e-03, 4.89377445827444e-03, 4.63411662659041e-03, 4.37386477306489e-03, 4.11553160090961e-03, 3.86141669794426e-03, 3.61354941296099e-03, 3.37363408340963e-03, 3.14300381550225e-03, 2.92260229728214e-03, 2.71300893899706e-03, 2.51449799586823e-03, 2.32709766187611e-03, 2.15061777817027e-03, 1.98464944816834e-03, 1.82857813626214e-03, 1.68165300318179e-03, 1.54310835733895e-03, 1.41228275586514e-03, 1.28868006786576e-03, 1.17195813550097e-03, 1.06186858866408e-03, 9.58178171913460e-04, 8.61784424088796e-04, 7.70822970196150e-04, 6.84803775328646e-04, 6.05224449657271e-04, 5.30573410259194e-04, 4.61158880781909e-04, 3.97093334981946e-04, 3.38448414382242e-04, 2.85235471544354e-04, 2.37398163733360e-04, 1.94823316952863e-04, 1.57355544295967e-04, 1.24801841870692e-04, 9.69299114998989e-05, 7.34667621884455e-05, 5.40998442772889e-05, 3.84806747959487e-05, 2.62303294391586e-05, 1.69461830239470e-05, 1.02094534568664e-05, 5.59331393645834e-06, 2.67154224786736e-06, 1.02790028222444e-06, 2.66783345637123e-07, 0.00000000000000e+00],
[ 6.50000000000000e+01, 6.49999999326244e+01, 6.49999978492380e+01, 6.49999837339651e+01, 6.49999318443763e+01, 6.49997935230758e+01, 6.49994907965528e+01, 6.49989109863949e+01, 6.49979025269115e+01, 6.49962721459662e+01, 6.49937835237308e+01, 6.49901574988375e+01, 6.49850738449871e+01, 6.49781745955895e+01, 6.49690688515843e+01, 6.49573389700461e+01, 6.49425479998533e+01, 6.49242482061091e+01, 6.49019905066585e+01, 6.48753346304162e+01, 6.48438597958397e+01, 6.48071756958442e+01, 6.47649335599893e+01, 6.47168370441465e+01, 6.46626526721064e+01, 6.46022195251169e+01, 6.45354578489967e+01, 6.44623762310584e+01, 6.43830769982804e+01, 6.42977595109835e+01, 6.42067210773729e+01, 6.41103552946279e+01, 6.40091477282205e+01, 6.39036689649252e+01, 6.37945652053522e+01, 6.36825466860638e+01, 6.35683743272136e+01, 6.34528450794573e+01, 6.33367764877845e+01, 6.32209909985468e+01, 6.31063005121951e+01, 6.29934916342552e+01, 6.28833120089897e+01, 6.27764580425978e+01, 6.26735642435138e+01, 6.25751943325512e+01, 6.24818342093765e+01, 6.23938868061100e+01, 6.23116688139691e+01, 6.22354092337461e+01, 6.21652496737240e+01, 6.21012462972903e+01, 6.20433733049868e+01, 6.19915278203658e+01, 6.19455360345862e+01, 6.19051604504808e+01, 6.18701080526331e+01, 6.18400392160106e+01, 6.18145771524656e+01, 6.17933176827197e+01, 6.17758391122545e+01, 6.17617119838745e+01, 6.17505084785876e+01, 6.17418112407615e+01, 6.17352214139589e+01, 6.17303656908308e+01, 6.17269022040563e+01, 6.17245251152358e+01, 6.17229677942193e+01, 6.17220045215291e+01, 6.17214506899911e+01, 6.17211615268470e+01, 6.17210294028110e+01, 6.17209798380371e+01, 6.17209663551710e+01, 6.17209643651265e+01, 6.17209641040746e+01, 6.17209580218793e+01, 6.17209168140933e+01, 6.17207653296422e+01, 6.17203615256335e+01, 6.17194777998860e+01, 6.17177853189355e+01, 6.17148419539167e+01, 6.17100843542723e+01, 6.17028245944760e+01, 6.16922517215749e+01, 6.16774384106505e+01, 6.16573528005742e+01, 6.16308754334208e+01, 6.15968210580930e+01, 6.15539648835897e+01, 6.15010726825857e+01, 6.14369339556969e+01, 6.13603971771536e+01, 6.12704059628338e+01, 6.11660348452042e+01, 6.10465232253785e+01, 6.09113060235839e+01, 6.07600395910870e+01, 6.05926216004582e+01, 6.04092039065450e+01, 6.02101977575287e+01, 5.99962711993399e+01, 5.97683390002893e+01, 5.95275458561256e+01, 5.92752439528460e+01, 5.90129661219729e+01, 5.87423958139267e+01, 5.84653349735243e+01, 5.81836706931423e+01, 5.78993413208047e+01, 5.76143025758856e+01, 5.73304942027895e+01, 5.70498077570878e+01, 5.67740562160401e+01, 5.65049461676588e+01, 5.62440533038601e+01, 5.59928018039197e+01, 5.57524479671516e+01, 5.55240681919293e+01, 5.53085511622875e+01, 5.51065939374957e+01, 5.49187015597019e+01, 5.47451897886413e+01, 5.45861906134414e+01, 5.44416602501007e+01, 5.43113893862665e+01, 5.41950154695796e+01, 5.40920368482055e+01, 5.40018285642543e+01, 5.39236595774115e+01, 5.38567111626110e+01, 5.38000961868681e+01, 5.37528789305587e+01, 5.37140950810162e+01, 5.36827714944731e+01, 5.36579452990646e+01, 5.36386818996537e+01, 5.36240914471008e+01, 5.36133433523146e+01, 5.36056784602029e+01, 5.36004185507792e+01, 5.35969729033087e+01, 5.35948417424593e+01, 5.35936164798200e+01, 5.35929767658279e+01, 5.35926844714598e+01, 5.35925748211426e+01, 5.35925449935403e+01, 5.35925405910466e+01, 5.35925402023902e+01, 5.35925325868594e+01, 5.35924809905073e+01, 5.35922913167015e+01, 5.35917857147655e+01, 5.35906792111601e+01, 5.35885601008677e+01, 5.35848748766626e+01, 5.35789183767531e+01, 5.35698297186284e+01, 5.35565944567036e+01, 5.35380532503794e+01, 5.35129171540626e+01, 5.34797894390806e+01, 5.34371936290071e+01, 5.33836071778920e+01, 5.33174999530822e+01, 5.32373764139118e+01, 5.31418201231229e+01, 5.30295390125893e+01, 5.28994096743847e+01, 5.27505188871316e+01, 5.25822006350203e+01, 5.23940670411123e+01, 5.21860319101543e+01, 5.19583259339822e+01, 5.17115030144358e+01, 5.14464375577659e+01, 5.11643129504796e+01, 5.08666017182357e+01, 5.05550381004960e+01, 5.02315839673267e+01, 4.98983891882001e+01, 4.95577477490309e+01, 4.92120510912641e+01, 4.88637402812983e+01, 4.85152586672423e+01, 4.81690066107490e+01, 4.78272996869549e+01, 4.74923314467103e+01, 4.71661414758623e+01, 4.68505891190409e+01, 4.65473329079152e+01, 4.62578154788675e+01, 4.59832535959638e+01, 4.57246328080726e+01, 4.54827062481651e+01, 4.52579971070995e+01, 4.50508043627574e+01, 4.48612114014433e+01, 4.46890972203969e+01, 4.45341499415066e+01, 4.43958823939415e+01, 4.42736495368530e+01, 4.41666674931720e+01, 4.40740339530386e+01, 4.39947496821235e+01, 4.39277408380237e+01, 4.38718817596198e+01, 4.38260178530315e+01, 4.37889881575622e+01, 4.37596471402432e+01, 4.37368852429460e+01, 4.37196476960057e+01, 4.37069511206941e+01, 4.36978974724252e+01, 4.36916849285226e+01, 4.36876153983010e+01, 4.36850984269157e+01, 4.36836513740916e+01, 4.36828958692802e+01, 4.36825506699629e+01, 4.36824211733074e+01, 4.36823859470326e+01, 4.36823807477066e+01, 4.36823803306563e+01, 4.36823726339636e+01, 4.36823204877309e+01, 4.36821287925620e+01, 4.36816178028435e+01, 4.36804995103886e+01, 4.36783578313997e+01, 4.36746333862341e+01, 4.36686135652800e+01, 4.36594284627162e+01, 4.36460531306072e+01, 4.36273164552155e+01, 4.36019167819477e+01, 4.35684442123557e+01, 4.35254092656813e+01, 4.34712773416792e+01, 4.34045081485264e+01, 4.33235989822829e+01, 4.32271304803608e+01, 4.31138132425708e+01, 4.29825335433955e+01, 4.28323962712778e+01, 4.26627632437362e+01, 4.24732851720107e+01, 4.22639257857310e+01, 4.20349769643640e+01, 4.17870641335462e+01, 4.15211416378375e+01, 4.12384782608186e+01, 4.09406334962526e+01, 4.06294255568486e+01, 4.03068924278763e+01, 3.99752475278381e+01, 3.96368317255098e+01, 3.92940635737716e+01, 3.89493896386460e+01, 3.86052367059343e+01, 3.82639674260500e+01, 3.79278406212091e+01, 3.75989770680318e+01, 3.72793311439783e+01, 3.69706683515104e+01, 3.66745484550652e+01, 3.63923137986642e+01, 3.61250823034808e+01, 3.58737446450910e+01, 3.56389651466753e+01, 3.54211859718267e+01, 3.52206342448450e+01, 3.50373317630163e+01, 3.48711069954023e+01, 3.47216090884184e+01, 3.45883236211764e+01, 3.44705898724411e+01, 3.43676193739237e+01, 3.42785155288290e+01, 3.42022940680223e+01, 3.41379040981136e+01, 3.40842494671107e+01, 3.40402101367173e+01, 3.40046632099800e+01, 3.39765032239738e+01, 3.39546612852300e+01, 3.39381226061984e+01, 3.39259419990561e+01, 3.39172569022526e+01, 3.39112975572854e+01, 3.39073940183790e+01, 3.39049797640764e+01, 3.39035917834545e+01, 3.39028671254825e+01, 3.39025360216712e+01, 3.39024118129419e+01, 3.39023780251316e+01, 3.39023730381206e+01, 3.39023726534744e+01, 3.39023657464620e+01, 3.39023189504358e+01, 3.39021469232779e+01, 3.39016883619158e+01, 3.39006848109879e+01, 3.38987628885951e+01, 3.38954206422072e+01, 3.38900186641320e+01, 3.38817764987521e+01, 3.38697747603448e+01, 3.38529632470824e+01, 3.38301751799396e+01, 3.38001475118826e+01, 3.37615470421753e+01, 3.37130018355917e+01, 3.36531371936760e+01, 3.35806151666286e+01, 3.34941763461583e+01, 3.33926824615910e+01, 3.32751581352936e+01, 3.31408300600019e+01, 3.29891618572179e+01, 3.28198829733952e+01, 3.26330101715953e+01, 3.24288604738032e+01, 3.22080547875206e+01, 3.19715118870667e+01, 3.17204328882415e+01, 3.14562768255175e+01, 3.11807283840684e+01, 3.08956592253043e+01, 3.06030846457916e+01, 3.03051174997621e+01, 3.00039213750346e+01, 2.97016649316172e+01, 2.94004790967798e+01, 2.91024184822524e+01, 2.88094279862282e+01, 2.85233151129264e+01, 2.82457281351696e+01, 2.79781398829842e+01, 2.77218366915216e+01, 2.74779118947648e+01, 2.72472632008908e+01, 2.70305933116415e+01, 2.68284132260278e+01, 2.66410477721234e+01, 2.64686430176466e+01, 2.63111753050549e+01, 2.61684617313725e+01, 2.60401719440613e+01, 2.59258411529071e+01, 2.58248842671008e+01, 2.57366110600243e+01, 2.56602422450304e+01, 2.55949263165771e+01, 2.55397569749470e+01, 2.54937909118910e+01, 2.54560656915973e+01, 2.54256174195424e+01, 2.54014978546235e+01, 2.53827905913537e+01, 2.53686259226514e+01, 2.53581939932696e+01, 2.53507558717458e+01, 2.53456522062679e+01, 2.53423091869345e+01, 2.53402416118664e+01, 2.53390529442993e+01, 2.53384323477100e+01, 2.53381487908618e+01, 2.53380424187140e+01, 2.53380134828932e+01, 2.53380092120261e+01, 2.53380088886489e+01, 2.53380031600545e+01, 2.53379643479834e+01, 2.53378216706904e+01, 2.53374413453033e+01, 2.53366090119477e+01, 2.53350149918930e+01, 2.53322429710357e+01, 2.53277626305441e+01, 2.53209266656594e+01, 2.53109725391695e+01, 2.52970292047513e+01, 2.52781289038267e+01, 2.52532239852047e+01, 2.52212085188853e+01, 2.51809442762667e+01, 2.51312904348476e+01, 2.50711361470379e+01, 2.49994349049690e+01, 2.49152394548503e+01, 2.48177358858115e+01, 2.47062754587895e+01, 2.45804027662944e+01, 2.44398789318499e+01, 2.42846987664950e+01, 2.41151010854061e+01, 2.39315717260688e+01, 2.37348391683840e+01, 2.35258630020734e+01, 2.33058157875325e+01, 2.30760590933310e+01, 2.28381146621963e+01, 2.25936317675056e+01, 2.23443518933547e+01, 2.20920719222908e+01, 2.18386070542068e+01, 2.15857546984736e+01, 2.13352605534382e+01, 2.10887879813325e+01, 2.08478915815826e+01, 2.06139955666123e+01, 2.03883771877196e+01, 2.01721551029917e+01, 1.99662822857463e+01, 1.97715428837801e+01, 1.95885523694756e+01, 1.94177603515786e+01, 1.92594555157584e+01, 1.91137722836831e+01, 1.89806988978677e+01, 1.88600867334481e+01, 1.87516607019941e+01, 1.86550306484140e+01, 1.85697036556201e+01, 1.84950971690556e+01, 1.84305528393043e+01, 1.83753509591035e+01, 1.83287253432987e+01, 1.82898784683138e+01, 1.82579966534230e+01, 1.82322650318883e+01, 1.82118820288249e+01, 1.81960730378794e+01, 1.81841029739014e+01, 1.81752873769290e+01, 1.81690017563866e+01, 1.81646888947675e+01, 1.81618638772503e+01, 1.81601166763058e+01, 1.81591121956751e+01, 1.81585877623194e+01, 1.81583481434875e+01, 1.81582582540455e+01, 1.81582338019240e+01, 1.81582301928416e+01, 1.81582299219066e+01, 1.81582251531441e+01, 1.81581928440756e+01, 1.81580740725072e+01, 1.81577574709977e+01, 1.81570645957548e+01, 1.81557376545390e+01, 1.81534300856279e+01, 1.81497004219581e+01, 1.81440098072225e+01, 1.81357234521142e+01, 1.81241162265131e+01, 1.81083824745281e+01, 1.80876500119054e+01, 1.80609981189259e+01, 1.80274791784468e+01, 1.79861434332644e+01, 1.79360661580500e+01, 1.78763763707751e+01, 1.78062860615554e+01, 1.77251188092876e+01, 1.76323366036480e+01, 1.75275637040031e+01, 1.74106064536816e+01, 1.72814681261344e+01, 1.71403580982517e+01, 1.69876949072789e+01, 1.68241030282917e+01, 1.66504034855853e+01, 1.64675986643976e+01, 1.62768519075854e+01, 1.60794626619196e+01, 1.58768380830893e+01, 1.56704621210295e+01, 1.54618631881599e+01, 1.52525815571675e+01, 1.50441376319723e+01, 1.48380021747973e+01, 1.46355694477645e+01, 1.44381340420174e+01, 1.42468719342821e+01, 1.40628260519279e+01, 1.38868963693234e+01, 1.37198343262613e+01, 1.35622411737497e+01, 1.34145697256892e+01, 1.32771289299793e+01, 1.31500906645664e+01, 1.30334982022341e+01, 1.29272758588301e+01, 1.28312394287789e+01, 1.27451071060859e+01, 1.26685106779516e+01, 1.26010068538732e+01, 1.25420886509195e+01, 1.24911967935112e+01, 1.24477311034657e+01, 1.24110618549021e+01, 1.23805410517425e+01, 1.23555135568370e+01, 1.23353279656485e+01, 1.23193470787855e+01, 1.23069577913531e+01, 1.22975801877331e+01, 1.22906756120215e+01, 1.22857534800566e+01, 1.22823766106452e+01, 1.22801648817189e+01, 1.22787970607476e+01, 1.22780107154246e+01, 1.22776001769164e+01, 1.22774125995355e+01, 1.22773422329546e+01, 1.22773230915546e+01, 1.22773202663246e+01, 1.22773200548787e+01, 1.22773163417818e+01, 1.22772911850147e+01, 1.22771987062339e+01, 1.22769521929586e+01, 1.22764127107773e+01, 1.22753795612093e+01, 1.22735829785214e+01, 1.22706794195516e+01, 1.22662497544728e+01, 1.22598006109354e+01, 1.22507690558312e+01, 1.22385307145586e+01, 1.22224113242927e+01, 1.22017015940502e+01, 1.21756751012416e+01, 1.21436087959025e+01, 1.21048055172035e+01, 1.20586177628206e+01, 1.20044718038267e+01, 1.19418911213011e+01, 1.18705180715924e+01, 1.17901326793937e+01, 1.17006675222683e+01, 1.16022178122714e+01, 1.14950459979065e+01, 1.13795804926430e+01, 1.12564084659743e+01, 1.11262629835364e+01, 1.09900051230912e+01, 1.08486019906718e+01, 1.07031017861723e+01, 1.05546071977468e+01, 1.04042484281801e+01, 1.02531570756335e+01, 1.01024419205224e+01, 9.95316743482567e+00, 9.80633556096763e+00, 9.66287103640298e+00, 9.52361029477206e+00, 9.38929377472300e+00, 9.26056132345086e+00, 9.13795029479874e+00, 9.02189590537302e+00, 8.91273341614986e+00, 8.81070173954816e+00, 8.71594812164720e+00, 8.62853360689662e+00, 8.54843905156719e+00, 8.47557150790574e+00, 8.40977085083667e+00, 8.35081656159561e+00, 8.29843461702100e+00, 8.25230445881949e+00, 8.21206603362140e+00, 8.17732690193510e+00, 8.14766941240726e+00, 8.12265792774533e+00, 8.10184607142623e+00, 8.08478394160759e+00, 8.07102521268214e+00, 8.06013401826984e+00, 8.05169148493705e+00, 8.04530176637590e+00, 8.04059741571232e+00, 8.03724393107156e+00, 8.03494331779743e+00, 8.03343653017357e+00, 8.03250468552101e+00, 8.03196898258258e+00, 8.03168930178107e+00, 8.03156151431991e+00, 8.03151357699441e+00, 8.03150053689496e+00, 8.03149861220415e+00, 8.03149846810247e+00, 8.03149593688431e+00, 8.03147878751796e+00, 8.03141574476800e+00, 8.03124769697874e+00, 8.03087993375922e+00, 8.03017564334169e+00, 8.02895093942999e+00, 8.02697166154154e+00, 8.02395216236027e+00, 8.01955625847487e+00, 8.01340047447162e+00, 8.00505965193464e+00, 7.99407492220039e+00, 7.97996395360650e+00, 7.96223328121462e+00, 7.94039241277039e+00, 7.91396928496494e+00, 7.88252652760546e+00, 7.84567789103149e+00, 7.80310411614788e+00, 7.75456748856050e+00, 7.69992432810400e+00, 7.63913472817654e+00, 7.57226897593689e+00, 7.49951024861836e+00, 7.42115338113730e+00, 7.33759971943537e+00, 7.24934829384102e+00, 7.15698374861481e+00, 7.06116163180845e+00, 6.96259177194742e+00, 6.86202053813000e+00, 6.76021279590712e+00, 6.65793433468281e+00, 6.55593545843264e+00, 6.45493630814617e+00, 6.35561433220634e+00, 6.25859415325016e+00, 6.16443991219545e+00, 6.07365001760927e+00, 5.98665410483999e+00, 5.90381192322678e+00, 5.82541382416150e+00, 5.75168251476727e+00, 5.68277576388792e+00, 5.61878978904727e+00, 5.55976310508758e+00, 5.50568066900832e+00, 5.45647820520990e+00, 5.41204663746943e+00, 5.37223658708280e+00, 5.33686292055781e+00, 5.30570934560379e+00, 5.27853306178496e+00, 5.25506947298953e+00, 5.23503696371350e+00, 5.21814173099011e+00, 5.20408264964320e+00, 5.19255613160657e+00, 5.18326092173462e+00, 5.17590275442827e+00, 5.17019877921638e+00, 5.16588165087606e+00, 5.16270317229648e+00, 5.16043737733398e+00, 5.15888294715226e+00, 5.15786486719527e+00, 5.15723525257636e+00, 5.15687329623904e+00, 5.15668432518038e+00, 5.15659798337516e+00, 5.15656559368546e+00, 5.15655678291417e+00, 5.15655548246314e+00, 5.15655538484001e+00, 5.15655366659138e+00, 5.15654202520994e+00, 5.15649923034870e+00, 5.15638515563942e+00, 5.15613550937208e+00, 5.15565742009981e+00, 5.15482606013028e+00, 5.15348247222198e+00, 5.15143274396608e+00, 5.14844864895787e+00, 5.14426984217867e+00, 5.13860765714503e+00, 5.13115050278203e+00, 5.12157079787788e+00, 5.10953331090468e+00, 5.09470469516491e+00, 5.07676392781593e+00, 5.05541328243710e+00, 5.03038939610965e+00, 5.00147394198235e+00, 4.96850339523187e+00, 4.93137739083940e+00, 4.89006521938865e+00, 4.84461009179327e+00, 4.79513092051855e+00, 4.74182150410755e+00, 4.68494715094587e+00, 4.62483892295258e+00, 4.56188580668022e+00, 4.49652521733716e+00, 4.42923230401393e+00, 4.36050855027545e+00, 4.29087015605639e+00, 4.22083665037524e+00, 4.15092012717416e+00, 4.08161542604400e+00, 4.01339150221833e+00, 3.94668415125325e+00, 3.88189017732368e+00, 3.81936302329897e+00, 3.75940981839751e+00, 3.70228974748973e+00, 3.64821360669659e+00, 3.59734438375399e+00, 3.54979868870750e+00, 3.50564885994061e+00, 3.46492558053630e+00, 3.42762085810960e+00, 3.39369124477435e+00, 3.36306120002517e+00, 3.33562652544782e+00, 3.31125782413525e+00, 3.28980395779906e+00, 3.27109548968243e+00, 3.25494811088607e+00, 3.24116605149042e+00, 3.22954547621713e+00, 3.21987785802722e+00, 3.21195331303948e+00, 3.20556386776045e+00, 3.20050661631577e+00, 3.19658671270675e+00, 3.19362013259564e+00, 3.19143613209626e+00, 3.18987932858629e+00, 3.18881133134055e+00, 3.18811185803040e+00, 3.18767928656618e+00, 3.18743060962692e+00, 3.18730078037705e+00, 3.18724146089670e+00, 3.18721920820492e+00, 3.18721315494200e+00, 3.18721226149340e+00, 3.18721219413777e+00, 3.18721100481712e+00, 3.18720294699988e+00, 3.18717332568828e+00, 3.18709436673804e+00, 3.18692157000590e+00, 3.18659065508503e+00, 3.18601522722906e+00, 3.18508527941493e+00, 3.18366663269553e+00, 3.18160140003364e+00, 3.17870953737547e+00, 3.17479151842858e+00, 3.16963213530501e+00, 3.16300538528410e+00, 3.15468035474590e+00, 3.14442795628720e+00, 3.13202831694349e+00, 3.11727855838605e+00, 3.10000065910971e+00, 3.08004904975945e+00, 3.05731757161460e+00, 3.03174542977128e+00, 3.00332179999545e+00, 2.97208880245118e+00, 2.93814263460121e+00, 2.90163275467629e+00, 2.86275911876152e+00, 2.82176758944554e+00, 2.77894374199440e+00, 2.73460538543112e+00, 2.68909418261200e+00, 2.64276678990726e+00, 2.59598594120526e+00, 2.54911187394642e+00, 2.50249444122906e+00, 2.45646618067232e+00, 2.41133652607743e+00, 2.36738726068810e+00, 2.32486922890434e+00, 2.28400025288636e+00, 2.24496414567968e+00, 2.20791067511460e+00, 2.17295631254727e+00, 2.14018559568333e+00, 2.10965294238715e+00, 2.08138476918622e+00, 2.05538179077313e+00, 2.03162140216289e+00, 2.01006007077284e+00, 1.99063568964586e+00, 1.97326986399126e+00, 1.95787012033809e+00, 1.94433204044597e+00, 1.93254133058591e+00, 1.92237584101007e+00, 1.91370755068517e+00, 1.90640452913237e+00, 1.90033288109690e+00, 1.89535867148517e+00, 1.89134981838993e+00, 1.88817793199112e+00, 1.88572006763348e+00, 1.88386035338827e+00, 1.88249144676772e+00, 1.88151577267471e+00, 1.88084649559671e+00, 1.88040818365614e+00, 1.88013713025136e+00, 1.87998131018125e+00, 1.87989996059209e+00, 1.87986279187061e+00, 1.87984884869343e+00, 1.87984505581998e+00, 1.87984449600013e+00, 1.87984445356656e+00, 1.87984370125650e+00, 1.87983860424997e+00, 1.87981986718684e+00, 1.87976992160949e+00, 1.87966061974447e+00, 1.87945130436201e+00, 1.87908733684121e+00, 1.87849915899994e+00, 1.87760195636027e+00, 1.87629598048991e+00, 1.87446757482916e+00, 1.87199093085927e+00, 1.86873057863184e+00, 1.86454458703309e+00, 1.85928841485868e+00, 1.85281931490368e+00, 1.84500115195527e+00, 1.83570945499358e+00, 1.82483648810641e+00, 1.81229609814284e+00, 1.79802808442873e+00, 1.78200184064303e+00, 1.76421904343553e+00, 1.74471520672028e+00, 1.72355998255579e+00, 1.70085616448200e+00, 1.67673743050216e+00, 1.65136494278709e+00, 1.62492299173068e+00, 1.59761392630811e+00, 1.56965264589327e+00, 1.54126093853934e+00, 1.51266193782486e+00, 1.48407493789401e+00, 1.45571075935290e+00, 1.42776780331139e+00, 1.40042887324012e+00, 1.37385878988227e+00, 1.34820277737018e+00, 1.32358556157261e+00, 1.30011109563628e+00, 1.27786281249197e+00, 1.25690429861461e+00, 1.23728028581729e+00, 1.21901786632547e+00, 1.20212784883586e+00, 1.18660618792413e+00, 1.17243543452422e+00, 1.15958617008744e+00, 1.14801840058220e+00, 1.13768289813560e+00, 1.12852248749343e+00, 1.12047328140934e+00, 1.11346587353237e+00, 1.10742649941641e+00, 1.10227817608982e+00, 1.09794182844743e+00, 1.09433740688510e+00, 1.09138499549737e+00, 1.08900590427475e+00, 1.08712373260472e+00, 1.08566538556752e+00, 1.08456201958776e+00, 1.08374989047820e+00, 1.08317107521600e+00, 1.08277403920968e+00, 1.08251402345322e+00, 1.08235323073448e+00, 1.08226079668341e+00, 1.08221253945963e+00, 1.08219049071936e+00, 1.08218221953778e+00, 1.08217996958170e+00, 1.08217963749311e+00, 1.08217961216154e+00, 1.08217916094760e+00, 1.08217610390996e+00, 1.08216486596487e+00, 1.08213491010301e+00, 1.08206935432846e+00, 1.08194381444424e+00, 1.08172552226990e+00, 1.08137276512234e+00, 1.08083468785401e+00, 1.08005149288741e+00, 1.07895506582317e+00, 1.07747004360035e+00, 1.07551532822474e+00, 1.07300603140320e+00, 1.06985581412243e+00, 1.06597956097882e+00, 1.06129630327628e+00, 1.05573227961487e+00, 1.04922400050994e+00, 1.04172116742077e+00, 1.03318928926318e+00, 1.02361184332893e+00, 1.01299184383179e+00, 1.00135270997787e+00, 9.88738364876909e-01, 9.75212543641646e-01, 9.60857339348550e-01, 9.45771064229973e-01, 9.30065545688791e-01, 9.13863008400838e-01, 8.97292712177261e-01, 8.80487519353835e-01, 8.63580555924187e-01, 8.46702109589742e-01, 8.29976878540947e-01, 8.13521650788573e-01, 7.97443458863640e-01, 7.81838221856987e-01, 7.66789858467646e-01, 7.52369832474130e-01, 7.38637076466020e-01, 7.25638230659982e-01, 7.13408130505087e-01, 7.01970478535202e-01, 6.91338641366514e-01, 6.81516520670769e-01, 6.72499456277549e-01, 6.64275129333674e-01, 6.56824442924545e-01, 6.50122366190999e-01, 6.44138735375861e-01, 6.38839011184785e-01, 6.34184996236570e-01, 6.30135519198234e-01, 6.26647093503934e-01, 6.23674558463122e-01, 6.21171709232641e-01, 6.19091919769289e-01, 6.17388759747101e-01, 6.16016602803938e-01, 6.14931219698718e-01, 6.14090346355823e-01, 6.13454213680924e-01, 6.12986023771143e-01, 6.12652355963689e-01, 6.12423486246982e-01, 6.12273604965992e-01, 6.12180920436842e-01, 6.12127639883425e-01, 6.12099823748821e-01, 6.12087114582010e-01, 6.12082346975981e-01, 6.12081050075432e-01, 6.12080858655787e-01, 6.12080843951333e-01, 6.12080580680313e-01, 6.12078796982686e-01, 6.12072239954244e-01, 6.12054761575322e-01, 6.12016511846662e-01, 6.11943263930129e-01, 6.11815900488949e-01, 6.11610088381354e-01, 6.11296166373037e-01, 6.10839267251336e-01, 6.10199691208464e-01, 6.09333541141353e-01, 6.08193622186462e-01, 6.06730597137627e-01, 6.04894376430906e-01, 6.02635706546209e-01, 5.99907904828285e-01, 5.96668673160352e-01, 5.92881909274965e-01, 5.88519424594665e-01, 5.83562473143729e-01, 5.78002998703454e-01, 5.71844517804124e-01, 5.65102574293064e-01, 5.57804726031236e-01, 5.49990053744224e-01, 5.41708213432017e-01, 5.33018083889388e-01, 5.23986086727419e-01, 5.14684275244250e-01, 5.05188298873406e-01, 4.95575351129929e-01, 4.85922201481642e-01, 4.76303396860501e-01, 4.66789698729485e-01, 4.57446799192823e-01, 4.48334336990527e-01, 4.39505213442374e-01, 4.31005191044761e-01, 4.22872744397523e-01, 4.15139124746872e-01, 4.07828595450908e-01, 4.00958795510064e-01, 3.94541191130926e-01, 3.88581580215643e-01, 3.83080620816116e-01, 3.78034361203817e-01, 3.73434755676683e-01, 3.69270156118802e-01, 3.65525774373126e-01, 3.62184114547674e-01, 3.59225377405224e-01, 3.56627841015694e-01, 3.54368222943449e-01, 3.52422029490853e-01, 3.50763897036131e-01, 3.49367929403456e-01, 3.48208033619116e-01, 3.47258254474895e-01, 3.46493106195581e-01, 3.45887897348305e-01, 3.45419043118000e-01, 3.45064357371228e-01, 3.44803315704283e-01, 3.44617280053001e-01, 3.44489675522553e-01, 3.44406110909034e-01, 3.44354435906368e-01, 3.44324730128108e-01, 3.44309221677790e-01, 3.44302135885556e-01, 3.44299477784006e-01, 3.44298754718196e-01, 3.44298647995274e-01, 3.44298639734490e-01, 3.44298491017341e-01, 3.44297483438140e-01, 3.44293779489959e-01, 3.44283906273944e-01, 3.44262299743150e-01, 3.44220923567491e-01, 3.44148979173454e-01, 3.44032722392626e-01, 3.43855400781123e-01, 3.43597323839376e-01, 3.43236075808434e-01, 3.42746877160025e-01, 3.42103096092856e-01, 3.41276905152498e-01, 3.40240070536836e-01, 3.38964852988951e-01, 3.37424989934959e-01, 3.35596719485778e-01, 3.33459799086526e-01, 3.30998466087337e-01, 3.28202285392095e-01, 3.25066831470195e-01, 3.21594158820517e-01, 3.17793026337611e-01, 3.13678856189725e-01, 3.09273425442186e-01, 3.04604306969708e-01, 2.99704093240209e-01, 2.94609450476628e-01, 2.89360060069285e-01, 2.83997508095524e-01, 2.78564182303683e-01, 2.73102229531474e-01, 2.67652616414892e-01, 2.62254323908039e-01, 2.56943693154838e-01, 2.51753928050633e-01, 2.46714749463052e-01, 2.41852188177603e-01, 2.37188498404208e-01, 2.32742171002712e-01, 2.28528025100147e-01, 2.24557357985631e-01, 2.20838135528723e-01, 2.17375208348130e-01, 2.14170542124009e-01, 2.11223453489552e-01, 2.08530845670580e-01, 2.06087440371320e-01, 2.03886004294944e-01, 2.01917570130405e-01, 2.00171652849079e-01, 1.98636462762153e-01, 1.97299117032301e-01, 1.96145851249227e-01, 1.95162232306146e-01, 1.94333373196244e-01, 1.93644149525753e-01, 1.93079416574989e-01, 1.92624224696047e-01, 1.92264029797471e-01, 1.91984894722140e-01, 1.91773676557872e-01, 1.91618194421011e-01, 1.91507372082285e-01, 1.91431350011769e-01, 1.91381562018194e-01, 1.91350772629946e-01, 1.91333072660186e-01, 1.91323831933474e-01, 1.91319609826792e-01, 1.91318025978148e-01, 1.91317595133782e-01, 1.91317531542083e-01, 1.91317526579923e-01, 1.91317436731694e-01, 1.91316827997284e-01, 1.91314590233972e-01, 1.91308625246094e-01, 1.91295571358623e-01, 1.91270572915087e-01, 1.91227104518887e-01, 1.91156859056917e-01, 1.91049707605131e-01, 1.90893738112777e-01, 1.90675378164642e-01, 1.90379604985580e-01, 1.89990243059498e-01, 1.89490346216047e-01, 1.88862656821328e-01, 1.88090129933826e-01, 1.87156505245625e-01, 1.86046904742369e-01, 1.84748429826229e-01, 1.83250728738557e-01, 1.81546504054594e-01, 1.79631931222460e-01, 1.77506962774459e-01, 1.75175498844041e-01, 1.72645412544028e-01, 1.69928427877380e-01, 1.67039857244518e-01, 1.63998214290899e-01, 1.60824724914676e-01, 1.57542764066037e-01, 1.54177248213614e-01, 1.50754013093689e-01, 1.47299203981504e-01, 1.43838701789619e-01, 1.40397603399450e-01, 1.36999769310634e-01, 1.33667446423672e-01, 1.30420968976409e-01, 1.27278536664164e-01, 1.24256065949268e-01, 1.21367108453086e-01, 1.18622828916628e-01, 1.16032034309342e-01, 1.13601245190040e-01, 1.11334800411527e-01, 1.09234986757940e-01, 1.07302186029216e-01, 1.05535033262782e-01, 1.03930581020649e-01, 1.02484465857998e-01, 1.01191074206142e-01, 1.00043705940848e-01, 9.90347348267591e-02, 9.81557657645880e-02, 9.73977892410574e-02, 9.67513335677239e-02, 9.62066154000070e-02, 9.57536887004693e-02, 9.53825918005009e-02, 9.50834915828551e-02, 9.48468231252946e-02, 9.46634224712000e-02, 9.45246496215901e-02, 9.44224984232420e-02, 9.43496898360365e-02, 9.42997451254683e-02, 9.42670358679010e-02, 9.42468082516683e-02, 9.42351799923671e-02, 9.42291091704419e-02, 9.42263354022685e-02, 9.42252948726292e-02, 9.42250118238945e-02, 9.42249700465823e-02, 9.42249667588868e-02, 9.42249068743308e-02, 9.42245011479486e-02, 9.42230096614907e-02, 9.42190339549711e-02, 9.42103334664921e-02, 9.41936719408716e-02, 9.41647003826915e-02, 9.41178826295328e-02, 9.40464690146143e-02, 9.39425228154818e-02, 9.37970031449566e-02, 9.35999065545872e-02, 9.33404677879963e-02, 9.30074178286446e-02, 9.25892946050482e-02, 9.20747985796595e-02, 9.14531820878182e-02, 9.07146579997971e-02, 8.98508103596640e-02, 8.88549874844529e-02, 8.77226569292446e-02, 8.64517020226430e-02, 8.50426415172336e-02, 8.34987572836070e-02, 8.18261197482002e-02, 8.00335066136739e-02, 7.81322168368916e-02, 7.61357882944509e-02, 7.40596334098882e-02, 7.19206116510804e-02, 6.97365608298099e-02, 6.75258104702721e-02, 6.53067003107708e-02, 6.30971255363448e-02, 6.09141276670329e-02, 5.87735461579655e-02, 5.66897406864229e-02, 5.46753884199944e-02, 5.27413552484575e-02, 5.08966360901155e-02, 4.91483570752109e-02, 4.75018311717292e-02, 4.59606576146427e-02, 4.45268542707709e-02, 4.32010111207006e-02, 4.19824537095934e-02, 4.08694074483265e-02, 3.98591564528772e-02, 3.89481931804738e-02, 3.81323567306140e-02, 3.74069585170815e-02, 3.67668945434616e-02, 3.62067445214223e-02, 3.57208592115796e-02, 3.53034380797510e-02, 3.49485994670573e-02, 3.46504455008159e-02, 3.44031237080884e-02, 3.42008865606829e-02, 3.40381497219937e-02, 3.39095489190413e-02, 3.38099948641286e-02, 3.37347251634696e-02, 3.36793515910898e-02, 3.36399008492521e-02, 3.36128468807715e-02, 3.35951325913157e-02, 3.35841793817906e-02, 3.35778832023692e-02, 3.35745962650407e-02, 3.35730944917429e-02, 3.35725311341548e-02, 3.35723778873529e-02, 3.35723552668838e-02, 3.35723534754209e-02, 3.35723205655540e-02, 3.35720975975562e-02, 3.35712779568809e-02, 3.35690931496657e-02, 3.35643120443336e-02, 3.35551567879975e-02, 3.35392392305778e-02, 3.35135216643891e-02, 3.34743053100226e-02, 3.34172498636178e-02, 3.33374267636373e-02, 3.32294081964619e-02, 3.30873925627628e-02, 3.29053656072205e-02, 3.26772944498259e-02, 3.23973493066632e-02, 3.20601452800133e-02, 3.16609938582104e-02, 3.11961519053057e-02, 3.06630537847354e-02, 3.00605114664023e-02, 2.93888672169658e-02, 2.86500856049514e-02, 2.78477758172215e-02, 2.69776512960647e-02, 2.60663685300671e-02, 2.51113411267596e-02, 2.41215220061122e-02, 2.31065882272077e-02, 2.20766160270784e-02, 2.10417488467414e-02, 2.00118780366694e-02, 1.89963540687899e-02, 1.80037425245274e-02, 1.70416344566036e-02, 1.61165157843470e-02, 1.52336959076251e-02, 1.43972920085389e-02, 1.36102624615404e-02, 1.28744803181604e-02, 1.21908362376075e-02, 1.15593599348784e-02, 1.09793503136899e-02, 1.04495064097421e-02, 9.96805323819672e-03, 9.53285808955684e-03, 9.14153387701333e-03, 8.79152728460042e-03, 8.48019093739083e-03, 8.20484034842650e-03, 7.96279749377504e-03, 7.75142325815636e-03, 7.56814083972301e-03, 7.41045189013789e-03, 7.27594699004896e-03, 7.16231204656026e-03, 7.06733220410172e-03, 6.98889473068220e-03, 6.92499204035357e-03, 6.87372560469157e-03, 6.83331108123753e-03, 6.80208461638188e-03, 6.77850998418705e-03, 6.76118600102882e-03, 6.74885350167642e-03, 6.74040107421092e-03, 6.73486873502327e-03, 6.73144878226140e-03, 6.72948319579803e-03, 6.72845714115056e-03, 6.72798836451705e-03, 6.72781251608967e-03, 6.72776468159952e-03, 6.72775762134847e-03, 6.72775705567119e-03, 6.72774662392328e-03, 6.72767594758739e-03, 6.72741613873029e-03, 6.72672362051685e-03, 6.72520824794276e-03, 6.72230684322983e-03, 6.71726351295376e-03, 6.70911815298325e-03, 6.69670457182292e-03, 6.67865963495728e-03, 6.65344469717866e-03, 6.61938029269038e-03, 6.57469454688453e-03, 6.51758502954181e-03, 6.44629279575781e-03, 6.35918620518794e-03, 6.25485086013100e-03, 6.13218078137834e-03, 5.99046488480326e-03, 5.82946208770352e-03, 5.64945814102202e-03, 5.45129775873932e-03, 5.23638700788354e-03, 5.00666330843558e-03, 4.76453356392881e-03, 4.51278430256252e-03, 4.25447042610472e-03, 3.99279062270352e-03, 3.73095774735077e-03, 3.47207223759938e-03, 3.21900664559394e-03, 2.97430946609722e-03, 2.74013540365559e-03, 2.51820588512430e-03, 2.30979850107360e-03, 2.11575966410749e-03, 1.93653372430447e-03, 1.77220414681756e-03, 1.62254509517368e-03, 1.48708150489085e-03, 1.36515239847768e-03, 1.25596947593465e-03, 1.15927366886126e-03, 1.07285460806850e-03, 9.96026413474738e-04, 9.29601931247248e-04, 8.70750357250927e-04, 8.18860106036823e-04, 7.74157235980510e-04, 7.35770914758405e-04, 7.02352433293640e-04, 6.73315491937387e-04, 6.48096484781036e-04, 6.27403923871839e-04, 6.10321892565935e-04, 5.95866163394021e-04, 5.83746327123436e-04, 5.73694340829434e-04, 5.65470475223151e-04, 5.58852606104665e-04, 5.53628613118917e-04, 5.49593980721852e-04, 5.46553464275235e-04, 5.44324807108865e-04, 5.42742553392372e-04, 5.41660710581482e-04, 5.40953947904654e-04, 5.40517624997960e-04, 5.40267064963018e-04, 5.40136333995633e-04, 5.40076621348362e-04, 5.40054224231312e-04, 5.40048131981052e-04, 5.40047232789220e-04, 5.40047160228596e-04, 5.40045815654366e-04, 5.40036705970692e-04, 5.40003217466278e-04, 5.39913946937531e-04, 5.39718568063563e-04, 5.39344347828319e-04, 5.38693439139593e-04, 5.37641067353434e-04, 5.36034740179048e-04, 5.33694643578997e-04, 5.30415356024605e-04, 5.25968516157604e-04, 5.20104558574985e-04, 5.12819437389567e-04, 5.04350218164973e-04, 4.93938989265530e-04, 4.81430826354767e-04, 4.66718300785374e-04, 4.49753177093726e-04, 4.30563986910215e-04, 4.09241867970422e-04, 3.86840326243032e-04, 3.63498718527921e-04, 3.38721030914257e-04, 3.12917987301375e-04, 2.86726416860006e-04, 2.61673608108026e-04, 2.36783376763583e-04, 2.12453599132800e-04, 1.89769609576729e-04, 1.68621353822118e-04, 1.48620443823688e-04, 1.30516935328728e-04, 1.14398028770779e-04, 9.96706660032487e-05, 8.67436690867862e-05, 7.53655185263824e-05, 6.49294549077695e-05, 5.62588199244374e-05, 4.86329105910954e-05, 4.15552073462777e-05, 3.59161666949603e-05, 3.06312865632208e-05, 2.48325506078118e-05, 1.96724103034285e-05, 1.53184933822104e-05, 1.10209367250036e-05, 7.04841822596336e-06, 3.56295750236768e-06, 2.84925516780277e-06, 1.97593669781782e-06, 4.20467941353309e-07, 1.56663683981151e-06, 1.86054380435507e-06, 1.73117292945335e-06, 1.45694104958238e-06, 1.21703462241385e-06, 1.08459417447193e-06, 1.03843829890593e-06, 1.01690538656385e-06, 9.72619659478369e-07, 8.88904466478380e-07, 7.71218594763923e-07, 6.34772391094884e-07, 4.95827251256742e-07, -1.42108547152020e-14, -1.42108547152020e-14, -2.84217094304040e-14, -1.42108547152020e-14, -1.42108547152020e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14],
[ 6.60000000000000e+01, 6.59999999310424e+01, 6.59999977987352e+01, 6.59999833520167e+01, 6.59999302439897e+01, 6.59997886747130e+01, 6.59994788397302e+01, 6.59988854147067e+01, 6.59978532747447e+01, 6.59961846088103e+01, 6.59936375466183e+01, 6.59899263690116e+01, 6.59847233253809e+01, 6.59776620351613e+01, 6.59683424071010e+01, 6.59563369717024e+01, 6.59411984903494e+01, 6.59224686796645e+01, 6.58996878710622e+01, 6.58724054116403e+01, 6.58401906009209e+01, 6.58026439454397e+01, 6.57594084969479e+01, 6.57101810182181e+01, 6.56547226932419e+01, 6.55928690685304e+01, 6.55245388844054e+01, 6.54497414367747e+01, 6.53685821089493e+01, 6.52812657368898e+01, 6.51880975246588e+01, 6.50894813106745e+01, 6.49859150957596e+01, 6.48779838725901e+01, 6.47663499311863e+01, 6.46517409432497e+01, 6.45349362367558e+01, 6.44167517512868e+01, 6.42980242082422e+01, 6.41795950371300e+01, 6.40622945728458e+01, 6.39469269858026e+01, 6.38342563355215e+01, 6.37249940577970e+01, 6.36197881138692e+01, 6.35192139534508e+01, 6.34237673760669e+01, 6.33338593188852e+01, 6.32498125541540e+01, 6.31718602443175e+01, 6.31001462758799e+01, 6.30347272719466e+01, 6.29755761659738e+01, 6.29225872039055e+01, 6.28755822273366e+01, 6.28343180759186e+01, 6.27984949327285e+01, 6.27677654219683e+01, 6.27417442547486e+01, 6.27200182066410e+01, 6.27021562011763e+01, 6.26877192675565e+01, 6.26762701395872e+01, 6.26673822671339e+01, 6.26606480219699e+01, 6.26556858971476e+01, 6.26521465230652e+01, 6.26497173539215e+01, 6.26481259145565e+01, 6.26471415387147e+01, 6.26465755741637e+01, 6.26462800762623e+01, 6.26461450577706e+01, 6.26460944071779e+01, 6.26460806289419e+01, 6.26460785953014e+01, 6.26460783285307e+01, 6.26460721130925e+01, 6.26460300025661e+01, 6.26458751995466e+01, 6.26454625494606e+01, 6.26445594643261e+01, 6.26428299079746e+01, 6.26398220689625e+01, 6.26349602633801e+01, 6.26275415131620e+01, 6.26167371359508e+01, 6.26015995592668e+01, 6.25810744338927e+01, 6.25540179686315e+01, 6.25192192413863e+01, 6.24754270613527e+01, 6.24213807667882e+01, 6.23558441467171e+01, 6.22776414796276e+01, 6.21856944975583e+01, 6.20790589242273e+01, 6.19569591205871e+01, 6.18188193245966e+01, 6.16642900198704e+01, 6.14932681315320e+01, 6.13059100357158e+01, 6.11026367696688e+01, 6.08841313050036e+01, 6.06513282372740e+01, 6.04053966786431e+01, 6.01477174504802e+01, 5.98798558175148e+01, 5.96035309819008e+01, 5.93205834034140e+01, 5.90329408016317e+01, 5.87425835072603e+01, 5.84515097253962e+01, 5.81617012768699e+01, 5.78750904709416e+01, 5.75935288717427e+01, 5.73187587779437e+01, 5.70523881848644e+01, 5.67958698273437e+01, 5.65504846430884e+01, 5.63173297128426e+01, 5.60973104908146e+01, 5.58911369794289e+01, 5.56993234368076e+01, 5.55221912155485e+01, 5.53598743852957e+01, 5.52123278574195e+01, 5.50793377850153e+01, 5.49605340440775e+01, 5.48554046101041e+01, 5.47633116320981e+01, 5.46835089786676e+01, 5.46151609944247e+01, 5.45573621640420e+01, 5.45091573401149e+01, 5.44695621527285e+01, 5.44375831863496e+01, 5.44122374862287e+01, 5.43925709446689e+01, 5.43776751197840e+01, 5.43667020577530e+01, 5.43588767253342e+01, 5.43535067128124e+01, 5.43499889377457e+01, 5.43478131647785e+01, 5.43465622531996e+01, 5.43459091477428e+01, 5.43456107346110e+01, 5.43454987889193e+01, 5.43454683369180e+01, 5.43454638422642e+01, 5.43454634454719e+01, 5.43454556705206e+01, 5.43454029940711e+01, 5.43452093497073e+01, 5.43446931636777e+01, 5.43435634968213e+01, 5.43414000252846e+01, 5.43376376538490e+01, 5.43315564563693e+01, 5.43222775227169e+01, 5.43087651585646e+01, 5.42898357301318e+01, 5.42641732672282e+01, 5.42303517321145e+01, 5.41868636285983e+01, 5.41321543687352e+01, 5.40646615415757e+01, 5.39828579531561e+01, 5.38852970484487e+01, 5.37706591079262e+01, 5.36377964597302e+01, 5.34857758878938e+01, 5.33139164663130e+01, 5.31218212144154e+01, 5.29094012450834e+01, 5.26768914322662e+01, 5.24248570247425e+01, 5.21541910287366e+01, 5.18661025385737e+01, 5.15620964937933e+01, 5.12439455890136e+01, 5.09136552814871e+01, 5.05734230531984e+01, 5.02255932956738e+01, 4.98726093778549e+01, 4.95169645920579e+01, 4.91611537075860e+01, 4.88076267677355e+01, 4.84587465424903e+01, 4.81167507239829e+01, 4.77837195732894e+01, 4.74615493503523e+01, 4.71519315312027e+01, 4.68563375684825e+01, 4.65760087927392e+01, 4.63119509759203e+01, 4.60649330666543e+01, 4.58354896368388e+01, 4.56239266297929e+01, 4.54303300555347e+01, 4.52545773280712e+01, 4.50963509773577e+01, 4.49551544927382e+01, 4.48303300651327e+01, 4.47210779928345e+01, 4.46264775016893e+01, 4.45455087060767e+01, 4.44770754043132e+01, 4.44200283632615e+01, 4.43731887052543e+01, 4.43353709698350e+01, 4.43054053877895e+01, 4.42821588802235e+01, 4.42645542855873e+01, 4.42515873264469e+01, 4.42423408582323e+01, 4.42359959954459e+01, 4.42318397864391e+01, 4.42292692036498e+01, 4.42277913282075e+01, 4.42270197308101e+01, 4.42266671785655e+01, 4.42265349235583e+01, 4.42264989469435e+01, 4.42264936368689e+01, 4.42264932109352e+01, 4.42264853502984e+01, 4.42264320933205e+01, 4.42262363149289e+01, 4.42257144407932e+01, 4.42245723278327e+01, 4.42223850288643e+01, 4.42185812474746e+01, 4.42124331908480e+01, 4.42030524142391e+01, 4.41893921185874e+01, 4.41702562089248e+01, 4.41443152420456e+01, 4.41101291847178e+01, 4.40661766681513e+01, 4.40108901636866e+01, 4.39426962265601e+01, 4.38600596721426e+01, 4.37615302803309e+01, 4.36457903909312e+01, 4.35117015801027e+01, 4.33583485185312e+01, 4.31850781248626e+01, 4.29915322538539e+01, 4.27776723974586e+01, 4.25437952159394e+01, 4.22905381307545e+01, 4.20188746689542e+01, 4.17300997152507e+01, 4.14258052718588e+01, 4.11078477254545e+01, 4.07783079627513e+01, 4.04394459546581e+01, 4.00936516358854e+01, 3.97433940267828e+01, 3.93911705540778e+01, 3.90394584052109e+01, 3.86906694912198e+01, 3.83471102183002e+01, 3.80109468307275e+01, 3.76841766568445e+01, 3.73686052287363e+01, 3.70658289931161e+01, 3.67772231908506e+01, 3.65039344336062e+01, 3.62468775141584e+01, 3.60067360199718e+01, 3.57839663564108e+01, 3.55788048171220e+01, 3.53912773640817e+01, 3.52212118015472e+01, 3.50682520494336e+01, 3.49318742430256e+01, 3.48114044055669e+01, 3.47060374548849e+01, 3.46148573112959e+01, 3.45368578690068e+01, 3.44709645760941e+01, 3.44160563398333e+01, 3.43709874374285e+01, 3.43346090713807e+01, 3.43057901691884e+01, 3.42834369946524e+01, 3.42665111184678e+01, 3.42540452940145e+01, 3.42451568040011e+01, 3.42390578868870e+01, 3.42350629188444e+01, 3.42325921154438e+01, 3.42311716233946e+01, 3.42304299911472e+01, 3.42300911315625e+01, 3.42299640133608e+01, 3.42299294341020e+01, 3.42299243302750e+01, 3.42299239366187e+01, 3.42299168678159e+01, 3.42298689756351e+01, 3.42296929188914e+01, 3.42292236160895e+01, 3.42281965575617e+01, 3.42262296144546e+01, 3.42228090740125e+01, 3.42172805448087e+01, 3.42088452758981e+01, 3.41965623176658e+01, 3.41793568157264e+01, 3.41560345687490e+01, 3.41253027936790e+01, 3.40857968268231e+01, 3.40361122493820e+01, 3.39748416682786e+01, 3.39006151193787e+01, 3.38121428068884e+01, 3.37082586699313e+01, 3.35879630969897e+01, 3.34504630122762e+01, 3.32952075528666e+01, 3.31219176529523e+01, 3.29306080548630e+01, 3.27216005693143e+01, 3.24955277945785e+01, 3.22533269536522e+01, 3.19962239926248e+01, 3.17257085719165e+01, 3.14435010428793e+01, 3.11515129030411e+01, 3.08518025325257e+01, 3.05465282039354e+01, 3.02379004077701e+01, 2.99281354373294e+01, 2.96194119396800e+01, 2.93138317895155e+01, 2.90133862229328e+01, 2.87199277294626e+01, 2.84351477939232e+01, 2.81605602460341e+01, 2.78974897402802e+01, 2.76470647576914e+01, 2.74102144846999e+01, 2.71876689602062e+01, 2.69799619635511e+01, 2.67874362173642e+01, 2.66102505795923e+01, 2.64483889848542e+01, 2.63016709599853e+01, 2.61697635809364e+01, 2.60521947600645e+01, 2.59483677576437e+01, 2.58575768025431e+01, 2.57790236873172e+01, 2.57118351746468e+01, 2.56550810170895e+01, 2.56077923524390e+01, 2.55689801950987e+01, 2.55376537028090e+01, 2.55128378614454e+01, 2.54935902024247e+01, 2.54790161514984e+01, 2.54682826079359e+01, 2.54606293719789e+01, 2.54553780774204e+01, 2.54519383451213e+01, 2.54498109505235e+01, 2.54485878904594e+01, 2.54479493372469e+01, 2.54476575757231e+01, 2.54475481256945e+01, 2.54475183526133e+01, 2.54475139581682e+01, 2.54475136254341e+01, 2.54475077310822e+01, 2.54474677959790e+01, 2.54473209902895e+01, 2.54469296600222e+01, 2.54460732422939e+01, 2.54444330960173e+01, 2.54415808557455e+01, 2.54369708445226e+01, 2.54299369980147e+01, 2.54196946546660e+01, 2.54053474519667e+01, 2.53858994337818e+01, 2.53602723155466e+01, 2.53273276723486e+01, 2.52858936118904e+01, 2.52347952763890e+01, 2.51728882954404e+01, 2.50990941009316e+01, 2.50124358338944e+01, 2.49120734420655e+01, 2.47973365051819e+01, 2.46677533478324e+01, 2.45230751146958e+01, 2.43632936879100e+01, 2.41886526076060e+01, 2.39996504906001e+01, 2.37970367982507e+01, 2.35818001505763e+01, 2.33551496929686e+01, 2.31184902780510e+01, 2.28733924257132e+01, 2.26215581779068e+01, 2.23647840848547e+01, 2.21049226540673e+01, 2.18438436571837e+01, 2.15833966999660e+01, 2.13253763864153e+01, 2.10714912239824e+01, 2.08233371218443e+01, 2.05823759598041e+01, 2.03499193108785e+01, 2.01271170543496e+01, 1.99149503718204e+01, 1.97142284994265e+01, 1.95255886039415e+01, 1.93494982246810e+01, 1.91862598350026e+01, 1.90360171912945e+01, 1.88987632316071e+01, 1.87743493521590e+01, 1.86624959291679e+01, 1.85628039717581e+01, 1.84747677955514e+01, 1.83977886008632e+01, 1.83311888269427e+01, 1.82742271355815e+01, 1.82261138540902e+01, 1.81860266797071e+01, 1.81531264162391e+01, 1.81265724814056e+01, 1.81055378930747e+01, 1.80892234180321e+01, 1.80768705519529e+01, 1.80677729973204e+01, 1.80612863198524e+01, 1.80568354950581e+01, 1.80539201049659e+01, 1.80521170094374e+01, 1.80510803940330e+01, 1.80505391831575e+01, 1.80502918984762e+01, 1.80501991332994e+01, 1.80501738989074e+01, 1.80501701743635e+01, 1.80501698947608e+01, 1.80501649734363e+01, 1.80501316307380e+01, 1.80500090594330e+01, 1.80496823291926e+01, 1.80489672873244e+01, 1.80475978938133e+01, 1.80452164982693e+01, 1.80413675064972e+01, 1.80354948148218e+01, 1.80269433091797e+01, 1.80149646307491e+01, 1.79987272978438e+01, 1.79773311429515e+01, 1.79498258737252e+01, 1.79152333994486e+01, 1.78725733851566e+01, 1.78208913128723e+01, 1.77592881553982e+01, 1.76869506174920e+01, 1.76031807879815e+01, 1.75074239894471e+01, 1.73992936212416e+01, 1.72785918727911e+01, 1.71453253358289e+01, 1.69997147569418e+01, 1.68421984291967e+01, 1.66734290031855e+01, 1.64942637828754e+01, 1.63057488428896e+01, 1.61090975497713e+01, 1.59056642845408e+01, 1.56969143449581e+01, 1.54843911506021e+01, 1.52696819763413e+01, 1.50543834903725e+01, 1.48400683605171e+01, 1.46282541083068e+01, 1.44203752330937e+01, 1.42177594066116e+01, 1.40216082708470e+01, 1.38329830848620e+01, 1.36527951880906e+01, 1.34818010047870e+01, 1.33206011261896e+01, 1.31696428839350e+01, 1.30292257713636e+01, 1.28995090717565e+01, 1.27805211020372e+01, 1.26721695621890e+01, 1.25742525795859e+01, 1.24864701401568e+01, 1.24084356940862e+01, 1.23396878048658e+01, 1.22797017722547e+01, 1.22279011999511e+01, 1.21836694975783e+01, 1.21463613056753e+01, 1.21153138147834e+01, 1.20898579194326e+01, 1.20693291095075e+01, 1.20530779601532e+01, 1.20404800421994e+01, 1.20309450428944e+01, 1.20239248658238e+01, 1.20189204725336e+01, 1.20154872386322e+01, 1.20132386245646e+01, 1.20118480048751e+01, 1.20110485571501e+01, 1.20106311793410e+01, 1.20104404773315e+01, 1.20103689386438e+01, 1.20103494784069e+01, 1.20103466061175e+01, 1.20103463911495e+01, 1.20103426162042e+01, 1.20103170404065e+01, 1.20102230212553e+01, 1.20099724021519e+01, 1.20094239356526e+01, 1.20083735845864e+01, 1.20065471025753e+01, 1.20035952565521e+01, 1.19990920080191e+01, 1.19925359142716e+01, 1.19833549420220e+01, 1.19709148000163e+01, 1.19545307912321e+01, 1.19334830575613e+01, 1.19070349414535e+01, 1.18744540239589e+01, 1.18350352246787e+01, 1.17881251775833e+01, 1.17331469418462e+01, 1.16696239850520e+01, 1.15972023040816e+01, 1.15156695417334e+01, 1.14249700260297e+01, 1.13252148095332e+01, 1.12166860155167e+01, 1.10998350953226e+01, 1.09752749467032e+01, 1.08437662086996e+01, 1.07061984020019e+01, 1.05635668908451e+01, 1.04169468727749e+01, 1.02674657334652e+01, 1.01162751241892e+01, 9.96452403207790e+00, 9.81333393356445e+00, 9.66377687541620e+00, 9.51685704727097e+00, 9.37349612695246e+00, 9.23452242297003e+00, 9.10066362761538e+00, 8.97254283968041e+00, 8.85067741981636e+00, 8.73548019864958e+00, 8.62726255823473e+00, 8.52623893972105e+00, 8.43253238297093e+00, 8.34618076764660e+00, 8.26714349245417e+00, 8.19530839440478e+00, 8.13049876946040e+00, 8.07248040740964e+00, 8.02096859577505e+00, 7.97563507910288e+00, 7.93611498070238e+00, 7.90201370377634e+00, 7.87291382835146e+00, 7.84838201038921e+00, 7.82797587136196e+00, 7.81125084234920e+00, 7.79776689874305e+00, 7.78709509270264e+00, 7.77882376350876e+00, 7.77256428387113e+00, 7.76795618568773e+00, 7.76467150388510e+00, 7.76241818315285e+00, 7.76094241007110e+00, 7.76002976173886e+00, 7.75950509998638e+00, 7.75923118518978e+00, 7.75910603260885e+00, 7.75905908377372e+00, 7.75904631257285e+00, 7.75904442757113e+00, 7.75904428644096e+00, 7.75904180741916e+00, 7.75902501169294e+00, 7.75896326899431e+00, 7.75879868691596e+00, 7.75843850959581e+00, 7.75774875197175e+00, 7.75654933600777e+00, 7.75461097123740e+00, 7.75165399341753e+00, 7.74734933983545e+00, 7.74132179500127e+00, 7.73315558305821e+00, 7.72240231071648e+00, 7.70859117555415e+00, 7.69124124979434e+00, 7.66987553242485e+00, 7.64403633899917e+00, 7.61330147781842e+00, 7.57730055500298e+00, 7.53573067209607e+00, 7.48837074085652e+00, 7.43509365113884e+00, 7.37587559537702e+00, 7.31080197756864e+00, 7.24006950956725e+00, 7.16398431025834e+00, 7.08295605600714e+00, 6.99748846291160e+00, 6.90816659203633e+00, 6.81564163974016e+00, 6.72061399305254e+00, 6.62381538747451e+00, 6.52599100053710e+00, 6.42788225357480e+00, 6.33021098567706e+00, 6.23366552019172e+00, 6.13888898003948e+00, 6.04647003877069e+00, 5.95693613441820e+00, 5.87074903530577e+00, 5.78830254010110e+00, 5.70992202317859e+00, 5.63586550079339e+00, 5.56632588968933e+00, 5.50143415085875e+00, 5.44126304932745e+00, 5.38583130832372e+00, 5.33510798645113e+00, 5.28901695460532e+00, 5.24744139211380e+00, 5.21022825709478e+00, 5.17719271352810e+00, 5.14812251687411e+00, 5.12278237154139e+00, 5.10091827760414e+00, 5.08226188159739e+00, 5.06653483783218e+00, 5.05345317353436e+00, 5.04273163449590e+00, 5.03408796933834e+00, 5.02724709161025e+00, 5.02194504156839e+00, 5.01793265541040e+00, 5.01497884056339e+00, 5.01287335269845e+00, 5.01142897428446e+00, 5.01048300598023e+00, 5.00989800062294e+00, 5.00956169400675e+00, 5.00938611552391e+00, 5.00930589315064e+00, 5.00927579911212e+00, 5.00926761281750e+00, 5.00926640453805e+00, 5.00926631383413e+00, 5.00926471736923e+00, 5.00925390109210e+00, 5.00921413942451e+00, 5.00910815027233e+00, 5.00887619994421e+00, 5.00843200394341e+00, 5.00765959539481e+00, 5.00641131748905e+00, 5.00450707804561e+00, 5.00173498166197e+00, 4.99785342603787e+00, 4.99259471193084e+00, 4.98567016912862e+00, 4.97677674293338e+00, 4.96560491737483e+00, 4.95184777486465e+00, 4.93521091142211e+00, 4.91542284806334e+00, 4.89224551019106e+00, 4.86548429645551e+00, 4.83499723495645e+00, 4.80070273464754e+00, 4.76258548729989e+00, 4.72070016028350e+00, 4.67517263802971e+00, 4.62619871118297e+00, 4.57404026448526e+00, 4.51901916289794e+00, 4.46150916617739e+00, 4.40192630330620e+00, 4.34071820209477e+00, 4.27835289304542e+00, 4.21530759211907e+00, 4.15205792012343e+00, 4.08906794523611e+00, 4.02678134878120e+00, 3.96561392153211e+00, 3.90594750606798e+00, 3.84812541604604e+00, 3.79244929001592e+00, 3.73917727842345e+00, 3.68852341918450e+00, 3.64065802989744e+00, 3.59570893255828e+00, 3.55376332780853e+00, 3.51487014784585e+00, 3.47904273732161e+00, 3.44626173685658e+00, 3.41647807138170e+00, 3.38961597285146e+00, 3.36557599194762e+00, 3.34423797469248e+00, 3.32546399643910e+00, 3.30910125700086e+00, 3.29498494662371e+00, 3.28294109334028e+00, 3.27278939849996e+00, 3.26434605971177e+00, 3.25742657002743e+00, 3.25184847005853e+00, 3.24743401709802e+00, 3.24401272349275e+00, 3.24142370674649e+00, 3.23951778725363e+00, 3.23815926707648e+00, 3.23722732536787e+00, 3.23661697309218e+00, 3.23623952137750e+00, 3.23602253348071e+00, 3.23590924895807e+00, 3.23585748894386e+00, 3.23583807207586e+00, 3.23583279022643e+00, 3.23583201063686e+00, 3.23583195186487e+00, 3.23583091410855e+00, 3.23582388316201e+00, 3.23579803673754e+00, 3.23572914026460e+00, 3.23557836508935e+00, 3.23528962435758e+00, 3.23478753862151e+00, 3.23397613343952e+00, 3.23273835960221e+00, 3.23093651213637e+00, 3.22841360563707e+00, 3.22499573940672e+00, 3.22049545520034e+00, 3.21471605248335e+00, 3.20745678125198e+00, 3.19851878203151e+00, 3.18771158932720e+00, 3.17485996252423e+00, 3.15981076204259e+00, 3.14243955408320e+00, 3.12265661010048e+00, 3.10041197185485e+00, 3.07569928242710e+00, 3.04855813833902e+00, 3.01907479543456e+00, 2.98738115601552e+00, 2.95365206898884e+00, 2.91810107895213e+00, 2.88097485428060e+00, 2.84254659935785e+00, 2.80310880521417e+00, 2.76296571216534e+00, 2.72242584726115e+00, 2.68179496157438e+00, 2.64136963355035e+00, 2.60143173260382e+00, 2.56224386034400e+00, 2.52404581313519e+00, 2.48705204552716e+00, 2.45145006365136e+00, 2.41739964287153e+00, 2.38503274457653e+00, 2.35445400114775e+00, 2.32574164296158e+00, 2.29894875357304e+00, 2.27410475588974e+00, 2.25121705060194e+00, 2.23027274644141e+00, 2.21124043870386e+00, 2.19407200712873e+00, 2.17870441631707e+00, 2.16506151125617e+00, 2.15305580720798e+00, 2.14259027726634e+00, 2.13356014238858e+00, 2.12585466780236e+00, 2.11935896658719e+00, 2.11395580624996e+00, 2.10952740769028e+00, 2.10595721864764e+00, 2.10313163621321e+00, 2.10094164600921e+00, 2.09928433994152e+00, 2.09806427070751e+00, 2.09719460005268e+00, 2.09659799948203e+00, 2.09620726686832e+00, 2.09596563001196e+00, 2.09582671828871e+00, 2.09575419543835e+00, 2.09572105951160e+00, 2.09570862914016e+00, 2.09570524778328e+00, 2.09570474870227e+00, 2.09570471087264e+00, 2.09570404018605e+00, 2.09569949618894e+00, 2.09568279202374e+00, 2.09563826522707e+00, 2.09554082132381e+00, 2.09535421188028e+00, 2.09502971929877e+00, 2.09450531424442e+00, 2.09370534399239e+00, 2.09254080054288e+00, 2.09091020526508e+00, 2.08870113123749e+00, 2.08579236463429e+00, 2.08205668208824e+00, 2.07736419213365e+00, 2.07158615652613e+00, 2.06459917323095e+00, 2.05628956979100e+00, 2.04655782692463e+00, 2.03532283120716e+00, 2.02252574606808e+00, 2.00813329488197e+00, 1.99214027015587e+00, 1.97457111848216e+00, 1.95548049984818e+00, 1.93495277804737e+00, 1.91310046094172e+00, 1.89006166921020e+00, 1.86599676434324e+00, 1.84108430659015e+00, 1.81551653876068e+00, 1.78949460176112e+00, 1.76322368394050e+00, 1.73690829150247e+00, 1.71074780472010e+00, 1.68493245757469e+00, 1.65963984907968e+00, 1.63503206436188e+00, 1.61125345321402e+00, 1.58842908364076e+00, 1.56666385837991e+00, 1.54604225447575e+00, 1.52662862124468e+00, 1.50846795224150e+00, 1.49158703381181e+00, 1.47599586758800e+00, 1.46168926699910e+00, 1.44864853764169e+00, 1.43684316646859e+00, 1.42623246298061e+00, 1.41676711465226e+00, 1.40839063666490e+00, 1.40104071114511e+00, 1.39465042258774e+00, 1.38914940361610e+00, 1.38446490878167e+00, 1.38052283414052e+00, 1.37724869747440e+00, 1.37456858896772e+00, 1.37241009567637e+00, 1.37070319599889e+00, 1.36938111333103e+00, 1.36838111184777e+00, 1.36764521251423e+00, 1.36712080446682e+00, 1.36676112614293e+00, 1.36652559210489e+00, 1.36637994531398e+00, 1.36629622038210e+00, 1.36625251060104e+00, 1.36623253973856e+00, 1.36622504805906e+00, 1.36622301014819e+00, 1.36622270935707e+00, 1.36622268641287e+00, 1.36622227772334e+00, 1.36621950879565e+00, 1.36620932998423e+00, 1.36618219746791e+00, 1.36612282094264e+00, 1.36600911669552e+00, 1.36581141191400e+00, 1.36549194322691e+00, 1.36500468895242e+00, 1.36429556975973e+00, 1.36330304569486e+00, 1.36195912798851e+00, 1.36019081107200e+00, 1.35792191336121e+00, 1.35507529464370e+00, 1.35157539390926e+00, 1.34735100544455e+00, 1.34233818488063e+00, 1.33648315313128e+00, 1.32974504767203e+00, 1.32209836036702e+00, 1.31353490177041e+00, 1.30406514555955e+00, 1.29371883447551e+00, 1.28254477040457e+00, 1.27060976391037e+00, 1.25799677871903e+00, 1.24480236881810e+00, 1.23113356317046e+00, 1.21710439832493e+00, 1.20283232584715e+00, 1.18843472490562e+00, 1.17402572914492e+00, 1.15971353372908e+00, 1.14559828952842e+00, 1.13177062593454e+00, 1.11831078160021e+00, 1.10528827205330e+00, 1.09276199010380e+00, 1.08078062088843e+00, 1.06938325633435e+00, 1.05860010936595e+00, 1.04845325089794e+00, 1.03895731744245e+00, 1.03012016008358e+00, 1.02194342426006e+00, 1.01442306330827e+00, 1.00754979720065e+00, 1.00130953216825e+00, 9.95683757976153e-01, 9.90649938566548e-01, 9.86181909457500e-01, 9.82250292317062e-01, 9.78822933927844e-01, 9.75865373544195e-01, 9.73341339529611e-01, 9.71213273183508e-01, 9.69442874839984e-01, 9.67991664667238e-01, 9.66821548162486e-01, 9.65895374201097e-01, 9.65177471768939e-01, 9.64634150309450e-01, 9.64234148081658e-01, 9.63949013162207e-01, 9.63753402810923e-01, 9.63625288870787e-01, 9.63546059643448e-01, 9.63500512150532e-01, 9.63476732673428e-01, 9.63465867729596e-01, 9.63461791929532e-01, 9.63460683214549e-01, 9.63460519570589e-01, 9.63460506999795e-01, 9.63460281930367e-01, 9.63458757052834e-01, 9.63453151456761e-01, 9.63438209110947e-01, 9.63405508737551e-01, 9.63342885851063e-01, 9.63233991133151e-01, 9.63058007112309e-01, 9.62789542490682e-01, 9.62398720188588e-01, 9.61851470974679e-01, 9.61110039105094e-01, 9.60133699446970e-01, 9.58879676975272e-01, 9.57304249398220e-01, 9.55364002374282e-01, 9.53017195077990e-01, 9.50225182840043e-01, 9.46953834578679e-01, 9.43174877192689e-01, 9.38867098281946e-01, 9.34017343336691e-01, 9.28621254040365e-01, 9.22683709856045e-01, 9.16218954077436e-01, 9.09250405862082e-01, 9.01810179021254e-01, 8.93938344329015e-01, 8.85681983288706e-01, 8.77094087021675e-01, 8.68232354527564e-01, 8.59157941060104e-01, 8.49934201208171e-01, 8.40625463907600e-01, 8.31295869173161e-01, 8.22008289513718e-01, 8.12823353030382e-01, 8.03798580051463e-01, 7.94987640654625e-01, 7.86439736410330e-01, 7.78199106086049e-01, 7.70304651915879e-01, 7.62789680465943e-01, 7.55681750217889e-01, 7.49002616812064e-01, 7.42768266423980e-01, 7.36989027898616e-01, 7.31669754888273e-01, 7.26810070146186e-01, 7.22404665140076e-01, 7.18443649098617e-01, 7.14912942362332e-01, 7.11794709379532e-01, 7.09067826815982e-01, 7.06708382014838e-01, 7.04690196469414e-01, 7.02985368103640e-01, 7.01564825067692e-01, 7.00398882549678e-01, 6.99457792890541e-01, 6.98712278199352e-01, 6.98134033826918e-01, 6.97696190595451e-01, 6.97373723703038e-01, 6.97143796808419e-01, 6.96986030987205e-01, 6.96882690037086e-01, 6.96818775938567e-01, 6.96782031053871e-01, 6.96762846731275e-01, 6.96754081214436e-01, 6.96750792959762e-01, 6.96749898474110e-01, 6.96749766449884e-01, 6.96749756230659e-01, 6.96749572256394e-01, 6.96748325803597e-01, 6.96743743711181e-01, 6.96731529533864e-01, 6.96704799099223e-01, 6.96653607094859e-01, 6.96564583693188e-01, 6.96420698000054e-01, 6.96201162527558e-01, 6.95881489865528e-01, 6.95433709155338e-01, 6.94826745740500e-01, 6.94026962394545e-01, 6.92998854784193e-01, 6.91705887383193e-01, 6.90111449110319e-01, 6.88179900863361e-01, 6.85877680346209e-01, 6.83174423730591e-01, 6.80044059415764e-01, 6.76465827074054e-01, 6.72425175813061e-01, 6.67914498940135e-01, 6.62933669490769e-01, 6.57490350056065e-01, 6.51600061867214e-01, 6.45286010662744e-01, 6.38578679511170e-01, 6.31515210402355e-01, 6.24138606087399e-01, 6.16496790610697e-01, 6.08641570838259e-01, 6.00627542005441e-01, 5.92510978190333e-01, 5.84348744250349e-01, 5.76197259882107e-01, 5.68111539870273e-01, 5.60144327984204e-01, 5.52345335909109e-01, 5.44760593402685e-01, 5.37431911688742e-01, 5.30396458920038e-01, 5.23686444234215e-01, 5.17328905317058e-01, 5.11345593303162e-01, 5.05752948144291e-01, 5.00562157167565e-01, 4.95779289368969e-01, 4.91405498007438e-01, 4.87437284241793e-01, 4.83866814842884e-01, 4.80682287350746e-01, 4.77868336361325e-01, 4.75406474839573e-01, 4.73275564400083e-01, 4.71452308321787e-01, 4.69911760653474e-01, 4.68627844137714e-01, 4.67573868890156e-01, 4.66723042906423e-01, 4.66048964646674e-01, 4.65526087294265e-01, 4.65130143927183e-01, 4.64838522889991e-01, 4.64630583192971e-01, 4.64487900836730e-01, 4.64394438567396e-01, 4.64336633660338e-01, 4.64303400822146e-01, 4.64286050066093e-01, 4.64278122312251e-01, 4.64275148329975e-01, 4.64274339333926e-01, 4.64274219927788e-01, 4.64274210610327e-01, 4.64274041902215e-01, 4.64272898879841e-01, 4.64268697007398e-01, 4.64257496348623e-01, 4.64232983932163e-01, 4.64186039535605e-01, 4.64104402252474e-01, 4.63972453133138e-01, 4.63771126784550e-01, 4.63477962017024e-01, 4.63067298357359e-01, 4.62510621423064e-01, 4.61777055709803e-01, 4.60833998274808e-01, 4.59647881170582e-01, 4.58185044465310e-01, 4.56412695560203e-01, 4.54299924663999e-01, 4.51818741185775e-01, 4.48945091962457e-01, 4.45659820139696e-01, 4.41949523563210e-01, 4.37807273948309e-01, 4.33233162891952e-01, 4.28234647758671e-01, 4.22826679161901e-01, 4.17031601566308e-01, 4.10878828746121e-01, 4.04404305748059e-01, 3.97649778013914e-01, 3.90661895946181e-01, 3.83491189160878e-01, 3.76190948821517e-01, 3.68816058756039e-01, 3.61421816561815e-01, 3.54062784664052e-01, 3.46791708407281e-01, 3.39658533861595e-01, 3.32709552315293e-01, 3.25986691674615e-01, 3.19526967552605e-01, 3.13362099122017e-01, 3.07518287303310e-01, 3.02016146037559e-01, 2.96870771688646e-01, 2.92091931366897e-01, 2.87684348357629e-01, 2.83648061918470e-01, 2.79978839371609e-01, 2.76668620429760e-01, 2.73705976723861e-01, 2.71076573142409e-01, 2.68763621422536e-01, 2.66748320045949e-01, 2.65010277556343e-01, 2.63527918689689e-01, 2.62278874069253e-01, 2.61240354653026e-01, 2.60389511714479e-01, 2.59703782063056e-01, 2.59161216694451e-01, 2.58740789361923e-01, 2.58422679941205e-01, 2.58188526162186e-01, 2.58021636489460e-01, 2.57907156779154e-01, 2.57832183876218e-01, 2.57785820521633e-01, 2.57759167728096e-01, 2.57745253012359e-01, 2.57738895365307e-01, 2.57736510408833e-01, 2.57735861644406e-01, 2.57735765888185e-01, 2.57735758352609e-01, 2.57735621093545e-01, 2.57734691144279e-01, 2.57731272568023e-01, 2.57722160018545e-01, 2.57702218034666e-01, 2.57664029079095e-01, 2.57597625456924e-01, 2.57490319141766e-01, 2.57326642210899e-01, 2.57088408891235e-01, 2.56754907956065e-01, 2.56303231193684e-01, 2.55708739729684e-01, 2.54945665037724e-01, 2.53987835539220e-01, 2.52809512923889e-01, 2.51386315023879e-01, 2.49696194679260e-01, 2.47720437095523e-01, 2.45444632335705e-01, 2.42859575441940e-01, 2.39962044827124e-01, 2.36755410484577e-01, 2.33250027533671e-01, 2.29463377757497e-01, 2.25419931990785e-01, 2.21150719181040e-01, 2.16692603200030e-01, 2.12087285394560e-01, 2.07380068643843e-01, 2.02618436340074e-01, 1.97850515950947e-01, 1.93123510020968e-01, 1.88482185645839e-01, 1.83967514479761e-01, 1.79615547417910e-01, 1.75456590464904e-01, 1.71514721938863e-01, 1.67807659136813e-01, 1.64346949716730e-01, 1.61138434770407e-01, 1.58182911421846e-01, 1.55476915151880e-01, 1.53013545592600e-01, 1.50783271658000e-01, 1.48774668802176e-01, 1.46975059159189e-01, 1.45371041443880e-01, 1.43948910148637e-01, 1.42694972274327e-01, 1.41595774922152e-01, 1.40638259302932e-01, 1.39809856952979e-01, 1.39098542939820e-01, 1.38492859166561e-01, 1.37981918913027e-01, 1.37555401702500e-01, 1.37203545561448e-01, 1.36917141797369e-01, 1.36687535577735e-01, 1.36506633872514e-01, 1.36366920749722e-01, 1.36261478632022e-01, 1.36184012982866e-01, 1.36128877050538e-01, 1.36091092804620e-01, 1.36066364083007e-01, 1.36051078229428e-01, 1.36042293108389e-01, 1.36037707274397e-01, 1.36035612149556e-01, 1.36034826225114e-01, 1.36034612437356e-01, 1.36034580882850e-01, 1.36034578377661e-01, 1.36034532466994e-01, 1.36034221416935e-01, 1.36033077986909e-01, 1.36030030197901e-01, 1.36023361077548e-01, 1.36010592329157e-01, 1.35988398051722e-01, 1.35952554867828e-01, 1.35897934993778e-01, 1.35818548716543e-01, 1.35707642174765e-01, 1.35557855010035e-01, 1.35361440116228e-01, 1.35110544212438e-01, 1.34797543293317e-01, 1.34415421369852e-01, 1.33958174762242e-01, 1.33421218252579e-01, 1.32801764564292e-01, 1.32099145922155e-01, 1.31315046787364e-01, 1.30453620894912e-01, 1.29521473569355e-01, 1.28527501412307e-01, 1.27482594597510e-01, 1.26399220371226e-01, 1.25290917914981e-01, 1.24171742683046e-01, 1.23055701549291e-01, 1.21956218384099e-01, 1.20885663750073e-01, 1.19854973615006e-01, 1.18873371932708e-01, 1.17948202107684e-01, 1.17084863834521e-01, 1.16286845218880e-01, 1.15555835601015e-01, 1.14891901907129e-01, 1.14293710254884e-01, 1.13758774599461e-01, 1.13283715258490e-01, 1.12864512148235e-01, 1.12496740374439e-01, 1.12175779116370e-01, 1.11896988080488e-01, 1.11655848812461e-01, 1.11448070708718e-01, 1.11269663674648e-01, 1.11116981052348e-01, 1.10986737606737e-01, 1.10876007910534e-01, 1.10782210407493e-01, 1.10703081909662e-01, 1.10636646526530e-01, 1.10581182230362e-01, 1.10535187527901e-01, 1.10497350052938e-01, 1.10466518306879e-01, 1.10441677256915e-01, 1.10421928066827e-01, 1.10406471897434e-01, 1.10394597469110e-01, 1.10385671913889e-01, 1.10379134338729e-01, 1.10374491458956e-01, 1.10371314635188e-01, 1.10369237654510e-01, 1.10367954641163e-01, 1.10367217560338e-01, 1.10366832889810e-01, 1.10366657165358e-01, 1.10366591250667e-01, 1.10366573320803e-01, 1.10366570674415e-01, 1.10366570462389e-01, 1.10366566552315e-01, 1.10366540061293e-01, 1.10366442681865e-01, 1.10366183137785e-01, 1.10365615302698e-01, 1.10364528484865e-01, 1.10362640545247e-01, 1.10359594617265e-01, 1.10354960290906e-01, 1.10348240171419e-01, 1.10338882671201e-01, 1.10326301681383e-01, 1.10309903344145e-01, 1.10289119472142e-01, 1.10263446238079e-01, 1.10232485642612e-01, 1.10195986074046e-01, 1.10153877180423e-01, 1.10106293504515e-01, 1.10053581124305e-01, 1.09996282121628e-01, 1.09935093223299e-01, 1.09870797463984e-01, 1.09804171067012e-01, 1.09735871552445e-01, 1.09666316761789e-01, 1.09595567308560e-01, 1.09523226298492e-01, 1.09448369710066e-01, 1.09369518701698e-01, 1.09284661685379e-01, 1.09191329613438e-01, 1.09086722740628e-01, 1.08967881394150e-01, 1.08831887735903e-01, 1.08676081386932e-01, 1.08498270263098e-01, 1.08296919223065e-01, 1.08071302110517e-01, 1.07821606080080e-01, 1.07548980210993e-01, 1.07255524025888e-01, 1.06944216463376e-01, 1.06618791701877e-01, 1.06283573236382e-01, 1.05943280098501e-01, 1.05602818939559e-01, 1.05267074306269e-01, 1.04940708340990e-01, 1.04627980600640e-01, 1.04332597804159e-01, 1.04057601117262e-01, 1.03805295036679e-01, 1.03577218041409e-01, 1.03374151906237e-01, 1.03196164496438e-01, 1.03042679775896e-01, 1.02912568298692e-01, 1.02804251306097e-01, 1.02715811667906e-01, 1.02645105363777e-01, 1.02589867988229e-01, 1.02547811846051e-01, 1.02516710406320e-01, 1.02494468093766e-01, 1.02479174451062e-01, 1.02469142591119e-01, 1.02462932515706e-01, 1.02459360342891e-01, 1.02457494767350e-01, 1.02456642238153e-01, 1.02456322403725e-01, 1.02456235398705e-01, 1.02456222557024e-01, 1.02456221518722e-01, 1.02456202246614e-01, 1.02456071674112e-01, 1.02455591660032e-01, 1.02454311994720e-01, 1.02451510850585e-01, 1.02446143918968e-01, 1.02436803248239e-01, 1.02421686006920e-01, 1.02398572512200e-01, 1.02364811983819e-01, 1.02317313748486e-01, 1.02252541058050e-01, 1.02166504487570e-01, 1.02054752025794e-01, 1.01912353507316e-01, 1.01733877960044e-01, 1.01513363679388e-01, 1.01244282425526e-01, 1.00919500999694e-01, 1.00531245675526e-01, 1.00071077233267e-01, 9.95298864039711e-02, 9.88979205909430e-02, 9.81648521165397e-02, 9.73198959084698e-02, 9.63519810873521e-02, 9.52499780183018e-02, 9.40029808479608e-02, 9.26006445811254e-02, 9.10335724335738e-02, 8.92937406893708e-02, 8.73749362710470e-02, 8.52731726851061e-02, 8.29870513566533e-02, 8.05180478188845e-02, 7.78707171906916e-02, 7.50528159706789e-02, 7.20753254146302e-02, 6.89523499375895e-02, 6.56790077311911e-02, 6.23218860059893e-02, 5.88768651024907e-02, 5.53673081884796e-02, 5.18177207679287e-02, 4.82532297174032e-02, 4.46990698996359e-02, 4.11800962644122e-02, 3.77203353816782e-02, 3.43425861250353e-02, 3.10680751982346e-02, 2.79161696964536e-02, 2.49041461120214e-02, 2.20470132063610e-02, 1.93573849549864e-02, 1.68453992351658e-02, 1.45186779120507e-02, 1.23823243090584e-02, 1.04389545478369e-02, 8.68875977029404e-03, 7.12959671099611e-03, 5.75710442861066e-03, 4.56484522196376e-03, 3.54446788168161e-03, 2.68589150556409e-03, 1.97750819687315e-03, 1.40640312483242e-03, 9.58590723342923e-04, 6.19266308291344e-04, 3.73073230576892e-04, 2.04387078611035e-04, 9.76209232135261e-05, 3.75604328581858e-05, 9.74850374291236e-06, 9.68749404250957e-07],
[ 6.70000000000000e+01, 6.69999999294297e+01, 6.69999977472564e+01, 6.69999829626861e+01, 6.69999286126718e+01, 6.69997837326438e+01, 6.69994666518135e+01, 6.69988593487866e+01, 6.69978030706721e+01, 6.69960953798255e+01, 6.69934887482550e+01, 6.69896907723242e+01, 6.69843660317937e+01, 6.69771395696936e+01, 6.69676019252463e+01, 6.69553156129976e+01, 6.69398229088609e+01, 6.69206547784212e+01, 6.68973407640264e+01, 6.68694196331779e+01, 6.68364505788257e+01, 6.67980247491837e+01, 6.67537768676322e+01, 6.67033966803220e+01, 6.66466399403858e+01, 6.65833386059349e+01, 6.65134098996884e+01, 6.64368638586915e+01, 6.63538090015331e+01, 6.62644557653740e+01, 6.61691174208719e+01, 6.60682082605276e+01, 6.59622389709084e+01, 6.58518092327654e+01, 6.57375977328581e+01, 6.56203499034361e+01, 6.55008638166610e+01, 6.53799747415323e+01, 6.52585389142058e+01, 6.51374170779916e+01, 6.50174583203924e+01, 6.48994846783441e+01, 6.47842769083508e+01, 6.46725617347793e+01, 6.45650008055044e+01, 6.44621815057785e+01, 6.43646097127048e+01, 6.42727045158684e+01, 6.41867948844712e+01, 6.41071182263656e+01, 6.40338207575902e+01, 6.39669595800606e+01, 6.39065063477853e+01, 6.38523523866216e+01, 6.38043151179139e+01, 6.37621456216933e+01, 6.37255371602882e+01, 6.36941344684658e+01, 6.36675436022055e+01, 6.36453421257609e+01, 6.36270894068251e+01, 6.36123367834572e+01, 6.36006373650240e+01, 6.35915552336917e+01, 6.35846738237038e+01, 6.35796032732311e+01, 6.35759865680791e+01, 6.35735043276616e+01, 6.35718781207141e+01, 6.35708722401306e+01, 6.35702939116696e+01, 6.35699919584555e+01, 6.35698539904238e+01, 6.35698022333481e+01, 6.35697881541210e+01, 6.35697860760549e+01, 6.35697858034564e+01, 6.35697794522395e+01, 6.35697364217926e+01, 6.35695782370487e+01, 6.35691565725324e+01, 6.35682337595655e+01, 6.35664664221612e+01, 6.35633928819127e+01, 6.35584248865486e+01, 6.35508441183413e+01, 6.35398038267777e+01, 6.35243358039569e+01, 6.35033627801412e+01, 6.34757161603345e+01, 6.34401588510686e+01, 6.33954127412975e+01, 6.33401902053368e+01, 6.32732287937967e+01, 6.31933280774657e+01, 6.30993874195656e+01, 6.29904432887785e+01, 6.28657046093926e+01, 6.27245846008421e+01, 6.25667276131319e+01, 6.23920296383275e+01, 6.22006514791131e+01, 6.19930239692834e+01, 6.17698451279100e+01, 6.15320696258878e+01, 6.12808913765366e+01, 6.10177203641232e+01, 6.07441549558929e+01, 6.04619509059186e+01, 6.01729880975150e+01, 5.98792358607173e+01, 5.95827175248666e+01, 5.92854747847222e+01, 5.89895324892045e+01, 5.86968645717989e+01, 5.84093619601229e+01, 5.81288033500407e+01, 5.78568296533502e+01, 5.75949227223627e+01, 5.73443886639918e+01, 5.71063457527784e+01, 5.68817167057545e+01, 5.66712249327524e+01, 5.64753943272927e+01, 5.62945521905755e+01, 5.61288349475130e+01, 5.59781963855930e+01, 5.58424182026811e+01, 5.57211226793291e+01, 5.56137872948268e+01, 5.55197610892077e+01, 5.54382825421906e+01, 5.53684987006300e+01, 5.53094852432530e+01, 5.52602671290430e+01, 5.52198394366974e+01, 5.51871879699851e+01, 5.51613091803445e+01, 5.51412289464133e+01, 5.51260197528840e+01, 5.51148158301640e+01, 5.51068258530760e+01, 5.51013428515544e+01, 5.50977510580838e+01, 5.50955295033927e+01, 5.50942522703921e+01, 5.50935854223527e+01, 5.50932807300235e+01, 5.50931664287722e+01, 5.50931353360001e+01, 5.50931307467699e+01, 5.50931303416282e+01, 5.50931224030765e+01, 5.50930686182091e+01, 5.50928708991780e+01, 5.50923438515331e+01, 5.50911904140540e+01, 5.50889814179985e+01, 5.50851398761222e+01, 5.50789307103134e+01, 5.50694565090496e+01, 5.50556597699224e+01, 5.50363319249437e+01, 5.50101292637760e+01, 5.49755956599431e+01, 5.49311917672545e+01, 5.48753300915802e+01, 5.48064150650664e+01, 5.47228869698024e+01, 5.46232682953956e+01, 5.45062108939828e+01, 5.43705421431971e+01, 5.42153082672371e+01, 5.40398130165984e+01, 5.38436500745383e+01, 5.36267278332067e+01, 5.33892855378426e+01, 5.31319001939240e+01, 5.28554840267796e+01, 5.25612726422785e+01, 5.22508043465833e+01, 5.19258913499027e+01, 5.15885838240138e+01, 5.12411280233706e+01, 5.08859199140356e+01, 5.05254559586496e+01, 5.01622828375498e+01, 4.97989479042095e+01, 4.94379520536136e+01, 4.90817064301973e+01, 4.87324940516962e+01, 4.83924370296608e+01, 4.80634696833209e+01, 4.77473175175381e+01, 4.74454817954132e+01, 4.71592292884539e+01, 4.68895867217248e+01, 4.66373394275649e+01, 4.64030337558360e+01, 4.61869828403526e+01, 4.59892753747438e+01, 4.58097870970755e+01, 4.56481947166200e+01, 4.55039920368950e+01, 4.53765080368526e+01, 4.52649266678073e+01, 4.51683081084017e+01, 4.50856111948422e+01, 4.50157167103677e+01, 4.49574511787225e+01, 4.49096107643708e+01, 4.48709848412390e+01, 4.48403787564522e+01, 4.48166352906652e+01, 4.47986543068275e+01, 4.47854100885556e+01, 4.47759659005444e+01, 4.47694853579768e+01, 4.47652402692215e+01, 4.47626147139861e+01, 4.47611052335195e+01, 4.47603171350661e+01, 4.47599570432649e+01, 4.47598219598956e+01, 4.47597852138964e+01, 4.47597797902622e+01, 4.47597793552196e+01, 4.47597713264777e+01, 4.47597169305626e+01, 4.47595169653124e+01, 4.47589839305093e+01, 4.47578173925002e+01, 4.47555833158714e+01, 4.47516981853275e+01, 4.47454186403698e+01, 4.47358372287834e+01, 4.47218847498410e+01, 4.47023395009462e+01, 4.46758435584629e+01, 4.46409260119931e+01, 4.45960328310349e+01, 4.45395627770428e+01, 4.44699084904278e+01, 4.43855015942269e+01, 4.42848603824573e+01, 4.41666384240450e+01, 4.40296722371884e+01, 4.38730260976268e+01, 4.36960320564815e+01, 4.34983233698766e+01, 4.32798597830125e+01, 4.30409434530181e+01, 4.27822247136788e+01, 4.25046973494016e+01, 4.22096835219579e+01, 4.18988089520562e+01, 4.15739693769880e+01, 4.12372896715320e+01, 4.08910773211136e+01, 4.05377721592494e+01, 4.01798944029463e+01, 3.98199930128039e+01, 3.94605962492782e+01, 3.91041659960835e+01, 3.87530570118400e+01, 3.84094818158502e+01, 3.80754814870161e+01, 3.77529023156825e+01, 3.74433780243876e+01, 3.71483171588853e+01, 3.68688952160461e+01, 3.66060510844086e+01, 3.63604873972000e+01, 3.61326744204064e+01, 3.59228571155969e+01, 3.57310650312400e+01, 3.55571246914634e+01, 3.54006741700246e+01, 3.52611795591292e+01, 3.51379530644641e+01, 3.50301724750120e+01, 3.49369017645960e+01, 3.48571125786911e+01, 3.47897063437252e+01, 3.47335367079047e+01, 3.46874319854787e+01, 3.46502172348640e+01, 3.46207355607574e+01, 3.45978681973280e+01, 3.45805529096743e+01, 3.45678002490559e+01, 3.45587072177391e+01, 3.45524679436842e+01, 3.45483810337563e+01, 3.45458533646535e+01, 3.45444001793377e+01, 3.45436414779297e+01, 3.45432948192293e+01, 3.45431647752980e+01, 3.45431294001687e+01, 3.45431241788727e+01, 3.45431237761561e+01, 3.45431165446588e+01, 3.45430675501990e+01, 3.45428874413563e+01, 3.45424073371124e+01, 3.45413566397215e+01, 3.45393444247063e+01, 3.45358451539802e+01, 3.45301893701894e+01, 3.45215599282580e+01, 3.45089942011275e+01, 3.44913925570628e+01, 3.44675332417714e+01, 3.44360936070321e+01, 3.43956774079070e+01, 3.43448476456656e+01, 3.42821641704296e+01, 3.42062249882580e+01, 3.41157099586219e+01, 3.40094253403557e+01, 3.38863474694857e+01, 3.37456637524879e+01, 3.35868091516636e+01, 3.34094964374025e+01, 3.32137386887357e+01, 3.29998628331345e+01, 3.27685134143111e+01, 3.25206462404793e+01, 3.22575120669703e+01, 3.19806309740985e+01, 3.16917585791244e+01, 3.13928456344792e+01, 3.10859928787056e+01, 3.07734031920462e+01, 3.04573331449312e+01, 3.01400459090844e+01, 2.98237672403453e+01, 2.95106458713321e+01, 2.92027192175375e+01, 2.89018848563788e+01, 2.86098778364186e+01, 2.83282535528953e+01, 2.80583757074074e+01, 2.78014087563590e+01, 2.75583142305503e+01, 2.73298503526997e+01, 2.71165744629242e+01, 2.69188478590453e+01, 2.67368427501099e+01, 2.65705510964256e+01, 2.64197951632130e+01, 2.62842396480153e+01, 2.61634052571574e+01, 2.60566836073517e+01, 2.59633533180780e+01, 2.58825971408069e+01, 2.58135199440047e+01, 2.57551673394243e+01, 2.57065446970136e+01, 2.56666362550542e+01, 2.56344239918841e+01, 2.56089058894521e+01, 2.55891131911565e+01, 2.55741262410990e+01, 2.55630884927628e+01, 2.55552182950109e+01, 2.55498181036997e+01, 2.55462808280357e+01, 2.55440931003096e+01, 2.55428353524701e+01, 2.55421786884696e+01, 2.55418786518303e+01, 2.55417660974964e+01, 2.55417354799667e+01, 2.55417309608827e+01, 2.55417306187113e+01, 2.55417245571788e+01, 2.55416834894020e+01, 2.55415325198726e+01, 2.55411300902456e+01, 2.55402493813773e+01, 2.55385627128883e+01, 2.55356295645521e+01, 2.55308887695062e+01, 2.55236553429493e+01, 2.55131223093825e+01, 2.54983677734359e+01, 2.54783673405835e+01, 2.54520118321094e+01, 2.54181300530060e+01, 2.53755161645365e+01, 2.53229609914399e+01, 2.52592863680938e+01, 2.51833814136440e+01, 2.50942394418206e+01, 2.49909940771272e+01, 2.48729530844422e+01, 2.47396284387006e+01, 2.45907612726623e+01, 2.44263405414401e+01, 2.42466145193111e+01, 2.40520945747299e+01, 2.38435510248318e+01, 2.36220012219487e+01, 2.33886903481437e+01, 2.31450656758636e+01, 2.28927452909985e+01, 2.26334824738230e+01, 2.23691270979561e+01, 2.21015855327657e+01, 2.18327806012885e+01, 2.15646131233714e+01, 2.12989264337250e+01, 2.10374749976270e+01, 2.07818978763286e+01, 2.05336973754190e+01, 2.02942228148533e+01, 2.00646590537959e+01, 1.98460192230096e+01, 1.96391410619531e+01, 1.94446862974538e+01, 1.92631425930643e+01, 1.90948277032345e+01, 1.89398955574824e+01, 1.87983440651123e+01, 1.86700244706639e+01, 1.85546521101823e+01, 1.84518184257761e+01, 1.83610040962152e+01, 1.82815931372554e+01, 1.82128878174083e+01, 1.81541242222310e+01, 1.81044882819861e+01, 1.80631320535702e+01, 1.80291900189592e+01, 1.80017951314015e+01, 1.79800943106153e+01, 1.79632630633839e+01, 1.79505188904090e+01, 1.79411331379240e+01, 1.79344409663237e+01, 1.79298491396250e+01, 1.79268413891076e+01, 1.79249811705851e+01, 1.79239117145382e+01, 1.79233533577159e+01, 1.79230982388708e+01, 1.79230025348223e+01, 1.79229765009853e+01, 1.79229726584449e+01, 1.79229723699842e+01, 1.79229672927485e+01, 1.79229328937277e+01, 1.79228064392682e+01, 1.79224693579526e+01, 1.79217316629536e+01, 1.79203188857735e+01, 1.79178620450023e+01, 1.79138911119305e+01, 1.79078323630711e+01, 1.78990099260131e+01, 1.78866517269292e+01, 1.78698999327906e+01, 1.78478258468969e+01, 1.78194490623322e+01, 1.77837605064720e+01, 1.77397488259507e+01, 1.76864293743026e+01, 1.76228748859393e+01, 1.75482467649233e+01, 1.74618258009752e+01, 1.73630410632279e+01, 1.72514957260619e+01, 1.71269886570896e+01, 1.69895307440940e+01, 1.68393551471778e+01, 1.66769209199619e+01, 1.65029097310620e+01, 1.63182157154255e+01, 1.61239287779589e+01, 1.59213119467132e+01, 1.57117736207469e+01, 1.54968357712092e+01, 1.52780993247543e+01, 1.50572080752528e+01, 1.48358125203210e+01, 1.46155349920829e+01, 1.43979373409090e+01, 1.41844922399150e+01, 1.39765589209137e+01, 1.37753638531592e+01, 1.35819865644671e+01, 1.33973505107581e+01, 1.32222186510729e+01, 1.30571931987902e+01, 1.29027189044024e+01, 1.27590891794548e+01, 1.26264543861374e+01, 1.25048316786262e+01, 1.23941158746511e+01, 1.22940909433007e+01, 1.22044418042001e+01, 1.21247662332932e+01, 1.20545867541272e+01, 1.19933624563869e+01, 1.19405007236962e+01, 1.18953688707187e+01, 1.18573056873156e+01, 1.18256328681745e+01, 1.17996662740389e+01, 1.17787269302255e+01, 1.17621516246847e+01, 1.17493029266597e+01, 1.17395784130404e+01, 1.17324188671482e+01, 1.17273152073328e+01, 1.17238139125283e+01, 1.17215207393958e+01, 1.17201025698982e+01, 1.17192872866104e+01, 1.17188616420345e+01, 1.17186671630592e+01, 1.17185942075317e+01, 1.17185743618829e+01, 1.17185714327076e+01, 1.17185712134822e+01, 1.17185673637738e+01, 1.17185412814470e+01, 1.17184454002606e+01, 1.17181898178033e+01, 1.17176304898790e+01, 1.17165593407723e+01, 1.17146966998291e+01, 1.17116864356114e+01, 1.17070941126883e+01, 1.17004084399858e+01, 1.16910462097056e+01, 1.16783608381254e+01, 1.16616545107929e+01, 1.16401938030688e+01, 1.16132284936001e+01, 1.15800131174090e+01, 1.15398306249515e+01, 1.14920173356041e+01, 1.14359882136758e+01, 1.13712613693749e+01, 1.12974806137498e+01, 1.12144348912483e+01, 1.11220734879005e+01, 1.10205160725045e+01, 1.09100568696249e+01, 1.07911625743801e+01, 1.06644639783193e+01, 1.05307416537530e+01, 1.03909064064021e+01, 1.02459755180393e+01, 1.00970460308885e+01, 9.94526645158091e+00, 9.79180826448086e+00, 9.63783854621438e+00, 9.48449478253800e+00, 9.33286273269813e+00, 9.18395789793267e+00, 9.03871086250024e+00, 8.89795651545946e+00, 8.76242694906090e+00, 8.63274767525638e+00, 8.50943670676987e+00, 8.39290600777204e+00, 8.28346482153407e+00, 8.18132441685665e+00, 8.08660385026488e+00, 7.99933640718370e+00, 7.91947645493495e+00, 7.84690650787445e+00, 7.78144436651183e+00, 7.72285024554446e+00, 7.67083384883696e+00, 7.62506138159753e+00, 7.58516251096357e+00, 7.55073729597160e+00, 7.52136310697876e+00, 7.49660154401363e+00, 7.47600534476749e+00, 7.45912524798167e+00, 7.44551674928465e+00, 7.43474665687308e+00, 7.42639932678925e+00, 7.42008243488061e+00, 7.41543212752211e+00, 7.41211738799822e+00, 7.40984346147447e+00, 7.40835419919229e+00, 7.40743321131826e+00, 7.40690375619351e+00, 7.40662733912514e+00, 7.40650104330417e+00, 7.40645366561134e+00, 7.40644077775185e+00, 7.40643887553157e+00, 7.40643873311226e+00, 7.40643623144587e+00, 7.40641928229968e+00, 7.40635697561995e+00, 7.40619089022886e+00, 7.40582742321808e+00, 7.40513136648727e+00, 7.40392099959104e+00, 7.40196494397135e+00, 7.39898099503306e+00, 7.39463711311107e+00, 7.38855470923123e+00, 7.38031430357205e+00, 7.36946356107410e+00, 7.35552761835980e+00, 7.33802150968184e+00, 7.31646438033696e+00, 7.29039505023623e+00, 7.25938836750188e+00, 7.22307168392255e+00, 7.18114070403842e+00, 7.13337392043186e+00, 7.07964486005355e+00, 7.01993143640374e+00, 6.95432183044315e+00, 6.88301650273365e+00, 6.80632615736219e+00, 6.72466571621752e+00, 6.63854459870581e+00, 6.54855381541901e+00, 6.45535055580797e+00, 6.35964106585703e+00, 6.26216266490373e+00, 6.16366574066355e+00, 6.06489649389952e+00, 5.96658108955081e+00, 5.86941172330421e+00, 5.77403494699325e+00, 5.68104242835393e+00, 5.59096416465692e+00, 5.50426403695854e+00, 5.42133748982503e+00, 5.34251105380936e+00, 5.26804339400967e+00, 5.19812756365896e+00, 5.13289416063125e+00, 5.07241511984009e+00, 5.01670791882968e+00, 4.96574002149872e+00, 4.91943343132304e+00, 4.87766926751409e+00, 4.84029231332921e+00, 4.80711551419814e+00, 4.77792442403864e+00, 4.75248161110827e+00, 4.73053104027682e+00, 4.71180244724118e+00, 4.69601571271526e+00, 4.68288523203633e+00, 4.67212425923035e+00, 4.66344918591497e+00, 4.65658369624406e+00, 4.65126272128177e+00, 4.64723610159728e+00, 4.64427185719148e+00, 4.64215896047367e+00, 4.64070951178538e+00, 4.63976022821329e+00, 4.63917317477603e+00, 4.63883569150734e+00, 4.63865949893516e+00, 4.63857899603175e+00, 4.63854879676517e+00, 4.63854058184675e+00, 4.63853936934254e+00, 4.63853927832147e+00, 4.63853767627450e+00, 4.63852682217836e+00, 4.63848692148932e+00, 4.63838056179710e+00, 4.63814780075460e+00, 4.63770205291861e+00, 4.63692694815684e+00, 4.63567431901331e+00, 4.63376345626367e+00, 4.63098174935123e+00, 4.62708679953813e+00, 4.62181005638639e+00, 4.61486198087798e+00, 4.60593868018824e+00, 4.59472989026558e+00, 4.58092810511380e+00, 4.56423857019937e+00, 4.54438977788090e+00, 4.52114403297769e+00, 4.49430760523103e+00, 4.46373996092110e+00, 4.42936157519161e+00, 4.39115987371789e+00, 4.34919293724650e+00, 4.30359072065889e+00, 4.25455368053297e+00, 4.20234885913731e+00, 4.14730362394241e+00, 4.08979739578869e+00, 4.03025180376315e+00, 3.96911977244282e+00, 3.90687407408387e+00, 3.84399586596943e+00, 3.78096368685576e+00, 3.71824331429337e+00, 3.65627879574895e+00, 3.59548486997919e+00, 3.53624089905088e+00, 3.47888634225401e+00, 3.42371772573048e+00, 3.37098699909620e+00, 3.32090112432477e+00, 3.27362271299372e+00, 3.22927151481456e+00, 3.18792656136431e+00, 3.14962878158835e+00, 3.11438392700034e+00, 3.08216567147300e+00, 3.05291878012469e+00, 3.02656227141845e+00, 3.00299252403434e+00, 2.98208630371040e+00, 2.96370370396617e+00, 2.94769100781721e+00, 2.93388348507820e+00, 2.92210814183451e+00, 2.91218643564072e+00, 2.90393696275342e+00, 2.89717811321890e+00, 2.89173067708933e+00, 2.88742037173725e+00, 2.88408024753922e+00, 2.88155291844348e+00, 2.87969255632524e+00, 2.87836658454427e+00, 2.87745700738344e+00, 2.87686131829678e+00, 2.87649294090825e+00, 2.87628117182864e+00, 2.87617061257355e+00, 2.87612009788781e+00, 2.87610114820674e+00, 2.87609599344549e+00, 2.87609523261395e+00, 2.87609517525611e+00, 2.87609416246970e+00, 2.87608730069881e+00, 2.87606207619464e+00, 2.87599483762999e+00, 2.87584769123929e+00, 2.87556590195275e+00, 2.87507591074949e+00, 2.87428406942310e+00, 2.87307618313797e+00, 2.87131793749994e+00, 2.86885626900620e+00, 2.86552171423342e+00, 2.86113174274314e+00, 2.85549504076008e+00, 2.84841666742862e+00, 2.83970395427005e+00, 2.82917296407902e+00, 2.81665527199125e+00, 2.80200478399836e+00, 2.78510427259685e+00, 2.76587129127543e+00, 2.74426313399429e+00, 2.72028053573068e+00, 2.69396986605315e+00, 2.66542364706767e+00, 2.63477932444786e+00, 2.60221632754933e+00, 2.56795156205395e+00, 2.53223357596420e+00, 2.49533571771039e+00, 2.45754865641676e+00, 2.41917265484203e+00, 2.38050997458707e+00, 2.34185775379244e+00, 2.30350163563844e+00, 2.26571034943179e+00, 2.22873136364275e+00, 2.19278765030462e+00, 2.15807552965867e+00, 2.12476350770989e+00, 2.09299197997689e+00, 2.06287365244965e+00, 2.03449452404837e+00, 2.00791528088197e+00, 1.98317296789521e+00, 1.96028282455191e+00, 1.93924019488028e+00, 1.92002244594849e+00, 1.90259085079535e+00, 1.88689241082525e+00, 1.87286160805489e+00, 1.86042208916494e+00, 1.84948829112798e+00, 1.83996702249414e+00, 1.83175901555262e+00, 1.82476046293677e+00, 1.81886454823140e+00, 1.81396297425080e+00, 1.80994748541355e+00, 1.80671137264656e+00, 1.80415094115759e+00, 1.80216691391938e+00, 1.80066573750049e+00, 1.79956075259736e+00, 1.79877318979089e+00, 1.79823295201963e+00, 1.79787914914999e+00, 1.79766035670633e+00, 1.79753457992088e+00, 1.79746891517580e+00, 1.79743891288454e+00, 1.79742765806581e+00, 1.79742459648946e+00, 1.79742414460748e+00, 1.79742411035544e+00, 1.79742350309692e+00, 1.79741938883512e+00, 1.79740426443236e+00, 1.79736394877969e+00, 1.79727572125425e+00, 1.79710676377955e+00, 1.79681297282056e+00, 1.79633820236289e+00, 1.79561399256708e+00, 1.79455983067999e+00, 1.79308398004615e+00, 1.79108489876181e+00, 1.78845325085149e+00, 1.78507448938420e+00, 1.78083196282081e+00, 1.77561046402059e+00, 1.76930010755153e+00, 1.76180038796950e+00, 1.75302424300897e+00, 1.74290192501120e+00, 1.73138447513199e+00, 1.71844660091115e+00, 1.70408878027912e+00, 1.68833845377888e+00, 1.67125021932104e+00, 1.65290500575722e+00, 1.63340826697350e+00, 1.61288730034427e+00, 1.59148784577384e+00, 1.56937015898470e+00, 1.54670477206157e+00, 1.52366815490085e+00, 1.50043847496795e+00, 1.47719162341242e+00, 1.45409763805924e+00, 1.43131761322370e+00, 1.40900114719798e+00, 1.38728434391838e+00, 1.36628835757747e+00, 1.34611844825963e+00, 1.32686350249617e+00, 1.30859596384065e+00, 1.29137211391220e+00, 1.27523264279183e+00, 1.26020344846525e+00, 1.24629660778194e+00, 1.23351146593646e+00, 1.22183579760888e+00, 1.21124700036466e+00, 1.20171328930296e+00, 1.19319487069481e+00, 1.18564508083176e+00, 1.17901148387010e+00, 1.17323692856353e+00, 1.16826056802837e+00, 1.16401884887941e+00, 1.16044647619773e+00, 1.15747735902093e+00, 1.15504553770813e+00, 1.15308609008510e+00, 1.15153600825074e+00, 1.15033503290891e+00, 1.14942642764639e+00, 1.14875767223420e+00, 1.14828105220121e+00, 1.14795412190067e+00, 1.14774002017464e+00, 1.14760762144917e+00, 1.14753151042683e+00, 1.14749177509282e+00, 1.14747362002819e+00, 1.14746680949132e+00, 1.14746495686531e+00, 1.14746468342177e+00, 1.14746466256362e+00, 1.14746429103164e+00, 1.14746177385059e+00, 1.14745252046912e+00, 1.14742785467648e+00, 1.14737387584161e+00, 1.14727050623435e+00, 1.14709076539540e+00, 1.14680030951452e+00, 1.14635727013885e+00, 1.14571242182494e+00, 1.14480970199544e+00, 1.14358709767107e+00, 1.14197790240603e+00, 1.13991233237347e+00, 1.13731947319374e+00, 1.13412950929902e+00, 1.13027616644347e+00, 1.12569927697065e+00, 1.12034735864229e+00, 1.11418008344528e+00, 1.10717050503366e+00, 1.09930691417523e+00, 1.09059420195710e+00, 1.08105463085306e+00, 1.07072794332805e+00, 1.05967077470727e+00, 1.04795537895039e+00, 1.03566771951844e+00, 1.02290501912643e+00, 1.00977289821492e+00, 9.96382259031776e-01, 9.82846087388367e-01, 9.69276345413562e-01, 9.55781115238665e-01, 9.42462126467930e-01, 9.29412762357686e-01, 9.16716595370147e-01, 9.04446457730430e-01, 8.92664012346671e-01, 8.81419758402515e-01, 8.70753386631506e-01, 8.60694392057482e-01, 8.51262855228512e-01, 8.42470313847500e-01, 8.34320661963545e-01, 8.26811030575286e-01, 8.19932619372551e-01, 8.13671463010223e-01, 8.08009126141570e-01, 8.02923329367786e-01, 7.98388513558493e-01, 7.94376353084388e-01, 7.90856229829672e-01, 7.87795679813570e-01, 7.85160823179581e-01, 7.82916786463304e-01, 7.81028123625816e-01, 7.79459239513784e-01, 7.78174816332466e-01, 7.77140240561565e-01, 7.76322024677114e-01, 7.75688215261482e-01, 7.75208776779635e-01, 7.74855938668026e-01, 7.74604492582313e-01, 7.74432026795594e-01, 7.74319085875263e-01, 7.74249245857263e-01, 7.74209098060425e-01, 7.74188138243161e-01, 7.74178561746638e-01, 7.74174969307282e-01, 7.74173992080009e-01, 7.74173847843414e-01, 7.74173836763467e-01, 7.74173638386188e-01, 7.74172294352880e-01, 7.74167353571499e-01, 7.74154183472376e-01, 7.74125362097806e-01, 7.74070169799131e-01, 7.73974202697985e-01, 7.73819128619436e-01, 7.73582604411786e-01, 7.73238371147002e-01, 7.72756540272724e-01, 7.72104078965683e-01, 7.71245496366618e-01, 7.70143723860912e-01, 7.68761172142561e-01, 7.67060935840149e-01, 7.65008103746723e-01, 7.62571120332453e-01, 7.59723133689064e-01, 7.56443257972748e-01, 7.52717676304655e-01, 7.48540514112563e-01, 7.43914423546016e-01, 7.38850836503289e-01, 7.33369865646225e-01, 7.27499857389475e-01, 7.21276625522052e-01, 7.14742416022148e-01, 7.07944670296797e-01, 7.00934663828932e-01, 6.93766099389634e-01, 6.86493728983478e-01, 6.79172067840867e-01, 6.71854248948733e-01, 6.64591049923828e-01, 6.57430107493070e-01, 6.50415320092179e-01, 6.43586427275892e-01, 6.36978746334350e-01, 6.30623041799012e-01, 6.24545502037265e-01, 6.18767798224994e-01, 6.13307203861737e-01, 6.08176756863600e-01, 6.03385450457878e-01, 5.98938443097850e-01, 5.94837281096133e-01, 5.91080130479824e-01, 5.87662016669967e-01, 5.84575072025459e-01, 5.81808792152316e-01, 5.79350302257026e-01, 5.77184634800446e-01, 5.75295019358478e-01, 5.73663184972389e-01, 5.72269674424163e-01, 5.71094168846656e-01, 5.70115819924339e-01, 5.69313585720025e-01, 5.68666564947634e-01, 5.68154323383382e-01, 5.67757205161115e-01, 5.67456621023339e-01, 5.67235305284015e-01, 5.67077533372029e-01, 5.66969292406995e-01, 5.66898398321513e-01, 5.66854554555064e-01, 5.66829349235206e-01, 5.66816189929398e-01, 5.66810177366449e-01, 5.66807921851790e-01, 5.66807308297612e-01, 5.66807217738315e-01, 5.66807210728641e-01, 5.66807084535327e-01, 5.66806229557315e-01, 5.66803086574296e-01, 5.66794708585704e-01, 5.66776373817277e-01, 5.66741261518075e-01, 5.66680203987204e-01, 5.66581526972996e-01, 5.66430989052634e-01, 5.66211828722231e-01, 5.65904925571971e-01, 5.65489078937546e-01, 5.64941403688962e-01, 5.64237838276753e-01, 5.63353754851747e-01, 5.62264655387665e-01, 5.60946931616101e-01, 5.59378660733174e-01, 5.57540403888780e-01, 5.55415971108033e-01, 5.52993115157662e-01, 5.50264118429780e-01, 5.47226241363376e-01, 5.43882008092723e-01, 5.40239314355048e-01, 5.36311353335464e-01, 5.32116365997368e-01, 5.27677232437270e-01, 5.23020928950473e-01, 5.18177881135415e-01, 5.13181246208575e-01, 5.08066157838044e-01, 5.02868964625691e-01, 4.97626489464324e-01, 4.92375332026043e-01, 4.87151231218903e-01, 4.81988499103153e-01, 4.76919532862325e-01, 4.71974407228927e-01, 4.67180546392598e-01, 4.62562471904619e-01, 4.58141621392969e-01, 4.53936231925624e-01, 4.49961281483411e-01, 4.46228482088344e-01, 4.42746318541921e-01, 4.39520127335044e-01, 4.36552210988680e-01, 4.33841983785754e-01, 4.31386145494074e-01, 4.29178880209264e-01, 4.27212077829481e-01, 4.25475575888157e-01, 4.23957419504376e-01, 4.22644137055926e-01, 4.21521028852297e-01, 4.20572465596834e-01, 4.19782192819838e-01, 4.19133636783002e-01, 4.18610206660873e-01, 4.18195587169208e-01, 4.17874015305415e-01, 4.17630534565731e-01, 4.17451219972762e-01, 4.17323367530145e-01, 4.17235642347492e-01, 4.17178180645095e-01, 4.17142642128383e-01, 4.17122210759402e-01, 4.17111543670700e-01, 4.17106669773148e-01, 4.17104841403216e-01, 4.17104344042002e-01, 4.17104270632521e-01, 4.17104264904268e-01, 4.17104161184525e-01, 4.17103458468091e-01, 4.17100875208931e-01, 4.17093989194839e-01, 4.17078919353372e-01, 4.17050058926762e-01, 4.16999870786100e-01, 4.16918754435628e-01, 4.16794992833061e-01, 4.16614785476042e-01, 4.16362372207999e-01, 4.16020249809847e-01, 4.15569480608994e-01, 4.14990089036039e-01, 4.14261538332354e-01, 4.13363275574142e-01, 4.12275329036888e-01, 4.10978937968665e-01, 4.09457191428771e-01, 4.07695650374222e-01, 4.05682926013384e-01, 4.03411187889418e-01, 4.00876577370639e-01, 3.98079506187401e-01, 3.95024825142755e-01, 3.91721854727621e-01, 3.88184276530637e-01, 3.84429891428880e-01, 3.80480256967957e-01, 3.76360221595746e-01, 3.72097377163243e-01, 3.67721453222813e-01, 3.63263677189479e-01, 3.58756123599846e-01, 3.54231073802168e-01, 3.49720404771787e-01, 3.45255022668397e-01, 3.40864353484278e-01, 3.36575899847617e-01, 3.32414869869709e-01, 3.28403880935696e-01, 3.24562738608435e-01, 3.20908288405249e-01, 3.17454336184269e-01, 3.14211631300765e-01, 3.11187905608222e-01, 3.08387960797688e-01, 3.05813796471796e-01, 3.03464771678065e-01, 3.01337793284787e-01, 2.99427525460629e-01, 2.97726615486809e-01, 2.96225932070428e-01, 2.94914813129921e-01, 2.93781320611657e-01, 2.92812500215632e-01, 2.91994643949067e-01, 2.91313553199743e-01, 2.90754799579474e-01, 2.90303980202339e-01, 2.89946963418998e-01, 2.89670120429506e-01, 2.89460537732339e-01, 2.89306205125825e-01, 2.89196174022507e-01, 2.89120681206981e-01, 2.89071233872818e-01, 2.89040652793020e-01, 2.89023071758265e-01, 2.89013892888036e-01, 2.89009698989716e-01, 2.89008125714190e-01, 2.89007697745149e-01, 2.89007634577814e-01, 2.89007629606829e-01, 2.89007539061373e-01, 2.89006925602777e-01, 2.89004670467023e-01, 2.88998659125639e-01, 2.88985503566494e-01, 2.88960309535724e-01, 2.88916498278368e-01, 2.88845691329158e-01, 2.88737665747817e-01, 2.88580385814342e-01, 2.88360115524270e-01, 2.88061614176769e-01, 2.87668414871945e-01, 2.87163182800299e-01, 2.86528146846919e-01, 2.85745594342927e-01, 2.84798414947304e-01, 2.83670675880643e-01, 2.82348207361778e-01, 2.80819174455758e-01, 2.79074609963018e-01, 2.77108882750113e-01, 2.74920077244317e-01, 2.72510262762736e-01, 2.69885635852631e-01, 2.67056524665477e-01, 2.64037251218141e-01, 2.60845854756681e-01, 2.57503686807382e-01, 2.54034895344148e-01, 2.50465821308751e-01, 2.46824335059117e-01, 2.43139142859874e-01, 2.39439094085199e-01, 2.35752518349656e-01, 2.32106618475655e-01, 2.28526940356318e-01, 2.25036934845747e-01, 2.21657620344615e-01, 2.18407348326863e-01, 2.15301668203480e-01, 2.12353283069604e-01, 2.09572084321366e-01, 2.06965250971152e-01, 2.04537398704858e-01, 2.02290764140741e-01, 2.00225411114559e-01, 1.98339447832026e-01, 1.96629246097004e-01, 1.95089656275627e-01, 1.93714213969329e-01, 1.92495336387239e-01, 1.91424508026415e-01, 1.90492456435294e-01, 1.89689319546844e-01, 1.89004806343448e-01, 1.88428352506520e-01, 1.87949272275233e-01, 1.87556907064277e-01, 1.87240770563264e-01, 1.86990689141979e-01, 1.86796935515844e-01, 1.86650352868156e-01, 1.86542466060374e-01, 1.86465576251763e-01, 1.86412835237263e-01, 1.86378296111883e-01, 1.86356937471501e-01, 1.86344659219785e-01, 1.86338249111188e-01, 1.86335320335644e-01, 1.86334221659877e-01, 1.86333922794290e-01, 1.86333878682376e-01, 1.86333875180253e-01, 1.86333810999017e-01, 1.86333376162324e-01, 1.86331777665245e-01, 1.86327516725257e-01, 1.86318192098611e-01, 1.86300335611961e-01, 1.86269287065258e-01, 1.86219115188720e-01, 1.86142590755907e-01, 1.86031217180442e-01, 1.85875322847338e-01, 1.85664217929144e-01, 1.85386416415312e-01, 1.85029921476726e-01, 1.84582569103370e-01, 1.84032421287966e-01, 1.83368196076671e-01, 1.82579717870766e-01, 1.81658367828291e-01, 1.80597511523501e-01, 1.79392879624217e-01, 1.78042877632024e-01, 1.76548802967460e-01, 1.74914951949219e-01, 1.73148605362542e-01, 1.71259888935438e-01, 1.69261513507536e-01, 1.67168408180189e-01, 1.64997267392934e-01, 1.62766038837304e-01, 1.60493382717135e-01, 1.58198133695890e-01, 1.55898794876975e-01, 1.53613088650857e-01, 1.51357582826563e-01, 1.49147402978386e-01, 1.46996034288122e-01, 1.44915209201145e-01, 1.42914871605072e-01, 1.41003204394877e-01, 1.39186705332449e-01, 1.37470295913616e-01, 1.35857449198511e-01, 1.34350324811166e-01, 1.32949902113438e-01, 1.31656105487195e-01, 1.30467918386756e-01, 1.29383485124592e-01, 1.28400201108150e-01, 1.27514793430009e-01, 1.26723394364689e-01, 1.26021610528724e-01, 1.25404590314389e-01, 1.24867091816270e-01, 1.24403552921621e-01, 1.24008164608952e-01, 1.23674947848556e-01, 1.23397833865610e-01, 1.23170746940545e-01, 1.22987688400116e-01, 1.22842820012551e-01, 1.22730544655951e-01, 1.22645581891845e-01, 1.22583035962307e-01, 1.22538453747339e-01, 1.22507870380304e-01, 1.22487840520392e-01, 1.22475453715197e-01, 1.22468332833449e-01, 1.22464615178771e-01, 1.22462916571578e-01, 1.22462279368605e-01, 1.22462106034234e-01, 1.22462080450475e-01, 1.22462078400716e-01, 1.22462040602358e-01, 1.22461784513234e-01, 1.22460843106197e-01, 1.22458333689138e-01, 1.22452842035472e-01, 1.22442325414170e-01, 1.22424038649086e-01, 1.22394487008066e-01, 1.22349409321401e-01, 1.22283794209935e-01, 1.22191931604718e-01, 1.22067500793463e-01, 1.21903694975771e-01, 1.21693380716124e-01, 1.21429288761988e-01, 1.21104230519521e-01, 1.20711332201111e-01, 1.20244276502433e-01, 1.19697539919727e-01, 1.19066612791769e-01, 1.18348189137606e-01, 1.17540314562689e-01, 1.16642482982911e-01, 1.15655676534146e-01, 1.14582347458750e-01, 1.13426345477066e-01, 1.12192798542367e-01, 1.10887958329045e-01, 1.09519023817057e-01, 1.08093956641824e-01, 1.06621300490417e-01, 1.05110014042637e-01, 1.03569323306402e-01, 1.02008595323142e-01, 1.00437231753119e-01, 9.88645782856281e-02, 9.72998444386377e-02, 9.57520281495903e-02, 9.42298404321349e-02, 9.27416269534689e-02, 9.12952852901299e-02, 8.98981784875019e-02, 8.85570471079689e-02, 8.72779230427483e-02, 8.60660489328495e-02, 8.49258071268224e-02, 8.38606617943043e-02, 8.28731172295818e-02, 8.19646946324468e-02, 8.11359288392310e-02, 8.03863856585281e-02, 7.97146996838194e-02, 7.91186317279653e-02, 7.85951443658490e-02, 7.81404934920715e-02, 7.77503333044933e-02, 7.74198317234465e-02, 7.71437929543310e-02, 7.69167837120079e-02, 7.67332595440138e-02, 7.65876877205613e-02, 7.64746632969207e-02, 7.63890151900739e-02, 7.63258994404055e-02, 7.62808772353765e-02, 7.62499757456254e-02, 7.62297303524377e-02, 7.62172074108065e-02, 7.62100072824126e-02, 7.62062479690542e-02, 7.62045302626149e-02, 7.62038858843681e-02, 7.62037105971700e-02, 7.62036847251011e-02, 7.62036826327233e-02, 7.62036438047744e-02, 7.62033807392299e-02, 7.62024136782458e-02, 7.61998358156291e-02, 7.61941940585018e-02, 7.61833887755330e-02, 7.61645962844426e-02, 7.61342172139621e-02, 7.60878531681612e-02, 7.60203133378923e-02, 7.59256518527138e-02, 7.57972357615699e-02, 7.56278425791521e-02, 7.54097853575786e-02, 7.51350622593066e-02, 7.47955266507745e-02, 7.43830728490451e-02, 7.38898318905257e-02, 7.33083711058953e-02, 7.26318909369610e-02, 7.18544123542415e-02, 7.09709484475667e-02, 6.99776542475519e-02, 6.88719495453256e-02, 6.76526103527380e-02, 6.63198256146984e-02, 6.48752168319078e-02, 6.33218193827787e-02, 6.16640255893941e-02, 5.99074910185209e-02, 5.80590071973290e-02, 5.61263458100427e-02, 5.41180813840327e-02, 5.20434012666868e-02, 4.99119130796544e-02, 4.77334604882031e-02, 4.55179577501639e-02, 4.32752520622302e-02, 4.10150201079915e-02, 3.87467016261667e-02, 3.64794689827050e-02, 3.42222278544853e-02, 3.19836405869864e-02, 2.97721619288609e-02, 2.75960762911041e-02, 2.54635246755015e-02, 2.33825107496557e-02, 2.13608810796728e-02, 1.94062783182432e-02, 1.75260658370036e-02, 1.57272247798232e-02, 1.40162312425360e-02, 1.23989247146454e-02, 1.08803770385606e-02, 9.46476907793681e-03, 8.15528227369100e-03, 6.95401182642286e-03, 5.86190612949622e-03, 4.87873448605569e-03, 4.00308320543274e-03, 3.23237884298067e-03, 2.56293635109728e-03, 1.99002926869696e-03, 1.50797883797793e-03, 1.11025908927331e-03, 7.89615329566118e-04, 5.38194020791138e-04, 3.47682693160323e-04, 2.09459328132766e-04, 1.14751681195457e-04, 5.48085778717677e-05, 2.10880398157087e-05, 5.47322862587407e-06, 5.43897556326556e-07],
[ 6.80000000000000e+01, 6.79999999277946e+01, 6.79999976950623e+01, 6.79999825679472e+01, 6.79999269586921e+01, 6.79997787219211e+01, 6.79994542945863e+01, 6.79988329207667e+01, 6.79977521691780e+01, 6.79960049112904e+01, 6.79933378828037e+01, 6.79894519027304e+01, 6.79840037746361e+01, 6.79766098459784e+01, 6.79668511562696e+01, 6.79542800649632e+01, 6.79384282169254e+01, 6.79188156776194e+01, 6.78949610510639e+01, 6.78663923794139e+01, 6.78326586108296e+01, 6.77933414088052e+01, 6.77480670582411e+01, 6.76965181993672e+01, 6.76384450903608e+01, 6.75736760660723e+01, 6.75021268293994e+01, 6.74238081914453e+01, 6.73388318754627e+01, 6.72474140256149e+01, 6.71498761198479e+01, 6.70466430773424e+01, 6.69382384706221e+01, 6.68252768910180e+01, 6.67084536608417e+01, 6.65885322217711e+01, 6.64663296430023e+01, 6.63427007741525e+01, 6.62185216107966e+01, 6.60946724441440e+01, 6.59720213346874e+01, 6.58514083902207e+01, 6.57336312508643e+01, 6.56194320973540e+01, 6.55094864123902e+01, 6.54043936448098e+01, 6.53046698567981e+01, 6.52107423770242e+01, 6.51229464372717e+01, 6.50415237353148e+01, 6.49666228402233e+01, 6.48983013355252e+01, 6.48365295784903e+01, 6.47811959384181e+01, 6.47321133619996e+01, 6.46890270988893e+01, 6.46516234054519e+01, 6.46195390295125e+01, 6.45923712645114e+01, 6.45696883486269e+01, 6.45510399742464e+01, 6.45359676667611e+01, 6.45240147901075e+01, 6.45147359407504e+01, 6.45077055026409e+01, 6.45025251535296e+01, 6.44988301379753e+01, 6.44962941541357e+01, 6.44946327392538e+01, 6.44936050815503e+01, 6.44930142325759e+01, 6.44927057422840e+01, 6.44925647873524e+01, 6.44925119097786e+01, 6.44924975257480e+01, 6.44924954026935e+01, 6.44924951241935e+01, 6.44924886354780e+01, 6.44924446734585e+01, 6.44922830641557e+01, 6.44918522710276e+01, 6.44909094803414e+01, 6.44891038833080e+01, 6.44859638095337e+01, 6.44808882790764e+01, 6.44731434399910e+01, 6.44618642442239e+01, 6.44460615860721e+01, 6.44246349831586e+01, 6.43963907194519e+01, 6.43600651936248e+01, 6.43143530255665e+01, 6.42579392721975e+01, 6.41895348957966e+01, 6.41079144213432e+01, 6.40119545249649e+01, 6.39006721293665e+01, 6.37732604654115e+01, 6.36291215176951e+01, 6.34678933327329e+01, 6.32894708523166e+01, 6.30940192481867e+01, 6.28819791612529e+01, 6.26540637463239e+01, 6.24112479259018e+01, 6.21547506880553e+01, 6.18860115568518e+01, 6.16066624819908e+01, 6.13184963431987e+01, 6.10234330950712e+01, 6.07234843700007e+01, 6.04207171952430e+01, 6.01172174238590e+01, 5.98150535387212e+01, 5.95162416210250e+01, 5.92227123998286e+01, 5.89362813334924e+01, 5.86586225669404e+01, 5.83912473652538e+01, 5.81354873010612e+01, 5.78924821525553e+01, 5.76631722226064e+01, 5.74482946538495e+01, 5.72483832856258e+01, 5.70637716438914e+01, 5.68945987333357e+01, 5.67408173774226e+01, 5.66022049063477e+01, 5.64783760180482e+01, 5.63687976356940e+01, 5.62728055630525e+01, 5.61896227039499e+01, 5.61183785698837e+01, 5.60581297552922e+01, 5.60078810165442e+01, 5.59666065512419e+01, 5.59332710416070e+01, 5.59068500022451e+01, 5.58863489611601e+01, 5.58708210060468e+01, 5.58593822476936e+01, 5.58512247901051e+01, 5.58456268529933e+01, 5.58419597657113e+01, 5.58396916403516e+01, 5.58383876323108e+01, 5.58377068048433e+01, 5.58373957250966e+01, 5.58372790276881e+01, 5.58372472831015e+01, 5.58372425976648e+01, 5.58372421840299e+01, 5.58372340790581e+01, 5.58371791666694e+01, 5.58369773027439e+01, 5.58364392062814e+01, 5.58352615884754e+01, 5.58330062832055e+01, 5.58290842060993e+01, 5.58227448655725e+01, 5.58130720290283e+01, 5.57989860084882e+01, 5.57792528690131e+01, 5.57525006768790e+01, 5.57172426900805e+01, 5.56719071511740e+01, 5.56148730753369e+01, 5.55445111433623e+01, 5.54592285244143e+01, 5.53575161867160e+01, 5.52379970304530e+01, 5.50994730226011e+01, 5.49409694527672e+01, 5.47617744801467e+01, 5.45614723097807e+01, 5.43399686108119e+01, 5.40975071432891e+01, 5.38346769544718e+01, 5.35524098999206e+01, 5.32519686084422e+01, 5.29349253317952e+01, 5.26031324079702e+01, 5.22586853386536e+01, 5.19038797489505e+01, 5.15411637528994e+01, 5.11730874612245e+01, 5.08022514940019e+01, 5.04312563603139e+01, 5.00626544206543e+01, 4.96989058675784e+01, 4.93423397861061e+01, 4.89951209447287e+01, 4.86592225786949e+01, 4.83364051048832e+01, 4.80282004766728e+01, 4.77359017509580e+01, 4.74605573843248e+01, 4.72029697788540e+01, 4.69636976358032e+01, 4.67430617270084e+01, 4.65411537448296e+01, 4.63578479336644e+01, 4.61928152360457e+01, 4.60455397035774e+01, 4.59153369281006e+01, 4.58013742424916e+01, 4.57026924242143e+01, 4.56182286091495e+01, 4.55468400896413e+01, 4.54873286312332e+01, 4.54384649002350e+01, 4.53990125529817e+01, 4.53677515020385e+01, 4.53434998495876e+01, 4.53251339685687e+01, 4.53116062219109e+01, 4.53019598423086e+01, 4.52953405508119e+01, 4.52910045715807e+01, 4.52883228001698e+01, 4.52867809995977e+01, 4.52859760267342e+01, 4.52856082247838e+01, 4.52854702490565e+01, 4.52854327162643e+01, 4.52854271765011e+01, 4.52854267321435e+01, 4.52854185314929e+01, 4.52853629708705e+01, 4.52851587240273e+01, 4.52846142760272e+01, 4.52834227603002e+01, 4.52811408475293e+01, 4.52771725269538e+01, 4.52707585178639e+01, 4.52609719298661e+01, 4.52467206524069e+01, 4.52267567934568e+01, 4.51996933003926e+01, 4.51640276802707e+01, 4.51181724914972e+01, 4.50604920077680e+01, 4.49893441661968e+01, 4.49031266183065e+01, 4.48003254237316e+01, 4.46795646849091e+01, 4.45396552415448e+01, 4.43796404499876e+01, 4.41988370838907e+01, 4.39968695193987e+01, 4.37736956101537e+01, 4.35296230020102e+01, 4.32653150609336e+01, 4.29817860594661e+01, 4.26803857551080e+01, 4.23627739694644e+01, 4.20308862181927e+01, 4.16868918322098e+01, 4.13331463340625e+01, 4.09721400679553e+01, 4.06064451986360e+01, 4.02386631638840e+01, 3.98713744728147e+01, 3.95070924020462e+01, 3.91482217031854e+01, 3.87970229716074e+01, 3.84555829128798e+01, 3.81257904313490e+01, 3.78093182713498e+01, 3.75076098484873e+01, 3.72218708817880e+01, 3.69530654418603e+01, 3.67019160410143e+01, 3.64689073980942e+01, 3.62542935137954e+01, 3.60581076965369e+01, 3.58801751892661e+01, 3.57201280655081e+01, 3.55774220865231e+01, 3.54513552360262e+01, 3.53410876690611e+01, 3.52456628223598e+01, 3.51640294315944e+01, 3.50950641852329e+01, 3.50375947164092e+01, 3.49904225965079e+01, 3.49523459518271e+01, 3.49221812835429e+01, 3.48987840374907e+01, 3.48810674500408e+01, 3.48680191947769e+01, 3.48587153756542e+01, 3.48523314578598e+01, 3.48481497977566e+01, 3.48455635259594e+01, 3.48440766487181e+01, 3.48433003567418e+01, 3.48429456606864e+01, 3.48428126016528e+01, 3.48427764063414e+01, 3.48427710639883e+01, 3.48427706519346e+01, 3.48427632527730e+01, 3.48427131223626e+01, 3.48425288376406e+01, 3.48420376019996e+01, 3.48409625436344e+01, 3.48389036737336e+01, 3.48353232675629e+01, 3.48295363414001e+01, 3.48207067936526e+01, 3.48078496566122e+01, 3.47898397623751e+01, 3.47654269585535e+01, 3.47332578135772e+01, 3.46919035269749e+01, 3.46398935098359e+01, 3.45757538319588e+01, 3.44980494571579e+01, 3.44054289237799e+01, 3.42966698943775e+01, 3.41707238193571e+01, 3.40267578563716e+01, 3.38641921789718e+01, 3.36827309071031e+01, 3.34823851027009e+01, 3.32634865908076e+01, 3.30266917759091e+01, 3.27729751019807e+01, 3.25036123237924e+01, 3.22201542822165e+01, 3.19243923701996e+01, 3.16183173000331e+01, 3.13040730995607e+01, 3.09839084439233e+01, 3.06601274506866e+01, 3.03350419268103e+01, 3.00109267729963e+01, 2.96899798610862e+01, 2.93742872542187e+01, 2.90657941936366e+01, 2.87662818814755e+01, 2.84773497825968e+01, 2.82004029682248e+01, 2.79366439278295e+01, 2.76870682657544e+01, 2.74524637490894e+01, 2.72334122550719e+01, 2.70302942558489e+01, 2.68432955589920e+01, 2.66724160849087e+01, 2.65174805049177e+01, 2.63781505880430e+01, 2.62539391140615e+01, 2.61442252084395e+01, 2.60482709439867e+01, 2.59652390357077e+01, 2.58942114301313e+01, 2.58342085589782e+01, 2.57842089905336e+01, 2.57431691725335e+01, 2.57100429207798e+01, 2.56838002718906e+01, 2.56634452909450e+01, 2.56480324096655e+01, 2.56366808721213e+01, 2.56285868856854e+01, 2.56230331167112e+01, 2.56193952330616e+01, 2.56171452774032e+01, 2.56158517526524e+01, 2.56151764092913e+01, 2.56148678377634e+01, 2.56147520816765e+01, 2.56147205931902e+01, 2.56147159455548e+01, 2.56147155936498e+01, 2.56147093596889e+01, 2.56146671236831e+01, 2.56145118596098e+01, 2.56140979822245e+01, 2.56131922198072e+01, 2.56114575692034e+01, 2.56084409746525e+01, 2.56035652951576e+01, 2.55961260360345e+01, 2.55852932165631e+01, 2.55701186242203e+01, 2.55495485635219e+01, 2.55224420416079e+01, 2.54875941428877e+01, 2.54437641338896e+01, 2.53897076134882e+01, 2.53242117939035e+01, 2.52461327796032e+01, 2.51544335231683e+01, 2.50482209994636e+01, 2.49267810711105e+01, 2.47896095340134e+01, 2.46364379389416e+01, 2.44672529817808e+01, 2.42823085286504e+01, 2.40821296715340e+01, 2.38675085691327e+01, 2.36394921901091e+01, 2.33993624207073e+01, 2.31486093132677e+01, 2.28888985315492e+01, 2.26220342910536e+01, 2.23499192909406e+01, 2.20745132707668e+01, 2.17977918711136e+01, 2.15217074010280e+01, 2.12481529000192e+01, 2.09789305407268e+01, 2.07157249977631e+01, 2.04600819803607e+01, 2.02133917643764e+01, 1.99768773131064e+01, 1.97515864599785e+01, 1.95383876189629e+01, 1.93379685500440e+01, 1.91508377946923e+01, 1.89773284781845e+01, 1.88176042357864e+01, 1.86716670558245e+01, 1.85393668501501e+01, 1.84204125694570e+01, 1.83143846837636e+01, 1.82207488503878e+01, 1.81388705930817e+01, 1.80680308148835e+01, 1.80074419612369e+01, 1.79562646371546e+01, 1.79136244620503e+01, 1.78786289193510e+01, 1.78503839276932e+01, 1.78280098301950e+01, 1.78106564725011e+01, 1.77975170236586e+01, 1.77878401906079e+01, 1.77809404903867e+01, 1.77762062759096e+01, 1.77731052616422e+01, 1.77711873632070e+01, 1.77700847469511e+01, 1.77695090775155e+01, 1.77692460483954e+01, 1.77691473769298e+01, 1.77691205358835e+01, 1.77691165742007e+01, 1.77691162767959e+01, 1.77691110421346e+01, 1.77690755765326e+01, 1.77689452012104e+01, 1.77685976683334e+01, 1.77678371004465e+01, 1.77663805193283e+01, 1.77638475043312e+01, 1.77597534566800e+01, 1.77535068723487e+01, 1.77444109400990e+01, 1.77316696803628e+01, 1.77143987221210e+01, 1.76916406767365e+01, 1.76623849094659e+01, 1.76255913332168e+01, 1.75802176603666e+01, 1.75252493559576e+01, 1.74597313515329e+01, 1.73828004180361e+01, 1.72937169744439e+01, 1.71918950411119e+01, 1.70769290450602e+01, 1.69486162552146e+01, 1.68069737689292e+01, 1.66522491799429e+01, 1.64849243193740e+01, 1.63057117587628e+01, 1.61155440798091e+01, 1.59155562327594e+01, 1.57070616100711e+01, 1.54915227414728e+01, 1.52705177585794e+01, 1.50457039683463e+01, 1.48187800000021e+01, 1.45914480352444e+01, 1.43653775858117e+01, 1.41421721433610e+01, 1.39233398019298e+01, 1.37102686625396e+01, 1.35042075011430e+01, 1.33062518480939e+01, 1.31173353216196e+01, 1.29382258053837e+01, 1.27695258777933e+01, 1.26116767942875e+01, 1.24649652898252e+01, 1.23295324961513e+01, 1.22053843417317e+01, 1.20924029045941e+01, 1.19903583036527e+01, 1.18989208288076e+01, 1.18176731137946e+01, 1.17461222413448e+01, 1.16837117336797e+01, 1.16298334212928e+01, 1.15838391999501e+01, 1.15450526820845e+01, 1.15127807276193e+01, 1.14863248050162e+01, 1.14649920908753e+01, 1.14481061709884e+01, 1.14350171626826e+01, 1.14251110427006e+01, 1.14178179411892e+01, 1.14126191541275e+01, 1.14090526358707e+01, 1.14067167610695e+01, 1.14052721900416e+01, 1.14044417311815e+01, 1.14040081643683e+01, 1.14038100658352e+01, 1.14037357525154e+01, 1.14037155375169e+01, 1.14037125538265e+01, 1.14037123305210e+01, 1.14037084091654e+01, 1.14036818414192e+01, 1.14035841757983e+01, 1.14033238368343e+01, 1.14027541001242e+01, 1.14016630196789e+01, 1.13997657265176e+01, 1.13966994780435e+01, 1.13920217912052e+01, 1.13852119356927e+01, 1.13756759924183e+01, 1.13627555930376e+01, 1.13457403447636e+01, 1.13238838094365e+01, 1.12964227476685e+01, 1.12625991623476e+01, 1.12216844893224e+01, 1.11730050991308e+01, 1.11159681080358e+01, 1.10500863673949e+01, 1.09750014258980e+01, 1.08905032559045e+01, 1.07965456150242e+01, 1.06932560823691e+01, 1.05809400619442e+01, 1.04600783699465e+01, 1.03313183951730e+01, 1.01954592111911e+01, 1.00534313895433e+01, 9.90627257886732e+00, 9.75510014397642e+00, 9.60108227974312e+00, 9.44540901814885e+00, 9.28926443862836e+00, 9.13380119070959e+00, 8.98011817269686e+00, 8.82924191443474e+00, 8.68211191911143e+00, 8.53956995636494e+00, 8.40235308591024e+00, 8.27109003717219e+00, 8.14630047678079e+00, 8.02839665626303e+00, 7.91768693669301e+00, 7.81438072367838e+00, 7.71859440346984e+00, 7.63035793937334e+00, 7.54962185930035e+00, 7.47626443456414e+00, 7.41009891302402e+00, 7.35088072385074e+00, 7.29831461496432e+00, 7.25206171671096e+00, 7.21174654627121e+00, 7.17696397676708e+00, 7.14728619365651e+00, 7.12226964983589e+00, 7.10146201147030e+00, 7.08440906097148e+00, 7.07066149421986e+00, 7.05978151889201e+00, 7.05134913260807e+00, 7.04496793653225e+00, 7.04027032475096e+00, 7.03692188439640e+00, 7.03462484748767e+00, 7.03312045328752e+00, 7.03219010903934e+00, 7.03165527566142e+00, 7.03137605087274e+00, 7.03124847222971e+00, 7.03120061331388e+00, 7.03118759455080e+00, 7.03118567300945e+00, 7.03118552914357e+00, 7.03118300206749e+00, 7.03116588076716e+00, 7.03110294123573e+00, 7.03093516893337e+00, 7.03056801035664e+00, 7.02986488459446e+00, 7.02864222696780e+00, 7.02666631263136e+00, 7.02365207922238e+00, 7.01926413196492e+00, 7.01312006906943e+00, 7.00479620662120e+00, 6.99383570775135e+00, 6.97975902933339e+00, 6.96207649141071e+00, 6.94030265342387e+00, 6.91397205356675e+00, 6.88265574285005e+00, 6.84597793579798e+00, 6.80363201863175e+00, 6.75539511645991e+00, 6.70114043419439e+00, 6.64084665802152e+00, 6.57460383553818e+00, 6.50261533637646e+00, 6.42519571771146e+00, 6.34276456145476e+00, 6.25583659008444e+00, 6.16500858377331e+00, 6.07094379349191e+00, 5.97435465919457e+00, 5.87598469198718e+00, 5.77659036441727e+00, 5.67692378024472e+00, 5.57771677576513e+00, 5.47966695368427e+00, 5.38342598372391e+00, 5.28959033720832e+00, 5.19869446954063e+00, 5.11120633531554e+00, 5.02752502253305e+00, 4.94798022756343e+00, 4.87283325997254e+00, 4.80227926189681e+00, 4.73645034436272e+00, 4.67541937609507e+00, 4.61920420255093e+00, 4.56777211867546e+00, 4.52104446394645e+00, 4.47890124962341e+00, 4.44118576376793e+00, 4.40770912839129e+00, 4.37825480443497e+00, 4.35258305403998e+00, 4.33043537586823e+00, 4.31153892853843e+00, 4.29561095024271e+00, 4.28236317031781e+00, 4.27150619225583e+00, 4.26275380892505e+00, 4.25582719142739e+00, 4.25045887496206e+00, 4.24639645020741e+00, 4.24340585881536e+00, 4.24127418803957e+00, 4.23981186320393e+00, 4.23885414796216e+00, 4.23826188073897e+00, 4.23792140036191e+00, 4.23774364311635e+00, 4.23766242532070e+00, 4.23763195787657e+00, 4.23762367000766e+00, 4.23762244673611e+00, 4.23762235490676e+00, 4.23762073863323e+00, 4.23760978815022e+00, 4.23756953313498e+00, 4.23746222895589e+00, 4.23722740101174e+00, 4.23677769513660e+00, 4.23599570832843e+00, 4.23473195870140e+00, 4.23280413528619e+00, 4.22999774560852e+00, 4.22606824990492e+00, 4.22074473331714e+00, 4.21373511970118e+00, 4.20473287188069e+00, 4.19342505369093e+00, 4.17950155118526e+00, 4.16266516811677e+00, 4.14264223046791e+00, 4.11919326422519e+00, 4.09212325850901e+00, 4.06129100107920e+00, 4.02661698210955e+00, 3.98808940903045e+00, 3.94576796025561e+00, 3.89978502422591e+00, 3.85034431345426e+00, 3.79771689863013e+00, 3.74223486090189e+00, 3.68428289684721e+00, 3.62428831816185e+00, 3.56270995831800e+00, 3.50002652766497e+00, 3.43672494773270e+00, 3.37328914997245e+00, 3.31018975172749e+00, 3.24787493211268e+00, 3.18676273196522e+00, 3.12723490343569e+00, 3.06963234294076e+00, 3.01425206117784e+00, 2.96134557906599e+00, 2.91111859058694e+00, 2.86373170289320e+00, 2.81930204987025e+00, 2.77790557574392e+00, 2.73957979777776e+00, 2.70432687863565e+00, 2.67211686648746e+00, 2.64289099140770e+00, 2.61656493736980e+00, 2.59303203792777e+00, 2.57216636876740e+00, 2.55382573049681e+00, 2.53785452961100e+00, 2.52408657424620e+00, 2.51234780425662e+00, 2.50245897276181e+00, 2.49423828938551e+00, 2.48750402494070e+00, 2.48207706452473e+00, 2.47778338222764e+00, 2.47445639735807e+00, 2.47193916065524e+00, 2.47008631066093e+00, 2.46876573629723e+00, 2.46785988241813e+00, 2.46726664092922e+00, 2.46689978079264e+00, 2.46668888519694e+00, 2.46657878233162e+00, 2.46652847625465e+00, 2.46650960484291e+00, 2.46650447137407e+00, 2.46650371368528e+00, 2.46650365656438e+00, 2.46650264796149e+00, 2.46649581453512e+00, 2.46647069423635e+00, 2.46640373350596e+00, 2.46625719545824e+00, 2.46597657240726e+00, 2.46548861300961e+00, 2.46470006559227e+00, 2.46349722877639e+00, 2.46174638718938e+00, 2.45929519245564e+00, 2.45597502626320e+00, 2.45160435183901e+00, 2.44599302191844e+00, 2.43894746546808e+00, 2.43027662342018e+00, 2.41979844828847e+00, 2.40734672790856e+00, 2.39277794497065e+00, 2.37597784742343e+00, 2.35686738609706e+00, 2.33540767996915e+00, 2.31160369950973e+00, 2.28550641505555e+00, 2.25721323771282e+00, 2.22686667934282e+00, 2.19465126759650e+00, 2.16078886185835e+00, 2.12553261595757e+00, 2.08915991407287e+00, 2.05196465996990e+00, 2.01424932220051e+00, 1.97631712836051e+00, 1.93846476267225e+00, 1.90097585877017e+00, 1.86411550141779e+00, 1.82812586571548e+00, 1.79322303864888e+00, 1.75959499281777e+00, 1.72740062115395e+00, 1.69676969737959e+00, 1.66780360063092e+00, 1.64057663291273e+00, 1.61513776232104e+00, 1.59151263993256e+00, 1.56970576034776e+00, 1.54970266171888e+00, 1.53147208783749e+00, 1.51496806028362e+00, 1.50013183121072e+00, 1.48689370612465e+00, 1.47517474057662e+00, 1.46488832495841e+00, 1.45594167775209e+00, 1.44823726998604e+00, 1.44167420272879e+00, 1.43614955571732e+00, 1.43155971922447e+00, 1.42780171363438e+00, 1.42477449257684e+00, 1.42238021658807e+00, 1.42052547585750e+00, 1.41912243342570e+00, 1.41808985489692e+00, 1.41735398788265e+00, 1.41684925438275e+00, 1.41651872228844e+00, 1.41631432805143e+00, 1.41619683093369e+00, 1.41613548953886e+00, 1.41610746276304e+00, 1.41609694905117e+00, 1.41609408907601e+00, 1.41609366695000e+00, 1.41609363495343e+00, 1.41609306768234e+00, 1.41608922434214e+00, 1.41607509588938e+00, 1.41603743517807e+00, 1.41595501845616e+00, 1.41579719139344e+00, 1.41552276266098e+00, 1.41507930374966e+00, 1.41440290841790e+00, 1.41341845813885e+00, 1.41204043020119e+00, 1.41017427167084e+00, 1.40771834434686e+00, 1.40456642262276e+00, 1.40061069789257e+00, 1.39574521074366e+00, 1.38986959753116e+00, 1.38289300388065e+00, 1.37473798783945e+00, 1.36534421386713e+00, 1.35467172958930e+00, 1.34270362346497e+00, 1.32944788505539e+00, 1.31493833030162e+00, 1.29923450973496e+00, 1.28242058329478e+00, 1.26460321511158e+00, 1.24590860802229e+00, 1.22647885364030e+00, 1.20646781363668e+00, 1.18603676773758e+00, 1.16535006268683e+00, 1.14457097564737e+00, 1.12385796911003e+00, 1.10336146777279e+00, 1.08322123706590e+00, 1.06356439369671e+00, 1.04450403535027e+00, 1.02613844254834e+00, 1.00855078202397e+00, 9.91809227717098e-01, 9.75967411417301e-01, 9.61065118281340e-01, 9.47129150788527e-01, 9.34174296110911e-01, 9.22204344621733e-01, 9.11213120008853e-01, 9.01185493289873e-01, 8.92098363391639e-01, 8.83921595597926e-01, 8.76618916001391e-01, 8.70148765161588e-01, 8.64465117553152e-01, 8.59518275203982e-01, 8.55255644293024e-01, 8.51622502536827e-01, 8.48562763099963e-01, 8.46019737710222e-01, 8.43936897883708e-01, 8.42258628955733e-01, 8.40930967307045e-01, 8.39902307134992e-01, 8.39124059721840e-01, 8.38551245744966e-01, 8.38143000048518e-01, 8.37862968647727e-01, 8.37679579646192e-01, 8.37566173147437e-01, 8.37500979944863e-01, 8.37466944457233e-01, 8.37451393644798e-01, 8.37445560043975e-01, 8.37443973167424e-01, 8.37443738947925e-01, 8.37443721081755e-01, 8.37443402844073e-01, 8.37441246739331e-01, 8.37433320705813e-01, 8.37412193083821e-01, 8.37365957186819e-01, 8.37277415227206e-01, 8.37123456644633e-01, 8.36874663278991e-01, 8.36495170669124e-01, 8.35942810653648e-01, 8.35169554895955e-01, 8.34122271415097e-01, 8.32743796219816e-01, 8.30974309452657e-01, 8.28752990131264e-01, 8.26019906142363e-01, 8.22718077634065e-01, 8.18795633905964e-01, 8.14207968268434e-01, 8.08919784290865e-01, 8.02906922442972e-01, 7.96157859960474e-01, 7.88674789630107e-01, 7.80474204808129e-01, 7.71586946889641e-01, 7.62057705080565e-01, 7.51943993364975e-01, 7.41314662406452e-01, 7.30248031447246e-01, 7.18829744543513e-01, 7.07150465318264e-01, 6.95303524721439e-01, 6.83382628117243e-01, 6.71479713274948e-01, 6.59683031858208e-01, 6.48075506136010e-01, 6.36733391911150e-01, 6.25725259604692e-01, 6.15111289018046e-01, 6.04942859994395e-01, 5.95262411125546e-01, 5.86103531670631e-01, 5.77491247727508e-01, 5.69442462122595e-01, 5.61966508144735e-01, 5.55065779798397e-01, 5.48736405317896e-01, 5.42968935861950e-01, 5.37749027156863e-01, 5.33058097943012e-01, 5.28873954976930e-01, 5.25171379684124e-01, 5.21922676047410e-01, 5.19098182737238e-01, 5.16666754731276e-01, 5.14596220696049e-01, 5.12853822276028e-01, 5.11406640283482e-01, 5.10222010799808e-01, 5.09267931634355e-01, 5.08513456718788e-01, 5.07929073143160e-01, 5.07487052963782e-01, 5.07161769902197e-01, 5.06929969832484e-01, 5.06770983673931e-01, 5.06666872032028e-01, 5.06602492635182e-01, 5.06565484175724e-01, 5.06546163382268e-01, 5.06537335770787e-01, 5.06534024264340e-01, 5.06533123457658e-01, 5.06532990500574e-01, 5.06532980287091e-01, 5.06532797423233e-01, 5.06531558495510e-01, 5.06527004093300e-01, 5.06514863937866e-01, 5.06488296563489e-01, 5.06437420920989e-01, 5.06348960449827e-01, 5.06206018904692e-01, 5.05988005662857e-01, 5.05670726119149e-01, 5.05226649756423e-01, 5.04625364204131e-01, 5.03834217761138e-01, 5.02819145275026e-01, 5.01545662964929e-01, 4.99980007036839e-01, 4.98090379367966e-01, 4.95848252035486e-01, 4.93229672189486e-01, 4.90216501020328e-01, 4.86797516625884e-01, 4.82969311517806e-01, 4.78736921985629e-01, 4.74114138675517e-01, 4.69123464996770e-01, 4.63795711166824e-01, 4.58169235116472e-01, 4.52288865034589e-01, 4.46204559841235e-01, 4.39969881298396e-01, 4.33640363127921e-01, 4.27271867301513e-01, 4.20919015182562e-01, 4.14633771764144e-01, 4.08464245914118e-01, 4.02453749983451e-01, 3.96640140466459e-01, 3.91055439897613e-01, 3.85725720978229e-01, 3.80671218758124e-01, 3.75906626621401e-01, 3.71441527155156e-01, 3.67280909315710e-01, 3.63425727703188e-01, 3.59873466944251e-01, 3.56618682830756e-01, 3.53653500773632e-01, 3.50968060379799e-01, 3.48550901951043e-01, 3.46389296146015e-01, 3.44469521903221e-01, 3.42777100124522e-01, 3.41296991790756e-01, 3.40013769375048e-01, 3.38911769883481e-01, 3.37975236790442e-01, 3.37188456721535e-01, 3.36535895096688e-01, 3.36002333191971e-01, 3.35573007295039e-01, 3.35233748906845e-01, 3.34971123365278e-01, 3.34772562923803e-01, 3.34626489296781e-01, 3.34522420060679e-01, 3.34451053129925e-01, 3.34404323838413e-01, 3.34375429934809e-01, 3.34358820994325e-01, 3.34350150264285e-01, 3.34346188680797e-01, 3.34344702577468e-01, 3.34344298323543e-01, 3.34344238656584e-01, 3.34344234038142e-01, 3.34344150892960e-01, 3.34343587573272e-01, 3.34341516768319e-01, 3.34335996898844e-01, 3.34323917490849e-01, 3.34300786628390e-01, 3.34260570151613e-01, 3.34195591874035e-01, 3.34096503567210e-01, 3.33952332634809e-01, 3.33750614054111e-01, 3.33477611067465e-01, 3.33118626054173e-01, 3.32658398877342e-01, 3.32081584785399e-01, 3.31373297850845e-01, 3.30519699358305e-01, 3.29508604145985e-01, 3.28330072476518e-01, 3.26976951472801e-01, 3.25445329372201e-01, 3.23734868486426e-01, 3.21848989087698e-01, 3.19794886247749e-01, 3.17583374133747e-01, 3.15228566101041e-01, 3.12747412435982e-01, 3.10159129033295e-01, 3.07484558102033e-01, 3.04745505174125e-01, 3.01964094966735e-01, 2.99162182532072e-01, 2.96360846804916e-01, 2.93579982711677e-01, 2.90837997107005e-01, 2.88151604369347e-01, 2.85535710462085e-01, 2.83003370039594e-01, 2.80565799613029e-01, 2.78232430394539e-01, 2.76010986521911e-01, 2.73907577237480e-01, 2.71926794660743e-01, 2.70071811643902e-01, 2.68344476588808e-01, 2.66745403943688e-01, 2.65274060404309e-01, 2.63928847686387e-01, 2.62707183208860e-01, 2.61605580226714e-01, 2.60619728949294e-01, 2.59744580038074e-01, 2.58974431636148e-01, 2.58303020765524e-01, 2.57723619555421e-01, 2.57229136348343e-01, 2.56812221277670e-01, 2.56465375438680e-01, 2.56181062293379e-01, 2.55951819485887e-01, 2.55770368818943e-01, 2.55629721786391e-01, 2.55523277799966e-01, 2.55444912120979e-01, 2.55389050534873e-01, 2.55350728001730e-01, 2.55325628885771e-01, 2.55310106901703e-01, 2.55301183588159e-01, 2.55296524897219e-01, 2.55294396321787e-01, 2.55293597822728e-01, 2.55293380611889e-01, 2.55293348552044e-01, 2.55293346050351e-01, 2.55293300753223e-01, 2.55292993858717e-01, 2.55291865685322e-01, 2.55288858414573e-01, 2.55282277204387e-01, 2.55269673891277e-01, 2.55247758161772e-01, 2.55212340641691e-01, 2.55158311831877e-01, 2.55079660172314e-01, 2.54969531677190e-01, 2.54820332492031e-01, 2.54623874323158e-01, 2.54371560951981e-01, 2.54054611989673e-01, 2.53664317726248e-01, 2.53192316531951e-01, 2.52630883994428e-01, 2.51973221091234e-01, 2.51213727493166e-01, 2.50348245832541e-01, 2.49374263627800e-01, 2.48291061584936e-01, 2.47099800080647e-01, 2.45803539502489e-01, 2.44407194375299e-01, 2.42917425377129e-01, 2.41342476992671e-01, 2.39691971310521e-01, 2.37976670126912e-01, 2.36208218024842e-01, 2.34398878557812e-01, 2.32561274293971e-01, 2.30708139546806e-01, 2.28852092414854e-01, 2.27005430529772e-01, 2.25179952871486e-01, 2.23386808285781e-01, 2.21636370012789e-01, 2.19938134620406e-01, 2.18300643215841e-01, 2.16731422623283e-01, 2.15236944294205e-01, 2.13822598975909e-01, 2.12492685525106e-01, 2.11250412643281e-01, 2.10097912676943e-01, 2.09036266923647e-01, 2.08065542094531e-01, 2.07184837692338e-01, 2.06392344074544e-01, 2.05685410891704e-01, 2.05060625434484e-01, 2.04513900207601e-01, 2.04040568787775e-01, 2.03635488736182e-01, 2.03293150037453e-01, 2.03007787243379e-01, 2.02773493228733e-01, 2.02584332230487e-01, 2.02434449663230e-01, 2.02318176093513e-01, 2.02230122733141e-01, 2.02165265887459e-01, 2.02119017977466e-01, 2.02087283048542e-01, 2.02066495078412e-01, 2.02053637892064e-01, 2.02046246062181e-01, 2.02042386795668e-01, 2.02040623450060e-01, 2.02039961955833e-01, 2.02039782013131e-01, 2.02039755453967e-01, 2.02039753363863e-01, 2.02039715293353e-01, 2.02039457359987e-01, 2.02038509169000e-01, 2.02035981634992e-01, 2.02030450163662e-01, 2.02019856638529e-01, 2.02001434109150e-01, 2.01971657596147e-01, 2.01926223868597e-01, 2.01860062401522e-01, 2.01767378962089e-01, 2.01641732389021e-01, 2.01476144099288e-01, 2.01263238682643e-01, 2.00995412639088e-01, 2.00665026923417e-01, 2.00264617556542e-01, 1.99787117245876e-01, 1.99226079857453e-01, 1.98575898831365e-01, 1.97832010361111e-01, 1.96991072460051e-01, 1.96051111956592e-01, 1.95011632968132e-01, 1.93873682397424e-01, 1.92639870304149e-01, 1.91314345412451e-01, 1.89902728287620e-01, 1.88412006642579e-01, 1.86850398657157e-01, 1.85227191026868e-01, 1.83552558704449e-01, 1.81837373030390e-01, 1.80093004299508e-01, 1.78331123935322e-01, 1.76563510497388e-01, 1.74801862856015e-01, 1.73057623120201e-01, 1.71341811336305e-01, 1.69664873584878e-01, 1.68036544856903e-01, 1.66465727938260e-01, 1.64960389419136e-01, 1.63527473828452e-01, 1.62172836740268e-01, 1.60901197496287e-01, 1.59716111933221e-01, 1.58619965209056e-01, 1.57613984499321e-01, 1.56698271004757e-01, 1.55871850383662e-01, 1.55132740408931e-01, 1.54478034354185e-01, 1.53903998337142e-01, 1.53406180590778e-01, 1.52979530393466e-01, 1.52618524169640e-01, 1.52317296076518e-01, 1.52069770229190e-01, 1.51869791595246e-01, 1.51711252526368e-01, 1.51588211900304e-01, 1.51495003935523e-01, 1.51426333924533e-01, 1.51377358412006e-01, 1.51343747724610e-01, 1.51321729230887e-01, 1.51308110260089e-01, 1.51300280218891e-01, 1.51296192090896e-01, 1.51294324159309e-01, 1.51293623428472e-01, 1.51293432812210e-01, 1.51293404677617e-01, 1.51293402443969e-01, 1.51293361509133e-01, 1.51293084169566e-01, 1.51292064637374e-01, 1.51289346919157e-01, 1.51283399159382e-01, 1.51272008104826e-01, 1.51252197766652e-01, 1.51220175774213e-01, 1.51171310247221e-01, 1.51100139257750e-01, 1.51000414112943e-01, 1.50865176758543e-01, 1.50686870586540e-01, 1.50457482836302e-01, 1.50168715627800e-01, 1.49812181494127e-01, 1.49379618137829e-01, 1.48863116088961e-01, 1.48255352066414e-01, 1.47549820219581e-01, 1.46741053124515e-01, 1.45824824483498e-01, 1.44798325951541e-01, 1.43660311377673e-01, 1.42411202951124e-01, 1.41053155201320e-01, 1.39590074412183e-01, 1.38027592667029e-01, 1.36372997342846e-01, 1.34635118349749e-01, 1.32824176714237e-01, 1.30951599219330e-01, 1.29029804734941e-01, 1.27071968606231e-01, 1.25091772011601e-01, 1.23103143544157e-01, 1.21120000385488e-01, 1.19155996300151e-01, 1.17224283260867e-01, 1.15337292813905e-01, 1.13506542334690e-01, 1.11742470155164e-01, 1.10054302248059e-01, 1.08449951817022e-01, 1.06935951869815e-01, 1.05517419723441e-01, 1.04198051480097e-01, 1.02980143850459e-01, 1.01864640301656e-01, 1.00851198343889e-01, 9.99382748045434e-02, 9.91232261097537e-02, 9.84024208420067e-02, 9.77713621040550e-02, 9.72248174484776e-02, 9.67569542882813e-02, 9.63614787715272e-02, 9.60317760778793e-02, 9.57610499935271e-02, 9.55424594678504e-02, 9.53692496885452e-02, 9.52348750690533e-02, 9.51331114634399e-02, 9.50581549421088e-02, 9.50047046019336e-02, 9.49680271586431e-02, 9.49440014805560e-02, 9.49291417552161e-02, 9.49205986144648e-02, 9.49161382455372e-02, 9.49141002479621e-02, 9.49133357217704e-02, 9.49131277518376e-02, 9.49130970558940e-02, 9.49130945965919e-02, 9.49130492453492e-02, 9.49127419839613e-02, 9.49116124578921e-02, 9.49086015506424e-02, 9.49020122259299e-02, 9.48893927741921e-02, 9.48674471085269e-02, 9.48319761856879e-02, 9.47778539694895e-02, 9.46990405957138e-02, 9.45886345237454e-02, 9.44389644543548e-02, 9.42417206463801e-02, 9.39881239783773e-02, 9.36691296920173e-02, 9.32756612611314e-02, 9.27988683086909e-02, 9.22304010214248e-02, 9.15626921781723e-02, 9.07892368115029e-02, 8.99048587622389e-02, 8.89059530550043e-02, 8.77906931946626e-02, 8.65591932105190e-02, 8.52136155814094e-02, 8.37582180483736e-02, 8.21993347142467e-02, 8.05452896532017e-02, 7.88062443770627e-02, 7.69939837602323e-02, 7.51216482030799e-02, 7.32034226872429e-02, 7.12541957069845e-02, 6.92892026347067e-02, 6.73236687311061e-02, 6.53724666555604e-02, 6.34498019827987e-02, 6.15689380026225e-02, 5.97419681813989e-02, 5.79796413690161e-02, 5.62912414559946e-02, 5.46845200205297e-02, 5.31656778123448e-02, 5.17393888818987e-02, 5.04088598715811e-02, 4.91759164415697e-02, 4.80411089349815e-02, 4.70038300641562e-02, 4.60624384654977e-02, 4.52143832602587e-02, 4.44563261159203e-02, 4.37842585943145e-02, 4.31936137007511e-02, 4.26793714378704e-02, 4.22361587922069e-02, 4.18583449294232e-02, 4.15401324646609e-02, 4.12756455443031e-02, 4.10590151677042e-02, 4.08844617508777e-02, 4.07463744430032e-02, 4.06393862147070e-02, 4.05584432970869e-02, 4.04988672134010e-02, 4.04564074512308e-02, 4.04272827947239e-02, 4.04082094864293e-02, 4.03964147077573e-02, 4.03896343342609e-02, 4.03860944996666e-02, 4.03844771499848e-02, 4.03838704311568e-02, 4.03837053893739e-02, 4.03836810295815e-02, 4.03836790595165e-02, 4.03836425010979e-02, 4.03833948127783e-02, 4.03824842893954e-02, 4.03800572076420e-02, 4.03747457978199e-02, 4.03645745760173e-02, 4.03468891253311e-02, 4.03183111519212e-02, 4.02747233202092e-02, 4.02112869419966e-02, 4.01224949921755e-02, 4.00022620858778e-02, 3.98440519187346e-02, 3.96410412088386e-02, 3.93863173893862e-02, 3.90731052284394e-02, 3.86950153013856e-02, 3.82463049627262e-02, 3.77221403590369e-02, 3.71188463284966e-02, 3.64341299849258e-02, 3.56672636169719e-02, 3.48192134195955e-02, 3.38927026179050e-02, 3.28922007311121e-02, 3.18238349251772e-02, 3.06952243533232e-02, 2.95152436971762e-02, 2.82937273248933e-02, 2.70411300542861e-02, 2.57681639547798e-02, 2.44854325482038e-02, 2.32030839479620e-02, 2.19305028931984e-02, 2.06760584706132e-02, 1.94469199056755e-02, 1.82489475774616e-02, 1.70866608451945e-02, 1.59632788622588e-02, 1.48808257549575e-02, 1.38402877328332e-02, 1.28418071009690e-02, 1.18848968287182e-02, 1.09686592546581e-02, 1.00919935561166e-02, 9.25377866828114e-03, 8.45302128364267e-03, 7.68896218229997e-03, 6.96113806945675e-03, 6.26939992420716e-03, 5.61389225482856e-03, 4.99500030424826e-03, 4.41327393029667e-03, 3.86933754042218e-03, 3.36379518363117e-03, 2.89713885950960e-03, 2.46966650232139e-03, 2.08141418674757e-03, 1.73210514986977e-03, 1.42111645727994e-03, 1.14746272188881e-03, 9.09795291050841e-04, 7.06414756052709e-04, 5.35294458174462e-04, 3.94112777641453e-04, 2.80292299237317e-04, 1.91044361443460e-04, 1.23417952536897e-04, 7.43523907260624e-05, 4.07337390129214e-05, 1.94555605759206e-05, 7.48568290021012e-06, 1.94284788790355e-06, -1.42108547152020e-14],
[ 6.90000000000000e+01, 6.89999999262531e+01, 6.89999976458518e+01, 6.89999821957718e+01, 6.89999253992553e+01, 6.89997739976140e+01, 6.89994426436977e+01, 6.89988080033420e+01, 6.89977041770560e+01, 6.89959196133745e+01, 6.89931956391674e+01, 6.89892266826614e+01, 6.89836622141768e+01, 6.89761103799876e+01, 6.89661432586274e+01, 6.89533036284364e+01, 6.89371131016054e+01, 6.89170814539200e+01, 6.88927169601581e+01, 6.88635375307316e+01, 6.88290824327298e+01, 6.87889243645300e+01, 6.87426816343937e+01, 6.86900301680159e+01, 6.86307150381066e+01, 6.85645611738709e+01, 6.84914828757474e+01, 6.84114917392852e+01, 6.83247025907660e+01, 6.82313370643330e+01, 6.81317245111549e+01, 6.80263000261236e+01, 6.79155995019111e+01, 6.78002517639566e+01, 6.76809679894652e+01, 6.75585287536426e+01, 6.74337691630743e+01, 6.73075626186162e+01, 6.71808037924881e+01, 6.70543914059466e+01, 6.69292113593548e+01, 6.68061207036761e+01, 6.66859328613113e+01, 6.65694044148376e+01, 6.64572236933927e+01, 6.63500013047613e+01, 6.62482626907140e+01, 6.61524427254173e+01, 6.60628823314261e+01, 6.59798270531916e+01, 6.59034275017718e+01, 6.58337415639751e+01, 6.57707382521908e+01, 6.57143030558395e+01, 6.56642446404754e+01, 6.56203027254340e+01, 6.55821569554416e+01, 6.55494365661107e+01, 6.55217306284300e+01, 6.54985986441469e+01, 6.54795812534346e+01, 6.54642108095683e+01, 6.54520215736317e+01, 6.54425592865088e+01, 6.54353898863592e+01, 6.54301071578769e+01, 6.54263391249991e+01, 6.54237530310420e+01, 6.54220587887685e+01, 6.54210108265072e+01, 6.54204083036971e+01, 6.54200937183986e+01, 6.54199499785598e+01, 6.54198960562632e+01, 6.54198813880420e+01, 6.54198792230415e+01, 6.54198789390391e+01, 6.54198723221237e+01, 6.54198274915309e+01, 6.54196626892632e+01, 6.54192233848791e+01, 6.54182619676001e+01, 6.54164206987299e+01, 6.54132185923221e+01, 6.54080428032876e+01, 6.54001450000079e+01, 6.53886430824360e+01, 6.53725284755382e+01, 6.53506790806380e+01, 6.53218778031212e+01, 6.52848363943621e+01, 6.52382241502075e+01, 6.51807008011040e+01, 6.51109527151942e+01, 6.50277313233736e+01, 6.49298924761438e+01, 6.48164352728135e+01, 6.46865387864278e+01, 6.45395950693902e+01, 6.43752368922810e+01, 6.41933588626330e+01, 6.39941308967219e+01, 6.37780034573932e+01, 6.35457044788624e+01, 6.32982284066781e+01, 6.30368182098204e+01, 6.27629415055551e+01, 6.24782620415939e+01, 6.21846077153415e+01, 6.18839361329880e+01, 6.15782985075445e+01, 6.12698025510229e+01, 6.09605749875082e+01, 6.06527244033055e+01, 6.03483053038475e+01, 6.00492843751639e+01, 5.97575099633172e+01, 5.94746856426591e+01, 5.92023484597255e+01, 5.89418520845339e+01, 5.86943547661244e+01, 5.84608117475370e+01, 5.82419716776040e+01, 5.80383765498830e+01, 5.78503647626988e+01, 5.76780769833390e+01, 5.75214645791844e+01, 5.73803004305290e+01, 5.72541919598529e+01, 5.71425962048935e+01, 5.70448367358864e+01, 5.69601221785506e+01, 5.68875660598482e+01, 5.68262076476275e+01, 5.67750334110869e+01, 5.67329986891879e+01, 5.66990491212034e+01, 5.66721413701349e+01, 5.66512626584721e+01, 5.66354486392604e+01, 5.66237991458059e+01, 5.66154914019471e+01, 5.66097903319763e+01, 5.66060556841152e+01, 5.66037457717605e+01, 5.66024177391541e+01, 5.66017243683522e+01, 5.66014075573734e+01, 5.66012887099690e+01, 5.66012563805303e+01, 5.66012516087706e+01, 5.66012511875150e+01, 5.66012429332197e+01, 5.66011870091425e+01, 5.66009814261390e+01, 5.66004334159494e+01, 5.65992341020393e+01, 5.65969372455409e+01, 5.65929429086415e+01, 5.65864867720848e+01, 5.65766357209828e+01, 5.65622901716435e+01, 5.65421934482263e+01, 5.65149483282269e+01, 5.64790406574984e+01, 5.64328696886228e+01, 5.63747845246441e+01, 5.63031257622476e+01, 5.62162711390334e+01, 5.61126837189604e+01, 5.59909609232131e+01, 5.58498825574408e+01, 5.56884559250635e+01, 5.55059561673312e+01, 5.53019601387406e+01, 5.50763723999156e+01, 5.48294422619967e+01, 5.45617712092678e+01, 5.42743104219441e+01, 5.39683484913237e+01, 5.36454897557831e+01, 5.33076239964427e+01, 5.29568885303888e+01, 5.25956240330721e+01, 5.22263256954828e+01, 5.18515915402249e+01, 5.14740698378236e+01, 5.10964075426206e+01, 5.07212014930128e+01, 5.03509538120415e+01, 4.99880325476276e+01, 4.96346381678522e+01, 4.92927761345376e+01, 4.89642354616095e+01, 4.86505729446021e+01, 4.83531026236250e+01, 4.80728899974953e+01, 4.78107505170964e+01, 4.75672519266053e+01, 4.73427200720688e+01, 4.71372478447286e+01, 4.69507069646679e+01, 4.67827623364616e+01, 4.66328887226859e+01, 4.65003894843043e+01, 4.63844171299301e+01, 4.62839953993714e+01, 4.61980425813030e+01, 4.61253957314527e+01, 4.60648354182493e+01, 4.60151105804003e+01, 4.59749630392902e+01, 4.59431511731024e+01, 4.59184722342252e+01, 4.58997827816664e+01, 4.58860167100907e+01, 4.58762003896870e+01, 4.58694644877925e+01, 4.58650521236007e+01, 4.58623231090077e+01, 4.58607541475697e+01, 4.58599349940755e+01, 4.58595607128296e+01, 4.58594203064870e+01, 4.58593821125078e+01, 4.58593764751547e+01, 4.58593760229693e+01, 4.58593676778540e+01, 4.58593111384619e+01, 4.58591032935570e+01, 4.58585492544371e+01, 4.58573367487375e+01, 4.58550146374989e+01, 4.58509764109533e+01, 4.58444494138983e+01, 4.58344904302762e+01, 4.58199881160976e+01, 4.57996726064765e+01, 4.57721324324355e+01, 4.57358386635340e+01, 4.56891759429215e+01, 4.56304798055668e+01, 4.55580793764422e+01, 4.54703442470956e+01, 4.53657340453305e+01, 4.52428489667735e+01, 4.51004793540789e+01, 4.49376523133729e+01, 4.47536733674615e+01, 4.45481612720986e+01, 4.43210743648934e+01, 4.40727271639067e+01, 4.38037963617879e+01, 4.35153158419081e+01, 4.32086608443824e+01, 4.28855219043040e+01, 4.25478696491268e+01, 4.21979119568617e+01, 4.18380453190578e+01, 4.14708024930777e+01, 4.10987986325015e+01, 4.07246780236097e+01, 4.03510633225660e+01, 3.99805088097520e+01, 3.96154587170323e+01, 3.92582112209912e+01, 3.89108883026317e+01, 3.85754113933698e+01, 3.82534825618247e+01, 3.79465709198722e+01, 3.76559039019690e+01, 3.73824630662024e+01, 3.71269840601369e+01, 3.68899603847568e+01, 3.66716505799567e+01, 3.64720884519014e+01, 3.62910959706538e+01, 3.61282984860186e+01, 3.59831419368457e+01, 3.58549117580373e+01, 3.57427532136292e+01, 3.56456928980807e+01, 3.55626611478388e+01, 3.54925150903211e+01, 3.54340620290641e+01, 3.53860828253850e+01, 3.53473548935435e+01, 3.53166743841409e+01, 3.52928770957776e+01, 3.52748576340062e+01, 3.52615863346834e+01, 3.52521234898668e+01, 3.52456304605282e+01, 3.52413773315594e+01, 3.52387468587167e+01, 3.52372345701333e+01, 3.52364450111087e+01, 3.52360842532207e+01, 3.52359489201830e+01, 3.52359121062872e+01, 3.52359066726323e+01, 3.52359062535366e+01, 3.52358987279220e+01, 3.52358477407742e+01, 3.52356603065969e+01, 3.52351606756791e+01, 3.52340672445318e+01, 3.52319731888455e+01, 3.52283315952075e+01, 3.52224457762733e+01, 3.52134653477661e+01, 3.52003885218007e+01, 3.51820709257212e+01, 3.51572410848743e+01, 3.51245225085078e+01, 3.50824620895791e+01, 3.50295642745184e+01, 3.49643301852671e+01, 3.48853005955581e+01, 3.47911013936234e+01, 3.46804899253533e+01, 3.45524004284854e+01, 3.44059866623016e+01, 3.42406598278661e+01, 3.40561199742603e+01, 3.38523793014921e+01, 3.36297760961844e+01, 3.33889784575642e+01, 3.31309774655819e+01, 3.28570699796329e+01, 3.25688317989112e+01, 3.22680824235204e+01, 3.19568430873409e+01, 3.16372900498592e+01, 3.13117053027968e+01, 3.09824268500340e+01, 3.06518005564594e+01, 3.03221352549053e+01, 2.99956623918955e+01, 2.96745010373338e+01, 2.93606286385436e+01, 2.90558575155383e+01, 2.87618168060074e+01, 2.84799393884623e+01, 2.82114532337054e+01, 2.79573766370760e+01, 2.77185168384617e+01, 2.74954716158098e+01, 2.72886335184652e+01, 2.70981964750081e+01, 2.69241645603079e+01, 2.67663627382134e+01, 2.66244494128421e+01, 2.64979306267627e+01, 2.63861757415476e+01, 2.62884344268132e+01, 2.62038547683164e+01, 2.61315022836833e+01, 2.60703796058511e+01, 2.60194465600445e+01, 2.59776403219559e+01, 2.59438953058317e+01, 2.59171623954169e+01, 2.58964271027679e+01, 2.58807262244593e+01, 2.58691625658136e+01, 2.58609173245949e+01, 2.58552597678104e+01, 2.58515538988340e+01, 2.58492618951801e+01, 2.58479441964004e+01, 2.58472562318418e+01, 2.58469418935445e+01, 2.58468239741368e+01, 2.58467918971742e+01, 2.58467871626809e+01, 2.58467868041993e+01, 2.58467804537342e+01, 2.58467374283957e+01, 2.58465792626496e+01, 2.58461576504549e+01, 2.58452349605188e+01, 2.58434678913949e+01, 2.58403949198417e+01, 2.58354281176578e+01, 2.58278498214163e+01, 2.58168145338886e+01, 2.58013563123355e+01, 2.57804017539547e+01, 2.57527885201698e+01, 2.57172891486761e+01, 2.56726396878298e+01, 2.56175724586930e+01, 2.55508520168589e+01, 2.54713131644345e+01, 2.53778996708836e+01, 2.52697022199191e+01, 2.51459940271861e+01, 2.50062625847529e+01, 2.48502360909536e+01, 2.46779033162759e+01, 2.44895259265247e+01, 2.42856426142328e+01, 2.40670647545995e+01, 2.38348636796947e+01, 2.35903500351128e+01, 2.33350460334285e+01, 2.30706517392011e+01, 2.27990067995989e+01, 2.25220492537272e+01, 2.22417731829116e+01, 2.19601869694174e+01, 2.16792737875432e+01, 2.14009556598388e+01, 2.11270620092338e+01, 2.08593031928861e+01, 2.05992490943209e+01, 2.03483125406984e+01, 2.01077371313678e+01, 1.98785890036537e+01, 1.96617520865150e+01, 1.94579264580455e+01, 1.92676294918629e+01, 1.90911995290641e+01, 1.89288018410424e+01, 1.87804366590464e+01, 1.86459490479350e+01, 1.85250404022167e+01, 1.84172813466499e+01, 1.83221258319453e+01, 1.82389262260294e+01, 1.81669492090252e+01, 1.81053922818084e+01, 1.80534006910751e+01, 1.80100845575032e+01, 1.79745359689112e+01, 1.79458457702397e+01, 1.79231197509054e+01, 1.79054939025970e+01, 1.78921484019696e+01, 1.78823199675247e+01, 1.78753122517496e+01, 1.78705039603866e+01, 1.78673544408351e+01, 1.78654065497437e+01, 1.78642866927483e+01, 1.78637020226909e+01, 1.78634348812489e+01, 1.78633346671315e+01, 1.78633074064486e+01, 1.78633033828284e+01, 1.78633030807739e+01, 1.78632977642733e+01, 1.78632617441992e+01, 1.78631293305935e+01, 1.78627763645137e+01, 1.78620039067665e+01, 1.78605245573331e+01, 1.78579519549035e+01, 1.78537939418792e+01, 1.78474498057864e+01, 1.78382119212296e+01, 1.78252720145029e+01, 1.78077321528686e+01, 1.77846204201795e+01, 1.77549110795170e+01, 1.77175488439585e+01, 1.76714766838176e+01, 1.76156664016323e+01, 1.75491510170832e+01, 1.74710578377418e+01, 1.73806409640771e+01, 1.72773119036653e+01, 1.71606669622788e+01, 1.70305101457544e+01, 1.68868704472218e+01, 1.67300126037891e+01, 1.65604406737738e+01, 1.63788940947049e+01, 1.61863362164871e+01, 1.59839356460588e+01, 1.57730410729631e+01, 1.55551505530761e+01, 1.53318764932519e+01, 1.51049077847711e+01, 1.48759706600702e+01, 1.46467898798296e+01, 1.44190517871261e+01, 1.41943705931736e+01, 1.39742589985110e+01, 1.37601039291342e+01, 1.35531478119328e+01, 1.33544754636286e+01, 1.31650063550046e+01, 1.29854917626381e+01, 1.28165161480269e+01, 1.26585020122513e+01, 1.25117174567381e+01, 1.23762857239779e+01, 1.22521960791603e+01, 1.21393155068393e+01, 1.20374008196604e+01, 1.19461108954915e+01, 1.18650188651200e+01, 1.17936241585492e+01, 1.17313643804678e+01, 1.16776270237412e+01, 1.16317610447862e+01, 1.15930883188954e+01, 1.15609149704803e+01, 1.15345425371512e+01, 1.15132788824229e+01, 1.14964487248890e+01, 1.14834036072503e+01, 1.14735310916970e+01, 1.14662629433074e+01, 1.14610820538491e+01, 1.14575278668490e+01, 1.14552000917083e+01, 1.14537605389906e+01, 1.14529329682702e+01, 1.14525009102282e+01, 1.14523035012710e+01, 1.14522294466682e+01, 1.14522093020502e+01, 1.14522063287479e+01, 1.14522061062199e+01, 1.14522021985171e+01, 1.14521757232718e+01, 1.14520783977138e+01, 1.14518189653843e+01, 1.14512512135923e+01, 1.14501639374544e+01, 1.14482732692501e+01, 1.14452177532505e+01, 1.14405565005488e+01, 1.14337707024827e+01, 1.14242687107807e+01, 1.14113948031422e+01, 1.13944416408751e+01, 1.13726662885544e+01, 1.13453095053281e+01, 1.13116178380679e+01, 1.12708678567760e+01, 1.12223916854211e+01, 1.11656028131532e+01, 1.11000210403344e+01, 1.10252953400554e+01, 1.09412234156971e+01, 1.08477668207918e+01, 1.07450606837662e+01, 1.06334173426244e+01, 1.05133235284653e+01, 1.03854311171747e+01, 1.02505418630182e+01, 1.01095868990795e+01, 9.96360210082486e+00, 9.81370062942901e+00, 9.66104408016035e+00, 9.50681365111194e+00, 9.35218262658320e+00, 9.19829125834289e+00, 9.04622485698356e+00, 8.89699560877482e+00, 8.75152834349374e+00, 8.61065022224043e+00, 8.47508410895577e+00, 8.34544524335818e+00, 8.22224074589194e+00, 8.10587145033015e+00, 7.99663556659485e+00, 7.89473371366333e+00, 7.80027491926505e+00, 7.71328325023295e+00, 7.63370480765907e+00, 7.56141488912906e+00, 7.49622518256112e+00, 7.43789091003582e+00, 7.38611788383447e+00, 7.34056946972401e+00, 7.30087347388994e+00, 7.26662897986792e+00, 7.23741316089071e+00, 7.21278808227214e+00, 7.19230748933252e+00, 7.17552355095508e+00, 7.16199349962324e+00, 7.15128607853538e+00, 7.14298767812596e+00, 7.13670802101735e+00, 7.13208523879307e+00, 7.12879017820650e+00, 7.12652977993584e+00, 7.12504939024084e+00, 7.12413389430562e+00, 7.12360759808968e+00, 7.12333283071736e+00, 7.12320728876144e+00, 7.12316019388643e+00, 7.12314738296184e+00, 7.12314549209707e+00, 7.12314535052794e+00, 7.12314286379552e+00, 7.12312601582984e+00, 7.12306408111034e+00, 7.12289898730549e+00, 7.12253769074672e+00, 7.12184579218469e+00, 7.12064266105179e+00, 7.11869831289172e+00, 7.11573225656989e+00, 7.11141449301827e+00, 7.10536880256862e+00, 7.09717840017891e+00, 7.08639396436071e+00, 7.07254395494992e+00, 7.05514702766951e+00, 7.03372623307685e+00, 7.00782456045049e+00, 6.97702126302273e+00, 6.94094829180567e+00, 6.89930608455370e+00, 6.85187791729714e+00, 6.79854203915119e+00, 6.73928088311558e+00, 6.67418677660211e+00, 6.60346375874592e+00, 6.52742533365876e+00, 6.44648823071753e+00, 6.36116248271484e+00, 6.27203834805562e+00, 6.17977077481032e+00, 6.08506221824365e+00, 5.98864467209652e+00, 5.89126175725990e+00, 5.79365163598017e+00, 5.69653139707876e+00, 5.60058340303947e+00, 5.50644391993455e+00, 5.41469418234578e+00, 5.32585389187972e+00, 5.24037702027689e+00, 5.15864969288177e+00, 5.08098986735086e+00, 5.00764849400183e+00, 4.93881184334370e+00, 4.87460470665332e+00, 4.81509421014934e+00, 4.76029402605617e+00, 4.71016880941633e+00, 4.66463873394351e+00, 4.62358404074739e+00, 4.58684954862802e+00, 4.55424910278518e+00, 4.52556995969088e+00, 4.50057711937099e+00, 4.47901762257044e+00, 4.46062482963283e+00, 4.44512269108705e+00, 4.43223000786347e+00, 4.42166466302639e+00, 4.41314778844411e+00, 4.40640781068733e+00, 4.40118430253793e+00, 4.39723155168241e+00, 4.39432174818130e+00, 4.39224768853070e+00, 4.39082489747625e+00, 4.38989307950628e+00, 4.38931682979144e+00, 4.38898555826107e+00, 4.38881260897068e+00, 4.38873358799397e+00, 4.38870394465644e+00, 4.38869588096489e+00, 4.38869469078146e+00, 4.38869460143599e+00, 4.38869302888084e+00, 4.38868237459648e+00, 4.38864320844063e+00, 4.38853880677409e+00, 4.38831033097510e+00, 4.38787279047621e+00, 4.38711196016777e+00, 4.38588240723874e+00, 4.38400676492516e+00, 4.38127636871148e+00, 4.37745334233001e+00, 4.37227418436869e+00, 4.36545485969010e+00, 4.35669734226563e+00, 4.34569748784622e+00, 4.33215403827410e+00, 4.31577847834265e+00, 4.29630538706796e+00, 4.27350285579965e+00, 4.24718249439901e+00, 4.21720852214921e+00, 4.18350544900477e+00, 4.14606389922723e+00, 4.10494421343064e+00, 4.06027758208768e+00, 4.01226460475115e+00, 3.96117132229259e+00, 3.90732292013998e+00, 3.85109543471738e+00, 3.79290590105522e+00, 3.73320144855559e+00, 3.67244788043767e+00, 3.61111826137925e+00, 3.54968199218580e+00, 3.48859477761940e+00, 3.42828980289278e+00, 3.36917033504943e+00, 3.31160386600750e+00, 3.25591782154780e+00, 3.20239678037632e+00, 3.15128108319365e+00, 3.10276666531145e+00, 3.05700591803507e+00, 3.01410937267957e+00, 2.97414800452383e+00, 2.93715596932513e+00, 2.90313360891818e+00, 2.87205059156787e+00, 2.84384908402671e+00, 2.81844688299132e+00, 2.79574046173022e+00, 2.77560791152264e+00, 2.75791177621591e+00, 2.74250179118106e+00, 2.72921754514709e+00, 2.71789108508909e+00, 2.70834948109105e+00, 2.70041736068622e+00, 2.69391941159391e+00, 2.68868283917168e+00, 2.68453975156227e+00, 2.68132943276207e+00, 2.67890045297536e+00, 2.67711255782633e+00, 2.67583827423429e+00, 2.67496417164318e+00, 2.67439172307006e+00, 2.67403772088940e+00, 2.67383421677322e+00, 2.67372797276138e+00, 2.67367942978875e+00, 2.67366121977226e+00, 2.67365626621836e+00, 2.67365553508462e+00, 2.67365547996563e+00, 2.67365450671170e+00, 2.67364791277934e+00, 2.67362367288153e+00, 2.67355905894365e+00, 2.67341765662501e+00, 2.67314686846208e+00, 2.67267600998956e+00, 2.67191509676304e+00, 2.67075441007489e+00, 2.66906491625409e+00, 2.66669959739544e+00, 2.66349572871185e+00, 2.65927810820595e+00, 2.65386320731943e+00, 2.64706416684018e+00, 2.63869651200353e+00, 2.62858440713313e+00, 2.61656721740123e+00, 2.60250609852234e+00, 2.58629030024936e+00, 2.56784285215751e+00, 2.54712530519772e+00, 2.52414123287764e+00, 2.49893825212263e+00, 2.47160840329796e+00, 2.44228682591184e+00, 2.41114877295718e+00, 2.37840511282178e+00, 2.34429656292360e+00, 2.30908697440530e+00, 2.27305603528990e+00, 2.23649177659048e+00, 2.19968325175842e+00, 2.16291371787169e+00, 2.12645458334175e+00, 2.09056030977418e+00, 2.05546437374534e+00, 2.02137631588944e+00, 1.98847983655210e+00, 1.95693184390885e+00, 1.92686232405943e+00, 1.89837488315865e+00, 1.87154780733657e+00, 1.84643549403788e+00, 1.82307012497661e+00, 1.80146347267082e+00, 1.78160875640985e+00, 1.76348248706486e+00, 1.74704626163293e+00, 1.73224848672817e+00, 1.71902602485930e+00, 1.70730576813358e+00, 1.69700615113553e+00, 1.68803861843625e+00, 1.68030906286396e+00, 1.67371924872160e+00, 1.66816823000146e+00, 1.66355376779059e+00, 1.65977374398909e+00, 1.65672756073657e+00, 1.65431750716992e+00, 1.65245006796253e+00, 1.65103714216110e+00, 1.64999713673825e+00, 1.64925589750322e+00, 1.64874744088631e+00, 1.64841445375015e+00, 1.64820853466590e+00, 1.64809015866281e+00, 1.64802835775798e+00, 1.64800012087920e+00, 1.64798952832636e+00, 1.64798664690225e+00, 1.64798622161037e+00, 1.64798618937382e+00, 1.64798561784824e+00, 1.64798174568209e+00, 1.64796751124712e+00, 1.64792956791068e+00, 1.64784653208457e+00, 1.64768751711767e+00, 1.64741101558218e+00, 1.64696418759908e+00, 1.64628260693438e+00, 1.64529050901983e+00, 1.64390157496139e+00, 1.64202027209387e+00, 1.63954375394538e+00, 1.63636430016045e+00, 1.63237225012645e+00, 1.62745935363954e+00, 1.62152242969466e+00, 1.61446719301506e+00, 1.60621208059864e+00, 1.59669189106690e+00, 1.58586104155550e+00, 1.57369625315495e+00, 1.56019849802905e+00, 1.54539407900999e+00, 1.52933476332805e+00, 1.51209695182888e+00, 1.49377992772759e+00, 1.47450328809778e+00, 1.45440371065467e+00, 1.43363124299253e+00, 1.41234531836639e+00, 1.39071070089955e+00, 1.36889354567153e+00, 1.34705772929790e+00, 1.32536156924911e+00, 1.30395501033679e+00, 1.28297731895061e+00, 1.26255529297022e+00, 1.24280196957231e+00, 1.22381579480026e+00, 1.20568020712697e+00, 1.18846358108098e+00, 1.17221947491730e+00, 1.15698712704950e+00, 1.14279214859478e+00, 1.12964736334993e+00, 1.11755375150229e+00, 1.10650145920820e+00, 1.09647084266724e+00, 1.08743352223135e+00, 1.07935342906295e+00, 1.07218783346599e+00, 1.06588834981353e+00, 1.06040191758660e+00, 1.05567176111735e+00, 1.05163833203433e+00, 1.04824023812614e+00, 1.04541516050874e+00, 1.04310075787727e+00, 1.04123555262601e+00, 1.03975978919325e+00, 1.03861625063159e+00, 1.03775101562093e+00, 1.03711413538565e+00, 1.03666020861984e+00, 1.03634883280832e+00, 1.03614491235329e+00, 1.03601880760743e+00, 1.03594631405205e+00, 1.03590846708725e+00, 1.03589117477144e+00, 1.03588468787127e+00, 1.03588292328133e+00, 1.03588266283171e+00, 1.03588264296474e+00, 1.03588228908775e+00, 1.03587989152157e+00, 1.03587107785005e+00, 1.03584758411012e+00, 1.03579617001706e+00, 1.03569771117455e+00, 1.03552650699596e+00, 1.03524984003110e+00, 1.03482781874702e+00, 1.03421353116235e+00, 1.03335353057921e+00, 1.03218866646516e+00, 1.03065526285839e+00, 1.02868663325691e+00, 1.02621490488110e+00, 1.02317310696878e+00, 1.01949745835370e+00, 1.01512977044759e+00, 1.01001986472954e+00, 1.00412789098624e+00, 9.97426425833709e-01, 9.89902232126340e-01, 9.81557569727656e-01, 9.72410966923732e-01, 9.62497388691560e-01, 9.51867771360469e-01, 9.40587930467615e-01, 9.28736886890604e-01, 9.16404692569770e-01, 9.03689868350185e-01, 8.90696590025314e-01, 8.77531772390668e-01, 8.64302203504707e-01, 8.51111871787296e-01, 8.38059607603384e-01, 8.25237130474918e-01, 8.12727556291591e-01, 8.00604380118315e-01, 7.88930914028498e-01, 7.77760129919002e-01, 7.67134837257913e-01, 7.57088116192179e-01, 7.47643926644088e-01, 7.38817821863336e-01, 7.30617707604537e-01, 7.23044602892941e-01, 7.16093372931653e-01, 7.09753417534372e-01, 7.04009308730079e-01, 6.98841378672171e-01, 6.94226263897932e-01, 6.90137414726138e-01, 6.86545579614304e-01, 6.83419274040460e-01, 6.80725242279053e-01, 6.78428918561295e-01, 6.76494891752583e-01, 6.74887374993844e-01, 6.73570678888936e-01, 6.72509683917326e-01, 6.71670304975649e-01, 6.71019938475396e-01, 6.70527880432331e-01, 6.70165702653065e-01, 6.69907573601662e-01, 6.69730510919436e-01, 6.69614553907252e-01, 6.69542846523015e-01, 6.69501624472801e-01, 6.69480103594481e-01, 6.69470270697943e-01, 6.69466582066576e-01, 6.69465578672018e-01, 6.69465430573155e-01, 6.69465419196527e-01, 6.69465215507259e-01, 6.69463835483924e-01, 6.69458762394726e-01, 6.69445239574316e-01, 6.69415646089590e-01, 6.69358974386057e-01, 6.69260432329949e-01, 6.69101190379834e-01, 6.68858292580055e-01, 6.68504747410594e-01, 6.68009810924431e-01, 6.67339469702029e-01, 6.66457124644481e-01, 6.65324468344721e-01, 6.63902538766365e-01, 6.62152920581093e-01, 6.60039053473056e-01, 6.57527595076459e-01, 6.54589776318105e-01, 6.51202680273926e-01, 6.47350373624050e-01, 6.43024823456003e-01, 6.38226541971108e-01, 6.32964917221926e-01, 6.27258208105076e-01, 6.21133204417575e-01, 6.14624575332172e-01, 6.07773949548942e-01, 6.00628785432932e-01, 5.93241098243240e-01, 5.85666113715519e-01, 5.77960913405946e-01, 5.70183128745128e-01, 5.62389729466375e-01, 5.54635939740891e-01, 5.46974303406486e-01, 5.39453909063440e-01, 5.32119776993341e-01, 5.25012402916133e-01, 5.18167448405066e-01, 5.11615564114948e-01, 5.05382329638593e-01, 4.99488292622630e-01, 4.93949089610140e-01, 4.88775631800564e-01, 4.83974340376605e-01, 4.79547418058758e-01, 4.75493145908061e-01, 4.71806196887698e-01, 4.68477960114868e-01, 4.65496871908755e-01, 4.62848751537493e-01, 4.60517140894339e-01, 4.58483648146128e-01, 4.56728295688151e-01, 4.55229872537089e-01, 4.53966290652545e-01, 4.52914943681904e-01, 4.52053065373136e-01, 4.51358083516993e-01, 4.50807963894249e-01, 4.50381537454049e-01, 4.50058802967590e-01, 4.49821196811911e-01, 4.49651821431672e-01, 4.49535624468240e-01, 4.49459521546942e-01, 4.49412457246638e-01, 4.49385400759979e-01, 4.49371275071286e-01, 4.49364820981259e-01, 4.49362399837980e-01, 4.49361741229126e-01, 4.49361644019859e-01, 4.49361636495468e-01, 4.49361501035469e-01, 4.49360583274768e-01, 4.49357209498402e-01, 4.49348216319009e-01, 4.49328535319381e-01, 4.49290845205212e-01, 4.49225306018675e-01, 4.49119388645613e-01, 4.48957811281488e-01, 4.48722592585327e-01, 4.48393228783544e-01, 4.47946998820925e-01, 4.47359397648469e-01, 4.46604692815669e-01, 4.45656593733332e-01, 4.44489016480603e-01, 4.43076920201207e-01, 4.41397184497006e-01, 4.39429491443164e-01, 4.37157171655443e-01, 4.34567971946535e-01, 4.31654703090203e-01, 4.28415730392572e-01, 4.24855277132423e-01, 4.20983521074049e-01, 4.16816476403937e-01, 4.12375666528760e-01, 4.07687605966018e-01, 4.02783120830733e-01, 3.97696546103120e-01, 3.92464843197729e-01, 3.87126682984032e-01, 3.81721537409490e-01, 3.76288817728295e-01, 3.70867089842633e-01, 3.65493388411011e-01, 3.60202642212357e-01, 3.55027214720721e-01, 3.49996556672309e-01, 3.45136962035255e-01, 3.40471415362586e-01, 3.36019516900748e-01, 3.31797471741197e-01, 3.27818130346870e-01, 3.24091069546512e-01, 3.20622705192150e-01, 3.17416429805732e-01, 3.14472770484301e-01, 3.11789563950342e-01, 3.09362146871351e-01, 3.07183560422459e-01, 3.05244768557401e-01, 3.03534889624487e-01, 3.02041440855248e-01, 3.00750594909815e-01, 2.99647447119099e-01, 2.98716291368763e-01, 2.97940901764491e-01, 2.97304816360622e-01, 2.96791618382457e-01, 2.96385209593524e-01, 2.96070069826897e-01, 2.95831496280314e-01, 2.95655816034682e-01, 2.95530565442192e-01, 2.95444630570728e-01, 2.95388343790393e-01, 2.95353532815582e-01, 2.95333520021927e-01, 2.95323071557434e-01, 2.95318297571683e-01, 2.95316506685452e-01, 2.95316019521039e-01, 2.95315947616587e-01, 2.95315942005757e-01, 2.95315840412485e-01, 2.95315152103299e-01, 2.95312621808918e-01, 2.95305876996764e-01, 2.95291116260117e-01, 2.95262848142514e-01, 2.95213691084115e-01, 2.95134244059270e-01, 2.95013036159901e-01, 2.94836561803464e-01, 2.94589406318238e-01, 2.94254464285970e-01, 2.93813250126803e-01, 2.93246296958856e-01, 2.92533635789496e-01, 2.91655342732341e-01, 2.90592137428519e-01, 2.89326011535621e-01, 2.87840862463327e-01, 2.86123104959017e-01, 2.84162232126221e-01, 2.81951298344282e-01, 2.79487299510279e-01, 2.76771430974762e-01, 2.73809210151654e-01, 2.70610458477435e-01, 2.67189145434784e-01, 2.63563104940332e-01, 2.59753640794528e-01, 2.55785042552915e-01, 2.51684035811081e-01, 2.47479191512454e-01, 2.43200317756745e-01, 2.38877855167544e-01, 2.34542293698638e-01, 2.30223625298976e-01, 2.25950843472518e-01, 2.21751497676919e-01, 2.17651307791215e-01, 2.13673841571108e-01, 2.09840256069171e-01, 2.06169102375185e-01, 2.02676191657957e-01, 1.99374519308989e-01, 1.96274242983463e-01, 1.93382709537943e-01, 1.90704525333146e-01, 1.88241664152784e-01, 1.85993607085365e-01, 1.83957509076350e-01, 1.82128387403097e-01, 1.80499327955417e-01, 1.79061705835650e-01, 1.77805417339343e-01, 1.76719120778614e-01, 1.75790483814765e-01, 1.75006434953076e-01, 1.74353416624072e-01, 1.73817636869174e-01, 1.73385316122491e-01, 1.73042925015395e-01, 1.72777408610003e-01, 1.72576392081780e-01, 1.72428362693523e-01, 1.72322822992783e-01, 1.72250410559087e-01, 1.72202980331065e-01, 1.72173646537615e-01, 1.72156782494525e-01, 1.72147977937058e-01, 1.72143955060420e-01, 1.72142445940224e-01, 1.72142035422908e-01, 1.72141974831433e-01, 1.72141970063151e-01, 1.72141883209960e-01, 1.72141294766931e-01, 1.72139131590370e-01, 1.72133365373270e-01, 1.72120746236487e-01, 1.72096579418394e-01, 1.72054554178402e-01, 1.71986633064179e-01, 1.71883008561508e-01, 1.71732132803612e-01, 1.71520824396197e-01, 1.71234454446505e-01, 1.70857211519532e-01, 1.70372442457193e-01, 1.69763062819328e-01, 1.69012027233123e-01, 1.68102846333340e-01, 1.67020133477806e-01, 1.65750161317092e-01, 1.64281405902642e-01, 1.62605054639400e-01, 1.60715454283334e-01, 1.58610476509153e-01, 1.56291781366775e-01, 1.53764963107989e-01, 1.51039568167093e-01, 1.48128981190396e-01, 1.45050181527381e-01, 1.41823379084457e-01, 1.38471544459932e-01, 1.35019853420843e-01, 1.31495069714234e-01, 1.27924892696299e-01, 1.24337297202658e-01, 1.20759892460285e-01, 1.17219324715762e-01, 1.13740744745641e-01, 1.10347356724887e-01, 1.07060059410699e-01, 1.03897184737846e-01, 1.00874333285617e-01, 9.80043011629022e-02, 9.52970889786400e-02, 9.27599808314028e-02, 9.03976796519004e-02, 8.82124847044992e-02, 8.62044975340126e-02, 8.43718440048207e-02, 8.27109021000325e-02, 8.12165275241341e-02, 7.98822715737515e-02, 7.87005879620324e-02, 7.76630271625152e-02, 7.67604183116362e-02, 7.59830397540497e-02, 7.53207799236719e-02, 7.47632904339071e-02, 7.43001330383208e-02, 7.39209215810206e-02, 7.36154592777609e-02, 7.33738707581750e-02, 7.31867273730558e-02, 7.30451634262863e-02, 7.29409803289798e-02, 7.28667352651939e-02, 7.28158108504005e-02, 7.27824624848523e-02, 7.27618406355219e-02, 7.27499860878993e-02, 7.27437972265363e-02, 7.27409695487893e-02, 7.27399087996190e-02, 7.27396202511130e-02, 7.27395776619773e-02, 7.27395742807460e-02, 7.27395123151240e-02, 7.27390924894564e-02, 7.27375491753719e-02, 7.27334353398419e-02, 7.27244326826479e-02, 7.27071929398591e-02, 7.26772173352828e-02, 7.26287807466264e-02, 7.25549066404767e-02, 7.24473981276788e-02, 7.22969294898803e-02, 7.20932011023905e-02, 7.18251587313716e-02, 7.14812756465477e-02, 7.10498928733188e-02, 7.05196092940099e-02, 6.98797093875214e-02, 6.91206124533608e-02, 6.82343235660170e-02, 6.72148636745362e-02, 6.60586546193400e-02, 6.47648347678569e-02, 6.33354827399586e-02, 6.17757304240456e-02, 6.00937521088269e-02, 5.83006237973365e-02, 5.64100551274720e-02, 5.44380050695565e-02, 5.24022008167009e-02, 5.03215860981356e-02, 4.82157297751229e-02, 4.61042276194377e-02, 4.40061296074958e-02, 4.19394221064948e-02, 3.99205892465204e-02, 3.79642708428491e-02, 3.60830259641887e-02, 3.42872026258618e-02, 3.25849064593484e-02, 3.09820557041860e-02, 2.94825068020828e-02, 2.80882336659971e-02, 2.67995434815589e-02, 2.56153122888492e-02, 2.45332248519645e-02, 2.35500058417131e-02, 2.26616329537990e-02, 2.18635264426723e-02, 2.11507127450687e-02, 2.05179619731553e-02, 1.99599003166213e-02, 1.94710992794853e-02, 1.90461444979775e-02, 1.86796875901649e-02, 1.83664848826339e-02, 1.81014268297162e-02, 1.78795615436655e-02, 1.76961152380670e-02, 1.75465116913074e-02, 1.74263921264384e-02, 1.73316362106419e-02, 1.72583842054763e-02, 1.72030596764472e-02, 1.71623916435095e-02, 1.71334346559604e-02, 1.71135850450241e-02, 1.71005915656224e-02, 1.70925587762127e-02, 1.70879418146512e-02, 1.70855316620901e-02, 1.70844305160642e-02, 1.70840174498181e-02, 1.70839050869347e-02, 1.70838885024409e-02, 1.70838871737544e-02, 1.70838626712282e-02, 1.70836966639882e-02, 1.70830864134359e-02, 1.70814597840092e-02, 1.70779003238550e-02, 1.70710849821631e-02, 1.70592375829415e-02, 1.70401012148886e-02, 1.70109327029735e-02, 1.69685222279696e-02, 1.69092408398228e-02, 1.68291179640931e-02, 1.67239499288314e-02, 1.65894389755010e-02, 1.64213601613739e-02, 1.62157510825267e-02, 1.59691166285825e-02, 1.56786382698613e-02, 1.53423750228541e-02, 1.49594416296281e-02, 1.45301490075980e-02, 1.40560930345544e-02, 1.35401804439113e-02, 1.29865849865780e-02, 1.24006326983306e-02, 1.17886213298988e-02, 1.11575847931391e-02, 1.05150180104943e-02, 9.86858043556538e-03, 9.22579797727963e-03, 8.59378348317819e-03, 7.97899510780553e-03, 7.38704872226492e-03, 6.82259411728126e-03, 6.28925599542640e-03, 5.78963240648989e-03, 5.32533886077147e-03, 4.89708748209239e-03, 4.50479416372218e-03, 4.14770728914959e-03, 3.82454821378531e-03, 3.53365035367403e-03, 3.27308414080107e-03, 3.04076057356895e-03, 2.83451392590450e-03, 2.65216860141493e-03, 2.49159383091069e-03, 2.35074650885281e-03, 2.22770095477642e-03, 2.12066540547085e-03, 2.02798714148855e-03, 1.94814966683055e-03, 1.87976546274626e-03, 1.82156687452562e-03, 1.77239649721628e-03, 1.73119758051143e-03, 1.69700465031042e-03, 1.66893462079543e-03, 1.64617882582263e-03, 1.62799638782474e-03, 1.61370915678560e-03, 1.60269821513737e-03, 1.59440175711723e-03, 1.58831403440729e-03, 1.58398499385726e-03, 1.58102020374429e-03, 1.57908065786216e-03, 1.57788207096132e-03, 1.57719332959516e-03, 1.57683383942242e-03, 1.57666960707559e-03, 1.57660800135773e-03, 1.57659124342047e-03, 1.57658877000699e-03, 1.57658856998921e-03, 1.57658485824186e-03, 1.57655971082704e-03, 1.57646726910343e-03, 1.57622087515108e-03, 1.57568175823997e-03, 1.57464970725130e-03, 1.57285628331749e-03, 1.56996117941333e-03, 1.56555238261546e-03, 1.55915081781188e-03, 1.55022011655603e-03, 1.53818202845457e-03, 1.52243773860050e-03, 1.50239494607263e-03, 1.47749998487257e-03, 1.44727355296936e-03, 1.41134784124120e-03, 1.36950216608511e-03, 1.32169378727554e-03, 1.26808057447647e-03, 1.20903265221273e-03, 1.14513116305659e-03, 1.07715373378880e-03, 1.00604743282418e-03, 9.32889967927508e-04, 8.58838696927933e-04, 7.85067337744749e-04, 7.12696753495834e-04, 6.42734898605113e-04, 5.76032223861489e-04, 5.13251609646659e-04, 4.54870633532778e-04, 4.01230591009494e-04, 3.52714572585455e-04, 3.09230788786863e-04, 2.70393449596895e-04, 2.35999143001209e-04, 2.06139886543610e-04, 1.80121874024053e-04, 1.57457735980415e-04, 1.37776059943917e-04, 1.20743625146247e-04, 1.06063569944581e-04, 9.34164187071929e-05, 8.25296490916116e-05, 7.31543517957789e-05, 6.49376085135600e-05, 5.76539381569319e-05, 5.11389479811442e-05, 4.52707123770324e-05, 3.99586980819322e-05, 3.51355316468016e-05, 3.08235107411292e-05, 2.69134165904461e-05, 2.33251104901910e-05, 2.00438148709736e-05, 1.70575025890685e-05, 1.43575840922949e-05, 1.19384466046313e-05, 9.79384209642831e-06, 7.91204421091152e-06, 6.27838248590251e-06, 4.87895054845922e-06, 3.69976422120999e-06, 2.72552087210443e-06, 1.93919296975764e-06, 1.32210908532215e-06, 8.54258800586649e-07, 5.14696921527502e-07, 2.81990750750083e-07, -1.42108547152020e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14],
[ 7.00000000000000e+01, 6.99999999247270e+01, 6.99999975971382e+01, 6.99999818273552e+01, 6.99999238555678e+01, 6.99997693210190e+01, 6.99994311104721e+01, 6.99987833375493e+01, 6.99976566695544e+01, 6.99958351766907e+01, 6.99930548314793e+01, 6.99890037356130e+01, 6.99833240996919e+01, 6.99756159507167e+01, 6.99654424952792e+01, 6.99523370251791e+01, 6.99358112181745e+01, 6.99153646602256e+01, 6.98904953961778e+01, 6.98607113013296e+01, 6.98255420536518e+01, 6.97845514719211e+01, 6.97373499653990e+01, 6.96836068139116e+01, 6.96230619636187e+01, 6.95555369866978e+01, 6.94809448189847e+01, 6.93992978670139e+01, 6.93107140745018e+01, 6.92154205666286e+01, 6.91137545538319e+01, 6.90061612756829e+01, 6.88931888945873e+01, 6.87754803980064e+01, 6.86537627223326e+01, 6.85288334557017e+01, 6.84015455963038e+01, 6.82727909261545e+01, 6.81434826019201e+01, 6.80145375640082e+01, 6.78868593275637e+01, 6.77613216527719e+01, 6.76387535073469e+01, 6.75199256417294e+01, 6.74055390063545e+01, 6.72962151570641e+01, 6.71924887233255e+01, 6.70948019558565e+01, 6.70035013250230e+01, 6.69188361070917e+01, 6.68409588695465e+01, 6.67699277465450e+01, 6.67057103788324e+01, 6.66481893771769e+01, 6.65971691534133e+01, 6.65523839478405e+01, 6.65135068659394e+01, 6.64801597215272e+01, 6.64519234682607e+01, 6.64283489878236e+01, 6.64089679922892e+01, 6.63933037912306e+01, 6.63808816722815e+01, 6.63712386480485e+01, 6.63639323333078e+01, 6.63585487347715e+01, 6.63547087614718e+01, 6.63520732966676e+01, 6.63503467113980e+01, 6.63492787442276e+01, 6.63486647198738e+01, 6.63483441295278e+01, 6.63481976458863e+01, 6.63481426942870e+01, 6.63481277460698e+01, 6.63481255397424e+01, 6.63481252503188e+01, 6.63481185070958e+01, 6.63480728207499e+01, 6.63479048726458e+01, 6.63474571826365e+01, 6.63464774137213e+01, 6.63446009996526e+01, 6.63413377766375e+01, 6.63360632100997e+01, 6.63280147036525e+01, 6.63162933593429e+01, 6.62998714246487e+01, 6.62776055114234e+01, 6.62482555042372e+01, 6.62105088906041e+01, 6.61630100450246e+01, 6.61043937859402e+01, 6.60333223051051e+01, 6.59485243509486e+01, 6.58488353436420e+01, 6.57332369273393e+01, 6.56008943475692e+01, 6.54511900064598e+01, 6.52837516229753e+01, 6.50984736300607e+01, 6.48955307795762e+01, 6.46753833784826e+01, 6.44387740972646e+01, 6.41867168026384e+01, 6.39204782916868e+01, 6.36415540775158e+01, 6.33516394658016e+01, 6.30525970835637e+01, 6.27464218388072e+01, 6.24352040928889e+01, 6.21210917042214e+01, 6.18062516038686e+01, 6.14928316839325e+01, 6.11829239528250e+01, 6.08785300385920e+01, 6.05815301128928e+01, 6.02936561253007e+01, 6.00164699107746e+01, 5.97513463473058e+01, 5.94994613954603e+01, 5.92617846193269e+01, 5.90390756921203e+01, 5.88318844063288e+01, 5.86405537901622e+01, 5.84652260306634e+01, 5.83058509850911e+01, 5.81621971103746e+01, 5.80338646543995e+01, 5.79203009393852e+01, 5.78208175357442e+01, 5.77346090826227e+01, 5.76607734647369e+01, 5.75983330081097e+01, 5.75462563126530e+01, 5.75034802995515e+01, 5.74689320184249e+01, 5.74415497358519e+01, 5.74203028157311e+01, 5.74042099057462e+01, 5.73923549650748e+01, 5.73839007078406e+01, 5.73780990950162e+01, 5.73742985836354e+01, 5.73719479340917e+01, 5.73705964805377e+01, 5.73698908815687e+01, 5.73695684833658e+01, 5.73694475399889e+01, 5.73694146403937e+01, 5.73694097844801e+01, 5.73694093557953e+01, 5.73694009559287e+01, 5.73693440455840e+01, 5.73691348369535e+01, 5.73685771621485e+01, 5.73673566973365e+01, 5.73650193338414e+01, 5.73609545534043e+01, 5.73543845572502e+01, 5.73443597739846e+01, 5.73297612273483e+01, 5.73093100780950e+01, 5.72815844611204e+01, 5.72450435168948e+01, 5.71980582651355e+01, 5.71389486922972e+01, 5.70660261318632e+01, 5.69776397225120e+01, 5.68722254547339e+01, 5.67483560866228e+01, 5.66047900512735e+01, 5.64405174159063e+01, 5.62548010032535e+01, 5.60472109528990e+01, 5.58176512723459e+01, 5.55663772776385e+01, 5.52940032150316e+01, 5.50014997526921e+01, 5.46901814101295e+01, 5.43616843457086e+01, 5.40179352566573e+01, 5.36611124724120e+01, 5.32936006408425e+01, 5.29179406969654e+01, 5.25367770249060e+01, 5.21528038292407e+01, 5.17687126865329e+01, 5.13871430442226e+01, 5.10106370973188e+01, 5.06416000554748e+01, 5.02822663780256e+01, 4.99346721617025e+01, 4.96006335563792e+01, 4.92817308760212e+01, 4.89792979605080e+01, 4.86944163095117e+01, 4.84279135259975e+01, 4.81803656491937e+01, 4.79521030059525e+01, 4.77432192535311e+01, 4.75535833205738e+01, 4.73828539752623e+01, 4.72304967609540e+01, 4.70958030411918e+01, 4.69779108883925e+01, 4.68758275340278e+01, 4.67884530729357e+01, 4.67146050813143e+01, 4.66530437686863e+01, 4.66024972415593e+01, 4.65616864146270e+01, 4.65293490689470e+01, 4.65042625307699e+01, 4.64852644345982e+01, 4.64712710439388e+01, 4.64612926361732e+01, 4.64544455154656e+01, 4.64499602992283e+01, 4.64471862269901e+01, 4.64455913613499e+01, 4.64447586833737e+01, 4.64443782226451e+01, 4.64442354981596e+01, 4.64441966735888e+01, 4.64441909431617e+01, 4.64441904835105e+01, 4.64441820006154e+01, 4.64441245277449e+01, 4.64439132512748e+01, 4.64433500648643e+01, 4.64421175405554e+01, 4.64397570913322e+01, 4.64356521948575e+01, 4.64290174418736e+01, 4.64188940495873e+01, 4.64041523376555e+01, 4.63835014983866e+01, 4.63555067995799e+01, 4.63186141352034e+01, 4.62711815855263e+01, 4.62115173678503e+01, 4.61379232600637e+01, 4.60487422758067e+01, 4.59424090813874e+01, 4.58175013941653e+01, 4.56727904153442e+01, 4.55072882512145e+01, 4.53202902852024e+01, 4.51114105895549e+01, 4.48806087100205e+01, 4.46282065076569e+01, 4.43548941769268e+01, 4.40617250502265e+01, 4.37500993160261e+01, 4.34217372931971e+01, 4.30786433931102e+01, 4.27230623389507e+01, 4.23574295688544e+01, 4.19843179894511e+01, 4.16063833311224e+01, 4.12263102598166e+01, 4.08467611257012e+01, 4.04703288179643e+01, 4.00994947215919e+01, 3.97365923197605e+01, 3.93837766206457e+01, 3.90429993393422e+01, 3.87159896254039e+01, 3.84042400610423e+01, 3.81089976261119e+01, 3.78312593056844e+01, 3.75717719921167e+01, 3.73310363071955e+01, 3.71093139488221e+01, 3.69066381584285e+01, 3.67228269134474e+01, 3.65574984721806e+01, 3.64100889307673e+01, 3.62798714861102e+01, 3.61659771270722e+01, 3.60674164930811e+01, 3.59831026409395e+01, 3.59118744462881e+01, 3.58525203374625e+01, 3.58038020201948e+01, 3.57644778070588e+01, 3.57333251220475e+01, 3.57091617148238e+01, 3.56908650973281e+01, 3.56773897129997e+01, 3.56677813698588e+01, 3.56611885152428e+01, 3.56568700020971e+01, 3.56541990922183e+01, 3.56526635565256e+01, 3.56518618604644e+01, 3.56514955570754e+01, 3.56513581437308e+01, 3.56513207639412e+01, 3.56513152467615e+01, 3.56513148212236e+01, 3.56513071799272e+01, 3.56512554090186e+01, 3.56510650936575e+01, 3.56505577825910e+01, 3.56494475437759e+01, 3.56473212999693e+01, 3.56436237327177e+01, 3.56376474503707e+01, 3.56285290079523e+01, 3.56152512410777e+01, 3.55966522286484e+01, 3.55714410256571e+01, 3.55382201049875e+01, 3.54955142148601e+01, 3.54418050993953e+01, 3.53755712510782e+01, 3.52953315782595e+01, 3.51996915956842e+01, 3.50873905028356e+01, 3.49573473271858e+01, 3.48087042004681e+01, 3.46408648257978e+01, 3.44535262957802e+01, 3.42467026422334e+01, 3.40207388326678e+01, 3.37763143629576e+01, 3.35144361059152e+01, 3.32364206295627e+01, 3.29438667578844e+01, 3.26386196673049e+01, 3.23227282493328e+01, 3.19983977822572e+01, 3.16679401102666e+01, 3.13337235103731e+01, 3.09981242404249e+01, 3.06634814322904e+01, 3.03320565696272e+01, 3.00059983278915e+01, 2.96873131145320e+01, 2.93778412784111e+01, 2.90792386895349e+01, 2.87929632314444e+01, 2.85202656877941e+01, 2.82621845171436e+01, 2.80195440661364e+01, 2.77929558437395e+01, 2.75828225480511e+01, 2.73893445913699e+01, 2.72125289057268e+01, 2.70521998321132e+01, 2.69080119065463e+01, 2.67794643588917e+01, 2.66659171383670e+01, 2.65666082733307e+01, 2.64806723614611e+01, 2.64071599685025e+01, 2.63450576886666e+01, 2.62933085879121e+01, 2.62508327145252e+01, 2.62165473228872e+01, 2.61893864203288e+01, 2.61683192184264e+01, 2.61523670538986e+01, 2.61406183447949e+01, 2.61322411682135e+01, 2.61264930881374e+01, 2.61227279261543e+01, 2.61203992519777e+01, 2.61190604712413e+01, 2.61183614999879e+01, 2.61180421326378e+01, 2.61179223266582e+01, 2.61178897365023e+01, 2.61178849262627e+01, 2.61178845620458e+01, 2.61178781099809e+01, 2.61178343962883e+01, 2.61176737000814e+01, 2.61172453426188e+01, 2.61163078908868e+01, 2.61145125513815e+01, 2.61113904181005e+01, 2.61063441593985e+01, 2.60986446362157e+01, 2.60874328370065e+01, 2.60717273892685e+01, 2.60504377605765e+01, 2.60223830910031e+01, 2.59863164032943e+01, 2.59409537198134e+01, 2.58850073826874e+01, 2.58172226368819e+01, 2.57364163104252e+01, 2.56415162303202e+01, 2.55315998668697e+01, 2.54059306220551e+01, 2.52639901840184e+01, 2.51055054675121e+01, 2.49304688486105e+01, 2.47391506713530e+01, 2.45321033373084e+01, 2.43101566653102e+01, 2.40744046065681e+01, 2.38261838012789e+01, 2.35670448505953e+01, 2.32987175355802e+01, 2.30230715201681e+01, 2.27420742968817e+01, 2.24577482348323e+01, 2.21721285371829e+01, 2.18872236989503e+01, 2.16049796995571e+01, 2.13272487272720e+01, 2.10557627944636e+01, 2.07921122381338e+01, 2.05377288553401e+01, 2.02938733042169e+01, 2.00616263813552e+01, 1.98418838220761e+01, 1.96353543206271e+01, 1.94425605057621e+01, 1.92638426238564e+01, 1.90993646802058e+01, 1.89491227793626e+01, 1.88129553973748e+01, 1.86905553191072e+01, 1.85814829841054e+01, 1.84851810022838e+01, 1.84009896211403e+01, 1.83281629436085e+01, 1.82658857052373e+01, 1.82132904180693e+01, 1.81694746755545e+01, 1.81335183894561e+01, 1.81045006993016e+01, 1.80815162621850e+01, 1.80636906010858e+01, 1.80501941688377e+01, 1.80402547773952e+01, 1.80331680518892e+01, 1.80283055983802e+01, 1.80251206236305e+01, 1.80231508132056e+01, 1.80220183576463e+01, 1.80214271108380e+01, 1.80211569646148e+01, 1.80210556233312e+01, 1.80210280560346e+01, 1.80210239871590e+01, 1.80210236817072e+01, 1.80210183054096e+01, 1.80209818802035e+01, 1.80208479773093e+01, 1.80204910414839e+01, 1.80197098968102e+01, 1.80182139138043e+01, 1.80156123946307e+01, 1.80114076691292e+01, 1.80049923198662e+01, 1.79956508673503e+01, 1.79825661439371e+01, 1.79648304637290e+01, 1.79414615536900e+01, 1.79114230476938e+01, 1.78736491623945e+01, 1.78270729770034e+01, 1.77706575372553e+01, 1.77034288094238e+01, 1.76245093381975e+01, 1.75331513287556e+01, 1.74287677940185e+01, 1.73109603955576e+01, 1.71795426691119e+01, 1.70345574651691e+01, 1.68762876473388e+01, 1.67052593662521e+01, 1.65222375500690e+01, 1.63282136072697e+01, 1.61243857046922e+01, 1.59121323442928e+01, 1.56929802953902e+01, 1.54685682235469e+01, 1.52406075706194e+01, 1.50108423622934e+01, 1.47810096340562e+01, 1.45528020675069e+01, 1.43278342217573e+01, 1.41076134481855e+01, 1.38935162210214e+01, 1.36867702381066e+01, 1.34884422835721e+01, 1.32994315300687e+01, 1.31204677160621e+01, 1.29521134753569e+01, 1.27947700216981e+01, 1.26486853917587e+01, 1.25139645095062e+01, 1.23905804353899e+01, 1.22783862865917e+01, 1.21771274435064e+01, 1.20864537797677e+01, 1.20059317593610e+01, 1.19350563289766e+01, 1.18732625940875e+01, 1.18199373029056e+01, 1.17744301747950e+01, 1.17360651015695e+01, 1.17041512250170e+01, 1.16779938563012e+01, 1.16569051574512e+01, 1.16402144571207e+01, 1.16272780274192e+01, 1.16174881109065e+01, 1.16102809612237e+01, 1.16051436508104e+01, 1.16016194069277e+01, 1.15993112634486e+01, 1.15978838596437e+01, 1.15970632759377e+01, 1.15966348664996e+01, 1.15964391247916e+01, 1.15963656956594e+01, 1.15963457211871e+01, 1.15963427729980e+01, 1.15963425523496e+01, 1.15963386776522e+01, 1.15963124260250e+01, 1.15962159225268e+01, 1.15959586816312e+01, 1.15953957263693e+01, 1.15943176385768e+01, 1.15924429565406e+01, 1.15894132987176e+01, 1.15847915476034e+01, 1.15780633728983e+01, 1.15686423034909e+01, 1.15558784688549e+01, 1.15390710183773e+01, 1.15174840898852e+01, 1.14903660370932e+01, 1.14569714445144e+01, 1.14165852665540e+01, 1.13685482383234e+01, 1.13122825361675e+01, 1.12473165352509e+01, 1.11733074393573e+01, 1.10900605615481e+01, 1.09975441256342e+01, 1.08958986419001e+01, 1.07854401807744e+01, 1.06666572090436e+01, 1.05402010385284e+01, 1.04068703328144e+01, 1.02675904856735e+01, 1.01233889883646e+01, 9.97536811175983e+00, 9.82467632414437e+00, 9.67247984179437e+00, 9.51993557706538e+00, 9.36816653047211e+00, 9.21824040029637e+00, 9.07115188976098e+00, 8.92780890938686e+00, 8.78902262567179e+00, 8.65550111243353e+00, 8.52784622439250e+00, 8.40655323219616e+00, 8.29201272690350e+00, 8.18451430981442e+00, 8.08425161961318e+00, 7.99132830311397e+00, 7.90576460010155e+00, 7.82750428031576e+00, 7.75642173664794e+00, 7.69232909952505e+00, 7.63498329068551e+00, 7.58409297829917e+00, 7.53932542847715e+00, 7.50031327003938e+00, 7.46666119987540e+00, 7.43795265580246e+00, 7.41375647348338e+00, 7.39363352523908e+00, 7.37714331350398e+00, 7.36385046269551e+00, 7.35333102320784e+00, 7.34517847308143e+00, 7.33900927960239e+00, 7.33446786734594e+00, 7.33123083316386e+00, 7.32901025374242e+00, 7.32755594810146e+00, 7.32665658521730e+00, 7.32613956428244e+00, 7.32586963955265e+00, 7.32574631026516e+00, 7.32570004544252e+00, 7.32568746031271e+00, 7.32568560277475e+00, 7.32568546370081e+00, 7.32568302079754e+00, 7.32566646978084e+00, 7.32560562667900e+00, 7.32544344273978e+00, 7.32508851442073e+00, 7.32440881206761e+00, 7.32322689108294e+00, 7.32131682641327e+00, 7.31840308381562e+00, 7.31416150728062e+00, 7.30822255816763e+00, 7.30017688485329e+00, 7.28958322930885e+00, 7.27597858757976e+00, 7.25889043526325e+00, 7.23785071008683e+00, 7.21241111800189e+00, 7.18215920630563e+00, 7.14673453918464e+00, 7.10584423109176e+00, 7.05927705455876e+00, 7.00691535212441e+00, 6.94874405336246e+00, 6.88485622786049e+00, 6.81545478684995e+00, 6.74085016672620e+00, 6.66145406819992e+00, 6.57776956344856e+00, 6.49037809825290e+00, 6.39992408722770e+00, 6.30709791345754e+00, 6.21261819137385e+00, 6.11721413303762e+00, 6.02160877945987e+00, 5.92650373218106e+00, 5.83256586191018e+00, 5.74041629821744e+00, 5.65062183425417e+00, 5.56368872793284e+00, 5.48005875660702e+00, 5.40010729190014e+00, 5.32414310601393e+00, 5.25240959770333e+00, 5.18508712954107e+00, 5.12229619125161e+00, 5.06410113986833e+00, 5.01051431017078e+00, 4.96150033343388e+00, 4.91698054541516e+00, 4.87683740330250e+00, 4.84091886451799e+00, 4.80904270692761e+00, 4.78100078968812e+00, 4.75656326653629e+00, 4.73548276892048e+00, 4.71749857536813e+00, 4.70234077652677e+00, 4.68973443334049e+00, 4.67940371005378e+00, 4.67107594566083e+00, 4.66448560874994e+00, 4.65937806325510e+00, 4.65551305824715e+00, 4.65266784523590e+00, 4.65063982286311e+00, 4.64924861221620e+00, 4.64833747658538e+00, 4.64777401697573e+00, 4.64745009810348e+00, 4.64728098745235e+00, 4.64720372035629e+00, 4.64717473495583e+00, 4.64716685023872e+00, 4.64716568647154e+00, 4.64716559910909e+00, 4.64716406145699e+00, 4.64715364364551e+00, 4.64711534678591e+00, 4.64701326232091e+00, 4.64678985754311e+00, 4.64636202820529e+00, 4.64561808427523e+00, 4.64441582042912e+00, 4.64258180565180e+00, 4.63991200485356e+00, 4.63617381781435e+00, 4.63110958501292e+00, 4.62444156428441e+00, 4.61587832588147e+00, 4.60512244678644e+00, 4.59187931014883e+00, 4.57586673657805e+00, 4.55682509672758e+00, 4.53452748677142e+00, 4.50878949845698e+00, 4.47947809164720e+00, 4.44651908634710e+00, 4.40990283702074e+00, 4.36968773452973e+00, 4.32600129582255e+00, 4.27903873985059e+00, 4.22905909799975e+00, 4.17637905472526e+00, 4.12136484530413e+00, 4.06442264096798e+00, 4.00598791900237e+00, 3.94651434297268e+00, 3.88646266682223e+00, 3.82629013082207e+00, 3.76644074466556e+00, 3.70733676236833e+00, 3.64937155437610e+00, 3.59290398310399e+00, 3.53825429642377e+00, 3.48570147503978e+00, 3.43548190797272e+00, 3.38778922728966e+00, 3.34277510875772e+00, 3.30055083773748e+00, 3.26118944670651e+00, 3.22472824896919e+00, 3.19117161875170e+00, 3.16049389752102e+00, 3.13264233692060e+00, 3.10754001769109e+00, 3.08508870949740e+00, 3.06517165749356e+00, 3.04765629707158e+00, 3.03239690834498e+00, 3.01923722666572e+00, 3.00801302527833e+00, 2.99855468170334e+00, 2.99068973141151e+00, 2.98424540173750e+00, 2.97905110685269e+00, 2.97494087211703e+00, 2.97175564445573e+00, 2.96934543572267e+00, 2.96757123937324e+00, 2.96630665802371e+00, 2.96543918117744e+00, 2.96487105873430e+00, 2.96451972665324e+00, 2.96431775569589e+00, 2.96421231158466e+00, 2.96416413396840e+00, 2.96414606098951e+00, 2.96414114471128e+00, 2.96414041907927e+00, 2.96414036437506e+00, 2.96413939844476e+00, 2.96413285413038e+00, 2.96410879662096e+00, 2.96404466876722e+00, 2.96390432974312e+00, 2.96363557604337e+00, 2.96316824965625e+00, 2.96241302946888e+00, 2.96126099145950e+00, 2.95958401002204e+00, 2.95723605557694e+00, 2.95405542109108e+00, 2.94986788111224e+00, 2.94449075078762e+00, 2.93773776930718e+00, 2.92942468363485e+00, 2.91937535690759e+00, 2.90742817542146e+00, 2.89344248368914e+00, 2.87730474424178e+00, 2.85893410316790e+00, 2.83828704840612e+00, 2.81536087830835e+00, 2.79019575319847e+00, 2.76287517983842e+00, 2.73352487217701e+00, 2.70231003332354e+00, 2.66943120366805e+00, 2.63511890853280e+00, 2.59962740691320e+00, 2.56322788438938e+00, 2.52620144510526e+00, 2.48883224041604e+00, 2.45140102940908e+00, 2.41417940565860e+00, 2.37742485342939e+00, 2.34137672345288e+00, 2.30625315074346e+00, 2.27224888024216e+00, 2.23953392360242e+00, 2.20825294310386e+00, 2.17852524548243e+00, 2.15044526703751e+00, 2.12408343867574e+00, 2.09948733245454e+00, 2.07668300694138e+00, 2.05567648516417e+00, 2.03645531462914e+00, 2.01899017296060e+00, 2.00323649475813e+00, 1.98913610513431e+00, 1.97661885308740e+00, 1.96560424341499e+00, 1.95600306930874e+00, 1.94771904911349e+00, 1.94065047002975e+00, 1.93469183890456e+00, 1.92973553591597e+00, 1.92567346126708e+00, 1.92239865845721e+00, 1.91980689090184e+00, 1.91779814230895e+00, 1.91627800598947e+00, 1.91515892483316e+00, 1.91436124254837e+00, 1.91381402829050e+00, 1.91345564111238e+00, 1.91323400762710e+00, 1.91310659550913e+00, 1.91304007638848e+00, 1.91300968359249e+00, 1.91299828226038e+00, 1.91299518082680e+00, 1.91299472306189e+00, 1.91299468836394e+00, 1.91299407319968e+00, 1.91298990537435e+00, 1.91297458405299e+00, 1.91293374338480e+00, 1.91284436635399e+00, 1.91267320543982e+00, 1.91237557642727e+00, 1.91189458587347e+00, 1.91116084578873e+00, 1.91009272045839e+00, 1.90859713926562e+00, 1.90657099508246e+00, 1.90390312961554e+00, 1.90047688468111e+00, 1.89617317191703e+00, 1.89087398378307e+00, 1.88446623749596e+00, 1.87684581320765e+00, 1.86792162129699e+00, 1.85761951444015e+00, 1.84588585134603e+00, 1.83269052320549e+00, 1.81802927231718e+00, 1.80192516472462e+00, 1.78442912295733e+00, 1.76561947741627e+00, 1.74560055071699e+00, 1.72450034310795e+00, 1.70246743403015e+00, 1.67966725126074e+00, 1.65627788285806e+00, 1.63248561807487e+00, 1.60848040284017e+00, 1.58445138558235e+00, 1.56058271255755e+00, 1.53704971054326e+00, 1.51401557010772e+00, 1.49162861531016e+00, 1.47002021590936e+00, 1.44930336640637e+00, 1.42957192358328e+00, 1.41090046249717e+00, 1.39334468264724e+00, 1.37694227390077e+00, 1.36171413789424e+00, 1.34766585613541e+00, 1.33478930072062e+00, 1.32306429599630e+00, 1.31246025728913e+00, 1.30293775328114e+00, 1.29444995911415e+00, 1.28694398577296e+00, 1.28036208630904e+00, 1.27464275031792e+00, 1.26972170466362e+00, 1.26553284106801e+00, 1.26200909043588e+00, 1.25908326037798e+00, 1.25668884707441e+00, 1.25476082613356e+00, 1.25323642015167e+00, 1.25205583393021e+00, 1.25116294236358e+00, 1.25050591139677e+00, 1.25003772958632e+00, 1.24971662695896e+00, 1.24950635915745e+00, 1.24937633821942e+00, 1.24930159649156e+00, 1.24926257672067e+00, 1.24924474875527e+00, 1.24923806094925e+00, 1.24923624171136e+00, 1.24923597319592e+00, 1.24923595271368e+00, 1.24923558787762e+00, 1.24923311606348e+00, 1.24922402947125e+00, 1.24919980839786e+00, 1.24914680343110e+00, 1.24904530123852e+00, 1.24886881517322e+00, 1.24858363952960e+00, 1.24814870330788e+00, 1.24751575522521e+00, 1.24662990573252e+00, 1.24543054320172e+00, 1.24385262957473e+00, 1.24182836520228e+00, 1.23928919334956e+00, 1.23616809245566e+00, 1.23240207989406e+00, 1.22793482653114e+00, 1.22271925923455e+00, 1.21672001145683e+00, 1.20991557303014e+00, 1.20229999203040e+00, 1.19388399599475e+00, 1.18469542780140e+00, 1.17477893253898e+00, 1.16419488332420e+00, 1.15301759201748e+00, 1.14133290920515e+00, 1.12923536961678e+00, 1.11682507713512e+00, 1.10420454170226e+00, 1.09147567524049e+00, 1.07873712534477e+00, 1.06608207814121e+00, 1.05359660293192e+00, 1.04135855050399e+00, 1.02943696347216e+00, 1.01789191771722e+00, 1.00677469232708e+00, 9.96128161154303e-01, 9.85987308883523e-01, 9.76379793403126e-01, 9.67326499096117e-01, 9.58842048053853e-01, 9.50935255239926e-01, 9.43609527858698e-01, 9.36863218417599e-01, 9.30689945827140e-01, 9.25078900366444e-01, 9.20015147524765e-01, 9.15479943547510e-01, 9.11451072669834e-01, 9.07903212975157e-01, 9.04808334831856e-01, 9.02136133048742e-01, 8.99854491275221e-01, 8.97929974738716e-01, 8.96328345140674e-01, 8.95015089423850e-01, 8.93955952209026e-01, 8.93117460044607e-01, 8.92467424309189e-01, 8.91975408766939e-01, 8.91613147507798e-01, 8.91354899403694e-01, 8.91177726341382e-01, 8.91061684362910e-01, 8.90989919409719e-01, 8.90948662525503e-01, 8.90927122956555e-01, 8.90917281405009e-01, 8.90913589508216e-01, 8.90912585223603e-01, 8.90912436993332e-01, 8.90912425606601e-01, 8.90912221736542e-01, 8.90910840487493e-01, 8.90905762880209e-01, 8.90892227922379e-01, 8.90862607398745e-01, 8.90805882086696e-01, 8.90707241090411e-01, 8.90547823879771e-01, 8.90304622118904e-01, 8.89950555304964e-01, 8.89454730377750e-01, 8.88782890559497e-01, 8.87898052533643e-01, 8.86761323573808e-01, 8.85332881474582e-01, 8.83573090454334e-01, 8.81443716225462e-01, 8.78909194073231e-01, 8.75937896177348e-01, 8.72503339734394e-01, 8.68585276743545e-01, 8.64170610264722e-01, 8.59254090633229e-01, 8.53838757901130e-01, 8.47936112390570e-01, 8.41566011921657e-01, 8.34756310099223e-01, 8.27542263298568e-01, 8.19965743526197e-01, 8.12074299706921e-01, 8.03920111451973e-01, 7.95558877801938e-01, 7.87048679861726e-01, 7.78448851640093e-01, 7.69818888498833e-01, 7.61217417796544e-01, 7.52701251683391e-01, 7.44324537520669e-01, 7.36138016988846e-01, 7.28188400615494e-01, 7.20517860308959e-01, 7.13163638710924e-01, 7.06157770990188e-01, 6.99526912260424e-01, 6.93292262199321e-01, 6.87469577661673e-01, 6.82069264008021e-01, 6.77096536347705e-01, 6.72551642714524e-01, 6.68430142163686e-01, 6.64723231714191e-01, 6.61418116827207e-01, 6.58498420606520e-01, 6.55944627077488e-01, 6.53734553727546e-01, 6.51843847993135e-01, 6.50246501599526e-01, 6.48915375677177e-01, 6.47822728481373e-01, 6.46940736442161e-01, 6.46241998289554e-01, 6.45700011253794e-01, 6.45289607950602e-01, 6.44987342621562e-01, 6.44771815984271e-01, 6.44623929088084e-01, 6.44527058265936e-01, 6.44467145466436e-01, 6.44432700855461e-01, 6.44414717461061e-01, 6.44406500653204e-01, 6.44403418237388e-01, 6.44402579744835e-01, 6.44402455985073e-01, 6.44402446405564e-01, 6.44402273947719e-01, 6.44401105520402e-01, 6.44396810256950e-01, 6.44385360654098e-01, 6.44360303440038e-01, 6.44312315730815e-01, 6.44228864242280e-01, 6.44093983214574e-01, 6.43888184251949e-01, 6.43588508420137e-01, 6.43168727583543e-01, 6.42599698010187e-01, 6.41849864649259e-01, 6.40885909176305e-01, 6.39673528970690e-01, 6.38178327819901e-01, 6.36366792640288e-01, 6.34207324297122e-01, 6.31671285228492e-01, 6.28734022615035e-01, 6.25375823852224e-01, 6.21582761546819e-01, 6.17347388452004e-01, 6.12669248709238e-01, 6.07555180215044e-01, 6.02019393324525e-01, 5.96083322641462e-01, 5.89775260370814e-01, 5.83129790624355e-01, 5.76187053262473e-01, 5.68991872611804e-01, 5.61592790302043e-01, 5.54041042432814e-01, 5.46389519554793e-01, 5.38691744044598e-01, 5.31000894051388e-01, 5.23368897036235e-01, 5.15845609691667e-01, 5.08478095259477e-01, 5.01310004305637e-01, 4.94381061021784e-01, 4.87726654095127e-01, 4.81377529014367e-01, 4.75359577191355e-01, 4.69693716309649e-01, 4.64395855722103e-01, 4.59476940401800e-01, 4.54943066836435e-01, 4.50795664294304e-01, 4.47031735040071e-01, 4.43644147292829e-01, 4.40621974947533e-01, 4.37950878259969e-01, 4.35613519765724e-01, 4.33590009614903e-01, 4.31858374219956e-01, 4.30395041632110e-01, 4.29175336402565e-01, 4.28173975907370e-01, 4.27365559300938e-01, 4.26725039517549e-01, 4.26228168176124e-01, 4.25851902971900e-01, 4.25574767256848e-01, 4.25377152089538e-01, 4.25241552110691e-01, 4.25152728169124e-01, 4.25097791638649e-01, 4.25066207747051e-01, 4.25049717865022e-01, 4.25042183445967e-01, 4.25039357016246e-01, 4.25038588157975e-01, 4.25038474676086e-01, 4.25038465820904e-01, 4.25038305483127e-01, 4.25037219170889e-01, 4.25033225770122e-01, 4.25022580813732e-01, 4.24999284505233e-01, 4.24954668974067e-01, 4.24877081210170e-01, 4.24751676123307e-01, 4.24560331850628e-01, 4.24281696747087e-01, 4.23891374363990e-01, 4.23362249081649e-01, 4.22664950855392e-01, 4.21768452761441e-01, 4.20640789748347e-01, 4.19249881378121e-01, 4.17564435630766e-01, 4.15554905412037e-01, 4.13194464678512e-01, 4.10459967549357e-01, 4.07332851861682e-01, 4.03799948713399e-01, 3.99854161838292e-01, 3.95494985191490e-01, 3.90728833683497e-01, 3.85569170172872e-01, 3.80036421033935e-01, 3.74157682192021e-01, 3.67966226799766e-01, 3.61500834132940e-01, 3.54804966357918e-01, 3.47925825282672e-01, 3.40913324895837e-01, 3.33819017397332e-01, 3.26695010566439e-01, 3.19592912791734e-01, 3.12562839021155e-01, 3.05652506455061e-01, 2.98906443232084e-01, 2.92365326963591e-01, 2.86065463141497e-01, 2.80038406626360e-01, 2.74310723060324e-01, 2.68903881564668e-01, 2.63834265785547e-01, 2.59113287443597e-01, 2.54747585069921e-01, 2.50739290485882e-01, 2.47086346601876e-01, 2.43782861989999e-01, 2.40819490098588e-01, 2.38183823603023e-01, 2.35860796921884e-01, 2.33833092133537e-01, 2.32081545222798e-01, 2.30585550669502e-01, 2.29323462831815e-01, 2.28272992412016e-01, 2.27411595622854e-01, 2.26716852637907e-01, 2.26166830685813e-01, 2.25740425921856e-01, 2.25417677167570e-01, 2.25180043910413e-01, 2.25010640730872e-01, 2.24894420653044e-01, 2.24818300821738e-01, 2.24771225365330e-01, 2.24744162224269e-01, 2.24730032991630e-01, 2.24723577266445e-01, 2.24721155507183e-01, 2.24720496730527e-01, 2.24720399496491e-01, 2.24720391844599e-01, 2.24720252467179e-01, 2.24719308165234e-01, 2.24715836819712e-01, 2.24706583550869e-01, 2.24686333322325e-01, 2.24647552953812e-01, 2.24580117454764e-01, 2.24471134203554e-01, 2.24304876861908e-01, 2.24062838922791e-01, 2.23723914302070e-01, 2.23264709257492e-01, 2.22659986032156e-01, 2.21883233919144e-01, 2.20907357960485e-01, 2.19705469378567e-01, 2.18251755367760e-01, 2.16522399449104e-01, 2.14496517717478e-01, 2.12157071558252e-01, 2.09491714359245e-01, 2.06493528925961e-01, 2.03161614127552e-01, 1.99501483988215e-01, 1.95525249981443e-01, 1.91251567415364e-01, 1.86705338972686e-01, 1.81917181903444e-01, 1.76922679080931e-01, 1.71761447024636e-01, 1.66476064963490e-01, 1.61110917060043e-01, 1.55711004277705e-01, 1.50320782621264e-01, 1.44983080602415e-01, 1.39738141189483e-01, 1.34622822999731e-01, 1.29669983168313e-01, 1.24908051408738e-01, 1.20360792468361e-01, 1.16047243507367e-01, 1.11981804633558e-01, 1.08174455306155e-01, 1.04631066636230e-01, 1.01353779518448e-01, 9.83414205784925e-02, 9.55899315424489e-02, 9.30927922268125e-02, 9.08414223542593e-02, 8.88255523476005e-02, 8.70335577849772e-02, 8.54527560700831e-02, 8.40696669390866e-02, 8.28702406395934e-02, 8.18400589848096e-02, 8.09645150653040e-02, 8.02289772833262e-02, 7.96189426700096e-02, 7.91201832865340e-02, 7.87188880334639e-02, 7.84018005524132e-02, 7.81563522499766e-02, 7.79707879557208e-02, 7.78342804826906e-02, 7.77370294988060e-02, 7.76703397228431e-02, 7.76266735659590e-02, 7.75996739370726e-02, 7.75841539642528e-02, 7.75760517584843e-02, 7.75723499350249e-02, 7.75709612761091e-02, 7.75705835293650e-02, 7.75705277748244e-02, 7.75705233483563e-02, 7.75704422275112e-02, 7.75698926231740e-02, 7.75678722387738e-02, 7.75624867838758e-02, 7.75507015991650e-02, 7.75281344406977e-02, 7.74888989755738e-02, 7.74255082314141e-02, 7.73288468158029e-02, 7.71882200106546e-02, 7.69914866365582e-02, 7.67252806449790e-02, 7.63753235360554e-02, 7.59268257832986e-02, 7.53649704764001e-02, 7.46754665134830e-02, 7.38451522333179e-02, 7.28626238971373e-02, 7.17188576019652e-02, 7.04077888193808e-02, 6.89268116017558e-02, 6.72771602661726e-02, 6.54641405056537e-02, 6.34971844840635e-02, 6.13897152168619e-02, 5.91588186372292e-02, 5.68247360115635e-02, 5.44102033299794e-02, 5.19396764037197e-02, 4.94384892315054e-02, 4.69319976955376e-02, 4.44447602975089e-02, 4.19998025568589e-02, 3.96180025959580e-02, 3.73176235469117e-02, 3.51140052325576e-02, 3.30194146441158e-02, 3.10430434075641e-02, 2.91911316820830e-02, 2.74671922892367e-02, 2.58723063877113e-02, 2.44054623594110e-02, 2.30639121595004e-02, 2.18435234844492e-02, 2.07391110000543e-02, 1.97447349121092e-02, 1.88539598707962e-02, 1.80600712582759e-02, 1.73562491552985e-02, 1.67357026836470e-02, 1.61917690327442e-02, 1.57179824121698e-02, 1.53081185554385e-02, 1.49562203703795e-02, 1.46566099984398e-02, 1.44038920054470e-02, 1.41929517506583e-02, 1.40189522189331e-02, 1.38773317935090e-02, 1.37638046223714e-02, 1.36743644195292e-02, 1.36052917749652e-02, 1.35531643544198e-02, 1.35148687912050e-02, 1.34876126381158e-02, 1.34689344870793e-02, 1.34567102930845e-02, 1.34491540602681e-02, 1.34448113424241e-02, 1.34425444484663e-02, 1.34415087763244e-02, 1.34411202745781e-02, 1.34410145940649e-02, 1.34409989958897e-02, 1.34409977462013e-02, 1.34409747009272e-02, 1.34408185667354e-02, 1.34402446128234e-02, 1.34387147533488e-02, 1.34353671552105e-02, 1.34289578736571e-02, 1.34178176149362e-02, 1.33998268508151e-02, 1.33724126156523e-02, 1.33325702924623e-02, 1.32769134193325e-02, 1.32017538814324e-02, 1.31032136717693e-02, 1.29773676276699e-02, 1.28204141717703e-02, 1.26288682092905e-02, 1.23997671837230e-02, 1.21308782253919e-02, 1.18208917973845e-02, 1.14695857568989e-02, 1.10779437703457e-02, 1.06482138816659e-02, 1.01838968304406e-02, 9.68965924077736e-03, 9.17117352662444e-03, 8.63489347572965e-03, 8.08778102307883e-03, 7.53700474872687e-03, 6.98963337228520e-03, 6.45234755263857e-03, 5.93119065139547e-03, 5.43137423925089e-03, 4.95714779819423e-03, 4.51173526819559e-03, 4.09733475035523e-03, 3.71517258348320e-03, 3.36559962798333e-03, 3.04821606643202e-03, 2.76201120480835e-03, 2.50550626468282e-03, 2.27689051732227e-03, 2.07414384715321e-03, 1.89514153967707e-03, 1.73773947177835e-03, 1.59983977837896e-03, 1.47943841800213e-03, 1.37465689036276e-03, 1.28376075421954e-03, 1.20516765635159e-03, 1.13744742037625e-03, 1.07931645035819e-03, 1.02962834915843e-03, 9.87362290501892e-04, 9.51610347257770e-04, 9.21564683821430e-04, 8.96505274070591e-04, 8.75788606450101e-04, 8.58837676261714e-04, 8.45133435262824e-04, 8.34207762821393e-04, 8.25637932152290e-04, 8.19042468393150e-04, 8.14078228259518e-04, 8.10438476079867e-04, 8.07851690609596e-04, 8.06080810519916e-04, 8.04922623458992e-04, 8.04207017992553e-04, 8.03795853627776e-04, 8.03581257684982e-04, 8.03483222782120e-04, 8.03446449026524e-04, 8.03436445934835e-04, 8.03434969512296e-04, 8.03434850098483e-04, 8.03432634327805e-04, 8.03417622250890e-04, 8.03362438418276e-04, 8.03215354437725e-04, 8.02893545412076e-04, 8.02277551471775e-04, 8.01207298877671e-04, 7.99480077077419e-04, 7.96850910518287e-04, 7.93035783416940e-04, 7.87718152693628e-04, 7.80559089349708e-04, 7.71211202447830e-04, 7.59336203842054e-04, 7.44625572423274e-04, 7.26823295252643e-04, 7.05749157447144e-04, 6.81320599170476e-04, 6.53570857878094e-04, 6.22661065861507e-04, 5.88884253232891e-04, 5.52659843734205e-04, 5.14518185738666e-04, 4.75075817647053e-04, 4.35003348499663e-04, 3.94988838110066e-04, 3.55700196280395e-04, 3.17750256982663e-04, 2.81667774515881e-04, 2.47876702644589e-04, 2.16684938038725e-04, 1.88282456448974e-04, 1.62747644679939e-04, 1.40059921378111e-04, 1.20116744739107e-04, 1.02753484583218e-04, 8.77642639380838e-05, 7.49207402321872e-05, 6.39861604980752e-05, 5.47249676401407e-05, 4.69104321041414e-05, 4.03312944712297e-05, 3.47961894817672e-05, 3.01353483536104e-05, 2.62006505948875e-05, 2.28647254942871e-05, 2.00192347818984e-05, 1.75730372404814e-05, 1.54508094851735e-05, 1.35916675958470e-05, 1.19473180291152e-05, 1.04800904239255e-05, 9.16127058303573e-06, 7.96961307969468e-06, 6.88982042618136e-06, 5.91106989134005e-06, 5.02575251459803e-06, 4.22844385639110e-06, 3.51505070739222e-06, 2.88212198995552e-06, 2.32634536700971e-06, 1.84422499671655e-06, 1.43190024459727e-06, 1.08505847151719e-06, 7.98909695731709e-07, 5.68206004913918e-07, 1.42108547152020e-14, 2.84217094304040e-14, 2.84217094304040e-14, 2.84217094304040e-14, 1.42108547152020e-14, 2.84217094304040e-14, 1.42108547152020e-14, 1.42108547152020e-14],
[ 7.10000000000000e+01, 7.09999999232093e+01, 7.09999975486875e+01, 7.09999814609262e+01, 7.09999223202083e+01, 7.09997646696530e+01, 7.09994196394632e+01, 7.09987588048094e+01, 7.09976094182912e+01, 7.09957511953487e+01, 7.09929147829244e+01, 7.09887819900758e+01, 7.09829878064202e+01, 7.09751241827588e+01, 7.09647455004466e+01, 7.09513756141968e+01, 7.09345163185231e+01, 7.09136570613935e+01, 7.08882857091093e+01, 7.08579001517957e+01, 7.08220205259618e+01, 7.07802018155669e+01, 7.07320465724734e+01, 7.06772174690301e+01, 7.06154493602223e+01, 7.05465604938557e+01, 7.04704624713272e+01, 7.03871685377987e+01, 7.02967997790652e+01, 7.01995888319325e+01, 7.00958807809428e+01, 6.99861310171538e+01, 6.98708999687936e+01, 6.97508447678855e+01, 6.96267080763637e+01, 6.94993044433663e+01, 6.93695046872307e+01, 6.92382188799719e+01, 6.91063785528430e+01, 6.89749187389997e+01, 6.88447604285895e+01, 6.87167939418039e+01, 6.85918636374415e+01, 6.84707542791621e+01, 6.83541792881248e+01, 6.82427710258573e+01, 6.81370731789500e+01, 6.80375352588397e+01, 6.79445091848461e+01, 6.78582478846892e+01, 6.77789058212471e+01, 6.77065413345334e+01, 6.76411206713439e+01, 6.75825235598449e+01, 6.75305501713191e+01, 6.74849292957428e+01, 6.74453275417794e+01, 6.74113593555650e+01, 6.73825976370672e+01, 6.73585847188483e+01, 6.73388434608839e+01, 6.73228882079117e+01, 6.73102353537504e+01, 6.73004132611797e+01, 6.72929712970991e+01, 6.72874877612707e+01, 6.72835765131065e+01, 6.72808921343600e+01, 6.72791335054806e+01, 6.72780457186047e+01, 6.72774202991966e+01, 6.72770937594046e+01, 6.72769445573662e+01, 6.72768885859950e+01, 6.72768733603744e+01, 6.72768711131029e+01, 6.72768708183082e+01, 6.72768639499470e+01, 6.72768174157719e+01, 6.72766463509608e+01, 6.72761903529765e+01, 6.72751924023718e+01, 6.72732811686164e+01, 6.72699573951980e+01, 6.72645849666576e+01, 6.72563871547429e+01, 6.72444484207895e+01, 6.72277220154387e+01, 6.72050434635388e+01, 6.71751498506917e+01, 6.71367046385984e+01, 6.70883275307470e+01, 6.70286286915631e+01, 6.69562463967210e+01, 6.68698869687921e+01, 6.67683656439057e+01, 6.66506468399851e+01, 6.65158821794680e+01, 6.63634445874400e+01, 6.61929568677453e+01, 6.60043133749440e+01, 6.57976937515994e+01, 6.55735681652636e+01, 6.53326940060664e+01, 6.50761045197831e+01, 6.48050902716920e+01, 6.45211745979590e+01, 6.42260842756351e+01, 6.39217165516870e+01, 6.36101034848612e+01, 6.32933743669786e+01, 6.29737168907645e+01, 6.26533377660589e+01, 6.23344236380089e+01, 6.20191033515021e+01, 6.17094127273833e+01, 6.14072629774566e+01, 6.11144136573761e+01, 6.08324506854912e+01, 6.05627695412929e+01, 6.03065634061385e+01, 6.00648157911931e+01, 5.98382971263321e+01, 5.96275648252923e+01, 5.94329664416876e+01, 5.92546456368185e+01, 5.90925507610191e+01, 5.89464458932193e+01, 5.88159241904827e+01, 5.87004233794583e+01, 5.85992431850352e+01, 5.85115644462205e+01, 5.84364696209725e+01, 5.83729643339292e+01, 5.83199995760496e+01, 5.82764941251563e+01, 5.82413567234351e+01, 5.82135075246341e+01, 5.81918983127389e+01, 5.81755309979650e+01, 5.81634739172412e+01, 5.81548755064354e+01, 5.81489749707808e+01, 5.81451096574073e+01, 5.81427189273513e+01, 5.81413444304305e+01, 5.81406268004420e+01, 5.81402989050972e+01, 5.81401758995414e+01, 5.81401424389825e+01, 5.81401375002717e+01, 5.81401370642776e+01, 5.81401285211866e+01, 5.81400706404762e+01, 5.81398578646760e+01, 5.81392906810848e+01, 5.81380494064180e+01, 5.81356721891456e+01, 5.81315381013532e+01, 5.81248560825904e+01, 5.81146603714357e+01, 5.80998129140825e+01, 5.80790130696686e+01, 5.80508147356916e+01, 5.80136507910690e+01, 5.79658644991383e+01, 5.79057472320690e+01, 5.78315815809583e+01, 5.77416886175295e+01, 5.76344777949056e+01, 5.75084977419083e+01, 5.73624860447815e+01, 5.71954160464583e+01, 5.70065387427938e+01, 5.67954180210909e+01, 5.65619577566262e+01, 5.63064196312445e+01, 5.60294309295891e+01, 5.57319819698213e+01, 5.54154132147073e+01, 5.50813924797197e+01, 5.47318830135624e+01, 5.43691035801017e+01, 5.39954820128115e+01, 5.36136040164606e+01, 5.32261592116145e+01, 5.28358865084803e+01, 5.24455208262978e+01, 5.20577429414031e+01, 5.16751338832846e+01, 5.13001348607689e+01, 5.09350132566976e+01, 5.05818348381238e+01, 5.02424420288685e+01, 4.99184378959608e+01, 4.96111754026109e+01, 4.93217514553473e+01, 4.90510052942577e+01, 4.87995208180107e+01, 4.85676324815636e+01, 4.83554344440053e+01, 4.81627926729696e+01, 4.79893597303469e+01, 4.78345919728325e+01, 4.76977689012593e+01, 4.75780143849761e+01, 4.74743194715439e+01, 4.73855664675457e+01, 4.73105539437918e+01, 4.72480222792203e+01, 4.71966793151837e+01, 4.71552256496068e+01, 4.71223790636175e+01, 4.70968975469994e+01, 4.70776003783746e+01, 4.70633867258208e+01, 4.70532512668646e+01, 4.70462963849800e+01, 4.70417405824423e+01, 4.70389228542286e+01, 4.70373028903438e+01, 4.70364571087080e+01, 4.70360706607831e+01, 4.70359256902893e+01, 4.70358862547494e+01, 4.70358804341445e+01, 4.70358799672600e+01, 4.70358713508724e+01, 4.70358129735713e+01, 4.70355983723196e+01, 4.70350263232721e+01, 4.70337744033490e+01, 4.70313768093964e+01, 4.70272073185286e+01, 4.70204681654553e+01, 4.70101854876200e+01, 4.69952118441389e+01, 4.69742361460139e+01, 4.69458011389082e+01, 4.69083283530335e+01, 4.68601501770447e+01, 4.67995484278047e+01, 4.67247984840197e+01, 4.66342177432030e+01, 4.65262168677554e+01, 4.63993520309009e+01, 4.62523761825024e+01, 4.60842872528843e+01, 4.58943712192944e+01, 4.56822380857299e+01, 4.54478490728495e+01, 4.51915336694333e+01, 4.49139956393966e+01, 4.46163075815475e+01, 4.42998941739800e+01, 4.39665047729194e+01, 4.36181765493046e+01, 4.32571898049858e+01, 4.28860174772555e+01, 4.25072710728799e+01, 4.21236453314124e+01, 4.17378637824068e+01, 4.13526270476961e+01, 4.09705653037811e+01, 4.05941958433898e+01, 4.02258862426644e+01, 3.98678233077682e+01, 3.95219877579761e+01, 3.91901344817459e+01, 3.88737781401162e+01, 3.85741838520618e+01, 3.82923626573739e+01, 3.80290714090066e+01, 3.77848167048749e+01, 3.75598624388418e+01, 3.73542405396780e+01, 3.71677644771523e+01, 3.70000451425228e+01, 3.68505087491067e+01, 3.67184164383736e+01, 3.66028853098664e+01, 3.65029106129748e+01, 3.64173888417177e+01, 3.63451414596242e+01, 3.62849389525398e+01, 3.62355248667897e+01, 3.61956394442498e+01, 3.61640424209893e+01, 3.61395345190894e+01, 3.61209771384551e+01, 3.61073097524482e+01, 3.60975645320460e+01, 3.60908777701216e+01, 3.60864977503639e+01, 3.60837888020856e+01, 3.60822313984397e+01, 3.60814182854375e+01, 3.60810467655770e+01, 3.60809073953576e+01, 3.60808694832509e+01, 3.60808638875024e+01, 3.60808634559044e+01, 3.60808557057902e+01, 3.60808031976246e+01, 3.60806101720338e+01, 3.60800956365280e+01, 3.60789695873977e+01, 3.60768130656870e+01, 3.60730628472247e+01, 3.60670014727497e+01, 3.60577532150334e+01, 3.60442864526395e+01, 3.60254227704973e+01, 3.59998529315648e+01, 3.59661596583183e+01, 3.59228469268290e+01, 3.58683752126274e+01, 3.58012018439397e+01, 3.57198253269386e+01, 3.56228322271603e+01, 3.55089449430295e+01, 3.53770685156085e+01, 3.52263345070719e+01, 3.50561399696747e+01, 3.48661796317499e+01, 3.46564696537354e+01, 3.44273616515452e+01, 3.41795461323434e+01, 3.39140450141889e+01, 3.36321934720757e+01, 3.33356119270502e+01, 3.30261695258441e+01, 3.27059408983281e+01, 3.23771582859634e+01, 3.20421612743993e+01, 3.17033463233984e+01, 3.13631180760927e+01, 3.10238440794433e+01, 3.06878141101205e+01, 3.03572048361938e+01, 3.00340501145156e+01, 2.97202168728204e+01, 2.94173862799903e+01, 2.91270397706346e+01, 2.88504494450621e+01, 2.85886723853239e+01, 2.83425484819506e+01, 2.81127014285256e+01, 2.78995425953610e+01, 2.77032775318662e+01, 2.75239148698391e+01, 2.73612774111525e+01, 2.72150151882630e+01, 2.70846202882465e+01, 2.69694432318750e+01, 2.68687106975936e+01, 2.67815443739577e+01, 2.67069807108779e+01, 2.66439913185654e+01, 2.65915037335537e+01, 2.65484222354790e+01, 2.65136483598777e+01, 2.64861007156972e+01, 2.64647336867312e+01, 2.64485545789813e+01, 2.64366387756297e+01, 2.64281424813330e+01, 2.64223126798215e+01, 2.64184939933529e+01, 2.64161322175700e+01, 2.64147744071043e+01, 2.64140655007356e+01, 2.64137415939763e+01, 2.64136200851121e+01, 2.64135870317306e+01, 2.64135821531199e+01, 2.64135817837262e+01, 2.64135752399538e+01, 2.64135309049302e+01, 2.64133679246490e+01, 2.64129334787162e+01, 2.64119827026862e+01, 2.64101618462078e+01, 2.64069953407396e+01, 2.64018773702879e+01, 2.63940684453827e+01, 2.63826973720144e+01, 2.63667688802624e+01, 2.63451770284433e+01, 2.63167243253145e+01, 2.62801463146804e+01, 2.62341411463213e+01, 2.61774034211473e+01, 2.61086613579629e+01, 2.60267160995850e+01, 2.59304817759108e+01, 2.58190247909376e+01, 2.56916007186903e+01, 2.55476871945648e+01, 2.53870112823334e+01, 2.52095699834039e+01, 2.50156428259897e+01, 2.48057958129342e+01, 2.45808763991819e+01, 2.43419995930994e+01, 2.40905257097605e+01, 2.38280307274524e+01, 2.35562705849649e+01, 2.32771410729464e+01, 2.29926351772709e+01, 2.27047997870205e+01, 2.24156935616526e+01, 2.21273474699355e+01, 2.18417291137757e+01, 2.15607115083164e+01, 2.12860465867944e+01, 2.10193433953959e+01, 2.07620507631883e+01, 2.05154441607024e+01, 2.02806164594644e+01, 2.00584723306504e+01, 1.98497260418052e+01, 1.96549024114531e+01, 1.94743406633545e+01, 1.93082008952945e+01, 1.91564728538546e+01, 1.90189866957851e+01, 1.88954254221093e+01, 1.87853386913842e+01, 1.86881577482701e+01, 1.86032112355926e+01, 1.85297416854346e+01, 1.84669225019377e+01, 1.84138752522195e+01, 1.83696870714413e+01, 1.83334279653553e+01, 1.83041677623760e+01, 1.82809924325312e+01, 1.82630194584507e+01, 1.82494119197522e+01, 1.82393909420694e+01, 1.82322461695709e+01, 1.82273439475721e+01, 1.82241329502459e+01, 1.82221470560906e+01, 1.82210053575168e+01, 1.82204092860473e+01, 1.82201369356308e+01, 1.82200347675160e+01, 1.82200069753049e+01, 1.82200028732325e+01, 1.82200025652886e+01, 1.82199971451274e+01, 1.82199604227405e+01, 1.82198254274018e+01, 1.82194655797019e+01, 1.82186780633834e+01, 1.82171698814468e+01, 1.82145471592864e+01, 1.82103081938303e+01, 1.82038406734601e+01, 1.81944234065357e+01, 1.81812327934985e+01, 1.81633541555148e+01, 1.81397978890122e+01, 1.81095202495889e+01, 1.80714483826652e+01, 1.80245090172600e+01, 1.79676600324826e+01, 1.78999239062482e+01, 1.78204218775916e+01, 1.77284075143033e+01, 1.76232982924546e+01, 1.75047037771899e+01, 1.73724490539366e+01, 1.72265921991929e+01, 1.70674347972396e+01, 1.68955247945423e+01, 1.67116513233407e+01, 1.65168315023536e+01, 1.63122896152450e+01, 1.60994294538486e+01, 1.58798009685350e+01, 1.56550626665951e+01, 1.54269414155855e+01, 1.51971914200741e+01, 1.49675541324107e+01, 1.47397207276035e+01, 1.45152985295371e+01, 1.42957824448254e+01, 1.40825320765100e+01, 1.38767547929895e+01, 1.36794946573363e+01, 1.34916268109517e+01, 1.33138566747374e+01, 1.31467231894127e+01, 1.29906052613450e+01, 1.28457305992317e+01, 1.27121862025770e+01, 1.25899298755213e+01, 1.24788022704622e+01, 1.23785390991075e+01, 1.22887832719383e+01, 1.22090968322266e+01, 1.21389726329410e+01, 1.20778457621313e+01, 1.20251047548997e+01, 1.19801026396173e+01, 1.19421678554914e+01, 1.19106150516351e+01, 1.18847557387196e+01, 1.18639087178729e+01, 1.18474101627584e+01, 1.18346231848569e+01, 1.18249466737886e+01, 1.18178231784216e+01, 1.18127455839606e+01, 1.18092623473947e+01, 1.18069810792931e+01, 1.18055703030376e+01, 1.18047592806620e+01, 1.18043358637207e+01, 1.18041424032675e+01, 1.18040698299356e+01, 1.18040500882639e+01, 1.18040471744358e+01, 1.18040469563590e+01, 1.18040431268210e+01, 1.18040171811562e+01, 1.18039218024213e+01, 1.18036675598321e+01, 1.18031111668003e+01, 1.18020456484163e+01, 1.18001928307950e+01, 1.17971985280635e+01, 1.17926307592682e+01, 1.17859812736515e+01, 1.17766705941233e+01, 1.17640567006042e+01, 1.17474473632686e+01, 1.17261159984537e+01, 1.16993207579860e+01, 1.16663263806427e+01, 1.16264281415928e+01, 1.15789770455813e+01, 1.15234052397271e+01, 1.14592504918459e+01, 1.13861785101174e+01, 1.13040018871129e+01, 1.12126945476890e+01, 1.11124007698471e+01, 1.10034381240002e+01, 1.08862940217945e+01, 1.07616159531372e+01, 1.06301958842031e+01, 1.04929496513906e+01, 1.03508924798647e+01, 1.02051119512941e+01, 1.00567398264868e+01, 9.90692409236966e+00, 9.75680246131607e+00, 9.60747832853901e+00, 9.45999992187605e+00, 9.31534309139770e+00, 9.17439791399794e+00, 9.03795905327274e+00, 8.90671963118928e+00, 8.78126823994383e+00, 8.66208864741510e+00, 8.54956172039412e+00, 8.44396909672604e+00, 8.34549817076474e+00, 8.25424800720538e+00, 8.17023585898816e+00, 8.09340402963206e+00, 8.02362688443870e+00, 7.96071787491583e+00, 7.90443649374659e+00, 7.85449512161882e+00, 7.81056576080782e+00, 7.77228667278597e+00, 7.73926894812454e+00, 7.71110303697527e+00, 7.68736525846124e+00, 7.66762428891437e+00, 7.65144760411234e+00, 7.63840782196900e+00, 7.62808886231534e+00, 7.62009181243715e+00, 7.61404036384165e+00, 7.60958566997100e+00, 7.60641046841228e+00, 7.60423231596663e+00, 7.60280580121410e+00, 7.60192362640218e+00, 7.60141648707713e+00, 7.60115172148898e+00, 7.60103074946755e+00, 7.60098536893866e+00, 7.60097302435836e+00, 7.60097120232501e+00, 7.60097106590930e+00, 7.60096866969920e+00, 7.60095243503567e+00, 7.60089275488201e+00, 7.60073367093910e+00, 7.60038552690334e+00, 7.59971881733030e+00, 7.59855949110577e+00, 7.59668594612980e+00, 7.59382792512181e+00, 7.58966748914770e+00, 7.58384220255763e+00, 7.57595060732624e+00, 7.56555999363078e+00, 7.55221638545556e+00, 7.53545655567652e+00, 7.51482176772365e+00, 7.48987281687567e+00, 7.46020582287873e+00, 7.42546811875239e+00, 7.38537350156350e+00, 7.33971607252161e+00, 7.28838190661047e+00, 7.23135786246291e+00, 7.16873697173330e+00, 7.10072002743146e+00, 7.02761320949259e+00, 6.94982182477990e+00, 6.86784047602335e+00, 6.78224018772035e+00, 6.69365318700844e+00, 6.60275614914983e+00, 6.51025276238389e+00, 6.41685644480393e+00, 6.32327396313772e+00, 6.23019057248877e+00, 6.13825713407029e+00, 6.04807949337152e+00, 5.96021023191165e+00, 5.87514275658201e+00, 5.79330757171567e+00, 5.71507049530420e+00, 5.64073253210785e+00, 5.57053109884090e+00, 5.50464230392200e+00, 5.44318400939898e+00, 5.38621943888153e+00, 5.33376113697891e+00, 5.28577512848554e+00, 5.24218516627586e+00, 5.20287699347715e+00, 5.16770257665841e+00, 5.13648429171327e+00, 5.10901906240771e+00, 5.08508246307150e+00, 5.06443280175463e+00, 5.04681519870930e+00, 5.03196566792526e+00, 5.01961519754803e+00, 5.00949380953475e+00, 5.00133456129532e+00, 4.99487743397296e+00, 4.98987303520175e+00, 4.98608603038839e+00, 4.98329820740506e+00, 4.98131107633590e+00, 4.97994790943625e+00, 4.97905513701211e+00, 4.97850303217569e+00, 4.97818564042828e+00, 4.97801993730917e+00, 4.97794422709345e+00, 4.97791582572461e+00, 4.97790809987751e+00, 4.97790695955918e+00, 4.97790687395700e+00, 4.97790536728718e+00, 4.97789515938490e+00, 4.97785763416844e+00, 4.97775760658011e+00, 4.97753870302009e+00, 4.97711949326151e+00, 4.97639053637543e+00, 4.97521248892447e+00, 4.97341540669345e+00, 4.97079935009418e+00, 4.96713637626488e+00, 4.96217396582485e+00, 4.95563988759504e+00, 4.94724844948384e+00, 4.93670801856585e+00, 4.92372962021280e+00, 4.90803634895340e+00, 4.88937324842597e+00, 4.86751725176731e+00, 4.84228672531123e+00, 4.81355013556343e+00, 4.78123336860237e+00, 4.74532527607275e+00, 4.70588110274321e+00, 4.66302356282240e+00, 4.61694146734483e+00, 4.56788595116645e+00, 4.51616449192439e+00, 4.46213304130721e+00, 4.40618668970241e+00, 4.34874935071061e+00, 4.29026297838077e+00, 4.23117681792949e+00, 4.17193714460799e+00, 4.11297787257112e+00, 4.05471232506167e+00, 3.99752635842550e+00, 3.94177293437539e+00, 3.88776814516861e+00, 3.83578862074702e+00, 3.78607018906885e+00, 3.73880762233163e+00, 3.69415528206655e+00, 3.65222847309553e+00, 3.61310532690754e+00, 3.57682905538633e+00, 3.54341044220209e+00, 3.51283046809047e+00, 3.48504299480486e+00, 3.45997745857127e+00, 3.43754154594914e+00, 3.41762384227208e+00, 3.40009645499762e+00, 3.38481762137904e+00, 3.37163431220836e+00, 3.36038484146488e+00, 3.35090148617513e+00, 3.34301311239935e+00, 3.33654779285781e+00, 3.33133539025194e+00, 3.32721006884259e+00, 3.32401268638864e+00, 3.32159301016137e+00, 3.31981169538169e+00, 3.31854196282755e+00, 3.31767091501450e+00, 3.31710043740642e+00, 3.31674764233324e+00, 3.31654482808021e+00, 3.31643894304838e+00, 3.31639056382477e+00, 3.31637241519218e+00, 3.31636747833204e+00, 3.31636674966212e+00, 3.31636669472890e+00, 3.31636572475460e+00, 3.31635915304064e+00, 3.31633499479281e+00, 3.31627059829552e+00, 3.31612967078370e+00, 3.31585978788300e+00, 3.31539049099354e+00, 3.31463206765395e+00, 3.31347509881907e+00, 3.31179084322289e+00, 3.30943251141910e+00, 3.30623745945968e+00, 3.30203030344661e+00, 3.29662692082577e+00, 3.28983926254038e+00, 3.28148085334588e+00, 3.27137280825799e+00, 3.25935014503197e+00, 3.24526813053721e+00, 3.22900836822369e+00, 3.21048431979305e+00, 3.18964596095292e+00, 3.16648330124553e+00, 3.14102855140936e+00, 3.11335679568987e+00, 3.08358511525800e+00, 3.05187020440906e+00, 3.01840461416876e+00, 2.98341183897476e+00, 2.94714052322341e+00, 2.90985810023140e+00, 2.87184418444097e+00, 2.83338401992314e+00, 2.79476224895659e+00, 2.75625721043674e+00, 2.71813591680863e+00, 2.68064979750469e+00, 2.64403124247740e+00, 2.60849093524835e+00, 2.57421593258528e+00, 2.54136842711419e+00, 2.51008511810963e+00, 2.48047711192682e+00, 2.45263027456313e+00, 2.42660596262455e+00, 2.40244206417196e+00, 2.38015428683558e+00, 2.35973763701385e+00, 2.34116804094253e+00, 2.32440406596180e+00, 2.30938870828648e+00, 2.29605122163761e+00, 2.28430896867367e+00, 2.27406928362439e+00, 2.26523133926752e+00, 2.25768801393106e+00, 2.25132775431260e+00, 2.24603642762303e+00, 2.24169915219743e+00, 2.23820208983786e+00, 2.23543417650730e+00, 2.23328876144187e+00, 2.23166511916207e+00, 2.23046979504775e+00, 2.22961774374151e+00, 2.22903322108164e+00, 2.22865039469126e+00, 2.22841364563101e+00, 2.22827754325273e+00, 2.22820648694460e+00, 2.22817402105001e+00, 2.22816184202379e+00, 2.22815852903825e+00, 2.22815804004868e+00, 2.22815800298396e+00, 2.22815734585869e+00, 2.22815289374111e+00, 2.22813652733022e+00, 2.22809290082714e+00, 2.22799742701473e+00, 2.22781458984333e+00, 2.22749665512742e+00, 2.22698284358165e+00, 2.22619902286992e+00, 2.22505796604477e+00, 2.22346021198236e+00, 2.22129554841030e+00, 2.21844511924162e+00, 2.21478413486456e+00, 2.21018513697240e+00, 2.20452173929155e+00, 2.19767273373253e+00, 2.18952642031151e+00, 2.17998499148342e+00, 2.16896878038024e+00, 2.15642017083208e+00, 2.14230696737097e+00, 2.12662503708741e+00, 2.10940006234533e+00, 2.09068828267061e+00, 2.07057615306503e+00, 2.04917890113454e+00, 2.02663802295092e+00, 2.00311781381100e+00, 1.97880108179284e+00, 1.95388423653593e+00, 1.92857198062509e+00, 1.90307185400769e+00, 1.87758889055613e+00, 1.85232063780563e+00, 1.82745276438972e+00, 1.80315543483448e+00, 1.77958057088939e+00, 1.75686004821689e+00, 1.73510480522610e+00, 1.71440477623898e+00, 1.69482951205485e+00, 1.67642932245482e+00, 1.65923676846420e+00, 1.64326834468072e+00, 1.62852621836757e+00, 1.61499992585135e+00, 1.60266796196038e+00, 1.59149923009106e+00, 1.58145434613290e+00, 1.57248680780050e+00, 1.56454405211315e+00, 1.55756842885778e+00, 1.55149811822766e+00, 1.54626801781956e+00, 1.54181061897937e+00, 1.53805688604075e+00, 1.53493714498880e+00, 1.53238198099595e+00, 1.53032313749715e+00, 1.52869440330390e+00, 1.52743246899023e+00, 1.52647772969289e+00, 1.52577500881803e+00, 1.52527417615519e+00, 1.52493063471816e+00, 1.52470565329730e+00, 1.52456652612491e+00, 1.52448654698482e+00, 1.52444479216483e+00, 1.52442571438493e+00, 1.52441855770641e+00, 1.52441661092156e+00, 1.52441632358037e+00, 1.52441630166213e+00, 1.52441591124719e+00, 1.52441326613142e+00, 1.52440354244924e+00, 1.52437762301390e+00, 1.52432090064495e+00, 1.52421227696402e+00, 1.52402339962178e+00, 1.52371817769290e+00, 1.52325261105234e+00, 1.52257496442830e+00, 1.52162630986811e+00, 1.52034145186116e+00, 1.51865023672767e+00, 1.51647923175409e+00, 1.51375374001391e+00, 1.51040009453608e+00, 1.50634815184742e+00, 1.50153388202945e+00, 1.49590193299761e+00, 1.48940803378683e+00, 1.48202109817484e+00, 1.47372489833438e+00, 1.46451919950825e+00, 1.45442028035481e+00, 1.44346080701726e+00, 1.43168907758182e+00, 1.41916770144340e+00, 1.40597181875478e+00, 1.39218699294675e+00, 1.37790692064041e+00, 1.36323109741016e+00, 1.34826255719780e+00, 1.33310577267308e+00, 1.31786476973366e+00, 1.30264147769007e+00, 1.28753431199944e+00, 1.27263697091442e+00, 1.25803742097888e+00, 1.24381704702625e+00, 1.23004994742671e+00, 1.21680236197162e+00, 1.20413222579515e+00, 1.19208884683952e+00, 1.18071270619718e+00, 1.17003538050081e+00, 1.16007958403740e+00, 1.15085932617814e+00, 1.14238017767011e+00, 1.13463963774426e+00, 1.12762759303152e+00, 1.12132685892273e+00, 1.11571379409330e+00, 1.11075897920547e+00, 1.10642795105464e+00, 1.10268198343199e+00, 1.09947890559000e+00, 1.09677394835936e+00, 1.09452060669527e+00, 1.09267150583266e+00, 1.09117925646586e+00, 1.08999728265486e+00, 1.08908060473074e+00, 1.08838655856536e+00, 1.08787543239197e+00, 1.08751100307484e+00, 1.08726095542082e+00, 1.08709717080876e+00, 1.08699587501616e+00, 1.08693763948081e+00, 1.08690723513416e+00, 1.08689334310303e+00, 1.08688813171703e+00, 1.08688671409317e+00, 1.08688650485485e+00, 1.08688648878160e+00, 1.08688620100341e+00, 1.08688425126411e+00, 1.08687708381872e+00, 1.08685797806083e+00, 1.08681616567814e+00, 1.08673609035081e+00, 1.08659683989690e+00, 1.08637177739487e+00, 1.08602839387528e+00, 1.08552840054054e+00, 1.08482807295196e+00, 1.08387885302473e+00, 1.08262820683841e+00, 1.08102072715549e+00, 1.07899945925722e+00, 1.07650741757833e+00, 1.07348924921533e+00, 1.06989298948767e+00, 1.06567184531224e+00, 1.06078593525109e+00, 1.05520391168243e+00, 1.04890439138093e+00, 1.04187712626857e+00, 1.03412385614463e+00, 1.02565879928370e+00, 1.01650875394120e+00, 1.00671280277419e+00, 9.96321631589893e-01, 9.85396492316369e-01, 9.74007856445311e-01, 9.62233818442684e-01, 9.50158318037822e-01, 9.37869255437676e-01, 9.25456574213712e-01, 9.13010383015276e-01, 9.00619179831580e-01, 8.88368232001270e-01, 8.76338152535865e-01, 8.64603699701192e-01, 8.53232813316822e-01, 8.42285888895574e-01, 8.31815280334666e-01, 8.21865013877996e-01, 8.12470690684520e-01, 8.03659552472070e-01, 7.95450684066267e-01, 7.87855327830428e-01, 7.80877287390183e-01, 7.74513401291273e-01, 7.68754070784553e-01, 7.63583829423737e-01, 7.58981945284816e-01, 7.54923049143670e-01, 7.51377783737709e-01, 7.48313470221817e-01, 7.45694788107556e-01, 7.43484464410855e-01, 7.41643966545496e-01, 7.40134191852206e-01, 7.38916144748529e-01, 7.37951590553038e-01, 7.37203673313559e-01, 7.36637483685470e-01, 7.36220562263028e-01, 7.35923323918428e-01, 7.35719389742712e-01, 7.35585815125106e-01, 7.35503205291536e-01, 7.35455713113396e-01, 7.35430917995558e-01, 7.35419588919527e-01, 7.35415338994912e-01, 7.35414182912635e-01, 7.35414012277332e-01, 7.35413999069465e-01, 7.35413761291085e-01, 7.35412150308207e-01, 7.35406228172351e-01, 7.35390441990830e-01, 7.35355894614813e-01, 7.35289733391610e-01, 7.35174682020869e-01, 7.34988737955391e-01, 7.34705054740090e-01, 7.34292025039991e-01, 7.33713575669896e-01, 7.32929680503901e-01, 7.31897090536776e-01, 7.30570272524858e-01, 7.28902538620574e-01, 7.26847339499727e-01, 7.24359683145295e-01, 7.21397631391611e-01, 7.17923817422076e-01, 7.13906920625632e-01, 7.09323031521023e-01, 7.04156839682611e-01, 6.98402582297803e-01, 6.92064700306787e-01, 6.85158162691522e-01, 6.77708436566007e-01, 6.69751099997328e-01, 6.61331114369844e-01, 6.52501791866030e-01, 6.43323509642997e-01, 6.33862234192577e-01, 6.24187926311777e-01, 6.14372898768409e-01, 6.04490195383491e-01, 5.94612052622935e-01, 5.84808494000114e-01, 5.75146094922033e-01, 5.65686942339411e-01, 5.56487800821202e-01, 5.47599485344762e-01, 5.39066431766926e-01, 5.30926448918805e-01, 5.23210631588512e-01, 5.15943411172245e-01, 5.09142720191520e-01, 5.02820247823436e-01, 4.96981765677688e-01, 4.91627505882917e-01, 4.86752576757809e-01, 4.82347404626694e-01, 4.78398193430991e-01, 4.74887396486949e-01, 4.71794196895573e-01, 4.69094994626417e-01, 4.66763899132530e-01, 4.64773226509038e-01, 4.63093999734213e-01, 4.61696449523259e-01, 4.60550511908437e-01, 4.59626317000414e-01, 4.58894661667344e-01, 4.58327457287481e-01, 4.57898142477575e-01, 4.57582049945529e-01, 4.57356716494829e-01, 4.57202125804301e-01, 4.57100874940423e-01, 4.57038257591506e-01, 4.57002259632787e-01, 4.56983465685951e-01, 4.56974878629325e-01, 4.56971657336496e-01, 4.56970781067554e-01, 4.56970651732135e-01, 4.56970641639856e-01, 4.56970458902774e-01, 4.56969220832761e-01, 4.56964669567043e-01, 4.56952537645563e-01, 4.56925987647779e-01, 4.56875142818319e-01, 4.56786728216940e-01, 4.56643840096177e-01, 4.56425859045879e-01, 4.56108519793958e-01, 4.55664147226116e-01, 4.55062063976840e-01, 4.54269169678895e-01, 4.53250685553840e-01, 4.51971050531697e-01, 4.50394946732928e-01, 4.48488423360203e-01, 4.46220079465235e-01, 4.43562258471886e-01, 4.40492201640438e-01, 4.36993104723001e-01, 4.33055022617538e-01, 4.28675571335873e-01, 4.23860385129259e-01, 4.18623298789825e-01, 4.12986240137215e-01, 4.06978834300446e-01, 4.00637738168768e-01, 3.94005738802790e-01, 3.87130662302283e-01, 3.80064148561445e-01, 3.72860351892527e-01, 3.65574627541946e-01, 3.58262260003670e-01, 3.50977281497379e-01, 3.43771419010594e-01, 3.36693197020722e-01, 3.29787211491222e-01, 3.23093579918151e-01, 3.16647562812264e-01, 3.10479344504557e-01, 3.04613955777370e-01, 2.99071317544673e-01, 2.93866383455764e-01, 2.89009359581925e-01, 2.84505980906530e-01, 2.80357826797143e-01, 2.76562660633275e-01, 2.73114781973732e-01, 2.70005382806616e-01, 2.67222902322203e-01, 2.64753377127462e-01, 2.62580785778923e-01, 2.60687387883777e-01, 2.59054058792202e-01, 2.57660621088803e-01, 2.56486173741109e-01, 2.55509418952215e-01, 2.54708985602761e-01, 2.54063746781313e-01, 2.53553127435737e-01, 2.53157396788225e-01, 2.52857938992506e-01, 2.52637494711578e-01, 2.52480365968594e-01, 2.52372576843371e-01, 2.52301983377251e-01, 2.52258327382421e-01, 2.52233230653346e-01, 2.52220128227677e-01, 2.52214141696228e-01, 2.52211895953735e-01, 2.52211285058465e-01, 2.52211194891643e-01, 2.52211187795922e-01, 2.52211058548824e-01, 2.52210182881512e-01, 2.52206963848096e-01, 2.52198383181465e-01, 2.52179605082162e-01, 2.52143644613753e-01, 2.52081114809741e-01, 2.51980065466228e-01, 2.51825925237824e-01, 2.51601556930538e-01, 2.51287433615673e-01, 2.50861940181082e-01, 2.50301801029664e-01, 2.49582629737972e-01, 2.48679590631752e-01, 2.47568155599510e-01, 2.46224932415217e-01, 2.44628533911808e-01, 2.42760451224186e-01, 2.40605889769284e-01, 2.38154524398965e-01, 2.35401130875829e-01, 2.32346054854233e-01, 2.28995486932448e-01, 2.25361522710926e-01, 2.21461999383180e-01, 2.17320114115722e-01, 2.12963843063122e-01, 2.08425192010992e-01, 2.03739319197396e-01, 1.98943576998175e-01, 1.94076521452772e-01, 1.89176937086685e-01, 1.84282919598473e-01, 1.79431051485892e-01, 1.74655696535069e-01, 1.69988429287031e-01, 1.65457606041912e-01, 1.61088075409396e-01, 1.56901019386410e-01, 1.52913910719050e-01, 1.49140568952930e-01, 1.45591295977482e-01, 1.42273071791436e-01, 1.39189792349981e-01, 1.36342533366090e-01, 1.33729826505459e-01, 1.31347937247895e-01, 1.29191136544932e-01, 1.27251961094316e-01, 1.25521459440264e-01, 1.23989423098124e-01, 1.22644603442268e-01, 1.21474916158064e-01, 1.20467635649248e-01, 1.19609581930405e-01, 1.18887302265279e-01, 1.18287249195944e-01, 1.17795955720169e-01, 1.17400207307142e-01, 1.17087209296884e-01, 1.16844747113433e-01, 1.16661335746784e-01, 1.16526354220213e-01, 1.16430160342304e-01, 1.16364181002183e-01, 1.16320973622706e-01, 1.16294255130981e-01, 1.16278895884861e-01, 1.16270877332013e-01, 1.16267213670682e-01, 1.16265839318075e-01, 1.16265465462064e-01, 1.16265410281770e-01, 1.16265405900862e-01, 1.16265325615487e-01, 1.16264781671148e-01, 1.16262782086309e-01, 1.16257452020861e-01, 1.16245787777601e-01, 1.16223451174989e-01, 1.16184613326411e-01, 1.16121856315687e-01, 1.16026140580260e-01, 1.15886844884628e-01, 1.15691884432493e-01, 1.15427910764296e-01, 1.15080594478030e-01, 1.14634988408596e-01, 1.14075964719163e-01, 1.13388714530885e-01, 1.12559293528605e-01, 1.11575191855010e-01, 1.10425902092118e-01, 1.09103455835609e-01, 1.07602897893045e-01, 1.05922667962716e-01, 1.04064863053694e-01, 1.02035359875444e-01, 9.98437846128724e-02, 9.75033272514736e-02, 9.50304080135425e-02, 9.24442134702019e-02, 8.97661284770237e-02, 8.70190963830453e-02, 8.42269434078560e-02, 8.14137034535349e-02, 7.86029770717676e-02, 7.58173533175039e-02, 7.30779164981072e-02, 7.04038521886474e-02, 6.78121591391374e-02, 6.53174665463183e-02, 6.29319501245220e-02, 6.06653358052256e-02, 5.85249768504212e-02, 5.65159886332083e-02, 5.46414251507201e-02, 5.29024822226205e-02, 5.12987140034511e-02, 4.98282515991519e-02, 4.84880149759590e-02, 4.72739117710148e-02, 4.61810189017484e-02, 4.52037449078944e-02, 4.43359726729398e-02, 4.35711835230705e-02, 4.29025646850505e-02, 4.23231027154287e-02, 4.18256658192036e-02, 4.14030779884911e-02, 4.10481876494941e-02, 4.07539330485491e-02, 4.05134059857772e-02, 4.03199147701798e-02, 4.01670464834041e-02, 4.00487278643737e-02, 3.99592834247642e-02, 3.98934888375635e-02, 3.98466172544119e-02, 3.98144760422383e-02, 3.97934314989925e-02, 3.97804194128923e-02, 3.97729398392102e-02, 3.97690351420437e-02, 3.97672511254257e-02, 3.97665818908308e-02, 3.97663998438844e-02, 3.97663729741708e-02, 3.97663708214111e-02, 3.97663311231895e-02, 3.97660621625704e-02, 3.97650734450536e-02, 3.97624379648676e-02, 3.97566706753025e-02, 3.97456271369236e-02, 3.97264270592501e-02, 3.96954073217302e-02, 3.96481088734220e-02, 3.95793015763104e-02, 3.94830504612145e-02, 3.93528258858424e-02, 3.91816586144529e-02, 3.89623388091422e-02, 3.86876553352948e-02, 3.83506687290236e-02, 3.79450078472843e-02, 3.74651769280518e-02, 3.69068569165165e-02, 3.62671829003318e-02, 3.55449787649178e-02, 3.47409310589768e-02, 3.38576867239766e-02, 3.28998637342579e-02, 3.18739695215697e-02, 3.07882287811481e-02, 2.96523291751214e-02, 2.84770997836006e-02, 2.72741421848934e-02, 2.60554372125910e-02, 2.48329514348740e-02, 2.36182661942763e-02, 2.24222488799768e-02, 2.12547814491302e-02, 2.01245556900744e-02, 1.90389389980510e-02, 1.80039091407878e-02, 1.70240521030536e-02, 1.61026138436142e-02, 1.52415947302416e-02, 1.44418745757946e-02, 1.37033564927549e-02, 1.30251189437018e-02, 1.24055671350334e-02, 1.18425769746437e-02, 1.13336267106945e-02, 1.08759129027760e-02, 1.04664488763433e-02, 1.01021453935886e-02, 9.77987457858376e-03, 9.49651891643555e-03, 9.24900751343216e-03, 9.03434189473273e-03, 8.84961353699509e-03, 8.69201521550167e-03, 8.55884813799435e-03, 8.44752667205739e-03, 8.35558218824417e-03, 8.28066715941134e-03, 8.22056024271944e-03, 8.17317267159012e-03, 8.13655592267537e-03, 8.10891030131700e-03, 8.08859382132709e-03, 8.07413055413519e-03, 8.06421750347397e-03, 8.05772903875379e-03, 8.05371798956855e-03, 8.05141266062037e-03, 8.05020925275812e-03, 8.04965944722369e-03, 8.04945320288653e-03, 8.04939710005215e-03, 8.04938881942974e-03, 8.04938814972900e-03, 8.04937572225128e-03, 8.04929152471345e-03, 8.04898201130300e-03, 8.04815700595896e-03, 8.04635172410428e-03, 8.04289524448620e-03, 8.03688706002958e-03, 8.02718338211150e-03, 8.01239490360217e-03, 7.99089769344619e-03, 7.96085873697905e-03, 7.92027728675748e-03, 7.86704258716497e-03, 7.79900764659658e-03, 7.71407755517828e-03, 7.61030943915841e-03, 7.48601962166617e-03, 7.33989209498986e-03, 7.17108119793863e-03, 6.97930061251384e-03, 6.76489062651342e-03, 6.52885622552901e-03, 6.27287003302968e-03, 5.99923605459196e-03, 5.71081236701332e-03, 5.41089531007799e-03, 5.10307813725319e-03, 4.79110876734978e-03, 4.47876490842702e-03, 4.16973402981569e-03, 3.86745959175983e-03, 3.57492916633362e-03, 3.29447169694674e-03, 3.02772128306117e-03, 2.77576361133924e-03, 2.53930687155446e-03, 2.31878213386949e-03, 2.11437374107959e-03, 1.92600544818333e-03, 1.75331564072678e-03, 1.59563638158033e-03, 1.45187118775425e-03, 1.32053488816553e-03, 1.20019949298467e-03, 1.08959571964817e-03, 9.87754046377631e-04, 8.93855325514892e-04, 8.06791350939307e-04, 7.25823457841557e-04, 6.50455606546529e-04, 5.80201253697510e-04, 5.14653200170301e-04, 4.53487645941664e-04, 3.97210565992623e-04, 3.45042022857456e-04, 2.97003999108369e-04, 2.53062679362870e-04, 2.13186151128752e-04, 1.77328728938164e-04, 1.45422125612527e-04, 1.17408902738703e-04, 9.31216404609359e-05, 7.23262388504509e-05, 5.48188401694460e-05, 4.03675717421947e-05, 2.87128476230691e-05, 1.95719783846471e-05, 1.26444812593718e-05, 7.61781575420173e-06, 4.17346272740815e-06, 1.99337540607303e-06, 7.66969193932709e-07, -1.42108547152020e-14, -2.84217094304040e-14],
[ 7.20000000000000e+01, 7.19999999215939e+01, 7.19999974971211e+01, 7.19999810709336e+01, 7.19999206861163e+01, 7.19997597191786e+01, 7.19994074308106e+01, 7.19987326945126e+01, 7.19975591286485e+01, 7.19956618139955e+01, 7.19927657297578e+01, 7.19885459883226e+01, 7.19826298951233e+01, 7.19746008074068e+01, 7.19640037170166e+01, 7.19503524397006e+01, 7.19331382584475e+01, 7.19118398414018e+01, 7.18859342351598e+01, 7.18549087194609e+01, 7.18182732960750e+01, 7.17755735691132e+01, 7.17264037524695e+01, 7.16704195105657e+01, 7.16073503014809e+01, 7.15370108506528e+01, 7.14593113456823e+01, 7.13742659179306e+01, 7.12819989750181e+01, 7.11827489791829e+01, 7.10768693353028e+01, 7.09648261594354e+01, 7.08471928379811e+01, 7.07246414473243e+01, 7.05979312683516e+01, 7.04678947824996e+01, 7.03354216603978e+01, 7.02014413392474e+01, 7.00669048250128e+01, 6.99327663506409e+01, 6.97999654776091e+01, 6.96694101547092e+01, 6.95419611564766e+01, 6.94184182252768e+01, 6.92995081452576e+01, 6.91858748899900e+01, 6.90780719125385e+01, 6.89765565881218e+01, 6.88816867745472e+01, 6.87937194219888e+01, 6.87128111385729e+01, 6.86390205987514e+01, 6.85723126650660e+01, 6.85125640787325e+01, 6.84595705592640e+01, 6.84130551375274e+01, 6.83726775301749e+01, 6.83380443467580e+01, 6.83087199048112e+01, 6.82842374138323e+01, 6.82641102775608e+01, 6.82478432565086e+01, 6.82349432305146e+01, 6.82249293052141e+01, 6.82173420175682e+01, 6.82117514144610e+01, 6.82077638049707e+01, 6.82050270209189e+01, 6.82032340609369e+01, 6.82021250393724e+01, 6.82014874113453e+01, 6.82011544973273e+01, 6.82010023828005e+01, 6.82009453188460e+01, 6.82009297960155e+01, 6.82009275048763e+01, 6.82009272043272e+01, 6.82009202018930e+01, 6.82008727593536e+01, 6.82006983553011e+01, 6.82002334561381e+01, 6.81992160255852e+01, 6.81972674856427e+01, 6.81938788372147e+01, 6.81884015554494e+01, 6.81800437675484e+01, 6.81678720985412e+01, 6.81508194319088e+01, 6.81276986754823e+01, 6.80972224478173e+01, 6.80580284063621e+01, 6.80087097278850e+01, 6.79478500274244e+01, 6.78740617706210e+01, 6.77860270050869e+01, 6.76825390232830e+01, 6.75625433914492e+01, 6.74251766614473e+01, 6.72698010538987e+01, 6.70960334899910e+01, 6.69037675753633e+01, 6.66931875037303e+01, 6.64647733248748e+01, 6.62192975566763e+01, 6.59578136370380e+01, 6.56816371265872e+01, 6.53923208224429e+01, 6.50916250028959e+01, 6.47814839206249e+01, 6.44639694737030e+01, 6.41412528090688e+01, 6.38155645405823e+01, 6.34891543335411e+01, 6.31642507907012e+01, 6.28430227801333e+01, 6.25275434555374e+01, 6.22197581453979e+01, 6.19214570106125e+01, 6.16342529546609e+01, 6.13595648306851e+01, 6.10986056384300e+01, 6.08523752055854e+01, 6.06216568047006e+01, 6.04070172235114e+01, 6.02088099216424e+01, 6.00271810182247e+01, 5.98620779330664e+01, 5.97132605400516e+01, 5.95803146907552e+01, 5.94626679397956e+01, 5.93596072620529e+01, 5.92702985037249e+01, 5.91938072594127e+01, 5.91291208190445e+01, 5.90751707833815e+01, 5.90308559068738e+01, 5.89950646938067e+01, 5.89666972504971e+01, 5.89446858855416e+01, 5.89280139545285e+01, 5.89157324675279e+01, 5.89069740185945e+01, 5.89009636568794e+01, 5.88970263978599e+01, 5.88945911684114e+01, 5.88931910874252e+01, 5.88924600998621e+01, 5.88921261012428e+01, 5.88920008061237e+01, 5.88919667227467e+01, 5.88919616921092e+01, 5.88919612479997e+01, 5.88919525458920e+01, 5.88918935878188e+01, 5.88916768515144e+01, 5.88910991106351e+01, 5.88898347314096e+01, 5.88874132654067e+01, 5.88832022263961e+01, 5.88763958278193e+01, 5.88660103282498e+01, 5.88508864830365e+01, 5.88296994275940e+01, 5.88009761178339e+01, 5.87631202234148e+01, 5.87144441097757e+01, 5.86532072593403e+01, 5.85776601802220e+01, 5.84860925477189e+01, 5.83768840412273e+01, 5.82485561027083e+01, 5.80998226796605e+01, 5.79296379498529e+01, 5.77372390728321e+01, 5.75221821775177e+01, 5.72843700638758e+01, 5.70240704438160e+01, 5.67419239389294e+01, 5.64389414594642e+01, 5.61164909902087e+01, 5.57762741994994e+01, 5.54202936718762e+01, 5.50508119453207e+01, 5.46703038981461e+01, 5.42814043456512e+01, 5.38868529250136e+01, 5.34894384218796e+01, 5.30919445963350e+01, 5.26970993042476e+01, 5.23075283209690e+01, 5.19257148203915e+01, 5.15539650123432e+01, 5.11943800534931e+01, 5.08488340569840e+01, 5.05189578435185e+01, 5.02061279893534e+01, 4.99114607094059e+01, 4.96358101379165e+01, 4.93797706104524e+01, 4.91436825929480e+01, 4.89276419375403e+01, 4.87315121687987e+01, 4.85549395183555e+01, 4.83973704324970e+01, 4.82580712768617e+01, 4.81361499549056e+01, 4.80305791415156e+01, 4.79402208094401e+01, 4.78638516942487e+01, 4.78001893047406e+01, 4.77479180429598e+01, 4.77057149553428e+01, 4.76722745991111e+01, 4.76463324812727e+01, 4.76266865168623e+01, 4.76122159628810e+01, 4.76018973180825e+01, 4.75948167378624e+01, 4.75901785976191e+01, 4.75873099446234e+01, 4.75856607032970e+01, 4.75847996359809e+01, 4.75844062038528e+01, 4.75842586133352e+01, 4.75842184650846e+01, 4.75842125392851e+01, 4.75842120639626e+01, 4.75842032918528e+01, 4.75841438595104e+01, 4.75839253798144e+01, 4.75833429922518e+01, 4.75820684467075e+01, 4.75796275218553e+01, 4.75753826777879e+01, 4.75685217328373e+01, 4.75580532273455e+01, 4.75428089897540e+01, 4.75214542488723e+01, 4.74925054365038e+01, 4.74543555938225e+01, 4.74053070326201e+01, 4.73436106123813e+01, 4.72675106847707e+01, 4.71752944428951e+01, 4.70653441134499e+01, 4.69361901696641e+01, 4.67865635478373e+01, 4.66154447450497e+01, 4.64221076802349e+01, 4.62061563268201e+01, 4.59675523734599e+01, 4.57066325296825e+01, 4.54241145456062e+01, 4.51210915330636e+01, 4.47990147304749e+01, 4.44596654157886e+01, 4.41051172089998e+01, 4.37376904813733e+01, 4.33599009586915e+01, 4.29744048236304e+01, 4.25839426499513e+01, 4.21912843274994e+01, 4.17991767910403e+01, 4.14102959156518e+01, 4.10272034744605e+01, 4.06523096493087e+01, 4.02878412859443e+01, 3.99358158953803e+01, 3.95980212939487e+01, 3.92760007072354e+01, 3.89710431062059e+01, 3.86841784829247e+01, 3.84161777097529e+01, 3.81675565701234e+01, 3.79385835121853e+01, 3.77292906654166e+01, 3.75394876746861e+01, 3.73687779403842e+01, 3.72165768978532e+01, 3.70821320143284e+01, 3.69645442181204e+01, 3.68627904965204e+01, 3.67757474026625e+01, 3.67022151970758e+01, 3.66409423192806e+01, 3.65906498429399e+01, 3.65500555206713e+01, 3.65178969783658e+01, 3.64929535807073e+01, 3.64740664661029e+01, 3.64601562460165e+01, 3.64502378848409e+01, 3.64434323241229e+01, 3.64389744891527e+01, 3.64362174143821e+01, 3.64346323424847e+01, 3.64338047840755e+01, 3.64334266639695e+01, 3.64332848177669e+01, 3.64332462321326e+01, 3.64332405369728e+01, 3.64332400977072e+01, 3.64332322099083e+01, 3.64331787689087e+01, 3.64329823141227e+01, 3.64324586376491e+01, 3.64313125837732e+01, 3.64291177507308e+01, 3.64253009090076e+01, 3.64191318548804e+01, 3.64097193068758e+01, 3.63960133233317e+01, 3.63768145671232e+01, 3.63507905645355e+01, 3.63164988959344e+01, 3.62724170154752e+01, 3.62169781284693e+01, 3.61486122658461e+01, 3.60657913983629e+01, 3.59670771469329e+01, 3.58511693919762e+01, 3.57169538887181e+01, 3.55635468813779e+01, 3.53903346986052e+01, 3.51970064206566e+01, 3.49835779423551e+01, 3.47504061111865e+01, 3.44981920820505e+01, 3.42279735732165e+01, 3.39411062958030e+01, 3.36392354172563e+01, 3.33242584583958e+01, 3.29982814636796e+01, 3.26635705813319e+01, 3.23225013129234e+01, 3.19775076300236e+01, 3.16310329221163e+01, 3.12854843726006e+01, 3.09431919141119e+01, 3.06063724541164e+01, 3.02770996455106e+01, 2.99572791477156e+01, 2.96486291019309e+01, 2.93526654272246e+01, 2.90706915115016e+01, 2.88037918932908e+01, 2.85528295767094e+01, 2.83184466696571e+01, 2.81010680711117e+01, 2.79009079538939e+01, 2.77179787976956e+01, 2.75521027294960e+01, 2.74029249300504e+01, 2.72699288685964e+01, 2.71524531329027e+01, 2.70497096256414e+01, 2.69608028972261e+01, 2.68847503765648e+01, 2.68205032428668e+01, 2.67669676538635e+01, 2.67230260106479e+01, 2.66875579005803e+01, 2.66594603223152e+01, 2.66376667665128e+01, 2.66211647076941e+01, 2.66090110618080e+01, 2.66003451839937e+01, 2.65943990237154e+01, 2.65905041199867e+01, 2.65880952058844e+01, 2.65867102952363e+01, 2.65859872400346e+01, 2.65856568685349e+01, 2.65855329345198e+01, 2.65854992214382e+01, 2.65854942454571e+01, 2.65854938686908e+01, 2.65854871943137e+01, 2.65854419744239e+01, 2.65852757412813e+01, 2.65848326244152e+01, 2.65838628722640e+01, 2.65820056742606e+01, 2.65787759706325e+01, 2.65735558553191e+01, 2.65655910823648e+01, 2.65539930755687e+01, 2.65377467117494e+01, 2.65157239962747e+01, 2.64867035729625e+01, 2.64493958085970e+01, 2.64024729677617e+01, 2.63446037531726e+01, 2.62744912413401e+01, 2.61909130086304e+01, 2.60927620373534e+01, 2.59790868355747e+01, 2.58491291172218e+01, 2.57023573863293e+01, 2.55384948604822e+01, 2.53575403555234e+01, 2.51597810303922e+01, 2.49457962447725e+01, 2.47164521954695e+01, 2.44728874492557e+01, 2.42164899563975e+01, 2.39488665814181e+01, 2.36718065899072e+01, 2.33872408392701e+01, 2.30971985926572e+01, 2.28037638753665e+01, 2.25090331145478e+01, 2.22150754757683e+01, 2.19238968979971e+01, 2.16374084128208e+01, 2.13573989860932e+01, 2.10855128815670e+01, 2.08232314181579e+01, 2.05718589464370e+01, 2.03325128640591e+01, 2.01061174876969e+01, 1.98934015801290e+01, 1.96948992923386e+01, 1.95109542310129e+01, 1.93417263156614e+01, 1.91872010589038e+01, 1.90472008950862e+01, 1.89213981966625e+01, 1.88093296498955e+01, 1.87104117034469e+01, 1.86239568464641e+01, 1.85491905088850e+01, 1.84852684000157e+01, 1.84312941088411e+01, 1.83863367806821e+01, 1.83494486618997e+01, 1.83196822716292e+01, 1.82961069226301e+01, 1.82778242785909e+01, 1.82639826089039e+01, 1.82537893895969e+01, 1.82465219050523e+01, 1.82415355319080e+01, 1.82382694347061e+01, 1.82362494709954e+01, 1.82350881884212e+01, 1.82344818930962e+01, 1.82342048714887e+01, 1.82341009510770e+01, 1.82340726822012e+01, 1.82340685097742e+01, 1.82340681965488e+01, 1.82340626834265e+01, 1.82340253312157e+01, 1.82338880205963e+01, 1.82335220013495e+01, 1.82327209794685e+01, 1.82311869355126e+01, 1.82285192476886e+01, 1.82242076288894e+01, 1.82176293121193e+01, 1.82080508309271e+01, 1.81946346372909e+01, 1.81764506754343e+01, 1.81524928836831e+01, 1.81217004269286e+01, 1.80829832715341e+01, 1.80352515080193e+01, 1.79774476137961e+01, 1.79085806413646e+01, 1.78277611322872e+01, 1.77342354109613e+01, 1.76274178213428e+01, 1.75069194486361e+01, 1.73725719265766e+01, 1.72244450736869e+01, 1.70628573268259e+01, 1.68883782392347e+01, 1.67018226692356e+01, 1.65042366860526e+01, 1.62968756381735e+01, 1.60811752411550e+01, 1.58587169170646e+01, 1.56311889270674e+01, 1.54003450537109e+01, 1.51679626873315e+01, 1.49358021384801e+01, 1.47055688355750e+01, 1.44788797895605e+01, 1.42572353446320e+01, 1.40419968255425e+01, 1.38343702808126e+01, 1.36353961475455e+01, 1.34459443588366e+01, 1.32667141979430e+01, 1.30982380800604e+01, 1.29408884066133e+01, 1.27948866736731e+01, 1.26603141060247e+01, 1.25371232109462e+01, 1.24251497820860e+01, 1.23241250184681e+01, 1.22336875452965e+01, 1.21533952245700e+01, 1.20827367207128e+01, 1.20211428383495e+01, 1.19679976768112e+01, 1.19226496511441e+01, 1.18844224153830e+01, 1.18526256944039e+01, 1.18265659900417e+01, 1.18055570799678e+01, 1.17889301789644e+01, 1.17760435866486e+01, 1.17662916081497e+01, 1.17591125089492e+01, 1.17539952554059e+01, 1.17504848005334e+01, 1.17481857010247e+01, 1.17467638953713e+01, 1.17459465317085e+01, 1.17455198039047e+01, 1.17453248306592e+01, 1.17452516898215e+01, 1.17452317937740e+01, 1.17452288571603e+01, 1.17452286373782e+01, 1.17452247778940e+01, 1.17451986293388e+01, 1.17451025047546e+01, 1.17448462739849e+01, 1.17442855297649e+01, 1.17432116779375e+01, 1.17413443671394e+01, 1.17383266362077e+01, 1.17337231135029e+01, 1.17270215478724e+01, 1.17176378815809e+01, 1.17049249864482e+01, 1.16881850716575e+01, 1.16666856330925e+01, 1.16396786505359e+01, 1.16064225554388e+01, 1.15662062976491e+01, 1.15183746483886e+01, 1.14623537066607e+01, 1.13976754472934e+01, 1.13240000811883e+01, 1.12411350095866e+01, 1.11490492560319e+01, 1.10478824554235e+01, 1.09379477617519e+01, 1.08197283861615e+01, 1.06938678660558e+01, 1.05611545578692e+01, 1.04225012019278e+01, 1.02789206913147e+01, 1.01314993599533e+01, 9.98136917307222e+00, 9.82968015564823e+00, 9.67757424574367e+00, 9.52616153569819e+00, 9.37649959717721e+00, 9.22957630907151e+00, 9.08629634867938e+00, 8.94747128703504e+00, 8.81381305994502e+00, 8.68593046992694e+00, 8.56432830695007e+00, 8.44940864989957e+00, 8.34147391678962e+00, 8.24073126098090e+00, 8.14729795525870e+00, 8.06120745930148e+00, 7.98241592381456e+00, 7.91080894241028e+00, 7.84620841691580e+00, 7.78837945049767e+00, 7.73703722367986e+00, 7.69185383944209e+00, 7.65246514417498e+00, 7.61847754102831e+00, 7.58947481144267e+00, 7.56502495034108e+00, 7.54468700207738e+00, 7.52801785981451e+00, 7.51457896298660e+00, 7.50394279864503e+00, 7.49569908572424e+00, 7.48946049946545e+00, 7.48486777900489e+00, 7.48159405655704e+00, 7.47934825303248e+00, 7.47787740275513e+00, 7.47696779859987e+00, 7.47644488681351e+00, 7.47617188561807e+00, 7.47604715046018e+00, 7.47600035821338e+00, 7.47598762960823e+00, 7.47598575089353e+00, 7.47598561023405e+00, 7.47598313948053e+00, 7.47596639977291e+00, 7.47590486300361e+00, 7.47574082984967e+00, 7.47538185379869e+00, 7.47469439660098e+00, 7.47349898063308e+00, 7.47156707966300e+00, 7.46861996102005e+00, 7.46432965699363e+00, 7.45832219854843e+00, 7.45018318711591e+00, 7.43946570775364e+00, 7.42570049803847e+00, 7.40840818233652e+00, 7.38711326384316e+00, 7.36135944326809e+00, 7.33072571247146e+00, 7.29484256537391e+00, 7.25340759001345e+00, 7.20619966735391e+00, 7.15309101477406e+00, 7.09405638131688e+00, 7.02917882821117e+00, 6.95865170571606e+00, 6.88277665329687e+00, 6.80195768650140e+00, 6.71669166942550e+00, 6.62755568451379e+00, 6.53519198203645e+00, 6.44029130498635e+00, 6.34357543291082e+00, 6.24577976928855e+00, 6.14763671778383e+00, 6.04986046566627e+00, 5.95313363485681e+00, 5.85809609127401e+00, 5.76533603929801e+00, 5.67538338476120e+00, 5.58870523642557e+00, 5.50570333642359e+00, 5.42671316370603e+00, 5.35200443670321e+00, 5.28178274584397e+00, 5.21619206660421e+00, 5.15531793338658e+00, 5.09919108898049e+00, 5.04779146010814e+00, 5.00105234419277e+00, 4.95886472431846e+00, 4.92108165721832e+00, 4.88752270215838e+00, 4.85797837613276e+00, 4.83221463240152e+00, 4.80997736487605e+00, 4.79099694026590e+00, 4.77499275365648e+00, 4.76167779206446e+00, 4.75076317565620e+00, 4.74196262917648e+00, 4.73499681843575e+00, 4.72959747029751e+00, 4.72551118134385e+00, 4.72250281195710e+00, 4.72035836028604e+00, 4.71888721532788e+00, 4.71792370042054e+00, 4.71732783742115e+00, 4.71698528674634e+00, 4.71680644772430e+00, 4.71672473545179e+00, 4.71669408247929e+00, 4.71668574413907e+00, 4.71668451341793e+00, 4.71668442102934e+00, 4.71668279491286e+00, 4.71667177774103e+00, 4.71663127755073e+00, 4.71652331967240e+00, 4.71628706034912e+00, 4.71583461024477e+00, 4.71504784191525e+00, 4.71377633915556e+00, 4.71183662699576e+00, 4.70901279763972e+00, 4.70505861803595e+00, 4.69970116715980e+00, 4.69264600389538e+00, 4.68358380885050e+00, 4.67219837598222e+00, 4.65817575466738e+00, 4.64121426382015e+00, 4.62103502268228e+00, 4.59739257534814e+00, 4.57008513607884e+00, 4.53896395791648e+00, 4.50394133447519e+00, 4.46499678781116e+00, 4.42218107404014e+00, 4.37561774879521e+00, 4.32550216861920e+00, 4.27209795074498e+00, 4.21573105950300e+00, 4.15678181990893e+00, 4.09567526664554e+00, 4.03287031155264e+00, 3.96884825058611e+00, 3.90410113165849e+00, 3.83912047099460e+00, 3.77438674351683e+00, 3.71035998996759e+00, 3.64747178844945e+00, 3.58611873924893e+00, 3.52665751688927e+00, 3.46940145871481e+00, 3.41461858963382e+00, 3.36253093074549e+00, 3.31331490635992e+00, 3.26710264860374e+00, 3.22398399924329e+00, 3.18400902149551e+00, 3.14719085692032e+00, 3.11350879044363e+00, 3.08291141688470e+00, 3.05531983227988e+00, 3.03063080064203e+00, 3.00871987001040e+00, 2.98944442974432e+00, 2.97264671349910e+00, 2.95815675911672e+00, 2.94579533802360e+00, 2.93537686320637e+00, 2.92671227722070e+00, 2.91961191098579e+00, 2.91388829149957e+00, 2.90935886338191e+00, 2.90584857667025e+00, 2.90319228288537e+00, 2.90123687425356e+00, 2.89984309807276e+00, 2.89888698017008e+00, 2.89826079841994e+00, 2.89787355913974e+00, 2.89765094517431e+00, 2.89753472357172e+00, 2.89748162163505e+00, 2.89746170137430e+00, 2.89745628259064e+00, 2.89745548278992e+00, 2.89745542249426e+00, 2.89745435783368e+00, 2.89744714460733e+00, 2.89742062810878e+00, 2.89734994550871e+00, 2.89719526166333e+00, 2.89689903622434e+00, 2.89638393722313e+00, 2.89555150640712e+00, 2.89428166981212e+00, 2.89243317232942e+00, 2.88984499603554e+00, 2.88633879738647e+00, 2.88172236684166e+00, 2.87579407543970e+00, 2.86834822654555e+00, 2.85918117885048e+00, 2.84809805148574e+00, 2.83491976790670e+00, 2.81949014720526e+00, 2.80168271554585e+00, 2.78140689224763e+00, 2.75861320949140e+00, 2.73329725475770e+00, 2.70550208139257e+00, 2.67531891263440e+00, 2.64288606245012e+00, 2.60838610446004e+00, 2.57204142822353e+00, 2.53410841993139e+00, 2.49487058277548e+00, 2.45463096385029e+00, 2.41370427547990e+00, 2.37240908906672e+00, 2.33106044212292e+00, 2.28996314008158e+00, 2.24940596157613e+00, 2.20965689747285e+00, 2.17095947780619e+00, 2.13353017320399e+00, 2.09755680269699e+00, 2.06319784015666e+00, 2.03058248722608e+00, 1.99981137020451e+00, 1.97095771962148e+00, 1.94406890139653e+00, 1.91916818467449e+00, 1.89625665103377e+00, 1.87531517058456e+00, 1.85630639078312e+00, 1.83917670232681e+00, 1.82385816242096e+00, 1.81027036850686e+00, 1.79832228496220e+00, 1.78791403127715e+00, 1.77893864287518e+00, 1.77128381531979e+00, 1.76483363947793e+00, 1.75947032975959e+00, 1.75507594039885e+00, 1.75153405655794e+00, 1.74873143858119e+00, 1.74655958980112e+00, 1.74491621170122e+00, 1.74370650569442e+00, 1.74284427887446e+00, 1.74225281222485e+00, 1.74186545406458e+00, 1.74162590882597e+00, 1.74148820118461e+00, 1.74141630744035e+00, 1.74138345906400e+00, 1.74137113657986e+00, 1.74136778457257e+00, 1.74136728982359e+00, 1.74136725232228e+00, 1.74136658745724e+00, 1.74136208290273e+00, 1.74134552374332e+00, 1.74130138356115e+00, 1.74120478617138e+00, 1.74101979950095e+00, 1.74069813393729e+00, 1.74017831152288e+00, 1.73938536524075e+00, 1.73823111862619e+00, 1.73661508414911e+00, 1.73442600327650e+00, 1.73154403103358e+00, 1.72784354278623e+00, 1.72319651102789e+00, 1.71747636612028e+00, 1.71056221908722e+00, 1.70234328947345e+00, 1.69272335049455e+00, 1.68162498115190e+00, 1.66899340451087e+00, 1.65479969606947e+00, 1.63904316788708e+00, 1.62175277287413e+00, 1.60298742724993e+00, 1.58283521350062e+00, 1.56141149544570e+00, 1.53885604460261e+00, 1.51532933632019e+00, 1.49100821952645e+00, 1.46608119156616e+00, 1.44074351788987e+00, 1.41519242601476e+00, 1.38962257693586e+00, 1.36422197918520e+00, 1.33916846588112e+00, 1.31462680824104e+00, 1.29074649440696e+00, 1.26766016332886e+00, 1.24548265197610e+00, 1.22431059127015e+00, 1.20422247184266e+00, 1.18527909427384e+00, 1.16752431863532e+00, 1.15098603349034e+00, 1.13567727348827e+00, 1.12159742591585e+00, 1.10873347879993e+00, 1.09706127538013e+00, 1.08654675119953e+00, 1.07714714012279e+00, 1.06881214389452e+00, 1.06148506616277e+00, 1.05510391610382e+00, 1.04960248890259e+00, 1.04491143046832e+00, 1.04095929209080e+00, 1.03767357754801e+00, 1.03498178082023e+00, 1.03281240747796e+00, 1.03109596747910e+00, 1.02976592205161e+00, 1.02875956306360e+00, 1.02801880026649e+00, 1.02749083041846e+00, 1.02712866280933e+00, 1.02689147820701e+00, 1.02674480264551e+00, 1.02666048351701e+00, 1.02661646271216e+00, 1.02659634955775e+00, 1.02658880447009e+00, 1.02658675202814e+00, 1.02658644909216e+00, 1.02658642598435e+00, 1.02658601438051e+00, 1.02658322570676e+00, 1.02657297428782e+00, 1.02654564808155e+00, 1.02648584693507e+00, 1.02637132666875e+00, 1.02617219400014e+00, 1.02585039377030e+00, 1.02535952566531e+00, 1.02464502217576e+00, 1.02364471231586e+00, 1.02228978600174e+00, 1.02050616143190e+00, 1.01821624202259e+00, 1.01534103052113e+00, 1.01180254645119e+00, 1.00752647023387e+00, 1.00244491495405e+00, 9.96499207067671e-01, 9.89642542889186e-01, 9.81842380919943e-01, 9.73082432983475e-01, 9.63364130882312e-01, 9.52707469887756e-01, 9.41151164483429e-01, 9.28752092788926e-01, 9.15584050343654e-01, 9.01735877219878e-01, 8.87309060561918e-01, 8.72414944010259e-01, 8.57171693563586e-01, 8.41701175137416e-01, 8.26125892735078e-01, 8.10566119380454e-01, 7.95137328338839e-01, 7.79948002751581e-01, 7.65097870773133e-01, 7.50676583465108e-01, 7.36762826325119e-01, 7.23423833949397e-01, 7.10715261727145e-01, 6.98681358758620e-01, 6.87355381948578e-01, 6.76760191663817e-01, 6.66908973457510e-01, 6.57806037088363e-01, 6.49447652369005e-01, 6.41822890348465e-01, 6.34914447213689e-01, 6.28699436504931e-01, 6.23150142375252e-01, 6.18234732431517e-01, 6.13917933038735e-01, 6.10161672821846e-01, 6.06925701497843e-01, 6.04168191223479e-01, 6.01846326505253e-01, 5.99916886588389e-01, 5.98336821364356e-01, 5.97063818485609e-01, 5.96056855862230e-01, 5.95276730355550e-01, 5.94686550599903e-01, 5.94252179768063e-01, 5.93942612987902e-01, 5.93730274179606e-01, 5.93591218378762e-01, 5.93505228093377e-01, 5.93455795758217e-01, 5.93429988651323e-01, 5.93418197405484e-01, 5.93413774140032e-01, 5.93412570908157e-01, 5.93412393313784e-01, 5.93412379671349e-01, 5.93412135415122e-01, 5.93410480545046e-01, 5.93404397097871e-01, 5.93388181092919e-01, 5.93352693938670e-01, 5.93284736298898e-01, 5.93166571492731e-01, 5.92975623602641e-01, 5.92684373148685e-01, 5.92260470139337e-01, 5.91667080122235e-01, 5.90863473098921e-01, 5.89805857466004e-01, 5.88448451285302e-01, 5.86744771209567e-01, 5.84649105688953e-01, 5.82118124436107e-01, 5.79112561717338e-01, 5.75598898369094e-01, 5.71550958195132e-01, 5.66951330228619e-01, 5.61792530582551e-01, 5.56077827034315e-01, 5.49821666076639e-01, 5.43049664994555e-01, 5.35798158791195e-01, 5.28113320978093e-01, 5.20049905475446e-01, 5.11669681254915e-01, 5.03039649456539e-01, 4.94230142853638e-01, 4.85312909065613e-01, 4.76359272205841e-01, 4.67438454011855e-01, 4.58616116903045e-01, 4.49953170159901e-01, 4.41504858840617e-01, 4.33320135181248e-01, 4.25441295612487e-01, 4.17903854096593e-01, 4.10736614555987e-01, 4.03961901509277e-01, 3.97595908060453e-01, 3.91649123295807e-01, 3.86126806052005e-01, 3.81029478071781e-01, 3.76353416008470e-01, 3.72091127976120e-01, 3.68231805946081e-01, 3.64761750017024e-01, 3.61664764321375e-01, 3.58922527064934e-01, 3.56514938970449e-01, 3.54420455287169e-01, 3.52616406638006e-01, 3.51079313415113e-01, 3.49785197325176e-01, 3.48709892151703e-01, 3.47829353975257e-01, 3.47119969109002e-01, 3.46558856015406e-01, 3.46124155622334e-01, 3.45795302908485e-01, 3.45553271521439e-01, 3.45380782648903e-01, 3.45262469462554e-01, 3.45184989228940e-01, 3.45137076605084e-01, 3.45109533623841e-01, 3.45095154291300e-01, 3.45088584389259e-01, 3.45086119813814e-01, 3.45085449391576e-01, 3.45085350438723e-01, 3.45085342779385e-01, 3.45085204889727e-01, 3.45084270668409e-01, 3.45080836389769e-01, 3.45071682000921e-01, 3.45051648540291e-01, 3.45013284731309e-01, 3.44946578082997e-01, 3.44838784783491e-01, 3.44674371388692e-01, 3.44435079599918e-01, 3.44100123046488e-01, 3.43646521683098e-01, 3.43049574953369e-01, 3.42283469125590e-01, 3.41322007207083e-01, 3.40139441822686e-01, 3.38711382890395e-01, 3.37015743573488e-01, 3.35033680767410e-01, 3.32750481330692e-01, 3.30156343388609e-01, 3.27247004114227e-01, 3.24024171820980e-01, 3.20495730856678e-01, 3.16675701952988e-01, 3.12583957083518e-01, 3.08245704842008e-01, 3.03690778026507e-01, 2.98952767787881e-01, 2.94068057048705e-01, 2.89074809190083e-01, 2.84011966193944e-01, 2.78918304118150e-01, 2.73831584105352e-01, 2.68787825520690e-01, 2.63820715794125e-01, 2.58961160432875e-01, 2.54236967440278e-01, 2.49672653556829e-01, 2.45289355450495e-01, 2.41104827045007e-01, 2.37133504195683e-01, 2.33386619400278e-01, 2.29872351631045e-01, 2.26595999191801e-01, 2.23560166355753e-01, 2.20764957172520e-01, 2.18208172120086e-01, 2.15885505163584e-01, 2.13790740255718e-01, 2.11915947372233e-01, 2.10251678851364e-01, 2.08787167136165e-01, 2.07510525056307e-01, 2.06408949572094e-01, 2.05468929478627e-01, 2.04676456953777e-01, 2.04017242069526e-01, 2.03476928515300e-01, 2.03041307864808e-01, 2.02696528820610e-01, 2.02429297078439e-01, 2.02227060832698e-01, 2.02078176575014e-01, 2.01972049770916e-01, 2.01899245276877e-01, 2.01851562989376e-01, 2.01822075183131e-01, 2.01805123246714e-01, 2.01796272989114e-01, 2.01792229274631e-01, 2.01790712344433e-01, 2.01790299703262e-01, 2.01790238798324e-01, 2.01790234045774e-01, 2.01790147993634e-01, 2.01789564978071e-01, 2.01787421758070e-01, 2.01781708773055e-01, 2.01769206312306e-01, 2.01745263628254e-01, 2.01703630303243e-01, 2.01636348378614e-01, 2.01533712797271e-01, 2.01384306508089e-01, 2.01175115022508e-01, 2.00891723172376e-01, 2.00518594198726e-01, 2.00039428040441e-01, 1.99437591823838e-01, 1.98696611204340e-01, 1.97800706647755e-01, 1.96735354336710e-01, 1.95487847634098e-01, 1.94047832454601e-01, 1.92407788979764e-01, 1.90563433269730e-01, 1.88514015633743e-01, 1.86262498005206e-01, 1.83815599628431e-01, 1.81183708456047e-01, 1.78380663967928e-01, 1.75423424792982e-01, 1.72331640770309e-01, 1.69127153350900e-01, 1.65833450252791e-01, 1.62475100123274e-01, 1.59077190996143e-01, 1.55664793093351e-01, 1.52262462562248e-01, 1.48893798502797e-01, 1.45581061498405e-01, 1.42344858047792e-01, 1.39203892025805e-01, 1.36174781644641e-01, 1.33271938328747e-01, 1.30507502391069e-01, 1.27891329300581e-01, 1.25431019672661e-01, 1.23131985875759e-01, 1.20997548280968e-01, 1.19029054635121e-01, 1.17226016653262e-01, 1.15586258692630e-01, 1.14106074236645e-01, 1.12780386886470e-01, 1.11602913568035e-01, 1.10566328555890e-01, 1.09662427640799e-01, 1.08882292184347e-01, 1.08216452970296e-01, 1.07655053649736e-01, 1.07188013335758e-01, 1.06805187392624e-01, 1.06496524994483e-01, 1.06252221360634e-01, 1.06062862042378e-01, 1.05919556086675e-01, 1.05814054607848e-01, 1.05738851131719e-01, 1.05687260238184e-01, 1.05653471399094e-01, 1.05632575576593e-01, 1.05620562982452e-01, 1.05614291451218e-01, 1.05611425963318e-01, 1.05610351025049e-01, 1.05610058616293e-01, 1.05610015457202e-01, 1.05610012061589e-01, 1.05609950196595e-01, 1.05609531053531e-01, 1.05607990243570e-01, 1.05603883051344e-01, 1.05594894754887e-01, 1.05577681843243e-01, 1.05547750793690e-01, 1.05499380565732e-01, 1.05425594247265e-01, 1.05318184454021e-01, 1.05167795981387e-01, 1.04964067806364e-01, 1.04695834676960e-01, 1.04351386303676e-01, 1.03918779511844e-01, 1.03386195652106e-01, 1.02742332448329e-01, 1.01976816267154e-01, 1.01080618078569e-01, 1.00046454258603e-01, 9.88691522722007e-02, 9.75459615656575e-02, 9.60767915080538e-02, 9.44643612741487e-02, 9.27142508479193e-02, 9.08348474998206e-02, 8.88371879066057e-02, 8.67347020755460e-02, 8.45428710115499e-02, 8.22788151161831e-02, 7.99608341164060e-02, 7.76079214088270e-02, 7.52392757283502e-02, 7.28738313776915e-02, 7.05298253217279e-02, 6.82244162888566e-02, 6.59733674443999e-02, 6.37908012854922e-02, 6.16890298970105e-02, 5.96575989379033e-02, 5.77493165715026e-02, 5.59468957941789e-02, 5.42551197135879e-02, 5.26770460913184e-02, 5.12141461127698e-02, 4.98664598055569e-02, 4.86327590083420e-02, 4.75107104265646e-02, 4.64970331676113e-02, 4.55876469854672e-02, 4.47778091018449e-02, 4.40622388268821e-02, 4.34352302710863e-02, 4.28907542425065e-02, 4.24225509650995e-02, 4.20242155228436e-02, 4.16892779281142e-02, 4.14112794519639e-02, 4.11838463892735e-02, 4.10007618326205e-02, 4.08560353666871e-02, 4.07439699398964e-02, 4.06592245796418e-02, 4.05968711484803e-02, 4.05524430318991e-02, 4.05219735343252e-02, 4.05020218573782e-02, 4.04896848328491e-02, 4.04825930643966e-02, 4.04788907529934e-02, 4.04771991889987e-02, 4.04765646334937e-02, 4.04763920198121e-02, 4.04763665424355e-02, 4.04763645197193e-02, 4.04763274509037e-02, 4.04760763048273e-02, 4.04751530725918e-02, 4.04726921313028e-02, 4.04673067099708e-02, 4.04569940894248e-02, 4.04390638153700e-02, 4.04100930142022e-02, 4.03659126539111e-02, 4.03016282226218e-02, 4.02116777174371e-02, 4.00899289582952e-02, 3.99298169850653e-02, 3.97245206147119e-02, 3.94671751311080e-02, 3.91511156186795e-02, 3.87701427761726e-02, 3.83188003651185e-02, 3.77926510354740e-02, 3.71885354365418e-02, 3.65047985867761e-02, 3.57414677372674e-02, 3.49003676279978e-02, 3.39851621633187e-02, 3.30013159568523e-02, 3.19559745146734e-02, 3.08577674774000e-02, 2.97165447492347e-02, 2.85430600625887e-02, 2.73486202502937e-02, 2.61447209160082e-02, 2.49426898915175e-02, 2.37533584506195e-02, 2.25867767130268e-02, 2.14519847159664e-02, 2.03568454163019e-02, 1.93079413624275e-02, 1.83105330743132e-02, 1.73685738048732e-02, 1.64847720548238e-02, 1.56606905607788e-02, 1.48968695217064e-02, 1.41929629881190e-02, 1.35478797728439e-02, 1.29599224647450e-02, 1.24269192886572e-02, 1.19463440883436e-02, 1.15154206741437e-02, 1.11312096915555e-02, 1.07906784846108e-02, 1.04907561193528e-02, 1.02283762262090e-02, 1.00005098992710e-02, 9.80419029733071e-03, 9.63653036528456e-03, 9.49473526567601e-03, 9.37611135022109e-03, 9.27807350210230e-03, 9.19815234742316e-03, 9.13400229390504e-03, 9.08341077892771e-03, 9.04430862581762e-03, 9.01478104823639e-03, 8.99307859481269e-03, 8.97762716289208e-03, 8.96703612232841e-03, 8.96010357980970e-03, 8.95581788897459e-03, 8.95335466962877e-03, 8.95206882913158e-03, 8.95148135954571e-03, 8.95126098613730e-03, 8.95120103984937e-03, 8.95119219194385e-03, 8.95119148302115e-03, 8.95117840939008e-03, 8.95108983404214e-03, 8.95076422725083e-03, 8.94989632158172e-03, 8.94799714698991e-03, 8.94436082680272e-03, 8.93803982060604e-03, 8.92783033073385e-03, 8.91226960453650e-03, 8.88964681340099e-03, 8.85802901220245e-03, 8.81530331598412e-03, 8.75923581932625e-03, 8.68754690391427e-03, 8.59800144299072e-03, 8.48851108075621e-03, 8.35724433872542e-03, 8.20273889790712e-03, 8.02400914925272e-03, 7.82064113049330e-03, 7.59286647753754e-03, 7.34160732254452e-03, 7.06848554942496e-03, 6.77579270559647e-03, 6.46642092351613e-03, 6.14375950247847e-03, 5.81156502076396e-03, 5.47381408730985e-03, 5.13454747134290e-03, 4.79771394202544e-03, 4.46702311838010e-03, 4.14581844829343e-03, 3.83698140730360e-03, 3.54287357426131e-03, 3.26531519044693e-03, 3.00559187216720e-03, 2.76448024467868e-03, 2.54228835596848e-03, 2.33891188786117e-03, 2.15390606840060e-03, 1.98656597119395e-03, 1.83600190891298e-03, 1.70188202019972e-03, 1.58175524062187e-03, 1.47510406004869e-03, 1.38080910910787e-03, 1.29783578317699e-03, 1.22511039278095e-03, 1.16156777622223e-03, 1.10626345249898e-03, 1.05835468144733e-03, 1.01708458579708e-03, 9.81776933329570e-04, 9.51791126979629e-04, 9.26483780688159e-04, 9.05263113352817e-04, 8.87619750514546e-04, 8.73109515893589e-04, 8.61332039619356e-04, 8.51916948477083e-04, 8.44518823058138e-04, 8.38818142753439e-04, 8.34523453150382e-04, 8.31372370853956e-04, 8.29131666634453e-04, 8.27597099373634e-04, 8.26593189231062e-04, 8.25972795482244e-04, 8.25616298072873e-04, 8.25430222775481e-04, 8.25345214522599e-04, 8.25313326785704e-04, 8.25304652693148e-04, 8.25303372423036e-04, 8.25303268854327e-04, 8.25301345358298e-04, 8.25288313748729e-04, 8.25240409611183e-04, 8.25112726190014e-04, 8.24833352396581e-04, 8.24298541147073e-04, 8.23369198926116e-04, 8.21869015837251e-04, 8.19584593187983e-04, 8.16267953609895e-04, 8.11641822096476e-04, 8.05407947140679e-04, 7.97258273209422e-04, 7.86887966341965e-04, 7.74009332857872e-04, 7.58367952926164e-04, 7.39763512854097e-04, 7.18074807750213e-04, 6.93288960107452e-04, 6.65542646331119e-04, 6.35110579565890e-04, 6.02215070202305e-04, 5.67084289343711e-04, 5.30115802320097e-04, 4.92075960167426e-04, 4.53821780297403e-04, 4.15554996095580e-04, 3.77771098314383e-04, 3.41093148080063e-04, 3.06554326357400e-04, 2.73829144987303e-04, 2.42998440299402e-04, 2.14966277667372e-04, 1.89529635463259e-04, 1.66527623846946e-04, 1.46565149265143e-04, 1.29030021312815e-04, 1.13314926224461e-04, 9.91496578421902e-05, 8.64183718363165e-05, 7.49737716034815e-05, 6.44883749600922e-05, 5.44852745747448e-05, 4.45520259830801e-05, 3.45889847181979e-05, 2.49232317912629e-05, 1.61763363735190e-05, 8.86839123381833e-06, 3.10841582518151e-06, 2.54464690385703e-06, 2.06941641067715e-06, 6.22482374978972e-07, 1.95029501526278e-06, 2.33367360635839e-06, 2.21490243745848e-06, 1.90160440638465e-06, 1.60282900196762e-06, 1.41323651803305e-06, 1.32425485332988e-06, 1.27590256226995e-06, 1.21138248232455e-06, 1.10432073086031e-06, 9.57443646143474e-07, 7.87969213433826e-07, 6.15531945413750e-07, 4.55940281085532e-07, 3.19154906947006e-07, 1.42108547152020e-14, 2.84217094304040e-14, 2.84217094304040e-14, 2.84217094304040e-14, 2.84217094304040e-14, 2.84217094304040e-14, 2.84217094304040e-14],
[ 7.30000000000000e+01, 7.29999999200099e+01, 7.29999974465562e+01, 7.29999806885151e+01, 7.29999190837597e+01, 7.29997548648455e+01, 7.29993954592543e+01, 7.29987070912719e+01, 7.29975098155699e+01, 7.29955741681792e+01, 7.29926195704313e+01, 7.29883145676835e+01, 7.29822789296529e+01, 7.29740875856159e+01, 7.29632763181601e+01, 7.29493490958837e+01, 7.29317868899389e+01, 7.29100577920776e+01, 7.28836282319405e+01, 7.28519750764087e+01, 7.28145983803451e+01, 7.27710345419133e+01, 7.27208695931544e+01, 7.26637523255108e+01, 7.25994069110560e+01, 7.25276446372925e+01, 7.24483743339301e+01, 7.23616110440633e+01, 7.22674824905353e+01, 7.21662329204823e+01, 7.20582239828000e+01, 7.19439324046070e+01, 7.18239443772744e+01, 7.16989467279080e+01, 7.15697151218823e+01, 7.14370996983687e+01, 7.13020086677297e+01, 7.11653904854409e+01, 7.10282152561202e+01, 7.08914560139527e+01, 7.07560704785386e+01, 7.06229838080883e+01, 7.04930727768201e+01, 7.03671517019802e+01, 7.02459603478080e+01, 7.01301539458959e+01, 7.00202953975624e+01, 6.99168496651027e+01, 6.98201803140010e+01, 6.97305481349686e+01, 6.96481117499855e+01, 6.95729300873700e+01, 6.95049665947194e+01, 6.94440950434112e+01, 6.93901067630059e+01, 6.93427191278070e+01, 6.93015851010169e+01, 6.92663036248844e+01, 6.92364306287969e+01, 6.92114904125016e+01, 6.91909871497713e+01, 6.91744162501148e+01, 6.91612753137841e+01, 6.91510744194228e+01, 6.91433454950555e+01, 6.91376505422517e+01, 6.91335885103103e+01, 6.91308006518889e+01, 6.91289742328526e+01, 6.91278445160407e+01, 6.91271949895593e+01, 6.91268558632603e+01, 6.91267009102420e+01, 6.91266427814634e+01, 6.91266269689741e+01, 6.91266246350819e+01, 6.91266243289245e+01, 6.91266171958236e+01, 6.91265688679989e+01, 6.91263912095460e+01, 6.91259176353637e+01, 6.91248812198323e+01, 6.91228963217240e+01, 6.91194444472004e+01, 6.91138649780921e+01, 6.91053512835979e+01, 6.90929526091874e+01, 6.90755819953279e+01, 6.90520303190841e+01, 6.90209863726251e+01, 6.89810626942715e+01, 6.89308266516515e+01, 6.88688360465867e+01, 6.87936782739676e+01, 6.87040118320333e+01, 6.85986087636768e+01, 6.84763964277713e+01, 6.83364968819554e+01, 6.81782621335307e+01, 6.80013036116475e+01, 6.78055144508080e+01, 6.75910835525017e+01, 6.73585008804657e+01, 6.71085539879207e+01, 6.68423162925891e+01, 6.65611280238541e+01, 6.62665710030216e+01, 6.59604384622290e+01, 6.56447009946192e+01, 6.53214695403271e+01, 6.49929561538770e+01, 6.46614332561703e+01, 6.43291921810838e+01, 6.39985020409527e+01, 6.36715701516583e+01, 6.33505053514827e+01, 6.30372854317388e+01, 6.27337295676944e+01, 6.24414761783357e+01, 6.21619661824705e+01, 6.18964312725392e+01, 6.16458866534967e+01, 6.14111276812980e+01, 6.11927299274453e+01, 6.09910523253285e+01, 6.08062431687046e+01, 6.06382488061167e+01, 6.04868249030123e+01, 6.03515501344127e+01, 6.02318421379565e+01, 6.01269755113825e+01, 6.00361015879199e+01, 5.99582696722137e+01, 5.98924493707532e+01, 5.98375536057939e+01, 5.97924618619629e+01, 5.97560431820640e+01, 5.97271784055460e+01, 5.97047811325015e+01, 5.96878169007521e+01, 5.96753200859583e+01, 5.96664080763655e+01, 5.96602923352136e+01, 5.96562860441224e+01, 5.96538081176248e+01, 5.96523834888848e+01, 5.96516396848365e+01, 5.96512998301805e+01, 5.96511723382470e+01, 5.96511376572821e+01, 5.96511325384418e+01, 5.96511320865457e+01, 5.96511232318628e+01, 5.96510632400706e+01, 5.96508427037018e+01, 5.96502548332174e+01, 5.96489682854530e+01, 5.96465043634876e+01, 5.96422194914547e+01, 5.96352937543952e+01, 5.96247261618423e+01, 5.96093371420328e+01, 5.95877785979195e+01, 5.95585516527587e+01, 5.95200319793633e+01, 5.94705023430979e+01, 5.94081916985032e+01, 5.93313198726306e+01, 5.92381465604905e+01, 5.91270230711294e+01, 5.89964450227426e+01, 5.88451040191667e+01, 5.86719362721032e+01, 5.84761661792060e+01, 5.82573430305302e+01, 5.80153692827108e+01, 5.77505191865887e+01, 5.74634469483676e+01, 5.71551840180822e+01, 5.68271255142617e+01, 5.64810062052584e+01, 5.61188668780068e+01, 5.57430123315333e+01, 5.53559626168691e+01, 5.49603994688768e+01, 5.45591100886152e+01, 5.41549304915494e+01, 5.37506905143283e+01, 5.33491622824093e+01, 5.29530135276258e+01, 5.25647666758745e+01, 5.21867641710949e+01, 5.18211401195342e+01, 5.14697980601063e+01, 5.11343944977212e+01, 5.08163277607054e+01, 5.05167317332026e+01, 5.02364740395330e+01, 4.99761582963074e+01, 4.97361300848104e+01, 4.95164863242627e+01, 4.93170877451745e+01, 4.91375741728009e+01, 4.89773823354557e+01, 4.88357659117548e+01, 4.87118175240830e+01, 4.86044923714134e+01, 4.85126331719629e+01, 4.84349960549351e+01, 4.83702770020897e+01, 4.83171383969620e+01, 4.82742351964757e+01, 4.82402402016653e+01, 4.82138678768975e+01, 4.81938961558396e+01, 4.81791856821672e+01, 4.81686959670040e+01, 4.81614980049713e+01, 4.81567829760785e+01, 4.81538667690179e+01, 4.81521901883168e+01, 4.81513148472354e+01, 4.81509148932724e+01, 4.81507648561839e+01, 4.81507240424061e+01, 4.81507180183762e+01, 4.81507175351744e+01, 4.81507086176517e+01, 4.81506482001151e+01, 4.81504260987419e+01, 4.81498340571231e+01, 4.81485383839422e+01, 4.81460569972491e+01, 4.81417417900492e+01, 4.81347671205092e+01, 4.81241251008956e+01, 4.81086282096270e+01, 4.80869195765493e+01, 4.80574910886077e+01, 4.80187092284671e+01, 4.79688482919519e+01, 4.79061303350607e+01, 4.78287708865623e+01, 4.77350291423685e+01, 4.76232610530772e+01, 4.74919734507059e+01, 4.73398771610437e+01, 4.71659369394390e+01, 4.69694160705307e+01, 4.67499135985119e+01, 4.65073924055914e+01, 4.62421967229624e+01, 4.59550581219206e+01, 4.56470895672663e+01, 4.53197676914150e+01, 4.49749040335055e+01, 4.46146065466575e+01, 4.42412331649025e+01, 4.38573395890282e+01, 4.34656236481345e+01, 4.30688685873852e+01, 4.26698874220042e+01, 4.22714701256367e+01, 4.18763349658510e+01, 4.14870848499523e+01, 4.11061691707134e+01, 4.07358513752815e+01, 4.03781823121256e+01, 4.00349793062162e+01, 3.97078108338774e+01, 3.93979865908488e+01, 3.91065526632016e+01, 3.88342914284145e+01, 3.85817257466024e+01, 3.83491269612003e+01, 3.81365262193456e+01, 3.79437286423849e+01, 3.77703299182149e+01, 3.76157349385397e+01, 3.74791781544762e+01, 3.73597453640118e+01, 3.72563966683971e+01, 3.71679903386899e+01, 3.70933073185216e+01, 3.70310760575009e+01, 3.69799973262563e+01, 3.69387686151047e+01, 3.69061076706292e+01, 3.68807746851329e+01, 3.68615926296721e+01, 3.68474652178153e+01, 3.68373920085477e+01, 3.68304802050450e+01, 3.68259527813189e+01, 3.68231526687004e+01, 3.68215428543302e+01, 3.68207023781250e+01, 3.68203183557724e+01, 3.68201742954335e+01, 3.68201351075004e+01, 3.68201293234425e+01, 3.68201288773203e+01, 3.68201208663976e+01, 3.68200665912163e+01, 3.68198670698929e+01, 3.68193352191790e+01, 3.68181712762913e+01, 3.68159421840850e+01, 3.68120657667449e+01, 3.68058004262210e+01, 3.67962409769819e+01, 3.67823211052328e+01, 3.67628227842787e+01, 3.67363927957586e+01, 3.67015662936651e+01, 3.66567971034973e+01, 3.66004941754754e+01, 3.65310633162288e+01, 3.64469530209427e+01, 3.63467029360773e+01, 3.62289932242581e+01, 3.60926929030483e+01, 3.59369051132827e+01, 3.57610072625910e+01, 3.55646841016745e+01, 3.53479520319796e+01, 3.51111733100048e+01, 3.48550592900690e+01, 3.45806624065473e+01, 3.42893571999526e+01, 3.39828112916310e+01, 3.36629477569300e+01, 3.33319007844397e+01, 3.29919667946545e+01, 3.26455532957378e+01, 3.22951276697469e+01, 3.19431678281719e+01, 3.15921162936731e+01, 3.12443388151337e+01, 3.09020881705342e+01, 3.05674734139003e+01, 3.02424345173617e+01, 2.99287221618303e+01, 2.96278823316558e+01, 2.93412453451933e+01, 2.90699189731323e+01, 2.88147853309325e+01, 2.85765012609802e+01, 2.83555019354909e+01, 2.81520074135163e+01, 2.79660318808912e+01, 2.77973952977650e+01, 2.76457371794051e+01, 2.75105322433600e+01, 2.73911076674625e+01, 2.72866617140387e+01, 2.71962834811266e+01, 2.71189735375743e+01, 2.70536651835813e+01, 2.69992460518488e+01, 2.69545797294554e+01, 2.69185270410581e+01, 2.68899665953937e+01, 2.68678141652098e+01, 2.68510404514326e+01, 2.68386867806017e+01, 2.68298783040934e+01, 2.68238343104539e+01, 2.68198753283454e+01, 2.68174267852286e+01, 2.68160190921555e+01, 2.68152841425687e+01, 2.68149483364477e+01, 2.68148223637220e+01, 2.68147880960620e+01, 2.68147830382263e+01, 2.68147826552622e+01, 2.68147758710920e+01, 2.68147299073381e+01, 2.68145609396764e+01, 2.68141105336178e+01, 2.68131248294076e+01, 2.68112370819174e+01, 2.68079542544008e+01, 2.68026482819415e+01, 2.67945525240097e+01, 2.67827638133404e+01, 2.67662503769618e+01, 2.67438657511092e+01, 2.67143686323499e+01, 2.66764484019793e+01, 2.66287558322718e+01, 2.65699382382014e+01, 2.64986780879993e+01, 2.64137338459182e+01, 2.63139816097039e+01, 2.61984559440253e+01, 2.60663882192078e+01, 2.59172407584585e+01, 2.57507351870272e+01, 2.55668735668541e+01, 2.53659511859352e+01, 2.51485602415092e+01, 2.49155840927581e+01, 2.46681822389351e+01, 2.44077666733285e+01, 2.41359707352017e+01, 2.38546119858072e+01, 2.35656509217422e+01, 2.32711474661886e+01, 2.29732171234147e+01, 2.26739884553654e+01, 2.23755631894901e+01, 2.20799798684613e+01, 2.17891815825693e+01, 2.15049880401712e+01, 2.12290720501367e+01, 2.09629403971030e+01, 2.07079190490851e+01, 2.04651426091428e+01, 2.02355478822900e+01, 2.00198713674096e+01, 1.98186504085342e+01, 1.96322276654129e+01, 1.94607585051997e+01, 1.93042208859622e+01, 1.91624273020148e+01, 1.90350383876780e+01, 1.89215778220230e+01, 1.88214482322906e+01, 1.87339478477449e+01, 1.86582877001550e+01, 1.85936091959660e+01, 1.85390018956277e+01, 1.84935213275899e+01, 1.84562066408378e+01, 1.84260978652620e+01, 1.84022525097124e+01, 1.83837611899890e+01, 1.83697619498816e+01, 1.83594529235261e+01, 1.83521029912779e+01, 1.83470601066912e+01, 1.83437570196963e+01, 1.83417141892220e+01, 1.83405397639777e+01, 1.83399266079952e+01, 1.83396464519252e+01, 1.83395413557052e+01, 1.83395127669846e+01, 1.83395085473493e+01, 1.83395082305800e+01, 1.83395026550803e+01, 1.83394648802554e+01, 1.83393260160851e+01, 1.83389558558215e+01, 1.83381457723371e+01, 1.83365943778166e+01, 1.83338965281403e+01, 1.83295361891179e+01, 1.83228836069444e+01, 1.83131971372198e+01, 1.82996299820417e+01, 1.82812419593915e+01, 1.82570162804763e+01, 1.82258811379598e+01, 1.81867357130294e+01, 1.81384799974785e+01, 1.80800476077988e+01, 1.80104405547182e+01, 1.79287647396964e+01, 1.78342647971774e+01, 1.77263568053018e+01, 1.76046573634947e+01, 1.74690075939596e+01, 1.73194907710920e+01, 1.71564425171324e+01, 1.69804528161986e+01, 1.67923594779011e+01, 1.65932331058551e+01, 1.63843540701857e+01, 1.61671824168296e+01, 1.59433220374228e+01, 1.57144807385499e+01, 1.54824280577901e+01, 1.52489527534149e+01, 1.50158218338534e+01, 1.47847427965476e+01, 1.45573304342167e+01, 1.43350791744700e+01, 1.41193414895974e+01, 1.39113124926534e+01, 1.37120204639026e+01, 1.35223227579788e+01, 1.33429063427059e+01, 1.31742921176275e+01, 1.30168421445332e+01, 1.28707689763793e+01, 1.27361463740130e+01, 1.26129208308491e+01, 1.25009234655244e+01, 1.23998819768815e+01, 1.23094324741653e+01, 1.22291310918499e+01, 1.21584653702046e+01, 1.20968654291934e+01, 1.20437149859355e+01, 1.19983622672019e+01, 1.19601308515102e+01, 1.19283304440146e+01, 1.19022675457185e+01, 1.18812559310189e+01, 1.18646267989239e+01, 1.18517384182386e+01, 1.18419850501743e+01, 1.18348049073035e+01, 1.18296868988099e+01, 1.18261759206295e+01, 1.18238764759742e+01, 1.18224544559100e+01, 1.18216369686516e+01, 1.18212101762234e+01, 1.18210151734285e+01, 1.18209420215023e+01, 1.18209221224381e+01, 1.18209191853791e+01, 1.18209189655637e+01, 1.18209151054943e+01, 1.18208889529741e+01, 1.18207928138120e+01, 1.18205365441665e+01, 1.18199757147847e+01, 1.18189016995431e+01, 1.18170341035636e+01, 1.18140159089921e+01, 1.18094116723618e+01, 1.18027090528475e+01, 1.17933238811442e+01, 1.17806088893831e+01, 1.17638661094119e+01, 1.17423628077816e+01, 1.17153506618128e+01, 1.16820876970819e+01, 1.16418623121005e+01, 1.15940185250083e+01, 1.15379814077806e+01, 1.14732815461593e+01, 1.13995772986102e+01, 1.13166736425648e+01, 1.12245365024537e+01, 1.11233016541046e+01, 1.10132775856646e+01, 1.08949420465755e+01, 1.07689324034462e+01, 1.06360303081484e+01, 1.04971415302243e+01, 1.03532720774367e+01, 1.02055018988193e+01, 1.00549575209493e+01, 9.90278491251762e+00, 9.75012372083258e+00, 9.59808380339890e+00, 9.44772472010442e+00, 9.30003858824709e+00, 9.15593645976266e+00, 9.01623817543592e+00, 8.88166549349280e+00, 8.75283818088661e+00, 8.63027269045527e+00, 8.51438301871679e+00, 8.40548333943954e+00, 8.30379202992191e+00, 8.20943674391053e+00, 8.12246023210631e+00, 8.04282666386301e+00, 7.97042825808639e+00, 7.90509208399794e+00, 7.84658694024694e+00, 7.79463026128890e+00, 7.74889503109574e+00, 7.70901670496013e+00, 7.67460014999030e+00, 7.64522661417440e+00, 7.62046072359625e+00, 7.59985748906959e+00, 7.58296927919463e+00, 7.56935268911050e+00, 7.55857520579568e+00, 7.55022154456525e+00, 7.54389951028578e+00, 7.53924522333880e+00, 7.53592754656090e+00, 7.53365155657094e+00, 7.53216092145379e+00, 7.53123907610694e+00, 7.53070912510350e+00, 7.53043244829546e+00, 7.53030603360405e+00, 7.53025861128754e+00, 7.53024571128636e+00, 7.53024380727391e+00, 7.53024366472039e+00, 7.53024116069699e+00, 7.53022419558044e+00, 7.53016183017333e+00, 7.52999558809265e+00, 7.52963177744043e+00, 7.52893505979698e+00, 7.52772353481076e+00, 7.52576558389126e+00, 7.52277868716359e+00, 7.51843039177503e+00, 7.51234156406817e+00, 7.50409200009402e+00, 7.49322839595968e+00, 7.47927459032586e+00, 7.46174388667507e+00, 7.44015314601366e+00, 7.41403821764011e+00, 7.38297015560909e+00, 7.34657156302757e+00, 7.30453232812278e+00, 7.25662397763173e+00, 7.20271188482712e+00, 7.14276463768768e+00, 7.07685999795750e+00, 7.00518705820329e+00, 6.92804441906328e+00, 6.84583444488567e+00, 6.75905389166201e+00, 6.66828141453772e+00, 6.57416263320258e+00, 6.47739354671458e+00, 6.37870313603655e+00, 6.27883597182004e+00, 6.17853556366506e+00, 6.07852905882915e+00, 5.97951374120890e+00, 5.88214561488776e+00, 5.78703019849321e+00, 5.69471552070920e+00, 5.60568720161112e+00, 5.52036543144600e+00, 5.43910361587952e+00, 5.36218843951217e+00, 5.28984110143189e+00, 5.22221949176659e+00, 5.15942110169087e+00, 5.10148648741276e+00, 5.04840313869072e+00, 5.00010963251843e+00, 4.95649998133202e+00, 4.91742811121664e+00, 4.88271242802755e+00, 4.85214044713392e+00, 4.82547347486846e+00, 4.80245133622856e+00, 4.78279714378705e+00, 4.76622209740255e+00, 4.75243029387777e+00, 4.74112351133034e+00, 4.73200591620903e+00, 4.72478862338696e+00, 4.71919402351735e+00, 4.71495977876518e+00, 4.71184237986942e+00, 4.70962015563023e+00, 4.70809563123441e+00, 4.70709714459444e+00, 4.70647964968502e+00, 4.70612466167719e+00, 4.70593932891794e+00, 4.70585464952826e+00, 4.70582288347553e+00, 4.70581424234935e+00, 4.70581296693761e+00, 4.70581287119417e+00, 4.70581118602922e+00, 4.70579976879503e+00, 4.70575779792676e+00, 4.70564591971139e+00, 4.70540108057530e+00, 4.70493219814105e+00, 4.70411685084311e+00, 4.70279914880024e+00, 4.70078892844872e+00, 4.69786238881058e+00, 4.69376425645486e+00, 4.68821152736268e+00, 4.68089878550047e+00, 4.67150503918974e+00, 4.65970194783874e+00, 4.64516323539267e+00, 4.62757500694984e+00, 4.60664660718425e+00, 4.58212159079079e+00, 4.55378832418242e+00, 4.52148971194961e+00, 4.48513154754880e+00, 4.44468902911474e+00, 4.40021105843871e+00, 4.35182205022183e+00, 4.29972111202343e+00, 4.24417860210639e+00, 4.18553022011102e+00, 4.12416892150561e+00, 4.06053505985487e+00, 3.99510524264234e+00, 3.92838043169790e+00, 3.86087382696316e+00, 3.79309904457752e+00, 3.72555904214789e+00, 3.65873616271209e+00, 3.59308357266653e+00, 3.52901826651745e+00, 3.46691571101039e+00, 3.40710611014522e+00, 3.34987219628550e+00, 3.29544839451626e+00, 3.24402116902387e+00, 3.19573034108029e+00, 3.15067116617220e+00, 3.10889696980627e+00, 3.07042216385696e+00, 3.03522549421317e+00, 3.00325340239010e+00, 2.97442341567556e+00, 2.94862750986644e+00, 2.92573541396843e+00, 2.90559784623147e+00, 2.88804968496896e+00, 2.87291308560627e+00, 2.86000055755130e+00, 2.84911801132853e+00, 2.84006777879240e+00, 2.83265159820210e+00, 2.82667354275030e+00, 2.82194285719160e+00, 2.81827665396821e+00, 2.81550240910848e+00, 2.81346019046515e+00, 2.81200454759414e+00, 2.81100599441793e+00, 2.81035202300207e+00, 2.80994759904530e+00, 2.80971510633583e+00, 2.80959372735289e+00, 2.80953826901549e+00, 2.80951746479538e+00, 2.80951180555410e+00, 2.80951097026234e+00, 2.80951090729106e+00, 2.80950979538640e+00, 2.80950226207455e+00, 2.80947456891359e+00, 2.80940074980691e+00, 2.80923920201376e+00, 2.80892983213762e+00, 2.80839187749180e+00, 2.80752251297490e+00, 2.80619634134733e+00, 2.80426584949552e+00, 2.80156289350136e+00, 2.79790124961508e+00, 2.79308023530923e+00, 2.78688936379382e+00, 2.77911394695599e+00, 2.76954150712108e+00, 2.75796880018810e+00, 2.74420919582388e+00, 2.72810010991468e+00, 2.70951014644196e+00, 2.68834558638549e+00, 2.66455586524756e+00, 2.63813771158328e+00, 2.60913767713662e+00, 2.57765287233170e+00, 2.54382982328131e+00, 2.50786147974679e+00, 2.46998251738700e+00, 2.43046318140084e+00, 2.38960200243503e+00, 2.34771777176984e+00, 2.30514118695541e+00, 2.26220657064422e+00, 2.21924402741931e+00, 2.17657234203406e+00, 2.13449284569741e+00, 2.09328439361815e+00, 2.05319951523646e+00, 2.01446172525129e+00, 1.97726392350825e+00, 1.94176776762129e+00, 1.90810387438522e+00, 1.87637269344403e+00, 1.84664589702452e+00, 1.81896813992623e+00, 1.79335906132431e+00, 1.76981542142771e+00, 1.74831328917567e+00, 1.72881021997726e+00, 1.71124738353528e+00, 1.69555162003486e+00, 1.68163741779179e+00, 1.66940881652404e+00, 1.65876124764776e+00, 1.64958332650603e+00, 1.64175861146560e+00, 1.63516734175201e+00, 1.62968816024116e+00, 1.62519981982065e+00, 1.62158286311011e+00, 1.61872125610762e+00, 1.61650394758487e+00, 1.61482631864222e+00, 1.61359148155402e+00, 1.61271138452388e+00, 1.61210767965878e+00, 1.61171231551999e+00, 1.61146782287675e+00, 1.61132727234509e+00, 1.61125389475512e+00, 1.61122036848856e+00, 1.61120779171917e+00, 1.61120437054149e+00, 1.61120386558315e+00, 1.61120382730800e+00, 1.61120314872321e+00, 1.61119855121602e+00, 1.61118165036227e+00, 1.61113659942127e+00, 1.61103800922798e+00, 1.61084920751766e+00, 1.61052091200298e+00, 1.60999038573459e+00, 1.60918113630454e+00, 1.60800321152435e+00, 1.60635413233355e+00, 1.60412048781753e+00, 1.60118019643194e+00, 1.59740541134737e+00, 1.59266601638154e+00, 1.58683362328239e+00, 1.57978594314964e+00, 1.57141136750695e+00, 1.56161356173979e+00, 1.55031584951608e+00, 1.53746515553736e+00, 1.52303527890587e+00, 1.50702929251760e+00, 1.48948090522295e+00, 1.47045468081406e+00, 1.45004507677871e+00, 1.42837434002597e+00, 1.40558936927297e+00, 1.38185771733320e+00, 1.35736295503767e+00, 1.33229964772146e+00, 1.30686820330031e+00, 1.28126983866804e+00, 1.25570188141339e+00, 1.23035358126076e+00, 1.20540255557705e+00, 1.18101194118668e+00, 1.15732827538058e+00, 1.13448008607536e+00, 1.11257713694486e+00, 1.09171024907589e+00, 1.07195160623009e+00, 1.05335544520874e+00, 1.03595903464404e+00, 1.01978385304115e+00, 1.00483688829021e+00, 9.91111994507293e-01, 9.78591256536490e-01, 9.67246326609256e-01, 9.57039710664588e-01, 9.47925993088589e-01, 9.39852997765939e-01, 9.32762890157775e-01, 9.26593229584142e-01, 9.21277983059014e-01, 9.16748512057481e-01, 9.12934541725761e-01, 9.09765118584332e-01, 9.07169558102439e-01, 9.05078378084355e-01, 9.03424208094933e-01, 9.02142659692089e-01, 9.01173137546877e-01, 9.00459568097929e-01, 8.99951020600227e-01, 8.99602195556980e-01, 8.99373757667263e-01, 8.99232494513200e-01, 8.99151287988730e-01, 8.99108892535864e-01, 8.99089522084921e-01, 8.99082255621764e-01, 8.99080278973443e-01, 8.99079987224482e-01, 8.99079964969999e-01, 8.99079568566165e-01, 8.99076882875079e-01, 8.99067010037840e-01, 8.99040693042309e-01, 8.98983100786580e-01, 8.98872811798185e-01, 8.98681040448253e-01, 8.98371146363090e-01, 8.97898463917556e-01, 8.97210483698316e-01, 8.96247411003358e-01, 8.94943117112462e-01, 8.93226486632884e-01, 8.91023148357988e-01, 8.88257557848746e-01, 8.84855377980614e-01, 8.80746080245444e-01, 8.75865666543504e-01, 8.70159390942689e-01, 8.63584346103622e-01, 8.56111772379492e-01, 8.47728951126328e-01, 8.38440558709976e-01, 8.28269384029014e-01, 8.17256348600722e-01, 8.05459811508200e-01, 7.92954187809030e-01, 7.79827953804585e-01, 7.66181151352797e-01, 7.52122532371672e-01, 7.37766501245630e-01, 7.23230015991319e-01, 7.08629599349663e-01, 6.94078590447930e-01, 6.79684739326191e-01, 6.65548213968322e-01, 6.51760056029644e-01, 6.38401090287317e-01, 6.25541266298185e-01, 6.13239390290090e-01, 6.01543191529004e-01, 5.90489660120269e-01, 5.80105591677153e-01, 5.70408277419077e-01, 5.61406284779707e-01, 5.53100282262378e-01, 5.45483871945123e-01, 5.38544402788474e-01, 5.32263747033284e-01, 5.26619030016349e-01, 5.21583310390170e-01, 5.17126212879759e-01, 5.13214519325075e-01, 5.09812725904581e-01, 5.06883575227121e-01, 5.04388571564292e-01, 5.02288486047163e-01, 5.00543856367543e-01, 4.99115482625939e-01, 4.97964917699065e-01, 4.97054947129229e-01, 4.96350050342869e-01, 4.95816832268630e-01, 4.95424412404375e-01, 4.95144757292820e-01, 4.94952942354388e-01, 4.94827330153072e-01, 4.94749654396827e-01, 4.94705002157232e-01, 4.94681690710337e-01, 4.94671039756611e-01, 4.94667044254797e-01, 4.94665957385095e-01, 4.94665796965521e-01, 4.94665784642464e-01, 4.94665564007761e-01, 4.94664069177261e-01, 4.94658574052337e-01, 4.94643926298252e-01, 4.94611871221053e-01, 4.94550486413132e-01, 4.94443751796354e-01, 4.94271278073484e-01, 4.94008214607248e-01, 4.93625355073661e-01, 4.93089455481709e-01, 4.92363773893857e-01, 4.91408834050944e-01, 4.90183405888317e-01, 4.88645684663126e-01, 4.86754637473467e-01, 4.84471472103522e-01, 4.81761169540192e-01, 4.78594009653932e-01, 4.74947011076267e-01, 4.70805202839358e-01, 4.66162648150458e-01, 4.61023150445740e-01, 4.55400588485162e-01, 4.49318849675208e-01, 4.42811357135795e-01, 4.35920213687581e-01, 4.28695012037352e-01, 4.21191382200291e-01, 4.13469362367266e-01, 4.05591686615168e-01, 3.97622081737978e-01, 3.89623656750146e-01, 3.81657453866907e-01, 3.73781211136262e-01, 3.66048366729032e-01, 3.58507315376457e-01, 3.51200910346904e-01, 3.44166190861017e-01, 3.37434305511721e-01, 3.31030597072157e-01, 3.24974812592870e-01, 3.19281404177417e-01, 3.13959889443538e-01, 3.09015245582756e-01, 3.04448316401491e-01, 3.00256217180277e-01, 2.96432727218757e-01, 2.92968664287756e-01, 2.89852238761696e-01, 2.87069387919303e-01, 2.84604092810000e-01, 2.82438681245353e-01, 2.80554120972894e-01, 2.78930307001758e-01, 2.77546346461904e-01, 2.76380843376813e-01, 2.75412184399613e-01, 2.74618825006627e-01, 2.73979573966358e-01, 2.73473872231051e-01, 2.73082060857405e-01, 2.72785631290645e-01, 2.72567450459078e-01, 2.72411952731801e-01, 2.72305290957519e-01, 2.72235439555331e-01, 2.72192243942015e-01, 2.72167412385883e-01, 2.72154448549173e-01, 2.72148525373197e-01, 2.72146303402863e-01, 2.72145698974725e-01, 2.72145609762447e-01, 2.72145602857066e-01, 2.72145478540807e-01, 2.72144636280871e-01, 2.72141540058968e-01, 2.72133286781241e-01, 2.72115225258673e-01, 2.72080637480997e-01, 2.72020495924721e-01, 2.71923309644905e-01, 2.71775070693039e-01, 2.71559310818688e-01, 2.71257276242508e-01, 2.70848225312079e-01, 2.70309849877989e-01, 2.69618816145879e-01, 2.68751414600587e-01, 2.67684301572956e-01, 2.66395307558383e-01, 2.64864280137871e-01, 2.63073923117290e-01, 2.61010589183456e-01, 2.58664981837143e-01, 2.56032724260507e-01, 2.53114758420750e-01, 2.49917546951593e-01, 2.46453062527166e-01, 2.42738563427750e-01, 2.38796168347591e-01, 2.34652256661562e-01, 2.30336730936301e-01, 2.25882185396614e-01, 2.21323026817714e-01, 2.16694592960920e-01, 2.12032308785837e-01, 2.07370913211690e-01, 2.02743780308055e-01, 1.98182349592315e-01, 1.93715671550876e-01, 1.89370067246472e-01, 1.85168895289181e-01, 1.81132415609952e-01, 1.77277737249739e-01, 1.73618836498576e-01, 1.70166631875333e-01, 1.66929103317827e-01, 1.63911444293475e-01, 1.61116237132873e-01, 1.58543643579407e-01, 1.56191604230983e-01, 1.54056042147602e-01, 1.52131067358397e-01, 1.50409180281002e-01, 1.48881473128611e-01, 1.47537829196111e-01, 1.46367120461633e-01, 1.45357404199842e-01, 1.44496119277107e-01, 1.43770282501620e-01, 1.43166684865392e-01, 1.42672086792032e-01, 1.42273410660252e-01, 1.41957927989154e-01, 1.41713437838447e-01, 1.41528432278903e-01, 1.41392244315909e-01, 1.41295173462808e-01, 1.41228584308521e-01, 1.41184973922194e-01, 1.41158004768315e-01, 1.41142500925568e-01, 1.41134406737848e-01, 1.41130708485832e-01, 1.41129321151780e-01, 1.41128943764031e-01, 1.41128888062397e-01, 1.41128883715908e-01, 1.41128805015512e-01, 1.41128271809208e-01, 1.41126311693995e-01, 1.41121086804503e-01, 1.41109652555926e-01, 1.41087755729558e-01, 1.41049680493168e-01, 1.40988150260497e-01, 1.40894292469696e-01, 1.40757672308226e-01, 1.40566400024014e-01, 1.40307314619122e-01, 1.39966244321215e-01, 1.39528341217854e-01, 1.38978483834535e-01, 1.38301737364372e-01, 1.37483856941913e-01, 1.36511815147372e-01, 1.35374331258220e-01, 1.34062377119292e-01, 1.32569633329638e-01, 1.30892870112277e-01, 1.29032229919758e-01, 1.26991393500333e-01, 1.24777617504151e-01, 1.22401639233786e-01, 1.19877452145758e-01, 1.17221963436677e-01, 1.14454551791596e-01, 1.11596548551702e-01, 1.08670668812749e-01, 1.05700420163330e-01, 1.02709516002534e-01, 9.97213179164760e-02, 9.67583278215045e-02, 9.38417459441609e-02, 9.09911056364479e-02, 8.82239909146421e-02, 8.55558378070356e-02, 8.29998163659837e-02, 8.05667867482924e-02, 7.82653202059862e-02, 7.61017742036643e-02, 7.40804101429973e-02, 7.22035422347176e-02, 7.04717067911531e-02, 6.88838424388649e-02, 6.74374733421388e-02, 6.61288892916474e-02, 6.49533183078006e-02, 6.39050891106905e-02, 6.29777822976934e-02, 6.21643702954344e-02, 6.14573470562760e-02, 6.08488490495915e-02, 6.03307693629915e-02, 5.98948666648766e-02, 5.95328704920917e-02, 5.92365837852213e-02, 5.89979829289007e-02, 5.88093148090252e-02, 5.86631896490388e-02, 5.85526677074455e-02, 5.84713374013432e-02, 5.84133820863144e-02, 5.83736326371422e-02, 5.83476031410015e-02, 5.83315074379414e-02, 5.83222548761739e-02, 5.83174244571438e-02, 5.83152174564390e-02, 5.83143895433977e-02, 5.83141643319607e-02, 5.83141310912936e-02, 5.83141284751889e-02, 5.83140808274010e-02, 5.83137580059514e-02, 5.83125712881696e-02, 5.83094079957931e-02, 5.83024855190217e-02, 5.82892293407156e-02, 5.82661804685358e-02, 5.82289373873408e-02, 5.81721372674480e-02, 5.80894806495422e-02, 5.79738030417190e-02, 5.78171957612028e-02, 5.76111767796448e-02, 5.73469103013622e-02, 5.70154712568893e-02, 5.66081479645959e-02, 5.61167730296006e-02, 5.55340693923370e-02, 5.48539955965452e-02, 5.40720722496388e-02, 5.31856705775198e-02, 5.21942443098311e-02, 5.10994880094131e-02, 4.99054084341708e-02, 4.86183004257015e-02, 4.72466247599641e-02, 4.58007919445294e-02, 4.42928623967163e-02, 4.27361791839758e-02, 4.11449540098658e-02, 3.95338298290540e-02, 3.79174442800405e-02, 3.63100168311234e-02, 3.47249795702993e-02, 3.31746671194395e-02, 3.16700759921673e-02, 3.02206982180593e-02, 2.88344287655491e-02, 2.75175417551168e-02, 2.62747267939289e-02, 2.51091743459995e-02, 2.40226978738463e-02, 2.30158807820828e-02, 2.20882369599025e-02, 2.12383752757006e-02, 2.04568155984504e-02, 1.97562367407045e-02, 1.91252644047637e-02, 1.85603805862655e-02, 1.80578221845451e-02, 1.76136647980485e-02, 1.72238908438089e-02, 1.68844440195954e-02, 1.65912724832395e-02, 1.63403632472665e-02, 1.61277702200806e-02, 1.59496381018869e-02, 1.58022239990601e-02, 1.56819181858197e-02, 1.55852649509143e-02, 1.55089839503120e-02, 1.54499919824787e-02, 1.54054246402922e-02, 1.53726569120352e-02, 1.53493215241838e-02, 1.53333236700206e-02, 1.53228507545293e-02, 1.53163759110271e-02, 1.53126542942346e-02, 1.53107114997511e-02, 1.53098238722720e-02, 1.53094909008615e-02, 1.53094003253358e-02, 1.53093869566163e-02, 1.53093858952360e-02, 1.53093664442281e-02, 1.53092346612453e-02, 1.53087502197593e-02, 1.53074589280919e-02, 1.53046332289364e-02, 1.52992226775979e-02, 1.52898168139615e-02, 1.52746228409484e-02, 1.52514604921521e-02, 1.52177763609842e-02, 1.51706796972491e-02, 1.51070011777392e-02, 1.50233753565345e-02, 1.49163463521802e-02, 1.47824948304844e-02, 1.46185825478824e-02, 1.44217087520673e-02, 1.41894707835917e-02, 1.39201195321874e-02, 1.36126992479859e-02, 1.32671608615027e-02, 1.28844386318008e-02, 1.24664817258235e-02, 1.20162351882271e-02, 1.15375684855792e-02, 1.10351540430997e-02, 1.05143024643439e-02, 9.98076491229938e-03, 9.44051594441930e-03, 8.89953156398349e-03, 8.36357719300906e-03, 7.83801873785706e-03, 7.32766717338507e-03, 6.83666353066315e-03, 6.36840731819177e-03, 5.92552768654286e-03, 5.51207905301965e-03, 5.12494995771817e-03, 4.76674140429623e-03, 4.43733855003359e-03, 4.13617193888172e-03, 3.86233133771441e-03, 3.61464924127120e-03, 3.39175728701946e-03, 3.19213137666452e-03, 3.01414449744186e-03, 2.85613124820827e-03, 2.71644943637739e-03, 2.59352290767367e-03, 2.48586187433375e-03, 2.39206669868963e-03, 2.31082368254931e-03, 2.24089974500430e-03, 2.18113929217623e-03, 2.13046292880392e-03, 2.08786594141941e-03, 2.05241521297239e-03, 2.02324481251992e-03, 1.99955135842345e-03, 1.98059015710328e-03, 1.96567261453140e-03, 1.95416495600398e-03, 1.94548801752603e-03, 1.93911775602373e-03, 1.93458609581398e-03, 1.93148173180191e-03, 1.92945052550897e-03, 1.92819515660858e-03, 1.92747373918678e-03, 1.92709717988748e-03, 1.92692514636406e-03, 1.92686061384961e-03, 1.92684305969237e-03, 1.92684046875513e-03, 1.92684026123402e-03, 1.92683643429348e-03, 1.92681050670274e-03, 1.92671519627652e-03, 1.92646115164052e-03, 1.92590527237257e-03, 1.92484104671564e-03, 1.92299144391939e-03, 1.92000493508715e-03, 1.91545525211723e-03, 1.90884551139447e-03, 1.89961730777100e-03, 1.88716528616339e-03, 1.87085748825666e-03, 1.85006138890742e-03, 1.82417494535514e-03, 1.79266118337296e-03, 1.75508399289015e-03, 1.71114223266500e-03, 1.66069936399538e-03, 1.60380674276439e-03, 1.54071982349535e-03, 1.47190690162802e-03, 1.39804908556584e-03, 1.32002818097021e-03, 1.23889697576374e-03, 1.15582615438825e-03, 1.07202864846556e-03, 9.88678583567548e-04, 9.06853310752354e-04, 8.27512029559330e-04, 7.51497838436421e-04, 6.79539316209343e-04, 6.12232310629679e-04, 5.49997610619357e-04, 4.93038178291272e-04, 4.41339596946477e-04, 3.94730410789634e-04, 3.52967130154980e-04, 3.15794933044344e-04, 2.82959306360908e-04, 2.54175608574769e-04, 2.29093776681566e-04, 2.07302386243668e-04, 1.88377192415601e-04, 1.71936944312279e-04, 1.57672417202548e-04, 1.45341883921901e-04, 1.34745513037160e-04, 1.25697604190123e-04, 1.18011961134812e-04, 1.11504266882889e-04, 1.06003495716323e-04, 1.01361444237114e-04, 9.74551612529240e-05, 9.41835278212011e-05, 9.14616998244355e-05, 8.92162801875429e-05, 8.73822644251732e-05, 8.59014841125827e-05, 8.47218322661547e-05, 8.37967209434964e-05, 8.30845549586456e-05, 8.25482326831661e-05, 8.21547461811178e-05, 8.18749165034660e-05, 8.16832464067829e-05, 8.15578400334971e-05, 8.14803339608261e-05, 8.14357935610133e-05, 8.14125446026992e-05, 8.14019231256680e-05, 8.13979388283315e-05, 8.13968550090749e-05, 8.13966950659051e-05, 8.13966820345513e-05, 8.13964391852551e-05, 8.13947939235504e-05, 8.13887461390550e-05, 8.13726281876370e-05, 8.13373704744436e-05, 8.12699088186264e-05, 8.11527818740387e-05, 8.09639742129775e-05, 8.06770612484797e-05, 8.02616976471882e-05, 7.96844514354689e-05, 7.89099249374203e-05, 7.79020425909493e-05, 7.66253813395679e-05, 7.50465492416197e-05, 7.31359155139444e-05, 7.08703592522397e-05, 6.82378211109835e-05, 6.52438683488299e-05, 6.19189080026672e-05, 5.83223559260659e-05, 5.45383613825834e-05, 5.06594074494160e-05, 4.67616671926407e-05, 4.28858488987771e-05, 3.90387248501156e-05, 3.52193586650174e-05, 3.14600396364995e-05, 2.78614823798762e-05, 2.45947719292872e-05, 2.18555449720270e-05, 1.97960900152339e-05, 1.84381183458981e-05, 1.74981163780785e-05, 1.63890484827789e-05, 1.46955508597557e-05, 1.25640549271111e-05, 1.05152825256027e-05, 8.75966374280779e-06, 7.13497277615716e-06, 5.62835148798513e-06, 4.42957239954467e-06, 3.56475895557651e-06, 2.80030798194275e-06, 2.07117575712346e-06, 1.55754364072891e-06, 1.26306922254571e-06, 9.90711299664326e-07, 8.30343935831479e-07, 7.96200282593418e-07, 7.03379512856372e-07, 4.67219791744355e-07, 3.71007487842689e-07, 5.81702991553357e-07, 7.09930901621192e-07, 7.18636030683228e-07, 6.38736509017690e-07, 5.07339606770074e-07, 1.42108547152020e-14, 1.42108547152020e-14, 2.84217094304040e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 2.84217094304040e-14, 2.84217094304040e-14, 1.42108547152020e-14, 2.84217094304040e-14, 1.42108547152020e-14, 1.42108547152020e-14, 2.84217094304040e-14, 2.84217094304040e-14],
[ 7.40000000000000e+01, 7.39999999184462e+01, 7.39999973966414e+01, 7.39999803110139e+01, 7.39999175020074e+01, 7.39997500729327e+01, 7.39993836416332e+01, 7.39986818172366e+01, 7.39974611365175e+01, 7.39954876491258e+01, 7.39924752898231e+01, 7.39880861210479e+01, 7.39819324730889e+01, 7.39735809545786e+01, 7.39625582555985e+01, 7.39483586217196e+01, 7.39304528413427e+01, 7.39082985611935e+01, 7.38813517247781e+01, 7.38490789135503e+01, 7.38109703567823e+01, 7.37665533594111e+01, 7.37154058736060e+01, 7.36571699073063e+01, 7.35915644221465e+01, 7.35183973282227e+01, 7.34375761418514e+01, 7.33491168453030e+01, 7.32531504858086e+01, 7.31499270847535e+01, 7.30398165027045e+01, 7.29233060216434e+01, 7.28009945556292e+01, 7.26735835720958e+01, 7.25418649809233e+01, 7.24067064088579e+01, 7.22690344062460e+01, 7.21298162194476e+01, 7.19900408000615e+01, 7.18506997122306e+01, 7.17127685485640e+01, 7.15771893842984e+01, 7.14448547006377e+01, 7.13165931037189e+01, 7.11931570652919e+01, 7.10752128218998e+01, 7.09633324948345e+01, 7.08579884342853e+01, 7.07595497465930e+01, 7.06682809307483e+01, 7.05843425260478e+01, 7.05077936540409e+01, 7.04385963219250e+01, 7.03766213394324e+01, 7.03216556857734e+01, 7.02734111468537e+01, 7.02315340258121e+01, 7.01956157124714e+01, 7.01652038804307e+01, 7.01398140653558e+01, 7.01189413658092e+01, 7.01020719999707e+01, 7.00886944490939e+01, 7.00783099225885e+01, 7.00704418910802e+01, 7.00646444531794e+01, 7.00605093291164e+01, 7.00576713095361e+01, 7.00558120297988e+01, 7.00546619878771e+01, 7.00540007757612e+01, 7.00536555483006e+01, 7.00534978075638e+01, 7.00534386330080e+01, 7.00534225360413e+01, 7.00534201601608e+01, 7.00534198484954e+01, 7.00534125870653e+01, 7.00533633897908e+01, 7.00531825351514e+01, 7.00527004411304e+01, 7.00516453802468e+01, 7.00496247744661e+01, 7.00461108052740e+01, 7.00404309780816e+01, 7.00317641693640e+01, 7.00191425585300e+01, 7.00014597028621e+01, 6.99774848512261e+01, 6.99458834094779e+01, 6.99052432676106e+01, 6.98541064774225e+01, 6.97910055338103e+01, 6.97145032694894e+01, 6.96232351324971e+01, 6.95159523934955e+01, 6.93915646466838e+01, 6.92491798508512e+01, 6.90881401362032e+01, 6.89080517068001e+01, 6.87088074161842e+01, 6.84906009829639e+01, 6.82539323130968e+01, 6.79996039455745e+01, 6.77287091559601e+01, 6.74426126533207e+01, 6.71429250291648e+01, 6.68314721466219e+01, 6.65102605355676e+01, 6.61814396739567e+01, 6.58472618955486e+01, 6.55100406551325e+01, 6.51721080281077e+01, 6.48357725656294e+01, 6.45032788499154e+01, 6.41767701644156e+01, 6.38582555292385e+01, 6.35495819667498e+01, 6.32524123590800e+01, 6.29682087818277e+01, 6.26982208636330e+01, 6.24434785765863e+01, 6.22047888822632e+01, 6.19827357747659e+01, 6.17776834034221e+01, 6.15897820729616e+01, 6.14189769858805e+01, 6.12650196105080e+01, 6.11274815408577e+01, 6.10057706749596e+01, 6.08991494886802e+01, 6.08067551294807e+01, 6.07276210030782e+01, 6.06606994773460e+01, 6.06048852830930e+01, 6.05590391518442e+01, 6.05220111982610e+01, 6.04926635319060e+01, 6.04698915725741e+01, 6.04526435483251e+01, 6.04399376781401e+01, 6.04308765834764e+01, 6.04246585353915e+01, 6.04205852254877e+01, 6.04180658473691e+01, 6.04166173869789e+01, 6.04158611403470e+01, 6.04155156004999e+01, 6.04153859758434e+01, 6.04153507147249e+01, 6.04153455102552e+01, 6.04153450507996e+01, 6.04153360479930e+01, 6.04152750526403e+01, 6.04150508270743e+01, 6.04144531225227e+01, 6.04131450530324e+01, 6.04106399139639e+01, 6.04062833637196e+01, 6.03992417720642e+01, 6.03884974048195e+01, 6.03728509604141e+01, 6.03509317951667e+01, 6.03212159674781e+01, 6.02820519936532e+01, 6.02316939398820e+01, 6.01683411801937e+01, 6.00901838387067e+01, 5.99954526222071e+01, 5.98824714579025e+01, 5.97497111073369e+01, 5.95958417582236e+01, 5.94197825253278e+01, 5.92207458349449e+01, 5.89982748276999e+01, 5.87522721794928e+01, 5.84830190864405e+01, 5.81911835569477e+01, 5.78778175760583e+01, 5.75443431376546e+01, 5.71925275737657e+01, 5.68244490467456e+01, 5.64424535018796e+01, 5.60491047805568e+01, 5.56471299242566e+01, 5.52393619048247e+01, 5.48286820528609e+01, 5.44179643056634e+01, 5.40100230772169e+01, 5.36075661167405e+01, 5.32131532405828e+01, 5.28291613666141e+01, 5.24577559057696e+01, 5.21008683003134e+01, 5.17601793434518e+01, 5.14371078502281e+01, 5.11328042453597e+01, 5.08481486602661e+01, 5.05837531665744e+01, 5.03399678040053e+01, 5.01168900822286e+01, 4.99143776496736e+01, 4.97320638298315e+01, 4.95693757290908e+01, 4.94255546197775e+01, 4.92996782965263e+01, 4.91906850914511e+01, 4.90973992122555e+01, 4.90185570370436e+01, 4.89528339613941e+01, 4.88988713501740e+01, 4.88553031029684e+01, 4.88207813032929e+01, 4.87940003937533e+01, 4.87737193076786e+01, 4.87587809973059e+01, 4.87481288328094e+01, 4.87408194070182e+01, 4.87360313671451e+01, 4.87330700047409e+01, 4.87313674638769e+01, 4.87304785691569e+01, 4.87300724224030e+01, 4.87299200621814e+01, 4.87298786164548e+01, 4.87298724991504e+01, 4.87298720084670e+01, 4.87298629528679e+01, 4.87298015998439e+01, 4.87295760595221e+01, 4.87289748509446e+01, 4.87276591161767e+01, 4.87251393094443e+01, 4.87207572904131e+01, 4.87136746377171e+01, 4.87028678678070e+01, 4.86871310922686e+01, 4.86650864699015e+01, 4.86352026038064e+01, 4.85958207954342e+01, 4.85451887965832e+01, 4.84815014003662e+01, 4.84029468921085e+01, 4.83077580557546e+01, 4.81942661209888e+01, 4.80609557656518e+01, 4.79065190838938e+01, 4.77299063184465e+01, 4.75303711562276e+01, 4.73075085129342e+01, 4.70612829867212e+01, 4.67920465350793e+01, 4.65005444046643e+01, 4.61879088959335e+01, 4.58556411426739e+01, 4.55055816954836e+01, 4.51398712759487e+01, 4.47609035644594e+01, 4.43712722439766e+01, 4.39737146946743e+01, 4.35710546930805e+01, 4.31661462260009e+01, 4.27618201396832e+01, 4.23608348937189e+01, 4.19658322640147e+01, 4.15792984994655e+01, 4.12035312002229e+01, 4.08406120320889e+01, 4.04923852840362e+01, 4.01604421799733e+01, 3.98461107539846e+01, 3.95504509912014e+01, 3.92742548370442e+01, 3.90180506014527e+01, 3.87821112428837e+01, 3.85664660122268e+01, 3.83709149643196e+01, 3.81950458940354e+01, 3.80382533123614e+01, 3.78997591335128e+01, 3.77786347875033e+01, 3.76738244976425e+01, 3.75841694666335e+01, 3.75084326988916e+01, 3.74453241535860e+01, 3.73935258777995e+01, 3.73517167185111e+01, 3.73185961629096e+01, 3.72929068160325e+01, 3.72734549996485e+01, 3.72591289523377e+01, 3.72489141320670e+01, 3.72419051713654e+01, 3.72373141114571e+01, 3.72344746430270e+01, 3.72328422031986e+01, 3.72319899145305e+01, 3.72316004949809e+01, 3.72314544099696e+01, 3.72314146712767e+01, 3.72314088059278e+01, 3.72314083535357e+01, 3.72314002300249e+01, 3.72313451920434e+01, 3.72311428665898e+01, 3.72306035411262e+01, 3.72294232400937e+01, 3.72271628207040e+01, 3.72232319270709e+01, 3.72168785435832e+01, 3.72071847743390e+01, 3.71930693423067e+01, 3.71732971452858e+01, 3.71464960211363e+01, 3.71111806586631e+01, 3.70657833420298e+01, 3.70086909385402e+01, 3.69382872399468e+01, 3.68529994595439e+01, 3.67513473900136e+01, 3.66319934636315e+01, 3.64937917529077e+01, 3.63358338319945e+01, 3.61574894100934e+01, 3.59584397642924e+01, 3.57387022483010e+01, 3.54986445316745e+01, 3.52389877150502e+01, 3.49607980416208e+01, 3.46654675427737e+01, 3.43546845665211e+01, 3.40303956860181e+01, 3.36947609183289e+01, 3.33501044559619e+01, 3.29988631982453e+01, 3.26435352631747e+01, 3.22866303866756e+01, 3.19306237230443e+01, 3.15779141107307e+01, 3.12307874271896e+01, 3.08913852799132e+01, 3.05616790015548e+01, 3.02434487435771e+01, 2.99382673810273e+01, 2.96474889225139e+01, 2.93722411322366e+01, 2.91134220885934e+01, 2.88717004111633e+01, 2.86475188812230e+01, 2.84411011655660e+01, 2.82524613380382e+01, 2.80814158855832e+01, 2.79275978894028e+01, 2.77904730860184e+01, 2.76693575328572e+01, 2.75634366220374e+01, 2.74717851981096e+01, 2.73933885361604e+01, 2.73271639239099e+01, 2.72719825658492e+01, 2.72266914919703e+01, 2.71901351129086e+01, 2.71611760231073e+01, 2.71387146201319e+01, 2.71217070875065e+01, 2.71091812855828e+01, 2.71002501138323e+01, 2.70941219506554e+01, 2.70901078434051e+01, 2.70876252098645e+01, 2.70861979189788e+01, 2.70854527377986e+01, 2.70851122568308e+01, 2.70849845304173e+01, 2.70849497857125e+01, 2.70849446574660e+01, 2.70849442691706e+01, 2.70849373905572e+01, 2.70848907869365e+01, 2.70847194670639e+01, 2.70842627909286e+01, 2.70832633650816e+01, 2.70813493400614e+01, 2.70780208189747e+01, 2.70726410023601e+01, 2.70644325969958e+01, 2.70524799028194e+01, 2.70357368591089e+01, 2.70130411748987e+01, 2.69831344865760e+01, 2.69446882771034e+01, 2.68963350589890e+01, 2.68367040736987e+01, 2.67644605049935e+01, 2.66783469583569e+01, 2.65772257423941e+01, 2.64601203217336e+01, 2.63262542149275e+01, 2.61750856024598e+01, 2.60063360012097e+01, 2.58200115574955e+01, 2.56164158082500e+01, 2.53961531481144e+01, 2.51601227006120e+01, 2.49095027972327e+01, 2.46457267833281e+01, 2.43704513490833e+01, 2.40855189757893e+01, 2.37929163417078e+01, 2.34947306116788e+01, 2.31931054323582e+01, 2.28901981993925e+01, 2.25881398152756e+01, 2.22889977931366e+01, 2.19947432485925e+01, 2.17072220956215e+01, 2.14281306236325e+01, 2.11589955544676e+01, 2.09011586227279e+01, 2.06557656600390e+01, 2.04237600795743e+01, 2.02058805544473e+01, 2.00026625771020e+01, 1.98144434947562e+01, 1.96413705531979e+01, 1.94834114555492e+01, 1.93403669538341e+01, 1.92118850325926e+01, 1.90974763046961e+01, 1.89965303078720e+01, 1.89083324550378e+01, 1.88320814434986e+01, 1.87669069615751e+01, 1.87118875438268e+01, 1.86660684182915e+01, 1.86284791642943e+01, 1.85981509625782e+01, 1.85741331772634e+01, 1.85555089685213e+01, 1.85414096027377e+01, 1.85310271094232e+01, 1.85236249358551e+01, 1.85185462742239e+01, 1.85152197826005e+01, 1.85131624888972e+01, 1.85119797527505e+01, 1.85113622588937e+01, 1.85110801210794e+01, 1.85109742814818e+01, 1.85109454905485e+01, 1.85109412410672e+01, 1.85109409220573e+01, 1.85109353071215e+01, 1.85108972651126e+01, 1.85107574187724e+01, 1.85103846406093e+01, 1.85095688289570e+01, 1.85080064682989e+01, 1.85052895609034e+01, 1.85008984530770e+01, 1.84941990050050e+01, 1.84844444653726e+01, 1.84707823069423e+01, 1.84522661531094e+01, 1.84278727751284e+01, 1.83965239641248e+01, 1.83571128830278e+01, 1.83085342867039e+01, 1.82497177733767e+01, 1.81796630102785e+01, 1.80974756778924e+01, 1.80024027183526e+01, 1.78938653728263e+01, 1.77714884661668e+01, 1.76351244569740e+01, 1.74848709238735e+01, 1.73210804040015e+01, 1.71443618296521e+01, 1.69555732087748e+01, 1.67558056424276e+01, 1.65463592390861e+01, 1.63287119384744e+01, 1.61044826599259e+01, 1.58753905059678e+01, 1.56432119491712e+01, 1.54097379871962e+01, 1.51767331594263e+01, 1.49458980876622e+01, 1.47188368589987e+01, 1.44970301507755e+01, 1.42818145523207e+01, 1.40743681129806e+01, 1.38757017803147e+01, 1.36866561136259e+01, 1.35079024788032e+01, 1.33399478493524e+01, 1.31831423429363e+01, 1.30376886932164e+01, 1.29036529711050e+01, 1.27809760063639e+01, 1.26694851017932e+01, 1.25689057644400e+01, 1.24788732922832e+01, 1.23989441456936e+01, 1.23286070987549e+01, 1.22672942065801e+01, 1.22143916428454e+01, 1.21692504595655e+01, 1.21311973018549e+01, 1.20995450776926e+01, 1.20736035404936e+01, 1.20526896948488e+01, 1.20361378876287e+01, 1.20233094023320e+01, 1.20136013385428e+01, 1.20064545346002e+01, 1.20013602832560e+01, 1.19978655992088e+01, 1.19955768245498e+01, 1.19941614024046e+01, 1.19933477079288e+01, 1.19929228955688e+01, 1.19927287974591e+01, 1.19926559849076e+01, 1.19926361781612e+01, 1.19926332547281e+01, 1.19926330359324e+01, 1.19926291937711e+01, 1.19926031625801e+01, 1.19925074694346e+01, 1.19922523886855e+01, 1.19916941610722e+01, 1.19906251281363e+01, 1.19887661951974e+01, 1.19857619990831e+01, 1.19811791127699e+01, 1.19745075646357e+01, 1.19651658806361e+01, 1.19525097686277e+01, 1.19358444512395e+01, 1.19144405150202e+01, 1.18875529800014e+01, 1.18544431102777e+01, 1.18144022923622e+01, 1.17667771181737e+01, 1.17109946417742e+01, 1.16465866538560e+01, 1.15732117558951e+01, 1.14906740341110e+01, 1.13989372429213e+01, 1.12981336102792e+01, 1.11885666639598e+01, 1.10707078278588e+01, 1.09451869201592e+01, 1.08127770638201e+01, 1.06743748558769e+01, 1.05309769015157e+01, 1.03836539775832e+01, 1.02335241374621e+01, 1.00817260093135e+01, 9.92939339011220e+00, 9.77763202556268e+00, 9.62749922161903e+00, 9.47998668606294e+00, 9.33600677159829e+00, 9.19638210024591e+00, 9.06183839930692e+00, 8.93300027208429e+00, 8.81038955803827e+00, 8.69442590141104e+00, 8.58542913870309e+00, 8.48362312837705e+00, 8.38914067606655e+00, 8.30202925095551e+00, 8.22225723949731e+00, 8.14972053687777e+00, 8.08424933035030e+00, 8.02561497796628e+00, 7.97353692817788e+00, 7.92768965808254e+00, 7.88770962943865e+00, 7.85320227165144e+00, 7.82374900014912e+00, 7.79891426806776e+00, 7.77825263063295e+00, 7.76131577722654e+00, 7.74765945838796e+00, 7.73685020670568e+00, 7.72847172459952e+00, 7.72213079119736e+00, 7.71746252739878e+00, 7.71413485479646e+00, 7.71185199166311e+00, 7.71035684807447e+00, 7.70943221181194e+00, 7.70890065540796e+00, 7.70862314017467e+00, 7.70849634234015e+00, 7.70844877628377e+00, 7.70843583718096e+00, 7.70843392739712e+00, 7.70843378441154e+00, 7.70843127279804e+00, 7.70841425625714e+00, 7.70835170180287e+00, 7.70818495574829e+00, 7.70782004193981e+00, 7.70712121077095e+00, 7.70590600768824e+00, 7.70394210484967e+00, 7.70094610992768e+00, 7.69658452967414e+00, 7.69047702002894e+00, 7.68220199635178e+00, 7.67130460435254e+00, 7.65730696325548e+00, 7.63972048834393e+00, 7.61805998352845e+00, 7.59185907209411e+00, 7.56068641433757e+00, 7.52416205579057e+00, 7.48197317184891e+00, 7.43388843635316e+00, 7.37977025336259e+00, 7.31958415938851e+00, 7.25340482845385e+00, 7.18141828872992e+00, 7.10392017478758e+00, 7.02131007573810e+00, 6.93408227514038e+00, 6.84281339096920e+00, 6.74814759264846e+00, 6.65078018155529e+00, 6.55144036309018e+00, 6.45087401245419e+00, 6.34982715106868e+00, 6.24903072126014e+00, 6.14918709246845e+00, 6.05095857235713e+00, 5.95495804757667e+00, 5.86174175331548e+00, 5.77180407408234e+00, 5.68557421106082e+00, 5.60341451078749e+00, 5.52562023085967e+00, 5.45242051557221e+00, 5.38398036318058e+00, 5.32040338328314e+00, 5.26173516496874e+00, 5.20796710189406e+00, 5.15904054768417e+00, 5.11485120244055e+00, 5.07525365716795e+00, 5.04006604611446e+00, 5.00907477601868e+00, 4.98203931499224e+00, 4.95869703151851e+00, 4.93876807554466e+00, 4.92196028909720e+00, 4.90797412396435e+00, 4.89650752992154e+00, 4.88726076028416e+00, 4.87994102408931e+00, 4.87426689792255e+00, 4.86997239729138e+00, 4.86681059927963e+00, 4.86455670641563e+00, 4.86301044714178e+00, 4.86199772125296e+00, 4.86137141877074e+00, 4.86101136688222e+00, 4.86082339020463e+00, 4.86073750276030e+00, 4.86070528352029e+00, 4.86069651911576e+00, 4.86069522550838e+00, 4.86069512839900e+00, 4.86069341919269e+00, 4.86068183907443e+00, 4.86063926942487e+00, 4.86052579506131e+00, 4.86027746270707e+00, 4.85980189000935e+00, 4.85897490724670e+00, 4.85763839641112e+00, 4.85559947174507e+00, 4.85263112073640e+00, 4.84847439328827e+00, 4.84284218745989e+00, 4.83542463130402e+00, 4.82589600117890e+00, 4.81392304798869e+00, 4.79917452622823e+00, 4.78133164042544e+00, 4.76009904534681e+00, 4.73521596736674e+00, 4.70646696269282e+00, 4.67369180147138e+00, 4.63679397159245e+00, 4.59574733616589e+00, 4.55060055459739e+00, 4.50147898545478e+00, 4.44858392255162e+00, 4.39218916344474e+00, 4.33263505952340e+00, 4.27032033638362e+00, 4.20569209072842e+00, 4.13923445656391e+00, 4.07145648324574e+00, 4.00287977893547e+00, 3.93402644683836e+00, 3.86540778298020e+00, 3.79751412046592e+00, 3.73080610497259e+00, 3.66570757908886e+00, 3.60260014810447e+00, 3.54181940482771e+00, 3.48365271193101e+00, 3.42833838088428e+00, 3.37606604808200e+00, 3.32697803054567e+00, 3.28117144316781e+00, 3.23870087334987e+00, 3.19958143305803e+00, 3.16379203874999e+00, 3.13127880265723e+00, 3.10195845147830e+00, 3.07572171828303e+00, 3.05243667866937e+00, 3.03195202190633e+00, 3.01410026142395e+00, 2.99870089650243e+00, 2.98556353865739e+00, 2.97449101260976e+00, 2.96528243371672e+00, 2.95773625240079e+00, 2.95165324270548e+00, 2.94683939802098e+00, 2.94310868371335e+00, 2.94028558528056e+00, 2.93820738302679e+00, 2.93672608112009e+00, 2.93570992094014e+00, 2.93504441605660e+00, 2.93463285874667e+00, 2.93439626494295e+00, 2.93427274478306e+00, 2.93421630811599e+00, 2.93419513688930e+00, 2.93418937781321e+00, 2.93418852778605e+00, 2.93418846370389e+00, 2.93418733218405e+00, 2.93417966597660e+00, 2.93415148427680e+00, 2.93407636290110e+00, 2.93391196510458e+00, 2.93359713705651e+00, 2.93304969036996e+00, 2.93216498356138e+00, 2.93081540177707e+00, 2.92885081847507e+00, 2.92610010257826e+00, 2.92237370842143e+00, 2.91746735234145e+00, 2.91116673832654e+00, 2.90325324601460e+00, 2.89351043900831e+00, 2.88173119287421e+00, 2.86772518463353e+00, 2.85132643453012e+00, 2.83240055253509e+00, 2.81085132252352e+00, 2.78662626148115e+00, 2.75972082273049e+00, 2.73018097158415e+00, 2.69810394643002e+00, 2.66363712218316e+00, 2.62697500767212e+00, 2.58835452342072e+00, 2.54804881049731e+00, 2.50635990468724e+00, 2.46361066562190e+00, 2.42013637347510e+00, 2.37627639607754e+00, 2.33236629016369e+00, 2.28873063830611e+00, 2.24567684612146e+00, 2.20349004135190e+00, 2.16242913561359e+00, 2.12272403759064e+00, 2.08457394779204e+00, 2.04814662200918e+00, 2.01357846363976e+00, 1.98097529282619e+00, 1.95041364059040e+00, 1.92194242600046e+00, 1.89558489094345e+00, 1.87134068757877e+00, 1.84918803567464e+00, 1.82908588891875e+00, 1.81097606954856e+00, 1.79478534828374e+00, 1.78042746094054e+00, 1.76780506391744e+00, 1.75681163784678e+00, 1.74733335216634e+00, 1.73925090339705e+00, 1.73244133687484e+00, 1.72677985606832e+00, 1.72214161604543e+00, 1.71840348886514e+00, 1.71544577949990e+00, 1.71315386222024e+00, 1.71141970006815e+00, 1.71014320491524e+00, 1.70923339328959e+00, 1.70860929409588e+00, 1.70820056870056e+00, 1.70794781146017e+00, 1.70780250919479e+00, 1.70772665065235e+00, 1.70769199079088e+00, 1.70767898876649e+00, 1.70767545190877e+00, 1.70767492987625e+00, 1.70767489030692e+00, 1.70767418877715e+00, 1.70766943581408e+00, 1.70765196348577e+00, 1.70760538918540e+00, 1.70750346506584e+00, 1.70730827813725e+00, 1.70696887757711e+00, 1.70642039965746e+00, 1.70558375331424e+00, 1.70436592000860e+00, 1.70266090917336e+00, 1.70035139415549e+00, 1.69731103226903e+00, 1.69340744587426e+00, 1.68850580945117e+00, 1.68247295145891e+00, 1.67518184136406e+00, 1.66651629456078e+00, 1.65637569478109e+00, 1.64467950922244e+00, 1.63137136013223e+00, 1.61642242135875e+00, 1.59983393137843e+00, 1.58163865557248e+00, 1.56190118783924e+00, 1.54071705060991e+00, 1.51821062684994e+00, 1.49453203057816e+00, 1.46985308671871e+00, 1.44436264063685e+00, 1.41826144825863e+00, 1.39175690736731e+00, 1.36505788010606e+00, 1.33836982868337e+00, 1.31189044515622e+00, 1.28580590716732e+00, 1.26028783992493e+00, 1.23549101524443e+00, 1.21155177481462e+00, 1.18858712946816e+00, 1.16669446034473e+00, 1.14595173156270e+00, 1.12641811663397e+00, 1.10813494106529e+00, 1.09112684977066e+00, 1.07540311837066e+00, 1.06095904054722e+00, 1.04777733791822e+00, 1.03582955320114e+00, 1.02507740081738e+00, 1.01547406087228e+00, 1.00696541217401e+00, 9.99491207366944e-01, 9.92986198244722e-01, 9.87381221898076e-01, 9.82604258678663e-01, 9.78581471246827e-01, 9.75238230541677e-01, 9.72500129761386e-01, 9.70293981844449e-01, 9.68548790025977e-01, 9.67196675369493e-01, 9.66173740297251e-01, 9.65420843575586e-01, 9.64884260371093e-01, 9.64516201160649e-01, 9.64275165562938e-01, 9.64126111480653e-01, 9.64040426034941e-01, 9.63995692207334e-01, 9.63975253339555e-01, 9.63967586077317e-01, 9.63965500401983e-01, 9.63965192560892e-01, 9.63965169078904e-01, 9.63964750810405e-01, 9.63961916983095e-01, 9.63951499582521e-01, 9.63923730989009e-01, 9.63862961987275e-01, 9.63746589305728e-01, 9.63544238849607e-01, 9.63217247426115e-01, 9.62718481678863e-01, 9.61992527632518e-01, 9.60976277036409e-01, 9.59599926883428e-01, 9.57788395460966e-01, 9.55463141725389e-01, 9.52544354753812e-01, 9.48953457163697e-01, 9.44615841987130e-01, 9.39463738477272e-01, 9.33439081173375e-01, 9.26496241007882e-01, 9.18604469999011e-01, 9.09749914335137e-01, 8.99937065706553e-01, 8.89189547589254e-01, 8.77550170381298e-01, 8.65080233961379e-01, 8.51858104367309e-01, 8.37977138227288e-01, 8.23543069669526e-01, 8.08671005692005e-01, 7.93482194606767e-01, 7.78100736952283e-01, 7.62650399600687e-01, 7.47251673553251e-01, 7.32019187089719e-01, 7.17059552080613e-01, 7.02469686034405e-01, 6.88335619107235e-01, 6.74731766422070e-01, 6.61720623326175e-01, 6.49352825451885e-01, 6.37667506616708e-01, 6.26692885046936e-01, 6.16447011046731e-01, 6.06938615754657e-01, 5.98168009691506e-01, 5.90127990170103e-01, 5.82804727263152e-01, 5.76178608102666e-01, 5.70225028242049e-01, 5.64915126297137e-01, 5.60216463917712e-01, 5.56093657289793e-01, 5.52508968883984e-01, 5.49422869171451e-01, 5.46794577689056e-01, 5.44582591343456e-01, 5.42745205421923e-01, 5.41241029669237e-01, 5.40029498264474e-01, 5.39071368880812e-01, 5.38329202538989e-01, 5.37767812974664e-01, 5.37354672009016e-01, 5.37060256178123e-01, 5.36858319791364e-01, 5.36726080725629e-01, 5.36644307573638e-01, 5.36597300104447e-01, 5.36572759121668e-01, 5.36561546407086e-01, 5.36557340173303e-01, 5.36556195979742e-01, 5.36556027099309e-01, 5.36556014126290e-01, 5.36555781854844e-01, 5.36554208183958e-01, 5.36548423236226e-01, 5.36533002942662e-01, 5.36499257291808e-01, 5.36434635255247e-01, 5.36322272391686e-01, 5.36140705472334e-01, 5.35863776419035e-01, 5.35460745073848e-01, 5.34896626391188e-01, 5.34132762106125e-01, 5.33127629387650e-01, 5.31837879198889e-01, 5.30219585095054e-01, 5.28229669391195e-01, 5.25827458843395e-01, 5.22976307446314e-01, 5.19645211242462e-01, 5.15810330930520e-01, 5.11456334276772e-01, 5.06577473256002e-01, 5.01178321237902e-01, 4.95274113291785e-01, 4.88890656724891e-01, 4.82063807277655e-01, 4.74838536224368e-01, 4.67267641855059e-01, 4.59410182418452e-01, 4.51329724111176e-01, 4.43092505540989e-01, 4.34765618849823e-01, 4.26415298072044e-01, 4.18105389013732e-01, 4.09896054310252e-01, 4.01842744939273e-01, 3.93995447797224e-01, 3.86398199999704e-01, 3.79088845710442e-01, 3.72099001194556e-01, 3.65454188426128e-01, 3.59174096434884e-01, 3.53272931792219e-01, 3.47759824218244e-01, 3.42639259250674e-01, 3.37911516400723e-01, 3.33573097556268e-01, 3.29617136110798e-01, 3.26033782117221e-01, 3.22810562568108e-01, 3.19932718681869e-01, 3.17383523899792e-01, 3.15144587279917e-01, 3.13196147232730e-01, 3.11517360200256e-01, 3.10086588046332e-01, 3.08881686701341e-01, 3.07880297094840e-01, 3.07060137708092e-01, 3.06399296300796e-01, 3.05876516624267e-01, 3.05471474354718e-01, 3.05165035189759e-01, 3.04939487169023e-01, 3.04778738905696e-01, 3.04668475616367e-01, 3.04596265638608e-01, 3.04551611502106e-01, 3.04525941496138e-01, 3.04512539929277e-01, 3.04506416754492e-01, 3.04504119758320e-01, 3.04503494921377e-01, 3.04503402696810e-01, 3.04503395558257e-01, 3.04503267044396e-01, 3.04502396345171e-01, 3.04499195577904e-01, 3.04490663624534e-01, 3.04471992246732e-01, 3.04436236598434e-01, 3.04374064340024e-01, 3.04273596534074e-01, 3.04120352252582e-01, 3.03897307181899e-01, 3.03585074259303e-01, 3.03162211279187e-01, 3.02605656268710e-01, 3.01891286136609e-01, 3.00994587674893e-01, 2.99891422668267e-01, 2.98558861093341e-01, 2.96976048852756e-01, 2.95125070054340e-01, 2.92991759464812e-01, 2.90566419351038e-01, 2.87844397153592e-01, 2.84826486631786e-01, 2.81519125078091e-01, 2.77934372184163e-01, 2.74089670905781e-01, 2.70007405643540e-01, 2.65714286548018e-01, 2.61240599271375e-01, 2.56619365948808e-01, 2.51885465133157e-01, 2.47074756007422e-01, 2.42223246214422e-01, 2.37366334199976e-01, 2.32538147352940e-01, 2.27770987653301e-01, 2.23094888005576e-01, 2.18537275611652e-01, 2.14122733940286e-01, 2.09872852051589e-01, 2.05806148917972e-01, 2.01938060485602e-01, 1.98280978068482e-01, 1.94844327898380e-01, 1.91634683045493e-01, 1.88655900357077e-01, 1.85909276457778e-01, 1.83393718148423e-01, 1.81105923676739e-01, 1.79040572315600e-01, 1.77190520472081e-01, 1.75547003190218e-01, 1.74099840387214e-01, 1.72837647471880e-01, 1.71748050096042e-01, 1.70817902675182e-01, 1.70033509974971e-01, 1.69380850529265e-01, 1.68845799966107e-01, 1.68414351537550e-01, 1.68072830353282e-01, 1.67808097082826e-01, 1.67607736311396e-01, 1.67460224378175e-01, 1.67355071469785e-01, 1.67282933011904e-01, 1.67235686024895e-01, 1.67206467058236e-01, 1.67189669547724e-01, 1.67180899876683e-01, 1.67176892974240e-01, 1.67175389852076e-01, 1.67174980966962e-01, 1.67174920616333e-01, 1.67174915907097e-01, 1.67174830638231e-01, 1.67174252929314e-01, 1.67172129216340e-01, 1.67166468222362e-01, 1.67154079504343e-01, 1.67130354504351e-01, 1.67089099274719e-01, 1.67022427222591e-01, 1.66920719239229e-01, 1.66772657367403e-01, 1.66565336612038e-01, 1.66284457512603e-01, 1.65914599577107e-01, 1.65439572580823e-01, 1.64842839090412e-01, 1.64107997500182e-01, 1.63219310592709e-01, 1.62162260504175e-01, 1.60924107393356e-01, 1.59494426552129e-01, 1.57865597577796e-01, 1.56033219895150e-01, 1.53996431510947e-01, 1.51758112359218e-01, 1.49324959620813e-01, 1.46707429500481e-01, 1.43919547459774e-01, 1.40978596195183e-01, 1.37904697118486e-01, 1.34720306248312e-01, 1.31449648900244e-01, 1.28118119193203e-01, 1.24751670173765e-01, 1.21376218571697e-01, 1.18017085348441e-01, 1.14698489776742e-01, 1.11443111120835e-01, 1.08271728062377e-01, 1.05202941641735e-01, 1.02252982769528e-01, 9.94356007101374e-02, 9.67620252694132e-02, 9.42409931243020e-02, 9.18788278787446e-02, 8.96795633359915e-02, 8.76450995768181e-02, 8.57753815775197e-02, 8.40685907013494e-02, 8.25213406062772e-02, 8.11288711399953e-02, 7.98852360075699e-02, 7.87834820355897e-02, 7.78158193224385e-02, 7.69737825571610e-02, 7.62483844066963e-02, 7.56302621719840e-02, 7.51098190115158e-02, 7.46773608828022e-02, 7.43232298308243e-02, 7.40379335837389e-02, 7.38122706545283e-02, 7.36374492588112e-02, 7.35051977023318e-02, 7.34078632680593e-02, 7.33384963712922e-02, 7.32909166862328e-02, 7.32597581229726e-02, 7.32404901884820e-02, 7.32294138692851e-02, 7.32236312764201e-02, 7.32209892145477e-02, 7.32199980965191e-02, 7.32197284903009e-02, 7.32196886957013e-02, 7.32196855643252e-02, 7.32196285253792e-02, 7.32192420644253e-02, 7.32178214055352e-02, 7.32140345083536e-02, 7.32057472956456e-02, 7.31898774766790e-02, 7.31622834522767e-02, 7.31176941711880e-02, 7.30496854590257e-02, 7.29507076017626e-02, 7.28121680114384e-02, 7.26245715263900e-02, 7.23777190743817e-02, 7.20609631692781e-02, 7.16635158597683e-02, 7.11748015219200e-02, 7.05848434442657e-02, 6.98846695476618e-02, 6.90667195694488e-02, 6.81252334988187e-02, 6.70565998230330e-02, 6.58596422699560e-02, 6.45358255811885e-02, 6.30893642807706e-02, 6.15272235314421e-02, 5.98590070825509e-02, 5.80967340675045e-02, 5.62545134222177e-02, 5.43481316612429e-02, 5.23945760319151e-02, 5.04115206139346e-02, 4.84168049957958e-02, 4.64279341619829e-02, 4.44457768710862e-02, 4.25197706989877e-02, 4.06455908621268e-02, 3.88355188748335e-02, 3.70999424262806e-02, 3.54472915967676e-02, 3.38840516670444e-02, 3.24148406997296e-02, 3.10425379844759e-02, 2.97684487313319e-02, 2.85924908931321e-02, 2.75133914258419e-02, 2.65288813425997e-02, 2.56358812758180e-02, 2.48306716718218e-02, 2.41090440081706e-02, 2.34664314199478e-02, 2.28980187686716e-02, 2.23988334625318e-02, 2.19638192350828e-02, 2.15878956398683e-02, 2.12660062588270e-02, 2.09931585988272e-02, 2.07644584137512e-02, 2.05751407814034e-02, 2.04205997352886e-02, 2.02964176393010e-02, 2.01983948456643e-02, 2.01225795396738e-02, 2.00652970897153e-02, 2.00231777362490e-02, 1.99931811017677e-02, 1.99726158150213e-02, 1.99591525284717e-02, 1.99508287702201e-02, 1.99460443880355e-02, 1.99435467856119e-02, 1.99424056734898e-02, 1.99419776131720e-02, 1.99418611713327e-02, 1.99418439847960e-02, 1.99418426203266e-02, 1.99418176145514e-02, 1.99416481973032e-02, 1.99410254098211e-02, 1.99393653505382e-02, 1.99357326735168e-02, 1.99287768966343e-02, 1.99166846039986e-02, 1.98971506106886e-02, 1.98673711087594e-02, 1.98240616618079e-02, 1.97635025700720e-02, 1.96816134958198e-02, 1.95740582317399e-02, 1.94363790609202e-02, 1.92641582891184e-02, 1.90532022951686e-02, 1.87997409917529e-02, 1.85006331414002e-02, 1.81535658374941e-02, 1.77572349688404e-02, 1.73114929726381e-02, 1.68174509095991e-02, 1.62775240050053e-02, 1.56954132607439e-02, 1.50760203195404e-02, 1.44252980341548e-02, 1.37500445864163e-02, 1.30576538803240e-02, 1.23558386882081e-02, 1.16523451762589e-02, 1.09546776997576e-02, 1.02698511197161e-02, 9.60418460537937e-03, 8.96314644354845e-03, 8.35125441123807e-03, 7.77203143636029e-03, 7.22801210423540e-03, 6.72079240563050e-03, 6.25111309896909e-03, 5.81896618945166e-03, 5.42371422172039e-03, 5.06421318409878e-03, 4.73893155584904e-03, 4.44606004147374e-03, 4.18360847726262e-03, 3.94948803933914e-03, 3.74157817151399e-03, 3.55777864179174e-03, 3.39604789517978e-03, 3.25442940948051e-03, 3.13106806160590e-03, 3.02421857152524e-03, 2.93224796581626e-03, 2.85363377741987e-03, 2.78695944875551e-03, 2.73090817233879e-03, 2.68425618641288e-03, 2.64586633271335e-03, 2.61468246982588e-03, 2.58972512320099e-03, 2.57008854576668e-03, 2.55493917546801e-03, 2.54351530875851e-03, 2.53512767430664e-03, 2.52916048843588e-03, 2.52507251322243e-03, 2.52239761913131e-03, 2.52074438145655e-03, 2.51979430703386e-03, 2.51929839073739e-03, 2.51907182720856e-03, 2.51898683939089e-03, 2.51896372103033e-03, 2.51896030883358e-03, 2.51896003544516e-03, 2.51895499410182e-03, 2.51892083849725e-03, 2.51879528209997e-03, 2.51846061932781e-03, 2.51772834253927e-03, 2.51632642182642e-03, 2.51388996170476e-03, 2.50995599709825e-03, 2.50396323356483e-03, 2.49525754777835e-03, 2.48310400715468e-03, 2.46670600557763e-03, 2.44523181046929e-03, 2.41784835387193e-03, 2.38376147702013e-03, 2.34226108250368e-03, 2.29276883190721e-03, 2.23488524731863e-03, 2.16843246752774e-03, 2.09348860860814e-03, 2.01040980839196e-03, 1.91983666998397e-03, 1.82268296802590e-03, 1.72010606581807e-03, 1.61346034515475e-03, 1.50423683263057e-03, 1.39399382108252e-03, 1.28428432915939e-03, 1.17658648807151e-03, 1.07224232590397e-03, 9.72409110005401e-04, 8.78025772067303e-04, 7.90098993022070e-04, 7.08534051810261e-04, 6.33429609095515e-04, 5.65813368297086e-04, 5.04861064015927e-04, 4.50199357018732e-04, 4.01600375511180e-04, 3.59079968745846e-04, 3.21600097336727e-04, 2.88676407180333e-04, 2.60202154464650e-04, 2.35625493985481e-04, 2.14235040132849e-04, 1.95654102796539e-04, 1.79771454128286e-04, 1.66328711713959e-04, 1.54769123810183e-04, 1.44823907845648e-04, 1.36274693105065e-04, 1.28934642319223e-04, 1.22867878062038e-04, 1.17934508850226e-04, 1.13771119970352e-04, 1.10274987576986e-04, 1.07365426416095e-04, 1.04974652870737e-04, 1.03040721754155e-04, 1.01502178978308e-04, 1.00298174459112e-04, 9.93721883020271e-05, 9.86744336870515e-05, 9.81617325521711e-05, 9.77964904791406e-05, 9.75459393259825e-05, 9.73818238065860e-05, 9.72803176750858e-05, 9.72219583559308e-05, 9.71914884360103e-05, 9.71775662179652e-05, 9.71723434730620e-05, 9.71709227570727e-05, 9.71707130759114e-05, 9.71706960655183e-05, 9.71703799024226e-05, 9.71682379713457e-05, 9.71603645751884e-05, 9.71393811539656e-05, 9.70934805479828e-05, 9.70056562437094e-05, 9.68531825975560e-05, 9.66074269825867e-05, 9.62340946415452e-05, 9.56940242673454e-05, 9.49446549327604e-05, 9.39421808254792e-05, 9.26439013539948e-05, 9.10092148131980e-05, 8.89982284206781e-05, 8.65718349558620e-05, 8.36986668844020e-05, 8.03696134283882e-05, 7.66211039859854e-05, 7.25280745257351e-05, 6.81160326223562e-05, 6.33825294471535e-05, 5.83559951365942e-05, 5.31097715708029e-05, 4.80838510981130e-05, 4.32139119652675e-05, 3.83242118147109e-05, 3.34565250454943e-05, 2.91259747200456e-05, 2.53268893857239e-05, 2.17116207466006e-05, 1.82740792951108e-05, 1.60920786385077e-05, 1.40232619543212e-05, 1.17534644488160e-05, 1.04279502153304e-05, 8.80361646693473e-06, 6.84185864940901e-06, 4.73836404069061e-06, 2.78707547352042e-06, 1.29244094182468e-06, 7.19532621928920e-07, 6.69252074203541e-07, 6.14583427704929e-07, 6.99218375643795e-07, 8.14638013935109e-07, 1.05842930508970e-06, 1.07122698977946e-06, 9.25765704096193e-07, 7.93551137689974e-07, 6.58782411733227e-07, 5.79632612129899e-07, 5.77685511871096e-07, 6.14344529026312e-07, 6.30760808917330e-07, 5.97012103753514e-07, 5.22184265605574e-07, 4.27066410679799e-07, 0.00000000000000e+00, 0.00000000000000e+00, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14],
[ 7.50000000000000e+01, 7.49999999168836e+01, 7.49999973467595e+01, 7.49999799337609e+01, 7.49999159212947e+01, 7.49997452841690e+01, 7.49993718317769e+01, 7.49986565598011e+01, 7.49974124894159e+01, 7.49954011867994e+01, 7.49923311036607e+01, 7.49878578236046e+01, 7.49815862420578e+01, 7.49730746519295e+01, 7.49618406559079e+01, 7.49473687816576e+01, 7.49291196397799e+01, 7.49065404366298e+01, 7.48790766339106e+01, 7.48461845317874e+01, 7.48073445382372e+01, 7.47620748700256e+01, 7.47099454061298e+01, 7.46505913803750e+01, 7.45837265572583e+01, 7.45091554878460e+01, 7.44267843994323e+01, 7.43366302442795e+01, 7.42388274310449e+01, 7.41336317975857e+01, 7.40214214616475e+01, 7.39026943061831e+01, 7.37780620113832e+01, 7.36482407222495e+01, 7.35140386207586e+01, 7.33763408362022e+01, 7.32360922590889e+01, 7.30942789109073e+01, 7.29519085585385e+01, 7.28099912495367e+01, 7.26695203901709e+01, 7.25314549033175e+01, 7.23967029009389e+01, 7.22661071983400e+01, 7.21404328947984e+01, 7.20203571544875e+01, 7.19064612464868e+01, 7.17992248437781e+01, 7.16990225369505e+01, 7.16061224860539e+01, 7.15206871103070e+01, 7.14427756969665e+01, 7.13723487949054e+01, 7.13092742433628e+01, 7.12533346707158e+01, 7.12042362815054e+01, 7.11616187324109e+01, 7.11250658799992e+01, 7.10941171657811e+01, 7.10682793885312e+01, 7.10470386012563e+01, 7.10298718619430e+01, 7.10162585645336e+01, 7.10056910805909e+01, 7.09976844536644e+01, 7.09917849080061e+01, 7.09875769611122e+01, 7.09846889652632e+01, 7.09827969459835e+01, 7.09816266539058e+01, 7.09809537992541e+01, 7.09806024931356e+01, 7.09804419749629e+01, 7.09803817584875e+01, 7.09803653780926e+01, 7.09803629603787e+01, 7.09803626432256e+01, 7.09803552539395e+01, 7.09803051904224e+01, 7.09801211513862e+01, 7.09796305689548e+01, 7.09785569315055e+01, 7.09765007498112e+01, 7.09729249151564e+01, 7.09671451005424e+01, 7.09583257436285e+01, 7.09454820216202e+01, 7.09274880821640e+01, 7.09030916286043e+01, 7.08709347714322e+01, 7.08295808503728e+01, 7.07775467050872e+01, 7.07133396310434e+01, 7.06354980078663e+01, 7.05426343414484e+01, 7.04334792342774e+01, 7.03069246127510e+01, 7.01620644234095e+01, 6.99982309931705e+01, 6.98150253606959e+01, 6.96123401446801e+01, 6.93903739162883e+01, 6.91496365538070e+01, 6.88909456138954e+01, 6.86154142709119e+01, 6.83244317684813e+01, 6.80196375364284e+01, 6.77028901411038e+01, 6.73762321063912e+01, 6.70418514625116e+01, 6.67020407619882e+01, 6.63591543296095e+01, 6.60155646992983e+01, 6.56736194635855e+01, 6.53355999866351e+01, 6.50036834715440e+01, 6.46799096540724e+01, 6.43661529510874e+01, 6.40641003480573e+01, 6.37752348219841e+01, 6.35008237796363e+01, 6.32419118800599e+01, 6.29993176641721e+01, 6.27736335541735e+01, 6.25652289361084e+01, 6.23742561519719e+01, 6.22006592861906e+01, 6.20441856400199e+01, 6.19043997612773e+01, 6.17806998513852e+01, 6.16723363185495e+01, 6.15784321918730e+01, 6.14980050594501e+01, 6.14299901451654e+01, 6.13732640946743e+01, 6.13266690018806e+01, 6.12890361749600e+01, 6.12592091181473e+01, 6.12360651951067e+01, 6.12185354447915e+01, 6.12056220438536e+01, 6.11964129526484e+01, 6.11900933452213e+01, 6.11859535065024e+01, 6.11833929799367e+01, 6.11819208622360e+01, 6.11811522640545e+01, 6.11808010806135e+01, 6.11806693388390e+01, 6.11806335018121e+01, 6.11806282123395e+01, 6.11806277453797e+01, 6.11806185955331e+01, 6.11805566039631e+01, 6.11803287161943e+01, 6.11797212495363e+01, 6.11783918158214e+01, 6.11758457613485e+01, 6.11714180577520e+01, 6.11642614603591e+01, 6.11533416148814e+01, 6.11374396365237e+01, 6.11151625041407e+01, 6.10849614021016e+01, 6.10451579013436e+01, 6.09939775987467e+01, 6.09295905347455e+01, 6.08501573929164e+01, 6.07538801683251e+01, 6.06390556958364e+01, 6.05041301817736e+01, 6.03477527096563e+01, 6.01688256171338e+01, 5.99665496819766e+01, 5.97404622128040e+01, 5.94904664037115e+01, 5.92168506582624e+01, 5.89202969895896e+01, 5.86018780351022e+01, 5.82630426713960e+01, 5.79055906716788e+01, 5.75316373107279e+01, 5.71435692783849e+01, 5.67439936814230e+01, 5.63356822474788e+01, 5.59215130397026e+01, 5.55044120051113e+01, 5.50872965009010e+01, 5.46730225962084e+01, 5.42643374897948e+01, 5.38638378919644e+01, 5.34739347641706e+01, 5.30968244448693e+01, 5.27344659392393e+01, 5.23885640094515e+01, 5.20605576476784e+01, 5.17516135142609e+01, 5.14626239489463e+01, 5.11942091929705e+01, 5.09467234833639e+01, 5.07202646957827e+01, 5.05146872205132e+01, 5.03296177610791e+01, 5.01644737477843e+01, 5.00184840590832e+01, 4.98907117398556e+01, 4.97800783948408e+01, 4.96853899156531e+01, 4.96053631703245e+01, 4.95386532464077e+01, 4.94838807953412e+01, 4.94396589815745e+01, 4.94046195004709e+01, 4.93774371002472e+01, 4.93568520310131e+01, 4.93416898532673e+01, 4.93308780725893e+01, 4.93234591284317e+01, 4.93185993524841e+01, 4.93155936234830e+01, 4.93138655759998e+01, 4.93129633644589e+01, 4.93125511331241e+01, 4.93123964903637e+01, 4.93123544237310e+01, 4.93123482147822e+01, 4.93123477167477e+01, 4.93123385254850e+01, 4.93122762533205e+01, 4.93120473341438e+01, 4.93114371187999e+01, 4.93101016730487e+01, 4.93075441178259e+01, 4.93030964551071e+01, 4.92959077081316e+01, 4.92849390719082e+01, 4.92689666158900e+01, 4.92465919013030e+01, 4.92162606665210e+01, 4.91762894919352e+01, 4.91249000805328e+01, 4.90602604855569e+01, 4.89805322912088e+01, 4.88839224213065e+01, 4.87687379347653e+01, 4.86334418907710e+01, 4.84767081577292e+01, 4.82974729245635e+01, 4.80949806720804e+01, 4.78688224893746e+01, 4.76189648788920e+01, 4.73457675763642e+01, 4.70499894011728e+01, 4.67327817235866e+01, 4.63956697557006e+01, 4.60405225035322e+01, 4.56695128106254e+01, 4.52850694222498e+01, 4.48898233448130e+01, 4.44865509194844e+01, 4.40781159537375e+01, 4.36674129842159e+01, 4.32573133457849e+01, 4.28506152840894e+01, 4.24499989536224e+01, 4.20579868373991e+01, 4.16769099122288e+01, 4.13088797377273e+01, 4.09557665295916e+01, 4.06191831597138e+01, 4.03004748979459e+01, 4.00007145808246e+01, 3.97207027784745e+01, 3.94609724489373e+01, 3.92217975289553e+01, 3.90032049120751e+01, 3.88049893010736e+01, 3.86267304795213e+01, 3.84678126127819e+01, 3.83274452493227e+01, 3.82046857393784e+01, 3.80984628140509e+01, 3.80076010717561e+01, 3.79308461016513e+01, 3.78668899388690e+01, 3.78143964994641e+01, 3.77720265905561e+01, 3.77384620404539e+01, 3.77124284518632e+01, 3.76927160554047e+01, 3.76781981363427e+01, 3.76678465288282e+01, 3.76607437212911e+01, 3.76560911938024e+01, 3.76532137108809e+01, 3.76515594167636e+01, 3.76506957182650e+01, 3.76503010854874e+01, 3.76501530448171e+01, 3.76501127741378e+01, 3.76501068302689e+01, 3.76501063718205e+01, 3.76500981395594e+01, 3.76500423647788e+01, 3.76498373307782e+01, 3.76492907853563e+01, 3.76480946837924e+01, 3.76458040051485e+01, 3.76418204924630e+01, 3.76353820680814e+01, 3.76255585599184e+01, 3.76112542396337e+01, 3.75912175150470e+01, 3.75640579319375e+01, 3.75282704210906e+01, 3.74822664740549e+01, 3.74244116484894e+01, 3.73530684992047e+01, 3.72666437177050e+01, 3.71636379604510e+01, 3.70426965780270e+01, 3.69026592504873e+01, 3.67426064150482e+01, 3.65619003645928e+01, 3.63602190164029e+01, 3.61375806079397e+01, 3.58943579663702e+01, 3.56312815036258e+01, 3.53494306782587e+01, 3.50502142959923e+01, 3.47353406398338e+01, 3.44067789704491e+01, 3.40667143631112e+01, 3.37174981047278e+01, 3.33615959386318e+01, 3.30015363172911e+01, 3.26398605329778e+01, 3.22790761962738e+01, 3.19216150878552e+01, 3.15697959855085e+01, 3.12257927168564e+01, 3.08916074363876e+01, 3.05690489746437e+01, 3.02597160380912e+01, 2.99649850190575e+01, 2.96860021747162e+01, 2.94236799297775e+01, 2.91786970395275e+01, 2.89515023203409e+01, 2.87423216231228e+01, 2.85511677020123e+01, 2.83778526233616e+01, 2.82220023700702e+01, 2.80830733199611e+01, 2.79603703067154e+01, 2.78530659996235e+01, 2.77602213568999e+01, 2.76808069118787e+01, 2.76137246403418e+01, 2.75578301316912e+01, 2.75119547501232e+01, 2.74749274295741e+01, 2.74455957041332e+01, 2.74228455403554e+01, 2.74056195156289e+01, 2.73929328827830e+01, 2.73838870793536e+01, 2.73776802825025e+01, 2.73736146775703e+01, 2.73711001976301e+01, 2.73696545992245e+01, 2.73688998601586e+01, 2.73685550121799e+01, 2.73684256475605e+01, 2.73683904572250e+01, 2.73683852632044e+01, 2.73683848699288e+01, 2.73683779030913e+01, 2.73683307017403e+01, 2.73681571845560e+01, 2.73676946512576e+01, 2.73666824074736e+01, 2.73647438358011e+01, 2.73613726317320e+01, 2.73559238384173e+01, 2.73476102158524e+01, 2.73355043654094e+01, 2.73185468961059e+01, 2.72955607608421e+01, 2.72652717062260e+01, 2.72263345678861e+01, 2.71773649069212e+01, 2.71169752291590e+01, 2.70438147686044e+01, 2.69566115657766e+01, 2.68542153500611e+01, 2.67356395641877e+01, 2.66001007696781e+01, 2.64470536629664e+01, 2.62762200261564e+01, 2.60876101402061e+01, 2.58815354995977e+01, 2.56586120749661e+01, 2.54197538526030e+01, 2.51661569054921e+01, 2.48992747769605e+01, 2.46207864334098e+01, 2.43325584119310e+01, 2.40366030045836e+01, 2.37350343581195e+01, 2.34300242335862e+01, 2.31237589075722e+01, 2.28183983731742e+01, 2.25160386842139e+01, 2.22186780313033e+01, 2.19281869606454e+01, 2.16462830320376e+01, 2.13745101293882e+01, 2.11142225525186e+01, 2.08665739132518e+01, 2.06325107296522e+01, 2.04127704724092e+01, 2.02078836866440e+01, 2.00181797101150e+01, 1.98437954478334e+01, 1.96846866476953e+01, 1.95406411478795e+01, 1.94112936244128e+01, 1.92961414433819e+01, 1.91945613032712e+01, 1.91058264270163e+01, 1.90291241216717e+01, 1.89635735606588e+01, 1.89082436574015e+01, 1.88621708909201e+01, 1.88243769173559e+01, 1.87938857620814e+01, 1.87697403418693e+01, 1.87510180229025e+01, 1.87368448853496e+01, 1.87264083450684e+01, 1.87189677825465e+01, 1.87138628513513e+01, 1.87105191838952e+01, 1.87084512797837e+01, 1.87072624479160e+01, 1.87066417727650e+01, 1.87063581816700e+01, 1.87062517969418e+01, 1.87062228577240e+01, 1.87062185863564e+01, 1.87062182657034e+01, 1.87062126218488e+01, 1.87061743839124e+01, 1.87060338173498e+01, 1.87056591195517e+01, 1.87048391079081e+01, 1.87032687079149e+01, 1.87005378330385e+01, 1.86961241849352e+01, 1.86893904289127e+01, 1.86795861121861e+01, 1.86658545874720e+01, 1.86472450776218e+01, 1.86227298650386e+01, 1.85912264113038e+01, 1.85516240095289e+01, 1.85028143500264e+01, 1.84437251487073e+01, 1.83733557609535e+01, 1.82908134986361e+01, 1.81953492034605e+01, 1.80863905250714e+01, 1.79635713245397e+01, 1.78267556861872e+01, 1.76760551805070e+01, 1.75118382783652e+01, 1.73347311638123e+01, 1.71456096135170e+01, 1.69455820807985e+01, 1.67359646100230e+01, 1.65182486757236e+01, 1.62940634503328e+01, 1.60651343162671e+01, 1.58332396199286e+01, 1.56001676963868e+01, 1.53676760696593e+01, 1.51374544685908e+01, 1.49110929235960e+01, 1.46900557690433e+01, 1.44756619196442e+01, 1.42690713644895e+01, 1.40712774676688e+01, 1.38831044041178e+01, 1.37052089023731e+01, 1.35380854071704e+01, 1.33820737987469e+01, 1.32373688908475e+01, 1.31040310527917e+01, 1.29819974414187e+01, 1.28710934691768e+01, 1.27710442627354e+01, 1.26814859745887e+01, 1.26019768943707e+01, 1.25320083660352e+01, 1.24710155526102e+01, 1.24183881039959e+01, 1.23734807779885e+01, 1.23356240435189e+01, 1.23041346614545e+01, 1.22783261959648e+01, 1.22575193624512e+01, 1.22410520706658e+01, 1.22282889783172e+01, 1.22186303354409e+01, 1.22115198769831e+01, 1.22064515134655e+01, 1.22029745792463e+01, 1.22006974253357e+01, 1.21992891880330e+01, 1.21984796233896e+01, 1.21980569669597e+01, 1.21978638538633e+01, 1.21977914108164e+01, 1.21977717045834e+01, 1.21977687959859e+01, 1.21977685783005e+01, 1.21977647556370e+01, 1.21977388565462e+01, 1.21976436490100e+01, 1.21973898626782e+01, 1.21968344676691e+01, 1.21957708587559e+01, 1.21939213558779e+01, 1.21909323948998e+01, 1.21863727385744e+01, 1.21797349833226e+01, 1.21704405682489e+01, 1.21578484040257e+01, 1.21412671269058e+01, 1.21199708450652e+01, 1.20932180816935e+01, 1.20602734367567e+01, 1.20204312968317e+01, 1.19730407341860e+01, 1.19175305706040e+01, 1.18534334588128e+01, 1.17804077750351e+01, 1.16982561371749e+01, 1.16069394749631e+01, 1.15065857821788e+01, 1.13974929668699e+01, 1.12801255619534e+01, 1.11551054347626e+01, 1.10231970032244e+01, 1.08852877909154e+01, 1.07423654009144e+01, 1.05954921370689e+01, 1.04457785426078e+01, 1.02943570659435e+01, 1.01423569202984e+01, 9.99088100370467e+00, 9.84098551771419e+00, 9.69366269294980e+00, 9.54982681737556e+00, 9.41030358119873e+00, 9.27582260601763e+00, 9.14701291555112e+00, 9.02440102867182e+00, 8.90841130922138e+00, 8.79936818788687e+00, 8.69749987596570e+00, 8.60294321550894e+00, 8.51574935079857e+00, 8.43588995734859e+00, 8.36326382118293e+00, 8.29770361772580e+00, 8.23898279160713e+00, 8.18682248245993e+00, 8.14089847502511e+00, 8.10084817342180e+00, 8.06627760916888e+00, 8.03676849115386e+00, 8.01188529459100e+00, 7.99118236693555e+00, 7.97421100396993e+00, 7.96052642140458e+00, 7.94969451910768e+00, 7.94129830934604e+00, 7.93494385994858e+00, 7.93026559056861e+00, 7.92693075718019e+00, 7.92464296782441e+00, 7.92314459177261e+00, 7.92221795406263e+00, 7.92168524623018e+00, 7.92140712961685e+00, 7.92128005695415e+00, 7.92123238779170e+00, 7.92121942064072e+00, 7.92121750671703e+00, 7.92121736342146e+00, 7.92121484636350e+00, 7.92119779293505e+00, 7.92113510287319e+00, 7.92096799529993e+00, 7.92060229012041e+00, 7.91990194262912e+00, 7.91868410029480e+00, 7.91671592540058e+00, 7.91371339728448e+00, 7.90934227129019e+00, 7.90322132548880e+00, 7.89492796789600e+00, 7.88400620388499e+00, 7.86997687457352e+00, 7.85234997283401e+00, 7.83063872746155e+00, 7.80437502403450e+00, 7.77312561209338e+00, 7.73650844379111e+00, 7.69420841179513e+00, 7.64599171641478e+00, 7.59171810406561e+00, 7.53135028766704e+00, 7.46495998516082e+00, 7.39273018916636e+00, 7.31495349625975e+00, 7.23202656013325e+00, 7.14444096689044e+00, 7.05277103997867e+00, 6.95765924613249e+00, 6.85979997704422e+00, 6.75992251701291e+00, 6.65877397622884e+00, 6.55710288270541e+00, 6.45564399944919e+00, 6.35510478640818e+00, 6.25615377744180e+00, 6.15941100567224e+00, 6.06544049540005e+00, 5.97474474914921e+00, 5.88776109388147e+00, 5.80485970792425e+00, 5.72634312569576e+00, 5.65244700709934e+00, 5.58334195945541e+00, 5.51913620973195e+00, 5.45987894178916e+00, 5.40556413566665e+00, 5.35613477184236e+00, 5.31148729094940e+00, 5.27147622664687e+00, 5.23591895424806e+00, 5.20460051858956e+00, 5.17727852011700e+00, 5.15368804738566e+00, 5.13354664678042e+00, 5.11655931645426e+00, 5.10242350201193e+00, 5.09083405754797e+00, 5.08148811891166e+00, 5.07408981842393e+00, 5.06835475376705e+00, 5.06401411043208e+00, 5.06081832876826e+00, 5.05854020477311e+00, 5.05697731920577e+00, 5.05595370266927e+00, 5.05532066457630e+00, 5.05495674032011e+00, 5.05476674190022e+00, 5.05467993070116e+00, 5.05464736492820e+00, 5.05463850625824e+00, 5.05463719873748e+00, 5.05463710058363e+00, 5.05463537299397e+00, 5.05462366832577e+00, 5.05458064081741e+00, 5.05446594596950e+00, 5.05421494261103e+00, 5.05373425460516e+00, 5.05289837623116e+00, 5.05154748759969e+00, 5.04948662559967e+00, 5.04648633050988e+00, 5.04228485679886e+00, 5.03659199695581e+00, 5.02909451793060e+00, 5.01946315018562e+00, 5.00736099997704e+00, 4.99245317840285e+00, 4.97441735991362e+00, 4.95295490410872e+00, 4.92780210492046e+00, 4.89874107868907e+00, 4.86560977496035e+00, 4.82831059758117e+00, 4.78681716287855e+00, 4.74117879705274e+00, 4.69152248317948e+00, 4.63805210235438e+00, 4.58104496327257e+00, 4.52084576762736e+00, 4.45785830223079e+00, 4.39253527072684e+00, 4.32536676845518e+00, 4.25686795695145e+00, 4.18756650706675e+00, 4.11799035297351e+00, 4.04865623818715e+00, 3.98005944682495e+00, 3.91266500818668e+00, 3.84690055066822e+00, 3.78315087187650e+00, 3.72175419402180e+00, 3.66299999342871e+00, 3.60712823393561e+00, 3.55432979705782e+00, 3.50474788582855e+00, 3.45848018128390e+00, 3.41558154669195e+00, 3.37606710054139e+00, 3.33991551084711e+00, 3.30707239685429e+00, 3.27745375677418e+00, 3.25094936955436e+00, 3.22742614336364e+00, 3.20673140254584e+00, 3.18869611785520e+00, 3.17313809180949e+00, 3.15986511229180e+00, 3.14867808368381e+00, 3.13937413664829e+00, 3.13174970624634e+00, 3.12560355461396e+00, 3.12073970030808e+00, 3.11697020311448e+00, 3.11411774200826e+00, 3.11201791635941e+00, 3.11052119741110e+00, 3.10949445919509e+00, 3.10882202561190e+00, 3.10840618313487e+00, 3.10816712577167e+00, 3.10804231940760e+00, 3.10798529506231e+00, 3.10796390337707e+00, 3.10795808433087e+00, 3.10795722545227e+00, 3.10795716070284e+00, 3.10795601740028e+00, 3.10794827136343e+00, 3.10791979620230e+00, 3.10784389256727e+00, 3.10767778281931e+00, 3.10735967619451e+00, 3.10680652808279e+00, 3.10591260638130e+00, 3.10454896518340e+00, 3.10256391022156e+00, 3.09978452008043e+00, 3.09601926072469e+00, 3.09106169708070e+00, 3.08469526359045e+00, 3.07669900609793e+00, 3.06685415164952e+00, 3.05495130374550e+00, 3.04079800264520e+00, 3.02422633905324e+00, 3.00510027112334e+00, 2.98332227535526e+00, 2.95883896678153e+00, 2.93164535606822e+00, 2.90178747126262e+00, 2.86936315714986e+00, 2.83452096949496e+00, 2.79745719604291e+00, 2.75841115046262e+00, 2.71765898851412e+00, 2.67550637779850e+00, 2.63228040625629e+00, 2.58832113635313e+00, 2.54397320177108e+00, 2.49957780517043e+00, 2.45546541575763e+00, 2.41194939211817e+00, 2.36932067745040e+00, 2.32784363845440e+00, 2.28775305147275e+00, 2.24925218373610e+00, 2.21251187539619e+00, 2.17767049941794e+00, 2.14483466018778e+00, 2.11408048609592e+00, 2.08545537437952e+00, 2.05898005624937e+00, 2.03465086501127e+00, 2.01244210795032e+00, 1.99230846271705e+00, 1.97418733947497e+00, 1.95800116986136e+00, 1.94365960172946e+00, 1.93106159370599e+00, 1.92009741509305e+00, 1.91065056411830e+00, 1.90259962086397e+00, 1.89582005054544e+00, 1.89018596861149e+00, 1.88557187208484e+00, 1.88185433253236e+00, 1.87891363606141e+00, 1.87663534585676e+00, 1.87491175407125e+00, 1.87364318332602e+00, 1.87273909445015e+00, 1.87211895692170e+00, 1.87171284197279e+00, 1.87146170536133e+00, 1.87131733693319e+00, 1.87124196655795e+00, 1.87120752988569e+00, 1.87119461161024e+00, 1.87119109753638e+00, 1.87119057886683e+00, 1.87119053955237e+00, 1.87118984254188e+00, 1.87118512019863e+00, 1.87116776044623e+00, 1.87112148634159e+00, 1.87102021974962e+00, 1.87082629419835e+00, 1.87048909386300e+00, 1.86994419003787e+00, 1.86911303968495e+00, 1.86790330185094e+00, 1.86620981583518e+00, 1.86391626882485e+00, 1.86089755931137e+00, 1.85702283540446e+00, 1.85215915428969e+00, 1.84617567153131e+00, 1.83894822873029e+00, 1.83036416822173e+00, 1.82032716789321e+00, 1.80876186213634e+00, 1.79561800068133e+00, 1.78087389920078e+00, 1.76453895641117e+00, 1.74665505244489e+00, 1.72729670092217e+00, 1.70656989874632e+00, 1.68460969774280e+00, 1.66157660423269e+00, 1.63765198935101e+00, 1.61303275749484e+00, 1.58792556668639e+00, 1.56254091825149e+00, 1.53708743134698e+00, 1.51176659019040e+00, 1.48676820078788e+00, 1.46226672483922e+00, 1.43841857918346e+00, 1.41536040903139e+00, 1.39320827169844e+00, 1.37205761222825e+00, 1.35198387766135e+00, 1.33304360348293e+00, 1.31527581133555e+00, 1.29870357648649e+00, 1.28333565103357e+00, 1.26916805905964e+00, 1.25618560878272e+00, 1.24436329158875e+00, 1.23366755748305e+00, 1.22405747081014e+00, 1.21548575957756e+00, 1.20789977716856e+00, 1.20124239749225e+00, 1.19545286443633e+00, 1.19046761446521e+00, 1.18622108780309e+00, 1.18264653921190e+00, 1.17967685421364e+00, 1.17724537099188e+00, 1.17528670243118e+00, 1.17373754713725e+00, 1.17253747318296e+00, 1.17162965411555e+00, 1.17096153379708e+00, 1.17048539524055e+00, 1.17015880895916e+00, 1.16994493855441e+00, 1.16981268525360e+00, 1.16973665865071e+00, 1.16969696762717e+00, 1.16967883286220e+00, 1.16967202994913e+00, 1.16967017939780e+00, 1.16966990626052e+00, 1.16966988542573e+00, 1.16966951430985e+00, 1.16966699994843e+00, 1.16965775693780e+00, 1.16963311883265e+00, 1.16957920080793e+00, 1.16947594848875e+00, 1.16929641420380e+00, 1.16900629909139e+00, 1.16856379601866e+00, 1.16791976364892e+00, 1.16701825554641e+00, 1.16579741923148e+00, 1.16419076797892e+00, 1.16212881252899e+00, 1.15954102083936e+00, 1.15635805222557e+00, 1.15251418908626e+00, 1.14794986695445e+00, 1.14261418448368e+00, 1.13646726205921e+00, 1.12948231376247e+00, 1.12164730447242e+00, 1.11296608284805e+00, 1.10345891112222e+00, 1.09316235167525e+00, 1.08212851436417e+00, 1.07042371271359e+00, 1.05812661628265e+00, 1.04532601646758e+00, 1.03211834076944e+00, 1.01860505513375e+00, 1.00489008624304e+00, 9.91077378058222e-01, 9.77268672816692e-01, 9.63561579679705e-01, 9.50047967459341e-01, 9.36812693763031e-01, 9.23932662977009e-01, 9.11476190438549e-01, 8.99502639925913e-01, 8.88062295792025e-01, 8.77196429002808e-01, 8.66937517236195e-01, 8.57309582277679e-01, 8.48328612487521e-01, 8.40003043482952e-01, 8.32334275838335e-01, 8.25317214129086e-01, 8.18940816703261e-01, 8.13188649928193e-01, 8.08039444178945e-01, 8.03467651437558e-01, 7.99444006033895e-01, 7.95936090810798e-01, 7.92908910893260e-01, 7.90325476375159e-01, 7.88147393721800e-01, 7.86335463660137e-01, 7.84850280955709e-01, 7.83652828946472e-01, 7.82705059222010e-01, 7.81970444627120e-01, 7.81414492042543e-01, 7.81005200356176e-01, 7.80713448841183e-01, 7.80513301907106e-01, 7.80382217914010e-01, 7.80301152386485e-01, 7.80254549398109e-01, 7.80230218913928e-01, 7.80219102223356e-01, 7.80214931986492e-01, 7.80213797582618e-01, 7.80213630147031e-01, 7.80213617284986e-01, 7.80213387000700e-01, 7.80211826791955e-01, 7.80206091316046e-01, 7.80190802769766e-01, 7.80157344803271e-01, 7.80093271258835e-01, 7.79981854536345e-01, 7.79801796181545e-01, 7.79527119664664e-01, 7.79127262048590e-01, 7.78567377052283e-01, 7.77808856591221e-01, 7.76810070967684e-01, 7.75527319351085e-01, 7.73915972102031e-01, 7.71931775185337e-01, 7.69532275036099e-01, 7.66678310741355e-01, 7.63335510481340e-01, 7.59475722165561e-01, 7.55078305366183e-01, 7.50131213987132e-01, 7.44631807118964e-01, 7.38587339067706e-01, 7.32015097750235e-01, 7.24942181981632e-01, 7.17404930624625e-01, 7.09448037893665e-01, 7.01123407186913e-01, 6.92488808965280e-01, 6.83606415367777e-01, 6.74541285177554e-01, 6.65359867919761e-01, 6.56128586383417e-01, 6.46912544220598e-01, 6.37774391142571e-01, 6.28773364170044e-01, 6.19964510694430e-01, 6.11398088663520e-01, 6.03119131488128e-01, 5.95167160347287e-01, 5.87576024200843e-01, 5.80373847552252e-01, 5.73583067282243e-01, 5.67220542143772e-01, 5.61297721269582e-01, 5.55820860903751e-01, 5.50791281239427e-01, 5.46205657551184e-01, 5.42056341649840e-01, 5.38331711030196e-01, 5.35016543926631e-01, 5.32092418865773e-01, 5.29538137244870e-01, 5.27330167012209e-01, 5.25443104735160e-01, 5.23850152268736e-01, 5.22523602961115e-01, 5.21435330939113e-01, 5.20557275612887e-01, 5.19861912244295e-01, 5.19322698363638e-01, 5.18914485117250e-01, 5.18613882398924e-01, 5.18399566940388e-01, 5.18252523467154e-01, 5.18156210562026e-01, 5.18096644981000e-01, 5.18062400736426e-01, 5.18044522168964e-01, 5.18036353307082e-01, 5.18033288885505e-01, 5.18032455288576e-01, 5.18032332251423e-01, 5.18032322727848e-01, 5.18032151276969e-01, 5.18030989672326e-01, 5.18026719495211e-01, 5.18015336804666e-01, 5.17990426236750e-01, 5.17942720177686e-01, 5.17859761000608e-01, 5.17725682440968e-01, 5.17521123941918e-01, 5.17223289119229e-01, 5.16806156237095e-01, 5.16240844599778e-01, 5.15496135925289e-01, 5.14539144026116e-01, 5.13336119539488e-01, 5.11853369239091e-01, 5.10058262011981e-01, 5.07920286455743e-01, 5.05412118943369e-01, 5.02510656675909e-01, 4.99197968423530e-01, 4.95462116907774e-01, 4.91297811386346e-01, 4.86706856872829e-01, 4.81698377057867e-01, 4.76288800508556e-01, 4.70501612955687e-01, 4.64366891167515e-01, 4.57920644875955e-01, 4.51204001543985e-01, 4.44262273916181e-01, 4.37143952180989e-01, 4.29899661503313e-01, 4.22581122265541e-01, 4.15240145339638e-01, 4.07927688864831e-01, 4.00692996976389e-01, 3.93582835188695e-01, 3.86640831958914e-01, 3.79906931458436e-01, 3.73416958757346e-01, 3.67202295434637e-01, 3.61289661004065e-01, 3.55700993453453e-01, 3.50453420623424e-01, 3.45559313099926e-01, 3.41026408763668e-01, 3.36857999109398e-01, 3.33053167863042e-01, 3.29607073197863e-01, 3.26511265860546e-01, 3.23754036625175e-01, 3.21320787554967e-01, 3.19194422435302e-01, 3.17355752346643e-01, 3.15783912600850e-01, 3.14456787150618e-01, 3.13351436119916e-01, 3.12444521360575e-01, 3.11712724021135e-01, 3.11133147144332e-01, 3.10683695433653e-01, 3.10343423687556e-01, 3.10092845115506e-01, 3.09914190920892e-01, 3.09791613218565e-01, 3.09711324559373e-01, 3.09661669026468e-01, 3.09633121967153e-01, 3.09618217812485e-01, 3.09611407977229e-01, 3.09608853372310e-01, 3.09608158457777e-01, 3.09608055889839e-01, 3.09608047886300e-01, 3.09607902968793e-01, 3.09606921131603e-01, 3.09603311797531e-01, 3.09593690663547e-01, 3.09572635176465e-01, 3.09532311895424e-01, 3.09462190922744e-01, 3.09348860956291e-01, 3.09175956613089e-01, 3.08924207416567e-01, 3.08571615122503e-01, 3.08093762770554e-01, 3.07464254902655e-01, 3.06655283754310e-01, 3.05638310948169e-01, 3.04384848460131e-01, 3.02867316659686e-01, 3.01059951439981e-01, 2.98939727327067e-01, 2.96487259523175e-01, 2.93687645605374e-01, 2.90531207500749e-01, 2.87014096662830e-01, 2.83138730149105e-01, 2.78914032368732e-01, 2.74355466232038e-01, 2.69484847689938e-01, 2.64329948478405e-01, 2.58923902513757e-01, 2.53304441080289e-01, 2.47512990069765e-01, 2.41593668570786e-01, 2.35592231724851e-01, 2.29555001802325e-01, 2.23527829916350e-01, 2.17555126878509e-01, 2.11678995754511e-01, 2.05938491201522e-01, 2.00369022260091e-01, 1.95001906591202e-01, 1.89864075842365e-01, 1.84977924471440e-01, 1.80361288406175e-01, 1.76027535652238e-01, 1.71985748488538e-01, 1.68240976134527e-01, 1.64794537532316e-01, 1.61644355843819e-01, 1.58785309060804e-01, 1.56209584390240e-01, 1.53907027465323e-01, 1.51865480649207e-01, 1.50071107509334e-01, 1.48508702787680e-01, 1.47161988777071e-01, 1.46013899907118e-01, 1.45046857567920e-01, 1.44243036826310e-01, 1.43584625830599e-01, 1.43054077492692e-01, 1.42634351645128e-01, 1.42309144460782e-01, 1.42063100663819e-01, 1.41882003096654e-01, 1.41752933663469e-01, 1.41664399618946e-01, 1.41606419649449e-01, 1.41570565177830e-01, 1.41549953748836e-01, 1.41539193111811e-01, 1.41534276576124e-01, 1.41532432228132e-01, 1.41531930522106e-01, 1.41531856471374e-01, 1.41531850643943e-01, 1.41531744498025e-01, 1.41531025344733e-01, 1.41528381674377e-01, 1.41521334721048e-01, 1.41505913149743e-01, 1.41476380964292e-01, 1.41425030407007e-01, 1.41342050800077e-01, 1.41215483283915e-01, 1.41031270105472e-01, 1.40773405331160e-01, 1.40424191430697e-01, 1.39964602946023e-01, 1.39374754337979e-01, 1.38634464106275e-01, 1.37723901546636e-01, 1.36624296335910e-01, 1.35318684957625e-01, 1.33792662358971e-01, 1.32035102806412e-01, 1.30038811331801e-01, 1.27801067010793e-01, 1.25324022009110e-01, 1.22614926040043e-01, 1.19686154464944e-01, 1.16555029272106e-01, 1.13243434798164e-01, 1.09777243268539e-01, 1.06185577832818e-01, 1.02499951546974e-01, 9.87533286350697e-02, 9.49791585634188e-02, 9.12104335817645e-02, 8.74788164961871e-02, 8.38138780443387e-02, 8.02424732303137e-02, 7.67882744765984e-02, 7.34714676709274e-02, 7.03086062594735e-02, 6.73126093653877e-02, 6.44928830967189e-02, 6.18555400044016e-02, 5.94036900025259e-02, 5.71377766617331e-02, 5.50559351471804e-02, 5.31543516488711e-02, 5.14276083873284e-02, 4.98690026946917e-02, 4.84708328900467e-02, 4.72246474208902e-02, 4.61214568756674e-02, 4.51519109134892e-02, 4.43064439044605e-02, 4.35753941701620e-02, 4.29491022282065e-02, 4.24179934554871e-02, 4.19726501825721e-02, 4.16038774945946e-02, 4.13027660345335e-02, 4.10607539610623e-02, 4.08696889967359e-02, 4.07218903017963e-02, 4.06102088088716e-02, 4.05280837440074e-02, 4.04695924030278e-02, 4.04294899086608e-02, 4.04032356629784e-02, 4.03870035334677e-02, 4.03776734297594e-02, 4.03728027850292e-02, 4.03705774637331e-02, 4.03697426877301e-02, 4.03695156101946e-02, 4.03694820940785e-02, 4.03694794331813e-02, 4.03694306684912e-02, 4.03691002814810e-02, 4.03678857592524e-02, 4.03646484067224e-02, 4.03575641379774e-02, 4.03439992073942e-02, 4.03204168495535e-02, 4.02823207379299e-02, 4.02242410493869e-02, 4.01397686028275e-02, 4.00216418573791e-02, 3.98618903330430e-02, 3.96520360937416e-02, 3.93833522088016e-02, 3.90471735759093e-02, 3.86352512693264e-02, 3.81401369438805e-02, 3.75555792048203e-02, 3.68769098024870e-02, 3.61013946551196e-02, 3.52285236604075e-02, 3.42602145260287e-02, 3.32009097039645e-02, 3.20575519090056e-02, 3.08394322280208e-02, 2.95579147111766e-02, 2.82260515268291e-02, 2.68581120667193e-02, 2.54690566524829e-02, 2.40739897979410e-02, 2.26876287918429e-02, 2.13238206351178e-02, 1.99951345402667e-02, 1.87125491322746e-02, 1.74852442891194e-02, 1.63204983905558e-02, 1.52236836358384e-02, 1.41983457938295e-02, 1.32463506640477e-02, 1.23680777100503e-02, 1.15626415538799e-02, 1.08281238625096e-02, 1.01618010883300e-02, 9.56035699265101e-03, 9.02007240739522e-03, 8.53698791485158e-03, 8.10703784121358e-03, 7.72615607925786e-03, 7.39035578344271e-03, 7.09578596872973e-03, 6.83876857009125e-03, 6.61581967789004e-03, 6.42365855632931e-03, 6.25920776928979e-03, 6.11958735976259e-03, 6.00210560246239e-03, 5.90424840586934e-03, 5.82366899510589e-03, 5.75817906654663e-03, 5.70574217969977e-03, 5.66446974447388e-03, 5.63261958539840e-03, 5.60859672934555e-03, 5.59095578260838e-03, 5.57840405342347e-03, 5.56980444810051e-03, 5.56417713202961e-03, 5.56069900299860e-03, 5.55870016560789e-03, 5.55765680738318e-03, 5.55718013748674e-03, 5.55700133010362e-03, 5.55695269090961e-03, 5.55694551188424e-03, 5.55694493672831e-03, 5.55693433021531e-03, 5.55686246990206e-03, 5.55659831022126e-03, 5.55589420638114e-03, 5.55435354054623e-03, 5.55140393211673e-03, 5.54627751307635e-03, 5.53799983904923e-03, 5.52538908561928e-03, 5.50706719218397e-03, 5.48148449432517e-03, 5.44695905374226e-03, 5.40173128568711e-03, 5.34403355256075e-03, 5.27217313978667e-03, 5.18462551373489e-03, 5.08013311257116e-03, 4.95780334435381e-03, 4.81719821661386e-03, 4.65840738033307e-03, 4.48209658567578e-03, 4.28952477716393e-03, 4.08252531677533e-03, 3.86344995312982e-03, 3.63507780701866e-03, 3.40049533605224e-03, 3.16295642332420e-03, 2.92573389739914e-03, 2.69197457382120e-03, 2.46456918478088e-03, 2.24604648306581e-03, 2.03849775884635e-03, 1.84353451940922e-03, 1.66227867954660e-03, 1.49538170687435e-03, 1.34306701953335e-03, 1.20518868344277e-03, 1.08129916209521e-03, 9.70719493892602e-04, 8.72606608439241e-04, 7.86014157867498e-04, 7.09944789562655e-04, 6.43392930697928e-04, 5.85377894566363e-04, 5.34967666027342e-04, 4.91294261507846e-04, 4.53562067846747e-04, 4.21050855678118e-04, 3.93115100791874e-04, 3.69180886593767e-04, 3.48741226233074e-04, 3.31350339607184e-04, 3.16617306651779e-04, 3.04199510779313e-04, 2.93796266149116e-04, 2.85142938466265e-04, 2.78005739971832e-04, 2.72177262303330e-04, 2.67472736311447e-04, 2.63726976015732e-04, 2.60791954218575e-04, 2.58534951655065e-04, 2.56837210471872e-04, 2.55593006031063e-04, 2.54709036966005e-04, 2.54104022289425e-04, 2.53708392989438e-04, 2.53463970764756e-04, 2.53323541755890e-04, 2.53250251105896e-04, 2.53216769934284e-04, 2.53204210935110e-04, 2.53200794674058e-04, 2.53200290444511e-04, 2.53200249673569e-04, 2.53199493002398e-04, 2.53194366379716e-04, 2.53175521223170e-04, 2.53125292942968e-04, 2.53015400033973e-04, 2.52805058565286e-04, 2.52439637634438e-04, 2.51849998079479e-04, 2.50952673525262e-04, 2.49651060983069e-04, 2.47837779937754e-04, 2.45398323173163e-04, 2.42216052185995e-04, 2.38178477488304e-04, 2.33184614714332e-04, 2.27153026514770e-04, 2.20029972339830e-04, 2.11796918918594e-04, 2.02476550128949e-04, 1.92136395952502e-04, 1.80889317419997e-04, 1.68890369593555e-04, 1.56330024651652e-04, 1.43424306358497e-04, 1.30402911167948e-04, 1.17496644435278e-04, 1.04925318765936e-04, 9.28867169562864e-05, 8.15467163448602e-05, 7.10307868843074e-05, 6.14181075917486e-05, 5.27408363950599e-05, 4.49908772708341e-05, 3.81335581778330e-05, 3.21234966094153e-05, 2.69162193262673e-05, 2.24718191503825e-05, 1.87536982423353e-05, 1.57338906348059e-05, 1.33376358917303e-05, 1.13709881759405e-05, 9.71826128193243e-06, 8.31091742270473e-06, 7.11902887928773e-06, 6.20473902301910e-06, 5.45499314341669e-06, 4.80547507208939e-06, 4.23117025150077e-06, 3.71802424581347e-06, 3.26470350842101e-06, 2.87044586855245e-06, 2.52100268482991e-06, 2.20082787905085e-06, 1.90496457719291e-06, 1.63622337367997e-06, 1.39793972664393e-06, 1.18742522658977e-06, 1.00117581780523e-06, 8.39190320789385e-07, 6.97805674576557e-07, 5.72264042375537e-07, 4.60554417713865e-07, 3.62261872055569e-07, 2.77399934134337e-07, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -2.84217094304040e-14],
[ 7.60000000000000e+01, 7.59999999152946e+01, 7.59999972960353e+01, 7.59999795501382e+01, 7.59999143138923e+01, 7.59997404145491e+01, 7.59993598225166e+01, 7.59986308759073e+01, 7.59973630209395e+01, 7.59953132646287e+01, 7.59921844830769e+01, 7.59876256717124e+01, 7.59812341656617e+01, 7.59725598018794e+01, 7.59611109425281e+01, 7.59463622337922e+01, 7.59277639374474e+01, 7.59047526444848e+01, 7.58767631597835e+01, 7.58432413315259e+01, 7.58036575847349e+01, 7.57575209003681e+01, 7.57043929557452e+01, 7.56439021064319e+01, 7.55757568448956e+01, 7.54997583219862e+01, 7.54158114721644e+01, 7.53239342538338e+01, 7.52242645144222e+01, 7.51170640264907e+01, 7.50027193221683e+01, 7.48817390781049e+01, 7.47547479640974e+01, 7.46224770511795e+01, 7.44857510605490e+01, 7.43454729033583e+01, 7.42026060955688e+01, 7.40581557194118e+01, 7.39131486380913e+01, 7.37686136549624e+01, 7.36255622503977e+01, 7.34849704407746e+01, 7.33477621979573e+01, 7.32147947570473e+01, 7.30868460353722e+01, 7.29646042936833e+01, 7.28486600948435e+01, 7.27395005564086e+01, 7.26375058496959e+01, 7.25429478661614e+01, 7.24559909486730e+01, 7.23766945672416e+01, 7.23050178032058e+01, 7.22408254907876e+01, 7.21838958491630e+01, 7.21339294212680e+01, 7.20905591176397e+01, 7.20533611452850e+01, 7.20218665838276e+01, 7.19955733551977e+01, 7.19739583201933e+01, 7.19564892267185e+01, 7.19426362316545e+01, 7.19318827222943e+01, 7.19237351749302e+01, 7.19177318080779e+01, 7.19134498160679e+01, 7.19105110050000e+01, 7.19085856965227e+01, 7.19073948142920e+01, 7.19067101216203e+01, 7.19063526347775e+01, 7.19061892925328e+01, 7.19061280166427e+01, 7.19061113480605e+01, 7.19061088878108e+01, 7.19061085650779e+01, 7.19061010457890e+01, 7.19060501014840e+01, 7.19058628245808e+01, 7.19053636112053e+01, 7.19042710852499e+01, 7.19021787302052e+01, 7.18985399910347e+01, 7.18926585093286e+01, 7.18836840407665e+01, 7.18706144733984e+01, 7.18523042165702e+01, 7.18274790614886e+01, 7.17947574240851e+01, 7.17526776688895e+01, 7.16997309808856e+01, 7.16343990050918e+01, 7.15551952183923e+01, 7.14607087465155e+01, 7.13496491077931e+01, 7.12208901772529e+01, 7.10735115483448e+01, 7.09068354569708e+01, 7.07204575525177e+01, 7.05142700701976e+01, 7.02884763725204e+01, 7.00435963489707e+01, 6.97804627249156e+01, 6.95002088462923e+01, 6.92042488900538e+01, 6.88942516448030e+01, 6.85721090066950e+01, 6.82399001984493e+01, 6.78998525473562e+01, 6.75542995663794e+01, 6.72056371495914e+01, 6.68562789205696e+01, 6.65086120711573e+01, 6.61649552492729e+01, 6.58275200565216e+01, 6.54983774377970e+01, 6.51794297414488e+01, 6.48723886475480e+01, 6.45787586701013e+01, 6.42998256472708e+01, 6.40366495605430e+01, 6.37900611121054e+01, 6.35606616508762e+01, 6.33488261944662e+01, 6.31547094025541e+01, 6.29782544053155e+01, 6.28192043883951e+01, 6.26771168010577e+01, 6.25513800030092e+01, 6.24412321092644e+01, 6.23457817373874e+01, 6.22640303097501e+01, 6.21948955156606e+01, 6.21372354945329e+01, 6.20898732625191e+01, 6.20516208729826e+01, 6.20213027784458e+01, 6.19977778513171e+01, 6.19799595259410e+01, 6.19668335480333e+01, 6.19574728611774e+01, 6.19510492243618e+01, 6.19468412386623e+01, 6.19442385627638e+01, 6.19427422123034e+01, 6.19419609621245e+01, 6.19416039978100e+01, 6.19414700874187e+01, 6.19414336604743e+01, 6.19414282839310e+01, 6.19414278092846e+01, 6.19414185088214e+01, 6.19413554968010e+01, 6.19411238577463e+01, 6.19405063915171e+01, 6.19391550738833e+01, 6.19365671086973e+01, 6.19320665207980e+01, 6.19247921197224e+01, 6.19136925263846e+01, 6.18975287955933e+01, 6.18748849829163e+01, 6.18441867910863e+01, 6.18037281861729e+01, 6.17517055971605e+01, 6.16862590087825e+01, 6.16055189364919e+01, 6.15076579507142e+01, 6.13909451173717e+01, 6.12538014696672e+01, 6.10948544497317e+01, 6.09129891818293e+01, 6.07073944766309e+01, 6.04776016214496e+01, 6.02235142734683e+01, 5.99454281202792e+01, 5.96440393784760e+01, 5.93204416439173e+01, 5.89761110723626e+01, 5.86128803496636e+01, 5.82329023994753e+01, 5.78386052554741e+01, 5.74326399581776e+01, 5.70178236717417e+01, 5.65970803987254e+01, 5.61733816618778e+01, 5.57496893148371e+01, 5.53289022701990e+01, 5.49138084574467e+01, 5.45070428231928e+01, 5.41110517345020e+01, 5.37280637924872e+01, 5.33600668272487e+01, 5.30087907178918e+01, 5.26756956357509e+01, 5.23619653118998e+01, 5.20685049525847e+01, 5.17959434494535e+01, 5.15446395471621e+01, 5.13146916388687e+01, 5.11059508636719e+01, 5.09180371826543e+01, 5.07503581131262e+01, 5.06021298027526e+01, 5.04724001235525e+01, 5.03600734570047e+01, 5.02639368232024e+01, 5.01826869783769e+01, 5.01149580673707e+01, 5.00593493739365e+01, 5.00144526668212e+01, 4.99788785992530e+01, 4.99512815898951e+01, 4.99303826005390e+01, 4.99149892348681e+01, 4.99040126171864e+01, 4.98964805718355e+01, 4.98915467127305e+01, 4.98884951654210e+01, 4.98867407766826e+01, 4.98858248125520e+01, 4.98854062975270e+01, 4.98852492975345e+01, 4.98852065896772e+01, 4.98852002860850e+01, 4.98851997804589e+01, 4.98851904490932e+01, 4.98851272277104e+01, 4.98848948191091e+01, 4.98842753022593e+01, 4.98829195004854e+01, 4.98803229614102e+01, 4.98758075065926e+01, 4.98685091926464e+01, 4.98573733909310e+01, 4.98411575366617e+01, 4.98184419162563e+01, 4.97876486494745e+01, 4.97470687771925e+01, 4.96948970859938e+01, 4.96292739908799e+01, 4.95483334664386e+01, 4.94502556799025e+01, 4.93333226575906e+01, 4.91959750346918e+01, 4.90368677248129e+01, 4.88549222267973e+01, 4.86493732840803e+01, 4.84198077405395e+01, 4.81661937006655e+01, 4.78888984943880e+01, 4.75886944513193e+01, 4.72667520797283e+01, 4.69246208877898e+01, 4.65641987347600e+01, 4.61876912037320e+01, 4.57975629836570e+01, 4.53964835755407e+01, 4.49872697524959e+01, 4.45728270973753e+01, 4.41560926521445e+01, 4.37399803155215e+01, 4.33273302091723e+01, 4.29208628704602e+01, 4.25231388552050e+01, 4.21365241398692e+01, 4.17631615662779e+01, 4.14049484375961e+01, 4.10635202304631e+01, 4.07402402338332e+01, 4.04361947747827e+01, 4.01521935656692e+01, 3.98887746222695e+01, 3.96462131664586e+01, 3.94245339369790e+01, 3.92235263773836e+01, 3.90427622366470e+01, 3.88816151900564e+01, 3.87392821528430e+01, 3.86148060072579e+01, 3.85070994901376e+01, 3.84149699909266e+01, 3.83371449912134e+01, 3.82722978400986e+01, 3.82190735108500e+01, 3.81761139301146e+01, 3.81420824188038e+01, 3.81156867409963e+01, 3.80957002306526e+01, 3.80809804613705e+01, 3.80704849460166e+01, 3.80632834030490e+01, 3.80585662045994e+01, 3.80556487254839e+01, 3.80539714376089e+01, 3.80530957343203e+01, 3.80526956164604e+01, 3.80525455181472e+01, 3.80525046877393e+01, 3.80524986612556e+01, 3.80524981964352e+01, 3.80524898497526e+01, 3.80524332997498e+01, 3.80522254159533e+01, 3.80516712740466e+01, 3.80504585478852e+01, 3.80481360316223e+01, 3.80440971545321e+01, 3.80375692505245e+01, 3.80276092273980e+01, 3.80131061442653e+01, 3.79927910447309e+01, 3.79652542034954e+01, 3.79289697214349e+01, 3.78823271477949e+01, 3.78236695208861e+01, 3.77513369087172e+01, 3.76637142122657e+01, 3.75592816862321e+01, 3.74366663595347e+01, 3.72946923277401e+01, 3.71324277695078e+01, 3.69492265332714e+01, 3.67447622665497e+01, 3.65190533264577e+01, 3.62724771117673e+01, 3.60057729758418e+01, 3.57200334831809e+01, 3.54166844143940e+01, 3.50974545496376e+01, 3.47643368093041e+01, 3.44195427470828e+01, 3.40654526313941e+01, 3.37045633952306e+01, 3.33394365878368e+01, 3.29726481588820e+01, 3.26067415037148e+01, 3.22441847645384e+01, 3.18873329806369e+01, 3.15383953574092e+01, 3.11994076996687e+01, 3.08722099239144e+01, 3.05584285021527e+01, 3.02594636633347e+01, 2.99764811582307e+01, 2.97104083621325e+01, 2.94619344439766e+01, 2.92315142784248e+01, 2.90193757321908e+01, 2.88255299290104e+01, 2.86497840947712e+01, 2.84917566039918e+01, 2.83508938839432e+01, 2.82264888732692e+01, 2.81177007680836e+01, 2.80235758124781e+01, 2.79430688976728e+01, 2.78750657235637e+01, 2.78184052500131e+01, 2.77719021269885e+01, 2.77343687482522e+01, 2.77046365292190e+01, 2.76815759725538e+01, 2.76641150613963e+01, 2.76512555151619e+01, 2.76420864606426e+01, 2.76357951137568e+01, 2.76316741348331e+01, 2.76291254107111e+01, 2.76276601261470e+01, 2.76268951093776e+01, 2.76265455654874e+01, 2.76264144392784e+01, 2.76263787697487e+01, 2.76263735050001e+01, 2.76263731063692e+01, 2.76263660446629e+01, 2.76263182005622e+01, 2.76261423205693e+01, 2.76256734889471e+01, 2.76246474617213e+01, 2.76226824942046e+01, 2.76192653909696e+01, 2.76137424216078e+01, 2.76053156461615e+01, 2.75930450781504e+01, 2.75758569768891e+01, 2.75525584004452e+01, 2.75218579631863e+01, 2.74823925266561e+01, 2.74327593120942e+01, 2.73715526641273e+01, 2.72974044296025e+01, 2.72090266594419e+01, 2.71052551147333e+01, 2.69850918831723e+01, 2.68477453104971e+01, 2.66926654432105e+01, 2.65195732780270e+01, 2.63284823271633e+01, 2.61197113346805e+01, 2.58938874048522e+01, 2.56519393044067e+01, 2.53950812394560e+01, 2.51247879364120e+01, 2.48427623182697e+01, 2.45508974087556e+01, 2.42512342766114e+01, 2.39459178376118e+01, 2.36371521841834e+01, 2.33271568635289e+01, 2.30181252412195e+01, 2.27121858270397e+01, 2.24113672366640e+01, 2.21175673180789e+01, 2.18325268628392e+01, 2.15578082179232e+01, 2.12947789897369e+01, 2.10446008795148e+01, 2.08082235176692e+01, 2.05863829934815e+01, 2.03796046287878e+01, 2.01882094385616e+01, 2.00123236674662e+01, 1.98518907898488e+01, 1.97066854034849e+01, 1.95763285216419e+01, 1.94603038586232e+01, 1.93579747963661e+01, 1.92686018017968e+01, 1.91913601278229e+01, 1.91253576701353e+01, 1.90696528660184e+01, 1.90232725118706e+01, 1.89852293473286e+01, 1.89545392118312e+01, 1.89302375313116e+01, 1.89113948460771e+01, 1.88971310531413e+01, 1.88866280137393e+01, 1.88791401743960e+01, 1.88740028707237e+01, 1.88706380278897e+01, 1.88685570389568e+01, 1.88673606885309e+01, 1.88667360891507e+01, 1.88664507053022e+01, 1.88663436480923e+01, 1.88663145259471e+01, 1.88663102275799e+01, 1.88663099049001e+01, 1.88663042253704e+01, 1.88662657457320e+01, 1.88661242906730e+01, 1.88657472246692e+01, 1.88649220312776e+01, 1.88633417114957e+01, 1.88605935981619e+01, 1.88561521207216e+01, 1.88493759814368e+01, 1.88395101170339e+01, 1.88256927145879e+01, 1.88069674226881e+01, 1.87823007451302e+01, 1.87506044230846e+01, 1.87107624046137e+01, 1.86616617732091e+01, 1.86022267696808e+01, 1.85314548084977e+01, 1.84484531782274e+01, 1.83524749459049e+01, 1.82429524770720e+01, 1.81195269551405e+01, 1.79820723498067e+01, 1.78307124525734e+01, 1.76658298683954e+01, 1.74880662176565e+01, 1.72983132444448e+01, 1.70976950187116e+01, 1.68875419266688e+01, 1.66693576251337e+01, 1.64447805485342e+01, 1.62155418614612e+01, 1.59834219126984e+01, 1.57502072501313e+01, 1.55176500994333e+01, 1.52874319120178e+01, 1.50611321858212e+01, 1.48402033041551e+01, 1.46259516751758e+01, 1.44195250350726e+01, 1.42219054382450e+01, 1.40339072185336e+01, 1.38561790720348e+01, 1.36892093754075e+01, 1.35333338955637e+01, 1.33887451441844e+01, 1.32555027602292e+01, 1.31335444449751e+01, 1.30226971112167e+01, 1.29226880302279e+01, 1.28331558606843e+01, 1.27536615203339e+01, 1.26836989136802e+01, 1.26227055587332e+01, 1.25700731653018e+01, 1.25251582090742e+01, 1.24872925228612e+01, 1.24557938921001e+01, 1.24299765995343e+01, 1.24091618177190e+01, 1.23926877017373e+01, 1.23799189924705e+01, 1.23702559070949e+01, 1.23631420719172e+01, 1.23580712462038e+01, 1.23545925962181e+01, 1.23523143068048e+01, 1.23509053626069e+01, 1.23500953899651e+01, 1.23496725200580e+01, 1.23494793093157e+01, 1.23494068296212e+01, 1.23493871134176e+01, 1.23493842033484e+01, 1.23493839855529e+01, 1.23493801609551e+01, 1.23493542487590e+01, 1.23492589930356e+01, 1.23490050781757e+01, 1.23484494014852e+01, 1.23473852515820e+01, 1.23455348031075e+01, 1.23425443008671e+01, 1.23379822621250e+01, 1.23313409710352e+01, 1.23220414696921e+01, 1.23094421613581e+01, 1.22928510291509e+01, 1.22715413357962e+01, 1.22447705079966e+01, 1.22118017275617e+01, 1.21719275604071e+01, 1.21244947681893e+01, 1.20689292839243e+01, 1.20047602128075e+01, 1.19316416626901e+01, 1.18493712318929e+01, 1.17579040952276e+01, 1.16573618327439e+01, 1.15480354291658e+01, 1.14303822133767e+01, 1.13050168756247e+01, 1.11726970590632e+01, 1.10343043355630e+01, 1.08908216129568e+01, 1.07433081621499e+01, 1.05928734914938e+01, 1.04406512397990e+01, 1.02877741267210e+01, 1.01353508148750e+01, 9.98444532796071e+00, 9.83605945612808e+00, 9.69111838040311e+00, 9.55045957225410e+00, 9.41482487697212e+00, 9.28485557188229e+00, 9.16109010252373e+00, 9.04396414150931e+00, 8.93381258512346e+00, 8.83087310022863e+00, 8.73529085556513e+00, 8.64712411215388e+00, 8.56635040100964e+00, 8.49287307598601e+00, 8.42652808892527e+00, 8.36709088811415e+00, 8.31428338559212e+00, 8.26778097191249e+00, 8.22721957766797e+00, 8.19220278967762e+00, 8.16230902718378e+00, 8.13709877126907e+00, 8.11612182093221e+00, 8.09892452413344e+00, 8.08505690417624e+00, 8.07407957375608e+00, 8.06557030367593e+00, 8.05913009331121e+00, 8.05438857782643e+00, 8.05100860481070e+00, 8.04868982160606e+00, 8.04717113446181e+00, 8.04623193111581e+00, 8.04569199789640e+00, 8.04541010857545e+00, 8.04528131203912e+00, 8.04523299617725e+00, 8.04521985310693e+00, 8.04521791321785e+00, 8.04521776797829e+00, 8.04521521677243e+00, 8.04519793198661e+00, 8.04513439142214e+00, 8.04496501665602e+00, 8.04459434945036e+00, 8.04388449782046e+00, 8.04265012375039e+00, 8.04065521854409e+00, 8.03761188107585e+00, 8.03318127673008e+00, 8.02697690793862e+00, 8.01857026811199e+00, 8.00749887707354e+00, 7.99327660666530e+00, 7.97540610073064e+00, 7.95339297741489e+00, 7.92676137976093e+00, 7.89507032186457e+00, 7.85793017371174e+00, 7.81501855088328e+00, 7.76609483801668e+00, 7.71101258761793e+00, 7.64972910486163e+00, 7.58231165505870e+00, 7.50893990743434e+00, 7.42990444384205e+00, 7.34560139533183e+00, 7.25652350072700e+00, 7.16324808652959e+00, 7.06642262638145e+00, 6.96674863680029e+00, 6.86496469815687e+00, 6.76182935907538e+00, 6.65810459978610e+00, 6.55454041184649e+00, 6.45186091616681e+00, 6.35075230473977e+00, 6.25185276601569e+00, 6.15574444651024e+00, 6.06294741445929e+00, 5.97391552446959e+00, 5.88903403299523e+00, 5.80861878070201e+00, 5.73291673741880e+00, 5.66210769697973e+00, 5.59630691165981e+00, 5.53556846780546e+00, 5.47988922400188e+00, 5.42921315859282e+00, 5.38343600212023e+00, 5.34241005968364e+00, 5.30594915584174e+00, 5.27383365832864e+00, 5.24581555486041e+00, 5.22162356859249e+00, 5.20096830191258e+00, 5.18354739541579e+00, 5.16905067989251e+00, 5.15716528527631e+00, 5.14758065347823e+00, 5.13999338390161e+00, 5.13411182335764e+00, 5.12966029821774e+00, 5.12638287786066e+00, 5.12404655630766e+00, 5.12244374434422e+00, 5.12139397768351e+00, 5.12074476741341e+00, 5.12037154598383e+00, 5.12017669366621e+00, 5.12008766469596e+00, 5.12005426696287e+00, 5.12004518197983e+00, 5.12004384105573e+00, 5.12004374039435e+00, 5.12004196866982e+00, 5.12002996498157e+00, 5.11998583824595e+00, 5.11986821327817e+00, 5.11961079750698e+00, 5.11911782928192e+00, 5.11826059650117e+00, 5.11687519615668e+00, 5.11476168400360e+00, 5.11168473756835e+00, 5.10737592333882e+00, 5.10153761837442e+00, 5.09384858620098e+00, 5.08397114597102e+00, 5.07155980298565e+00, 5.05627112991934e+00, 5.03777460543300e+00, 5.01576403607744e+00, 4.98996911572671e+00, 4.96016662229601e+00, 4.92619072212099e+00, 4.88794185477518e+00, 4.84539370956271e+00, 4.79859788039660e+00, 4.74768589533788e+00, 4.69286845391416e+00, 4.63443185930932e+00, 4.57273179120334e+00, 4.50818471514879e+00, 4.44125735326377e+00, 4.37245473812999e+00, 4.30230742975601e+00, 4.23135849088784e+00, 4.16015078961084e+00, 4.08921513484019e+00, 4.01905965796838e+00, 3.95016074285563e+00, 3.88295568764924e+00, 3.81783716631251e+00, 3.75514945432489e+00, 3.69518629843618e+00, 3.63819024839755e+00, 3.58435323022172e+00, 3.53381812430568e+00, 3.48668111448792e+00, 3.44299459159826e+00, 3.40277042273419e+00, 3.36598343101184e+00, 3.33257496608672e+00, 3.30245648025608e+00, 3.27551305614479e+00, 3.25160685826664e+00, 3.23058050118735e+00, 3.21226034114214e+00, 3.19645970574724e+00, 3.18298207818803e+00, 3.17162424856551e+00, 3.16217943678865e+00, 3.15444037960512e+00, 3.14820236035222e+00, 3.14326614522660e+00, 3.13944077582541e+00, 3.13654615589154e+00, 3.13441536195305e+00, 3.13289660395975e+00, 3.13185476380168e+00, 3.13117244698562e+00, 3.13075049549531e+00, 3.13050792722422e+00, 3.13038128818876e+00, 3.13032342655815e+00, 3.13030172079080e+00, 3.13029581630786e+00, 3.13029494481900e+00, 3.13029487911888e+00, 3.13029371903015e+00, 3.13028585926482e+00, 3.13025696603377e+00, 3.13017994803916e+00, 3.13001139985339e+00, 3.12968862451974e+00, 3.12912736121251e+00, 3.12822033320924e+00, 3.12683671878877e+00, 3.12482263181147e+00, 3.12200267816726e+00, 3.11818262740499e+00, 3.11315320434352e+00, 3.10669496258519e+00, 3.09858415087088e+00, 3.08859942569747e+00, 3.07652920261880e+00, 3.06217937870636e+00, 3.04538110551995e+00, 3.02599825210056e+00, 3.00393417733156e+00, 2.97913743589089e+00, 2.95160607529425e+00, 2.92139024372284e+00, 2.88859291659699e+00, 2.85336865796297e+00, 2.81592045159505e+00, 2.77649475535497e+00, 2.73537503947534e+00, 2.69287415490159e+00, 2.64932593409181e+00, 2.60507644977200e+00, 2.56047534707749e+00, 2.51586762498070e+00, 2.47158618044676e+00, 2.42794535157016e+00, 2.38523561257722e+00, 2.34371949183007e+00, 2.30362871008624e+00, 2.26516247457316e+00, 2.22848681730970e+00, 2.19373483421296e+00, 2.16100766421012e+00, 2.13037604323577e+00, 2.10188227445464e+00, 2.07554247083073e+00, 2.05134894671187e+00, 2.02927265895299e+00, 2.00926562303289e+00, 1.99126325372586e+00, 1.97518660166446e+00, 1.96094447548067e+00, 1.94843545345171e+00, 1.93754979838367e+00, 1.92817129482727e+00, 1.92017902888001e+00, 1.91344912823679e+00, 1.90785647441989e+00, 1.90327639100320e+00, 1.89958630200391e+00, 1.89666734439075e+00, 1.89440590882505e+00, 1.89269507427315e+00, 1.89143589587394e+00, 1.89053850212611e+00, 1.88992295755122e+00, 1.88951985067436e+00, 1.88927057431037e+00, 1.88912727529501e+00, 1.88905246323648e+00, 1.88901828166071e+00, 1.88900545908024e+00, 1.88900197103771e+00, 1.88900145621029e+00, 1.88900141718710e+00, 1.88900072533986e+00, 1.88899603797847e+00, 1.88897880682278e+00, 1.88893287550655e+00, 1.88883235908291e+00, 1.88863987013303e+00, 1.88830516786663e+00, 1.88776430105511e+00, 1.88693930890619e+00, 1.88573853522293e+00, 1.88405759928426e+00, 1.88178105068201e+00, 1.87878471384238e+00, 1.87493870058181e+00, 1.87011103590919e+00, 1.86417180443797e+00, 1.85699768429787e+00, 1.84847669550408e+00, 1.83851295437437e+00, 1.82703119932160e+00, 1.81398084065977e+00, 1.79933929168190e+00, 1.78311436249881e+00, 1.76534554223409e+00, 1.74610405704935e+00, 1.72549166662087e+00, 1.70363824358584e+00, 1.68069826132175e+00, 1.65684638710276e+00, 1.63227243288605e+00, 1.60717594921078e+00, 1.58176075603824e+00, 1.55622968791100e+00, 1.53077979260729e+00, 1.50559816801204e+00, 1.48085855828297e+00, 1.45671876508453e+00, 1.43331886954755e+00, 1.41078021095140e+00, 1.38920503204730e+00, 1.36867667934405e+00, 1.34926023854473e+00, 1.33100348829514e+00, 1.31393806645728e+00, 1.29808075915437e+00, 1.28343484111868e+00, 1.26999141427063e+00, 1.25773070849093e+00, 1.24662332332852e+00, 1.23663140149293e+00, 1.22770973431687e+00, 1.21980680604133e+00, 1.21286578798002e+00, 1.20682549561207e+00, 1.20162132169155e+00, 1.19718615681576e+00, 1.19345130583066e+00, 1.19034740426369e+00, 1.18780533398572e+00, 1.18575713187745e+00, 1.18413687981743e+00, 1.18288155925210e+00, 1.18193184938882e+00, 1.18123284508056e+00, 1.18073466908679e+00, 1.18039295382614e+00, 1.18016917007381e+00, 1.18003078422259e+00, 1.17995123148114e+00, 1.17990969934158e+00, 1.17989072332036e+00, 1.17988360481733e+00, 1.17988166841735e+00, 1.17988138260895e+00, 1.17988136080763e+00, 1.17988097247533e+00, 1.17987834146982e+00, 1.17986866965983e+00, 1.17984288850771e+00, 1.17978646882405e+00, 1.17967842503612e+00, 1.17949055680725e+00, 1.17918696770047e+00, 1.17872389676222e+00, 1.17804989391654e+00, 1.17710636328231e+00, 1.17582848927816e+00, 1.17414654808586e+00, 1.17198759137386e+00, 1.16927747022169e+00, 1.16594314557911e+00, 1.16191520862962e+00, 1.15713051203676e+00, 1.15153479369641e+00, 1.14508516101084e+00, 1.13775229845697e+00, 1.12952226637390e+00, 1.12039777546730e+00, 1.11039884916805e+00, 1.09956282282421e+00, 1.08794367147483e+00, 1.07561070237793e+00, 1.06264668987400e+00, 1.04914556422855e+00, 1.03520978948593e+00, 1.02094757624374e+00, 1.00647007346463e+00, 9.91888670381982e-01, 9.77312517812607e-01, 9.62846351003179e-01, 9.48588666861383e-01, 9.34630280054577e-01, 9.21053257335060e-01, 9.07930209107931e-01, 8.95323902424579e-01, 8.83287150310110e-01, 8.71862928137389e-01, 8.61084667821970e-01, 8.50976683943202e-01, 8.41554691469241e-01, 8.32826381627342e-01, 8.24792029797266e-01, 8.17445116467908e-01, 8.10772948812740e-01, 8.04757275999307e-01, 7.99374895777731e-01, 7.94598253130602e-01, 7.90396033822276e-01, 7.86733756629062e-01, 7.83574367965755e-01, 7.80878841674166e-01, 7.78606785056098e-01, 7.76717049977947e-01, 7.75168345238257e-01, 7.73919843574532e-01, 7.72931773920007e-01, 7.72165987038377e-01, 7.71586480697252e-01, 7.71159869303688e-01, 7.70855782588896e-01, 7.70647178605316e-01, 7.70510558023688e-01, 7.70426069431139e-01, 7.70377498899535e-01, 7.70352141290473e-01, 7.70340555323187e-01, 7.70336209048494e-01, 7.70335026758431e-01, 7.70334852254976e-01, 7.70334838850005e-01, 7.70334598844826e-01, 7.70332972775847e-01, 7.70326995193614e-01, 7.70311061302891e-01, 7.70276191128005e-01, 7.70209413465224e-01, 7.70093295602706e-01, 7.69905642564837e-01, 7.69619386978164e-01, 7.69202687318341e-01, 7.68619247963315e-01, 7.67828868811776e-01, 7.66788224936690e-01, 7.65451867680682e-01, 7.63773427822457e-01, 7.61706989316082e-01, 7.59208589318689e-01, 7.56237787820353e-01, 7.52759239478209e-01, 7.48744192689486e-01, 7.44171837904162e-01, 7.39030429776776e-01, 7.33318116553249e-01, 7.27043424933228e-01, 7.20225368594853e-01, 7.12893171919831e-01, 7.05085624964653e-01, 6.96850108888697e-01, 6.88241350503716e-01, 6.79319978454004e-01, 6.70150960638409e-01, 6.60802002595034e-01, 6.51341980311727e-01, 6.41839469636139e-01, 6.32361419910481e-01, 6.22972003570524e-01, 6.13731657992943e-01, 6.04696322261717e-01, 5.95916860674379e-01, 5.87438657135408e-01, 5.79301360046713e-01, 5.71538755524813e-01, 5.64178747168114e-01, 5.57243422509842e-01, 5.50749189094617e-01, 5.44706966285119e-01, 5.39122422037877e-01, 5.33996246723120e-01, 5.29324458447860e-01, 5.25098736210154e-01, 5.21306778554930e-01, 5.17932686240002e-01, 5.14957367788540e-01, 5.12358966741246e-01, 5.10113308969309e-01, 5.08194367612163e-01, 5.06574742119440e-01, 5.05226146572085e-01, 5.04119901022719e-01, 5.03227418135737e-01, 5.02520676046430e-01, 5.01972667226127e-01, 5.01557812368077e-01, 5.01252328015497e-01, 5.01034536928259e-01, 5.00885111084486e-01, 5.00787238743840e-01, 5.00726709115284e-01, 5.00691910778073e-01, 5.00673742964622e-01, 5.00665441953075e-01, 5.00662327959077e-01, 5.00661480877412e-01, 5.00661355849928e-01, 5.00661346172294e-01, 5.00661171947925e-01, 5.00659991552610e-01, 5.00655652300026e-01, 5.00644085487949e-01, 5.00618772016168e-01, 5.00570294507696e-01, 5.00485994273191e-01, 5.00349749549429e-01, 5.00141889197451e-01, 4.99839253353471e-01, 4.99415409232114e-01, 4.98841026224241e-01, 4.98084409478068e-01, 4.97112185243338e-01, 4.95890124456537e-01, 4.94384083574062e-01, 4.92561033927856e-01, 4.90390143473434e-01, 4.87843868460203e-01, 4.84899008077150e-01, 4.81537673276108e-01, 4.77748122335129e-01, 4.73525420580557e-01, 4.68871889927385e-01, 4.63797324977790e-01, 4.58318965397069e-01, 4.52461227949300e-01, 4.46255214611298e-01, 4.39738024383175e-01, 4.32951904860488e-01, 4.25943284819951e-01, 4.18761730942492e-01, 4.11458870699803e-01, 4.04087319990836e-01, 3.96699649093179e-01, 3.89347414628844e-01, 3.82080279155090e-01, 3.74945234108026e-01, 3.67985936405645e-01, 3.61242164163670e-01, 3.54749392716016e-01, 3.48538488450359e-01, 3.42635514868590e-01, 3.37061642776348e-01, 3.31833154633330e-01, 3.26961531887548e-01, 3.22453613580393e-01, 3.18311814616635e-01, 3.14534392763292e-01, 3.11115754551864e-01, 3.08046791649417e-01, 3.05315240763179e-01, 3.02906061571846e-01, 3.00801828384579e-01, 2.98983132091266e-01, 2.97428989412296e-01, 2.96117256457151e-01, 2.95025043184168e-01, 2.94129124595216e-01, 2.93406343509702e-01, 2.92833998683918e-01, 2.92390211031957e-01, 2.92054259911353e-01, 2.91806881005826e-01, 2.91630517366528e-01, 2.91509515736180e-01, 2.91430261386424e-01, 2.91381246319588e-01, 2.91353067736850e-01, 2.91338356039162e-01, 2.91331634157132e-01, 2.91329112549406e-01, 2.91328426611187e-01, 2.91328325368156e-01, 2.91328317467986e-01, 2.91328174422432e-01, 2.91327205268033e-01, 2.91323642559249e-01, 2.91314145726318e-01, 2.91293362345968e-01, 2.91253560476576e-01, 2.91184347165611e-01, 2.91072486699264e-01, 2.90901830361250e-01, 2.90653367306504e-01, 2.90305403514992e-01, 2.89833872547192e-01, 2.89212777870389e-01, 2.88414761825493e-01, 2.87411790898290e-01, 2.86175941010328e-01, 2.84680260335136e-01, 2.82899681091507e-01, 2.80811946365759e-01, 2.78398513841765e-01, 2.75645395911084e-01, 2.72543895466754e-01, 2.69091199054898e-01, 2.65290794055588e-01, 2.61152684024609e-01, 2.56693385836599e-01, 2.51935703202008e-01, 2.46908282714926e-01, 2.41644969981110e-01, 2.36183993757138e-01, 2.30567014662995e-01, 2.24838081322503e-01, 2.19042540336972e-01, 2.13225947112662e-01, 2.07433022274614e-01, 2.01706693462043e-01, 1.96087255170042e-01, 1.90611670609442e-01, 1.85313030034735e-01, 1.80220170421407e-01, 1.75357452488754e-01, 1.70744683483690e-01, 1.66397168303405e-01, 1.62325867675293e-01, 1.58537640255787e-01, 1.55035545495664e-01, 1.51819185650780e-01, 1.48885068007331e-01, 1.46226971825214e-01, 1.43836308275766e-01, 1.41702465405515e-01, 1.39813133608428e-01, 1.38154610019456e-01, 1.36712082519949e-01, 1.35469895600266e-01, 1.34411801149938e-01, 1.33521197386642e-01, 1.32781358671494e-01, 1.32175658013665e-01, 1.31687782776709e-01, 1.31301942627360e-01, 1.31003067275799e-01, 1.30776990208659e-01, 1.30610613556797e-01, 1.30492048586390e-01, 1.30410726135921e-01, 1.30357471673221e-01, 1.30324540502201e-01, 1.30305609944187e-01, 1.30295726949669e-01, 1.30291211433544e-01, 1.30289517524645e-01, 1.30289056741958e-01, 1.30288988731465e-01, 1.30288983379344e-01, 1.30288885891616e-01, 1.30288225398871e-01, 1.30285797372622e-01, 1.30279325260886e-01, 1.30265161785204e-01, 1.30238039267326e-01, 1.30190879967813e-01, 1.30114676763384e-01, 1.29998453679377e-01, 1.29829314685480e-01, 1.29592587577818e-01, 1.29272067516510e-01, 1.28850361702774e-01, 1.28309332663633e-01, 1.27630632672336e-01, 1.26796316114280e-01, 1.25789510410300e-01, 1.24595119893399e-01, 1.23200531391817e-01, 1.21596285877871e-01, 1.19776678085955e-01, 1.17740246088545e-01, 1.15490115873641e-01, 1.13034172146200e-01, 1.10385035700517e-01, 1.07559839239002e-01, 1.04579806557936e-01, 1.01469653437448e-01, 9.82568410664442e-02, 9.49707231336276e-02, 9.16416347339037e-02, 8.82999742361079e-02, 8.49753279414358e-02, 8.16956819496113e-02, 7.84867568298893e-02, 7.53714895253097e-02, 7.23696746799902e-02, 6.94977655687268e-02, 6.67688241844218e-02, 6.41926016661785e-02, 6.17757246271111e-02, 5.95219601808452e-02, 5.74325323437677e-02, 5.55064645292163e-02, 5.37409264296116e-02, 5.21315680499157e-02, 5.06728284127433e-02, 4.93582110232893e-02, 4.81805222250671e-02, 4.71320719028512e-02, 4.62048385275580e-02, 4.53906023087995e-02, 4.46810512998468e-02, 4.40678657966060e-02, 4.35427863921802e-02, 4.30976707035171e-02, 4.27245431613699e-02, 4.24156414310488e-02, 4.21634620728213e-02, 4.19608070198763e-02, 4.18008314058937e-02, 4.16770922752221e-02, 4.15835968138936e-02, 4.15148480126248e-02, 4.14658851580469e-02, 4.14323162941628e-02, 4.14103398113497e-02, 4.13967526126129e-02, 4.13889428403564e-02, 4.13848658756706e-02, 4.13830031773443e-02, 4.13823044311670e-02, 4.13821143567930e-02, 4.13820863022778e-02, 4.13820840749395e-02, 4.13820432566325e-02, 4.13817667075023e-02, 4.13807500969057e-02, 4.13780402876540e-02, 4.13721104508511e-02, 4.13607560752212e-02, 4.13410168881398e-02, 4.13091296676100e-02, 4.12605168307891e-02, 4.11898154200117e-02, 4.10909505254864e-02, 4.09572561264753e-02, 4.07816446593046e-02, 4.05568242425005e-02, 4.02755594000155e-02, 3.99309674397870e-02, 3.95168386215801e-02, 3.90279642818570e-02, 3.84604536910729e-02, 3.78120181771351e-02, 3.70822005133107e-02, 3.62725291679169e-02, 3.53865809432818e-02, 3.44299416665024e-02, 3.34100624560278e-02, 3.23360178576166e-02, 3.12181807681498e-02, 3.00678364206135e-02, 2.88967627757870e-02, 2.77168067632516e-02, 2.65394846882714e-02, 2.53756310336826e-02, 2.42351135506169e-02, 2.31266249634388e-02, 2.20575539294288e-02, 2.10339310860093e-02, 2.00604407961862e-02, 1.91404859086219e-02, 1.82762914671457e-02, 1.74690335522740e-02, 1.67189808805404e-02, 1.60256389477240e-02, 1.53878889473589e-02, 1.48041160847754e-02, 1.42723240096672e-02, 1.37902338018847e-02, 1.33553672445430e-02, 1.29651150500791e-02, 1.26167913315811e-02, 1.23076759967091e-02, 1.20350469338177e-02, 1.17962039094976e-02, 1.15884860397415e-02, 1.14092845704334e-02, 1.12560525295180e-02, 1.11263126079137e-02, 1.10176643940605e-02, 1.09277918327138e-02, 1.08544715033645e-02, 1.07955820298429e-02, 1.07491146460603e-02, 1.07131846712178e-02, 1.06860434028420e-02, 1.06660897383364e-02, 1.06518806944251e-02, 1.06421399267589e-02, 1.06357633619609e-02, 1.06318211440168e-02, 1.06295552566422e-02, 1.06283724032181e-02, 1.06278319806279e-02, 1.06276292547420e-02, 1.06275741089235e-02, 1.06275659695143e-02, 1.06275653173924e-02, 1.06275532919824e-02, 1.06274718183101e-02, 1.06271723169016e-02, 1.06263739901777e-02, 1.06246270442512e-02, 1.06212820951868e-02, 1.06154672696164e-02, 1.06060745176535e-02, 1.05917566453257e-02, 1.05709365094100e-02, 1.05418296466695e-02, 1.05024812794170e-02, 1.04508181027825e-02, 1.03847144848714e-02, 1.03020717012328e-02, 1.02009076252898e-02, 1.00794530006283e-02, 9.93624917832392e-03, 9.77024119356429e-03, 9.58085948172993e-03, 9.36808356685503e-03, 9.13248180960125e-03, 8.87522279172970e-03, 8.59805603340646e-03, 8.30326227067246e-03, 7.99357615029805e-03, 7.67208658261609e-03, 7.34212180957172e-03, 7.00712726197139e-03, 6.67054440158665e-03, 6.33569803014211e-03, 6.00569817817131e-03, 5.68336092423749e-03, 5.37115058060067e-03, 5.07114382286034e-03, 4.78501467652848e-03, 4.51403793148586e-03, 4.25910765190451e-03, 4.02076706279786e-03, 3.79924618759730e-03, 3.59450401899153e-03, 3.40627249261161e-03, 3.23409994949486e-03, 3.07739214082403e-03, 2.93544926590528e-03, 2.80749810775660e-03, 2.69271897538204e-03, 2.59026768367221e-03, 2.49929305257979e-03, 2.41895043571105e-03, 2.34841174712130e-03, 2.28687243526338e-03, 2.23355589491803e-03, 2.18771590429867e-03, 2.14863773321383e-03, 2.11563850679397e-03, 2.08806727313515e-03, 2.06530509066738e-03, 2.04676535334158e-03, 2.03189448964736e-03, 2.02017308808422e-03, 2.01111742464377e-03, 2.00428129275565e-03, 1.99925797249989e-03, 1.99568212620704e-03, 1.99323137688623e-03, 1.99162732084801e-03, 1.99063574335412e-03, 1.99006584654171e-03, 1.98976835598330e-03, 1.98963244079664e-03, 1.98958145605843e-03, 1.98956758713109e-03, 1.98956554012852e-03, 1.98956537457207e-03, 1.98956230252634e-03, 1.98954148899588e-03, 1.98946497778252e-03, 1.98926103824704e-03, 1.98881477876967e-03, 1.98796035780902e-03, 1.98647519860629e-03, 1.98407662743705e-03, 1.98042136413790e-03, 1.97510828772351e-03, 1.96768485999144e-03, 1.95765750260080e-03, 1.94450606763041e-03, 1.92770231430472e-03, 1.90673199902847e-03, 1.88111982022576e-03, 1.85045606096423e-03, 1.81442338424631e-03, 1.77282191802419e-03, 1.72559057851629e-03, 1.67282259083379e-03, 1.61477343584693e-03, 1.55186001393304e-03, 1.48465060371450e-03, 1.41384601828065e-03, 1.34025303943019e-03, 1.26475173046003e-03, 1.18825865375527e-03, 1.11168821240426e-03, 1.03591444297990e-03, 9.61736445262318e-04, 8.89851884409154e-04, 8.20841734579858e-04, 7.55165314274109e-04, 6.93162220471777e-04, 6.35057470930178e-04, 5.80966665353344e-04, 5.30923025422680e-04, 4.84989059643226e-04, 4.42789294737622e-04, 4.03954536224660e-04, 3.68343262465487e-04, 3.35655401457302e-04, 3.05445056739018e-04, 2.77902579767897e-04, 2.52526100581463e-04, 2.28916221132636e-04, 2.06861828502269e-04, 1.86208655776454e-04, 1.66855517917952e-04, 1.48736220907608e-04, 1.31914360991914e-04, 1.16422019132756e-04, 1.01955226938344e-04, 8.85393581029348e-05, 7.61894963261511e-05, 6.49119427151845e-05, 5.46993693717468e-05, 4.55301365605010e-05, 3.73731400316046e-05, 3.01936399011993e-05, 2.39533086130450e-05, 1.86079743116352e-05, 1.41062128591329e-05, 1.03890297538101e-05, 7.39033646368625e-06, 5.03795101280957e-06, 3.25491856756344e-06, 1.96101632354839e-06, 1.07436885343759e-06, 5.13155242742869e-07, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14],
[ 7.70000000000000e+01, 7.69999999136345e+01, 7.69999972430436e+01, 7.69999791493660e+01, 7.69999126346331e+01, 7.69997353272399e+01, 7.69993472764034e+01, 7.69986040438747e+01, 7.69973113411391e+01, 7.69952214123306e+01, 7.69920313089011e+01, 7.69873831440067e+01, 7.69808663552820e+01, 7.69720219470704e+01, 7.69603486301336e+01, 7.69453107308432e+01, 7.69263477045064e+01, 7.69028850591631e+01, 7.68743464757375e+01, 7.68401668949720e+01, 7.67998063269480e+01, 7.67527641204007e+01, 7.66985934022468e+01, 7.66369153604504e+01, 7.65674329964828e+01, 7.64899439221591e+01, 7.64043517285634e+01, 7.63106754240242e+01, 7.62090564364612e+01, 7.60997627137174e+01, 7.59831895398565e+01, 7.58598568151262e+01, 7.57304027140465e+01, 7.55955738247505e+01, 7.54562120637459e+01, 7.53132388330798e+01, 7.51676370234164e+01, 7.50204315542433e+01, 7.48726691760171e+01, 7.47253982407305e+01, 7.45796490855647e+01, 7.44364155814820e+01, 7.42966382888138e+01, 7.41611895482497e+01, 7.40308607286480e+01, 7.39063517598010e+01, 7.37882630020773e+01, 7.36770894460261e+01, 7.35732171915930e+01, 7.34769221253286e+01, 7.33883706912005e+01, 7.33076226329236e+01, 7.32346355702923e+01, 7.31692712568686e+01, 7.31113033503750e+01, 7.30604265098642e+01, 7.30162666153765e+01, 7.29783918870047e+01, 7.29463246620755e+01, 7.29195535727221e+01, 7.28975458528253e+01, 7.28797594944931e+01, 7.28656549712477e+01, 7.28547062490382e+01, 7.28464108179846e+01, 7.28402984979452e+01, 7.28359387996973e+01, 7.28329466604008e+01, 7.28309864162349e+01, 7.28297739253741e+01, 7.28290768090942e+01, 7.28287128357787e+01, 7.28285465297585e+01, 7.28284841420454e+01, 7.28284671710197e+01, 7.28284646661299e+01, 7.28284643375412e+01, 7.28284566818183e+01, 7.28284048131534e+01, 7.28282141382073e+01, 7.28277058669107e+01, 7.28265935179867e+01, 7.28244631998534e+01, 7.28207584433888e+01, 7.28147702624160e+01, 7.28056330017015e+01, 7.27923263995231e+01, 7.27736841404960e+01, 7.27484090022433e+01, 7.27150945051920e+01, 7.26722527581146e+01, 7.26183479548801e+01, 7.25518347247217e+01, 7.24712002770345e+01, 7.23750090244412e+01, 7.22619481321630e+01, 7.21308722513317e+01, 7.19808455783344e+01, 7.18111793740327e+01, 7.16214632048273e+01, 7.14115884481169e+01, 7.11817630301128e+01, 7.09325168951881e+01, 7.06646982728313e+01, 7.03794613213907e+01, 7.00782461012832e+01, 6.97627520101224e+01, 6.94349057994345e+01, 6.90968251509754e+01, 6.87507786302106e+01, 6.83991427724866e+01, 6.80443571671516e+01, 6.76888786739682e+01, 6.73351362276189e+01, 6.69854878968141e+01, 6.66421818213433e+01, 6.63073223068002e+01, 6.59828417933645e+01, 6.56704788019176e+01, 6.53717614728683e+01, 6.50879960524691e+01, 6.48202596491758e+01, 6.45693967048744e+01, 6.43360188053561e+01, 6.41205076140202e+01, 6.39230208132677e+01, 6.37435009742522e+01, 6.35816872619833e+01, 6.34371298392501e+01, 6.33092067763657e+01, 6.31971432151098e+01, 6.31000324795410e+01, 6.30168587750048e+01, 6.29465210695451e+01, 6.28878577088675e+01, 6.28396712776903e+01, 6.28007531884802e+01, 6.27699074559186e+01, 6.27459731051823e+01, 6.27278446675534e+01, 6.27144902407993e+01, 6.27049666360938e+01, 6.26984311986078e+01, 6.26941499744299e+01, 6.26915019997860e+01, 6.26899796057856e+01, 6.26891847581664e+01, 6.26888215809856e+01, 6.26886853399202e+01, 6.26886482789747e+01, 6.26886428088541e+01, 6.26886423259466e+01, 6.26886328636113e+01, 6.26885687548837e+01, 6.26883330842131e+01, 6.26877048711462e+01, 6.26863300341676e+01, 6.26836970260367e+01, 6.26791181062454e+01, 6.26717170950921e+01, 6.26604243135449e+01, 6.26439792510273e+01, 6.26209413157509e+01, 6.25897088053194e+01, 6.25485459863186e+01, 6.24956178905521e+01, 6.24290321269718e+01, 6.23468866823614e+01, 6.22473223570619e+01, 6.21285781770320e+01, 6.19890478670476e+01, 6.18273352894205e+01, 6.16423066721064e+01, 6.14331374847715e+01, 6.11993519743679e+01, 6.09408536329442e+01, 6.06579452195129e+01, 6.03513373705618e+01, 6.00221452891735e+01, 5.96718734869775e+01, 5.93023890580134e+01, 5.89158844781773e+01, 5.85148314247886e+01, 5.81019275563657e+01, 5.76800385273935e+01, 5.72521376816086e+01, 5.68212458345371e+01, 5.63903733210865e+01, 5.59624660856971e+01, 5.55403571002361e+01, 5.51267238896440e+01, 5.47240524991184e+01, 5.43346078956447e+01, 5.39604105754113e+01, 5.36032190337575e+01, 5.32645177155993e+01, 5.29455100676220e+01, 5.26471163310499e+01, 5.23699757288740e+01, 5.21144527084842e+01, 5.18806469015332e+01, 5.16684064620780e+01, 5.14773444451537e+01, 5.13068578915962e+01, 5.11561492890371e+01, 5.10242500796996e+01, 5.09100458789647e+01, 5.08123030518092e+01, 5.07296962662308e+01, 5.06608366049707e+01, 5.06042997725669e+01, 5.05586538889837e+01, 5.05224863197823e+01, 5.04944289624308e+01, 5.04731813950172e+01, 5.04575313025496e+01, 5.04463716309194e+01, 5.04387139812793e+01, 5.04336978476344e+01, 5.04305954152280e+01, 5.04288117720772e+01, 5.04278805343519e+01, 5.04274550406599e+01, 5.04272954227227e+01, 5.04272520027204e+01, 5.04272455940171e+01, 5.04272450799599e+01, 5.04272355929955e+01, 5.04271713174092e+01, 5.04269350334450e+01, 5.04263051863073e+01, 5.04249267769855e+01, 5.04222869418830e+01, 5.04176961953219e+01, 5.04102761910084e+01, 5.03989547217008e+01, 5.03824685175726e+01, 5.03593742187577e+01, 5.03280676817764e+01, 5.02868115296609e+01, 5.02337705711080e+01, 5.01670543986370e+01, 5.00847661387881e+01, 4.99850559842047e+01, 4.98661778092454e+01, 4.97265468834017e+01, 4.95647964783808e+01, 4.93798310425032e+01, 4.91708736128730e+01, 4.89375052667752e+01, 4.86796946841011e+01, 4.83978162965312e+01, 4.80926560200861e+01, 4.77654041784605e+01, 4.74176358877645e+01, 4.70512798401029e+01, 4.66685770341013e+01, 4.62720314890610e+01, 4.58643552853912e+01, 4.54484103601381e+01, 4.50271493556034e+01, 4.46035575195419e+01, 4.41805972682737e+01, 4.37611566353482e+01, 4.33480024997213e+01, 4.29437392393608e+01, 4.25507732721616e+01, 4.21712837910667e+01, 4.18071998432269e+01, 4.14601837307641e+01, 4.11316205303098e+01, 4.08226133600055e+01, 4.05339838883507e+01, 4.02662774948725e+01, 4.00197724629400e+01, 3.97944926043979e+01, 3.95902227708598e+01, 3.94065267810102e+01, 3.92427673710835e+01, 3.90981278437004e+01, 3.89716351394630e+01, 3.88621840814166e+01, 3.87685625437483e+01, 3.86894772750333e+01, 3.86235800673496e+01, 3.85694939116267e+01, 3.85258387236470e+01, 3.84912561716110e+01, 3.84644330925119e+01, 3.84441229575893e+01, 3.84291648426259e+01, 3.84184993809898e+01, 3.84111812282997e+01, 3.84063876472723e+01, 3.84034229272126e+01, 3.84017184800219e+01, 3.84008285969882e+01, 3.84004220002549e+01, 3.84002694714876e+01, 3.84002279799368e+01, 3.84002218558698e+01, 3.84002213835228e+01, 3.84002129016872e+01, 3.84001554360032e+01, 3.83999441860659e+01, 3.83993810712612e+01, 3.83981487081224e+01, 3.83957885846607e+01, 3.83916843082152e+01, 3.83850507011287e+01, 3.83749293996922e+01, 3.83601914729411e+01, 3.83395474143449e+01, 3.83115646663239e+01, 3.82746926110869e+01, 3.82272947004096e+01, 3.81676871046620e+01, 3.80941829459187e+01, 3.80051408555795e+01, 3.78990162836703e+01, 3.77744137098768e+01, 3.76301376933271e+01, 3.74652405773579e+01, 3.72790646619498e+01, 3.70712767884192e+01, 3.68418935563090e+01, 3.65912958062386e+01, 3.63202315350109e+01, 3.60298070257092e+01, 3.57214666275097e+01, 3.53969622494624e+01, 3.50583141781282e+01, 3.47077652345368e+01, 3.43477305101331e+01, 3.39807449466327e+01, 3.36094108623299e+01, 3.32363472170656e+01, 3.28641420102687e+01, 3.24953087895862e+01, 3.21322478724687e+01, 3.17772125899956e+01, 3.14322806643102e+01, 3.10993307156979e+01, 3.07800238333271e+01, 3.04757901017609e+01, 3.01878199281639e+01, 2.99170599520159e+01, 2.96642132443053e+01, 2.94297434304828e+01, 2.92138823166568e+01, 2.90166405725960e+01, 2.88378210307994e+01, 2.86770341928680e+01, 2.85337155820125e+01, 2.84071446313830e+01, 2.82964648410385e+01, 2.82007049640267e+01, 2.81188009903606e+01, 2.80496186863829e+01, 2.79919764187451e+01, 2.79446679517281e+01, 2.79064848599779e+01, 2.78762381526649e+01, 2.78527786664707e+01, 2.78350157600406e+01, 2.78219338370621e+01, 2.78126062429002e+01, 2.78062061228831e+01, 2.78020138989200e+01, 2.77994211130364e+01, 2.77979304975005e+01, 2.77971522557290e+01, 2.77967966692370e+01, 2.77966632762413e+01, 2.77966269900908e+01, 2.77966216343303e+01, 2.77966212288082e+01, 2.77966140450259e+01, 2.77965653738425e+01, 2.77963864534094e+01, 2.77959095171151e+01, 2.77948657531398e+01, 2.77928668181129e+01, 2.77893906466170e+01, 2.77837722113313e+01, 2.77751997875706e+01, 2.77627171581016e+01, 2.77452320555814e+01, 2.77215309767804e+01, 2.76903003118905e+01, 2.76501535128849e+01, 2.75996637796869e+01, 2.75374014787770e+01, 2.74619752376777e+01, 2.73720753970949e+01, 2.72665182709188e+01, 2.71442894856736e+01, 2.70045845683543e+01, 2.68468449453959e+01, 2.66707876209998e+01, 2.64764270275524e+01, 2.62640878815392e+01, 2.60344083201791e+01, 2.57883331084539e+01, 2.55270972513349e+01, 2.52522008688926e+01, 2.49653766353896e+01, 2.46685513966258e+01, 2.43638037315617e+01, 2.40533192135453e+01, 2.37393449847180e+01, 2.34241450389561e+01, 2.31099573734766e+01, 2.27989539614458e+01, 2.24932043346738e+01, 2.21946434367142e+01, 2.19050442864078e+01, 2.16259958537992e+01, 2.13588863802908e+01, 2.11048921760529e+01, 2.08649717178442e+01, 2.06398646745316e+01, 2.04300953296378e+01, 2.02359797667696e+01, 2.00576361410642e+01, 1.98949973741244e+01, 1.97478256699477e+01, 1.96157283394921e+01, 1.94981745251066e+01, 1.93945125179423e+01, 1.93039874495897e+01, 1.92257592053508e+01, 1.91589204464666e+01, 1.91025146415944e+01, 1.90555539962020e+01, 1.90170371370377e+01, 1.89859663638587e+01, 1.89613642295957e+01, 1.89422891608713e+01, 1.89278497907401e+01, 1.89172176511892e+01, 1.89096378693125e+01, 1.89044375311989e+01, 1.89010314222405e+01, 1.88989249202973e+01, 1.88977139053500e+01, 1.88970816506167e+01, 1.88967927691680e+01, 1.88966843999188e+01, 1.88966549208701e+01, 1.88966505698248e+01, 1.88966502431905e+01, 1.88966444940561e+01, 1.88966055428367e+01, 1.88964623542163e+01, 1.88960806673286e+01, 1.88952453620275e+01, 1.88936456797260e+01, 1.88908639041847e+01, 1.88863680448401e+01, 1.88795089917689e+01, 1.88695225237968e+01, 1.88555364450606e+01, 1.88365829958545e+01, 1.88116165271167e+01, 1.87795362428473e+01, 1.87392136027622e+01, 1.86895237443422e+01, 1.86293800394253e+01, 1.85577706603381e+01, 1.84737958126720e+01, 1.83767041168992e+01, 1.82659265102171e+01, 1.81411060126186e+01, 1.80021217722834e+01, 1.78491059836547e+01, 1.76824525573855e+01, 1.75028168055048e+01, 1.73111058682367e+01, 1.71084601215317e+01, 1.68962263285035e+01, 1.66759237896837e+01, 1.64492051606534e+01, 1.62178138993992e+01, 1.59835404480267e+01, 1.57481792284661e+01, 1.55134883434663e+01, 1.52811535466888e+01, 1.50527576203095e+01, 1.48297558268969e+01, 1.46134576380294e+01, 1.44050145321521e+01, 1.42054133324126e+01, 1.40154743390356e+01, 1.38358534012236e+01, 1.36670470583542e+01, 1.35093999385455e+01, 1.33631137101819e+01, 1.32282570152904e+01, 1.31047759530531e+01, 1.29925048127588e+01, 1.28911768689000e+01, 1.28004351421740e+01, 1.27198430974330e+01, 1.26488952938658e+01, 1.25870280257945e+01, 1.25336299968883e+01, 1.24880530590191e+01, 1.24496230222183e+01, 1.24176505072312e+01, 1.23914417702567e+01, 1.23703093841821e+01, 1.23535826158463e+01, 1.23406172986399e+01, 1.23308049680628e+01, 1.23235810083305e+01, 1.23184315536349e+01, 1.23148989000617e+01, 1.23125852139409e+01, 1.23111543685934e+01, 1.23103318016132e+01, 1.23099023553537e+01, 1.23097061395996e+01, 1.23096325325855e+01, 1.23096125097202e+01, 1.23096095543882e+01, 1.23096093332052e+01, 1.23096054491202e+01, 1.23095791338879e+01, 1.23094823965406e+01, 1.23092245320561e+01, 1.23086602109030e+01, 1.23075795027308e+01, 1.23057002496188e+01, 1.23026631654550e+01, 1.22980299920197e+01, 1.22912849859302e+01, 1.22818399394960e+01, 1.22690428481999e+01, 1.22521902241699e+01, 1.22305429166943e+01, 1.22033451386647e+01, 1.21698462169032e+01, 1.21293243943523e+01, 1.20811118274060e+01, 1.20246197603399e+01, 1.19593627410668e+01, 1.18849806880501e+01, 1.18012576433342e+01, 1.17081361606856e+01, 1.16057264806136e+01, 1.14943099242847e+01, 1.13743362738587e+01, 1.12464152666218e+01, 1.11113026789334e+01, 1.09698817786528e+01, 1.08231411529403e+01, 1.06721500545616e+01, 1.05180324496931e+01, 1.03619409021428e+01, 1.02050313109335e+01, 1.00484393533646e+01, 9.89325929683388e+00, 9.74052564877077e+00, 9.59119792811117e+00, 9.44614867161245e+00, 9.30615463822360e+00, 9.17189104758560e+00, 9.04392858740873e+00, 8.92273285190088e+00, 8.80866583169420e+00, 8.70198906484787e+00, 8.60286807563863e+00, 8.51137776735486e+00, 8.42750848967783e+00, 8.35117256241372e+00, 8.28221109799460e+00, 8.22040101954481e+00, 8.16546221556752e+00, 8.11706480442513e+00, 8.07483650120788e+00, 8.03837008680927e+00, 8.00723097525371e+00, 7.98096486222362e+00, 7.95910541731297e+00, 7.94118195708369e+00, 7.92672700804374e+00, 7.91528364091143e+00, 7.90641243275856e+00, 7.89969789453383e+00, 7.89475419046083e+00, 7.89122997477354e+00, 7.88881218136051e+00, 7.88722862341173e+00, 7.88624929240849e+00, 7.88568628713132e+00, 7.88539235135120e+00, 7.88525805051333e+00, 7.88520766976963e+00, 7.88519396499871e+00, 7.88519194220372e+00, 7.88519179075696e+00, 7.88518913051901e+00, 7.88517110702362e+00, 7.88510485086927e+00, 7.88492823721943e+00, 7.88454172729432e+00, 7.88380153266417e+00, 7.88251438467319e+00, 7.88043416347892e+00, 7.87726060837376e+00, 7.87264029130708e+00, 7.86616998625587e+00, 7.85740250556451e+00, 7.84585499831569e+00, 7.83101961411732e+00, 7.81237632918423e+00, 7.78940761354528e+00, 7.76161449475173e+00, 7.72853345357687e+00, 7.68975348219891e+00, 7.64493255788187e+00, 7.59381274763031e+00, 7.53623317202265e+00, 7.47214012571490e+00, 7.40159377849803e+00, 7.32477105804405e+00, 7.24196453044557e+00, 7.15357732839976e+00, 7.06011440692865e+00, 6.96217061002899e+00, 6.86041618902345e+00, 6.75558051152777e+00, 6.64843473430987e+00, 6.53977418823189e+00, 6.43040115055621e+00, 6.32110857498999e+00, 6.21266522932285e+00, 6.10580256819979e+00, 6.00120355393167e+00, 5.89949353620698e+00, 5.80123321358658e+00, 5.70691362535302e+00, 5.61695306086838e+00, 5.53169572476570e+00, 5.45141196034739e+00, 5.37629981103321e+00, 5.30648769081751e+00, 5.24203793891263e+00, 5.18295104953884e+00, 5.12917039270613e+00, 5.08058727269309e+00, 5.03704620435541e+00, 4.99835032006850e+00, 4.96426684911967e+00, 4.93453263443028e+00, 4.90885966709402e+00, 4.88694062664432e+00, 4.86845441429186e+00, 4.85307165840258e+00, 4.84046015765367e+00, 4.83029020954750e+00, 4.82223975256218e+00, 4.81599923162187e+00, 4.81127608117961e+00, 4.80779871017687e+00, 4.80531987016856e+00, 4.80361929305906e+00, 4.80250549851552e+00, 4.80181669276811e+00, 4.80142070899281e+00, 4.80121397300508e+00, 4.80111951436221e+00, 4.80108407978349e+00, 4.80107444073020e+00, 4.80107301802653e+00, 4.80107291122609e+00, 4.80107103144856e+00, 4.80105829568600e+00, 4.80101147777678e+00, 4.80088667919702e+00, 4.80061356450059e+00, 4.80009053229418e+00, 4.79918102219003e+00, 4.79771113906526e+00, 4.79546875340924e+00, 4.79220420912230e+00, 4.78763273595902e+00, 4.78143862091299e+00, 4.77328113960034e+00, 4.76280218432531e+00, 4.74963545052782e+00, 4.73341695993044e+00, 4.71379661103434e+00, 4.69045036164590e+00, 4.66309257135850e+00, 4.63148797274657e+00, 4.59546270675435e+00, 4.55491385747609e+00, 4.50981695898427e+00, 4.46023102350634e+00, 4.40630075361241e+00, 4.34825574466051e+00, 4.28640664750340e+00, 4.22113843271094e+00, 4.15290106234062e+00, 4.08219801984521e+00, 4.00957326101071e+00, 3.93559721987889e+00, 3.86085252844249e+00, 3.78592008705721e+00, 3.71136605805333e+00, 3.63773025597192e+00, 3.56551628510847e+00, 3.49518364103476e+00, 3.42714185984717e+00, 3.36174667795515e+00, 3.29929806460032e+00, 3.24003991413950e+00, 3.18416113731327e+00, 3.13179786922615e+00, 3.08303651331299e+00, 3.03791736046774e+00, 2.99643855545065e+00, 2.95856022340735e+00, 2.92420861311059e+00, 2.89328015647031e+00, 2.86564538298208e+00, 2.84115266103490e+00, 2.81963176412123e+00, 2.80089727840252e+00, 2.78475187874722e+00, 2.77098950366533e+00, 2.75939845626299e+00, 2.74976444946913e+00, 2.74187360063048e+00, 2.73551536462452e+00, 2.73048537754401e+00, 2.72658816647704e+00, 2.72363966661713e+00, 2.72146947642031e+00, 2.71992277600793e+00, 2.71886183431874e+00, 2.71816703695364e+00, 2.71773737899201e+00, 2.71749038450588e+00, 2.71736143581757e+00, 2.71730251916837e+00, 2.71728041767260e+00, 2.71727440554622e+00, 2.71727351816958e+00, 2.71727345127169e+00, 2.71727227003387e+00, 2.71726426698226e+00, 2.71723484704404e+00, 2.71715642525875e+00, 2.71698480601435e+00, 2.71665615346269e+00, 2.71608468285081e+00, 2.71516119264533e+00, 2.71375254543581e+00, 2.71170219031576e+00, 2.70883179900142e+00, 2.70494406050408e+00, 2.69982664313630e+00, 2.69325728711061e+00, 2.68500993619342e+00, 2.67486175448474e+00, 2.66260080775369e+00, 2.64803412290250e+00, 2.63099578047907e+00, 2.61135465095596e+00, 2.58902136294596e+00, 2.56395409673797e+00, 2.53616283342676e+00, 2.50571175923038e+00, 2.47271962332910e+00, 2.43735796884303e+00, 2.39984729018512e+00, 2.36045130355254e+00, 2.31946963770284e+00, 2.27722934750059e+00, 2.23407571395114e+00, 2.19036281657947e+00, 2.14644434676413e+00, 2.10266507821653e+00, 2.05935333108687e+00, 2.01681466936213e+00, 1.97532696821193e+00, 1.93513688888491e+00, 1.89645771188641e+00, 1.85946841007687e+00, 1.82431379480525e+00, 1.79110554042980e+00, 1.75992388372457e+00, 1.73081980148949e+00, 1.70381748824305e+00, 1.67891698213833e+00, 1.65609681749383e+00, 1.63531661345300e+00, 1.61651953788754e+00, 1.59963461205541e+00, 1.58457884367870e+00, 1.57125919349778e+00, 1.55957439285923e+00, 1.54941663765277e+00, 1.54067318726280e+00, 1.53322789659737e+00, 1.52696270525428e+00, 1.52175910110519e+00, 1.51749956673117e+00, 1.51406900700964e+00, 1.51135614557863e+00, 1.50925486778701e+00, 1.50766547895803e+00, 1.50649584017692e+00, 1.50566234003790e+00, 1.50509066032424e+00, 1.50471629664186e+00, 1.50448480146686e+00, 1.50435172648072e+00, 1.50428225275407e+00, 1.50425051041292e+00, 1.50423860289082e+00, 1.50423536376798e+00, 1.50423488568065e+00, 1.50423484944231e+00, 1.50423420696802e+00, 1.50422985411625e+00, 1.50421385266323e+00, 1.50417119935912e+00, 1.50407785717168e+00, 1.50389910922441e+00, 1.50358830731233e+00, 1.50308608226257e+00, 1.50232007771645e+00, 1.50120526123048e+00, 1.49964485574927e+00, 1.49753191924533e+00, 1.49475157956566e+00, 1.49118390463607e+00, 1.48670735524986e+00, 1.48120272978443e+00, 1.47455746956953e+00, 1.46667015362559e+00, 1.45745497640004e+00, 1.44684597676026e+00, 1.43480077554695e+00, 1.42130358625819e+00, 1.40636729106336e+00, 1.39003442217621e+00, 1.37237695383988e+00, 1.35349488743410e+00, 1.33351369412581e+00, 1.31258075762838e+00, 1.29086102574090e+00, 1.26853212654378e+00, 1.24577922898146e+00, 1.22278992665238e+00, 1.19974939963025e+00, 1.17683606640711e+00, 1.15421788273906e+00, 1.13204938322549e+00, 1.11046950152408e+00, 1.08960015173466e+00, 1.06954551055840e+00, 1.05039190937515e+00, 1.03220822767375e+00, 1.01504667320980e+00, 9.98943837809634e-01, 9.83921928377001e-01, 9.69990087824229e-01, 9.57145738016948e-01, 9.45375894493807e-01, 9.34658419299524e-01, 9.24963192829352e-01, 9.16253197626219e-01, 9.08485516415496e-01, 9.01612253346769e-01, 8.95581391622073e-01, 8.90337602675245e-01, 8.85823022112703e-01, 8.81978006020390e-01, 8.78741878278120e-01, 8.76053675496578e-01, 8.73852891421393e-01, 8.72080217471108e-01, 8.70678270862072e-01, 8.69592296913723e-01, 8.68770828017105e-01, 8.68166278758196e-01, 8.67735455128042e-01, 8.67439955836645e-01, 8.67246445566877e-01, 8.67126783494243e-01, 8.67057995351914e-01, 8.67022083386104e-01, 8.67005675287615e-01, 8.66999520104102e-01, 8.66997845750689e-01, 8.66997598619847e-01, 8.66997579768793e-01, 8.66997243988294e-01, 8.66994969028909e-01, 8.66986606083060e-01, 8.66964313889383e-01, 8.66915529742016e-01, 8.66822109058091e-01, 8.66659670963017e-01, 8.66397184875510e-01, 8.65996830023349e-01, 8.65414155722974e-01, 8.64598564500128e-01, 8.63494132062002e-01, 8.62040767188276e-01, 8.60175700457759e-01, 8.57835273434603e-01, 8.54956980096304e-01, 8.51481691110067e-01, 8.47355970881779e-01, 8.42534379450640e-01, 8.36981638864557e-01, 8.30674539133739e-01, 8.23603464151233e-01, 8.15773434050698e-01, 8.07204586973796e-01, 7.97932058318807e-01, 7.88005256052145e-01, 7.77486572427634e-01, 7.66449610935410e-01, 7.54977038285531e-01, 7.43158191506097e-01, 7.31086578082611e-01, 7.18857402457289e-01, 7.06565236727130e-01, 6.94301929817911e-01, 6.82154821243913e-01, 6.70205296328959e-01, 6.58527692565059e-01, 6.47188543914211e-01, 6.36246132654293e-01, 6.25750307227335e-01, 6.15742519106817e-01, 6.06256031059530e-01, 5.97316252176100e-01, 5.88941160458432e-01, 5.81141780477154e-01, 5.73922690739209e-01, 5.67282542270718e-01, 5.61214576083913e-01, 5.55707132415421e-01, 5.50744148796596e-01, 5.46305647139931e-01, 5.42368212154472e-01, 5.38905464616207e-01, 5.35888533411310e-01, 5.33286529929299e-01, 5.31067027410771e-01, 5.29196546346867e-01, 5.27641045101987e-01, 5.26366412720449e-01, 5.25338958532814e-01, 5.24525890876248e-01, 5.23895775171468e-01, 5.23418959946483e-01, 5.23067958343546e-01, 5.22817772333568e-01, 5.22646147389580e-01, 5.22533746766953e-01, 5.22464236762190e-01, 5.22424277258281e-01, 5.22403415331581e-01, 5.22393883467231e-01, 5.22390307755629e-01, 5.22389335077250e-01, 5.22389191512005e-01, 5.22389180483628e-01, 5.22388983029700e-01, 5.22387645251413e-01, 5.22382727454172e-01, 5.22369618544715e-01, 5.22340930689097e-01, 5.22285992588564e-01, 5.22190462826842e-01, 5.22036082955907e-01, 5.21800587846485e-01, 5.21457788119577e-01, 5.20977835925947e-01, 5.20327680656280e-01, 5.19471715099101e-01, 5.18372604978751e-01, 5.16992285759272e-01, 5.15293100399390e-01, 5.13239040996282e-01, 5.10797046855288e-01, 5.07938302605325e-01, 5.04639473784763e-01, 5.00883815032537e-01, 4.96662088533867e-01, 4.91973238114227e-01, 4.86824777135695e-01, 4.81232865221116e-01, 4.75222068295125e-01, 4.68824816526620e-01, 4.62080593402092e-01, 4.55034904450343e-01, 4.47738084663825e-01, 4.40244008692105e-01, 4.32608767437031e-01, 4.24889369447328e-01, 4.17142516670680e-01, 4.09423493087786e-01, 4.01785192934099e-01, 3.94277303821298e-01, 3.86945649983474e-01, 3.79831692621977e-01, 3.72972178106750e-01, 3.66398920555739e-01, 3.60138702836167e-01, 3.54213279010423e-01, 3.48639461357763e-01, 3.43429276044589e-01, 3.38590173025224e-01, 3.34125277625091e-01, 3.30033673308350e-01, 3.26310707226369e-01, 3.22948312157877e-01, 3.19935340284019e-01, 3.17257905803018e-01, 3.14899734606172e-01, 3.12842520052939e-01, 3.11066284268037e-01, 3.09549744328820e-01, 3.08270682240860e-01, 3.07206316769111e-01, 3.06333674082722e-01, 3.05629952896723e-01, 3.05072878483642e-01, 3.04641038724839e-01, 3.04314194417969e-01, 3.04073555480585e-01, 3.03902014595380e-01, 3.03784330292544e-01, 3.03707252479455e-01, 3.03659584975705e-01, 3.03632181613239e-01, 3.03617874796899e-01, 3.03611337941959e-01, 3.03608885749696e-01, 3.03608218694691e-01, 3.03608120238778e-01, 3.03608112617894e-01, 3.03607975420718e-01, 3.03607045890331e-01, 3.03603628846673e-01, 3.03594520327692e-01, 3.03574586885460e-01, 3.03536413218808e-01, 3.03470032834426e-01, 3.03362755068846e-01, 3.03199100191051e-01, 3.02960852326194e-01, 3.02627237441641e-01, 3.02175230457109e-01, 3.01579991546703e-01, 3.00815426825523e-01, 2.99854862898115e-01, 2.98671818370806e-01, 2.97240848741211e-01, 2.95538434576855e-01, 2.93543877211476e-01, 2.91240162018411e-01, 2.88614747333611e-01, 2.85660237833483e-01, 2.82374904924623e-01, 2.78763023453124e-01, 2.74835003415731e-01, 2.70607306643214e-01, 2.66102150665787e-01, 2.61347014067454e-01, 2.56373968524514e-01, 2.51218871494714e-01, 2.45920459544209e-01, 2.40519385263980e-01, 2.35057240657383e-01, 2.29575607085565e-01, 2.24115166850950e-01, 2.18714904913071e-01, 2.13411421732999e-01, 2.08238370462055e-01, 2.03226024188851e-01, 1.98400972179769e-01, 1.93785938331899e-01, 1.89399710610360e-01, 1.85257167154575e-01, 1.81369382991548e-01, 1.77743800776454e-01, 1.74384449517817e-01, 1.71292196618310e-01, 1.68465020538505e-01, 1.65898293729967e-01, 1.63585067961321e-01, 1.61516356577025e-01, 1.59681410412574e-01, 1.58067985921861e-01, 1.56662605457981e-01, 1.55450810548899e-01, 1.54417409409533e-01, 1.53546719858966e-01, 1.52822808320522e-01, 1.52229724757348e-01, 1.51751732341509e-01, 1.51373529494080e-01, 1.51080460794773e-01, 1.50858712275067e-01, 1.50695485888420e-01, 1.50579147596829e-01, 1.50499343577906e-01, 1.50447079571762e-01, 1.50414759328342e-01, 1.50396179428625e-01, 1.50386479345599e-01, 1.50382047365838e-01, 1.50380384788264e-01, 1.50379932527898e-01, 1.50379865775264e-01, 1.50379860566431e-01, 1.50379766252129e-01, 1.50379127259399e-01, 1.50376778264629e-01, 1.50370516784577e-01, 1.50356814082116e-01, 1.50330573275539e-01, 1.50284945011776e-01, 1.50211210278059e-01, 1.50098738980034e-01, 1.49935031696359e-01, 1.49705850395293e-01, 1.49395441717232e-01, 1.48986853566981e-01, 1.48462342148449e-01, 1.47803862231910e-01, 1.46993628501718e-01, 1.46014730553333e-01, 1.44851778887741e-01, 1.43491554588621e-01, 1.41923631824810e-01, 1.40140940457101e-01, 1.38140236312438e-01, 1.35922449423873e-01, 1.33492885784449e-01, 1.30861265705136e-01, 1.28041591199931e-01, 1.25051845190853e-01, 1.21913535818479e-01, 1.18651108796385e-01, 1.15291258662495e-01, 1.11862175254956e-01, 1.08392764333004e-01, 1.04911880854431e-01, 1.01447610197894e-01, 9.80266270452574e-02, 9.46736543840956e-02, 9.14110369383536e-02, 8.82584350871412e-02, 8.52326377008126e-02, 8.23474858619022e-02, 7.96138944907341e-02, 7.70399556068071e-02, 7.46311053020321e-02, 7.23903363274587e-02, 7.03184392532563e-02, 6.84142571461308e-02, 6.66749412812635e-02, 6.50961982514247e-02, 6.36725216817098e-02, 6.23974044122662e-02, 6.12635293340418e-02, 6.02629389846498e-02, 5.93871854841126e-02, 5.86274634091808e-02, 5.79747287757044e-02, 5.74198074595387e-02, 5.69534961844482e-02, 5.65666586991966e-02, 5.62503190285497e-02, 5.59957527852220e-02, 5.57945765560675e-02, 5.56388344086542e-02, 5.55210796884751e-02, 5.54344495710950e-02, 5.53727293618351e-02, 5.53304033450530e-02, 5.53026890978856e-02, 5.52855525904050e-02, 5.52757020663108e-02, 5.52705595723069e-02, 5.52682100081512e-02, 5.52673286191094e-02, 5.52670888611999e-02, 5.52670534734716e-02, 5.52670506886273e-02, 5.52669999630950e-02, 5.52666562909820e-02, 5.52653929269979e-02, 5.52620253477869e-02, 5.52546559159737e-02, 5.52405442312391e-02, 5.52160090839067e-02, 5.51763676985502e-02, 5.51159176531826e-02, 5.50279663791287e-02, 5.49049123653447e-02, 5.47383809689421e-02, 5.45194159565341e-02, 5.42387254903076e-02, 5.38869782332085e-02, 5.34551416762099e-02, 5.29348509028296e-02, 5.23187921362762e-02, 5.16010819861208e-02, 5.07776208136477e-02, 4.98463975375785e-02, 4.88077239017883e-02, 4.76643789494773e-02, 4.64216491970006e-02, 4.50872565077702e-02, 4.36711734041637e-02, 4.21853337741709e-02, 4.06432547622444e-02, 3.90595922090284e-02, 3.74496565902120e-02, 3.58289185220286e-02, 3.42125323908959e-02, 3.26149037032906e-02, 3.10493207867211e-02, 2.95276651488621e-02, 2.80602078667016e-02, 2.66554925924680e-02, 2.53202998184321e-02, 2.40596824446300e-02, 2.28770596927887e-02, 2.17743549779357e-02, 2.07521632583081e-02, 1.98099343282081e-02, 1.89461602100636e-02, 1.81585569928444e-02, 1.74442339082646e-02, 1.67998448760898e-02, 1.62217199466710e-02, 1.57059758675757e-02, 1.52486063649206e-02, 1.48455537052286e-02, 1.44927637676773e-02, 1.41862272647018e-02, 1.39220099313917e-02, 1.36962744851843e-02, 1.35052969643965e-02, 1.33454797330046e-02, 1.32133630262388e-02, 1.31056364461131e-02, 1.30191513206483e-02, 1.29509343394147e-02, 1.28982023947515e-02, 1.28583781182243e-02, 1.28291052349425e-02, 1.28082625886918e-02, 1.27939755407738e-02, 1.27846234235420e-02, 1.27788418372887e-02, 1.27755188049008e-02, 1.27737841161633e-02, 1.27729915759147e-02, 1.27726942753412e-02, 1.27726134031718e-02, 1.27726014666507e-02, 1.27726005189999e-02, 1.27725831517722e-02, 1.27724654868189e-02, 1.27720329448948e-02, 1.27708799997350e-02, 1.27683570753732e-02, 1.27635264000503e-02, 1.27551290272407e-02, 1.27415652612228e-02, 1.27208907307761e-02, 1.26908302777622e-02, 1.26488115992345e-02, 1.25920201145817e-02, 1.25174757576758e-02, 1.24221312682096e-02, 1.23029900753124e-02, 1.21572400888255e-02, 1.19823977685343e-02, 1.17764549295600e-02, 1.15380191242167e-02, 1.12664374022273e-02, 1.09618930688526e-02, 1.06254659399241e-02, 1.02591486191272e-02, 9.86581441588896e-03, 9.44913640836376e-03, 9.01346139801262e-03, 8.56364655317066e-03, 8.10486986685532e-03, 7.64242774253887e-03, 7.18153384215725e-03, 6.72713274501291e-03, 6.28374006268473e-03, 5.85531763329072e-03, 5.44518862103871e-03, 5.05599330892892e-03, 4.68968277108672e-03, 4.34754502538226e-03, 4.03025690667391e-03, 3.73795450944669e-03, 3.47031516109553e-03, 3.22664421597096e-03, 3.00596059308589e-03, 2.80707615260667e-03, 2.62866570813003e-03, 2.46932629063679e-03, 2.32762565325118e-03, 2.20214067039137e-03, 2.09148642281320e-03, 1.99433678120897e-03, 1.90943747105621e-03, 1.83561288656620e-03, 1.77176811817503e-03, 1.71688762709721e-03, 1.67003177105585e-03, 1.63033209315699e-03, 1.59698604694825e-03, 1.56925167625843e-03, 1.54644266403636e-03, 1.52792406839808e-03, 1.51310895476797e-03, 1.50145601293161e-03, 1.49246813376180e-03, 1.48569181864389e-03, 1.48071721518761e-03, 1.47717851444895e-03, 1.47475441110601e-03, 1.47316831932187e-03, 1.47218805437888e-03, 1.47162473118101e-03, 1.47133069292238e-03, 1.47119635977333e-03, 1.47114596924780e-03, 1.47113226202578e-03, 1.47113023889744e-03, 1.47113007679422e-03, 1.47112708778252e-03, 1.47110683676033e-03, 1.47103239392266e-03, 1.47083397176573e-03, 1.47039980673469e-03, 1.46956862363368e-03, 1.46812411183816e-03, 1.46579187185125e-03, 1.46223932146938e-03, 1.45707905642212e-03, 1.44987612968350e-03, 1.44015961284083e-03, 1.42743861857753e-03, 1.41122267469029e-03, 1.39104595088213e-03, 1.36649436537084e-03, 1.33723408660558e-03, 1.30303946627919e-03, 1.26381808581755e-03, 1.21963046646556e-03, 1.17070215495119e-03, 1.11742637010082e-03, 1.06035613295319e-03, 1.00018569966664e-03, 9.37722074013436e-04, 8.73848346373052e-04, 8.09481574904680e-04, 7.45528816594287e-04, 6.82845482387506e-04, 6.22200061386025e-04, 5.64248135177081e-04, 5.09516583335312e-04, 4.58396796005900e-04, 4.11144873808666e-04, 3.67887611972151e-04, 3.28634251260951e-04, 2.93294234751329e-04, 2.61700232229600e-04, 2.33633026226698e-04, 2.08842344122218e-04, 1.87060428132213e-04, 1.68011033849780e-04, 1.51416653139336e-04, 1.37004383930162e-04, 1.24514229483452e-04, 1.13711375533398e-04, 1.04392644189488e-04, 9.63801500546424e-05, 8.95121803381471e-05, 8.36427780797067e-05, 7.86451380037079e-05, 7.44093382962774e-05, 7.08366482768952e-05, 6.78372694409290e-05, 6.53315493792661e-05, 6.32507967139873e-05, 6.15362483387116e-05, 6.01370134916124e-05, 5.90081134816955e-05, 5.81092889007095e-05, 5.74045273822321e-05, 5.68618242624552e-05, 5.64528503588235e-05, 5.61525517213113e-05, 5.59388261223148e-05, 5.57923399213678e-05, 5.56964510423086e-05, 5.56371689413027e-05, 5.56030945233488e-05, 5.55853065691281e-05, 5.55771795518467e-05, 5.55741308829738e-05, 5.55733015801252e-05, 5.55731791678227e-05, 5.55731692344352e-05, 5.55729840101549e-05, 5.55717291348401e-05, 5.55671164619298e-05, 5.55548230778413e-05, 5.55279312663970e-05, 5.54764756799386e-05, 5.53871347932500e-05, 5.52431078801874e-05, 5.50242189802930e-05, 5.47072794319092e-05, 5.42667152245713e-05, 5.36754282904894e-05, 5.29058223577294e-05, 5.19308847515276e-05, 5.07252328674213e-05, 4.92664429003753e-05, 4.75391816365800e-05, 4.55545405060320e-05, 4.34034088669932e-05, 4.11418623826876e-05, 3.86699413610359e-05, 3.59342194542478e-05, 3.29378577816897e-05, 2.98035408405894e-05, 2.69163619321944e-05, 2.40526904065064e-05, 2.10458554477100e-05, 1.79036633483065e-05, 1.55056349342431e-05, 1.32129934513614e-05, 1.07667722346605e-05, 8.37128240505081e-06, 6.84561727837263e-06, 5.10430804467887e-06, 3.57905406644932e-06, 3.53026943855639e-06, 4.23164421192723e-06, 4.71810308511067e-06, 4.48430816391010e-06, 3.62814836307734e-06, 2.45871196113967e-06, 1.34786002092824e-06, 6.42274216033911e-07, 4.21356617152924e-07, 5.81966389745503e-07, 7.79714682153099e-07, 8.94486120728288e-07, 9.35289065751022e-07, 8.63075172219396e-07, 6.53586582188836e-07, 1.42108547152020e-14, 4.30168185516777e-07, 5.96140651509813e-07, 6.49478010927851e-07, 6.22756132884206e-07, 5.55020861270350e-07, 4.76184538911184e-07, 4.04025655598161e-07, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14],
[ 7.80000000000000e+01, 7.79999999120077e+01, 7.79999971911105e+01, 7.79999787566008e+01, 7.79999109889233e+01, 7.79997303415678e+01, 7.79993349809397e+01, 7.79985777478837e+01, 7.79972606937286e+01, 7.79951313947852e+01, 7.79918811940058e+01, 7.79871454593338e+01, 7.79805058878837e+01, 7.79714948264668e+01, 7.79596015250534e+01, 7.79442801930634e+01, 7.79249596905636e+01, 7.79010546578869e+01, 7.78719778666056e+01, 7.78371535592067e+01, 7.77960315298871e+01, 7.77481016796276e+01, 7.76929087507513e+01, 7.76300669071846e+01, 7.75592737776358e+01, 7.74803235251570e+01, 7.73931184574744e+01, 7.72976786605097e+01, 7.71941491359490e+01, 7.70828039637125e+01, 7.69640470979827e+01, 7.68384095401081e+01, 7.67065428043643e+01, 7.65692087873457e+01, 7.64272663483062e+01, 7.62816550847226e+01, 7.61333769261652e+01, 7.59834762574948e+01, 7.58330193143604e+01, 7.56830735725516e+01, 7.55346877870174e+01, 7.53888732394188e+01, 7.52465866395074e+01, 7.51087150089109e+01, 7.49760627667905e+01, 7.48493411423241e+01, 7.47291599623208e+01, 7.46160218035602e+01, 7.45103184564697e+01, 7.44123296160494e+01, 7.43222236937036e+01, 7.42400606263049e+01, 7.41657965435553e+01, 7.40992901395376e+01, 7.40403105781378e+01, 7.39885467443961e+01, 7.39436176350505e+01, 7.39050836622810e+01, 7.38724586259724e+01, 7.38452220929555e+01, 7.38228319080135e+01, 7.38047365523584e+01, 7.37903870621205e+01, 7.37792482233046e+01, 7.37708087715809e+01, 7.37645903457281e+01, 7.37601549726825e+01, 7.37571108996172e+01, 7.37551166334254e+01, 7.37538830991151e+01, 7.37531738841901e+01, 7.37528035940735e+01, 7.37526344018050e+01, 7.37525709313524e+01, 7.37525536657945e+01, 7.37525511174323e+01, 7.37525507831410e+01, 7.37525429945531e+01, 7.37524902257082e+01, 7.37522962416076e+01, 7.37517791493390e+01, 7.37506474960278e+01, 7.37484802080460e+01, 7.37447111618352e+01, 7.37386190745642e+01, 7.37293232849794e+01, 7.37157858599373e+01, 7.36968203081303e+01, 7.36711070072210e+01, 7.36372152523896e+01, 7.35936316129831e+01, 7.35387940414294e+01, 7.34711309195328e+01, 7.33891039599450e+01, 7.32912536177395e+01, 7.31762454269647e+01, 7.30429154844571e+01, 7.28903131885128e+01, 7.27177393363087e+01, 7.25247778203090e+01, 7.23113194554219e+01, 7.20775769057253e+01, 7.18240902202227e+01, 7.15517230580547e+01, 7.12616501932701e+01, 7.09553372518507e+01, 7.06345137983701e+01, 7.03011408640175e+01, 6.99573738635717e+01, 6.96055217031814e+01, 6.92480028522791e+01, 6.88872993080917e+01, 6.85259096920315e+01, 6.81663030575187e+01, 6.78108751817104e+01, 6.74619090173447e+01, 6.71215405674059e+01, 6.67917308228532e+01, 6.64742437641914e+01, 6.61706299513410e+01, 6.58822150031823e+01, 6.56100922794903e+01, 6.53551192333118e+01, 6.51179170968340e+01, 6.48988737229010e+01, 6.46981494947130e+01, 6.45156862392771e+01, 6.43512190548052e+01, 6.42042909104709e+01, 6.40742698158360e+01, 6.39603682967832e+01, 6.38616648589937e+01, 6.37771270693864e+01, 6.37056358397270e+01, 6.36460104543179e+01, 6.35970338458264e+01, 6.35574775916806e+01, 6.35261260808342e+01, 6.35017992904454e+01, 6.34833736175441e+01, 6.34698002350009e+01, 6.34601204860564e+01, 6.34534778979834e+01, 6.34491264822870e+01, 6.34464350937824e+01, 6.34448877400229e+01, 6.34440798608409e+01, 6.34437107293586e+01, 6.34435722546174e+01, 6.34435345860567e+01, 6.34435290262534e+01, 6.34435285354286e+01, 6.34435189179581e+01, 6.34434537581665e+01, 6.34432142236706e+01, 6.34425757110445e+01, 6.34411783336324e+01, 6.34385021574910e+01, 6.34338481669678e+01, 6.34263258183407e+01, 6.34148478973665e+01, 6.33981332319508e+01, 6.33747176206590e+01, 6.33429731174617e+01, 6.33011355602080e+01, 6.32473399449291e+01, 6.31796629347416e+01, 6.30961714611756e+01, 6.29949760439485e+01, 6.28742871453005e+01, 6.27324726138930e+01, 6.25681140885865e+01, 6.23800601480814e+01, 6.21674740238110e+01, 6.19298738440719e+01, 6.16671636379079e+01, 6.13796536788467e+01, 6.10680691687089e+01, 6.07335467307142e+01, 6.03776186854679e+01, 6.00021856129714e+01, 5.96094782436727e+01, 5.92020102428355e+01, 5.87825239083151e+01, 5.83539311335240e+01, 5.79192521400902e+01, 5.74815544267511e+01, 5.70438941181185e+01, 5.66092614747803e+01, 5.61805318194145e+01, 5.57604226256209e+01, 5.53514570776197e+01, 5.49559340821358e+01, 5.45759045079721e+01, 5.42131533259230e+01, 5.38691872886885e+01, 5.35452277926403e+01, 5.32422085742910e+01, 5.29607779001936e+01, 5.27013049069973e+01, 5.24638897424008e+01, 5.22483771532588e+01, 5.20543731675233e+01, 5.18812645219195e+01, 5.17282404940405e+01, 5.15943168011165e+01, 5.14783612234057e+01, 5.13791205948063e+01, 5.12952487759367e+01, 5.12253351870023e+01, 5.11679334328292e+01, 5.11215895057032e+01, 5.10848690093462e+01, 5.10563828160905e+01, 5.10348105553442e+01, 5.10189213401167e+01, 5.10075911734821e+01, 5.09998165402404e+01, 5.09947237802925e+01, 5.09915739566734e+01, 5.09897630679794e+01, 5.09888176055376e+01, 5.09883856124352e+01, 5.09882235563416e+01, 5.09881794731013e+01, 5.09881729665055e+01, 5.09881724445960e+01, 5.09881628127189e+01, 5.09880975553272e+01, 5.09878576621454e+01, 5.09872181941772e+01, 5.09858187299971e+01, 5.09831385726833e+01, 5.09784777065884e+01, 5.09709443735019e+01, 5.09594499989744e+01, 5.09427120385355e+01, 5.09192651257948e+01, 5.08874806853250e+01, 5.08455949194054e+01, 5.07917447885810e+01, 5.07240112853490e+01, 5.06404689575033e+01, 5.05392402883066e+01, 5.04185532062881e+01, 5.02767997043572e+01, 5.01125933247579e+01, 4.99248231411085e+01, 4.97127018650128e+01, 4.94758058385757e+01, 4.92141049517053e+01, 4.89279809391189e+01, 4.86182330496373e+01, 4.82860707109333e+01, 4.79330934952842e+01, 4.75612593721461e+01, 4.71728428465350e+01, 4.67703850594641e+01, 4.63566382097132e+01, 4.59345067156299e+01, 4.55069873855302e+01, 4.50771105633659e+01, 4.46478838463879e+01, 4.42222396135600e+01, 4.38029873069425e+01, 4.33927711814088e+01, 4.29940340570690e+01, 4.26089874389476e+01, 4.22395881849084e+01, 4.18875217011388e+01, 4.15541914394284e+01, 4.12407142865135e+01, 4.09479212961335e+01, 4.06763631331330e+01, 4.04263195777351e+01, 4.01978124680345e+01, 3.99906215240553e+01, 3.98043025793248e+01, 3.96382078291348e+01, 3.94915077754363e+01, 3.93632145981260e+01, 3.92522067072485e+01, 3.91572542300538e+01, 3.90770451635464e+01, 3.90102118818959e+01, 3.89553576350614e+01, 3.89110826173041e+01, 3.88760091294262e+01, 3.88488053140133e+01, 3.88282069155192e+01, 3.88130365125044e+01, 3.88022196918925e+01, 3.87947976869238e+01, 3.87899360814017e+01, 3.87869292903246e+01, 3.87852006562833e+01, 3.87842981454776e+01, 3.87838857789982e+01, 3.87837310857979e+01, 3.87836890054686e+01, 3.87836827944991e+01, 3.87836823154494e+01, 3.87836737132539e+01, 3.87836154321136e+01, 3.87834011844736e+01, 3.87828300789050e+01, 3.87815802282269e+01, 3.87791866142358e+01, 3.87750240981035e+01, 3.87682963618672e+01, 3.87580314459182e+01, 3.87430844065742e+01, 3.87221474519143e+01, 3.86937677180666e+01, 3.86563726182213e+01, 3.86083024316450e+01, 3.85478495029294e+01, 3.84733031010515e+01, 3.83829986580665e+01, 3.82753697889312e+01, 3.81490012126417e+01, 3.80026804792062e+01, 3.78354462858601e+01, 3.76466311649730e+01, 3.74358964639055e+01, 3.72032578215547e+01, 3.69490997717911e+01, 3.66741786492284e+01, 3.63796136008213e+01, 3.60668661667950e+01, 3.57377095259879e+01, 3.53941890410139e+01, 3.50385761317413e+01, 3.46733177124631e+01, 3.43009834354347e+01, 3.39242128078608e+01, 3.35456639355803e+01, 3.31679652581611e+01, 3.27936712449053e+01, 3.24252226761463e+01, 3.20649118723865e+01, 3.17148530600758e+01, 3.13769579576585e+01, 3.10529165963414e+01, 3.07441833251649e+01, 3.04519678698985e+01, 3.01772312176282e+01, 2.99206859958069e+01, 2.96828009250463e+01, 2.94638088661619e+01, 2.92637179627589e+01, 2.90823253995122e+01, 2.89192333433736e+01, 2.87738666960449e+01, 2.86454923468648e+01, 2.85332396642912e+01, 2.84361219942118e+01, 2.83530589414877e+01, 2.82828991981808e+01, 2.82244436512940e+01, 2.81764684598575e+01, 2.81377477422243e+01, 2.81070754664117e+01, 2.80832860961723e+01, 2.80652735196270e+01, 2.80520077811885e+01, 2.80425491551317e+01, 2.80360591426589e+01, 2.80318080436877e+01, 2.80291788476910e+01, 2.80276673005616e+01, 2.80268781307877e+01, 2.80265175512412e+01, 2.80263822851850e+01, 2.80263454895172e+01, 2.80263400585530e+01, 2.80263396473367e+01, 2.80263323626823e+01, 2.80262830080762e+01, 2.80261015753147e+01, 2.80256179421200e+01, 2.80245595223693e+01, 2.80225325206504e+01, 2.80190075436667e+01, 2.80133102330503e+01, 2.80046174818998e+01, 2.79919596780091e+01, 2.79742292763473e+01, 2.79501958377870e+01, 2.79185274772139e+01, 2.78778184411352e+01, 2.78266222853341e+01, 2.77634898541036e+01, 2.76870109862256e+01, 2.75958586061912e+01, 2.74888336233470e+01, 2.73649088803572e+01, 2.72232702895008e+01, 2.70633532922028e+01, 2.68848728895821e+01, 2.66878457268488e+01, 2.64726030678871e+01, 2.62397939505498e+01, 2.59903783356304e+01, 2.57256106071026e+01, 2.54470142926956e+01, 2.51563492962133e+01, 2.48555732214660e+01, 2.45467985009287e+01, 2.42322470294078e+01, 2.39142038812897e+01, 2.35949715118108e+01, 2.32768256574032e+01, 2.29619739872062e+01, 2.26525184209059e+01, 2.23504218998425e+01, 2.20574802537170e+01, 2.17752996255681e+01, 2.15052796997826e+01, 2.12486027353040e+01, 2.10062281640092e+01, 2.07788923006440e+01, 2.05671125492563e+01, 2.03711953953774e+01, 2.01912474455425e+01, 2.00271888083356e+01, 1.98787681893583e+01, 1.97455791784841e+01, 1.96270773236369e+01, 1.95225976956451e+01, 1.94313727417180e+01, 1.93525502931872e+01, 1.92852116328250e+01, 1.92283895383409e+01, 1.91810862044947e+01, 1.91422909118525e+01, 1.91109972622575e+01, 1.90862197472545e+01, 1.90670093638854e+01, 1.90524679500155e+01, 1.90417608852171e+01, 1.90341277982662e+01, 1.90288909415623e+01, 1.90254609370731e+01, 1.90233396662325e+01, 1.90221201639374e+01, 1.90214834790014e+01, 1.90211925735752e+01, 1.90210834450975e+01, 1.90210537595236e+01, 1.90210493779957e+01, 1.90210490490730e+01, 1.90210432596614e+01, 1.90210040355591e+01, 1.90208598438145e+01, 1.90204754831352e+01, 1.90196343271662e+01, 1.90180234434287e+01, 1.90152221987999e+01, 1.90106948998907e+01, 1.90037879434268e+01, 1.89937318641680e+01, 1.89796485612741e+01, 1.89605638534375e+01, 1.89354253545816e+01, 1.89031254737413e+01, 1.88625291264905e+01, 1.88125055068431e+01, 1.87519630185154e+01, 1.86798862180493e+01, 1.85953733986782e+01, 1.84976732646140e+01, 1.83862190327890e+01, 1.82606582733152e+01, 1.81208768769463e+01, 1.79670157267378e+01, 1.77994789520695e+01, 1.76189330459460e+01, 1.74262966096033e+01, 1.72227210197029e+01, 1.70095628514733e+01, 1.67883493889822e+01, 1.65607389633685e+01, 1.63284781388850e+01, 1.60933578846676e+01, 1.58571708147988e+01, 1.56216713591471e+01, 1.53885403720507e+01, 1.51593552409888e+01, 1.49355660775825e+01, 1.47184781129333e+01, 1.45092400239770e+01, 1.43088376178478e+01, 1.41180921104329e+01, 1.39376621502199e+01, 1.37680487435499e+01, 1.36096023095392e+01, 1.34625312072092e+01, 1.33269112108463e+01, 1.32026955440212e+01, 1.30897252053559e+01, 1.29877394226221e+01, 1.28963861528910e+01, 1.28152326048975e+01, 1.27437757968850e+01, 1.26814531810728e+01, 1.26276533669189e+01, 1.25817269619480e+01, 1.25429975235649e+01, 1.25107725806296e+01, 1.24843546425878e+01, 1.24630520699991e+01, 1.24461896370567e+01, 1.24331185780981e+01, 1.24232258800461e+01, 1.24159425647428e+01, 1.24107507020949e+01, 1.24071889086104e+01, 1.24048561167655e+01, 1.24034134477044e+01, 1.24025840806403e+01, 1.24021510833633e+01, 1.24019532449462e+01, 1.24018790291870e+01, 1.24018588407258e+01, 1.24018558609521e+01, 1.24018556379397e+01, 1.24018517217318e+01, 1.24018251888610e+01, 1.24017276514348e+01, 1.24014676541083e+01, 1.24008986646439e+01, 1.23998090136497e+01, 1.23979142012362e+01, 1.23948519485360e+01, 1.23901803249764e+01, 1.23833792247098e+01, 1.23738553935479e+01, 1.23609511167022e+01, 1.23439565639274e+01, 1.23221256504629e+01, 1.22946951106362e+01, 1.22609063012400e+01, 1.22200290635133e+01, 1.21713867899379e+01, 1.21143816830367e+01, 1.20485190778248e+01, 1.19734296468598e+01, 1.18888883326379e+01, 1.17948289651854e+01, 1.16913537223909e+01, 1.15787368652930e+01, 1.14574225079983e+01, 1.13280165318297e+01, 1.11912730914604e+01, 1.10480764543113e+01, 1.08994191372729e+01, 1.07463774414343e+01, 1.05900855329889e+01, 1.04317091852030e+01, 1.02724201985578e+01, 1.01133723734497e+01, 9.95567974058211e+00, 9.80039757299257e+00, 9.64850652080744e+00, 9.50090003257587e+00, 9.35837506183331e+00, 9.22162591138063e+00, 9.09124094827575e+00, 8.96770183660574e+00, 8.85138488687549e+00, 8.74256411018375e+00, 8.64141558755156e+00, 8.54802281155452e+00, 8.46238271881508e+00, 8.38441219821817e+00, 8.31395492292499e+00, 8.25078840879178e+00, 8.19463124449047e+00, 8.14515046826092e+00, 8.10196908296474e+00, 8.06467370593860e+00, 8.03282234451535e+00, 8.00595227372277e+00, 7.98358797153786e+00, 7.96524904135273e+00, 7.95045802341484e+00, 7.93874796956005e+00, 7.92966963127915e+00, 7.92279809270595e+00, 7.91773866986564e+00, 7.91413189735167e+00, 7.91165743465260e+00, 7.91003674688552e+00, 7.90903444799469e+00, 7.90845823688403e+00, 7.90815740601067e+00, 7.90801995464317e+00, 7.90796839201010e+00, 7.90795436573502e+00, 7.90795229548662e+00, 7.90795214048703e+00, 7.90794941784168e+00, 7.90793097152631e+00, 7.90786316103177e+00, 7.90768240401067e+00, 7.90728682605867e+00, 7.90652926369168e+00, 7.90521190874041e+00, 7.90308285346637e+00, 7.89983476167150e+00, 7.89510585000097e+00, 7.88848331298605e+00, 7.87950926244558e+00, 7.86768917444336e+00, 7.85250274404143e+00, 7.83341694044304e+00, 7.80990093611896e+00, 7.78144245934730e+00, 7.74756499928893e+00, 7.70784518762675e+00, 7.66192960355674e+00, 7.60955021190603e+00, 7.55053765757660e+00, 7.48483170942100e+00, 7.41248827324515e+00, 7.33368257021094e+00, 7.24870828993319e+00, 7.15797275788314e+00, 7.06198838208213e+00, 6.96136084250843e+00, 6.85677464002440e+00, 6.74897671866721e+00, 6.63875891293986e+00, 6.52693995545653e+00, 6.41434772107129e+00, 6.30180229532870e+00, 6.19010035109419e+00, 6.08000120808681e+00, 5.97221484257167e+00, 5.86739201230954e+00, 5.76611656676138e+00, 5.66889992578407e+00, 5.57617763332119e+00, 5.48830782818129e+00, 5.40557142438061e+00, 5.32817376067885e+00, 5.25624746369452e+00, 5.18985627069797e+00, 5.12899957470859e+00, 5.07361748254131e+00, 5.02359621191358e+00, 4.97877369239160e+00, 4.93894527283707e+00, 4.90386947172249e+00, 4.87327373363672e+00, 4.84686017384044e+00, 4.82431130210901e+00, 4.80529571745622e+00, 4.78947375758040e+00, 4.77650307260642e+00, 4.76604407401665e+00, 4.75776518903903e+00, 4.75134783081040e+00, 4.74649097792980e+00, 4.74291524585166e+00, 4.74036632875435e+00, 4.73861769520249e+00, 4.73747243447322e+00, 4.73676417238379e+00, 4.73635700461526e+00, 4.73614443000216e+00, 4.73604730374068e+00, 4.73601086846190e+00, 4.73600095719540e+00, 4.73599949431370e+00, 4.73599938449713e+00, 4.73599745163354e+00, 4.73598435620566e+00, 4.73593621613917e+00, 4.73580789325320e+00, 4.73552706610330e+00, 4.73498926517586e+00, 4.73405407699441e+00, 4.73254270430704e+00, 4.73023704662879e+00, 4.72688043748704e+00, 4.72218013977425e+00, 4.71581165656791e+00, 4.70742485991623e+00, 4.69665187356333e+00, 4.68311656790767e+00, 4.66644543891925e+00, 4.64627955151582e+00, 4.62228713819285e+00, 4.59417636328671e+00, 4.56170770082514e+00, 4.52470533816458e+00, 4.48306701603694e+00, 4.43677175339920e+00, 4.38588498440583e+00, 4.33056075267785e+00, 4.27104075836988e+00, 4.20765022601908e+00, 4.14079074252365e+00, 4.07093038992960e+00, 3.99859165207873e+00, 3.92433769429221e+00, 3.84875769104059e+00, 3.77245190225241e+00, 3.69601717388596e+00, 3.62003346707735e+00, 3.54505191150895e+00, 3.47158474479636e+00, 3.40009735445003e+00, 3.33100249596593e+00, 3.26465663167397e+00, 3.20135822901108e+00, 3.14134777916090e+00, 3.08480924916874e+00, 3.03187266118604e+00, 2.98261749754271e+00, 2.93707665456920e+00, 2.89524070561845e+00, 2.85706227886759e+00, 2.82246040318351e+00, 2.79132472154055e+00, 2.76351951314574e+00, 2.73888750051393e+00, 2.71725344505155e+00, 2.69842755377273e+00, 2.68220873064315e+00, 2.66838770922354e+00, 2.65675009958491e+00, 2.64707937300585e+00, 2.63915979407501e+00, 2.63277929304932e+00, 2.62773225334841e+00, 2.62382217164573e+00, 2.62086413287619e+00, 2.61868703118061e+00, 2.61713546162287e+00, 2.61607120730373e+00, 2.61537425258517e+00, 2.61494326530334e+00, 2.61469550829142e+00, 2.61456616198944e+00, 2.61450706378000e+00, 2.61448489419290e+00, 2.61447886354578e+00, 2.61447797343556e+00, 2.61447790633160e+00, 2.61447672145501e+00, 2.61446869375091e+00, 2.61443918319875e+00, 2.61436051995661e+00, 2.61418837272900e+00, 2.61385871070941e+00, 2.61328548989064e+00, 2.61235918502248e+00, 2.61094627696390e+00, 2.60888979018195e+00, 2.60601095482825e+00, 2.60211204021346e+00, 2.59698036990804e+00, 2.59039348216213e+00, 2.58212534304982e+00, 2.57195345543896e+00, 2.55966663806332e+00, 2.54507318089318e+00, 2.52800902240118e+00, 2.50834554880517e+00, 2.48599659249223e+00, 2.46092421297845e+00, 2.43314288299403e+00, 2.40272177531217e+00, 2.36978494948285e+00, 2.33450936432784e+00, 2.29712078104684e+00, 2.25788775992642e+00, 2.21711407720726e+00, 2.17512998523692e+00, 2.13228279930308e+00, 2.08892731359279e+00, 2.04541652663377e+00, 2.00209309818659e+00, 1.95928187351524e+00, 1.91728370821725e+00, 1.87637071891297e+00, 1.83678298267797e+00, 1.79872661957907e+00, 1.76237312360117e+00, 1.72785976016846e+00, 1.69529082318820e+00, 1.66473953878611e+00, 1.63625041300335e+00, 1.60984184239834e+00, 1.58550883544051e+00, 1.56322572494132e+00, 1.54294878436578e+00, 1.52461869134517e+00, 1.50816280847023e+00, 1.49349727355900e+00, 1.48052890867173e+00, 1.46915696917264e+00, 1.45927476136528e+00, 1.45077116006041e+00, 1.44353205638289e+00, 1.43744176175940e+00, 1.43238438698927e+00, 1.42824520627799e+00, 1.42491200587185e+00, 1.42227640629308e+00, 1.42023513700570e+00, 1.41869123350030e+00, 1.41755512007960e+00, 1.41674553772296e+00, 1.41619027577066e+00, 1.41582666900062e+00, 1.41560182785736e+00, 1.41547257873054e+00, 1.41540510257829e+00, 1.41537427297528e+00, 1.41536270785760e+00, 1.41535956187752e+00, 1.41535909753790e+00, 1.41535906234162e+00, 1.41535843834208e+00, 1.41535421065969e+00, 1.41533866934513e+00, 1.41529724262202e+00, 1.41520658488237e+00, 1.41503297837687e+00, 1.41473111873601e+00, 1.41424335015596e+00, 1.41349941074616e+00, 1.41241674123368e+00, 1.41090139975776e+00, 1.40884961059729e+00, 1.40614995425098e+00, 1.40268617967783e+00, 1.39834058679679e+00, 1.39299788962120e+00, 1.38654942991036e+00, 1.37889757135792e+00, 1.36996006945742e+00, 1.35967418716425e+00, 1.34800031604756e+00, 1.33492487061504e+00, 1.32046225199829e+00, 1.30465572593749e+00, 1.28757712602921e+00, 1.26932537093199e+00, 1.25002386608438e+00, 1.22981693789662e+00, 1.20886551300532e+00, 1.18734230022542e+00, 1.16542675404877e+00, 1.14330009480148e+00, 1.12114063392603e+00, 1.09911960798864e+00, 1.07739766836730e+00, 1.05612211223642e+00, 1.03542488110364e+00, 1.01542130118688e+00, 9.96209499000742e-01, 9.77870397391186e-01, 9.60468181921883e-01, 9.44051123619957e-01, 9.28652649408079e-01, 9.14292563469914e-01, 9.00978338721757e-01, 8.88706415203060e-01, 8.77463459733562e-01, 8.67227557308794e-01, 8.57969318614991e-01, 8.49652899328959e-01, 8.42236935438180e-01, 8.35675404780417e-01, 8.29918428587902e-01, 8.24913028312437e-01, 8.20603852702419e-01, 8.16933888290620e-01, 8.13845163412921e-01, 8.11279451898798e-01, 8.09178977939084e-01, 8.07487118664866e-01, 8.06149095998506e-01, 8.05112644726549e-01, 8.04328639860472e-01, 8.03751663538605e-01, 8.03340490274337e-01, 8.03058469471495e-01, 8.02873785890611e-01, 8.02759582101231e-01, 8.02693931696112e-01, 8.02659657844089e-01, 8.02643998197837e-01, 8.02638123782117e-01, 8.02636525804118e-01, 8.02636289946093e-01, 8.02636271954924e-01, 8.02635951490998e-01, 8.02633780303623e-01, 8.02625798832665e-01, 8.02604523496356e-01, 8.02557964642176e-01, 8.02468805376179e-01, 8.02313777031372e-01, 8.02063264676320e-01, 8.01681173156751e-01, 8.01125080251637e-01, 8.00346698022580e-01, 7.99292655713018e-01, 7.97905607069453e-01, 7.96125651388508e-01, 7.93892041008107e-01, 7.91145128933394e-01, 7.87828489993132e-01, 7.83891129137345e-01, 7.79289673466096e-01, 7.73990432828825e-01, 7.67971209738349e-01, 7.61222744752047e-01, 7.53749699265171e-01, 7.45571103427594e-01, 7.36720230803101e-01, 7.27243900297950e-01, 7.17201245734316e-01, 7.06662029846413e-01, 6.95704608414189e-01, 6.84413668719188e-01, 6.72877872983392e-01, 6.61187532071907e-01, 6.49432419169756e-01, 6.37699810188224e-01, 6.26072810706333e-01, 6.14629001702895e-01, 6.03439411080572e-01, 5.92567797106057e-01, 5.82070214517515e-01, 5.71994824423186e-01, 5.62381904758070e-01, 5.53264018043791e-01, 5.44666296359722e-01, 5.36606808626090e-01, 5.29096981503955e-01, 5.22142051634958e-01, 5.15741533011223e-01, 5.09889688633635e-01, 5.04576000111030e-01, 4.99785632414401e-01, 4.95499893651868e-01, 4.91696691526144e-01, 4.88350989150234e-01, 4.85435263199918e-01, 4.82919967047536e-01, 4.80774000620968e-01, 4.78965187347441e-01, 4.77460756772857e-01, 4.76227829406525e-01, 4.75233898178857e-01, 4.74447298771636e-01, 4.73837659172560e-01, 4.73376317295305e-01, 4.73036694564726e-01, 4.72794613131100e-01, 4.72628544939255e-01, 4.72519782265707e-01, 4.72452521509112e-01, 4.72413854864456e-01, 4.72393667858611e-01, 4.72384444355612e-01, 4.72380984318477e-01, 4.72380043106099e-01, 4.72379904185203e-01, 4.72379893513590e-01, 4.72379702447256e-01, 4.72378407945769e-01, 4.72373649236800e-01, 4.72360964384563e-01, 4.72333204497971e-01, 4.72280043301723e-01, 4.72187602754644e-01, 4.72038213630100e-01, 4.71810327855323e-01, 4.71478596748966e-01, 4.71014125840455e-01, 4.70384912436842e-01, 4.69556466288012e-01, 4.68492606469482e-01, 4.67156419009939e-01, 4.65511350129560e-01, 4.63522399796744e-01, 4.61157370493922e-01, 4.58388117668420e-01, 4.55191742490712e-01, 4.51551665347694e-01, 4.47458520800637e-01, 4.42910821920506e-01, 4.37915353777569e-01, 4.32487271558188e-01, 4.26649896901083e-01, 4.20434224778234e-01, 4.13878170678700e-01, 4.07025602250201e-01, 3.99925209632073e-01, 3.92629273819807e-01, 3.85192392555012e-01, 3.77670219048994e-01, 3.70118261357192e-01, 3.62590780658081e-01, 3.55139816268064e-01, 3.47814354981708e-01, 3.40659653026293e-01, 3.33716711027620e-01, 3.27021896105634e-01, 3.20606700553768e-01, 3.14497623387567e-01, 3.08716159186758e-01, 3.03278877901633e-01, 2.98197579449948e-01, 2.93479507813458e-01, 2.89127610779062e-01, 2.85140833297191e-01, 2.81514434487079e-01, 2.78240320453833e-01, 2.75307387148871e-01, 2.72701869368717e-01, 2.70407693537464e-01, 2.68406833060013e-01, 2.66679665713511e-01, 2.65205332734041e-01, 2.63962098973565e-01, 2.62927712795531e-01, 2.62079763336516e-01, 2.61396031503480e-01, 2.60854829740325e-01, 2.60435324338403e-01, 2.60117833033945e-01, 2.59884089970015e-01, 2.59717469909830e-01, 2.59603163945272e-01, 2.59528299868961e-01, 2.59482001842841e-01, 2.59455385921783e-01, 2.59441490257899e-01, 2.59435141270373e-01, 2.59432759555040e-01, 2.59432111671572e-01, 2.59432016045380e-01, 2.59432008643515e-01, 2.59431875389481e-01, 2.59430972574492e-01, 2.59427653740303e-01, 2.59418807018108e-01, 2.59399446544080e-01, 2.59362370304174e-01, 2.59297898742744e-01, 2.59193707187308e-01, 2.59034763616683e-01, 2.58803381416257e-01, 2.58479394365963e-01, 2.58040458030166e-01, 2.57462477828753e-01, 2.56720159294488e-01, 2.55787670399855e-01, 2.54639399543080e-01, 2.53250786152620e-01, 2.51599194390494e-01, 2.49664794724524e-01, 2.47431413879468e-01, 2.44887311527080e-01, 2.42025842578528e-01, 2.38845967416964e-01, 2.35352578870277e-01, 2.31556623855440e-01, 2.27475008773752e-01, 2.23130289994856e-01, 2.18550163082014e-01, 2.13766775716522e-01, 2.08815898635834e-01, 2.03735995600724e-01, 1.98567237038276e-01, 1.93350502480399e-01, 1.88126414421760e-01, 1.82934441197347e-01, 1.77812099531891e-01, 1.72794279247540e-01, 1.67912703957882e-01, 1.63195533108222e-01, 1.58667103018047e-01, 1.54347798078035e-01, 1.50254038229917e-01, 1.46398365444412e-01, 1.42789610101630e-01, 1.39433117848682e-01, 1.36331018440231e-01, 1.33482519975033e-01, 1.30884214515362e-01, 1.28530384004904e-01, 1.26413298416566e-01, 1.24523500933805e-01, 1.22850077524220e-01, 1.21380910375564e-01, 1.20102916251440e-01, 1.19002271851571e-01, 1.18064628727836e-01, 1.17275320254308e-01, 1.16619562639045e-01, 1.16082651090167e-01, 1.15650151122821e-01, 1.15308083736593e-01, 1.15043101944224e-01, 1.14842655017114e-01, 1.14695135954904e-01, 1.14590007179075e-01, 1.14517899362355e-01, 1.14470678662656e-01, 1.14441478417945e-01, 1.14424692523613e-01, 1.14415929164721e-01, 1.14411925202887e-01, 1.14410423192993e-01, 1.14410014611238e-01, 1.14409954305486e-01, 1.14409949599747e-01, 1.14409864394190e-01, 1.14409287114825e-01, 1.14407164987014e-01, 1.14401508267164e-01, 1.14389129149842e-01, 1.14365423482241e-01, 1.14324204832315e-01, 1.14257599853829e-01, 1.14156013149326e-01, 1.14008168830836e-01, 1.13801234566168e-01, 1.13521031913933e-01, 1.13152334054533e-01, 1.12679248534590e-01, 1.12085678370462e-01, 1.11355849930760e-01, 1.10474890712524e-01, 1.09429434843605e-01, 1.08208229416050e-01, 1.06802711165841e-01, 1.05207521159372e-01, 1.03420925520282e-01, 1.01445113156814e-01, 9.92863469946457e-02, 9.69549531471046e-02, 9.44651422206988e-02, 9.18346677473494e-02, 8.90843375677832e-02, 8.62374038122056e-02, 8.33188649612140e-02, 8.03547185547018e-02, 7.73712049890065e-02, 7.43940814283377e-02, 7.14479604477418e-02, 6.85557412459161e-02, 6.57381529403409e-02, 6.30134204867261e-02, 6.03970550050121e-02, 5.79017624809239e-02, 5.55374585002824e-02, 5.33113722167968e-02, 5.12282202674470e-02, 4.92904307123325e-02, 4.74983979767671e-02, 4.58507517992075e-02, 4.43446259214397e-02, 4.29759153413158e-02, 4.17395141003567e-02, 4.06295285825564e-02, 3.96394639786877e-02, 3.87623838083897e-02, 3.79910441225064e-02, 3.73180052315689e-02, 3.67357245576727e-02, 3.62366345476772e-02, 3.58132095708896e-02, 3.54580254143144e-02, 3.51638144348243e-02, 3.49235186896379e-02, 3.47303425057532e-02, 3.45778050302528e-02, 3.44597923990762e-02, 3.43706083353226e-02, 3.43050213075458e-02, 3.42583058974526e-02, 3.42262757811085e-02, 3.42053057448908e-02, 3.41923404274667e-02, 3.41848879794924e-02, 3.41809975135163e-02, 3.41792200150479e-02, 3.41785532281875e-02, 3.41783718473181e-02, 3.41783450759436e-02, 3.41783429691418e-02, 3.41783045944624e-02, 3.41780446011057e-02, 3.41770888492476e-02, 3.41745412515877e-02, 3.41689663323450e-02, 3.41582913654292e-02, 3.41397327349711e-02, 3.41097510845714e-02, 3.40640396924670e-02, 3.39975502304668e-02, 3.39045595129193e-02, 3.37787798848552e-02, 3.36135144221430e-02, 3.34018560422749e-02, 3.31369269541284e-02, 3.28121516956656e-02, 3.24215535315204e-02, 3.19600605217261e-02, 3.14238045473445e-02, 3.08103944513789e-02, 3.01191436817447e-02, 2.93512337687076e-02, 2.85097978268993e-02, 2.75999129943045e-02, 2.66284969962101e-02, 2.56041112740917e-02, 2.45366805918650e-02, 2.34371458698064e-02, 2.23170723945998e-02, 2.11882388568796e-02, 2.00622334830598e-02, 1.89500817827053e-02, 1.78619264130049e-02, 1.68067739762563e-02, 1.57923170473140e-02, 1.48248332405245e-02, 1.39091573761334e-02, 1.30487182282906e-02, 1.22456281147407e-02, 1.15008117694941e-02, 1.08141605187768e-02, 1.01846986654834e-02, 9.61075091065311e-03, 9.09010214444095e-03, 8.62014350809659e-03, 8.19800088926570e-03, 7.82064385184356e-03, 7.48497448543617e-03, 7.18789688451693e-03, 6.92636894534360e-03, 6.69743883506158e-03, 6.49826880962223e-03, 6.32614907236473e-03, 6.17850418592525e-03, 6.05289428303024e-03, 5.94701303195677e-03, 5.85868401614675e-03, 5.78585687463828e-03, 5.72660421565274e-03, 5.67911996824932e-03, 5.64171949152126e-03, 5.61284143368823e-03, 5.59105104591140e-03, 5.57504441343326e-03, 5.56365289024541e-03, 5.55584691663569e-03, 5.55073837141151e-03, 5.54758066337513e-03, 5.54576589158273e-03, 5.54481859028044e-03, 5.54438580000749e-03, 5.54422345183525e-03, 5.54417928978523e-03, 5.54417277157881e-03, 5.54417225409054e-03, 5.54416277034875e-03, 5.54409851693549e-03, 5.54386231958404e-03, 5.54323274245405e-03, 5.54185512258698e-03, 5.53921756004172e-03, 5.53463314525970e-03, 5.52722976097186e-03, 5.51594884305473e-03, 5.49955447660011e-03, 5.47665409054332e-03, 5.44573173479534e-03, 5.40519442695597e-03, 5.35343130616184e-03, 5.28888432833696e-03, 5.21012802319376e-03, 5.11595451040137e-03, 5.00545869209645e-03, 4.87811750757317e-03, 4.73385656218284e-03, 4.57309752418666e-03, 4.39678053712100e-03, 4.20635753123122e-03, 4.00375462736235e-03, 3.79130458659915e-03, 3.57165317669228e-03, 3.34764604680515e-03, 3.12220483749570e-03, 2.89820239724747e-03, 2.67834682398416e-03, 2.46508254936373e-03, 2.26051418522388e-03, 2.06635603204575e-03, 1.88390768904867e-03, 1.71405438011618e-03, 1.55728917933118e-03, 1.41375289389600e-03, 1.28328595108940e-03, 1.16548585793907e-03, 1.05976436040578e-03, 9.65400334507649e-04, 8.81586743147977e-04, 8.07471466060861e-04, 7.42191968072348e-04, 6.84903224467348e-04, 6.34798143678950e-04, 5.91120453776739e-04, 5.53171213425685e-04, 5.20310892042630e-04, 4.91958871904785e-04, 4.67591515516119e-04, 4.46739219867709e-04, 4.28982551227364e-04, 4.13947617730059e-04, 4.01301031445200e-04, 3.90744898979278e-04, 3.82012202834403e-04, 3.74862775643692e-04, 3.69079929427585e-04, 3.64467723883877e-04, 3.60848829245697e-04, 3.58062930686742e-04, 3.55965608662245e-04, 3.54427611696906e-04, 3.53334417496853e-04, 3.52585965046615e-04, 3.52096434397708e-04, 3.51793957847235e-04, 3.51620160259358e-04, 3.51529450185240e-04, 3.51488010380763e-04, 3.51472465936808e-04, 3.51468237553831e-04, 3.51467613455725e-04, 3.51467563447727e-04, 3.51466641319576e-04, 3.51460393730463e-04, 3.51437427823953e-04, 3.51376215476762e-04, 3.51242285887565e-04, 3.50985917762614e-04, 3.50540476304673e-04, 3.49821556383745e-04, 3.48727109511060e-04, 3.47138737822661e-04, 3.44924328643970e-04, 3.41942161739439e-04, 3.38046547113890e-04, 3.33094935044187e-04, 3.26956288517977e-04, 3.19520333420087e-04, 3.10707118813980e-04, 3.00476147018003e-04, 2.88834189476006e-04, 2.75840807262284e-04, 2.61610589447514e-04, 2.46311283746081e-04, 2.30157413028564e-04, 2.13399691233462e-04, 1.96311450167741e-04, 1.79174005054961e-04, 1.62262924746415e-04, 1.45836276360001e-04, 1.30124565046685e-04, 1.15321419059455e-04, 1.01574828335060e-04, 8.89804926487159e-05, 7.75804626300669e-05, 6.73701360227597e-05, 5.83125695641229e-05, 5.03528696214062e-05, 4.34255253196625e-05, 3.74547982744389e-05, 3.23512176692020e-05, 2.80084708492723e-05, 2.43127006882560e-05, 2.11653246964261e-05, 1.84927668840373e-05, 1.62345800589492e-05, 1.43406863770679e-05, 1.27766075479485e-05, 1.15021544075944e-05, 1.04495827883966e-05, 9.55043554995427e-06, 8.77289225797995e-06, 8.11370074416118e-06, 7.56645938793099e-06, 7.11215733417703e-06, 6.73105542148278e-06, 6.41029753012390e-06, 6.14224597939028e-06, 5.92122188436406e-06, 5.74175389544962e-06, 5.59798603205763e-06, 5.48396570820842e-06, 5.39434778090708e-06, 5.32479860737567e-06, 5.27187418697395e-06, 5.23267658536497e-06, 5.20460822883706e-06, 5.18529979842697e-06, 5.17263637789256e-06, 5.16480031365063e-06, 5.16029453478950e-06, 5.15794195621311e-06, 5.15686704716245e-06, 5.15646381415991e-06, 5.15635412057236e-06, 5.15633792019798e-06, 5.15633668385362e-06, 5.15631414543805e-06, 5.15616143559328e-06, 5.15560013525374e-06, 5.15410438595154e-06, 5.15083347352174e-06, 5.14457880740338e-06, 5.13373107935422e-06, 5.11627443700036e-06, 5.08981385394236e-06, 5.05164092601262e-06, 4.99884525595462e-06, 4.92849353861402e-06, 4.83793778016661e-06, 4.72537340101553e-06, 4.59081313408660e-06, 4.43757605239625e-06, 4.27384738088676e-06, 4.11186819349041e-06, 3.95921058782278e-06, 3.80074108363715e-06, 3.58876322081869e-06, 3.26875364464740e-06, 2.86514450920095e-06, 2.60151459485769e-06, 2.52636606035139e-06, 2.28887989806026e-06, 1.69992033249855e-06, 1.29942637272507e-06, 1.83800163711112e-06, 1.89898635483132e-06, 1.20170670925290e-06, 1.29565641771023e-06, 1.62013397186911e-06, 1.17577009461911e-06, 2.25622113703139e-06, 3.79123368077217e-06, 4.55418440026278e-06, 4.90014758725010e-06, 4.49688906201118e-06, 3.41754456201215e-06, 2.06427768034700e-06, 9.24780650279899e-07, 5.94829586475498e-07, 5.59282028689267e-07, 1.01213085201834e-06, 1.21275735409654e-06, 1.37338132333298e-06, 1.53276498338073e-06, 1.38774248625850e-06, 9.52417138933015e-07, 6.11695639918253e-07, 9.03702087384772e-07, 1.00781117851056e-06, 9.46952980029891e-07, 8.00125889099945e-07, 6.49241101768894e-07, 5.40733452680797e-07, 4.81968768895058e-07, 4.64424928736662e-07, 4.57411289289666e-07, 4.37733547187236e-07, 3.99226934177932e-07, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14],
[ 7.90000000000000e+01, 7.89999999104072e+01, 7.89999971400218e+01, 7.89999783702209e+01, 7.89999093699683e+01, 7.89997254369486e+01, 7.89993228853618e+01, 7.89985518793696e+01, 7.89972108696142e+01, 7.89950428403830e+01, 7.89917335187408e+01, 7.89869116366901e+01, 7.89801512759519e+01, 7.89709762653352e+01, 7.89588665458835e+01, 7.89432663716346e+01, 7.89235941755864e+01, 7.88992539016759e+01, 7.88696475828893e+01, 7.88341889299321e+01, 7.87923176796496e+01, 7.87435144324540e+01, 7.86873156788662e+01, 7.86233286745388e+01, 7.85512457719326e+01, 7.84708577607157e+01, 7.83820657178135e+01, 7.82848908348358e+01, 7.81794816891209e+01, 7.80661184663984e+01, 7.79452137344045e+01, 7.78173095064972e+01, 7.76830705130050e+01, 7.75432737990477e+01, 7.73987949696535e+01, 7.72505915840834e+01, 7.70996843422759e+01, 7.69471367943799e+01, 7.67940343344929e+01, 7.66414632150674e+01, 7.64904902486459e+01, 7.63421437624283e+01, 7.61973962537720e+01, 7.60571490749712e+01, 7.59222193644528e+01, 7.57933293458671e+01, 7.56710980395462e+01, 7.55560353722990e+01, 7.54485386290549e+01, 7.53488911597940e+01, 7.52572632335065e+01, 7.51737149139902e+01, 7.50982008172140e+01, 7.50305765947832e+01, 7.49706069716234e+01, 7.49179751480405e+01, 7.48722933570929e+01, 7.48331143484909e+01, 7.47999435510871e+01, 7.47722516487260e+01, 7.47494872901854e+01, 7.47310896445838e+01, 7.47165005102928e+01, 7.47051756892755e+01, 7.46965953507812e+01, 7.46902731290458e+01, 7.46857637291977e+01, 7.46826688536069e+01, 7.46806413065880e+01, 7.46793871873240e+01, 7.46786661373756e+01, 7.46782896681016e+01, 7.46781176524774e+01, 7.46780531228790e+01, 7.46780355692070e+01, 7.46780329783197e+01, 7.46780326384500e+01, 7.46780247198923e+01, 7.46779710704830e+01, 7.46777738493375e+01, 7.46772481283180e+01, 7.46760975913109e+01, 7.46738941392294e+01, 7.46700622047110e+01, 7.46638684768684e+01, 7.46544176168644e+01, 7.46406544081328e+01, 7.46213726285488e+01, 7.45952307534153e+01, 7.45607743960772e+01, 7.45164651670428e+01, 7.44607153846396e+01, 7.43919278053339e+01, 7.43085392685252e+01, 7.42090668822395e+01, 7.40921551318206e+01, 7.39566220990592e+01, 7.38015028655266e+01, 7.36260881748672e+01, 7.34299565734465e+01, 7.32129985511719e+01, 7.29754316527061e+01, 7.27178060787504e+01, 7.24410008712277e+01, 7.21462112813962e+01, 7.18349282707463e+01, 7.15089112438655e+01, 7.11701550747869e+01, 7.08208523439703e+01, 7.04633515755668e+01, 7.01001122734618e+01, 6.97336577576368e+01, 6.93665271542654e+01, 6.90012282471352e+01, 6.86401930651084e+01, 6.82857379230245e+01, 6.79400291460525e+01, 6.76050550279718e+01, 6.72826039150269e+01, 6.69742478505799e+01, 6.66813310361589e+01, 6.64049624212176e+01, 6.61460119205495e+01, 6.59051099652322e+01, 6.56826502482079e+01, 6.54787956035794e+01, 6.52934869675942e+01, 6.51264553322278e+01, 6.49772365428080e+01, 6.48451887261187e+01, 6.47295120737937e+01, 6.46292706504432e+01, 6.45434158464144e+01, 6.44708110500040e+01, 6.44102570724937e+01, 6.43605178220166e+01, 6.43203456908108e+01, 6.42885060977639e+01, 6.42638006178265e+01, 6.42450881354206e+01, 6.42313034834392e+01, 6.42214730749064e+01, 6.42147271015475e+01, 6.42103079615795e+01, 6.42075746854973e+01, 6.42060032495745e+01, 6.42051827970527e+01, 6.42048079206437e+01, 6.42046672907717e+01, 6.42046290359623e+01, 6.42046233896299e+01, 6.42046228911662e+01, 6.42046131240157e+01, 6.42045469501200e+01, 6.42043036876690e+01, 6.42036552376722e+01, 6.42022361125122e+01, 6.41995182866210e+01, 6.41947918661750e+01, 6.41871524497775e+01, 6.41754959073809e+01, 6.41585211376060e+01, 6.41347411702633e+01, 6.41025027572655e+01, 6.40600143384461e+01, 6.40053819791701e+01, 6.39366525586508e+01, 6.38518631520116e+01, 6.37490952122816e+01, 6.36265318436173e+01, 6.34825161911439e+01, 6.33156087836778e+01, 6.31246415772207e+01, 6.29087664751244e+01, 6.26674962491076e+01, 6.24007360455776e+01, 6.21088040165116e+01, 6.17924400426330e+01, 6.14528020003212e+01, 6.10914495489338e+01, 6.07103159696527e+01, 6.03116691514011e+01, 5.98980633594373e+01, 5.94722838860883e+01, 5.90372870094034e+01, 5.85961378202674e+01, 5.81519483943623e+01, 5.77078184946683e+01, 5.72667805454988e+01, 5.68317501000113e+01, 5.64054825148776e+01, 5.59905361170860e+01, 5.55892418365375e+01, 5.52036790879909e+01, 5.48356575942823e+01, 5.44867048139062e+01, 5.41580586352486e+01, 5.38506650026373e+01, 5.35651801350512e+01, 5.33019769869860e+01, 5.30611555885046e+01, 5.28425568941048e+01, 5.26457797707484e+01, 5.24702007630317e+01, 5.23149962836148e+01, 5.21791668838685e+01, 5.20615632579266e+01, 5.19609136195311e+01, 5.18758520642722e+01, 5.18049474916673e+01, 5.17467326158159e+01, 5.16997325455706e+01, 5.16624923717474e+01, 5.16336031666399e+01, 5.16117257864275e+01, 5.15956118754161e+01, 5.15841215062767e+01, 5.15762369544448e+01, 5.15710721971814e+01, 5.15678778457559e+01, 5.15660413578869e+01, 5.15650825303374e+01, 5.15646444305998e+01, 5.15644800836953e+01, 5.15644353772988e+01, 5.15644287787265e+01, 5.15644282494393e+01, 5.15644184814070e+01, 5.15643523015442e+01, 5.15641090172643e+01, 5.15634605099023e+01, 5.15620412633744e+01, 5.15593232210505e+01, 5.15545964743358e+01, 5.15469566652997e+01, 5.15352998451467e+01, 5.15183253671726e+01, 5.14945472025555e+01, 5.14623138451330e+01, 5.14198367137216e+01, 5.13652266669388e+01, 5.12965379196835e+01, 5.12118183019087e+01, 5.11091644448767e+01, 5.09867801396149e+01, 5.08430358135341e+01, 5.06765268434242e+01, 5.04861282949618e+01, 5.02710436749805e+01, 5.00308454200840e+01, 4.97655051306372e+01, 4.94754119879294e+01, 4.91613783468666e+01, 4.88246321460313e+01, 4.84667964763073e+01, 4.80898573394699e+01, 4.76961212397353e+01, 4.72881647141042e+01, 4.68687781669378e+01, 4.64409064101318e+01, 4.60075881472746e+01, 4.55718963434500e+01, 4.51368810753471e+01, 4.47055161307192e+01, 4.42806503588284e+01, 4.38649645616007e+01, 4.34609345301774e+01, 4.30708006412115e+01, 4.26965442142490e+01, 4.23398706003240e+01, 4.20021987441371e+01, 4.16846567658370e+01, 4.13880829666834e+01, 4.11130315873193e+01, 4.08597826364350e+01, 4.06283551490467e+01, 4.04185233092838e+01, 4.02298349630483e+01, 4.00616321340541e+01, 3.99130732297633e+01, 3.97831566736690e+01, 3.96707457238618e+01, 3.95745942351737e+01, 3.94933730965603e+01, 3.94256970318363e+01, 3.93701513968386e+01, 3.93253185467402e+01, 3.92898032911288e+01, 3.92622569090237e+01, 3.92413991680996e+01, 3.92260377877871e+01, 3.92150848088032e+01, 3.92075693841989e+01, 3.92026465890192e+01, 3.91996019545835e+01, 3.91978515643772e+01, 3.91969376948951e+01, 3.91965201385431e+01, 3.91963634984424e+01, 3.91963208885090e+01, 3.91963145993711e+01, 3.91963141142923e+01, 3.91963054038334e+01, 3.91962463891932e+01, 3.91960294451327e+01, 3.91954511519160e+01, 3.91941855713437e+01, 3.91917618330974e+01, 3.91875469318008e+01, 3.91807345303385e+01, 3.91703404430219e+01, 3.91552053296405e+01, 3.91340049634211e+01, 3.91052682373158e+01, 3.90674028399259e+01, 3.90187282635378e+01, 3.89575155052716e+01, 3.88820324968038e+01, 3.87905939633309e+01, 3.86816140893735e+01, 3.85536600838400e+01, 3.84055045188237e+01, 3.82361741954748e+01, 3.80449932921943e+01, 3.78316186942145e+01, 3.75960656971029e+01, 3.73387227133132e+01, 3.70603541678285e+01, 3.67620914070577e+01, 3.64454121111401e+01, 3.61121093311945e+01, 3.57642518059677e+01, 3.54041375915958e+01, 3.50342432276480e+01, 3.46571706535997e+01, 3.42755939044985e+01, 3.38922073021731e+01, 3.35096764845835e+01, 3.31305932472880e+01, 3.27574348587723e+01, 3.23925282804512e+01, 3.20380195685637e+01, 3.16958486330139e+01, 3.13677294431721e+01, 3.10551356747413e+01, 3.07592916735602e+01, 3.04811684788079e+01, 3.02214845195973e+01, 2.99807104984515e+01, 2.97590779192231e+01, 2.95565907105250e+01, 2.93730394318357e+01, 2.92080176134082e+01, 2.90609398555935e+01, 2.89310613826318e+01, 2.88174987991539e+01, 2.87192518284506e+01, 2.86352258183822e+01, 2.85642547854482e+01, 2.85051247342364e+01, 2.84565969439259e+01, 2.84174308623322e+01, 2.83864061981313e+01, 2.83623437603139e+01, 2.83441245670309e+01, 2.83307067392442e+01, 2.83211397119828e+01, 2.83145753396677e+01, 2.83102755419255e+01, 2.83076162303652e+01, 2.83060873707119e+01, 2.83052891625112e+01, 2.83049244532949e+01, 2.83047876380668e+01, 2.83047504209877e+01, 2.83047449278241e+01, 2.83047445118983e+01, 2.83047371438147e+01, 2.83046872239640e+01, 2.83045037133113e+01, 2.83040145412808e+01, 2.83029440002246e+01, 2.83008937859087e+01, 2.82973284456319e+01, 2.82915659068847e+01, 2.82827736563280e+01, 2.82699710183814e+01, 2.82520378408104e+01, 2.82277297271372e+01, 2.81956997592383e+01, 2.81545264271857e+01, 2.81027472298985e+01, 2.80388971367003e+01, 2.79615508188469e+01, 2.78693672889895e+01, 2.77611353471708e+01, 2.76358180487487e+01, 2.74925943072953e+01, 2.73308957459997e+01, 2.71504370303199e+01, 2.69512381591025e+01, 2.67336375551224e+01, 2.64982952583139e+01, 2.62461860499666e+01, 2.59785828749552e+01, 2.56970314259250e+01, 2.54033171556317e+01, 2.50994262538577e+01, 2.47875022513106e+01, 2.44697999110816e+01, 2.41486379773470e+01, 2.38263522174668e+01, 2.35052500545199e+01, 2.31875679586252e+01, 2.28754326405577e+01, 2.25708269496865e+01, 2.22755611992524e+01, 2.19912504160586e+01, 2.17192977463634e+01, 2.14608839677873e+01, 2.12169627890017e+01, 2.09882613946527e+01, 2.07752855342466e+01, 2.05783283706388e+01, 2.03974822943116e+01, 2.02326529620935e+01, 2.00835749157063e+01, 1.99498282567721e+01, 1.98308559819952e+01, 1.97259816996806e+01, 1.96344275453202e+01, 1.95553321828774e+01, 1.94877688169603e+01, 1.94307631500494e+01, 1.93833112019101e+01, 1.93443968708400e+01, 1.93130090654905e+01, 1.92881581794397e+01, 1.92688916264547e+01, 1.92543081100506e+01, 1.92435702730954e+01, 1.92359153668980e+01, 1.92306635975091e+01, 1.92272238506952e+01, 1.92250965646960e+01, 1.92238736077702e+01, 1.92232351202175e+01, 1.92229433913918e+01, 1.92228339540659e+01, 1.92228041844813e+01, 1.92227997905538e+01, 1.92227994607003e+01, 1.92227936549047e+01, 1.92227543198005e+01, 1.92226097200241e+01, 1.92222242718577e+01, 1.92213807368419e+01, 1.92197653003638e+01, 1.92169561491762e+01, 1.92124160998711e+01, 1.92054897576732e+01, 1.91954055977762e+01, 1.91812832548400e+01, 1.91621461750586e+01, 1.91369396251789e+01, 1.91045538621139e+01, 1.90638520468771e+01, 1.90137022434562e+01, 1.89530125878369e+01, 1.88807684604581e+01, 1.87960702667960e+01, 1.86981702481048e+01, 1.85865066307311e+01, 1.84607333991656e+01, 1.83207440617168e+01, 1.81666879777643e+01, 1.79989781316869e+01, 1.78182896591935e+01, 1.76255489332890e+01, 1.74219135643336e+01, 1.72087442170909e+01, 1.69875696473667e+01, 1.67600467621565e+01, 1.65279177676503e+01, 1.62929665606827e+01, 1.60569764326309e+01, 1.58216909040658e+01, 1.55887791282174e+01, 1.53598068414954e+01, 1.51362133568178e+01, 1.49192946445934e+01, 1.47101921701349e+01, 1.45098868818196e+01, 1.43191975801691e+01, 1.41387828368911e+01, 1.39691456560379e+01, 1.38106401522602e+01, 1.36634796385295e+01, 1.35277456458962e+01, 1.34033975244182e+01, 1.32902823867738e+01, 1.31881452487834e+01, 1.30966392923663e+01, 1.30153362269171e+01, 1.29437367564279e+01, 1.28812811739507e+01, 1.28273601042569e+01, 1.27813254017844e+01, 1.27425011861988e+01, 1.27101949644048e+01, 1.26837087483073e+01, 1.26623500351829e+01, 1.26454424757693e+01, 1.26323360179678e+01, 1.26224162852534e+01, 1.26151129320146e+01, 1.26099067159454e+01, 1.26063350420498e+01, 1.26039957642730e+01, 1.26025490782636e+01, 1.26017173998964e+01, 1.26012831953432e+01, 1.26010848051820e+01, 1.26010103824237e+01, 1.26009901376516e+01, 1.26009871495665e+01, 1.26009869259321e+01, 1.26009829988007e+01, 1.26009563919212e+01, 1.26008585824188e+01, 1.26005978597416e+01, 1.26000272823812e+01, 1.25989345885229e+01, 1.25970344787435e+01, 1.25939636485215e+01, 1.25892789006602e+01, 1.25824586096711e+01, 1.25729077376783e+01, 1.25599665101929e+01, 1.25429227465986e+01, 1.25210277026416e+01, 1.24935151218384e+01, 1.24596230146130e+01, 1.24186174977370e+01, 1.23698178463141e+01, 1.23126217537586e+01, 1.22465296814913e+01, 1.21711671282773e+01, 1.20863036743961e+01, 1.19918677664297e+01, 1.18879564033170e+01, 1.17748391517409e+01, 1.16529562371612e+01, 1.15229107966476e+01, 1.13854557085926e+01, 1.12414757018071e+01, 1.10919656685617e+01, 1.09380062491271e+01, 1.07807378166778e+01, 1.06213339779497e+01, 1.04609756296235e+01, 1.03008264877832e+01, 1.01420108513773e+01, 9.98559418136720e+00, 9.83256688453463e+00, 9.68383149440226e+00, 9.54019325288701e+00, 9.40235392829501e+00, 9.27090857092122e+00, 9.14634481639131e+00, 9.02904430271735e+00, 8.91928576684181e+00, 8.81724942153899e+00, 8.72302227201101e+00, 8.63660410102435e+00, 8.55791392147688e+00, 8.48679675837678e+00, 8.42303067399941e+00, 8.36633398868575e+00, 8.31637267530029e+00, 8.27276791866471e+00, 8.23510383364439e+00, 8.20293532844411e+00, 8.17579608463596e+00, 8.15320660421420e+00, 8.13468224850710e+00, 8.11974116632128e+00, 8.10791198175480e+00, 8.09874108835275e+00, 8.09179937832410e+00, 8.08668822577607e+00, 8.08304454322962e+00, 8.08054474221217e+00, 8.07890745174161e+00, 8.07789488230245e+00, 8.07731276580481e+00, 8.07700885154308e+00, 8.07686999129423e+00, 8.07681790013208e+00, 8.07680373008344e+00, 8.07680163861434e+00, 8.07680148202596e+00, 8.07679873147259e+00, 8.07678009607608e+00, 8.07671159049424e+00, 8.07652898058720e+00, 8.07612934744455e+00, 8.07536401815500e+00, 8.07403315412149e+00, 8.07188225830028e+00, 8.06860082813493e+00, 8.06382333926997e+00, 8.05713269384974e+00, 8.04806620367644e+00, 8.03612410053491e+00, 8.02078047258409e+00, 8.00149641775235e+00, 7.97773508592741e+00, 7.94897815762042e+00, 7.91474318674743e+00, 7.87460113055320e+00, 7.82819331307869e+00, 7.77524703217381e+00, 7.71558903383945e+00, 7.64915614734086e+00, 7.57600249984051e+00, 7.49630290318753e+00, 7.41035221425811e+00, 7.31856069499315e+00, 7.22144561763552e+00, 7.11961955410479e+00, 7.01377593966316e+00, 6.90467260043002e+00, 6.79311397992484e+00, 6.67993279648064e+00, 6.56597182035429e+00, 6.45206638756704e+00, 6.33902817692949e+00, 6.22763067505498e+00, 6.11859664678055e+00, 6.01258781911015e+00, 5.91019687900257e+00, 5.81194178293552e+00, 5.71826228368344e+00, 5.62951850182387e+00, 5.54599131015082e+00, 5.46788426100299e+00, 5.39532677012248e+00, 5.32837827460730e+00, 5.26703310360308e+00, 5.21122583417829e+00, 5.16083694635863e+00, 5.11569863561395e+00, 5.07560068376695e+00, 5.04029632673172e+00, 5.00950808708050e+00, 4.98293355960531e+00, 4.96025114819349e+00, 4.94112575278207e+00, 4.92521439698027e+00, 4.91217177189334e+00, 4.90165565197880e+00, 4.89333211699359e+00, 4.88688049296317e+00, 4.88199790726671e+00, 4.87840334075439e+00, 4.87584105515239e+00, 4.87408327806263e+00, 4.87293204101550e+00, 4.87222008783712e+00, 4.87181079979874e+00, 4.87159711871840e+00, 4.87149948701814e+00, 4.87146286215049e+00, 4.87145289931297e+00, 4.87145142881955e+00, 4.87145131843158e+00, 4.87144937551088e+00, 4.87143621194542e+00, 4.87138782140923e+00, 4.87125883095111e+00, 4.87097654328218e+00, 4.87043594699648e+00, 4.86949590301633e+00, 4.86797669644082e+00, 4.86565912036276e+00, 4.86228523050143e+00, 4.85756087458350e+00, 4.85116005610085e+00, 4.84273113656442e+00, 4.83190481273962e+00, 4.81830372607264e+00, 4.80155347289606e+00, 4.78129469037805e+00, 4.75719580090745e+00, 4.72896591471957e+00, 4.69636732609150e+00, 4.65922700141810e+00, 4.61744645580991e+00, 4.57100945403678e+00, 4.51998705371061e+00, 4.46453963136389e+00, 4.40491568885875e+00, 4.34144741750468e+00, 4.27454318626086e+00, 4.20467730249594e+00, 4.13237755293392e+00, 4.05821115438837e+00, 3.98276981799799e+00, 3.90665465115862e+00, 3.83046158810168e+00, 3.75476795862676e+00, 3.68012068519946e+00, 3.60702645512494e+00, 3.53594406184607e+00, 3.46727896204293e+00, 3.40137996525394e+00, 3.33853786883122e+00, 3.27898577779901e+00, 3.22290080735482e+00, 3.17040685286361e+00, 3.12157812351262e+00, 3.07644316534939e+00, 3.03498914095405e+00, 2.99716618065567e+00, 2.96289166903098e+00, 2.93205437656376e+00, 2.90451838704956e+00, 2.88012680488160e+00, 2.85870525185577e+00, 2.84006518035446e+00, 2.82400703896043e+00, 2.81032332832805e+00, 2.79880158038934e+00, 2.78922728381561e+00, 2.78138676440368e+00, 2.77507001219709e+00, 2.77007342929606e+00, 2.76620245513782e+00, 2.76327401118979e+00, 2.76111869599544e+00, 2.75958265557614e+00, 2.75852905414210e+00, 2.75783907622632e+00, 2.75741240349231e+00, 2.75716712680247e+00, 2.75703907541882e+00, 2.75698056886063e+00, 2.75695862122110e+00, 2.75695265094897e+00, 2.75695176974996e+00, 2.75695170331781e+00, 2.75695053030347e+00, 2.75694258296768e+00, 2.75691336785704e+00, 2.75683549214615e+00, 2.75666506837415e+00, 2.75633870683504e+00, 2.75577122512985e+00, 2.75485419498017e+00, 2.75345543499095e+00, 2.75141954319196e+00, 2.74856954302578e+00, 2.74470968901771e+00, 2.73962944197052e+00, 2.73310857720718e+00, 2.72492333333263e+00, 2.71485344495493e+00, 2.70268983432061e+00, 2.68824266919738e+00, 2.67134943438091e+00, 2.65188261956922e+00, 2.62975660462547e+00, 2.60493333079280e+00, 2.57742638717325e+00, 2.54730321622246e+00, 2.51468524655604e+00, 2.47974588829466e+00, 2.44270646430800e+00, 2.40383028668896e+00, 2.36341520785059e+00, 2.32178506760866e+00, 2.27928051266271e+00, 2.23624967878098e+00, 2.19303919956553e+00, 2.14998594445804e+00, 2.10740980177863e+00, 2.06560772120751e+00, 2.02484912563516e+00, 1.98537270486359e+00, 1.94738452098478e+00, 1.91105729219555e+00, 1.87653068015221e+00, 1.84391238498741e+00, 1.81327984917557e+00, 1.78468238278471e+00, 1.75814354411240e+00, 1.73366363723915e+00, 1.71122221812738e+00, 1.69078053075768e+00, 1.67228382239183e+00, 1.65566351106591e+00, 1.64083919810712e+00, 1.62772053355307e+00, 1.61620895285623e+00, 1.60619930941805e+00, 1.59758142968057e+00, 1.59024161615054e+00, 1.58406411935967e+00, 1.57893259293996e+00, 1.57473153735982e+00, 1.57134772813993e+00, 1.56867161432662e+00, 1.56659866347496e+00, 1.56503062120932e+00, 1.56387664735715e+00, 1.56305428733249e+00, 1.56249023731856e+00, 1.56212086504620e+00, 1.56189245446942e+00, 1.56176115199317e+00, 1.56169260343620e+00, 1.56166128375786e+00, 1.56164953478229e+00, 1.56164633878711e+00, 1.56164586706534e+00, 1.56164583130948e+00, 1.56164519738944e+00, 1.56164090249364e+00, 1.56162511408704e+00, 1.56158302864260e+00, 1.56149092897944e+00, 1.56131455976086e+00, 1.56100789180522e+00, 1.56051234114892e+00, 1.55975650301654e+00, 1.55865645311960e+00, 1.55711665789771e+00, 1.55503152023931e+00, 1.55228756683317e+00, 1.54876625700585e+00, 1.54434736080350e+00, 1.53891281724425e+00, 1.53235094427883e+00, 1.52456083333192e+00, 1.51545672755039e+00, 1.50497215873116e+00, 1.49306360787334e+00, 1.47971346202503e+00, 1.46493206753766e+00, 1.44875872667807e+00, 1.43126154787807e+00, 1.41253613435742e+00, 1.39270317419874e+00, 1.37190506904361e+00, 1.35030180058769e+00, 1.32806627763811e+00, 1.30537942776765e+00, 1.28242529559159e+00, 1.25938638636912e+00, 1.23643945334793e+00, 1.21375187593057e+00, 1.19147871977633e+00, 1.16976051536552e+00, 1.14872174320290e+00, 1.12846997500188e+00, 1.10909559248502e+00, 1.09067198896986e+00, 1.07325615262825e+00, 1.05688953241362e+00, 1.04159909599744e+00, 1.02739850147870e+00, 1.01428931918690e+00, 1.00226225499185e+00, 9.91298340967532e-01, 9.81370072212826e-01, 9.72442479622160e-01, 9.64474137185178e-01, 9.57418108929417e-01, 9.51222844970900e-01, 9.45833038451894e-01, 9.41190455609188e-01, 9.37234750051061e-01, 9.33904269767311e-01, 9.31136861726557e-01, 9.28870674436567e-01, 9.27044953900705e-01, 9.25600823380549e-01, 9.24482032676636e-01, 9.23635658677640e-01, 9.23012736083834e-01, 9.22568795800231e-01, 9.22264288738177e-01, 9.22064874741295e-01, 9.21941560001997e-01, 9.21870671439422e-01, 9.21833662720061e-01, 9.21816753472726e-01, 9.21810410285630e-01, 9.21808684790165e-01, 9.21808430110858e-01, 9.21808410684008e-01, 9.21808064647252e-01, 9.21805720200098e-01, 9.21797101806661e-01, 9.21774128658669e-01, 9.21723854133731e-01, 9.21627578716738e-01, 9.21460174710219e-01, 9.21189658347458e-01, 9.20777041614386e-01, 9.20176491728157e-01, 9.19335820230472e-01, 9.18197315407710e-01, 9.16698920715120e-01, 9.14775747745963e-01, 9.12361895123567e-01, 9.09392525085053e-01, 9.05806128638332e-01, 9.01546889772078e-01, 8.96567041516064e-01, 8.90829094170883e-01, 8.84307811133695e-01, 8.76991812314969e-01, 8.68884700124795e-01, 8.60005628124327e-01, 8.50389266031513e-01, 8.40085153896325e-01, 8.29156479049999e-01, 8.17678347608165e-01, 8.05735653862428e-01, 7.93420672680085e-01, 7.80830510228753e-01, 7.68064546679469e-01, 7.55221992208718e-01, 7.42399656983551e-01, 7.29690009928092e-01, 7.17179573170512e-01, 7.04947672101568e-01, 6.93065537258491e-01, 6.81595735296540e-01, 6.70591892819402e-01, 6.60098668785707e-01, 6.50151928031661e-01, 6.40779069251963e-01, 6.31999464528633e-01, 6.23824973160453e-01, 6.16260499215329e-01, 6.09304569159207e-01, 6.02949912558941e-01, 5.97184034826938e-01, 5.91989776048280e-01, 5.87345853984942e-01, 5.83227392353692e-01, 5.79606437446245e-01, 5.76452467153686e-01, 5.73732896556294e-01, 5.71413583537947e-01, 5.69459336498625e-01, 5.67834424304621e-01, 5.66503086291746e-01, 5.65430037604969e-01, 5.64580962622543e-01, 5.63922986888187e-01, 5.63425116081476e-01, 5.63058629291746e-01, 5.62797413390342e-01, 5.62618225727206e-01, 5.62500873749400e-01, 5.62428302405451e-01, 5.62386583239515e-01, 5.62364802698852e-01, 5.62354851132369e-01, 5.62351117979247e-01, 5.62350102473246e-01, 5.62349952586771e-01, 5.62349941072810e-01, 5.62349734924894e-01, 5.62348338243638e-01, 5.62343203915219e-01, 5.62329517831031e-01, 5.62299566933305e-01, 5.62242210303751e-01, 5.62142475787880e-01, 5.61981302839214e-01, 5.61735450090623e-01, 5.61377583387738e-01, 5.60876556353250e-01, 5.60197890673393e-01, 5.59304456940524e-01, 5.58157348896827e-01, 5.56716934357695e-01, 5.54944055268123e-01, 5.52801337895318e-01, 5.50254563023998e-01, 5.47274036391073e-01, 5.43835892788664e-01, 5.39923264539652e-01, 5.35527247398619e-01, 5.30647604853073e-01, 5.25293165146678e-01, 5.19481883244836e-01, 5.13240560868354e-01, 5.06604239605039e-01, 4.99615302737908e-01, 4.92322338705264e-01, 4.84778831382769e-01, 4.77041748686219e-01, 4.69170101160699e-01, 4.61223536815965e-01, 4.53261028658389e-01, 4.45339698671816e-01, 4.37513808026694e-01, 4.29833929530488e-01, 4.22346305948011e-01, 4.15092387612418e-01, 4.08108535113939e-01, 4.01425867842462e-01, 3.95070236558652e-01, 3.89062297593910e-01, 3.83417667270635e-01, 3.78147137217809e-01, 3.73256934006562e-01, 3.68749009579119e-01, 3.64621352013245e-01, 3.60868309040271e-01, 3.57480919274579e-01, 3.54447248226933e-01, 3.51752727805646e-01, 3.49380499138263e-01, 3.47311759170140e-01, 3.45526111628516e-01, 3.44001922622326e-01, 3.42716680423905e-01, 3.41647357930015e-01, 3.40770775010455e-01, 3.40063956541499e-01, 3.39504480510172e-01, 3.39070809302456e-01, 3.38742596286338e-01, 3.38500959194747e-01, 3.38328711697685e-01, 3.38210544994638e-01, 3.38133152272675e-01, 3.38085290430556e-01, 3.38057775488110e-01, 3.38043410459079e-01, 3.38036847015900e-01, 3.38034384851042e-01, 3.38033715083355e-01, 3.38033616227079e-01, 3.38033608575202e-01, 3.38033470820108e-01, 3.38032537509818e-01, 3.38029106571952e-01, 3.38019961024202e-01, 3.37999946586763e-01, 3.37961617962975e-01, 3.37894968601958e-01, 3.37787257426086e-01, 3.37622944421682e-01, 3.37383745030237e-01, 3.37048810773553e-01, 3.36595040323061e-01, 3.35997521137728e-01, 3.35230096765954e-01, 3.34266048976247e-01, 3.33078877243210e-01, 3.31643151139190e-01, 3.29935404413746e-01, 3.27935033671508e-01, 3.25625160341758e-01, 3.22993412799235e-01, 3.20032586622858e-01, 3.16741145373456e-01, 3.13123531873345e-01, 3.09190270317572e-01, 3.04957851790149e-01, 3.00448408773605e-01, 2.95689196767384e-01, 2.90711911969311e-01, 2.85551882149065e-01, 2.80247172741355e-01, 2.74837651620757e-01, 2.69364054211024e-01, 2.63867086087259e-01, 2.58386593816667e-01, 2.52960827295567e-01, 2.47625809078045e-01, 2.42414818814609e-01, 2.37357994409308e-01, 2.32482046148448e-01, 2.27810075968890e-01, 2.23361491204727e-01, 2.19152000471482e-01, 2.15193678660867e-01, 2.11495088149093e-01, 2.08061444086397e-01, 2.04894812865476e-01, 2.01994334401959e-01, 1.99356460557354e-01, 1.96975203765703e-01, 1.94842391574241e-01, 1.92947924275117e-01, 1.91280034012166e-01, 1.89825544629059e-01, 1.88570132048639e-01, 1.87498585124061e-01, 1.86595066694622e-01, 1.85843374055750e-01, 1.85227197276959e-01, 1.84730372864749e-01, 1.84337129271782e-01, 1.84032319810484e-01, 1.83801637754584e-01, 1.83631807903524e-01, 1.83510748725823e-01, 1.83427699438980e-01, 1.83373307044064e-01, 1.83339669389227e-01, 1.83320331732560e-01, 1.83310235922974e-01, 1.83305623107557e-01, 1.83303892688244e-01, 1.83303421972909e-01, 1.83303352496338e-01, 1.83303347074954e-01, 1.83303248911997e-01, 1.83302583844082e-01, 1.83300138991470e-01, 1.83293621968815e-01, 1.83279359912675e-01, 1.83252047474454e-01, 1.83204554374640e-01, 1.83127802187272e-01, 1.83010718922645e-01, 1.82840278606889e-01, 1.82601631300798e-01, 1.82278326732430e-01, 1.81852631831916e-01, 1.81305938898134e-01, 1.80619256921432e-01, 1.79773773866387e-01, 1.78751472733126e-01, 1.77535779338797e-01, 1.76112215456882e-01, 1.74469027738795e-01, 1.72597761225632e-01, 1.70493746648958e-01, 1.68156473363496e-01, 1.65589824669823e-01, 1.62802159222338e-01, 1.59806230684012e-01, 1.56618947084610e-01, 1.53260980651908e-01, 1.49756247395814e-01, 1.46131282710883e-01, 1.42414544184945e-01, 1.38635675358472e-01, 1.34824764319120e-01, 1.31011628919836e-01, 1.27225156432146e-01, 1.23492720077252e-01, 1.19839688655603e-01, 1.16289038966400e-01, 1.12861074373683e-01, 1.09573247152412e-01, 1.06440077452433e-01, 1.03473158050321e-01, 1.00681231606714e-01, 9.80703258937865e-02, 9.56439323020817e-02, 9.34032137078447e-02, 9.13472292701840e-02, 8.94731657005678e-02, 8.77765667829635e-02, 8.62515552149006e-02, 8.48910430216989e-02, 8.36869287348208e-02, 8.26302811224480e-02, 8.17115104648281e-02, 8.09205291421193e-02, 8.02469036620153e-02, 7.96800002175644e-02, 7.92091254879637e-02, 7.88236637425683e-02, 7.85132104634840e-02, 7.82677017588327e-02, 7.80775378944867e-02, 7.79336984255252e-02, 7.78278457470947e-02, 7.77524134806811e-02, 7.77006760166756e-02, 7.76667957662909e-02, 7.76458452286164e-02, 7.76338018081049e-02, 7.76275143642522e-02, 7.76246416493365e-02, 7.76235640061458e-02, 7.76232708621336e-02, 7.76232275947280e-02, 7.76232241897787e-02, 7.76231621693455e-02, 7.76227419724052e-02, 7.76211972938228e-02, 7.76170798234546e-02, 7.76080692237997e-02, 7.75908143171051e-02, 7.75608124884997e-02, 7.75123339055881e-02, 7.74383966492564e-02, 7.73307981529427e-02, 7.71802073350756e-02, 7.69763203607710e-02, 7.67080809682881e-02, 7.63639636844147e-02, 7.59323150305420e-02, 7.54017440927868e-02, 7.47615497970884e-02, 7.40021682150740e-02, 7.31156196273872e-02, 7.20959323505355e-02, 7.09395189432485e-02, 6.96454807353746e-02, 6.82158189122362e-02, 6.66555346883229e-02, 6.49726072271761e-02, 6.31778454829117e-02, 6.12846184158116e-02, 5.93084762975167e-02, 5.72666832564579e-02, 5.51776870710796e-02, 5.30605559224995e-02, 5.09344130345113e-02, 4.88178988151589e-02, 4.67286865050767e-02, 4.46830719089775e-02, 4.26956511810204e-02, 4.07790935756935e-02, 3.89440092701250e-02, 3.71989064179417e-02, 3.55502269311501e-02, 3.40024473243687e-02, 3.25582293090037e-02, 3.12186045598963e-02, 2.99831789592133e-02, 2.88503433687737e-02, 2.78174802858899e-02, 2.68811582938469e-02, 2.60373087638754e-02, 2.52813815961304e-02, 2.46084787868597e-02, 2.40134662229252e-02, 2.34910653296652e-02, 2.30359270522342e-02, 2.26426911595041e-02, 2.23060340637886e-02, 2.20207082888209e-02, 2.17815764426916e-02, 2.15836421093911e-02, 2.14220795075306e-02, 2.12922631224615e-02, 2.11897978429647e-02, 2.11105494691282e-02, 2.10506748526313e-02, 2.10066504253064e-02, 2.09752975089117e-02, 2.09538026021363e-02, 2.09397308303210e-02, 2.09310309121093e-02, 2.09260303320775e-02, 2.09234198713375e-02, 2.09222271971043e-02, 2.09217797946195e-02, 2.09216580913250e-02, 2.09216401282077e-02, 2.09216387020490e-02, 2.09216125663829e-02, 2.09214354939178e-02, 2.09207845655897e-02, 2.09190494967686e-02, 2.09152526820020e-02, 2.09079826336023e-02, 2.08953440451722e-02, 2.08749276962124e-02, 2.08438033672564e-02, 2.07985387774698e-02, 2.07352471966828e-02, 2.06496657165189e-02, 2.05372651037692e-02, 2.03933906416012e-02, 2.02134313897488e-02, 1.99930129327015e-02, 1.97282060944985e-02, 1.94157415248242e-02, 1.90532178264959e-02, 1.86392893548373e-02, 1.81738193279273e-02, 1.76579847199037e-02, 1.70943217107435e-02, 1.64867041861925e-02, 1.58402526583643e-02, 1.51611765290198e-02, 1.44565582241398e-02, 1.37340927044534e-02, 1.30017996008860e-02, 1.22677272853906e-02, 1.15396683467708e-02, 1.08249042026074e-02, 1.01299931624226e-02, 9.46061159908140e-03, 8.82145262031031e-03, 8.21618153062786e-03, 7.64744313494248e-03, 7.11691293335548e-03, 6.62538250441003e-03, 6.17286868173039e-03, 5.75873622605627e-03, 5.38182444347513e-03, 5.04056973754530e-03, 4.73311832026013e-03, 4.45742556271966e-03, 4.21134024944081e-03, 3.99267343757970e-03, 3.79925260763514e-03, 3.62896211650821e-03, 3.47977117905884e-03, 3.34975121621994e-03, 3.23708489023034e-03, 3.14006899778008e-03, 3.05711297046685e-03, 2.98673446313558e-03, 2.92755332992556e-03, 2.87828505045695e-03, 2.83773441600488e-03, 2.80479006943324e-03, 2.77842029015574e-03, 2.75767021136630e-03, 2.74166045807078e-03, 2.72958702045401e-03, 2.72072203365781e-03, 2.71441502739833e-03, 2.71009414407786e-03, 2.70726680433597e-03, 2.70551932814556e-03, 2.70451509085490e-03, 2.70399090102558e-03, 2.70375142001456e-03, 2.70366158660806e-03, 2.70363715013389e-03, 2.70363354336212e-03, 2.70363325441281e-03, 2.70362792572598e-03, 2.70359182347590e-03, 2.70345911104641e-03, 2.70310537395346e-03, 2.70233135829301e-03, 2.70084952128968e-03, 2.69827414446411e-03, 2.69411581055579e-03, 2.68778108771528e-03, 2.67857827631701e-03, 2.66573001320580e-03, 2.64839336036005e-03, 2.62568769160509e-03, 2.59673020740081e-03, 2.56067825162631e-03, 2.51677680407170e-03, 2.46440865217323e-03, 2.40314392708285e-03, 2.33278508493129e-03, 2.25340316943345e-03, 2.16536138421475e-03, 2.06932260219617e-03, 1.96623845536692e-03, 1.85731926291055e-03, 1.74398632586303e-03, 1.62781011920288e-03, 1.51043815192509e-03, 1.39351610062022e-03, 1.27861007787544e-03, 1.16714317276490e-03, 1.06035055915754e-03, 9.59243021796397e-04, 8.64582232480871e-04, 7.76884933728184e-04, 6.96436992583926e-04, 6.23296416833341e-04, 5.57328013286451e-04, 4.98268873073471e-04, 4.45753495370127e-04, 3.99325329453859e-04, 3.58495128011782e-04, 3.22774643848334e-04, 2.91650009785371e-04, 2.64599463974946e-04, 2.41157082854215e-04, 2.20914884977219e-04, 2.03481377212711e-04, 1.88487845449004e-04, 1.75623750450882e-04, 1.64632968392198e-04, 1.55282417892977e-04, 1.47351288021014e-04, 1.40643241635985e-04, 1.34993433789532e-04, 1.30263012891874e-04, 1.26331215923869e-04, 1.23090938203063e-04, 1.20446812857722e-04, 1.18313897374378e-04, 1.16616439072459e-04, 1.15286789664992e-04, 1.14264506919426e-04, 1.13495645422290e-04, 1.12932247049002e-04, 1.12532014341582e-04, 1.12258110178232e-04, 1.12079013874222e-04, 1.11968373488480e-04, 1.11904809173780e-04, 1.11871635340322e-04, 1.11856480813799e-04, 1.11850796272961e-04, 1.11849250004070e-04, 1.11849021777743e-04, 1.11849003289421e-04, 1.11848660409919e-04, 1.11846337318866e-04, 1.11837797831527e-04, 1.11815037769247e-04, 1.11765243531181e-04, 1.11669941460946e-04, 1.11504397096951e-04, 1.11237332461656e-04, 1.10831035954106e-04, 1.10241940305400e-04, 1.09421733952786e-04, 1.08319047186001e-04, 1.06881717471197e-04, 1.05059591462009e-04, 1.02807765529178e-04, 1.00090093596350e-04, 9.68826933132050e-05, 9.31770827037326e-05, 8.89826898600177e-05, 8.43291534522450e-05, 7.92696548899130e-05, 7.38846688932426e-05, 6.82783560392863e-05, 6.25574592305611e-05, 5.68069322213205e-05, 5.11003216132622e-05, 4.55365481002445e-05, 4.02307402680435e-05, 3.52456096379683e-05, 3.05962086883937e-05, 2.63503105770724e-05, 2.25847357739895e-05, 1.92203574158611e-05, 1.62142753481476e-05, 1.38795589350593e-05, 1.22790233803016e-05, 1.08470046882303e-05, 9.16973736764248e-06, 7.18141038191789e-06, 5.05598222844128e-06, 3.09714749846535e-06, 1.78608220835486e-06, 1.20165141481721e-06, 8.53297621006277e-07, 7.92443501040907e-07, 8.29672814006699e-07, 9.81065241489887e-07, 1.09073167209317e-06, 1.00900960831041e-06, 7.54645384404284e-07, 6.55118142844913e-07, 8.58824918736900e-07, 8.80172549955205e-07, 7.71486313055902e-07, 6.12389797538526e-07, 4.80568729699371e-07, 3.99824656938108e-07, 3.57965575403796e-07, 3.56867772666192e-07, 3.67099701747975e-07, 3.61721745889554e-07, 3.35706062060126e-07, -2.84217094304040e-14, -4.26325641456060e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14],
[ 8.00000000000000e+01, 7.99999999088022e+01, 7.99999970887843e+01, 7.99999779827160e+01, 7.99999077462994e+01, 7.99997205180484e+01, 7.99993107545637e+01, 7.99985259355261e+01, 7.99971609003955e+01, 7.99949540280352e+01, 7.99915854132047e+01, 7.99866771325119e+01, 7.99797956298696e+01, 7.99704561908781e+01, 7.99581294199200e+01, 7.99422495857565e+01, 7.99222246627505e+01, 7.98974478658400e+01, 7.98673104566078e+01, 7.98312155819923e+01, 7.97885928916408e+01, 7.97389136592692e+01, 7.96817061030174e+01, 7.96165705572376e+01, 7.95431940947047e+01, 7.94613641397253e+01, 7.93709805593736e+01, 7.92720656856309e+01, 7.91647717198258e+01, 7.90493850143825e+01, 7.89263268218515e+01, 7.87961502461083e+01, 7.86595333154504e+01, 7.85172683046451e+01, 7.83702476406740e+01, 7.82194469121256e+01, 7.80659056453290e+01, 7.79107065983700e+01, 7.77549543523468e+01, 7.75997539511842e+01, 7.74461902673505e+01, 7.72953086653653e+01, 7.71480974137257e+01, 7.70054721730867e+01, 7.68682627752759e+01, 7.67372024109737e+01, 7.66129192666039e+01, 7.64959305927449e+01, 7.63866391444949e+01, 7.62853319048087e+01, 7.61921809807105e+01, 7.61072465457498e+01, 7.60304816871671e+01, 7.59617390010053e+01, 7.59007787617685e+01, 7.58472784749134e+01, 7.58008436008039e+01, 7.57610192185595e+01, 7.57273023786080e+01, 7.56991548750187e+01, 7.56760161542917e+01, 7.56573160676375e+01, 7.56424871702811e+01, 7.56309762751662e+01, 7.56222549805553e+01, 7.56158289119944e+01, 7.56112454490865e+01, 7.56080997461241e+01, 7.56060389020173e+01, 7.56047641877480e+01, 7.56040312969995e+01, 7.56036486455527e+01, 7.56034738051952e+01, 7.56034082159340e+01, 7.56033903740074e+01, 7.56033877405743e+01, 7.56033873951235e+01, 7.56033793465326e+01, 7.56033248161292e+01, 7.56031243563594e+01, 7.56025900023798e+01, 7.56014205724677e+01, 7.55991809386614e+01, 7.55952860852208e+01, 7.55889906674204e+01, 7.55793846618705e+01, 7.55653955605354e+01, 7.55457974015551e+01, 7.55192267483937e+01, 7.54842055229227e+01, 7.54391703674860e+01, 7.53825079577952e+01, 7.53125954177330e+01, 7.52278447078373e+01, 7.51267495853447e+01, 7.50079334851426e+01, 7.48701964743588e+01, 7.47125593207923e+01, 7.45343027212136e+01, 7.43349998885852e+01, 7.41145410104999e+01, 7.38731485505855e+01, 7.36113829223006e+01, 7.33301386410224e+01, 7.30306315600741e+01, 7.27143781346450e+01, 7.23831677914757e+01, 7.20390294336243e+01, 7.16841929676296e+01, 7.13210466349160e+01, 7.09520909794030e+01, 7.05798905364427e+01, 7.02070247197101e+01, 6.98360397449819e+01, 6.94694035624744e+01, 6.91094655433416e+01, 6.87584221017362e+01, 6.84182887007573e+01, 6.80908780204517e+01, 6.77777836379954e+01, 6.74803684392886e+01, 6.71997570840492e+01, 6.69368320617669e+01, 6.66922330907988e+01, 6.64663597607704e+01, 6.62593773818973e+01, 6.60712259988243e+01, 6.59016324779999e+01, 6.57501255110830e+01, 6.56160533087058e+01, 6.54986036967626e+01, 6.53968262729116e+01, 6.53096562327968e+01, 6.52359394316832e+01, 6.51744582066507e+01, 6.51239574476157e+01, 6.50831703741319e+01, 6.50508434522017e+01, 6.50257598748769e+01, 6.50067610359507e+01, 6.49927654506994e+01, 6.49827846236004e+01, 6.49759354309750e+01, 6.49714486757655e+01, 6.49686735796508e+01, 6.49670781004356e+01, 6.49662450948799e+01, 6.49658644828246e+01, 6.49657217013038e+01, 6.49656828611928e+01, 6.49656771284710e+01, 6.49656766223808e+01, 6.49656667057922e+01, 6.49655995194309e+01, 6.49653525350485e+01, 6.49646941637414e+01, 6.49632533259980e+01, 6.49604939177646e+01, 6.49556951848173e+01, 6.49479388909682e+01, 6.49361040188054e+01, 6.49188695721225e+01, 6.48947258530063e+01, 6.48619943596496e+01, 6.48188561905089e+01, 6.47633885465471e+01, 6.46936086006514e+01, 6.46075236624437e+01, 6.45031862247063e+01, 6.43787521576010e+01, 6.42325400459957e+01, 6.40630894714089e+01, 6.38692159475117e+01, 6.36500602426579e+01, 6.34051299690527e+01, 6.31343315786375e+01, 6.28379912645932e+01, 6.25168637052351e+01, 6.21721280865212e+01, 6.18053713862385e+01, 6.14185594823122e+01, 6.10139972359144e+01, 6.05942792573157e+01, 6.01622335322199e+01, 5.97208604051300e+01, 5.92732695314946e+01, 5.88226172994304e+01, 5.83720469039452e+01, 5.79246327910942e+01, 5.74833306606552e+01, 5.70509337099409e+01, 5.66300353846041e+01, 5.62229986075654e+01, 5.58319312826634e+01, 5.54586677879645e+01, 5.51047561470444e+01, 5.47714505606722e+01, 5.44597089743896e+01, 5.41701953419757e+01, 5.39032862239544e+01, 5.36590813418595e+01, 5.34374176995642e+01, 5.32378868849346e+01, 5.30598551759352e+01, 5.29024860893399e+01, 5.27647650205514e+01, 5.26455256238383e+01, 5.25434775699235e+01, 5.24572352914666e+01, 5.23853472884209e+01, 5.23263255186216e+01, 5.22786743499995e+01, 5.22409185062200e+01, 5.22116294042609e+01, 5.21894492670449e+01, 5.21731124022466e+01, 5.21614630737508e+01, 5.21534694568305e+01, 5.21482332615417e+01, 5.21449947283062e+01, 5.21431328402096e+01, 5.21421607514307e+01, 5.21417165925229e+01, 5.21415499726137e+01, 5.21415046479048e+01, 5.21414979580707e+01, 5.21414974214633e+01, 5.21414875183341e+01, 5.21414204231704e+01, 5.21411737741517e+01, 5.21405162976614e+01, 5.21390774225692e+01, 5.21363217897850e+01, 5.21315296747860e+01, 5.21237842181687e+01, 5.21119662169867e+01, 5.20947570644025e+01, 5.20706502323499e+01, 5.20379714666039e+01, 5.19949076022946e+01, 5.19395436098084e+01, 5.18699071500290e+01, 5.17840195635699e+01, 5.16799518570817e+01, 5.15558839031283e+01, 5.14101647657728e+01, 5.12413718318211e+01, 5.10483662971030e+01, 5.08303425536535e+01, 5.05868691651266e+01, 5.03179194118636e+01, 5.00238898290187e+01, 4.97056057325026e+01, 4.93643133947374e+01, 4.90016592461028e+01, 4.86196571746401e+01, 4.82206456031060e+01, 4.78072364689000e+01, 4.73822584696230e+01, 4.69486969540599e+01, 4.65096326696710e+01, 4.60681812932941e+01, 4.56274353521196e+01, 4.51904098486896e+01, 4.47599926610505e+01, 4.43389005853817e+01, 4.39296416924821e+01, 4.35344844537688e+01, 4.31554338479950e+01, 4.27942143998699e+01, 4.24522598538659e+01, 4.21307089810353e+01, 4.18304068759096e+01, 4.15519110331319e+01, 4.12955014941996e+01, 4.10611944081640e+01, 4.08487584360071e+01, 4.06577335261430e+01, 4.04874516807965e+01, 4.03370594074819e+01, 4.02055415992224e+01, 4.00917466088904e+01, 3.99944122780607e+01, 3.99121926526624e+01, 3.98436850718607e+01, 3.97874572595777e+01, 3.97420739871145e+01, 3.97061228180431e+01, 3.96782384003074e+01, 3.96571247421734e+01, 3.96415749040596e+01, 3.96304875615336e+01, 3.96228799480395e+01, 3.96178967690261e+01, 3.96148147894778e+01, 3.96130429295057e+01, 3.96121178508606e+01, 3.96116951729472e+01, 3.96115366115717e+01, 3.96114934790049e+01, 3.96114871127273e+01, 3.96114866216988e+01, 3.96114778044016e+01, 3.96114180659158e+01, 3.96111984609247e+01, 3.96106130746609e+01, 3.96093319712127e+01, 3.96068785051255e+01, 3.96026119078738e+01, 3.95957159547046e+01, 3.95851943939661e+01, 3.95698736767715e+01, 3.95484133547359e+01, 3.95193243123449e+01, 3.94809947642482e+01, 3.94317236751720e+01, 3.93697609543253e+01, 3.92933534458753e+01, 3.92007953973878e+01, 3.90904817605149e+01, 3.89609623894091e+01, 3.88109949823739e+01, 3.86395944913996e+01, 3.84460767292061e+01, 3.82300940533247e+01, 3.79916613090395e+01, 3.77311706601413e+01, 3.74493945040205e+01, 3.71474763142526e+01, 3.68269099238877e+01, 3.64895083918183e+01, 3.61373641184655e+01, 3.57728022417288e+01, 3.53983295169725e+01, 3.50165808621122e+01, 3.46302655581870e+01, 3.42421147905902e+01, 3.38548318631693e+01, 3.34710460800914e+01, 3.30932710129354e+01, 3.27238676684379e+01, 3.23650129328285e+01, 3.20186735605258e+01, 3.16865858642548e+01, 3.13702411287744e+01, 3.10708766100840e+01, 3.07894718138206e+01, 3.05267495978915e+01, 3.02831815403262e+01, 3.00589969676124e+01, 2.98541950504903e+01, 2.96685594302478e+01, 2.95016749195816e+01, 2.93529459083410e+01, 2.92216161801512e+01, 2.91067899008859e+01, 2.90074535696734e+01, 2.89224987273704e+01, 2.88507451991115e+01, 2.87909646114200e+01, 2.87419038764082e+01, 2.87023082824443e+01, 2.86709437792571e+01, 2.86466180026647e+01, 2.86281995563649e+01, 2.86146350609433e+01, 2.86049634974683e+01, 2.85983274169019e+01, 2.85939806570779e+01, 2.85912923039887e+01, 2.85897467492610e+01, 2.85889398249940e+01, 2.85885711333959e+01, 2.85884328242489e+01, 2.85883952007885e+01, 2.85883896476439e+01, 2.85883892271765e+01, 2.85883817786393e+01, 2.85883313137040e+01, 2.85881457992705e+01, 2.85876512859542e+01, 2.85865690559200e+01, 2.85844964569447e+01, 2.85808921930980e+01, 2.85750667530158e+01, 2.85661785527559e+01, 2.85532362486482e+01, 2.85351075322081e+01, 2.85105345496922e+01, 2.84781558899043e+01, 2.84365348544073e+01, 2.83841934670407e+01, 2.83196514019707e+01, 2.82414687241180e+01, 2.81482910606985e+01, 2.80388955801952e+01, 2.79122359704970e+01, 2.77674845065378e+01, 2.76040693019016e+01, 2.74217049642761e+01, 2.72204151268238e+01, 2.70005456985280e+01, 2.67627681429595e+01, 2.65080726180975e+01, 2.62377513396063e+01, 2.59533730119468e+01, 2.56567495580680e+01, 2.53498966391288e+01, 2.50349895860236e+01, 2.47143163848698e+01, 2.43902293056091e+01, 2.40650966742396e+01, 2.37412561895680e+01, 2.34209710790941e+01, 2.31063902624819e+01, 2.27995135248797e+01, 2.25021624817927e+01, 2.22159578429185e+01, 2.19423031720402e+01, 2.16823750240402e+01, 2.14371190528166e+01, 2.12072514553310e+01, 2.09932649664157e+01, 2.07954385523088e+01, 2.06138499617782e+01, 2.04483903667048e+01, 2.02987804387692e+01, 2.01645873441781e+01, 2.00452422750330e+01, 1.99400582587939e+01, 1.98482480858340e+01, 1.97689422634441e+01, 1.97012069410528e+01, 1.96440617574158e+01, 1.95964975402072e+01, 1.95574937477827e+01, 1.95260354890712e+01, 1.95011298984698e+01, 1.94818215862487e+01, 1.94672068388758e+01, 1.94564462144046e+01, 1.94487751706676e+01, 1.94435123815540e+01, 1.94400654399431e+01, 1.94379337133952e+01, 1.94367082067593e+01, 1.94360683889458e+01, 1.94357760525225e+01, 1.94356663872996e+01, 1.94356365557248e+01, 1.94356321526477e+01, 1.94356318221073e+01, 1.94356260042223e+01, 1.94355865872119e+01, 1.94354416863603e+01, 1.94350554357970e+01, 1.94342101455629e+01, 1.94325913507254e+01, 1.94297763690011e+01, 1.94252269219691e+01, 1.94182863028786e+01, 1.94081814873149e+01, 1.93940304764341e+01, 1.93748550306878e+01, 1.93495987905498e+01, 1.93171505876308e+01, 1.92763725262424e+01, 1.92261321680189e+01, 1.91653378918533e+01, 1.90929762444544e+01, 1.90081498639134e+01, 1.89101143732660e+01, 1.87983125272843e+01, 1.86724038758111e+01, 1.85322882980984e+01, 1.83781219742642e+01, 1.82103246912217e+01, 1.80295778181615e+01, 1.78368128048902e+01, 1.76331906169609e+01, 1.74200730769852e+01, 1.71989875793715e+01, 1.69715870351052e+01, 1.67396071422063e+01, 1.65048231404570e+01, 1.62690080915573e+01, 1.60338944469194e+01, 1.58011402643632e+01, 1.55723009656150e+01, 1.53488070471092e+01, 1.51319477204486e+01, 1.49228601061495e+01, 1.47225233573398e+01, 1.45317569528651e+01, 1.43512223604158e+01, 1.41814273083865e+01, 1.40227319946657e+01, 1.38753566768885e+01, 1.37393902118026e+01, 1.36147992272006e+01, 1.35014377101284e+01, 1.33990568762608e+01, 1.33073152470098e+01, 1.32257889043156e+01, 1.31539819199512e+01, 1.30913369682706e+01, 1.30372461301820e+01, 1.29910618831224e+01, 1.29521082484755e+01, 1.29196920360832e+01, 1.28931140876799e+01, 1.28716803802166e+01, 1.28547128096892e+01, 1.28415594400722e+01, 1.28316039741855e+01, 1.28242741873286e+01, 1.28190490631587e+01, 1.28154643863506e+01, 1.28131165785256e+01, 1.28116646118075e+01, 1.28108298957546e+01, 1.28103941047329e+01, 1.28101949895818e+01, 1.28101202948354e+01, 1.28100999760742e+01, 1.28100969770683e+01, 1.28100967526166e+01, 1.28100928111325e+01, 1.28100661070105e+01, 1.28099679400227e+01, 1.28097062643336e+01, 1.28091336008868e+01, 1.28080369102264e+01, 1.28061298446647e+01, 1.28030477577985e+01, 1.27983458010969e+01, 1.27915003786551e+01, 1.27819141582774e+01, 1.27689247456466e+01, 1.27518170149794e+01, 1.27298389526198e+01, 1.27022207107070e+01, 1.26681963914754e+01, 1.26270278983627e+01, 1.25780300117470e+01, 1.25205956921116e+01, 1.24542205008438e+01, 1.23785249770762e+01, 1.22932738325887e+01, 1.21983909336878e+01, 1.20939692279086e+01, 1.19802750329246e+01, 1.18577464143084e+01, 1.17269857100295e+01, 1.15887465823716e+01, 1.14439162639782e+01, 1.12934938920369e+01, 1.11385659802523e+01, 1.09802801585865e+01, 1.08198183194900e+01, 1.06583702544500e+01, 1.04971087553706e+01, 1.03371670006211e+01, 1.01796188550148e+01, 1.00254624981619e+01, 9.87560757197932e+00, 9.73086582443213e+00, 9.59194504243997e+00, 9.45944592852001e+00, 9.33386149217411e+00, 9.21557849818879e+00, 9.10488053160006e+00, 9.00195228955067e+00, 8.90688477867231e+00, 8.81968116979402e+00, 8.74026313015865e+00, 8.66847751134220e+00, 8.60410331641654e+00, 8.54685890241980e+00, 8.49640939483314e+00, 8.45237430076371e+00, 8.41433530813345e+00, 8.38184425050306e+00, 8.35443120237926e+00, 8.33161264930651e+00, 8.31289965241713e+00, 8.29780590047180e+00, 8.28585551619695e+00, 8.27659046055355e+00, 8.26957736111412e+00, 8.26441358139638e+00, 8.26073234872126e+00, 8.25820677010819e+00, 8.25655258916944e+00, 8.25552957119687e+00, 8.25494144698393e+00, 8.25463439590801e+00, 8.25449410236679e+00, 8.25444147351816e+00, 8.25442715720490e+00, 8.25442504414740e+00, 8.25442488594270e+00, 8.25442210699779e+00, 8.25440327924409e+00, 8.25433406653183e+00, 8.25414957163611e+00, 8.25374581304601e+00, 8.25297258228098e+00, 8.25162797516866e+00, 8.24945486058166e+00, 8.24613951494453e+00, 8.24131260852599e+00, 8.23455267763791e+00, 8.22539215276049e+00, 8.21332593415528e+00, 8.19782241275988e+00, 8.17833672599480e+00, 8.15432591896115e+00, 8.12526555758390e+00, 8.09066722057506e+00, 8.05009619291515e+00, 8.00318860730661e+00, 7.94966724373586e+00, 7.88935521053510e+00, 7.82218679850298e+00, 7.74821492219255e+00, 7.66761473223242e+00, 7.58068318606205e+00, 7.48783458362878e+00, 7.38959228932463e+00, 7.28657705285875e+00, 7.17949249514676e+00, 7.06910843201391e+00, 6.95624276640912e+00, 6.84174269217046e+00, 6.72646592572862e+00, 6.61126262412553e+00, 6.49695856544550e+00, 6.38434006717624e+00, 6.27414100432999e+00, 6.16703216760034e+00, 6.06361307834882e+00, 5.96440625857821e+00, 5.86985384733576e+00, 5.78031636665308e+00, 5.69607337498296e+00, 5.61732570659626e+00, 5.54419898136661e+00, 5.47674807821056e+00, 5.41496229274027e+00, 5.35877093993842e+00, 5.30804921010522e+00, 5.26262413552026e+00, 5.22228057157137e+00, 5.18676713594100e+00, 5.15580208034247e+00, 5.12907908991042e+00, 5.10627301530729e+00, 5.08704554243579e+00, 5.07105079560422e+00, 5.05794085394658e+00, 5.04737114017922e+00, 5.03900561801602e+00, 5.03252171251867e+00, 5.02761484900707e+00, 5.02400249328345e+00, 5.02142757071769e+00, 5.01966114540649e+00, 5.01850425355312e+00, 5.01778880696924e+00, 5.01737751190798e+00, 5.01716278336909e+00, 5.01706467316500e+00, 5.01702786880838e+00, 5.01701785714687e+00, 5.01701637944717e+00, 5.01701626851825e+00, 5.01701431607611e+00, 5.01700108800223e+00, 5.01695246033511e+00, 5.01682283784075e+00, 5.01653916732032e+00, 5.01599592402778e+00, 5.01505128100837e+00, 5.01352465224298e+00, 5.01119577806027e+00, 5.00780549367664e+00, 5.00305828677058e+00, 4.99662670540297e+00, 4.98815762145728e+00, 4.97728028611039e+00, 4.96361603312050e+00, 4.94678939527040e+00, 4.92644030358544e+00, 4.90223694450866e+00, 4.87388876535839e+00, 4.84115905247397e+00, 4.80387646887635e+00, 4.76194493725927e+00, 4.71535129544202e+00, 4.66417023718751e+00, 4.60856617920045e+00, 4.54879185808217e+00, 4.48518364749296e+00, 4.41815378073555e+00, 4.34817985045150e+00, 4.27579211815701e+00, 4.20155928707028e+00, 4.12607346128026e+00, 4.04993502744315e+00, 3.97373815274447e+00, 3.89805750170751e+00, 3.82343664636872e+00, 3.75037849431168e+00, 3.67933790295687e+00, 3.61071650109022e+00, 3.54485961172394e+00, 3.48205507190653e+00, 3.42253367863204e+00, 3.36647095521643e+00, 3.31398992600081e+00, 3.26516460359530e+00, 3.22002392576753e+00, 3.17855592218794e+00, 3.14071193890865e+00, 3.10641079605780e+00, 3.07554279831470e+00, 3.04797355595879e+00, 3.02354760526572e+00, 3.00209184013235e+00, 2.98341878198562e+00, 2.96732972260745e+00, 2.95361777513658e+00, 2.94207086304961e+00, 2.93247466645349e+00, 2.92461553078125e+00, 2.91828332638532e+00, 2.91327423010580e+00, 2.90939338326426e+00, 2.90645736628196e+00, 2.90429641969689e+00, 2.90275633594079e+00, 2.90169994662709e+00, 2.90100813660186e+00, 2.90058032838067e+00, 2.90033439806760e+00, 2.90020600519260e+00, 2.90014734254912e+00, 2.90012533634773e+00, 2.90011935014451e+00, 2.90011846659408e+00, 2.90011839998466e+00, 2.90011722384018e+00, 2.90010925529691e+00, 2.90007996221996e+00, 2.90000187863782e+00, 2.89983099973308e+00, 2.89950376574944e+00, 2.89893476431733e+00, 2.89801527106421e+00, 2.89661273671328e+00, 2.89457131377424e+00, 2.89171349557378e+00, 2.88784291283305e+00, 2.88274829662045e+00, 2.87620857035118e+00, 2.86799897778600e+00, 2.85789809045835e+00, 2.84569547015603e+00, 2.83119969530399e+00, 2.81424640110214e+00, 2.79470593967676e+00, 2.77249024579116e+00, 2.74755850205538e+00, 2.71992123883095e+00, 2.68964257841617e+00, 2.65684043681750e+00, 2.62168462158239e+00, 2.58439289955771e+00, 2.54522524080421e+00, 2.50447656083223e+00, 2.46246837116969e+00, 2.41953979986195e+00, 2.37603845526272e+00, 2.33231157977605e+00, 2.28869788086240e+00, 2.24552034367441e+00, 2.20308023383760e+00, 2.16165240081513e+00, 2.12148190124083e+00, 2.08278188444578e+00, 2.04573262324166e+00, 2.01048153327372e+00, 1.97714400311270e+00, 1.94580485233118e+00, 1.91652024282551e+00, 1.88931988601448e+00, 1.86420941183570e+00, 1.84117279166936e+00, 1.82017473400911e+00, 1.80116299703755e+00, 1.78407058495644e+00, 1.76881781414370e+00, 1.75531425050164e+00, 1.74346053054138e+00, 1.73315008584564e+00, 1.72427079374587e+00, 1.71670657664832e+00, 1.71033896886730e+00, 1.70504866359990e+00, 1.70071704444055e+00, 1.69722769632725e+00, 1.69446788085573e+00, 1.69232995137584e+00, 1.69071267507032e+00, 1.68952242312618e+00, 1.68867418681585e+00, 1.68809237726627e+00, 1.68771137009749e+00, 1.68747576283869e+00, 1.68734032265374e+00, 1.68726961374399e+00, 1.68723730696072e+00, 1.68722518768448e+00, 1.68722189095820e+00, 1.68722140436870e+00, 1.68722136748588e+00, 1.68722071358597e+00, 1.68721628332341e+00, 1.68719999729244e+00, 1.68715658534985e+00, 1.68706158260538e+00, 1.68687965336120e+00, 1.68656331560685e+00, 1.68605213366131e+00, 1.68527244042967e+00, 1.68413764264081e+00, 1.68254915124720e+00, 1.68039796371625e+00, 1.67756690420786e+00, 1.67393350102763e+00, 1.66937344838952e+00, 1.66376456247266e+00, 1.65699110216275e+00, 1.64894828595861e+00, 1.63954680241429e+00, 1.62871708677173e+00, 1.61641312557686e+00, 1.60261555766354e+00, 1.58733386593946e+00, 1.57060749972831e+00, 1.55250582935149e+00, 1.53312690812157e+00, 1.51259509509740e+00, 1.49105766699699e+00, 1.46868061196352e+00, 1.44564384509128e+00, 1.42213611160433e+00, 1.39834984686949e+00, 1.37447624428290e+00, 1.35070074607705e+00, 1.32719912345324e+00, 1.30413425710381e+00, 1.28165367300726e+00, 1.25988783642330e+00, 1.23894916310279e+00, 1.21893167321393e+00, 1.19991119128099e+00, 1.18194598419572e+00, 1.16507772780476e+00, 1.14933269880093e+00, 1.13472310048630e+00, 1.12124844627327e+00, 1.10889694160771e+00, 1.09764682175046e+00, 1.08746761834887e+00, 1.07832134117929e+00, 1.07016357239196e+00, 1.06294447887346e+00, 1.05660975398267e+00, 1.05110150307648e+00, 1.04635908816194e+00, 1.04231994597994e+00, 1.03892039115790e+00, 1.03609641210515e+00, 1.03378446242932e+00, 1.03192224522125e+00, 1.03044948200990e+00, 1.02930865297211e+00, 1.02844569053134e+00, 1.02781060520718e+00, 1.02735802081911e+00, 1.02704759614119e+00, 1.02684431193950e+00, 1.02671860593946e+00, 1.02664634343800e+00, 1.02660861763036e+00, 1.02659138079237e+00, 1.02658491472327e+00, 1.02658315580172e+00, 1.02658289618883e+00, 1.02658287638565e+00, 1.02658252364554e+00, 1.02658013378272e+00, 1.02657134844196e+00, 1.02654793031691e+00, 1.02649668218933e+00, 1.02639854304533e+00, 1.02622790062580e+00, 1.02595215717109e+00, 1.02553158192372e+00, 1.02491947966297e+00, 1.02406269852649e+00, 1.02290249194409e+00, 1.02137573808282e+00, 1.01941650543124e+00, 1.01695793509820e+00, 1.01393438967145e+00, 1.01028379631160e+00, 1.00595009002086e+00, 1.00088564411223e+00, 9.95053561483417e-01, 9.88429694950781e-01, 9.81004269682202e-01, 9.72782996781717e-01, 9.63787594101603e-01, 9.54055666679736e-01, 9.43639941696333e-01, 9.32606897307011e-01, 9.21034866429096e-01, 9.09011731012811e-01, 8.96632345887426e-01, 8.83995841746568e-01, 8.71202953781037e-01, 8.58353507168175e-01, 8.45544165794109e-01, 8.32866519838106e-01, 8.20405555085614e-01, 8.08238515655745e-01, 7.96434145071373e-01, 7.85052270082062e-01, 7.74143678131708e-01, 7.63750232645648e-01, 7.53905169505643e-01, 7.44633521855178e-01, 7.35952627266528e-01, 7.27872679891689e-01, 7.20397299324389e-01, 7.13524096607316e-01, 7.07245225524844e-01, 7.01547913678880e-01, 6.96414972732342e-01, 6.91825290641475e-01, 6.87754310805261e-01, 6.84174504002854e-01, 6.81055838950243e-01, 6.78366256462596e-01, 6.76072150720529e-01, 6.74138859162937e-01, 6.72531160206631e-01, 6.71213775475962e-01, 6.70151870663162e-01, 6.69311546703554e-01, 6.68660310808605e-01, 6.68167515233506e-01, 6.67804750626772e-01, 6.67546180552179e-01, 6.67368804377944e-01, 6.67252637220500e-01, 6.67180797964889e-01, 6.67139499442712e-01, 6.67117938447944e-01, 6.67108087178221e-01, 6.67104391647371e-01, 6.67103386375317e-01, 6.67103237999328e-01, 6.67103226601412e-01, 6.67103022530966e-01, 6.67101639924823e-01, 6.67096557336208e-01, 6.67083009158418e-01, 6.67053359999258e-01, 6.66996580981447e-01, 6.66897850142448e-01, 6.66738297263805e-01, 6.66494911243433e-01, 6.66140625315037e-01, 6.65644591730327e-01, 6.64972652639165e-01, 6.64088007568864e-01, 6.62952069969108e-01, 6.61525495812654e-01, 6.59769356552587e-01, 6.57646417481629e-01, 6.55122471683413e-01, 6.52167670523880e-01, 6.48757785346845e-01, 6.44875332975516e-01, 6.40510500745876e-01, 6.35661815540928e-01, 6.30336515383419e-01, 6.24550600544438e-01, 6.18328562105376e-01, 6.11702807273559e-01, 6.04712820157317e-01, 5.97404112052331e-01, 5.89827025055158e-01, 5.82035456291194e-01, 5.74085567405746e-01, 5.66034536186109e-01, 5.57939395803785e-01, 5.49855993974433e-01, 5.41838091074197e-01, 5.33936604345755e-01, 5.26198995738497e-01, 5.18668794106432e-01, 5.11385238417375e-01, 5.04383026954798e-01, 4.97692157662726e-01, 4.91337846171959e-01, 4.85340510085607e-01, 4.79715810330163e-01, 4.74474742475834e-01, 4.69623772713703e-01, 4.65165014565230e-01, 4.61096443396443e-01, 4.57412146446188e-01, 4.54102606411055e-01, 4.51155016703751e-01, 4.48553626347760e-01, 4.46280112111182e-01, 4.44313974925365e-01, 4.42632956894130e-01, 4.41213474297442e-01, 4.40031060966319e-01, 4.39060815312942e-01, 4.38277843215502e-01, 4.37657687977563e-01, 4.37176737807903e-01, 4.36812600802043e-01, 4.36544437346029e-01, 4.36353240274414e-01, 4.36222054040513e-01, 4.36136125596988e-01, 4.36082981599498e-01, 4.36052428852577e-01, 4.36036477499727e-01, 4.36029189179905e-01, 4.36026455077112e-01, 4.36025711334608e-01, 4.36025601559777e-01, 4.36025593062766e-01, 4.36025440092777e-01, 4.36024403698653e-01, 4.36020593806305e-01, 4.36010438047319e-01, 4.35988212494607e-01, 4.35945648198839e-01, 4.35871629458447e-01, 4.35751997980660e-01, 4.35569475017118e-01, 4.35303711103856e-01, 4.34931470094696e-01, 4.34426950630737e-01, 4.33762243948806e-01, 4.32907921960492e-01, 4.31833743894501e-01, 4.30509463673999e-01, 4.28905713922546e-01, 4.26994936529169e-01, 4.24752324625189e-01, 4.22156737257183e-01, 4.19191546561038e-01, 4.15845378287941e-01, 4.12112710302168e-01, 4.07994300050007e-01, 4.03497420548575e-01, 3.98635894435571e-01, 3.93429926141565e-01, 3.87905742337068e-01, 3.82095059612737e-01, 3.76034405249840e-01, 3.69764321609594e-01, 3.63328487097377e-01, 3.56772787075656e-01, 3.50144366889410e-01, 3.43490696760057e-01, 3.36858675084613e-01, 3.30293792944516e-01, 3.23839378580729e-01, 3.17535936353806e-01, 3.11420590369110e-01, 3.05526638607944e-01, 2.99883219194641e-01, 2.94515086514934e-01, 2.89442491468236e-01, 2.84681157358079e-01, 2.80242340930130e-01, 2.76132966918070e-01, 2.72355824138259e-01, 2.68909811592806e-01, 2.65790224045986e-01, 2.62989067937696e-01, 2.60495400080487e-01, 2.58295683148063e-01, 2.56374153327187e-01, 2.54713196535050e-01, 2.53293730213159e-01, 2.52095587870912e-01, 2.51097903285512e-01, 2.50279490645667e-01, 2.49619216062285e-01, 2.49096354898427e-01, 2.48690928442869e-01, 2.48384012714268e-01, 2.48158011768922e-01, 2.47996887894104e-01, 2.47886341564367e-01, 2.47813935035964e-01, 2.47769154923361e-01, 2.47743410966066e-01, 2.47729970346981e-01, 2.47723829227780e-01, 2.47721525483669e-01, 2.47720898809590e-01, 2.47720806313794e-01, 2.47720799096214e-01, 2.47720668409599e-01, 2.47719782988796e-01, 2.47716528092838e-01, 2.47707851770500e-01, 2.47688864029755e-01, 2.47652500918974e-01, 2.47589267306367e-01, 2.47487070732788e-01, 2.47331157057573e-01, 2.47104156639111e-01, 2.46786247385657e-01, 2.46355438065748e-01, 2.45787971694426e-01, 2.45058844576263e-01, 2.44142431725905e-01, 2.43013204042569e-01, 2.41646517032933e-01, 2.40019445405622e-01, 2.38111632940800e-01, 2.35906123148624e-01, 2.33390133856247e-01, 2.30555738421046e-01, 2.27400418057002e-01, 2.23927453906754e-01, 2.20146133917908e-01, 2.16071757995522e-01, 2.11725434818931e-01, 2.07133674480161e-01, 2.02327791975492e-01, 1.97343146765590e-01, 1.92218252350955e-01, 1.86993796415251e-01, 1.81711616060269e-01, 1.76413673685317e-01, 1.71141077104863e-01, 1.65933182756291e-01, 1.60826813788617e-01, 1.55855616110060e-01, 1.51049565905851e-01, 1.46434632559888e-01, 1.42032592103945e-01, 1.37860978914077e-01, 1.33933157804620e-01, 1.30258495136829e-01, 1.26842606040228e-01, 1.23687655136351e-01, 1.20792689925196e-01, 1.18153988834905e-01, 1.15765409429130e-01, 1.13618726012646e-01, 1.11703949542715e-01, 1.10009626069981e-01, 1.08523112718473e-01, 1.07230832350666e-01, 1.06118509504626e-01, 1.05171390937784e-01, 1.04374454216185e-01, 1.03712607330010e-01, 1.03170881409028e-01, 1.02734617385053e-01, 1.02389646053069e-01, 1.02122459565024e-01, 1.01920371105322e-01, 1.01771658473723e-01, 1.01665686652694e-01, 1.01593004234473e-01, 1.01545408859721e-01, 1.01515977562727e-01, 1.01499059067990e-01, 1.01490226547000e-01, 1.01486190999765e-01, 1.01484677143617e-01, 1.01484265339622e-01, 1.01484204558261e-01, 1.01484199775058e-01, 1.01484112649729e-01, 1.01483522363921e-01, 1.01481352425367e-01, 1.01475568277436e-01, 1.01462910377322e-01, 1.01438671153900e-01, 1.01396525730763e-01, 1.01328425754474e-01, 1.01224564986666e-01, 1.01073424364571e-01, 1.00861902838901e-01, 1.00575538276374e-01, 1.00198819911753e-01, 9.97155901549434e-02, 9.91095290072082e-02, 9.83647090486528e-02, 9.74662032118232e-02, 9.64007217831977e-02, 9.51572498366318e-02, 9.37276522382291e-02, 9.21072111224959e-02, 9.02950608944764e-02, 8.82944887640491e-02, 8.61130747048833e-02, 8.37626533711671e-02, 8.12590913631368e-02, 7.86218854514829e-02, 7.58735998380899e-02, 7.30391720132673e-02, 7.01451260658814e-02, 6.72187384115119e-02, 6.42872031751693e-02, 6.13768426972001e-02, 5.85124031107966e-02, 5.57164663924539e-02, 5.30089997642875e-02, 5.04070520423028e-02, 4.79245956672969e-02, 4.55725037360821e-02, 4.33586440839520e-02, 4.12880677248921e-02, 3.93632667636012e-02, 3.75844770022269e-02, 3.59500024200088e-02, 3.44565419542988e-02, 3.30995029934513e-02, 3.18732902023839e-02, 3.07715623495852e-02, 2.97874534112168e-02, 2.89137572392093e-02, 2.81430774265914e-02, 2.74679456950508e-02, 2.68809132179086e-02, 2.63746198563979e-02, 2.59418464152645e-02, 2.55755548018897e-02, 2.52689204764209e-02, 2.50153608778021e-02, 2.48085626613772e-02, 2.46425096476770e-02, 2.45115124106121e-02, 2.44102394908481e-02, 2.43337493600251e-02, 2.42775215430129e-02, 2.42374847788653e-02, 2.42100398019858e-02, 2.41920742780621e-02, 2.41809676283253e-02, 2.41745839010434e-02, 2.41712514533106e-02, 2.41697289293086e-02, 2.41691577939918e-02, 2.41690024328705e-02, 2.41689795019369e-02, 2.41689776813985e-02, 2.41689443177364e-02, 2.41687182749786e-02, 2.41678873306199e-02, 2.41656724361974e-02, 2.41608256920500e-02, 2.41515455056458e-02, 2.41354131548945e-02, 2.41093550150282e-02, 2.40696347046310e-02, 2.40118792533082e-02, 2.39311429525486e-02, 2.38220116769980e-02, 2.36787490205188e-02, 2.34954834978254e-02, 2.32664333078105e-02, 2.29861618451821e-02, 2.26498535092645e-02, 2.22535957580305e-02, 2.17946502571920e-02, 2.12716939063995e-02, 2.06850099934286e-02, 2.00366111259029e-02, 1.93302791009273e-02, 1.85715123972727e-02, 1.77673790878714e-02, 1.69262809570654e-02, 1.60576425305976e-02, 1.51715455854031e-02, 1.42783345776678e-02, 1.33882206546758e-02, 1.25109112128001e-02, 1.16552884775558e-02, 1.08291548697252e-02, 1.00390558354491e-02, 9.29018334510090e-03, 8.58635634635618e-03, 7.93006883164082e-03, 7.32259229907584e-03, 6.76411739200944e-03, 6.25391927944463e-03, 5.79053257135342e-03, 5.37192382586227e-03, 4.99565252071932e-03, 4.65901429157611e-03, 4.35916293334060e-03, 4.09320990934248e-03, 3.85830183847702e-03, 3.65167764968533e-03, 3.47070786496317e-03, 3.31291886719498e-03, 3.17600510017257e-03, 3.05783201416432e-03, 2.95643229306108e-03, 2.86999755645923e-03, 2.79686737289353e-03, 2.73551708114894e-03, 2.68454560288944e-03, 2.64266414056635e-03, 2.60868638599732e-03, 2.58152061809369e-03, 2.56016383866609e-03, 2.54369788937936e-03, 2.53128731199581e-03, 2.52217856437653e-03, 2.51570009071145e-03, 2.51126267686175e-03, 2.50835950055261e-03, 2.50656531656546e-03, 2.50553429403055e-03, 2.50499613933641e-03, 2.50475028211383e-03, 2.50465805747524e-03, 2.50463297058445e-03, 2.50462926783257e-03, 2.50462897119519e-03, 2.50462350059877e-03, 2.50458643685647e-03, 2.50445019050005e-03, 2.50408703750793e-03, 2.50329243699809e-03, 2.50177126039830e-03, 2.49912773075778e-03, 2.49485993796839e-03, 2.48835985550500e-03, 2.47891980637860e-03, 2.46574626511631e-03, 2.44798169573812e-03, 2.42473476612304e-03, 2.39511872530329e-03, 2.35829697928125e-03, 2.31353399325940e-03, 2.26024867137653e-03, 2.19806644568621e-03, 2.12686561387443e-03, 2.04681316969868e-03, 1.95838562460438e-03, 1.86237120719568e-03, 1.75985133918743e-03, 1.65216129136070e-03, 1.54083217516643e-03, 1.42751859284829e-03, 1.31391799071423e-03, 1.20168872321358e-03, 1.09237385653671e-03, 9.87336827620311e-04, 8.87713431012571e-04, 7.94382566652985e-04, 7.07956095823192e-04, 6.28786300367778e-04, 5.56987966845668e-04, 4.92471082480961e-04, 4.34979623733511e-04, 3.84132071445720e-04, 3.39460093101707e-04, 3.00443056843847e-04, 2.66537308135639e-04, 2.37200010047900e-04, 2.11907632063912e-04, 1.90169149732355e-04, 1.71534134906892e-04, 1.55596335645214e-04, 1.41993866250800e-04, 1.30407325386273e-04, 1.20556906111347e-04, 1.12199190681395e-04, 1.05123939405871e-04, 9.91507284311410e-05, 9.41251961990019e-05, 8.99151060309578e-05, 8.64067450692119e-05, 8.35019242941826e-05, 8.11154789488455e-05, 7.91731330167522e-05, 7.76097167118905e-05, 7.63677502675364e-05, 7.53963500983446e-05, 7.46503868640502e-05, 7.40898453841510e-05, 7.36793565891958e-05, 7.33878723622183e-05, 7.31884450573261e-05, 7.30580679686454e-05, 7.29775326391291e-05, 7.29312665868065e-05, 7.29071213214638e-05, 7.28960913818355e-05, 7.28919540335937e-05, 7.28908286049545e-05, 7.28906624942738e-05, 7.28906490792269e-05, 7.28903999061004e-05, 7.28887118555122e-05, 7.28825066005356e-05, 7.28659679936072e-05, 7.28297852674586e-05, 7.27605354882144e-05, 7.26402486463940e-05, 7.24462061469922e-05, 7.21510301247008e-05, 7.17231250604300e-05, 7.11275298073133e-05, 7.03272252593479e-05, 6.92849156536113e-05, 6.79652583102097e-05, 6.63374572127395e-05, 6.43780673073024e-05, 6.20737840364427e-05, 5.94239254070317e-05, 5.64422518749552e-05, 5.31577234426095e-05, 4.96138264196588e-05, 4.58663692057826e-05, 4.19801936999420e-05, 3.80256592080741e-05, 3.40753268943672e-05, 3.02003196708256e-05, 2.64661654512111e-05, 2.29295097966542e-05, 1.96356576793733e-05, 1.66126693983415e-05, 1.38625811274551e-05, 1.13647239743386e-05, 9.09964175832556e-06, 7.06817425566442e-06, 5.27349297385626e-06, 3.68425506280801e-06, 2.28690771564288e-06, 1.61627545480769e-06, 1.94024489985622e-06, 1.98183835209420e-06, 1.64896302123907e-06, 1.38611525812848e-06, 1.18732864962112e-06, 2.84217094304040e-14, 1.66046302751965e-06, 2.65229402884870e-06, 3.13548257224738e-06, 3.06614803946559e-06, 2.52154512736524e-06, 1.66491477671116e-06, 6.54494584750864e-07, 6.49620247372695e-07, 9.01219351590044e-07, 8.33648414300114e-07, 6.45327276060925e-07, 4.71970636795049e-07, 1.42108547152020e-14, 1.42108547152020e-14, 3.97331248791488e-07, 4.43888609424903e-07, 4.60021780668285e-07, 4.39778403915625e-07, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 0.00000000000000e+00, 0.00000000000000e+00, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 0.00000000000000e+00],
[ 8.10000000000000e+01, 8.09999999072056e+01, 8.09999970378178e+01, 8.09999775972602e+01, 8.09999061312162e+01, 8.09997156251582e+01, 8.09992986879083e+01, 8.09985001288547e+01, 8.09971111953491e+01, 8.09948656851312e+01, 8.09914380903201e+01, 8.09864438670859e+01, 8.09794418614882e+01, 8.09699388603591e+01, 8.09573961796289e+01, 8.09412381538661e+01, 8.09208623517331e+01, 8.08956513128628e+01, 8.08649855808072e+01, 8.08282577908065e+01, 8.07848875552762e+01, 8.07343368685517e+01, 8.06761257207891e+01, 8.06098475665104e+01, 8.05351842375163e+01, 8.04519198289034e+01, 8.03599530315259e+01, 8.02593073485202e+01, 8.01501386322739e+01, 8.00327394237333e+01, 7.99075396746683e+01, 7.97751035837405e+01, 7.96361224683465e+01, 7.94914038079467e+01, 7.93418568079268e+01, 7.91884750223525e+01, 7.90323167191730e+01, 7.88744837593550e+01, 7.87160997875591e+01, 7.85582885004106e+01, 7.84021526801485e+01, 7.82487545716123e+01, 7.80990980553748e+01, 7.79541129440256e+01, 7.78146416133487e+01, 7.76814280823832e+01, 7.75551095788568e+01, 7.74362105685890e+01, 7.73251391862274e+01, 7.72221859759563e+01, 7.71275248303152e+01, 7.70412159991301e+01, 7.69632110258416e+01, 7.68933594532437e+01, 7.68314171237808e+01, 7.67770558808807e+01, 7.67298744576982e+01, 7.66894103189703e+01, 7.66551522015844e+01, 7.66265530812968e+01, 7.66030432782645e+01, 7.65840434041352e+01, 7.65689768497694e+01, 7.65572815164738e+01, 7.65484205058426e+01, 7.65418915045392e+01, 7.65372346307317e+01, 7.65340385480722e+01, 7.65319447002009e+01, 7.65306495723735e+01, 7.65299049451524e+01, 7.65295161660050e+01, 7.65293385258037e+01, 7.65292718862170e+01, 7.65292537585757e+01, 7.65292510829717e+01, 7.65292507319889e+01, 7.65292425545105e+01, 7.65291871508759e+01, 7.65289834810191e+01, 7.65284405701508e+01, 7.65272524138227e+01, 7.65249769171703e+01, 7.65210196994407e+01, 7.65146234883873e+01, 7.65048637060891e+01, 7.64906507066760e+01, 7.64707389648552e+01, 7.64437433294292e+01, 7.64081622462317e+01, 7.63624076197406e+01, 7.63048407240821e+01, 7.62338132975223e+01, 7.61477126692698e+01, 7.60450094880595e+01, 7.59243063693441e+01, 7.57843855794338e+01, 7.56242537637053e+01, 7.54431817367608e+01, 7.52407375138753e+01, 7.50168110870104e+01, 7.47716299187939e+01, 7.45057646930880e+01, 7.42201254386336e+01, 7.39159486355702e+01, 7.35947762397350e+01, 7.32584276782758e+01, 7.29089658120853e+01, 7.25486577238636e+01, 7.21799311112119e+01, 7.18053271594164e+01, 7.14274510735304e+01, 7.10489218780651e+01, 7.06723234561195e+01, 7.03001588888481e+01, 6.99348098539035e+01, 6.95785021988436e+01, 6.92332780239918e+01, 6.89009739369190e+01, 6.85832047496636e+01, 6.82813518121748e+01, 6.79965553242025e+01, 6.77297102083344e+01, 6.74814653455106e+01, 6.72522261114786e+01, 6.70421601997359e+01, 6.68512066951171e+01, 6.66790883024998e+01, 6.65253265623268e+01, 6.63892598141015e+01, 6.62700636069989e+01, 6.61667732035811e+01, 6.60783077760712e+01, 6.60034958522118e+01, 6.59411015281391e+01, 6.58898509292860e+01, 6.58484583691056e+01, 6.58156516325558e+01, 6.57901958006494e+01, 6.57709150378045e+01, 6.57567117885240e+01, 6.57465828763360e+01, 6.57396320667450e+01, 6.57350787463620e+01, 6.57322624798635e+01, 6.57306433309170e+01, 6.57297979673821e+01, 6.57294117088079e+01, 6.57292668090738e+01, 6.57292273927565e+01, 6.57292215749879e+01, 6.57292210613897e+01, 6.57292109976851e+01, 6.57291428145917e+01, 6.57288921661162e+01, 6.57282240276802e+01, 6.57267618148034e+01, 6.57239614705920e+01, 6.57190915495107e+01, 6.57112201963857e+01, 6.56992097706983e+01, 6.56817196943040e+01, 6.56572179017488e+01, 6.56240010422209e+01, 6.55802233180034e+01, 6.55239335461871e+01, 6.54531197030709e+01, 6.53657598647871e+01, 6.52598781104579e+01, 6.51336036288894e+01, 6.49852309938060e+01, 6.48132793739831e+01, 6.46165483477167e+01, 6.43941680121120e+01, 6.41456412218657e+01, 6.38708760533798e+01, 6.35702069536839e+01, 6.32444034818906e+01, 6.28946660670257e+01, 6.25226087750738e+01, 6.21302296822703e+01, 6.17198700627111e+01, 6.12941641710673e+01, 6.08559818747260e+01, 6.04083666988521e+01, 5.99544719420286e+01, 5.94974973819990e+01, 5.90406287468617e+01, 5.85869816427979e+01, 5.81395510934001e+01, 5.77011673446686e+01, 5.72744581868783e+01, 5.68618177672212e+01, 5.64653817076602e+01, 5.60870082691345e+01, 5.57282652766077e+01, 5.53904225062860e+01, 5.50744492179300e+01, 5.47810164877554e+01, 5.45105039672385e+01, 5.42630106695397e+01, 5.40383693749768e+01, 5.38361642512946e+01, 5.36557512994502e+01, 5.34962812540739e+01, 5.33567245818568e+01, 5.32358982244654e+01, 5.31324937215006e+01, 5.30451063228238e+01, 5.29722646604275e+01, 5.29124605023355e+01, 5.28641780607931e+01, 5.28259222811592e+01, 5.27962455035580e+01, 5.27737718731886e+01, 5.27572188828284e+01, 5.27454154665080e+01, 5.27373161284955e+01, 5.27320106861910e+01, 5.27287293264743e+01, 5.27268428174250e+01, 5.27258578742960e+01, 5.27254078421309e+01, 5.27252390189604e+01, 5.27251930949109e+01, 5.27251863166155e+01, 5.27251857729124e+01, 5.27251757388316e+01, 5.27251077564517e+01, 5.27248578459371e+01, 5.27241916755314e+01, 5.27227337741963e+01, 5.27199417045647e+01, 5.27150862277214e+01, 5.27072383672609e+01, 5.26952641361105e+01, 5.26778275223755e+01, 5.26534021351850e+01, 5.26202916831925e+01, 5.25766591933251e+01, 5.25205645748476e+01, 5.24500097976599e+01, 5.23629905936993e+01, 5.22575532226458e+01, 5.21318544904773e+01, 5.19842228995729e+01, 5.18132185726367e+01, 5.16176894600105e+01, 5.13968213373026e+01, 5.11501792464961e+01, 5.08777383369944e+01, 5.05799025185007e+01, 5.02575099252773e+01, 4.99118248747156e+01, 4.95445167288021e+01, 4.91576267667863e+01, 4.87535247758432e+01, 4.83348574955859e+01, 4.79044912693737e+01, 4.74654512589584e+01, 4.70208594117166e+01, 4.65738731038666e+01, 4.61276260935780e+01, 4.56851731551417e+01, 4.52494395420408e+01, 4.48231762239140e+01, 4.44089216291777e+01, 4.40089703799769e+01, 4.36253492292780e+01, 4.32598001226021e+01, 4.29137700419245e+01, 4.25884070781893e+01, 4.22845620422189e+01, 4.20027948667249e+01, 4.17433850657269e+01, 4.15063455835025e+01, 4.12914394609320e+01, 4.10981988514083e+01, 4.09259460141416e+01, 4.07738159878821e+01, 4.06407806964185e+01, 4.05256742568123e+01, 4.04272192538053e+01, 4.03440537131935e+01, 4.02747584588435e+01, 4.02178844791047e+01, 4.01719798659390e+01, 4.01356158317014e+01, 4.01074112615634e+01, 4.00860552309220e+01, 4.00703269125331e+01, 4.00591123217223e+01, 4.00514174020153e+01, 4.00463770376750e+01, 4.00432596913073e+01, 4.00414674989406e+01, 4.00405318049574e+01, 4.00401042768092e+01, 4.00399438959433e+01, 4.00399002684310e+01, 4.00398938291006e+01, 4.00398933324375e+01, 4.00398844139620e+01, 4.00398239899784e+01, 4.00396018650273e+01, 4.00390097614934e+01, 4.00377139575575e+01, 4.00352323386343e+01, 4.00309167844789e+01, 4.00239417069193e+01, 4.00132994284493e+01, 3.99978029460335e+01, 3.99760964523838e+01, 3.99466737833420e+01, 3.99079047207283e+01, 3.98580688038211e+01, 3.97953959926593e+01, 3.97181131915135e+01, 3.96244952965536e+01, 3.95129190998453e+01, 3.93819180896729e+01, 3.92302359654906e+01, 3.90568765654592e+01, 3.88611479127276e+01, 3.86426982425794e+01, 3.84015421833061e+01, 3.81380757205832e+01, 3.78530791520523e+01, 3.75477078922081e+01, 3.72234716597394e+01, 3.68822032043921e+01, 3.65260182437814e+01, 3.61572686304250e+01, 3.57784909268830e+01, 3.53923525338904e+01, 3.50015973256066e+01, 3.46089924543472e+01, 3.42172776609195e+01, 3.38291181245840e+01, 3.34470616443096e+01, 3.30735007656294e+01, 3.27106403336440e+01, 3.23604708278346e+01, 3.20247476877535e+01, 3.17049766580398e+01, 3.14024049773801e+01, 3.11180180371783e+01, 3.08525409746958e+01, 3.06064445667631e+01, 3.03799547621859e+01, 3.01730652253762e+01, 2.99855523409964e+01, 2.98169922261384e+01, 2.96667793918860e+01, 2.95341467750916e+01, 2.94181869155410e+01, 2.93178740806899e+01, 2.92320871410763e+01, 2.91596329780457e+01, 2.90992701666987e+01, 2.90497326269835e+01, 2.90097528811305e+01, 2.89780845029885e+01, 2.89535233010633e+01, 2.89349267485626e+01, 2.89212311659746e+01, 2.89114661787052e+01, 2.89047660162725e+01, 2.89003772905345e+01, 2.88976629862041e+01, 2.88961025131337e+01, 2.88952878004502e+01, 2.88949155503294e+01, 2.88947759062689e+01, 2.88947379196807e+01, 2.88947323129392e+01, 2.88947318884137e+01, 2.88947243679860e+01, 2.88946734159828e+01, 2.88944861110444e+01, 2.88939868249630e+01, 2.88928941501706e+01, 2.88908015494664e+01, 2.88871625063014e+01, 2.88812808639447e+01, 2.88723069371615e+01, 2.88592398575341e+01, 2.88409364689573e+01, 2.88161269186422e+01, 2.87834368877253e+01, 2.87414161732550e+01, 2.86885730726431e+01, 2.86234137401621e+01, 2.85444853957702e+01, 2.84504219879014e+01, 2.83399906668579e+01, 2.82121372398511e+01, 2.80660286783777e+01, 2.79010907559950e+01, 2.77170390249632e+01, 2.75139015979201e+01, 2.72920325758185e+01, 2.70521154301038e+01, 2.67951561652803e+01, 2.65224666074375e+01, 2.62356386334115e+01, 2.59365105316857e+01, 2.56271269465698e+01, 2.53096940021444e+01, 2.49865312531748e+01, 2.46600220981359e+01, 2.43325642423576e+01, 2.40065217300745e+01, 2.36841799680929e+01, 2.33677050242675e+01, 2.30591082839886e+01, 2.27602172812234e+01, 2.24726531980967e+01, 2.21978151754483e+01, 2.19368712327165e+01, 2.16907552956752e+01, 2.14601696039303e+01, 2.12455916324816e+01, 2.10472846146525e+01, 2.08653107865556e+01, 2.06995465670522e+01, 2.05496990192148e+01, 2.04153230872548e+01, 2.02958392474645e+01, 2.01905513383165e+01, 2.00986644339078e+01, 2.00193026915750e+01, 1.99515271379111e+01, 1.98943533598759e+01, 1.98467690438374e+01, 1.98077512615007e+01, 1.97762833451008e+01, 1.97513711328691e+01, 1.97320583075645e+01, 1.97174405032850e+01, 1.97066778253440e+01, 1.96990054196904e+01, 1.96937417452413e+01, 1.96902942453667e+01, 1.96881621821170e+01, 1.96869364848719e+01, 1.96862965684019e+01, 1.96860041870977e+01, 1.96858945050702e+01, 1.96858646689270e+01, 1.96858602651758e+01, 1.96858599345848e+01, 1.96858541158091e+01, 1.96858146927651e+01, 1.96856697697525e+01, 1.96852834602640e+01, 1.96844380418226e+01, 1.96828190043234e+01, 1.96800036095866e+01, 1.96754535191045e+01, 1.96685119755464e+01, 1.96584059373888e+01, 1.96442534596829e+01, 1.96250764818411e+01, 1.95998190208091e+01, 1.95673705728462e+01, 1.95265943008446e+01, 1.94763593328494e+01, 1.94155762327525e+01, 1.93432344428526e+01, 1.92584402614830e+01, 1.91604537315848e+01, 1.90487227030240e+01, 1.89229123155928e+01, 1.87829282487211e+01, 1.86289323072892e+01, 1.84613492588034e+01, 1.82808642907820e+01, 1.80884109900370e+01, 1.78851503166255e+01, 1.76724416041450e+01, 1.74518071102654e+01, 1.72248920153998e+01, 1.69934219825803e+01, 1.67591604253993e+01, 1.65238674836012e+01, 1.62892624016602e+01, 1.60569905885725e+01, 1.58285961637133e+01, 1.56055003229241e+01, 1.53889854424920e+01, 1.51801845126706e+01, 1.49800752743625e+01, 1.47894783216608e+01, 1.46090584139550e+01, 1.44393282906688e+01, 1.42806543735533e+01, 1.41332638525518e+01, 1.39972527638320e+01, 1.38725947712443e+01, 1.37591504494734e+01, 1.36566769368213e+01, 1.35648378785438e+01, 1.34832136194941e+01, 1.34113116290068e+01, 1.33485771526242e+01, 1.32944040851959e+01, 1.32481460488091e+01, 1.32091276378832e+01, 1.31766557641233e+01, 1.31500309980722e+01, 1.31285587646693e+01, 1.31115602110516e+01, 1.30983825297251e+01, 1.30884084931417e+01, 1.30810649402490e+01, 1.30758299546356e+01, 1.30722384892710e+01, 1.30698862250366e+01, 1.30684314982815e+01, 1.30675951941246e+01, 1.30671585735760e+01, 1.30669590793174e+01, 1.30668842423401e+01, 1.30668638848873e+01, 1.30668608801706e+01, 1.30668606552915e+01, 1.30668567063018e+01, 1.30668299513277e+01, 1.30667315973940e+01, 1.30664694233095e+01, 1.30658956688015e+01, 1.30647968873444e+01, 1.30628861818541e+01, 1.30597982011233e+01, 1.30550872261895e+01, 1.30482286168309e+01, 1.30386238152236e+01, 1.30256090123744e+01, 1.30084674699015e+01, 1.29864453534842e+01, 1.29587707761029e+01, 1.29246755740887e+01, 1.28834191563227e+01, 1.28343135902484e+01, 1.27767489346289e+01, 1.27102177168490e+01, 1.26343373999000e+01, 1.25488697051814e+01, 1.24537357595222e+01, 1.23490262172260e+01, 1.22350057596513e+01, 1.21121116762049e+01, 1.19809465557462e+01, 1.18422654380228e+01, 1.16969580645694e+01, 1.15460271066631e+01, 1.13905634209525e+01, 1.12317194848885e+01, 1.10706821930631e+01, 1.09086461541931e+01, 1.07467885209767e+01, 1.05862462182770e+01, 1.04280962204515e+01, 1.02733392838793e+01, 1.01228872895378e+01, 9.97755411933503e+00, 9.83804980266966e+00, 9.70497754210079e+00, 9.57883316345337e+00, 9.46000652945314e+00, 9.34878449265722e+00, 9.24535502567842e+00, 9.14981223861194e+00, 9.06216206341034e+00, 8.98232844581428e+00, 8.91015993452788e+00, 8.84543659448074e+00, 8.78787719733751e+00, 8.73714665922643e+00, 8.69286370400265e+00, 8.65460873094040e+00, 8.62193185907354e+00, 8.59436110708299e+00, 8.57141064859988e+00, 8.55258905949046e+00, 8.53740744806156e+00, 8.52538733359518e+00, 8.51606811587267e+00, 8.50901396112235e+00, 8.50381992061706e+00, 8.50011709894673e+00, 8.49757670102353e+00, 8.49591281044756e+00, 8.49488378625919e+00, 8.49429220865791e+00, 8.49398335448308e+00, 8.49384223706390e+00, 8.49378929914580e+00, 8.49377489875775e+00, 8.49377277329098e+00, 8.49377261415718e+00, 8.49376981889247e+00, 8.49375088056944e+00, 8.49368126139015e+00, 8.49349568298423e+00, 8.49308955306523e+00, 8.49231178056417e+00, 8.49095927421799e+00, 8.48877338937817e+00, 8.48543855231550e+00, 8.48058324868252e+00, 8.47378351060289e+00, 8.46456897249838e+00, 8.45243148713327e+00, 8.43683619952881e+00, 8.41723486831621e+00, 8.39308110516440e+00, 8.36384707932514e+00, 8.32904111501587e+00, 8.28822550543568e+00, 8.24103379089075e+00, 8.18718671153141e+00, 8.12650605691631e+00, 8.05892570012905e+00, 7.98449922286522e+00, 7.90340370284481e+00, 7.81593943328083e+00, 7.72252555895797e+00, 7.62369182608167e+00, 7.52006683602117e+00, 7.41236335263956e+00, 7.30136133049901e+00, 7.18788940351735e+00, 7.07280560156481e+00, 6.95697804957889e+00, 6.84126635476332e+00, 6.72650430739073e+00, 6.61348441503608e+00, 6.50294466492781e+00, 6.39555777199736e+00, 6.29192302994447e+00, 6.19256074876111e+00, 6.09790914386372e+00, 6.00832344685304e+00, 5.92407694089674e+00, 5.84536358657680e+00, 5.77230189539347e+00, 5.70493972402258e+00, 5.64325969720332e+00, 5.58718501432560e+00, 5.53658544797482e+00, 5.49128339615447e+00, 5.45105989898461e+00, 5.41566057195315e+00, 5.38480143908971e+00, 5.35817466968591e+00, 5.33545423132873e+00, 5.31630147079461e+00, 5.30037062420182e+00, 5.28731424070710e+00, 5.27678848234414e+00, 5.26845823898283e+00, 5.26200197459471e+00, 5.25711620170601e+00, 5.25351946748033e+00, 5.25095572918275e+00, 5.24919700003377e+00, 5.24804515907684e+00, 5.24733284020105e+00, 5.24692334462007e+00, 5.24670955596611e+00, 5.24661187529318e+00, 5.24657523208307e+00, 5.24656526425871e+00, 5.24656379302934e+00, 5.24656368258611e+00, 5.24656173869305e+00, 5.24654856854093e+00, 5.24650015380909e+00, 5.24637109898869e+00, 5.24608867114736e+00, 5.24554780903092e+00, 5.24460731095408e+00, 5.24308739240595e+00, 5.24076878230358e+00, 5.23739349962383e+00, 5.23266741747277e+00, 5.22626467713955e+00, 5.21783395833954e+00, 5.20700654263487e+00, 5.19340602537318e+00, 5.17665943977467e+00, 5.15640945959275e+00, 5.13232725080813e+00, 5.10412545665686e+00, 5.07157073361293e+00, 5.03449521847811e+00, 4.99280630699437e+00, 4.94649416831707e+00, 4.89563650939185e+00, 4.84040023618083e+00, 4.78103982716294e+00, 4.71789242626211e+00, 4.65136986137030e+00, 4.58194798302405e+00, 4.51015387787626e+00, 4.43655162819385e+00, 4.36172735111508e+00, 4.28627425528498e+00, 4.21077839989583e+00, 4.13580574048193e+00, 4.06189091031095e+00, 3.98952803191706e+00, 3.91916369668611e+00, 3.85119210607783e+00, 3.78595224722677e+00, 3.72372688511692e+00, 3.66474309563249e+00, 3.60917403705488e+00, 3.55714165768127e+00, 3.50872005821316e+00, 3.46393926289764e+00, 3.42278919692041e+00, 3.38522371395743e+00, 3.35116456296032e+00, 3.32050522422745e+00, 3.29311457971478e+00, 3.26884041033817e+00, 3.24751273333548e+00, 3.22894700566806e+00, 3.21294722535013e+00, 3.19930896212929e+00, 3.18782234292023e+00, 3.17827500680859e+00, 3.17045503044923e+00, 3.16415380860072e+00, 3.15916885781647e+00, 3.15530649547323e+00, 3.15238433287017e+00, 3.15023351145929e+00, 3.14870060650250e+00, 3.14764912334303e+00, 3.14696051831844e+00, 3.14653468888457e+00, 3.14628989500812e+00, 3.14616209511614e+00, 3.14610370333398e+00, 3.14608179872950e+00, 3.14607584016190e+00, 3.14607496069037e+00, 3.14607489438848e+00, 3.14607372367364e+00, 3.14606579191664e+00, 3.14603663406209e+00, 3.14595891087325e+00, 3.14578882037647e+00, 3.14546309512305e+00, 3.14489671373813e+00, 3.14398144538070e+00, 3.14258533386185e+00, 3.14055321289118e+00, 3.13770832349752e+00, 3.13385507723406e+00, 3.12878297273986e+00, 3.12227162769025e+00, 3.11409683336872e+00, 3.10403747677380e+00, 3.09188310881910e+00, 3.07744187200650e+00, 3.06054844359679e+00, 3.04107160821374e+00, 3.01892105428574e+00, 2.99405299775111e+00, 2.96647427754915e+00, 2.93624464064344e+00, 2.90347703572441e+00, 2.86833585640856e+00, 2.83103320561032e+00, 2.79182337993393e+00, 2.75099588357890e+00, 2.70886736444434e+00, 2.66577291345014e+00, 2.62205717865372e+00, 2.57806572029533e+00, 2.53413697727777e+00, 2.49059513842761e+00, 2.44774412311922e+00, 2.40586278514944e+00, 2.36520136934794e+00, 2.32597917830621e+00, 2.28838335037049e+00, 2.25256861095312e+00, 2.21865783670781e+00, 2.18674326432333e+00, 2.15688817999349e+00, 2.12912893907993e+00, 2.10347718518892e+00, 2.07992216117169e+00, 2.05843302913838e+00, 2.03896114059187e+00, 2.02144221981372e+00, 2.00579844265873e+00, 1.99194040829302e+00, 1.97976901280714e+00, 1.96917724097524e+00, 1.96005189583462e+00, 1.95227528552145e+00, 1.94572688334176e+00, 1.94028497093576e+00, 1.93582826626638e+00, 1.93223752879517e+00, 1.92939712443626e+00, 1.92719652358943e+00, 1.92553169762199e+00, 1.92430637340046e+00, 1.92343310253014e+00, 1.92283410228437e+00, 1.92244182897242e+00, 1.92219925156924e+00, 1.92205980338539e+00, 1.92198700169011e+00, 1.92195373863902e+00, 1.92194126062449e+00, 1.92193786631184e+00, 1.92193736531875e+00, 1.92193732734418e+00, 1.92193665408810e+00, 1.92193209268432e+00, 1.92191532455894e+00, 1.92187062748809e+00, 1.92177281206081e+00, 1.92158549538392e+00, 1.92125978630875e+00, 1.92073345102538e+00, 1.91993062127749e+00, 1.91876209886421e+00, 1.91712629844520e+00, 1.91491085481070e+00, 1.91199489974051e+00, 1.90825198679860e+00, 1.90355361001919e+00, 1.89777322548802e+00, 1.89079064540763e+00, 1.88249663552021e+00, 1.87279751277036e+00, 1.86161951531756e+00, 1.84891270583863e+00, 1.83465417506287e+00, 1.81885033769964e+00, 1.80153815729028e+00, 1.78278519757453e+00, 1.76268847090324e+00, 1.74137213239646e+00, 1.71898414426006e+00, 1.69569210029798e+00, 1.67167844966301e+00, 1.64713538678477e+00, 1.62225967930920e+00, 1.59724768864808e+00, 1.57229080172849e+00, 1.54757144295928e+00, 1.52325977852669e+00, 1.49951116721510e+00, 1.47646435862499e+00, 1.45424039518308e+00, 1.43294214128260e+00, 1.41265434211438e+00, 1.39344410561385e+00, 1.37536170172741e+00, 1.35844158153695e+00, 1.34270353212891e+00, 1.32815389908016e+00, 1.31478682506838e+00, 1.30258546891223e+00, 1.29152318332135e+00, 1.28156464123732e+00, 1.27266690966115e+00, 1.26478047630989e+00, 1.25785023848158e+00, 1.25181646536436e+00, 1.24661574496730e+00, 1.24218192515085e+00, 1.23844705517963e+00, 1.23534233011124e+00, 1.23279903549481e+00, 1.23074948464523e+00, 1.22912793557008e+00, 1.22787146987547e+00, 1.22692081207555e+00, 1.22622106506843e+00, 1.22572233643264e+00, 1.22538023085913e+00, 1.22515618653827e+00, 1.22501763758883e+00, 1.22493799040845e+00, 1.22489640876785e+00, 1.22487741008457e+00, 1.22487028307300e+00, 1.22486834435780e+00, 1.22486805820768e+00, 1.22486803638029e+00, 1.22486764758365e+00, 1.22486501343182e+00, 1.22485533005104e+00, 1.22482951802000e+00, 1.22477303057866e+00, 1.22466485633731e+00, 1.22447675909171e+00, 1.22417279404162e+00, 1.22370913571632e+00, 1.22303424777201e+00, 1.22208941795508e+00, 1.22080967229238e+00, 1.21912507043250e+00, 1.21696236872195e+00, 1.21424701914400e+00, 1.21090545130247e+00, 1.20686756242766e+00, 1.20206931876896e+00, 1.19645535306128e+00, 1.18998142961071e+00, 1.18261664343476e+00, 1.17434522477494e+00, 1.16516783616748e+00, 1.15510227580442e+00, 1.14418353637558e+00, 1.13246320983586e+00, 1.12000827147688e+00, 1.10689931678637e+00, 1.09322835762185e+00, 1.07909630693651e+00, 1.06461029185039e+00, 1.04988093309179e+00, 1.03501971614635e+00, 1.02013655843712e+00, 1.00533765078006e+00, 9.90723623558523e-01, 9.76388061485707e-01, 9.62416367688661e-01, 9.48884959503033e-01, 9.35860765325074e-01, 9.23400983957237e-01, 9.11553064461970e-01, 9.00354864710948e-01, 8.89834949610815e-01, 8.80012994492887e-01, 8.70900264590787e-01, 8.62500147275142e-01, 8.54808719309673e-01, 8.47815336513833e-01, 8.41503237656241e-01, 8.35850158029700e-01, 8.30828950909179e-01, 8.26408216937864e-01, 8.22552942433092e-01, 8.19225147685344e-01, 8.16384545603810e-01, 8.13989209637199e-01, 8.11996247895394e-01, 8.10362477988022e-01, 8.09045094474996e-01, 8.08002318226926e-01, 8.07194014660340e-01, 8.06582265993356e-01, 8.06131881580669e-01, 8.05810830216373e-01, 8.05590579147903e-01, 8.05446326461379e-01, 8.05357116414825e-01, 8.05305831066789e-01, 8.05279055948631e-01, 8.05266822279961e-01, 8.05262233025317e-01, 8.05260984638366e-01, 8.05260800379131e-01, 8.05260786224707e-01, 8.05260532801839e-01, 8.05258815825610e-01, 8.05252504055872e-01, 8.05235679321996e-01, 8.05198859446790e-01, 8.05128347446313e-01, 8.05005734199227e-01, 8.04807579297616e-01, 8.04505291840883e-01, 8.04065229293286e-01, 8.03449027897642e-01, 8.02614172208351e-01, 8.01514803811699e-01, 8.00102760107336e-01, 7.98328823203619e-01, 7.96144146888707e-01, 7.93501816953494e-01, 7.90358487855997e-01, 7.86676028079683e-01, 7.82423098927566e-01, 7.77576588227788e-01, 7.72122822544759e-01, 7.66058489531702e-01, 7.59391215906362e-01, 7.52139765316684e-01, 7.44333842548599e-01, 7.36013514063913e-01, 7.27228277478147e-01, 7.18035832140089e-01, 7.08500617721867e-01, 6.98692196587658e-01, 6.88683558354668e-01, 6.78549421891631e-01, 6.68364602004146e-01, 6.58202496588316e-01, 6.48133736588392e-01, 6.38225027089078e-01, 6.28538194510924e-01, 6.19129443037195e-01, 6.10048813642848e-01, 6.01339831664092e-01, 5.93039323749380e-01, 5.85177382088432e-01, 5.77777452746474e-01, 5.70856525385011e-01, 5.64425403268430e-01, 5.58489034880964e-01, 5.53046891383730e-01, 5.48093377238089e-01, 5.43618264361740e-01, 5.39607142967171e-01, 5.36041884596983e-01, 5.32901114703648e-01, 5.30160693345692e-01, 5.27794203156247e-01, 5.25773443694419e-01, 5.24068930658856e-01, 5.22650397321215e-01, 5.21487294044775e-01, 5.20549280048087e-01, 5.19806699832813e-01, 5.19231035108405e-01, 5.18795321800525e-01, 5.18474520992712e-01, 5.18245832554612e-01, 5.18088940836719e-01, 5.17986183184050e-01, 5.17922634096195e-01, 5.17886100532010e-01, 5.17867026965902e-01, 5.17858312152356e-01, 5.17855042934059e-01, 5.17854153628264e-01, 5.17854022368638e-01, 5.17854012208602e-01, 5.17853829299796e-01, 5.17852590066454e-01, 5.17848034524050e-01, 5.17835891201017e-01, 5.17809316243699e-01, 5.17758423585121e-01, 5.17669925718380e-01, 5.17526902775259e-01, 5.17308715433472e-01, 5.16991074512845e-01, 5.16546276776936e-01, 5.15943612224746e-01, 5.15149942864483e-01, 5.14130446525030e-01, 5.12849511730224e-01, 5.11271761273335e-01, 5.09363173329490e-01, 5.07092260376851e-01, 5.04431258683752e-01, 5.01357275560736e-01, 4.97853338853943e-01, 4.93909293991535e-01, 4.89522498707927e-01, 4.84698274384442e-01, 4.79450085309722e-01, 4.73799432186283e-01, 4.67775462634606e-01, 4.61414317810366e-01, 4.54758249079120e-01, 4.47854550691403e-01, 4.40754362619671e-01, 4.33511401655053e-01, 4.26180678489771e-01, 4.18817254243791e-01, 4.11475082485026e-01, 4.04205973209500e-01, 3.97058704535780e-01, 3.90078297024260e-01, 3.83305455406841e-01, 3.76776173753186e-01, 3.70521493116385e-01, 3.64567395683139e-01, 3.58934816385968e-01, 3.53639751654654e-01, 3.48693445213030e-01, 3.44102632234780e-01, 3.39869825405970e-01, 3.35993629167220e-01, 3.32469071325690e-01, 3.29287944088605e-01, 3.26439149178867e-01, 3.23909043906269e-01, 3.21681786789441e-01, 3.19739682500014e-01, 3.18063526516994e-01, 3.16632949953316e-01, 3.15426764598413e-01, 3.14423307385951e-01, 3.13600782348374e-01, 3.12937596781609e-01, 3.12412686955057e-01, 3.12005827408569e-01, 3.11697916825096e-01, 3.11471232783376e-01, 3.11309647484677e-01, 3.11198796874478e-01, 3.11126196466944e-01, 3.11081298595155e-01, 3.11055487680704e-01, 3.11042012319348e-01, 3.11035855375295e-01, 3.11033545702699e-01, 3.11032917416668e-01, 3.11032824682997e-01, 3.11032817446829e-01, 3.11032686424099e-01, 3.11031798726532e-01, 3.11028535466420e-01, 3.11019836890623e-01, 3.11000800663905e-01, 3.10964345520702e-01, 3.10900954440541e-01, 3.10798510273329e-01, 3.10642235277285e-01, 3.10414744208487e-01, 3.10096218253790e-01, 3.09664704058434e-01, 3.09096538217787e-01, 3.08366892805239e-01, 3.07450431827050e-01, 3.06322062105195e-01, 3.04957755337966e-01, 3.03335411476695e-01, 3.01435727733292e-01, 2.99243033209933e-01, 2.96746047021969e-01, 2.93938518451796e-01, 2.90819711469979e-01, 2.87394702925141e-01, 2.83674473488446e-01, 2.79675782333044e-01, 2.75420829518993e-01, 2.70936722946175e-01, 2.66254778311975e-01, 2.61409689693465e-01, 2.56438614378141e-01, 2.51380217990516e-01, 2.46273724824334e-01, 2.41158013991964e-01, 2.36070795247727e-01, 2.31047890005968e-01, 2.26122634081221e-01, 2.21325409877792e-01, 2.16683307843937e-01, 2.12219910464441e-01, 2.07955187159499e-01, 2.03905485247034e-01, 2.00083600504385e-01, 1.96498910616839e-01, 1.93157555644575e-01, 1.90062651273081e-01, 1.87214522748093e-01, 1.84610949775703e-01, 1.82247415079104e-01, 1.80117351576556e-01, 1.78212385157778e-01, 1.76522571703686e-01, 1.75036628266966e-01, 1.73742159185480e-01, 1.72625878336717e-01, 1.71673828779376e-01, 1.70871600707017e-01, 1.70204548012933e-01, 1.69658002908875e-01, 1.69217487038580e-01, 1.68868916476953e-01, 1.68598797014766e-01, 1.68394405294180e-01, 1.68243950780763e-01, 1.68136713306879e-01, 1.68063151047761e-01, 1.68014974316534e-01, 1.67985181466705e-01, 1.67968054420328e-01, 1.67959112815012e-01, 1.67955027380216e-01, 1.67953494802063e-01, 1.67953077904528e-01, 1.67953016371357e-01, 1.67953011528965e-01, 1.67952923325970e-01, 1.67952325738312e-01, 1.67950128952057e-01, 1.67944273197534e-01, 1.67931458382299e-01, 1.67906917858772e-01, 1.67864246002765e-01, 1.67795288537334e-01, 1.67690103639302e-01, 1.67537000714390e-01, 1.67322662187161e-01, 1.67032351585220e-01, 1.66650208484327e-01, 1.66159627453183e-01, 1.65543714015499e-01, 1.64785805963035e-01, 1.63870043368462e-01, 1.62781965759166e-01, 1.61509110615839e-01, 1.60041584216643e-01, 1.58372574396324e-01, 1.56498775477161e-01, 1.54420698702609e-01, 1.52142846964097e-01, 1.49673740149538e-01, 1.47025786469527e-01, 1.44215004817880e-01, 1.41260612657376e-01, 1.38184502158737e-01, 1.35010633575661e-01, 1.31764378571376e-01, 1.28471847206782e-01, 1.25159230663499e-01, 1.21852187900018e-01, 1.18575298924313e-01, 1.15351600901988e-01, 1.12202216601631e-01, 1.09146078312946e-01, 1.06199744820358e-01, 1.03377304579823e-01, 1.00690355055619e-01, 9.81480462269673e-02, 9.57571754533149e-02, 9.35223210138645e-02, 9.14460024988131e-02, 8.95288576112705e-02, 8.77698266395726e-02, 8.61663377057340e-02, 8.47144877412802e-02, 8.34092158751645e-02, 8.22444674560359e-02, 8.12133482143906e-02, 8.03082690673307e-02, 7.95210827582338e-02, 7.88432139097637e-02, 7.82657841589014e-02, 7.77797338652135e-02, 7.73759414750117e-02, 7.70453410334397e-02, 7.67790376241066e-02, 7.65684197497052e-02, 7.64052669193092e-02, 7.62818500535758e-02, 7.61910218222397e-02, 7.61262937475919e-02, 7.60818968793586e-02, 7.60528230861013e-02, 7.60348445085981e-02, 7.60245094472936e-02, 7.60191138562334e-02, 7.60166486214615e-02, 7.60157238357380e-02, 7.60154722724451e-02, 7.60154351422671e-02, 7.60154321944100e-02, 7.60153781713342e-02, 7.60150121575833e-02, 7.60136666605575e-02, 7.60100801227708e-02, 7.60022313935877e-02, 7.59872013684202e-02, 7.59610678779694e-02, 7.59188395439452e-02, 7.58544338634835e-02, 7.57607041599613e-02, 7.56295191410601e-02, 7.54518975251273e-02, 7.52181984383924e-02, 7.49183659930992e-02, 7.45422236376783e-02, 7.40798106163538e-02, 7.35217493803759e-02, 7.28596293424033e-02, 7.20863893323838e-02, 7.11966789052383e-02, 7.01871776717695e-02, 6.90568523950219e-02, 6.78071338955846e-02, 6.64419998244767e-02, 6.49679548564706e-02, 6.33939063782805e-02, 6.17309406887045e-02, 5.99920113947263e-02, 5.81915574056211e-02, 5.63450721329133e-02, 5.44686478299212e-02, 5.25785193142667e-02, 5.06906297156888e-02, 4.88202376825484e-02, 4.69815811127461e-02, 4.51876074651523e-02, 4.34497755711902e-02, 4.17779290645797e-02, 4.01802374255169e-02, 3.86631974272547e-02, 3.72316855748664e-02, 3.58890509252490e-02, 3.46372373834498e-02, 3.34769250343925e-02, 3.24076811158847e-02, 3.14281126721028e-02, 3.05360145679998e-02, 2.97285082381080e-02, 2.90021681588257e-02, 2.83531344768164e-02, 2.77772114555290e-02, 2.72699523886644e-02, 2.68267323584723e-02, 2.64428106921883e-02, 2.61133852095696e-02, 2.58336403796591e-02, 2.55987913471216e-02, 2.54041254759017e-02, 2.52450426300896e-02, 2.51170949036066e-02, 2.50160259665790e-02, 2.49378096550430e-02, 2.48786869402409e-02, 2.48352000112106e-02, 2.48042219264448e-02, 2.47829801624562e-02, 2.47690724203835e-02, 2.47604732437452e-02, 2.47555303364919e-02, 2.47529499154666e-02, 2.47517709504592e-02, 2.47513286882253e-02, 2.47512083829236e-02, 2.47511906261622e-02, 2.47511892034993e-02, 2.47511629690251e-02, 2.47509852271577e-02, 2.47503318363158e-02, 2.47485901906828e-02, 2.47447789179631e-02, 2.47374809330836e-02, 2.47247929875982e-02, 2.47042947839873e-02, 2.46730406143598e-02, 2.46275762758899e-02, 2.45639836280134e-02, 2.44779545081855e-02, 2.43648947366353e-02, 2.42200575702611e-02, 2.40387042107386e-02, 2.38162868918579e-02, 2.35486478017179e-02, 2.32322248415500e-02, 2.28642532547809e-02, 2.24429507729553e-02, 2.19676734139398e-02, 2.14390296611100e-02, 2.08589425681680e-02, 2.02306523426046e-02, 1.95586559617027e-02, 1.88485850152205e-02, 1.81070277655948e-02, 1.73413058171263e-02, 1.65592192788750e-02, 1.57687765457979e-02, 1.49779255725520e-02, 1.41943025751061e-02, 1.34250116630170e-02, 1.26764456787640e-02, 1.19541547520896e-02, 1.12627645816730e-02, 1.06059425559266e-02, 9.98640770235681e-03, 9.40597776325092e-03, 8.86564354590291e-03, 8.36566230732672e-03, 7.90566432782214e-03, 7.48476253210129e-03, 7.10165580670719e-03, 6.75472756407203e-03, 6.44213937462723e-03, 6.16190818615792e-03, 5.91196434022834e-03, 5.69020196776648e-03, 5.49452457175903e-03, 5.32287717244628e-03, 5.17326357481807e-03, 5.04375579313887e-03, 4.93250092516462e-03, 4.83772687175588e-03, 4.75774636495885e-03, 4.69095816191611e-03, 4.63584573789433e-03, 4.59097572421285e-03, 4.55499809389437e-03, 4.52664835930250e-03, 4.50475099940206e-03, 4.48822342134747e-03, 4.47608010958334e-03, 4.46743664274152e-03, 4.46151308827325e-03, 4.45763617733519e-03, 4.45523965214534e-03, 4.45386230001077e-03, 4.45314331751945e-03, 4.45281483555959e-03, 4.45269161514261e-03, 4.45265809672435e-03, 4.45265314952792e-03, 4.45265274959183e-03, 4.45264532480394e-03, 4.45259502232886e-03, 4.45241010805830e-03, 4.45191722410243e-03, 4.45083870704366e-03, 4.44877378239994e-03, 4.44518462691690e-03, 4.43938833335267e-03, 4.43055584537433e-03, 4.41771891067333e-03, 4.39978600688562e-03, 4.37556798166838e-03, 4.34381375390558e-03, 4.30325587778668e-03, 4.25266504009869e-03, 4.19091177847974e-03, 4.11703288298781e-03, 4.03029890311757e-03, 3.93027702953930e-03, 3.81688036549122e-03, 3.69039578227159e-03, 3.55149691944234e-03, 3.40126282912934e-03, 3.24120021959118e-03, 3.07318251483935e-03, 2.89920632350515e-03, 2.72131098509476e-03, 2.54180308947127e-03, 2.36333916288345e-03, 2.18777864175479e-03, 2.01663428352106e-03, 1.85121414799028e-03, 1.69547173243245e-03, 1.54743134602597e-03, 1.40952340223066e-03, 1.28234250615833e-03, 1.16485187784576e-03, 1.05672103947541e-03, 9.57558788940105e-04, 8.66969646281746e-04, 7.84556385582391e-04, 7.09900726050705e-04, 6.42484681961264e-04, 5.81504345674944e-04, 5.26359735019355e-04, 4.76100837929039e-04, 4.29927832357180e-04, 3.87281058948474e-04, 3.47729555514320e-04, 3.10953727350238e-04, 2.76725747056616e-04, 2.45443624919517e-04, 2.16422028486818e-04, 1.89430788978484e-04, 1.64459077694801e-04, 1.41499405785339e-04, 1.20537602725790e-04, 1.01550362558100e-04, 8.45045820483392e-05, 6.93504027537983e-05, 5.60212311029318e-05, 4.44405015400662e-05, 3.45226699636214e-05, 2.61706687609831e-05, 1.92744171698678e-05, 1.37111198199591e-05, 9.34683893660804e-06, 6.03882496363894e-06, 3.63826600846551e-06, 1.99327469374566e-06, 9.52056581127181e-07, 3.66313329891454e-07, -1.42108547152020e-14, -2.84217094304040e-14],
[ 8.20000000000000e+01, 8.19999999055960e+01, 8.19999969864383e+01, 8.19999772086810e+01, 8.19999045030460e+01, 8.19997106926210e+01, 8.19992865234764e+01, 8.19984741130692e+01, 8.19970610875319e+01, 8.19947766263490e+01, 8.19912895735782e+01, 8.19862087112608e+01, 8.19790852259579e+01, 8.19694173367394e+01, 8.19566569956351e+01, 8.19402185222909e+01, 8.19194889951104e+01, 8.18938401917673e+01, 8.18626418510499e+01, 8.18252760119995e+01, 8.17811521701258e+01, 8.17297229683914e+01, 8.16705001077069e+01, 8.16030701153624e+01, 8.15271095516777e+01, 8.14423991716464e+01, 8.13488365007666e+01, 8.12464462472571e+01, 8.11353879718028e+01, 8.10159604834784e+01, 8.08886025330948e+01, 8.07538895308964e+01, 8.06125262130834e+01, 8.04653354018771e+01, 8.03132432228894e+01, 8.01572613369529e+01, 7.99984668907584e+01, 7.98379809783258e+01, 7.96769464292513e+01, 7.95165057044566e+01, 7.93577795975088e+01, 7.92018473253137e+01, 7.90497284629425e+01, 7.89023670485428e+01, 7.87606180670881e+01, 7.86252364230127e+01, 7.84968684341416e+01, 7.83760458218340e+01, 7.82631821316973e+01, 7.81585714912300e+01, 7.80623895908536e+01, 7.79746967590574e+01, 7.78954429878234e+01, 7.78244747491598e+01, 7.77615434264579e+01, 7.77063151653378e+01, 7.76583819280822e+01, 7.76172735145970e+01, 7.75824702922499e+01, 7.75534163583413e+01, 7.75295328438081e+01, 7.75102310565692e+01, 7.74949251590817e+01, 7.74830440784445e+01, 7.74740423597108e+01, 7.74674096945683e+01, 7.74626788883435e+01, 7.74594320680345e+01, 7.74573049818812e+01, 7.74559892954329e+01, 7.74552328483342e+01, 7.74548378979390e+01, 7.74546574379965e+01, 7.74545897406199e+01, 7.74545713252333e+01, 7.74545686071587e+01, 7.74545682506046e+01, 7.74545599433228e+01, 7.74545036602512e+01, 7.74542967574940e+01, 7.74537452289243e+01, 7.74525382130899e+01, 7.74502265986947e+01, 7.74462065733697e+01, 7.74397088524563e+01, 7.74297941997863e+01, 7.74153557092888e+01, 7.73951281521289e+01, 7.73677045062277e+01, 7.73315595713833e+01, 7.72850803333979e+01, 7.72266024767145e+01, 7.71544521625754e+01, 7.70669918984809e+01, 7.69626690399452e+01, 7.68400652088388e+01, 7.66979447123729e+01, 7.65352999369572e+01, 7.63513917069648e+01, 7.61457827683130e+01, 7.59183628912290e+01, 7.56693645669161e+01, 7.53993688448837e+01, 7.51093014360209e+01, 7.48004196924543e+01, 7.44742913866341e+01, 7.41327663159065e+01, 7.37779416932186e+01, 7.34121221566637e+01, 7.30377751816678e+01, 7.26574828238649e+01, 7.22738910784903e+01, 7.18896586044013e+01, 7.15074069173228e+01, 7.11296741923209e+01, 7.07588744300875e+01, 7.03972630203993e+01, 7.00469089131184e+01, 6.97096729429043e+01, 6.93871915083134e+01, 6.90808647852287e+01, 6.87918488477806e+01, 6.85210513313128e+01, 6.82691304896415e+01, 6.80364976218919e+01, 6.78233228732832e+01, 6.76295443772557e+01, 6.74548806361402e+01, 6.72988459594002e+01, 6.71607687063946e+01, 6.70398120193477e+01, 6.69349966807980e+01, 6.68452256851100e+01, 6.67693100726869e+01, 6.67059955368658e+01, 6.66539892774889e+01, 6.66119865439514e+01, 6.65786962874922e+01, 6.65528653316146e+01, 6.65333004748283e+01, 6.65188879647947e+01, 6.65086098297805e+01, 6.65015566229284e+01, 6.64969362264206e+01, 6.64940784735910e+01, 6.64924354732942e+01, 6.64915776569643e+01, 6.64911857085503e+01, 6.64910386743519e+01, 6.64909986774078e+01, 6.64909927739398e+01, 6.64909922527760e+01, 6.64909820408268e+01, 6.64909128533544e+01, 6.64906585126755e+01, 6.64899805321809e+01, 6.64884967802001e+01, 6.64856551860300e+01, 6.64807135309337e+01, 6.64727262366033e+01, 6.64605389126363e+01, 6.64427912502521e+01, 6.64179286477285e+01, 6.63842227196441e+01, 6.63398005740764e+01, 6.62826824395135e+01, 6.62108268911487e+01, 6.61221825752004e+01, 6.60147449772984e+01, 6.58866164502363e+01, 6.57360674350480e+01, 6.55615966056900e+01, 6.53619875662197e+01, 6.51363597469498e+01, 6.48842112885792e+01, 6.46054519657564e+01, 6.43004245708663e+01, 6.39699136385130e+01, 6.36151409249083e+01, 6.32377476480660e+01, 6.28397641234996e+01, 6.24235680629108e+01, 6.19918333896873e+01, 6.15474719000718e+01, 6.10935703963418e+01, 6.06333259901337e+01, 6.01699821085911e+01, 5.97067673667420e+01, 5.92468389687690e+01, 5.87932317605928e+01, 5.83488135629187e+01, 5.79162470269016e+01, 5.74979579953309e+01, 5.70961102070243e+01, 5.67125861152263e+01, 5.63489735616139e+01, 5.60065580246880e+01, 5.56863201298023e+01, 5.53889380680782e+01, 5.51147945322233e+01, 5.48639877494766e+01, 5.46363461819388e+01, 5.44314464723286e+01, 5.42486342330460e+01, 5.40870472996744e+01, 5.39456410879368e+01, 5.38232156988697e+01, 5.37184444069227e+01, 5.36299031394395e+01, 5.35561005160133e+01, 5.34955079673021e+01, 5.34465894012932e+01, 5.34078298378648e+01, 5.33777623970580e+01, 5.33549930095632e+01, 5.33382222252232e+01, 5.33262635309113e+01, 5.33180576548999e+01, 5.33126824303649e+01, 5.33093579131012e+01, 5.33074465927495e+01, 5.33064486958851e+01, 5.33059927450513e+01, 5.33058217015839e+01, 5.33057751735598e+01, 5.33057683061190e+01, 5.33057677552653e+01, 5.33057575892204e+01, 5.33056887127644e+01, 5.33054355155357e+01, 5.33047605839913e+01, 5.33032835092919e+01, 5.33004547210384e+01, 5.32955353922829e+01, 5.32875843357380e+01, 5.32754526654605e+01, 5.32577868290327e+01, 5.32330404175017e+01, 5.31994949289609e+01, 5.31552893928494e+01, 5.30984584550295e+01, 5.30269781823442e+01, 5.29388184795015e+01, 5.28320006373264e+01, 5.27046581727113e+01, 5.25550988053405e+01, 5.23818651758198e+01, 5.21837917753568e+01, 5.19600555558814e+01, 5.17102178410688e+01, 5.14342554716777e+01, 5.11325795875368e+01, 5.08060410516808e+01, 5.04559222198590e+01, 5.00839154932761e+01, 4.96920897921300e+01, 4.92828466758494e+01, 4.88588682478776e+01, 4.84230591832152e+01, 4.79784852132133e+01, 4.75283102431447e+01, 4.70757340359288e+01, 4.66239321373540e+01, 4.61759994829613e+01, 4.57348989165921e+01, 4.53034156417131e+01, 4.48841183904395e+01, 4.44793278177133e+01, 4.40910923186835e+01, 4.37211711548560e+01, 4.33710244958137e+01, 4.30418097700598e+01, 4.27343835887790e+01, 4.24493084614346e+01, 4.21868635494013e+01, 4.19470587819416e+01, 4.17296517638028e+01, 4.15341670136673e+01, 4.13599171707267e+01, 4.12060258817192e+01, 4.10714521273668e+01, 4.09550157642015e+01, 4.08554240474842e+01, 4.07712988676832e+01, 4.07012043826280e+01, 4.06436746667339e+01, 4.05972409348924e+01, 4.05604578392579e+01, 4.05319282896215e+01, 4.05103262191551e+01, 4.04944167127889e+01, 4.04830729395319e+01, 4.04752893847882e+01, 4.04701909639611e+01, 4.04670377116858e+01, 4.04652248768953e+01, 4.04642784056795e+01, 4.04638459533185e+01, 4.04636837252095e+01, 4.04636395952022e+01, 4.04636330817045e+01, 4.04636325793209e+01, 4.04636235581237e+01, 4.04635624381862e+01, 4.04633377548361e+01, 4.04627388315628e+01, 4.04614281028586e+01, 4.04589179014666e+01, 4.04545526429312e+01, 4.04474972319451e+01, 4.04367323888647e+01, 4.04210574472269e+01, 4.03991009989794e+01, 4.03693395580363e+01, 4.03301241704823e+01, 4.02797146199692e+01, 4.02163205627750e+01, 4.01381485876889e+01, 4.00434538470630e+01, 3.99305945693708e+01, 3.97980874683513e+01, 3.96446618406194e+01, 3.94693100239802e+01, 3.92713319002066e+01, 3.90503712879356e+01, 3.88064423901110e+01, 3.85399449262552e+01, 3.82516671648395e+01, 3.79427767294764e+01, 3.76147997248160e+01, 3.72695893470267e+01, 3.69092856456782e+01, 3.65362684393089e+01, 3.61531055316925e+01, 3.57624983369292e+01, 3.53672268366888e+01, 3.49700955213098e+01, 3.45738816726033e+01, 3.41812870821089e+01, 3.37948940895889e+01, 3.34171266672239e+01, 3.30502171363793e+01, 3.26961789503534e+01, 3.23567857842903e+01, 3.20335569424417e+01, 3.17277488473708e+01, 3.14403521533305e+01, 3.11720938626237e+01, 3.09234437399284e+01, 3.06946243157725e+01, 3.04856238299646e+01, 3.02962115632651e+01, 3.01259551149201e+01, 2.99742392843029e+01, 2.98402862938465e+01, 2.97231771418009e+01, 2.96218738963869e+01, 2.95352427401077e+01, 2.94620775485476e+01, 2.94011237471207e+01, 2.93511021377370e+01, 2.93107323315360e+01, 2.92787553703305e+01, 2.92539550749225e+01, 2.92351776293631e+01, 2.92213489020355e+01, 2.92114890212674e+01, 2.92047237672779e+01, 2.92002924137053e+01, 2.91975517485351e+01, 2.91959761215652e+01, 2.91951534974606e+01, 2.91947776326102e+01, 2.91946366325501e+01, 2.91945982770982e+01, 2.91945926159132e+01, 2.91945921872654e+01, 2.91945845938116e+01, 2.91945331470468e+01, 2.91943440233197e+01, 2.91938398890748e+01, 2.91927366044975e+01, 2.91906236859436e+01, 2.91869493134895e+01, 2.91810105793332e+01, 2.91719495673426e+01, 2.91587557306011e+01, 2.91402748879279e+01, 2.91152249874524e+01, 2.90822185818541e+01, 2.90397917247727e+01, 2.89864387339098e+01, 2.89206519812040e+01, 2.88409655774977e+01, 2.87460015372005e+01, 2.86345167611083e+01, 2.85054489890070e+01, 2.83579597741214e+01, 2.81914725412998e+01, 2.80057039247047e+01, 2.78006868417096e+01, 2.75767841360082e+01, 2.73346920873416e+01, 2.70754335968002e+01, 2.68003413665049e+01, 2.65110318526684e+01, 2.62093711447006e+01, 2.58974341922376e+01, 2.55774589700476e+01, 2.52517972570349e+01, 2.49228637343667e+01, 2.45930850964015e+01, 2.42648508193766e+01, 2.39404671355221e+01, 2.36221155972935e+01, 2.33118173759979e+01, 2.30114041240321e+01, 2.27224958608995e+01, 2.24464859550980e+01, 2.21845329080466e+01, 2.19375583402484e+01, 2.17062503607054e+01, 2.14910713800602e+01, 2.12922694029686e+01, 2.11098918908500e+01, 2.09438014002710e+01, 2.07936923501971e+01, 2.06591084301192e+01, 2.05394603115772e+01, 2.04340434540528e+01, 2.03420558941178e+01, 2.02626159703736e+01, 2.01947799662230e+01, 2.01375596508962e+01, 2.00899396715334e+01, 2.00508947020161e+01, 2.00194061950870e+01, 1.99944785209388e+01, 1.99751542157681e+01, 1.99605280151307e+01, 1.99497593157312e+01, 1.99420827004245e+01, 1.99368161777387e+01, 1.99333668299110e+01, 1.99312336307543e+01, 1.99300072828883e+01, 1.99293670274353e+01, 1.99290744914067e+01, 1.99289647513625e+01, 1.99289348994399e+01, 1.99289304933597e+01, 1.99289301625939e+01, 1.99289243407409e+01, 1.99288848968492e+01, 1.99287398972142e+01, 1.99283533835987e+01, 1.99275075190424e+01, 1.99258876295229e+01, 1.99230707604442e+01, 1.99185183066885e+01, 1.99115732040596e+01, 1.99014620841204e+01, 1.98873026881515e+01, 1.98681167034043e+01, 1.98428480212106e+01, 1.98103862194522e+01, 1.97695948428470e+01, 1.97193437997617e+01, 1.96585449256858e+01, 1.95861894984633e+01, 1.95013862509858e+01, 1.94033982385365e+01, 1.92916768063016e+01, 1.91658908914145e+01, 1.90259500013182e+01, 1.88720194452973e+01, 1.87045267559365e+01, 1.85241587053566e+01, 1.83318488663939e+01, 1.81287562475589e+01, 1.79162360896489e+01, 1.76958043951473e+01, 1.74690981174348e+01, 1.72378331265811e+01, 1.70037620731980e+01, 1.67686340967504e+01, 1.65341579990228e+01, 1.63019700754691e+01, 1.60736073257198e+01, 1.58504863083078e+01, 1.56338875124257e+01, 1.54249448230611e+01, 1.52246394673539e+01, 1.50337977438029e+01, 1.48530918333844e+01, 1.46830430476704e+01, 1.45240269582760e+01, 1.43762799532136e+01, 1.42399068641671e+01, 1.41148893959587e+01, 1.40010951623772e+01, 1.38982871910720e+01, 1.38061338057648e+01, 1.37242188280331e+01, 1.36520520642433e+01, 1.35890800560893e+01, 1.35346970755156e+01, 1.34882563365336e+01, 1.34490813780340e+01, 1.34164775443808e+01, 1.33897434564296e+01, 1.33681823276045e+01, 1.33511129414082e+01, 1.33378800722498e+01, 1.33278641048022e+01, 1.33204895919525e+01, 1.33152324907275e+01, 1.33116258312026e+01, 1.33092636058401e+01, 1.33078027149704e+01, 1.33069628658322e+01, 1.33065243941280e+01, 1.33063240539799e+01, 1.33062488996671e+01, 1.33062284558901e+01, 1.33062254384321e+01, 1.33062252125994e+01, 1.33062212468642e+01, 1.33061943784366e+01, 1.33060956074282e+01, 1.33058323215170e+01, 1.33052561335061e+01, 1.33041526904604e+01, 1.33022338748658e+01, 1.32991327764963e+01, 1.32944017644961e+01, 1.32875139297076e+01, 1.32778680937333e+01, 1.32647974896007e+01, 1.32475821059255e+01, 1.32254645506042e+01, 1.31976691326000e+01, 1.31634236862939e+01, 1.31219834813401e+01, 1.30726563852998e+01, 1.30148282931245e+01, 1.29479877250379e+01, 1.28717484399753e+01, 1.27858689293616e+01, 1.26902677532286e+01, 1.25850338567375e+01, 1.24704312502228e+01, 1.23468977318282e+01, 1.22150376547332e+01, 1.20756090646331e+01, 1.19295058325338e+01, 1.17777356620882e+01, 1.16213950438781e+01, 1.14616423504093e+01, 1.12996703082949e+01, 1.11366790448518e+01, 1.09738507864758e+01, 1.08123270940298e+01, 1.06531892736592e+01, 1.04974423267448e+01, 1.03460025327120e+01, 1.01996885248250e+01, 1.00592155455486e+01, 9.92519246495851e+00, 9.79812110948663e+00, 9.67839746481134e+00, 9.56631436681255e+00, 9.46206535978244e+00, 9.36574946749778e+00, 9.27737668201419e+00, 9.19687402381902e+00, 9.12409206523084e+00, 9.05881183807249e+00, 9.00075206820850e+00, 8.94957669450524e+00, 8.90490263844416e+00, 8.86630779284020e+00, 8.83333919366348e+00, 8.80552132788414e+00, 8.78236451316097e+00, 8.76337326334463e+00, 8.74805452910782e+00, 8.73592567802240e+00, 8.72652205588649e+00, 8.71940395391654e+00, 8.71416279717538e+00, 8.71042637037684e+00, 8.70786290928064e+00, 8.70618390959358e+00, 8.70514553992643e+00, 8.70454858922648e+00, 8.70423692970087e+00, 8.70409453046889e+00, 8.70404111169452e+00, 8.70402658050152e+00, 8.70402443572820e+00, 8.70402427514895e+00, 8.70402145449360e+00, 8.70400234414518e+00, 8.70393209258046e+00, 8.70374482845142e+00, 8.70333500928909e+00, 8.70255017114147e+00, 8.70118537662681e+00, 8.69897962830024e+00, 8.69561447832695e+00, 8.69071502239781e+00, 8.68385341317899e+00, 8.67455496379775e+00, 8.66230683285654e+00, 8.64656918834187e+00, 8.62678863951193e+00, 8.60241360680631e+00, 8.57291117616177e+00, 8.53778486467071e+00, 8.49659262024444e+00, 8.44896430084211e+00, 8.39461784039557e+00, 8.33337331794982e+00, 8.26516420878660e+00, 8.19004521096912e+00, 8.10819620141031e+00, 8.01992207033072e+00, 7.92564839603625e+00, 7.82591313593416e+00, 7.72135470814277e+00, 7.61269700743111e+00, 7.50073203013579e+00, 7.38630086990443e+00, 7.27027388782541e+00, 7.15353085742692e+00, 7.03694183962753e+00, 6.92134945905384e+00, 6.80755313711843e+00, 6.69629569697261e+00, 6.58825260117177e+00, 6.48402392637385e+00, 6.38412903264604e+00, 6.28900375870850e+00, 6.19899987672669e+00, 6.11438647549404e+00, 6.03535290957421e+00, 5.96201295130983e+00, 5.89440980709226e+00, 5.83252170201203e+00, 5.77626779066398e+00, 5.72551420963636e+00, 5.68008014329205e+00, 5.63974382442098e+00, 5.60424843216568e+00, 5.57330787955505e+00, 5.54661250138984e+00, 5.52383466038646e+00, 5.50463428641346e+00, 5.48866435193666e+00, 5.47557626846759e+00, 5.46502516627666e+00, 5.45667499548894e+00, 5.45020336362036e+00, 5.44530600522181e+00, 5.44170076589292e+00, 5.43913097733339e+00, 5.43736810348720e+00, 5.43621355059136e+00, 5.43549955560609e+00, 5.43508909681435e+00, 5.43487480538927e+00, 5.43477689502159e+00, 5.43474016564916e+00, 5.43473017438694e+00, 5.43472869969820e+00, 5.43472858899528e+00, 5.43472664053148e+00, 5.43471343941201e+00, 5.43466491084349e+00, 5.43453555259602e+00, 5.43425246081193e+00, 5.43371032754844e+00, 5.43276762013154e+00, 5.43124413390144e+00, 5.42892008795086e+00, 5.42553690622617e+00, 5.42079979157289e+00, 5.41438215554265e+00, 5.40593190999927e+00, 5.39507955669009e+00, 5.38144792842395e+00, 5.36466334273518e+00, 5.34436783058524e+00, 5.32023200568028e+00, 5.29196805316678e+00, 5.25934224976604e+00, 5.22218639082787e+00, 5.18040750204673e+00, 5.13399526067879e+00, 5.08302664499345e+00, 5.02766746838711e+00, 4.96817062778288e+00, 4.90487109143743e+00, 4.83817785217236e+00, 4.76856325963949e+00, 4.69655030136461e+00, 4.62269851188985e+00, 4.54758924243734e+00, 4.47181101687042e+00, 4.39594563709805e+00, 4.32055559251815e+00, 4.24617318827765e+00, 4.17329165287678e+00, 4.10235833345631e+00, 4.03376995091226e+00, 3.96786977664969e+00, 3.90494651352508e+00, 3.84523461610759e+00, 3.78891576704353e+00, 3.73612123191931e+00, 3.68693483837403e+00, 3.64139636004369e+00, 3.59950512662081e+00, 3.56122372335368e+00, 3.52648168331129e+00, 3.49517911141950e+00, 3.46719020923237e+00, 3.44236669292771e+00, 3.42054111387411e+00, 3.40153010142092e+00, 3.38513755166001e+00, 3.37115778434747e+00, 3.35937868366746e+00, 3.34958482796196e+00, 3.34156060000231e+00, 3.33509325407371e+00, 3.32997590044513e+00, 3.32601035314960e+00, 3.32300977483244e+00, 3.32080104404878e+00, 3.31922676686746e+00, 3.31814685666744e+00, 3.31743961384100e+00, 3.31700225047888e+00, 3.31675082325283e+00, 3.31661955943321e+00, 3.31655958478923e+00, 3.31653708637195e+00, 3.31653096627038e+00, 3.31653006295669e+00, 3.31652999485736e+00, 3.31652879240480e+00, 3.31652064561905e+00, 3.31649069728005e+00, 3.31641086682814e+00, 3.31623616401932e+00, 3.31590160326712e+00, 3.31531984946004e+00, 3.31437971526775e+00, 3.31294561703798e+00, 3.31085808007084e+00, 3.30793536270562e+00, 3.30397624052306e+00, 3.29876395557697e+00, 3.29207129011397e+00, 3.28366667000053e+00, 3.27332114178662e+00, 3.26081600239498e+00, 3.24595079689870e+00, 3.22855134416238e+00, 3.20847740944177e+00, 3.18562962434683e+00, 3.15995526356640e+00, 3.13145252769681e+00, 3.10017305233190e+00, 3.06622246133935e+00, 3.02975889932975e+00, 2.99098960414604e+00, 2.95016570261544e+00, 2.90757551981174e+00, 2.86353677353738e+00, 2.81838807479156e+00, 2.77248016885032e+00, 2.72616733160683e+00, 2.67979928694308e+00, 2.63371394059222e+00, 2.58823114295494e+00, 2.54364760648893e+00, 2.50023302053249e+00, 2.45822733408386e+00, 2.41783911947569e+00, 2.37924488927072e+00, 2.34258921526192e+00, 2.30798549069814e+00, 2.27551718206080e+00, 2.24523943150122e+00, 2.21718089188279e+00, 2.19134570003035e+00, 2.16771551762960e+00, 2.14625159131340e+00, 2.12689680257850e+00, 2.10957769363786e+00, 2.09420646689618e+00, 2.08068296348792e+00, 2.06889663045033e+00, 2.05872848692802e+00, 2.05005309768472e+00, 2.04274055756308e+00, 2.03665848387692e+00, 2.03167400562663e+00, 2.02765572956538e+00, 2.02447565426274e+00, 2.02201099521170e+00, 2.02014587750242e+00, 2.01877284837205e+00, 2.01779416062944e+00, 2.01712277992883e+00, 2.01668307425288e+00, 2.01641115258766e+00, 2.01625483117655e+00, 2.01617321921037e+00, 2.01613593046240e+00, 2.01612194223578e+00, 2.01611813710561e+00, 2.01611757547661e+00, 2.01611753290588e+00, 2.01611677816459e+00, 2.01611166468501e+00, 2.01609286705067e+00, 2.01604275990069e+00, 2.01593310386923e+00, 2.01572310803628e+00, 2.01535795038404e+00, 2.01476783064638e+00, 2.01386762146406e+00, 2.01255717318625e+00, 2.01072231404201e+00, 2.00823656973060e+00, 2.00496360389813e+00, 2.00076035302232e+00, 1.99548079612599e+00, 1.98898026255665e+00, 1.98112014188752e+00, 1.97177282189641e+00, 1.96082664745762e+00, 1.94819066935590e+00, 1.93379894167117e+00, 1.91761413281854e+00, 1.89963024035875e+00, 1.87987424302992e+00, 1.85840658244180e+00, 1.83532043663092e+00, 1.81073982161637e+00, 1.78481662786129e+00, 1.75772675909661e+00, 1.72966558568747e+00, 1.70084295024473e+00, 1.67147796875530e+00, 1.64179385788688e+00, 1.61201299198198e+00, 1.58235235633860e+00, 1.55301952157537e+00, 1.52420922151882e+00, 1.49610057742299e+00, 1.46885497657743e+00, 1.44261458459593e+00, 1.41750144825029e+00, 1.39361712948259e+00, 1.37104280078914e+00, 1.34983972697609e+00, 1.33005005771793e+00, 1.31169785870728e+00, 1.29479031569558e+00, 1.27931905454793e+00, 1.26526153068893e+00, 1.25258245212093e+00, 1.24123521072576e+00, 1.23116330609146e+00, 1.22230175404896e+00, 1.21457847804014e+00, 1.20791568511963e+00, 1.20223122975007e+00, 1.19743996768862e+00, 1.19345509944371e+00, 1.19018949841242e+00, 1.18755701341200e+00, 1.18547372950212e+00, 1.18385916540461e+00, 1.18263738110970e+00, 1.18173796599011e+00, 1.18109687639550e+00, 1.18065709257149e+00, 1.18036906794580e+00, 1.18019094924409e+00, 1.18008855322699e+00, 1.18003509459777e+00, 1.18001066920989e+00, 1.18000150645238e+00, 1.17999901396479e+00, 1.17999864607900e+00, 1.17999861801684e+00, 1.17999811816465e+00, 1.17999473159534e+00, 1.17998228224933e+00, 1.17994909717905e+00, 1.17987647406497e+00, 1.17973739824659e+00, 1.17949556342448e+00, 1.17910474559217e+00, 1.17850857427491e+00, 1.17764073829427e+00, 1.17642565440985e+00, 1.17477961564853e+00, 1.17261242134869e+00, 1.16982947262942e+00, 1.16633429525568e+00, 1.16203142738262e+00, 1.15682958371754e+00, 1.15064498214778e+00, 1.14340469626445e+00, 1.13504988016138e+00, 1.12553870305246e+00, 1.11484883283944e+00, 1.10297932115016e+00, 1.08995176777373e+00, 1.07581067872448e+00, 1.06062297692043e+00, 1.04447667410579e+00, 1.02747876291107e+00, 1.00975243436287e+00, 9.91433764619316e-01, 9.72668041945482e-01, 9.53605918898447e-01, 9.34399574701047e-01, 9.15199059594102e-01, 8.96148968565115e-01, 8.77385559202338e-01, 8.59034391015257e-01, 8.41208525031519e-01, 8.24007286169177e-01, 8.07515559521761e-01, 7.91803567162503e-01, 7.76927055316492e-01, 7.62927812816969e-01, 7.49834439956032e-01, 7.37663290932730e-01, 7.26419521575494e-01, 7.16098185280529e-01, 7.06685332686462e-01, 6.98159083249180e-01, 6.90490648626820e-01, 6.83645297969349e-01, 6.77583263432652e-01, 6.72260590326488e-01, 6.67629940253065e-01, 6.63641357506847e-01, 6.60243009083544e-01, 6.57381907139410e-01, 6.55004619947007e-01, 6.53057973646369e-01, 6.51489742755359e-01, 6.50249322874785e-01, 6.49288374704454e-01, 6.48561424775352e-01, 6.48026405561268e-01, 6.47645116159595e-01, 6.47383584722462e-01, 6.47212315357777e-01, 6.47106405243591e-01, 6.47045522018544e-01, 6.47013736814955e-01, 6.46999214195858e-01, 6.46993766309066e-01, 6.46992284356401e-01, 6.46992065623465e-01, 6.46992048820849e-01, 6.46991747984231e-01, 6.46989709774545e-01, 6.46982217138742e-01, 6.46962244819605e-01, 6.46918537394583e-01, 6.46834838221011e-01, 6.46689302723772e-01, 6.46454127173442e-01, 6.46095421306597e-01, 6.45573349432908e-01, 6.44842559543065e-01, 6.43852912830781e-01, 6.42550516496186e-01, 6.40879050435544e-01, 6.38781363485506e-01, 6.36201297759968e-01, 6.33085681266849e-01, 6.29386410895918e-01, 6.25062531916313e-01, 6.20082208440721e-01, 6.14424473993125e-01, 6.08080654093669e-01, 6.01055364673343e-01, 5.93367011198509e-01, 5.85047742494012e-01, 5.76142848135689e-01, 5.66709625706878e-01, 5.56815780395908e-01, 5.46537450528263e-01, 5.35956975383669e-01, 5.25160533777481e-01, 5.14235782488782e-01, 5.03269613260997e-01, 4.92346127685636e-01, 4.81544903656655e-01, 4.70939598582447e-01, 4.60596906416527e-01, 4.50575860566971e-01, 4.40927454795698e-01, 4.31694540314950e-01, 4.22911949538360e-01, 4.14606794739058e-01, 4.06798892124570e-01, 3.99501267259112e-01, 3.92720705050706e-01, 3.86458315544374e-01, 3.80710094658170e-01, 3.75467466172083e-01, 3.70717797394420e-01, 3.66444885840764e-01, 3.62629417953784e-01, 3.59249403443712e-01, 3.56280590343729e-01, 3.53696866482835e-01, 3.51470652900062e-01, 3.49573293878379e-01, 3.47975446878252e-01, 3.46647473817683e-01, 3.45559833006874e-01, 3.44683468750816e-01, 3.43990193340119e-01, 3.43453054042030e-01, 3.43046675955705e-01, 3.42747570379927e-01, 3.42534397797138e-01, 3.42388174802949e-01, 3.42292415339045e-01, 3.42233198388968e-01, 3.42199156769340e-01, 3.42181384627580e-01, 3.42173264538005e-01, 3.42170218436124e-01, 3.42169389824846e-01, 3.42169267523630e-01, 3.42169258056984e-01, 3.42169087631689e-01, 3.42167932976508e-01, 3.42163688362206e-01, 3.42152373934368e-01, 3.42127613375055e-01, 3.42080196965554e-01, 3.41997748892211e-01, 3.41864516278690e-01, 3.41661295777541e-01, 3.41365511381923e-01, 3.40951454145397e-01, 3.40390690418360e-01, 3.39652639758327e-01, 3.38705316703795e-01, 3.37516222150413e-01, 3.36053360435102e-01, 3.34286347998926e-01, 3.32187569554677e-01, 3.29733329144617e-01, 3.26904937576373e-01, 3.23689675644786e-01, 3.20081575183849e-01, 3.16081967762699e-01, 3.11699763519613e-01, 3.06951439292419e-01, 3.01860734299936e-01, 2.96458071163656e-01, 2.90779737914576e-01, 2.84866880888202e-01, 2.78764367676061e-01, 2.72519582905844e-01, 2.66181217699668e-01, 2.59798107066615e-01, 2.53418159582154e-01, 2.47087412066122e-01, 2.40849230069728e-01, 2.34743663977483e-01, 2.28806961152372e-01, 2.23071227142640e-01, 2.17564223580538e-01, 2.12309286921979e-01, 2.07325350370766e-01, 2.02627050893113e-01, 1.98224903797225e-01, 1.94125528586071e-01, 1.90331911433702e-01, 1.86843691538627e-01, 1.83657460705845e-01, 1.80767067735331e-01, 1.78163921455280e-01, 1.75837288392444e-01, 1.73774582984478e-01, 1.71961649797993e-01, 1.70383038355666e-01, 1.69022271858680e-01, 1.67862111313909e-01, 1.66884816346126e-01, 1.66072403343961e-01, 1.65406900620681e-01, 1.64870599076735e-01, 1.64446295548046e-01, 1.64117524759845e-01, 1.63868774709201e-01, 1.63685679507282e-01, 1.63555183319815e-01, 1.63465669127888e-01, 1.63407046612519e-01, 1.63370794531971e-01, 1.63349954441003e-01, 1.63339074399588e-01, 1.63334103301978e-01, 1.63332238485083e-01, 1.63331731210945e-01, 1.63331656338400e-01, 1.63331650495962e-01, 1.63331544709067e-01, 1.63330827988119e-01, 1.63328193258124e-01, 1.63321170127915e-01, 1.63305800649297e-01, 1.63276368060210e-01, 1.63225190177002e-01, 1.63142488243849e-01, 1.63016341008912e-01, 1.62832732450696e-01, 1.62575699753546e-01, 1.62227585662109e-01, 1.61769396073325e-01, 1.61181259573496e-01, 1.60442980618669e-01, 1.59534672349864e-01, 1.58437448954203e-01, 1.57134151511258e-01, 1.55610076001011e-01, 1.53853668275119e-01, 1.51857148950185e-01, 1.49617031899140e-01, 1.47134503593904e-01, 1.44415636974855e-01, 1.41471422436965e-01, 1.38317609219712e-01, 1.34974362018383e-01, 1.31465748909136e-01, 1.27819086679196e-01, 1.24064177562275e-01, 1.20232476679476e-01, 1.16356231943200e-01, 1.12467637740494e-01, 1.08598040451213e-01, 1.04777228026308e-01, 1.01032828013530e-01, 9.73898295178088e-02, 9.38702357728829e-02, 9.04928462577317e-02, 8.72731610367623e-02, 8.42233951092481e-02, 8.13525866769425e-02, 7.86667803642160e-02, 7.61692649313517e-02, 7.38608453638392e-02, 7.17401314042405e-02, 6.98038279970774e-02, 6.80470167927183e-02, 6.64634211248938e-02, 6.50456495349090e-02, 6.37854151657820e-02, 6.26737304076954e-02, 6.17010780551936e-02, 6.08575617330160e-02, 6.01330392925519e-02, 5.95172431943638e-02, 5.89998916775869e-02, 5.85707938968341e-02, 5.82199513516883e-02, 5.79376569064181e-02, 5.77145916042809e-02, 5.75419183768133e-02, 5.74113707469479e-02, 5.73153337881678e-02, 5.72469140416132e-02, 5.71999948386974e-02, 5.71692735805556e-02, 5.71502779578736e-02, 5.71393588353715e-02, 5.71336584971078e-02, 5.71310540632055e-02, 5.71300770657501e-02, 5.71298113003706e-02, 5.71297720741057e-02, 5.71297689870818e-02, 5.71297127591350e-02, 5.71293318074879e-02, 5.71279314014248e-02, 5.71241985323354e-02, 5.71160297406266e-02, 5.71003874485427e-02, 5.70731913834521e-02, 5.70292514523771e-02, 5.69622481099117e-02, 5.68647657472496e-02, 5.67283837800403e-02, 5.65438287519129e-02, 5.63011887952882e-02, 5.59901890838717e-02, 5.56005235001749e-02, 5.51222337204962e-02, 5.45461225416659e-02, 5.38641838992646e-02, 5.30700281248215e-02, 5.21592780888653e-02, 5.11299105114205e-02, 4.99825173502586e-02, 4.87204651154087e-02, 4.73499353210372e-02, 4.58798368269839e-02, 4.43215899505418e-02, 4.26887918904697e-02, 4.09967819448127e-02, 3.92621320172140e-02, 3.75020923483049e-02, 3.57340243379980e-02, 3.39748523928449e-02, 3.22405652717066e-02, 3.05457940086313e-02, 2.89034869654472e-02, 2.73246923178050e-02, 2.58184459482038e-02, 2.43917522100361e-02, 2.30496403136726e-02, 2.17952807321353e-02, 2.06301506168955e-02, 1.95542389769798e-02, 1.85662788006482e-02, 1.76639871310158e-02, 1.68442921006005e-02, 1.61035317651539e-02, 1.54376201141275e-02, 1.48421845479305e-02, 1.43126822427462e-02, 1.38444999521141e-02, 1.34330369234732e-02, 1.30737682015507e-02, 1.27622874423849e-02, 1.24943322768303e-02, 1.22657977836980e-02, 1.20727435910197e-02, 1.19113987174728e-02, 1.17781663977325e-02, 1.16696298615011e-02, 1.15825594949825e-02, 1.15139215971851e-02, 1.14608886772629e-02, 1.14208510506444e-02, 1.13914289133419e-02, 1.13704838676512e-02, 1.13561284659340e-02, 1.13467324167971e-02, 1.13409239858697e-02, 1.13375856352178e-02, 1.13358429833568e-02, 1.13350468109701e-02, 1.13347481478456e-02, 1.13346669060235e-02, 1.13346549176896e-02, 1.13346539633454e-02, 1.13346365164233e-02, 1.13345183186908e-02, 1.13340838107661e-02, 1.13329256361965e-02, 1.13303913082348e-02, 1.13255389438507e-02, 1.13171043322069e-02, 1.13034816623667e-02, 1.12827202993060e-02, 1.12525399500214e-02, 1.12103663059600e-02, 1.11533887459672e-02, 1.10786409722579e-02, 1.09831040097106e-02, 1.08638297095780e-02, 1.07180807523406e-02, 1.05434813119274e-02, 1.03381702205354e-02, 1.01009467157951e-02, 9.82784528714831e-03, 9.52669582139265e-03, 9.19507249105322e-03, 8.83529437828656e-03, 8.45056944220346e-03, 8.04489518196760e-03, 7.62291182851982e-03, 7.18971754380959e-03, 6.75065878077419e-03, 6.31111124627637e-03, 5.87626748747994e-03, 5.45094592813200e-03, 5.03943342755520e-03, 4.64536956111772e-03, 4.27167641100823e-03, 3.92053332014086e-03, 3.59339233268940e-03, 3.29102729733677e-03, 3.01360801296369e-03, 2.76079033658050e-03, 2.53181367548905e-03, 2.32559849254699e-03, 2.14083806790200e-03, 1.97608051456655e-03, 1.82979870886868e-03, 1.70044722585772e-03, 1.58650647823322e-03, 1.48651503567976e-03, 1.39974899727235e-03, 1.32360676448684e-03, 1.25741186108996e-03, 1.19984896818437e-03, 1.15081565459718e-03, 1.10933772461408e-03, 1.07382915692256e-03, 1.04383159691679e-03, 1.01870498087919e-03, 9.97863108651131e-04, 9.80766474938832e-04, 9.66916863788470e-04, 9.55855926974891e-04, 9.47165346701695e-04, 9.40465929176071e-04, 9.35415646992510e-04, 9.31707748875965e-04, 9.29069553208706e-04, 9.27261900457665e-04, 9.26078918325857e-04, 9.25347683875088e-04, 9.24927429380773e-04, 9.24708056672330e-04, 9.24607832118340e-04, 9.24570235753208e-04, 9.24560008741082e-04, 9.24558499320938e-04, 9.24558378372353e-04, 9.24556151886691e-04, 9.24541067291784e-04, 9.24485616707216e-04, 9.24337820521259e-04, 9.24014446667343e-04, 9.23395432323559e-04, 9.22319849266273e-04, 9.20583785287477e-04, 9.17940525255290e-04, 9.14103387898990e-04, 9.08751500119820e-04, 9.01538683763192e-04, 8.92105572361857e-04, 8.80094973695122e-04, 8.65169938265353e-04, 8.46807202009359e-04, 8.25394191664941e-04, 8.00767793563750e-04, 7.72485865240924e-04, 7.40739714316874e-04, 7.05740329863147e-04, 6.67817658012382e-04, 6.27416726985075e-04, 5.84983786296789e-04, 5.42299443750949e-04, 4.98858412100844e-04, 4.55447124522834e-04, 4.12689106326525e-04, 3.71880940804203e-04, 3.33052941073220e-04, 2.96341442876269e-04, 2.62109967053448e-04, 2.31391276102499e-04, 2.03282988735509e-04, 1.77769075150991e-04, 1.55561665252435e-04, 1.35996226802604e-04, 1.18592808490803e-04, 1.03503355092016e-04, 9.09084843954133e-05, 7.98077286390253e-05, 6.99625037583473e-05, 6.20900504770816e-05, 5.54345574101944e-05, 4.95269640623519e-05, 4.42199729775439e-05, 4.01816106858632e-05, 3.69164067706151e-05, 3.39873956676229e-05, 3.13597006424970e-05, 2.90138451077837e-05, 2.69674917774410e-05, 2.56665725544281e-05, 2.45699164480584e-05, 2.36203448196193e-05, 2.28168001541462e-05, 2.21493735779177e-05, 2.16030590109995e-05, 2.11607486306775e-05, 2.08057917205906e-05, 2.05236719494906e-05, 2.03024690250686e-05, 2.01324005928427e-05, 2.00050717324984e-05, 1.99128989777364e-05, 1.98488631042437e-05, 1.98065265379910e-05, 1.97801752790383e-05, 1.97649651170195e-05, 1.97570059299323e-05, 1.97533651089543e-05, 1.97519986357975e-05, 1.97516268514164e-05, 1.97515719833063e-05, 1.97515675068871e-05, 1.97514836912660e-05, 1.97509159534093e-05, 1.97488292457138e-05, 1.97432696893429e-05, 1.97311170069270e-05, 1.97078973940279e-05, 1.96676853221334e-05, 1.96031280808029e-05, 1.95056221201639e-05, 1.93656419469335e-05, 1.91731583356614e-05, 1.89179800145212e-05, 1.85897389854972e-05, 1.81771978304823e-05, 1.76667589641966e-05, 1.70408911372988e-05, 1.62830756096355e-05, 1.55069260472374e-05, 1.48557249843861e-05, 1.41206856199005e-05, 1.32519190714220e-05, 1.22156762216719e-05, 1.09797433509584e-05, 9.50652096776139e-06, 8.83359288650354e-06, 8.02368404606568e-06, 6.99047470220648e-06, 5.69079024614894e-06, 5.04860788907990e-06, 4.56446296936974e-06, 3.77302330889506e-06, 2.72576458826279e-06, 2.82004484120080e-06, 2.53554817675194e-06, 1.61107229246227e-06, 2.44997441711803e-06, 3.47362210106894e-06, 4.15406341858215e-06, 4.12567543150999e-06, 3.45798294176802e-06, 2.42517756987581e-06, 1.38214390688063e-06, 6.76326777693248e-07, 4.81707758126504e-07, 5.98573407728509e-07, 7.04658404515612e-07, 7.35533888018836e-07, 7.57728386702183e-07, 7.35615145686097e-07, 5.98803012508142e-07, 0.00000000000000e+00, -1.42108547152020e-14, 5.23466582080800e-07, 6.17911027234186e-07, 6.20360907532813e-07, 5.67819725461050e-07, 4.90892588800307e-07, 4.10637753134324e-07, -1.42108547152020e-14, -1.42108547152020e-14, 0.00000000000000e+00, -1.42108547152020e-14, 0.00000000000000e+00, -1.42108547152020e-14, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, -1.42108547152020e-14, 0.00000000000000e+00, -1.42108547152020e-14, 0.00000000000000e+00, 0.00000000000000e+00],
[ 8.30000000000000e+01, 8.29999999039731e+01, 8.29999969346316e+01, 8.29999768168712e+01, 8.29999028613394e+01, 8.29997057190748e+01, 8.29992742579094e+01, 8.29984478809856e+01, 8.29970105631039e+01, 8.29946868270803e+01, 8.29911398219162e+01, 8.29859715999618e+01, 8.29787256244664e+01, 8.29688914752892e+01, 8.29559116623845e+01, 8.29391904067762e+01, 8.29181042088778e+01, 8.28920139944337e+01, 8.28602786073714e+01, 8.28222694027492e+01, 8.27773856764785e+01, 8.27250706456445e+01, 8.26648276589456e+01, 8.25962362684935e+01, 8.25189677336250e+01, 8.24327994612812e+01, 8.23376278276970e+01, 8.22334787878699e+01, 8.21205156784396e+01, 8.19990436692008e+01, 8.18695104250726e+01, 8.17325027016530e+01, 8.15887388016029e+01, 8.14390570459965e+01, 8.12844006395809e+01, 8.11257995063817e+01, 8.09643498212012e+01, 8.08011920498742e+01, 8.06374883326914e+01, 8.04744000063810e+01, 8.03130659728500e+01, 8.01545825041045e+01, 7.99999849397761e+01, 7.98502316018914e+01, 7.97061901324162e+01, 7.95686263595239e+01, 7.94381957208263e+01, 7.93154372147877e+01, 7.92007698117041e+01, 7.90944912283917e+01, 7.89967789514494e+01, 7.89076933786269e+01, 7.88271829334176e+01, 7.87550909925784e+01, 7.86911644488946e+01, 7.86350637120575e+01, 7.85863739294803e+01, 7.85446171872102e+01, 7.85092654300198e+01, 7.84797538207078e+01, 7.84554942431184e+01, 7.84358886429108e+01, 7.84203418961038e+01, 7.84082738991535e+01, 7.83991305867578e+01, 7.83923936053372e+01, 7.83875884013680e+01, 7.83842905240633e+01, 7.83821299904216e+01, 7.83807936159524e+01, 7.83800252745793e+01, 7.83796241140928e+01, 7.83794408166609e+01, 7.83793720548363e+01, 7.83793533498933e+01, 7.83793505890808e+01, 7.83793502269204e+01, 7.83793417890182e+01, 7.83792846209742e+01, 7.83790744649679e+01, 7.83785142644546e+01, 7.83772882704063e+01, 7.83749403109306e+01, 7.83708570826706e+01, 7.83642572129308e+01, 7.83541867149959e+01, 7.83395213138727e+01, 7.83189759529998e+01, 7.82911216014334e+01, 7.82544092634958e+01, 7.82072008483721e+01, 7.81478062878373e+01, 7.80745260019026e+01, 7.79856975149422e+01, 7.78797447346671e+01, 7.77552281455890e+01, 7.76108939667320e+01, 7.74457202150672e+01, 7.72589576371380e+01, 7.70501636497283e+01, 7.68192277753527e+01, 7.65663875486247e+01, 7.62922344478300e+01, 7.59977099839443e+01, 7.56840925570061e+01, 7.53529759868498e+01, 7.50062407143903e+01, 7.46460185980193e+01, 7.42746521135600e+01, 7.38946487518710e+01, 7.35086316049751e+01, 7.31192875430554e+01, 7.27293148769311e+01, 7.23413727422669e+01, 7.19580344144540e+01, 7.15817462888171e+01, 7.12147934620013e+01, 7.08592719931724e+01, 7.05170672771803e+01, 7.01898376691165e+01, 6.98790025381252e+01, 6.95857341638355e+01, 6.93109531667366e+01, 6.90553273763878e+01, 6.88192741475372e+01, 6.86029661442080e+01, 6.84063405593282e+01, 6.82291116574707e+01, 6.80707864455547e+01, 6.79306832034688e+01, 6.78079525465853e+01, 6.77016006427862e+01, 6.76105141639212e+01, 6.75334865122078e+01, 6.74692448243469e+01, 6.74164772205138e+01, 6.73738597341768e+01, 6.73400823354082e+01, 6.73138734492022e+01, 6.72940223754410e+01, 6.72793990421277e+01, 6.72689705707207e+01, 6.72618142027936e+01, 6.72571262299599e+01, 6.72542266814828e+01, 6.72525596520225e+01, 6.72516892900670e+01, 6.72512916093989e+01, 6.72511424248254e+01, 6.72511018429263e+01, 6.72510958531202e+01, 6.72510953243343e+01, 6.72510849630355e+01, 6.72510147636972e+01, 6.72507567032927e+01, 6.72500688073768e+01, 6.72485633557481e+01, 6.72456802041906e+01, 6.72406662805591e+01, 6.72325621813677e+01, 6.72201966420358e+01, 6.72021894772151e+01, 6.71769633830691e+01, 6.71427647567031e+01, 6.70976934162007e+01, 6.70397407979957e+01, 6.69668358713850e+01, 6.68768976537335e+01, 6.67678928518180e+01, 6.66378968184275e+01, 6.64851557264946e+01, 6.63081476541761e+01, 6.61056401683577e+01, 6.58767420082555e+01, 6.56209466119042e+01, 6.53381654925872e+01, 6.50287498480468e+01, 6.46934992572670e+01, 6.43336568717053e+01, 6.39508911225559e+01, 6.35472646187491e+01, 6.31251915639290e+01, 6.26873856190268e+01, 6.22368006117711e+01, 6.17765667786348e+01, 6.13099252732278e+01, 6.08401634824663e+01, 6.03705532987025e+01, 5.99042939809171e+01, 5.94444606961404e+01, 5.89939593490582e+01, 5.85554879381208e+01, 5.81315044355292e+01, 5.77242010564960e+01, 5.73354847207123e+01, 5.69669634748587e+01, 5.66199386104374e+01, 5.62954021652651e+01, 5.59940394440661e+01, 5.57162361458803e+01, 5.54620896550744e+01, 5.52314240442916e+01, 5.50238083498950e+01, 5.48385777057899e+01, 5.46748569497969e+01, 5.45315863383584e+01, 5.44075490132771e+01, 5.43013998548699e+01, 5.42116953293357e+01, 5.41369238971606e+01, 5.40755364991155e+01, 5.40259765833715e+01, 5.39867090887897e+01, 5.39562477629435e+01, 5.39331801757898e+01, 5.39161897969001e+01, 5.39040745398699e+01, 5.38957612439109e+01, 5.38903156592559e+01, 5.38869476269583e+01, 5.38850112897550e+01, 5.38840003318315e+01, 5.38835384132901e+01, 5.38833651311291e+01, 5.38833179941261e+01, 5.38833110368013e+01, 5.38833104787378e+01, 5.38833001796353e+01, 5.38832304016951e+01, 5.38829738905194e+01, 5.38822901252415e+01, 5.38807937182978e+01, 5.38779279071180e+01, 5.38729441971662e+01, 5.38648890888142e+01, 5.38525986731634e+01, 5.38347017106324e+01, 5.38096316058520e+01, 5.37756474579392e+01, 5.37308640926310e+01, 5.36732906712066e+01, 5.36008771245173e+01, 5.35115672887488e+01, 5.34033572395846e+01, 5.32743569567161e+01, 5.31228531300972e+01, 5.29473706750931e+01, 5.27467303877505e+01, 5.25201001719683e+01, 5.22670374276711e+01, 5.19875205118982e+01, 5.16819676668452e+01, 5.13512424265337e+01, 5.09966452239253e+01, 5.06198916613998e+01, 5.02230786048523e+01, 4.98086398388533e+01, 4.93792934165380e+01, 4.89379830256567e+01, 4.84878156871030e+01, 4.80319979573446e+01, 4.75737725910217e+01, 4.71163573929336e+01, 4.66628877768960e+01, 4.62163643457684e+01, 4.57796065858831e+01, 4.53552135056359e+01, 4.49455317365079e+01, 4.45526312737071e+01, 4.41782886985696e+01, 4.38239774356134e+01, 4.34908643848393e+01, 4.31798121493893e+01, 4.28913860473363e+01, 4.26258651377318e+01, 4.23832565810837e+01, 4.21633127678140e+01, 4.19655507628626e+01, 4.17892737140635e+01, 4.16335939460778e+01, 4.14974575059751e+01, 4.13796699407323e+01, 4.12789230737362e+01, 4.11938225115455e+01, 4.11229155596744e+01, 4.10647191637606e+01, 4.10177474274429e+01, 4.09805381980609e+01, 4.09516781632473e+01, 4.09298258724063e+01, 4.09137320927064e+01, 4.09022569337432e+01, 4.08943832305825e+01, 4.08892257612578e+01, 4.08860359892590e+01, 4.08842021590588e+01, 4.08832447262904e+01, 4.08828072654925e+01, 4.08826431585438e+01, 4.08825985174463e+01, 4.08825919285128e+01, 4.08825914203109e+01, 4.08825822946350e+01, 4.08825204668391e+01, 4.08822931813280e+01, 4.08816873216560e+01, 4.08803614128509e+01, 4.08778221398940e+01, 4.08734063261136e+01, 4.08662692056097e+01, 4.08553796962721e+01, 4.08395232306962e+01, 4.08173125261183e+01, 4.07872064658804e+01, 4.07475370196235e+01, 4.06965438460590e+01, 4.06324159040466e+01, 4.05533390540254e+01, 4.04575482786464e+01, 4.03433828115069e+01, 4.02093421646889e+01, 4.00541408212716e+01, 3.98767592401771e+01, 3.96764888355327e+01, 3.94529687603532e+01, 3.92062126506007e+01, 3.89366239595500e+01, 3.86449991040913e+01, 3.83325183064668e+01, 3.80007246854993e+01, 3.76514927631220e+01, 3.72869880421189e+01, 3.69096196331685e+01, 3.65219880441523e+01, 3.61268302048795e+01, 3.57269636279002e+01, 3.53252313610790e+01, 3.49244491309245e+01, 3.45273558507088e+01, 3.41365684880039e+01, 3.37545421361205e+01, 3.33835359784837e+01, 3.30255856410250e+01, 3.26824821817611e+01, 3.23557576837476e+01, 3.20466771354166e+01, 3.17562360462663e+01, 3.14851630909972e+01, 3.12339270149308e+01, 3.10027470584947e+01, 3.07916062436368e+01, 3.06002669798648e+01, 3.04282885653184e+01, 3.02750462600870e+01, 3.01397516847837e+01, 3.00214743439497e+01, 2.99191640922624e+01, 2.98316743550983e+01, 2.97577858881447e+01, 2.96962308184113e+01, 2.96457166566047e+01, 2.96049499143994e+01, 2.95726589060921e+01, 2.95476152691131e+01, 2.95286537082807e+01, 2.95146894601062e+01, 2.95047329901101e+01, 2.94979014803162e+01, 2.94934267359849e+01, 2.94906592380015e+01, 2.94890681857477e+01, 2.94882375085216e+01, 2.94878579641943e+01, 2.94877155838452e+01, 2.94876768529221e+01, 2.94876711363184e+01, 2.94876707034745e+01, 2.94876630356866e+01, 2.94876110852974e+01, 2.94874201102015e+01, 2.94869110409444e+01, 2.94857969565047e+01, 2.94836633560372e+01, 2.94799530210514e+01, 2.94739561712535e+01, 2.94648065110873e+01, 2.94514836395875e+01, 2.94328221490091e+01, 2.94075275630988e+01, 2.93741990602178e+01, 2.93313586887051e+01, 2.92774865147028e+01, 2.92110608540534e+01, 2.91306024434210e+01, 2.90347211207699e+01, 2.89221633357004e+01, 2.87918586225265e+01, 2.86429630696989e+01, 2.84748978304477e+01, 2.82873808552906e+01, 2.80804502891260e+01, 2.78544783509890e+01, 2.76101749749564e+01, 2.73485809954981e+01, 2.70710511630489e+01, 2.67792277318019e+01, 2.64750057396920e+01, 2.61604913861807e+01, 2.58379551106411e+01, 2.55097810981701e+01, 2.51784150066859e+01, 2.48463117264990e+01, 2.45158849455259e+01, 2.41894601847813e+01, 2.38692327744137e+01, 2.35572319553108e+01, 2.32552919274266e+01, 2.29650302534090e+01, 2.26878336067733e+01, 2.24248504723263e+01, 2.21769901004409e+01, 2.19449268099776e+01, 2.17291086345656e+01, 2.15297693057531e+01, 2.13469426452064e+01, 2.11804785716346e+01, 2.10300600903437e+01, 2.08952208008566e+01, 2.07753626123763e+01, 2.06697734852660e+01, 2.05776451119362e+01, 2.04980905100059e+01, 2.04301615254291e+01, 2.03728662371578e+01, 2.03251862233770e+01, 2.02860935991129e+01, 2.02545676735187e+01, 2.02296110101564e+01, 2.02102646128310e+01, 2.01956219102189e+01, 2.01848411809483e+01, 2.01771560517472e+01, 2.01718837177511e+01, 2.01684305767961e+01, 2.01662950369889e+01, 2.01650673452952e+01, 2.01644263887688e+01, 2.01641335325340e+01, 2.01640236723889e+01, 2.01639937877977e+01, 2.01639893768958e+01, 2.01639890457680e+01, 2.01639832175440e+01, 2.01639437304883e+01, 2.01637985721896e+01, 2.01634116357264e+01, 2.01625648462400e+01, 2.01609431871385e+01, 2.01581232462744e+01, 2.01535658425275e+01, 2.01466132226231e+01, 2.01364912323847e+01, 2.01223167602663e+01, 2.01031106179531e+01, 2.00778158581048e+01, 2.00453213304677e+01, 2.00044900460149e+01, 1.99541916609331e+01, 1.98933381203180e+01, 1.98209212332705e+01, 1.97360507094898e+01, 1.96379909985042e+01, 1.95261951630936e+01, 1.94003340125356e+01, 1.92603188372690e+01, 1.91063163330521e+01, 1.89387546756852e+01, 1.87583201882656e+01, 1.85659445982228e+01, 1.83627834645042e+01, 1.81501869110103e+01, 1.79296642734406e+01, 1.77028446020931e+01, 1.74714351265255e+01, 1.72371797646182e+01, 1.70018195587817e+01, 1.67670565799489e+01, 1.65345224075129e+01, 1.63057518306375e+01, 1.60821619807335e+01, 1.58650367413625e+01, 1.56555160168953e+01, 1.54545892815315e+01, 1.52630927654154e+01, 1.50817096431368e+01, 1.49109726462862e+01, 1.47512686021351e+01, 1.46028444867410e+01, 1.44658146615950e+01, 1.43401690333032e+01, 1.42257819349245e+01, 1.41224215767395e+01, 1.40297599548878e+01, 1.39473831392593e+01, 1.38748018869930e+01, 1.38114625439893e+01, 1.37567582028413e+01, 1.37100400807981e+01, 1.36706290658475e+01, 1.36378273538152e+01, 1.36109300666669e+01, 1.35892367050475e+01, 1.35720622502826e+01, 1.35587476967809e+01, 1.35486697691994e+01, 1.35412495636481e+01, 1.35359598516378e+01, 1.35323308012640e+01, 1.35299539027929e+01, 1.35284839343641e+01, 1.35276388655558e+01, 1.35271976684307e+01, 1.35269960829507e+01, 1.35269204614605e+01, 1.35268998905989e+01, 1.35268968543834e+01, 1.35268966271469e+01, 1.35268926367595e+01, 1.35268656013089e+01, 1.35267662163002e+01, 1.35265012936446e+01, 1.35259215234286e+01, 1.35248112191519e+01, 1.35228804690673e+01, 1.35197600740133e+01, 1.35149996023947e+01, 1.35080688329478e+01, 1.34983627818742e+01, 1.34852104189202e+01, 1.34678870637832e+01, 1.34456303183555e+01, 1.34176592330586e+01, 1.33831962317831e+01, 1.33414911387816e+01, 1.32918464753241e+01, 1.32336430403097e+01, 1.31663646752163e+01, 1.30896210569383e+01, 1.30031673760246e+01, 1.29069198504308e+01, 1.28009661961359e+01, 1.26855704172688e+01, 1.25611715733465e+01, 1.24283765076939e+01, 1.22879468540718e+01, 1.21407809529145e+01, 1.19878915813218e+01, 1.18303806120847e+01, 1.16694118500071e+01, 1.15061833362835e+01, 1.13419003574567e+01, 1.11777502474392e+01, 1.10148798441912e+01, 1.08543761844516e+01, 1.06972507270355e+01, 1.05444271259699e+01, 1.03967323603359e+01, 1.02548908852535e+01, 1.01195213988633e+01, 9.99113581091325e+00, 9.87014002996249e+00, 9.75683623811013e+00, 9.65142637871305e+00, 9.55401663426620e+00, 9.46462271473207e+00, 9.38317581125423e+00, 9.30952909818907e+00, 9.24346468959847e+00, 9.18470097576515e+00, 9.13290028087714e+00, 9.08767679397327e+00, 9.04860473043632e+00, 9.01522667984425e+00, 8.98706208747333e+00, 8.96361580159773e+00, 8.94438659819272e+00, 8.92887557074357e+00, 8.91659424819819e+00, 8.90707228163720e+00, 8.89986452296583e+00, 8.89455730959128e+00, 8.89077376978661e+00, 8.88817797560634e+00, 8.88647779414846e+00, 8.88542632295389e+00, 8.88482183967973e+00, 8.88450624732754e+00, 8.88436205112177e+00, 8.88430795823518e+00, 8.88429324366682e+00, 8.88429107182768e+00, 8.88429090922196e+00, 8.88428805297153e+00, 8.88426870146090e+00, 8.88419756335566e+00, 8.88400793601463e+00, 8.88359294491684e+00, 8.88279820149647e+00, 8.88141618038770e+00, 8.87918258598273e+00, 8.87577494172265e+00, 8.87081359185483e+00, 8.86386525209187e+00, 8.85444918016435e+00, 8.84204595762967e+00, 8.82610877961336e+00, 8.80607704039173e+00, 8.78139188325204e+00, 8.75151325897114e+00, 8.71593791724577e+00, 8.67421765028446e+00, 8.62597702933675e+00, 8.57092983451417e+00, 8.50889338492206e+00, 8.43980003502304e+00, 8.36370521424550e+00, 8.28079154442393e+00, 8.19136876280714e+00, 8.09586939239813e+00, 7.99484031982324e+00, 7.88893064764643e+00, 7.77887636902186e+00, 7.66548255690581e+00, 7.54960386019155e+00, 7.43212415070795e+00, 7.31393616650391e+00, 7.19592194945838e+00, 7.07893478321604e+00, 6.96378320818006e+00, 6.85121753438717e+00, 6.74191910363132e+00, 6.63649238344341e+00, 6.53545982113684e+00, 6.43925925739224e+00, 6.34824360336752e+00, 6.26268242630837e+00, 6.18276506510543e+00, 6.10860490476107e+00, 6.04024447057964e+00, 5.97766105133397e+00, 5.92077261801832e+00, 5.86944386426467e+00, 5.82349225059755e+00, 5.78269398344622e+00, 5.74678989866020e+00, 5.71549124688229e+00, 5.68848539422794e+00, 5.66544145682406e+00, 5.64601588301811e+00, 5.62985798412093e+00, 5.61661539543285e+00, 5.60593942635515e+00, 5.59749023415240e+00, 5.59094173300683e+00, 5.58598613092391e+00, 5.58233797405209e+00, 5.57973757285085e+00, 5.57795368842200e+00, 5.57678537056715e+00, 5.57606286128404e+00, 5.57564750719543e+00, 5.57543065986192e+00, 5.57533158165059e+00, 5.57529441417520e+00, 5.57528430373792e+00, 5.57528281145918e+00, 5.57528269943580e+00, 5.57528072773081e+00, 5.57526736914882e+00, 5.57521826172889e+00, 5.57508736045781e+00, 5.57480089171140e+00, 5.57425229080448e+00, 5.57329833498368e+00, 5.57175666533257e+00, 5.56940486865450e+00, 5.56598126267893e+00, 5.56118749259409e+00, 5.55469300151108e+00, 5.54614137996612e+00, 5.53515852889026e+00, 5.52136248707933e+00, 5.50437468054452e+00, 5.48383225198531e+00, 5.45940103104863e+00, 5.43078861899922e+00, 5.39775699505742e+00, 5.36013401611527e+00, 5.31782318560155e+00, 5.27081111684198e+00, 5.21917221305424e+00, 5.16307022668060e+00, 5.10275653618750e+00, 5.03856517489915e+00, 4.97090484642301e+00, 4.90024834576850e+00, 4.82711995634600e+00, 4.75208149616716e+00, 4.67571773290915e+00, 4.59862187510909e+00, 4.52138178070223e+00, 4.44456741545584e+00, 4.36871995761523e+00, 4.29434279774591e+00, 4.22189453978152e+00, 4.15178398307901e+00, 4.08436696406032e+00, 4.01994486353638e+00, 3.95876454190781e+00, 3.90101944604109e+00, 3.84685163397323e+00, 3.79635448132125e+00, 3.74957586120387e+00, 3.70652162314187e+00, 3.66715923219375e+00, 3.63142146476034e+00, 3.59921009005620e+00, 3.57039949477232e+00, 3.54484023197249e+00, 3.52236249316222e+00, 3.50277951442841e+00, 3.48589093353651e+00, 3.47148611512647e+00, 3.45934745617932e+00, 3.44925367451330e+00, 3.44098307025773e+00, 3.43431673531290e+00, 3.42904167018951e+00, 3.42495375287608e+00, 3.42186049204662e+00, 3.41958348839351e+00, 3.41796052428857e+00, 3.41684720408192e+00, 3.41611807540495e+00, 3.41566717558548e+00, 3.41540796592190e+00, 3.41527263887620e+00, 3.41521080768959e+00, 3.41518761281569e+00, 3.41518130326071e+00, 3.41518037198408e+00, 3.41518030177667e+00, 3.41517906210105e+00, 3.41517066312333e+00, 3.41513978769864e+00, 3.41505748596522e+00, 3.41487737475967e+00, 3.41453245606036e+00, 3.41393268900894e+00, 3.41296343878916e+00, 3.41148491216998e+00, 3.40933267258397e+00, 3.40631930558375e+00, 3.40223727667518e+00, 3.39686298639995e+00, 3.38996198134731e+00, 3.38129522481019e+00, 3.37062626879724e+00, 3.35772910344433e+00, 3.34239639554730e+00, 3.32444777129939e+00, 3.30373775643314e+00, 3.28016296672993e+00, 3.25366814893368e+00, 3.22425070981981e+00, 3.19196343966421e+00, 3.15691523209604e+00, 3.11926971821102e+00, 3.07924185892834e+00, 3.03709266435406e+00, 2.99312232074710e+00, 2.94766209444727e+00, 2.90106544042546e+00, 2.85369876709746e+00, 2.80593229858427e+00, 2.75813143402570e+00, 2.71064893695286e+00, 2.66381820407267e+00, 2.61794777097772e+00, 2.57331712104198e+00, 2.53017378085825e+00, 2.48873161701417e+00, 2.44917019850674e+00, 2.41163505794485e+00, 2.37623867192229e+00, 2.34306198386615e+00, 2.31215630754176e+00, 2.28354547214114e+00, 2.25722809667222e+00, 2.23317990900280e+00, 2.21135605100977e+00, 2.19169333425627e+00, 2.17411242956517e+00, 2.15851998839494e+00, 2.14481070402618e+00, 2.13286932642902e+00, 2.12257264663609e+00, 2.11379146491458e+00, 2.10639255250940e+00, 2.10024060979687e+00, 2.09520021500384e+00, 2.09113774796815e+00, 2.08792326357150e+00, 2.08543228034536e+00, 2.08354744220941e+00, 2.08216000614790e+00, 2.08117110652134e+00, 2.08049274807134e+00, 2.08004848466261e+00, 2.07977374924563e+00, 2.07961581194583e+00, 2.07953335684468e+00, 2.07949568297654e+00, 2.07948155029673e+00, 2.07947770587351e+00, 2.07947713844503e+00, 2.07947709543471e+00, 2.07947633289982e+00, 2.07947116661826e+00, 2.07945217489011e+00, 2.07940155044804e+00, 2.07929076279629e+00, 2.07907860154076e+00, 2.07870968372603e+00, 2.07811350140298e+00, 2.07720407750375e+00, 2.07588028778365e+00, 2.07402689304696e+00, 2.07151630798546e+00, 2.06821110976529e+00, 2.06396726048219e+00, 2.05863798294125e+00, 2.05207819003159e+00, 2.04414932645903e+00, 2.03472444105367e+00, 2.02369327245411e+00, 2.01096710532416e+00, 1.99648314289503e+00, 1.98020814819023e+00, 1.96214113281614e+00, 1.94231491850971e+00, 1.92079646000785e+00, 1.89768589309011e+00, 1.87311435180519e+00, 1.84724067592310e+00, 1.82024719579771e+00, 1.79233483070720e+00, 1.76371776434766e+00, 1.73461796629221e+00, 1.70525981251238e+00, 1.67586502536932e+00, 1.64664810912576e+00, 1.61781240675904e+00, 1.58954685301143e+00, 1.56202345151401e+00, 1.53539546346741e+00, 1.50979626341986e+00, 1.48533879467981e+00, 1.46211554249049e+00, 1.44019893637487e+00, 1.41964209280049e+00, 1.40047981416126e+00, 1.38272976865795e+00, 1.36639378669666e+00, 1.35145922174344e+00, 1.33790033616086e+00, 1.32567968456058e+00, 1.31474947796403e+00, 1.30505292107236e+00, 1.29652552187301e+00, 1.28909637747418e+00, 1.28268944241842e+00, 1.27722478586784e+00, 1.27261984218360e+00, 1.26879065584465e+00, 1.26565311678243e+00, 1.26312417653264e+00, 1.26112302968306e+00, 1.25957223951143e+00, 1.25839878204268e+00, 1.25753497953787e+00, 1.25691929308729e+00, 1.25649694479569e+00, 1.25622034311758e+00, 1.25604929012333e+00, 1.25595095655450e+00, 1.25589961899610e+00, 1.25587616276732e+00, 1.25586736357879e+00, 1.25586496999114e+00, 1.25586461670279e+00, 1.25586458975411e+00, 1.25586410973570e+00, 1.25586085754327e+00, 1.25584890218312e+00, 1.25581703391413e+00, 1.25574729267049e+00, 1.25561373629675e+00, 1.25538150082848e+00, 1.25500620076315e+00, 1.25443371212853e+00, 1.25360037466305e+00, 1.25243364092242e+00, 1.25085318892820e+00, 1.24877250048530e+00, 1.24610088920049e+00, 1.24274594064249e+00, 1.23861630268009e+00, 1.23362473816498e+00, 1.22769132678233e+00, 1.22074668060672e+00, 1.21273502149299e+00, 1.20361696064704e+00, 1.19337182381210e+00, 1.18199938078402e+00, 1.16952086546817e+00, 1.15597921096719e+00, 1.14143847037738e+00, 1.12598244405206e+00, 1.10971258339730e+00, 1.09274528509179e+00, 1.07520872387114e+00, 1.05723939372280e+00, 1.03897853503454e+00, 1.02056861905697e+00, 1.00215004259761e+00, 9.83858157935558e-01, 9.65820729023676e-01, 9.48155868842761e-01, 9.30970477756503e-01, 9.14359171744522e-01, 8.98403664462606e-01, 8.83172549261246e-01, 8.68721416744165e-01, 8.55093239601132e-01, 8.42318958199243e-01, 8.30418206385701e-01, 8.19400125690436e-01, 8.09264226269093e-01, 8.00001263363697e-01, 7.91594107910484e-01, 7.84018598592482e-01, 7.77244369764261e-01, 7.71235655102302e-01, 7.65952070535192e-01, 7.61349382063017e-01, 7.57380264618860e-01, 7.53995057341555e-01, 7.51142518723100e-01, 7.48770582304942e-01, 7.46827110184924e-01, 7.45260637853448e-01, 7.44021100116726e-01, 7.43060524412471e-01, 7.42333675006932e-01, 7.41798629670100e-01, 7.41417269707000e-01, 7.41155664825939e-01, 7.40984336307150e-01, 7.40878385229848e-01, 7.40817476939753e-01, 7.40785678210287e-01, 7.40771149310717e-01, 7.40765699051636e-01, 7.40764216452121e-01, 7.40763997623645e-01, 7.40763980813711e-01, 7.40763679845685e-01, 7.40761640744950e-01, 7.40754144822915e-01, 7.40734163662026e-01, 7.40690436471525e-01, 7.40606697852442e-01, 7.40461088775731e-01, 7.40225780918635e-01, 7.39866841308611e-01, 7.39344360165148e-01, 7.38612859479431e-01, 7.37621992779211e-01, 7.36317537300906e-01, 7.34642668220943e-01, 7.32539490753268e-01, 7.29950790252829e-01, 7.26821943867932e-01, 7.23102921092163e-01, 7.18750286470225e-01, 7.13729107603029e-01, 7.08014667300020e-01, 7.01593881718281e-01, 6.94466337412649e-01, 6.86644879297077e-01, 6.78155707438265e-01, 6.69037971225478e-01, 6.59342881864561e-01, 6.49132395027593e-01, 6.38477541637855e-01, 6.27456503566307e-01, 6.16152540819982e-01, 6.04651877176593e-01, 5.93041642940747e-01, 5.81407958334637e-01, 5.69834221401081e-01, 5.58399642809519e-01, 5.47178049028375e-01, 5.36236956859170e-01, 5.25636907540303e-01, 5.15431038032048e-01, 5.05664860607951e-01, 4.96376218982292e-01, 4.87595389147998e-01, 4.79345295073259e-01, 4.71641812653601e-01, 4.64494139230297e-01, 4.57905210101529e-01, 4.51872147464471e-01, 4.46386730927117e-01, 4.41435882004640e-01, 4.37002157780938e-01, 4.33064251126453e-01, 4.29597496474273e-01, 4.26574381139986e-01, 4.23965062516402e-01, 4.21737891189508e-01, 4.19859939152332e-01, 4.18297530916419e-01, 4.17016773558586e-01, 4.15984079747901e-01, 4.15166675766784e-01, 4.14533084672172e-01, 4.14053573249461e-01, 4.13700550481337e-01, 4.13448905040241e-01, 4.13276269914775e-01, 4.13163203728203e-01, 4.13093280556637e-01, 4.13053082989819e-01, 4.13032096618281e-01, 4.13022507858756e-01, 4.13018910798101e-01, 4.13017932311703e-01, 4.13017787889203e-01, 4.13017776710333e-01, 4.13017575459236e-01, 4.13016211954513e-01, 4.13011199581135e-01, 4.12997838543461e-01, 4.12968598794421e-01, 4.12912603304292e-01, 4.12815233333248e-01, 4.12657875441255e-01, 4.12417827670382e-01, 4.12068379485447e-01, 4.11579076417425e-01, 4.10916175663360e-01, 4.10043292942390e-01, 4.08922233579233e-01, 4.07513992142214e-01, 4.05779895285818e-01, 4.03682852275310e-01, 4.01188667794713e-01, 3.98267363053193e-01, 3.94894444998897e-01, 3.91052060670674e-01, 3.86729975186682e-01, 3.81926318016468e-01, 3.76648052917417e-01, 3.70911141571042e-01, 3.64740388339101e-01, 3.58168972098269e-01, 3.51237689077237e-01, 3.43993946374738e-01, 3.36490558077401e-01, 3.28784403802231e-01, 3.20935012790045e-01, 3.13003135609407e-01, 3.05049360724595e-01, 2.97132825495709e-01, 2.89310061564137e-01, 2.81634003939601e-01, 2.74153182252405e-01, 2.66911102226189e-01, 2.59945815986242e-01, 2.53289671735430e-01, 2.46969226872181e-01, 2.41005303945684e-01, 2.35413165980816e-01, 2.30202786587526e-01, 2.25379190725420e-01, 2.20942843765286e-01, 2.16890069266029e-01, 2.13213479317517e-01, 2.09902405049334e-01, 2.06943318646367e-01, 2.04320241673784e-01, 2.02015137472614e-01, 2.00008287690821e-01, 1.98278654568597e-01, 1.96804231373093e-01, 1.95562383404209e-01, 1.94530181350700e-01, 1.93684727589201e-01, 1.93003474450492e-01, 1.92464531714947e-01, 1.92046958841814e-01, 1.91731035884601e-01, 1.91498505876297e-01, 1.91332780828802e-01, 1.91219103478005e-01, 1.91144657559292e-01, 1.91098620690241e-01, 1.91072155788703e-01, 1.91058339225989e-01, 1.91052026438754e-01, 1.91049658312849e-01, 1.91049014126918e-01, 1.91048919046494e-01, 1.91048911627220e-01, 1.91048777288771e-01, 1.91047867127196e-01, 1.91044521293293e-01, 1.91035602653713e-01, 1.91016085064433e-01, 1.90978708980509e-01, 1.90913719275713e-01, 1.90808699092713e-01, 1.90648512267387e-01, 1.90415365201730e-01, 1.90088996784681e-01, 1.89647001856557e-01, 1.89065289584889e-01, 1.88318672860561e-01, 1.87381578453682e-01, 1.86228860380297e-01, 1.84836691116843e-01, 1.83183497537712e-01, 1.81250901492348e-01, 1.79024619613841e-01, 1.76495274097306e-01, 1.73659066507469e-01, 1.70518270634446e-01, 1.67081508123118e-01, 1.63363781755720e-01, 1.59386255143716e-01, 1.55175783094663e-01, 1.50764212720858e-01, 1.46187490007222e-01, 1.41484618696225e-01, 1.36696526858898e-01, 1.31864900669285e-01, 1.27031044413627e-01, 1.22234820863881e-01, 1.17513717489857e-01, 1.12902072588938e-01, 1.08430482499429e-01, 1.04125397906415e-01, 1.00008905021483e-01, 9.60986770633099e-02, 9.24080736188557e-02, 8.89463604151643e-02, 8.57190197517497e-02, 8.27281220384464e-02, 7.99727310822931e-02, 7.74493194033568e-02, 7.51521743586778e-02, 7.30737806942585e-02, 7.12051698925649e-02, 6.95362310230792e-02, 6.80559815196489e-02, 6.67527992929422e-02, 6.56146197980405e-02, 6.46291031378468e-02, 6.37837770457708e-02, 6.30661617338717e-02, 6.24638822055914e-02, 6.19647728107964e-02, 6.15569776697953e-02, 6.12290492203869e-02, 6.09700456629696e-02, 6.07696266093285e-02, 6.06181449013548e-02, 6.05067314603218e-02, 6.04273692544979e-02, 6.03729520922087e-02, 6.03373239981977e-02, 6.03152954014092e-02, 6.03026332117480e-02, 6.02960230069272e-02, 6.02930028866098e-02, 6.02918699573394e-02, 6.02915617752160e-02, 6.02915162882596e-02, 6.02915127086305e-02, 6.02914475066285e-02, 6.02910057547490e-02, 6.02893818457488e-02, 6.02850532359582e-02, 6.02755808663602e-02, 6.02574427780525e-02, 6.02259087281425e-02, 6.01749634349886e-02, 6.00972858136259e-02, 5.99842910488064e-02, 5.98262414003159e-02, 5.96124300341927e-02, 5.93314397421381e-02, 5.89714750305461e-02, 5.85207617277632e-02, 5.79680031165708e-02, 5.73028759706773e-02, 5.65165442595799e-02, 5.56021633304198e-02, 5.45553438160340e-02, 5.33745430704187e-02, 5.20613531925846e-02, 5.06206589998897e-02, 4.90606466331798e-02, 4.73926533813369e-02, 4.56308609549012e-02, 4.37918466400049e-02, 4.18940181696428e-02, 3.99569674240894e-02, 3.80007841056909e-02, 3.60453726295447e-02, 3.41098134557001e-02, 3.22118043341959e-02, 3.03672082858100e-02, 2.85897247686790e-02, 2.68906896694574e-02, 2.52789997200438e-02, 2.37611486378171e-02, 2.23413563291075e-02, 2.10217691098222e-02, 1.98027079607783e-02, 1.86829429612487e-02, 1.76599746942827e-02, 1.67303069990368e-02, 1.58896994132647e-02, 1.51333915510037e-02, 1.44562951766716e-02, 1.38531526725671e-02, 1.33186628731607e-02, 1.28475768665908e-02, 1.24347674058782e-02, 1.20752761266942e-02, 1.17643429427545e-02, 1.14974218791559e-02, 1.12701872948691e-02, 1.10785340025927e-02, 1.09185742644371e-02, 1.07866340617164e-02, 1.06792504284385e-02, 1.05931710232170e-02, 1.05253565110388e-02, 1.04729857561523e-02, 1.04334633150529e-02, 1.04044282894336e-02, 1.03837632783126e-02, 1.03696019780557e-02, 1.03603339303504e-02, 1.03546050126084e-02, 1.03513124927446e-02, 1.03495938020046e-02, 1.03488085867838e-02, 1.03485140365933e-02, 1.03484339128528e-02, 1.03484220867784e-02, 1.03484211478815e-02, 1.03484039414070e-02, 1.03482873655878e-02, 1.03478588292631e-02, 1.03467165755688e-02, 1.03442171206041e-02, 1.03394316651020e-02, 1.03311137841331e-02, 1.03176807894840e-02, 1.02972112317872e-02, 1.02674609878051e-02, 1.02259001682654e-02, 1.01697725847174e-02, 1.00961786342992e-02, 1.00021811430508e-02, 9.88493194783757e-03, 9.74181486911618e-03, 9.57059840312979e-03, 9.36958920544839e-03, 9.13777558839968e-03, 8.87494918320897e-03, 8.58179296650974e-03, 8.25992525760455e-03, 7.91189211655308e-03, 7.54110465102542e-03, 7.15172266862396e-03, 6.74849129168820e-03, 6.33654188392541e-03, 5.92117222824129e-03, 5.50762281520178e-03, 5.10086606664117e-03, 4.70542336971391e-03, 4.32522127475465e-03, 3.96349368824644e-03, 3.62273204963515e-03, 3.30468098502479e-03, 3.01037331905718e-03, 2.74019592347940e-03, 2.49397676620333e-03, 2.27108360147099e-03, 2.07052576575961e-03, 1.89105214985830e-03, 1.73124027409699e-03, 1.58957328673637e-03, 1.46450341327409e-03, 1.35450168808404e-03, 1.25809474279492e-03, 1.17389013746561e-03, 1.10059208975599e-03, 1.03700942700868e-03, 9.82057452603158e-04, 9.34755356695405e-04, 8.94220636041609e-04, 8.59661659845301e-04, 8.30369223791649e-04, 8.05707772883579e-04, 7.85106848780970e-04, 7.68053162531146e-04, 7.54083545032813e-04, 7.42778919729403e-04, 7.33759362049113e-04, 7.26680242195243e-04, 7.21229383529476e-04, 7.17125113624206e-04, 7.14115037226293e-04, 7.11975325430103e-04, 7.10510294979372e-04, 7.09552045961459e-04, 7.08959938705789e-04, 7.08619719787862e-04, 7.08442148152244e-04, 7.08361026411808e-04, 7.08330596808082e-04, 7.08322319411536e-04, 7.08321097690146e-04, 7.08320999805778e-04, 7.08319194799856e-04, 7.08306965705674e-04, 7.08262011855254e-04, 7.08142193289518e-04, 7.07880033758101e-04, 7.07378198612219e-04, 7.06506227572845e-04, 7.05098834117734e-04, 7.02956110359310e-04, 6.99845992485848e-04, 6.95509319569965e-04, 6.89667749341538e-04, 6.82034651305230e-04, 6.72328877371342e-04, 6.60291008102831e-04, 6.45701311668745e-04, 6.28398268318620e-04, 6.08296172032396e-04, 5.85400089818222e-04, 5.59816406877189e-04, 5.31757341178718e-04, 5.01538194782825e-04, 4.69566775379349e-04, 4.36325431536488e-04, 4.02347303435135e-04, 3.68189002102781e-04, 3.34401553999442e-04, 3.01501375219004e-04, 2.69945020605178e-04, 2.40112422986272e-04, 2.12297555947316e-04, 1.86700507100568e-04, 1.63426407866041e-04, 1.42503075139189e-04, 1.23896700657156e-04, 1.07501608653138e-04, 9.31608033596376e-05, 8.07355397114407e-05, 7.00685067869244e-05, 6.08913674398082e-05, 5.30090973143160e-05, 4.64558095814027e-05, 4.10145339770907e-05, 3.61834962063767e-05, 3.19485720439161e-05, 2.89222422225066e-05, 2.65406087436304e-05, 2.41516233216998e-05, 2.16751507053914e-05, 1.96832646963685e-05, 1.89179006468976e-05, 1.81968260619669e-05, 1.73654461121941e-05, 1.64827372657328e-05, 1.56262944699392e-05, 1.48585412773627e-05, 1.42110878016410e-05, 1.36842224094380e-05, 1.32588294263769e-05, 1.29123350660620e-05, 1.26300550959968e-05, 1.24087817283680e-05, 1.22520555407846e-05, 1.21578085838792e-05, 1.21107554349464e-05, 1.20908933496366e-05, 1.20837182606692e-05, 1.20815757895798e-05, 1.20811206159033e-05, 1.20811040460467e-05, 1.20811463517612e-05, 1.20811702686296e-05, 1.20811775161656e-05, 1.20811786246122e-05, 1.20811781556540e-05, 1.20811628505635e-05, 1.20810593102760e-05, 1.20806805057327e-05, 1.20796843816606e-05, 1.20775717533661e-05, 1.20737689286443e-05, 1.20678343051850e-05, 1.20597276946910e-05, 1.20498476974262e-05, 1.20383866857310e-05, 1.20239811280953e-05, 1.20025523813183e-05, 1.19672718028596e-05, 1.19092690340494e-05, 1.18176451735508e-05, 1.16775239007438e-05, 1.14658243148824e-05, 1.11456517970510e-05, 1.06609005712244e-05, 9.93175734720353e-06, 8.84694748037873e-06, 7.22425284038763e-06, 7.43396243763073e-06, 7.50755857836793e-06, 7.06349790391414e-06, 5.86448317108079e-06, 4.44659283971305e-06, 5.32378018647250e-06, 5.67420633501570e-06, 5.14622466596393e-06, 3.36352786689531e-06, 2.32819986933919e-06, 4.50211099689568e-06, 6.69878947689995e-06, 1.15555047983662e-05, 1.64315081008226e-05, 1.91162530143174e-05, 1.90111439337670e-05, 1.63097789283029e-05, 1.19335584400915e-05, 7.21862262764716e-06, 3.54557613491124e-06, 2.03292445632997e-06, 2.20304224285428e-06, 2.51518112293070e-06, 2.15287563776201e-06, 2.29339384816285e-06, 2.85146595047081e-06, 2.56655197006239e-06, 1.46296639513821e-06, 1.13016794500709e-06, 2.09222201874582e-06, 2.43737135008359e-06, 2.33034751317973e-06, 1.99368237474573e-06, 1.62340138842865e-06, 1.35006052914832e-06, 1.21074795345066e-06, 1.16045505649254e-06, 1.12993593859301e-06, 1.07260088100247e-06, 9.73727495079402e-07, 8.40410720570617e-07, 6.89080323468261e-07, 5.36800058625886e-07, 3.96859832108021e-07, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, 0.00000000000000e+00, -1.42108547152020e-14],
[ 8.40000000000000e+01, 8.39999999023545e+01, 8.39999968829621e+01, 8.39999764260989e+01, 8.39999012239797e+01, 8.39997007586976e+01, 8.39992620248174e+01, 8.39984217183486e+01, 8.39969601724104e+01, 8.39945972654372e+01, 8.39909904663564e+01, 8.39857351154527e+01, 8.39783669727587e+01, 8.39683670011970e+01, 8.39551682927026e+01, 8.39381649950321e+01, 8.39167230567291e+01, 8.38901925779650e+01, 8.38579215339320e+01, 8.38192706212703e+01, 8.37736289611675e+01, 8.37204303691309e+01, 8.36591698657399e+01, 8.35894200523915e+01, 8.35108469129532e+01, 8.34232245334948e+01, 8.33264481701898e+01, 8.32205450559127e+01, 8.31056823354746e+01, 8.29821715711947e+01, 8.28504693712258e+01, 8.27111738601072e+01, 8.25650169218319e+01, 8.24128523793355e+01, 8.22556405049234e+01, 8.20944294577124e+01, 8.19303343951020e+01, 8.17645150921182e+01, 8.15981529215001e+01, 8.14324280044315e+01, 8.12684972500037e+01, 8.11074738781229e+01, 8.09504088836373e+01, 8.07982747647057e+01, 8.06519517174902e+01, 8.05122163988665e+01, 8.03797332811473e+01, 8.02550485663282e+01, 8.01385865883070e+01, 8.00306486050757e+01, 7.99314138642399e+01, 7.98409428102850e+01, 7.97591822877393e+01, 7.96859725788684e+01, 7.96210560968735e+01, 7.95640875357001e+01, 7.95146452560419e+01, 7.94722436649441e+01, 7.94363463248405e+01, 7.94063795083636e+01, 7.93817458993633e+01, 7.93618381298173e+01, 7.93460518381420e+01, 7.93337979381077e+01, 7.93245138001057e+01, 7.93176730685371e+01, 7.93127938707290e+01, 7.93094452136868e+01, 7.93072514142195e+01, 7.93058944641041e+01, 7.93051142930688e+01, 7.93047069562270e+01, 7.93045208367204e+01, 7.93044510162306e+01, 7.93044320233043e+01, 7.93044292199859e+01, 7.93044288522497e+01, 7.93044202844367e+01, 7.93043622362279e+01, 7.93041488446462e+01, 7.93035800193020e+01, 7.93023351501499e+01, 7.92999510430807e+01, 7.92958049553774e+01, 7.92891034921467e+01, 7.92788779966377e+01, 7.92639869200592e+01, 7.92431254872614e+01, 7.92148427804647e+01, 7.91775661417953e+01, 7.91296325462240e+01, 7.90693263218058e+01, 7.89949222998375e+01, 7.89047331743694e+01, 7.87971595549412e+01, 7.86707409317675e+01, 7.85242055691561e+01, 7.83565172363191e+01, 7.81669167109823e+01, 7.79549561780555e+01, 7.77205250007709e+01, 7.74638658413979e+01, 7.71855806923354e+01, 7.68866269549044e+01, 7.65683041716565e+01, 7.62322323004405e+01, 7.58803224938025e+01, 7.55147412719662e+01, 7.51378688772544e+01, 7.47522526225990e+01, 7.43605562994461e+01, 7.39655071751756e+01, 7.35698426268049e+01, 7.31762587746512e+01, 7.27873633794940e+01, 7.24056346986094e+01, 7.20333871227510e+01, 7.16727435343401e+01, 7.13256137093950e+01, 7.09936778530413e+01, 7.06783744573491e+01, 7.03808919446413e+01, 7.01021638486996e+01, 6.98428674891702e+01, 6.96034261811029e+01, 6.93840150115670e+01, 6.91845701476837e+01, 6.90048015515418e+01, 6.88442088912136e+01, 6.87021003642379e+01, 6.85776140918017e+01, 6.84697416948985e+01, 6.83773536232117e+01, 6.82992257696224e+01, 6.82340668664063e+01, 6.81805461238979e+01, 6.81373205410915e+01, 6.81030612940611e+01, 6.80764785965806e+01, 6.80563444322354e+01, 6.80415125823161e+01, 6.80309354213682e+01, 6.80236770233935e+01, 6.80189222155423e+01, 6.80159813300804e+01, 6.80142905352090e+01, 6.80134077653498e+01, 6.80130044153584e+01, 6.80128531040180e+01, 6.80128119435859e+01, 6.80128058683895e+01, 6.80128053320653e+01, 6.80127948230565e+01, 6.80127236229615e+01, 6.80124618836701e+01, 6.80117641811992e+01, 6.80102372681986e+01, 6.80073130156768e+01, 6.80022276175330e+01, 6.79940079976377e+01, 6.79814662041723e+01, 6.79632023968594e+01, 6.79376168253933e+01, 6.79029309574659e+01, 6.78572176392536e+01, 6.77984398601674e+01, 6.77244973518851e+01, 6.76332798901322e+01, 6.75227258039870e+01, 6.73908838554276e+01, 6.72359763593571e+01, 6.70564612001788e+01, 6.68510902904526e+01, 6.66189620280775e+01, 6.63595654484904e+01, 6.60728140350513e+01, 6.57590675338260e+01, 6.54191406040861e+01, 6.50542977069681e+01, 6.46662342726412e+01, 6.42570448632489e+01, 6.38291797218290e+01, 6.33853917058739e+01, 6.29286760764134e+01, 6.24622058825172e+01, 6.19892657054116e+01, 6.15131863068524e+01, 6.10372823110463e+01, 6.05647945226548e+01, 6.00988379428191e+01, 5.96423560745900e+01, 5.91980817583366e+01, 5.87685045550034e+01, 5.83558445747773e+01, 5.79620325881021e+01, 5.75886962141536e+01, 5.72371519332903e+01, 5.69084026084131e+01, 5.66031401343655e+01, 5.63217527792499e+01, 5.60643367489572e+01, 5.58307115007630e+01, 5.56204383496692e+01, 5.54328419426670e+01, 5.52670342096464e+01, 5.51219404249167e+01, 5.49963270231096e+01, 5.48888308043514e+01, 5.47979891364021e+01, 5.47222707192665e+01, 5.46601064260318e+01, 5.46099196791534e+01, 5.45701557715965e+01, 5.45393095046620e+01, 5.45159504959640e+01, 5.44987455177099e+01, 5.44864772612884e+01, 5.44780589911660e+01, 5.44725446487922e+01, 5.44691340926468e+01, 5.44671733084826e+01, 5.44661495870506e+01, 5.44656818367587e+01, 5.44655063669071e+01, 5.44654586347992e+01, 5.44654515896382e+01, 5.44654510245291e+01, 5.44654405954005e+01, 5.44653699365142e+01, 5.44651101868921e+01, 5.44644177891405e+01, 5.44629024904189e+01, 5.44600004999121e+01, 5.44549538759589e+01, 5.44467970876279e+01, 5.44343515465142e+01, 5.44162287312333e+01, 5.43908423243671e+01, 5.43564295441068e+01, 5.43110815765172e+01, 5.42527826983100e+01, 5.41794573281938e+01, 5.40890238673812e+01, 5.39794538038049e+01, 5.38488341839562e+01, 5.36954312306082e+01, 5.35177526369028e+01, 5.33146059302888e+01, 5.30851503025705e+01, 5.28289394657310e+01, 5.25459534259062e+01, 5.22366175626573e+01, 5.19018080314482e+01, 5.15428432275959e+01, 5.11614617949232e+01, 5.07597883552009e+01, 5.03402887000488e+01, 4.99057165696208e+01, 4.94590543222892e+01, 4.90034497991096e+01, 4.85421515612708e+01, 4.80784444928877e+01, 4.76155875637796e+01, 4.71567553535124e+01, 4.67049847350931e+01, 4.62631278773284e+01, 4.58338124301743e+01, 4.54194094108245e+01, 4.50220089370046e+01, 4.46434035994435e+01, 4.42850789694403e+01, 4.39482105294293e+01, 4.36336662056504e+01, 4.33420136655492e+01, 4.30735315982441e+01, 4.28282242978993e+01, 4.26058389907352e+01, 4.24058854646355e+01, 4.22276576602412e+01, 4.20702569549415e+01, 4.19326169127131e+01, 4.18135292838458e+01, 4.17116710224795e+01, 4.16256320515563e+01, 4.15539434503083e+01, 4.14951056755052e+01, 4.14476163615644e+01, 4.14099971836855e+01, 4.13808192197080e+01, 4.13587262171867e+01, 4.13424551679778e+01, 4.13308536176193e+01, 4.13228931931213e+01, 4.13176789202203e+01, 4.13144540169680e+01, 4.13125999896167e+01, 4.13116320120661e+01, 4.13111897332589e+01, 4.13110238189066e+01, 4.13109786861515e+01, 4.13109720246503e+01, 4.13109715108512e+01, 4.13109622846692e+01, 4.13108997759288e+01, 4.13106699871948e+01, 4.13100574548581e+01, 4.13087169431308e+01, 4.13061497039901e+01, 4.13016852572767e+01, 4.12944695343449e+01, 4.12834600997172e+01, 4.12674290143016e+01, 4.12449737255647e+01, 4.12145361590362e+01, 4.11744299373045e+01, 4.11228753659047e+01, 4.10580415034955e+01, 4.09780942858845e+01, 4.08812493161084e+01, 4.07658275891377e+01, 4.06303121188706e+01, 4.04734032094226e+01, 4.02940699948250e+01, 4.00915958893008e+01, 3.98656157632507e+01, 3.96161429930951e+01, 3.93435850138196e+01, 3.90487465999049e+01, 3.87328207637492e+01, 3.83973678278688e+01, 3.80442838304662e+01, 3.76757599022141e+01, 3.72942345628086e+01, 3.69023410147839e+01, 3.65028514772458e+01, 3.60986204489828e+01, 3.56925285783589e+01, 3.52874286014340e+01, 3.48860946226476e+01, 3.44911758554791e+01, 3.41051557875569e+01, 3.37303175484837e+01, 3.33687160125971e+01, 3.30221568639520e+01, 3.26921825185224e+01, 3.23800644894219e+01, 3.20868015440386e+01, 3.18131228674237e+01, 3.15594954170648e+01, 3.13261347099901e+01, 3.11130183913684e+01, 3.09199020612056e+01, 3.07463369566738e+01, 3.05916891864556e+01, 3.04551602836037e+01, 3.03358088841711e+01, 3.02325733527322e+01, 3.01442951666216e+01, 3.00697428423839e+01, 3.00076361450136e+01, 2.99566702680049e+01, 2.99155396156624e+01, 2.98829607648129e+01, 2.98576941376090e+01, 2.98385638870035e+01, 2.98244754874167e+01, 2.98144305394480e+01, 2.98075383416564e+01, 2.98030238546160e+01, 2.98002317804486e+01, 2.97986266004063e+01, 2.97977885475137e+01, 2.97974056332549e+01, 2.97972619887397e+01, 2.97972229139339e+01, 2.97972171465739e+01, 2.97972167098868e+01, 2.97972089740185e+01, 2.97971565623751e+01, 2.97969638916684e+01, 2.97964503026002e+01, 2.97953263269921e+01, 2.97931737850104e+01, 2.97894305145059e+01, 2.97833804427198e+01, 2.97741496037756e+01, 2.97607085801314e+01, 2.97418816978958e+01, 2.97163631298052e+01, 2.96827398520183e+01, 2.96395211605862e+01, 2.95851741833699e+01, 2.95181645313492e+01, 2.94370009335683e+01, 2.93402824120006e+01, 2.92267463006874e+01, 2.90953152245186e+01, 2.89451410533322e+01, 2.87756438585251e+01, 2.85865440353479e+01, 2.83778860153235e+01, 2.81500523658789e+01, 2.79037675298570e+01, 2.76400909565179e+01, 2.73603998739110e+01, 2.70663624101496e+01, 2.67599021601249e+01, 2.64431556025853e+01, 2.61184240027050e+01, 2.57881215969373e+01, 2.54547219575857e+01, 2.51207044723272e+01, 2.47885028367383e+01, 2.44604573289194e+01, 2.41387724029896e+01, 2.38254808053214e+01, 2.35224150056233e+01, 2.32311862828635e+01, 2.29531713612600e+01, 2.26895061008099e+01, 2.24410854464361e+01, 2.22085686500042e+01, 2.19923887027648e+01, 2.17927649396410e+01, 2.16097178783375e+01, 2.14430855079770e+01, 2.12925404166501e+01, 2.11576073215416e+01, 2.10376807215140e+01, 2.09320425186903e+01, 2.08398795467916e+01, 2.07603009983872e+01, 2.06923557628295e+01, 2.06350496757995e+01, 2.05873626459671e+01, 2.05482655710874e+01, 2.05167368923171e+01, 2.04917785692552e+01, 2.04724311967322e+01, 2.04577879347296e+01, 2.04470068911324e+01, 2.04393215880154e+01, 2.04340491587310e+01, 2.04305959659267e+01, 2.04284603982276e+01, 2.04272326919410e+01, 2.04265917282138e+01, 2.04262988687843e+01, 2.04261890074561e+01, 2.04261591225446e+01, 2.04261547115955e+01, 2.04261543804641e+01, 2.04261485521777e+01, 2.04261090647000e+01, 2.04259639048592e+01, 2.04255769643570e+01, 2.04247301663956e+01, 2.04231084924578e+01, 2.04202885301515e+01, 2.04157311034241e+01, 2.04087784761608e+01, 2.03986565347549e+01, 2.03844822488777e+01, 2.03652765762475e+01, 2.03399828116183e+01, 2.03074901799095e+01, 2.02666622397832e+01, 2.02163694036154e+01, 2.01555246051730e+01, 2.00831208757601e+01, 1.99982693466156e+01, 1.99002360066961e+01, 1.97884754381199e+01, 1.96626597512328e+01, 1.95227010657275e+01, 1.93687661415632e+01, 1.92012821482232e+01, 1.90209330524926e+01, 1.88286466674146e+01, 1.86255729893825e+01, 1.84130549993390e+01, 1.81925935599457e+01, 1.79658083534751e+01, 1.77343969417715e+01, 1.75000939796970e+01, 1.72646323924700e+01, 1.70297079739510e+01, 1.67969484320521e+01, 1.65678874595639e+01, 1.63439439987784e+01, 1.61264065366777e+01, 1.59164220348010e+01, 1.57149889650527e+01, 1.55229538750068e+01, 1.53410109202477e+01, 1.51697038518635e+01, 1.50094300132038e+01, 1.48604459671268e+01, 1.47228744360412e+01, 1.45967122904249e+01, 1.44818393684678e+01, 1.43780279519256e+01, 1.42849527621591e+01, 1.42022013751537e+01, 1.41292849833039e+01, 1.40656494525937e+01, 1.40106866343694e+01, 1.39637458898617e+01, 1.39241457728113e+01, 1.38911857920941e+01, 1.38641581444633e+01, 1.38423592707222e+01, 1.38251010508343e+01, 1.38117214190167e+01, 1.38015941530623e+01, 1.37941375768771e+01, 1.37888219145367e+01, 1.37851750499654e+01, 1.37827864790996e+01, 1.37813092901281e+01, 1.37804600696624e+01, 1.37800167048349e+01, 1.37798141288767e+01, 1.37797381358179e+01, 1.37797174638800e+01, 1.37797144127459e+01, 1.37797141843928e+01, 1.37797101743984e+01, 1.37796830061068e+01, 1.37795831327574e+01, 1.37793169083390e+01, 1.37787342891180e+01, 1.37776185281605e+01, 1.37756782873149e+01, 1.37725425485227e+01, 1.37677586562179e+01, 1.37607937621553e+01, 1.37510398695914e+01, 1.37378225814210e+01, 1.37204135434611e+01, 1.36980464381014e+01, 1.36699362262734e+01, 1.36353011619712e+01, 1.35933869223040e+01, 1.35434920201178e+01, 1.34849935116609e+01, 1.34173718962261e+01, 1.33402340452736e+01, 1.32533330091717e+01, 1.31565836387614e+01, 1.30500731274249e+01, 1.29340658200114e+01, 1.28090019329751e+01, 1.26754901644096e+01, 1.25342945184093e+01, 1.23863159986632e+01, 1.22325701146858e+01, 1.20741613654318e+01, 1.19122559973789e+01, 1.17480543619975e+01, 1.15827641156591e+01, 1.14175753225342e+01, 1.12536382626820e+01, 1.10920444510650e+01, 1.09338110816495e+01, 1.07798688623697e+01, 1.06310530262092e+01, 1.04880971979289e+01, 1.03516297560153e+01, 1.02221723364196e+01, 1.01001401575522e+01, 9.98584388743963e+00, 9.87949281322904e+00, 9.78119910607190e+00, 9.69098300114052e+00, 9.60877873544413e+00, 9.53444110756180e+00, 9.46775254460320e+00, 9.40843058240254e+00, 9.35613568376566e+00, 9.31047933454627e+00, 9.27103236648088e+00, 9.23733345761840e+00, 9.20889775524043e+00, 9.18522555274447e+00, 9.16581093239722e+00, 9.15015026232395e+00, 9.13775041137797e+00, 9.12813652275538e+00, 9.12085916953492e+00, 9.11550070557595e+00, 9.11168062564377e+00, 9.10905976063795e+00, 9.10734315774569e+00, 9.10628153051712e+00, 9.10567120852406e+00, 9.10535256783233e+00, 9.10520697881456e+00, 9.10515236343606e+00, 9.10513750673721e+00, 9.10513531391989e+00, 9.10513514974356e+00, 9.10513226590410e+00, 9.10511272747391e+00, 9.10504090223299e+00, 9.10484944324327e+00, 9.10443044364125e+00, 9.10362802351059e+00, 9.10223265271226e+00, 9.09997748213870e+00, 9.09653691901809e+00, 9.09152763780985e+00, 9.08451216449006e+00, 9.07500510609979e+00, 9.06248201728403e+00, 9.04639080032209e+00, 9.02616542581239e+00, 9.00124164107631e+00, 8.97107420850686e+00, 8.93515509512567e+00, 8.89303192818966e+00, 8.84432595149651e+00, 8.78874867432614e+00, 8.72611640886652e+00, 8.65636194792464e+00, 8.57954274320090e+00, 8.49584510047428e+00, 8.40558410171145e+00, 8.30919918151876e+00, 8.20724551048473e+00, 8.10038155448474e+00, 7.98935337185635e+00, 7.87497636673955e+00, 7.75811532695718e+00, 7.63966363196214e+00, 7.52052251752269e+00, 7.40158122981373e+00, 7.28369879760656e+00, 7.16768800683310e+00, 7.05430199026387e+00, 6.94422366215747e+00, 6.83805804981930e+00, 6.73632741575527e+00, 6.63946893675731e+00, 6.54783461641750e+00, 6.46169305690964e+00, 6.38123270168597e+00, 6.30656617704844e+00, 6.23773539954068e+00, 6.17471716942524e+00, 6.11742903036799e+00, 6.06573523529241e+00, 6.01945271319319e+00, 5.97835697809082e+00, 5.94218795725786e+00, 5.91065574049912e+00, 5.88344626566604e+00, 5.86022695845266e+00, 5.84065233809687e+00, 5.82436958654569e+00, 5.81102405892582e+00, 5.80026469004122e+00, 5.79174922753924e+00, 5.78514919985527e+00, 5.78015450849287e+00, 5.77647752179517e+00, 5.77385654283577e+00, 5.77205852850102e+00, 5.77088095059061e+00, 5.77015271234946e+00, 5.76973406399294e+00, 5.76951549655561e+00, 5.76941563236915e+00, 5.76937817003997e+00, 5.76936797939463e+00, 5.76936647527732e+00, 5.76936636236525e+00, 5.76936437501830e+00, 5.76935091045932e+00, 5.76930141345272e+00, 5.76916947365420e+00, 5.76888073195127e+00, 5.76832777742996e+00, 5.76736624866119e+00, 5.76581233379800e+00, 5.76344184118122e+00, 5.75999098400887e+00, 5.75515898683555e+00, 5.74861257581418e+00, 5.73999235663480e+00, 5.72892101299198e+00, 5.71501317460222e+00, 5.69788670980664e+00, 5.67717509846354e+00, 5.65254044330965e+00, 5.62368659133416e+00, 5.59037177120631e+00, 5.55242011863986e+00, 5.50973146751652e+00, 5.46228883636670e+00, 5.41016313879270e+00, 5.35351478867942e+00, 5.29259204709496e+00, 5.22772615323669e+00, 5.15932347880697e+00, 5.08785512479243e+00, 5.01384452419856e+00, 4.93785371028653e+00, 4.86046894971216e+00, 4.78228642313884e+00, 4.70389856867011e+00, 4.62588159755737e+00, 4.54878456204334e+00, 4.47312021751478e+00, 4.39935778928923e+00, 4.32791763891511e+00, 4.25916773227064e+00, 4.19342174446527e+00, 4.13093859382380e+00, 4.07192317620225e+00, 4.01652806763160e+00, 3.96485597368832e+00, 3.91696272421792e+00, 3.87286063875659e+00, 3.83252211837853e+00, 3.79588335134957e+00, 3.76284805084290e+00, 3.73329117134038e+00, 3.70706257477718e+00, 3.68399063690937e+00, 3.66388579806855e+00, 3.64654407009144e+00, 3.63175051282296e+00, 3.61928268962812e+00, 3.60891410258593e+00, 3.60041759557470e+00, 3.59356869863814e+00, 3.58814887138483e+00, 3.58394858834534e+00, 3.58077019679983e+00, 3.57843046905140e+00, 3.57676276762105e+00, 3.57561874414753e+00, 3.57486950114594e+00, 3.57440615993754e+00, 3.57413979720124e+00, 3.57400073545924e+00, 3.57393719782532e+00, 3.57391336279866e+00, 3.57390687910609e+00, 3.57390592212711e+00, 3.57390584998201e+00, 3.57390457609249e+00, 3.57389594531040e+00, 3.57386421774548e+00, 3.57377964449771e+00, 3.57359456202987e+00, 3.57324012235240e+00, 3.57262379693900e+00, 3.57162778051693e+00, 3.57010840683010e+00, 3.56789667090922e+00, 3.56479992968721e+00, 3.56060482326144e+00, 3.55508142128544e+00, 3.54798855209989e+00, 3.53908021659278e+00, 3.52811292613468e+00, 3.51485373761463e+00, 3.49908869355481e+00, 3.48063131777980e+00, 3.45933077415431e+00, 3.43507927434324e+00, 3.40781832609018e+00, 3.37754344964773e+00, 3.34430705715404e+00, 3.30821928483273e+00, 3.26944668429522e+00, 3.22820880745537e+00, 3.18477284831789e+00, 3.13944662243232e+00, 3.09257026038070e+00, 3.04450705685007e+00, 2.99563394644368e+00, 2.94633207001137e+00, 2.89697785340644e+00, 2.84793495019565e+00, 2.79954730967098e+00, 2.75213353181852e+00, 2.70598257215735e+00, 2.66135077082103e+00, 2.61846010880176e+00, 2.57749754375605e+00, 2.53861524889757e+00, 2.50193156934692e+00, 2.46753251712263e+00, 2.43547364414970e+00, 2.40578215764867e+00, 2.37845917007829e+00, 2.35348200339212e+00, 2.33080649265487e+00, 2.31036925581601e+00, 2.29208991406844e+00, 2.27587326056248e+00, 2.26161138439731e+00, 2.24918576199767e+00, 2.23846932951410e+00, 2.22932854812116e+00, 2.22162546946151e+00, 2.21521980151797e+00, 2.20997096650744e+00, 2.20574013272353e+00, 2.20239219243020e+00, 2.19979764882260e+00, 2.19783436761156e+00, 2.19638914377467e+00, 2.19535903212019e+00, 2.19465239196462e+00, 2.19418960158258e+00, 2.19390340697636e+00, 2.19373888143863e+00, 2.19365298658434e+00, 2.19361374104197e+00, 2.19359901876919e+00, 2.19359501396205e+00, 2.19359442286121e+00, 2.19359437805659e+00, 2.19359358370976e+00, 2.19358820189761e+00, 2.19356841785441e+00, 2.19351568137816e+00, 2.19340027152396e+00, 2.19317925794000e+00, 2.19279494504531e+00, 2.19217387801483e+00, 2.19122648088145e+00, 2.18984738617269e+00, 2.18791650131311e+00, 2.18530083842063e+00, 2.18185711015091e+00, 2.17743506438512e+00, 2.17188149501364e+00, 2.16504482599531e+00, 2.15678012351698e+00, 2.14695434978000e+00, 2.13545163592606e+00, 2.12217832557459e+00, 2.10706752891807e+00, 2.09008293394952e+00, 2.07122164816647e+00, 2.05051589074367e+00, 2.02803341888435e+00, 2.00387664761936e+00, 1.97818050267246e+00, 1.95110912324041e+00, 1.92285159806795e+00, 1.89361696792133e+00, 1.86362875670609e+00, 1.83311930098503e+00, 1.80232413508648e+00, 1.77147665993708e+00, 1.74080328299821e+00, 1.71051916936050e+00, 1.68082469491172e+00, 1.65190264546278e+00, 1.62391616369055e+00, 1.59700741062508e+00, 1.57129688117601e+00, 1.54688329414894e+00, 1.52384396607620e+00, 1.50223557427172e+00, 1.48209521678913e+00, 1.46344168417242e+00, 1.44627686867634e+00, 1.43058724963316e+00, 1.41634540756237e+00, 1.40351153333063e+00, 1.39203491124471e+00, 1.38185536570052e+00, 1.37290466944272e+00, 1.36510791735118e+00, 1.35838487288851e+00, 1.35265129500634e+00, 1.34782025165002e+00, 1.34380342237642e+00, 1.34051238747386e+00, 1.33785989489382e+00, 1.33576108989227e+00, 1.33413468617815e+00, 1.33290405221750e+00, 1.33199818272104e+00, 1.33135252371839e+00, 1.33090962029864e+00, 1.33061955918491e+00, 1.33044018370485e+00, 1.33033706610193e+00, 1.33028323101158e+00, 1.33025863367756e+00, 1.33024940642713e+00, 1.33024689639684e+00, 1.33024652592184e+00, 1.33024649766217e+00, 1.33024599429196e+00, 1.33024258388802e+00, 1.33023004692994e+00, 1.33019662836797e+00, 1.33012349451619e+00, 1.32998344154417e+00, 1.32973991047834e+00, 1.32934635920725e+00, 1.32874603668452e+00, 1.32787219779445e+00, 1.32664878849174e+00, 1.32499161910050e+00, 1.32281002843364e+00, 1.32000902231754e+00, 1.31649184729228e+00, 1.31216293441707e+00, 1.30693112065022e+00, 1.30071302829251e+00, 1.29343645914597e+00, 1.28504364232717e+00, 1.27549416601545e+00, 1.26476742623518e+00, 1.25286444154870e+00, 1.23980891141004e+00, 1.22564743650412e+00, 1.21044886873349e+00, 1.19430281241158e+00, 1.17731735168891e+00, 1.15961612714867e+00, 1.14133492227354e+00, 1.12261794468063e+00, 1.10361399580843e+00, 1.08447271607562e+00, 1.06534107207229e+00, 1.04636022114730e+00, 1.02766285076311e+00, 1.00937104946399e+00, 9.91594727250700e-01, 9.74430568864719e-01, 9.57961476235965e-01, 9.42256437286431e-01, 9.27370747507865e-01, 9.13346507476277e-01, 9.00213322386108e-01, 8.87989137145198e-01, 8.76681150891656e-01, 8.66286766459979e-01, 8.56794542089588e-01, 8.48185123613121e-01, 8.40432144868345e-01, 8.33503091810655e-01, 8.27360131621560e-01, 8.21960912035351e-01, 8.17259338244895e-01, 8.13206335273506e-01, 8.09750602814233e-01, 8.06839367469451e-01, 8.04419134325187e-01, 8.02436436140496e-01, 8.00838574424318e-01, 7.99574342637982e-01, 7.98594718041755e-01, 7.97853505633284e-01, 7.97307915519596e-01, 7.96919054170672e-01, 7.96652310493130e-01, 7.96477619593858e-01, 7.96369590412979e-01, 7.96307487902567e-01, 7.96275065822556e-01, 7.96260252140939e-01, 7.96254695055481e-01, 7.96253183397084e-01, 7.96252960279602e-01, 7.96252943140189e-01, 7.96252636273238e-01, 7.96250557206605e-01, 7.96242914369245e-01, 7.96222541613133e-01, 7.96177957563373e-01, 7.96092578482558e-01, 7.95944118275472e-01, 7.95704206652886e-01, 7.95338253321134e-01, 7.94805582018256e-01, 7.94059852809028e-01, 7.93049783792995e-01, 7.91720173845476e-01, 7.90013215967704e-01, 7.87870076333377e-01, 7.85232697665990e-01, 7.82045768112809e-01, 7.78258779723998e-01, 7.73828085811260e-01, 7.68718855869423e-01, 7.62906822352647e-01, 7.56379716972006e-01, 7.49138306168589e-01, 7.41196955879090e-01, 7.32583683383751e-01, 7.23339686580957e-01, 7.13518375309476e-01, 7.03183961792348e-01, 6.92409694485221e-01, 6.81275838819587e-01, 6.69867517860425e-01, 6.58272525374841e-01, 6.46579214148773e-01, 6.34874545562667e-01, 6.23242365062083e-01, 6.11761945040655e-01, 6.00506814348407e-01, 5.89543874139679e-01, 5.78932784336558e-01, 5.68725594121986e-01, 5.58966583465818e-01, 5.49692280159547e-01, 5.40931617367363e-01, 5.32706199404572e-01, 5.25030647500742e-01, 5.17913002006139e-01, 5.11355162336358e-01, 5.05353350558508e-01, 4.99898588675023e-01, 4.94977183224805e-01, 4.90571213724905e-01, 4.86659023691075e-01, 4.83215714487514e-01, 4.80213643076922e-01, 4.77622924887513e-01, 4.75411942526520e-01, 4.73547860011124e-01, 4.71997140655432e-01, 4.70726064865573e-01, 4.69701242018431e-01, 4.68890108508702e-01, 4.68261402148002e-01, 4.67785601585859e-01, 4.67435318478664e-01, 4.67185629908528e-01, 4.67014339144654e-01, 4.66902154273996e-01, 4.66832776462425e-01, 4.66792892527437e-01, 4.66772069930116e-01, 4.66762556007296e-01, 4.66758987021564e-01, 4.66758016172349e-01, 4.66757872877082e-01, 4.66757861785453e-01, 4.66757662105167e-01, 4.66756309242797e-01, 4.66751335992669e-01, 4.66738079248557e-01, 4.66709067771276e-01, 4.66653509555201e-01, 4.66556900339853e-01, 4.66400772922242e-01, 4.66162604674508e-01, 4.65815897864161e-01, 4.65330443765069e-01, 4.64672776868028e-01, 4.63806819528259e-01, 4.62694710005749e-01, 4.61297798121137e-01, 4.59577782950831e-01, 4.57497956681337e-01, 4.55024508746192e-01, 4.52127835714734e-01, 4.48783796220653e-01, 4.44974847592448e-01, 4.40691002624007e-01, 4.35930551531072e-01, 4.30700505457907e-01, 4.25016733199257e-01, 4.18903780798431e-01, 4.12394382693307e-01, 4.05528691262077e-01, 3.98353267250940e-01, 3.90919885285541e-01, 3.83284215671239e-01, 3.75504445770218e-01, 3.67639901777025e-01, 3.59749725497096e-01, 3.51891651818349e-01, 3.44120922079668e-01, 3.36489357521899e-01, 3.29044606323251e-01, 3.21829568017463e-01, 3.14881990797574e-01, 3.08234230556280e-01, 3.01913155593169e-01, 2.95940177698668e-01, 2.90331388688529e-01, 2.85097781232267e-01, 2.80245533768735e-01, 2.75776341177107e-01, 2.71687775404246e-01, 2.67973663166515e-01, 2.64624470894532e-01, 2.61627690041053e-01, 2.58968218529930e-01, 2.56628736335472e-01, 2.54590074836443e-01, 2.52831580625070e-01, 2.51331474853330e-01, 2.50067208994096e-01, 2.49015817156561e-01, 2.48154263930346e-01, 2.47459785283652e-01, 2.46910218468770e-01, 2.46484315370225e-01, 2.46162032439443e-01, 2.45924789451578e-01, 2.45755688923722e-01, 2.45639688229161e-01, 2.45563717261660e-01, 2.45516735917207e-01, 2.45489727588577e-01, 2.45475627178109e-01, 2.45469184668835e-01, 2.45466767874916e-01, 2.45466110449641e-01, 2.45466013415097e-01, 2.45466005843326e-01, 2.45465868743949e-01, 2.45464939876243e-01, 2.45461525273356e-01, 2.45452423295902e-01, 2.45432504349381e-01, 2.45394359132533e-01, 2.45328030378374e-01, 2.45220841850966e-01, 2.45057336921846e-01, 2.44819337227341e-01, 2.44486129506924e-01, 2.44034785550610e-01, 2.43440616063353e-01, 2.42677754100669e-01, 2.41719857567375e-01, 2.40540913294240e-01, 2.39116117789251e-01, 2.37422802456805e-01, 2.35441364597406e-01, 2.33156160625299e-01, 2.30556315449562e-01, 2.27636402493900e-01, 2.24396952780552e-01, 2.20844758924784e-01, 2.16992950439234e-01, 2.12860829694066e-01, 2.08473472168663e-01, 2.03861108997700e-01, 1.99058322962941e-01, 1.94103099820992e-01, 1.89035784278246e-01, 1.83897993475469e-01, 1.78731540403433e-01, 1.73577415516533e-01, 1.68474867570168e-01, 1.63460615247772e-01, 1.58568210465290e-01, 1.53827563351186e-01, 1.49264628707456e-01, 1.44901244990294e-01, 1.40755109998210e-01, 1.36839872756781e-01, 1.33165318546176e-01, 1.29737623437563e-01, 1.26559655770421e-01, 1.23631304315339e-01, 1.20949816015894e-01, 1.18510129788987e-01, 1.16305196546392e-01, 1.14326279100126e-01, 1.12563228731744e-01, 1.11004737806724e-01, 1.09638569829642e-01, 1.08451769736789e-01, 1.07430858024813e-01, 1.06562012551720e-01, 1.05831241584326e-01, 1.05224550981546e-01, 1.04728107385952e-01, 1.04328398061753e-01, 1.04012386675393e-01, 1.03767662993135e-01, 1.03582583290645e-01, 1.03446397340605e-01, 1.03349357264250e-01, 1.03282803361395e-01, 1.03239222302037e-01, 1.03212273757194e-01, 1.03196782617786e-01, 1.03188695310465e-01, 1.03185000258819e-01, 1.03183614134466e-01, 1.03183237076664e-01, 1.03183181423788e-01, 1.03183177044144e-01, 1.03183097270104e-01, 1.03182556790102e-01, 1.03180569942992e-01, 1.03175273849132e-01, 1.03163684041007e-01, 1.03141490331993e-01, 1.03102902045080e-01, 1.03040551261316e-01, 1.02945462161998e-01, 1.02807093691226e-01, 1.02613461484268e-01, 1.02351343109106e-01, 1.02006567992944e-01, 1.01564389855753e-01, 1.01009935052900e-01, 1.00328715084316e-01, 9.95071859568810e-02, 9.85333315533836e-02, 9.73972432598060e-02, 9.60916645003920e-02, 9.46124671889521e-02, 9.29590279673391e-02, 9.11344757938508e-02, 8.91457889730844e-02, 8.70037287357093e-02, 8.47226072688727e-02, 8.23198996576906e-02, 7.98157203504246e-02, 7.72321943098007e-02, 7.45927599381560e-02, 7.19214444547731e-02, 6.92421523812214e-02, 6.65780043259900e-02, 6.39507569384392e-02, 6.13803265915038e-02, 5.88844300698526e-02, 5.64783462930336e-02, 5.41747947716829e-02, 5.19839197391576e-02, 4.99133641096847e-02, 4.79684146799428e-02, 4.61521991619946e-02, 4.44659163729000e-02, 4.29090828012164e-02, 4.14797814057408e-02, 4.01749015113495e-02, 3.89903617408009e-02, 3.79213108210195e-02, 3.69623036540361e-02, 3.61074521510432e-02, 3.53505519623667e-02, 3.46851874160450e-02, 3.41048177463250e-02, 3.36028481037260e-02, 3.31726889390183e-02, 3.28078071939473e-02, 3.25017723550900e-02, 3.22482998813456e-02, 3.20412938426813e-02, 3.18748898571073e-02, 3.17434986304903e-02, 3.16418496445152e-02, 3.15650338502564e-02, 3.15085436634917e-02, 3.14683081644489e-02, 3.14407212128174e-02, 3.14226602168759e-02, 3.14114935396788e-02, 3.14050749628905e-02, 3.14017242220217e-02, 3.14001933172818e-02, 3.13996190344170e-02, 3.13994628167222e-02, 3.13994397593689e-02, 3.13994379287834e-02, 3.13994043811618e-02, 3.13991770917568e-02, 3.13983415620243e-02, 3.13961144241119e-02, 3.13912407814456e-02, 3.13819086827749e-02, 3.13656848097708e-02, 3.13394754000456e-02, 3.12995163265981e-02, 3.12413960152895e-02, 3.11601143974656e-02, 3.10501802400154e-02, 3.09057478660719e-02, 3.07207924061146e-02, 3.04893203017258e-02, 3.02056089117713e-02, 2.98644659385303e-02, 2.94614963054585e-02, 2.89933614669877e-02, 2.84580143508890e-02, 2.78548926389561e-02, 2.71850541996201e-02, 2.64512413413769e-02, 2.56578650644030e-02, 2.48109063052908e-02, 2.39177377062276e-02, 2.29868759520571e-02, 2.20276804090531e-02, 2.10500179668429e-02, 2.00639161397618e-02, 1.90792264302218e-02, 1.81053178125268e-02, 1.71508163405036e-02, 1.62234018825700e-02, 1.53296674953083e-02, 1.44750416207984e-02, 1.36637687064791e-02, 1.28989404049094e-02, 1.21825673946745e-02, 1.15156810102661e-02, 1.08984540459716e-02, 1.03303310035585e-02, 9.81015944326202e-03, 9.33631578531902e-03, 8.90682073227822e-03, 8.51944126706883e-03, 8.17177776903577e-03, 7.86133604283634e-03, 7.58558492424299e-03, 7.34200068583846e-03, 7.12809982744034e-03, 6.94146205917434e-03, 6.77974538930926e-03, 6.64069521481281e-03, 6.52214920152971e-03, 6.42203954045328e-03, 6.33839391541358e-03, 6.26933623652803e-03, 6.21308791193087e-03, 6.16797014711779e-03, 6.13240748656096e-03, 6.10493254549738e-03, 6.08419164014151e-03, 6.06895082057690e-03, 6.05810165882303e-03, 6.05066605882598e-03, 6.04579933776961e-03, 6.04279088348392e-03, 6.04106181235409e-03, 6.04015922387191e-03, 6.03974685627406e-03, 6.03959216823569e-03, 6.03955008982382e-03, 6.03954387915451e-03, 6.03954338154722e-03, 6.03953420572623e-03, 6.03947203858013e-03, 6.03924350973273e-03, 6.03863436873553e-03, 6.03730144588610e-03, 6.03474938019133e-03, 6.03031332190085e-03, 6.02314885394151e-03, 6.01223040233378e-03, 5.99635938347376e-03, 5.97418322031729e-03, 5.94422609755441e-03, 5.90493187509367e-03, 5.85471890813949e-03, 5.79204562893665e-03, 5.71548467330274e-03, 5.62380216295821e-03, 5.51603762254160e-03, 5.39157908541199e-03, 5.25022741426540e-03, 5.09224389824681e-03, 4.91837589784438e-03, 4.72985670865000e-03, 4.52837780730420e-03, 4.31603403016823e-03, 4.09524472357248e-03, 3.86865616881948e-03, 3.63903233868257e-03, 3.40914212729615e-03, 3.18165146414628e-03, 2.95902790949754e-03, 2.74346341224430e-03, 2.53681865932265e-03, 2.34059046591995e-03, 2.15590122988374e-03, 1.98350711248452e-03, 1.82382207536591e-03, 1.67695582575789e-03, 1.54275968885997e-03, 1.42087347910547e-03, 1.31077473484709e-03, 1.21182855887980e-03, 1.12332365621626e-03, 1.04449749262869e-03, 9.74572825811038e-04, 9.12792521958750e-04, 8.58422002934844e-04, 8.10745119679268e-04, 7.69090731836286e-04, 7.32857912083773e-04, 7.01501812514493e-04, 6.74507613979358e-04, 6.51390786032380e-04, 6.31711896630804e-04, 6.15079105955374e-04, 6.01138975554250e-04, 5.89567949987213e-04, 5.80068343367657e-04, 5.72366535436686e-04, 5.66211610561140e-04, 5.61374111214263e-04, 5.57644922452027e-04, 5.54834369310697e-04, 5.52771652309048e-04, 5.51304637141925e-04, 5.50299855021308e-04, 5.49642504338976e-04, 5.49236270273923e-04, 5.49002834361545e-04, 5.48880991303236e-04, 5.48825327413738e-04, 5.48804447149109e-04, 5.48798767312064e-04, 5.48797928985323e-04, 5.48797861199546e-04, 5.48796603339952e-04, 5.48788081303542e-04, 5.48756754355395e-04, 5.48673255963195e-04, 5.48490560930759e-04, 5.48140828087185e-04, 5.47533108473885e-04, 5.46552127048017e-04, 5.45058361339557e-04, 5.42889642730415e-04, 5.39864487905106e-04, 5.35787321837233e-04, 5.30455682167030e-04, 5.23669395789739e-04, 5.15241574191805e-04, 5.05011000498712e-04, 4.92854855238534e-04, 4.78699440421337e-04, 4.62525329595564e-04, 4.44366172246191e-04, 4.24310917040316e-04, 4.02529480695080e-04, 3.79341226121710e-04, 3.55246141708676e-04, 3.30571949291425e-04, 3.05380353083251e-04, 2.79897636261239e-04, 2.55054800774701e-04, 2.31550753582610e-04, 2.08638745121448e-04, 1.86353039225651e-04, 1.66524616034280e-04, 1.48331020383807e-04, 1.30748927517743e-04, 1.16284695181434e-04, 1.04072764244734e-04, 9.27687721343773e-05, 8.21151684675669e-05, 7.20404203917724e-05, 6.26685702940222e-05, 5.42346293173068e-05, 4.69688846465033e-05, 4.15592269149556e-05, 3.71090248307837e-05, 3.31569979579172e-05, 2.93523454644173e-05, 2.61886271601952e-05, 2.36194680240942e-05, 2.10746570417086e-05, 1.85181869341022e-05, 1.63094754839221e-05, 1.47330738968776e-05, 1.31217071981382e-05, 1.15211347946342e-05, 9.98310697752913e-06, 8.55672010402486e-06, 7.27480360751542e-06, 6.14514151209278e-06, 5.15306473403143e-06, 4.27610677888879e-06, 3.49715496383851e-06, 2.80745376812774e-06, 2.20376678328194e-06, 1.68482930007485e-06, 1.24890026143021e-06, 8.92550332309838e-07, 6.10355030516985e-07, 3.95116984464039e-07, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14],
[ 8.50000000000000e+01, 8.49999999007236e+01, 8.49999968308997e+01, 8.49999760323559e+01, 8.49998995741727e+01, 8.49996957606110e+01, 8.49992496987273e+01, 8.49983953568164e+01, 8.49969093986262e+01, 8.49945070228887e+01, 8.49908399752467e+01, 8.49854968328341e+01, 8.49780055937999e+01, 8.49678385384526e+01, 8.49544192688600e+01, 8.49371317826039e+01, 8.49153313957721e+01, 8.48883573003569e+01, 8.48555465201375e+01, 8.48162490128371e+01, 8.47698436491542e+01, 8.47157547748213e+01, 8.46534690247231e+01, 8.45825520057367e+01, 8.45026643992898e+01, 8.44135769631855e+01, 8.43151838478519e+01, 8.42075136013244e+01, 8.40907372369780e+01, 8.39651727920405e+01, 8.38312859198884e+01, 8.36896862320865e+01, 8.35411193237787e+01, 8.33864546565009e+01, 8.32266697089470e+01, 8.30628310118215e+01, 8.28960728356107e+01, 8.27275743863056e+01, 8.25585363804756e+01, 8.23901578240372e+01, 8.22236137225022e+01, 8.20600343224706e+01, 8.19004863432226e+01, 8.17459565195840e+01, 8.15973376545171e+01, 8.14554172787723e+01, 8.13208689373157e+01, 8.11942460663621e+01, 8.10759783866015e+01, 8.09663707125664e+01, 8.08656040600812e+01, 8.07737389191878e+01, 8.06907205457939e+01, 8.06163861096564e+01, 8.05504735183463e+01, 8.04926317165436e+01, 8.04424322379882e+01, 8.03993817647107e+01, 8.03629354260958e+01, 8.03325105503799e+01, 8.03075005649102e+01, 8.02872887304535e+01, 8.02712613905065e+01, 8.02588204202205e+01, 8.02493945722087e+01, 8.02424494387889e+01, 8.02374957822801e+01, 8.02340960264450e+01, 8.02318687521234e+01, 8.02304910970667e+01, 8.02296990220310e+01, 8.02292854700207e+01, 8.02290965107048e+01, 8.02290256248968e+01, 8.02290063421774e+01, 8.02290034960862e+01, 8.02290031227390e+01, 8.02289944241988e+01, 8.02289354902941e+01, 8.02287188428038e+01, 8.02281413384274e+01, 8.02268774755510e+01, 8.02244569937080e+01, 8.02202476514684e+01, 8.02134439561038e+01, 8.02030624884441e+01, 8.01879443169499e+01, 8.01667648232172e+01, 8.01380510637612e+01, 8.01002065675915e+01, 8.00515432152304e+01, 7.99903195647095e+01, 7.99147846898972e+01, 7.98232262873719e+01, 7.97140215071515e+01, 7.95856886940760e+01, 7.94369380217449e+01, 7.92667188960475e+01, 7.90742620369015e+01, 7.88591143421173e+01, 7.86211650024595e+01, 7.83606618456142e+01, 7.80782174751281e+01, 7.77748053445771e+01, 7.74517463656980e+01, 7.71106869167170e+01, 7.67535691794734e+01, 7.63825946573975e+01, 7.60001816457775e+01, 7.56089174942242e+01, 7.52115068128115e+01, 7.48107172905179e+01, 7.44093253292588e+01, 7.40100639787061e+01, 7.36155754745387e+01, 7.32283700170568e+01, 7.28507914839672e+01, 7.24849898752595e+01, 7.21328997098738e+01, 7.17962234272886e+01, 7.14764190065428e+01, 7.11746913242451e+01, 7.08919870683041e+01, 7.06289932129974e+01, 7.03861391257899e+01, 7.01636023458573e+01, 6.99613179921415e+01, 6.97789916620342e+01, 6.96161155930042e+01, 6.94719877874285e+01, 6.93457337451034e+01, 6.92363304036227e+01, 6.91426318485359e+01, 6.90633963189241e+01, 6.89973139980396e+01, 6.89430350436191e+01, 6.88991972810126e+01, 6.88644529583089e+01, 6.88374939507439e+01, 6.88170748063311e+01, 6.88020330496552e+01, 6.87913062086974e+01, 6.87839451014165e+01, 6.87791230137724e+01, 6.87761405163047e+01, 6.87744257979314e+01, 6.87735305376361e+01, 6.87731214806130e+01, 6.87729680283602e+01, 6.87729262855473e+01, 6.87729201243927e+01, 6.87729195804800e+01, 6.87729089227788e+01, 6.87728367152709e+01, 6.87725712726241e+01, 6.87718636983766e+01, 6.87703151812786e+01, 6.87673495544966e+01, 6.87621922067041e+01, 6.87538562984992e+01, 6.87411370805308e+01, 6.87226149282727e+01, 6.86966674982850e+01, 6.86614911676164e+01, 6.86151316386456e+01, 6.85555232761971e+01, 6.84805363970401e+01, 6.83880313649208e+01, 6.82759179748566e+01, 6.81422182624878e+01, 6.79851305759995e+01, 6.78030925285381e+01, 6.75948403338861e+01, 6.73594620359239e+01, 6.70964422817516e+01, 6.68056965580426e+01, 6.64875932013093e+01, 6.61429619918564e+01, 6.57730887319692e+01, 6.53796958701348e+01, 6.49649099332166e+01, 6.45312172195519e+01, 6.40814098226210e+01, 6.36185245226802e+01, 6.31457773360493e+01, 6.26664965111785e+01, 6.21840565147939e+01, 6.17018151157139e+01, 6.12230551382345e+01, 6.07509319206932e+01, 6.02884270593485e+01, 5.98383086869160e+01, 5.94030983302439e+01, 5.89850442811252e+01, 5.85861013526895e+01, 5.82079168414189e+01, 5.78518224500340e+01, 5.75188318482999e+01, 5.72096434703579e+01, 5.69246480855710e+01, 5.66639406471918e+01, 5.64273359221590e+01, 5.62143874297610e+01, 5.60244092550424e+01, 5.58565003416522e+01, 5.57095708974415e+01, 5.55823705574282e+01, 5.54735179399163e+01, 5.53815312034634e+01, 5.53048591687855e+01, 5.52419125163344e+01, 5.51910945142155e+01, 5.51508306799453e+01, 5.51195967409250e+01, 5.50959442394551e+01, 5.50785231345393e+01, 5.50661007887528e+01, 5.50575767961030e+01, 5.50519932056023e+01, 5.50485398222196e+01, 5.50465544166438e+01, 5.50455178406151e+01, 5.50450442169516e+01, 5.50448665437955e+01, 5.50448182123354e+01, 5.50448110787112e+01, 5.50448105065063e+01, 5.50447999464235e+01, 5.50447284003036e+01, 5.50444653891214e+01, 5.50437642972815e+01, 5.50422299719311e+01, 5.50392915438076e+01, 5.50341815565780e+01, 5.50259223618323e+01, 5.50133205862010e+01, 5.49949703017483e+01, 5.49692653255955e+01, 5.49344208341506e+01, 5.48885041972397e+01, 5.48294746168853e+01, 5.47552307984198e+01, 5.46636654983423e+01, 5.45527254012440e+01, 5.44204744016589e+01, 5.42651580360903e+01, 5.40852665593952e+01, 5.38795940219572e+01, 5.36472907094005e+01, 5.33879064762804e+01, 5.31014228476445e+01, 5.27882722690562e+01, 5.24493435280872e+01, 5.20859730987998e+01, 5.16999229074055e+01, 5.12933457043402e+01, 5.08687397825309e+01, 5.04288951536931e+01, 4.99768334715770e+01, 4.95157440009372e+01, 4.90489178288792e+01, 4.85796823596297e+01, 4.81113379626483e+01, 4.76470984635163e+01, 4.71900369580943e+01, 4.67430381675161e+01, 4.63087582228341e+01, 4.58895923863314e+01, 4.54876508171148e+01, 4.51047421179787e+01, 4.47423641012309e+01, 4.44017010104173e+01, 4.40836263397605e+01, 4.37887103923491e+01, 4.35172317880768e+01, 4.32691922445175e+01, 4.30443340812003e+01, 4.28421600184263e+01, 4.26619549411535e+01, 4.25028093686054e+01, 4.23636444085786e+01, 4.22432379832070e+01, 4.21402520938943e+01, 4.20532608524122e+01, 4.19807789488686e+01, 4.19212901620250e+01, 4.18732754504043e+01, 4.18352401010901e+01, 4.18057393643320e+01, 4.17834019728135e+01, 4.17669509404428e+01, 4.17552210610285e+01, 4.17471725843305e+01, 4.17419006355185e+01, 4.17386400612625e+01, 4.17367655263722e+01, 4.17357868419642e+01, 4.17353396710886e+01, 4.17351719215499e+01, 4.17351262895792e+01, 4.17351195543949e+01, 4.17351190349127e+01, 4.17351097066794e+01, 4.17350465065271e+01, 4.17348141760899e+01, 4.17341948685097e+01, 4.17328395293470e+01, 4.17302438939673e+01, 4.17257300661853e+01, 4.17184345309254e+01, 4.17073033232937e+01, 4.16910949238772e+01, 4.16683912699076e+01, 4.16376170584829e+01, 4.15970672669268e+01, 4.15449425252184e+01, 4.14793916496102e+01, 4.13985602946535e+01, 4.13006443194804e+01, 4.11839461170174e+01, 4.10469318513538e+01, 4.08882873218077e+01, 4.07069700552810e+01, 4.05022552495943e+01, 4.02737733683738e+01, 4.00215375270299e+01, 3.97459592957142e+01, 3.94478521455908e+01, 3.91284224280815e+01, 3.87892484391347e+01, 3.84322487147746e+01, 3.80596411713891e+01, 3.76738950061332e+01, 3.72776774007008e+01, 3.68737970484244e+01, 3.64651463978191e+01, 3.60546443321686e+01, 3.56451808312672e+01, 3.52395650082755e+01, 3.48404777703027e+01, 3.44504301816493e+01, 3.40717283776166e+01, 3.37064455692750e+01, 3.33564013142834e+01, 3.30231478544728e+01, 3.27079629975654e+01, 3.24118487964999e+01, 3.21355351759850e+01, 3.18794876622926e+01, 3.16439184578595e+01, 3.14288002287416e+01, 3.12338821068777e+01, 3.10587075278603e+01, 3.09026336171358e+01, 3.07648519002371e+01, 3.06444101473410e+01, 3.05402351724537e+01, 3.04511563964229e+01, 3.03759299542604e+01, 3.03132630846697e+01, 3.02618384875867e+01, 3.02203382792329e+01, 3.01874671197921e+01, 3.01619740429902e+01, 3.01426724861920e+01, 3.01284580100065e+01, 3.01183232123301e+01, 3.01113693858005e+01, 3.01068145400591e+01, 3.01039975088183e+01, 3.01023779820410e+01, 3.01015324391802e+01, 3.01011461027662e+01, 3.01010011744943e+01, 3.01009617504766e+01, 3.01009559315737e+01, 3.01009554909839e+01, 3.01009476859802e+01, 3.01008948059346e+01, 3.01007004133407e+01, 3.01001822344193e+01, 3.00990482143657e+01, 3.00968764373561e+01, 3.00930997210060e+01, 3.00869956026678e+01, 3.00776823274920e+01, 3.00641213220986e+01, 3.00451264883297e+01, 3.00193804739482e+01, 2.99854578675640e+01, 2.99418550223373e+01, 2.98870259400698e+01, 2.98194233523253e+01, 2.97375438323042e+01, 2.96399754802385e+01, 2.95254464706331e+01, 2.93928725588320e+01, 2.92414015434670e+01, 2.90704526919151e+01, 2.88797492706322e+01, 2.86693425812506e+01, 2.84396262723532e+01, 2.81913401478123e+01, 2.79255631876004e+01, 2.76436959949744e+01, 2.73474333485065e+01, 2.70387279434381e+01, 2.67197467427627e+01, 2.63928216236195e+01, 2.60603962020241e+01, 2.57249708475662e+01, 2.53890479497317e+01, 2.50550794522344e+01, 2.47254185147731e+01, 2.44022768865985e+01, 2.40876891944922e+01, 2.37834848902436e+01, 2.34912681157078e+01, 2.32124052793269e+01, 2.29480197441812e+01, 2.26989927383167e+01, 2.24659694291345e+01, 2.22493690525344e+01, 2.20493980369948e+01, 2.18660651864533e+01, 2.16991981540960e+01, 2.15484606241103e+01, 2.14133697971652e+01, 2.12933139313425e+01, 2.11875698134019e+01, 2.10953201211388e+01, 2.10156706860350e+01, 2.09476676793072e+01, 2.08903147287223e+01, 2.08425899342931e+01, 2.08034626950816e+01, 2.07719101940762e+01, 2.07469333209354e+01, 2.07275717505904e+01, 2.07129178461269e+01, 2.07021290229663e+01, 2.06944382027289e+01, 2.06891620020975e+01, 2.06857063452059e+01, 2.06835692559875e+01, 2.06823406758139e+01, 2.06816992560809e+01, 2.06814061883537e+01, 2.06812962488948e+01, 2.06812663427307e+01, 2.06812619286448e+01, 2.06812615972779e+01, 2.06812557648468e+01, 2.06812162492883e+01, 2.06810709862252e+01, 2.06806837706120e+01, 2.06798363707901e+01, 2.06782135450326e+01, 2.06753915822474e+01, 2.06708309290217e+01, 2.06638733950061e+01, 2.06537443432263e+01, 2.06395601653813e+01, 2.06203412083039e+01, 2.05950301513284e+01, 2.05625156326984e+01, 2.05216606877606e+01, 2.04713352992429e+01, 2.04104520831374e+01, 2.03380038614499e+01, 2.02531016293911e+01, 2.01550112369091e+01, 2.00431870008565e+01, 1.99173004696100e+01, 1.97772626947737e+01, 1.96232386319734e+01, 1.94556526878997e+01, 1.92751849309739e+01, 1.90827580498911e+01, 1.88795157266029e+01, 1.86667936295239e+01, 1.84460846710018e+01, 1.82190004622990e+01, 1.79872310097206e+01, 1.77525046216805e+01, 1.75165497589454e+01, 1.72810602019009e+01, 1.70476644863566e+01, 1.68179001329176e+01, 1.65931928157755e+01, 1.63748403196164e+01, 1.61640009324214e+01, 1.59616858123730e+01, 1.57687548305043e+01, 1.55859154027386e+01, 1.54137238623436e+01, 1.52525889693048e+01, 1.51027771972560e+01, 1.49644194787138e+01, 1.48375191269535e+01, 1.47219606906239e+01, 1.46175195367683e+01, 1.45238719988558e+01, 1.44406059662875e+01, 1.43672318270151e+01, 1.43031937016486e+01, 1.42478809227775e+01, 1.42006397155289e+01, 1.41607850245591e+01, 1.41276124101028e+01, 1.41004099040577e+01, 1.40784696799936e+01, 1.40610993526976e+01, 1.40476326879352e+01, 1.40374394759235e+01, 1.40299343065169e+01, 1.40245839833374e+01, 1.40209133299431e+01, 1.40185091740793e+01, 1.40170223450774e+01, 1.40161675820934e+01, 1.40157213234392e+01, 1.40155174252373e+01, 1.40154409361544e+01, 1.40154201292854e+01, 1.40154170582358e+01, 1.40154168283922e+01, 1.40154127922234e+01, 1.40153854465969e+01, 1.40152849213430e+01, 1.40150169591671e+01, 1.40144305368066e+01, 1.40133074920004e+01, 1.40113545832884e+01, 1.40081983666610e+01, 1.40033832225190e+01, 1.39963728054956e+01, 1.39865551147626e+01, 1.39732512896072e+01, 1.39557281211472e+01, 1.39332141345828e+01, 1.39049189386055e+01, 1.38700553642928e+01, 1.38278637338042e+01, 1.37776374222155e+01, 1.37187487197801e+01, 1.36506738844152e+01, 1.35730162124517e+01, 1.34855259640101e+01, 1.33881160668726e+01, 1.32808726914433e+01, 1.31640600333432e+01, 1.30381189454423e+01, 1.29036594068197e+01, 1.27614471759875e+01, 1.26123853198596e+01, 1.24574916070258e+01, 1.22978729737845e+01, 1.21346983891029e+01, 1.19691714454377e+01, 1.18025038866316e+01, 1.16358910705316e+01, 1.14704900876574e+01, 1.13074009628156e+01, 1.11476510974584e+01, 1.09921828987137e+01, 1.08418444003423e+01, 1.06973826076048e+01, 1.05594392759255e+01, 1.04285488416961e+01, 1.03051382443942e+01, 1.01895284006960e+01, 1.00819371086838e+01, 9.98248317402857e+00, 9.89119156305901e+00, 9.80799940291567e+00, 9.73276266812452e+00, 9.66526341568589e+00, 9.60521745532263e+00, 9.55228236510958e+00, 9.50606578234085e+00, 9.46613391263323e+00, 9.43202020519752e+00, 9.40323413830193e+00, 9.37927004667361e+00, 9.35961590346061e+00, 9.34376194577492e+00, 9.33120900774450e+00, 9.32147640170703e+00, 9.31410916996600e+00, 9.30868451934681e+00, 9.30481725094427e+00, 9.30216400937195e+00, 9.30042619988468e+00, 9.29935145722399e+00, 9.29873359516378e+00, 9.29841101787878e+00, 9.29826363019939e+00, 9.29820834008082e+00, 9.29819329983633e+00, 9.29819107992806e+00, 9.29819091372345e+00, 9.29818799425588e+00, 9.29816821443995e+00, 9.29809550184012e+00, 9.29790167748163e+00, 9.29747750134962e+00, 9.29666516762809e+00, 9.29525255727515e+00, 9.29296952371394e+00, 9.28948645011559e+00, 9.28441527213188e+00, 9.27731310558453e+00, 9.26768855199582e+00, 9.25501067391731e+00, 9.23872053590988e+00, 9.21824509671717e+00, 9.19301311698327e+00, 9.16247262077745e+00, 9.12610932675966e+00, 9.08346535665487e+00, 9.03415744645517e+00, 8.97789384073526e+00, 8.91448905198004e+00, 8.84387572067952e+00, 8.76611291921691e+00, 8.68139039904759e+00, 8.59002847707177e+00, 8.49247348006338e+00, 8.38928889950945e+00, 8.28114263619440e+00, 8.16879091781051e+00, 8.05305963902748e+00, 7.93482398984330e+00, 7.81498729635142e+00, 7.69445999447032e+00, 7.57413959304017e+00, 7.45489236486075e+00, 7.33753734476167e+00, 7.22283302870113e+00, 7.11146697526253e+00, 7.00404832844542e+00, 6.90110312329573e+00, 6.80307211412504e+00, 6.71031078389815e+00, 6.62309115265062e+00, 6.54160499807371e+00, 6.46596812538353e+00, 6.39622536768890e+00, 6.33235605376058e+00, 6.27427973986272e+00, 6.22186206015759e+00, 6.17492060182086e+00, 6.13323075366100e+00, 6.09653150925367e+00, 6.06453122691786e+00, 6.03691335953945e+00, 6.01334216809569e+00, 5.99346842498180e+00, 5.97693509845726e+00, 5.96338298959881e+00, 5.95245627020957e+00, 5.94380784652196e+00, 5.93710445166438e+00, 5.93203135208708e+00, 5.92829654157863e+00, 5.92563429281529e+00, 5.92380794164249e+00, 5.92261179380041e+00, 5.92187206708624e+00, 5.92144681271625e+00, 5.92122479597239e+00, 5.92112335568432e+00, 5.92108530209127e+00, 5.92107495060650e+00, 5.92107342274959e+00, 5.92107330805541e+00, 5.92107128934198e+00, 5.92105761226993e+00, 5.92100733403498e+00, 5.92087331170818e+00, 5.92058001213708e+00, 5.92001832758027e+00, 5.91904161353511e+00, 5.91746314536407e+00, 5.91505516679167e+00, 5.91154967441601e+00, 5.90664104578907e+00, 5.89999057200112e+00, 5.89123289704987e+00, 5.87998429456766e+00, 5.86585262823195e+00, 5.84844874795564e+00, 5.82739897461911e+00, 5.80235822886749e+00, 5.77302327348572e+00, 5.73914547433078e+00, 5.70054245191656e+00, 5.65710800312401e+00, 5.60881972563470e+00, 5.55574387756364e+00, 5.49803714701798e+00, 5.43594518102564e+00, 5.36979791577710e+00, 5.30000194258774e+00, 5.22703031781829e+00, 5.15141036347754e+00, 5.07371009603661e+00, 4.99452395793244e+00, 4.91445851000225e+00, 4.83411868064391e+00, 4.75409507045096e+00, 4.67495269335693e+00, 4.59722141074512e+00, 4.52138819522770e+00, 4.44789125428277e+00, 4.37711595552705e+00, 4.30939242692939e+00, 4.24499465642825e+00, 4.18414088469461e+00, 4.12699507017233e+00, 4.07366920487962e+00, 4.02422627055857e+00, 3.97868364526344e+00, 3.93701679783592e+00, 3.89916313919630e+00, 3.86502593221442e+00, 3.83447819350179e+00, 3.80736654854292e+00, 3.78351502443904e+00, 3.76272878105307e+00, 3.74479779100859e+00, 3.72950048186068e+00, 3.71660735035510e+00, 3.70588454996538e+00, 3.69709744010348e+00, 3.69001407004838e+00, 3.68440855438917e+00, 3.68006428137325e+00, 3.67677688266363e+00, 3.67435688415743e+00, 3.67263195389820e+00, 3.67144866551011e+00, 3.67067370424732e+00, 3.67019445739324e+00, 3.66991895051900e+00, 3.66977511472135e+00, 3.66970939578391e+00, 3.66968474247531e+00, 3.66967803619025e+00, 3.66967704635708e+00, 3.66967697173517e+00, 3.66967565411154e+00, 3.66966672702461e+00, 3.66963391021258e+00, 3.66954643343898e+00, 3.66935499671155e+00, 3.66898838800363e+00, 3.66835090097304e+00, 3.66732068290690e+00, 3.66574912850095e+00, 3.66346141684542e+00, 3.66025826549141e+00, 3.65591894497913e+00, 3.65020555825446e+00, 3.64286854119972e+00, 3.63365328336077e+00, 3.62230770360333e+00, 3.60859054724700e+00, 3.59228010420726e+00, 3.57318298816594e+00, 3.55114257191163e+00, 3.52604665079338e+00, 3.49783391069384e+00, 3.46649881290865e+00, 3.43209457667548e+00, 3.39473403812207e+00, 3.35458828574330e+00, 3.31188310760089e+00, 3.26689342240033e+00, 3.21993599260202e+00, 3.17136082048337e+00, 3.12154169744930e+00, 3.07086640630631e+00, 3.01972706368541e+00, 2.96851103838242e+00, 2.91759279884610e+00, 2.86732694069863e+00, 2.81804253602061e+00, 2.77003884268665e+00, 2.72358232434611e+00, 2.67890486594760e+00, 2.63620302807543e+00, 2.59563816421500e+00, 2.55733722430888e+00, 2.52139408038977e+00, 2.48787123051740e+00, 2.45680176138566e+00, 2.42819147466190e+00, 2.40202110538496e+00, 2.37824858149830e+00, 2.35681129134392e+00, 2.33762834052888e+00, 2.32060279096829e+00, 2.30562388310572e+00, 2.29256924732569e+00, 2.28130711244066e+00, 2.27169851797221e+00, 2.26359953298277e+00, 2.25686347783832e+00, 2.25134313705396e+00, 2.24689294200920e+00, 2.24337109268397e+00, 2.24064157859064e+00, 2.23857605171560e+00, 2.23705549940885e+00, 2.23597166347881e+00, 2.23522815372769e+00, 2.23474120995824e+00, 2.23444007591165e+00, 2.23426696115195e+00, 2.23417658179680e+00, 2.23413528721403e+00, 2.23411979626982e+00, 2.23411558236491e+00, 2.23411496040163e+00, 2.23411491325770e+00, 2.23411407743660e+00, 2.23410841463003e+00, 2.23408759761887e+00, 2.23403210760773e+00, 2.23391067160100e+00, 2.23367811679699e+00, 2.23327373231984e+00, 2.23262022077877e+00, 2.23162331281404e+00, 2.23017210614054e+00, 2.22814017559489e+00, 2.22538748120975e+00, 2.22176307644693e+00, 2.21710858792586e+00, 2.21126240151031e+00, 2.20406444864639e+00, 2.19536144364852e+00, 2.18501238055742e+00, 2.17289406146401e+00, 2.15890640146678e+00, 2.14297724319209e+00, 2.12506641963481e+00, 2.10516882990011e+00, 2.08331633799335e+00, 2.05957836739162e+00, 2.03406113876342e+00, 2.00690557820042e+00, 1.97828400127148e+00, 1.94839574698845e+00, 1.91746198958623e+00, 1.88571999105658e+00, 1.85341707212397e+00, 1.82080457446203e+00, 1.78813206482113e+00, 1.75564199594019e+00, 1.72356499382717e+00, 1.69211589052276e+00, 1.66149056991185e+00, 1.63186364515937e+00, 1.60338694293900e+00, 1.57618873403629e+00, 1.55037362353963e+00, 1.52602299718293e+00, 1.50319591314509e+00, 1.48193032969850e+00, 1.46224456697901e+00, 1.44413891397443e+00, 1.42759730769215e+00, 1.41258902857604e+00, 1.39907037307036e+00, 1.38698627956617e+00, 1.37627189695408e+00, 1.36685409511708e+00, 1.35865292367495e+00, 1.35158302912372e+00, 1.34555504136942e+00, 1.34047693885400e+00, 1.33625539745113e+00, 1.33279712261512e+00, 1.33001015751297e+00, 1.32780515274116e+00, 1.32609657642827e+00, 1.32480383774001e+00, 1.32385229266340e+00, 1.32317409894962e+00, 1.32270888757731e+00, 1.32240422117867e+00, 1.32221581544032e+00, 1.32210750721211e+00, 1.32205096240187e+00, 1.32202512703250e+00, 1.32201543536169e+00, 1.32201279899868e+00, 1.32201240987726e+00, 1.32201238019523e+00, 1.32201185148988e+00, 1.32200826943692e+00, 1.32199510148537e+00, 1.32196000097655e+00, 1.32188318647924e+00, 1.32173608560394e+00, 1.32148030122080e+00, 1.32106695350862e+00, 1.32043644584741e+00, 1.31951869563245e+00, 1.31823386186406e+00, 1.31649358903626e+00, 1.31420277070852e+00, 1.31126181580765e+00, 1.30756937627321e+00, 1.30302546688812e+00, 1.29753487854170e+00, 1.29101075701453e+00, 1.28337819355039e+00, 1.27457765422972e+00, 1.26456806569874e+00, 1.25332937782164e+00, 1.24086444098874e+00, 1.22720006735700e+00, 1.21238718975250e+00, 1.19650008613560e+00, 1.17963469677208e+00, 1.16190611998165e+00, 1.14344542474524e+00, 1.12439595931447e+00, 1.10490936035956e+00, 1.08514147504684e+00, 1.06524839880345e+00, 1.04538280647544e+00, 1.02569071783863e+00, 1.00630879471322e+00, 9.87362221292287e-01, 9.68963176326952e-01, 9.51209869164543e-01, 9.34186083655689e-01, 9.17961155601517e-01, 9.02590300442313e-01, 8.88115207085832e-01, 8.74564819355868e-01, 8.61956236505847e-01, 8.50295676661020e-01, 8.39579460273654e-01, 8.29794983427178e-01, 8.20921662233204e-01, 8.12931839119926e-01, 8.05791649284174e-01, 7.99461850957670e-01, 7.93898626532012e-01, 7.89054363185457e-01, 7.84878421674179e-01, 7.81317900628750e-01, 7.78318401269019e-01, 7.75824794172465e-01, 7.73781985867103e-01, 7.72135678865766e-01, 7.70833114626839e-01, 7.69823785152838e-01, 7.69060095849596e-01, 7.68497960176447e-01, 7.68097305769530e-01, 7.67822472285928e-01, 7.67642483254022e-01, 7.67531177655584e-01, 7.67467191622998e-01, 7.67433786201494e-01, 7.67418523229153e-01, 7.67412797600201e-01, 7.67411240093978e-01, 7.67411010209472e-01, 7.67410992550239e-01, 7.67410676376159e-01, 7.67408534252496e-01, 7.67400659611695e-01, 7.67379668959151e-01, 7.67333732691355e-01, 7.67245764077970e-01, 7.67092801048364e-01, 7.66845612728773e-01, 7.66468559424453e-01, 7.65919730558352e-01, 7.65151379489183e-01, 7.64110666623210e-01, 7.62740712369364e-01, 7.60981949019055e-01, 7.58773745635239e-01, 7.56056263012127e-01, 7.52772477720399e-01, 7.48870296679925e-01, 7.44304668509017e-01, 7.39039587200821e-01, 7.33049879505813e-01, 7.26322671362169e-01, 7.18858441633586e-01, 7.10671593051842e-01, 7.01790499191105e-01, 6.92257019982492e-01, 6.82125513336956e-01, 6.71461403171492e-01, 6.60339391025886e-01, 6.48841416808779e-01, 6.37054482503189e-01, 6.25068450779708e-01, 6.12973919610241e-01, 6.00860256370964e-01, 5.88813853351269e-01, 5.76916643882200e-01, 5.65244896926771e-01, 5.53868289722985e-01, 5.42849243953199e-01, 5.32242501194176e-01, 5.22094907780797e-01, 5.12445377015425e-01, 5.03324997049745e-01, 4.94757254936246e-01, 4.86758350588275e-01, 4.79337578162912e-01, 4.72497756305273e-01, 4.66235692518410e-01, 4.60542670492373e-01, 4.55404952431536e-01, 4.50804291187652e-01, 4.46718449264580e-01, 4.43121723453501e-01, 4.39985474931902e-01, 4.37278665084150e-01, 4.34968397073646e-01, 4.33020462351536e-01, 4.31399889899097e-01, 4.30071494196241e-01, 4.29000415850226e-01, 4.28152646703808e-01, 4.27495529287881e-01, 4.26998218912843e-01, 4.26632095701251e-01, 4.26371113619652e-01, 4.26192074170189e-01, 4.26074813889784e-01, 4.26002297126061e-01, 4.25960608601301e-01, 4.25938843844406e-01, 4.25928899440819e-01, 4.25925168966842e-01, 4.25924154188891e-01, 4.25924004409850e-01, 4.25923992816351e-01, 4.25923784100974e-01, 4.25922370024594e-01, 4.25917171745468e-01, 4.25903315158166e-01, 4.25872990945962e-01, 4.25814918713385e-01, 4.25713937690034e-01, 4.25550744505898e-01, 4.25301796169336e-01, 4.24939393436574e-01, 4.24431955992631e-01, 4.23744495999998e-01, 4.22839290386989e-01, 4.21676744626694e-01, 4.20216431759457e-01, 4.18418280332631e-01, 4.16243874329965e-01, 4.13657817868128e-01, 4.10629108480606e-01, 4.07132456329521e-01, 4.03149483769468e-01, 3.98669741187177e-01, 3.93691481404147e-01, 3.88222146057302e-01, 3.82278532583527e-01, 3.75886628492736e-01, 3.69081118908412e-01, 3.61904592100544e-01, 3.54406484253900e-01, 3.46641817646045e-01, 3.38669794888403e-01, 3.30552315590126e-01, 3.22352480930334e-01, 3.14133146766352e-01, 3.05955577911035e-01, 2.97878246044860e-01, 2.89955802342874e-01, 2.82238244175744e-01, 2.74770283944136e-01, 2.67590917850811e-01, 2.60733183677516e-01, 2.54224089737306e-01, 2.48084692285346e-01, 2.42330295813446e-01, 2.36970749715354e-01, 2.32010815564891e-01, 2.27450581388084e-01, 2.23285902472568e-01, 2.19508852057103e-01, 2.16108169308427e-01, 2.13069695988324e-01, 2.10376796855300e-01, 2.08010761922111e-01, 2.05951191052094e-01, 2.04176362943954e-01, 2.02663591307299e-01, 2.01389571007397e-01, 2.00330716246214e-01, 1.99463491579280e-01, 1.98764734912615e-01, 1.98211969773965e-01, 1.97783702313487e-01, 1.97459696862921e-01, 1.97221222656069e-01, 1.97051263633753e-01, 1.96934683225265e-01, 1.96858336655168e-01, 1.96811124643943e-01, 1.96783984268492e-01, 1.96769815081183e-01, 1.96763341184578e-01, 1.96760912622196e-01, 1.96760251996224e-01, 1.96760154489269e-01, 1.96760146880649e-01, 1.96760009113802e-01, 1.96759075724358e-01, 1.96755644503241e-01, 1.96746498260097e-01, 1.96726482602983e-01, 1.96688152801585e-01, 1.96621505018214e-01, 1.96513806131605e-01, 1.96349535086043e-01, 1.96110447003363e-01, 1.95775768969241e-01, 1.95322533213741e-01, 1.94726049146638e-01, 1.93960510262784e-01, 1.92999725324654e-01, 1.91817955658664e-01, 1.90390832276051e-01, 1.88696318462846e-01, 1.86715676255190e-01, 1.84434389709651e-01, 1.81842994965720e-01, 1.78937767521688e-01, 1.75721221386496e-01, 1.72202382936248e-01, 1.68396814061666e-01, 1.64326373761739e-01, 1.60018723547026e-01, 1.55506598435693e-01, 1.50826880429406e-01, 1.46019523722288e-01, 1.41126389373468e-01, 1.36190051033083e-01, 1.31252632316887e-01, 1.26354730871810e-01, 1.21534474820947e-01, 1.16826745238797e-01, 1.12262584902041e-01, 1.07868800107212e-01, 1.03667750022055e-01, 9.96773077469015e-02, 9.59109695900793e-02, 9.23780842499866e-02, 8.90841715826411e-02, 8.60313010989131e-02, 8.32185027721408e-02, 8.06421865572702e-02, 7.82965516352192e-02, 7.61739712807241e-02, 7.42653439953074e-02, 7.25604058541620e-02, 7.10480026899205e-02, 6.97163236781080e-02, 6.85531000576702e-02, 6.75457741426300e-02, 6.66816445055929e-02, 6.59479933263327e-02, 6.53322014828177e-02, 6.48218561198064e-02, 6.44048542605731e-02, 6.40695046407984e-02, 6.38046284507823e-02, 6.35996581955283e-02, 6.34447325351175e-02, 6.33307838607351e-02, 6.32496145904042e-02, 6.31939577968410e-02, 6.31575178412902e-02, 6.31349871779179e-02, 6.31220363628415e-02, 6.31152754733222e-02, 6.31121865047817e-02, 6.31110277483344e-02, 6.31107125405634e-02, 6.31106660166552e-02, 6.31106623554132e-02, 6.31105956670268e-02, 6.31101438445114e-02, 6.31084829149131e-02, 6.31040556228584e-02, 6.30943672950366e-02, 6.30758156421081e-02, 6.30435624658645e-02, 6.29914550504509e-02, 6.29120047307481e-02, 6.27964295981656e-02, 6.26347673235728e-02, 6.24160624377339e-02, 6.21286299301289e-02, 6.17603935831283e-02, 6.12992930512064e-02, 6.07337484700849e-02, 6.00531656673837e-02, 5.92484593523324e-02, 5.83125666485671e-02, 5.72409197412753e-02, 5.60318449748110e-02, 5.46868570506973e-02, 5.32108213773199e-02, 5.16119650807241e-02, 4.99017272431956e-02, 4.80944507176986e-02, 4.62069301695180e-02, 4.42578424492126e-02, 4.22670946737185e-02, 4.02551313977568e-02, 3.82422443274493e-02, 3.62479259944308e-02, 3.42903030367552e-02, 3.23856760505663e-02, 3.05481825594569e-02, 2.87895888228320e-02, 2.71192062167103e-02, 2.55439197279230e-02, 2.40683102497741e-02, 2.26948489964656e-02, 2.14241413130907e-02, 2.02551981104619e-02, 1.91857156544160e-02, 1.82123479588796e-02, 1.73309600048839e-02, 1.65368539221760e-02, 1.58249637543548e-02, 1.51900173006254e-02, 1.46266657540224e-02, 1.41295834858113e-02, 1.36935414233506e-02, 1.33134581000292e-02, 1.29844326952906e-02, 1.27017643120553e-02, 1.24609614485678e-02, 1.22577451870001e-02, 1.20880490958228e-02, 1.19480182644907e-02, 1.18340092774218e-02, 1.17425923103980e-02, 1.16705559163393e-02, 1.16149144808162e-02, 1.15729178004358e-02, 1.15420617958648e-02, 1.15200990420448e-02, 1.15050476059224e-02, 1.14951966354226e-02, 1.14891072486358e-02, 1.14856075106360e-02, 1.14837806379029e-02, 1.14829459945298e-02, 1.14826329022719e-02, 1.14825477346585e-02, 1.14825351641343e-02, 1.14825341661202e-02, 1.14825158764944e-02, 1.14823919621330e-02, 1.14819364487460e-02, 1.14807222853130e-02, 1.14780654637485e-02, 1.14729786558883e-02, 1.14641367941175e-02, 1.14498571046937e-02, 1.14280961650479e-02, 1.13964665233652e-02, 1.13522751909869e-02, 1.12925857973210e-02, 1.12143052850939e-02, 1.11142946635283e-02, 1.09895015243211e-02, 1.08371098372402e-02, 1.06547001517043e-02, 1.04404110065133e-02, 1.01930904383067e-02, 9.91242535818060e-03, 9.59903658666406e-03, 9.25452873740085e-03, 8.88148698247448e-03, 8.48341684537957e-03, 8.06462815387476e-03, 7.63006954989010e-03, 7.18512481626021e-03, 6.73538606127977e-03, 6.28642095296073e-03, 5.84355138414594e-03, 5.41165913566033e-03, 4.99503057889683e-03, 4.59724781899240e-03, 4.22112870637648e-03, 3.86871361328645e-03, 3.54129332097841e-03, 3.23946999314728e-03, 2.96324194641784e-03, 2.71210261374222e-03, 2.48514463966387e-03, 2.28116139192025e-03, 2.09874016242395e-03, 1.93634358437578e-03, 1.79237779482833e-03, 1.66524724293993e-03, 1.55339677607458e-03, 1.45534199798192e-03, 1.36968917280456e-03, 1.29514625699301e-03, 1.23052691374426e-03, 1.17474943536422e-03, 1.12683228226729e-03, 1.08588754471839e-03, 1.05111324036500e-03, 1.02178509358453e-03, 9.97248297849751e-04, 9.76909679977211e-04, 9.60230608455959e-04, 9.46720893992392e-04, 9.35933817984846e-04, 9.27462316084871e-04, 9.20936247922555e-04, 9.16020611413160e-04, 9.12414505208403e-04, 9.09850605196993e-04, 9.08094899159551e-04, 9.06946419149790e-04, 9.06236725484177e-04, 9.05828928182473e-04, 9.05616080984828e-04, 9.05518842998276e-04, 9.05482367869581e-04, 9.05472445950295e-04, 9.05470981507506e-04, 9.05470864196900e-04, 9.05468700437950e-04, 9.05454040847076e-04, 9.05400152447555e-04, 9.05256519089903e-04, 9.04942248865837e-04, 9.04340643899104e-04, 9.03295261238668e-04, 9.01607826719442e-04, 8.99038394223339e-04, 8.95308155548946e-04, 8.90105282053355e-04, 8.83094096380432e-04, 8.73927711509737e-04, 8.62264026309845e-04, 8.47784630579440e-04, 8.30215774072940e-04, 8.09350129813424e-04, 7.85067697719910e-04, 7.57353914551118e-04, 7.26312933736040e-04, 6.92174166474047e-04, 6.55290583623014e-04, 6.16127995954230e-04, 5.75245541725167e-04, 5.33268768563744e-04, 4.90857718119742e-04, 4.48673065392313e-04, 4.07343599803767e-04, 3.67438082619742e-04, 3.29443369366800e-04, 2.93748931795790e-04, 2.60637546602993e-04, 2.30283565699096e-04, 2.02759674294839e-04, 1.78049139123004e-04, 1.56061724595702e-04, 1.36655841842526e-04, 1.19659358858826e-04, 1.04874451551495e-04, 9.20822534595800e-05, 8.10695542128315e-05, 7.16361206798410e-05, 6.35630164111944e-05, 5.66386757583359e-05, 5.07415313677484e-05, 4.57969835281347e-05, 4.16320875160636e-05, 3.80344971375735e-05, 3.49393279464039e-05, 3.24096873214330e-05, 3.04029916833315e-05, 2.87244074002047e-05, 2.72321681933363e-05, 2.58962759005499e-05, 2.47384895004643e-05, 2.37771516111707e-05, 2.30043501403543e-05, 2.23928880842550e-05, 2.19128797169788e-05, 2.15401362311241e-05, 2.12560556178687e-05, 2.10450804019047e-05, 2.08929995579865e-05, 2.07866766430698e-05, 2.07145646555773e-05, 2.06671834774852e-05, 2.06371833115782e-05, 2.06190703693210e-05, 2.06088168823726e-05, 2.06035112029213e-05, 2.06010980008386e-05, 2.06001945031176e-05, 2.05999489111264e-05, 2.05999126876577e-05, 2.05999094617937e-05, 2.05998463087553e-05, 2.05994184341307e-05, 2.05978458609479e-05, 2.05936565293996e-05, 2.05845013425687e-05, 2.05670181259165e-05, 2.05367711316740e-05, 2.04883029653047e-05, 2.04153432434850e-05, 2.03112133334571e-05, 2.01694182919709e-05, 1.99842725407962e-05, 1.97510944985879e-05, 1.94651181857353e-05, 1.91183013384943e-05, 1.86943484692392e-05, 1.81640499050673e-05, 1.74836713853210e-05, 1.65984801725472e-05, 1.54546214758966e-05, 1.40409812843245e-05, 1.27009065522543e-05, 1.20798992497839e-05, 1.13443535383340e-05, 1.00839069148151e-05, 8.43784296478134e-06, 8.50106688687902e-06, 8.56418358807787e-06, 8.08838106536314e-06, 7.08055368647820e-06, 5.96314563949818e-06, 5.64179322282143e-06, 6.74562394920031e-06, 8.87400723570408e-06, 1.11697964086943e-05, 1.27593767444978e-05, 1.30032335334818e-05, 1.17087545845607e-05, 9.16701971220846e-06, 6.02352635326042e-06, 3.05245991683023e-06, 9.88394901924039e-07, 7.11894912797106e-07, 1.25920826121728e-06, 1.44241880661866e-06, 1.11538389546695e-06, 1.50414298616397e-06, 1.90114977272060e-06, 1.66602374918057e-06, 8.92392108653439e-07, 1.01156938114855e-06, 1.31492291188806e-06, 1.42465314922902e-06, 1.26973846192868e-06, 9.91904883562711e-07, 7.31343632764947e-07, 5.82696202400257e-07, 5.61126825004976e-07, 6.05428326139190e-07, 6.46628180334119e-07, 6.49998597168633e-07, 6.10250992849615e-07, 5.37265407274390e-07, 4.45832256446010e-07, 3.49868443549894e-07, 0.00000000000000e+00, 0.00000000000000e+00, -1.42108547152020e-14, 0.00000000000000e+00, -1.42108547152020e-14, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, -1.42108547152020e-14],
[ 8.60000000000000e+01, 8.59999998990832e+01, 8.59999967785357e+01, 8.59999756363306e+01, 8.59998979148026e+01, 8.59996907335531e+01, 8.59992373011886e+01, 8.59983688424757e+01, 8.59968583305141e+01, 8.59944162571883e+01, 8.59906886116396e+01, 8.59852571685647e+01, 8.59776421190870e+01, 8.59673070103231e+01, 8.59536658990895e+01, 8.59360925734401e+01, 8.59139316547380e+01, 8.58865113628978e+01, 8.58531577062986e+01, 8.58132098414881e+01, 8.57660363302554e+01, 8.57110519963639e+01, 8.56477350456492e+01, 8.55756440586860e+01, 8.54944343969260e+01, 8.54038734890026e+01, 8.53038543971593e+01, 8.51944070216163e+01, 8.50757063007448e+01, 8.49480768212771e+01, 8.48119933721397e+01, 8.46680771545018e+01, 8.45170875852617e+01, 8.43599098785989e+01, 8.41975388325715e+01, 8.40310594573765e+01, 8.38616252362425e+01, 8.36904348953842e+01, 8.35187085729925e+01, 8.33476642259523e+01, 8.31784950115675e+01, 8.30123482488424e+01, 8.28503064189742e+01, 8.26933705241340e+01, 8.25424459991711e+01, 8.23983312691047e+01, 8.22617089678022e+01, 8.21331397780184e+01, 8.20130588155899e+01, 8.19017744557720e+01, 8.17994694823413e+01, 8.17062044259190e+01, 8.16219229439271e+01, 8.15464590788130e+01, 8.14795462128880e+01, 8.14208275173772e+01, 8.13698676707385e+01, 8.13261655980761e+01, 8.12891679608928e+01, 8.12582831060195e+01, 8.12328951659066e+01, 8.12123779911424e+01, 8.11961085915658e+01, 8.11834797659665e+01, 8.11739116131372e+01, 8.11668616395953e+01, 8.11618332117829e+01, 8.11583821426099e+01, 8.11561212528787e+01, 8.11547228059416e+01, 8.11539187769201e+01, 8.11534989836433e+01, 8.11533071725935e+01, 8.11532352169939e+01, 8.11532156432646e+01, 8.11532127542209e+01, 8.11532123752393e+01, 8.11532035454229e+01, 8.11531437221032e+01, 8.11529238050323e+01, 8.11523375851790e+01, 8.11510546488195e+01, 8.11485976394524e+01, 8.11443247770334e+01, 8.11374184201613e+01, 8.11268803251723e+01, 8.11115341046109e+01, 8.10900352131586e+01, 8.10608885885329e+01, 8.10224738456040e+01, 8.09730776633954e+01, 8.09109327190914e+01, 8.08342622170695e+01, 8.07413287458940e+01, 8.06304858899972e+01, 8.05002307504514e+01, 8.03492553229846e+01, 8.01764945783862e+01, 7.99811691274240e+01, 7.97628205560912e+01, 7.95213378919185e+01, 7.92569741791493e+01, 7.89703527327997e+01, 7.86624632125466e+01, 7.83346481050518e+01, 7.79885804557644e+01, 7.76262337417391e+01, 7.72498447020297e+01, 7.68618698856038e+01, 7.64649367934540e+01, 7.60617908646275e+01, 7.56552401234674e+01, 7.52480998505674e+01, 7.48431398759800e+01, 7.44430368190907e+01, 7.40503328341395e+01, 7.36674014138419e+01, 7.32964199051310e+01, 7.29393478628583e+01, 7.25979102718080e+01, 7.22735848859735e+01, 7.19675932726839e+01, 7.16808954446393e+01, 7.14141881337774e+01, 7.11679068019725e+01, 7.09422314325246e+01, 7.07370960505589e+01, 7.05522018169919e+01, 7.03870334504109e+01, 7.02408786606928e+01, 7.01128502252073e+01, 7.00019102970434e+01, 6.99068964987538e+01, 6.98265493203812e+01, 6.97595403053225e+01, 6.97045004727214e+01, 6.96600483933762e+01, 6.96248173117549e+01, 6.95974806943718e+01, 6.95767755891351e+01, 6.95615232052555e+01, 6.95506461715593e+01, 6.95431820036260e+01, 6.95382924062531e+01, 6.95352681546507e+01, 6.95335294310922e+01, 6.95326216377229e+01, 6.95322068541880e+01, 6.95320512537154e+01, 6.95320089265339e+01, 6.95320026791277e+01, 6.95320021276006e+01, 6.95319913206995e+01, 6.95319181023403e+01, 6.95316489436987e+01, 6.95309314639845e+01, 6.95293612690724e+01, 6.95263541269030e+01, 6.95211245841943e+01, 6.95126719916753e+01, 6.94997747459793e+01, 6.94809933725253e+01, 6.94546828608384e+01, 6.94190144172798e+01, 6.93720065169484e+01, 6.93115648165992e+01, 6.92355301386758e+01, 6.91417333640895e+01, 6.90280556960959e+01, 6.88924924036656e+01, 6.87332178485285e+01, 6.85486493749605e+01, 6.83375075216088e+01, 6.80988700194363e+01, 6.78322171788995e+01, 6.75374665428054e+01, 6.72149950812423e+01, 6.68656477188291e+01, 6.64907315954630e+01, 6.60919961463362e+01, 6.56715998096340e+01, 6.52320648780813e+01, 6.47762226335321e+01, 6.43071513651349e+01, 6.38281101057544e+01, 6.33424708954596e+01, 6.28536521096367e+01, 6.23650549352884e+01, 6.18800045373483e+01, 6.14016969278473e+01, 6.09331521126572e+01, 6.04771737806492e+01, 6.00363156123130e+01, 5.96128541819697e+01, 5.92087683622060e+01, 5.88257250732575e+01, 5.84650711370697e+01, 5.81278309003002e+01, 5.78147092000638e+01, 5.75260991797281e+01, 5.72620944310140e+01, 5.70225049435608e+01, 5.68068763750767e+01, 5.66145122003009e+01, 5.64444983410001e+01, 5.62957299108195e+01, 5.61669397211908e+01, 5.60567281853283e+01, 5.59635942280643e+01, 5.58859667640687e+01, 5.58222362519745e+01, 5.57707857742899e+01, 5.57300210404862e+01, 5.56983986710087e+01, 5.56744521002641e+01, 5.56568144428284e+01, 5.56442377033643e+01, 5.56356077790583e+01, 5.56299548033022e+01, 5.56264585078144e+01, 5.56244484320159e+01, 5.56233989759008e+01, 5.56229194672179e+01, 5.56227395863876e+01, 5.56226906543865e+01, 5.56226834321238e+01, 5.56226828528089e+01, 5.56226721615122e+01, 5.56225997263990e+01, 5.56223334471877e+01, 5.56216236440147e+01, 5.56200702542891e+01, 5.56170953162364e+01, 5.56119218399037e+01, 5.56035600351726e+01, 5.55908017139421e+01, 5.55722235062955e+01, 5.55461993227775e+01, 5.55109222512543e+01, 5.54644357929711e+01, 5.54046740173531e+01, 5.53295098528055e+01, 5.52368103417097e+01, 5.51244972897306e+01, 5.49906113573567e+01, 5.48333773061977e+01, 5.46512678584511e+01, 5.44430634897199e+01, 5.42079054834703e+01, 5.39453397514349e+01, 5.36553492761324e+01, 5.33383735493827e+01, 5.29953140332046e+01, 5.26275254041659e+01, 5.22367930888507e+01, 5.18252982786782e+01, 5.13955721570800e+01, 5.09504414370673e+01, 5.04929674865677e+01, 5.00263813440912e+01, 4.95540168516380e+01, 4.90792440061134e+01, 4.86054044823006e+01, 4.81357511070469e+01, 4.76733928432729e+01, 4.72212465514214e+01, 4.67819964311589e+01, 4.63580616297410e+01, 4.59515720785058e+01, 4.55643522357540e+01, 4.51979121152412e+01, 4.48534447888132e+01, 4.45318294719471e+01, 4.42336393164517e+01, 4.39591531184304e+01, 4.37083702715068e+01, 4.34810284277575e+01, 4.32766234507202e+01, 4.30944313427037e+01, 4.29335318956768e+01, 4.27928338497406e+01, 4.26711013475544e+01, 4.25669814511656e+01, 4.24790324447714e+01, 4.24057525890355e+01, 4.23456089262422e+01, 4.22970656677815e+01, 4.22586116334463e+01, 4.22287861630048e+01, 4.22062028912383e+01, 4.21895707738920e+01, 4.21777117779917e+01, 4.21695747079274e+01, 4.21642447283451e+01, 4.21609482634626e+01, 4.21590530947179e+01, 4.21580636374890e+01, 4.21576115444020e+01, 4.21574419483686e+01, 4.21573958141063e+01, 4.21573890047848e+01, 4.21573884795844e+01, 4.21573790486710e+01, 4.21573151528461e+01, 4.21570802650434e+01, 4.21564541404661e+01, 4.21550838824773e+01, 4.21524596757793e+01, 4.21478961622914e+01, 4.21405203218663e+01, 4.21292665881839e+01, 4.21128797758820e+01, 4.20899262127801e+01, 4.20588132551865e+01, 4.20178171102905e+01, 4.19651185961619e+01, 4.18988461403335e+01, 4.18171249620986e+01, 4.17181310184125e+01, 4.16001479427218e+01, 4.14616249001865e+01, 4.13012330551322e+01, 4.11179182304941e+01, 4.09109473629155e+01, 4.06799465393925e+01, 4.04249287453755e+01, 4.01463099449522e+01, 3.98449127162516e+01, 3.95219573268746e+01, 3.91790407906063e+01, 3.88181050316359e+01, 3.84413957401291e+01, 3.80514137997012e+01, 3.76508612998097e+01, 3.72425841413885e+01, 3.68295131503705e+01, 3.64146054830297e+01, 3.60007879753703e+01, 3.55909039625032e+01, 3.51876649493332e+01, 3.47936083119976e+01, 3.44110619205720e+01, 3.40421161986717e+01, 3.36886037135874e+01, 3.33520859859485e+01, 3.30338468859034e+01, 3.27348917855589e+01, 3.24559515721592e+01, 3.21974906695203e+01, 3.19597183262340e+01, 3.17426025670671e+01, 3.15458863376403e+01, 3.13691054840603e+01, 3.12116082921035e+01, 3.10725763650851e+01, 3.09510466488012e+01, 3.08459344193349e+01, 3.07560570379728e+01, 3.06801582495664e+01, 3.06169327592857e+01, 3.05650507716451e+01, 3.05231821199265e+01, 3.04900195597436e+01, 3.04643007542488e+01, 3.04448284471542e+01, 3.04304883094347e+01, 3.04202639609779e+01, 3.04132487122706e+01, 3.04086536437753e+01, 3.04058117397980e+01, 3.04041779148609e+01, 3.04033249074186e+01, 3.04029351604548e+01, 3.04027889527807e+01, 3.04027491807359e+01, 3.04027433104649e+01, 3.04027428659857e+01, 3.04027349920811e+01, 3.04026816452224e+01, 3.04024855365833e+01, 3.04019627833863e+01, 3.04008187530016e+01, 3.03986278063872e+01, 3.03948177582441e+01, 3.03886597784526e+01, 3.03792643512808e+01, 3.03655837819923e+01, 3.03464215900253e+01, 3.03204489488642e+01, 3.02862281212338e+01, 3.02422425931453e+01, 3.01869333345730e+01, 3.01187403164536e+01, 3.00361481075420e+01, 2.99377340805529e+01, 2.98222174997434e+01, 2.96885075688225e+01, 2.95357484150497e+01, 2.93633589939173e+01, 2.91710660310902e+01, 2.89589283741027e+01, 2.87273514914495e+01, 2.84770913040568e+01, 2.82092470278170e+01, 2.79252432077448e+01, 2.76268016009537e+01, 2.73159039935747e+01, 2.69947474034280e+01, 2.66656934206624e+01, 2.63312136685001e+01, 2.59938335163965e+01, 2.56560762320087e+01, 2.53204096968982e+01, 2.49891976195830e+01, 2.46646568584907e+01, 2.43488220367573e+01, 2.40435181306249e+01, 2.37503411965024e+01, 2.34706469242709e+01, 2.32055463135332e+01, 2.29559074961997e+01, 2.27223625840798e+01, 2.25053183962814e+01, 2.23049699963739e+01, 2.21213161138446e+01, 2.19541757071302e+01, 2.18032051183277e+01, 2.16679154504739e+01, 2.15476899519151e+01, 2.14418013102991e+01, 2.13494288381099e+01, 2.12696755734470e+01, 2.12015853276338e+01, 2.11441596905875e+01, 2.10963749619863e+01, 2.10571989179688e+01, 2.10256072565064e+01, 2.10005994968951e+01, 2.09812140470952e+01, 2.09665421035021e+01, 2.09557400169394e+01, 2.09480397506743e+01, 2.09427570737986e+01, 2.09392971770438e+01, 2.09371574664457e+01, 2.09359273795162e+01, 2.09352851732008e+01, 2.09349917460958e+01, 2.09348816718249e+01, 2.09348517289888e+01, 2.09348473094902e+01, 2.09348469777171e+01, 2.09348411381340e+01, 2.09348015741206e+01, 2.09346561329336e+01, 2.09342684425221e+01, 2.09334200036892e+01, 2.09317951883821e+01, 2.09289697666258e+01, 2.09244035250995e+01, 2.09174374700958e+01, 2.09072960220480e+01, 2.08930945019886e+01, 2.08738520759098e+01, 2.08485101557660e+01, 2.08159560531060e+01, 2.07750514444882e+01, 2.07246649437444e+01, 2.06637077978279e+01, 2.05911714496713e+01, 2.05061654677647e+01, 2.04079541560831e+01, 2.02959900580412e+01, 2.01699425796222e+01, 2.00297200975783e+01, 1.98754841950383e+01, 1.97076550706571e+01, 1.95269076738297e+01, 1.93341586868435e+01, 1.91305450520064e+01, 1.89173952683245e+01, 1.86961951009103e+01, 1.84685496112686e+01, 1.82361435021999e+01, 1.80007016768671e+01, 1.77639516630180e+01, 1.75275891968405e+01, 1.72932478544037e+01, 1.70624732199496e+01, 1.68367017361271e+01, 1.66172441197527e+01, 1.64052730551496e+01, 1.62018147852443e+01, 1.60077441872173e+01, 1.58237829208045e+01, 1.56505002536038e+01, 1.54883161870972e+01, 1.53375065257713e+01, 1.51982095516043e+01, 1.50704339913092e+01, 1.49540679970669e+01, 1.48488889032410e+01, 1.47545735688784e+01, 1.46707091637313e+01, 1.45968042984522e+01, 1.45323004325526e+01, 1.44765835132322e+01, 1.44289958027345e+01, 1.43888478417905e+01, 1.43554304738837e+01, 1.43280268226506e+01, 1.43059240767068e+01, 1.42884248970581e+01, 1.42748582266593e+01, 1.42645892540812e+01, 1.42570282675535e+01, 1.42516381349035e+01, 1.42479401609451e+01, 1.42455181071397e+01, 1.42440202077761e+01, 1.42431590800211e+01, 1.42427094982632e+01, 1.42425040816804e+01, 1.42424270229960e+01, 1.42424060611812e+01, 1.42424029672619e+01, 1.42424027357067e+01, 1.42423986694811e+01, 1.42423711202151e+01, 1.42422698463587e+01, 1.42419998886662e+01, 1.42414090990803e+01, 1.42402776901849e+01, 1.42383102352184e+01, 1.42351305052032e+01, 1.42302794786855e+01, 1.42232167981110e+01, 1.42133258713136e+01, 1.41999227230825e+01, 1.41822685871175e+01, 1.41595860914006e+01, 1.41310787313309e+01, 1.40959531496354e+01, 1.40534435588434e+01, 1.40028374636933e+01, 1.39435016831618e+01, 1.38749075525933e+01, 1.37966541231667e+01, 1.37084881835735e+01, 1.36103200171869e+01, 1.35022339801448e+01, 1.33844932363690e+01, 1.32575383007343e+01, 1.31219793993610e+01, 1.29785830273635e+01, 1.28282534354363e+01, 1.26720100716393e+01, 1.25109622105431e+01, 1.23462820939710e+01, 1.21791778761207e+01, 1.20108675202786e+01, 1.18425545632783e+01, 1.16754063896470e+01, 1.15105353856878e+01, 1.13489831122013e+01, 1.11917074644010e+01, 1.10395726815706e+01, 1.08933420153780e+01, 1.07536728457913e+01, 1.06211140292451e+01, 1.04961052627662e+01, 1.03789782450068e+01, 1.02699594111008e+01, 1.01691740164843e+01, 1.00766513491048e+01, 9.99233086206279e+00, 9.91606903966932e+00, 9.84764683712827e+00, 9.78677756411113e+00, 9.73311511147185e+00, 9.68626244455950e+00, 9.64578030317203e+00, 9.61119605541056e+00, 9.58201265004419e+00, 9.55771760017889e+00, 9.53779191179218e+00, 9.52171884664148e+00, 9.50899238333696e+00, 9.49912521642848e+00, 9.49165611457144e+00, 9.48615644823427e+00, 9.48223569733869e+00, 9.47954576113298e+00, 9.47778391686570e+00, 9.47669430970758e+00, 9.47606790206022e+00, 9.47574086321531e+00, 9.47559143701388e+00, 9.47553538217559e+00, 9.47552013390879e+00, 9.47551788329687e+00, 9.47551771479347e+00, 9.47551475494662e+00, 9.47549470155511e+00, 9.47542098326323e+00, 9.47522447810466e+00, 9.47479443513630e+00, 9.47397086583271e+00, 9.47253871705833e+00, 9.47022410516007e+00, 9.46669285289170e+00, 9.46155152564866e+00, 9.45435110842098e+00, 9.44459339740600e+00, 9.43174009831111e+00, 9.41522452613923e+00, 9.39446568964553e+00, 9.36888442100371e+00, 9.33792108349284e+00, 9.30105426576087e+00, 9.25781976105257e+00, 9.20782904541947e+00, 9.15078642180566e+00, 9.08650399664219e+00, 9.01491370840999e+00, 8.93607573518021e+00, 8.85018276679129e+00, 8.75755982825645e+00, 8.65865957094908e+00, 8.55405319061761e+00, 8.44441736821504e+00, 8.33051784272284e+00, 8.21319039793629e+00, 8.09332016387586e+00, 7.97182018876883e+00, 7.84961022564175e+00, 7.72759660107141e+00, 7.60665390162362e+00, 7.48760904065507e+00, 7.37122807297027e+00, 7.25820592709158e+00, 7.14915904290075e+00, 7.04462075053715e+00, 6.94503911366019e+00, 6.85077688960732e+00, 6.76211322847182e+00, 6.67924673665564e+00, 6.60229955978572e+00, 6.53132218616150e+00, 6.46629872690825e+00, 6.40715248591789e+00, 6.35375168633766e+00, 6.30591526733818e+00, 6.26341870306931e+00, 6.22599982408408e+00, 6.19336463983548e+00, 6.16519316943965e+00, 6.14114528748034e+00, 6.12086658327833e+00, 6.10399421718145e+00, 6.09016273777063e+00, 6.07900980144889e+00, 6.07018171292881e+00, 6.06333868401803e+00, 6.05815969113894e+00, 6.05434680129224e+00, 6.05162883334047e+00, 6.04976422759368e+00, 6.04854301202718e+00, 6.04778777754458e+00, 6.04735360623481e+00, 6.04712693360588e+00, 6.04702336590523e+00, 6.04698451423006e+00, 6.04697394564607e+00, 6.04697238574570e+00, 6.04697226864609e+00, 6.04697020759454e+00, 6.04695624367494e+00, 6.04690491094890e+00, 6.04676807766549e+00, 6.04646862602709e+00, 6.04589515815506e+00, 6.04489794853731e+00, 6.04328634255488e+00, 6.04082777622071e+00, 6.03724856233761e+00, 6.03223655069512e+00, 6.02544572235091e+00, 6.01650271854602e+00, 6.00501523240973e+00, 5.99058210688926e+00, 5.97280488788081e+00, 5.95130048224885e+00, 5.92571447350130e+00, 5.89573456244671e+00, 5.86110353637058e+00, 5.82163113816330e+00, 5.77720421480252e+00, 5.72779457783489e+00, 5.67346410767627e+00, 5.61436677396688e+00, 5.55074741579013e+00, 5.48293731368749e+00, 5.41134677271089e+00, 5.33645510452567e+00, 5.25879853132730e+00, 5.17895662410858e+00, 5.09753792763323e+00, 5.01516541564125e+00, 4.93246236902314e+00, 4.85003918722975e+00, 4.76848154067756e+00, 4.68834016037380e+00, 4.61012244945189e+00, 4.53428599634215e+00, 4.46123397529979e+00, 4.39131233984897e+00, 4.32480865038910e+00, 4.26195233023226e+00, 4.20291611557184e+00, 4.14781845429512e+00, 4.09672661488008e+00, 4.04966028727640e+00, 4.00659548897187e+00, 3.96746862707020e+00, 3.93218060680209e+00, 3.90060091459922e+00, 3.87257163667492e+00, 3.84791139997708e+00, 3.82641924038110e+00, 3.80787841287874e+00, 3.79206016073636e+00, 3.77872745606433e+00, 3.76763871419671e+00, 3.75855147020918e+00, 3.75122598943675e+00, 3.74542876672179e+00, 3.74093585308312e+00, 3.73753593522605e+00, 3.73503308431134e+00, 3.73324908684944e+00, 3.73202527325087e+00, 3.73122376867761e+00, 3.73072810608947e+00, 3.73044316188589e+00, 3.73029439897716e+00, 3.73022642880775e+00, 3.73020093098526e+00, 3.73019399497201e+00, 3.73019297123143e+00, 3.73019289405333e+00, 3.73019153129363e+00, 3.73018229840319e+00, 3.73014835742653e+00, 3.73005788405290e+00, 3.72985988939094e+00, 3.72948072165050e+00, 3.72882139497143e+00, 3.72775587987228e+00, 3.72613047458199e+00, 3.72376435756841e+00, 3.72045139719903e+00, 3.71596326287987e+00, 3.71005384170230e+00, 3.70246491477617e+00, 3.69293298834307e+00, 3.68119710823214e+00, 3.66700741571981e+00, 3.65013413355621e+00, 3.63037660942324e+00, 3.60757199782907e+00, 3.58160313786051e+00, 3.55240518966822e+00, 3.51997063124848e+00, 3.48435229001495e+00, 3.44566418796109e+00, 3.40408010794260e+00, 3.35982993107322e+00, 3.31319393803294e+00, 3.26449539571378e+00, 3.21409185144074e+00, 3.16236561937967e+00, 3.10971396192548e+00, 3.05653944307952e+00, 3.00324086732377e+00, 2.95020512733373e+00, 2.89780018070212e+00, 2.84636927324004e+00, 2.79622643545505e+00, 2.74765320649925e+00, 2.70089648885634e+00, 2.65616740619615e+00, 2.61364102270990e+00, 2.57345678047302e+00, 2.53571951782605e+00, 2.50050094316352e+00, 2.46784145266305e+00, 2.43775219602092e+00, 2.41021731039197e+00, 2.38519625890291e+00, 2.36262622579842e+00, 2.34242453491075e+00, 2.32449107107685e+00, 2.30871069473052e+00, 2.29495564761785e+00, 2.28308795203408e+00, 2.27296180698974e+00, 2.26442598236724e+00, 2.25732620677579e+00, 2.25150753706055e+00, 2.24681668808878e+00, 2.24310429151473e+00, 2.24022704279035e+00, 2.23804968782740e+00, 2.23644679540959e+00, 2.23530425950237e+00, 2.23452047752579e+00, 2.23400715662679e+00, 2.23368970981701e+00, 2.23350721699464e+00, 2.23341194149111e+00, 2.23336840982807e+00, 2.23335207967952e+00, 2.23334763749082e+00, 2.23334698183336e+00, 2.23334693213543e+00, 2.23334605103460e+00, 2.23334008145137e+00, 2.23331813669856e+00, 2.23325964055660e+00, 2.23313162577898e+00, 2.23288647212162e+00, 2.23246017901060e+00, 2.23177125965969e+00, 2.23072033356604e+00, 2.22919048201982e+00, 2.22704841549006e+00, 2.22414648135451e+00, 2.22032551447930e+00, 2.21541850112420e+00, 2.20925498878633e+00, 2.20166613209152e+00, 2.19249021998462e+00, 2.18157848561414e+00, 2.16880096168343e+00, 2.15405211532628e+00, 2.13725598229291e+00, 2.11837052409150e+00, 2.09739095586247e+00, 2.07435183723375e+00, 2.04932778094047e+00, 2.02243271009941e+00, 1.99381767847687e+00, 1.96366735169882e+00, 1.93219532390543e+00, 1.89963850747489e+00, 1.86625087829648e+00, 1.83229688277693e+00, 1.79804481451700e+00, 1.76376044958829e+00, 1.72970119246288e+00, 1.69611093411737e+00, 1.66321576471425e+00, 1.63122062096799e+00, 1.60030688805006e+00, 1.57063092220521e+00, 1.54232341657217e+00, 1.51548950110443e+00, 1.49020944863422e+00, 1.46653985238520e+00, 1.44451514396664e+00, 1.42414933277131e+00, 1.40543786515840e+00, 1.38835952229738e+00, 1.37287829682910e+00, 1.35894520876201e+00, 1.34650003894711e+00, 1.33547297321989e+00, 1.32578616142514e+00, 1.31735520293500e+00, 1.31009057406538e+00, 1.30389901328833e+00, 1.29868487776558e+00, 1.29435148001004e+00, 1.29080240702422e+00, 1.28794281673279e+00, 1.28568069863304e+00, 1.28392807806695e+00, 1.28260213708855e+00, 1.28162622021071e+00, 1.28093069089464e+00, 1.28045360484674e+00, 1.28014116914284e+00, 1.27994796179881e+00, 1.27983689430123e+00, 1.27977890924512e+00, 1.27975241589348e+00, 1.27974247740342e+00, 1.27973977390077e+00, 1.27973937486972e+00, 1.27973934443179e+00, 1.27973880226209e+00, 1.27973512898674e+00, 1.27972162570045e+00, 1.27968563137634e+00, 1.27960686111136e+00, 1.27945601595306e+00, 1.27919372412875e+00, 1.27876986912516e+00, 1.27812335476490e+00, 1.27718235052684e+00, 1.27586505100386e+00, 1.27408097067276e+00, 1.27173277828365e+00, 1.26871865371223e+00, 1.26493512414689e+00, 1.26028030682436e+00, 1.25465745382058e+00, 1.24797866313828e+00, 1.24016859267596e+00, 1.23116799320263e+00, 1.22093686672916e+00, 1.20945706062719e+00, 1.19673412731645e+00, 1.18279831451783e+00, 1.16770460022575e+00, 1.15153174602348e+00, 1.13438040683970e+00, 1.11637039839323e+00, 1.09763727887797e+00, 1.07832844313560e+00, 1.05859895149688e+00, 1.03860731970902e+00, 1.01851148145275e+00, 9.98465103740202e-01, 9.78614392598558e-01, 9.59095477485107e-01, 9.40032413524904e-01, 9.21535795877418e-01, 9.03701943970816e-01, 8.86612587053804e-01, 8.70334966986974e-01, 8.54922268593228e-01, 8.40414290466668e-01, 8.26838277690314e-01, 8.14209850164119e-01, 8.02533974159473e-01, 7.91805938669711e-01, 7.82012310967531e-01, 7.73131856808718e-01, 7.65136419602229e-01, 7.57991759539408e-01, 7.51658358244399e-01, 7.46092197175997e-01, 7.41245519008515e-01, 7.37067580784782e-01, 7.33505405995672e-01, 7.30504540118915e-01, 7.28009810775006e-01, 7.25966089768960e-01, 7.24319050157803e-01, 7.23015907407060e-01, 7.22006129994853e-01, 7.21242101807263e-01, 7.20679716644597e-01, 7.20278884369677e-01, 7.20003928843923e-01, 7.19823859869763e-01, 7.19712504827626e-01, 7.19648490368783e-01, 7.19615070105874e-01, 7.19599800352299e-01, 7.19594072179461e-01, 7.19592513981240e-01, 7.19592283994587e-01, 7.19592266327524e-01, 7.19591950012955e-01, 7.19589806937577e-01, 7.19581928798021e-01, 7.19560928819092e-01, 7.19514972140502e-01, 7.19426964437076e-01, 7.19273933426166e-01, 7.19026635219947e-01, 7.18649414222554e-01, 7.18100341081694e-01, 7.17331647599181e-01, 7.16290469971909e-01, 7.14919901830910e-01, 7.13160346026271e-01, 7.10951139048291e-01, 7.08232404893494e-01, 7.04947077103299e-01, 7.01043010148439e-01, 6.96475086239900e-01, 6.91207213160212e-01, 6.85214104861842e-01, 6.78482740959026e-01, 6.71013414599685e-01, 6.62820300209873e-01, 6.53931501705387e-01, 6.44388575303438e-01, 6.34245555599719e-01, 6.23567545390699e-01, 6.12428955421208e-01, 6.00911497306555e-01, 5.89102040052481e-01, 5.77090438016882e-01, 5.64967427219372e-01, 5.52822669915940e-01, 5.40743007029860e-01, 5.28810957012993e-01, 5.17103480174953e-01, 5.05691010939003e-01, 4.94636747599515e-01, 4.83996180058085e-01, 4.73816830340610e-01, 4.64138177841207e-01, 4.54991740524079e-01, 4.46401284138389e-01, 4.38383133369285e-01, 4.30946561405847e-01, 4.24094237402599e-01, 4.17822714558113e-01, 4.12122944881006e-01, 4.06980810016577e-01, 4.02377660614363e-01, 3.98290859479118e-01, 3.94694326020357e-01, 3.91559081173213e-01, 3.88853792918397e-01, 3.86545322738542e-01, 3.84599272824076e-01, 3.82980532656802e-01, 3.81653821881940e-01, 3.80584224311590e-01, 3.79737705700904e-01, 3.79081605848441e-01, 3.78585093833877e-01, 3.78219574046398e-01, 3.77959030251304e-01, 3.77780295413061e-01, 3.77663236375497e-01, 3.77590844756043e-01, 3.77549228413471e-01, 3.77527501410867e-01, 3.77517574273213e-01, 3.77513850278859e-01, 3.77512837263765e-01, 3.77512687744925e-01, 3.77512676171520e-01, 3.77512467818732e-01, 3.77511056199097e-01, 3.77505866952831e-01, 3.77492034456907e-01, 3.77461763034461e-01, 3.77403792155022e-01, 3.77302988182151e-01, 3.77140083299579e-01, 3.76891579948051e-01, 3.76529836231441e-01, 3.76023344036639e-01, 3.75337206777559e-01, 3.74433817485397e-01, 3.73273730301506e-01, 3.71816709340436e-01, 3.70022928673436e-01, 3.67854286393097e-01, 3.65275785187379e-01, 3.62256922606576e-01, 3.58773027395799e-01, 3.54806474982865e-01, 3.50347716339101e-01, 3.45396060452032e-01, 3.39960161532602e-01, 3.34058177215582e-01, 3.27717582252632e-01, 3.20974642000451e-01, 3.13873569626210e-01, 3.06465408693001e-01, 2.98806697200945e-01, 2.90957979219598e-01, 2.82982235416256e-01, 2.74943304018493e-01, 2.66904359430853e-01, 2.58926507561483e-01, 2.51067545836349e-01, 2.43380922926605e-01, 2.35914919463895e-01, 2.28712057486646e-01, 2.21808733952813e-01, 2.15235063097552e-01, 2.09014904215195e-01, 2.03166045869509e-01, 1.97700514608414e-01, 1.92624975792967e-01, 1.87941195787019e-01, 1.83646538029365e-01, 1.79734469907459e-01, 1.76195062356101e-01, 1.73015469249947e-01, 1.70180378550896e-01, 1.67672431507626e-01, 1.65472609773119e-01, 1.63560592981128e-01, 1.61915091050943e-01, 1.60514156290361e-01, 1.59335480301536e-01, 1.58356679878509e-01, 1.57555574670539e-01, 1.56910457547866e-01, 1.56400356551586e-01, 1.56005285246692e-01, 1.55706476436038e-01, 1.55486592724486e-01, 1.55329906501606e-01, 1.55222441646103e-01, 1.55152069690516e-01, 1.55108554303112e-01, 1.55083539659245e-01, 1.55070480452096e-01, 1.55064513750347e-01, 1.55062275460040e-01, 1.55061666593227e-01, 1.55061576725856e-01, 1.55061569713354e-01, 1.55061442740418e-01, 1.55060582481440e-01, 1.55057420099482e-01, 1.55048990511347e-01, 1.55030543381628e-01, 1.54995218054722e-01, 1.54933796820487e-01, 1.54834550105207e-01, 1.54683186177053e-01, 1.54462916551509e-01, 1.54154646229586e-01, 1.53737294929414e-01, 1.53188251352432e-01, 1.52483957118989e-01, 1.51600610322859e-01, 1.50514970897433e-01, 1.49205241590892e-01, 1.47651989965510e-01, 1.45839069320218e-01, 1.43754490735802e-01, 1.41391195518892e-01, 1.38747677983332e-01, 1.35828413282141e-01, 1.32644053999243e-01, 1.29211372048474e-01, 1.25552938210788e-01, 1.21696549046618e-01, 1.17674428310039e-01, 1.13522245642812e-01, 1.09278007636277e-01, 1.04980884071026e-01, 1.00670034547235e-01, 9.63834976942195e-02, 9.21571972096586e-02, 8.80241071807006e-02, 8.40136049065023e-02, 8.01510243586563e-02, 7.64574090011365e-02, 7.29494502098902e-02, 6.96395878642306e-02, 6.65362432763317e-02, 6.36441515163710e-02, 6.09647600667387e-02, 5.84966630461281e-02, 5.62360442907845e-02, 5.41771076634490e-02, 5.23124784312188e-02, 5.06335648994707e-02, 4.91308743355745e-02, 4.77942813402450e-02, 4.66132501148877e-02, 4.55770145160699e-02, 4.46747214286205e-02, 4.38955439136919e-02, 4.32287708930374e-02, 4.26638799276589e-02, 4.21905990295102e-02, 4.17989625082100e-02, 4.14793646821749e-02, 4.12226139653313e-02, 4.10199884607749e-02, 4.08632928423600e-02, 4.07449150737875e-02, 4.06578804906559e-02, 4.05959000284639e-02, 4.05534089751711e-02, 4.05255925923171e-02, 4.05083952793461e-02, 4.04985106149525e-02, 4.04933505334952e-02, 4.04909929875998e-02, 4.04901086130565e-02, 4.04898680438492e-02, 4.04898325364229e-02, 4.04898297421283e-02, 4.04897788450285e-02, 4.04894340107660e-02, 4.04881663808823e-02, 4.04847874783485e-02, 4.04773935090645e-02, 4.04632357643919e-02, 4.04386234350511e-02, 4.03988651297453e-02, 4.03382553124629e-02, 4.02501115735134e-02, 4.01268679518978e-02, 3.99602282213039e-02, 3.97413809638465e-02, 3.94612752672714e-02, 3.91109519836874e-02, 3.86819208252263e-02, 3.81665684572567e-02, 3.75585776754406e-02, 3.68533333578398e-02, 3.60482878823234e-02, 3.51432577872970e-02, 3.41406251735492e-02, 3.30454219662784e-02, 3.18652825708057e-02, 3.06102601260676e-02, 2.92925125470305e-02, 2.79258756182941e-02, 2.65253502039542e-02, 2.51065379319186e-02, 2.36850635609471e-02, 2.22760221956833e-02, 2.08934856701575e-02, 1.95500953922760e-02, 1.82567597524326e-02, 1.70224641059775e-02, 1.58541916130588e-02, 1.47569449363232e-02, 1.37338526869542e-02, 1.27863408754365e-02, 1.19143483862700e-02, 1.11165663063275e-02, 1.03906832853369e-02, 9.73362244745601e-03, 9.14175915747251e-03, 8.61111269917103e-03, 8.13750827612125e-03, 7.71670847809958e-03, 7.34451540128589e-03, 7.01684600514341e-03, 6.72978413511771e-03, 6.47961304615308e-03, 6.26283232875835e-03, 6.07616296439062e-03, 5.91654389992868e-03, 5.78112310542167e-03, 5.66724561784326e-03, 5.57244060554751e-03, 5.49440903310483e-03, 5.43101306679716e-03, 5.38026794215796e-03, 5.34033661919864e-03, 5.30952718492017e-03, 5.28629263777702e-03, 5.26923241632460e-03, 5.25709482981540e-03, 5.24877942423529e-03, 5.24333828099088e-03, 5.23997530129350e-03, 5.23804266636319e-03, 5.23703387264618e-03, 5.23657299576996e-03, 5.23640011290638e-03, 5.23635308535120e-03, 5.23634614420132e-03, 5.23634559314701e-03, 5.23633549403257e-03, 5.23626707158087e-03, 5.23601554952791e-03, 5.23534513146728e-03, 5.23387817831633e-03, 5.23106970841525e-03, 5.22618863647040e-03, 5.21830725621442e-03, 5.20630055095239e-03, 5.18885693017523e-03, 5.16450187569717e-03, 5.13163566111530e-03, 5.08858571998871e-03, 5.03367333818971e-03, 4.96529313249994e-03, 4.88200231072256e-03, 4.78261511555900e-03, 4.66629633349669e-03, 4.53264655588725e-03, 4.38177128155814e-03, 4.21432619300788e-03, 4.03153216822716e-03, 3.83515581452798e-03, 3.62745435457157e-03, 3.41108721546846e-03, 3.18900019361479e-03, 2.96429107561380e-03, 2.74006761895862e-03, 2.51930950759061e-03, 2.30474517536550e-03, 2.09875235118773e-03, 1.90328816239571e-03, 1.71985115386519e-03, 1.54947420945462e-03, 1.39274460052263e-03, 1.24984553099239e-03, 1.12061265190277e-03, 1.00459894413518e-03, 9.01141914667392e-04, 8.09428043694993e-04, 7.28550731878386e-04, 6.57559467995839e-04, 5.95499345990902e-04, 5.41441168550705e-04, 4.94503039419669e-04, 4.53864607678156e-04, 4.18775153718798e-04, 3.88556682878516e-04, 3.62603183589272e-04, 3.40377186162755e-04, 3.21404652822821e-04, 3.05269028757493e-04, 2.91605031094377e-04, 2.80092530928755e-04, 2.70450735783356e-04, 2.62432806380275e-04, 2.55821006120982e-04, 2.50422456488764e-04, 2.46065540238760e-04, 2.42596958173635e-04, 2.39879409662080e-04, 2.37789839104607e-04, 2.36218170257985e-04, 2.35066436857778e-04, 2.34248209139309e-04, 2.33688211977778e-04, 2.33322030581462e-04, 2.33095805640460e-04, 2.32965833035337e-04, 2.32898000078308e-04, 2.32867012201154e-04, 2.32855388503594e-04, 2.32852226645264e-04, 2.32851759975006e-04, 2.32851722586247e-04, 2.32851033032944e-04, 2.32846361299721e-04, 2.32829188234973e-04, 2.32783416521443e-04, 2.32683273836187e-04, 2.32491593962436e-04, 2.32158590577569e-04, 2.31621252211767e-04, 2.30803502830668e-04, 2.29617279117633e-04, 2.27964668411573e-04, 2.25741217633413e-04, 2.22840460324392e-04, 2.19159606089647e-04, 2.14606202007417e-04, 2.09105414043620e-04, 2.02607410898281e-04, 1.95094189010092e-04, 1.86585085586444e-04, 1.77140216280236e-04, 1.66861167386401e-04, 1.55888486148115e-04, 1.44395843761913e-04, 1.32581171897073e-04, 1.20655531645752e-04, 1.08830874509636e-04, 9.73080907868962e-05, 8.62666707348581e-05, 7.58568743464139e-05, 6.61947044733324e-05, 5.73595560240392e-05, 4.93942429358185e-05, 4.23071340378556e-05, 3.60764718294604e-05, 3.06569561985270e-05, 2.59875442765178e-05, 2.19988260568016e-05, 1.86189962789740e-05, 1.57768341750852e-05, 1.34017916906259e-05, 1.14257967283038e-05, 9.78588535360814e-06, 8.42385257726619e-06, 7.29194168513914e-06, 6.35850928176751e-06, 5.59214694817456e-06, 4.95743485373623e-06, 4.44568038915349e-06, 4.05799887914782e-06, 3.74265245284278e-06, 3.43731942109571e-06, 3.17287324946847e-06, 3.04009928697724e-06, 3.02677123897865e-06, 3.01385647105690e-06, 2.92340575924754e-06, 2.76414282041060e-06, 2.59256204060421e-06, 2.46233382483751e-06, 2.39051848893723e-06, 2.36079394255739e-06, 2.34992585035343e-06, 2.34395017173483e-06, 2.33819687878167e-06, 2.33240072589069e-06, 2.32731282778786e-06, 2.32342075889846e-06, 2.32077925943486e-06, 2.31917773874102e-06, 2.31831734254229e-06, 2.31791868543496e-06, 2.31776822090524e-06, 2.31772720837853e-06, 2.31772115455442e-06, 2.31772096981331e-06, 2.31772108350015e-06, 2.31772189351886e-06, 2.31772510517203e-06, 2.31773555015025e-06, 2.31776775194703e-06, 2.31786471260875e-06, 2.31814013318399e-06, 2.31885590551428e-06, 2.32053088211615e-06, 2.32403073141541e-06, 2.33049244968697e-06, 2.34080654593072e-06, 2.35432794681856e-06, 2.36677172438249e-06, 2.36809668763271e-06, 2.34235054108467e-06, 2.27222139415062e-06, 2.15099544220720e-06, 2.00230533664580e-06, 1.89128161309782e-06, 1.87187735889438e-06, 1.89793267679761e-06, 1.88082729835060e-06, 1.82235785928242e-06, 1.82975966822596e-06, 1.93915552415547e-06, 2.04985373386535e-06, 2.11710582220803e-06, 2.18026809761795e-06, 2.24538609927549e-06, 2.30522583422044e-06, 2.41441530590691e-06, 2.58891847693121e-06, 2.76485648953440e-06, 2.88861538422225e-06, 2.89526106200810e-06, 2.70903584009830e-06, 2.34547339061919e-06, 1.90308388425819e-06, 1.38586779030447e-06, 9.07668422200913e-07, 1.13813416646735e-06, 1.10719891210920e-06, 4.23389835191301e-07, 1.79553644841235e-06, 2.75803246552186e-06, 3.20147486831956e-06, 3.08625615730307e-06, 2.49662990370325e-06, 1.60032057294757e-06, 5.56076329871757e-07, 7.34181696770975e-07, 9.34386648054897e-07, 8.24942503641068e-07, 5.93647470736869e-07, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14, 4.42459736404999e-07, 4.66079029592947e-07, 4.48506455086317e-07, 4.00334485561871e-07, -4.26325641456060e-14, -5.68434188608080e-14, -2.84217094304040e-14, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14],
[ 8.70000000000000e+01, 8.69999998974743e+01, 8.69999967271750e+01, 8.69999752478941e+01, 8.69998962872302e+01, 8.69996858028254e+01, 8.69992251412117e+01, 8.69983428361872e+01, 8.69968082408854e+01, 8.69943272304385e+01, 8.69905401475718e+01, 8.69850220944482e+01, 8.69772856039421e+01, 8.69667856559209e+01, 8.69529269426214e+01, 8.69350732348703e+01, 8.69125586595755e+01, 8.68847006687168e+01, 8.68508144589205e+01, 8.68102285842465e+01, 8.67623014869176e+01, 8.67064386445879e+01, 8.66421099926505e+01, 8.65688672233914e+01, 8.64863604927964e+01, 8.63943539887701e+01, 8.62927397453950e+01, 8.61815490443996e+01, 8.60609607454274e+01, 8.59313059454774e+01, 8.57930684917390e+01, 8.56468810572232e+01, 8.54935167203203e+01, 8.53338762438326e+01, 8.51689714973053e+01, 8.49999056800801e+01, 8.48278511584078e+01, 8.46540258145008e+01, 8.44796688159654e+01, 8.43060166583780e+01, 8.41342802273878e+01, 8.39656234892112e+01, 8.38011442694979e+01, 8.36418574371072e+01, 8.34886806832459e+01, 8.33424229840832e+01, 8.32037757577484e+01, 8.30733066721346e+01, 8.29514560234857e+01, 8.28385355818208e+01, 8.27347297825621e+01, 8.26400991299639e+01, 8.25545856640123e+01, 8.24780203265556e+01, 8.24101320438152e+01, 8.23505583212342e+01, 8.22988571235752e+01, 8.22545197894183e+01, 8.22169847061340e+01, 8.21856514505447e+01, 8.21598950834650e+01, 8.21390802747112e+01, 8.21225749305041e+01, 8.21097629987805e+01, 8.21000561407996e+01, 8.20929039802401e+01, 8.20878026738841e+01, 8.20843015905989e+01, 8.20820079367057e+01, 8.20805892244529e+01, 8.20797735442031e+01, 8.20793476677366e+01, 8.20791530771819e+01, 8.20790800788868e+01, 8.20790602215183e+01, 8.20790572906101e+01, 8.20790569061367e+01, 8.20790479483692e+01, 8.20789872581618e+01, 8.20787641543241e+01, 8.20781694397490e+01, 8.20768679130434e+01, 8.20743753014927e+01, 8.20700405283067e+01, 8.20630341118023e+01, 8.20523433608348e+01, 8.20367748772685e+01, 8.20149646993833e+01, 8.19853962164189e+01, 8.19464257513310e+01, 8.18963154455994e+01, 8.18332727890115e+01, 8.17554958253888e+01, 8.16612227442650e+01, 8.15487842570828e+01, 8.14166568804670e+01, 8.12635150416283e+01, 8.10882798199252e+01, 8.08901621810432e+01, 8.06686987722766e+01, 8.04237787317952e+01, 8.01556604897052e+01, 7.98649781339562e+01, 7.95527374807935e+01, 7.92203024254850e+01, 7.88693723862851e+01, 7.85019516943705e+01, 7.81203117119075e+01, 7.77269464320527e+01, 7.73245224840010e+01, 7.69158249026863e+01, 7.65037006378875e+01, 7.60910023248942e+01, 7.56805350179627e+01, 7.52750082137513e+01, 7.48769946262485e+01, 7.44888961127198e+01, 7.41129162612843e+01, 7.37510386824186e+01, 7.34050100259097e+01, 7.30763270201315e+01, 7.27662271932032e+01, 7.24756832258708e+01, 7.22054010352560e+01, 7.19558217046758e+01, 7.17271273033912e+01, 7.15192505317014e+01, 7.13318880178427e+01, 7.11645170023898e+01, 7.10164150776182e+01, 7.08866825995800e+01, 7.07742673523573e+01, 7.06779910104358e+01, 7.05965769119543e+01, 7.05286786211173e+01, 7.04729087232983e+01, 7.04278672643554e+01, 7.03921692207760e+01, 7.03644703744143e+01, 7.03434909695796e+01, 7.03280365551038e+01, 7.03170154625322e+01, 7.03094524447922e+01, 7.03044980968174e+01, 7.03014337978987e+01, 7.02996720505472e+01, 7.02987522365197e+01, 7.02983319606097e+01, 7.02981742997514e+01, 7.02981314120946e+01, 7.02981250819633e+01, 7.02981245231332e+01, 7.02981135731325e+01, 7.02980393852530e+01, 7.02977666625549e+01, 7.02970396823948e+01, 7.02954486960929e+01, 7.02924017363227e+01, 7.02871029517185e+01, 7.02785384492447e+01, 7.02654704638982e+01, 7.02464404892915e+01, 7.02197817871337e+01, 7.01836414439950e+01, 7.01360118567898e+01, 7.00747712041372e+01, 6.99977321040546e+01, 6.99026972804918e+01, 6.97875206800644e+01, 6.96501721203346e+01, 6.94888032407806e+01, 6.93018122969465e+01, 6.90879052139327e+01, 6.88461503173743e+01, 6.85760242991414e+01, 6.82774472524958e+01, 6.79508050208519e+01, 6.75969576334586e+01, 6.72172332326724e+01, 6.68134076051550e+01, 6.63876701735372e+01, 6.59425780279721e+01, 6.54810002043971e+01, 6.50060548693432e+01, 6.45210422858030e+01, 6.40293763832239e+01, 6.35345174590264e+01, 6.30399080689650e+01, 6.25489136187531e+01, 6.20647686505066e+01, 6.15905293987458e+01, 6.11290329021408e+01, 6.06828627851753e+01, 6.02543217259246e+01, 5.98454105534706e+01, 5.94578138362671e+01, 5.90928917199492e+01, 5.87516776602270e+01, 5.84348815952314e+01, 5.81428980322485e+01, 5.78758184963666e+01, 5.76334478009641e+01, 5.74153236403792e+01, 5.72207390574847e+01, 5.70487673879151e+01, 5.68982893169407e+01, 5.67680216980545e+01, 5.66565477725264e+01, 5.65623483985083e+01, 5.64838338513894e+01, 5.64193757003648e+01, 5.63673382069877e+01, 5.63261086376682e+01, 5.62941258414218e+01, 5.62699064238142e+01, 5.62520678539545e+01, 5.62393478777748e+01, 5.62306196797305e+01, 5.62249023359671e+01, 5.62213662318676e+01, 5.62193332702282e+01, 5.62182718656873e+01, 5.62177868976804e+01, 5.62176049688709e+01, 5.62175554797700e+01, 5.62175481752804e+01, 5.62175475893700e+01, 5.62175367763510e+01, 5.62174635165515e+01, 5.62171942057114e+01, 5.62164763213705e+01, 5.62149052464162e+01, 5.62118964398377e+01, 5.62066640683655e+01, 5.61982070801681e+01, 5.61853035478787e+01, 5.61665139290384e+01, 5.61401936820705e+01, 5.61045154293745e+01, 5.60575005717175e+01, 5.59970599287162e+01, 5.59210426128952e+01, 5.58272919501871e+01, 5.57137068558785e+01, 5.55783066873364e+01, 5.54192972549306e+01, 5.52351354156971e+01, 5.50245895358320e+01, 5.47867931191390e+01, 5.45212890807614e+01, 5.42280625062228e+01, 5.39075602633441e+01, 5.35606964954752e+01, 5.31888437633922e+01, 5.27938103482171e+01, 5.23778049016058e+01, 5.19433901662245e+01, 5.14934278512175e+01, 5.10310169336942e+01, 5.05594277017918e+01, 5.00820338073304e+01, 4.96022444987364e+01, 4.91234390750225e+01, 4.86489054292144e+01, 4.81817843107785e+01, 4.77250206137670e+01, 4.72813225954341e+01, 4.68531294804484e+01, 4.64425874584415e+01, 4.60515336906320e+01, 4.56814876460546e+01, 4.53336489099534e+01, 4.50089005440076e+01, 4.47078171102108e+01, 4.44306765675051e+01, 4.41774753809125e+01, 4.39479463193970e+01, 4.37415785408492e+01, 4.35576396581795e+01, 4.33951995440846e+01, 4.32531556630986e+01, 4.31302597206302e+01, 4.30251453942003e+01, 4.29363568672973e+01, 4.28623778270895e+01, 4.28016605200472e+01, 4.27526543911712e+01, 4.27138337701328e+01, 4.26837240183892e+01, 4.26609255219980e+01, 4.26441349112325e+01, 4.26321629144715e+01, 4.26239483123976e+01, 4.26185675490918e+01, 4.26152396763376e+01, 4.26133264511347e+01, 4.26123275667925e+01, 4.26118711663762e+01, 4.26116999545143e+01, 4.26116533807074e+01, 4.26116465065101e+01, 4.26116459763058e+01, 4.26116364555394e+01, 4.26115719509467e+01, 4.26113348252507e+01, 4.26107027352847e+01, 4.26093194222537e+01, 4.26066702138376e+01, 4.26020632229316e+01, 4.25946171135282e+01, 4.25832561712368e+01, 4.25667132602279e+01, 4.25435410644465e+01, 4.25121317929679e+01, 4.24707452728347e+01, 4.24175450562446e+01, 4.23506418360518e+01, 4.22681431041088e+01, 4.21682076182176e+01, 4.20491028898178e+01, 4.19092635965110e+01, 4.17473485950536e+01, 4.15622940951695e+01, 4.13533605808495e+01, 4.11201712516674e+01, 4.08627401045729e+01, 4.05814882702096e+01, 4.02772478206639e+01, 3.99512529239477e+01, 3.96051188700078e+01, 3.92408100689996e+01, 3.88605985726231e+01, 3.84670149650181e+01, 3.80627936128924e+01, 3.76508142854154e+01, 3.72340420996553e+01, 3.68154676614020e+01, 3.63980491772488e+01, 3.59846582036138e+01, 3.55780305379227e+01, 3.51807235072203e+01, 3.47950805521380e+01, 3.44232035609919e+01, 3.40669329394110e+01, 3.37278349819858e+01, 3.34071958084633e+01, 3.31060209683142e+01, 3.28250397951689e+01, 3.25647136696487e+01, 3.23252474787266e+01, 3.21066037013441e+01, 3.19085186771969e+01, 3.17305207164030e+01, 3.15719497802684e+01, 3.14319785104590e+01, 3.13096344093464e+01, 3.12038229809968e+01, 3.11133516320031e+01, 3.10369541055569e+01, 3.09733151828384e+01, 3.09210953362349e+01, 3.08789549638549e+01, 3.08455777803857e+01, 3.08196928924833e+01, 3.08000950546063e+01, 3.07856625898670e+01, 3.07753724749812e+01, 3.07683121316120e+01, 3.07636875387496e+01, 3.07608273801063e+01, 3.07591830622763e+01, 3.07583245770997e+01, 3.07579323274283e+01, 3.07577851809208e+01, 3.07577451534922e+01, 3.07577392455272e+01, 3.07577387981939e+01, 3.07577308737296e+01, 3.07576771843224e+01, 3.07574798164511e+01, 3.07569537066992e+01, 3.07558023310618e+01, 3.07535973192720e+01, 3.07497628175014e+01, 3.07435653298159e+01, 3.07341096604632e+01, 3.07203414518307e+01, 3.07010566618171e+01, 3.06749181443102e+01, 3.06404792835054e+01, 3.05962143857423e+01, 3.05405552543581e+01, 3.04719330721457e+01, 3.03888244067901e+01, 3.02897998574253e+01, 3.01735736001213e+01, 3.00390518939273e+01, 2.98853785032048e+01, 2.97119749974925e+01, 2.95185740190174e+01, 2.93052438599593e+01, 2.90724030529903e+01, 2.88208241236704e+01, 2.85516261483413e+01, 2.82662562707469e+01, 2.79664608235193e+01, 2.76542471539201e+01, 2.73318376520297e+01, 2.70016178135748e+01, 2.66660804272040e+01, 2.63277681405237e+01, 2.59892167091150e+01, 2.56529011475389e+01, 2.53211867703693e+01, 2.49962867419973e+01, 2.46802272756252e+01, 2.43748210828906e+01, 2.40816491348722e+01, 2.38020503108038e+01, 2.35371181287814e+01, 2.32877035001392e+01, 2.30544223317167e+01, 2.28376668048947e+01, 2.26376192608848e+01, 2.24542677859106e+01, 2.22874227852213e+01, 2.21367340334449e+01, 2.20017078696641e+01, 2.18817243553084e+01, 2.17760543247219e+01, 2.16838763305491e+01, 2.16042935210191e+01, 2.15363504881994e+01, 2.14790501010573e+01, 2.14313702910578e+01, 2.13922806977758e+01, 2.13607590144712e+01, 2.13358068057163e+01, 2.13164645077623e+01, 2.13018252737723e+01, 2.12910472959124e+01, 2.12833642289080e+01, 2.12780933576916e+01, 2.12746411958135e+01, 2.12725062697865e+01, 2.12712789338039e+01, 2.12706381638154e+01, 2.12703453929994e+01, 2.12702355649283e+01, 2.12702056890649e+01, 2.12702012794513e+01, 2.12702009484202e+01, 2.12701951218985e+01, 2.12701556463774e+01, 2.12700105304997e+01, 2.12696237072562e+01, 2.12687771662648e+01, 2.12671559858018e+01, 2.12643368858487e+01, 2.12597808641028e+01, 2.12528304068503e+01, 2.12427116812247e+01, 2.12285420086594e+01, 2.12093427849240e+01, 2.11840578446848e+01, 2.11515770653368e+01, 2.11107647673750e+01, 2.10604922037469e+01, 2.09996731517091e+01, 2.09273013473028e+01, 2.08424882597114e+01, 2.07444995189535e+01, 2.06327882144478e+01, 2.05070232992887e+01, 2.03671114830549e+01, 2.02132112800295e+01, 2.00457382899508e+01, 1.98653612980592e+01, 1.96729893476330e+01, 1.94697505067436e+01, 1.92569635617833e+01, 1.90361042673312e+01, 1.88087680223978e+01, 1.85766309057391e+01, 1.83414108924546e+01, 1.81048308204751e+01, 1.78685843279237e+01, 1.76343055985451e+01, 1.74035433869250e+01, 1.71777394887282e+01, 1.69582115943018e+01, 1.67461403170232e+01, 1.65425601055602e+01, 1.63483537093249e+01, 1.61642498475300e+01, 1.59908237201634e+01, 1.58284999888054e+01, 1.56775578489092e+01, 1.55381378188807e+01, 1.54102498902666e+01, 1.52937827192908e+01, 1.51885135901535e+01, 1.50941189387615e+01, 1.50101852841228e+01, 1.49362204660829e+01, 1.48716651266981e+01, 1.48159043949884e+01, 1.47682797401428e+01, 1.47281009475285e+01, 1.46946581476924e+01, 1.46672337945958e+01, 1.46451144499010e+01, 1.46276021898856e+01, 1.46140254151602e+01, 1.46037488152536e+01, 1.45961822241384e+01, 1.45907881018799e+01, 1.45870873935836e+01, 1.45846635500982e+01, 1.45831645444046e+01, 1.45823027807979e+01, 1.45818528671198e+01, 1.45816472988917e+01, 1.45815701833218e+01, 1.45815492060330e+01, 1.45815461098297e+01, 1.45815458781035e+01, 1.45815418088763e+01, 1.45815142392735e+01, 1.45814128906581e+01, 1.45811427336948e+01, 1.45805515080557e+01, 1.45794192642432e+01, 1.45774503579013e+01, 1.45742682835566e+01, 1.45694136836614e+01, 1.45623458073164e+01, 1.45524476174355e+01, 1.45390346514622e+01, 1.45213676258608e+01, 1.44986686369550e+01, 1.44701406514781e+01, 1.44349898041530e+01, 1.43924498354929e+01, 1.43418078235641e+01, 1.42824302047256e+01, 1.42137879583701e+01, 1.41354797673631e+01, 1.40472519747083e+01, 1.39490142485340e+01, 1.38408500452415e+01, 1.37230212193093e+01, 1.35959664529621e+01, 1.34602935454480e+01, 1.33167659777057e+01, 1.31662845160728e+01, 1.30098648999817e+01, 1.28486128403594e+01, 1.26836976173342e+01, 1.25163255059492e+01, 1.23477140956289e+01, 1.21790683390166e+01, 1.20115589126225e+01, 1.18463032363466e+01, 1.16843493090840e+01, 1.15266623833155e+01, 1.13741144170291e+01, 1.12274761913836e+01, 1.10874119505037e+01, 1.09544763933450e+01, 1.08291138211671e+01, 1.07116592184270e+01, 1.06023410240298e+01, 1.05012853386344e+01, 1.04085213153294e+01, 1.03239874960479e+01, 1.02475388825829e+01, 1.01789545651023e+01, 1.01179457678389e+01, 1.00641642063635e+01, 1.00172106794719e+01, 9.97664383830713e+00, 9.94198908432094e+00, 9.91274754579659e+00, 9.88840507088409e+00, 9.86844115543528e+00, 9.85233769915544e+00, 9.83958745702640e+00, 9.82970202813974e+00, 9.82221920447405e+00, 9.81670949082059e+00, 9.81278160639972e+00, 9.81008679017505e+00, 9.80832175585215e+00, 9.80723017828865e+00, 9.80660263872549e+00, 9.80627500917149e+00, 9.80612531312796e+00, 9.80606915707195e+00, 9.80605388127236e+00, 9.80605162659668e+00, 9.80605145778902e+00, 9.80604849259781e+00, 9.80602840299795e+00, 9.80595455160599e+00, 9.80575769169521e+00, 9.80532687258197e+00, 9.80450181771802e+00, 9.80306708820220e+00, 9.80074831228744e+00, 9.79721072376664e+00, 9.79206020699995e+00, 9.78484699130503e+00, 9.77507206987397e+00, 9.76219633620012e+00, 9.74565233328397e+00, 9.72485839856439e+00, 9.69923486389389e+00, 9.66822184091436e+00, 9.63129799645952e+00, 9.58799961079006e+00, 9.53793912538983e+00, 9.48082233838338e+00, 9.41646340431377e+00, 9.34479684769990e+00, 9.26588590858410e+00, 9.17992669996930e+00, 9.08724786300259e+00, 8.98830564250366e+00, 8.88367455575772e+00, 8.77403407201487e+00, 8.66015193909067e+00, 8.54286496858481e+00, 8.42305820770507e+00, 8.30164347388634e+00, 8.17953820536421e+00, 8.05764549099153e+00, 7.93683599736482e+00, 7.81793232793412e+00, 7.70169614763857e+00, 7.58881820844879e+00, 7.47991123403483e+00, 7.37550547872512e+00, 7.27604667399915e+00, 7.18189601599087e+00, 7.09333182595282e+00, 7.01055252518184e+00, 6.93368059790043e+00, 6.86276726145611e+00, 6.79779761548225e+00, 6.73869609450132e+00, 6.68533209766312e+00, 6.63752571217675e+00, 6.59505348190578e+00, 6.55765419874658e+00, 6.52503471153187e+00, 6.49687575543381e+00, 6.47283780466786e+00, 6.45256694355240e+00, 6.43570073685810e+00, 6.42187406146876e+00, 6.41072483959765e+00, 6.40189959137011e+00, 6.39505870384936e+00, 6.38988129688119e+00, 6.38606955557776e+00, 6.38335239653026e+00, 6.38148834100197e+00, 6.38026748372398e+00, 6.37951246999656e+00, 6.37907842531142e+00, 6.37885181870858e+00, 6.37874828115675e+00, 6.37870944078837e+00, 6.37869887527981e+00, 6.37869731583335e+00, 6.37869719876778e+00, 6.37869513831598e+00, 6.37868117845962e+00, 6.37862986066851e+00, 6.37849306718178e+00, 6.37819370256442e+00, 6.37762040106828e+00, 6.37662347990492e+00, 6.37501233779619e+00, 6.37255447365014e+00, 6.36897627024386e+00, 6.36396565020998e+00, 6.35717666411614e+00, 6.34823601109112e+00, 6.33675142033512e+00, 6.32232173753880e+00, 6.30454846631635e+00, 6.28304841611802e+00, 6.25746701187200e+00, 6.22749173583748e+00, 6.19286510885836e+00, 6.15339658624335e+00, 6.10897275102090e+00, 6.05956523928917e+00, 6.00523592927078e+00, 5.94613906274002e+00, 5.88252013479965e+00, 5.81471157143230e+00, 5.74312539731834e+00, 5.66824326250925e+00, 5.59060433142074e+00, 5.51079163130549e+00, 5.42941750528858e+00, 5.34710881792405e+00, 5.26449252407669e+00, 5.18218214240537e+00, 5.10076558146237e+00, 5.02079465780228e+00, 4.94277652914224e+00, 4.86716714848161e+00, 4.79436673380940e+00, 4.72471714904167e+00, 4.65850101108225e+00, 4.59594228005403e+00, 4.53720805734093e+00, 4.48241130889511e+00, 4.43161424632689e+00, 4.38483213046065e+00, 4.34203730500735e+00, 4.30316331541746e+00, 4.26810901433996e+00, 4.23674259630553e+00, 4.20890553769038e+00, 4.18441644248644e+00, 4.16307480975674e+00, 4.14466474546431e+00, 4.12895864064815e+00, 4.11572083093391e+00, 4.10471124051053e+00, 4.09568899845428e+00, 4.08841599819594e+00, 4.08266035360548e+00, 4.07819968922827e+00, 4.07482418919821e+00, 4.07233932066930e+00, 4.07056814435431e+00, 4.06935312765013e+00, 4.06855738508125e+00, 4.06806528608027e+00, 4.06778239058958e+00, 4.06763469729465e+00, 4.06756721584084e+00, 4.06754190135247e+00, 4.06753501521057e+00, 4.06753399883091e+00, 4.06753392220772e+00, 4.06753256924652e+00, 4.06752340274251e+00, 4.06748970580924e+00, 4.06739988296448e+00, 4.06720331196352e+00, 4.06682687068219e+00, 4.06617228526702e+00, 4.06511443310747e+00, 4.06350071909188e+00, 4.06115162477320e+00, 4.05786250579695e+00, 4.05340668142726e+00, 4.04753981971882e+00, 4.04000557199602e+00, 4.03054235112823e+00, 4.01889108145507e+00, 4.00480367768299e+00, 3.98805194096005e+00, 3.96843649942222e+00, 3.94579537544307e+00, 3.92001174020909e+00, 3.89102042450649e+00, 3.85881279676700e+00, 3.82343969602567e+00, 3.78501221482510e+00, 3.74370025722564e+00, 3.69972893812721e+00, 3.65337302781731e+00, 3.60494976534572e+00, 3.55481045321775e+00, 3.50333129535883e+00, 3.45090394737032e+00, 3.39792621589156e+00, 3.34479328074211e+00, 3.29188973100291e+00, 3.23958261658484e+00, 3.18821563078322e+00, 3.13810446429959e+00, 3.08953331104473e+00, 3.04275246136268e+00, 2.99797688767183e+00, 2.95538570850914e+00, 2.91512240715798e+00, 2.87729567846009e+00, 2.84198078061702e+00, 2.80922127671803e+00, 2.77903106247777e+00, 2.75139659128176e+00, 2.72627922405080e+00, 2.70361764847166e+00, 2.68333032860683e+00, 2.66531796067279e+00, 2.64946592293467e+00, 2.63564671651538e+00, 2.62372239906510e+00, 2.61354701458045e+00, 2.60496902039441e+00, 2.59783370693549e+00, 2.59198559797400e+00, 2.58727080962163e+00, 2.58353933635185e+00, 2.58064722285584e+00, 2.57845857272088e+00, 2.57684733968964e+00, 2.57569884540953e+00, 2.57491096961274e+00, 2.57439496474763e+00, 2.57407585701280e+00, 2.57389240897812e+00, 2.57379663466747e+00, 2.57375287507251e+00, 2.57373645941512e+00, 2.57373199396558e+00, 2.57373133487486e+00, 2.57373128491669e+00, 2.57373039920211e+00, 2.57372439835973e+00, 2.57370233869278e+00, 2.57364353621480e+00, 2.57351485094114e+00, 2.57326841286964e+00, 2.57283988510655e+00, 2.57214735115824e+00, 2.57109090337921e+00, 2.56955299712962e+00, 2.56739961929077e+00, 2.56448229930130e+00, 2.56064096486823e+00, 2.55570761248926e+00, 2.54951072512412e+00, 2.54188032694306e+00, 2.53265352030972e+00, 2.52168030632737e+00, 2.50882945156994e+00, 2.49399413466696e+00, 2.47709709175712e+00, 2.45809498321925e+00, 2.43698172781917e+00, 2.41379059475150e+00, 2.38859490696223e+00, 2.36150728630294e+00, 2.33267745630485e+00, 2.30228870431563e+00, 2.27055318377712e+00, 2.23770630253136e+00, 2.20400048869392e+00, 2.16969864835777e+00, 2.13506762813682e+00, 2.10037197171752e+00, 2.06586821674668e+00, 2.03179992181629e+00, 1.99839354926922e+00, 1.96585526449215e+00, 1.93436865213718e+00, 1.90409329894237e+00, 1.87516415447479e+00, 1.84769155639512e+00, 1.82176179530529e+00, 1.79743809418467e+00, 1.77476188633281e+00, 1.75375429079047e+00, 1.73441770266231e+00, 1.71673743525572e+00, 1.70068336964943e+00, 1.68621158394254e+00, 1.67326594825109e+00, 1.66177968215452e+00, 1.65167687869058e+00, 1.64287400327834e+00, 1.63528137736705e+00, 1.62880465548697e+00, 1.62334630109434e+00, 1.61880706159552e+00, 1.61508743669054e+00, 1.61208912725201e+00, 1.60971644494772e+00, 1.60787765636593e+00, 1.60648623012683e+00, 1.60546195194628e+00, 1.60473187131045e+00, 1.60423104463241e+00, 1.60390304358850e+00, 1.60370020364878e+00, 1.60358359627206e+00, 1.60352271829250e+00, 1.60349490300153e+00, 1.60348446858394e+00, 1.60348163017488e+00, 1.60348121123187e+00, 1.60348117927504e+00, 1.60348061005058e+00, 1.60347675347433e+00, 1.60346257634068e+00, 1.60342478568059e+00, 1.60334208363867e+00, 1.60318370658935e+00, 1.60290831027154e+00, 1.60246325754140e+00, 1.60178435914580e+00, 1.60079611011795e+00, 1.59941245423194e+00, 1.59753809561855e+00, 1.59507035945379e+00, 1.59190158221155e+00, 1.58792198646179e+00, 1.58302296641750e+00, 1.57710067999059e+00, 1.57005981340200e+00, 1.56181735843646e+00, 1.55230622362424e+00, 1.54147849226446e+00, 1.52930814495852e+00, 1.51579308370836e+00, 1.50095632855135e+00, 1.48484630419019e+00, 1.46753618935558e+00, 1.44912236050651e+00, 1.42972201797146e+00, 1.40947013093997e+00, 1.38851587303240e+00, 1.36701873950828e+00, 1.34514453972990e+00, 1.32306144572264e+00, 1.30093625283487e+00, 1.27893097597477e+00, 1.25719986931485e+00, 1.23588692278412e+00, 1.21512385807647e+00, 1.19502862190400e+00, 1.17570435517001e+00, 1.15723880307169e+00, 1.13970412185218e+00, 1.12315703196359e+00, 1.10763926399179e+00, 1.09317824241120e+00, 1.07978795296297e+00, 1.06746994219579e+00, 1.05621440244373e+00, 1.04600130202110e+00, 1.03680152824100e+00, 1.02857801937104e+00, 1.02128687007743e+00, 1.01487840255066e+00, 1.00929820173592e+00, 1.00448811747948e+00, 1.00038723873205e+00, 9.96932845221181e-01, 9.94061340402453e-01, 9.91709166364785e-01, 9.89813697157643e-01, 9.88314102253938e-01, 9.87152167126283e-01, 9.86273053745848e-01, 9.85625980704839e-01, 9.85164801015344e-01, 9.84848455708843e-01, 9.84641283255726e-01, 9.84513168473228e-01, 9.84439519760414e-01, 9.84401069819484e-01, 9.84383502026333e-01, 9.84376911789539e-01, 9.84375119089961e-01, 9.84374854491449e-01, 9.84374834165536e-01, 9.84374470247189e-01, 9.84372004649572e-01, 9.84362940891884e-01, 9.84338780548626e-01, 9.84285907786415e-01, 9.84184656091969e-01, 9.84008597230257e-01, 9.83724089474975e-01, 9.83290118538264e-01, 9.82658459969542e-01, 9.81774185549256e-01, 9.80576527803024e-01, 9.79000105716239e-01, 9.76976500762760e-01, 9.74436155537788e-01, 9.71310548105464e-01, 9.67534574643096e-01, 9.63049052579080e-01, 9.57803238092581e-01, 9.51757237698089e-01, 9.44884185808391e-01, 9.37172060467645e-01, 9.28625019105198e-01, 9.19264155585935e-01, 9.09127608480162e-01, 8.98269986824161e-01, 8.86761121291613e-01, 8.74684192525891e-01, 8.62133330758070e-01, 8.49210817846114e-01, 8.36024050632034e-01, 8.22682439576141e-01, 8.09294416542883e-01, 7.95964709559200e-01, 7.82792011680201e-01, 7.69867129435099e-01, 7.57271649260289e-01, 7.45077114325923e-01, 7.33344665268518e-01, 7.22125070859377e-01, 7.11459060394517e-01, 7.01377867860643e-01, 6.91903906057405e-01, 6.83051503214529e-01, 6.74827651611309e-01, 6.67232734364433e-01, 6.60261210996850e-01, 6.53902253734714e-01, 6.48140334582394e-01, 6.42955768453305e-01, 6.38325220561399e-01, 6.34222187485975e-01, 6.30617461313648e-01, 6.27479585400110e-01, 6.24775308829612e-01, 6.22470044739899e-01, 6.20528335426656e-01, 6.18914324632968e-01, 6.17592234758831e-01, 6.16526844009272e-01, 6.15683955887647e-01, 6.15030851098311e-01, 6.14536710042842e-01, 6.14172992847287e-01, 6.13913763404241e-01, 6.13735944355625e-01, 6.13619491317522e-01, 6.13547476920274e-01, 6.13506078284757e-01, 6.13484465188876e-01, 6.13474590151839e-01, 6.13470885711209e-01, 6.13469878016062e-01, 6.13469729282443e-01, 6.13469717769888e-01, 6.13469510511337e-01, 6.13468106305760e-01, 6.13462944320389e-01, 6.13449184538396e-01, 6.13419072484135e-01, 6.13361407715800e-01, 6.13261138892483e-01, 6.13099106512038e-01, 6.12851952336769e-01, 6.12492211817340e-01, 6.11988602154270e-01, 6.11306513639704e-01, 6.10408705290951e-01, 6.09256197347136e-01, 6.07809343010231e-01, 6.06029050243819e-01, 6.03878112231314e-01, 6.01322593338963e-01, 5.98333207507821e-01, 5.94886619418205e-01, 5.90966596925000e-01, 5.86564947162074e-01, 5.81682178716036e-01, 5.76327847904295e-01, 5.70520567117768e-01, 5.64287675344588e-01, 5.57664592915060e-01, 5.50693901771339e-01, 5.43424207209583e-01, 5.35908845912871e-01, 5.28204507966734e-01, 5.20369838044331e-01, 5.12464074261871e-01, 5.04545773762160e-01, 4.96671663245081e-01, 4.88895641511874e-01, 4.81267950389551e-01, 4.73834520632622e-01, 4.66636490854455e-01, 4.59709890415255e-01, 4.53085471628569e-01, 4.46788672733007e-01, 4.40839690834267e-01, 4.35253643376214e-01, 4.30040797468678e-01, 4.25206848312456e-01, 4.20753230678514e-01, 4.16677450569139e-01, 4.12973427473062e-01, 4.09631840739706e-01, 4.06640476322224e-01, 4.03984572325498e-01, 4.01647163367912e-01, 3.99609424699790e-01, 3.97851017339619e-01, 3.96350435248905e-01, 3.95085354844937e-01, 3.94032986051045e-01, 3.93170422718256e-01, 3.92474988752269e-01, 3.91924574776738e-01, 3.91497958806099e-01, 3.91175103318560e-01, 3.90937420440082e-01, 3.90767996770165e-01, 3.90651769760311e-01, 3.90575648519558e-01, 3.90528573428398e-01, 3.90501510924594e-01, 3.90487382148478e-01, 3.90480926660118e-01, 3.90478504994292e-01, 3.90477846243485e-01, 3.90477749013272e-01, 3.90477741426253e-01, 3.90477604050403e-01, 3.90476673309550e-01, 3.90473251818719e-01, 3.90464131464611e-01, 3.90444172220342e-01, 3.90405949518140e-01, 3.90339485044535e-01, 3.90232074544684e-01, 3.90068224699760e-01, 3.89829709228920e-01, 3.89495751895822e-01, 3.89043340871567e-01, 3.88447674764549e-01, 3.87682735484276e-01, 3.86721976995531e-01, 3.85539112138474e-01, 3.84108972424769e-01, 3.82408408640117e-01, 3.80417193915406e-01, 3.78118886479399e-01, 3.75501607358700e-01, 3.72558689481608e-01, 3.69289159331942e-01, 3.65698020464123e-01, 3.61796319376495e-01, 3.57600987567096e-01, 3.53134467844782e-01, 3.48424146732114e-01, 3.43501626670786e-01, 3.38401880525495e-01, 3.33162335759255e-01, 3.27821936299515e-01, 3.22420226743986e-01, 3.16996496858238e-01, 3.11589015345206e-01, 3.06234371836837e-01, 3.00966936175826e-01, 2.95818435322559e-01, 2.90817641337341e-01, 2.85990159179022e-01, 2.81358300517283e-01, 2.76941029092171e-01, 2.72753963927101e-01, 2.68809428402037e-01, 2.65116535354338e-01, 2.61681300615336e-01, 2.58506779448908e-01, 2.55593222086077e-01, 2.52938245890874e-01, 2.50537022649496e-01, 2.48382480093539e-01, 2.46465517089902e-01, 2.44775232009104e-01, 2.43299163649823e-01, 2.42023543780178e-01, 2.40933559875657e-01, 2.40013626011375e-01, 2.39247659134548e-01, 2.38619357141999e-01, 2.38112474371377e-01, 2.37711089351748e-01, 2.37399859020741e-01, 2.37164253176104e-01, 2.36990762759959e-01, 2.36867075725115e-01, 2.36782214738625e-01, 2.36726631841790e-01, 2.36692256385581e-01, 2.36672494039126e-01, 2.36662176350222e-01, 2.36657462121343e-01, 2.36655693652722e-01, 2.36655212586498e-01, 2.36655141582133e-01, 2.36655135994425e-01, 2.36655034215190e-01, 2.36654344646212e-01, 2.36651809721508e-01, 2.36645052572513e-01, 2.36630264867102e-01, 2.36601945216492e-01, 2.36552698903026e-01, 2.36473108588442e-01, 2.36351684374412e-01, 2.36174900033362e-01, 2.35927320315042e-01, 2.35591821856943e-01, 2.35149907307729e-01, 2.34582108768677e-01, 2.33868472610069e-01, 2.32989113262022e-01, 2.31924818964245e-01, 2.30657688047074e-01, 2.29171770559176e-01, 2.27453687441965e-01, 2.25493198435316e-01, 2.23283690822171e-01, 2.20822564132803e-01, 2.18111490925480e-01, 2.15156540370089e-01, 2.11968158981207e-01, 2.08561010729980e-01, 2.04953686130807e-01, 2.01168296085498e-01, 1.97229970813751e-01, 1.93166286924580e-01, 1.89006646679530e-01, 1.84781633052566e-01, 1.80522362712210e-01, 1.76259856914015e-01, 1.72024447808482e-01, 1.67845235001650e-01, 1.63749604427267e-01, 1.59762818689956e-01, 1.55907685018633e-01, 1.52204303855186e-01, 1.48669897978621e-01, 1.45318719083434e-01, 1.42162026070238e-01, 1.39208127162178e-01, 1.36462476484823e-01, 1.33927815017188e-01, 1.31604345829203e-01, 1.29489934182814e-01, 1.27580324226827e-01, 1.25869365492207e-01, 1.24349243999404e-01, 1.23010714345355e-01, 1.21843330498351e-01, 1.20835674083793e-01, 1.19975579612785e-01, 1.19250356384015e-01, 1.18647006681812e-01, 1.18152439463515e-01, 1.17753678075559e-01, 1.17438059733018e-01, 1.17193423706809e-01, 1.17008284444395e-01, 1.16871985356951e-01, 1.16774828774794e-01, 1.16708177690086e-01, 1.16664525373764e-01, 1.16637529747933e-01, 1.16622010500336e-01, 1.16613908215825e-01, 1.16610206252091e-01, 1.16608817523613e-01, 1.16608439756391e-01, 1.16608383998667e-01, 1.16608379571986e-01, 1.16608298446508e-01, 1.16607748810679e-01, 1.16605728298282e-01, 1.16600342407025e-01, 1.16588555783196e-01, 1.16565984015722e-01, 1.16526734737732e-01, 1.16463306105629e-01, 1.16366549841118e-01, 1.16225704934450e-01, 1.16028506685282e-01, 1.15761373937161e-01, 1.15409675013979e-01, 1.14958069984382e-01, 1.14390923452476e-01, 1.13692778210392e-01, 1.12848875983346e-01, 1.11845707390017e-01, 1.10671569528492e-01, 1.09317106594872e-01, 1.07775807117278e-01, 1.06044431003994e-01, 1.04123340982454e-01, 1.02016716227453e-01, 9.97326310063187e-02, 9.72829878037089e-02, 9.46833022047855e-02, 9.19523453454190e-02, 8.91116582543674e-02, 8.61849603327300e-02, 8.31974807874474e-02, 8.01752465014403e-02, 7.71443621113406e-02, 7.41303176415187e-02, 7.11573560113123e-02, 6.82479273127541e-02, 6.54222495155636e-02, 6.26979870800994e-02, 6.00900506470623e-02, 5.76105131872566e-02, 5.52686315353981e-02, 5.30709573541941e-02, 5.10215184418428e-02, 4.91220502909897e-02, 4.73722583803351e-02, 4.57700936666328e-02, 4.43120267284911e-02, 4.29933091819947e-02, 4.18082146505299e-02, 4.07502541201978e-02, 3.98123638183705e-02, 3.89870658240454e-02, 3.82666033775223e-02, 3.76430542794708e-02, 3.71084262433641e-02, 3.66413323847610e-02, 3.62611084471638e-02, 3.59461000634553e-02, 3.56887819857405e-02, 3.54818986858021e-02, 3.53185238309095e-02, 3.51921185496877e-02, 3.50965873934683e-02, 3.50263301349685e-02, 3.49762870003190e-02, 3.49419746422797e-02, 3.49195101614583e-02, 3.49056207562910e-02, 3.48976371075622e-02, 3.48934693237766e-02, 3.48915651203328e-02, 3.48908508027108e-02, 3.48906564923510e-02, 3.48906278126009e-02, 3.48906255148336e-02, 3.48905831424275e-02, 3.48902960638782e-02, 3.48892407453576e-02, 3.48864277506067e-02, 3.48802720719448e-02, 3.48684850948473e-02, 3.48479933377064e-02, 3.48148890363831e-02, 3.47644175100896e-02, 3.46910059512879e-02, 3.45883376881488e-02, 3.44494749279676e-02, 3.42670313784623e-02, 3.40333938740969e-02, 3.37409892016467e-02, 3.33825888157406e-02, 3.29516402751580e-02, 3.24426103492073e-02, 3.18513212789497e-02, 3.11752591349119e-02, 3.04138320931315e-02, 2.95685571880284e-02, 2.86431569583527e-02, 2.76435524135650e-02, 2.65777456574909e-02, 2.54555937526959e-02, 2.42884841831312e-02, 2.30889305873632e-02, 2.18701142858322e-02, 2.06454016351358e-02, 1.94278688169902e-02, 1.82298641117313e-02, 1.70626332526069e-02, 1.59360267429918e-02, 1.48582999642457e-02, 1.38360085502001e-02, 1.28739938450195e-02, 1.19754470784983e-02, 1.11420366711457e-02, 1.03740809713884e-02, 9.67074858181149e-03, 9.03026988372346e-03, 8.45014593062388e-03, 7.92734403765394e-03, 7.45847269111266e-03, 7.03993148106008e-03, 6.66803439158059e-03, 6.33910685186834e-03, 6.04955843412824e-03, 5.79593403253398e-03, 5.57494685303084e-03, 5.38349669020022e-03, 5.21867686555311e-03, 5.07777293510969e-03, 4.95825591457333e-03, 4.85777236313822e-03, 4.77413324543363e-03, 4.70530307133288e-03, 4.64939040743673e-03, 4.60464046473419e-03, 4.56943010148336e-03, 4.54226524330181e-03, 4.52178042317541e-03, 4.50673989438144e-03, 4.49603958104206e-03, 4.48870901499276e-03, 4.48391237178214e-03, 4.48094776483288e-03, 4.47924407683331e-03, 4.47835479148750e-03, 4.47794851380934e-03, 4.47779611215537e-03, 4.47775465589473e-03, 4.47774853707017e-03, 4.47774804221979e-03, 4.47773885926495e-03, 4.47767664387300e-03, 4.47744793945049e-03, 4.47683834278223e-03, 4.47550448576806e-03, 4.47295087263910e-03, 4.46851287890127e-03, 4.46134730921699e-03, 4.45043199336226e-03, 4.43457592190555e-03, 4.41244132076690e-03, 4.38257876899684e-03, 4.34347591075834e-03, 4.29361946200402e-03, 4.23156906373379e-03, 4.15604014014548e-03, 4.06599140656283e-03, 3.96071122662534e-03, 3.83989588345912e-03, 3.70371226669874e-03, 3.55283771956749e-03, 3.38847098413453e-03, 3.21231033925073e-03, 3.02649797998811e-03, 2.83458024064487e-03, 2.63727217647158e-03, 2.43841289385216e-03, 2.24082874987630e-03, 2.04721713967615e-03, 1.86001101481281e-03, 1.68129535649086e-03, 1.51275559599640e-03, 1.35561611656954e-03, 1.21062735385635e-03, 1.07811510932265e-03, 9.58031117136215e-04, 8.50021944813761e-04, 7.53522937685602e-04, 6.67819827938843e-04, 5.92057444848137e-04, 5.25344488792712e-04, 4.66933354488219e-04, 4.15536365892422e-04, 3.70084959371297e-04, 3.30734451708281e-04, 2.95774201518384e-04, 2.64456616406505e-04, 2.36175432249297e-04, 2.10468100263483e-04, 1.86993865554541e-04, 1.65475845349761e-04, 1.46307021665848e-04, 1.28681589501412e-04, 1.12373105054075e-04, 9.73601553226899e-05, 8.36294160677653e-05, 7.11734515732587e-05, 5.99633179234615e-05, 4.99291680426950e-05, 4.10072948540119e-05, 3.31501893668928e-05, 2.63142417082918e-05, 2.04524433939923e-05, 1.55108261026271e-05, 1.14270748525769e-05, 8.13057829418540e-06, 5.54340523706287e-06, 3.58182231252613e-06, 2.15808761083736e-06, 1.18237122137543e-06, 5.64748361853162e-07, 0.00000000000000e+00, 1.42108547152020e-14, 0.00000000000000e+00],
[ 8.80000000000000e+01, 8.79999998958584e+01, 8.79999966755927e+01, 8.79999748577816e+01, 8.79998946526349e+01, 8.79996808508218e+01, 8.79992129287642e+01, 8.79983167176768e+01, 8.79967579350980e+01, 8.79942378194624e+01, 8.79903910426622e+01, 8.79847860054293e+01, 8.79769275491535e+01, 8.79662620491685e+01, 8.79521847922382e+01, 8.79340494880586e+01, 8.79111797229335e+01, 8.78828821334766e+01, 8.78484610570363e+01, 8.78072343990131e+01, 8.77585504389781e+01, 8.77018052704917e+01, 8.76364605278724e+01, 8.75620609942262e+01, 8.74782516112736e+01, 8.73847933315729e+01, 8.72815771821656e+01, 8.71686358639006e+01, 8.70461522115296e+01, 8.69144639009617e+01, 8.67740639185461e+01, 8.66255964987609e+01, 8.64698484756499e+01, 8.63077362548886e+01, 8.61402888675889e+01, 8.59686277844986e+01, 8.57939443266159e+01, 8.56174755917363e+01, 8.54404798238587e+01, 8.52642120921676e+01, 8.50899010348409e+01, 8.49187272805736e+01, 8.47518040078200e+01, 8.45901599554906e+01, 8.44347250711746e+01, 8.42863188801398e+01, 8.41456415814844e+01, 8.40132678241398e+01, 8.38896430799575e+01, 8.37750825080952e+01, 8.36697721889082e+01, 8.35737725921781e+01, 8.34870241306724e+01, 8.34093546339719e+01, 8.33404885585458e+01, 8.32800577284046e+01, 8.32276133770905e+01, 8.31826392374737e+01, 8.31445654022394e+01, 8.31127826566377e+01, 8.30866569676694e+01, 8.30655438020051e+01, 8.30488019401006e+01, 8.30358064575266e+01, 8.30259605575098e+01, 8.30187059617430e+01, 8.30135315998370e+01, 8.30099803809658e+01, 8.30076538833331e+01, 8.30062148565441e+01, 8.30053874967676e+01, 8.30049555223365e+01, 8.30047581455261e+01, 8.30046841020029e+01, 8.30046639603063e+01, 8.30046609874318e+01, 8.30046605974534e+01, 8.30046515114239e+01, 8.30045899522235e+01, 8.30043636538789e+01, 8.30037604239500e+01, 8.30024402617457e+01, 8.29999119615406e+01, 8.29955151272174e+01, 8.29884084080516e+01, 8.29775646303233e+01, 8.29617733438893e+01, 8.29396511234395e+01, 8.29096597573321e+01, 8.28701322198921e+01, 8.28193060551671e+01, 8.27553635038285e+01, 8.26764773870958e+01, 8.25808614347362e+01, 8.24668234275670e+01, 8.23328192452406e+01, 8.21775057014442e+01, 8.19997899496830e+01, 8.17988732907458e+01, 8.15742874332288e+01, 8.13259216520819e+01, 8.10540398223632e+01, 8.07592869028243e+01, 8.04426850053857e+01, 8.01056196102740e+01, 7.97498167087304e+01, 7.93773116857700e+01, 7.89904106925445e+01, 7.85916452623622e+01, 7.81837211506337e+01, 7.77694628804232e+01, 7.73517561341356e+01, 7.69334906720884e+01, 7.65175065693746e+01, 7.61065460809988e+01, 7.57032124805615e+01, 7.53099361099459e+01, 7.49289470107332e+01, 7.45622531088159e+01, 7.42116229799103e+01, 7.38785725521496e+01, 7.35643554825002e+01, 7.32699572233509e+01, 7.29960929201136e+01, 7.27432092707186e+01, 7.25114903867660e+01, 7.23008675762027e+01, 7.21110328541762e+01, 7.19414558984549e+01, 7.17914041004995e+01, 7.16599653171733e+01, 7.15460728930559e+01, 7.14485324922265e+01, 7.13660502466827e+01, 7.12972616947449e+01, 7.12407609480614e+01, 7.11951294934353e+01, 7.11589640102320e+01, 7.11309025707035e+01, 7.11096485941640e+01, 7.10939919506981e+01, 7.10828266588307e+01, 7.10751646954337e+01, 7.10701455342927e+01, 7.10670411494504e+01, 7.10652563561404e+01, 7.10643245098910e+01, 7.10638987363241e+01, 7.10637390130992e+01, 7.10636955644281e+01, 7.10636891514923e+01, 7.10636885853522e+01, 7.10636774921143e+01, 7.10636023337824e+01, 7.10633260436016e+01, 7.10625895538657e+01, 7.10609777562066e+01, 7.10578909406462e+01, 7.10525228478938e+01, 7.10438463288090e+01, 7.10306074420934e+01, 7.10113286330354e+01, 7.09843214179582e+01, 7.09477087462224e+01, 7.08994569209183e+01, 7.08374166307033e+01, 7.07593722835091e+01, 7.06630984493245e+01, 7.05464218321275e+01, 7.04072868248546e+01, 7.02438223849980e+01, 7.00544077321819e+01, 6.98377342402341e+01, 6.95928608956791e+01, 6.93192608341295e+01, 6.90168567483097e+01, 6.86860433810225e+01, 6.83276958614505e+01, 6.79431632952682e+01, 6.75342477495263e+01, 6.71031695381546e+01, 6.66525204504223e+01, 6.61852071946209e+01, 6.57043877722176e+01, 6.52134036918627e+01, 6.47157108556773e+01, 6.42148116315270e+01, 6.37141901413293e+01, 6.32172522504885e+01, 6.27272712376968e+01, 6.22473397266246e+01, 6.17803281939242e+01, 6.13288502100606e+01, 6.08952344734036e+01, 6.04815036147330e+01, 6.00893596479547e+01, 5.97201758187378e+01, 5.93749944727106e+01, 5.90545304542251e+01, 5.87591794763742e+01, 5.84890308811246e+01, 5.82438842296859e+01, 5.80232692127076e+01, 5.78264684295334e+01, 5.76525426393723e+01, 5.75003581235882e+01, 5.73686158115433e+01, 5.72558818115927e+01, 5.71606189565312e+01, 5.70812189241131e+01, 5.70160344347977e+01, 5.69634109681583e+01, 5.69217173843098e+01, 5.68893747950930e+01, 5.68648830087721e+01, 5.68468438776615e+01, 5.68339809146509e+01, 5.68251546140811e+01, 5.68193730143503e+01, 5.68157971708897e+01, 5.68137413632073e+01, 5.68126680310146e+01, 5.68121776131745e+01, 5.68119936399459e+01, 5.68119435947137e+01, 5.68119362081403e+01, 5.68119356156457e+01, 5.68119246811159e+01, 5.68118505980637e+01, 5.68115782608648e+01, 5.68108523094027e+01, 5.68092635799556e+01, 5.68062209636956e+01, 5.68009297993566e+01, 5.67923777928362e+01, 5.67793293010827e+01, 5.67603286365920e+01, 5.67337128365791e+01, 5.66976340915047e+01, 5.66500917366666e+01, 5.65889733768965e+01, 5.65121043420982e+01, 5.64173042712368e+01, 5.63024492128329e+01, 5.61655372369890e+01, 5.60047552097143e+01, 5.58185441208645e+01, 5.56056602185184e+01, 5.53652292166011e+01, 5.50967910307533e+01, 5.48003328663156e+01, 5.44763090185323e+01, 5.41256464127816e+01, 5.37497356542397e+01, 5.33504080989593e+01, 5.29299001261445e+01, 5.24908063226433e+01, 5.20360236531176e+01, 5.15686888872253e+01, 5.10921116224444e+01, 5.06097052225626e+01, 5.01249179204763e+01, 4.96411662175315e+01, 4.91617725344164e+01, 4.86899088069532e+01, 4.82285473625354e+01, 4.77804199734835e+01, 4.73479855026743e+01, 4.69334060903151e+01, 4.65385314336925e+01, 4.61648904236206e+01, 4.58136892383337e+01, 4.54858149505790e+01, 4.51818437524106e+01, 4.49020530119898e+01, 4.46464365147627e+01, 4.44147223804993e+01, 4.42063932688527e+01, 4.40207085785305e+01, 4.38567284047614e+01, 4.37133390469171e+01, 4.35892798560499e+01, 4.34831711852564e+01, 4.33935431594070e+01, 4.33188649206093e+01, 4.32575739379790e+01, 4.32081049015452e+01, 4.31689176574668e+01, 4.31385235923245e+01, 4.31155098448665e+01, 4.30985607200867e+01, 4.30864757072212e+01, 4.30781835624171e+01, 4.30727520082526e+01, 4.30693927231695e+01, 4.30674614389308e+01, 4.30664531261252e+01, 4.30659924177618e+01, 4.30658195898395e+01, 4.30657725764249e+01, 4.30657656373423e+01, 4.30657651021336e+01, 4.30657554915011e+01, 4.30656903780532e+01, 4.30654510141426e+01, 4.30648129579301e+01, 4.30634165879756e+01, 4.30607423742225e+01, 4.30560918995561e+01, 4.30485755107930e+01, 4.30371073437494e+01, 4.30204083098907e+01, 4.29970174464049e+01, 4.29653118114760e+01, 4.29235348475027e+01, 4.28698328356118e+01, 4.28022987286324e+01, 4.27190222867874e+01, 4.26181450682042e+01, 4.24979184697440e+01, 4.23567627035163e+01, 4.21933243651052e+01, 4.20065301348405e+01, 4.17956341817331e+01, 4.15602570284297e+01, 4.13004139863030e+01, 4.10165317648899e+01, 4.07094524616832e+01, 4.03804247922756e+01, 4.00310830631139e+01, 3.96634149572133e+01, 3.92797196488765e+01, 3.88825580634776e+01, 3.84746972591932e+01, 3.80590509609590e+01, 3.76386182657577e+01, 3.72164224968021e+01, 3.67954521206832e+01, 3.63786055337092e+01, 3.59686413304602e+01, 3.55681353551602e+01, 3.51794454045625e+01, 3.48046839445867e+01, 3.44456987006561e+01, 3.41040605668682e+01, 3.37810580087804e+01, 3.34776970225359e+01, 3.31947057332929e+01, 3.29325428203776e+01, 3.26914090957555e+01, 3.24712616994671e+01, 3.22718304908322e+01, 3.20926363017435e+01, 3.19330107803966e+01, 3.17921175948362e+01, 3.16689747893189e+01, 3.15624780943749e+01, 3.14714249839424e+01, 3.13945392502798e+01, 3.13304958306884e+01, 3.12779455722577e+01, 3.12355395665404e+01, 3.12019526314999e+01, 3.11759054704234e+01, 3.11561850040793e+01, 3.11416623598449e+01, 3.11313080149826e+01, 3.11242036337204e+01, 3.11195502098197e+01, 3.11166722257623e+01, 3.11150176619742e+01, 3.11141538280313e+01, 3.11137591345893e+01, 3.11136110713593e+01, 3.11135707945619e+01, 3.11135648497906e+01, 3.11135643996705e+01, 3.11135564258373e+01, 3.11135024019496e+01, 3.11133038045059e+01, 3.11127744172693e+01, 3.11116158694631e+01, 3.11093971241287e+01, 3.11055387459244e+01, 3.10993026845475e+01, 3.10897882014966e+01, 3.10759344403261e+01, 3.10565299888830e+01, 3.10302296009944e+01, 3.09955780306967e+01, 3.09510406830915e+01, 3.08950405052747e+01, 3.08260002369796e+01, 3.07423888281339e+01, 3.06427705298835e+01, 3.05258549018110e+01, 3.03905457783617e+01, 3.02359871279917e+01, 3.00616037402826e+01, 2.98671348011795e+01, 2.96526586646650e+01, 2.94186074877353e+01, 2.91657708403348e+01, 2.88952879015123e+01, 2.86086283742904e+01, 2.83075627647953e+01, 2.79941231524860e+01, 2.76705560097281e+01, 2.73392689942028e+01, 2.70027739178705e+01, 2.66636282681024e+01, 2.63243776949221e+01, 2.59875017630858e+01, 2.56553649932046e+01, 2.53301747977706e+01, 2.50139473941142e+01, 2.47084822031645e+01, 2.44153446841826e+01, 2.41358570704878e+01, 2.38710961034536e+01, 2.36218966341120e+01, 2.33888598736887e+01, 2.31723651079599e+01, 2.29725838152855e+01, 2.27894953097992e+01, 2.26229032364817e+01, 2.24724524464742e+01, 2.23376459595918e+01, 2.22178618649006e+01, 2.21123701144368e+01, 2.20203492295798e+01, 2.19409029673843e+01, 2.18730769903370e+01, 2.18158755534520e+01, 2.17682781736988e+01, 2.17292561849726e+01, 2.16977890137871e+01, 2.16728799432162e+01, 2.16535710718704e+01, 2.16389571270041e+01, 2.16281977616135e+01, 2.16205279588143e+01, 2.16152661854456e+01, 2.16118199813376e+01, 2.16096887394277e+01, 2.16084635212617e+01, 2.16078238569103e+01, 2.16075315912567e+01, 2.16074219526873e+01, 2.16073921283728e+01, 2.16073877263677e+01, 2.16073873959078e+01, 2.16073815794394e+01, 2.16073421720308e+01, 2.16071973065399e+01, 2.16068111507257e+01, 2.16059660703331e+01, 2.16043476868853e+01, 2.16015334503140e+01, 2.15969852873147e+01, 2.15900468163136e+01, 2.15799455345992e+01, 2.15658002764609e+01, 2.15466341066974e+01, 2.15213926462570e+01, 2.14889676233259e+01, 2.14482252053773e+01, 2.13980384025907e+01, 2.13373225540928e+01, 2.12650726355814e+01, 2.11804008854777e+01, 2.10825730656546e+01, 2.09710415810684e+01, 2.08454737055053e+01, 2.07057733150578e+01, 2.05520948211707e+01, 2.03848484096488e+01, 2.02046962022578e+01, 2.00125395190831e+01, 1.98094979769099e+01, 1.95968816519034e+01, 1.93761579095527e+01, 1.91489147220017e+01, 1.89168223360392e+01, 1.86815950339167e+01, 1.84449544778310e+01, 1.82085957981843e+01, 1.79741572319000e+01, 1.77431937897251e+01, 1.75171551638025e+01, 1.72973678912707e+01, 1.70850216596480e+01, 1.68811595565765e+01, 1.66866720085460e+01, 1.65022941041951e+01, 1.63286059511084e+01, 1.61660356735584e+01, 1.60148646307772e+01, 1.58752344296833e+01, 1.57471553269833e+01, 1.56305156616155e+01, 1.55250920226071e+01, 1.54305599295026e+01, 1.53465048721921e+01, 1.52724336156421e+01, 1.52077857170836e+01, 1.51519452261144e+01, 1.51042525420629e+01, 1.50640163899177e+01, 1.50305258494991e+01, 1.50030623364617e+01, 1.49809113926790e+01, 1.49633741022519e+01, 1.49497779124036e+01, 1.49394866101761e+01, 1.49319091898715e+01, 1.49265073454272e+01, 1.49228013381142e+01, 1.49203740234403e+01, 1.49188728708152e+01, 1.49180098728872e+01, 1.49175593147678e+01, 1.49173534520859e+01, 1.49172762260556e+01, 1.49172552187189e+01, 1.49172521180806e+01, 1.49172518860225e+01, 1.49172478109665e+01, 1.49172202018728e+01, 1.49171187080847e+01, 1.49168481641422e+01, 1.49162560915994e+01, 1.49151222258291e+01, 1.49131504987750e+01, 1.49099638650938e+01, 1.49051023079603e+01, 1.48980242991751e+01, 1.48881119124141e+01, 1.48746796945501e+01, 1.48569872847262e+01, 1.48342556327300e+01, 1.48056865083802e+01, 1.47704848168449e+01, 1.47278830498670e+01, 1.46771670226752e+01, 1.46177018871935e+01, 1.45489572923767e+01, 1.44705305004938e+01, 1.43821662799242e+01, 1.42837724914466e+01, 1.41754304694586e+01, 1.40573995659830e+01, 1.39301155570374e+01, 1.37941829809917e+01, 1.36503618514435e+01, 1.34995495228979e+01, 1.33427587476257e+01, 1.31810931168341e+01, 1.30157211150573e+01, 1.28478499399507e+01, 1.26787000762587e+01, 1.25094814010673e+01, 1.23413713781800e+01, 1.21754957046700e+01, 1.20129116193291e+01, 1.18545939718831e+01, 1.17014240733443e+01, 1.15541812875707e+01, 1.14135372703704e+01, 1.12800527097506e+01, 1.11541763699000e+01, 1.10362461970323e+01, 1.09264922132950e+01, 1.08250409100448e+01, 1.07319208555162e+01, 1.06470692527181e+01, 1.05703392172355e+01, 1.05015075860102e+01, 1.04402831112937e+01, 1.03863149335486e+01, 1.03392012589547e+01, 1.02984981886294e+01, 1.02637286563990e+01, 1.02343914300396e+01, 1.02099701186098e+01, 1.01899421080391e+01, 1.01737873214792e+01, 1.01609966734786e+01, 1.01510800614256e+01, 1.01435737173949e+01, 1.01380467316469e+01, 1.01341065579296e+01, 1.01314033219189e+01, 1.01296327779631e+01, 1.01285377950158e+01, 1.01279082984241e+01, 1.01275796473363e+01, 1.01274294845938e+01, 1.01273731534707e+01, 1.01273578300494e+01, 1.01273555683449e+01, 1.01273553990110e+01, 1.01273524245759e+01, 1.01273322723481e+01, 1.01272581907338e+01, 1.01270607171719e+01, 1.01266285552864e+01, 1.01258009295735e+01, 1.01243617315855e+01, 1.01220357394887e+01, 1.01184871552355e+01, 1.01133206584693e+01, 1.01060851211633e+01, 1.00962800592086e+01, 1.00833648145761e+01, 1.00667703633296e+01, 1.00459135315527e+01, 1.00202132764937e+01, 9.98910855990245e+00, 9.95207721329128e+00, 9.90865508145879e+00, 9.85845464312128e+00, 9.80118225780058e+00, 9.73665318650826e+00, 9.66480358727308e+00, 9.58569879763770e+00, 9.49953738188948e+00, 9.40665063227333e+00, 9.30749745677652e+00, 9.20265484312102e+00, 9.09280433865065e+00, 8.97871520764748e+00, 8.86122510181174e+00, 8.74121919082184e+00, 8.61960873884088e+00, 8.49731007797004e+00, 8.37522482749947e+00, 8.25422205228385e+00, 8.13512286393407e+00, 8.01868776642941e+00, 7.90560685399258e+00, 7.79649280058064e+00, 7.69187644833610e+00, 7.59220471161440e+00, 7.49784046277118e+00, 7.40906405047076e+00, 7.32607611316102e+00, 7.24900138119220e+00, 7.17789320314182e+00, 7.11273857894841e+00, 7.05346352975941e+00, 6.99993867870937e+00, 6.95198495601316e+00, 6.90937937439544e+00, 6.87186084608130e+00, 6.83913602980647e+00, 6.81088520534865e+00, 6.78676817406928e+00, 6.76643017743488e+00, 6.74950781246918e+00, 6.73563490503859e+00, 6.72444828070307e+00, 6.71559335079071e+00, 6.70872941079318e+00, 6.70353453152060e+00, 6.69970991287175e+00, 6.69698356729806e+00, 6.69511320616489e+00, 6.69388821757722e+00, 6.69313064834618e+00, 6.69269513434445e+00, 6.69246776058083e+00, 6.69236387249661e+00, 6.69232490062979e+00, 6.69231429935030e+00, 6.69231273462414e+00, 6.69231261716224e+00, 6.69231054973449e+00, 6.69229654261501e+00, 6.69224505107879e+00, 6.69210779444451e+00, 6.69180741620502e+00, 6.69123217336727e+00, 6.69023187577652e+00, 6.68861527534747e+00, 6.68614908048357e+00, 6.68255874088028e+00, 6.67753111021548e+00, 6.67071904665092e+00, 6.66174795171176e+00, 6.65022417604698e+00, 6.63574513638969e+00, 6.61791089431766e+00, 6.59633684900662e+00, 6.57066710015070e+00, 6.54058795250000e+00, 6.50584096991905e+00, 6.46623495417522e+00, 6.42165622981527e+00, 6.37207666633034e+00, 6.31755896283677e+00, 6.25825885419550e+00, 6.19442406145718e+00, 6.12638999045913e+00, 6.05457236484186e+00, 5.97945714818074e+00, 5.90158825086252e+00, 5.82155362090626e+00, 5.73997037859053e+00, 5.65746967114295e+00, 5.57468189789532e+00, 5.49222289263405e+00, 5.41068155449722e+00, 5.33060929922533e+00, 5.25251156790669e+00, 5.17684149108655e+00, 5.10399567368792e+00, 5.03431195175951e+00, 4.96806888472800e+00, 4.90548669204895e+00, 4.84672932182139e+00, 4.79190734753966e+00, 4.74108142088913e+00, 4.69426605492848e+00, 4.65143356480390e+00, 4.61251804533110e+00, 4.57741931139005e+00, 4.54600676520533e+00, 4.51812318306038e+00, 4.49358843288113e+00, 4.47220314420971e+00, 4.45375235451132e+00, 4.43800915177117e+00, 4.42473832421042e+00, 4.41370001494128e+00, 4.40465336378148e+00, 4.39736010159876e+00, 4.39158804586734e+00, 4.38711443103553e+00, 4.38372899524568e+00, 4.38123673719791e+00, 4.37946025454157e+00, 4.37824157877452e+00, 4.37744343142444e+00, 4.37694984194758e+00, 4.37666608846833e+00, 4.37651794690299e+00, 4.37645026055696e+00, 4.37642486919451e+00, 4.37641796213985e+00, 4.37641694267349e+00, 4.37641686581759e+00, 4.37641550874747e+00, 4.37640631440441e+00, 4.37637251512453e+00, 4.37628241940551e+00, 4.37608525094223e+00, 4.37570766436200e+00, 4.37505108385824e+00, 4.37398999795467e+00, 4.37237132819678e+00, 4.37001497039601e+00, 4.36671558249218e+00, 4.36224566138954e+00, 4.35635991077909e+00, 4.34880085218751e+00, 4.33930557264840e+00, 4.32761343632937e+00, 4.31347451778898e+00, 4.29665844664197e+00, 4.27696329410739e+00, 4.25422408883628e+00, 4.22832053006169e+00, 4.19918347663348e+00, 4.16679983446171e+00, 4.13121554219569e+00, 4.09253646120422e+00, 4.05092710243296e+00, 4.00660725739759e+00, 3.95984672954735e+00, 3.91095847197683e+00, 3.86029051710045e+00, 3.80821712702830e+00, 3.75512959899194e+00, 3.70142713239500e+00, 3.64750811084018e+00, 3.59376208354239e+00, 3.54056265535941e+00, 3.48826142095956e+00, 3.43718301160590e+00, 3.38762126546568e+00, 3.33983648524801e+00, 3.29405371028500e+00, 3.25046190342289e+00, 3.20921393569196e+00, 3.17042724303835e+00, 3.13418502864607e+00, 3.10053789050971e+00, 3.06950576559539e+00, 3.04108009754682e+00, 3.01522615273181e+00, 2.99188542778411e+00, 2.97097810915679e+00, 2.95240556032724e+00, 2.93605282427599e+00, 2.92179113714108e+00, 2.90948045330363e+00, 2.89897198265248e+00, 2.89011073774050e+00, 2.88273808253692e+00, 2.87669426624784e+00, 2.87182091612286e+00, 2.86796345329708e+00, 2.86497338659271e+00, 2.86271043185431e+00, 2.86104439974375e+00, 2.85985679370455e+00, 2.85904206248985e+00, 2.85850845836295e+00, 2.85817846261307e+00, 2.85798875381631e+00, 2.85788971046547e+00, 2.85784445713722e+00, 2.85782748111565e+00, 2.85782286323223e+00, 2.85782218164255e+00, 2.85782212997900e+00, 2.85782121402930e+00, 2.85781500833909e+00, 2.85779219562214e+00, 2.85773138572975e+00, 2.85759830697663e+00, 2.85734345370838e+00, 2.85690028822334e+00, 2.85618408609058e+00, 2.85509150316452e+00, 2.85350092940263e+00, 2.85127367793984e+00, 2.84825603668631e+00, 2.84428218310158e+00, 2.83917793030112e+00, 2.83276523452022e+00, 2.82486735137044e+00, 2.81531448350822e+00, 2.80394971857361e+00, 2.79063501769335e+00, 2.77525698608558e+00, 2.75773214288733e+00, 2.73801141101393e+00, 2.71608357197871e+00, 2.69197747546077e+00, 2.66576285697015e+00, 2.63754969484779e+00, 2.60748612369433e+00, 2.57575500752252e+00, 2.54256935450287e+00, 2.50816681884551e+00, 2.47280357848096e+00, 2.43674789648657e+00, 2.40027366911949e+00, 2.36365423606480e+00, 2.32715668362108e+00, 2.29103681514304e+00, 2.25553490189371e+00, 2.22087226793745e+00, 2.18724871007218e+00, 2.15484071157441e+00, 2.12380037833843e+00, 2.09425500771556e+00, 2.06630719256137e+00, 2.04003536346956e+00, 2.01549467846468e+00, 1.99271817928327e+00, 1.97171814498289e+00, 1.95248758569001e+00, 1.93500183101128e+00, 1.91922017850148e+00, 1.90508757732020e+00, 1.89253633058973e+00, 1.88148780678158e+00, 1.87185415548103e+00, 1.86354002590647e+00, 1.85644428745556e+00, 1.85046175030669e+00, 1.84548488086952e+00, 1.84140550199204e+00, 1.83811646180862e+00, 1.83551324860254e+00, 1.83349552281128e+00, 1.83196853207644e+00, 1.83084437173953e+00, 1.83004305197922e+00, 1.82949333424303e+00, 1.82913330388070e+00, 1.82891065280445e+00, 1.82878265520773e+00, 1.82871583027969e+00, 1.82868529772701e+00, 1.82867384396238e+00, 1.82867072826539e+00, 1.82867026839521e+00, 1.82867023331649e+00, 1.82866960848366e+00, 1.82866537515224e+00, 1.82864981302238e+00, 1.82860833043812e+00, 1.82851754854383e+00, 1.82834369697294e+00, 1.82804138834008e+00, 1.82755283282084e+00, 1.82680754699332e+00, 1.82572260232270e+00, 1.82420344747057e+00, 1.82214532423491e+00, 1.81943527874428e+00, 1.81595474721712e+00, 1.81158266933231e+00, 1.80619905287283e+00, 1.79968888236833e+00, 1.79194623426818e+00, 1.78287843459276e+00, 1.77241007520149e+00, 1.76048669481919e+00, 1.74707793321923e+00, 1.73217998282760e+00, 1.71581719139321e+00, 1.69804271056604e+00, 1.67893813504681e+00, 1.65861213115373e+00, 1.63719810748890e+00, 1.61485102947594e+00, 1.59174352043129e+00, 1.56806142244019e+00, 1.54399900997018e+00, 1.51975405830468e+00, 1.49552296851040e+00, 1.47149614170750e+00, 1.44785377841481e+00, 1.42476225376362e+00, 1.40237118643428e+00, 1.38081127887203e+00, 1.36019296049032e+00, 1.34060581744936e+00, 1.32211874669687e+00, 1.30478073318463e+00, 1.28862212176645e+00, 1.27365624183876e+00, 1.25988124366950e+00, 1.24728201876789e+00, 1.23583209915088e+00, 1.22549545778190e+00, 1.21622816075087e+00, 1.20797984765279e+00, 1.20069503794642e+00, 1.19431427679071e+00, 1.18877514385308e+00, 1.18401315338517e+00, 1.17996257433954e+00, 1.17655719643820e+00, 1.17373106286476e+00, 1.17141918349154e+00, 1.16955823502349e+00, 1.16808724679599e+00, 1.16694826377821e+00, 1.16608697212789e+00, 1.16545326786283e+00, 1.16500174622324e+00, 1.16469208833989e+00, 1.16448932299281e+00, 1.16436394446352e+00, 1.16429187250196e+00, 1.16425424683385e+00, 1.16423705590125e+00, 1.16423060707727e+00, 1.16422885284912e+00, 1.16422859392904e+00, 1.16422857403936e+00, 1.16422821793097e+00, 1.16422580524869e+00, 1.16421693603904e+00, 1.16419329448682e+00, 1.16414155808721e+00, 1.16404248649866e+00, 1.16387023085801e+00, 1.16359190231485e+00, 1.16316743595331e+00, 1.16254978249212e+00, 1.16168545439311e+00, 1.16051544441245e+00, 1.15897652244350e+00, 1.15700290024832e+00, 1.15452823334169e+00, 1.15148790548565e+00, 1.14782151532316e+00, 1.14347545874233e+00, 1.13840547741206e+00, 1.13257902683515e+00, 1.12597730960142e+00, 1.11859682432235e+00, 1.11045030009527e+00, 1.10156692091326e+00, 1.09199179282224e+00, 1.08178466516685e+00, 1.07101798001472e+00, 1.05977438312784e+00, 1.04814387728987e+00, 1.03622082687400e+00, 1.02410102626393e+00, 1.01187902308573e+00, 9.99645843716479e-01, 9.87487210804403e-01, 9.75482280559660e-01, 9.63702871444653e-01, 9.52213113487531e-01, 9.41069422944040e-01, 9.30320700551050e-01, 9.20008659940024e-01, 9.10168210812927e-01, 9.00827843779211e-01, 8.92009985725878e-01, 8.83731313163352e-01, 8.76003024676393e-01, 8.68831082240760e-01, 8.62216435430256e-01, 8.56155243516355e-01, 8.50639109287755e-01, 8.45655336050328e-01, 8.41187216413118e-01, 8.37214358567792e-01, 8.33713053046040e-01, 8.30656680468891e-01, 8.28016158564779e-01, 8.25760424673618e-01, 8.23856948019511e-01, 8.22272264191810e-01, 8.20972522528834e-01, 8.19924035489834e-01, 8.19093817710453e-01, 8.18450101359986e-01, 8.17962813776788e-01, 8.17604003256776e-01, 8.17348199405259e-01, 8.17172695687375e-01, 8.17057743739781e-01, 8.16986651586504e-01, 8.16945781031563e-01, 8.16924443026792e-01, 8.16914693540852e-01, 8.16911036176066e-01, 8.16910041284558e-01, 8.16909894440670e-01, 8.16909883074345e-01, 8.16909678449051e-01, 8.16908292083141e-01, 8.16903195662931e-01, 8.16889610535569e-01, 8.16859880110059e-01, 8.16802943912208e-01, 8.16703934899536e-01, 8.16543919430558e-01, 8.16299796673533e-01, 8.15944371082409e-01, 8.15446607654309e-01, 8.14772074803258e-01, 8.13883573650457e-01, 8.12741945268357e-01, 8.11307039011311e-01, 8.09538815841975e-01, 8.07398551101429e-01, 8.04850092319768e-01, 8.01861120440833e-01, 7.98404358325627e-01, 7.94458669537946e-01, 7.90009993800425e-01, 7.85052073210494e-01, 7.79586934785257e-01, 7.73625109043167e-01, 7.67185579621355e-01, 7.60295473728632e-01, 7.52989516103113e-01, 7.45309279070071e-01, 7.37302267829861e-01, 7.29020883340880e-01, 7.20521305564816e-01, 7.11862338055994e-01, 7.03104251535649e-01, 6.94307659703554e-01, 6.85532455470053e-01, 6.76836830306982e-01, 6.68276393753430e-01, 6.59903404518559e-01, 6.51766119374543e-01, 6.43908261391203e-01, 6.36368605248535e-01, 6.29180674498002e-01, 6.22372543747360e-01, 6.15966737730446e-01, 6.09980218939157e-01, 6.04424455737089e-01, 5.99305563444275e-01, 5.94624511593750e-01, 5.90377391264610e-01, 5.86555736978227e-01, 5.83146898028843e-01, 5.80134454257845e-01, 5.77498671149627e-01, 5.75216988727647e-01, 5.73264538078547e-01, 5.71614678472457e-01, 5.70239547039236e-01, 5.69110611883815e-01, 5.68199218478398e-01, 5.67477118263810e-01, 5.66916967743467e-01, 5.66492786067869e-01, 5.66180359281830e-01, 5.65957580105959e-01, 5.65804713381254e-01, 5.65704579111241e-01, 5.65642647337711e-01, 5.65607041788567e-01, 5.65588452213504e-01, 5.65579958421566e-01, 5.65576772097003e-01, 5.65575905338520e-01, 5.65575777406750e-01, 5.65575767424036e-01, 5.65575586670121e-01, 5.65574362035676e-01, 5.65569860147662e-01, 5.65557859750086e-01, 5.65531597086917e-01, 5.65481300613243e-01, 5.65393833584494e-01, 5.65252460820233e-01, 5.65036753322985e-01, 5.64722641395960e-01, 5.64282623358807e-01, 5.63686132819768e-01, 5.62900062659082e-01, 5.61889438397046e-01, 5.60618227540459e-01, 5.59050265003748e-01, 5.57150268103939e-01, 5.54884908366120e-01, 5.52223901986082e-01, 5.49141076862810e-01, 5.45615372195599e-01, 5.41631727192211e-01, 5.37181818709030e-01, 5.32264613637381e-01, 5.26886710252967e-01, 5.21062452964927e-01, 5.14813816114241e-01, 5.08170063731114e-01, 5.01167202529132e-01, 4.93847254086162e-01, 4.86257378585648e-01, 4.78448886411854e-01, 4.70476175373790e-01, 4.62395630681300e-01, 4.54264522486795e-01, 4.46139932356388e-01, 4.38077735910340e-01, 4.30131664435166e-01, 4.22352463742371e-01, 4.14787164048263e-01, 4.07478470233286e-01, 4.00464277552288e-01, 3.93777313781229e-01, 3.87444905018242e-01, 3.81488859054031e-01, 3.75925457539310e-01, 3.70765546229848e-01, 3.66014711444521e-01, 3.61673530515446e-01, 3.57737884352701e-01, 3.54199321125790e-01, 3.51045461284443e-01, 3.48260435476092e-01, 3.45825348165761e-01, 3.43718760747649e-01, 3.41917188541814e-01, 3.40395606232121e-01, 3.39127956035284e-01, 3.38087652259389e-01, 3.37248075034282e-01, 3.36583045026188e-01, 3.36067270057555e-01, 3.35676753903243e-01, 3.35389157281199e-01, 3.35184101305970e-01, 3.35043404498279e-01, 3.34951245851883e-01, 3.34894248410450e-01, 3.34861480209440e-01, 3.34844372159168e-01, 3.34836555327712e-01, 3.34833622960787e-01, 3.34832825285588e-01, 3.34832707550376e-01, 3.34832698285126e-01, 3.34832529520853e-01, 3.34831386118267e-01, 3.34827182851626e-01, 3.34815978503173e-01, 3.34791458147720e-01, 3.34744499067583e-01, 3.34662837879364e-01, 3.34530854503384e-01, 3.34329486333175e-01, 3.34036283292932e-01, 3.33625611294551e-01, 3.33069007821038e-01, 3.32335688884910e-01, 3.31393201413377e-01, 3.30208209260050e-01, 3.28747394695213e-01, 3.26978450670381e-01, 3.24871132773509e-01, 3.22398334068566e-01, 3.19537141471699e-01, 3.16269829471395e-01, 3.12584746299507e-01, 3.08477049423573e-01, 3.03949251595768e-01, 2.99011545596741e-01, 2.93681884972870e-01, 2.87985809017542e-01, 2.81956012367317e-01, 2.75631672144499e-01, 2.69057557792294e-01, 2.62282959822514e-01, 2.55360482863679e-01, 2.48344754986334e-01, 2.41291108758858e-01, 2.34254289524614e-01, 2.27287242934267e-01, 2.20440027064839e-01, 2.13758885069893e-01, 2.07285503062806e-01, 2.01056465840892e-01, 1.95102911168348e-01, 1.89450372629381e-01, 1.84118792318174e-01, 1.79122678354773e-01, 1.74471378602689e-01, 1.70169440922820e-01, 1.66217031502669e-01, 1.62610385763472e-01, 1.59342270509939e-01, 1.56402440783438e-01, 1.53778079797689e-01, 1.51454214953546e-01, 1.49414106930664e-01, 1.47639612027277e-01, 1.46111520155927e-01, 1.44809872179096e-01, 1.43714260632436e-01, 1.42804117440335e-01, 1.42058991132970e-01, 1.41458814509718e-01, 1.40984161868147e-01, 1.40616493047858e-01, 1.40338379831022e-01, 1.40133708887774e-01, 1.39987854602879e-01, 1.39887814874058e-01, 1.39822303374785e-01, 1.39781792803547e-01, 1.39758505221238e-01, 1.39746347577130e-01, 1.39740792779861e-01, 1.39738709004789e-01, 1.39738142169449e-01, 1.39738058505813e-01, 1.39738051863588e-01, 1.39737930135993e-01, 1.39737105415108e-01, 1.39734073673097e-01, 1.39725992310161e-01, 1.39708307196528e-01, 1.39674440968264e-01, 1.39615556326689e-01, 1.39520407296885e-01, 1.39375290412218e-01, 1.39164106436155e-01, 1.38868541274931e-01, 1.38468371941940e-01, 1.37941899594253e-01, 1.37266506652608e-01, 1.36419328841669e-01, 1.35378025820856e-01, 1.34121626284255e-01, 1.32631415549824e-01, 1.30891826459205e-01, 1.28891288702363e-01, 1.26622988328776e-01, 1.24085488980697e-01, 1.21283169866430e-01, 1.18226442942770e-01, 1.14931723057197e-01, 1.11421139316008e-01, 1.07721992642951e-01, 1.03865981947109e-01, 9.98882378596306e-02, 9.58262169116892e-02, 9.17185187942522e-02, 8.76036938541063e-02, 8.35191067390895e-02, 7.94999152918336e-02, 7.55782122727879e-02, 7.17823626805227e-02, 6.81365530669495e-02, 6.46605531159423e-02, 6.13696754614210e-02, 5.82749084580172e-02, 5.53831890593557e-02, 5.26977792465715e-02, 5.02187092644704e-02, 4.79432536116633e-02, 4.58664104668287e-02, 4.39813611540671e-02, 4.22798925755359e-02, 4.07527716456713e-02, 3.93900662105722e-02, 3.81814114676331e-02, 3.71162244182557e-02, 3.61838713976397e-02, 3.53737953246593e-02, 3.46756101244665e-02, 3.40791699416059e-02, 3.35746204090555e-02, 3.31524384931186e-02, 3.28034663951797e-02, 3.25189437491389e-02, 3.22905409921788e-02, 3.21103953834552e-02, 3.19711497845816e-02, 3.18659930755842e-02, 3.17887000422843e-02, 3.17336678029250e-02, 3.16959453995764e-02, 3.16712530895700e-02, 3.16559881349576e-02, 3.16472144686628e-02, 3.16426344526946e-02, 3.16405419490735e-02, 3.16397570019404e-02, 3.16395434794714e-02, 3.16395119640873e-02, 3.16395094391169e-02, 3.16394628772230e-02, 3.16391474148929e-02, 3.16379877611297e-02, 3.16348966969855e-02, 3.16281327245349e-02, 3.16151817658721e-02, 3.15926688812738e-02, 3.15563060738526e-02, 3.15008822882419e-02, 3.14203015378638e-02, 3.13076744959204e-02, 3.11554676482899e-02, 3.09557120287849e-02, 3.07002705114883e-02, 3.03811586116609e-02, 2.99909088941490e-02, 2.95229637500540e-02, 2.89720760157905e-02, 2.83346923726526e-02, 2.76092914641737e-02, 2.67966479605803e-02, 2.58999959725799e-02, 2.49250705408883e-02, 2.38800142487605e-02, 2.27751466756700e-02, 2.16226063354270e-02, 2.04358864948944e-02, 1.92292963176044e-02, 1.80173857262531e-02, 1.68143752500924e-02, 1.56336305252580e-02, 1.44872153351798e-02, 1.33855480214464e-02, 1.23371751115400e-02, 1.13486646351504e-02, 1.04246112826303e-02, 9.56773741900463e-03, 8.77906866386979e-03, 8.05816043373397e-03, 7.40335223571265e-03, 6.81202900388200e-03, 6.28087264111343e-03, 5.80609142322430e-03, 5.38361940709819e-03, 5.00928199420514e-03, 4.67892706538464e-03, 4.38852351057051e-03, 4.13423055954354e-03, 3.91244217968278e-03, 3.71981110397712e-03, 3.55325694103215e-03, 3.40996245277836e-03, 3.28736157644016e-03, 3.18312220811379e-03, 3.09512621203112e-03, 3.02144860438602e-03, 2.96033740006862e-03, 2.91019520518887e-03, 2.86956328145038e-03, 2.83710849492991e-03, 2.81161328034329e-03, 2.79196850377161e-03, 2.77716888837176e-03, 2.76631048733122e-03, 2.75858955157560e-03, 2.75330205678870e-03, 2.74984312977722e-03, 2.74770565192739e-03, 2.74647741332501e-03, 2.74583633574821e-03, 2.74554346178491e-03, 2.74543360106350e-03, 2.74540371687237e-03, 2.74539930605044e-03, 2.74539894935799e-03, 2.74539232972870e-03, 2.74534748126598e-03, 2.74518261879564e-03, 2.74474319631679e-03, 2.74378173560308e-03, 2.74194120729021e-03, 2.73874296527765e-03, 2.73358033399518e-03, 2.72571904095287e-03, 2.71430572617248e-03, 2.69838568974023e-03, 2.67693079358367e-03, 2.64887795569280e-03, 2.61317792949001e-03, 2.56885304837340e-03, 2.51506139373703e-03, 2.45116353501373e-03, 2.37678678007569e-03, 2.29188099508804e-03, 2.19675975098710e-03, 2.09212102816991e-03, 1.97904306850205e-03, 1.85895315789253e-03, 1.73356993178686e-03, 1.60482284428554e-03, 1.47475524809693e-03, 1.34541961179480e-03, 1.21877437317153e-03, 1.09659160294484e-03, 9.80383093079240e-04, 8.71349974701729e-04, 7.70357966544566e-04, 6.77937365594516e-04, 5.94304378324750e-04, 5.19398678434868e-04, 4.52931300614523e-04, 3.94437088914401e-04, 3.43326708886593e-04, 2.98934424975528e-04, 2.60559172943431e-04, 2.27497720857173e-04, 1.99069785978168e-04, 1.74635785000987e-04, 1.53608393475224e-04, 1.35459248681968e-04, 1.19721994266797e-04, 1.05992596942883e-04, 9.39276091429520e-05, 8.32408490936132e-05, 7.36988482543666e-05, 6.51153993231901e-05, 5.73455576358128e-05, 5.02794112833271e-05, 4.38358592020904e-05, 3.79565901909018e-05, 3.26004537924973e-05, 2.77384018545490e-05, 2.33491420544851e-05, 1.94155837078824e-05, 1.59220916771119e-05, 1.28525048523898e-05, 1.01888289094632e-05, 7.91048671544559e-06, 5.99400365786096e-06, 4.41301422426932e-06, 3.13849568556179e-06, 2.13915662072850e-06, 1.38192923770930e-06, 8.32534055916767e-07, 4.56101147960908e-07, -5.68434188608080e-14, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14],
[ 8.90000000000000e+01, 8.89999998942264e+01, 8.89999966234972e+01, 8.89999744637876e+01, 8.89998930017761e+01, 8.89996758495477e+01, 8.89992005948074e+01, 8.89982903392964e+01, 8.89967071287835e+01, 8.89941475188678e+01, 8.89902404541770e+01, 8.89845475673186e+01, 8.89765659316438e+01, 8.89657332323055e+01, 8.89514352569839e+01, 8.89330155541460e+01, 8.89097870647322e+01, 8.88810455027359e+01, 8.88460842389889e+01, 8.88042104258615e+01, 8.87547620819485e+01, 8.86971258279012e+01, 8.86307549213177e+01, 8.85551871773624e+01, 8.84700622851745e+01, 8.83751379472376e+01, 8.82703041948406e+01, 8.81555951866797e+01, 8.80311977989602e+01, 8.78974563792709e+01, 8.77548731697582e+01, 8.76041041031209e+01, 8.74459499212954e+01, 8.72813428354610e+01, 8.71113292062216e+01, 8.69370489442737e+01, 8.67597124906010e+01, 8.65805763175393e+01, 8.64009178961666e+01, 8.62220110105845e+01, 8.60451021830050e+01, 8.58713888263014e+01, 8.57019995837899e+01, 8.55379771669624e+01, 8.53802638727146e+01, 8.52296898583945e+01, 8.50869641765224e+01, 8.49526685182292e+01, 8.48272535800045e+01, 8.47110379462485e+01, 8.46042093647900e+01, 8.45068282795236e+01, 8.44188334705543e+01, 8.43400496359949e+01, 8.42701967302362e+01, 8.42089008513731e+01, 8.41557064463623e+01, 8.41100895776465e+01, 8.40714719708840e+01, 8.40392355416511e+01, 8.40127370812133e+01, 8.39913227693025e+01, 8.39743421768465e+01, 8.39611614251146e+01, 8.39511751808335e+01, 8.39438171901545e+01, 8.39385690880798e+01, 8.39349672637101e+01, 8.39326076144631e+01, 8.39311480826966e+01, 8.39303089338971e+01, 8.39298708043314e+01, 8.39296706151429e+01, 8.39295955165921e+01, 8.39295750879018e+01, 8.39295720726675e+01, 8.39295716771324e+01, 8.39295624616385e+01, 8.39295000252978e+01, 8.39292705024961e+01, 8.39286586773760e+01, 8.39273197049455e+01, 8.39247553814413e+01, 8.39202959039671e+01, 8.39130879412791e+01, 8.39020897008519e+01, 8.38860735197510e+01, 8.38636363248850e+01, 8.38332180982123e+01, 8.37931282415694e+01, 8.37415794631898e+01, 8.36767285062776e+01, 8.35967227163160e+01, 8.34997511110995e+01, 8.33840982957005e+01, 8.32481992813633e+01, 8.30906930575956e+01, 8.29104726699169e+01, 8.27067296091697e+01, 8.24789905467204e+01, 8.22271448523312e+01, 8.19514618704073e+01, 8.16525975291146e+01, 8.13315904121905e+01, 8.09898478340892e+01, 8.06291226664655e+01, 8.02514816871923e+01, 7.98592661714067e+01, 7.94550454862115e+01, 7.90415647378935e+01, 7.86216880875476e+01, 7.81983400485221e+01, 7.77744476017460e+01, 7.73528859957416e+01, 7.69364305030880e+01, 7.65277153445822e+01, 7.61291998499296e+01, 7.57431410922156e+01, 7.53715719114131e+01, 7.50162833762259e+01, 7.46788111103937e+01, 7.43604253014253e+01, 7.40621244733309e+01, 7.37846332017679e+01, 7.35284039134759e+01, 7.32936228017560e+01, 7.30802197596964e+01, 7.28878821165755e+01, 7.27160718740304e+01, 7.25640460767604e+01, 7.24308799103282e+01, 7.23154920869371e+01, 7.22166720513512e+01, 7.21331085088257e+01, 7.20634187436501e+01, 7.20061781621001e+01, 7.19599494607450e+01, 7.19233107949715e+01, 7.18948823085695e+01, 7.18733503883967e+01, 7.18574890327604e+01, 7.18461777711341e+01, 7.18384156473153e+01, 7.18333308773257e+01, 7.18301859144562e+01, 7.18283777922750e+01, 7.18274337660973e+01, 7.18270024273852e+01, 7.18268406164765e+01, 7.18267965999042e+01, 7.18267901031474e+01, 7.18267895296074e+01, 7.18267782913741e+01, 7.18267021506767e+01, 7.18264222492196e+01, 7.18256761331707e+01, 7.18240432686814e+01, 7.18209161080824e+01, 7.18154778564163e+01, 7.18066879457672e+01, 7.17932760600951e+01, 7.17737453627889e+01, 7.17463853583097e+01, 7.17092945632728e+01, 7.16604128677806e+01, 7.15975631346513e+01, 7.15185012173316e+01, 7.14209731880748e+01, 7.13027781747223e+01, 7.11618348319331e+01, 7.09962491502215e+01, 7.08043810642927e+01, 7.05849071889071e+01, 7.03368770075753e+01, 7.00597599798203e+01, 6.97534813203469e+01, 6.94184446338400e+01, 6.90555401506468e+01, 6.86661379815948e+01, 6.82520665631589e+01, 6.78155772487627e+01, 6.73592967505569e+01, 6.68861697667202e+01, 6.63993945607405e+01, 6.59023544320155e+01, 6.53985479150958e+01, 6.48915202049286e+01, 6.43847978110774e+01, 6.38818279021638e+01, 6.33859233117161e+01, 6.29002138014320e+01, 6.24276039317407e+01, 6.19707377433791e+01, 6.15319703559519e+01, 6.11133464918319e+01, 6.07165858105351e+01, 6.03430747923067e+01, 5.99938647630824e+01, 5.96696755349447e+01, 5.93709040673272e+01, 5.90976375399856e+01, 5.88496702600682e+01, 5.86265238845230e+01, 5.84274705056734e+01, 5.82515582054227e+01, 5.80976387212808e+01, 5.79643968802311e+01, 5.78503814441841e+01, 5.77540369766190e+01, 5.76737362894499e+01, 5.76078129689619e+01, 5.75545934174623e+01, 5.75124277910135e+01, 5.74797191711002e+01, 5.74549502865411e+01, 5.74367071074255e+01, 5.74236986696213e+01, 5.74147725585169e+01, 5.74089255836383e+01, 5.74053093084748e+01, 5.74032302566617e+01, 5.74021447889621e+01, 5.74016488263142e+01, 5.74014627730377e+01, 5.74014121619821e+01, 5.74014046918943e+01, 5.74014040927009e+01, 5.74013930345428e+01, 5.74013181138905e+01, 5.74010426975901e+01, 5.74003085384110e+01, 5.73987018467645e+01, 5.73956248314657e+01, 5.73902738491674e+01, 5.73816251669783e+01, 5.73684291858025e+01, 5.73492137890231e+01, 5.73222972665585e+01, 5.72858110120213e+01, 5.72377318961501e+01, 5.71759238816863e+01, 5.70981880675653e+01, 5.70023199446598e+01, 5.68861722301216e+01, 5.67477212490666e+01, 5.65851344838754e+01, 5.63968366495032e+01, 5.61815715147428e+01, 5.59384567062904e+01, 5.56670289262873e+01, 5.53672773903283e+01, 5.50396638366872e+01, 5.46851281306403e+01, 5.43050792308737e+01, 5.39013720248037e+01, 5.34762712025971e+01, 5.30324038683944e+01, 5.25727029546796e+01, 5.21003437194069e+01, 5.16186756981852e+01, 5.11311524939596e+01, 5.06412617380218e+01, 5.01524574479516e+01, 4.96680968203053e+01, 4.91913832068736e+01, 4.87253166287960e+01, 4.82726527062955e+01, 4.78358703721484e+01, 4.74171482551765e+01, 4.70183492218399e+01, 4.66410122861946e+01, 4.62863509524586e+01, 4.59552570280152e+01, 4.56483090095365e+01, 4.53657842659964e+01, 4.51076743862245e+01, 4.48737031988108e+01, 4.46633470910401e+01, 4.44758573419084e+01, 4.43102842393894e+01, 4.41655027753047e+01, 4.40402397059829e+01, 4.39331017379561e+01, 4.38426045503130e+01, 4.37672023044642e+01, 4.37053172239181e+01, 4.36553687577087e+01, 4.36158017783158e+01, 4.35851132154027e+01, 4.35618764972926e+01, 4.35447631687378e+01, 4.35325610805888e+01, 4.35241886067526e+01, 4.35187044361557e+01, 4.35153126094814e+01, 4.35133626169182e+01, 4.35123445366403e+01, 4.35118793654287e+01, 4.35117048633365e+01, 4.35116573945070e+01, 4.35116503882061e+01, 4.35116498478128e+01, 4.35116401440830e+01, 4.35115743998867e+01, 4.35113327172797e+01, 4.35106884802824e+01, 4.35092785838745e+01, 4.35065784655100e+01, 4.35018829430229e+01, 4.34942937463887e+01, 4.34827144953321e+01, 4.34658537161331e+01, 4.34422363035425e+01, 4.34102236115164e+01, 4.33680420949126e+01, 4.33138201219701e+01, 4.32456322379429e+01, 4.31615497939270e+01, 4.30596964794136e+01, 4.29383069375891e+01, 4.27957863301439e+01, 4.26307684880267e+01, 4.24421701701710e+01, 4.22292389819375e+01, 4.19915926957736e+01, 4.17292480690525e+01, 4.14426377493493e+01, 4.11326144568959e+01, 4.08004422828392e+01, 4.04477755773284e+01, 4.00766264638092e+01, 3.96893224614637e+01, 3.92884560083409e+01, 3.88768278634963e+01, 3.84573864587385e+01, 3.80331653056075e+01, 3.76072205623623e+01, 3.71825708216887e+01, 3.67621410577915e+01, 3.63487124284645e+01, 3.59448792416376e+01, 3.55530138891580e+01, 3.51752399922579e+01, 3.48134134868247e+01, 3.44691109849542e+01, 3.41436245251951e+01, 3.38379617616900e+01, 3.35528507000730e+01, 3.32887482101684e+01, 3.30458516841569e+01, 3.28241133337377e+01, 3.26232567188478e+01, 3.24427951737880e+01, 3.22820518493271e+01, 3.21401811264757e+01, 3.20161911816528e+01, 3.19089674939041e+01, 3.18172970814081e+01, 3.17398932357374e+01, 3.16754204886762e+01, 3.16225195003801e+01, 3.15798315039952e+01, 3.15460218869599e+01, 3.15198024405797e+01, 3.14999517747299e+01, 3.14853333806282e+01, 3.14749108369076e+01, 3.14677596958979e+01, 3.14630756586979e+01, 3.14601787471866e+01, 3.14585133039529e+01, 3.14576437905279e+01, 3.14572465022137e+01, 3.14570974655785e+01, 3.14570569239932e+01, 3.14570509401400e+01, 3.14570504870607e+01, 3.14570424608060e+01, 3.14569880817561e+01, 3.14567881787128e+01, 3.14562553113348e+01, 3.14550891478723e+01, 3.14528558197461e+01, 3.14489720885226e+01, 3.14426950676074e+01, 3.14331181323178e+01, 3.14191735240981e+01, 3.13996420023456e+01, 3.13731697152911e+01, 3.13382922454833e+01, 3.12934655343610e+01, 3.12371031071498e+01, 3.11676187123948e+01, 3.10834731744884e+01, 3.09832239530850e+01, 3.08655756354715e+01, 3.07294293840974e+01, 3.05739292477216e+01, 3.03985032420429e+01, 3.02028972264449e+01, 2.99871998482306e+01, 2.97518571827675e+01, 2.94976761448812e+01, 2.92258162542934e+01, 2.89377698744861e+01, 2.86353315808446e+01, 2.83205578250632e+01, 2.79957185263135e+01, 2.76632426129871e+01, 2.73256598361522e+01, 2.69855413482831e+01, 2.66454415607919e+01, 2.63078436433764e+01, 2.59751107074193e+01, 2.56494442486111e+01, 2.53328508577909e+01, 2.50271176069104e+01, 2.47337959465535e+01, 2.44541934716809e+01, 2.41893725638572e+01, 2.39401547183041e+01, 2.37071293069904e+01, 2.34906655911136e+01, 2.32909269439368e+01, 2.31078864414574e+01, 2.29413431906332e+01, 2.27909389667813e+01, 2.26561749056776e+01, 2.25364281321828e+01, 2.24309683026969e+01, 2.23389740947110e+01, 2.22595496972177e+01, 2.21917413462205e+01, 2.21345539160651e+01, 2.20869675260084e+01, 2.20479540585796e+01, 2.20164934181686e+01, 2.19915892912123e+01, 2.19722841095967e+01, 2.19576728723610e+01, 2.19469154527185e+01, 2.19392470118793e+01, 2.19339861606769e+01, 2.19305405551289e+01, 2.19284096812331e+01, 2.19271846738853e+01, 2.19265451193812e+01, 2.19262529038676e+01, 2.19261432840995e+01, 2.19261134648986e+01, 2.19261090636483e+01, 2.19261087332450e+01, 2.19261029177738e+01, 2.19260635171212e+01, 2.19259186764615e+01, 2.19255325867999e+01, 2.19246876509889e+01, 2.19230695436955e+01, 2.19202557850496e+01, 2.19157083882826e+01, 2.19087710714718e+01, 2.18986714379546e+01, 2.18845284230453e+01, 2.18653651699836e+01, 2.18401273302247e+01, 2.18077065795274e+01, 2.17669689033646e+01, 2.17167869401033e+01, 2.16560753918369e+01, 2.15838282408433e+01, 2.14991562701018e+01, 2.14013232082045e+01, 2.12897787314768e+01, 2.11641865841835e+01, 2.10244462375600e+01, 2.08707068032471e+01, 2.07033723334159e+01, 2.05230981481320e+01, 2.03307783844480e+01, 2.01275255050554e+01, 1.99146429781622e+01, 1.96935926928238e+01, 1.94659588700845e+01, 1.92334102585063e+01, 1.89976622775865e+01, 1.87604405315164e+01, 1.85234468099803e+01, 1.82883283753520e+01, 1.80566510497187e+01, 1.78298763851329e+01, 1.76093430301071e+01, 1.73962522819172e+01, 1.71916577174765e+01, 1.69964587074749e+01, 1.68113975306989e+01, 1.66370597209811e+01, 1.64738772093061e+01, 1.63221337814092e+01, 1.61819723656702e+01, 1.60534036983153e+01, 1.59363159759849e+01, 1.58304851873823e+01, 1.57355859021127e+01, 1.56512023737507e+01, 1.55768398769953e+01, 1.55119362409136e+01, 1.54558735607089e+01, 1.54079900707187e+01, 1.53675921445898e+01, 1.53339663587922e+01, 1.53063915172101e+01, 1.52841504920378e+01, 1.52665416941631e+01, 1.52528899490712e+01, 1.52425565261634e+01, 1.52349480538435e+01, 1.52295240525055e+01, 1.52258028343073e+01, 1.52233655526283e+01, 1.52218582342553e+01, 1.52209916910973e+01, 1.52205392818924e+01, 1.52203325733985e+01, 1.52202550300692e+01, 1.52202339364189e+01, 1.52202308230409e+01, 1.52202305900293e+01, 1.52202264982299e+01, 1.52201987756973e+01, 1.52200968648923e+01, 1.52198252093114e+01, 1.52192307038495e+01, 1.52180921782671e+01, 1.52161123462520e+01, 1.52129126087367e+01, 1.52080310485832e+01, 1.52009238917011e+01, 1.51909706333854e+01, 1.51774829342244e+01, 1.51597172740766e+01, 1.51368912138119e+01, 1.51082029539487e+01, 1.50728537017846e+01, 1.50300721730478e+01, 1.49791403735505e+01, 1.49194196473827e+01, 1.48503758595346e+01, 1.47716025212307e+01, 1.46828406819289e+01, 1.45839945137074e+01, 1.44751417043983e+01, 1.43565380478012e+01, 1.42286159538319e+01, 1.40919769689890e+01, 1.39473787605308e+01, 1.37957173358684e+01, 1.36380055057139e+01, 1.34753487311834e+01, 1.33089195148661e+01, 1.31399314170059e+01, 1.29696136294917e+01, 1.27991868587908e+01, 1.26298410878237e+01, 1.24627156283659e+01, 1.22988817477174e+01, 1.21393280517610e+01, 1.19849487206471e+01, 1.18365346130569e+01, 1.16947671750548e+01, 1.15602150108558e+01, 1.14333329009873e+01, 1.13144629956614e+01, 1.12038378739033e+01, 1.11015851452420e+01, 1.10077332800233e+01, 1.09222183828663e+01, 1.08448916655751e+01, 1.07755274241804e+01, 1.07138313732553e+01, 1.06594492337791e+01, 1.06119755046331e+01, 1.05709623698981e+01, 1.05359287035918e+01, 1.05063691307723e+01, 1.04817630905946e+01, 1.04615838253672e+01, 1.04453071930052e+01, 1.04324201719901e+01, 1.04224289016672e+01, 1.04148660799294e+01, 1.04092975281375e+01, 1.04053277318914e+01, 1.04026041774493e+01, 1.04008203275609e+01, 1.03997171164541e+01, 1.03990828898452e+01, 1.03987517693659e+01, 1.03986004783617e+01, 1.03985437239925e+01, 1.03985282854384e+01, 1.03985260067405e+01, 1.03985258361343e+01, 1.03985228393508e+01, 1.03985025357090e+01, 1.03984278974838e+01, 1.03982289402215e+01, 1.03977935313962e+01, 1.03969596877973e+01, 1.03955096780795e+01, 1.03931662146428e+01, 1.03895909811501e+01, 1.03843856965813e+01, 1.03770958612031e+01, 1.03672172614888e+01, 1.03542052276877e+01, 1.03374865384032e+01, 1.03164737521174e+01, 1.02905816193949e+01, 1.02592450975713e+01, 1.02219383608359e+01, 1.01781940836851e+01, 1.01276221870282e+01, 1.00699271859713e+01, 1.00049232770924e+01, 9.93254635820861e+00, 9.85286228788959e+00, 9.76607086229082e+00, 9.67250520423373e+00, 9.57262650915251e+00, 9.46701435532194e+00, 9.35635303968307e+00, 9.24141462314822e+00, 9.12303954024846e+00, 9.00211573197492e+00, 8.87955728933827e+00, 8.75628354870702e+00, 8.63319946719500e+00, 8.51117794341963e+00, 8.39104455675599e+00, 8.27356499904526e+00, 8.15943528674477e+00, 8.04927468405505e+00, 7.94362114757560e+00, 7.84292902305329e+00, 7.74756868200235e+00, 7.65782777414984e+00, 7.57391378304906e+00, 7.49595759920224e+00, 7.42401786118074e+00, 7.35808585559333e+00, 7.29809080781862e+00, 7.24390543474257e+00, 7.19535166655302e+00, 7.15220647553718e+00, 7.11420777463621e+00, 7.08106036628975e+00, 7.05244193221407e+00, 7.02800905699095e+00, 7.00740327299867e+00, 6.99025710216530e+00, 6.97620005268860e+00, 6.96486450817096e+00, 6.95589142483411e+00, 6.94893573209011e+00, 6.94367131521057e+00, 6.93979544837993e+00, 6.93703254381106e+00, 6.93513708896052e+00, 6.93389565954563e+00, 6.93312792054134e+00, 6.93268655931746e+00, 6.93245613259936e+00, 6.93235084955450e+00, 6.93231135438319e+00, 6.93230061075158e+00, 6.93229902501452e+00, 6.93229890597539e+00, 6.93229681078648e+00, 6.93228261558122e+00, 6.93223043261807e+00, 6.93209133286619e+00, 6.93178692087081e+00, 6.93120395237874e+00, 6.93019021815530e+00, 6.92855189611772e+00, 6.92605254875708e+00, 6.92241391218874e+00, 6.91731858491464e+00, 6.91041467625745e+00, 6.90132241456789e+00, 6.88964264281120e+00, 6.87496704461081e+00, 6.85688984984134e+00, 6.83502067022738e+00, 6.80899801913004e+00, 6.77850298456985e+00, 6.74327246030114e+00, 6.70311130593606e+00, 6.65790281161743e+00, 6.60761689039953e+00, 6.55231551298697e+00, 6.49215503069956e+00, 6.42738519476062e+00, 6.35834486085182e+00, 6.28545455281855e+00, 6.20920623364702e+00, 6.13015078200073e+00, 6.04888378798903e+00, 5.96603035495976e+00, 5.88222962074047e+00, 5.79811969077592e+00, 5.71432360899962e+00, 5.63143688540278e+00, 5.55001696127403e+00, 5.47057483685339e+00, 5.39356892727605e+00, 5.31940106740682e+00, 5.24841446856790e+00, 5.18089334957494e+00, 5.11706392400782e+00, 5.05709642214387e+00, 5.00110785169838e+00, 4.94916524633818e+00, 4.90128920474450e+00, 4.85745757741074e+00, 4.81760920752392e+00, 4.78164767292904e+00, 4.74944500712361e+00, 4.72084539866862e+00, 4.69566888126306e+00, 4.67371503220201e+00, 4.65476669620918e+00, 4.63859374579536e+00, 4.62495687934231e+00, 4.61361144504005e+00, 4.60431126366323e+00, 4.59681240711345e+00, 4.59087687395628e+00, 4.58627608919790e+00, 4.58279414462649e+00, 4.58023068943100e+00, 4.57840337949827e+00, 4.57714979841639e+00, 4.57632877395004e+00, 4.57582103025993e+00, 4.57552913756825e+00, 4.57537674603614e+00, 4.57530711771628e+00, 4.57528099783156e+00, 4.57527389259923e+00, 4.57527284388216e+00, 4.57527276482111e+00, 4.57527136881370e+00, 4.57526191066391e+00, 4.57522714159046e+00, 4.57513446066422e+00, 4.57493163406340e+00, 4.57454320965347e+00, 4.57386777637849e+00, 4.57277620409138e+00, 4.57111098301675e+00, 4.56868676182835e+00, 4.56529215776165e+00, 4.56069287959409e+00, 4.55463616330675e+00, 4.54685647007410e+00, 4.53708233731730e+00, 4.52504420780004e+00, 4.51048299275652e+00, 4.49315905815490e+00, 4.47286126514454e+00, 4.44941565398689e+00, 4.42269334256170e+00, 4.39261722166945e+00, 4.35916707295246e+00, 4.32238281080160e+00, 4.28236565240172e+00, 4.23927714142346e+00, 4.19333607916133e+00, 4.14481353924613e+00, 4.09402624634333e+00, 4.04132867620571e+00, 3.98710427899348e+00, 3.93175623947344e+00, 3.87569817019904e+00, 3.81934509342983e+00, 3.76310501181668e+00, 3.70737130383843e+00, 3.65251611343149e+00, 3.59888483848799e+00, 3.54679176288295e+00, 3.49651682350866e+00, 3.44830345898066e+00, 3.40235745138516e+00, 3.35884664738190e+00, 3.31790143032326e+00, 3.27961581029457e+00, 3.24404900295399e+00, 3.21122737903362e+00, 3.18114668229863e+00, 3.15377443249592e+00, 3.12905244932574e+00, 3.10689945199228e+00, 3.08721370505828e+00, 3.06987569417603e+00, 3.05475082421007e+00, 3.04169213707907e+00, 3.03054304740262e+00, 3.02114009109692e+00, 3.01331567600076e+00, 3.00690081521655e+00, 3.00172781406569e+00, 2.99763287144420e+00, 2.99445854703383e+00, 2.99205603836199e+00, 2.99028720707899e+00, 2.98902629280119e+00, 2.98816125592906e+00, 2.98759469813007e+00, 2.98724432043699e+00, 2.98704289356290e+00, 2.98693773217930e+00, 2.98688968343932e+00, 2.98687165875714e+00, 2.98686675561210e+00, 2.98686603191840e+00, 2.98686597706339e+00, 2.98686500453246e+00, 2.98685841549593e+00, 2.98683419355507e+00, 2.98676962717204e+00, 2.98662832734401e+00, 2.98635772940069e+00, 2.98588718275172e+00, 2.98512672258457e+00, 2.98396660590143e+00, 2.98227768010490e+00, 2.97991263457999e+00, 2.97670816329466e+00, 2.97248803846227e+00, 2.96706706132611e+00, 2.96025581641575e+00, 2.95186611139329e+00, 2.94171693814036e+00, 2.92964074537541e+00, 2.91548977304994e+00, 2.89914216875628e+00, 2.88050759101480e+00, 2.85953200748015e+00, 2.83620142024499e+00, 2.81054429593573e+00, 2.78263254320237e+00, 2.75258096010839e+00, 2.72054516236547e+00, 2.68671809251879e+00, 2.65132529185701e+00, 2.61461918347844e+00, 2.57687266076695e+00, 2.53837229717384e+00, 2.49941149018858e+00, 2.46028382705397e+00, 2.42127691679339e+00, 2.38266687859947e+00, 2.34471361717740e+00, 2.30765695731121e+00, 2.27171365752494e+00, 2.23707527934482e+00, 2.20390685570214e+00, 2.17234627936161e+00, 2.14250431882775e+00, 2.11446516338229e+00, 2.08828739906809e+00, 2.06400532207147e+00, 2.04163050386384e+00, 2.02115353270054e+00, 2.00254586785974e+00, 1.98576175561968e+00, 1.97074016868667e+00, 1.95740674285271e+00, 1.94567569532931e+00, 1.93545171780553e+00, 1.92663184328322e+00, 1.91910728882776e+00, 1.91276527647319e+00, 1.90749083182776e+00, 1.90316855489316e+00, 1.89968435090654e+00, 1.89692710146910e+00, 1.89479024875912e+00, 1.89317325916301e+00, 1.89198292804039e+00, 1.89113448524431e+00, 1.89055246189321e+00, 1.89017128290237e+00, 1.88993555679694e+00, 1.88980004393601e+00, 1.88972929582098e+00, 1.88969697083644e+00, 1.88968484468566e+00, 1.88968154608496e+00, 1.88968105921862e+00, 1.88968102208065e+00, 1.88968036056775e+00, 1.88967587872420e+00, 1.88965940304908e+00, 1.88961548537742e+00, 1.88951937480493e+00, 1.88933531990624e+00, 1.88901527301158e+00, 1.88849806228150e+00, 1.88770909013498e+00, 1.88656060759214e+00, 1.88495260245709e+00, 1.88277432411093e+00, 1.87990644827623e+00, 1.87622386106503e+00, 1.87159901309032e+00, 1.86590576227377e+00, 1.85902358988558e+00, 1.85084204075623e+00, 1.84126520858739e+00, 1.83021606428323e+00, 1.81764041257927e+00, 1.80351026274811e+00, 1.78782641456483e+00, 1.77062009133847e+00, 1.75195349638106e+00, 1.73191922495801e+00, 1.71063852648517e+00, 1.68825847672080e+00, 1.66494818202412e+00, 1.64089419287212e+00, 1.61629534794132e+00, 1.59135730024654e+00, 1.56628699104002e+00, 1.54128733419117e+00, 1.51655235327168e+00, 1.49226297636875e+00, 1.46858364202912e+00, 1.44565980772660e+00, 1.42361638555573e+00, 1.40255706529915e+00, 1.38256442948529e+00, 1.36370072419984e+00, 1.34600912657757e+00, 1.32951534550426e+00, 1.31422940376608e+00, 1.30014747338203e+00, 1.28725366595582e+00, 1.27552171162873e+00, 1.26491648965934e+00, 1.25539539824298e+00, 1.24690956972343e+00, 1.23940494974926e+00, 1.23282326581175e+00, 1.22710291295699e+00, 1.22217978335445e+00, 1.21798806281164e+00, 1.21446101206472e+00, 1.21153174439966e+00, 1.20913400438562e+00, 1.20720294565727e+00, 1.20567589915281e+00, 1.20449311736373e+00, 1.20359847533916e+00, 1.20294010574504e+00, 1.20247094349043e+00, 1.20214915549411e+00, 1.20193843314699e+00, 1.20180812885003e+00, 1.20173322344156e+00, 1.20169411798875e+00, 1.20167625082294e+00, 1.20166954830316e+00, 1.20166772506197e+00, 1.20166745595563e+00, 1.20166743528345e+00, 1.20166706516528e+00, 1.20166455756393e+00, 1.20165533941065e+00, 1.20163076761753e+00, 1.20157699497565e+00, 1.20147402208062e+00, 1.20129497684091e+00, 1.20100566054070e+00, 1.20056439621504e+00, 1.19992221068739e+00, 1.19902337529058e+00, 1.19780632126076e+00, 1.19620493337452e+00, 1.19415020918053e+00, 1.19157225120952e+00, 1.18840253651605e+00, 1.18457638324593e+00, 1.18003550983001e+00, 1.17473056172180e+00, 1.16862346652373e+00, 1.16168947402190e+00, 1.15391874556936e+00, 1.14531737870675e+00, 1.13590778742525e+00, 1.12572840355713e+00, 1.11483271592796e+00, 1.10328771512005e+00, 1.09117185637841e+00, 1.07857268534151e+00, 1.06558428674157e+00, 1.05230471359225e+00, 1.03883353535204e+00, 1.02526961247877e+00, 1.01170916764175e+00, 9.98244186816066e-01, 9.84961151633513e-01, 9.71940080841804e-01, 9.59253844478468e-01, 9.46967708513881e-01, 9.35139068232402e-01, 9.23817333098683e-01, 9.13043932129455e-01, 9.02852415265386e-01, 8.93268631987098e-01, 8.84310973076467e-01, 8.75990665003044e-01, 8.68312109114100e-01, 8.61273259866635e-01, 8.54866037953784e-01, 8.49076775459963e-01, 8.43886691162126e-01, 8.39272394761920e-01, 8.35206419134039e-01, 8.31657779564026e-01, 8.28592558387911e-01, 8.25974512435778e-01, 8.23765699254736e-01, 8.21927116317326e-01, 8.20419345424796e-01, 8.19203192434813e-01, 8.18240310460524e-01, 8.17493792990859e-01, 8.16928722161990e-01, 8.16512656844168e-01, 8.16216045433421e-01, 8.16012549338410e-01, 8.15879265149832e-01, 8.15796836312984e-01, 8.15749448665983e-01, 8.15724708260746e-01, 8.15713404214833e-01, 8.15709163684971e-01, 8.15708010158801e-01, 8.15707839900767e-01, 8.15707826722075e-01, 8.15707589469483e-01, 8.15705982048911e-01, 8.15700073011897e-01, 8.15684321773332e-01, 8.15649851002277e-01, 8.15583836996638e-01, 8.15469043231843e-01, 8.15283519794093e-01, 8.15000488479697e-01, 8.14588429782219e-01, 8.14011383554742e-01, 8.13229469568583e-01, 8.12199627278744e-01, 8.10876565744479e-01, 8.09213904929763e-01, 8.07165478846727e-01, 8.04686759823085e-01, 8.01736352460637e-01, 7.98277496713254e-01, 7.94279513151068e-01, 7.89719121014201e-01, 7.84581561936065e-01, 7.78861469609339e-01, 7.72563437956322e-01, 7.65702256703293e-01, 7.58302802253269e-01, 7.50399591650023e-01, 7.42036026376439e-01, 7.33263369025025e-01, 7.24139508208324e-01, 7.14727574629649e-01, 7.05094473784499e-01, 6.95309398594176e-01, 6.85442379112843e-01, 6.75562917311112e-01, 6.65738743998048e-01, 6.56034723373537e-01, 6.46511919572291e-01, 6.37226829713242e-01, 6.28230779970608e-01, 6.19569475305795e-01, 6.11282689733358e-01, 6.03404082117379e-01, 5.95961122143819e-01, 5.88975111868493e-01, 5.82461289694095e-01, 5.76429005424657e-01, 5.70881956915116e-01, 5.65818480581910e-01, 5.61231889545098e-01, 5.57110854362492e-01, 5.53439822153365e-01, 5.50199470373613e-01, 5.47367191593835e-01, 5.44917605351728e-01, 5.42823092524415e-01, 5.41054346736487e-01, 5.39580936149235e-01, 5.38371867659734e-01, 5.37396144185507e-01, 5.36623304456981e-01, 5.36023933726383e-01, 5.35570133167809e-01, 5.35235935612889e-01, 5.34997655731217e-01, 5.34834163881243e-01, 5.34727074619553e-01, 5.34660843221843e-01, 5.34622766425898e-01, 5.34602886814270e-01, 5.34593803634380e-01, 5.34590396216586e-01, 5.34589469316032e-01, 5.34589332507437e-01, 5.34589321832030e-01, 5.34589128536112e-01, 5.34587818928131e-01, 5.34583004673308e-01, 5.34570171657748e-01, 5.34542087019631e-01, 5.34488301970981e-01, 5.34394770549525e-01, 5.34243602204498e-01, 5.34012963505106e-01, 5.33677141172504e-01, 5.33206774935877e-01, 5.32569264257674e-01, 5.31729347651833e-01, 5.30649847085513e-01, 5.29292562885189e-01, 5.27619296884339e-01, 5.25592973648330e-01, 5.23178822051776e-01, 5.20345572967216e-01, 5.17066624105482e-01, 5.13321120859956e-01, 5.09094902920680e-01, 5.04381270747487e-01, 4.99181533672640e-01, 4.93505311991839e-01, 4.87370578079265e-01, 4.80803435225269e-01, 4.73837646332385e-01, 4.66513936648894e-01, 4.58879104440044e-01, 4.50984980308675e-01, 4.42887279595340e-01, 4.34644393092441e-01, 4.26316159651122e-01, 4.17962660729714e-01, 4.09643072131956e-01, 4.01414602621855e-01, 3.93331543160301e-01, 3.85444444447032e-01, 3.77799434435005e-01, 3.70437681657350e-01, 3.63395004712217e-01, 3.56701623267114e-01, 3.50382041664844e-01, 3.44455052825097e-01, 3.38933847782400e-01, 3.33826214947408e-01, 3.29134813002511e-01, 3.24857502118320e-01, 3.20987719709848e-01, 3.17514888979659e-01, 3.14424850748608e-01, 3.11700311276525e-01, 3.09321300698542e-01, 3.07265638161297e-01, 3.05509400630740e-01, 3.04027392616419e-01, 3.02793613747966e-01, 3.01781720342191e-01, 3.00965475958265e-01, 3.00319184633025e-01, 2.99818099217632e-01, 2.99438796193087e-01, 2.99159507701532e-01, 2.98960401425902e-01, 2.98823799468821e-01, 2.98734328544782e-01, 2.98678995572615e-01, 2.98647185044146e-01, 2.98630577207987e-01, 2.98622988979389e-01, 2.98620142377544e-01, 2.98619368033229e-01, 2.98619253741677e-01, 2.98619244747428e-01, 2.98619080919352e-01, 2.98617970960521e-01, 2.98613890642443e-01, 2.98603014073691e-01, 2.98579211279389e-01, 2.98533627270388e-01, 2.98454360020216e-01, 2.98326253074364e-01, 2.98130816523951e-01, 2.97846287670268e-01, 2.97447839733479e-01, 2.96907943263903e-01, 2.96196880367660e-01, 2.95283406402447e-01, 2.94135547481247e-01, 2.92721515120874e-01, 2.91010712015890e-01, 2.88974795681483e-01, 2.86588760174880e-01, 2.83831990933393e-01, 2.80689244607359e-01, 2.77151505187931e-01, 2.73216670139945e-01, 2.68890025816788e-01, 2.64184480057210e-01, 2.59120531145882e-01, 2.53725965603778e-01, 2.48035291681603e-01, 2.42088929942170e-01, 2.35932195862461e-01, 2.29614120910966e-01, 2.23186167155191e-01, 2.16700895404571e-01, 2.10210647789467e-01, 2.03766302441537e-01, 1.97416150885076e-01, 1.91204938550499e-01, 1.85173096449006e-01, 1.79356178670119e-01, 1.73784507186710e-01, 1.68483013586425e-01, 1.63471257667780e-01, 1.58763595907175e-01, 1.54369468825379e-01, 1.50293775149763e-01, 1.46537301997938e-01, 1.43097183555767e-01, 1.39967365257078e-01, 1.37139055680052e-01, 1.34601153702604e-01, 1.32340643471380e-01, 1.30342954112734e-01, 1.28592284652925e-01, 1.27071897217959e-01, 1.25764383237225e-01, 1.24651908125969e-01, 1.23716439862193e-01, 1.22939966129280e-01, 1.22304703417214e-01, 1.21793299827559e-01, 1.21389031491432e-01, 1.21075990665148e-01, 1.20839261894346e-01, 1.20665081292245e-01, 1.20540973094236e-01, 1.20455857318589e-01, 1.20400122620225e-01, 1.20365659259193e-01, 1.20345848445240e-01, 1.20335506052172e-01, 1.20330780672873e-01, 1.20329008043498e-01, 1.20328525847540e-01, 1.20328454676510e-01, 1.20328449026090e-01, 1.20328345474817e-01, 1.20327643901120e-01, 1.20325064862385e-01, 1.20318190249321e-01, 1.20303146124812e-01, 1.20274337884709e-01, 1.20224249683432e-01, 1.20143319529333e-01, 1.20019900760013e-01, 1.19840319571267e-01, 1.19589036658951e-01, 1.19248918593314e-01, 1.18801621060996e-01, 1.18228081445992e-01, 1.17509112375018e-01, 1.16626081003730e-01, 1.15561651345544e-01, 1.14300559437780e-01, 1.12830384365651e-01, 1.11142273001974e-01, 1.09231573661916e-01, 1.07098334493728e-01, 1.04747626855584e-01, 1.02189662314956e-01, 9.94396839356568e-02, 9.65176273665378e-02, 9.34475636371701e-02, 9.02569518797378e-02, 8.69757446742199e-02, 8.36353996978119e-02, 8.02678575572173e-02, 7.69045463687377e-02, 7.35754688218293e-02, 7.03084178174436e-02, 6.71283536347005e-02, 6.40569606418637e-02, 6.11123866301142e-02, 5.83091545236272e-02, 5.56582257120084e-02, 5.31671872248296e-02, 5.08405315400182e-02, 4.86799976365830e-02, 4.66849443010346e-02, 4.48527308577127e-02, 4.31790856032848e-02, 4.16584475820656e-02, 4.02842724082291e-02, 3.90492972823040e-02, 3.79457639821368e-02, 3.69656013904915e-02, 3.61005710970801e-02, 3.53423808798539e-02, 3.46827715443538e-02, 3.41135827996055e-02, 3.36268036768956e-02, 3.32146125412152e-02, 3.28694110738468e-02, 3.25838557762239e-02, 3.23508896069313e-02, 3.21637753654898e-02, 3.20161314254506e-02, 3.19019694490805e-02, 3.18157328421194e-02, 3.17523339864181e-02, 3.17071877722981e-02, 3.16762386783438e-02, 3.16559786380139e-02, 3.16434531876553e-02, 3.16362538832067e-02, 3.16324956574050e-02, 3.16307785980285e-02, 3.16301344867611e-02, 3.16299592744969e-02, 3.16299334135834e-02, 3.16299313416124e-02, 3.16298931338395e-02, 3.16296342715390e-02, 3.16286826804912e-02, 3.16261461967144e-02, 3.16205957151681e-02, 3.16099679906756e-02, 3.15914928953021e-02, 3.15616499611764e-02, 3.15161590132078e-02, 3.14500095008867e-02, 3.13575326118070e-02, 3.12325192324465e-02, 3.10683851570985e-02, 3.08583825416804e-02, 3.05958534530077e-02, 3.02745175989401e-02, 2.98887822193876e-02, 2.94340581017991e-02, 2.89070623060894e-02, 2.83060860568014e-02, 2.76312059507973e-02, 2.68844185534789e-02, 2.60696827639322e-02, 2.51928608132062e-02, 2.42615568475344e-02, 2.32848608363696e-02, 2.22730139415717e-02, 2.12370183814699e-02, 2.01882193001808e-02, 1.91378876559583e-02, 1.80968315624170e-02, 1.70750592236004e-02, 1.60815103298830e-02, 1.51238654791115e-02, 1.42084358501791e-02, 1.33401288315866e-02, 1.25224802182942e-02, 1.17577402370301e-02, 1.10469990797526e-02, 1.03903376108008e-02, 9.78699012267725e-03, 9.23550803989315e-03, 8.73391591112238e-03, 8.27985355152805e-03, 7.87070054154526e-03, 7.50368130063350e-03, 7.17595055165532e-03, 6.88466016846689e-03, 6.62700918529424e-03, 6.40027920564989e-03, 6.20185764636005e-03, 6.02925126200660e-03, 5.88009227023178e-03, 5.75213919296402e-03, 5.64327427369449e-03, 5.55149905176222e-03, 5.47492938564176e-03, 5.41179092591904e-03, 5.36041574996204e-03, 5.31924058824984e-03, 5.28680680075411e-03, 5.26176200905581e-03, 5.24286306296062e-03, 5.22897983269388e-03, 5.21909917871710e-03, 5.21232837077434e-03, 5.20789721367976e-03, 5.20515818824663e-03, 5.20358402908982e-03, 5.20276232361994e-03, 5.20238691323982e-03, 5.20224608932551e-03, 5.20220778231817e-03, 5.20220212828804e-03, 5.20220167102536e-03, 5.20219318566717e-03, 5.20213569640759e-03, 5.20192436438549e-03, 5.20136106580082e-03, 5.20012847846374e-03, 5.19776860745935e-03, 5.19366690285494e-03, 5.18704317538266e-03, 5.17695055971501e-03, 5.16228377034622e-03, 5.14179778808455e-03, 5.11413785780235e-03, 5.07788121402086e-03, 5.03159025485900e-03, 4.97387595368082e-03, 4.90346917830209e-03, 4.81929637656719e-03, 4.72055493398216e-03, 4.60678260478176e-03, 4.47791496226557e-03, 4.33432498643072e-03, 4.17683981326888e-03, 4.00673131055385e-03, 3.82567938844147e-03, 3.63570954102954e-03, 3.43910870051900e-03, 3.23832568953719e-03, 3.03586404670853e-03, 2.83417555945675e-03, 2.63556240145135e-03, 2.44209445052945e-03, 2.25554639777670e-03, 2.07735698472788e-03, 1.90861046874602e-03, 1.75003851826716e-03, 1.60203937335268e-03, 1.46471034108231e-03, 1.33788948078006e-03, 1.22120254192737e-03, 1.11411170875897e-03, 1.01596335480281e-03, 9.26032707909030e-04, 8.43563992503960e-04, 7.67805171662417e-04, 6.98036779681388e-04, 6.33594547039706e-04, 5.73885718821998e-04, 5.18399190056584e-04, 4.66709761781203e-04, 4.18477019863417e-04, 3.73439642288531e-04, 3.31406205916096e-04, 2.92243628052802e-04, 2.55864294857133e-04, 2.22212842189151e-04, 1.91253461267138e-04, 1.62958442416539e-04, 1.37298448848355e-04, 1.14234785300482e-04, 9.37137280203615e-05, 7.56628200804244e-05, 5.99889203130033e-05, 4.65777274314405e-05, 3.52944839789870e-05, 2.59855854665147e-05, 1.84808624084098e-05, 1.25963556598663e-05, 8.13746034111773e-06, 4.90236264738542e-06, 2.68574486028683e-06, 1.28278603028775e-06, 4.93562225756250e-07, 0.00000000000000e+00, 0.00000000000000e+00],
[ 9.00000000000000e+01, 8.99999998925484e+01, 8.99999965699330e+01, 8.99999740586856e+01, 8.99998913043740e+01, 8.99996707072712e+01, 8.99991879131181e+01, 8.99982632172368e+01, 8.99966548901297e+01, 8.99940546725942e+01, 8.99900856206528e+01, 8.99843024082369e+01, 8.99761941221455e+01, 8.99651895132051e+01, 8.99506646036065e+01, 8.99319524962882e+01, 8.99083551899130e+01, 8.98791571724356e+01, 8.98436405450887e+01, 8.98011014119189e+01, 8.97508672509703e+01, 8.96923149543618e+01, 8.96248891796178e+01, 8.95481205910054e+01, 8.94616434901321e+01, 8.93652122491703e+01, 8.92587158835453e+01, 8.91421900535578e+01, 8.90158257862651e+01, 8.88799742756028e+01, 8.87351472569189e+01, 8.85820126567482e+01, 8.84213854725473e+01, 8.82542141132138e+01, 8.80815626974847e+01, 8.79045900326609e+01, 8.77245261561060e+01, 8.75426474029671e+01, 8.73602509642263e+01, 8.71786298295056e+01, 8.69990488871212e+01, 8.68227228016894e+01, 8.66507961287013e+01, 8.64843259737001e+01, 8.63242673730777e+01, 8.61714614699609e+01, 8.60266264826488e+01, 8.58903514111677e+01, 8.57630923940852e+01, 8.56451716065325e+01, 8.55367785757006e+01, 8.54379737773871e+01, 8.53486943634263e+01, 8.52687618533495e+01, 8.51978916038838e+01, 8.51357038472268e+01, 8.50817360643536e+01, 8.50354564342000e+01, 8.49962780749294e+01, 8.49635737712602e+01, 8.49366908636684e+01, 8.49149659628514e+01, 8.48977391476871e+01, 8.48843673084212e+01, 8.48742363100300e+01, 8.48667716743138e+01, 8.48614475134475e+01, 8.48577934920616e+01, 8.48553996484566e+01, 8.48539189666973e+01, 8.48530676580137e+01, 8.48526231796785e+01, 8.48524200896358e+01, 8.48523439028665e+01, 8.48523231781536e+01, 8.48523201192270e+01, 8.48523197179604e+01, 8.48523103689291e+01, 8.48522470278533e+01, 8.48520141791527e+01, 8.48513934884306e+01, 8.48500351139647e+01, 8.48474336336913e+01, 8.48429095416891e+01, 8.48355971485016e+01, 8.48244395811288e+01, 8.48081914187530e+01, 8.47854293156890e+01, 8.47545707497844e+01, 8.47139007895869e+01, 8.46616064962197e+01, 8.45958182687038e+01, 8.45146571118555e+01, 8.44162864672930e+01, 8.42989669211522e+01, 8.41611118153489e+01, 8.40013415785020e+01, 8.38185344980177e+01, 8.36118717138098e+01, 8.33808744505650e+01, 8.31254319164452e+01, 8.28458188411441e+01, 8.25427022255500e+01, 8.22171374236765e+01, 8.18705540750366e+01, 8.15047325988308e+01, 8.11217719792934e+01, 8.07240495350039e+01, 8.03141734496067e+01, 7.98949291934518e+01, 7.94692215981196e+01, 7.90400150757838e+01, 7.86102749692566e+01, 7.81829129582162e+01, 7.77607387338779e+01, 7.73464190034318e+01, 7.69424437212174e+01, 7.65510986600681e+01, 7.61744431983995e+01, 7.58142924096787e+01, 7.54722029602180e+01, 7.51494627174715e+01, 7.48470842135485e+01, 7.45658021754855e+01, 7.43060752704993e+01, 7.40680920863028e+01, 7.38517812276612e+01, 7.36568252921879e+01, 7.34826784017333e+01, 7.33285869078201e+01, 7.31936128514161e+01, 7.30766597289706e+01, 7.29765000902133e+01, 7.28918044641781e+01, 7.28211710770697e+01, 7.27631557906090e+01, 7.27163016561155e+01, 7.26791674528335e+01, 7.26503545643916e+01, 7.26285315500417e+01, 7.26124557918552e+01, 7.26009916483195e+01, 7.25931246199755e+01, 7.25879711330508e+01, 7.25847836696462e+01, 7.25829511132459e+01, 7.25819943300206e+01, 7.25815571624663e+01, 7.25813931649508e+01, 7.25813485535679e+01, 7.25813419690182e+01, 7.25813413877278e+01, 7.25813299976285e+01, 7.25812528280170e+01, 7.25809691441652e+01, 7.25802129456592e+01, 7.25785580160545e+01, 7.25753885984847e+01, 7.25698768627347e+01, 7.25609681853470e+01, 7.25473750975504e+01, 7.25275805629324e+01, 7.24998510230997e+01, 7.24622593901765e+01, 7.24127178664456e+01, 7.23490201336016e+01, 7.22688920819047e+01, 7.21700498544195e+01, 7.20502635819749e+01, 7.19074248054979e+01, 7.17396152535706e+01, 7.15451743955141e+01, 7.13227630526997e+01, 7.10714203457735e+01, 7.07906113972312e+01, 7.04802635023098e+01, 7.01407889229581e+01, 6.97730930380999e+01, 6.93785672774254e+01, 6.89590670409552e+01, 6.85168756098666e+01, 6.80546558132588e+01, 6.75753918454555e+01, 6.70823240471909e+01, 6.65788796154500e+01, 6.60686020807771e+01, 6.55550820320507e+01, 6.50418910669322e+01, 6.45325204111110e+01, 6.40303251777699e+01, 6.35384748868812e+01, 6.30599106374796e+01, 6.25973091884031e+01, 6.21530540993568e+01, 6.17292139682861e+01, 6.13275276534629e+01, 6.09493961995228e+01, 6.05958810250138e+01, 6.02677078060314e+01, 5.99652754256822e+01, 5.96886693543854e+01, 5.94376788682925e+01, 5.92118175814043e+01, 5.90103468398533e+01, 5.88323015875254e+01, 5.86765183504382e+01, 5.85416649990212e+01, 5.84262719332842e+01, 5.83287642996358e+01, 5.82474947955755e+01, 5.81807765565859e+01, 5.81269155559359e+01, 5.80842418907231e+01, 5.80511392841761e+01, 5.80260721122506e+01, 5.80076092679493e+01, 5.79944442138953e+01, 5.79854106445750e+01, 5.79794932838275e+01, 5.79758334774338e+01, 5.79737293993637e+01, 5.79726308656672e+01, 5.79721289330493e+01, 5.79719406402296e+01, 5.79718894199637e+01, 5.79718818599579e+01, 5.79718812535519e+01, 5.79718700622856e+01, 5.79717942398068e+01, 5.79715155082998e+01, 5.79707725120240e+01, 5.79691464807430e+01, 5.79660324282495e+01, 5.79606170395822e+01, 5.79518642639301e+01, 5.79385094715022e+01, 5.79190628476384e+01, 5.78918224802952e+01, 5.78548973416241e+01, 5.78062400659961e+01, 5.77436890843154e+01, 5.76650192919908e+01, 5.75680000166757e+01, 5.74504586309731e+01, 5.73103477517121e+01, 5.71458136147327e+01, 5.69552629498613e+01, 5.67374255423475e+01, 5.64914096867091e+01, 5.62167479379753e+01, 5.59134309483166e+01, 5.55819277272725e+01, 5.52231913412203e+01, 5.48386498116759e+01, 5.44301827095659e+01, 5.40000846027376e+01, 5.35510170437764e+01, 5.30859511620737e+01, 5.26081031574239e+01, 5.21208651123293e+01, 5.16277335779335e+01, 5.11322383586715e+01, 5.06378738150535e+01, 5.01480348000997e+01, 4.96659590247564e+01, 4.91946772148768e+01, 4.87369719101308e+01, 4.82953452201399e+01, 4.78719953601495e+01, 4.74688013931921e+01, 4.70873153411731e+01, 4.67287606998357e+01, 4.63940363848523e+01, 4.60837252162156e+01, 4.57981061787014e+01, 4.55371698438471e+01, 4.53006364781933e+01, 4.50879764775878e+01, 4.48984328506497e+01, 4.47310455245877e+01, 4.45846772656172e+01, 4.44580409980946e+01, 4.43497282757199e+01, 4.42582386095693e+01, 4.41820092964408e+01, 4.41194453227067e+01, 4.40689488498694e+01, 4.40289477251932e+01, 4.39979224112625e+01, 4.39744306989860e+01, 4.39571295654232e+01, 4.39447935653286e+01, 4.39363292058402e+01, 4.39307848471290e+01, 4.39273557952415e+01, 4.39253844014472e+01, 4.39243551476731e+01, 4.39238848711704e+01, 4.39237084539025e+01, 4.39236604640984e+01, 4.39236533809027e+01, 4.39236528345785e+01, 4.39236430243494e+01, 4.39235765586047e+01, 4.39233322235083e+01, 4.39226809159434e+01, 4.39212555457457e+01, 4.39185257931252e+01, 4.39137787361133e+01, 4.39061062451677e+01, 4.38943999050757e+01, 4.38773540638966e+01, 4.38534774191844e+01, 4.38211133271146e+01, 4.37784687557839e+01, 4.37236514984621e+01, 4.36547149196744e+01, 4.35697091370627e+01, 4.34667371630001e+01, 4.33440141673653e+01, 4.31999277083459e+01, 4.30330965465548e+01, 4.28424255432032e+01, 4.26271541735650e+01, 4.23868963788603e+01, 4.21216698327331e+01, 4.18319131928580e+01, 4.15184905045242e+01, 4.11826825667603e+01, 4.08261657014498e+01, 4.04509789258080e+01, 4.00594809795282e+01, 3.96542989862966e+01, 3.92382707475154e+01, 3.88143828023206e+01, 3.83857064690548e+01, 3.79553341156490e+01, 3.75263178670538e+01, 3.71016128034099e+01, 3.66840263943993e+01, 3.62761754496086e+01, 3.58804512901281e+01, 3.54989932541204e+01, 3.51336701402684e+01, 3.47860688419031e+01, 3.44574892545526e+01, 3.41489445246847e+01, 3.38611657927927e+01, 3.35946107119080e+01, 3.33494751505043e+01, 3.31257075950653e+01, 3.29230258480790e+01, 3.27409356768214e+01, 3.25787511137977e+01, 3.24356161454426e+01, 3.23105275523419e+01, 3.22023586798178e+01, 3.21098839193262e+01, 3.20318036666536e+01, 3.19667694922837e+01, 3.19134092149421e+01, 3.18703515160655e+01, 3.18362496775010e+01, 3.18098039749530e+01, 3.17897822235542e+01, 3.17750379566508e+01, 3.17645257300847e+01, 3.17573130852624e+01, 3.17525887758093e+01, 3.17496669625247e+01, 3.17479872049840e+01, 3.17471102187058e+01, 3.17467095160986e+01, 3.17465591986629e+01, 3.17465183086706e+01, 3.17465122733933e+01, 3.17465118164203e+01, 3.17465037211895e+01, 3.17464488748175e+01, 3.17462472538616e+01, 3.17457098072583e+01, 3.17445336227749e+01, 3.17422811051187e+01, 3.17383640091661e+01, 3.17320330781719e+01, 3.17223739276933e+01, 3.17083096874101e+01, 3.16886107597723e+01, 3.16619118709872e+01, 3.16267363720523e+01, 3.15815274939735e+01, 3.15246859749701e+01, 3.14546131669131e+01, 3.13697584081983e+01, 3.12686691413598e+01, 3.11500419812360e+01, 3.10127727308343e+01, 3.08560032234787e+01, 3.06791628626200e+01, 3.04820028474806e+01, 3.02646213149719e+01, 3.00274779857073e+01, 2.97713973537915e+01, 2.94975599788789e+01, 2.92074819945469e+01, 2.89029835097515e+01, 2.85861471223012e+01, 2.82592682577534e+01, 2.79247994649068e+01, 2.75852911079106e+01, 2.72433310615121e+01, 2.69014860116330e+01, 2.65622467735231e+01, 2.62279796712720e+01, 2.59008855081926e+01, 2.55829670531693e+01, 2.52760053428341e+01, 2.49815445235609e+01, 2.47008844887367e+01, 2.44350802421373e+01, 2.41849467488988e+01, 2.39510680098862e+01, 2.37338091848093e+01, 2.35333307572140e+01, 2.33496039427007e+01, 2.31824267576146e+01, 2.30314403644857e+01, 2.28961454771470e+01, 2.27759187352361e+01, 2.26700290432805e+01, 2.25776539164020e+01, 2.24978958876584e+01, 2.24297990169341e+01, 2.23723655040930e+01, 2.23245723557857e+01, 2.22853879917659e+01, 2.22537886087602e+01, 2.22287740538490e+01, 2.22093829009286e+01, 2.21947063788429e+01, 2.21839007732564e+01, 2.21761979202826e+01, 2.21709134306879e+01, 2.21674523297158e+01, 2.21653118676271e+01, 2.21640813463281e+01, 2.21634389125534e+01, 2.21631453813650e+01, 2.21630352680237e+01, 2.21630053145572e+01, 2.21630008934895e+01, 2.21630005615985e+01, 2.21629947199421e+01, 2.21629551418804e+01, 2.21628096490353e+01, 2.21624218208053e+01, 2.21615730797668e+01, 2.21599476834231e+01, 2.21571212440952e+01, 2.21525533386322e+01, 2.21455846986414e+01, 2.21354393855524e+01, 2.21212322470620e+01, 2.21019818160105e+01, 2.20766286442187e+01, 2.20440588596025e+01, 2.20031324965992e+01, 2.19527158848432e+01, 2.18917171029190e+01, 2.18191232332588e+01, 2.17340379167401e+01, 2.16357175305023e+01, 2.15236042290076e+01, 2.13973541210691e+01, 2.12568590199103e+01, 2.11022605012550e+01, 2.09339554213859e+01, 2.07525925511703e+01, 2.05590605263171e+01, 2.03544678421797e+01, 2.01401160757450e+01, 1.99174678492887e+01, 1.96881112288375e+01, 1.94537222696723e+01, 1.92160272997087e+01, 1.89767663093648e+01, 1.87376585437036e+01, 1.85003711173547e+01, 1.82664912296775e+01, 1.80375023614518e+01, 1.78147646802050e+01, 1.75994997517412e+01, 1.73927795314859e+01, 1.71955194799829e+01, 1.70084755145551e+01, 1.68322443876797e+01, 1.66672669909196e+01, 1.65138340368653e+01, 1.63720935766201e+01, 1.62420598619369e+01, 1.61236231454978e+01, 1.60165601124596e+01, 1.59205447345347e+01, 1.58351594219086e+01, 1.57599064107985e+01, 1.56942193628129e+01, 1.56374751673867e+01, 1.55890059334360e+01, 1.55481111350410e+01, 1.55140698428078e+01, 1.54861529320068e+01, 1.54636351150321e+01, 1.54458066034803e+01, 1.54319841684022e+01, 1.54215213399358e+01, 1.54138174730154e+01, 1.54083254067595e+01, 1.54045574630577e+01, 1.54020895650201e+01, 1.54005633071888e+01, 1.53996858741819e+01, 1.53992277790470e+01, 1.53990184725027e+01, 1.53989399545384e+01, 1.53989185957622e+01, 1.53989154432521e+01, 1.53989152073118e+01, 1.53989110640825e+01, 1.53988829931041e+01, 1.53987798013665e+01, 1.53985047312191e+01, 1.53979027527050e+01, 1.53967499139641e+01, 1.53947451869586e+01, 1.53915052009946e+01, 1.53865622038931e+01, 1.53793655274345e+01, 1.53692867561605e+01, 1.53556287028829e+01, 1.53376381770653e+01, 1.53145223921375e+01, 1.52854686959906e+01, 1.52496671302965e+01, 1.52063351379508e+01, 1.51547435571410e+01, 1.50942428821297e+01, 1.50242886537695e+01, 1.49444647859972e+01, 1.48545036542517e+01, 1.47543018783263e+01, 1.46439309273007e+01, 1.45236419487481e+01, 1.43938645575592e+01, 1.42551996801319e+01, 1.41084068992155e+01, 1.39543870444276e+01, 1.37941609911986e+01, 1.36288457482252e+01, 1.34596289300307e+01, 1.32877426439173e+01, 1.31144376981060e+01, 1.29409588914831e+01, 1.27685220001649e+01, 1.25982929454519e+01, 1.24313695131655e+01, 1.22687658895823e+01, 1.21114001762382e+01, 1.19600849401945e+01, 1.18155207491906e+01, 1.16782925388814e+01, 1.15488685710682e+01, 1.14276016759280e+01, 1.13147324330905e+01, 1.12103939371784e+01, 1.11146178101918e+01, 1.10273411599632e+01, 1.09484142333188e+01, 1.08776085668675e+01, 1.08146254906688e+01, 1.07591048850406e+01, 1.07106341248526e+01, 1.06687571668734e+01, 1.06329837436832e+01, 1.06027986231813e+01, 1.05776708776493e+01, 1.05570630833152e+01, 1.05404403435770e+01, 1.05272790000121e+01, 1.05170748686356e+01, 1.05093508180498e+01, 1.05036634941727e+01, 1.04996089954922e+01, 1.04968273146754e+01, 1.04950053872034e+01, 1.04938786246764e+01, 1.04932308576679e+01, 1.04928926676846e+01, 1.04927381465211e+01, 1.04926801804005e+01, 1.04926644122199e+01, 1.04926620848698e+01, 1.04926619106210e+01, 1.04926588498535e+01, 1.04926381127107e+01, 1.04925618808863e+01, 1.04923586756479e+01, 1.04919139700986e+01, 1.04910623216546e+01, 1.04895813475747e+01, 1.04871878332294e+01, 1.04835362238986e+01, 1.04782197043072e+01, 1.04707740132295e+01, 1.04606840706200e+01, 1.04473934096336e+01, 1.04303163045999e+01, 1.04088523694223e+01, 1.03824032723574e+01, 1.03503910788561e+01, 1.03122776030223e+01, 1.02675840314464e+01, 1.02159099932540e+01, 1.01569511997120e+01, 1.00905147766003e+01, 1.00165314704119e+01, 9.93506402806136e+00, 9.84631122600902e+00, 9.75060724899285e+00, 9.64841637530041e+00, 9.54032319424458e+00, 9.42701883888543e+00, 9.30928393947212e+00, 9.18796916971218e+00, 9.06397435364342e+00, 8.93822711847328e+00, 8.81166202044959e+00, 8.68520094788968e+00, 8.55973543636190e+00, 8.43611133815146e+00, 8.31511609423461e+00, 8.19746868135020e+00, 8.08381216280448e+00, 7.97470866583623e+00, 7.87063654072779e+00, 7.77198942263622e+00, 7.67907690904502e+00, 7.59212657601581e+00, 7.51128707828195e+00, 7.43663210658724e+00, 7.36816500715534e+00, 7.30582390065422e+00, 7.24948717004325e+00, 7.19897921700891e+00, 7.15407641399125e+00, 7.11451320167912e+00, 7.07998829896708e+00, 7.05017100264482e+00, 7.02470755694222e+00, 7.00322756849758e+00, 6.98535043105977e+00, 6.97069170763389e+00, 6.95886939775386e+00, 6.94950999643528e+00, 6.94225423166591e+00, 6.93676235154047e+00, 6.93271882261033e+00, 6.92983629947997e+00, 6.92785873326115e+00, 6.92656350351011e+00, 6.92576248419149e+00, 6.92530198769326e+00, 6.92506156994173e+00, 6.92495172175231e+00, 6.92491051401197e+00, 6.92489930451768e+00, 6.92489765002016e+00, 6.92489752581925e+00, 6.92489533977914e+00, 6.92488052904228e+00, 6.92482608330582e+00, 6.92468095172227e+00, 6.92436333865832e+00, 6.92375508642414e+00, 6.92269737693861e+00, 6.92098796141001e+00, 6.91838010017122e+00, 6.91458336225121e+00, 6.90926639402871e+00, 6.90206171623784e+00, 6.89257254690835e+00, 6.88038157369687e+00, 6.86506151316598e+00, 6.84618719948878e+00, 6.82334884557980e+00, 6.79616602269409e+00, 6.76430181873629e+00, 6.72747657042903e+00, 6.68548052942808e+00, 6.63818482515450e+00, 6.58555013234692e+00, 6.52763253993675e+00, 6.46458624616373e+00, 6.39666286493946e+00, 6.32420730898471e+00, 6.24765040284167e+00, 6.16749855959125e+00, 6.08432101586079e+00, 5.99873524892219e+00, 5.91139128776332e+00, 5.82295566958032e+00, 5.73409577960410e+00, 5.64546524485780e+00, 5.55769093600964e+00, 5.47136197695177e+00, 5.38702098614722e+00, 5.30515759796339e+00, 5.22620415691169e+00, 5.15053335929873e+00, 5.07845754399416e+00, 5.01022930687250e+00, 4.94604312468911e+00, 4.88603771204691e+00, 4.83029888710283e+00, 4.77886277692065e+00, 4.73171924429832e+00, 4.68881546031304e+00, 4.65005957935641e+00, 4.61532449647756e+00, 4.58445168166708e+00, 4.55725509370623e+00, 4.53352517859298e+00, 4.51303295524963e+00, 4.49553418491011e+00, 4.48077361087523e+00, 4.46848924287197e+00, 4.45841664588330e+00, 4.45029317806691e+00, 4.44386210749767e+00, 4.43887652432991e+00, 4.43510295498673e+00, 4.43232457943974e+00, 4.43034395258492e+00, 4.42898513679893e+00, 4.42809516513999e+00, 4.42754477296637e+00, 4.42722835908205e+00, 4.42706316453793e+00, 4.42698768624726e+00, 4.42695937180920e+00, 4.42695166960257e+00, 4.42695053277325e+00, 4.42695044706956e+00, 4.42694893377084e+00, 4.42693868095412e+00, 4.42690099058612e+00, 4.42680052244701e+00, 4.42658065314194e+00, 4.42615958769474e+00, 4.42542738443647e+00, 4.42424403812004e+00, 4.42243874757430e+00, 4.41981047162338e+00, 4.41612984893108e+00, 4.41114252230042e+00, 4.40457386444012e+00, 4.39613504968385e+00, 4.38553035510388e+00, 4.37246550686991e+00, 4.35665681723516e+00, 4.33784078952374e+00, 4.31578380963253e+00, 4.29029150020038e+00, 4.26121729477427e+00, 4.22846979954640e+00, 4.19201855239552e+00, 4.15189786227833e+00, 4.10820851173510e+00, 4.06111722299089e+00, 4.01085391290030e+00, 3.95770688199539e+00, 3.90201618742647e+00, 3.84416553065473e+00, 3.78457304411072e+00, 3.72368138620762e+00, 3.66194755363753e+00, 3.59983279806396e+00, 3.53779299592132e+00, 3.47626976940680e+00, 3.41568259760078e+00, 3.35642209216839e+00, 3.29884454537033e+00, 3.24326779241747e+00, 3.18996836901725e+00, 3.13917989168452e+00, 3.09109254590901e+00, 3.04585353751601e+00, 3.00356834621526e+00, 2.96430261681878e+00, 2.92808453121879e+00, 2.89490752047415e+00, 2.86473319838259e+00, 2.83749442281309e+00, 2.81309841619237e+00, 2.79142989968075e+00, 2.77235421507477e+00, 2.75572042323513e+00, 2.74136437728573e+00, 2.72911177286906e+00, 2.71878117669733e+00, 2.71018702918238e+00, 2.70314260803312e+00, 2.69746292856983e+00, 2.69296754447147e+00, 2.68948320115651e+00, 2.68684628437141e+00, 2.68490500005575e+00, 2.68352121913119e+00, 2.68257192314822e+00, 2.68195019391410e+00, 2.68156570207307e+00, 2.68134466544566e+00, 2.68122926672736e+00, 2.68117654062988e+00, 2.68115676133617e+00, 2.68115138089681e+00, 2.68115058675552e+00, 2.68115052656070e+00, 2.68114945935933e+00, 2.68114222891788e+00, 2.68111564912311e+00, 2.68104479770903e+00, 2.68088974396402e+00, 2.68059280838261e+00, 2.68007646899518e+00, 2.67924201878829e+00, 2.67796906639440e+00, 2.67611595714486e+00, 2.67352117244030e+00, 2.67000574092462e+00, 2.66537666408946e+00, 2.65943132113799e+00, 2.65196277345675e+00, 2.64276583919434e+00, 2.63164375572998e+00, 2.61841519592299e+00, 2.60292135763457e+00, 2.58503281036204e+00, 2.56465576316634e+00, 2.54173741893879e+00, 2.51627010450814e+00, 2.48829391508737e+00, 2.45789768348997e+00, 2.42521817510759e+00, 2.39043751209627e+00, 2.35377893600803e+00, 2.31550111780878e+00, 2.27589130867835e+00, 2.23525768647728e+00, 2.19392128600840e+00, 2.15220790405864e+00, 2.11044034390289e+00, 2.06893131281804e+00, 2.02797721690439e+00, 1.98785301821138e+00, 1.94880823807186e+00, 1.91106411504617e+00, 1.87481186167579e+00, 1.84021191493594e+00, 1.80739404230064e+00, 1.77645814822688e+00, 1.74747562274642e+00, 1.72049108195722e+00, 1.69552436641222e+00, 1.67257268464192e+00, 1.65161281260120e+00, 1.63260328346800e+00, 1.61548652426453e+00, 1.60019091504546e+00, 1.58663276217064e+00, 1.57471818907538e+00, 1.56434495586288e+00, 1.55540422307558e+00, 1.54778227542553e+00, 1.54136221848319e+00, 1.53602565587722e+00, 1.53165434708740e+00, 1.52813183717214e+00, 1.52534504059260e+00, 1.52318575254694e+00, 1.52155205377878e+00, 1.52034956945779e+00, 1.51949254009223e+00, 1.51890466294692e+00, 1.51851966626779e+00, 1.51828158561831e+00, 1.51814472139145e+00, 1.51807326839793e+00, 1.51804062149836e+00, 1.51802837461038e+00, 1.51802504316852e+00, 1.51802455145499e+00, 1.51802451394728e+00, 1.51802384584843e+00, 1.51801931938523e+00, 1.51800267969928e+00, 1.51795832495513e+00, 1.51786125847021e+00, 1.51767537522166e+00, 1.51735215611956e+00, 1.51682983786152e+00, 1.51603311945479e+00, 1.51487345782147e+00, 1.51324999328942e+00, 1.51105113014064e+00, 1.50815677683302e+00, 1.50444122450989e+00, 1.49977661109861e+00, 1.49403688266430e+00, 1.48710212570323e+00, 1.47886310669720e+00, 1.46922582228179e+00, 1.45811583904070e+00, 1.44548219038650e+00, 1.43130060265651e+00, 1.41557584550058e+00, 1.39834304296980e+00, 1.37966783928846e+00, 1.35964538272631e+00, 1.33839816607910e+00, 1.31607283575065e+00, 1.29283614598158e+00, 1.26887028405149e+00, 1.24436782184604e+00, 1.21952655706390e+00, 1.19454449420446e+00, 1.16961518434907e+00, 1.14492359836936e+00, 1.12064265621403e+00, 1.09693048102977e+00, 1.07392839608144e+00, 1.05175963859413e+00, 1.03052873018618e+00, 1.01032141953962e+00, 9.91205099208187e-01, 9.73229593928195e-01, 9.56428220849503e-01, 9.40819030850804e-01, 9.26406152648099e-01, 9.13181176013325e-01, 9.01124525629839e-01, 8.90206791773565e-01, 8.80389997289996e-01, 8.71628791679939e-01, 8.63871572192295e-01, 8.57061538506983e-01, 8.51137691878222e-01, 8.46035791594616e-01, 8.41689281486836e-01, 8.38030197222892e-01, 8.34990061593899e-01, 8.32500770283147e-01, 8.30495465163978e-01, 8.28909386462641e-01, 8.27680689663950e-01, 8.26751208330435e-01, 8.26067140522341e-01, 8.25579634626237e-01, 8.25245250391930e-01, 8.25026272940789e-01, 8.24890861378435e-01, 8.24813019174798e-01, 8.24772380269266e-01, 8.24753812414968e-01, 8.24746847036053e-01, 8.24744952290160e-01, 8.24744672629876e-01, 8.24744651146958e-01, 8.24744266513292e-01, 8.24741660566886e-01, 8.24732080882768e-01, 8.24706545349926e-01, 8.24650663428741e-01, 8.24543650404976e-01, 8.24357577453327e-01, 8.24056897435185e-01, 8.23598282814217e-01, 8.22930807364514e-01, 8.21996496792963e-01, 8.20731264295603e-01, 8.19066234796821e-01, 8.16929445828208e-01, 8.14247893753318e-01, 8.10949871968646e-01, 8.06967524093466e-01, 8.02239511934715e-01, 7.96713677651198e-01, 7.90349564797367e-01, 7.83120656486986e-01, 7.75016192944534e-01, 7.66042446418055e-01, 7.56223358706023e-01, 7.45600483797915e-01, 7.34232222302509e-01, 7.22192381281587e-01, 7.09568138083512e-01, 6.96457525150535e-01, 6.82966580705823e-01, 6.69206325195930e-01, 6.55289724511562e-01, 6.41328789158138e-01, 6.27431935967849e-01, 6.13701708904316e-01, 6.00232921724782e-01, 5.87111251362714e-01, 5.74412279966509e-01, 5.62200957843459e-01, 5.50531440421977e-01, 5.39447240154189e-01, 5.28981628654506e-01, 5.19158224385876e-01, 5.09991705625055e-01, 5.01488595920947e-01, 4.93648078523663e-01, 4.86462806198332e-01, 4.79919682562056e-01, 4.74000599968264e-01, 4.68683126602940e-01, 4.63941141646941e-01, 4.59745422024866e-01, 4.56064187440518e-01, 4.52863612179740e-01, 4.50108312678466e-01, 4.47761819259966e-01, 4.45787038913011e-01, 4.44146713697549e-01, 4.42803876532494e-01, 4.41722302964749e-01, 4.40866954284431e-01, 4.40204404296779e-01, 4.39703239444938e-01, 4.39334420041632e-01, 4.39071589312533e-01, 4.38891316921968e-01, 4.38773264694490e-01, 4.38700264332596e-01, 4.38658299928264e-01, 4.38636391758450e-01, 4.38626381970167e-01, 4.38622626991076e-01, 4.38621605549301e-01, 4.38621454786784e-01, 4.38621443117142e-01, 4.38621233031313e-01, 4.38619809671025e-01, 4.38614577278699e-01, 4.38600629877143e-01, 4.38570107524015e-01, 4.38511658141394e-01, 4.38410028489457e-01, 4.38245806377878e-01, 4.37995334381910e-01, 4.37630812839899e-01, 4.37120606355435e-01, 4.36429762987260e-01, 4.35520748409701e-01, 4.34354388350485e-01, 4.32891001595863e-01, 4.31091693168327e-01, 4.28919763703917e-01, 4.26342177748225e-01, 4.23331022143373e-01, 4.19864877549671e-01, 4.15930023030398e-01, 4.11521396785616e-01, 4.06643246197419e-01, 4.01309417155048e-01, 3.95543255053283e-01, 3.89377115909951e-01, 3.82851513096938e-01, 3.76013950325785e-01, 3.68917512032297e-01, 3.61619296030739e-01, 3.54178779036459e-01, 3.46656203236691e-01, 3.39111062414773e-01, 3.31600750916010e-01, 3.24179420194014e-01, 3.16897068124320e-01, 3.09798867813882e-01, 3.02924726876441e-01, 2.96309056078329e-01, 2.89980718253190e-01, 2.83963124288334e-01, 2.78274442264063e-01, 2.72927887731029e-01, 2.67932066811852e-01, 2.63291348530558e-01, 2.59006247855069e-01, 2.55073805895620e-01, 2.51487958213573e-01, 2.48239886074984e-01, 2.45318348645512e-01, 2.42709996551653e-01, 2.40399668955689e-01, 2.38370677361772e-01, 2.36605079851941e-01, 2.35083949411191e-01, 2.33787639502992e-01, 2.32696049170173e-01, 2.31788888732837e-01, 2.31045945723892e-01, 2.30447349141727e-01, 2.29973828528458e-01, 2.29606962922276e-01, 2.29329413513653e-01, 2.29125132977728e-01, 2.28979544056259e-01, 2.28879680095389e-01, 2.28814280928901e-01, 2.28773838719803e-01, 2.28750590053565e-01, 2.28738452615204e-01, 2.28732907024565e-01, 2.28730826699064e-01, 2.28730260801683e-01, 2.28730177276475e-01, 2.28730170758865e-01, 2.28730052746727e-01, 2.28729253198068e-01, 2.28726313987480e-01, 2.28718479245174e-01, 2.28701333681698e-01, 2.28668500092809e-01, 2.28611409140370e-01, 2.28519153594988e-01, 2.28378438110497e-01, 2.28173634124644e-01, 2.27886947435351e-01, 2.27498703172316e-01, 2.26987749080081e-01, 2.26331973141086e-01, 2.25508925626855e-01, 2.24496528873061e-01, 2.23273850856145e-01, 2.21821911627231e-01, 2.20124485630990e-01, 2.18168858780103e-01, 2.15946497696663e-01, 2.13453590404470e-01, 2.10691423245407e-01, 2.07666567734819e-01, 2.04390862807813e-01, 2.00881191346824e-01, 1.97159063628874e-01, 1.93250032907841e-01, 1.89182978418302e-01, 1.84989297682620e-01, 1.80702052655136e-01, 1.76355113037019e-01, 1.71982335612583e-01, 1.67616811592282e-01, 1.63290205728671e-01, 1.59032202392012e-01, 1.54870065679788e-01, 1.50828313596634e-01, 1.46928500733551e-01, 1.43189099811337e-01, 1.39625469845939e-01, 1.36249897330629e-01, 1.33071696466033e-01, 1.30097354886061e-01, 1.27330712342015e-01, 1.24773161253827e-01, 1.22423859755031e-01, 1.20279949689191e-01, 1.18336773843723e-01, 1.16588088442029e-01, 1.15026268502277e-01, 1.13642505053036e-01, 1.12426994316849e-01, 1.11369119780377e-01, 1.10457628534633e-01, 1.09680803386155e-01, 1.09026632035068e-01, 1.08482974129444e-01, 1.08037726296800e-01, 1.07678984394710e-01, 1.07395201292434e-01, 1.07175337591357e-01, 1.07009001903180e-01, 1.06886576721678e-01, 1.06799325616834e-01, 1.06739477499417e-01, 1.06700284069362e-01, 1.06676047257068e-01, 1.06662114456967e-01, 1.06654840558235e-01, 1.06651517117442e-01, 1.06650270390361e-01, 1.06649931251326e-01, 1.06649881195153e-01, 1.06649877255961e-01, 1.06649805504361e-01, 1.06649319377325e-01, 1.06647532332360e-01, 1.06642768798494e-01, 1.06632344286169e-01, 1.06612381496461e-01, 1.06577670413969e-01, 1.06521579825426e-01, 1.06436027046641e-01, 1.06311511760936e-01, 1.06137218655306e-01, 1.05901191852041e-01, 1.05590581862174e-01, 1.05191962869341e-01, 1.04691714601032e-01, 1.04076458974603e-01, 1.03333537356860e-01, 1.02451509998303e-01, 1.01420655447797e-01, 1.00233445018631e-01, 9.88849661380442e-02, 9.73732690583944e-02, 9.56996141344462e-02, 9.38686016593238e-02, 9.18881728074865e-02, 8.97694780434932e-02, 8.75266177095000e-02, 8.51762676295635e-02, 8.27372096787400e-02, 8.02297927149880e-02, 7.76753526036771e-02, 7.50956210884794e-02, 7.25121520610799e-02, 6.99457906600713e-02, 6.74162060361425e-02, 6.49415030520260e-02, 6.25379221450970e-02, 6.02196305570430e-02, 5.79986026089756e-02, 5.58845820876286e-02, 5.38851163863967e-02, 5.20056498872350e-02, 5.02496630719662e-02, 4.86188438293027e-02, 4.71132781909489e-02, 4.57316491587250e-02, 4.44714341982717e-02, 4.33290941812317e-02, 4.23002488023769e-02, 4.13798355685060e-02, 4.05622512121084e-02, 3.98414758027741e-02, 3.92111809019156e-02, 3.86648238660143e-02, 3.81957308602949e-02, 3.77971712857885e-02, 3.74624261921355e-02, 3.71848528714764e-02, 3.69579472539669e-02, 3.67754050458018e-02, 3.66311817823686e-02, 3.65195512473946e-02, 3.64351610137987e-02, 3.63730833272768e-02, 3.63288591635751e-02, 3.62985331392167e-02, 3.62786769974690e-02, 3.62663996745312e-02, 3.62593424425057e-02, 3.62556582241780e-02, 3.62539749411610e-02, 3.62533434943515e-02, 3.62531717265853e-02, 3.62531463742641e-02, 3.62531443612681e-02, 3.62531074742378e-02, 3.62528575593899e-02, 3.62519388554858e-02, 3.62494899978429e-02, 3.62441310871731e-02, 3.62338694869493e-02, 3.62160287209576e-02, 3.61872046906342e-02, 3.61432532668573e-02, 3.60793129825367e-02, 3.59898660063038e-02, 3.58688396644453e-02, 3.57097494673724e-02, 3.55058827353076e-02, 3.52505196046593e-02, 3.49371854318434e-02, 3.45599256122142e-02, 3.41135908315380e-02, 3.35941181284767e-02, 3.29987911959648e-02, 3.23264625299089e-02, 3.15777206057533e-02, 3.07549874635384e-02, 2.98625359097713e-02, 2.89064207763516e-02, 2.78943248255246e-02, 2.68353262976291e-02, 2.57396010284765e-02, 2.46180768806710e-02, 2.34820615424240e-02, 2.23428663113197e-02, 2.12114481835783e-02, 2.00980902821044e-02, 1.90121365665590e-02, 1.79617910985286e-02, 1.69539860028465e-02, 1.59943166132877e-02, 1.50870379930126e-02, 1.42351143175574e-02, 1.34403109655636e-02, 1.26989507257207e-02, 1.20201062961343e-02, 1.13977109781160e-02, 1.08301307765544e-02, 1.03152393835728e-02, 9.85054722502809e-03, 9.43331477230913e-03, 9.06064816462049e-03, 8.72957672231678e-03, 8.43711340110076e-03, 8.18030026796635e-03, 7.95624148102547e-03, 7.76212616241878e-03, 7.59524328979921e-03, 7.45299053642157e-03, 7.33287889956102e-03, 7.23253486823694e-03, 7.14970167966555e-03, 7.08224087621545e-03, 7.02813495247767e-03, 6.98549144826188e-03, 6.95254845913951e-03, 6.92768121788845e-03, 6.90940915478677e-03, 6.89640267070502e-03, 6.88748875438705e-03, 6.88165455446210e-03, 6.87804807648718e-03, 6.87597531316442e-03, 6.87489331895108e-03, 6.87439898680964e-03, 6.87421355223705e-03, 6.87416311014033e-03, 6.87415566501670e-03, 6.87415506850186e-03, 6.87414406827713e-03, 6.87406954051539e-03, 6.87379557381007e-03, 6.87306531960985e-03, 6.87146738137301e-03, 6.86840792484134e-03, 6.86308996425566e-03, 6.85450132731091e-03, 6.84141283650774e-03, 6.82238822017212e-03, 6.79580712751715e-03, 6.75990230577383e-03, 6.71281145140767e-03, 6.65264343832916e-03, 6.57755756240874e-03, 6.48585316801586e-03, 6.37606564500004e-03, 6.24706343926107e-03, 6.09813957890992e-03, 5.92909046488899e-03, 5.74027451129666e-03, 5.53264384832630e-03, 5.30774388876409e-03, 5.06767814258069e-03, 4.81503900434177e-03, 4.55280876811059e-03, 4.28423807005629e-03, 4.01271071089582e-03, 3.74160430709480e-03, 3.47415598098166e-03, 3.21334191608003e-03, 2.96177903342709e-03, 2.72270625671922e-03, 2.49575235203281e-03, 2.28321752159388e-03, 2.08589881353305e-03, 1.90416189343523e-03, 1.73799345418502e-03, 1.58707098256627e-03, 1.45083041597616e-03, 1.32853177983350e-03, 1.21931687581878e-03, 1.12224799525507e-03, 1.03634174638501e-03, 9.60606282404797e-04, 8.94070666134894e-04, 8.35810730507092e-04, 7.84973109432485e-04, 7.40782236917426e-04, 7.02529495967497e-04, 6.69564250358690e-04, 6.41293859700909e-04, 6.17180991795863e-04, 5.96735926691849e-04, 5.79512799404824e-04, 5.65110790788026e-04, 5.53173225057435e-04, 5.43381970771861e-04, 5.35450263683401e-04, 5.29117128607481e-04, 5.24144180445774e-04, 5.20314089683893e-04, 5.17429868608588e-04, 5.15314518182208e-04, 5.13810843997931e-04, 5.12781330854750e-04, 5.12107966400777e-04, 5.11691901095901e-04, 5.11452838125592e-04, 5.11328064419558e-04, 5.11271063132313e-04, 5.11249681423465e-04, 5.11243865204847e-04, 5.11243006755535e-04, 5.11242937349721e-04, 5.11241649434169e-04, 5.11232923628313e-04, 5.11200847668647e-04, 5.11115352551883e-04, 5.10928287283718e-04, 5.10570183394066e-04, 5.09947902898489e-04, 5.08943384133431e-04, 5.07413721507533e-04, 5.05192835134949e-04, 5.02094992441471e-04, 4.97920422162679e-04, 4.92463182752090e-04, 4.85521269084188e-04, 4.76908633970652e-04, 4.66468414856536e-04, 4.54086344589655e-04, 4.39703195695529e-04, 4.23324998308772e-04, 4.05029449780159e-04, 3.84966824896082e-04, 3.63355117926289e-04, 3.40472549950732e-04, 3.16651501321985e-04, 2.92270762855651e-04, 2.67737078900154e-04, 2.43456324341196e-04, 2.19804801631085e-04, 1.97107030658117e-04, 1.75624128388563e-04, 1.55557580058030e-04, 1.37060252342280e-04, 1.20232664031050e-04, 1.05099868221714e-04, 9.15928009845857e-05, 7.95592649751597e-05, 6.88162995317043e-05, 5.92248322277555e-05, 5.07305728092433e-05, 4.33230282226305e-05, 3.69366494936685e-05, 3.14084346086929e-05, 2.64221237955553e-05, 2.15412383681723e-05, 1.68678526364374e-05, 1.24563358383512e-05, 8.35107451280237e-06, 4.79355639981804e-06, 1.93378340895833e-06, 1.31878425690957e-06, 1.07258581749647e-06, 3.68708569453702e-07, 1.01880503677876e-06, 1.19126501374467e-06, 1.11080674969344e-06, 9.39601122240674e-07, 7.87580873407023e-07, 7.01062887742410e-07, 6.69919785423190e-07, 6.55644555536128e-07, 6.27109713491336e-07, 5.73233208456259e-07, 4.97432694146482e-07, 4.09490468200602e-07, 3.19896500400318e-07, -1.42108547152020e-14, -4.26325641456060e-14, -4.26325641456060e-14, -2.84217094304040e-14, -2.84217094304040e-14, -1.42108547152020e-14, -2.84217094304040e-14, -2.84217094304040e-14, -1.42108547152020e-14],
[ 9.10000000000000e+01, 9.09999998907847e+01, 9.09999965136331e+01, 9.09999736328937e+01, 9.09998895202803e+01, 9.09996653023638e+01, 9.09991745837434e+01, 9.09982347100077e+01, 9.09965999836314e+01, 9.09939570848381e+01, 9.09899228805980e+01, 9.09840447313753e+01, 9.09758033306518e+01, 9.09646180405804e+01, 9.09498546222793e+01, 9.09308352045211e+01, 9.09068502918572e+01, 9.08771725825127e+01, 9.08410723451633e+01, 9.07978340868881e+01, 9.07467742251253e+01, 9.06872594466235e+01, 9.06187253898040e+01, 9.05406952209193e+01, 9.04527975919550e+01, 9.03547833792922e+01, 9.02465405231288e+01, 9.01281062390443e+01, 8.99996758755899e+01, 8.98616077613404e+01, 8.97144235281624e+01, 8.95588036089828e+01, 8.93955778699747e+01, 8.92257116206701e+01, 8.90502875179176e+01, 8.88704841088461e+01, 8.86875519192255e+01, 8.85027880731787e+01, 8.83175104273014e+01, 8.81330321276361e+01, 8.79506373706585e+01, 8.77715589922685e+01, 8.75969583439020e+01, 8.74279077604024e+01, 8.72653757922931e+01, 8.71102152712787e+01, 8.69631542023140e+01, 8.68247894246098e+01, 8.66955829515524e+01, 8.65758608791895e+01, 8.64658147388436e+01, 8.63655051569753e+01, 8.62748676716111e+01, 8.61937205378423e+01, 8.61217743346679e+01, 8.60586431621815e+01, 8.60038571927854e+01, 8.59568763140830e+01, 8.59171045758914e+01, 8.58839051310953e+01, 8.58566153415056e+01, 8.58345617071835e+01, 8.58170742723880e+01, 8.58035001648051e+01, 8.57932159380846e+01, 8.57856384116373e+01, 8.57802337363005e+01, 8.57765244594761e+01, 8.57740944176640e+01, 8.57725913463779e+01, 8.57717271651207e+01, 8.57712759659044e+01, 8.57710698049813e+01, 8.57709924662101e+01, 8.57709714281238e+01, 8.57709683229440e+01, 8.57709679156099e+01, 8.57709584252142e+01, 8.57708941263740e+01, 8.57706577568322e+01, 8.57700276808545e+01, 8.57686487670603e+01, 8.57660079517999e+01, 8.57614154568087e+01, 8.57539925085613e+01, 8.57426662662988e+01, 8.57261725036354e+01, 8.57030664012583e+01, 8.56717415915360e+01, 8.56304573467586e+01, 8.55773735203966e+01, 8.55105925379213e+01, 8.54282073980435e+01, 8.53283543006404e+01, 8.52092681854860e+01, 8.50693391756049e+01, 8.49071677073702e+01, 8.47216160371231e+01, 8.45118538786173e+01, 8.42773961699620e+01, 8.40181313879842e+01, 8.37343393785625e+01, 8.34266982706787e+01, 8.30962805827435e+01, 8.27445390137211e+01, 8.23732825913261e+01, 8.19846438645983e+01, 8.15810378106531e+01, 8.11651132571740e+01, 8.07396980423155e+01, 8.03077398305790e+01, 7.98722452588358e+01, 7.94362205407565e+01, 7.90026164966218e+01, 7.85742801415558e+01, 7.81539137310974e+01, 7.77440409898274e+01, 7.73469795250168e+01, 7.69648182784911e+01, 7.65993991555676e+01, 7.62523024252187e+01, 7.59248358796347e+01, 7.56180279584743e+01, 7.53326250782525e+01, 7.50690933174694e+01, 7.48276244630713e+01, 7.46081462773189e+01, 7.44103367249983e+01, 7.42336418169817e+01, 7.40772966723320e+01, 7.39403493669362e+01, 7.38216871113813e+01, 7.37200642763753e+01, 7.36341317559774e+01, 7.35624671262759e+01, 7.35036050218472e+01, 7.34560671183526e+01, 7.34183910822074e+01, 7.33891578330806e+01, 7.33670164674310e+01, 7.33507062158496e+01, 7.33390748566958e+01, 7.33310930845917e+01, 7.33258644339667e+01, 7.33226304821327e+01, 7.33207711985848e+01, 7.33198004611139e+01, 7.33193569176886e+01, 7.33191905283541e+01, 7.33191452663374e+01, 7.33191385857555e+01, 7.33191379959873e+01, 7.33191264397694e+01, 7.33190481446798e+01, 7.33187603234515e+01, 7.33179930962201e+01, 7.33163140305255e+01, 7.33130983894789e+01, 7.33075062707873e+01, 7.32984676733971e+01, 7.32846763594234e+01, 7.32645931834796e+01, 7.32364593346773e+01, 7.31983196734651e+01, 7.31480560425853e+01, 7.30834300885933e+01, 7.30021347525668e+01, 7.29018531873472e+01, 7.27803234525522e+01, 7.26354069529374e+01, 7.24651582504535e+01, 7.22678936271430e+01, 7.20422556342412e+01, 7.17872708561079e+01, 7.15023982612310e+01, 7.11875658126778e+01, 7.08431934642718e+01, 7.04702012646592e+01, 7.00700020066004e+01, 6.96444786554545e+01, 6.91959476117048e+01, 6.87271096306951e+01, 6.82409908502738e+01, 6.77408767818423e+01, 6.72302422498531e+01, 6.67126801156421e+01, 6.61918312458087e+01, 6.56713176790546e+01, 6.51546804201288e+01, 6.46453228387987e+01, 6.41464603246354e+01, 6.36610766412218e+01, 6.31918872923877e+01, 6.27413101008944e+01, 6.23114430634134e+01, 6.19040493722858e+01, 6.15205493025899e+01, 6.11620184873794e+01, 6.08291919775979e+01, 6.05224734236886e+01, 6.02419487212295e+01, 5.99874035158398e+01, 5.97583440392990e+01, 5.95540208267123e+01, 5.93734549270182e+01, 5.92154662568077e+01, 5.90787037572621e+01, 5.89616769974887e+01, 5.88627888289830e+01, 5.87803686413827e+01, 5.87127057061374e+01, 5.86580820298446e+01, 5.86148040807291e+01, 5.85812327079078e+01, 5.85558105509511e+01, 5.85370862429112e+01, 5.85237347477897e+01, 5.85145732454488e+01, 5.85085720827224e+01, 5.85048604457274e+01, 5.85027265693813e+01, 5.85016124780592e+01, 5.85011034369788e+01, 5.85009124775206e+01, 5.85008605318636e+01, 5.85008528647914e+01, 5.85008522497974e+01, 5.85008409000383e+01, 5.85007640037470e+01, 5.85004813247903e+01, 5.84997278060519e+01, 5.84980787465639e+01, 5.84949205920876e+01, 5.84894285090390e+01, 5.84805517730901e+01, 5.84670078427822e+01, 5.84472858008232e+01, 5.84196596254751e+01, 5.83822114953368e+01, 5.83328650281584e+01, 5.82694280068685e+01, 5.81896437580246e+01, 5.80912499308065e+01, 5.79720429979310e+01, 5.78299463909957e+01, 5.76630798259488e+01, 5.74698271077976e+01, 5.72488995652479e+01, 5.69993922881438e+01, 5.67208305442754e+01, 5.64132041409057e+01, 5.60769880519268e+01, 5.57131483122456e+01, 5.53231329252588e+01, 5.49088482650093e+01, 5.44726221136657e+01, 5.40171550086795e+01, 5.35454619634765e+01, 5.30608068810474e+01, 5.25666321278873e+01, 5.20664858002133e+01, 5.15639492012068e+01, 5.10625669425702e+01, 5.05657818623217e+01, 5.00768765996677e+01, 4.95989231988504e+01, 4.91347415690179e+01, 4.86868670698937e+01, 4.82575269912813e+01, 4.78486253028936e+01, 4.74617347990999e+01, 4.70980956521733e+01, 4.67586193956221e+01, 4.64438974509129e+01, 4.61542134480150e+01, 4.58895587393516e+01, 4.56496506437703e+01, 4.54339530678184e+01, 4.52416992299588e+01, 4.50719162590403e+01, 4.49234514540453e+01, 4.47949999818631e+01, 4.46851337579198e+01, 4.45923312054168e+01, 4.45150075275140e+01, 4.44515450584593e+01, 4.44003231906025e+01, 4.43597473112869e+01, 4.43282761339874e+01, 4.43044467787407e+01, 4.42868969540212e+01, 4.42743836203077e+01, 4.42657975771057e+01, 4.42601735099707e+01, 4.42566951592932e+01, 4.42546954228011e+01, 4.42536513713558e+01, 4.42531743336226e+01, 4.42529953799754e+01, 4.42529467002138e+01, 4.42529395151818e+01, 4.42529389610030e+01, 4.42529290097306e+01, 4.42528615883963e+01, 4.42526137404493e+01, 4.42519530688956e+01, 4.42505072057430e+01, 4.42477382063001e+01, 4.42429228976025e+01, 4.42351400907508e+01, 4.42232654278965e+01, 4.42059744718390e+01, 4.41817544539542e+01, 4.41489248684051e+01, 4.41056668326320e+01, 4.40500608246062e+01, 4.39801320600693e+01, 4.38939023985219e+01, 4.37894472832911e+01, 4.36649558543844e+01, 4.35187920549521e+01, 4.33495543182454e+01, 4.31561313061472e+01, 4.29377512005450e+01, 4.26940222409393e+01, 4.24249625548704e+01, 4.21310178215434e+01, 4.18130659038914e+01, 4.14724082263767e+01, 4.11107483058975e+01, 4.07301584084172e+01, 4.03330357692162e+01, 3.99220501692340e+01, 3.95000849155075e+01, 3.90701734526264e+01, 3.86354339494704e+01, 3.81990042531577e+01, 3.77639795458075e+01, 3.73333548343236e+01, 3.69099740223883e+01, 3.64964867765175e+01, 3.60953137769050e+01, 3.57086203441638e+01, 3.53382979543408e+01, 3.49859528545558e+01, 3.46529008724127e+01, 3.43401675315265e+01, 3.40484926820959e+01, 3.37783389749416e+01, 3.35299036157917e+01, 3.33031329213315e+01, 3.30977392617342e+01, 3.29132200237618e+01, 3.27488782707510e+01, 3.26038448134245e+01, 3.24771014369290e+01, 3.23675050508146e+01, 3.22738125355026e+01, 3.21947060483070e+01, 3.21288185238973e+01, 3.20747590607339e+01, 3.20311378315903e+01, 3.19965900998805e+01, 3.19697988723673e+01, 3.19495156812336e+01, 3.19345789719891e+01, 3.19239295840456e+01, 3.19166228515343e+01, 3.19118369237654e+01, 3.19088770055568e+01, 3.19071753427050e+01, 3.19062869202391e+01, 3.19058809924253e+01, 3.19057287148475e+01, 3.19056872916503e+01, 3.19056811776730e+01, 3.19056807147411e+01, 3.19056725139488e+01, 3.19056169523820e+01, 3.19054127023046e+01, 3.19048682475066e+01, 3.19036767261562e+01, 3.19013948380816e+01, 3.18974266714746e+01, 3.18910132094934e+01, 3.18812281670362e+01, 3.18669806780548e+01, 3.18470251987916e+01, 3.18199788066518e+01, 3.17843458537533e+01, 3.17385496770913e+01, 3.16809707767080e+01, 3.16099905578850e+01, 3.15240394080274e+01, 3.14216475643890e+01, 3.13014969505602e+01, 3.11624719453576e+01, 3.10037069240266e+01, 3.08246284001734e+01, 3.06249897109415e+01, 3.04048964301125e+01, 3.01648210550503e+01, 2.99056059745598e+01, 2.96284542594679e+01, 2.93349083960927e+01, 2.90268176737998e+01, 2.87062955104266e+01, 2.83756685211830e+01, 2.80374195726809e+01, 2.76941273765631e+01, 2.73484053305328e+01, 2.70028422807147e+01, 2.66599476479558e+01, 2.63221029463153e+01, 2.59915211656840e+01, 2.56702148551763e+01, 2.53599731036698e+01, 2.50623470399025e+01, 2.47786430226336e+01, 2.45099223933836e+01, 2.42570065258199e+01, 2.40204859097719e+01, 2.38007321215070e+01, 2.35979117158459e+01, 2.34120012915643e+01, 2.32428031974040e+01, 2.30899615391978e+01, 2.29529783057283e+01, 2.28312295465800e+01, 2.27239816096779e+01, 2.26304074832218e+01, 2.25496032917667e+01, 2.24806049752609e+01, 2.24224051388600e+01, 2.23739700059863e+01, 2.23342563429961e+01, 2.23022281565347e+01, 2.22768728998998e+01, 2.22572168681876e+01, 2.22423394191614e+01, 2.22313856325471e+01, 2.22235770186611e+01, 2.22182199102651e+01, 2.22147112197724e+01, 2.22125413158074e+01, 2.22112938650163e+01, 2.22106425915525e+01, 2.22103450212161e+01, 2.22102333926151e+01, 2.22102030269578e+01, 2.22101985450515e+01, 2.22101982085934e+01, 2.22101922865495e+01, 2.22101521638493e+01, 2.22100046688353e+01, 2.22096115034011e+01, 2.22087510812244e+01, 2.22071033110487e+01, 2.22042379538746e+01, 2.21996071211248e+01, 2.21925424079585e+01, 2.21822570674865e+01, 2.21678535218029e+01, 2.21483363682598e+01, 2.21226308690716e+01, 2.20896067065865e+01, 2.20481065471221e+01, 2.19969786905250e+01, 2.19351128045205e+01, 2.18614774729943e+01, 2.17751580516450e+01, 2.16753931522268e+01, 2.15616079968176e+01, 2.14334429204096e+01, 2.12907754682971e+01, 2.11337348350450e+01, 2.09627078082988e+01, 2.07783358802052e+01, 2.05815037240342e+01, 2.03733197475387e+01, 2.01550898716861e+01, 1.99282859979128e+01, 1.96945107924446e+01, 1.94554604305889e+01, 1.92128868299064e+01, 1.89685606999341e+01, 1.87242364960490e+01, 1.84816201290368e+01, 1.82423400766117e+01, 1.80079223744068e+01, 1.77797698199114e+01, 1.75591455833951e+01, 1.73471612687215e+01, 1.71447693008550e+01, 1.69527593480464e+01, 1.67717583380297e+01, 1.66022335229730e+01, 1.64444980030215e+01, 1.62987181350744e+01, 1.61649223209919e+01, 1.60430107685062e+01, 1.59327659276060e+01, 1.58338634072389e+01, 1.57458832599583e+01, 1.56683215799816e+01, 1.56006023923271e+01, 1.55420898196795e+01, 1.54921005033498e+01, 1.54499162296047e+01, 1.54147966771381e+01, 1.53859921598182e+01, 1.53627561951893e+01, 1.53443576876190e+01, 1.53300924794378e+01, 1.53192939976934e+01, 1.53113427115150e+01, 1.53056741181193e+01, 1.53017849956540e+01, 1.52992376985687e+01, 1.52976623248143e+01, 1.52967566513083e+01, 1.52962838110084e+01, 1.52960677670299e+01, 1.52959867215816e+01, 1.52959646752647e+01, 1.52959614212748e+01, 1.52959611777396e+01, 1.52959569011392e+01, 1.52959279265502e+01, 1.52958214130225e+01, 1.52955374880335e+01, 1.52949161300590e+01, 1.52937261740521e+01, 1.52916568895404e+01, 1.52883125339025e+01, 1.52832102280794e+01, 1.52757814358270e+01, 1.52653772470545e+01, 1.52512775670026e+01, 1.52327041929062e+01, 1.52088376169514e+01, 1.51788372301511e+01, 1.51418644212267e+01, 1.50971078768986e+01, 1.50438102085544e+01, 1.49812948720360e+01, 1.49089922312727e+01, 1.48264635615752e+01, 1.47334218104718e+01, 1.46297480427984e+01, 1.45155026930906e+01, 1.43909310217934e+01, 1.42564625003968e+01, 1.41127042024100e+01, 1.39604286145810e+01, 1.38005565693145e+01, 1.36341362065195e+01, 1.34623189873062e+01, 1.32863338063896e+01, 1.31074602030042e+01, 1.29270015780044e+01, 1.27462592134960e+01, 1.25665077787765e+01, 1.23889728993880e+01, 1.22148112620657e+01, 1.20450936202200e+01, 1.18807909465539e+01, 1.17227638510459e+01, 1.15717552496815e+01, 1.14283861424261e+01, 1.12931542497858e+01, 1.11664351757527e+01, 1.10484857166678e+01, 1.09394489212538e+01, 1.08393605230150e+01, 1.07481564054146e+01, 1.06656808142485e+01, 1.05916950918412e+01, 1.05258867664867e+01, 1.04678788819621e+01, 1.04172394918181e+01, 1.03734912691956e+01, 1.03361211944824e+01, 1.03045902809389e+01, 1.02783432843712e+01, 1.02568183197846e+01, 1.02394562791167e+01, 1.02257099134633e+01, 1.02150524146002e+01, 1.02069853078766e+01, 1.02010454550292e+01, 1.01968109637036e+01, 1.01939058120499e+01, 1.01920030220610e+01, 1.01908262535159e+01, 1.01901497394860e+01, 1.01897965413946e+01, 1.01896351630003e+01, 1.01895746245114e+01, 1.01895581565858e+01, 1.01895557259548e+01, 1.01895555439734e+01, 1.01895523473782e+01, 1.01895306899835e+01, 1.01894510752272e+01, 1.01892388524225e+01, 1.01887744125977e+01, 1.01878849723713e+01, 1.01863382834394e+01, 1.01838385708723e+01, 1.01800249630741e+01, 1.01744726278872e+01, 1.01666967701824e+01, 1.01561595735076e+01, 1.01422800796384e+01, 1.01244468941453e+01, 1.01020334842845e+01, 1.00744157011183e+01, 1.00409910171491e+01, 1.00011988330828e+01, 9.95454108419729e+00, 9.90060228114206e+00, 9.83906806488778e+00, 9.76974135244943e+00, 9.69255520709319e+00, 9.60758168729741e+00, 9.51503611001594e+00, 9.41527639644451e+00, 9.30879743685082e+00, 9.19622069421888e+00, 9.07827954036757e+00, 8.95580105815610e+00, 8.82968522721795e+00, 8.70088252178667e+00, 8.57037097905949e+00, 8.43913374611273e+00, 8.30813799284557e+00, 8.17831590556528e+00, 8.05054827292824e+00, 7.92565096614642e+00, 7.80436441913527e+00, 7.68734604685861e+00, 7.57516541033689e+00, 7.46830184720757e+00, 7.36714423497556e+00, 7.27199253460533e+00, 7.18306076792513e+00, 7.10048110646389e+00, 7.02430878563831e+00, 6.95452760155163e+00, 6.89105579398102e+00, 6.83375216513983e+00, 6.78242232681789e+00, 6.73682500629441e+00, 6.69667837217328e+00, 6.66166636367163e+00, 6.63144502009679e+00, 6.60564881108471e+00, 6.58389696310218e+00, 6.56579976483727e+00, 6.55096481511204e+00, 6.53900315404694e+00, 6.52953519394430e+00, 6.52219634347402e+00, 6.51664219994130e+00, 6.51255317212652e+00, 6.50963839234279e+00, 6.50763878219254e+00, 6.50632915240251e+00, 6.50551924256379e+00, 6.50505364020529e+00, 6.50481055827380e+00, 6.50469949315004e+00, 6.50465782895247e+00, 6.50464649529607e+00, 6.50464482247270e+00, 6.50464469689612e+00, 6.50464248664262e+00, 6.50462751185900e+00, 6.50457246310411e+00, 6.50442572436222e+00, 6.50410459542871e+00, 6.50348961508415e+00, 6.50242022163175e+00, 6.50069196539781e+00, 6.49805546179503e+00, 6.49421724170232e+00, 6.48884261584001e+00, 6.48156061901496e+00, 6.47197103648733e+00, 6.45965343723039e+00, 6.44417804811007e+00, 6.42511820175018e+00, 6.40206398425151e+00, 6.37463660458590e+00, 6.34250291493639e+00, 6.30538944115077e+00, 6.26309524528375e+00, 6.21550294670614e+00, 6.16258728002374e+00, 6.10442066814478e+00, 6.04117543313777e+00, 5.97312244673755e+00, 5.90062622287161e+00, 5.82413666000532e+00, 5.74417783420172e+00, 5.66133440845151e+00, 5.57623634672504e+00, 5.48954269297110e+00, 5.40192519121288e+00, 5.31405248317327e+00, 5.22657552965561e+00, 5.14011477106702e+00, 5.05524938483990e+00, 4.97250882954567e+00, 4.89236670423608e+00, 4.81523681211095e+00, 4.74147121094661e+00, 4.67135996414112e+00, 4.60513227525082e+00, 4.54295869026824e+00, 4.48495407749137e+00, 4.43118113568141e+00, 4.38165422912618e+00, 4.33634339683030e+00, 4.29517842801842e+00, 4.25805293498209e+00, 4.22482838591061e+00, 4.19533808446603e+00, 4.16939109969547e+00, 4.14677615982028e+00, 4.12726552699856e+00, 4.11061886789460e+00, 4.09658712749993e+00, 4.08491640200290e+00, 4.07535179168347e+00, 4.06764119812520e+00, 4.06153901299449e+00, 4.05680962985357e+00, 4.05323069757721e+00, 4.05059602541934e+00, 4.04871804683373e+00, 4.04742975257119e+00, 4.04658601361785e+00, 4.04606423087483e+00, 4.04576427022326e+00, 4.04560766738895e+00, 4.04553611509583e+00, 4.04550927349558e+00, 4.04550197194224e+00, 4.04550089424873e+00, 4.04550081300319e+00, 4.04549937842363e+00, 4.04548965894310e+00, 4.04545392921128e+00, 4.04535868768453e+00, 4.04515025809401e+00, 4.04475110643853e+00, 4.04405702825382e+00, 4.04293534732383e+00, 4.04122425082051e+00, 4.03873336574044e+00, 4.03524565568708e+00, 4.03052068366773e+00, 4.02429924406334e+00, 4.01630931446367e+00, 4.00627321602137e+00, 3.99391580134002e+00, 3.97897341547781e+00, 3.96120330403684e+00, 3.94039307961918e+00, 3.91636981204492e+00, 3.88900828630842e+00, 3.85823798148920e+00, 3.82404836723782e+00, 3.78649219181780e+00, 3.74568654258300e+00, 3.70181158776485e+00, 3.65510704600854e+00, 3.60586656421333e+00, 3.55443030230897e+00, 3.50117611533433e+00, 3.44650978191990e+00, 3.39085475161147e+00, 3.33464187301732e+00, 3.27829952526666e+00, 3.22224451341421e+00, 3.16687401165430e+00, 3.11255875382538e+00, 3.05963758530029e+00, 3.00841340962232e+00, 2.95915049173317e+00, 2.91207302076261e+00, 2.86736479128841e+00, 2.82516983362071e+00, 2.78559381062848e+00, 2.74870599938811e+00, 2.71454168808226e+00, 2.68310483916615e+00, 2.65437089570456e+00, 2.62828963598021e+00, 2.60478800937258e+00, 2.58377291203543e+00, 2.56513388256393e+00, 2.54874571470643e+00, 2.53447099578909e+00, 2.52216258584725e+00, 2.51166605376987e+00, 2.50282208360811e+00, 2.49546885731947e+00, 2.48944441053847e+00, 2.48458894653869e+00, 2.48074708153673e+00, 2.47776998308456e+00, 2.47551735367215e+00, 2.47385920487694e+00, 2.47267736429886e+00, 2.47186665866832e+00, 2.47133572210545e+00, 2.47100738834649e+00, 2.47081863925368e+00, 2.47072009818019e+00, 2.47067507462643e+00, 2.47065818484650e+00, 2.47065359042716e+00, 2.47065291230089e+00, 2.47065286089986e+00, 2.47065194960449e+00, 2.47064577545002e+00, 2.47062307869000e+00, 2.47056257811357e+00, 2.47043017739331e+00, 2.47017662686400e+00, 2.46973574007382e+00, 2.46902325641463e+00, 2.46793643150150e+00, 2.46635442475792e+00, 2.46413953762037e+00, 2.46113933456706e+00, 2.45718965167082e+00, 2.45211846294676e+00, 2.44575053360488e+00, 2.43791274269545e+00, 2.42843990806611e+00, 2.41718089778506e+00, 2.40400476899950e+00, 2.38880664296468e+00, 2.37151300899879e+00, 2.35208615484929e+00, 2.33052744918579e+00, 2.30687925407742e+00, 2.28122531898973e+00, 2.25368959782708e+00, 2.22443352920801e+00, 2.19365191827715e+00, 2.16156764630163e+00, 2.12842550335284e+00, 2.09448548293935e+00, 2.06001589199440e+00, 2.02528661509770e+00, 1.99056283159224e+00, 1.95609942448749e+00, 1.92213624858275e+00, 1.88889435046771e+00, 1.85657316260742e+00, 1.82534863360097e+00, 1.79537221070957e+00, 1.76677056035169e+00, 1.73964589689440e+00, 1.71407678764298e+00, 1.69011930949331e+00, 1.66780844705454e+00, 1.64715964017758e+00, 1.62817040826768e+00, 1.61082199767885e+00, 1.59508101567211e+00, 1.58090102916695e+00, 1.56822411850705e+00, 1.55698238560863e+00, 1.54709942222671e+00, 1.53849174777839e+00, 1.53107022737910e+00, 1.52474147967867e+00, 1.51940928099367e+00, 1.51497596744937e+00, 1.51134383078890e+00, 1.50841649668422e+00, 1.50610026737999e+00, 1.50430540395244e+00, 1.50294731801189e+00, 1.50194763891750e+00, 1.50123512098961e+00, 1.50074635612687e+00, 1.50042626077786e+00, 1.50022831226703e+00, 1.50011451769207e+00, 1.50005510845619e+00, 1.50002796429378e+00, 1.50001778164736e+00, 1.50001501172686e+00, 1.50001460289263e+00, 1.50001457170691e+00, 1.50001401621745e+00, 1.50001025269815e+00, 1.49999641765622e+00, 1.49995953892279e+00, 1.49987883279113e+00, 1.49972427884242e+00, 1.49945553309944e+00, 1.49902123529419e+00, 1.49835876020238e+00, 1.49739445556699e+00, 1.49604439931883e+00, 1.49421569575729e+00, 1.49180831336497e+00, 1.48871744560159e+00, 1.48483635049161e+00, 1.48005959592474e+00, 1.47428660698438e+00, 1.46742538175407e+00, 1.45939621604468e+00, 1.45013525876580e+00, 1.43959771155900e+00, 1.42776049143822e+00, 1.41462419493885e+00, 1.40021423637401e+00, 1.38458107908217e+00, 1.36779953311925e+00, 1.34996715051891e+00, 1.33120180428497e+00, 1.31163858430293e+00, 1.29142617813152e+00, 1.27072292469832e+00, 1.24969273378744e+00, 1.22850105522068e+00, 1.20731106145477e+00, 1.18628017929329e+00, 1.16555707390695e+00, 1.14527915432269e+00, 1.12557063628255e+00, 1.10654116758045e+00, 1.08828499388662e+00, 1.07088062059570e+00, 1.05439090911463e+00, 1.03886353473736e+00, 1.02433172800838e+00, 1.01081522199230e+00, 9.98321333416527e-01, 9.86846115100192e-01, 9.76375529041533e-01, 9.66886602566959e-01, 9.58348542743451e-01, 9.50723795784214e-01, 9.43969047722888e-01, 9.38036169809649e-01, 9.32873116781309e-01, 9.28424788471659e-01, 9.24633865395833e-01, 9.21441627282675e-01, 9.18788760419474e-01, 9.16616155517119e-01, 9.14865693031160e-01, 9.13481007926066e-01, 9.12408221193445e-01, 9.11596621457406e-01, 9.10999277112168e-01, 9.10573557950912e-01, 9.10281545371248e-01, 9.10090312059268e-01, 9.09972055491608e-01, 9.09904074440519e-01, 9.09868583578103e-01, 9.09852367823376e-01, 9.09846284784649e-01, 9.09844630055417e-01, 9.09844385821032e-01, 9.09844367059463e-01, 9.09844031149163e-01, 9.09841755310168e-01, 9.09833389125794e-01, 9.09811088261236e-01, 9.09762284944520e-01, 9.09668826808371e-01, 9.09506321267983e-01, 9.09243719963641e-01, 9.08843174576319e-01, 9.08260191099828e-01, 9.07444103872393e-01, 9.06338882745473e-01, 9.04884276191979e-01, 9.03017279634099e-01, 9.00673901855654e-01, 8.97791183575976e-01, 8.94309402154889e-01, 8.90174376575331e-01, 8.85339769362872e-01, 8.79769269268323e-01, 8.73438532664906e-01, 8.66336764615696e-01, 8.58467833643346e-01, 8.49850837496248e-01, 8.40520069498169e-01, 8.30524373920966e-01, 8.19925920631277e-01, 8.08798469693940e-01, 7.97225231180988e-01, 7.85296450187900e-01, 7.73106859251769e-01, 7.60753138978075e-01, 7.48331513611276e-01, 7.35935584196682e-01, 7.23654471767958e-01, 7.11571311003880e-01, 6.99762105005945e-01, 6.88294927213605e-01, 6.77229438589691e-01, 6.66616677341040e-01, 6.56499073804298e-01, 6.46910643348861e-01, 6.37877313673343e-01, 6.29417348299214e-01, 6.21541834302434e-01, 6.14255208641879e-01, 6.07555803406754e-01, 6.01436395696140e-01, 5.95884752565951e-01, 5.90884165498522e-01, 5.86413972144044e-01, 5.82450065640685e-01, 5.78965393619640e-01, 5.75930450037021e-01, 5.73313763250610e-01, 5.71082383308337e-01, 5.69202370299180e-01, 5.67639283936714e-01, 5.66358672435214e-01, 5.65326556374060e-01, 5.64509900826579e-01, 5.63877066776271e-01, 5.63398230976532e-01, 5.63045762137904e-01, 5.62794540814977e-01, 5.62622210728321e-01, 5.62509350535223e-01, 5.62439557215384e-01, 5.62399435148876e-01, 5.62378488441439e-01, 5.62368917860965e-01, 5.62365327628982e-01, 5.62364351001023e-01, 5.62364206852877e-01, 5.62364195695196e-01, 5.62363994826413e-01, 5.62362633912173e-01, 5.62357631067812e-01, 5.62344295477970e-01, 5.62315111659288e-01, 5.62259224195827e-01, 5.62162044941630e-01, 5.62005002967979e-01, 5.61765455461270e-01, 5.61416775118502e-01, 5.60928626045552e-01, 5.60267435414971e-01, 5.59397061916187e-01, 5.58279654179969e-01, 5.56876682929726e-01, 5.55150119925329e-01, 5.53063725443764e-01, 5.50584395021033e-01, 5.47683506635053e-01, 5.44338202736853e-01, 5.40532538770805e-01, 5.36258432022237e-01, 5.31516352268497e-01, 5.26315708630648e-01, 5.20674904376961e-01, 5.14621051743632e-01, 5.08189360196411e-01, 5.01422231874599e-01, 4.94368115285823e-01, 4.87080181103963e-01, 4.79614891204420e-01, 4.72030533559987e-01, 4.64385791649534e-01, 4.56738408470414e-01, 4.49143993301988e-01, 4.41655005437966e-01, 4.34319934596360e-01, 4.27182683902174e-01, 4.20282149261055e-01, 4.13651979332286e-01, 4.07320493551936e-01, 4.01310731804486e-01, 3.95640608154423e-01, 3.90323142084512e-01, 3.85366743378043e-01, 3.80775530534336e-01, 3.76549666862815e-01, 3.72685702694596e-01, 3.69176916122655e-01, 3.66013648094437e-01, 3.63183630398780e-01, 3.60672307065300e-01, 3.58463150944104e-01, 3.56537977812167e-01, 3.54877260336366e-01, 3.53460443704790e-01, 3.52266263806740e-01, 3.51273067606328e-01, 3.50459133913262e-01, 3.49802991227008e-01, 3.49283727833139e-01, 3.48881287989769e-01, 3.48576746979219e-01, 3.48352557125423e-01, 3.48192756682906e-01, 3.48083133845023e-01, 3.48011339015841e-01, 3.47966939911345e-01, 3.47941415928133e-01, 3.47928090425711e-01, 3.47922001965969e-01, 3.47919717986230e-01, 3.47919096689481e-01, 3.47919004987403e-01, 3.47918997831755e-01, 3.47918868266575e-01, 3.47917990444245e-01, 3.47914763487921e-01, 3.47906161695249e-01, 3.47887337329240e-01, 3.47851288128524e-01, 3.47788603622817e-01, 3.47687303173856e-01, 3.47532777262202e-01, 3.47307841769137e-01, 3.46992912677521e-01, 3.46566305545593e-01, 3.46004660125800e-01, 3.45283485517982e-01, 3.44377815294052e-01, 3.43262955309058e-01, 3.41915299810552e-01, 3.40313184559832e-01, 3.38437739721499e-01, 3.36273701086313e-01, 3.33810136558540e-01, 3.31041046387242e-01, 3.27965800681966e-01, 3.24589386240362e-01, 3.20922446079521e-01, 3.16981108319467e-01, 3.12786614894648e-01, 3.08364773502433e-01, 3.03745266847201e-01, 2.98960860508743e-01, 2.94046554027219e-01, 2.89038718980038e-01, 2.83974263388345e-01, 2.78889854637953e-01, 2.73821224386111e-01, 2.68802569870076e-01, 2.63866057691615e-01, 2.59041429275939e-01, 2.54355702169491e-01, 2.49832958172334e-01, 2.45494207775678e-01, 2.41357320127122e-01, 2.37437008380468e-01, 2.33744861422423e-01, 2.30289414293082e-01, 2.27076250904219e-01, 2.24108133786686e-01, 2.21385156538560e-01, 2.18904915420993e-01, 2.16662697209713e-01, 2.14651680976502e-01, 2.12863151950003e-01, 2.11286725968378e-01, 2.09910583257368e-01, 2.08721710320091e-01, 2.07706148590233e-01, 2.06849248174848e-01, 2.06135924508430e-01, 2.05550915094165e-01, 2.05079032778173e-01, 2.04705411259553e-01, 2.04415737872367e-01, 2.04196468168576e-01, 2.04035016567374e-01, 2.03919917374918e-01, 2.03840950863338e-01, 2.03789229837753e-01, 2.03757243192513e-01, 2.03738854312462e-01, 2.03729253736626e-01, 2.03724867171729e-01, 2.03723221622624e-01, 2.03722773993690e-01, 2.03722707924570e-01, 2.03722702725187e-01, 2.03722608020328e-01, 2.03721966380996e-01, 2.03719607651621e-01, 2.03713320181208e-01, 2.03699560389126e-01, 2.03673209442684e-01, 2.03627387120619e-01, 2.03553331958275e-01, 2.03440355614987e-01, 2.03275878002913e-01, 2.03045547967534e-01, 2.02733452120057e-01, 2.02322411702070e-01, 2.01794364079589e-01, 2.01130821645393e-01, 2.00313396685061e-01, 1.99324376360764e-01, 1.98147327725735e-01, 1.96767709011411e-01, 1.95173460781518e-01, 1.93355549342570e-01, 1.91308435364874e-01, 1.89030443156753e-01, 1.86524010386293e-01, 1.83795803964102e-01, 1.80856694788019e-01, 1.77721591484769e-01, 1.74409140512424e-01, 1.70941306460634e-01, 1.67342851699715e-01, 1.63640738457573e-01, 1.59863478856423e-01, 1.56040459423721e-01, 1.52201266183042e-01, 1.48375034805966e-01, 1.44589847707422e-01, 1.40872196639691e-01, 1.37246525454813e-01, 1.33734863334595e-01, 1.30356553980320e-01, 1.27128081266306e-01, 1.24062987069806e-01, 1.21171872972226e-01, 1.18462474562904e-01, 1.15939795326298e-01, 1.13606286373440e-01, 1.11462058422717e-01, 1.09505113349428e-01, 1.07731584217589e-01, 1.06135974856045e-01, 1.04711392477071e-01, 1.03449769280218e-01, 1.02342071091414e-01, 1.01378492820700e-01, 1.00548641690978e-01, 9.98417099364701e-02, 9.92466388846509e-02, 9.87522761995052e-02, 9.83475274860695e-02, 9.80215027011866e-02, 9.77636568211864e-02, 9.75639232391785e-02, 9.74128374647449e-02, 9.73016479439792e-02, 9.72224103990982e-02, 9.71680619391009e-02, 9.71324714273578e-02, 9.71104631368718e-02, 9.70978116048542e-02, 9.70912066698162e-02, 9.70881888903534e-02, 9.70870568290394e-02, 9.70867488814520e-02, 9.70867034288290e-02, 9.70866998199540e-02, 9.70866336888889e-02, 9.70861856386023e-02, 9.70845385706269e-02, 9.70801481708037e-02, 9.70705402890388e-02, 9.70521415853085e-02, 9.70201509000646e-02, 9.69684583792656e-02, 9.68896188238375e-02, 9.67748850078465e-02, 9.66143056319311e-02, 9.63968910771769e-02, 9.61108480305342e-02, 9.57438813796756e-02, 9.52835584045175e-02, 9.47177264844470e-02, 9.40349712986119e-02, 9.32250982925353e-02, 9.22796163143573e-02, 9.11921992326086e-02, 8.99590995451405e-02, 8.85794877882375e-02, 8.70556933602131e-02, 8.53933261457342e-02, 8.36012640999115e-02, 8.16914996031386e-02, 7.96788460828566e-02, 7.75805161561749e-02, 7.54155917481114e-02, 7.32044150031754e-02, 7.09679347587553e-02, 6.87270463208307e-02, 6.65019618922855e-02, 6.43116458618351e-02, 6.21733429419038e-02, 6.01022198411130e-02, 5.81111317084151e-02, 5.61900280273306e-02, 5.43907631043652e-02, 5.26952881535436e-02, 5.11072517218025e-02, 4.96284039342214e-02, 4.82588553182381e-02, 4.69973430288206e-02, 4.58414875196951e-02, 4.47880275317516e-02, 4.38330260613924e-02, 4.29720441991890e-02, 4.22002830829342e-02, 4.15126965865937e-02, 4.09040788204464e-02, 4.03691311977212e-02, 3.99025139167009e-02, 3.94988863982633e-02, 3.91529406607845e-02, 3.88594309239352e-02, 3.86132019880705e-02, 3.84092181847961e-02, 3.82425939659470e-02, 3.81086265113169e-02, 3.80028301015329e-02, 3.79209714416930e-02, 3.78591046492716e-02, 3.78136042594690e-02, 3.77811943748014e-02, 3.77589720078362e-02, 3.77444227484176e-02, 3.77354271248578e-02, 3.77302564056095e-02, 3.77275570768205e-02, 3.77263237888457e-02, 3.77258611492834e-02, 3.77257353009952e-02, 3.77257167260865e-02, 3.77257152378832e-02, 3.77256877946905e-02, 3.77255018634912e-02, 3.77248183687158e-02, 3.77229964839501e-02, 3.77190096447180e-02, 3.77113755623384e-02, 3.76981035567638e-02, 3.76766624699769e-02, 3.76439723202822e-02, 3.75964226445120e-02, 3.75299200471915e-02, 3.74399667380629e-02, 3.73217706940210e-02, 3.71703864749833e-02, 3.69808836562413e-02, 3.67485374002996e-02, 3.64690330680730e-02, 3.61386742514895e-02, 3.57545815715525e-02, 3.53148684381210e-02, 3.48187800857573e-02, 3.42667838468884e-02, 3.36606018542653e-02, 3.30031819702299e-02, 3.22986082273644e-02, 3.15519577023338e-02, 3.07691156855725e-02, 2.99565644525188e-02, 2.91211623317480e-02, 2.82699289253259e-02, 2.74098495381878e-02, 2.65477077716412e-02, 2.56899506915147e-02, 2.48425868532394e-02, 2.40111143757957e-02, 2.32004744970880e-02, 2.24150255430970e-02, 2.16585326816841e-02, 2.09341697732413e-02, 2.02445306728265e-02, 1.95916481948331e-02, 1.89770194977115e-02, 1.84016368790765e-02, 1.78660229833270e-02, 1.73702693260935e-02, 1.69140769433653e-02, 1.64967979470276e-02, 1.61174768448262e-02, 1.57748906593582e-02, 1.54675871291374e-02, 1.51939205597103e-02, 1.49520851762617e-02, 1.47401460773864e-02, 1.45560680825838e-02, 1.43977428855351e-02, 1.42630149718173e-02, 1.41497067305352e-02, 1.40556430989562e-02, 1.39786759357747e-02, 1.39167081412097e-02, 1.38677173483757e-02, 1.38297788177937e-02, 1.38010869954996e-02, 1.37799750627892e-02, 1.37649317224344e-02, 1.37546144485867e-02, 1.37478584728967e-02, 1.37436808916505e-02, 1.37412794473590e-02, 1.37400257495415e-02, 1.37394529412234e-02, 1.37392380637067e-02, 1.37391796120738e-02, 1.37391709847350e-02, 1.37391702870246e-02, 1.37391573394012e-02, 1.37390696175288e-02, 1.37387471452115e-02, 1.37378875720628e-02, 1.37360065169787e-02, 1.37324044538758e-02, 1.37261416322048e-02, 1.37160224586381e-02, 1.37005906766205e-02, 1.36781365490179e-02, 1.36467170413965e-02, 1.36041896883654e-02, 1.35482603737245e-02, 1.34765446501177e-02, 1.33866414658286e-02, 1.32762172824101e-02, 1.31430976125984e-02, 1.29853620657343e-02, 1.28014381671875e-02, 1.25901886335384e-02, 1.23509865495492e-02, 1.20837730924279e-02, 1.17890931252731e-02, 1.14681051265677e-02, 1.11225634616403e-02, 1.07547728184869e-02, 1.03675165586736e-02, 9.96396260043753e-03, 9.54755207838787e-03, 9.12187725982960e-03, 8.69055592518464e-03, 8.25710957688841e-03, 7.82485241583686e-03, 7.39679707368168e-03, 6.97558171631840e-03, 6.56342147881617e-03, 6.16208541111973e-03, 5.77289835410966e-03, 5.39676555891333e-03, 5.03421651998792e-03, 4.68546349321741e-03, 4.35046956459928e-03, 4.02902092757529e-03, 3.72079819744897e-03, 3.42544209117079e-03, 3.14260958703017e-03, 2.87201769454271e-03, 2.61347313653459e-03, 2.36688747698111e-03, 2.13227840131935e-03, 1.90975886430067e-03, 1.69951656677370e-03, 1.50178666251577e-03, 1.31682072355943e-03, 1.14485483673832e-03, 9.86079325613787e-04, 8.40612054901158e-04, 7.08476661401392e-04, 5.89586444078805e-04, 4.83734099844924e-04, 3.90587058589631e-04, 3.09687871947517e-04, 2.40458945725663e-04, 1.82278668376057e-04, 1.34206677842030e-04, 9.54491980991179e-05, 6.50579526251249e-05, 4.20288668721014e-05, 2.53201413187298e-05, 1.38715960531499e-05, 6.62546479190951e-06, 2.54920166753436e-06, 6.61624568465413e-07, -1.42108547152020e-14],
[ 9.20000000000000e+01, 9.19999998890293e+01, 9.19999964575968e+01, 9.19999732090961e+01, 9.19998877445428e+01, 9.19996599227710e+01, 9.19991613167965e+01, 9.19982063362831e+01, 9.19965453342415e+01, 9.19938599539705e+01, 9.19897609022565e+01, 9.19837882601127e+01, 9.19754143665800e+01, 9.19640492383751e+01, 9.19490484224469e+01, 9.19297231231093e+01, 9.19053524023355e+01, 9.18751972209520e+01, 9.18385160672101e+01, 9.17945819022824e+01, 9.17427001323241e+01, 9.16822272858968e+01, 9.16125900272424e+01, 9.15333040673811e+01, 9.14439924495540e+01, 9.13444025935060e+01, 9.12344214015439e+01, 9.11140876794495e+01, 9.09836011285116e+01, 9.08433272374927e+01, 9.06937975519203e+01, 9.05357050166484e+01, 9.03698943571328e+01, 9.01973477560329e+01, 9.00191663603068e+01, 8.98365483870563e+01, 8.96507647586952e+01, 8.94631332760066e+01, 8.92749923310027e+01, 8.90876750818324e+01, 8.89024848792226e+01, 8.87206725717535e+01, 8.85434161484093e+01, 8.83718030197289e+01, 8.82068151055883e+01, 8.80493167936561e+01, 8.79000457576794e+01, 8.77596065747892e+01, 8.76284670496901e+01, 8.75069571341778e+01, 8.73952703169289e+01, 8.72934673463149e+01, 8.72014821351057e+01, 8.71191296787886e+01, 8.70461157984847e+01, 8.69820484955639e+01, 8.69264506791062e+01, 8.68787740006990e+01, 8.68384135052902e+01, 8.68047227835981e+01, 8.67770292926410e+01, 8.67546494979620e+01, 8.67369034856841e+01, 8.67231286960197e+01, 8.67126924433849e+01, 8.67050029125033e+01, 8.66995183550267e+01, 8.66957542568242e+01, 8.66932883012068e+01, 8.66917630163221e+01, 8.66908860636454e+01, 8.66904281963631e+01, 8.66902189886923e+01, 8.66901405069721e+01, 8.66901191579752e+01, 8.66901160069056e+01, 8.66901155935517e+01, 8.66901059629026e+01, 8.66900407138247e+01, 8.66898008511107e+01, 8.66891614636348e+01, 8.66877621719367e+01, 8.66850823308665e+01, 8.66804219706439e+01, 8.66728893367523e+01, 8.66613957467307e+01, 8.66446583175699e+01, 8.66212109272469e+01, 8.65894235538134e+01, 8.65475298822062e+01, 8.64936627816782e+01, 8.64258969382747e+01, 8.63422975850663e+01, 8.62409739222975e+01, 8.61201354823100e+01, 8.59781494003341e+01, 8.58135963396621e+01, 8.56253227296873e+01, 8.54124870453450e+01, 8.51745981085949e+01, 8.49115438196666e+01, 8.46236092813036e+01, 8.43114838773743e+01, 8.39762573995652e+01, 8.36194056859354e+01, 8.32427664022645e+01, 8.28485056117885e+01, 8.24390757848152e+01, 8.20171660836027e+01, 8.15856462491666e+01, 8.11475061761635e+01, 8.07057940348538e+01, 8.02635562008468e+01, 7.98237819806012e+01, 7.93893551642263e+01, 7.89630131284823e+01, 7.85473130451538e+01, 7.81446040979940e+01, 7.77570045561124e+01, 7.73863829087299e+01, 7.70343427504634e+01, 7.67022114913640e+01, 7.63910331534396e+01, 7.61015655174979e+01, 7.58342817683155e+01, 7.55893766256630e+01, 7.53667767961073e+01, 7.51661554616223e+01, 7.49869504406669e+01, 7.48283856081250e+01, 7.46894951304654e+01, 7.45691500503601e+01, 7.44660867325642e+01, 7.43789366557607e+01, 7.43062570025782e+01, 7.42465614642907e+01, 7.41983506420595e+01, 7.41601413984279e+01, 7.41304944969813e+01, 7.41080398701749e+01, 7.40914988798808e+01, 7.40797029853184e+01, 7.40716083099555e+01, 7.40663057018522e+01, 7.40630260078655e+01, 7.40611404263004e+01, 7.40601559586562e+01, 7.40597061417469e+01, 7.40595373990032e+01, 7.40594914968013e+01, 7.40594847217294e+01, 7.40594841236195e+01, 7.40594724039507e+01, 7.40593930014571e+01, 7.40591011092957e+01, 7.40583230304593e+01, 7.40566202163335e+01, 7.40533590943564e+01, 7.40476878844366e+01, 7.40385214559198e+01, 7.40245351055670e+01, 7.40041679388268e+01, 7.39756363067536e+01, 7.39369574841577e+01, 7.38859834673802e+01, 7.38204444222555e+01, 7.37380009293518e+01, 7.36363037660643e+01, 7.35130595524032e+01, 7.33661001949032e+01, 7.31934537213694e+01, 7.29934138403094e+01, 7.27646054130535e+01, 7.25060430183357e+01, 7.22171799349589e+01, 7.18979451753873e+01, 7.15487666695696e+01, 7.11705793118936e+01, 7.07648173203891e+01, 7.03333911749128e+01, 6.98786502383901e+01, 6.94033329410824e+01, 6.89105070314296e+01, 6.84035027868222e+01, 6.78858421852950e+01, 6.73611668680670e+01, 6.68331673328851e+01, 6.63055152906655e+01, 6.57818006057053e+01, 6.52654738120326e+01, 6.47597948960163e+01, 6.42677888451190e+01, 6.37922083341536e+01, 6.33355037948679e+01, 6.28998009534609e+01, 6.24868857195956e+01, 6.20981960962342e+01, 6.17348205923674e+01, 6.13975024942335e+01, 6.10866492994514e+01, 6.08023466358328e+01, 6.05443760510215e+01, 6.03122361440884e+01, 6.01051665924784e+01, 5.99221746911879e+01, 5.97620640575446e+01, 5.96234651625262e+01, 5.95048673305073e+01, 5.94046518084353e+01, 5.93211254489289e+01, 5.92525544867029e+01, 5.91971978216638e+01, 5.91533391628964e+01, 5.91193173433983e+01, 5.90935540930975e+01, 5.90745785635474e+01, 5.90610479361034e+01, 5.90517635184718e+01, 5.90456818417476e+01, 5.90419204081829e+01, 5.90397579031155e+01, 5.90386288648355e+01, 5.90381129943243e+01, 5.90379194729040e+01, 5.90378668303305e+01, 5.90378590603949e+01, 5.90378584371500e+01, 5.90378469351196e+01, 5.90377690071676e+01, 5.90374825357174e+01, 5.90367189075830e+01, 5.90350477239258e+01, 5.90318471990058e+01, 5.90262814333580e+01, 5.90172856066214e+01, 5.90035599721087e+01, 5.89835733450274e+01, 5.89555765530231e+01, 5.89176260550743e+01, 5.88676176279282e+01, 5.88033296669929e+01, 5.87224752552395e+01, 5.86227617308334e+01, 5.85019560517386e+01, 5.83579538414468e+01, 5.81888496390915e+01, 5.79930056082855e+01, 5.77691158204855e+01, 5.75162632530758e+01, 5.72339668501863e+01, 5.69222163880332e+01, 5.65814934465433e+01, 5.62127774722268e+01, 5.58175366617276e+01, 5.53977041304205e+01, 5.49556404907715e+01, 5.44940845062368e+01, 5.40160938921785e+01, 5.35249786154208e+01, 5.30242292204892e+01, 5.25174427992007e+01, 5.20082492188196e+01, 5.15002401128147e+01, 5.09969028939923e+01, 5.05015616642447e+01, 5.00173263889859e+01, 4.95470511283484e+01, 4.90933015410244e+01, 4.86583313703187e+01, 4.82440672383197e+01, 4.78521008380485e+01, 4.74836875208746e+01, 4.71397503008597e+01, 4.68208884007783e+01, 4.65273896070641e+01, 4.62592458501224e+01, 4.60161715598072e+01, 4.57976244510459e+01, 4.56028284672850e+01, 4.54307986503617e+01, 4.52803677177614e+01, 4.51502141159975e+01, 4.50388912860755e+01, 4.49448578277641e+01, 4.48665081881262e+01, 4.48022034316152e+01, 4.47503015800684e+01, 4.47091869479554e+01, 4.46772978485433e+01, 4.46531520173133e+01, 4.46353690962640e+01, 4.46226895513282e+01, 4.46139894575172e+01, 4.46082906824711e+01, 4.46047661259560e+01, 4.46027398250223e+01, 4.46016819043491e+01, 4.46011985296066e+01, 4.46010171987212e+01, 4.46009678722924e+01, 4.46009605918137e+01, 4.46009600302732e+01, 4.46009499468069e+01, 4.46008816298403e+01, 4.46006304894520e+01, 4.45999610414449e+01, 4.45984959711686e+01, 4.45956901875170e+01, 4.45908109099313e+01, 4.45829247100971e+01, 4.45708922890693e+01, 4.45533716055761e+01, 4.45288297764659e+01, 4.44955639395461e+01, 4.44517309977541e+01, 4.43953858503283e+01, 4.43245273653099e+01, 4.42371509690092e+01, 4.41313063404698e+01, 4.40051583285208e+01, 4.38570488879409e+01, 4.36855575949832e+01, 4.34895581852525e+01, 4.32682685863710e+01, 4.30212921095092e+01, 4.27486478162465e+01, 4.24507885694605e+01, 4.21286058695741e+01, 4.17834212178220e+01, 4.14169643795262e+01, 4.10313394935167e+01, 4.06289804574390e+01, 4.02125974039277e+01, 3.97851163788999e+01, 3.93496145547595e+01, 3.89092534596648e+01, 3.84672127564884e+01, 3.80266270184422e+01, 3.75905276808524e+01, 3.71617918910809e+01, 3.67430993759993e+01, 3.63368977936313e+01, 3.59453764463072e+01, 3.55704477973760e+01, 3.52137359894847e+01, 3.48765714903666e+01, 3.45599910362875e+01, 3.42647421408313e+01, 3.39912915397404e+01, 3.37398370271008e+01, 3.35103222005664e+01, 3.33024536808571e+01, 3.31157204126303e+01, 3.29494146955579e+01, 3.28026546362410e+01, 3.26744077494351e+01, 3.25635154649878e+01, 3.24687183093302e+01, 3.23886815238787e+01, 3.23220208568029e+01, 3.22673282220233e+01, 3.22231968654670e+01, 3.21882456210441e+01, 3.21611417860920e+01, 3.21406221068372e+01, 3.21255113465191e+01, 3.21147379182190e+01, 3.21073461047441e+01, 3.21025044599224e+01, 3.20995100872285e+01, 3.20977886179616e+01, 3.20968898551953e+01, 3.20964792029118e+01, 3.20963251530383e+01, 3.20962832477352e+01, 3.20962770626002e+01, 3.20962765942804e+01, 3.20962682980429e+01, 3.20962120898215e+01, 3.20960054625850e+01, 3.20954546712433e+01, 3.20942492829776e+01, 3.20919408398473e+01, 3.20879264992427e+01, 3.20814384224996e+01, 3.20715395709026e+01, 3.20571264372666e+01, 3.20369390847807e+01, 3.20095786770910e+01, 3.19735324602630e+01, 3.19272058967905e+01, 3.18689613571865e+01, 3.17971624545757e+01, 3.17102227770557e+01, 3.16066574523991e+01, 3.14851356955797e+01, 3.13445322695981e+01, 3.11839756611084e+01, 3.10028907566768e+01, 3.08010339172942e+01, 3.05785185916260e+01, 3.03358299747982e+01, 3.00738276915297e+01, 2.97937360347068e+01, 2.94971218929278e+01, 2.91858611202846e+01, 2.88620947037508e+01, 2.85281766303054e+01, 2.81866158064696e+01, 2.78400146950778e+01, 2.74910074695562e+01, 2.71422004186942e+01, 2.67961170599085e+01, 2.64551499585457e+01, 2.61215206543573e+01, 2.57972484338617e+01, 2.54841280368810e+01, 2.51837158186771e+01, 2.48973234585071e+01, 2.46260180375579e+01, 2.43706272035819e+01, 2.41317481733516e+01, 2.39097594606406e+01, 2.37048344153493e+01, 2.35169558801629e+01, 2.33459314842165e+01, 2.31914092781317e+01, 2.30528935605632e+01, 2.29297608498022e+01, 2.28212760169692e+01, 2.27266086247013e+01, 2.26448495129227e+01, 2.25750276473130e+01, 2.25161272021376e+01, 2.24671047925099e+01, 2.24269067070968e+01, 2.23944859259800e+01, 2.23688186452155e+01, 2.23489199750573e+01, 2.23338584381273e+01, 2.23227688717758e+01, 2.23148633392305e+01, 2.23094396791435e+01, 2.23058873732362e+01, 2.23036904852870e+01, 2.23024275180701e+01, 2.23017681426721e+01, 2.23014668702754e+01, 2.23013538528682e+01, 2.23013231094186e+01, 2.23013185717508e+01, 2.23013182311067e+01, 2.23013122353839e+01, 2.23012716134966e+01, 2.23011222833976e+01, 2.23007242261441e+01, 2.22998530975453e+01, 2.22981848202429e+01, 2.22952837914630e+01, 2.22905952782850e+01, 2.22834424975752e+01, 2.22730287874036e+01, 2.22584451650214e+01, 2.22386834272106e+01, 2.22126547769508e+01, 2.21792137534174e+01, 2.21371870021809e+01, 2.20854061566006e+01, 2.20227438238681e+01, 2.19481514005652e+01, 2.18606972090615e+01, 2.17596032767819e+01, 2.16442790039934e+01, 2.15143500059625e+01, 2.13696805858800e+01, 2.12103885954969e+01, 2.10368518537325e+01, 2.08497057854127e+01, 2.06498324647418e+01, 2.04383417457115e+01, 2.02165455807559e+01, 1.99859269276210e+01, 1.97481048008193e+01, 1.95047970409434e+01, 1.92577822782881e+01, 1.90088623978568e+01, 1.87598266153084e+01, 1.85124180826002e+01, 1.82683037741390e+01, 1.80290482548390e+01, 1.77960917823555e+01, 1.75707330270692e+01, 1.73541164963400e+01, 1.71472245334367e+01, 1.69508735519652e+01, 1.67657139958394e+01, 1.65922334093704e+01, 1.64307619733252e+01, 1.62814799046113e+01, 1.61444262098215e+01, 1.60195084006627e+01, 1.59065128984586e+01, 1.58051159581077e+01, 1.57148950196873e+01, 1.56353404456430e+01, 1.55658676250448e+01, 1.55058294278001e+01, 1.54545289757354e+01, 1.54112326686472e+01, 1.53751833659309e+01, 1.53456135820891e+01, 1.53217585110243e+01, 1.53028686533062e+01, 1.52882217863375e+01, 1.52771339931461e+01, 1.52689694546294e+01, 1.52631487149604e+01, 1.52591551519788e+01, 1.52565394239000e+01, 1.52549217193240e+01, 1.52539917066993e+01, 1.52535061582866e+01, 1.52532843076610e+01, 1.52532010839057e+01, 1.52531784450333e+01, 1.52531751035834e+01, 1.52531748535025e+01, 1.52531704619565e+01, 1.52531407085928e+01, 1.52530313321885e+01, 1.52527397756816e+01, 1.52521017156057e+01, 1.52508797703510e+01, 1.52487548477389e+01, 1.52453205432297e+01, 1.52400809424878e+01, 1.52324521133934e+01, 1.52217674894251e+01, 1.52072872449324e+01, 1.51882116401644e+01, 1.51636981687293e+01, 1.51328821739097e+01, 1.50949004180899e+01, 1.50489169007948e+01, 1.49941500390570e+01, 1.49299001660552e+01, 1.48555761892568e+01, 1.47707201963210e+01, 1.46750288212190e+01, 1.45683702933323e+01, 1.44507962883163e+01, 1.43225479696299e+01, 1.41840559309675e+01, 1.40359340905654e+01, 1.38789679126142e+01, 1.37140976055068e+01, 1.35423971475604e+01, 1.33650501086840e+01, 1.31833232769842e+01, 1.29985390799015e+01, 1.28120477320678e+01, 1.26251999655177e+01, 1.24393211126865e+01, 1.22556872200404e+01, 1.20755037654133e+01, 1.18998874295498e+01, 1.17298512304502e+01, 1.15662931728290e+01, 1.14099884052694e+01, 1.12615847286270e+01, 1.11216011746149e+01, 1.09904292834554e+01, 1.08683366588143e+01, 1.07554723661612e+01, 1.06518737617870e+01, 1.05574743855460e+01, 1.04721126113173e+01, 1.03955408158279e+01, 1.03274348907692e+01, 1.02674039788770e+01, 1.02150003576467e+01, 1.01697294223963e+01, 1.01310597329286e+01, 1.00984330859831e+01, 1.00712745610514e+01, 1.00490024629099e+01, 1.00310380539985e+01, 1.00168149375510e+01, 1.00057879222876e+01, 9.99744117547468e+00, 9.99129545670785e+00, 9.98691422256537e+00, 9.98390840394524e+00, 9.98193968385111e+00, 9.98072214276696e+00, 9.98002218948484e+00, 9.97965675425525e+00, 9.97948978467164e+00, 9.97942714874438e+00, 9.97941011026555e+00, 9.97940759542220e+00, 9.97940740713578e+00, 9.97940409979034e+00, 9.97938169204681e+00, 9.97929931893653e+00, 9.97907974343019e+00, 9.97859921268905e+00, 9.97767895754937e+00, 9.97607868472031e+00, 9.97349237572058e+00, 9.96954666497265e+00, 9.96380201954477e+00, 9.95575688193561e+00, 9.94485486203882e+00, 9.93049497258535e+00, 9.91204479282808e+00, 9.88885631895828e+00, 9.86028412031762e+00, 9.82570527446784e+00, 9.78454041111299e+00, 9.73627506671320e+00, 9.68048045185229e+00, 9.61683267561615e+00, 9.54512946741188e+00, 9.46530349544334e+00, 9.37743150604857e+00, 9.28173869658760e+00, 9.17859797690959e+00, 9.06852405429203e+00, 8.95216257224263e+00, 8.83027481933837e+00, 8.70371877439503e+00, 8.57342744546271e+00, 8.44038557514349e+00, 8.30560581480789e+00, 8.17010541667230e+00, 8.03488436630296e+00, 7.90090569760481e+00, 7.76907852093943e+00, 7.64024407661577e+00, 7.51516492167907e+00, 7.39451718336075e+00, 7.27888567684397e+00, 7.16876159092469e+00, 7.06454239043502e+00, 6.96653356341723e+00, 6.87495184695470e+00, 6.78992959111905e+00, 6.71151995919746e+00, 6.63970270877033e+00, 6.57439034786158e+00, 6.51543450967715e+00, 6.46263243542684e+00, 6.41573349498465e+00, 6.37444570778128e+00, 6.33844224997719e+00, 6.30736794786542e+00, 6.28084576146436e+00, 6.25848325692246e+00, 6.23987905284767e+00, 6.22462920579142e+00, 6.21233347614277e+00, 6.20260139028927e+00, 6.19505799090383e+00, 6.18934914743276e+00, 6.18514628580222e+00, 6.18215039204274e+00, 6.18009515023411e+00, 6.17874909130529e+00, 6.17791665525617e+00, 6.17743810386878e+00, 6.17718826175584e+00, 6.17707410793581e+00, 6.17703128507561e+00, 6.17701963623625e+00, 6.17701791689254e+00, 6.17701778782373e+00, 6.17701551610432e+00, 6.17700012488064e+00, 6.17694354525739e+00, 6.17679272585488e+00, 6.17646266686249e+00, 6.17583058579346e+00, 6.17473145843925e+00, 6.17295515613228e+00, 6.17024537391141e+00, 6.16630051163193e+00, 6.16077662897277e+00, 6.15329254304002e+00, 6.14343707165499e+00, 6.13077834533885e+00, 6.11487501706704e+00, 6.09528909391098e+00, 6.07160000410269e+00, 6.04341940475145e+00, 6.01040613937502e+00, 5.97228068169207e+00, 5.92883836369398e+00, 5.87996069105643e+00, 5.82562410327685e+00, 5.76590564065110e+00, 5.70098513097253e+00, 5.63114369586307e+00, 5.55675858555655e+00, 5.47829456445496e+00, 5.39629226966555e+00, 5.31135413405688e+00, 5.22412859026075e+00, 5.13529334298910e+00, 5.04553850962088e+00, 4.95555038417926e+00, 4.86599648370574e+00, 4.77751239950578e+00, 4.69069081341887e+00, 4.60607286801780e+00, 4.52414191645290e+00, 4.44531953727299e+00, 4.36996359237601e+00, 4.29836803710958e+00, 4.23076415969381e+00, 4.16732292732624e+00, 4.10815814063540e+00, 4.05333013790764e+00, 4.00284983784411e+00, 3.95668295831064e+00, 3.91475429433856e+00, 3.87695197894591e+00, 3.84313168386868e+00, 3.81312074354025e+00, 3.78672220459747e+00, 3.76371881503981e+00, 3.74387697226805e+00, 3.72695064807014e+00, 3.71268530186093e+00, 3.70082178200239e+00, 3.69110019996512e+00, 3.68326374483215e+00, 3.67706238778985e+00, 3.67225640952849e+00, 3.66861966961461e+00, 3.66594252749219e+00, 3.66403432110165e+00, 3.66272531202583e+00, 3.66186801586149e+00, 3.66133785286470e+00, 3.66103307590330e+00, 3.66087395895757e+00, 3.66080125804582e+00, 3.66077398557519e+00, 3.66076656681619e+00, 3.66076547182342e+00, 3.66076538927371e+00, 3.66076393166612e+00, 3.66075405616813e+00, 3.66071775290989e+00, 3.66062098265128e+00, 3.66040920786973e+00, 3.66000365131089e+00, 3.65929843975820e+00, 3.65815877727111e+00, 3.65642027602567e+00, 3.65388955273822e+00, 3.65034617260682e+00, 3.64554598880490e+00, 3.63922588214912e+00, 3.63110985175469e+00, 3.62091634366081e+00, 3.60836663257717e+00, 3.59319399599543e+00, 3.57515334573014e+00, 3.55403091684408e+00, 3.52965356603700e+00, 3.50189720887956e+00, 3.47069393421160e+00, 3.43603737829848e+00, 3.39798602084734e+00, 3.35666417530570e+00, 3.31226057841104e+00, 3.26502462693946e+00, 3.21526044959025e+00, 3.16331912584347e+00, 3.10958946055952e+00, 3.05448778581064e+00, 2.99844728703987e+00, 2.94190734040239e+00, 2.88530330671483e+00, 2.82905716169094e+00, 2.77356925997007e+00, 2.71921143978402e+00, 2.66632158327528e+00, 2.61519966076742e+00, 2.56610521076989e+00, 2.51925614487996e+00, 2.47482872030753e+00, 2.43295849331953e+00, 2.39374205393071e+00, 2.35723934392462e+00, 2.32347637414193e+00, 2.29244817976894e+00, 2.26412188077853e+00, 2.23843974556810e+00, 2.21532218646142e+00, 2.19467064389492e+00, 2.17637034019849e+00, 2.16029290287617e+00, 2.14629887068601e+00, 2.13424010351238e+00, 2.12396211928534e+00, 2.11530637856063e+00, 2.10811253060368e+00, 2.10222062487561e+00, 2.09747327981599e+00, 2.09371778797293e+00, 2.09080812412668e+00, 2.08860681233178e+00, 2.08698659990665e+00, 2.08583188224345e+00, 2.08503982251064e+00, 2.08452111511617e+00, 2.08420035100688e+00, 2.08401595592298e+00, 2.08391968867633e+00, 2.08387570419937e+00, 2.08385920423672e+00, 2.08385471585939e+00, 2.08385405338474e+00, 2.08385400317009e+00, 2.08385311290814e+00, 2.08384708125909e+00, 2.08382490837856e+00, 2.08376580438365e+00, 2.08363646061241e+00, 2.08338876666917e+00, 2.08295807130054e+00, 2.08226207786571e+00, 2.08120045635921e+00, 2.07965524254870e+00, 2.07749207943594e+00, 2.07456233549097e+00, 2.07070610670414e+00, 2.06575607474359e+00, 2.05954215155603e+00, 2.05189679290304e+00, 2.04266081219482e+00, 2.03168947548892e+00, 2.01885861366443e+00, 2.00407045420323e+00, 1.98725885827768e+00, 1.96839365365798e+00, 1.94748378324971e+00, 1.92457904332124e+00, 1.89977026220741e+00, 1.87318786404006e+00, 1.84499886487471e+00, 1.81540245082465e+00, 1.78462437941066e+00, 1.75291051715371e+00, 1.72051987153769e+00, 1.68771749007566e+00, 1.65476758317958e+00, 1.62192718425653e+00, 1.58944059623451e+00, 1.55753479672947e+00, 1.52641589307787e+00, 1.49626664164859e+00, 1.46724497977905e+00, 1.43948346770451e+00, 1.41308950391675e+00, 1.38814616025230e+00, 1.36471348070710e+00, 1.34283009745626e+00, 1.32251503530800e+00, 1.30376959839728e+00, 1.28657925732573e+00, 1.27091547882070e+00, 1.25673746166319e+00, 1.24399376110331e+00, 1.23262379875823e+00, 1.22255926596439e+00, 1.21372543588419e+00, 1.20604240364207e+00, 1.19942627475530e+00, 1.19379032053500e+00, 1.18904611538440e+00, 1.18510466546168e+00, 1.18187753148320e+00, 1.17927794104499e+00, 1.17722187830381e+00, 1.17562913178482e+00, 1.17442427507224e+00, 1.17353755075193e+00, 1.17290562567908e+00, 1.17247218575680e+00, 1.17218834106265e+00, 1.17201281725555e+00, 1.17191191642478e+00, 1.17185923939766e+00, 1.17183517135403e+00, 1.17182614268489e+00, 1.17182368667599e+00, 1.17182332417453e+00, 1.17182329652306e+00, 1.17182280398666e+00, 1.17181946698446e+00, 1.17180719987545e+00, 1.17177450075907e+00, 1.17170294204135e+00, 1.17156590810995e+00, 1.17132763490862e+00, 1.17094260247727e+00, 1.17035532696171e+00, 1.16950059252376e+00, 1.16830415513193e+00, 1.16668393846987e+00, 1.16455172636277e+00, 1.16181533573069e+00, 1.15838122922331e+00, 1.15415749820612e+00, 1.14905711635122e+00, 1.14300133429796e+00, 1.13592305995668e+00, 1.12777005067286e+00, 1.11850773617181e+00, 1.10812149775599e+00, 1.09661825110767e+00, 1.08402721691054e+00, 1.07039981296512e+00, 1.05580865921129e+00, 1.04034574731361e+00, 1.02411988279039e+00, 1.00725355398581e+00, 9.89879413700493e-01, 9.72136573262674e-01, 9.54166904887387e-01, 9.36111528315209e-01, 9.18107625740348e-01, 9.00285689784425e-01, 8.82767267779883e-01, 8.65663226350065e-01, 8.49072526537270e-01, 8.33081473510958e-01, 8.17763386839118e-01, 8.03178627021879e-01, 7.89374910349892e-01, 7.76387845699375e-01, 7.64241632097395e-01, 7.52949863423424e-01, 7.42516395347735e-01, 7.32936238703005e-01, 7.24196452347371e-01, 7.16277016797761e-01, 7.09151677219324e-01, 7.02788750569468e-01, 6.97151896692404e-01, 6.92200856869903e-01, 6.87892165717003e-01, 6.84179843378374e-01, 6.81016074780302e-01, 6.78351881333668e-01, 6.76137788123938e-01, 6.74324486489908e-01, 6.72863488256567e-01, 6.71707764074611e-01, 6.70812354677324e-01, 6.70134940750401e-01, 6.69636354850894e-01, 6.69281017678401e-01, 6.69037281180820e-01, 6.68877662540169e-01, 6.68778955981807e-01, 6.68722213408955e-01, 6.68692589802546e-01, 6.68679054790189e-01, 6.68673977381516e-01, 6.68672596207202e-01, 6.68672392348896e-01, 6.68672376688946e-01, 6.68672096310317e-01, 6.68670196706202e-01, 6.68663213593987e-01, 6.68644599437599e-01, 6.68603864115667e-01, 6.68525856097560e-01, 6.68390215134082e-01, 6.68171025246565e-01, 6.67836694095413e-01, 6.67350081266832e-01, 6.66668893179576e-01, 6.65746355690231e-01, 6.64532166652734e-01, 6.62973719421558e-01, 6.61017574644788e-01, 6.58611142120606e-01, 6.55704517905122e-01, 6.52252405618796e-01, 6.48216036732393e-01, 6.43564994431287e-01, 6.38278841293584e-01, 6.32348453924678e-01, 6.25776978622696e-01, 6.18580340948995e-01, 6.10787267558919e-01, 6.02438808691801e-01, 5.93587381512521e-01, 5.84295384933100e-01, 5.74633462665147e-01, 5.64678510671484e-01, 5.54511536377959e-01, 5.44215479459425e-01, 5.33873098184046e-01, 5.23565012449410e-01, 5.13367976583083e-01, 5.03353433794118e-01, 4.93586381953520e-01, 4.84124559067027e-01, 4.75017937950639e-01, 4.66308504361209e-01, 4.58030281813990e-01, 4.50209559700070e-01, 4.42865278843897e-01, 4.36009529728011e-01, 4.29648122467611e-01, 4.23781193369621e-01, 4.18403819711088e-01, 4.13506621482114e-01, 4.09076335671330e-01, 4.05096354807640e-01, 4.01547226643132e-01, 3.98407115937289e-01, 3.95652232238632e-01, 3.93257229394266e-01, 3.91195583320979e-01, 3.89439954452158e-01, 3.87962540358757e-01, 3.86735422474104e-01, 3.85730908796873e-01, 3.84921872083666e-01, 3.84282080577023e-01, 3.83786515947790e-01, 3.83411671081248e-01, 3.83135818789285e-01, 3.82939241654043e-01, 3.82804413107721e-01, 3.82716120578735e-01, 3.82661523055418e-01, 3.82630137639978e-01, 3.82613752419118e-01, 3.82606266057820e-01, 3.82603457694586e-01, 3.82602693754450e-01, 3.82602580998594e-01, 3.82602572270812e-01, 3.82602415146835e-01, 3.82601350610429e-01, 3.82597437285455e-01, 3.82587005978294e-01, 3.82564178233380e-01, 3.82520463879587e-01, 3.82444455230157e-01, 3.82321634516927e-01, 3.82134309797777e-01, 3.81861692739392e-01, 3.81480129044959e-01, 3.80963488550989e-01, 3.80283716876988e-01, 3.79411543815749e-01, 3.78317335423304e-01, 3.76972067274608e-01, 3.75348386166323e-01, 3.73421717520657e-01, 3.71171366971424e-01, 3.68581558321182e-01, 3.65642347433194e-01, 3.62350353622944e-01, 3.58709257263840e-01, 3.54730024545816e-01, 3.50430836873457e-01, 3.45836721854141e-01, 3.40978903303991e-01, 3.35893907033849e-01, 3.30622475293467e-01, 3.25208353964058e-01, 3.19697021869217e-01, 3.14134430689890e-01, 3.08565817476008e-01, 3.03034640846093e-01, 2.97581678254232e-01, 2.92244306892030e-01, 2.87055976452550e-01, 2.82045869352402e-01, 2.77238733906344e-01, 2.72654868759361e-01, 2.68310232608115e-01, 2.64216651605352e-01, 2.60382097375356e-01, 2.56811010737437e-01, 2.53504649502915e-01, 2.50461442597754e-01, 2.47677336867653e-01, 2.45146126926755e-01, 2.42859762083967e-01, 2.40808627563695e-01, 2.38981799839365e-01, 2.37367277884815e-01, 2.35952193522991e-01, 2.34723004848036e-01, 2.33665676964662e-01, 2.32765854088655e-01, 2.32009026453241e-01, 2.31380694542878e-01, 2.30866532013479e-01, 2.30452547348861e-01, 2.30125242947040e-01, 2.29871769035427e-01, 2.29680068685383e-01, 2.29539009336733e-01, 2.29438495733262e-01, 2.29369559072751e-01, 2.29324417513354e-01, 2.29296503940219e-01, 2.29280458032562e-01, 2.29272081094081e-01, 2.29268253709321e-01, 2.29266817942602e-01, 2.29266427380850e-01, 2.29266369734830e-01, 2.29266365236597e-01, 2.29266283788974e-01, 2.29265731970131e-01, 2.29263703439941e-01, 2.29258296232047e-01, 2.29246463242148e-01, 2.29223803767113e-01, 2.29184405371399e-01, 2.29120744863295e-01, 2.29023656539866e-01, 2.28882374963675e-01, 2.28684658195405e-01, 2.28416995432937e-01, 2.28064900237385e-01, 2.27613286859778e-01, 2.27046922630095e-01, 2.26350944089219e-01, 2.25511418864869e-01, 2.24515929712695e-01, 2.23354152305006e-01, 2.22018394964081e-01, 2.20504067288417e-01, 2.18810046046386e-01, 2.16938911084569e-01, 2.14897031225632e-01, 2.12694489725948e-01, 2.10344849960023e-01, 2.07864773466341e-01, 2.05273513081366e-01, 2.02592312446924e-01, 1.99843748792730e-01, 1.97051058051002e-01, 1.94237479979179e-01, 1.91425656404888e-01, 1.88637108692774e-01, 1.85891812031102e-01, 1.83207875213014e-01, 1.80601326225400e-01, 1.78085996913111e-01, 1.75673494715681e-01, 1.73373246132186e-01, 1.71192595090659e-01, 1.69136939553482e-01, 1.67209891155025e-01, 1.65413445040500e-01, 1.63748149923890e-01, 1.62213271305887e-01, 1.60806943483720e-01, 1.59526308264773e-01, 1.58367640096444e-01, 1.57326458648413e-01, 1.56397630763706e-01, 1.55575464177190e-01, 1.54853795543815e-01, 1.54226075188532e-01, 1.53685450653228e-01, 1.53224850638367e-01, 1.52837070366417e-01, 1.52514858770573e-01, 1.52251007271545e-01, 1.52038439270413e-01, 1.51870298897236e-01, 1.51740037034742e-01, 1.51641492226588e-01, 1.51568963806412e-01, 1.51517274477357e-01, 1.51481819649590e-01, 1.51458601113632e-01, 1.51444243080050e-01, 1.51435989229014e-01, 1.51431680146786e-01, 1.51429711330735e-01, 1.51428972766226e-01, 1.51428771859372e-01, 1.51428742205979e-01, 1.51428739872358e-01, 1.51428697366512e-01, 1.51428409383485e-01, 1.51427350733172e-01, 1.51424528804114e-01, 1.51418353304507e-01, 1.51406527338892e-01, 1.51385964559935e-01, 1.51352736818481e-01, 1.51302056345031e-01, 1.51228295966732e-01, 1.51125050118949e-01, 1.50985238359382e-01, 1.50801251662088e-01, 1.50565139910384e-01, 1.50268836738150e-01, 1.49904415266732e-01, 1.49464365528146e-01, 1.48941881710812e-01, 1.48331145146543e-01, 1.47627587536718e-01, 1.46828118635725e-01, 1.45931303721270e-01, 1.44937478777535e-01, 1.43848795289713e-01, 1.42669191550652e-01, 1.41404292874284e-01, 1.40061248429717e-01, 1.38648516901384e-01, 1.37175616308667e-01, 1.35652854780616e-01, 1.34091058802809e-01, 1.32501313581173e-01, 1.30894727009334e-01, 1.29282224753979e-01, 1.27674379800098e-01, 1.26081276095789e-01, 1.24512403236849e-01, 1.22976577639221e-01, 1.21481885174219e-01, 1.20035640400005e-01, 1.18644358006421e-01, 1.17313732824968e-01, 1.16048625800289e-01, 1.14853054609725e-01, 1.13730188879643e-01, 1.12682350831278e-01, 1.11711022513916e-01, 1.10816860669289e-01, 1.09999719981289e-01, 1.09258685212765e-01, 1.08592112554433e-01, 1.07997680330868e-01, 1.07472448939021e-01, 1.07012929551701e-01, 1.06615160747637e-01, 1.06274791891181e-01, 1.05987171815627e-01, 1.05747441127050e-01, 1.05550626258378e-01, 1.05391733232736e-01, 1.05265838966503e-01, 1.05168177867625e-01, 1.05094221471205e-01, 1.05039748931759e-01, 1.05000906346575e-01, 1.04974253137570e-01, 1.04956794058012e-01, 1.04945995805920e-01, 1.04939787709483e-01, 1.04936546473482e-01, 1.04935065514226e-01, 1.04934509953608e-01, 1.04934358827450e-01, 1.04934336521552e-01, 1.04934334750553e-01, 1.04934302296243e-01, 1.04934082413834e-01, 1.04933274102194e-01, 1.04931119437268e-01, 1.04926403991271e-01, 1.04917373292110e-01, 1.04901668654989e-01, 1.04876285314020e-01, 1.04837555308720e-01, 1.04781157037422e-01, 1.04702152737886e-01, 1.04595054400889e-01, 1.04453917733423e-01, 1.04272462764882e-01, 1.04044218550172e-01, 1.03762688206515e-01, 1.03421529293570e-01, 1.03014743418115e-01, 1.02536868027670e-01, 1.01983162798604e-01, 1.01349782912763e-01, 1.00633931924989e-01, 9.98339878183714e-02, 9.89495971484189e-02, 9.79817337444331e-02, 9.69327201619166e-02, 9.58062118353524e-02, 9.46071456312723e-02, 9.33416559991116e-02, 9.20169630249461e-02, 9.06412371110719e-02, 8.92234447226912e-02, 8.77731791048717e-02, 8.63004795646418e-02, 8.48156434286835e-02, 8.33290358453951e-02, 8.18509034058224e-02, 8.03911973684563e-02, 7.89594106229288e-02, 7.75644310205621e-02, 7.62144124204127e-02, 7.49166655198650e-02, 7.36775711494744e-02, 7.25025187294079e-02, 7.13958708914646e-02, 7.03609527820959e-02, 6.94000622937807e-02, 6.85144977665573e-02, 6.77045994808338e-02, 6.69465015176627e-02, 6.62869567608908e-02, 6.56986959973693e-02, 6.51789099087523e-02, 6.47241943197230e-02, 6.43306437482778e-02, 6.39939476785685e-02, 6.37094883866212e-02, 6.34724391372572e-02, 6.32778614835843e-02, 6.31208002623111e-02, 6.29963747185798e-02, 6.28998640445957e-02, 6.28267855112057e-02, 6.27729633370961e-02, 6.27345864995448e-02, 6.27082538526338e-02, 6.26910051890803e-02, 6.26803372495885e-02, 6.26742041321648e-02, 6.26710020598011e-02, 6.26695389990886e-02, 6.26689901533268e-02, 6.26688408538598e-02, 6.26688188175564e-02, 6.26688170520424e-02, 6.26687844948464e-02, 6.26685639151816e-02, 6.26677530410973e-02, 6.26655915499015e-02, 6.26608611863020e-02, 6.26518020010565e-02, 6.26360480865173e-02, 6.26105858011243e-02, 6.25717370276817e-02, 6.25151694888046e-02, 6.24359355302033e-02, 6.23285400585729e-02, 6.21870374775852e-02, 6.20051564963831e-02, 6.17764506018403e-02, 6.14944708200653e-02, 6.11529561864046e-02, 6.07460361648293e-02, 6.02684381811684e-02, 5.97156925483091e-02, 5.90843264527479e-02, 5.83720384197761e-02, 5.75778448489928e-02, 5.67021908559155e-02, 5.57470187868887e-02, 5.47157893799550e-02, 5.36134525690812e-02, 5.24463672855546e-02, 5.12221721700143e-02, 4.99496117081435e-02, 4.86383247578885e-02, 4.72986045478763e-02, 4.59411408082531e-02, 4.45767555920895e-02, 4.32161444576593e-02, 4.18696339868490e-02, 4.05469651688293e-02, 3.92571101243533e-02, 3.80081271855772e-02, 3.68070567214289e-02, 3.56598575514511e-02, 3.45713815370630e-02, 3.35453821419662e-02, 3.25845515104959e-02, 3.16905799535476e-02, 3.08642316220897e-02, 3.01054305066657e-02, 2.94133516127175e-02, 2.87865131012950e-02, 2.82228662268267e-02, 2.77198809443178e-02, 2.72746260086478e-02, 2.68838431912855e-02, 2.65440158545687e-02, 2.62514325383734e-02, 2.60022464296412e-02, 2.57925316175545e-02, 2.56183369160112e-02, 2.54757377905435e-02, 2.53608866004242e-02, 2.52700609964194e-02, 2.51997099439478e-02, 2.51464965063377e-02, 2.51073362601346e-02, 2.50794300498143e-02, 2.50602897426688e-02, 2.50477557223121e-02, 2.50400050573631e-02, 2.50355495861783e-02, 2.50332235422945e-02, 2.50321607795740e-02, 2.50317621047600e-02, 2.50316536559581e-02, 2.50316376491497e-02, 2.50316363546261e-02, 2.50316123321284e-02, 2.50314495764030e-02, 2.50308512733426e-02, 2.50292564488888e-02, 2.50257663761175e-02, 2.50190830882673e-02, 2.50074627319776e-02, 2.49886863491184e-02, 2.49600504929361e-02, 2.49183798258912e-02, 2.48600634630805e-02, 2.47811162676896e-02, 2.46772655244882e-02, 2.45440623908877e-02, 2.43770162477404e-02, 2.41717485822619e-02, 2.39241614089138e-02, 2.36306135887929e-02, 2.32880969011120e-02, 2.28944025285358e-02, 2.24482679304714e-02, 2.19494940627527e-02, 2.13990236831165e-02, 2.07989731236893e-02, 2.01526123739768e-02, 1.94642914794656e-02, 1.87393148812021e-02, 1.79837690931350e-02, 1.72043126698327e-02, 1.64079403829760e-02, 1.56017355750038e-02, 1.47926255529711e-02, 1.39871545193984e-02, 1.31912869670288e-02, 1.24102518774976e-02, 1.16484347664567e-02, 1.09093209468796e-02, 1.01954896834684e-02, 9.50865548004742e-03, 8.84974983864595e-03, 8.21903463601359e-03, 7.61623689913904e-03, 7.04069423389342e-03, 6.49150041243729e-03, 5.96764154731488e-03, 5.46811476930031e-03, 4.99202327596038e-03, 4.53864390634351e-03, 4.10746584338995e-03, 3.69820142519472e-03, 3.31077212211994e-03, 2.94527431805136e-03, 2.60193050772273e-03, 2.28103186965711e-03, 1.98287794158603e-03, 1.70771842398665e-03, 1.45570110510107e-03, 1.22682869759672e-03, 1.02092616054961e-03, 8.37618985613631e-04, 6.76322048036582e-04, 5.36238014149149e-04, 4.16363953632981e-04, 3.15504700864722e-04, 2.32291584225663e-04, 1.65205336756458e-04, 1.12602260742278e-04, 7.27430013398589e-05, 4.38235760640282e-05, 2.40086174869703e-05, 1.14671798598920e-05, 4.41208936763360e-06, 1.14512177162851e-06, -1.42108547152020e-14],
[ 9.30000000000000e+01, 9.29999998872877e+01, 9.29999964019988e+01, 9.29999727886128e+01, 9.29998859826922e+01, 9.29996545852482e+01, 9.29991481535984e+01, 9.29981781844330e+01, 9.29964911121597e+01, 9.29937635824777e+01, 9.29896001900194e+01, 9.29835337929852e+01, 9.29750284407778e+01, 9.29634848768615e+01, 9.29482485124740e+01, 9.29286197108153e+01, 9.29038661779042e+01, 9.28732372253257e+01, 9.28359796488184e+01, 9.27913549496114e+01, 9.27386576050707e+01, 9.26772340633432e+01, 9.26065020865928e+01, 9.25259699963688e+01, 9.24352552861051e+01, 9.23341019705596e+01, 9.22223959578588e+01, 9.21001776787156e+01, 9.19676512112992e+01, 9.18251892158851e+01, 9.16733331473803e+01, 9.15127884395399e+01, 9.13444146321877e+01, 9.11692107115773e+01, 9.09882962187417e+01, 9.08028889175475e+01, 9.06142799774592e+01, 9.04238077022498e+01, 9.02328308253342e+01, 9.00427023075023e+01, 8.98547444345041e+01, 8.96702258446739e+01, 8.94903409439601e+01, 8.93161920060205e+01, 8.91487741205396e+01, 8.89889630488629e+01, 8.88375059718441e+01, 8.86950150659129e+01, 8.85619638131655e+01, 8.84386859327881e+01, 8.83253768082262e+01, 8.82220972725665e+01, 8.81287796006213e+01, 8.80452355387290e+01, 8.79711661820145e+01, 8.79061734843699e+01, 8.78497731598161e+01, 8.78014087066352e+01, 8.77604662593220e+01, 8.77262899497105e+01, 8.76981974392962e+01, 8.76754952715225e+01, 8.76574936872070e+01, 8.76435205497710e+01, 8.76329340406041e+01, 8.76251338094435e+01, 8.76195703002525e+01, 8.76157520193447e+01, 8.76132505683832e+01, 8.76117033287534e+01, 8.76108137534903e+01, 8.76103492958516e+01, 8.76101370769403e+01, 8.76100574655915e+01, 8.76100358093074e+01, 8.76100326128828e+01, 8.76100321935793e+01, 8.76100224243113e+01, 8.76099562360700e+01, 8.76097129208947e+01, 8.76090643304428e+01, 8.76076448983494e+01, 8.76049264863681e+01, 8.76001990521144e+01, 8.75925580117123e+01, 8.75808990263997e+01, 8.75639207755491e+01, 8.75401360856364e+01, 8.75078915611920e+01, 8.74653956064004e+01, 8.74107544338398e+01, 8.73420153327978e+01, 8.72572161218817e+01, 8.71544393541912e+01, 8.70318695009310e+01, 8.68878510421934e+01, 8.67209451802560e+01, 8.65299828030825e+01, 8.63141114012422e+01, 8.60728339011081e+01, 8.58060378116412e+01, 8.55140136417550e+01, 8.51974621414034e+01, 8.48574904425847e+01, 8.44955975323827e+01, 8.41136496457155e+01, 8.37138461825693e+01, 8.32986767882846e+01, 8.28708704763791e+01, 8.24333382383852e+01, 8.19891114043757e+01, 8.15412787985278e+01, 8.10929260705037e+01, 8.06470801897547e+01, 8.02066610110741e+01, 7.97744404470204e+01, 7.93530086353429e+01, 7.89447459203418e+01, 7.85517995079288e+01, 7.81760640780448e+01, 7.78191661437940e+01, 7.74824523162658e+01, 7.71669817883335e+01, 7.68735233189423e+01, 7.66025568585078e+01, 7.63542797817013e+01, 7.61286175367300e+01, 7.59252384027657e+01, 7.57435719710992e+01, 7.55828309212767e+01, 7.54420356377563e+01, 7.53200411936494e+01, 7.52155662075835e+01, 7.51272230534614e+01, 7.50535488704325e+01, 7.49930367842167e+01, 7.49441667155301e+01, 7.49054351224287e+01, 7.48753830069232e+01, 7.48526215179029e+01, 7.48358545068999e+01, 7.48238974437000e+01, 7.48156921765175e+01, 7.48103171255687e+01, 7.48069926264246e+01, 7.48050812856096e+01, 7.48040833691028e+01, 7.48036274072486e+01, 7.48034563593159e+01, 7.48034098300460e+01, 7.48034029624202e+01, 7.48034023561395e+01, 7.48033904763689e+01, 7.48033099891618e+01, 7.48030141094806e+01, 7.48022254013944e+01, 7.48004993254862e+01, 7.47971936546850e+01, 7.47914449748382e+01, 7.47821533366161e+01, 7.47679759518441e+01, 7.47473306266361e+01, 7.47184093907532e+01, 7.46792025122548e+01, 7.46275327750537e+01, 7.45610995441967e+01, 7.44775317545650e+01, 7.43744485449202e+01, 7.42495258398247e+01, 7.41005667827619e+01, 7.39255735755702e+01, 7.37228180148340e+01, 7.34909078661128e+01, 7.32288462074438e+01, 7.29360810220678e+01, 7.26125426351130e+01, 7.22586670684512e+01, 7.18754040194843e+01, 7.14642089267068e+01, 7.10270194227658e+01, 7.05662173282528e+01, 7.00845781212540e+01, 6.95852104356351e+01, 6.90714885146846e+01, 6.85469806322954e+01, 6.80153763020538e+01, 6.74804146927667e+01, 6.69458161635386e+01, 6.64152183354153e+01, 6.58921177138834e+01, 6.53798175987143e+01, 6.48813828421329e+01, 6.43996018863741e+01, 6.39369563685481e+01, 6.34955983916269e+01, 6.30773353304568e+01, 6.26836218062060e+01, 6.23155582664043e+01, 6.19738954841843e+01, 6.16590442499652e+01, 6.13710895595817e+01, 6.11098086789648e+01, 6.08746925582348e+01, 6.06649701538528e+01, 6.04796352811680e+01, 6.03174756544180e+01, 6.01771037762296e+01, 6.00569893170115e+01, 5.99554925814830e+01, 5.98708986012567e+01, 5.98014513258958e+01, 5.97453873177412e+01, 5.97009682958915e+01, 5.96665118298608e+01, 5.96404194609209e+01, 5.96212015351491e+01, 5.96074980711730e+01, 5.95980950596633e+01, 5.95919357002551e+01, 5.95881262214911e+01, 5.95859360946756e+01, 5.95847926352002e+01, 5.95842701754906e+01, 5.95840741822287e+01, 5.95840208672539e+01, 5.95840129980733e+01, 5.95840123668677e+01, 5.95840007179224e+01, 5.95839217946002e+01, 5.95836316640642e+01, 5.95828582821556e+01, 5.95811657526263e+01, 5.95779243479006e+01, 5.95722874922175e+01, 5.95631767658573e+01, 5.95492758239740e+01, 5.95290339309777e+01, 5.95006795853180e+01, 5.94622444434710e+01, 5.94115974410598e+01, 5.93464886520872e+01, 5.92646020286629e+01, 5.91636157351811e+01, 5.90412683528520e+01, 5.88954288113182e+01, 5.87241675392079e+01, 5.85258260542342e+01, 5.82990820746593e+01, 5.80430072601929e+01, 5.77571149017173e+01, 5.74413952771521e+01, 5.70963369542444e+01, 5.67229330061134e+01, 5.63226718500770e+01, 5.58975131554881e+01, 5.54498499301543e+01, 5.49824584462192e+01, 5.44984380910557e+01, 5.40011435349941e+01, 5.34941118119252e+01, 5.29809870188240e+01, 5.24654453462112e+01, 5.19511230300110e+01, 5.14415495443966e+01, 5.09400879334032e+01, 5.04498836360460e+01, 4.99738225547370e+01, 4.95144985250079e+01, 4.90741898371656e+01, 4.86548440876697e+01, 4.82580704196734e+01, 4.78851381386403e+01, 4.75369807292032e+01, 4.72142044126530e+01, 4.69171005307822e+01, 4.66456611895768e+01, 4.63995977248299e+01, 4.61783616508210e+01, 4.59811678200865e+01, 4.58070195587063e+01, 4.56547355510012e+01, 4.55229782339079e+01, 4.54102834282086e+01, 4.53150908848381e+01, 4.52357753636906e+01, 4.51706777945890e+01, 4.51181360012708e+01, 4.50765144062432e+01, 4.50442320845098e+01, 4.50197885047258e+01, 4.50017862936815e+01, 4.49889503889459e+01, 4.49801430076013e+01, 4.49743739561725e+01, 4.49708059352285e+01, 4.49687546461275e+01, 4.49676836792524e+01, 4.49671943435636e+01, 4.49670107765167e+01, 4.49669608417974e+01, 4.49669534715363e+01, 4.49669529030708e+01, 4.49669426952559e+01, 4.49668735358084e+01, 4.49666192983713e+01, 4.49659415947670e+01, 4.49644584573065e+01, 4.49616180727930e+01, 4.49566786238846e+01, 4.49486951708439e+01, 4.49365143639962e+01, 4.49187776102197e+01, 4.48939331196429e+01, 4.48602570252661e+01, 4.48158834937467e+01, 4.47588434287428e+01, 4.46871110129609e+01, 4.45986569522551e+01, 4.44915068934397e+01, 4.43638031132246e+01, 4.42138672513348e+01, 4.40402616220059e+01, 4.38418465189222e+01, 4.36178309567579e+01, 4.33678144829101e+01, 4.30918180438964e+01, 4.27903023811238e+01, 4.24641730213596e+01, 4.21147715671276e+01, 4.17438536275375e+01, 4.13535543152171e+01, 4.09463427418041e+01, 4.05249673645303e+01, 4.00923943747262e+01, 3.96517415794296e+01, 3.92062103968384e+01, 3.87590186292731e+01, 3.83133365458991e+01, 3.78722284693808e+01, 3.74386015288182e+01, 3.70151625866881e+01, 3.66043836826067e+01, 3.62084757760179e+01, 3.58293701882201e+01, 3.54687069557478e+01, 3.51278292734549e+01, 3.48077832626177e+01, 3.45093223868247e+01, 3.42329159180552e+01, 3.39787609138014e+01, 3.37467972074271e+01, 3.35367249490553e+01, 3.33480242723592e+01, 3.31799767071361e+01, 3.30316880058854e+01, 3.29021120982590e+01, 3.27900759224392e+01, 3.26943049006497e+01, 3.26134488232948e+01, 3.25461078822446e+01, 3.24908585515632e+01, 3.24462789593813e+01, 3.24109733355753e+01, 3.23835950653997e+01, 3.23628678382138e+01, 3.23476043609762e+01, 3.23367221144448e+01, 3.23292556697398e+01, 3.23243651549199e+01, 3.23213405633606e+01, 3.23196017230810e+01, 3.23186938916130e+01, 3.23182790958862e+01, 3.23181234916838e+01, 3.23180811635674e+01, 3.23180749160261e+01, 3.23180744429811e+01, 3.23180660630369e+01, 3.23180092876912e+01, 3.23178005756597e+01, 3.23172442271255e+01, 3.23160266776165e+01, 3.23136949463064e+01, 3.23096401137989e+01, 3.23030866086722e+01, 3.22930879703309e+01, 3.22785296237886e+01, 3.22581390442363e+01, 3.22305034997444e+01, 3.21940953346466e+01, 3.21473044924614e+01, 3.20884776785839e+01, 3.20159632381240e+01, 3.19281604882615e+01, 3.18235719184880e+01, 3.17008563820434e+01, 3.15588811760292e+01, 3.13967707735515e+01, 3.12139499517564e+01, 3.10101791698521e+01, 3.07855802959880e+01, 3.05406511548913e+01, 3.02762678528829e+01, 2.99936744084301e+01, 2.96944598439019e+01, 2.93805235424229e+01, 2.90540303035115e+01, 2.87173570993598e+01, 2.83730339938098e+01, 2.80236819920945e+01, 2.76719507019481e+01, 2.73204585823325e+01, 2.69717382359130e+01, 2.66281886958889e+01, 2.62920360246230e+01, 2.59653028570134e+01, 2.56497868667969e+01, 2.53470475798243e+01, 2.50584005535646e+01, 2.47849177072491e+01, 2.45274325152193e+01, 2.42865488389829e+01, 2.40626523309031e+01, 2.38559235514406e+01, 2.36663521644175e+01, 2.34937517822221e+01, 2.33377752072444e+01, 2.31979299488150e+01, 2.30735939854618e+01, 2.29640317938814e+01, 2.28684106843156e+01, 2.27858174732121e+01, 2.27152754940176e+01, 2.26557619009764e+01, 2.26062251637581e+01, 2.25656025873258e+01, 2.25328376265210e+01, 2.25068967034631e+01, 2.24867851832933e+01, 2.24715621251853e+01, 2.24603534055649e+01, 2.24523628127915e+01, 2.24468807393907e+01, 2.24432901496605e+01, 2.24410695754904e+01, 2.24397929878680e+01, 2.24391265004679e+01, 2.24388219783279e+01, 2.24387077417957e+01, 2.24386766667114e+01, 2.24386720800949e+01, 2.24386717357762e+01, 2.24386656753762e+01, 2.24386246152908e+01, 2.24384736743095e+01, 2.24380713228840e+01, 2.24371907958282e+01, 2.24355045164186e+01, 2.24325721726277e+01, 2.24278330213668e+01, 2.24206029202727e+01, 2.24100764945205e+01, 2.23953347327253e+01, 2.23753581652252e+01, 2.23490456049566e+01, 2.23152382234149e+01, 2.22727484937769e+01, 2.22203932677275e+01, 2.21570299760228e+01, 2.20815946759584e+01, 2.19931404377717e+01, 2.18908743955612e+01, 2.17741917148652e+01, 2.16427047715743e+01, 2.14962660081991e+01, 2.13349832321774e+01, 2.11592265285611e+01, 2.09696264411748e+01, 2.07670635843121e+01, 2.05526503273122e+01, 2.03277055962979e+01, 2.00937241233781e+01, 1.98523416269567e+01, 1.96052974345355e+01, 1.93543959893856e+01, 1.91014685539379e+01, 1.88483362735451e+01, 1.85967756185672e+01, 1.83484870844218e+01, 1.81050678840546e+01, 1.78679891934457e+01, 1.76385782936796e+01, 1.74180056970673e+01, 1.72072770757320e+01, 1.70072295685485e+01, 1.68185318650323e+01, 1.66416873765084e+01, 1.64770398076016e+01, 1.63247805168672e+01, 1.61849571750518e+01, 1.60574833626433e+01, 1.59421488710587e+01, 1.58386305696549e+01, 1.57465037683809e+01, 1.56652540443249e+01, 1.55942895139686e+01, 1.55329535268990e+01, 1.54805377359907e+01, 1.54362954678014e+01, 1.53994552786795e+01, 1.53692345400863e+01, 1.53448528541780e+01, 1.53255450612096e+01, 1.53105735674216e+01, 1.52992396992988e+01, 1.52908937806679e+01, 1.52849436355058e+01, 1.52808612430361e+01, 1.52781873128428e+01, 1.52765336050617e+01, 1.52755828915747e+01, 1.52750865346936e+01, 1.52748597454177e+01, 1.52747746689763e+01, 1.52747515261258e+01, 1.52747481102897e+01, 1.52747478546416e+01, 1.52747433653323e+01, 1.52747129496080e+01, 1.52746011382806e+01, 1.52743030910455e+01, 1.52736508255755e+01, 1.52724016730148e+01, 1.52702294295432e+01, 1.52667186228100e+01, 1.52613622525067e+01, 1.52535632936233e+01, 1.52426401655894e+01, 1.52278362669270e+01, 1.52083335503085e+01, 1.51832699658492e+01, 1.51517604326356e+01, 1.51129208151776e+01, 1.50658941920989e+01, 1.50098785225993e+01, 1.49441546591034e+01, 1.48681135409501e+01, 1.47812813523532e+01, 1.46833414529437e+01, 1.45741519990530e+01, 1.44537583670395e+01, 1.43223997539486e+01, 1.41805096425371e+01, 1.40287101464430e+01, 1.38678005636754e+01, 1.36987407331985e+01, 1.35226299905331e+01, 1.33406826475733e+01, 1.31542009854674e+01, 1.29645467618999e+01, 1.27731122114878e+01, 1.25812914718543e+01, 1.23904533023372e+01, 1.22019158747593e+01, 1.20169243012632e+01, 1.18366314203809e+01, 1.16620821928026e+01, 1.14942018731699e+01, 1.13337879391653e+01, 1.11815055912317e+01, 1.10378865002428e+01, 1.09033303860858e+01, 1.07781089608380e+01, 1.06623717634367e+01, 1.05561534412655e+01, 1.04593820880314e+01, 1.03718883159914e+01, 1.02934148139865e+01, 1.02236262124865e+01, 1.01621191366565e+01, 1.01084323742372e+01, 1.00620571146974e+01, 1.00224472293101e+01, 9.98902955960537e+00, 9.96121416628002e+00, 9.93840446521844e+00, 9.92000714555816e+00, 9.90544173088108e+00, 9.89414961291112e+00, 9.88560236165458e+00, 9.87930910032738e+00, 9.87482273046490e+00, 9.87174480400122e+00, 9.86972886528407e+00, 9.86848212607039e+00, 9.86776538839760e+00, 9.86739119067765e+00, 9.86722021754814e+00, 9.86715607977010e+00, 9.86713863275310e+00, 9.86713605761045e+00, 9.86713586480940e+00, 9.86713247816256e+00, 9.86710953314109e+00, 9.86702518495193e+00, 9.86680034470021e+00, 9.86630829264766e+00, 9.86536597456278e+00, 9.86372733954627e+00, 9.86107904171453e+00, 9.85703878583878e+00, 9.85115654620040e+00, 9.84291882553495e+00, 9.83175604386685e+00, 9.81705305272709e+00, 9.79816265751077e+00, 9.77442190052366e+00, 9.74517071327747e+00, 9.70977239562235e+00, 9.66763523120811e+00, 9.61823441597156e+00, 9.56113337287374e+00, 9.49600346611857e+00, 9.42264112413461e+00, 9.34098144175273e+00, 9.25110746207253e+00, 9.15325453466812e+00, 9.04780939894403e+00, 8.93530393242465e+00, 8.81640381052004e+00, 8.69189262020925e+00, 8.56265222796722e+00, 8.42964039798542e+00, 8.29386677261904e+00, 8.15636835448683e+00, 8.01818557064600e+00, 7.88033986565858e+00, 7.74381358189389e+00, 7.60953266619107e+00, 7.47835251660116e+00, 7.35104707273656e+00, 7.22830107403024e+00, 7.11070527098794e+00, 6.99875427797657e+00, 6.89284669988669e+00, 6.79328714355928e+00, 6.70028973112149e+00, 6.61398275910575e+00, 6.53441418785138e+00, 6.46155769457739e+00, 6.39531907600042e+00, 6.33554283860299e+00, 6.28201886345525e+00, 6.23448907525979e+00, 6.19265407999369e+00, 6.15617976072544e+00, 6.12470383609454e+00, 6.09784239044781e+00, 6.07519637931507e+00, 6.05635810003064e+00, 6.04091759673896e+00, 6.02846894414142e+00, 6.01861632789355e+00, 6.01097981447978e+00, 6.00520068256345e+00, 6.00094617384141e+00, 5.99791351639945e+00, 5.99583307880422e+00, 5.99447052911319e+00, 5.99362789910960e+00, 5.99314348891664e+00, 5.99289058847411e+00, 5.99277503739103e+00, 5.99273169038753e+00, 5.99271989897018e+00, 5.99271815858233e+00, 5.99271802793376e+00, 5.99271572840935e+00, 5.99270014880344e+00, 5.99264287667872e+00, 5.99249021140989e+00, 5.99215611323258e+00, 5.99151629835930e+00, 5.99040372723104e+00, 5.98860571064665e+00, 5.98586283203156e+00, 5.98186985146930e+00, 5.97627871570576e+00, 5.96870374540090e+00, 5.95872900414028e+00, 5.94591777194952e+00, 5.92982394987223e+00, 5.91000511448907e+00, 5.88603682762923e+00, 5.85752769511927e+00, 5.82413456953417e+00, 5.78557721706437e+00, 5.74165172913250e+00, 5.69224196480320e+00, 5.63732836638796e+00, 5.57699359912526e+00, 5.51142462192702e+00, 5.44091098976406e+00, 5.36583940438992e+00, 5.28668475072644e+00, 5.20399806233891e+00, 5.11839203324541e+00, 5.03052482025696e+00, 4.94108295020068e+00, 4.85076415527676e+00, 4.76026090877835e+00, 4.67024532945395e+00, 4.58135597794228e+00, 4.49418689884604e+00, 4.40927908506185e+00, 4.32711437454466e+00, 4.24811164853958e+00, 4.17262509438235e+00, 4.10094422929264e+00, 4.03329535256550e+00, 3.96984409631582e+00, 3.91069877119835e+00, 3.85591424461383e+00, 3.80549613711116e+00, 3.75940517204083e+00, 3.71756155993013e+00, 3.67984934007853e+00, 3.64612063623193e+00, 3.61619981035798e+00, 3.58988751840070e+00, 3.56696468456401e+00, 3.54719641644373e+00, 3.53033588263189e+00, 3.51612816790107e+00, 3.50431410961737e+00, 3.49463410378995e+00, 3.48683185155809e+00, 3.48065799858733e+00, 3.47587360256338e+00, 3.47225334951396e+00, 3.46958842968732e+00, 3.46768897949711e+00, 3.46638599848569e+00, 3.46553265966759e+00, 3.46500494768812e+00, 3.46470158104847e+00, 3.46454320077464e+00, 3.46447083653591e+00, 3.46444369037610e+00, 3.46443630597796e+00, 3.46443521605683e+00, 3.46443513388947e+00, 3.46443368303304e+00, 3.46442385327572e+00, 3.46438771817297e+00, 3.46429139621668e+00, 3.46408060284543e+00, 3.46367692699893e+00, 3.46297498975811e+00, 3.46184062946438e+00, 3.46011024198530e+00, 3.45759138500883e+00, 3.45406472948201e+00, 3.44928740753717e+00, 3.44299776273924e+00, 3.43492145432359e+00, 3.42477880262240e+00, 3.41229319015125e+00, 3.39720025583779e+00, 3.37925754453920e+00, 3.35825420780967e+00, 3.33402030318197e+00, 3.30643521612845e+00, 3.27543473797304e+00, 3.24101637811621e+00, 3.20324256993248e+00, 3.16224154211122e+00, 3.11820576232088e+00, 3.07138800587926e+00, 3.02209524493622e+00, 2.97068068012481e+00, 2.91753433551348e+00, 2.86307270145204e+00, 2.80772793537011e+00, 2.75193711888564e+00, 2.69613202537738e+00, 2.64072978256208e+00, 2.58612472790186e+00, 2.53268165934395e+00, 2.48073058786508e+00, 2.43056300836996e+00, 2.38242962707611e+00, 2.33653942043183e+00, 2.29305985502528e+00, 2.25211807039544e+00, 2.21380281621971e+00, 2.17816693990534e+00, 2.14523023717950e+00, 2.11498250345554e+00, 2.08738665408787e+00, 2.06238181392148e+00, 2.03988630807359e+00, 2.01980051453671e+00, 2.00200956345938e+00, 1.98638588692343e+00, 1.97279163626256e+00, 1.96108099144573e+00, 1.95110238907988e+00, 1.94270069274017e+00, 1.93571932238767e+00, 1.93000234953898e+00, 1.92539655270964e+00, 1.92175341466778e+00, 1.91893103046377e+00, 1.91679588428198e+00, 1.91522444501472e+00, 1.91410452600542e+00, 1.91333635426388e+00, 1.91283329886416e+00, 1.91252221703093e+00, 1.91234338903320e+00, 1.91225002852154e+00, 1.91220737220476e+00, 1.91219137048874e+00, 1.91218701764733e+00, 1.91218637517754e+00, 1.91218632647923e+00, 1.91218546310067e+00, 1.91217961359084e+00, 1.91215811027780e+00, 1.91210079113843e+00, 1.91197535365437e+00, 1.91173514138715e+00, 1.91131745880810e+00, 1.91064250311585e+00, 1.90961299350653e+00, 1.90811456754030e+00, 1.90601699948157e+00, 1.90317627529444e+00, 1.89943753208597e+00, 1.89463883547121e+00, 1.88861572672987e+00, 1.88120642403409e+00, 1.87225751107891e+00, 1.86162989610074e+00, 1.84920477959065e+00, 1.83488933569524e+00, 1.81862179598092e+00, 1.80037562964020e+00, 1.78016254426278e+00, 1.75803408636668e+00, 1.73408169837774e+00, 1.70843518303307e+00, 1.68125962909028e+00, 1.65275095395369e+00, 1.62313030922326e+00, 1.59263766511727e+00, 1.56152493241819e+00, 1.53004899251891e+00, 1.49846498746767e+00, 1.46702017634580e+00, 1.43594859842275e+00, 1.40546670565291e+00, 1.37577004602322e+00, 1.34703100312865e+00, 1.31939753254470e+00, 1.29299278623093e+00, 1.26791548406899e+00, 1.24424087627979e+00, 1.22202213977457e+00, 1.20129206233577e+00, 1.18206488734160e+00, 1.16433821509509e+00, 1.14809488170965e+00, 1.13330476061243e+00, 1.11992645343437e+00, 1.10790885539764e+00, 1.09719259484079e+00, 1.08771135719250e+00, 1.07939311070434e+00, 1.07216125492386e+00, 1.06593571362782e+00, 1.06063399217248e+00, 1.05617221538330e+00, 1.05246615664942e+00, 1.04943226227236e+00, 1.04698866785021e+00, 1.04505619609277e+00, 1.04355931853158e+00, 1.04242705767628e+00, 1.04159380181288e+00, 1.04100000228082e+00, 1.04059272301464e+00, 1.04032601451817e+00, 1.04016108916916e+00, 1.04006628153662e+00, 1.04001678574795e+00, 1.03999417124749e+00, 1.03998568785404e+00, 1.03998338017348e+00, 1.03998303956496e+00, 1.03998301358349e+00, 1.03998255079340e+00, 1.03997941532688e+00, 1.03996788908469e+00, 1.03993716484997e+00, 1.03986992812629e+00, 1.03974117138372e+00, 1.03951729244920e+00, 1.03915552542388e+00, 1.03860374837453e+00, 1.03780070646512e+00, 1.03667668131487e+00, 1.03515462628121e+00, 1.03315177228076e+00, 1.03058168921494e+00, 1.02735676413869e+00, 1.02339102981264e+00, 1.01860324787071e+00, 1.01292012202806e+00, 1.00627949176102e+00, 9.98633339280488e-01, 9.89950435852393e-01, 9.80218460319847e-01, 9.69445444466402e-01, 9.57660436252084e-01, 9.44913320541602e-01, 9.31273793422676e-01, 9.16829544845442e-01, 9.01683758754203e-01, 8.85952084143611e-01, 8.69759259876119e-01, 8.53235588031993e-01, 8.36513444810009e-01, 8.19723996652542e-01, 8.02994256240595e-01, 7.86444573216841e-01, 7.70186613033772e-01, 7.54321838649531e-01, 7.38940477212324e-01, 7.24120929285320e-01, 7.09929562118106e-01, 6.96420820405592e-01, 6.83637586597314e-01, 6.71611726478716e-01, 6.60364762754071e-01, 6.49908628234229e-01, 6.40246459779007e-01, 6.31373403515752e-01, 6.23277410492662e-01, 6.15940009510595e-01, 6.09337050247859e-01, 6.03439414905111e-01, 5.98213700451225e-01, 5.93622876182764e-01, 5.89626922769980e-01, 5.86183459306412e-01, 5.83248364193381e-01, 5.80776394079109e-01, 5.78721802679539e-01, 5.77038958327535e-01, 5.75682955755696e-01, 5.74610214191210e-01, 5.73779050624552e-01, 5.73150214408628e-01, 5.72687367435890e-01, 5.72357493260569e-01, 5.72131218841562e-01, 5.71983034138242e-01, 5.71891397551369e-01, 5.71838719002287e-01, 5.71811217035219e-01, 5.71798651383759e-01, 5.71793937610380e-01, 5.71792655353150e-01, 5.71792466094806e-01, 5.71792451556377e-01, 5.71792191257956e-01, 5.71790427699682e-01, 5.71783944702418e-01, 5.71766663638144e-01, 5.71728845599509e-01, 5.71656423929298e-01, 5.71530495798072e-01, 5.71326999397044e-01, 5.71016600741601e-01, 5.70564810731042e-01, 5.69932348571996e-01, 5.69075761497899e-01, 5.67948302530340e-01, 5.66501057611504e-01, 5.64684300855717e-01, 5.62449042345634e-01, 5.59748717692287e-01, 5.56540953756141e-01, 5.52789332101526e-01, 5.48465062703059e-01, 5.43548476828718e-01, 5.38030251201079e-01, 5.31912286122520e-01, 5.25208177998593e-01, 5.17943250395589e-01, 5.10154135341182e-01, 5.01887925325718e-01, 4.93200943479721e-01, 4.84157202003189e-01, 4.74826635050391e-01, 4.65283200766535e-01, 4.55602947869565e-01, 4.45862135806351e-01, 4.36135485500174e-01, 4.26494621828283e-01, 4.17006751090355e-01, 4.07733598565144e-01, 3.98730614234339e-01, 3.90046439934522e-01, 3.81722619260856e-01, 3.73793522801861e-01, 3.66286455744515e-01, 3.59221912335542e-01, 3.52613941736394e-01, 3.46470591990041e-01, 3.40794402581665e-01, 3.35582920863331e-01, 3.30829222893911e-01, 3.26522424553829e-01, 3.22648173774951e-01, 3.19189119109282e-01, 3.16125353476295e-01, 3.13434834675519e-01, 3.11093786082338e-01, 3.09077081875245e-01, 3.07358621225717e-01, 3.05911695211876e-01, 3.04709348918621e-01, 3.03724739413184e-01, 3.02931488205758e-01, 3.02304024616404e-01, 3.01817914369181e-01, 3.01450165927449e-01, 3.01179505752202e-01, 3.00986612962959e-01, 3.00854303907528e-01, 3.00767657945869e-01, 3.00714077295723e-01, 3.00683275975089e-01, 3.00667195557509e-01, 3.00659848429788e-01, 3.00657092292425e-01, 3.00656342558497e-01, 3.00656231899438e-01, 3.00656223333959e-01, 3.00656069131833e-01, 3.00655024391034e-01, 3.00651183833182e-01, 3.00640946467809e-01, 3.00618543014664e-01, 3.00575640674225e-01, 3.00501042383075e-01, 3.00380496514393e-01, 3.00196631550321e-01, 2.99929028349382e-01, 2.99554440002495e-01, 2.99047165629460e-01, 2.98379579578040e-01, 2.97522811206292e-01, 2.96447562785460e-01, 2.95125044302424e-01, 2.93527994571065e-01, 2.91631748844665e-01, 2.89415305018110e-01, 2.86862334611428e-01, 2.83962082083079e-01, 2.80710097497391e-01, 2.77108753655511e-01, 2.73167509472444e-01, 2.68902896045518e-01, 2.64338219333879e-01, 2.59502992040495e-01, 2.54432125289284e-01, 2.49164926182871e-01, 2.43743958756284e-01, 2.38213832155125e-01, 2.32619980653169e-01, 2.27007495639981e-01, 2.21420060811539e-01, 2.15899029798507e-01, 2.10482671896344e-01, 2.05205597934295e-01, 2.00098365907621e-01, 1.95187255708078e-01, 1.90494194621678e-01, 1.86036810363873e-01, 1.81828586134358e-01, 1.77879092124726e-01, 1.74194269589194e-01, 1.70776746436289e-01, 1.67626166807111e-01, 1.64739520862327e-01, 1.62111464710577e-01, 1.59734623860743e-01, 1.57599876622811e-01, 1.55696616413522e-01, 1.54012993890873e-01, 1.52536141239722e-01, 1.51252381781006e-01, 1.50147428424162e-01, 1.49206574375654e-01, 1.48414879010076e-01, 1.47757350970224e-01, 1.47219129466563e-01, 1.46785663486270e-01, 1.46442887304573e-01, 1.46177389425233e-01, 1.45976570979045e-01, 1.45828788779056e-01, 1.45723477759248e-01, 1.45651247468024e-01, 1.45603947674118e-01, 1.45574698957674e-01, 1.45557885350527e-01, 1.45549107567575e-01, 1.45545097025405e-01, 1.45543592548677e-01, 1.45543183296027e-01, 1.45543122891254e-01, 1.45543118177770e-01, 1.45543032832322e-01, 1.45542454605263e-01, 1.45540328994741e-01, 1.45534663000021e-01, 1.45522263629360e-01, 1.45498519344358e-01, 1.45457234069966e-01, 1.45390522831889e-01, 1.45288777374105e-01, 1.45140709129521e-01, 1.44933475493360e-01, 1.44652893304951e-01, 1.44283740620395e-01, 1.43810144164689e-01, 1.43216045298402e-01, 1.42485732075897e-01, 1.41604419330392e-01, 1.40558853175804e-01, 1.39337911489264e-01, 1.37933168514607e-01, 1.36339390369997e-01, 1.34554929469459e-01, 1.32581989946289e-01, 1.30426743021545e-01, 1.28099280447913e-01, 1.25613404883168e-01, 1.22986267253168e-01, 1.20237871680601e-01, 1.17390477249614e-01, 1.14467931852786e-01, 1.11494976093411e-01, 1.08496554618355e-01, 1.05497168677203e-01, 1.02520297798748e-01, 9.95879110513158e-02, 9.67200802590611e-02, 9.39346995977814e-02, 9.12473089136512e-02, 8.86710124396757e-02, 8.62164806013652e-02, 8.38920202762239e-02, 8.17036979855317e-02, 7.96555007685384e-02, 7.77495206851597e-02, 7.59861507928861e-02, 7.43642827862772e-02, 7.28814989624311e-02, 7.15342534999621e-02, 7.03180400444268e-02, 6.92275442508645e-02, 6.82567812815904e-02, 6.73992193294026e-02, 6.66478910198549e-02, 6.59954950151729e-02, 6.54344902907127e-02, 6.49571854213349e-02, 6.45558248548781e-02, 6.42226736290326e-02, 6.39501013552746e-02, 6.37306655913505e-02, 6.35571939969992e-02, 6.34228639648740e-02, 6.33212777984369e-02, 6.32465310284118e-02, 6.31932711642804e-02, 6.31567441076015e-02, 6.31328256187942e-02, 6.31180356259620e-02, 6.31095337645604e-02, 6.31050952923857e-02, 6.31030673784352e-02, 6.31023066476928e-02, 6.31020997114433e-02, 6.31020691680959e-02, 6.31020667644862e-02, 6.31020229830881e-02, 6.31017263580844e-02, 6.31006359415096e-02, 6.30977293545243e-02, 6.30913686808725e-02, 6.30791884595823e-02, 6.30580108001055e-02, 6.30237923997896e-02, 6.29716079304643e-02, 6.28956737977830e-02, 6.27894155682611e-02, 6.26455812925997e-02, 6.24564014389080e-02, 6.22137941280698e-02, 6.19096118421680e-02, 6.15359228391554e-02, 6.10853173472208e-02, 6.05512255033887e-02, 5.99282313184375e-02, 5.92123650883138e-02, 5.84013560265362e-02, 5.74948277564857e-02, 5.64944218207586e-02, 5.54038384637607e-02, 5.42287893228206e-02, 5.29768628163652e-02, 5.16573093003245e-02, 5.02807588008096e-02, 4.88588886946957e-02, 4.74040616304165e-02, 4.59289549772421e-02, 4.44462021365695e-02, 4.29680633636451e-02, 4.15061398046106e-02, 4.00711398725235e-02, 3.86727024897340e-02, 3.73192775851834e-02, 3.60180608094254e-02, 3.47749767849876e-02, 3.35947033778723e-02, 3.24807284874709e-02, 3.14354307170390e-02, 3.04601758844996e-02, 2.95554223941394e-02, 2.87208297024364e-02, 2.79553652744937e-02, 2.72574065095483e-02, 2.66248351703524e-02, 2.60551228963806e-02, 2.55454073393082e-02, 2.50925592192601e-02, 2.46932411073573e-02, 2.43439590316825e-02, 2.40411081549325e-02, 2.37810138434753e-02, 2.35599694412514e-02, 2.33742719591277e-02, 2.32202566725732e-02, 2.30943313023744e-02, 2.29930100721560e-02, 2.29129475337544e-02, 2.28509716700529e-02, 2.28041154513789e-02, 2.27696457602349e-02, 2.27450884315914e-02, 2.27282480931734e-02, 2.27172215452782e-02, 2.27104035866006e-02, 2.27064844651892e-02, 2.27044384837001e-02, 2.27035036928243e-02, 2.27031530261286e-02, 2.27030576368463e-02, 2.27030435575983e-02, 2.27030424398151e-02, 2.27030219548681e-02, 2.27028831666587e-02, 2.27023729713238e-02, 2.27010130145970e-02, 2.26980369689045e-02, 2.26923381995192e-02, 2.26824302259843e-02, 2.26664223038142e-02, 2.26420123371298e-02, 2.26064991457235e-02, 2.25568157916314e-02, 2.24895851626883e-02, 2.24011982615622e-02, 2.22879146290040e-02, 2.21459830412272e-02, 2.19717791121212e-02, 2.17619548029404e-02, 2.15135932362358e-02, 2.12243608195877e-02, 2.08926477066456e-02, 2.05176872729282e-02, 2.00996457362663e-02, 1.96396744172205e-02, 1.91399193933677e-02, 1.86034862552447e-02, 1.80343609587794e-02, 1.74372909507241e-02, 1.68176334709074e-02, 1.61811800495286e-02, 1.55339677008044e-02, 1.48820880837235e-02, 1.42315056875191e-02, 1.35878946198744e-02, 1.29565009496417e-02, 1.23420345234422e-02, 1.17485917213713e-02, 1.11796091714211e-02, 1.06378474599467e-02, 1.01254023894910e-02, 9.64373913089389e-03, 9.19374271663287e-03, 8.77577811807839e-03, 8.38975491409144e-03, 8.03519393963370e-03, 7.71129446741270e-03, 7.41699986883759e-03, 7.15105852380304e-03, 6.91207667370008e-03, 6.69856151563408e-03, 6.50895518944594e-03, 6.34166186524965e-03, 6.19507010887332e-03, 6.06757165972738e-03, 5.95757681664111e-03, 5.86352638757148e-03, 5.78390051333599e-03, 5.71722521029017e-03, 5.66207778791750e-03, 5.61709219176976e-03, 5.58096491391780e-03, 5.55246164324785e-03, 5.53042448271412e-03, 5.51377936847075e-03, 5.50154323231311e-03, 5.49283040780324e-03, 5.48685776077207e-03, 5.48294803665783e-03, 5.48053095931778e-03, 5.47914170269337e-03, 5.47841647924940e-03, 5.47808514004089e-03, 5.47796084674701e-03, 5.47792703625305e-03, 5.47792204594089e-03, 5.47792164604743e-03, 5.47791427132438e-03, 5.47786430662711e-03, 5.47768063407261e-03, 5.47719105253464e-03, 5.47611972743312e-03, 5.47406844124509e-03, 5.47050258070669e-03, 5.46474280527320e-03, 5.45596333894593e-03, 5.44319778781244e-03, 5.42535327438998e-03, 5.40123346830512e-03, 5.36957074901068e-03, 5.32906726627402e-03, 5.27844407535838e-03, 5.21649686831438e-03, 5.14215614512636e-03, 5.05454900806512e-03, 4.95305912893684e-03, 4.83738080758656e-03, 4.70756244887127e-03, 4.56403448427523e-03, 4.40761734515149e-03, 4.23950712253429e-03, 4.06123976048889e-03, 3.87463760488060e-03, 3.68174348689365e-03, 3.48474714495239e-03, 3.28590705265697e-03, 3.08746840424590e-03, 2.89157868476764e-03, 2.70020840193297e-03, 2.51508880545259e-03, 2.33767365917004e-03, 2.16912633023014e-03, 2.01032715045812e-03, 1.86188435546342e-03, 1.72414208192606e-03, 1.59720056477397e-03, 1.48094270447530e-03, 1.37506653616981e-03, 1.27915383828281e-03, 1.19271087419293e-03, 1.11516143147128e-03, 1.04591606665849e-03, 9.84428796726888e-04, 9.29993817990749e-04, 8.81915920231791e-04, 8.39810746910530e-04, 8.03146413829836e-04, 7.71228862120665e-04, 7.43547105102493e-04, 7.19680780662202e-04, 6.99488415136784e-04, 6.82561641866641e-04, 6.68306426064191e-04, 6.56392806050121e-04, 6.46543754371010e-04, 6.38509454077507e-04, 6.32058170594973e-04, 6.26971565267809e-04, 6.23043498293896e-04, 6.20081145640938e-04, 6.17907115923799e-04, 6.16361517984387e-04, 6.15303403435519e-04, 6.14611446778213e-04, 6.14183958788317e-04, 6.13938359578015e-04, 6.13810183025976e-04, 6.13751629288117e-04, 6.13729665602136e-04, 6.13723691088808e-04, 6.13722809248429e-04, 6.13722738123101e-04, 6.13721421089508e-04, 6.13712497795404e-04, 6.13679694836833e-04, 6.13592254637751e-04, 6.13400895787208e-04, 6.13034429548520e-04, 6.12397175373758e-04, 6.11367321937450e-04, 6.09796396730644e-04, 6.07510061072958e-04, 6.04310524636276e-04, 5.99980955882984e-04, 5.94293179119632e-04, 5.86906821922639e-04, 5.78320164223101e-04, 5.67685699806475e-04, 5.54968163896774e-04, 5.40067257233545e-04, 5.22942265021697e-04, 5.03617865348360e-04, 4.82187454423411e-04, 4.59136551967276e-04, 4.34638008329102e-04, 4.08637417052660e-04, 3.81490628981851e-04, 3.53724722543802e-04, 3.26228073490142e-04, 2.98916415943040e-04, 2.72163167522876e-04, 2.46540200819823e-04, 2.22201732356098e-04, 1.99109333948400e-04, 1.77675021348023e-04, 1.58145767684914e-04, 1.40347894983961e-04, 1.24243339030272e-04, 1.09601051377695e-04, 9.61639064911424e-05, 8.43272897412817e-05, 7.37346816350737e-05, 6.41380062802455e-05, 5.55350141127064e-05, 4.73354895262901e-05, 3.88241686550828e-05, 3.04507758386308e-05, 2.24605418281953e-05, 1.49360437688983e-05, 8.39728906498749e-06, 3.09379731788795e-06, 2.65235165386457e-06, 2.07705465982144e-06, 5.65860219126080e-07, 1.84508446920972e-06, 2.21213603879278e-06, 2.08054808581437e-06, 1.76648804028900e-06, 1.48190831339434e-06, 1.31752705101462e-06, 1.25320576671584e-06, 1.21940712460855e-06, 1.16117051618403e-06, 1.05840504716070e-06, 9.16865275257805e-07, 7.53982234869000e-07, 5.88640673981899e-07, 4.35849671021060e-07, 3.05014694390593e-07, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14, -4.26325641456060e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14],
[ 9.40000000000000e+01, 9.39999998856023e+01, 9.39999963481977e+01, 9.39999723817199e+01, 9.39998842777860e+01, 9.39996494202370e+01, 9.39991354158362e+01, 9.39981509424296e+01, 9.39964386424139e+01, 9.39936703252637e+01, 9.39894446706699e+01, 9.39832875467883e+01, 9.39746549800626e+01, 9.39629387382924e+01, 9.39474744212050e+01, 9.39275518962149e+01, 9.39024278733262e+01, 9.38713403820890e+01, 9.38335248922717e+01, 9.37882318027968e+01, 9.37347450026949e+01, 9.36724011749229e+01, 9.36006094619686e+01, 9.35188710385062e+01, 9.34267980446531e+01, 9.33241312349505e+01, 9.32107556114031e+01, 9.30867132565436e+01, 9.29522125871604e+01, 9.28076333281403e+01, 9.26535266653463e+01, 9.24906102694614e+01, 9.23197581683197e+01, 9.21419857517768e+01, 9.19584304840195e+01, 9.17703291387727e+01, 9.15789925369542e+01, 9.13857788406174e+01, 9.11920664423432e+01, 9.09992273989488e+01, 9.08086022143670e+01, 9.06214766041619e+01, 9.04390606973628e+01, 9.02624709690767e+01, 9.00927150617203e+01, 8.99306795486852e+01, 8.97771206208553e+01, 8.96326576286728e+01, 8.94977693834419e+01, 8.93727931040486e+01, 8.92579258830205e+01, 8.91532285341755e+01, 8.90586316700700e+01, 8.89739438396531e+01, 8.88988615347761e+01, 8.88329808491301e+01, 8.87758105459741e+01, 8.87267862631458e+01, 8.86852855569707e+01, 8.86506434625169e+01, 8.86221682279394e+01, 8.85991568671282e+01, 8.85809101691376e+01, 8.85667468063378e+01, 8.85560161970318e+01, 8.85481098031038e+01, 8.85424705793158e+01, 8.85386003377191e+01, 8.85360648472693e+01, 8.85344965534574e+01, 8.85335948734022e+01, 8.85331240957599e+01, 8.85329089891385e+01, 8.85328282945021e+01, 8.85328063435367e+01, 8.85328031036178e+01, 8.85328026786088e+01, 8.85327927764085e+01, 8.85327256875317e+01, 8.85324790615282e+01, 8.85318216456489e+01, 8.85303828994501e+01, 8.85276274991307e+01, 8.85228357433588e+01, 8.85150907463259e+01, 8.85032731577576e+01, 8.84860639808798e+01, 8.84619558637957e+01, 8.84292730129797e+01, 8.83861996162280e+01, 8.83308163654552e+01, 8.82611443401995e+01, 8.81751951591000e+01, 8.80710259443684e+01, 8.79467972969194e+01, 8.78008321794061e+01, 8.76316733903215e+01, 8.74381372270784e+01, 8.72193610169249e+01, 8.69748424614619e+01, 8.67044691818250e+01, 8.64085374146198e+01, 8.60877594018745e+01, 8.57432595311557e+01, 8.53765596235899e+01, 8.49895539126725e+01, 8.45844742793826e+01, 8.41638463755948e+01, 8.37304375706004e+01, 8.32871982962523e+01, 8.28371992411181e+01, 8.23835676217497e+01, 8.19294260168141e+01, 8.14778367263725e+01, 8.10317534193661e+01, 8.05939804077351e+01, 8.01671387736644e+01, 7.97536381002281e+01, 7.93556526940057e+01, 7.89751016733675e+01, 7.86136328153988e+01, 7.82726104023607e+01, 7.79531074265904e+01, 7.76559024467496e+01, 7.73814812236417e+01, 7.71300430774010e+01, 7.69015117478073e+01, 7.66955504247132e+01, 7.65115805446097e+01, 7.63488039104389e+01, 7.62062276705739e+01, 7.60826916771229e+01, 7.59768977250835e+01, 7.58874401483946e+01, 7.58128372164516e+01, 7.57515627379343e+01, 7.57020772425946e+01, 7.56628580818594e+01, 7.56324277719002e+01, 7.56093799040040e+01, 7.55924019714176e+01, 7.55802945125538e+01, 7.55719860488307e+01, 7.55665434006179e+01, 7.55631770937842e+01, 7.55612417171967e+01, 7.55602312517047e+01, 7.55597695560822e+01, 7.55595963572098e+01, 7.55595492428316e+01, 7.55595422888450e+01, 7.55595416749403e+01, 7.55595296457809e+01, 7.55594481464432e+01, 7.55591485460666e+01, 7.55583499200011e+01, 7.55566021389827e+01, 7.55532549008546e+01, 7.55474339368182e+01, 7.55380254728068e+01, 7.55236698511348e+01, 7.55027650145400e+01, 7.54734803190163e+01, 7.54337808681078e+01, 7.53814622458367e+01, 7.53141951678668e+01, 7.52295791758592e+01, 7.51252040800652e+01, 7.49987174291990e+01, 7.48478958806857e+01, 7.46707179897235e+01, 7.44654356656120e+01, 7.42306413902954e+01, 7.39653282837021e+01, 7.36689402520250e+01, 7.33414097778995e+01, 7.29831814043165e+01, 7.25952196137389e+01, 7.21790005815520e+01, 7.17364881402403e+01, 7.12700951568989e+01, 7.07826323125254e+01, 7.02772468820135e+01, 6.97573544699413e+01, 6.92265667204910e+01, 6.86886178084157e+01, 6.81472921067298e+01, 6.76063549266441e+01, 6.70694877483558e+01, 6.65402289852855e+01, 6.60219210708108e+01, 6.55176644930484e+01, 6.50302792676163e+01, 6.45622741730569e+01, 6.41158238534543e+01, 6.36927536331854e+01, 6.32945316335680e+01, 6.29222675790174e+01, 6.25767175630895e+01, 6.22582940178699e+01, 6.19670801757492e+01, 6.17028484007008e+01, 6.14650818662193e+01, 6.12529991456424e+01, 6.10655813438362e+01, 6.09016014317948e+01, 6.07596554480126e+01, 6.06381952063293e+01, 6.05355621047339e+01, 6.04500215696981e+01, 6.03797976028819e+01, 6.03231068289773e+01, 6.02781913827950e+01, 6.02433499283005e+01, 6.02169660794898e+01, 6.01975334990114e+01, 6.01836769897646e+01, 6.01741689695193e+01, 6.01679408283824e+01, 6.01640888103807e+01, 6.01618742275694e+01, 6.01607179998142e+01, 6.01601897061617e+01, 6.01599915243847e+01, 6.01599376140805e+01, 6.01599296570305e+01, 6.01599290187767e+01, 6.01599172397562e+01, 6.01598374351554e+01, 6.01595440649494e+01, 6.01587620472934e+01, 6.01570506187623e+01, 6.01537730206477e+01, 6.01480732257973e+01, 6.01388607768723e+01, 6.01248046400010e+01, 6.01043367831921e+01, 6.00756659603010e+01, 6.00368019122791e+01, 5.99855898829214e+01, 5.99197549850870e+01, 5.98369555498971e+01, 5.97348441577508e+01, 5.96111346066602e+01, 5.94636726494119e+01, 5.92905079623541e+01, 5.90899645350766e+01, 5.88607065310502e+01, 5.86017966970117e+01, 5.83127446130449e+01, 5.79935424765160e+01, 5.76446866791125e+01, 5.72671841225609e+01, 5.68625429636498e+01, 5.64327482157983e+01, 5.59802233038590e+01, 5.55077792330381e+01, 5.50185534786782e+01, 5.45159410369075e+01, 5.40035203069158e+01, 5.34849766036456e+01, 5.29640261083306e+01, 5.24443429274167e+01, 5.19294916290449e+01, 5.14228671662290e+01, 5.09276435162308e+01, 5.04467317339497e+01, 4.99827475131799e+01, 4.95379878448663e+01, 4.91144160027418e+01, 4.87136538883179e+01, 4.83369807137354e+01, 4.79853370570335e+01, 4.76593334468575e+01, 4.73592627825426e+01, 4.70851160406163e+01, 4.68366008416196e+01, 4.66131625436230e+01, 4.64140075898413e+01, 4.62381288699709e+01, 4.60843328620104e+01, 4.59512683069655e+01, 4.58374561360605e+01, 4.57413203219495e+01, 4.56612192654830e+01, 4.55954772624559e+01, 4.55424155262346e+01, 4.55003821791038e+01, 4.54677805750627e+01, 4.54430952870312e+01, 4.54249150885563e+01, 4.54119522890677e+01, 4.54030578451946e+01, 4.53972317685732e+01, 4.53936284800594e+01, 4.53915569156979e+01, 4.53904753633359e+01, 4.53899811910566e+01, 4.53897958096387e+01, 4.53897453813665e+01, 4.53897379382581e+01, 4.53897373641739e+01, 4.53897270554654e+01, 4.53896572124490e+01, 4.53894004621432e+01, 4.53887160601759e+01, 4.53872182636265e+01, 4.53843498056431e+01, 4.53793615381107e+01, 4.53712991848970e+01, 4.53589980042048e+01, 4.53410859900364e+01, 4.53159960437914e+01, 4.52819873113638e+01, 4.52371756040120e+01, 4.51795725014022e+01, 4.51071323765854e+01, 4.50178061962440e+01, 4.49096005541721e+01, 4.47806400181844e+01, 4.46292305431422e+01, 4.44539214612644e+01, 4.42535634395827e+01, 4.40273598205922e+01, 4.37749089506169e+01, 4.34962354489765e+01, 4.31918088586730e+01, 4.28625487079074e+01, 4.25098156521521e+01, 4.21353890080313e+01, 4.17414315909208e+01, 4.13304433021477e+01, 4.09052053694615e+01, 4.04687175244290e+01, 4.00241306941025e+01, 3.95746779641060e+01, 3.91236065895510e+01, 3.86741136416721e+01, 3.82292874645006e+01, 3.77920565156498e+01, 3.73651464739494e+01, 3.69510458385361e+01, 3.65519797267807e+01, 3.61698912562029e+01, 3.58064297589040e+01, 3.54629450716092e+01, 3.51404872025732e+01, 3.48398107449960e+01, 3.45613834582712e+01, 3.43053984703718e+01, 3.40717895774707e+01, 3.38602491433613e+01, 3.36702481393919e+01, 3.35010579160624e+01, 3.33517733545506e+01, 3.32213371012035e+01, 3.31085646311070e+01, 3.30121699108463e+01, 3.29307914315109e+01, 3.28630183606828e+01, 3.28074165199921e+01, 3.27625538391712e+01, 3.27270248767335e+01, 3.26994739407993e+01, 3.26786163004819e+01, 3.26632569568817e+01, 3.26523064495077e+01, 3.26447932126487e+01, 3.26398720677604e+01, 3.26368285401400e+01, 3.26350788160761e+01, 3.26341653030377e+01, 3.26337479115249e+01, 3.26335913335811e+01, 3.26335487405861e+01, 3.26335424539493e+01, 3.26335419779441e+01, 3.26335335455605e+01, 3.26334764149313e+01, 3.26332663968581e+01, 3.26327065670358e+01, 3.26314813995198e+01, 3.26291350815242e+01, 3.26250548911293e+01, 3.26184604236988e+01, 3.26083993410278e+01, 3.25937501850855e+01, 3.25732326398580e+01, 3.25454254341680e+01, 3.25087918507405e+01, 3.24617125404800e+01, 3.24025250388303e+01, 3.23295690519339e+01, 3.22412362392751e+01, 3.21360228879288e+01, 3.20125835776570e+01, 3.18697837046193e+01, 3.17067485923889e+01, 3.15229068960426e+01, 3.13180261143958e+01, 3.10922382730732e+01, 3.08460542220144e+01, 3.05803654896641e+01, 3.02964332278078e+01, 2.99958644338070e+01, 2.96805763127471e+01, 2.93527502970706e+01, 2.90147778269772e+01, 2.86692004597104e+01, 2.83186471703232e+01, 2.79657717911955e+01, 2.76131933926195e+01, 2.72634420400910e+01, 2.69189118143376e+01, 2.65818223134563e+01, 2.62541891549013e+01, 2.59378033417260e+01, 2.56342188216243e+01, 2.53447471928784e+01, 2.50704583123385e+01, 2.48121855234733e+01, 2.45705343137807e+01, 2.43458933869201e+01, 2.41384473520443e+01, 2.39481904540468e+01, 2.37749409675565e+01, 2.36183560394524e+01, 2.34779468839124e+01, 2.33530943116665e+01, 2.32430646160753e+01, 2.31470258491652e+01, 2.30640645070137e+01, 2.29932026113621e+01, 2.29334151277462e+01, 2.28836476040394e+01, 2.28428338512983e+01, 2.28099134256342e+01, 2.27838486103342e+01, 2.27636405467122e+01, 2.27483441252591e+01, 2.27370812302107e+01, 2.27290519343069e+01, 2.27235432685086e+01, 2.27199352441422e+01, 2.27177038807396e+01, 2.27164210880962e+01, 2.27157513604534e+01, 2.27154453576685e+01, 2.27153305656708e+01, 2.27152993394841e+01, 2.27152947305651e+01, 2.27152943845721e+01, 2.27152882947034e+01, 2.27152470349615e+01, 2.27150953600075e+01, 2.27146910519698e+01, 2.27138062423559e+01, 2.27121117591875e+01, 2.27091651422702e+01, 2.27044029037943e+01, 2.26971375341791e+01, 2.26865596600683e+01, 2.26717456471141e+01, 2.26516707987822e+01, 2.26252281290259e+01, 2.25912524790908e+01, 2.25485495089425e+01, 2.24959288295000e+01, 2.24322402671451e+01, 2.23564119874477e+01, 2.22674889765888e+01, 2.21646702152582e+01, 2.20473428086067e+01, 2.19151113793573e+01, 2.17678212012795e+01, 2.16055738443438e+01, 2.14287345018240e+01, 2.12379306382582e+01, 2.10340420884749e+01, 2.08181832002387e+01, 2.05916779997205e+01, 2.03560296374499e+01, 2.01128855310740e+01, 1.98639996708468e+01, 1.96111935222128e+01, 1.93563168813920e+01, 1.91012099419777e+01, 1.88476677235273e+01, 1.85974078874800e+01, 1.83520427996032e+01, 1.81130564722151e+01, 1.78817867322764e+01, 1.76594126374775e+01, 1.74469468470689e+01, 1.72452323983440e+01, 1.70549431815375e+01, 1.68765873584309e+01, 1.67105130196232e+01, 1.65569154906076e+01, 1.64158458404613e+01, 1.62872202875425e+01, 1.61708303137364e+01, 1.60663533837920e+01, 1.59733642190909e+01, 1.58913466004845e+01, 1.58197056787482e+01, 1.57577807593018e+01, 1.57048585044754e+01, 1.56601864649577e+01, 1.56229868147593e+01, 1.55924701235514e+01, 1.55678489593925e+01, 1.55483510768286e+01, 1.55332319136619e+01, 1.55217860978943e+01, 1.55133576577503e+01, 1.55073486348211e+01, 1.55032258247749e+01, 1.55005254119292e+01, 1.54988553220329e+01, 1.54978951891971e+01, 1.54973939142017e+01, 1.54971648777153e+01, 1.54970789582542e+01, 1.54970555860806e+01, 1.54970521363968e+01, 1.54970518782155e+01, 1.54970473444215e+01, 1.54970166273059e+01, 1.54969037080248e+01, 1.54966027073349e+01, 1.54959439780113e+01, 1.54946824452867e+01, 1.54924886691308e+01, 1.54889430505285e+01, 1.54835335434964e+01, 1.54756571629022e+01, 1.54646254918524e+01, 1.54496742879290e+01, 1.54299771620328e+01, 1.54046631559177e+01, 1.53728378763411e+01, 1.53336076604273e+01, 1.52861060578464e+01, 1.52295217343825e+01, 1.51631267453605e+01, 1.50863040148764e+01, 1.49985728056930e+01, 1.48996109891011e+01, 1.47892730311995e+01, 1.46676027996640e+01, 1.45348405504015e+01, 1.43914237540492e+01, 1.42379817392329e+01, 1.40753244326398e+01, 1.39044257396773e+01, 1.37264023174252e+01, 1.35424886387362e+01, 1.33540093373745e+01, 1.31623498688143e+01, 1.29689265291380e+01, 1.27751568501652e+01, 1.25824313317562e+01, 1.23920873781526e+01, 1.22053861712989e+01, 1.20234930424784e+01, 1.18474617036418e+01, 1.16782224873210e+01, 1.15165745383311e+01, 1.13631817203903e+01, 1.12185718610073e+01, 1.10831388667443e+01, 1.09571471994905e+01, 1.08407382078204e+01, 1.07339378467273e+01, 1.06366653827688e+01, 1.05487427584393e+01, 1.04699043690998e+01, 1.03998070797344e+01, 1.03380403711599e+01, 1.02841365523528e+01, 1.02375810055234e+01, 1.01978224434010e+01, 1.01642831551987e+01, 1.01363692012772e+01, 1.01134804898466e+01, 1.00950206359854e+01, 1.00804064680225e+01, 1.00690770132748e+01, 1.00605017684165e+01, 1.00541880430294e+01, 1.00496871609907e+01, 1.00465993150409e+01, 1.00445768956101e+01, 1.00433261548821e+01, 1.00426071190024e+01, 1.00422317220688e+01, 1.00420602011588e+01, 1.00419958579137e+01, 1.00419783550070e+01, 1.00419757716152e+01, 1.00419755781965e+01, 1.00419721807011e+01, 1.00419491621795e+01, 1.00418645438108e+01, 1.00416389834730e+01, 1.00411453561040e+01, 1.00402000233004e+01, 1.00385561525498e+01, 1.00358994108682e+01, 1.00318463157216e+01, 1.00259454755536e+01, 1.00176818899069e+01, 1.00064844017150e+01, 9.99173629917362e+00, 9.97278895039156e+00, 9.94897822170377e+00, 9.91964328386113e+00, 9.88414725628314e+00, 9.84189898832958e+00, 9.79237514096653e+00, 9.73514162676470e+00, 9.66987340554851e+00, 9.59637163019006e+00, 9.51457720133791e+00, 9.42457992507114e+00, 9.32662267040638e+00, 9.22110018342876e+00, 9.10855251333982e+00, 8.98965331869563e+00, 8.86519362179544e+00, 8.73606183765935e+00, 8.60322109697732e+00, 8.46768499262940e+00, 8.33049289942147e+00, 8.19268594989278e+00, 8.05528460877051e+00, 7.91926859568039e+00, 7.78555968462511e+00, 7.65500768381794e+00, 7.52837969128908e+00, 7.40635254493138e+00, 7.28950824834453e+00, 7.17833205799162e+00, 7.07321286028393e+00, 6.97444544399860e+00, 6.88223427747531e+00, 6.79669842482647e+00, 6.71787727494515e+00, 6.64573680610648e+00, 6.58017616278953e+00, 6.52103437590581e+00, 6.46809710934127e+00, 6.42110336156301e+00, 6.37975208860549e+00, 6.34370874233001e+00, 6.31261173452029e+00, 6.28607884299973e+00, 6.26371357117425e+00, 6.24511145856970e+00, 6.22986631902252e+00, 6.21757635768256e+00, 6.20785009072840e+00, 6.20031196568665e+00, 6.19460755843133e+00, 6.19040820797589e+00, 6.18741494417768e+00, 6.18536156781819e+00, 6.18401675767242e+00, 6.18318510461106e+00, 6.18270700705416e+00, 6.18245740295225e+00, 6.18234335812588e+00, 6.18230057619250e+00, 6.18228893848995e+00, 6.18228722079016e+00, 6.18228709184476e+00, 6.18228482229746e+00, 6.18226944579169e+00, 6.18221292029699e+00, 6.18206224536515e+00, 6.18173250347586e+00, 6.18110103327085e+00, 6.18000297939756e+00, 6.17822844214824e+00, 6.17552142447792e+00, 6.17158074223563e+00, 6.16606302284357e+00, 6.15858786535756e+00, 6.14874516844965e+00, 6.13610455111224e+00, 6.12022669434310e+00, 6.10067632365363e+00, 6.07703643755140e+00, 6.04892327446206e+00, 6.01600141034336e+00, 5.97799830322272e+00, 5.93471756064251e+00, 5.88605021132206e+00, 5.83198331955030e+00, 5.77260539132158e+00, 5.70810818067552e+00, 5.63878470294793e+00, 5.56502348352481e+00, 5.48729929761974e+00, 5.40616086861620e+00, 5.32221617057260e+00, 5.23611610856031e+00, 5.14853741739034e+00, 5.06016561991792e+00, 4.97167882232728e+00, 4.88373300435778e+00, 4.79694930219317e+00, 4.71190359976964e+00, 4.62911856108391e+00, 4.54905807046931e+00, 4.47212391392557e+00, 4.39865444023201e+00, 4.32892488691738e+00, 4.26314903890470e+00, 4.20148189908380e+00, 4.14402308125158e+00, 4.09082067842169e+00, 4.04187540666121e+00, 3.99714487150098e+00, 3.95654784747730e+00, 3.91996849968992e+00, 3.88726050847107e+00, 3.85825108390421e+00, 3.83274487584389e+00, 3.81052779726504e+00, 3.79137078436561e+00, 3.77503351619099e+00, 3.76126811021007e+00, 3.74982279907709e+00, 3.74044557888055e+00, 3.73288780189341e+00, 3.72690766880682e+00, 3.72227355838264e+00, 3.71876711814650e+00, 3.71618602975829e+00, 3.71434635835045e+00, 3.71308439727575e+00, 3.71225792867199e+00, 3.71174683573365e+00, 3.71145302369514e+00, 3.71129963187200e+00, 3.71122954691761e+00, 3.71120325579928e+00, 3.71119610399346e+00, 3.71119504840249e+00, 3.71119496882322e+00, 3.71119356366565e+00, 3.71118404352477e+00, 3.71114904660612e+00, 3.71105575864027e+00, 3.71085160513114e+00, 3.71046064560007e+00, 3.70978082271718e+00, 3.70868220665695e+00, 3.70700635996955e+00, 3.70456692925532e+00, 3.70115154505443e+00, 3.69652507859969e+00, 3.69043426171029e+00, 3.68261362319855e+00, 3.67279263205768e+00, 3.66070386640801e+00, 3.64609195164829e+00, 3.62872293732916e+00, 3.60839371734900e+00, 3.58494105039476e+00, 3.55824971511352e+00, 3.52825934378681e+00, 3.49496952298507e+00, 3.45844282968076e+00, 3.41880558214828e+00, 3.37624621809857e+00, 3.33101135611446e+00, 3.28339973716743e+00, 3.23375436764637e+00, 3.18245328282494e+00, 3.12989941240889e+00, 3.07651005430725e+00, 3.02270644983116e+00, 2.96890390750848e+00, 2.91550285060613e+00, 2.86288107378425e+00, 2.81138739605150e+00, 2.76133679891916e+00, 2.71300704782745e+00, 2.66663671740471e+00, 2.62242448086181e+00, 2.58052948274407e+00, 2.54107259235006e+00, 2.50413833071212e+00, 2.46977727422913e+00, 2.43800875922490e+00, 2.40882374002976e+00, 2.38218768499415e+00, 2.35804342699274e+00, 2.33631391498646e+00, 2.31690483931932e+00, 2.29970712455200e+00, 2.28459929925762e+00, 2.27144976225389e+00, 2.26011896947513e+00, 2.25046156558324e+00, 2.24232848013348e+00, 2.23556900043411e+00, 2.23003282306237e+00, 2.22557207431318e+00, 2.22204327772224e+00, 2.21930923532287e+00, 2.21724077954052e+00, 2.21571834558698e+00, 2.21463331068364e+00, 2.21388904695901e+00, 2.21340163962617e+00, 2.21310023088346e+00, 2.21292696235261e+00, 2.21283650391949e+00, 2.21279517348034e+00, 2.21277966912967e+00, 2.21277545158203e+00, 2.21277482908127e+00, 2.21277478189658e+00, 2.21277394535323e+00, 2.21276827765519e+00, 2.21274744269026e+00, 2.21269190503527e+00, 2.21257036584838e+00, 2.21233761760834e+00, 2.21193290978131e+00, 2.21127891062184e+00, 2.21028134199469e+00, 2.20882935332956e+00, 2.20679668629251e+00, 2.20404366209952e+00, 2.20041999742061e+00, 2.19576842185634e+00, 2.18992903016566e+00, 2.18274425705347e+00, 2.17406431392834e+00, 2.16375287944415e+00, 2.15169279372479e+00, 2.13779147542645e+00, 2.12198576660516e+00, 2.10424591718669e+00, 2.08457845133384e+00, 2.06302771229788e+00, 2.03967595758205e+00, 2.01464196665607e+00, 1.98807822082891e+00, 1.96016680955779e+00, 1.93111429966673e+00, 1.90114586517080e+00, 1.87049900971434e+00, 1.83941721850140e+00, 1.80814385320465e+00, 1.77691655624780e+00, 1.74596236714819e+00, 1.71549368165542e+00, 1.68570511257420e+00, 1.65677124655248e+00, 1.62884523885840e+00, 1.60205815097169e+00, 1.57651891407463e+00, 1.55231479376519e+00, 1.52951223481504e+00, 1.50815797629093e+00, 1.48828034362052e+00, 1.46989064243914e+00, 1.45298459720107e+00, 1.43754379419128e+00, 1.42353710296184e+00, 1.41092206204310e+00, 1.39964622403257e+00, 1.38964846197824e+00, 1.38086024353358e+00, 1.37320688183037e+00, 1.36660877254947e+00, 1.36098262540862e+00, 1.35624269540823e+00, 1.35230201491231e+00, 1.34907362231078e+00, 1.34647177702122e+00, 1.34441314443637e+00, 1.34281792865758e+00, 1.34161092604435e+00, 1.34072246928180e+00, 1.34008923026478e+00, 1.33965485091274e+00, 1.33937037417304e+00, 1.33919445284846e+00, 1.33909332119374e+00, 1.33904052299123e+00, 1.33901639942981e+00, 1.33900734990948e+00, 1.33900488822624e+00, 1.33900452488716e+00, 1.33900449717180e+00, 1.33900400349729e+00, 1.33900065878322e+00, 1.33898836330887e+00, 1.33895558846031e+00, 1.33888386339245e+00, 1.33874650853602e+00, 1.33850766990719e+00, 1.33812170390421e+00, 1.33753295712502e+00, 1.33667597917128e+00, 1.33547619768753e+00, 1.33385107327261e+00, 1.33171173673279e+00, 1.32896509208071e+00, 1.32551634583906e+00, 1.32127189735756e+00, 1.31614249749768e+00, 1.31004655642073e+00, 1.30291345814739e+00, 1.29468672314897e+00, 1.28532685343716e+00, 1.27481369973378e+00, 1.26314820844905e+00, 1.25035343694474e+00, 1.23647476678320e+00, 1.22157929269770e+00, 1.20575441514310e+00, 1.18910571144936e+00, 1.17175420023824e+00, 1.15383314255286e+00, 1.13548453944654e+00, 1.11685548974255e+00, 1.09809456493771e+00, 1.07934834331279e+00, 1.06075822494617e+00, 1.04245762582237e+00, 1.02456962417592e+00, 1.00720510650582e+00, 9.90461434837457e-01, 9.74421631332646e-01, 9.59154052172266e-01, 9.44712501157483e-01, 9.31136716368044e-01, 9.18453152014763e-01, 9.06675973271675e-01, 8.95808184404601e-01, 8.85842819013121e-01, 8.76764134010756e-01, 8.68548764055433e-01, 8.61166808523095e-01, 8.54582837158034e-01, 8.48756812126155e-01, 8.43644932780776e-01, 8.39200414920583e-01, 8.35374218901734e-01, 8.32115741070780e-01, 8.29373481110096e-01, 8.27095694537135e-01, 8.25231035263272e-01, 8.23729188252841e-01, 8.22541487359686e-01, 8.21621508763513e-01, 8.20925626457878e-01, 8.20413513292777e-01, 8.20048569417651e-01, 8.19798259787717e-01, 8.19634343753975e-01, 8.19532982596527e-01, 8.19474714989084e-01, 8.19444295495146e-01, 8.19430396907805e-01, 8.19425183121382e-01, 8.19423764850100e-01, 8.19423555516408e-01, 8.19423539435846e-01, 8.19423251526615e-01, 8.19421300901936e-01, 8.19414130240673e-01, 8.19395016207892e-01, 8.19353187225531e-01, 8.19273085892618e-01, 8.19133808315925e-01, 8.18908750569548e-01, 8.18565489979733e-01, 8.18065925508932e-01, 8.17366696657160e-01, 8.16419893432453e-01, 8.15174060569746e-01, 8.13575486992605e-01, 8.11569756540550e-01, 8.09103518695267e-01, 8.06126419407946e-01, 8.02593113669303e-01, 7.98465265135761e-01, 7.93713426175430e-01, 7.88318686398966e-01, 7.82273981053422e-01, 7.75584963850918e-01, 7.68270372109725e-01, 7.60361844414106e-01, 7.51903189826749e-01, 7.42949149166506e-01, 7.33563728212829e-01, 7.23818214866824e-01, 7.13789012868361e-01, 7.03555430760034e-01, 6.93197555805440e-01, 6.82794320453979e-01, 6.72421837840588e-01, 6.62152048134487e-01, 6.52051684759300e-01, 6.42181542901454e-01, 6.32596014706778e-01, 6.23342846473804e-01, 6.14463071668055e-01, 6.05991077445822e-01, 5.97954769144053e-01, 5.90375804807266e-01, 5.83269878895635e-01, 5.76647040150235e-01, 5.70512033030099e-01, 5.64864655359145e-01, 5.59700127140772e-01, 5.55009467220145e-01, 5.50779875820055e-01, 5.46995122068466e-01, 5.43635936499740e-01, 5.40680409120142e-01, 5.38104393919156e-01, 5.35881920627403e-01, 5.33985614029930e-01, 5.32387120240855e-01, 5.31057538073227e-01, 5.29967852070826e-01, 5.29089362036927e-01, 5.28394102135579e-01, 5.27855241033549e-01, 5.27447453259200e-01, 5.27147251144797e-01, 5.26933266518341e-01, 5.26786471808464e-01, 5.26690331449359e-01, 5.26630876387031e-01, 5.26596696996521e-01, 5.26578852670710e-01, 5.26570699541352e-01, 5.26567641035683e-01, 5.26566809049385e-01, 5.26566686249978e-01, 5.26566676744821e-01, 5.26566505625297e-01, 5.26565346266253e-01, 5.26561084353659e-01, 5.26549723766337e-01, 5.26524861941084e-01, 5.26477250648995e-01, 5.26394460712510e-01, 5.26260667610870e-01, 5.26056573029066e-01, 5.25759475030881e-01, 5.25343496450319e-01, 5.24779977062195e-01, 5.24038029904503e-01, 5.23085255677813e-01, 5.21888601492620e-01, 5.20415341635115e-01, 5.18634148986806e-01, 5.16516217002476e-01, 5.14036384659249e-01, 5.11174211542979e-01, 5.07914948184094e-01, 5.04250348595889e-01, 5.00179278017697e-01, 4.95708078933447e-01, 4.90850671811089e-01, 4.85628382535211e-01, 4.80069504744165e-01, 4.74208620730096e-01, 4.68085717856482e-01, 4.61745147550516e-01, 4.55234480201696e-01, 4.48603311533603e-01, 4.41902074374980e-01, 4.35180904718749e-01, 4.28488603213054e-01, 4.21871723622786e-01, 4.15373809246859e-01, 4.09034787683083e-01, 4.02890524526796e-01, 3.96972528243282e-01, 3.91307792032478e-01, 3.85918754231668e-01, 3.80823356666141e-01, 3.76035180140036e-01, 3.71563637591379e-01, 3.67414207866616e-01, 3.63588696140994e-01, 3.60085510302937e-01, 3.56899945797622e-01, 3.54024474244383e-01, 3.51449033449057e-01, 3.49161318156106e-01, 3.47147072007147e-01, 3.45390381724044e-01, 3.43873974570556e-01, 3.42579519739559e-01, 3.41487933547754e-01, 3.40579687284986e-01, 3.39835115358682e-01, 3.39234720097750e-01, 3.38759468341010e-01, 3.38391073843184e-01, 3.38112258692775e-01, 3.37906986444153e-01, 3.37760659599610e-01, 3.37660274478793e-01, 3.37594527398338e-01, 3.37553867421704e-01, 3.37530492662481e-01, 3.37518289131665e-01, 3.37512713283303e-01, 3.37510621597488e-01, 3.37510052608934e-01, 3.37509968627444e-01, 3.37509962074222e-01, 3.37509843417379e-01, 3.37509039500347e-01, 3.37506084224444e-01, 3.37498206606909e-01, 3.37480966952782e-01, 3.37447952175793e-01, 3.37390543028107e-01, 3.37297764852863e-01, 3.37156232107361e-01, 3.36950195212239e-01, 3.36661696116821e-01, 3.36270836181527e-01, 3.35756156433106e-01, 3.35095125880599e-01, 3.34264728419356e-01, 3.33242133066932e-01, 3.32005426189426e-01, 3.30534378471484e-01, 3.28811214265343e-01, 3.26821347300879e-01, 3.24554045208927e-01, 3.22002986430135e-01, 3.19166677148758e-01, 3.16048702865288e-01, 3.12657798693579e-01, 3.09007733662099e-01, 3.05117016145758e-01, 3.01008438842317e-01, 2.96708491239087e-01, 2.92246674312111e-01, 2.87654755648887e-01, 2.82966003136707e-01, 2.78214432136835e-01, 2.73434095378974e-01, 2.68658437613240e-01, 2.63919729356715e-01, 2.59248586784807e-01, 2.54673578642354e-01, 2.50220916405866e-01, 2.45914220941231e-01, 2.41774357430785e-01, 2.37819330062464e-01, 2.34064228481870e-01, 2.30521218925091e-01, 2.27199574000636e-01, 2.24105736098196e-01, 2.21243410285027e-01, 2.18613683278221e-01, 2.16215165648961e-01, 2.14044154841076e-01, 2.12094816899750e-01, 2.10359385022471e-01, 2.08828373177809e-01, 2.07490803070385e-01, 2.06334442647446e-01, 2.05346054124092e-01, 2.04511649142731e-01, 2.03816748193375e-01, 2.03246640833527e-01, 2.02786642611557e-01, 2.02422343981567e-01, 2.02139845970237e-01, 2.01925976993593e-01, 2.01768485090682e-01, 2.01656199994744e-01, 2.01579159929651e-01, 2.01528698809810e-01, 2.01497490607991e-01, 2.01479548993163e-01, 2.01470181853921e-01, 2.01465901930789e-01, 2.01464296383847e-01, 2.01463859636178e-01, 2.01463795173154e-01, 2.01463790100178e-01, 2.01463697697420e-01, 2.01463071655326e-01, 2.01460770260951e-01, 2.01454635608755e-01, 2.01441210174877e-01, 2.01415499254992e-01, 2.01370788976561e-01, 2.01298528573133e-01, 2.01188284411487e-01, 2.01027771829089e-01, 2.00802969098206e-01, 2.00498315702021e-01, 2.00096994522170e-01, 1.99581294469482e-01, 1.98933046577821e-01, 1.98134122735254e-01, 1.97166982255496e-01, 1.96015247690170e-01, 1.94664288006464e-01, 1.93101784902765e-01, 1.91318256965261e-01, 1.89307516860822e-01, 1.87067038942700e-01, 1.84598218456287e-01, 1.81906508718527e-01, 1.79001428790272e-01, 1.75896440735443e-01, 1.72608702004894e-01, 1.69158704284229e-01, 1.65569814911677e-01, 1.61867740493477e-01, 1.58079934592109e-01, 1.54234972462632e-01, 1.50361915930560e-01, 1.46489690747416e-01, 1.42646497139751e-01, 1.38859271773399e-01, 1.35153216047897e-01, 1.31551401744090e-01, 1.28074460901288e-01, 1.24740362660532e-01, 1.21564275871336e-01, 1.18558512605730e-01, 1.15732544557446e-01, 1.13093081879668e-01, 1.10644202624655e-01, 1.08387520677937e-01, 1.06322380762862e-01, 1.04446070456177e-01, 1.02754040853497e-01, 1.01240129354309e-01, 9.98967799149995e-02, 9.87152578467914e-02, 9.76858577635085e-02, 9.67981044634598e-02, 9.60409472368440e-02, 9.54029483671377e-02, 9.48724664797993e-02, 9.44378348707602e-02, 9.40875342991490e-02, 9.38103588428305e-02, 9.35955725835527e-02, 9.34330541246879e-02, 9.33134253349692e-02, 9.32281604294900e-02, 9.31696714661996e-02, 9.31313666966531e-02, 9.31076788602354e-02, 9.30940614500884e-02, 9.30869521510687e-02, 9.30837039030195e-02, 9.30824853811032e-02, 9.30821539141107e-02, 9.30821049905006e-02, 9.30821011065888e-02, 9.30820299242328e-02, 9.30815476536964e-02, 9.30797747844423e-02, 9.30750490279166e-02, 9.30647071408117e-02, 9.30449023911706e-02, 9.30104655287209e-02, 9.29548165108685e-02, 9.28699335722030e-02, 9.27463853049915e-02, 9.25734302367829e-02, 9.23391867649173e-02, 9.20308741935827e-02, 9.16351229338517e-02, 9.11383487142530e-02, 9.05271819518703e-02, 8.97889394928484e-02, 8.89121219403677e-02, 8.78869161569753e-02, 8.67056796252115e-02, 8.53633814682979e-02, 8.38579747042445e-02, 8.21906756464870e-02, 8.03661298094624e-02, 7.83924488804786e-02, 7.62811102271144e-02, 7.40467183721023e-02, 7.17066363923493e-02, 6.92805034119601e-02, 6.67896618074053e-02, 6.42565237510553e-02, 6.17039107365684e-02, 5.91544013066851e-02, 5.66297206937776e-02, 5.41502014676212e-02, 5.17343372953860e-02, 4.93984434878030e-02, 4.71564300091245e-02, 4.50196854761487e-02, 4.29970648539069e-02, 4.10949680932333e-02, 3.93174929947833e-02, 3.76535634691919e-02, 3.61310584484329e-02, 3.47336227864048e-02, 3.34584954953385e-02, 3.23018136125057e-02, 3.12588414087998e-02, 3.03241752662018e-02, 2.94919214360050e-02, 2.87558462641471e-02, 2.81095003182088e-02, 2.75463191753573e-02, 2.70597044609673e-02, 2.66430891225724e-02, 2.62899909463954e-02, 2.59940580402684e-02, 2.57491094864548e-02, 2.55491736683382e-02, 2.53885259638906e-02, 2.52617266342412e-02, 2.51636588821782e-02, 2.50895662733512e-02, 2.50350880588428e-02, 2.49962904635908e-02, 2.49696917417452e-02, 2.49522787722043e-02, 2.49415131661834e-02, 2.49353252668385e-02, 2.49320949897793e-02, 2.49306191326468e-02, 2.49300655011098e-02, 2.49299149011932e-02, 2.49298926730006e-02, 2.49298908921105e-02, 2.49298580514647e-02, 2.49296355518283e-02, 2.49288176310358e-02, 2.49266374410979e-02, 2.49218665884428e-02, 2.49127315196631e-02, 2.48968508299185e-02, 2.48711975644511e-02, 2.48320905344741e-02, 2.47752184842085e-02, 2.46957005014110e-02, 2.45881852338385e-02, 2.44469901392392e-02, 2.42662800848734e-02, 2.40402821173973e-02, 2.37635302254802e-02, 2.34311306181070e-02, 2.30390347395968e-02, 2.25843043412794e-02, 2.20653508822011e-02, 2.14821307865947e-02, 2.08362790092735e-02, 2.01311661631962e-02, 1.93718691183165e-02, 1.85650511984363e-02, 1.77187553221074e-02, 1.68421208731786e-02, 1.59450418604905e-02, 1.50377891581996e-02, 1.41306226116598e-02, 1.32334191448962e-02, 1.23553406864119e-02, 1.15045610963875e-02, 1.06880650084520e-02, 9.91152445861587e-03, 9.17925227454930e-03, 8.49422523346277e-03, 7.85816553450047e-03, 7.27166645503985e-03, 6.73434716270549e-03, 6.24502230471080e-03, 5.80187379992481e-03, 5.40261477486581e-03, 5.04463837805247e-03, 4.72514693130677e-03, 4.44125927587891e-03, 4.19009611094623e-03, 3.96884454809765e-03, 3.77480404644359e-03, 3.60541641366297e-03, 3.45828274825521e-03, 3.33117015034645e-03, 3.22201082175866e-03, 3.12889588316523e-03, 3.05006589844936e-03, 2.98389975483815e-03, 2.92890321291850e-03, 2.88369812807332e-03, 2.84701305329804e-03, 2.81767566384872e-03, 2.79460719640667e-03, 2.77681886991843e-03, 2.76341005655922e-03, 2.75356780602465e-03, 2.74656720091571e-03, 2.74177194435765e-03, 2.73863455707613e-03, 2.73669559240375e-03, 2.73558135937435e-03, 2.73499976829328e-03, 2.73473406623737e-03, 2.73463439735622e-03, 2.73460728548969e-03, 2.73460328385511e-03, 2.73460296023131e-03, 2.73459695470990e-03, 2.73455626668806e-03, 2.73440669745639e-03, 2.73400803325785e-03, 2.73313573114820e-03, 2.73146579941397e-03, 2.72856374822084e-03, 2.72387855861211e-03, 2.71674268225297e-03, 2.70637910992377e-03, 2.69191647929290e-03, 2.67241298956833e-03, 2.64688950076675e-03, 2.61437159166178e-03, 2.57393953393148e-03, 2.52478414991231e-03, 2.46626545261108e-03, 2.39796995744257e-03, 2.31976178713467e-03, 2.23182235315278e-03, 2.13467365536246e-03, 2.02918119335038e-03, 1.91653411752668e-03, 1.79820242520634e-03, 1.67587346109599e-03, 1.55137235648795e-03, 1.42657295536708e-03, 1.30330689992775e-03, 1.18327868199231e-03, 1.06799358304954e-03, 9.58703685881801e-04, 8.56374849561803e-04, 7.61675086991431e-04, 6.74982550393111e-04, 5.96669558831309e-04, 5.26103785574605e-04, 4.63161334906204e-04, 4.07337767995841e-04, 3.58593815050767e-04, 3.15818556373415e-04, 2.78427593897845e-04, 2.45823565805381e-04, 2.17447569369256e-04, 1.92696176142704e-04, 1.71143598663548e-04, 1.52298823124397e-04, 1.35631205395725e-04, 1.20768870743859e-04, 1.07414382043203e-04, 9.53365379956495e-05, 8.43589100441022e-05, 7.43542379524342e-05, 6.52984796403189e-05, 5.70082973041508e-05, 4.94020577406218e-05, 4.24502255498282e-05, 3.61299803302018e-05, 3.04215786997020e-05, 2.53062444670604e-05, 2.07630855868501e-05, 1.67691974866102e-05, 1.33005481046666e-05, 1.03309322696532e-05, 7.83082305133576e-06, 5.76690337084074e-06, 4.10215980650719e-06, 2.79634008393259e-06, 1.80662780735474e-06, 1.08844255919394e-06, 5.96314706058365e-07, 2.84819634543965e-07, 2.84217094304040e-14, 2.84217094304040e-14, 2.84217094304040e-14],
[ 9.50000000000000e+01, 9.49999998839179e+01, 9.49999962944297e+01, 9.49999719750765e+01, 9.49998825739254e+01, 9.49996442583934e+01, 9.49991226858842e+01, 9.49981237171232e+01, 9.49963862048060e+01, 9.49935771251081e+01, 9.49892892463169e+01, 9.49830414506517e+01, 9.49742817461916e+01, 9.49623929300284e+01, 9.49467007955450e+01, 9.49264847195810e+01, 9.49009904212414e+01, 9.48694446529943e+01, 9.48310715634975e+01, 9.47851104543970e+01, 9.47308346321108e+01, 9.46675710216229e+01, 9.45947201555226e+01, 9.45117760751880e+01, 9.44183455860694e+01, 9.43141662072806e+01, 9.41991220662523e+01, 9.40732569355573e+01, 9.39367836145127e+01, 9.37900889402484e+01, 9.36337338780535e+01, 9.34684483812637e+01, 9.32951210047822e+01, 9.31147835706459e+01, 9.29285914810514e+01, 9.27378005184331e+01, 9.25437411369617e+01, 9.23477913220041e+01, 9.21513490751147e+01, 9.19558054862917e+01, 9.17625192054259e+01, 9.15727929473162e+01, 9.13878524839026e+01, 9.12088284126826e+01, 9.10367408536418e+01, 9.08724871231246e+01, 9.07168323605683e+01, 9.05704030375353e+01, 9.04336832507026e+01, 9.03070136839610e+01, 9.01905931131600e+01, 9.00844823156149e+01, 8.99886102323712e+01, 8.99027822130774e+01, 8.98266901510447e+01, 8.97599242903986e+01, 8.97019864593948e+01, 8.96523044555064e+01, 8.96102472804674e+01, 8.95751408988264e+01, 8.95462841734878e+01, 8.95229646179175e+01, 8.95044735988063e+01, 8.94901206264295e+01, 8.94792463838686e+01, 8.94712341713670e+01, 8.94655194785762e+01, 8.94615974448912e+01, 8.94590280254282e+01, 8.94574387457755e+01, 8.94565250002233e+01, 8.94560479230947e+01, 8.94558299381381e+01, 8.94557481637310e+01, 8.94557259190412e+01, 8.94557226357691e+01, 8.94557222050730e+01, 8.94557121703721e+01, 8.94556441837837e+01, 8.94553942577009e+01, 8.94547280450476e+01, 8.94532700474548e+01, 8.94504777789133e+01, 8.94456219105563e+01, 8.94377732947196e+01, 8.94257976186604e+01, 8.94083582674699e+01, 8.93839277775564e+01, 8.93508080320887e+01, 8.93071590844329e+01, 8.92510361938982e+01, 8.91804343231215e+01, 8.90933389869613e+01, 8.89877819747721e+01, 8.88619001156547e+01, 8.87139949526550e+01, 8.85425909771589e+01, 8.83464899919223e+01, 8.81248192574565e+01, 8.78770713504231e+01, 8.76031341103314e+01, 8.73033096164537e+01, 8.69783217261784e+01, 8.66293122080358e+01, 8.62578258299588e+01, 8.58657848995680e+01, 8.54554537848336e+01, 8.50293940475540e+01, 8.45904111914573e+01, 8.41414947443791e+01, 8.36857543193656e+01, 8.32263550628200e+01, 8.27664560635152e+01, 8.23091546360587e+01, 8.18574380770424e+01, 8.14141430292958e+01, 8.09819215283979e+01, 8.05632124309402e+01, 8.01602171592080e+01, 7.97748792364907e+01, 7.94088676120063e+01, 7.90635640948453e+01, 7.87400552955405e+01, 7.84391293737419e+01, 7.81612777034433e+01, 7.79067013704748e+01, 7.76753222555376e+01, 7.74667983450086e+01, 7.72805428464759e+01, 7.71157466527024e+01, 7.69714036810860e+01, 7.68463386030621e+01, 7.67392364610293e+01, 7.66486736455659e+01, 7.65731496730924e+01, 7.65111191667842e+01, 7.64610234064722e+01, 7.64213207825682e+01, 7.63905154710713e+01, 7.63671836473749e+01, 7.63499965807503e+01, 7.63377400023244e+01, 7.63293292184049e+01, 7.63238195472834e+01, 7.63204117881152e+01, 7.63184525800969e+01, 7.63174296723185e+01, 7.63169622916825e+01, 7.63167869601604e+01, 7.63167392656493e+01, 7.63167322260363e+01, 7.63167316045725e+01, 7.63167194272946e+01, 7.63166369244329e+01, 7.63163336350000e+01, 7.63155251752972e+01, 7.63137558737666e+01, 7.63103674219070e+01, 7.63044747887987e+01, 7.62949504937617e+01, 7.62804181543135e+01, 7.62592560212862e+01, 7.62296109744904e+01, 7.61894231752412e+01, 7.61364612525256e+01, 7.60683675372134e+01, 7.59827124585285e+01, 7.58770567908792e+01, 7.57490200064875e+01, 7.55963525765123e+01, 7.54170097022552e+01, 7.52092236826000e+01, 7.49715719667593e+01, 7.47030379304160e+01, 7.44030615683731e+01, 7.40715776281183e+01, 7.37090392153715e+01, 7.33164255705287e+01, 7.28952335128357e+01, 7.24474529249258e+01, 7.19755275289198e+01, 7.14823029935460e+01, 7.09709650132684e+01, 7.04449703383389e+01, 6.99079737760013e+01, 6.93637539538402e+01, 6.88161402186067e+01, 6.82689425524749e+01, 6.77258859339103e+01, 6.71905502223155e+01, 6.66663164152198e+01, 6.61563199718394e+01, 6.56634117507233e+01, 6.51901269173253e+01, 6.47386619234859e+01, 6.43108593715052e+01, 6.39082003029385e+01, 6.35318032477133e+01, 6.31824292615917e+01, 6.28604921687145e+01, 6.25660732870963e+01, 6.22989400145429e+01, 6.20585677587214e+01, 6.18441647853104e+01, 6.16546996200692e+01, 6.14889306705464e+01, 6.13454377325785e+01, 6.12226550200808e+01, 6.11189053095215e+01, 6.10324347291275e+01, 6.09614476540970e+01, 6.09041411001701e+01, 6.08587379465861e+01, 6.08235182735445e+01, 6.07968480761758e+01, 6.07772046230240e+01, 6.07631977666899e+01, 6.07535865898106e+01, 6.07472908804437e+01, 6.07433970738539e+01, 6.07411584666474e+01, 6.07399896959618e+01, 6.07394556713337e+01, 6.07392553396687e+01, 6.07392008445428e+01, 6.07391928011744e+01, 6.07391921559967e+01, 6.07391802491968e+01, 6.07390995788721e+01, 6.07388030261768e+01, 6.07380125251993e+01, 6.07362825312601e+01, 6.07329693786794e+01, 6.07272077559149e+01, 6.07178953809384e+01, 6.07036867911602e+01, 6.06829969656028e+01, 6.06540152660389e+01, 6.06147299117942e+01, 6.05629628848286e+01, 6.04964147962536e+01, 6.04127188371916e+01, 6.03095024980589e+01, 6.01844552918226e+01, 6.00354002880194e+01, 5.98603668917983e+01, 5.96576620263394e+01, 5.94259367371592e+01, 5.91642452655644e+01, 5.88720938549148e+01, 5.85494769571316e+01, 5.81968990749963e+01, 5.78153811632318e+01, 5.74064512573999e+01, 5.69721197392856e+01, 5.65148403249602e+01, 5.60374584414981e+01, 5.55431491274313e+01, 5.50353469527207e+01, 5.45176707097064e+01, 5.39938457690553e+01, 5.34676270015300e+01, 5.29427250099135e+01, 5.24227380806928e+01, 5.19110917660471e+01, 5.14109873918226e+01, 5.09253601317623e+01, 5.04568466758857e+01, 5.00077620224990e+01, 4.95800845813686e+01, 4.91754485985277e+01, 4.87951428798052e+01, 4.84401148609150e+01, 4.81109792019055e+01, 4.78080302333305e+01, 4.75312577222748e+01, 4.72803655423370e+01, 4.70547929169726e+01, 4.68537379607500e+01, 4.66761832714481e+01, 4.65209233313016e+01, 4.63865934612061e+01, 4.62717000397933e+01, 4.61746516523936e+01, 4.60937907761247e+01, 4.60274255409940e+01, 4.59738610387037e+01, 4.59314295877094e+01, 4.58985193126380e+01, 4.58736003659761e+01, 4.58552481167488e+01, 4.58421626597851e+01, 4.58331840628869e+01, 4.58273028677915e+01, 4.58236654913344e+01, 4.58215743300176e+01, 4.58204825463448e+01, 4.58199836993104e+01, 4.58197965642328e+01, 4.58197456589229e+01, 4.58197381454046e+01, 4.58197375658898e+01, 4.58197271596638e+01, 4.58196566559519e+01, 4.58193974768637e+01, 4.58187066006837e+01, 4.58171946356280e+01, 4.58142990438340e+01, 4.58092635921759e+01, 4.58011249817089e+01, 4.57887074623434e+01, 4.57706260693813e+01, 4.57452989178294e+01, 4.57109687526367e+01, 4.56657336733820e+01, 4.56075866291165e+01, 4.55344629170916e+01, 4.54442945290813e+01, 4.53350697899344e+01, 4.52048963516074e+01, 4.50520652750730e+01, 4.48751136879897e+01, 4.46728833820525e+01, 4.44445727376508e+01, 4.41897795493653e+01, 4.39085326716874e+01, 4.36013108894828e+01, 4.32690480052851e+01, 4.29131237787273e+01, 4.25353410048028e+01, 4.21378896383797e+01, 4.17232994377663e+01, 4.12943830975878e+01, 4.08541722602665e+01, 4.04058491139359e+01, 3.99526764611776e+01, 3.94979291240637e+01, 3.90448292957705e+01, 3.85964879606217e+01, 3.81558538479741e+01, 3.77256706770942e+01, 3.73084428182988e+01, 3.69064090336570e+01, 3.65215236983775e+01, 3.61554448103187e+01, 3.58095281045513e+01, 3.54848266368461e+01, 3.51820952411774e+01, 3.49017992870714e+01, 3.46441271693284e+01, 3.44090059704063e+01, 3.41961197577841e+01, 3.40049300205229e+01, 3.38346978082960e+01, 3.36845072035936e+01, 3.35532898223445e+01, 3.34398500891307e+01, 3.33428910626363e+01, 3.32610405910170e+01, 3.31928775556769e+01, 3.31369579193919e+01, 3.30918402376523e+01, 3.30561102292958e+01, 3.30284039436636e+01, 3.30074290162150e+01, 3.29919834812405e+01, 3.29809716155481e+01, 3.29734163246684e+01, 3.29684676541630e+01, 3.29654071108492e+01, 3.29636476071557e+01, 3.29627289890540e+01, 3.29623092651748e+01, 3.29621518123088e+01, 3.29621089813164e+01, 3.29621026595519e+01, 3.29621021808870e+01, 3.29620937013857e+01, 3.29620362515296e+01, 3.29618250599631e+01, 3.29612621021839e+01, 3.29600300899489e+01, 3.29576706663262e+01, 3.29535676939895e+01, 3.29469364286661e+01, 3.29368192579215e+01, 3.29220885531771e+01, 3.29014570240066e+01, 3.28734957720909e+01, 3.28366600128524e+01, 3.27893221638874e+01, 3.27298116937322e+01, 3.26564607909669e+01, 3.25676545674976e+01, 3.24618841726199e+01, 3.23378008927137e+01, 3.21942690743353e+01, 3.20304155647385e+01, 3.18456733380632e+01, 3.16398170847840e+01, 3.14129887937965e+01, 3.11657117469282e+01, 3.08988918598438e+01, 3.06138059164135e+01, 3.03120769220215e+01, 2.99956375036040e+01, 2.96666829620063e+01, 2.93276161815304e+01, 2.89809870661108e+01, 2.86294294494507e+01, 2.82755984790989e+01, 2.79221112865178e+01, 2.75714933419550e+01, 2.72261323017853e+01, 2.68882404605154e+01, 2.65598262068961e+01, 2.62426742379541e+01, 2.59383337726974e+01, 2.56481136670403e+01, 2.53730831701710e+01, 2.51140770598948e+01, 2.48717040118577e+01, 2.46463572488625e+01, 2.44382267377662e+01, 2.42473124175630e+01, 2.40734381298884e+01, 2.39162660705201e+01, 2.37753116848421e+01, 2.36499589948830e+01, 2.35394763764698e+01, 2.34430328088590e+01, 2.33597146018234e+01, 2.32885425714418e+01, 2.32284895897201e+01, 2.31784983782581e+01, 2.31374993561226e+01, 2.31044282909771e+01, 2.30782434449740e+01, 2.30579418579222e+01, 2.30425743748681e+01, 2.30312590080921e+01, 2.30231922283228e+01, 2.30176578089227e+01, 2.30140329003292e+01, 2.30117910884984e+01, 2.30105022869010e+01, 2.30098294214204e+01, 2.30095219847893e+01, 2.30094066548835e+01, 2.30093752823706e+01, 2.30093706518540e+01, 2.30093703042397e+01, 2.30093641858336e+01, 2.30093227327462e+01, 2.30091703470201e+01, 2.30087641442215e+01, 2.30078751874438e+01, 2.30061727599459e+01, 2.30032123214693e+01, 2.29984277267562e+01, 2.29911282066169e+01, 2.29805005179978e+01, 2.29656165544556e+01, 2.29454465664959e+01, 2.29188779676472e+01, 2.28847394948237e+01, 2.28418302525499e+01, 2.27889529076795e+01, 2.27249500284220e+01, 2.26487422992784e+01, 2.25593671174971e+01, 2.24560159151659e+01, 2.23380684811775e+01, 2.22051226004219e+01, 2.20570174945655e+01, 2.18938498362088e+01, 2.17159814963588e+01, 2.15240386391684e+01, 2.13189022522862e+01, 2.11016906474579e+01, 2.08737348420868e+01, 2.06365480111714e+01, 2.03917903735262e+01, 2.01412309591773e+01, 1.98867077214449e+01, 1.96300874329877e+01, 1.93732267535214e+01, 1.91179357727309e+01, 1.88659451939852e+01, 1.86188781092554e+01, 1.83782270148660e+01, 1.81453363513961e+01, 1.79213904675251e+01, 1.77074065691369e+01, 1.75042319757940e+01, 1.73125448944669e+01, 1.71328579299211e+01, 1.69655236509743e+01, 1.68107416789819e+01, 1.66685669194976e+01, 1.65389186924326e+01, 1.64215906169294e+01, 1.63162611729120e+01, 1.62225048968758e+01, 1.61398041822082e+01, 1.60675616508050e+01, 1.60051130477825e+01, 1.59517405876912e+01, 1.59066866507285e+01, 1.58691676925613e+01, 1.58383881933098e+01, 1.58135544324413e+01, 1.57938878398361e+01, 1.57786376426714e+01, 1.57670925067123e+01, 1.57585908625047e+01, 1.57525296145231e+01, 1.57483709561386e+01, 1.57456470555838e+01, 1.57439624366605e+01, 1.57429939501066e+01, 1.57424883134361e+01, 1.57422572840002e+01, 1.57421706169045e+01, 1.57421470413552e+01, 1.57421435616536e+01, 1.57421433012256e+01, 1.57421387279802e+01, 1.57421077435752e+01, 1.57419938417042e+01, 1.57416902217469e+01, 1.57410257600006e+01, 1.57397532481573e+01, 1.57375403765848e+01, 1.57339638876432e+01, 1.57285072631057e+01, 1.57205622372900e+01, 1.57094343404208e+01, 1.56943525715348e+01, 1.56744831741141e+01, 1.56489473395546e+01, 1.56168424952609e+01, 1.55772666509671e+01, 1.55293450883057e+01, 1.54722584982257e+01, 1.54052715153516e+01, 1.53277604860526e+01, 1.52392392551917e+01, 1.51393817786596e+01, 1.50280404714155e+01, 1.49052593812872e+01, 1.47712815247539e+01, 1.46265500112004e+01, 1.44717028902035e+01, 1.43075619552869e+01, 1.41351160047720e+01, 1.39554992815063e+01, 1.37699659830304e+01, 1.35798618537365e+01, 1.33865939451322e+01, 1.31915996624182e+01, 1.29963162045094e+01, 1.28021514466327e+01, 1.26104572060931e+01, 1.24225056731598e+01, 1.22394695877444e+01, 1.20624065136201e+01, 1.18922473258266e+01, 1.17297888057346e+01, 1.15756900515692e+01, 1.14304722737089e+01, 1.12945214599677e+01, 1.11680933650959e+01, 1.10513202937598e+01, 1.09442191963881e+01, 1.08467006701482e+01, 1.07585785409694e+01, 1.06795797866727e+01, 1.06093546378695e+01, 1.05474867568122e+01, 1.04935034415067e+01, 1.04468858317857e+01, 1.04070791059225e+01, 1.03735026522276e+01, 1.03455601824274e+01, 1.03226497257450e+01, 1.03041734083689e+01, 1.02895468866379e+01, 1.02782082681447e+01, 1.02696263272525e+01, 1.02633078039013e+01, 1.02588035699416e+01, 1.02557134573278e+01, 1.02536895678441e+01, 1.02524379237327e+01, 1.02517183704990e+01, 1.02513427040407e+01, 1.02511710601151e+01, 1.02511066707441e+01, 1.02510891552920e+01, 1.02510865700486e+01, 1.02510863764914e+01, 1.02510829765608e+01, 1.02510599415423e+01, 1.02509752625418e+01, 1.02507495406810e+01, 1.02502555603260e+01, 1.02493095534398e+01, 1.02476645164848e+01, 1.02450059060959e+01, 1.02409499982796e+01, 1.02350451457039e+01, 1.02267761054610e+01, 1.02155715321098e+01, 1.02008146350163e+01, 1.01818568836556e+01, 1.01580345105203e+01, 1.01286874124319e+01, 1.00931798945443e+01, 1.00509225476066e+01, 1.00013944113610e+01, 9.94416447023534e+00, 9.87891146685725e+00, 9.80544101769308e+00, 9.72369908276376e+00, 9.63378098155597e+00, 9.53593535627300e+00, 9.43056274980613e+00, 9.31820876998745e+00, 9.19955212850041e+00, 9.07538814366343e+00, 8.94660855284602e+00, 8.81417866844669e+00, 8.67911301486448e+00, 8.54245059666468e+00, 8.40523087505228e+00, 8.26847138543765e+00, 8.13314773476804e+00, 8.00017649782818e+00, 7.87040131029870e+00, 7.74458225224900e+00, 7.62338844220290e+00, 7.50739362596410e+00, 7.39707444765294e+00, 7.29281103070086e+00, 7.19488946947212e+00, 7.10350583240668e+00, 7.01877129959900e+00, 6.94071809619915e+00, 6.86930593280879e+00, 6.80442872021428e+00, 6.74592138369488e+00, 6.69356665752700e+00, 6.64710178954995e+00, 6.60622512596031e+00, 6.57060257596783e+00, 6.53987397366099e+00, 6.51365936038779e+00, 6.49156520599242e+00, 6.47319057290579e+00, 6.45813320550288e+00, 6.44599550090703e+00, 6.43639028943869e+00, 6.42894632619951e+00, 6.42331337280787e+00, 6.41916673270552e+00, 6.41621109685417e+00, 6.41418356039979e+00, 6.41285568546959e+00, 6.41203451017526e+00, 6.41156243767959e+00, 6.41131597960452e+00, 6.41120337231553e+00, 6.41116112966695e+00, 6.41114963866433e+00, 6.41114794261721e+00, 6.41114781529723e+00, 6.41114557435897e+00, 6.41113039168430e+00, 6.41107457874188e+00, 6.41092580327842e+00, 6.41060021866264e+00, 6.40997671141106e+00, 6.40889250917739e+00, 6.40714037058765e+00, 6.40446755341347e+00, 6.40057672655463e+00, 6.39512895138586e+00, 6.38774880628847e+00, 6.37803166217286e+00, 6.36555303518745e+00, 6.34987984611742e+00, 6.33058330711810e+00, 6.30725304104992e+00, 6.27951192515940e+00, 6.24703104979760e+00, 6.20954410621438e+00, 6.16686047704975e+00, 6.11887630904378e+00, 6.06558290629852e+00, 6.00707189577665e+00, 5.94353678034663e+00, 5.87527069824283e+00, 5.80266043558946e+00, 5.72617697104624e+00, 5.64636304731930e+00, 5.56381844259734e+00, 5.47918373854658e+00, 5.39312343854439e+00, 5.30630927599590e+00, 5.21940447186367e+00, 5.13304956504466e+00, 5.04785026742977e+00, 4.96436760924279e+00, 4.88311046109455e+00, 4.80453036474321e+00, 4.72901848612905e+00, 4.65690442607040e+00, 4.58845658405035e+00, 4.52388376251854e+00, 4.46333771484368e+00, 4.40691637123057e+00, 4.35466751653493e+00, 4.30659273665796e+00, 4.26265149242116e+00, 4.22276521916015e+00, 4.18682138527166e+00, 4.15467747269800e+00, 4.12616486626251e+00, 4.10109265652707e+00, 4.07925137226417e+00, 4.06041666376400e+00, 4.04435295730296e+00, 4.03081709472015e+00, 4.01956196100248e+00, 4.01034008816760e+00, 4.00290720691601e+00, 3.99702570006473e+00, 3.99246789535397e+00, 3.98901912151501e+00, 3.98648044204418e+00, 3.98467097719326e+00, 3.98342972708807e+00, 3.98261681791536e+00, 3.98211410848143e+00, 3.98182511527128e+00, 3.98167423906293e+00, 3.98160530345714e+00, 3.98157944349002e+00, 3.98157240896690e+00, 3.98157137068657e+00, 3.98157129241233e+00, 3.98156991029794e+00, 3.98156054627765e+00, 3.98152612326884e+00, 3.98143436508961e+00, 3.98123355924693e+00, 3.98084900999952e+00, 3.98018033184296e+00, 3.97909972107374e+00, 3.97745132814406e+00, 3.97505183468716e+00, 3.97169231453236e+00, 3.96714142577702e+00, 3.96114993938234e+00, 3.95345655791118e+00, 3.94379491626682e+00, 3.93190158659725e+00, 3.91752483575117e+00, 3.90043381152377e+00, 3.88042777064328e+00, 3.85734491505514e+00, 3.83107038135991e+00, 3.80154293757018e+00, 3.76875998527304e+00, 3.73278054369864e+00, 3.69372600078714e+00, 3.65177854678223e+00, 3.60717734670180e+00, 3.56021264615504e+00, 3.51121812744077e+00, 3.46056192863682e+00, 3.40863679970984e+00, 3.35584989285525e+00, 3.30261266982673e+00, 3.24933136119029e+00, 3.19639833836328e+00, 3.14418466788949e+00, 3.09303401821485e+00, 3.04325799142332e+00, 2.99513286383724e+00, 2.94889764600785e+00, 2.90475331805168e+00, 2.86286306179139e+00, 2.82335329587916e+00, 2.78631532148461e+00, 2.75180740062170e+00, 2.71985711270870e+00, 2.69046386353638e+00, 2.66360145098460e+00, 2.63922062081868e+00, 2.61725157176004e+00, 2.59760639054375e+00, 2.58018141425421e+00, 2.56485952874118e+00, 2.55151241854171e+00, 2.54000278586518e+00, 2.53018655435251e+00, 2.52191506810817e+00, 2.51503728861529e+00, 2.50940198235182e+00, 2.50485988108535e+00, 2.50126578586581e+00, 2.49848057562482e+00, 2.49637307299258e+00, 2.49482171432612e+00, 2.49371596872618e+00, 2.49295745246258e+00, 2.49246069089325e+00, 2.49215348945719e+00, 2.49197688811846e+00, 2.49188468890942e+00, 2.49184256292418e+00, 2.49182676010901e+00, 2.49182246136918e+00, 2.49182182688452e+00, 2.49182177879146e+00, 2.49182092614362e+00, 2.49181514933395e+00, 2.49179391324614e+00, 2.49173730621371e+00, 2.49161342605493e+00, 2.49137619202362e+00, 2.49096367538674e+00, 2.49029703368069e+00, 2.48928012451624e+00, 2.48779986320088e+00, 2.48572737326101e+00, 2.48291995893405e+00, 2.47922390280719e+00, 2.47447805942639e+00, 2.46851817709033e+00, 2.46118183642481e+00, 2.45231384815121e+00, 2.44177190736779e+00, 2.42943226229194e+00, 2.41519512693448e+00, 2.39898955463437e+00, 2.38077749686047e+00, 2.36055680149782e+00, 2.33836295681910e+00, 2.31426945849839e+00, 2.28838676158912e+00, 2.26085986939809e+00, 2.23186469752218e+00, 2.20160342500547e+00, 2.17029909818702e+00, 2.13818978160047e+00, 2.10552255287853e+00, 2.07254761709876e+00, 2.03951277537648e+00, 2.00665842964416e+00, 1.97421324791817e+00, 1.94239055873167e+00, 1.91138549495410e+00, 1.88137286891278e+00, 1.85250573340785e+00, 1.82491456595262e+00, 1.79870700437591e+00, 1.77396805844825e+00, 1.75076072231968e+00, 1.72912691479046e+00, 1.70908867796429e+00, 1.69064956944128e+00, 1.67379618899167e+00, 1.65849978776220e+00, 1.64471791648091e+00, 1.63239607853737e+00, 1.62146936363953e+00, 1.61186404721272e+00, 1.60349914898758e+00, 1.59628795058721e+00, 1.59013947583539e+00, 1.58495993871145e+00, 1.58065416241031e+00, 1.57712696915895e+00, 1.57428453486310e+00, 1.57203569605626e+00, 1.57029318984632e+00, 1.56897480146370e+00, 1.56800438939891e+00, 1.56731275561850e+00, 1.56683832840055e+00, 1.56652762810781e+00, 1.56633549163118e+00, 1.56622503894198e+00, 1.56616737462963e+00, 1.56614102777330e+00, 1.56613114422977e+00, 1.56612845567308e+00, 1.56612805884799e+00, 1.56612802857833e+00, 1.56612748940589e+00, 1.56612383643704e+00, 1.56611040779377e+00, 1.56607461239632e+00, 1.56599627725267e+00, 1.56584626454338e+00, 1.56558541766724e+00, 1.56516389088989e+00, 1.56452091179736e+00, 1.56358501879892e+00, 1.56227480668004e+00, 1.56050020087058e+00, 1.55816426498433e+00, 1.55516552593859e+00, 1.55140077668329e+00, 1.54676828900786e+00, 1.54117133948982e+00, 1.53452192253691e+00, 1.52674449832665e+00, 1.51777960321763e+00, 1.50758713876863e+00, 1.49614915521299e+00, 1.48347195756030e+00, 1.46958738771599e+00, 1.45455317311925e+00, 1.43845227923849e+00, 1.42139125681435e+00, 1.40349763152422e+00, 1.38491644017367e+00, 1.36580607010463e+00, 1.34633360383248e+00, 1.32666990551587e+00, 1.30698470613616e+00, 1.28744194674756e+00, 1.26819562117763e+00, 1.24938632028278e+00, 1.23113862138230e+00, 1.21355939436036e+00, 1.19673701875575e+00, 1.18074143413189e+00, 1.16562488867201e+00, 1.15142321498161e+00, 1.13815744996640e+00, 1.12583562545569e+00, 1.11445458262368e+00, 1.10400169920938e+00, 1.09445645710449e+00, 1.08579181352022e+00, 1.07797536812853e+00, 1.07097033978737e+00, 1.06473637979094e+00, 1.05923025512389e+00, 1.05440643649220e+00, 1.05021762353395e+00, 1.04661523495383e+00, 1.04354988542867e+00, 1.04097186473891e+00, 1.03883162816786e+00, 1.03708030107190e+00, 1.03567019485065e+00, 1.03455532647956e+00, 1.03369192945065e+00, 1.03303894055558e+00, 1.03255844460013e+00, 1.03221605801735e+00, 1.03198123255974e+00, 1.03182746184207e+00, 1.03173237642093e+00, 1.03167771717213e+00, 1.03164918169198e+00, 1.03163614395466e+00, 1.03163125310742e+00, 1.03162992268385e+00, 1.03162972631632e+00, 1.03162971123180e+00, 1.03162944115573e+00, 1.03162761135343e+00, 1.03162088485064e+00, 1.03160295480163e+00, 1.03156371704168e+00, 1.03148857860317e+00, 1.03135793276677e+00, 1.03114682946990e+00, 1.03082486767833e+00, 1.03035633318601e+00, 1.02970060065199e+00, 1.02881281186687e+00, 1.02764483269907e+00, 1.02614647856225e+00, 1.02426698261880e+00, 1.02195666288696e+00, 1.01916872519375e+00, 1.01586112041966e+00, 1.01199835916489e+00, 1.00755317755348e+00, 1.00250794695764e+00, 9.96855729863896e-01, 9.90600904587311e-01, 9.83759312073218e-01, 9.76357915770819e-01, 9.68434006031828e-01, 9.60034018179471e-01, 9.51212062797495e-01, 9.42028283546392e-01, 9.32547159747784e-01, 9.22835858657592e-01, 9.12962718968330e-01, 9.02995917666317e-01, 8.93002342524184e-01, 8.83046667095812e-01, 8.73190607248077e-01, 8.63492329036873e-01, 8.54005976242490e-01, 8.44781289985590e-01, 8.35863299964501e-01, 8.27292074629156e-01, 8.19102524354079e-01, 8.11324256504207e-01, 8.03981484020781e-01, 7.97092990093759e-01, 7.90672151159683e-01, 7.84727019438492e-01, 7.79260464960174e-01, 7.74270375836394e-01, 7.69749914556343e-01, 7.65687827344181e-01, 7.62068803054206e-01, 7.58873877595150e-01, 7.56080879365641e-01, 7.53664910562705e-01, 7.51598858444595e-01, 7.49853929673392e-01, 7.48400199761349e-01, 7.47207168458786e-01, 7.46244310744871e-01, 7.45481612031156e-01, 7.44890075382258e-01, 7.44442188129085e-01, 7.44112335299477e-01, 7.43877147908051e-01, 7.43715775372976e-01, 7.43610073162088e-01, 7.43544699164588e-01, 7.43507115137575e-01, 7.43487492751882e-01, 7.43478527090630e-01, 7.43475163756514e-01, 7.43474248847619e-01, 7.43474113808944e-01, 7.43474103356419e-01, 7.43473915181568e-01, 7.43472640269431e-01, 7.43467953558579e-01, 7.43455460516245e-01, 7.43428119839692e-01, 7.43375759330178e-01, 7.43284704454553e-01, 7.43137536877214e-01, 7.42912997642009e-01, 7.42586046618371e-01, 7.42128086177047e-01, 7.41507352697639e-01, 7.40689474328192e-01, 7.39638187386859e-01, 7.38316196989658e-01, 7.36686160140053e-01, 7.34711762011173e-01, 7.32358849053881e-01, 7.29596576544779e-01, 7.26398523977878e-01, 7.22743729964350e-01, 7.18617599529168e-01, 7.14012641092580e-01, 7.08928997857299e-01, 7.03374748301556e-01, 6.97365962212004e-01, 6.90926511207934e-01, 6.84087644982498e-01, 6.76887355576739e-01, 6.69369561142361e-01, 6.61583147309202e-01, 6.53580908173552e-01, 6.45418430027490e-01, 6.37152959437060e-01, 6.28842293525125e-01, 6.20543724845717e-01, 6.12313066681622e-01, 6.04203777611332e-01, 5.96266197397412e-01, 5.88546900149964e-01, 5.81088165662351e-01, 5.73927565964155e-01, 5.67097661490521e-01, 5.60625799689348e-01, 5.54534008176645e-01, 5.48838974462569e-01, 5.43552104587178e-01, 5.38679653531247e-01, 5.34222920862206e-01, 5.30178505632335e-01, 5.26538614993484e-01, 5.23291421281570e-01, 5.20421462421965e-01, 5.17910080388162e-01, 5.15735892108125e-01, 5.13875286654340e-01, 5.12302941806681e-01, 5.10992352185554e-01, 5.09916360186921e-01, 5.09047680002553e-01, 5.08359404180155e-01, 5.07825481581236e-01, 5.07421155337980e-01, 5.07123349582187e-01, 5.06910994390367e-01, 5.06765279588109e-01, 5.06669829777351e-01, 5.06610795141142e-01, 5.06576855152971e-01, 5.06559135149303e-01, 5.06551038671361e-01, 5.06548001392943e-01, 5.06547175178625e-01, 5.06547053231088e-01, 5.06547043715344e-01, 5.06546871416504e-01, 5.06545704066454e-01, 5.06541412761678e-01, 5.06529973699728e-01, 5.06504939495741e-01, 5.06456995630003e-01, 5.06373619686173e-01, 5.06238858881446e-01, 5.06033238883106e-01, 5.05733814003008e-01, 5.05314365482533e-01, 5.04745750646478e-01, 5.03996401155987e-01, 5.03032963420551e-01, 5.01821068508065e-01, 5.00326212795571e-01, 4.98514724417873e-01, 4.96354784702220e-01, 4.93817468710944e-01, 4.90877765286797e-01, 4.87515535117481e-01, 4.83716365719957e-01, 4.79472285126477e-01, 4.74782301432995e-01, 4.69652742974930e-01, 4.64097383197426e-01, 4.58137344559916e-01, 4.51800786222961e-01, 4.45122389987944e-01, 4.38142667299715e-01, 4.30907116597993e-01, 4.23465264684225e-01, 4.15869628075697e-01, 4.08174630748704e-01, 4.00435513535172e-01, 3.92707268064825e-01, 3.85043624825769e-01, 3.77496120877623e-01, 3.70113268160694e-01, 3.62939838348083e-01, 3.56016274937886e-01, 3.49378237979636e-01, 3.43056281709252e-01, 3.37075660697209e-01, 3.31456256153032e-01, 3.26212610984925e-01, 3.21354060213949e-01, 3.16884942415868e-01, 3.12804877929437e-01, 3.09109100464468e-01, 3.05788830227868e-01, 3.02831678502571e-01, 3.00222075492300e-01, 2.97941714951378e-01, 2.95970010466078e-01, 2.94284559117656e-01, 2.92861608590982e-01, 2.91676523606000e-01, 2.90704246931597e-01, 2.89919749311224e-01, 2.89298461562623e-01, 2.88816681080320e-01, 2.88451944154602e-01, 2.88183355077919e-01, 2.87991863061421e-01, 2.87860478604784e-01, 2.87774422167246e-01, 2.87721199743856e-01, 2.87690602164375e-01, 2.87674627478879e-01, 2.87667328515070e-01, 2.87664590422082e-01, 2.87663845594423e-01, 2.87663735659422e-01, 2.87663727008038e-01, 2.87663569424723e-01, 2.87662501775202e-01, 2.87658576987198e-01, 2.87648114979291e-01, 2.87625219313199e-01, 2.87581372072552e-01, 2.87505123541791e-01, 2.87381891527374e-01, 2.87193883139281e-01, 2.86920149443588e-01, 2.86536780489826e-01, 2.86017244687841e-01, 2.85332872260327e-01, 2.84453477495930e-01, 2.83348108817634e-01, 2.81985909426510e-01, 2.80337064751180e-01, 2.78373806525451e-01, 2.76071437509700e-01, 2.73409336202619e-01, 2.70371897898201e-01, 2.66949367605932e-01, 2.63138522029351e-01, 2.58943162174930e-01, 2.54374385218739e-01, 2.49450613739469e-01, 2.44197371866306e-01, 2.38646810624346e-01, 2.32836997966999e-01, 2.26811001755124e-01, 2.20615805325551e-01, 2.14301104386621e-01, 2.07918040021667e-01, 2.01517925018734e-01, 1.95151019333352e-01, 1.88865405333701e-01, 1.82706005048544e-01, 1.76713770740719e-01, 1.70925067796958e-01, 1.65371256307736e-01, 1.60078465907944e-01, 1.55067548395110e-01, 1.50354184988885e-01, 1.45949120175885e-01, 1.41858491912998e-01, 1.38084228277066e-01, 1.34624482996529e-01, 1.31474086122850e-01, 1.28624990803402e-01, 1.26066702162873e-01, 1.23786679220331e-01, 1.21770705208178e-01, 1.20003225372557e-01, 1.18467654175461e-01, 1.17146655734118e-01, 1.16022402333527e-01, 1.15076816008909e-01, 1.14291797624873e-01, 1.13649446728090e-01, 1.13132273887786e-01, 1.12723405450822e-01, 1.12406778817729e-01, 1.12167324678367e-01, 1.11991131301082e-01, 1.11865585082015e-01, 1.11779481228467e-01, 1.11723098709760e-01, 1.11688234448224e-01, 1.11668193070173e-01, 1.11657730276946e-01, 1.11652949880110e-01, 1.11651156610847e-01, 1.11650668800252e-01, 1.11650596800516e-01, 1.11650591084285e-01, 1.11650486327264e-01, 1.11649776584372e-01, 1.11647167514306e-01, 1.11640212843767e-01, 1.11624993490523e-01, 1.11595849571074e-01, 1.11545177322583e-01, 1.11463302394725e-01, 1.11338440235329e-01, 1.11156753223185e-01, 1.10902512568103e-01, 1.10558370558678e-01, 1.10105745287683e-01, 1.09525315391068e-01, 1.08797616600455e-01, 1.07903725195811e-01, 1.06826006118013e-01, 1.05548896111841e-01, 1.04059685545906e-01, 1.02349257324860e-01, 1.00412738422094e-01, 9.82500197719958e-02, 9.58661041064914e-02, 9.32712489978798e-02, 9.04808836601916e-02, 8.75152922640297e-02, 8.43990725075230e-02, 8.11603944680570e-02, 7.78300996742018e-02, 7.44406922518976e-02, 7.10252815582209e-02, 6.76165380457121e-02, 6.42457209682448e-02, 6.09418283888914e-02, 5.77309078462775e-02, 5.46355514874932e-02, 5.16745841925399e-02, 4.88629388532900e-02, 4.62117008880085e-02, 4.37282951852751e-02, 4.14167833746859e-02, 3.92782375281939e-02, 3.73111576319332e-02, 3.55119037131573e-02, 3.38751185445858e-02, 3.23941225920805e-02, 3.10612686533744e-02, 2.98682489610371e-02, 2.88063520723227e-02, 2.78666704935944e-02, 2.70402626608330e-02, 2.63182746883928e-02, 2.56920283237747e-02, 2.51530819409993e-02, 2.46932713082231e-02, 2.43047363921391e-02, 2.39799397148204e-02, 2.37116808387725e-02, 2.34931104874505e-02, 2.33177466730581e-02, 2.31794940537497e-02, 2.30726666385550e-02, 2.29920129560099e-02, 2.29327419640413e-02, 2.28905473538106e-02, 2.28616275337572e-02, 2.28426984929087e-02, 2.28309969371594e-02, 2.28242715407703e-02, 2.28207608103759e-02, 2.28191568502893e-02, 2.28185551691524e-02, 2.28183914991575e-02, 2.28183673418556e-02, 2.28183654064082e-02, 2.28183297156477e-02, 2.28180879065150e-02, 2.28171990066386e-02, 2.28148296495760e-02, 2.28096449967126e-02, 2.27997181689830e-02, 2.27824628497899e-02, 2.27545938428619e-02, 2.27121204512457e-02, 2.26503774473485e-02, 2.25640979590764e-02, 2.24475316184112e-02, 2.22946096356793e-02, 2.20991559726826e-02, 2.18551404744147e-02, 2.15569658217589e-02, 2.11997757710805e-02, 2.07797678255162e-02, 2.02944898253747e-02, 1.97430976295720e-02, 1.91265507164076e-02, 1.84477246385910e-02, 1.77114240157437e-02, 1.69242869450983e-02, 1.60945807483301e-02, 1.52318988550348e-02, 1.43467781107489e-02, 1.34502635923610e-02, 1.25534529948368e-02, 1.16670540984103e-02, 1.08009865412129e-02, 9.96405350502982e-03, 9.16370086086715e-03, 8.40587202861798e-03, 7.69495758876815e-03, 7.03383069270558e-03, 6.42395338662993e-03, 5.86553549311475e-03, 5.35772667691958e-03, 4.89882339937253e-03, 4.48647507725752e-03, 4.11787726071111e-03, 3.78994342911199e-03, 3.49945058911771e-03, 3.24315692061816e-03, 3.01789208639036e-03, 2.82062246820658e-03, 2.64849457896332e-03, 2.49886034235658e-03, 2.36928796749680e-03, 2.25756190828008e-03, 2.16167500083486e-03, 2.07981540377489e-03, 2.01035049168752e-03, 1.95180940798423e-03, 1.90286558915886e-03, 1.86232023293087e-03, 1.82908739347454e-03, 1.80218114194020e-03, 1.78070501944205e-03, 1.76384382470474e-03, 1.75085761709681e-03, 1.74107767432474e-03, 1.73390402861173e-03, 1.72880411649601e-03, 1.72531202485970e-03, 1.72302780117661e-03, 1.72161632281131e-03, 1.72080528609797e-03, 1.72038197455038e-03, 1.72018858802403e-03, 1.72011604658451e-03, 1.72009631398851e-03, 1.72009340153068e-03, 1.72009316598576e-03, 1.72008879502528e-03, 1.72005918150830e-03, 1.71995032270900e-03, 1.71966017349234e-03, 1.71902533324442e-03, 1.71781009042604e-03, 1.71569850017761e-03, 1.71229024201125e-03, 1.70710107433081e-03, 1.69956873384081e-03, 1.68906508099553e-03, 1.67491512260654e-03, 1.65642320766324e-03, 1.63290616634981e-03, 1.60373244949596e-03, 1.56836546558736e-03, 1.52640839739604e-03, 1.47764694411023e-03, 1.42208584715320e-03, 1.35997489307726e-03, 1.29182048506493e-03, 1.21837990147355e-03, 1.14063696901212e-03, 1.05975989609419e-03, 9.77044150786810e-04, 8.93845175838237e-04, 8.11507071702522e-04, 7.31293890012807e-04, 6.54329757850292e-04, 5.81552770171356e-04, 5.13685694684796e-04, 4.51224384818261e-04, 3.94442745829338e-04, 3.43411450515418e-04, 2.98026560557219e-04, 2.58043853150980e-04, 2.23114974303940e-04, 1.92822310026486e-04, 1.66710184345220e-04, 1.44310665902481e-04, 1.25163452437960e-04, 1.08830399568660e-04, 9.49053220580254e-05, 8.30198375751934e-05, 7.28464952288732e-05, 6.40994566367681e-05, 5.65324383927646e-05, 4.99357966816660e-05, 4.41350281477071e-05, 3.89893387620077e-05, 3.43877549511262e-05, 3.02436854298094e-05, 2.64905408471350e-05, 2.30787596109394e-05, 1.99728679177724e-05, 1.71480635486887e-05, 1.45869309449154e-05, 1.22767899028986e-05, 1.02076599262091e-05, 8.37068301962063e-06, 6.75695575580448e-06, 5.35675947332948e-06, 4.15912759876846e-06, 3.15165632969183e-06, 2.32048088832926e-06, 1.65036888688519e-06, 1.12490104697827e-06, 7.26717559018653e-07, 4.37811038977998e-07, 1.42108547152020e-14, 0.00000000000000e+00, 1.42108547152020e-14, -1.42108547152020e-14, 0.00000000000000e+00],
[ 9.60000000000000e+01, 9.59999998822164e+01, 9.59999962401143e+01, 9.59999715642933e+01, 9.59998808527188e+01, 9.59996390439999e+01, 9.59991098263354e+01, 9.59980962146505e+01, 9.59963332333581e+01, 9.59934829761288e+01, 9.59891322396642e+01, 9.59827928491266e+01, 9.59739047126024e+01, 9.59618415651675e+01, 9.59459192940819e+01, 9.59254066789527e+01, 9.58995383364078e+01, 9.58675296276653e+01, 9.58285932662087e+01, 9.57819573455080e+01, 9.57268844850331e+01, 9.56626917577137e+01, 9.55887710062916e+01, 9.55046090769389e+01, 9.54098074002649e+01, 9.53041002447282e+01, 9.51873708753035e+01, 9.50596647953645e+01, 9.49211992561233e+01, 9.47723683034498e+01, 9.46137428028241e+01, 9.44460651312437e+01, 9.42702385271114e+01, 9.40873114113376e+01, 9.38984572960637e+01, 9.37049511451544e+01, 9.35081432159415e+01, 9.33094314815979e+01, 9.31102337100924e+01, 9.29119601741556e+01, 9.27159878110036e+01, 9.25236364678129e+01, 9.23361476843042e+01, 9.21546662967166e+01, 9.19802250098801e+01, 9.18137319804190e+01, 9.16559613825705e+01, 9.15075468829116e+01, 9.13689779237645e+01, 9.12405986995503e+01, 9.11226096993580e+01, 9.10150716778094e+01, 9.09179119020523e+01, 9.08309325041974e+01, 9.07538207456980e+01, 9.06861609738745e+01, 9.06274480223165e+01, 9.05771017778201e+01, 9.05344826085477e+01, 9.04989073230024e+01, 9.04696653089535e+01, 9.04460344873902e+01, 9.04272967105439e+01, 9.04127522364578e+01, 9.04017329266395e+01, 9.03936138387245e+01, 9.03878229230087e+01, 9.03838485797488e+01, 9.03812448922263e+01, 9.03796344170089e+01, 9.03787084853766e+01, 9.03782250458027e+01, 9.03780041537435e+01, 9.03779212887740e+01, 9.03778987474240e+01, 9.03778954203655e+01, 9.03778949839256e+01, 9.03778848153996e+01, 9.03778159221275e+01, 9.03775626629825e+01, 9.03768875656334e+01, 9.03754101242260e+01, 9.03725806189297e+01, 9.03676599970310e+01, 9.03597067262321e+01, 9.03475713812720e+01, 9.03298995519938e+01, 9.03051434601819e+01, 9.02715824392380e+01, 9.02273521613884e+01, 9.01704821906045e+01, 9.00989410989547e+01, 9.00106880187767e+01, 8.99037291293155e+01, 8.97761772193030e+01, 8.96263121601826e+01, 8.94526399093549e+01, 8.92539475824634e+01, 8.90293522252617e+01, 8.87783411962831e+01, 8.85008025250075e+01, 8.81970441776775e+01, 8.78678017475060e+01, 8.75142345765923e+01, 8.71379106301931e+01, 8.67407805731323e+01, 8.63251415423592e+01, 8.58935912564392e+01, 8.54489735430366e+01, 8.49943171603670e+01, 8.45327707581157e+01, 8.40675375596910e+01, 8.36018134086727e+01, 8.31387310200741e+01, 8.26813118522724e+01, 8.22324255296133e+01, 8.17947557499580e+01, 8.13707713452859e+01, 8.09627014931931e+01, 8.05725146632836e+01, 8.02019014045639e+01, 7.98522613675290e+01, 7.95246949930860e+01, 7.92200001667229e+01, 7.89386739285817e+01, 7.86809191248265e+01, 7.84466557242748e+01, 7.82355364178272e+01, 7.80469660591244e+01, 7.78801244773794e+01, 7.77339921812289e+01, 7.76073784628204e+01, 7.74989513960987e+01, 7.74072691989571e+01, 7.73308123960538e+01, 7.72680161810253e+01, 7.72173023388326e+01, 7.71771100573369e+01, 7.71459249384296e+01, 7.71223055191778e+01, 7.71049066374332e+01, 7.70924990275342e+01, 7.70839846114346e+01, 7.70784070579682e+01, 7.70749573149027e+01, 7.70729739698459e+01, 7.70719384602226e+01, 7.70714653216688e+01, 7.70712878301479e+01, 7.70712395480643e+01, 7.70712324217268e+01, 7.70712317926068e+01, 7.70712194653110e+01, 7.70711359460561e+01, 7.70708289202584e+01, 7.70700105008108e+01, 7.70682194027844e+01, 7.70647892086487e+01, 7.70588239873163e+01, 7.70491823756349e+01, 7.70344710516528e+01, 7.70130483213657e+01, 7.69830382993893e+01, 7.69423558835336e+01, 7.68887423996069e+01, 7.68198114255252e+01, 7.67331038979756e+01, 7.66261511724597e+01, 7.64965442681320e+01, 7.63420071092333e+01, 7.61604712073387e+01, 7.59501489478103e+01, 7.57096024833694e+01, 7.54378052263868e+01, 7.51341930903759e+01, 7.47987029713647e+01, 7.44317964805809e+01, 7.40344676259696e+01, 7.36082339578515e+01, 7.31551115874497e+01, 7.26775753765835e+01, 7.21785063860190e+01, 7.16611292613699e+01, 7.11289425547832e+01, 7.05856450009800e+01, 7.00350605215223e+01, 6.94810643105404e+01, 6.89275118762160e+01, 6.83781724823978e+01, 6.78366681153434e+01, 6.73064188912600e+01, 6.67905956695392e+01, 6.62920804745986e+01, 6.58134351062122e+01, 6.53568780285624e+01, 6.49242693101178e+01, 6.45171030997184e+01, 6.41365069213804e+01, 6.37832469758071e+01, 6.34577386427949e+01, 6.31600614556212e+01, 6.28899779286879e+01, 6.26469557307542e+01, 6.24301927865692e+01, 6.22386449493162e+01, 6.20710559129460e+01, 6.19259890298569e+01, 6.18018606703206e+01, 6.16969747111936e+01, 6.16095576788977e+01, 6.15377940019777e+01, 6.14798607589118e+01, 6.14339612449335e+01, 6.13983566352512e+01, 6.13713949986647e+01, 6.13515369215748e+01, 6.13373770423568e+01, 6.13276608723252e+01, 6.13212963915372e+01, 6.13173600521250e+01, 6.13150969926250e+01, 6.13139154555966e+01, 6.13133755979115e+01, 6.13131730780640e+01, 6.13131179876992e+01, 6.13131098564747e+01, 6.13131092042499e+01, 6.13130971673946e+01, 6.13130156159253e+01, 6.13127158240523e+01, 6.13119166886410e+01, 6.13101677985794e+01, 6.13068184581414e+01, 6.13009939058840e+01, 6.12915798241273e+01, 6.12772160638047e+01, 6.12563003108281e+01, 6.12270021866703e+01, 6.11872880008589e+01, 6.11349560513852e+01, 6.10676819993312e+01, 6.09830734309705e+01, 6.08787322766023e+01, 6.07523233016480e+01, 6.06016464519788e+01, 6.04247104590260e+01, 6.02198048317779e+01, 5.99855672219869e+01, 5.97210431782515e+01, 5.94257355226886e+01, 5.90996409896626e+01, 5.87432723358092e+01, 5.83576648193242e+01, 5.79443666945477e+01, 5.75054141124881e+01, 5.70432915061565e+01, 5.65608791375867e+01, 5.60613899774510e+01, 5.55482984762492e+01, 5.50252640644858e+01, 5.44960523724468e+01, 5.39644571605741e+01, 5.34342257713228e+01, 5.29089905429526e+01, 5.23922080875409e+01, 5.18871076874982e+01, 5.13966493893143e+01, 5.09234917566442e+01, 5.04699687564461e+01, 5.00380749292926e+01, 4.96294578402601e+01, 4.92454167930233e+01, 4.88869068736082e+01, 4.85545475254979e+01, 4.82486350057501e+01, 4.79691582063517e+01, 4.77158174328984e+01, 4.74880458102700e+01, 4.72850330342706e+01, 4.71057512131645e+01, 4.69489825472511e+01, 4.68133485806253e+01, 4.66973407287535e+01, 4.65993517402319e+01, 4.65177076936958e+01, 4.64507000653854e+01, 4.63966173350165e+01, 4.63537755343554e+01, 4.63205470920451e+01, 4.62953872975562e+01, 4.62768577035565e+01, 4.62636458147532e+01, 4.62545804752251e+01, 4.62486424654835e+01, 4.62449699520738e+01, 4.62428585907279e+01, 4.62417562608862e+01, 4.62412525952353e+01, 4.62410636525308e+01, 4.62410122555028e+01, 4.62410046694080e+01, 4.62410040842953e+01, 4.62409935775507e+01, 4.62409223928111e+01, 4.62406607101969e+01, 4.62399631605659e+01, 4.62384365909789e+01, 4.62355130303472e+01, 4.62304289424843e+01, 4.62222117279280e+01, 4.62096742894468e+01, 4.61914183046984e+01, 4.61658466470004e+01, 4.61311851552394e+01, 4.60855136717375e+01, 4.60268059410443e+01, 4.59529775972306e+01, 4.58619410735311e+01, 4.57516658652484e+01, 4.56202421916222e+01, 4.54659457678538e+01, 4.52873011505709e+01, 4.50831409928956e+01, 4.48526585659426e+01, 4.45954510864791e+01, 4.43115517334992e+01, 4.40014487196728e+01, 4.36660903717905e+01, 4.33068758232688e+01, 4.29256315871541e+01, 4.25245749234211e+01, 4.21062655145731e+01, 4.16735475007952e+01, 4.12294843788358e+01, 4.07772896012668e+01, 4.03202558717057e+01, 3.98616860617399e+01, 3.94048283483530e+01, 3.89528176140002e+01, 3.85086244559601e+01, 3.80750124476706e+01, 3.76545037061741e+01, 3.72493524196896e+01, 3.68615257823987e+01, 3.64926917202715e+01, 3.61442128018115e+01, 3.58171457519577e+01, 3.55122459953937e+01, 3.52299766444944e+01, 3.49705213312674e+01, 3.47338002796908e+01, 3.45194890367537e+01, 3.43270393295768e+01, 3.41557015858234e+01, 3.40045487333974e+01, 3.38725009698558e+01, 3.37583512503494e+01, 3.36607912771461e+01, 3.35784377802070e+01, 3.35098588575396e+01, 3.34536001006014e+01, 3.34082101712043e+01, 3.33722654313719e+01, 3.33443931664684e+01, 3.33232928944464e+01, 3.33077552289097e+01, 3.32966777674901e+01, 3.32890775137859e+01, 3.32840994117480e+01, 3.32810206737476e+01, 3.32792507125302e+01, 3.32783266354169e+01, 3.32779044174435e+01, 3.32777460289846e+01, 3.32777029434910e+01, 3.32776965841627e+01, 3.32776961026535e+01, 3.32776875727674e+01, 3.32776297815471e+01, 3.32774173351083e+01, 3.32768510324417e+01, 3.32756117007188e+01, 3.32732382621478e+01, 3.32691109262945e+01, 3.32624403061376e+01, 3.32522631443753e+01, 3.32374452040028e+01, 3.32166917177402e+01, 3.31885655980740e+01, 3.31515133784207e+01, 3.31038985842635e+01, 3.30440419238312e+01, 3.29702673497758e+01, 3.28809526917827e+01, 3.27745832169602e+01, 3.26498061671314e+01, 3.25054840795529e+01, 3.23407445495290e+01, 3.21550240654313e+01, 3.19481036571071e+01, 3.17201343560734e+01, 3.14716508674304e+01, 3.12035723847058e+01, 3.09171901143702e+01, 3.06141417808643e+01, 3.02963741105671e+01, 2.99660949909242e+01, 2.96257176097387e+01, 2.92777993391473e+01, 2.89249783854993e+01, 2.85699112434903e+01, 2.82152137605160e+01, 2.78634081584210e+01, 2.75168777306096e+01, 2.71778302136236e+01, 2.68482701148588e+01, 2.65299796465039e+01, 2.62245074326984e+01, 2.59331638541824e+01, 2.56570217719981e+01, 2.53969214022538e+01, 2.51534782544384e+01, 2.49270932482152e+01, 2.47179643446584e+01, 2.45260992348448e+01, 2.43513288016892e+01, 2.41933212016037e+01, 2.40515965011976e+01, 2.39255418559519e+01, 2.38144272395007e+01, 2.37174217304680e+01, 2.36336103440889e+01, 2.35620113620962e+01, 2.35015940696940e+01, 2.34512967557012e+01, 2.34100447742666e+01, 2.33767684077048e+01, 2.33504202144196e+01, 2.33299914985683e+01, 2.33145275041350e+01, 2.33031409201404e+01, 2.32950232895458e+01, 2.32894539442825e+01, 2.32858061431943e+01, 2.32835501668836e+01, 2.32822532199111e+01, 2.32815761011707e+01, 2.32812667210441e+01, 2.32811506620424e+01, 2.32811190911954e+01, 2.32811144314051e+01, 2.32811140815932e+01, 2.32811079245070e+01, 2.32810662093552e+01, 2.32809128602388e+01, 2.32805040892901e+01, 2.32796095116439e+01, 2.32778963174613e+01, 2.32749171492191e+01, 2.32701022650205e+01, 2.32627564894749e+01, 2.32520613587486e+01, 2.32370827504053e+01, 2.32167841459857e+01, 2.31900455003341e+01, 2.31556874844592e+01, 2.31125006303846e+01, 2.30592786447006e+01, 2.29948548864092e+01, 2.29181407444392e+01, 2.28281644262074e+01, 2.27241085085559e+01, 2.26053445327487e+01, 2.24714629664683e+01, 2.23222970179470e+01, 2.21579390663335e+01, 2.19787488483258e+01, 2.17853529800485e+01, 2.15786358519406e+01, 2.13597223684900e+01, 2.11299533769645e+01, 2.08908549180063e+01, 2.06441026327046e+01, 2.03914827877762e+01, 2.01348514513815e+01, 1.98760933796370e+01, 1.96170821549885e+01, 1.93596430320250e+01, 1.91055197666193e+01, 1.88563464142327e+01, 1.86136246952949e+01, 1.83787070869276e+01, 1.81527853817235e+01, 1.79368841264509e+01, 1.77318581623025e+01, 1.75383934414419e+01, 1.73570103649446e+01, 1.71880690278608e+01, 1.70317759197512e+01, 1.68881917773534e+01, 1.67572404007250e+01, 1.66387183209083e+01, 1.65323052508945e+01, 1.64375752710172e+01, 1.63540087030088e+01, 1.62810046195332e+01, 1.62178939212248e+01, 1.61639528924008e+01, 1.61184171201617e+01, 1.60804956301401e+01, 1.60493850569748e+01, 1.60242836309003e+01, 1.60044047267547e+01, 1.59889896919118e+01, 1.59773196490323e+01, 1.59687259616960e+01, 1.59625990587605e+01, 1.59583953384093e+01, 1.59556419155846e+01, 1.59539390356392e+01, 1.59529600498901e+01, 1.59524489314268e+01, 1.59522153972535e+01, 1.59521277905345e+01, 1.59521039593840e+01, 1.59521004419561e+01, 1.59521001787047e+01, 1.59520955558770e+01, 1.59520642355446e+01, 1.59519490987639e+01, 1.59516421869512e+01, 1.59509705208577e+01, 1.59496842110548e+01, 1.59474473422995e+01, 1.59438320610608e+01, 1.59383162334614e+01, 1.59302849675028e+01, 1.59190362054192e+01, 1.59037904858959e+01, 1.58837048489668e+01, 1.58578907076954e+01, 1.58254353420149e+01, 1.57854264867278e+01, 1.57369792970765e+01, 1.56792647948999e+01, 1.56115387426601e+01, 1.55331697794458e+01, 1.54436655992458e+01, 1.53426959702522e+01, 1.52301114909003e+01, 1.51059571513243e+01, 1.49704800062613e+01, 1.48241305477518e+01, 1.46675576692774e+01, 1.45015974130431e+01, 1.43272559686768e+01, 1.41456876309494e+01, 1.39581686192015e+01, 1.37660678095141e+01, 1.35708155310392e+01, 1.33738716275950e+01, 1.31766939798408e+01, 1.29807086170443e+01, 1.27872824188021e+01, 1.25976992208204e+01, 1.24131399079266e+01, 1.22346668220983e+01, 1.20632125577623e+01, 1.18995729849031e+01, 1.17444041520678e+01, 1.15982225882519e+01, 1.14614084486187e+01, 1.13342109302465e+01, 1.12167554112391e+01, 1.11090518271416e+01, 1.10110038794755e+01, 1.09224187601849e+01, 1.08430171627997e+01, 1.07724434286614e+01, 1.07102757396409e+01, 1.06560363146381e+01, 1.06092015948855e+01, 1.05692124132028e+01, 1.05354841365764e+01, 1.05074167523368e+01, 1.04844048391005e+01, 1.04658473283982e+01, 1.04511569257128e+01, 1.04397690248595e+01, 1.04311499213802e+01, 1.04248041126030e+01, 1.04202804670806e+01, 1.04171770560668e+01, 1.04151444650041e+01, 1.04138874428150e+01, 1.04131647989328e+01, 1.04127875192289e+01, 1.04126151382783e+01, 1.04125504724368e+01, 1.04125328817792e+01, 1.04125302854357e+01, 1.04125300910474e+01, 1.04125266765187e+01, 1.04125035425965e+01, 1.04124185000238e+01, 1.04121918090742e+01, 1.04116957082042e+01, 1.04107456414730e+01, 1.04090935481516e+01, 1.04064235428273e+01, 1.04023502729289e+01, 1.03964201903936e+01, 1.03881159118290e+01, 1.03768637634774e+01, 1.03620445106713e+01, 1.03430071548512e+01, 1.03190855455051e+01, 1.02896174035062e+01, 1.02539651935925e+01, 1.02115381279099e+01, 1.01618144432070e+01, 1.01043629868008e+01, 1.00388630863082e+01, 9.96512167894974e+00, 9.88308674748536e+00, 9.79285625493510e+00, 9.69468198491856e+00, 9.58896796619784e+00, 9.47626346851784e+00, 9.35725087581685e+00, 9.23272904304628e+00, 9.10359299575798e+00, 8.97081101375809e+00, 8.83540023638432e+00, 8.69840193283214e+00, 8.56085750287380e+00, 8.42378612701097e+00, 8.28816479238759e+00, 8.15491120554016e+00, 8.02486988741153e+00, 7.89880154760444e+00, 7.77737566536567e+00, 7.66116607032363e+00, 7.55064921804382e+00, 7.44620479244557e+00, 7.34811823578610e+00, 7.25658480318057e+00, 7.17171475798042e+00, 7.09353936191880e+00, 7.02201736454744e+00, 6.95704175472551e+00, 6.89844659742839e+00, 6.84601383702496e+00, 6.79947999932193e+00, 6.75854276605774e+00, 6.72286742526238e+00, 6.69209321822521e+00, 6.66583960897425e+00, 6.64371249624251e+00, 6.62531037262059e+00, 6.61023041324695e+00, 6.59807444960920e+00, 6.58845475573806e+00, 6.58099954725661e+00, 6.57535807130796e+00, 6.57120514991674e+00, 6.56824503291134e+00, 6.56621442047494e+00, 6.56488453014791e+00, 6.56406210815669e+00, 6.56358931884581e+00, 6.56334248650286e+00, 6.56322970820194e+00, 6.56318740139992e+00, 6.56317589294589e+00, 6.56317419432297e+00, 6.56317406680962e+00, 6.56317182246806e+00, 6.56315661673537e+00, 6.56310071902864e+00, 6.56295171761067e+00, 6.56262563847844e+00, 6.56200118408974e+00, 6.56091533452810e+00, 6.55916053274450e+00, 6.55648365054745e+00, 6.55258690089643e+00, 6.54713082201414e+00, 6.53973940693706e+00, 6.53000738589175e+00, 6.51750958699394e+00, 6.50181220372086e+00, 6.48248568833822e+00, 6.45911887464848e+00, 6.43133381950793e+00, 6.39880075125481e+00, 6.36125243657129e+00, 6.31849723736919e+00, 6.27043013631221e+00, 6.21704107030085e+00, 6.15842002734094e+00, 6.09475852930470e+00, 6.02634733044597e+00, 5.95357039273637e+00, 5.87689543353635e+00, 5.79686155630968e+00, 5.71406464942004e+00, 5.62914135377810e+00, 5.54275244611112e+00, 5.45556645835113e+00, 5.36824426179135e+00, 5.28142520223399e+00, 5.19571519995684e+00, 5.11167704842545e+00, 5.02982297853816e+00, 4.95060941576168e+00, 4.87443375382811e+00, 4.80163290210739e+00, 4.73248333063354e+00, 4.66720233048407e+00, 4.60595022038072e+00, 4.54883325622707e+00, 4.49590703337100e+00, 4.44718020766558e+00, 4.40261839808167e+00, 4.36214816872925e+00, 4.32566102023999e+00, 4.29301734845802e+00, 4.26405035143060e+00, 4.23856988315038e+00, 4.21636626398845e+00, 4.19721406314014e+00, 4.18087586785524e+00, 4.16710604822113e+00, 4.15565451560599e+00, 4.14627045865511e+00, 4.13870602432837e+00, 4.13271989444880e+00, 4.12808069228285e+00, 4.12457014047519e+00, 4.12198588274208e+00, 4.12014387833354e+00, 4.11888028120615e+00, 4.11805272539003e+00, 4.11754095387256e+00, 4.11724674957954e+00, 4.11709315234364e+00, 4.11702297339278e+00, 4.11699664699023e+00, 4.11698948558407e+00, 4.11698842857602e+00, 4.11698834888993e+00, 4.11698694184599e+00, 4.11697740892370e+00, 4.11694236500524e+00, 4.11684895164706e+00, 4.11664452317173e+00, 4.11625303494671e+00, 4.11557228608643e+00, 4.11447215574849e+00, 4.11279395660492e+00, 4.11035100958081e+00, 4.10693051877509e+00, 4.10229679271980e+00, 4.09619581619383e+00, 4.08836112483323e+00, 4.07852087313293e+00, 4.06640591707077e+00, 4.05175865929786e+00, 4.03434233323318e+00, 4.01395033956356e+00, 3.99041520244165e+00, 3.96361669074311e+00, 3.93348865829384e+00, 3.90002419959198e+00, 3.86327879420482e+00, 3.82337121963522e+00, 3.78048214101287e+00, 3.73485042540287e+00, 3.68676736606857e+00, 3.63656912494021e+00, 3.58462779877894e+00, 3.53134157812335e+00, 3.47712449400161e+00, 3.42239623567372e+00, 3.36757247722485e+00, 3.31305607857804e+00, 3.25922943632413e+00, 3.20644816136401e+00, 3.15503616307632e+00, 3.10528213159309e+00, 3.05743733677849e+00, 3.01171460823831e+00, 2.96828832626164e+00, 2.92729523798620e+00, 2.88883591363502e+00, 2.85297667076952e+00, 2.81975181617453e+00, 2.78916608145433e+00, 2.76119715643148e+00, 2.73579825148798e+00, 2.71290064429790e+00, 2.69241618687016e+00, 2.67423976487584e+00, 2.65825171268889e+00, 2.64432019449686e+00, 2.63230356449149e+00, 2.62205271790953e+00, 2.61341344005665e+00, 2.60622875302809e+00, 2.60034125037127e+00, 2.59559539927980e+00, 2.59183977902241e+00, 2.58892921421356e+00, 2.58672675324101e+00, 2.58510543660374e+00, 2.58394979783678e+00, 2.58315704159435e+00, 2.58263784949733e+00, 2.58231677432593e+00, 2.58213219650371e+00, 2.58203583271714e+00, 2.58199180387133e+00, 2.58197528722249e+00, 2.58197079430215e+00, 2.58197013115684e+00, 2.58197008089134e+00, 2.58196918972804e+00, 2.58196315197043e+00, 2.58194095660799e+00, 2.58188179248303e+00, 2.58175231609567e+00, 2.58150436425649e+00, 2.58107320813433e+00, 2.58037643704279e+00, 2.57931355050664e+00, 2.57776632485947e+00, 2.57560000468713e+00, 2.57266534862084e+00, 2.56880153225805e+00, 2.56383987779066e+00, 2.55760834047322e+00, 2.54993663763227e+00, 2.54066185896156e+00, 2.52963435101373e+00, 2.51672362870900e+00, 2.50182403747893e+00, 2.48485987635055e+00, 2.46578969891510e+00, 2.44460953801230e+00, 2.42135485097045e+00, 2.39610105245866e+00, 2.36896258582230e+00, 2.34009057345665e+00, 2.30966917349511e+00, 2.27791084519552e+00, 2.24505078174900e+00, 2.21134080222878e+00, 2.17704300262830e+00, 2.14242345118241e+00, 2.10774617979263e+00, 2.07326767737628e+00, 2.03923203873489e+00, 2.00586686986547e+00, 1.97338000187891e+00, 1.94195702351355e+00, 1.91175960770691e+00, 1.88292458073566e+00, 1.85556366234728e+00, 1.82976379125616e+00, 1.80558794169804e+00, 1.78307633302184e+00, 1.76224793529713e+00, 1.74310217935290e+00, 1.72562078906279e+00, 1.70976966627235e+00, 1.69550077345370e+00, 1.68275397470423e+00, 1.67145881076753e+00, 1.66153619715715e+00, 1.65290004525474e+00, 1.64545881380900e+00, 1.63911700229892e+00, 1.63377659819220e+00, 1.62933848757042e+00, 1.62570383347855e+00, 1.62277541943938e+00, 1.62045894771597e+00, 1.61866427402008e+00, 1.61730655334364e+00, 1.61630726623986e+00, 1.61559509184521e+00, 1.61510659365001e+00, 1.61478668666368e+00, 1.61458886008288e+00, 1.61447513748813e+00, 1.61441576637887e+00, 1.61438863976164e+00, 1.61437846371722e+00, 1.61437569559455e+00, 1.61437528702572e+00, 1.61437525586028e+00, 1.61437470073147e+00, 1.61437093965657e+00, 1.61435711361349e+00, 1.61432025896708e+00, 1.61423960605346e+00, 1.61408515595683e+00, 1.61381659686762e+00, 1.61338261702163e+00, 1.61272066589821e+00, 1.61175720842297e+00, 1.61040850750686e+00, 1.60858195743907e+00, 1.60617797365808e+00, 1.60309242279713e+00, 1.59921955072645e+00, 1.59445533636988e+00, 1.58870116692802e+00, 1.58186769814252e+00, 1.57387873438788e+00, 1.56467494104911e+00, 1.55421718920891e+00, 1.54248933302239e+00, 1.52950023528349e+00, 1.51528488724662e+00, 1.49990451389284e+00, 1.48344561306128e+00, 1.46601794231508e+00, 1.44775153604731e+00, 1.42879290139719e+00, 1.40930059899932e+00, 1.38944045762162e+00, 1.36938069537395e+00, 1.34928722091745e+00, 1.32931936475939e+00, 1.30962624497083e+00, 1.29034390836593e+00, 1.27159331498929e+00, 1.25347916001100e+00, 1.23608946214848e+00, 1.21949579899614e+00, 1.20375404134863e+00, 1.18890543113591e+00, 1.17497785793536e+00, 1.16198721195968e+00, 1.14993872095283e+00, 1.13882820913140e+00, 1.12864324413331e+00, 1.11936416049517e+00, 1.11096496461644e+00, 1.10341413672995e+00, 1.09667535098936e+00, 1.09070813655305e+00, 1.08546850160592e+00, 1.08090953954644e+00, 1.07698203277630e+00, 1.07363506513958e+00, 1.07081664938460e+00, 1.06847437126568e+00, 1.06655604722044e+00, 1.06501038809294e+00, 1.06378765728493e+00, 1.06284030819857e+00, 1.06212358308719e+00, 1.06159605367300e+00, 1.06122008330594e+00, 1.06096219115651e+00, 1.06079330100376e+00, 1.06068886054021e+00, 1.06062882159439e+00, 1.06059747699932e+00, 1.06058315565646e+00, 1.06057778326863e+00, 1.06057632185286e+00, 1.06057610615110e+00, 1.06057608958133e+00, 1.06057579291370e+00, 1.06057378294901e+00, 1.06056639414034e+00, 1.06054669855173e+00, 1.06050359656936e+00, 1.06042105621819e+00, 1.06027753381288e+00, 1.06004560626722e+00, 1.05969184244988e+00, 1.05917693711544e+00, 1.05845612565048e+00, 1.05747989060939e+00, 1.05619496129717e+00, 1.05454559522040e+00, 1.05247511515685e+00, 1.04992765845527e+00, 1.04685007710691e+00, 1.04319390982690e+00, 1.03891733300064e+00, 1.03398698823221e+00, 1.02837958255219e+00, 1.02208316464068e+00, 1.01509799718669e+00, 1.00743697084647e+00, 9.99125536852944e-01, 9.90201169632186e-01, 9.80712403603789e-01, 9.70717515597130e-01, 9.60282942823241e-01, 9.49481534460361e-01, 9.38390732798879e-01, 9.27090769339060e-01, 9.15662945118584e-01, 9.04188046079966e-01, 8.92744926316595e-01, 8.81409276540140e-01, 8.70252583060122e-01, 8.59341274017794e-01, 8.48736044050639e-01, 8.38491345211324e-01, 8.28655030098986e-01, 8.19268132238562e-01, 8.10364768457006e-01, 8.01972148207042e-01, 7.94110675440976e-01, 7.86794129703665e-01, 7.80029914544954e-01, 7.73819363039451e-01, 7.68158092010495e-01, 7.63036398324175e-01, 7.58439692194628e-01, 7.54348963688514e-01, 7.50741279437975e-01, 7.47590306903348e-01, 7.44866863353693e-01, 7.42539486074264e-01, 7.40575019223300e-01, 7.38939211334284e-01, 7.37597315820523e-01, 7.36514685123680e-01, 7.35657347527763e-01, 7.34992554299623e-01, 7.34489283886575e-01, 7.34118689546619e-01, 7.33854477122733e-01, 7.33673200763263e-01, 7.33554466246744e-01, 7.33481034134655e-01, 7.33438818132541e-01, 7.33416777624299e-01, 7.33406707145321e-01, 7.33402929362796e-01, 7.33401901714529e-01, 7.33401750035796e-01, 7.33401738295242e-01, 7.33401526932667e-01, 7.33400094920526e-01, 7.33394830698643e-01, 7.33380798263681e-01, 7.33350088865777e-01, 7.33291277624929e-01, 7.33189007292580e-01, 7.33023719051388e-01, 7.32771547509302e-01, 7.32404393767197e-01, 7.31890186441163e-01, 7.31193335611536e-01, 7.30275378692525e-01, 7.29095810100503e-01, 7.27613078422294e-01, 7.25785725805281e-01, 7.23573634994750e-01, 7.20939340515301e-01, 7.17849352800499e-01, 7.14275438555120e-01, 7.10195798192785e-01, 7.05596082532850e-01, 7.00470196445508e-01, 6.94820846733109e-01, 6.88659804676263e-01, 6.82007869346464e-01, 6.74894534635072e-01, 6.67357379453563e-01, 6.59441215225741e-01, 6.51197036361452e-01, 6.42680826998287e-01, 6.33952280525989e-01, 6.25073487373555e-01, 6.16107641800369e-01, 6.07117810879160e-01, 5.98165799582418e-01, 5.89311135990499e-01, 5.80610191123370e-01, 5.72115439513908e-01, 5.63874859865322e-01, 5.55931470151762e-01, 5.48322988269518e-01, 5.41081607586975e-01, 5.34233876146345e-01, 5.27800668487117e-01, 5.21797239782316e-01, 5.16233352960057e-01, 5.11113470551905e-01, 5.06437004051904e-01, 5.02198614509041e-01, 4.98388558860654e-01, 4.94993077103899e-01, 4.91994815763206e-01, 4.89373283216750e-01, 4.87105332269962e-01, 4.85165664909971e-01, 4.83527353449773e-01, 4.82162371328954e-01, 4.81042125745390e-01, 4.80137983156297e-01, 4.79421777627820e-01, 4.78866291161211e-01, 4.78445694614081e-01, 4.78135937778291e-01, 4.77915077661606e-01, 4.77763535093800e-01, 4.77664271441171e-01, 4.77602879416494e-01, 4.77567584621497e-01, 4.77549157419290e-01, 4.77540737846780e-01, 4.77537579369354e-01, 4.77536720186407e-01, 4.77536593372804e-01, 4.77536583477317e-01, 4.77536404303237e-01, 4.77535190372635e-01, 4.77530727836381e-01, 4.77518832361596e-01, 4.77492799448683e-01, 4.77442943469612e-01, 4.77356244015141e-01, 4.77216116018667e-01, 4.77002317655007e-01, 4.76691007109565e-01, 4.76254955868484e-01, 4.75663922072670e-01, 4.74885182619786e-01, 4.73884217056650e-01, 4.72625529952182e-01, 4.71073591573514e-01, 4.69193869649146e-01, 4.66953918276133e-01, 4.64324484208831e-01, 4.61280586501559e-01, 4.57802523386363e-01, 4.53876760841666e-01, 4.49496660814745e-01, 4.44663013477580e-01, 4.39384346863946e-01, 4.33676998097582e-01, 4.27564942292676e-01, 4.21079387094139e-01, 4.14258151780217e-01, 4.07144859088376e-01, 3.99787974929438e-01, 3.92239735697416e-01, 3.84555005003222e-01, 3.76790101589378e-01, 3.69001638253891e-01, 3.61245408180324e-01, 3.53575350470237e-01, 3.46042621194556e-01, 3.38694790184491e-01, 3.31575177326769e-01, 3.24722335584255e-01, 3.18169681640640e-01, 3.11945269288998e-01, 3.06071695766690e-01, 3.00566127441328e-01, 2.95440428730700e-01, 2.90701376932617e-01, 2.86350945667508e-01, 2.82386640715103e-01, 2.78801873902893e-01, 2.75586363086902e-01, 2.72726548838534e-01, 2.70206020931752e-01, 2.68005949865739e-01, 2.66105520280348e-01, 2.64482364124248e-01, 2.63112991787835e-01, 2.61973219157142e-01, 2.61038587778032e-01, 2.60284774192129e-01, 2.59687983193203e-01, 2.59225318452849e-01, 2.58875122868346e-01, 2.58617280266009e-01, 2.58433469883883e-01, 2.58307365443287e-01, 2.58224771625791e-01, 2.58173692366753e-01, 2.58144327472010e-01, 2.58128996526736e-01, 2.58121991731443e-01, 2.58119363997409e-01, 2.58118649190664e-01, 2.58118543686692e-01, 2.58118535384000e-01, 2.58118384152311e-01, 2.58117359535902e-01, 2.58113592946444e-01, 2.58103552670363e-01, 2.58081580119224e-01, 2.58039501387088e-01, 2.57966330287942e-01, 2.57848077538540e-01, 2.57667678970648e-01, 2.57405053489521e-01, 2.57037298786059e-01, 2.56539029401850e-01, 2.55882857511764e-01, 2.55040011660270e-01, 2.53981082710268e-01, 2.52676879597573e-01, 2.51099370446227e-01, 2.49222677637420e-01, 2.47024089101572e-01, 2.44485043056628e-01, 2.41592040280153e-01, 2.38337437348022e-01, 2.34720076512701e-01, 2.30745713228515e-01, 2.26427210695817e-01, 2.21784481834263e-01, 2.16844172177346e-01, 2.11639091434463e-01, 2.06207415837056e-01, 2.00591696741270e-01, 1.94837722186847e-01, 1.88993286249897e-01, 1.83106925359127e-01, 1.77226680894947e-01, 1.71398943357929e-01, 1.65667425580850e-01, 1.60072301653628e-01, 1.54649535522893e-01, 1.49430409870021e-01, 1.44441253100737e-01, 1.39703351145158e-01, 1.35233022018468e-01, 1.31041825159201e-01, 1.27136874536106e-01, 1.23521224211572e-01, 1.20194297037784e-01, 1.17152330846338e-01, 1.14388821227507e-01, 1.11894945176957e-01, 1.09659955024057e-01, 1.07671536778369e-01, 1.05916131087000e-01, 1.04379218228118e-01, 1.03045570897265e-01, 1.01899479975913e-01, 1.00924959062397e-01, 1.00105933396307e-01, 9.94264180426967e-02, 9.88706889642259e-02, 9.84234490469760e-02, 9.80699894138013e-02, 9.77963446249390e-02, 9.75894387817817e-02, 9.74372182646874e-02, 9.73287659687685e-02, 9.72543915344630e-02, 9.72056922420848e-02, 9.71755799387353e-02, 9.71582705290075e-02, 9.71492340880502e-02, 9.71451054075914e-02, 9.71435566203240e-02, 9.71431353148375e-02, 9.71430731311074e-02, 9.71430681942280e-02, 9.71429777191730e-02, 9.71423647387297e-02, 9.71401113777119e-02, 9.71341048973642e-02, 9.71209606377386e-02, 9.70957907803154e-02, 9.70520294880544e-02, 9.69813245777829e-02, 9.68735053458545e-02, 9.67166353235456e-02, 9.64971573588684e-02, 9.62001362691609e-02, 9.58096011743663e-02, 9.53089853714317e-02, 9.46816562437931e-02, 9.39115213794253e-02, 9.29836901589596e-02, 9.18851631354727e-02, 9.06055152885870e-02, 8.91375345389918e-02, 8.74777745972324e-02, 8.56269820223190e-02, 8.35903617562224e-02, 8.13776534864843e-02, 7.90030026246455e-02, 7.64846236483834e-02, 7.38442687917171e-02, 7.11065300258014e-02, 6.82980152905799e-02, 6.54464494838720e-02, 6.25797556336778e-02, 5.97251714040112e-02, 5.69084507366853e-02, 5.41531908031203e-02, 5.14803118462765e-02, 4.89077035369263e-02, 4.64500377916153e-02, 4.41187360459452e-02, 4.19220698039595e-02, 3.98653674506875e-02, 3.79512978504266e-02, 3.61802017468449e-02, 3.45504447358138e-02, 3.30587698064733e-02, 3.17006323896578e-02, 3.04705059016896e-02, 2.93621504559241e-02, 2.83688414390042e-02, 2.74835578567263e-02, 2.66991327277992e-02, 2.60083694103059e-02, 2.54041287067963e-02, 2.48793920391535e-02, 2.44273060304181e-02, 2.40412135764387e-02, 2.37146760105276e-02, 2.34414903199109e-02, 2.32157046147421e-02, 2.30316342169772e-02, 2.28838798663702e-02, 2.27673486716213e-02, 2.26772776046005e-02, 2.26092585893554e-02, 2.25592636182483e-02, 2.25236678774223e-02, 2.24992686129326e-02, 2.24832974396350e-02, 2.24734239839108e-02, 2.24677491399774e-02, 2.24647867697882e-02, 2.24634333319216e-02, 2.24629256257742e-02, 2.24627875188332e-02, 2.24627671345701e-02, 2.24627655014160e-02, 2.24627353850479e-02, 2.24625313429954e-02, 2.24617812753252e-02, 2.24597819660772e-02, 2.24554070125578e-02, 2.24470303147797e-02, 2.24324689913118e-02, 2.24089496520463e-02, 2.23731019874407e-02, 2.23209835139215e-02, 2.22481389196787e-02, 2.21496966265704e-02, 2.20205038008885e-02, 2.18552990162948e-02, 2.16489190984248e-02, 2.13965334699395e-02, 2.10938958143885e-02, 2.07375994596930e-02, 2.03253200382107e-02, 1.98560272399533e-02, 1.93301473394030e-02, 1.87496600016175e-02, 1.81181167777709e-02, 1.74405744771775e-02, 1.67234437088837e-02, 1.59742604815278e-02, 1.52013958459634e-02, 1.44137241964870e-02, 1.36202742600915e-02, 1.28298875900725e-02, 1.20509075042889e-02, 1.12909172180338e-02, 1.05565400694729e-02, 9.85330811352014e-03, 9.18559891991322e-03, 8.55663496652426e-03, 7.96853606011894e-03, 7.42241287150591e-03, 6.91848880489943e-03, 6.45623780128801e-03, 6.03452705045981e-03, 5.65175565634490e-03, 5.30598270633220e-03, 4.99504054707245e-03, 4.71663108240250e-03, 4.46840447244767e-03, 4.24802076543074e-03, 4.05319582432639e-03, 3.88173347099041e-03, 3.73154607504489e-03, 3.60066587472829e-03, 3.48724918796961e-03, 3.38957543118568e-03, 3.30604259494294e-03, 3.23516057004269e-03, 3.17554348647775e-03, 3.12590200795171e-03, 3.08503630733981e-03, 3.05183023310462e-03, 3.02524696459727e-03, 3.00432625782321e-03, 2.98818320207772e-03, 2.97600825297195e-03, 2.96706817876213e-03, 2.96070746163934e-03, 2.95634964042790e-03, 2.95349806798129e-03, 2.95173558951944e-03, 2.95072272207619e-03, 2.95019402496166e-03, 2.94995248424357e-03, 2.94986187806501e-03, 2.94983723138387e-03, 2.94983359358980e-03, 2.94983329939669e-03, 2.94982783992737e-03, 2.94979085137470e-03, 2.94965488093624e-03, 2.94929245980313e-03, 2.94849944242515e-03, 2.94698122658588e-03, 2.94434262329446e-03, 2.94008219628950e-03, 2.93359193430831e-03, 2.92416312855437e-03, 2.91099926317884e-03, 2.89323654787665e-03, 2.86997239351194e-03, 2.84030162730176e-03, 2.80335956769306e-03, 2.75837026201486e-03, 2.70469731228218e-03, 2.64189388785496e-03, 2.56974790039521e-03, 2.48831804468352e-03, 2.39795662103859e-03, 2.29931581368703e-03, 2.19333538760225e-03, 2.08121146405915e-03, 1.96434794563061e-03, 1.84429402524700e-03, 1.72267275664240e-03, 1.60110662439195e-03, 1.48114623891615e-03, 1.36420765112177e-03, 1.25152247564131e-03, 1.14410332658110e-03, 1.04272533550898e-03, 9.47923001831441e-04, 8.60000446152753e-04, 7.79052291505877e-04, 7.04991905664087e-04, 6.37583700267896e-04, 5.76476581017005e-04, 5.21236365500499e-04, 4.71375850864320e-04, 4.26381801972298e-04, 3.85738273905645e-04, 3.48945889214747e-04, 3.15536897375068e-04, 2.85085647760752e-04, 2.57214585630550e-04, 2.31597284297891e-04, 2.07959903363530e-04, 1.86080559757329e-04, 1.65785702336052e-04, 1.46944241805613e-04, 1.29460914095603e-04, 1.13269406142535e-04, 9.83251501054383e-05, 8.45980960804127e-05, 7.20661338391437e-05, 6.07096146154618e-05, 5.05070554339682e-05, 4.14319532922036e-05, 3.34506216859154e-05, 2.65209430381219e-05, 2.05918912854486e-05, 1.56036624332501e-05, 1.14882655282145e-05, 8.17045503254121e-06, 5.56891392022862e-06, 3.59762213975046e-06, 2.16736842162391e-06, 1.18738729781853e-06, 5.67129191608728e-07, -7.10542735760100e-14, -7.10542735760100e-14, -7.10542735760100e-14],
[ 9.70000000000000e+01, 9.69999998804240e+01, 9.69999961828977e+01, 9.69999711315693e+01, 9.69998790395787e+01, 9.69996335510951e+01, 9.69990962799381e+01, 9.69980672432527e+01, 9.69962774327541e+01, 9.69933837989788e+01, 9.69889668484370e+01, 9.69825309730070e+01, 9.69735075494548e+01, 9.69612607690766e+01, 9.69450960875518e+01, 9.69242711267919e+01, 9.68980088159495e+01, 9.68655125274954e+01, 9.68259829431016e+01, 9.67786363666835e+01, 9.67227241797586e+01, 9.66575530980382e+01, 9.65825058304844e+01, 9.64970616601958e+01, 9.64008163648815e+01, 9.62935007861702e+01, 9.61749972621409e+01, 9.60453530813677e+01, 9.59047901241620e+01, 9.57537099457349e+01, 9.55926937329559e+01, 9.54224968222767e+01, 9.52440377771691e+01, 9.50583823536537e+01, 9.48667229919027e+01, 9.46703547231862e+01, 9.44706485472443e+01, 9.42690234026585e+01, 9.40669178247847e+01, 9.38657622784946e+01, 9.36669529913477e+01, 9.34718279248532e+01, 9.32816453329632e+01, 9.30975651875393e+01, 9.29206336121050e+01, 9.27517703618074e+01, 9.25917593169825e+01, 9.24412419138174e+01, 9.23007134103310e+01, 9.21705218713335e+01, 9.20508697455751e+01, 9.19418178972400e+01, 9.18432919394849e+01, 9.17550906987319e+01, 9.16768966149854e+01, 9.16082878564499e+01, 9.15487518975538e+01, 9.14977002797590e+01, 9.14544842459787e+01, 9.14184109138417e+01, 9.13887596322050e+01, 9.13647981509868e+01, 9.13457982282274e+01, 9.13310503017259e+01, 9.13198768668264e+01, 9.13116442276422e+01, 9.13057723264285e+01, 9.13017424045087e+01, 9.12991023070500e+01, 9.12974693113519e+01, 9.12965304318579e+01, 9.12960402320949e+01, 9.12958162511951e+01, 9.12957322274860e+01, 9.12957093709300e+01, 9.12957059973476e+01, 9.12957055548048e+01, 9.12956952440877e+01, 9.12956253874496e+01, 9.12953685868715e+01, 9.12946840493774e+01, 9.12931859485209e+01, 9.12903168782042e+01, 9.12853274533257e+01, 9.12772629810793e+01, 9.12649579752443e+01, 9.12470391104848e+01, 9.12219370100599e+01, 9.11879070236187e+01, 9.11430588781853e+01, 9.10853947739468e+01, 9.10128551503630e+01, 9.09233709767839e+01, 9.08149210421705e+01, 9.06855923564040e+01, 9.05336414657144e+01, 9.03575542688479e+01, 9.01561018426596e+01, 8.99283898824971e+01, 8.96738996499301e+01, 8.93925187792356e+01, 8.90845608627170e+01, 8.87507733141372e+01, 8.83923334883532e+01, 8.80108333351687e+01, 8.76082529895984e+01, 8.71869237615714e+01, 8.67494811825336e+01, 8.62988092813667e+01, 8.58379781335244e+01, 8.53701777334980e+01, 8.48986519376331e+01, 8.44266361691749e+01, 8.39573016306157e+01, 8.34937072433886e+01, 8.30387590420813e+01, 8.25951758337132e+01, 8.21654597796908e+01, 8.17518709780811e+01, 8.13564057480692e+01, 8.09807788300609e+01, 8.06264099650526e+01, 8.02944153134041e+01, 7.99856040067186e+01, 7.97004798997951e+01, 7.94392483774198e+01, 7.92018279104741e+01, 7.89878659545871e+01, 7.87967587317223e+01, 7.86276744132443e+01, 7.84795792151076e+01, 7.83512659088347e+01, 7.82413842379799e+01, 7.81484727057185e+01, 7.80709911658898e+01, 7.80073536109122e+01, 7.79559605110143e+01, 7.79152300266561e+01, 7.78836273965119e+01, 7.78596918031025e+01, 7.78420600421197e+01, 7.78294863730623e+01, 7.78208580093676e+01, 7.78152058148785e+01, 7.78117099071101e+01, 7.78097000211791e+01, 7.78086506546088e+01, 7.78081711846515e+01, 7.78079913179944e+01, 7.78079423898155e+01, 7.78079351681156e+01, 7.78079345305770e+01, 7.78079220383213e+01, 7.78078374014390e+01, 7.78075262671267e+01, 7.78066968959019e+01, 7.78048818302528e+01, 7.78014057355343e+01, 7.77953606935865e+01, 7.77855900729432e+01, 7.77706819179700e+01, 7.77489725890183e+01, 7.77185611412868e+01, 7.76773346458644e+01, 7.76230043283212e+01, 7.75531520277412e+01, 7.74652860672997e+01, 7.73569051884741e+01, 7.72255687545685e+01, 7.70689710025801e+01, 7.68850167483085e+01, 7.66718956634413e+01, 7.64281520796722e+01, 7.61527472636048e+01, 7.58451112693848e+01, 7.55051818255603e+01, 7.51334282478559e+01, 7.47308590739252e+01, 7.42990129534191e+01, 7.38399332368940e+01, 7.33561276064576e+01, 7.28505148799267e+01, 7.23263617007234e+01, 7.17872121266105e+01, 7.12368131311203e+01, 7.06790387741085e+01, 7.01178153778563e+01, 6.95570495822463e+01, 6.90005607498960e+01, 6.84520189016155e+01, 6.79148891720242e+01, 6.73923836243063e+01, 6.68874210805177e+01, 6.64025953662896e+01, 6.59401520424910e+01, 6.55019733513170e+01, 6.50895708068923e+01, 6.47040846628865e+01, 6.43462894105492e+01, 6.40166044848498e+01, 6.37151094480775e+01, 6.34415630391323e+01, 6.31954255905193e+01, 6.29758844040644e+01, 6.27818817324450e+01, 6.26121450363803e+01, 6.24652191803669e+01, 6.23395001983065e+01, 6.22332702098520e+01, 6.21447330046367e+01, 6.20720497411162e+01, 6.20133741365538e+01, 6.19668864623187e+01, 6.19308256120121e+01, 6.19035184864846e+01, 6.18834059460336e+01, 6.18690646206672e+01, 6.18592239466038e+01, 6.18527779106711e+01, 6.18487911305746e+01, 6.18464990719803e+01, 6.18453023946056e+01, 6.18447556191240e+01, 6.18445505041647e+01, 6.18444947078658e+01, 6.18444864724469e+01, 6.18444858118645e+01, 6.18444736207675e+01, 6.18443910242886e+01, 6.18440873908492e+01, 6.18432780152276e+01, 6.18415067146932e+01, 6.18381144554297e+01, 6.18322152666959e+01, 6.18226805518015e+01, 6.18081327325325e+01, 6.17869489624732e+01, 6.17572754082341e+01, 6.17170523190684e+01, 6.16640497795572e+01, 6.15959136656519e+01, 6.15102209061136e+01, 6.14045427019042e+01, 6.12765138968879e+01, 6.11239062546227e+01, 6.09447030155059e+01, 6.07371718272971e+01, 6.04999329999839e+01, 6.02320200654495e+01, 5.99329298415834e+01, 5.96026596077637e+01, 5.92417295698917e+01, 5.88511894835894e+01, 5.84326090553934e+01, 5.79880524933050e+01, 5.75200382793367e+01, 5.70314858550464e+01, 5.65256514311920e+01, 5.60060555477558e+01, 5.54764053100032e+01, 5.49405143874305e+01, 5.44022238539275e+01, 5.38653267424405e+01, 5.33334987818023e+01, 5.28102372085106e+01, 5.22988088687509e+01, 5.18022081342747e+01, 5.13231245381543e+01, 5.08639195598626e+01, 5.04266116858073e+01, 5.00128687372691e+01, 4.96240064604447e+01, 4.92609924666068e+01, 4.89244547475094e+01, 4.86146941348570e+01, 4.83317001993820e+01, 4.80751701842116e+01, 4.78445306372942e+01, 4.76389614518460e+01, 4.74574220462505e+01, 4.72986794187785e+01, 4.71613377996321e+01, 4.70438695940785e+01, 4.69446472669229e+01, 4.68619757624892e+01, 4.67941249895349e+01, 4.67393618328093e+01, 4.66959810894224e+01, 4.66623346768351e+01, 4.66368584280974e+01, 4.66180957860909e+01, 4.66047177373527e+01, 4.65955383905439e+01, 4.65895257048717e+01, 4.65858070068320e+01, 4.65836690937317e+01, 4.65825529014297e+01, 4.65820429018944e+01, 4.65818515831293e+01, 4.65817995397551e+01, 4.65817918582609e+01, 4.65817912657902e+01, 4.65817806269174e+01, 4.65817085469902e+01, 4.65814435735735e+01, 4.65807372518846e+01, 4.65791914849308e+01, 4.65762311593046e+01, 4.65710831376215e+01, 4.65627625912104e+01, 4.65500674984372e+01, 4.65315819605343e+01, 4.65056887818550e+01, 4.64705915186662e+01, 4.64243459150171e+01, 4.63649003145130e+01, 4.62901442671352e+01, 4.61979641518713e+01, 4.60863042281836e+01, 4.59532311393193e+01, 4.57969995513481e+01, 4.56161163596253e+01, 4.54094007633867e+01, 4.51760375264519e+01, 4.49156209221513e+01, 4.46281872019559e+01, 4.43142339110027e+01, 4.39747249663805e+01, 4.36110810739595e+01, 4.32251557447474e+01, 4.28191978465900e+01, 4.23958022637479e+01, 4.19578508101254e+01, 4.15084460185481e+01, 4.10508407598644e+01, 4.05883667719140e+01, 4.01243650492633e+01, 3.96621206498402e+01, 3.92048038666123e+01, 3.87554190023001e+01, 3.83167613080530e+01, 3.78913821135707e+01, 3.74815618356082e+01, 3.70892903864369e+01, 3.67162544526625e+01, 3.63638311092374e+01, 3.60330872254006e+01, 3.57247840912651e+01, 3.54393866535355e+01, 3.51770767157133e+01, 3.49377694502085e+01, 3.47211325961919e+01, 3.45266077761651e+01, 3.43534334460133e+01, 3.42006690832536e+01, 3.40672203014573e+01, 3.39518646431731e+01, 3.38532778411449e+01, 3.37700603450113e+01, 3.37007638892921e+01, 3.36439178333649e+01, 3.35980549431412e+01, 3.35617362168608e+01, 3.35335742941258e+01, 3.35122549379969e+01, 3.34965560534425e+01, 3.34853637083819e+01, 3.34776846600213e+01, 3.34726549600982e+01, 3.34695443158008e+01, 3.34677560133458e+01, 3.34668223609512e+01, 3.34663957680662e+01, 3.34662357384454e+01, 3.34661922065176e+01, 3.34661857812966e+01, 3.34661852947983e+01, 3.34661766765291e+01, 3.34661182865014e+01, 3.34659036387966e+01, 3.34653314684629e+01, 3.34640792960023e+01, 3.34616812678295e+01, 3.34575111765456e+01, 3.34507714686074e+01, 3.34404889344641e+01, 3.34255176427138e+01, 3.34045495189230e+01, 3.33761327748738e+01, 3.33386981598990e+01, 3.32905927305883e+01, 3.32301205212207e+01, 3.31555891536656e+01, 3.30653610678143e+01, 3.29579077040222e+01, 3.28318646549926e+01, 3.26860855564389e+01, 3.25196923340012e+01, 3.23321193948515e+01, 3.21231494656673e+01, 3.18929390437680e+01, 3.16420318431154e+01, 3.13713591671997e+01, 3.10822267999292e+01, 3.07762887352686e+01, 3.04555088174682e+01, 3.01221120781710e+01, 2.97785281706006e+01, 2.94273297509561e+01, 2.90711688890287e+01, 2.87127145700809e+01, 2.83545940742428e+01, 2.79993405191959e+01, 2.76493481907656e+01, 2.73068365511095e+01, 2.69738230993265e+01, 2.66521046476795e+01, 2.63432461262109e+01, 2.60485757643558e+01, 2.57691854126252e+01, 2.55059348276357e+01, 2.52594589027867e+01, 2.50301770355571e+01, 2.48183040386434e+01, 2.46238621962300e+01, 2.44466942223394e+01, 2.42864769906229e+01, 2.41427359770664e+01, 2.40148603956999e+01, 2.39021190200260e+01, 2.38036766759684e+01, 2.37186113703186e+01, 2.36459319850347e+01, 2.35845964245613e+01, 2.35335300527357e+01, 2.34916442004756e+01, 2.34578544688208e+01, 2.34310984983702e+01, 2.34103528307064e+01, 2.33946484551621e+01, 2.33830846200192e+01, 2.33748404947267e+01, 2.33691843012672e+01, 2.33654795888349e+01, 2.33631884048508e+01, 2.33618712134140e+01, 2.33611835242097e+01, 2.33608693141157e+01, 2.33607514431885e+01, 2.33607193794503e+01, 2.33607146469102e+01, 2.33607142916369e+01, 2.33607080384246e+01, 2.33606656720031e+01, 2.33605099287287e+01, 2.33600947756682e+01, 2.33591862300863e+01, 2.33574462824169e+01, 2.33544205795646e+01, 2.33495304560298e+01, 2.33420698190106e+01, 2.33312072964586e+01, 2.33159939388831e+01, 2.32953766210411e+01, 2.32682171140877e+01, 2.32333165904307e+01, 2.31894450847183e+01, 2.31353751728023e+01, 2.30699188601422e+01, 2.29919664120318e+01, 2.29005256351726e+01, 2.27947599607179e+01, 2.26740236086856e+01, 2.25378921523085e+01, 2.23861869577240e+01, 2.22189922447901e+01, 2.20366638791797e+01, 2.18398294315751e+01, 2.16293794863002e+01, 2.14064506085467e+01, 2.11724007543242e+01, 2.09287782128658e+01, 2.06772854059387e+01, 2.04197390417504e+01, 2.01580282427475e+01, 1.98940723353347e+01, 1.96297799868958e+01, 1.93670112705367e+01, 1.91075440035548e+01, 1.88530453394830e+01, 1.86050491319996e+01, 1.83649391036189e+01, 1.81339374280326e+01, 1.79130980412646e+01, 1.77033038635949e+01, 1.75052671277499e+01, 1.73195321238543e+01, 1.71464798323642e+01, 1.69863340753317e+01, 1.68391689449775e+01, 1.67049173566925e+01, 1.65833806246978e+01, 1.64742389824187e+01, 1.63770629764132e+01, 1.62913256600414e+01, 1.62164155050375e+01, 1.61516499369304e+01, 1.60962893834337e+01, 1.60495517027145e+01, 1.60106268306759e+01, 1.59786914539906e+01, 1.59529234807929e+01, 1.59325160469325e+01, 1.59166907664749e+01, 1.59047099148569e+01, 1.58958872256239e+01, 1.58895969900069e+01, 1.58852811745729e+01, 1.58824543161399e+01, 1.58807060138398e+01, 1.58797009129039e+01, 1.58791761594251e+01, 1.58789363951786e+01, 1.58788464513244e+01, 1.58788219844145e+01, 1.58788183731498e+01, 1.58788181028755e+01, 1.58788133567215e+01, 1.58787812008343e+01, 1.58786629924559e+01, 1.58783478928141e+01, 1.58776583074664e+01, 1.58763376786507e+01, 1.58740411241639e+01, 1.58703293623692e+01, 1.58646662988501e+01, 1.58564205745689e+01, 1.58448712849564e+01, 1.58292179707457e+01, 1.58085948519497e+01, 1.57820891252547e+01, 1.57487629741208e+01, 1.57076787554480e+01, 1.56579266361911e+01, 1.55986537710791e+01, 1.55290939548488e+01, 1.54485965666993e+01, 1.53566535674391e+01, 1.52529233236250e+01, 1.51372501240116e+01, 1.50096784197721e+01, 1.48704610506144e+01, 1.47200609964239e+01, 1.45591464966869e+01, 1.43885796857032e+01, 1.42093991816900e+01, 1.40227973285110e+01, 1.38300930108836e+01, 1.36327011411710e+01, 1.34321000420803e+01, 1.32297980173322e+01, 1.30273004035622e+01, 1.28260783253348e+01, 1.26275402307885e+01, 1.24330070761120e+01, 1.22436917699159e+01, 1.20606832078899e+01, 1.18849349513533e+01, 1.17172583561782e+01, 1.15583197610161e+01, 1.14086412073022e+01, 1.12686040907305e+01, 1.11384551295793e+01, 1.10183140687575e+01, 1.09081826062360e+01, 1.08079541166412e+01, 1.07174238425579e+01, 1.06362993171337e+01, 1.05642108642220e+01, 1.05007220894802e+01, 1.04453403247763e+01, 1.03975270180444e+01, 1.03567080719643e+01, 1.03222841292762e+01, 1.02936407828866e+01, 1.02701586586231e+01, 1.02512232815696e+01, 1.02362345977609e+01, 1.02246159861458e+01, 1.02158225655033e+01, 1.02093485812572e+01, 1.02047336509137e+01, 1.02015676560170e+01, 1.01994940936459e+01, 1.01982117406633e+01, 1.01974745369769e+01, 1.01970896565806e+01, 1.01969138030022e+01, 1.01968478344890e+01, 1.01968298894760e+01, 1.01968272408305e+01, 1.01968270425263e+01, 1.01968235592138e+01, 1.01967999592720e+01, 1.01967132035814e+01, 1.01964819462391e+01, 1.01959758526934e+01, 1.01950066517172e+01, 1.01933212929360e+01, 1.01905975463631e+01, 1.01864423387795e+01, 1.01803930711037e+01, 1.01719220975949e+01, 1.01604444680867e+01, 1.01453289353751e+01, 1.01259121098441e+01, 1.01015155032892e+01, 1.00714650477462e+01, 1.00351125105627e+01, 9.99185806516944e+00, 9.94117313218959e+00, 9.88262249372652e+00, 9.81588462106706e+00, 9.74076915684371e+00, 9.65723056685999e+00, 9.56537712785091e+00, 9.46547464099601e+00, 9.35794454394269e+00, 9.24335641524219e+00, 9.12241519703791e+00, 8.99594377409163e+00, 8.86486181020912e+00, 8.73016193237413e+00, 8.59288445265823e+00, 8.45409182340116e+00, 8.31484393864098e+00, 8.17617524027436e+00, 8.03907438325868e+00, 7.90446698502295e+00, 7.77320175342959e+00, 7.64604007419636e+00, 7.52364895605503e+00, 7.40659708770725e+00, 7.29535365747743e+00, 7.19028952336015e+00, 7.09168029429290e+00, 6.99971088776627e+00, 6.91448115820721e+00, 6.83601223839555e+00, 6.76425329587101e+00, 6.69908847153236e+00, 6.64034383273896e+00, 6.58779423344626e+00, 6.54117002560996e+00, 6.50016360682915e+00, 6.46443581763265e+00, 6.43362221761609e+00, 6.40733927331581e+00, 6.38519048346308e+00, 6.36677245087115e+00, 6.35168088690361e+00, 6.33951650686340e+00, 6.32989074558442e+00, 6.32243119497475e+00, 6.31678664214658e+00, 6.31263157068412e+00, 6.30966998063549e+00, 6.30763838632927e+00, 6.30630786557636e+00, 6.30548505871978e+00, 6.30501204988875e+00, 6.30476510344097e+00, 6.30465227311976e+00, 6.30460994682174e+00, 6.30459843306610e+00, 6.30459673366074e+00, 6.30459660608868e+00, 6.30459436071331e+00, 6.30457914797726e+00, 6.30452322453678e+00, 6.30437415460888e+00, 6.30404792598483e+00, 6.30342318698928e+00, 6.30233684777426e+00, 6.30058126875154e+00, 6.29790323439133e+00, 6.29400487982124e+00, 6.28854669776575e+00, 6.28115270131636e+00, 6.27141775123894e+00, 6.25891697360682e+00, 6.24321709500029e+00, 6.22388941134935e+00, 6.20052398859630e+00, 6.17274457745917e+00, 6.14022362180339e+00, 6.10269666296333e+00, 6.05997540316856e+00, 6.01195870048335e+00, 5.95864083201927e+00, 5.90011648289934e+00, 5.83658209035750e+00, 5.76833338370182e+00, 5.69575919417059e+00, 5.61933184260076e+00, 5.53959462466176e+00, 5.45714708228282e+00, 5.37262885958248e+00, 5.28670298293908e+00, 5.20003937703069e+00, 5.11329933934067e+00, 5.02712155930988e+00, 4.94211010389640e+00, 4.85882461872157e+00, 4.77777283113332e+00, 4.69940530188596e+00, 4.62411226368245e+00, 4.55222231012647e+00, 4.48400265593929e+00, 4.41966067409746e+00, 4.35934642201974e+00, 4.30315589113874e+00, 4.25113474674386e+00, 4.20328236335803e+00, 4.15955600147711e+00, 4.11987501141377e+00, 4.08412498706080e+00, 4.05216182498116e+00, 4.02381567115857e+00, 3.99889475823574e+00, 3.97718914972688e+00, 3.95847441447432e+00, 3.94251525481273e+00, 3.92906910611948e+00, 3.91788971458006e+00, 3.90873068528309e+00, 3.90134897562956e+00, 3.89550829112993e+00, 3.89098232371531e+00, 3.88755775842949e+00, 3.88503696437859e+00, 3.88324028137150e+00, 3.88200781563810e+00, 3.88120066666336e+00, 3.88070152223951e+00, 3.88041457944060e+00, 3.88026477398833e+00, 3.88019632767845e+00, 3.88017065127258e+00, 3.88016366668349e+00, 3.88016263577333e+00, 3.88016255805472e+00, 3.88016118575122e+00, 3.88015188820157e+00, 3.88011770955147e+00, 3.88002660278647e+00, 3.87982722276880e+00, 3.87944540499372e+00, 3.87878147954103e+00, 3.87770855752436e+00, 3.87607191274392e+00, 3.87368956276997e+00, 3.87035412943469e+00, 3.86583602669285e+00, 3.85988798232347e+00, 3.85225084805781e+00, 3.84266059074641e+00, 3.83085628708689e+00, 3.81658887008187e+00, 3.79963030252119e+00, 3.77978278866296e+00, 3.75688758803510e+00, 3.73083297280583e+00, 3.70156087895829e+00, 3.66907184534679e+00, 3.63342791370660e+00, 3.59475327270572e+00, 3.55323256191858e+00, 3.50910689563284e+00, 3.46266780826858e+00, 3.41424944940088e+00, 3.36421945517399e+00, 3.31296898576183e+00, 3.26090244119308e+00, 3.20842735057829e+00, 3.15594487702545e+00, 3.10384130011067e+00, 3.05248073959189e+00, 3.00219927878931e+00, 2.95330054381442e+00, 2.90605270408852e+00, 2.86068678643049e+00, 2.81739614275504e+00, 2.77633688076746e+00, 2.73762905634045e+00, 2.70135843224280e+00, 2.66757862635338e+00, 2.63631349900665e+00, 2.60755965956501e+00, 2.58128900327851e+00, 2.55745121840978e+00, 2.53597622877560e+00, 2.51677655734080e+00, 2.49974961193227e+00, 2.48477990456877e+00, 2.47174122162052e+00, 2.46049876346444e+00, 2.45091127001557e+00, 2.44283314307316e+00, 2.43611656847699e+00, 2.43061363134174e+00, 2.42617840694051e+00, 2.42266899901894e+00, 2.41994948737758e+00, 2.41789173838579e+00, 2.41637702654160e+00, 2.41529741296637e+00, 2.41455682827826e+00, 2.41407181278541e+00, 2.41377187619669e+00, 2.41359945154930e+00, 2.41350943298596e+00, 2.41346830336401e+00, 2.41345287432046e+00, 2.41344867725574e+00, 2.41344805777810e+00, 2.41344801082255e+00, 2.41344717834178e+00, 2.41344153816725e+00, 2.41342080436507e+00, 2.41336553624541e+00, 2.41324458628313e+00, 2.41301296400016e+00, 2.41261020646401e+00, 2.41195933865677e+00, 2.41096649808351e+00, 2.40952128672626e+00, 2.40749789864748e+00, 2.40475705196661e+00, 2.40114872852388e+00, 2.39651569272654e+00, 2.39069772304535e+00, 2.38353644663640e+00, 2.37488062204602e+00, 2.36459167056178e+00, 2.35254921812175e+00, 2.33865638193713e+00, 2.32284452411126e+00, 2.30507720256277e+00, 2.28535307967438e+00, 2.26370760098976e+00, 2.24021332681181e+00, 2.21497888282023e+00, 2.18814658381640e+00, 2.15988886835927e+00, 2.13040375264484e+00, 2.09990956244275e+00, 2.06863922797861e+00, 2.03683442746980e+00, 2.00473984308509e+00, 1.97259775360610e+00, 1.94064313783504e+00, 1.90909940891552e+00, 1.87817484841138e+00, 1.84805976464391e+00, 1.81892436472168e+00, 1.79091730420961e+00, 1.76416486130459e+00, 1.73877067178644e+00, 1.71481595489126e+00, 1.69236015705886e+00, 1.67144193935528e+00, 1.65208043503213e+00, 1.63427670633902e+00, 1.61801533463941e+00, 1.60326608520720e+00, 1.58998559756269e+00, 1.57811906322911e+00, 1.56760186443718e+00, 1.55836115852117e+00, 1.55031740251741e+00, 1.54338581995816e+00, 1.53747781650256e+00, 1.53250235264427e+00, 1.52836728039490e+00, 1.52498064696596e+00, 1.52225196269985e+00, 1.52009342362788e+00, 1.51842107193953e+00, 1.51715587121409e+00, 1.51622466828995e+00, 1.51556101077803e+00, 1.51510578889733e+00, 1.51480767270611e+00, 1.51462332083460e+00, 1.51451734415554e+00, 1.51446201690818e+00, 1.51443673791823e+00, 1.51442725497506e+00, 1.51442467539202e+00, 1.51442429465123e+00, 1.51442426560850e+00, 1.51442374829006e+00, 1.51442024338562e+00, 1.51440735904671e+00, 1.51437301459508e+00, 1.51429785496413e+00, 1.51415392432891e+00, 1.51390365606895e+00, 1.51349923208441e+00, 1.51288235949832e+00, 1.51198450630568e+00, 1.51072762878742e+00, 1.50902541166039e+00, 1.50678502628406e+00, 1.50390939239178e+00, 1.50029990485159e+00, 1.49585955959925e+00, 1.49049638354757e+00, 1.48412704406627e+00, 1.47668048721027e+00, 1.46810143321532e+00, 1.45835354580133e+00, 1.44742209099795e+00, 1.43531591317668e+00, 1.42206858127679e+00, 1.40773859613769e+00, 1.39240859851822e+00, 1.37618357391474e+00, 1.35918811110207e+00, 1.34156283232188e+00, 1.32346016979582e+00, 1.30503971099148e+00, 1.28646336889869e+00, 1.26789064874640e+00, 1.24947427532857e+00, 1.23135641369298e+00, 1.21366566184670e+00, 1.19651492265155e+00, 1.18000018203577e+00, 1.16420014289028e+00, 1.14917659914643e+00, 1.13497539049465e+00, 1.12162775862974e+00, 1.10915192971923e+00, 1.09755477008189e+00, 1.08683339581698e+00, 1.07697665509083e+00, 1.06796643794891e+00, 1.05977879886915e+00, 1.05238489990420e+00, 1.04575179699265e+00, 1.03984309980763e+00, 1.03461953785755e+00, 1.03003946405686e+00, 1.02605932306022e+00, 1.02263410640236e+00, 1.01971781068050e+00, 1.01726390914854e+00, 1.01522584144625e+00, 1.01355752092464e+00, 1.01221385424718e+00, 1.01115126373911e+00, 1.01032819942328e+00, 1.00970562495561e+00, 1.00924745989545e+00, 1.00892096005663e+00, 1.00869701818391e+00, 1.00855036892379e+00, 1.00845968496054e+00, 1.00840755511580e+00, 1.00838033992324e+00, 1.00836790537022e+00, 1.00836324078750e+00, 1.00836197191239e+00, 1.00836178462924e+00, 1.00836177024254e+00, 1.00836151266076e+00, 1.00835976750862e+00, 1.00835335218146e+00, 1.00833625155647e+00, 1.00829882868686e+00, 1.00822716489270e+00, 1.00810255808830e+00, 1.00790120633583e+00, 1.00759410130708e+00, 1.00714715317329e+00, 1.00652156390326e+00, 1.00567445947441e+00, 1.00455978259141e+00, 1.00312943587926e+00, 1.00133465123241e+00, 9.99127544630326e-01, 9.96462798409567e-01, 9.93299396466455e-01, 9.89602324355062e-01, 9.85344138178434e-01, 9.80506305777709e-01, 9.75080232575493e-01, 9.69067902935720e-01, 9.62482094970483e-01, 9.55346159661872e-01, 9.47693389945229e-01, 9.39566037285985e-01, 9.31014057744719e-01, 9.22093683212140e-01, 9.12865915001035e-01, 9.03395027164720e-01, 8.93747148649112e-01, 8.83988970800814e-01, 8.74186604186619e-01, 8.64404589686629e-01, 8.54705055529948e-01, 8.45147004848471e-01, 8.35785716590891e-01, 8.26672244607892e-01, 8.17853003541643e-01, 8.09369434239869e-01, 8.01257744729583e-01, 7.93548724857175e-01, 7.86267633543062e-01, 7.79434157496851e-01, 7.73062439595662e-01, 7.67161174311383e-01, 7.61733766840152e-01, 7.56778552076256e-01, 7.52289069296154e-01, 7.48254388317761e-01, 7.44659482868940e-01, 7.41485646827016e-01, 7.38710948776827e-01, 7.36310719912041e-01, 7.34258069642522e-01, 7.32524422373359e-01, 7.31080067838846e-01, 7.29894716179842e-01, 7.28938047751726e-01, 7.28180246565827e-01, 7.27592505426117e-01, 7.27147490356145e-01, 7.26819751928048e-01, 7.26586071688700e-01, 7.26425733074691e-01, 7.26320708014484e-01, 7.26255752782052e-01, 7.26218409494479e-01, 7.26198912793933e-01, 7.26190004558646e-01, 7.26186662766921e-01, 7.26185753718042e-01, 7.26185619544339e-01, 7.26185609158762e-01, 7.26185422189161e-01, 7.26184155442908e-01, 7.26179498750554e-01, 7.26167085725834e-01, 7.26139920162311e-01, 7.26087895001513e-01, 7.25997423257638e-01, 7.25851198063154e-01, 7.25628096412379e-01, 7.25303238133790e-01, 7.24848208003863e-01, 7.24231444589648e-01, 7.23418794283944e-01, 7.22374223034748e-01, 7.21060671550688e-01, 7.19441032500569e-01, 7.17479220799092e-01, 7.15141301015947e-01, 7.12396629920036e-01, 7.09218967879011e-01, 7.05587510941783e-01, 7.01487796439622e-01, 6.96912439098810e-01, 6.91861661881987e-01, 6.86343595635634e-01, 6.80374333381067e-01, 6.73977737791787e-01, 6.67185013004612e-01, 6.60034063408901e-01, 6.52568671584319e-01, 6.44837534489056e-01, 6.36893200980040e-01, 6.28790954710624e-01, 6.20587684612744e-01, 6.12340780972659e-01, 6.04107089178342e-01, 5.95941946274721e-01, 5.87898318233300e-01, 5.80026048975668e-01, 5.72371226176614e-01, 5.64975664015023e-01, 5.57876499446209e-01, 5.51105896179408e-01, 5.44690849184406e-01, 5.38653081988102e-01, 5.33009029014806e-01, 5.27769895555181e-01, 5.22941788450254e-01, 5.18525911120676e-01, 5.14518817072897e-01, 5.10912716415589e-01, 5.07695830177511e-01, 5.04852787303818e-01, 5.02365059093975e-01, 5.00211425523077e-01, 4.98368467353671e-01, 4.96811077224805e-01, 4.95512982036331e-01, 4.94447268000002e-01, 4.93586898791577e-01, 4.92905216414073e-01, 4.92376413783845e-01, 4.91975967782807e-01, 4.91681021679881e-01, 4.91470706475170e-01, 4.91326391897516e-01, 4.91231859481744e-01, 4.91173392317123e-01, 4.91139778603028e-01, 4.91122228953770e-01, 4.91114210314450e-01, 4.91111202236354e-01, 4.91110383965264e-01, 4.91110263190137e-01, 4.91110253765868e-01, 4.91110083123544e-01, 4.91108926996418e-01, 4.91104676948552e-01, 4.91093347863966e-01, 4.91068554351102e-01, 4.91021071469220e-01, 4.90938497279188e-01, 4.90805032587716e-01, 4.90601390780299e-01, 4.90304847768684e-01, 4.89889438736725e-01, 4.89326304473607e-01, 4.88584185573970e-01, 4.87630057656190e-01, 4.86429895068753e-01, 4.84949544497454e-01, 4.83155683739497e-01, 4.81016835073603e-01, 4.78504397619034e-01, 4.75593659367732e-01, 4.72264747703164e-01, 4.68503477593757e-01, 4.64302059497967e-01, 4.59659634339161e-01, 4.54582610427920e-01, 4.49084786403134e-01, 4.43187254414710e-01, 4.36918088060281e-01, 4.30311829217672e-01, 4.23408796211035e-01, 4.16254242246367e-01, 4.08897397531376e-01, 4.01390430963559e-01, 3.93787367907677e-01, 3.86142999656684e-01, 3.78511817972523e-01, 3.70947004878502e-01, 3.63499503832344e-01, 3.56217193705135e-01, 3.49144181790962e-01, 3.42320226559593e-01, 3.35780295282660e-01, 3.29554256290280e-01, 3.23666700751858e-01, 3.18136884800353e-01, 3.12978779749741e-01, 3.08201216210236e-01, 3.03808107098575e-01, 2.99798734768757e-01, 2.96168088564144e-01, 2.92907240762801e-01, 2.90003750872131e-01, 2.87442090254643e-01, 2.85204080888860e-01, 2.83269343503520e-01, 2.81615751245965e-01, 2.80219885411057e-01, 2.79057489582399e-01, 2.78103917910670e-01, 2.77334572303644e-01, 2.76725322198132e-01, 2.76252899508506e-01, 2.75895260479771e-01, 2.75631905675027e-01, 2.75444149321174e-01, 2.75315329798218e-01, 2.75230954206194e-01, 2.75178771643269e-01, 2.75148771992392e-01, 2.75133109512041e-01, 2.75125953204764e-01, 2.75123268628363e-01, 2.75122538358630e-01, 2.75122430572353e-01, 2.75122422090050e-01, 2.75122267586795e-01, 2.75121220804976e-01, 2.75117372729625e-01, 2.75107115215391e-01, 2.75084667108786e-01, 2.75041677113052e-01, 2.74966919698144e-01, 2.74846098689679e-01, 2.74661771218007e-01, 2.74393402378152e-01, 2.74017558117578e-01, 2.73508240408702e-01, 2.72837364572680e-01, 2.71975373661363e-01, 2.70891979119895e-01, 2.69557010701732e-01, 2.67941352065591e-01, 2.66017932048285e-01, 2.63762735760849e-01, 2.61155794948777e-01, 2.58182114034028e-01, 2.54832487407398e-01, 2.51104165239994e-01, 2.47001329519179e-01, 2.42535349178851e-01, 2.37724792827933e-01, 2.32595189207188e-01, 2.27178538432497e-01, 2.21512590466091e-01, 2.15639920136766e-01, 2.09606839426925e-01, 2.03462196721233e-01, 1.97256118496796e-01, 1.91038750984234e-01, 1.84859057436611e-01, 1.78763720973137e-01, 1.72796194052708e-01, 1.66995924352761e-01, 1.61397774259683e-01, 1.56031638498519e-01, 1.50922252736791e-01, 1.46089176204612e-01, 1.41546924086413e-01, 1.37305220941030e-01, 1.33369344656003e-01, 1.29740531131986e-01, 1.26416412541062e-01, 1.23391466032388e-01, 1.20657454579359e-01, 1.18203846743270e-01, 1.16018207016680e-01, 1.14086552774012e-01, 1.12393677468603e-01, 1.10923442448808e-01, 1.09659041581764e-01, 1.08583243792324e-01, 1.07678618727377e-01, 1.06927750153133e-01, 1.06313440530982e-01, 1.05818908661121e-01, 1.05427980512630e-01, 1.05125271558393e-01, 1.04896357283323e-01, 1.04727927198525e-01, 1.04607916802962e-01, 1.04525611581792e-01, 1.04471717355310e-01, 1.04438392080539e-01, 1.04419235493452e-01, 1.04409234648472e-01, 1.04404665321340e-01, 1.04402951231890e-01, 1.04402484960161e-01, 1.04402416139493e-01, 1.04402410675675e-01, 1.04402310544131e-01, 1.04401632139457e-01, 1.04399138272015e-01, 1.04392490689889e-01, 1.04377943394724e-01, 1.04350086556551e-01, 1.04301652583331e-01, 1.04223395383755e-01, 1.04104053196664e-01, 1.03930404365727e-01, 1.03687423890449e-01, 1.03358546241523e-01, 1.02926036587604e-01, 1.02371468104607e-01, 1.01676297430970e-01, 1.00822523756804e-01, 9.97934098460718e-02, 9.85742360492736e-02, 9.71530517839341e-02, 9.55213838736597e-02, 9.36748583863789e-02, 9.16136929451596e-02, 8.93430204212819e-02, 8.68730126409218e-02, 8.42187839949418e-02, 8.14000689173469e-02, 7.84406829403252e-02, 7.53677929431262e-02, 7.22110366374551e-02, 6.90015426663422e-02, 6.57709096346792e-02, 6.25502041329042e-02, 5.93690342016941e-02, 5.62547462311045e-02, 5.32317811007630e-02, 5.03212109723989e-02, 4.75404632328917e-02, 4.49032242554921e-02, 4.24195041964310e-02, 4.00958358085575e-02, 3.79355755613346e-02, 3.59392740592739e-02, 3.41050843352519e-02, 3.24291803248968e-02, 3.09061628991003e-02, 2.95294364822780e-02, 2.82915448890009e-02, 2.71844601052607e-02, 2.61998220435800e-02, 2.53291306776049e-02, 2.45638944134612e-02, 2.38957401554813e-02, 2.33164914020847e-02, 2.28182210001791e-02, 2.23932850272490e-02, 2.20343437738677e-02, 2.17343750588270e-02, 2.14866841996297e-02, 2.12849139423952e-02, 2.11230565814304e-02, 2.09954694171728e-02, 2.08968936662046e-02, 2.08224759992532e-02, 2.07677910962900e-02, 2.07288630242743e-02, 2.07021828950644e-02, 2.06847201784797e-02, 2.06739252207342e-02, 2.06677209363022e-02, 2.06644822466728e-02, 2.06630025781323e-02, 2.06624475220707e-02, 2.06622965351357e-02, 2.06622742498439e-02, 2.06622724643779e-02, 2.06622395393339e-02, 2.06620164684495e-02, 2.06611964514565e-02, 2.06590107043922e-02, 2.06542278448296e-02, 2.06450703742433e-02, 2.06291525818187e-02, 2.06034443220915e-02, 2.05642652070850e-02, 2.05073136777401e-02, 2.04277350115376e-02, 2.03202314108921e-02, 2.01792157398160e-02, 1.99990081305828e-02, 1.97740715623667e-02, 1.94992787419750e-02, 1.91701984772266e-02, 1.87833856667226e-02, 1.83366556085645e-02, 1.78293211903053e-02, 1.72623712672788e-02, 1.66385706110361e-02, 1.59624663868811e-02, 1.52402930023499e-02, 1.44797757563708e-02, 1.36898430498889e-02, 1.28802658015701e-02, 1.20612499092374e-02, 1.12430120694142e-02, 1.04353703579392e-02, 9.64737855852604e-03, 8.88702771440819e-03, 8.16103064616414e-03, 7.47469637320819e-03, 6.83189272403695e-03, 6.23508800535433e-03, 5.68545720391000e-03, 5.18303519298513e-03, 4.72689874955279e-03, 4.31536046038161e-03, 3.94616020358285e-03, 3.61664321499688e-03, 3.32391725942216e-03, 3.06498473830175e-03, 2.83684841053855e-03, 2.63659161508656e-03, 2.46143539000343e-03, 2.30877572069232e-03, 2.17620444688293e-03, 2.06151728872328e-03, 1.96271217363631e-03, 1.87798066725975e-03, 1.80569489000959e-03, 1.74439187547648e-03, 1.69275691398241e-03, 1.64960705247097e-03, 1.61387559951720e-03, 1.58459821754775e-03, 1.56090096608352e-03, 1.54199047493364e-03, 1.52714626966599e-03, 1.51571513094950e-03, 1.50710724852843e-03, 1.50079382844126e-03, 1.49630573524462e-03, 1.49323270484558e-03, 1.49122265064250e-03, 1.48998061051486e-03, 1.48926694079421e-03, 1.48889445110001e-03, 1.48872428263758e-03, 1.48866045063301e-03, 1.48864308717123e-03, 1.48864052438569e-03, 1.48864031712037e-03, 1.48863647098096e-03, 1.48861041309090e-03, 1.48851462481048e-03, 1.48825931395891e-03, 1.48770070174464e-03, 1.48663138796223e-03, 1.48477339189412e-03, 1.48177453274911e-03, 1.47720887626690e-03, 1.47058200320771e-03, 1.46134181464674e-03, 1.44889543688009e-03, 1.43263248861558e-03, 1.41195450083842e-03, 1.38630963921571e-03, 1.35523110556335e-03, 1.31837677295721e-03, 1.27556686047114e-03, 1.22681593437335e-03, 1.17235539391913e-03, 1.11264298881508e-03, 1.04835687641014e-03, 9.80373203830709e-04, 9.09728012430833e-04, 8.37566130655887e-04, 7.65081335543982e-04, 6.93453161275670e-04, 6.23786178522323e-04, 5.57057317749354e-04, 4.94075874954092e-04, 4.35459213065315e-04, 3.81624897940469e-04, 3.32797374312577e-04, 2.89025141483990e-04, 2.50204076849059e-04, 2.16104651968863e-04, 1.86403171568372e-04, 1.60716087719948e-04, 1.38632223581681e-04, 1.19736810844984e-04, 1.03626637354637e-04, 8.99202593842574e-05, 7.82656304636475e-05, 6.83448880920423e-05, 5.98764494696979e-05, 5.26147492081463e-05, 4.63481395627241e-05, 4.08971149141735e-05, 3.61137344384588e-05, 3.18795870697386e-05, 2.81009257179221e-05, 2.47037241081216e-05, 2.16308784075636e-05, 1.88402570273638e-05, 1.63018862480158e-05, 1.39946479151831e-05, 1.19035644416954e-05, 1.00178187381061e-05, 8.32909569226103e-06, 6.83011441537928e-06, 5.51352428601604e-06, 4.37125838459451e-06, 3.39424029505153e-06, 2.57227971189877e-06, 1.89405820094635e-06, 1.34718010258439e-06, 9.18291263474202e-07, 5.93261333392547e-07, 1.42108547152020e-14, 2.84217094304040e-14, 1.42108547152020e-14, 0.00000000000000e+00, 1.42108547152020e-14, 0.00000000000000e+00],
[ 9.80000000000000e+01, 9.79999998785993e+01, 9.79999961246480e+01, 9.79999706910312e+01, 9.79998771936972e+01, 9.79996279590007e+01, 9.79990824889245e+01, 9.79980377487038e+01, 9.79962206245457e+01, 9.79932828310083e+01, 9.79887984708882e+01, 9.79822643686844e+01, 9.79731032160358e+01, 9.79606694885610e+01, 9.79442580227227e+01, 9.79231150824199e+01, 9.78964517000519e+01, 9.78634590454397e+01, 9.78233255548209e+01, 9.77752555346205e+01, 9.77184889319827e+01, 9.76523219266258e+01, 9.75761279388646e+01, 9.74893785639256e+01, 9.73916638376492e+01, 9.72827111267117e+01, 9.71624018389146e+01, 9.70307850918318e+01, 9.68880874865718e+01, 9.67347182261643e+01, 9.65712690012107e+01, 9.63985083293499e+01, 9.62173703545919e+01, 9.60289384509193e+01, 9.58344242902185e+01, 9.56351432893802e+01, 9.54324875175676e+01, 9.52278972095935e+01, 9.50228319985761e+01, 9.48187428678217e+01, 9.46170456542156e+01, 9.44190967422298e+01, 9.42261713952781e+01, 9.40394449995052e+01, 9.38599773558986e+01, 9.36887000536064e+01, 9.35264068878906e+01, 9.33737472435950e+01, 9.32312223409825e+01, 9.30991842271721e+01, 9.29778373864696e+01, 9.28672428319189e+01, 9.27673245256852e+01, 9.26778779563740e+01, 9.25985806772819e+01, 9.25290045818518e+01, 9.24686296627128e+01, 9.24168589702800e+01, 9.23730344577454e+01, 9.23364533732060e+01, 9.23063848384546e+01, 9.22820862393715e+01, 9.22628190465511e+01, 9.22478636882674e+01, 9.22365331122718e+01, 9.22281846989769e+01, 9.22222302265061e+01, 9.22181436374533e+01, 9.22154664169045e+01, 9.22138104595869e+01, 9.22128583785920e+01, 9.22123612862093e+01, 9.22121341559583e+01, 9.22120489508098e+01, 9.22120257728727e+01, 9.22120223518551e+01, 9.22120219030898e+01, 9.22120114473960e+01, 9.22119406085194e+01, 9.22116801971303e+01, 9.22109860345521e+01, 9.22094668694971e+01, 9.22065574589595e+01, 9.22014978828432e+01, 9.21933200295288e+01, 9.21808420361046e+01, 9.21626712892870e+01, 9.21372163865478e+01, 9.21027082167194e+01, 9.20572300415099e+01, 9.19987561428449e+01, 9.19251982491731e+01, 9.18344585765054e+01, 9.17244879345136e+01, 9.15933469809673e+01, 9.14392683946894e+01, 9.12607175207279e+01, 9.10564489659257e+01, 9.08255567247757e+01, 9.05675157088344e+01, 9.02822130166794e+01, 8.99699678505119e+01, 8.96315395585718e+01, 8.92681237491906e+01, 8.88813367094803e+01, 8.84731884831949e+01, 8.80460450439225e+01, 8.76025802469771e+01, 8.71457188365196e+01, 8.66785727309648e+01, 8.62043738440221e+01, 8.57264073426972e+01, 8.52479490610593e+01, 8.47722096959016e+01, 8.43022867958343e+01, 8.38411240720896e+01, 8.33914767347945e+01, 8.29558815221853e+01, 8.25366305952832e+01, 8.21357491239461e+01, 8.17549768836714e+01, 8.13957543911822e+01, 8.10592140606137e+01, 8.07461766642095e+01, 8.04571531375799e+01, 8.01923515521677e+01, 7.99516889196208e+01, 7.97348073973803e+01, 7.95410944184926e+01, 7.93697062524208e+01, 7.92195944997912e+01, 7.90895350194727e+01, 7.89781587734970e+01, 7.88839840513486e+01, 7.88054495012623e+01, 7.87409473563312e+01, 7.86888562032255e+01, 7.86475726078043e+01, 7.86155408916730e+01, 7.85912803530608e+01, 7.85734092493429e+01, 7.85606649105631e+01, 7.85519194347734e+01, 7.85461905260154e+01, 7.85426471711239e+01, 7.85406100069853e+01, 7.85395463984935e+01, 7.85390604212301e+01, 7.85388781134490e+01, 7.85388285212243e+01, 7.85388212015126e+01, 7.85388205553214e+01, 7.85388078935228e+01, 7.85387221079621e+01, 7.85384067509882e+01, 7.85375661237085e+01, 7.85357264245132e+01, 7.85322031538168e+01, 7.85260760731043e+01, 7.85161728577718e+01, 7.85010623993979e+01, 7.84790585001102e+01, 7.84482344540750e+01, 7.84064487230618e+01, 7.83513815803443e+01, 7.82805822193715e+01, 7.81915254057591e+01, 7.80816763054048e+01, 7.79485616680136e+01, 7.77898451115494e+01, 7.76034038724330e+01, 7.73874040948692e+01, 7.71403715659886e+01, 7.68612547924809e+01, 7.65494774822568e+01, 7.62049778539406e+01, 7.58282327468206e+01, 7.54202652264604e+01, 7.49826352374124e+01, 7.45174137804475e+01, 7.40271419996082e+01, 7.35147773516622e+01, 7.29836295992610e+01, 7.24372896517220e+01, 7.18795542601112e+01, 7.13143493060594e+01, 7.07456540080768e+01, 7.01774279257504e+01, 6.96135422688029e+01, 6.90577167557776e+01, 6.85134630919871e+01, 6.79840359801236e+01, 6.74723923683297e+01, 6.69811593448321e+01, 6.65126107248574e+01, 6.60686520060613e+01, 6.56508130649393e+01, 6.52602477785647e+01, 6.48977396950629e+01, 6.45637129193282e+01, 6.42582474860849e+01, 6.39810986182688e+01, 6.37317193834486e+01, 6.35092863471751e+01, 6.33127278738838e+01, 6.31407547444979e+01, 6.29918927495838e+01, 6.28645168830850e+01, 6.27568867096932e+01, 6.26671824143634e+01, 6.25935409714409e+01, 6.25340918002014e+01, 6.24869912108830e+01, 6.24504548983863e+01, 6.24227877173420e+01, 6.24024099787474e+01, 6.23878795496286e+01, 6.23779091155668e+01, 6.23713780809589e+01, 6.23673387301424e+01, 6.23650164477744e+01, 6.23638039906251e+01, 6.23632500051793e+01, 6.23630421855045e+01, 6.23629856534566e+01, 6.23629773094427e+01, 6.23629766401496e+01, 6.23629642882967e+01, 6.23628806026724e+01, 6.23625729654179e+01, 6.23617529170848e+01, 6.23599582595293e+01, 6.23565212685738e+01, 6.23505442904178e+01, 6.23408838454323e+01, 6.23261441882226e+01, 6.23046810689840e+01, 6.22746162016134e+01, 6.22338626618545e+01, 6.21801611087148e+01, 6.21111263431843e+01, 6.20243032947317e+01, 6.19172310709750e+01, 6.17875132411660e+01, 6.16328920804110e+01, 6.14513241167780e+01, 6.12410540391631e+01, 6.10006838801269e+01, 6.07292344171091e+01, 6.04261959552910e+01, 6.00915660638335e+01, 5.97258724099551e+01, 5.93301795280135e+01, 5.89060791164976e+01, 5.84556642163080e+01, 5.79814883403704e+01, 5.74865112652658e+01, 5.69740337425666e+01, 5.64476238288357e+01, 5.59110378510316e+01, 5.53681391895893e+01, 5.48228180403020e+01, 5.42789150835187e+01, 5.37401515475714e+01, 5.32100675429901e+01, 5.26919698400434e+01, 5.21888895581136e+01, 5.17035496205733e+01, 5.12383413668058e+01, 5.07953094302498e+01, 5.03761438773777e+01, 4.99821786202016e+01, 4.96143952153263e+01, 4.92734312990488e+01, 4.89595930452880e+01, 4.86728711505714e+01, 4.84129599400819e+01, 4.81792792514006e+01, 4.79709987921740e+01, 4.77870646886451e+01, 4.76262279463517e+01, 4.74870745331621e+01, 4.73680567682342e+01, 4.72675256589143e+01, 4.71837637728717e+01, 4.71150181686916e+01, 4.70595328404688e+01, 4.70155800680136e+01, 4.69814900123197e+01, 4.69556778642095e+01, 4.69366678498770e+01, 4.69231134259583e+01, 4.69138130618184e+01, 4.69077211080899e+01, 4.69039533851533e+01, 4.69017872873514e+01, 4.69006563800216e+01, 4.69001396570517e+01, 4.68999458160915e+01, 4.68998930866187e+01, 4.68998853038579e+01, 4.68998847035765e+01, 4.68998739244492e+01, 4.68998008942778e+01, 4.68995324276637e+01, 4.68988167944039e+01, 4.68972506493587e+01, 4.68942512974558e+01, 4.68890354093978e+01, 4.68806051746316e+01, 4.68677427280142e+01, 4.68490135108834e+01, 4.68227790196923e+01, 4.67872191494936e+01, 4.67403640504134e+01, 4.66801350814566e+01, 4.66043940717279e+01, 4.65109996958510e+01, 4.63978693574738e+01, 4.62630445795561e+01, 4.61047575559941e+01, 4.59214962622562e+01, 4.57120653874847e+01, 4.54756403642028e+01, 4.52118119493018e+01, 4.49206191504501e+01, 4.46025687778833e+01, 4.42586405012823e+01, 4.38902769657034e+01, 4.34993592290027e+01, 4.30881684897757e+01, 4.26593357473689e+01, 4.22157816397235e+01, 4.17606491955544e+01, 4.12972325557318e+01, 4.08289048025189e+01, 4.03590478443564e+01, 3.98909868491030e+01, 3.94279310760072e+01, 3.89729222514220e+01, 3.85287909976246e+01, 3.80981213500028e+01, 3.76832231102150e+01, 3.72861116456189e+01, 3.69084946916368e+01, 3.65517656817098e+01, 3.62170030838502e+01, 3.59049751590285e+01, 3.56161494909073e+01, 3.53507065911251e+01, 3.51085568761882e+01, 3.48893603465120e+01, 3.46925483694509e+01, 3.45173470626546e+01, 3.43628018749683e+01, 3.42278030533794e+01, 3.41111117537484e+01, 3.40113865926692e+01, 3.39272104452312e+01, 3.38571172707162e+01, 3.37996187010222e+01, 3.37532300631876e+01, 3.37164954377282e+01, 3.36880112891143e+01, 3.36664481537527e+01, 3.36505698431549e+01, 3.36392496223204e+01, 3.36314828597965e+01, 3.36263957173110e+01, 3.36232495510573e+01, 3.36214408283880e+01, 3.36204965152025e+01, 3.36200650514056e+01, 3.36199031945525e+01, 3.36198591655746e+01, 3.36198526669901e+01, 3.36198521749369e+01, 3.36198434582639e+01, 3.36197844015366e+01, 3.36195673029844e+01, 3.36189885996758e+01, 3.36177221303879e+01, 3.36152967237218e+01, 3.36110790261443e+01, 3.36042623875650e+01, 3.35938625085765e+01, 3.35787204186876e+01, 3.35575131922588e+01, 3.35287726116169e+01, 3.34909117496078e+01, 3.34422591644544e+01, 3.33811000809323e+01, 3.33057235825080e+01, 3.32144744750247e+01, 3.31058081263304e+01, 3.29783462658938e+01, 3.28309314751125e+01, 3.26626779438869e+01, 3.24730160397513e+01, 3.22617283530249e+01, 3.20289751555903e+01, 3.17753076404919e+01, 3.15016678798585e+01, 3.12093751215529e+01, 3.09000987992760e+01, 3.05758194033915e+01, 3.02387790876167e+01, 2.98914245017887e+01, 2.95363447772053e+01, 2.91762077948780e+01, 2.88136978083131e+01, 2.84514571738020e+01, 2.80920344018896e+01, 2.77378400554317e+01, 2.73911112747125e+01, 2.70538850040438e+01, 2.67279794079990e+01, 2.64149825514031e+01, 2.61162471931043e+01, 2.58328904944735e+01, 2.55657975308807e+01, 2.53156276679782e+01, 2.50828230749698e+01, 2.48676188540705e+01, 2.46700544428749e+01, 2.44899860820754e+01, 2.43271002337146e+01, 2.41809278904620e+01, 2.40508597424707e+01, 2.39361621730248e+01, 2.38359940433489e+01, 2.37494242042834e+01, 2.36754496400907e+01, 2.36130141087090e+01, 2.35610270947740e+01, 2.35183828388922e+01, 2.34839791522878e+01, 2.34567356743874e+01, 2.34356111871926e+01, 2.34196195696724e+01, 2.34078439627432e+01, 2.33994487245393e+01, 2.33936887889632e+01, 2.33899160983398e+01, 2.33875828617404e+01, 2.33862414904929e+01, 2.33855411761656e+01, 2.33852211973015e+01, 2.33851011622694e+01, 2.33850685098385e+01, 2.33850636904084e+01, 2.33850633286122e+01, 2.33850569605902e+01, 2.33850138163145e+01, 2.33848552135431e+01, 2.33844324379409e+01, 2.33835072097373e+01, 2.33817353094971e+01, 2.33786540303609e+01, 2.33736740526083e+01, 2.33660762526370e+01, 2.33550138587178e+01, 2.33395202451569e+01, 2.33185225094670e+01, 2.32908607995714e+01, 2.32553131482081e+01, 2.32106253319655e+01, 2.31555450104977e+01, 2.30888591313227e+01, 2.30094333270479e+01, 2.29162518094289e+01, 2.28084561049889e+01, 2.26853809050158e+01, 2.25465853375832e+01, 2.23918781194714e+01, 2.22213353067507e+01, 2.20353097154432e+01, 2.18344314973361e+01, 2.16195997935119e+01, 2.13919658134479e+01, 2.11529080732016e+01, 2.09040008580624e+01, 2.06469772512531e+01, 2.03836882941907e+01, 2.01160600130954e+01, 1.98460501440463e+01, 1.95756063810220e+01, 1.93066278225931e+01, 1.90409309842895e+01, 1.87802212951832e+01, 1.85260704731118e+01, 1.82798996662283e+01, 1.80429678496932e+01, 1.78163647319061e+01, 1.76010073616201e+01, 1.73976396987109e+01, 1.72068345577917e+01, 1.70289974971426e+01, 1.68643723650415e+01, 1.67130483134398e+01, 1.65749681451830e+01, 1.64499378858954e+01, 1.63376374779594e+01, 1.62376324919778e+01, 1.61493867465444e+01, 1.60722757217693e+01, 1.60056006447805e+01, 1.59486031141849e+01, 1.59004801133463e+01, 1.58603992385461e+01, 1.58275139384276e+01, 1.58009785278721e+01, 1.57799627061670e+01, 1.57636652802773e+01, 1.57513267737396e+01, 1.57422405942769e+01, 1.57357624419337e+01, 1.57313176663041e+01, 1.57284063266510e+01, 1.57266057711254e+01, 1.57255706280701e+01, 1.57250301894038e+01, 1.57247832583528e+01, 1.57246906259645e+01, 1.57246654277065e+01, 1.57246617084963e+01, 1.57246614301430e+01, 1.57246565421202e+01, 1.57246234250497e+01, 1.57245016832511e+01, 1.57241771647404e+01, 1.57234669660830e+01, 1.57221068589914e+01, 1.57197416473199e+01, 1.57159189061924e+01, 1.57100864886356e+01, 1.57015941069105e+01, 1.56896992025758e+01, 1.56735772078213e+01, 1.56523361682223e+01, 1.56250355429564e+01, 1.55907088246138e+01, 1.55483894324582e+01, 1.54971391397451e+01, 1.54360781106474e+01, 1.53644154614999e+01, 1.52814791417416e+01, 1.51867438682487e+01, 1.50798558551736e+01, 1.49606531662276e+01, 1.48291806761684e+01, 1.46856988537691e+01, 1.45306858539120e+01, 1.43648327120542e+01, 1.41890317499727e+01, 1.40043586093827e+01, 1.38120486151105e+01, 1.36134684202147e+01, 1.34100840912359e+01, 1.32034269416254e+01, 1.29950585037072e+01, 1.27865360336605e+01, 1.25793798636999e+01, 1.23750437524790e+01, 1.21748891503576e+01, 1.19801640119302e+01, 1.17919864822768e+01, 1.16113334860688e+01, 1.14390339872497e+01, 1.12757664816839e+01, 1.11220601467619e+01, 1.09782990017338e+01, 1.08447284238899e+01, 1.07214634065422e+01, 1.06084980204861e+01, 1.05057156364243e+01, 1.04128995685303e+01, 1.03297438982297e+01, 1.02558643245998e+01, 1.01908089584389e+01, 1.01340690283313e+01, 1.00850894981279e+01, 1.00432796068808e+01, 1.00080233363284e+01, 9.97868979035269e+00, 9.95464343903799e+00, 9.93525414122293e+00, 9.91990681831169e+00, 9.90801061328733e+00, 9.89900733686052e+00, 9.89237898149437e+00, 9.88765407681379e+00, 9.88441266860951e+00, 9.88228972887481e+00, 9.88097684517517e+00, 9.88022209186026e+00, 9.87982804970207e+00, 9.87964801020475e+00, 9.87958047143562e+00, 9.87956209927758e+00, 9.87955938758643e+00, 9.87955918456200e+00, 9.87955561833633e+00, 9.87953145664676e+00, 9.87944263592952e+00, 9.87920587411710e+00, 9.87868773484783e+00, 9.87769546745534e+00, 9.87597000407283e+00, 9.87318145952412e+00, 9.86892743753387e+00, 9.86273439529720e+00, 9.85406225386978e+00, 9.84231236023318e+00, 9.82683880499059e+00, 9.80696297579885e+00, 9.78199108172915e+00, 9.75123422195956e+00, 9.71403040147982e+00, 9.66976772850833e+00, 9.61790787791644e+00, 9.55800878890132e+00, 9.48974550037340e+00, 9.41292802889139e+00, 9.32751527186399e+00, 9.23362407681189e+00, 9.13153285103152e+00, 9.02167938123691e+00, 8.90465286740692e+00, 8.78118052039345e+00, 8.65210939695231e+00, 8.51838441788004e+00, 8.38102370942721e+00, 8.24109250879842e+00, 8.09967687643386e+00, 7.95785836770055e+00, 7.81669065138672e+00, 7.67717884545250e+00, 7.54026209788475e+00, 7.40679969694922e+00, 7.27756077103196e+00, 7.15321744808627e+00, 7.03434119668930e+00, 6.92140196758307e+00, 6.81476969473341e+00, 6.71471769382397e+00, 6.62142750732619e+00, 6.53499478158318e+00, 6.45543581521959e+00, 6.38269448246420e+00, 6.31664930314591e+00, 6.25712049766935e+00, 6.20387692579014e+00, 6.15664285927285e+00, 6.11510457844155e+00, 6.07891681016132e+00, 6.04770903973973e+00, 6.02109173217868e+00, 5.99866249032897e+00, 5.98001216054269e+00, 5.96473087257107e+00, 5.95241397230181e+00, 5.94266777632019e+00, 5.93511504920296e+00, 5.92940008084027e+00, 5.92519322458395e+00, 5.92219474976930e+00, 5.92013786554934e+00, 5.91879078750047e+00, 5.91795774360594e+00, 5.91747885053077e+00, 5.91722883227779e+00, 5.91711459849202e+00, 5.91707174571711e+00, 5.91706008874803e+00, 5.91705836820471e+00, 5.91705823904586e+00, 5.91705596574150e+00, 5.91704056378303e+00, 5.91698394474804e+00, 5.91683302067555e+00, 5.91650273458022e+00, 5.91587022611007e+00, 5.91477037905594e+00, 5.91299297674534e+00, 5.91028166664539e+00, 5.90633490446790e+00, 5.90080900774893e+00, 5.89332339612284e+00, 5.88346802767188e+00, 5.87081295636351e+00, 5.85491983520124e+00, 5.83535507640796e+00, 5.81170425978048e+00, 5.78358726228986e+00, 5.75067347748525e+00, 5.71269641509244e+00, 5.66946693204258e+00, 5.62088435662248e+00, 5.56694483413827e+00, 5.50774634642185e+00, 5.44349003303273e+00, 5.37447765675731e+00, 5.30110529191009e+00, 5.22385354923200e+00, 5.14327486317288e+00, 5.05997853549862e+00, 4.97461433796488e+00, 4.88785551789908e+00, 4.80038202377948e+00, 4.71286468103706e+00, 4.62595091541375e+00, 4.54025246018466e+00, 4.45633531318936e+00, 4.37471204685966e+00, 4.29583643235362e+00, 4.22010022571791e+00, 4.14783188306168e+00, 4.07929692237497e+00, 4.01469962838277e+00, 3.95418579867189e+00, 3.89784624872419e+00, 3.84572082510137e+00, 3.79780271510189e+00, 3.75404288373777e+00, 3.71435451158418e+00, 3.67861734735190e+00, 3.64668192494878e+00, 3.61837362488754e+00, 3.59349658321955e+00, 3.57183746720560e+00, 3.55316914556471e+00, 3.53725428263377e+00, 3.52384888073462e+00, 3.51270578443638e+00, 3.50357814547139e+00, 3.49622282934254e+00, 3.49040372587416e+00, 3.48589490795359e+00, 3.48248356731254e+00, 3.47997264508349e+00, 3.47818306940083e+00, 3.47695551340624e+00, 3.47615159501325e+00, 3.47565445440668e+00, 3.47536866562149e+00, 3.47521946325385e+00, 3.47515129263198e+00, 3.47512571966786e+00, 3.47511876321946e+00, 3.47511773646288e+00, 3.47511765905739e+00, 3.47511629228300e+00, 3.47510703219471e+00, 3.47507299126980e+00, 3.47498225173099e+00, 3.47478367589378e+00, 3.47440340019372e+00, 3.47374216259846e+00, 3.47267360146101e+00, 3.47104364984556e+00, 3.46867113140820e+00, 3.46534964005127e+00, 3.46085075368394e+00, 3.45492859071341e+00, 3.44732566536044e+00, 3.43777993482702e+00, 3.42603285973524e+00, 3.41183822305526e+00, 3.39497137783667e+00, 3.37523852794332e+00, 3.35248559702549e+00, 3.32660621732911e+00, 3.29754837837437e+00, 3.26531932003201e+00, 3.22998833530748e+00, 3.19168726107402e+00, 3.15060857169026e+00, 3.10700113907062e+00, 3.06116386952930e+00, 3.01343755860394e+00, 2.96419540781359e+00, 2.91383271295621e+00, 2.86275625750788e+00, 2.81137392705850e+00, 2.76008500600362e+00, 2.70927153388911e+00, 2.65929099607618e+00, 2.61047051278233e+00, 2.56310258255974e+00, 2.51744233976413e+00, 2.47370620701763e+00, 2.43207176679421e+00, 2.39267864204847e+00, 2.35563016290097e+00, 2.32099560166012e+00, 2.28881277771228e+00, 2.25909086247951e+00, 2.23181324835586e+00, 2.20694038051033e+00, 2.18441248373291e+00, 2.16415214602084e+00, 2.14606674509149e+00, 2.13005072287351e+00, 2.11598772618471e+00, 2.10375263951735e+00, 2.09321353859414e+00, 2.08423359175063e+00, 2.07667293092346e+00, 2.07039050584464e+00, 2.06524592476863e+00, 2.06110127357991e+00, 2.05782289338474e+00, 2.05528308566099e+00, 2.05336170469205e+00, 2.05194759023789e+00, 2.05093978993143e+00, 2.05024852122911e+00, 2.04979582708140e+00, 2.04951588766993e+00, 2.04935496207888e+00, 2.04927094783240e+00, 2.04923256180899e+00, 2.04921816201249e+00, 2.04921424493061e+00, 2.04921366677797e+00, 2.04921362295480e+00, 2.04921284600837e+00, 2.04920758208830e+00, 2.04918823145429e+00, 2.04913665048294e+00, 2.04902377041334e+00, 2.04880760546460e+00, 2.04843173626165e+00, 2.04782434835695e+00, 2.04689789814879e+00, 2.04554946867829e+00, 2.04366186435873e+00, 2.04110547528437e+00, 2.03774091720972e+00, 2.03342242190288e+00, 2.02800191450490e+00, 2.02133367101449e+00, 2.01327940258261e+00, 2.00371356774090e+00, 1.99252867385387e+00, 1.97964030040043e+00, 1.96499156445340e+00, 1.94855675722796e+00, 1.93034391222395e+00, 1.91039612008935e+00, 1.88879147967290e+00, 1.86564166267858e+00, 1.84108916243247e+00, 1.81530338586229e+00, 1.78847582238741e+00, 1.76081457617262e+00, 1.73253857411697e+00, 1.70387175960410e+00, 1.67503755369044e+00, 1.64625381643184e+00, 1.61772847891137e+00, 1.58965594944765e+00, 1.56221433310081e+00, 1.53556344802151e+00, 1.50984357925870e+00, 1.48517488186640e+00, 1.46165732996263e+00, 1.43937110475183e+00, 1.41837731956261e+00, 1.39871899065209e+00, 1.38042217618309e+00, 1.36349722029779e+00, 1.34794005318439e+00, 1.33373351068191e+00, 1.32084864797345e+00, 1.30924603123452e+00, 1.29887699880476e+00, 1.28968488960801e+00, 1.28160624118294e+00, 1.27457196276505e+00, 1.26850849033119e+00, 1.26333893033684e+00, 1.25898419708450e+00, 1.25536414539310e+00, 1.25239869575681e+00, 1.25000894387038e+00, 1.24811824074618e+00, 1.24665322420719e+00, 1.24554477788992e+00, 1.24472889056442e+00, 1.24414738702414e+00, 1.24374850231631e+00, 1.24348727379351e+00, 1.24332573028680e+00, 1.24323286435130e+00, 1.24318438157073e+00, 1.24316222975476e+00, 1.24315391990319e+00, 1.24315165942768e+00, 1.24315132578644e+00, 1.24315130033642e+00, 1.24315084701284e+00, 1.24314777568210e+00, 1.24313648519230e+00, 1.24310638925230e+00, 1.24304052685341e+00, 1.24291439949101e+00, 1.24269508488679e+00, 1.24234067307367e+00, 1.24180006607850e+00, 1.24101317650077e+00, 1.23991155236456e+00, 1.23841944501149e+00, 1.23645532289298e+00, 1.23393381658686e+00, 1.23076805931551e+00, 1.22687236338442e+00, 1.22216514766433e+00, 1.21657200654533e+00, 1.21002878925876e+00, 1.20248454290451e+00, 1.19390416561542e+00, 1.18427062010153e+00, 1.17358657338264e+00, 1.16187535548717e+00, 1.14918116646710e+00, 1.13556850412759e+00, 1.12112083042003e+00, 1.10593853828212e+00, 1.09013631905104e+00, 1.07384006064416e+00, 1.05718342698707e+00, 1.04030427938579e+00, 1.02334110129533e+00, 1.00642958022709e+00, 9.89699485352801e-01, 9.73271957476499e-01, 9.57257300187919e-01, 9.41753328223939e-01, 9.26844293149088e-01, 9.12600370135266e-01, 8.99077656303746e-01, 8.86318604287894e-01, 8.74352797150735e-01, 8.63197963893839e-01, 8.52861138148867e-01, 8.43339874346412e-01, 8.34623452833725e-01, 8.26694024940096e-01, 8.19527668111832e-01, 8.13095337962508e-01, 8.07363717289618e-01, 8.02295971443158e-01, 7.97852425104338e-01, 7.93991178065752e-01, 7.90668677644589e-01, 7.87840263535330e-01, 7.85460697774838e-01, 7.83484688501346e-01, 7.81867411718281e-01, 7.80565030641426e-01, 7.79535207703333e-01, 7.78737600191306e-01, 7.78134327069594e-01, 7.77690392034245e-01, 7.77374046476368e-01, 7.77157075932493e-01, 7.77014994833678e-01, 7.76927136877603e-01, 7.76876631977743e-01, 7.76850265231886e-01, 7.76838218359245e-01, 7.76833699211764e-01, 7.76832469898835e-01, 7.76832288454969e-01, 7.76832274516863e-01, 7.76832024966211e-01, 7.76830334226460e-01, 7.76824118926669e-01, 7.76807551514736e-01, 7.76771295637104e-01, 7.76701867034575e-01, 7.76581148072566e-01, 7.76386082341645e-01, 7.76088573661752e-01, 7.75655610912196e-01, 7.75049635890198e-01, 7.74229165334489e-01, 7.73149669866712e-01, 7.71764701617116e-01, 7.70027248741883e-01, 7.67891279364761e-01, 7.65313420649193e-01, 7.62254702205553e-01, 7.58682278776590e-01, 7.54571037290134e-01, 7.49904990048648e-01, 7.44678360771161e-01, 7.38896284333947e-01, 7.32575064149273e-01, 7.25741961574357e-01, 7.18434526601470e-01, 7.10699514326791e-01, 7.02591462846129e-01, 6.94171031137884e-01, 6.85503207210843e-01, 6.76655496209364e-01, 6.67696186283962e-01, 6.58692769697524e-01, 6.49710571829331e-01, 6.40811615613657e-01, 6.32053726929598e-01, 6.23489869656041e-01, 6.15167688189587e-01, 6.07129229668288e-01, 5.99410816721033e-01, 5.92043042841510e-01, 5.85050865206881e-01, 5.78453773036173e-01, 5.72266012898794e-01, 5.66496855525969e-01, 5.61150891601869e-01, 5.56228346734898e-01, 5.51725408344510e-01, 5.47634559519977e-01, 5.43944916960385e-01, 5.40642571816932e-01, 5.37710933561044e-01, 5.35131077843943e-01, 5.32882099671966e-01, 5.30941473098821e-01, 5.29285418063736e-01, 5.27889274037264e-01, 5.26727878854473e-01, 5.25775949615849e-01, 5.25008460937556e-01, 5.24401014266275e-01, 5.23930190576422e-01, 5.23573877681045e-01, 5.23311562735202e-01, 5.23124580396953e-01, 5.22996307600536e-01, 5.22912297008418e-01, 5.22860342910818e-01, 5.22830475547536e-01, 5.22814882406564e-01, 5.22807757843410e-01, 5.22805085185425e-01, 5.22804358158723e-01, 5.22804250851166e-01, 5.22804242545078e-01, 5.22804093013235e-01, 5.22803079913601e-01, 5.22799355663437e-01, 5.22789428267245e-01, 5.22767702846195e-01, 5.22726097762188e-01, 5.22653751478884e-01, 5.22536834893714e-01, 5.22358482188395e-01, 5.22098851141848e-01, 5.21735321129697e-01, 5.21242833496572e-01, 5.20594374457815e-01, 5.19761595069681e-01, 5.18715556150326e-01, 5.17427578561254e-01, 5.15870171407201e-01, 5.14018003129763e-01, 5.11848873963089e-01, 5.09344643681430e-01, 5.06492066837197e-01, 5.03283489412027e-01, 4.99717366289701e-01, 4.95798568080943e-01, 4.91538457973277e-01, 4.86954733410272e-01, 4.82071042170574e-01, 4.76916396357623e-01, 4.71524419534475e-01, 4.65932470649250e-01, 4.60180692810525e-01, 4.54311035220272e-01, 4.48366292973859e-01, 4.42389202741921e-01, 4.36421623590832e-01, 4.30503822527598e-01, 4.24673874879460e-01, 4.18967181245989e-01, 4.13416096122504e-01, 4.08049658672056e-01, 4.02893413503293e-01, 3.97969308421906e-01, 3.93295656558664e-01, 3.88887151588705e-01, 3.84754926538776e-01, 3.80906648609212e-01, 3.77346644282994e-01, 3.74076050608735e-01, 3.71092989847753e-01, 3.68392765636244e-01, 3.65968079439895e-01, 3.63809266392110e-01, 3.61904549643882e-01, 3.60240312144242e-01, 3.58801384348098e-01, 3.57571345745697e-01, 3.56532837360518e-01, 3.55667881514449e-01, 3.54958204265131e-01, 3.54385555041247e-01, 3.53932017211122e-01, 3.53580302692308e-01, 3.53314023320095e-01, 3.53117931608537e-01, 3.52978123806835e-01, 3.52882198809269e-01, 3.52819367514201e-01, 3.52780508621052e-01, 3.52758168539708e-01, 3.52746504981710e-01, 3.52741175800631e-01, 3.52739176640000e-01, 3.52738632819765e-01, 3.52738552553049e-01, 3.52738546289658e-01, 3.52738432881381e-01, 3.52737664523644e-01, 3.52734839961656e-01, 3.52727310734650e-01, 3.52710833297678e-01, 3.52679277387793e-01, 3.52624402499899e-01, 3.52535713053157e-01, 3.52400401321987e-01, 3.52203385413674e-01, 3.51927447404563e-01, 3.51553474104819e-01, 3.51060799730973e-01, 3.50427645990777e-01, 3.49631650773603e-01, 3.48650471929730e-01, 3.47462447772287e-01, 3.46047291312814e-01, 3.44386791304473e-01, 3.42465490411087e-01, 3.40271309719469e-01, 3.37796089723028e-01, 3.35036021000036e-01, 3.31991943008262e-01, 3.28669496366970e-01, 3.25079122098970e-01, 3.21235909796727e-01, 3.17159304744422e-01, 3.12872690937084e-01, 3.08402872134451e-01, 3.03779476264268e-01, 2.99034309595726e-01, 2.94200686335330e-01, 2.89312757049089e-01, 2.84404856099741e-01, 2.79510884645532e-01, 2.74663742112679e-01, 2.69894815690009e-01, 2.65233534390816e-01, 2.60706991577933e-01, 2.56339637539526e-01, 2.52153041752464e-01, 2.48165722891159e-01, 2.44393043407953e-01, 2.40847164556286e-01, 2.37537056986881e-01, 2.34468561506162e-01, 2.31644494279806e-01, 2.29064790732011e-01, 2.26726682627771e-01, 2.24624903266744e-01, 2.22751916263377e-01, 2.21098163932410e-01, 2.19652331756180e-01, 2.18401625716069e-01, 2.17332059386138e-01, 2.16428747592971e-01, 2.15676203153691e-01, 2.15058632744700e-01, 2.14560227395467e-01, 2.14165442517540e-01, 2.13859261869047e-01, 2.13627439498779e-01, 2.13456713603193e-01, 2.13334986413159e-01, 2.13251464750556e-01, 2.13196756755039e-01, 2.13162921459585e-01, 2.13143469330120e-01, 2.13133313509104e-01, 2.13128673217724e-01, 2.13126932482155e-01, 2.13126458959650e-01, 2.13126389068719e-01, 2.13126383568621e-01, 2.13126283385378e-01, 2.13125604629340e-01, 2.13123109451644e-01, 2.13116458234367e-01, 2.13101902260519e-01, 2.13074026030299e-01, 2.13025549661609e-01, 2.12947200653872e-01, 2.12827662943909e-01, 2.12653609591811e-01, 2.12409823538238e-01, 2.12079408605987e-01, 2.11644090225789e-01, 2.11084602231125e-01, 2.10381152528868e-01, 2.09513956614671e-01, 2.08463823947426e-01, 2.07212778375123e-01, 2.05744690429214e-01, 2.04045896710326e-01, 2.02105780107487e-01, 1.99917284503172e-01, 1.97477339090682e-01, 1.94787170505407e-01, 1.91852485515469e-01, 1.88683512743722e-01, 1.85294898411399e-01, 1.81705457978637e-01, 1.77937792407050e-01, 1.74017784220268e-01, 1.69973994213294e-01, 1.65836984151582e-01, 1.61638593724319e-01, 1.57411201152954e-01, 1.53186996273007e-01, 1.48997292892773e-01, 1.44871904095481e-01, 1.40838600017361e-01, 1.36922662488175e-01, 1.33146544897969e-01, 1.29529639264845e-01, 1.26088146553798e-01, 1.22835041583087e-01, 1.19780120662725e-01, 1.16930118263923e-01, 1.14288878180218e-01, 1.11857564608471e-01, 1.09634899364877e-01, 1.07617413085691e-01, 1.05799700583106e-01, 1.04174673177212e-01, 1.02733803390748e-01, 1.01467359586522e-01, 1.00364629822039e-01, 9.94141353814086e-02, 9.86038351520619e-02, 9.79213222611577e-02, 9.73540142072977e-02, 9.68893371735362e-02, 9.65149043784095e-02, 9.62186873355364e-02, 9.59891778785078e-02, 9.58155378933441e-02, 9.56877329956711e-02, 9.55966459825248e-02, 9.55341658332287e-02, 9.54932483563624e-02, 9.54679452632377e-02, 9.54533994488571e-02, 9.54458054989686e-02, 9.54423358235914e-02, 9.54410342387604e-02, 9.54406801770915e-02, 9.54406279183928e-02, 9.54406237694485e-02, 9.54405477349383e-02, 9.54400325906590e-02, 9.54381388755650e-02, 9.54330910019081e-02, 9.54220442368694e-02, 9.54008898280279e-02, 9.53641067275584e-02, 9.53046677960998e-02, 9.52140076710037e-02, 9.50820585648984e-02, 9.48973590015356e-02, 9.46472387651482e-02, 9.43180810271684e-02, 9.38956596569795e-02, 9.33655461247866e-02, 9.27135762677551e-02, 9.19263627169045e-02, 9.09918342922538e-02, 8.98997795691230e-02, 8.86423685694382e-02, 8.72146245972800e-02, 8.56148180209999e-02, 8.38447555747877e-02, 8.19099426305030e-02, 7.98196018351405e-02, 7.75865393206203e-02, 7.52268589796898e-02, 7.27595353959032e-02, 7.02058659226026e-02, 6.75888309044552e-02, 6.49323969926741e-02, 6.22608012956789e-02, 5.95978538217850e-02, 5.69662929025014e-02, 5.43872235839160e-02, 5.18796624678970e-02, 4.94602040849799e-02, 4.71428139663601e-02, 4.49387434989035e-02, 4.28565533451746e-02, 4.09022271654749e-02, 3.90793556560283e-02, 3.73893712906153e-02, 3.58318151528749e-02, 3.44046182687947e-02, 3.31043813261971e-02, 3.19266393078550e-02, 3.08661014611857e-02, 2.99168614285321e-02, 2.90725762043422e-02, 2.83266151962351e-02, 2.76721820577990e-02, 2.71024125969603e-02, 2.66104524151984e-02, 2.61895182091934e-02, 2.58329468141767e-02, 2.55342359529180e-02, 2.52870802078462e-02, 2.50854049963891e-02, 2.49234004170944e-02, 2.47955558662767e-02, 2.46966953977363e-02, 2.46220129769199e-02, 2.45671061098705e-02, 2.45280058441892e-02, 2.45012008722938e-02, 2.44836534335917e-02, 2.44728049110563e-02, 2.44665694284834e-02, 2.44633143335165e-02, 2.44618271425310e-02, 2.44612692602288e-02, 2.44611175040887e-02, 2.44610951052238e-02, 2.44610933106202e-02, 2.44610602173481e-02, 2.44608360063268e-02, 2.44600117948721e-02, 2.44578148417958e-02, 2.44530073310329e-02, 2.44438021633471e-02, 2.44277999016873e-02, 2.44019510299438e-02, 2.43625476562670e-02, 2.43052486209479e-02, 2.42251415235728e-02, 2.41168443291286e-02, 2.39746478260514e-02, 2.37926982168233e-02, 2.35652165219875e-02, 2.32867483727830e-02, 2.29524343591692e-02, 2.25582877005337e-02, 2.21014629943141e-02, 2.15804976169380e-02, 2.09955064623699e-02, 2.03483115759866e-02, 1.96424912679589e-02, 1.88833386358738e-02, 1.80777267842984e-02, 1.72338864474142e-02, 1.63611096777316e-02, 1.54693992083708e-02, 1.45690862312478e-02, 1.36704401542858e-02, 1.27832937985630e-02, 1.19167074261242e-02, 1.10786941213945e-02, 1.02760249061191e-02, 9.51412257410311e-03, 8.79703990406711e-03, 8.12750586787558e-03, 7.50701871596959e-03, 6.93596924325846e-03, 6.41378644888846e-03, 5.93910316769097e-03, 5.50993633052599e-03, 5.12386823901068e-03, 4.77820986608890e-03, 4.47013097746662e-03, 4.19675346533666e-03, 3.95521606958482e-03, 3.74272326791925e-03, 3.55658715457707e-03, 3.39426338356930e-03, 3.25337703863227e-03, 3.13173451415594e-03, 3.02732148607276e-03, 2.93829109156718e-03, 2.86294793723130e-03, 2.79973239732101e-03, 2.74720735376377e-03, 2.70404763800514e-03, 2.66903161885068e-03, 2.64103442296459e-03, 2.61902261401303e-03, 2.60205037461958e-03, 2.58925720747527e-03, 2.57986697754120e-03, 2.57318789162753e-03, 2.56861284790944e-03, 2.56561952232914e-03, 2.56376958181193e-03, 2.56270650338308e-03, 2.56215161212481e-03, 2.56189810772867e-03, 2.56180301433062e-03, 2.56177714709338e-03, 2.56177332914831e-03, 2.56177302044591e-03, 2.56176729251933e-03, 2.56172848531833e-03, 2.56158583007959e-03, 2.56120559515693e-03, 2.56037362058237e-03, 2.55878090372619e-03, 2.55601307632958e-03, 2.55154468433716e-03, 2.54473924842102e-03, 2.53485610782889e-03, 2.52106499256399e-03, 2.50246907701523e-03, 2.47813690258170e-03, 2.44714298655424e-03, 2.40861615003496e-03, 2.36179361709787e-03, 2.30607779732850e-03, 2.24109145914042e-03, 2.16672591574252e-03, 2.08317625832422e-03, 1.99095817200146e-03, 1.89090308680306e-03, 1.78413242154818e-03, 1.67201619544244e-03, 1.55612343085920e-03, 1.43816845799449e-03, 1.31994830948656e-03, 1.20325755072770e-03, 1.08976887449330e-03, 9.80889485973080e-04, 8.77639420224341e-04, 7.80620214285932e-04, 6.90115079734710e-04, 6.06285728352418e-04, 5.29362233251618e-04, 4.59718358499117e-04, 3.97784270717239e-04, 3.43810417788859e-04, 2.99608821421771e-04, 2.63704827361266e-04, 2.30915265319709e-04, 2.04355420322599e-04, 1.80255082938174e-04, 1.59565314774568e-04, 1.41899430914805e-04, 1.26313511216836e-04, 1.12894810357034e-04, 1.00816532111025e-04, 8.95898691766206e-05, 7.91338639203332e-05, 7.00322620872385e-05, 6.22094445787980e-05, 5.48419381658505e-05, 4.78692012961801e-05, 4.13687255189643e-05, 3.54316468929028e-05, 3.01188079987469e-05, 2.54241918327125e-05, 2.12752948414163e-05, 1.75916348439387e-05, 1.43233153409028e-05, 1.14461864626492e-05, 8.94778825966114e-06, 6.81729737550540e-06, 5.03999186207693e-06, 3.59500441504679e-06, 2.45517701102926e-06, 1.58806398076194e-06, 9.57411799618058e-07, 5.24714465655052e-07, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -4.26325641456060e-14],
[ 9.90000000000000e+01, 9.89999998767549e+01, 9.89999960657723e+01, 9.89999702457590e+01, 9.89998753279797e+01, 9.89996223068129e+01, 9.89990685497108e+01, 9.89980079372026e+01, 9.89961632058696e+01, 9.89931807780262e+01, 9.89886282839446e+01, 9.89819948994335e+01, 9.89726945376968e+01, 9.89600718543211e+01, 9.89434109525490e+01, 9.89219466164415e+01, 9.88948778542455e+01, 9.88613835028225e+01, 9.88206396228233e+01, 9.87718383966496e+01, 9.87142082179879e+01, 9.86470346230535e+01, 9.85696816518919e+01, 9.84816131404417e+01, 9.83824133355333e+01, 9.82718061095405e+01, 9.81496719510561e+01, 9.80160618495647e+01, 9.78712072017162e+01, 9.77155249633844e+01, 9.75496174611874e+01, 9.73742665492404e+01, 9.71904221252925e+01, 9.69991853669516e+01, 9.68017873706409e+01, 9.65995641341177e+01, 9.63939289897767e+01, 9.61863436581640e+01, 9.59782890534479e+01, 9.57712368533004e+01, 9.55666226718750e+01, 9.53658214761378e+01, 9.51701256896189e+01, 9.49807262538334e+01, 9.47986967777507e+01, 9.46249808031589e+01, 9.44603821454817e+01, 9.43055582284407e+01, 9.41610163082138e+01, 9.40271124700388e+01, 9.39040532707663e+01, 9.37918998899531e+01, 9.36905746370685e+01, 9.35998696423263e+01, 9.35194575338438e+01, 9.34489038753563e+01, 9.33876811080770e+01, 9.33351837092045e+01, 9.32907442498367e+01, 9.32536500084653e+01, 9.32231597746235e+01, 9.31985204624055e+01, 9.31789831471553e+01, 9.31638181421067e+01, 9.31523287463322e+01, 9.31438633217611e+01, 9.31378253954718e+01, 9.31336815335025e+01, 9.31309667929670e+01, 9.31292876285218e+01, 9.31283222048682e+01, 9.31278181461642e+01, 9.31275878328862e+01, 9.31275014336651e+01, 9.31274779309104e+01, 9.31274744619505e+01, 9.31274740068961e+01, 9.31274634046753e+01, 9.31273915730572e+01, 9.31271275122463e+01, 9.31264236216606e+01, 9.31248831671545e+01, 9.31219329849822e+01, 9.31168025073081e+01, 9.31085100599729e+01, 9.30958572277234e+01, 9.30774319019978e+01, 9.30516204169347e+01, 9.30166289354641e+01, 9.29705139652072e+01, 9.29112215623329e+01, 9.28366344240359e+01, 9.27446256868882e+01, 9.26331178570225e+01, 9.25001449261445e+01, 9.23439154111661e+01, 9.21628738382266e+01, 9.19557581187423e+01, 9.17216503717224e+01, 9.14600190458526e+01, 9.11707506628423e+01, 9.08541700723539e+01, 9.05110486749865e+01, 9.01426005239857e+01, 8.97504664915323e+01, 8.93366868068640e+01, 8.89036623802455e+01, 8.84541056318798e+01, 8.79909822192914e+01, 8.75174460758675e+01, 8.70367712260014e+01, 8.65522844195384e+01, 8.60673023079811e+01, 8.55850756468948e+01, 8.51087413170859e+01, 8.46412815003432e+01, 8.41854886249104e+01, 8.37439347779733e+01, 8.33189448661005e+01, 8.29125734780460e+01, 8.25265858724783e+01, 8.21624436769090e+01, 8.18212957948864e+01, 8.15039747906237e+01, 8.12109987612323e+01, 8.09425784858231e+01, 8.06986294863027e+01, 8.04787885457962e+01, 8.02824341910940e+01, 8.01087106348866e+01, 7.99565546736145e+01, 7.98247250344701e+01, 7.97118336530925e+01, 7.96163783394330e+01, 7.95367762546802e+01, 7.94713975813024e+01, 7.94185987271794e+01, 7.93767543703111e+01, 7.93442876296367e+01, 7.93196976464258e+01, 7.93015838846817e+01, 7.92886665114942e+01, 7.92798023006780e+01, 7.92739956144148e+01, 7.92704041547439e+01, 7.92683393342807e+01, 7.92672612864475e+01, 7.92667687116731e+01, 7.92665839289288e+01, 7.92665336634532e+01, 7.92665262443711e+01, 7.92665255894074e+01, 7.92665127557156e+01, 7.92664258055534e+01, 7.92661061673821e+01, 7.92652541280115e+01, 7.92633894537819e+01, 7.92598183530887e+01, 7.92536080961788e+01, 7.92435704475255e+01, 7.92282548798714e+01, 7.92059523232888e+01, 7.91747099516757e+01, 7.91323572171050e+01, 7.90765430050992e+01, 7.90047834006406e+01, 7.89145191306494e+01, 7.88031812961056e+01, 7.86682635462745e+01, 7.85073984065620e+01, 7.83184350842691e+01, 7.80995157798961e+01, 7.78491473621185e+01, 7.75662652541360e+01, 7.72502865519778e+01, 7.69011497646511e+01, 7.65193391306565e+01, 7.61058922060371e+01, 7.56623902937572e+01, 7.51909322249706e+01, 7.46940929178041e+01, 7.41748689236701e+01, 7.36366137273215e+01, 7.30829658314465e+01, 7.25177726229347e+01, 7.19450127437455e+01, 7.13687192815880e+01, 7.07929056749998e+01, 7.02214958848903e+01, 6.96582601497249e+01, 6.91067574781179e+01, 6.85702858656082e+01, 6.80518409826030e+01, 6.75540837437080e+01, 6.70793167684844e+01, 6.66294693529483e+01, 6.62060902654717e+01, 6.58103475060617e+01, 6.54430341274407e+01, 6.51045792785163e+01, 6.47950637491295e+01, 6.45142394260299e+01, 6.42615521840307e+01, 6.40361678184350e+01, 6.38370006716213e+01, 6.36627446208276e+01, 6.35119060806954e+01, 6.33828386382777e+01, 6.32737788850843e+01, 6.31828829455111e+01, 6.31082631295163e+01, 6.30480240663720e+01, 6.30002976132789e+01, 6.29632757855104e+01, 6.29352409312593e+01, 6.29145923811394e+01, 6.28998688442269e+01, 6.28897659020200e+01, 6.28831480682965e+01, 6.28790550330972e+01, 6.28767018866542e+01, 6.28754733154193e+01, 6.28749119672560e+01, 6.28747013855622e+01, 6.28746441021773e+01, 6.28746356472676e+01, 6.28746349690793e+01, 6.28746224530645e+01, 6.28745376552194e+01, 6.28742259293221e+01, 6.28733949821554e+01, 6.28715764727029e+01, 6.28680938022579e+01, 6.28620373862936e+01, 6.28522485465631e+01, 6.28373129847853e+01, 6.28155645926793e+01, 6.27851001103517e+01, 6.27438048591400e+01, 6.26893894406782e+01, 6.26194369098671e+01, 6.25314595005332e+01, 6.24229635218353e+01, 6.22915205731071e+01, 6.21348427758867e+01, 6.19508593327105e+01, 6.17377914347185e+01, 6.14942223945229e+01, 6.12191599092897e+01, 6.09120875791036e+01, 6.05730032150377e+01, 6.02024420456391e+01, 5.98014836261410e+01, 5.93717420162834e+01, 5.89153395638828e+01, 5.84348653653256e+01, 5.79333201385535e+01, 5.74140498182867e+01, 5.68806706494038e+01, 5.63369888879098e+01, 5.57869183853529e+01, 5.52343992955612e+01, 5.46833208803239e+01, 5.41374509127575e+01, 5.36003735328519e+01, 5.30754366821181e+01, 5.25657095315055e+01, 5.20739497079607e+01, 5.16025796797942e+01, 5.11536713995861e+01, 5.07289382089209e+01, 5.03297330400198e+01, 4.99570520544796e+01, 4.96115429927998e+01, 4.92935176373887e+01, 4.90029678988171e+01, 4.87395851152075e+01, 4.85027822101854e+01, 4.82917183905513e+01, 4.81053260846502e+01, 4.79423398280006e+01, 4.78013267939293e+01, 4.76807186428966e+01, 4.75788443247188e+01, 4.74939634145479e+01, 4.74242994999631e+01, 4.73680730687666e+01, 4.73235332826112e+01, 4.72889879689179e+01, 4.72628311311578e+01, 4.72435672730064e+01, 4.72298318608428e+01, 4.72204073146997e+01, 4.72142340202375e+01, 4.72104159906056e+01, 4.72082209712197e+01, 4.72070749641501e+01, 4.72065513419693e+01, 4.72063549128744e+01, 4.72063014793660e+01, 4.72062935926910e+01, 4.72062929843947e+01, 4.72062820613463e+01, 4.72062080560877e+01, 4.72059360049517e+01, 4.72052108166849e+01, 4.72036237608046e+01, 4.72005843623953e+01, 4.71952988338947e+01, 4.71867560437316e+01, 4.71737218696213e+01, 4.71547426063323e+01, 4.71281578875133e+01, 4.70921233321535e+01, 4.70446428334109e+01, 4.69836100695080e+01, 4.69068584373888e+01, 4.68122182013633e+01, 4.66975792306563e+01, 4.65609572990857e+01, 4.64005615706005e+01, 4.62148606324997e+01, 4.60026442984619e+01, 4.57630784135873e+01, 4.54957500689584e+01, 4.52007009737635e+01, 4.48784472227188e+01, 4.45299843058943e+01, 4.41567768996278e+01, 4.37607337115510e+01, 4.33441683918622e+01, 4.29097482288819e+01, 4.24604329754601e+01, 4.19994066476675e+01, 4.15300054321120e+01, 4.10556448732494e+01, 4.05797492613346e+01, 4.01056856383803e+01, 3.96367041797521e+01, 3.91758860254547e+01, 3.87260990502769e+01, 3.82899616459193e+01, 3.78698143432297e+01, 3.74676989791832e+01, 3.70853450443831e+01, 3.67241627805074e+01, 3.63852425118223e+01, 3.60693595977309e+01, 3.57769843069370e+01, 3.55082958616058e+01, 3.52631998960127e+01, 3.50413486197434e+01, 3.48421630604899e+01, 3.46648568690040e+01, 3.45084612800543e+01, 3.43718509215557e+01, 3.42537702370367e+01, 3.41528603272854e+01, 3.40676860235998e+01, 3.39967629803147e+01, 3.39385845244947e+01, 3.38916479346345e+01, 3.38544797481443e+01, 3.38256596300091e+01, 3.38038422825537e+01, 3.37877768476012e+01, 3.37763232542748e+01, 3.37684650022843e+01, 3.37633179426729e+01, 3.37601347233501e+01, 3.37583046999886e+01, 3.37573492662535e+01, 3.37569127214615e+01, 3.37567489585638e+01, 3.37567044110968e+01, 3.37566978359844e+01, 3.37566973381368e+01, 3.37566885188155e+01, 3.37566287666318e+01, 3.37564091115179e+01, 3.37558235934271e+01, 3.37545422105235e+01, 3.37520882439110e+01, 3.37478208847642e+01, 3.37409239917383e+01, 3.37304016919574e+01, 3.37150814026773e+01, 3.36936246862405e+01, 3.36645461512253e+01, 3.36262403722105e+01, 3.35770165170386e+01, 3.35151400465563e+01, 3.34388804965765e+01, 3.33465639813223e+01, 3.32366286948381e+01, 3.31076813603750e+01, 3.29585523195304e+01, 3.27883467951153e+01, 3.25964898330621e+01, 3.23827625508385e+01, 3.21473276041584e+01, 3.18907422290439e+01, 3.16139578071390e+01, 3.13183056086457e+01, 3.10054691449681e+01, 3.06774443546730e+01, 3.03364895842107e+01, 2.99850679370379e+01, 2.96257849835738e+01, 2.92613249973467e+01, 2.88943887840913e+01, 2.85276358101730e+01, 2.81636327619235e+01, 2.78048099573007e+01, 2.74534262825497e+01, 2.71115426354856e+01, 2.67810033012572e+01, 2.64634243120779e+01, 2.61601876593798e+01, 2.58724402129102e+01, 2.56010963124719e+01, 2.53468431817054e+01, 2.51101485204473e+01, 2.48912698254896e+01, 2.46902651469204e+01, 2.45070051007424e+01, 2.43411860306146e+01, 2.41923442503286e+01, 2.40598713132910e+01, 2.39430302535799e+01, 2.38409727300840e+01, 2.37527569832878e+01, 2.36773664842173e+01, 2.36137291171802e+01, 2.35607366930021e+01, 2.35172645393023e+01, 2.34821908621995e+01, 2.34544155240460e+01, 2.34328778395702e+01, 2.34165729635510e+01, 2.34045664318624e+01, 2.33960064283611e+01, 2.33901333850177e+01, 2.33862865823423e+01, 2.33839074998271e+01, 2.33825397680847e+01, 2.33818256901459e+01, 2.33814994223326e+01, 2.33813770280595e+01, 2.33813437338541e+01, 2.33813388196993e+01, 2.33813384507921e+01, 2.33813319576082e+01, 2.33812879653404e+01, 2.33811262452389e+01, 2.33806951598120e+01, 2.33797517449133e+01, 2.33779450117302e+01, 2.33748031476592e+01, 2.33697252212947e+01, 2.33619779104691e+01, 2.33506976698573e+01, 2.33348986312159e+01, 2.33134863797367e+01, 2.32852775701485e+01, 2.32490251345442e+01, 2.32034485928992e+01, 2.31472687146547e+01, 2.30792455093081e+01, 2.29982182652765e+01, 2.29031461334617e+01, 2.27931475909849e+01, 2.26675370457996e+01, 2.25258568726388e+01, 2.23679033133160e+01, 2.21937449252773e+01, 2.20037326038451e+01, 2.17985006075931e+01, 2.15789584494570e+01, 2.13462739471449e+01, 2.11018481319215e+01, 2.08472830822813e+01, 2.05843440731758e+01, 2.03149177059000e+01, 2.00409678907469e+01, 1.97644916602125e+01, 1.94874767514729e+01, 1.92118626793617e+01, 1.89395066253074e+01, 1.86721549427103e+01, 1.84114205168104e+01, 1.81587657234240e+01, 1.79154903894077e+01, 1.76827240018117e+01, 1.74614214217784e+01, 1.72523614760865e+01, 1.70561479572387e+01, 1.68732127090649e+01, 1.67038205802751e+01, 1.65480760877278e+01, 1.64059316538963e+01, 1.62771972842243e+01, 1.61615515433434e+01, 1.60585536831099e+01, 1.59676567731951e+01, 1.58882216854776e+01, 1.58195317834088e+01, 1.57608081631440e+01, 1.57112252817410e+01, 1.56699267880065e+01, 1.56360413443994e+01, 1.56086981963328e+01, 1.55870422121545e+01, 1.55702480877254e+01, 1.55575333887443e+01, 1.55481700962185e+01, 1.55414943292459e+01, 1.55369139466303e+01, 1.55339137752351e+01, 1.55320582770941e+01, 1.55309915462152e+01, 1.55304346154886e+01, 1.55301801489948e+01, 1.55300846897881e+01, 1.55300587225656e+01, 1.55300548898579e+01, 1.55300546030102e+01, 1.55300495658215e+01, 1.55300154381306e+01, 1.55298899811783e+01, 1.55295555593856e+01, 1.55288236874039e+01, 1.55274220724086e+01, 1.55249846753490e+01, 1.55210452559659e+01, 1.55150347977473e+01, 1.55062831282849e+01, 1.54940249545255e+01, 1.54774104178982e+01, 1.54555201386856e+01, 1.54273845615510e+01, 1.53920072367897e+01, 1.53483914801185e+01, 1.52955696570162e+01, 1.52326341488367e+01, 1.51587688930223e+01, 1.50732802658033e+01, 1.49756260087988e+01, 1.48654409034347e+01, 1.47425579757597e+01, 1.46070241685867e+01, 1.44591096400354e+01, 1.42993101235310e+01, 1.41283420963571e+01, 1.39471308331002e+01, 1.37567917488363e+01, 1.35586057485154e+01, 1.33539895787358e+01, 1.31444624110123e+01, 1.29316100562123e+01, 1.27170483027639e+01, 1.25023868741611e+01, 1.22891954084167e+01, 1.20789726775298e+01, 1.18731200042033e+01, 1.16729195214831e+01, 1.14795175905443e+01, 1.12939133753428e+01, 1.11169522987995e+01, 1.09493238937656e+01, 1.07915634231510e+01, 1.06440565771982e+01, 1.05070465537960e+01, 1.03806428765400e+01, 1.02648313892334e+01, 1.01594849691340e+01, 1.00643746107865e+01, 9.97918063683440e+00, 9.90350388389025e+00, 9.83687678532763e+00, 9.77877432611440e+00, 9.72862487687297e+00, 9.68582092608645e+00, 9.64972972281349e+00, 9.61970382043269e+00, 9.59509147855049e+00, 9.57524683960997e+00, 9.55953975364203e+00, 9.54736508384421e+00, 9.53815129192523e+00, 9.53136807955289e+00, 9.52653285405218e+00, 9.52321579477366e+00, 9.52104332191803e+00, 9.51969981116275e+00, 9.51892745283099e+00, 9.51852421999233e+00, 9.51833998136270e+00, 9.51827086738086e+00, 9.51825206672927e+00, 9.51824929179351e+00, 9.51824908403393e+00, 9.51824543463337e+00, 9.51822070942281e+00, 9.51812981716000e+00, 9.51788753350051e+00, 9.51735731047792e+00, 9.51634190383476e+00, 9.51457620918227e+00, 9.51172266190189e+00, 9.50736951302945e+00, 9.50103224106286e+00, 9.49215829375798e+00, 9.48013527038761e+00, 9.46430255017198e+00, 9.44396624487993e+00, 9.41841720382703e+00, 9.38695163200356e+00, 9.34889370511814e+00, 9.30361939113359e+00, 9.25058053189331e+00, 9.18932811822765e+00, 9.11953362518683e+00, 9.04100727627936e+00, 8.95371218760974e+00, 8.85777350833858e+00, 8.75348191786594e+00, 8.64129114798743e+00, 8.52180954654875e+00, 8.39578605767510e+00, 8.26409132912028e+00, 8.12769493751067e+00, 7.98763992087289e+00, 7.84501590809141e+00, 7.70093213205101e+00, 7.55649151462030e+00, 7.41276683515819e+00, 7.27077976465003e+00, 7.13148329236719e+00, 6.99574781714652e+00, 6.86435094251665e+00, 6.73797081914995e+00, 6.61718272798933e+00, 6.50245849536203e+00, 6.39416827520897e+00, 6.29258421772896e+00, 6.19788556088716e+00, 6.11016472320063e+00, 6.02943403486343e+00, 5.95563281208875e+00, 5.88863454998062e+00, 5.82825407693161e+00, 5.77425457434795e+00, 5.72635441650337e+00, 5.68423382469206e+00, 5.64754135670429e+00, 5.61590026692264e+00, 5.58891477463214e+00, 5.56617626966300e+00, 5.54726946694811e+00, 5.53177849716860e+00, 5.51929289194035e+00, 5.50941339183257e+00, 5.50175747691041e+00, 5.49596449542172e+00, 5.49170024939363e+00, 5.48866088843398e+00, 5.48657596637159e+00, 5.48521053001774e+00, 5.48436613478873e+00, 5.48388071670600e+00, 5.48362729206701e+00, 5.48351150193382e+00, 5.48346806532869e+00, 5.48345624954455e+00, 5.48345450556047e+00, 5.48345437464195e+00, 5.48345207036603e+00, 5.48343645857105e+00, 5.48337906816106e+00, 5.48322608793367e+00, 5.48289130223540e+00, 5.48225017741365e+00, 5.48113534934996e+00, 5.47933374152312e+00, 5.47658551813350e+00, 5.47258504546865e+00, 5.46698399386328e+00, 5.45939665908386e+00, 5.44940751301706e+00, 5.43658090768157e+00, 5.42047275433740e+00, 5.40064388400562e+00, 5.37667467324410e+00, 5.34818039877744e+00, 5.31482667827244e+00, 5.27634427531312e+00, 5.23254250738754e+00, 5.18332050723141e+00, 5.12867565682180e+00, 5.06870864046692e+00, 5.00362474276211e+00, 4.93373123578563e+00, 4.85943093916667e+00, 4.78121227436782e+00, 4.69963634778409e+00, 4.61532176562424e+00, 4.52892799184295e+00, 4.44113810079337e+00, 4.35264174907374e+00, 4.26411910423549e+00, 4.17622633561801e+00, 4.08958311211994e+00, 4.00476238150952e+00, 3.92228254242897e+00, 3.84260197642821e+00, 3.76611579147711e+00, 3.69315454417824e+00, 3.62398465516583e+00, 3.55881020795297e+00, 3.49777582111348e+00, 3.44097030180585e+00, 3.38843081993009e+00, 3.34014738177407e+00, 3.29606742564651e+00, 3.25610040621942e+00, 3.22012227633276e+00, 3.18797981271101e+00, 3.15949476384077e+00, 3.13446782311942e+00, 3.11268244772266e+00, 3.09390855329779e+00, 3.07790611680559e+00, 3.06442871521357e+00, 3.05322701724853e+00, 3.04405223033163e+00, 3.03665948671510e+00, 3.03081113350278e+00, 3.02627987257652e+00, 3.02285168036167e+00, 3.02032842559335e+00, 3.01853009720422e+00, 3.01729655510155e+00, 3.01648872430316e+00, 3.01598916739236e+00, 3.01570199063647e+00, 3.01555206396080e+00, 3.01548356247376e+00, 3.01545786540315e+00, 3.01545087519595e+00, 3.01544984345671e+00, 3.01544976567560e+00, 3.01544839226850e+00, 3.01543908724308e+00, 3.01540488113231e+00, 3.01531370132845e+00, 3.01511416228536e+00, 3.01473204309356e+00, 3.01406760329006e+00, 3.01299387630699e+00, 3.01135606613678e+00, 3.00897215499255e+00, 3.00563480591576e+00, 3.00111461225042e+00, 2.99516470380115e+00, 2.98752666628312e+00, 2.97793766662470e+00, 2.96613860382735e+00, 2.95188302743718e+00, 2.93494648925521e+00, 2.91513592634793e+00, 2.89229862326216e+00, 2.86633027696162e+00, 2.83718169629043e+00, 2.80486371284402e+00, 2.76944996226254e+00, 2.73107730997842e+00, 2.68994383482739e+00, 2.64630443562612e+00, 2.60046427572706e+00, 2.55277041439442e+00, 2.50360207905419e+00, 2.45336009988711e+00, 2.40245605313916e+00, 2.35130164201959e+00, 2.30029878869880e+00, 2.24983082580161e+00, 2.20025507123543e+00, 2.15189695732121e+00, 2.10504577458411e+00, 2.05995199114673e+00, 2.01682602701911e+00, 1.97583830255114e+00, 1.93712034309264e+00, 1.90076670635993e+00, 1.86683750224098e+00, 1.83536129279037e+00, 1.80633818852048e+00, 1.77974299143567e+00, 1.75552827172768e+00, 1.73362730053688e+00, 1.71395679338181e+00, 1.69641944626395e+00, 1.68090626823077e+00, 1.66729873004428e+00, 1.65547075865419e+00, 1.64529061179863e+00, 1.63662266680097e+00, 1.62932915318912e+00, 1.62327185091044e+00, 1.61831376552153e+00, 1.61432077975111e+00, 1.61116326829173e+00, 1.60871765063614e+00, 1.60686784629409e+00, 1.60550658878071e+00, 1.60453655016346e+00, 1.60387122724872e+00, 1.60343554391213e+00, 1.60316613149709e+00, 1.60301126012695e+00, 1.60293040738024e+00, 1.60289346602080e+00, 1.60287960819170e+00, 1.60287583854036e+00, 1.60287528214823e+00, 1.60287523997447e+00, 1.60287449227081e+00, 1.60286942647578e+00, 1.60285080418484e+00, 1.60280116482039e+00, 1.60269253447707e+00, 1.60248451043515e+00, 1.60212280513706e+00, 1.60153832812927e+00, 1.60064687793513e+00, 1.59934950696713e+00, 1.59753361001768e+00, 1.59507476912134e+00, 1.59183936337537e+00, 1.58768792088794e+00, 1.58247915149703e+00, 1.57607455452165e+00, 1.56834344815354e+00, 1.55916822009090e+00, 1.54844955767727e+00, 1.53611138577718e+00, 1.52210522741494e+00, 1.50641371034570e+00, 1.48905297481402e+00, 1.47007379371863e+00, 1.44956129307410e+00, 1.42763325185159e+00, 1.40443705736102e+00, 1.38014548534420e+00, 1.35495155297563e+00, 1.32906274965697e+00, 1.30269497928769e+00, 1.27606654663795e+00, 1.24939249136219e+00, 1.22287952121030e+00, 1.19672172859362e+00, 1.17109720038535e+00, 1.14616555797392e+00, 1.12206640007980e+00, 1.09891856954725e+00, 1.07682012971026e+00, 1.05584891625008e+00, 1.03606352514183e+00, 1.01750460359408e+00, 1.00019632551364e+00, 9.84147952684509e-01, 9.69355404616337e-01, 9.55802781586002e-01, 9.43463805140041e-01, 9.32303157251255e-01, 9.22277712976665e-01, 9.13337671769014e-01, 9.05427599727332e-01, 8.98487399346280e-01, 8.92453225091401e-01, 8.87258362758374e-01, 8.82834088408373e-01, 8.79110519045852e-01, 8.76017462443230e-01, 8.73485267965734e-01, 8.71445674279542e-01, 8.69832643844120e-01, 8.68583168532098e-01, 8.67638026030534e-01, 8.66942463273602e-01, 8.66446781397244e-01, 8.66106796830806e-01, 8.65884155255586e-01, 8.65746479191188e-01, 8.65667335681138e-01, 8.65626017537892e-01, 8.65607139371122e-01, 8.65600057589262e-01, 8.65598131179539e-01, 8.65597846845731e-01, 8.65597825156883e-01, 8.65597438828132e-01, 8.65594821397536e-01, 8.65585199498412e-01, 8.65559551451994e-01, 8.65503423379394e-01, 8.65395939259173e-01, 8.65209048043809e-01, 8.64907048104925e-01, 8.64446425774602e-01, 8.63776040036569e-01, 8.62837678793753e-01, 8.61567002837504e-01, 8.59894881015137e-01, 8.57749103754074e-01, 8.55056442128102e-01, 8.51744996742866e-01, 8.47746756282234e-01, 8.43000261738609e-01, 8.37453251856843e-01, 8.31065151169597e-01, 8.23809257056439e-01, 8.15674488739234e-01, 8.06666580081625e-01, 7.96808628987577e-01, 7.86140956871250e-01, 7.74720278307683e-01, 7.62618228687501e-01, 7.49919341216568e-01, 7.36718599133084e-01, 7.23118711005043e-01, 7.09227264676244e-01, 6.95153909109479e-01, 6.81007695108420e-01, 6.66894679093744e-01, 6.52915862790024e-01, 6.39165509799298e-01, 6.25729850887964e-01, 6.12686165672940e-01, 6.00102210427593e-01, 5.88035950096270e-01, 5.76535546700640e-01, 5.65639555082555e-01, 5.55377279121359e-01, 5.45769246018125e-01, 5.36827761965910e-01, 5.28557518784993e-01, 5.20956227375436e-01, 5.14015259791236e-01, 5.07720287163593e-01, 5.02051905470438e-01, 4.96986245183180e-01, 4.92495564065010e-01, 4.88548824800532e-01, 4.85112260670874e-01, 4.82149933135759e-01, 4.79624284958618e-01, 4.77496691468517e-01, 4.75728010794072e-01, 4.74279131585220e-01, 4.73111514056882e-01, 4.72187717387882e-01, 4.71471903849547e-01, 4.70930307794177e-01, 4.70531656053481e-01, 4.70247525586359e-01, 4.70052624514722e-01, 4.69924984047523e-01, 4.69846051174713e-01, 4.69800675278975e-01, 4.69776985745980e-01, 4.69766161979408e-01, 4.69762101639489e-01, 4.69760997131274e-01, 4.69760834108257e-01, 4.69760821585183e-01, 4.69760597369827e-01, 4.69759078279239e-01, 4.69753493965172e-01, 4.69738608415440e-01, 4.69706032673898e-01, 4.69643649756776e-01, 4.69535176370243e-01, 4.69359884148048e-01, 4.69092500850934e-01, 4.68703309037465e-01, 4.68158455792107e-01, 4.67420481819019e-01, 4.66449071273800e-01, 4.65202014941255e-01, 4.63636368824908e-01, 4.61709778253692e-01, 4.59381924920237e-01, 4.56616041884431e-01, 4.53380430806305e-01, 4.49649907953116e-01, 4.45407102229737e-01, 4.40643530676468e-01, 4.35360385109490e-01, 4.29568977689172e-01, 4.23290812276136e-01, 4.16557270869305e-01, 4.09408928119063e-01, 4.01894529585675e-01, 3.94069688926351e-01, 3.85995373845930e-01, 3.77736259403676e-01, 3.69359029835650e-01, 3.60930706819616e-01, 3.52517073988537e-01, 3.44181255707753e-01, 3.35982494006188e-01, 3.27975152401095e-01, 3.20207960362865e-01, 3.12723498336894e-01, 3.05557911332372e-01, 2.98740829607183e-01, 2.92295468169073e-01, 2.86238872715131e-01, 2.80582278125280e-01, 2.75331546439048e-01, 2.70487653979956e-01, 2.66047201436777e-01, 2.62002925704053e-01, 2.58344197603989e-01, 2.55057494814054e-01, 2.52126844079910e-01, 2.49534230860235e-01, 2.47259977767783e-01, 2.45283095451583e-01, 2.43581610887844e-01, 2.42132878460538e-01, 2.40913878811114e-01, 2.39901509358646e-01, 2.39072868789407e-01, 2.38405535870598e-01, 2.37877840841136e-01, 2.37469125573497e-01, 2.37159986877217e-01, 2.36932495906970e-01, 2.36770385791345e-01, 2.36659199415300e-01, 2.36586389809972e-01, 2.36541366808169e-01, 2.36515485427105e-01, 2.36501973702474e-01, 2.36495800241130e-01, 2.36493484388120e-01, 2.36492854422323e-01, 2.36492761440800e-01, 2.36492754243642e-01, 2.36492624674938e-01, 2.36491746829358e-01, 2.36488519796410e-01, 2.36479917866561e-01, 2.36461093541237e-01, 2.36425045724800e-01, 2.36362367715302e-01, 2.36261088747938e-01, 2.36106621774411e-01, 2.35881828680817e-01, 2.35567212011574e-01, 2.35141239205191e-01, 2.34580801132367e-01, 2.33861801217515e-01, 2.32959864636001e-01, 2.31851149225648e-01, 2.30513231282202e-01, 2.28926031002786e-01, 2.27072734878135e-01, 2.24940666796982e-01, 2.22522056992972e-01, 2.19814659056084e-01, 2.16822170547644e-01, 2.13554422363970e-01, 2.10027315416681e-01, 2.06262499430053e-01, 2.02286806246676e-01, 1.98131467250377e-01, 1.93831159581066e-01, 1.89422937152159e-01, 1.84945108911393e-01, 1.80436127693156e-01, 1.75933548429853e-01, 1.71473105116803e-01, 1.67087943046525e-01, 1.62808028109282e-01, 1.58659740164723e-01, 1.54665644155585e-01, 1.50844421879242e-01, 1.47210939776997e-01, 1.43776423924791e-01, 1.40548712417242e-01, 1.37532557009777e-01, 1.34729949465182e-01, 1.32140452697399e-01, 1.29761521761580e-01, 1.27588804470392e-01, 1.25616415615397e-01, 1.23837182292476e-01, 1.22242860611095e-01, 1.20824326091963e-01, 1.19571741368119e-01, 1.18474705489149e-01, 1.17522389309485e-01, 1.16703661237537e-01, 1.16007207122408e-01, 1.15421647325206e-01, 1.14935653114344e-01, 1.14538063487160e-01, 1.14218002418198e-01, 1.13964995430237e-01, 1.13769083354413e-01, 1.13620930262769e-01, 1.13511921889227e-01, 1.13434250465644e-01, 1.13380981828925e-01, 1.13346100922087e-01, 1.13324532411127e-01, 1.13312134025946e-01, 1.13305661348022e-01, 1.13302704018537e-01, 1.13301594636795e-01, 1.13301292859433e-01, 1.13301248317768e-01, 1.13301244842134e-01, 1.13301181909591e-01, 1.13300755533132e-01, 1.13299188139990e-01, 1.13295010135772e-01, 1.13285867134536e-01, 1.13268358951274e-01, 1.13237917573670e-01, 1.13188730976049e-01, 1.13113719183502e-01, 1.13004568322069e-01, 1.12851827356010e-01, 1.12645070659710e-01, 1.12373127360229e-01, 1.12024375432995e-01, 1.11587094849298e-01, 1.11049869789213e-01, 1.10402025336114e-01, 1.09634079587096e-01, 1.08738188301473e-01, 1.07708556677878e-01, 1.06541792171427e-01, 1.05237173858953e-01, 1.03796817901241e-01, 1.02225724980116e-01, 1.00531703701662e-01, 9.87251730606289e-02, 9.68188561460295e-02, 9.48273852518469e-02, 9.27668444511482e-02, 9.06542787835463e-02, 8.85071992553890e-02, 8.63431101412289e-02, 8.41790803227980e-02, 8.20313744888352e-02, 7.99151537328981e-02, 7.78442490267111e-02, 7.58310056935017e-02, 7.38861928218029e-02, 7.20189688874768e-02, 7.02368937267011e-02, 6.85459770552370e-02, 6.69507543865677e-02, 6.54543820470224e-02, 6.40587439411036e-02, 6.27645639136034e-02, 6.15715190068329e-02, 6.04783504098947e-02, 5.94829701339847e-02, 5.85825622829219e-02, 5.77736783222065e-02, 5.70523261962421e-02, 5.64140536301778e-02, 5.58540264276388e-02, 5.53671029123564e-02, 5.49479057605566e-02, 5.45908923404141e-02, 5.42904243872329e-02, 5.40408374785386e-02, 5.38365103732872e-02, 5.36719338616365e-02, 5.35417783403744e-02, 5.34409589078422e-02, 5.33646963964429e-02, 5.33085724736679e-02, 5.32685767846317e-02, 5.32411441040637e-02, 5.32231796275937e-02, 5.32120708501793e-02, 5.32056849394849e-02, 5.32023510732529e-02, 5.32008278375429e-02, 5.32002564249581e-02, 5.32001009874108e-02, 5.32000780451796e-02, 5.32000762397189e-02, 5.32000433539821e-02, 5.31998205485280e-02, 5.31990014971342e-02, 5.31968182452829e-02, 5.31920404470156e-02, 5.31828911389596e-02, 5.31669827534955e-02, 5.31412768489616e-02, 5.31020707569212e-02, 5.30450139623042e-02, 5.29651564795728e-02, 5.28570306960319e-02, 5.27147670639749e-02, 5.25322425992272e-02, 5.23032593916497e-02, 5.20217483139049e-02, 5.16819909555295e-02, 5.12788507003563e-02, 5.08080020563568e-02, 5.02661461045903e-02, 4.96511995165037e-02, 4.89624451791286e-02, 4.82006341332806e-02, 4.73680311961715e-02, 4.64684000873063e-02, 4.55069277898303e-02, 4.44900918893296e-02, 4.34254783705086e-02, 4.23215604683236e-02, 4.11874513462465e-02, 4.00326443760264e-02, 3.88667545315826e-02, 3.76992730303414e-02, 3.65393452109117e-02, 3.53955791541267e-02, 3.42758900415987e-02, 3.31873828003069e-02, 3.21362731840935e-02, 3.11278451765986e-02, 3.01664407378297e-02, 2.92554767785731e-02, 2.83974839299646e-02, 2.75941619240569e-02, 2.68464467968386e-02, 2.61545854133800e-02, 2.55182130821083e-02, 2.49364305439741e-02, 2.44078775328092e-02, 2.39308012475163e-02, 2.35031191127888e-02, 2.31224758744588e-02, 2.27862953708211e-02, 2.24918274234227e-02, 2.22361903929169e-02, 2.20164101065734e-02, 2.18294560212939e-02, 2.16722755286156e-02, 2.15418271895231e-02, 2.14351134217452e-02, 2.13492128228268e-02, 2.12813119633353e-02, 2.12287361711105e-02, 2.11889785721695e-02, 2.11597264647310e-02, 2.11388839881295e-02, 2.11245900160293e-02, 2.11152302608468e-02, 2.11094427225618e-02, 2.11061158435513e-02, 2.11043790233560e-02, 2.11035854810859e-02, 2.11032878000594e-02, 2.11032068239945e-02, 2.11031948721114e-02, 2.11031939232100e-02, 2.11031765332024e-02, 2.11030587136207e-02, 2.11026255998377e-02, 2.11014711046147e-02, 2.10989446576804e-02, 2.10941067366690e-02, 2.10856952006537e-02, 2.10721043489599e-02, 2.10513785125954e-02, 2.10212218100025e-02, 2.09790254196633e-02, 2.09219132964762e-02, 2.08468066496152e-02, 2.07505066864258e-02, 2.06297941172409e-02, 2.04815427508720e-02, 2.03028432573973e-02, 2.00911319323609e-02, 1.98443181736110e-02, 1.95609035114188e-02, 1.92400845826626e-02, 1.88818326057856e-02, 1.84869428935315e-02, 1.80570498025929e-02, 1.75946050835734e-02, 1.71028203781844e-02, 1.65855770299146e-02, 1.60473080964891e-02, 1.54928586695462e-02, 1.49273318327658e-02, 1.43559290407183e-02, 1.37837947578134e-02, 1.32158747280755e-02, 1.26567947239522e-02, 1.21107630480708e-02, 1.15814974364241e-02, 1.10721764814912e-02, 1.05854162143970e-02, 1.01232717018149e-02, 9.68726039172907e-03, 9.27840031903315e-03, 8.89725534631225e-03, 8.54398254907096e-03, 8.21838126644536e-03, 7.91994537712526e-03, 7.64791848175150e-03, 7.40134810122584e-03, 7.17913342494114e-03, 6.98006320254763e-03, 6.80284436967327e-03, 6.64612483211613e-03, 6.50851381089979e-03, 6.38860108283268e-03, 6.28497440099807e-03, 6.19623378368317e-03, 6.12100217693978e-03, 6.05793323772730e-03, 6.00571771126113e-03, 5.96308979571347e-03, 5.92883428167568e-03, 5.90179458011164e-03, 5.88088129410380e-03, 5.86508080385784e-03, 5.85346331708081e-03, 5.84518988316063e-03, 5.83951791385573e-03, 5.83580478844681e-03, 5.83350916446079e-03, 5.83218968503729e-03, 5.83150087811646e-03, 5.83118617487344e-03, 5.83106812180745e-03, 5.83103600878587e-03, 5.83103126896845e-03, 5.83103088906967e-03, 5.83102388245038e-03, 5.83097641171548e-03, 5.83080190675389e-03, 5.83033675978584e-03, 5.82931889448446e-03, 5.82736993375477e-03, 5.82398184096178e-03, 5.81850890790747e-03, 5.81016594892958e-03, 5.79803350798613e-03, 5.78107076294998e-03, 5.75813659148139e-03, 5.72801893019914e-03, 5.68947212720161e-03, 5.64126150237598e-03, 5.58221385510649e-03, 5.51127223091896e-03, 5.42755283541396e-03, 5.33040142744312e-03, 5.21944567120158e-03, 5.09463883855688e-03, 4.95628940636550e-03, 4.80507142236775e-03, 4.64201285289789e-03, 4.46846332746986e-03, 4.28604700158530e-03, 4.09660795227751e-03, 3.90215291682239e-03, 3.70479070652152e-03, 3.50666379968345e-03, 3.30986987694359e-03, 3.11637914924745e-03, 2.92796105078708e-03, 2.74613407398761e-03, 2.57214401732142e-03, 2.40696498811133e-03, 2.25131259000477e-03, 2.10566262130385e-03, 1.97027490050061e-03, 1.84522202208370e-03, 1.73041717295064e-03, 1.62563256427006e-03, 1.53050752520301e-03, 1.44455800854359e-03, 1.36720325618001e-03, 1.29781416173103e-03, 1.23577041205181e-03, 1.18050369744083e-03, 1.13151004104850e-03, 1.08833239697503e-03, 1.05053133631827e-03, 1.01766422127980e-03, 9.89282377702239e-04, 9.64942477907016e-04, 9.44221819466406e-04, 9.26728677612232e-04, 9.12104496379129e-04, 9.00019862584145e-04, 8.90168418550275e-04, 8.82262125628586e-04, 8.76029236167142e-04, 8.71214620033811e-04, 8.67581383374727e-04, 8.64912804900086e-04, 8.63014006327489e-04, 8.61713120528407e-04, 8.60861906929244e-04, 8.60335814664381e-04, 8.60033486475231e-04, 8.59875679765310e-04, 8.59803584788210e-04, 8.59776540806934e-04, 8.59769184316406e-04, 8.59768098507629e-04, 8.59768010897710e-04, 8.59766387279137e-04, 8.59755387082828e-04, 8.59714949513091e-04, 8.59607160336395e-04, 8.59371279275933e-04, 8.58919589205698e-04, 8.58134259132726e-04, 8.56865416437813e-04, 8.54930645530771e-04, 8.52116181832230e-04, 8.48180167523083e-04, 8.42858463201424e-04, 8.35873601317871e-04, 8.26947363535169e-04, 8.15816936579949e-04, 8.02253474958547e-04, 7.86080317595861e-04, 7.67186795172847e-04, 7.45533794329845e-04, 7.21150141643534e-04, 6.94124175211641e-04, 6.64599991310411e-04, 6.32788305622967e-04, 5.98993592632269e-04, 5.63642447289681e-04, 5.27282472091883e-04, 4.90525815777687e-04, 4.53948869207466e-04, 4.18003969585357e-04, 3.83000592250937e-04, 3.49169004550731e-04, 3.16764523773827e-04, 2.86117066565339e-04, 2.57540792802047e-04, 2.31173271259877e-04, 2.06926105647653e-04, 1.84557388010376e-04, 1.63846991725336e-04, 1.44844218539220e-04, 1.27627299605138e-04, 1.11921934646375e-04, 9.75871142969709e-05, 8.36503097758623e-05, 6.89757035274852e-05, 5.41988114690639e-05, 3.98614491530225e-05, 2.63505172881651e-05, 1.46429652971847e-05, 5.37361306385264e-06, 4.69426036886489e-06, 3.64698045984824e-06, 9.37834656156156e-07, 3.22846852895964e-06, 3.87053798078796e-06, 3.63768562294808e-06, 3.08563761564074e-06, 2.58743354208946e-06, 2.30240785015212e-06, 2.19391671407720e-06, 2.13841764207245e-06, 2.03880475169171e-06, 1.85984984568677e-06, 1.61194250836161e-06, 1.32599647884035e-06, 1.03542380713861e-06, 7.66759114867455e-07, 5.36631901582041e-07, 3.52392504510135e-07, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -2.84217094304040e-14],
[ 1.00000000000000e+02, 9.99999998749552e+01, 9.99999960083207e+01, 9.99999698112573e+01, 9.99998735073914e+01, 9.99996167913432e+01, 9.99990549476609e+01, 9.99979788467661e+01, 9.99961071759374e+01, 9.99930811931296e+01, 9.99884622123666e+01, 9.99817319452060e+01, 9.99722957377016e+01, 9.99594886611498e+01, 9.99425843423529e+01, 9.99208063589990e+01, 9.98933419801742e+01, 9.98593580005103e+01, 9.98180183953248e+01, 9.97685035063630e+01, 9.97100304439568e+01, 9.96418743515250e+01, 9.95633901144279e+01, 9.94740340045472e+01, 9.93733846398678e+01, 9.92611625192880e+01, 9.91372472897493e+01, 9.90016918432151e+01, 9.88547323518717e+01, 9.86967934504317e+01, 9.85284879704001e+01, 9.83506109115766e+01, 9.81641276734317e+01, 9.79701569238030e+01, 9.77699488105315e+01, 9.75648594831592e+01, 9.73563230582388e+01, 9.71458222210817e+01, 9.69348586139662e+01, 9.67249240353122e+01, 9.65174732943534e+01, 9.63138993621229e+01, 9.61155112596093e+01, 9.59235149480222e+01, 9.57389973456839e+01, 9.55629134941192e+01, 9.53960768288882e+01, 9.52391524710226e+01, 9.50926534335119e+01, 9.49569396255504e+01, 9.48322195283159e+01, 9.47185544052067e+01, 9.46158648941448e+01, 9.45239398089538e+01, 9.44424469513172e+01, 9.43709457056185e+01, 9.43089011575988e+01, 9.42556994460211e+01, 9.42106640262039e+01, 9.41730724972230e+01, 9.41421736226000e+01, 9.41172041591910e+01, 9.40974051024354e+01, 9.40820369596256e+01, 9.40703936775964e+01, 9.40618148779651e+01, 9.40556960919833e+01, 9.40514967377552e+01, 9.40487456439135e+01, 9.40470439940257e+01, 9.40460656426590e+01, 9.40455548342868e+01, 9.40453214369756e+01, 9.40452338808184e+01, 9.40452100633482e+01, 9.40452065479369e+01, 9.40452060867891e+01, 9.40451953425984e+01, 9.40451225491138e+01, 9.40448549523821e+01, 9.40441416363458e+01, 9.40425805545540e+01, 9.40395908689778e+01, 9.40343916955365e+01, 9.40259882202021e+01, 9.40131659921089e+01, 9.39944940176680e+01, 9.39683370674437e+01, 9.39328773602876e+01, 9.38861455028977e+01, 9.38260602364395e+01, 9.37504761787752e+01, 9.36572383614913e+01, 9.35442419638710e+01, 9.34094952691849e+01, 9.32511835494464e+01, 9.30677313672642e+01, 9.28578607126583e+01, 9.26206425040160e+01, 9.23555392871007e+01, 9.20624374376552e+01, 9.17416677407789e+01, 9.13940137786770e+01, 9.10207079997715e+01, 9.06234156062484e+01, 9.02042065211115e+01, 8.97655158323458e+01, 8.93100934798734e+01, 8.88409447093065e+01, 8.83612639047250e+01, 8.78743654729247e+01, 8.73836159475200e+01, 8.68923710135234e+01, 8.64039197718925e+01, 8.59214368083133e+01, 8.54479412174682e+01, 8.49862611281984e+01, 8.45390024757875e+01, 8.41085214224347e+01, 8.36969005111762e+01, 8.33059290745348e+01, 8.29370885350036e+01, 8.25915431024662e+01, 8.22701361175845e+01, 8.19733920182202e+01, 8.17015236834076e+01, 8.14544447597672e+01, 8.12317864936075e+01, 8.10329185595687e+01, 8.08569733717386e+01, 8.07028733671113e+01, 8.05693607511459e+01, 8.04550291839801e+01, 8.03583568616315e+01, 8.02777404111708e+01, 8.02115289769634e+01, 8.01580578328699e+01, 8.01156808197968e+01, 8.00828008862552e+01, 8.00578980080656e+01, 8.00395537872715e+01, 8.00264720832120e+01, 8.00174951119384e+01, 8.00116145628158e+01, 8.00079774199387e+01, 8.00058863354946e+01, 8.00047945753011e+01, 8.00042957351914e+01, 8.00041086020963e+01, 8.00040576972676e+01, 8.00040501838183e+01, 8.00040495205238e+01, 8.00040365235935e+01, 8.00039484674669e+01, 8.00036247636563e+01, 8.00027618868046e+01, 8.00008734951168e+01, 7.99972569730536e+01, 7.99909677292514e+01, 7.99808024201398e+01, 7.99652920813253e+01, 7.99427059304790e+01, 7.99110663514405e+01, 7.98681752735042e+01, 7.98116518180255e+01, 7.97389806959341e+01, 7.96475704095957e+01, 7.95348198531824e+01, 7.93981914378852e+01, 7.92352884202740e+01, 7.90439337183950e+01, 7.88222471985482e+01, 7.85687182435405e+01, 7.82822704036786e+01, 7.79623151099814e+01, 7.76087918087140e+01, 7.72221924558673e+01, 7.68035690686573e+01, 7.63545239234141e+01, 7.58771829435459e+01, 7.53741537419612e+01, 7.48484705622610e+01, 7.43035289056572e+01, 7.37430128771325e+01, 7.31708182356860e+01, 7.25909738551373e+01, 7.20075639056924e+01, 7.14246526717288e+01, 7.08462136109482e+01, 7.02760640510445e+01, 6.97178067641209e+01, 6.91747794755371e+01, 6.86500130876559e+01, 6.81461990184531e+01, 6.76656656188557e+01, 6.72103632244512e+01, 6.67818570945336e+01, 6.63813273342862e+01, 6.60095748778990e+01, 6.56670326918100e+01, 6.53537814870845e+01, 6.50695693645815e+01, 6.48138349284486e+01, 6.45857334808072e+01, 6.43841659521297e+01, 6.42078102318048e+01, 6.40551545471398e+01, 6.39245325016411e+01, 6.38141593295245e+01, 6.37221688578118e+01, 6.36466505956325e+01, 6.35856862990248e+01, 6.35373852961633e+01, 6.34999178105293e+01, 6.34715454958991e+01, 6.34506484039273e+01, 6.34357476475192e+01, 6.34255231035904e+01, 6.34188256167740e+01, 6.34146833176483e+01, 6.34123018487736e+01, 6.34110584905127e+01, 6.34104903859976e+01, 6.34102772697569e+01, 6.34102192969132e+01, 6.34102107402408e+01, 6.34102100538899e+01, 6.34101973872332e+01, 6.34101115687674e+01, 6.34097960909607e+01, 6.34089551425846e+01, 6.34071147457478e+01, 6.34035901584115e+01, 6.33974608488354e+01, 6.33875541942245e+01, 6.33724388766505e+01, 6.33504287402786e+01, 6.33195976282971e+01, 6.32778054286415e+01, 6.32227352192792e+01, 6.31519410151748e+01, 6.30629051850719e+01, 6.29531041401114e+01, 6.28200804204843e+01, 6.26615188522716e+01, 6.24753240530317e+01, 6.22596962737771e+01, 6.20132024171695e+01, 6.17348390991758e+01, 6.14240848413840e+01, 6.10809388911371e+01, 6.07059447425208e+01, 6.03001971303093e+01, 5.98653320375173e+01, 5.94035000408504e+01, 5.89173240713509e+01, 5.84098433566741e+01, 5.78844459125614e+01, 5.73447924403926e+01, 5.67947348337827e+01, 5.62382326606189e+01, 5.56792709303889e+01, 5.51217821621995e+01, 5.45695752539191e+01, 5.40262729758612e+01, 5.34952591639747e+01, 5.29796359695158e+01, 5.24821909228620e+01, 5.20053731433378e+01, 5.15512777881337e+01, 5.11216377583827e+01, 5.07178217231233e+01, 5.03408376298975e+01, 4.99913409992835e+01, 4.96696474199375e+01, 4.93757487566326e+01, 4.91093326542173e+01, 4.88698049694202e+01, 4.86563147952388e+01, 4.84677817626196e+01, 4.83029253119123e+01, 4.81602956208493e+01, 4.80383058547843e+01, 4.79352653678525e+01, 4.78494134318543e+01, 4.77789530067379e+01, 4.77220839986154e+01, 4.76770353861132e+01, 4.76420955422922e+01, 4.76156400461346e+01, 4.75961562724159e+01, 4.75822640775408e+01, 4.75727319648479e+01, 4.75664882161741e+01, 4.75626266140868e+01, 4.75604065451056e+01, 4.75592474599594e+01, 4.75587178623133e+01, 4.75585191916177e+01, 4.75584651483398e+01, 4.75584571716641e+01, 4.75584565564261e+01, 4.75584455087267e+01, 4.75583706589396e+01, 4.75580955032320e+01, 4.75573620393523e+01, 4.75557568726727e+01, 4.75526827905957e+01, 4.75473369490349e+01, 4.75386966826588e+01, 4.75255137969671e+01, 4.75063180204199e+01, 4.74794300823769e+01, 4.74429846308637e+01, 4.73949629080769e+01, 4.73332347596482e+01, 4.72556091702714e+01, 4.71598921049376e+01, 4.70439500114805e+01, 4.69057769340526e+01, 4.67435628322994e+01, 4.65557604340127e+01, 4.63411478048118e+01, 4.60988838248964e+01, 4.58285539362674e+01, 4.55302038651487e+01, 4.52043595191606e+01, 4.48520318795604e+01, 4.44747064196558e+01, 4.40743173418636e+01, 4.36532076968027e+01, 4.32140771832558e+01, 4.27599200734716e+01, 4.22939561978353e+01, 4.18195581859356e+01, 4.13401781438745e+01, 4.08592766416871e+01, 4.03802563450550e+01, 3.99064019641547e+01, 3.94408275432322e+01, 3.89864315851218e+01, 3.85458601425073e+01, 3.81214777945085e+01, 3.77153463049909e+01, 3.73292106650433e+01, 3.69644921164732e+01, 3.66222876282096e+01, 3.63033751704505e+01, 3.60082240293968e+01, 3.57370093514020e+01, 3.54896301098316e+01, 3.52657297473733e+01, 3.50647188467025e+01, 3.48857993034182e+01, 3.47279895966763e+01, 3.45901508577484e+01, 3.44710135128213e+01, 3.43692043174598e+01, 3.42832736053774e+01, 3.42117225470271e+01, 3.41530301609631e+01, 3.41056797521219e+01, 3.40681843766435e+01, 3.40391108634260e+01, 3.40171018685967e+01, 3.40008954093943e+01, 3.39893413253750e+01, 3.39814141513950e+01, 3.39762219593617e+01, 3.39730108319276e+01, 3.39711647657324e+01, 3.39702009566444e+01, 3.39697605851877e+01, 3.39695953867927e+01, 3.39695504488372e+01, 3.39695438160896e+01, 3.39695433138779e+01, 3.39695344172495e+01, 3.39694741412991e+01, 3.39692525607748e+01, 3.39686619103376e+01, 3.39673692958788e+01, 3.39648938212551e+01, 3.39605890650026e+01, 3.39536317430867e+01, 3.39430172784948e+01, 3.39275628613621e+01, 3.39059184174915e+01, 3.38765857014694e+01, 3.38379454873788e+01, 3.37882925429652e+01, 3.37258777446794e+01, 3.36489563304237e+01, 3.35558409104031e+01, 3.34449574876076e+01, 3.33149024073591e+01, 3.31644978928123e+01, 3.29928436634725e+01, 3.27993621064993e+01, 3.25838345984657e+01, 3.23464268704519e+01, 3.20877017706920e+01, 3.18086183902643e+01, 3.15105172464107e+01, 3.11950920172328e+01, 3.08643491289791e+01, 3.05205572409971e+01, 3.01661892783437e+01, 2.98038600591303e+01, 2.94362627024541e+01, 2.90661068627591e+01, 2.86960614349607e+01, 2.83287037673021e+01, 2.79664766916239e+01, 2.76116539341596e+01, 2.72663138000156e+01, 2.69323205044444e+01, 2.66113121927184e+01, 2.63046945495310e+01, 2.60136389187281e+01, 2.57390839858374e+01, 2.54817402649313e+01, 2.52420968301229e+01, 2.50204299068669e+01, 2.48168130719980e+01, 2.46311289009714e+01, 2.44630819522756e+01, 2.43122130026510e+01, 2.41779144524527e+01, 2.40594468153472e+01, 2.39559561941772e+01, 2.38664926261195e+01, 2.37900291545507e+01, 2.37254814515319e+01, 2.36717277736428e+01, 2.36276289866406e+01, 2.35920483442824e+01, 2.35638706582084e+01, 2.35420204544403e+01, 2.35254786834622e+01, 2.35132975402013e+01, 2.35046129616025e+01, 2.34986544053401e+01, 2.34947515739823e+01, 2.34923378328820e+01, 2.34909501734539e+01, 2.34902256907595e+01, 2.34898946687659e+01, 2.34897704910095e+01, 2.34897367116509e+01, 2.34897317258884e+01, 2.34897313516056e+01, 2.34897247638048e+01, 2.34896801304927e+01, 2.34895160538301e+01, 2.34890786865663e+01, 2.34881215234608e+01, 2.34862884586699e+01, 2.34831007969984e+01, 2.34779488315294e+01, 2.34700885120346e+01, 2.34586436241994e+01, 2.34426137742998e+01, 2.34208883225313e+01, 2.33922662267280e+01, 2.33554815452626e+01, 2.33092341060061e+01, 2.32522245850766e+01, 2.31831929683087e+01, 2.31009591092038e+01, 2.30044638731862e+01, 2.28928091947315e+01, 2.27652952950351e+01, 2.26214533312144e+01, 2.24610718818523e+01, 2.22842159143041e+01, 2.20912372109037e+01, 2.18827756291264e+01, 2.16597510060045e+01, 2.14233459627981e+01, 2.11749803016761e+01, 2.09162780979927e+01, 2.06490289668116e+01, 2.03751452999880e+01, 2.00966174949081e+01, 1.98154692796152e+01, 1.95337151355027e+01, 1.92533215089206e+01, 1.89761730164624e+01, 1.87040442673477e+01, 1.84385773632466e+01, 1.81812646966199e+01, 1.79334364147413e+01, 1.76962518497290e+01, 1.74706942872292e+01, 1.72575685875250e+01, 1.70575013185102e+01, 1.68709431695882e+01, 1.66981734757594e+01, 1.65393066989222e+01, 1.63943007054597e+01, 1.62629666627179e+01, 1.61449803641876e+01, 1.60398947895623e+01, 1.59471537110183e+01, 1.58661061670413e+01, 1.57960216344583e+01, 1.57361057330160e+01, 1.56855162914614e+01, 1.56433795883830e+01, 1.56088065560432e+01, 1.55809087039966e+01, 1.55588134857864e+01, 1.55416788015939e+01, 1.55287063076609e+01, 1.55191531943937e+01, 1.55123421030138e+01, 1.55076688776775e+01, 1.55046078966477e+01, 1.55027147910429e+01, 1.55016264399182e+01, 1.55010582215656e+01, 1.55007985976830e+01, 1.55007012037651e+01, 1.55006747102546e+01, 1.55006707998678e+01, 1.55006705072065e+01, 1.55006653679272e+01, 1.55006305485571e+01, 1.55005025489224e+01, 1.55001613492992e+01, 1.54994146443381e+01, 1.54979846229382e+01, 1.54954978285534e+01, 1.54914785744202e+01, 1.54853463182999e+01, 1.54764173191559e+01, 1.54639108005226e+01, 1.54469597288261e+01, 1.54246261770367e+01, 1.53959210840173e+01, 1.53598280397411e+01, 1.53153305316688e+01, 1.52614418873124e+01, 1.51972369553471e+01, 1.51218843983235e+01, 1.50346783410392e+01, 1.49350680459694e+01, 1.48226842835874e+01, 1.46973611382724e+01, 1.45591521404262e+01, 1.44083398360936e+01, 1.42454381845860e+01, 1.40711874961061e+01, 1.38865419669899e+01, 1.36926502213476e+01, 1.34908296061255e+01, 1.32825352932260e+01, 1.30693254978938e+01, 1.28528243073927e+01, 1.26346837099486e+01, 1.24165464079645e+01, 1.22000108873980e+01, 1.19866000044710e+01, 1.17777340613777e+01, 1.15747090042239e+01, 1.13786800243050e+01, 1.11906505128044e+01, 1.10114660380416e+01, 1.08418128028961e+01, 1.06822199063098e+01, 1.05330646748379e+01, 1.03945803381068e+01, 1.02668653808829e+01, 1.01498939975407e+01, 1.00435271859315e+01, 9.94752413310846e+00, 9.86155365421362e+00, 9.78520554038688e+00, 9.71800174702956e+00, 9.65940740778041e+00, 9.60884169168693e+00, 9.56568853231533e+00, 9.52930725012034e+00, 9.49904306638938e+00, 9.47423747231504e+00, 9.45423837464232e+00, 9.43840989466236e+00, 9.42614165500702e+00, 9.41685735348372e+00, 9.41002239930349e+00, 9.40515037783339e+00, 9.40180811756692e+00, 9.39961915801393e+00, 9.39826545877672e+00, 9.39748724580879e+00, 9.39708095711411e+00, 9.39689532241943e+00, 9.39682568475460e+00, 9.39680674165133e+00, 9.39680394569005e+00, 9.39680373635635e+00, 9.39680005930460e+00, 9.39677514675435e+00, 9.39668356582911e+00, 9.39643944658634e+00, 9.39590520713455e+00, 9.39488211123073e+00, 9.39310305328078e+00, 9.39022792988474e+00, 9.38584191466613e+00, 9.37945690092999e+00, 9.37051631058824e+00, 9.35840338343722e+00, 9.34245295464237e+00, 9.32196659829114e+00, 9.29623086198380e+00, 9.26453814609643e+00, 9.22620960011372e+00, 9.18061922991171e+00, 9.12721825013553e+00, 9.06555859304235e+00, 8.99531441768779e+00, 8.91630046712847e+00, 8.82848620744629e+00, 8.73200485449784e+00, 8.62715664679638e+00, 8.51440604038702e+00, 8.39437285940271e+00, 8.26781780280714e+00, 8.13562304918408e+00, 7.99876898439439e+00, 7.85830827482334e+00, 7.71533860524606e+00, 7.57097539092197e+00, 7.42632566675691e+00, 7.28246417111798e+00, 7.14041240387527e+00, 7.00112117569896e+00, 6.86545690513267e+00, 6.73419168327855e+00, 6.60799692832265e+00, 6.48744030198834e+00, 6.37298545954559e+00, 6.26499415161820e+00, 6.16373018378485e+00, 6.06936476102973e+00, 5.98198278988059e+00, 5.90158977312143e+00, 5.82811900255268e+00, 5.76143882776182e+00, 5.70135984791784e+00, 5.64764193517689e+00, 5.60000104959963e+00, 5.55811584483989e+00, 5.52163409051313e+00, 5.49017895111493e+00, 5.46335516330343e+00, 5.44075514451946e+00, 5.42196504802143e+00, 5.40657075464166e+00, 5.39416376248511e+00, 5.38434690525146e+00, 5.37673980086568e+00, 5.37098390762149e+00, 5.36674704777090e+00, 5.36372725061145e+00, 5.36165577007132e+00, 5.36029914608540e+00, 5.35946020420099e+00, 5.35897792237186e+00, 5.35872613546532e+00, 5.35861109369989e+00, 5.35856793784538e+00, 5.35855619843333e+00, 5.35855446572170e+00, 5.35855433564934e+00, 5.35855204626735e+00, 5.35853653538136e+00, 5.35847951593054e+00, 5.35832752459829e+00, 5.35799490334777e+00, 5.35735792475566e+00, 5.35625031031547e+00, 5.35446037046685e+00, 5.35172997065388e+00, 5.34775549682266e+00, 5.34219095333827e+00, 5.33465327304688e+00, 5.32472985034563e+00, 5.31198822212251e+00, 5.29598771880129e+00, 5.27629279167525e+00, 5.25248759951835e+00, 5.22419131648319e+00, 5.19107351642113e+00, 5.15286890930200e+00, 5.10939066646216e+00, 5.06054158385720e+00, 5.00632240297301e+00, 4.94683673822691e+00, 4.88229224127117e+00, 4.81299785334653e+00, 4.73935723770195e+00, 4.66185872254910e+00, 4.58106229793844e+00, 4.49758437686869e+00, 4.41208113690968e+00, 4.32523129616466e+00, 4.23771914728347e+00, 4.15021858388462e+00, 4.06337871944926e+00, 3.97781153716696e+00, 3.89408183856459e+00, 3.81269959535537e+00, 3.73411466546017e+00, 3.65871371871351e+00, 3.58681913401719e+00, 3.51868957748205e+00, 3.45452194746841e+00, 3.39445437278400e+00, 3.33856996929016e+00, 3.28690109243601e+00, 3.23943386383988e+00, 3.19611279467057e+00, 3.15684537367734e+00, 3.12150653043915e+00, 3.08994292258477e+00, 3.06197702782225e+00, 3.03741104659498e+00, 3.01603063851648e+00, 2.99760852531637e+00, 2.98190799513836e+00, 2.96868633830165e+00, 2.95769823405301e+00, 2.94869909268506e+00, 2.94144833924065e+00, 2.93571260563759e+00, 2.93126877931904e+00, 2.92790684034560e+00, 2.92543240691938e+00, 2.92366890307703e+00, 2.92245926265895e+00, 2.92166709102864e+00, 2.92117722011916e+00, 2.92089561234496e+00, 2.92074859332263e+00, 2.92068142039903e+00, 2.92065622172372e+00, 2.92064936709265e+00, 2.92064835536415e+00, 2.92064827909161e+00, 2.92064693232197e+00, 2.92063780776968e+00, 2.92060426509990e+00, 2.92051485380775e+00, 2.92031918521866e+00, 2.91994447882851e+00, 2.91929293126070e+00, 2.91824004507762e+00, 2.91663404133065e+00, 2.91429646245444e+00, 2.91102404883976e+00, 2.90659194095980e+00, 2.90075821740538e+00, 2.89326972654273e+00, 2.88386910600676e+00, 2.87230281189350e+00, 2.85832990233158e+00, 2.84173124410901e+00, 2.82231874385749e+00, 2.79994415550613e+00, 2.77450699168998e+00, 2.74596107543677e+00, 2.71431931386627e+00, 2.67965635805686e+00, 2.64210892842942e+00, 2.60187372437052e+00, 2.55920298822420e+00, 2.51439794302587e+00, 2.46780045607794e+00, 2.41978338409358e+00, 2.37074012104333e+00, 2.32107389243660e+00, 2.27118731986144e+00, 2.22147272200975e+00, 2.17230353149739e+00, 2.12402710110855e+00, 2.07695905993461e+00, 2.03137926985366e+00, 1.98752933477770e+00, 1.94561153544984e+00, 1.90578900490981e+00, 1.86818692496038e+00, 1.83289451067331e+00, 1.79996755507233e+00, 1.76943132551564e+00, 1.74128363247887e+00, 1.71549792608947e+00, 1.69202631211292e+00, 1.67080241412138e+00, 1.65174404010891e+00, 1.63475563845556e+00, 1.61973054914704e+00, 1.60655307131972e+00, 1.59510037767350e+00, 1.58524431048644e+00, 1.57685309344494e+00, 1.56979298894900e+00, 1.56392992272761e+00, 1.55913108734769e+00, 1.55526652444080e+00, 1.55221067319091e+00, 1.54984386085216e+00, 1.54805370081564e+00, 1.54673635596122e+00, 1.54579762049671e+00, 1.54515377274261e+00, 1.54473215459169e+00, 1.54447144053537e+00, 1.54432156971824e+00, 1.54424332767037e+00, 1.54420757915156e+00, 1.54419416879664e+00, 1.54419052086898e+00, 1.54418998244300e+00, 1.54418994163109e+00, 1.54418921807114e+00, 1.54418431585309e+00, 1.54416629488719e+00, 1.54411825842401e+00, 1.54401313593740e+00, 1.54381182965577e+00, 1.54346180595775e+00, 1.54289620779743e+00, 1.54203355807461e+00, 1.54077811535267e+00, 1.53902093118928e+00, 1.53664164113613e+00, 1.53351099792440e+00, 1.52949412473801e+00, 1.52445442888549e+00, 1.51825807284487e+00, 1.51077885311200e+00, 1.50190329138829e+00, 1.49153570233689e+00, 1.47960297296369e+00, 1.46605877605633e+00, 1.45088694846643e+00, 1.43410379685587e+00, 1.41575914872095e+00, 1.39593604186925e+00, 1.37474903489168e+00, 1.35234121600740e+00, 1.32888007814945e+00, 1.30455250460653e+00, 1.27955916382166e+00, 1.25410863876097e+00, 1.22841161385352e+00, 1.20267541283432e+00, 1.17709912908126e+00, 1.15186952365659e+00, 1.12715779364432e+00, 1.10311724260106e+00, 1.07988182270375e+00, 1.05756546909289e+00, 1.03626211325653e+00, 1.01604624417563e+00, 9.96973881729872e-01, 9.79083833797148e-01, 9.62399123308614e-01, 9.46928491020614e-01, 9.32667901102789e-01, 9.19601997601902e-01, 9.07705478883031e-01, 8.96944373354287e-01, 8.87277212741125e-01, 8.78656108858436e-01, 8.71027746427004e-01, 8.64334308324729e-01, 8.58514351114323e-01, 8.53503648117169e-01, 8.49236015048064e-01, 8.45644129612197e-01, 8.42660351807524e-01, 8.40217546300110e-01, 8.38249902497964e-01, 8.36693742223844e-01, 8.35488299590622e-01, 8.34576453232074e-01, 8.33905387833425e-01, 8.33427160275662e-01, 8.33099145884802e-01, 8.32884342359890e-01, 8.32751512869521e-01, 8.32675155329426e-01, 8.32635291618189e-01, 8.32617077971634e-01, 8.32610245470704e-01, 8.32608386871087e-01, 8.32608112545898e-01, 8.32608091620486e-01, 8.32607718890600e-01, 8.32605193594162e-01, 8.32595910387383e-01, 8.32571165151037e-01, 8.32517012757776e-01, 8.32413311922636e-01, 8.32232998691396e-01, 8.31941627354368e-01, 8.31497214167683e-01, 8.30850414692136e-01, 8.29945059146254e-01, 8.28719061188053e-01, 8.27105703360246e-01, 8.25035286693634e-01, 8.22437112733908e-01, 8.19241744215518e-01, 8.15383467114458e-01, 8.10802853931690e-01, 8.05449308412065e-01, 7.99283458385389e-01, 7.92279258794608e-01, 7.84425673368801e-01, 7.75727821806186e-01, 7.66207509237560e-01, 7.55903093974169e-01, 7.44868694423502e-01, 7.33172781847642e-01, 7.20896247266623e-01, 7.08130063663717e-01, 6.94972685354131e-01, 6.81527333309830e-01, 6.67899308701806e-01, 6.54193458998662e-01, 6.40511894966608e-01, 6.26952026765210e-01, 6.13604956835658e-01, 6.00554239584028e-01, 5.87874995130591e-01, 5.75633347686235e-01, 5.63886148518250e-01, 5.52680938336906e-01, 5.42056103207443e-01, 5.32041180554600e-01, 5.22657276332936e-01, 5.13917560030677e-01, 5.05827810144808e-01, 4.98386988626521e-01, 4.91587828252733e-01, 4.85417421761269e-01, 4.79857805810951e-01, 4.74886536335546e-01, 4.70477254627795e-01, 4.66600245481899e-01, 4.63222989924077e-01, 4.60310715464161e-01, 4.57826946424831e-01, 4.55734055803504e-01, 4.53993818391254e-01, 4.52567962652665e-01, 4.51418716349025e-01, 4.50509338284661e-01, 4.49804626118947e-01, 4.49271388164519e-01, 4.48878865717205e-01, 4.48599091925274e-01, 4.48407173631892e-01, 4.48281484061724e-01, 4.48203756624721e-01, 4.48159073348322e-01, 4.48135745312939e-01, 4.48125086691988e-01, 4.48121088299757e-01, 4.48120000642419e-01, 4.48119840106543e-01, 4.48119827774534e-01, 4.48119606979873e-01, 4.48118111064900e-01, 4.48112611944040e-01, 4.48097953466672e-01, 4.48065874558594e-01, 4.48004442719807e-01, 4.47897621960919e-01, 4.47724997318787e-01, 4.47461675759328e-01, 4.47078380309364e-01, 4.46541750346867e-01, 4.45814855788285e-01, 4.44857926158605e-01, 4.43629287064624e-01, 4.42086486466053e-01, 4.40187581700087e-01, 4.37892546117226e-01, 4.35164742417939e-01, 4.31972399570697e-01, 4.28290022906197e-01, 4.24099663910127e-01, 4.19391978382805e-01, 4.14167009500801e-01, 4.08434645737785e-01, 4.02214721697831e-01, 3.95536751144675e-01, 3.88439303882492e-01, 3.80969059497033e-01, 3.73179589292008e-01, 3.65129931488781e-01, 3.56883032959942e-01, 3.48504133201757e-01, 3.40059163308368e-01, 3.31613225278488e-01, 3.23229206227111e-01, 3.14966569263930e-01, 3.06880349141437e-01, 2.99020367323621e-01, 2.91430668722683e-01, 2.84149171614445e-01, 2.77207513585523e-01, 2.70631069996327e-01, 2.64439117411570e-01, 2.58645112639812e-01, 2.53257058206231e-01, 2.48277926924416e-01, 2.43706121344346e-01, 2.39535947817430e-01, 2.35758089327163e-01, 2.32360065712967e-01, 2.29326674146492e-01, 2.26640406453626e-01, 2.24281842929685e-01, 2.22230024552857e-01, 2.20462806917979e-01, 2.18957199788832e-01, 2.17689695964353e-01, 2.16636592271342e-01, 2.15774304072454e-01, 2.15079672883292e-01, 2.14530264718533e-01, 2.14104654837726e-01, 2.13782692837114e-01, 2.13545740721045e-01, 2.13376875831372e-01, 2.13261050421124e-01, 2.13185200269265e-01, 2.13138296023544e-01, 2.13111332842189e-01, 2.13097256241468e-01, 2.13090824665599e-01, 2.13088411981971e-01, 2.13087755675645e-01, 2.13087658806302e-01, 2.13087651308214e-01, 2.13087516321906e-01, 2.13086601770968e-01, 2.13083239803368e-01, 2.13074278173224e-01, 2.13054666581058e-01, 2.13017110780626e-01, 2.12951809517378e-01, 2.12846288375943e-01, 2.12685343398434e-01, 2.12451105718046e-01, 2.12123236187708e-01, 2.11679255831683e-01, 2.11095013688620e-01, 2.10345288126717e-01, 2.09404511012124e-01, 2.08247596410729e-01, 2.06850847228594e-01, 2.05192904991648e-01, 2.03255700655788e-01, 2.01025358850998e-01, 1.98493005212526e-01, 1.95655427203604e-01, 1.92515543567026e-01, 1.89082646319690e-01, 1.85372391627382e-01, 1.81406531071460e-01, 1.77212391471599e-01, 1.72822128013408e-01, 1.68271790350104e-01, 1.63600253145788e-01, 1.58848070090215e-01, 1.54056313094927e-01, 1.49265456091854e-01, 1.44514356050408e-01, 1.39839373427236e-01, 1.35273661514759e-01, 1.30846640441845e-01, 1.26583658262348e-01, 1.22505829787599e-01, 1.18630034419638e-01, 1.14969047699304e-01, 1.11531777705210e-01, 1.08323576623690e-01, 1.05346599306259e-01, 1.02600183849702e-01, 1.00081233537168e-01, 9.77845842734411e-02, 9.57033464377730e-02, 9.38292144975179e-02, 9.21527415393939e-02, 9.06635789571055e-02, 8.93506838501708e-02, 8.82024982640388e-02, 8.72071053032499e-02, 8.63523674593409e-02, 8.56260523069636e-02, 8.50159501237187e-02, 8.45099870726642e-02, 8.40963364367724e-02, 8.37635291032086e-02, 8.35005631541463e-02, 8.32970111274278e-02, 8.31431223602976e-02, 8.30299169146969e-02, 8.29492669804210e-02, 8.28939614189039e-02, 8.28577492719802e-02, 8.28353586080937e-02, 8.28224879703328e-02, 8.28157688511197e-02, 8.28126989473361e-02, 8.28115473394035e-02, 8.28112340759191e-02, 8.28111878389564e-02, 8.28111842309909e-02, 8.28111189031517e-02, 8.28106762983225e-02, 8.28090492509261e-02, 8.28047122539175e-02, 8.27952214201986e-02, 8.27770475535203e-02, 8.27454499782334e-02, 8.26943985022268e-02, 8.26165505116450e-02, 8.25032895879900e-02, 8.23448310212598e-02, 8.21303979853809e-02, 8.18484697805815e-02, 8.14871003299942e-02, 8.10343010465147e-02, 8.04784774031759e-02, 7.98089033487770e-02, 7.90162125845910e-02, 7.80928812586836e-02, 7.70336735284047e-02, 7.58360203429476e-02, 7.45003032342026e-02, 7.30300191449231e-02, 7.14318092895923e-02, 6.97153442645799e-02, 6.78930682467467e-02, 6.59798160089338e-02, 6.39923263722579e-02, 6.19486834141867e-02, 5.98677213397991e-02, 5.77684299158534e-02, 5.56693947982296e-02, 5.35883014832450e-02, 5.15415238907906e-02, 4.95438098509595e-02, 4.76080671228800e-02, 4.57452459689591e-02, 4.39643083984720e-02, 4.22722702910363e-02, 4.06743007010846e-02, 3.91738624369680e-02, 3.77728790799807e-02, 3.64719154968896e-02, 3.52703612088590e-02, 3.41666084073609e-02, 3.31582187425283e-02, 3.22420751109433e-02, 3.14145164584403e-02, 3.06714550547582e-02, 3.00084768018394e-02, 2.94209259411957e-02, 2.89039760727263e-02, 2.84526897244319e-02, 2.80620688576505e-02, 2.77270986701694e-02, 2.74427868985470e-02, 2.72042005330917e-02, 2.70065014692449e-02, 2.68449821488161e-02, 2.67151017219334e-02, 2.66125227172864e-02, 2.65331476790749e-02, 2.64731547547967e-02, 2.64290308319062e-02, 2.63976005607134e-02, 2.63760494861458e-02, 2.63619395587682e-02, 2.63532155015582e-02, 2.63482008566456e-02, 2.63455829984309e-02, 2.63443869318394e-02, 2.63439382547404e-02, 2.63438162045446e-02, 2.63437981902257e-02, 2.63437967725508e-02, 2.63437709504188e-02, 2.63435960022491e-02, 2.63429528812225e-02, 2.63412386104562e-02, 2.63374872439783e-02, 2.63303039746603e-02, 2.63178154827131e-02, 2.62976395462999e-02, 2.62668768297800e-02, 2.62221275559824e-02, 2.61595353811401e-02, 2.60748601360490e-02, 2.59635801018590e-02, 2.58210231037737e-02, 2.56425239343940e-02, 2.54236035278410e-02, 2.51601630417042e-02, 2.48486837895285e-02, 2.44864220852179e-02, 2.40715868261105e-02, 2.36034873398694e-02, 2.30826398658621e-02, 2.25108231098403e-02, 2.18910765140805e-02, 2.12276389537607e-02, 2.05258300927511e-02, 1.97918810933828e-02, 1.90327252527283e-02, 1.82557619689447e-02, 1.74686089192875e-02, 1.66788573466476e-02, 1.58938440182510e-02, 1.51204510375322e-02, 1.43649416598066e-02, 1.36328369924144e-02, 1.29288352790269e-02, 1.22567726026546e-02, 1.16196214300146e-02, 1.10195215575715e-02, 1.04578367822086e-02, 9.93523007161912e-03, 9.45175013681876e-03, 9.00692301586048e-03, 8.59984337084541e-03, 8.22926144901714e-03, 7.89366286866766e-03, 7.59133944434609e-03, 7.32045013579352e-03, 7.07907190871992e-03, 6.86524085938345e-03, 6.67698439492881e-03, 6.51234557423663e-03, 6.36940090397786e-03, 6.24627295967173e-03, 6.14113917886527e-03, 6.05223807468747e-03, 5.97787396384319e-03, 5.91642110724422e-03, 5.86632793881847e-03, 5.82612182142839e-03, 5.79441452769913e-03, 5.76990841317127e-03, 5.75140303929800e-03, 5.73780183052008e-03, 5.72811822104313e-03, 5.72148067519151e-03, 5.71713595388701e-03, 5.71445004902671e-03, 5.71290631258137e-03, 5.71210045679038e-03, 5.71173228057376e-03, 5.71159416914213e-03, 5.71155659990552e-03, 5.71155105478738e-03, 5.71155061454931e-03, 5.71154254659234e-03, 5.71148788516496e-03, 5.71128694735989e-03, 5.71075134654109e-03, 5.70957932721683e-03, 5.70733526788558e-03, 5.70343438758414e-03, 5.69713370857983e-03, 5.68753033546443e-03, 5.67356807940200e-03, 5.65405335105140e-03, 5.62768102605560e-03, 5.59307061968184e-03, 5.54881256883277e-03, 5.49352370951794e-03, 5.42591018545124e-03, 5.34483509134986e-03, 5.24938724318247e-03, 5.13894670260129e-03, 5.01324220662980e-03, 4.87239559676311e-03, 4.71694880046414e-03, 4.54786991988954e-03, 4.36653647111029e-03, 4.17469564087014e-03, 3.97440337334842e-03, 3.76794593810814e-03, 3.55774917473184e-03, 3.34628174319107e-03, 3.13595933874922e-03, 2.92905683143374e-03, 2.72763447732416e-03, 2.53348259900577e-03, 2.34808659914165e-03, 2.17261143770031e-03, 2.00790262476858e-03, 1.85450002638277e-03, 1.71266133804693e-03, 1.58239317470077e-03, 1.46348833243337e-03, 1.35556737744480e-03, 1.25812169486039e-03, 1.17055443686809e-03, 1.09221618190247e-03, 1.02243351177833e-03, 9.60530411902027e-04, 9.05843603732137e-04, 8.57733287006113e-04, 8.15590509873232e-04, 7.78841937957964e-04, 7.46952478763774e-04, 7.19426105590060e-04, 6.95805233149827e-04, 6.75669023721071e-04, 6.58630998302101e-04, 6.44336290619663e-04, 6.32458833081273e-04, 6.22698707474001e-04, 6.14779833995271e-04, 6.08448105538173e-04, 6.03470010688056e-04, 5.99631731674322e-04, 5.96738657947071e-04, 5.94615219057459e-04, 5.93104917015808e-04, 5.92070422655411e-04, 5.91393597616729e-04, 5.90975311880015e-04, 5.90734944182714e-04, 5.90609481051274e-04, 5.90552162861968e-04, 5.90530661966682e-04, 5.90524813304683e-04, 5.90523950052102e-04, 5.90523880873661e-04, 5.90522605506294e-04, 5.90513964738193e-04, 5.90482201261011e-04, 5.90397538402954e-04, 5.90212290546788e-04, 5.89857652045112e-04, 5.89241349658209e-04, 5.88246359768618e-04, 5.86730904359456e-04, 5.84529945285794e-04, 5.81458380636946e-04, 5.77316104013903e-04, 5.71895002110523e-04, 5.64987836327191e-04, 5.56398777035838e-04, 5.45955144190202e-04, 5.33519679507322e-04, 5.19002462610274e-04, 5.02371441598370e-04, 4.83660523357798e-04, 4.62974306643105e-04, 4.40488846123799e-04, 4.16448243271361e-04, 3.91157227838335e-04, 3.64970019944622e-04, 3.38275566761581e-04, 3.11478941625865e-04, 2.84978877871822e-04, 2.59142733085582e-04, 2.34282688992948e-04, 2.10639500565435e-04, 1.88380233552721e-04, 1.67612080943513e-04, 1.48406241365251e-04, 1.30818585518000e-04, 1.14893376164105e-04, 1.00645394482513e-04, 8.80307931652169e-05, 7.69280414090190e-05, 6.71475914089115e-05, 5.84717783027600e-05, 5.07064320345307e-05, 4.37179455730075e-05, 3.74399578930706e-05, 3.18525523539392e-05, 2.69489272142209e-05, 2.27051641985554e-05, 1.90613621668945e-05, 1.59045418683945e-05, 1.29747013346559e-05, 1.30141100385117e-05, 1.33199533109973e-05, 1.33934349406672e-05, 1.33715801950984e-05, 1.33060215574687e-05, 1.32198830868901e-05, 1.31257643545268e-05, 1.30323665388232e-05, 1.29460105995349e-05, 1.28706085433805e-05, 1.28078131638176e-05, 1.27575791708523e-05, 1.27188255305555e-05, 1.26899653736245e-05, 1.26692529534012e-05, 1.26549872732085e-05, 1.26456225899574e-05, 1.26398236659497e-05, 1.26364875683294e-05, 1.26347451754327e-05, 1.26339489412430e-05, 1.26336502290769e-05, 1.26335689571988e-05, 1.26335569774483e-05, 1.26335559826884e-05, 1.26335373948905e-05, 1.26334115293503e-05, 1.26329487528665e-05, 1.26317146253996e-05, 1.26290111950311e-05, 1.26238238209453e-05, 1.26147719896608e-05, 1.26000597475695e-05, 1.25774206907181e-05, 1.25440535754251e-05, 1.24965491323792e-05, 1.24308164117792e-05, 1.23420155517806e-05, 1.22244695717200e-05, 1.20714480829065e-05, 1.18746669954817e-05, 1.16235701170808e-05, 1.13051496128946e-05, 1.09057628208120e-05, 1.04156972753344e-05, 9.83422309275284e-06, 9.17005618816802e-06, 8.43557290863828e-06, 7.64373329786849e-06, 6.81791993883962e-06, 6.00074132250938e-06, 5.23194928803150e-06, 4.51882792162905e-06, 3.86321231360398e-06, 3.29027854206743e-06, 2.79904963917943e-06, 2.35404384341109e-06, 1.97314898287004e-06, 1.70876332106218e-06, 1.50583156255379e-06, 1.29119739256112e-06, 1.02156482739701e-06, 7.41841162721357e-07, 8.33455388260518e-07, 9.34052096113192e-07, 7.58325242600222e-07, 8.11394556876621e-07, 9.94331401216186e-07, 7.19769616353005e-07, 1.00500128041858e-06, 1.90019243007100e-06, 2.46150601412864e-06, 2.60012072317295e-06, 2.33767681834252e-06, 1.78152076557581e-06, 1.07301157470374e-06, 1.42108547152020e-14, 5.38430072083429e-07, 6.67018767330774e-07, 6.23149858824945e-07, 5.24315240113538e-07, 4.37346116655135e-07, 3.88175081411646e-07, 1.42108547152020e-14, 1.42108547152020e-14, 2.84217094304040e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 2.84217094304040e-14],
[ 1.01000000000000e+02, 1.00999999873216e+02, 1.00999995952792e+02, 1.00999969391299e+02, 1.00999871747740e+02, 1.00999611460482e+02, 1.00999041800879e+02, 1.00997950729976e+02, 1.00996053021218e+02, 1.00992984940895e+02, 1.00988301697828e+02, 1.00981477788427e+02, 1.00971910276996e+02, 1.00958924969037e+02, 1.00941785361913e+02, 1.00919704196586e+02, 1.00891857388322e+02, 1.00857400082706e+02, 1.00815484562287e+02, 1.00765279711290e+02, 1.00705991721554e+02, 1.00636885681667e+02, 1.00557307625220e+02, 1.00466706520367e+02, 1.00364655567140e+02, 1.00250872046262e+02, 1.00125234857194e+02, 9.99877988224282e+01, 9.98388048471343e+01, 9.96786851277873e+01, 9.95080628059863e+01, 9.93277457525452e+01, 9.91387145133698e+01, 9.89421048117395e+01, 9.87391853359721e+01, 9.85313318061373e+01, 9.83199984797683e+01, 9.81066883126449e+01, 9.78929229425823e+01, 9.76802135322569e+01, 9.74700333208055e+01, 9.72637925249194e+01, 9.70628160264475e+01, 9.68683241056109e+01, 9.66814163380602e+01, 9.65030586727802e+01, 9.63340736421907e+01, 9.61751336176739e+01, 9.60267570037414e+01, 9.58893072533268e+01, 9.57629945782846e+01, 9.56478802184137e+01, 9.55438831167364e+01, 9.54507888276313e+01, 9.53682604582612e+01, 9.52958514138192e+01, 9.52330196850651e+01, 9.51791433842378e+01, 9.51335372045542e+01, 9.50954694509761e+01, 9.50641792675862e+01, 9.50388936715589e+01, 9.50188439970216e+01, 9.50032813556100e+01, 9.49914907353954e+01, 9.49828033868998e+01, 9.49766071842954e+01, 9.49723547011956e+01, 9.49695688025291e+01, 9.49678456250584e+01, 9.49668548967166e+01, 9.49663376262297e+01, 9.49661012762751e+01, 9.49660126124713e+01, 9.49659884936936e+01, 9.49659849338099e+01, 9.49659844668283e+01, 9.49659735867162e+01, 9.49658998723441e+01, 9.49656288903358e+01, 9.49649065504186e+01, 9.49633257202025e+01, 9.49602982144846e+01, 9.49550332729152e+01, 9.49465235025843e+01, 9.49335391028335e+01, 9.49146310049830e+01, 9.48881433447418e+01, 9.48522354344929e+01, 9.48049131122649e+01, 9.47440690128468e+01, 9.46675309379428e+01, 9.45731171070918e+01, 9.44586966684003e+01, 9.43222534666781e+01, 9.41619507444046e+01, 9.39761942330062e+01, 9.37636910234742e+01, 9.35235017211673e+01, 9.32550836994204e+01, 9.29583237411265e+01, 9.26335589230235e+01, 9.22815851472573e+01, 9.19036531531516e+01, 9.15014520961797e+01, 9.10770809107691e+01, 9.06330078447361e+01, 9.01720189893208e+01, 8.96971574729587e+01, 8.92116561394650e+01, 8.87188675867594e+01, 8.82221958420204e+01, 8.77250333253037e+01, 8.72307052337936e+01, 8.67424216758360e+01, 8.62632365319704e+01, 8.57960115385968e+01, 8.53433844080379e+01, 8.49077405163234e+01, 8.44911883754778e+01, 8.40955395046035e+01, 8.37222933796641e+01, 8.33726279677191e+01, 8.30473960691680e+01, 8.27471274089057e+01, 8.24720361949621e+01, 8.22220337196460e+01, 8.19967455045830e+01, 8.17955324661307e+01, 8.16175155785630e+01, 8.14616035202363e+01, 8.13265227899395e+01, 8.12108497701669e+01, 8.11130441896046e+01, 8.10314834009659e+01, 8.09644968473226e+01, 8.09104000467131e+01, 8.08675273882422e+01, 8.08342630103125e+01, 8.08090690295784e+01, 8.07905104129943e+01, 8.07772758384874e+01, 8.07681939737335e+01, 8.07622447163078e+01, 8.07585650787863e+01, 8.07564495637005e+01, 8.07553450483824e+01, 8.07548403803217e+01, 8.07546510609563e+01, 8.07545995614086e+01, 8.07545919601800e+01, 8.07545912891362e+01, 8.07545781403634e+01, 8.07544890554816e+01, 8.07541615698623e+01, 8.07532886121323e+01, 8.07513781588666e+01, 8.07477193869228e+01, 8.07413566723528e+01, 8.07310726206641e+01, 8.07153811228423e+01, 8.06925312108759e+01, 8.06605222296489e+01, 8.06171305428271e+01, 8.05599476442820e+01, 8.04864291529610e+01, 8.03939537330731e+01, 8.02798905156000e+01, 8.01416731223123e+01, 7.99768779384726e+01, 7.97833038803481e+01, 7.95590505971539e+01, 7.93025918725137e+01, 7.90128409821662e+01, 7.86892049485048e+01, 7.83316250227078e+01, 7.79406013196526e+01, 7.75172003067160e+01, 7.70630447567074e+01, 7.65802867417382e+01, 7.60715651695189e+01, 7.55399501374800e+01, 7.49888769081993e+01, 7.44220725383128e+01, 7.38434781310212e+01, 7.32571694027495e+01, 7.26672778733194e+01, 7.20779146231094e+01, 7.14930982834930e+01, 7.09166887416349e+01, 7.03523278873538e+01, 6.98033885233981e+01, 6.92729322426257e+01, 6.87636766492343e+01, 6.82779718308483e+01, 6.78177855670365e+01, 6.73846964653470e+01, 6.69798940803187e+01, 6.66041850772775e+01, 6.62580046031099e+01, 6.59414321661166e+01, 6.56542114635771e+01, 6.53957737041705e+01, 6.51652640442935e+01, 6.49615707937913e+01, 6.47833570528332e+01, 6.46290944231976e+01, 6.44970983988196e+01, 6.43855649862546e+01, 6.42926080400741e+01, 6.42162967263904e+01, 6.41546924561641e+01, 6.41058845662215e+01, 6.40680239780982e+01, 6.40393540408751e+01, 6.40182377709888e+01, 6.40031807446623e+01, 6.39928489796577e+01, 6.39860812620972e+01, 6.39818955276027e+01, 6.39794890875557e+01, 6.39782326920495e+01, 6.39776586306996e+01, 6.39774432798429e+01, 6.39773846991293e+01, 6.39773760527366e+01, 6.39773753591890e+01, 6.39773625597171e+01, 6.39772758414087e+01, 6.39769570556888e+01, 6.39761072896555e+01, 6.39742475957190e+01, 6.39706860526155e+01, 6.39644924781760e+01, 6.39544819585610e+01, 6.39392081777443e+01, 6.39169673199314e+01, 6.38858130694845e+01, 6.38435829406349e+01, 6.37879358273561e+01, 6.37164002709399e+01, 6.36264325041084e+01, 6.35154828592760e+01, 6.33810686475371e+01, 6.32208511559260e+01, 6.30327140124962e+01, 6.28148398743335e+01, 6.25657822433208e+01, 6.22845292404411e+01, 6.19705563888341e+01, 6.16238658661843e+01, 6.12450102646279e+01, 6.08350995995618e+01, 6.03957910854504e+01, 5.99292619943056e+01, 5.94381666859092e+01, 5.89255796134361e+01, 5.83949267356222e+01, 5.78499082769232e+01, 5.72944161322680e+01, 5.67324493692825e+01, 5.61680312009029e+01, 5.56051304720221e+01, 5.50475901510249e+01, 5.44990646085191e+01, 5.39629666991039e+01, 5.34424249424945e+01, 5.29402505141401e+01, 5.24589133518938e+01, 5.20005264705673e+01, 5.15668375205822e+01, 5.11592266802572e+01, 5.07787100802713e+01, 5.04259480805624e+01, 5.01012578279533e+01, 4.98046296068215e+01, 4.95357465559087e+01, 4.92940073675035e+01, 4.90785516161533e+01, 4.88882873853184e+01, 4.87219208713748e+01, 4.85779876425349e+01, 4.84548852128406e+01, 4.83509065570354e+01, 4.82642741418967e+01, 4.81931739872606e+01, 4.81357892017769e+01, 4.80903323724813e+01, 4.80550761326208e+01, 4.80283811978382e+01, 4.80087211547673e+01, 4.79947033143283e+01, 4.79850850078977e+01, 4.79787848082586e+01, 4.79748882957886e+01, 4.79726481576036e+01, 4.79714785947566e+01, 4.79709442098155e+01, 4.79707437432488e+01, 4.79706892114512e+01, 4.79706811626710e+01, 4.79706805418717e+01, 4.79706693943076e+01, 4.79705938679231e+01, 4.79703162249801e+01, 4.79695761311053e+01, 4.79679564551544e+01, 4.79648545872795e+01, 4.79594604297166e+01, 4.79507420818718e+01, 4.79374400865751e+01, 4.79180709231838e+01, 4.78909402190568e+01, 4.78541658970145e+01, 4.78057111775691e+01, 4.77434270094862e+01, 4.76651031148778e+01, 4.75685264171835e+01, 4.74515451919388e+01, 4.73121368689675e+01, 4.71484770545551e+01, 4.69590070701171e+01, 4.67424971551397e+01, 4.64981024851740e+01, 4.62254093274139e+01, 4.59244689992153e+01, 4.55958177960520e+01, 4.52404816890769e+01, 4.48599653238674e+01, 4.44562256407468e+01, 4.40316312380688e+01, 4.35889093600872e+01, 4.31310830460112e+01, 4.26614014521435e+01, 4.21832665832048e+01, 4.17001595985120e+01, 4.12155695041320e+01, 4.07329264789756e+01, 4.02555414341820e+01, 3.97865528001070e+01, 3.93288810629744e+01, 3.88851912578777e+01, 3.84578634320068e+01, 3.80489709595616e+01, 3.76602664626936e+01, 3.72931749447993e+01, 3.69487935791175e+01, 3.66278974426877e+01, 3.63309503732502e+01, 3.60581200756704e+01, 3.58092966210655e+01, 3.55841135576108e+01, 3.53819709686302e+01, 3.52020599484722e+01, 3.50433880982007e+01, 3.49048057539196e+01, 3.47850327391434e+01, 3.46826854736376e+01, 3.45963042745984e+01, 3.45243806562503e+01, 3.44653843784332e+01, 3.44177899231135e+01, 3.43801020007073e+01, 3.43508796165901e+01, 3.43287581725157e+01, 3.43124690467816e+01, 3.43008560976302e+01, 3.42928885705482e+01, 3.42876699626906e+01, 3.42844425042109e+01, 3.42825870513951e+01, 3.42816183422569e+01, 3.42811757320664e+01, 3.42810096938676e+01, 3.42809645274673e+01, 3.42809578610018e+01, 3.42809573562371e+01, 3.42809484143822e+01, 3.42808878320171e+01, 3.42806651250950e+01, 3.42800714722091e+01, 3.42787722875457e+01, 3.42762842324189e+01, 3.42719576055262e+01, 3.42649649535173e+01, 3.42542966281231e+01, 3.42387638786671e+01, 3.42170099022699e+01, 3.41875290711444e+01, 3.41486943111633e+01, 3.40987923158393e+01, 3.40360659475773e+01, 3.39587628127721e+01, 3.38651886155845e+01, 3.37537635208829e+01, 3.36230794197426e+01, 3.34719557246816e+01, 3.32994911606490e+01, 3.31051089925518e+01, 3.28885932644448e+01, 3.26501139321042e+01, 3.23902392483124e+01, 3.21099343914539e+01, 3.18105460785004e+01, 3.14937737217975e+01, 3.11616285091630e+01, 3.08163825325369e+01, 3.04605106835252e+01, 3.00966284053269e+01, 2.97274284920129e+01, 2.93556199433761e+01, 2.89838714418834e+01, 2.86147613816126e+01, 2.82507356399004e+01, 2.78940735443848e+01, 2.75468618456823e+01, 2.72109760268868e+01, 2.68880679960720e+01, 2.65795591098338e+01, 2.62866375274810e+01, 2.60102590432829e+01, 2.57511507329279e+01, 2.55098169347971e+01, 2.52865472380346e+01, 2.50814262561721e+01, 2.48943450293375e+01, 2.47250139297434e+01, 2.45729769561615e+01, 2.44376273032976e+01, 2.43182240871504e+01, 2.42139100992768e+01, 2.41237304502618e+01, 2.40466519433502e+01, 2.39815829912899e+01, 2.39273938525477e+01, 2.38829369186901e+01, 2.38470667362483e+01, 2.38186593986604e+01, 2.37966309026721e+01, 2.37799540348114e+01, 2.37676733426674e+01, 2.37589177569402e+01, 2.37529104661340e+01, 2.37489757068132e+01, 2.37465422168041e+01, 2.37451432027735e+01, 2.37444127916807e+01, 2.37440790608907e+01, 2.37439538669611e+01, 2.37439198111778e+01, 2.37439147846156e+01, 2.37439144072700e+01, 2.37439077655596e+01, 2.37438627670014e+01, 2.37436973476505e+01, 2.37432564012322e+01, 2.37422914050294e+01, 2.37404433381351e+01, 2.37372295852284e+01, 2.37320354428262e+01, 2.37241107559394e+01, 2.37125721077159e+01, 2.36964108582837e+01, 2.36745071766487e+01, 2.36456500272404e+01, 2.36085628591308e+01, 2.35619345035843e+01, 2.35044545220834e+01, 2.34348519757064e+01, 2.33519363266307e+01, 2.32546389567423e+01, 2.31420536217336e+01, 2.30134740748853e+01, 2.28684271106349e+01, 2.27066994027694e+01, 2.25283567433965e+01, 2.23337546137734e+01, 2.21235394156212e+01, 2.18986401373164e+01, 2.16602506999859e+01, 2.14098037038516e+01, 2.11489367557482e+01, 2.08794529793318e+01, 2.06032776519210e+01, 2.03224131238774e+01, 2.00388942022542e+01, 1.97547459837163e+01, 1.94719457106588e+01, 1.91923896644817e+01, 1.89178655115942e+01, 1.86500300009853e+01, 1.83903915646213e+01, 1.81402972213225e+01, 1.79009231993423e+01, 1.76732688047337e+01, 1.74581531994543e+01, 1.72562148625731e+01, 1.70679135668287e+01, 1.68935347134677e+01, 1.67331958478047e+01, 1.65868551468351e+01, 1.64543216448511e+01, 1.63352669521683e+01, 1.62292382271422e+01, 1.61356721786338e+01, 1.60539098982053e+01, 1.59832123414229e+01, 1.59227762898107e+01, 1.58717506255626e+01, 1.58292527389863e+01, 1.57943848651501e+01, 1.57662501147220e+01, 1.57439679293565e+01, 1.57266886597814e+01, 1.57136069406508e+01, 1.57039735253478e+01, 1.56971052502050e+01, 1.56923928234520e+01, 1.56893061801652e+01, 1.56873972092721e+01, 1.56862997391928e+01, 1.56857267605079e+01, 1.56854649617234e+01, 1.56853667519464e+01, 1.56853400365044e+01, 1.56853360933611e+01, 1.56853357982482e+01, 1.56853306159182e+01, 1.56852955048741e+01, 1.56851664330288e+01, 1.56848223753897e+01, 1.56840694162596e+01, 1.56826274194053e+01, 1.56801198059552e+01, 1.56760669197558e+01, 1.56698833899572e+01, 1.56608798182546e+01, 1.56482690201571e+01, 1.56311769324739e+01, 1.56086581602463e+01, 1.55797159748170e+01, 1.55433263925298e+01, 1.54984657661317e+01, 1.54441411175703e+01, 1.53794222444650e+01, 1.53034744587739e+01, 1.52155906820605e+01, 1.51152215434840e+01, 1.50020021174002e+01, 1.48757740054592e+01, 1.47366016152037e+01, 1.45847817086246e+01, 1.44208455792552e+01, 1.42455535499539e+01, 1.40598818474035e+01, 1.38650022835157e+01, 1.36622555369653e+01, 1.34531191572226e+01, 1.32391716849468e+01, 1.30220544731995e+01, 1.28034328837980e+01, 1.25849585097519e+01, 1.23682339367324e+01, 1.21547813157998e+01, 1.19460157008604e+01, 1.17432237416778e+01, 1.15475479548117e+01, 1.13599764565027e+01, 1.11813377616410e+01, 1.10123000491068e+01, 1.08533741718587e+01, 1.07049196457096e+01, 1.05671528717607e+01, 1.04401569176865e+01, 1.03238922850252e+01, 1.02182082070685e+01, 1.01228541412288e+01, 1.00374912305011e+01, 9.96170360369284e+00, 9.89500945916633e+00, 9.83687192996435e+00, 9.78670975920369e+00, 9.74390782472994e+00, 9.70782754350296e+00, 9.67781716214861e+00, 9.65322190439036e+00, 9.63339390303825e+00, 9.61770179862371e+00, 9.60553984352327e+00, 9.59633631425996e+00, 9.58956100984346e+00, 9.58473160386843e+00, 9.58141862478381e+00, 9.57924886297010e+00, 9.57790704414424e+00, 9.57713566378199e+00, 9.57673294306255e+00, 9.57654893877400e+00, 9.57647991275815e+00, 9.57646113604076e+00, 9.57645836463780e+00, 9.57645815714275e+00, 9.57645451238839e+00, 9.57642981865888e+00, 9.57633904215764e+00, 9.57609706733616e+00, 9.57556752151822e+00, 9.57455341686152e+00, 9.57279000220186e+00, 9.56994018247984e+00, 9.56559282207435e+00, 9.55926419718480e+00, 9.55040279716226e+00, 9.53839759091045e+00, 9.52258976839586e+00, 9.50228783694089e+00, 9.47678579815772e+00, 9.44538395863512e+00, 9.40741174453908e+00, 9.36225171005100e+00, 9.30936376850116e+00, 9.24830855177346e+00, 9.17876873676943e+00, 9.10056718368179e+00, 9.01368082058990e+00, 8.91824938578333e+00, 8.81457839706425e+00, 8.70313603976825e+00, 8.58454402684731e+00, 8.45956285288015e+00, 8.32907220434649e+00, 8.19404756798573e+00, 8.05553427137318e+00, 7.91462027936178e+00, 7.77240905371556e+00, 7.62999367048594e+00, 7.48843320001481e+00, 7.34873211386441e+00, 7.21182321952497e+00, 7.07855436355676e+00, 6.94967890815509e+00, 6.82584978982030e+00, 6.70761682004955e+00, 6.59542678893710e+00, 6.48962588076436e+00, 6.39046390023994e+00, 6.29809983120768e+00, 6.21260829770445e+00, 6.13398656160194e+00, 6.06216176377281e+00, 5.99699818998458e+00, 5.93830441308691e+00, 5.88584022546490e+00, 5.83932332744139e+00, 5.79843577672969e+00, 5.76283023052666e+00, 5.73213602552796e+00, 5.70596514280511e+00, 5.68391809539952e+00, 5.66558975842081e+00, 5.65057513655890e+00, 5.63847503477479e+00, 5.62890156733522e+00, 5.62148341126726e+00, 5.61587068564805e+00, 5.61173932058109e+00, 5.60879477139883e+00, 5.60677493600511e+00, 5.60545214685429e+00, 5.60463413338300e+00, 5.60416388431801e+00, 5.60391837989020e+00, 5.60380620870775e+00, 5.60376412971881e+00, 5.60375268324131e+00, 5.60375099376620e+00, 5.60375086693958e+00, 5.60374863468485e+00, 5.60373351084479e+00, 5.60367791422132e+00, 5.60352971566769e+00, 5.60320539504040e+00, 5.60258431397763e+00, 5.60150434811102e+00, 5.59975910278403e+00, 5.59709691289881e+00, 5.59322179772580e+00, 5.58779650226319e+00, 5.58044770471545e+00, 5.57077340205906e+00, 5.55835240091793e+00, 5.54275573956400e+00, 5.52355975196639e+00, 5.50036036270235e+00, 5.47278808157046e+00, 5.44052306080854e+00, 5.40330949923995e+00, 5.36096863955321e+00, 5.31340961813565e+00, 5.26063749801698e+00, 5.20275794511076e+00, 5.13997818962521e+00, 5.07260413481764e+00, 5.00103371498214e+00, 4.92574684097458e+00, 4.84729248152547e+00, 4.76627359168754e+00, 4.68333070131217e+00, 4.59912500939591e+00, 4.51432179577230e+00, 4.42957486887659e+00, 4.34551263188396e+00, 4.26272618725575e+00, 4.18175972979763e+00, 4.10310331680255e+00, 4.02718796300138e+00, 3.95438289563070e+00, 3.88499472428211e+00, 3.81926823096168e+00, 3.75738846492084e+00, 3.69948382966628e+00, 3.64562987074974e+00, 3.59585350706240e+00, 3.55013749040246e+00, 3.50842492367531e+00, 3.47062371360735e+00, 3.43661087645923e+00, 3.40623665280010e+00, 3.37932841850267e+00, 3.35569440286835e+00, 3.33512724082293e+00, 3.31740739448044e+00, 3.30230648046992e+00, 3.28959053397980e+00, 3.27902322952252e+00, 3.27036906323153e+00, 3.26339648359269e+00, 3.25788093856734e+00, 3.25360778887290e+00, 3.25037502152607e+00, 3.24799568624682e+00, 3.24629997130579e+00, 3.24513683576406e+00, 3.24437512216203e+00, 3.24390408730572e+00, 3.24363330803351e+00, 3.24349194236105e+00, 3.24342735247595e+00, 3.24340312278372e+00, 3.24339653173915e+00, 3.24339555891548e+00, 3.24339548557590e+00, 3.24339419059470e+00, 3.24338541691633e+00, 3.24335316409302e+00, 3.24326719103327e+00, 3.24307904680620e+00, 3.24271874998828e+00, 3.24209225928038e+00, 3.24107986759188e+00, 3.23953563903272e+00, 3.23728799169909e+00, 3.23414150676173e+00, 3.22988001402102e+00, 3.22427096394306e+00, 3.21707104518815e+00, 3.20803294503867e+00, 3.19691307988522e+00, 3.18348004797588e+00, 3.16752348286754e+00, 3.14886292095653e+00, 3.12735624850993e+00, 3.10290727099326e+00, 3.07547195695581e+00, 3.04506295427221e+00, 3.01175205829090e+00, 2.97567042511078e+00, 2.93700646017136e+00, 2.89600146046692e+00, 2.85294323384315e+00, 2.80815804674418e+00, 2.76200135004517e+00, 2.71484779235202e+00, 2.66708104713540e+00, 2.61908395503099e+00, 2.57122942091658e+00, 2.52387241585279e+00, 2.47734332767018e+00, 2.43194279246471e+00, 2.38793803319695e+00, 2.34556063959036e+00, 2.30500565144654e+00, 2.26643175819848e+00, 2.22996240106451e+00, 2.19568755833055e+00, 2.16366600530128e+00, 2.13392786375510e+00, 2.10647728663884e+00, 2.08129515800096e+00, 2.05834172230105e+00, 2.03755908867549e+00, 2.01887358280800e+00, 2.00219794085191e+00, 1.98743335609569e+00, 1.97447139989352e+00, 1.96319584420883e+00, 1.95348441449688e+00, 1.94521049920959e+00, 1.93824483660362e+00, 1.93245719146283e+00, 1.92771802453464e+00, 1.92390014670804e+00, 1.92088033905517e+00, 1.91854090968529e+00, 1.91677114977922e+00, 1.91546864498190e+00, 1.91454039519115e+00, 1.91390369614442e+00, 1.91348674024157e+00, 1.91322890161679e+00, 1.91308068112380e+00, 1.91300329989211e+00, 1.91296794450420e+00, 1.91295468159664e+00, 1.91295107377546e+00, 1.91295054126901e+00, 1.91295050090575e+00, 1.91294978530070e+00, 1.91294493697686e+00, 1.91292711411305e+00, 1.91287960557358e+00, 1.91277563770586e+00, 1.91257653986304e+00, 1.91223034791223e+00, 1.91167091938371e+00, 1.91081762707978e+00, 1.90957568895638e+00, 1.90783717971472e+00, 1.90548275250823e+00, 1.90238407637173e+00, 1.89840696577654e+00, 1.89341514324244e+00, 1.88727453537588e+00, 1.87985795944662e+00, 1.87105001526629e+00, 1.86075196029758e+00, 1.84888631973658e+00, 1.83540097275332e+00, 1.82027246513381e+00, 1.80350832938288e+00, 1.78514824552839e+00, 1.76526394614302e+00, 1.74395785143425e+00, 1.72136050649884e+00, 1.69762697382932e+00, 1.67293240108546e+00, 1.64746702993487e+00, 1.62143093220153e+00, 1.59502875387804e+00, 1.56846471843251e+00, 1.54193809375482e+00, 1.51563926924877e+00, 1.48974652863927e+00, 1.46442354690871e+00, 1.43981759160077e+00, 1.41605837257882e+00, 1.39325746105845e+00, 1.37150818740614e+00, 1.35088592562711e+00, 1.33144867792025e+00, 1.31323788242129e+00, 1.29627937898499e+00, 1.28058447989486e+00, 1.26615110370567e+00, 1.25296494053929e+00, 1.24100062594216e+00, 1.23022290793830e+00, 1.22058779826982e+00, 1.21204370405688e+00, 1.20453254019613e+00, 1.19799082565071e+00, 1.19235076825306e+00, 1.18754134264812e+00, 1.18348936454896e+00, 1.18012056165821e+00, 1.17736063765463e+00, 1.17513632091308e+00, 1.17337638456006e+00, 1.17201261958905e+00, 1.17098073859853e+00, 1.17022118476969e+00, 1.16967981938276e+00, 1.16930846175248e+00, 1.16906525805823e+00, 1.16891486006868e+00, 1.16882840096299e+00, 1.16878326291153e+00, 1.16876263928118e+00, 1.16875490269689e+00, 1.16875279816325e+00, 1.16875248753867e+00, 1.16875246384431e+00, 1.16875204179387e+00, 1.16874918234255e+00, 1.16873867073835e+00, 1.16871065097794e+00, 1.16864933202638e+00, 1.16853190506306e+00, 1.16832771787840e+00, 1.16799774889033e+00, 1.16749441885322e+00, 1.16676177273069e+00, 1.16573605679346e+00, 1.16434670615000e+00, 1.16251774507853e+00, 1.16016958641622e+00, 1.15722119698012e+00, 1.15359257418885e+00, 1.14920745596442e+00, 1.14399616346063e+00, 1.13789845649627e+00, 1.13086626729510e+00, 1.12286617165300e+00, 1.11388145982568e+00, 1.10391368318494e+00, 1.09298357674159e+00, 1.08113129040066e+00, 1.06841590057319e+00, 1.05491421505643e+00, 1.04071892424535e+00, 1.02593618746083e+00, 1.01068277200328e+00, 9.95082883009530e-01, 9.79264833797288e-01, 9.63357709255561e-01, 9.47488169391818e-01, 9.31777526803643e-01, 9.16339211048196e-01, 9.01276705344515e-01, 8.86682008137626e-01, 8.72634636099818e-01, 8.59201149396526e-01, 8.46435148230242e-01, 8.34377665245569e-01, 8.23057863705714e-01, 8.12493947184677e-01, 8.02694191911954e-01, 7.93658025567098e-01, 7.85377093282349e-01, 7.77836269916307e-01, 7.71014594894908e-01, 7.64886120471942e-01, 7.59420675359664e-01, 7.54584553209213e-01, 7.50341139712262e-01, 7.46651493701918e-01, 7.43474897159700e-01, 7.40769387049255e-01, 7.38492278863916e-01, 7.36600688078411e-01, 7.35052051641617e-01, 7.33804647503561e-01, 7.32818106184666e-01, 7.32053904807103e-01, 7.31475831054809e-01, 7.31050402431308e-01, 7.30747225128610e-01, 7.30539276933598e-01, 7.30403099927386e-01, 7.30318891226986e-01, 7.30270483528287e-01, 7.30245211490427e-01, 7.30233664748638e-01, 7.30229333209380e-01, 7.30228154929662e-01, 7.30227981018189e-01, 7.30227967658692e-01, 7.30227728467753e-01, 7.30226107916181e-01, 7.30220150630103e-01, 7.30204270945151e-01, 7.30169519924601e-01, 7.30102972473219e-01, 7.29987261259225e-01, 7.29800282406060e-01, 7.29515095753868e-01, 7.29100039575457e-01, 7.28519075514171e-01, 7.27732373715668e-01, 7.26697140235842e-01, 7.25368678573190e-01, 7.23701664635399e-01, 7.21651600035230e-01, 7.19176393203995e-01, 7.16238002766502e-01, 7.12804064682473e-01, 7.08849415788094e-01, 7.04357423484467e-01, 6.99321035941551e-01, 6.93743480101205e-01, 6.87638555690967e-01, 6.81030500887729e-01, 6.73953436488134e-01, 6.66450426833464e-01, 6.58572223351115e-01, 6.50375776831581e-01, 6.41922614997554e-01, 6.33277181718086e-01, 6.24505224367581e-01, 6.15672298854264e-01, 6.06842441157482e-01, 5.98077033297642e-01, 5.89433873324495e-01, 5.80966444829443e-01, 5.72723372148118e-01, 5.64748042332766e-01, 5.57078373110627e-01, 5.49746706230223e-01, 5.42779806844493e-01, 5.36198951216818e-01, 5.30020086718878e-01, 5.24254049725414e-01, 5.18906828634258e-01, 5.13979860922916e-01, 5.09470354936823e-01, 5.05371628972128e-01, 5.01673462095809e-01, 4.98362452930621e-01, 4.95422384198520e-01, 4.92834592051295e-01, 4.90578340032002e-01, 4.88631197851348e-01, 4.86969425015872e-01, 4.85568358730731e-01, 4.84402804486209e-01, 4.83447426416234e-01, 4.82677133017091e-01, 4.82067452279381e-01, 4.81594888874582e-01, 4.81237254906418e-01, 4.80973965027616e-01, 4.80786286552728e-01, 4.80657535636652e-01, 4.80573211663696e-01, 4.80521063669116e-01, 4.80491084805706e-01, 4.80475433443345e-01, 4.80468282276476e-01, 4.80465599638208e-01, 4.80464869896622e-01, 4.80464762188348e-01, 4.80464753851280e-01, 4.80464603761021e-01, 4.80463586878173e-01, 4.80459848720372e-01, 4.80449884250007e-01, 4.80428077685573e-01, 4.80386317173881e-01, 4.80313700512340e-01, 4.80196346690207e-01, 4.80017326320734e-01, 4.79756721927998e-01, 4.79391826337960e-01, 4.78897483913457e-01, 4.78246574893888e-01, 4.77410637569989e-01, 4.76360616481685e-01, 4.75067717490234e-01, 4.73504342854781e-01, 4.71645071945503e-01, 4.69467646711507e-01, 4.66953916327114e-01, 4.64090693363531e-01, 4.60870474999595e-01, 4.57291987526517e-01, 4.53360520653177e-01, 4.49088029393906e-01, 4.44492994735910e-01, 4.39600048673555e-01, 4.34439383252112e-01, 4.29045975716178e-01, 4.23458671626506e-01, 4.17719174127186e-01, 4.11870990026117e-01, 4.05958382010667e-01, 4.00025371546775e-01, 3.94114829495777e-01, 3.88267682142356e-01, 3.82522250162481e-01, 3.76913728060373e-01, 3.71473802619960e-01, 3.66230401587529e-01, 3.61207558501832e-01, 3.56425376420574e-01, 3.51900072121225e-01, 3.47644082859901e-01, 3.43666219526540e-01, 3.39971852577534e-01, 3.36563120024749e-01, 3.33439149661302e-01, 3.30596290343991e-01, 3.28028349365155e-01, 3.25726834647753e-01, 3.23681201663717e-01, 3.21879105631467e-01, 3.20306659741817e-01, 3.18948699960302e-01, 3.17789056425084e-01, 3.16810830679330e-01, 3.15996677018589e-01, 3.15329085172408e-01, 3.14790660460474e-01, 3.14364396544931e-01, 3.14033935034161e-01, 3.13783805559140e-01, 3.13599639617721e-01, 3.13468351524122e-01, 3.13378280246482e-01, 3.13319286769342e-01, 3.13282802856776e-01, 3.13261828652330e-01, 3.13250878349891e-01, 3.13245875094239e-01, 3.13243998204811e-01, 3.13243487645821e-01, 3.13243412288386e-01, 3.13243406408148e-01, 3.13243299936175e-01, 3.13242578573366e-01, 3.13239926773264e-01, 3.13232858093045e-01, 3.13217388688813e-01, 3.13187763808500e-01, 3.13136248644213e-01, 3.13052993825494e-01, 3.12925984555775e-01, 3.12741080825816e-01, 3.12482154164186e-01, 3.12131323881488e-01, 3.11669292622469e-01, 3.11075777210988e-01, 3.10330026283580e-01, 3.09411411209268e-01, 3.08300071566634e-01, 3.06977591406408e-01, 3.05427678194803e-01, 3.03636813286630e-01, 3.01594841575309e-01, 2.99295469052851e-01, 2.96736640611343e-01, 2.93920776476170e-01, 2.90854853786868e-01, 2.87550329343006e-01, 2.84022909474672e-01, 2.80292182339181e-01, 2.76381135696667e-01, 2.72315588599000e-01, 2.68123567998487e-01, 2.63834661015466e-01, 2.59479370851182e-01, 2.55088499752802e-01, 2.50692576836585e-01, 2.46321342759785e-01, 2.42003297887010e-01, 2.37765316198661e-01, 2.33632323982192e-01, 2.29627040339267e-01, 2.25769775544151e-01, 2.22078283001522e-01, 2.18567660646528e-01, 2.15250297862212e-01, 2.12135864213806e-01, 2.09231336476222e-01, 2.06541060562444e-01, 2.04066845068525e-01, 2.01808083246775e-01, 1.99761900328511e-01, 1.97923323246329e-01, 1.96285469969780e-01, 1.94839755833414e-01, 1.93576114373272e-01, 1.92483230239276e-01, 1.91548781680467e-01, 1.90759689887273e-01, 1.90102372123263e-01, 1.89562995121776e-01, 1.89127724700924e-01, 1.88782967033845e-01, 1.88515596566802e-01, 1.88313165278046e-01, 1.88164087884260e-01, 1.88057797774491e-01, 1.87984868916033e-01, 1.87937099739770e-01, 1.87907556046042e-01, 1.87890571236025e-01, 1.87881703599814e-01, 1.87877651895661e-01, 1.87876131960266e-01, 1.87875718500834e-01, 1.87875657475089e-01, 1.87875652672602e-01, 1.87875565196990e-01, 1.87874972536648e-01, 1.87872793856002e-01, 1.87866986305352e-01, 1.87854276683481e-01, 1.87829936469484e-01, 1.87787609360399e-01, 1.87719199451791e-01, 1.87614827032135e-01, 1.87462857579973e-01, 1.87250007894107e-01, 1.86961531290237e-01, 1.86581481403081e-01, 1.86093051333657e-01, 1.85478981699404e-01, 1.84722027694363e-01, 1.83805471703138e-01, 1.82713664610674e-01, 1.81432575983393e-01, 1.79950331114142e-01, 1.78257711819768e-01, 1.76348598090513e-01, 1.74220329332854e-01, 1.71873967006931e-01, 1.69314444742056e-01, 1.66550597240047e-01, 1.63595065034571e-01, 1.60464078046004e-01, 1.57177126428849e-01, 1.53756532107067e-01, 1.50226938387263e-01, 1.46614738049067e-01, 1.42947462336210e-01, 1.39253154334497e-01, 1.35559750264648e-01, 1.31894491101917e-01, 1.28283384512400e-01, 1.24750733554237e-01, 1.21318744223359e-01, 1.18007219306193e-01, 1.14833341483759e-01, 1.11811544335609e-01, 1.08953465915818e-01, 1.06267976025620e-01, 1.03761265686600e-01, 1.01436985914788e-01, 9.92964229186555e-02, 9.73386978926527e-02, 9.55609812360620e-02, 9.39587128796830e-02, 9.25258223078771e-02, 9.12549437100267e-02, 9.01376235449476e-02, 8.91645194068929e-02, 8.83255902945876e-02, 8.76102791874871e-02, 8.70076890247162e-02, 8.65067530977512e-02, 8.60964002701934e-02, 8.57657148280282e-02, 8.55040898404553e-02, 8.53013721171152e-02, 8.51479960029593e-02, 8.50351027176401e-02, 8.49546415665685e-02, 8.48994492925357e-02, 8.48633041959630e-02, 8.48409521669709e-02, 8.48281027600706e-02, 8.48213944425567e-02, 8.48183294077955e-02, 8.48171796155981e-02, 8.48168668450455e-02, 8.48168206823345e-02, 8.48168170165451e-02, 8.48167498490113e-02, 8.48162947809641e-02, 8.48146219111499e-02, 8.48101627214675e-02, 8.48004042142350e-02, 8.47817167426967e-02, 8.47492229566882e-02, 8.46967147805486e-02, 8.46166247555828e-02, 8.45000571862897e-02, 8.43368834682252e-02, 8.41159044002211e-02, 8.38250803279550e-02, 8.34518272923930e-02, 8.29833743624846e-02, 8.24071736218883e-02, 8.17113505430029e-02, 8.08851785215836e-02, 7.99195577964298e-02, 7.88074761390760e-02, 7.75444267830494e-02, 7.61287588680375e-02, 7.45619370037787e-02, 7.28486900193701e-02, 7.09970342546171e-02, 6.90181636568354e-02, 6.69262069148147e-02, 6.47378600844917e-02, 6.24719111828398e-02, 6.01486802631683e-02, 5.77894045200935e-02, 5.54156019891536e-02, 5.30484489093368e-02, 5.07082039675879e-02, 4.84137068735180e-02, 4.61819708278313e-02, 4.40278795360172e-02, 4.19639923926951e-02, 4.00004555718425e-02, 3.81315752126028e-02, 3.63912952571610e-02, 3.47676067654561e-02, 3.32619067236664e-02, 3.18738938113938e-02, 3.06018177078755e-02, 2.94427267579920e-02, 2.83927030746014e-02, 2.74470773487963e-02, 2.66006186579233e-02, 2.58476972103807e-02, 2.51824201396147e-02, 2.45987421179592e-02, 2.40905537150411e-02, 2.36517511206529e-02, 2.32762911460895e-02, 2.29582353794058e-02, 2.26917870589176e-02, 2.24713237125940e-02, 2.22914279400612e-02, 2.21469179482625e-02, 2.20328786436568e-02, 2.19446932902514e-02, 2.18780750151808e-02, 2.18290968373935e-02, 2.17942184489317e-02, 2.17703077316997e-02, 2.17546549560126e-02, 2.17449777837402e-02, 2.17394155645536e-02, 2.17365119311950e-02, 2.17351853162171e-02, 2.17346876699338e-02, 2.17345522993071e-02, 2.17345323188880e-02, 2.17345307181063e-02, 2.17345011984236e-02, 2.17343011989470e-02, 2.17335659904165e-02, 2.17316062763331e-02, 2.17273179060697e-02, 2.17191067579989e-02, 2.17048324903288e-02, 2.16817748673890e-02, 2.16466263467936e-02, 2.15955144589230e-02, 2.15240571187962e-02, 2.14274532534375e-02, 2.13006098968265e-02, 2.11383051328653e-02, 2.09353839387063e-02, 2.06869811889021e-02, 2.03887630010939e-02, 2.00371745330443e-02, 1.96296796600421e-02, 1.91649760994324e-02, 1.86431689311064e-02, 1.80658864321259e-02, 1.74363248820129e-02, 1.67592134646526e-02, 1.60406962940698e-02, 1.52881353696017e-02, 1.45098451704513e-02, 1.37147757799596e-02, 1.29121660777258e-02, 1.21111910173397e-02, 1.13206269877253e-02, 1.05485567578114e-02, 9.80213091821724e-03, 9.08739672874503e-03, 8.40919869496304e-03, 7.77114887658570e-03, 7.17565959459421e-03, 6.62402734445777e-03, 6.11655455242044e-03, 5.65269528267720e-03, 5.23121185779019e-03, 4.85033120905598e-03, 4.50789220472814e-03, 4.20147781981939e-03, 3.92852849894609e-03, 3.68643520879175e-03, 3.47261237179453e-03, 3.28455208367018e-03, 3.11986178529366e-03, 2.97628794965021e-03, 2.85172844913006e-03, 2.74423617071307e-03, 2.65201622089251e-03, 2.57341877099293e-03, 2.50692927531304e-03, 2.45115747948432e-03, 2.40482633681438e-03, 2.36676167479288e-03, 2.33588320138267e-03, 2.31119720990591e-03, 2.29179113051714e-03, 2.27682988698064e-03, 2.26555384902838e-03, 2.25727803214681e-03, 2.25139209132408e-03, 2.24736059020358e-03, 2.24472300814682e-03, 2.24309297368563e-03, 2.24215628271907e-03, 2.24166736656173e-03, 2.24144400434056e-03, 2.24136021793697e-03, 2.24133742642607e-03, 2.24133406247518e-03, 2.24133379040836e-03, 2.24132874190275e-03, 2.24129453764021e-03, 2.24116880285408e-03, 2.24083366833838e-03, 2.24010037773326e-03, 2.23869658684350e-03, 2.23625709777764e-03, 2.23231883619235e-03, 2.22632093833397e-03, 2.21761083746230e-03, 2.20545718470078e-03, 2.18907026372506e-03, 2.16763022156385e-03, 2.14032291475519e-03, 2.10638246129236e-03, 2.06513873017400e-03, 2.01606707553026e-03, 1.95883675425534e-03, 1.89335381240596e-03, 1.81979395429721e-03, 1.73862116155021e-03, 1.65058868952883e-03, 1.55672051923261e-03, 1.45827325576420e-03, 1.35668059681393e-03, 1.25348453285312e-03, 1.15025904860033e-03, 1.04853299400531e-03, 9.49718815917322e-04, 8.55052983737892e-04, 7.65552358203081e-04, 6.81988728516103e-04, 6.04881615998920e-04, 5.34771356129227e-04, 4.71264134816352e-04, 4.14332287263619e-04, 3.63663447799922e-04, 3.18978683807813e-04, 2.80083497798955e-04, 2.46037876109995e-04, 2.16352908353201e-04, 1.90587743219339e-04, 1.68216011971367e-04, 1.48666962758170e-04, 1.32040586237281e-04, 1.17477754372430e-04, 1.04577595095634e-04, 9.30369600808945e-05, 8.26282205395046e-05, 7.31885640021801e-05, 6.45993233092668e-05, 5.69265496892513e-05, 5.00413749335848e-05, 4.36746678360578e-05, 3.78206838718143e-05, 3.24754439873232e-05, 2.76316749818761e-05, 2.32734327312301e-05, 1.93780557822265e-05, 1.59185720178812e-05, 1.28725525883056e-05, 1.02214869031059e-05, 7.94695201022932e-06, 6.02837549479318e-06, 4.44207087468840e-06, 3.16106506659253e-06, 2.15541255954577e-06, 1.39278547806043e-06, 8.39198420976572e-07, 4.59787926843092e-07, 2.84217094304040e-14, 4.26325641456060e-14, 2.84217094304040e-14, 2.84217094304040e-14],
[ 1.02000000000000e+02, 1.01999999871472e+02, 1.01999995897120e+02, 1.01999968970257e+02, 1.01999869983547e+02, 1.01999606115860e+02, 1.01999028620160e+02, 1.01997922540602e+02, 1.01995998726683e+02, 1.01992888440114e+02, 1.01988140768693e+02, 1.01981222974932e+02, 1.01971523819784e+02, 1.01958359818666e+02, 1.01940984312787e+02, 1.01918599177622e+02, 1.01890368944312e+02, 1.01855437078242e+02, 1.01812944138074e+02, 1.01762047520595e+02, 1.01701942471883e+02, 1.01631884002824e+02, 1.01551209278736e+02, 1.01459359956004e+02, 1.01355903819139e+02, 1.01240554945259e+02, 1.01113191514081e+02, 1.00973870319666e+02, 1.00822837053547e+02, 1.00660531537554e+02, 1.00487587294010e+02, 1.00304825138448e+02, 1.00113240835665e+02, 9.99139872313782e+01, 9.97083516123160e+01, 9.94977293153400e+01, 9.92835947721852e+01, 9.90674712291907e+01, 9.88509003277395e+01, 9.86354125926838e+01, 9.84224996833501e+01, 9.82135890473680e+01, 9.80100214101774e+01, 9.78130313532327e+01, 9.76237310927210e+01, 9.74430974701968e+01, 9.72719621023392e+01, 9.71110046005403e+01, 9.69607487524339e+01, 9.68215615477573e+01, 9.66936549230453e+01, 9.65770900889532e+01, 9.64717842881069e+01, 9.63775198096965e+01, 9.62939550601896e+01, 9.62206374588998e+01, 9.61570178943962e+01, 9.61024664447212e+01, 9.60562890329510e+01, 9.60177446616500e+01, 9.59860628470716e+01, 9.59604608583492e+01, 9.59401603601103e+01, 9.59244030604463e+01, 9.59124649811998e+01, 9.59036689948600e+01, 9.58973953122002e+01, 9.58930896567092e+01, 9.58902689247143e+01, 9.58885242020326e+01, 9.58875210865817e+01, 9.58869973486892e+01, 9.58867580436736e+01, 9.58866682713164e+01, 9.58866438509846e+01, 9.58866402465921e+01, 9.58866397737718e+01, 9.58866287576270e+01, 9.58865541216136e+01, 9.58862797515572e+01, 9.58855483803711e+01, 9.58839477855640e+01, 9.58808824287476e+01, 9.58755516652299e+01, 9.58669355129528e+01, 9.58537888089710e+01, 9.58346443948649e+01, 9.58078257548776e+01, 9.57714692766770e+01, 9.57235560102055e+01, 9.56619524639139e+01, 9.55844596036454e+01, 9.54888688184660e+01, 9.53730232094630e+01, 9.52348821714123e+01, 9.50725869121683e+01, 9.48845243362496e+01, 9.46693866528995e+01, 9.44262241890157e+01, 9.41544892017538e+01, 9.38540689624284e+01, 9.35253069460384e+01, 9.31690115013458e+01, 9.27864517917750e+01, 9.23793410428359e+01, 9.19498072703696e+01, 9.15003518746900e+01, 9.10337969947284e+01, 9.05532234474243e+01, 9.00619022885974e+01, 8.95632240694005e+01, 8.90606301519710e+01, 8.85575496616662e+01, 8.80573440017069e+01, 8.75632590221115e+01, 8.70783836604904e+01, 8.66056135222720e+01, 8.61476182999562e+01, 8.57068127019141e+01, 8.52853312387321e+01, 8.48850075684821e+01, 8.45073591161097e+01, 8.41535774669009e+01, 8.38245247278295e+01, 8.35207357593731e+01, 8.32424259598713e+01, 8.29895041480070e+01, 8.27615900238178e+01, 8.25580356713401e+01, 8.23779505726305e+01, 8.22202296144262e+01, 8.20835835726081e+01, 8.19665715497301e+01, 8.18676348160700e+01, 8.17851314675985e+01, 8.17173712701198e+01, 8.16626500142650e+01, 8.16192826683782e+01, 8.15856345929038e+01, 8.15601500773109e+01, 8.15413774842205e+01, 8.15279903388370e+01, 8.15188037864478e+01, 8.15127859556820e+01, 8.15090639070384e+01, 8.15069240093463e+01, 8.15058067639877e+01, 8.15052962794385e+01, 8.15051047781028e+01, 8.15050526850059e+01, 8.15050449961707e+01, 8.15050443173929e+01, 8.15050310170762e+01, 8.15049409054625e+01, 8.15046096454735e+01, 8.15037266266956e+01, 8.15017941552575e+01, 8.14980932165875e+01, 8.14916571759425e+01, 8.14812546157697e+01, 8.14653823166579e+01, 8.14422691656277e+01, 8.14098915157935e+01, 8.13660002186917e+01, 8.13081592002216e+01, 8.12337950524007e+01, 8.11402566711731e+01, 8.10248834980285e+01, 8.08850804413388e+01, 8.07183970913376e+01, 8.05226084363245e+01, 8.02957939764337e+01, 8.00364119545945e+01, 7.97433654173898e+01, 7.94160570083309e+01, 7.90544297969885e+01, 7.86589920567858e+01, 7.82308246972892e+01, 7.77715709820455e+01, 7.72834091406383e+01, 7.67690094109531e+01, 7.62314778143754e+01, 7.56742894797972e+01, 7.51012145439869e+01, 7.45162395827647e+01, 7.39234872502036e+01, 7.33271364407608e+01, 7.27313449549113e+01, 7.21401764049699e+01, 7.15575329332997e+01, 7.09870951585678e+01, 7.04322705301165e+01, 6.98961509064461e+01, 6.93814797007454e+01, 6.88906284348290e+01, 6.84255821139316e+01, 6.79879325538043e+01, 6.75788786808345e+01, 6.71992328574225e+01, 6.68494324029508e+01, 6.65295556284658e+01, 6.62393418394403e+01, 6.59782148646053e+01, 6.57453097345789e+01, 6.55395021652358e+01, 6.53594405034807e+01, 6.52035797728379e+01, 6.50702174172637e+01, 6.49575302874187e+01, 6.48636123481670e+01, 6.47865125142994e+01, 6.47242719497612e+01, 6.46749601015762e+01, 6.46367086914224e+01, 6.46077428635114e+01, 6.45864086941078e+01, 6.45711963108874e+01, 6.45607579520084e+01, 6.45539204148582e+01, 6.45496914994120e+01, 6.45472602345039e+01, 6.45459908781536e+01, 6.45454108948801e+01, 6.45451933225044e+01, 6.45451341374837e+01, 6.45451254018965e+01, 6.45451247011944e+01, 6.45451117696857e+01, 6.45450241568088e+01, 6.45447020825640e+01, 6.45438435505584e+01, 6.45419646726468e+01, 6.45383663904764e+01, 6.45321089281829e+01, 6.45219951533567e+01, 6.45065638407597e+01, 6.44840936201442e+01, 6.44526181389704e+01, 6.44099526755557e+01, 6.43537320927337e+01, 6.42814596250986e+01, 6.41905655494916e+01, 6.40784743120472e+01, 6.39426781988381e+01, 6.37808151730430e+01, 6.35907480990208e+01, 6.33706422754877e+01, 6.31190380469304e+01, 6.28349152865977e+01, 6.25177467630966e+01, 6.21675378133985e+01, 6.17848503248332e+01, 6.13708097369246e+01, 6.09270945602561e+01, 6.04559087227302e+01, 5.99599378487911e+01, 5.94422913177707e+01, 5.89064326005984e+01, 5.83561009032140e+01, 5.77952275059958e+01, 5.72278503340896e+01, 5.66580301873124e+01, 5.60897716928829e+01, 5.55269514542447e+01, 5.49732551316387e+01, 5.44321244092749e+01, 5.39067140868766e+01, 5.33998589636274e+01, 5.29140498030929e+01, 5.24514174772149e+01, 5.20137243498586e+01, 5.16023620219079e+01, 5.12183546670725e+01, 5.08623672999503e+01, 5.05347184130091e+01, 5.02353964906098e+01, 4.99640799595554e+01, 4.97201601737859e+01, 4.95027670610025e+01, 4.93107970826442e+01, 4.91429431737179e+01, 4.89977263315904e+01, 4.88735285092993e+01, 4.87686264373589e+01, 4.86812259493372e+01, 4.86094963245586e+01, 4.85516040926557e+01, 4.85057456777739e+01, 4.84701782043725e+01, 4.84432477510493e+01, 4.84234143318069e+01, 4.84092729118553e+01, 4.83995698308316e+01, 4.83932141104712e+01, 4.83892832632173e+01, 4.83870233868372e+01, 4.83858435191338e+01, 4.83853044258781e+01, 4.83851021930697e+01, 4.83850471808125e+01, 4.83850390611175e+01, 4.83850384348485e+01, 4.83850271890674e+01, 4.83849509972487e+01, 4.83846709081079e+01, 4.83839242936340e+01, 4.83822903478481e+01, 4.83791611528216e+01, 4.83737194771508e+01, 4.83649243382632e+01, 4.83515052041829e+01, 4.83319655279007e+01, 4.83045960916757e+01, 4.82674983825440e+01, 4.82186179187630e+01, 4.81557870983295e+01, 4.80767767493474e+01, 4.79793551395828e+01, 4.78613527690258e+01, 4.77207308526702e+01, 4.75556510352254e+01, 4.73645436022786e+01, 4.71461712991995e+01, 4.68996858689312e+01, 4.66246745906929e+01, 4.63211944472671e+01, 4.59897920578292e+01, 4.56315081619275e+01, 4.52478661938588e+01, 4.48408453029888e+01, 4.44128390040337e+01, 4.39666014204493e+01, 4.35051837400766e+01, 4.30318639549898e+01, 4.25500731386382e+01, 4.20633213926250e+01, 4.15751262031488e+01, 4.10889453759314e+01, 4.06081160966960e+01, 4.01358011105759e+01, 3.96749425963667e+01, 3.92282240330787e+01, 3.87980401704914e+01, 3.83864750610382e+01, 3.79952879430760e+01, 3.76259065736194e+01, 3.72794274095803e+01, 3.69566218628956e+01, 3.66579477374881e+01, 3.63835649127005e+01, 3.61333543697822e+01, 3.59069397518236e+01, 3.57037107809992e+01, 3.55228480052320e+01, 3.53633484868957e+01, 3.52240521617280e+01, 3.51036686760663e+01, 3.50008045504473e+01, 3.49139905185541e+01, 3.48417088575690e+01, 3.47824204672919e+01, 3.47345913808351e+01, 3.46967183102136e+01, 3.46673527566830e+01, 3.46451231585729e+01, 3.46287545174724e+01, 3.46170849436728e+01, 3.46090785977240e+01, 3.46038345776409e+01, 3.46005914082765e+01, 3.45987269251889e+01, 3.45977535019939e+01, 3.45973087380394e+01, 3.45971418919100e+01, 3.45970965057349e+01, 3.45970898068312e+01, 3.45970892996104e+01, 3.45970803142455e+01, 3.45970194370949e+01, 3.45967956465237e+01, 3.45961991051320e+01, 3.45948935995762e+01, 3.45923934412223e+01, 3.45880457730903e+01, 3.45810191299436e+01, 3.45702989829039e+01, 3.45546908622159e+01, 3.45328314845853e+01, 3.45032081073398e+01, 3.44641860844706e+01, 3.44140443067076e+01, 3.43510178717540e+01, 3.42733469609385e+01, 3.41793305116853e+01, 3.40673828956489e+01, 3.39360914706472e+01, 3.37842726051741e+01, 3.36110236124466e+01, 3.34157680085848e+01, 3.31982916511329e+01, 3.29587676328415e+01, 3.26977682998614e+01, 3.24162634145171e+01, 3.21156042535312e+01, 3.17974942678923e+01, 3.14639477601090e+01, 3.11172387783220e+01, 3.07598430040748e+01, 3.03943757524622e+01, 3.00235292657665e+01, 2.96500122565462e+01, 2.92764941770783e+01, 2.89055560311814e+01, 2.85396488000741e+01, 2.81810598315878e+01, 2.78318869336029e+01, 2.74940194791281e+01, 2.71691255936634e+01, 2.68586444389540e+01, 2.65637826866005e+01, 2.62855144329590e+01, 2.60245839882968e+01, 2.57815111367385e+01, 2.55565985862570e+01, 2.53499414047316e+01, 2.51614382764218e+01, 2.49908044263924e+01, 2.48375860615629e+01, 2.47011761751529e+01, 2.45808315600901e+01, 2.44756908756768e+01, 2.43847936071301e+01, 2.43070997456431e+01, 2.42415099945939e+01, 2.41868862747386e+01, 2.41420722592662e+01, 2.41059136221899e+01, 2.40772776359621e+01, 2.40550717125970e+01, 2.40382604532200e+01, 2.40258807595029e+01, 2.40170545712932e+01, 2.40109988305244e+01, 2.40070323327060e+01, 2.40045792121789e+01, 2.40031689120287e+01, 2.40024326084134e+01, 2.40020961852469e+01, 2.40019699813084e+01, 2.40019356507777e+01, 2.40019305836633e+01, 2.40019302032734e+01, 2.40019235079805e+01, 2.40018781463928e+01, 2.40017113925045e+01, 2.40012668886797e+01, 2.40002941070809e+01, 2.39984311298434e+01, 2.39951914465443e+01, 2.39899553905870e+01, 2.39819667462824e+01, 2.39703349521848e+01, 2.39540431989365e+01, 2.39319625665774e+01, 2.39028721631010e+01, 2.38654850115886e+01, 2.38184791900400e+01, 2.37605334635701e+01, 2.36903663761589e+01, 2.36067775072237e+01, 2.35086893696540e+01, 2.33951882550893e+01, 2.32655622420982e+01, 2.31193345914231e+01, 2.29562908692749e+01, 2.27764983645910e+01, 2.25803166894054e+01, 2.23683988560189e+01, 2.21416825897361e+01, 2.19013721396081e+01, 2.16489113684792e+01, 2.13859494086065e+01, 2.11143006194677e+01, 2.08359009247416e+01, 2.05527627726626e+01, 2.02669309060499e+01, 1.99804408308591e+01, 1.96952813778276e+01, 1.94133621587480e+01, 1.91364861523740e+01, 1.88663272240000e+01, 1.86044121435228e+01, 1.83521066105677e+01, 1.81106048641271e+01, 1.78809225725803e+01, 1.76638928045880e+01, 1.74601649375286e+01, 1.72702063628260e+01, 1.70943068119396e+01, 1.69325850760338e+01, 1.67849978473354e+01, 1.66513503841658e+01, 1.65313086989803e+01, 1.64244129865571e+01, 1.63300920403779e+01, 1.62476784402317e+01, 1.61764243247359e+01, 1.61155175824402e+01, 1.60640983008295e+01, 1.60212753030810e+01, 1.59861425796836e+01, 1.59577953898562e+01, 1.59353457714224e+01, 1.59179371634337e+01, 1.59047578194107e+01, 1.58950526759942e+01, 1.58881333462264e+01, 1.58833859311596e+01, 1.58802763886428e+01, 1.58783532626504e+01, 1.58772476572776e+01, 1.58766704319571e+01, 1.58764066930144e+01, 1.58763077554436e+01, 1.58762808420266e+01, 1.58762768696627e+01, 1.58762765723629e+01, 1.58762713516293e+01, 1.58762359803967e+01, 1.58761059520872e+01, 1.58757593449929e+01, 1.58750008070866e+01, 1.58735481286784e+01, 1.58710219476518e+01, 1.58669390721051e+01, 1.58607098353759e+01, 1.58516398160614e+01, 1.58389361573393e+01, 1.58217186020698e+01, 1.57990352200824e+01, 1.57698826408215e+01, 1.57332304200116e+01, 1.56880489686092e+01, 1.56333402652832e+01, 1.55681703729318e+01, 1.54917026010619e+01, 1.54032300163372e+01, 1.53022059198133e+01, 1.51882708951233e+01, 1.50612750963274e+01, 1.49212945904237e+01, 1.47686407942721e+01, 1.46038623394139e+01, 1.44277390463115e+01, 1.42412680731396e+01, 1.40456427014575e+01, 1.38422246071995e+01, 1.36325108136222e+01, 1.34180968054796e+01, 1.32006374747163e+01, 1.29818076467047e+01, 1.27632638914511e+01, 1.25466091583250e+01, 1.23333615022337e+01, 1.21249278232289e+01, 1.19225831582175e+01, 1.17274556830801e+01, 1.15405172423299e+01, 1.13625789486516e+01, 1.11942912015473e+01, 1.10361473661273e+01, 1.08884903229097e+01, 1.07515211334140e+01, 1.06253091469398e+01, 1.05098029833000e+01, 1.04048419482875e+01, 1.03101675601135e+01, 1.02254349762026e+01, 1.01502242040320e+01, 1.00840510533953e+01, 1.00263778389952e+01, 9.97662387170823e+00, 9.93417578557180e+00, 9.89839773784853e+00, 9.86864149452919e+00, 9.84425637713254e+00, 9.82459900296038e+00, 9.80904270477716e+00, 9.79698647205576e+00, 9.78786321918751e+00, 9.78114716061107e+00, 9.77636006204600e+00, 9.77307614307416e+00, 9.77092542992375e+00, 9.76959539761702e+00, 9.76883079529529e+00, 9.76843161389304e+00, 9.76824922687769e+00, 9.76818080757785e+00, 9.76816219590394e+00, 9.76815944886111e+00, 9.76815924318991e+00, 9.76815563047235e+00, 9.76813115379751e+00, 9.76804117522443e+00, 9.76780132747982e+00, 9.76727643719093e+00, 9.76627125024315e+00, 9.76452334922826e+00, 9.76169861890703e+00, 9.75738957511621e+00, 9.75111682225203e+00, 9.74233384000566e+00, 9.73043521681805e+00, 9.71476834170072e+00, 9.69464843557219e+00, 9.66937664890273e+00, 9.63826077881892e+00, 9.60063797476928e+00, 9.55589862056107e+00, 9.50351041882534e+00, 9.44304158070587e+00, 9.37418195780438e+00, 9.29676096154994e+00, 9.21076120817312e+00, 9.11632700833998e+00, 9.01376708245137e+00, 8.90355120855952e+00, 8.78630087347992e+00, 8.66277436617985e+00, 8.53384709069640e+00, 8.40048815096254e+00, 8.26373444647940e+00, 8.12466360110014e+00, 7.98436702524636e+00, 7.84392429483533e+00, 7.70437983799877e+00, 7.56672267923493e+00, 7.43186972758862e+00, 7.30065283636215e+00, 7.17380962749999e+00, 7.05197787873415e+00, 6.93569312398267e+00, 6.82538901972113e+00, 6.72139997977581e+00, 6.62396557250366e+00, 6.53323619976807e+00, 6.44927962764227e+00, 6.37208800547390e+00, 6.30158508462812e+00, 6.23763342388970e+00, 6.18004143961005e+00, 6.12857022120347e+00, 6.08294008390691e+00, 6.04283686938963e+00, 6.00791803038108e+00, 5.97781854826594e+00, 5.95215673346846e+00, 5.93053994877712e+00, 5.91257027733876e+00, 5.89785013204063e+00, 5.88598777390058e+00, 5.87660267665167e+00, 5.86933064582684e+00, 5.86382857617717e+00, 5.85977871378985e+00, 5.85689228092650e+00, 5.85491232378082e+00, 5.85361565657045e+00, 5.85281379917531e+00, 5.85235283848753e+00, 5.85211218349860e+00, 5.85200222808054e+00, 5.85196098030384e+00, 5.85194975993650e+00, 5.85194810383479e+00, 5.85194797951345e+00, 5.85194579135403e+00, 5.85193096626622e+00, 5.85187646788856e+00, 5.85173119686173e+00, 5.85141328308875e+00, 5.85080447204699e+00, 5.84974584425110e+00, 5.84803508799178e+00, 5.84542552266487e+00, 5.84162704095063e+00, 5.83630909968687e+00, 5.82910583820139e+00, 5.81962333631550e+00, 5.80744894063463e+00, 5.79216248757194e+00, 5.77334913795495e+00, 5.75061341728491e+00, 5.72359393707065e+00, 5.69197816800674e+00, 5.65551655839441e+00, 5.61403525416168e+00, 5.56744669093779e+00, 5.51575740036517e+00, 5.45907250264239e+00, 5.39759653858586e+00, 5.33163051367669e+00, 5.26156526426819e+00, 5.18787148958162e+00, 5.11108699918675e+00, 5.03180188417633e+00, 4.95064241688740e+00, 4.86825451226085e+00, 4.78528754563061e+00, 4.70237922632867e+00, 4.62014208901076e+00, 4.53915200295859e+00, 4.45993893179097e+00, 4.38298001770256e+00, 4.30869492738010e+00, 4.23744328844764e+00, 4.16952396867471e+00, 4.10517590467092e+00, 4.04458016922889e+00, 3.98786297213053e+00, 3.93509931265676e+00, 3.88631703774686e+00, 3.84150110264953e+00, 3.80059787657191e+00, 3.76351938065990e+00, 3.73014738685544e+00, 3.70033734181455e+00, 3.67392210887104e+00, 3.65071554235804e+00, 3.63051592229554e+00, 3.61310928376034e+00, 3.59827267473300e+00, 3.58577736968027e+00, 3.57539205462906e+00, 3.56688598427213e+00, 3.56003209415427e+00, 3.55461003279790e+00, 3.55040906140296e+00, 3.54723075413170e+00, 3.54489142147605e+00, 3.54322417402423e+00, 3.54208054492291e+00, 3.54133159677851e+00, 3.54086845239894e+00, 3.54060220782537e+00, 3.54046320922686e+00, 3.54039970077518e+00, 3.54037587674942e+00, 3.54036939605444e+00, 3.54036843951808e+00, 3.54036836740636e+00, 3.54036709410612e+00, 3.54035846731863e+00, 3.54032675447129e+00, 3.54024222070839e+00, 3.54005722593696e+00, 3.53970295913098e+00, 3.53908694974218e+00, 3.53809148542834e+00, 3.53657305245819e+00, 3.53436289858841e+00, 3.53126879600050e+00, 3.52707805198223e+00, 3.52156177568035e+00, 3.51448035945000e+00, 3.50559007333571e+00, 3.49465060296873e+00, 3.48143328848880e+00, 3.46572975075576e+00, 3.44736052842913e+00, 3.42618330365018e+00, 3.40210027306183e+00, 3.37506423128328e+00, 3.34508297951929e+00, 3.31222175271868e+00, 3.27660347022999e+00, 3.23840674858671e+00, 3.19786175882930e+00, 3.15524415063857e+00, 3.11086738750619e+00, 3.06507392914438e+00, 3.01822575091808e+00, 2.97069470158164e+00, 2.92285317137390e+00, 2.87506547851173e+00, 2.82768029272276e+00, 2.78102431099198e+00, 2.73539729473598e+00, 2.69106847946406e+00, 2.64827428554574e+00, 2.60721719685228e+00, 2.56806563453497e+00, 2.53095463505431e+00, 2.49598714172654e+00, 2.46323573323406e+00, 2.43274463607064e+00, 2.40453189634594e+00, 2.37859161606457e+00, 2.35489618716956e+00, 2.33339848150290e+00, 2.31403397541952e+00, 2.29672280376496e+00, 2.28137174936478e+00, 2.26787618139751e+00, 2.25612195943825e+00, 2.24598732001323e+00, 2.23734475963229e+00, 2.23006292292227e+00, 2.22400849716027e+00, 2.21904810576613e+00, 2.21505018380810e+00, 2.21188680904577e+00, 2.20943545327620e+00, 2.20758061157700e+00, 2.20621526220673e+00, 2.20524210805458e+00, 2.20457455203835e+00, 2.20413736388116e+00, 2.20386700409493e+00, 2.20371158230179e+00, 2.20363044050822e+00, 2.20359336669806e+00, 2.20357945912021e+00, 2.20357567593014e+00, 2.20357511753950e+00, 2.20357507521427e+00, 2.20357432482486e+00, 2.20356924083082e+00, 2.20355055159965e+00, 2.20350073349940e+00, 2.20339171048606e+00, 2.20318292864383e+00, 2.20281988740416e+00, 2.20223320250261e+00, 2.20133826772519e+00, 2.20003557233713e+00, 2.19821171772142e+00, 2.19574115906043e+00, 2.19248867491409e+00, 2.18831253876145e+00, 2.18306833217434e+00, 2.17661330042883e+00, 2.16881111029481e+00, 2.15953682982214e+00, 2.14868191541711e+00, 2.13615896706266e+00, 2.12190600268551e+00, 2.10589001094236e+00, 2.08810956987462e+00, 2.06859636639942e+00, 2.04741551528063e+00, 2.02466465042012e+00, 2.00047183862235e+00, 1.97499243830498e+00, 1.94840508542080e+00, 1.92090703040971e+00, 1.89270907030851e+00, 1.86403031930369e+00, 1.83509304195181e+00, 1.80611774100190e+00, 1.77731865214815e+00, 1.74889975676017e+00, 1.72105138508383e+00, 1.69394744920558e+00, 1.66774331801631e+00, 1.64257432487069e+00, 1.61855488122364e+00, 1.59577815481151e+00, 1.57431625803130e+00, 1.55422088094360e+00, 1.53552429438885e+00, 1.51824064305875e+00, 1.50236744695869e+00, 1.48788723304688e+00, 1.47476922679459e+00, 1.46297104515297e+00, 1.45244034658091e+00, 1.44311640878013e+00, 1.43493161901340e+00, 1.42781287402198e+00, 1.42168289567786e+00, 1.41646147411741e+00, 1.41206665214304e+00, 1.40841586343853e+00, 1.40542703317973e+00, 1.40301964368369e+00, 1.40111576067486e+00, 1.39964100844470e+00, 1.39852547549924e+00, 1.39770452698816e+00, 1.39711949690424e+00, 1.39671823214020e+00, 1.39645546216620e+00, 1.39629297226837e+00, 1.39619956465373e+00, 1.39615079976490e+00, 1.39612851921054e+00, 1.39612016109054e+00, 1.39611788748724e+00, 1.39611755190845e+00, 1.39611752631066e+00, 1.39611707035455e+00, 1.39611398118913e+00, 1.39610262515373e+00, 1.39607235462051e+00, 1.39600611077408e+00, 1.39587925536831e+00, 1.39565868243848e+00, 1.39530225766138e+00, 1.39475862903495e+00, 1.39396744717446e+00, 1.39286002601364e+00, 1.39136046423351e+00, 1.38938723361062e+00, 1.38685522226615e+00, 1.38367819854076e+00, 1.37977163555203e+00, 1.37505580870621e+00, 1.36945905049586e+00, 1.36292102133035e+00, 1.35539583477399e+00, 1.34685486330532e+00, 1.33728904915910e+00, 1.32671055592134e+00, 1.31515362132808e+00, 1.30267451003084e+00, 1.28935051554198e+00, 1.27527802045645e+00, 1.26056968936236e+00, 1.24535093434282e+00, 1.22975585228271e+00, 1.21392287929952e+00, 1.19799043362318e+00, 1.18209281867217e+00, 1.16635663036131e+00, 1.15089785854057e+00, 1.13581979840926e+00, 1.12121180431804e+00, 1.10714883808916e+00, 1.09369169851880e+00, 1.08088777621981e+00, 1.06877216143749e+00, 1.05736893976270e+00, 1.04669253576746e+00, 1.03674899971206e+00, 1.02753717009408e+00, 1.01904967910869e+00, 1.01127379554073e+00, 1.00419211890249e+00, 9.97783150111289e-01, 9.92021769004964e-01, 9.86879649233558e-01, 9.82325638204841e-01, 9.78326125173609e-01, 9.74845415247060e-01, 9.71846121672570e-01, 9.69289583638812e-01, 9.67136312117830e-01, 9.65346462050462e-01, 9.63880325430878e-01, 9.62698836572656e-01, 9.61764078060682e-01, 9.61039773662350e-01, 9.60491752879150e-01, 9.60088370965678e-01, 9.59800868233600e-01, 9.59603653364624e-01, 9.59474497305820e-01, 9.59394627067084e-01, 9.59348712265694e-01, 9.59324741371177e-01, 9.59313789047187e-01, 9.59309680480757e-01, 9.59308562853749e-01, 9.59308397894446e-01, 9.59308385222613e-01, 9.59308158344129e-01, 9.59306621210501e-01, 9.59300970568265e-01, 9.59285908211527e-01, 9.59252945529911e-01, 9.59189821566682e-01, 9.59080059459140e-01, 9.58902684200396e-01, 9.58632122306696e-01, 9.58238300910836e-01, 9.57686959644803e-01, 9.56940183017934e-01, 9.55957153477272e-01, 9.54695115764508e-01, 9.53110531642324e-01, 9.51160391026789e-01, 9.48803631969056e-01, 9.46002609165362e-01, 9.42724540475453e-01, 9.38942855151694e-01, 9.34638367763142e-01, 9.29800209181124e-01, 9.24426460552752e-01, 9.18524456823974e-01, 9.12110750814648e-01, 9.05210753967111e-01, 8.97858092258303e-01, 8.90093732376599e-01, 8.81964942183643e-01, 8.73524150285277e-01, 8.64827763323049e-01, 8.55934988655449e-01, 8.46906697218841e-01, 8.37804349141123e-01, 8.28688994958057e-01, 8.19620358844162e-01, 8.10656006943915e-01, 8.01850602841128e-01, 7.93255252340060e-01, 7.84916940073842e-01, 7.76878060344870e-01, 7.69176043784469e-01, 7.61843079969935e-01, 7.54905934336961e-01, 7.48385855912204e-01, 7.42298570860058e-01, 7.36654355763946e-01, 7.31458184005433e-01, 7.26709938505309e-01, 7.22404684329305e-01, 7.18532995075293e-01, 7.15081327392483e-01, 7.12032438296603e-01, 7.09365840040860e-01, 7.07058287119693e-01, 7.05084289505621e-01, 7.03416645472402e-01, 7.02026986398835e-01, 7.00886324863276e-01, 6.99965596239451e-01, 6.99236183014236e-01, 6.98670410292905e-01, 6.98242000563212e-01, 6.97926475859575e-01, 6.97701496077428e-01, 6.97547123381966e-01, 6.97446004419419e-01, 6.97383464332319e-01, 6.97347509298538e-01, 6.97328737329229e-01, 6.97320160215412e-01, 6.97316942636476e-01, 6.97316067376264e-01, 6.97315938189661e-01, 6.97315928190122e-01, 6.97315748170070e-01, 6.97314528507576e-01, 6.97310044898558e-01, 6.97298093237251e-01, 6.97271937289372e-01, 6.97221845407270e-01, 6.97134734850167e-01, 6.96993940078414e-01, 6.96779118837753e-01, 6.96466306872225e-01, 6.96028128598329e-01, 6.95434166940601e-01, 6.94651490687136e-01, 6.93645332153892e-01, 6.92379901719249e-01, 6.90819319099020e-01, 6.88928634403950e-01, 6.86674905519382e-01, 6.84028292733643e-01, 6.80963127436669e-01, 6.77458909698331e-01, 6.73501190090420e-01, 6.69082294508428e-01, 6.64201856978735e-01, 6.58867134208435e-01, 6.53093086369239e-01, 6.46902220494084e-01, 6.40324204960834e-01, 6.33395274865904e-01, 6.26157457757344e-01, 6.18657656486093e-01, 6.10946630385413e-01, 6.03077917439975e-01, 5.95106738704303e-01, 5.87088922401051e-01, 5.79079879504050e-01, 5.71133655935512e-01, 5.63302079529052e-01, 5.55634013282386e-01, 5.48174720627401e-01, 5.40965343760774e-01, 5.34042492587290e-01, 5.27437939455567e-01, 5.21178413436857e-01, 5.15285487188422e-01, 5.09775549237787e-01, 5.04659854636060e-01, 4.99944647215699e-01, 4.95631347049553e-01, 4.91716797073778e-01, 4.88193563154596e-01, 4.85050282106187e-01, 4.82272052263625e-01, 4.79840861148247e-01, 4.77736044506713e-01, 4.75934770550737e-01, 4.74412542582812e-01, 4.73143712400741e-01, 4.72101995993270e-01, 4.71260982157190e-01, 4.70594623882789e-01, 4.70077701784604e-01, 4.69686248601462e-01, 4.69397923949060e-01, 4.69192329147120e-01, 4.69051253094420e-01, 4.68958841821433e-01, 4.68901686465657e-01, 4.68868826899609e-01, 4.68851670980229e-01, 4.68843832238548e-01, 4.68840891646138e-01, 4.68840091732773e-01, 4.68839973667215e-01, 4.68839964454361e-01, 4.68839797640342e-01, 4.68838667450655e-01, 4.68834512751343e-01, 4.68823437829542e-01, 4.68799200538456e-01, 4.68752782860065e-01, 4.68672060981149e-01, 4.68541589884680e-01, 4.68342515032688e-01, 4.68052620905752e-01, 4.67646522897240e-01, 4.67096005261084e-01, 4.66370503423917e-01, 4.65437723993944e-01, 4.64264390293707e-01, 4.62817095384239e-01, 4.61063238603941e-01, 4.58972016002235e-01, 4.56515430163407e-01, 4.53669281297067e-01, 4.50414099597040e-01, 4.46735979128519e-01, 4.42627276119126e-01, 4.38087139505683e-01, 4.33121848681651e-01, 4.27744942123311e-01, 4.21977130284972e-01, 4.15845996093267e-01, 4.09385495792392e-01, 4.02635281167150e-01, 3.95639870840284e-01, 3.88447703171934e-01, 3.81110106229073e-01, 3.73680221452645e-01, 3.66211917206471e-01, 3.58758726556815e-01, 3.51372840597193e-01, 3.44104184576295e-01, 3.36999599181311e-01, 3.30102143781858e-01, 3.23450532499351e-01, 3.17078707945896e-01, 3.11015551721027e-01, 3.05284725609255e-01, 2.99904633192256e-01, 2.94888488495303e-01, 2.90244476431283e-01, 2.85975989171774e-01, 2.82081923030177e-01, 2.78557021764584e-01, 2.75392254120064e-01, 2.72575215633253e-01, 2.70090546928714e-01, 2.67920362703748e-01, 2.66044687134467e-01, 2.64441892430710e-01, 2.63089137668885e-01, 2.61962804874599e-01, 2.61038928691221e-01, 2.60293614999966e-01, 2.59703442715207e-01, 2.59245841856426e-01, 2.58899440072796e-01, 2.58644369234460e-01, 2.58462523625411e-01, 2.58337761759492e-01, 2.58256044909871e-01, 2.58205507063408e-01, 2.58176453095956e-01, 2.58161284386645e-01, 2.58154353696369e-01, 2.58151753758142e-01, 2.58151046512182e-01, 2.58150942124175e-01, 2.58150933909292e-01, 2.58150784277206e-01, 2.58149770498193e-01, 2.58146043745640e-01, 2.58136109640247e-01, 2.58114369340731e-01, 2.58072735013499e-01, 2.58000335534589e-01, 2.57883326697581e-01, 2.57704818350163e-01, 2.57444928660007e-01, 2.57080972988916e-01, 2.56587791502739e-01, 2.55938215556625e-01, 2.55103668023537e-01, 2.54054887124397e-01, 2.52762757118816e-01, 2.51199222707612e-01, 2.49338257572987e-01, 2.47156851630251e-01, 2.44635976837785e-01, 2.41761488365327e-01, 2.38524917054363e-01, 2.34924110806176e-01, 2.30963687008426e-01, 2.26655265355234e-01, 2.22017460175650e-01, 2.17075623173983e-01, 2.11861340584733e-01, 2.06411702271268e-01, 2.00768373244344e-01, 1.94976509417302e-01, 1.89083568181672e-01, 1.83138069795689e-01, 1.77188367109125e-01, 1.71281478645923e-01, 1.65462033768605e-01, 1.59771369183161e-01, 1.54246804353846e-01, 1.48921110629743e-01, 1.43822176234679e-01, 1.38972857828463e-01, 1.34390999977271e-01, 1.30089597114832e-01, 1.26077068651867e-01, 1.22357616679324e-01, 1.18931636868567e-01, 1.15796156167676e-01, 1.12945275146700e-01, 1.10370597765325e-01, 1.08061636409033e-01, 1.06006184841945e-01, 1.04190655950433e-01, 1.02600384602638e-01, 1.01219898523581e-01, 1.00033161759853e-01, 9.90237961185301e-02, 9.81752859929372e-02, 9.74711713483600e-02, 9.68952324750489e-02, 9.64316685778641e-02, 9.60652705391709e-02, 9.57815864302631e-02, 9.55670767348664e-02, 9.54092549345802e-02, 9.52968082209225e-02, 9.52196927264595e-02, 9.51691978496854e-02, 9.51379749720758e-02, 9.51200270662014e-02, 9.51106572651241e-02, 9.51063762679212e-02, 9.51047703411092e-02, 9.51043334922588e-02, 9.51042690143566e-02, 9.51042638953083e-02, 9.51041700822941e-02, 9.51035344867392e-02, 9.51011979900613e-02, 9.50949698997618e-02, 9.50813406481501e-02, 9.50552419577519e-02, 9.50098653697324e-02, 9.49365495862935e-02, 9.48247464964567e-02, 9.46620748572258e-02, 9.44344691683341e-02, 9.41264290707693e-02, 9.37213714138920e-02, 9.32020828383031e-02, 9.25512653150236e-02, 9.17521607199774e-02, 9.07892335662694e-02, 8.96488840067775e-02, 8.83201568716032e-02, 8.67954076366289e-02, 8.50708836600518e-02, 8.31471795049339e-02, 8.10295291852015e-02, 7.87279058862111e-02, 7.62569108412521e-02, 7.36354468504459e-02, 7.08861872355584e-02, 6.80348663418187e-02, 6.51094313940064e-02, 6.21391060412293e-02, 5.91534220534697e-02, 5.61812766541010e-02, 5.32500688163964e-02, 5.03849591104739e-02, 4.76082855148121e-02, 4.49391535331358e-02, 4.23932046355588e-02, 3.99825540020089e-02, 3.77158779769502e-02, 3.55986242787338e-02, 3.36333140740237e-02, 3.18199043178424e-02, 3.01561807241058e-02, 2.86381556356048e-02, 2.72604501208775e-02, 2.60166451120512e-02, 2.48995917299197e-02, 2.39016757047352e-02, 2.30150347519924e-02, 2.22317308110718e-02, 2.15438812158482e-02, 2.09437542360291e-02, 2.04238351290513e-02, 1.99768690174693e-02, 1.95958866782604e-02, 1.92742188121855e-02, 1.90055036348440e-02, 1.87836917689594e-02, 1.86030514666982e-02, 1.84581761999709e-02, 1.83439956679621e-02, 1.82557903246021e-02, 1.81892086756505e-02, 1.81402858793831e-02, 1.81054616495544e-02, 1.80815951425899e-02, 1.80659744307263e-02, 1.80563183180453e-02, 1.80507686311557e-02, 1.80478716640522e-02, 1.80465481230812e-02, 1.80460516344851e-02, 1.80459165792115e-02, 1.80458966453898e-02, 1.80458950483171e-02, 1.80458655974292e-02, 1.80456660643102e-02, 1.80449325736163e-02, 1.80429774650861e-02, 1.80386993082351e-02, 1.80305082344603e-02, 1.80162704847788e-02, 1.79932761993200e-02, 1.79582345723475e-02, 1.79073004401715e-02, 1.78361360089241e-02, 1.77400106027363e-02, 1.76139398700315e-02, 1.74528637345759e-02, 1.72518595089173e-02, 1.70063831179732e-02, 1.67125275765301e-02, 1.63672841355407e-02, 1.59687883959378e-02, 1.55165317754751e-02, 1.50115185627300e-02, 1.44563508084303e-02, 1.38552276337549e-02, 1.32138519731768e-02, 1.25392457375142e-02, 1.18394829692647e-02, 1.11233586497406e-02, 1.04000172686654e-02, 9.67856912791376e-03, 8.96772305218008e-03, 8.27546163723980e-03, 7.60877981441865e-03, 6.97350020610088e-03, 6.37417059212453e-03, 5.81404095126459e-03, 5.29511095024304e-03, 4.81823408748028e-03, 4.38326222666774e-03, 3.98921388399742e-03, 3.63445098875559e-03, 3.31685138796445e-03, 3.03396754426899e-03, 2.78316517001542e-03, 2.56173856034536e-03, 2.36700188256123e-03, 2.19635753953185e-03, 2.04734395282458e-03, 1.91766578532793e-03, 1.80520985938415e-03, 1.70804994748153e-03, 1.62444333435019e-03, 1.55282166076631e-03, 1.49177813372603e-03, 1.44005277311976e-03, 1.39651699124954e-03, 1.36015847883186e-03, 1.33006710218808e-03, 1.30542229172192e-03, 1.28548221734093e-03, 1.26957488841128e-03, 1.25709118130146e-03, 1.24747967920769e-03, 1.24024310832738e-03, 1.23493606972147e-03, 1.23116370377829e-03, 1.22858088587918e-03, 1.22689154288480e-03, 1.22584770062417e-03, 1.22524792331546e-03, 1.22493488073872e-03, 1.22479187075442e-03, 1.22473822628422e-03, 1.22472363401016e-03, 1.22472148022723e-03, 1.22472130605900e-03, 1.22471807374325e-03, 1.22469617451770e-03, 1.22461567355003e-03, 1.22440110952482e-03, 1.22393165301560e-03, 1.22303301529314e-03, 1.22147161415853e-03, 1.21895156215146e-03, 1.21511509662753e-03, 1.20954709834109e-03, 1.20178431006934e-03, 1.19132973621561e-03, 1.17767244492484e-03, 1.16031258714600e-03, 1.13879089379054e-03, 1.11272124733830e-03, 1.08182421874403e-03, 1.04595882271497e-03, 1.00514930720408e-03, 9.59603692166411e-04, 9.09721117295703e-04, 8.56085892706915e-04, 7.99447428192934e-04, 7.40686799289847e-04, 6.80772357426918e-04, 6.20708231679146e-04, 5.61480531459324e-04, 5.04006357942899e-04, 4.49090296598342e-04, 3.97391966799887e-04, 3.49406666074970e-04, 3.05459458971313e-04, 2.65711500901489e-04, 2.30176196609477e-04, 1.98742197952129e-04, 1.71200283475059e-04, 1.47271429483453e-04, 1.26633385306718e-04, 1.08943197716371e-04, 9.38544923343443e-05, 8.10304266707362e-05, 7.01536321656704e-05, 6.09330452334689e-05, 5.31074268650400e-05, 4.64466272234176e-05, 4.07512270754751e-05, 3.58504712920649e-05, 3.15998262436779e-05, 2.78796420616345e-05, 2.45937800826823e-05, 2.16665408032668e-05, 1.90387974328132e-05, 1.66650935966572e-05, 1.45116597707329e-05, 1.25543282081253e-05, 1.07761488266078e-05, 9.16520220073380e-06, 7.71287650991326e-06, 6.41251327238024e-06, 5.25831846687197e-06, 4.24455174652394e-06, 3.36500619368962e-06, 2.61272700186055e-06, 1.97988892125522e-06, 1.45777303828254e-06, 1.03681497876096e-06, 7.06709187170418e-07, 4.56558254313677e-07, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14],
[ 1.03000000000000e+02, 1.02999999869733e+02, 1.02999995841611e+02, 1.02999968550448e+02, 1.02999868224524e+02, 1.02999600786900e+02, 1.02999015478065e+02, 1.02997894433822e+02, 1.02995944591204e+02, 1.02992792221957e+02, 1.02987980310686e+02, 1.02980968906990e+02, 1.02971138492409e+02, 1.02957796318841e+02, 1.02940185600074e+02, 1.02917497376460e+02, 1.02888884826407e+02, 1.02853479766938e+02, 1.02810411064626e+02, 1.02758824660260e+02, 1.02697904885140e+02, 1.02626896703177e+02, 1.02545128442333e+02, 1.02452034478960e+02, 1.02347177215249e+02, 1.02230267559820e+02, 1.02101183009663e+02, 1.01959982368637e+02, 1.01806916152539e+02, 1.01642431843705e+02, 1.01467173374405e+02, 1.01281974524543e+02, 1.01087846284176e+02, 1.00885958611275e+02, 1.00677617361849e+02, 1.00464237440313e+02, 1.00247313383542e+02, 1.00028388641258e+02, 9.98090247560756e+01, 9.95907715014016e+01, 9.93751388362308e+01, 9.91635713161930e+01, 9.89574253891774e+01, 9.87579498222054e+01, 9.85662693647899e+01, 9.83833716545323e+01, 9.82100973080530e+01, 9.80471331055299e+01, 9.78950081599711e+01, 9.77540929536345e+01, 9.76246011166177e+01, 9.75065938119669e+01, 9.73999865754242e+01, 9.73045584356766e+01, 9.72199631134339e+01, 9.71457420662906e+01, 9.70813391128809e+01, 9.70261163361847e+01, 9.69793709338602e+01, 9.69403526550476e+01, 9.69082814400340e+01, 9.68823648633178e+01, 9.68618149736705e+01, 9.68458641282974e+01, 9.68337794333693e+01, 9.68248754308223e+01, 9.68185247116184e+01, 9.68141661881842e+01, 9.68113108223397e+01, 9.68095446778431e+01, 9.68085292462325e+01, 9.68079990779790e+01, 9.68077568348292e+01, 9.68076659602685e+01, 9.68076412401098e+01, 9.68076375914634e+01, 9.68076371128380e+01, 9.68076259614397e+01, 9.68075504090641e+01, 9.68072726703668e+01, 9.68065323196456e+01, 9.68049120734707e+01, 9.68018090824145e+01, 9.67964128741362e+01, 9.67876909497339e+01, 9.67743828724031e+01, 9.67550034984807e+01, 9.67278557806823e+01, 9.66910533161579e+01, 9.66425525135233e+01, 9.65801939120618e+01, 9.65017518068253e+01, 9.64049909266286e+01, 9.62877284981001e+01, 9.61478996382194e+01, 9.59836236898775e+01, 9.57932688961883e+01, 9.55755127452574e+01, 9.53293954413032e+01, 9.50543642766137e+01, 9.47503071572507e+01, 9.44175740944469e+01, 9.40569860044094e+01, 9.36698305617176e+01, 9.32578450899600e+01, 9.28231866244254e+01, 9.23683895371756e+01, 9.18963117012076e+01, 9.14100711887413e+01, 9.09129767609192e+01, 9.04084564126965e+01, 8.98999884022296e+01, 8.93910382410450e+01, 8.88850033473166e+01, 8.83851652177831e+01, 8.78946477929913e+01, 8.74163804776710e+01, 8.69530648189033e+01, 8.65071446589598e+01, 8.60807802401505e+01, 8.56758270433016e+01, 8.52938201024988e+01, 8.49359642839065e+01, 8.46031306886711e+01, 8.42958590423039e+01, 8.40143657157993e+01, 8.37585568952230e+01, 8.35280463601938e+01, 8.33221773220512e+01, 8.31400477847600e+01, 8.29805389066333e+01, 8.28423458465816e+01, 8.27240105690876e+01, 8.26239560568236e+01, 8.25405213417787e+01, 8.24719967204070e+01, 8.24166584725116e+01, 8.23728023648303e+01, 8.23387751960051e+01, 8.23130036365010e+01, 8.22940196405172e+01, 8.22804817606326e+01, 8.22711917812857e+01, 8.22651062032386e+01, 8.22613422544833e+01, 8.22591782680084e+01, 8.22580484460243e+01, 8.22575322150859e+01, 8.22573385580772e+01, 8.22572858785846e+01, 8.22572781031986e+01, 8.22572774167800e+01, 8.22572639667458e+01, 8.22571728407748e+01, 8.22568378519055e+01, 8.22559448933414e+01, 8.22539906691234e+01, 8.22502480720407e+01, 8.22437395894539e+01, 8.22332199502525e+01, 8.22171690319512e+01, 8.21937958216307e+01, 8.21610539619042e+01, 8.21166691068881e+01, 8.20581779585600e+01, 8.19829784500530e+01, 8.18883900945411e+01, 8.17717230392738e+01, 8.16303538754455e+01, 8.14618057856532e+01, 8.12638301981404e+01, 8.10344868011826e+01, 8.07722185922850e+01, 8.04759186316611e+01, 8.01449853654693e+01, 7.97793637961457e+01, 7.93795704013831e+01, 7.89467005130302e+01, 7.84824178075886e+01, 7.79889265474273e+01, 7.74689281404002e+01, 7.69255643440413e+01, 7.63623499385156e+01, 7.57830978882889e+01, 7.51918399309407e+01, 7.45927452603717e+01, 7.39900396319966e+01, 7.33879269170384e+01, 7.27905149219162e+01, 7.22017471410195e+01, 7.16253419450939e+01, 7.10647404361151e+01, 7.05230637852398e+01, 7.00030803508178e+01, 6.95071823447198e+01, 6.90373713845911e+01, 6.85952520074178e+01, 6.81820321368564e+01, 6.77985295535762e+01, 6.74451835518967e+01, 6.71220711191718e+01, 6.68289271081720e+01, 6.65651679700931e+01, 6.63299186748916e+01, 6.61220424717109e+01, 6.59401731418221e+01, 6.57827493749792e+01, 6.56480508609581e+01, 6.55342356342432e+01, 6.54393781447450e+01, 6.53615074558179e+01, 6.52986448989875e+01, 6.52488404503328e+01, 6.52102070447560e+01, 6.51809520196532e+01, 6.51594048859805e+01, 6.51440406677108e+01, 6.51334981328889e+01, 6.51265923605772e+01, 6.51223212443577e+01, 6.51198657181881e+01, 6.51185836952707e+01, 6.51179979245376e+01, 6.51177781810860e+01, 6.51177184054799e+01, 6.51177095827236e+01, 6.51177088750294e+01, 6.51176958144820e+01, 6.51176073273494e+01, 6.51172820392523e+01, 6.51164149403296e+01, 6.51145173140963e+01, 6.51108831272869e+01, 6.51045632285291e+01, 6.50943485451634e+01, 6.50787632832454e+01, 6.50560689206817e+01, 6.50242795305235e+01, 6.49811886732907e+01, 6.49244077482611e+01, 6.48514152925570e+01, 6.47596162687707e+01, 6.46464099004556e+01, 6.45092641232558e+01, 6.43457942502049e+01, 6.41538430424899e+01, 6.39315590748380e+01, 6.36774701285818e+01, 6.33905483676446e+01, 6.30702642704658e+01, 6.27166267021787e+01, 6.23302070937940e+01, 6.19121464094868e+01, 6.14641443804435e+01, 6.09884313140113e+01, 6.04877236050331e+01, 5.99651648430416e+01, 5.94242550865268e+01, 5.88687714209128e+01, 5.83026832802890e+01, 5.77300661443736e+01, 5.71550170871521e+01, 5.65815752511606e+01, 5.60136496952830e+01, 5.54549563001119e+01, 5.49089646236945e+01, 5.43788548898587e+01, 5.38674847411209e+01, 5.33773650343893e+01, 5.29106437911668e+01, 5.24690973924952e+01, 5.20541281757671e+01, 5.16667676928872e+01, 5.13076849898527e+01, 5.09771993485900e+01, 5.06752969904394e+01, 5.04016512832282e+01, 5.01556460281834e+01, 4.99364014336843e+01, 4.97428024100931e+01, 4.95735288399477e+01, 4.94270874853814e+01, 4.93018451851461e+01, 4.91960629647821e+01, 4.91079306361576e+01, 4.90356014009052e+01, 4.89772259030243e+01, 4.89309851078566e+01, 4.88951213274338e+01, 4.88679666753840e+01, 4.88479682265326e+01, 4.88337091833573e+01, 4.88239254170821e+01, 4.88175168556873e+01, 4.88135533313945e+01, 4.88112746700192e+01, 4.88100849951620e+01, 4.88095414210069e+01, 4.88093375072697e+01, 4.88092820377602e+01, 4.88092738505757e+01, 4.88092732191013e+01, 4.88092618798473e+01, 4.88091850547367e+01, 4.88089026375575e+01, 4.88081498174622e+01, 4.88065022912025e+01, 4.88033470893889e+01, 4.87978601922201e+01, 4.87889919755231e+01, 4.87754613712557e+01, 4.87557594429570e+01, 4.87281628592571e+01, 4.86907574904568e+01, 4.86414720489846e+01, 4.85781213426251e+01, 4.84984583140385e+01, 4.84002336130117e+01, 4.82812610092147e+01, 4.81394865311639e+01, 4.79730588460825e+01, 4.77803981129766e+01, 4.75602603836839e+01, 4.73117946238124e+01, 4.70345895964011e+01, 4.67287082009798e+01, 4.63947073798925e+01, 4.60336423690874e+01, 4.56470548472789e+01, 4.52369453801686e+01, 4.48057314085066e+01, 4.43561928201299e+01, 4.38914077950062e+01, 4.34146820360571e+01, 4.29294746352255e+01, 4.24393236592669e+01, 4.19477741223434e+01, 4.14583104482126e+01, 4.09742949411967e+01, 4.04989132858923e+01, 4.00351277269705e+01, 3.95856383269314e+01, 3.91528525078128e+01, 3.87388628964891e+01, 3.83454332810099e+01, 3.79739922499477e+01, 3.76256338560942e+01, 3.73011244571024e+01, 3.70009147688393e+01, 3.67251561359839e+01, 3.64737200744779e+01, 3.62462202535089e+01, 3.60420362350445e+01, 3.58603384497769e+01, 3.57001140366512e+01, 3.55601932922252e+01, 3.54392765562368e+01, 3.53359613976989e+01, 3.52487699636223e+01, 3.51761763160162e+01, 3.51166335207362e+01, 3.50686001742795e+01, 3.50305659727701e+01, 3.50010758521195e+01, 3.49787521699557e+01, 3.49623143671848e+01, 3.49505955466286e+01, 3.49425554419939e+01, 3.49372893232647e+01, 3.49340324920024e+01, 3.49321601564923e+01, 3.49311826341503e+01, 3.49307359973818e+01, 3.49305684487146e+01, 3.49305228714340e+01, 3.49305161443235e+01, 3.49305156349670e+01, 3.49305066117680e+01, 3.49304454782863e+01, 3.49302207454277e+01, 3.49296216923418e+01, 3.49283106905028e+01, 3.49258000080167e+01, 3.49214340442964e+01, 3.49143778466160e+01, 3.49036126444953e+01, 3.48879389998423e+01, 3.48659880029501e+01, 3.48362407391213e+01, 3.47970560016360e+01, 3.47467059311827e+01, 3.46834189222773e+01, 3.46054287629981e+01, 3.45110285826804e+01, 3.43986277978260e+01, 3.42668099006374e+01, 3.41143886626740e+01, 3.39404601643542e+01, 3.37444480423101e+01, 3.35261394962251e+01, 3.32857099281485e+01, 3.30237345982314e+01, 3.27411863509833e+01, 3.24394192555564e+01, 3.21201388533158e+01, 3.17853605414945e+01, 3.14373583594692e+01, 3.10786070018509e+01, 3.07117201922918e+01, 3.03393885739213e+01, 2.99643200057978e+01, 2.95891846418264e+01, 2.92165664891886e+01, 2.88489224008317e+01, 2.84885487572696e+01, 2.81375555247404e+01, 2.77978469929742e+01, 2.74711083082643e+01, 2.71587969003906e+01, 2.68621380039697e+01, 2.65821236359327e+01, 2.63195145575474e+01, 2.60748448853007e+01, 2.58484291042185e+01, 2.56403712817931e+01, 2.54505762935586e+01, 2.52787628686969e+01, 2.51244782591822e+01, 2.49871143361266e+01, 2.48659249230635e+01, 2.47600441842336e+01, 2.46685058907912e+01, 2.45902633838745e+01, 2.45242100372414e+01, 2.44691999929299e+01, 2.44240689031300e+01, 2.43876543643715e+01, 2.43588156820321e+01, 2.43364525606376e+01, 2.43195222850414e+01, 2.43070549452239e+01, 2.42981662676635e+01, 2.42920676516933e+01, 2.42880730704725e+01, 2.42856025813980e+01, 2.42841822960344e+01, 2.42834407792314e+01, 2.42831019741148e+01, 2.42829748766244e+01, 2.42829403030259e+01, 2.42829352000352e+01, 2.42829348169521e+01, 2.42829280742550e+01, 2.42828823914971e+01, 2.42827144569539e+01, 2.42822668059411e+01, 2.42812871368214e+01, 2.42794109692466e+01, 2.42761483481238e+01, 2.42708752193249e+01, 2.42628300127075e+01, 2.42511158610954e+01, 2.42347087557948e+01, 2.42124717850676e+01, 2.41831754178734e+01, 2.41455235797202e+01, 2.40981850230757e+01, 2.40398292289335e+01, 2.39691658017858e+01, 2.38849860557708e+01, 2.37862052575547e+01, 2.36719038158287e+01, 2.35413656111887e+01, 2.33941116621056e+01, 2.32299274338398e+01, 2.30488823197421e+01, 2.28513401519387e+01, 2.26379600173426e+01, 2.24096871465299e+01, 2.21677341838830e+01, 2.19135537071597e+01, 2.16488034014144e+01, 2.13753057491224e+01, 2.10950044072339e+01, 2.08099195380460e+01, 2.05221042079326e+01, 2.02336035822508e+01, 1.99464181029181e+01, 1.96624712567058e+01, 1.93835820497952e+01, 1.91114419806185e+01, 1.88475961677279e+01, 1.85934283024602e+01, 1.83501491854210e+01, 1.81187887000668e+01, 1.79001911289807e+01, 1.76950137136892e+01, 1.75037283076579e+01, 1.73266258984266e+01, 1.71638237039691e+01, 1.70152744988646e+01, 1.68807778065239e+01, 1.67599926041236e+01, 1.66524512201385e+01, 1.65575741502302e+01, 1.64746855650364e+01, 1.64030293239654e+01, 1.63417853359282e+01, 1.62900861176969e+01, 1.62470333929175e+01, 1.62117145520749e+01, 1.61832187602831e+01, 1.61606524614237e+01, 1.61431539902897e+01, 1.61299069753917e+01, 1.61201521995782e+01, 1.61131975880067e+01, 1.61084260159124e+01, 1.61053006727722e+01, 1.61033677835224e+01, 1.61022565682944e+01, 1.61016764150085e+01, 1.61014113384563e+01, 1.61013118991348e+01, 1.61012848492325e+01, 1.61012808567237e+01, 1.61012805579162e+01, 1.61012753107071e+01, 1.61012397600999e+01, 1.61011090724105e+01, 1.61007607078047e+01, 1.60999983244067e+01, 1.60985382844077e+01, 1.60959993108612e+01, 1.60918957843226e+01, 1.60856350990399e+01, 1.60765194163198e+01, 1.60637520543907e+01, 1.60464486358378e+01, 1.60236529726092e+01, 1.59943575036799e+01, 1.59575279135048e+01, 1.59121313557814e+01, 1.58571674960378e+01, 1.57917013811658e+01, 1.57148969601450e+01, 1.56260499352653e+01, 1.55246185341833e+01, 1.54102507747238e+01, 1.52828068566202e+01, 1.51423754617462e+01, 1.49892829746474e+01, 1.48240949396765e+01, 1.46476094351555e+01, 1.44608424488968e+01, 1.42650057587560e+01, 1.40614782285671e+01, 1.38517717931251e+01, 1.36374936947778e+01, 1.34203067205896e+01, 1.32018892522991e+01, 1.29838968725389e+01, 1.27679270759891e+01, 1.25554883345534e+01, 1.23479743951840e+01, 1.21466442888599e+01, 1.19526081412667e+01, 1.17668185359184e+01, 1.15900669142091e+01, 1.14229843169178e+01, 1.12660456784940e+01, 1.11195768697454e+01, 1.09837637307427e+01, 1.08586624255630e+01, 1.07442105658518e+01, 1.06402386752815e+01, 1.05464816892862e+01, 1.04625902948930e+01, 1.03881420081856e+01, 1.03226519586795e+01, 1.02655833994618e+01, 1.02163579896603e+01, 1.01743659031073e+01, 1.01389758062865e+01, 1.01095447228767e+01, 1.00854277651413e+01, 1.00659876682943e+01, 1.00506040174434e+01, 1.00386820125495e+01, 1.00296605797246e+01, 1.00230196113217e+01, 1.00182861059954e+01, 1.00150389854054e+01, 1.00129123872373e+01, 1.00115972738817e+01, 1.00108412501312e+01, 1.00104465479965e+01, 1.00102662077051e+01, 1.00101985562198e+01, 1.00101801534118e+01, 1.00101774371971e+01, 1.00101772338341e+01, 1.00101736616596e+01, 1.00101494596712e+01, 1.00100604908845e+01, 1.00098233349369e+01, 1.00093043359715e+01, 1.00083104331136e+01, 1.00065821599198e+01, 1.00037891659706e+01, 9.99952857420823e+00, 9.99332643827998e+00, 9.98464250059918e+00, 9.97287836940524e+00, 9.95738912784357e+00, 9.93749825787549e+00, 9.91251560746089e+00, 9.88175795554990e+00, 9.84457154496629e+00, 9.80035577162519e+00, 9.74858705672195e+00, 9.68884180577327e+00, 9.62081729388258e+00, 9.54434932680944e+00, 9.45942562318621e+00, 9.36619404716711e+00, 9.26496508556519e+00, 9.15620829138157e+00, 9.04054277963881e+00, 8.91872222839794e+00, 8.79161517255613e+00, 8.66018164817756e+00, 8.52544742595096e+00, 8.38847715015851e+00, 8.25034767308543e+00, 8.11212275481029e+00, 7.97483010483207e+00, 7.83944150062655e+00, 7.70685645607249e+00, 7.57788965477087e+00, 7.45326213002593e+00, 7.33359597917739e+00, 7.21941225340569e+00, 7.11113156771481e+00, 7.00907692742520e+00, 6.91347826149556e+00, 6.82447818129987e+00, 6.74213853693358e+00, 6.66644741243819e+00, 6.59732627799804e+00, 6.53463709399756e+00, 6.47818923325779e+00, 6.42774614987732e+00, 6.38303177345307e+00, 6.34373664485597e+00, 6.30952383395369e+00, 6.28003469119017e+00, 6.25489448477187e+00, 6.23371796481867e+00, 6.21611487702333e+00, 6.20169542327005e+00, 6.19007563772747e+00, 6.18088261682968e+00, 6.17375951308277e+00, 6.16837017855407e+00, 6.16440332672877e+00, 6.16157607321162e+00, 6.15963671787085e+00, 6.15836664397011e+00, 6.15758123315993e+00, 6.15712972751196e+00, 6.15689400887295e+00, 6.15678630890639e+00, 6.15674590722622e+00, 6.15673491701749e+00, 6.15673329488673e+00, 6.15673317311557e+00, 6.15673102984093e+00, 6.15671650885429e+00, 6.15666312838320e+00, 6.15652083727380e+00, 6.15620944492159e+00, 6.15561312296693e+00, 6.15457621312798e+00, 6.15290055705562e+00, 6.15034454185327e+00, 6.14662402651909e+00, 6.14141527695004e+00, 6.13435998702911e+00, 6.12507239789645e+00, 6.11314844522974e+00, 6.09817676568129e+00, 6.07975128165489e+00, 6.05748496454949e+00, 6.03102425969699e+00, 6.00006355326441e+00, 5.96435898556669e+00, 5.92374087949568e+00, 5.87812406779652e+00, 5.82751547507820e+00, 5.77201844002340e+00, 5.71183344351437e+00, 5.64725512557960e+00, 5.57866570875422e+00, 5.50652517487303e+00, 5.43135874352186e+00, 5.35374235353301e+00, 5.27428694020708e+00, 5.19362232443969e+00, 5.11238148806251e+00, 5.03118591235086e+00, 4.95063251903213e+00, 4.87128259308622e+00, 4.79365290184721e+00, 4.71820907062937e+00, 4.64536114268049e+00, 4.57546114762630e+00, 4.50880243044557e+00, 4.44562045162358e+00, 4.38609475519826e+00, 4.33035181007531e+00, 4.27846845569836e+00, 4.23047572030305e+00, 4.18636282339087e+00, 4.14608121928090e+00, 4.10954858208436e+00, 4.07665267156874e+00, 4.04725505238125e+00, 4.02119466498419e+00, 3.99829126504315e+00, 3.97834875898569e+00, 3.96115846745816e+00, 3.94650234613695e+00, 3.93415618568964e+00, 3.92389280068025e+00, 3.91548520207746e+00, 3.90870973109962e+00, 3.90334911487128e+00, 3.89919538829653e+00, 3.89605261317050e+00, 3.89373931622880e+00, 3.89209056369025e+00, 3.89095959162663e+00, 3.89021891945704e+00, 3.88976088774525e+00, 3.88949758047646e+00, 3.88936011484370e+00, 3.88929730668613e+00, 3.88927374534374e+00, 3.88926733610299e+00, 3.88926639011305e+00, 3.88926631879642e+00, 3.88926505953509e+00, 3.88925652786249e+00, 3.88922516465426e+00, 3.88914156279928e+00, 3.88895860696982e+00, 3.88860824300539e+00, 3.88799901450473e+00, 3.88701449329245e+00, 3.88551271727221e+00, 3.88332673239685e+00, 3.88026631325094e+00, 3.87612090712167e+00, 3.87066380776812e+00, 3.86365751646967e+00, 3.85486018961628e+00, 3.84403300598524e+00, 3.83094821666080e+00, 3.81539757187883e+00, 3.79720075902075e+00, 3.77621344245682e+00, 3.75233447666970e+00, 3.72551187534600e+00, 3.69574716446213e+00, 3.66309782663892e+00, 3.62767765272535e+00, 3.58965494606588e+00, 3.54924866326357e+00, 3.50672270873422e+00, 3.46237871532128e+00, 3.41654772816810e+00, 3.36958125613971e+00, 3.32184216137685e+00, 3.27369582513963e+00, 3.22550196353329e+00, 3.17760737983285e+00, 3.13033984234761e+00, 3.08400317934520e+00, 3.03887359505066e+00, 2.99519713995785e+00, 2.95318821821574e+00, 2.91302898513248e+00, 2.87486947679096e+00, 2.83882831762860e+00, 2.80499386610616e+00, 2.77342567887376e+00, 2.74415619641323e+00, 2.71719257526391e+00, 2.69251861191431e+00, 2.67009672043794e+00, 2.64986993975741e+00, 2.63176395717578e+00, 2.61568914277758e+00, 2.60154259471000e+00, 2.58921019835257e+00, 2.57856870305319e+00, 2.56948781849604e+00, 2.56183232899104e+00, 2.55546421826074e+00, 2.55024479004801e+00, 2.54603676164892e+00, 2.54270629901869e+00, 2.54012495424183e+00, 2.53817145976141e+00, 2.53673332963966e+00, 2.53570821692993e+00, 2.53500497841745e+00, 2.53454440367646e+00, 2.53425957443362e+00, 2.53409583214125e+00, 2.53401034571765e+00, 2.53397128667781e+00, 2.53395663435033e+00, 2.53395264856832e+00, 2.53395206027547e+00, 2.53395201568367e+00, 2.53395122511034e+00, 2.53394586886337e+00, 2.53392617878893e+00, 2.53387369271054e+00, 2.53375883039767e+00, 2.53353886370330e+00, 2.53315636598465e+00, 2.53253821833810e+00, 2.53159524116130e+00, 2.53022250837383e+00, 2.52830038734299e+00, 2.52569632916413e+00, 2.52226741060559e+00, 2.51786360031636e+00, 2.51233168805162e+00, 2.50551977781726e+00, 2.49728220609470e+00, 2.48748470779596e+00, 2.47600961920972e+00, 2.46276088320431e+00, 2.44766861139672e+00, 2.43069296401112e+00, 2.41182713225942e+00, 2.39109924970310e+00, 2.36857311532289e+00, 2.34434767700594e+00, 2.31855529355229e+00, 2.29135885941822e+00, 2.26294793332201e+00, 2.23353405533435e+00, 2.20334546526965e+00, 2.17262144854318e+00, 2.14160653639711e+00, 2.11054477860628e+00, 2.07967429127885e+00, 2.04922226192186e+00, 2.01940056890355e+00, 1.99040214211689e+00, 1.96239815514001e+00, 1.93553609649747e+00, 1.90993872045320e+00, 1.88570382974125e+00, 1.86290479876027e+00, 1.84159171122772e+00, 1.82179296517711e+00, 1.80351719237805e+00, 1.78675534807752e+00, 1.77148284746987e+00, 1.75766165318770e+00, 1.74524224870984e+00, 1.73416546182099e+00, 1.72436412717043e+00, 1.71576459590969e+00, 1.70828811284319e+00, 1.70185208787622e+00, 1.69637128967193e+00, 1.69175898645031e+00, 1.68792805291763e+00, 1.68479205446238e+00, 1.68226631094164e+00, 1.68026893343820e+00, 1.67872181903211e+00, 1.67755158155376e+00, 1.67669039102366e+00, 1.67607669148131e+00, 1.67565576645119e+00, 1.67538012347914e+00, 1.67520967389233e+00, 1.67511169085050e+00, 1.67506053735106e+00, 1.67503716546061e+00, 1.67502839794918e+00, 1.67502601298212e+00, 1.67502566096630e+00, 1.67502563411466e+00, 1.67502515582532e+00, 1.67502191534912e+00, 1.67501000308381e+00, 1.67497824988092e+00, 1.67490876142580e+00, 1.67477569288603e+00, 1.67454431738470e+00, 1.67417043844510e+00, 1.67360019234219e+00, 1.67277028060224e+00, 1.67160866482601e+00, 1.67003574463293e+00, 1.66796602400272e+00, 1.66531025111951e+00, 1.66197799201387e+00, 1.65788056957987e+00, 1.65293426847232e+00, 1.64706367538633e+00, 1.64020499647866e+00, 1.63230917292979e+00, 1.62334460574739e+00, 1.61329930538356e+00, 1.60218230312709e+00, 1.59002420045583e+00, 1.57687678830601e+00, 1.56281173663425e+00, 1.54791842913687e+00, 1.53230108978700e+00, 1.51607540711213e+00, 1.49936489966990e+00, 1.48229727549992e+00, 1.46500101746047e+00, 1.44760237881738e+00, 1.43022290781231e+00, 1.41297754798786e+00, 1.39597329496706e+00, 1.37930833990606e+00, 1.36307160024298e+00, 1.34734252996121e+00, 1.33219111055722e+00, 1.31767794408754e+00, 1.30385439449095e+00, 1.29076274736475e+00, 1.27843637795004e+00, 1.26689993068145e+00, 1.25616952131831e+00, 1.24625297544769e+00, 1.23715011649922e+00, 1.22885311378126e+00, 1.22134689759429e+00, 1.21460964497432e+00, 1.20861333647413e+00, 1.20332438172828e+00, 1.19870430930146e+00, 1.19471051434083e+00, 1.19129705567195e+00, 1.18841549206708e+00, 1.18601574541509e+00, 1.18404697645158e+00, 1.18245845666011e+00, 1.18120041808945e+00, 1.18022486133812e+00, 1.17948630105337e+00, 1.17894242816746e+00, 1.17855466891325e+00, 1.17828862251426e+00, 1.17811436234825e+00, 1.17800658926301e+00, 1.17794463041938e+00, 1.17791228231279e+00, 1.17789750219423e+00, 1.17789195766123e+00, 1.17789044941382e+00, 1.17789022679966e+00, 1.17789020969890e+00, 1.17788990352416e+00, 1.17788782914553e+00, 1.17788020351676e+00, 1.17785987643973e+00, 1.17781539137998e+00, 1.17773019813073e+00, 1.17758204936302e+00, 1.17734260963093e+00, 1.17697730133322e+00, 1.17644540728857e+00, 1.17570044370889e+00, 1.17469081020205e+00, 1.17336071478030e+00, 1.17165136162464e+00, 1.16950237768287e+00, 1.16685344147854e+00, 1.16364606451474e+00, 1.15982546341496e+00, 1.15534245071935e+00, 1.15015526538666e+00, 1.14423126172277e+00, 1.13754837846110e+00, 1.13009631826529e+00, 1.12187738150637e+00, 1.11290691563011e+00, 1.10321336112636e+00, 1.09283789521898e+00, 1.08183369323294e+00, 1.07026484389344e+00, 1.05820496783774e+00, 1.04573559814057e+00, 1.03294438777738e+00, 1.01992321191987e+00, 1.00676623300663e+00, 9.93567993805186e-01, 9.80421598300637e-01, 9.67417032402594e-01, 9.54639666535186e-01, 9.42168970769885e-01, 9.30077461087166e-01, 9.18429883519593e-01, 9.07282632205238e-01, 8.96683388477641e-01, 8.86670961485294e-01, 8.77275306640499e-01, 8.68517696339467e-01, 8.60411017560764e-01, 8.52960172688839e-01, 8.46162562717396e-01, 8.40008635359496e-01, 8.34482484075778e-01, 8.29562487256851e-01, 8.25221979477547e-01, 8.21429948692469e-01, 8.18151754365417e-01, 8.15349861803256e-01, 8.12984587454849e-01, 8.11014848758219e-01, 8.09398910449843e-01, 8.08095117304589e-01, 8.07062601297133e-01, 8.06261949418797e-01, 8.05655817094220e-01, 8.05209471528073e-01, 8.04891249541697e-01, 8.04672915624423e-01, 8.04529908040024e-01, 8.04441463831566e-01, 8.04390617308883e-01, 8.04364070874996e-01, 8.04351941597687e-01, 8.04347391488250e-01, 8.04346153748270e-01, 8.04345971060442e-01, 8.04345956919676e-01, 8.04345702346254e-01, 8.04343977574462e-01, 8.04337637137721e-01, 8.04320735913862e-01, 8.04283748265973e-01, 8.04212913518853e-01, 8.04089734502625e-01, 8.03890653066730e-01, 8.03586923161021e-01, 8.03144698138709e-01, 8.02525344167037e-01, 8.01685985823497e-01, 8.00580282939720e-01, 7.99159429465604e-01, 7.97373355637873e-01, 7.95172104332053e-01, 7.92507341648076e-01, 7.89333951246320e-01, 7.85611652606235e-01, 7.81306576223955e-01, 7.76392724794960e-01, 7.70853249497392e-01, 7.64681475187302e-01, 7.57881617816693e-01, 7.50469151393915e-01, 7.42470799512958e-01, 7.33924146615522e-01, 7.24876885103910e-01, 7.15385734412365e-01, 7.05515085437909e-01, 6.95335436867509e-01, 6.84921697901672e-01, 6.74351434235589e-01, 6.63703131093087e-01, 6.53054539342918e-01, 6.42481159413265e-01, 6.32054904228738e-01, 6.21842968140598e-01, 6.11906915066569e-01, 6.02301986803880e-01, 5.93076622399806e-01, 5.84272171891158e-01, 5.75922782705916e-01, 5.68055434369668e-01, 5.60690096545798e-01, 5.53839986445041e-01, 5.47511903829985e-01, 5.41706624792695e-01, 5.36419338816259e-01, 5.31640117021809e-01, 5.27354402681752e-01, 5.23543517841560e-01, 5.20185182087488e-01, 5.17254041027186e-01, 5.14722202872164e-01, 5.12559781628639e-01, 5.10735444869198e-01, 5.09216962968594e-01, 5.07971755180051e-01, 5.06967426175564e-01, 5.06172284869820e-01, 5.05555835699923e-01, 5.05089231243559e-01, 5.04745674308282e-01, 5.04500757554823e-01, 5.04332729414074e-01, 5.04222676544984e-01, 5.04154615308110e-01, 5.04115487586716e-01, 5.04095059602051e-01, 5.04085725935582e-01, 5.04082224562893e-01, 5.04081272105367e-01, 5.04081131524657e-01, 5.04081120554929e-01, 5.04080921929429e-01, 5.04079576213343e-01, 5.04074629229166e-01, 5.04061442466892e-01, 5.04032583972204e-01, 5.03977318081454e-01, 5.03881215174701e-01, 5.03725900593977e-01, 5.03488959481473e-01, 5.03144011381394e-01, 5.02660964813501e-01, 5.02006457453987e-01, 5.01144481872473e-01, 5.00037189885646e-01, 4.98645860554788e-01, 4.96932007921728e-01, 4.94858595202629e-01, 4.92391313013485e-01, 4.89499871126341e-01, 4.86159247192987e-01, 4.82350832745837e-01, 4.78063417365021e-01, 4.73293956669949e-01, 4.68048078825049e-01, 4.62340297143939e-01, 4.56193912256694e-01, 4.49640604895123e-01, 4.42719738125447e-01, 4.35477404228209e-01, 4.27965264956924e-01, 4.20239243499168e-01, 4.12358131474477e-01, 4.04382174609310e-01, 3.96371696690451e-01, 3.88385813769347e-01, 3.80481280388864e-01, 3.72711497956672e-01, 3.65125703401375e-01, 3.57768344878934e-01, 3.50678641304555e-01, 3.43890314363918e-01, 3.37431475666818e-01, 3.31324647872847e-01, 3.25586896802932e-01, 3.20230051480024e-01, 3.15260990376714e-01, 3.10681974504831e-01, 3.06491010993341e-01, 3.02682234119104e-01, 2.99246294089784e-01, 2.96170746985254e-01, 2.93440441965046e-01, 2.91037904013066e-01, 2.88943712038986e-01, 2.87136873055374e-01, 2.85595193408000e-01, 2.84295647705605e-01, 2.83214745251527e-01, 2.82328892545834e-01, 2.81614748938580e-01, 2.81049570935011e-01, 2.80611539155430e-01, 2.80280060698516e-01, 2.80036038803360e-01, 2.79862101371990e-01, 2.79742780177330e-01, 2.79664633472194e-01, 2.79616306197013e-01, 2.79588523980081e-01, 2.79574019497730e-01, 2.79567392357293e-01, 2.79564906300607e-01, 2.79564230034111e-01, 2.79564130218631e-01, 2.79564122363581e-01, 2.79563979285882e-01, 2.79563009914256e-01, 2.79559446414552e-01, 2.79549947532288e-01, 2.79529159966003e-01, 2.79489351225692e-01, 2.79420129550616e-01, 2.79308265186486e-01, 2.79137625796565e-01, 2.78889236853630e-01, 2.78541475337832e-01, 2.78070401747598e-01, 2.77450231137692e-01, 2.76653938544470e-01, 2.75653987766205e-01, 2.74423165242581e-01, 2.72935493103859e-01, 2.71167187919417e-01, 2.69097624998182e-01, 2.66710263100606e-01, 2.63993481921887e-01, 2.60941285368403e-01, 2.57553827889069e-01, 2.53837728988955e-01, 2.49806152171629e-01, 2.45478638135594e-01, 2.40880696945950e-01, 2.36043178751544e-01, 2.31001456023606e-01, 2.25794460998316e-01, 2.20463629083312e-01, 2.15051801935502e-01, 2.09602142717770e-01, 2.04157111113020e-01, 1.98757537796496e-01, 1.93441828224707e-01, 1.88245314867771e-01, 1.83199766402083e-01, 1.78333052754567e-01, 1.73668956861718e-01, 1.69227117937055e-01, 1.65023087025759e-01, 1.61068473571945e-01, 1.57371161393669e-01, 1.53935573542029e-01, 1.50762967656263e-01, 1.47851746253394e-01, 1.45197769566821e-01, 1.42794661783554e-01, 1.40634104592252e-01, 1.38706114687267e-01, 1.36999304166181e-01, 1.35501124545470e-01, 1.34198096368351e-01, 1.33076027086929e-01, 1.32120220091153e-01, 1.31315677478426e-01, 1.30647298480966e-01, 1.30100074480183e-01, 1.29659280342253e-01, 1.29310660524027e-01, 1.29040607142883e-01, 1.28836326103979e-01, 1.28685986541967e-01, 1.28578848357563e-01, 1.28505362572241e-01, 1.28457239615756e-01, 1.28427481483385e-01, 1.28410374901264e-01, 1.28401444126681e-01, 1.28397363674026e-01, 1.28395832970256e-01, 1.28395416583118e-01, 1.28395355125292e-01, 1.28395350246024e-01, 1.28395260827119e-01, 1.28394655001770e-01, 1.28392427937428e-01, 1.28386491506674e-01, 1.28373500306438e-01, 1.28348622647309e-01, 1.28305366583803e-01, 1.28235470452353e-01, 1.28128866633986e-01, 1.27973726227538e-01, 1.27756590838075e-01, 1.27462595568872e-01, 1.27075784399537e-01, 1.26579515341248e-01, 1.25956948095180e-01, 1.25191601555997e-01, 1.24267962705545e-01, 1.23172122714010e-01, 1.21892411007778e-01, 1.20419994358670e-01, 1.18749406362411e-01, 1.16878973551238e-01, 1.14811108141652e-01, 1.12552444050522e-01, 1.10113801948017e-01, 1.07509980025497e-01, 1.04759378822948e-01, 1.01883479720215e-01, 9.89062063976291e-02, 9.58532057449162e-02, 9.27510886845511e-02, 8.96266719021810e-02, 8.65062586720455e-02, 8.34149913341093e-02, 8.03763003123663e-02, 7.74114658236584e-02, 7.45392995835346e-02, 7.17759456871505e-02, 6.91347930106758e-02, 6.66264862629333e-02, 6.42590193402413e-02, 6.20378928804115e-02, 5.99663177264205e-02, 5.80454471318603e-02, 5.62746225940600e-02, 5.46516207923702e-02, 5.31728918887069e-02, 5.18337821851560e-02, 5.06287366816736e-02, 4.95514793516492e-02, 4.85951708816827e-02, 4.77525451489385e-02, 4.70160268107804e-02, 4.63778330640423e-02, 4.58300629370996e-02, 4.53647774564274e-02, 4.49740737303017e-02, 4.46501554687444e-02, 4.43854017576655e-02, 4.41724350856845e-02, 4.40041887454043e-02, 4.38739728681554e-02, 4.37755375685640e-02, 4.37031310410987e-02, 4.36515500196606e-02, 4.36161798217540e-02, 4.35930212739208e-02, 4.35787021464051e-02, 4.35704712874951e-02, 4.35661743905627e-02, 4.35642111834937e-02, 4.35634747297797e-02, 4.35632743977692e-02, 4.35632448292296e-02, 4.35632424602375e-02, 4.35631987746774e-02, 4.35629027994366e-02, 4.35618147745487e-02, 4.35589145866686e-02, 4.35525680382938e-02, 4.35404153318899e-02, 4.35192869708345e-02, 4.34851521423951e-02, 4.34331044502727e-02, 4.33573895021482e-02, 4.32514782039419e-02, 4.31081885320026e-02, 4.29198569246552e-02, 4.26785581806399e-02, 4.23763698728266e-02, 4.20056738827412e-02, 4.15594839553108e-02, 4.10317845057619e-02, 4.04178627246239e-02, 3.97146138092808e-02, 3.89207983796922e-02, 3.80372321649958e-02, 3.70668910596521e-02, 3.60149195597046e-02, 3.48885370712253e-02, 3.36968440623480e-02, 3.24505377567448e-02, 3.11615541914136e-02, 2.98426591246397e-02, 2.85070137435213e-02, 2.71677419521126e-02, 2.58375242380566e-02, 2.45282392192223e-02, 2.32506687550682e-02, 2.20142767282283e-02, 2.08270657301455e-02, 1.96955101664429e-02, 1.86245590336114e-02, 1.76176974155169e-02, 1.66770533099765e-02, 1.58035360380353e-02, 1.49969937618692e-02, 1.42563795546664e-02, 1.35799171787454e-02, 1.29652590830744e-02, 1.24096305499819e-02, 1.19099558225599e-02, 1.14629643246502e-02, 1.10652771987390e-02, 1.07134757980276e-02, 1.04041543689277e-02, 1.01339592340821e-02, 9.89961671456285e-03, 9.69795202006196e-03, 9.52590136311926e-03, 9.38051950490149e-03, 9.25898473784059e-03, 9.15860296581172e-03, 9.07681211464251e-03, 9.01118765669651e-03, 8.95944959837891e-03, 8.91947088240386e-03, 8.88928680333834e-03, 8.86710474465247e-03, 8.85131332904621e-03, 8.84048995067133e-03, 8.83340563296997e-03, 8.82902623489201e-03, 8.82650919820094e-03, 8.82519527610270e-03, 8.82459497942989e-03, 8.82436979472345e-03, 8.82430853974370e-03, 8.82429949864161e-03, 8.82429876743629e-03, 8.82428519857115e-03, 8.82419326798356e-03, 8.82385532761987e-03, 8.82295455036797e-03, 8.82098346350801e-03, 8.81720952398268e-03, 8.81064951349231e-03, 8.80005458020605e-03, 8.78390778717630e-03, 8.76043599069476e-03, 8.72763769713458e-03, 8.68332816371264e-03, 8.62520235078534e-03, 8.55091536575969e-03, 8.45817876177080e-03, 8.34486952845737e-03, 8.20914695239594e-03, 8.04957089772529e-03, 7.86521366666193e-03, 7.65575666919460e-03, 7.42156290600349e-03, 7.16371700055163e-03, 6.88402643986308e-03, 6.58498087418025e-03, 6.26967053199223e-03, 5.94166925476713e-03, 5.60489116098495e-03, 5.26343142850294e-03, 4.92140098222649e-03, 4.58276324785345e-03, 4.25118070278074e-03, 3.92988104898961e-03, 3.62155610744708e-03, 3.32830652014593e-03, 3.05163749928283e-03, 2.79249670175830e-03, 2.55133425396536e-03, 2.32816554080273e-03, 2.12262780283368e-03, 1.93403217106436e-03, 1.76141251213835e-03, 1.60356174394849e-03, 1.45910169173646e-03, 1.32620238088066e-03, 1.20452806679339e-03, 1.09236676290436e-03, 9.88657733941523e-04, 8.92475525304803e-04, 8.03041505761826e-04, 7.19725182747766e-04, 6.42036716897110e-04, 5.69612420775911e-04, 5.02195250660975e-04, 4.39612362484354e-04, 3.81751697631216e-04, 3.28539311766463e-04, 2.79918799577672e-04, 2.35833748675418e-04, 1.96213734710682e-04, 1.60963990239793e-04, 1.29958576337685e-04, 1.03077746416602e-04, 8.00403742715616e-05, 6.06551482036366e-05, 4.46597209133870e-05, 3.17630255324275e-05, 2.16498734602055e-05, 1.39864057047134e-05, 8.42609998130683e-06, 4.61623591263560e-06, 2.20484675139687e-06, 8.48333158387504e-07, 2.84217094304040e-14, 1.42108547152020e-14],
[ 1.04000000000000e+02, 1.03999999867898e+02, 1.03999995783027e+02, 1.03999968107378e+02, 1.03999866368033e+02, 1.03999595162661e+02, 1.03999001607770e+02, 1.03997864769682e+02, 1.03995887456244e+02, 1.03992690672923e+02, 1.03987810963349e+02, 1.03980700765343e+02, 1.03970731823767e+02, 1.03957201616277e+02, 1.03939342672227e+02, 1.03916334602757e+02, 1.03887318613500e+02, 1.03851414239893e+02, 1.03807738024189e+02, 1.03755423835204e+02, 1.03693644505765e+02, 1.03621634417744e+02, 1.03538712591581e+02, 1.03444305733937e+02, 1.03337970569962e+02, 1.03219414652610e+02, 1.03088514726739e+02, 1.02945331661674e+02, 1.02790120982210e+02, 1.02623338145483e+02, 1.02445637934565e+02, 1.02257867654906e+02, 1.02061054194278e+02, 1.01856385395315e+02, 1.01645186542681e+02, 1.01428893040565e+02, 1.01209020521220e+02, 1.00987133670776e+02, 1.00764814993425e+02, 1.00543634583176e+02, 1.00325121766677e+02, 1.00110739255661e+02, 9.99018602329140e+01, 9.96997486123119e+01, 9.95055425718857e+01, 9.93202413604026e+01, 9.91446953168617e+01, 9.89795990091625e+01, 9.88254873824642e+01, 9.86827347999712e+01, 9.85515568518976e+01, 9.84320147976196e+01, 9.83240224893530e+01, 9.82273556027688e+01, 9.81416629716406e+01, 9.80664797914427e+01, 9.80012424226134e+01, 9.79453044898697e+01, 9.78979539414018e+01, 9.78584307028664e+01, 9.78259445376859e+01, 9.77996927090673e+01, 9.77788770320942e+01, 9.77627199077791e+01, 9.77504789463036e+01, 9.77414598146399e+01, 9.77350269845421e+01, 9.77306121100811e+01, 9.77277198283042e+01, 9.77259308503500e+01, 9.77249022909337e+01, 9.77243652685403e+01, 9.77241198936221e+01, 9.77240278442179e+01, 9.77240028044724e+01, 9.77239991086557e+01, 9.77239986238425e+01, 9.77239873282770e+01, 9.77239107991469e+01, 9.77236294697985e+01, 9.77228795477350e+01, 9.77212383549878e+01, 9.77180952490301e+01, 9.77126292815074e+01, 9.77037946100357e+01, 9.76903145134789e+01, 9.76706846698011e+01, 9.76431861324587e+01, 9.76059081806231e+01, 9.75567809158289e+01, 9.74936171317042e+01, 9.74141625983744e+01, 9.73161534904939e+01, 9.71973792683698e+01, 9.70557489262221e+01, 9.68893581908290e+01, 9.66965550345091e+01, 9.64760008044211e+01, 9.62267243981830e+01, 9.59481672381726e+01, 9.56402172763122e+01, 9.53032308158864e+01, 9.49380414575233e+01, 9.45459558666219e+01, 9.41287362930488e+01, 9.36885699416569e+01, 9.32280255975856e+01, 9.27499985780182e+01, 9.22576461867298e+01, 9.17543171529219e+01, 9.12434794969645e+01, 9.07286512950293e+01, 9.02133376932917e+01, 8.97009756381186e+01, 8.91948859478381e+01, 8.86982312787427e+01, 8.82139784652919e+01, 8.77448643572679e+01, 8.72933651235408e+01, 8.68616696270050e+01, 8.64516577262654e+01, 8.60648842675479e+01, 8.57025692376017e+01, 8.53655942011563e+01, 8.50545048443122e+01, 8.47695192326271e+01, 8.45105412725929e+01, 8.42771788177686e+01, 8.40687658586116e+01, 8.38843882525554e+01, 8.37229124690259e+01, 8.35830168310883e+01, 8.34632247260331e+01, 8.33619392312367e+01, 8.32774785624695e+01, 8.32081117051817e+01, 8.31520935422522e+01, 8.31076987518638e+01, 8.30732537240816e+01, 8.30471657411547e+01, 8.30279486900012e+01, 8.30142446294468e+01, 8.30048406210122e+01, 8.29986803493898e+01, 8.29948702037560e+01, 8.29926796580219e+01, 8.29915359695247e+01, 8.29910134028194e+01, 8.29908173690390e+01, 8.29907640430062e+01, 8.29907561721921e+01, 8.29907554773491e+01, 8.29907418622414e+01, 8.29906496178737e+01, 8.29903105176630e+01, 8.29894065997962e+01, 8.29874283915350e+01, 8.29836398626100e+01, 8.29770515056331e+01, 8.29664027721245e+01, 8.29501548963327e+01, 8.29264949111218e+01, 8.28933513945263e+01, 8.28484221755564e+01, 8.27892138686440e+01, 8.27130926968373e+01, 8.26173456096996e+01, 8.24992502158535e+01, 8.23561515539642e+01, 8.21855432498867e+01, 8.19851501888449e+01, 8.17530095110617e+01, 8.14875465587149e+01, 8.11876423987596e+01, 8.08526897485488e+01, 8.04826345540307e+01, 8.00780011092269e+01, 7.96398994316521e+01, 7.91700145631905e+01, 7.86705784623072e+01, 7.81443260825941e+01, 7.75944379828954e+01, 7.70244722983147e+01, 7.64382890840504e+01, 7.58399699581674e+01, 7.52337357090306e+01, 7.46238642199800e+01, 7.40146107985248e+01, 7.34101328164552e+01, 7.28144204301451e+01, 7.22312349665874e+01, 7.16640562466453e+01, 7.11160396495735e+01, 7.05899831590126e+01, 7.00883040809879e+01, 6.96130246993307e+01, 6.91657658958598e+01, 6.87477477094583e+01, 6.83597958890077e+01, 6.80023536420007e+01, 6.76754979357716e+01, 6.73789598369340e+01, 6.71121484641397e+01, 6.68741781810475e+01, 6.66638986773069e+01, 6.64799275822497e+01, 6.63206852335148e+01, 6.61844311838293e+01, 6.60693019758599e+01, 6.59733496504053e+01, 6.58945803817330e+01, 6.58309925618420e+01, 6.57806135906003e+01, 6.57415345795530e+01, 6.57119421521621e+01, 6.56901465296617e+01, 6.56746051350431e+01, 6.56639410306880e+01, 6.56569556275109e+01, 6.56526352615660e+01, 6.56501514212862e+01, 6.56488546157330e+01, 6.56482620906646e+01, 6.56480398134241e+01, 6.56479793485657e+01, 6.56479704240771e+01, 6.56479697082227e+01, 6.56479564970786e+01, 6.56478669896309e+01, 6.56475379507494e+01, 6.56466608536260e+01, 6.56447413466427e+01, 6.56410652558394e+01, 6.56346724864991e+01, 6.56243400270906e+01, 6.56085750724915e+01, 6.55856190659380e+01, 6.55534632033835e+01, 6.55098756424247e+01, 6.54524403050357e+01, 6.53786067576511e+01, 6.52857501988907e+01, 6.51712400982417e+01, 6.50325155317284e+01, 6.48671647857705e+01, 6.46730063880754e+01, 6.44481684179227e+01, 6.41911627887399e+01, 6.39009512158801e+01, 6.35769997995199e+01, 6.32193195649801e+01, 6.28284908890204e+01, 6.24056704627945e+01, 6.19525802526422e+01, 6.14714787693635e+01, 6.09651157991237e+01, 6.04366725423079e+01, 5.98896898072320e+01, 5.93279874644793e+01, 5.87555787297948e+01, 5.81765829573210e+01, 5.75951404590267e+01, 5.70153324267480e+01, 5.64411083735343e+01, 5.58762227253722e+01, 5.53241813980372e+01, 5.47881984946144e+01, 5.42711627320547e+01, 5.37756128782070e+01, 5.33037213370755e+01, 5.28572850116545e+01, 5.24377226418629e+01, 5.20460779085254e+01, 5.16830276798787e+01, 5.13488948415189e+01, 5.10436651957343e+01, 5.07670079503181e+01, 5.05182993484038e+01, 5.02966490232652e+01, 5.01009286936737e+01, 4.99298028408411e+01, 4.97817610205588e+01, 4.96551514584693e+01, 4.95482155499099e+01, 4.94591228394055e+01, 4.93860059930536e+01, 4.93269952069218e+01, 4.92802514252718e+01, 4.92439976839081e+01, 4.92165478559981e+01, 4.91963320688771e+01, 4.91819180870736e+01, 4.91720280226487e+01, 4.91655498392292e+01, 4.91615432577040e+01, 4.91592398431246e+01, 4.91580372449889e+01, 4.91574877661094e+01, 4.91572816373109e+01, 4.91572255652514e+01, 4.91572172891319e+01, 4.91572166507979e+01, 4.91572051883688e+01, 4.91571275287290e+01, 4.91568420437378e+01, 4.91560810460271e+01, 4.91544156234848e+01, 4.91512261490247e+01, 4.91456796541786e+01, 4.91367151191614e+01, 4.91230375738795e+01, 4.91031217188800e+01, 4.90752255556521e+01, 4.90374142548497e+01, 4.89875941834373e+01, 4.89235566558074e+01, 4.88430305738270e+01, 4.87437426883686e+01, 4.86234837704376e+01, 4.84801785519271e+01, 4.83119569191224e+01, 4.81172235543463e+01, 4.78947230594335e+01, 4.76435975898196e+01, 4.73634342001855e+01, 4.70542994583871e+01, 4.67167594155516e+01, 4.63518837040363e+01, 4.59612333353637e+01, 4.55468326379464e+01, 4.51111266458134e+01, 4.46569260464621e+01, 4.41873424308402e+01, 4.37057169793403e+01, 4.32155458124937e+01, 4.27204050359178e+01, 4.22238780818730e+01, 4.17294874075792e+01, 4.12406320741589e+01, 4.07605322850910e+01, 4.02921816344330e+01, 3.98383075729728e+01, 3.94013403893678e+01, 3.89833907760197e+01, 3.85862357890622e+01, 3.82113127345062e+01, 3.78597202550794e+01, 3.75322256941323e+01, 3.72292777017696e+01, 3.69510230326841e+01, 3.66973265550357e+01, 3.64677936221173e+01, 3.62617941246322e+01, 3.60784877130150e+01, 3.59168498332541e+01, 3.57756983402041e+01, 3.56537205309091e+01, 3.55495004749812e+01, 3.54615465126764e+01, 3.53883187506878e+01, 3.53282563199246e+01, 3.52798040791466e+01, 3.52414383643279e+01, 3.52116913069856e+01, 3.51891731855806e+01, 3.51725922411624e+01, 3.51607713882539e+01, 3.51526612883814e+01, 3.51473493273968e+01, 3.51440641462970e+01, 3.51421755130615e+01, 3.51411894820165e+01, 3.51407389575919e+01, 3.51405699505373e+01, 3.51405239765406e+01, 3.51405171908757e+01, 3.51405166770856e+01, 3.51405075753464e+01, 3.51404459097439e+01, 3.51402192207584e+01, 3.51396149533987e+01, 3.51382925404627e+01, 3.51357600051800e+01, 3.51313560418215e+01, 3.51242384334701e+01, 3.51133795496532e+01, 3.50975695273264e+01, 3.50754275692971e+01, 3.50454215867475e+01, 3.50058961606795e+01, 3.49551084979715e+01, 3.48912717138322e+01, 3.48126043933047e+01, 3.47173849875304e+01, 3.46040092110538e+01, 3.44710482564900e+01, 3.43173053687257e+01, 3.41418681596627e+01, 3.39441540302256e+01, 3.37239462248714e+01, 3.34814183886299e+01, 3.32171460248165e+01, 3.29321039406741e+01, 3.26276495753118e+01, 3.23054929666102e+01, 3.19676549526772e+01, 3.16164159322824e+01, 3.12542580441490e+01, 3.08838039011159e+01, 3.05077549973474e+01, 3.01288326017482e+01, 2.97497234091844e+01, 2.93730315303030e+01, 2.90012376681298e+01, 2.86366656609946e+01, 2.82814560493194e+01, 2.79375459927968e+01, 2.76066547258730e+01, 2.72902737579306e+01, 2.69896611426834e+01, 2.67058392973153e+01, 2.64395959953549e+01, 2.61914882583609e+01, 2.59618489225500e+01, 2.57507956670465e+01, 2.55582422786482e+01, 2.53839119121565e+01, 2.52273520979266e+01, 2.50879512537918e+01, 2.49649564742997e+01, 2.48574923893575e+01, 2.47645808989013e+01, 2.46851615935186e+01, 2.46181126593336e+01, 2.45622720385128e+01, 2.45164585770434e+01, 2.44794928438587e+01, 2.44502172563065e+01, 2.44275151033700e+01, 2.44103280268371e+01, 2.43976715078036e+01, 2.43886479160964e+01, 2.43824567161084e+01, 2.43784014846251e+01, 2.43758934826525e+01, 2.43744516300233e+01, 2.43736988528527e+01, 2.43733549027100e+01, 2.43732258751312e+01, 2.43731907765007e+01, 2.43731855960164e+01, 2.43731852071158e+01, 2.43731783620246e+01, 2.43731319855310e+01, 2.43729615007367e+01, 2.43725070516493e+01, 2.43715125048994e+01, 2.43696078440786e+01, 2.43662956704469e+01, 2.43609424448796e+01, 2.43527750138556e+01, 2.43408828535640e+01, 2.43242263361488e+01, 2.43016511657839e+01, 2.42719089458365e+01, 2.42336836204013e+01, 2.41856232864066e+01, 2.41263766036742e+01, 2.40546327526227e+01, 2.39691636209096e+01, 2.38688666634950e+01, 2.37528066996532e+01, 2.36202548087642e+01, 2.34707224837661e+01, 2.33039893098729e+01, 2.31201226613957e+01, 2.29194882477162e+01, 2.27027507793438e+01, 2.24708645476633e+01, 2.22250542896170e+01, 2.19667873010354e+01, 2.16977383138256e+01, 2.14197490925761e+01, 2.11347849614795e+01, 2.08448904869440e+01, 2.05521463027845e+01, 2.02586286217794e+01, 1.99663724321837e+01, 1.96773388560533e+01, 1.93933867526704e+01, 1.91162484344010e+01, 1.88475093092822e+01, 1.85885913139757e+01, 1.83407400782745e+01, 1.81050158097067e+01, 1.78822878763025e+01, 1.76732329999617e+01, 1.74783368744026e+01, 1.72978989190849e+01, 1.71320397982382e+01, 1.69807112868435e+01, 1.68437080569027e+01, 1.67206809826238e+01, 1.66111516118504e+01, 1.65145275107174e+01, 1.64301182472950e+01, 1.63571518283841e+01, 1.62947914353446e+01, 1.62421523169801e+01, 1.61983186902933e+01, 1.61623604762679e+01, 1.61333496626090e+01, 1.61103760447734e+01, 1.60925620574201e+01, 1.60790763772466e+01, 1.60691459609153e+01, 1.60620661829409e+01, 1.60572087607422e+01, 1.60540271982872e+01, 1.60520595445271e+01, 1.60509283448595e+01, 1.60503377583890e+01, 1.60500679149414e+01, 1.60499666874111e+01, 1.60499391510744e+01, 1.60499350867690e+01, 1.60499347825881e+01, 1.60499294410192e+01, 1.60498932511121e+01, 1.60497602133007e+01, 1.60494055842273e+01, 1.60486294916752e+01, 1.60471431988463e+01, 1.60445585771705e+01, 1.60403812866049e+01, 1.60340080920834e+01, 1.60247286615965e+01, 1.60117320915862e+01, 1.59941182850720e+01, 1.59709141641675e+01, 1.59410945303001e+01, 1.59036071942975e+01, 1.58574017899356e+01, 1.58014614679504e+01, 1.57348364560726e+01, 1.56566782805595e+01, 1.55662732938316e+01, 1.54630740588384e+01, 1.53467271191092e+01, 1.52170957453082e+01, 1.50742764000108e+01, 1.49186079011353e+01, 1.47506725828360e+01, 1.45712891356648e+01, 1.43814972340776e+01, 1.41825345018225e+01, 1.39758067927311e+01, 1.37628531414826e+01, 1.35453070318995e+01, 1.33248558096665e+01, 1.31032001119369e+01, 1.28820150918877e+01, 1.26629149920821e+01, 1.24474222925873e+01, 1.22369422664947e+01, 1.20327433610829e+01, 1.18359434304284e+01, 1.16475015101683e+01, 1.14682145706969e+01, 1.12987185211802e+01, 1.11394926608369e+01, 1.09908667740962e+01, 1.08530301250885e+01, 1.07260417052312e+01, 1.06098412072600e+01, 1.05042603246048e+01, 1.04090340948152e+01, 1.03238121117346e+01, 1.02481695183571e+01, 1.01816177583356e+01, 1.01236151083140e+01, 1.00735770363282e+01, 1.00308864351163e+01, 9.99490376569356e+00, 9.96497711902794e+00, 9.94045216570206e+00, 9.92068191917873e+00, 9.90503619217323e+00, 9.89291058232439e+00, 9.88373478738056e+00, 9.87698002555652e+00, 9.87216532683583e+00, 9.86886246788858e+00, 9.86669934763324e+00, 9.86536164143867e+00, 9.86459262718915e+00, 9.86419114230331e+00, 9.86400770279363e+00, 9.86393888866462e+00, 9.86392016958742e+00, 9.86391740669205e+00, 9.86391719983396e+00, 9.86391356626829e+00, 9.86388894834434e+00, 9.86379845052352e+00, 9.86355721864719e+00, 9.86302929917979e+00, 9.86201831083787e+00, 9.86026032066036e+00, 9.85741928223115e+00, 9.85308535290964e+00, 9.84677635616626e+00, 9.83794259007135e+00, 9.82597509900231e+00, 9.81021741879982e+00, 9.78998067384586e+00, 9.76456174870077e+00, 9.73326408173861e+00, 9.69542044278347e+00, 9.65041687460842e+00, 9.59771681653507e+00, 9.53688430664668e+00, 9.46760509658698e+00, 9.38970452595227e+00, 9.30316110243926e+00, 9.20811492127248e+00, 9.10487032492772e+00, 8.99389253325012e+00, 8.87579833714344e+00, 8.75134131255396e+00, 8.62139234045490e+00, 8.48691648141869e+00, 8.34894742666336e+00, 8.20856081898654e+00, 8.06684770701733e+00, 7.92488927644196e+00, 7.78373381262323e+00, 7.64437661542902e+00, 7.50774333501145e+00, 7.37467695018564e+00, 7.24592838779955e+00, 7.12215059533543e+00, 7.00389573755064e+00, 6.89161509373091e+00, 6.78566118326418e+00, 6.68629163861642e+00, 6.59367436900506e+00, 6.50789360654352e+00, 6.42895649060087e+00, 6.35679991745371e+00, 6.29129745410435e+00, 6.23226618202969e+00, 6.17947339481891e+00, 6.13264312079085e+00, 6.09146247657156e+00, 6.05558787999145e+00, 6.02465116093880e+00, 5.99826560792206e+00, 5.97603197739339e+00, 5.95754447411667e+00, 5.94239668609819e+00, 5.93018742924626e+00, 5.92052642763828e+00, 5.91303972783506e+00, 5.90737472284464e+00, 5.90320464560848e+00, 5.90023238528251e+00, 5.89819348345024e+00, 5.89685818218781e+00, 5.89603242110741e+00, 5.89555771469604e+00, 5.89530988219153e+00, 5.89519664707687e+00, 5.89515416893551e+00, 5.89514261387558e+00, 5.89514090837382e+00, 5.89514078034411e+00, 5.89513852691377e+00, 5.89512325960440e+00, 5.89506713555197e+00, 5.89491753090726e+00, 5.89459013227638e+00, 5.89396315337315e+00, 5.89287292142434e+00, 5.89111105735800e+00, 5.88842344928221e+00, 5.88451118793176e+00, 5.87903359265682e+00, 5.87161340400080e+00, 5.86184415129418e+00, 5.84929961953029e+00, 5.83354523962024e+00, 5.81415111310417e+00, 5.79070626269589e+00, 5.76283358283042e+00, 5.73020486131317e+00, 5.69255516724860e+00, 5.64969586440030e+00, 5.60152552350758e+00, 5.54803807808133e+00, 5.48932769608453e+00, 5.42559001799381e+00, 5.35711962652208e+00, 5.28430384566832e+00, 5.20761319444617e+00, 5.12758902119748e+00, 5.04482899853066e+00, 4.95997125327611e+00, 4.87367793477921e+00, 4.78661899080981e+00, 4.69945683289386e+00, 4.61283244646299e+00, 4.52735335261539e+00, 4.44358367384756e+00, 4.36203640974900e+00, 4.28316790059479e+00, 4.20737435327453e+00, 4.13499022769278e+00, 4.06628823240416e+00, 4.00148065353295e+00, 3.94072173744242e+00, 3.88411086101382e+00, 3.83169624943217e+00, 3.78347903585443e+00, 3.73941749638568e+00, 3.69943133402774e+00, 3.66340592386506e+00, 3.63119646648741e+00, 3.60263202588052e+00, 3.57751945066674e+00, 3.55564719307310e+00, 3.53678904822549e+00, 3.52070783758830e+00, 3.50715905523829e+00, 3.49589448513534e+00, 3.48666578290519e+00, 3.47922799838335e+00, 3.47334299699659e+00, 3.46878272077286e+00, 3.46533221516145e+00, 3.46279233753114e+00, 3.46098205851379e+00, 3.45974026914020e+00, 3.45892701529283e+00, 3.45842409608017e+00, 3.45813498344138e+00, 3.45798404522104e+00, 3.45791508135923e+00, 3.45788921080485e+00, 3.45788217340294e+00, 3.45788113469774e+00, 3.45788105639147e+00, 3.45787967371155e+00, 3.45787030586015e+00, 3.45783586877559e+00, 3.45774407313510e+00, 3.45754318561129e+00, 3.45715848109225e+00, 3.45648953654094e+00, 3.45540850494439e+00, 3.45375949314621e+00, 3.45135914878882e+00, 3.44799853721473e+00, 3.44344635697259e+00, 3.43745350082875e+00, 3.42975891696193e+00, 3.42009666262440e+00, 3.40820397195024e+00, 3.39383008463571e+00, 3.37674550864267e+00, 3.35675132517034e+00, 3.33368809603176e+00, 3.30744391023200e+00, 3.27796111454612e+00, 3.24524131618091e+00, 3.20934832445008e+00, 3.17040880885052e+00, 3.12861058488012e+00, 3.08419858486585e+00, 3.03746871543491e+00, 2.98875993226432e+00, 2.93844496421224e+00, 2.88692018392828e+00, 2.83459514595944e+00, 2.78188229636918e+00, 2.72918730456468e+00, 2.67690038650547e+00, 2.62538888904481e+00, 2.57499129881619e+00, 2.52601273606341e+00, 2.47872190251005e+00, 2.43334937860681e+00, 2.39008711233501e+00, 2.34908890962927e+00, 2.31047172376007e+00, 2.27431754457126e+00, 2.24067570442635e+00, 2.20956544203493e+00, 2.18097859425451e+00, 2.15488231631467e+00, 2.13122176021079e+00, 2.10992266748599e+00, 2.09089385511074e+00, 2.07402959104095e+00, 2.05921186903073e+00, 2.04631260043162e+00, 2.03519574425896e+00, 2.02571939614288e+00, 2.01773785243262e+00, 2.01110365833355e+00, 2.00566963930220e+00, 2.00129090388822e+00, 1.99782679477640e+00, 1.99514275398012e+00, 1.99311205900194e+00, 1.99161738025106e+00, 1.99055210686423e+00, 1.98982138883625e+00, 1.98934284821917e+00, 1.98904692093215e+00, 1.98887680291276e+00, 1.98878798910391e+00, 1.98874741005540e+00, 1.98873218756894e+00, 1.98872804669443e+00, 1.98872743551041e+00, 1.98872738918351e+00, 1.98872656784810e+00, 1.98872100318572e+00, 1.98870054698531e+00, 1.98864601893513e+00, 1.98852668902838e+00, 1.98829817099958e+00, 1.98790081696481e+00, 1.98725869655914e+00, 1.98627923555169e+00, 1.98485357821366e+00, 1.98285772337431e+00, 1.98015446480278e+00, 1.97659614117244e+00, 1.97202816876035e+00, 1.96629329138219e+00, 1.95923643810673e+00, 1.95071003252032e+00, 1.94057955144262e+00, 1.92872909076448e+00, 1.91506666677635e+00, 1.89952896813277e+00, 1.88208528064514e+00, 1.86274033680399e+00, 1.84153589419412e+00, 1.81855091883310e+00, 1.79390033520760e+00, 1.76773239650171e+00, 1.74022481714054e+00, 1.71157988636489e+00, 1.68201883860574e+00, 1.65177578894685e+00, 1.62109154811239e+00, 1.59020761258338e+00, 1.55936058580549e+00, 1.52877723207459e+00, 1.49867030252013e+00, 1.46923520938778e+00, 1.44064756628764e+00, 1.41306156240401e+00, 1.38660910031443e+00, 1.36139960086651e+00, 1.33752036400753e+00, 1.31503737011427e+00, 1.29399641025523e+00, 1.27442444377110e+00, 1.25633109548467e+00, 1.23971022089411e+00, 1.22454148431810e+00, 1.21079191096493e+00, 1.19841738840961e+00, 1.18736410540498e+00, 1.17756992596249e+00, 1.16896570403449e+00, 1.16147654888188e+00, 1.15502305338454e+00, 1.14952249732562e+00, 1.14489003531028e+00, 1.14103987481133e+00, 1.13788644430163e+00, 1.13534554504540e+00, 1.13333547344911e+00, 1.13177809453435e+00, 1.13059984169642e+00, 1.12973261402000e+00, 1.12911454049741e+00, 1.12869058084902e+00, 1.12841293539489e+00, 1.12824124148489e+00, 1.12814254105093e+00, 1.12809101241933e+00, 1.12806746899346e+00, 1.12805863711139e+00, 1.12805623463194e+00, 1.12805588003124e+00, 1.12805585298243e+00, 1.12805537118095e+00, 1.12805210690856e+00, 1.12804010715216e+00, 1.12800812062358e+00, 1.12793812100045e+00, 1.12780407144218e+00, 1.12757098346248e+00, 1.12719431931045e+00, 1.12661978224632e+00, 1.12578353291687e+00, 1.12461286167468e+00, 1.12302733534929e+00, 1.12094042197214e+00, 1.11826157789486e+00, 1.11489875872037e+00, 1.11076128923594e+00, 1.10576299959928e+00, 1.09982550762624e+00, 1.09288150301855e+00, 1.08487787193054e+00, 1.07577849249388e+00, 1.06556653628425e+00, 1.05424612861889e+00, 1.04184325189584e+00, 1.02840581910996e+00, 1.01400289574785e+00, 9.98723102761986e-01, 9.82672285892249e-01, 9.65970581997937e-01, 9.48749046917911e-01, 9.31146028794089e-01, 9.13303474713132e-01, 8.95363347713143e-01, 8.77464308036835e-01, 8.59738780442427e-01, 8.42310492393551e-01, 8.25292529954922e-01, 8.08785922630705e-01, 7.92878737773407e-01, 7.77645641183454e-01, 7.63147863793876e-01, 7.49433504763601e-01, 7.36538098158334e-01, 7.24485372559485e-01, 7.13288139114653e-01, 7.02949252410647e-01, 6.93462598901007e-01, 6.84814078423997e-01, 6.76982554780750e-01, 6.69940760810462e-01, 6.63656151490059e-01, 6.58091705067207e-01, 6.53206677001918e-01, 6.48957314537157e-01, 6.45297541119277e-01, 6.42179619773628e-01, 6.39554803093390e-01, 6.37373974948403e-01, 6.35588285633958e-01, 6.34149778271194e-01, 6.33012000179605e-01, 6.32130589035668e-01, 6.31463820266148e-01, 6.30973099637714e-01, 6.30623383668663e-01, 6.30383510490873e-01, 6.30226425208079e-01, 6.30129286575965e-01, 6.30073445801060e-01, 6.30044293129330e-01, 6.30030973317147e-01, 6.30025976641747e-01, 6.30024617429228e-01, 6.30024416812475e-01, 6.30024401401542e-01, 6.30024125481185e-01, 6.30022256082754e-01, 6.30015384012538e-01, 6.29997065874008e-01, 6.29956978484756e-01, 6.29880211736989e-01, 6.29746730610293e-01, 6.29531034968792e-01, 6.29202043233718e-01, 6.28723222639053e-01, 6.28052984075396e-01, 6.27145352993281e-01, 6.25950919032206e-01, 6.24418055709114e-01, 6.22494387687922e-01, 6.20128467313322e-01, 6.17271605148929e-01, 6.13879782603306e-01, 6.09915560129011e-01, 6.05349883922216e-01, 6.00163689457531e-01, 5.94349203118242e-01, 5.87910854493032e-01, 5.80865731525904e-01, 5.73243537472450e-01, 5.65086040347268e-01, 5.56446039253345e-01, 5.47385904266264e-01, 5.37975774073161e-01, 5.28291515567304e-01, 5.18412560267663e-01, 5.08419733140229e-01, 4.98393180740962e-01, 4.88410489212555e-01, 4.78545060906242e-01, 4.68864793959511e-01, 4.59431084670484e-01, 4.50298150227084e-01, 4.41512650926967e-01, 4.33113577418666e-01, 4.25132359982030e-01, 4.17593153144381e-01, 4.10513249297267e-01, 4.03903578500334e-01, 3.97769257332101e-01, 3.92110156522321e-01, 3.86921464358039e-01, 3.82194229859948e-01, 3.77915876026051e-01, 3.74070678768248e-01, 3.70640211395269e-01, 3.67603757603291e-01, 3.64938697966835e-01, 3.62620875967579e-01, 3.60624949767697e-01, 3.58924735346491e-01, 3.57493545409014e-01, 3.56304526772760e-01, 3.55330996891013e-01, 3.54546777931475e-01, 3.53926524566660e-01, 3.53446039517507e-01, 3.53082569108807e-01, 3.52815069798453e-01, 3.52624435979990e-01, 3.52493679414437e-01, 3.52408051465531e-01, 3.52355100863790e-01, 3.52324661912974e-01, 3.52308770729863e-01, 3.52301510075179e-01, 3.52298786378512e-01, 3.52298045469368e-01, 3.52297936112834e-01, 3.52297927648166e-01, 3.52297775261093e-01, 3.52296742817714e-01, 3.52292947467433e-01, 3.52282830617057e-01, 3.52260690949748e-01, 3.52218293950372e-01, 3.52144574949918e-01, 3.52025451582605e-01, 3.51843760172741e-01, 3.51579328715530e-01, 3.51209196508350e-01, 3.50707986830002e-01, 3.50048434096280e-01, 3.49202060489134e-01, 3.48139989193953e-01, 3.46833872351979e-01, 3.45256902194038e-01, 3.43384864398047e-01, 3.41197184565488e-01, 3.38677913017861e-01, 3.35816591004757e-01, 3.32608943776705e-01, 3.29057353254768e-01, 3.25171075072788e-01, 3.20966180761346e-01, 3.16465224345222e-01, 3.11696651746928e-01, 3.06693989051041e-01, 3.01494859945379e-01, 2.96139892022524e-01, 2.90671575279504e-01, 2.85133134053041e-01, 2.79567466483101e-01, 2.74016194686936e-01, 2.68518855755858e-01, 2.63112250118198e-01, 2.57829951193756e-01, 2.52701969690179e-01, 2.47754557995577e-01, 2.43010135099567e-01, 2.38487310150063e-01, 2.34200982679084e-01, 2.30162499122216e-01, 2.26379847901370e-01, 2.22857878507952e-01, 2.19598533295041e-01, 2.16601083788944e-01, 2.13862366086744e-01, 2.11377012224460e-01, 2.09137676244808e-01, 2.07135255063662e-01, 2.05359105161207e-01, 2.03797256652464e-01, 2.02436626466181e-01, 2.01263232225784e-01, 2.00262408013771e-01, 1.99419022552974e-01, 1.98717699494338e-01, 1.98143038519646e-01, 1.97679834916684e-01, 1.97313294240260e-01, 1.97029237728160e-01, 1.96814293378324e-01, 1.96656067100932e-01, 1.96543288199379e-01, 1.96465923656859e-01, 1.96415256322851e-01, 1.96383923088050e-01, 1.96365910457274e-01, 1.96356506492421e-01, 1.96352209800651e-01, 1.96350597972540e-01, 1.96350159517095e-01, 1.96350094802000e-01, 1.96350089752187e-01, 1.96349998316734e-01, 1.96349378828671e-01, 1.96347101533789e-01, 1.96341031169453e-01, 1.96327746668771e-01, 1.96302306575944e-01, 1.96258070155821e-01, 1.96186583383877e-01, 1.96077538040356e-01, 1.95918810902981e-01, 1.95696588403806e-01, 1.95395579940495e-01, 1.94999320191101e-01, 1.94490557206777e-01, 1.93851718771384e-01, 1.93065444674232e-01, 1.92115167428838e-01, 1.90985719025235e-01, 1.89663937069739e-01, 1.88139240744775e-01, 1.86404145979267e-01, 1.84454690478503e-01, 1.82290743020445e-01, 1.79916177549771e-01, 1.77338900644259e-01, 1.74570730134334e-01, 1.71627132085050e-01, 1.68526832016454e-01, 1.65291323278836e-01, 1.61944300295374e-01, 1.58511046642971e-01, 1.55017807698314e-01, 1.51491175149673e-01, 1.47957506610723e-01, 1.44442398538914e-01, 1.40970225344802e-01, 1.37563752552296e-01, 1.34243827466065e-01, 1.31029147118340e-01, 1.27936100232375e-01, 1.24978677501304e-01, 1.22168442678387e-01, 1.19514555890163e-01, 1.17023840236044e-01, 1.14700882980074e-01, 1.12548163242280e-01, 1.10566198863609e-01, 1.08753705992754e-01, 1.07107765945472e-01, 1.05623995040588e-01, 1.04296714359748e-01, 1.03119117567573e-01, 1.02083435933110e-01, 1.01181100433962e-01, 1.00402901270201e-01, 9.97391452805374e-02, 9.91798116679803e-02, 9.87147061307354e-02, 9.83336129957308e-02, 9.80264443197143e-02, 9.77833842305671e-02, 9.75950260966556e-02, 9.74524995278045e-02, 9.73475837919295e-02, 9.72728040429871e-02, 9.72215068186983e-02, 9.71879116327443e-02, 9.71671360916844e-02, 9.71551928253547e-02, 9.71489575401989e-02, 9.71461086270438e-02, 9.71450399077156e-02, 9.71447491908179e-02, 9.71447062818385e-02, 9.71447029048562e-02, 9.71446413978470e-02, 9.71442246796101e-02, 9.71426927852264e-02, 9.71386093663114e-02, 9.71296731527360e-02, 9.71125601863463e-02, 9.70828035742954e-02, 9.70347169793797e-02, 9.69613674510157e-02, 9.68546023896550e-02, 9.67051343766343e-02, 9.65026861679803e-02, 9.62361962936455e-02, 9.58940833417756e-02, 9.54645642004266e-02, 9.49360183182506e-02, 9.42973866366685e-02, 9.35385904979285e-02, 9.26509528429307e-02, 9.16276017837276e-02, 9.04638355075917e-02, 8.91574277427765e-02, 8.77088548927532e-02, 8.61214294288999e-02, 8.44013290049759e-02, 8.25575166649202e-02, 8.06015538707214e-02, 7.85473143981363e-02, 7.64106129331736e-02, 7.42087671739142e-02, 7.19601160057124e-02, 6.96835184977829e-02, 6.73978585155197e-02, 6.51215775632465e-02, 6.28722543314666e-02, 6.06662446216717e-02, 5.85183907900273e-02, 5.64418062316463e-02, 5.44477368636649e-02, 5.25454974648341e-02, 5.07424756974331e-02, 4.90441924136888e-02, 4.74544046730614e-02, 4.59752387316712e-02, 4.46073427758904e-02, 4.33500514928937e-02, 4.22015553579342e-02, 4.11590675995654e-02, 4.02189825778976e-02, 3.93770206640482e-02, 3.86283578574904e-02, 3.79677401043494e-02, 3.73895842853642e-02, 3.68880678493753e-02, 3.64572093725712e-02, 3.60909417676822e-02, 3.57831800338175e-02, 3.55278853201639e-02, 3.53191265934214e-02, 3.51383290374088e-02, 3.50057489141733e-02, 3.49031128407518e-02, 3.48255120926098e-02, 3.47684231965104e-02, 3.47277501652030e-02, 3.46998578867783e-02, 3.46815946421373e-02, 3.46703019875889e-02, 3.46638106770456e-02, 3.46604218729283e-02, 3.46588735562818e-02, 3.46582927382997e-02, 3.46581347425712e-02, 3.46581114227718e-02, 3.46581095713532e-02, 3.46580756418291e-02, 3.46578457647979e-02, 3.46570007201734e-02, 3.46547482005661e-02, 3.46498189162219e-02, 3.46403798956203e-02, 3.46239689537526e-02, 3.45974541582450e-02, 3.45570219296718e-02, 3.44981971020673e-02, 3.44158977523392e-02, 3.43045267933491e-02, 3.41581011324052e-02, 3.39704175757873e-02, 3.37352526098442e-02, 3.34465907760233e-02, 3.30988737267290e-02, 3.26872594151979e-02, 3.22078785228683e-02, 3.16580734801875e-02, 3.10366046173129e-02, 3.03438083706880e-02, 2.95816942531815e-02, 2.87539704972772e-02, 2.78659927269445e-02, 2.69246353248889e-02, 2.59380907929341e-02, 2.49156077655641e-02, 2.38671828955290e-02, 2.28032251647505e-02, 2.17342129618316e-02, 2.06703642548689e-02, 1.96213382588155e-02, 1.85959833367377e-02, 1.76021410685507e-02, 1.66465113260870e-02, 1.57345785540457e-02, 1.48705955883202e-02, 1.40576181823917e-02, 1.32975808997742e-02, 1.25914034170478e-02, 1.19391159867490e-02, 1.13399939153851e-02, 1.07926928932471e-02, 1.02953790284630e-02, 9.84584896150409e-03, 9.44163657491970e-03, 9.08010402015691e-03, 8.75851628536850e-03, 8.47410007284566e-03, 8.22408885959192e-03, 8.00575642605850e-03, 7.81644100482026e-03, 7.65356190656519e-03, 7.51463030219668e-03, 7.39725581621542e-03, 7.29915057353026e-03, 7.21813219726641e-03, 7.15212694510114e-03, 7.09917375435509e-03, 7.05742953583410e-03, 7.02517567538052e-03, 7.00082539860603e-03, 6.98293142231421e-03, 6.97019315566649e-03, 6.96146262190211e-03, 6.95574825361689e-03, 6.95221577458938e-03, 6.95018551452620e-03, 6.94912569959172e-03, 6.94864149878072e-03, 6.94845986441806e-03, 6.94841045603312e-03, 6.94840316344880e-03, 6.94840257916951e-03, 6.94839180422946e-03, 6.94831880277036e-03, 6.94805044659574e-03, 6.94733514554002e-03, 6.94576991935492e-03, 6.94277306240565e-03, 6.93756381294008e-03, 6.92915047957854e-03, 6.91632850217161e-03, 6.89768989639106e-03, 6.87164539057505e-03, 6.83646025800044e-03, 6.79030432459626e-03, 6.73131586452769e-03, 6.65767808924045e-03, 6.56770574006771e-03, 6.45993799935241e-03, 6.33323267054209e-03, 6.18685548640485e-03, 6.02055765328657e-03, 5.83463450952593e-03, 5.62995868283167e-03, 5.40798256848518e-03, 5.17070739869041e-03, 4.92061941181987e-03, 4.66059707022737e-03, 4.39379608988588e-03, 4.12352061196941e-03, 3.85308918038163e-03, 3.58570397344238e-03, 3.32433167747581e-03, 3.07160435568221e-03, 2.82974751058873e-03, 2.60053917007497e-03, 2.38529873794846e-03, 2.18489995366156e-03, 1.99980113649190e-03, 1.83078595713937e-03, 1.67618679100201e-03, 1.53562867930646e-03, 1.40970914260663e-03, 1.29695414308628e-03, 1.19699127554895e-03, 1.10795678008913e-03, 1.02925793129316e-03, 9.59900834473615e-04, 8.98968972919079e-04, 8.45643150370279e-04, 7.99212574889907e-04, 7.59011024044298e-04, 7.24327657948720e-04, 6.94545778785027e-04, 6.69165352391587e-04, 6.47662528578508e-04, 6.29529142585739e-04, 6.14337188039826e-04, 6.01721819478485e-04, 5.91360960029874e-04, 5.82961788353487e-04, 5.76253526247683e-04, 5.70985575407690e-04, 5.66928062653460e-04, 5.63872356906359e-04, 5.61631042714339e-04, 5.60037703110083e-04, 5.58946732780896e-04, 5.58233141603637e-04, 5.57792208866204e-04, 5.57538853556139e-04, 5.57406619179801e-04, 5.57346209291154e-04, 5.57323548946442e-04, 5.57317384902944e-04, 5.57316475109815e-04, 5.57316401469166e-04, 5.57315034626527e-04, 5.57305774094630e-04, 5.57271732475328e-04, 5.57180998200124e-04, 5.56982469944955e-04, 5.56602423614549e-04, 5.55942017271605e-04, 5.54875960062873e-04, 5.53252590023590e-04, 5.50895616896696e-04, 5.47607777235726e-04, 5.43176606996099e-04, 5.37382436135658e-04, 5.30008528926373e-04, 5.20853052250914e-04, 5.09742371463062e-04, 4.96545100716617e-04, 4.81186180067539e-04, 4.63659976688291e-04, 4.44041862550648e-04, 4.22499425667411e-04, 3.99294964509522e-04, 3.74730821476987e-04, 3.49126455375881e-04, 3.22900728875197e-04, 2.96471586622715e-04, 2.70208032759456e-04, 2.44476885725931e-04, 2.19629146769762e-04, 1.95992901922182e-04, 1.73878029201546e-04, 1.53494041597924e-04, 1.34977302181483e-04, 1.18312870782233e-04, 1.03361062826934e-04, 8.99437839620987e-05, 7.78878865759225e-05, 6.72062137141438e-05, 5.81407779236542e-05, 5.02205941046441e-05, 4.32306105864200e-05, 3.70207197590844e-05, 3.13313494189060e-05, 2.58101056687110e-05, 2.03227696999875e-05, 1.49856463451670e-05, 1.00429302136718e-05, 5.78053668220946e-06, 2.38623120196735e-06, 1.33513958644471e-06, 1.14772394965712e-06, 4.83560199882049e-07, 1.16669963290406e-06, 1.35394485312190e-06, 1.26816475187752e-06, 1.08814195698415e-06, 9.19419207434657e-07, 8.08051396461451e-07, 7.61685910788401e-07, 7.44873986491257e-07, 7.14630004949868e-07, 6.54980681247253e-07, 5.69403070471708e-07, 4.69283776283191e-07, 3.66877614510486e-07, -2.84217094304040e-14, -1.42108547152020e-14, -2.84217094304040e-14, -1.42108547152020e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -4.26325641456060e-14],
[ 1.05000000000000e+02, 1.04999999866103e+02, 1.04999995725727e+02, 1.04999967674027e+02, 1.04999864552263e+02, 1.04999589661789e+02, 1.04998988041712e+02, 1.04997835756189e+02, 1.04995831574403e+02, 1.04992591350932e+02, 1.04987645329437e+02, 1.04980438502367e+02, 1.04970334068517e+02, 1.04956619943958e+02, 1.04938518205221e+02, 1.04915197280864e+02, 1.04885786660750e+02, 1.04849393860912e+02, 1.04805123361090e+02, 1.04752097213855e+02, 1.04689476997598e+02, 1.04616486739126e+02, 1.04532436356307e+02, 1.04436745064531e+02, 1.04328964059717e+02, 1.04208797652618e+02, 1.04076121911541e+02, 1.03930999805458e+02, 1.03773691857407e+02, 1.03604661440068e+02, 1.03424574076166e+02, 1.03234290430728e+02, 1.03034853066442e+02, 1.02827467430358e+02, 1.02613477899288e+02, 1.02394339987746e+02, 1.02171589986499e+02, 1.01946813341462e+02, 1.01721613011469e+02, 1.01497578884615e+02, 1.01276259120597e+02, 1.01059134056165e+02, 1.00847593092664e+02, 1.00642914799603e+02, 1.00446250326668e+02, 1.00258610119154e+02, 1.00080853872459e+02, 9.99136836298499e+01, 9.97576399134454e+01, 9.96131007715209e+01, 9.94802836186943e+01, 9.93592497347148e+01, 9.92499112704836e+01, 9.91520405863379e+01, 9.90652817185404e+01, 9.89891637369562e+01, 9.89231157219569e+01, 9.88664830536071e+01, 9.88185446730726e+01, 9.87785309468217e+01, 9.87456417404321e+01, 9.87190642924783e+01, 9.86979904718053e+01, 9.86816330050336e+01, 9.86692402766519e+01, 9.86601093323305e+01, 9.86535967573633e+01, 9.86491271559681e+01, 9.86461990223621e+01, 9.86443878691847e+01, 9.86433465604268e+01, 9.86428028814988e+01, 9.86425544651026e+01, 9.86424612747265e+01, 9.86424359246083e+01, 9.86424321829811e+01, 9.86424316921586e+01, 9.86424202565822e+01, 9.86423427788582e+01, 9.86420579623809e+01, 9.86412987449322e+01, 9.86396372095631e+01, 9.86364551454395e+01, 9.86309214305518e+01, 9.86219772645651e+01, 9.86083301133583e+01, 9.85884570332181e+01, 9.85606178167630e+01, 9.85228781385998e+01, 9.84731425718913e+01, 9.84091969961908e+01, 9.83287595263336e+01, 9.82295386737255e+01, 9.81092970263078e+01, 9.79659183334624e+01, 9.77974755484750e+01, 9.76022971614090e+01, 9.73790290952203e+01, 9.71266895695449e+01, 9.68447146622286e+01, 9.65329927784134e+01, 9.61918867866396e+01, 9.58222430912974e+01, 9.54253872889400e+01, 9.50031062865881e+01, 9.45576169483993e+01, 9.40915216975405e+01, 9.36077522528894e+01, 9.31095038696597e+01, 9.26001637918017e+01, 9.20832385250145e+01, 9.15622844202104e+01, 9.10408447670433e+01, 9.05223946153833e+01, 9.00102927265347e+01, 8.95077391042928e+01, 8.90177366252927e+01, 8.85430560252556e+01, 8.80862043668514e+01, 8.76493977162379e+01, 8.72345389495539e+01, 8.68432014655231e+01, 8.64766192517897e+01, 8.61356833884898e+01, 8.58209447679343e+01, 8.55326226026429e+01, 8.52706181831949e+01, 8.50345333091902e+01, 8.48236928218709e+01, 8.46371706896301e+01, 8.44738191187528e+01, 8.43323001699776e+01, 8.42111193520810e+01, 8.41086606369556e+01, 8.40232223003130e+01, 8.39530529441856e+01, 8.38963870090811e+01, 8.38514790427196e+01, 8.38166359663696e+01, 8.37902465757538e+01, 8.37708075368506e+01, 8.37569451913672e+01, 8.37474325736739e+01, 8.37412011595464e+01, 8.37373470136622e+01, 8.37351311716611e+01, 8.37339742760900e+01, 8.37334456749309e+01, 8.37332473774133e+01, 8.37331934355886e+01, 8.37331854738849e+01, 8.37331847710180e+01, 8.37331709986876e+01, 8.37330776891122e+01, 8.37327346730900e+01, 8.37318203171404e+01, 8.37298192655587e+01, 8.37259869895868e+01, 8.37193225581452e+01, 8.37085508734887e+01, 8.36921154166875e+01, 8.36681823174841e+01, 8.36346562944243e+01, 8.35892086975001e+01, 8.35293175219001e+01, 8.34523188468850e+01, 8.33554686935169e+01, 8.32360138021474e+01, 8.30912693272541e+01, 8.29187009641243e+01, 8.27160085968567e+01, 8.24812082322958e+01, 8.22127088023304e+01, 8.19093804157979e+01, 8.15706108502366e+01, 8.11963475076792e+01, 8.07871227121652e+01, 8.03440610683174e+01, 7.98688685686630e+01, 7.93638041413408e+01, 7.88316352580854e+01, 7.82755799633114e+01, 7.76992381550534e+01, 7.71065151191746e+01, 7.65015402306855e+01, 7.58885834898581e+01, 7.52719722773434e+01, 7.46560104836760e+01, 7.40449020161805e+01, 7.34426805548859e+01, 7.28531472211727e+01, 7.22798174606159e+01, 7.17258779189031e+01, 7.11941534839183e+01, 7.06870841019253e+01, 7.02067105622158e+01, 6.97546682348431e+01, 6.93321877238839e+01, 6.89401015023171e+01, 6.85788557518238e+01, 6.82485267854322e+01, 6.79488415527636e+01, 6.76792018086201e+01, 6.74387115702339e+01, 6.72262075048236e+01, 6.70402918839285e+01, 6.68793677182469e+01, 6.67416756483454e+01, 6.66253321141533e+01, 6.65283682621476e+01, 6.64487689779037e+01, 6.63845113595106e+01, 6.63336018820799e+01, 6.62941114539028e+01, 6.62642075393169e+01, 6.62421825295747e+01, 6.62264775864983e+01, 6.62157012673250e+01, 6.62086423625775e+01, 6.62042765384147e+01, 6.62017665640351e+01, 6.62004561140864e+01, 6.61998573547719e+01, 6.61996327388490e+01, 6.61995716378121e+01, 6.61995626194249e+01, 6.61995618960387e+01, 6.61995485458941e+01, 6.61994580966980e+01, 6.61991255958536e+01, 6.61982392704351e+01, 6.61962995676923e+01, 6.61925848001243e+01, 6.61861247732276e+01, 6.61756836121955e+01, 6.61597528157208e+01, 6.61365553454416e+01, 6.61040613012276e+01, 6.60600154278779e+01, 6.60019763421683e+01, 6.59273669591351e+01, 6.58335351383007e+01, 6.57178230781884e+01, 6.55776434845846e+01, 6.54105600576430e+01, 6.52143694254234e+01, 6.49871813405022e+01, 6.47274937933142e+01, 6.44342597135484e+01, 6.41069421470918e+01, 6.37455552096411e+01, 6.33506887086224e+01, 6.29235150557851e+01, 6.24657779174411e+01, 6.19797629188604e+01, 6.14682515865618e+01, 6.09344605312214e+01, 6.03819685958824e+01, 5.98146352637947e+01, 5.92365139780176e+01, 5.86517641183808e+01, 5.80645651828490e+01, 5.74790362438638e+01, 5.68991630585524e+01, 5.63287344073268e+01, 5.57712884376621e+01, 5.52300691054136e+01, 5.47079923040762e+01, 5.42076209728629e+01, 5.37311483524021e+01, 5.32803885590483e+01, 5.28567737152247e+01, 5.24613569549737e+01, 5.20948206925556e+01, 5.17574895894938e+01, 5.14493476875807e+01, 5.11700592025448e+01, 5.09189925034464e+01, 5.06952468386090e+01, 5.04976814062827e+01, 5.03249463999202e+01, 5.01755156759581e+01, 5.00477206902295e+01, 4.99397853248806e+01, 4.98498611820663e+01, 4.97760628585155e+01, 4.97165026438599e+01, 4.96693240148930e+01, 4.96327332379220e+01, 4.96050283521173e+01, 4.95846247969672e+01, 4.95700769731601e+01, 4.95600950920941e+01, 4.95535567750481e+01, 4.95495130056899e+01, 4.95471882126438e+01, 4.95459744532667e+01, 4.95454198747799e+01, 4.95452118329516e+01, 4.95451552405025e+01, 4.95451468875746e+01, 4.95451462433165e+01, 4.95451346745077e+01, 4.95450562941308e+01, 4.95447681596423e+01, 4.95440000994095e+01, 4.95423192210391e+01, 4.95391001480806e+01, 4.95335021851455e+01, 4.95244544746208e+01, 4.95106500487990e+01, 4.94905494943314e+01, 4.94623947244744e+01, 4.94242330913743e+01, 4.93739517600441e+01, 4.93093219059914e+01, 4.92280518938975e+01, 4.91278481572184e+01, 4.90064820485040e+01, 4.88618604962597e+01, 4.86920979215001e+01, 4.84955865742836e+01, 4.82710622850558e+01, 4.80176626193253e+01, 4.77349745985397e+01, 4.74230695129312e+01, 4.70825228962924e+01, 4.67144184353736e+01, 4.63203354097278e+01, 4.59023201473587e+01, 4.54628428668660e+01, 4.50047420726322e+01, 4.45311592847919e+01, 4.40454672419444e+01, 4.35511947716771e+01, 4.30519513006503e+01, 4.25513535515339e+01, 4.20529564640941e+01, 4.15601898922178e+01, 4.10763022333196e+01, 4.06043118472956e+01, 4.01469668790805e+01, 3.97067138584173e+01, 3.92856751769317e+01, 3.88856352339128e+01, 3.85080347267969e+01, 3.81539722837340e+01, 3.78242124340219e+01, 3.75191988116283e+01, 3.72390714903343e+01, 3.69836874405484e+01, 3.67526432500566e+01, 3.65452994323705e+01, 3.63608058277414e+01, 3.61981277605084e+01, 3.60560727373633e+01, 3.59333175471813e+01, 3.58284356535570e+01, 3.57399247602415e+01, 3.56662343847294e+01, 3.56057932058233e+01, 3.55570358677544e+01, 3.55184288374385e+01, 3.54884948335365e+01, 3.54658352862011e+01, 3.54491502532757e+01, 3.54372552187127e+01, 3.54290942358589e+01, 3.54237489526931e+01, 3.54204431655493e+01, 3.54185426867157e+01, 3.54175504714418e+01, 3.54170971214468e+01, 3.54169270544321e+01, 3.54168807921014e+01, 3.54168739638790e+01, 3.54168734468666e+01, 3.54168642880444e+01, 3.54168022356958e+01, 3.54165741249974e+01, 3.54159660679235e+01, 3.54146353615462e+01, 3.54120869443082e+01, 3.54076553651489e+01, 3.54004931304948e+01, 3.53895661770422e+01, 3.53736570782420e+01, 3.53513764218432e+01, 3.53211825867810e+01, 3.52814098935889e+01, 3.52303048003828e+01, 3.51660694688328e+01, 3.50869116410776e+01, 3.49910993669879e+01, 3.48770187273794e+01, 3.47432323454350e+01, 3.45885362029270e+01, 3.44120121180197e+01, 3.42130732322118e+01, 3.39915000218256e+01, 3.37474647077211e+01, 3.34815424816289e+01, 3.31947086742111e+01, 3.28883218124421e+01, 3.25640933854877e+01, 3.22240459765566e+01, 3.18704621339223e+01, 3.15058268638060e+01, 3.11327668678769e+01, 3.07539895899907e+01, 3.03722247953041e+01, 2.99901708393195e+01, 2.96104470888524e+01, 2.92355532419461e+01, 2.88678356633029e+01, 2.85094603817964e+01, 2.81623921212536e+01, 2.78283786451106e+01, 2.75089397448856e+01, 2.72053603287488e+01, 2.69186872079041e+01, 2.66497292895229e+01, 2.63990609442513e+01, 2.61670283256708e+01, 2.59537583960369e+01, 2.57591703799588e+01, 2.55829893445590e+01, 2.54247616011995e+01, 2.52838716404612e+01, 2.51595603417557e+01, 2.50509442313526e+01, 2.49570355876247e+01, 2.48767632026052e+01, 2.48089936010203e+01, 2.47525524921285e+01, 2.47062461894379e+01, 2.46688826843638e+01, 2.46392920090932e+01, 2.46163454787363e+01, 2.45989733703564e+01, 2.45861805828272e+01, 2.45770598312877e+01, 2.45708019659393e+01, 2.45667030674815e+01, 2.45641680587171e+01, 2.45627106796943e+01, 2.45619497962976e+01, 2.45616021423421e+01, 2.45614717253334e+01, 2.45614362487442e+01, 2.45614310124739e+01, 2.45614306193855e+01, 2.45614237005830e+01, 2.45613768246851e+01, 2.45612045040276e+01, 2.45607451612034e+01, 2.45597399046315e+01, 2.45578147332255e+01, 2.45544668915326e+01, 2.45490560169526e+01, 2.45408006275615e+01, 2.45287803891562e+01, 2.45119444687035e+01, 2.44891261234482e+01, 2.44590634866218e+01, 2.44204262905336e+01, 2.43718480181912e+01, 2.43119627027927e+01, 2.42394453132357e+01, 2.41530543913201e+01, 2.40516753649034e+01, 2.39343627754679e+01, 2.38003795525026e+01, 2.36492314616160e+01, 2.34806949633138e+01, 2.32948369522087e+01, 2.30920252007075e+01, 2.28729287961948e+01, 2.26385084147976e+01, 2.23899968828080e+01, 2.21288710869938e+01, 2.18568168387575e+01, 2.15756886961751e+01, 2.12874669320182e+01, 2.09942137673502e+01, 2.06980306861911e+01, 2.04010181835008e+01, 2.01052387922131e+01, 1.98126838012929e+01, 1.95252437917586e+01, 1.92446829957349e+01, 1.89726174838821e+01, 1.87104972394754e+01, 1.84595922169059e+01, 1.82209824680438e+01, 1.79955523435730e+01, 1.77839886554696e+01, 1.75867825511882e+01, 1.74042347296988e+01, 1.72364635456587e+01, 1.70834155103249e+01, 1.69448777049864e+01, 1.68204916655218e+01, 1.67097683618437e+01, 1.66121039693565e+01, 1.65267961986405e+01, 1.64530610048897e+01, 1.63900495343005e+01, 1.63368651783081e+01, 1.62925805991505e+01, 1.62562545651535e+01, 1.62269483967352e+01, 1.62037417809858e+01, 1.61857476708658e+01, 1.61721259514608e+01, 1.61620955363665e+01, 1.61549445568169e+01, 1.61500383274373e+01, 1.61468248162389e+01, 1.61448374113305e+01, 1.61436948594440e+01, 1.61430983468843e+01, 1.61428257959385e+01, 1.61427235527614e+01, 1.61426957401469e+01, 1.61426916350636e+01, 1.61426913278308e+01, 1.61426859326691e+01, 1.61426493796634e+01, 1.61425150070816e+01, 1.61421568201411e+01, 1.61413729419982e+01, 1.61398717415695e+01, 1.61372612040220e+01, 1.61330420495376e+01, 1.61266050355958e+01, 1.61172327945882e+01, 1.61041064581651e+01, 1.60863171982755e+01, 1.60628826692234e+01, 1.60327681640014e+01, 1.59949121033990e+01, 1.59482552633450e+01, 1.58917729241317e+01, 1.58245089080194e+01, 1.57456102757979e+01, 1.56543612966115e+01, 1.55502152071128e+01, 1.54328222521528e+01, 1.53020525619931e+01, 1.51580125769331e+01, 1.50010539787895e+01, 1.48317744217729e+01, 1.46510097571405e+01, 1.44598178932592e+01, 1.42594548954128e+01, 1.40513443726110e+01, 1.38370415840164e+01, 1.36181939888722e+01, 1.33965001307719e+01, 1.31736687710869e+01, 1.29513800644952e+01, 1.27312503165331e+01, 1.25148015097545e+01, 1.23034363735937e+01, 1.20984193499182e+01, 1.19008634148728e+01, 1.17117223918542e+01, 1.15317881509684e+01, 1.13616919437828e+01, 1.12019090631095e+01, 1.10527660318298e+01, 1.09144495937214e+01, 1.07870168833009e+01, 1.06704062730666e+01, 1.05644485207341e+01, 1.04688779554415e+01, 1.03833435433649e+01, 1.03074197556132e+01, 1.02406172227964e+01, 1.01823932010049e+01, 1.01321618939755e+01, 1.00893046776043e+01, 1.00531802580605e+01, 1.00231347664545e+01, 9.99851175482706e+00, 9.97866201410905e+00, 9.96295308901820e+00, 9.95077832215688e+00, 9.94156522428906e+00, 9.93478294392932e+00, 9.92994860017511e+00, 9.92663225022677e+00, 9.92446028793434e+00, 9.92311711112576e+00, 9.92234495106693e+00, 9.92194182356754e+00, 9.92175763348445e+00, 9.92168853778142e+00, 9.92166974210846e+00, 9.92166696790773e+00, 9.92166676020318e+00, 9.92166311176942e+00, 9.92163839311178e+00, 9.92154752497858e+00, 9.92130530595000e+00, 9.92077522594428e+00, 9.91976009922389e+00, 9.91799491021268e+00, 9.91514223074851e+00, 9.91079052614695e+00, 9.90445561626987e+00, 9.89558549230375e+00, 9.88356860552922e+00, 9.86774563710397e+00, 9.84742462573070e+00, 9.82189917391769e+00, 9.79046927812756e+00, 9.75246414264754e+00, 9.70726615513534e+00, 9.65433504081960e+00, 9.59323109171068e+00, 9.52363630618576e+00, 9.44537228920370e+00, 9.35841386442671e+00, 9.26289753833323e+00, 9.15912422439432e+00, 9.04755596355058e+00, 8.92880673770510e+00, 8.80362783230609e+00, 8.67288852783300e+00, 8.53755315750372e+00, 8.39865573746030e+00, 8.25727344493167e+00, 8.11450019035640e+00, 7.97142141294334e+00, 7.82909104535092e+00, 7.68851136604043e+00, 7.55061621198320e+00, 7.41625778212675e+00, 7.28619704160209e+00, 7.16109755136006e+00, 7.04152240586258e+00, 6.92793386582417e+00, 6.82069522263835e+00, 6.72007442130143e+00, 6.62624899205923e+00, 6.53931188891978e+00, 6.45927789656582e+00, 6.38609033771706e+00, 6.31962788361976e+00, 6.25971133573900e+00, 6.20611030323140e+00, 6.15854974618438e+00, 6.11671638790652e+00, 6.08026502059064e+00, 6.04882473792381e+00, 6.02200512663887e+00, 5.99940243792614e+00, 5.98060574074495e+00, 5.96520303443646e+00, 5.95278726999646e+00, 5.94296220053725e+00, 5.93534795460263e+00, 5.92958620382468e+00, 5.92534478140109e+00, 5.92232160203054e+00, 5.92024773857408e+00, 5.91888952625391e+00, 5.91804959113544e+00, 5.91756673451836e+00, 5.91731464642478e+00, 5.91719946679426e+00, 5.91715625918130e+00, 5.91714450568588e+00, 5.91714277089541e+00, 5.91714264066702e+00, 5.91714034853831e+00, 5.91712481904113e+00, 5.91706773114915e+00, 5.91691555718975e+00, 5.91658253530962e+00, 5.91594478580541e+00, 5.91483581929126e+00, 5.91304366337269e+00, 5.91030980908407e+00, 5.90633014658124e+00, 5.90075801855221e+00, 5.89320946643532e+00, 5.88327067620202e+00, 5.87050754578162e+00, 5.85447719573769e+00, 5.83474113172713e+00, 5.81087964772512e+00, 5.78250694211309e+00, 5.74928631606147e+00, 5.71094474810313e+00, 5.66728610301099e+00, 5.61820224744194e+00, 5.56368141539514e+00, 5.50381329352766e+00, 5.43879047304947e+00, 5.36890612779479e+00, 5.29454800817221e+00, 5.21618906598199e+00, 5.13437522351825e+00, 5.04971095315585e+00, 4.96284342794695e+00, 4.87444603443254e+00, 4.78520200859134e+00, 4.69578887400040e+00, 4.60686424180049e+00, 4.51905339077732e+00, 4.43293889804995e+00, 4.34905244937221e+00, 4.26786883241238e+00, 4.18980201255613e+00, 4.11520311161362e+00, 4.04436005563842e+00, 3.97749862743640e+00, 3.91478464958999e+00, 3.85632703154749e+00, 3.80218143572986e+00, 3.75235434876829e+00, 3.70680738109976e+00, 3.66546165770676e+00, 3.62820220172954e+00, 3.59488224848387e+00, 3.56532745815359e+00, 3.53934001974281e+00, 3.51670265598081e+00, 3.49718254848584e+00, 3.48053520479873e+00, 3.46650828448216e+00, 3.45484539132671e+00, 3.44528982410876e+00, 3.43758826089804e+00, 3.43149433340184e+00, 3.42677203015478e+00, 3.42319885239114e+00, 3.42056863587197e+00, 3.41869394715475e+00, 3.41740796469368e+00, 3.41656576407139e+00, 3.41604494228113e+00, 3.41574553738960e+00, 3.41558922567188e+00, 3.41551780661153e+00, 3.41549101502694e+00, 3.41548372708243e+00, 3.41548265139768e+00, 3.41548257030355e+00, 3.41548113839799e+00, 3.41547143703579e+00, 3.41543577393067e+00, 3.41534071017333e+00, 3.41513267047515e+00, 3.41473426876134e+00, 3.41404150490851e+00, 3.41292197562558e+00, 3.41121422715698e+00, 3.40872835815198e+00, 3.40524795577137e+00, 3.40053341484983e+00, 3.39432664698182e+00, 3.38635713279957e+00, 3.37634920679378e+00, 3.36403039177567e+00, 3.34914052338942e+00, 3.33144132974405e+00, 3.31072606458976e+00, 3.28682874272300e+00, 3.25963250153701e+00, 3.22907661953890e+00, 3.19516176533520e+00, 3.15795312951430e+00, 3.11758120327173e+00, 3.07424010359914e+00, 3.02818349396232e+00, 2.97971829806684e+00, 2.92919653855689e+00, 2.87700573977010e+00, 2.82355840450592e+00, 2.76928110394476e+00, 2.71460370691828e+00, 2.65994922377465e+00, 2.60572465892244e+00, 2.55231316502415e+00, 2.50006768192870e+00, 2.44930613545516e+00, 2.40030817401534e+00, 2.35331334125760e+00, 2.30852052405851e+00, 2.26608847822304e+00, 2.22613721784940e+00, 2.18875005552526e+00, 2.15397609547459e+00, 2.12183300629998e+00, 2.09230993009001e+00, 2.06537041693603e+00, 2.04095530556633e+00, 2.01898549981928e+00, 1.99936461566439e+00, 1.98198149362939e+00, 1.96671258642724e+00, 1.95342424126002e+00, 1.94197490091240e+00, 1.93221724773106e+00, 1.92400031047417e+00, 1.91717154650166e+00, 1.91157890169974e+00, 1.90707283884878e+00, 1.90350831291698e+00, 1.90074666010820e+00, 1.89865735752541e+00, 1.89711960304975e+00, 1.89602366131332e+00, 1.89527192202773e+00, 1.89477962163141e+00, 1.89447518807734e+00, 1.89430018104933e+00, 1.89420881510007e+00, 1.89416707004040e+00, 1.89415141015627e+00, 1.89414715030027e+00, 1.89414652155487e+00, 1.89414647389685e+00, 1.89414562896174e+00, 1.89413990440839e+00, 1.89411886044071e+00, 1.89406276568648e+00, 1.89394000743019e+00, 1.89370492501962e+00, 1.89329615959403e+00, 1.89263560642226e+00, 1.89162804782437e+00, 1.89016153384938e+00, 1.88810856326239e+00, 1.88532809754918e+00, 1.88166841436181e+00, 1.87697077341649e+00, 1.87107382747875e+00, 1.86381866502299e+00, 1.85505432203612e+00, 1.84464355214752e+00, 1.83246860178723e+00, 1.81843670596530e+00, 1.80248500597013e+00, 1.78458459725995e+00, 1.76474344665624e+00, 1.74300797263362e+00, 1.71946315803918e+00, 1.69423115504142e+00, 1.66746843924253e+00, 1.63936166412938e+00, 1.61012244887307e+00, 1.57998139381888e+00, 1.54918165333255e+00, 1.51797240278154e+00, 1.48660251656730e+00, 1.45531473155494e+00, 1.42434051137698e+00, 1.39389575937716e+00, 1.36417745877971e+00, 1.33536125439298e+00, 1.30759993552229e+00, 1.28102273765292e+00, 1.25573535191195e+00, 1.23182051586154e+00, 1.20933905523718e+00, 1.18833125159205e+00, 1.16881842295679e+00, 1.15080462114710e+00, 1.13427836808933e+00, 1.11921437271417e+00, 1.10557518822736e+00, 1.09331278594149e+00, 1.08237003572276e+00, 1.07268209412611e+00, 1.06417770932921e+00, 1.05678045705196e+00, 1.05040992390856e+00, 1.04498285430952e+00, 1.04041427442520e+00, 1.03661860222546e+00, 1.03351074667941e+00, 1.03100719236900e+00, 1.02902705862982e+00, 1.02749311550889e+00, 1.02633273295146e+00, 1.02547873527597e+00, 1.02487013065085e+00, 1.02445268527221e+00, 1.02417931439575e+00, 1.02401026720891e+00, 1.02391308943862e+00, 1.02386235608229e+00, 1.02383917609453e+00, 1.02383048056204e+00, 1.02382811517404e+00, 1.02382776604799e+00, 1.02382773941679e+00, 1.02382726505381e+00, 1.02382405117902e+00, 1.02381223669653e+00, 1.02378074409505e+00, 1.02371182569738e+00, 1.02357984788055e+00, 1.02335036601033e+00, 1.02297953923819e+00, 1.02241392990958e+00, 1.02159072648969e+00, 1.02043842081524e+00, 1.01887795925263e+00, 1.01682437223754e+00, 1.01418886725389e+00, 1.01088134666371e+00, 1.00681328470388e+00, 1.00190086899730e+00, 9.96068283461454e-01, 9.89250984569381e-01, 9.81398804902071e-01, 9.72478710116505e-01, 9.62477040474866e-01, 9.51401087403511e-01, 9.39279889015253e-01, 9.26164174141832e-01, 9.12125438434614e-01, 8.97254193401395e-01, 8.81657484071980e-01, 8.65455817699626e-01, 8.48779679862758e-01, 8.31765832524610e-01, 8.14553590053578e-01, 7.97281254944949e-01, 7.80082867793453e-01, 7.63085389897796e-01, 7.46406396202886e-01, 7.30152315435717e-01, 7.14417216948661e-01, 6.99282112633384e-01, 6.84814718872261e-01, 6.71069608337476e-01, 6.58088674119782e-01, 6.45901828111832e-01, 6.34527860354822e-01, 6.23975394612955e-01, 6.14243886268440e-01, 6.05324620409831e-01, 5.97201679668075e-01, 5.89852862160470e-01, 5.83250539317376e-01, 5.77362451097571e-01, 5.72152442026137e-01, 5.67581145612365e-01, 5.63606627116513e-01, 5.60184995468418e-01, 5.57270994575575e-01, 5.54818582494917e-01, 5.52781504206138e-01, 5.51113860273247e-01, 5.49770669794711e-01, 5.48708422035673e-01, 5.47885607331381e-01, 5.47263214584234e-01, 5.46805180246494e-01, 5.46478772348934e-01, 5.46254893071406e-01, 5.46108284632709e-01, 5.46017625854759e-01, 5.45965510469969e-01, 5.45938302821654e-01, 5.45925871714502e-01, 5.45921208424303e-01, 5.45919939900742e-01, 5.45919752669505e-01, 5.45919738286784e-01, 5.45919480776377e-01, 5.45917736107754e-01, 5.45911322557956e-01, 5.45894226669844e-01, 5.45856814162320e-01, 5.45785170196680e-01, 5.45660597829411e-01, 5.45459301641188e-01, 5.45152281258609e-01, 5.44705456407144e-01, 5.44080040772911e-01, 5.43233175796587e-01, 5.42118827134686e-01, 5.40688935640006e-01, 5.38894801378788e-01, 5.36688663865533e-01, 5.34025425277378e-01, 5.30864447301269e-01, 5.27171338243434e-01, 5.22919637073187e-01, 5.18092297089254e-01, 5.12682875408757e-01, 5.06696346282411e-01, 5.00149476157560e-01, 4.93070725166177e-01, 4.85499670993789e-01, 4.77485983793841e-01, 4.69088011508930e-01, 4.60371060357588e-01, 4.51405472724261e-01, 4.42264612682990e-01, 4.33022867578813e-01, 4.23753763403951e-01, 4.14528274059634e-01, 4.05413382541937e-01, 3.96470928378747e-01, 3.87756752821858e-01, 3.79320133383231e-01, 3.71203483640485e-01, 3.63442283424192e-01, 3.56065198503742e-01, 3.49094347169000e-01, 3.42545672820705e-01, 3.36429385859418e-01, 3.30750443848672e-01, 3.25509045288101e-01, 3.20701118709735e-01, 3.16318794734528e-01, 3.12350853900369e-01, 3.08783147342851e-01, 3.05598990726679e-01, 3.02779534209350e-01, 3.00304112730018e-01, 2.98150581635568e-01, 2.96295642672746e-01, 2.94715164781451e-01, 2.93384503015631e-01, 2.92278817399065e-01, 2.91373391711105e-01, 2.90643950219291e-01, 2.90066968381993e-01, 2.89619971687401e-01, 2.89281815237317e-01, 2.89032935570859e-01, 2.88855565683448e-01, 2.88733904309083e-01, 2.88654231339905e-01, 2.88604962723639e-01, 2.88576640225642e-01, 2.88561853927916e-01, 2.88555098083080e-01, 2.88552563754536e-01, 2.88551874357793e-01, 2.88551772604379e-01, 2.88551764728240e-01, 2.88551622936012e-01, 2.88550662274062e-01, 2.88547130797383e-01, 2.88537717317311e-01, 2.88517116857946e-01, 2.88477667239917e-01, 2.88409072552668e-01, 2.88298228195615e-01, 2.88129160846864e-01, 2.87883094956214e-01, 2.87538654913376e-01, 2.87072208642414e-01, 2.86458353790636e-01, 2.85670541783063e-01, 2.84681827828919e-01, 2.83465726747352e-01, 2.81997145725200e-01, 2.80253356589100e-01, 2.78214962828812e-01, 2.75866811509729e-01, 2.73198798371624e-01, 2.70206516613456e-01, 2.66891706472535e-01, 2.63262473560118e-01, 2.59333258230356e-01, 2.55124554732305e-01, 2.50662395820214e-01, 2.45977634057851e-01, 2.41105063574722e-01, 2.36082434275033e-01, 2.30949413827432e-01, 2.25746551212595e-01, 2.20514289835435e-01, 2.15292069322928e-01, 2.10117544461824e-01, 2.05025938613417e-01, 2.00049538507841e-01, 1.95217328396978e-01, 1.90554754596462e-01, 1.86083606619519e-01, 1.81821998276149e-01, 1.77784430994720e-01, 1.73981921850213e-01, 1.70422179976256e-01, 1.67109816863274e-01, 1.64046578217537e-01, 1.61231587373294e-01, 1.58661592547844e-01, 1.56331212399209e-01, 1.54233176306747e-01, 1.52358557488157e-01, 1.50696998450258e-01, 1.49236929318803e-01, 1.47965780281652e-01, 1.46870189709801e-01, 1.45936209493684e-01, 1.45149508778729e-01, 1.44495576640480e-01, 1.43959923372933e-01, 1.43528279049505e-01, 1.43186786948178e-01, 1.42922188413721e-01, 1.42721994862711e-01, 1.42574642016569e-01, 1.42469621154888e-01, 1.42397582265474e-01, 1.42350404452941e-01, 1.42321229835801e-01, 1.42304458364350e-01, 1.42295702445409e-01, 1.42291701862391e-01, 1.42290201116680e-01, 1.42289792878501e-01, 1.42289732623439e-01, 1.42289727921636e-01, 1.42289642787702e-01, 1.42289065993467e-01, 1.42286945646688e-01, 1.42281293656737e-01, 1.42268924800675e-01, 1.42245238439685e-01, 1.42204052284086e-01, 1.42137496922572e-01, 1.42035979000397e-01, 1.41888219848255e-01, 1.41681374890894e-01, 1.41401237157012e-01, 1.41032525545015e-01, 1.40559255103980e-01, 1.39965182482243e-01, 1.39234315017106e-01, 1.38351466960160e-01, 1.37302841462215e-01, 1.36076612697039e-01, 1.34663479452740e-01, 1.33057160212928e-01, 1.31254800612069e-01, 1.29257267406160e-01, 1.27069308693251e-01, 1.24699567691906e-01, 1.22160446301791e-01, 1.19467824091799e-01, 1.16640647371469e-01, 1.13700410737337e-01, 1.10670559258310e-01, 1.07575842846487e-01, 1.04441655213421e-01, 1.01293388280283e-01, 9.81558293646572e-02, 9.50526234350235e-02, 9.20058168061786e-02, 8.90354924301420e-02, 8.61595009192655e-02, 8.33932860058582e-02, 8.07497985637013e-02, 7.82394897567116e-02, 7.58703714364941e-02, 7.36481305838055e-02, 7.15762843012300e-02, 6.96563624469064e-02, 6.78881062433732e-02, 6.62696728812620e-02, 6.47978380806080e-02, 6.34681906135341e-02, 6.22753147932968e-02, 6.12129587772330e-02, 6.02741881180435e-02, 5.94515252558523e-02, 5.87370765487663e-02, 5.81226489815094e-02, 5.75998588923312e-02, 5.71602349489098e-02, 5.67953172229352e-02, 5.64967536229801e-02, 5.62563942013981e-02, 5.60663830323307e-02, 5.59192465440077e-02, 5.58079764484773e-02, 5.57261048271016e-02, 5.56677685554234e-02, 5.56277601228601e-02, 5.56015620490058e-02, 5.55853625055533e-02, 5.55760503904708e-02, 5.55711889213200e-02, 5.55689677430138e-02, 5.55681345132371e-02, 5.55679078555329e-02, 5.55678744013761e-02, 5.55678717687300e-02, 5.55678238147692e-02, 5.55674989204533e-02, 5.55663045837775e-02, 5.55631209904988e-02, 5.55561541291780e-02, 5.55428131157640e-02, 5.55196172350492e-02, 5.54821379451198e-02, 5.54249807507858e-02, 5.53418113589856e-02, 5.52254297669492e-02, 5.50678947946608e-02, 5.48606999551140e-02, 5.45949993999244e-02, 5.42618799871093e-02, 5.38526723777579e-02, 5.33592906662221e-02, 5.27745866608882e-02, 5.20927019279043e-02, 5.13093984876747e-02, 5.04223480231332e-02, 4.94313599474765e-02, 4.83385308890263e-02, 4.71483020936319e-02, 4.58674167001476e-02, 4.45047753780443e-02, 4.30711958160401e-02, 4.15790883263583e-02, 4.00420657033322e-02, 3.84745098869388e-02, 3.68911205084288e-02, 3.53064707820039e-02, 3.37345944014089e-02, 3.21886232955677e-02, 3.06804908477858e-02, 2.92207092576007e-02, 2.78182239973006e-02, 2.64803432960150e-02, 2.52127364762060e-02, 2.40194916667775e-02, 2.29032209212789e-02, 2.18651993989880e-02, 2.09055253406092e-02, 2.00232890757519e-02, 1.92167416006157e-02, 1.84834557677220e-02, 1.78204750626207e-02, 1.72244464249900e-02, 1.66917349079370e-02, 1.62185195583788e-02, 1.58008712407707e-02, 1.54348141158067e-02, 1.51163732790423e-02, 1.48416114612928e-02, 1.46066575287733e-02, 1.44077289538558e-02, 1.42411503404105e-02, 1.41033698363060e-02, 1.39909750940035e-02, 1.39007096806125e-02, 1.38294903618430e-02, 1.37744251375835e-02, 1.37328314938543e-02, 1.37022540644978e-02, 1.36804805136563e-02, 1.36655543673925e-02, 1.36557834706821e-02, 1.36497428189131e-02, 1.36462708266265e-02, 1.36444583587263e-02, 1.36436302806828e-02, 1.36433196477554e-02, 1.36432351487628e-02, 1.36432226788514e-02, 1.36432216879427e-02, 1.36432035402692e-02, 1.36430806021934e-02, 1.36426286661901e-02, 1.36414240273410e-02, 1.36387879655047e-02, 1.36337405930931e-02, 1.36249663002843e-02, 1.36107931194687e-02, 1.35891882722490e-02, 1.35577720848516e-02, 1.35138522720837e-02, 1.34544800105800e-02, 1.33765285448959e-02, 1.32767939477816e-02, 1.31521160684116e-02, 1.29995160565670e-02, 1.28163446622978e-02, 1.26004338396370e-02, 1.23502425798421e-02, 1.20649871553695e-02, 1.17447452786053e-02, 1.13905239084744e-02, 1.10042825035066e-02, 1.05889079752757e-02, 1.01447816979316e-02, 9.68366675940047e-03, 9.20667267671149e-03, 8.71920914779878e-03, 8.22685379057475e-03, 7.73514174393597e-03, 7.24936740050452e-03, 6.77441066302720e-03, 6.31459723294370e-03, 5.87359883952843e-03, 5.45437549681083e-03, 5.05915829744197e-03, 4.69139623990600e-03, 4.34817892025308e-03, 4.03161432262777e-03, 3.74145594673791e-03, 3.47703845598346e-03, 3.23737844686889e-03, 3.02125126417252e-03, 2.82727574591490e-03, 2.65398312370735e-03, 2.49984851542706e-03, 2.36333712871328e-03, 2.24295980247291e-03, 2.13728266388102e-03, 2.04492774757625e-03, 1.96460836068013e-03, 1.89515078126590e-03, 1.83546731197737e-03, 1.78452525904049e-03, 1.74135662389574e-03, 1.70508045628992e-03, 1.67490103420675e-03, 1.65008717465298e-03, 1.62995394606469e-03, 1.61385828231175e-03, 1.60120371796779e-03, 1.59144446367065e-03, 1.58408572264079e-03, 1.57868231976011e-03, 1.57483765450195e-03, 1.57220340862807e-03, 1.57047955154610e-03, 1.56941402957500e-03, 1.56880167187978e-03, 1.56848202719573e-03, 1.56833599282891e-03, 1.56828121257035e-03, 1.56826631122442e-03, 1.56826411179622e-03, 1.56826393590848e-03, 1.56826069549254e-03, 1.56823874127099e-03, 1.56815803734389e-03, 1.56794292591655e-03, 1.56747223924469e-03, 1.56657112077596e-03, 1.56500500742141e-03, 1.56247624042294e-03, 1.55862376894333e-03, 1.55302639676336e-03, 1.54521005219976e-03, 1.53465974480582e-03, 1.52083721130225e-03, 1.50320533069248e-03, 1.48125931032439e-03, 1.45456130016441e-03, 1.42276982681722e-03, 1.38565565282533e-03, 1.34311315510161e-03, 1.29518967855802e-03, 1.24212620546871e-03, 1.18437109667013e-03, 1.12254735233819e-03, 1.05740887869388e-03, 9.89830196175490e-04, 9.20795633646776e-04, 8.51333918689079e-04, 7.82436782728269e-04, 7.15027444854854e-04, 6.49937974287695e-04, 5.87866336573484e-04, 5.29368242737860e-04, 4.74870409178152e-04, 4.24654657265933e-04, 3.78848528541198e-04, 3.37445052664975e-04, 3.00316374463705e-04, 2.67244285822699e-04, 2.37970179341573e-04, 2.12214596473359e-04, 1.89683141812225e-04, 1.70079601033990e-04, 1.53098529807494e-04, 1.38425952030730e-04, 1.25770763091282e-04, 1.14873736038135e-04, 1.05499282682331e-04, 9.74502549269118e-05, 9.05733891727323e-05, 8.47274766186956e-05, 7.97651995867454e-05, 7.55575086657245e-05, 7.20102975435566e-05, 6.90466266064504e-05, 6.65863664863764e-05, 6.45489772495011e-05, 6.28661720440959e-05, 6.14857719511974e-05, 6.03668564593818e-05, 5.94739455124227e-05, 5.87740853035257e-05, 5.82364010455194e-05, 5.78325250302214e-05, 5.75369677164872e-05, 5.73272427146776e-05, 5.71838381802081e-05, 5.70901279814962e-05, 5.70322597894801e-05, 5.69990219219108e-05, 5.69816776447851e-05, 5.69737549653837e-05, 5.69707832340782e-05, 5.69699748638186e-05, 5.69698555636933e-05, 5.69698458861012e-05, 5.69696658629937e-05, 5.69684462021769e-05, 5.69639627769902e-05, 5.69520131250556e-05, 5.69258694298469e-05, 5.68758310492967e-05, 5.67889080258510e-05, 5.66486719009163e-05, 5.64353188963196e-05, 5.61260009845910e-05, 5.56954922785735e-05, 5.51172603451278e-05, 5.43649479709529e-05, 5.34140529424576e-05, 5.22431991214489e-05, 5.08342356653202e-05, 4.91714385759678e-05, 4.72424767536950e-05, 4.50447740263371e-05, 4.25965563550790e-05, 3.99432113482590e-05, 3.71464433897017e-05, 3.42616236395088e-05, 3.13329036458754e-05, 2.84092490261401e-05, 2.55413458489784e-05, 2.27582030589701e-05, 2.00742657199271e-05, 1.75246873226342e-05, 1.51734399480574e-05, 1.30914999090237e-05, 1.13498696094894e-05, 1.00096672923655e-05, 9.05287453178971e-06, 8.28841949385151e-06, 7.40647872987665e-06, 6.19200871199155e-06, 4.64026489055414e-06, 2.93649918603478e-06, 1.36656669269541e-06, 7.36039396542765e-07, 1.15161631697447e-06, 1.14767739489707e-06, 6.37790648738701e-07, 1.05957293783376e-06, 1.92713706326231e-06, 2.43755584961036e-06, 2.52261769162487e-06, 2.21224431129485e-06, 1.63357508142781e-06, 1.00155691029613e-06, 4.87802651605307e-07, 6.29043938715768e-07, 6.61484889974417e-07, 5.78283760432896e-07, 4.95777385367546e-07, 4.21612909917712e-07, 3.61957503969279e-07, 3.46963545894141e-07, 3.64059729918154e-07, 3.66566268894530e-07, 2.84217094304040e-14, 2.84217094304040e-14, 1.42108547152020e-14, 2.84217094304040e-14, 2.84217094304040e-14, 2.84217094304040e-14, 2.84217094304040e-14, 1.42108547152020e-14, 2.84217094304040e-14, 2.84217094304040e-14, 2.84217094304040e-14, 2.84217094304040e-14, 2.84217094304040e-14],
[ 1.06000000000000e+02, 1.05999999864329e+02, 1.05999995669117e+02, 1.05999967245888e+02, 1.05999862758336e+02, 1.05999584227087e+02, 1.05998974638838e+02, 1.05997807091681e+02, 1.05995776364684e+02, 1.05992493223410e+02, 1.05987481687143e+02, 1.05980179392140e+02, 1.05969941093231e+02, 1.05956045258733e+02, 1.05937703636227e+02, 1.05914073603472e+02, 1.05884273071765e+02, 1.05847397677615e+02, 1.05802539977954e+02, 1.05748810345809e+02, 1.05685359236165e+02, 1.05611400443740e+02, 1.05526234896669e+02, 1.05429274420014e+02, 1.05320064767945e+02, 1.05198307081506e+02, 1.05063876808239e+02, 1.04916839053777e+02, 1.04757459355073e+02, 1.04586208991637e+02, 1.04403764189356e+02, 1.04210998905197e+02, 1.04008971275020e+02, 1.03798904212423e+02, 1.03582161011710e+02, 1.03360217087319e+02, 1.03134629145275e+02, 1.02907003119812e+02, 1.02678962130870e+02, 1.02452115552247e+02, 1.02228030061248e+02, 1.02008203305010e+02, 1.01794040597089e+02, 1.01586834871347e+02, 1.01387749978723e+02, 1.01197807316311e+02, 1.01017875720568e+02, 1.00848664526898e+02, 1.00690719685041e+02, 1.00544422813857e+02, 1.00409993072943e+02, 1.00287491717581e+02, 1.00176829186004e+02, 1.00077774543670e+02, 9.99899670793764e+01, 9.99129298143058e+01, 9.98460846494295e+01, 9.97887688410615e+01, 9.97402524606931e+01, 9.96997564654329e+01, 9.96664709812652e+01, 9.96395733847780e+01, 9.96182457617184e+01, 9.96016913242864e+01, 9.95891493847453e+01, 9.95799085115044e+01, 9.95733175355758e+01, 9.95687941297420e+01, 9.95658307487275e+01, 9.95639977942117e+01, 9.95629439510843e+01, 9.95623937278641e+01, 9.95621423212760e+01, 9.95620480091675e+01, 9.95620223539101e+01, 9.95620185672450e+01, 9.95620180705145e+01, 9.95620064972881e+01, 9.95619280869658e+01, 9.95616398421617e+01, 9.95608714860750e+01, 9.95591899511371e+01, 9.95559695858928e+01, 9.95503692664276e+01, 9.95413174534597e+01, 9.95275060676517e+01, 9.95073938599007e+01, 9.94792197247674e+01, 9.94410261377909e+01, 9.93906925865409e+01, 9.93259785096120e+01, 9.92445748617562e+01, 9.91441629991478e+01, 9.90224791482139e+01, 9.88773823169017e+01, 9.87069231708113e+01, 9.85094111763627e+01, 9.82834772549986e+01, 9.80281293273043e+01, 9.77427984543588e+01, 9.74273737629639e+01, 9.70822248851921e+01, 9.67082111414512e+01, 9.63066770630522e+01, 9.58794340802338e+01, 9.54287284145869e+01, 9.49571956354749e+01, 9.44678031811292e+01, 9.39637834167480e+01, 9.34485611638045e+01, 9.29256804605360e+01, 9.23987350353638e+01, 9.18713055175720e+01, 9.13469043452367e+01, 9.08289275579862e+01, 9.03206118435573e+01, 8.98249954183390e+01, 8.93448821448021e+01, 8.88828091693893e+01, 8.84410189247940e+01, 8.80214364751293e+01, 8.76256529849894e+01, 8.72549157313096e+01, 8.69101246984201e+01, 8.65918354916219e+01, 8.63002681052685e+01, 8.60353209805516e+01, 8.57965897595066e+01, 8.55833901544763e+01, 8.53947843799579e+01, 8.52296106176747e+01, 8.50865149950127e+01, 8.49639855474555e+01, 8.48603876080553e+01, 8.47740000254128e+01, 8.47030515623254e+01, 8.46457567776557e+01, 8.46003506519285e+01, 8.45651211904099e+01, 8.45384392328698e+01, 8.45187847224665e+01, 8.45047687409507e+01, 8.44951507051436e+01, 8.44888502393604e+01, 8.44849533865928e+01, 8.44827129919717e+01, 8.44815432776035e+01, 8.44810088194035e+01, 8.44808083247032e+01, 8.44807537851912e+01, 8.44807457352701e+01, 8.44807450246153e+01, 8.44807310996844e+01, 8.44806367562191e+01, 8.44802899395132e+01, 8.44793654523749e+01, 8.44773422291304e+01, 8.44734674925224e+01, 8.44667292240262e+01, 8.44558382053599e+01, 8.44392206885455e+01, 8.44150225191805e+01, 8.43811252660465e+01, 8.43351745929696e+01, 8.42746207408513e+01, 8.41967705679848e+01, 8.40988501304201e+01, 8.39780762845649e+01, 8.38317352837889e+01, 8.36572658507684e+01, 8.34523437762266e+01, 8.32149647656185e+01, 8.29435220717095e+01, 8.26368754521753e+01, 8.22944082071593e+01, 8.19160694968237e+01, 8.15023998067584e+01, 8.10545382858950e+01, 8.05742116623968e+01, 8.00637054532726e+01, 7.95258191095921e+01, 7.89638074700474e+01, 7.83813113517212e+01, 7.77822802673159e+01, 7.71708901718888e+01, 7.65514589139559e+01, 7.59283618150194e+01, 7.53059496098898e+01, 7.46884708535291e+01, 7.40800007685287e+01, 7.34843782682285e+01, 7.29051524747837e+01, 7.23455394727124e+01, 7.18083893943708e+01, 7.12961633595848e+01, 7.08109193959449e+01, 7.03543062886543e+01, 6.99275643181715e+01, 6.95315319683962e+01, 6.91666578527629e+01, 6.88330172570940e+01, 6.85303328114530e+01, 6.82579988751220e+01, 6.80151092564418e+01, 6.78004879016248e+01, 6.76127221801329e+01, 6.74501983718358e+01, 6.73111389239088e+01, 6.71936409941244e+01, 6.70957157339772e+01, 6.70153276941297e+01, 6.69504336623191e+01, 6.68990201781016e+01, 6.68591389185225e+01, 6.68289391227637e+01, 6.68066962297372e+01, 6.67908359461473e+01, 6.67799530466626e+01, 6.67728243322374e+01, 6.67684153336611e+01, 6.67658805383207e+01, 6.67645571296051e+01, 6.67639524493256e+01, 6.67637256122447e+01, 6.67636639069991e+01, 6.67636547994319e+01, 6.67636540688924e+01, 6.67636405867325e+01, 6.67635492431137e+01, 6.67632134542813e+01, 6.67623183643308e+01, 6.67603594808099e+01, 6.67566079804698e+01, 6.67500840772056e+01, 6.67395396809788e+01, 6.67234513867276e+01, 6.67000246106707e+01, 6.66672094306601e+01, 6.66227283796645e+01, 6.65641160812742e+01, 6.64887702018532e+01, 6.63940127309712e+01, 6.62771601040817e+01, 6.61356001729669e+01, 6.59668735435429e+01, 6.57687563778863e+01, 6.55393414417892e+01, 6.52771140124851e+01, 6.49810192764001e+01, 6.46505180621487e+01, 6.42856281691575e+01, 6.38869491477131e+01, 6.34556691264864e+01, 6.29935531233888e+01, 6.25029131661102e+01, 6.19865614408999e+01, 6.14477485322143e+01, 6.08900895572969e+01, 6.03174815773721e+01, 5.97340160174303e+01, 5.91438898969852e+01, 5.85513194416321e+01, 5.79604591321400e+01, 5.73753285264044e+01, 5.67997483700706e+01, 5.62372867165352e+01, 5.56912151108731e+01, 5.51644744175355e+01, 5.46596495996321e+01, 5.41789526553956e+01, 5.37242129271376e+01, 5.32968740592162e+01, 5.28979969486519e+01, 5.25282680820390e+01, 5.21880126826920e+01, 5.18772121120565e+01, 5.15955249912792e+01, 5.13423115400105e+01, 5.11166606702681e+01, 5.09174194175138e+01, 5.07432243297063e+01, 5.05925344587482e+01, 5.04636656008152e+01, 5.03548254098024e+01, 5.02641489630179e+01, 5.01897342954516e+01, 5.01296773464799e+01, 5.00821056905447e+01, 5.00452103617391e+01, 5.00172750415510e+01, 4.99967018681579e+01, 4.99820331512617e+01, 4.99719683422223e+01, 4.99653757154961e+01, 4.99612983609304e+01, 4.99589542608270e+01, 4.99577304217458e+01, 4.99571712378319e+01, 4.99569614683608e+01, 4.99569044059521e+01, 4.99568959836592e+01, 4.99568953340509e+01, 4.99568836691716e+01, 4.99568046379032e+01, 4.99565141106788e+01, 4.99557396723677e+01, 4.99540448361207e+01, 4.99507990336591e+01, 4.99451545925909e+01, 4.99360317736637e+01, 4.99221127805441e+01, 4.99018454665664e+01, 4.98734572398243e+01, 4.98349794025869e+01, 4.97842818475964e+01, 4.97191176705296e+01, 4.96371768490052e+01, 4.95361476959530e+01, 4.94137843395166e+01, 4.92679780418754e+01, 4.90968297810514e+01, 4.88987212220284e+01, 4.86723810347460e+01, 4.84169435095505e+01, 4.81319965981993e+01, 4.78176168796335e+01, 4.74743895078061e+01, 4.71034119204833e+01, 4.67062809322123e+01, 4.62850637423545e+01, 4.58422542828953e+01, 4.53807171195809e+01, 4.49036217114442e+01, 4.44143701559003e+01, 4.39165215724375e+01, 4.34137160414835e+01, 4.29096006051573e+01, 4.24077593666904e+01, 4.19116492910932e+01, 4.14245429569548e+01, 4.09494792277543e+01, 4.04892225549957e+01, 4.00462313470756e+01, 3.96226355145191e+01, 3.92202229465156e+01, 3.88404343248571e+01, 3.84843653876957e+01, 3.81527755563588e+01, 3.78461017530969e+01, 3.75644762628469e+01, 3.73077476065140e+01, 3.70755035652857e+01, 3.68670956913229e+01, 3.66816648300863e+01, 3.65181673413513e+01, 3.63754018260399e+01, 3.62520362386326e+01, 3.61466352907338e+01, 3.60576880355012e+01, 3.59836354732427e+01, 3.59228979454084e+01, 3.58739019982431e+01, 3.58351063095245e+01, 3.58050261928175e+01, 3.57822561333838e+01, 3.57654897766967e+01, 3.57535367907447e+01, 3.57453360606489e+01, 3.57399647491782e+01, 3.57366428669569e+01, 3.57347331358803e+01, 3.57337360903416e+01, 3.57332805334136e+01, 3.57331096385109e+01, 3.57330631509754e+01, 3.57330562895133e+01, 3.57330557699841e+01, 3.57330465665768e+01, 3.57329842121584e+01, 3.57327549910257e+01, 3.57321439739900e+01, 3.57308067900545e+01, 3.57282459684588e+01, 3.57237928209573e+01, 3.57165957339956e+01, 3.57056156229399e+01, 3.56896291598755e+01, 3.56672402150152e+01, 3.56368997413136e+01, 3.55969340758159e+01, 3.55455813265078e+01, 3.54810351624929e+01, 3.54014949380092e+01, 3.53052206751735e+01, 3.51905910327653e+01, 3.50561620322779e+01, 3.49007240359779e+01, 3.47233543139279e+01, 3.45234625333747e+01, 3.43008266814340e+01, 3.40556173037097e+01, 3.37884085021228e+01, 3.35001748580348e+01, 3.31922742821755e+01, 3.28664176704445e+01, 3.25246270784469e+01, 3.21691848262903e+01, 3.18025764254687e+01, 3.14274304219392e+01, 3.10464581515534e+01, 3.06623960291907e+01, 3.02779524089839e+01, 2.98957603602925e+01, 2.95183370150818e+01, 2.91480495577112e+01, 2.87870875154569e+01, 2.84374407900796e+01, 2.81008828255867e+01, 2.77789583807549e+01, 2.74729754996991e+01, 2.71840013896500e+01, 2.69128619839408e+01, 2.66601449791843e+01, 2.64262061012895e+01, 2.62111783007698e+01, 2.60149835301128e+01, 2.58373467329610e+01, 2.56778116824975e+01, 2.55357583396766e+01, 2.54104214488799e+01, 2.53009101354580e+01, 2.52062283047561e+01, 2.51252956581741e+01, 2.50569691359177e+01, 2.50000645698179e+01, 2.49533782875227e+01, 2.49157083580230e+01, 2.48858751154009e+01, 2.48627405505675e+01, 2.48452261268251e+01, 2.48323285604972e+01, 2.48231331172197e+01, 2.48168240104034e+01, 2.48126915512387e+01, 2.48101357874487e+01, 2.48086664766742e+01, 2.48078993639228e+01, 2.48075488637453e+01, 2.48074173790240e+01, 2.48073816119914e+01, 2.48073763328523e+01, 2.48073759365456e+01, 2.48073689610997e+01, 2.48073217014336e+01, 2.48071479700069e+01, 2.48066848666128e+01, 2.48056713802425e+01, 2.48037304482355e+01, 2.48003552001252e+01, 2.47949000332487e+01, 2.47865770734203e+01, 2.47744584632344e+01, 2.47574847880892e+01, 2.47344797907811e+01, 2.47041713355440e+01, 2.46652183601710e+01, 2.46162433026661e+01, 2.45558692139973e+01, 2.44827604837145e+01, 2.43956658286724e+01, 2.42934619494164e+01, 2.41751960690782e+01, 2.40401254610260e+01, 2.38877520663685e+01, 2.37178504176361e+01, 2.35304873297607e+01, 2.33260321933877e+01, 2.31051571963496e+01, 2.28688273800618e+01, 2.26182810643423e+01, 2.23550017836454e+01, 2.20806833922834e+01, 2.17971903369861e+01, 2.15065152038991e+01, 2.12107355102738e+01, 2.09119713732710e+01, 2.06123452444064e+01, 2.03139444632018e+01, 2.00187870521921e+01, 1.97287909907848e+01, 1.94457471521389e+01, 1.91712961084974e+01, 1.89069090384995e+01, 1.86538729534917e+01, 1.84132803778742e+01, 1.81860234800136e+01, 1.79727924833462e+01, 1.77740780257356e+01, 1.75901770072618e+01, 1.74212013892517e+01, 1.72670893847257e+01, 1.71276185062979e+01, 1.70024199990460e+01, 1.68909942673017e+01, 1.67927269904913e+01, 1.67069057012949e+01, 1.66327366603712e+01, 1.65693619003730e+01, 1.65158763263425e+01, 1.64713447512886e+01, 1.64348187187909e+01, 1.64053529246377e+01, 1.63820210036655e+01, 1.63639304034466e+01, 1.63502360303300e+01, 1.63401523317650e+01, 1.63329634765851e+01, 1.63280313148942e+01, 1.63248008421687e+01, 1.63228029567780e+01, 1.63216543829608e+01, 1.63210547273698e+01, 1.63207807405660e+01, 1.63206779587844e+01, 1.63206499996598e+01, 1.63206458729520e+01, 1.63206455641008e+01, 1.63206401405188e+01, 1.63206033949635e+01, 1.63204683145707e+01, 1.63201082410322e+01, 1.63193202349599e+01, 1.63178111322870e+01, 1.63151868627947e+01, 1.63109455410260e+01, 1.63044747699021e+01, 1.62950535141743e+01, 1.62818588003281e+01, 1.62639773774581e+01, 1.62404223265761e+01, 1.62101544324457e+01, 1.61721079340021e+01, 1.61252200521730e+01, 1.60684634671333e+01, 1.60008806944142e+01, 1.59216191078935e+01, 1.58299651965432e+01, 1.57253765401041e+01, 1.56075099638786e+01, 1.54762443978143e+01, 1.53316971273211e+01, 1.51742323827830e+01, 1.50044615631463e+01, 1.48232348092625e+01, 1.46316241093294e+01, 1.44308985987635e+01, 1.42224931716260e+01, 1.40079719095738e+01, 1.37889881190420e+01, 1.35672429179203e+01, 1.33444443128718e+01, 1.31222685586119e+01, 1.29023253102992e+01, 1.26861277049451e+01, 1.24750680828308e+01, 1.22703996332630e+01, 1.20732238633754e+01, 1.18844834758227e+01, 1.17049600186499e+01, 1.15352755417923e+01, 1.13758974513311e+01, 1.12271457789099e+01, 1.10892021601702e+01, 1.09621199232752e+01, 1.08458348095828e+01, 1.07401759699039e+01, 1.06448769919930e+01, 1.05595868118161e+01, 1.04838804392779e+01, 1.04172694869953e+01, 1.03592125282768e+01, 1.03091253285614e+01, 1.02663909947676e+01, 1.02303700714915e+01, 1.02004105845512e+01, 1.01758579943398e+01, 1.01560649776935e+01, 1.01404009117472e+01, 1.01282608910336e+01, 1.01190740742531e+01, 1.01123111336874e+01, 1.01074905713306e+01, 1.01041836735209e+01, 1.01020179008400e+01, 1.01006785514542e+01, 1.00999085915728e+01, 1.00995066127031e+01, 1.00993229474244e+01, 1.00992540485914e+01, 1.00992353064716e+01, 1.00992325401752e+01, 1.00992323330625e+01, 1.00992286950240e+01, 1.00992040467953e+01, 1.00991134375652e+01, 1.00988719086656e+01, 1.00983433389314e+01, 1.00973311043355e+01, 1.00955709440441e+01, 1.00927263904195e+01, 1.00883870793290e+01, 1.00820702060417e+01, 1.00732253266375e+01, 1.00612426206014e+01, 1.00454646232434e+01, 1.00252013047101e+01, 9.99974821654708e+00, 9.96840725180112e+00, 9.93050937975933e+00, 9.88543853527439e+00, 9.83265568256149e+00, 9.77172195382249e+00, 9.70231970324585e+00, 9.62427033293598e+00, 9.53754784911652e+00, 9.44228729584106e+00, 9.33878748050836e+00, 9.22750773150034e+00, 9.10905878542879e+00, 8.98418825673591e+00, 8.85376146194886e+00, 8.71873862503216e+00, 8.58014965753905e+00, 8.43906777703496e+00, 8.29658320045183e+00, 8.15377803672912e+00, 8.01170332386563e+00, 7.87135893204614e+00, 7.73367681045463e+00, 7.59950781279530e+00, 7.46961211410181e+00, 7.34465304337951e+00, 7.22519401224031e+00, 7.11169812301857e+00, 7.00452999007354e+00, 6.90395930010592e+00, 6.81016566339194e+00, 6.72324435835041e+00, 6.64321263714336e+00, 6.57001633136376e+00, 6.50353656724801e+00, 6.44359646413436e+00, 6.38996774474357e+00, 6.34237722944006e+00, 6.30051321820419e+00, 6.26403178364779e+00, 6.23256300659708e+00, 6.20571718351405e+00, 6.18309102362768e+00, 6.16427383472723e+00, 6.14885367211035e+00, 6.13642339746426e+00, 6.12658656605338e+00, 6.11896303419364e+00, 6.11319415729812e+00, 6.10894743422496e+00, 6.10592044822396e+00, 6.10384395974819e+00, 6.10248402219774e+00, 6.10164301777567e+00, 6.10115954561343e+00, 6.10090713591720e+00, 6.10079180929066e+00, 6.10074854652582e+00, 6.10073677802694e+00, 6.10073504102195e+00, 6.10073491062735e+00, 6.10073261557264e+00, 6.10071706625119e+00, 6.10065990547788e+00, 6.10050753720400e+00, 6.10017408987012e+00, 6.09953552480016e+00, 6.09842513759747e+00, 6.09663067898897e+00, 6.09389329591723e+00, 6.08990846169812e+00, 6.08432902287359e+00, 6.07677043695695e+00, 6.06681820702120e+00, 6.05403743465750e+00, 6.03798431281247e+00, 6.01821926756220e+00, 5.99432133911853e+00, 5.96590327635326e+00, 5.93262671735299e+00, 5.89421675376903e+00, 5.85047514150504e+00, 5.80129143480606e+00, 5.74665139104059e+00, 5.68664211942533e+00, 5.62145362182918e+00, 5.55137658412417e+00, 5.47679650351992e+00, 5.39818445921033e+00, 5.31608502895682e+00, 5.23110200476954e+00, 5.14388265443978e+00, 5.05510130755982e+00, 4.96544301760446e+00, 4.87558797466238e+00, 4.78619722965672e+00, 4.69790015518782e+00, 4.61128392474117e+00, 4.52688515306615e+00, 4.44518371528483e+00, 4.36659865690244e+00, 4.29148602475406e+00, 4.22013839112694e+00, 4.15278580910588e+00, 4.08959792449429e+00, 4.03068697535473e+00, 3.97611143054830e+00, 3.92588004957538e+00, 3.87995618348457e+00, 3.83826217684917e+00, 3.80068377051322e+00, 3.76707444129637e+00, 3.73725964608511e+00, 3.71104096234976e+00, 3.68820013433530e+00, 3.66850304373969e+00, 3.65170362585232e+00, 3.63754774752640e+00, 3.62577705301665e+00, 3.61613276896828e+00, 3.60835944229677e+00, 3.60220856614752e+00, 3.59744203147146e+00, 3.59383532685739e+00, 3.59118039882452e+00, 3.58928808015763e+00, 3.58798999596192e+00, 3.58713986623884e+00, 3.58661413962979e+00, 3.58631191461426e+00, 3.58615413043891e+00, 3.58608203857800e+00, 3.58605499459901e+00, 3.58604763799677e+00, 3.58604655217826e+00, 3.58604647032017e+00, 3.58604502492501e+00, 3.58603523216850e+00, 3.58599923308620e+00, 3.58590327372150e+00, 3.58569327392988e+00, 3.58529111810081e+00, 3.58459182486553e+00, 3.58346173982829e+00, 3.58173787977597e+00, 3.57922853719856e+00, 3.57571522837810e+00, 3.57095603474552e+00, 3.56469034395138e+00, 3.55664494320651e+00, 3.54654135325742e+00, 3.53410421885569e+00, 3.51907049465638e+00, 3.50119908992761e+00, 3.48028056861409e+00, 3.45614645139143e+00, 3.42867764146128e+00, 3.39781150270224e+00, 3.36354716155995e+00, 3.32594868324661e+00, 3.28514588471492e+00, 3.24133268353381e+00, 3.19476303175686e+00, 3.14574463345285e+00, 3.09463077960905e+00, 3.04181074188239e+00, 2.98769923753116e+00, 2.93272550651888e+00, 2.87732252782335e+00, 2.82191684968149e+00, 2.76691942594827e+00, 2.71271774851093e+00, 2.65966945524133e+00, 2.60809748512591e+00, 2.55828675595231e+00, 2.51048226158160e+00, 2.46488842878274e+00, 2.42166953848769e+00, 2.38095100152776e+00, 2.34282128127154e+00, 2.30733427110746e+00, 2.27451195921553e+00, 2.24434724264921e+00, 2.21680678408123e+00, 2.19183383504500e+00, 2.16935097724220e+00, 2.14926275722844e+00, 2.13145820882005e+00, 2.11581327155686e+00, 2.10219312249831e+00, 2.09045444272304e+00, 2.08044763952878e+00, 2.07201904101865e+00, 2.06501307218691e+00, 2.05927441159133e+00, 2.05465011616255e+00, 2.05099168969724e+00, 2.04815705921632e+00, 2.04601241373986e+00, 2.04443385314400e+00, 2.04330879144621e+00, 2.04253705966087e+00, 2.04203165848686e+00, 2.04171912035382e+00, 2.04153945321364e+00, 2.04144565403841e+00, 2.04140279716573e+00, 2.04138672019376e+00, 2.04138234687896e+00, 2.04138170138727e+00, 2.04138165245989e+00, 2.04138078502035e+00, 2.04137490799583e+00, 2.04135330352325e+00, 2.04129571463130e+00, 2.04116968630173e+00, 2.04092834058326e+00, 2.04050868093982e+00, 2.03983051371232e+00, 2.03879606576668e+00, 2.03729036607081e+00, 2.03518244379704e+00, 2.03232737519988e+00, 2.02856918479128e+00, 2.02374457256583e+00, 2.01768739839521e+00, 2.01023380852051e+00, 2.00122783991922e+00, 1.99052729009692e+00, 1.97800959749527e+00, 1.96357744674303e+00, 1.94716379879327e+00, 1.92873605298388e+00, 1.90829907876572e+00, 1.88589690920399e+00, 1.86161296336371e+00, 1.83556875442999e+00, 1.80792113669855e+00, 1.77885823796782e+00, 1.74859430502076e+00, 1.71736375090423e+00, 1.68541472826689e+00, 1.65300256102449e+00, 1.62038334835543e+00, 1.58780801466699e+00, 1.55551702292078e+00, 1.52373590372397e+00, 1.49267168591881e+00, 1.46251025208579e+00, 1.43341458892007e+00, 1.40552386058802e+00, 1.37895320397963e+00, 1.35379412789922e+00, 1.33011539232173e+00, 1.30796424692284e+00, 1.28736791794283e+00, 1.26833524687270e+00, 1.25085840146149e+00, 1.23491459745674e+00, 1.22046778696247e+00, 1.20747028533783e+00, 1.19586432245771e+00, 1.18558351547398e+00, 1.17655426870975e+00, 1.16869711191390e+00, 1.16192799085810e+00, 1.15615952434472e+00, 1.15130223938711e+00, 1.14726579200199e+00, 1.14396017518118e+00, 1.14129690875156e+00, 1.13919019860845e+00, 1.13755804588948e+00, 1.13632328071215e+00, 1.13541449073811e+00, 1.13476681256178e+00, 1.13432255408610e+00, 1.13403161878367e+00, 1.13385170794820e+00, 1.13374828440105e+00, 1.13369429011161e+00, 1.13366962015954e+00, 1.13366036568723e+00, 1.13365784825378e+00, 1.13365747668614e+00, 1.13365744834314e+00, 1.13365694348842e+00, 1.13365352302733e+00, 1.13364094911086e+00, 1.13360743212809e+00, 1.13353408337609e+00, 1.13339362072652e+00, 1.13314938311649e+00, 1.13275470568711e+00, 1.13215270252900e+00, 1.13127649762596e+00, 1.13004993590081e+00, 1.12838879438726e+00, 1.12620249773327e+00, 1.12339632205681e+00, 1.11987404671964e+00, 1.11554098565452e+00, 1.11030730006885e+00, 1.10409146504963e+00, 1.09682373686202e+00, 1.08844944896313e+00, 1.07893195625951e+00, 1.06825505159838e+00, 1.05642469742493e+00, 1.04346994887273e+00, 1.02944299039669e+00, 1.01441826274757e+00, 9.98490715619894e-01, 9.81773277940263e-01, 9.64393686855345e-01, 9.46490853270788e-01, 9.28210963081909e-01, 9.09703517753655e-01, 8.91117506374016e-01, 8.72597876176741e-01, 8.54282433493623e-01, 8.36299266467336e-01, 8.18764738939336e-01, 8.01782065524137e-01, 7.85440443915206e-01, 7.69814693821615e-01, 7.54965333450528e-01, 7.40939014044471e-01, 7.27769229868969e-01, 7.15477223965792e-01, 7.04073017448380e-01, 6.93556500597154e-01, 6.83918536097920e-01, 6.75142037263839e-01, 6.67202996040885e-01, 6.60071446343665e-01, 6.53712357356298e-01, 6.48086458628185e-01, 6.43151004010477e-01, 6.38860484752811e-01, 6.35167303528448e-01, 6.32022420950292e-01, 6.29375984504847e-01, 6.27177947018765e-01, 6.25378678080722e-01, 6.23929567597571e-01, 6.22783616227977e-01, 6.21896003185896e-01, 6.21224618215521e-01, 6.20730541759201e-01, 6.20378455755002e-01, 6.20136967313840e-01, 6.19978828818589e-01, 6.19881040710652e-01, 6.19824827207012e-01, 6.19795480127408e-01, 6.19782071532072e-01, 6.19777041558052e-01, 6.19775673288174e-01, 6.19775471334592e-01, 6.19775455820928e-01, 6.19775178062014e-01, 6.19773296207072e-01, 6.19766378350050e-01, 6.19747938196468e-01, 6.19707583961429e-01, 6.19630306866355e-01, 6.19495940413742e-01, 6.19278819681085e-01, 6.18947667476732e-01, 6.18465730930097e-01, 6.17791187424601e-01, 6.16877832165841e-01, 6.15676050622426e-01, 6.14134067302288e-01, 6.12199447867042e-01, 6.09820814926238e-01, 6.06949719972334e-01, 6.03542596315052e-01, 5.99562702457348e-01, 5.94981954264199e-01, 5.89782539586565e-01, 5.83958212391764e-01, 5.77515175845576e-01, 5.70472485082362e-01, 5.62861929312078e-01, 5.54727387037801e-01, 5.46123684214024e-01, 5.37115019439398e-01, 5.27773049114060e-01, 5.18174745904417e-01, 5.08400153923546e-01, 4.98530163167416e-01, 4.88644414712240e-01, 4.78819428916907e-01, 4.69127024110193e-01, 4.59633066052078e-01, 4.50396561795642e-01, 4.41469087915763e-01, 4.32894524150399e-01, 4.24709050240267e-01, 4.16941356283274e-01, 4.09613014733694e-01, 4.02738964330283e-01, 3.96328061561576e-01, 3.90383662574578e-01, 3.84904206616184e-01, 3.79883780280167e-01, 3.75312649368695e-01, 3.71177751658621e-01, 3.67463149078759e-01, 3.64150441713278e-01, 3.61219148702119e-01, 3.58647062640586e-01, 3.56410584632755e-01, 3.54485046887561e-01, 3.52845028809185e-01, 3.51464671068925e-01, 3.50317990284367e-01, 3.49379194813125e-01, 3.48622999923450e-01, 3.48024938388065e-01, 3.47561660508958e-01, 3.47211215879085e-01, 3.46953307969159e-01, 3.46769512017786e-01, 3.46643446787709e-01, 3.46560891563854e-01, 3.46509841283293e-01, 3.46480494815211e-01, 3.46465174002702e-01, 3.46458173954147e-01, 3.46455548019691e-01, 3.46454833704257e-01, 3.46454728272874e-01, 3.46454720112064e-01, 3.46454573194663e-01, 3.46453577809157e-01, 3.46449918687313e-01, 3.46440164970275e-01, 3.46418820002441e-01, 3.46377944913669e-01, 3.46306872421650e-01, 3.46192026226376e-01, 3.46016860055670e-01, 3.45761928642460e-01, 3.45405100393251e-01, 3.44921917945115e-01, 3.44286107948662e-01, 3.43470235104263e-01, 3.42446487748504e-01, 3.41187573407339e-01, 3.39667693263308e-01, 3.37863555264107e-01, 3.35755377708423e-01, 3.33327829765963e-01, 3.30570853658799e-01, 3.27480316014899e-01, 3.24058443608550e-01, 3.20314011096144e-01, 3.16262264532369e-01, 3.11924582866922e-01, 3.07327898300386e-01, 3.02503913196034e-01, 2.97488164283223e-01, 2.92318992748022e-01, 2.87036480830849e-01, 2.81681411919934e-01, 2.76294302767255e-01, 2.70914544845610e-01, 2.65579678734326e-01, 2.60324812463992e-01, 2.55182183343265e-01, 2.50180853859788e-01, 2.45346526203619e-01, 2.40701456746933e-01, 2.36264451033364e-01, 2.32050920887232e-01, 2.28072987497740e-01, 2.24339617171267e-01, 2.20856779405096e-01, 2.17627619691996e-01, 2.14652641835315e-01, 2.11929896471446e-01, 2.09455173964429e-01, 2.07222200909115e-01, 2.05222840212343e-01, 2.03447295170051e-01, 2.01884318161746e-01, 2.00521424566460e-01, 1.99345112283126e-01, 1.98341086822452e-01, 1.97494491349374e-01, 1.96790140314590e-01, 1.96212754468007e-01, 1.95747194148254e-01, 1.95378686859655e-01, 1.95093044364768e-01, 1.94876863912199e-01, 1.94717707867227e-01, 1.94604255977083e-01, 1.94526424825028e-01, 1.94475449716805e-01, 1.94443925285881e-01, 1.94425802442254e-01, 1.94416340851035e-01, 1.94412017809839e-01, 1.94410396094000e-01, 1.94409954948554e-01, 1.94409889836436e-01, 1.94409884755643e-01, 1.94409792759188e-01, 1.94409169470177e-01, 1.94406878200468e-01, 1.94400770567725e-01, 1.94387404421519e-01, 1.94361807643972e-01, 1.94317297734585e-01, 1.94245366211433e-01, 1.94135635810810e-01, 1.93975897161337e-01, 1.93752229934077e-01, 1.93449211296667e-01, 1.93050211734374e-01, 1.92537774882197e-01, 1.91894073977664e-01, 1.91101433024627e-01, 1.90142896035084e-01, 1.89002823167712e-01, 1.87667488711597e-01, 1.86125653195560e-01, 1.84369080938652e-01, 1.82392975459166e-01, 1.80196308476113e-01, 1.77782023643630e-01, 1.75157103236657e-01, 1.72332494091052e-01, 1.69322897387403e-01, 1.66146434519661e-01, 1.62824207602370e-01, 1.59379777658913e-01, 1.55838585995042e-01, 1.52227344782233e-01, 1.48573421736046e-01, 1.44904241406920e-01, 1.41246722411793e-01, 1.37626766321262e-01, 1.34068810136355e-01, 1.30595450534784e-01, 1.27227144457294e-01, 1.23981987236050e-01, 1.20875566408998e-01, 1.17920886704226e-01, 1.15128359495259e-01, 1.12505848402918e-01, 1.10058761700145e-01, 1.07790181820647e-01, 1.05701022489583e-01, 1.03790204775649e-01, 1.02054844504593e-01, 1.00490444901581e-01, 9.90910898348005e-02, 9.78496345287283e-02, 9.67578919661634e-02, 9.58068142680446e-02, 9.49866691681933e-02, 9.42872120884743e-02, 9.36978544693261e-02, 9.32078287348901e-02, 9.28063498381420e-02, 9.24827726428390e-02, 9.22267436351945e-02, 9.20283447061934e-02, 9.18782260663420e-02, 9.17677248301345e-02, 9.16889655986779e-02, 9.16349393342273e-02, 9.15995572207038e-02, 9.15776767681962e-02, 9.15650983682781e-02, 9.15585315099747e-02, 9.15555311035234e-02, 9.15544055556268e-02, 9.15540993792519e-02, 9.15540541888049e-02, 9.15540506323111e-02, 9.15539858542900e-02, 9.15535469770106e-02, 9.15519336253254e-02, 9.15476330734748e-02, 9.15382216994516e-02, 9.15201988504180e-02, 9.14888602783606e-02, 9.14382177102908e-02, 9.13609705218477e-02, 9.12485346120349e-02, 9.10911325222941e-02, 9.08779472932650e-02, 9.05973406740799e-02, 9.02371337784018e-02, 8.97849453461816e-02, 8.92285794176075e-02, 8.85564505762346e-02, 8.77580313679118e-02, 8.68243032609683e-02, 8.57481900136321e-02, 8.45249508953856e-02, 8.31525112338340e-02, 8.16317093904502e-02, 7.99664425535838e-02, 7.81636987365459e-02, 7.62334684797850e-02, 7.41885368373119e-02, 7.20441635727695e-02, 6.98176663939449e-02, 6.75279281555845e-02, 6.51948534062399e-02, 6.28388023871196e-02, 6.04800310799334e-02, 5.81381642242036e-02, 5.58317246745332e-02, 5.35777374088156e-02, 5.13914203075387e-02, 4.92859676653126e-02, 4.72724261129969e-02, 4.53596570925896e-02, 4.35394655044377e-02, 4.18482057222036e-02, 4.02715962106015e-02, 3.88107420467918e-02, 3.74652352875984e-02, 3.62333643461312e-02, 3.51123200509136e-02, 3.40983905599614e-02, 3.31871397778229e-02, 3.23735662049813e-02, 3.16522411316811e-02, 3.10174266998615e-02, 3.04631755725637e-02, 2.99834147654252e-02, 2.95720166353135e-02, 2.92228601259836e-02, 2.89298851933637e-02, 2.86871429292717e-02, 2.84888433277786e-02, 2.83294019500033e-02, 2.82034859969826e-02, 2.81060595564355e-02, 2.80324271073056e-02, 2.79782738004855e-02, 2.79397006314213e-02, 2.79132524171217e-02, 2.78959364980977e-02, 2.78852303074331e-02, 2.78790763561716e-02, 2.78758637422101e-02, 2.78743959416090e-02, 2.78738453301202e-02, 2.78736955514916e-02, 2.78736734445175e-02, 2.78736716894059e-02, 2.78736395245716e-02, 2.78734216038004e-02, 2.78726205131647e-02, 2.78704851704106e-02, 2.78658123842774e-02, 2.78568648162576e-02, 2.78413092274548e-02, 2.78161788232438e-02, 2.77778633956842e-02, 2.77221307276534e-02, 2.76441822625486e-02, 2.75387452524143e-02, 2.74002023807185e-02, 2.72227581525044e-02, 2.70006391485964e-02, 2.67283226276760e-02, 2.64007850887822e-02, 2.60137595340950e-02, 2.55639876268532e-02, 2.50494510981980e-02, 2.44695659929164e-02, 2.38253239638766e-02, 2.31193670080927e-02, 2.23559857831788e-02, 2.15410367374460e-02, 2.06817793074663e-02, 1.97866407858101e-02, 1.88649224345028e-02, 1.79264652984443e-02, 1.69812973439889e-02, 1.60392846046165e-02, 1.51098078540883e-02, 1.42014831602069e-02, 1.33219399851612e-02, 1.24776649280278e-02, 1.16739134313946e-02, 1.09146864180047e-02, 1.02027644004039e-02, 9.53978848598069e-03, 8.92637605247160e-03, 8.36225862889250e-03, 7.84643041362187e-03, 7.37729750996152e-03, 6.95281999936981e-03, 6.57064112839123e-03, 6.22819998206126e-03, 5.92282590697835e-03, 5.65181452043362e-03, 5.41248631201086e-03, 5.20222960682304e-03, 5.01853007354214e-03, 4.85898912387484e-03, 4.72133354627147e-03, 4.60341860205915e-03, 4.50322661717451e-03, 4.41886284718862e-03, 4.34855010104229e-03, 4.29062329762075e-03, 4.24352481924473e-03, 4.20580122191438e-03, 4.17610157604997e-03, 4.15317744278809e-03, 4.13588425357148e-03, 4.12318365921749e-03, 4.11414626420026e-03, 4.10795406997977e-03, 4.10390192381271e-03, 4.10139730959713e-03, 4.09995791636675e-03, 4.09920657165230e-03, 4.09886330990616e-03, 4.09873454606213e-03, 4.09869951968744e-03, 4.09869434987797e-03, 4.09869393567419e-03, 4.09868629760979e-03, 4.09863454896708e-03, 4.09844431989370e-03, 4.09793727281738e-03, 4.09682778428078e-03, 4.09470363842956e-03, 4.09101178749438e-03, 4.08505029339779e-03, 4.07596761434093e-03, 4.06277040615066e-03, 4.04434091838368e-03, 4.01946483276561e-03, 3.98686996649644e-03, 3.94527561672930e-03, 3.89345145602249e-03, 3.83028383804174e-03, 3.75484623050681e-03, 3.66646938874737e-03, 3.56480599350562e-03, 3.44988398667567e-03, 3.32214292552635e-03, 3.18244846427262e-03, 3.03208159633073e-03, 2.87270146041863e-03, 2.70628308867060e-03, 2.53503408661970e-03, 2.36129646404493e-03, 2.18744132433812e-03, 2.01576468903397e-03, 1.84839240168344e-03, 1.68720095398101e-03, 1.53375932957545e-03, 1.38929464772275e-03, 1.25468169453313e-03, 1.13045379715970e-03, 1.01683058663582e-03, 9.13757495936807e-04, 8.20952253661744e-04, 7.37954475553693e-04, 6.64175008594725e-04, 5.98941856438273e-04, 5.41539887151998e-04, 4.91242651136758e-04, 4.47507107551814e-04, 4.09137064764309e-04, 3.76003509913403e-04, 3.47570933755037e-04, 3.22892548837217e-04, 3.01569652890521e-04, 2.83309960693146e-04, 2.67912987098384e-04, 2.54776918637845e-04, 2.43581394329340e-04, 2.34062184560457e-04, 2.25997386237964e-04, 2.19260694322543e-04, 2.13945650003211e-04, 2.09641905371427e-04, 2.06151130910825e-04, 2.03354333478956e-04, 2.01148021702124e-04, 1.99439939166268e-04, 1.98147064907062e-04, 1.97194432317360e-04, 1.96514631952027e-04, 1.96047791888532e-04, 1.95741783940662e-04, 1.95552427683765e-04, 1.95443528454575e-04, 1.95386661673069e-04, 1.95360676073619e-04, 1.95350927540971e-04, 1.95348275639162e-04, 1.95347884229591e-04, 1.95347853377825e-04, 1.95347290755876e-04, 1.95343478921473e-04, 1.95329465668692e-04, 1.95292107378009e-04, 1.95210329238193e-04, 1.95053638094578e-04, 1.94780924132942e-04, 1.94339588091452e-04, 1.93665064614379e-04, 1.92680896304864e-04, 1.91299633272024e-04, 1.89424932173665e-04, 1.86955210836004e-04, 1.83788959475351e-04, 1.79830991427821e-04, 1.74996938866911e-04, 1.69461871692533e-04, 1.63412467415469e-04, 1.56425218889922e-04, 1.48535489387314e-04, 1.39813988113247e-04, 1.30367611291149e-04, 1.20667132989638e-04, 1.10950829650847e-04, 1.00943738701176e-04, 9.08492532687433e-05, 8.10722604143166e-05, 7.19980315437851e-05, 6.32694023892100e-05, 5.49937385301291e-05, 4.77889956158606e-05, 4.11670764890459e-05, 3.51001444016674e-05, 3.00442755758468e-05, 2.56618246652351e-05, 2.17645937539146e-05, 1.86795811032425e-05, 1.60889630507199e-05, 1.37239877489037e-05, 1.18769318788736e-05, 1.03095687080668e-05, 8.67232235179927e-06, 7.32114457946409e-06, 6.00398367112120e-06, 4.37686263410342e-06, 3.45298948900563e-06, 2.84617084389538e-06, 2.48794408719277e-06, 2.82321647659955e-06, 3.00530571450963e-06, 2.90156664561891e-06, 2.68186002472248e-06, 2.39163102833118e-06, 2.04723976082732e-06, 1.70154795853250e-06, 1.40578299578920e-06, 1.18058875386851e-06, 1.01754859826997e-06, 8.97271988264947e-07, 7.97900554516673e-07, 7.01658407820105e-07, 6.01139973355203e-07, 4.97605086025033e-07, 3.96218538867288e-07, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -1.42108547152020e-14, -1.42108547152020e-14, -2.84217094304040e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14],
[ 1.07000000000000e+02, 1.06999999862556e+02, 1.06999995612530e+02, 1.06999966817924e+02, 1.06999860965142e+02, 1.06999578794607e+02, 1.06998961241440e+02, 1.06997778438878e+02, 1.06995721177487e+02, 1.06992395135858e+02, 1.06987318111354e+02, 1.06979920386874e+02, 1.06969548276420e+02, 1.06955470803895e+02, 1.06936889391347e+02, 1.06912950369112e+02, 1.06882760073170e+02, 1.06845402263648e+02, 1.06799957577847e+02, 1.06745524712861e+02, 1.06681243004941e+02, 1.06606316023380e+02, 1.06520035716395e+02, 1.06421806532992e+02, 1.06311168805551e+02, 1.06187820531986e+02, 1.06051636572635e+02, 1.05902684209861e+02, 1.05741234039649e+02, 1.05567765296028e+02, 1.05382964954975e+02, 1.05187720307678e+02, 1.04983105096816e+02, 1.04770359723955e+02, 1.04550866407373e+02, 1.04326120451476e+02, 1.04097698951054e+02, 1.03867228286910e+02, 1.03636351685511e+02, 1.03406697942208e+02, 1.03179852181834e+02, 1.02957329289542e+02, 1.02740550419731e+02, 1.02530822802966e+02, 1.02329322929531e+02, 1.02137083093472e+02, 1.01954981225233e+02, 1.01783733913254e+02, 1.01623892503630e+02, 1.01475842162008e+02, 1.01339803776137e+02, 1.01215838566460e+02, 1.01103855254078e+02, 1.01003619610492e+02, 1.00914766182818e+02, 1.00836811953679e+02, 1.00769171658571e+02, 1.00711174447218e+02, 1.00662081541249e+02, 1.00621104510258e+02, 1.00587423763900e+02, 1.00560206840829e+02, 1.00538626067933e+02, 1.00521875166847e+02, 1.00509184400554e+02, 1.00499833881787e+02, 1.00493164707108e+02, 1.00488587635655e+02, 1.00485589098193e+02, 1.00483734398609e+02, 1.00482668053466e+02, 1.00482111302847e+02, 1.00481856913786e+02, 1.00481761482841e+02, 1.00481735523232e+02, 1.00481731691645e+02, 1.00481731189022e+02, 1.00481719478501e+02, 1.00481640137988e+02, 1.00481348473707e+02, 1.00480571002575e+02, 1.00478869519703e+02, 1.00475610952238e+02, 1.00469944200226e+02, 1.00460785018439e+02, 1.00446809822712e+02, 1.00426459106335e+02, 1.00397950920755e+02, 1.00359304604089e+02, 1.00308374626008e+02, 1.00242894057097e+02, 1.00160526769347e+02, 1.00058927044543e+02, 9.99358048312450e+01, 9.97889944820222e+01, 9.96165244633399e+01, 9.94166853097447e+01, 9.91880930375119e+01, 9.89297453708004e+01, 9.86410684640589e+01, 9.83219522826578e+01, 9.79727733409758e+01, 9.75944039845371e+01, 9.71882077590228e+01, 9.67560206408823e+01, 9.63001181464748e+01, 9.58231688226057e+01, 9.53281755532778e+01, 9.48184074675396e+01, 9.42973266067040e+01, 9.37685142446226e+01, 9.32356013078972e+01, 9.27022057225389e+01, 9.21718773832660e+01, 9.16480497317470e+01, 9.11339962546657e+01, 9.06327905634654e+01, 9.01472696159858e+01, 8.96800005218222e+01, 8.92332518858107e+01, 8.88089707168302e+01, 8.84087656803283e+01, 8.80338970797936e+01, 8.76852735619611e+01, 8.73634552369316e+01, 8.70686627134773e+01, 8.68007914595404e+01, 8.65594308788606e+01, 8.63438875147288e+01, 8.61532118243384e+01, 8.59862279936782e+01, 8.58415662730661e+01, 8.57176973036330e+01, 8.56129678765068e+01, 8.55256375235855e+01, 8.54539152880986e+01, 8.53959959722327e+01, 8.53500951159447e+01, 8.53144819334834e+01, 8.52875094290859e+01, 8.52676409364353e+01, 8.52534723815323e+01, 8.52437496571223e+01, 8.52373806176959e+01, 8.52334413537493e+01, 8.52311765765343e+01, 8.52299941321452e+01, 8.52294538574684e+01, 8.52292511808029e+01, 8.52291960477430e+01, 8.52291879102155e+01, 8.52291871918267e+01, 8.52291731153523e+01, 8.52290777451577e+01, 8.52287271540914e+01, 8.52277926059346e+01, 8.52257473646012e+01, 8.52218304615014e+01, 8.52150188675698e+01, 8.52040093420733e+01, 8.51872110279571e+01, 8.51627496284618e+01, 8.51284837253588e+01, 8.50820334783970e+01, 8.50208215732206e+01, 8.49421258599906e+01, 8.48431426525286e+01, 8.47210591514765e+01, 8.45731329374579e+01, 8.43967759832945e+01, 8.41896401972821e+01, 8.39497011763151e+01, 8.36753366628030e+01, 8.33653962030834e+01, 8.30192587277841e+01, 8.26368752308330e+01, 8.22187944057031e+01, 8.17661699687335e+01, 8.12807493916838e+01, 8.07648447810870e+01, 8.02212875649325e+01, 7.96533693676641e+01, 7.90647718975961e+01, 7.84594888233092e+01, 7.78417425343344e+01, 7.72158984751555e+01, 7.65863795261321e+01, 7.59575827502050e+01, 7.53338007193816e+01, 7.47191494962194e+01, 7.41175050677757e+01, 7.35324495556664e+01, 7.29672278912162e+01, 7.24247149673274e+01, 7.19073927027377e+01, 7.14173360827860e+01, 7.09562070984238e+01, 7.05252555456793e+01, 7.01253257903889e+01, 6.97568687716442e+01, 6.94199586630364e+01, 6.91143137142057e+01, 6.88393208576102e+01, 6.85940636964510e+01, 6.83773534987996e+01, 6.81877628158557e+01, 6.80236613208242e+01, 6.78832534291475e+01, 6.77646172109001e+01, 6.76657440438050e+01, 6.75845783846771e+01, 6.75190569645292e+01, 6.74671466462612e+01, 6.74268801328818e+01, 6.73963886875606e+01, 6.73739310323620e+01, 6.73579176360638e+01, 6.73469296860357e+01, 6.73397321644784e+01, 6.73352806117414e+01, 6.73327213520992e+01, 6.73313851708469e+01, 6.73307746546974e+01, 6.73305456283816e+01, 6.73304833276110e+01, 6.73304741321456e+01, 6.73304733945556e+01, 6.73304597822777e+01, 6.73303675570912e+01, 6.73300285275277e+01, 6.73291247990190e+01, 6.73271470104322e+01, 6.73233593054240e+01, 6.73167724444273e+01, 6.73061262984564e+01, 6.72898827745183e+01, 6.72662299988778e+01, 6.72330983213137e+01, 6.71881883930296e+01, 6.71290112071629e+01, 6.70529395723841e+01, 6.69572700222659e+01, 6.68392936601102e+01, 6.66963739248468e+01, 6.65260287720320e+01, 6.63260143358737e+01, 6.60944068178877e+01, 6.58296791773107e+01, 6.55307692110998e+01, 6.51971358260663e+01, 6.48288007228217e+01, 6.44263733123178e+01, 6.39910574366106e+01, 6.35246393216352e+01, 6.30294571020812e+01, 6.25083531757892e+01, 6.19646115132886e+01, 6.14018828069281e+01, 6.08241009266902e+01, 6.02353944894384e+01, 5.96399973932464e+01, 5.90421619009519e+01, 5.84460773083474e+01, 5.78557964841706e+01, 5.72751717385158e+01, 5.67078006882155e+01, 5.61569821433642e+01, 5.56256815934589e+01, 5.51165056266596e+01, 5.46316845308793e+01, 5.41730623389653e+01, 5.37420936323243e+01, 5.33398464666974e+01, 5.29670108134561e+01, 5.26239119227273e+01, 5.23105280238844e+01, 5.20265117973847e+01, 5.17712150858833e+01, 5.15437163599257e+01, 5.13428505058019e+01, 5.11672405491036e+01, 5.10153309568220e+01, 5.08854221664394e+01, 5.07757059698488e+01, 5.06843013349478e+01, 5.06092901840520e+01, 5.05487525743147e+01, 5.05008006513076e+01, 5.04636106836258e+01, 5.04354524442536e+01, 5.04147151924805e+01, 5.03999295351243e+01, 5.03897845114633e+01, 5.03831393529175e+01, 5.03790295129255e+01, 5.03766667381142e+01, 5.03754331494969e+01, 5.03748695110180e+01, 5.03746580704985e+01, 5.03746005535253e+01, 5.03745920641396e+01, 5.03745914093565e+01, 5.03745796515538e+01, 5.03744999907162e+01, 5.03742071491396e+01, 5.03734265416962e+01, 5.03717182048663e+01, 5.03684465487532e+01, 5.03627571529127e+01, 5.03535616884014e+01, 5.03395318872245e+01, 5.03191032910463e+01, 5.02904892863884e+01, 5.02517056660294e+01, 5.02006056400951e+01, 5.01349248534861e+01, 5.00523355530414e+01, 4.99505086002053e+01, 4.98271815637339e+01, 4.96802306813992e+01, 4.95077440857878e+01, 4.93080933869553e+01, 4.90800005331212e+01, 4.88225968636937e+01, 4.85354714505742e+01, 4.82187062039772e+01, 4.78728957915045e+01, 4.74991511592387e+01, 4.70990863069096e+01, 4.66747888910741e+01, 4.62287761272639e+01, 4.57639382390036e+01, 4.52834722673811e+01, 4.47908093457716e+01, 4.42895385471775e+01, 4.37833301735882e+01, 4.32758609720307e+01, 4.27707433373911e+01, 4.22714601769170e+01, 4.17813067924222e+01, 4.13033408604472e+01, 4.08403413105862e+01, 4.03947765788459e+01, 3.99687823380283e+01, 3.95641484070198e+01, 3.91823141639345e+01, 3.88243714855010e+01, 3.84910740438523e+01, 3.81828517257629e+01, 3.78998289887412e+01, 3.76418461063620e+01, 3.74084824464622e+01, 3.71990811347807e+01, 3.70127746534217e+01, 3.68485110869111e+01, 3.67050808464796e+01, 3.65811437713710e+01, 3.64752565264116e+01, 3.63859001939173e+01, 3.63115079040179e+01, 3.62504922707976e+01, 3.62012723131361e+01, 3.61622994496771e+01, 3.61320820775190e+01, 3.61092081836571e+01, 3.60923654051563e+01, 3.60803579546679e+01, 3.60721198658228e+01, 3.60667240887935e+01, 3.60633870773120e+01, 3.60614686490064e+01, 3.60604670629020e+01, 3.60600094313910e+01, 3.60598377582471e+01, 3.60597910590117e+01, 3.60597841663031e+01, 3.60597836444081e+01, 3.60597743990893e+01, 3.60597117607151e+01, 3.60594814957356e+01, 3.60588676962236e+01, 3.60575244230894e+01, 3.60549519406485e+01, 3.60504785170202e+01, 3.60432486642758e+01, 3.60322185742917e+01, 3.60161593649959e+01, 3.59936685793046e+01, 3.59631901672785e+01, 3.59230429241049e+01, 3.58714570492815e+01, 3.58066181383713e+01, 3.57267175278488e+01, 3.56300075042699e+01, 3.55148594881357e+01, 3.53798229445708e+01, 3.52236824962535e+01, 3.50455105588973e+01, 3.48447128222131e+01, 3.46210640867802e+01, 3.43747323522260e+01, 3.41062896278835e+01, 3.38167086743729e+01, 3.35073457304304e+01, 3.31799101595492e+01, 3.28364227765863e+01, 3.24791652925520e+01, 3.21106237642876e+01, 3.17334290993773e+01, 3.13502975304318e+01, 3.09639735689834e+01, 3.05771773536586e+01, 3.01925576258253e+01, 2.98126509117232e+01, 2.94398469587826e+01, 2.90763601230826e+01, 2.87242062445276e+01, 2.83851845419463e+01, 2.80608641487281e+01, 2.77525750207693e+01, 2.74614030265448e+01, 2.71881890458901e+01, 2.69335318614788e+01, 2.66977945491529e+01, 2.64811139929791e+01, 2.62834130962309e+01, 2.61044152452570e+01, 2.59436606092395e+01, 2.58005239134526e+01, 2.56742333897945e+01, 2.55638906693933e+01, 2.54684914256200e+01, 2.53869465951717e+01, 2.53181039988057e+01, 2.52607701547943e+01, 2.52137320330440e+01, 2.51757784434279e+01, 2.51457206961967e+01, 2.51224121232300e+01, 2.51047660136001e+01, 2.50917715015427e+01, 2.50825069539610e+01, 2.50761504405825e+01, 2.50719869331322e+01, 2.50694119683699e+01, 2.50679316193458e+01, 2.50671587437500e+01, 2.50668056105080e+01, 2.50666731380376e+01, 2.50666371023137e+01, 2.50666317835163e+01, 2.50666313842325e+01, 2.50666243563852e+01, 2.50665767416926e+01, 2.50664017051538e+01, 2.50659351228366e+01, 2.50649140230612e+01, 2.50629585109824e+01, 2.50595579095988e+01, 2.50540617693655e+01, 2.50456763041990e+01, 2.50334667001583e+01, 2.50163656096138e+01, 2.49931879831706e+01, 2.49626521999324e+01, 2.49234072322845e+01, 2.48740653265018e+01, 2.48132394023955e+01, 2.47395840867685e+01, 2.46518390150617e+01, 2.45488727862687e+01, 2.44297257638681e+01, 2.42936498064888e+01, 2.41401430101593e+01, 2.39689776678088e+01, 2.37802199113905e+01, 2.35742398965051e+01, 2.33517119030842e+01, 2.31136043249802e+01, 2.28611601525184e+01, 2.25958691432061e+01, 2.23194333457114e+01, 2.20337279186445e+01, 2.17407592284924e+01, 2.14426220314059e+01, 2.11414572071213e+01, 2.08394111216068e+01, 2.05385973502280e+01, 2.02410612623303e+01, 1.99487478637512e+01, 1.96634732791811e+01, 1.93869002694220e+01, 1.91205181610884e+01, 1.88656274844271e+01, 1.86233294661940e+01, 1.83945203315367e+01, 1.81798901666284e+01, 1.79799259168863e+01, 1.77949179689713e+01, 1.76249697000066e+01, 1.74700093734002e+01, 1.73298038066333e+01, 1.72039733163986e+01, 1.70920075431570e+01, 1.69932818546353e+01, 1.69070741133847e+01, 1.68325816586283e+01, 1.67689383926790e+01, 1.67152318762687e+01, 1.66705203273157e+01, 1.66338493884403e+01, 1.66042684860688e+01, 1.65808465553988e+01, 1.65626868583282e+01, 1.65489405828824e+01, 1.65388188889913e+01, 1.65316030615077e+01, 1.65266524500715e+01, 1.65234099176167e+01, 1.65214045835987e+01, 1.65202517309647e+01, 1.65196498424334e+01, 1.65193748356063e+01, 1.65192716712153e+01, 1.65192436080150e+01, 1.65192394659460e+01, 1.65192391559452e+01, 1.65192337121746e+01, 1.65191968298398e+01, 1.65190612466532e+01, 1.65186998330213e+01, 1.65179088950513e+01, 1.65163941807544e+01, 1.65137601629612e+01, 1.65095031131663e+01, 1.65030084111033e+01, 1.64935524520602e+01, 1.64803094134609e+01, 1.64623630195983e+01, 1.64387232950484e+01, 1.64083481216894e+01, 1.63701692129994e+01, 1.63231218979031e+01, 1.62661778747417e+01, 1.61983798680101e+01, 1.61188769139198e+01, 1.60269588353196e+01, 1.59220883621039e+01, 1.58039293281704e+01, 1.56723694444025e+01, 1.55275363170238e+01, 1.53698056521461e+01, 1.51998009514717e+01, 1.50183844423897e+01, 1.48266394701738e+01, 1.46258450745658e+01, 1.44174439360494e+01, 1.42030052654217e+01, 1.39841844846635e+01, 1.37626816783837e+01, 1.35402007695503e+01, 1.33184111959361e+01, 1.30989135587111e+01, 1.28832103212493e+01, 1.26726822024814e+01, 1.24685704835798e+01, 1.22719650709112e+01, 1.20837978610120e+01, 1.19048407490209e+01, 1.17357075107170e+01, 1.15768587593694e+01, 1.14286092147202e+01, 1.12911366028774e+01, 1.11644916138639e+01, 1.10486084621542e+01, 1.09433157126227e+01, 1.08483471415194e+01, 1.07633524941133e+01, 1.06879080746035e+01, 1.06215271585698e+01, 1.05636702535610e+01, 1.05137552501246e+01, 1.04711675050954e+01, 1.04352698832745e+01, 1.04054127553627e+01, 1.03809439123645e+01, 1.03612183134234e+01, 1.03456075393528e+01, 1.03335087824316e+01, 1.03243531686665e+01, 1.03176131856548e+01, 1.03128089805565e+01, 1.03095133005624e+01, 1.03073548732838e+01, 1.03060200658455e+01, 1.03052527168233e+01, 1.03048521009675e+01, 1.03046690584406e+01, 1.03046003932199e+01, 1.03045817146481e+01, 1.03045789577312e+01, 1.03045787513207e+01, 1.03045751256175e+01, 1.03045505609622e+01, 1.03044602589541e+01, 1.03042195489790e+01, 1.03036927713561e+01, 1.03026839685429e+01, 1.03009297751327e+01, 1.02980948628380e+01, 1.02937702555696e+01, 1.02874747787634e+01, 1.02786598419248e+01, 1.02667176693548e+01, 1.02509929873308e+01, 1.02307980445899e+01, 1.02054306877260e+01, 1.01741950388369e+01, 1.01364241386939e+01, 1.00915037384988e+01, 1.00388962642530e+01, 9.97816385927007e+00, 9.90898935146237e+00, 9.83119400862508e+00, 9.74475104674953e+00, 9.64979404448007e+00, 9.54661968208946e+00, 9.43568454657159e+00, 9.31759609810371e+00, 9.19309824471274e+00, 9.06305228848872e+00, 8.92841425921017e+00, 8.79020981888054e+00, 8.64950799287884e+00, 8.50739496053427e+00, 8.36494902993010e+00, 8.22321774585561e+00, 8.08319785767156e+00, 7.94581862866366e+00, 7.81192872272931e+00, 7.68228667783680e+00, 7.55755478432512e+00, 7.43829604020966e+00, 7.32497375982855e+00, 7.21795336530076e+00, 7.11750588651927e+00, 7.02381272568510e+00, 6.93697129621482e+00, 6.85700121296431e+00, 6.78385078229400e+00, 6.71740360976908e+00, 6.65748520558526e+00, 6.60386952031595e+00, 6.55628538487080e+00, 6.51442285818572e+00, 6.47793950430493e+00, 6.44646662775419e+00, 6.41961549334583e+00, 6.39698354501664e+00, 6.37816061950774e+00, 6.36273512654182e+00, 6.35030013985900e+00, 6.34045931554060e+00, 6.33283252814887e+00, 6.32706109398902e+00, 6.32281243668223e+00, 6.31978404517959e+00, 6.31770657961933e+00, 6.31634599645214e+00, 6.31550459051176e+00, 6.31502088674470e+00, 6.31476835590603e+00, 6.31465297387740e+00, 6.31460969032103e+00, 6.31459791616557e+00, 6.31459617832570e+00, 6.31459604786839e+00, 6.31459375171049e+00, 6.31457819491453e+00, 6.31452100665911e+00, 6.31436856508961e+00, 6.31403495715550e+00, 6.31339608377114e+00, 6.31228515807318e+00, 6.31048982284642e+00, 6.30775108734359e+00, 6.30376425156400e+00, 6.29818194475145e+00, 6.29061935143318e+00, 6.28066163021823e+00, 6.26787344647730e+00, 6.25181044053288e+00, 6.23203234125126e+00, 6.20811731703253e+00, 6.17967704113519e+00, 6.14637184742936e+00, 6.10792527871007e+00, 6.06413729495146e+00, 6.01489542340151e+00, 5.96018320199299e+00, 5.90008539208513e+00, 5.83478960920486e+00, 5.76458422783934e+00, 5.68985263991057e+00, 5.61106416509246e+00, 5.52876210370344e+00, 5.44354957202189e+00, 5.35607385386157e+00, 5.26701003656858e+00, 5.17704467680167e+00, 5.08686016989637e+00, 4.99712038846448e+00, 4.90845802490371e+00, 4.82146393210246e+00, 4.73667861853926e+00, 4.65458592750335e+00, 4.57560882197940e+00, 4.50010711100667e+00, 4.42837689191524e+00, 4.36065144566344e+00, 4.29710330779865e+00, 4.23784724231631e+00, 4.18294386599885e+00, 4.13240370234604e+00, 4.08619148259019e+00, 4.04423055247895e+00, 4.00640728400496e+00, 3.97257542828308e+00, 3.94256037726477e+00, 3.91616332659662e+00, 3.89316534896538e+00, 3.87333139656448e+00, 3.85641425316845e+00, 3.84215845141729e+00, 3.83030416034003e+00, 3.82059103324538e+00, 3.81276198848074e+00, 3.80656687700903e+00, 3.80176597315540e+00, 3.79813321009343e+00, 3.79545907134707e+00, 3.79355304513776e+00, 3.79224555068618e+00, 3.79138925489546e+00, 3.79085971388071e+00, 3.79055529563648e+00, 3.79039636630404e+00, 3.79032375118989e+00, 3.79029651091710e+00, 3.79028910091786e+00, 3.79028800721804e+00, 3.79028792476579e+00, 3.79028646887936e+00, 3.79027660504293e+00, 3.79024034466134e+00, 3.79014368875370e+00, 3.78993216451818e+00, 3.78952708890331e+00, 3.78882271721773e+00, 3.78768442158804e+00, 3.78594802826949e+00, 3.78342042295432e+00, 3.77988150740160e+00, 3.77508755691170e+00, 3.76877598468181e+00, 3.76067146497570e+00, 3.75049330266410e+00, 3.73796386402586e+00, 3.72281780665318e+00, 3.70481177066115e+00, 3.68373412651428e+00, 3.65941432488063e+00, 3.63173136910301e+00, 3.60062093789551e+00, 3.56608072889354e+00, 3.52817367323159e+00, 3.48702878365903e+00, 3.44283953583006e+00, 3.39585983274530e+00, 3.34639775205135e+00, 3.29480741072295e+00, 3.24147938871721e+00, 3.18683022286109e+00, 3.13129150935082e+00, 3.07529913765686e+00, 3.01928312499986e+00, 2.96365843739028e+00, 2.90881708141285e+00, 2.85512164210006e+00, 2.80290033712289e+00, 2.75244356475839e+00, 2.70400184845641e+00, 2.65778502708037e+00, 2.61396250709250e+00, 2.57266437905923e+00, 2.53398320253980e+00, 2.49797627681048e+00, 2.46466823614126e+00, 2.43405383410226e+00, 2.40610080889279e+00, 2.38075274887420e+00, 2.35793190284232e+00, 2.33754190204887e+00, 2.31947037984911e+00, 2.30359148963913e+00, 2.28976833215619e+00, 2.27785530914035e+00, 2.26770042186192e+00, 2.25914753039140e+00, 2.25203858322413e+00, 2.24621581770454e+00, 2.24152392056124e+00, 2.23781212587336e+00, 2.23493621614374e+00, 2.23276038207365e+00, 2.23115888923667e+00, 2.23001749605379e+00, 2.22923456789717e+00, 2.22872183703814e+00, 2.22840476734362e+00, 2.22822249554604e+00, 2.22812733666188e+00, 2.22808385856284e+00, 2.22806754855323e+00, 2.22806311184708e+00, 2.22806245699897e+00, 2.22806240736239e+00, 2.22806152734930e+00, 2.22805556513747e+00, 2.22803364751108e+00, 2.22797522389585e+00, 2.22784736894540e+00, 2.22760252560734e+00, 2.22717678538480e+00, 2.22648879519851e+00, 2.22543937142818e+00, 2.22391189034364e+00, 2.22177350890098e+00, 2.21887724938236e+00, 2.21506495440740e+00, 2.21017108477464e+00, 2.20402729158759e+00, 2.19646764748805e+00, 2.18733437214566e+00, 2.17648383829760e+00, 2.16379260148936e+00, 2.14916316468604e+00, 2.13252917340897e+00, 2.11385974232539e+00, 2.09316264277878e+00, 2.07048613259225e+00, 2.04591928177152e+00, 2.01959073492839e+00, 1.99166594569010e+00, 1.96234301133960e+00, 1.93184731890597e+00, 1.90042527974981e+00, 1.86833747352061e+00, 1.83585154207169e+00, 1.80323516996063e+00, 1.77074946299999e+00, 1.73864299364398e+00, 1.70714672600454e+00, 1.67646996828999e+00, 1.64679743078413e+00, 1.61828739765507e+00, 1.59107095569136e+00, 1.56525216739087e+00, 1.54090903403217e+00, 1.51809506948119e+00, 1.49684129850867e+00, 1.47715850296171e+00, 1.45903956183263e+00, 1.44246176241192e+00, 1.42738899433550e+00, 1.41377377206881e+00, 1.40155906102807e+00, 1.39067990626751e+00, 1.38106487982711e+00, 1.37263737368421e+00, 1.36531677056463e+00, 1.35901952564551e+00, 1.35366018944049e+00, 1.34915239681689e+00, 1.34540983995056e+00, 1.34234723476592e+00, 1.33988128164849e+00, 1.33793161254296e+00, 1.33642170853092e+00, 1.33527976517253e+00, 1.33443947781497e+00, 1.33384071615158e+00, 1.33343005685910e+00, 1.33316114527209e+00, 1.33299486168397e+00, 1.33289927468105e+00, 1.33284937242071e+00, 1.33282657229815e+00, 1.33281801928732e+00, 1.33281569267065e+00, 1.33281534926725e+00, 1.33281532307261e+00, 1.33281485648510e+00, 1.33281169529118e+00, 1.33280007448259e+00, 1.33276909824734e+00, 1.33270131042595e+00, 1.33257149979535e+00, 1.33234579292730e+00, 1.33198108432039e+00, 1.33142484968346e+00, 1.33061538290946e+00, 1.32948248946622e+00, 1.32794865762376e+00, 1.32593071323753e+00, 1.32334194317139e+00, 1.32009464692726e+00, 1.31610304650292e+00, 1.31128645273307e+00, 1.30557255514925e+00, 1.29890067529287e+00, 1.29122480447958e+00, 1.28251624020321e+00, 1.27276564390296e+00, 1.26198436844943e+00, 1.25020494608172e+00, 1.23748068383721e+00, 1.22388437849322e+00, 1.20950622950964e+00, 1.19445108835002e+00, 1.17883522828058e+00, 1.16278284463826e+00, 1.14642249905576e+00, 1.12988370326404e+00, 1.11329380319050e+00, 1.09677527869223e+00, 1.08044352568878e+00, 1.06440514218342e+00, 1.04875670233777e+00, 1.03358397573203e+00, 1.01896153244081e+00, 1.00495266728413e+00, 9.91609576476350e-01, 9.78973724629839e-01, 9.67076347666207e-01, 9.55939046097825e-01, 9.45574432290186e-01, 9.35986804028104e-01, 9.27172824601016e-01, 9.19122196498407e-01, 9.11818321579190e-01, 9.05238945226841e-01, 8.99356785537663e-01, 8.94140151039949e-01, 8.89553551856224e-01, 8.85558309652566e-01, 8.82113171245308e-01, 8.79174929443366e-01, 8.76699052711942e-01, 8.74640322689444e-01, 8.72953475647719e-01, 8.71593840860186e-01, 8.70517965764137e-01, 8.69684215026027e-01, 8.69053328392539e-01, 8.68588920778748e-01, 8.68257907595194e-01, 8.68030838986215e-01, 8.67882128479224e-01, 8.67790164479231e-01, 8.67737296930187e-01, 8.67709696064665e-01, 8.67697085174640e-01, 8.67692354422488e-01, 8.67691067545834e-01, 8.67690877605639e-01, 8.67690863014843e-01, 8.67690601778591e-01, 8.67688831866261e-01, 8.67682325505385e-01, 8.67664982120772e-01, 8.67627027485128e-01, 8.67554343408770e-01, 8.67427956426042e-01, 8.67223711616077e-01, 8.66912154904469e-01, 8.66458643643981e-01, 8.65823700588564e-01, 8.64963620155507e-01, 8.63831327774164e-01, 8.62377482956873e-01, 8.60551804573319e-01, 8.58304583083751e-01, 8.55588330014854e-01, 8.52359500968802e-01, 8.48580216486383e-01, 8.44219896825877e-01, 8.39256723767065e-01, 8.33678846113145e-01, 8.27485256185653e-01, 8.20686281975270e-01, 8.13303662483136e-01, 8.05370200090906e-01, 7.96929010920707e-01, 7.88032419291582e-01, 7.78740563018900e-01, 7.69119790530908e-01, 7.59240937569786e-01, 7.49177570539970e-01, 7.39004276184133e-01, 7.28795064698360e-01, 7.18621937519643e-01, 7.08553653800223e-01, 6.98654712820272e-01, 6.88984554719426e-01, 6.79596969911117e-01, 6.70539698860097e-01, 6.61854198583157e-01, 6.53575549961502e-01, 6.45732480180158e-01, 6.38347476679527e-01, 6.31436972240039e-01, 6.25011584624417e-01, 6.19076398079073e-01, 6.13631277593271e-01, 6.08671209898432e-01, 6.04186667640164e-01, 6.00163994928195e-01, 5.96585813575942e-01, 5.93431449830391e-01, 5.90677381320873e-01, 5.88297703400357e-01, 5.86264613085447e-01, 5.84548907511618e-01, 5.83120492293645e-01, 5.81948893531987e-01, 5.81003765544025e-01, 5.80255384862824e-01, 5.79675119770002e-01, 5.79235863749005e-01, 5.78912420884492e-01, 5.78681831487842e-01, 5.78523627152620e-01, 5.78420006047125e-01, 5.78355921483862e-01, 5.78319079567478e-01, 5.78299844866592e-01, 5.78291056395656e-01, 5.78287759540487e-01, 5.78286862716269e-01, 5.78286730346903e-01, 5.78286720100948e-01, 5.78286535645759e-01, 5.78285285935152e-01, 5.78280691873559e-01, 5.78268445844657e-01, 5.78241645976618e-01, 5.78190322044875e-01, 5.78101072492615e-01, 5.77956830062234e-01, 5.77736771190999e-01, 5.77416381502687e-01, 5.76967686249546e-01, 5.76359650243887e-01, 5.75558746509174e-01, 5.74529686511809e-01, 5.73236297464547e-01, 5.71642524069375e-01, 5.69713523639379e-01, 5.67416815463375e-01, 5.64723438371118e-01, 5.61609065622591e-01, 5.58055024331168e-01, 5.54049168301717e-01, 5.49586558739236e-01, 5.44669916637815e-01, 5.39309823168693e-01, 5.33524658976319e-01, 5.27340288563806e-01, 5.20789510382698e-01, 5.13911305431066e-01, 5.06749926023105e-01, 4.99353871336211e-01, 4.91774797306221e-01, 4.84066405887773e-01, 4.76283353443009e-01, 4.68480211084554e-01, 4.60710502192867e-01, 4.53025834910548e-01, 4.45475140789270e-01, 4.38104025256195e-01, 4.30954231247512e-01, 4.24063214136538e-01, 4.17463823766013e-01, 4.11184087766117e-01, 4.05247089218278e-01, 3.99670930976527e-01, 3.94468778519766e-01, 3.89648973044515e-01, 3.85215206613253e-01, 3.81166751521192e-01, 3.77498736597431e-01, 3.74202463839765e-01, 3.71265759511076e-01, 3.68673354496096e-01, 3.66407289237529e-01, 3.64447338854418e-01, 3.62771454042232e-01, 3.61356213041901e-01, 3.60177279366823e-01, 3.59209859154461e-01, 3.58429151059582e-01, 3.57810780651590e-01, 3.57331210453054e-01, 3.56968116195659e-01, 3.56700719690963e-01, 3.56510069009431e-01, 3.56379257482672e-01, 3.56293574400283e-01, 3.56240582122425e-01, 3.56210116590418e-01, 3.56194210769132e-01, 3.56186943252496e-01, 3.56184216953594e-01, 3.56183475333950e-01, 3.56183365872440e-01, 3.56183357330963e-01, 3.56183202673577e-01, 3.56182154847062e-01, 3.56178302928242e-01, 3.56168035144336e-01, 3.56145564438606e-01, 3.56102530687195e-01, 3.56027695690997e-01, 3.55906745284869e-01, 3.55722210824595e-01, 3.55453519858230e-01, 3.55077182852796e-01, 3.54567119325168e-01, 3.53895122529622e-01, 3.53031456955691e-01, 3.51945577345433e-01, 3.50606951919104e-01, 3.48985966294791e-01, 3.47054878615836e-01, 3.44788791168540e-01, 3.42166599835295e-01, 3.39171880617386e-01, 3.35793672607196e-01, 3.32027119444334e-01, 3.27873936473225e-01, 3.23342678291453e-01, 3.18448790654315e-01, 3.13214441115349e-01, 3.07668133563979e-01, 3.01844122198062e-01, 2.95781649749628e-01, 2.89524042416716e-01, 2.83117699566660e-01, 2.76611019646623e-01, 2.70053304803284e-01, 2.63493685524139e-01, 2.56980103340098e-01, 2.50558384540369e-01, 2.44271431322943e-01, 2.38158549296671e-01, 2.32254922286486e-01, 2.26591237520068e-01, 2.21193457016355e-01, 2.16082724817326e-01, 2.11275394943755e-01, 2.06783161808787e-01, 2.02613273317695e-01, 1.98768806891351e-01, 1.95248989934811e-01, 1.92049548516408e-01, 1.89163070884433e-01, 1.86579375582880e-01, 1.84285877025346e-01, 1.82267944182257e-01, 1.80509250329536e-01, 1.78992113473527e-01, 1.77697828044955e-01, 1.76606988758394e-01, 1.75699807219459e-01, 1.74956421045408e-01, 1.74357194082859e-01, 1.73883004935846e-01, 1.73515519628793e-01, 1.73237443007125e-01, 1.73032742574847e-01, 1.72886838013198e-01, 1.72786749696471e-01, 1.72721200153617e-01, 1.72680663589873e-01, 1.72657360211801e-01, 1.72645194073525e-01, 1.72639635338726e-01, 1.72637550077411e-01, 1.72636982836906e-01, 1.72636899113428e-01, 1.72636892524750e-01, 1.72636772513712e-01, 1.72635959421953e-01, 1.72632970422640e-01, 1.72625002939725e-01, 1.72607566764341e-01, 1.72574176188874e-01, 1.72516115289000e-01, 1.72422288454968e-01, 1.72279167041467e-01, 1.72070841439378e-01, 1.71779185814373e-01, 1.71384140030014e-01, 1.70864109736755e-01, 1.70196481179985e-01, 1.69358241983460e-01, 1.68326693145445e-01, 1.67080231039563e-01, 1.65599171780784e-01, 1.63866584457224e-01, 1.61869095075957e-01, 1.59597620273985e-01, 1.57047989477761e-01, 1.54221416677444e-01, 1.51124788510231e-01, 1.47770743819663e-01, 1.44177530868546e-01, 1.40368641206493e-01, 1.36372232849425e-01, 1.32220368749955e-01, 1.27948108312850e-01, 1.23592498819121e-01, 1.19191519196420e-01, 1.14783030106366e-01, 1.10403781719612e-01, 1.06088524158181e-01, 1.01869256114213e-01, 9.77746356401212e-02, 9.38295647575274e-02, 9.00549475984320e-02, 8.64676113243803e-02, 8.30803708232395e-02, 7.99022125302713e-02, 7.69385697257263e-02, 7.41916611386699e-02, 7.16608662779521e-02, 6.93431141063172e-02, 6.72332658873671e-02, 6.53244776832764e-02, 6.36085325816680e-02, 6.20761369554685e-02, 6.07171787123235e-02, 5.95209484696113e-02, 5.84763268446551e-02, 5.75719425705046e-02, 5.67963069642445e-02, 5.61379304682816e-02, 5.55854266457203e-02, 5.51276082555603e-02, 5.47535789602875e-02, 5.44528229382877e-02, 5.42152932861484e-02, 5.40314987160002e-02, 5.38925867843858e-02, 5.37904208425033e-02, 5.37176471537038e-02, 5.36677482524226e-02, 5.36350786445325e-02, 5.36148793709117e-02, 5.36032687276986e-02, 5.35972074864191e-02, 5.35944381832820e-02, 5.35933993428017e-02, 5.35931167549535e-02, 5.35930750456544e-02, 5.35930717342552e-02, 5.35930110487612e-02, 5.35925998968594e-02, 5.35910884760114e-02, 5.35870597141752e-02, 5.35782435333232e-02, 5.35613620437658e-02, 5.35320129757224e-02, 5.34845985128811e-02, 5.34123064690135e-02, 5.33071502931506e-02, 5.31600735158833e-02, 5.29611227596689e-02, 5.26996911521280e-02, 5.23648307910065e-02, 5.19456288041766e-02, 5.14316366645886e-02, 5.08133370596795e-02, 5.00826272507879e-02, 4.92332931083155e-02, 4.82614445707128e-02, 4.71658818459275e-02, 4.59483628250439e-02, 4.46137462442806e-02, 4.31699921079058e-02, 4.16280103751347e-02, 4.00013601236395e-02, 3.83058132164820e-02, 3.65588076112431e-02, 3.47788245328502e-02, 3.29847296582244e-02, 3.11951204867427e-02, 2.94277199928246e-02, 2.76988508504985e-02, 2.60230158578452e-02, 2.44125998594171e-02, 2.28776977483136e-02, 2.14260632126866e-02, 2.00631647355465e-02, 1.87923295983126e-02, 1.76149535375600e-02, 1.65307531171806e-02, 1.55380393499911e-02, 1.46339940077382e-02, 1.38149337850137e-02, 1.30765515089450e-02, 1.24141275128409e-02, 1.18227078016133e-02, 1.12972485066223e-02, 1.08327282578102e-02, 1.04242315169216e-02, 1.00670067442081e-02, 9.75650365855074e-03, 9.48839391980982e-03, 9.25857939434138e-03, 9.06319181751769e-03, 8.89858719261838e-03, 8.76133771660648e-03, 8.64822344387051e-03, 8.55622531555866e-03, 8.48252061571486e-03, 8.42448136960172e-03, 8.37967569188436e-03, 8.34587163545564e-03, 8.32104270789102e-03, 8.30337393337288e-03, 8.29126716364215e-03, 8.28334429120048e-03, 8.27844709752412e-03, 8.27563266241782e-03, 8.27416355352284e-03, 8.27349236942609e-03, 8.27324059545731e-03, 8.27317210780620e-03, 8.27316199918471e-03, 8.27316118930810e-03, 8.27314625452402e-03, 8.27304506972837e-03, 8.27267311204594e-03, 8.27168167339210e-03, 8.26951225862160e-03, 8.26535881307677e-03, 8.25813982575596e-03, 8.24648245971105e-03, 8.22872095015725e-03, 8.20291151227082e-03, 8.16686582568593e-03, 8.11820471297153e-03, 8.05443281660700e-03, 7.97303384827330e-03, 7.87158433156776e-03, 7.74788176157415e-03, 7.60008092514397e-03, 7.42683002094680e-03, 7.22739651584448e-03, 7.00177172987537e-03, 6.75074328943026e-03, 6.47592605504599e-03, 6.17974497311025e-03, 5.86536733798937e-03, 5.53658674502344e-03, 5.19766591946791e-03, 4.85314988972618e-03, 4.50766399966085e-03, 4.16571262672960e-03, 3.83149407498706e-03, 3.50874503725151e-03, 3.20062448155056e-03, 2.90964218544332e-03, 2.63763207672696e-03, 2.38576601840634e-03, 2.15460064151785e-03, 1.94414863459258e-03, 1.75396600745614e-03, 1.58324728191417e-03, 1.43092082866758e-03, 1.29573718574250e-03, 1.17634507017783e-03, 1.07135297525929e-03, 9.79377428436123e-04, 8.99080470674107e-04, 8.29198172056067e-04, 7.68560283717079e-04, 7.16100417932353e-04, 6.70857267067504e-04, 6.31969307448799e-04, 5.98666522961366e-04, 5.70262107629560e-04, 5.46145389307640e-04, 5.25775579092169e-04, 5.08675335368025e-04, 4.94423571097968e-04, 4.82647753955234e-04, 4.73016464781040e-04, 4.65232962497453e-04, 4.59030148320494e-04, 4.54166933309352e-04, 4.50425770992524e-04, 4.47611046169527e-04, 4.45548044567090e-04, 4.44082288950653e-04, 4.43079074287311e-04, 4.42423057890551e-04, 4.42017769756831e-04, 4.41784919047450e-04, 4.41663393445424e-04, 4.41607877348815e-04, 4.41587052947057e-04, 4.41581388358259e-04, 4.41580552276832e-04, 4.41580484633164e-04, 4.41579228976252e-04, 4.41570721790185e-04, 4.41539449724360e-04, 4.41456099181892e-04, 4.41273736029757e-04, 4.40924670499498e-04, 4.40318211019530e-04, 4.39339536313810e-04, 4.37849940993829e-04, 4.35688723882777e-04, 4.32676978661561e-04, 4.28623493675673e-04, 4.23332861572590e-04, 4.16615717554691e-04, 4.08300766210346e-04, 3.98247926511885e-04, 3.86361573703198e-04, 3.72602587304982e-04, 3.56997854311203e-04, 3.39646092257340e-04, 3.20719248335877e-04, 3.00459087753779e-04, 2.79168839355748e-04, 2.57200049233575e-04, 2.34935070736242e-04, 2.12765665196457e-04, 1.91068725257537e-04, 1.70182629346982e-04, 1.50390377726239e-04, 1.31912891063735e-04, 1.14908693717553e-04, 9.94743262197062e-05, 8.56421037696009e-05, 7.33725641310912e-05, 6.25575770953901e-05, 5.30702617709267e-05, 4.48425839181255e-05, 3.78542706442886e-05, 3.19834503699212e-05, 2.69880849685933e-05, 2.26357694828039e-05, 1.87180790334196e-05, 1.53416263231065e-05, 1.23255350246154e-05, 8.88586599501195e-06, 6.99880487786686e-06, 5.46601862083662e-06, 4.35000120546647e-06, 4.65278938577285e-06, 4.68592918423383e-06, 3.97177403499427e-06, 3.69317615422915e-06, 3.43923092316345e-06, 2.92304969207180e-06, 2.35602416864822e-06, 1.88003767220835e-06, 1.55286046776837e-06, 1.35954257984849e-06, 1.25518739935160e-06, 1.17557215162378e-06, 1.07763014511875e-06, 9.50994390791493e-07, 8.03363576551419e-07, 6.48523055701844e-07, 4.99721437563494e-07, -1.42108547152020e-14, -1.42108547152020e-14, -1.42108547152020e-14, -2.84217094304040e-14, -1.42108547152020e-14, -1.42108547152020e-14, -2.84217094304040e-14, -1.42108547152020e-14, -2.84217094304040e-14],
[ 1.08000000000000e+02, 1.07999999860715e+02, 1.07999995553752e+02, 1.07999966373389e+02, 1.07999859102511e+02, 1.07999573151765e+02, 1.07998947325263e+02, 1.07997748676590e+02, 1.07995663853414e+02, 1.07992293250493e+02, 1.07987148202609e+02, 1.07979651355104e+02, 1.07969140255315e+02, 1.07954874118762e+02, 1.07936043644943e+02, 1.07911783693270e+02, 1.07881188582349e+02, 1.07843329744435e+02, 1.07797275446476e+02, 1.07742112270674e+02, 1.07676968018992e+02, 1.07601035655079e+02, 1.07513597814277e+02, 1.07414051295307e+02, 1.07301930803856e+02, 1.07176931068303e+02, 1.07038926321162e+02, 1.06887986071640e+02, 1.06724386117925e+02, 1.06548613884344e+02, 1.06361367422241e+02, 1.06163547766276e+02, 1.05956244751726e+02, 1.05740716821565e+02, 1.05518365729414e+02, 1.05290707328806e+02, 1.05059339800000e+02, 1.04825910694437e+02, 1.04592084086456e+02, 1.04359508941464e+02, 1.04129789577280e+02, 1.03904458849083e+02, 1.03684954460124e+02, 1.03472598611040e+02, 1.03268581059752e+02, 1.03073945570498e+02, 1.02889579676757e+02, 1.02716207656920e+02, 1.02554386611779e+02, 1.02404505528872e+02, 1.02266787213251e+02, 1.02141292953009e+02, 1.02027929769293e+02, 1.01926460074783e+02, 1.01836513533102e+02, 1.01757600876245e+02, 1.01689129399972e+02, 1.01630419820130e+02, 1.01580724138162e+02, 1.01539244133308e+02, 1.01505150074262e+02, 1.01477599225999e+02, 1.01455753719991e+02, 1.01438797359621e+02, 1.01425950948577e+02, 1.01416485759280e+02, 1.01409734801025e+02, 1.01405101603297e+02, 1.01402066297193e+02, 1.01400188855275e+02, 1.01399109434724e+02, 1.01398545857325e+02, 1.01398288348997e+02, 1.01398191747898e+02, 1.01398165469978e+02, 1.01398161591409e+02, 1.01398161082623e+02, 1.01398149228510e+02, 1.01398068915142e+02, 1.01397773674544e+02, 1.01396986670280e+02, 1.01395264324469e+02, 1.01391965802308e+02, 1.01386229569765e+02, 1.01376958091696e+02, 1.01362811565283e+02, 1.01342211383380e+02, 1.01313353788200e+02, 1.01274233901585e+02, 1.01222679998939e+02, 1.01156397528540e+02, 1.01073021970962e+02, 1.00970179197622e+02, 1.00845551545894e+02, 1.00696947414589e+02, 1.00522371841551e+02, 1.00320095303699e+02, 1.00088717924932e+02, 9.98272264178222e+01, 9.95350414175147e+01, 9.92120533427156e+01, 9.88586454488416e+01, 9.84757032143723e+01, 9.80646095483170e+01, 9.76272255435891e+01, 9.71658567772566e+01, 9.66832057148018e+01, 9.61823118001052e+01, 9.56664822362144e+01, 9.51392178341987e+01, 9.46041389376644e+01, 9.40649158079825e+01, 9.35252060792093e+01, 9.29885997139158e+01, 9.24585702622256e+01, 9.19384307021655e+01, 9.14312926258756e+01, 9.09400284992677e+01, 9.04672375947722e+01, 9.00152166555320e+01, 8.95859363590234e+01, 8.91810243495541e+01, 8.88017551873755e+01, 8.84490471621455e+01, 8.81234656179297e+01, 8.78252322552954e+01, 8.75542397965286e+01, 8.73100713903849e+01, 8.70920241598320e+01, 8.68991363331055e+01, 8.67302174270447e+01, 8.65838809623828e+01, 8.64585791803577e+01, 8.63526392003352e+01, 8.62643000138864e+01, 8.61917496586426e+01, 8.61331618630225e+01, 8.60867314086752e+01, 8.60507074290949e+01, 8.60234238573706e+01, 8.60033262591304e+01, 8.59889943422152e+01, 8.59791595239605e+01, 8.59727170591254e+01, 8.59687323828426e+01, 8.59664414974821e+01, 8.59652454221251e+01, 8.59646989193094e+01, 8.59644939062479e+01, 8.59644381376310e+01, 8.59644299062967e+01, 8.59644291796265e+01, 8.59644149408829e+01, 8.59643184712907e+01, 8.59639638387266e+01, 8.59630185174501e+01, 8.59609496995580e+01, 8.59569876451040e+01, 8.59500975343905e+01, 8.59389611097335e+01, 8.59219691895059e+01, 8.58972258996924e+01, 8.58625651907072e+01, 8.58155798826952e+01, 8.57536631056039e+01, 8.56740615698360e+01, 8.55739396245713e+01, 8.54504525475818e+01, 8.53008269855987e+01, 8.51224459603514e+01, 8.49129354123486e+01, 8.46702489171749e+01, 8.43927470230956e+01, 8.40792676652008e+01, 8.37291843410827e+01, 8.33424492003228e+01, 8.29196188956684e+01, 8.24618619288618e+01, 8.19709472270211e+01, 8.14492147051527e+01, 8.08995294897822e+01, 8.03252221892852e+01, 7.97300180286396e+01, 7.91179578143861e+01, 7.84933136238492e+01, 7.78605019325206e+01, 7.72239967155469e+01, 7.65882449397343e+01, 7.59575867734704e+01, 7.53361826874689e+01, 7.47279492950071e+01, 7.41365052447843e+01, 7.35651277913715e+01, 7.30167199647913e+01, 7.24937876916059e+01, 7.19984258787856e+01, 7.15323123665342e+01, 7.10967087259892e+01, 7.06924670344668e+01, 7.03200419294471e+01, 6.99795073792764e+01, 6.96705777004998e+01, 6.93926324043668e+01, 6.91447444798814e+01, 6.89257117272701e+01, 6.87340907486021e+01, 6.85682331821950e+01, 6.84263237333873e+01, 6.83064195057284e+01, 6.82064900751118e+01, 6.81244576789228e+01, 6.80582368194006e+01, 6.80057725135008e+01, 6.79650763698863e+01, 6.79342596464637e+01, 6.79115624471581e+01, 6.78953782603071e+01, 6.78842731262536e+01, 6.78769988481606e+01, 6.78724998240824e+01, 6.78699132729751e+01, 6.78685628430728e+01, 6.78679458165870e+01, 6.78677143480176e+01, 6.78676513828946e+01, 6.78676420893723e+01, 6.78676413439168e+01, 6.78676275864827e+01, 6.78675343778414e+01, 6.78671917329978e+01, 6.78662783675073e+01, 6.78642794887618e+01, 6.78604513941151e+01, 6.78537942968836e+01, 6.78430346350163e+01, 6.78266179239331e+01, 6.78027129892559e+01, 6.77692281493587e+01, 6.77238396039392e+01, 6.76640319172180e+01, 6.75871500613617e+01, 6.74904620127289e+01, 6.73712303848547e+01, 6.72267910620155e+01, 6.70546362996313e+01, 6.68524993239583e+01, 6.66184371381865e+01, 6.63509080676191e+01, 6.60488405868914e+01, 6.57116901866250e+01, 6.53394814567974e+01, 6.49328331720688e+01, 6.44929649272268e+01, 6.40216847448703e+01, 6.35213580125997e+01, 6.29948590495362e+01, 6.24455074932148e+01, 6.18769924717669e+01, 6.12932881108547e+01, 6.06985642507758e+01, 6.00970962665514e+01, 5.94931775812460e+01, 5.88910378805810e+01, 5.82947692660048e+01, 5.77082617471488e+01, 5.71351486987333e+01, 5.65787622881279e+01, 5.60420984643911e+01, 5.55277908803256e+01, 5.50380930480661e+01, 5.45748680412557e+01, 5.41395850950537e+01, 5.37333224832460e+01, 5.33567760593070e+01, 5.30102728439442e+01, 5.26937890414552e+01, 5.24069718842297e+01, 5.21491647433246e+01, 5.19194349983054e+01, 5.17166042203371e+01, 5.15392802757324e+01, 5.13858909915903e+01, 5.12547190336415e+01, 5.11439376267454e+01, 5.10516467030578e+01, 5.09759089978996e+01, 5.09147855377004e+01, 5.08663698885890e+01, 5.08288204692982e+01, 5.08003901886381e+01, 5.07794526549910e+01, 5.07645242299005e+01, 5.07542812636595e+01, 5.07475719581637e+01, 5.07434224479213e+01, 5.07410368673833e+01, 5.07397913723325e+01, 5.07392222937540e+01, 5.07390088124732e+01, 5.07389507403642e+01, 5.07389421690417e+01, 5.07389415079388e+01, 5.07389296366535e+01, 5.07388492069553e+01, 5.07385535389736e+01, 5.07377653974317e+01, 5.07360405727142e+01, 5.07327373414843e+01, 5.07269930398821e+01, 5.07177088428146e+01, 5.07035436803215e+01, 5.06829180310815e+01, 5.06540281064514e+01, 5.06148706683119e+01, 5.05632784044706e+01, 5.04969654144693e+01, 5.04135819405552e+01, 5.03107770254428e+01, 5.01862673113037e+01, 5.00379097427920e+01, 4.98637755374639e+01, 4.96622224801825e+01, 4.94319624242244e+01, 4.91721208757877e+01, 4.88822857254039e+01, 4.85625425801361e+01, 4.82134947377159e+01, 4.78362666013998e+01, 4.74324902141908e+01, 4.70042755231491e+01, 4.65541658801382e+01, 4.60850810471740e+01, 4.56002505150406e+01, 4.51031402092650e+01, 4.45973756478625e+01, 4.40866643877644e+01, 4.35747202467220e+01, 4.30651914112033e+01, 4.25615941999655e+01, 4.20672539574009e+01, 4.15852542683819e+01, 4.11183953720944e+01, 4.06691622785277e+01, 4.02397026644800e+01, 3.98318141851468e+01, 3.94469404382404e+01, 3.90861745123464e+01, 3.87502688730430e+01, 3.84396502967196e+01, 3.81544386368809e+01, 3.78944683689208e+01, 3.76593120682990e+01, 3.74483051969743e+01, 3.72605717742708e+01, 3.70950506711386e+01, 3.69505223805541e+01, 3.68256361790944e+01, 3.67189376088004e+01, 3.66288961814181e+01, 3.65539331482615e+01, 3.64924490986634e+01, 3.64428510590938e+01, 3.64035786742853e+01, 3.63731289714431e+01, 3.63500791482767e+01, 3.63331067931946e+01, 3.63210069476016e+01, 3.63127054592934e+01, 3.63072681529619e+01, 3.63039054561505e+01, 3.63019722609484e+01, 3.63009629650818e+01, 3.63005018108914e+01, 3.63003288162647e+01, 3.63002817575534e+01, 3.63002748117870e+01, 3.63002742858746e+01, 3.63002649693883e+01, 3.63002018488435e+01, 3.62999698113524e+01, 3.62993512869651e+01, 3.62979976735052e+01, 3.62954053879044e+01, 3.62908975257848e+01, 3.62836120095623e+01, 3.62724969862912e+01, 3.62563140938953e+01, 3.62336500404393e+01, 3.62029367273283e+01, 3.61624798869416e+01, 3.61104958949911e+01, 3.60451560604177e+01, 3.59646373010116e+01, 3.58671776998953e+01, 3.57511350339509e+01, 3.56150460051173e+01, 3.54576836290012e+01, 3.52781100832432e+01, 3.50757223273498e+01, 3.48502880034448e+01, 3.46019695255346e+01, 3.43313348550957e+01, 3.40393542111374e+01, 3.37273828170408e+01, 3.33971306666477e+01, 3.30506211065022e+01, 3.26901406861151e+01, 3.23181831435462e+01, 3.19373905193312e+01, 3.15504942205968e+01, 3.11602584301928e+01, 3.07694276566741e+01, 3.03806795602064e+01, 2.99965835792301e+01, 2.96195654118347e+01, 2.92518771202370e+01, 2.88955725222594e+01, 2.85524875630205e+01, 2.82242254519883e+01, 2.79121464345195e+01, 2.76173620944763e+01, 2.73407340403373e+01, 2.70828767279574e+01, 2.68441640547350e+01, 2.66247392609693e+01, 2.64245276217596e+01, 2.62432514157780e+01, 2.60804467073866e+01, 2.59354815569234e+01, 2.58075753587306e+01, 2.56958190791318e+01, 2.55991962150915e+01, 2.55166043137000e+01, 2.54468768836720e+01, 2.53888054974636e+01, 2.53411618333744e+01, 2.53027193491652e+01, 2.52722742205046e+01, 2.52486651268624e+01, 2.52307914314081e+01, 2.52176292859913e+01, 2.52082452017455e+01, 2.52018066626764e+01, 2.51975894239910e+01, 2.51949812267871e+01, 2.51934817717863e+01, 2.51926989209680e+01, 2.51923412299231e+01, 2.51922070476568e+01, 2.51921705468263e+01, 2.51921651593802e+01, 2.51921647549429e+01, 2.51921576363883e+01, 2.51921094071409e+01, 2.51919321114312e+01, 2.51914595069731e+01, 2.51904252277861e+01, 2.51884444751593e+01, 2.51849999790029e+01, 2.51794328897834e+01, 2.51709391654685e+01, 2.51585719047862e+01, 2.51412499427372e+01, 2.51177728615166e+01, 2.50868423756884e+01, 2.50470898229642e+01, 2.49971092335451e+01, 2.49354951689926e+01, 2.48608842290838e+01, 2.47719988408141e+01, 2.46676916913131e+01, 2.45469889726940e+01, 2.44091304992956e+01, 2.42536047615559e+01, 2.40801771153824e+01, 2.38889095814822e+01, 2.36801711425712e+01, 2.34546379573094e+01, 2.32132835173843e+01, 2.29573593967497e+01, 2.26883678025620e+01, 2.24080275569007e+01, 2.21182353566558e+01, 2.18210241559658e+01, 2.15185203233452e+01, 2.12129009217796e+01, 2.09063521418937e+01, 2.06010296678043e+01, 2.02990216113559e+01, 2.00023146011680e+01, 1.97127636094522e+01, 1.94320660825209e+01, 1.91617408656150e+01, 1.89031122632314e+01, 1.86572993657973e+01, 1.84252105350750e+01, 1.82075427128746e+01, 1.80047850337268e+01, 1.78172261027388e+01, 1.76449642511232e+01, 1.74879200976268e+01, 1.73458508096993e+01, 1.72183655552205e+01, 1.71049417453312e+01, 1.70049417752365e+01, 1.69176300605389e+01, 1.68421902338170e+01, 1.67777424060392e+01, 1.67233604098432e+01, 1.66780889294537e+01, 1.66409603900152e+01, 1.66110114338502e+01, 1.65872987599811e+01, 1.65689140538119e+01, 1.65549976934252e+01, 1.65447508938529e+01, 1.65374459458025e+01, 1.65324342236183e+01, 1.65291516795432e+01, 1.65271216061917e+01, 1.65259545330677e+01, 1.65253452207885e+01, 1.65250668221334e+01, 1.65249623853738e+01, 1.65249339760606e+01, 1.65249297829062e+01, 1.65249294690820e+01, 1.65249239581717e+01, 1.65248866209563e+01, 1.65247493655956e+01, 1.65243834946760e+01, 1.65235828026221e+01, 1.65220494103149e+01, 1.65193829182738e+01, 1.65150734001673e+01, 1.65084986886024e+01, 1.64989263217170e+01, 1.64855204175598e+01, 1.64673536189498e+01, 1.64434241007044e+01, 1.64126774521865e+01, 1.63740330425895e+01, 1.63264142498536e+01, 1.62687816966917e+01, 1.62001684033786e+01, 1.61197155548972e+01, 1.60267074102471e+01, 1.59206037751569e+01, 1.58010684352534e+01, 1.56679920199070e+01, 1.55215079459732e+01, 1.53620003755578e+01, 1.51901035029698e+01, 1.50066919428665e+01, 1.48128624937080e+01, 1.46099080588361e+01, 1.43992849771470e+01, 1.41825754011630e+01, 1.39614466225272e+01, 1.37376093556565e+01, 1.35127769388900e+01, 1.32886272084150e+01, 1.30667684722195e+01, 1.28487106025122e+01, 1.26358418257622e+01, 1.24294113685593e+01, 1.22305177547215e+01, 1.20401022704219e+01, 1.18589469301321e+01, 1.16876761836855e+01, 1.15267615905158e+01, 1.13765287324872e+01, 1.12371657217751e+01, 1.11087327668961e+01, 1.09911723736210e+01, 1.08843198674855e+01, 1.07879140239096e+01, 1.07016076763256e+01, 1.06249782399912e+01, 1.05575381384621e+01, 1.04987451510155e+01, 1.04480127133015e+01, 1.04047202013811e+01, 1.03682232128638e+01, 1.03378638305105e+01, 1.03129808164461e+01, 1.02929196426606e+01, 1.02770422198545e+01, 1.02647361462355e+01, 1.02554232648886e+01, 1.02485672967088e+01, 1.02436803087206e+01, 1.02403277869126e+01, 1.02381321091166e+01, 1.02367742560932e+01, 1.02359936555135e+01, 1.02355861203861e+01, 1.02353999162123e+01, 1.02353300649205e+01, 1.02353110637059e+01, 1.02353082591675e+01, 1.02353080491916e+01, 1.02353043608599e+01, 1.02352793718863e+01, 1.02351875100232e+01, 1.02349426419242e+01, 1.02344067637241e+01, 1.02333805298218e+01, 1.02315960161062e+01, 1.02287120780072e+01, 1.02243126212030e+01, 1.02179080502939e+01, 1.02089400940360e+01, 1.01967901202978e+01, 1.01807909460040e+01, 1.01602420150118e+01, 1.01344276610904e+01, 1.01026379987904e+01, 1.00641918012152e+01, 1.00184605441745e+01, 9.96489263811765e+00, 9.90303675182231e+00, 9.83256307388029e+00, 9.75328137515423e+00, 9.66515483696315e+00, 9.56830879652559e+00, 9.46303382385622e+00, 9.34978286372090e+00, 9.22916252525749e+00, 9.10191894874866e+00, 8.96891899182603e+00, 8.83112772831214e+00, 8.68958342193105e+00, 8.54537121401272e+00, 8.39959674896423e+00, 8.25336086249138e+00, 8.10773629094955e+00, 7.96374714580932e+00, 7.82235165718400e+00, 7.68442844687256e+00, 7.55076636473787e+00, 7.42205772909344e+00, 7.29889466350726e+00, 7.18176812428710e+00, 7.07106916397072e+00, 6.96709197034991e+00, 6.87003824848895e+00, 6.78002256410727e+00, 6.69707833031590e+00, 6.62116418743584e+00, 6.55217059109056e+00, 6.48992648270253e+00, 6.43420596642520e+00, 6.38473495614281e+00, 6.34119778499843e+00, 6.30324378799045e+00, 6.27049387586666e+00, 6.24254711651201e+00, 6.21898732926721e+00, 6.19938967952308e+00, 6.18332723732080e+00, 6.17037743677027e+00, 6.16012834543375e+00, 6.15218462715272e+00, 6.14617306087308e+00, 6.14174746434624e+00, 6.13859286717368e+00, 6.13642878382308e+00, 6.13501145439173e+00, 6.13413494848558e+00, 6.13363106411751e+00, 6.13336799672061e+00, 6.13324780034509e+00, 6.13320271074384e+00, 6.13319044529875e+00, 6.13318863494548e+00, 6.13318849904468e+00, 6.13318610707692e+00, 6.13316990115430e+00, 6.13311032663177e+00, 6.13295152408153e+00, 6.13260399494779e+00, 6.13193845946435e+00, 6.13078116397236e+00, 6.12891087161562e+00, 6.12605774288681e+00, 6.12190427851036e+00, 6.11608845389087e+00, 6.10820911887805e+00, 6.09783366642878e+00, 6.08450788760676e+00, 6.06776782890820e+00, 6.04715335433238e+00, 6.02222299504130e+00, 5.99256955286677e+00, 5.95783582166837e+00, 5.91772971522445e+00, 5.87203805420378e+00, 5.82063827780586e+00, 5.76350741362859e+00, 5.70072776218301e+00, 5.63248892353755e+00, 5.55908599976847e+00, 5.48091403013728e+00, 5.39845893553840e+00, 5.31228544444525e+00, 5.22302262730301e+00, 5.13134776859599e+00, 5.03796935084922e+00, 4.94360991434827e+00, 4.84898949743723e+00, 4.75481026542293e+00, 4.66174281347870e+00, 4.57041449256928e+00, 4.48139996828412e+00, 4.39521408988911e+00, 4.31230702833336e+00, 4.23306154288157e+00, 4.15779216002686e+00, 4.08674599695584e+00, 4.02010493478464e+00, 3.95798884207019e+00, 3.90045956337707e+00, 3.84752541668905e+00, 3.79914598251459e+00, 3.75523701200322e+00, 3.71567532703244e+00, 3.68030362846547e+00, 3.64893516683645e+00, 3.62135826062644e+00, 3.59734066986400e+00, 3.57663384653135e+00, 3.55897708831307e+00, 3.54410161924699e+00, 3.53173461094835e+00, 3.52160314279152e+00, 3.51343808057331e+00, 3.50697783280538e+00, 3.50197192404914e+00, 3.49818430773780e+00, 3.49539632864963e+00, 3.49340923915247e+00, 3.49204617455226e+00, 3.49115350172949e+00, 3.49060147141867e+00, 3.49028412699130e+00, 3.49011844987622e+00, 3.49004275183852e+00, 3.49001435508607e+00, 3.49000663049928e+00, 3.49000549036712e+00, 3.49000540441443e+00, 3.49000388671928e+00, 3.48999360412095e+00, 3.48995580434674e+00, 3.48985504513961e+00, 3.48963454172610e+00, 3.48921227294862e+00, 3.48847801171711e+00, 3.48729143217744e+00, 3.48548143009353e+00, 3.48284677276799e+00, 3.47915816828814e+00, 3.47416180876593e+00, 3.46758439666392e+00, 3.45913960622883e+00, 3.44853586379703e+00, 3.43548525337332e+00, 3.41971327151359e+00, 3.40096907427198e+00, 3.37903578657784e+00, 3.35374038962219e+00, 3.32496267332286e+00, 3.29264274603162e+00, 3.25678663706550e+00, 3.21746961035529e+00, 3.17483692606227e+00, 3.12910193344483e+00, 3.08054154068908e+00, 3.02948927134177e+00, 2.97632626703303e+00, 2.92147071810612e+00, 2.86536628635665e+00, 2.80847012057396e+00, 2.75124105449423e+00, 2.69412852191088e+00, 2.63756263334572e+00, 2.58194574428879e+00, 2.52764571940791e+00, 2.47499097281346e+00, 2.42426725211881e+00, 2.37571604149802e+00, 2.32953439074520e+00, 2.28587593479001e+00, 2.24485284980523e+00, 2.20653849454332e+00, 2.17097050430380e+00, 2.13815413499697e+00, 2.10806569138772e+00, 2.08065591263481e+00, 2.05585322637916e+00, 2.03356681744627e+00, 2.01368948709367e+00, 1.99610030273701e+00, 1.98066705584519e+00, 1.96724855723610e+00, 1.95569680464325e+00, 1.94585905768577e+00, 1.93757985094126e+00, 1.93070296751304e+00, 1.92507338424568e+00, 1.92053918663801e+00, 1.91695343769167e+00, 1.91417597162906e+00, 1.91207507182638e+00, 1.91052898354562e+00, 1.90942720703252e+00, 1.90867151590692e+00, 1.90817664976521e+00, 1.90787063839369e+00, 1.90769472738941e+00, 1.90760289038587e+00, 1.90756093030373e+00, 1.90754518979061e+00, 1.90754090800469e+00, 1.90754027602263e+00, 1.90754022811927e+00, 1.90753937883460e+00, 1.90753362481374e+00, 1.90751247254141e+00, 1.90745608924784e+00, 1.90733270034788e+00, 1.90709641331225e+00, 1.90668556283720e+00, 1.90602166599042e+00, 1.90500906840388e+00, 1.90353535237323e+00, 1.90147256338803e+00, 1.89867929154646e+00, 1.89500361710131e+00, 1.89028689422811e+00, 1.88436830412256e+00, 1.87709005911162e+00, 1.86830308644646e+00, 1.85787296813557e+00, 1.84568586708428e+00, 1.83165413611376e+00, 1.81572129119687e+00, 1.79786603844404e+00, 1.77810507888982e+00, 1.75649447596896e+00, 1.73312945436331e+00, 1.70814259900089e+00, 1.68170053008984e+00, 1.65399923340092e+00, 1.62525831382382e+00, 1.59571450538115e+00, 1.56561480620316e+00, 1.53520961005594e+00, 1.50474617846889e+00, 1.47446274443692e+00, 1.44458346771503e+00, 1.41531438174007e+00, 1.38684039197189e+00, 1.35932331247587e+00, 1.33290086740929e+00, 1.30768653991969e+00, 1.28377012372158e+00, 1.26121882128085e+00, 1.24007873471672e+00, 1.22037660807355e+00, 1.20212169913187e+00, 1.18530768222912e+00, 1.16991450794031e+00, 1.15591016883556e+00, 1.14325234145009e+00, 1.13188989221513e+00, 1.12176424902664e+00, 1.11281065033779e+00, 1.10495929034316e+00, 1.09813638229471e+00, 1.09226516263971e+00, 1.08726685690048e+00, 1.08306162444570e+00, 1.07956949395660e+00, 1.07671129492238e+00, 1.07440958338280e+00, 1.07258955289475e+00, 1.07117991487588e+00, 1.07011372662313e+00, 1.06932914092963e+00, 1.06877004876765e+00, 1.06838658626629e+00, 1.06813547932070e+00, 1.06798020354881e+00, 1.06789094368438e+00, 1.06784434439427e+00, 1.06782305334245e+00, 1.06781506642263e+00, 1.06781289379518e+00, 1.06781257312022e+00, 1.06781254865926e+00, 1.06781211295325e+00, 1.06780916098519e+00, 1.06779830930380e+00, 1.06776938320805e+00, 1.06770608170541e+00, 1.06758486152722e+00, 1.06737408919598e+00, 1.06703350714552e+00, 1.06651405617272e+00, 1.06575809138896e+00, 1.06470002149297e+00, 1.06326739069871e+00, 1.06138240823563e+00, 1.05896391152726e+00, 1.05592972601701e+00, 1.05219935793963e+00, 1.04769692776644e+00, 1.04235422405908e+00, 1.03611373321043e+00, 1.02893148355791e+00, 1.02077953601359e+00, 1.01164796033065e+00, 1.00154615777893e+00, 9.90503426902094e-01, 9.78568716758772e-01, 9.65809567328236e-01, 9.52310293927326e-01, 9.38169525316098e-01, 9.23497247813003e-01, 9.08411535669458e-01, 8.93035158650065e-01, 8.77492251074514e-01, 8.61905204582285e-01, 8.46391913444805e-01, 8.31063461222527e-01, 8.16022295985931e-01, 8.01360902634926e-01, 7.87160948407703e-01, 7.73492853439294e-01, 7.60415722829137e-01, 7.47977569585387e-01, 7.36215757716565e-01, 7.25157599906112e-01, 7.14821052823524e-01, 7.05215463551710e-01, 6.96342331502521e-01, 6.88196060570476e-01, 6.80764685501529e-01, 6.74030564181749e-01, 6.67971033662369e-01, 6.62559032258756e-01, 6.57763693104584e-01, 6.53550916262489e-01, 6.49883927043803e-01, 6.46723827729446e-01, 6.44030148558343e-01, 6.41761401806534e-01, 6.39875640177920e-01, 6.38331017738366e-01, 6.37086348448761e-01, 6.36101654209426e-01, 6.35338691458870e-01, 6.34761443019656e-01, 6.34336560281241e-01, 6.34033740149818e-01, 6.33826021603610e-01, 6.33689988221818e-01, 6.33605865656250e-01, 6.33557506545557e-01, 6.33532259604948e-01, 6.33520724268649e-01, 6.33516396998033e-01, 6.33515219878547e-01, 6.33515046138285e-01, 6.33515032791934e-01, 6.33514793836468e-01, 6.33513174879766e-01, 6.33507223450053e-01, 6.33491359325703e-01, 6.33456642103766e-01, 6.33390158402293e-01, 6.33274554988702e-01, 6.33087742153322e-01, 6.32802789253262e-01, 6.32388031240026e-01, 6.31807400774960e-01, 6.31020994816922e-01, 6.29985876989039e-01, 6.28657107393778e-01, 6.26988979911346e-01, 6.24936433787354e-01, 6.22456592304232e-01, 6.19510367740446e-01, 6.16064060187497e-01, 6.12090869793235e-01, 6.07572239227551e-01, 6.02498946833066e-01, 5.96871881578451e-01, 5.90702448252131e-01, 5.84012574052821e-01, 5.76834313731936e-01, 5.69209077034003e-01, 5.61186526505963e-01, 5.52823213232386e-01, 5.44181030829392e-01, 5.35325573192480e-01, 5.26324479195779e-01, 5.17245838862209e-01, 5.08156722204944e-01, 4.99121876000544e-01, 4.90202617183783e-01, 4.81455936008800e-01, 4.72933808802281e-01, 4.64682709713486e-01, 4.56743303530132e-01, 4.49150297193498e-01, 4.41932425677720e-01, 4.35112547874368e-01, 4.28707829518572e-01, 4.22729992543978e-01, 4.17185613166112e-01, 4.12076454145563e-01, 4.07399819826296e-01, 4.03148925497675e-01, 3.99313275275105e-01, 3.95879044951684e-01, 3.92829468083548e-01, 3.90145224891171e-01, 3.87804834351826e-01, 3.85785050119622e-01, 3.84061260649588e-01, 3.82607893173216e-01, 3.81398820044993e-01, 3.80407764555116e-01, 3.79608701711575e-01, 3.78976247877986e-01, 3.78486031673589e-01, 3.78115037357048e-01, 3.77841911173007e-01, 3.77647220957968e-01, 3.77513659753802e-01, 3.77426185290730e-01, 3.77372088938387e-01, 3.77340989995957e-01, 3.77324753861032e-01, 3.77317335508948e-01, 3.77314552640655e-01, 3.77313795634109e-01, 3.77313683901562e-01, 3.77313675252992e-01, 3.77313519554974e-01, 3.77312464678823e-01, 3.77308586854028e-01, 3.77298250085502e-01, 3.77275628770263e-01, 3.77232307968939e-01, 3.77156978124987e-01, 3.77035239546657e-01, 3.76849530208190e-01, 3.76579188236022e-01, 3.76200657633206e-01, 3.75687842154989e-01, 3.75012607609179e-01, 3.74145427125072e-01, 3.73056157161145e-01, 3.71714924423557e-01, 3.70093095873372e-01, 3.68164296236472e-01, 3.65905430684734e-01, 3.63297665502145e-01, 3.60327317413109e-01, 3.56986603494690e-01, 3.53274208561515e-01, 3.49195635534443e-01, 3.44763316048883e-01, 3.39996472468130e-01, 3.34920737272000e-01, 3.29567550091070e-01, 3.23973365121176e-01, 3.18178711211161e-01, 3.12227152878293e-01, 3.06164202619570e-01, 3.00036233325940e-01, 2.93889434904926e-01, 2.87768852159431e-01, 2.81717532467880e-01, 2.75775802740782e-01, 2.69980686252950e-01, 2.64365461838366e-01, 2.58959360992748e-01, 2.53787392911747e-01, 2.48870283535425e-01, 2.44224512255585e-01, 2.39862428947163e-01, 2.35792434168388e-01, 2.32019206484793e-01, 2.28543962651472e-01, 2.25364738613891e-01, 2.22476681748859e-01, 2.19872347255375e-01, 2.17541993937942e-01, 2.15473876651942e-01, 2.13654534300588e-01, 2.12069073431707e-01, 2.10701448160108e-01, 2.09534737340462e-01, 2.08551419668893e-01, 2.07733646752445e-01, 2.07063513239504e-01, 2.06523321956141e-01, 2.06095840769962e-01, 2.05764546734628e-01, 2.05513852084749e-01, 2.05329305971702e-01, 2.05197765546842e-01, 2.05107530173706e-01, 2.05048433204553e-01, 2.05011886868675e-01, 2.04990877324860e-01, 2.04979908731062e-01, 2.04974897154230e-01, 2.04973017149129e-01, 2.04972505743157e-01, 2.04972430260739e-01, 2.04972424370723e-01, 2.04972317722152e-01, 2.04971595163215e-01, 2.04968938970026e-01, 2.04961858610801e-01, 2.04946363806101e-01, 2.04916690889618e-01, 2.04865094095680e-01, 2.04781712462221e-01, 2.04654521909987e-01, 2.04469380632275e-01, 2.04210174054779e-01, 2.03859063148428e-01, 2.03396836666570e-01, 2.02803363861747e-01, 2.02058139436744e-01, 2.01140907053372e-01, 2.00032341952436e-01, 1.98714767569911e-01, 1.97172876022393e-01, 1.95394418601666e-01, 1.93370830568270e-01, 1.91097755067815e-01, 1.88575434211216e-01, 1.85808941283838e-01, 1.82808236397150e-01, 1.79588038070150e-01, 1.76167514382897e-01, 1.72569808460182e-01, 1.68821423066490e-01, 1.64951497078818e-01, 1.60991011861668e-01, 1.56971967760668e-01, 1.52926570084020e-01, 1.48886460371983e-01, 1.44882022984788e-01, 1.40941789714859e-01, 1.37091957003349e-01, 1.33356022218194e-01, 1.29754538077592e-01, 1.26304978211991e-01, 1.23021702285840e-01, 1.19916006038011e-01, 1.16996239933101e-01, 1.14267979731622e-01, 1.11734233055031e-01, 1.09395667725053e-01, 1.07250849963222e-01, 1.05296483080821e-01, 1.03527639788737e-01, 1.01937983563317e-01, 1.00519976562396e-01, 9.92650733562073e-02, 9.81639011601345e-02, 9.72064282615293e-02, 9.63821228873627e-02, 9.56801048871512e-02, 9.50892923613793e-02, 9.45985448318538e-02, 9.41968037951000e-02, 9.38732306031511e-02, 9.36173406500416e-02, 9.34191318944073e-02, 9.32692049102712e-02, 9.31588710138556e-02, 9.30802446346064e-02, 9.30263160359885e-02, 9.29910007625097e-02, 9.29691627854083e-02, 9.29566091979410e-02, 9.29500554074849e-02, 9.29470609981138e-02, 9.29459377033481e-02, 9.29456321410385e-02, 9.29455870407025e-02, 9.29455834915416e-02, 9.29455188438055e-02, 9.29450808466470e-02, 9.29434707328625e-02, 9.29391788380656e-02, 9.29297865244649e-02, 9.29118006265526e-02, 9.28805277126656e-02, 9.28299950241467e-02, 9.27529244201111e-02, 9.26407649169505e-02, 9.24837883440262e-02, 9.22712511071495e-02, 9.19916229444908e-02, 9.16328808038998e-02, 9.11828625768436e-02, 9.06296714990305e-02, 8.99621178002405e-02, 8.91701800023554e-02, 8.82454645695532e-02, 8.71816399074277e-02, 8.59748194790484e-02, 8.46238694500130e-02, 8.31306190205225e-02, 8.14999564430821e-02, 7.97398003894898e-02, 7.78609443026426e-02, 7.58767799244708e-02, 7.38029144887093e-02, 7.16567032421551e-02, 6.94567242476438e-02, 6.72222252855192e-02, 6.49725729106052e-02, 6.27267314933988e-02, 6.05027958465598e-02, 5.83175954203057e-02, 5.61863816841992e-02, 5.41226037850180e-02, 5.21377714396607e-02, 5.02413988228625e-02, 4.84410193622153e-02, 4.67422590396325e-02, 4.51489548938326e-02, 4.36633055768993e-02, 4.22860416897777e-02, 4.10166049805269e-02, 3.98533272471155e-02, 3.87936018509123e-02, 3.78340429315074e-02, 3.69706294654151e-02, 3.61988330199523e-02, 3.55137293426253e-02, 3.49100948475893e-02, 3.43824897067151e-02, 3.39253296920674e-02, 3.35329491573333e-02, 3.31996575728226e-02, 3.29197918371165e-02, 3.26877662032814e-02, 3.24981211267215e-02, 3.23455717272054e-02, 3.22250559151058e-02, 3.21317816117244e-02, 3.20612719408757e-02, 3.20094068259351e-02, 3.19724591252992e-02, 3.19471233113404e-02, 3.19305347539114e-02, 3.19202779085970e-02, 3.19143821135555e-02, 3.19113042303343e-02, 3.19098979778403e-02, 3.19093704533344e-02, 3.19092269547525e-02, 3.19092057747099e-02, 3.19092040931821e-02, 3.19091732768157e-02, 3.19089644919188e-02, 3.19081969843893e-02, 3.19061511511478e-02, 3.19016741997302e-02, 3.18931014693646e-02, 3.18781970714070e-02, 3.18541174049329e-02, 3.18174009570242e-02, 3.17639875335516e-02, 3.16892696023388e-02, 3.15881776640907e-02, 3.14553004227207e-02, 3.12850389475869e-02, 3.10717920142451e-02, 3.08101674458925e-02, 3.04952117105017e-02, 3.01226474870191e-02, 2.96891066902134e-02, 2.91923448684202e-02, 2.86314222919373e-02, 2.80068377258402e-02, 2.73206030061033e-02, 2.65762501065012e-02, 2.57787671132377e-02, 2.49344648655665e-02, 2.40507812683148e-02, 2.31360348043523e-02, 2.21991422194918e-02, 2.12493176330923e-02, 2.02957714050171e-02, 1.93474267023674e-02, 1.84126694633306e-02, 1.74991433121221e-02, 1.66135957060902e-02, 1.57617766278264e-02, 1.49483876449779e-02, 1.41770772021061e-02, 1.34504765709238e-02, 1.27702690568299e-02, 1.21372831038542e-02, 1.15515991330852e-02, 1.10126611979240e-02, 1.05193872763891e-02, 1.00702745787231e-02, 9.66349738570216e-03, 9.29699490018265e-03, 8.96854667043101e-03, 8.67583425176122e-03, 8.41648962652641e-03, 8.18813243424188e-03, 7.98839850293120e-03, 7.81496165124906e-03, 7.66554997986191e-03, 7.53795751316488e-03, 7.43005217967152e-03, 7.33978141249736e-03, 7.26517675265370e-03, 7.20435861853730e-03, 7.15554196622747e-03, 7.11704308548633e-03, 7.08728738327125e-03, 7.06481773738687e-03, 7.04830283301305e-03, 7.03654479424642e-03, 7.02848537208922e-03, 7.02320995070238e-03, 7.01994869388045e-03, 7.01807427361700e-03, 7.01709579617216e-03, 7.01664875344932e-03, 7.01648105741981e-03, 7.01643544051933e-03, 7.01642870757269e-03, 7.01642816810022e-03, 7.01641821915189e-03, 7.01635081385632e-03, 7.01610302915867e-03, 7.01544255851161e-03, 7.01399729921093e-03, 7.01123008515481e-03, 7.00641984860795e-03, 6.99865051791448e-03, 6.98680896940118e-03, 6.96959330993252e-03, 6.94553262883346e-03, 6.91301906989850e-03, 6.87035260135360e-03, 6.81579818061095e-03, 6.74765414264300e-03, 6.66432963357977e-03, 6.56442784756450e-03, 6.44683079079300e-03, 6.31078037034172e-03, 6.15594987679913e-03, 5.98249953969798e-03, 5.79111004239508e-03, 5.58298901664500e-03, 5.35984781025434e-03, 5.12384902314977e-03, 4.87752864296453e-03, 4.62369897908843e-03, 4.36533930141536e-03, 4.10548053660875e-03, 3.84708998380745e-03, 3.59296294151079e-03, 3.34562964332008e-03, 3.10728543144023e-03, 2.87974804999180e-03, 2.66444000054378e-03, 2.46238974970936e-03, 2.27424610983462e-03, 2.10030517759208e-03, 1.94055297981777e-03, 1.79472210665210e-03, 1.66235130875236e-03, 1.54283626201845e-03, 1.43546767394298e-03, 1.33945913729860e-03, 1.25397180458720e-03, 1.17814653955861e-03, 1.11114420455749e-03, 1.05217672528113e-03, 1.00051357387088e-03, 9.55471685742282e-04, 9.16408225350551e-04, 8.82720288274186e-04, 8.53843133043597e-04, 8.29247024967117e-04, 8.08439767197910e-04, 7.90973327838174e-04, 7.76445647758806e-04, 7.64494741630983e-04, 7.54790061705535e-04, 7.47026282070351e-04, 7.40920675596612e-04, 7.36212674880221e-04, 7.32664163720642e-04, 7.30059876147493e-04, 7.28207763330602e-04, 7.26939268673732e-04, 7.26109414415532e-04, 7.25596584786103e-04, 7.25301898626185e-04, 7.25148086630156e-04, 7.25077817975261e-04, 7.25051459284032e-04, 7.25044289239918e-04, 7.25043230943356e-04, 7.25043145394011e-04, 7.25041558439443e-04, 7.25030806592031e-04, 7.24991282609722e-04, 7.24885933266251e-04, 7.24655415083930e-04, 7.24214084158348e-04, 7.23447042304315e-04, 7.22208476432229e-04, 7.20321550005565e-04, 7.17580128863915e-04, 7.13752647357069e-04, 7.08588428594226e-04, 7.01826722092846e-04, 6.93208523273370e-04, 6.82490820608450e-04, 6.69462423360301e-04, 6.53960455323954e-04, 6.35887297363524e-04, 6.15228208872054e-04, 5.92065846120704e-04, 5.66570156465218e-04, 5.38938666409194e-04, 5.09367946349926e-04, 4.78147142473517e-04, 4.45821805939772e-04, 4.13144066712334e-04, 3.80400994004049e-04, 3.47856020141535e-04, 3.16027222893922e-04, 2.86034994829265e-04, 2.57404427486563e-04, 2.30027008981892e-04, 2.04833553965500e-04, 1.82104071711819e-04, 1.60819991066319e-04, 1.43081612563378e-04, 1.27368706557718e-04, 1.13055853262267e-04, 9.98856472165244e-05, 8.77573925066599e-05, 7.65837149572235e-05, 6.61448640499884e-05, 5.60960550330947e-05, 4.61412655283766e-05, 3.62396456523584e-05, 2.66631043928101e-05, 1.78655509586179e-05, 1.03402752102966e-05, 4.54921278958409e-06, 2.94681547075015e-06, 2.57767179334678e-06, 1.86870708773768e-06, 2.28043276706558e-06, 2.53862052090881e-06, 2.41433463088470e-06, 2.10688119750557e-06, 1.78774756420808e-06, 1.54726757273238e-06, 1.39421607059376e-06, 1.29879316546067e-06, 1.22043499573010e-06, 1.11438274075226e-06, 9.69321050092731e-07, 7.99792118755249e-07, 6.25859243541527e-07, 4.64114734199939e-07, 3.25106896070793e-07, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 2.84217094304040e-14, 1.42108547152020e-14, 2.84217094304040e-14],
[ 1.09000000000000e+02, 1.08999999858897e+02, 1.08999995495697e+02, 1.08999965934329e+02, 1.08999857262824e+02, 1.08999567578434e+02, 1.08998933580504e+02, 1.08997719280896e+02, 1.08995607235376e+02, 1.08992192619866e+02, 1.08986980385976e+02, 1.08979385635177e+02, 1.08968737255399e+02, 1.08954284773369e+02, 1.08935208296183e+02, 1.08910631350801e+02, 1.08879636382634e+02, 1.08841282643088e+02, 1.08794626175619e+02, 1.08738741591286e+02, 1.08672745295590e+02, 1.08595819775925e+02, 1.08507238473529e+02, 1.08406390643117e+02, 1.08292805455786e+02, 1.08166174446659e+02, 1.08026371279033e+02, 1.07873467727667e+02, 1.07707744809003e+02, 1.07529698127840e+02, 1.07340036771616e+02, 1.07139675446234e+02, 1.06929719971351e+02, 1.06711446685125e+02, 1.06486276691739e+02, 1.06255746171743e+02, 1.06021474134598e+02, 1.05785129016933e+02, 1.05548395432829e+02, 1.05312942194526e+02, 1.05080392482630e+02, 1.04852296793333e+02, 1.04630109058633e+02, 1.04415166145013e+02, 1.04208670795658e+02, 1.04011677989392e+02, 1.03825084637712e+02, 1.03649622517341e+02, 1.03485854327416e+02, 1.03334172757302e+02, 1.03194802445838e+02, 1.03067804701374e+02, 1.02953084832726e+02, 1.02850401914700e+02, 1.02759380779415e+02, 1.02679525988477e+02, 1.02610237503120e+02, 1.02550827731843e+02, 1.02500539599783e+02, 1.02458565252884e+02, 1.02424064984857e+02, 1.02396185957655e+02, 1.02374080278568e+02, 1.02356922000664e+02, 1.02343922629487e+02, 1.02334344748411e+02, 1.02327513418287e+02, 1.02322825063324e+02, 1.02319753623498e+02, 1.02317853831998e+02, 1.02316761561861e+02, 1.02316191275588e+02, 1.02315930701867e+02, 1.02315832950825e+02, 1.02315806360092e+02, 1.02315802435353e+02, 1.02315801920510e+02, 1.02315789925286e+02, 1.02315708655865e+02, 1.02315409900719e+02, 1.02314613527983e+02, 1.02312870679599e+02, 1.02309532892949e+02, 1.02303728380350e+02, 1.02294346546813e+02, 1.02280031652188e+02, 1.02259186321218e+02, 1.02229985368117e+02, 1.02190400124665e+02, 1.02138233137551e+02, 1.02071162730414e+02, 1.01986796513636e+02, 1.01882732483426e+02, 1.01756625904708e+02, 1.01606259754180e+02, 1.01429616155025e+02, 1.01224946012695e+02, 1.00990834007662e+02, 1.00726256243841e+02, 1.00430628185148e+02, 1.00103840986658e+02, 9.97462848492847e+01, 9.93588584913964e+01, 9.89429641693619e+01, 9.85004879238445e+01, 9.80337650434750e+01, 9.75455313686451e+01, 9.70388621766576e+01, 9.65171018802983e+01, 9.59837891298366e+01, 9.54425824190404e+01, 9.48971904914162e+01, 9.43513099194076e+01, 9.38085700225927e+01, 9.32724837601720e+01, 9.27464028664278e+01, 9.22334761144265e+01, 9.17366106100019e+01, 9.12584368707175e+01, 9.08012788440875e+01, 9.03671299649310e+01, 8.99576360052601e+01, 8.95740850225308e+01, 8.92174043049407e+01, 8.88881639166991e+01, 8.85865862748510e+01, 8.83125611208963e+01, 8.80656652503611e+01, 8.78451863973895e+01, 8.76501507124794e+01, 8.74793533021072e+01, 8.73313913100786e+01, 8.72046990095066e+01, 8.70975843434443e+01, 8.70082663065321e+01, 8.69349125064494e+01, 8.68756761904136e+01, 8.68287319766251e+01, 8.67923095013579e+01, 8.67647241864553e+01, 8.67444043550324e+01, 8.67299139790291e+01, 8.67199704324152e+01, 8.67134567472356e+01, 8.67094280226295e+01, 8.67071118133881e+01, 8.67059025165470e+01, 8.67053499726948e+01, 8.67051426934270e+01, 8.67050863083466e+01, 8.67050779860235e+01, 8.67050772513207e+01, 8.67050628551828e+01, 8.67049653192213e+01, 8.67046067665751e+01, 8.67036509958341e+01, 8.67015593097235e+01, 8.66975534605800e+01, 8.66905871926754e+01, 8.66793276834950e+01, 8.66621479795426e+01, 8.66371312822235e+01, 8.66020876588869e+01, 8.65545834258980e+01, 8.64919830691862e+01, 8.64115031318577e+01, 8.63102770135329e+01, 8.61854291059236e+01, 8.60341561572252e+01, 8.58538132470509e+01, 8.56420013047018e+01, 8.53966527623536e+01, 8.51161117478060e+01, 8.47992052307488e+01, 8.44453017733749e+01, 8.40543550144358e+01, 8.36269297248025e+01, 8.31642091708141e+01, 8.26679835342263e+01, 8.21406201603636e+01, 8.15850173209834e+01, 8.10045438788979e+01, 8.04029676635773e+01, 7.97843755130716e+01, 7.91530878788205e+01, 7.85135707392973e+01, 7.78703474295515e+01, 7.72279129078113e+01, 7.65906529019342e+01, 7.59627702009785e+01, 7.53482199800341e+01, 7.47506554467848e+01, 7.41733843594451e+01, 7.36193362429077e+01, 7.30910395749047e+01, 7.25906079082060e+01, 7.21197338281611e+01, 7.16796897418314e+01, 7.12713346623370e+01, 7.08951263172119e+01, 7.05511380354155e+01, 7.02390799473829e+01, 6.99583240756356e+01, 6.97079329129186e+01, 6.94866910896479e+01, 6.92931397260875e+01, 6.91256130464944e+01, 6.89822768004656e+01, 6.88611679897721e+01, 6.87602353383157e+01, 6.86773798725688e+01, 6.86104949066232e+01, 6.85575046583767e+01, 6.85164006709490e+01, 6.84852751855447e+01, 6.84623506168732e+01, 6.84460043259575e+01, 6.84347879708645e+01, 6.84274408433452e+01, 6.84228967648581e+01, 6.84202843119687e+01, 6.84189203590102e+01, 6.84182971537303e+01, 6.84180633672822e+01, 6.84179997716401e+01, 6.84179903850544e+01, 6.84179896321341e+01, 6.84179757369361e+01, 6.84178815949242e+01, 6.84175355189165e+01, 6.84166130072394e+01, 6.84145941124785e+01, 6.84107276854962e+01, 6.84040039302923e+01, 6.83931365375508e+01, 6.83765554696830e+01, 6.83524112423271e+01, 6.83185912789455e+01, 6.82727485983570e+01, 6.82123427237168e+01, 6.81346922740350e+01, 6.80370382212456e+01, 6.79166162815387e+01, 6.77707363836338e+01, 6.75968666530640e+01, 6.73927189119867e+01, 6.71563323636079e+01, 6.68861519518586e+01, 6.65810978948602e+01, 6.62406231050861e+01, 6.58647556322022e+01, 6.54541238800782e+01, 6.50099631252651e+01, 6.45341027567565e+01, 6.40289346151581e+01, 6.34973637767637e+01, 6.29427440408816e+01, 6.23688011653778e+01, 6.17795474790301e+01, 6.11791918088615e+01, 6.05720486487642e+01, 5.99624501579096e+01, 5.93546639634450e+01, 5.87528189516655e+01, 5.81608403940163e+01, 5.75823949945338e+01, 5.70208458546372e+01, 5.64792169667738e+01, 5.59601666532469e+01, 5.54659693060055e+01, 5.49985047906006e+01, 5.45592548980747e+01, 5.41493062328519e+01, 5.37693589095778e+01, 5.34197404113268e+01, 5.31004239540804e+01, 5.28110507203995e+01, 5.25509553705902e+01, 5.23191943043022e+01, 5.21145762155211e+01, 5.19356945445414e+01, 5.17809614698094e+01, 5.16486430934724e+01, 5.15368954553365e+01, 5.14438009637408e+01, 5.13674047653470e+01, 5.13057504983482e+01, 5.12569147959379e+01, 5.12190398403091e+01, 5.11903632226997e+01, 5.11692443512781e+01, 5.11541866728506e+01, 5.11438550402819e+01, 5.11370876654737e+01, 5.11329022444348e+01, 5.11304960197161e+01, 5.11292397468214e+01, 5.11286657438208e+01, 5.11284504152295e+01, 5.11283918406082e+01, 5.11283831951159e+01, 5.11283825282924e+01, 5.11283705542820e+01, 5.11282894286070e+01, 5.11279912021497e+01, 5.11271962407220e+01, 5.11254564912094e+01, 5.11221246785838e+01, 5.11163306779899e+01, 5.11069661656719e+01, 5.10926784888829e+01, 5.10718745026144e+01, 5.10427348924997e+01, 5.10032392307489e+01, 5.09512016893669e+01, 5.08843169601354e+01, 5.08002155081038e+01, 5.06965268268633e+01, 5.05709488909004e+01, 5.04213215428680e+01, 5.02457011488370e+01, 5.00424335438179e+01, 4.98102221141881e+01, 4.95481878591749e+01, 4.92559184660926e+01, 4.89335038347750e+01, 4.85815560882288e+01, 4.82012128801149e+01, 4.77941237030877e+01, 4.73624198402227e+01, 4.69086694920412e+01, 4.64358203562748e+01, 4.59471324543607e+01, 4.54461042445804e+01, 4.49363950486056e+01, 4.44217466103217e+01, 4.39059062955986e+01, 4.33925541128412e+01, 4.28852354306699e+01, 4.23873009861978e+01, 4.19018554777724e+01, 4.14317156800848e+01, 4.09793785923524e+01, 4.05469996531023e+01, 4.01363805783221e+01, 3.97489659645377e+01, 3.93858474963672e+01, 3.90477744376581e+01, 3.87351690672193e+01, 3.84481458217707e+01, 3.81865330930245e+01, 3.79498968514241e+01, 3.77375654984204e+01, 3.75486555534204e+01, 3.73820979422981e+01, 3.72366647629917e+01, 3.71109964594064e+01, 3.70036293422125e+01, 3.69130233621412e+01, 3.68375899777832e+01, 3.67757198762260e+01, 3.67258102118546e+01, 3.66862909368519e+01, 3.66556497164532e+01, 3.66324548620328e+01, 3.66153756834580e+01, 3.66031996647181e+01, 3.65948459069950e+01, 3.65893743615161e+01, 3.65859904880430e+01, 3.65840451179791e+01, 3.65830294656257e+01, 3.65825654070784e+01, 3.65823913229229e+01, 3.65823439678331e+01, 3.65823369783218e+01, 3.65823364490971e+01, 3.65823270739350e+01, 3.65822635558527e+01, 3.65820300569702e+01, 3.65814076370382e+01, 3.65800454982129e+01, 3.65774368852838e+01, 3.65729006290650e+01, 3.65655692184183e+01, 3.65543841663938e+01, 3.65380992922928e+01, 3.65152923660954e+01, 3.64843853462244e+01, 3.64436731791179e+01, 3.63913608162888e+01, 3.63256077440944e+01, 3.62445789235720e+01, 3.61465006215075e+01, 3.60297192073205e+01, 3.58927606289533e+01, 3.57343880053332e+01, 3.55536546246444e+01, 3.53499496538741e+01, 3.51230340730170e+01, 3.48730647580089e+01, 3.46006052397603e+01, 3.43066224278729e+01, 3.39924694470058e+01, 3.36598556103155e+01, 3.33108053538480e+01, 3.29476085843215e+01, 3.25727652731222e+01, 3.21889272173192e+01, 3.17988396852682e+01, 3.14052852200794e+01, 3.10110312801960e+01, 3.06187827660798e+01, 3.02311399252366e+01, 2.98505617244400e+01, 2.94793345601652e+01, 2.91195461260569e+01, 2.87730643093418e+01, 2.84415210687722e+01, 2.81263012869805e+01, 2.78285365528979e+01, 2.75491037166951e+01, 2.72886279036907e+01, 2.70474895220133e+01, 2.68258346931364e+01, 2.66235884971815e+01, 2.64404704557330e+01, 2.62760117556505e+01, 2.61295738215787e+01, 2.60003679469416e+01, 2.58874757742991e+01, 2.57898704658268e+01, 2.57064384209506e+01, 2.56360013840139e+01, 2.55773387467737e+01, 2.55292097965322e+01, 2.54903755992782e+01, 2.54596201466107e+01, 2.54357703431714e+01, 2.54177143747269e+01, 2.54044179816267e+01, 2.53949381723280e+01, 2.53884339493078e+01, 2.53841736851018e+01, 2.53815388772674e+01, 2.53800241234330e+01, 2.53792332851391e+01, 2.53788719445264e+01, 2.53787363931773e+01, 2.53786995199227e+01, 2.53786940775076e+01, 2.53786936689438e+01, 2.53786864777574e+01, 2.53786377564186e+01, 2.53784586517276e+01, 2.53779812251799e+01, 2.53769363929187e+01, 2.53749354297035e+01, 2.53714557865697e+01, 2.53658318888771e+01, 2.53572514845874e+01, 2.53447579982008e+01, 2.53272592088664e+01, 2.53035424065266e+01, 2.52722959833640e+01, 2.52321371876280e+01, 2.51816455053946e+01, 2.51194008504121e+01, 2.50440254462152e+01, 2.49542279974019e+01, 2.48488484926998e+01, 2.47269017889526e+01, 2.45876180209983e+01, 2.44304778936964e+01, 2.42552410588102e+01, 2.40619660700224e+01, 2.38510208376883e+01, 2.36230830451075e+01, 2.33791305925215e+01, 2.31204227365473e+01, 2.28484731135964e+01, 2.25650162043238e+01, 2.22719689674630e+01, 2.19713893446754e+01, 2.16654331603334e+01, 2.13563106918675e+01, 2.10462439521993e+01, 2.07374255640556e+01, 2.04319800299633e+01, 2.01319281820119e+01, 1.98391555791199e+01, 1.95553855556833e+01, 1.92821574853287e+01, 1.90208106080200e+01, 1.87724735042307e+01, 1.85380590260185e+01, 1.83182642516279e+01, 1.81135748479145e+01, 1.79242731189602e+01, 1.77504489904117e+01, 1.75920132163661e+01, 1.74487121810434e+01, 1.73201437805441e+01, 1.72057739913176e+01, 1.71049538455422e+01, 1.70169366278766e+01, 1.69408951759648e+01, 1.68759392058165e+01, 1.68211325933532e+01, 1.67755105283528e+01, 1.67380964221010e+01, 1.67079184020033e+01, 1.66840251727308e+01, 1.66655009718677e+01, 1.66514793058145e+01, 1.66411551252234e+01, 1.66337950933967e+01, 1.66287456188211e+01, 1.66254383651581e+01, 1.66233930171013e+01, 1.66222171650286e+01, 1.66216032700692e+01, 1.66213227777146e+01, 1.66212175555627e+01, 1.66211889326064e+01, 1.66211847079187e+01, 1.66211843917345e+01, 1.66211788393813e+01, 1.66211412213853e+01, 1.66210029338617e+01, 1.66206343117024e+01, 1.66198275992745e+01, 1.66182826797084e+01, 1.66155961494196e+01, 1.66112542649830e+01, 1.66046302196899e+01, 1.65949861235080e+01, 1.65814799580818e+01, 1.65631776529831e+01, 1.65390702767439e+01, 1.65080961544815e+01, 1.64691675146979e+01, 1.64212010367680e+01, 1.63631514281074e+01, 1.62940469209359e+01, 1.62130253617617e+01, 1.61193693934862e+01, 1.60125391222308e+01, 1.58922006386564e+01, 1.57582488423513e+01, 1.56108232065920e+01, 1.54503154191812e+01, 1.52773682322041e+01, 1.50928653271118e+01, 1.48979125182994e+01, 1.46938111361819e+01, 1.44820249024600e+01, 1.42641419890789e+01, 1.40418341986708e+01, 1.38168152922236e+01, 1.35908004123905e+01, 1.33654683222542e+01, 1.31424278325567e+01, 1.29231893718388e+01, 1.27091422152173e+01, 1.25015374767072e+01, 1.23014766243149e+01, 1.21099050189132e+01, 1.19276098139092e+01, 1.17552214766254e+01, 1.15932181889601e+01, 1.14419324350629e+01, 1.13015591681337e+01, 1.11721650503216e+01, 1.10536983661480e+01, 1.09459993118852e+01, 1.08488104550745e+01, 1.07617872364528e+01, 1.06845084490858e+01, 1.06164866756126e+01, 1.05571786939680e+01, 1.05059958751038e+01, 1.04623145939914e+01, 1.04254866590760e+01, 1.03948497375896e+01, 1.03697377176328e+01, 1.03494909062480e+01, 1.03334659198921e+01, 1.03210450840864e+01, 1.03116451268565e+01, 1.03047249297636e+01, 1.02997920939642e+01, 1.02964080888229e+01, 1.02941917777423e+01, 1.02928211591945e+01, 1.02920332180958e+01, 1.02916218500889e+01, 1.02914338945341e+01, 1.02913633862216e+01, 1.02913442062798e+01, 1.02913413753616e+01, 1.02913411634106e+01, 1.02913374403860e+01, 1.02913122163621e+01, 1.02912194904196e+01, 1.02909723189266e+01, 1.02904313994209e+01, 1.02893955093446e+01, 1.02875941988581e+01, 1.02846831003530e+01, 1.02802421738738e+01, 1.02737771542649e+01, 1.02647243982828e+01, 1.02524592437428e+01, 1.02363078842962e+01, 1.02155626307723e+01, 1.01895002741831e+01, 1.01574030913066e+01, 1.01185818504568e+01, 1.00723999961418e+01, 1.00182980338957e+01, 9.95581701967002e+00, 9.88462000037602e+00, 9.80451026882727e+00, 9.71544539661285e+00, 9.61754619298414e+00, 9.51109999748692e+00, 9.39655803004230e+00, 9.27452686801227e+00, 9.14575446464077e+00, 9.01111143606165e+00, 8.87156859761853e+00, 8.72817190450304e+00, 8.58201603567940e+00, 8.43421785213603e+00, 8.28589086770928e+00, 8.13812170715497e+00, 7.99194931116313e+00, 7.84834740424944e+00, 7.70821049262362e+00, 7.57234342752869e+00, 7.44145437348388e+00, 7.31615087253832e+00, 7.19693859997224e+00, 7.08422236177501e+00, 6.97830888212306e+00, 6.87941095929574e+00, 6.78765261961003e+00, 6.70307496079704e+00, 6.62564244075888e+00, 6.55524942924862e+00, 6.49172689522798e+00, 6.43484914953372e+00, 6.38434060003404e+00, 6.33988250423758e+00, 6.30111972218666e+00, 6.26766748053484e+00, 6.23911815737245e+00, 6.21504808739593e+00, 6.19502436964896e+00, 6.17861163702095e+00, 6.16537872017399e+00, 6.15490511116414e+00, 6.14678710653587e+00, 6.14064348891058e+00, 6.13612059263134e+00, 6.13289659493682e+00, 6.13068488074468e+00, 6.12923634686121e+00, 6.12834053973477e+00, 6.12782555821722e+00, 6.12755669685754e+00, 6.12743385311538e+00, 6.12738777038622e+00, 6.12737523478539e+00, 6.12737338455766e+00, 6.12737324566355e+00, 6.12737080101080e+00, 6.12735423813930e+00, 6.12729335143018e+00, 6.12713105103535e+00, 6.12677586676735e+00, 6.12609567006966e+00, 6.12491287636276e+00, 6.12300136610305e+00, 6.12008533448332e+00, 6.11584024472199e+00, 6.10989601557179e+00, 6.10184251710183e+00, 6.09123737753252e+00, 6.07761601668852e+00, 6.06050371916727e+00, 6.03942944582393e+00, 6.01394096176074e+00, 5.98362074165757e+00, 5.94810201028567e+00, 5.90708419993442e+00, 5.86034706940785e+00, 5.80776274111493e+00, 5.74930497927909e+00, 5.68505515341846e+00, 5.61520450046690e+00, 5.54005250347090e+00, 5.46000142713270e+00, 5.37554727051655e+00, 5.28726759529620e+00, 5.19580684739930e+00, 5.10185989936041e+00, 5.00615459483892e+00, 4.90943407628899e+00, 4.81243962730198e+00, 4.71589467166001e+00, 4.62049045243883e+00, 4.52687377790332e+00, 4.43563707742591e+00, 4.34731087027741e+00, 4.26235862158921e+00, 4.18117385005961e+00, 4.10407926602856e+00, 4.03132765904205e+00, 3.96310422131928e+00, 3.89952998581502e+00, 3.84066607129677e+00, 3.78651845731795e+00, 3.73704305390601e+00, 3.69215087895721e+00, 3.65171320597537e+00, 3.61556659191744e+00, 3.58351773643551e+00, 3.55534815758612e+00, 3.53081869384572e+00, 3.50967385752998e+00, 3.49164607062984e+00, 3.47645981138631e+00, 3.46383568983647e+00, 3.45349445469216e+00, 3.44516091419401e+00, 3.43856773217917e+00, 3.43345903978177e+00, 3.42959378517189e+00, 3.42674873054698e+00, 3.42472099884114e+00, 3.42333007338488e+00, 3.42241916243724e+00, 3.42185585680686e+00, 3.42153203165772e+00, 3.42136297144754e+00, 3.42128572774297e+00, 3.42125675117323e+00, 3.42124886886359e+00, 3.42124770545192e+00, 3.42124761774420e+00, 3.42124606906037e+00, 3.42123557650991e+00, 3.42119700493959e+00, 3.42109418848565e+00, 3.42086918323474e+00, 3.42043829442123e+00, 3.41968904763549e+00, 3.41847826024149e+00, 3.41663135324679e+00, 3.41394302352509e+00, 3.41017936936611e+00, 3.40508152649510e+00, 3.39837082499973e+00, 3.38975541891698e+00, 3.37893826987140e+00, 3.36562628624750e+00, 3.34954033414974e+00, 3.33042575216861e+00, 3.30806292678288e+00, 3.28227742814522e+00, 3.25294917604727e+00, 3.22002011071179e+00, 3.18349988771361e+00, 3.14346920192932e+00, 3.10008046851270e+00, 3.05355574130699e+00, 3.00418191853169e+00, 2.95230345685933e+00, 2.89831297196469e+00, 2.84264023119523e+00, 2.78574013020081e+00, 2.72808028289312e+00, 2.67012884130395e+00, 2.61234310275498e+00, 2.55515936507607e+00, 2.49898436864387e+00, 2.44418853052679e+00, 2.39110104441627e+00, 2.34000680171430e+00, 2.29114499244805e+00, 2.24470917431601e+00, 2.20084855535825e+00, 2.15967021876830e+00, 2.12124202339280e+00, 2.08559593546018e+00, 2.05273158063444e+00, 2.02261984548171e+00, 1.99520639945855e+00, 1.97041504911218e+00, 1.94815087283115e+00, 1.92830311560155e+00, 1.91074784795786e+00, 1.89535041138843e+00, 1.88196768399425e+00, 1.87045020561932e+00, 1.86064420157616e+00, 1.85239353922192e+00, 1.84554164286313e+00, 1.83993338074792e+00, 1.83541692432060e+00, 1.83184556563346e+00, 1.82907946505522e+00, 1.82698728939650e+00, 1.82544769141688e+00, 1.82435057631021e+00, 1.82359809983353e+00, 1.82310534651508e+00, 1.82280064470552e+00, 1.82262548756013e+00, 1.82253404442626e+00, 1.82249226437202e+00, 1.82247659140421e+00, 1.82247232799327e+00, 1.82247169872333e+00, 1.82247165102552e+00, 1.82247080538556e+00, 1.82246507605861e+00, 1.82244401457018e+00, 1.82238787332381e+00, 1.82226501439493e+00, 1.82202974330089e+00, 1.82162066263396e+00, 1.82095963452632e+00, 1.81995143359708e+00, 1.81848416207920e+00, 1.81643048465743e+00, 1.81364972048885e+00, 1.80999080246650e+00, 1.80529607816410e+00, 1.79940588318014e+00, 1.79216376720444e+00, 1.78342219899345e+00, 1.77304852301270e+00, 1.76093089348579e+00, 1.74698387738158e+00, 1.73115340274090e+00, 1.71342073781420e+00, 1.69380522270225e+00, 1.67236553847690e+00, 1.64919938551672e+00, 1.62444154601242e+00, 1.59826041552060e+00, 1.57085319383698e+00, 1.54244001517266e+00, 1.51325736228088e+00, 1.48355114258089e+00, 1.45356980427368e+00, 1.42355783894267e+00, 1.39374995983533e+00, 1.36436617013899e+00, 1.33560785246566e+00, 1.30765492855886e+00, 1.28066406460586e+00, 1.25476783795291e+00, 1.23007473843707e+00, 1.20666985254039e+00, 1.18461606975976e+00, 1.16395565524840e+00, 1.14471204749984e+00, 1.12689176106026e+00, 1.11048629870980e+00, 1.09547400256085e+00, 1.08182179706695e+00, 1.06948679768554e+00, 1.05841777615289e+00, 1.04855648672698e+00, 1.03983886738932e+00, 1.03219613612806e+00, 1.02555580541468e+00, 1.01984263825365e+00, 1.01497956713830e+00, 1.01088859330960e+00, 1.00749167829640e+00, 1.00471163325159e+00, 1.00247300454654e+00, 1.00070294693920e+00, 9.99332068912381e-01, 9.98295229016094e-01, 9.97532257737646e-01, 9.96988576986880e-01, 9.96615689017361e-01, 9.96371508630531e-01, 9.96220516761795e-01, 9.96133719750802e-01, 9.96088406302135e-01, 9.96067702764208e-01, 9.96059936241480e-01, 9.96057823567412e-01, 9.96057511741469e-01, 9.96057487955483e-01, 9.96057064272733e-01, 9.96054193764110e-01, 9.96043641536218e-01, 9.96015513673512e-01, 9.95953959084318e-01, 9.95836084454240e-01, 9.95631130058342e-01, 9.95299951333806e-01, 9.94794847558069e-01, 9.94059774057021e-01, 9.93030967241253e-01, 9.91638001531811e-01, 9.89805283106492e-01, 9.87453966906358e-01, 9.84504260557131e-01, 9.80878052570688e-01, 9.76501774026573e-01, 9.71309375362296e-01, 9.65245276082072e-01, 9.58267128631448e-01, 9.50348231749004e-01, 9.41479435924407e-01, 9.31670405463265e-01, 9.20950137588633e-01, 9.09366686505294e-01, 8.96986095071554e-01, 8.83890592979512e-01, 8.70176171936379e-01, 8.55949689500449e-01, 8.41325679575377e-01, 8.26423056763829e-01, 8.11361893855164e-01, 7.96260428915446e-01, 7.81232424710140e-01, 7.66384963405784e-01, 7.51816718650574e-01, 7.37616709566453e-01, 7.23863510086730e-01, 7.10624864259017e-01, 6.97957644063678e-01, 6.85908080349407e-01, 6.74512198268616e-01, 6.63796394354179e-01, 6.53778101301512e-01, 6.44466496989125e-01, 6.35863224982572e-01, 6.27963103799829e-01, 6.20754810990178e-01, 6.14221535307536e-01, 6.08341595871181e-01, 6.03089031262485e-01, 5.98434164145914e-01, 5.94344148390817e-01, 5.90783505976091e-01, 5.87714660337681e-01, 5.85098471415108e-01, 5.82894775606135e-01, 5.81062931293999e-01, 5.79562367731043e-01, 5.78353132027033e-01, 5.77396426009685e-01, 5.76655122019289e-01, 5.76094244502926e-01, 5.75681402800441e-01, 5.75387159945450e-01, 5.75185322763105e-01, 5.75053140075710e-01, 5.74971398382061e-01, 5.74924407818131e-01, 5.74899875318991e-01, 5.74888666402472e-01, 5.74884461580822e-01, 5.74883317770215e-01, 5.74883148946256e-01, 5.74883135977572e-01, 5.74882903783802e-01, 5.74881330638476e-01, 5.74875547614553e-01, 5.74860132385339e-01, 5.74826397493936e-01, 5.74761794821896e-01, 5.74649461748706e-01, 5.74467932543584e-01, 5.74191036169836e-01, 5.73787998626898e-01, 5.73223763843700e-01, 5.72459541576890e-01, 5.71453583427925e-01, 5.70162178782226e-01, 5.68540851254951e-01, 5.66545723495139e-01, 5.64135004730844e-01, 5.61270542401942e-01, 5.57919368093934e-01, 5.54055160376123e-01, 5.49659544583378e-01, 5.44723153220048e-01, 5.39246381010770e-01, 5.33239785343923e-01, 5.26724104693912e-01, 5.19729892508678e-01, 5.12296789408822e-01, 5.04472479620333e-01, 4.96311395920159e-01, 4.87873249267537e-01, 4.79221463929989e-01, 4.70421596503726e-01, 4.61539808868224e-01, 4.52641452504011e-01, 4.43789806690162e-01, 4.35044997763384e-01, 4.26463112388191e-01, 4.18095505707299e-01, 4.09988295844329e-01, 4.02182029607886e-01, 3.94711500174211e-01, 3.87605695591233e-01, 3.80887856679891e-01, 3.74575623848756e-01, 3.68681254089466e-01, 3.63211891668087e-01, 3.58169878529011e-01, 3.53553093008088e-01, 3.49355307973084e-01, 3.45566561869347e-01, 3.42173538251600e-01, 3.39159951148687e-01, 3.36506934969691e-01, 3.34193438556326e-01, 3.32196623390416e-01, 3.30492265866752e-01, 3.29055162968061e-01, 3.27859539689541e-01, 3.26879455249895e-01, 3.26089203619290e-01, 3.25463702342603e-01, 3.24978862196687e-01, 3.24611929065313e-01, 3.24341788684720e-01, 3.24149224734015e-01, 3.24017121190622e-01, 3.23930600968239e-01, 3.23877094568445e-01, 3.23846334718368e-01, 3.23830275599434e-01, 3.23822938122689e-01, 3.23820185592723e-01, 3.23819436838875e-01, 3.23819326324411e-01, 3.23819317770131e-01, 3.23819163769500e-01, 3.23818120393284e-01, 3.23814284842939e-01, 3.23804060758363e-01, 3.23781686027417e-01, 3.23738837381285e-01, 3.23664328358930e-01, 3.23543915748914e-01, 3.23360227862892e-01, 3.23092825817938e-01, 3.22718406224141e-01, 3.22211150089615e-01, 3.21543218205321e-01, 3.20685387672157e-01, 3.19607817625624e-01, 3.18280924814943e-01, 3.16676341910508e-01, 3.14767923847938e-01, 3.12532760920007e-01, 3.09952152529263e-01, 3.07012493308633e-01, 3.03706024333181e-01, 3.00031406722510e-01, 2.95994083017561e-01, 2.91606402828975e-01, 2.86887502520827e-01, 2.81862942947754e-01, 2.76564123207251e-01, 2.71027500755963e-01, 2.65293658041415e-01, 2.59406262320738e-01, 2.53410968275048e-01, 2.47354312460672e-01, 2.41282644971960e-01, 2.35241137567030e-01, 2.29272899688837e-01, 2.23418225079499e-01, 2.17713982750070e-01, 2.12193157528063e-01, 2.06884537739015e-01, 2.01812541146282e-01, 1.96997165314997e-01, 1.92454045194467e-01, 1.88194598910172e-01, 1.84226242376468e-01, 1.80552654161119e-01, 1.77174073779881e-01, 1.74087619000545e-01, 1.71287610519528e-01, 1.68765895293106e-01, 1.66512162634874e-01, 1.64514249740449e-01, 1.62758435426497e-01, 1.61229722476861e-01, 1.59912110023612e-01, 1.58788857852230e-01, 1.57842744435669e-01, 1.57056319939926e-01, 1.56412154497673e-01, 1.55893080836506e-01, 1.55482429010533e-01, 1.55164249666669e-01, 1.54923521123195e-01, 1.54746334679970e-01, 1.54620052123747e-01, 1.54533429408673e-01, 1.54476701009841e-01, 1.54441620446519e-01, 1.54421453889782e-01, 1.54410925502162e-01, 1.54406115079908e-01, 1.54404310538226e-01, 1.54403819660430e-01, 1.54403747207922e-01, 1.54403741554361e-01, 1.54403639186782e-01, 1.54402945632157e-01, 1.54400396064204e-01, 1.54393599945664e-01, 1.54378727281255e-01, 1.54350246169031e-01, 1.54300722969481e-01, 1.54220695600642e-01, 1.54098629484452e-01, 1.53920964387225e-01, 1.53672258654481e-01, 1.53335434965499e-01, 1.52892128581698e-01, 1.52323135050452e-01, 1.51608949469008e-01, 1.50730383863319e-01, 1.49669243292124e-01, 1.48409035397222e-01, 1.46935682864154e-01, 1.45238204273795e-01, 1.43309326757546e-01, 1.41145994249200e-01, 1.38749738302394e-01, 1.36126884474450e-01, 1.33288575912999e-01, 1.30250606432398e-01, 1.27033067165087e-01, 1.23659822754476e-01, 1.20157843896678e-01, 1.16556431802934e-01, 1.12886376049417e-01, 1.09179089851708e-01, 1.05465765969910e-01, 1.01776592477776e-01, 9.81400610401408e-02, 9.45823918680162e-02, 9.11270900386967e-02, 8.77946383382380e-02, 8.46023231071911e-02, 8.15641824411415e-02, 7.86910608672287e-02, 7.59907513367608e-02, 7.34682038760752e-02, 7.11257803042145e-02, 6.89635358153851e-02, 6.69795106552726e-02, 6.51700182364436e-02, 6.35299194359931e-02, 6.20528761388641e-02, 6.07315800806276e-02, 5.95579555849355e-02, 5.85233368425122e-02, 5.76186219361432e-02, 5.68344068798154e-02, 5.61611035239338e-02, 5.55890453108390e-02, 5.51085846036443e-02, 5.47101847247973e-02, 5.43845090095232e-02, 5.41225081835819e-02, 5.39155063016494e-02, 5.37552844183722e-02, 5.36341601979444e-02, 5.35450608808503e-02, 5.34815864898377e-02, 5.34380599192161e-02, 5.34095606384994e-02, 5.33919391446176e-02, 5.33818099834065e-02, 5.33765220684757e-02, 5.33741060723543e-02, 5.33731997643088e-02, 5.33729532280347e-02, 5.33729168398480e-02, 5.33729139762471e-02, 5.33728618166549e-02, 5.33725084285095e-02, 5.33712093488532e-02, 5.33677465757023e-02, 5.33601688700998e-02, 5.33456585296648e-02, 5.33204307847939e-02, 5.32796717709232e-02, 5.32175207605263e-02, 5.31271018788573e-02, 5.30006097392715e-02, 5.28294521668755e-02, 5.26044512985209e-02, 5.23161017719502e-02, 5.19548814426116e-02, 5.15116062072281e-02, 5.09778163004171e-02, 5.03461772411526e-02, 4.96108749068895e-02, 4.87679815491617e-02, 4.78157684525939e-02, 4.67549418097661e-02, 4.55887814720484e-02, 4.43231675338183e-02, 4.29664869078010e-02, 4.15294205695886e-02, 4.00246211561068e-02, 3.84662991075970e-02, 3.68697425314224e-02, 3.52508005878889e-02, 3.36253619499729e-02, 3.20088587268259e-02, 3.04158225438869e-02, 2.88595138530923e-02, 2.73516386512966e-02, 2.59021592391377e-02, 2.45191981371562e-02, 2.32090276229542e-02, 2.19761323922398e-02, 2.08233300699874e-02, 1.97519335819436e-02, 1.87619401046959e-02, 1.78522327255308e-02, 1.70207827028293e-02, 1.62648423377902e-02, 1.55811210394177e-02, 1.49659400032505e-02, 1.44153635796300e-02, 1.39253074527659e-02, 1.34916250442103e-02, 1.31101742719721e-02, 1.27768672086859e-02, 1.24877054672936e-02, 1.22388043091206e-02, 1.20264084662409e-02, 1.18469024641001e-02, 1.16968178594306e-02, 1.15728393390953e-02, 1.14718111197902e-02, 1.13907445782644e-02, 1.13268275447211e-02, 1.12774352169822e-02, 1.12401422216521e-02, 1.12127349843831e-02, 1.11932233003955e-02, 1.11798498406444e-02, 1.11710962979146e-02, 1.11656849757225e-02, 1.11625748323121e-02, 1.11609512977964e-02, 1.11602095463894e-02, 1.11599312987209e-02, 1.11598556094492e-02, 1.11598444378842e-02, 1.11598435509563e-02, 1.11598272967370e-02, 1.11597171725037e-02, 1.11593123512250e-02, 1.11582333002929e-02, 1.11558720978593e-02, 1.11513511629511e-02, 1.11434925049281e-02, 1.11307996424443e-02, 1.11114543684891e-02, 1.10833304840838e-02, 1.10440264134013e-02, 1.09909181654899e-02, 1.09212333493360e-02, 1.08321458326230e-02, 1.07208891599413e-02, 1.05848850728734e-02, 1.04218815372548e-02, 1.02300927830896e-02, 1.00083322631122e-02, 9.75612842516682e-03, 9.47381305309136e-03, 9.16257288884026e-03, 8.82445740633386e-03, 8.46233889572545e-03, 8.07982511561534e-03, 7.68112913678465e-03, 7.27090496287985e-03, 6.85406047543324e-03, 6.43556090936670e-03, 6.02023645544136e-03, 5.61260691576138e-03, 5.21673468996653e-03, 4.83611459260658e-03, 4.47360499580896e-03, 4.13140000281942e-03, 3.81103795500337e-03, 3.51343891465206e-03, 3.23896340266572e-03, 2.98748581637653e-03, 2.75847705806598e-03, 2.55109097321338e-03, 2.36424865430251e-03, 2.19671485041317e-03, 2.04716252484616e-03, 1.91422449297818e-03, 1.79653354267373e-03, 1.69275333469443e-03, 1.60160179775914e-03, 1.52186771599361e-03, 1.45242075089413e-03, 1.39221546915280e-03, 1.34029059080376e-03, 1.29576505067064e-03, 1.25783236441634e-03, 1.22575437208639e-03, 1.19885498907024e-03, 1.17651430431920e-03, 1.15816323868501e-03, 1.14327893288646e-03, 1.13138099438004e-03, 1.12202866941402e-03, 1.11481892393783e-03, 1.10938533286742e-03, 1.10539760545691e-03, 1.10256152035504e-03, 1.10061901416714e-03, 1.09934815783674e-03, 1.09856277030929e-03, 1.09811145344452e-03, 1.09787588422705e-03, 1.09776826413110e-03, 1.09772789429030e-03, 1.09771691292337e-03, 1.09771529210434e-03, 1.09771516227397e-03, 1.09771276792969e-03, 1.09769654604008e-03, 1.09763691435205e-03, 1.09747797138482e-03, 1.09713019541857e-03, 1.09646441907785e-03, 1.09530742990671e-03, 1.09343957853980e-03, 1.09059480284657e-03, 1.08646348832053e-03, 1.08069856111115e-03, 1.07292512933554e-03, 1.06275383264176e-03, 1.04979780833503e-03, 1.03369283591803e-03, 1.01411978796762e-03, 9.90828043441638e-04, 9.63658088366515e-04, 9.32561256192344e-04, 8.97614557629822e-04, 8.59028878934964e-04, 8.17149419020780e-04, 7.72447901354667e-04, 7.25506598655556e-04, 6.76994478766346e-04, 6.27636088708527e-04, 5.78174661157504e-04, 5.29332608479649e-04, 4.81774481642105e-04, 4.36078278696073e-04, 3.92719302922728e-04, 3.52066214333036e-04, 3.14383696107257e-04, 2.79835059828315e-04, 2.48483055031556e-04, 2.20292870281469e-04, 1.95143299393408e-04, 1.72850994474061e-04, 1.53204099760273e-04, 1.35987870535814e-04, 1.20991044113339e-04, 1.08003912117738e-04, 9.68149266356022e-05, 8.72014801984733e-05, 7.89354396886210e-05, 7.18176251126579e-05, 6.56985639153618e-05, 6.04577323883859e-05, 5.59824325279124e-05, 5.21758842495501e-05, 4.89628995410385e-05, 4.62728735044493e-05, 4.40280883111654e-05, 4.21551178249047e-05, 4.05977869490926e-05, 3.93146124224586e-05, 3.82690515721151e-05, 3.74247857024557e-05, 3.67474724498607e-05, 3.62077418429863e-05, 3.57820148906285e-05, 3.54513918523480e-05, 3.52000538441644e-05, 3.50140903009333e-05, 3.48809420813723e-05, 3.47892834895447e-05, 3.47291047546605e-05, 3.46918283042896e-05, 3.46703775733204e-05, 3.46591723570100e-05, 3.46540512197180e-05, 3.46521298695279e-05, 3.46516072085024e-05, 3.46515300719830e-05, 3.46515239471046e-05, 3.46514122213648e-05, 3.46506551522907e-05, 3.46478721695576e-05, 3.46404540465528e-05, 3.46242210014225e-05, 3.45931371299457e-05, 3.45390914588961e-05, 3.44517472683492e-05, 3.43184461542023e-05, 3.41241431840444e-05, 3.38513994364575e-05, 3.34806582884539e-05, 3.29914212500171e-05, 3.23652576099676e-05, 3.15910487387328e-05, 3.06701641932250e-05, 2.96142058857640e-05, 2.84289673970761e-05, 2.70997389009153e-05, 2.56046206175142e-05, 2.39417999381430e-05, 2.21345499085146e-05, 2.02289200927908e-05, 1.83746717539179e-05, 1.68071638171341e-05, 1.53291097575448e-05, 1.36507566566024e-05, 1.17660144951515e-05, 1.04237639959592e-05, 9.33559958582464e-06, 7.88643200166916e-06, 6.38738302427555e-06, 6.26145937587808e-06, 5.56747994551188e-06, 4.81459106538296e-06, 5.70619646111936e-06, 6.35761098521925e-06, 6.58982386880780e-06, 6.21227862040996e-06, 5.17522863674458e-06, 3.71241279140122e-06, 2.23826496892343e-06, 1.39225733164494e-06, 1.11813101000280e-06, 1.04902461828260e-06, 2.18911473837125e-06, 3.24150944663870e-06, 3.94981370277492e-06, 4.22938630606495e-06, 4.10060225419784e-06, 3.66129135898063e-06, 3.04601456946330e-06, 2.38867953328281e-06, 1.79717542891922e-06, 1.33817844982786e-06, 1.02664246526274e-06, 8.31264102885143e-07, 7.10900792455504e-07, 6.37201225117678e-07, 5.84537602321689e-07, 5.31007202653200e-07, 4.67206760390582e-07, 3.94373941503545e-07, -1.42108547152020e-14, 0.00000000000000e+00, 0.00000000000000e+00, -1.42108547152020e-14, 0.00000000000000e+00, 0.00000000000000e+00, -1.42108547152020e-14, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, -1.42108547152020e-14],
[ 1.10000000000000e+02, 1.09999999857071e+02, 1.09999995437424e+02, 1.09999965493615e+02, 1.09999855416205e+02, 1.09999561984102e+02, 1.09998919783956e+02, 1.09997689774437e+02, 1.09995550403987e+02, 1.09992091610005e+02, 1.09986811936829e+02, 1.09979118913539e+02, 1.09968332735859e+02, 1.09953693204941e+02, 1.09934369795153e+02, 1.09909474657753e+02, 1.09878078319576e+02, 1.09839227804683e+02, 1.09791966886699e+02, 1.09735358160909e+02, 1.09668506595306e+02, 1.09590584165721e+02, 1.09500855092082e+02, 1.09398701068415e+02, 1.09283645727338e+02, 1.09155377421432e+02, 1.09013769271146e+02, 1.08858895358849e+02, 1.08691041975835e+02, 1.08510712976112e+02, 1.08318628560682e+02, 1.08115717188714e+02, 1.07903100746435e+02, 1.07682073545388e+02, 1.07454076111141e+02, 1.07220665012417e+02, 1.06983480138341e+02, 1.06744210850731e+02, 1.06504562334189e+02, 1.06266223271271e+02, 1.06030835723768e+02, 1.05799967844317e+02, 1.05575089807842e+02, 1.05357553160769e+02, 1.05148573646087e+02, 1.04949217472148e+02, 1.04760390943310e+02, 1.04582833348588e+02, 1.04417112997642e+02, 1.04263626291182e+02, 1.04122599707970e+02, 1.03994094578808e+02, 1.03878014498163e+02, 1.03774115196652e+02, 1.03682016664418e+02, 1.03601217278349e+02, 1.03531109647399e+02, 1.03470997852077e+02, 1.03420115718245e+02, 1.03377645733885e+02, 1.03342738191978e+02, 1.03314530125231e+02, 1.03292163590633e+02, 1.03274802865478e+02, 1.03261650132834e+02, 1.03251959264274e+02, 1.03245047351368e+02, 1.03240303694422e+02, 1.03237196026045e+02, 1.03235273826309e+02, 1.03234168672861e+02, 1.03233591660107e+02, 1.03233328012949e+02, 1.03233229108946e+02, 1.03233202204579e+02, 1.03233198233548e+02, 1.03233197712632e+02, 1.03233185575926e+02, 1.03233103347943e+02, 1.03232801069026e+02, 1.03231995303229e+02, 1.03230231898464e+02, 1.03226854744273e+02, 1.03220981772405e+02, 1.03211489293737e+02, 1.03197005588928e+02, 1.03175914465294e+02, 1.03146369252707e+02, 1.03106317430299e+02, 1.03053535748325e+02, 1.02985675334439e+02, 1.02900315855723e+02, 1.02795027360606e+02, 1.02667437972308e+02, 1.02515305182783e+02, 1.02336588148460e+02, 1.02129518166284e+02, 1.01892664456189e+02, 1.01624992521006e+02, 1.01325912689167e+02, 1.00995316916510e+02, 1.00633602437656e+02, 1.00241681310487e+02, 9.98209752290182e+01, 9.93733952359676e+01, 9.89013063246416e+01, 9.84074776302946e+01, 9.78950201238879e+01, 9.73673152723848e+01, 9.68279394593559e+01, 9.62805893365856e+01, 9.57290122872215e+01, 9.51769441209365e+01, 9.46280539056404e+01, 9.40858944239685e+01, 9.35538565380627e+01, 9.30351264858662e+01, 9.25326461910904e+01, 9.20490774928150e+01, 9.15867715365700e+01, 9.11477444500855e+01, 9.07336600346210e+01, 9.03458197320973e+01, 8.99851597162972e+01, 8.96522546669957e+01, 8.93473276256401e+01, 8.90702652743198e+01, 8.88206379891399e+01, 8.85977240596587e+01, 8.84005375110030e+01, 8.82278589976177e+01, 8.80782692488775e+01, 8.79501845350684e+01, 8.78418935901047e+01, 8.77515953801909e+01, 8.76774370526163e+01, 8.76175513439949e+01, 8.75700926808307e+01, 8.75332711752976e+01, 8.75053837127410e+01, 8.74848413504033e+01, 8.74701923031016e+01, 8.74601398825565e+01, 8.74535548816950e+01, 8.74494820498078e+01, 8.74471404827922e+01, 8.74459179467721e+01, 8.74453593537988e+01, 8.74451498052917e+01, 8.74450928029228e+01, 8.74450843894892e+01, 8.74450836467431e+01, 8.74450690930002e+01, 8.74449704892421e+01, 8.74446080112676e+01, 8.74436417770780e+01, 8.74415271921459e+01, 8.74374774897089e+01, 8.74304349627147e+01, 8.74190522043537e+01, 8.74016844655374e+01, 8.73763939952077e+01, 8.73409669458977e+01, 8.72929430950635e+01, 8.72296582467780e+01, 8.71482987371256e+01, 8.70459669754810e+01, 8.69197564268196e+01, 8.67668339011034e+01, 8.65845264980558e+01, 8.63704101008979e+01, 8.61223959676039e+01, 8.58388117804838e+01, 8.55184735272126e+01, 8.51607448304496e+01, 8.47655808323078e+01, 8.43335544618443e+01, 8.38658638243672e+01, 8.33643204724965e+01, 8.28313193438353e+01, 8.22697920599391e+01, 8.16831459721324e+01, 8.10751917536817e+01, 8.04500624851572e+01, 7.98121271379823e+01, 7.91659012434039e+01, 7.85159574352591e+01, 7.78668384997417e+01, 7.72229754910471e+01, 7.65886132630878e+01, 7.59677453312595e+01, 7.53640593133013e+01, 7.47808934133852e+01, 7.42212036794160e+01, 7.36875412303461e+01, 7.31820383841789e+01, 7.27064025890191e+01, 7.22619171805715e+01, 7.18494481633499e+01, 7.14694563711946e+01, 7.11220144756386e+01, 7.08068283776727e+01, 7.05232625525550e+01, 7.02703689322645e+01, 7.00469189142449e+01, 6.98514380802129e+01, 6.96822431930166e+01, 6.95374810098046e+01, 6.94151684044789e+01, 6.93132332326037e+01, 6.92295553017732e+01, 6.91620067367608e+01, 6.91084909604132e+01, 6.90669794579689e+01, 6.90355454638959e+01, 6.90123937148154e+01, 6.89958854558051e+01, 6.89845579735657e+01, 6.89771380583855e+01, 6.89725489637943e+01, 6.89699106311805e+01, 6.89685331666926e+01, 6.89679037878877e+01, 6.89676676855379e+01, 6.89676034599144e+01, 6.89675939803448e+01, 6.89675932199661e+01, 6.89675791871216e+01, 6.89674841125353e+01, 6.89671346082906e+01, 6.89662029582321e+01, 6.89641640645483e+01, 6.89602593380084e+01, 6.89534689819700e+01, 6.89424939510712e+01, 6.89257486687546e+01, 6.89013653578619e+01, 6.88672105676970e+01, 6.88209141610760e+01, 6.87599106498149e+01, 6.86814923352358e+01, 6.85828732270722e+01, 6.84612621941557e+01, 6.83139432681805e+01, 6.81383605120835e+01, 6.79322044191580e+01, 6.76934964733890e+01, 6.74206683190203e+01, 6.71126319929898e+01, 6.67688378885172e+01, 6.63893175449960e+01, 6.59747089832535e+01, 6.55262630949063e+01, 6.50458305064272e+01, 6.45358293203223e+01, 6.39991951274473e+01, 6.34393156173897e+01, 6.28599529107197e+01, 6.22651573167115e+01, 6.16591765107693e+01, 6.10463640832872e+01, 6.04310910390594e+01, 5.98176631829857e+01, 5.92102465221602e+01, 5.86128019797442e+01, 5.80290299766109e+01, 5.74623248767386e+01, 5.69157389389589e+01, 5.63919552441999e+01, 5.58932690130553e+01, 5.54215767260209e+01, 5.49783724575976e+01, 5.45647508136976e+01, 5.41814158237158e+01, 5.38286951033852e+01, 5.35065585921930e+01, 5.32146411907708e+01, 5.29522686778930e+01, 5.27184863619243e+01, 5.25120900011989e+01, 5.23316585955952e+01, 5.21755886952100e+01, 5.20421298848172e+01, 5.19294210836540e+01, 5.18355272527166e+01, 5.17584760334718e+01, 5.16962937624312e+01, 5.16470402264815e+01, 5.16088414556392e+01, 5.15799198038448e+01, 5.15586205538175e+01, 5.15434343057883e+01, 5.15330144759504e+01, 5.15261893390339e+01, 5.15219681971801e+01, 5.15195414374972e+01, 5.15182744437782e+01, 5.15176955424078e+01, 5.15174783762791e+01, 5.15174193018041e+01, 5.15174105825345e+01, 5.15174099100205e+01, 5.15173978338285e+01, 5.15173160158586e+01, 5.15170152444613e+01, 5.15162134992282e+01, 5.15144589038798e+01, 5.15110986610687e+01, 5.15052552244179e+01, 5.14958108217648e+01, 5.14814012788575e+01, 5.14604198990674e+01, 5.14310319243870e+01, 5.13911998281000e+01, 5.13387193638176e+01, 5.12712659169228e+01, 5.11864502770641e+01, 5.10818824865348e+01, 5.09552419406531e+01, 5.08043514530641e+01, 5.06272525890242e+01, 5.04222792552893e+01, 5.01881263583195e+01, 4.99239103401306e+01, 4.96292185999011e+01, 4.93041452206206e+01, 4.89493110352135e+01, 4.85658668542069e+01, 4.81554795808934e+01, 4.77203018805671e+01, 4.72629269523599e+01, 4.67863306791453e+01, 4.62938039275571e+01, 4.57888780039071e+01, 4.52752462645740e+01, 4.47566846992691e+01, 4.42369740387984e+01, 4.37198256547119e+01, 4.32088132437728e+01, 4.27073120090775e+01, 4.22184467228751e+01, 4.17450496521427e+01, 4.12896288458512e+01, 4.08543467579658e+01, 4.04410086728865e+01, 4.00510599750448e+01, 3.96855910113554e+01, 3.93453481567645e+01, 3.90307497029269e+01, 3.87419053190392e+01, 3.84786380405003e+01, 3.82405079817416e+01, 3.80268372063897e+01, 3.78367353931639e+01, 3.76691260929832e+01, 3.75227734749928e+01, 3.73963095075176e+01, 3.72882615201857e+01, 3.71970800543962e+01, 3.71211668410523e+01, 3.70589026576282e+01, 3.70086747217809e+01, 3.69689031861719e+01, 3.69380662187375e+01, 3.69147230932856e+01, 3.68975346846212e+01, 3.68852807661293e+01, 3.68768735491437e+01, 3.68713669828726e+01, 3.68679614484114e+01, 3.68660036247316e+01, 3.68649814702651e+01, 3.68645144408022e+01, 3.68643392421461e+01, 3.68642915838845e+01, 3.68642845496256e+01, 3.68642840170128e+01, 3.68642745818299e+01, 3.68642106570980e+01, 3.68639756633247e+01, 3.68633492585345e+01, 3.68619783988031e+01, 3.68593530837503e+01, 3.68547877806865e+01, 3.68474094179905e+01, 3.68361527173067e+01, 3.68197634892612e+01, 3.67968103406055e+01, 3.67657050235000e+01, 3.67247313930662e+01, 3.66720826240986e+01, 3.66059059745908e+01, 3.65243539829448e+01, 3.64256405668051e+01, 3.63081000828281e+01, 3.61702470446091e+01, 3.60108339217596e+01, 3.58289042989048e+01, 3.56238386967761e+01, 3.53953905755102e+01, 3.51437104627783e+01, 3.48693567643197e+01, 3.45732925842870e+01, 3.42568687444580e+01, 3.39217940606648e+01, 3.35700947158651e+01, 3.32040651688273e+01, 3.28262133817369e+01, 3.24392032017090e+01, 3.20457965002773e+01, 3.16487972207339e+01, 3.12509989038559e+01, 3.08551366735206e+01, 3.04638441700132e+01, 3.00796155890271e+01, 2.97047728344199e+01, 2.93414377859189e+01, 2.89915097460246e+01, 2.86566481816779e+01, 2.83382608548384e+01, 2.80374973210980e+01, 2.77552475887014e+01, 2.74921455224863e+01, 2.72485764045234e+01, 2.70246879660718e+01, 2.68204041971748e+01, 2.66354413080390e+01, 2.64693253311992e+01, 2.63214109823200e+01, 2.61909015128188e+01, 2.60768693720101e+01, 2.59782775431453e+01, 2.58940014274945e+01, 2.58228511292288e+01, 2.57635939493754e+01, 2.57149768382407e+01, 2.56757484909817e+01, 2.56446807085167e+01, 2.56205885931047e+01, 2.56023491112766e+01, 2.55889175418452e+01, 2.55793413374792e+01, 2.55727709669726e+01, 2.55684673719374e+01, 2.55658057639793e+01, 2.55642756021387e+01, 2.55634767193036e+01, 2.55631117030341e+01, 2.55629747728128e+01, 2.55629375244709e+01, 2.55629320266936e+01, 2.55629316139738e+01, 2.55629243496362e+01, 2.55628751326869e+01, 2.55626942060763e+01, 2.55622119229390e+01, 2.55611564620423e+01, 2.55591351433051e+01, 2.55556201004488e+01, 2.55499389838871e+01, 2.55412712689429e+01, 2.55286506287219e+01, 2.55109736924605e+01, 2.54870153436039e+01, 2.54554505130303e+01, 2.54148821900161e+01, 2.53638751090750e+01, 2.53009942823137e+01, 2.52248472480680e+01, 2.51341286171031e+01, 2.50276652426357e+01, 2.49044601485093e+01, 2.47637332503865e+01, 2.46049569243483e+01, 2.44278846349561e+01, 2.42325711375343e+01, 2.40193832068305e+01, 2.37890003850432e+01, 2.35424058336011e+01, 2.32808679454878e+01, 2.30059138540069e+01, 2.27192962986538e+01, 2.24229554503042e+01, 2.21189772709698e+01, 2.18095498427842e+01, 2.14969189220120e+01, 2.11833438239851e+01, 2.08710546585324e+01, 2.05622119054578e+01, 2.02588693061291e+01, 1.99629409995418e+01, 1.96761737103925e+01, 1.94001245901089e+01, 1.91361450352981e+01, 1.88853704980846e+01, 1.86487160033079e+01, 1.84268768377585e+01, 1.82203337027235e+01, 1.80293615331238e+01, 1.78540411797977e+01, 1.76942732107301e+01, 1.75497931913737e+01, 1.74201879317086e+01, 1.73049123186456e+01, 1.72033064712320e+01, 1.71146130521796e+01, 1.70379946364804e+01, 1.69725510744055e+01, 1.69173367933653e+01, 1.68713779648536e+01, 1.68336894247034e+01, 1.68032911840972e+01, 1.67792243127698e+01, 1.67605659223472e+01, 1.67464429340624e+01, 1.67360442875327e+01, 1.67286312407131e+01, 1.67235454285415e+01, 1.67202143899341e+01, 1.67181543382155e+01, 1.67169700352059e+01, 1.67163517287364e+01, 1.67160692208726e+01, 1.67159632426573e+01, 1.67159344140355e+01, 1.67159301589922e+01, 1.67159298405361e+01, 1.67159242482875e+01, 1.67158863599956e+01, 1.67157470788505e+01, 1.67153758081744e+01, 1.67145633001073e+01, 1.67130072834027e+01, 1.67103014619903e+01, 1.67059284164662e+01, 1.66992568708422e+01, 1.66895437030619e+01, 1.66759409751153e+01, 1.66575081320304e+01, 1.66332293644792e+01, 1.66020359453105e+01, 1.65628331375068e+01, 1.65145310356462e+01, 1.64560784555742e+01, 1.63864987430553e+01, 1.63049261511505e+01, 1.62106412600650e+01, 1.61031038049689e+01, 1.59819812576670e+01, 1.58471715932309e+01, 1.56988188717091e+01, 1.55373205769693e+01, 1.53633260673103e+01, 1.51777259812294e+01, 1.49816329707026e+01, 1.47763546586351e+01, 1.45633601873813e+01, 1.43442420937441e+01, 1.41206754738384e+01, 1.38943764654425e+01, 1.36670619730560e+01, 1.34404123104405e+01, 1.32160380740706e+01, 1.29954521377263e+01, 1.27800472255237e+01, 1.25710791244518e+01, 1.23696552719292e+01, 1.21767282173192e+01, 1.19930933116647e+01, 1.18193899179615e+01, 1.16561054385262e+01, 1.15035815072619e+01, 1.13620217747892e+01, 1.12315008086149e+01, 1.11119737278371e+01, 1.10032862849453e+01, 1.09051851914161e+01, 1.08173285561584e+01, 1.07392963646178e+01, 1.06706009703894e+01, 1.06106975998730e+01, 1.05589948837505e+01, 1.05148654274008e+01, 1.04776564170511e+01, 1.04467002315514e+01, 1.04213249939693e+01, 1.04008649562752e+01, 1.03846705682774e+01, 1.03721180430167e+01, 1.03626181993434e+01, 1.03556243422756e+01, 1.03506389360824e+01, 1.03472188358407e+01, 1.03449788710763e+01, 1.03435936191272e+01, 1.03427972637078e+01, 1.03423815022123e+01, 1.03421915391324e+01, 1.03421202777111e+01, 1.03421008929038e+01, 1.03420980317479e+01, 1.03420978175330e+01, 1.03420940547418e+01, 1.03420685612926e+01, 1.03419748449048e+01, 1.03417250331750e+01, 1.03411783352054e+01, 1.03401313772605e+01, 1.03383108152174e+01, 1.03353685891591e+01, 1.03308801411487e+01, 1.03243458628808e+01, 1.03151959714079e+01, 1.03027989241336e+01, 1.02864733751188e+01, 1.02655035416335e+01, 1.02391576938414e+01, 1.02067093063811e+01, 1.01674602275448e+01, 1.01207650432112e+01, 1.00660556555820e+01, 1.00028649799224e+01, 9.93084860430076e+00, 9.84980327289101e+00, 9.75968115183530e+00, 9.66059901866396e+00, 9.55284177298955e+00, 9.43685997944715e+00, 9.31326149799472e+00, 9.18279760174369e+00, 9.04634429802205e+00, 8.90487982728104e+00, 8.75945949600478e+00, 8.61118909148244e+00, 8.46119812461544e+00, 8.31061405707139e+00, 8.16053850421511e+00, 8.01202618510582e+00, 7.86606713983758e+00, 7.72357247884878e+00, 7.58536369319179e+00, 7.45216535874448e+00, 7.32460092294124e+00, 7.20319117318724e+00, 7.08835494743307e+00, 6.98041164987814e+00, 6.87958516658358e+00, 6.78600882535950e+00, 6.69973110211669e+00, 6.62072183507971e+00, 6.54887876446217e+00, 6.48403426586175e+00, 6.42596218932742e+00, 6.37438475192164e+00, 6.32897945902050e+00, 6.28938604807634e+00, 6.25521345781300e+00, 6.22604682586824e+00, 6.20145450922809e+00, 6.18099510547930e+00, 6.16422443058455e+00, 6.15070238277534e+00, 6.13999959490714e+00, 6.13170375213316e+00, 6.12542543094666e+00, 6.12080330218096e+00, 6.11750853658494e+00, 6.11524825848505e+00, 6.11376791125033e+00, 6.11285242721925e+00, 6.11232613288122e+00, 6.11205136504901e+00, 6.11192582255413e+00, 6.11187872742370e+00, 6.11186591642462e+00, 6.11186402554866e+00, 6.11186388360315e+00, 6.11186138524303e+00, 6.11184445849622e+00, 6.11178223414001e+00, 6.11161636804795e+00, 6.11125338023095e+00, 6.11055823839052e+00, 6.10934945366505e+00, 6.10739593155445e+00, 6.10441579175991e+00, 6.10007733475611e+00, 6.09400228757129e+00, 6.08577140357632e+00, 6.07493241865306e+00, 6.06101027771685e+00, 6.04351944191600e+00, 6.02197797117154e+00, 5.99592295517328e+00, 5.96492674747871e+00, 5.92861335321834e+00, 5.88667424355822e+00, 5.83888283159021e+00, 5.78510685454752e+00, 5.72531797188743e+00, 5.65959800793850e+00, 5.58814143522683e+00, 5.51125389798334e+00, 5.42934679772343e+00, 5.34292818479192e+00, 5.25259040222399e+00, 5.15899509480987e+00, 5.06285631512584e+00, 4.96492252330907e+00, 4.86595828777561e+00, 4.76672645380930e+00, 4.66797146336327e+00, 4.57040439218767e+00, 4.47469013050006e+00, 4.38143698231590e+00, 4.29118880751579e+00, 4.20441969006745e+00, 4.12153099423395e+00, 4.04285057446180e+00, 3.96863383759990e+00, 3.89906631891709e+00, 3.83426742419107e+00, 3.77429500495306e+00, 3.71915046743125e+00, 3.66878416185946e+00, 3.62310085170839e+00, 3.58196511674271e+00, 3.54520659522419e+00, 3.51262501574179e+00, 3.48399500579708e+00, 3.45907069110785e+00, 3.43759011616984e+00, 3.41927952317842e+00, 3.40385752379859e+00, 3.39103918779763e+00, 3.38054005594257e+00, 3.37208006385657e+00, 3.36538734098527e+00, 3.36020182681749e+00, 3.35627862734893e+00, 3.35339102057090e+00, 3.35133301220301e+00, 3.34992134309172e+00, 3.34899685810304e+00, 3.34842516262584e+00, 3.34809651593392e+00, 3.34792493896201e+00, 3.34784654544654e+00, 3.34781713756219e+00, 3.34780913792630e+00, 3.34780795719760e+00, 3.34780786818438e+00, 3.34780629644888e+00, 3.34779564772070e+00, 3.34775650203676e+00, 3.34765215530003e+00, 3.34742380153625e+00, 3.34698650173496e+00, 3.34622611186752e+00, 3.34499732954278e+00, 3.34312300293111e+00, 3.34039482526362e+00, 3.33657551107606e+00, 3.33140251397718e+00, 3.32459329799455e+00, 3.31585211431143e+00, 3.30487816273657e+00, 3.29137493473267e+00, 3.27506044664614e+00, 3.25567798447915e+00, 3.23300690350102e+00, 3.20687296665392e+00, 3.17715767448986e+00, 3.14380604433768e+00, 3.10683234281903e+00, 3.06632336499297e+00, 3.02243898177956e+00, 2.97540983639495e+00, 2.92553224743111e+00, 2.87316055499385e+00, 2.81869730989531e+00, 2.76258183831554e+00, 2.70527780289704e+00, 2.64726041823624e+00, 2.58900396249894e+00, 2.53097016180097e+00, 2.47359791965846e+00, 2.41729473343510e+00, 2.36242999809051e+00, 2.30933025897593e+00, 2.25827635200136e+00, 2.20950226986136e+00, 2.16319552188270e+00, 2.11949871340032e+00, 2.07851205620589e+00, 2.04029653017228e+00, 2.00487744207028e+00, 1.97224816499511e+00, 1.94237388524741e+00, 1.91519522834912e+00, 1.89063167855932e+00, 1.86858474430682e+00, 1.84894085385520e+00, 1.83157399053189e+00, 1.81634809483734e+00, 1.80311927196324e+00, 1.79173784820759e+00, 1.78205031917074e+00, 1.77390122724319e+00, 1.76713499664872e+00, 1.76159774216447e+00, 1.75713905367832e+00, 1.75361374412056e+00, 1.75088353423163e+00, 1.74881863530481e+00, 1.74729918159154e+00, 1.74621645840716e+00, 1.74547387077490e+00, 1.74498760097424e+00, 1.74468691147382e+00, 1.74451406187578e+00, 1.74442382374134e+00, 1.74438259431878e+00, 1.74436712792119e+00, 1.74436292070325e+00, 1.74436229972731e+00, 1.74436225265821e+00, 1.74436141816409e+00, 1.74435576435229e+00, 1.74433498047118e+00, 1.74427957926720e+00, 1.74415834012228e+00, 1.74392617194846e+00, 1.74352248991408e+00, 1.74287019448450e+00, 1.74187533429847e+00, 1.74042752479757e+00, 1.73840118064896e+00, 1.73565760011981e+00, 1.73204791216490e+00, 1.72741686118120e+00, 1.72160736026514e+00, 1.71446569283307e+00, 1.70584718763713e+00, 1.69562213809988e+00, 1.68368168936597e+00, 1.66994338210459e+00, 1.65435602732528e+00, 1.63690359652205e+00, 1.61760784929960e+00, 1.59652948605962e+00, 1.57376770250636e+00, 1.54945812829541e+00, 1.52376924393724e+00, 1.49689747643947e+00, 1.46906126369151e+00, 1.44049444076673e+00, 1.41143933201403e+00, 1.38213992915075e+00, 1.35283550001483e+00, 1.32375491138947e+00, 1.29511187111875e+00, 1.26710120936875e+00, 1.23989623571121e+00, 1.21364713541494e+00, 1.18848031023319e+00, 1.16449852873055e+00, 1.14178172898892e+00, 1.12038831061297e+00, 1.10035676027010e+00, 1.08170747188716e+00, 1.06444464540468e+00, 1.04855817336542e+00, 1.03402544994759e+00, 1.02081306043418e+00, 1.00887832933860e+00, 9.98170721871290e-01, 9.88633105981108e-01, 9.80202890993681e-01, 9.72813064225477e-01, 9.66393149301638e-01, 9.60870109679249e-01, 9.56169218508094e-01, 9.52214911850305e-01, 9.48931636831290e-01, 9.46244699908178e-01, 9.44081113550041e-01, 9.42370432686886e-01, 9.41045565788855e-01, 9.40043539885025e-01, 9.39306194687546e-01, 9.38780778657701e-01, 9.38420419613379e-01, 9.38184444456240e-01, 9.38038526722011e-01, 9.37954646669695e-01, 9.37910856088479e-01, 9.37890848350335e-01, 9.37883342844373e-01, 9.37881301173022e-01, 9.37880999826916e-01, 9.37880976840361e-01, 9.37880567396704e-01, 9.37877793360215e-01, 9.37867595772602e-01, 9.37840413239883e-01, 9.37780927427113e-01, 9.37667014540239e-01, 9.37468948988069e-01, 9.37148902689941e-01, 9.36660780083258e-01, 9.35950424145105e-01, 9.34956221852772e-01, 9.33610127579044e-01, 9.31839109184281e-01, 9.29567003566035e-01, 9.26716746230909e-01, 9.23212913833936e-01, 9.18984491200007e-01, 9.13967747531856e-01, 9.08109083430674e-01, 9.01367694459353e-01, 8.93717891560399e-01, 8.85150926254624e-01, 8.75676190407688e-01, 8.65321695893030e-01, 8.54133786151024e-01, 8.42176085084063e-01, 8.29527743288139e-01, 8.16281091201517e-01, 8.02538847757845e-01, 7.88411057399244e-01, 7.74011935747225e-01, 7.59456795092760e-01, 7.44859197570548e-01, 7.30328450447900e-01, 7.15967519238589e-01, 7.01871395192697e-01, 6.88125918237631e-01, 6.74807027586070e-01, 6.61980391536432e-01, 6.49701355719301e-01, 6.38015144413288e-01, 6.26957251108252e-01, 6.16553960491117e-01, 6.06822952726446e-01, 5.97773950794533e-01, 5.89409381556706e-01, 5.81725030344799e-01, 5.74710676742001e-01, 5.68350705633478e-01, 5.62624692533163e-01, 5.57507965726288e-01, 5.52972150045093e-01, 5.48985698272091e-01, 5.45514416383639e-01, 5.42521988237525e-01, 5.39970503990062e-01, 5.37820994616411e-01, 5.36033972530120e-01, 5.34569975598686e-01, 5.33390109004117e-01, 5.32456576595507e-01, 5.31733190844704e-01, 5.31185848463309e-01, 5.30782957385000e-01, 5.30495800329618e-01, 5.30298820666701e-01, 5.30169817827002e-01, 5.30090042022323e-01, 5.30044181399916e-01, 5.30020238759718e-01, 5.30009299338090e-01, 5.30005195610627e-01, 5.30004079299815e-01, 5.30003914534774e-01, 5.30003901877890e-01, 5.30003675266556e-01, 5.30002139943079e-01, 5.29996495954322e-01, 5.29981451327430e-01, 5.29948527417076e-01, 5.29885477590796e-01, 5.29775844053574e-01, 5.29598675665611e-01, 5.29328427256672e-01, 5.28935057919824e-01, 5.28384341734991e-01, 5.27638398960917e-01, 5.26656448628501e-01, 5.25395774495209e-01, 5.23812885540764e-01, 5.21864839977297e-01, 5.19510688856414e-01, 5.16712982891704e-01, 5.13439275477467e-01, 5.09663547602955e-01, 5.05367477884164e-01, 5.00541484342278e-01, 4.95185474326391e-01, 4.89309254774057e-01, 4.82932575643829e-01, 4.76084802883804e-01, 4.68804241284289e-01, 4.61137149430044e-01, 4.53136506462613e-01, 4.44860601872520e-01, 4.36371524350250e-01, 4.27733624036904e-01, 4.19012015339561e-01, 4.10271176348317e-01, 4.01573687563513e-01, 3.92979138769135e-01, 3.84543219821708e-01, 3.76316999779021e-01, 3.68346389652544e-01, 3.60671777247333e-01, 3.53327817907527e-01, 3.46343362221091e-01, 3.39741500519054e-01, 3.33539703993480e-01, 3.27750043171463e-01, 3.22379466077365e-01, 3.17430120493142e-01, 3.12899707111626e-01, 3.08781852915487e-01, 3.05066496652202e-01, 3.01740280671865e-01, 2.98786945517179e-01, 2.96187725390340e-01, 2.93921743877675e-01, 2.91966410030795e-01, 2.90297815061905e-01, 2.88891129524785e-01, 2.87720999986007e-01, 2.86761942938469e-01, 2.85988732205823e-01, 2.85376774494651e-01, 2.84902466241604e-01, 2.84543523653852e-01, 2.84279277010754e-01, 2.84090920011664e-01, 2.83961705302730e-01, 2.83877078322163e-01, 2.83824743237773e-01, 2.83794656923234e-01, 2.83778949492770e-01, 2.83771772715028e-01, 2.83769080470378e-01, 2.83768348115757e-01, 2.83768240021786e-01, 2.83768231654847e-01, 2.83768081027148e-01, 2.83767060503138e-01, 2.83763308960857e-01, 2.83753308818362e-01, 2.83731424217109e-01, 2.83689514373222e-01, 2.83616638390626e-01, 2.83498866429568e-01, 2.83319210454891e-01, 2.83057685703000e-01, 2.82691511318205e-01, 2.82195455114774e-01, 2.81542322955985e-01, 2.80703587714214e-01, 2.79650146229841e-01, 2.78353185326097e-01, 2.76785130168207e-01, 2.74920640662387e-01, 2.72737614922647e-01, 2.70218153905418e-01, 2.67349438921485e-01, 2.64124474515782e-01, 2.60542653524396e-01, 2.56610108962832e-01, 2.52339828344716e-01, 2.47751519262479e-01, 2.42871229461443e-01, 2.37730738932768e-01, 2.32366754483024e-01, 2.26819947719164e-01, 2.21133884637638e-01, 2.15353898610985e-01, 2.09525958490232e-01, 2.03695580092472e-01, 1.97906823087237e-01, 1.92201406974050e-01, 1.86617970278476e-01, 1.81191487113367e-01, 1.75952845608307e-01, 1.70928584045228e-01, 1.66140773323164e-01, 1.61607028911178e-01, 1.57340631853202e-01, 1.53350736626052e-01, 1.49642643546358e-01, 1.46218114704112e-01, 1.43075714743404e-01, 1.40211160859081e-01, 1.37617669793457e-01, 1.35286293090772e-01, 1.33206235150936e-01, 1.31365151517713e-01, 1.29749427208750e-01, 1.28344436661138e-01, 1.27134787993953e-01, 1.26104554786778e-01, 1.25237498484253e-01, 1.24517283939369e-01, 1.23927689602283e-01, 1.23452812573873e-01, 1.23077267312382e-01, 1.22786375354650e-01, 1.22566342139137e-01, 1.22404416026342e-01, 1.22289024013710e-01, 1.22209878512820e-01, 1.22158049922604e-01, 1.22126000586121e-01, 1.22107576991326e-01, 1.22097958676108e-01, 1.22093564093873e-01, 1.22091915551394e-01, 1.22091467109470e-01, 1.22091400920439e-01, 1.22091395755604e-01, 1.22091302237635e-01, 1.22090668640482e-01, 1.22088339484648e-01, 1.22082130919623e-01, 1.22068544181758e-01, 1.22042526079454e-01, 1.21997287034716e-01, 1.21924186575754e-01, 1.21812695392165e-01, 1.21650441989800e-01, 1.21423350447117e-01, 1.21115873587598e-01, 1.20711322887445e-01, 1.20192292537055e-01, 1.19541170279732e-01, 1.18740722108598e-01, 1.17774731923063e-01, 1.16628671294791e-01, 1.15290369169870e-01, 1.13750647315584e-01, 1.12003885279620e-01, 1.10048479133440e-01, 1.07887161681475e-01, 1.05527158198882e-01, 1.02980160827599e-01, 1.00262115874301e-01, 9.73928304714633e-02, 9.43954172423531e-02, 9.12956065542687e-02, 8.81209645595931e-02, 8.49000606631307e-02, 8.16616298383934e-02, 7.84337732853828e-02, 7.52432356402721e-02, 7.21147890344156e-02, 6.90707447165551e-02, 6.61306027476058e-02, 6.33108404537381e-02, 6.06248317098448e-02, 5.80828823104440e-02, 5.56923619970604e-02, 5.34579112079427e-02, 5.13817001386769e-02, 4.94637189043772e-02, 4.77020800382917e-02, 4.60933177799348e-02, 4.46326721768315e-02, 4.33143496018573e-02, 4.21317546208684e-02, 4.10776910594279e-02, 4.01445325118459e-02, 3.93243643751191e-02, 3.86091007771086e-02, 3.79905805483105e-02, 3.74606467162408e-02, 3.70112139458882e-02, 3.66343279755483e-02, 3.63222204668574e-02, 3.60673618666425e-02, 3.58625139310078e-02, 3.57007825596298e-02, 3.55756706003376e-02, 3.54811293840243e-02, 3.54116070060968e-02, 3.53620908391719e-02, 3.53281414896998e-02, 3.53059154160604e-02, 3.52921737080010e-02, 3.52842750574069e-02, 3.52801516754795e-02, 3.52782677653778e-02, 3.52775610613776e-02, 3.52773688222499e-02, 3.52773404482036e-02, 3.52773382152947e-02, 3.52772975433311e-02, 3.52770219856922e-02, 3.52760090189435e-02, 3.52733089164019e-02, 3.52674002896691e-02, 3.52560864079976e-02, 3.52364172470203e-02, 3.52046421537011e-02, 3.51561978959012e-02, 3.50857365833974e-02, 3.49871973678830e-02, 3.48539248094681e-02, 3.46788352032377e-02, 3.44546299148192e-02, 3.41740518699396e-02, 3.38301778844823e-02, 3.34167357346757e-02, 3.29284311048070e-02, 3.23612662675998e-02, 3.17128300686278e-02, 3.09825379935234e-02, 3.01718022002149e-02, 2.92841146000740e-02, 2.83250313133152e-02, 2.73020537527771e-02, 2.62244095511619e-02, 2.51027446920205e-02, 2.39487455739322e-02, 2.27747154423241e-02, 2.15931329685333e-02, 2.04162213619412e-02, 1.92555542567874e-02, 1.81217200747597e-02, 1.70240602906091e-02, 1.59704898787680e-02, 1.49674010981471e-02, 1.40196454726436e-02, 1.31305839022531e-02, 1.23021915699013e-02, 1.15352027169990e-02, 1.08292802982106e-02, 1.01831967251940e-02, 9.59501401338514e-03, 9.06225425688945e-03, 8.58205406206025e-03, 8.15129906594336e-03, 7.76673676693918e-03, 7.42506756316175e-03, 7.12301514690239e-03, 6.85737828999322e-03, 6.62506660225404e-03, 6.42312309415161e-03, 6.24873638071222e-03, 6.09924519891081e-03, 5.97213764278592e-03, 5.86504720271819e-03, 5.77574735123676e-03, 5.70214606393904e-03, 5.64228130204469e-03, 5.59431812311573e-03, 5.55654773687309e-03, 5.52738849380319e-03, 5.50538849994098e-03, 5.48922930647677e-03, 5.47772993809303e-03, 5.46985041441417e-03, 5.46469388939386e-03, 5.46150658503564e-03, 5.45967482503329e-03, 5.45871866208358e-03, 5.45828182461605e-03, 5.45811795848294e-03, 5.45807338355075e-03, 5.45806680440819e-03, 5.45806628207401e-03, 5.45805670965649e-03, 5.45799185556461e-03, 5.45775345027266e-03, 5.45711798925197e-03, 5.45572750093015e-03, 5.45306532586665e-03, 5.44843821000995e-03, 5.44096607836764e-03, 5.42958090767343e-03, 5.41303611299782e-03, 5.38992775051383e-03, 5.35872855236619e-03, 5.31783529628171e-03, 5.26562923695906e-03, 5.20054828598404e-03, 5.12116836947030e-03, 5.02629002244248e-03, 4.91502495651730e-03, 4.78687627396823e-03, 4.64180541744952e-03, 4.48027904805315e-03, 4.30328996101537e-03, 4.11234790148285e-03, 3.90943860800519e-03, 3.69695234047640e-03, 3.47758618661942e-03, 3.25422718333357e-03, 3.02982533418117e-03, 2.80726661350172e-03, 2.58925581124458e-03, 2.37821758851453e-03, 2.17622164461773e-03, 1.98493494320928e-03, 1.80560109785688e-03, 1.63904473605214e-03, 1.48569709570268e-03, 1.34563813446675e-03, 1.21864985220554e-03, 1.10427530367474e-03, 1.00187805733754e-03, 9.10697775481140e-04, 8.29899024409997e-04, 7.58611975300028e-04, 6.95964874665833e-04, 6.41108839772642e-04, 5.93235772143430e-04, 5.51590243688338e-04, 5.15476277769267e-04, 4.84260032564521e-04, 4.57369413936703e-04, 4.34291547307453e-04, 4.14568859653741e-04, 3.97794337047230e-04, 3.83606382726498e-04, 3.71683609074580e-04, 3.61739829045860e-04, 3.53519448552220e-04, 3.46793395905820e-04, 3.41355662129672e-04, 3.37020474404426e-04, 3.33620082727748e-04, 3.31003110034089e-04, 3.29033391963662e-04, 3.27589213398483e-04, 3.26562834430888e-04, 3.25860190756089e-04, 3.25400650268648e-04, 3.25116715814033e-04, 3.24953576154030e-04, 3.24868429942171e-04, 3.24829532189597e-04, 3.24814941308205e-04, 3.24810972344380e-04, 3.24810386544527e-04, 3.24810339591863e-04, 3.24809474065546e-04, 3.24803610112667e-04, 3.24782054377692e-04, 3.24724600744730e-04, 3.24598895872441e-04, 3.24358274554015e-04, 3.23940202974882e-04, 3.23265482890633e-04, 3.22238386701201e-04, 3.20747900346419e-04, 3.18670241327368e-04, 3.15872783701820e-04, 3.12219449753570e-04, 3.07577515044954e-04, 3.01825622074148e-04, 2.94862616058822e-04, 2.86616626681280e-04, 2.77053651132064e-04, 2.66184781722245e-04, 2.54071193296568e-04, 2.40826082929857e-04, 2.26612937225923e-04, 2.11639791302787e-04, 1.96149533422840e-04, 1.80406810756040e-04, 1.64682705843688e-04, 1.49239011392410e-04, 1.34314389185874e-04, 1.20114545694605e-04, 1.06807488563732e-04, 9.45231173972161e-05, 8.33545662430879e-05, 7.33580228171604e-05, 6.45494393722856e-05, 5.69003591692763e-05, 5.03381729828334e-05, 4.47555920146669e-05, 4.00296771800868e-05, 3.60439683078084e-05, 3.27031203966044e-05, 2.99339577338742e-05, 2.76756999113559e-05, 2.58663475420917e-05, 2.44330749410437e-05, 2.32929783265945e-05, 2.23628104976115e-05, 2.15699060248653e-05, 2.08590605978998e-05, 2.01946577789158e-05, 1.95589600338053e-05, 1.89482041008660e-05, 1.83675987273091e-05, 1.78256773466501e-05, 1.73296791956545e-05, 1.68838698328955e-05, 1.64902242261178e-05, 1.61492397978691e-05, 1.58598980419811e-05, 1.56194450937619e-05, 1.54236620488746e-05, 1.52675011975134e-05, 1.51456642640824e-05, 1.50529394886689e-05, 1.49843612007317e-05, 1.49353041223321e-05, 1.49015640289463e-05, 1.48794278658215e-05, 1.48657230596427e-05, 1.48578391332421e-05, 1.48537215665101e-05, 1.48518398930264e-05, 1.48511339546076e-05, 1.48509419233278e-05, 1.48509135726727e-05, 1.48509112989359e-05, 1.48508693769145e-05, 1.48505853445613e-05, 1.48495411735894e-05, 1.48467574803135e-05, 1.48406639794985e-05, 1.48289884691621e-05, 1.48086668332326e-05, 1.47757739910048e-05, 1.47254750686443e-05, 1.46519964374647e-05, 1.45486221043711e-05, 1.44077372539186e-05, 1.42209649851566e-05, 1.39794573357221e-05, 1.36743613694534e-05, 1.32973660811331e-05, 1.28411498536707e-05, 1.22997420390902e-05, 1.16692957590203e-05, 1.09497762963429e-05, 1.01467822304357e-05, 9.27171731746057e-06, 8.34122668891268e-06, 7.38054318105696e-06, 6.42875855305647e-06, 5.52952808163809e-06, 4.70593694501531e-06, 3.96931601187589e-06, 3.36463881467353e-06, 2.90267564651003e-06, 2.45151771594010e-06, 1.97278745872609e-06, 1.70934437448977e-06, 1.41780238038791e-06, 1.20127387503999e-06, 2.64106188296864e-06, 4.01348195566698e-06, 4.85474144795717e-06, 5.03987553202023e-06, 4.62391156474951e-06, 3.81485973832696e-06, 2.93693886987967e-06, 2.21505510467068e-06, 1.58726723498148e-06, 1.65122317241639e-06, 2.05699160460426e-06, 2.39044319982895e-06, 2.46550638394183e-06, 2.28740849195219e-06, 1.95785771950341e-06, 1.60457736342323e-06, 1.30347611104753e-06, 1.04775897113996e-06, 8.24652630626588e-07, 6.49746596081968e-07, 5.27771803149335e-07, 4.40880256746823e-07, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, -1.42108547152020e-14, 0.00000000000000e+00, -1.42108547152020e-14, 1.42108547152020e-14, 0.00000000000000e+00, 1.42108547152020e-14, 0.00000000000000e+00],
[ 1.11000000000000e+02, 1.10999999855272e+02, 1.10999995379994e+02, 1.10999965059277e+02, 1.10999853596300e+02, 1.10999556470700e+02, 1.10998906186989e+02, 1.10997660694805e+02, 1.10995494394641e+02, 1.10991992061059e+02, 1.10986645923599e+02, 1.10978856048055e+02, 1.10967934062895e+02, 1.10953110183031e+02, 1.10933543401797e+02, 1.10908334656195e+02, 1.10876542722655e+02, 1.10837202569425e+02, 1.10789345871665e+02, 1.10732023376542e+02, 1.10664328775090e+02, 1.10585423681893e+02, 1.10494563232841e+02, 1.10391121683040e+02, 1.10274617230619e+02, 1.10144735129656e+02, 1.10001348019513e+02, 1.09844532327019e+02, 1.09674579627222e+02, 1.09492002000903e+02, 1.09297530705269e+02, 1.09092107857137e+02, 1.08876871272727e+02, 1.08653133057927e+02, 1.08422352938307e+02, 1.08186107609062e+02, 1.07946057540995e+02, 1.07703912692709e+02, 1.07461398467905e+02, 1.07220223053413e+02, 1.06982047020419e+02, 1.06748455809257e+02, 1.06520935480362e+02, 1.06300851921418e+02, 1.06089433561629e+02, 1.05887757555573e+02, 1.05696739351541e+02, 1.05517125539275e+02, 1.05349489866762e+02, 1.05194232314365e+02, 1.05051581109852e+02, 1.04921597555898e+02, 1.04804183521168e+02, 1.04699091417900e+02, 1.04605936454817e+02, 1.04524210916262e+02, 1.04453300179069e+02, 1.04392500139777e+02, 1.04341035688431e+02, 1.04298079833251e+02, 1.04262773054642e+02, 1.04234242449401e+02, 1.04211620218118e+02, 1.04194061052431e+02, 1.04180757995314e+02, 1.04170956377711e+02, 1.04163965478984e+02, 1.04159167616250e+02, 1.04156024437551e+02, 1.04154080273862e+02, 1.04152962492554e+02, 1.04152378886696e+02, 1.04152112227043e+02, 1.04152012192941e+02, 1.04151984981159e+02, 1.04151980964754e+02, 1.04151980437886e+02, 1.04151968162503e+02, 1.04151884994966e+02, 1.04151579262152e+02, 1.04150764289555e+02, 1.04148980736055e+02, 1.04145564995083e+02, 1.04139624921909e+02, 1.04130023993513e+02, 1.04115374829542e+02, 1.04094042795717e+02, 1.04064160168638e+02, 1.04023651056606e+02, 1.03970266939556e+02, 1.03901632311318e+02, 1.03815299484094e+02, 1.03708811162216e+02, 1.03579768934289e+02, 1.03425905405777e+02, 1.03245157343576e+02, 1.03035736980532e+02, 1.02796198576513e+02, 1.02525497479139e+02, 1.02223039261687e+02, 1.01888716982914e+02, 1.01522935118804e+02, 1.01126619158096e+02, 1.00701210179879e+02, 1.00248644002172e+02, 9.97713148971947e+01, 9.92720246630476e+01, 9.87539191482144e+01, 9.82204159286905e+01, 9.76751281242576e+01, 9.71217895703403e+01, 9.65641853837823e+01, 9.60060897767608e+01, 9.54512107678880e+01, 9.49031401526693e+01, 9.43653070551313e+01, 9.38409342370166e+01, 9.33329974303644e+01, 9.28441887453400e+01, 9.23768854735273e+01, 9.19331254242663e+01, 9.15145894959374e+01, 9.11225916931197e+01, 9.07580763861121e+01, 9.04216223278127e+01, 9.01134527946428e+01, 8.98334511730903e+01, 8.95811813322553e+01, 8.93559121697820e+01, 8.91566457671947e+01, 8.89821486245082e+01, 8.88309854552758e+01, 8.87015550106439e+01, 8.85921273674961e+01, 8.85008820671204e+01, 8.84259464343404e+01, 8.83654333508558e+01, 8.83174777090139e+01, 8.82802707414405e+01, 8.82520914151156e+01, 8.82313341013994e+01, 8.82165317900769e+01, 8.82063742072255e+01, 8.81997203225098e+01, 8.81956048876154e+01, 8.81932388277324e+01, 8.81920035041210e+01, 8.81914390683672e+01, 8.81912273280275e+01, 8.81911697294265e+01, 8.81911612279903e+01, 8.81911604774753e+01, 8.81911457715035e+01, 8.81910461363731e+01, 8.81906798669700e+01, 8.81897035262649e+01, 8.81875668236807e+01, 8.81834747640967e+01, 8.81763585801796e+01, 8.81648567793852e+01, 8.81473074264598e+01, 8.81217525374560e+01, 8.80859551749308e+01, 8.80374294985259e+01, 8.79734836347980e+01, 8.78912747836372e+01, 8.77878754813178e+01, 8.76603494064026e+01, 8.75058345685467e+01, 8.73216311957522e+01, 8.71052911752071e+01, 8.68547055541915e+01, 8.65681864191378e+01, 8.62445394863787e+01, 8.58831239893355e+01, 8.54838969466860e+01, 8.50474396304479e+01, 8.45749649751377e+01, 8.40683056978516e+01, 8.35298839250713e+01, 8.29626640260837e+01, 8.23700910343394e+01, 8.17560174454317e+01, 8.11246213319016e+01, 8.04803186938481e+01, 7.98276728823518e+01, 7.91713038742634e+01, 7.85158001490512e+01, 7.78656358412803e+01, 7.72250955945718e+01, 7.65982090423115e+01, 7.59886961102074e+01, 7.53999235102810e+01, 7.48348720585954e+01, 7.42961139448307e+01, 7.37857988586652e+01, 7.33056478871638e+01, 7.28569542390122e+01, 7.24405900277183e+01, 7.20570184943472e+01, 7.17063111489113e+01, 7.13881693637393e+01, 7.11019499779755e+01, 7.08466944838967e+01, 7.06211613700307e+01, 7.04238611934680e+01, 7.02530939408564e+01, 7.01069882104088e+01, 6.99835417037260e+01, 6.98806624572908e+01, 6.97962101733720e+01, 6.97280369358690e+01, 6.96740265274129e+01, 6.96321315098140e+01, 6.96004072005425e+01, 6.95770416818813e+01, 6.95603810230045e+01, 6.95489489817977e+01, 6.95414605825641e+01, 6.95368291340581e+01, 6.95341664524144e+01, 6.95327762756157e+01, 6.95321410884686e+01, 6.95319028072119e+01, 6.95318379888733e+01, 6.95318284218201e+01, 6.95318276544241e+01, 6.95318134920757e+01, 6.95317175400809e+01, 6.95313648103967e+01, 6.95304245625507e+01, 6.95283668530738e+01, 6.95244260930136e+01, 6.95175730774332e+01, 6.95064967802559e+01, 6.94895970095000e+01, 6.94649887861377e+01, 6.94305190366474e+01, 6.93837958665462e+01, 6.93222303035452e+01, 6.92430899628093e+01, 6.91435635985908e+01, 6.90208349809058e+01, 6.88721639978165e+01, 6.86949723679912e+01, 6.84869308968928e+01, 6.82460448690999e+01, 6.79707339827708e+01, 6.76599032357459e+01, 6.73130013879968e+01, 6.69300640562996e+01, 6.65117391299971e+01, 6.60592930009487e+01, 6.55745970325343e+01, 6.50600946978040e+01, 6.45187508319233e+01, 6.39539853952305e+01, 6.33695949478261e+01, 6.27696656097703e+01, 6.21584815503036e+01, 6.15404329752318e+01, 6.09199271748437e+01, 6.03013055245202e+01, 5.96887685131204e+01, 5.90863100471317e+01, 5.84976615631851e+01, 5.79262459540996e+01, 5.73751409907222e+01, 5.68470517671264e+01, 5.63442916442069e+01, 5.58687711495791e+01, 5.54219942652569e+01, 5.50050614851289e+01, 5.46186789637890e+01, 5.42631730305295e+01, 5.39385093281140e+01, 5.36443158640305e+01, 5.33799093270557e+01, 5.31443241088610e+01, 5.29363435598988e+01, 5.27545330834665e+01, 5.25972747192746e+01, 5.24628028819064e+01, 5.23492408999595e+01, 5.22546379528385e+01, 5.21770059319558e+01, 5.21143556716089e+01, 5.20647319133260e+01, 5.20262462975797e+01, 5.19971076293881e+01, 5.19756486488010e+01, 5.19603485605872e+01, 5.19498506434581e+01, 5.19429743681987e+01, 5.19387216027578e+01, 5.19362766638485e+01, 5.19350001792819e+01, 5.19344169415442e+01, 5.19341981487072e+01, 5.19341386317293e+01, 5.19341298471472e+01, 5.19341291695957e+01, 5.19341170029460e+01, 5.19340345721122e+01, 5.19337315477729e+01, 5.19329237971051e+01, 5.19311560594221e+01, 5.19277706490487e+01, 5.19218834507419e+01, 5.19123683308963e+01, 5.18978509221690e+01, 5.18767125443269e+01, 5.18471047922670e+01, 5.18069750451425e+01, 5.17541028219938e+01, 5.16861465269972e+01, 5.16006996956104e+01, 5.14953553839917e+01, 5.13677768598433e+01, 5.12157722841997e+01, 5.10373706590689e+01, 5.08308959981726e+01, 5.05950365003356e+01, 5.03289055052507e+01, 5.00320912165518e+01, 4.97046925984186e+01, 4.93473394794556e+01, 4.89611956970161e+01, 4.85479450263199e+01, 4.81097605784798e+01, 4.76492592230742e+01, 4.71694433008024e+01, 4.66736323723209e+01, 4.61653879782317e+01, 4.56484343920593e+01, 4.51265782014317e+01, 4.46036293301154e+01, 4.40833258689935e+01, 4.35692648294061e+01, 4.30648406425112e+01, 4.25731928658138e+01, 4.20971641014993e+01, 4.16392685941944e+01, 4.12016714069339e+01, 4.07861775420635e+01, 4.03942299453893e+01, 4.00269150533302e+01, 3.96849744302261e+01, 3.93688210826350e+01, 3.90785591941822e+01, 3.88140062526474e+01, 3.85747167951095e+01, 3.83600072394347e+01, 3.81689814749120e+01, 3.80005570368741e+01, 3.78534917850672e+01, 3.77264110459444e+01, 3.76178351720257e+01, 3.75262074264420e+01, 3.74499220281311e+01, 3.73873521034460e+01, 3.73368771935469e+01, 3.72969098739103e+01, 3.72659209621208e+01, 3.72424627314847e+01, 3.72251895183432e+01, 3.72128751158239e+01, 3.72044263893260e+01, 3.71988926296077e+01, 3.71954702751711e+01, 3.71935027810148e+01, 3.71924755774959e+01, 3.71920062410352e+01, 3.71918301769376e+01, 3.71917822832542e+01, 3.71917752142475e+01, 3.71917746790037e+01, 3.71917651972130e+01, 3.71917009567056e+01, 3.71914648021050e+01, 3.71908353029477e+01, 3.71894576711279e+01, 3.71868193861735e+01, 3.71822315266386e+01, 3.71748167020871e+01, 3.71635043583170e+01, 3.71470340826158e+01, 3.71239673578774e+01, 3.70927079958070e+01, 3.70515312131688e+01, 3.69986209977912e+01, 3.69321150455682e+01, 3.68501561465897e+01, 3.67509484775289e+01, 3.66328168474238e+01, 3.64942665818945e+01, 3.63340414585761e+01, 3.61511769668490e+01, 3.59450461956130e+01, 3.57153958804067e+01, 3.54623705744586e+01, 3.51865235330214e+01, 3.48888136761163e+01, 3.45705888553384e+01, 3.42335565087196e+01, 3.38797435467764e+01, 3.35114478807907e+01, 3.31311843119707e+01, 3.27416275175045e+01, 3.23455546161477e+01, 3.19457893402741e+01, 3.15451492873840e+01, 3.11463971875612e+01, 3.07521967021506e+01, 3.03650730172484e+01, 2.99873784107714e+01, 2.96212629990228e+01, 2.92686509232091e+01, 2.89312222366545e+01, 2.86104006513023e+01, 2.83073470985504e+01, 2.80229587999100e+01, 2.77578732947453e+01, 2.75124766964409e+01, 2.72869153781497e+01, 2.70811103239459e+01, 2.68947734931251e+01, 2.67274256948351e+01, 2.65784156196110e+01, 2.64469397969217e+01, 2.63320633298986e+01, 2.62327412974266e+01, 2.61478407143347e+01, 2.60761629106348e+01, 2.60164661397699e+01, 2.59674881625626e+01, 2.59279684862705e+01, 2.58966698744596e+01, 2.58723986902638e+01, 2.58540235992382e+01, 2.58404921436803e+01, 2.58308447118522e+01, 2.58242254650946e+01, 2.58198898533888e+01, 2.58172084432710e+01, 2.58156668967675e+01, 2.58148620699998e+01, 2.58144943378726e+01, 2.58143563888351e+01, 2.58143188633502e+01, 2.58143133246671e+01, 2.58143129088764e+01, 2.58143055904891e+01, 2.58142560073450e+01, 2.58140737345619e+01, 2.58135878630098e+01, 2.58125245488887e+01, 2.58104881900325e+01, 2.58069469914823e+01, 2.58012235981968e+01, 2.57924913739062e+01, 2.57797767883524e+01, 2.57619682354460e+01, 2.57378314368596e+01, 2.57060313849929e+01, 2.56651605443947e+01, 2.56137727640140e+01, 2.55504220618911e+01, 2.54737051430772e+01, 2.53823062210079e+01, 2.52750424579778e+01, 2.51509081511085e+01, 2.50091156961103e+01, 2.48491313885741e+01, 2.46707042894257e+01, 2.44738866917140e+01, 2.42590451656998e+01, 2.40268616924276e+01, 2.37783249666345e+01, 2.35147124891738e+01, 2.32375645094279e+01, 2.29486511706075e+01, 2.26499343408815e+01, 2.23435256062924e+01, 2.20316418133722e+01, 2.17165594449265e+01, 2.14005690393548e+01, 2.10859308355910e+01, 2.07748328200339e+01, 2.04693523247685e+01, 2.01714222338672e+01, 1.98828026713779e+01, 1.96050587735161e+01, 1.93395448170870e+01, 1.90873946301550e+01, 1.88495178954604e+01, 1.86266017089622e+01, 1.84191165964761e+01, 1.82273261250936e+01, 1.80512992631164e+01, 1.78909247234986e+01, 1.77459266480386e+01, 1.76158811298992e+01, 1.75002332108501e+01, 1.73983141121038e+01, 1.73093585541076e+01, 1.72325220862199e+01, 1.71668983808763e+01, 1.71115364507167e+01, 1.70654577256014e+01, 1.70276728854696e+01, 1.69971982916435e+01, 1.69730718010718e+01, 1.69543676928002e+01, 1.69402103909151e+01, 1.69297866396063e+01, 1.69223557786704e+01, 1.69172577846746e+01, 1.69139187848997e+01, 1.69118538166038e+01, 1.69106666895025e+01, 1.69100469093136e+01, 1.69097637282577e+01, 1.69096574975310e+01, 1.69096286002225e+01, 1.69096243350412e+01, 1.69096240158264e+01, 1.69096184102540e+01, 1.69095804316919e+01, 1.69094408187202e+01, 1.69090686636384e+01, 1.69082542206923e+01, 1.69066945008417e+01, 1.69039822470694e+01, 1.68995988251076e+01, 1.68929114950205e+01, 1.68831754454861e+01, 1.68695408698906e+01, 1.68510652362257e+01, 1.68267307470080e+01, 1.67954667989852e+01, 1.67561770367139e+01, 1.67077703550861e+01, 1.66491949544650e+01, 1.65794743041286e+01, 1.64977436453442e+01, 1.64032854874993e+01, 1.62955624428685e+01, 1.61742457293793e+01, 1.60392377628215e+01, 1.58906874691437e+01, 1.57289972721870e+01, 1.55548211385804e+01, 1.53690535628303e+01, 1.51728099133310e+01, 1.49673990867434e+01, 1.47542898828903e+01, 1.45350728670847e+01, 1.43114196943541e+01, 1.40850419103338e+01, 1.38576511180537e+01, 1.36309221314834e+01, 1.34064603661703e+01, 1.31857742953474e+01, 1.29702533780937e+01, 1.27611514883179e+01, 1.25595755695845e+01, 1.23664790251117e+01, 1.21826592240120e+01, 1.20087584526607e+01, 1.18452676469149e+01, 1.16925322885797e+01, 1.15507599219011e+01, 1.14200288303775e+01, 1.13002975019692e+01, 1.11914145959136e+01, 1.10931292029074e+01, 1.10051012595925e+01, 1.09269120359141e+01, 1.08580746582946e+01, 1.07980446614369e+01, 1.07462305763471e+01, 1.07020045620186e+01, 1.06647130742075e+01, 1.06336875388287e+01, 1.06082549625423e+01, 1.05877483726611e+01, 1.05715169367162e+01, 1.05589355732754e+01, 1.05494138342653e+01, 1.05424038190447e+01, 1.05374068749343e+01, 1.05339788498316e+01, 1.05317336905234e+01, 1.05303452244623e+01, 1.05295470207317e+01, 1.05291302941015e+01, 1.05289398900096e+01, 1.05288684631439e+01, 1.05288490333312e+01, 1.05288461655325e+01, 1.05288459508203e+01, 1.05288421792930e+01, 1.05288166266556e+01, 1.05287226926823e+01, 1.05284723009252e+01, 1.05279243334558e+01, 1.05268749437910e+01, 1.05250501514787e+01, 1.05221010841904e+01, 1.05176021886233e+01, 1.05110526769440e+01, 1.05018814068088e+01, 1.04894553058589e+01, 1.04730913424144e+01, 1.04520719110569e+01, 1.04256633459641e+01, 1.03931371010366e+01, 1.03537929531365e+01, 1.03069834064806e+01, 1.02521383191720e+01, 1.01887886556855e+01, 1.01165882100795e+01, 1.00353321588751e+01, 9.94497139907133e+00, 9.84562180665988e+00, 9.73756780590170e+00, 9.62125995204084e+00, 9.49730657510756e+00, 9.36645988051177e+00, 9.22959722276083e+00, 9.08769853394477e+00, 8.94182107630638e+00, 8.79307278374222e+00, 8.64258545512288e+00, 8.49148896728158e+00, 8.34088750133795e+00, 8.19183854484004e+00, 8.04533517171767e+00, 7.90229184152902e+00, 7.76353372523860e+00, 7.62978937598655e+00, 7.50168642973094e+00, 7.37974994248422e+00, 7.26440294091218e+00, 7.15596876996055e+00, 7.05467485271841e+00, 6.96065752327695e+00, 6.87396764498295e+00, 6.79457677904959e+00, 6.72238371893867e+00, 6.65722125250274e+00, 6.59886305538149e+00, 6.54703065452959e+00, 6.50140042895521e+00, 6.46161063478115e+00, 6.42726845284534e+00, 6.39795705891066e+00, 6.37324270942111e+00, 6.35268182055728e+00, 6.33582799674464e+00, 6.32223893899973e+00, 6.31148313628917e+00, 6.30314621742524e+00, 6.29683681994018e+00, 6.29219181861681e+00, 6.28888075211532e+00, 6.28660929284163e+00, 6.28512162331926e+00, 6.28420161128719e+00, 6.28367271398734e+00, 6.28339658722953e+00, 6.28327042384286e+00, 6.28322309579656e+00, 6.28321022143891e+00, 6.28320832121136e+00, 6.28320817856383e+00, 6.28320566784774e+00, 6.28318865738741e+00, 6.28312612529295e+00, 6.28295943889464e+00, 6.28259465591299e+00, 6.28189607633067e+00, 6.28068131401177e+00, 6.27871813232412e+00, 6.27572325866115e+00, 6.27136335671287e+00, 6.26525828983790e+00, 6.25698675102699e+00, 6.24609426230538e+00, 6.23210345777947e+00, 6.21452646061263e+00, 6.19287904824556e+00, 6.16669617825912e+00, 6.13554832832271e+00, 6.09905799886673e+00, 6.05691564883496e+00, 6.00889429513315e+00, 5.95486201499715e+00, 5.89479165324052e+00, 5.82876715350540e+00, 5.75698609836184e+00, 5.67975824560590e+00, 5.59750007104009e+00, 5.51072555251521e+00, 5.42003363719660e+00, 5.32609300733739e+00, 5.22962488683766e+00, 5.13138470411822e+00, 5.03214344408738e+00, 4.93266948567485e+00, 4.83371163791278e+00, 4.73598396625648e+00, 4.64015285328821e+00, 4.54682657687816e+00, 4.45654752721042e+00, 4.36978703383225e+00, 4.28694264507187e+00, 4.20833760194778e+00, 4.13422218083741e+00, 4.06477654403878e+00, 4.00011473229422e+00, 3.94028945343496e+00, 3.88529736029224e+00, 3.83508456220875e+00, 3.78955217148977e+00, 3.74856174342386e+00, 3.71194052167129e+00, 3.67948644668574e+00, 3.65097292136122e+00, 3.62615335425642e+00, 3.60476551633278e+00, 3.58653575261022e+00, 3.57118308647946e+00, 3.55842324302517e+00, 3.54797260037287e+00, 3.53955205682084e+00, 3.53289077859330e+00, 3.52772977078602e+00, 3.52382519474607e+00, 3.52095134079480e+00, 3.51890315753685e+00, 3.51749823910619e+00, 3.51657818001114e+00, 3.51600922343849e+00, 3.51568215191986e+00, 3.51551139749719e+00, 3.51543337984933e+00, 3.51540411297205e+00, 3.51539615169409e+00, 3.51539497662695e+00, 3.51539488804057e+00, 3.51539332384151e+00, 3.51538272617417e+00, 3.51534376819878e+00, 3.51523992185234e+00, 3.51501266332319e+00, 3.51457746157368e+00, 3.51382072194484e+00, 3.51259784395776e+00, 3.51073253682956e+00, 3.50801751611350e+00, 3.50421667790248e+00, 3.49906881071745e+00, 3.49229285740883e+00, 3.48359467892476e+00, 3.47267519891540e+00, 3.45923972506124e+00, 3.44300815418336e+00, 3.42372568028631e+00, 3.40117354623904e+00, 3.37517932045327e+00, 3.34562614927900e+00, 3.31246044209806e+00, 3.27569749457470e+00, 3.23542464739543e+00, 3.19180170929612e+00, 3.14505853528709e+00, 3.09548983029879e+00, 3.04344742843857e+00, 2.98933046110022e+00, 2.93357395704669e+00, 2.87663650163456e+00, 2.81898761331000e+00, 2.76109547244579e+00, 2.70341556588278e+00, 2.64638070079634e+00, 2.59039270779624e+00, 2.53581601079908e+00, 2.48297310458727e+00, 2.43214186181410e+00, 2.38355449749720e+00, 2.33739795461055e+00, 2.29381543929230e+00, 2.25290882548495e+00, 2.21474166160016e+00, 2.17934254021607e+00, 2.14670863000619e+00, 2.11680921180564e+00, 2.08958910367883e+00, 2.06497189989965e+00, 2.04286298380762e+00, 2.02315230337881e+00, 2.00571692057457e+00, 1.99042336112993e+00, 1.97712980076140e+00, 1.96568812736413e+00, 1.95594591729032e+00, 1.94774835801083e+00, 1.94094014017899e+00, 1.93536733025284e+00, 1.93087922138677e+00, 1.92733014636275e+00, 1.92458122303299e+00, 1.92250199123436e+00, 1.92097189146619e+00, 1.91988153068394e+00, 1.91913367996015e+00, 1.91864395275674e+00, 1.91834112099244e+00, 1.91816703842211e+00, 1.91807615616369e+00, 1.91803463234392e+00, 1.91801905549296e+00, 1.91801481822762e+00, 1.91801419281671e+00, 1.91801414541142e+00, 1.91801330495734e+00, 1.91800761076543e+00, 1.91798667843354e+00, 1.91793088144374e+00, 1.91780877577321e+00, 1.91757494672709e+00, 1.91716837212751e+00, 1.91651139001077e+00, 1.91550935159721e+00, 1.91405103029332e+00, 1.91200984330430e+00, 1.90924592242244e+00, 1.90560904327178e+00, 1.90094238688916e+00, 1.89508706413187e+00, 1.88788728346130e+00, 1.87919598911232e+00, 1.86888074398999e+00, 1.85682958556130e+00, 1.84295654992498e+00, 1.82720654537226e+00, 1.80955926710998e+00, 1.79003188216301e+00, 1.76868027745402e+00, 1.74559875077497e+00, 1.72091812650716e+00, 1.69480238557027e+00, 1.66744400093353e+00, 1.63905825520273e+00, 1.60987687642651e+00, 1.58014135683324e+00, 1.55009631531544e+00, 1.51998323084212e+00, 1.49003481674220e+00, 1.46047023331178e+00, 1.43149125756707e+00, 1.40327945269956e+00, 1.37599431272172e+00, 1.34977230448789e+00, 1.32472669188851e+00, 1.30094800555169e+00, 1.27850501420599e+00, 1.25744605829357e+00, 1.23780061935335e+00, 1.21958101703861e+00, 1.20278414665169e+00, 1.18739319157127e+00, 1.17337926528717e+00, 1.16070295586968e+00, 1.14931576096110e+00, 1.13916141351520e+00, 1.13017710750555e+00, 1.12229463880118e+00, 1.11544147959778e+00, 1.10954180546734e+00, 1.10451749255215e+00, 1.10028909901403e+00, 1.09677683991872e+00, 1.09390155869561e+00, 1.09158569162086e+00, 1.08975421493192e+00, 1.08833555773184e+00, 1.08726245832905e+00, 1.08647273759117e+00, 1.08590996070055e+00, 1.08552395868028e+00, 1.08527118333357e+00, 1.08511487373353e+00, 1.08502501883285e+00, 1.08497810868070e+00, 1.08495667554793e+00, 1.08494863532127e+00, 1.08494644819235e+00, 1.08494612537699e+00, 1.08494610075277e+00, 1.08494566213851e+00, 1.08494269046655e+00, 1.08493176634677e+00, 1.08490264712111e+00, 1.08483892277660e+00, 1.08471689210930e+00, 1.08450470815218e+00, 1.08416183869184e+00, 1.08363888376761e+00, 1.08287778674661e+00, 1.08181246809428e+00, 1.08036990045537e+00, 1.07847162937686e+00, 1.07603572551227e+00, 1.07297913152171e+00, 1.06922034090194e+00, 1.06468231821924e+00, 1.05929554305726e+00, 1.05300103646226e+00, 1.04575321213375e+00, 1.03752238829718e+00, 1.02829680263223e+00, 1.01808399308769e+00, 1.00691144146974e+00, 9.94826422049144e-01, 9.81895050058768e-01, 9.68200579555429e-01, 9.53841050898944e-01, 9.38926429637576e-01, 9.23575406681849e-01, 9.07912041928867e-01, 8.92062429703657e-01, 8.76151546258143e-01, 8.60300410409764e-01, 8.44623652437377e-01, 8.29227548009641e-01, 8.14208537210391e-01, 7.99652216772799e-01, 7.85632768467224e-01, 7.72212769089776e-01, 7.59443317572320e-01, 7.47364411545547e-01, 7.36005507933939e-01, 7.25386208360732e-01, 7.15517018830525e-01, 7.06400143043808e-01, 6.98030278746842e-01, 6.90395395971507e-01, 6.83477484371494e-01, 6.77253263835851e-01, 6.71694858037426e-01, 6.66770434554934e-01, 6.62444817761937e-01, 6.58680081914639e-01, 6.55436131928823e-01, 6.52671278360202e-01, 6.50342811251605e-01, 6.48407574964637e-01, 6.46822543071650e-01, 6.45545389070676e-01, 6.44535045357756e-01, 6.43752239807910e-01, 6.43159996756339e-01, 6.42724087381339e-01, 6.42413413682036e-01, 6.42200310557598e-01, 6.42060751983848e-01, 6.41974449900431e-01, 6.41924838016763e-01, 6.41898937083482e-01, 6.41887102948033e-01, 6.41882663589996e-01, 6.41881455980254e-01, 6.41881277739714e-01, 6.41881264047655e-01, 6.41881018902680e-01, 6.41879358011224e-01, 6.41873252426620e-01, 6.41856977395648e-01, 6.41821360983883e-01, 6.41753155491756e-01, 6.41634558700858e-01, 6.41442910005210e-01, 6.41150584069237e-01, 6.40725101504984e-01, 6.40129471776447e-01, 6.39322777693181e-01, 6.38261003084239e-01, 6.36898095334658e-01, 6.35187242480811e-01, 6.33082330902951e-01, 6.30539535145516e-01, 6.27518977265652e-01, 6.23986380916790e-01, 6.19914636842381e-01, 6.15285193229212e-01, 6.10089187735980e-01, 6.04328248595323e-01, 5.98014909750262e-01, 5.91172608325323e-01, 5.83835259766289e-01, 5.76046433995046e-01, 5.67858181970323e-01, 5.59329583376936e-01, 5.50525100699033e-01, 5.41512831507987e-01, 5.32362749340535e-01, 5.23145014973352e-01, 5.13928425897433e-01, 5.04779054468429e-01, 4.95759106718054e-01, 4.86926016091417e-01, 4.78331770890875e-01, 4.70022461888121e-01, 4.62038027794904e-01, 4.54412170993322e-01, 4.47172413721049e-01, 4.40340265201243e-01, 4.33931472345520e-01, 4.27956330008456e-01, 4.22420030779193e-01, 4.17323038513871e-01, 4.12661473901352e-01, 4.08427504079455e-01, 4.04609731524204e-01, 4.01193580036391e-01, 3.98161677603269e-01, 3.95494237222252e-01, 3.93169437447185e-01, 3.91163804501389e-01, 3.89452597345354e-01, 3.88010196159655e-01, 3.86810493397022e-01, 3.85827284979399e-01, 3.85034657492341e-01, 3.84407365509375e-01, 3.83921191613908e-01, 3.83553280432878e-01, 3.83282437196570e-01, 3.83089381106856e-01, 3.82956944208445e-01, 3.82870207535660e-01, 3.82816568020331e-01, 3.82785731901492e-01, 3.82769633038620e-01, 3.82762277419715e-01, 3.82759518086800e-01, 3.82758767482599e-01, 3.82758656695060e-01, 3.82758648119633e-01, 3.82758493738450e-01, 3.82757447784130e-01, 3.82753602758186e-01, 3.82743353429547e-01, 3.82720923525937e-01, 3.82677969498289e-01, 3.82603278094521e-01, 3.82482573062191e-01, 3.82298444605524e-01, 3.82030413090206e-01, 3.81655135715164e-01, 3.81146761234334e-01, 3.80477433126529e-01, 3.79617935778370e-01, 3.78538471306072e-01, 3.77209546825114e-01, 3.75602943737618e-01, 3.73692732603843e-01, 3.71456290247039e-01, 3.68875270829278e-01, 3.65936480623844e-01, 3.62632607789465e-01, 3.58962763979875e-01, 3.54932803996775e-01, 3.50555402309411e-01, 3.45849880049059e-01, 3.40841791630240e-01, 3.35562294896192e-01, 3.30047341173625e-01, 3.24336730689410e-01, 3.18473083766165e-01, 3.12500778924104e-01, 3.06464905842844e-01, 3.00410274842946e-01, 2.94380516140393e-01, 2.88417292677920e-01, 2.82559640841853e-01, 2.76843444645309e-01, 2.71301041566886e-01, 2.65960952496116e-01, 2.60847724211118e-01, 2.55981870390229e-01, 2.51379896091734e-01, 2.47054390656018e-01, 2.43014174815698e-01, 2.39264489205439e-01, 2.35807213234111e-01, 2.32641105228538e-01, 2.29762056733122e-01, 2.27163355719242e-01, 2.24835955133059e-01, 2.22768744619856e-01, 2.20948824352874e-01, 2.19361780634472e-01, 2.17991963303660e-01, 2.16822764976357e-01, 2.15836901788308e-01, 2.15016694645954e-01, 2.14344349093992e-01, 2.13802230864005e-01, 2.13373133086989e-01, 2.13040530147609e-01, 2.12788812342168e-01, 2.12603494983853e-01, 2.12471395458920e-01, 2.12380772028936e-01, 2.12321418917711e-01, 2.12284713389053e-01, 2.12263612055580e-01, 2.12252595461081e-01, 2.12247561934447e-01, 2.12245673692365e-01, 2.12245160045427e-01, 2.12245084232237e-01, 2.12245078316414e-01, 2.12244971200519e-01, 2.12244245475063e-01, 2.12241577639531e-01, 2.12234466230726e-01, 2.12218903397314e-01, 2.12189099900399e-01, 2.12137275095799e-01, 2.12053522437969e-01, 2.11925759839716e-01, 2.11739772718914e-01, 2.11479355638332e-01, 2.11126555955587e-01, 2.10662019730762e-01, 2.10065436231744e-01, 2.09316072757176e-01, 2.08393386314995e-01, 2.07277693232953e-01, 2.05950872455062e-01, 2.04397073616647e-01, 2.02603397576027e-01, 2.00560515469178e-01, 1.98263192998539e-01, 1.95710689818853e-01, 1.92907009523168e-01, 1.89860983538154e-01, 1.86586181606842e-01, 1.83100651645489e-01, 1.79426501694635e-01, 1.75589345548289e-01, 1.71617640693640e-01, 1.67541951899551e-01, 1.63394175903420e-01, 1.59206762181938e-01, 1.55011962049755e-01, 1.50841133810346e-01, 1.46724126004614e-01, 1.42688754564361e-01, 1.38760383365522e-01, 1.34961611643476e-01, 1.31312066258104e-01, 1.27828292139881e-01, 1.24523730660144e-01, 1.21408773296551e-01, 1.18490876790148e-01, 1.15774725844176e-01, 1.13262430049289e-01, 1.10953742934839e-01, 1.08846292696413e-01, 1.06935816116462e-01, 1.05216389330295e-01, 1.03680651209160e-01, 1.02320017054211e-01, 1.01124881888950e-01, 1.00084813830705e-01, 9.91887387954620e-02, 9.84251181526474e-02, 9.77821209210816e-02, 9.72477917256356e-02, 9.68102150774115e-02, 9.64576756816768e-02, 9.61788135167012e-02, 9.59627714674411e-02, 9.57993324496300e-02, 9.56790423116445e-02, 9.55933144362575e-02, 9.55345119290740e-02, 9.54960035972903e-02, 9.54721905809635e-02, 9.54585014526828e-02, 9.54513547817015e-02, 9.54480894743313e-02, 9.54468645554556e-02, 9.54465313488129e-02, 9.54464821682564e-02, 9.54464782979727e-02, 9.54464078014610e-02, 9.54459301779025e-02, 9.54441743927674e-02, 9.54394941876870e-02, 9.54292520463298e-02, 9.54096385494410e-02, 9.53755349861041e-02, 9.53204265697138e-02, 9.52363729741137e-02, 9.51140420605299e-02, 9.49428114882522e-02, 9.47109412621074e-02, 9.44058180618725e-02, 9.40142693509642e-02, 9.35229417923580e-02, 9.29187345095244e-02, 9.21892734402263e-02, 9.13234087677210e-02, 9.03117135990499e-02, 8.91469591642675e-02, 8.78245403071531e-02, 8.63428253250902e-02, 8.47034065516681e-02, 8.29112325020844e-02, 8.09746087203820e-02, 7.89050622186096e-02, 7.67170729393740e-02, 7.44276842452223e-02, 7.20560122913412e-02, 6.96226805753213e-02, 6.71492103878819e-02, 6.46573998418063e-02, 6.21687233833228e-02, 5.97037803085669e-02, 5.72818153758305e-02, 5.49203280159674e-02, 5.26347797851514e-02, 5.04384031438292e-02, 4.83421085668425e-02, 4.63544815055315e-02, 4.44818561897620e-02, 4.27284502968206e-02, 4.10965435736586e-02, 3.95866844335018e-02, 3.81979106131070e-02, 3.69279723094138e-02, 3.57735483699457e-02, 3.47304481609427e-02, 3.37937939364394e-02, 3.29581809094890e-02, 3.22178145096359e-02, 3.15666261255529e-02, 3.09983697987803e-02, 3.05067029155452e-02, 3.00852541340362e-02, 2.97276817403684e-02, 2.94277254076292e-02, 2.91792539403701e-02, 2.89763110160806e-02, 2.88131602258517e-02, 2.86843299322470e-02, 2.85846576811082e-02, 2.85093331924458e-02, 2.84539383732465e-02, 2.84144823825585e-02, 2.83874295721773e-02, 2.83697181364460e-02, 2.83587675293973e-02, 2.83524731304965e-02, 2.83491872097983e-02, 2.83476859191722e-02, 2.83471227451599e-02, 2.83469695493181e-02, 2.83469469379725e-02, 2.83469451428005e-02, 2.83469122441744e-02, 2.83466893517215e-02, 2.83458699852872e-02, 2.83436859302242e-02, 2.83389065595685e-02, 2.83297549509598e-02, 2.83138447886842e-02, 2.82881419830829e-02, 2.82489548343960e-02, 2.81919563361441e-02, 2.81122417449780e-02, 2.80044237263724e-02, 2.78627661173516e-02, 2.76813555601905e-02, 2.74543079525245e-02, 2.71760039047138e-02, 2.68413443730395e-02, 2.64460146322421e-02, 2.59867421277988e-02, 2.54615319248330e-02, 2.48698628315225e-02, 2.42128281054903e-02, 2.34932070620175e-02, 2.27154578194018e-02, 2.18856266106826e-02, 2.10111752306119e-02, 2.01007348267126e-02, 1.91638007365782e-02, 1.82103884861533e-02, 1.72506743050604e-02, 1.62946437669973e-02, 1.53517694508309e-02, 1.44307339551943e-02, 1.35392098300287e-02, 1.26837039983627e-02, 1.18694705716251e-02, 1.11004913389934e-02, 1.03795171505112e-02, 9.70815729175456e-03, 9.08700057618717e-03, 8.51575322859333e-03, 7.99338336865674e-03, 7.51826627612218e-03, 7.08832617556254e-03, 6.70116938727006e-03, 6.35420235505535e-03, 6.04472887197005e-03, 5.77002473400512e-03, 5.52739251483558e-03, 5.31420103301627e-03, 5.12791333244422e-03, 4.96610554208132e-03, 4.82647803524117e-03, 4.70686005795073e-03, 4.60520922021601e-03, 4.51960754631386e-03, 4.44825584668251e-03, 4.38946793838113e-03, 4.34166580701856e-03, 4.30337632052158e-03, 4.27322970143962e-03, 4.24995968043618e-03, 4.23240504782996e-03, 4.21951215469107e-03, 4.21033777493562e-03, 4.20405164939552e-03, 4.19993800268514e-03, 4.19739536239661e-03, 4.19593411126584e-03, 4.19517135570402e-03, 4.19482288049267e-03, 4.19469216092239e-03, 4.19465660253593e-03, 4.19465135421149e-03, 4.19465093371230e-03, 4.19464317897678e-03, 4.19459063975580e-03, 4.19439750452000e-03, 4.19388271116361e-03, 4.19275627243110e-03, 4.19059967377677e-03, 4.18685141472963e-03, 4.18079882244626e-03, 4.17157731642703e-03, 4.15817830244691e-03, 4.13946677828392e-03, 4.11420948807972e-03, 4.08111402741440e-03, 4.03887865458330e-03, 3.98625171604294e-03, 3.92209859252546e-03, 3.84547299367455e-03, 3.75568837367268e-03, 3.65238429016301e-03, 3.53558177529578e-03, 3.40572144406792e-03, 3.26367859209142e-03, 3.11075129103244e-03, 2.94861994119344e-03, 2.77927935678690e-03, 2.60494948790324e-03, 2.42797928055438e-03, 2.25075746831749e-03, 2.07562343418033e-03, 1.90476160992148e-03, 1.74010326747975e-03, 1.58327978849115e-03, 1.43556920991728e-03, 1.29780628900278e-03, 1.17047409095505e-03, 1.05389367418240e-03, 9.48081908646259e-04, 8.52604786501843e-04, 7.67215484643202e-04, 6.91614343921287e-04, 6.24492060566695e-04, 5.65018305849208e-04, 5.13186064821980e-04, 4.68164084551859e-04, 4.28615474930893e-04, 3.93841224507696e-04, 3.64268934433198e-04, 3.39120926284409e-04, 3.17208235486532e-04, 2.98170328107972e-04, 2.81719883446385e-04, 2.67698499783364e-04, 2.56095658670574e-04, 2.46427144176664e-04, 2.38302256974521e-04, 2.31514139954925e-04, 2.25893448416059e-04, 2.21287598634490e-04, 2.17557587035344e-04, 2.14577134983074e-04, 2.12232397501566e-04, 2.10421428135987e-04, 2.09053162848249e-04, 2.08046287454522e-04, 2.07328415768870e-04, 2.06835719836818e-04, 2.06512885384313e-04, 2.06313163630512e-04, 2.06198318977613e-04, 2.06138352069729e-04, 2.06110950898619e-04, 2.06100671462650e-04, 2.06097875164346e-04, 2.06097462438493e-04, 2.06097428986141e-04, 2.06096807232825e-04, 2.06092594794427e-04, 2.06077110462388e-04, 2.06035842523988e-04, 2.05945567174126e-04, 2.05772825296435e-04, 2.05472875762780e-04, 2.04989252907239e-04, 2.04254042657226e-04, 2.03188939963184e-04, 2.01707080236702e-04, 1.99715646417076e-04, 1.97119456998962e-04, 1.93826024414534e-04, 1.89752636487128e-04, 1.84837959665174e-04, 1.79076349809293e-04, 1.72568922977234e-04, 1.65344986555738e-04, 1.57326583433814e-04, 1.48489388692497e-04, 1.38852674183454e-04, 1.28477578996922e-04, 1.17606377557422e-04, 1.07573793485471e-04, 9.74952652796901e-05, 8.72118235974995e-05, 7.68284873089442e-05, 6.80697951906950e-05, 5.99201072475353e-05, 5.18060181917690e-05, 4.41886064805885e-05, 3.87581556822170e-05, 3.31801739719140e-05, 2.80583619911567e-05, 2.55161993294450e-05, 2.31335912701525e-05, 2.05779679447460e-05, 1.77426080654186e-05, 1.47273490966882e-05, 1.17612683823154e-05, 9.10541638177165e-06, 7.07691430079649e-06, 5.51664945191987e-06, 4.30223408898200e-06, 3.45262490952791e-06, 3.12214447717452e-06, 3.23379295252835e-06, 3.14410465307446e-06, 2.71202743817867e-06, 2.10932942934505e-06, 1.59688458722940e-06, 1.33547716529847e-06, 1.19975865686683e-06, 1.05253604942845e-06, 8.91146640924489e-07, 7.51525277564724e-07, 6.54641667097167e-07, 5.97871391505578e-07, 5.59101039243615e-07, 5.16793420501926e-07, 4.61674403595680e-07, 3.94695916838828e-07, 0.00000000000000e+00, 0.00000000000000e+00, -1.42108547152020e-14, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 1.42108547152020e-14, 0.00000000000000e+00, 1.42108547152020e-14, 0.00000000000000e+00, 0.00000000000000e+00],
[ 1.12000000000000e+02, 1.11999999853469e+02, 1.11999995322428e+02, 1.11999964623909e+02, 1.11999851772081e+02, 1.11999550944228e+02, 1.11998892557787e+02, 1.11997631546227e+02, 1.11995438252485e+02, 1.11991892276019e+02, 1.11986479516535e+02, 1.11978592558730e+02, 1.11967534443278e+02, 1.11952525775762e+02, 1.11932715043085e+02, 1.11907191940652e+02, 1.11875003465709e+02, 1.11835172501155e+02, 1.11786718594299e+02, 1.11728680616140e+02, 1.11660140955520e+02, 1.11580250845137e+02, 1.11488256322805e+02, 1.11383524199342e+02, 1.11265567243641e+02, 1.11134067628641e+02, 1.10988897542877e+02, 1.10830135800572e+02, 1.10658079314747e+02, 1.10473248455899e+02, 1.10276385605558e+02, 1.10068446607275e+02, 1.09850585272951e+02, 1.09624131561135e+02, 1.09390564445199e+02, 1.09151480782181e+02, 1.08908561646941e+02, 1.08663537605027e+02, 1.08418154278984e+02, 1.08174139351779e+02, 1.07933171890763e+02, 1.07696854608328e+02, 1.07466689434613e+02, 1.07244056584224e+02, 1.07030197160671e+02, 1.06826199255620e+02, 1.06632987454750e+02, 1.06451315644079e+02, 1.06281763006898e+02, 1.06124733100905e+02, 1.05980455900642e+02, 1.05848992678011e+02, 1.05730243572477e+02, 1.05623957673577e+02, 1.05529745403307e+02, 1.05447092947235e+02, 1.05375378443033e+02, 1.05313889595583e+02, 1.05261842350960e+02, 1.05218400229168e+02, 1.05182693889472e+02, 1.05153840484243e+02, 1.05130962349346e+02, 1.05113204582768e+02, 1.05099751079840e+02, 1.05089838623876e+02, 1.05082768675668e+02, 1.05077916563478e+02, 1.05074737845857e+02, 1.05072771700531e+02, 1.05071641281196e+02, 1.05071051076926e+02, 1.05070781402353e+02, 1.05070680237242e+02, 1.05070652717797e+02, 1.05070648655982e+02, 1.05070648123158e+02, 1.05070635708986e+02, 1.05070551601139e+02, 1.05070242411647e+02, 1.05069418224840e+02, 1.05067614506387e+02, 1.05064160147583e+02, 1.05058152919125e+02, 1.05048443453814e+02, 1.05033628697794e+02, 1.05012055560571e+02, 1.04981835248973e+02, 1.04940868482460e+02, 1.04886881451692e+02, 1.04817471997541e+02, 1.04730165059047e+02, 1.04622475980320e+02, 1.04491979803104e+02, 1.04336384240175e+02, 1.04153603671450e+02, 1.03941831280184e+02, 1.03699606396141e+02, 1.03425874260260e+02, 1.03120035759414e+02, 1.02781985142518e+02, 1.02412134225345e+02, 1.02011422022258e+02, 1.01581309066020e+02, 1.01123755965792e+02, 1.00641186213565e+02, 1.00136434130640e+02, 9.96126802449913e+01, 9.90733780367254e+01, 9.85221772177227e+01, 9.79628487822645e+01, 9.73992156995177e+01, 9.68350908251025e+01, 9.62742214347777e+01, 9.57202386392805e+01, 9.51766100637993e+01, 9.46465951355745e+01, 9.41332034312892e+01, 9.36391572759645e+01, 9.31668599828059e+01, 9.27183708776327e+01, 9.22953877750356e+01, 9.18992370649247e+01, 9.15308711534073e+01, 9.11908727297035e+01, 9.08794651950955e+01, 9.05965285566981e+01, 9.03416201169447e+01, 9.01139993428537e+01, 8.99126563511602e+01, 8.97363434804930e+01, 8.95836094329011e+01, 8.94528354534227e+01, 8.93422729814896e+01, 8.92500821577780e+01, 8.91743705121407e+01, 8.91132311007482e+01, 8.90647793119578e+01, 8.90271875288502e+01, 8.89987168290788e+01, 8.89777449255298e+01, 8.89627896081971e+01, 8.89525270401797e+01, 8.89458043876876e+01, 8.89416464215959e+01, 8.89392559101357e+01, 8.89380078204925e+01, 8.89374375518093e+01, 8.89372236233325e+01, 8.89371654295049e+01, 8.89371568402144e+01, 8.89371560819435e+01, 8.89371412239996e+01, 8.89370405592357e+01, 8.89366705047967e+01, 8.89356840746174e+01, 8.89335252916774e+01, 8.89293909463828e+01, 8.89222012297995e+01, 8.89105805874797e+01, 8.88928499271624e+01, 8.88670310664873e+01, 8.88308640176342e+01, 8.87818373667621e+01, 8.87172316113926e+01, 8.86341748670663e+01, 8.85297098511685e+01, 8.84008705112183e+01, 8.82447661116836e+01, 8.80586700621782e+01, 8.78401103038821e+01, 8.75869577188529e+01, 8.72975088379872e+01, 8.69705591420766e+01, 8.66054635088111e+01, 8.62021808688396e+01, 8.57613008804122e+01, 8.52840513651640e+01, 8.47722862828730e+01, 8.42284550491339e+01, 8.36555548978931e+01, 8.30570686635187e+01, 8.24368907599778e+01, 8.17992442938443e+01, 8.11485922510984e+01, 8.04895456541486e+01, 7.98267715670662e+01, 7.91649038220223e+01, 7.85084592523472e+01, 7.78617619229499e+01, 7.72288772795041e+01, 7.66135573439567e+01, 7.60191972246642e+01, 7.54488024778575e+01, 7.49049663886531e+01, 7.43898560619421e+01, 7.39052062583353e+01, 7.34523200683274e+01, 7.30320756922208e+01, 7.26449387289416e+01, 7.22909794598157e+01, 7.19698946548980e+01, 7.16810334478880e+01, 7.14234268348624e+01, 7.11958203575317e+01, 7.09967095320648e+01, 7.08243775748791e+01, 7.06769349523443e+01, 7.05523602395821e+01, 7.04485417153845e+01, 7.03633190501090e+01, 7.02945243685799e+01, 7.02400218998707e+01, 7.01977453706136e+01, 7.01657322682106e+01, 7.01421541037204e+01, 7.01253418476694e+01, 7.01138057989552e+01, 7.01062492771926e+01, 7.01015756984866e+01, 7.00988887964706e+01, 7.00974859745444e+01, 7.00968450097547e+01, 7.00966045611081e+01, 7.00965391531872e+01, 7.00965294991129e+01, 7.00965287247368e+01, 7.00965144335688e+01, 7.00964176088035e+01, 7.00960616707290e+01, 7.00951128705454e+01, 7.00930364447539e+01, 7.00890598417369e+01, 7.00821444981608e+01, 7.00709674709891e+01, 7.00539140310253e+01, 7.00290820902351e+01, 6.99942990596120e+01, 6.99471514086762e+01, 6.98850268154814e+01, 6.98051683555198e+01, 6.97047396846683e+01, 6.95808996400422e+01, 6.94308841384292e+01, 6.92520927298488e+01, 6.90421767064806e+01, 6.87991253210830e+01, 6.85213464785538e+01, 6.82077382658083e+01, 6.78577479013171e+01, 6.74714151217198e+01, 6.70493976657717e+01, 6.65929773354722e+01, 6.61040460663875e+01, 6.55850724678028e+01, 6.50390503309148e+01, 6.44694315709028e+01, 6.38800468785259e+01, 6.32750179207669e+01, 6.26586651760564e+01, 6.20354153829474e+01, 6.14097121414368e+01, 6.07859325122415e+01, 6.01683116353321e+01, 5.95608765742097e+01, 5.89673899047061e+01, 5.83913030736737e+01, 5.78357192592285e+01, 5.73033653245485e+01, 5.67965724013363e+01, 5.63172646022655e+01, 5.58669553064537e+01, 5.54467503836639e+01, 5.50573576411593e+01, 5.46991017195500e+01, 5.43719436510964e+01, 5.40755043311893e+01, 5.38090912315205e+01, 5.35717277827025e+01, 5.33621849532984e+01, 5.31790146330933e+01, 5.30205844788528e+01, 5.28851138954392e+01, 5.27707108045805e+01, 5.26754088029794e+01, 5.25972042391620e+01, 5.25340926549026e+01, 5.24841039536936e+01, 5.24453355872351e+01, 5.24159830022638e+01, 5.23943665736525e+01, 5.23789542725903e+01, 5.23683793846917e+01, 5.23614527024027e+01, 5.23571687656598e+01, 5.23547059075422e+01, 5.23534200678806e+01, 5.23528325558017e+01, 5.23526121595247e+01, 5.23525522063733e+01, 5.23525433574131e+01, 5.23525426748961e+01, 5.23525304190825e+01, 5.23524473841512e+01, 5.23521421390960e+01, 5.23513284688959e+01, 5.23495477768723e+01, 5.23461375589616e+01, 5.23402072250672e+01, 5.23306223998727e+01, 5.23159986690331e+01, 5.22947055406992e+01, 5.22648811589292e+01, 5.22244580264187e+01, 5.21711996626199e+01, 5.21027477374287e+01, 5.20166787843902e+01, 5.19105691234260e+01, 5.17820661334724e+01, 5.16289635415043e+01, 5.14492779753231e+01, 5.12413237069354e+01, 5.10037823352617e+01, 5.07357641599590e+01, 5.04368582098209e+01, 5.01071683200986e+01, 4.97473332916353e+01, 4.93585299736753e+01, 4.89424590278792e+01, 4.85013140674849e+01, 4.80377357258239e+01, 4.75547529033220e+01, 4.70557139120463e+01, 4.65442104685237e+01, 4.60239975139511e+01, 4.54989117321284e+01, 4.49727914561929e+01, 4.44494004441085e+01, 4.39323577584506e+01, 4.34250756777469e+01, 4.29307071611404e+01, 4.24521038760263e+01, 4.19917852076619e+01, 4.15519180614727e+01, 4.11343067181847e+01, 4.07403915762534e+01, 4.03712553570210e+01, 4.00276352645085e+01, 3.97099396626505e+01, 3.94182680169215e+01, 3.91524330954369e+01, 3.89119846901326e+01, 3.86962343645360e+01, 3.85042809364781e+01, 3.83350365495554e+01, 3.81872532737447e+01, 3.80595502075445e+01, 3.79504410395571e+01, 3.78583619765366e+01, 3.77816998681481e+01, 3.77188202664481e+01, 3.76680950605133e+01, 3.76279292334278e+01, 3.75967862091824e+01, 3.75732111993738e+01, 3.75558519311053e+01, 3.75434761436188e+01, 3.75349852849997e+01, 3.75294239220580e+01, 3.75259844934366e+01, 3.75240071823889e+01, 3.75229748532976e+01, 3.75225031748589e+01, 3.75223262321944e+01, 3.75222780995185e+01, 3.75222709952371e+01, 3.75222704573224e+01, 3.75222609282169e+01, 3.75221963671445e+01, 3.75219590341076e+01, 3.75213263936302e+01, 3.75199418868975e+01, 3.75172904348526e+01, 3.75126796750972e+01, 3.75052278308550e+01, 3.74938589875360e+01, 3.74773064055353e+01, 3.74541243187361e+01, 3.74227084481870e+01, 3.73813251925248e+01, 3.73281491376408e+01, 3.72613081606113e+01, 3.71789349978370e+01, 3.70792237246825e+01, 3.69604891830578e+01, 3.68212270317717e+01, 3.66601718246570e+01, 3.64763503865571e+01, 3.62691277951551e+01, 3.60382435132969e+01, 3.57838356593677e+01, 3.55064520359629e+01, 3.52070473161544e+01, 3.48869666427855e+01, 3.45479167401661e+01, 3.41919263716922e+01, 3.38212985114790e+01, 3.34385568694348e+01, 3.30463893957647e+01, 3.26475911220242e+01, 3.22450082487721e+01, 3.18414848735245e+01, 3.14398132807409e+01, 3.10426883755527e+01, 3.06526666714743e+01, 3.02721302155201e+01, 2.99032558786632e+01, 2.95479904622125e+01, 2.92080319957701e+01, 2.88848174018120e+01, 2.85795164043655e+01, 2.82930312351504e+01, 2.80260014205895e+01, 2.77788127763030e+01, 2.75516097113954e+01, 2.73443100335312e+01, 2.71566216040952e+01, 2.69880603723462e+01, 2.68379694796139e+01, 2.67055392462405e+01, 2.65898279271462e+01, 2.64897831491734e+01, 2.64042639328318e+01, 2.63320631625981e+01, 2.62719303130780e+01, 2.62225941716906e+01, 2.61827852297135e+01, 2.61512573495009e+01, 2.61268082629102e+01, 2.61082984203401e+01, 2.60946676963420e+01, 2.60849494702837e+01, 2.60782816410905e+01, 2.60739142036539e+01, 2.60712131088683e+01, 2.60696602450422e+01, 2.60688495094397e+01, 2.60684790774777e+01, 2.60683401156326e+01, 2.60683023146388e+01, 2.60682967352913e+01, 2.60682963164479e+01, 2.60682889443296e+01, 2.60682389971498e+01, 2.60680553861303e+01, 2.60675659473077e+01, 2.60664948261839e+01, 2.60644435155289e+01, 2.60608763142483e+01, 2.60551108897918e+01, 2.60463145268279e+01, 2.60335065271231e+01, 2.60155670846505e+01, 2.59912527905521e+01, 2.59592187203917e+01, 2.59180468194941e+01, 2.58662800335850e+01, 2.58024613394321e+01, 2.57251765281023e+01, 2.56330993025645e+01, 2.55250369979930e+01, 2.53999750470128e+01, 2.52571182232382e+01, 2.50959267304790e+01, 2.49161453781500e+01, 2.47178243970002e+01, 2.45013308852306e+01, 2.42673503943903e+01, 2.40168787103106e+01, 2.37512043914240e+01, 2.34718830359161e+01, 2.31807045234180e+01, 2.28796546131888e+01, 2.25708723097035e+01, 2.22566043798744e+01, 2.19391583744342e+01, 2.16208554979132e+01, 2.13039846832267e+01, 2.09907592263304e+01, 2.06832772804685e+01, 2.03834873636790e+01, 2.00931597863268e+01, 1.98138645742744e+01, 1.95469560867801e+01, 1.92935641553703e+01, 1.90545912469678e+01, 1.88307149149073e+01, 1.86223946611083e+01, 1.84298822903959e+01, 1.82532348791874e+01, 1.80923295830465e+01, 1.79468796459143e+01, 1.78164511247862e+01, 1.77004799880534e+01, 1.75982893698621e+01, 1.75091068581771e+01, 1.74320817569698e+01, 1.73663022930165e+01, 1.73108127379526e+01, 1.72646303912762e+01, 1.72267623261148e+01, 1.71962217438888e+01, 1.71720437240693e+01, 1.71533000986599e+01, 1.71391131349643e+01, 1.71286676808714e+01, 1.71212214190526e+01, 1.71161128930671e+01, 1.71127670101511e+01, 1.71106977909584e+01, 1.71095082221067e+01, 1.71088871677111e+01, 1.71086034045981e+01, 1.71084969555442e+01, 1.71084679988475e+01, 1.71084637249007e+01, 1.71084634050298e+01, 1.71084577879373e+01, 1.71084197313254e+01, 1.71082798314482e+01, 1.71079069116861e+01, 1.71070907957857e+01, 1.71055278740582e+01, 1.71028100585452e+01, 1.70984176643656e+01, 1.70917166853247e+01, 1.70819608484710e+01, 1.70682987294522e+01, 1.70497860828728e+01, 1.70254033849242e+01, 1.69940783972461e+01, 1.69547133426800e+01, 1.69062160413315e+01, 1.68475341003431e+01, 1.67776909993472e+01, 1.66958226864670e+01, 1.66012131206385e+01, 1.64933270892464e+01, 1.63718386178272e+01, 1.62366533878062e+01, 1.60879237974847e+01, 1.59260556381754e+01, 1.57517057959935e+01, 1.55657709013469e+01, 1.53693673916076e+01, 1.51638039783317e+01, 1.49505479665428e+01, 1.47311872124789e+01, 1.45073896926536e+01, 1.42808626746550e+01, 1.40533133345772e+01, 1.38264123844848e+01, 1.36017618993134e+01, 1.33808681175937e+01, 1.31651195843721e+01, 1.29557706480897e+01, 1.27539300416175e+01, 1.25605540808864e+01, 1.23764438988363e+01, 1.22022460850332e+01, 1.20384561056695e+01, 1.18854239184585e+01, 1.17433612585997e+01, 1.16123501455077e+01, 1.14923522384456e+01, 1.13832187477434e+01, 1.12847006832213e+01, 1.11964592895661e+01, 1.11180765768157e+01, 1.10490659002734e+01, 1.09888825761691e+01, 1.09369345361548e+01, 1.08925930251816e+01, 1.08552033344481e+01, 1.08240955359412e+01, 1.07985951505278e+01, 1.07780336413031e+01, 1.07617585821511e+01, 1.07491433127262e+01, 1.07395958597056e+01, 1.07325668841636e+01, 1.07275564094165e+01, 1.07241190946821e+01, 1.07218678479902e+01, 1.07204756160595e+01, 1.07196752469602e+01, 1.07192573897059e+01, 1.07190664689981e+01, 1.07189948483280e+01, 1.07189753657954e+01, 1.07189724902154e+01, 1.07189722749206e+01, 1.07189684931598e+01, 1.07189428711888e+01, 1.07188486823357e+01, 1.07185976111454e+01, 1.07180481566671e+01, 1.07169959188721e+01, 1.07151661726052e+01, 1.07122090956405e+01, 1.07076979726549e+01, 1.07011306421291e+01, 1.06919343842812e+01, 1.06794743608156e+01, 1.06630656079679e+01, 1.06419884512284e+01, 1.06155070542906e+01, 1.05828906409730e+01, 1.05434367461474e+01, 1.04964956733147e+01, 1.04414951790993e+01, 1.03779642871404e+01, 1.03055550738589e+01, 1.02240612814982e+01, 1.01334327090678e+01, 1.00337845107306e+01, 9.92540078597607e+00, 9.80873215965916e+00, 9.68438739809721e+00, 9.55311946032066e+00, 9.41580671001175e+00, 9.27343028423694e+00, 9.12704880552801e+00, 8.97777171791543e+00, 8.82673251901309e+00, 8.67506305443402e+00, 8.52386985416551e+00, 8.37421324884147e+00, 8.22708973828802e+00, 8.08341782653316e+00, 7.94402731256919e+00, 7.80965185114368e+00, 7.68092447864821e+00, 7.55837573232688e+00, 7.44243396697848e+00, 7.33342748017274e+00, 7.23158808359861e+00, 7.13705579587229e+00, 7.04988437520598e+00, 6.97004745535740e+00, 6.89744509325087e+00, 6.83191058014074e+00, 6.77321740877267e+00, 6.72108632509833e+00, 6.67519242303960e+00, 6.63517226300812e+00, 6.60063100817223e+00, 6.57114957621032e+00, 6.54629179861236e+00, 6.52561156539230e+00, 6.50865991201678e+00, 6.49499197977150e+00, 6.48417375349760e+00, 6.47578845473902e+00, 6.46944244694592e+00, 6.46477049530857e+00, 6.46144021930486e+00, 6.45915558259296e+00, 6.45765928294789e+00, 6.45673393395744e+00, 6.45620196858648e+00, 6.45592424006266e+00, 6.45579734482583e+00, 6.45574974223888e+00, 6.45573679319966e+00, 6.45573488194928e+00, 6.45573473847428e+00, 6.45573221319405e+00, 6.45571510405954e+00, 6.45565220922936e+00, 6.45548455592019e+00, 6.45511765692467e+00, 6.45441502510269e+00, 6.45319321647000e+00, 6.45121864759005e+00, 6.44820640346605e+00, 6.44382121590881e+00, 6.43768074664963e+00, 6.42936125136153e+00, 6.41840562782200e+00, 6.40433376248306e+00, 6.38665498542308e+00, 6.36488232725743e+00, 6.33854814915487e+00, 6.30722059750575e+00, 6.27052022914718e+00, 6.22813607367792e+00, 6.17984035828660e+00, 6.12550112751106e+00, 6.06509205137496e+00, 5.99869883099913e+00, 5.92652177561369e+00, 5.84887432765625e+00, 5.76617753746308e+00, 5.67895071739697e+00, 5.58779871818540e+00, 5.49339645068335e+00, 5.39647141075622e+00, 5.29778504471992e+00, 5.19811381413076e+00, 5.09823078306209e+00, 4.99888846426326e+00, 4.90080353253757e+00, 4.80464385694059e+00, 4.71101813232170e+00, 4.62046821993010e+00, 4.53346414970808e+00, 4.45040160437657e+00, 4.37160160483074e+00, 4.29731205106361e+00, 4.22771074235246e+00, 4.16290950108879e+00, 4.10295905051528e+00, 4.04785434070490e+00, 3.99754007217666e+00, 3.95191622599557e+00, 3.91084346756082e+00, 3.87414834434480e+00, 3.84162824274401e+00, 3.81305610422818e+00, 3.78818492540455e+00, 3.76675208049389e+00, 3.74848350866860e+00, 3.73309780382499e+00, 3.72031023210428e+00, 3.70983668459769e+00, 3.70139755117252e+00, 3.69472147842410e+00, 3.68954895264896e+00, 3.68563562966649e+00, 3.68275531928697e+00, 3.68070252486669e+00, 3.67929443877432e+00, 3.67837230313741e+00, 3.67780206161801e+00, 3.67747425114409e+00, 3.67730311085083e+00, 3.67722491687962e+00, 3.67719558385485e+00, 3.67718760458290e+00, 3.67718642685986e+00, 3.67718633807326e+00, 3.67718477033881e+00, 3.67717414871849e+00, 3.67713510268797e+00, 3.67703102160725e+00, 3.67680324930852e+00, 3.67636706340154e+00, 3.67560861160153e+00, 3.67438296436892e+00, 3.67251342743492e+00, 3.66979223754828e+00, 3.66598273765975e+00, 3.66082309127634e+00, 3.65403154773257e+00, 3.64531320943757e+00, 3.63436817903887e+00, 3.62090088114159e+00, 3.60463026422791e+00, 3.58530050048724e+00, 3.56269172302790e+00, 3.53663028113282e+00, 3.50699796453627e+00, 3.47373965534304e+00, 3.43686891643610e+00, 3.39647111900328e+00, 3.35270384503806e+00, 3.30579446397689e+00, 3.25603496213699e+00, 3.20377428240488e+00, 3.14940859194706e+00, 3.09337002140745e+00, 3.03611449790439e+00, 2.97810931959943e+00, 2.91982109166480e+00, 2.86170456840873e+00, 2.80419283546047e+00, 2.74768913368317e+00, 2.69256048808325e+00, 2.63913317426447e+00, 2.58768994277962e+00, 2.53846883499466e+00, 2.49166336567794e+00, 2.44742381676792e+00, 2.40585938031171e+00, 2.36704090144779e+00, 2.33100399897835e+00, 2.29775237616644e+00, 2.26726117324706e+00, 2.23948025205969e+00, 2.21433733952230e+00, 2.19174098860458e+00, 2.17158334201795e+00, 2.15374270456304e+00, 2.13808594491223e+00, 2.12447075674662e+00, 2.11274781299792e+00, 2.10276284595916e+00, 2.09435868084019e+00, 2.08737724168017e+00, 2.08166153723980e+00, 2.07705762155594e+00, 2.07341651033953e+00, 2.07059602148672e+00, 2.06846249681878e+00, 2.06689235384562e+00, 2.06577341175633e+00, 2.06500593559582e+00, 2.06450334693911e+00, 2.06419255817418e+00, 2.06401390018476e+00, 2.06392062887438e+00, 2.06387801341519e+00, 2.06386202704260e+00, 2.06385767837648e+00, 2.06385703652303e+00, 2.06385698787142e+00, 2.06385612532112e+00, 2.06385028142358e+00, 2.06382879875220e+00, 2.06377153471423e+00, 2.06364621821581e+00, 2.06340623917332e+00, 2.06298896693555e+00, 2.06231468723125e+00, 2.06128623935108e+00, 2.05978942473251e+00, 2.05769424343387e+00, 2.05485699454722e+00, 2.05112324902966e+00, 2.04633166791551e+00, 2.04031859556393e+00, 2.03292330792138e+00, 2.02399374264409e+00, 2.01339248570669e+00, 2.00100274341092e+00, 1.98673399573057e+00, 1.97052701273837e+00, 1.95235792532104e+00, 1.93224107716992e+00, 1.91023044680544e+00, 1.88641951240530e+00, 1.86093953146903e+00, 1.83395631237208e+00, 1.80566565494414e+00, 1.77628772200208e+00, 1.74606066484455e+00, 1.71523385776794e+00, 1.68406109816915e+00, 1.65279410198166e+00, 1.62167657435465e+00, 1.59093906995047e+00, 1.56079478405445e+00, 1.53143634250773e+00, 1.50303359162606e+00, 1.47573233329072e+00, 1.44965390788936e+00, 1.42489549955559e+00, 1.40153102357186e+00, 1.37961245319104e+00, 1.35917145021355e+00, 1.34022117787886e+00, 1.32275819346776e+00, 1.30676433916922e+00, 1.29220857130230e+00, 1.27904868837413e+00, 1.26723293660670e+00, 1.25670148677347e+00, 1.24738778808522e+00, 1.23921981335529e+00, 1.23212121484553e+00, 1.22601241228759e+00, 1.22081163390406e+00, 1.21643592822120e+00, 1.21280215950942e+00, 1.20982799331050e+00, 1.20743287125703e+00, 1.20553896685340e+00, 1.20407210668465e+00, 1.20296263527128e+00, 1.20214619706194e+00, 1.20156440633504e+00, 1.20116537538850e+00, 1.20090407348053e+00, 1.20074249345970e+00, 1.20064960960116e+00, 1.20060111835332e+00, 1.20057896287183e+00, 1.20057065167808e+00, 1.20056839084054e+00, 1.20056805714600e+00, 1.20056803169192e+00, 1.20056757829593e+00, 1.20056450647611e+00, 1.20055321420944e+00, 1.20052311369483e+00, 1.20045724210544e+00, 1.20033110028776e+00, 1.20011177038729e+00, 1.19975736026537e+00, 1.19921681870434e+00, 1.19843016009850e+00, 1.19732912972466e+00, 1.19583832984992e+00, 1.19387681205902e+00, 1.19136012178961e+00, 1.18820275721698e+00, 1.18432097712538e+00, 1.17963586288205e+00, 1.17407651058392e+00, 1.16758320405968e+00, 1.16011040124501e+00, 1.15162935895997e+00, 1.14213022708589e+00, 1.13162346407827e+00, 1.12014046144775e+00, 1.10773331311503e+00, 1.09447372233510e+00, 1.08045109867413e+00, 1.06576995406984e+00, 1.05054675423317e+00, 1.03490641456884e+00, 1.01897864522799e+00, 1.00289434694713e+00, 9.86782239367272e-01, 9.70765869949034e-01, 9.54961109146979e-01, 9.39474191498235e-01, 9.24400317812186e-01, 9.09822794922121e-01, 8.95812659346831e-01, 8.82428711087940e-01, 8.69717873634571e-01, 8.57715794924019e-01, 8.46447609663159e-01, 8.35928793880711e-01, 8.26166055730340e-01, 8.17158220554290e-01, 8.08897081631258e-01, 8.01368199912346e-01, 7.94551645859883e-01, 7.88422684053998e-01, 7.82952406567688e-01, 7.78108324436403e-01, 7.73854928133119e-01, 7.70154228101234e-01, 7.66966285369122e-01, 7.64249740327571e-01, 7.61962345113432e-01, 7.60061501913896e-01, 7.58504806084431e-01, 7.57250589455822e-01, 7.56258455812016e-01, 7.55489797468329e-01, 7.54908279403082e-01, 7.54480275709540e-01, 7.54175242420018e-01, 7.53966011138331e-01, 7.53828989442539e-01, 7.53744256642918e-01, 7.53695547053638e-01, 7.53670117231763e-01, 7.53658498358362e-01, 7.53654139753763e-01, 7.53652954111047e-01, 7.53652779112784e-01, 7.53652765669798e-01, 7.53652524984119e-01, 7.53650894305238e-01, 7.53644899785954e-01, 7.53628920819594e-01, 7.53593952364724e-01, 7.53526987889927e-01, 7.53410549557429e-01, 7.53222390354608e-01, 7.52935390547208e-01, 7.52517667753963e-01, 7.51932915713439e-01, 7.51140980990499e-01, 7.50098679070021e-01, 7.48760841286710e-01, 7.47081571902868e-01, 7.45015680797863e-01, 7.42520242543890e-01, 7.39556218417519e-01, 7.36090065792069e-01, 7.32095251200192e-01, 7.27553580879530e-01, 7.22456267126162e-01, 7.16804660856539e-01, 7.10610600020175e-01, 7.03896348450584e-01, 6.96694127977395e-01, 6.89045275087224e-01, 6.80999078878813e-01, 6.72611376627842e-01, 6.63942994915473e-01, 6.55058127106614e-01, 6.46022732368735e-01, 6.36903028900434e-01, 6.27764136855447e-01, 6.18668907194575e-01, 6.09676953835532e-01, 6.00843889979274e-01, 5.92220756655308e-01, 5.83853622893642e-01, 5.75783332386024e-01, 5.68045370456147e-01, 5.60669826754889e-01, 5.53681432396715e-01, 5.47099654390536e-01, 5.40938834500437e-01, 5.35208363606543e-01, 5.29912885936639e-01, 5.25052530083485e-01, 5.20623165507857e-01, 5.16616684328554e-01, 5.13021308714585e-01, 5.09821924235808e-01, 5.07000439185575e-01, 5.04536169237724e-01, 5.02406245900588e-01, 5.00586046126912e-01, 4.99049639179233e-01, 4.97770245487644e-01, 4.96720700836036e-01, 4.95873917857480e-01, 4.95203335604074e-01, 4.94683346989163e-01, 4.94289693283960e-01, 4.93999814690483e-01, 4.93793146378010e-01, 4.93651350311538e-01, 4.93558474717148e-01, 4.93501035081763e-01, 4.93468013080886e-01, 4.93450772645232e-01, 4.93442895353425e-01, 4.93439940310225e-01, 4.93439136466904e-01, 4.93439017821331e-01, 4.93439008637665e-01, 4.93438843306393e-01, 4.93437723162970e-01, 4.93433605401847e-01, 4.93422628997322e-01, 4.93398607580787e-01, 4.93352604369946e-01, 4.93272606512619e-01, 4.93143314371608e-01, 4.92946059117898e-01, 4.92658859415229e-01, 4.92256624824819e-01, 4.91711509683427e-01, 4.90993416485423e-01, 4.90070642177216e-01, 4.88910654313443e-01, 4.87480976952185e-01, 4.85750158878204e-01, 4.83688789826189e-01, 4.81270524541017e-01, 4.78473070539977e-01, 4.75279094051075e-01, 4.71677000319175e-01, 4.67661549512314e-01, 4.63234277624593e-01, 4.58403702440862e-01, 4.53185306808010e-01, 4.47601303934121e-01, 4.41680200960860e-01, 4.35456186554475e-01, 4.28968375007159e-01, 4.22259943045262e-01, 4.15377196349425e-01, 4.08368601202071e-01, 4.01283813381980e-01, 3.94172732158040e-01, 3.87084602620277e-01, 3.80067185073315e-01, 3.73166006033642e-01, 3.66423701584395e-01, 3.59879460393159e-01, 3.53568570510788e-01, 3.47522071077094e-01, 3.41766507264154e-01, 3.36323784246758e-01, 3.31211113805793e-01, 3.26441045456960e-01, 3.22021572841493e-01, 3.17956305561964e-01, 3.14244696669718e-01, 3.10882316533124e-01, 3.07861164703695e-01, 3.05170012486016e-01, 3.02794770042297e-01, 3.00718872855228e-01, 2.98923683112903e-01, 2.97388901975879e-01, 2.96092988703876e-01, 2.95013582274578e-01, 2.94127920479824e-01, 2.93413250638821e-01, 2.92847225153778e-01, 2.92408274292853e-01, 2.92075947962871e-01, 2.91831217955249e-01, 2.91656732310727e-01, 2.91537014108656e-01, 2.91458598158940e-01, 2.91410100720427e-01, 2.91382219415027e-01, 2.91367662833650e-01, 2.91361011805179e-01, 2.91358516773755e-01, 2.91357838064641e-01, 2.91357737888575e-01, 2.91357730071681e-01, 2.91357588533643e-01, 2.91356629592855e-01, 2.91353104428055e-01, 2.91343707655741e-01, 2.91323143165130e-01, 2.91283760148843e-01, 2.91215274141535e-01, 2.91104586277200e-01, 2.90935712110340e-01, 2.90689829134990e-01, 2.90345449483596e-01, 2.89878721090048e-01, 2.89263856777950e-01, 2.88473686255244e-01, 2.87480320913303e-01, 2.86255915784537e-01, 2.84773507272575e-01, 2.83007899691256e-01, 2.80936568684226e-01, 2.78540545734955e-01, 2.75805245697313e-01, 2.72721198985153e-01, 2.69284652022307e-01, 2.65498003860813e-01, 2.61370053395282e-01, 2.56916040008363e-01, 2.52157470262617e-01, 2.47121733777576e-01, 2.41841521992328e-01, 2.36354073411860e-01, 2.30700277506202e-01, 2.24923676108460e-01, 2.19069405494437e-01, 2.13183124034316e-01, 2.07309969283372e-01, 2.01493584726919e-01, 1.95775250431595e-01, 1.90193144093016e-01, 1.84781750075814e-01, 1.79571424761193e-01, 1.74588117601644e-01, 1.69853239394925e-01, 1.65383662945757e-01, 1.61191836801990e-01, 1.57285990235309e-01, 1.53670406996241e-01, 1.50345746352926e-01, 1.47309392170882e-01, 1.44555813896275e-01, 1.42076926865073e-01, 1.39862443007630e-01, 1.37900206443192e-01, 1.36176511429795e-01, 1.34676402490499e-01, 1.33383958188276e-01, 1.32282560949676e-01, 1.31355155567505e-01, 1.30584498632302e-01, 1.29953400256852e-01, 1.29444958222464e-01, 1.29042783241559e-01, 1.28731212576469e-01, 1.28495507940869e-01, 1.28322032588756e-01, 1.28198401885768e-01, 1.28113601539951e-01, 1.28058068074580e-01, 1.28023727040159e-01, 1.28003985815710e-01, 1.27993679541120e-01, 1.27988970615874e-01, 1.27987204151097e-01, 1.27986723631324e-01, 1.27986652707662e-01, 1.27986647126278e-01, 1.27986545462761e-01, 1.27985856678364e-01, 1.27983324648127e-01, 1.27976575292877e-01, 1.27961805041039e-01, 1.27933520337962e-01, 1.27884339562357e-01, 1.27804867971832e-01, 1.27683655398172e-01, 1.27507245212243e-01, 1.27260321401621e-01, 1.26925958272750e-01, 1.26485974146959e-01, 1.25921386369839e-01, 1.25212960013627e-01, 1.24341836954187e-01, 1.23290225846858e-01, 1.22042127344798e-01, 1.20584063270130e-01, 1.18905774010059e-01, 1.17000845833260e-01, 1.14867229694582e-01, 1.12507615849893e-01, 1.09929634420595e-01, 1.07145860790425e-01, 1.04173615913041e-01, 1.01034564436645e-01, 9.77541269322586e-02, 9.43607351985349e-02, 9.08849703507286e-02, 8.73586310818126e-02, 8.38137833085142e-02, 8.02818420077784e-02, 7.67927315197454e-02, 7.33741625286939e-02, 7.00510533035299e-02, 6.68451108142989e-02, 6.37745753157759e-02, 6.08541210688855e-02, 5.80948969453061e-02, 5.55046842785458e-02, 5.30881456674024e-02, 5.08471373820498e-02, 4.87810591895936e-02, 4.68872182830040e-02, 4.51611879665279e-02, 4.35971462473077e-02, 4.21881840181015e-02, 4.09265767222848e-02, 3.98040170382785e-02, 3.88118090886422e-02, 3.79410269316907e-02, 3.71826416525494e-02, 3.65276222992605e-02, 3.59670162796135e-02, 3.54920147401288e-02, 3.50940079787563e-02, 3.47646351791013e-02, 3.44958317805037e-02, 3.42798766889985e-02, 3.41094403672741e-02, 3.39776336952866e-02, 3.38780564455163e-02, 3.38048433444271e-02, 3.37527050552495e-02, 3.37169610710220e-02, 3.36935614695193e-02, 3.36790947554135e-02, 3.36707795662790e-02, 3.36664387914567e-02, 3.36644555705021e-02, 3.36637116144374e-02, 3.36635092421034e-02, 3.36634793724215e-02, 3.36634770009994e-02, 3.36634335417472e-02, 3.36631390998008e-02, 3.36620567151300e-02, 3.36591715900596e-02, 3.36528581487698e-02, 3.36407693923491e-02, 3.36197539488410e-02, 3.35858062097003e-02, 3.35340548813576e-02, 3.34587950350453e-02, 3.33535681458841e-02, 3.32112935038538e-02, 3.30244525955266e-02, 3.27853255017061e-02, 3.24862750188970e-02, 3.21200702157967e-02, 3.16802367401721e-02, 3.11614168202823e-02, 3.05597180938548e-02, 2.98730277654613e-02, 2.91012677339921e-02, 2.82465677247785e-02, 2.73133373547410e-02, 2.63082243828165e-02, 2.52399547136690e-02, 2.41190592415421e-02, 2.29575022886905e-02, 2.17682350403123e-02, 2.05647039050092e-02, 1.93603472888384e-02, 1.81681144054266e-02, 1.70000364720124e-02, 1.58668744612100e-02, 1.47778593970713e-02, 1.37405321631547e-02, 1.27606811175838e-02, 1.18423684658211e-02, 1.09880309354793e-02, 1.01986370519001e-02, 9.47388215301714e-03, 8.81240296968144e-03, 8.21199577509901e-03, 7.66982529609095e-03, 7.18261519854480e-03, 6.74681444237990e-03, 6.35873674784193e-03, 6.01467269630973e-03, 5.71097566418644e-03, 5.44412399081295e-03, 5.21076257406605e-03, 5.00772748949885e-03, 4.83205726501978e-03, 4.68099419366297e-03, 4.55197866284607e-03, 4.44263902727471e-03, 4.35077912646875e-03, 4.27436515161617e-03, 4.21151319250157e-03, 4.16047843091860e-03, 4.11964659065234e-03, 4.08752791557276e-03, 4.06275363748421e-03, 4.04407462265510e-03, 4.03036166157733e-03, 4.02060669826199e-03, 4.01392419328772e-03, 4.00955178447759e-03, 4.00684945273611e-03, 4.00529651338388e-03, 4.00448592264979e-03, 4.00411559897407e-03, 4.00397668457231e-03, 4.00393889711381e-03, 4.00393331980808e-03, 4.00393287294776e-03, 4.00392463281207e-03, 4.00386880492931e-03, 4.00366358118731e-03, 4.00311657155328e-03, 4.00191966521390e-03, 3.99962825404998e-03, 3.99564599717905e-03, 3.98921638986849e-03, 3.97942247720096e-03, 3.96519607481594e-03, 3.94533776761818e-03, 3.91854868583152e-03, 3.88347455262306e-03, 3.83876171639486e-03, 3.78312382699164e-03, 3.71541654176610e-03, 3.63471626637590e-03, 3.54039762883929e-03, 3.43220337391870e-03, 3.31029989048659e-03, 3.17531185055486e-03, 3.02833057327234e-03, 2.87089272208618e-03, 2.70492865053029e-03, 2.53268283212549e-03, 2.35661193025294e-03, 2.17926870227814e-03, 2.00318161623159e-03, 1.83074041704856e-03, 1.66409679856372e-03, 1.50508705021934e-03, 1.35518063160589e-03, 1.21545576848803e-03, 1.08660083080281e-03, 9.68938443534739e-04, 8.62467690410540e-04, 7.66918307107289e-04, 6.81809882664197e-04, 6.06509549939460e-04, 5.40283717626266e-04, 4.82342294276350e-04, 4.31875993101016e-04, 3.88087560722283e-04, 3.50216624909194e-04, 3.17556948132847e-04, 2.89465382380172e-04, 2.65363569852184e-04, 2.44735151028408e-04, 2.27121571995781e-04, 2.12118295749519e-04, 1.99371482253241e-04, 1.88574298363164e-04, 1.79462228672378e-04, 1.71807494254494e-04, 1.65413251480118e-04, 1.60108309387397e-04, 1.55742782226298e-04, 1.52184700255020e-04, 1.49317375957025e-04, 1.47037302014041e-04, 1.45252440205468e-04, 1.43880846209754e-04, 1.42849612828400e-04, 1.42094107744128e-04, 1.41557457766339e-04, 1.41190208168496e-04, 1.40950077692992e-04, 1.40801731191686e-04, 1.40716503508997e-04, 1.40672023476895e-04, 1.40651703901540e-04, 1.40644081952246e-04, 1.40642008645386e-04, 1.40641702614630e-04, 1.40641677873532e-04, 1.40641218834503e-04, 1.40638108689473e-04, 1.40626676056854e-04, 1.40596204843746e-04, 1.40529539791601e-04, 1.40401946552515e-04, 1.40180305535864e-04, 1.39822730673700e-04, 1.39278714655688e-04, 1.38489908991346e-04, 1.37391639057682e-04, 1.35915231055606e-04, 1.33991178941528e-04, 1.31553109966376e-04, 1.28542415055222e-04, 1.24913306777330e-04, 1.20637952733205e-04, 1.15711203150681e-04, 1.10154286488751e-04, 1.04016706799825e-04, 9.73755562085898e-05, 9.03317619673771e-05, 8.30035706798071e-05, 7.55186051293322e-05, 6.80064851508178e-05, 6.05938152915542e-05, 5.34021800433493e-05, 4.65468733921170e-05, 4.01295580161332e-05, 3.42189462543274e-05, 2.88285900893470e-05, 2.39140782554159e-05, 1.93931355170207e-05, 1.51740617866380e-05, 1.12275757828684e-05, 7.81920509496103e-06, 5.89672656303719e-06, 5.61590314873683e-06, 5.25032608322817e-06, 4.20194126604656e-06, 2.74550775714033e-06, 1.34772282933682e-06, 6.08697774850953e-07, 4.42456851601492e-07, 6.65458969706378e-07, 1.12287271747391e-06, 1.47328083244247e-06, 1.65178282429679e-06, 1.60720809105896e-06, 1.35829679948074e-06, 1.01603134794459e-06, 8.06750520609967e-07, 8.27618023890864e-07, 8.54987192155932e-07, 8.10636606729531e-07, 7.19901862566985e-07, 6.17271268765762e-07, 5.25200547940585e-07, 4.52354527169518e-07, 3.95979810718927e-07, -5.68434188608080e-14, -5.68434188608080e-14, -7.10542735760100e-14, -7.10542735760100e-14, -7.10542735760100e-14, -7.10542735760100e-14, -5.68434188608080e-14, -7.10542735760100e-14, -7.10542735760100e-14, -5.68434188608080e-14, -7.10542735760100e-14, -7.10542735760100e-14, -5.68434188608080e-14, -7.10542735760100e-14],
[ 1.13000000000000e+02, 1.12999999851673e+02, 1.12999995265095e+02, 1.12999964190302e+02, 1.12999849955243e+02, 1.12999545440114e+02, 1.12998878983724e+02, 1.12997602515564e+02, 1.12995382337414e+02, 1.12991792894508e+02, 1.12986313782212e+02, 1.12978330134151e+02, 1.12967136437519e+02, 1.12951943726725e+02, 1.12931890023618e+02, 1.12906053825875e+02, 1.12873470397080e+02, 1.12833150581121e+02, 1.12784101843839e+02, 1.12725351225887e+02, 1.12655969858794e+02, 1.12575098637201e+02, 1.12481974543766e+02, 1.12375956987314e+02, 1.12256553349445e+02, 1.12123442763565e+02, 1.11976497008166e+02, 1.11815797323609e+02, 1.11641645995532e+02, 1.11454571711810e+02, 1.11255327995475e+02, 1.11044884419760e+02, 1.10824410777441e+02, 1.10595254844378e+02, 1.10358914784259e+02, 1.10117007536226e+02, 1.09871234678675e+02, 1.09623347265705e+02, 1.09375111006518e+02, 1.09128272939005e+02, 1.08884530481477e+02, 1.08645503473996e+02, 1.08412709577115e+02, 1.08187543201706e+02, 1.07971258006231e+02, 1.07764952913273e+02, 1.07569561554063e+02, 1.07385845033994e+02, 1.07214387909852e+02, 1.07055597269826e+02, 1.06909704803019e+02, 1.06776771732458e+02, 1.06656696463797e+02, 1.06549224771969e+02, 1.06453962312137e+02, 1.06370389201722e+02, 1.06297876379373e+02, 1.06235703406611e+02, 1.06183077340501e+02, 1.06139152272910e+02, 1.06103049105519e+02, 1.06073875111695e+02, 1.06050742828289e+02, 1.06032787824138e+02, 1.06019184908864e+02, 1.06009162376324e+02, 1.06002013922139e+02, 1.05997107933585e+02, 1.05993893921501e+02, 1.05991905945756e+02, 1.05990762975342e+02, 1.05990166218057e+02, 1.05989893549308e+02, 1.05989791260970e+02, 1.05989763435979e+02, 1.05989759329066e+02, 1.05989758790326e+02, 1.05989746238321e+02, 1.05989661196633e+02, 1.05989348574246e+02, 1.05988515236628e+02, 1.05986691491977e+02, 1.05983198781100e+02, 1.05977124859603e+02, 1.05967307604724e+02, 1.05952328397487e+02, 1.05930515819287e+02, 1.05899960154241e+02, 1.05858538897583e+02, 1.05803953130842e+02, 1.05733774235058e+02, 1.05645499979277e+02, 1.05536618557430e+02, 1.05404678678092e+02, 1.05247363375657e+02, 1.05062564855297e+02, 1.04848457458644e+02, 1.04603565787140e+02, 1.04326825168542e+02, 1.04017631985021e+02, 1.03675881838912e+02, 1.03301994018632e+02, 1.02896921147699e+02, 1.02462143221269e+02, 1.01999645545860e+02, 1.01511880616410e+02, 1.01001714936798e+02, 1.00472363279306e+02, 9.99273145573482e+01, 9.93702546414065e+01, 9.88049913484729e+01, 9.82353852839529e+01, 9.76652878282450e+01, 9.70984854373932e+01, 9.65386484380308e+01, 9.59892827845538e+01, 9.54536843013202e+01, 9.49348960477671e+01, 9.44356701307411e+01, 9.39584354130477e+01, 9.35052722595195e+01, 9.30778949474182e+01, 9.26776418448646e+01, 9.23054730481144e+01, 9.19619749068508e+01, 9.16473707441241e+01, 9.13615370563840e+01, 9.11040245162732e+01, 9.08740831599180e+01, 9.06706911955537e+01, 9.04925869064320e+01, 9.03383031317190e+01, 9.02062037943658e+01, 9.00945219085589e+01, 9.00013984476192e+01, 8.99249213937645e+01, 8.98631642323360e+01, 8.98142231034122e+01, 8.97762517913687e+01, 8.97474937252012e+01, 8.97263101852077e+01, 8.97112039688616e+01, 8.97008378619646e+01, 8.96940473892991e+01, 8.96898474781786e+01, 8.96874328521695e+01, 8.96861721724649e+01, 8.96855961512572e+01, 8.96853800648065e+01, 8.96853212839573e+01, 8.96853126080236e+01, 8.96853118421038e+01, 8.96852968342827e+01, 8.96851951540788e+01, 8.96848213667807e+01, 8.96838249862139e+01, 8.96816444273180e+01, 8.96774683793207e+01, 8.96702061440707e+01, 8.96584682995032e+01, 8.96405588338537e+01, 8.96144796468005e+01, 8.95779480216613e+01, 8.95284273302001e+01, 8.94631708325199e+01, 8.93792779774478e+01, 8.92737620992993e+01, 8.91436278595558e+01, 8.89859562217749e+01, 8.87979942101924e+01, 8.85772462309824e+01, 8.83215633795550e+01, 8.80292269679800e+01, 8.76990225286180e+01, 8.73303008155859e+01, 8.69230228461443e+01, 8.64777867819599e+01, 8.59958353932061e+01, 8.54790438893165e+01, 8.49298889256395e+01, 8.43514004869803e+01, 8.37470990140224e+01, 8.31209205395108e+01, 8.24771327715169e+01, 8.18202450926116e+01, 8.11549154408423e+01, 8.04858570586662e+01, 7.98177481083127e+01, 7.91551470456392e+01, 7.85024162951059e+01, 7.78636561270921e+01, 7.72426497843452e+01, 7.66428200194815e+01, 7.60671964886716e+01, 7.55183930199903e+01, 7.49985936434963e+01, 7.45095463476403e+01, 7.40525636961706e+01, 7.36285296078825e+01, 7.32379117216450e+01, 7.28807788356482e+01, 7.25568229391647e+01, 7.22653853672828e+01, 7.20054866171368e+01, 7.17758593717837e+01, 7.15749842817895e+01, 7.14011280485449e+01, 7.12523833317065e+01, 7.11267099631513e+01, 7.10219768923693e+01, 7.09360042179103e+01, 7.08666045839460e+01, 7.08116231498268e+01, 7.07689752842033e+01, 7.07366811040872e+01, 7.07128959820060e+01, 7.06959361876964e+01, 7.06842989179984e+01, 7.06766760996057e+01, 7.06719615202351e+01, 7.06692510473174e+01, 7.06678359193750e+01, 7.06671893318919e+01, 7.06669467739834e+01, 7.06668807922925e+01, 7.06668710535310e+01, 7.06668702723618e+01, 7.06668558558293e+01, 7.06667581817007e+01, 7.06663991212848e+01, 7.06654419981432e+01, 7.06633473580563e+01, 7.06593358736131e+01, 7.06523598744778e+01, 7.06410848212894e+01, 7.06238818398212e+01, 7.05988321965621e+01, 7.05637443191577e+01, 7.05161836366561e+01, 7.04535151287658e+01, 7.03729580285793e+01, 7.02716516249331e+01, 7.01467305736317e+01, 6.99954075764305e+01, 6.98150607582739e+01, 6.96033226099452e+01, 6.93581670118855e+01, 6.90779906605715e+01, 6.87616852186192e+01, 6.84086967273236e+01, 6.80190692619244e+01, 6.75934704633216e+01, 6.71331974155841e+01, 6.66401623115178e+01, 6.61168584002811e+01, 6.55663077699728e+01, 6.49919935001460e+01, 6.43977795315565e+01, 6.37878221523706e+01, 6.31664772208929e+01, 6.25382071055029e+01, 6.19074908516215e+01, 6.12787403715536e+01, 6.06562246269033e+01, 6.00440029754869e+01, 5.94458681979991e+01, 5.88652992611504e+01, 5.83054236073988e+01, 5.77689886322506e+01, 5.72583419451955e+01, 5.67754199487694e+01, 5.63217441830077e+01, 5.58984247748444e+01, 5.55061702308553e+01, 5.51453027476401e+01, 5.48157782061047e+01, 5.45172100651029e+01, 5.42488964619021e+01, 5.40098499389177e+01, 5.37988293246248e+01, 5.36143733828153e+01, 5.34548358969103e+01, 5.33184218705263e+01, 5.32032245034831e+01, 5.31072625498076e+01, 5.30285175898391e+01, 5.29649706629192e+01, 5.29146376219497e+01, 5.28756024980556e+01, 5.28460481137146e+01, 5.28242831654758e+01, 5.28087650198247e+01, 5.27981175317832e+01, 5.27911433057871e+01, 5.27868299688192e+01, 5.27843502097643e+01, 5.27830555466596e+01, 5.27824640031610e+01, 5.27822420945728e+01, 5.27821817300375e+01, 5.27821728203578e+01, 5.27821721331576e+01, 5.27821597932477e+01, 5.27820761885521e+01, 5.27817688489942e+01, 5.27809495957030e+01, 5.27791566856496e+01, 5.27757230703907e+01, 5.27697520534446e+01, 5.27601014874015e+01, 5.27453774846330e+01, 5.27239384184650e+01, 5.26939097568391e+01, 5.26532099893597e+01, 5.25995875745563e+01, 5.25306684448090e+01, 5.24440131658956e+01, 5.23371823692808e+01, 5.22078085804405e+01, 5.20536720875667e+01, 5.18727780717000e+01, 5.16634318962379e+01, 5.14243092756048e+01, 5.11545180489498e+01, 5.08536485026967e+01, 5.05218096254376e+01, 5.01596493266475e+01, 4.97683574668842e+01, 4.93496514647259e+01, 4.89057451769036e+01, 4.84393025972081e+01, 4.79533786027104e+01, 4.74513494410131e+01, 4.69368358940054e+01, 4.64136221096874e+01, 4.58855730250923e+01, 4.53565531646389e+01, 4.48303494131718e+01, 4.43106001190190e+01, 4.38007325464650e+01, 4.33039102425109e+01, 4.28229913133462e+01, 4.23604979640308e+01, 4.19185970123995e+01, 4.14990905255134e+01, 4.11034153102119e+01, 4.07326497545326e+01, 4.03875264651390e+01, 4.00684492489120e+01, 3.97755131977332e+01, 3.95085269019014e+01, 3.92670360923341e+01, 3.90503482588469e+01, 3.88575579890798e+01, 3.86875729101627e+01, 3.85391401926341e+01, 3.84108735993197e+01, 3.83012810399988e+01, 3.82087925361936e+01, 3.81317884198768e+01, 3.80686274955245e+01, 3.80176747965369e+01, 3.79773284739414e+01, 3.79460452763905e+01, 3.79223640239015e+01, 3.79049264505781e+01, 3.78924947989913e+01, 3.78839655939683e+01, 3.78783791064694e+01, 3.78749241362348e+01, 3.78729378892393e+01, 3.78719008944690e+01, 3.78714270841724e+01, 3.78712493417630e+01, 3.78712009915360e+01, 3.78711938551445e+01, 3.78711933147985e+01, 3.78711837426230e+01, 3.78711188897451e+01, 3.78708804839939e+01, 3.78702449840051e+01, 3.78688542190527e+01, 3.78661907808149e+01, 3.78615591739797e+01, 3.78540736272619e+01, 3.78426533428009e+01, 3.78260258140141e+01, 3.78027386616834e+01, 3.77711802167844e+01, 3.77296088092390e+01, 3.76761904011237e+01, 3.76090438337543e+01, 3.75262925514578e+01, 3.74261212410462e+01, 3.73068354148492e+01, 3.71669216045590e+01, 3.70051055661518e+01, 3.68204057662269e+01, 3.66121794650419e+01, 3.63801589565911e+01, 3.61244759772141e+01, 3.58456729326453e+01, 3.55447003728545e+01, 3.52229009923452e+01, 3.48819812596483e+01, 3.45239724861341e+01, 3.41511836442956e+01, 3.37661484821117e+01, 3.33715694406124e+01, 3.29702606058268e+01, 3.25650914994590e+01, 3.21589330470182e+01, 3.17546066665345e+01, 3.13548371689916e+01, 3.09622100684146e+01, 3.05791339178623e+01, 3.02078083266071e+01, 2.98501982767392e+01, 2.95080151826601e+01, 2.91827048246195e+01, 2.88754418987155e+01, 2.85871305548042e+01, 2.83184100278317e+01, 2.80696643552954e+01, 2.78410352119908e+01, 2.76324370418767e+01, 2.74435738681418e+01, 2.72739573634026e+01, 2.71229259266317e+01, 2.69896646256253e+01, 2.68732259225336e+01, 2.67725511129024e+01, 2.66864923866696e+01, 2.66138353733397e+01, 2.65533219723127e+01, 2.65036732008017e+01, 2.64636117226317e+01, 2.64318836577286e+01, 2.64072792202310e+01, 2.63886516985256e+01, 2.63749342780400e+01, 2.63651542210665e+01, 2.63584439593168e+01, 2.63540487243404e+01, 2.63513304361775e+01, 2.63497676872785e+01, 2.63489517906062e+01, 2.63485790004698e+01, 2.63484391539858e+01, 2.63484011123481e+01, 2.63483954974820e+01, 2.63483950759722e+01, 2.63483876569224e+01, 2.63483373917749e+01, 2.63481526118689e+01, 2.63476600572002e+01, 2.63465821170025e+01, 2.63445177465405e+01, 2.63409278326902e+01, 2.63351256947644e+01, 2.63262733065091e+01, 2.63133837065061e+01, 2.62953299222811e+01, 2.62708605628492e+01, 2.62386220309537e+01, 2.61971870682361e+01, 2.61450890766433e+01, 2.60808613657490e+01, 2.60030801730050e+01, 2.59104100134112e+01, 2.58016496633688e+01, 2.56757769005129e+01, 2.55319900367638e+01, 2.53697443203907e+01, 2.51887814587910e+01, 2.49891508255230e+01, 2.47712213420013e+01, 2.45356835255606e+01, 2.42835417157723e+01, 2.40160969699310e+01, 2.37349215064365e+01, 2.34418258450024e+01, 2.31388199497372e+01, 2.28280697566217e+01, 2.25118505041648e+01, 2.21924983204017e+01, 2.18723615621898e+01, 2.15537534367676e+01, 2.12389074239924e+01, 2.09299369216365e+01, 2.06288003321471e+01, 2.03372725000606e+01, 2.00569230239613e+01, 1.97891015526481e+01, 1.95349297842601e+01, 1.92952995644060e+01, 1.90708762539674e+01, 1.88621064195408e+01, 1.86692288827350e+01, 1.84922882298204e+01, 1.83311500052274e+01, 1.81855169648266e+01, 1.80549459243487e+01, 1.79388648863669e+01, 1.78365902533382e+01, 1.77473440270881e+01, 1.76702709541408e+01, 1.76044556022697e+01, 1.75489393498960e+01, 1.75027372416397e+01, 1.74648546167357e+01, 1.74343033592663e+01, 1.74101175577215e+01, 1.73913683037404e+01, 1.73771773130717e+01, 1.73667290219169e+01, 1.73592808036302e+01, 1.73541709670960e+01, 1.73508242397757e+01, 1.73487545038926e+01, 1.73475646399106e+01, 1.73469434319857e+01, 1.73466595988506e+01, 1.73465531235495e+01, 1.73465241597149e+01, 1.73465198847146e+01, 1.73465195647649e+01, 1.73465139462879e+01, 1.73464758802966e+01, 1.73463359459518e+01, 1.73459629344063e+01, 1.73451466181201e+01, 1.73435833144672e+01, 1.73408648405291e+01, 1.73364713975695e+01, 1.73297688549650e+01, 1.73200108202718e+01, 1.73063457793523e+01, 1.72878294625076e+01, 1.72634424347795e+01, 1.72321127187739e+01, 1.71927430379108e+01, 1.71442420230660e+01, 1.70855584675928e+01, 1.70157174614940e+01, 1.69338570064291e+01, 1.68392635336489e+01, 1.67314046418728e+01, 1.66099573642564e+01, 1.64748303800133e+01, 1.63261788150729e+01, 1.61644106236907e+01, 1.59901839918867e+01, 1.58043957227522e+01, 1.56081611097738e+01, 1.54027863261685e+01, 1.51897348030622e+01, 1.49705893906219e+01, 1.47470122614205e+01, 1.45207045118366e+01, 1.42933672552619e+01, 1.40666657111246e+01, 1.38421974217559e+01, 1.36214653261415e+01, 1.34058560331352e+01, 1.31966233029841e+01, 1.29948764859280e+01, 1.28015734858774e+01, 1.26175177094992e+01, 1.24433584133336e+01, 1.22795938586910e+01, 1.21265767124235e+01, 1.19845211807273e+01, 1.18535114254934e+01, 1.17335108831152e+01, 1.16243721797976e+01, 1.15258474112511e+01, 1.14375986240908e+01, 1.13592083973233e+01, 1.12901904715360e+01, 1.12300004082158e+01, 1.11780462806155e+01, 1.11336994005591e+01, 1.10963050736090e+01, 1.10651933502395e+01, 1.10396897061771e+01, 1.10191255446923e+01, 1.10028483716891e+01, 1.09902314554786e+01, 1.09806827515740e+01, 1.09736528526950e+01, 1.09686417185492e+01, 1.09652039508780e+01, 1.09629524072895e+01, 1.09615599916525e+01, 1.09607595169102e+01, 1.09603416044920e+01, 1.09601506585774e+01, 1.09600790284511e+01, 1.09600595433462e+01, 1.09600566673865e+01, 1.09600564520633e+01, 1.09600526698031e+01, 1.09600270444491e+01, 1.09599328431595e+01, 1.09596817388167e+01, 1.09591322117778e+01, 1.09580798349932e+01, 1.09562498469368e+01, 1.09532923789497e+01, 1.09487806588270e+01, 1.09422124576755e+01, 1.09330149781573e+01, 1.09205532949380e+01, 1.09041423488810e+01, 1.08830623634768e+01, 1.08565773962043e+01, 1.08239565637300e+01, 1.07844972970645e+01, 1.07375498041986e+01, 1.06825417599610e+01, 1.06190021244234e+01, 1.05465829303807e+01, 1.04650778925431e+01, 1.03744367857807e+01, 1.02747747188621e+01, 1.01663756862618e+01, 1.00496900971636e+01, 9.92532633311178e+00, 9.79403674363182e+00, 9.65669881997724e+00, 9.51429255976240e+00, 9.36787522363477e+00, 9.21855477215067e+00, 9.06746325114578e+00, 8.91573127480818e+00, 8.76446455756735e+00, 8.61472319838052e+00, 8.46750415757894e+00, 8.32372711832214e+00, 8.18422371522428e+00, 8.04972995430165e+00, 7.92088154270043e+00, 7.79821178731058e+00, 7.68215169787175e+00, 7.57303193165299e+00, 7.47108623435108e+00, 7.37645605944832e+00, 7.28919608231580e+00, 7.20928036377353e+00, 7.13660895909317e+00, 7.07101481089704e+00, 7.01227080602727e+00, 6.96009691487602e+00, 6.91416736448546e+00, 6.87411782182089e+00, 6.83955257942849e+00, 6.81005174141217e+00, 6.78517840335975e+00, 6.76448580646056e+00, 6.74752442533295e+00, 6.73384892346655e+00, 6.72302488260867e+00, 6.71463518609555e+00, 6.70828591422760e+00, 6.70361159521387e+00, 6.70027965026689e+00, 6.69799387761130e+00, 6.69649683796699e+00, 6.69557103292068e+00, 6.69503880591867e+00, 6.69476094096117e+00, 6.69463398342354e+00, 6.69458635747135e+00, 6.69457340207681e+00, 6.69457148988840e+00, 6.69457134634301e+00, 6.69456881982340e+00, 6.69455170229222e+00, 6.69448877659882e+00, 6.69432104104739e+00, 6.69395396220831e+00, 6.69325098650729e+00, 6.69202858156640e+00, 6.69005305346560e+00, 6.68703935671127e+00, 6.68265207759489e+00, 6.67650872600839e+00, 6.66818541271476e+00, 6.65722491611055e+00, 6.64314705343229e+00, 6.62546116684369e+00, 6.60368041809892e+00, 6.57733746254608e+00, 6.54600095294711e+00, 6.50929221703261e+00, 6.46690137218278e+00, 6.41860209813903e+00, 6.36426429416578e+00, 6.30386390674143e+00, 6.23748932859692e+00, 6.16534393473066e+00, 6.08774452511221e+00, 6.00511567161165e+00, 5.91798019977143e+00, 5.82694625509671e+00, 5.73269159069498e+00, 5.63594585367328e+00, 5.53747173159005e+00, 5.43804584263994e+00, 5.33844021485415e+00, 5.23940510646116e+00, 5.14165378233268e+00, 5.04584969428957e+00, 4.95259633206538e+00, 4.86242983316835e+00, 4.77581427831075e+00, 4.69313946580672e+00, 4.61472086047030e+00, 4.54080135261287e+00, 4.47155443913192e+00, 4.40708844653959e+00, 4.34745144817731e+00, 4.29263657705555e+00, 4.24258749418320e+00, 4.19720383328817e+00, 4.15634650119706e+00, 4.11984276501038e+00, 4.08749110012896e+00, 4.05906580585685e+00, 4.03432141735014e+00, 4.01299695439204e+00, 3.99482004965215e+00, 3.97951099289922e+00, 3.96678671454880e+00, 3.95636471365735e+00, 3.94796691395378e+00, 3.94132340881556e+00, 3.93617603441486e+00, 3.93228169170017e+00, 3.92941532436484e+00, 3.92737245304602e+00, 3.92597116673477e+00, 3.92505348114901e+00, 3.92448599030287e+00, 3.92415976067052e+00, 3.92398944556734e+00, 3.92391162859768e+00, 3.92388243699357e+00, 3.92387449619093e+00, 3.92387332414589e+00, 3.92387323578733e+00, 3.92387167561112e+00, 3.92386110519897e+00, 3.92382224741139e+00, 3.92371866808979e+00, 3.92349199373933e+00, 3.92305791000044e+00, 3.92230311221486e+00, 3.92108336633959e+00, 3.91922282270839e+00, 3.91651470488337e+00, 3.91272346874028e+00, 3.90758849052948e+00, 3.90082929482861e+00, 3.89215227286381e+00, 3.88125876890689e+00, 3.86785432961547e+00, 3.85165882278568e+00, 3.83241704481290e+00, 3.80990935887397e+00, 3.78396184821956e+00, 3.75445544067261e+00, 3.72133346961282e+00, 3.68460718847274e+00, 3.64435885079899e+00, 3.60074210177308e+00, 3.55397958999301e+00, 3.50435788604847e+00, 3.45221996987397e+00, 3.39795570432824e+00, 3.34199083214347e+00, 3.28477510578612e+00, 3.22677017927592e+00, 3.16843785862235e+00, 3.11022923043700e+00, 3.05257507834558e+00, 2.99587786853769e+00, 2.94050545390193e+00, 2.88678652362356e+00, 2.83500772154504e+00, 2.78541227772921e+00, 2.73819994539627e+00, 2.69352800840717e+00, 2.65151311919428e+00, 2.61223373875845e+00, 2.57573297402921e+00, 2.54202163884075e+00, 2.51108139905895e+00, 2.48286789691434e+00, 2.45731378210856e+00, 2.43433160624076e+00, 2.41381656159177e+00, 2.39564906478847e+00, 2.37969720013848e+00, 2.36581904650001e+00, 2.35386491563710e+00, 2.34367952946583e+00, 2.33510415893009e+00, 2.32797873912138e+00, 2.32214396448906e+00, 2.31744335554876e+00, 2.31372527546948e+00, 2.31084486247515e+00, 2.30866583329809e+00, 2.30706210505170e+00, 2.30591917874111e+00, 2.30513522780824e+00, 2.30462183985574e+00, 2.30430436884915e+00, 2.30412186809139e+00, 2.30402659017807e+00, 2.30398305780973e+00, 2.30396672746060e+00, 2.30396228522339e+00, 2.30396162955900e+00, 2.30396157986054e+00, 2.30396069875049e+00, 2.30395472910737e+00, 2.30393278417451e+00, 2.30387428786084e+00, 2.30374627427103e+00, 2.30350112887977e+00, 2.30307486888853e+00, 2.30238605338046e+00, 2.30133540545597e+00, 2.29980621746667e+00, 2.29766559531210e+00, 2.29476657698133e+00, 2.29095113278679e+00, 2.28605401928547e+00, 2.27990741588837e+00, 2.27234622402212e+00, 2.26321385626477e+00, 2.25236829138011e+00, 2.23968812603741e+00, 2.22507832125800e+00, 2.20847532712850e+00, 2.18985127777738e+00, 2.16921698264593e+00, 2.14662349944000e+00, 2.12216215534239e+00, 2.09596297950689e+00, 2.06819161260121e+00, 2.03904485807476e+00, 2.00874512503711e+00, 1.97753407594831e+00, 1.94566582836379e+00, 1.91340006674474e+00, 1.88099539926607e+00, 1.84870324994652e+00, 1.81676251464509e+00, 1.78539513787877e+00, 1.75480269342029e+00, 1.72516398184943e+00, 1.69663359791565e+00, 1.66934137328701e+00, 1.64339256778695e+00, 1.61886866464417e+00, 1.59582862130100e+00, 1.57431043460549e+00, 1.55433289485094e+00, 1.53589742404975e+00, 1.51898991717243e+00, 1.50358252843868e+00, 1.48963536630387e+00, 1.47709807934429e+00, 1.46591133016604e+00, 1.45600816553699e+00, 1.44731529827163e+00, 1.43975432027898e+00, 1.43324286701481e+00, 1.42769575180898e+00, 1.42302608463483e+00, 1.41914638432910e+00, 1.41596968655981e+00, 1.41341064249701e+00, 1.41138659572819e+00, 1.40981861805292e+00, 1.40863247896418e+00, 1.40775951941062e+00, 1.40713739827981e+00, 1.40671068024470e+00, 1.40643123629643e+00, 1.40625843335162e+00, 1.40615909646085e+00, 1.40610723586849e+00, 1.40608354084169e+00, 1.40607465210098e+00, 1.40607223415570e+00, 1.40607187727232e+00, 1.40607185004946e+00, 1.40607136514646e+00, 1.40606807986153e+00, 1.40605600286860e+00, 1.40602381051566e+00, 1.40595336075494e+00, 1.40581845029330e+00, 1.40558386883588e+00, 1.40520480059013e+00, 1.40462661866285e+00, 1.40378511195897e+00, 1.40260717584084e+00, 1.40101198622369e+00, 1.39891266119305e+00, 1.39621839417585e+00, 1.39283701827135e+00, 1.38867793336755e+00, 1.38365529780309e+00, 1.37769135710018e+00, 1.37071975690888e+00, 1.36268866935575e+00, 1.35356355495118e+00, 1.34332938885623e+00, 1.33199220209812e+00, 1.31957982494365e+00, 1.30614176873571e+00, 1.29174823979379e+00, 1.27648833871926e+00, 1.26046755419840e+00, 1.24380470601382e+00, 1.22662852259029e+00, 1.20907405119837e+00, 1.19127909356050e+00, 1.17338083810563e+00, 1.15551282647422e+00, 1.13780235118880e+00, 1.12036833897814e+00, 1.10331973476305e+00, 1.08675436823685e+00, 1.07075826028499e+00, 1.05540531070376e+00, 1.04075730115649e+00, 1.02686414661805e+00, 1.01376433293271e+00, 1.00148548574924e+00, 9.90045025434114e-01, 9.79450872355557e-01, 9.69702176301169e-01, 9.60790052187050e-01, 9.52698311332597e-01, 9.45404183302415e-01, 9.38879027652320e-01, 9.33089037922713e-01, 9.27995941991668e-01, 9.23557703534797e-01, 9.19729228942060e-01, 9.16463082735717e-01, 9.13710212445622e-01, 9.11420681189014e-01, 9.09544403062128e-01, 9.08031873104733e-01, 9.06834880304004e-01, 9.05907189137167e-01, 9.05205172796855e-01, 9.04688379755555e-01, 9.04320014922973e-01, 9.04067317468531e-01, 9.03901819466611e-01, 9.03799472812878e-01, 9.03740636181951e-01, 9.03709918892517e-01, 9.03695884076939e-01, 9.03690619160201e-01, 9.03689186977772e-01, 9.03688975590725e-01, 9.03688959352422e-01, 9.03688668619054e-01, 9.03686698859303e-01, 9.03679457838152e-01, 9.03660156119244e-01, 9.03617915717874e-01, 9.03537023891275e-01, 9.03396363574373e-01, 9.03169049295855e-01, 9.02822294208818e-01, 9.02317529842620e-01, 9.01610794939202e-01, 9.00653402859803e-01, 8.99392888184238e-01, 8.97774222084010e-01, 8.95741272916496e-01, 8.93238473704187e-01, 8.90212642600659e-01, 8.86614887387807e-01, 8.82402512113671e-01, 8.77540834949031e-01, 8.72004822895860e-01, 8.65780452422783e-01, 8.58865716041578e-01, 8.51271212995812e-01, 8.43020286345322e-01, 8.34148696656854e-01, 8.24703851501965e-01, 8.14743637061397e-01, 8.04334920637700e-01, 7.93551808744780e-01, 7.82473753567999e-01, 7.71183600908600e-01, 7.59765666096030e-01, 7.48303912331437e-01, 7.36880290424395e-01, 7.25573281865763e-01, 7.14456670372911e-01, 7.03598551797711e-01, 6.93060579479393e-01, 6.82897432191155e-01, 6.73156484833584e-01, 6.63877657759798e-01, 6.55093418708177e-01, 6.46828911315268e-01, 6.39102185624111e-01, 6.31924508453480e-01, 6.25300734567972e-01, 6.19229722954472e-01, 6.13704785903479e-01, 6.08714161788129e-01, 6.04241505266813e-01, 6.00266390968386e-01, 5.96764828463705e-01, 5.93709787420920e-01, 5.91071732266087e-01, 5.88819165439517e-01, 5.86919177507539e-01, 5.85338001054339e-01, 5.84041563569713e-01, 5.82996032632067e-01, 5.82168344751906e-01, 5.81526707491022e-01, 5.81041063110931e-01, 5.80683501209066e-01, 5.80428607718829e-01, 5.80253738370871e-01, 5.80139206258096e-01, 5.80068375473473e-01, 5.80027655774771e-01, 5.80006396703382e-01, 5.79996683322250e-01, 5.79993039508039e-01, 5.79992048303183e-01, 5.79991902003485e-01, 5.79991890679267e-01, 5.79991686812278e-01, 5.79990305584261e-01, 5.79985228055747e-01, 5.79971693319806e-01, 5.79942073341812e-01, 5.79885349305485e-01, 5.79786711253050e-01, 5.79627300761885e-01, 5.79384113976062e-01, 5.79030079344548e-01, 5.78534320683403e-01, 5.77862611447131e-01, 5.76978020166038e-01, 5.75841739784593e-01, 5.74414085185111e-01, 5.72655633764143e-01, 5.70528474061021e-01, 5.67997517854891e-01, 5.65031822789081e-01, 5.61605866473244e-01, 5.57700710154663e-01, 5.53304991236345e-01, 5.48415689602209e-01, 5.43038622856386e-01, 5.37188639635758e-01, 5.30889497038672e-01, 5.24173426469602e-01, 5.17080410166443e-01, 5.09657206722366e-01, 5.01956176648946e-01, 4.94033967493053e-01, 4.85950121769733e-01, 4.77765670095224e-01, 4.69541766937766e-01, 4.61338418231250e-01, 4.53213339775814e-01, 4.45220973980369e-01, 4.37411681097899e-01, 4.29831110506370e-01, 4.22519748421720e-01, 4.15512631088632e-01, 4.08839207152099e-01, 4.02523329558079e-01, 3.96583355800587e-01, 3.91032335366702e-01, 3.85878264504939e-01, 3.81124390616321e-01, 3.76769551304236e-01, 3.72808536111123e-01, 3.69232461958546e-01, 3.66029156068151e-01, 3.63183542518641e-01, 3.60678030465095e-01, 3.58492903347440e-01, 3.56606709113080e-01, 3.54996651587712e-01, 3.53638982696651e-01, 3.52509394344963e-01, 3.51583407521701e-01, 3.50836754737287e-01, 3.50245750389746e-01, 3.49787642245374e-01, 3.49440936078153e-01, 3.49185684781887e-01, 3.49003733064151e-01, 3.48878909227054e-01, 3.48797156555975e-01, 3.48746598442659e-01, 3.48717533476261e-01, 3.48702359213746e-01, 3.48695426029352e-01, 3.48692825162473e-01, 3.48692117664569e-01, 3.48692013239400e-01, 3.48692005090911e-01, 3.48691857549355e-01, 3.48690857934201e-01, 3.48687183253844e-01, 3.48677387987522e-01, 3.48655951704188e-01, 3.48614900256933e-01, 3.48543516451741e-01, 3.48428154684726e-01, 3.48252172369087e-01, 3.47995988899044e-01, 3.47637280245877e-01, 3.47151313862781e-01, 3.46511424234592e-01, 3.45689624064050e-01, 3.44657339773406e-01, 3.43386252917554e-01, 3.41849221629289e-01, 3.40021248901664e-01, 3.37880458065626e-01, 3.35409031020376e-01, 3.32594062392303e-01, 3.29428283468744e-01, 3.25910613859776e-01, 3.22046506424812e-01, 3.17848061689290e-01, 3.13333901010139e-01, 3.08528802031077e-01, 3.03463114209151e-01, 2.98171985072784e-01, 2.92694438212038e-01, 2.87072350933116e-01, 2.81349382564144e-01, 2.75569903555024e-01, 2.69777971167514e-01, 2.64016390419584e-01, 2.58325889958996e-01, 2.52744432686825e-01, 2.47306671167507e-01, 2.42043548925366e-01, 2.36982041207767e-01, 2.32145023015391e-01, 2.27551248277862e-01, 2.23215421907440e-01, 2.19148345891242e-01, 2.15357121293323e-01, 2.11845389706724e-01, 2.08613599999040e-01, 2.05659288840891e-01, 2.02977366238471e-01, 2.00560399912163e-01, 1.98398894715865e-01, 1.96481565269082e-01, 1.94795601508304e-01, 1.93326927918662e-01, 1.92060457775128e-01, 1.90980343824634e-01, 1.90070226518188e-01, 1.89313480219070e-01, 1.88693456852604e-01, 1.88193725328247e-01, 1.87798303864838e-01, 1.87491881211059e-01, 1.87260021789825e-01, 1.87089349120541e-01, 1.86967701571191e-01, 1.86884254623735e-01, 1.86829604420154e-01, 1.86795808373887e-01, 1.86776380022152e-01, 1.86766236967230e-01, 1.86761602589144e-01, 1.86759864084848e-01, 1.86759391170568e-01, 1.86759321369451e-01, 1.86759315876401e-01, 1.86759215821937e-01, 1.86758537938942e-01, 1.86756045979990e-01, 1.86749403413842e-01, 1.86734866733914e-01, 1.86707028837716e-01, 1.86658623462151e-01, 1.86580400820972e-01, 1.86461083579957e-01, 1.86287411002525e-01, 1.86044277358803e-01, 1.85714968349686e-01, 1.85281496223482e-01, 1.84725030376697e-01, 1.84026415554442e-01, 1.83166764449155e-01, 1.82128105833698e-01, 1.80894063801304e-01, 1.79450538782703e-01, 1.77786357400535e-01, 1.75893856517575e-01, 1.73769367543642e-01, 1.71413570479018e-01, 1.68831693300845e-01, 1.66033540807931e-01, 1.63033347285761e-01, 1.59849458416573e-01, 1.56503858685767e-01, 1.53021570064709e-01, 1.49429955068314e-01, 1.45757961759983e-01, 1.42035349617871e-01, 1.38291933473852e-01, 1.34556878403572e-01, 1.30858072144662e-01, 1.27221594138831e-01, 1.23671292440832e-01, 1.20228472231247e-01, 1.16911693079544e-01, 1.13736666801742e-01, 1.10716243925651e-01, 1.07860474431405e-01, 1.05176727447869e-01, 1.02669854757593e-01, 1.00342384030583e-01, 9.81947294130379e-02, 9.62254091800787e-02, 9.44312624143180e-02, 9.28076589044053e-02, 9.13486985408554e-02, 9.00473983200811e-02, 8.88958665847213e-02, 8.78854652955283e-02, 8.70069619309675e-02, 8.62506730467629e-02, 8.56066016227572e-02, 8.50645701059420e-02, 8.46143505769703e-02, 8.42457927772244e-02, 8.39489499059312e-02, 8.37142012067176e-02, 8.35323694963819e-02, 8.33948310235684e-02, 8.32936144632299e-02, 8.32214855121833e-02, 8.31720134992651e-02, 8.31396166765046e-02, 8.31195834053489e-02, 8.31080672613922e-02, 8.31020550858312e-02, 8.30993081392251e-02, 8.30982776754041e-02, 8.30979973652717e-02, 8.30979559921303e-02, 8.30979527074192e-02, 8.30978925109918e-02, 8.30974846718817e-02, 8.30959854219486e-02, 8.30919890466078e-02, 8.30832434550643e-02, 8.30664960502219e-02, 8.30373767073809e-02, 8.29903242815817e-02, 8.29185625758839e-02, 8.28141310229569e-02, 8.26679743419163e-02, 8.24700940008114e-02, 8.22097623337044e-02, 8.18757975618354e-02, 8.14568947615300e-02, 8.09420041089481e-02, 8.03207437384543e-02, 7.95838306046619e-02, 7.87235092718532e-02, 7.77339560358143e-02, 7.66116346861736e-02, 7.53555809050255e-02, 7.39675949885736e-02, 7.24523272430275e-02, 7.08172467744532e-02, 6.90724919333832e-02, 6.72306086768231e-02, 6.53061907606087e-02, 6.33154421833098e-02, 6.12756870154811e-02, 5.92048542378336e-02, 5.71209653344908e-02, 5.50416502820212e-02, 5.29837136255225e-02, 5.09627670939494e-02, 4.89929393099544e-02, 4.70866672073811e-02, 4.52545683115915e-02, 4.35053884473007e-02, 4.18460159581997e-02, 4.02815512384933e-02, 3.88154192608425e-02, 3.74495126986290e-02, 3.61843539876361e-02, 3.50192660167039e-02, 3.39525428504572e-02, 3.29816137743251e-02, 3.21031958470428e-02, 3.13134319310109e-02, 3.06080127586057e-02, 2.99822829212673e-02, 2.94313317089347e-02, 2.89500704759433e-02, 2.85332986859004e-02, 2.81757610097202e-02, 2.78721978488647e-02, 2.76173914607369e-02, 2.74062095104028e-02, 2.72336473969972e-02, 2.70948701419513e-02, 2.69852540251776e-02, 2.69004275633620e-02, 2.68363108856704e-02, 2.67891521228876e-02, 2.67555591213977e-02, 2.67325246515497e-02, 2.67174433151922e-02, 2.67081185646703e-02, 2.67027586101847e-02, 2.66999604777709e-02, 2.66986820443407e-02, 2.66982024689639e-02, 2.66980720136729e-02, 2.66980527587890e-02, 2.66980512160870e-02, 2.66980227682438e-02, 2.66978300305425e-02, 2.66971215141751e-02, 2.66952329300238e-02, 2.66911001159116e-02, 2.66831864556849e-02, 2.66694282049400e-02, 2.66472010678882e-02, 2.66133112563409e-02, 2.65640140418810e-02, 2.64950623939484e-02, 2.64017875887106e-02, 2.62792125934652e-02, 2.61221975253108e-02, 2.59256145503599e-02, 2.56845473011111e-02, 2.53945073902884e-02, 2.50516581244966e-02, 2.46530333684518e-02, 2.41967380090955e-02, 2.36821159438847e-02, 2.31098722156560e-02, 2.24821379708828e-02, 2.18024702898418e-02, 2.10757833881132e-02, 2.03082128023340e-02, 1.95069194128763e-02, 1.86798449454528e-02, 1.78354343531595e-02, 1.69823427452656e-02, 1.61291450930747e-02, 1.52840658456768e-02, 1.44547429862456e-02, 1.36480373287640e-02, 1.28698935941145e-02, 1.21252554062323e-02, 1.14180321628510e-02, 1.07511124148090e-02, 1.01264160785632e-02, 9.54497629240336e-03, 9.00704152677179e-03, 8.51218925961916e-03, 8.05944307190032e-03, 7.64738667044185e-03, 7.27427057506702e-03, 6.93810746767554e-03, 6.63675373985484e-03, 6.36797824994062e-03, 6.12951810880702e-03, 5.91911953735291e-03, 5.73456705758701e-03, 5.57370676254720e-03, 5.43446308577700e-03, 5.31484624924872e-03, 5.21295362273122e-03, 5.12697072443302e-03, 5.05517317104420e-03, 4.99592777056534e-03, 4.94769234371972e-03, 4.90901571150459e-03, 4.87853911080549e-03, 4.85499921094856e-03, 4.83723224340338e-03, 4.82417861454110e-03, 4.81488738824964e-03, 4.80852001572885e-03, 4.80435267900248e-03, 4.80177664630332e-03, 4.80029613315480e-03, 4.79952330239541e-03, 4.79917021947074e-03, 4.79903777063839e-03, 4.79900174185843e-03, 4.79899642407133e-03, 4.79899599402245e-03, 4.79898801306433e-03, 4.79893394113162e-03, 4.79873517093665e-03, 4.79820535382203e-03, 4.79704601956144e-03, 4.79482635972772e-03, 4.79096824220449e-03, 4.78473756984954e-03, 4.77524312906041e-03, 4.76144406643186e-03, 4.74216703766217e-03, 4.71613384432601e-03, 4.68199995943053e-03, 4.63840368894353e-03, 4.58402480855113e-03, 4.51765044340391e-03, 4.43824503447843e-03, 4.34502085128941e-03, 4.23750547524548e-03, 4.11560157735380e-03, 3.97963104649079e-03, 3.83035318198210e-03, 3.66895263523759e-03, 3.49700542136588e-03, 3.31642955192990e-03, 3.12940844810328e-03, 2.93828655330231e-03, 2.74547348304566e-03, 2.55336986084842e-03, 2.36427769706893e-03, 2.18029158969557e-03, 2.00321390421720e-03, 1.83449594960905e-03, 1.67519812075057e-03, 1.52600862496399e-03, 1.38730233260276e-03, 1.25918595674079e-03, 1.14154360915109e-03, 1.03409644829355e-03, 9.36405231200865e-04, 8.47811229604645e-04, 7.67837143300198e-04, 6.95593225003677e-04, 6.29573252183491e-04, 5.69933113354182e-04, 5.16068862211227e-04, 4.66438411280024e-04, 4.20432754864919e-04, 3.77606047308632e-04, 3.37670478955943e-04, 3.00454502919933e-04, 2.65926085063484e-04, 2.34451989939544e-04, 2.05270046208739e-04, 1.78216569295842e-04, 1.53321804020834e-04, 1.30601072882541e-04, 1.10039829976927e-04, 9.15912413717024e-05, 7.51871147883776e-05, 6.07514100181561e-05, 4.82030846598036e-05, 3.74520305967962e-05, 2.83951866464349e-05, 2.09148763303801e-05, 1.48791587122332e-05, 1.01435809796158e-05, 6.55378576652765e-06, 3.94858912500240e-06, 2.16330970204126e-06, 1.03327555223132e-06, 3.97563809428902e-07, -1.42108547152020e-14, 1.42108547152020e-14],
[ 1.14000000000000e+02, 1.13999999849866e+02, 1.13999995207422e+02, 1.13999963754124e+02, 1.13999848127631e+02, 1.13999539903364e+02, 1.13998865329173e+02, 1.13997573312763e+02, 1.13995326090785e+02, 1.13991692923683e+02, 1.13986147065076e+02, 1.13978066153296e+02, 1.13966736071265e+02, 1.13951358225356e+02, 1.13931060110169e+02, 1.13904908959108e+02, 1.13871928232443e+02, 1.13831116663818e+02, 1.13781469566609e+02, 1.13722002082503e+02, 1.13651774021923e+02, 1.13569915885217e+02, 1.13475655554100e+02, 1.13368345002969e+02, 1.13247486209692e+02, 1.13112755269817e+02, 1.12964023572849e+02, 1.12801374825881e+02, 1.12625116746215e+02, 1.12435786414263e+02, 1.12234148582434e+02, 1.12021186650142e+02, 1.11798086491854e+02, 1.11566213801868e+02, 1.11327086032346e+02, 1.11082340297510e+02, 1.10833698766026e+02, 1.10582933061048e+02, 1.10331829053526e+02, 1.10082153207220e+02, 1.09835621359499e+02, 1.09593870544387e+02, 1.09358434217839e+02, 1.09130721050555e+02, 1.08911997317382e+02, 1.08703372829921e+02, 1.08505790318269e+02, 1.08320018154148e+02, 1.08146646306996e+02, 1.07986085425666e+02, 1.07838568934160e+02, 1.07704158016713e+02, 1.07582749344940e+02, 1.07474085368952e+02, 1.07377766957504e+02, 1.07293268131808e+02, 1.07219952596004e+02, 1.07157091726530e+02, 1.07103883644762e+02, 1.07059472964085e+02, 1.07022970775907e+02, 1.06993474420828e+02, 1.06970086583067e+02, 1.06951933249976e+02, 1.06938180095431e+02, 1.06928046876990e+02, 1.06920819482211e+02, 1.06915859319001e+02, 1.06912609817032e+02, 1.06910599889992e+02, 1.06909444299000e+02, 1.06908840952415e+02, 1.06908565272910e+02, 1.06908461855125e+02, 1.06908433722897e+02, 1.06908429570636e+02, 1.06908429025947e+02, 1.06908416335345e+02, 1.06908330354645e+02, 1.06908014280355e+02, 1.06907171741253e+02, 1.06905327859509e+02, 1.06901796584178e+02, 1.06895655600292e+02, 1.06885729958838e+02, 1.06870585389400e+02, 1.06848532043515e+02, 1.06817639166268e+02, 1.06775760899295e+02, 1.06720573074788e+02, 1.06649620465821e+02, 1.06560373519002e+02, 1.06450293125663e+02, 1.06316901513901e+02, 1.06157856903351e+02, 1.05971029205526e+02, 1.05754573826000e+02, 1.05507000575010e+02, 1.05227234841968e+02, 1.04914668520643e+02, 1.04569198623708e+02, 1.04191252001927e+02, 1.03781794993973e+02, 1.03342327155353e+02, 1.02874858552780e+02, 1.02381870692308e+02, 1.01866262214923e+02, 1.01331282069194e+02, 1.00780454570060e+02, 1.00217501819225e+02, 9.96462686828207e+01, 9.90706537900343e+01, 9.84945475550069e+01, 9.79217761734307e+01, 9.73560497212864e+01, 9.68009129307724e+01, 9.62596983564669e+01, 9.57354827549006e+01, 9.52310481263324e+01, 9.47488489174095e+01, 9.42909865154838e+01, 9.38591916159079e+01, 9.34548145087484e+01, 9.30788229224327e+01, 9.27318068119937e+01, 9.24139893707092e+01, 9.21252435347772e+01, 9.18651132967633e+01, 9.16328392081567e+01, 9.14273875092186e+01, 9.12474823611242e+01, 9.10916406656022e+01, 9.09582089412863e+01, 9.08454016880742e+01, 9.07513406174611e+01, 9.06740940658793e+01, 9.06117158479683e+01, 9.05622827559677e+01, 9.05239298783008e+01, 9.04948829022446e+01, 9.04734865883034e+01, 9.04582286614915e+01, 9.04477584587457e+01, 9.04408998009942e+01, 9.04366577191221e+01, 9.04342188487822e+01, 9.04329455112544e+01, 9.04323637065605e+01, 9.04321454505208e+01, 9.04320860794905e+01, 9.04320773164473e+01, 9.04320765428374e+01, 9.04320613843323e+01, 9.04319586832230e+01, 9.04315811429740e+01, 9.04305747584613e+01, 9.04283723063948e+01, 9.04241543311993e+01, 9.04168191868360e+01, 9.04049635088285e+01, 9.03868742744959e+01, 9.03605333576437e+01, 9.03236351890632e+01, 9.02736177880669e+01, 9.02077070268312e+01, 9.01229735269806e+01, 9.00164010722087e+01, 8.98849648666376e+01, 8.97257174014149e+01, 8.95358791475907e+01, 8.93129308164065e+01, 8.90547035692462e+01, 8.87594633699299e+01, 8.84259856973966e+01, 8.80536171094393e+01, 8.76423206785413e+01, 8.71927030896037e+01, 8.67060221414988e+01, 8.61841744398538e+01, 8.56296640925144e+01, 8.50455541046251e+01, 8.44354028288505e+01, 8.38031882279302e+01, 8.31532228924167e+01, 8.24900628200137e+01, 8.18184130023622e+01, 8.11430329219710e+01, 8.04686450852431e+01, 7.97998495835004e+01, 7.91410472623459e+01, 7.84963733641525e+01, 7.78696425980705e+01, 7.72643056905295e+01, 7.66834167763933e+01, 7.61296106113403e+01, 7.56050885011414e+01, 7.51116119498641e+01, 7.46505032050425e+01, 7.42226520355548e+01, 7.38285281801990e+01, 7.34681989545465e+01, 7.31413515215285e+01, 7.28473193383242e+01, 7.25851123002139e+01, 7.23534501127340e+01, 7.21507984314907e+01, 7.19754073067866e+01, 7.18253514514559e+01, 7.16985718119664e+01, 7.15929178659598e+01, 7.15061899988166e+01, 7.14361812354475e+01, 7.13807175312220e+01, 7.13376957685070e+01, 7.13051185731309e+01, 7.12811250672434e+01, 7.12640167181400e+01, 7.12522775301430e+01, 7.12445879584411e+01, 7.12398320959922e+01, 7.12370978898213e+01, 7.12356703711030e+01, 7.12350181221998e+01, 7.12347734405016e+01, 7.12347068810888e+01, 7.12346970570568e+01, 7.12346962690479e+01, 7.12346817262872e+01, 7.12345831969455e+01, 7.12342209926828e+01, 7.12332554892589e+01, 7.12311425094273e+01, 7.12270959033223e+01, 7.12200588309007e+01, 7.12086850766578e+01, 7.11913315240240e+01, 7.11660626795850e+01, 7.11306678572832e+01, 7.10826913014164e+01, 7.10194751379539e+01, 7.09382145946048e+01, 7.08360244266915e+01, 7.07100149431703e+01, 7.05573754705812e+01, 7.03754625579268e+01, 7.01618897560340e+01, 6.99146154482544e+01, 6.96320250111779e+01, 6.93130035823354e+01, 6.89569959312165e+01, 6.85640503776786e+01, 6.81348443665877e+01, 6.76706901598602e+01, 6.71735201010493e+01, 6.66458519820957e+01, 6.60907361209461e+01, 6.55116867531044e+01, 6.49126011525143e+01, 6.42976704346973e+01, 6.36712861892170e+01, 6.30379469166434e+01, 6.24021677453396e+01, 6.17683961733966e+01, 6.11409357578117e+01, 6.05238788970288e+01, 5.99210492303286e+01, 5.93359537540654e+01, 5.87717445123620e+01, 5.82311895959611e+01, 5.77166531029393e+01, 5.72300836235867e+01, 5.67730106899461e+01, 5.63465484937222e+01, 5.59514060583578e+01, 5.55879029842197e+01, 5.52559898862176e+01, 5.49552727068073e+01, 5.46850401946354e+01, 5.44442939636948e+01, 5.42317806647467e+01, 5.40460258913242e+01, 5.38853694962997e+01, 5.37480020086105e+01, 5.36320018158403e+01, 5.35353727234946e+01, 5.34560814251559e+01, 5.33920943300860e+01, 5.33414131079041e+01, 5.33021082354782e+01, 5.32723497801688e+01, 5.32504346355510e+01, 5.32348094477993e+01, 5.32240885369950e+01, 5.32170662280381e+01, 5.32127231571640e+01, 5.32102263052780e+01, 5.32089227185267e+01, 5.32083270978206e+01, 5.32081036597441e+01, 5.32080428791521e+01, 5.32080339080634e+01, 5.32080332161267e+01, 5.32080207911653e+01, 5.32079366102337e+01, 5.32076271523820e+01, 5.32068022525785e+01, 5.32049969856833e+01, 5.32015397071734e+01, 5.31955275445771e+01, 5.31858104894898e+01, 5.31709850720010e+01, 5.31493984012975e+01, 5.31191631198971e+01, 5.30781835358127e+01, 5.30241928601569e+01, 5.29548010848239e+01, 5.28675525901554e+01, 5.27599920888243e+01, 5.26297370122344e+01, 5.24745539619060e+01, 5.22924364210439e+01, 5.20816805960841e+01, 5.18409560799785e+01, 5.15693680381110e+01, 5.12665078416065e+01, 5.09324895202994e+01, 5.05679700634188e+01, 5.01741524173358e+01, 4.97527709471975e+01, 4.93060600543219e+01, 4.88367074804080e+01, 4.83477945045550e+01, 4.78427257057563e+01, 4.73251512221891e+01, 4.67988845273240e+01, 4.62678187154001e+01, 4.57358441865542e+01, 4.52067704548738e+01, 4.46842545497473e+01, 4.41717381092045e+01, 4.36723947557199e+01, 4.31890887177538e+01, 4.27243449704268e+01, 4.22803304977515e+01, 4.18588457105493e+01, 4.14613246521252e+01, 4.10888424174449e+01, 4.07421281940197e+01, 4.04215824681038e+01, 4.01272971761649e+01, 3.98590778641229e+01, 3.96164671982162e+01, 3.93987694173370e+01, 3.92050755074158e+01, 3.90342890066226e+01, 3.88851524176028e+01, 3.87562742170609e+01, 3.86461564237804e+01, 3.85532226243536e+01, 3.84758462719970e+01, 3.84123789777474e+01, 3.83611784144761e+01, 3.83206353614922e+01, 3.82891993395900e+01, 3.82654022311389e+01, 3.82478792540258e+01, 3.82353866671294e+01, 3.82268156314298e+01, 3.82212017350396e+01, 3.82177298095747e+01, 3.82157338136489e+01, 3.82146917286645e+01, 3.82142155925170e+01, 3.82140369775861e+01, 3.82139883900108e+01, 3.82139812185871e+01, 3.82139806755885e+01, 3.82139710564237e+01, 3.82139058851857e+01, 3.82136663091004e+01, 3.82130276893673e+01, 3.82116300966087e+01, 3.82089535810696e+01, 3.82042992288258e+01, 3.81967769090618e+01, 3.81853004928802e+01, 3.81685911750974e+01, 3.81451893489598e+01, 3.81134752611226e+01, 3.80716984034856e+01, 3.80180152763531e+01, 3.79505347868256e+01, 3.78673701384620e+01, 3.77666956436908e+01, 3.76468064793737e+01, 3.75061790467332e+01, 3.73435293329943e+01, 3.71578665472039e+01, 3.69485393543637e+01, 3.67152722846109e+01, 3.64581903520285e+01, 3.61778305595593e+01, 3.58751397433973e+01, 3.55514590474336e+01, 3.52084961234157e+01, 3.48482868289411e+01, 3.44731486607051e+01, 3.40856283647658e+01, 3.36884461065438e+01, 3.32844383106169e+01, 3.28765008871119e+01, 3.24675341606905e+01, 3.20603905101216e+01, 3.16578255663379e+01, 3.12624537946611e+01, 3.08767093305207e+01, 3.05028129423914e+01, 3.01427458681817e+01, 2.97982309745189e+01, 2.94707212590587e+01, 2.91613952496904e+01, 2.88711584639209e+01, 2.86006498604318e+01, 2.83502521698315e+01, 2.81201051041014e+01, 2.79101206515846e+01, 2.77199998987665e+01, 2.75492510293369e+01, 2.73972083066837e+01, 2.72630519407568e+01, 2.71458287804526e+01, 2.70444737702974e+01, 2.69578320777314e+01, 2.68846817451635e+01, 2.68237566572844e+01, 2.67737695453415e+01, 2.67334346818491e+01, 2.67014898570400e+01, 2.66767171778367e+01, 2.66579621966829e+01, 2.66441508661352e+01, 2.66343038294598e+01, 2.66275475997482e+01, 2.66231222503376e+01, 2.66203853354138e+01, 2.66188118772244e+01, 2.66179903891192e+01, 2.66176150441609e+01, 2.66174742392670e+01, 2.66174359369178e+01, 2.66174302835712e+01, 2.66174298591727e+01, 2.66174223892778e+01, 2.66173717796464e+01, 2.66171857333776e+01, 2.66166898030191e+01, 2.66156044750466e+01, 2.66135259554985e+01, 2.66099114343479e+01, 2.66040695193220e+01, 2.65951564284779e+01, 2.65821784111804e+01, 2.65640007241181e+01, 2.65393633172674e+01, 2.65069031800189e+01, 2.64651830584522e+01, 2.64127259837301e+01, 2.63480547571851e+01, 2.62697352348031e+01, 2.61764219639141e+01, 2.60669044748202e+01, 2.59401523497029e+01, 2.57953571095482e+01, 2.56319690004632e+01, 2.54497269359454e+01, 2.52486801576922e+01, 2.50292005917016e+01, 2.47919853580659e+01, 2.45380493893182e+01, 2.42687085698108e+01, 2.39855541859883e+01, 2.36904197564726e+01, 2.33853415001828e+01, 2.30725138286939e+01, 2.27542413485467e+01, 2.24328889504342e+01, 2.21108316415805e+01, 2.17904058186195e+01, 2.14738636429328e+01, 2.11633320365540e+01, 2.08607775528238e+01, 2.05679780077171e+01, 2.02865013250502e+01, 2.00176916051137e+01, 1.97626620254036e+01, 1.95222938664269e+01, 1.92972407491913e+01, 1.90879370784945e+01, 1.88946096955154e+01, 1.87172918316030e+01, 1.85558385949658e+01, 1.84099433861034e+01, 1.82791548032438e+01, 1.81628937484718e+01, 1.80604705674140e+01, 1.79711021444116e+01, 1.78939289295158e+01, 1.78280318950048e+01, 1.77724494112894e+01, 1.77261940004257e+01, 1.76882688763507e+01, 1.76576841214334e+01, 1.76334722863275e+01, 1.76147031417545e+01, 1.76004972636333e+01, 1.75900381029146e+01, 1.75825821832161e+01, 1.75774670857337e+01, 1.75741169226504e+01, 1.75720450658970e+01, 1.75708539840121e+01, 1.75702321406341e+01, 1.75699480172470e+01, 1.75698414330772e+01, 1.75698124396291e+01, 1.75698081602580e+01, 1.75698078399811e+01, 1.75698022157596e+01, 1.75697641108497e+01, 1.75696240334446e+01, 1.75692506406220e+01, 1.75684334902699e+01, 1.75668685906208e+01, 1.75641473454203e+01, 1.75597494345897e+01, 1.75530401017433e+01, 1.75432722369496e+01, 1.75295935402902e+01, 1.75110589233644e+01, 1.74866481472375e+01, 1.74552885040955e+01, 1.74158821274739e+01, 1.73673372686096e+01, 1.73086026159568e+01, 1.72387034784388e+01, 1.71567784225231e+01, 1.70621147737148e+01, 1.69541812901702e+01, 1.68326563129325e+01, 1.66974498108180e+01, 1.65487179757769e+01, 1.63868693817091e+01, 1.62125621771093e+01, 1.60266923062297e+01, 1.58303732993903e+01, 1.56249086881632e+01, 1.54117585326200e+01, 1.51925018506743e+01, 1.49687968843077e+01, 1.47423411157267e+01, 1.45148327724739e+01, 1.42879352681914e+01, 1.40632456617142e+01, 1.38422678313807e+01, 1.36263906978634e+01, 1.34168715187833e+01, 1.32148240378659e+01, 1.30212111023997e+01, 1.28368412574411e+01, 1.26623687712146e+01, 1.24982965309461e+01, 1.23449812618912e+01, 1.22026405574750e+01, 1.20713612603006e+01, 1.19511087981314e+01, 1.18417371513830e+01, 1.17429992040899e+01, 1.16545573032462e+01, 1.15759939165954e+01, 1.15068223319342e+01, 1.14464973786981e+01, 1.13944261734517e+01, 1.13499788948391e+01, 1.13124995818417e+01, 1.12813169242597e+01, 1.12557549794728e+01, 1.12351437087463e+01, 1.12188291840261e+01, 1.12061832769232e+01, 1.11966126098531e+01, 1.11895665290491e+01, 1.11845438535370e+01, 1.11810981651179e+01, 1.11788414325318e+01, 1.11774458073409e+01, 1.11766434872980e+01, 1.11762246114300e+01, 1.11760332252989e+01, 1.11759614300308e+01, 1.11759419000032e+01, 1.11759390174130e+01, 1.11759388015933e+01, 1.11759350106133e+01, 1.11759093261801e+01, 1.11758149077087e+01, 1.11755632244338e+01, 1.11750124303894e+01, 1.11739576270359e+01, 1.11721234188371e+01, 1.11691591291220e+01, 1.11646369986454e+01, 1.11580536342938e+01, 1.11488349062911e+01, 1.11363444047614e+01, 1.11198954570082e+01, 1.10987665735758e+01, 1.10722200350770e+01, 1.10395231576456e+01, 1.09999715916718e+01, 1.09529138293725e+01, 1.08977759383808e+01, 1.08340854195049e+01, 1.07614930255019e+01, 1.06797913896617e+01, 1.05889294082216e+01, 1.04890215012829e+01, 1.03803511358758e+01, 1.02633683154338e+01, 1.01386810969884e+01, 1.00070415594831e+01, 9.86932697949048e+00, 9.72651724142287e+00, 9.57966969105601e+00, 9.42989271706310e+00, 9.27831931195973e+00, 9.12608173269743e+00, 8.97428817630323e+00, 8.82400213986658e+00, 8.67622488080585e+00, 8.53188116158698e+00, 8.39180827193793e+00, 8.25674817878577e+00, 8.12734255770626e+00, 8.00413040178717e+00, 7.88754787479985e+00, 7.77793006696163e+00, 7.67551431742514e+00, 7.58044478450790e+00, 7.49277797069824e+00, 7.41248894325285e+00, 7.33947803101316e+00, 7.27357782154277e+00, 7.21456032685165e+00, 7.16214422760484e+00, 7.11600214173144e+00, 7.07576789120029e+00, 7.04104375863758e+00, 7.01140773259168e+00, 6.98642073673064e+00, 6.96563382515772e+00, 6.94859530523802e+00, 6.93485772341538e+00, 6.92398462150783e+00, 6.91555694418116e+00, 6.90917895596786e+00, 6.90448351125403e+00, 6.90113651543686e+00, 6.89884042146331e+00, 6.89733662369601e+00, 6.89640663994092e+00, 6.89587201091604e+00, 6.89559289197899e+00, 6.89546536150891e+00, 6.89541752063346e+00, 6.89540450677487e+00, 6.89540258595731e+00, 6.89540244176412e+00, 6.89539990384307e+00, 6.89538270906553e+00, 6.89531949940856e+00, 6.89515100693184e+00, 6.89478227166821e+00, 6.89407612405283e+00, 6.89284820418112e+00, 6.89086376532182e+00, 6.88783647965863e+00, 6.88342942802826e+00, 6.87725840965227e+00, 6.86889764956535e+00, 6.85788790816174e+00, 6.84374690776686e+00, 6.82598188607663e+00, 6.80410396887544e+00, 6.77764393068288e+00, 6.74616879070840e+00, 6.70929858381209e+00, 6.66672256441910e+00, 6.61821405762370e+00, 6.56364317618286e+00, 6.50298668115646e+00, 6.43633437888273e+00, 6.36389161315913e+00, 6.28597761842335e+00, 6.20301973196860e+00, 6.11554370203288e+00, 6.02416055403870e+00, 5.92955067036885e+00, 5.83244588389840e+00, 5.73361047072251e+00, 5.63382194750979e+00, 5.53385253429165e+00, 5.43445204103303e+00, 5.33633278814004e+00, 5.24015699319222e+00, 5.14652686663541e+00, 5.05597747565044e+00, 4.96897227320473e+00, 4.88590105960294e+00, 4.80708005288952e+00, 4.73275369341488e+00, 4.66309779361228e+00, 4.59822366032051e+00, 4.53818285589206e+00, 4.48297231756051e+00, 4.43253961445791e+00, 4.38678818194033e+00, 4.34558242873209e+00, 4.30875266063454e+00, 4.27609980331587e+00, 4.24739993522302e+00, 4.22240865994020e+00, 4.20086535590733e+00, 4.18249734122303e+00, 4.16702398349402e+00, 4.15416077076796e+00, 4.14362334113231e+00, 4.13513144738400e+00, 4.12841281123062e+00, 4.12320680080647e+00, 4.11926784786966e+00, 4.11636850870440e+00, 4.11430206695795e+00, 4.11288457836535e+00, 4.11195626690704e+00, 4.11138219906533e+00, 4.11105218647424e+00, 4.11087989579192e+00, 4.11080117604196e+00, 4.11077164575404e+00, 4.11076361281937e+00, 4.11076242717576e+00, 4.11076233779202e+00, 4.11076075951395e+00, 4.11075006645819e+00, 4.11071075780808e+00, 4.11060597656147e+00, 4.11037667136654e+00, 4.10993754747821e+00, 4.10917397929748e+00, 4.10794004326678e+00, 4.10605781344147e+00, 4.10331804109072e+00, 4.09948231194895e+00, 4.09428673785605e+00, 4.08744719144289e+00, 4.07866603210958e+00, 4.06764019912565e+00, 4.05407046553312e+00, 4.03767155916098e+00, 4.01818277121725e+00, 3.99537859722403e+00, 3.96907889916949e+00, 3.93915805120314e+00, 3.90555254191547e+00, 3.86826655902449e+00, 3.82737517749875e+00, 3.78302490494187e+00, 3.73543149838392e+00, 3.68487514008022e+00, 3.63169322953166e+00, 3.57627119748852e+00, 3.51903186026334e+00, 3.46042389887791e+00, 3.40091006304776e+00, 3.34095566683560e+00, 3.28101786875474e+00, 3.22153612598318e+00, 3.16292409381617e+00, 3.10556312084914e+00, 3.04979737882071e+00, 2.99593057154733e+00, 2.94422409445086e+00, 2.89489646616539e+00, 2.84812382552251e+00, 2.80404127814074e+00, 2.76274488331144e+00, 2.72429408999324e+00, 2.68871445670717e+00, 2.65600052048897e+00, 2.62611871180685e+00, 2.59901024303659e+00, 2.57459392583841e+00, 2.55276889630463e+00, 2.53341724527228e+00, 2.51640656438803e+00, 2.50159242639640e+00, 2.48882082097991e+00, 2.47793056579658e+00, 2.46875570676667e+00, 2.46112791292114e+00, 2.45487886011222e+00, 2.44984258558696e+00, 2.44585778290137e+00, 2.44276999502139e+00, 2.44043365381658e+00, 2.43871390750624e+00, 2.43748817478489e+00, 2.43664736587381e+00, 2.43609671680642e+00, 2.43575619365714e+00, 2.43556043757080e+00, 2.43545823841676e+00, 2.43541154350176e+00, 2.43539402674395e+00, 2.43538926177213e+00, 2.43538855847268e+00, 2.43538850516356e+00, 2.43538756003943e+00, 2.43538115669035e+00, 2.43535761739206e+00, 2.43529487095824e+00, 2.43515755543926e+00, 2.43489459326784e+00, 2.43443734200696e+00, 2.43369841558349e+00, 2.43257126094906e+00, 2.43093056245023e+00, 2.42863352946522e+00, 2.42552210063653e+00, 2.42142606907252e+00, 2.41616709652303e+00, 2.40956354097304e+00, 2.40143597275180e+00, 2.39161320191204e+00, 2.37993858846303e+00, 2.36627636233254e+00, 2.35051764759334e+00, 2.33258587126610e+00, 2.31244124555803e+00, 2.29008404635351e+00, 2.26555646996671e+00, 2.23894293123098e+00, 2.21036876243961e+00, 2.17999737550551e+00, 2.14802604865670e+00, 2.11468058381990e+00, 2.08020914295764e+00, 2.04487560531804e+00, 2.00895279085384e+00, 1.97271586992392e+00, 1.93643623119047e+00, 1.90037601613440e+00, 1.86478345853993e+00, 1.82988909885708e+00, 1.79590288304017e+00, 1.76301210739281e+00, 1.73138013672845e+00, 1.70114580226748e+00, 1.67242337619309e+00, 1.64530301907322e+00, 1.61985160178415e+00, 1.59611381287803e+00, 1.57411347380045e+00, 1.55385499677975e+00, 1.53532493278220e+00, 1.51849356910056e+00, 1.50331654750900e+00, 1.48973648409996e+00, 1.47768458058303e+00, 1.46708222365631e+00, 1.45784257380289e+00, 1.44987214734958e+00, 1.44307239581249e+00, 1.43734128453707e+00, 1.43257486867859e+00, 1.42866885906624e+00, 1.42552016399588e+00, 1.42302838615382e+00, 1.42109724739589e+00, 1.41963590871035e+00, 1.41856014903369e+00, 1.41779336520023e+00, 1.41726735653273e+00, 1.41692286155066e+00, 1.41670982085591e+00, 1.41658734909103e+00, 1.41652340938202e+00, 1.41649419516378e+00, 1.41648323596068e+00, 1.41648025479725e+00, 1.41647981478397e+00, 1.41647978121993e+00, 1.41647918336699e+00, 1.41647513282868e+00, 1.41646024267074e+00, 1.41642055135581e+00, 1.41633368980828e+00, 1.41616734692711e+00, 1.41587809833479e+00, 1.41541065765672e+00, 1.41469760306710e+00, 1.41365962361546e+00, 1.41220631919361e+00, 1.41023757409910e+00, 1.40764550630915e+00, 1.40431697243534e+00, 1.40013658203607e+00, 1.39499014537070e+00, 1.38876844740847e+00, 1.38137121036800e+00, 1.37271108028212e+00, 1.36271745340611e+00, 1.35133994894346e+00, 1.33855133814092e+00, 1.32434975775010e+00, 1.30876006807460e+00, 1.29183426051333e+00, 1.27365087326812e+00, 1.25431343208049e+00, 1.23394799027915e+00, 1.21269989393583e+00, 1.19072993920147e+00, 1.16821011684691e+00, 1.14531915211661e+00, 1.12223804619791e+00, 1.09914581024988e+00, 1.07621555639470e+00, 1.05361107541033e+00, 1.03148399147470e+00, 1.00997154360677e+00, 9.89195004578889e-01, 9.69258713680716e-01, 9.50249671766159e-01, 9.32237626688050e-01, 9.15275564896447e-01, 8.99400520279769e-01, 8.84634613303191e-01, 8.70986240779175e-01, 8.58451347632140e-01, 8.47014725216823e-01, 8.36651294692885e-01, 8.27327347456119e-01, 8.19001726772669e-01, 8.11626944943711e-01, 8.05150238182932e-01, 7.99514566779621e-01, 7.94659571075812e-01, 7.90522494468689e-01, 7.87039083320323e-01, 7.84144470655008e-01, 7.81774046257183e-01, 7.79864310704639e-01, 7.78353705478651e-01, 7.77183406100860e-01, 7.76298060753447e-01, 7.75646453508827e-01, 7.75182069492345e-01, 7.74863539283842e-01, 7.74654941721124e-01, 7.74525947937860e-01, 7.74451794707332e-01, 7.74413081600571e-01, 7.74395393619074e-01, 7.74388758304269e-01, 7.74386953343196e-01, 7.74386686934932e-01, 7.74386666470036e-01, 7.74386300062659e-01, 7.74383817601901e-01, 7.74374691859251e-01, 7.74350366331618e-01, 7.74297132294151e-01, 7.74195189582926e-01, 7.74017931753392e-01, 7.73731493430901e-01, 7.73294594220715e-01, 7.72658709001092e-01, 7.71768588174268e-01, 7.70563142852083e-01, 7.68976698444732e-01, 7.66940605382729e-01, 7.64385177788455e-01, 7.61241910387739e-01, 7.57445901958576e-01, 7.52938391881031e-01, 7.47669297094575e-01, 7.41599622479271e-01, 7.34703610822109e-01, 7.26970501171309e-01, 7.18405777792611e-01, 7.09031816263433e-01, 6.98887867294843e-01, 6.88029360133484e-01, 6.76526552213204e-01, 6.64462595737518e-01, 6.51931130574013e-01, 6.39033542224837e-01, 6.25876040771630e-01, 6.12566720150141e-01, 5.99212747148485e-01, 5.85917807995912e-01, 5.72779910427158e-01, 5.59889604512350e-01, 5.47328650306710e-01, 5.35169128042028e-01, 5.23472959843346e-01, 5.12291792418992e-01, 5.01667178337797e-01, 4.91630988924840e-01, 4.82205993322509e-01, 4.73406544344542e-01, 4.65239320758414e-01, 4.57704086057390e-01, 4.50794434386935e-01, 4.44498504170141e-01, 4.38799648558060e-01, 4.33677058823733e-01, 4.29106342145374e-01, 4.25060058949484e-01, 4.21508227249305e-01, 4.18418802397724e-01, 4.15758140562858e-01, 4.13491453202610e-01, 4.11583258039585e-01, 4.09997829683860e-01, 4.08699650301131e-01, 4.07653857762753e-01, 4.06826685755377e-01, 4.06185887586759e-01, 4.05701133129455e-01, 4.05344366700959e-01, 4.05090112866318e-01, 4.04915717287167e-01, 4.04801510877334e-01, 4.04730887619962e-01, 4.04690289336472e-01, 4.04669094266097e-01, 4.04659410267129e-01, 4.04655777497752e-01, 4.04654789299514e-01, 4.04654643443621e-01, 4.04654632153779e-01, 4.04654428905332e-01, 4.04653051868920e-01, 4.04647989764513e-01, 4.04634496260215e-01, 4.04604967109179e-01, 4.04548419287181e-01, 4.04450094774120e-01, 4.04291210104233e-01, 4.04048870941836e-01, 4.03696168442551e-01, 4.03202470658883e-01, 4.02533917362831e-01, 4.01654120038430e-01, 4.00525060252832e-01, 3.99108169183179e-01, 3.97365559121312e-01, 3.95261365035665e-01, 3.92763141847865e-01, 3.89843252397242e-01, 3.86480173681647e-01, 3.82659646377519e-01, 3.78375596012091e-01, 3.73630764038722e-01, 3.68437003194870e-01, 3.62815212741467e-01, 3.56794913513127e-01, 3.50413487586579e-01, 3.43715130039783e-01, 3.36749578210984e-01, 3.29570695224930e-01, 3.22234988434545e-01, 3.14800140000230e-01, 3.07323617221925e-01, 2.99861416281914e-01, 2.92466976886288e-01, 2.85190288994045e-01, 2.78077198076801e-01, 2.71168903279460e-01, 2.64501633949592e-01, 2.58106484254284e-01, 2.52009382629723e-01, 2.46231172053214e-01, 2.40787777986128e-01, 2.35690442757857e-01, 2.30946007689752e-01, 2.26557227059246e-01, 2.22523100859235e-01, 2.18839216089080e-01, 2.15498088941615e-01, 2.12489502676561e-01, 2.09800838129496e-01, 2.07417395639666e-01, 2.05322708624649e-01, 2.03498850035544e-01, 2.01926733456716e-01, 2.00586410663504e-01, 1.99457367033517e-01, 1.98518815372736e-01, 1.97749987546345e-01, 1.97130421909179e-01, 1.96640243039411e-01, 1.96260428848603e-01, 1.95973058914845e-01, 1.95761537009858e-01, 1.95610780375887e-01, 1.95507368435599e-01, 1.95439644314817e-01, 1.95397763805474e-01, 1.95373688118622e-01, 1.95361118860191e-01, 1.95355375958030e-01, 1.95353221612677e-01, 1.95352635579965e-01, 1.95352549082813e-01, 1.95352542333296e-01, 1.95352420122134e-01, 1.95351592124339e-01, 1.95348548330756e-01, 1.95340434801437e-01, 1.95322679083006e-01, 1.95288676835048e-01, 1.95229553139669e-01, 1.95134010976972e-01, 1.94988278146909e-01, 1.94776162356533e-01, 1.94479222074406e-01, 1.94077057883746e-01, 1.93547725259904e-01, 1.92868264864913e-01, 1.92015340616834e-01, 1.90965969151364e-01, 1.89698317230011e-01, 1.88192536744310e-01, 1.86431600935251e-01, 1.84402101098200e-01, 1.82094961131028e-01, 1.79506028405811e-01, 1.76636503906721e-01, 1.73493182302934e-01, 1.70088483128808e-01, 1.66440266655471e-01, 1.62571441227584e-01, 1.58509381556925e-01, 1.54285188528959e-01, 1.49932829544909e-01, 1.45488203692679e-01, 1.40988177924214e-01, 1.36469639076893e-01, 1.31968602428486e-01, 1.27519411083213e-01, 1.23154052479961e-01, 1.18901609388203e-01, 1.14787853666044e-01, 1.10834982543665e-01, 1.07061489880849e-01, 1.03482159088031e-01, 1.00108160285075e-01, 9.69472317430586e-02, 9.40039246135029e-02, 9.12798903321601e-02, 8.87741917572384e-02, 8.64836217855895e-02, 8.44030164757186e-02, 8.25255531537437e-02, 8.08430272531098e-02, 7.93461045678470e-02, 7.80245480931541e-02, 7.68674206455415e-02, 7.58632659528615e-02, 7.50002718354352e-02, 7.42664194768992e-02, 7.36496226404171e-02, 7.31378601169581e-02, 7.27193037858171e-02, 7.23824435369664e-02, 7.21162090580378e-02, 7.19100872366880e-02, 7.17542327921592e-02, 7.16395688261713e-02, 7.15578733702671e-02, 7.15018477548881e-02, 7.14651627830278e-02, 7.14424792251833e-02, 7.14294400399638e-02, 7.14226328765903e-02, 7.14195227339758e-02, 7.14183560293975e-02, 7.14180386590613e-02, 7.14179918158919e-02, 7.14179881295678e-02, 7.14179209836345e-02, 7.14174660607938e-02, 7.14157937309494e-02, 7.14113360223365e-02, 7.14015809889332e-02, 7.13829010514928e-02, 7.13504231131878e-02, 7.12979479295370e-02, 7.12179257956223e-02, 7.11014949639832e-02, 7.09385881699234e-02, 7.07181110393407e-02, 7.04281938355820e-02, 7.00565148845982e-02, 6.95906900788117e-02, 6.90187182555917e-02, 6.83294672196695e-02, 6.75131801710620e-02, 6.65619778343824e-02, 6.54703282835101e-02, 6.42354549089390e-02, 6.28576536778240e-02, 6.13404941013584e-02, 5.96908841595223e-02, 5.79189875882378e-02, 5.60379917989451e-02, 5.40637354136919e-02, 5.20142148539549e-02, 4.99089984188998e-02, 4.77685827413126e-02, 4.56137296920787e-02, 4.34648215789224e-02, 4.13412692432615e-02, 3.92610020864339e-02, 3.72400616834767e-02, 3.52923118923627e-02, 3.34292687012265e-02, 3.16600435830168e-02, 2.99913864505328e-02, 2.84278097979609e-02, 2.69717744490237e-02, 2.56239182186562e-02, 2.43833101707196e-02, 2.32477141229310e-02, 2.22138462942070e-02, 2.12776146229601e-02, 2.04343315056832e-02, 1.96788963578314e-02, 1.90059480579663e-02, 1.84099892661891e-02, 1.78854851085646e-02, 1.74269387066630e-02, 1.70289464903846e-02, 1.66862370621743e-02, 1.63936977229184e-02, 1.61463926063448e-02, 1.59395758422392e-02, 1.57687026413242e-02, 1.56294404235808e-02, 1.55176812944404e-02, 1.54295564389173e-02, 1.53614523671450e-02, 1.53100284270806e-02, 1.52722345219587e-02, 1.52453275568405e-02, 1.52268849117974e-02, 1.52148131774084e-02, 1.52073505245909e-02, 1.52030613579512e-02, 1.52008223520852e-02, 1.51997994065596e-02, 1.51994156754824e-02, 1.51993112931024e-02, 1.51992958866458e-02, 1.51992946635318e-02, 1.51992722467469e-02, 1.51991203755131e-02, 1.51985620872637e-02, 1.51970739704410e-02, 1.51938176346533e-02, 1.51875827855719e-02, 1.51767447810300e-02, 1.51592396300941e-02, 1.51325594048757e-02, 1.50937709719443e-02, 1.50395606476792e-02, 1.49663067931272e-02, 1.48701813100445e-02, 1.47472795240162e-02, 1.45937759099013e-02, 1.44061007522822e-02, 1.41811301490975e-02, 1.39163793071049e-02, 1.36101869138514e-02, 1.32618770117148e-02, 1.28718840343396e-02, 1.24418272654623e-02, 1.19745239468045e-02, 1.14739361017797e-02, 1.09450526360320e-02, 1.03937124286944e-02, 9.82637881172366e-03, 9.24988325999720e-03, 8.66811713410698e-03, 8.09437998265139e-03, 7.53145405057865e-03, 6.98502912982235e-03, 6.46000516536560e-03, 5.96039045350949e-03, 5.48925626628716e-03, 5.04874353622142e-03, 4.64011363141026e-03, 4.26383306148637e-03, 3.91968112410268e-03, 3.60686979786351e-03, 3.32416633467858e-03, 3.07001075424296e-03, 2.84262255566148e-03, 2.64009314510361e-03, 2.46046243384512e-03, 2.30177956080979e-03, 2.16214868574127e-03, 2.03976138382700e-03, 1.93291749990010e-03, 1.84003648872988e-03, 1.75966129809524e-03, 1.69045673887069e-03, 1.63194916336806e-03, 1.58178101742124e-03, 1.53927164784307e-03, 1.50351345523347e-03, 1.47369908965800e-03, 1.44865092906343e-03, 1.42895806122567e-03, 1.41387552747574e-03, 1.40167447176509e-03, 1.39224520432890e-03, 1.38512030430604e-03, 1.37987867829281e-03, 1.37614313911172e-03, 1.37358041435220e-03, 1.37190179243873e-03, 1.37086356883742e-03, 1.37026666102713e-03, 1.36995501011938e-03, 1.36981261131552e-03, 1.36975919214422e-03, 1.36974466072104e-03, 1.36974251604727e-03, 1.36974234622755e-03, 1.36973923856942e-03, 1.36971818498921e-03, 1.36964078919277e-03, 1.36943447684246e-03, 1.36898295183130e-03, 1.36811817549187e-03, 1.36661419909956e-03, 1.36418321282861e-03, 1.36047430078179e-03, 1.35507563132364e-03, 1.34752104763436e-03, 1.33730208276006e-03, 1.32388618490609e-03, 1.30674139292353e-03, 1.28536678030855e-03, 1.25932676581897e-03, 1.22793374268326e-03, 1.19282111161567e-03, 1.15230433502234e-03, 1.10661189178529e-03, 1.05603912444963e-03, 1.00109385263636e-03, 9.42953740121766e-04, 8.82337171475456e-04, 8.19658254528122e-04, 7.55953478872584e-04, 6.92605528385570e-04, 6.30410181543084e-04, 5.69979264597009e-04, 5.12273821485110e-04, 4.57964603256755e-04, 4.07289784362774e-04, 3.60720878546772e-04, 3.18375174686025e-04, 2.80093297305939e-04, 2.46083423775190e-04, 2.16003240510076e-04, 1.89353992340102e-04, 1.66301534207491e-04, 1.46409789607560e-04, 1.28973273987754e-04, 1.13962253038835e-04, 1.01502416256949e-04, 9.06257264574606e-05, 8.11698231899527e-05, 7.34167206246639e-05, 6.68420985334706e-05, 6.10945983652300e-05, 5.62957745131598e-05, 5.24322283581569e-05, 4.90916774111838e-05, 4.60977639988869e-05, 4.34900695012175e-05, 4.16088846151297e-05, 4.02037725706350e-05, 3.89928711683751e-05, 3.79217860171366e-05, 3.69821969030681e-05, 3.61769009344926e-05, 3.55075421367701e-05, 3.49692494836518e-05, 3.45501624110511e-05, 3.42338881722526e-05, 3.40025325868965e-05, 3.38388507117315e-05, 3.37273555146567e-05, 3.36547583827951e-05, 3.36100545439422e-05, 3.35844273706698e-05, 3.35710699346237e-05, 3.35649720284437e-05, 3.35626853740223e-05, 3.35620634075440e-05, 3.35619716054225e-05, 3.35619643863083e-05, 3.35618326658960e-05, 3.35609401815873e-05, 3.35576588668118e-05, 3.35489080924845e-05, 3.35297371236720e-05, 3.34929463576827e-05, 3.34287343974893e-05, 3.33243529979654e-05, 3.31637754413805e-05, 3.29274961785586e-05, 3.25929826061611e-05, 3.21380180849928e-05, 3.15583146743847e-05, 3.09278223085130e-05, 3.03151169447347e-05, 2.96046105461301e-05, 2.87041494999585e-05, 2.75643238865086e-05, 2.61557987926153e-05, 2.44777969271581e-05, 2.25979578090119e-05, 2.08066722819922e-05, 1.92044899165467e-05, 1.75053527158298e-05, 1.57249154284500e-05, 1.40273442497119e-05, 1.24804513035315e-05, 1.10514191504762e-05, 9.87526848916787e-06, 8.85684768547890e-06, 7.84267108144832e-06, 7.22691284238408e-06, 6.96817960488261e-06, 6.37878102338618e-06, 6.37173886275377e-06, 6.52639501197427e-06, 6.21082537577422e-06, 5.37727018468104e-06, 4.15860010605229e-06, 2.79742639008873e-06, 1.54231095450541e-06, 1.05506521208554e-06, 1.12949901165393e-06, 8.28425285703815e-07, 1.11742330943798e-06, 1.87938438500623e-06, 2.41571731862678e-06, 2.55006291638438e-06, 2.29412388819128e-06, 1.76202318868945e-06, 1.11905615085561e-06, 6.56792906283954e-07, 7.09531519760276e-07, 7.19932685910862e-07, 6.49890466775105e-07, 5.53656036572647e-07, 4.66921179054225e-07, 4.08178323141328e-07, 3.83198994313716e-07, 3.80381791842410e-07, 3.73032776224136e-07, 3.47461920569003e-07, -4.26325641456060e-14, -5.68434188608080e-14, -4.26325641456060e-14, -5.68434188608080e-14, -4.26325641456060e-14, -5.68434188608080e-14, -4.26325641456060e-14, -5.68434188608080e-14, -5.68434188608080e-14, -5.68434188608080e-14, -4.26325641456060e-14, -4.26325641456060e-14],
[ 1.15000000000000e+02, 1.14999999848049e+02, 1.14999995149411e+02, 1.14999963315393e+02, 1.14999846289319e+02, 1.14999534334197e+02, 1.14998851594677e+02, 1.14997543938981e+02, 1.14995269514828e+02, 1.14991592367495e+02, 1.14985979371692e+02, 1.14977800626500e+02, 1.14966333360071e+02, 1.14950769294173e+02, 1.14930225334245e+02, 1.14903757383123e+02, 1.14870377028335e+02, 1.14829070822238e+02, 1.14778821854950e+02, 1.14718633300860e+02, 1.14647553586008e+02, 1.14564702761135e+02, 1.14469299562813e+02, 1.14360688501114e+02, 1.14238366137550e+02, 1.14102005536866e+02, 1.13951477727794e+02, 1.13786868933795e+02, 1.13608492373705e+02, 1.13416893607937e+02, 1.13212848719407e+02, 1.12997355043682e+02, 1.12771614650534e+02, 1.12537011264967e+02, 1.12295081734324e+02, 1.12047483446001e+02, 1.11795959246731e+02, 1.11542301405892e+02, 1.11288316023506e+02, 1.11035789048223e+02, 1.10786454789064e+02, 1.10541967521971e+02, 1.10303876543057e+02, 1.10073604825243e+02, 1.09852431300029e+02, 1.09641476705843e+02, 1.09441692906218e+02, 1.09253855569545e+02, 1.09078560102929e+02, 1.08916220734592e+02, 1.08767072635054e+02, 1.08631176953773e+02, 1.08508428624541e+02, 1.08398566761160e+02, 1.08301187427147e+02, 1.08215758521936e+02, 1.08141636483618e+02, 1.08078084466957e+02, 1.08024291616992e+02, 1.07979393024985e+02, 1.07942489926509e+02, 1.07912669682981e+02, 1.07889025079738e+02, 1.07870672477521e+02, 1.07856768371341e+02, 1.07846523942098e+02, 1.07839217232325e+02, 1.07834202637501e+02, 1.07830917477303e+02, 1.07828885494840e+02, 1.07827717223403e+02, 1.07827107256260e+02, 1.07826828551716e+02, 1.07826723999127e+02, 1.07826695558203e+02, 1.07826691360380e+02, 1.07826690809714e+02, 1.07826677979858e+02, 1.07826591055692e+02, 1.07826271513123e+02, 1.07825419728888e+02, 1.07823555614523e+02, 1.07819985591785e+02, 1.07813777227344e+02, 1.07803742685039e+02, 1.07788431968661e+02, 1.07766136712251e+02, 1.07734905021661e+02, 1.07692567573485e+02, 1.07636774829755e+02, 1.07565044827841e+02, 1.07474820560256e+02, 1.07363535483729e+02, 1.07228685217617e+02, 1.07067903046903e+02, 1.06879036483027e+02, 1.06660221908004e+02, 1.06409954277721e+02, 1.06127149009150e+02, 1.05811193505375e+02, 1.05461986217972e+02, 1.05079961612765e+02, 1.04666099806712e+02, 1.04221919969257e+02, 1.03749456949951e+02, 1.03251221244029e+02, 1.02730143568636e+02, 1.02189506980896e+02, 1.01632871177785e+02, 1.01063994580808e+02, 1.00486759340846e+02, 9.99051024924157e+01, 9.93229539683418e+01, 9.87441802248608e+01, 9.81725315759142e+01, 9.76115919380015e+01, 9.70647308968092e+01, 9.65350591016377e+01, 9.60253885528739e+01, 9.55381993219641e+01, 9.50756138171760e+01, 9.46393791268616e+01, 9.42308574277908e+01, 9.38510240429009e+01, 9.35004724956898e+01, 9.31794258136257e+01, 9.28877533357612e+01, 9.26249923344417e+01, 9.23903738308173e+01, 9.21828520441402e+01, 9.20011369521199e+01, 9.18437294490697e+01, 9.17089585712221e+01, 9.15950202190673e+01, 9.15000167516326e+01, 9.14219967652064e+01, 9.13589943076168e+01, 9.13090667274127e+01, 9.12703303234253e+01, 9.12409929516020e+01, 9.12193827686859e+01, 9.12039723502655e+01, 9.11933975155103e+01, 9.11864703213957e+01, 9.11821858514840e+01, 9.11797226118817e+01, 9.11784365513015e+01, 9.11778489333167e+01, 9.11776284965072e+01, 9.11775685322542e+01, 9.11775596816526e+01, 9.11775589003130e+01, 9.11775435903474e+01, 9.11774398630714e+01, 9.11770585505306e+01, 9.11760421105156e+01, 9.11738176524446e+01, 9.11695575339467e+01, 9.11621491046193e+01, 9.11501749855794e+01, 9.11319050552424e+01, 9.11053010576833e+01, 9.10680344521653e+01, 9.10175177727300e+01, 9.09509493588899e+01, 9.08653708508330e+01, 9.07577363207878e+01, 9.06249913513433e+01, 9.04641597973495e+01, 9.02724354169964e+01, 9.00472750754058e+01, 8.97864898619720e+01, 8.94883302729139e+01, 8.91515616392671e+01, 8.87755262601400e+01, 8.83601892410825e+01, 8.79061658165141e+01, 8.74147288956974e+01, 8.68877966210677e+01, 8.63279007498838e+01, 8.57381375496714e+01, 8.51221035515162e+01, 8.44838189104679e+01, 8.38276413267558e+01, 8.31581735802702e+01, 8.24801678134091e+01, 8.17984297878927e+01, 8.11177263686364e+01, 8.04426993178367e+01, 7.97777880028369e+01, 7.91271628313881e+01, 7.84946702674888e+01, 7.78837893722848e+01, 7.72975991537486e+01, 7.67387556799855e+01, 7.62094778712455e+01, 7.57115410163739e+01, 7.52462772368280e+01, 7.48145822649045e+01, 7.44169279855577e+01, 7.40533802238664e+01, 7.37236212676147e+01, 7.34269766177812e+01, 7.31624454689810e+01, 7.29287344364025e+01, 7.27242940583216e+01, 7.25473576050714e+01, 7.23959817094092e+01, 7.22680882964499e+01, 7.21615072348800e+01, 7.20740190601779e+01, 7.20033970432461e+01, 7.19474478043962e+01, 7.19040496140078e+01, 7.18711874880514e+01, 7.18469841881810e+01, 7.18297262790124e+01, 7.18178844830463e+01, 7.18101277063276e+01, 7.18053302814467e+01, 7.18025721814252e+01, 7.18011321880858e+01, 7.18004742394531e+01, 7.18002274195940e+01, 7.18001602785501e+01, 7.18001503686705e+01, 7.18001495737756e+01, 7.18001349039327e+01, 7.18000355135907e+01, 7.17996701442077e+01, 7.17986962037942e+01, 7.17965647601179e+01, 7.17924827946929e+01, 7.17853842357210e+01, 7.17739111125474e+01, 7.17564059698898e+01, 7.17309164410224e+01, 7.16952125968302e+01, 7.16468173534276e+01, 7.15830498281480e+01, 7.15010810802340e+01, 7.13980011640142e+01, 7.12708958737941e+01, 7.11169309968091e+01, 7.09334413493476e+01, 7.07180213955422e+01, 7.04686138863420e+01, 7.01835927543123e+01, 6.98618364969831e+01, 6.95027885030856e+01, 6.91065012305109e+01, 6.86736618215678e+01, 6.82055976106448e+01, 6.77042609945239e+01, 6.71721942322911e+01, 6.66124758396965e+01, 6.60286512475089e+01, 6.54246512033715e+01, 6.48047019177574e+01, 6.41732311210820e+01, 6.35347739952722e+01, 6.28938824168405e+01, 6.22550402066833e+01, 6.16225862663850e+01, 6.10006467308809e+01, 6.03930766809839e+01, 5.98034115701794e+01, 5.92348282982558e+01, 5.86901157400025e+01, 5.81716544361194e+01, 5.76814050277579e+01, 5.72209048579785e+01, 5.67912719987478e+01, 5.63932158307342e+01, 5.60270532375739e+01, 5.56927294885157e+01, 5.53898429634875e+01, 5.51176729978388e+01, 5.48752102607420e+01, 5.46611892054604e+01, 5.44741222233574e+01, 5.43123351873350e+01, 5.41740040824723e+01, 5.40571923955045e+01, 5.39598888775755e+01, 5.38800452158845e+01, 5.38156130603131e+01, 5.37645797625747e+01, 5.37250021095967e+01, 5.36950372808157e+01, 5.36729702403625e+01, 5.36572367968426e+01, 5.36464416295874e+01, 5.36393706911740e+01, 5.36349975482366e+01, 5.36324834090537e+01, 5.36311707971100e+01, 5.36305710527953e+01, 5.36303460678232e+01, 5.36302848664393e+01, 5.36302758332427e+01, 5.36302751365156e+01, 5.36302626255346e+01, 5.36301778618087e+01, 5.36298662615535e+01, 5.36290356509671e+01, 5.36272178865597e+01, 5.36237366753386e+01, 5.36176828983116e+01, 5.36078985960035e+01, 5.35929706058024e+01, 5.35712346417918e+01, 5.35407903699214e+01, 5.34995277461625e+01, 5.34451645457738e+01, 5.33752946154462e+01, 5.32874459311517e+01, 5.31791470560739e+01, 5.30480000880590e+01, 5.28917576974656e+01, 5.27084014252197e+01, 5.24962180833575e+01, 5.22538709225144e+01, 5.19804622428644e+01, 5.16755843542661e+01, 5.13393562456606e+01, 5.09724439857919e+01, 5.05760637019042e+01, 5.01519668990434e+01, 4.97024088014068e+01, 4.92301012284654e+01, 4.87381521899126e+01, 4.82299948576163e+01, 4.77093088539932e+01, 4.71799369207230e+01, 4.66458000443413e+01, 4.61108140441414e+01, 4.55788104706243e+01, 4.50534643921723e+01, 4.45382312334601e+01, 4.40362942640966e+01, 4.35505236519146e+01, 4.30834472616431e+01, 4.26372326862490e+01, 4.22136794306698e+01, 4.18142197856974e+01, 4.14399267555180e+01, 4.10915274205577e+01, 4.07694202847998e+01, 4.04736954169463e+01, 4.02041564910173e+01, 3.99603441173695e+01, 3.97415600975238e+01, 3.95468924185060e+01, 3.93752409199993e+01, 3.92253436244303e+01, 3.90958037249698e+01, 3.89851171900445e+01, 3.88917008762333e+01, 3.88139209546319e+01, 3.87501213584376e+01, 3.86986518605901e+01, 3.86578952983847e+01, 3.86262933852830e+01, 3.86023704963482e+01, 3.85847547894948e+01, 3.85721960351187e+01, 3.85635795745338e+01, 3.85579359131326e+01, 3.85544455745116e+01, 3.85524389912305e+01, 3.85513913782226e+01, 3.85509127161682e+01, 3.85507331536650e+01, 3.85506843083254e+01, 3.85506770988561e+01, 3.85506765529769e+01, 3.85506668827809e+01, 3.85506013657983e+01, 3.85503605187095e+01, 3.85497185108711e+01, 3.85483135029394e+01, 3.85456227849698e+01, 3.85409437299235e+01, 3.85333814713390e+01, 3.85218440882244e+01, 3.85050459299864e+01, 3.84815195313873e+01, 3.84496363435505e+01, 3.84076362348845e+01, 3.83536653922789e+01, 3.82858218810996e+01, 3.82022077136582e+01, 3.81009858508376e+01, 3.79804401509474e+01, 3.78390359224370e+01, 3.76754784765564e+01, 3.74887669561556e+01, 3.72782407747301e+01, 3.70436162589093e+01, 3.67850115503940e+01, 3.65029584662514e+01, 3.61984007877713e+01, 3.58726792708780e+01, 3.55275044522528e+01, 3.51649189702285e+01, 3.47872515508034e+01, 3.43970649855253e+01, 3.39971003577276e+01, 3.35902195166511e+01, 3.31793474544196e+01, 3.27674159195654e+01, 3.23573093900622e+01, 3.19518144600737e+01, 3.15535737288910e+01, 3.11650453214825e+01, 3.07884691044672e+01, 3.04258404117175e+01, 3.00788916588055e+01, 2.97490816852177e+01, 2.94375921455866e+01, 2.91453298969135e+01, 2.88729341646837e+01, 2.86207873130765e+01, 2.83890282343794e+01, 2.81775676295975e+01, 2.79861047046524e+01, 2.78141450092023e+01, 2.76610192806380e+01, 2.75259032270430e+01, 2.74078382027155e+01, 2.73057527131122e+01, 2.72184846453783e+01, 2.71448040651367e+01, 2.70834363563170e+01, 2.70330854134469e+01, 2.69924565295858e+01, 2.69602785628419e+01, 2.69353249154931e+01, 2.69164328275320e+01, 2.69025204759780e+01, 2.68926013864416e+01, 2.68857957064382e+01, 2.68813379609878e+01, 2.68785810079520e+01, 2.68769960289935e+01, 2.68761685257636e+01, 2.68757904323893e+01, 2.68756485964605e+01, 2.68756100136441e+01, 2.68756043189010e+01, 2.68756038913948e+01, 2.68755963668019e+01, 2.68755453865830e+01, 2.68753579779904e+01, 2.68748584161412e+01, 2.68737651405594e+01, 2.68716713996958e+01, 2.68680304063823e+01, 2.68621456994875e+01, 2.68531673047089e+01, 2.68400941667463e+01, 2.68217831809817e+01, 2.67969649799132e+01, 2.67642664237353e+01, 2.67222395041592e+01, 2.66693960986639e+01, 2.66042477174450e+01, 2.65253490824588e+01, 2.64313440887334e+01, 2.63210124493771e+01, 2.61933151470287e+01, 2.60474367341956e+01, 2.58828225651516e+01, 2.56992092134147e+01, 2.54966466258984e+01, 2.52755109642150e+01, 2.50365075458964e+01, 2.47806637748631e+01, 2.45093123942991e+01, 2.42240657724588e+01, 2.39267822305369e+01, 2.36195256510570e+01, 2.33045197878553e+01, 2.29840988557944e+01, 2.26606561157855e+01, 2.23365922729294e+01, 2.20142655401290e+01, 2.16959451502716e+01, 2.13837699033842e+01, 2.10797130116079e+01, 2.07855540822211e+01, 2.05028586062567e+01, 2.02329648555636e+01, 1.99769776881289e+01, 1.97357684589264e+01, 1.95099800501260e+01, 1.93000359688680e+01, 1.91061524960380e+01, 1.89283529792673e+01, 1.87664835184435e+01, 1.86202294650329e+01, 1.84891323252507e+01, 1.83726068056461e+01, 1.82699578582580e+01, 1.81803976666221e+01, 1.81030625630874e+01, 1.80370298844389e+01, 1.79813347609826e+01, 1.79349867995169e+01, 1.78969865693515e+01, 1.78663417396167e+01, 1.78420826527256e+01, 1.78232770601402e+01, 1.78090436993204e+01, 1.77985643607751e+01, 1.77910940860680e+01, 1.77859691542475e+01, 1.77826125561842e+01, 1.77805367222162e+01, 1.77793433547138e+01, 1.77787203182944e+01, 1.77784356498554e+01, 1.77783288612272e+01, 1.77782998121624e+01, 1.77782955245824e+01, 1.77782952036912e+01, 1.77782895686811e+01, 1.77782513906770e+01, 1.77781110445761e+01, 1.77777369355541e+01, 1.77769182180442e+01, 1.77753503179709e+01, 1.77726238577269e+01, 1.77682175252814e+01, 1.77614953603411e+01, 1.77517088474572e+01, 1.77380041029003e+01, 1.77194343132017e+01, 1.76949774237174e+01, 1.76635588830148e+01, 1.76240790247439e+01, 1.75754444192914e+01, 1.75166022649223e+01, 1.74465766299141e+01, 1.73645051258450e+01, 1.72696744132824e+01, 1.71615528407251e+01, 1.70398185190997e+01, 1.69043812542915e+01, 1.67553970059056e+01, 1.65932739055086e+01, 1.64186693311131e+01, 1.62324780613688e+01, 1.60358120759421e+01, 1.58299730746737e+01, 1.56164192051160e+01, 1.53967277731342e+01, 1.51725558383928e+01, 1.49456005613789e+01, 1.47175609890224e+01, 1.44901026780301e+01, 1.42648262052067e+01, 1.40432402495312e+01, 1.38267395909863e+01, 1.36165880821487e+01, 1.34139064217722e+01, 1.32196643944547e+01, 1.30346771292234e+01, 1.28596048625910e+01, 1.26949556597554e+01, 1.25410905452067e+01, 1.23982305171703e+01, 1.22664649656534e+01, 1.21457610771103e+01, 1.20359738841427e+01, 1.19368566992636e+01, 1.18480717503222e+01, 1.17692009051699e+01, 1.16997564294670e+01, 1.16391917609971e+01, 1.15869123051486e+01, 1.15422862597041e+01, 1.15046554644504e+01, 1.14733462451444e+01, 1.14476801855740e+01, 1.14269847199528e+01, 1.14106033951189e+01, 1.13979056125623e+01, 1.13882956285631e+01, 1.13812205706175e+01, 1.13761772229250e+01, 1.13727173449513e+01, 1.13704513155135e+01, 1.13690499395105e+01, 1.13682443129486e+01, 1.13678237106716e+01, 1.13676315357051e+01, 1.13675594445136e+01, 1.13675398339874e+01, 1.13675369395157e+01, 1.13675367228065e+01, 1.13675329162007e+01, 1.13675071259013e+01, 1.13674123182520e+01, 1.13671595975550e+01, 1.13666065330531e+01, 1.13655473811767e+01, 1.13637056098426e+01, 1.13607290933528e+01, 1.13561883012243e+01, 1.13495777489618e+01, 1.13403209093274e+01, 1.13277786947497e+01, 1.13112615115746e+01, 1.12900447530817e+01, 1.12633874415832e+01, 1.12305535552378e+01, 1.11908353914371e+01, 1.11435781389767e+01, 1.10882046724419e+01, 1.10242394630605e+01, 1.09513304392999e+01, 1.08692676435009e+01, 1.07779976280774e+01, 1.06776327185489e+01, 1.05684545336041e+01, 1.04509114772286e+01, 1.03256102783251e+01, 1.01933020162991e+01, 1.00548634010340e+01, 9.91127433882575e+00, 9.76359298568195e+00, 9.61292955135481e+00, 9.46042007190130e+00, 9.30720123081929e+00, 9.15438710606341e+00, 9.00304848548799e+00, 8.85419515816560e+00, 8.70876137793992e+00, 8.56759452233136e+00, 8.43144683812667e+00, 8.30097007091165e+00, 8.17671271232044e+00, 8.05911955875951e+00, 7.94853325362396e+00, 7.84519747862464e+00, 7.74926146735120e+00, 7.66078553457605e+00, 7.57974734650303e+00, 7.50604869766846e+00, 7.43952260609670e+00, 7.37994058564102e+00, 7.32701999921612e+00, 7.28043143529720e+00, 7.23980607985098e+00, 7.20474307498336e+00, 7.17481686330309e+00, 7.14958451359765e+00, 7.12859301012696e+00, 7.11138646670348e+00, 7.09751320041539e+00, 7.08653257149103e+00, 7.07802146872471e+00, 7.07158029736213e+00, 7.06683831132432e+00, 7.06345812646869e+00, 7.06113925774156e+00, 7.05962054105679e+00, 7.05868132995290e+00, 7.05814139591537e+00, 7.05785950722462e+00, 7.05773071121777e+00, 7.05768239559360e+00, 7.05766925259161e+00, 7.05766731271275e+00, 7.05766716708867e+00, 7.05766460398246e+00, 7.05764723857175e+00, 7.05758340164826e+00, 7.05741323710217e+00, 7.05704084251919e+00, 7.05632768673023e+00, 7.05508757915018e+00, 7.05308344054522e+00, 7.05002609502802e+00, 7.04557526643680e+00, 7.03934291598506e+00, 7.03089899921315e+00, 7.01977964636063e+00, 7.00549768006105e+00, 6.98755527841739e+00, 6.96545847323588e+00, 6.93873304851174e+00, 6.90694128200936e+00, 6.86969886409014e+00, 6.82669124521392e+00, 6.77768861900135e+00, 6.72255875170740e+00, 6.66127692807518e+00, 6.59393239937307e+00, 6.52073088756450e+00, 6.44199290962877e+00, 6.35814792246433e+00, 6.26972453237300e+00, 6.17733724313847e+00, 6.08167041324060e+00, 5.98346023887892e+00, 5.88347566319503e+00, 5.78249912764143e+00, 5.68130803001854e+00, 5.58065764323999e+00, 5.48126609308181e+00, 5.38380180964080e+00, 5.28887367514847e+00, 5.19702390836869e+00, 5.10872356804430e+00, 5.02437043499090e+00, 4.94428894931626e+00, 4.86873183562263e+00, 4.79788304061682e+00, 4.73186162733677e+00, 4.67072631007392e+00, 4.61448036599015e+00, 4.56307671633073e+00, 4.51642302641214e+00, 4.47438672515818e+00, 4.43679988923751e+00, 4.40346397235243e+00, 4.37415438633739e+00, 4.34862495745674e+00, 4.32661228908121e+00, 4.30784006152335e+00, 4.29202329223283e+00, 4.27887256605902e+00, 4.26809822737860e+00, 4.25941450529561e+00, 4.25254352176880e+00, 4.24721911242369e+00, 4.24319037296742e+00, 4.24022483236375e+00, 4.23811114872593e+00, 4.23666122621646e+00, 4.23571166145408e+00, 4.23512444566606e+00, 4.23478687305987e+00, 4.23461063500531e+00, 4.23453011158544e+00, 4.23449990466528e+00, 4.23449168766898e+00, 4.23449047485819e+00, 4.23449038342635e+00, 4.23448876898453e+00, 4.23447783091331e+00, 4.23443762155267e+00, 4.23433043928391e+00, 4.23409587921245e+00, 4.23364669046448e+00, 4.23286561583383e+00, 4.23160337513040e+00, 4.22967793634312e+00, 4.22687519693390e+00, 4.22295116902630e+00, 4.21763572460713e+00, 4.21063790813876e+00, 4.20165276309396e+00, 4.19036954624384e+00, 4.17648112133604e+00, 4.15969423659691e+00, 4.13974030498525e+00, 4.11638623083269e+00, 4.08944477097319e+00, 4.05878389203944e+00, 4.02433459601244e+00, 3.98609673793844e+00, 3.94414245317159e+00, 3.89861694165491e+00, 3.84973651365742e+00, 3.79778397100033e+00, 3.74310156374891e+00, 3.68608190823747e+00, 3.62715736414670e+00, 3.56678843662901e+00, 3.50545179032922e+00, 3.44362843756862e+00, 3.38179259995310e+00, 3.32040165158791e+00, 3.25988744470095e+00, 3.20064920615199e+00, 3.14304808564937e+00, 3.08740334071840e+00, 3.03399006425303e+00, 2.98303830029791e+00, 2.93473335321025e+00, 2.88921707380231e+00, 2.84658990164075e+00, 2.80691345272839e+00, 2.77021346307590e+00, 2.73648292764214e+00, 2.70568530724400e+00, 2.67775771007217e+00, 2.65261398669048e+00, 2.63014770580925e+00, 2.61023500135187e+00, 2.59273729866517e+00, 2.57750393895661e+00, 2.56437472637909e+00, 2.55318242211668e+00, 2.54375520506302e+00, 2.53591911009407e+00, 2.52950044352916e+00, 2.52432816226798e+00, 2.52023618949575e+00, 2.51706562701588e+00, 2.51466681342681e+00, 2.51290116962232e+00, 2.51164276937450e+00, 2.51077957365331e+00, 2.51021427307292e+00, 2.50986469322667e+00, 2.50966373206057e+00, 2.50955881585631e+00, 2.50951087960610e+00, 2.50949289719797e+00, 2.50948800555963e+00, 2.50948728356451e+00, 2.50948722883830e+00, 2.50948625859016e+00, 2.50947968502300e+00, 2.50945551999632e+00, 2.50939110568764e+00, 2.50925014050758e+00, 2.50898019049377e+00, 2.50851079262260e+00, 2.50775224830123e+00, 2.50659519543551e+00, 2.50491103478497e+00, 2.50255326936406e+00, 2.49935979288280e+00, 2.49515613332247e+00, 2.48975961996260e+00, 2.48298439671288e+00, 2.47464715289823e+00, 2.46457338758394e+00, 2.45260396943797e+00, 2.43860170652766e+00, 2.42245760552835e+00, 2.40409648365892e+00, 2.38348160427837e+00, 2.36061804146246e+00, 2.33555454016866e+00, 2.30838372358510e+00, 2.27924060142576e+00, 2.24829944301234e+00, 2.21576918606841e+00, 2.18188764518888e+00, 2.14691485334714e+00, 2.11112590889627e+00, 2.07480370657269e+00, 2.03823190553592e+00, 2.00168843584272e+00, 1.96543977507763e+00, 1.92973614854314e+00, 1.89480772854000e+00, 1.86086183833442e+00, 1.82808110950005e+00, 1.79662249990211e+00, 1.76661705379334e+00, 1.73817027371788e+00, 1.71136297356732e+00, 1.68625249023560e+00, 1.66287414502897e+00, 1.64124286285738e+00, 1.62135487533037e+00, 1.60318945176856e+00, 1.58671061881527e+00, 1.57186884411887e+00, 1.55860267202141e+00, 1.54684030907796e+00, 1.53650116441359e+00, 1.52749735437342e+00, 1.51973518270177e+00, 1.51311660675303e+00, 1.50754069724736e+00, 1.50290509419123e+00, 1.49910745525150e+00, 1.49604688566173e+00, 1.49362533128843e+00, 1.49174890948748e+00, 1.49032914651964e+00, 1.48928408621221e+00, 1.48853923276759e+00, 1.48802829147634e+00, 1.48769367471940e+00, 1.48748674692477e+00, 1.48736779071834e+00, 1.48730568680959e+00, 1.48727731146629e+00, 1.48726666696749e+00, 1.48726377141254e+00, 1.48726334403491e+00, 1.48726331143473e+00, 1.48726273075013e+00, 1.48725879653020e+00, 1.48724433397922e+00, 1.48720578257064e+00, 1.48712141600080e+00, 1.48695985252846e+00, 1.48667891920942e+00, 1.48622492852978e+00, 1.48553241991381e+00, 1.48452441088013e+00, 1.48311319239441e+00, 1.48120168932320e+00, 1.47868538906268e+00, 1.47545481897500e+00, 1.47139852634993e+00, 1.46640648410069e+00, 1.46037381304077e+00, 1.45320467996260e+00, 1.44481620313050e+00, 1.43514217684823e+00, 1.42413641797549e+00, 1.41177554245155e+00, 1.39806100055539e+00, 1.38302023557614e+00, 1.36670687962987e+00, 1.34919995855508e+00, 1.33060213979604e+00, 1.31103711692910e+00, 1.29064627620407e+00, 1.26958482938163e+00, 1.24801762013793e+00, 1.22611481722937e+00, 1.20404769728344e+00, 1.18198469596467e+00, 1.16008787195756e+00, 1.13850988787101e+00, 1.11739156996717e+00, 1.09686006827042e+00, 1.07702760308558e+00, 1.05799075528309e+00, 1.03983023696836e+00, 1.02261106652128e+00, 1.00638306690480e+00, 9.91181607492933e-01, 9.77028516019487e-01, 9.63933097048681e-01, 9.51893205115610e-01, 9.40896333066092e-01, 9.30920688079283e-01, 9.21936238607486e-01, 9.13905724509220e-01, 9.06785629704984e-01, 9.00527121660190e-01, 8.95076964936621e-01, 8.90378417088513e-01, 8.86372114513591e-01, 8.82996953753207e-01, 8.80190970459040e-01, 8.77892214135173e-01, 8.76039612188322e-01, 8.74573812167682e-01, 8.73437986759896e-01, 8.72578582526842e-01, 8.71945990904351e-01, 8.71495118917082e-01, 8.71185837609261e-01, 8.70983288410045e-01, 8.70858031473247e-01, 8.70786025233528e-01, 8.70748432665025e-01, 8.70731256578921e-01, 8.70724813279580e-01, 8.70723060550006e-01, 8.70722801850960e-01, 8.70722781978259e-01, 8.70722426173813e-01, 8.70720015548784e-01, 8.70711153872406e-01, 8.70687532175111e-01, 8.70635838080730e-01, 8.70536843117506e-01, 8.70364706978847e-01, 8.70086534669042e-01, 8.69662218808301e-01, 8.69044594673568e-01, 8.68179929289241e-01, 8.67008757488492e-01, 8.65467066859196e-01, 8.63487819596088e-01, 8.61002782581778e-01, 8.57944618055754e-01, 8.54249167086849e-01, 8.49857838362510e-01, 8.44719997611378e-01, 8.38795240587075e-01, 8.32055427199492e-01, 8.24486357896376e-01, 8.16088986780599e-01, 8.06880089107651e-01, 7.96892332387671e-01, 7.86173737746068e-01, 7.74786557944935e-01, 7.62805636508190e-01, 7.50316344768322e-01, 7.37412217081896e-01, 7.24192416811476e-01, 7.10759166269597e-01, 6.97215263473282e-01, 6.83661789356904e-01, 6.70196083988372e-01, 6.56910042633029e-01, 6.43888755352037e-01, 6.31209489762369e-01, 6.18940997348531e-01, 6.07143110109831e-01, 5.95866586336911e-01, 5.85153161300497e-01, 5.75035759578327e-01, 5.65538829487764e-01, 5.56678765520360e-01, 5.48464390856310e-01, 5.40897478267865e-01, 5.33973293523104e-01, 5.27681150497941e-01, 5.22004971455218e-01, 5.16923849309421e-01, 5.12412611174952e-01, 5.08442384117700e-01, 5.04981164838327e-01, 5.01994395053586e-01, 4.99445543669069e-01, 4.97296695530196e-01, 4.95509144705423e-01, 4.94043988038882e-01, 4.92862712282118e-01, 4.91927765687436e-01, 4.91203102737586e-01, 4.90654688932423e-01, 4.90250951456389e-01, 4.89963161286042e-01, 4.89765732972430e-01, 4.89636429982170e-01, 4.89556466058289e-01, 4.89510496425609e-01, 4.89486496622689e-01, 4.89475531025974e-01, 4.89471417470099e-01, 4.89470298484889e-01, 4.89470133325057e-01, 4.89470120541014e-01, 4.89469890392698e-01, 4.89468331105314e-01, 4.89462599017500e-01, 4.89447319511086e-01, 4.89413881370794e-01, 4.89349845959495e-01, 4.89238496063265e-01, 4.89058547094942e-01, 4.88784040762127e-01, 4.88384438079649e-01, 4.87824924449512e-01, 4.87066934129516e-01, 4.86068894506346e-01, 4.84787182039526e-01, 4.83177271619994e-01, 4.81195049722302e-01, 4.78798249779274e-01, 4.75947956626896e-01, 4.72610116859997e-01, 4.68756984835750e-01, 4.64368431118658e-01, 4.59433042362761e-01, 4.53948949457569e-01, 4.47924334054434e-01, 4.41377581454347e-01, 4.34337068732390e-01, 4.26840598902970e-01, 4.18934512725883e-01, 4.10672527425703e-01, 4.02114364619464e-01, 3.93324237269638e-01, 3.84369267391065e-01, 3.75317903085445e-01, 3.66238396277240e-01, 3.57197392538026e-01, 3.48258672874039e-01, 3.39482075419895e-01, 3.30922613479487e-01, 3.22629795872118e-01, 3.14647146432947e-01, 3.07011911982102e-01, 2.99754942195634e-01, 2.92900720605928e-01, 2.86467523386762e-01, 2.80467681559116e-01, 2.74907922637951e-01, 2.69789769335645e-01, 2.65109975485188e-01, 2.60860982562463e-01, 2.57031383759710e-01, 2.53606386191521e-01, 2.50568265223066e-01, 2.47896807864237e-01, 2.45569744497232e-01, 2.43563169782064e-01, 2.41851954369139e-01, 2.40410149050604e-01, 2.39211382282946e-01, 2.38229250736495e-01, 2.37437700849469e-01, 2.36811397481347e-01, 2.36326073902489e-01, 2.35958855730345e-01, 2.35688550233533e-01, 2.35495891830240e-01, 2.35363734710489e-01, 2.35277184356406e-01, 2.35223661288444e-01, 2.35192892536929e-01, 2.35176828968221e-01, 2.35169489503448e-01, 2.35166736235044e-01, 2.35165987280979e-01, 2.35165876736986e-01, 2.35165868111082e-01, 2.35165711924296e-01, 2.35164653736973e-01, 2.35160763745057e-01, 2.35150394582632e-01, 2.35127702573322e-01, 2.35084247147554e-01, 2.35008685579984e-01, 2.34886578895683e-01, 2.34700323225610e-01, 2.34429218874993e-01, 2.34049686681914e-01, 2.33535637638298e-01, 2.32858997028103e-01, 2.31990378383898e-01, 2.30899895409067e-01, 2.29558091878744e-01, 2.27936960844374e-01, 2.26011015875045e-01, 2.23758369386672e-01, 2.21161767252624e-01, 2.18209525759775e-01, 2.14896317328936e-01, 2.11223755744314e-01, 2.07200740033002e-01, 2.02843528256295e-01, 1.98175527531546e-01, 1.93226803422974e-01, 1.88033329005293e-01, 1.82636009904357e-01, 1.77079535039979e-01, 1.71411112477145e-01, 1.65679154950411e-01, 1.59931979947814e-01, 1.54216584877290e-01, 1.48577549398524e-01, 1.43056105444785e-01, 1.37689401985625e-01, 1.32509977484901e-01, 1.27545439530607e-01, 1.22818339305653e-01, 1.18346219203147e-01, 1.14141805397111e-01, 1.10213313657397e-01, 1.06564835942066e-01, 1.03196776889760e-01, 1.00106312708931e-01, 9.72878495137763e-02, 9.47334633093675e-02, 9.24333090743232e-02, 9.03759913334312e-02, 8.85488929725398e-02, 8.69384626453922e-02, 8.55304638779302e-02, 8.43101908719888e-02, 8.32626570879853e-02, 8.23727630149023e-02, 8.16254492119981e-02, 8.10058398397757e-02, 8.04993806121246e-02, 8.00919735307986e-02, 7.97701090486100e-02, 7.95209945980702e-02, 7.93326768593090e-02, 7.91941538630851e-02, 7.90954721397981e-02, 7.90278037162153e-02, 7.89834978671990e-02, 7.89561031365622e-02, 7.89403562001354e-02, 7.89321355513266e-02, 7.89283796221980e-02, 7.89269706683200e-02, 7.89265874010425e-02, 7.89265308316942e-02, 7.89265263799024e-02, 7.89264452921685e-02, 7.89258959121213e-02, 7.89238763529028e-02, 7.89184931029041e-02, 7.89067127701912e-02, 7.88841550059232e-02, 7.88449361961909e-02, 7.87815732256831e-02, 7.86849562121290e-02, 7.85443984162271e-02, 7.83477703197661e-02, 7.80817229029651e-02, 7.77320022271653e-02, 7.72838534123395e-02, 7.67225069841260e-02, 7.60337345194984e-02, 7.52044539108567e-02, 7.42233579617420e-02, 7.30815341602380e-02, 7.17730391828866e-02, 7.02953897910845e-02, 6.86499329785448e-02, 6.68420629250051e-02, 6.48812605261782e-02, 6.27809425527488e-02, 6.05581209335355e-02, 5.82328869576116e-02, 5.58277488219545e-02, 5.33668623725987e-02, 5.08752027724455e-02, 4.83777282801725e-02, 4.58985860116599e-02, 4.34604037665167e-02, 4.10837025838049e-02, 3.87864529048585e-02, 3.65837845284602e-02, 3.44878483823834e-02, 3.25078177140341e-02, 3.06500084490438e-02, 2.89180935709368e-02, 2.73133844013245e-02, 2.58351522418394e-02, 2.44809663963252e-02, 2.32470284583854e-02, 2.21284872857979e-02, 2.11197237306635e-02, 2.02145985350768e-02, 1.94066605481424e-02, 1.86893154320273e-02, 1.80559572632859e-02, 1.75000669561172e-02, 1.70152823313998e-02, 1.65954450481394e-02, 1.62346296210671e-02, 1.59271594664006e-02, 1.56676144362251e-02, 1.54508336767805e-02, 1.52719169322779e-02, 1.51262266449805e-02, 1.50093924061423e-02, 1.49173185247236e-02, 1.48461947295573e-02, 1.47925093424703e-02, 1.47530636936466e-02, 1.47249861275469e-02, 1.47057436987836e-02, 1.46931495967522e-02, 1.46853644689600e-02, 1.46808901172051e-02, 1.46785544863235e-02, 1.46774874049242e-02, 1.46770871200488e-02, 1.46769782341778e-02, 1.46769621629090e-02, 1.46769608869448e-02, 1.46769375038502e-02, 1.46767790805740e-02, 1.46761967110507e-02, 1.46746444132901e-02, 1.46712476816617e-02, 1.46647441969776e-02, 1.46534397847944e-02, 1.46351828105793e-02, 1.46073602764289e-02, 1.45669188336512e-02, 1.45104136373675e-02, 1.44340873090698e-02, 1.43339801239648e-02, 1.42060708306957e-02, 1.40464452310738e-02, 1.38514868981190e-02, 1.36180814006650e-02, 1.33438224650604e-02, 1.30272060679886e-02, 1.26677969870599e-02, 1.22663522791413e-02, 1.18248878208078e-02, 1.13466775288487e-02, 1.08361800050574e-02, 1.02988936344843e-02, 9.74114788384384e-03, 9.16984480625160e-03, 8.59216967064924e-03, 8.01529247090116e-03, 7.44608241843991e-03, 6.89085537919709e-03, 6.35516998785590e-03, 5.84368257898404e-03, 5.36006496541575e-03, 4.90698327310213e-03, 4.48613112396856e-03, 4.09830685623547e-03, 3.74352240064013e-03, 3.42113095017282e-03, 3.12996152985079e-03, 2.86845064078989e-03, 2.63476378420080e-03, 2.42690230993503e-03, 2.24279327423460e-03, 2.08036168316994e-03, 1.93758569777458e-03, 1.81253621914834e-03, 1.70340284167025e-03, 1.60850848163818e-03, 1.52631506557555e-03, 1.45542252890607e-03, 1.39456310238018e-03, 1.34259253424318e-03, 1.29847957730078e-03, 1.26129478216797e-03, 1.23019938673963e-03, 1.20443487958255e-03, 1.18331363967172e-03, 1.16621090779745e-03, 1.15255821719984e-03, 1.14183829943215e-03, 1.13358138388264e-03, 1.12736272198788e-03, 1.12280109760832e-03, 1.11955802887564e-03, 1.11733733407959e-03, 1.11588472593382e-03, 1.11498711461877e-03, 1.11447134334242e-03, 1.11420214139457e-03, 1.11407915841255e-03, 1.11403302612700e-03, 1.11402047727438e-03, 1.11401862510263e-03, 1.11401847671289e-03, 1.11401574024228e-03, 1.11399720037753e-03, 1.11392904815943e-03, 1.11374739660164e-03, 1.11334994436163e-03, 1.11258911174161e-03, 1.11126706829623e-03, 1.10913311348781e-03, 1.10588391332556e-03, 1.10116711366004e-03, 1.09458882154456e-03, 1.08572534017526e-03, 1.07413933919531e-03, 1.05940032048579e-03, 1.04110880607777e-03, 1.01892315187513e-03, 9.92587333556116e-04, 9.61957540226877e-04, 9.27025045484697e-04, 8.87932707897221e-04, 8.44982666791338e-04, 7.98633394538228e-04, 7.49485243559889e-04, 6.98254911327467e-04, 6.45740670222494e-04, 5.92781494546557e-04, 5.40214058432298e-04, 4.88831723075123e-04, 4.39348952539831e-04, 3.92373261760781e-04, 3.48385422512365e-04, 3.07727980057848e-04, 2.70602333372949e-04, 2.37075167419221e-04, 2.07094999737478e-04, 1.80517893866750e-04, 1.57138000801638e-04, 1.36716025309624e-04, 1.18999193404079e-04, 1.03729796052221e-04, 9.06454635298815e-05, 7.94789598614898e-05, 6.99631394383005e-05, 6.18422458131818e-05, 5.48873159971208e-05, 4.89072794067624e-05, 4.37494194329702e-05, 3.92955386274707e-05, 3.54572264882336e-05, 3.21633800268728e-05, 2.93468883398873e-05, 2.69464226221316e-05, 2.49152195834768e-05, 2.32173838128347e-05, 2.18149518786959e-05, 2.06637832462775e-05, 1.97212305721450e-05, 1.89529678493727e-05, 1.83328551344175e-05, 1.78395545873400e-05, 1.74539925694717e-05, 1.71585456030243e-05, 1.69371264178153e-05, 1.67753863706821e-05, 1.66607735820890e-05, 1.65824801001691e-05, 1.65313511928389e-05, 1.64997910729880e-05, 1.64816683678737e-05, 1.64722128772610e-05, 1.64678939711393e-05, 1.64662740473887e-05, 1.64658334114165e-05, 1.64657683541236e-05, 1.64657631387399e-05, 1.64656665049279e-05, 1.64650117966403e-05, 1.64626050889183e-05, 1.64561905648952e-05, 1.64421565784778e-05, 1.64152958177510e-05, 1.63686355421078e-05, 1.62933608578442e-05, 1.61788615571368e-05, 1.60129524431341e-05, 1.57823502178189e-05, 1.54735359956248e-05, 1.50741430218204e-05, 1.45748404491997e-05, 1.39711034989887e-05, 1.32633897180767e-05, 1.24545885142879e-05, 1.15472748518641e-05, 1.05471306284244e-05, 9.47537176898550e-06, 8.38364725552765e-06, 7.34907933974682e-06, 6.41796719946797e-06, 5.56952979025027e-06, 4.78501426925959e-06, 4.11361934027354e-06, 3.60446746583420e-06, 3.14141517776534e-06, 2.63969019442811e-06, 2.22119824400124e-06, 1.87090668646306e-06, 1.43854272494082e-06, 1.11028386129419e-06, 9.80548222173638e-07, 8.80194733099415e-07, 9.69130510952709e-07, 1.08338761606319e-06, 1.10393543195642e-06, 1.09217307908693e-06, 1.04173098236515e-06, 8.69408495418611e-07, 8.99312453839229e-07, 9.90090128993870e-07, 7.41140453897060e-07, 8.46320688197011e-07, 1.55671067147978e-06, 2.03202674242675e-06, 2.16178801792921e-06, 1.96151938780531e-06, 1.52203242009818e-06, 9.72583009684058e-07, 4.95308029258013e-07, 4.92047774969251e-07, 5.54618239334559e-07, 5.11760617882828e-07, 4.29775582233560e-07, -2.84217094304040e-14, -4.26325641456060e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -2.84217094304040e-14, -1.42108547152020e-14],
[ 1.16000000000000e+02, 1.15999999846236e+02, 1.15999995091534e+02, 1.15999962877675e+02, 1.15999844455253e+02, 1.15999528777891e+02, 1.15998837891895e+02, 1.15997514633017e+02, 1.15995213069469e+02, 1.15991492043359e+02, 1.15985812065114e+02, 1.15977535711779e+02, 1.15965931576357e+02, 1.15950181717767e+02, 1.15929392475807e+02, 1.15902608447588e+02, 1.15868829373588e+02, 1.15827029650995e+02, 1.15776180172903e+02, 1.15715272172606e+02, 1.15643342718049e+02, 1.15559501437095e+02, 1.15462957948931e+02, 1.15353049328779e+02, 1.15229266753594e+02, 1.15091280291718e+02, 1.14938960647816e+02, 1.14772396599636e+02, 1.14591906904680e+02, 1.14398045636855e+02, 1.14191600235955e+02, 1.13973581989217e+02, 1.13745209162896e+02, 1.13507883496803e+02, 1.13263161198928e+02, 1.13012719876548e+02, 1.12758322983789e+02, 1.12501783350905e+02, 1.12244927210664e+02, 1.11989559893535e+02, 1.11737434074695e+02, 1.11490221168071e+02, 1.11249486210906e+02, 1.11016666386789e+02, 1.10793053201534e+02, 1.10579778248279e+02, 1.10377802462532e+02, 1.10187908758532e+02, 1.10010697940636e+02, 1.09846587786030e+02, 1.09695815190924e+02, 1.09558441258305e+02, 1.09434359181114e+02, 1.09323304741995e+02, 1.09224869212019e+02, 1.09138514388624e+02, 1.09063589469913e+02, 1.08999349420462e+02, 1.08944974444957e+02, 1.08899590152007e+02, 1.08862287963240e+02, 1.08832145304123e+02, 1.08808245104633e+02, 1.08789694141694e+02, 1.08775639772598e+02, 1.08765284640274e+02, 1.08757898977781e+02, 1.08752830200040e+02, 1.08749509544571e+02, 1.08747455607493e+02, 1.08746274713573e+02, 1.08745658156136e+02, 1.08745376440382e+02, 1.08745270758176e+02, 1.08745242009969e+02, 1.08745237766790e+02, 1.08745237210175e+02, 1.08745224241702e+02, 1.08745136378382e+02, 1.08744813383389e+02, 1.08743952396287e+02, 1.08742068141803e+02, 1.08738459548835e+02, 1.08732184112023e+02, 1.08722041167172e+02, 1.08706565064619e+02, 1.08684029006473e+02, 1.08652460057726e+02, 1.08609665540259e+02, 1.08553270667052e+02, 1.08480766870196e+02, 1.08389569826237e+02, 1.08277085701481e+02, 1.08140783655270e+02, 1.07978272190012e+02, 1.07787376571669e+02, 1.07566214315377e+02, 1.07313265680995e+02, 1.07027436271802e+02, 1.06708109155949e+02, 1.06355184369065e+02, 1.05969104112784e+02, 1.05550862357621e+02, 1.05101997890255e+02, 1.04624570248733e+02, 1.04121118711754e+02, 1.03594605766216e+02, 1.03048348212682e+02, 1.02485940772991e+02, 1.01911177908680e+02, 1.01327978896972e+02, 1.00740319139998e+02, 1.00152168128217e+02, 9.95674326266437e+01, 9.89899031802551e+01, 9.84232027858546e+01, 9.78707378466393e+01, 9.73356525943080e+01, 9.68207886508861e+01, 9.63286513005120e+01, 9.58613835594076e+01, 9.54207485217372e+01, 9.50081199084325e+01, 9.46244803504130e+01, 9.42704267142629e+01, 9.39461816979264e+01, 9.36516109386861e+01, 9.33862449387154e+01, 9.31493051881697e+01, 9.29397339281684e+01, 9.27562270335208e+01, 9.25972695036103e+01, 9.24611730310289e+01, 9.23461150763903e+01, 9.22501788212006e+01, 9.21713933067819e+01, 9.21077730045923e+01, 9.20573560105048e+01, 9.20182400210047e+01, 9.19886152402744e+01, 9.19667933897717e+01, 9.19512320502141e+01, 9.19405536614362e+01, 9.19335586372448e+01, 9.19292322161857e+01, 9.19267448585188e+01, 9.19254462060423e+01, 9.19248528346949e+01, 9.19246302395969e+01, 9.19245696882369e+01, 9.19245607509795e+01, 9.19245599619898e+01, 9.19245445021251e+01, 9.19244397592611e+01, 9.19240547133174e+01, 9.19230283214654e+01, 9.19207820842671e+01, 9.19164802569965e+01, 9.19089992984571e+01, 9.18969079601541e+01, 9.18784591985878e+01, 9.18515948370723e+01, 9.18139636028677e+01, 9.17629528121985e+01, 9.16957335635027e+01, 9.16093188263790e+01, 9.15006332858381e+01, 9.13665932339055e+01, 9.12041942195355e+01, 9.10106036102772e+01, 9.07832547316083e+01, 9.05199388846392e+01, 9.02188913529916e+01, 8.98788675418251e+01, 8.94992056784063e+01, 8.90798730528908e+01, 8.86214935667876e+01, 8.81253553246333e+01, 8.75933980566945e+01, 8.70281811803078e+01, 8.64328341814129e+01, 8.58109916482233e+01, 8.51667157009360e+01, 8.45044087885690e+01, 8.38287199610823e+01, 8.31444478510442e+01, 8.24564437191792e+01, 8.17695179418174e+01, 8.10883531034466e+01, 8.04174263043153e+01, 7.97609424297126e+01, 7.91227791246027e+01, 7.85064433116171e+01, 7.79150384698799e+01, 7.73512416168715e+01, 7.68172889385308e+01, 7.63149691624757e+01, 7.58456239426377e+01, 7.54101546493641e+01, 7.50090350208968e+01, 7.46423291482890e+01, 7.43097142640456e+01, 7.40105078057846e+01, 7.37436982378640e+01, 7.35079791330493e+01, 7.33017860338130e+01, 7.31233356187780e+01, 7.29706666866904e+01, 7.28416824349652e+01, 7.27341934536914e+01, 7.26459607844579e+01, 7.25747383149981e+01, 7.25183137059296e+01, 7.24745469860114e+01, 7.24414059181836e+01, 7.24169972394977e+01, 7.23995929207371e+01, 7.23876506796484e+01, 7.23798281151341e+01, 7.23749900045382e+01, 7.23722085147682e+01, 7.23707563100164e+01, 7.23700927819214e+01, 7.23698438690188e+01, 7.23697761586168e+01, 7.23697661647012e+01, 7.23697653630656e+01, 7.23697505688220e+01, 7.23696503356479e+01, 7.23692818679332e+01, 7.23682996685682e+01, 7.23661501506997e+01, 7.23620335722655e+01, 7.23548748248337e+01, 7.23433044314575e+01, 7.23256509025208e+01, 7.22999453577129e+01, 7.22639390377830e+01, 7.22151339926773e+01, 7.21508268362126e+01, 7.20681649996033e+01, 7.19642144024512e+01, 7.18360369053812e+01, 7.16807753392843e+01, 7.14957433583938e+01, 7.12785168826434e+01, 7.10270235275424e+01, 7.07396262143996e+01, 7.04151971498958e+01, 7.00531785883445e+01, 6.96536272518259e+01, 6.92172399727030e+01, 6.87453590101713e+01, 6.82399565287582e+01, 6.77035988448339e+01, 6.71393921623342e+01, 6.65509125315963e+01, 6.59421235702169e+01, 6.53172859873509e+01, 6.46808630907500e+01, 6.40374262212264e+01, 6.33915635096840e+01, 6.27477946029439e+01, 6.21104932022409e+01, 6.14838185386773e+01, 6.08716563616875e+01, 6.02775696603015e+01, 5.97047591309976e+01, 5.91560332745608e+01, 5.86337878757485e+01, 5.81399944552582e+01, 5.76761970880991e+01, 5.72435167921912e+01, 5.68426625504609e+01, 5.64739479699752e+01, 5.61373126093051e+01, 5.58323471038071e+01, 5.55583213578714e+01, 5.53142152211848e+01, 5.50987511963057e+01, 5.49104288203827e+01, 5.47475604168665e+01, 5.46083079230678e+01, 5.44907204708241e+01, 5.43927723379806e+01, 5.43124008075122e+01, 5.42475433798217e+01, 5.41961736937121e+01, 5.41563354344692e+01, 5.41261734544886e+01, 5.41039613125001e+01, 5.40881244588558e+01, 5.40772583605689e+01, 5.40701409711827e+01, 5.40657391036224e+01, 5.40632084518336e+01, 5.40618872191694e+01, 5.40612835360558e+01, 5.40610570735171e+01, 5.40609954702005e+01, 5.40609863776794e+01, 5.40609856763767e+01, 5.40609730832314e+01, 5.40608877628311e+01, 5.40605741161926e+01, 5.40597380507794e+01, 5.40579083490231e+01, 5.40544042779071e+01, 5.40483107522001e+01, 5.40384622185277e+01, 5.40234362578182e+01, 5.40015577034800e+01, 5.39709138334462e+01, 5.39293809070596e+01, 5.38746619756260e+01, 5.38043354957375e+01, 5.37159138214544e+01, 5.36069101583778e+01, 5.34749120528860e+01, 5.33176589967324e+01, 5.31331212925378e+01, 5.29195769962184e+01, 5.26756835747900e+01, 5.24005409326696e+01, 5.20937426935444e+01, 5.17554130847317e+01, 5.13862274375935e+01, 5.09874151437626e+01, 5.05607448203692e+01, 5.01084923504930e+01, 4.96333932910544e+01, 4.91385818127759e+01, 4.86275188241624e+01, 4.81039122401571e+01, 4.75716325183632e+01, 4.70346266361657e+01, 4.64968336358506e+01, 4.59621047088331e+01, 4.54341304933966e+01, 4.49163777976830e+01, 4.44120373356556e+01, 4.39239833245886e+01, 4.34547450200076e+01, 4.30064895545332e+01, 4.25810148867280e+01, 4.21797513094578e+01, 4.18037698284219e+01, 4.14537957765552e+01, 4.11302262289986e+01, 4.08331500654457e+01, 4.05623698339680e+01, 4.03174248571382e+01, 4.00976152577691e+01, 3.99020267537651e+01, 3.97295561775493e+01, 3.95789377213416e+01, 3.94487699052704e+01, 3.93375432221802e+01, 3.92436683419194e+01, 3.91655046688013e+01, 3.91013889478604e+01, 3.90496635169645e+01, 3.90087037109489e+01, 3.89769438486437e+01, 3.89529011814194e+01, 3.89351971591965e+01, 3.89225753817417e+01, 3.89139156523194e+01, 3.89082436375199e+01, 3.89047357585347e+01, 3.89027190895863e+01, 3.89016662104577e+01, 3.89011851421566e+01, 3.89010046769690e+01, 3.89009555860756e+01, 3.89009483403630e+01, 3.89009477917396e+01, 3.89009380729297e+01, 3.89008722265806e+01, 3.89006301686951e+01, 3.88999849332364e+01, 3.88985728613261e+01, 3.88958686134192e+01, 3.88911660247709e+01, 3.88835657163820e+01, 3.88719702468098e+01, 3.88550874388038e+01, 3.88314423306478e+01, 3.87993979768472e+01, 3.87571850494065e+01, 3.87029398662676e+01, 3.86347501006883e+01, 3.85507070158833e+01, 3.84489626442464e+01, 3.83277899207418e+01, 3.81856434246916e+01, 3.80212181271243e+01, 3.78335034257580e+01, 3.76218298132019e+01, 3.73859057882728e+01, 3.71258430860937e+01, 3.68421689438715e+01, 3.65358248817313e+01, 3.62081522831048e+01, 3.58608658138159e+01, 3.54960163314527e+01, 3.51159453358284e+01, 3.47232331656386e+01, 3.43206430752124e+01, 3.39110630981812e+01, 3.34974473248665e+01, 3.30827579918455e+01, 3.26699096740369e+01, 3.22617168842079e+01, 3.18608464517734e+01, 3.14697760550445e+01, 3.10907601073740e+01, 3.07258037995766e+01, 3.03766455243361e+01, 3.00447472788126e+01, 2.97312921111424e+01, 2.94371873589396e+01, 2.91630723593445e+01, 2.89093294492622e+01, 2.86760973343910e+01, 2.84632861950298e+01, 2.82705941484496e+01, 2.80975248690260e+01, 2.79434062742323e+01, 2.78074102294270e+01, 2.76885732248225e+01, 2.75858179498861e+01, 2.74979756449688e+01, 2.74238090541596e+01, 2.73620357415558e+01, 2.73113514686952e+01, 2.72704532674539e+01, 2.72380617847189e+01, 2.72129424278059e+01, 2.71939248084392e+01, 2.71799199732957e+01, 2.71699349247526e+01, 2.71630839789474e+01, 2.71585965798163e+01, 2.71558212853162e+01, 2.71542257612032e+01, 2.71533927522685e+01, 2.71530121432503e+01, 2.71528693636099e+01, 2.71528305240803e+01, 2.71528247914470e+01, 2.71528243610963e+01, 2.71528167864380e+01, 2.71527654670186e+01, 2.71525768114858e+01, 2.71520739257264e+01, 2.71509733757181e+01, 2.71488657029596e+01, 2.71452004804082e+01, 2.71392766084571e+01, 2.71302384473441e+01, 2.71170782608056e+01, 2.70986453007320e+01, 2.70736616887399e+01, 2.70407450437940e+01, 2.69984375638939e+01, 2.69452409973329e+01, 2.68796566438217e+01, 2.68002292230339e+01, 2.67055931594548e+01, 2.65945195844959e+01, 2.64659621787572e+01, 2.63190998965661e+01, 2.61533746523940e+01, 2.59685222136750e+01, 2.57645948305876e+01, 2.55419745173558e+01, 2.53013763444635e+01, 2.50438415628743e+01, 2.47707208192238e+01, 2.44836481074534e+01, 2.41845064285527e+01, 2.38753864036874e+01, 2.35585393219572e+01, 2.32363263115497e+01, 2.29111654954108e+01, 2.25854791048413e+01, 2.22616425412327e+01, 2.19419372655022e+01, 2.16285091426262e+01, 2.13233334874993e+01, 2.10281875869567e+01, 2.07446309676742e+01, 2.04739932018063e+01, 2.02173686439821e+01, 1.99756172090382e+01, 1.97493701424093e+01, 1.95390396973387e+01, 1.93448316934707e+01, 1.91667600607602e+01, 1.90046626405417e+01, 1.88582176949822e+01, 1.87269607459207e+01, 1.86103015099255e+01, 1.85075408101133e+01, 1.84178874237090e+01, 1.83404748679341e+01, 1.82743781385387e+01, 1.82186303996371e+01, 1.81722395860350e+01, 1.81342048261718e+01, 1.81035325318664e+01, 1.80792519372196e+01, 1.80604298102516e+01, 1.80461840137334e+01, 1.80356955619859e+01, 1.80282188126508e+01, 1.80230894493490e+01, 1.80197299534215e+01, 1.80176523291128e+01, 1.80164579329891e+01, 1.80158343597239e+01, 1.80155494460383e+01, 1.80154425654166e+01, 1.80154134913280e+01, 1.80154092000545e+01, 1.80154088788869e+01, 1.80154032390227e+01, 1.80153650281314e+01, 1.80152245611377e+01, 1.80148501298921e+01, 1.80140307073679e+01, 1.80124614577356e+01, 1.80097326525576e+01, 1.80053225353209e+01, 1.79985946081175e+01, 1.79887997312512e+01, 1.79750833233190e+01, 1.79564978200418e+01, 1.79320203899080e+01, 1.79005757111562e+01, 1.78610633891735e+01, 1.78123893425131e+01, 1.77535002216659e+01, 1.76834196654283e+01, 1.76012849681280e+01, 1.75063825531260e+01, 1.73981805503816e+01, 1.72763567818876e+01, 1.71408205850304e+01, 1.69917271564166e+01, 1.68294834698864e+01, 1.66547452897398e+01, 1.64684053266659e+01, 1.62715731213848e+01, 1.60655477361267e+01, 1.58517847357736e+01, 1.56318592085368e+01, 1.54074266881439e+01, 1.51801837952287e+01, 1.49518302359304e+01, 1.47240335178700e+01, 1.44983974127491e+01, 1.42764348542308e+01, 1.40595456435329e+01, 1.38489990638528e+01, 1.36459212858301e+01, 1.34512872778434e+01, 1.32659168109582e+01, 1.30904740628222e+01, 1.29254702740842e+01, 1.27712688933595e+01, 1.26280926610674e+01, 1.24960321257289e+01, 1.23750551529610e+01, 1.22650170694544e+01, 1.21656711723563e+01, 1.20766794196478e+01, 1.19976231917373e+01, 1.19280140732990e+01, 1.18673046444691e+01, 1.18148992911332e+01, 1.17701650461463e+01, 1.17324424591383e+01, 1.17010564651502e+01, 1.16753271854412e+01, 1.16545805515643e+01, 1.16381586006487e+01, 1.16254292502642e+01, 1.16157953295913e+01, 1.16087026237198e+01, 1.16036466827735e+01, 1.16001781590833e+01, 1.15979064643473e+01, 1.15965015836411e+01, 1.15956939418971e+01, 1.15952722874187e+01, 1.15950796316712e+01, 1.15950073601186e+01, 1.15949877005294e+01, 1.15949847988162e+01, 1.15949845815648e+01, 1.15949807654354e+01, 1.15949549106117e+01, 1.15948598657626e+01, 1.15946065127629e+01, 1.15940520644055e+01, 1.15929902619872e+01, 1.15911438804190e+01, 1.15881599100917e+01, 1.15836077392857e+01, 1.15769806053672e+01, 1.15677005132800e+01, 1.15551267316432e+01, 1.15385678665062e+01, 1.15172973786784e+01, 1.14905722535645e+01, 1.14576543573567e+01, 1.14178338292969e+01, 1.13704536798173e+01, 1.13149346055513e+01, 1.12507989133885e+01, 1.11776923857203e+01, 1.10954029337474e+01, 1.10038749855243e+01, 1.09032187424073e+01, 1.07937137039718e+01, 1.06758061890459e+01, 1.05501009416575e+01, 1.04173472711762e+01, 1.02784204991412e+01, 1.01342997377122e+01, 9.98604318158802e+00, 9.83476214520385e+00, 9.68159502418173e+00, 9.52768222287489e+00, 9.37414289672516e+00, 9.22205414041007e+00, 9.07243303806166e+00, 8.92622179944216e+00, 8.78427604524153e+00, 8.64735617679943e+00, 8.51612166692352e+00, 8.39112803420082e+00, 8.27282620950359e+00, 8.16156396838394e+00, 8.05758908620841e+00, 7.96105387396389e+00, 7.87202077064961e+00, 7.79046870072970e+00, 7.71629994858726e+00, 7.64934735171292e+00, 7.58938166558721e+00, 7.53611900127711e+00, 7.48922827782049e+00, 7.44833866270753e+00, 7.41304699352540e+00, 7.38292518157641e+00, 7.35752759452696e+00, 7.33639840131082e+00, 7.31907884078763e+00, 7.30511434881105e+00, 7.29406144957548e+00, 7.28549428972640e+00, 7.27901067101973e+00, 7.27423742223007e+00, 7.27083494585926e+00, 7.26850078146759e+00, 7.26697204561293e+00, 7.26602663783756e+00, 7.26548314125621e+00, 7.26519939257531e+00, 7.26506974671872e+00, 7.26502111228578e+00, 7.26500788256004e+00, 7.26500592988094e+00, 7.26500578329593e+00, 7.26500320327712e+00, 7.26498572328097e+00, 7.26492146512916e+00, 7.26475017774128e+00, 7.26437532583813e+00, 7.26365746395473e+00, 7.26240917233552e+00, 7.26039180587357e+00, 7.25731427716997e+00, 7.25283405785602e+00, 7.24656053571117e+00, 7.23806080584622e+00, 7.22686789982556e+00, 7.21249136581245e+00, 7.19443000631347e+00, 7.17218646106325e+00, 7.14528319707803e+00, 7.11327934478827e+00, 7.07578770962722e+00, 7.03249120500480e+00, 6.98315790761356e+00, 6.92765393999416e+00, 6.86595344512115e+00, 6.79814503520055e+00, 6.72443426762541e+00, 6.64514191463741e+00, 6.56069803379364e+00, 6.47163209401992e+00, 6.37855964527836e+00, 6.28216621782722e+00, 6.18318928201478e+00, 6.08239917918719e+00, 5.98057994314981e+00, 5.87851087185990e+00, 5.77694959009503e+00, 5.67661718112826e+00, 5.57818577804079e+00, 5.48226881327571e+00, 5.38941394665034e+00, 5.30009854153330e+00, 5.21472744512093e+00, 5.13363275492166e+00, 5.05707521790158e+00, 4.98524690582636e+00, 4.91827483265911e+00, 4.85622521936769e+00, 4.79910816065559e+00, 4.74688250070587e+00, 4.69946077628244e+00, 4.65671413218658e+00, 4.61847715405055e+00, 4.58455259561356e+00, 4.55471600145263e+00, 4.52872024156660e+00, 4.50629998148480e+00, 4.48717611123108e+00, 4.47106014931452e+00, 4.45765862503693e+00, 4.44667742518519e+00, 4.43782607127953e+00, 4.43082187286147e+00, 4.42539388284611e+00, 4.42128656473820e+00, 4.41826307035409e+00, 4.41610802208885e+00, 4.41462969669340e+00, 4.41366151831785e+00, 4.41306278687931e+00, 4.41271859257998e+00, 4.41253889701680e+00, 4.41245679374416e+00, 4.41242599415233e+00, 4.41241761593359e+00, 4.41241637932667e+00, 4.41241628610088e+00, 4.41241463998261e+00, 4.41240348729835e+00, 4.41236248898926e+00, 4.41225320360982e+00, 4.41201404060901e+00, 4.41155603546270e+00, 4.41075962516778e+00, 4.40947258755672e+00, 4.40750928990771e+00, 4.40465136975772e+00, 4.40064994153001e+00, 4.39522938429744e+00, 4.38809271679868e+00, 4.37892850467340e+00, 4.36741917218404e+00, 4.35325050874000e+00, 4.33612207379421e+00, 4.31575811876299e+00, 4.29191856991204e+00, 4.26440956100450e+00, 4.23309297805348e+00, 4.19789448825045e+00, 4.15880957543573e+00, 4.11590719543184e+00, 4.06933079131254e+00, 4.01929656156464e+00, 3.96608903961892e+00, 3.91005420582023e+00, 3.85159049715320e+00, 3.79113819289367e+00, 3.72916772692128e+00, 3.66616750593809e+00, 3.60263179861899e+00, 3.53904920934070e+00, 3.47589216997308e+00, 3.41360778392662e+00, 3.35261024789622e+00, 3.29327496733478e+00, 3.23593437918775e+00, 3.18087540592718e+00, 3.12833839301067e+00, 3.07851733042972e+00, 3.03156112903595e+00, 2.98757571298410e+00, 2.94662669833642e+00, 2.90874245074730e+00, 2.87391734761648e+00, 2.84211510751457e+00, 2.81327208789945e+00, 2.78730048784591e+00, 2.76409142339271e+00, 2.74351786780602e+00, 2.72543746699426e+00, 2.70969525151236e+00, 2.69612627148894e+00, 2.68455818008520e+00, 2.67481378559519e+00, 2.66671358299260e+00, 2.66007826367752e+00, 2.65473118853178e+00, 2.65050079537312e+00, 2.64722289876417e+00, 2.64474282911371e+00, 2.64291735022228e+00, 2.64161629078126e+00, 2.64072382643782e+00, 2.64013935509864e+00, 2.63977791895371e+00, 2.63957014161304e+00, 2.63946166675710e+00, 2.63941210453085e+00, 2.63939351216321e+00, 2.63938845460109e+00, 2.63938770811598e+00, 2.63938765153344e+00, 2.63938664837457e+00, 2.63937985183247e+00, 2.63935486712708e+00, 2.63928826786136e+00, 2.63914252098441e+00, 2.63886341355682e+00, 2.63837809120292e+00, 2.63759380958562e+00, 2.63639748993388e+00, 2.63465615644664e+00, 2.63221831580493e+00, 2.62891631547511e+00, 2.62456968662990e+00, 2.61898943859431e+00, 2.61198322493922e+00, 2.60336124818760e+00, 2.59294271352920e+00, 2.58056258641118e+00, 2.56607836006610e+00, 2.54937650333520e+00, 2.53037824281560e+00, 2.50904434159597e+00, 2.48537857272461e+00, 2.45942964918125e+00, 2.43129146010655e+00, 2.40110156849290e+00, 2.36903803868371e+00, 2.33531477140465e+00, 2.30017561812966e+00, 2.26388761553578e+00, 2.22673371806269e+00, 2.18900540989394e+00, 2.15099554927814e+00, 2.11299174418670e+00, 2.07527048764398e+00, 2.03809220336046e+00, 2.00169727650449e+00, 1.96630307751761e+00, 1.93210193314002e+00, 1.89925995995547e+00, 1.86791665134426e+00, 1.83818509688865e+00, 1.81015271148415e+00, 1.78388235705881e+00, 1.75941375055193e+00, 1.73676506572760e+00, 1.71593465198919e+00, 1.69690280945646e+00, 1.67963357525875e+00, 1.66407649056471e+00, 1.65016833072953e+00, 1.63783479164042e+00, 1.62699213354286e+00, 1.61754878912974e+00, 1.60940694539657e+00, 1.60246410877743e+00, 1.59661466058907e+00, 1.59175140517411e+00, 1.58776710684006e+00, 1.58455600434472e+00, 1.58201528397723e+00, 1.58004648498463e+00, 1.57855680494177e+00, 1.57746026835200e+00, 1.57667871985537e+00, 1.57614260428265e+00, 1.57579149956702e+00, 1.57557437508312e+00, 1.57544955695292e+00, 1.57538439263813e+00, 1.57535461898065e+00, 1.57534344992764e+00, 1.57534041168140e+00, 1.57533996324285e+00, 1.57533992903616e+00, 1.57533931973569e+00, 1.57533519163951e+00, 1.57532001638052e+00, 1.57527956516680e+00, 1.57519104098475e+00, 1.57502151543635e+00, 1.57472673677941e+00, 1.57425037044825e+00, 1.57352372697201e+00, 1.57246602410135e+00, 1.57098521923452e+00, 1.56897943397736e+00, 1.56633897408865e+00, 1.56294892472852e+00, 1.55869227297441e+00, 1.55345347790566e+00, 1.54712237495892e+00, 1.53959826835991e+00, 1.53079403659322e+00, 1.52064005478601e+00, 1.50908772815660e+00, 1.49611243520503e+00, 1.48171569973680e+00, 1.46592644701310e+00, 1.44880124928302e+00, 1.43042352580272e+00, 1.41090172688922e+00, 1.39036659455782e+00, 1.36896764790322e+00, 1.34686908452284e+00, 1.32424531639383e+00, 1.30127636802575e+00, 1.27814335671054e+00, 1.25502425133344e+00, 1.23209007090256e+00, 1.20950164094997e+00, 1.18740697978038e+00, 1.16593934147103e+00, 1.14521590217278e+00, 1.12533704325922e+00, 1.10638616075269e+00, 1.08842991564404e+00, 1.07151883369326e+00, 1.05568816479787e+00, 1.04095891935879e+00, 1.02733901040828e+00, 1.01482444380120e+00, 1.00340051296723e+00, 9.93042968337335e-01, 9.83719143726148e-01, 9.75389032113142e-01, 9.68006311144279e-01, 9.61519324203437e-01, 9.55872026154807e-01, 9.51004904011612e-01, 9.46855882083270e-01, 9.43361218873321e-01, 9.40456399467678e-01, 9.38077022720122e-01, 9.36159677597999e-01, 9.34642798015716e-01, 9.33467480787840e-01, 9.32578247410035e-01, 9.31923727613508e-01, 9.31457241360363e-01, 9.31137256366569e-01, 9.30927700433017e-01, 9.30798111758008e-01, 9.30723615769821e-01, 9.30684723493883e-01, 9.30666953598532e-01, 9.30660287546928e-01, 9.30658474223918e-01, 9.30658206581441e-01, 9.30658186021688e-01, 9.30657817916796e-01, 9.30655323954724e-01, 9.30646155926709e-01, 9.30621717641401e-01, 9.30568236630918e-01, 9.30465820148996e-01, 9.30287735961087e-01, 9.29999955392873e-01, 9.29560992424172e-01, 9.28922067852426e-01, 9.28027621121913e-01, 9.26816183711480e-01, 9.25221616468491e-01, 9.23174698667040e-01, 9.20605038933431e-01, 9.17443258092419e-01, 9.13623372605414e-01, 9.09085286327254e-01, 9.03777280019625e-01, 8.97658374898299e-01, 8.90700440865814e-01, 8.82889923939103e-01, 8.74229081812814e-01, 8.64736641396206e-01, 8.54447826061900e-01, 8.43413740456498e-01, 8.31700143220715e-01, 8.19385678483457e-01, 8.06559671259990e-01, 7.93319616364727e-01, 7.79768502895521e-01, 7.66012116100399e-01, 7.52156446467524e-01, 7.38305314532468e-01, 7.24558292432917e-01, 7.11008973256270e-01, 6.97743610093724e-01, 6.84840121142230e-01, 6.72367436961892e-01, 6.60385151896293e-01, 6.48943433613312e-01, 6.38083141993704e-01, 6.27836110064635e-01, 6.18225544096376e-01, 6.09266506167771e-01, 6.00966449460287e-01, 5.93325783508121e-01, 5.86338453094314e-01, 5.79992520137154e-01, 5.74270742597207e-01, 5.69151148108858e-01, 5.64607602709017e-01, 5.60610376745970e-01, 5.57126710858128e-01, 5.54121384886017e-01, 5.51557291797451e-01, 5.49396017264399e-01, 5.47598423548152e-01, 5.46125233983801e-01, 5.44937611781208e-01, 5.43997724295366e-01, 5.43269281586390e-01, 5.42718036220649e-01, 5.42312230073208e-01, 5.42022973547503e-01, 5.41824543249362e-01, 5.41694585767431e-01, 5.41614217781728e-01, 5.41568016101053e-01, 5.41543895219959e-01, 5.41532874317994e-01, 5.41528740017853e-01, 5.41527615389953e-01, 5.41527449397279e-01, 5.41527436548776e-01, 5.41527205239930e-01, 5.41525638089766e-01, 5.41519877099390e-01, 5.41504520563279e-01, 5.41470913916100e-01, 5.41406556064743e-01, 5.41294646289586e-01, 5.41113794681308e-01, 5.40837916544390e-01, 5.40436328009321e-01, 5.39874055937901e-01, 5.39112369754776e-01, 5.38109535810150e-01, 5.36821786122900e-01, 5.35204482935910e-01, 5.33213448786427e-01, 5.30806419430689e-01, 5.27944564978341e-01, 5.24594014250269e-01, 5.20727310077987e-01, 5.16324720340563e-01, 5.11375332019895e-01, 5.05877863954353e-01, 4.99841148090326e-01, 4.93284247921920e-01, 4.86236204813110e-01, 4.78735425882022e-01, 4.70828748773357e-01, 4.62570236824845e-01, 4.54019771213240e-01, 4.45241513756542e-01, 4.36302315043136e-01, 4.27270138099459e-01, 4.18212559050161e-01, 4.09195394615963e-01, 4.00281493296035e-01, 3.91529713989499e-01, 3.82994103617861e-01, 3.74723274687881e-01, 3.66759975038335e-01, 3.59140835361529e-01, 3.51896275430164e-01, 3.45050547132999e-01, 3.38621891219688e-01, 3.32622784836062e-01, 3.27060258241474e-01, 3.21936261276576e-01, 3.17248062931043e-01, 3.12988670482824e-01, 3.09147257900548e-01, 3.05709596295259e-01, 3.02658481981410e-01, 2.99974160003984e-01, 2.97634742695891e-01, 2.95616623873144e-01, 2.93894889640882e-01, 2.92443726494184e-01, 2.91236826530763e-01, 2.90247788267195e-01, 2.89450509916421e-01, 2.88819570227062e-01, 2.88330590294905e-01, 2.87960568331215e-01, 2.87688178382027e-01, 2.87494023585836e-01, 2.87360834819353e-01, 2.87273606549078e-01, 2.87219663348708e-01, 2.87188652762197e-01, 2.87172462846627e-01, 2.87165065633033e-01, 2.87162290697822e-01, 2.87161535849563e-01, 2.87161424435624e-01, 2.87161415741778e-01, 2.87161258325824e-01, 2.87160191810344e-01, 2.87156271201084e-01, 2.87145820407233e-01, 2.87122949663654e-01, 2.87079151611564e-01, 2.87002993184274e-01, 2.86879919049326e-01, 2.86692180686316e-01, 2.86418903067101e-01, 2.86036298135642e-01, 2.85518030636823e-01, 2.84835737123501e-01, 2.83959693074650e-01, 2.82859616027324e-01, 2.81505584671791e-01, 2.79869045423737e-01, 2.77923869698299e-01, 2.75647417782793e-01, 2.73021559726885e-01, 2.70033600925544e-01, 2.66677060772281e-01, 2.62952257345574e-01, 2.58866659591604e-01, 2.54434980460857e-01, 2.49678999094101e-01, 2.44627116216989e-01, 2.39313662977779e-01, 2.33777998100379e-01, 2.28063440134875e-01, 2.22216089790763e-01, 2.16283601272096e-01, 2.10313961090236e-01, 2.04354328342518e-01, 1.98449982594880e-01, 1.92643415238948e-01, 1.86973588561784e-01, 1.81475374840517e-01, 1.76179176516754e-01, 1.71110718704213e-01, 1.66290997492581e-01, 1.61736362024598e-01, 1.57458705184524e-01, 1.53465736773697e-01, 1.49761313933652e-01, 1.46345805898378e-01, 1.43216473470801e-01, 1.40367847499235e-01, 1.37792094693694e-01, 1.35479363052809e-01, 1.33418102719858e-01, 1.31595361087321e-01, 1.29997053318775e-01, 1.28608211120024e-01, 1.27413213566669e-01, 1.26396004125283e-01, 1.25540297750149e-01, 1.24829781188666e-01, 1.24248308488973e-01, 1.23780092302894e-01, 1.23409890048407e-01, 1.23123182487745e-01, 1.22906340929958e-01, 1.22746778212104e-01, 1.22633077958781e-01, 1.22555096441985e-01, 1.22504031691207e-01, 1.22472455325848e-01, 1.22454303838481e-01, 1.22444827644685e-01, 1.22440498011855e-01, 1.22438873836245e-01, 1.22438432022903e-01, 1.22438366812261e-01, 1.22438361680466e-01, 1.22438268206082e-01, 1.22437634904543e-01, 1.22435306837346e-01, 1.22429101190122e-01, 1.22415520918096e-01, 1.22389515506512e-01, 1.22344299493875e-01, 1.22271238844618e-01, 1.22159814546436e-01, 1.21997671796478e-01, 1.21770761630245e-01, 1.21463579632589e-01, 1.21059503267531e-01, 1.20541225266336e-01, 1.19891275425246e-01, 1.19092617251042e-01, 1.18129299495607e-01, 1.16987136262864e-01, 1.15654383727545e-01, 1.14122377332691e-01, 1.12386091382689e-01, 1.10444583841087e-01, 1.08301293253433e-01, 1.05964162069810e-01, 1.03445570857801e-01, 1.00762080187437e-01, 9.79339902122121e-02, 9.49847408398057e-02, 9.19401865341456e-02, 8.88277880322335e-02, 8.56757677581328e-02, 8.25122760961818e-02, 7.93646120896625e-02, 7.62585351715899e-02, 7.32176951773056e-02, 7.02631972804681e-02, 6.74133077934300e-02, 6.46832969932376e-02, 6.20854070550649e-02, 5.96289272857007e-02, 5.73203553231707e-02, 5.51636216936799e-02, 5.31603557521976e-02, 5.13101730964109e-02, 4.96109675355143e-02, 4.80591941581281e-02, 4.66501336154863e-02, 4.53781311455401e-02, 4.42368069119397e-02, 4.32192367946840e-02, 4.23181047801791e-02, 4.15258295607401e-02, 4.08346689079337e-02, 4.02368058972939e-02, 3.97244212035446e-02, 3.92897555154548e-02, 3.89251656897756e-02, 3.86231776285371e-02, 3.83765380719865e-02, 3.81782666082842e-02, 3.80217082721686e-02, 3.79005861983615e-02, 3.78090529792274e-02, 3.77417387060746e-02, 3.76937932065857e-02, 3.76609197641358e-02, 3.76393976369087e-02, 3.76260909860946e-02, 3.76184423443249e-02, 3.76144494588573e-02, 3.76126251666449e-02, 3.76119408262241e-02, 3.76117546704364e-02, 3.76117271942604e-02, 3.76117250128800e-02, 3.76116850360830e-02, 3.76114141881061e-02, 3.76104185342854e-02, 3.76077645785244e-02, 3.76019569266361e-02, 3.75908363618009e-02, 3.75715031859301e-02, 3.75402706108332e-02, 3.74926528567556e-02, 3.74233923294724e-02, 3.73265296682490e-02, 3.71955194418945e-02, 3.70233926932997e-02, 3.68029653155162e-02, 3.65270883779800e-02, 3.61889331227871e-02, 3.57822996397061e-02, 3.53019345715353e-02, 3.47438400506377e-02, 3.41055539408615e-02, 3.33863808464372e-02, 3.25875546226939e-02, 3.17123164623609e-02, 3.07658979307064e-02, 2.97554051857531e-02, 2.86896083702146e-02, 2.75786479238462e-02, 2.64336763968771e-02, 2.52664593928671e-02, 2.40889619131366e-02, 2.29129463658069e-02, 2.17496059759128e-02, 2.06092527800052e-02, 1.95010734989012e-02, 1.84329601146089e-02, 1.74114156440481e-02, 1.64415300322389e-02, 1.55270167944366e-02, 1.46702983519162e-02, 1.38726269724998e-02, 1.31342285681484e-02, 1.24544578469568e-02, 1.18319550069828e-02, 1.12647960474561e-02, 1.07506307839031e-02, 1.02868047371061e-02, 9.87046305650097e-03, 9.49863628781600e-03, 9.16830893355325e-03, 8.87647239842693e-03, 8.62016423135969e-03, 8.39649574760415e-03, 8.20267022750443e-03, 8.03599392254739e-03, 7.89388201312136e-03, 7.77386144164893e-03, 7.67357223357124e-03, 7.59076856711260e-03, 7.52332050213056e-03, 7.46921693693992e-03, 7.42657003722513e-03, 7.39362106811825e-03, 7.36874727860481e-03, 7.35046924373250e-03, 7.33745789132456e-03, 7.32854033509511e-03, 7.32270361703513e-03, 7.31909552816035e-03, 7.31702182005733e-03, 7.31593932715668e-03, 7.31544476595047e-03, 7.31525924520326e-03, 7.31520877967284e-03, 7.31520133108177e-03, 7.31520073431113e-03, 7.31518972956735e-03, 7.31511517115280e-03, 7.31484109161329e-03, 7.31411053553188e-03, 7.31251193067806e-03, 7.30945117473425e-03, 7.30413088332682e-03, 7.29553828860219e-03, 7.28244330724692e-03, 7.26340826828675e-03, 7.23681065244364e-03, 7.20087988068485e-03, 7.15374865005458e-03, 7.09351851367046e-03, 7.01833833672083e-03, 6.92649298036940e-03, 6.81649816833385e-03, 6.68719613990731e-03, 6.53784558581094e-03, 6.36819872561034e-03, 6.17855841964854e-03, 5.96980905059752e-03, 5.74341659661570e-03, 5.50139573033448e-03, 5.24624465643342e-03, 4.98085136764814e-03, 4.70837765415411e-03, 4.43212921899772e-03, 4.15542141367098e-03, 3.88145031418219e-03, 3.61317810433093e-03, 3.35324004328186e-03, 3.10387770834097e-03, 2.86690003281365e-03, 2.64367063356019e-03, 2.43511779238759e-03, 2.24176260321940e-03, 2.06376107333028e-03, 1.90095638077992e-03, 1.75293705513013e-03, 1.61909615155764e-03, 1.49868648732365e-03, 1.39086809133460e-03, 1.29474685010678e-03, 1.20940619875398e-03, 1.13393343805512e-03, 1.06744115659296e-03, 1.00908439559078e-03, 9.58072255073716e-04, 9.13671613091083e-04, 8.75205510183719e-04, 8.42051501891206e-04, 8.13640214687439e-04, 7.89451291453247e-04, 7.69007771722841e-04, 7.51872138138765e-04, 7.37644179096719e-04, 7.25958553630335e-04, 7.16481307321715e-04, 7.08906351732708e-04, 7.02952914679145e-04, 6.98364117795336e-04, 6.94906343170487e-04, 6.92369034425155e-04, 6.90564703134555e-04, 6.89328981493986e-04, 6.88520580126806e-04, 6.88021010333273e-04, 6.87733943877333e-04, 6.87584108945316e-04, 6.87515657176618e-04, 6.87489980023770e-04, 6.87482995374467e-04, 6.87481964448011e-04, 6.87481881072927e-04, 6.87480333567692e-04, 6.87469849225408e-04, 6.87431308833197e-04, 6.87328583069302e-04, 6.87103815025125e-04, 6.86673528832671e-04, 6.85925794172704e-04, 6.84718696589925e-04, 6.82880384133000e-04, 6.80210970713802e-04, 6.76486564586298e-04, 6.71465640422753e-04, 6.64897865206626e-04, 6.56535286410076e-04, 6.46145478711446e-04, 6.33525866163609e-04, 6.18518216015218e-04, 6.01022567067844e-04, 5.81010487124445e-04, 5.58537330476838e-04, 5.33750687296219e-04, 5.06889308169889e-04, 4.78270133584147e-04, 4.48273008430533e-04, 4.17335028984667e-04, 3.85940734261681e-04, 3.54577627405206e-04, 3.23683775818040e-04, 2.93658063370117e-04, 2.64878787305634e-04, 2.37647203164215e-04, 2.12182495985758e-04, 1.88724794043083e-04, 1.67397947834047e-04, 1.48129845811695e-04, 1.30868511391213e-04, 1.15455004589649e-04, 1.01711984910935e-04, 8.95679322781007e-05, 7.89157059983836e-05, 6.94851173079769e-05, 6.15862476109896e-05, 5.47716704772938e-05, 4.84629160695249e-05, 4.32951498936518e-05, 3.91940416903935e-05, 3.53987042984727e-05, 3.17860295382388e-05, 2.82997212366354e-05, 2.49499553177657e-05, 2.17810911209426e-05, 1.88707456629800e-05, 1.67608781538320e-05, 1.47379314938689e-05, 1.27909302136686e-05, 1.09792095628336e-05, 9.33777742773145e-06, 7.87668270163522e-06, 6.58418733223698e-06, 5.43996314661399e-06, 4.42775314013488e-06, 3.53821799592424e-06, 2.76626974482497e-06, 2.10800565980662e-06, 1.55872267271207e-06, 1.11199787511396e-06, 7.59513014259028e-07, 4.91305968353117e-07, 2.96211354111620e-07, 1.42108547152020e-14, -1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14],
[ 1.17000000000000e+02, 1.16999999844413e+02, 1.16999995033347e+02, 1.16999962437608e+02, 1.16999842611347e+02, 1.16999523191775e+02, 1.16998824115597e+02, 1.16997485169825e+02, 1.16995156321268e+02, 1.16991391180939e+02, 1.16985643860808e+02, 1.16977269375466e+02, 1.16965527636337e+02, 1.16949590987487e+02, 1.16928555146163e+02, 1.16901453342829e+02, 1.16867273407122e+02, 1.16824977515824e+02, 1.16773524300016e+02, 1.16711892989221e+02, 1.16639109235280e+02, 1.16554272192510e+02, 1.16456582322232e+02, 1.16345369237392e+02, 1.16220118718629e+02, 1.16080497843862e+02, 1.15926377018586e+02, 1.15757847618676e+02, 1.15575234001817e+02, 1.15379098832021e+02, 1.15170240994013e+02, 1.14949685821839e+02, 1.14718667875916e+02, 1.14478607006896e+02, 1.14231078874461e+02, 1.13977781389629e+02, 1.13720498689678e+02, 1.13461064233638e+02, 1.13201324448200e+02, 1.12943104101736e+02, 1.12688174288271e+02, 1.12438223610437e+02, 1.12194832896234e+02, 1.11959453588681e+02, 1.11733389815390e+02, 1.11517784069496e+02, 1.11313606400300e+02, 1.11121647004877e+02, 1.10942512115665e+02, 1.10776623082398e+02, 1.10624218542516e+02, 1.10485359559602e+02, 1.10359937584268e+02, 1.10247685058217e+02, 1.10148188442497e+02, 1.10060903407933e+02, 1.09985171881855e+02, 1.09920240602714e+02, 1.09865280794818e+02, 1.09819408541106e+02, 1.09781705404350e+02, 1.09751238828325e+02, 1.09727081842088e+02, 1.09708331594329e+02, 1.09694126262185e+02, 1.09683659910918e+02, 1.09676194927764e+02, 1.09671071714586e+02, 1.09667715398415e+02, 1.09665639404396e+02, 1.09664445829161e+02, 1.09663822650714e+02, 1.09663537909716e+02, 1.09663431092628e+02, 1.09663402035704e+02, 1.09663397746960e+02, 1.09663397184367e+02, 1.09663384076631e+02, 1.09663295269780e+02, 1.09662968806272e+02, 1.09662098573408e+02, 1.09660194084932e+02, 1.09656546741954e+02, 1.09650203920127e+02, 1.09639952067401e+02, 1.09624309807629e+02, 1.09601531824947e+02, 1.09569624045178e+02, 1.09526370326262e+02, 1.09469370513917e+02, 1.09396089310306e+02, 1.09303914948092e+02, 1.09190226175785e+02, 1.09052465570333e+02, 1.08888216739240e+02, 1.08695282606361e+02, 1.08471761744956e+02, 1.08216119671217e+02, 1.07927252158998e+02, 1.07604537959501e+02, 1.07247878741154e+02, 1.06857724511144e+02, 1.06435083166865e+02, 1.05981513166064e+02, 1.05499098747463e+02, 1.04990407933418e+02, 1.04458434902070e+02, 1.03906530122333e+02, 1.03338323331805e+02, 1.02757645149188e+02, 1.02168452252306e+02, 1.01574758827712e+02, 1.00980574427348e+02, 1.00389846641035e+02, 9.98064067044552e+01, 9.92339170534230e+01, 9.86758211507026e+01, 9.81352969439567e+01, 9.76152157243512e+01, 9.71181079774223e+01, 9.66461372825509e+01, 9.62010826810443e+01, 9.57843293783069e+01, 9.53968672591866e+01, 9.50392964868193e+01, 9.47118393893690e+01, 9.44143578654996e+01, 9.41463756104237e+01, 9.39071045435201e+01, 9.36954748826704e+01, 9.35101683479190e+01, 9.33496539845399e+01, 9.32122260752084e+01, 9.30960435681204e+01, 9.29991703897331e+01, 9.29196159454605e+01, 9.28553750477618e+01, 9.28044664572886e+01, 9.27649691874411e+01, 9.27350557133030e+01, 9.27130212485376e+01, 9.26973083124987e+01, 9.26865259061419e+01, 9.26794627481591e+01, 9.26750941880828e+01, 9.26725826043331e+01, 9.26712713035617e+01, 9.26706721530818e+01, 9.26704473900281e+01, 9.26703862489312e+01, 9.26703772246299e+01, 9.26703764279559e+01, 9.26703608175207e+01, 9.26702550545198e+01, 9.26698662584506e+01, 9.26688298701842e+01, 9.26665617562992e+01, 9.26622180334042e+01, 9.26546642206745e+01, 9.26424551377436e+01, 9.26238267430984e+01, 9.25967008494465e+01, 9.25587033490677e+01, 9.25071962253838e+01, 9.24393232103820e+01, 9.23520684693736e+01, 9.22423271606126e+01, 9.21069861430289e+01, 9.19430125173493e+01, 9.17475471219102e+01, 9.15179996117418e+01, 9.12521413812515e+01, 9.09481924008705e+01, 9.06048980735758e+01, 9.02215925102887e+01, 8.97982451813014e+01, 8.93354886987504e+01, 8.88346264599429e+01, 8.82976199358550e+01, 8.77270564062588e+01, 8.71260988119931e+01, 8.64984200441100e+01, 8.58481244116683e+01, 8.51796592836974e+01, 8.44977200788991e+01, 8.38071519456974e+01, 8.31128516200235e+01, 8.24196729597669e+01, 8.17323393866588e+01, 8.10553658352204e+01, 8.03929918744250e+01, 7.97491266308881e+01, 7.91273052500096e+01, 7.85306560582787e+01, 7.79618773696968e+01, 7.74232229215093e+01, 7.69164950874264e+01, 7.64430451807747e+01, 7.60037802649270e+01, 7.55991759284779e+01, 7.52292944829726e+01, 7.48938080314523e+01, 7.45920258562230e+01, 7.43229255894073e+01, 7.40851876542849e+01, 7.38772324882205e+01, 7.36972600680632e+01, 7.35432912484670e+01, 7.34132103895143e+01, 7.33048086939862e+01, 7.32158276024207e+01, 7.31440015145945e+01, 7.30870990300146e+01, 7.30429618388783e+01, 7.30095403597639e+01, 7.29849252204849e+01, 7.29673737210134e+01, 7.29553305057949e+01, 7.29474418070144e+01, 7.29425627964361e+01, 7.29397577937388e+01, 7.29382933132654e+01, 7.29376241763250e+01, 7.29373731593562e+01, 7.29373048765977e+01, 7.29372947982035e+01, 7.29372939897917e+01, 7.29372790704924e+01, 7.29371779900475e+01, 7.29368064076843e+01, 7.29358159058737e+01, 7.29336482186294e+01, 7.29294968449056e+01, 7.29222775920511e+01, 7.29106094162398e+01, 7.28928067196297e+01, 7.28668840213854e+01, 7.28305736329792e+01, 7.27813566282834e+01, 7.27165069989205e+01, 7.26331484230926e+01, 7.25283225571723e+01, 7.23990671989787e+01, 7.22425020960133e+01, 7.20559196176435e+01, 7.18368770222527e+01, 7.15832866779837e+01, 7.12935003869346e+01, 7.09663839582372e+01, 7.06013784029892e+01, 7.01985445937808e+01, 6.97585889340265e+01, 6.92828684874130e+01, 6.87733750750180e+01, 6.82326989864256e+01, 6.76639740819208e+01, 6.70708070813490e+01, 6.64571946329414e+01, 6.58274322375744e+01, 6.51860192128308e+01, 6.45375636172818e+01, 6.38866904861068e+01, 6.32379559781474e+01, 6.25957692504264e+01, 6.19643231907708e+01, 6.13475346301053e+01, 6.07489943301164e+01, 6.01719268462894e+01, 5.96191602209147e+01, 5.90931052982685e+01, 5.85957442477623e+01, 5.81286276482854e+01, 5.76928792742824e+01, 5.72892075791404e+01, 5.69179228221948e+01, 5.65789588318383e+01, 5.62718985153029e+01, 5.59960023808041e+01, 5.57502394956862e+01, 5.55333204389418e+01, 5.53437319025309e+01, 5.51797726472549e+01, 5.50395905263797e+01, 5.49212202590114e+01, 5.48226215733733e+01, 5.47417172573332e+01, 5.46764305606525e+01, 5.46247213020505e+01, 5.45846199560005e+01, 5.45542589402894e+01, 5.45319003053781e+01, 5.45159590477322e+01, 5.45050213356007e+01, 5.44978570476027e+01, 5.44934261784152e+01, 5.44908788547159e+01, 5.44895489181442e+01, 5.44889412582192e+01, 5.44887133038536e+01, 5.44886512947249e+01, 5.44886421423070e+01, 5.44886414363844e+01, 5.44886287602819e+01, 5.44885428778344e+01, 5.44882271650613e+01, 5.44873855921690e+01, 5.44855438378212e+01, 5.44820166860515e+01, 5.44758830275904e+01, 5.44659696413239e+01, 5.44508447611565e+01, 5.44288222323219e+01, 5.43979768205360e+01, 5.43561709482496e+01, 5.43010927874958e+01, 5.42303052357525e+01, 5.41413048442091e+01, 5.40315892702306e+01, 5.38987313113499e+01, 5.37404570806407e+01, 5.35547254451941e+01, 5.33398055180063e+01, 5.30943488160956e+01, 5.28174527146105e+01, 5.25087120645548e+01, 5.21682563060230e+01, 5.17967700786730e+01, 5.13954961577509e+01, 5.09662204545142e+01, 5.05112397283748e+01, 5.00333134819810e+01, 4.95356021888049e+01, 4.90215945085664e+01, 4.84950264858049e+01, 4.79597959274963e+01, 4.74198752405069e+01, 4.68792259815335e+01, 4.63417182090749e+01, 4.58110573965888e+01, 4.52907211500896e+01, 4.47839072894942e+01, 4.42934940615210e+01, 4.38220124457608e+01, 4.33716297966011e+01, 4.29441435167655e+01, 4.25409831317747e+01, 4.21632190338981e+01, 4.18115762566840e+01, 4.14864518704275e+01, 4.11879348905514e+01, 4.09158279063098e+01, 4.06696699223929e+01, 4.04487601342028e+01, 4.02521825178979e+01, 4.00788312096955e+01, 3.99274366834036e+01, 3.97965927219025e+01, 3.96847841289138e+01, 3.95904150525795e+01, 3.95118377015852e+01, 3.94473811362823e+01, 3.93953797194556e+01, 3.93542007217987e+01, 3.93222705034335e+01, 3.92980986422068e+01, 3.92802993584308e+01, 3.92676095991334e+01, 3.92589031955327e+01, 3.92532005954899e+01, 3.92496737952986e+01, 3.92476462466512e+01, 3.92465876867926e+01, 3.92461040227940e+01, 3.92459225838505e+01, 3.92458732280699e+01, 3.92458659432606e+01, 3.92458653916768e+01, 3.92458556204256e+01, 3.92457894187777e+01, 3.92455460547646e+01, 3.92448973375593e+01, 3.92434776454760e+01, 3.92407588022792e+01, 3.92360308268294e+01, 3.92283894719198e+01, 3.92167313401757e+01, 3.91997572121797e+01, 3.91759840365895e+01, 3.91437658054693e+01, 3.91013232634285e+01, 3.90467820734247e+01, 3.89782186884928e+01, 3.88937127688300e+01, 3.87914045589348e+01, 3.86695552309244e+01, 3.85266078470663e+01, 3.83612463409786e+01, 3.81724498063318e+01, 3.79595394501767e+01, 3.77222158361189e+01, 3.74605845091819e+01, 3.71751687309909e+01, 3.68669088044678e+01, 3.65371482517081e+01, 3.61876078349182e+01, 3.58203489904619e+01, 3.54377286168088e+01, 3.50423472973752e+01, 3.46369929797784e+01, 3.42245819520172e+01, 3.38080987570748e+01, 3.33905365633726e+01, 3.29748395027598e+01, 3.25638485687254e+01, 3.21602527337512e+01, 3.17665468658950e+01, 3.13849977057871e+01, 3.10176186021990e+01, 3.06661529984799e+01, 3.03320659807265e+01, 3.00165427023073e+01, 2.97204922774988e+01, 2.94445557829408e+01, 2.91891172388553e+01, 2.89543167545330e+01, 2.87400653220870e+01, 2.85460609743647e+01, 2.83718061707978e+01, 2.82166263481980e+01, 2.80796895922215e+01, 2.79600273695167e+01, 2.78565562251764e+01, 2.77681003037066e+01, 2.76934144985600e+01, 2.76312079775095e+01, 2.75801677707114e+01, 2.75389820481428e+01, 2.75063626574121e+01, 2.74810664470277e+01, 2.74619148698226e+01, 2.74478113518068e+01, 2.74377559275843e+01, 2.74308566872030e+01, 2.74263376512450e+01, 2.74235427890619e+01, 2.74219360149374e+01, 2.74210971323274e+01, 2.74207138395430e+01, 2.74205700531235e+01, 2.74205309397251e+01, 2.74205251666693e+01, 2.74205247332841e+01, 2.74205171052146e+01, 2.74204654239266e+01, 2.74202754381238e+01, 2.74197690063324e+01, 2.74186606958166e+01, 2.74165381603630e+01, 2.74128470902277e+01, 2.74068814384179e+01, 2.73977795235181e+01, 2.73845264867981e+01, 2.73659634353351e+01, 2.73408034280751e+01, 2.73076542540640e+01, 2.72650477101524e+01, 2.72114748123688e+01, 2.71454260796029e+01, 2.70654357254502e+01, 2.69701283055345e+01, 2.68582661197402e+01, 2.67287954900842e+01, 2.65808899520558e+01, 2.64139884299957e+01, 2.62278266237693e+01, 2.60224601080230e+01, 2.57982780146464e+01, 2.55560065995687e+01, 2.52967024475321e+01, 2.50217355079500e+01, 2.47327625583808e+01, 2.44316920517682e+01, 2.41206416233303e+01, 2.38018898190386e+01, 2.34778238576729e+01, 2.31508854347907e+01, 2.28235166880141e+01, 2.24981084329089e+01, 2.21769526213814e+01, 2.18622006661050e+01, 2.15558288390637e+01, 2.12596114384257e+01, 2.09751018881182e+01, 2.07036214511038e+01, 2.04462548502865e+01, 2.02038518294225e+01, 1.99770335570830e+01, 1.97662027669133e+01, 1.95715566114831e+01, 1.93931013534517e+01, 1.92306681955689e+01, 1.90839297339032e+01, 1.89524166870331e+01, 1.88355346951919e+01, 1.87325810909579e+01, 1.86427616151461e+01, 1.85652070894043e+01, 1.84989900640304e+01, 1.84431414403718e+01, 1.83966670273407e+01, 1.83585639371357e+01, 1.83278366626918e+01, 1.83035126154452e+01, 1.82846568434584e+01, 1.82703856032524e+01, 1.82598784294974e+01, 1.82523883394433e+01, 1.82472498263393e+01, 1.82438843387843e+01, 1.82418030094699e+01, 1.82406064835345e+01, 1.82399817983730e+01, 1.82396963766659e+01, 1.82395893054699e+01, 1.82395601795406e+01, 1.82395558806157e+01, 1.82395555588754e+01, 1.82395499089549e+01, 1.82395116299317e+01, 1.82393709124793e+01, 1.82389958136136e+01, 1.82381749300740e+01, 1.82366028826319e+01, 1.82338692126915e+01, 1.82294512344388e+01, 1.82227113172784e+01, 1.82128989900781e+01, 1.81991581552339e+01, 1.81805395708309e+01, 1.81560185983120e+01, 1.81245180186282e+01, 1.80849354932977e+01, 1.80361749947410e+01, 1.79771812651145e+01, 1.79069761029484e+01, 1.78246950454888e+01, 1.77296228381885e+01, 1.76212259879820e+01, 1.74991807072611e+01, 1.73633946869981e+01, 1.72140213953559e+01, 1.70514659735293e+01, 1.68763822693478e+01, 1.66896610729879e+01, 1.64924101491028e+01, 1.62859271426735e+01, 1.60716668226256e+01, 1.58512043805460e+01, 1.56261966031584e+01, 1.53983426897246e+01, 1.51693463124245e+01, 1.49408802561620e+01, 1.47145546664043e+01, 1.44918896179476e+01, 1.42742924219253e+01, 1.40630398283266e+01, 1.38592650615405e+01, 1.36639494457547e+01, 1.34779182330913e+01, 1.33018401397612e+01, 1.31362300261026e+01, 1.29814541273166e+01, 1.28377372529499e+01, 1.27051714205923e+01, 1.25837254645551e+01, 1.24732552523523e+01, 1.23735142387578e+01, 1.22841641785341e+01, 1.22047858966618e+01, 1.21348900741607e+01, 1.20739280463080e+01, 1.20213026283527e+01, 1.19763789833852e+01, 1.19384955305388e+01, 1.19069748625180e+01, 1.18811346033632e+01, 1.18602980944761e+01, 1.18438047535929e+01, 1.18310199119659e+01, 1.18213439037200e+01, 1.18142201619284e+01, 1.18091420713380e+01, 1.18056583397247e+01, 1.18033766790433e+01, 1.18019656328977e+01, 1.18011544459815e+01, 1.18007309404130e+01, 1.18005374388436e+01, 1.18004648499870e+01, 1.18004451040828e+01, 1.18004421896296e+01, 1.18004419714243e+01, 1.18004381385402e+01, 1.18004121702008e+01, 1.18003167080517e+01, 1.18000622426528e+01, 1.17995053596815e+01, 1.17984388940907e+01, 1.17965844013603e+01, 1.17935873162200e+01, 1.17890151233834e+01, 1.17823588088581e+01, 1.17730377902077e+01, 1.17604084354733e+01, 1.17437761697272e+01, 1.17224110336233e+01, 1.16955664008765e+01, 1.16625003861322e+01, 1.16224992902859e+01, 1.15749022503754e+01, 1.15191261026786e+01, 1.14546893496892e+01, 1.13812340631363e+01, 1.12985445721833e+01, 1.12065618885720e+01, 1.11053930104668e+01, 1.09953145155346e+01, 1.08767701820474e+01, 1.07503627357532e+01, 1.06168401748317e+01, 1.04770774389058e+01, 1.03320544286727e+01, 1.01828315282068e+01, 1.00305238244696e+01, 9.87627516534620e+00, 9.72123306866271e+00, 9.56652531743153e+00, 9.41323887894572e+00, 9.26240159005467e+00, 9.11496687255421e+00, 8.97180158752131e+00, 8.83367700631155e+00, 8.70126276706341e+00, 8.57512359732371e+00, 8.45571851557540e+00, 8.34340217791939e+00, 8.23842801194391e+00, 8.14095277791246e+00, 8.05104221607399e+00, 7.96867747457848e+00, 7.89376206017990e+00, 7.82612910794457e+00, 7.76554882092627e+00, 7.71173598126558e+00, 7.66435747653190e+00, 7.62303981670216e+00, 7.58737663656632e+00, 7.55693618531780e+00, 7.53126880039218e+00, 7.50991434786927e+00, 7.49240958926322e+00, 7.47829540712628e+00, 7.46712379277471e+00, 7.45846447194850e+00, 7.45191102157820e+00, 7.44708631595617e+00, 7.44364713577434e+00, 7.44128778015420e+00, 7.43974254040293e+00, 7.43878692416141e+00, 7.43823755822073e+00, 7.43795074505874e+00, 7.43781969898197e+00, 7.43777053927269e+00, 7.43775716665851e+00, 7.43775519288936e+00, 7.43775504472116e+00, 7.43775243683667e+00, 7.43773476804620e+00, 7.43766981586363e+00, 7.43749667842268e+00, 7.43711777761237e+00, 7.43639216116094e+00, 7.43513038294742e+00, 7.43309121497400e+00, 7.42998041384485e+00, 7.42545172710221e+00, 7.41911027754091e+00, 7.41051840368233e+00, 7.39920395945533e+00, 7.38467098455284e+00, 7.36641254958192e+00, 7.34392546021819e+00, 7.31672637819116e+00, 7.28436879299318e+00, 7.24646016799545e+00, 7.20267850079246e+00, 7.15278749260261e+00, 7.09664952610589e+00, 7.03423571223051e+00, 6.96563238595655e+00, 6.89104360485148e+00, 6.81078942107558e+00, 6.72529994159393e+00, 6.63510544178698e+00, 6.54082303218500e+00, 6.44314057508241e+00, 6.34279868925654e+00, 6.24057175474553e+00, 6.13724883100491e+00, 6.03361533429454e+00, 5.93043619496731e+00, 5.82844104937131e+00, 5.72831183444322e+00, 5.63067296599306e+00, 5.53608411168378e+00, 5.44503542954412e+00, 5.35794503951028e+00, 5.27515843036838e+00, 5.19694947442801e+00, 5.12352272120977e+00, 5.05501666198091e+00, 4.99150769178303e+00, 4.93301453827593e+00, 4.87950297223655e+00, 4.83089065922660e+00, 4.78705205324702e+00, 4.74782326943529e+00, 4.71300690291760e+00, 4.68237678406707e+00, 4.65568267621164e+00, 4.63265493013400e+00, 4.61300911067654e+00, 4.59645060493342e+00, 4.58267920978670e+00, 4.57139368022881e+00, 4.56229620066645e+00, 4.55509672115136e+00, 4.54951708132764e+00, 4.54529482890200e+00, 4.54218662854748e+00, 4.53997115287375e+00, 4.53845135044421e+00, 4.53745599714189e+00, 4.53684045610663e+00, 4.53648659692854e+00, 4.53630185512903e+00, 4.53621744612113e+00, 4.53618578155739e+00, 4.53617716804428e+00, 4.53617589670836e+00, 4.53617580086444e+00, 4.53617410851636e+00, 4.53616264261819e+00, 4.53612049289548e+00, 4.53600813822392e+00, 4.53576225791784e+00, 4.53529138742998e+00, 4.53447260138876e+00, 4.53314939108810e+00, 4.53113088470523e+00, 4.52819253503351e+00, 4.52407836927929e+00, 4.51850485626345e+00, 4.51116639631867e+00, 4.50174237729819e+00, 4.48990566683679e+00, 4.47533232862844e+00, 4.45771226342330e+00, 4.43676039023683e+00, 4.41222790813973e+00, 4.38391312313767e+00, 4.35167129698554e+00, 4.31542298264503e+00, 4.27516035880727e+00, 4.23095116366213e+00, 4.18293995139534e+00, 4.13134654468402e+00, 4.07646172013109e+00, 4.01864032658079e+00, 3.95829218411399e+00, 3.89587123174293e+00, 3.83186347539018e+00, 3.76677432981305e+00, 3.70111594808041e+00, 3.63539509295079e+00, 3.57010203171770e+00, 3.50570083740479e+00, 3.44262136356708e+00, 3.38125303720102e+00, 3.32194049446653e+00, 3.26498097664322e+00, 3.21062331695869e+00, 3.15906828802228e+00, 3.11047004661187e+00, 3.06493840611122e+00, 3.02254168271158e+00, 2.98330989348464e+00, 2.94723812593594e+00, 2.91428994348298e+00, 2.88440073453302e+00, 2.85748095104570e+00, 2.83341921371849e+00, 2.81208528447998e+00, 2.79333292290463e+00, 2.77700265204815e+00, 2.76292446187402e+00, 2.75092047580529e+00, 2.74080759894757e+00, 2.73240015614765e+00, 2.72551251530358e+00, 2.71996167730266e+00, 2.71556979979614e+00, 2.71216660892198e+00, 2.70959164225005e+00, 2.70769625873841e+00, 2.70634534824499e+00, 2.70541867472382e+00, 2.70481179385462e+00, 2.70443649728200e+00, 2.70422075121863e+00, 2.70410811586878e+00, 2.70405665265504e+00, 2.70403734715755e+00, 2.70403209560649e+00, 2.70403132048898e+00, 2.70403126173616e+00, 2.70403022009985e+00, 2.70402316286737e+00, 2.70399721983478e+00, 2.70392806600771e+00, 2.70377672846998e+00, 2.70348691400605e+00, 2.70298297123571e+00, 2.70216859194657e+00, 2.70092634539400e+00, 2.69911812542654e+00, 2.69658657188559e+00, 2.69315750323044e+00, 2.68864336538962e+00, 2.68284766167932e+00, 2.67557028048999e+00, 2.66661358287766e+00, 2.65578905429187e+00, 2.64292426804010e+00, 2.62786985861754e+00, 2.61050616729919e+00, 2.59074920683264e+00, 2.56855560191406e+00, 2.54392620033204e+00, 2.51690811598009e+00, 2.48759505543100e+00, 2.45612588687038e+00, 2.42268152350206e+00, 2.38748030121674e+00, 2.35077212092368e+00, 2.31283169032010e+00, 2.27395123266376e+00, 2.23443303051233e+00, 2.19458214397051e+00, 2.15469959259816e+00, 2.11507622633067e+00, 2.07598744205031e+00, 2.03768883594331e+00, 2.00041282245834e+00, 1.96436620129992e+00, 1.92972861536475e+00, 1.89665181456428e+00, 1.86525962215285e+00, 1.83564849036098e+00, 1.80788852966548e+00, 1.78202489977292e+00, 1.75807945919370e+00, 1.73605258291508e+00, 1.71592507286657e+00, 1.69766010231345e+00, 1.68120515177975e+00, 1.66649390949331e+00, 1.65344812274938e+00, 1.64197939733484e+00, 1.63199094980804e+00, 1.62337932179972e+00, 1.61603606662452e+00, 1.60984941660665e+00, 1.60470593504969e+00, 1.60049215029535e+00, 1.59709616152561e+00, 1.59440919767107e+00, 1.59232710283902e+00, 1.59075171491276e+00, 1.58959209915625e+00, 1.58876559640885e+00, 1.58819864617558e+00, 1.58782734876564e+00, 1.58759773747173e+00, 1.58746574121049e+00, 1.58739682942254e+00, 1.58736534354978e+00, 1.58735353219153e+00, 1.58735031922321e+00, 1.58734984499608e+00, 1.58734980882225e+00, 1.58734916448245e+00, 1.58734479898976e+00, 1.58732875104060e+00, 1.58728597358933e+00, 1.58719235866617e+00, 1.58701308440195e+00, 1.58670135472023e+00, 1.58619759659970e+00, 1.58542917312798e+00, 1.58431066204911e+00, 1.58274473822522e+00, 1.58062368341835e+00, 1.57783152721943e+00, 1.57424679834475e+00, 1.56974583602955e+00, 1.56420657779742e+00, 1.55751270430872e+00, 1.54955798702260e+00, 1.54025065351456e+00, 1.52951756234621e+00, 1.51730796817370e+00, 1.50359666143711e+00, 1.48838628736954e+00, 1.47170868635753e+00, 1.45362515004159e+00, 1.43422555118602e+00, 1.41362637488766e+00, 1.39196774778750e+00, 1.36940962406761e+00, 1.34612733632171e+00, 1.32230675144977e+00, 1.29813928408404e+00, 1.27381701244907e+00, 1.24952811588322e+00, 1.22545281320123e+00, 1.20175993159475e+00, 1.17860418232101e+00, 1.15612416733207e+00, 1.13444109473880e+00, 1.11365814384681e+00, 1.09386039422719e+00, 1.07511521824469e+00, 1.05747303178188e+00, 1.04096830183354e+00, 1.02562071997663e+00, 1.01143646512583e+00, 9.98409495311279e-01, 9.86522824688677e-01, 9.75749757281875e-01, 9.66055062194400e-01, 9.57396085727481e-01, 9.49723803841991e-01, 9.42983823750069e-01, 9.37117346303268e-01, 9.32062101528146e-01, 9.27753268449919e-01, 9.24124387569719e-01, 9.21108270367384e-01, 9.18637905355880e-01, 9.16647354911376e-01, 9.15072631760935e-01, 9.13852539065104e-01, 9.12929453915453e-01, 9.12250031174281e-01, 9.11765803241963e-01, 9.11433651764554e-01, 9.11216129570718e-01, 9.11081615174766e-01, 9.11004287765209e-01, 9.10963917342031e-01, 9.10945472094113e-01, 9.10938552698497e-01, 9.10936670460188e-01, 9.10936392645937e-01, 9.10936371304814e-01, 9.10935989210117e-01, 9.10933400465183e-01, 9.10923884008483e-01, 9.10898516962845e-01, 9.10843003498726e-01, 9.10736695067754e-01, 9.10551844132769e-01, 9.10253130417502e-01, 9.09797494928995e-01, 9.09134312097223e-01, 9.08205923709801e-01, 9.06948549242799e-01, 9.05293575135843e-01, 9.03169210206883e-01, 9.00502475827778e-01, 8.97221478302697e-01, 8.93257888344806e-01, 8.88549530501066e-01, 8.83042966207256e-01, 8.76695940522112e-01, 8.69479557063713e-01, 8.61380050318303e-01, 8.52400040419894e-01, 8.42559182517149e-01, 8.31894159260003e-01, 8.20458007643211e-01, 8.08318816220719e-01, 7.95557870817532e-01, 7.82267361759381e-01, 7.68547789659337e-01, 7.54505217770912e-01, 7.40248516445760e-01, 7.25886730692281e-01, 7.11526678026104e-01, 6.97270854399179e-01, 6.83215694871706e-01, 6.69450206325905e-01, 6.56054964542250e-01, 6.43101448923247e-01, 6.30651675553480e-01, 6.18758082735155e-01, 6.07463621651902e-01, 5.96802007111208e-01, 5.86798088093772e-01, 5.77468303945551e-01, 5.68821198588836e-01, 5.60857971476381e-01, 5.53573049786877e-01, 5.46954671359629e-01, 5.40985472018860e-01, 5.35643074227949e-01, 5.30900676461982e-01, 5.26727644314036e-01, 5.23090105180970e-01, 5.19951548433525e-01, 5.17273432306595e-01, 5.15015797411863e-01, 5.13137884880706e-01, 5.11598754830317e-01, 5.10357898293989e-01, 5.09375833187505e-01, 5.08614672532701e-01, 5.08038651276578e-01, 5.07614596850189e-01, 5.07312328292755e-01, 5.07104969441841e-01, 5.06969163397613e-01, 5.06885178160005e-01, 5.06836896867782e-01, 5.06811690227281e-01, 5.06800173229536e-01, 5.06795852826201e-01, 5.06794677573652e-01, 5.06794504108910e-01, 5.06794490682012e-01, 5.06794248960887e-01, 5.06792611266007e-01, 5.06786590945694e-01, 5.06770543130244e-01, 5.06735423630460e-01, 5.06668168467385e-01, 5.06551220186878e-01, 5.06362224864901e-01, 5.06073921026641e-01, 5.05654238340625e-01, 5.05066619596292e-01, 5.04270573795296e-01, 5.03222460904396e-01, 5.01876499778746e-01, 5.00185980041067e-01, 4.98104646654568e-01, 4.95588213246918e-01, 4.92595947957369e-01, 4.89092264970367e-01, 4.85048247381542e-01, 4.80443023952674e-01, 4.75264924683913e-01, 4.69512348466225e-01, 4.63194290176261e-01, 4.56330493480266e-01, 4.48951217691317e-01, 4.41096630160047e-01, 4.32815857617143e-01, 4.24165748553804e-01, 4.15209412551121e-01, 4.06014610577628e-01, 3.96652072535474e-01, 3.87193815288825e-01, 3.77711527054686e-01, 3.68275073629832e-01, 3.58951169730233e-01, 3.49802245874997e-01, 3.40885528666476e-01, 3.32252340687276e-01, 3.23947616022664e-01, 3.16009618928149e-01, 3.08469846564975e-01, 3.01353092092810e-01, 2.94677641703956e-01, 2.88455578298894e-01, 2.82693165231294e-01, 2.77391285627814e-01, 2.72545915884180e-01, 2.68148615705073e-01, 2.64187021131193e-01, 2.60645331051322e-01, 2.57504781435145e-01, 2.54744104711634e-01, 2.52339974182775e-01, 2.50267435004403e-01, 2.48500324047029e-01, 2.47011680898297e-01, 2.45774151476141e-01, 2.44760384324991e-01, 2.43943417850033e-01, 2.43297054722660e-01, 2.42796217689175e-01, 2.42417279263634e-01, 2.42138356491893e-01, 2.41939561302132e-01, 2.41803197019280e-01, 2.41713892459842e-01, 2.41658666611386e-01, 2.41626919133068e-01, 2.41610344639682e-01, 2.41602771742649e-01, 2.41599930907910e-01, 2.41599158133894e-01, 2.41599044074150e-01, 2.41599035173849e-01, 2.41598874019687e-01, 2.41597782177593e-01, 2.41593768468803e-01, 2.41583069535324e-01, 2.41559655893710e-01, 2.41514818699969e-01, 2.41436854990326e-01, 2.41310867856868e-01, 2.41118696436516e-01, 2.40838988546358e-01, 2.40447426041356e-01, 2.39917109243279e-01, 2.39219101868699e-01, 2.38323131648514e-01, 2.37198434306833e-01, 2.35814719986408e-01, 2.34143232031755e-01, 2.32157858957379e-01, 2.29836252314740e-01, 2.27160897004893e-01, 2.24120077340743e-01, 2.20708682649587e-01, 2.16928800948025e-01, 2.12790058315008e-01, 2.08309674648291e-01, 2.03512222646935e-01, 1.98429094825030e-01, 1.93097701575837e-01, 1.87560440121345e-01, 1.81863488067805e-01, 1.76055485008774e-01, 1.70186170386387e-01, 1.64305045366035e-01, 1.58460121074739e-01, 1.52696805947656e-01, 1.47056972242581e-01, 1.41578227360057e-01, 1.36293400841950e-01, 1.31230244095534e-01, 1.26411328049713e-01, 1.21854114811939e-01, 1.17571173309969e-01, 1.13570505882223e-01, 1.09855952541949e-01, 1.06427641695745e-01, 1.03282459850405e-01, 1.00414517666422e-01, 9.78155950354846e-02, 9.54755531806484e-02, 9.33827067254214e-02, 9.15241529923350e-02, 8.98860593169815e-02, 8.84539118295464e-02, 8.72127309690143e-02, 8.61472599885502e-02, 8.52421329726241e-02, 8.44820285026771e-02, 8.38518141957110e-02, 8.33366860176028e-02, 8.29223046732466e-02, 8.25949296395407e-02, 8.23415496811748e-02, 8.21500071144499e-02, 8.20091118011703e-02, 8.19087399676590e-02, 8.18399125414544e-02, 8.17948478120627e-02, 8.17669838486381e-02, 8.17509671861814e-02, 8.17426057265322e-02, 8.17387854621927e-02, 8.17373523743186e-02, 8.17369625420383e-02, 8.17369050036945e-02, 8.17369004757040e-02, 8.17368179990297e-02, 8.17362592086113e-02, 8.17342050563212e-02, 8.17287295962359e-02, 8.17167474763636e-02, 8.16938033125609e-02, 8.16539126980729e-02, 8.15894643072994e-02, 8.14911921415131e-02, 8.13482262540930e-02, 8.11482290570069e-02, 8.08776223083214e-02, 8.05219068968057e-02, 8.00660734428931e-02, 7.94950965177890e-02, 7.87944991208605e-02, 7.79509673245542e-02, 7.69529882787765e-02, 7.57914788231773e-02, 7.44603676379825e-02, 7.29570920183704e-02, 7.12829716782011e-02, 6.94434268824011e-02, 6.74480166695872e-02, 6.53102844738385e-02, 6.30474121327040e-02, 6.06796977333914e-02, 5.82298864422199e-02, 5.57223948259349e-02, 5.31824768968505e-02, 5.06353833258544e-02, 4.81055636937526e-02, 4.56159556293159e-02, 4.31873951110475e-02, 4.08381703748972e-02, 3.85837292391926e-02, 3.64365376390623e-02, 3.44060769196801e-02, 3.24989597493186e-02, 3.07191397359219e-02, 2.90681879216237e-02, 2.75456099246441e-02, 2.61491800333147e-02, 2.48752723771872e-02, 2.37191737633253e-02, 2.26753673262294e-02, 2.17377803854788e-02, 2.08999935700689e-02, 2.01554112253461e-02, 1.94973953330759e-02, 1.89193667002456e-02, 1.84148780891320e-02, 1.79776643841478e-02, 1.76016749253165e-02, 1.72810928853266e-02, 1.70103461055220e-02, 1.67841131982982e-02, 1.65973280170988e-02, 1.64451848264378e-02, 1.63231457055559e-02, 1.62269509233397e-02, 1.61526322626457e-02, 1.60965285860186e-02, 1.60553023621333e-02, 1.60259554468922e-02, 1.60058421646170e-02, 1.59926776820924e-02, 1.59845398073912e-02, 1.59798626649632e-02, 1.59774211601729e-02, 1.59763057043421e-02, 1.59758872725888e-02, 1.59757734503785e-02, 1.59757566504908e-02, 1.59757553167026e-02, 1.59757308735067e-02, 1.59755652679934e-02, 1.59749564958673e-02, 1.59733338189056e-02, 1.59697830664811e-02, 1.59629846229450e-02, 1.59511673040242e-02, 1.59320814146611e-02, 1.59029943426674e-02, 1.58607119784477e-02, 1.58016289442884e-02, 1.57218099326002e-02, 1.56171032776911e-02, 1.54832861404088e-02, 1.53162383683849e-02, 1.51121393014222e-02, 1.48676787366213e-02, 1.45802702828348e-02, 1.42482528547703e-02, 1.38710645500026e-02, 1.34493730630680e-02, 1.29851484365702e-02, 1.24816674223922e-02, 1.19434438592947e-02, 1.13760857804692e-02, 1.07860867238116e-02, 1.01805650497795e-02, 9.56697009429774e-03, 8.95277697942731e-03, 8.34519245044874e-03, 7.75089216536173e-03, 7.17580578398724e-03, 6.62496065669416e-03, 6.10238873498758e-03, 5.61109537387949e-03, 5.15308367286593e-03, 4.72942437731660e-03, 4.34035933095345e-03, 3.98542598172469e-03, 3.66359125531801e-03, 3.37338486149008e-03, 3.11302437145855e-03, 2.88052681182194e-03, 2.67380375372284e-03, 2.49073875640704e-03, 2.32924747413676e-03, 2.18732175942193e-03, 2.06305973344456e-03, 1.95468410595367e-03, 1.86055107941741e-03, 1.77915203902046e-03, 1.70910998832596e-03, 1.64917240910256e-03, 1.59820193759685e-03, 1.55516598255190e-03, 1.51912616807692e-03, 1.48922827087006e-03, 1.46469313273201e-03, 1.44480886523013e-03, 1.42892451640364e-03, 1.41644524050832e-03, 1.40682889535526e-03, 1.39958389192429e-03, 1.39426803626463e-03, 1.39048804069830e-03, 1.38789934358385e-03, 1.38620586616867e-03, 1.38515935479688e-03, 1.38455800427550e-03, 1.38424412912741e-03, 1.38410073613215e-03, 1.38404694759231e-03, 1.38403231608208e-03, 1.38403015651534e-03, 1.38402998348397e-03, 1.38402679286287e-03, 1.38400517609227e-03, 1.38392571304280e-03, 1.38371391255987e-03, 1.38325048804688e-03, 1.38236334133524e-03, 1.38082173256748e-03, 1.37833315923785e-03, 1.37454350928579e-03, 1.36904106965119e-03, 1.36136493729566e-03, 1.35101826245432e-03, 1.33748652754662e-03, 1.32026071410962e-03, 1.29886473342822e-03, 1.27288592301511e-03, 1.24200679869091e-03, 1.20603568454669e-03, 1.16493343480784e-03, 1.11883331834406e-03, 1.06805135702359e-03, 1.01308503960240e-03, 9.54599357740449e-04, 8.93400422853574e-04, 8.30398340383454e-04, 7.66562317636499e-04, 7.02871948760730e-04, 6.40269065755206e-04, 5.79614330533218e-04, 5.21651957711811e-04, 4.66984911000168e-04, 4.16061752687824e-04, 3.69174915348935e-04, 3.26468908966149e-04, 2.87956479809282e-04, 2.53540229891769e-04, 2.23036262084975e-04, 1.96197311041146e-04, 1.72734968728605e-04, 1.52339432546000e-04, 1.34694303255856e-04, 1.19488843395743e-04, 1.06429961348908e-04, 9.52481741478550e-05, 8.56962288651175e-05, 7.75505535131060e-05, 7.06167566164595e-05, 6.47278256735717e-05, 5.97355852249848e-05, 5.55077216830568e-05, 5.19313529991905e-05, 4.89138067791828e-05, 4.63780675801218e-05, 4.42577443493519e-05, 4.24945831412060e-05, 4.10376857331585e-05, 3.98429253039012e-05, 3.88720644224350e-05, 3.80917592224250e-05, 3.74727135152852e-05, 3.69890707361265e-05, 3.66179875612715e-05, 3.63393084654717e-05, 3.61353047821922e-05, 3.59904737621264e-05, 3.58913940203820e-05, 3.58266231472726e-05, 3.57866143616548e-05, 3.57636302794617e-05, 3.57516354512200e-05, 3.57461560014372e-05, 3.57441006570980e-05, 3.57435415736518e-05, 3.57434590654293e-05, 3.57434523294842e-05, 3.57433270323781e-05, 3.57424780048632e-05, 3.57393571022158e-05, 3.57310396736921e-05, 3.57128459995693e-05, 3.56780365393661e-05, 3.56176066276248e-05, 3.55202104884711e-05, 3.53722455344041e-05, 3.51581237651999e-05, 3.48607494657927e-05, 3.44622100527658e-05, 3.39446523867082e-05, 3.32911967575455e-05, 3.24865041534395e-05, 3.15166095532504e-05, 3.03688791092327e-05, 2.90359409831353e-05, 2.75288849991284e-05, 2.58950654483669e-05, 2.42035175546107e-05, 2.24633165970545e-05, 2.05897919300924e-05, 1.85271829025169e-05, 1.64660367261149e-05, 1.48253808021082e-05, 1.32812728992349e-05, 1.14255821443976e-05, 9.49607709799238e-06, 8.67535985094037e-06, 7.64297276134585e-06, 6.00656910876296e-06, 5.08308633584420e-06, 4.63554614782424e-06, 3.25764000308482e-06, 5.83652933983103e-06, 9.98439917054839e-06, 1.34075949205226e-05, 1.54380659012077e-05, 1.58867792663386e-05, 1.49559494957430e-05, 1.31082845342689e-05, 1.08724996863430e-05, 8.66189004966600e-06, 6.67943375276536e-06, 4.94327083799817e-06, 3.39664686066499e-06, 2.02960643491679e-06, 1.12475188984718e-06, 1.18291123385461e-06, 1.14680329943440e-06, 7.81044207087689e-07, 1.42108547152020e-14, 7.27959502455633e-07, 9.53212563103989e-07, 9.78148733565831e-07, 8.71070909624905e-07, 6.90852061779879e-07, 4.85648200765354e-07, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 1.42108547152020e-14, 1.42108547152020e-14, 1.42108547152020e-14, 0.00000000000000e+00, 0.00000000000000e+00, 1.42108547152020e-14, 0.00000000000000e+00],
[ 1.18000000000000e+02, 1.17999999842583e+02, 1.17999994974926e+02, 1.17999961995780e+02, 1.17999840760058e+02, 1.17999517583294e+02, 1.17998810284141e+02, 1.17997455588664e+02, 1.17995099345841e+02, 1.17991289914626e+02, 1.17985474982869e+02, 1.17977001972344e+02, 1.17965122077985e+02, 1.17948997889868e+02, 1.17927714459801e+02, 1.17900293605636e+02, 1.17865711197989e+02, 1.17822917143926e+02, 1.17770857763461e+02, 1.17708500235211e+02, 1.17634858751686e+02, 1.17549021956623e+02, 1.17450181125116e+02, 1.17337658390640e+02, 1.17210934133778e+02, 1.17069672453485e+02, 1.16913743483654e+02, 1.16743241241779e+02, 1.16558495743675e+02, 1.16360078313195e+02, 1.16148799357824e+02, 1.15925698340027e+02, 1.15692026195387e+02, 1.15449220961841e+02, 1.15198877819605e+02, 1.14942715042899e+02, 1.14682537501747e+02, 1.14420199324426e+02, 1.14157567164535e+02, 1.13896485255982e+02, 1.13638743136108e+02, 1.13386046619442e+02, 1.13139992347959e+02, 1.12902046047977e+02, 1.12673524493371e+02, 1.12455581101714e+02, 1.12249195059557e+02, 1.12055163868114e+02, 1.11874099205942e+02, 1.11706426009114e+02, 1.11552384665146e+02, 1.11412036201691e+02, 1.11285270325005e+02, 1.11171816128456e+02, 1.11071255250656e+02, 1.10983037218871e+02, 1.10906496668810e+02, 1.10840872088715e+02, 1.10785325695910e+02, 1.10738964019250e+02, 1.10700858733143e+02, 1.10670067269745e+02, 1.10645652727495e+02, 1.10626702597963e+02, 1.10612345850600e+02, 1.10601767947281e+02, 1.10594223405909e+02, 1.10589045594271e+02, 1.10585653510602e+02, 1.10583555393588e+02, 1.10582349099063e+02, 1.10581719279778e+02, 1.10581431504475e+02, 1.10581323549107e+02, 1.10581294182542e+02, 1.10581289848096e+02, 1.10581289279508e+02, 1.10581276032091e+02, 1.10581186278884e+02, 1.10580856336474e+02, 1.10579976830161e+02, 1.10578052047100e+02, 1.10574365838074e+02, 1.10567955429451e+02, 1.10557594342746e+02, 1.10541785428273e+02, 1.10518764796819e+02, 1.10486517171902e+02, 1.10442802877582e+02, 1.10385196316975e+02, 1.10311135383434e+02, 1.10217980785658e+02, 1.10103083775933e+02, 1.09963860275448e+02, 1.09797868932458e+02, 1.09602890277827e+02, 1.09377003910194e+02, 1.09118660591893e+02, 1.08826746282850e+02, 1.08500635458799e+02, 1.08140231483919e+02, 1.07745992244130e+02, 1.07318939628619e+02, 1.06860651799608e+02, 1.06373237677495e+02, 1.05859293949578e+02, 1.05321846364272e+02, 1.04764278941607e+02, 1.04190256385411e+02, 1.03603645547657e+02, 1.03008440733975e+02, 1.02408695273151e+02, 1.01808459208078e+02, 1.01211721380177e+02, 1.00622354075488e+02, 1.00044059422606e+02, 9.94803180819156e+01, 9.89343417522027e+01, 9.84090313515766e+01, 9.79069424763953e+01, 9.74302591562177e+01, 9.69807762637879e+01, 9.65598903831285e+01, 9.61685985632711e+01, 9.58075041917319e+01, 9.54768291706510e+01, 9.51764316164922e+01, 9.49058283828183e+01, 9.46642217889057e+01, 9.44505300025378e+01, 9.42634205624859e+01, 9.41013465324102e+01, 9.39625847558797e+01, 9.38452756376300e+01, 9.37474638163810e+01, 9.36671390277617e+01, 9.36022763907748e+01, 9.35508752964868e+01, 9.35109960416266e+01, 9.34807933400209e+01, 9.34585458673906e+01, 9.34426810540797e+01, 9.34317944374090e+01, 9.34246630193575e+01, 9.34202522421523e+01, 9.34177163874083e+01, 9.34163924148926e+01, 9.34157874745657e+01, 9.34155605395370e+01, 9.34154988076099e+01, 9.34154896961032e+01, 9.34154888917306e+01, 9.34154731304458e+01, 9.34153663454166e+01, 9.34149737922689e+01, 9.34139273890541e+01, 9.34116373579193e+01, 9.34072516617113e+01, 9.33996248596723e+01, 9.33872978136779e+01, 9.33684894525597e+01, 9.33411015410609e+01, 9.33027370936589e+01, 9.32507327131531e+01, 9.31822047129308e+01, 9.30941083983978e+01, 9.29833093430040e+01, 9.28466649133293e+01, 9.26811137028455e+01, 9.24837699635933e+01, 9.22520196271632e+01, 9.19836141351384e+01, 9.16767581091709e+01, 9.13301869297488e+01, 9.09432305922715e+01, 9.05158607756567e+01, 9.00487188650059e+01, 8.95431236505754e+01, 8.90010584813712e+01, 8.84251386660280e+01, 8.78185607786935e+01, 8.71850361779419e+01, 8.65287114823286e+01, 8.58540790302270e+01, 8.51658805727443e+01, 8.44690076591392e+01, 8.37684023378395e+01, 8.30689617867916e+01, 8.23754501573508e+01, 8.16924202040329e+01, 8.10241462721436e+01, 8.03745691542407e+01, 7.97472524574553e+01, 7.91453496038206e+01, 7.85715804205740e+01, 7.80282163543732e+01, 7.75170735136114e+01, 7.70395128931977e+01, 7.65964472176254e+01, 7.61883538561791e+01, 7.58152932497137e+01, 7.54769322727062e+01, 7.51725719550308e+01, 7.49011790076289e+01, 7.46614206266182e+01, 7.44517020786772e+01, 7.42702065847796e+01, 7.41149370114702e+01, 7.39837588458827e+01, 7.38744438746161e+01, 7.37847139135619e+01, 7.37122838550211e+01, 7.36549032210162e+01, 7.36103953492662e+01, 7.35766933020401e+01, 7.35518715875986e+01, 7.35341728261990e+01, 7.35220285813483e+01, 7.35140737120444e+01, 7.35091537790984e+01, 7.35063252506175e+01, 7.35048484877142e+01, 7.35041737388642e+01, 7.35039206166799e+01, 7.35038517612522e+01, 7.35038415983333e+01, 7.35038407831415e+01, 7.35038257387181e+01, 7.35037238105397e+01, 7.35033491118269e+01, 7.35023503030368e+01, 7.35001644364987e+01, 7.34959782485076e+01, 7.34886984572781e+01, 7.34769324459075e+01, 7.34589805008880e+01, 7.34328405323240e+01, 7.33962259130437e+01, 7.33465967310264e+01, 7.32812043459052e+01, 7.31971486736093e+01, 7.30914470989992e+01, 7.29611133499907e+01, 7.28032440845377e+01, 7.26151103809548e+01, 7.23942508279073e+01, 7.21385625328946e+01, 7.18463861561173e+01, 7.15165810719147e+01, 7.11485869911740e+01, 7.07424688563521e+01, 7.02989425368525e+01, 6.98193797757865e+01, 6.93057919170878e+01, 6.87607931003632e+01, 6.81875447555152e+01, 6.75896842512393e+01, 6.69712413396751e+01, 6.63365464998026e+01, 6.56901353620647e+01, 6.50366531053892e+01, 6.43807621326768e+01, 6.37270555829108e+01, 6.30799784774065e+01, 6.24437576499027e+01, 6.18223411405167e+01, 6.12193474346458e+01, 6.06380247361059e+01, 6.00812202969579e+01, 5.95513596243307e+01, 5.90504351333812e+01, 5.85800035470434e+01, 5.81411911123445e+01, 5.77347055590321e+01, 5.73608536920310e+01, 5.70195635766541e+01, 5.67104104137756e+01, 5.64326453721399e+01, 5.61852268112324e+01, 5.59668534656747e+01, 5.57759992573879e+01, 5.56109494506183e+01, 5.54698378694685e+01, 5.53506848637553e+01, 5.52514356450063e+01, 5.51699985298904e+01, 5.51042825340950e+01, 5.50522336671076e+01, 5.50118691991734e+01, 5.49813091170065e+01, 5.49588039642662e+01, 5.49427582836873e+01, 5.49317489441594e+01, 5.49245377484254e+01, 5.49200778718698e+01, 5.49175138729259e+01, 5.49161752306961e+01, 5.49155635931576e+01, 5.49153341466638e+01, 5.49152717316412e+01, 5.49152625193142e+01, 5.49152618087709e+01, 5.49152490496943e+01, 5.49151626050862e+01, 5.49148448257606e+01, 5.49139977442739e+01, 5.49121439348840e+01, 5.49085936977317e+01, 5.49024198981808e+01, 5.48924416455471e+01, 5.48772178241936e+01, 5.48550512876898e+01, 5.48240042846483e+01, 5.47819253937605e+01, 5.47264878981871e+01, 5.46552391230874e+01, 5.45656597990785e+01, 5.44552320123798e+01, 5.43215137833283e+01, 5.41622178131937e+01, 5.39752914976980e+01, 5.37589949722920e+01, 5.35119737766767e+01, 5.32333227447769e+01, 5.29226379673413e+01, 5.25800541418794e+01, 5.22062652964530e+01, 5.18025276999719e+01, 5.13706446794303e+01, 5.09129339701910e+01, 5.04321790507453e+01, 4.99315666023840e+01, 4.94146127630194e+01, 4.88850812183632e+01, 4.83468964119977e+01, 4.78040552708887e+01, 4.72605408256685e+01, 4.67202409267374e+01, 4.61868748864202e+01, 4.56639303045930e+01, 4.51546115914425e+01, 4.46618008607336e+01, 4.41880310330944e+01, 4.37354702670830e+01, 4.33059163083473e+01, 4.29007990552392e+01, 4.25211895788208e+01, 4.21678139652612e+01, 4.18410706065072e+01, 4.15410498835048e+01, 4.12675555066981e+01, 4.10201270592291e+01, 4.07980635058001e+01, 4.06004475770941e+01, 4.04261710196905e+01, 4.02739607244719e+01, 4.01424057246740e+01, 4.00299849997221e+01, 3.99350959431655e+01, 3.98560832612074e+01, 3.97912679703943e+01, 3.97389760663728e+01, 3.96975663476026e+01, 3.96654568058918e+01, 3.96411489467119e+01, 3.96232493827972e+01, 3.96104880594300e+01, 3.96017325218486e+01, 3.95959977245215e+01, 3.95924510057958e+01, 3.95904120039661e+01, 3.95893474639328e+01, 3.95888610674103e+01, 3.95886786033812e+01, 3.95886289687506e+01, 3.95886216427834e+01, 3.95886210880833e+01, 3.95886112616264e+01, 3.95885446859507e+01, 3.95882999469597e+01, 3.95876475644727e+01, 3.95862198505253e+01, 3.95834856426412e+01, 3.95787309420556e+01, 3.95710463765966e+01, 3.95593222787070e+01, 3.95422520150423e+01, 3.95183440173303e+01, 3.94859427363191e+01, 3.94432584648262e+01, 3.93884056491998e+01, 3.93194489342668e+01, 3.92344557768667e+01, 3.91315540388549e+01, 3.90089925632727e+01, 3.88652023866797e+01, 3.86988559904308e+01, 3.85089218870455e+01, 3.82947119099024e+01, 3.80559188447381e+01, 3.77926425066066e+01, 3.75054029954590e+01, 3.71951405990728e+01, 3.68632025732885e+01, 3.65113177260758e+01, 3.61415602813430e+01, 3.57563048461030e+01, 3.53581744410040e+01, 3.49499835214417e+01, 3.45346777990258e+01, 3.41152725706546e+01, 3.36947912494946e+01, 3.32762058800791e+01, 3.28623815395049e+01, 3.24560265494343e+01, 3.20596502185314e+01, 3.16755293410450e+01, 3.13056839490611e+01, 3.09518620123471e+01, 3.06155320971504e+01, 3.02978825829620e+01, 2.99998259408427e+01, 2.97220067424426e+01, 2.94648123804845e+01, 2.92283858211921e+01, 2.90126399944014e+01, 2.88172736221616e+01, 2.86417883930224e+01, 2.84855074279261e+01, 2.83475949796283e+01, 2.82270772807045e+01, 2.81228644173715e+01, 2.80337730628200e+01, 2.79585498558084e+01, 2.78958951580212e+01, 2.78444868679648e+01, 2.78030039124142e+01, 2.77701489828640e+01, 2.77446700396117e+01, 2.77253800761101e+01, 2.77111746268211e+01, 2.77010465176062e+01, 2.76940974014396e+01, 2.76895456944021e+01, 2.76867306254633e+01, 2.76851122341223e+01, 2.76842672861806e+01, 2.76838812220737e+01, 2.76837363960319e+01, 2.76836969998307e+01, 2.76836911850337e+01, 2.76836907485150e+01, 2.76836830652920e+01, 2.76836310103309e+01, 2.76834396508650e+01, 2.76829295573818e+01, 2.76818132332914e+01, 2.76796753506857e+01, 2.76759575911102e+01, 2.76699488004435e+01, 2.76607810622373e+01, 2.76474321705283e+01, 2.76287348362062e+01, 2.76033927852754e+01, 2.75700036987317e+01, 2.75270887010306e+01, 2.74731278303342e+01, 2.74066006276769e+01, 2.73260306790289e+01, 2.72300326552896e+01, 2.71173601465269e+01, 2.69869524062350e+01, 2.68379780348406e+01, 2.66698736581795e+01, 2.64823758040281e+01, 2.62755444417718e+01, 2.60497770066619e+01, 2.58058121500765e+01, 2.55447229062650e+01, 2.52678994138545e+01, 2.49770217568716e+01, 2.46740238867192e+01, 2.43610499520656e+01, 2.40404046949412e+01, 2.37144998553786e+01, 2.33857987369086e+01, 2.30567611851190e+01, 2.27297911870653e+01, 2.24071890917889e+01, 2.20911100882987e+01, 2.17835300920856e+01, 2.14862196416968e+01, 2.12007258600721e+01, 2.09283620537298e+01, 2.06702041526954e+01, 2.04270929587954e+01, 2.01996410700554e+01, 1.99882433664449e+01, 1.97930900478042e+01, 1.96141813753553e+01, 1.94513434527644e+01, 1.93042445661993e+01, 1.91724117674431e+01, 1.90552475191770e+01, 1.89520463221003e+01, 1.88620113088908e+01, 1.87842708221484e+01, 1.87178949960964e+01, 1.86619123396187e+01, 1.86153262765215e+01, 1.85771315435377e+01, 1.85463302837878e+01, 1.85219476096863e+01, 1.85030463512288e+01, 1.84887406595087e+01, 1.84782081067647e+01, 1.84706999176130e+01, 1.84655489840506e+01, 1.84621753600029e+01, 1.84600889981433e+01, 1.84588895788383e+01, 1.84582633830317e+01, 1.84579772710880e+01, 1.84578699409588e+01, 1.84578407445932e+01, 1.84578364352720e+01, 1.84578361127536e+01, 1.84578304491697e+01, 1.84577920775748e+01, 1.84576510198180e+01, 1.84572750138197e+01, 1.84564521450105e+01, 1.84548762954176e+01, 1.84521360131187e+01, 1.84477073464852e+01, 1.84409511188651e+01, 1.84311150358452e+01, 1.84173409133336e+01, 1.83986771844760e+01, 1.83740966819233e+01, 1.83425194967588e+01, 1.83028404879769e+01, 1.82539607634742e+01, 1.81948221876971e+01, 1.81244437109426e+01, 1.80419580844465e+01, 1.79466473504984e+01, 1.78379754047552e+01, 1.77156159419661e+01, 1.75794742321368e+01, 1.74297014359303e+01, 1.72667005457228e+01, 1.70911235067313e+01, 1.69038595916311e+01, 1.67060156229276e+01, 1.64988891078705e+01, 1.62839357238406e+01, 1.60627328340592e+01, 1.58369408093105e+01, 1.56082638872401e+01, 1.53784121409061e+01, 1.51490658881332e+01, 1.49218435907260e+01, 1.46982739999842e+01, 1.44797730240572e+01, 1.42676255347244e+01, 1.40629721002491e+01, 1.38668004290021e+01, 1.36799411392238e+01, 1.35030673399492e+01, 1.33366974238765e+01, 1.31812004393292e+01, 1.30368034244957e+01, 1.29036001455240e+01, 1.27815607684670e+01, 1.26705420988883e+01, 1.25702981281293e+01, 1.24804907206276e+01, 1.24007003548004e+01, 1.23304368871110e+01, 1.22691503441626e+01, 1.22162417621882e+01, 1.21710740894905e+01, 1.21329831482906e+01, 1.21012886214910e+01, 1.20753049907942e+01, 1.20543523093886e+01, 1.20377666491999e+01, 1.20249100237238e+01, 1.20151795567865e+01, 1.20080156488636e+01, 1.20029088887025e+01, 1.19994054707251e+01, 1.19971109084994e+01, 1.19956918804802e+01, 1.19948761038171e+01, 1.19944502017078e+01, 1.19942556050753e+01, 1.19941826054124e+01, 1.19941627477579e+01, 1.19941598168106e+01, 1.19941595973705e+01, 1.19941557427944e+01, 1.19941296274884e+01, 1.19940336250689e+01, 1.19937777194624e+01, 1.19932176844029e+01, 1.19921451813076e+01, 1.19902801865557e+01, 1.19872661199624e+01, 1.19826679998746e+01, 1.19759738937592e+01, 1.19665998623860e+01, 1.19538985064535e+01, 1.19371711132303e+01, 1.19156832656893e+01, 1.18886836187858e+01, 1.18554253718102e+01, 1.18151897813137e+01, 1.17673108794519e+01, 1.17112004048328e+01, 1.16463718362583e+01, 1.15724623630958e+01, 1.14892516453097e+01, 1.13966763212703e+01, 1.12948394134105e+01, 1.11840140511468e+01, 1.10646412569092e+01, 1.09373218950835e+01, 1.08028032299898e+01, 1.06619608421640e+01, 1.05157768822534e+01, 1.03653157798333e+01, 1.02116985655315e+01, 1.00560769181619e+01, 9.89960793381715e+00, 9.74343045562537e+00, 9.58864362459737e+00, 9.43628813175062e+00, 9.28733048060633e+00, 9.14265041250519e+00, 9.00303150697212e+00, 8.86915484686153e+00, 8.74159553524393e+00, 8.62082177083168e+00, 8.50719613399511e+00, 8.40097870765132e+00, 8.30233165591204e+00, 8.21132490535184e+00, 8.12794261406987e+00, 8.05209016610674e+00, 7.98360148667805e+00, 7.92224653101844e+00, 7.86773885136982e+00, 7.81974318909496e+00, 7.77788306972299e+00, 7.74174839689574e+00, 7.71090304675165e+00, 7.68489245825504e+00, 7.66325119912767e+00, 7.64551046374702e+00, 7.63120543150151e+00, 7.61988238480447e+00, 7.61110545855531e+00, 7.60446287049201e+00, 7.59957246743971e+00, 7.59608641817142e+00, 7.59369489086694e+00, 7.59212857240551e+00, 7.59115991731599e+00, 7.59060305446728e+00, 7.59031232699263e+00, 7.59017949237671e+00, 7.59012966171535e+00, 7.59011610658510e+00, 7.59011410587691e+00, 7.59011395568643e+00, 7.59011131220817e+00, 7.59009340226430e+00, 7.59002756357113e+00, 7.58985206297947e+00, 7.58946799027383e+00, 7.58873246826585e+00, 7.58745346183234e+00, 7.58538644209537e+00, 7.58223313066162e+00, 7.57764251057806e+00, 7.57121424261192e+00, 7.56250456465816e+00, 7.55103467605414e+00, 7.53630151609180e+00, 7.51779073776440e+00, 7.49499155700505e+00, 7.46741303049517e+00, 7.43460119053337e+00, 7.39615635482761e+00, 7.35174984513591e+00, 7.30113930408729e+00, 7.24418180505295e+00, 7.18084401270600e+00, 7.11120877371619e+00, 7.03547769334126e+00, 6.95396947339592e+00, 6.86711403322417e+00, 6.77544268667086e+00, 6.67957488162646e+00, 6.58020220255904e+00, 6.47807047258651e+00, 6.37396085860819e+00, 6.26867087745444e+00, 6.16299612808398e+00, 6.05771344727012e+00, 5.95356602201778e+00, 5.85125081151965e+00, 5.75140845428760e+00, 5.65461567808023e+00, 5.56138010197735e+00, 5.47213722623674e+00, 5.38724934611737e+00, 5.30700609670258e+00, 5.23162633098424e+00, 5.16126104669968e+00, 5.09599710286756e+00, 5.03586149995557e+00, 4.98082603455896e+00, 4.93081217769128e+00, 4.88569606317502e+00, 4.84531350739267e+00, 4.80946501224044e+00, 4.77792072814481e+00, 4.75042537236280e+00, 4.72670310877446e+00, 4.70646239875337e+00, 4.68940082875645e+00, 4.67520990983277e+00, 4.66357982865232e+00, 4.65420411065095e+00, 4.64678413556483e+00, 4.64103342621181e+00, 4.63668161510485e+00, 4.63347798236532e+00, 4.63119445406120e+00, 4.62962795358253e+00, 4.62860201033745e+00, 4.62796754953277e+00, 4.62760281298678e+00, 4.62741239211115e+00, 4.62732538826137e+00, 4.62729275027883e+00, 4.62728387197181e+00, 4.62728256155283e+00, 4.62728246276248e+00, 4.62728071838875e+00, 4.62726890000876e+00, 4.62722545452415e+00, 4.62710964581441e+00, 4.62685620636672e+00, 4.62637085897514e+00, 4.62552689671790e+00, 4.62416299293214e+00, 4.62208239329200e+00, 4.61905361774448e+00, 4.61481276885849e+00, 4.60906750289520e+00, 4.60150266874389e+00, 4.59178755681968e+00, 4.57958462553268e+00, 4.56455948929330e+00, 4.54639186362134e+00, 4.52478707623756e+00, 4.49948767620330e+00, 4.47028461530856e+00, 4.43702744588475e+00, 4.39963298439879e+00, 4.35809193510904e+00, 4.31247305336788e+00, 4.26292455002240e+00, 4.20967258866466e+00, 4.15301689456041e+00, 4.09332366417350e+00, 4.03101612303239e+00, 3.96656321409404e+00, 3.90046699793251e+00, 3.83324940239332e+00, 3.76543896881887e+00, 3.69755820450148e+00, 3.63011207067679e+00, 3.56357802025856e+00, 3.49839786149968e+00, 3.43497157732472e+00, 3.37365309051339e+00, 3.31474784592115e+00, 3.25851199241733e+00, 3.20515289382362e+00, 3.15483067899731e+00, 3.10766055096586e+00, 3.06371560566043e+00, 3.02302995359690e+00, 2.98560198503954e+00, 2.95139766486511e+00, 2.92035378377409e+00, 2.89238112580976e+00, 2.86736753784770e+00, 2.84518090513208e+00, 2.82567204875700e+00, 2.80867756700555e+00, 2.79402264341272e+00, 2.78152384098642e+00, 2.77099189487144e+00, 2.76223450558285e+00, 2.75505912258824e+00, 2.74927569444129e+00, 2.74469934795682e+00, 2.74115294625577e+00, 2.73846946510098e+00, 2.73649411990507e+00, 2.73508617303681e+00, 2.73412035318648e+00, 2.73348782578499e+00, 2.73309666557142e+00, 2.73287179870280e+00, 2.73275440124233e+00, 2.73270076212879e+00, 2.73268064036776e+00, 2.73267516677241e+00, 2.73267435888162e+00, 2.73267429764462e+00, 2.73267321196612e+00, 2.73266585634097e+00, 2.73263881638100e+00, 2.73256673851374e+00, 2.73240900161034e+00, 2.73210693092430e+00, 2.73158167238033e+00, 2.73073283549418e+00, 2.72943800150753e+00, 2.72755317854580e+00, 2.72491426626850e+00, 2.72133956700759e+00, 2.71663334743860e+00, 2.71059041367718e+00, 2.70300161361530e+00, 2.69366012491422e+00, 2.68236832850255e+00, 2.66894501036523e+00, 2.65323258476997e+00, 2.63510399643611e+00, 2.61446894383249e+00, 2.59127907583203e+00, 2.56553185200347e+00, 2.53727282230734e+00, 2.50659617067957e+00, 2.47364347132994e+00, 2.43860071648247e+00, 2.40169377868210e+00, 2.36318255940350e+00, 2.32335414064808e+00, 2.28251529315462e+00, 2.24098470327559e+00, 2.19908526337315e+00, 2.15713673296766e+00, 2.11544902606217e+00, 2.07431632021103e+00, 2.03401212030587e+00, 1.99478534893251e+00, 1.95685747869109e+00, 1.92042067242511e+00, 1.88563685660806e+00, 1.85263762234901e+00, 1.82152482818483e+00, 1.79237176893588e+00, 1.76522477466042e+00, 1.74010511176084e+00, 1.71701107272820e+00, 1.69592015975927e+00, 1.67679128841742e+00, 1.65956695869080e+00, 1.64417536058616e+00, 1.63053239854933e+00, 1.61854363267344e+00, 1.60810614438839e+00, 1.59911033998287e+00, 1.59144170703426e+00, 1.58498253697974e+00, 1.57961362219774e+00, 1.57521592878744e+00, 1.57167223754477e+00, 1.56886873634353e+00, 1.56669653817298e+00, 1.56505309137941e+00, 1.56384344304149e+00, 1.56298131354734e+00, 1.56238994078842e+00, 1.56200265610170e+00, 1.56176316105143e+00, 1.56162548390695e+00, 1.56155360656170e+00, 1.56152076578711e+00, 1.56150844617228e+00, 1.56150509494718e+00, 1.56150460031371e+00, 1.56150456258329e+00, 1.56150389051706e+00, 1.56149933717434e+00, 1.56148259867797e+00, 1.56143798055400e+00, 1.56134033773147e+00, 1.56115335096786e+00, 1.56082821359718e+00, 1.56030279693441e+00, 1.55950135595113e+00, 1.55833482816955e+00, 1.55670176709236e+00, 1.55448993581820e+00, 1.55157856586015e+00, 1.54784126007657e+00, 1.54314948715088e+00, 1.53737657920615e+00, 1.53040210585658e+00, 1.52211646021964e+00, 1.51242545887057e+00, 1.50125473262580e+00, 1.48855367254340e+00, 1.47429869912122e+00, 1.45849564453978e+00, 1.44118107828257e+00, 1.42242246373601e+00, 1.40231710339452e+00, 1.38098990716404e+00, 1.35859009482866e+00, 1.33528701253714e+00, 1.31126529734343e+00, 1.28671965814529e+00, 1.26184955286443e+00, 1.23685403013445e+00, 1.21192697151812e+00, 1.18725292199689e+00, 1.16300363829465e+00, 1.13933542319673e+00, 1.11638725570293e+00, 1.09427967673039e+00, 1.07311435158861e+00, 1.05297420514172e+00, 1.03392401326840e+00, 1.01601133337013e+00, 9.99267664824160e-01, 9.83709744622900e-01, 9.69340901240926e-01, 9.56152408652969e-01, 9.44124800559180e-01, 9.33229121008011e-01, 9.23428101021031e-01, 9.14677261226629e-01, 9.06925947916832e-01, 9.00118314551847e-01, 8.94194262864886e-01, 8.89090357701320e-01, 8.84740727916409e-01, 8.81077962396830e-01, 8.78034005905207e-01, 8.75541054329560e-01, 8.73532443423713e-01, 8.71943519649065e-01, 8.70712476694251e-01, 8.69781137071413e-01, 8.69095655259613e-01, 8.68607117506571e-01, 8.68272013831955e-01, 8.68052560074574e-01, 8.67916851922061e-01, 8.67838838511091e-01, 8.67798110017461e-01, 8.67779501183364e-01, 8.67772520424040e-01, 8.67770621493605e-01, 8.67770341215660e-01, 8.67770319685292e-01, 8.67769934202101e-01, 8.67767322499787e-01, 8.67757721651500e-01, 8.67732129665129e-01, 8.67676124005953e-01, 8.67568873277591e-01, 8.67382384664694e-01, 8.67081026667833e-01, 8.66621362916064e-01, 8.65952327932916e-01, 8.65015769011663e-01, 8.63747369145457e-01, 8.62077953663601e-01, 8.59935167507047e-01, 8.57245491017679e-01, 8.53936540360053e-01, 8.49939575548959e-01, 8.45192116481570e-01, 8.39640547873600e-01, 8.33242580354991e-01, 8.25969429868536e-01, 8.17807583042338e-01, 8.08760033442539e-01, 7.98846902208439e-01, 7.88105394604415e-01, 7.76589088063858e-01, 7.64366592844738e-01, 7.51519668490317e-01, 7.38140913258476e-01, 7.24331165991714e-01, 7.10196768652310e-01, 6.95846832954786e-01, 6.81390638011905e-01, 6.66935260850366e-01, 6.52583511949544e-01, 6.38432217510612e-01, 6.24570862278432e-01, 6.11080583721488e-01, 5.98033491393679e-01, 5.85492274485674e-01, 5.73510055263810e-01, 5.62130445170794e-01, 5.51387762554469e-01, 5.41307375113647e-01, 5.31906135245606e-01, 5.23192881861320e-01, 5.15168987474453e-01, 5.07828934223696e-01, 5.01160906847261e-01, 4.95147394437339e-01, 4.89765796027655e-01, 4.84989027674871e-01, 4.80786130633305e-01, 4.77122881444672e-01, 4.73962405233593e-01, 4.71265793201866e-01, 4.68992724285400e-01, 4.67102089258319e-01, 4.65552613378392e-01, 4.64303471158203e-01, 4.63314884251474e-01, 4.62548691021624e-01, 4.61968874377959e-01, 4.61542033165657e-01, 4.61237781980245e-01, 4.61029064874339e-01, 4.60892370081964e-01, 4.60807835548835e-01, 4.60759238591024e-01, 4.60733867180878e-01, 4.60722274906786e-01, 4.60717926266014e-01, 4.60716743332327e-01, 4.60716568733872e-01, 4.60716555219250e-01, 4.60716311918290e-01, 4.60714663519994e-01, 4.60708603853675e-01, 4.60692451163482e-01, 4.60657102184456e-01, 4.60589407681752e-01, 4.60471695826044e-01, 4.60281467547247e-01, 4.59991285339456e-01, 4.59568873641018e-01, 4.58977444560546e-01, 4.58176257030729e-01, 4.57121410153903e-01, 4.55766862412190e-01, 4.54065657614606e-01, 4.51971326305014e-01, 4.49439418549218e-01, 4.46429111576521e-01, 4.42904824950020e-01, 4.38837768183006e-01, 4.34207342341125e-01, 4.29002319202951e-01, 4.23221729516726e-01, 4.16875405623813e-01, 4.09984142329208e-01, 4.02579461829902e-01, 3.94702991754684e-01, 3.86405487751929e-01, 3.77745551569177e-01, 3.68788110624379e-01, 3.59602734716390e-01, 3.50261869461320e-01, 3.40839064583477e-01, 3.31407269071960e-01, 3.22037255458611e-01, 3.12796223187718e-01, 3.03746617335634e-01, 2.94945184807219e-01, 2.86442276471661e-01, 2.78281391243922e-01, 2.70498947461391e-01, 2.63124258468579e-01, 2.56179683360713e-01, 2.49680920413667e-01, 2.43637409731136e-01, 2.38052812799367e-01, 2.32925539550337e-01, 2.28249297722655e-01, 2.24013644268425e-01, 2.20204523806999e-01, 2.16804784244516e-01, 2.13794664316694e-01, 2.11152251707915e-01, 2.08853913381489e-01, 2.06874701731252e-01, 2.05188741123976e-01, 2.03769599398910e-01, 2.02590648037358e-01, 2.01625413181361e-01, 2.00847917659829e-01, 2.00233011912545e-01, 1.99756689427900e-01, 1.99396380269619e-01, 1.99131214686417e-01, 1.98942247848876e-01, 1.98812636569613e-01, 1.98727759479453e-01, 1.98675273534093e-01, 1.98645101812204e-01, 1.98629350175153e-01, 1.98622153286692e-01, 1.98619453512023e-01, 1.98618719110357e-01, 1.98618610714306e-01, 1.98618602255948e-01, 1.98618449103989e-01, 1.98617411478324e-01, 1.98613597079003e-01, 1.98603429463120e-01, 1.98581178758076e-01, 1.98538569343924e-01, 1.98464481692113e-01, 1.98344764481789e-01, 1.98162171797478e-01, 1.97896438430746e-01, 1.97524503781622e-01, 1.97020891292169e-01, 1.96358245492291e-01, 1.95508022444642e-01, 1.94441321662836e-01, 1.93129838677834e-01, 1.91546907830343e-01, 1.89668595308049e-01, 1.87474793870763e-01, 1.84950264179122e-01, 1.82085564224991e-01, 1.78877808953828e-01, 1.75331207357218e-01, 1.71457334224428e-01, 1.67275107972955e-01, 1.62810463568448e-01, 1.58095729064115e-01, 1.53168733976784e-01, 1.48071695687577e-01, 1.42849944550264e-01, 1.37550557994203e-01, 1.32220977742250e-01, 1.26907682102370e-01, 1.21654977570614e-01, 1.16503961733414e-01, 1.11491694128588e-01, 1.06650594988750e-01, 1.02008075309072e-01, 9.75863869171008e-02, 9.34026692885226e-02, 8.94691614153516e-02, 8.57935422933025e-02, 8.23793623310820e-02, 7.92265296499437e-02, 7.63318191155946e-02, 7.36893772490248e-02, 7.12912021817544e-02, 6.91275839406984e-02, 6.71874961010985e-02, 6.54589348980892e-02, 6.39292060492522e-02, 6.25851627328160e-02, 6.14134004063942e-02, 6.04004155140103e-02, 5.95327357183635e-02, 5.87970292245785e-02, 5.81802001434681e-02, 5.76694757884155e-02, 5.72524904203249e-02, 5.69173683626332e-02, 5.66528077222870e-02, 5.64481642888239e-02, 5.62935336687076e-02, 5.61798284518744e-02, 5.60988463059999e-02, 5.60433244191643e-02, 5.60069757010382e-02, 5.59845026023282e-02, 5.59715852748752e-02, 5.59648419830552e-02, 5.59617610811358e-02, 5.59606053549402e-02, 5.59602909718961e-02, 5.59602445697038e-02, 5.59602409180684e-02, 5.59601744041913e-02, 5.59597237642180e-02, 5.59580671848323e-02, 5.59536515118708e-02, 5.59439887291830e-02, 5.59254864462702e-02, 5.58933205268346e-02, 5.58413579036738e-02, 5.57621374518078e-02, 5.56469163499145e-02, 5.54857884235105e-02, 5.52678792908949e-02, 5.49816205083715e-02, 5.46151011966742e-02, 5.41564908168368e-02, 5.35945210215374e-02, 5.29190082103241e-02, 5.21213921525145e-02, 5.11952605698127e-02, 5.01368257487371e-02, 4.89453179158801e-02, 4.76232619261907e-02, 4.61766091335392e-02, 4.46147050511883e-02, 4.29500849821949e-02, 4.11981031392799e-02, 3.93764144584026e-02, 3.75043407304076e-02, 3.56021623027516e-02, 3.36903822235257e-02, 3.17890106326928e-02, 2.99169134331549e-02, 2.80912614329054e-02, 2.63271054191421e-02, 2.46370904763467e-02, 2.30313108037166e-02, 2.15172956128242e-02, 2.01001083403725e-02, 1.87825359153067e-02, 1.75653422930822e-02, 1.64475606685244e-02, 1.54268011539642e-02, 1.44995545380624e-02, 1.36614772633976e-02, 1.29076473386078e-02, 1.22327850938717e-02, 1.16314362599610e-02, 1.10981177034262e-02, 1.06274282671706e-02, 1.02141285756687e-02, 9.85319444043853e-03, 9.53984876294101e-03, 9.26957672220396e-03, 9.03812868675402e-03, 8.84151481348283e-03, 8.67599474933911e-03, 8.53806527513257e-03, 8.42444813650900e-03, 8.33207971139416e-03, 8.25810357804357e-03, 8.19986648451732e-03, 8.15491769809285e-03, 8.12101124635944e-03, 8.09611017335499e-03, 8.07839164259860e-03, 8.06625154119445e-03, 8.05830719173173e-03, 8.05339685584272e-03, 8.05057491385242e-03, 8.04910189974350e-03, 8.04842893455771e-03, 8.04817649294876e-03, 8.04810782373977e-03, 8.04809768833081e-03, 8.04809688368380e-03, 8.04808213703723e-03, 8.04798222699787e-03, 8.04761495555795e-03, 8.04663600945332e-03, 8.04449393831419e-03, 8.04039287430669e-03, 8.03326502635571e-03, 8.02175509282677e-03, 8.00421883113245e-03, 7.97873802952154e-03, 7.94315395481249e-03, 7.89512089843925e-03, 7.83218062613855e-03, 7.75185729226280e-03, 7.65177070995549e-03, 7.52976384853810e-03, 7.38403822975897e-03, 7.21328878250915e-03, 7.01682802521475e-03, 6.79468854400511e-03, 6.54769295192636e-03, 6.27748206244405e-03, 5.98649492188486e-03, 5.67789842108368e-03, 5.35546901708983e-03, 5.02343402250460e-03, 4.68628425352335e-03, 4.34857288527724e-03, 4.01471663478503e-03, 3.68881467100834e-03, 3.37449809437373e-03, 3.07481894130035e-03, 2.79218315510832e-03, 2.52832750936705e-03, 2.28433658237748e-03, 2.06069285161448e-03, 1.85735098663997e-03, 1.67382664101012e-03, 1.50929057124927e-03, 1.36266057893408e-03, 1.23268604073701e-03, 1.11802193778487e-03, 1.01729081769975e-03, 9.29132031799895e-04, 8.52238273921557e-04, 7.85380251855372e-04, 7.27421200608092e-04, 6.77323540500652e-04, 6.34149996258770e-04, 5.97061004668831e-04, 5.65309598982822e-04, 5.38234502144519e-04, 5.15252014736234e-04, 4.95847322582676e-04, 4.79565855698638e-04, 4.66005202198971e-04, 4.54807881723696e-04, 4.45655103050058e-04, 4.38261520670835e-04, 4.32370948757921e-04, 4.27752968789719e-04, 4.24200353037918e-04, 4.21527207819850e-04, 4.19567714146751e-04, 4.18175318472436e-04, 4.17222208355383e-04, 4.16598903612453e-04, 4.16213802409970e-04, 4.15992541618948e-04, 4.15877062437175e-04, 4.15824307921753e-04, 4.15804519334984e-04, 4.15799136476380e-04, 4.15798341990126e-04, 4.15798278339707e-04, 4.15797104665216e-04, 4.15789152796719e-04, 4.15759922020698e-04, 4.15682012246066e-04, 4.15511552702696e-04, 4.15185271378959e-04, 4.14618394103172e-04, 4.13703587781811e-04, 4.12311183708880e-04, 4.10290930418000e-04, 4.07475510826316e-04, 4.03686010116644e-04, 3.98739416169747e-04, 3.92458075396007e-04, 3.84680804287996e-04, 3.75275092522998e-04, 3.64149549227477e-04, 3.51265492795960e-04, 3.36646422525178e-04, 3.20384097307169e-04, 3.02640122029629e-04, 2.83642325797473e-04, 2.63675800312058e-04, 2.43069181209421e-04, 2.22177409227697e-04, 2.01362574927089e-04, 1.80974495677333e-04, 1.61332577505391e-04, 1.42710339233076e-04, 1.25323737563576e-04, 1.09324459117488e-04, 9.47994490019255e-05, 8.17767641621003e-05, 7.02356780379887e-05, 6.01182605066697e-05, 5.13396480101846e-05, 4.37941486097770e-05, 3.73590740565533e-05, 3.19024982786686e-05, 2.72936117937661e-05, 2.34107315719712e-05, 2.01499725420717e-05, 1.74291723880060e-05, 1.51762814226686e-05, 1.33177309749044e-05, 1.17815960010148e-05, 1.05024951579935e-05, 9.42854815377814e-06, 8.53018096336200e-06, 7.78425511782643e-06, 7.15192363998085e-06, 6.60054888612649e-06, 6.13858874487505e-06, 5.79554404112059e-06, 5.56730499567948e-06, 5.40715386421198e-06, 5.26777989762195e-06, 5.13007816493882e-06, 4.99788413321767e-06, 4.88137135334910e-06, 4.78676976456427e-06, 4.71459708251132e-06, 4.66190151371393e-06, 4.62466245210180e-06, 4.59909560390770e-06, 4.58208226916668e-06, 4.57118783003807e-06, 4.56455302355607e-06, 4.56077520993858e-06, 4.55881361460797e-06, 4.55791982290066e-06, 4.55758492989844e-06, 4.55749388095228e-06, 4.55748042327286e-06, 4.55747924377192e-06, 4.55745619376557e-06, 4.55730013015909e-06, 4.55672636689997e-06, 4.55519695208295e-06, 4.55185015368897e-06, 4.54544169770088e-06, 4.53430253344322e-06, 4.51632104159216e-06, 4.48897354488054e-06, 4.44946761035681e-06, 4.39514786876316e-06, 4.32442831765911e-06, 4.23845533248368e-06, 4.14290347805490e-06, 4.04716672619543e-06, 3.95643593265049e-06, 3.85750941234164e-06, 3.71409238653087e-06, 3.49217097550536e-06, 3.21249487456043e-06, 2.97349006928016e-06, 2.82510424653992e-06, 2.61679841173645e-06, 2.18880849445213e-06, 1.79640814224058e-06, 1.97424655823397e-06, 2.06889494336338e-06, 1.57289566971031e-06, 1.23076834768199e-06, 1.88972387604736e-06, 1.83245350626748e-06, 7.27532636801698e-07, 1.57931809496858e-06, 1.86710083482922e-06, 9.20169071605415e-07, 3.53897846139262e-06, 5.75524072132794e-06, 7.15374306992089e-06, 7.52741181031524e-06, 6.92202959839960e-06, 5.62318598440470e-06, 4.08632988069257e-06, 2.77971614082162e-06, 1.78152171770307e-06, 1.53608699804408e-06, 2.56737250481365e-06, 3.30701469408723e-06, 3.47974379621974e-06, 3.10923107349481e-06, 2.34049765879263e-06, 1.35814785551247e-06, 0.00000000000000e+00, 8.68374286255857e-07, 1.04374298359744e-06, 1.00207557807153e-06, 8.80573992390055e-07, 7.49564023294624e-07, 6.41439001469735e-07, 5.71116103742497e-07, 5.31269634507225e-07, 4.97134166721480e-07, 4.51193741923817e-07, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00, 0.00000000000000e+00]])
